i2d = {
  "n00001740": "entity",
  "n00001930": "physical entity",
  "n00002137": "abstraction, abstract entity",
  "n00002452": "thing",
  "n00002684": "object, physical object",
  "n00003553": "whole, unit",
  "n00003993": "congener",
  "n00004258": "living thing, animate thing",
  "n00004475": "organism, being",
  "n00005787": "benthos",
  "n00005930": "dwarf",
  "n00006024": "heterotroph",
  "n00006150": "parent",
  "n00006269": "life",
  "n00006400": "biont",
  "n00006484": "cell",
  "n00007347": "causal agent, cause, causal agency",
  "n00007846": "person, individual, someone, somebody, mortal, soul",
  "n00015388": "animal, animate being, beast, brute, creature, fauna",
  "n00017222": "plant, flora, plant life",
  "n00019046": "native",
  "n00019128": "natural object",
  "n00019613": "substance",
  "n00020090": "substance",
  "n00020827": "matter",
  "n00021265": "food, nutrient",
  "n00021734": "nutrient",
  "n00021939": "artifact, artefact",
  "n00022903": "article",
  "n00023100": "psychological feature",
  "n00023271": "cognition, knowledge, noesis",
  "n00023773": "motivation, motive, need",
  "n00024264": "attribute",
  "n00024720": "state",
  "n00026192": "feeling",
  "n00027167": "location",
  "n00027807": "shape, form",
  "n00028270": "time",
  "n00028651": "space, infinite",
  "n00029007": "absolute space",
  "n00029114": "phase space",
  "n00029378": "event",
  "n00029677": "process, physical process",
  "n00030358": "act, deed, human action, human activity",
  "n00031264": "group, grouping",
  "n00031921": "relation",
  "n00032613": "possession",
  "n00032823": "social relation",
  "n00033020": "communication",
  "n00033615": "measure, quantity, amount",
  "n00034213": "phenomenon",
  "n00034479": "thing",
  "n00034574": "kindness, benignity",
  "n00034777": "abdominoplasty, tummy tuck",
  "n00034939": "abort",
  "n00035189": "accomplishment, achievement",
  "n00035697": "agon",
  "n00035891": "alienation",
  "n00036061": "application",
  "n00036299": "beachhead, foothold",
  "n00036580": "cakewalk",
  "n00036762": "feat, effort, exploit",
  "n00037006": "masterpiece",
  "n00037090": "masterstroke",
  "n00037200": "credit",
  "n00037396": "action",
  "n00038175": "res gestae",
  "n00038262": "course, course of action",
  "n00038573": "blind alley",
  "n00038863": "collision course",
  "n00039021": "interaction",
  "n00039211": "interplay",
  "n00039297": "contact",
  "n00039545": "brush",
  "n00039740": "eye contact",
  "n00039916": "fetch",
  "n00039990": "placement",
  "n00040152": "interchange, reciprocation, give-and-take",
  "n00040420": "reciprocity",
  "n00040545": "cross-fertilization, cross-fertilisation",
  "n00040804": "dealings, traffic",
  "n00040962": "relation",
  "n00041188": "playing",
  "n00041468": "play, swordplay",
  "n00041614": "boondoggle",
  "n00041740": "bowling",
  "n00041899": "acquiring, getting",
  "n00042311": "causing, causation",
  "n00042541": "delivery, obstetrical delivery",
  "n00042757": "departure, going, going away, leaving",
  "n00043116": "derring-do",
  "n00043195": "discovery, find, uncovering",
  "n00043609": "disposal, disposition",
  "n00043902": "hit",
  "n00044150": "implementation, effectuation",
  "n00044455": "egress, egression, emergence",
  "n00044673": "equalization, equalisation, leveling",
  "n00044900": "exhumation, disinterment, digging up",
  "n00045114": "mitzvah, mitsvah",
  "n00045250": "propulsion, actuation",
  "n00045646": "rally, rallying",
  "n00045907": "recovery, retrieval",
  "n00046177": "running away",
  "n00046344": "stunt",
  "n00046522": "touch, touching",
  "n00047018": "tour de force",
  "n00047106": "performance",
  "n00047356": "overachievement",
  "n00047550": "underachievement",
  "n00047745": "record, track record",
  "n00047941": "fait accompli, accomplished fact",
  "n00048051": "going, sledding",
  "n00048225": "arrival, reaching",
  "n00048374": "arrival",
  "n00048656": "attainment",
  "n00048828": "advent, coming",
  "n00049003": "entrance, entering, entry, ingress, incoming",
  "n00049344": "incursion",
  "n00049530": "intrusion",
  "n00049636": "irruption",
  "n00049789": "entree",
  "n00049909": "entail",
  "n00050037": "registration, enrollment, enrolment",
  "n00050195": "appearance",
  "n00050484": "apparition",
  "n00050693": "emergence, emersion",
  "n00050887": "reappearance, return",
  "n00051077": "comeback",
  "n00051192": "return, homecoming",
  "n00051385": "repatriation",
  "n00051525": "penetration",
  "n00051712": "interpenetration, permeation",
  "n00051897": "market penetration",
  "n00052052": "anchorage",
  "n00052146": "docking, moorage, dockage, tying up",
  "n00052334": "landing",
  "n00052500": "landing",
  "n00052791": "forced landing, emergency landing",
  "n00053020": "breaking away",
  "n00053097": "farewell, leave, leave-taking, parting",
  "n00053405": "French leave",
  "n00053525": "valediction",
  "n00053609": "disappearance, disappearing",
  "n00053806": "vanishing",
  "n00053913": "withdrawal",
  "n00054328": "effacement, self-effacement",
  "n00054483": "retreat",
  "n00054652": "retirement, retreat",
  "n00054821": "evacuation",
  "n00055038": "medical evacuation, medevac, medivac",
  "n00055227": "decampment",
  "n00055315": "desertion, abandonment, defection",
  "n00055633": "abscondment, decampment",
  "n00055793": "absence without leave, unauthorized absence",
  "n00055932": "deviationism",
  "n00056087": "emigration, out-migration, expatriation",
  "n00056311": "immigration, in-migration",
  "n00056551": "aliyah",
  "n00056688": "pullback",
  "n00056912": "retreat",
  "n00057162": "standdown, stand-down",
  "n00057306": "disengagement, fallback, pullout",
  "n00057486": "receding, recession",
  "n00057657": "sailing",
  "n00057748": "amphibious landing",
  "n00058002": "debarkation, disembarkation, disembarkment",
  "n00058247": "going ashore",
  "n00058337": "boarding, embarkation, embarkment",
  "n00058519": "exit",
  "n00058608": "elopement",
  "n00058743": "escape, flight",
  "n00059127": "evasion",
  "n00059376": "slip, elusion, eluding",
  "n00059552": "maneuver, manoeuvre, evasive action",
  "n00059728": "clinch",
  "n00059895": "dodge",
  "n00059989": "break, breakout, jailbreak, gaolbreak, prisonbreak, prison-breaking",
  "n00060201": "getaway, lam",
  "n00060414": "exodus, hegira, hejira",
  "n00060548": "Hegira, Hejira",
  "n00060747": "skedaddle",
  "n00060817": "Underground Railroad, Underground Railway",
  "n00061014": "close call, close shave, squeak, squeaker, narrow escape",
  "n00061171": "surfacing",
  "n00061290": "dispatch, despatch, shipment",
  "n00061453": "reshipment",
  "n00061598": "consummation",
  "n00061792": "consummation",
  "n00061917": "realization, realisation, fruition",
  "n00062133": "orgasm, climax, sexual climax, coming",
  "n00062331": "male orgasm",
  "n00062451": "fulfillment, fulfilment",
  "n00062671": "self-fulfillment, self-realization, self-realisation",
  "n00062806": "attainment",
  "n00063014": "record",
  "n00063310": "track record",
  "n00063559": "world record",
  "n00063652": "success",
  "n00064018": "winning",
  "n00064151": "blockbuster, megahit, smash hit",
  "n00064370": "sleeper",
  "n00064504": "hit, smash, smasher, strike, bang",
  "n00064789": "bell ringer, bull's eye, mark, home run",
  "n00065098": "ennoblement",
  "n00065216": "conquest",
  "n00065336": "coup",
  "n00065418": "flying colors, flying colours",
  "n00065575": "passing, pass, qualifying",
  "n00065855": "credit, course credit",
  "n00066075": "semester hour, credit hour",
  "n00066216": "nonaccomplishment, nonachievement",
  "n00066397": "failure",
  "n00066636": "failure",
  "n00066901": "failing, flunk",
  "n00067153": "naught",
  "n00067254": "cut",
  "n00067397": "default",
  "n00067526": "loss",
  "n00067707": "capitulation, fall, surrender",
  "n00067990": "frustration, thwarting, foiling",
  "n00068170": "overturn, upset",
  "n00068333": "backsliding, lapse, lapsing, relapse, relapsing, reversion, reverting",
  "n00068617": "recidivism",
  "n00068755": "disappointment, dashing hopes",
  "n00068901": "breach",
  "n00069060": "copout",
  "n00069173": "breach of contract",
  "n00069444": "anticipatory breach, constructive breach",
  "n00069604": "breach of duty",
  "n00069685": "breach of the covenant of warranty",
  "n00069815": "breach of promise",
  "n00069918": "breach of trust",
  "n00070077": "breach of trust with fraudulent intent",
  "n00070215": "breach of warranty",
  "n00070363": "leaning",
  "n00070462": "material breach",
  "n00070641": "motivation, motivating",
  "n00070807": "partial breach",
  "n00070965": "mistake, error, fault",
  "n00071546": "double fault",
  "n00071700": "footfault",
  "n00071864": "bobble",
  "n00072068": "error, misplay",
  "n00072261": "blot, smear, smirch, spot, stain",
  "n00072473": "confusion, mix-up",
  "n00072691": "incursion",
  "n00072808": "miscalculation, misreckoning, misestimation",
  "n00073032": "backfire, boomerang",
  "n00073177": "rounding, rounding error",
  "n00073525": "truncation error",
  "n00073713": "distortion",
  "n00073828": "slip, slip-up, miscue, parapraxis",
  "n00074092": "Freudian slip",
  "n00074324": "offside",
  "n00074524": "oversight, lapse",
  "n00074624": "omission, skip",
  "n00074790": "blunder, blooper, bloomer, bungle, pratfall, foul-up, fuckup, flub, botch, boner, boo-boo",
  "n00075283": "snafu",
  "n00075471": "spectacle",
  "n00075618": "ballup, balls-up, cockup, mess-up",
  "n00075790": "bull",
  "n00075912": "fumble, muff",
  "n00076072": "fluff",
  "n00076196": "faux pas, gaffe, solecism, slip, gaucherie",
  "n00076323": "howler",
  "n00076393": "clanger",
  "n00076563": "trip, trip-up, stumble, misstep",
  "n00076884": "spill, tumble, fall",
  "n00077169": "pratfall",
  "n00077249": "wipeout",
  "n00077419": "acquisition",
  "n00077856": "obtainment, obtention",
  "n00077981": "catching, contracting",
  "n00078217": "incurring",
  "n00078393": "moneymaking",
  "n00078536": "annexation",
  "n00078821": "pork-barreling",
  "n00079018": "purchase",
  "n00079212": "redemption, repurchase, buyback",
  "n00079398": "trading",
  "n00079603": "bond trading, bond-trading activity",
  "n00079752": "program trading",
  "n00079908": "short sale, short selling",
  "n00080096": "short covering",
  "n00080242": "insider trading",
  "n00080474": "naked option",
  "n00080619": "covered option",
  "n00080743": "call option, call",
  "n00080968": "put option, put",
  "n00081174": "straddle",
  "n00081395": "incentive option, incentive stock option",
  "n00081572": "buying, purchasing",
  "n00081836": "shopping",
  "n00082081": "marketing",
  "n00082223": "mail-order buying, catalog buying",
  "n00082347": "viatication, viaticus",
  "n00082525": "acceptance",
  "n00082754": "succession, taking over",
  "n00082870": "assumption, laying claim",
  "n00083260": "assumption",
  "n00083448": "position",
  "n00083585": "inheritance, heritage",
  "n00083729": "procurement, procurance, procural",
  "n00083975": "appropriation",
  "n00084371": "borrowing, adoption",
  "n00084642": "naturalization, naturalisation",
  "n00084895": "misappropriation",
  "n00085041": "preemption, pre-emption",
  "n00085219": "seizure",
  "n00085432": "usurpation",
  "n00085678": "confiscation, arrogation",
  "n00085829": "distress, distraint",
  "n00086140": "expropriation",
  "n00086297": "impoundment, impounding, internment, poundage",
  "n00086547": "drug bust, drugs bust",
  "n00086654": "impress, impressment",
  "n00086809": "occupation, occupancy, moving in",
  "n00087073": "preoccupancy, preoccupation",
  "n00087218": "sequestration, requisition",
  "n00087423": "grant, subsidization, subsidisation",
  "n00087663": "award, awarding",
  "n00087849": "addiction",
  "n00088154": "block grant",
  "n00088367": "grant-in-aid",
  "n00088481": "capture, gaining control, seizure",
  "n00088725": "apprehension, arrest, catch, collar, pinch, taking into custody",
  "n00089027": "conquest, conquering, subjection, subjugation",
  "n00089234": "enslavement",
  "n00089351": "restitution, return, restoration, regaining",
  "n00089657": "clawback",
  "n00089891": "repossession",
  "n00090076": "foreclosure",
  "n00090253": "reception, receipt",
  "n00090393": "appointment",
  "n00090610": "comb-out",
  "n00090779": "giving",
  "n00091013": "abandonment",
  "n00091234": "discard, throwing away",
  "n00091404": "staging",
  "n00091503": "discard",
  "n00091670": "mine disposal",
  "n00091779": "minesweeping",
  "n00091893": "sewage disposal",
  "n00091977": "bait and switch",
  "n00092212": "private treaty",
  "n00092366": "auction, auction sale, vendue",
  "n00092663": "bootlegging",
  "n00092847": "bootlegging",
  "n00093006": "capitalization, capitalisation",
  "n00093127": "overcapitalization, overcapitalisation",
  "n00093354": "reclamation",
  "n00093483": "rescue, deliverance, delivery, saving",
  "n00093874": "lifesaving",
  "n00094001": "redemption, salvation",
  "n00094240": "absolution, remission, remittal, remission of sin",
  "n00094500": "indulgence",
  "n00094788": "conversion, rebirth, spiritual rebirth",
  "n00094954": "proselytism",
  "n00095121": "expiation, atonement, propitiation",
  "n00095329": "reparation, amends",
  "n00095502": "liberation, release, freeing",
  "n00095873": "jail delivery",
  "n00095971": "reclamation, reformation",
  "n00096211": "salvage",
  "n00096343": "salvage",
  "n00096513": "salvation",
  "n00096720": "search and rescue mission",
  "n00096851": "ransom",
  "n00096969": "recapture, retaking",
  "n00097122": "recapture",
  "n00097244": "invocation",
  "n00097348": "instrumentation",
  "n00097504": "performance, execution, carrying out, carrying into action",
  "n00097923": "specific performance",
  "n00098051": "linguistic performance",
  "n00098385": "mechanism, mechanics",
  "n00098625": "service",
  "n00098822": "curb service",
  "n00098939": "self-service",
  "n00099061": "valet parking",
  "n00099267": "dramatic production, dramatic performance",
  "n00099439": "encore",
  "n00099588": "extemporization, extemporisation, improvisation",
  "n00099799": "juggle, juggling",
  "n00099951": "magic trick, conjuring trick, trick, magic, legerdemain, conjuration, thaumaturgy, illusion, deception",
  "n00100253": "musical performance",
  "n00100437": "one-night stand",
  "n00100543": "rendition, rendering, interpretation",
  "n00100889": "reinterpretation",
  "n00101003": "spin",
  "n00101191": "playing",
  "n00101410": "bowing",
  "n00101570": "spiccato, spiccato bowing",
  "n00101708": "piping",
  "n00101809": "stopping",
  "n00102039": "double stopping",
  "n00102162": "transposition",
  "n00102374": "jam session",
  "n00102457": "automation, mechanization, mechanisation",
  "n00102779": "computerization, cybernation",
  "n00102927": "motorization, motorisation",
  "n00103140": "launching, launch",
  "n00103291": "launching",
  "n00103419": "rocket firing, rocket launching",
  "n00103590": "blastoff, shot",
  "n00103737": "moon shot",
  "n00103834": "drive, thrust, driving force",
  "n00104088": "firewall",
  "n00104249": "impulse, impulsion, impetus",
  "n00104409": "roll, bowl",
  "n00104539": "throw",
  "n00104976": "bowling",
  "n00105164": "fling",
  "n00105271": "heave, heaving",
  "n00105479": "hurl, cast",
  "n00105624": "leaner",
  "n00105820": "pass, toss, flip",
  "n00106104": "pitch",
  "n00106272": "pitch, delivery",
  "n00106726": "ringer",
  "n00106888": "shy",
  "n00107016": "slinging",
  "n00107137": "throw-in",
  "n00107279": "balk",
  "n00107400": "ball",
  "n00107551": "beanball, beaner",
  "n00107684": "change-up, change-of-pace, change-of-pace ball, off-speed pitch",
  "n00107875": "curve, curve ball, breaking ball, bender",
  "n00108081": "duster",
  "n00108181": "fastball, heater, smoke, hummer, bullet",
  "n00108427": "knuckleball, knuckler",
  "n00108545": "overhand pitch",
  "n00108665": "passed ball",
  "n00108818": "screwball",
  "n00108965": "sinker",
  "n00109081": "slider",
  "n00109227": "spitball, spitter",
  "n00109414": "strike",
  "n00109771": "submarine ball, submarine pitch",
  "n00109892": "wild pitch",
  "n00110057": "basketball shot",
  "n00110415": "bank shot",
  "n00110554": "dunk, dunk shot, stuff shot",
  "n00110745": "slam dunk",
  "n00110834": "finger-roll",
  "n00110964": "foul shot, free throw, penalty free throw, charity toss, charity throw, charity shot",
  "n00111222": "one-and-one",
  "n00111358": "hook shot, hook",
  "n00111503": "jumper, jump shot",
  "n00111660": "lay-up, layup",
  "n00111839": "pivot shot",
  "n00111962": "set shot",
  "n00112072": "scoop shot",
  "n00112185": "tip in",
  "n00112312": "push, pushing",
  "n00112674": "depression",
  "n00112828": "click, mouse click",
  "n00112997": "nudge, jog",
  "n00113113": "press, pressure, pressing",
  "n00113532": "impression",
  "n00113726": "shove",
  "n00113954": "bundling",
  "n00114095": "jostle, jostling",
  "n00114298": "elbowing",
  "n00114431": "pull, pulling",
  "n00114871": "drag",
  "n00115036": "draw, haul, haulage",
  "n00115292": "tow, towage",
  "n00115500": "tug, jerk",
  "n00115667": "draft, draught, drawing",
  "n00115803": "extirpation, excision, deracination",
  "n00116008": "pluck",
  "n00116122": "traction",
  "n00116376": "lift, raise, heave",
  "n00116687": "expulsion, projection, ejection, forcing out",
  "n00116989": "defenestration",
  "n00117124": "accommodation reflex",
  "n00117267": "Babinski, Babinski reflex, Babinski sign",
  "n00117578": "belch, belching, burp, burping, eructation",
  "n00117810": "belching",
  "n00117959": "blink, eye blink, blinking, wink, winking, nictitation, nictation",
  "n00118268": "blush, flush",
  "n00118445": "coughing up",
  "n00118552": "spit, spitting, expectoration",
  "n00118733": "vomit, vomiting, emesis, regurgitation, disgorgement, puking",
  "n00119079": "rumination",
  "n00119210": "hematemesis, haematemesis",
  "n00119297": "hyperemesis",
  "n00119402": "hyperemesis gravidarum",
  "n00119568": "jump, jumping",
  "n00119893": "header",
  "n00120010": "hop",
  "n00120202": "leap, leaping, spring, saltation, bound, bounce",
  "n00120515": "vault, hurdle",
  "n00120647": "jumping up and down",
  "n00120804": "lob",
  "n00120943": "centering, snap",
  "n00121166": "sending",
  "n00121366": "transmission, transmittal, transmitting",
  "n00121645": "forwarding",
  "n00121904": "referral",
  "n00122106": "remission, remitment, remit",
  "n00122338": "mailing, posting",
  "n00122530": "wheeling, rolling",
  "n00122661": "shooting, shot",
  "n00122954": "shoot",
  "n00123143": "countershot",
  "n00123234": "discharge, firing, firing off",
  "n00123430": "gun",
  "n00123652": "fire control",
  "n00123783": "gunfire, gunshot",
  "n00124008": "enfilade, enfilade fire",
  "n00124151": "snipe",
  "n00124256": "headshot",
  "n00124342": "skeet, skeet shooting, trapshooting",
  "n00124536": "shellfire",
  "n00124617": "gunfight, gunplay, shootout",
  "n00124766": "potshot",
  "n00124880": "contact, physical contact",
  "n00125126": "rub, wipe",
  "n00125315": "scuff",
  "n00125436": "tap, pat, dab",
  "n00125629": "hit, hitting, striking",
  "n00126100": "contusion",
  "n00126236": "crash, smash",
  "n00126447": "impingement, impaction",
  "n00126584": "batting",
  "n00126721": "fielding",
  "n00126886": "catching",
  "n00127021": "golfing",
  "n00127151": "pitching",
  "n00127286": "base on balls, walk, pass",
  "n00127531": "best",
  "n00127672": "worst",
  "n00127866": "fair ball",
  "n00128091": "foul ball",
  "n00128324": "snick",
  "n00128477": "bunt",
  "n00128638": "fly, fly ball",
  "n00128867": "blast",
  "n00128976": "pop fly, pop-fly, pop-up",
  "n00129089": "grounder, ground ball, groundball, hopper",
  "n00129317": "chop, chopper",
  "n00129435": "roller",
  "n00129527": "out",
  "n00129743": "force out, force-out, force play, force",
  "n00130093": "putout",
  "n00130347": "strikeout",
  "n00130512": "whiff",
  "n00130673": "fielder's choice",
  "n00130846": "sacrifice",
  "n00130987": "sacrifice fly",
  "n00131090": "base hit, safety",
  "n00131347": "daisy cutter",
  "n00131485": "header",
  "n00131608": "liner, line drive",
  "n00131791": "scorcher, screamer",
  "n00131877": "line-drive single, line single",
  "n00131991": "line-drive double, line double",
  "n00132105": "line-drive triple, line triple",
  "n00132219": "plunk, plunker",
  "n00132355": "homer, home run",
  "n00132497": "solo homer, solo blast",
  "n00132601": "single, bingle",
  "n00132756": "double, two-base hit, two-bagger, two-baser",
  "n00132982": "triple, three-base hit, three-bagger",
  "n00133160": "backhander",
  "n00133234": "clip",
  "n00133338": "knock, belt, rap, whack, whang",
  "n00133567": "thwack",
  "n00133668": "smack, smacking, slap",
  "n00133875": "smacker",
  "n00133981": "knockdown",
  "n00134099": "knockout, KO, kayo",
  "n00134246": "technical knockout, TKO",
  "n00134391": "swat",
  "n00134472": "spank",
  "n00134574": "whip, lash, whiplash",
  "n00134780": "punch, clout, poke, lick, biff, slug",
  "n00135148": "box",
  "n00135311": "dig, jab",
  "n00135504": "counterpunch, parry, counter",
  "n00135637": "haymaker, knockout punch, KO punch, Sunday punch",
  "n00135799": "hook",
  "n00135952": "jab",
  "n00136048": "rabbit punch",
  "n00136152": "sucker punch",
  "n00136230": "roundhouse",
  "n00136329": "kick, boot, kicking",
  "n00136691": "goal-kick",
  "n00136876": "goal-kick",
  "n00136984": "punt, punting",
  "n00137279": "place kick, place-kicking",
  "n00137534": "free kick",
  "n00137709": "corner kick",
  "n00137877": "dropkick",
  "n00138078": "kiss",
  "n00138221": "kiss, buss, osculation",
  "n00138441": "laying on",
  "n00138599": "smack, smooch",
  "n00138734": "smacker",
  "n00138799": "soul kiss, deep kiss, French kiss",
  "n00138956": "catch, grab, snatch, snap",
  "n00139544": "fair catch",
  "n00139758": "interception",
  "n00139919": "reception",
  "n00140112": "rebound",
  "n00140264": "shoestring catch",
  "n00140393": "mesh, meshing, interlock, interlocking",
  "n00140652": "handling, manipulation",
  "n00140900": "fingering",
  "n00141027": "grope",
  "n00141176": "audit",
  "n00141396": "autopsy, necropsy, postmortem, post-mortem, PM, postmortem examination, post-mortem examination",
  "n00141669": "check-in",
  "n00141806": "check, checkout, check-out procedure",
  "n00142140": "spot check",
  "n00142361": "checkup, medical checkup, medical examination, medical exam, medical, health check",
  "n00142665": "comparison, comparing",
  "n00143057": "fine-tooth comb, fine-toothed comb",
  "n00143251": "follow-up, followup, reexamination, review",
  "n00143536": "going-over",
  "n00143626": "once-over, look-over",
  "n00143763": "ophthalmoscopy",
  "n00143885": "palpation, tactual exploration",
  "n00144182": "ballottement",
  "n00144445": "tickle, tickling, titillation",
  "n00144632": "stroke, stroking",
  "n00144778": "caress",
  "n00145024": "tag",
  "n00145218": "joining, connection, connexion",
  "n00145779": "hit",
  "n00145929": "interconnection",
  "n00146111": "intersection",
  "n00146277": "approximation, bringing close together",
  "n00146443": "concatenation",
  "n00146572": "convergence, converging, convergency",
  "n00146856": "merging, meeting, coming together",
  "n00147091": "concourse, confluence",
  "n00147187": "encounter, coming upon",
  "n00147314": "articulation",
  "n00147454": "junction, adjunction",
  "n00147595": "fastening, attachment",
  "n00147862": "loosening, laxation",
  "n00148057": "tightening",
  "n00148242": "ligation",
  "n00148446": "tubal ligation",
  "n00148653": "bonding, soldering",
  "n00148762": "doweling",
  "n00148836": "grounding, earthing",
  "n00148978": "linkage",
  "n00149084": "tying, ligature",
  "n00149262": "untying, undoing, unfastening",
  "n00149508": "welding",
  "n00149699": "butt welding, butt-welding",
  "n00149895": "spot welding, spot-welding",
  "n00150097": "flare",
  "n00150228": "Texas leaguer",
  "n00150379": "flash welding, flash butt welding",
  "n00150591": "lick, lap",
  "n00150762": "grazing, shaving, skimming",
  "n00150932": "tracing",
  "n00151087": "detection, catching, espial, spying, spotting",
  "n00151314": "self-discovery",
  "n00151409": "breakthrough",
  "n00151497": "determination, finding",
  "n00151914": "rediscovery",
  "n00152018": "designation, identification",
  "n00152230": "Bertillon system",
  "n00152338": "fingerprinting",
  "n00152519": "genetic profiling, genetic fingerprinting",
  "n00152727": "diagnosis, diagnosing",
  "n00152990": "blood typing",
  "n00153105": "medical diagnosis",
  "n00153288": "prenatal diagnosis",
  "n00153499": "differential diagnosis",
  "n00153665": "prognosis, prospect, medical prognosis",
  "n00153809": "resolution, solving",
  "n00153961": "validation, proof, substantiation",
  "n00154233": "authentication, certification",
  "n00154433": "documentation, support",
  "n00154689": "monetization, monetisation",
  "n00154894": "probate",
  "n00155085": "demonetization, demonetisation",
  "n00155298": "falsification, falsifying, disproof, refutation, refutal",
  "n00155487": "localization, localisation, location, locating, fix",
  "n00155797": "echolocation, echo sounding",
  "n00155970": "predetermination",
  "n00156119": "rectification",
  "n00156293": "redetermination",
  "n00156390": "trigger, induction, initiation",
  "n00156625": "fomentation, instigation",
  "n00156812": "compulsion, coercion",
  "n00157081": "influence",
  "n00157318": "cross-pollination",
  "n00157463": "exposure",
  "n00157687": "overexposure",
  "n00157877": "underexposure",
  "n00157957": "impingement, encroachment, impact",
  "n00158185": "manipulation, use",
  "n00158443": "mind game",
  "n00158805": "autosuggestion, auto-suggestion, self-suggestion",
  "n00158996": "hypnotism, mesmerism, suggestion",
  "n00159177": "inducement, inducing",
  "n00159396": "corruption",
  "n00159620": "enticement, temptation",
  "n00159899": "blandishment, wheedling",
  "n00160073": "ingratiation, insinuation",
  "n00160297": "leading astray, leading off",
  "n00160415": "seduction",
  "n00160532": "seduction, conquest",
  "n00160688": "sexual conquest, score",
  "n00160922": "cuckoldry",
  "n00161044": "solicitation, allurement",
  "n00161243": "choice, selection, option, pick",
  "n00161603": "casting",
  "n00161739": "coloration, colouration",
  "n00161888": "sampling",
  "n00162055": "random sampling",
  "n00162236": "lucky dip",
  "n00162417": "stratified sampling, representative sampling, proportional sampling",
  "n00162632": "decision, determination, conclusion",
  "n00163047": "volition, willing",
  "n00163233": "intention",
  "n00163406": "about-face, volte-face, reversal, policy change",
  "n00163630": "adulteration",
  "n00163779": "appointment, assignment, designation, naming",
  "n00164152": "nomination",
  "n00164345": "assignment, assigning",
  "n00164579": "allocation, storage allocation",
  "n00164801": "call",
  "n00164999": "co-option, co-optation",
  "n00165178": "delegacy",
  "n00165298": "ordination, ordinance",
  "n00165563": "recognition",
  "n00165793": "laying on of hands",
  "n00165942": "move",
  "n00166172": "move",
  "n00166355": "chess move",
  "n00166552": "castle, castling",
  "n00166702": "capture",
  "n00166865": "en passant",
  "n00167063": "exchange",
  "n00167278": "exchange",
  "n00167446": "check",
  "n00167580": "discovered check",
  "n00167764": "checkmate, mate",
  "n00167950": "gambit",
  "n00168121": "demarche",
  "n00168237": "maneuver, manoeuvre, tactical maneuver, tactical manoeuvre",
  "n00168505": "parking",
  "n00168658": "move, relocation",
  "n00168911": "flit",
  "n00169047": "downshift",
  "n00169147": "downshift",
  "n00169305": "bank",
  "n00169522": "vertical bank",
  "n00169651": "chandelle",
  "n00169811": "loop, loop-the-loop",
  "n00170005": "inside loop",
  "n00170172": "outside loop",
  "n00170340": "roll",
  "n00170536": "barrel roll",
  "n00170643": "snap roll",
  "n00170710": "slip, sideslip",
  "n00170844": "flight maneuver, airplane maneuver",
  "n00171050": "straight-arm",
  "n00171249": "device, gimmick, twist",
  "n00171507": "mnemonic",
  "n00171618": "trick, fast one",
  "n00171882": "shtik, schtik, shtick, schtick",
  "n00172073": "feint",
  "n00172217": "juke, fake",
  "n00172347": "footwork",
  "n00172490": "ploy, gambit, stratagem",
  "n00172598": "ruse, artifice",
  "n00172710": "means, agency, way",
  "n00173172": "dint",
  "n00173283": "escape",
  "n00173487": "fast track",
  "n00173761": "instrument, tool",
  "n00174003": "road",
  "n00174127": "royal road",
  "n00174260": "stepping stone",
  "n00174412": "measure, step",
  "n00174663": "countermeasure",
  "n00174782": "bear hug",
  "n00174950": "proxy fight",
  "n00175174": "leveraged buyout",
  "n00175433": "bust-up takeover",
  "n00175605": "shark repellent, porcupine provision",
  "n00175875": "golden parachute",
  "n00176052": "greenmail",
  "n00176312": "pac-man strategy",
  "n00176449": "poison pill",
  "n00176609": "suicide pill",
  "n00176766": "safe harbor",
  "n00176994": "scorched-earth policy",
  "n00177127": "diagnostic procedure, diagnostic technique",
  "n00177448": "expedient",
  "n00177638": "backstop",
  "n00177783": "emergency procedure",
  "n00178024": "experimental procedure",
  "n00178177": "double-blind procedure, double-blind experiment, double-blind study",
  "n00178551": "makeshift, stopgap, make-do",
  "n00178700": "crutch",
  "n00178832": "improvisation, temporary expedient",
  "n00178972": "pis aller, last resort",
  "n00179125": "desperate measure",
  "n00179275": "open sesame",
  "n00179380": "salvation",
  "n00179576": "tooth",
  "n00179683": "voice",
  "n00179916": "wings",
  "n00180054": "casting lots, drawing lots, sortition",
  "n00180228": "resolution",
  "n00180413": "adoption, acceptance, acceptation, espousal",
  "n00180770": "embrace, bosom",
  "n00180962": "election",
  "n00181191": "co-option, co-optation",
  "n00181373": "reelection",
  "n00181495": "plebiscite",
  "n00181640": "referendum",
  "n00181781": "election",
  "n00182213": "vote",
  "n00182435": "general election",
  "n00182571": "primary, primary election",
  "n00182723": "direct primary",
  "n00182897": "closed primary",
  "n00183084": "open primary",
  "n00183240": "by-election, bye-election",
  "n00183357": "runoff",
  "n00183505": "vote, ballot, voting, balloting",
  "n00183995": "block vote",
  "n00184135": "cumulative vote",
  "n00184362": "secret ballot",
  "n00184501": "split ticket",
  "n00184652": "straight ticket",
  "n00184802": "multiple voting",
  "n00184981": "casting vote",
  "n00185104": "reconciliation, balancing",
  "n00185307": "equation, equating",
  "n00185438": "breech delivery, breech birth, breech presentation",
  "n00185612": "frank breech, frank breech delivery",
  "n00185778": "cesarean delivery, caesarean delivery, caesarian delivery, cesarean section, cesarian section, caesarean section, caesarian section, C-section, cesarean, cesarian, caesarean, caesarian, abdominal delivery",
  "n00186251": "forceps delivery",
  "n00186549": "midwifery",
  "n00186634": "score",
  "n00187056": "bowling score",
  "n00187144": "bull's eye",
  "n00187337": "goal",
  "n00187499": "own goal",
  "n00187710": "strike, ten-strike",
  "n00187890": "spare",
  "n00188007": "open frame, break",
  "n00188183": "audible",
  "n00188341": "football score",
  "n00188449": "touchback",
  "n00188620": "safety",
  "n00188766": "touchdown",
  "n00188934": "field goal",
  "n00189106": "conversion",
  "n00189257": "point after, point after touchdown, extra point",
  "n00189476": "baseball score",
  "n00189565": "run, tally",
  "n00189896": "earned run",
  "n00190040": "unearned run",
  "n00190180": "run batted in, rbi",
  "n00190338": "basketball score",
  "n00190431": "basket, field goal",
  "n00190579": "hat trick",
  "n00190783": "solution",
  "n00190931": "Russian roulette",
  "n00191142": "change",
  "n00191980": "filtration",
  "n00192149": "percolation",
  "n00192300": "reduction, simplification",
  "n00192457": "schematization, schematisation",
  "n00192613": "economy, saving",
  "n00192910": "retrenchment, curtailment, downsizing",
  "n00193099": "economy of scale",
  "n00193225": "accommodation",
  "n00193406": "adaptation",
  "n00193622": "dark adaptation",
  "n00193811": "light adaptation",
  "n00194005": "take-up",
  "n00194127": "readjustment",
  "n00194257": "domestication",
  "n00194414": "decimalization, decimalisation",
  "n00194645": "metrification, metrication",
  "n00194831": "habituation",
  "n00194969": "variation, variance",
  "n00195194": "variation",
  "n00195415": "turning",
  "n00195569": "diversification, variegation",
  "n00195938": "flux",
  "n00196084": "switch, switching, shift",
  "n00196402": "switcheroo",
  "n00196485": "substitution, exchange, commutation",
  "n00196848": "novation",
  "n00197136": "pitching change",
  "n00197229": "superannuation",
  "n00197419": "supersedure, supersession",
  "n00197610": "supplanting, displacement",
  "n00197772": "replacement, replacing",
  "n00198118": "subrogation",
  "n00198270": "weaning, ablactation",
  "n00198451": "promotion",
  "n00198631": "preferment",
  "n00198793": "demotion",
  "n00198919": "investment, investiture",
  "n00199130": "change of state",
  "n00199707": "alteration, modification, adjustment",
  "n00200041": "distraction, misdirection",
  "n00200244": "aeration",
  "n00200379": "modulation",
  "n00200563": "qualification",
  "n00200768": "reorganization",
  "n00201058": "passage, transition",
  "n00201269": "fossilization, fossilisation",
  "n00201397": "segue",
  "n00201516": "meddling, tampering",
  "n00201671": "transfer, transference",
  "n00201923": "prohibition, inhibition, forbiddance",
  "n00202284": "resistance, opposition",
  "n00202620": "lockout",
  "n00202816": "reaction",
  "n00203020": "backlash",
  "n00203208": "whitelash, white backlash",
  "n00203342": "rejection",
  "n00203649": "brush-off",
  "n00203753": "avoidance, turning away, shunning, dodging",
  "n00204022": "aversion, averting",
  "n00204199": "escape",
  "n00204334": "near thing",
  "n00204439": "abandonment, forsaking, desertion",
  "n00204659": "exposure",
  "n00204814": "apostasy, tergiversation",
  "n00204943": "bolt",
  "n00205079": "renunciation, forgoing, forswearing",
  "n00205349": "nonacceptance, turndown",
  "n00205543": "forsaking, giving up",
  "n00205649": "abnegation, self-abnegation, denial, self-denial, self-renunciation",
  "n00205891": "forfeit, forfeiture, sacrifice",
  "n00206130": "boycott",
  "n00206302": "banishment, proscription",
  "n00206600": "anathematization, anathematisation",
  "n00206779": "disbarment",
  "n00206927": "ejection, exclusion, expulsion, riddance",
  "n00207306": "deportation",
  "n00207434": "ostracism",
  "n00207622": "barring, blackball",
  "n00207761": "exile, deportation, expatriation, transportation",
  "n00208141": "Babylonian Captivity",
  "n00208277": "excommunication, excision",
  "n00208521": "relegation",
  "n00208694": "rustication",
  "n00208797": "ouster, ousting",
  "n00208943": "deposition, dethronement",
  "n00209132": "suspension, temporary removal",
  "n00209301": "rustication",
  "n00209446": "displacement",
  "n00209546": "veto",
  "n00209680": "pocket veto",
  "n00209789": "write-in",
  "n00209943": "termination, ending, conclusion",
  "n00210518": "finish, finishing",
  "n00210797": "finale, close, closing curtain, finis",
  "n00210940": "release, tone ending",
  "n00211110": "completion, culmination, closing, windup, mop up",
  "n00211462": "finalization, finalisation",
  "n00211593": "follow-through",
  "n00211776": "follow-through",
  "n00212065": "graduation",
  "n00212205": "retirement",
  "n00212377": "hibernation",
  "n00212551": "rustication",
  "n00212678": "swan song, last hurrah",
  "n00212808": "relinquishment, relinquishing",
  "n00213052": "cession, ceding",
  "n00213186": "handover",
  "n00213343": "surrender",
  "n00213482": "extradition",
  "n00213694": "release, waiver, discharge",
  "n00213903": "exemption, immunity, granting immunity",
  "n00214148": "fix",
  "n00214315": "official immunity",
  "n00214526": "sovereign immunity",
  "n00214794": "transactional immunity",
  "n00215013": "use immunity, testimonial immunity",
  "n00215314": "dissolution, breakup",
  "n00215534": "splitsville",
  "n00215683": "overthrow",
  "n00215838": "subversion, subversive activity",
  "n00216038": "adjournment, dissolution",
  "n00216174": "dismissal, dismission, discharge, firing, liberation, release, sack, sacking",
  "n00216607": "conge, congee",
  "n00216723": "removal",
  "n00216834": "purge",
  "n00217014": "destruction, devastation",
  "n00217499": "disaster",
  "n00217593": "kill",
  "n00217773": "laying waste, ruin, ruining, ruination, wrecking",
  "n00218045": "razing, leveling, tearing down, demolishing",
  "n00218208": "annihilation, obliteration",
  "n00218427": "decimation",
  "n00218602": "atomization, atomisation",
  "n00218753": "pulverization, pulverisation",
  "n00218902": "vaporization, vaporisation",
  "n00219012": "killing, kill, putting to death",
  "n00219463": "deathblow, coup de grace",
  "n00219575": "death",
  "n00219738": "drive-by killing",
  "n00219856": "euthanasia, mercy killing",
  "n00220023": "homicide",
  "n00220219": "honor killing",
  "n00220409": "manslaughter",
  "n00220522": "murder, slaying, execution",
  "n00221056": "assassination",
  "n00221178": "bloodshed, gore",
  "n00221343": "chance-medley",
  "n00221480": "contract killing",
  "n00221596": "parricide",
  "n00221727": "mariticide",
  "n00221819": "matricide",
  "n00221900": "patricide",
  "n00221981": "fratricide",
  "n00222064": "uxoricide",
  "n00222155": "filicide",
  "n00222248": "dispatch, despatch",
  "n00222376": "fell",
  "n00222485": "suicide, self-destruction, self-annihilation",
  "n00222766": "self-destruction",
  "n00222939": "assisted suicide",
  "n00223144": "physician-assisted suicide",
  "n00223268": "felo-de-se",
  "n00223362": "harakiri, hara-kiri, harikari, seppuku",
  "n00223575": "suttee",
  "n00223720": "elimination, liquidation",
  "n00223854": "slaughter",
  "n00223983": "slaughter, massacre, mass murder, carnage, butchery",
  "n00224260": "bloodbath, bloodletting, bloodshed, battue",
  "n00224599": "lynching",
  "n00224738": "poisoning",
  "n00224936": "gassing",
  "n00225070": "regicide",
  "n00225150": "shooting",
  "n00225361": "drive-by shooting",
  "n00225484": "wing shooting",
  "n00225593": "suffocation, asphyxiation",
  "n00225786": "choking, strangling, strangulation, throttling",
  "n00226107": "spasm",
  "n00226319": "squeeze",
  "n00226511": "bronchospasm",
  "n00226711": "cardiospasm",
  "n00226951": "heave, retch",
  "n00227137": "laryngismus",
  "n00227264": "strangulation",
  "n00227484": "carjacking",
  "n00227595": "sacrifice, ritual killing",
  "n00227848": "hecatomb",
  "n00227969": "immolation",
  "n00228078": "electrocution",
  "n00228181": "decapitation, beheading",
  "n00228283": "abolition, abolishment",
  "n00228535": "liquidation, settlement",
  "n00228762": "viatical settlement, viaticus settlement",
  "n00228911": "withdrawal, drug withdrawal",
  "n00229034": "cold turkey",
  "n00229260": "closure, closedown, closing, shutdown",
  "n00229587": "plant closing",
  "n00229688": "bank closing",
  "n00229814": "layoff",
  "n00229934": "extinction, extinguishing, quenching",
  "n00230172": "fade, disappearance",
  "n00230324": "abortion",
  "n00230475": "spontaneous abortion, miscarriage, stillbirth",
  "n00230703": "habitual abortion",
  "n00230824": "imminent abortion, threatened abortion",
  "n00230997": "incomplete abortion, partial abortion",
  "n00231161": "induced abortion",
  "n00231315": "aborticide, feticide",
  "n00231412": "therapeutic abortion",
  "n00231567": "nullification, override",
  "n00231887": "abrogation, repeal, annulment",
  "n00232147": "derogation",
  "n00232386": "cancellation",
  "n00232604": "write-off",
  "n00232765": "attainder, civil death",
  "n00232863": "recission, rescission",
  "n00233212": "vitiation",
  "n00233386": "neutralization, neutralisation, counteraction",
  "n00233614": "deactivation, defusing",
  "n00233795": "deactivation, inactivation",
  "n00233980": "honorable discharge",
  "n00234105": "dishonorable discharge",
  "n00234277": "Section Eight",
  "n00234423": "neutralization, neutralisation",
  "n00234675": "neutralization, neutralisation",
  "n00234892": "reversal",
  "n00235074": "undoing",
  "n00235208": "regression, regress, reversion, retrogression, retroversion",
  "n00235435": "beginning, start, commencement",
  "n00236194": "springboard, jumping-off point, point of departure",
  "n00236581": "accession, rise to power",
  "n00236821": "activation",
  "n00236953": "attack, tone-beginning",
  "n00237078": "constitution, establishment, formation, organization, organisation",
  "n00237705": "re-establishment",
  "n00237869": "Creation",
  "n00238022": "introduction, debut, first appearance, launching, unveiling, entry",
  "n00238326": "induction of labor",
  "n00238527": "induction",
  "n00238778": "hypnogenesis",
  "n00238871": "product introduction",
  "n00239024": "face-off",
  "n00239230": "first step, initiative, opening move, opening",
  "n00239483": "groundbreaking, groundbreaking ceremony",
  "n00239666": "housing start",
  "n00239766": "icebreaker",
  "n00239910": "inauguration, startup",
  "n00240184": "initiation, founding, foundation, institution, origination, creation, innovation, introduction, instauration",
  "n00240754": "authorship, paternity",
  "n00240938": "installation, installing, installment, instalment",
  "n00241245": "jump ball",
  "n00241507": "kickoff",
  "n00241699": "start, starting",
  "n00242003": "resumption, recommencement",
  "n00242146": "scrum, scrummage",
  "n00242431": "startup",
  "n00242583": "unionization, unionisation",
  "n00242808": "arousal, rousing",
  "n00243102": "reveille",
  "n00243237": "ushering in",
  "n00243373": "inauguration, inaugural",
  "n00243577": "curtain raiser",
  "n00243662": "first base",
  "n00243813": "peace initiative",
  "n00243918": "cooking, cookery, preparation",
  "n00246341": "baking",
  "n00246460": "shirring",
  "n00246552": "toasting, browning",
  "n00246754": "broil, broiling, grilling",
  "n00246940": "frying, sauteing",
  "n00247081": "fusion cooking",
  "n00247229": "braising",
  "n00247342": "poaching",
  "n00247442": "roasting",
  "n00247654": "barbecuing",
  "n00247792": "boiling, stewing, simmering",
  "n00247955": "basting",
  "n00248063": "tenderization, tenderisation",
  "n00248252": "percolation",
  "n00248368": "seasoning",
  "n00248534": "salting",
  "n00248636": "sweetening",
  "n00248748": "infusion",
  "n00248977": "improvement",
  "n00249501": "advancement, progress",
  "n00249780": "forwarding, furtherance, promotion",
  "n00249987": "stride",
  "n00250143": "work flow, workflow",
  "n00250259": "development",
  "n00250597": "broadening",
  "n00250710": "elaboration, working out",
  "n00250882": "product development",
  "n00251013": "cleaning, cleansing, cleanup",
  "n00251520": "disinfestation",
  "n00251636": "spring-cleaning",
  "n00251780": "scrub, scrubbing, scouring",
  "n00252020": "swabbing, mopping",
  "n00252169": "dry cleaning",
  "n00252307": "sweeping",
  "n00252430": "purge, purging",
  "n00252662": "purge, purging, purgation",
  "n00252894": "purification",
  "n00253070": "purification, purgation",
  "n00253270": "purification",
  "n00253395": "catharsis, katharsis, abreaction",
  "n00253577": "catharsis, katharsis, purgation",
  "n00253773": "high colonic",
  "n00253919": "sterilization, sterilisation",
  "n00254166": "pasteurization, pasteurisation",
  "n00254415": "sanitation, sanitization, sanitisation",
  "n00254597": "depilation, epilation",
  "n00254769": "shave, shaving",
  "n00254922": "tonsure",
  "n00255065": "electrolysis",
  "n00255214": "washup, bathing",
  "n00255450": "ablution",
  "n00255600": "dishwashing, washup",
  "n00255710": "wash, washing, lavation",
  "n00256122": "washing-up",
  "n00256217": "window-washing",
  "n00256309": "rinse",
  "n00256407": "rinse",
  "n00256501": "soak, soaking",
  "n00256620": "brush, brushing",
  "n00256746": "comb, combing",
  "n00256961": "comb-out, teasing",
  "n00257115": "shampoo",
  "n00257228": "hair care, haircare, hairdressing",
  "n00257468": "hairweaving",
  "n00257580": "shower, shower bath",
  "n00257770": "bath",
  "n00257969": "bubble bath",
  "n00258093": "mikvah",
  "n00258301": "mud bath",
  "n00258403": "sponge bath",
  "n00258530": "Turkish bath, steam bath, vapor bath, vapour bath",
  "n00258695": "rubdown",
  "n00258854": "correction, rectification",
  "n00259177": "redress, remedy, remediation",
  "n00259392": "salve",
  "n00259544": "retribution",
  "n00259643": "recompense, compensation",
  "n00259894": "indemnification",
  "n00260051": "optimization, optimisation",
  "n00260494": "perfection",
  "n00260622": "reform",
  "n00260881": "land reform",
  "n00261029": "amelioration, melioration, betterment",
  "n00261258": "self-improvement, self-reformation",
  "n00261405": "reform",
  "n00261604": "beautification",
  "n00261797": "beauty treatment",
  "n00261972": "glamorization, glamorisation, glamourization, glamourisation",
  "n00262249": "decoration",
  "n00262596": "adornment",
  "n00262743": "ornamentation, embellishment",
  "n00262908": "window dressing",
  "n00263000": "trimming",
  "n00263164": "tessellation",
  "n00263272": "figuration",
  "n00263353": "tattoo",
  "n00263492": "titivation, tittivation",
  "n00263642": "marking",
  "n00263813": "lineation",
  "n00263947": "mottling",
  "n00264088": "striping",
  "n00264192": "clearing, clarification",
  "n00264366": "enrichment",
  "n00264529": "fortification",
  "n00264775": "humanization, humanisation",
  "n00264913": "modernization, modernisation",
  "n00265119": "renovation, redevelopment, overhaul",
  "n00265472": "face lift, facelift, face lifting",
  "n00265773": "moralization, moralisation",
  "n00265992": "enhancement, sweetening",
  "n00266149": "upturn",
  "n00266253": "worsening",
  "n00266401": "downturn, downswing",
  "n00266560": "downspin",
  "n00266645": "ventilation, airing",
  "n00266806": "repair, fix, fixing, fixture, mend, mending, reparation",
  "n00267217": "darning",
  "n00267349": "patching",
  "n00267522": "care, maintenance, upkeep",
  "n00267821": "camera care",
  "n00267917": "car care",
  "n00268025": "oil change",
  "n00268112": "overhaul, inspection and repair, service",
  "n00268341": "interim overhaul",
  "n00268457": "band aid, quick fix, quickie, quicky",
  "n00268557": "restoration",
  "n00268824": "gentrification",
  "n00269018": "reclamation, renewal, rehabilitation",
  "n00269258": "reconstruction",
  "n00269471": "anastylosis",
  "n00269674": "makeover",
  "n00269856": "reassembly, refabrication",
  "n00269963": "re-formation, regeneration",
  "n00270186": "rebuilding",
  "n00270275": "restitution",
  "n00270403": "pump priming",
  "n00270539": "scheduled maintenance",
  "n00270648": "steam fitting",
  "n00270800": "coaching, coaching job",
  "n00270919": "engagement, booking",
  "n00271155": "gig",
  "n00271263": "degradation, debasement",
  "n00271636": "dehumanization, dehumanisation",
  "n00271879": "brutalization, brutalisation, animalization, animalisation",
  "n00272123": "barbarization, barbarisation",
  "n00272303": "bastardization, bastardisation",
  "n00272448": "corruption, subversion",
  "n00272713": "demoralization, demoralisation",
  "n00272878": "stultification, constipation, impairment, deadening",
  "n00273077": "popularization, popularisation, vulgarization, vulgarisation",
  "n00273319": "profanation",
  "n00273449": "humiliation, abasement",
  "n00273601": "comedown",
  "n00273690": "change of color",
  "n00273877": "whitening, lightening",
  "n00274009": "bleach",
  "n00274206": "etiolation",
  "n00274437": "blackening, darkening",
  "n00274586": "obfuscation",
  "n00274707": "discoloration, discolouration",
  "n00274941": "coloring, colouring",
  "n00275151": "tinting",
  "n00275307": "hair coloring",
  "n00275424": "dyeing",
  "n00275572": "staining",
  "n00275751": "Gram's method, Gram method, Gram's procedure, Gram's stain, Gram stain",
  "n00276189": "environmentalism",
  "n00276342": "fixation, fixing",
  "n00276620": "soiling, soilure, dirtying",
  "n00276813": "staining, spotting, maculation",
  "n00276987": "contamination, pollution",
  "n00277267": "dust contamination",
  "n00277376": "wetting",
  "n00277569": "submersion, immersion, ducking, dousing",
  "n00277811": "drenching, soaking, souse, sousing",
  "n00278040": "moistening, dampening",
  "n00278221": "splash, splashing",
  "n00278403": "watering",
  "n00278555": "sprinkle, sprinkling, sparge",
  "n00278810": "chew, chewing, mastication, manduction",
  "n00279136": "chomping",
  "n00279235": "mumbling, gumming",
  "n00279377": "rumination",
  "n00279534": "bruxism",
  "n00279674": "defoliation",
  "n00279835": "motion, movement, move",
  "n00280586": "movement",
  "n00280853": "approach, approaching, coming",
  "n00281132": "access",
  "n00281291": "back door, backdoor",
  "n00281462": "closing, closure",
  "n00281752": "landing approach",
  "n00281898": "overshoot, wave-off, go-around",
  "n00282050": "progress, progression, procession, advance, advancement, forward motion, onward motion",
  "n00282485": "push",
  "n00282613": "career, life history",
  "n00282840": "march",
  "n00282953": "plain sailing, clear sailing, easy going",
  "n00283127": "locomotion, travel",
  "n00283450": "brachiation",
  "n00283568": "walk, walking",
  "n00283951": "ambulation",
  "n00284101": "amble, promenade, saunter, stroll, perambulation",
  "n00284409": "ramble, meander",
  "n00284544": "constitutional",
  "n00284665": "foot",
  "n00284798": "walk",
  "n00285141": "sleepwalking, somnambulism, somnambulation, noctambulism, noctambulation",
  "n00285387": "sleep talking, somniloquy, somniloquism",
  "n00285557": "step",
  "n00285889": "pace, stride, tread",
  "n00286112": "pas",
  "n00286240": "trip",
  "n00286360": "sidestep",
  "n00286497": "gait",
  "n00286756": "hitch, hobble, limp",
  "n00286957": "gait",
  "n00287142": "walk",
  "n00287276": "rack, single-foot",
  "n00287449": "jog trot",
  "n00287561": "trot",
  "n00287762": "rising trot",
  "n00287889": "sitting trot",
  "n00288000": "dressage",
  "n00288190": "curvet, vaulting",
  "n00288384": "piaffe",
  "n00288486": "canter, lope",
  "n00288671": "gallop",
  "n00288880": "footstep",
  "n00288970": "hike, hiking, tramp",
  "n00289175": "trudge",
  "n00289267": "flounce",
  "n00289388": "lurch, stumble, stagger",
  "n00289555": "pacing",
  "n00289659": "roll",
  "n00289737": "saunter",
  "n00289906": "skip",
  "n00290015": "stalk, angry walk",
  "n00290125": "strut, prance, swagger",
  "n00290276": "lurch, lunge",
  "n00290406": "waddle",
  "n00290579": "march, marching",
  "n00291004": "countermarch",
  "n00291154": "goose step",
  "n00291286": "last mile",
  "n00291400": "lockstep",
  "n00291663": "promenade",
  "n00291796": "quick march",
  "n00291876": "routemarch",
  "n00291965": "plodding, plod",
  "n00292125": "prowl",
  "n00292269": "moonwalk",
  "n00292386": "perambulation",
  "n00292599": "turn",
  "n00292712": "shamble, shambling, shuffle, shuffling",
  "n00292981": "space walk",
  "n00293125": "moonwalk",
  "n00293307": "wading",
  "n00293417": "walkabout",
  "n00293545": "walkabout",
  "n00293657": "walkabout",
  "n00293753": "walk-through",
  "n00293916": "run, running",
  "n00294190": "jog, trot, lope",
  "n00294366": "dogtrot",
  "n00294452": "dash, sprint",
  "n00294577": "fast break",
  "n00294748": "break",
  "n00294868": "crawl, crawling, creep, creeping",
  "n00295172": "lap, circle, circuit",
  "n00295422": "pace lap",
  "n00295563": "victory lap, lap of honour",
  "n00295701": "travel, traveling, travelling",
  "n00296263": "circumnavigation",
  "n00296478": "peregrination",
  "n00296585": "procession",
  "n00296946": "traversal, traverse",
  "n00297062": "wandering, roving, vagabondage",
  "n00297318": "wayfaring",
  "n00297404": "drifting",
  "n00297532": "crossing",
  "n00297657": "ford, fording",
  "n00297863": "shallow fording",
  "n00297951": "deep fording",
  "n00298047": "traversal, traverse",
  "n00298161": "tourism, touristry",
  "n00298351": "ecotourism",
  "n00298497": "driving",
  "n00299112": "motoring",
  "n00299217": "riding, horseback riding",
  "n00299680": "roping",
  "n00299826": "bronco busting",
  "n00299933": "endurance riding",
  "n00300037": "pack riding",
  "n00300113": "trail riding",
  "n00300206": "calf roping",
  "n00300333": "steer roping",
  "n00300441": "air travel, aviation, air",
  "n00301192": "flight",
  "n00301443": "connecting flight",
  "n00301598": "direct flight",
  "n00301728": "domestic flight",
  "n00301857": "international flight",
  "n00302001": "nonstop flight, nonstop",
  "n00302185": "redeye, redeye flight",
  "n00302394": "flight, flying",
  "n00302861": "acrobatics, aerobatics, stunting, stunt flying",
  "n00303049": "blind flying, blind landing",
  "n00303221": "ballooning",
  "n00303297": "flyover, fly-by, flypast",
  "n00303495": "glide, gliding, sailplaning, soaring, sailing",
  "n00303748": "hang gliding",
  "n00303849": "jump, parachuting",
  "n00304040": "skydiving",
  "n00304208": "maiden flight",
  "n00304349": "parasailing, paragliding",
  "n00304442": "overflight",
  "n00304592": "pass",
  "n00304729": "solo",
  "n00304851": "sortie",
  "n00305005": "touchdown",
  "n00305153": "aircraft landing, airplane landing",
  "n00305326": "ground-controlled approach, GCA",
  "n00305519": "crash landing",
  "n00305707": "three-point landing",
  "n00305858": "instrument landing",
  "n00305980": "splashdown",
  "n00306102": "takeoff",
  "n00306210": "tailspin, spin",
  "n00306318": "terrain flight, low level flight",
  "n00306426": "journey, journeying",
  "n00306900": "stage, leg",
  "n00307102": "staging",
  "n00307176": "leg",
  "n00307314": "fare-stage",
  "n00307455": "commute",
  "n00307631": "drive, ride",
  "n00308073": "long haul",
  "n00308208": "mush",
  "n00308279": "odyssey",
  "n00308370": "trip",
  "n00308681": "junket",
  "n00308779": "round trip",
  "n00308871": "run",
  "n00309011": "run",
  "n00309115": "passage, transit",
  "n00309276": "lift",
  "n00309368": "joyride",
  "n00309540": "spin",
  "n00309647": "expedition",
  "n00309802": "scouting trip",
  "n00309906": "campaign, hunting expedition, safari",
  "n00310063": "exploration, geographic expedition",
  "n00310201": "digression, excursion",
  "n00310347": "trek",
  "n00310425": "schlep, shlep",
  "n00310516": "trek",
  "n00310666": "tour, circuit",
  "n00311091": "grand tour",
  "n00311195": "grand tour",
  "n00311381": "itineration",
  "n00311528": "on the road, on tour",
  "n00311687": "pilgrimage, pilgrim's journey",
  "n00311809": "excursion, jaunt, outing, junket, pleasure trip, expedition, sashay",
  "n00312160": "junketing",
  "n00312266": "airing",
  "n00312403": "field trip",
  "n00312553": "voyage",
  "n00312691": "way",
  "n00312784": "ocean trip, voyage",
  "n00312932": "cruise, sail",
  "n00313096": "maiden voyage",
  "n00313245": "crossing",
  "n00313385": "lockage",
  "n00313502": "spaceflight, space travel, spacefaring",
  "n00313647": "water travel, seafaring",
  "n00313806": "sailing",
  "n00313971": "luff",
  "n00314094": "beat",
  "n00314271": "ministry",
  "n00314395": "tack",
  "n00314469": "seafaring, navigation, sailing",
  "n00315307": "cabotage",
  "n00315390": "boating, yachting",
  "n00315534": "bareboating",
  "n00315700": "commutation, commuting",
  "n00315830": "displacement, deracination",
  "n00315986": "transportation, transport, transfer, transferral, conveyance",
  "n00316460": "transshipment",
  "n00316594": "airlift, lift",
  "n00316806": "Berlin airlift",
  "n00316989": "connection, connexion",
  "n00317207": "delivery, bringing",
  "n00317485": "cattle drive",
  "n00317594": "drive",
  "n00317709": "airdrop",
  "n00317917": "consignment",
  "n00318035": "passage, handing over",
  "n00318186": "post",
  "n00318391": "service, serving, service of process",
  "n00318567": "relay",
  "n00318735": "carry",
  "n00319176": "pickup",
  "n00319312": "packing, backpacking",
  "n00319513": "piggyback",
  "n00319608": "fireman's carry",
  "n00319717": "portage",
  "n00319825": "porterage",
  "n00319939": "pursuit, chase, pursual, following",
  "n00320284": "trailing, tracking",
  "n00320486": "shadowing, tailing",
  "n00320625": "stalk, stalking",
  "n00320742": "wild-goose chase",
  "n00320852": "insertion, introduction, intromission",
  "n00321195": "cannulation, canulation, cannulization, cannulisation, canulization, canulisation, intubation",
  "n00321562": "catheterization, catheterisation",
  "n00321731": "instillation, instillment, instilment",
  "n00321956": "enclosure, enclosing, envelopment, inclosure",
  "n00322228": "packing, boxing",
  "n00322391": "bundling",
  "n00322488": "encasement, incasement",
  "n00322634": "injection",
  "n00322778": "epidural injection",
  "n00322962": "intradermal injection",
  "n00323056": "intramuscular injection",
  "n00323152": "intravenous injection",
  "n00323262": "fix",
  "n00323436": "subcutaneous injection",
  "n00323532": "infusion",
  "n00323766": "exchange transfusion",
  "n00323923": "transfusion",
  "n00324056": "transfusion, blood transfusion",
  "n00324233": "perfusion",
  "n00324384": "rise, ascent, ascension, ascending",
  "n00324699": "levitation",
  "n00324834": "heave, heaving",
  "n00324978": "funambulism, tightrope walking",
  "n00325110": "climb, mount",
  "n00325364": "scaling",
  "n00325502": "clamber",
  "n00325631": "escalade",
  "n00325785": "mountain climbing, mountaineering",
  "n00325975": "Alpinism",
  "n00326094": "rock climbing",
  "n00326291": "soar, zoom",
  "n00326440": "descent",
  "n00326677": "dive, nose dive, nosedive",
  "n00326809": "rappel, abseil",
  "n00327033": "swoop",
  "n00327134": "power dive",
  "n00327279": "crash dive",
  "n00327366": "drop",
  "n00327510": "flop, collapse",
  "n00327683": "lowering, letting down",
  "n00327824": "swing, swinging, vacillation",
  "n00328015": "return",
  "n00328230": "reentry",
  "n00328327": "remand",
  "n00328502": "slide, glide, coast",
  "n00328885": "slippage",
  "n00329031": "skid, slip, sideslip",
  "n00329227": "flow, stream",
  "n00329468": "snowboarding",
  "n00329619": "spill, spillage, release",
  "n00329819": "flood, overflow, outpouring",
  "n00329943": "effusion",
  "n00330035": "crawl",
  "n00330160": "speed, speeding, hurrying",
  "n00330457": "acceleration, quickening, speedup",
  "n00330674": "deceleration",
  "n00330836": "scud, scudding",
  "n00330984": "translation, displacement",
  "n00331102": "transplant, transplantation, transplanting",
  "n00331531": "troop movement",
  "n00331655": "shift, shifting",
  "n00331950": "motion, movement, move, motility",
  "n00333037": "abduction",
  "n00333203": "adduction",
  "n00333366": "agitation",
  "n00333572": "body English",
  "n00333733": "circumduction",
  "n00333829": "disturbance",
  "n00334174": "fetal movement, foetal movement",
  "n00334356": "flit, dart",
  "n00334509": "gesture",
  "n00334711": "headshake, headshaking",
  "n00334935": "jab, jabbing, poke, poking, thrust, thrusting",
  "n00335285": "mudra",
  "n00335384": "inclination, inclining",
  "n00335653": "inversion, eversion, everting",
  "n00335814": "inversion, upending",
  "n00335988": "jerk, jerking, jolt, saccade",
  "n00336274": "bob",
  "n00336430": "nod",
  "n00336561": "nutation",
  "n00336654": "stoop",
  "n00336805": "kick, kicking",
  "n00337078": "kneel, kneeling",
  "n00337210": "lurch, pitch, pitching",
  "n00337486": "eye movement",
  "n00337605": "nystagmus",
  "n00337844": "physiological nystagmus",
  "n00338088": "rotational nystagmus",
  "n00338271": "saccade",
  "n00338404": "post-rotational nystagmus",
  "n00338641": "opening",
  "n00338821": "rearrangement",
  "n00338994": "juggle, juggling",
  "n00339173": "musical chairs",
  "n00339463": "reordering",
  "n00339646": "permutation",
  "n00339776": "transposition, reversal",
  "n00339907": "transposition",
  "n00340192": "passing, overtaking",
  "n00340463": "shuffle, shuffling, make",
  "n00340662": "reshuffle, reshuffling",
  "n00340838": "riffle",
  "n00340989": "twiddle",
  "n00341109": "prostration",
  "n00341243": "reach, reaching, stretch",
  "n00341435": "reciprocation",
  "n00341548": "reclining",
  "n00341695": "retraction",
  "n00341890": "retroflection, retroflexion",
  "n00342028": "rotation, rotary motion",
  "n00342424": "circumvolution",
  "n00342565": "feather, feathering",
  "n00342755": "gyration, whirling",
  "n00342916": "pivot",
  "n00343091": "pronation",
  "n00343249": "spin, twirl, twist, twisting, whirl",
  "n00343606": "spiral",
  "n00343730": "pirouette",
  "n00343894": "birling, logrolling",
  "n00344040": "shutting, closing",
  "n00344259": "sitting",
  "n00344421": "sitting, posing",
  "n00344699": "snap",
  "n00344942": "squat, squatting",
  "n00345149": "sweep",
  "n00345297": "supination",
  "n00345454": "twist, turn",
  "n00345641": "wind, winding, twist",
  "n00345817": "toss",
  "n00345926": "vibration, quiver, quivering",
  "n00346095": "wave",
  "n00346296": "change of direction, reorientation",
  "n00346532": "turn",
  "n00346693": "reversion, reverse, reversal, turnabout, turnaround",
  "n00346936": "about-face, about turn",
  "n00347087": "u-turn",
  "n00347180": "shaking",
  "n00347359": "joggle, jiggle",
  "n00347490": "stirring",
  "n00347652": "wag, waggle, shake",
  "n00347853": "worrying",
  "n00348008": "rock, careen, sway, tilt",
  "n00348312": "upset, overturn, turnover",
  "n00348571": "waver, flutter, flicker",
  "n00348801": "tremor, shudder",
  "n00348953": "outreach",
  "n00349080": "standing",
  "n00349213": "straddle, span",
  "n00349369": "stroke",
  "n00349520": "keystroke, key stroke",
  "n00349705": "wiggle, wriggle, squirm",
  "n00349886": "change of course",
  "n00350030": "turn, turning",
  "n00350380": "diversion, deviation, digression, deflection, deflexion, divagation",
  "n00350878": "red herring",
  "n00351000": "right",
  "n00351168": "left",
  "n00351334": "tack, tacking",
  "n00351485": "change of magnitude",
  "n00351638": "decrease, diminution, reduction, step-down",
  "n00352331": "cut",
  "n00352585": "budget cut",
  "n00352683": "pay cut, salary cut",
  "n00352778": "cost cutting",
  "n00352861": "price cutting, price cut",
  "n00353013": "spending cut",
  "n00353100": "tax cut",
  "n00353249": "moderation, mitigation",
  "n00353469": "lowering",
  "n00353611": "tapering",
  "n00353782": "cutback",
  "n00353903": "service cutback",
  "n00353992": "devaluation",
  "n00354183": "devitalization, devitalisation",
  "n00354342": "evisceration",
  "n00354583": "extenuation, mitigation, palliation",
  "n00354785": "spasmolysis",
  "n00354884": "easing, easement, alleviation, relief",
  "n00355252": "de-escalation",
  "n00355420": "detente",
  "n00355547": "palliation",
  "n00355691": "liberalization, liberalisation, relaxation",
  "n00355919": "minimization, minimisation",
  "n00356199": "depletion",
  "n00356367": "consumption, using up, expenditure",
  "n00356516": "burnup",
  "n00356621": "exhaustion",
  "n00356790": "compression, compressing",
  "n00357023": "squeeze, squeezing",
  "n00357275": "pinch, tweak",
  "n00357451": "decompression, decompressing",
  "n00357680": "condensing, condensation",
  "n00357906": "thickening, inspissation",
  "n00358089": "crush, crunch, compaction",
  "n00358290": "grind, mill, pulverization, pulverisation",
  "n00358528": "expression",
  "n00358702": "extrusion, expulsion",
  "n00358931": "shortening",
  "n00359135": "abbreviation",
  "n00359238": "cut, cutting, cutting off",
  "n00359492": "severance, severing",
  "n00359614": "clip, clipping, snip",
  "n00359822": "haircut",
  "n00359903": "trim, trimming, clipping",
  "n00360143": "pruning",
  "n00360242": "shearing",
  "n00360404": "sheepshearing",
  "n00360485": "shrinking",
  "n00360601": "miniaturization, miniaturisation",
  "n00360757": "subtraction, deduction",
  "n00361041": "bite",
  "n00361192": "withholding",
  "n00361333": "abatement",
  "n00361459": "abatement of a nuisance, nuisance abatement",
  "n00361684": "asbestos abatement",
  "n00361795": "attrition",
  "n00361932": "deflation",
  "n00362103": "discount, price reduction, deduction",
  "n00362258": "rollback",
  "n00362355": "weakening",
  "n00362546": "wilt, wilting",
  "n00362659": "dilution",
  "n00362862": "etiolation",
  "n00363052": "cutting, thinning",
  "n00363260": "increase, step-up",
  "n00363788": "addition",
  "n00364053": "retrofit",
  "n00364260": "advance, rise",
  "n00364440": "appreciation",
  "n00364600": "depreciation",
  "n00364787": "surge, upsurge",
  "n00365012": "fluoridation, fluoridization, fluoridisation",
  "n00365231": "augmentation",
  "n00365329": "amplification",
  "n00365471": "contraction",
  "n00365709": "expansion, enlargement",
  "n00365995": "dilation, dilatation",
  "n00366207": "vasodilation",
  "n00366317": "distention, distension",
  "n00366521": "stretching",
  "n00366675": "tension",
  "n00366846": "escalation",
  "n00367066": "maximization, maximisation, maximation",
  "n00367280": "inflation",
  "n00367427": "magnification",
  "n00367552": "exaggeration",
  "n00367768": "extension",
  "n00367976": "spread, spreading",
  "n00368302": "circulation",
  "n00368515": "recirculation",
  "n00368592": "dispersion, dispersal, dissemination, diffusion",
  "n00368939": "crop-dusting, spraying",
  "n00369138": "scatter, scattering, strewing",
  "n00369399": "contracture",
  "n00369532": "extension",
  "n00369694": "hyperextension",
  "n00369802": "contraction, muscular contraction, muscle contraction",
  "n00370127": "tetanus",
  "n00370261": "truncation",
  "n00370458": "uterine contraction",
  "n00370688": "Braxton-Hicks contraction, false labor",
  "n00370885": "vaginismus",
  "n00371059": "stretch",
  "n00371314": "expansion, expanding upon",
  "n00371487": "amplification, elaboration",
  "n00371735": "annotation, annotating",
  "n00371846": "supplementation, subjunction, subjoining",
  "n00372013": "accumulation, accrual, accruement",
  "n00372226": "buildup",
  "n00372448": "deposit, deposition",
  "n00372607": "repositing, reposition, storage, warehousing",
  "n00372977": "stockpiling",
  "n00373130": "inclusion",
  "n00373278": "incorporation",
  "n00373418": "annexation, appropriation",
  "n00373544": "aggrandizement, aggrandisement, elevation",
  "n00373862": "self-aggrandizement, self-aggrandisement, ego trip",
  "n00374063": "strengthening",
  "n00374224": "intensification",
  "n00374446": "roughness",
  "n00374677": "intensification",
  "n00374835": "aggravation, exacerbation",
  "n00375071": "concentration",
  "n00375348": "pervaporation",
  "n00375625": "focalization, focalisation, focusing",
  "n00375938": "refocusing",
  "n00376063": "change of integrity",
  "n00376400": "breakage, break, breaking",
  "n00376715": "rupture",
  "n00376825": "smashing, shattering",
  "n00376994": "fracture, crack, cracking",
  "n00377169": "chip, chipping, splintering",
  "n00377364": "explosion, burst",
  "n00377686": "detonation",
  "n00377813": "percussion",
  "n00377907": "fulmination",
  "n00378069": "burning, combustion",
  "n00378296": "arson, incendiarism, fire-raising",
  "n00378479": "ignition, firing, lighting, kindling, inflammation",
  "n00378706": "incineration",
  "n00378880": "cremation",
  "n00378985": "combination, combining, compounding",
  "n00379422": "attachment, affixation",
  "n00379588": "graft, grafting",
  "n00379754": "confusion",
  "n00379993": "babel",
  "n00380083": "mix, commixture, admixture, mixture, intermixture, mixing",
  "n00380568": "fusion",
  "n00380696": "blend, blending",
  "n00380881": "confluence, conflux, merging",
  "n00380994": "homogenization, homogenisation",
  "n00381326": "interspersion, interspersal",
  "n00381567": "temperance",
  "n00381680": "union, unification, uniting, conjugation, jointure",
  "n00382109": "coalescence, coalescency, coalition, concretion, conglutination",
  "n00382474": "reunion, reunification",
  "n00382739": "tribalization, tribalisation",
  "n00382906": "detribalization, detribalisation",
  "n00383071": "umbrella",
  "n00383281": "homecoming",
  "n00383390": "opening",
  "n00383606": "separation",
  "n00383952": "break, interruption, disruption, gap",
  "n00384329": "cut-in, insert",
  "n00384510": "cut-in, insert",
  "n00384658": "avulsion",
  "n00384802": "dissociation",
  "n00384933": "secession, withdrawal",
  "n00385104": "Secession",
  "n00385266": "breakaway, breaking away",
  "n00385501": "disunion",
  "n00385649": "disconnection, disjunction",
  "n00385791": "division",
  "n00386164": "parcellation",
  "n00386345": "cleavage",
  "n00386450": "bisection",
  "n00386553": "quartering",
  "n00386676": "schism, split",
  "n00386915": "cut, cutting",
  "n00387214": "dissection",
  "n00387326": "scission",
  "n00387424": "slicing",
  "n00387526": "undercut",
  "n00387657": "cut, cutting",
  "n00387897": "notch, nick, snick",
  "n00388065": "slash, gash",
  "n00388210": "atomization, atomisation, fragmentation",
  "n00388392": "branching, ramification, fork, forking",
  "n00388710": "bifurcation",
  "n00388842": "trifurcation",
  "n00388959": "divarication",
  "n00389043": "fibrillation",
  "n00389135": "dichotomization, dichotomisation",
  "n00389308": "quantization, quantisation",
  "n00389508": "fractionation",
  "n00389610": "pairing",
  "n00389763": "buddy system",
  "n00389953": "match-up, matchup",
  "n00390198": "punctuation",
  "n00390445": "hyphenation",
  "n00390581": "syllabication, syllabification",
  "n00390735": "word division, hyphenation",
  "n00390906": "detachment, disengagement",
  "n00391086": "tear",
  "n00391293": "laceration",
  "n00391407": "rent, rip, split",
  "n00391599": "removal, remotion",
  "n00392093": "drawing, drawing off",
  "n00392335": "derivation",
  "n00392468": "derivation",
  "n00392582": "derivation",
  "n00392709": "abscission, cutting off",
  "n00392848": "abstraction",
  "n00392950": "extraction",
  "n00393161": "threshing",
  "n00393369": "ablation, extirpation, cutting out, excision",
  "n00394242": "autotomy",
  "n00394485": "decontamination",
  "n00394610": "deletion",
  "n00394803": "denudation, stripping, uncovering, baring, husking",
  "n00395017": "dermabrasion",
  "n00395199": "dislodgment, dislodgement",
  "n00395333": "elimination, riddance",
  "n00395503": "elimination",
  "n00395654": "circumcision",
  "n00395797": "emptying, voidance, evacuation",
  "n00396029": "drain, drainage",
  "n00396213": "bank withdrawal",
  "n00396344": "bank run",
  "n00396509": "disinvestment",
  "n00396642": "rinse, rinsing",
  "n00396825": "bowdlerization, bowdlerisation",
  "n00397010": "expurgation, castration",
  "n00397191": "bowdlerization, bowdlerisation",
  "n00397347": "censoring, censorship",
  "n00397545": "Bowdlerism",
  "n00397647": "Comstockery",
  "n00397760": "expunction, expunging, erasure",
  "n00397953": "division, partition, partitioning, segmentation, sectionalization, sectionalisation",
  "n00398427": "subdivision",
  "n00398585": "septation",
  "n00398704": "transformation, translation",
  "n00399030": "transformation",
  "n00399223": "permutation",
  "n00399393": "revision, alteration",
  "n00399717": "transfiguration",
  "n00399846": "transmogrification",
  "n00400083": "conversion",
  "n00400278": "afforestation",
  "n00400449": "reforestation, re-afforestation",
  "n00400645": "rehabilitation",
  "n00400851": "correctional rehabilitation",
  "n00400995": "physical rehabilitation, physical restoration, therapeutic rehabilitation",
  "n00401194": "urban renewal",
  "n00401314": "vocational rehabilitation",
  "n00401459": "reinstatement",
  "n00401639": "rejuvenation",
  "n00401783": "refreshment, recreation",
  "n00402128": "metamorphosis",
  "n00402308": "transfiguration, metamorphosis",
  "n00402535": "filling",
  "n00402671": "saturation",
  "n00402789": "hardening",
  "n00402951": "annealing, tempering",
  "n00403092": "damage, harm, hurt, scathe",
  "n00403334": "impairment",
  "n00403466": "defacement, disfigurement, disfiguration",
  "n00403783": "wound, wounding",
  "n00403911": "burn",
  "n00404058": "scald",
  "n00404170": "updating",
  "n00404403": "change of shape",
  "n00404726": "contortion, deformation",
  "n00404959": "convolution",
  "n00405106": "angulation",
  "n00405206": "bending",
  "n00405360": "flexion, flexure",
  "n00405628": "flex",
  "n00405766": "crouch",
  "n00405892": "dorsiflexion",
  "n00406007": "elongation",
  "n00406114": "hunch",
  "n00406234": "incurvation",
  "n00406365": "involution, enfolding",
  "n00406485": "corrugation",
  "n00406612": "fold, folding",
  "n00406800": "plication, pleating",
  "n00406939": "indentation",
  "n00407090": "protrusion, projection, jut, jutting",
  "n00407270": "widening, broadening",
  "n00407407": "narrowing",
  "n00407535": "activity",
  "n00409075": "domesticity",
  "n00409211": "operation",
  "n00409483": "operation",
  "n00409823": "rescue operation",
  "n00409943": "undercover operation",
  "n00410096": "buy-and-bust operation",
  "n00410247": "practice, pattern",
  "n00411048": "practice, praxis",
  "n00411235": "biologism",
  "n00411384": "cooperation",
  "n00411553": "featherbedding",
  "n00411717": "formalism",
  "n00411859": "mycophagy",
  "n00412006": "one-upmanship",
  "n00412130": "pluralism",
  "n00412271": "symbolism, symbolization, symbolisation",
  "n00412449": "modernism",
  "n00412572": "occult, occult arts",
  "n00412727": "ornamentalism",
  "n00412839": "cannibalism",
  "n00412982": "anthropophagy",
  "n00413086": "careerism",
  "n00413239": "custom, usage, usance",
  "n00413554": "Americanism",
  "n00413674": "Anglicism, Britishism",
  "n00413795": "consuetude",
  "n00413904": "couvade",
  "n00414071": "Germanism",
  "n00414179": "habit, use",
  "n00414599": "hijab",
  "n00414790": "ritual",
  "n00414881": "second nature",
  "n00415000": "habitude",
  "n00415098": "round, daily round",
  "n00415226": "fashion",
  "n00415333": "lobbyism",
  "n00415442": "slavery, slaveholding",
  "n00415540": "peonage",
  "n00415676": "way, path, way of life",
  "n00415988": "ambages",
  "n00416129": "primrose path",
  "n00416216": "straight and narrow, strait and narrow",
  "n00416409": "Sunnah, Sunna, hadith",
  "n00416646": "warpath",
  "n00416737": "line of least resistance, path of least resistance",
  "n00416914": "unwritten law",
  "n00417005": "lynch law",
  "n00417131": "chokehold, choke hold",
  "n00417397": "embrace, embracing, embracement",
  "n00417643": "cuddle, nestle, snuggle",
  "n00417859": "hug, clinch, squeeze",
  "n00418025": "mistreatment",
  "n00418305": "nonconformism",
  "n00418394": "annoyance, annoying, irritation, vexation",
  "n00418615": "disregard, neglect",
  "n00418787": "despite",
  "n00418903": "exploitation, victimization, victimisation, using",
  "n00419315": "blaxploitation",
  "n00419464": "sexploitation",
  "n00419644": "harassment, molestation",
  "n00419908": "maltreatment, ill-treatment, ill-usage, abuse",
  "n00420218": "child abuse",
  "n00420336": "child neglect",
  "n00420477": "persecution",
  "n00420712": "repression",
  "n00420877": "impalement",
  "n00421047": "oppression, subjugation",
  "n00421210": "pogrom",
  "n00421321": "rendition",
  "n00421437": "torture, torturing",
  "n00422114": "bastinado, falanga",
  "n00422261": "boot",
  "n00422391": "burning",
  "n00422551": "crucifixion, excruciation",
  "n00422719": "genital torture",
  "n00422834": "judicial torture",
  "n00423028": "kia quen",
  "n00423222": "kittee",
  "n00423423": "nail pulling, nail removal",
  "n00423563": "picket, piquet",
  "n00423769": "prolonged interrogation",
  "n00423916": "rack",
  "n00424034": "sensory deprivation",
  "n00424186": "sleep deprivation",
  "n00424324": "strappado, strapado",
  "n00424599": "cruelty, inhuman treatment",
  "n00424767": "atrocity, inhumanity",
  "n00424934": "brutality, barbarity, barbarism, savagery",
  "n00425090": "outrage",
  "n00425182": "baiting",
  "n00425278": "badgering, worrying, torment, bedevilment",
  "n00425451": "exasperation",
  "n00425615": "red flag",
  "n00425781": "sexual harassment",
  "n00425905": "tease, teasing, ribbing, tantalization",
  "n00426275": "witch-hunt",
  "n00426388": "McCarthyism",
  "n00426526": "colonialism",
  "n00426757": "neocolonialism",
  "n00426928": "diversion, recreation",
  "n00427580": "antic, joke, prank, trick, caper, put-on",
  "n00427853": "bathing",
  "n00428000": "celebration, festivity",
  "n00428270": "dancing, dance, terpsichore, saltation",
  "n00429048": "entertainment, amusement",
  "n00429322": "escapade, lark",
  "n00429440": "escape, escapism",
  "n00429713": "eurythmy, eurhythmy, eurythmics, eurhythmics",
  "n00429949": "fun, merriment, playfulness",
  "n00430140": "gambling, gaming, play",
  "n00430606": "game",
  "n00431005": "jest, joke, jocularity",
  "n00431292": "nightlife, night life",
  "n00431552": "pastime, interest, pursuit",
  "n00431893": "play, child's play",
  "n00432164": "house",
  "n00432356": "doctor",
  "n00432587": "fireman",
  "n00432689": "avocation, by-line, hobby, pursuit, sideline, spare-time activity",
  "n00432881": "cup of tea, bag, dish",
  "n00433113": "confectionery",
  "n00433216": "sport",
  "n00433458": "contact sport",
  "n00433661": "outdoor sport, field sport",
  "n00433802": "gymnastics, gymnastic exercise",
  "n00434075": "acrobatics, tumbling",
  "n00434231": "backbend",
  "n00434396": "back circle",
  "n00434568": "walkover",
  "n00434657": "cartwheel",
  "n00434844": "crucifix",
  "n00435013": "dip",
  "n00435182": "double leg circle",
  "n00435401": "grand circle",
  "n00435563": "cardiopulmonary exercise",
  "n00435778": "gymnastic exercise",
  "n00436187": "handstand",
  "n00436339": "hang",
  "n00436609": "bent hang",
  "n00436702": "inverted hang",
  "n00436817": "lever hang",
  "n00436953": "reverse hang",
  "n00437067": "straight hang",
  "n00437219": "piked reverse hang",
  "n00437321": "kick up",
  "n00437487": "handspring",
  "n00437639": "headstand",
  "n00437788": "tumble",
  "n00437929": "split",
  "n00438065": "acrobatic stunt, acrobatic feat",
  "n00438338": "kip, upstart",
  "n00438606": "long fly",
  "n00438725": "scissors",
  "n00438893": "straddle",
  "n00439043": "split, stock split, split up",
  "n00439284": "reverse split, reverse stock split, split down",
  "n00439484": "somersault, somerset, summersault, summerset, somersaulting, flip",
  "n00439749": "flip-flop",
  "n00439826": "track and field",
  "n00440039": "track, running",
  "n00440218": "jumping",
  "n00440382": "broad jump, long jump",
  "n00440509": "high jump",
  "n00440643": "Fosbury flop",
  "n00440747": "skiing",
  "n00440941": "cross-country skiing",
  "n00441073": "ski jumping",
  "n00441235": "kick turn",
  "n00441501": "stem turn, stem",
  "n00441686": "telemark",
  "n00441824": "water sport, aquatics",
  "n00442115": "swimming, swim",
  "n00442437": "bathe",
  "n00442569": "sea bathing",
  "n00442654": "skinny-dip",
  "n00442759": "sun bathing",
  "n00442847": "dip, plunge",
  "n00442981": "dive, diving",
  "n00443231": "floating, natation",
  "n00443375": "dead-man's float, prone float",
  "n00443517": "belly flop, belly flopper, belly whop, belly whopper",
  "n00443692": "cliff diving",
  "n00443803": "flip",
  "n00443917": "gainer, full gainer",
  "n00444142": "half gainer",
  "n00444340": "jackknife",
  "n00444490": "swan dive, swallow dive",
  "n00444651": "skin diving, skin-dive",
  "n00444846": "scuba diving",
  "n00444937": "snorkeling, snorkel diving",
  "n00445055": "surfing, surfboarding, surfriding",
  "n00445226": "water-skiing",
  "n00445351": "rowing, row",
  "n00445529": "crab",
  "n00445685": "sculling",
  "n00445802": "boxing, pugilism, fisticuffs",
  "n00446311": "professional boxing",
  "n00446411": "in-fighting",
  "n00446493": "fight",
  "n00446632": "rope-a-dope",
  "n00446804": "spar, sparring",
  "n00446980": "archery",
  "n00447073": "sledding",
  "n00447221": "tobogganing",
  "n00447361": "luging",
  "n00447463": "bobsledding",
  "n00447540": "wrestling, rassling, grappling",
  "n00447879": "flying mare",
  "n00447957": "Greco-Roman wrestling",
  "n00448126": "professional wrestling",
  "n00448232": "sumo",
  "n00448466": "skating",
  "n00448640": "ice skating",
  "n00448748": "figure skating",
  "n00448872": "rollerblading",
  "n00448958": "roller skating",
  "n00449054": "skateboarding",
  "n00449168": "speed skating",
  "n00449295": "racing",
  "n00449517": "auto racing, car racing",
  "n00449695": "boat racing",
  "n00449796": "hydroplane racing",
  "n00449892": "camel racing",
  "n00449977": "greyhound racing",
  "n00450070": "horse racing",
  "n00450230": "thoroughbred racing",
  "n00450335": "riding, horseback riding, equitation",
  "n00450700": "equestrian sport",
  "n00450866": "pony-trekking",
  "n00450998": "showjumping, stadium jumping",
  "n00451186": "cross-country riding, cross-country jumping",
  "n00451370": "cycling",
  "n00451563": "bicycling",
  "n00451635": "motorcycling",
  "n00451768": "dune cycling",
  "n00451866": "blood sport",
  "n00452034": "bullfighting, tauromachy",
  "n00452152": "cockfighting",
  "n00452293": "hunt, hunting",
  "n00452734": "battue",
  "n00452864": "beagling",
  "n00452947": "canned hunt",
  "n00453126": "coursing",
  "n00453313": "deer hunting, deer hunt",
  "n00453396": "ducking, duck hunting",
  "n00453478": "fox hunting, foxhunt",
  "n00453631": "pigsticking",
  "n00453731": "farming, land",
  "n00453935": "fishing, sportfishing",
  "n00454121": "fishing",
  "n00454237": "angling",
  "n00454395": "fly-fishing",
  "n00454493": "troll, trolling",
  "n00454624": "casting, cast",
  "n00454855": "bait casting",
  "n00454983": "fly casting",
  "n00455076": "overcast",
  "n00455173": "surf casting, surf fishing",
  "n00455348": "follow-up, followup",
  "n00455599": "game",
  "n00456199": "game",
  "n00456465": "day game",
  "n00456563": "night game",
  "n00456690": "away game, road game",
  "n00456804": "home game",
  "n00456899": "exhibition game, practice game",
  "n00457038": "follow-on",
  "n00457228": "innings",
  "n00457382": "turn, play",
  "n00457723": "attack",
  "n00457890": "opening, chess opening",
  "n00458094": "counterattack, counterplay",
  "n00458286": "down",
  "n00458456": "bat, at-bat",
  "n00458641": "catch",
  "n00458800": "party game",
  "n00458890": "computer game, video game",
  "n00459013": "virtual reality",
  "n00459284": "pinball, pinball game",
  "n00459465": "pachinko",
  "n00459570": "guessing game",
  "n00459727": "charades",
  "n00459845": "ducks and drakes",
  "n00459979": "mind game",
  "n00460078": "paper chase, hare and hounds",
  "n00460344": "ring-around-the-rosy, ring-around-a-rosy, ring-a-rosy",
  "n00460541": "prisoner's base",
  "n00460706": "treasure hunt",
  "n00460844": "nightcap",
  "n00460951": "twin bill, doubleheader, double feature",
  "n00461162": "playoff game",
  "n00461294": "cup tie",
  "n00461402": "war game",
  "n00461611": "curling",
  "n00461782": "bowling",
  "n00462116": "frame",
  "n00462241": "tenpins, tenpin bowling",
  "n00462383": "ninepins, skittles",
  "n00462557": "duckpins",
  "n00462686": "candlepins, candlepin bowling",
  "n00462804": "lawn bowling, bowls",
  "n00462997": "bocce, bocci, boccie",
  "n00463116": "pall-mall",
  "n00463246": "athletic game",
  "n00463543": "ice hockey, hockey, hockey game",
  "n00464037": "goalkeeper, goalie, goaltender, netkeeper",
  "n00464277": "tetherball",
  "n00464478": "water polo",
  "n00464651": "outdoor game",
  "n00464894": "golf, golf game",
  "n00466273": "professional golf",
  "n00466377": "round of golf, round",
  "n00466524": "medal play, stroke play",
  "n00466630": "match play",
  "n00466712": "miniature golf",
  "n00466880": "croquet",
  "n00467122": "paintball",
  "n00467320": "quoits, horseshoes",
  "n00467536": "shuffleboard, shovelboard",
  "n00467719": "field game",
  "n00467995": "field hockey, hockey",
  "n00468299": "shinny, shinney",
  "n00468480": "football, football game",
  "n00469651": "American football, American football game",
  "n00470554": "professional football",
  "n00470682": "touch football",
  "n00470830": "hurling",
  "n00470966": "rugby, rugby football, rugger",
  "n00471277": "knock on",
  "n00471437": "ball game, ballgame",
  "n00471613": "baseball, baseball game",
  "n00474568": "ball",
  "n00474657": "professional baseball",
  "n00474769": "hardball",
  "n00474881": "perfect game",
  "n00475014": "no-hit game, no-hitter",
  "n00475142": "one-hitter, 1-hitter",
  "n00475273": "two-hitter, 2-hitter",
  "n00475403": "three-hitter, 3-hitter",
  "n00475535": "four-hitter, 4-hitter",
  "n00475661": "five-hitter, 5-hitter",
  "n00475787": "softball, softball game",
  "n00476140": "rounders",
  "n00476235": "stickball, stickball game",
  "n00476389": "cricket",
  "n00476788": "run-up",
  "n00476952": "Chinaman",
  "n00477097": "googly, wrong 'un, bosie, bosie ball",
  "n00477268": "no ball",
  "n00477392": "lacrosse",
  "n00477639": "polo",
  "n00477827": "pushball",
  "n00477986": "ultimate frisbee",
  "n00478262": "soccer, association football",
  "n00478647": "dribble, dribbling",
  "n00478855": "double dribble",
  "n00479076": "court game",
  "n00479440": "handball",
  "n00479616": "racquetball",
  "n00479734": "fives",
  "n00479887": "squash, squash racquets, squash rackets",
  "n00480211": "volleyball, volleyball game",
  "n00480366": "jai alai, pelota",
  "n00480508": "badminton",
  "n00480885": "battledore, battledore and shuttlecock",
  "n00480993": "basketball, basketball game, hoops",
  "n00481666": "tip-off, tap-off",
  "n00481803": "professional basketball",
  "n00481938": "deck tennis",
  "n00482122": "netball",
  "n00482298": "tennis, lawn tennis",
  "n00482892": "break, break of serve",
  "n00483098": "equalizer",
  "n00483205": "professional tennis",
  "n00483313": "singles",
  "n00483409": "singles",
  "n00483508": "doubles",
  "n00483605": "doubles",
  "n00483705": "royal tennis, real tennis, court tennis",
  "n00483848": "pallone",
  "n00483935": "child's game",
  "n00484362": "blindman's bluff, blindman's buff",
  "n00484532": "cat and mouse, cat and rat",
  "n00484802": "cat's cradle",
  "n00484910": "hide-and-seek, hide and go seek",
  "n00485081": "hopscotch",
  "n00485262": "jacks, jackstones, knucklebones",
  "n00485450": "jackstraws, spillikins",
  "n00485632": "jump rope",
  "n00485815": "double Dutch",
  "n00485998": "leapfrog",
  "n00486130": "leapfrog",
  "n00486358": "marbles",
  "n00486490": "mumblety-peg, mumble-the-peg",
  "n00486670": "musical chairs, going to Jerusalem",
  "n00487072": "peekaboo, bopeep",
  "n00487228": "pillow fight",
  "n00487326": "post office",
  "n00487453": "spin the bottle",
  "n00487617": "spin the plate, spin the platter",
  "n00487874": "tag",
  "n00488051": "tiddlywinks",
  "n00488225": "card game, cards",
  "n00489475": "cut, cutting",
  "n00489730": "all fours, high-low-jack",
  "n00489933": "baccarat, chemin de fer",
  "n00490155": "beggar-my-neighbor, beggar-my-neighbour, strip-Jack-naked",
  "n00490350": "blackjack, twenty-one, vingt-et-un",
  "n00490569": "bridge",
  "n00491037": "bridge whist",
  "n00491161": "auction, auction bridge",
  "n00491366": "contract, contract bridge",
  "n00491561": "no-trump",
  "n00491713": "casino, cassino",
  "n00491901": "cribbage, crib",
  "n00492083": "crib",
  "n00492195": "ecarte",
  "n00492309": "euchre, five hundred",
  "n00492497": "fantan, sevens, parliament",
  "n00492724": "faro",
  "n00492871": "Go Fish",
  "n00493031": "monte, four-card monte, three-card monte",
  "n00493308": "Michigan, Chicago, Newmarket, boodle, stops",
  "n00493742": "Napoleon, nap",
  "n00493858": "old maid",
  "n00494100": "pinochle, pinocle, penuchle, bezique",
  "n00494293": "piquet",
  "n00494406": "pisha paysha",
  "n00494768": "poker, poker game",
  "n00495103": "rouge et noir, trente-et-quarante",
  "n00495331": "rummy, rum",
  "n00495524": "solitaire, patience",
  "n00495677": "canfield",
  "n00495774": "klondike",
  "n00496167": "whist, long whist, short whist",
  "n00496437": "dummy whist",
  "n00496599": "hearts, Black Maria",
  "n00496760": "Russian bank, crapette",
  "n00496877": "gin, gin rummy, knock rummy",
  "n00497060": "canasta, basket rummy, meld",
  "n00497321": "bolivia",
  "n00497425": "samba",
  "n00497536": "draw, draw poker",
  "n00497713": "high-low",
  "n00497819": "penny ante, penny ante poker",
  "n00497924": "straight poker",
  "n00498086": "strip poker",
  "n00498220": "stud, stud poker",
  "n00498427": "cinch",
  "n00498558": "pitch, auction pitch",
  "n00498700": "seven-up, old sledge",
  "n00498828": "royal casino",
  "n00498947": "spade casino",
  "n00499066": "table game",
  "n00499263": "table tennis, Ping-Pong",
  "n00499477": "dominoes, dominos",
  "n00499599": "nim",
  "n00499924": "billiards",
  "n00500280": "break",
  "n00500449": "carom, cannon",
  "n00500669": "masse, masse shot",
  "n00500927": "miscue",
  "n00501080": "pool, pocket billiards",
  "n00501304": "snooker",
  "n00501479": "bagatelle, bar billiards",
  "n00501722": "parlor game, parlour game",
  "n00501870": "word game",
  "n00502030": "anagrams",
  "n00502161": "Scrabble",
  "n00502415": "board game",
  "n00502822": "backgammon",
  "n00502952": "checkers, draughts",
  "n00503237": "chess, chess game",
  "n00503981": "Chinese checkers, Chinese chequers",
  "n00504209": "darts",
  "n00504325": "go, go game",
  "n00504526": "halma",
  "n00504660": "lotto, bingo, beano, keno",
  "n00504844": "tombola",
  "n00504975": "ludo",
  "n00505126": "Mah-Jongg, mahjong",
  "n00505238": "Monopoly",
  "n00505446": "pachisi, parchesi, parchisi",
  "n00505615": "Parcheesi",
  "n00505726": "shogi",
  "n00505871": "shovel board, shove-halfpenny, shove-ha'penny",
  "n00506060": "snakes and ladders",
  "n00506207": "ticktacktoe, ticktacktoo, tick-tack-toe, tic-tac-toe, tit-tat-toe, noughts and crosses",
  "n00506554": "sporting life",
  "n00506658": "bet, wager",
  "n00506919": "daily double",
  "n00507025": "exacta, perfecta",
  "n00507162": "parimutuel",
  "n00507271": "parlay",
  "n00507425": "place bet",
  "n00507539": "superfecta",
  "n00507673": "game of chance, gambling game",
  "n00507887": "fantan, fan tan",
  "n00508091": "lottery, drawing",
  "n00508340": "lucky dip",
  "n00508547": "numbers pool, numbers game, numbers racket, numbers",
  "n00508672": "raffle",
  "n00508800": "sweepstakes",
  "n00508952": "craps",
  "n00509302": "crap shooting, crapshoot, crap game",
  "n00509399": "roulette",
  "n00509566": "banking game",
  "n00509706": "zero-sum game",
  "n00509846": "merrymaking, conviviality, jollification",
  "n00510050": "jinks, high jinks, hijinks, high jinx",
  "n00510189": "revel, revelry",
  "n00510383": "sexcapade",
  "n00510475": "spree, fling",
  "n00510624": "spending spree",
  "n00510723": "bust, tear, binge, bout",
  "n00510922": "piss-up",
  "n00511041": "carouse, carousal, bender, toot, booze-up",
  "n00511212": "orgy, debauch, debauchery, saturnalia, riot, bacchanal, bacchanalia, drunken revelry",
  "n00511555": "carnival",
  "n00511676": "Dionysia, Bacchanalia",
  "n00511817": "play, frolic, romp, gambol, caper",
  "n00512267": "caper, capriole",
  "n00512351": "capriole",
  "n00512522": "flirt, flirting, flirtation, coquetry, dalliance, toying",
  "n00512843": "folly, foolery, tomfoolery, craziness, lunacy, indulgence",
  "n00513089": "game",
  "n00513251": "meshugaas, mishegaas, mishegoss",
  "n00513401": "buffoonery, clowning, japery, frivolity, harlequinade, prank",
  "n00513597": "shtik, schtik, shtick, schtick",
  "n00513761": "horseplay",
  "n00513841": "teasing",
  "n00513952": "word play",
  "n00514041": "dirty trick",
  "n00514128": "practical joke",
  "n00514356": "April fool",
  "n00514471": "hotfoot",
  "n00514658": "rag",
  "n00514787": "snipe hunt",
  "n00515069": "drollery, waggery",
  "n00515159": "leg-pull, leg-pulling",
  "n00515297": "pleasantry",
  "n00515414": "beguilement, distraction",
  "n00515578": "edutainment",
  "n00515684": "extravaganza",
  "n00515791": "militainment",
  "n00515952": "nightlife, night life",
  "n00516086": "celebration, solemnization, solemnisation",
  "n00516381": "Isthmian Games",
  "n00516559": "Nemean Games",
  "n00516720": "Olympian Games, Olympic Games",
  "n00516900": "Pythian Games",
  "n00517080": "Royal National Eisteddfod",
  "n00517231": "eisteddfod",
  "n00517418": "film festival",
  "n00517564": "feria",
  "n00517728": "festival, fete",
  "n00518039": "jazz festival",
  "n00518152": "Kwanzaa, Kwanza",
  "n00518303": "Oktoberfest",
  "n00518442": "Saturnalia",
  "n00518555": "sheepshearing",
  "n00518669": "gala, gala affair, jamboree, blowout",
  "n00518770": "Ludi Saeculares, secular games",
  "n00519118": "victory celebration",
  "n00519251": "whoopee",
  "n00519333": "carnival, fair, funfair",
  "n00519492": "dog show",
  "n00519579": "horseshow",
  "n00519669": "raree-show",
  "n00519739": "circus",
  "n00519941": "three-ring circus",
  "n00520059": "Mardi Gras, Fat Tuesday",
  "n00520257": "show",
  "n00520672": "cabaret, floorshow, floor show",
  "n00520783": "ice show",
  "n00520880": "interlude, intermezzo, entr'acte",
  "n00521085": "parade",
  "n00521209": "display, exhibit, showing",
  "n00521450": "light show",
  "n00521562": "presentation, presentment, demonstration",
  "n00521970": "demonstration",
  "n00522145": "exhibition",
  "n00522338": "repudiation, debunking",
  "n00522537": "exposure",
  "n00522736": "production",
  "n00522939": "rodeo",
  "n00523078": "road show",
  "n00523148": "sideshow",
  "n00523263": "Wild West Show, Buffalo Bill's Wild West Show",
  "n00523513": "sport, athletics",
  "n00526259": "adagio",
  "n00526412": "break dancing, break dance",
  "n00526663": "courante",
  "n00526793": "nautch, nauch, nautch dance",
  "n00526948": "pavane, pavan",
  "n00527061": "phrase",
  "n00527198": "saraband",
  "n00527319": "skank",
  "n00527498": "slam dancing, slam dance",
  "n00527695": "step dancing, hoofing",
  "n00527872": "tap dancing, tap dance",
  "n00528059": "toe dancing, toe dance",
  "n00528156": "soft-shoe, soft-shoe shuffle, soft-shoe dancing",
  "n00528298": "buck-and-wing",
  "n00528397": "stage dancing, choreography",
  "n00528667": "ballet, concert dance",
  "n00529101": "pas seul, variation",
  "n00529224": "pas de deux, duet",
  "n00529400": "pas de trois",
  "n00529511": "pas de quatre",
  "n00529622": "classical ballet",
  "n00529784": "modern ballet",
  "n00529902": "comedy ballet",
  "n00530018": "modern dance",
  "n00530208": "clog dance, clog dancing, clog",
  "n00530386": "apache dance",
  "n00530554": "belly dance, belly dancing, danse du ventre",
  "n00530755": "bolero",
  "n00530874": "cakewalk",
  "n00531077": "cancan",
  "n00531201": "nude dancing",
  "n00531355": "fan dance",
  "n00531490": "strip, striptease, strip show",
  "n00531745": "bubble dance",
  "n00531886": "interpretive dance, interpretive dancing, interpretative dance, interpretative dancing",
  "n00532110": "social dancing",
  "n00532441": "jitterbug",
  "n00532567": "lindy, lindy hop",
  "n00532739": "fandango",
  "n00532896": "farandole",
  "n00533036": "flamenco, gypsy dancing",
  "n00533224": "gavotte",
  "n00533322": "habanera",
  "n00533404": "shag",
  "n00533535": "shimmy",
  "n00533691": "stomp",
  "n00533805": "tarantella, tarantelle",
  "n00533922": "dance step, step",
  "n00534152": "chasse, sashay",
  "n00534344": "glissade",
  "n00534480": "turnout",
  "n00534631": "twist",
  "n00534849": "ballroom dancing, ballroom dance",
  "n00535321": "beguine",
  "n00535455": "carioca",
  "n00535557": "cha-cha, cha-cha-cha",
  "n00535732": "one-step",
  "n00535873": "turkey trot",
  "n00535956": "fox-trot, foxtrot",
  "n00536120": "two-step",
  "n00536249": "bunny hug",
  "n00536359": "Charleston",
  "n00536516": "conga",
  "n00536655": "cotillion, cotilion",
  "n00536778": "minuet",
  "n00536872": "paso doble",
  "n00536964": "quickstep",
  "n00537085": "rumba, rhumba",
  "n00537192": "samba",
  "n00537297": "round dance, round dancing",
  "n00537423": "tango",
  "n00537534": "waltz, valse",
  "n00537682": "folk dancing, folk dance",
  "n00538052": "mazurka",
  "n00538144": "polka",
  "n00538266": "schottische",
  "n00538368": "morris dance, morris dancing",
  "n00538510": "sword dance, sword dancing",
  "n00538668": "mambo",
  "n00538791": "highland fling",
  "n00538876": "hornpipe",
  "n00538991": "jig",
  "n00539121": "country-dance, country dancing, contredanse, contra danse, contradance",
  "n00539510": "longways, longways dance",
  "n00539674": "Virginia reel, reel",
  "n00539827": "round dance, ring dance",
  "n00539951": "square dance, square dancing",
  "n00540211": "reel, Scottish reel",
  "n00540396": "eightsome",
  "n00540485": "quadrille",
  "n00540616": "lancers",
  "n00540701": "do-si-do",
  "n00540895": "promenade",
  "n00541035": "sashay",
  "n00541178": "swing",
  "n00541338": "landler",
  "n00541479": "ritual dancing, ritual dance, ceremonial dance",
  "n00541779": "rumba, rhumba",
  "n00541972": "apache devil dance",
  "n00542065": "corn dance",
  "n00542149": "danse macabre, dance of death",
  "n00542323": "ghost dance",
  "n00542458": "hula, hula-hula, Hawaiian dancing",
  "n00542584": "pyrrhic",
  "n00542711": "rain dance",
  "n00542841": "snake dance",
  "n00542992": "sun dance",
  "n00543114": "war dance",
  "n00543233": "music",
  "n00544441": "bell ringing, carillon, carillon playing",
  "n00544605": "change ringing",
  "n00544731": "instrumental music",
  "n00544842": "intonation",
  "n00545059": "percussion",
  "n00545194": "drumming",
  "n00545344": "vocal music",
  "n00545501": "singing, vocalizing",
  "n00546070": "a cappella singing, a capella singing",
  "n00546216": "bel canto",
  "n00546299": "coloratura",
  "n00546389": "song, strain",
  "n00546613": "carol",
  "n00546738": "lullaby, cradlesong",
  "n00546861": "caroling",
  "n00546972": "crooning",
  "n00547101": "crooning",
  "n00547244": "scat, scat singing",
  "n00547454": "whistling",
  "n00547616": "beat",
  "n00547765": "bow",
  "n00547960": "down-bow",
  "n00548068": "up-bow",
  "n00548173": "officiation",
  "n00548326": "acting, playing, playacting, performing",
  "n00548802": "portrayal, characterization, enactment, personation",
  "n00549106": "impression",
  "n00549284": "impersonation, personation",
  "n00549472": "apery, mimicry",
  "n00549610": "parody, mockery, takeoff",
  "n00549766": "method acting, method",
  "n00550016": "mime, pantomime, dumb show",
  "n00550242": "panto",
  "n00550341": "business, stage business, byplay",
  "n00550545": "shtik, schtik, shtick, schtick",
  "n00550771": "performance",
  "n00551215": "program, programme",
  "n00551407": "bill",
  "n00551508": "skit",
  "n00551585": "hamming, overacting",
  "n00551714": "heroics",
  "n00551913": "reenactment",
  "n00552097": "roleplaying",
  "n00552219": "card trick",
  "n00552312": "prestidigitation, sleight of hand",
  "n00552436": "liveliness, animation",
  "n00552568": "brouhaha",
  "n00552679": "circus, carnival",
  "n00552922": "disorganization, disorganisation",
  "n00553173": "disruption, perturbation",
  "n00553362": "dislocation, breakdown",
  "n00553655": "surprise, surprisal",
  "n00553823": "commotion, din, ruction, ruckus, rumpus, tumult",
  "n00553995": "furor, furore",
  "n00554107": "havoc, mayhem",
  "n00554200": "melee, scrimmage, battle royal",
  "n00554300": "agitation, excitement, turmoil, upheaval, hullabaloo",
  "n00554433": "outburst, tumultuous disturbance",
  "n00554541": "rampage, violent disorder",
  "n00554726": "wilding",
  "n00554850": "upset, derangement, overthrow",
  "n00555138": "bustle, hustle, flurry, ado, fuss, stir",
  "n00555325": "burst, fit",
  "n00555525": "fits and starts",
  "n00555648": "haste, hurry, rush, rushing",
  "n00555983": "dash, bolt",
  "n00556142": "scamper, scramble, scurry",
  "n00556313": "maneuver, manoeuvre, play",
  "n00556834": "takeaway",
  "n00556992": "figure",
  "n00557184": "figure eight",
  "n00557287": "spread eagle",
  "n00557419": "completion, pass completion",
  "n00557588": "play",
  "n00557865": "ball hawking",
  "n00558008": "assist",
  "n00558181": "icing, icing the puck",
  "n00558396": "power play",
  "n00558630": "football play",
  "n00558883": "run, running, running play, running game",
  "n00559329": "draw, draw play",
  "n00559555": "end run, sweep",
  "n00559724": "return",
  "n00559916": "reverse",
  "n00560141": "double reverse",
  "n00560293": "rush, rushing",
  "n00560529": "pass, passing play, passing game, passing",
  "n00560866": "power play",
  "n00561078": "handoff",
  "n00561226": "forward pass, aerial",
  "n00561376": "flare pass, flare",
  "n00561570": "screen pass",
  "n00561707": "lateral pass, lateral",
  "n00561820": "spot pass",
  "n00561985": "tackle",
  "n00562179": "jugglery",
  "n00562280": "obstruction",
  "n00562398": "blocking, block",
  "n00562643": "interference",
  "n00562823": "trap block",
  "n00562935": "check",
  "n00563097": "crosscheck",
  "n00563212": "poke check",
  "n00563330": "razzle-dazzle, razzle, razzmatazz, razmataz",
  "n00563494": "basketball play",
  "n00563665": "pick",
  "n00563823": "switch",
  "n00563998": "give-and-go",
  "n00564177": "baseball play",
  "n00564373": "double play",
  "n00564477": "triple play",
  "n00564583": "pick-off",
  "n00564710": "squeeze play",
  "n00564897": "suicide squeeze play, suicide squeeze",
  "n00565051": "safety squeeze play, safety squeeze",
  "n00565219": "footwork",
  "n00565302": "stroke, shot",
  "n00565809": "cut, undercut",
  "n00566040": "swipe",
  "n00566135": "tennis stroke, tennis shot",
  "n00566298": "return",
  "n00566690": "backhand, backhand stroke, backhand shot",
  "n00566889": "chop, chop shot",
  "n00567044": "drive",
  "n00567280": "drop shot, dink",
  "n00567418": "forehand, forehand stroke, forehand shot",
  "n00567685": "forehand drive",
  "n00567896": "get",
  "n00568057": "backhand drive",
  "n00568166": "two-handed backhand",
  "n00568286": "ground stroke",
  "n00568430": "serve, service",
  "n00568701": "ace",
  "n00568813": "fault",
  "n00569073": "let, net ball",
  "n00569234": "half volley",
  "n00569361": "lob",
  "n00569474": "overhead, smash",
  "n00569608": "passing shot",
  "n00569733": "volley",
  "n00569898": "stroke",
  "n00570066": "swimming stroke",
  "n00570366": "crawl, front crawl, Australian crawl",
  "n00570572": "dog paddle",
  "n00570683": "sidestroke",
  "n00570854": "butterfly, butterfly stroke",
  "n00571060": "breaststroke",
  "n00571283": "backstroke",
  "n00571444": "baseball swing, swing, cut",
  "n00571609": "golf stroke, golf shot, swing",
  "n00571956": "downswing",
  "n00572043": "slice, fade, slicing",
  "n00572285": "hook, draw, hooking",
  "n00572489": "drive, driving",
  "n00572706": "explosion",
  "n00572838": "putt, putting",
  "n00573106": "clock golf",
  "n00573268": "approach, approach shot",
  "n00573530": "chip, chip shot",
  "n00573664": "pitch, pitch shot",
  "n00573775": "sclaff",
  "n00573945": "shank",
  "n00574077": "teeoff",
  "n00574227": "swimming kick",
  "n00574430": "flutter kick",
  "n00574608": "frog kick",
  "n00574790": "dolphin kick",
  "n00574925": "scissors kick",
  "n00575083": "thrash",
  "n00575214": "treading water",
  "n00575365": "cinch, breeze, picnic, snap, duck soup, child's play, pushover, walkover, piece of cake",
  "n00575657": "doddle",
  "n00575741": "work",
  "n00576451": "action",
  "n00576717": "job",
  "n00576877": "job",
  "n00577068": "operation, procedure",
  "n00577357": "works, deeds",
  "n00577525": "service",
  "n00577749": "consulting service",
  "n00577931": "advisory service",
  "n00578154": "attestation service, attestation report",
  "n00578405": "financial audit",
  "n00578549": "facility",
  "n00578708": "laundering",
  "n00578795": "shining, polishing",
  "n00579082": "shoeshine",
  "n00579201": "national service",
  "n00579334": "utility",
  "n00579564": "service",
  "n00579834": "socage",
  "n00579977": "military service",
  "n00580081": "knight's service",
  "n00580190": "heavy lifting",
  "n00580370": "housecleaning",
  "n00580565": "housecleaning",
  "n00580880": "housewifery",
  "n00580961": "housework, housekeeping",
  "n00581090": "ironing",
  "n00581241": "workload, work load",
  "n00581382": "case load",
  "n00581517": "piecework",
  "n00581621": "busywork, make-work",
  "n00581772": "logging",
  "n00581885": "loose end, unfinished business",
  "n00581991": "nightwork",
  "n00582071": "paperwork",
  "n00582195": "welfare work, social service",
  "n00582388": "occupation, business, job, line of work, line",
  "n00582868": "occupation",
  "n00583089": "game, biz",
  "n00583246": "career, calling, vocation",
  "n00583461": "specialization, specialisation, specialty, speciality, specialism",
  "n00583747": "specialization, specialisation",
  "n00583933": "spiritualization, spiritualisation",
  "n00584160": "lifework",
  "n00584248": "walk of life, walk",
  "n00584367": "employment, work",
  "n00584769": "job",
  "n00584891": "service",
  "n00585018": "telecommuting, teleworking",
  "n00585174": "services",
  "n00585406": "facility",
  "n00585660": "public service",
  "n00585810": "minister",
  "n00585964": "cabinet minister",
  "n00586105": "appointment",
  "n00586262": "position, post, berth, office, spot, billet, place, situation",
  "n00588473": "academicianship",
  "n00588598": "accountantship",
  "n00588703": "admiralty",
  "n00588780": "ambassadorship",
  "n00588881": "apostleship",
  "n00588998": "apprenticeship",
  "n00589103": "associateship",
  "n00589217": "attorneyship",
  "n00589318": "bailiffship",
  "n00589415": "baronetage",
  "n00589494": "bishopry, episcopate",
  "n00589596": "cadetship",
  "n00589691": "caliphate",
  "n00589769": "captainship, captaincy",
  "n00589948": "cardinalship",
  "n00590047": "chairmanship",
  "n00590148": "chancellorship",
  "n00590269": "chaplaincy, chaplainship",
  "n00590383": "chieftaincy, chieftainship",
  "n00590518": "clerkship",
  "n00590626": "commandership, commandery",
  "n00590806": "comptrollership",
  "n00590913": "consulship",
  "n00591006": "controllership",
  "n00591111": "councillorship, councilorship",
  "n00591236": "counselorship, counsellorship",
  "n00591446": "curacy",
  "n00591523": "curatorship",
  "n00591622": "custodianship",
  "n00591725": "deanship, deanery",
  "n00591858": "directorship",
  "n00592001": "discipleship",
  "n00592102": "editorship",
  "n00592199": "eldership",
  "n00592292": "emirate",
  "n00592367": "fatherhood",
  "n00592446": "fatherhood",
  "n00592535": "foremanship",
  "n00592652": "generalship, generalcy",
  "n00592795": "governorship",
  "n00592894": "headmastership",
  "n00592999": "headmistressship",
  "n00593108": "headship",
  "n00593219": "headship",
  "n00593389": "hot seat",
  "n00593512": "incumbency",
  "n00593613": "inspectorship",
  "n00593732": "instructorship",
  "n00593837": "internship",
  "n00593944": "judgeship, judicature",
  "n00594070": "khanate",
  "n00594146": "lectureship",
  "n00594260": "legation, legateship",
  "n00594374": "legislatorship",
  "n00594477": "librarianship",
  "n00594580": "lieutenancy",
  "n00594738": "magistracy, magistrature",
  "n00594836": "managership",
  "n00594953": "manhood",
  "n00595032": "marshalship",
  "n00595146": "mastership",
  "n00595333": "mayoralty",
  "n00595410": "messiahship",
  "n00595545": "moderatorship",
  "n00595684": "overlordship",
  "n00595785": "pastorship, pastorate",
  "n00595894": "peasanthood",
  "n00596081": "plum",
  "n00596193": "praetorship",
  "n00596290": "precentorship",
  "n00596393": "preceptorship",
  "n00596496": "prefecture",
  "n00596592": "prelacy, prelature",
  "n00596692": "premiership",
  "n00596807": "presidency, presidentship",
  "n00597265": "President of the United States, President, Chief Executive",
  "n00597532": "primateship",
  "n00597629": "principalship",
  "n00597728": "priorship",
  "n00597821": "proconsulship, proconsulate",
  "n00597957": "proctorship",
  "n00598056": "professorship, chair",
  "n00598215": "protectorship",
  "n00598318": "public office",
  "n00598439": "bully pulpit",
  "n00598678": "rabbinate",
  "n00598767": "receivership",
  "n00598868": "rectorship, rectorate",
  "n00598970": "regency",
  "n00599064": "residency",
  "n00599234": "rulership",
  "n00599329": "sainthood",
  "n00599472": "secretaryship",
  "n00599917": "Attorney General, Attorney General of the United States",
  "n00600200": "Secretary of Agriculture, Agriculture Secretary",
  "n00600435": "Secretary of Commerce, Commerce Secretary",
  "n00600655": "Secretary of Defense, Defense Secretary",
  "n00600871": "Secretary of Education, Education Secretary",
  "n00601088": "Secretary of Energy, Energy Secretary",
  "n00601296": "Secretary of Health and Human Services",
  "n00601557": "Secretary of Housing and Urban Development",
  "n00601822": "Secretary of Labor, Labor Secretary",
  "n00602026": "Secretary of State",
  "n00602220": "Secretary of the Interior, Interior Secretary",
  "n00602448": "Secretary of the Treasury, Treasury Secretary",
  "n00602669": "Secretary of Transportation, Transportation Secretary",
  "n00602909": "Secretary of Veterans Affairs",
  "n00603131": "Secretary of War, War Secretary",
  "n00603319": "Secretary of the Navy, Navy Secretary",
  "n00603512": "Secretary of Commerce and Labor",
  "n00603684": "Secretary of Health Education and Welfare",
  "n00603866": "seigniory, seigneury, feudal lordship",
  "n00603995": "seismography",
  "n00604131": "senatorship",
  "n00604228": "sinecure",
  "n00604321": "solicitorship",
  "n00604424": "speakership",
  "n00604523": "stewardship",
  "n00604694": "studentship",
  "n00604811": "teachership",
  "n00604910": "thaneship",
  "n00605023": "throne",
  "n00605246": "treasurership",
  "n00605349": "tribuneship",
  "n00605430": "trusteeship",
  "n00605511": "vice-presidency",
  "n00605616": "viceroyship",
  "n00605715": "viziership",
  "n00605812": "wardenship",
  "n00605909": "wardership",
  "n00606006": "womanhood",
  "n00606119": "treadmill, salt mine",
  "n00606227": "business life, professional life",
  "n00606370": "trade, craft",
  "n00607075": "airplane mechanics",
  "n00607185": "auto mechanics",
  "n00607293": "basketry",
  "n00607374": "bookbinding",
  "n00607458": "bricklaying",
  "n00607542": "cabinetwork, cabinetry",
  "n00607775": "carpentry, woodworking, woodwork",
  "n00608037": "drafting, mechanical drawing",
  "n00608162": "dressmaking",
  "n00608265": "electrical work",
  "n00608354": "interior decoration, interior design",
  "n00608551": "furnishing",
  "n00608645": "lighting",
  "n00608789": "lumbering",
  "n00608896": "masonry",
  "n00609159": "oculism",
  "n00609236": "painting, house painting",
  "n00609423": "papermaking",
  "n00609506": "piloting, pilotage",
  "n00609669": "plumbing, plumbery",
  "n00609868": "pottery",
  "n00609953": "profession",
  "n00610222": "metier, medium",
  "n00610373": "learned profession",
  "n00610592": "literature",
  "n00610738": "architecture",
  "n00611143": "law, practice of law",
  "n00611433": "education",
  "n00611674": "journalism",
  "n00611874": "newspapering",
  "n00611972": "politics",
  "n00612160": "medicine, practice of medicine",
  "n00613282": "preventive medicine",
  "n00613473": "alternative medicine",
  "n00613685": "herbal medicine",
  "n00613819": "complementary medicine",
  "n00613973": "theology",
  "n00614224": "writing, committal to writing",
  "n00614489": "cryptography, coding, secret writing, steganography",
  "n00614730": "handwriting",
  "n00615011": "inscription",
  "n00615182": "notation",
  "n00615336": "superscription",
  "n00615462": "stenography",
  "n00615659": "subscription",
  "n00615887": "encoding, encryption",
  "n00616083": "compression",
  "n00616279": "image compression",
  "n00616398": "MPEG",
  "n00616641": "decompression",
  "n00616807": "data encryption",
  "n00616951": "recoding",
  "n00617059": "decoding, decryption, decipherment",
  "n00617337": "triangulation",
  "n00617601": "cabinetmaking, joinery",
  "n00617692": "pyrotechnics, pyrotechny",
  "n00617865": "shoemaking, shoe repairing, cobbling",
  "n00617989": "roofing",
  "n00618119": "sheet-metal work",
  "n00618264": "shingling",
  "n00618436": "tailoring",
  "n00618536": "tool-and-die work",
  "n00618642": "couture",
  "n00618734": "accountancy, accounting",
  "n00619011": "cost accounting",
  "n00619142": "costing",
  "n00619230": "bookkeeping, clerking",
  "n00619419": "single entry, single-entry bookkeeping",
  "n00619578": "double entry, double-entry bookkeeping",
  "n00619738": "inventory accounting",
  "n00619889": "inventory control",
  "n00620084": "first in first out, FIFO",
  "n00620258": "last in first out, LIFO",
  "n00620424": "butchery, butchering",
  "n00620554": "photography",
  "n00620752": "labor, labour, toil",
  "n00621198": "strikebreaking",
  "n00621326": "corvee",
  "n00621476": "drudgery, plodding, grind, donkeywork",
  "n00621627": "effort, elbow grease, exertion, travail, sweat",
  "n00622068": "struggle",
  "n00622266": "wrestle, wrestling, grapple, grappling, hand-to-hand struggle",
  "n00622584": "hunt, hunting",
  "n00622859": "hackwork",
  "n00622957": "haymaking",
  "n00623052": "haymaking",
  "n00623162": "manual labor, manual labour",
  "n00623370": "overwork, overworking",
  "n00623545": "slavery",
  "n00623670": "subbing, substituting",
  "n00623862": "trouble, difficulty",
  "n00624147": "the devil",
  "n00624285": "tsuris",
  "n00624436": "least effort, least resistance",
  "n00624553": "strain, straining",
  "n00624738": "exercise, exercising, physical exercise, physical exertion, workout",
  "n00625427": "pull",
  "n00625551": "conditioner",
  "n00625699": "set, exercise set",
  "n00625858": "aerobics, aerobic exercise",
  "n00625993": "bodybuilding, anaerobic exercise, muscle building, musclebuilding",
  "n00626188": "weightlift, weightlifting",
  "n00626423": "jerk",
  "n00626574": "bench press",
  "n00626728": "incline bench press",
  "n00626838": "clean and jerk, clean",
  "n00627013": "press, military press",
  "n00627196": "snatch",
  "n00627321": "weight gaining",
  "n00627437": "calisthenics, callisthenics",
  "n00627664": "calisthenics, callisthenics",
  "n00627849": "isometrics, isometric exercise",
  "n00628131": "isotonic exercise",
  "n00628390": "jogging",
  "n00628539": "Kegel exercises, pubococcygeus exercises",
  "n00628692": "stretch, stretching",
  "n00628883": "pandiculation",
  "n00629009": "power walking",
  "n00629176": "arm exercise",
  "n00629318": "pushup, press-up",
  "n00629496": "widegrip pushup",
  "n00629597": "pull-up, chin-up",
  "n00629803": "back exercise",
  "n00629911": "leg exercise",
  "n00630071": "knee bend, squat, squatting",
  "n00630268": "leg curl, leg curling",
  "n00630401": "leg extensor",
  "n00630526": "neck exercise",
  "n00630634": "stomach exercise, tummy crunch",
  "n00630786": "sit-up",
  "n00630960": "yoga",
  "n00631168": "hatha yoga",
  "n00631378": "practice",
  "n00631712": "consultancy",
  "n00631887": "cosmetology",
  "n00632017": "dental practice",
  "n00632104": "law practice",
  "n00632201": "medical practice",
  "n00632342": "family practice, family medicine",
  "n00632531": "group practice",
  "n00632716": "optometry",
  "n00632820": "private practice",
  "n00633108": "quackery, empiricism",
  "n00633329": "application, diligence",
  "n00633479": "overkill",
  "n00633613": "supererogation",
  "n00633717": "overexertion",
  "n00633864": "investigation, investigating",
  "n00634276": "analysis",
  "n00634586": "count, counting, numeration, enumeration, reckoning, tally",
  "n00635012": "police work, police investigation",
  "n00635205": "detection, detecting, detective work, sleuthing",
  "n00635422": "forensics",
  "n00635544": "roundup",
  "n00635699": "empiricism",
  "n00635850": "examination, scrutiny",
  "n00636461": "examination, testing",
  "n00636728": "inquiry, enquiry",
  "n00636921": "research",
  "n00637145": "eleven-plus, 11-plus",
  "n00637354": "search",
  "n00637546": "operations research",
  "n00637677": "means test",
  "n00637810": "inquest",
  "n00637912": "big science",
  "n00638080": "biological research",
  "n00638243": "cloning",
  "n00638448": "reproductive cloning",
  "n00638602": "human reproductive cloning",
  "n00638770": "somatic cell nuclear transplantation, somatic cell nuclear transfer, SCNT, nuclear transplantation",
  "n00639007": "therapeutic cloning, biomedical cloning",
  "n00639267": "stem-cell research",
  "n00639398": "embryonic stem-cell research",
  "n00639556": "experiment, experimentation",
  "n00639833": "field work",
  "n00639975": "testing",
  "n00640188": "marketing research, market research",
  "n00640434": "market analysis",
  "n00640560": "product research",
  "n00640721": "consumer research",
  "n00640889": "microscopy",
  "n00641109": "electron microscopy",
  "n00641222": "electron spin resonance, ESR, electron paramagnetic resonance",
  "n00641421": "trial and error",
  "n00641522": "probe",
  "n00641693": "Human Genome Project",
  "n00641820": "scientific research, research project",
  "n00642045": "endoscopy",
  "n00642325": "celioscopy",
  "n00642446": "colonoscopy",
  "n00642604": "culdoscopy",
  "n00642762": "gastroscopy",
  "n00642910": "hysteroscopy",
  "n00643067": "proctoscopy",
  "n00643208": "sigmoidoscopy, flexible sigmoidoscopy",
  "n00643389": "gonioscopy",
  "n00643621": "keratoscopy",
  "n00643772": "rhinoscopy",
  "n00643943": "scan",
  "n00644156": "scanning",
  "n00644366": "search",
  "n00644503": "survey, study",
  "n00644702": "testing",
  "n00644967": "screening",
  "n00645162": "genetic screening",
  "n00645365": "time and motion study, time-and-motion study, time-motion study, motion study, time study, work study",
  "n00645622": "dark ground illumination, dark field illumination",
  "n00645843": "fluorescence microscopy",
  "n00646024": "indirect immunofluorescence",
  "n00646187": "anatomy",
  "n00646332": "urinalysis, uranalysis",
  "n00646504": "scatology",
  "n00646680": "case study",
  "n00646833": "chemical analysis, qualitative analysis",
  "n00647147": "polarography",
  "n00647270": "quantitative analysis, quantitative chemical analysis",
  "n00647536": "colorimetry, colorimetric analysis",
  "n00647713": "volumetric analysis",
  "n00647919": "acidimetry",
  "n00648087": "alkalimetry",
  "n00648237": "titration",
  "n00648542": "volumetric analysis",
  "n00648692": "gravimetric analysis",
  "n00648790": "cost analysis",
  "n00648931": "dissection",
  "n00649090": "fundamental analysis, fundamentals analysis",
  "n00649266": "technical analysis, technical analysis of stock trends",
  "n00649482": "spectroscopy, spectrometry, spectroscopic analysis, spectrum analysis, spectrographic analysis",
  "n00649760": "dialysis",
  "n00649992": "apheresis, pheresis",
  "n00650267": "plasmapheresis",
  "n00650402": "plateletpheresis",
  "n00650543": "hemodialysis, haemodialysis",
  "n00650743": "mass spectroscopy",
  "n00650912": "microwave spectroscopy",
  "n00651088": "likening",
  "n00651176": "analogy",
  "n00651531": "collation",
  "n00651670": "confrontation",
  "n00651813": "contrast",
  "n00651954": "lighterage",
  "n00652069": "visitation",
  "n00652311": "site visit",
  "n00652466": "surveillance",
  "n00652659": "tabulation",
  "n00652803": "blood count",
  "n00652975": "complete blood count, CBC, blood profile",
  "n00653171": "differential blood count",
  "n00653388": "census, nose count, nosecount",
  "n00653518": "countdown",
  "n00653719": "miscount",
  "n00653811": "poll",
  "n00653958": "recount",
  "n00654113": "sperm count",
  "n00654234": "spying, undercover work",
  "n00654400": "wiretap, tap",
  "n00654563": "espionage",
  "n00654683": "counterespionage",
  "n00654765": "electronic surveillance",
  "n00654885": "care, attention, aid, tending",
  "n00655540": "maternalism",
  "n00655727": "babysitting, baby sitting",
  "n00655878": "pet sitting",
  "n00656033": "primary care",
  "n00656192": "aftercare",
  "n00656292": "dental care",
  "n00656386": "brush, brushing",
  "n00656524": "first aid",
  "n00656643": "eyedrop, eye-drop",
  "n00656835": "adrenergic agonist eyedrop",
  "n00657080": "beta blocker eyedrop",
  "n00657257": "miotic eyedrop",
  "n00657420": "topical prostaglandin eyedrop",
  "n00657604": "medical care, medical aid",
  "n00658082": "treatment, intervention",
  "n00658627": "hospitalization, hospitalisation, hospital care",
  "n00658796": "incubation",
  "n00658946": "livery",
  "n00659048": "massage",
  "n00659349": "cardiac massage, heart massage",
  "n00659667": "effleurage",
  "n00659834": "petrissage",
  "n00659946": "reflexology",
  "n00660173": "Swedish massage",
  "n00660324": "tapotement",
  "n00660570": "makeover",
  "n00660783": "manicure",
  "n00660957": "pedicure",
  "n00661091": "therapy",
  "n00661847": "modality",
  "n00662017": "diathermy",
  "n00662196": "aromatherapy",
  "n00662340": "chemotherapy",
  "n00662527": "correction",
  "n00662681": "electrotherapy, galvanism, electric healing, electrical healing",
  "n00662972": "heliotherapy, insolation",
  "n00663094": "hormone replacement therapy, hormone-replacement therapy, HRT",
  "n00663333": "immunotherapy",
  "n00663488": "infrared therapy",
  "n00663714": "inflation therapy",
  "n00663878": "iontophoresis, ionic medication, iontotherapy, electromotive drug administration, EMDA",
  "n00664110": "medication",
  "n00664285": "antipyresis",
  "n00664388": "megavitamin therapy",
  "n00664577": "occupational therapy",
  "n00664849": "nourishment",
  "n00664997": "nursing care",
  "n00665079": "nursing",
  "n00665221": "tender loving care, TLC",
  "n00665358": "nurturance",
  "n00665476": "personal care",
  "n00665679": "skin care, skincare",
  "n00665781": "facial",
  "n00665970": "adenoidectomy",
  "n00666107": "adrenalectomy, suprarenalectomy",
  "n00666232": "appendectomy, appendicectomy",
  "n00666350": "amputation",
  "n00666468": "angioplasty",
  "n00666600": "arthrodesis",
  "n00666733": "arthroplasty",
  "n00666884": "arthroscopy",
  "n00667096": "autoplasty",
  "n00667246": "brain surgery",
  "n00667384": "psychosurgery",
  "n00667612": "split-brain technique",
  "n00667847": "castration",
  "n00668112": "cautery, cauterization, cauterisation",
  "n00668404": "chemosurgery",
  "n00668552": "colostomy",
  "n00668736": "craniotomy",
  "n00668829": "cryosurgery",
  "n00669000": "cholecystectomy",
  "n00669155": "clitoridectomy, female circumcision",
  "n00669262": "laparoscopic cholecystectomy, lap choly",
  "n00669481": "curettage, curettement",
  "n00669687": "suction curettage, vacuum aspiration",
  "n00669932": "debridement",
  "n00670105": "decortication",
  "n00670250": "dilation and curettage, dilatation and curettage, D and C",
  "n00670703": "disembowelment, evisceration",
  "n00670897": "electrosurgery",
  "n00671022": "enterostomy, enterotomy",
  "n00671190": "enucleation",
  "n00671351": "operation, surgery, surgical operation, surgical procedure, surgical process",
  "n00673109": "wrong-site surgery",
  "n00673234": "embolectomy",
  "n00673347": "endarterectomy",
  "n00673494": "enervation",
  "n00673578": "evisceration",
  "n00673710": "exenteration",
  "n00673863": "eye operation, eye surgery",
  "n00674158": "face lift, facelift, lift, face lifting, cosmetic surgery, rhytidectomy, rhytidoplasty, nip and tuck",
  "n00674562": "fenestration",
  "n00674760": "gastrectomy",
  "n00674864": "gastroenterostomy",
  "n00675064": "gastrostomy",
  "n00675219": "heart surgery",
  "n00675357": "closed-heart surgery",
  "n00675540": "open-heart surgery",
  "n00675808": "coronary bypass, coronary bypass surgery, coronary artery bypass graft, CABG",
  "n00676160": "port-access coronary bypass surgery",
  "n00676453": "minimally invasive coronary bypass surgery",
  "n00676693": "hemorrhoidectomy, haemorrhoidectomy",
  "n00676834": "hemostasis, haemostasis, hemostasia, haemostasia",
  "n00677038": "hypophysectomy",
  "n00677174": "hysterectomy",
  "n00677299": "hysterotomy",
  "n00677434": "radical hysterectomy, panhysterectomy",
  "n00677613": "total hysterectomy",
  "n00677719": "gastromy",
  "n00677808": "implantation",
  "n00678010": "incision, section, surgical incision",
  "n00678425": "cataract surgery",
  "n00678632": "intracapsular surgery",
  "n00678752": "extracapsular surgery",
  "n00678988": "cyclodestructive surgery",
  "n00679140": "phacoemulsification",
  "n00679379": "filtration surgery",
  "n00679566": "iridectomy",
  "n00679724": "iridotomy",
  "n00679910": "keratotomy",
  "n00680018": "radial keratotomy",
  "n00680183": "laser-assisted subepithelial keratomileusis, LASEK",
  "n00680336": "laser trabecular surgery",
  "n00680511": "laser-assisted in situ keratomileusis, LASIK",
  "n00680658": "vitrectomy",
  "n00680803": "perineotomy",
  "n00680914": "episiotomy",
  "n00681071": "ileostomy",
  "n00681299": "intestinal bypass",
  "n00681441": "jejunostomy",
  "n00681613": "keratoplasty, corneal graft, corneal transplant",
  "n00681869": "epikeratophakia",
  "n00682080": "lipectomy, selective lipectomy",
  "n00682243": "liposuction, suction lipectomy",
  "n00682436": "mastopexy",
  "n00682538": "neuroplasty",
  "n00682625": "otoplasty",
  "n00682738": "laminectomy",
  "n00682856": "laparotomy",
  "n00683012": "laparoscopy",
  "n00683214": "laryngectomy",
  "n00683358": "lithotomy",
  "n00683470": "cholelithotomy",
  "n00683590": "lobectomy",
  "n00683722": "amygdalotomy",
  "n00683915": "callosotomy, callosectomy",
  "n00684128": "lobotomy, leukotomy, leucotomy, prefrontal lobotomy, prefrontal leukotomy, prefrontal leucotomy, frontal lobotomy",
  "n00684450": "transorbital lobotomy",
  "n00684648": "lumpectomy",
  "n00684837": "major surgery",
  "n00684988": "microsurgery",
  "n00685196": "robotic telesurgery",
  "n00685349": "minor surgery",
  "n00685508": "mastectomy",
  "n00685675": "modified radical mastectomy",
  "n00685837": "radical mastectomy",
  "n00686027": "simple mastectomy",
  "n00686168": "mastoidectomy",
  "n00686282": "meniscectomy",
  "n00686385": "nephrectomy",
  "n00686471": "neurectomy",
  "n00686570": "oophorectomy, ovariectomy",
  "n00686682": "oophorosalpingectomy",
  "n00686826": "ophthalmectomy",
  "n00686913": "orchidectomy, orchiectomy",
  "n00687027": "pancreatectomy",
  "n00687135": "pneumonectomy",
  "n00687252": "prostatectomy",
  "n00687365": "salpingectomy",
  "n00687472": "septectomy",
  "n00687619": "sigmoidectomy",
  "n00687769": "splenectomy",
  "n00687857": "stapedectomy",
  "n00687964": "sympathectomy",
  "n00688098": "thrombectomy",
  "n00688220": "thyroidectomy",
  "n00688317": "tonsillectomy",
  "n00688462": "myotomy",
  "n00688557": "myringectomy",
  "n00688647": "myringoplasty",
  "n00688766": "myringotomy",
  "n00688915": "neurosurgery",
  "n00689062": "nose job, rhinoplasty",
  "n00689186": "orchiopexy",
  "n00689302": "orchotomy",
  "n00689467": "osteotomy",
  "n00689550": "ostomy",
  "n00689729": "palatopharyngoplasty, PPP, uvulopalatopharyngoplasty, UPPP",
  "n00689995": "phalloplasty",
  "n00690132": "phlebectomy",
  "n00690281": "photocoagulation",
  "n00690501": "plastic surgery, reconstructive surgery, anaplasty",
  "n00690848": "polypectomy",
  "n00690933": "proctoplasty, rectoplasty",
  "n00691050": "resection",
  "n00691190": "rhinotomy",
  "n00691330": "rhizotomy",
  "n00691541": "sclerotomy",
  "n00691648": "sex-change operation, transsexual surgery",
  "n00691879": "Shirodkar's operation, purse-string operation",
  "n00692130": "sterilization, sterilisation",
  "n00692349": "castration, emasculation",
  "n00692506": "neutering, fixing, altering",
  "n00692726": "spaying",
  "n00692840": "strabotomy",
  "n00692991": "taxis",
  "n00693109": "Michelson-Morley experiment",
  "n00693399": "tracheostomy, tracheotomy",
  "n00693679": "transplant, transplantation, organ transplant",
  "n00694088": "transurethral resection of the prostate, TURP",
  "n00694276": "trephination",
  "n00694420": "tympanoplasty",
  "n00694578": "uranoplasty",
  "n00694681": "justice",
  "n00694866": "administration, judicature",
  "n00694990": "administration, giving medication",
  "n00695160": "drip feed",
  "n00695300": "sedation, drugging",
  "n00695448": "irrigation",
  "n00695717": "douche",
  "n00695944": "enema, clyster",
  "n00696147": "colonic irrigation, colonic",
  "n00696300": "barium enema",
  "n00696487": "lavage",
  "n00696650": "gastric lavage",
  "n00696882": "dressing, bandaging, binding",
  "n00697024": "holistic medicine",
  "n00697249": "hospice",
  "n00697365": "injection, shot",
  "n00697614": "cryocautery",
  "n00697734": "electrocautery",
  "n00697887": "thermocautery",
  "n00698004": "bloodletting",
  "n00698234": "nephrotomy",
  "n00698348": "thoracotomy",
  "n00698474": "valvotomy, valvulotomy",
  "n00698609": "venesection, phlebotomy",
  "n00698794": "cupping",
  "n00698959": "defibrillation",
  "n00699146": "detoxification",
  "n00699320": "detoxification",
  "n00699517": "fusion, spinal fusion",
  "n00699736": "faith healing, faith cure",
  "n00699872": "laying on of hands",
  "n00700000": "physical therapy, physiotherapy, physiatrics",
  "n00700260": "rehabilitation",
  "n00700421": "phytotherapy, herbal therapy, botanical medicine",
  "n00700652": "psychotherapy",
  "n00700979": "behavior therapy, behavior modification",
  "n00701325": "assertiveness training",
  "n00701491": "aversion therapy",
  "n00701755": "desensitization technique, desensitisation technique, desensitization procedure, desensitisation procedure, systematic desensitization, systematic desensitisation",
  "n00702202": "exposure therapy",
  "n00702418": "implosion therapy, flooding",
  "n00702659": "reciprocal inhibition, reciprocal-inhibition therapy",
  "n00702992": "token economy",
  "n00703236": "client-centered therapy",
  "n00703422": "crisis intervention",
  "n00703688": "group therapy, group psychotherapy",
  "n00703926": "family therapy",
  "n00704064": "hypnotherapy",
  "n00704159": "play therapy",
  "n00704305": "psychoanalysis, analysis, depth psychology",
  "n00705319": "hypnoanalysis",
  "n00705433": "self-analysis",
  "n00705580": "radiotherapy, radiation therapy, radiation, actinotherapy, irradiation",
  "n00705886": "phototherapy",
  "n00706019": "radium therapy, Curietherapy",
  "n00706133": "X-ray therapy",
  "n00706222": "chrysotherapy",
  "n00706371": "shock therapy, shock treatment",
  "n00706605": "electroconvulsive therapy, electroshock, electroshock therapy, ECT",
  "n00706847": "insulin shock, insulin shock therapy, insulin shock treatment",
  "n00707028": "metrazol shock, metrazol shock therapy, metrazol shock treatment",
  "n00707213": "speech therapy",
  "n00707344": "refrigeration",
  "n00707537": "thermotherapy",
  "n00707785": "thrombolytic therapy",
  "n00707967": "chiropractic",
  "n00708168": "fomentation",
  "n00708332": "naturopathy",
  "n00708489": "naprapathy",
  "n00708676": "orthodontic treatment",
  "n00708957": "orthoptics",
  "n00709205": "osteopathy",
  "n00709395": "osteoclasis",
  "n00709521": "disinfection",
  "n00709659": "chlorination",
  "n00709843": "digitalization, digitalisation",
  "n00710015": "anticoagulation",
  "n00710155": "acupuncture, stylostixis",
  "n00710338": "acupressure, G-Jo, shiatsu",
  "n00710550": "autogenic therapy, autogenic training, autogenics",
  "n00710692": "allopathy",
  "n00710889": "homeopathy, homoeopathy",
  "n00711127": "hydropathy, hydrotherapy",
  "n00711265": "intensive care",
  "n00711413": "rest-cure",
  "n00711523": "stalk, stalking, still hunt",
  "n00711727": "deerstalking",
  "n00711799": "birdnesting",
  "n00711897": "predation",
  "n00712031": "friction, detrition, rubbing",
  "n00712225": "application, coating, covering",
  "n00712833": "anointing, anointment",
  "n00712979": "fumigation",
  "n00713135": "foliation",
  "n00713250": "galvanization, galvanisation",
  "n00713493": "bodywork",
  "n00713594": "handling",
  "n00713784": "materials handling",
  "n00713952": "loading",
  "n00714173": "unloading",
  "n00714357": "picking",
  "n00714477": "pickings, taking",
  "n00714693": "planking",
  "n00714806": "wiring",
  "n00714944": "handicraft",
  "n00715674": "sewing, stitching",
  "n00715843": "baking",
  "n00715949": "blind stitching",
  "n00716055": "suturing",
  "n00716179": "vasectomy",
  "n00716397": "vasotomy, vasosection",
  "n00716520": "vasovasostomy",
  "n00716676": "vulvectomy",
  "n00716777": "vivisection",
  "n00716945": "lubrication",
  "n00717081": "paving, pavage",
  "n00717208": "painting",
  "n00717468": "spraying",
  "n00717640": "spray painting",
  "n00717748": "spatter, spattering, splash, splashing, splattering",
  "n00718066": "finger-painting",
  "n00718195": "tinning, tin-plating",
  "n00718325": "tinning",
  "n00718573": "papering, paperhanging",
  "n00718689": "pargeting, pargetting",
  "n00718815": "plastering, daubing",
  "n00719016": "plating",
  "n00719148": "scumble",
  "n00719274": "tiling",
  "n00719388": "waxing",
  "n00719494": "duty",
  "n00719705": "job, task, chore",
  "n00720166": "ball-buster, ball-breaker",
  "n00720331": "paper route",
  "n00720431": "stint",
  "n00720565": "function, office, part, role",
  "n00720951": "capacity",
  "n00721141": "hat",
  "n00721282": "portfolio",
  "n00721431": "stead, position, place, lieu",
  "n00721660": "behalf",
  "n00721929": "second fiddle",
  "n00722061": "role",
  "n00722253": "gender role",
  "n00722479": "position",
  "n00722984": "pitcher, mound",
  "n00723241": "catcher",
  "n00723547": "first base, first",
  "n00723783": "second base, second",
  "n00723984": "shortstop, short",
  "n00724168": "third base, third",
  "n00724433": "left field, leftfield",
  "n00724684": "center field, centerfield",
  "n00724898": "right field, rightfield",
  "n00725152": "steal",
  "n00725383": "forward",
  "n00725506": "center",
  "n00725673": "guard",
  "n00725775": "back",
  "n00726100": "lineman",
  "n00726369": "linebacker, line backer",
  "n00726567": "quarterback, signal caller, field general",
  "n00726867": "fullback",
  "n00727002": "halfback",
  "n00727137": "tailback",
  "n00727333": "wingback",
  "n00727500": "center",
  "n00727743": "guard",
  "n00727901": "tackle",
  "n00728065": "end",
  "n00728218": "mid-off",
  "n00728342": "mid-on",
  "n00728464": "center",
  "n00728641": "school assignment, schoolwork",
  "n00728849": "classroom project",
  "n00728975": "classwork",
  "n00729108": "homework, prep, preparation",
  "n00729285": "lesson",
  "n00729478": "language lesson",
  "n00729637": "French lesson",
  "n00729731": "German lesson",
  "n00729825": "Hebrew lesson",
  "n00729919": "exercise, example",
  "n00730134": "reading assignment",
  "n00730247": "assignment, duty assignment",
  "n00730538": "guard duty, guard, sentry duty, sentry go",
  "n00730708": "fatigue duty, fatigue",
  "n00730984": "mission, missionary work",
  "n00731136": "da'wah, dawah",
  "n00731222": "mission, charge, commission",
  "n00731569": "fool's errand",
  "n00731648": "mission impossible",
  "n00731756": "suicide mission, martyr operation, sacrifice operation",
  "n00731947": "errand",
  "n00732076": "reassignment",
  "n00732201": "secondment",
  "n00732369": "sea-duty, service abroad, shipboard duty",
  "n00732492": "shore duty",
  "n00732576": "scut work, shitwork",
  "n00732746": "wrongdoing, wrongful conduct, misconduct, actus reus",
  "n00733317": "brutalization, brutalisation",
  "n00733483": "trespass, encroachment, violation, intrusion, usurpation",
  "n00733753": "inroad",
  "n00733883": "tort, civil wrong",
  "n00734107": "alienation of affection",
  "n00734303": "invasion of privacy",
  "n00734482": "trespass",
  "n00734783": "continuing trespass",
  "n00735017": "trespass de bonis asportatis",
  "n00735198": "trespass on the case",
  "n00735553": "trespass quare clausum fregit",
  "n00735686": "trespass viet armis",
  "n00735832": "malversation",
  "n00735936": "misbehavior, misbehaviour, misdeed",
  "n00736219": "delinquency, juvenile delinquency",
  "n00736375": "mischief, mischief-making, mischievousness, deviltry, devilry, devilment, rascality, roguery, roguishness, shenanigan",
  "n00736786": "hell, blaze",
  "n00736894": "monkey business",
  "n00736989": "ruffianism",
  "n00737070": "familiarity, impropriety, indecorum, liberty",
  "n00737188": "abnormality, irregularity",
  "n00737399": "deviation, deviance",
  "n00737536": "indecency, impropriety",
  "n00737705": "paraphilia",
  "n00737894": "exhibitionism, immodesty",
  "n00738058": "fetishism, fetichism",
  "n00738270": "pedophilia, paedophilia",
  "n00738372": "voyeurism",
  "n00738598": "zoophilia, zoophilism",
  "n00738697": "obscenity",
  "n00738785": "indiscretion, peccadillo",
  "n00738872": "infantilism",
  "n00738966": "dereliction",
  "n00739078": "nonfeasance",
  "n00739270": "negligence, carelessness, neglect, nonperformance",
  "n00739632": "comparative negligence",
  "n00739850": "concurrent negligence",
  "n00740048": "contributory negligence",
  "n00740342": "criminal negligence, culpable negligence",
  "n00740609": "neglect of duty",
  "n00740712": "evasion, escape, dodging",
  "n00741158": "escape mechanism",
  "n00741272": "malingering, skulking",
  "n00741478": "shirking, slacking, soldiering, goofing off, goldbricking",
  "n00741685": "circumvention",
  "n00741797": "tax evasion",
  "n00741925": "malfeasance",
  "n00742020": "misfeasance",
  "n00742130": "malpractice",
  "n00742314": "malpractice",
  "n00742474": "perversion",
  "n00742645": "waste, wastefulness, dissipation",
  "n00743155": "waste of effort, waste of energy",
  "n00743251": "waste of material",
  "n00743348": "waste of money",
  "n00743500": "waste of time",
  "n00743641": "extravagance, prodigality, lavishness, highlife, high life",
  "n00743822": "squandering",
  "n00744004": "squandermania",
  "n00744131": "wrong, legal injury, damage",
  "n00744305": "injury",
  "n00744443": "injury",
  "n00744616": "injustice, unfairness, iniquity, shabbiness",
  "n00744758": "infliction",
  "n00744862": "transgression",
  "n00745005": "transgression, evildoing",
  "n00745431": "abomination",
  "n00745637": "evil, immorality, wickedness, iniquity",
  "n00745863": "villainy",
  "n00745943": "deviltry, devilry",
  "n00746033": "enormity",
  "n00746116": "foul play",
  "n00746232": "irreverence, violation",
  "n00746375": "sexual immorality",
  "n00746587": "profanation, desecration, blasphemy, sacrilege",
  "n00746866": "depravity, turpitude",
  "n00747029": "vice",
  "n00747215": "pornography, porno, porn, erotica, smut",
  "n00747519": "child pornography, kiddie porn, kiddy porn",
  "n00747671": "intemperance, intemperateness, self-indulgence",
  "n00748011": "intemperance, intemperateness",
  "n00748155": "prostitution, harlotry, whoredom",
  "n00748307": "profligacy, dissipation, dissolution, licentiousness, looseness",
  "n00748515": "drink, drinking, boozing, drunkenness, crapulence",
  "n00748834": "drinking bout",
  "n00748919": "package tour, package holiday",
  "n00749105": "pub crawl",
  "n00749232": "whistle-stop tour",
  "n00749491": "jag",
  "n00749574": "dishonesty, knavery",
  "n00749767": "treachery, betrayal, treason, perfidy",
  "n00749991": "double cross, double-crossing",
  "n00750216": "sellout",
  "n00750306": "charlatanism, quackery",
  "n00750405": "plagiarism, plagiarization, plagiarisation, piracy",
  "n00750730": "trick",
  "n00750890": "falsification, falsehood",
  "n00751145": "falsification, misrepresentation",
  "n00751398": "frame-up, setup",
  "n00751529": "distortion, overrefinement, straining, torture, twisting",
  "n00751779": "equivocation, tergiversation",
  "n00751944": "lying, prevarication, fabrication",
  "n00752144": "fibbing, paltering",
  "n00752298": "fakery",
  "n00752431": "deception, deceit, dissembling, dissimulation",
  "n00752798": "indirection",
  "n00752954": "trickery, chicanery, chicane, guile, wile, shenanigan",
  "n00753240": "duplicity, double-dealing",
  "n00753472": "sophistication",
  "n00753685": "fraud, fraudulence, dupery, hoax, humbug, put-on",
  "n00753973": "goldbrick",
  "n00754118": "jugglery",
  "n00754280": "scam, cozenage",
  "n00754424": "cheat, cheating",
  "n00754593": "gerrymander",
  "n00754767": "delusion, illusion, head game",
  "n00754956": "pretense, pretence, pretending, simulation, feigning",
  "n00755277": "appearance, show",
  "n00755500": "make-believe, pretend",
  "n00755673": "affectation, mannerism, pose, affectedness",
  "n00755863": "attitude",
  "n00756034": "radical chic",
  "n00756194": "masquerade",
  "n00756331": "imposture, impersonation",
  "n00756470": "obscurantism",
  "n00756598": "bluff, four flush",
  "n00756780": "take-in",
  "n00756919": "fall",
  "n00757080": "sin, sinning",
  "n00757337": "actual sin",
  "n00757483": "original sin",
  "n00757730": "mortal sin, deadly sin",
  "n00758033": "venial sin",
  "n00758175": "pride, superbia",
  "n00758335": "envy, invidia",
  "n00758525": "avarice, greed, covetousness, rapacity, avaritia",
  "n00758795": "sloth, laziness, acedia",
  "n00758972": "wrath, anger, ire, ira",
  "n00759186": "gluttony, overeating, gula",
  "n00759335": "lust, luxuria",
  "n00759500": "terror",
  "n00759694": "terrorism, act of terrorism, terrorist act",
  "n00763132": "bioterrorism, biological terrorism",
  "n00763282": "chemical terrorism",
  "n00763630": "cyber-terrorism, cyberwar",
  "n00763787": "domestic terrorism",
  "n00764031": "ecoterrorism, ecological terrorism, eco-warfare, ecological warfare",
  "n00764258": "international terrorism",
  "n00764436": "narcoterrorism",
  "n00764588": "nuclear terrorism",
  "n00764891": "state-sponsored terrorism",
  "n00765081": "theoterrorism",
  "n00765193": "terrorization, terrorisation",
  "n00765343": "barratry",
  "n00765488": "champerty",
  "n00765791": "maintenance, criminal maintenance",
  "n00766234": "crime, offense, criminal offense, criminal offence, offence, law-breaking",
  "n00767477": "crime",
  "n00767633": "inside job",
  "n00767826": "assault",
  "n00768063": "aggravated assault",
  "n00768203": "battery, assault and battery",
  "n00768353": "capital offense",
  "n00768483": "cybercrime",
  "n00768701": "felony",
  "n00768921": "forgery",
  "n00769092": "fraud",
  "n00769401": "barratry",
  "n00769695": "Had crime",
  "n00769944": "hijack, highjack",
  "n00770151": "mayhem",
  "n00770270": "misdemeanor, misdemeanour, infraction, violation, infringement",
  "n00770543": "violation, infringement",
  "n00770834": "copyright infringement, infringement of copyright",
  "n00770997": "foul",
  "n00771198": "personal foul",
  "n00771356": "technical foul, technical",
  "n00771604": "patent infringement",
  "n00771713": "disorderly conduct, disorderly behavior, disturbance of the peace, breach of the peace",
  "n00772026": "false pretense, false pretence",
  "n00772253": "indecent exposure, public nudity",
  "n00772381": "perjury, bearing false witness, lying under oath",
  "n00772550": "resisting arrest",
  "n00772813": "sedition",
  "n00773047": "molestation",
  "n00773235": "perpetration, commission, committal",
  "n00773402": "rape, violation, assault, ravishment",
  "n00773673": "date rape",
  "n00773814": "attack, attempt",
  "n00774009": "mugging",
  "n00774107": "sexual assault, sexual abuse, sex crime, sex offense",
  "n00774506": "Tazir crime",
  "n00774796": "statutory offense, statutory offence, regulatory offense, regulatory offence",
  "n00775006": "thuggery",
  "n00775095": "bigamy",
  "n00775286": "capture, seizure",
  "n00775460": "abduction",
  "n00775702": "kidnapping, snatch",
  "n00775943": "racket, fraudulent scheme, illegitimate enterprise",
  "n00776165": "racketeering",
  "n00776262": "bribery, graft",
  "n00776466": "barratry",
  "n00776581": "commercial bribery",
  "n00776732": "embezzlement, peculation, defalcation, misapplication, misappropriation",
  "n00777069": "identity theft",
  "n00777324": "raid",
  "n00777439": "plunderage",
  "n00777587": "mail fraud",
  "n00777679": "election fraud",
  "n00777806": "constructive fraud, legal fraud",
  "n00778017": "extrinsic fraud, collateral fraud",
  "n00778213": "fraud in fact, positive fraud",
  "n00778405": "fraud in law",
  "n00778575": "fraud in the factum",
  "n00778809": "fraud in the inducement",
  "n00779035": "intrinsic fraud",
  "n00779248": "bunco, bunco game, bunko, bunko game, con, confidence trick, confidence game, con game, gyp, hustle, sting, flimflam",
  "n00779599": "sting operation",
  "n00779809": "pyramiding",
  "n00780148": "swindle, cheat, rig",
  "n00780466": "holdout",
  "n00780615": "swiz",
  "n00780731": "shell game, thimblerig",
  "n00780889": "larceny, theft, thievery, thieving, stealing",
  "n00781355": "pilferage",
  "n00781480": "shoplifting, shrinkage",
  "n00781685": "robbery",
  "n00781912": "armed robbery, heist, holdup, stickup",
  "n00782072": "treason, high treason, lese majesty",
  "n00782241": "vice crime",
  "n00782338": "victimless crime",
  "n00782518": "war crime",
  "n00782629": "biopiracy",
  "n00782792": "caper, job",
  "n00782927": "dacoity, dakoity",
  "n00783063": "heist, rip-off",
  "n00783199": "highjacking, hijacking",
  "n00783419": "highway robbery",
  "n00783527": "piracy, buccaneering",
  "n00783762": "rolling",
  "n00783902": "grand larceny, grand theft",
  "n00784083": "petit larceny, petty larceny, petty",
  "n00784271": "skimming",
  "n00784388": "extortion",
  "n00784598": "blackmail",
  "n00784755": "protection, tribute",
  "n00784934": "shakedown",
  "n00785045": "burglary",
  "n00785263": "housebreaking, break-in, breaking and entering",
  "n00785482": "home invasion",
  "n00785596": "joint venture",
  "n00785818": "foreign direct investment",
  "n00785959": "experiment",
  "n00786102": "forlorn hope",
  "n00786195": "attempt, effort, endeavor, endeavour, try",
  "n00786887": "bid, play",
  "n00787061": "crack, fling, go, pass, whirl, offer",
  "n00787218": "essay",
  "n00787307": "foray",
  "n00787465": "contribution, part, share",
  "n00787727": "end",
  "n00787832": "liberation",
  "n00788007": "mug's game",
  "n00788097": "power play, squeeze play, squeeze",
  "n00788362": "seeking",
  "n00788473": "shot, stab",
  "n00788632": "shot",
  "n00788766": "striving, nisus, pains, strain",
  "n00788973": "struggle, battle",
  "n00789237": "duel",
  "n00789391": "scramble, scuffle",
  "n00789534": "buyout",
  "n00789735": "strategic buyout",
  "n00789906": "takeover",
  "n00790086": "anti-takeover defense",
  "n00790205": "takeover attempt",
  "n00790347": "takeover bid",
  "n00790498": "two-tier bid",
  "n00790677": "any-and-all bid",
  "n00790817": "hostile takeover",
  "n00790947": "friendly takeover",
  "n00791078": "test, trial, run",
  "n00791527": "assay",
  "n00791875": "enzyme-linked-immunosorbent serologic assay, ELISA",
  "n00792142": "immunohistochemistry",
  "n00792356": "clinical trial, clinical test",
  "n00792720": "phase I clinical trial, phase I",
  "n00792954": "phase II clinical trial, phase II",
  "n00793212": "phase III clinical trial, phase III",
  "n00793529": "phase IV clinical trial, phase IV",
  "n00793844": "double blind",
  "n00794123": "preclinical trial, preclinical test, preclinical phase",
  "n00794367": "test, trial",
  "n00794614": "audition, tryout",
  "n00794772": "screen test",
  "n00794870": "field trial",
  "n00795008": "fitting, try-on, trying on",
  "n00795161": "MOT, MOT test, Ministry of Transportation test",
  "n00795352": "pilot project, pilot program",
  "n00795508": "spadework",
  "n00795625": "timework",
  "n00795720": "undertaking, project, task, labor",
  "n00796197": "written assignment, writing assignment",
  "n00796315": "adventure, escapade, risky venture, dangerous undertaking",
  "n00796586": "assignment",
  "n00796767": "baby",
  "n00796886": "enterprise, endeavor, endeavour",
  "n00797207": "labor of love, labour of love",
  "n00797361": "marathon, endurance contest",
  "n00797468": "no-brainer",
  "n00797581": "proposition",
  "n00797723": "tall order, large order",
  "n00797878": "venture",
  "n00798108": "speleology, spelaeology",
  "n00798245": "campaign, cause, crusade, drive, movement, effort",
  "n00798959": "advertising campaign, ad campaign, ad blitz",
  "n00799125": "anti-war movement",
  "n00799236": "charm campaign",
  "n00799409": "consumerism",
  "n00799537": "campaigning, candidacy, candidature, electioneering, political campaign",
  "n00799809": "front-porch campaigning, front-porch campaign",
  "n00800121": "hustings",
  "n00800270": "fund-raising campaign, fund-raising drive, fund-raising effort",
  "n00800421": "feminist movement, feminism, women's liberation movement, women's lib",
  "n00800657": "gay liberation movement, gay lib",
  "n00800825": "lost cause",
  "n00800940": "reform",
  "n00801125": "war",
  "n00801277": "youth movement, youth crusade",
  "n00801450": "whispering campaign",
  "n00801598": "stumping",
  "n00801724": "sales campaign",
  "n00801834": "public-relations campaign",
  "n00801966": "sally, sallying forth",
  "n00802111": "self-help",
  "n00802238": "risk, peril, danger",
  "n00802629": "chance",
  "n00802785": "crapshoot",
  "n00802962": "gamble",
  "n00803091": "long shot",
  "n00803208": "raise",
  "n00803394": "doubling, double",
  "n00803617": "control",
  "n00804180": "crowd control",
  "n00804271": "damage control",
  "n00804379": "federalization, federalisation",
  "n00804535": "flight control",
  "n00804708": "flood control",
  "n00804909": "imperialism",
  "n00805034": "regulation, regulating",
  "n00805337": "deregulation, deregulating",
  "n00805524": "devaluation",
  "n00805766": "gun control",
  "n00805868": "indexation",
  "n00806075": "internal control",
  "n00806366": "management control",
  "n00806484": "quality control",
  "n00806621": "acceptance sampling",
  "n00806902": "regulation, regularization, regularisation",
  "n00807078": "timing",
  "n00807273": "coordination",
  "n00807500": "synchronization, synchronisation, synchronizing",
  "n00807768": "load-shedding",
  "n00807925": "proration",
  "n00808182": "limitation, restriction",
  "n00808485": "arms control",
  "n00808614": "hold-down",
  "n00808767": "freeze",
  "n00808967": "clampdown",
  "n00809074": "hire",
  "n00809201": "hiring freeze",
  "n00809279": "price freeze",
  "n00809373": "wage freeze",
  "n00809465": "possession, ownership",
  "n00809702": "possession",
  "n00809886": "actual possession",
  "n00810026": "constructive possession",
  "n00810234": "criminal possession",
  "n00810498": "illegal possession",
  "n00810598": "retention, keeping, holding",
  "n00810794": "withholding",
  "n00811036": "power trip",
  "n00811221": "defecation reflex, rectal reflex",
  "n00811355": "storage",
  "n00811491": "filing",
  "n00811661": "storage",
  "n00811825": "cold storage",
  "n00811921": "stowage, stowing",
  "n00812055": "tankage",
  "n00812154": "riot control, riot control operation",
  "n00812274": "grasping, taking hold, seizing, prehension",
  "n00812526": "clasp, clench, clutch, clutches, grasp, grip, hold",
  "n00812977": "wrestling hold",
  "n00813220": "bear hug",
  "n00813336": "nelson",
  "n00813571": "full nelson",
  "n00813800": "half nelson",
  "n00813975": "hammerlock",
  "n00814106": "headlock",
  "n00814274": "Japanese stranglehold",
  "n00814458": "lock",
  "n00814665": "scissors, scissors hold, scissor hold, scissor grip, scissors grip",
  "n00814899": "stranglehold",
  "n00815036": "toehold",
  "n00815173": "steering, steerage",
  "n00815320": "steering, guidance, direction",
  "n00815644": "aim",
  "n00815801": "navigation, pilotage, piloting",
  "n00816419": "instrument flying",
  "n00816530": "celestial navigation, astronavigation",
  "n00816665": "celestial guidance",
  "n00816841": "inertial guidance, inertial navigation",
  "n00817017": "command guidance",
  "n00817191": "terrestrial guidance",
  "n00817394": "dead reckoning",
  "n00817507": "fire watching",
  "n00817680": "protection",
  "n00818170": "air cover",
  "n00818362": "shielding",
  "n00818466": "guardianship, keeping, safekeeping",
  "n00818678": "hands, custody",
  "n00819024": "preservation, saving",
  "n00819274": "conservation",
  "n00819524": "conservancy",
  "n00819641": "soil conservation",
  "n00819756": "oil conservation",
  "n00819858": "water conservation",
  "n00819958": "self-preservation",
  "n00820091": "reservation",
  "n00820232": "Manhattan Project",
  "n00820406": "embalmment",
  "n00820583": "mummification",
  "n00820721": "momism, overprotection, overshielding",
  "n00820845": "security intelligence",
  "n00821083": "censoring, censorship, security review",
  "n00821295": "military censorship",
  "n00821518": "civil censorship",
  "n00821752": "field press censorship",
  "n00821973": "prisoner of war censorship",
  "n00822159": "armed forces censorship",
  "n00822350": "primary censorship",
  "n00822544": "secondary censorship",
  "n00822755": "national censorship",
  "n00822970": "precaution, safeguard, guard",
  "n00823316": "security, security measures",
  "n00823532": "defense, defence",
  "n00823750": "defense, defence",
  "n00823884": "inoculation, vaccination",
  "n00824054": "inoculating, vaccinating",
  "n00824357": "ring vaccination",
  "n00824614": "variolation, variolization",
  "n00824808": "patrol",
  "n00825005": "airborne patrol",
  "n00825096": "round-the-clock patrol",
  "n00825192": "self-defense, self-defence, self-protection",
  "n00825331": "aikido",
  "n00825443": "martial art",
  "n00825773": "judo",
  "n00825951": "jujutsu, jujitsu, jiujitsu",
  "n00826213": "ninjutsu, ninjitsu",
  "n00826397": "karate",
  "n00826606": "kung fu",
  "n00826681": "tae kwon do, taekwondo",
  "n00826789": "t'ai chi, tai chi, t'ai chi chuan, tai chi chuan, taichi, taichichuan",
  "n00827010": "insulation",
  "n00827240": "lining",
  "n00827379": "lining, facing",
  "n00827535": "babbitting",
  "n00827638": "locking, lockup",
  "n00827782": "escort, accompaniment",
  "n00827974": "convoy",
  "n00828082": "covering",
  "n00828237": "dressing, grooming",
  "n00828462": "investment",
  "n00828559": "primping",
  "n00828704": "toilet, toilette",
  "n00828862": "dressing",
  "n00828990": "immunization, immunisation",
  "n00829170": "sensitizing, sensitising, sensitization, sensitisation",
  "n00829378": "care, charge, tutelage, guardianship",
  "n00829721": "ruggedization, ruggedisation",
  "n00829918": "umbrella",
  "n00830099": "waterproofing, sealing",
  "n00830257": "wear, wearing",
  "n00830448": "control",
  "n00830733": "motor control",
  "n00830811": "respiration, internal respiration, cellular respiration",
  "n00831191": "breathing, external respiration, respiration, ventilation",
  "n00831782": "respiration",
  "n00831919": "artificial respiration",
  "n00832092": "cardiography, electrocardiography",
  "n00832306": "echocardiography",
  "n00832471": "echoencephalography",
  "n00832626": "cardiopulmonary resuscitation, CPR, cardiac resuscitation, mouth-to-mouth resuscitation, kiss of life",
  "n00833119": "Heimlich maneuver, Heimlich manoeuvere",
  "n00833296": "abdominal breathing",
  "n00833491": "eupnea, eupnoea",
  "n00833578": "hyperpnea",
  "n00833772": "hypopnea",
  "n00833870": "hyperventilation",
  "n00834135": "panting, heaving",
  "n00834290": "periodic breathing, Cheyne-Stokes respiration",
  "n00834460": "puffing, huffing, snorting",
  "n00834636": "smoke, smoking",
  "n00834866": "puffing",
  "n00835032": "breath",
  "n00835267": "exhalation, expiration, breathing out",
  "n00835501": "blow, puff",
  "n00835830": "insufflation",
  "n00835976": "snore, snoring, stertor",
  "n00836149": "snuffle, sniffle, snivel",
  "n00836407": "wheeze",
  "n00836537": "wind",
  "n00836639": "second wind",
  "n00836788": "inhalation, inspiration, aspiration, intake, breathing in",
  "n00837098": "gasp, pant",
  "n00837293": "yawn, yawning, oscitance, oscitancy",
  "n00837675": "puff, drag, pull",
  "n00837965": "toke",
  "n00838098": "consumption, ingestion, intake, uptake",
  "n00838367": "eating, feeding",
  "n00838816": "bite, chomp",
  "n00839023": "browse, browsing",
  "n00839163": "coprophagy, coprophagia",
  "n00839292": "electric shock, electrical shock, shock",
  "n00839597": "fart, farting, flatus, wind, breaking wind",
  "n00839778": "swallow, drink, deglutition",
  "n00840057": "aerophagia",
  "n00840189": "gulp, draft, draught, swig",
  "n00840363": "gulp, gulping",
  "n00840517": "dining",
  "n00840630": "engorgement",
  "n00840751": "feasting, banqueting",
  "n00840898": "geophagy, geophagia",
  "n00841091": "graze, grazing",
  "n00841297": "lunching",
  "n00841393": "munch",
  "n00841527": "Dutch treat",
  "n00841628": "repletion, surfeit",
  "n00841777": "supping",
  "n00841901": "tasting, savoring, savouring, relishing, degustation",
  "n00842197": "nibble",
  "n00842281": "nip, pinch",
  "n00842419": "necrophagia, necrophagy",
  "n00842519": "omophagia",
  "n00842597": "scatophagy",
  "n00842692": "sucking, suck, suction",
  "n00842997": "suckling, lactation",
  "n00843128": "drinking, imbibing, imbibition",
  "n00843325": "gulping, swilling, guzzling",
  "n00843494": "sip",
  "n00843575": "potation",
  "n00843681": "bondage",
  "n00843831": "outercourse",
  "n00844048": "safe sex",
  "n00844254": "sexual activity, sexual practice, sex, sex activity",
  "n00844847": "conception",
  "n00844994": "defloration",
  "n00845178": "insemination",
  "n00845352": "artificial insemination, AI",
  "n00845523": "sexual intercourse, intercourse, sex act, copulation, coitus, coition, sexual congress, congress, sexual relation, relation, carnal knowledge",
  "n00846021": "fuck, fucking, screw, screwing, ass, nooky, nookie, piece of ass, piece of tail, roll in the hay, shag, shtup",
  "n00846326": "pleasure",
  "n00846432": "hank panky",
  "n00846515": "sexual love, lovemaking, making love, love, love life",
  "n00846817": "penetration",
  "n00846961": "statutory rape, carnal abuse",
  "n00847167": "carnal abuse",
  "n00847340": "coupling, mating, pairing, conjugation, union, sexual union",
  "n00847770": "assortative mating",
  "n00847932": "disassortative mating",
  "n00848098": "unlawful carnal knowledge, criminal congress",
  "n00848282": "extramarital sex, free love",
  "n00848466": "adultery, criminal conversation, fornication",
  "n00848745": "fornication",
  "n00848891": "incest",
  "n00849059": "coitus interruptus, withdrawal method, withdrawal, pulling out, onanism",
  "n00849294": "sodomy, buggery, anal sex, anal intercourse",
  "n00849523": "reproduction, procreation, breeding, facts of life",
  "n00849768": "miscegenation, crossbreeding, interbreeding",
  "n00849982": "generation, multiplication, propagation",
  "n00850260": "biogenesis, biogeny",
  "n00850425": "hybridization, hybridisation, crossbreeding, crossing, cross, interbreeding, hybridizing",
  "n00850873": "dihybrid cross",
  "n00850986": "monohybrid cross",
  "n00851146": "reciprocal cross, reciprocal",
  "n00851316": "testcross, test-cross",
  "n00851587": "inbreeding",
  "n00851689": "natural family planning",
  "n00851994": "birth control, birth prevention, family planning",
  "n00852181": "contraception, contraceptive method",
  "n00852400": "oral contraception",
  "n00852522": "basal body temperature method of family planning, basal body temperature method",
  "n00852825": "ovulation method of family planning, ovulation method",
  "n00853145": "rhythm method of birth control, rhythm method, rhythm, calendar method of birth control, calendar method",
  "n00853487": "surgical contraception",
  "n00853649": "servicing, service",
  "n00853835": "foreplay, arousal, stimulation",
  "n00854000": "caressing, cuddling, fondling, hugging, kissing, necking, petting, smooching, snuggling",
  "n00854393": "snogging",
  "n00854538": "feel",
  "n00854717": "perversion, sexual perversion",
  "n00854876": "oral sex, head",
  "n00855055": "cunnilingus, cunnilinctus",
  "n00855169": "fellatio, fellation",
  "n00855301": "cock sucking, blowjob",
  "n00855407": "soixante-neuf, sixty-nine",
  "n00855527": "autoeroticism, autoerotism",
  "n00855674": "masturbation, onanism",
  "n00855936": "self-stimulation, self-abuse",
  "n00856076": "frottage",
  "n00856193": "jacking off, jerking off, hand job, wank",
  "n00856342": "promiscuity, promiscuousness, sleeping around",
  "n00856559": "one-night stand",
  "n00856733": "lechery",
  "n00856847": "homosexuality, homosexualism, homoeroticism, queerness, gayness",
  "n00857160": "bisexuality",
  "n00857275": "inversion, sexual inversion",
  "n00857424": "lesbianism, sapphism",
  "n00857530": "tribadism",
  "n00857664": "heterosexuality, heterosexualism, straightness",
  "n00857872": "pederasty, paederasty",
  "n00858060": "bestiality, zooerastia, zooerasty",
  "n00858188": "sleeping",
  "n00858377": "nap, catnap, cat sleep, forty winks, short sleep, snooze",
  "n00858631": "siesta",
  "n00858742": "zizz",
  "n00858849": "doze, drowse",
  "n00859001": "reaction, response",
  "n00859490": "automatism",
  "n00859691": "rebound",
  "n00859863": "overreaction",
  "n00860011": "galvanic skin response, GSR, psychogalvanic response, electrodermal response, electrical skin response, Fere phenomenon, Tarchanoff phenomenon",
  "n00860434": "immune response, immune reaction, immunologic response",
  "n00860857": "anamnestic response, anamnestic reaction",
  "n00861046": "humoral immune response",
  "n00861199": "cell-mediated immune response",
  "n00861382": "complement fixation",
  "n00861611": "tropism",
  "n00861894": "ergotropism",
  "n00861990": "geotropism",
  "n00862079": "heliotropism",
  "n00862170": "meteortropism",
  "n00862310": "neurotropism",
  "n00862399": "phototropism",
  "n00862488": "trophotropism",
  "n00862595": "thermotropism",
  "n00862686": "taxis",
  "n00862859": "chemotaxis",
  "n00863017": "negative chemotaxis",
  "n00863121": "positive chemotaxis",
  "n00863222": "kinesis",
  "n00863417": "double take",
  "n00863513": "reflex, reflex response, reflex action, instinctive reflex, innate reflex, inborn reflex, unconditioned reflex, physiological reaction",
  "n00864226": "conditional reflex, conditioned reflex, acquired reflex, conditional reaction, conditioned reaction, conditional response, conditioned response",
  "n00864535": "learned reaction, learned response",
  "n00864680": "conditioned avoidance, conditioned avoidance response",
  "n00864859": "knee jerk, knee-jerk reflex, patellar reflex",
  "n00865031": "startle response, startle reaction",
  "n00865284": "startle reflex, Moro reflex",
  "n00865471": "wince, flinch",
  "n00865600": "passage, passing",
  "n00865808": "light reflex, pupillary reflex, miosis, myosis",
  "n00866079": "mydriasis",
  "n00866273": "micturition reflex",
  "n00866423": "pharyngeal reflex, gag reflex",
  "n00866606": "pilomotor reflex, gooseflesh, goose bump, goosebump, goose pimple, goose skin, horripilation",
  "n00866882": "plantar reflex",
  "n00867163": "rooting reflex",
  "n00867357": "startle, jump, start",
  "n00867606": "stretch reflex, myotactic reflex",
  "n00867790": "suckling reflex",
  "n00867983": "tremble, shiver, shake",
  "n00868196": "crying, weeping, tears",
  "n00868523": "snivel, sniveling",
  "n00868669": "sob, sobbing",
  "n00868799": "wailing, bawling",
  "n00868910": "calculation, computation, computing",
  "n00869260": "transposition",
  "n00869467": "number crunching",
  "n00869583": "mathematical process, mathematical operation, operation",
  "n00870101": "recalculation",
  "n00870312": "permutation",
  "n00870453": "combination",
  "n00870640": "differentiation",
  "n00870786": "maximization",
  "n00870912": "division",
  "n00871123": "long division",
  "n00871263": "short division",
  "n00871418": "integration",
  "n00871576": "multiplication, times",
  "n00871862": "subtraction, minus",
  "n00872107": "summation, addition, plus",
  "n00872411": "exponentiation, involution",
  "n00872541": "arithmetic operation",
  "n00872722": "matrix operation",
  "n00872900": "matrix addition",
  "n00872986": "matrix multiplication",
  "n00873084": "matrix inversion",
  "n00873238": "matrix transposition",
  "n00873381": "construction",
  "n00873671": "quadrature",
  "n00873798": "relaxation, relaxation method",
  "n00874067": "judgment, judgement, assessment",
  "n00874457": "adjudication",
  "n00874621": "disapproval",
  "n00874806": "evaluation, rating",
  "n00874977": "marking, grading, scoring",
  "n00875246": "estimate, estimation",
  "n00875519": "appraisal",
  "n00875671": "logistic assessment",
  "n00875838": "value judgment, value judgement",
  "n00876062": "moralism",
  "n00876239": "percussion, pleximetry",
  "n00876377": "succussion",
  "n00876542": "auscultation",
  "n00876737": "sensory activity",
  "n00876874": "sensing, perception",
  "n00877127": "look, looking, looking at",
  "n00877625": "glance, glimpse, coup d'oeil",
  "n00877786": "eye-beaming",
  "n00877924": "side-glance, side-look",
  "n00878052": "scrutiny",
  "n00878221": "peek, peep",
  "n00878328": "squint",
  "n00878456": "stare",
  "n00878648": "gaze, regard",
  "n00878797": "glare, glower",
  "n00878926": "contemplation",
  "n00879037": "gape",
  "n00879156": "evil eye",
  "n00879271": "inspection, review",
  "n00879531": "resurvey",
  "n00879607": "sightseeing, rubber-necking",
  "n00879759": "observation, observance, watching",
  "n00880046": "monitoring",
  "n00880269": "sighting",
  "n00880431": "landfall",
  "n00880563": "stargazing",
  "n00880662": "watch, vigil",
  "n00880893": "stakeout",
  "n00881035": "surveillance of disease",
  "n00881257": "listening watch, continuous receiver watch",
  "n00881441": "spying",
  "n00881545": "lookout, outlook",
  "n00881649": "view, survey, sight",
  "n00881914": "eyeful",
  "n00882045": "dekko",
  "n00882159": "listening, hearing",
  "n00882460": "relistening, rehearing",
  "n00882554": "lipreading",
  "n00882702": "taste, tasting",
  "n00882961": "smell, smelling",
  "n00883139": "sniff, snuff",
  "n00883297": "education, instruction, teaching, pedagogy, didactics, educational activity",
  "n00884202": "coeducation",
  "n00884311": "continuing education",
  "n00884466": "course, course of study, course of instruction, class",
  "n00885088": "coursework",
  "n00885297": "adult education",
  "n00885457": "art class",
  "n00885574": "childbirth-preparation class",
  "n00885764": "life class",
  "n00885858": "elementary education",
  "n00886039": "extension, extension service, university extension",
  "n00886272": "extracurricular activity",
  "n00886456": "dramatics",
  "n00886582": "athletics",
  "n00886699": "higher education",
  "n00886807": "secondary education",
  "n00886968": "spectator sport",
  "n00887081": "teaching, instruction, pedagogy",
  "n00887544": "team sport",
  "n00887702": "team teaching",
  "n00887882": "catechesis",
  "n00888037": "catechetical instruction",
  "n00888162": "language teaching",
  "n00888284": "teaching reading",
  "n00888409": "phonics",
  "n00888546": "whole-word method",
  "n00888693": "schooling",
  "n00888796": "indoctrination",
  "n00888960": "brainwashing",
  "n00889082": "inculcation, ingraining, instilling",
  "n00889294": "tutelage, tuition, tutorship",
  "n00889472": "lesson",
  "n00889682": "dance lesson",
  "n00889760": "music lesson",
  "n00889883": "piano lesson",
  "n00889971": "violin lesson",
  "n00890059": "tennis lesson",
  "n00890145": "golf lesson",
  "n00890227": "history lesson",
  "n00890320": "correspondence course",
  "n00890441": "course of lectures",
  "n00890568": "directed study",
  "n00890808": "elective course, elective",
  "n00890941": "extension course",
  "n00891071": "work-study program",
  "n00891224": "home study",
  "n00891345": "industrial arts",
  "n00891512": "orientation course, orientation",
  "n00891642": "propaedeutic, propaedeutics",
  "n00891850": "refresher course, refresher",
  "n00892036": "required course",
  "n00892145": "seminar",
  "n00892254": "shop class, shop",
  "n00892413": "workshop",
  "n00892538": "sleep-learning, hypnopedia",
  "n00892711": "spoonfeeding",
  "n00892861": "lecture, lecturing",
  "n00893088": "lecture demonstration",
  "n00893243": "talk",
  "n00893402": "chalk talk",
  "n00893498": "athletic training",
  "n00893649": "fartlek",
  "n00893836": "discipline",
  "n00893955": "training, preparation, grooming",
  "n00894359": "drill",
  "n00894552": "exercise, practice, drill, practice session, recitation",
  "n00894979": "fire drill",
  "n00895135": "manual of arms, manual",
  "n00895292": "order arms",
  "n00895501": "military training",
  "n00895680": "basic training",
  "n00895879": "retraining",
  "n00895983": "schooling",
  "n00896114": "skull session, skull practice",
  "n00896228": "toilet training",
  "n00896348": "military drill",
  "n00896526": "close-order drill",
  "n00896688": "square-bashing",
  "n00896832": "rehearsal",
  "n00897026": "rehearsal, dry run",
  "n00897365": "dress rehearsal",
  "n00897506": "run-through",
  "n00897608": "walk-through",
  "n00897811": "review, brushup",
  "n00897989": "rub up",
  "n00898127": "scrimmage",
  "n00898286": "shadowboxing",
  "n00898425": "target practice",
  "n00898518": "representation",
  "n00898804": "model, modelling, modeling",
  "n00899049": "simulation",
  "n00899292": "dramatization, dramatisation",
  "n00899501": "guerrilla theater, street theater",
  "n00899657": "puppetry",
  "n00899761": "pageant, pageantry",
  "n00899927": "figuration",
  "n00900070": "symbolizing, symbolising",
  "n00900207": "schematization, schematisation, diagramming",
  "n00900375": "pictorial representation, picturing",
  "n00900581": "typification",
  "n00900726": "depiction, delineation, portrayal",
  "n00900957": "portraiture",
  "n00901083": "imaging, tomography",
  "n00901316": "X-raying, X-radiation",
  "n00901476": "computerized tomography, computed tomography, CT, computerized axial tomography, computed axial tomography, CAT",
  "n00901789": "sonography, ultrasonography, echography, ultrasound",
  "n00902108": "A-scan ultrasonography",
  "n00902240": "B-scan ultrasonography",
  "n00902376": "positron emission tomography, PET",
  "n00902579": "magnetic resonance imaging, MRI",
  "n00902757": "functional magnetic resonance imaging, fMRI",
  "n00902975": "blood-oxygenation level dependent functional magnetic resonance imaging, BOLD FMRI",
  "n00903201": "fluoroscopy",
  "n00903309": "radioscopy, radiology",
  "n00903559": "photography, picture taking",
  "n00904428": "radiography",
  "n00904623": "roentgenography, X-ray photography",
  "n00904904": "xerography",
  "n00905192": "xeroradiography",
  "n00905336": "angiography",
  "n00905528": "lymphangiography, lymphography",
  "n00905742": "arteriography",
  "n00905842": "arthrography",
  "n00906046": "venography",
  "n00906140": "cholangiography",
  "n00906290": "encephalography",
  "n00906470": "myelography",
  "n00906664": "pyelography",
  "n00906829": "intravenous pyelography, IVP",
  "n00906975": "telephotography",
  "n00907071": "telephotography",
  "n00907235": "radiophotography",
  "n00907340": "exposure",
  "n00907500": "overexposure",
  "n00907642": "underexposure",
  "n00907788": "time exposure",
  "n00907919": "filming, cinematography, motion-picture photography",
  "n00908133": "take",
  "n00908292": "retake",
  "n00908405": "animation",
  "n00908492": "creation, creative activity",
  "n00908772": "re-creation",
  "n00908909": "creating from raw materials",
  "n00909193": "spinning",
  "n00909281": "weaving",
  "n00909386": "netting",
  "n00909471": "knitting",
  "n00909565": "crocheting",
  "n00909672": "lace making, tatting",
  "n00909796": "mintage",
  "n00909899": "molding, casting",
  "n00910070": "needlework, needlecraft",
  "n00910203": "recording, transcription",
  "n00910533": "lip synchronization, lip synchronisation, lip synch, lip sync",
  "n00910856": "mastering",
  "n00911048": "construction, building",
  "n00911572": "crenelation, crenellation",
  "n00911752": "erecting, erection",
  "n00911905": "house-raising",
  "n00912001": "fabrication, assembly",
  "n00912165": "dry walling",
  "n00912274": "dismantling, dismantlement, disassembly",
  "n00912576": "grading, leveling",
  "n00912733": "road construction",
  "n00912822": "shipbuilding, ship building",
  "n00912960": "production",
  "n00913274": "rustication",
  "n00913427": "cottage industry",
  "n00913585": "production",
  "n00913705": "production",
  "n00914182": "mass production",
  "n00914343": "overproduction, overrun",
  "n00914509": "underproduction",
  "n00914632": "output, yield",
  "n00914795": "capacity",
  "n00914929": "breeding",
  "n00915140": "brewing",
  "n00915347": "autosexing",
  "n00915497": "cattle breeding",
  "n00915574": "dog breeding",
  "n00915646": "horse breeding",
  "n00915722": "cultivation",
  "n00916023": "cultivation",
  "n00916147": "aquaculture",
  "n00916285": "beekeeping, apiculture",
  "n00916464": "farming, agriculture, husbandry",
  "n00917211": "animal husbandry",
  "n00917614": "arboriculture, tree farming",
  "n00917759": "culture",
  "n00917961": "cranberry culture",
  "n00918055": "monoculture",
  "n00918176": "tillage",
  "n00918289": "dairying, dairy farming",
  "n00918383": "gardening, horticulture",
  "n00918644": "plowing, ploughing",
  "n00918820": "tilling",
  "n00918976": "hydroponics, aquiculture, tank farming",
  "n00919201": "drip culture",
  "n00919396": "mixed farming",
  "n00919513": "planting",
  "n00919692": "insemination",
  "n00919877": "stratification",
  "n00920062": "ranching",
  "n00920211": "strip cropping",
  "n00920357": "subsistence farming",
  "n00920510": "culture",
  "n00920727": "starter",
  "n00920956": "naturalization, naturalisation",
  "n00921160": "landscaping, landscape gardening",
  "n00921289": "market gardening",
  "n00921399": "flower gardening, floriculture",
  "n00921530": "tree surgery",
  "n00921627": "roundup",
  "n00921790": "harvest, harvest time",
  "n00921926": "haying, haying time",
  "n00922054": "rainmaking",
  "n00922144": "generation",
  "n00922327": "mining, excavation",
  "n00922727": "placer mining",
  "n00922848": "strip mining, opencast mining",
  "n00922968": "quarrying",
  "n00923130": "boring, drilling, oil production",
  "n00923303": "sericulture",
  "n00923444": "industry, manufacture",
  "n00923802": "industrialization, industrialisation, industrial enterprise",
  "n00923995": "devising, fashioning, making",
  "n00924438": "foliation",
  "n00924579": "mapmaking, cartography",
  "n00924714": "moviemaking, movie making, film making",
  "n00924825": "fabrication, manufacture, manufacturing",
  "n00925207": "formation, shaping",
  "n00925489": "filing",
  "n00925622": "forging",
  "n00925732": "metalworking, metalwork",
  "n00925866": "granulation",
  "n00926026": "grooving, rifling",
  "n00926179": "turning",
  "n00926277": "newspeak",
  "n00926468": "prefabrication",
  "n00926668": "confection, concoction",
  "n00926923": "lamination",
  "n00927062": "tanning",
  "n00927161": "veneering",
  "n00927261": "creating by mental acts",
  "n00927516": "formation",
  "n00927694": "affixation",
  "n00927845": "prefixation",
  "n00927961": "suffixation",
  "n00928077": "design, designing",
  "n00928371": "planning",
  "n00928542": "city planning, town planning, urban planning",
  "n00928751": "zoning",
  "n00928947": "programming, programing, computer programming, computer programing",
  "n00929285": "logic programming, logic programing",
  "n00929432": "object-oriented programming, object-oriented programing",
  "n00929586": "verbal creation",
  "n00929718": "writing, authorship, composition, penning",
  "n00930481": "adoxography",
  "n00930736": "drafting",
  "n00930868": "dramatization, dramatisation",
  "n00931040": "fabrication, fictionalization, fictionalisation",
  "n00931217": "historiography",
  "n00931300": "metrification",
  "n00931453": "novelization, novelisation",
  "n00931608": "redaction",
  "n00931721": "lexicography",
  "n00931847": "realization, realisation, actualization, actualisation",
  "n00932088": "objectification",
  "n00932298": "depersonalization, depersonalisation, reification",
  "n00932624": "externalization, externalisation, exteriorization, exteriorisation",
  "n00932804": "hypostatization, hypostatisation, reification",
  "n00933000": "embodiment",
  "n00933154": "soul",
  "n00933259": "personification, incarnation",
  "n00933420": "art, artistic creation, artistic production",
  "n00935005": "arts and crafts",
  "n00935247": "ceramics",
  "n00935453": "decalcomania",
  "n00935608": "decantation",
  "n00935786": "decoupage",
  "n00935940": "drawing, draftsmanship, drafting",
  "n00936194": "glyptography",
  "n00936296": "gastronomy",
  "n00936456": "origami",
  "n00936620": "painting",
  "n00936901": "distemper",
  "n00937105": "fresco",
  "n00937250": "impasto",
  "n00937394": "perfumery",
  "n00937476": "printmaking",
  "n00937656": "sculpture, carving",
  "n00937895": "modeling, modelling, molding, moulding",
  "n00938134": "topiary",
  "n00938240": "pyrography",
  "n00938419": "tracing",
  "n00938538": "oil painting",
  "n00938642": "watercolor, water-color, watercolour, water-colour",
  "n00938791": "engraving, etching",
  "n00938992": "steel engraving",
  "n00939091": "aquatint",
  "n00939231": "serigraphy",
  "n00939338": "lithography",
  "n00939452": "composing, composition",
  "n00939628": "arrangement, arranging, transcription",
  "n00939818": "orchestration, instrumentation",
  "n00940043": "realization, realisation",
  "n00940217": "recapitulation",
  "n00940412": "invention",
  "n00940560": "neologism, neology, coinage",
  "n00940709": "devisal, contrivance",
  "n00940842": "conceptualization, conceptualisation, formulation",
  "n00941140": "approach, attack, plan of attack",
  "n00941451": "framing",
  "n00941604": "avenue",
  "n00941777": "creating by removal",
  "n00941974": "excavation, digging, dig",
  "n00942234": "carving, cutting",
  "n00942440": "petroglyph",
  "n00942574": "truncation",
  "n00942799": "drilling, boring",
  "n00942900": "gouge",
  "n00942988": "puncture",
  "n00943187": "centesis",
  "n00943471": "abdominocentesis, paracentesis",
  "n00943600": "amniocentesis, amnio",
  "n00943861": "arthrocentesis",
  "n00943963": "celiocentesis",
  "n00944068": "lumbar puncture, spinal puncture, spinal tap",
  "n00944296": "thoracocentesis, thoracentesis",
  "n00944456": "fetoscopy, foetoscopy",
  "n00944633": "perforation",
  "n00944789": "prick, pricking",
  "n00944965": "venipuncture",
  "n00945205": "film editing, cutting",
  "n00945401": "search, hunt, hunting",
  "n00945777": "exploration",
  "n00945916": "foraging, forage",
  "n00946060": "frisk, frisking",
  "n00946281": "strip search",
  "n00946432": "looking, looking for",
  "n00946529": "manhunt",
  "n00946650": "quest, seeking",
  "n00946806": "ransacking, rummage",
  "n00947041": "probe",
  "n00947128": "use, usage, utilization, utilisation, employment, exercise",
  "n00947596": "play",
  "n00947719": "misuse, abuse",
  "n00947923": "substance abuse, drug abuse, habit",
  "n00948064": "alcohol abuse, alcoholic abuse, alcoholism abuse",
  "n00948206": "exploitation, development",
  "n00948640": "land development",
  "n00948737": "water development, water project, water program",
  "n00948868": "recycling",
  "n00949042": "bottle collection",
  "n00949134": "application, practical application",
  "n00949516": "misapplication",
  "n00949619": "technology, engineering",
  "n00949948": "aeronautical engineering",
  "n00950070": "automotive technology, automotive engineering",
  "n00950217": "chemical engineering",
  "n00950356": "communications technology",
  "n00950526": "digital communications technology",
  "n00950705": "computer technology",
  "n00950858": "high technology, high tech",
  "n00951037": "rail technology, railroading",
  "n00951233": "magnetic levitation, maglev",
  "n00951433": "overexploitation, overuse, overutilization, overutilisation",
  "n00951626": "capitalization, capitalisation",
  "n00951781": "commercialization, commercialisation",
  "n00952214": "capitalization, capitalisation",
  "n00952386": "market capitalization, market capitalisation",
  "n00952615": "electrification",
  "n00952787": "unitization, unitisation",
  "n00952963": "military action, action",
  "n00953280": "limited war",
  "n00953410": "psychological warfare, war of nerves",
  "n00953559": "battle, conflict, fight, engagement",
  "n00954086": "blockade, encirclement",
  "n00954311": "defense, defence, defensive measure",
  "n00954751": "electronic warfare, EW",
  "n00955060": "operation, military operation",
  "n00955565": "combined operation",
  "n00955806": "police action",
  "n00955987": "resistance",
  "n00956172": "saber rattling, sabre rattling",
  "n00956340": "Armageddon",
  "n00956485": "pitched battle",
  "n00958477": "naval battle",
  "n00958896": "conflict, struggle, battle",
  "n00959376": "brush, clash, encounter, skirmish",
  "n00959537": "close-quarter fighting",
  "n00959645": "contretemps",
  "n00959800": "class struggle, class war, class warfare",
  "n00959992": "maneuver, manoeuvre, simulated military operation",
  "n00960192": "air defense",
  "n00960395": "active air defense",
  "n00960546": "passive air defense",
  "n00960688": "civil defense",
  "n00960851": "stand",
  "n00961001": "repulsion, standoff",
  "n00961177": "hasty defense, hasty defence",
  "n00961328": "deliberate defense, deliberate defence",
  "n00961594": "biological defense, biological defence, biodefense, biodefence",
  "n00961794": "chemical defense, chemical defence",
  "n00961962": "mining, minelaying",
  "n00962129": "rebellion, insurrection, revolt, rising, uprising",
  "n00962567": "civil war",
  "n00962722": "revolution",
  "n00963057": "counterrevolution",
  "n00963241": "insurgency, insurgence",
  "n00963447": "intifada, intifadah",
  "n00963749": "pacification, counterinsurgency",
  "n00963896": "mutiny",
  "n00964105": "Peasant's Revolt, Great Revolt",
  "n00964343": "combat, armed combat",
  "n00964569": "aggression, hostility",
  "n00964781": "hostilities, belligerency",
  "n00964952": "trench warfare",
  "n00965183": "meat grinder",
  "n00965404": "violence, force",
  "n00965718": "domestic violence",
  "n00965895": "plundering, pillage, pillaging",
  "n00966294": "banditry",
  "n00966384": "rape, rapine",
  "n00966504": "rustling",
  "n00966599": "looting, robbery",
  "n00966718": "defloration",
  "n00966869": "spoil, spoliation, spoilation, despoilation, despoilment, despoliation",
  "n00967157": "ravaging, devastation",
  "n00967310": "depredation, predation",
  "n00967446": "sack",
  "n00967622": "chemical warfare, chemical operations",
  "n00967780": "biological warfare, BW, biological attack, biologic attack, bioattack",
  "n00967993": "biological warfare defense, biological warfare defence, BW defense, BW defence",
  "n00968155": "campaign, military campaign",
  "n00968479": "expedition, military expedition, hostile expedition",
  "n00968715": "Crusade",
  "n00969087": "First Crusade",
  "n00969225": "Second Crusade",
  "n00969421": "Third Crusade",
  "n00969684": "Fourth Crusade",
  "n00969858": "Fifth Crusade",
  "n00970079": "Sixth Crusade",
  "n00970331": "Seventh Crusade",
  "n00970478": "naval campaign",
  "n00970645": "mission, military mission",
  "n00970903": "combat mission",
  "n00971004": "search mission",
  "n00971096": "search and destroy mission",
  "n00971309": "sortie, sally",
  "n00971463": "support, reinforcement, reenforcement",
  "n00971802": "dogfight",
  "n00971984": "close support",
  "n00972112": "direct support",
  "n00972319": "amphibious demonstration",
  "n00972521": "diversionary landing",
  "n00972621": "attack, onslaught, onset, onrush",
  "n00973077": "war, warfare",
  "n00974111": "air raid, air attack",
  "n00974224": "dogfight",
  "n00974358": "ground attack",
  "n00974444": "assault",
  "n00974630": "storm",
  "n00974762": "charge",
  "n00974994": "countercharge",
  "n00975074": "banzai attack, banzai charge",
  "n00975270": "diversion, diversionary attack",
  "n00975452": "penetration, incursion",
  "n00975658": "interpenetration",
  "n00975781": "blitz, blitzkrieg",
  "n00975955": "breakthrough",
  "n00976085": "safety blitz, linebacker blitzing, blitz",
  "n00976270": "mousetrap, trap play",
  "n00976531": "invasion",
  "n00976698": "infiltration",
  "n00976953": "foray, raid, maraud",
  "n00977132": "inroad",
  "n00977214": "swoop",
  "n00977301": "strike",
  "n00977551": "first strike",
  "n00977857": "surgical strike",
  "n00978005": "preventive strike, preventive attack",
  "n00978173": "counterattack, countermove",
  "n00978413": "bombing, bombardment",
  "n00978580": "bombardment",
  "n00978764": "bombing run",
  "n00978993": "carpet bombing, area bombing, saturation bombing",
  "n00979180": "dive-bombing",
  "n00979348": "loft bombing, toss bombing",
  "n00979557": "over-the-shoulder bombing",
  "n00979742": "bombing",
  "n00979902": "suicide bombing",
  "n00980038": "offense, offence, offensive",
  "n00980200": "counteroffensive",
  "n00980394": "dirty war",
  "n00980779": "rollback, push back",
  "n00980904": "peacekeeping, peacekeeping mission, peacekeeping operation",
  "n00981180": "amphibious operation",
  "n00981369": "amphibious assault",
  "n00981711": "information gathering",
  "n00981830": "intelligence, intelligence activity, intelligence operation",
  "n00982153": "current intelligence",
  "n00982347": "tactical intelligence, combat intelligence",
  "n00982536": "terrain intelligence",
  "n00982679": "strategic intelligence",
  "n00982852": "signals intelligence, SIGINT",
  "n00983102": "electronics intelligence, ELINT",
  "n00983287": "communications intelligence, COMINT",
  "n00983482": "telemetry intelligence, TELINT",
  "n00983651": "clandestine operation",
  "n00983811": "exfiltration operation",
  "n00983982": "psychological operation, psyop",
  "n00984195": "covert operation",
  "n00984357": "black operation",
  "n00984490": "overt operation",
  "n00984609": "reconnaissance, reconnaissance mission",
  "n00985008": "recce, recco, reccy",
  "n00985106": "scouting, exploratory survey, reconnoitering, reconnoitring",
  "n00985356": "air reconnaissance",
  "n00985516": "reconnaissance by fire",
  "n00985746": "reconnaissance in force",
  "n00985921": "shufti",
  "n00986080": "electronic reconnaissance",
  "n00986275": "counterintelligence",
  "n00986663": "countersubversion",
  "n00986813": "counter-sabotage",
  "n00986938": "fire, firing",
  "n00987775": "antiaircraft fire",
  "n00987863": "barrage, barrage fire, battery, bombardment, shelling",
  "n00988177": "broadside",
  "n00988320": "fusillade, salvo, volley, burst",
  "n00988556": "call fire",
  "n00988696": "close supporting fire",
  "n00988893": "cover, covering fire",
  "n00989122": "deep supporting fire",
  "n00989385": "direct supporting fire",
  "n00989583": "concentrated fire, massed fire",
  "n00989773": "counterfire",
  "n00989937": "counterbattery fire",
  "n00990071": "counterbombardment",
  "n00990195": "countermortar fire",
  "n00990319": "counterpreparation fire",
  "n00990474": "crossfire",
  "n00990590": "destruction fire",
  "n00990719": "direct fire",
  "n00990843": "distributed fire",
  "n00990963": "friendly fire, fratricide",
  "n00991070": "hostile fire",
  "n00991164": "grazing fire",
  "n00991336": "harassing fire",
  "n00991496": "indirect fire",
  "n00991642": "interdiction fire",
  "n00991772": "neutralization fire",
  "n00991914": "observed fire",
  "n00992102": "preparation fire",
  "n00992221": "radar fire",
  "n00992331": "dating, geological dating",
  "n00992544": "potassium-argon dating",
  "n00992732": "radiocarbon dating, carbon dating, carbon-14 dating",
  "n00992995": "rubidium-strontium dating",
  "n00993228": "registration fire",
  "n00993377": "scheduled fire",
  "n00993488": "scouring",
  "n00993651": "searching fire",
  "n00993787": "shakedown",
  "n00993956": "supporting fire",
  "n00994144": "suppressive fire",
  "n00994321": "unobserved fire",
  "n00994449": "artillery fire, cannon fire",
  "n00994623": "cannonade, drumfire",
  "n00994745": "high-angle fire",
  "n00994895": "mortar fire",
  "n00994989": "zone fire",
  "n00995134": "electronic countermeasures, ECM",
  "n00995356": "electronic counter-countermeasures, ECCM",
  "n00995588": "electronic warfare-support measures, ESM",
  "n00995855": "electromagnetic intrusion",
  "n00996056": "germ warfare, bacteriological warfare",
  "n00996180": "information warfare, IW",
  "n00996513": "jihad, jehad",
  "n00996673": "jihad, jehad, international jihad",
  "n00996817": "world war",
  "n00996969": "measurement, measuring, measure, mensuration",
  "n00998037": "actinometry",
  "n00998196": "algometry",
  "n00998329": "anemography",
  "n00998423": "anemometry",
  "n00998514": "angulation",
  "n00998604": "anthropometry",
  "n00998768": "arterial blood gases",
  "n00998982": "audiometry",
  "n00999089": "bathymetry, plumbing",
  "n00999245": "calibration, standardization, standardisation",
  "n00999588": "tuning",
  "n00999787": "adjustment, registration, readjustment",
  "n01000068": "alignment",
  "n01000276": "collimation",
  "n01000411": "temperament",
  "n01000610": "equal temperament",
  "n01000843": "tune",
  "n01000969": "tune-up",
  "n01001097": "synchronization, synchronisation, synchronizing, synchronising",
  "n01001344": "camber",
  "n01001489": "toe-in",
  "n01001640": "voicing",
  "n01001814": "calorimetry",
  "n01001923": "cephalometry",
  "n01002008": "densitometry",
  "n01002165": "dosimetry",
  "n01002284": "fetometry, foetometry",
  "n01002413": "hydrometry, gravimetry",
  "n01002554": "hypsometry, hypsography",
  "n01002677": "mental measurement",
  "n01002872": "micrometry",
  "n01002956": "observation",
  "n01003113": "pelvimetry",
  "n01003272": "photometry",
  "n01003435": "cytophotometry",
  "n01003570": "quantification",
  "n01003729": "gradation, graduation",
  "n01003936": "shading, blending",
  "n01004072": "divergence, divergency",
  "n01004377": "radioactive dating",
  "n01004582": "reading, meter reading",
  "n01004775": "sampling",
  "n01004961": "sounding",
  "n01005100": "sound ranging",
  "n01005284": "scaling",
  "n01005473": "spirometry",
  "n01005579": "surveying",
  "n01005814": "scalage",
  "n01005957": "scalage",
  "n01006054": "electromyography",
  "n01006188": "mammography",
  "n01006313": "thermography",
  "n01006536": "mammothermography",
  "n01006675": "test, mental test, mental testing, psychometric test",
  "n01007053": "intelligence test, IQ test",
  "n01007273": "Stanford-Binet test",
  "n01007372": "Binet-Simon Scale",
  "n01007463": "personality test",
  "n01007609": "projective test, projective device, projective technique",
  "n01007907": "Rorschach, Rorschach test, inkblot test",
  "n01008097": "Thematic Apperception Test, TAT",
  "n01008271": "sub-test",
  "n01008378": "organization, organisation",
  "n01008801": "orchestration",
  "n01009001": "randomization, randomisation",
  "n01009190": "systematization, systematisation, rationalization, rationalisation",
  "n01009507": "codification",
  "n01009637": "formalization, formalisation",
  "n01009871": "order, ordering",
  "n01010245": "rank order",
  "n01010334": "scaling, grading",
  "n01010458": "succession, sequence",
  "n01010684": "alternation",
  "n01010901": "layout",
  "n01011028": "alphabetization, alphabetisation",
  "n01011166": "listing, itemization, itemisation",
  "n01011425": "inventory, inventorying, stocktaking, stock-taking",
  "n01011720": "stocktake, stock-take",
  "n01011883": "roll call",
  "n01012030": "mail call",
  "n01012125": "muster call",
  "n01012257": "attendance check",
  "n01012360": "grouping",
  "n01012608": "phrasing",
  "n01012712": "categorization, categorisation, classification, compartmentalization, compartmentalisation, assortment",
  "n01013156": "indexing",
  "n01013316": "reclassification",
  "n01013434": "relegation",
  "n01013604": "stratification",
  "n01013770": "taxonomy",
  "n01013971": "typology",
  "n01014066": "collection, collecting, assembling, aggregation",
  "n01014490": "agglomeration",
  "n01014607": "collation",
  "n01014731": "compilation, compiling",
  "n01014990": "gather, gathering",
  "n01015175": "centralization, centralisation",
  "n01015310": "harvest, harvesting, harvest home",
  "n01015477": "haying",
  "n01015551": "bottle collection",
  "n01015689": "conchology, shell collecting",
  "n01015843": "garbage collection, garbage pickup, trash collection, trash pickup",
  "n01015996": "numismatics, numismatology, coin collecting, coin collection",
  "n01016201": "pickup",
  "n01016420": "philately, stamp collecting, stamp collection",
  "n01016628": "aerophilately",
  "n01016748": "tax collection",
  "n01016832": "sorting",
  "n01016973": "territorialization, territorialisation",
  "n01017167": "triage",
  "n01017320": "support, supporting",
  "n01017550": "shoring, shoring up, propping up",
  "n01017701": "suspension, dangling, hanging",
  "n01017987": "continuance, continuation",
  "n01018366": "prolongation, protraction, perpetuation, lengthening",
  "n01018630": "repetition, repeating",
  "n01018951": "echolalia",
  "n01019129": "iteration",
  "n01019248": "redundancy",
  "n01019372": "reduplication, reiteration",
  "n01019524": "copying",
  "n01019703": "duplication, gemination",
  "n01019901": "reproduction, replication",
  "n01020117": "replay, instant replay, action replay",
  "n01020362": "sound reproduction",
  "n01020488": "high fidelity, hi-fi",
  "n01020628": "headroom, dynamic headroom",
  "n01020770": "playback",
  "n01020936": "imitation",
  "n01021119": "echo",
  "n01021270": "emulation",
  "n01021396": "mimicry",
  "n01021579": "perseverance, persistence, perseveration",
  "n01021889": "abidance",
  "n01022008": "pursuance, prosecution",
  "n01022178": "survival",
  "n01022292": "hangover, holdover",
  "n01022483": "discontinuance, discontinuation",
  "n01022824": "disfranchisement",
  "n01022992": "disinheritance",
  "n01023138": "phase-out",
  "n01023242": "intervention",
  "n01023636": "procedure",
  "n01023820": "procedure, process",
  "n01024392": "medical procedure",
  "n01024547": "dental procedure",
  "n01024643": "mapping, chromosome mapping",
  "n01024811": "operating procedure",
  "n01024968": "standing operating procedure, standard operating procedure, SOP, standard procedure",
  "n01025254": "lockstep",
  "n01025411": "stiffening",
  "n01025563": "bureaucratic procedure, red tape",
  "n01025678": "objection",
  "n01026020": "recusation",
  "n01026191": "indirection",
  "n01026316": "rigmarole, rigamarole",
  "n01026482": "routine, modus operandi",
  "n01026633": "rat race",
  "n01026744": "rut, groove",
  "n01026897": "ceremony",
  "n01027231": "tea ceremony, chanoyu",
  "n01027379": "ceremony",
  "n01027662": "lustrum",
  "n01027859": "ritual, rite",
  "n01028082": "religious ceremony, religious ritual",
  "n01028381": "military ceremony",
  "n01028534": "agape, love feast",
  "n01028655": "worship",
  "n01029114": "deification, exaltation, apotheosis",
  "n01029323": "ancestor worship",
  "n01029406": "rite, religious rite",
  "n01029671": "vigil, watch",
  "n01029883": "wake, viewing",
  "n01030033": "agrypnia",
  "n01030152": "last rites",
  "n01030280": "orgy",
  "n01030488": "popery, papism",
  "n01030642": "quotation",
  "n01030820": "ritual",
  "n01031023": "ritualism",
  "n01031194": "circumcision",
  "n01031392": "Berith, Berit, Brith, Bris, Briss",
  "n01031563": "nudism, naturism",
  "n01031705": "systematism",
  "n01031858": "transvestism, transvestitism, cross dressing",
  "n01032040": "service, religious service, divine service",
  "n01032368": "church service, church",
  "n01032810": "devotional",
  "n01032892": "prayer meeting, prayer service",
  "n01033024": "chapel service, chapel",
  "n01033184": "liturgy",
  "n01033345": "Christian liturgy",
  "n01033458": "office",
  "n01033714": "Divine Office",
  "n01033903": "Little Office",
  "n01034077": "Office of the Dead",
  "n01034233": "committal service",
  "n01034385": "none",
  "n01034571": "vesper",
  "n01034685": "placebo",
  "n01034816": "watch night",
  "n01034925": "sacrament",
  "n01035504": "Last Supper, Lord's Supper",
  "n01035667": "Seder, Passover supper",
  "n01035853": "Holy Eucharist, Eucharist, sacrament of the Eucharist, Holy Sacrament, Liturgy, Eucharistic liturgy, Lord's Supper",
  "n01036194": "Offertory",
  "n01036333": "Communion, Holy Communion, sacramental manduction, manduction",
  "n01036624": "intercommunion",
  "n01036778": "betrothal, espousal",
  "n01036903": "matrimony",
  "n01036996": "marriage, wedding, marriage ceremony",
  "n01037294": "rite of passage",
  "n01037469": "bridal, espousal",
  "n01037577": "civil marriage",
  "n01037710": "love match",
  "n01037819": "baptism",
  "n01038147": "affusion",
  "n01038261": "aspersion, sprinkling",
  "n01038375": "christening",
  "n01038485": "immersion",
  "n01038629": "trine immersion",
  "n01038761": "confirmation",
  "n01038895": "confirmation",
  "n01039140": "penance",
  "n01039307": "confession",
  "n01039560": "shrift",
  "n01039637": "anointing of the sick, extreme unction, last rites",
  "n01039822": "holy order",
  "n01039925": "sanctification",
  "n01040128": "beatification",
  "n01040390": "canonization, canonisation",
  "n01040646": "consecration",
  "n01040943": "communalism",
  "n01041111": "consecration",
  "n01041349": "Oblation, religious offering",
  "n01041498": "oblation, offering",
  "n01041674": "unction, inunction",
  "n01041835": "libation",
  "n01041968": "prayer, supplication",
  "n01042242": "Mass",
  "n01042465": "High Mass",
  "n01042578": "Low Mass",
  "n01042661": "Requiem",
  "n01042764": "devotion",
  "n01042998": "bhakti",
  "n01043189": "novena",
  "n01043333": "Stations, Stations of the Cross",
  "n01043693": "blessing, benediction",
  "n01043820": "idolization, idolisation",
  "n01043989": "adoration, latria",
  "n01044084": "idolatry, idol worship",
  "n01044268": "bardolatry",
  "n01044363": "iconolatry",
  "n01044448": "idolatry, devotion, veneration, cultism",
  "n01044761": "idiolatry, autolatry, self-worship",
  "n01044867": "bibliolatry, Bible-worship",
  "n01044983": "verbolatry, grammatolatry, word-worship",
  "n01045091": "symbolatry, symbololatry, symbol-worship",
  "n01045202": "anthropolatry, worship of man",
  "n01045306": "gyneolatry, gynaeolatry, woman-worship",
  "n01045413": "lordolatry",
  "n01045520": "thaumatolatry, miracle-worship",
  "n01045621": "topolatry, place-worship",
  "n01045714": "arborolatry, tree-worship",
  "n01045807": "astrolatry, worship of heavenly bodies",
  "n01045924": "cosmolatry",
  "n01046006": "diabolatry, demonolatry, devil-worship",
  "n01046167": "pyrolatry, fire-worship",
  "n01046257": "hagiolatry, hierolatry",
  "n01046348": "heliolatry, sun-worship",
  "n01046441": "zoolatry, animal-worship",
  "n01046571": "ichthyolatry, fish-worship",
  "n01046664": "monolatry",
  "n01046792": "ophiolatry, serpent-worship",
  "n01046888": "moon-worship, selenolatry",
  "n01046984": "energizing, activating, activation",
  "n01047191": "electrification",
  "n01047338": "revival, resurgence, revitalization, revitalisation, revivification",
  "n01047803": "rebirth, Renaissance, Renascence",
  "n01047937": "regeneration",
  "n01048059": "resurrection",
  "n01048210": "resuscitation",
  "n01048466": "vivification, invigoration, animation",
  "n01048697": "presentation",
  "n01048912": "concealment, concealing, hiding",
  "n01049266": "disguise, camouflage",
  "n01049475": "mask",
  "n01049685": "cover, covering, screening, masking",
  "n01049992": "cover",
  "n01050187": "cover-up",
  "n01050356": "blue wall of silence, blue wall, wall of silence",
  "n01050627": "burying, burial",
  "n01050763": "reburying, reburial",
  "n01050872": "smoke screen, smokescreen",
  "n01051082": "stealth, stealing",
  "n01051221": "money laundering",
  "n01051331": "placement, location, locating, position, positioning, emplacement",
  "n01051801": "juxtaposition, apposition, collocation",
  "n01052078": "tessellation",
  "n01052215": "interposition, intervention",
  "n01052372": "orientation",
  "n01052450": "planting",
  "n01052618": "implantation",
  "n01052739": "repositioning",
  "n01052853": "set",
  "n01053067": "superposition",
  "n01053207": "fingering",
  "n01053404": "superposition",
  "n01053617": "stay",
  "n01053920": "residency, residence, abidance",
  "n01054227": "lodging",
  "n01054335": "occupancy, tenancy",
  "n01054545": "inhabitancy, inhabitation, habitation",
  "n01054876": "cohabitation",
  "n01055065": "concubinage",
  "n01055165": "camping, encampment, bivouacking, tenting",
  "n01055360": "sojourn, visit",
  "n01055493": "call",
  "n01055688": "round",
  "n01055954": "call",
  "n01056236": "visiting",
  "n01056411": "stop, stopover, layover",
  "n01056713": "night-stop",
  "n01056804": "pit stop",
  "n01056941": "pit stop",
  "n01057053": "stand",
  "n01057200": "provision, supply, supplying",
  "n01057631": "irrigation",
  "n01057759": "feeding, alimentation",
  "n01058049": "infant feeding",
  "n01058181": "demand feeding",
  "n01058291": "forced feeding, gavage",
  "n01058539": "nasogastric feeding",
  "n01058715": "gastrogavage",
  "n01058870": "nursing, breast feeding",
  "n01058983": "intravenous feeding, IV",
  "n01059096": "overfeeding",
  "n01059189": "spoonfeeding",
  "n01059306": "schedule feeding",
  "n01059445": "total parenteral nutrition, TPN, hyperalimentation",
  "n01059719": "fueling, refueling",
  "n01059900": "healthcare, health care",
  "n01060101": "healthcare delivery, health care delivery, care delivery",
  "n01060234": "issue, issuing, issuance",
  "n01060530": "stock issue",
  "n01060745": "logistics",
  "n01060924": "purveyance",
  "n01061029": "stocking",
  "n01061203": "subvention",
  "n01061333": "demand",
  "n01061526": "exaction",
  "n01061726": "extortion",
  "n01061945": "claim",
  "n01062136": "insurance claim",
  "n01062255": "drain",
  "n01062468": "brain drain",
  "n01062583": "inactivity",
  "n01062817": "pause",
  "n01062997": "respite, recess, break, time out",
  "n01063234": "spring break",
  "n01063350": "hesitation, waver, falter, faltering",
  "n01063578": "intermission",
  "n01063697": "freeze, halt",
  "n01063939": "wait, waiting",
  "n01064148": "rest, ease, repose, relaxation",
  "n01064468": "bedrest, bed rest",
  "n01064613": "laziness",
  "n01064758": "lie-in",
  "n01064863": "quiescence, quiescency, dormancy, sleeping",
  "n01065057": "vegetation",
  "n01065286": "leisure",
  "n01065441": "idleness, idling, loafing",
  "n01065607": "dolce far niente",
  "n01065687": "free time, spare time",
  "n01065805": "vacationing",
  "n01065945": "busman's holiday",
  "n01066061": "caravanning",
  "n01066163": "delay, holdup",
  "n01066545": "demurrage",
  "n01066689": "forbearance",
  "n01066881": "postponement, deferment, deferral",
  "n01067070": "adjournment",
  "n01067192": "prorogation",
  "n01067362": "procrastination, cunctation, shillyshally",
  "n01067577": "slowdown, lag, retardation",
  "n01067819": "dalliance, dawdling, trifling",
  "n01068012": "filibuster",
  "n01068184": "interjection, interposition, interpolation, interpellation",
  "n01068423": "tarriance, lingering",
  "n01068528": "breaking off, abruption",
  "n01068633": "heckling, barracking",
  "n01068773": "abstinence",
  "n01069012": "asceticism, ascesis",
  "n01069125": "chastity, celibacy, sexual abstention",
  "n01069311": "mortification",
  "n01069578": "self-denial, self-discipline, self-control",
  "n01069777": "sobriety, temperance",
  "n01069867": "teetotaling, teetotalism",
  "n01069980": "fast, fasting",
  "n01070187": "diet, dieting",
  "n01070432": "traffic control",
  "n01070566": "point duty",
  "n01070708": "price-fixing",
  "n01070892": "inhibition, suppression",
  "n01071090": "tolerance",
  "n01071213": "lenience, leniency",
  "n01071411": "clemency, mercifulness, mercy",
  "n01071746": "pleasure",
  "n01071921": "luxuriation",
  "n01072072": "enjoyment, delectation",
  "n01072236": "lamentation, mourning",
  "n01072402": "laughter",
  "n01072565": "satisfaction",
  "n01072780": "gratification",
  "n01072984": "satiation",
  "n01073097": "self-gratification, head trip",
  "n01073241": "indulgence, indulging, pampering, humoring",
  "n01073556": "pleasing",
  "n01073655": "overindulgence, excess",
  "n01073824": "orgy, binge, splurge",
  "n01073995": "hindrance, hinderance, interference",
  "n01074252": "antagonism",
  "n01074498": "obstruction",
  "n01074694": "blockage, closure, occlusion",
  "n01074970": "naval blockade",
  "n01075117": "siege, besieging, beleaguering, military blockade",
  "n01075567": "relief",
  "n01075725": "stall, stalling",
  "n01075893": "stonewalling",
  "n01076046": "stop, stoppage",
  "n01076359": "complication",
  "n01076488": "deterrence",
  "n01076724": "discouragement",
  "n01076863": "nuclear deterrence",
  "n01077190": "countermine",
  "n01077350": "prevention, bar",
  "n01077738": "averting",
  "n01077881": "debarment",
  "n01078086": "disqualification",
  "n01078279": "interception",
  "n01078572": "nonproliferation, non-proliferation",
  "n01079042": "obviation, forestalling, preclusion",
  "n01079295": "prophylaxis",
  "n01079396": "save",
  "n01079604": "suppression, crushing, quelling, stifling",
  "n01079951": "tax avoidance",
  "n01080062": "recusation, recusal",
  "n01080366": "group action",
  "n01081197": "social activity",
  "n01081346": "communalism",
  "n01081456": "confederation, alliance",
  "n01081628": "association",
  "n01081867": "fraternization, fraternisation",
  "n01082061": "affiliation",
  "n01082274": "reaffiliation",
  "n01082350": "mingling",
  "n01082548": "decolonization, decolonisation",
  "n01082720": "disbandment",
  "n01082886": "disestablishment",
  "n01083077": "distribution",
  "n01083350": "redistribution",
  "n01083504": "dispensation",
  "n01083645": "allotment, apportionment, apportioning, allocation, parceling, parcelling, assignation",
  "n01084180": "reallotment, reapportionment, reallocation",
  "n01084489": "reshuffle",
  "n01084637": "deal",
  "n01084848": "new deal",
  "n01084932": "rationing",
  "n01085098": "parcel, portion, share",
  "n01085337": "deal",
  "n01085567": "misdeal",
  "n01085656": "revenue sharing",
  "n01085793": "sharing, share-out",
  "n01085937": "generosity, unselfishness",
  "n01086081": "giving, gift",
  "n01086356": "bestowal, bestowment, conferral, conferment",
  "n01086572": "accordance, accordance of rights",
  "n01086744": "endowment",
  "n01086945": "social welfare, welfare, public assistance",
  "n01087178": "social insurance",
  "n01087498": "national insurance",
  "n01087740": "supplementary benefit, social assistance, national assistance",
  "n01087939": "Social Security",
  "n01088168": "relief",
  "n01088304": "dole, pogy, pogey",
  "n01088437": "unemployment compensation",
  "n01088563": "old-age insurance",
  "n01088656": "survivors insurance",
  "n01088757": "disability insurance",
  "n01088857": "health care",
  "n01089009": "Medicare",
  "n01089179": "Medicaid",
  "n01089297": "primary health care",
  "n01089483": "philanthropy, philanthropic gift",
  "n01089635": "charity",
  "n01089778": "contribution, donation",
  "n01090018": "subscription",
  "n01090117": "alms",
  "n01090225": "alms-giving, almsgiving",
  "n01090342": "handout",
  "n01090446": "commerce, commercialism, mercantilism",
  "n01091905": "trade",
  "n01092315": "fair trade",
  "n01092403": "fair trade",
  "n01092643": "free trade",
  "n01092769": "North American Free Trade Agreement, NAFTA",
  "n01092974": "e-commerce",
  "n01093085": "exchange, interchange",
  "n01093380": "conversion",
  "n01093532": "unitization, unitisation",
  "n01093666": "lending, loaning",
  "n01093855": "usury",
  "n01093965": "arbitrage",
  "n01094293": "risk arbitrage, takeover arbitrage",
  "n01094575": "initial public offering, IPO, initial offering",
  "n01094725": "commercial enterprise, business enterprise, business",
  "n01095753": "business activity, commercial activity",
  "n01095966": "operation",
  "n01096245": "business",
  "n01096454": "trade, patronage",
  "n01096674": "wash",
  "n01096873": "custom",
  "n01096997": "land-office business",
  "n01097119": "field, field of operation, line of business",
  "n01097292": "market, marketplace, market place",
  "n01097739": "black market",
  "n01097920": "buyer's market, buyers' market, soft market",
  "n01098071": "grey market, gray market",
  "n01098265": "seller's market, sellers' market",
  "n01098404": "labor market",
  "n01098541": "employee-owned enterprise, employee-owned business",
  "n01098698": "finance",
  "n01098968": "corporate finance",
  "n01099109": "financing, funding",
  "n01099231": "high finance",
  "n01099436": "investing, investment",
  "n01099768": "foreign direct investment",
  "n01099946": "leverage, leveraging",
  "n01100147": "flotation, floatation",
  "n01100273": "banking",
  "n01100491": "home banking",
  "n01100658": "banking",
  "n01100877": "cooperative, co-op",
  "n01101114": "discount business",
  "n01101227": "real-estate business",
  "n01101329": "advertising, publicizing",
  "n01101571": "hard sell",
  "n01101661": "soft sell",
  "n01101753": "circularization, circularisation",
  "n01101958": "publication, publishing",
  "n01102256": "desktop publishing",
  "n01102436": "publication",
  "n01102574": "republication, republishing",
  "n01102712": "contribution",
  "n01102862": "serialization, serialisation",
  "n01103000": "typography",
  "n01103159": "printing",
  "n01103374": "gravure",
  "n01103492": "photogravure, rotogravure",
  "n01103614": "issue, publication",
  "n01103788": "packaging",
  "n01103964": "meatpacking, meat packing, meat-packing business",
  "n01104174": "unitization, unitisation",
  "n01104324": "catering",
  "n01104406": "agribusiness, agriculture, factory farm",
  "n01104544": "truck farming",
  "n01104637": "construction, building",
  "n01105038": "jerry-building",
  "n01105152": "slating",
  "n01105259": "transportation, shipping, transport",
  "n01105635": "air transportation, air transport",
  "n01105737": "navigation",
  "n01105909": "hauling, trucking, truckage",
  "n01106118": "cartage, carting",
  "n01106272": "freight, freightage",
  "n01106460": "express, expressage",
  "n01106587": "ferry, ferrying",
  "n01106715": "carriage trade",
  "n01106808": "transaction, dealing, dealings",
  "n01107359": "affairs",
  "n01107549": "world affairs, international affairs",
  "n01107726": "operations, trading operations",
  "n01107932": "transfer, transference",
  "n01108150": "alienation",
  "n01108402": "conveyance, conveyance of title, conveyancing, conveying",
  "n01108641": "quitclaim",
  "n01108753": "delivery, livery, legal transfer",
  "n01108971": "bailment",
  "n01109114": "lend-lease, lease-lend",
  "n01109311": "secularization, secularisation",
  "n01109467": "exchange",
  "n01109687": "barter, swap, swop, trade",
  "n01109934": "horse trade, horse trading",
  "n01110063": "logrolling",
  "n01110274": "deal, trade, business deal",
  "n01110633": "arms deal",
  "n01110720": "penny ante",
  "n01110811": "downtick",
  "n01110973": "uptick",
  "n01111133": "borrowing",
  "n01111236": "pawn",
  "n01111375": "rental, renting",
  "n01111569": "Seward's Folly",
  "n01111750": "importing, importation",
  "n01111952": "exporting, exportation",
  "n01112132": "smuggling",
  "n01112289": "gunrunning",
  "n01112420": "marketing",
  "n01112739": "direct marketing",
  "n01112885": "distribution",
  "n01113068": "selling, merchandising, marketing",
  "n01113595": "distribution channel, channel",
  "n01113867": "traffic",
  "n01114055": "drug traffic, drug trafficking, narcotraffic",
  "n01114172": "simony, barratry",
  "n01114284": "slave trade, slave traffic",
  "n01114458": "retail",
  "n01114646": "wholesale",
  "n01114824": "sale",
  "n01115017": "divestiture",
  "n01115162": "sell",
  "n01115348": "syndication",
  "n01115589": "dumping",
  "n01115734": "dutch auction",
  "n01115866": "retailing",
  "n01116026": "telemarketing, teleselling, telecommerce",
  "n01116206": "telemetry",
  "n01116360": "thermometry",
  "n01116466": "thermogravimetry",
  "n01116696": "tonometry",
  "n01116876": "telephone order",
  "n01116968": "vending, peddling, hawking, vendition",
  "n01117164": "venture",
  "n01117318": "viscometry, viscosimetry",
  "n01117454": "resale",
  "n01117541": "sale",
  "n01117723": "sale, cut-rate sale, sales event",
  "n01118182": "bazaar, fair",
  "n01118338": "book fair, bookfair",
  "n01118495": "craft fair",
  "n01118614": "car boot sale, boot sale",
  "n01118776": "clearance sale, inventory-clearance sale",
  "n01118890": "closeout",
  "n01119012": "fire sale",
  "n01119116": "fire sale",
  "n01119250": "garage sale, yard sale",
  "n01119401": "going-out-of-business sale",
  "n01119620": "realization, realisation",
  "n01119850": "rummage sale, jumble sale",
  "n01119949": "selloff",
  "n01120176": "white sale",
  "n01120259": "undertaking",
  "n01120348": "upholstery",
  "n01120448": "payment, defrayal, defrayment",
  "n01120855": "evasion, nonpayment",
  "n01121070": "amortization, amortisation",
  "n01121245": "fee splitting",
  "n01121390": "overpayment",
  "n01121492": "prepayment",
  "n01121585": "ransom",
  "n01121690": "refund, repayment",
  "n01121855": "remuneration",
  "n01122037": "rendering",
  "n01122149": "spending, disbursement, disbursal, outlay",
  "n01122387": "tribute",
  "n01122488": "underpayment",
  "n01122601": "expending, expenditure",
  "n01122754": "deficit spending, compensatory spending, pump priming",
  "n01122938": "amortization, amortisation",
  "n01123095": "migration",
  "n01123304": "gold rush",
  "n01123420": "stampede",
  "n01123598": "social control",
  "n01124110": "auto limitation",
  "n01124246": "sanction",
  "n01124384": "population control",
  "n01124508": "politics",
  "n01124794": "government, governing, governance, government activity, administration",
  "n01125562": "misgovernment, misrule",
  "n01125693": "legislation, legislating, lawmaking",
  "n01125959": "criminalization, criminalisation",
  "n01126151": "decriminalization, decriminalisation",
  "n01126335": "trust busting",
  "n01126564": "winemaking, wine making",
  "n01126709": "viticulture, viniculture",
  "n01126856": "enactment, passage",
  "n01127019": "enforcement",
  "n01127245": "coercion",
  "n01127379": "execution, implementation, carrying out",
  "n01127623": "imposition, infliction",
  "n01127874": "protection, trade protection",
  "n01128137": "law enforcement",
  "n01128266": "vigilantism",
  "n01128390": "domination",
  "n01128547": "bossism",
  "n01128655": "mastery, subordination",
  "n01128806": "monopolization, monopolisation",
  "n01128984": "socialization, socialisation, acculturation, enculturation",
  "n01129363": "cultivation",
  "n01129532": "breeding, bringing up, fostering, fosterage, nurture, raising, rearing, upbringing",
  "n01129920": "duty, responsibility, obligation",
  "n01130560": "moral obligation",
  "n01130735": "noblesse oblige",
  "n01130905": "burden of proof",
  "n01131004": "civic duty, civic responsibility",
  "n01131135": "jury duty",
  "n01131224": "filial duty",
  "n01131330": "imperative",
  "n01131425": "incumbency",
  "n01131515": "legal duty",
  "n01131656": "fiduciary duty",
  "n01131794": "due care, ordinary care, reasonable care",
  "n01131993": "foster care",
  "n01132148": "great care",
  "n01132327": "providence",
  "n01132472": "slight care",
  "n01132590": "line of duty",
  "n01132709": "white man's burden",
  "n01132864": "obedience, respect",
  "n01133106": "occupation, military control",
  "n01133281": "management, direction",
  "n01133760": "conducting",
  "n01133933": "database management",
  "n01134037": "finance",
  "n01134244": "homemaking",
  "n01134330": "misconduct",
  "n01134479": "mismanagement, misdirection",
  "n01134699": "screwup",
  "n01134861": "treatment, handling",
  "n01135163": "bioremediation",
  "n01135372": "dealing",
  "n01135529": "supervision, supervising, superintendence, oversight",
  "n01135795": "invigilation",
  "n01135952": "administration, disposal",
  "n01136265": "conducting",
  "n01136375": "line management",
  "n01136519": "organization, organisation",
  "n01136985": "running",
  "n01137206": "administrivia",
  "n01137455": "polity",
  "n01137597": "nonprofit organization, nonprofit, not-for-profit",
  "n01137760": "rationalization, rationalisation",
  "n01137987": "reorganization, reorganisation, shake-up, shakeup",
  "n01138399": "self-organization, self-organisation",
  "n01138547": "syndication",
  "n01138670": "authorization, authorisation, empowerment",
  "n01139000": "sanction",
  "n01139194": "license, permission, permit",
  "n01139490": "benefit of clergy",
  "n01139636": "name",
  "n01139749": "nihil obstat",
  "n01139830": "certification, enfranchisement",
  "n01140029": "disenfranchisement",
  "n01140193": "accreditation",
  "n01140471": "commission, commissioning",
  "n01140658": "mandate",
  "n01140839": "delegating, delegation, relegating, relegation, deputation",
  "n01141160": "devolution, devolvement",
  "n01141366": "loan approval",
  "n01141482": "rubber stamp",
  "n01141593": "clearance",
  "n01141729": "conge, congee",
  "n01141841": "allowance",
  "n01142014": "dispensation",
  "n01142150": "variance",
  "n01142324": "toleration",
  "n01142519": "channelization, channelisation, canalization, canalisation",
  "n01142843": "canalization, canalisation",
  "n01143040": "preparation, readying",
  "n01143409": "deployment",
  "n01143580": "groundwork",
  "n01143760": "redeployment, redisposition",
  "n01143957": "makeready",
  "n01144046": "priming",
  "n01144133": "planning",
  "n01144355": "scheduling, programming, programing",
  "n01144555": "turnaround, turnround",
  "n01144716": "warm-up, tune-up, prolusion",
  "n01144876": "guidance, steering",
  "n01145015": "coup d'etat, coup, putsch, takeover",
  "n01145219": "countercoup",
  "n01145359": "restraint",
  "n01145688": "collar, leash",
  "n01145905": "damper",
  "n01146039": "bridle, check, curb",
  "n01146288": "immobilization, immobilisation, immobilizing",
  "n01146576": "confinement",
  "n01146768": "imprisonment, internment",
  "n01146993": "lockdown",
  "n01147135": "house arrest",
  "n01147222": "false imprisonment",
  "n01147347": "custody",
  "n01147451": "containment",
  "n01147676": "ring containment",
  "n01147950": "suppression, curtailment",
  "n01148182": "crackdown",
  "n01148283": "regimentation",
  "n01148398": "reimposition",
  "n01148491": "restraint of trade",
  "n01148614": "restriction, confinement",
  "n01148904": "classification",
  "n01149115": "declassification",
  "n01149303": "stipulation, specification",
  "n01149480": "circumscription",
  "n01149621": "constraint",
  "n01149793": "swaddling clothes",
  "n01149911": "constriction",
  "n01150096": "vasoconstriction",
  "n01150200": "privation, deprivation",
  "n01150467": "pauperization, pauperisation, impoverishment",
  "n01150662": "starvation, starving",
  "n01150938": "appeasement, calming",
  "n01151097": "pacification, mollification",
  "n01151407": "placation, conciliation, propitiation",
  "n01151605": "internationalization, internationalisation",
  "n01151788": "nationalization, nationalisation, communization, communisation",
  "n01152033": "denationalization, denationalisation, privatization, privatisation",
  "n01152300": "nationalization, nationalisation",
  "n01152461": "nationalization, nationalisation",
  "n01152583": "detribalization, detribalisation",
  "n01152787": "collectivization, collectivisation",
  "n01152973": "communization, communisation",
  "n01153139": "communization, communisation",
  "n01153305": "federation",
  "n01153548": "discrimination, favoritism, favouritism",
  "n01153861": "patronage",
  "n01154083": "nomenklatura",
  "n01154237": "ableism, ablism, able-bodiedism, able-bodism",
  "n01154373": "ageism, agism",
  "n01154487": "cronyism",
  "n01154661": "fattism, fatism",
  "n01154772": "heterosexism",
  "n01154900": "nepotism",
  "n01155044": "racism, racialism, racial discrimination",
  "n01155253": "racial profiling",
  "n01155465": "secularization, secularisation",
  "n01155722": "rollover",
  "n01155893": "sexism",
  "n01156070": "male chauvinism, chauvinism, antifeminism",
  "n01156262": "sexual discrimination",
  "n01156438": "mobilization, mobilisation, militarization, militarisation",
  "n01156899": "arming, armament, equipping",
  "n01157138": "outfitting",
  "n01157275": "refit",
  "n01157384": "rearmament",
  "n01157557": "disarming, disarmament",
  "n01157850": "conscription, muster, draft, selective service",
  "n01158064": "levy, levy en masse",
  "n01158190": "demobilization, demobilisation",
  "n01158543": "remilitarization, remilitarisation",
  "n01158690": "standardization, standardisation, normalization, normalisation",
  "n01159025": "stabilization, stabilisation",
  "n01159461": "destabilization, destabilisation",
  "n01159776": "stylization, stylisation",
  "n01159964": "conventionalization, conventionalisation",
  "n01160170": "taxation",
  "n01160342": "punishment, penalty, penalization, penalisation",
  "n01160729": "beating, thrashing, licking, drubbing, lacing, trouncing, whacking",
  "n01161017": "castigation, chastisement",
  "n01161161": "corporal punishment",
  "n01161411": "cruel and unusual punishment",
  "n01161635": "detention",
  "n01161821": "discipline, correction",
  "n01162062": "economic strangulation",
  "n01162257": "self-flagellation",
  "n01162376": "imprisonment",
  "n01162529": "music, medicine",
  "n01162672": "self-punishment",
  "n01162784": "spanking",
  "n01162928": "stick",
  "n01163047": "whipping, tanning, flogging, lashing, flagellation",
  "n01163316": "flagellation",
  "n01163429": "horsewhipping",
  "n01163586": "electric shock",
  "n01163779": "execution, executing, capital punishment, death penalty",
  "n01164063": "gauntlet, gantlet",
  "n01164275": "kick in the butt",
  "n01164394": "stoning, lapidation",
  "n01164618": "burning, burning at the stake",
  "n01164730": "auto-da-fe",
  "n01164874": "hanging",
  "n01165112": "electrocution, burning",
  "n01165224": "decapitation, beheading",
  "n01165337": "crucifixion",
  "n01165537": "penance, self-mortification, self-abasement",
  "n01165692": "commitment, committal, consignment",
  "n01165919": "commutation, re-sentencing",
  "n01166092": "corrections",
  "n01166258": "exchange, interchange",
  "n01166517": "rally, exchange",
  "n01166760": "tradeoff, trade-off",
  "n01166926": "submission, compliance",
  "n01167146": "obedience, obeisance",
  "n01167385": "truckling",
  "n01167548": "prostration",
  "n01167710": "strife",
  "n01167847": "tug-of-war",
  "n01167952": "turf war",
  "n01168199": "countercurrent, crosscurrent",
  "n01168369": "direct action",
  "n01168569": "competition, contention, rivalry",
  "n01168840": "battle of wits",
  "n01168961": "contest",
  "n01169112": "bidding contest",
  "n01169200": "popularity contest",
  "n01169317": "resistance",
  "n01169598": "nonresistance",
  "n01169744": "confrontation, opposition",
  "n01170047": "sales resistance",
  "n01170175": "discord, discordance",
  "n01170320": "defiance",
  "n01170406": "road rage",
  "n01170502": "riot, public violence",
  "n01170666": "race riot",
  "n01170813": "dispute, contravention",
  "n01170962": "fight, fighting, combat, scrap",
  "n01171644": "fencing",
  "n01172031": "in-fighting",
  "n01172173": "set-to",
  "n01172252": "shock, impact",
  "n01172441": "hassle, scuffle, tussle, dogfight, rough-and-tumble",
  "n01172598": "aggro",
  "n01172784": "duel, affaire d'honneur",
  "n01173038": "blow",
  "n01173569": "counterblow",
  "n01173660": "swing",
  "n01173826": "fistfight, fisticuffs, slugfest",
  "n01173965": "stab, thrust, knife thrust",
  "n01174251": "lunge, straight thrust, passado",
  "n01174495": "parry",
  "n01174645": "remise",
  "n01174825": "riposte",
  "n01174988": "stinger",
  "n01175099": "thump",
  "n01175195": "uppercut",
  "n01175316": "hammer, pound, hammering, pounding",
  "n01175633": "shot",
  "n01175831": "cheap shot",
  "n01176031": "wallop",
  "n01176115": "battering, banging",
  "n01176219": "beating, whipping",
  "n01176335": "affray, disturbance, fray, ruffle",
  "n01176431": "brawl, free-for-all",
  "n01176540": "knife fight, snickersnee, cut-and-thrust",
  "n01176649": "rumble, gang fight",
  "n01176757": "single combat",
  "n01176931": "obstructionism",
  "n01177033": "protest, objection, dissent",
  "n01177327": "rebellion",
  "n01177583": "punch-up",
  "n01177703": "demonstration, manifestation",
  "n01177990": "counterdemonstration",
  "n01178195": "walkout",
  "n01178415": "Boston Tea Party",
  "n01178728": "peace march",
  "n01178850": "sit-in",
  "n01179004": "work-in",
  "n01179167": "protest march",
  "n01179393": "insubordination, rebelliousness",
  "n01179564": "contumacy",
  "n01179707": "disobedience, noncompliance",
  "n01179927": "civil disobedience",
  "n01180200": "contempt",
  "n01180398": "contempt of Congress",
  "n01180557": "contempt of court",
  "n01180718": "civil contempt",
  "n01180858": "contumacy",
  "n01181066": "criminal contempt",
  "n01181212": "obstruction of justice",
  "n01181475": "due process, due process of law",
  "n01181902": "legal action, action, action at law",
  "n01182312": "action",
  "n01182654": "lawsuit, suit, case, cause, causa",
  "n01183031": "civil suit",
  "n01183166": "class action, class-action suit",
  "n01183373": "countersuit",
  "n01183497": "criminal suit",
  "n01183638": "moot",
  "n01183798": "paternity suit, bastardy proceeding",
  "n01184045": "antitrust case",
  "n01184230": "civil action",
  "n01184407": "counterclaim",
  "n01184565": "custody case",
  "n01184720": "lis pendens",
  "n01184814": "proceeding, legal proceeding, proceedings",
  "n01185292": "adoption",
  "n01185611": "appeal",
  "n01185981": "reversal",
  "n01186192": "affirmation",
  "n01186397": "bankruptcy",
  "n01186578": "receivership",
  "n01186810": "litigation, judicial proceeding",
  "n01187085": "custody battle",
  "n01187230": "vexatious litigation",
  "n01187463": "presentment, notification",
  "n01187620": "naturalization, naturalisation",
  "n01187810": "judgment, judgement, judicial decision",
  "n01188273": "confession of judgment, confession of judgement, cognovit judgment, cognovit judgement",
  "n01188537": "default judgment, default judgement, judgment by default, judgement by default",
  "n01188783": "non prosequitur, non pros",
  "n01189001": "final judgment, final decision",
  "n01189282": "conviction, judgment of conviction, condemnation, sentence",
  "n01189650": "judgment in personam, judgement in personam, personal judgment, personal judgement",
  "n01189929": "judgment in rem, judgement in rem",
  "n01190172": "judgment of dismissal, judgement of dismissal, dismissal",
  "n01190364": "judgment on the merits, judgement on the merits",
  "n01190561": "summary judgment, summary judgement, judgment on the pleadings, judgement on the pleadings",
  "n01190884": "arbitration, arbitrament, arbitrement",
  "n01191158": "opinion, ruling",
  "n01191364": "Bakke decision",
  "n01191610": "fatwa",
  "n01191755": "umpirage, officiation, officiating, refereeing",
  "n01191975": "finding",
  "n01192150": "verdict, finding of fact",
  "n01192463": "finding of law, conclusion of law",
  "n01192633": "compromise verdict",
  "n01192814": "directed verdict",
  "n01193044": "false verdict",
  "n01193176": "general verdict",
  "n01193362": "partial verdict",
  "n01193535": "quotient verdict",
  "n01193672": "special verdict",
  "n01193886": "acquittal",
  "n01194021": "murder conviction",
  "n01194125": "rape conviction",
  "n01194225": "robbery conviction",
  "n01194331": "eviction, dispossession, legal ouster",
  "n01194615": "ouster",
  "n01194711": "actual eviction",
  "n01194904": "eviction, constructive eviction",
  "n01195187": "retaliatory eviction",
  "n01195380": "legalization, legalisation, legitimation",
  "n01195584": "legitimation",
  "n01195867": "trial",
  "n01196316": "court-martial",
  "n01196477": "ordeal, trial by ordeal",
  "n01196759": "Scopes trial",
  "n01197119": "show trial",
  "n01197258": "review",
  "n01197479": "bill of review",
  "n01197658": "judicial review",
  "n01197928": "plea",
  "n01198068": "double jeopardy",
  "n01198307": "prosecution, criminal prosecution",
  "n01198588": "test case, test suit",
  "n01198750": "defense, defence, denial, demurrer",
  "n01199035": "entrapment",
  "n01199252": "mistrial",
  "n01199365": "retrial",
  "n01199697": "hearing",
  "n01200000": "administrative hearing",
  "n01200266": "competence hearing",
  "n01200502": "fair hearing",
  "n01200857": "quo warranto",
  "n01201021": "separation",
  "n01201271": "divorce, divorcement",
  "n01201429": "legal separation, separation",
  "n01201645": "quarantine",
  "n01201773": "seclusion",
  "n01201906": "cocooning",
  "n01202029": "isolation, closing off",
  "n01202184": "segregation, sequestration",
  "n01202415": "integration, integrating, desegregation",
  "n01202665": "separationism, separatism",
  "n01202799": "withdrawal",
  "n01202904": "cooperation",
  "n01203277": "brainstorming",
  "n01203494": "teamwork",
  "n01203676": "conformity, conformation, compliance, abidance",
  "n01204055": "formality",
  "n01204294": "line",
  "n01204419": "honoring, observance",
  "n01204623": "punctilio",
  "n01204711": "nonobservance",
  "n01204845": "nonconformity, nonconformance",
  "n01205010": "keeping",
  "n01205156": "collaboration, coaction",
  "n01205341": "collaboration, collaborationism, quislingism",
  "n01205564": "compromise, via media",
  "n01205702": "concurrence, concurrency",
  "n01205827": "reconciliation, rapprochement",
  "n01205961": "selflessness, self-sacrifice",
  "n01206153": "commitment, allegiance, loyalty, dedication",
  "n01206553": "devotion",
  "n01206774": "cultism",
  "n01206910": "hobbyism",
  "n01206986": "enlistment",
  "n01207108": "reenlistment",
  "n01207187": "faith",
  "n01207342": "fetish, fetich",
  "n01207521": "party spirit",
  "n01207609": "aid, assist, assistance, help",
  "n01208291": "facilitation",
  "n01208460": "hand, helping hand",
  "n01208597": "recourse, resort, refuge",
  "n01208797": "thanks",
  "n01208924": "social work",
  "n01209079": "casework",
  "n01209220": "relief, succor, succour, ministration",
  "n01209487": "lift",
  "n01209576": "service",
  "n01209791": "disservice, ill service, ill turn",
  "n01209963": "childcare, child care",
  "n01210102": "community service",
  "n01210281": "community service, public service",
  "n01210439": "daycare, day care",
  "n01210547": "help desk, helpdesk",
  "n01210699": "seating",
  "n01210816": "accommodation",
  "n01211019": "boost, encouragement",
  "n01211185": "comfort",
  "n01211339": "boost",
  "n01211489": "morale building, morale booster",
  "n01211667": "consolation, comfort, solace",
  "n01211888": "simplification",
  "n01212037": "oversimplification, simplism",
  "n01212225": "rationalization, rationalisation",
  "n01212519": "support",
  "n01212882": "attachment, adherence, adhesion",
  "n01213234": "ecclesiasticism",
  "n01213406": "kabbalism, cabalism",
  "n01213548": "royalism",
  "n01213702": "traditionalism",
  "n01213886": "backing, backup, championship, patronage",
  "n01214171": "advocacy, protagonism",
  "n01214408": "drumbeat",
  "n01214557": "insistence, insistency",
  "n01214746": "urging",
  "n01214863": "auspices, protection, aegis",
  "n01215036": "sponsorship",
  "n01215168": "endorsement, indorsement",
  "n01215392": "blessing, approval, approving",
  "n01215719": "reassurance",
  "n01215902": "support",
  "n01216191": "sustenance, sustentation, sustainment, maintenance, upkeep",
  "n01216515": "logistic support, logistic assistance",
  "n01216738": "integrated logistic support",
  "n01216908": "mutual aid, international logistic support",
  "n01217052": "interdepartmental support, interagency support",
  "n01217306": "inter-service support",
  "n01217499": "representation",
  "n01217720": "proportional representation",
  "n01217859": "employment, engagement",
  "n01218050": "shape-up",
  "n01218213": "call-back",
  "n01218327": "booking, reservation",
  "n01218593": "admiration, appreciation",
  "n01218766": "adoration, idolization, idolisation",
  "n01218932": "glorification",
  "n01219075": "idealization, idealisation, glorification",
  "n01219306": "sentimentalization, sentimentalisation, romanticization, romanticisation",
  "n01219551": "reward, reinforcement",
  "n01219722": "carrot",
  "n01219893": "disparagement, dispraise",
  "n01220074": "belittling",
  "n01220152": "deprecation, denigration",
  "n01220336": "aspersion, calumny, slander, defamation, denigration",
  "n01220619": "attack",
  "n01220754": "detraction",
  "n01220984": "behavior, behaviour, conduct, doings",
  "n01221254": "behavior, behaviour",
  "n01221464": "territoriality",
  "n01221611": "aggression",
  "n01221790": "aggravation, irritation, provocation",
  "n01222034": "last straw",
  "n01222157": "exacerbation",
  "n01222333": "bitchery",
  "n01222477": "bullying, intimidation",
  "n01222666": "terrorization, terrorisation, frightening",
  "n01222859": "twit, taunt, taunting",
  "n01223032": "raising hell, hell raising",
  "n01223143": "self-assertion",
  "n01223257": "condemnation",
  "n01223488": "stigmatization, stigmatisation, branding",
  "n01223672": "bohemianism",
  "n01223766": "dirty pool",
  "n01223877": "dirty tricks",
  "n01224031": "discourtesy, offense, offence, offensive activity",
  "n01224346": "easiness",
  "n01224517": "derision, ridicule",
  "n01224700": "mock",
  "n01224828": "indelicacy",
  "n01224914": "insolence",
  "n01225027": "insult, affront",
  "n01225299": "indignity",
  "n01225397": "scandalization, scandalisation, outrage",
  "n01225562": "presumption",
  "n01225783": "rebuff, slight",
  "n01225997": "snub, cut, cold shoulder",
  "n01226181": "silent treatment",
  "n01226289": "the way of the world, the ways of the world",
  "n01226679": "benevolence, benefaction",
  "n01226837": "cupboard love",
  "n01226941": "favor, favour",
  "n01227083": "turn, good turn",
  "n01227190": "forgiveness, pardon",
  "n01227351": "condonation",
  "n01227495": "mercy",
  "n01227691": "exculpation",
  "n01227805": "endearment",
  "n01227908": "politeness, civility",
  "n01228102": "reverence",
  "n01228245": "courtesy",
  "n01228380": "gesture",
  "n01228544": "beau geste",
  "n01228645": "attention",
  "n01228792": "gallantry",
  "n01228877": "deference, respect",
  "n01229223": "court, homage",
  "n01229350": "last respects",
  "n01229549": "props",
  "n01229712": "devoir",
  "n01229793": "consideration, thoughtfulness",
  "n01229938": "assembly, assemblage, gathering",
  "n01230283": "mobilization, mobilisation",
  "n01230566": "economic mobilization, economic mobilisation",
  "n01230685": "rallying",
  "n01230850": "convocation, calling together",
  "n01230965": "meeting, coming together",
  "n01231296": "service call",
  "n01231435": "assignation, tryst",
  "n01231548": "rendezvous",
  "n01231686": "congregation, congregating",
  "n01231819": "convention, convening",
  "n01231980": "concentration",
  "n01232091": "session",
  "n01232246": "course session, class period, recitation",
  "n01232412": "socialization, socialisation, socializing, socialising",
  "n01232684": "visit",
  "n01232842": "visit",
  "n01232978": "flying visit",
  "n01233156": "visit",
  "n01233397": "attendance, attending",
  "n01233627": "appearance, appearing, coming into court",
  "n01233802": "presence",
  "n01233917": "turnout",
  "n01234090": "nonattendance",
  "n01234243": "nonappearance",
  "n01234345": "absence",
  "n01234528": "absenteeism",
  "n01234612": "truancy, hooky",
  "n01234729": "return, paying back, getting even",
  "n01234952": "answer",
  "n01235137": "requital, payment",
  "n01235258": "retaliation, revenge",
  "n01235463": "vengeance, retribution, payback",
  "n01235946": "reprisal",
  "n01236049": "reciprocation",
  "n01236173": "feud",
  "n01236296": "war, warfare",
  "n01236491": "drug war",
  "n01236633": "trench warfare",
  "n01236947": "vendetta, blood feud",
  "n01237080": "tit for tat",
  "n01237167": "aggression",
  "n01237294": "democratization, democratisation",
  "n01237415": "consolidation, integration",
  "n01237872": "centralization, centralisation",
  "n01238058": "decentralization, decentralisation",
  "n01238267": "incorporation",
  "n01238424": "amalgamation, merger, uniting",
  "n01238607": "vertical integration, vertical combination",
  "n01238847": "horizontal integration, horizontal combination",
  "n01239064": "engagement, participation, involvement, involution",
  "n01239487": "non-engagement, nonparticipation, non-involvement",
  "n01239717": "isolation",
  "n01239868": "commitment",
  "n01240064": "incurrence",
  "n01240210": "intervention, intercession",
  "n01240432": "mediation, intermediation",
  "n01240628": "matchmaking",
  "n01240745": "group participation",
  "n01240850": "neutrality",
  "n01240979": "annulment, invalidation",
  "n01241216": "dissolution of marriage",
  "n01241331": "vindication, exoneration",
  "n01241594": "whitewash",
  "n01241767": "justification",
  "n01241997": "rehabilitation",
  "n01242164": "job action",
  "n01242354": "go-slow",
  "n01242539": "work to rule",
  "n01242716": "passive resistance, nonviolent resistance, nonviolence",
  "n01242962": "hunger strike",
  "n01243089": "Ramadan",
  "n01243262": "Satyagraha",
  "n01243453": "recusancy",
  "n01243674": "strike, work stoppage",
  "n01243969": "sit-down, sit-down strike",
  "n01244127": "sympathy strike, sympathetic strike",
  "n01244341": "walkout",
  "n01244451": "wildcat strike",
  "n01244593": "unsnarling, untangling, disentanglement, extrication",
  "n01244895": "sabotage",
  "n01245061": "extermination, liquidation",
  "n01245159": "genocide, race murder, racial extermination",
  "n01245318": "holocaust",
  "n01245471": "Holocaust, final solution",
  "n01245618": "throw",
  "n01245813": "cast, roll",
  "n01245950": "natural",
  "n01246086": "flip, toss",
  "n01246206": "flip",
  "n01246334": "strafe",
  "n01246541": "surprise attack, coup de main",
  "n01246697": "terrorist attack",
  "n01246926": "ambush, ambuscade, lying in wait, trap",
  "n01247156": "pre-emptive strike",
  "n01247306": "dry-gulching",
  "n01247413": "emancipation",
  "n01247647": "clearing",
  "n01247807": "manumission",
  "n01247969": "radio observation",
  "n01248075": "stupidity, betise, folly, foolishness, imbecility",
  "n01248191": "admission, admittance",
  "n01248418": "readmission",
  "n01248608": "matriculation, matric",
  "n01248752": "remarriage",
  "n01248852": "renewal",
  "n01248961": "self-renewal",
  "n01249060": "replication",
  "n01249315": "amnesty, pardon, free pardon",
  "n01249483": "demolition",
  "n01249616": "spoliation",
  "n01249816": "vandalism, hooliganism, malicious mischief",
  "n01249991": "recession, ceding back",
  "n01250101": "amendment",
  "n01250208": "emendation",
  "n01250335": "hit",
  "n01250492": "infanticide",
  "n01250569": "shoot-down",
  "n01250671": "tyrannicide",
  "n01250745": "thuggee",
  "n01250826": "transmutation, transubstantiation",
  "n01251010": "barrage jamming",
  "n01251138": "spot jamming, selective jamming",
  "n01251270": "electronic deception",
  "n01251489": "manipulative electronic deception, electronic manipulative deception",
  "n01251724": "simulative electronic deception, electronic simulative deception",
  "n01251923": "imitative electronic deception, electronic imitative deception",
  "n01252124": "waste, permissive waste",
  "n01252280": "colonization, colonisation, settlement",
  "n01252566": "resettlement, relocation",
  "n01252800": "dismount",
  "n01252918": "radiation",
  "n01253060": "emission, emanation",
  "n01253277": "discharge, venting",
  "n01253379": "jamming, electronic jamming, jam",
  "n01253665": "vacation",
  "n01253778": "harmonization, harmonisation",
  "n01253944": "humming",
  "n01254051": "winnow, winnowing, sifting",
  "n01254253": "separation",
  "n01254473": "teleportation",
  "n01254685": "intonation, chanting",
  "n01254882": "cantillation",
  "n01254978": "intonation",
  "n01255125": "fixed intonation",
  "n01255302": "karaoke",
  "n01255452": "part-singing",
  "n01255549": "psalmody, hymnody",
  "n01255648": "singalong, singsong",
  "n01255808": "solfege, solfeggio",
  "n01255935": "solmization, solfege, solfeggio",
  "n01256124": "yodeling",
  "n01256270": "lead",
  "n01256417": "leadership, leading",
  "n01256615": "helm",
  "n01256743": "lead",
  "n01257033": "trend setting",
  "n01257145": "precession, precedence, precedency",
  "n01257418": "solo",
  "n01257542": "flood, flowage",
  "n01257701": "parole",
  "n01257969": "population",
  "n01258161": "pounce",
  "n01258251": "probation",
  "n01258579": "quarter",
  "n01258719": "recall",
  "n01258852": "revocation",
  "n01259034": "reprieve, respite",
  "n01259211": "revoke, renege",
  "n01259380": "ruff, trumping",
  "n01259594": "trick",
  "n01259773": "awakening, wakening, waking up",
  "n01260008": "buzz",
  "n01260182": "fixation",
  "n01260309": "immobilization, immobilisation",
  "n01260556": "fun",
  "n01260731": "sin, hell",
  "n01260867": "excitation, excitement",
  "n01261113": "hair-raiser, chiller",
  "n01261293": "thrill",
  "n01261490": "incitation, incitement",
  "n01261712": "inflammation, inflaming",
  "n01261845": "inspiration, stirring",
  "n01261974": "stimulation",
  "n01262109": "galvanization, galvanisation",
  "n01262345": "titillation",
  "n01262441": "deforestation, disforestation",
  "n01262574": "skimming",
  "n01262713": "withdrawal",
  "n01262868": "withdrawal",
  "n01263018": "spoil, spoiling, spoilage",
  "n01263257": "swerve, swerving, veering",
  "n01263418": "three-point turn",
  "n01263582": "face saver, face saving",
  "n01263711": "recruitment, enlisting",
  "n01263924": "smooth",
  "n01264050": "reference, consultation",
  "n01264243": "emphasizing, accenting, accentuation",
  "n01264447": "release, outlet, vent",
  "n01264667": "last",
  "n01264795": "slapshot",
  "n01264933": "headshot",
  "n01265071": "cornhusking",
  "n01265176": "palpebration",
  "n01265325": "bank examination",
  "n01265475": "beatification",
  "n01265632": "equilibration",
  "n01265734": "ethnic cleansing",
  "n01265921": "jumpstart, jump-start",
  "n01266152": "mystification, obfuscation",
  "n01266364": "negotiation",
  "n01266491": "proclamation, promulgation",
  "n01266745": "socialization, socialisation",
  "n01266945": "stabilization, stabilisation",
  "n01267135": "stupefaction",
  "n01267344": "transfusion reaction",
  "n01267588": "upgrade",
  "n01267808": "vampirism",
  "n01267901": "version",
  "n01268020": "vulgarization, vulgarisation",
  "n01268183": "witching",
  "n01268271": "xenotransplant, xenotransplantation",
  "n01268457": "Actium",
  "n01268633": "Aegates Isles, Aegadean Isles",
  "n01268886": "Aegospotami, Aegospotamos",
  "n01269161": "Agincourt",
  "n01269360": "Alamo",
  "n01269633": "Atlanta, battle of Atlanta",
  "n01269857": "Austerlitz, battle of Austerlitz",
  "n01270153": "Bannockburn",
  "n01270343": "Bataan, Corregidor",
  "n01270628": "Battle of Britain",
  "n01270860": "Battle of Kerbala",
  "n01271107": "Battle of the Ardennes Bulge, Battle of the Bulge, Ardennes counteroffensive",
  "n01271428": "Battle of the Marne, Belleau Wood, Chateau-Thierry, Marne River",
  "n01271669": "Bismarck Sea, battle of the Bismarck Sea",
  "n01271915": "Blenheim",
  "n01272134": "Borodino",
  "n01272367": "Bosworth Field",
  "n01272582": "Bouvines",
  "n01272787": "Boyne, battle of Boyne",
  "n01273081": "Brunanburh, battle of Brunanburh",
  "n01273230": "Buena Vista",
  "n01273491": "Bull Run, Battle of Bull Run",
  "n01273735": "Bunker Hill, battle of Bunker Hill",
  "n01274000": "Cannae",
  "n01274171": "Caporetto, battle of Caporetto",
  "n01274374": "Caudine Forks",
  "n01274531": "Chaeronea",
  "n01274733": "Chalons, Chalons-sur-Marne",
  "n01274909": "Chancellorsville",
  "n01275142": "Chapultepec",
  "n01275389": "Chattanooga, battle of Chattanooga",
  "n01275697": "Chickamauga, battle of Chickamauga",
  "n01275934": "Chino-Japanese War, Sino-Japanese War",
  "n01276194": "Coral Sea, battle of the Coral Sea",
  "n01276436": "Cowpens, battle of Cowpens",
  "n01276634": "Crecy, battle of Crecy",
  "n01276875": "Cunaxa, battle of Cunaxa",
  "n01277065": "Cynoscephalae, battle of Cynoscephalae",
  "n01277288": "Dardanelles, Dardanelles campaign",
  "n01277540": "Dien Bien Phu",
  "n01277755": "Drogheda",
  "n01277938": "Dunkirk, Dunkerque",
  "n01278232": "El Alamein, Al Alamayn, Battle of El Alamein",
  "n01278509": "Eniwetok",
  "n01278692": "Flodden, Battle of Flodden Field",
  "n01278873": "Fontenoy, Battle of Fontenoy",
  "n01279120": "Fort Ticonderoga, Ticonderoga",
  "n01279342": "Fredericksburg, Battle of Fredericksburg",
  "n01279615": "Gettysburg, Battle of Gettysburg",
  "n01279866": "Granicus, Battle of Granicus River",
  "n01280055": "Guadalcanal, Battle of Guadalcanal",
  "n01280308": "Hampton Roads",
  "n01280514": "Hastings, battle of Hastings",
  "n01280792": "Hohenlinden, battle of Hohenlinden",
  "n01280990": "Inchon",
  "n01281154": "Indian Mutiny, Sepoy Mutiny",
  "n01281427": "Ipsus, battle of Ipsus",
  "n01281638": "Issus, battle of Issus",
  "n01281813": "Ivry, battle of Ivry, Ivry la Bataille",
  "n01282022": "Iwo, Iwo Jima, invasion of Iwo",
  "n01282289": "Jena, Battle of Jena",
  "n01282466": "Jutland, battle of Jutland",
  "n01282711": "Kennesaw Mountain",
  "n01282966": "Kwajalein",
  "n01283185": "Lake Trasimenus, Battle of Lake Trasimenus",
  "n01283389": "Langside, battle of Langside",
  "n01283565": "Lepanto, Battle of Lepanto",
  "n01283753": "Leuctra, battle of Leuctra",
  "n01283935": "Lexington, Concord, Lexington and Concord",
  "n01284124": "Leyte, Leyte Island, Leyte invasion",
  "n01284444": "Little Bighorn, Battle of Little Bighorn, Battle of the Little Bighorn, Custer's Last Stand",
  "n01284928": "Lucknow",
  "n01285101": "Lule Burgas, battle of Lule Burgas",
  "n01285305": "Lutzen, battle of Lutzen",
  "n01285567": "Macedonian War",
  "n01285784": "Magenta, Battle of Magenta",
  "n01286000": "Maldon, Battle of Maldon",
  "n01286181": "Manila Bay",
  "n01286393": "Mantinea, Mantineia",
  "n01286569": "Marathon, battle of Marathon",
  "n01286745": "Marengo",
  "n01286938": "Marston Moor, battle of Marston Moor",
  "n01287179": "Metaurus River",
  "n01287431": "Meuse, Meuse River, Argonne, Argonne Forest, Meuse-Argonne, Meuse-Argonne operation",
  "n01287782": "Midway, Battle of Midway",
  "n01288057": "Minden, battle of Minden",
  "n01288272": "Monmouth Court House, Battle of Monmouth Court House, Battle of Monmouth",
  "n01288549": "Naseby, Battle of Naseby",
  "n01288795": "Navarino, battle of Navarino",
  "n01289061": "Okinawa, Okinawa campaign",
  "n01289444": "Omdurman, battle of Omdurman",
  "n01289631": "Operation Desert Storm",
  "n01289830": "Orleans, siege of Orleans",
  "n01289997": "Panipat, battle of Panipat",
  "n01290162": "Passero, Cape Passero",
  "n01290435": "Petersburg, Petersburg Campaign",
  "n01290711": "Pharsalus, battle of Pharsalus",
  "n01290840": "Philippi, battle of Philippi",
  "n01290997": "Philippine Sea, battle of the Philippine Sea",
  "n01291310": "Plassey, battle of Plassey",
  "n01291517": "Plataea, battle of Plataea",
  "n01291674": "Plevna, Pleven",
  "n01291839": "Poitiers, battle of Poitiers",
  "n01292017": "Port Arthur",
  "n01292200": "Battle of Puebla",
  "n01292343": "Pydna, Battle of Pydna",
  "n01292553": "Ravenna, Battle of Ravenna",
  "n01292735": "Rocroi, Battle of Rocroi",
  "n01292928": "Rossbach, battle of Rossbach",
  "n01293167": "Saint-Mihiel, St Mihiel, battle of St Mihiel",
  "n01293438": "Saipan",
  "n01293650": "Salerno",
  "n01293832": "Santiago, Santiago de Cuba",
  "n01294127": "Saratoga, battle of Saratoga",
  "n01294330": "Sempatch, battle of Sempatch",
  "n01294502": "Shiloh, battle of Shiloh, battle of Pittsburgh Landing",
  "n01294791": "Soissons, battle of Soissons-Reims, battle of the Chemin-des-Dames, battle of the Aisne",
  "n01295153": "Solferino, battle of Solferino",
  "n01295373": "Somme, Somme River, Battle of the Somme",
  "n01295528": "Somme, Somme River, Battle of the Somme",
  "n01295684": "Battle of the Spanish Armada",
  "n01295918": "Spotsylvania, battle of Spotsylvania Courthouse",
  "n01296127": "Syracuse, siege of Syracuse",
  "n01296296": "Syracuse, siege of Syracuse",
  "n01296505": "Tannenberg, battle of Tannenberg",
  "n01296697": "Tarawa, Makin, Tarawa-Makin",
  "n01296946": "Tertry, battle of Tertry",
  "n01297095": "Teutoburger Wald, battle of Teutoburger Wald",
  "n01297291": "Tewkesbury, battle of Tewkesbury",
  "n01297484": "Thermopylae, battle of Thermopylae",
  "n01297706": "Trafalgar, battle of Trafalgar",
  "n01297978": "Trasimeno, battle of Trasimeno",
  "n01298162": "Tsushima",
  "n01298371": "Valmy, battle of Valmy",
  "n01298573": "Verdun, battle of Verdun",
  "n01298797": "Vicksburg, siege of Vicksburg",
  "n01299037": "Wagram, battle of Wagram",
  "n01299224": "Battle of Wake, Battle of Wake Island",
  "n01299476": "Waterloo, Battle of Waterloo",
  "n01299735": "Wilderness Campaign",
  "n01299994": "Yalu River",
  "n01300242": "Yorktown, siege of Yorktown",
  "n01300508": "Ypres, battle of Ypres, first battle of Ypres",
  "n01300782": "Ypres, battle of Ypres, second battle of Ypres",
  "n01301080": "Ypres, battle of Ypres, third battle of Ypres",
  "n01301423": "Zama, battle of Zama",
  "n01301630": "American Civil War, United States Civil War, War between the States",
  "n01302086": "American Revolution, American Revolutionary War, War of American Independence, American War of Independence",
  "n01302449": "Arab-Israeli War, Six-Day War, Six Day War",
  "n01302683": "Arab-Israeli War, Yom Kippur War",
  "n01302935": "Balkan Wars",
  "n01303252": "Boer War",
  "n01303582": "Chinese Revolution",
  "n01303739": "Crimean War",
  "n01303934": "Cuban Revolution",
  "n01304121": "English Civil War",
  "n01304356": "English Revolution, Glorious Revolution, Bloodless Revolution",
  "n01304629": "Franco-Prussian War",
  "n01304820": "French and Indian War",
  "n01305007": "French Revolution",
  "n01305147": "Hundred Years' War",
  "n01305310": "Iran-Iraq War, Gulf War",
  "n01305551": "Korean War",
  "n01305796": "Mexican Revolution",
  "n01306007": "Mexican War",
  "n01306358": "Napoleonic Wars",
  "n01306736": "Norman Conquest",
  "n01306911": "Peloponnesian War",
  "n01307090": "Persian Gulf War, Gulf War",
  "n01307299": "Punic War",
  "n01307606": "Restoration",
  "n01307754": "Russian Revolution, February Revolution",
  "n01308008": "Russian Revolution, October Revolution",
  "n01308260": "Russo-Japanese War",
  "n01308438": "Seven Years' War",
  "n01308668": "Spanish-American War, Spanish War",
  "n01308837": "Spanish Civil War",
  "n01309109": "Thirty Years' War",
  "n01309395": "Trojan War",
  "n01309807": "Vietnam War, Vietnam",
  "n01310074": "War of Greek Independence",
  "n01310347": "War of the Austrian Succession",
  "n01310536": "War of the Grand Alliance, War of the League of Augsburg",
  "n01310789": "War of the Spanish Succession",
  "n01311045": "War of the Roses, Wars of the Roses",
  "n01311344": "War of 1812",
  "n01311520": "World War I, World War 1, Great War, First World War, War to End War",
  "n01312096": "World War II, World War 2, Second World War",
  "n01313093": "Animalia, kingdom Animalia, animal kingdom",
  "n01313888": "recombinant",
  "n01314026": "conspecific",
  "n01314145": "carrier",
  "n01314388": "pest",
  "n01314663": "critter",
  "n01314781": "creepy-crawly",
  "n01314910": "darter",
  "n01315062": "denizen",
  "n01315213": "peeper",
  "n01315330": "homeotherm, homoiotherm, homotherm",
  "n01315581": "poikilotherm, ectotherm",
  "n01315805": "range animal",
  "n01315980": "vermin",
  "n01316288": "varmint, varment",
  "n01316422": "scavenger",
  "n01316579": "bottom-feeder, bottom-dweller",
  "n01316734": "bottom-feeder",
  "n01316838": "bottom lurkers",
  "n01316949": "work animal",
  "n01317089": "beast of burden, jument",
  "n01317294": "draft animal",
  "n01317391": "pack animal, sumpter",
  "n01317541": "domestic animal, domesticated animal",
  "n01317813": "feeder",
  "n01317916": "feeder",
  "n01318053": "stocker",
  "n01318279": "hatchling",
  "n01318381": "head",
  "n01318478": "migrator",
  "n01318660": "molter, moulter",
  "n01318894": "pet",
  "n01319001": "stayer",
  "n01319187": "stunt",
  "n01319339": "pollard",
  "n01319467": "marine animal, marine creature, sea animal, sea creature",
  "n01319685": "by-catch, bycatch",
  "n01319932": "amphidiploid",
  "n01320093": "diploid",
  "n01320314": "haploid",
  "n01320479": "heteroploid",
  "n01320692": "polyploid",
  "n01320872": "female",
  "n01321123": "hen",
  "n01321230": "male",
  "n01321456": "adult",
  "n01321579": "young, offspring",
  "n01321770": "orphan",
  "n01321854": "young mammal",
  "n01322221": "baby",
  "n01322343": "pup, whelp",
  "n01322508": "wolf pup, wolf cub",
  "n01322604": "puppy",
  "n01322685": "cub, young carnivore",
  "n01322898": "lion cub",
  "n01322983": "bear cub",
  "n01323068": "tiger cub",
  "n01323155": "kit",
  "n01323261": "suckling",
  "n01323355": "sire",
  "n01323493": "dam",
  "n01323599": "thoroughbred, purebred, pureblood",
  "n01323781": "giant",
  "n01323901": "vent",
  "n01324019": "animalcule, animalculum",
  "n01324142": "survivor",
  "n01324305": "mutant",
  "n01324431": "carnivore",
  "n01324610": "herbivore",
  "n01324799": "insectivore",
  "n01324916": "acrodont",
  "n01325060": "pleurodont",
  "n01325206": "form genus",
  "n01325417": "horn",
  "n01325658": "antler",
  "n01325775": "tuft",
  "n01325853": "horn",
  "n01326015": "crest",
  "n01326198": "topknot",
  "n01326291": "microorganism, micro-organism",
  "n01326751": "monad",
  "n01326897": "aerobe",
  "n01327028": "anaerobe",
  "n01327204": "obligate anaerobe",
  "n01327322": "hybrid, crossbreed, cross",
  "n01327765": "dihybrid",
  "n01327909": "monohybrid",
  "n01328121": "polymorph",
  "n01328302": "relative, congener, congenator, congeneric",
  "n01328515": "intestinal flora",
  "n01328702": "virus",
  "n01329186": "arbovirus, arborvirus",
  "n01329794": "capsid",
  "n01329935": "virion",
  "n01330126": "adenovirus",
  "n01330314": "parainfluenza virus",
  "n01330497": "arenavirus",
  "n01330696": "Junin virus",
  "n01330852": "Lassa virus",
  "n01330966": "lymphocytic choriomeningitis virus",
  "n01331188": "Machupo virus",
  "n01331345": "Bunyaviridae",
  "n01331522": "bunyavirus",
  "n01331659": "Filoviridae",
  "n01331867": "filovirus",
  "n01332009": "Ebola virus",
  "n01332181": "Marburg virus",
  "n01332347": "Togaviridae",
  "n01332488": "alphavirus",
  "n01332653": "Flaviviridae",
  "n01332795": "flavivirus",
  "n01332940": "West Nile virus, West Nile encephalitis virus",
  "n01333082": "Arenaviridae",
  "n01333301": "Rhabdoviridae",
  "n01333483": "vesiculovirus",
  "n01333610": "Reoviridae",
  "n01333731": "poxvirus",
  "n01333890": "myxoma virus",
  "n01334036": "variola virus, smallpox virus",
  "n01334217": "variola major, variola major virus",
  "n01334368": "variola minor, variola minor virus",
  "n01334518": "tobacco mosaic virus, TMV",
  "n01334690": "viroid, virusoid",
  "n01334862": "bacteriophage, phage",
  "n01335218": "coliphage",
  "n01335333": "typhoid bacteriophage",
  "n01335460": "plant virus",
  "n01335659": "animal virus",
  "n01336076": "hepadnavirus",
  "n01336282": "retrovirus",
  "n01336599": "human T-cell leukemia virus-1, HTLV-1",
  "n01336718": "human immunodeficiency virus, HIV",
  "n01336910": "myxovirus",
  "n01337078": "orthomyxovirus",
  "n01337191": "paramyxovirus",
  "n01337329": "respiratory syncytial virus",
  "n01337583": "picornavirus",
  "n01337734": "poliovirus",
  "n01337822": "hepatitis A virus",
  "n01337915": "enterovirus",
  "n01338178": "coxsackievirus, Coxsackie virus",
  "n01338333": "echovirus",
  "n01338537": "rhinovirus",
  "n01338685": "herpes, herpes virus",
  "n01338908": "herpes simplex, herpes simplex virus",
  "n01339083": "herpes simplex 1, HS1, HSV-1, HSV-I",
  "n01339206": "herpes simplex 2, HS2, HSV-2, HSV-II",
  "n01339336": "herpes zoster, herpes zoster virus",
  "n01339471": "herpes varicella zoster, herpes varicella zoster virus",
  "n01339623": "Epstein-Barr virus, EBV",
  "n01339801": "cytomegalovirus, CMV",
  "n01340014": "varicella zoster virus",
  "n01340155": "papovavirus",
  "n01340331": "human papilloma virus",
  "n01340522": "polyoma, polyoma virus",
  "n01340648": "rhabdovirus",
  "n01340785": "lyssavirus",
  "n01340935": "reovirus",
  "n01341090": "rotavirus",
  "n01341183": "parvovirus, parvo",
  "n01341410": "slow virus",
  "n01341580": "onion yellow-dwarf virus",
  "n01341727": "potato yellow-dwarf virus",
  "n01341876": "Monera, kingdom Monera, Prokayotae, kingdom Prokaryotae",
  "n01342269": "moneran, moneron",
  "n01342529": "animal order",
  "n01347199": "protoctist order",
  "n01347431": "division Archaebacteria",
  "n01347583": "archaebacteria, archaebacterium, archaeobacteria, archeobacteria",
  "n01347858": "methanogen",
  "n01348075": "halophile, halophil",
  "n01348232": "halobacteria, halobacterium, halobacter",
  "n01348388": "thermoacidophile",
  "n01348530": "bacteria, bacterium",
  "n01349333": "acidophil, acidophile",
  "n01349495": "probiotic, probiotic bacterium, probiotic microflora, probiotic flora",
  "n01349735": "bacteroid",
  "n01349948": "bacillus, B",
  "n01350226": "Bacillus anthracis, anthrax bacillus",
  "n01350485": "Bacillus subtilis, Bacillus globigii, grass bacillus, hay bacillus",
  "n01350701": "Yersinia pestis",
  "n01350855": "coccus, cocci",
  "n01350994": "coccobacillus",
  "n01351170": "Brucella",
  "n01351315": "spirillum, spirilla",
  "n01351453": "Heliobacter, genus Heliobacter",
  "n01351688": "Heliobacter pylori, H. pylori",
  "n01351959": "bacteria order",
  "n01352059": "bacteria family",
  "n01352574": "bacteria genus",
  "n01353411": "bacteria species",
  "n01353658": "Pseudomonas pyocanea",
  "n01353773": "Aerobacter, genus Aerobacter",
  "n01353932": "Aerobacter aerogenes",
  "n01354149": "Rhizobiaceae, family Rhizobiaceae",
  "n01354324": "Rhizobium, genus Rhizobium",
  "n01354521": "Agrobacterium, genus Agrobacterium",
  "n01354724": "Agrobacterium tumefaciens",
  "n01354869": "division Eubacteria",
  "n01355326": "eubacteria, eubacterium, true bacteria",
  "n01356086": "Eubacteriales, order Eubacteriales",
  "n01356459": "Bacillaceae, family Bacillaceae",
  "n01356670": "genus Bacillus",
  "n01356888": "genus Clostridium",
  "n01357156": "clostridium, clostridia",
  "n01357328": "botulinus, botulinum, Clostridium botulinum",
  "n01357507": "clostridium perfringens",
  "n01357707": "Cyanophyta, division Cyanophyta",
  "n01357967": "Schizophyta, division Schizophyta",
  "n01358135": "Schizomycetes, class Schizomycetes",
  "n01358259": "class Cyanobacteria, Cyanophyceae, class Cyanophyceae",
  "n01358572": "cyanobacteria, blue-green algae",
  "n01358904": "Myxophyceae, family Myxophyceae, Schizophyceae, family Schizophyceae",
  "n01359070": "Nostocaceae, family Nostocaceae",
  "n01359203": "genus Nostoc",
  "n01359365": "nostoc",
  "n01359488": "Oscillatoriaceae, family Oscillatoriaceae",
  "n01359631": "genus Trichodesmium",
  "n01359762": "trichodesmium",
  "n01359945": "phototrophic bacteria, phototropic bacteria",
  "n01360180": "purple bacteria",
  "n01360330": "Pseudomonadales, order Pseudomonadales",
  "n01360712": "Pseudomonodaceae, family Pseudomonodaceae",
  "n01360937": "Pseudomonas, genus Pseudomonas",
  "n01361113": "ring rot bacteria, Pseudomonas solanacearum",
  "n01361261": "pseudomonad",
  "n01361465": "Xanthomonas, genus Xanthomonas",
  "n01361683": "xanthomonad",
  "n01361840": "Athiorhodaceae, family Athiorhodaceae",
  "n01361973": "Nitrobacteriaceae, family Nitrobacteriaceae",
  "n01362196": "Nitrobacter, genus Nitrobacter",
  "n01362336": "nitric bacteria, nitrobacteria",
  "n01362480": "Nitrosomonas, genus Nitrosomonas",
  "n01362623": "nitrosobacteria, nitrous bacteria",
  "n01362769": "Thiobacteriaceae, family Thiobacteriaceae",
  "n01362999": "genus Thiobacillus",
  "n01363121": "thiobacillus",
  "n01363288": "thiobacteria, sulphur bacteria, sulfur bacteria",
  "n01363423": "Spirillaceae, family Spirillaceae",
  "n01363600": "genus Spirillum",
  "n01363719": "spirillum",
  "n01363887": "ratbite fever bacterium, Spirillum minus",
  "n01364008": "genus Vibrio",
  "n01364162": "vibrio, vibrion",
  "n01364329": "comma bacillus, Vibrio comma",
  "n01364472": "Vibrio fetus",
  "n01364587": "Bacteroidaceae, family Bacteroidaceae",
  "n01364866": "Bacteroides, genus Bacteroides",
  "n01365122": "Calymmatobacterium, genus Calymmatobacterium",
  "n01365322": "Calymmatobacterium granulomatis",
  "n01365474": "Francisella, genus Francisella",
  "n01365684": "Francisella tularensis",
  "n01365885": "gonococcus, Neisseria gonorrhoeae",
  "n01366015": "Corynebacteriaceae, family Corynebacteriaceae",
  "n01366276": "corynebacterium",
  "n01366415": "genus Corynebacterium",
  "n01366700": "Corynebacterium diphtheriae, C. diphtheriae, Klebs-Loeffler bacillus",
  "n01366881": "genus Listeria",
  "n01367083": "listeria",
  "n01367208": "Listeria monocytogenes, L. monocytogenes",
  "n01367430": "Enterobacteriaceae, family Enterobacteriaceae",
  "n01367772": "enteric bacteria, enterobacteria, enterics, entric",
  "n01368095": "genus Escherichia",
  "n01368216": "escherichia",
  "n01368338": "Escherichia coli, E. coli",
  "n01368552": "genus Klebsiella",
  "n01368672": "klebsiella",
  "n01368853": "genus Salmonella",
  "n01368973": "salmonella",
  "n01369210": "Salmonella enteritidis, Gartner's bacillus",
  "n01369358": "Salmonella typhimurium",
  "n01369484": "typhoid bacillus, Salmonella typhosa, Salmonella typhi",
  "n01369633": "genus Serratia, Serratia",
  "n01369819": "Serratia marcescens",
  "n01370142": "genus Shigella",
  "n01370260": "shigella",
  "n01370465": "shiga bacillus, Shigella dysentariae",
  "n01370581": "genus Erwinia",
  "n01370698": "erwinia",
  "n01370816": "endospore-forming bacteria",
  "n01370913": "Rickettsiales, order Rickettsiales",
  "n01371092": "Rickettsiaceae, family Rickettsiaceae",
  "n01371342": "genus Rickettsia",
  "n01371483": "rickettsia",
  "n01371777": "tumor virus",
  "n01371928": "wound tumor virus, WTV",
  "n01372038": "vector",
  "n01372199": "cosmid",
  "n01372372": "Chlamydiaceae, family Chlamydiaceae",
  "n01372544": "genus Chlamydia",
  "n01372709": "chlamydia",
  "n01372944": "Chlamydia psittaci, C. psittaci",
  "n01373090": "Chlamydia trachomatis, C. trachomatis",
  "n01373272": "Mycoplasmatales, order Mycoplasmatales",
  "n01373440": "Mycoplasmataceae, family Mycoplasmataceae",
  "n01373675": "genus Mycoplasma",
  "n01373826": "mycoplasma",
  "n01374063": "pleuropneumonialike organism, PPLO",
  "n01374224": "Legionella pneumophilia, legionella",
  "n01374457": "nitrobacterium",
  "n01374703": "nitrate bacterium, nitric bacterium",
  "n01374846": "nitrite bacterium, nitrous bacterium",
  "n01374989": "Actinomycetales, order Actinomycetales",
  "n01375204": "actinomycete",
  "n01375460": "Actinomycetaceae, family Actinomycetaceae",
  "n01375617": "genus Actinomyces",
  "n01375760": "actinomyces",
  "n01375913": "Streptomycetaceae, family Streptomycetaceae",
  "n01376092": "genus Streptomyces",
  "n01376237": "streptomyces",
  "n01376437": "Streptomyces erythreus",
  "n01376543": "Streptomyces griseus",
  "n01376647": "potato scab bacteria, Streptomyces scabies",
  "n01376801": "Mycobacteriaceae, family Mycobacteriaceae",
  "n01376948": "genus Mycobacterium",
  "n01377092": "mycobacteria, mycobacterium",
  "n01377278": "tubercle bacillus, Mycobacterium tuberculosis",
  "n01377392": "penicillin-resistant bacteria",
  "n01377510": "pus-forming bacteria",
  "n01377602": "rod",
  "n01377694": "streptobacillus",
  "n01377804": "leprosy bacillus, Mycobacterium leprae",
  "n01377906": "order Myxobacteria, Myxobacterales, order Myxobacterales, Myxobacteriales, order Myxobacteriales",
  "n01378137": "Polyangiaceae, family Polyangiaceae, Myxobacteriaceae, family Myxobacteriaceae",
  "n01378346": "Polyangium, genus Polyangium",
  "n01378545": "myxobacteria, myxobacterium, myxobacter, gliding bacteria, slime bacteria",
  "n01378800": "Micrococcaceae, family Micrococcaceae",
  "n01379116": "Micrococcus, genus Micrococcus",
  "n01379252": "genus Staphylococcus",
  "n01379389": "staphylococcus, staphylococci, staph",
  "n01379636": "Lactobacillaceae, family Lactobacillaceae, Lactobacteriaceae, family Lactobacteriaceae",
  "n01379954": "genus Lactobacillus",
  "n01380118": "lactobacillus",
  "n01380298": "acidophilus, Lactobacillus acidophilus",
  "n01380489": "genus Diplococcus",
  "n01380610": "diplococcus",
  "n01380754": "pneumococcus, Diplococcus pneumoniae",
  "n01380902": "genus Streptococcus",
  "n01381044": "streptococcus, streptococci, strep",
  "n01381285": "Streptococcus anhemolyticus",
  "n01381399": "Spirochaetales, order Spirochaetales",
  "n01381604": "Spirochaetaceae, family Spirochaetaceae",
  "n01381829": "Spirochaeta, genus Spirochaeta",
  "n01382033": "spirochete, spirochaete",
  "n01382273": "Treponemataceae, family Treponemataceae",
  "n01382482": "genus Treponema",
  "n01382699": "treponema",
  "n01382839": "genus Borrelia",
  "n01383027": "borrelia",
  "n01383151": "Borrelia burgdorferi, Lime disease spirochete",
  "n01383332": "genus Leptospira",
  "n01383502": "leptospira",
  "n01383638": "plankton",
  "n01383896": "phytoplankton",
  "n01384084": "planktonic algae",
  "n01384164": "zooplankton",
  "n01384313": "nekton",
  "n01384491": "microbe, bug, germ",
  "n01384687": "parasite",
  "n01385017": "endoparasite, entoparasite, entozoan, entozoon, endozoan",
  "n01385330": "ectoparasite, ectozoan, ectozoon, epizoan, epizoon",
  "n01385527": "host",
  "n01385749": "intermediate host",
  "n01385878": "definitive host",
  "n01386007": "pathogen",
  "n01386182": "commensal",
  "n01386354": "myrmecophile",
  "n01386494": "Protoctista, kingdom Protoctista",
  "n01387065": "protoctist",
  "n01387208": "Protista, division Protista",
  "n01387451": "protist, protistan",
  "n01387617": "protoctist family",
  "n01388130": "protoctist genus",
  "n01388992": "Pyrrophyta, phylum Pyrrophyta",
  "n01389188": "Protozoa, phylum Protozoa",
  "n01389507": "protozoan, protozoon",
  "n01389875": "Sarcodina, class Sarcodina",
  "n01390123": "sarcodinian, sarcodine",
  "n01390287": "Actinopoda, subclass Actinopoda",
  "n01390466": "actinopod",
  "n01390627": "Heliozoa, order Heliozoa",
  "n01390763": "heliozoan",
  "n01390899": "Radiolaria, order Radiolaria",
  "n01391028": "radiolarian",
  "n01391174": "Rhizopoda, subclass Rhizopoda",
  "n01391391": "rhizopod, rhizopodan",
  "n01391569": "Amoebida, order Amoebida, Amoebina, order Amoebina",
  "n01391779": "genus Amoeba",
  "n01391933": "Endamoebidae, family Endamoebidae",
  "n01392122": "Endamoeba, genus Endamoeba",
  "n01392275": "endameba",
  "n01392380": "ameba, amoeba",
  "n01392692": "Endamoeba histolytica",
  "n01392843": "Foraminifera, order Foraminifera",
  "n01393030": "foram, foraminifer",
  "n01393237": "Globigerinidae, family Globigerinidae",
  "n01393364": "genus Globigerina",
  "n01393486": "globigerina",
  "n01393604": "Nummulitidae, family Nummulitidae",
  "n01393753": "nummulite",
  "n01393873": "Testacea, order Testacea",
  "n01394040": "testacean",
  "n01394193": "Arcellidae, family Arcellidae",
  "n01394366": "genus Arcella",
  "n01394492": "arcella",
  "n01394635": "genus Difflugia",
  "n01394771": "difflugia",
  "n01394901": "Ciliata, class Ciliata, Ciliophora, class Ciliophora",
  "n01395254": "ciliate, ciliated protozoan, ciliophoran",
  "n01395531": "Infusoria, subclass Infusoria",
  "n01395773": "infusorian",
  "n01395885": "genus Paramecium",
  "n01396048": "paramecium, paramecia",
  "n01396170": "genus Tetrahymena",
  "n01396322": "tetrahymena",
  "n01396458": "genus Stentor",
  "n01396617": "stentor",
  "n01396776": "genus Vorticella",
  "n01396951": "vorticella",
  "n01397114": "alga, algae",
  "n01397497": "seaweed",
  "n01397690": "arame",
  "n01397778": "wrack",
  "n01397871": "seagrass",
  "n01398064": "sea wrack, wrack",
  "n01398212": "chlorophyll, chlorophyl",
  "n01398481": "chlorophyll a",
  "n01398632": "chlorophyll b",
  "n01398803": "chlorophyll c, chlorofucin",
  "n01398941": "chlorophyll d",
  "n01399065": "bacteriochlorophyll",
  "n01399236": "phycobilin",
  "n01399366": "phycoerythrin",
  "n01399450": "phycocyanin",
  "n01399529": "Heterokontophyta, division Heterokontophyta",
  "n01399772": "Chrysophyta, division Chrysophyta",
  "n01400247": "golden algae",
  "n01400391": "yellow-green algae",
  "n01400575": "Chrysophyceae, class Chrysophyceae, Heterokontae, class Heterokontae",
  "n01400772": "Xanthophyceae, class Xanthophyceae",
  "n01400891": "Bacillariophyceae, class Bacillariophyceae, Diatomophyceae, class Diatomophyceae",
  "n01401106": "diatom",
  "n01401296": "Heterotrichales, order Heterotrichales",
  "n01401517": "Tribonemaceae, family Tribonemaceae",
  "n01401686": "Tribonema, genus Tribonema, genus Conferva",
  "n01401842": "conferva",
  "n01402039": "confervoid algae",
  "n01402169": "Phaeophyceae, class Phaeophyceae",
  "n01402381": "Phaeophyta, division Phaeophyta",
  "n01402600": "brown algae",
  "n01402831": "Laminariales, order Laminariales",
  "n01403052": "Laminariaceae, family Laminariaceae",
  "n01403284": "Laminaria, genus Laminaria",
  "n01403457": "kelp",
  "n01403595": "sea tangle, tang",
  "n01403713": "tang, sea tang",
  "n01403805": "Fucales, order Fucales",
  "n01403968": "Cyclosporeae, class Cyclosporeae",
  "n01404129": "Fucaceae, family Fucaceae",
  "n01404365": "fucoid, fucoid algae",
  "n01404495": "fucoid",
  "n01404628": "rockweed",
  "n01404813": "genus Fucus",
  "n01405007": "fucus",
  "n01405107": "serrated wrack, Fucus serratus, tang",
  "n01405250": "bladderwrack, black rockweed, bladder fucus, tang, Fucus vesiculosus",
  "n01405442": "Ascophyllum, genus Ascophyllum",
  "n01405616": "bladderwrack, Ascophyllum nodosum",
  "n01405737": "genus Sargassum",
  "n01405858": "gulfweed, sargassum, sargasso, Sargassum bacciferum",
  "n01406092": "Euglenophyta, division Euglenophyta",
  "n01406262": "Euglenophyceae, class Euglenophyceae",
  "n01406426": "Euglenaceae, family Euglenaceae",
  "n01406565": "genus Euglena",
  "n01406736": "euglena",
  "n01406904": "euglenoid, euglenophyte, euglenid",
  "n01407065": "Chlorophyta, division Chlorophyta",
  "n01407465": "Chlorophyceae, class Chlorophyceae",
  "n01407798": "green algae, chlorophyte",
  "n01408088": "Ulvophyceae, class Ulvophyceae",
  "n01408253": "Ulvales, order Ulvales",
  "n01408383": "Ulvaceae, family Ulvaceae, sea-lettuce family",
  "n01408547": "Ulva, genus Ulva",
  "n01408743": "sea lettuce, laver",
  "n01408880": "Volvocales, order Volvocales",
  "n01409065": "Volvocaceae, family Volvocaceae",
  "n01409244": "Volvox, genus Volvox",
  "n01409477": "Chlamydomonadaceae, family Chlamydomonadaceae",
  "n01409665": "Chlamydomonas, genus Chlamydomonas",
  "n01409940": "Zygnematales, order Zygnematales, Zygnemales, order Zygnemales",
  "n01410109": "Zygnemataceae, family Zygnemataceae",
  "n01410330": "Zygnema, genus Zygnema",
  "n01410457": "pond scum",
  "n01410568": "genus Spirogyra",
  "n01410689": "spirogyra",
  "n01410847": "Chlorococcales, order Chlorococcales",
  "n01411036": "Chlorococcum, genus Chlorococcum",
  "n01411240": "genus Chlorella",
  "n01411450": "chlorella",
  "n01411556": "Oedogoniales, order oedogoniales",
  "n01411727": "Oedogoniaceae, family Oedogoniaceae",
  "n01411871": "Oedogonium, genus Oedogonium",
  "n01412085": "Charophyceae, class Charophyceae",
  "n01412279": "Charales, order Charales",
  "n01412479": "Characeae, family Characeae",
  "n01412694": "stonewort",
  "n01412925": "Chara, genus Chara",
  "n01413071": "Nitella, genus Nitella",
  "n01413188": "Desmidiaceae, family Desmidiaceae",
  "n01413324": "Desmidium, genus Desmidium",
  "n01413457": "desmid",
  "n01413551": "Rhodophyta, division Rhodophyta",
  "n01413744": "Rhodophyceae, class Rhodophyceae",
  "n01413942": "red algae",
  "n01414216": "sea moss",
  "n01414359": "Gigartinaceae, family Gigartinaceae",
  "n01414502": "Chondrus, genus Chondrus",
  "n01414633": "Irish moss, carrageen, carageen, carragheen, Chondrus crispus",
  "n01414841": "Rhodymeniaceae, family Rhodymeniaceae",
  "n01414986": "Rhodymenia, genus Rhodymenia",
  "n01415139": "dulse, Rhodymenia palmata",
  "n01415256": "Bangiaceae, family Bangiaceae",
  "n01415393": "Porphyra, genus Porphyra",
  "n01415524": "red laver, laver",
  "n01415626": "eukaryote, eucaryote",
  "n01415920": "prokaryote, procaryote",
  "n01416213": "zooid",
  "n01416354": "Mastigophora, class Mastigophora, Flagellata, class Flagellata",
  "n01416585": "flagellate, flagellate protozoan, flagellated protozoan, mastigophoran, mastigophore",
  "n01417041": "Dinoflagellata, order Dinoflagellata, Cilioflagellata, order Cilioflagellata",
  "n01417361": "dinoflagellate",
  "n01417553": "genus Noctiluca",
  "n01417674": "noctiluca, Noctiluca miliaris",
  "n01417807": "Peridiniidae, family Peridiniidae",
  "n01417982": "Peridinium, genus Peridinium",
  "n01418114": "peridinian",
  "n01418237": "Zoomastigina, subclass Zoomastigina",
  "n01418498": "Leishmania, genus Leishmania",
  "n01418620": "zoomastigote, zooflagellate",
  "n01418794": "Hypermastigina, order Hypermastigina",
  "n01418947": "hypermastigote",
  "n01419082": "Polymastigina, order Polymastigina",
  "n01419332": "polymastigote",
  "n01419444": "genus Costia",
  "n01419573": "costia, Costia necatrix",
  "n01419740": "genus Giardia",
  "n01419888": "giardia",
  "n01420000": "Chilomastix, genus Chilomastix",
  "n01420164": "Hexamita, genus Hexamita",
  "n01420314": "genus Trichomonas",
  "n01420525": "trichomonad",
  "n01420655": "Phytomastigina, subclass Phytomastigina",
  "n01420860": "plantlike flagellate",
  "n01421012": "Cryptophyta, phylum Cryptophyta",
  "n01421164": "Cryptophyceae, class Cryptophyceae",
  "n01421333": "cryptomonad, cryptophyte",
  "n01421496": "Sporozoa, class Sporozoa",
  "n01421807": "sporozoan",
  "n01422185": "sporozoite",
  "n01422335": "trophozoite",
  "n01422450": "merozoite",
  "n01422594": "Telosporidia, subclass Telosporidia",
  "n01422835": "Coccidia, order Coccidia",
  "n01423001": "Eimeriidae, family Eimeriidae",
  "n01423160": "genus Eimeria",
  "n01423302": "coccidium, eimeria",
  "n01423464": "Gregarinida, order Gregarinida",
  "n01423617": "gregarine",
  "n01423757": "Haemosporidia, order Haemosporidia",
  "n01423971": "haemosporidian",
  "n01424165": "Plasmodiidae, family Plasmodiidae",
  "n01424282": "genus Plasmodium",
  "n01424420": "plasmodium, Plasmodium vivax, malaria parasite",
  "n01424607": "Haemoproteidae, family Haemoproteidae",
  "n01424782": "haemoproteid",
  "n01424938": "Haemoproteus, genus Haemoproteus",
  "n01425076": "genus Leucocytozoon, genus Leucocytozoan",
  "n01425223": "leucocytozoan, leucocytozoon",
  "n01425336": "Babesiidae, family Babesiidae",
  "n01425501": "genus Babesia, genus Piroplasma",
  "n01425634": "piroplasm",
  "n01425817": "Acnidosporidia, subclass Acnidosporidia",
  "n01425983": "Sarcosporidia, order Sarcosporidia",
  "n01426160": "Sarcocystis, genus Sarcocystis",
  "n01426314": "sarcosporidian, sarcocystidean, sarcocystieian",
  "n01426466": "Haplosporidia, order Haplosporidia",
  "n01426625": "haplosporidian",
  "n01426784": "Cnidosporidia, subclass Cnidosporidia",
  "n01427005": "Actinomyxidia, order Actinomyxidia",
  "n01427143": "actinomyxidian",
  "n01427241": "Mycrosporidia, order Mycrosporidia",
  "n01427399": "microsporidian",
  "n01427547": "Myxosporidia, order Myxosporidia",
  "n01427703": "myxosporidian",
  "n01427848": "pseudopod, pseudopodium",
  "n01428003": "plasmodium",
  "n01428155": "Malacopterygii, superorder Malacopterygii",
  "n01428580": "soft-finned fish, malacopterygian",
  "n01429172": "Ostariophysi, order Ostariophysi",
  "n01429349": "fish family",
  "n01432517": "fish genus",
  "n01437805": "Cypriniformes, order Cypriniformes",
  "n01438208": "cypriniform fish",
  "n01438461": "Cobitidae, family Cobitidae",
  "n01438581": "loach",
  "n01438720": "Cyprinidae, family Cyprinidae",
  "n01439121": "cyprinid, cyprinid fish",
  "n01439514": "carp",
  "n01439657": "Cyprinus, genus Cyprinus",
  "n01439808": "domestic carp, Cyprinus carpio",
  "n01440160": "leather carp",
  "n01440242": "mirror carp",
  "n01440344": "Abramis, genus Abramis",
  "n01440467": "European bream, Abramis brama",
  "n01440655": "Tinca, genus Tinca",
  "n01440764": "tench, Tinca tinca",
  "n01440949": "Leuciscus, genus Leuciscus",
  "n01441117": "dace, Leuciscus leuciscus",
  "n01441272": "chub, Leuciscus cephalus",
  "n01441425": "shiner",
  "n01441625": "Notropis, genus Notropis",
  "n01441742": "emerald shiner, Notropis atherinoides",
  "n01441910": "common shiner, silversides, Notropis cornutus",
  "n01442055": "Notemigonus, genus Notemigonus",
  "n01442166": "golden shiner, Notemigonus crysoleucas",
  "n01442335": "Rutilus, genus Rutilus",
  "n01442450": "roach, Rutilus rutilus",
  "n01442591": "Scardinius, genus Scardinius",
  "n01442710": "rudd, Scardinius erythrophthalmus",
  "n01442855": "Phoxinus, genus Phoxinus",
  "n01442972": "minnow, Phoxinus phoxinus",
  "n01443126": "Gobio, genus Gobio",
  "n01443243": "gudgeon, Gobio gobio",
  "n01443398": "Carassius, genus Carassius",
  "n01443537": "goldfish, Carassius auratus",
  "n01443736": "silverfish",
  "n01443831": "crucian carp, Carassius carassius, Carassius vulgaris",
  "n01443998": "Electrophoridae, family Electrophoridae",
  "n01444164": "Electrophorus, genus Electrophorus",
  "n01444339": "electric eel, Electrophorus electric",
  "n01444520": "Catostomidae, family Catostomidae",
  "n01444783": "catostomid",
  "n01444922": "sucker",
  "n01445173": "Catostomus, genus Catostomus",
  "n01445305": "Ictiobus, genus Ictiobus",
  "n01445429": "buffalo fish, buffalofish",
  "n01445593": "black buffalo, Ictiobus niger",
  "n01445718": "Hypentelium, genus Hypentelium",
  "n01445857": "hog sucker, hog molly, Hypentelium nigricans",
  "n01445998": "Maxostoma, genus Maxostoma",
  "n01446152": "redhorse, redhorse sucker",
  "n01446283": "Cyprinodontidae, family Cyprinodontidae",
  "n01446589": "cyprinodont",
  "n01446760": "killifish",
  "n01447001": "Fundulus, genus Fundulus",
  "n01447139": "mummichog, Fundulus heteroclitus",
  "n01447331": "striped killifish, mayfish, may fish, Fundulus majalis",
  "n01447551": "genus Rivulus",
  "n01447658": "rivulus",
  "n01447822": "Jordanella, genus Jordanella",
  "n01447946": "flagfish, American flagfish, Jordanella floridae",
  "n01448165": "Xyphophorus, genus Xyphophorus",
  "n01448291": "swordtail, helleri, topminnow, Xyphophorus helleri",
  "n01448496": "Lebistes, genus Lebistes",
  "n01448594": "guppy, rainbow fish, Lebistes reticulatus",
  "n01448767": "Poeciliidae, family Poeciliidae",
  "n01448951": "topminnow, poeciliid fish, poeciliid, live-bearer",
  "n01449252": "Gambusia, genus Gambusia",
  "n01449374": "mosquitofish, Gambusia affinis",
  "n01449586": "Platypoecilus, genus Platypoecilus",
  "n01449712": "platy, Platypoecilus maculatus",
  "n01449857": "Mollienesia, genus Mollienesia",
  "n01449980": "mollie, molly",
  "n01450081": "Berycomorphi, order Berycomorphi",
  "n01450281": "Holocentridae, family Holocentridae",
  "n01450453": "Holocentrus, genus Holocentrus",
  "n01450661": "squirrelfish",
  "n01450950": "reef squirrelfish, Holocentrus coruscus",
  "n01451115": "deepwater squirrelfish, Holocentrus bullisi",
  "n01451295": "Holocentrus ascensionis",
  "n01451426": "soldierfish, soldier-fish",
  "n01451524": "Anomalopidae, family Anomalopidae",
  "n01451726": "genus Anomalops",
  "n01451863": "anomalops, flashlight fish",
  "n01452051": "Krypterophaneron, genus Krypterophaneron",
  "n01452200": "Photoblepharon, genus Photoblepharon",
  "n01452345": "flashlight fish, Photoblepharon palpebratus",
  "n01452496": "Zeomorphi, order Zeomorphi",
  "n01452633": "Zeidae, family Zeidae",
  "n01452798": "dory",
  "n01452954": "Zeus, genus Zeus",
  "n01453087": "John Dory, Zeus faber",
  "n01453188": "Caproidae, family Caproidae",
  "n01453330": "Capros, genus Capros",
  "n01453475": "boarfish, Capros aper",
  "n01453591": "Antigonia, genus Antigonia",
  "n01453742": "boarfish",
  "n01453852": "Solenichthyes, order Solenichthyes",
  "n01454260": "Fistulariidae, family Fistulariidae",
  "n01454393": "Fistularia, genus Fistularia",
  "n01454545": "cornetfish",
  "n01454702": "Gasterosteidae, family Gasterosteidae",
  "n01454856": "stickleback, prickleback",
  "n01455141": "Gasterosteus, genus gasterosteus",
  "n01455317": "three-spined stickleback, Gasterosteus aculeatus",
  "n01455461": "ten-spined stickleback, Gasterosteus pungitius",
  "n01455592": "Syngnathidae, family Syngnathidae",
  "n01455778": "pipefish, needlefish",
  "n01455986": "Syngnathus, genus Syngnathus",
  "n01456137": "dwarf pipefish, Syngnathus hildebrandi",
  "n01456296": "Cosmocampus, genus Cosmocampus",
  "n01456454": "deepwater pipefish, Cosmocampus profundus",
  "n01456631": "Hippocampus, genus Hippocampus",
  "n01456756": "seahorse, sea horse",
  "n01456939": "Macrorhamphosidae, family Macrorhamphosidae",
  "n01457082": "snipefish, bellows fish",
  "n01457276": "Centriscidae, family Centriscidae",
  "n01457407": "shrimpfish, shrimp-fish",
  "n01457576": "Aulostomidae, family Aulostomidae",
  "n01457708": "Aulostomus, genus Aulostomus",
  "n01457852": "trumpetfish, Aulostomus maculatus",
  "n01458010": "cytostome",
  "n01458105": "cilium",
  "n01458302": "flagellum",
  "n01458509": "flame cell",
  "n01458616": "investment",
  "n01458746": "pellicle",
  "n01458842": "embryo, conceptus, fertilized egg",
  "n01459242": "blastocoel, blastocoele, blastocele, segmentation cavity, cleavage cavity",
  "n01459480": "blastoderm, germinal disc, blastodisc, germinal area",
  "n01459664": "blastomere",
  "n01459791": "fetus, foetus",
  "n01460108": "monster, teras",
  "n01460303": "abortus",
  "n01460457": "egg",
  "n01460813": "ovipositor",
  "n01460963": "chalaza",
  "n01461146": "nit",
  "n01461315": "spawn",
  "n01461445": "roe",
  "n01461532": "roe",
  "n01461646": "blastula, blastosphere",
  "n01462042": "blastocyst, blastodermic vessicle",
  "n01462209": "trophoblast",
  "n01462544": "gastrula",
  "n01462803": "morula",
  "n01462945": "archenteron",
  "n01463115": "blastopore",
  "n01463259": "layer",
  "n01463419": "embryonic tissue",
  "n01463519": "germ layer",
  "n01463739": "ectoderm, exoderm, ectoblast",
  "n01463933": "neural tube",
  "n01464077": "mesoderm, mesoblast",
  "n01464318": "chordamesoderm, chordomesoderm",
  "n01464441": "mesenchyme",
  "n01464573": "endoderm, entoderm, endoblast, entoblast, hypoblast",
  "n01464766": "silkworm seed",
  "n01464844": "yolk, vitellus",
  "n01465046": "yolk sac",
  "n01465243": "yolk sac, vitelline sac, umbilical vesicle, vesicula umbilicus",
  "n01465472": "fang",
  "n01465593": "fang",
  "n01465713": "tusk",
  "n01465994": "Chordata, phylum Chordata",
  "n01466257": "chordate",
  "n01466472": "notochord",
  "n01466701": "urochord",
  "n01466828": "chordate family",
  "n01466996": "chordate genus",
  "n01467180": "Cephalochordata, subphylum Cephalochordata",
  "n01467336": "cephalochordate",
  "n01467504": "Amphioxidae, family Amphioxidae, Branchiostomidae, family Branchiostomidae",
  "n01467675": "genus Amphioxus",
  "n01467804": "lancelet, amphioxus",
  "n01467986": "Urochordata, subphylum Urochordata, Urochorda, subphylum Urochorda, Tunicata, subphylum Tunicata",
  "n01468238": "tunicate, urochordate, urochord",
  "n01468532": "Ascidiaceae, class Ascidiaceae",
  "n01468712": "ascidian",
  "n01468913": "siphon, syphon",
  "n01469103": "sea squirt",
  "n01469222": "Thaliacea, class Thaliacea",
  "n01469425": "Salpidae, family Salpidae",
  "n01469586": "genus Salpa",
  "n01469723": "salp, salpa",
  "n01469886": "Doliolidae, family Doliolidae",
  "n01470018": "genus Doliolum",
  "n01470145": "doliolum",
  "n01470287": "Larvacea, class Larvacea",
  "n01470479": "larvacean",
  "n01470586": "genus Appendicularia",
  "n01470733": "appendicularia",
  "n01470895": "ascidian tadpole",
  "n01471070": "Vertebrata, subphylum Vertebrata, Craniata, subphylum Craniata",
  "n01471682": "vertebrate, craniate",
  "n01472303": "Amniota",
  "n01472502": "amniote",
  "n01472638": "amnion, amniotic sac, amnios",
  "n01472939": "chorion",
  "n01473188": "chorionic villus",
  "n01473393": "allantois",
  "n01473620": "chorioallantois, chorioallantoic membrane",
  "n01473806": "aquatic vertebrate",
  "n01473990": "Agnatha, superclass Agnatha",
  "n01474283": "jawless vertebrate, jawless fish, agnathan",
  "n01474641": "Ostracodermi, order Ostracodermi",
  "n01474864": "ostracoderm",
  "n01475025": "Heterostraci, suborder Heterostraci",
  "n01475232": "heterostracan",
  "n01475421": "Osteostraci, suborder Osteostraci, Cephalaspida, suborder Cephalaspida",
  "n01475648": "osteostracan, cephalaspid",
  "n01475794": "Anaspida, order Anaspida",
  "n01475940": "anaspid",
  "n01476135": "Conodonta, order Conodonta, Conodontophorida, order Conodontophorida",
  "n01476418": "conodont",
  "n01476696": "conodont",
  "n01476829": "Cyclostomata, order Cyclostomata",
  "n01477080": "cyclostome",
  "n01477184": "Petromyzoniformes, suborder Petromyzoniformes, Hyperoartia, suborder Hyperoartia",
  "n01477373": "Petromyzontidae, family Petromyzontidae",
  "n01477525": "lamprey, lamprey eel, lamper eel",
  "n01477745": "Petromyzon, genus Petromyzon",
  "n01477875": "sea lamprey, Petromyzon marinus",
  "n01478073": "Myxiniformes, suborder Myxiniformes, Hyperotreta, suborder Hyperotreta, Myxinoidei, Myxinoidea, suborder Myxinoidei",
  "n01478300": "Myxinidae, family Myxinidae",
  "n01478511": "hagfish, hag, slime eels",
  "n01478816": "Myxine, genus Myxine",
  "n01478969": "Myxine glutinosa",
  "n01479066": "genus Eptatretus",
  "n01479213": "eptatretus",
  "n01479329": "Myxinikela, genus Myxinikela",
  "n01479459": "Myxinikela siroka",
  "n01479643": "Gnathostomata, superclass Gnathostomata",
  "n01479820": "gnathostome",
  "n01479937": "Placodermi, class Placodermi",
  "n01480106": "placoderm",
  "n01480336": "Chondrichthyes, class Chondrichthyes",
  "n01480516": "cartilaginous fish, chondrichthian",
  "n01480715": "Holocephali, subclass Holocephali",
  "n01480880": "holocephalan, holocephalian",
  "n01481063": "Chimaeridae, family Chimaeridae",
  "n01481203": "genus Chimaera",
  "n01481331": "chimaera",
  "n01481498": "rabbitfish, Chimaera monstrosa",
  "n01481599": "Elasmobranchii, subclass Elasmobranchii, Selachii, subclass Selachii",
  "n01482071": "elasmobranch, selachian",
  "n01482330": "shark",
  "n01482754": "Hexanchidae, family Hexanchidae",
  "n01482887": "Hexanchus, genus Hexanchus",
  "n01483021": "cow shark, six-gilled shark, Hexanchus griseus",
  "n01483188": "Lamnidae, family Lamnidae",
  "n01483370": "Isuridae, family Isuridae",
  "n01483522": "mackerel shark",
  "n01483707": "Lamna, genus Lamna",
  "n01483830": "porbeagle, Lamna nasus",
  "n01483980": "Isurus, genus Isurus",
  "n01484097": "mako, mako shark",
  "n01484285": "shortfin mako, Isurus oxyrhincus",
  "n01484447": "longfin mako, Isurus paucus",
  "n01484562": "bonito shark, blue pointed, Isurus glaucus",
  "n01484717": "Carcharodon, genus Carcharodon",
  "n01484850": "great white shark, white shark, man-eater, man-eating shark, Carcharodon carcharias",
  "n01485073": "Cetorhinus, genus Cetorhinus",
  "n01485306": "Cetorhinidae, family Cetorhinidae",
  "n01485479": "basking shark, Cetorhinus maximus",
  "n01485673": "Alopiidae, family Alopiidae",
  "n01485801": "Alopius, genus Alopius",
  "n01486010": "thresher, thrasher, thresher shark, fox shark, Alopius vulpinus",
  "n01486241": "Orectolobidae, family Orectolobidae",
  "n01486411": "Orectolobus, genus Orectolobus",
  "n01486540": "carpet shark, Orectolobus barbatus",
  "n01486706": "Ginglymostoma, genus Ginglymostoma",
  "n01486838": "nurse shark, Ginglymostoma cirratum",
  "n01487077": "Carchariidae, family Carchariidae, Odontaspididae, family Odontaspididae",
  "n01487312": "Carcharias, genus Carcharias, Odontaspis, genus Odontaspis",
  "n01487506": "sand tiger, sand shark, Carcharias taurus, Odontaspis taurus",
  "n01487743": "Rhincodontidae, family Rhincodontidae",
  "n01487914": "Rhincodon, genus Rhincodon",
  "n01488038": "whale shark, Rhincodon typus",
  "n01488234": "Scyliorhinidae, family Scyliorhinidae",
  "n01488385": "cat shark",
  "n01488539": "Carcharhinidae, family Carcharhinidae",
  "n01488918": "requiem shark",
  "n01489275": "Carcharhinus, genus Carcharhinus",
  "n01489501": "bull shark, cub shark, Carcharhinus leucas",
  "n01489709": "sandbar shark, Carcharhinus plumbeus",
  "n01489920": "blacktip shark, sandbar shark, Carcharhinus limbatus",
  "n01490112": "whitetip shark, oceanic whitetip shark, white-tipped shark, Carcharinus longimanus",
  "n01490360": "dusky shark, Carcharhinus obscurus",
  "n01490546": "Negaprion, genus Negaprion",
  "n01490670": "lemon shark, Negaprion brevirostris",
  "n01490885": "Prionace, genus Prionace",
  "n01491006": "blue shark, great blue shark, Prionace glauca",
  "n01491235": "Galeocerdo, genus Galeocerdo",
  "n01491361": "tiger shark, Galeocerdo cuvieri",
  "n01491520": "Galeorhinus, genus Galeorhinus",
  "n01491661": "soupfin shark, soupfin, soup-fin, Galeorhinus zyopterus",
  "n01491874": "dogfish",
  "n01491991": "Triakidae, family Triakidae",
  "n01492212": "Mustelus, genus Mustelus",
  "n01492357": "smooth dogfish",
  "n01492569": "smoothhound, smoothhound shark, Mustelus mustelus",
  "n01492708": "American smooth dogfish, Mustelus canis",
  "n01492860": "Florida smoothhound, Mustelus norrisi",
  "n01493012": "Triaenodon, genus Triaenodon",
  "n01493146": "whitetip shark, reef whitetip shark, Triaenodon obseus",
  "n01493366": "Squalidae, family Squalidae",
  "n01493541": "spiny dogfish",
  "n01493687": "Squalus, genus Squalus",
  "n01493829": "Atlantic spiny dogfish, Squalus acanthias",
  "n01494041": "Pacific spiny dogfish, Squalus suckleyi",
  "n01494188": "Sphyrnidae, family Sphyrnidae",
  "n01494339": "Sphyrna, genus Sphyrna",
  "n01494475": "hammerhead, hammerhead shark",
  "n01494757": "smooth hammerhead, Sphyrna zygaena",
  "n01494882": "smalleye hammerhead, Sphyrna tudes",
  "n01495006": "shovelhead, bonnethead, bonnet shark, Sphyrna tiburo",
  "n01495192": "Squatinidae, family Squatinidae",
  "n01495340": "Squatina, genus Squatina",
  "n01495493": "angel shark, angelfish, Squatina squatina, monkfish",
  "n01495701": "ray",
  "n01496037": "Torpediniformes, order Torpediniformes",
  "n01496199": "Torpedinidae, family Torpedinidae",
  "n01496331": "electric ray, crampfish, numbfish, torpedo",
  "n01496617": "Rajiformes, order Rajiformes, Batoidei, order Batoidei",
  "n01496944": "Pristidae, family Pristidae",
  "n01497118": "sawfish",
  "n01497278": "Pristis, genus Pristis",
  "n01497413": "smalltooth sawfish, Pristis pectinatus",
  "n01497579": "Rhinobatidae, family Rhinobatidae",
  "n01497738": "guitarfish",
  "n01497878": "Dasyatidae, family Dasyatidae",
  "n01498041": "stingray",
  "n01498268": "Dasyatis, genus Dasyatis",
  "n01498406": "roughtail stingray, Dasyatis centroura",
  "n01498577": "Gymnura, genus Gymnura",
  "n01498699": "butterfly ray",
  "n01498822": "Myliobatidae, family Myliobatidae",
  "n01498989": "eagle ray",
  "n01499261": "Aetobatus, genus Aetobatus",
  "n01499396": "spotted eagle ray, spotted ray, Aetobatus narinari",
  "n01499595": "Rhinoptera, genus Rhinoptera",
  "n01499732": "cownose ray, cow-nosed ray, Rhinoptera bonasus",
  "n01499898": "Mobulidae, family Mobulidae",
  "n01500091": "manta, manta ray, devilfish",
  "n01500360": "genus Manta",
  "n01500476": "Atlantic manta, Manta birostris",
  "n01500721": "Mobula, genus Mobula",
  "n01500854": "devil ray, Mobula hypostoma",
  "n01500995": "Rajidae, family Rajidae",
  "n01501160": "skate",
  "n01501450": "Raja, genus Raja",
  "n01501641": "grey skate, gray skate, Raja batis",
  "n01501777": "little skate, Raja erinacea",
  "n01501948": "thorny skate, Raja radiata",
  "n01502101": "barndoor skate, Raja laevis",
  "n01502262": "Aves, class Aves",
  "n01503061": "bird",
  "n01503976": "dickeybird, dickey-bird, dickybird, dicky-bird",
  "n01504179": "fledgling, fledgeling",
  "n01504344": "nestling, baby bird",
  "n01504437": "bird family",
  "n01507175": "bird genus",
  "n01514431": "breast",
  "n01514549": "throat",
  "n01514668": "cock",
  "n01514752": "gamecock, fighting cock",
  "n01514859": "hen",
  "n01514926": "nester",
  "n01515078": "night bird",
  "n01515217": "night raven",
  "n01515303": "bird of passage",
  "n01515398": "genus Protoavis",
  "n01515583": "protoavis",
  "n01515811": "Archaeornithes, subclass Archaeornithes",
  "n01516064": "genus Archaeopteryx, genus Archeopteryx",
  "n01516212": "archaeopteryx, archeopteryx, Archaeopteryx lithographica",
  "n01516487": "genus Sinornis",
  "n01516609": "Sinornis",
  "n01516878": "genus Ibero-mesornis",
  "n01517036": "Ibero-mesornis",
  "n01517265": "genus Archaeornis",
  "n01517389": "archaeornis",
  "n01517565": "ratite, ratite bird, flightless bird",
  "n01517966": "carinate, carinate bird, flying bird",
  "n01518170": "Ratitae, superorder Ratitae",
  "n01518347": "Struthioniformes, order Struthioniformes",
  "n01518564": "Struthionidae, family Struthionidae",
  "n01518718": "Struthio, genus Struthio",
  "n01518878": "ostrich, Struthio camelus",
  "n01519046": "Casuariiformes, order Casuariiformes",
  "n01519228": "Casuaridae, family Casuaridae",
  "n01519401": "Casuarius, genus Casuarius",
  "n01519563": "cassowary",
  "n01519719": "Dromaius, genus Dromaius",
  "n01519873": "emu, Dromaius novaehollandiae, Emu novaehollandiae",
  "n01520058": "Apterygiformes, order Apterygiformes",
  "n01520284": "Apterygidae, family Apterygidae",
  "n01520442": "genus Apteryx",
  "n01520576": "kiwi, apteryx",
  "n01520789": "Rheiformes, order Rheiformes",
  "n01521014": "Rheidae, family Rheidae",
  "n01521197": "genus Rhea",
  "n01521399": "rhea, Rhea americana",
  "n01521602": "Pterocnemia, genus Pterocnemia",
  "n01521756": "rhea, nandu, Pterocnemia pennata",
  "n01521980": "Aepyorniformes, order Aepyorniformes",
  "n01522147": "Aepyornidae, family Aepyornidae",
  "n01522305": "genus Aepyornis",
  "n01522450": "elephant bird, aepyornis",
  "n01522594": "Dinornithiformes, order Dinornithiformes",
  "n01522789": "Dinornithidae, family Dinornithidae",
  "n01522952": "Dinornis, genus Dinornis",
  "n01523105": "moa",
  "n01523248": "giant moa, Dinornis giganteus",
  "n01523379": "genus Anomalopteryx",
  "n01523493": "anomalopteryx, Anomalopteryx oweni",
  "n01523656": "Insessores, order Insessores, perching bird, percher",
  "n01523908": "Passeriformes, order Passeriformes",
  "n01524359": "passerine, passeriform bird",
  "n01524761": "nonpasserine bird",
  "n01524885": "Oscines, suborder Oscines, Passeres, suborder Passeres",
  "n01525720": "oscine, oscine bird",
  "n01526521": "songbird, songster",
  "n01526635": "Meliphagidae, family Meliphagidae",
  "n01526766": "honey eater, honeysucker",
  "n01526925": "Prunellidae, family Prunellidae",
  "n01527055": "Prunella, genus Prunella",
  "n01527194": "accentor",
  "n01527347": "hedge sparrow, sparrow, dunnock, Prunella modularis",
  "n01527480": "Alaudidae, family Alaudidae",
  "n01527617": "lark",
  "n01527774": "Alauda, genus Alauda",
  "n01527917": "skylark, Alauda arvensis",
  "n01528087": "Motacillidae, family Motacillidae",
  "n01528244": "Motacilla, genus Motacilla",
  "n01528396": "wagtail",
  "n01528542": "Anthus, genus Anthus",
  "n01528654": "pipit, titlark, lark",
  "n01528845": "meadow pipit, Anthus pratensis",
  "n01529036": "Fringillidae, family Fringillidae",
  "n01529672": "finch",
  "n01530256": "Fringilla, genus Fringilla",
  "n01530439": "chaffinch, Fringilla coelebs",
  "n01530575": "brambling, Fringilla montifringilla",
  "n01530691": "Carduelinae, subfamily Carduelinae",
  "n01530846": "Carduelis, genus Carduelis",
  "n01531178": "goldfinch, Carduelis carduelis",
  "n01531344": "linnet, lintwhite, Carduelis cannabina",
  "n01531512": "siskin, Carduelis spinus",
  "n01531639": "red siskin, Carduelis cucullata",
  "n01531811": "redpoll, Carduelis flammea",
  "n01531971": "redpoll, Carduelis hornemanni",
  "n01532107": "Spinus, genus Spinus",
  "n01532325": "New World goldfinch, goldfinch, yellowbird, Spinus tristis",
  "n01532511": "pine siskin, pine finch, Spinus pinus",
  "n01532664": "Carpodacus, genus Carpodacus",
  "n01532829": "house finch, linnet, Carpodacus mexicanus",
  "n01533000": "purple finch, Carpodacus purpureus",
  "n01533169": "Serinus, genus Serinus",
  "n01533339": "canary, canary bird",
  "n01533481": "common canary, Serinus canaria",
  "n01533651": "serin",
  "n01533780": "Loxia, genus Loxia",
  "n01533893": "crossbill, Loxia curvirostra",
  "n01534034": "Pyrrhula, genus Pyrrhula",
  "n01534155": "bullfinch, Pyrrhula pyrrhula",
  "n01534321": "genus Junco",
  "n01534433": "junco, snowbird",
  "n01534582": "dark-eyed junco, slate-colored junco, Junco hyemalis",
  "n01534762": "New World sparrow",
  "n01535005": "Pooecetes, genus Pooecetes",
  "n01535140": "vesper sparrow, grass finch, Pooecetes gramineus",
  "n01535310": "Zonotrichia, genus Zonotrichia",
  "n01535469": "white-throated sparrow, whitethroat, Zonotrichia albicollis",
  "n01535690": "white-crowned sparrow, Zonotrichia leucophrys",
  "n01535842": "Spizella, genus Spizella",
  "n01536035": "chipping sparrow, Spizella passerina",
  "n01536186": "field sparrow, Spizella pusilla",
  "n01536334": "tree sparrow, Spizella arborea",
  "n01536474": "Melospiza, genus Melospiza",
  "n01536644": "song sparrow, Melospiza melodia",
  "n01536780": "swamp sparrow, Melospiza georgiana",
  "n01536916": "Emberizidae, subfamily Emberizidae, subfamily Emberizinae",
  "n01537134": "bunting",
  "n01537360": "Passerina, genus Passerina",
  "n01537544": "indigo bunting, indigo finch, indigo bird, Passerina cyanea",
  "n01537710": "Emberiza, genus Emberiza",
  "n01537895": "ortolan, ortolan bunting, Emberiza hortulana",
  "n01538059": "reed bunting, Emberiza schoeniclus",
  "n01538200": "yellowhammer, yellow bunting, Emberiza citrinella",
  "n01538362": "yellow-breasted bunting, Emberiza aureola",
  "n01538498": "Plectrophenax, genus Plectrophenax",
  "n01538630": "snow bunting, snowbird, snowflake, Plectrophenax nivalis",
  "n01538775": "Coerebidae, family Coerebidae, Dacninae, family Dacninae",
  "n01538955": "honeycreeper",
  "n01539136": "Coereba, genus Coereba",
  "n01539272": "banana quit",
  "n01539377": "Passeridae, family Passeridae",
  "n01539573": "sparrow, true sparrow",
  "n01539772": "Passer, genus Passer",
  "n01539925": "English sparrow, house sparrow, Passer domesticus",
  "n01540090": "tree sparrow, Passer montanus",
  "n01540233": "grosbeak, grossbeak",
  "n01540432": "Hesperiphona, genus Hesperiphona",
  "n01540566": "evening grosbeak, Hesperiphona vespertina",
  "n01540697": "Coccothraustes, genus Coccothraustes",
  "n01540832": "hawfinch, Coccothraustes coccothraustes",
  "n01540969": "Pinicola, genus Pinicola",
  "n01541102": "pine grosbeak, Pinicola enucleator",
  "n01541261": "Richmondena, genus Richmondena",
  "n01541386": "cardinal, cardinal grosbeak, Richmondena Cardinalis, Cardinalis cardinalis, redbird",
  "n01541618": "genus Pyrrhuloxia",
  "n01541760": "pyrrhuloxia, Pyrrhuloxia sinuata",
  "n01541922": "towhee",
  "n01542055": "Pipilo, genus Pipilo",
  "n01542168": "chewink, cheewink, Pipilo erythrophthalmus",
  "n01542316": "Chlorura, genus Chlorura",
  "n01542433": "green-tailed towhee, Chlorura chlorura",
  "n01542567": "Ploceidae, family Ploceidae",
  "n01542786": "weaver, weaverbird, weaver finch",
  "n01543059": "Ploceus, genus Ploceus",
  "n01543175": "baya, Ploceus philippinus",
  "n01543272": "Vidua, genus Vidua",
  "n01543383": "whydah, whidah, widow bird",
  "n01543508": "Padda, genus Padda",
  "n01543632": "Java sparrow, Java finch, ricebird, Padda oryzivora",
  "n01543817": "Estrilda, genus Estrilda",
  "n01543936": "avadavat, amadavat",
  "n01544067": "Poephila, genus Poephila",
  "n01544208": "grassfinch, grass finch",
  "n01544389": "zebra finch, Poephila castanotis",
  "n01544544": "Drepanididae, family Drepanididae",
  "n01544704": "honeycreeper, Hawaiian honeycreeper",
  "n01544877": "Drepanis, genus Drepanis",
  "n01545010": "mamo",
  "n01545149": "Menurae, suborder Menurae",
  "n01545303": "Menuridae, family Menuridae",
  "n01545425": "Menura, genus Menura",
  "n01545574": "lyrebird",
  "n01545752": "Atrichornithidae, family Atrichornithidae",
  "n01545889": "Atrichornis, genus Atrichornis",
  "n01546039": "scrubbird, scrub-bird, scrub bird",
  "n01546223": "Eurylaimi, suborder Eurylaimi",
  "n01546348": "Eurylaimidae, family Eurylaimidae",
  "n01546506": "broadbill",
  "n01546660": "Tyranni, suborder Tyranni",
  "n01546921": "tyrannid",
  "n01547143": "Clamatores, suborder Clamatores",
  "n01547459": "Tyrannidae, superfamily Tyrannidae",
  "n01547832": "New World flycatcher, flycatcher, tyrant flycatcher, tyrant bird",
  "n01548143": "Tyrannus, genus Tyrannus",
  "n01548301": "kingbird, Tyrannus tyrannus",
  "n01548492": "Arkansas kingbird, western kingbird",
  "n01548694": "Cassin's kingbird, Tyrannus vociferans",
  "n01548865": "eastern kingbird",
  "n01549053": "grey kingbird, gray kingbird, petchary, Tyrannus domenicensis domenicensis",
  "n01549314": "Contopus, genus Contopus",
  "n01549430": "pewee, peewee, peewit, pewit, wood pewee, Contopus virens",
  "n01549641": "western wood pewee, Contopus sordidulus",
  "n01549769": "Sayornis, genus Sayornis",
  "n01549886": "phoebe, phoebe bird, Sayornis phoebe",
  "n01550033": "Pyrocephalus, genus Pyrocephalus",
  "n01550172": "vermillion flycatcher, firebird, Pyrocephalus rubinus mexicanus",
  "n01550429": "Cotingidae, family Cotingidae",
  "n01550625": "genus Cotinga",
  "n01550761": "cotinga, chatterer",
  "n01550953": "Rupicola, genus Rupicola",
  "n01551080": "cock of the rock, Rupicola rupicola",
  "n01551300": "cock of the rock, Rupicola peruviana",
  "n01551430": "Pipridae, family Pipridae",
  "n01551549": "Pipra, genus Pipra",
  "n01551711": "manakin",
  "n01551915": "Procnias, genus Procnias",
  "n01552034": "bellbird",
  "n01552192": "Cephalopterus, genus Cephalopterus",
  "n01552333": "umbrella bird, Cephalopterus ornatus",
  "n01552523": "Furnariidae, family Furnariidae",
  "n01552654": "Furnarius, genus Furnarius",
  "n01552813": "ovenbird",
  "n01552956": "Formicariidae, family Formicariidae",
  "n01553142": "antbird, ant bird",
  "n01553380": "Formicarius, genus Formicarius",
  "n01553527": "ant thrush",
  "n01553620": "Thamnophilus, genus Thamnophilus",
  "n01553762": "ant shrike",
  "n01553879": "Hylophylax, genus Hylophylax",
  "n01554017": "spotted antbird, Hylophylax naevioides",
  "n01554139": "Dendrocolaptidae, family Dendrocolaptidae",
  "n01554311": "Dendrocolaptes, genus Dendrocolaptes",
  "n01554448": "woodhewer, woodcreeper, wood-creeper, tree creeper",
  "n01554708": "Pittidae, family Pittidae",
  "n01554825": "genus Pitta",
  "n01555004": "pitta",
  "n01555172": "Muscivora, genus Muscivora",
  "n01555305": "scissortail, scissortailed flycatcher, Muscivora-forficata",
  "n01555586": "Muscicapidae, family Muscicapidae",
  "n01555809": "Old World flycatcher, true flycatcher, flycatcher",
  "n01556040": "Muscicapa, genus Muscicapa",
  "n01556182": "spotted flycatcher, Muscicapa striata, Muscicapa grisola",
  "n01556368": "Pachycephala, genus Pachycephala",
  "n01556514": "thickhead, whistler",
  "n01556671": "Turdidae, family Turdidae",
  "n01557028": "Turdinae, subfamily Turdinae",
  "n01557185": "thrush",
  "n01557697": "Turdus, genus Turdus",
  "n01557962": "missel thrush, mistle thrush, mistletoe thrush, Turdus viscivorus",
  "n01558149": "song thrush, mavis, throstle, Turdus philomelos",
  "n01558307": "fieldfare, snowbird, Turdus pilaris",
  "n01558461": "redwing, Turdus iliacus",
  "n01558594": "blackbird, merl, merle, ouzel, ousel, European blackbird, Turdus merula",
  "n01558765": "ring ouzel, ring blackbird, ring thrush, Turdus torquatus",
  "n01558993": "robin, American robin, Turdus migratorius",
  "n01559160": "clay-colored robin, Turdus greyi",
  "n01559294": "Hylocichla, genus Hylocichla",
  "n01559477": "hermit thrush, Hylocichla guttata",
  "n01559639": "veery, Wilson's thrush, Hylocichla fuscescens",
  "n01559804": "wood thrush, Hylocichla mustelina",
  "n01559964": "Luscinia, genus Luscinia",
  "n01560105": "nightingale, Luscinia megarhynchos",
  "n01560280": "thrush nightingale, Luscinia luscinia",
  "n01560419": "bulbul",
  "n01560511": "Saxicola, genus Saxicola",
  "n01560636": "Old World chat, chat",
  "n01560793": "stonechat, Saxicola torquata",
  "n01560935": "whinchat, Saxicola rubetra",
  "n01561059": "Myadestes, genus Myadestes",
  "n01561181": "solitaire",
  "n01561318": "Phoenicurus, genus Phoenicurus",
  "n01561452": "redstart, redtail",
  "n01561613": "Oenanthe, genus Oenanthe",
  "n01561732": "wheatear",
  "n01561884": "Sialia, genus Sialia",
  "n01562014": "bluebird",
  "n01562116": "Erithacus, genus Erithacus",
  "n01562265": "robin, redbreast, robin redbreast, Old World robin, Erithacus rubecola",
  "n01562451": "bluethroat, Erithacus svecicus",
  "n01562584": "Sylviidae, family Sylviidae",
  "n01562978": "Sylviinae, subfamily Sylviinae",
  "n01563128": "warbler",
  "n01563313": "Polioptila, genus Polioptila",
  "n01563449": "gnatcatcher",
  "n01563579": "Regulus, genus Regulus",
  "n01563746": "kinglet",
  "n01563945": "goldcrest, golden-crested kinglet, Regulus regulus",
  "n01564101": "gold-crowned kinglet, Regulus satrata",
  "n01564217": "ruby-crowned kinglet, ruby-crowned wren, Regulus calendula",
  "n01564394": "Old World warbler, true warbler",
  "n01564630": "Silvia, genus Silvia",
  "n01564773": "blackcap, Silvia atricapilla",
  "n01564914": "greater whitethroat, whitethroat, Sylvia communis",
  "n01565078": "lesser whitethroat, whitethroat, Sylvia curruca",
  "n01565238": "Phylloscopus, genus Phylloscopus",
  "n01565345": "wood warbler, Phylloscopus sibilatrix",
  "n01565480": "Acrocephalus, genus Acrocephalus",
  "n01565599": "sedge warbler, sedge bird, sedge wren, reedbird, Acrocephalus schoenobaenus",
  "n01565804": "Prinia, genus Prinia",
  "n01565930": "wren warbler",
  "n01566082": "Orthotomus, genus Orthotomus",
  "n01566207": "tailorbird, Orthotomus sutorius",
  "n01566386": "Timaliidae, family Timaliidae",
  "n01566509": "Timalia, genus Timalia",
  "n01566645": "babbler, cackler",
  "n01566888": "Parulidae, family Parulidae",
  "n01567133": "New World warbler, wood warbler",
  "n01567530": "Parula, genus Parula",
  "n01567678": "parula warbler, northern parula, Parula americana",
  "n01567879": "Wilson's warbler, Wilson's blackcap, Wilsonia pusilla",
  "n01568019": "Setophaga, genus Setophaga",
  "n01568132": "flycatching warbler",
  "n01568294": "American redstart, redstart, Setophaga ruticilla",
  "n01568493": "Dendroica, genus Dendroica",
  "n01568720": "Cape May warbler, Dendroica tigrina",
  "n01568892": "yellow warbler, golden warbler, yellowbird, Dendroica petechia",
  "n01569060": "Blackburn, Blackburnian warbler, Dendroica fusca",
  "n01569262": "Audubon's warbler, Audubon warbler, Dendroica auduboni",
  "n01569423": "myrtle warbler, myrtle bird, Dendroica coronata",
  "n01569566": "blackpoll, Dendroica striate",
  "n01569713": "Icteria, genus Icteria",
  "n01569836": "New World chat, chat",
  "n01569971": "yellow-breasted chat, Icteria virens",
  "n01570112": "Seiurus, genus Seiurus",
  "n01570267": "ovenbird, Seiurus aurocapillus",
  "n01570421": "water thrush",
  "n01570549": "Geothlypis, genus Geothlypis",
  "n01570676": "yellowthroat",
  "n01570839": "common yellowthroat, Maryland yellowthroat, Geothlypis trichas",
  "n01570969": "Paradisaeidae, family Paradisaeidae",
  "n01571126": "bird of paradise",
  "n01571297": "Ptloris, genus Ptloris",
  "n01571410": "riflebird, Ptloris paradisea",
  "n01571578": "Icteridae, family Icteridae",
  "n01571904": "New World oriole, American oriole, oriole",
  "n01572174": "Icterus, genus Icterus",
  "n01572328": "northern oriole, Icterus galbula",
  "n01572489": "Baltimore oriole, Baltimore bird, hangbird, firebird, Icterus galbula galbula",
  "n01572654": "Bullock's oriole, Icterus galbula bullockii",
  "n01572782": "orchard oriole, Icterus spurius",
  "n01572910": "Sturnella, genus Sturnella",
  "n01573074": "meadowlark, lark",
  "n01573240": "eastern meadowlark, Sturnella magna",
  "n01573360": "western meadowlark, Sturnella neglecta",
  "n01573483": "Cacicus, genus Cacicus",
  "n01573627": "cacique, cazique",
  "n01573775": "Dolichonyx, genus Dolichonyx",
  "n01573898": "bobolink, ricebird, reedbird, Dolichonyx oryzivorus",
  "n01574045": "New World blackbird, blackbird",
  "n01574270": "Quiscalus, genus Quiscalus",
  "n01574390": "grackle, crow blackbird",
  "n01574560": "purple grackle, Quiscalus quiscula",
  "n01574671": "Euphagus, genus Euphagus",
  "n01574801": "rusty blackbird, rusty grackle, Euphagus carilonus",
  "n01574997": "Molothrus, genus Molothrus",
  "n01575117": "cowbird",
  "n01575270": "Agelaius, genus Agelaius",
  "n01575401": "red-winged blackbird, redwing, Agelaius phoeniceus",
  "n01575577": "Oriolidae, family Oriolidae",
  "n01575745": "Old World oriole, oriole",
  "n01575941": "Oriolus, genus Oriolus",
  "n01576076": "golden oriole, Oriolus oriolus",
  "n01576212": "Sphecotheres, genus Sphecotheres",
  "n01576358": "fig-bird",
  "n01576506": "Sturnidae, family Sturnidae",
  "n01576695": "starling",
  "n01576863": "Sturnus, genus Sturnus",
  "n01577035": "common starling, Sturnus vulgaris",
  "n01577265": "Pastor, subgenus Pastor",
  "n01577458": "rose-colored starling, rose-colored pastor, Pastor sturnus, Pastor roseus",
  "n01577659": "myna, mynah, mina, minah, myna bird, mynah bird",
  "n01577818": "Acridotheres, genus Acridotheres",
  "n01577941": "crested myna, Acridotheres tristis",
  "n01578086": "Gracula, genus Gracula",
  "n01578180": "hill myna, Indian grackle, grackle, Gracula religiosa",
  "n01578341": "Corvidae, family Corvidae",
  "n01578575": "corvine bird",
  "n01578821": "Corvus, genus Corvus",
  "n01579028": "crow",
  "n01579149": "American crow, Corvus brachyrhyncos",
  "n01579260": "raven, Corvus corax",
  "n01579410": "rook, Corvus frugilegus",
  "n01579578": "jackdaw, daw, Corvus monedula",
  "n01579729": "chough",
  "n01579868": "Garrulinae, subfamily Garrulinae",
  "n01580077": "jay",
  "n01580225": "Garrulus, genus Garrulus",
  "n01580379": "Old World jay",
  "n01580490": "common European jay, Garullus garullus",
  "n01580644": "Cyanocitta, genus Cyanocitta",
  "n01580772": "New World jay",
  "n01580870": "blue jay, jaybird, Cyanocitta cristata",
  "n01581041": "Perisoreus, genus Perisoreus",
  "n01581166": "Canada jay, grey jay, gray jay, camp robber, whisker jack, Perisoreus canadensis",
  "n01581434": "Rocky Mountain jay, Perisoreus canadensis capitalis",
  "n01581607": "Nucifraga, genus Nucifraga",
  "n01581730": "nutcracker",
  "n01581874": "common nutcracker, Nucifraga caryocatactes",
  "n01581984": "Clark's nutcracker, Nucifraga columbiana",
  "n01582111": "Pica, genus Pica",
  "n01582220": "magpie",
  "n01582398": "European magpie, Pica pica",
  "n01582498": "American magpie, Pica pica hudsonia",
  "n01582625": "Cracticidae, family Cracticidae",
  "n01582856": "Australian magpie",
  "n01583043": "Cracticus, genus Cracticus",
  "n01583209": "butcherbird",
  "n01583373": "Strepera, genus Strepera",
  "n01583495": "currawong, bell magpie",
  "n01583636": "Gymnorhina, genus Gymnorhina",
  "n01583828": "piping crow, piping crow-shrike, Gymnorhina tibicen",
  "n01584004": "Troglodytidae, family Troglodytidae",
  "n01584225": "wren, jenny wren",
  "n01584529": "Troglodytes, genus Troglodytes",
  "n01584695": "winter wren, Troglodytes troglodytes",
  "n01584853": "house wren, Troglodytes aedon",
  "n01584994": "Cistothorus, genus Cistothorus",
  "n01585121": "marsh wren",
  "n01585287": "long-billed marsh wren, Cistothorus palustris",
  "n01585422": "sedge wren, short-billed marsh wren, Cistothorus platensis",
  "n01585577": "Salpinctes, genus Salpinctes",
  "n01585715": "rock wren, Salpinctes obsoletus",
  "n01585890": "Thryothorus, genus Thryothorus",
  "n01586020": "Carolina wren, Thryothorus ludovicianus",
  "n01586170": "Campylorhynchus, genus Campylorhynchus, Heleodytes, genus Heleodytes",
  "n01586374": "cactus wren",
  "n01586541": "Mimidae, family Mimidae",
  "n01586791": "Mimus, genus Mimus",
  "n01586941": "mockingbird, mocker, Mimus polyglotktos",
  "n01587148": "Melanotis, genus Melanotis",
  "n01587278": "blue mockingbird, Melanotis caerulescens",
  "n01587406": "Dumetella, genus Dumetella",
  "n01587526": "catbird, grey catbird, gray catbird, Dumetella carolinensis",
  "n01587713": "Toxostoma, genus Toxostoma",
  "n01587834": "thrasher, mocking thrush",
  "n01588002": "brown thrasher, brown thrush, Toxostoma rufums",
  "n01588172": "Xenicidae, family Xenicidae, Acanthisittidae, family Acanthisittidae",
  "n01588431": "New Zealand wren",
  "n01588589": "Xenicus, genus Xenicus",
  "n01588725": "rock wren, Xenicus gilviventris",
  "n01588858": "Acanthisitta, genus Acanthisitta",
  "n01588996": "rifleman bird, Acanthisitta chloris",
  "n01589125": "Certhiidae, family Certhiidae",
  "n01589286": "creeper, tree creeper",
  "n01589582": "Certhia, genus Certhia",
  "n01589718": "brown creeper, American creeper, Certhia americana",
  "n01589893": "European creeper, Certhia familiaris",
  "n01590042": "Tichodroma, genus Tichodroma",
  "n01590220": "wall creeper, tichodrome, Tichodroma muriaria",
  "n01590443": "Sittidae, family Sittidae",
  "n01590583": "nuthatch, nutcracker",
  "n01590837": "Sitta, genus Sitta",
  "n01591005": "European nuthatch, Sitta europaea",
  "n01591123": "red-breasted nuthatch, Sitta canadensis",
  "n01591301": "white-breasted nuthatch, Sitta carolinensis",
  "n01591490": "Paridae, family Paridae",
  "n01591697": "titmouse, tit",
  "n01591910": "Parus, genus Parus",
  "n01592084": "chickadee",
  "n01592257": "black-capped chickadee, blackcap, Parus atricapillus",
  "n01592387": "tufted titmouse, Parus bicolor",
  "n01592540": "Carolina chickadee, Parus carolinensis",
  "n01592694": "blue tit, tomtit, Parus caeruleus",
  "n01592892": "Psaltriparus, genus Psaltriparus",
  "n01593028": "bushtit, bush tit",
  "n01593156": "Chamaea, genus Chamaea",
  "n01593282": "wren-tit, Chamaea fasciata",
  "n01593423": "Auriparus, genus Auriparus",
  "n01593553": "verdin, Auriparus flaviceps",
  "n01593705": "Irenidae, family Irenidae",
  "n01593857": "Irena, genus Irena",
  "n01594004": "fairy bluebird, bluebird",
  "n01594157": "Hirundinidae, family Hirundinidae",
  "n01594372": "swallow",
  "n01594611": "Hirundo, genus Hirundo",
  "n01594787": "barn swallow, chimney swallow, Hirundo rustica",
  "n01594968": "cliff swallow, Hirundo pyrrhonota",
  "n01595168": "tree swallow, tree martin, Hirundo nigricans",
  "n01595330": "Iridoprocne, genus Iridoprocne",
  "n01595450": "white-bellied swallow, tree swallow, Iridoprocne bicolor",
  "n01595624": "martin",
  "n01595841": "Delichon, genus Delichon",
  "n01595974": "house martin, Delichon urbica",
  "n01596142": "Riparia, genus Riparia",
  "n01596273": "bank martin, bank swallow, sand martin, Riparia riparia",
  "n01596479": "Progne, genus Progne",
  "n01596608": "purple martin, Progne subis",
  "n01596761": "Artamidae, family Artamidae",
  "n01596887": "Artamus, genus Artamus",
  "n01597022": "wood swallow, swallow shrike",
  "n01597194": "Thraupidae, family Thraupidae",
  "n01597336": "tanager",
  "n01597551": "Piranga, genus Piranga",
  "n01597737": "scarlet tanager, Piranga olivacea, redbird, firebird",
  "n01597906": "western tanager, Piranga ludoviciana",
  "n01598074": "summer tanager, summer redbird, Piranga rubra",
  "n01598271": "hepatic tanager, Piranga flava hepatica",
  "n01598432": "Laniidae, family Laniidae",
  "n01598588": "shrike",
  "n01598820": "Lanius, genus Lanius",
  "n01598988": "butcherbird",
  "n01599159": "European shrike, Lanius excubitor",
  "n01599269": "northern shrike, Lanius borealis",
  "n01599388": "white-rumped shrike, Lanius ludovicianus excubitorides",
  "n01599556": "loggerhead shrike, Lanius lucovicianus",
  "n01599741": "migrant shrike, Lanius ludovicianus migrans",
  "n01599919": "Malaconotinae, subfamily Malaconotinae",
  "n01600085": "bush shrike",
  "n01600197": "Chlorophoneus, genus Chlorophoneus",
  "n01600341": "black-fronted bush shrike, Chlorophoneus nigrifrons",
  "n01600480": "Ptilonorhynchidae, family Ptilonorhynchidae",
  "n01600657": "bowerbird, catbird",
  "n01600909": "Ptilonorhynchus, genus Ptilonorhynchus",
  "n01601068": "satin bowerbird, satin bird, Ptilonorhynchus violaceus",
  "n01601268": "Chlamydera, genus Chlamydera",
  "n01601410": "great bowerbird, Chlamydera nuchalis",
  "n01601550": "Cinclidae, family Cinclidae",
  "n01601694": "water ouzel, dipper",
  "n01601919": "Cinclus, genus Cinclus",
  "n01602080": "European water ouzel, Cinclus aquaticus",
  "n01602209": "American water ouzel, Cinclus mexicanus",
  "n01602353": "Vireonidae, family Vireonidae",
  "n01602506": "genus Vireo",
  "n01602630": "vireo",
  "n01602832": "red-eyed vireo, Vireo olivaceous",
  "n01603000": "solitary vireo, Vireo solitarius",
  "n01603152": "blue-headed vireo, Vireo solitarius solitarius",
  "n01603316": "Bombycillidae, family Bombycillidae",
  "n01603478": "Bombycilla, genus bombycilla",
  "n01603600": "waxwing",
  "n01603812": "cedar waxwing, cedarbird, Bombycilla cedrorun",
  "n01603953": "Bohemian waxwing, Bombycilla garrulus",
  "n01604123": "Raptores, order Raptores",
  "n01604330": "bird of prey, raptor, raptorial bird",
  "n01604625": "Falconiformes, order Falconiformes",
  "n01604968": "Accipitriformes, order Accipitriformes",
  "n01605119": "Accipitridae, family Accipitridae",
  "n01605630": "hawk",
  "n01606097": "eyas",
  "n01606177": "tiercel, tercel, tercelet",
  "n01606335": "Accipiter, genus Accipiter",
  "n01606522": "goshawk, Accipiter gentilis",
  "n01606672": "sparrow hawk, Accipiter nisus",
  "n01606809": "Cooper's hawk, blue darter, Accipiter cooperii",
  "n01606978": "chicken hawk, hen hawk",
  "n01607103": "Buteo, genus Buteo",
  "n01607309": "buteonine",
  "n01607429": "redtail, red-tailed hawk, Buteo jamaicensis",
  "n01607600": "rough-legged hawk, roughleg, Buteo lagopus",
  "n01607812": "red-shouldered hawk, Buteo lineatus",
  "n01607962": "buzzard, Buteo buteo",
  "n01608086": "Pernis, genus Pernis",
  "n01608265": "honey buzzard, Pernis apivorus",
  "n01608432": "kite",
  "n01608685": "Milvus, genus-Milvus",
  "n01608814": "black kite, Milvus migrans",
  "n01608934": "Elanoides, genus Elanoides",
  "n01609062": "swallow-tailed kite, swallow-tailed hawk, Elanoides forficatus",
  "n01609236": "Elanus, genus Elanus",
  "n01609391": "white-tailed kite, Elanus leucurus",
  "n01609549": "Circus, genus Circus",
  "n01609751": "harrier",
  "n01609956": "marsh harrier, Circus Aeruginosus",
  "n01610100": "Montagu's harrier, Circus pygargus",
  "n01610226": "marsh hawk, northern harrier, hen harrier, Circus cyaneus",
  "n01610426": "Circaetus, genus Circaetus",
  "n01610552": "harrier eagle, short-toed eagle",
  "n01610758": "Falconidae, family Falconidae",
  "n01610955": "falcon",
  "n01611252": "Falco, genus Falco",
  "n01611472": "peregrine, peregrine falcon, Falco peregrinus",
  "n01611674": "falcon-gentle, falcon-gentil",
  "n01611800": "gyrfalcon, gerfalcon, Falco rusticolus",
  "n01611969": "kestrel, Falco tinnunculus",
  "n01612122": "sparrow hawk, American kestrel, kestrel, Falco sparverius",
  "n01612275": "pigeon hawk, merlin, Falco columbarius",
  "n01612476": "hobby, Falco subbuteo",
  "n01612628": "caracara",
  "n01612803": "Polyborus, genus Polyborus",
  "n01612955": "Audubon's caracara, Polyborus cheriway audubonii",
  "n01613177": "carancha, Polyborus plancus",
  "n01613294": "eagle, bird of Jove",
  "n01613615": "young bird",
  "n01613807": "eaglet",
  "n01613909": "Harpia, genus Harpia",
  "n01614038": "harpy, harpy eagle, Harpia harpyja",
  "n01614195": "Aquila, genus Aquila",
  "n01614343": "golden eagle, Aquila chrysaetos",
  "n01614556": "tawny eagle, Aquila rapax",
  "n01614690": "ringtail",
  "n01614769": "Haliaeetus, genus Haliaeetus",
  "n01614925": "bald eagle, American eagle, Haliaeetus leucocephalus",
  "n01615121": "sea eagle",
  "n01615303": "Kamchatkan sea eagle, Stellar's sea eagle, Haliaeetus pelagicus",
  "n01615458": "ern, erne, grey sea eagle, gray sea eagle, European sea eagle, white-tailed sea eagle, Haliatus albicilla",
  "n01615703": "fishing eagle, Haliaeetus leucorhyphus",
  "n01615825": "Pandionidae, family Pandionidae",
  "n01615949": "Pandion, genus Pandion",
  "n01616086": "osprey, fish hawk, fish eagle, sea eagle, Pandion haliaetus",
  "n01616318": "vulture",
  "n01616551": "Aegypiidae, family Aegypiidae",
  "n01616764": "Old World vulture",
  "n01616970": "Gyps, genus Gyps",
  "n01617095": "griffon vulture, griffon, Gyps fulvus",
  "n01617289": "Gypaetus, genus Gypaetus",
  "n01617443": "bearded vulture, lammergeier, lammergeyer, Gypaetus barbatus",
  "n01617633": "Neophron, genus Neophron",
  "n01617766": "Egyptian vulture, Pharaoh's chicken, Neophron percnopterus",
  "n01617949": "Aegypius, genus Aegypius",
  "n01618082": "black vulture, Aegypius monachus",
  "n01618220": "Sagittariidae, family Sagittariidae",
  "n01618356": "Sagittarius, genus Sagittarius",
  "n01618503": "secretary bird, Sagittarius serpentarius",
  "n01618671": "Cathartidae, family Cathartidae",
  "n01618922": "New World vulture, cathartid",
  "n01619152": "Cathartes, genus Cathartes",
  "n01619310": "buzzard, turkey buzzard, turkey vulture, Cathartes aura",
  "n01619536": "condor",
  "n01619675": "Vultur, genus Vultur",
  "n01619835": "Andean condor, Vultur gryphus",
  "n01620003": "Gymnogyps, genus Gymnogyps",
  "n01620135": "California condor, Gymnogyps californianus",
  "n01620282": "Coragyps, genus Coragyps",
  "n01620414": "black vulture, carrion crow, Coragyps atratus",
  "n01620575": "Sarcorhamphus, genus Sarcorhamphus",
  "n01620735": "king vulture, Sarcorhamphus papa",
  "n01620967": "Strigiformes, order Strigiformes",
  "n01621127": "owl, bird of Minerva, bird of night, hooter",
  "n01621635": "owlet",
  "n01621714": "Strigidae, family Strigidae",
  "n01621994": "Athene, genus Athene",
  "n01622120": "little owl, Athene noctua",
  "n01622230": "Bubo, genus Bubo",
  "n01622352": "horned owl",
  "n01622483": "great horned owl, Bubo virginianus",
  "n01622596": "Strix, genus Strix",
  "n01622779": "great grey owl, great gray owl, Strix nebulosa",
  "n01622959": "tawny owl, Strix aluco",
  "n01623110": "barred owl, Strix varia",
  "n01623284": "Otus, genus Otus",
  "n01623425": "screech owl, Otus asio",
  "n01623615": "screech owl",
  "n01623706": "scops owl",
  "n01623880": "spotted owl, Strix occidentalis",
  "n01624115": "Old World scops owl, Otus scops",
  "n01624212": "Oriental scops owl, Otus sunia",
  "n01624305": "hoot owl",
  "n01624406": "Surnia, genus Surnia",
  "n01624537": "hawk owl, Surnia ulula",
  "n01624707": "Asio, genus Asio",
  "n01624833": "long-eared owl, Asio otus",
  "n01624987": "Sceloglaux, genus Sceloglaux",
  "n01625121": "laughing owl, laughing jackass, Sceloglaux albifacies",
  "n01625275": "Tytonidae, family Tytonidae",
  "n01625417": "Tyto, genus Tyto",
  "n01625562": "barn owl, Tyto alba",
  "n01625747": "amphibia, class Amphibia",
  "n01626134": "amphibian family",
  "n01626600": "amphibian genus",
  "n01627424": "amphibian",
  "n01627786": "Hynerpeton, genus Hynerpeton",
  "n01627976": "Hynerpeton bassetti",
  "n01628148": "genus Ichthyostega",
  "n01628331": "Ichyostega",
  "n01628450": "Urodella, order Urodella, Caudata, order Caudata",
  "n01628770": "urodele, caudate",
  "n01628885": "Salamandridae, family Salamandridae",
  "n01629093": "Salamandra, genus Salamandra",
  "n01629276": "salamander",
  "n01629819": "European fire salamander, Salamandra salamandra",
  "n01629962": "spotted salamander, fire salamander, Salamandra maculosa",
  "n01630148": "alpine salamander, Salamandra atra",
  "n01630284": "newt, triton",
  "n01630533": "Triturus, genus Triturus",
  "n01630670": "common newt, Triturus vulgaris",
  "n01630795": "Notophthalmus, genus Notophthalmus",
  "n01630901": "red eft, Notophthalmus viridescens",
  "n01631035": "Taricha, genus Taricha",
  "n01631175": "Pacific newt",
  "n01631354": "rough-skinned newt, Taricha granulosa",
  "n01631512": "California newt, Taricha torosa",
  "n01631663": "eft",
  "n01631759": "Ambystomatidae, family Ambystomatidae",
  "n01631903": "Ambystoma, genus Ambystoma",
  "n01632047": "ambystomid, ambystomid salamander",
  "n01632308": "mole salamander, Ambystoma talpoideum",
  "n01632458": "spotted salamander, Ambystoma maculatum",
  "n01632601": "tiger salamander, Ambystoma tigrinum",
  "n01632777": "axolotl, mud puppy, Ambystoma mexicanum",
  "n01632952": "waterdog",
  "n01633047": "Cryptobranchidae, family Cryptobranchidae",
  "n01633250": "Cryptobranchus, genus Cryptobranchus",
  "n01633406": "hellbender, mud puppy, Cryptobranchus alleganiensis",
  "n01633578": "Megalobatrachus, genus Megalobatrachus",
  "n01633781": "giant salamander, Megalobatrachus maximus",
  "n01633949": "Proteidae, family Proteidae",
  "n01634092": "Proteus, genus Proteus",
  "n01634227": "olm, Proteus anguinus",
  "n01634392": "Necturus, genus Necturus",
  "n01634522": "mud puppy, Necturus maculosus",
  "n01634684": "Dicamptodontidae, family Dicamptodontidae",
  "n01634891": "genus Dicamptodon",
  "n01635027": "dicamptodon, dicamptodontid",
  "n01635176": "Pacific giant salamander, Dicamptodon ensatus",
  "n01635343": "Rhyacotriton, genus Rhyacotriton",
  "n01635480": "olympic salamander, Rhyacotriton olympicus",
  "n01635659": "Plethodontidae, family Plethodontidae",
  "n01635964": "Plethodon, genus Plethodon",
  "n01636127": "lungless salamander, plethodont",
  "n01636352": "eastern red-backed salamander, Plethodon cinereus",
  "n01636510": "western red-backed salamander, Plethodon vehiculum",
  "n01636675": "Desmograthus, genus Desmograthus",
  "n01636829": "dusky salamander",
  "n01636984": "Aneides, genus Aneides",
  "n01637112": "climbing salamander",
  "n01637338": "arboreal salamander, Aneides lugubris",
  "n01637478": "Batrachoseps, genus Batrachoseps",
  "n01637615": "slender salamander, worm salamander",
  "n01637796": "Hydromantes, genus Hydromantes",
  "n01637932": "web-toed salamander",
  "n01638194": "Shasta salamander, Hydromantes shastae",
  "n01638329": "limestone salamander, Hydromantes brunus",
  "n01638482": "Amphiumidae, family Amphiumidae",
  "n01638611": "genus Amphiuma",
  "n01638722": "amphiuma, congo snake, congo eel, blind eel",
  "n01638952": "Sirenidae, family Sirenidae",
  "n01639071": "genus Siren",
  "n01639187": "siren",
  "n01639369": "Salientia, order Salientia, Anura, order Anura, Batrachia, order Batrachia",
  "n01639765": "frog, toad, toad frog, anuran, batrachian, salientian",
  "n01640383": "Ranidae, family Ranidae",
  "n01640567": "Rana, genus Rana",
  "n01640846": "true frog, ranid",
  "n01641206": "wood-frog, wood frog, Rana sylvatica",
  "n01641391": "leopard frog, spring frog, Rana pipiens",
  "n01641577": "bullfrog, Rana catesbeiana",
  "n01641739": "green frog, spring frog, Rana clamitans",
  "n01641930": "cascades frog, Rana cascadae",
  "n01642097": "goliath frog, Rana goliath",
  "n01642257": "pickerel frog, Rana palustris",
  "n01642391": "tarahumara frog, Rana tarahumarae",
  "n01642539": "grass frog, Rana temporaria",
  "n01642671": "Leptodactylidae, family Leptodactylidae",
  "n01642943": "leptodactylid frog, leptodactylid",
  "n01643092": "Eleutherodactylus, genus Eleutherodactylus",
  "n01643255": "robber frog",
  "n01643374": "Hylactophryne, genus Hylactophryne",
  "n01643507": "barking frog, robber frog, Hylactophryne augusti",
  "n01643687": "Leptodactylus, genus Leptodactylus",
  "n01643896": "crapaud, South American bullfrog, Leptodactylus pentadactylus",
  "n01644104": "Polypedatidae, family Polypedatidae",
  "n01644245": "Polypedates, genus Polypedates",
  "n01644373": "tree frog, tree-frog",
  "n01644542": "Ascaphidae, family Ascaphidae",
  "n01644699": "Ascaphus, genus Ascaphus",
  "n01644900": "tailed frog, bell toad, ribbed toad, tailed toad, Ascaphus trui",
  "n01645093": "Leiopelmatidae, family Leiopelmatidae, Liopelmidae, family Liopelmidae",
  "n01645278": "Leiopelma, genus Leiopelma, Liopelma, genus Liopelma",
  "n01645466": "Liopelma hamiltoni",
  "n01645634": "Bufonidae, family Bufonidae",
  "n01645776": "true toad",
  "n01646134": "genus Bufo",
  "n01646292": "bufo",
  "n01646388": "agua, agua toad, Bufo marinus",
  "n01646555": "European toad, Bufo bufo",
  "n01646648": "natterjack, Bufo calamita",
  "n01646802": "American toad, Bufo americanus",
  "n01646902": "Eurasian green toad, Bufo viridis",
  "n01647033": "American green toad, Bufo debilis",
  "n01647180": "Yosemite toad, Bufo canorus",
  "n01647303": "Texas toad, Bufo speciosus",
  "n01647466": "southwestern toad, Bufo microscaphus",
  "n01647640": "western toad, Bufo boreas",
  "n01647803": "Discoglossidae, family Discoglossidae",
  "n01648001": "Alytes, genus Alytes",
  "n01648139": "obstetrical toad, midwife toad, Alytes obstetricans",
  "n01648356": "midwife toad, Alytes cisternasi",
  "n01648494": "Bombina, genus Bombina",
  "n01648620": "fire-bellied toad, Bombina bombina",
  "n01648818": "Pelobatidae, family Pelobatidae",
  "n01648993": "Scaphiopus, genus Scaphiopus",
  "n01649170": "spadefoot, spadefoot toad",
  "n01649412": "western spadefoot, Scaphiopus hammondii",
  "n01649556": "southern spadefoot, Scaphiopus multiplicatus",
  "n01649726": "plains spadefoot, Scaphiopus bombifrons",
  "n01649948": "Hylidae, family Hylidae",
  "n01650167": "tree toad, tree frog, tree-frog",
  "n01650509": "Hyla, genus Hyla",
  "n01650690": "spring peeper, Hyla crucifer",
  "n01650901": "Pacific tree toad, Hyla regilla",
  "n01651059": "canyon treefrog, Hyla arenicolor",
  "n01651285": "chameleon tree frog",
  "n01651370": "Acris, genus Acris",
  "n01651487": "cricket frog",
  "n01651641": "northern cricket frog, Acris crepitans",
  "n01651778": "eastern cricket frog, Acris gryllus",
  "n01651900": "Pseudacris, genus Pseudacris",
  "n01652026": "chorus frog",
  "n01652163": "Pternohyla, genus Pternohyla",
  "n01652297": "lowland burrowing treefrog, northern casque-headed frog, Pternohyla fodiens",
  "n01652583": "Microhylidae, family Microhylidae, Brevicipitidae, family Brevicipitidae",
  "n01652850": "Gastrophryne, genus Gastrophryne",
  "n01653026": "western narrow-mouthed toad, Gastrophryne olivacea",
  "n01653223": "eastern narrow-mouthed toad, Gastrophryne carolinensis",
  "n01653384": "Hypopachus, genus Hypopachus",
  "n01653509": "sheep frog",
  "n01653610": "Pipidae, family Pipidae",
  "n01653773": "tongueless frog",
  "n01653975": "Pipa, genus Pipa",
  "n01654083": "Surinam toad, Pipa pipa, Pipa americana",
  "n01654245": "Xenopodidae, family Xenopodidae",
  "n01654429": "Xenopus, genus Xenopus",
  "n01654637": "African clawed frog, Xenopus laevis",
  "n01654863": "South American poison toad",
  "n01654957": "Gymnophiona, order Gymnophiona",
  "n01655116": "Caeciliidae, family Caeciliidae, Caeciliadae, family Caeciliadae",
  "n01655344": "caecilian, blindworm",
  "n01655577": "Labyrinthodontia, superorder Labyrinthodontia, Labyrinthodonta, superorder Labyrinthodonta",
  "n01655951": "labyrinthodont",
  "n01656078": "Stereospondyli, order Stereospondyli",
  "n01656340": "Stegocephalia, order Stegocephalia",
  "n01656576": "Temnospondyli, order Temnospondyli",
  "n01656813": "reptile family",
  "n01657723": "reptile genus",
  "n01660719": "Reptilia, class Reptilia",
  "n01661091": "reptile, reptilian",
  "n01661404": "Anapsida, subclass Anapsida",
  "n01661592": "anapsid, anapsid reptile",
  "n01661818": "diapsid, diapsid reptile",
  "n01662060": "Diapsida, subclass Diapsida",
  "n01662274": "Chelonia, order Chelonia, Testudinata, order Testudinata, Testudines, order Testudines",
  "n01662622": "chelonian, chelonian reptile",
  "n01662784": "turtle",
  "n01663169": "Cheloniidae, family Cheloniidae, Chelonidae, family Chelonidae",
  "n01663401": "sea turtle, marine turtle",
  "n01663659": "Chelonia, genus Chelonia",
  "n01663782": "green turtle, Chelonia mydas",
  "n01663939": "Caretta, genus Caretta",
  "n01664065": "loggerhead, loggerhead turtle, Caretta caretta",
  "n01664244": "Lepidochelys, genus Lepidochelys",
  "n01664369": "ridley",
  "n01664492": "Atlantic ridley, bastard ridley, bastard turtle, Lepidochelys kempii",
  "n01664674": "Pacific ridley, olive ridley, Lepidochelys olivacea",
  "n01664862": "Eretmochelys, genus Eretmochelys",
  "n01664990": "hawksbill turtle, hawksbill, hawkbill, tortoiseshell turtle, Eretmochelys imbricata",
  "n01665238": "Dermochelyidae, family Dermochelyidae",
  "n01665372": "Dermochelys, genus Dermochelys",
  "n01665541": "leatherback turtle, leatherback, leathery turtle, Dermochelys coriacea",
  "n01665761": "Chelydridae, family Chelydridae",
  "n01665932": "snapping turtle",
  "n01666102": "Chelydra, genus Chelydra",
  "n01666228": "common snapping turtle, snapper, Chelydra serpentina",
  "n01666431": "Macroclemys, genus Macroclemys",
  "n01666585": "alligator snapping turtle, alligator snapper, Macroclemys temmincki",
  "n01666802": "Kinosternidae, family Kinosternidae",
  "n01666967": "Kinosternon, genus Kinosternon",
  "n01667114": "mud turtle",
  "n01667302": "Sternotherus, genus Sternotherus",
  "n01667432": "musk turtle, stinkpot",
  "n01667570": "Emydidae, family Emydidae",
  "n01667778": "terrapin",
  "n01667959": "Malaclemys, genus Malaclemys",
  "n01668091": "diamondback terrapin, Malaclemys centrata",
  "n01668257": "Pseudemys, genus Pseudemys",
  "n01668436": "red-bellied terrapin, red-bellied turtle, redbelly, Pseudemys rubriventris",
  "n01668665": "slider, yellow-bellied terrapin, Pseudemys scripta",
  "n01668892": "cooter, river cooter, Pseudemys concinna",
  "n01669068": "Terrapene, genus Terrapene",
  "n01669191": "box turtle, box tortoise",
  "n01669372": "Western box turtle, Terrapene ornata",
  "n01669527": "Chrysemys, genus Chrysemys",
  "n01669654": "painted turtle, painted terrapin, painted tortoise, Chrysemys picta",
  "n01669883": "Testudinidae, family Testudinidae",
  "n01670092": "tortoise",
  "n01670378": "Testudo, genus Testudo",
  "n01670535": "European tortoise, Testudo graeca",
  "n01670673": "Geochelone, genus Geochelone",
  "n01670802": "giant tortoise",
  "n01670961": "Gopherus, genus Gopherus",
  "n01671125": "gopher tortoise, gopher turtle, gopher, Gopherus polypemus",
  "n01671312": "Xerobates, genus Xerobates",
  "n01671479": "desert tortoise, Gopherus agassizii",
  "n01671705": "Texas tortoise",
  "n01671874": "Trionychidae, family Trionychidae",
  "n01672032": "soft-shelled turtle, pancake turtle",
  "n01672275": "Trionyx, genus Trionyx",
  "n01672432": "spiny softshell, Trionyx spiniferus",
  "n01672611": "smooth softshell, Trionyx muticus",
  "n01672767": "Lepidosauria, subclass Lepidosauria",
  "n01672950": "Rhynchocephalia, order Rhynchocephalia",
  "n01673118": "Sphenodon, genus Sphenodon",
  "n01673282": "tuatara, Sphenodon punctatum",
  "n01673503": "Squamata, order Squamata",
  "n01673668": "Sauria, suborder Sauria, Lacertilia, suborder Lacertilia",
  "n01674216": "saurian",
  "n01674464": "lizard",
  "n01674850": "Gekkonidae, family Gekkonidae",
  "n01674990": "gecko",
  "n01675225": "Ptychozoon, genus Ptychozoon",
  "n01675352": "flying gecko, fringed gecko, Ptychozoon homalocephalum",
  "n01675599": "Coleonyx, genus Coleonyx",
  "n01675722": "banded gecko",
  "n01675964": "Pygopodidae, family Pygopodidae",
  "n01676113": "Pygopus, genus Pygopus",
  "n01676313": "Iguanidae, family Iguanidae, Iguania, family Iguania",
  "n01676755": "iguanid, iguanid lizard",
  "n01677242": "genus Iguana",
  "n01677366": "common iguana, iguana, Iguana iguana",
  "n01677613": "Amblyrhynchus, genus Amblyrhynchus",
  "n01677747": "marine iguana, Amblyrhynchus cristatus",
  "n01677913": "Dipsosaurus, genus Dipsosaurus",
  "n01678043": "desert iguana, Dipsosaurus dorsalis",
  "n01678237": "Sauromalus, genus Sauromalus",
  "n01678343": "chuckwalla, Sauromalus obesus",
  "n01678522": "Callisaurus, genus Callisaurus",
  "n01678657": "zebra-tailed lizard, gridiron-tailed lizard, Callisaurus draconoides",
  "n01678887": "Uma, genus Uma",
  "n01679005": "fringe-toed lizard, Uma notata",
  "n01679178": "Holbrookia, genus Holbrookia",
  "n01679307": "earless lizard",
  "n01679494": "Crotaphytus, genus Crotaphytus",
  "n01679626": "collared lizard",
  "n01679837": "Gambelia, genus Gambelia",
  "n01679962": "leopard lizard",
  "n01680137": "Sceloporus, genus Sceloporus",
  "n01680264": "spiny lizard",
  "n01680478": "fence lizard",
  "n01680655": "western fence lizard, swift, blue-belly, Sceloporus occidentalis",
  "n01680813": "eastern fence lizard, pine lizard, Sceloporus undulatus",
  "n01680983": "sagebrush lizard, Sceloporus graciosus",
  "n01681200": "Uta, genus Uta",
  "n01681328": "side-blotched lizard, sand lizard, Uta stansburiana",
  "n01681513": "Urosaurus, genus Urosaurus",
  "n01681653": "tree lizard, Urosaurus ornatus",
  "n01681812": "Phrynosoma, genus Phrynosoma",
  "n01681940": "horned lizard, horned toad, horny frog",
  "n01682172": "Texas horned lizard, Phrynosoma cornutum",
  "n01682293": "Basiliscus, genus Basiliscus",
  "n01682435": "basilisk",
  "n01682588": "Anolis, genus Anolis",
  "n01682714": "American chameleon, anole, Anolis carolinensis",
  "n01682920": "Amphisbaenidae, family Amphisbaenidae",
  "n01683036": "Amphisbaena, genus Amphisbaena, Amphisbaenia, genus Amphisbaenia",
  "n01683201": "worm lizard",
  "n01683428": "Xantusiidae, family Xantusiidae",
  "n01683558": "night lizard",
  "n01683724": "Scincidae, family Scincidae",
  "n01683900": "Scincus, genus Scincus",
  "n01684012": "Scincella, genus Scincella",
  "n01684133": "skink, scincid, scincid lizard",
  "n01684435": "Eumeces, genus Eumeces",
  "n01684578": "western skink, Eumeces skiltonianus",
  "n01684741": "mountain skink, Eumeces callicephalus",
  "n01684941": "Cordylidae, family Cordylidae",
  "n01685107": "Cordylus, genus Cordylus",
  "n01685277": "Teiidae, family Teiidae",
  "n01685439": "teiid lizard, teiid",
  "n01685679": "Cnemidophorus, genus Cnemidophorus",
  "n01685808": "whiptail, whiptail lizard",
  "n01686044": "racerunner, race runner, six-lined racerunner, Cnemidophorus sexlineatus",
  "n01686220": "plateau striped whiptail, Cnemidophorus velox",
  "n01686403": "Chihuahuan spotted whiptail, Cnemidophorus exsanguis",
  "n01686609": "western whiptail, Cnemidophorus tigris",
  "n01686808": "checkered whiptail, Cnemidophorus tesselatus",
  "n01687009": "Tupinambis, genus Tupinambis",
  "n01687128": "teju",
  "n01687290": "caiman lizard",
  "n01687441": "Agamidae, family Agamidae",
  "n01687665": "agamid, agamid lizard",
  "n01687856": "genus Agama",
  "n01687978": "agama",
  "n01688106": "Chlamydosaurus, genus Chlamydosaurus",
  "n01688243": "frilled lizard, Chlamydosaurus kingi",
  "n01688428": "Draco, genus Draco",
  "n01688589": "dragon, flying dragon, flying lizard",
  "n01688812": "genus Moloch",
  "n01688961": "moloch",
  "n01689081": "mountain devil, spiny lizard, Moloch horridus",
  "n01689226": "Anguidae, family Anguidae",
  "n01689411": "anguid lizard",
  "n01689678": "Gerrhonotus, genus Gerrhonotus",
  "n01689811": "alligator lizard",
  "n01690005": "Anguis, genus Anguis",
  "n01690149": "blindworm, slowworm, Anguis fragilis",
  "n01690339": "Ophisaurus, genus Ophisaurus",
  "n01690466": "glass lizard, glass snake, joint snake",
  "n01690703": "Xenosauridae, family Xenosauridae",
  "n01690857": "Xenosaurus, genus Xenosaurus",
  "n01691085": "Anniellidae, family Anniellidae",
  "n01691217": "legless lizard",
  "n01691384": "Lanthanotidae, family Lanthanotidae",
  "n01691525": "Lanthanotus, genus Lanthanotus",
  "n01691652": "Lanthanotus borneensis",
  "n01691782": "Helodermatidae, family Helodermatidae",
  "n01691951": "venomous lizard",
  "n01692143": "Heloderma, genus Heloderma",
  "n01692333": "Gila monster, Heloderma suspectum",
  "n01692523": "beaded lizard, Mexican beaded lizard, Heloderma horridum",
  "n01692713": "Lacertidae, family Lacertidae",
  "n01692864": "lacertid lizard, lacertid",
  "n01693020": "Lacerta, genus Lacerta",
  "n01693175": "sand lizard, Lacerta agilis",
  "n01693334": "green lizard, Lacerta viridis",
  "n01693472": "Chamaeleontidae, family Chamaeleontidae, Chamaeleonidae, family Chamaeleonidae, Rhiptoglossa, family Rhiptoglossa",
  "n01693783": "chameleon, chamaeleon",
  "n01693995": "Chamaeleo, genus Chamaeleo, genus Chamaeleon",
  "n01694178": "African chameleon, Chamaeleo chamaeleon",
  "n01694311": "horned chameleon, Chamaeleo oweni",
  "n01694430": "Varanidae, family Varanidae",
  "n01694558": "Varanus, genus Varanus",
  "n01694709": "monitor, monitor lizard, varan",
  "n01694955": "African monitor, Varanus niloticus",
  "n01695060": "Komodo dragon, Komodo lizard, dragon lizard, giant lizard, Varanus komodoensis",
  "n01695259": "Archosauria, subclass Archosauria",
  "n01695681": "archosaur, archosaurian, archosaurian reptile",
  "n01696026": "Saurosuchus, genus Saurosuchus",
  "n01696151": "Proterochampsa, genus Proterochampsa",
  "n01696282": "Crocodylia, order Crocodylia, Crocodilia, order Crocodilia",
  "n01696525": "Loricata, order Loricata",
  "n01696633": "crocodilian reptile, crocodilian",
  "n01696849": "Crocodylidae, family Crocodylidae",
  "n01697002": "Crocodylus, genus Crocodylus, Crocodilus, genus Crocodilus",
  "n01697178": "crocodile",
  "n01697457": "African crocodile, Nile crocodile, Crocodylus niloticus",
  "n01697611": "Asian crocodile, Crocodylus porosus",
  "n01697749": "Morlett's crocodile",
  "n01697837": "Tomistoma, genus Tomistoma",
  "n01697978": "false gavial, Tomistoma schlegeli",
  "n01698144": "Alligatoridae, family Alligatoridae",
  "n01698303": "genus Alligator",
  "n01698434": "alligator, gator",
  "n01698640": "American alligator, Alligator mississipiensis",
  "n01698782": "Chinese alligator, Alligator sinensis",
  "n01698936": "genus Caiman",
  "n01699040": "caiman, cayman",
  "n01699254": "spectacled caiman, Caiman sclerops",
  "n01699415": "Gavialidae, family Gavialidae",
  "n01699537": "Gavialis, genus Gavialis",
  "n01699675": "gavial, Gavialis gangeticus",
  "n01699831": "dinosaur",
  "n01700076": "Ornithischia, order Ornithischia",
  "n01700470": "ornithischian, ornithischian dinosaur",
  "n01700754": "genus Pisanosaurus",
  "n01700924": "pisanosaur, pisanosaurus",
  "n01701052": "genus Staurikosaurus",
  "n01701205": "staurikosaur, staurikosaurus",
  "n01701334": "Thyreophora, suborder Thyreophora, thyreophoran",
  "n01701551": "armored dinosaur",
  "n01701697": "genus Stegosaurus",
  "n01701859": "stegosaur, stegosaurus, Stegosaur stenops",
  "n01702087": "genus Ankylosaurus",
  "n01702256": "ankylosaur, ankylosaurus",
  "n01702479": "Edmontonia",
  "n01702623": "Marginocephalia, suborder Marginocephalia, marginocephalian",
  "n01702862": "suborder Pachycephalosaurus",
  "n01703011": "bone-headed dinosaur",
  "n01703161": "pachycephalosaur, pachycephalosaurus",
  "n01703341": "Ceratopsia, suborder Ceratopsia",
  "n01703569": "ceratopsian, horned dinosaur",
  "n01703866": "Ceratopsidae, family Ceratopsidae",
  "n01703996": "genus Protoceratops",
  "n01704103": "protoceratops",
  "n01704184": "genus Triceratops",
  "n01704323": "triceratops",
  "n01704497": "genus Styracosaurus",
  "n01704626": "styracosaur, styracosaurus",
  "n01704847": "genus Psittacosaurus",
  "n01705010": "psittacosaur, psittacosaurus",
  "n01705247": "Euronithopoda, suborder Euronithopoda, euronithopod, Ornithopoda, suborder Ornithopoda",
  "n01705591": "ornithopod, ornithopod dinosaur",
  "n01705717": "Hadrosauridae, family Hadrosauridae",
  "n01705934": "hadrosaur, hadrosaurus, duck-billed dinosaur",
  "n01706247": "genus Anatotitan",
  "n01706401": "anatotitan",
  "n01706533": "genus Corythosaurus",
  "n01706686": "corythosaur, corythosaurus",
  "n01706865": "genus Edmontosaurus",
  "n01707000": "edmontosaurus",
  "n01707149": "genus Trachodon",
  "n01707294": "trachodon, trachodont",
  "n01707433": "Iguanodontidae, family Iguanodontidae",
  "n01707566": "genus Iguanodon",
  "n01707698": "iguanodon",
  "n01707895": "Saurischia, order Saurischia",
  "n01708106": "saurischian, saurischian dinosaur",
  "n01708332": "Sauropodomorpha, suborder Sauropodomorpha",
  "n01708648": "Prosauropoda, suborder Prosauropoda",
  "n01708778": "Sauropoda, suborder Sauropoda",
  "n01708998": "sauropod, sauropod dinosaur",
  "n01709278": "genus Apatosaurus, genus Brontosaurus",
  "n01709484": "apatosaur, apatosaurus, brontosaur, brontosaurus, thunder lizard, Apatosaurus excelsus",
  "n01709723": "genus Barosaurus",
  "n01709876": "barosaur, barosaurus",
  "n01710035": "genus Diplodocus",
  "n01710177": "diplodocus",
  "n01710348": "Titanosauridae, family Titanosauridae",
  "n01710529": "Titanosaurus, genus Titanosaurus",
  "n01710747": "titanosaur, titanosaurian",
  "n01710993": "genus Argentinosaurus",
  "n01711160": "argentinosaur",
  "n01711297": "Seismosaurus, genus Seismosaurus",
  "n01711496": "ground-shaker, seismosaur",
  "n01711662": "Theropoda, suborder Theropoda",
  "n01712008": "theropod, theropod dinosaur, bird-footed dinosaur",
  "n01712432": "suborder Ceratosaura",
  "n01712606": "genus Ceratosaurus",
  "n01712752": "ceratosaur, ceratosaurus",
  "n01713040": "genus Coelophysis",
  "n01713170": "coelophysis",
  "n01713310": "Carnosaura, suborder Carnosaura",
  "n01713508": "carnosaur",
  "n01713635": "genus Tyrannosaurus",
  "n01713764": "tyrannosaur, tyrannosaurus, Tyrannosaurus rex",
  "n01714059": "genus Allosaurus, genus Antrodemus",
  "n01714231": "allosaur, allosaurus",
  "n01714404": "genus Compsognathus",
  "n01714536": "compsognathus",
  "n01714686": "genus Herrerasaurus",
  "n01714843": "herrerasaur, herrerasaurus",
  "n01714982": "genus Eoraptor",
  "n01715134": "eoraptor",
  "n01715249": "Megalosauridae, family Megalosauridae",
  "n01715383": "genus Megalosaurus",
  "n01715518": "megalosaur, megalosaurus",
  "n01715692": "Ornithomimida, suborder Ornithomimida",
  "n01715888": "ornithomimid",
  "n01716122": "genus Struthiomimus",
  "n01716298": "struthiomimus",
  "n01716459": "genus Deinocheirus",
  "n01716594": "deinocheirus",
  "n01716732": "Maniraptora, suborder Maniraptora",
  "n01717016": "maniraptor",
  "n01717229": "oviraptorid",
  "n01717335": "genus Velociraptor",
  "n01717467": "velociraptor",
  "n01717666": "Dromaeosauridae, family Dromaeosauridae",
  "n01717860": "dromaeosaur",
  "n01717957": "genus Deinonychus",
  "n01718096": "deinonychus",
  "n01718276": "genus Utahraptor",
  "n01718414": "utahraptor, superslasher",
  "n01718632": "genus Mononychus",
  "n01718808": "Mononychus olecranus",
  "n01719175": "Synapsida, subclass Synapsida",
  "n01719403": "synapsid, synapsid reptile",
  "n01719645": "Therapsida, order Therapsida",
  "n01719914": "therapsid, protomammal",
  "n01720117": "Chronoperates, genus Chronoperates",
  "n01720266": "Chronoperates paradoxus",
  "n01720496": "Cynodontia, division Cynodontia",
  "n01720767": "cynodont",
  "n01720867": "Exaeretodon, genus Exaeretodon",
  "n01721010": "Dicynodontia, division Dicynodontia",
  "n01721174": "dicynodont",
  "n01721269": "Ischigualastia, genus Ischigualastia",
  "n01721404": "Ictodosauria, order Ictodosauria",
  "n01721567": "ictodosaur",
  "n01721718": "Pelycosauria, order Pelycosauria",
  "n01721898": "pelycosaur",
  "n01722085": "Edaphosauridae, family Edaphosauridae",
  "n01722254": "genus Edaphosaurus",
  "n01722389": "edaphosaurus",
  "n01722538": "genus Dimetrodon",
  "n01722670": "dimetrodon",
  "n01722828": "Pterosauria, order Pterosauria",
  "n01722998": "pterosaur, flying reptile",
  "n01723259": "Pterodactylidae, family Pterodactylidae",
  "n01723425": "Pterodactylus, genus Pterodactylus",
  "n01723579": "pterodactyl",
  "n01723678": "Thecodontia, order Thecodontia",
  "n01723883": "thecodont, thecodont reptile",
  "n01724055": "Ichthyosauria, order Ichthyosauria",
  "n01724231": "ichthyosaur",
  "n01724470": "Ichthyosauridae, family Ichthyosauridae",
  "n01724703": "genus Ichthyosaurus",
  "n01724840": "ichthyosaurus",
  "n01724947": "genus Stenopterygius",
  "n01725086": "stenopterygius, Stenopterygius quadrisicissus",
  "n01725240": "Sauropterygia, order Sauropterygia",
  "n01725427": "Plesiosauria, suborder Plesiosauria",
  "n01725570": "genus Plesiosaurus",
  "n01725713": "plesiosaur, plesiosaurus",
  "n01725933": "Nothosauria, suborder Nothosauria",
  "n01726079": "genus Nothosaurus",
  "n01726203": "nothosaur",
  "n01726390": "Serpentes, suborder Serpentes, Ophidia, suborder Ophidia",
  "n01726692": "snake, serpent, ophidian",
  "n01726960": "Colubridae, family Colubridae",
  "n01727646": "colubrid snake, colubrid",
  "n01728266": "hoop snake",
  "n01728445": "Carphophis, genus Carphophis",
  "n01728572": "thunder snake, worm snake, Carphophis amoenus",
  "n01728738": "Diadophis, genus Diadophis",
  "n01728920": "ringneck snake, ring-necked snake, ring snake",
  "n01729133": "Heterodon, genus Heterodon",
  "n01729322": "hognose snake, puff adder, sand viper",
  "n01729533": "Phyllorhynchus, genus Phyllorhynchus",
  "n01729672": "leaf-nosed snake",
  "n01729838": "Opheodrys, genus Opheodrys",
  "n01729977": "green snake, grass snake",
  "n01730185": "smooth green snake, Opheodrys vernalis",
  "n01730307": "rough green snake, Opheodrys aestivus",
  "n01730429": "Chlorophis, genus Chlorophis",
  "n01730563": "green snake",
  "n01730679": "Coluber, genus Coluber",
  "n01730812": "racer",
  "n01730960": "blacksnake, black racer, Coluber constrictor",
  "n01731137": "blue racer, Coluber constrictor flaviventris",
  "n01731277": "horseshoe whipsnake, Coluber hippocrepis",
  "n01731418": "Masticophis, genus Masticophis",
  "n01731545": "whip-snake, whip snake, whipsnake",
  "n01731764": "coachwhip, coachwhip snake, Masticophis flagellum",
  "n01731941": "California whipsnake, striped racer, Masticophis lateralis",
  "n01732093": "Sonoran whipsnake, Masticophis bilineatus",
  "n01732244": "rat snake",
  "n01732445": "Elaphe, genus Elaphe",
  "n01732614": "corn snake, red rat snake, Elaphe guttata",
  "n01732789": "black rat snake, blacksnake, pilot blacksnake, mountain blacksnake, Elaphe obsoleta",
  "n01732989": "chicken snake",
  "n01733094": "Ptyas, genus Ptyas",
  "n01733214": "Indian rat snake, Ptyas mucosus",
  "n01733346": "Arizona, genus Arizona",
  "n01733466": "glossy snake, Arizona elegans",
  "n01733634": "Pituophis, genus Pituophis",
  "n01733757": "bull snake, bull-snake",
  "n01733957": "gopher snake, Pituophis melanoleucus",
  "n01734104": "pine snake",
  "n01734273": "Lampropeltis, genus Lampropeltis",
  "n01734418": "king snake, kingsnake",
  "n01734637": "common kingsnake, Lampropeltis getulus",
  "n01734808": "milk snake, house snake, milk adder, checkered adder, Lampropeltis triangulum",
  "n01735062": "Thamnophis, genus Thamnophis",
  "n01735189": "garter snake, grass snake",
  "n01735439": "common garter snake, Thamnophis sirtalis",
  "n01735577": "ribbon snake, Thamnophis sauritus",
  "n01735728": "Western ribbon snake, Thamnophis proximus",
  "n01735898": "Tropidoclonion, genus Tropidoclonion",
  "n01736032": "lined snake, Tropidoclonion lineatum",
  "n01736256": "Sonora, genus Sonora",
  "n01736375": "ground snake, Sonora semiannulata",
  "n01736569": "Potamophis, genus Potamophis",
  "n01736696": "Haldea, genus Haldea",
  "n01736796": "eastern ground snake, Potamophis striatula, Haldea striatula",
  "n01737021": "water snake",
  "n01737197": "Natrix, genus Natrix",
  "n01737356": "Nerodia, genus Nerodia",
  "n01737472": "common water snake, banded water snake, Natrix sipedon, Nerodia sipedon",
  "n01737728": "water moccasin",
  "n01737875": "grass snake, ring snake, ringed snake, Natrix natrix",
  "n01738065": "viperine grass snake, Natrix maura",
  "n01738175": "Storeria, genus Storeria",
  "n01738306": "red-bellied snake, Storeria occipitamaculata",
  "n01738470": "Chilomeniscus, genus Chilomeniscus",
  "n01738601": "sand snake",
  "n01738731": "banded sand snake, Chilomeniscus cinctus",
  "n01738965": "Tantilla, genus Tantilla",
  "n01739094": "black-headed snake",
  "n01739260": "Oxybelis, genus Oxybelis",
  "n01739381": "vine snake",
  "n01739518": "Trimorphodon, genus Trimorphodon",
  "n01739647": "lyre snake",
  "n01739871": "Sonoran lyre snake, Trimorphodon lambda",
  "n01740005": "Hypsiglena, genus Hypsiglena",
  "n01740131": "night snake, Hypsiglena torquata",
  "n01740283": "Typhlopidae, family Typhlopidae",
  "n01740393": "Leptotyphlopidae, family Leptotyphlopidae",
  "n01740551": "blind snake, worm snake",
  "n01740721": "Leptotyphlops, genus Leptotyphlops",
  "n01740885": "western blind snake, Leptotyphlops humilis",
  "n01741116": "Drymarchon, genus Drymarchon",
  "n01741232": "indigo snake, gopher snake, Drymarchon corais",
  "n01741442": "eastern indigo snake, Drymarchon corais couperi",
  "n01741562": "constrictor",
  "n01741744": "Boidae, family Boidae",
  "n01741943": "boa",
  "n01742172": "boa constrictor, Constrictor constrictor",
  "n01742310": "Charina, genus Charina",
  "n01742447": "rubber boa, tow-headed snake, Charina bottae",
  "n01742680": "Lichanura, genus Lichanura",
  "n01742821": "rosy boa, Lichanura trivirgata",
  "n01742967": "Eunectes, genus Eunectes",
  "n01743086": "anaconda, Eunectes murinus",
  "n01743223": "Pythoninae, subfamily Pythoninae",
  "n01743449": "Pythonidae, family Pythonidae",
  "n01743605": "python",
  "n01743787": "genus Python",
  "n01743936": "carpet snake, Python variegatus, Morelia spilotes variegatus",
  "n01744100": "reticulated python, Python reticulatus",
  "n01744270": "Indian python, Python molurus",
  "n01744401": "rock python, rock snake, Python sebae",
  "n01744555": "amethystine python",
  "n01744657": "Elapidae, family Elapidae",
  "n01745125": "elapid, elapid snake",
  "n01745484": "coral snake, harlequin-snake, New World coral snake",
  "n01745780": "Micrurus, genus Micrurus",
  "n01745902": "eastern coral snake, Micrurus fulvius",
  "n01746063": "Micruroides, genus Micruroides",
  "n01746191": "western coral snake, Micruroides euryxanthus",
  "n01746359": "coral snake, Old World coral snake",
  "n01746565": "Calliophis, genus Calliophis, Callophis, genus Callophis",
  "n01746727": "Asian coral snake",
  "n01746818": "Aspidelaps, genus Aspidelaps",
  "n01746952": "African coral snake, Aspidelaps lubricus",
  "n01747144": "Rhynchoelaps, genus Rhynchoelaps",
  "n01747285": "Australian coral snake, Rhynchoelaps australis",
  "n01747466": "Denisonia, genus Denisonia",
  "n01747589": "copperhead, Denisonia superba",
  "n01747739": "Naja, genus Naja",
  "n01747885": "cobra",
  "n01748100": "hood",
  "n01748264": "Indian cobra, Naja naja",
  "n01748389": "asp, Egyptian cobra, Naja haje",
  "n01748560": "Ophiophagus, genus Ophiophagus",
  "n01748686": "black-necked cobra, spitting cobra, Naja nigricollis",
  "n01748906": "hamadryad, king cobra, Ophiophagus hannah, Naja hannah",
  "n01749141": "Hemachatus, genus Hemachatus",
  "n01749244": "ringhals, rinkhals, spitting snake, Hemachatus haemachatus",
  "n01749428": "Dendroaspis, genus Dendroaspis, Dendraspis, genus Dendraspis",
  "n01749582": "mamba",
  "n01749742": "black mamba, Dendroaspis augusticeps",
  "n01749939": "green mamba",
  "n01750027": "Acanthophis, genus Acanthophis",
  "n01750167": "death adder, Acanthophis antarcticus",
  "n01750315": "Notechis, genus Notechis",
  "n01750437": "tiger snake, Notechis scutatus",
  "n01750598": "Pseudechis, genus Pseudechis",
  "n01750743": "Australian blacksnake, Pseudechis porphyriacus",
  "n01750920": "Bungarus, genus Bungarus",
  "n01751036": "krait",
  "n01751215": "banded krait, banded adder, Bungarus fasciatus",
  "n01751353": "Oxyuranus, genus Oxyuranus",
  "n01751472": "taipan, Oxyuranus scutellatus",
  "n01751621": "Hydrophidae, family Hydrophidae",
  "n01751748": "sea snake",
  "n01751979": "Viperidae, family Viperidae",
  "n01752165": "viper",
  "n01752433": "Vipera, genus Vipera",
  "n01752585": "adder, common viper, Vipera berus",
  "n01752736": "asp, asp viper, Vipera aspis",
  "n01752889": "Bitis, genus Bitis",
  "n01753032": "puff adder, Bitis arietans",
  "n01753180": "gaboon viper, Bitis gabonica",
  "n01753354": "genus Cerastes, Aspis, genus Aspis",
  "n01753488": "horned viper, cerastes, sand viper, horned asp, Cerastes cornutus",
  "n01753721": "Crotalidae, family Crotalidae",
  "n01753959": "pit viper",
  "n01754190": "Agkistrodon, genus Agkistrodon, Ancistrodon, genus Ancistrodon",
  "n01754370": "copperhead, Agkistrodon contortrix",
  "n01754533": "water moccasin, cottonmouth, cottonmouth moccasin, Agkistrodon piscivorus",
  "n01754737": "rattle",
  "n01754876": "rattlesnake, rattler",
  "n01755274": "Crotalus, genus Crotalus",
  "n01755581": "diamondback, diamondback rattlesnake, Crotalus adamanteus",
  "n01755740": "timber rattlesnake, banded rattlesnake, Crotalus horridus horridus",
  "n01755952": "canebrake rattlesnake, canebrake rattler, Crotalus horridus atricaudatus",
  "n01756089": "prairie rattlesnake, prairie rattler, Western rattlesnake, Crotalus viridis",
  "n01756291": "sidewinder, horned rattlesnake, Crotalus cerastes",
  "n01756508": "Western diamondback, Western diamondback rattlesnake, Crotalus atrox",
  "n01756733": "rock rattlesnake, Crotalus lepidus",
  "n01756916": "tiger rattlesnake, Crotalus tigris",
  "n01757115": "Mojave rattlesnake, Crotalus scutulatus",
  "n01757343": "speckled rattlesnake, Crotalus mitchellii",
  "n01757547": "Sistrurus, genus Sistrurus",
  "n01757677": "massasauga, massasauga rattler, Sistrurus catenatus",
  "n01757901": "ground rattler, massasauga, Sistrurus miliaris",
  "n01758019": "Bothrops, genus Bothrops",
  "n01758141": "fer-de-lance, Bothrops atrops",
  "n01758308": "beak, bill, neb, nib, pecker",
  "n01758510": "beak",
  "n01758637": "cere",
  "n01758757": "carcase, carcass",
  "n01758895": "carrion",
  "n01759009": "roadkill",
  "n01759182": "arthropod family",
  "n01762525": "arthropod genus",
  "n01767199": "Arthropoda, phylum Arthropoda",
  "n01767661": "arthropod",
  "n01768244": "trilobite",
  "n01768402": "Chelicerata, superclass Chelicerata",
  "n01768596": "chelicera",
  "n01768808": "mouthpart",
  "n01768969": "Arachnida, class Arachnida",
  "n01769347": "arachnid, arachnoid",
  "n01769635": "Phalangida, order Phalangida, Opiliones, order Opiliones",
  "n01769789": "Phalangiidae, family Phalangiidae",
  "n01769930": "Phalangium, genus Phalangium",
  "n01770081": "harvestman, daddy longlegs, Phalangium opilio",
  "n01770263": "Scorpionida, order Scorpionida",
  "n01770393": "scorpion",
  "n01770553": "Chelonethida, order Chelonethida, Pseudoscorpionida, order Pseudoscorpionida, Pseudoscorpiones, order Pseudoscorpiones",
  "n01770795": "false scorpion, pseudoscorpion",
  "n01770967": "Chelifer, genus Chelifer",
  "n01771100": "book scorpion, Chelifer cancroides",
  "n01771246": "Pedipalpi, order Pedipalpi, Uropygi, order Uropygi",
  "n01771417": "whip-scorpion, whip scorpion",
  "n01771624": "Mastigoproctus, genus Mastigoproctus",
  "n01771766": "vinegarroon, Mastigoproctus giganteus",
  "n01771966": "Araneae, order Araneae, Araneida, order Araneida",
  "n01772222": "spider",
  "n01772664": "orb-weaving spider",
  "n01772782": "Argiopidae, family Argiopidae, orb-weaver",
  "n01772985": "Argiope, genus Argiope",
  "n01773157": "black and gold garden spider, Argiope aurantia",
  "n01773319": "Aranea, genus Aranea, Araneus, genus Araneus",
  "n01773549": "barn spider, Araneus cavaticus",
  "n01773797": "garden spider, Aranea diademata",
  "n01773930": "Theridiidae, family Theridiidae",
  "n01774097": "comb-footed spider, theridiid",
  "n01774252": "Latrodectus, genus Latrodectus",
  "n01774384": "black widow, Latrodectus mactans",
  "n01774595": "Theraphosidae, family Theraphosidae",
  "n01774750": "tarantula",
  "n01774918": "Lycosidae, family Lycosidae",
  "n01775062": "wolf spider, hunting spider",
  "n01775230": "Lycosa, genus Lycosa",
  "n01775370": "European wolf spider, tarantula, Lycosa tarentula",
  "n01775592": "Ctenizidae, family Ctenizidae",
  "n01775730": "trap-door spider",
  "n01775879": "Acarina, order Acarina",
  "n01776192": "acarine",
  "n01776313": "tick",
  "n01776546": "Ixodidae, family Ixodidae",
  "n01776705": "hard tick, ixodid",
  "n01777032": "Ixodes, genus Ixodes",
  "n01777304": "Ixodes dammini, deer tick",
  "n01777467": "Ixodes neotomae",
  "n01777649": "Ixodes pacificus, western black-legged tick",
  "n01777909": "Ixodes scapularis, black-legged tick",
  "n01778217": "sheep-tick, sheep tick, Ixodes ricinus",
  "n01778487": "Ixodes persulcatus",
  "n01778621": "Ixodes dentatus",
  "n01778801": "Ixodes spinipalpis",
  "n01778984": "Dermacentor, genus Dermacentor",
  "n01779148": "wood tick, American dog tick, Dermacentor variabilis",
  "n01779340": "Argasidae, family Argasidae",
  "n01779463": "soft tick, argasid",
  "n01779629": "mite",
  "n01779939": "web-spinning mite",
  "n01780026": "Acaridae, family Acaridae",
  "n01780142": "acarid",
  "n01780302": "Trombidiidae, family Trombidiidae",
  "n01780426": "trombidiid",
  "n01780551": "Trombiculidae, family Trombiculidae",
  "n01780696": "trombiculid",
  "n01780919": "Trombicula, genus Trombicula",
  "n01781071": "harvest mite, chigger, jigger, redbug",
  "n01781274": "Sarcoptidae, family Sarcoptidae",
  "n01781410": "Sarcoptes, genus Sarcoptes",
  "n01781570": "acarus, genus Acarus",
  "n01781698": "itch mite, sarcoptid",
  "n01781875": "rust mite",
  "n01782050": "Tetranychidae, family Tetranychidae",
  "n01782209": "spider mite, tetranychid",
  "n01782378": "Panonychus, genus Panonychus",
  "n01782516": "red spider, red spider mite, Panonychus ulmi",
  "n01782675": "superclass Myriapoda",
  "n01783017": "myriapod",
  "n01783205": "Pauropoda, class Pauropoda",
  "n01783384": "Symphyla, class Symphyla",
  "n01783571": "Scutigerella, genus Scutigerella",
  "n01783706": "garden centipede, garden symphilid, symphilid, Scutigerella immaculata",
  "n01783936": "Tardigrada, class Tardigrada",
  "n01784293": "tardigrade",
  "n01784427": "Chilopoda, class Chilopoda",
  "n01784675": "centipede",
  "n01784925": "prehensor",
  "n01785092": "toxicognath",
  "n01785234": "fang",
  "n01785392": "Scutigeridae, family Scutigeridae",
  "n01785532": "Scutigera, genus Scutigera",
  "n01785667": "house centipede, Scutigera coleoptrata",
  "n01785831": "Geophilomorpha, order Geophilomorpha",
  "n01786048": "Geophilidae, family Geophilidae",
  "n01786219": "Geophilus, genus Geophilus",
  "n01786402": "Diplopoda, class Diplopoda, Myriapoda, class Myriapoda",
  "n01786646": "millipede, millepede, milliped",
  "n01786879": "Pycnogonida, order Pycnogonida",
  "n01787006": "sea spider, pycnogonid",
  "n01787191": "Merostomata, class Merostomata",
  "n01787401": "Xiphosura, order Xiphosura",
  "n01787546": "Limulidae, family Limulidae",
  "n01787693": "Limulus, genus Limulus",
  "n01787835": "horseshoe crab, king crab, Limulus polyphemus, Xiphosurus polyphemus",
  "n01788157": "Tachypleus, genus Tachypleus",
  "n01788291": "Asian horseshoe crab",
  "n01788420": "Eurypterida, order Eurypterida",
  "n01788579": "eurypterid",
  "n01788730": "Pentastomida, subphylum Pentastomida",
  "n01788864": "tongue worm, pentastomid",
  "n01789064": "Galliformes, order Galliformes",
  "n01789386": "gallinaceous bird, gallinacean",
  "n01789740": "domestic fowl, fowl, poultry",
  "n01790171": "Dorking",
  "n01790304": "Plymouth Rock",
  "n01790398": "Cornish, Cornish fowl",
  "n01790557": "Rock Cornish",
  "n01790711": "game fowl",
  "n01790812": "cochin, cochin china",
  "n01790943": "Gallus, genus Gallus",
  "n01791107": "jungle fowl, gallina",
  "n01791314": "jungle cock",
  "n01791388": "jungle hen",
  "n01791463": "red jungle fowl, Gallus gallus",
  "n01791625": "chicken, Gallus gallus",
  "n01791954": "bantam",
  "n01792042": "chick, biddy",
  "n01792158": "cock, rooster",
  "n01792255": "comb, cockscomb, coxcomb",
  "n01792429": "cockerel",
  "n01792530": "capon",
  "n01792640": "hen, biddy",
  "n01792808": "cackler",
  "n01792955": "brood hen, broody, broody hen, setting hen, sitter",
  "n01793085": "mother hen",
  "n01793159": "layer",
  "n01793249": "pullet",
  "n01793340": "spring chicken",
  "n01793435": "Rhode Island red",
  "n01793565": "Dominique, Dominick",
  "n01793715": "Orpington",
  "n01793818": "Meleagrididae, family Meleagrididae",
  "n01793988": "Meleagris, genus Meleagris",
  "n01794158": "turkey, Meleagris gallopavo",
  "n01794344": "turkey cock, gobbler, tom, tom turkey",
  "n01794460": "Agriocharis, genus Agriocharis",
  "n01794651": "ocellated turkey, Agriocharis ocellata",
  "n01794813": "Tetraonidae, family Tetraonidae",
  "n01795088": "grouse",
  "n01795425": "Lyrurus, genus Lyrurus",
  "n01795545": "black grouse",
  "n01795735": "European black grouse, heathfowl, Lyrurus tetrix",
  "n01795900": "Asian black grouse, Lyrurus mlokosiewiczi",
  "n01796019": "blackcock, black cock",
  "n01796105": "greyhen, grayhen, grey hen, gray hen, heath hen",
  "n01796222": "Lagopus, genus Lagopus",
  "n01796340": "ptarmigan",
  "n01796519": "red grouse, moorfowl, moorbird, moor-bird, moorgame, Lagopus scoticus",
  "n01796729": "moorhen",
  "n01796800": "moorcock",
  "n01796870": "Tetrao, genus Tetrao",
  "n01797020": "capercaillie, capercailzie, horse of the wood, Tetrao urogallus",
  "n01797180": "Canachites, genus Canachites",
  "n01797307": "spruce grouse, Canachites canadensis",
  "n01797472": "Centrocercus, genus Centrocercus",
  "n01797601": "sage grouse, sage hen, Centrocercus urophasianus",
  "n01797767": "Bonasa, genus Bonasa",
  "n01797886": "ruffed grouse, partridge, Bonasa umbellus",
  "n01798052": "Pedioecetes, genus Pedioecetes",
  "n01798168": "sharp-tailed grouse, sprigtail, sprig tail, Pedioecetes phasianellus",
  "n01798352": "Tympanuchus, genus Tympanuchus",
  "n01798484": "prairie chicken, prairie grouse, prairie fowl",
  "n01798706": "greater prairie chicken, Tympanuchus cupido",
  "n01798839": "lesser prairie chicken, Tympanuchus pallidicinctus",
  "n01798979": "heath hen, Tympanuchus cupido cupido",
  "n01799086": "Cracidae, family Cracidae",
  "n01799302": "guan",
  "n01799540": "Crax, genus Crax",
  "n01799679": "curassow",
  "n01799876": "Penelope, genus Penelope",
  "n01800042": "Pipile, genus Pipile",
  "n01800195": "piping guan",
  "n01800286": "Ortalis, genus Ortalis",
  "n01800424": "chachalaca",
  "n01800633": "Texas chachalaca, Ortilis vetula macalli",
  "n01800759": "Megapodiidae, family Megapodiidae",
  "n01800963": "Megapodius, genus-Megapodius",
  "n01801088": "megapode, mound bird, mound-bird, mound builder, scrub fowl",
  "n01801371": "genus Leipoa",
  "n01801479": "mallee fowl, leipoa, lowan, Leipoa ocellata",
  "n01801672": "mallee hen",
  "n01801753": "Alectura, genus Alectura",
  "n01801876": "brush turkey, Alectura lathami",
  "n01802033": "Macrocephalon, genus Macrocephalon",
  "n01802159": "maleo, Macrocephalon maleo",
  "n01802309": "Phasianidae, family Phasianidae",
  "n01802721": "phasianid",
  "n01802895": "Phasianus, genus Phasianus",
  "n01803078": "pheasant",
  "n01803362": "ring-necked pheasant, Phasianus colchicus",
  "n01803548": "genus Afropavo",
  "n01803641": "afropavo, Congo peafowl, Afropavo congensis",
  "n01803764": "Argusianus, genus Argusianus",
  "n01803893": "argus, argus pheasant",
  "n01804029": "Chrysolophus, genus Chrysolophus",
  "n01804163": "golden pheasant, Chrysolophus pictus",
  "n01804340": "Colinus, genus Colinus",
  "n01804478": "bobwhite, bobwhite quail, partridge",
  "n01804653": "northern bobwhite, Colinus virginianus",
  "n01804796": "Coturnix, genus Coturnix",
  "n01804921": "Old World quail",
  "n01805070": "migratory quail, Coturnix coturnix, Coturnix communis",
  "n01805199": "Lophophorus, genus Lophophorus",
  "n01805321": "monal, monaul",
  "n01805446": "Odontophorus, genus Odontophorus",
  "n01805692": "Pavo, genus Pavo",
  "n01805801": "peafowl, bird of Juno",
  "n01806061": "peachick, pea-chick",
  "n01806143": "peacock",
  "n01806297": "peahen",
  "n01806364": "blue peafowl, Pavo cristatus",
  "n01806467": "green peafowl, Pavo muticus",
  "n01806567": "quail",
  "n01806740": "Lofortyx, genus Lofortyx",
  "n01806847": "California quail, Lofortyx californicus",
  "n01806984": "genus Tragopan",
  "n01807105": "tragopan",
  "n01807265": "Perdicidae, subfamily Perdicidae, Perdicinae, subfamily Perdicinae",
  "n01807496": "partridge",
  "n01807701": "Perdix, genus Perdix",
  "n01807828": "Hungarian partridge, grey partridge, gray partridge, Perdix perdix",
  "n01807988": "Alectoris, genus Alectoris",
  "n01808140": "red-legged partridge, Alectoris ruffa",
  "n01808291": "Greek partridge, rock partridge, Alectoris graeca",
  "n01808447": "Oreortyx, genus Oreortyx",
  "n01808596": "mountain quail, mountain partridge, Oreortyx picta palmeri",
  "n01808785": "Numididae, subfamily Numididae, Numidinae, subfamily Numidinae",
  "n01808989": "Numida, genus Numida",
  "n01809106": "guinea fowl, guinea, Numida meleagris",
  "n01809371": "guinea hen",
  "n01809446": "Opisthocomidae, family Opisthocomidae",
  "n01809592": "Opisthocomus, genus Opisthocomus",
  "n01809752": "hoatzin, hoactzin, stinkbird, Opisthocomus hoazin",
  "n01809977": "Tinamiformes, order Tinamiformes",
  "n01810132": "Tinamidae, family Tinamidae",
  "n01810268": "tinamou, partridge",
  "n01810466": "Columbiformes, order Columbiformes",
  "n01810700": "columbiform bird",
  "n01810946": "Raphidae, family Raphidae",
  "n01811104": "Raphus, genus Raphus",
  "n01811243": "dodo, Raphus cucullatus",
  "n01811394": "Pezophaps, genus Pezophaps",
  "n01811542": "solitaire, Pezophaps solitaria",
  "n01811682": "Columbidae, family Columbidae",
  "n01811909": "pigeon",
  "n01812187": "pouter pigeon, pouter",
  "n01812337": "dove",
  "n01812471": "Columba, genus Columba",
  "n01812662": "rock dove, rock pigeon, Columba livia",
  "n01812866": "band-tailed pigeon, band-tail pigeon, bandtail, Columba fasciata",
  "n01813088": "wood pigeon, ringdove, cushat, Columba palumbus",
  "n01813256": "Streptopelia, genus Streptopelia",
  "n01813385": "turtledove",
  "n01813532": "Streptopelia turtur",
  "n01813658": "ringdove, Streptopelia risoria",
  "n01813811": "Stictopelia, genus Stictopelia",
  "n01813948": "Australian turtledove, turtledove, Stictopelia cuneata",
  "n01814091": "Zenaidura, genus Zenaidura",
  "n01814217": "mourning dove, Zenaidura macroura",
  "n01814370": "domestic pigeon",
  "n01814549": "squab",
  "n01814620": "fairy swallow",
  "n01814755": "roller, tumbler, tumbler pigeon",
  "n01814921": "homing pigeon, homer",
  "n01815036": "carrier pigeon",
  "n01815135": "Ectopistes, genus Ectopistes",
  "n01815270": "passenger pigeon, Ectopistes migratorius",
  "n01815431": "Pteroclididae, family Pteroclididae",
  "n01815601": "sandgrouse, sand grouse",
  "n01815855": "Pterocles, genus Pterocles",
  "n01816017": "painted sandgrouse, Pterocles indicus",
  "n01816140": "pin-tailed sandgrouse, pin-tailed grouse, Pterocles alchata",
  "n01816336": "Syrrhaptes, genus Syrrhaptes",
  "n01816474": "pallas's sandgrouse, Syrrhaptes paradoxus",
  "n01816635": "Psittaciformes, order Psittaciformes",
  "n01816887": "parrot",
  "n01817263": "popinjay",
  "n01817346": "poll, poll parrot",
  "n01817424": "Psittacidae, family Psittacidae",
  "n01817772": "Psittacus, genus Psittacus",
  "n01817953": "African grey, African gray, Psittacus erithacus",
  "n01818169": "Amazona, genus Amazona",
  "n01818299": "amazon",
  "n01818409": "Ara, genus Ara",
  "n01818515": "macaw",
  "n01818704": "Nestor, genus Nestor",
  "n01818832": "kea, Nestor notabilis",
  "n01818959": "Kakatoe, genus Kakatoe, Cacatua, genus Cacatua",
  "n01819115": "cockatoo",
  "n01819313": "sulphur-crested cockatoo, Kakatoe galerita, Cacatua galerita",
  "n01819465": "pink cockatoo, Kakatoe leadbeateri",
  "n01819600": "Nymphicus, genus Nymphicus",
  "n01819734": "cockateel, cockatiel, cockatoo parrot, Nymphicus hollandicus",
  "n01819918": "Agapornis, genus Agapornis",
  "n01820052": "lovebird",
  "n01820190": "Loriinae, subfamily Loriinae",
  "n01820348": "lory",
  "n01820546": "lorikeet",
  "n01820664": "Glossopsitta, genus Glossopsitta",
  "n01820801": "varied Lorikeet, Glossopsitta versicolor",
  "n01820937": "Trichoglossus, genus Trichoglossus",
  "n01821076": "rainbow lorikeet, Trichoglossus moluccanus",
  "n01821203": "parakeet, parrakeet, parroket, paraquet, paroquet, parroquet",
  "n01821418": "Conuropsis, genus Conuropsis",
  "n01821554": "Carolina parakeet, Conuropsis carolinensis",
  "n01821727": "Melopsittacus, genus Melopsittacus",
  "n01821869": "budgerigar, budgereegah, budgerygah, budgie, grass parakeet, lovebird, shell parakeet, Melopsittacus undulatus",
  "n01822164": "Psittacula, genus Psittacula",
  "n01822300": "ring-necked parakeet, Psittacula krameri",
  "n01822423": "Cuculiformes, order Cuculiformes",
  "n01822602": "cuculiform bird",
  "n01822773": "Cuculidae, family Cuculidae",
  "n01823013": "cuckoo",
  "n01823279": "Cuculus, genus Cuculus",
  "n01823414": "European cuckoo, Cuculus canorus",
  "n01823610": "Coccyzus, genus Coccyzus",
  "n01823740": "black-billed cuckoo, Coccyzus erythropthalmus",
  "n01823912": "Geococcyx, genus Geococcyx",
  "n01824035": "roadrunner, chaparral cock, Geococcyx californianus",
  "n01824227": "Crotophaga, genus Crotophaga",
  "n01824344": "ani",
  "n01824443": "Centropus, genus Centropus",
  "n01824575": "coucal",
  "n01824749": "crow pheasant, Centropus sinensis",
  "n01824862": "pheasant coucal, pheasant cuckoo, Centropus phasianinus",
  "n01825009": "Musophagidae, family Musophagidae",
  "n01825155": "Musophaga, genus Musophaga",
  "n01825278": "touraco, turaco, turacou, turakoo",
  "n01825417": "Coraciiformes, order Coraciiformes",
  "n01825758": "Picariae, order Picariae",
  "n01825930": "coraciiform bird",
  "n01826223": "Coraciidae, family Coraciidae",
  "n01826364": "roller",
  "n01826542": "Coracias, genus Coracias",
  "n01826680": "European roller, Coracias garrulus",
  "n01826844": "ground roller",
  "n01826998": "Alcedinidae, family Alcedinidae",
  "n01827202": "halcyon",
  "n01827403": "kingfisher",
  "n01827658": "Alcedo, genus Alcedo",
  "n01827793": "Eurasian kingfisher, Alcedo atthis",
  "n01827948": "Ceryle, genus Ceryle",
  "n01828096": "belted kingfisher, Ceryle alcyon",
  "n01828267": "Dacelo, genus Dacelo",
  "n01828397": "Halcyon, genus Halcyon",
  "n01828556": "kookaburra, laughing jackass, Dacelo gigas",
  "n01828714": "Meropidae, family Meropidae",
  "n01828856": "Merops, genus Merops",
  "n01828970": "bee eater",
  "n01829143": "Bucerotidae, family Bucerotidae",
  "n01829288": "Buceros, genus Buceros",
  "n01829413": "hornbill",
  "n01829602": "Upupidae, family Upupidae",
  "n01829739": "Upupa, genus Upupa",
  "n01829869": "hoopoe, hoopoo",
  "n01830042": "Euopean hoopoe, Upupa epops",
  "n01830183": "Phoeniculidae, family Phoeniculidae",
  "n01830316": "Phoeniculus, genus Phoeniculus",
  "n01830479": "wood hoopoe",
  "n01830623": "Momotidae, family Momotidae",
  "n01830799": "Momotus, genus Momotus",
  "n01830915": "motmot, momot",
  "n01831078": "Todidae, family Todidae",
  "n01831231": "Todus, genus Todus",
  "n01831360": "tody",
  "n01831519": "Apodiformes, order Apodiformes",
  "n01831712": "apodiform bird",
  "n01831930": "Apodidae, family Apodidae",
  "n01832167": "swift",
  "n01832381": "Apus, genus Apus",
  "n01832493": "European swift, Apus apus",
  "n01832684": "Chateura, genus Chateura",
  "n01832813": "chimney swift, chimney swallow, Chateura pelagica",
  "n01832979": "Collocalia, genus Collocalia",
  "n01833112": "swiftlet, Collocalia inexpectata",
  "n01833283": "Hemiprocnidae, family Hemiprocnidae",
  "n01833415": "tree swift, crested swift",
  "n01833619": "Trochilidae, family Trochilidae",
  "n01833805": "hummingbird",
  "n01834039": "Archilochus, genus Archilochus",
  "n01834177": "Archilochus colubris",
  "n01834284": "Chalcostigma, genus Chalcostigma",
  "n01834412": "Ramphomicron, genus Ramphomicron",
  "n01834540": "thornbill",
  "n01834702": "Caprimulgiformes, order Caprimulgiformes",
  "n01834918": "caprimulgiform bird",
  "n01835087": "Caprimulgidae, family Caprimulgidae",
  "n01835276": "goatsucker, nightjar, caprimulgid",
  "n01835584": "Caprimulgus, genus Caprimulgus",
  "n01835769": "European goatsucker, European nightjar, Caprimulgus europaeus",
  "n01835918": "chuck-will's-widow, Caprimulgus carolinensis",
  "n01836087": "whippoorwill, Caprimulgus vociferus",
  "n01836246": "Chordeiles, genus Chordeiles",
  "n01836384": "nighthawk, bullbat, mosquito hawk",
  "n01836527": "Phalaenoptilus, genus Phalaenoptilus",
  "n01836673": "poorwill, Phalaenoptilus nuttallii",
  "n01836809": "Podargidae, family Podargidae",
  "n01836953": "Podargus, genus Podargus",
  "n01837072": "frogmouth",
  "n01837230": "Steatornithidae, family Steatornithidae",
  "n01837363": "Steatornis, genus Steatornis",
  "n01837526": "oilbird, guacharo, Steatornis caripensis",
  "n01837746": "Piciformes, order Piciformes",
  "n01838038": "piciform bird",
  "n01838326": "Picidae, family Picidae",
  "n01838598": "woodpecker, peckerwood, pecker",
  "n01838961": "Picus, genus Picus",
  "n01839086": "green woodpecker, Picus viridis",
  "n01839221": "Picoides, genus Picoides",
  "n01839330": "downy woodpecker",
  "n01839470": "Colaptes, genus Colaptes",
  "n01839598": "flicker",
  "n01839750": "yellow-shafted flicker, Colaptes auratus, yellowhammer",
  "n01839949": "gilded flicker, Colaptes chrysoides",
  "n01840120": "red-shafted flicker, Colaptes caper collaris",
  "n01840278": "Campephilus, genus Campephilus",
  "n01840412": "ivorybill, ivory-billed woodpecker, Campephilus principalis",
  "n01840643": "Melanerpes, genus Melanerpes",
  "n01840775": "redheaded woodpecker, redhead, Melanerpes erythrocephalus",
  "n01840968": "Sphyrapicus, genus Sphyrapicus",
  "n01841102": "sapsucker",
  "n01841288": "yellow-bellied sapsucker, Sphyrapicus varius",
  "n01841441": "red-breasted sapsucker, Sphyrapicus varius ruber",
  "n01841569": "Jynx, genus Jynx",
  "n01841679": "wryneck",
  "n01841815": "Picumnus, genus Picumnus",
  "n01841943": "piculet",
  "n01842111": "Capitonidae, family Capitonidae",
  "n01842235": "barbet",
  "n01842380": "Bucconidae, family Bucconidae",
  "n01842504": "puffbird",
  "n01842655": "Indicatoridae, family Indicatoridae",
  "n01842788": "honey guide",
  "n01842942": "Galbulidae, family Galbulidae",
  "n01843065": "jacamar",
  "n01843238": "Ramphastidae, family Ramphastidae",
  "n01843383": "toucan",
  "n01843576": "Aulacorhyncus, genus Aulacorhyncus",
  "n01843719": "toucanet",
  "n01843805": "Trogoniformes, order Trogoniformes",
  "n01843932": "Trogonidae, family Trogonidae",
  "n01844125": "genus Trogon",
  "n01844231": "trogon",
  "n01844414": "Pharomacrus, genus Pharomacrus",
  "n01844551": "quetzal, quetzal bird",
  "n01844746": "resplendent quetzel, resplendent trogon, Pharomacrus mocino",
  "n01844917": "aquatic bird",
  "n01845132": "waterfowl, water bird, waterbird",
  "n01845272": "Anseriformes, order Anseriformes",
  "n01845477": "anseriform bird",
  "n01845627": "Anatidae, family Anatidae",
  "n01846156": "Anseres, suborder Anseres",
  "n01846331": "duck",
  "n01847000": "drake",
  "n01847089": "quack-quack",
  "n01847170": "duckling",
  "n01847253": "diving duck",
  "n01847407": "dabbling duck, dabbler",
  "n01847565": "Anas, genus Anas",
  "n01847806": "mallard, Anas platyrhynchos",
  "n01847978": "black duck, Anas rubripes",
  "n01848123": "teal",
  "n01848323": "greenwing, green-winged teal, Anas crecca",
  "n01848453": "bluewing, blue-winged teal, Anas discors",
  "n01848555": "garganey, Anas querquedula",
  "n01848648": "widgeon, wigeon, Anas penelope",
  "n01848840": "American widgeon, baldpate, Anas americana",
  "n01848976": "shoveler, shoveller, broadbill, Anas clypeata",
  "n01849157": "pintail, pin-tailed duck, Anas acuta",
  "n01849348": "Tadorna, genus Tadorna",
  "n01849466": "sheldrake",
  "n01849676": "shelduck",
  "n01849747": "Oxyura, genus Oxyura",
  "n01849863": "ruddy duck, Oxyura jamaicensis",
  "n01850035": "Bucephala, genus Bucephala",
  "n01850192": "bufflehead, butterball, dipper, Bucephela albeola",
  "n01850373": "goldeneye, whistler, Bucephela clangula",
  "n01850553": "Barrow's goldeneye, Bucephala islandica",
  "n01850676": "Aythya, genus Aythya",
  "n01850873": "canvasback, canvasback duck, Aythya valisineria",
  "n01851038": "pochard, Aythya ferina",
  "n01851207": "redhead, Aythya americana",
  "n01851375": "scaup, scaup duck, bluebill, broadbill",
  "n01851573": "greater scaup, Aythya marila",
  "n01851731": "lesser scaup, lesser scaup duck, lake duck, Aythya affinis",
  "n01851895": "wild duck",
  "n01851996": "Aix, genus Aix",
  "n01852142": "wood duck, summer duck, wood widgeon, Aix sponsa",
  "n01852329": "wood drake",
  "n01852400": "mandarin duck, Aix galericulata",
  "n01852544": "Cairina, genus Cairina",
  "n01852671": "muscovy duck, musk duck, Cairina moschata",
  "n01852861": "sea duck",
  "n01853072": "Somateria, genus Somateria",
  "n01853195": "eider, eider duck",
  "n01853379": "Melanitta, genus Melanitta",
  "n01853498": "scoter, scooter",
  "n01853666": "common scoter, Melanitta nigra",
  "n01853763": "Clangula, genus Clangula",
  "n01853870": "old squaw, oldwife, Clangula hyemalis",
  "n01854047": "Merginae, subfamily Merginae",
  "n01854223": "Mergus, genus Mergus",
  "n01854415": "merganser, fish duck, sawbill, sheldrake",
  "n01854700": "goosander, Mergus merganser",
  "n01854838": "American merganser, Mergus merganser americanus",
  "n01855032": "red-breasted merganser, Mergus serrator",
  "n01855188": "smew, Mergus albellus",
  "n01855343": "Lophodytes, genus Lophodytes",
  "n01855476": "hooded merganser, hooded sheldrake, Lophodytes cucullatus",
  "n01855672": "goose",
  "n01856072": "gosling",
  "n01856155": "gander",
  "n01856225": "Anser, genus Anser",
  "n01856380": "Chinese goose, Anser cygnoides",
  "n01856553": "greylag, graylag, greylag goose, graylag goose, Anser anser",
  "n01856748": "Chen, subgenus Chen",
  "n01856890": "blue goose, Chen caerulescens",
  "n01857079": "snow goose",
  "n01857171": "Branta, genus Branta",
  "n01857325": "brant, brant goose, brent, brent goose",
  "n01857512": "common brant goose, Branta bernicla",
  "n01857632": "honker, Canada goose, Canadian goose, Branta canadensis",
  "n01857851": "barnacle goose, barnacle, Branta leucopsis",
  "n01858023": "Anserinae, subfamily Anserinae",
  "n01858162": "genus Coscoroba",
  "n01858281": "coscoroba",
  "n01858441": "swan",
  "n01858780": "cob",
  "n01858845": "pen",
  "n01858906": "cygnet",
  "n01858989": "Cygnus, genus Cygnus",
  "n01859190": "mute swan, Cygnus olor",
  "n01859325": "whooper, whooper swan, Cygnus cygnus",
  "n01859496": "tundra swan, Cygnus columbianus",
  "n01859689": "whistling swan, Cygnus columbianus columbianus",
  "n01859852": "Bewick's swan, Cygnus columbianus bewickii",
  "n01860002": "trumpeter, trumpeter swan, Cygnus buccinator",
  "n01860187": "black swan, Cygnus atratus",
  "n01860337": "Anhimidae, family Anhimidae",
  "n01860497": "screamer",
  "n01860713": "Anhima, genus Anhima",
  "n01860864": "horned screamer, Anhima cornuta",
  "n01861025": "Chauna, genus Chauna",
  "n01861148": "crested screamer",
  "n01861330": "chaja, Chauna torquata",
  "n01861465": "Mammalia, class Mammalia",
  "n01861778": "mammal, mammalian",
  "n01862399": "female mammal",
  "n01862557": "mammal family",
  "n01864707": "mammal genus",
  "n01871265": "tusker",
  "n01871406": "Prototheria, subclass Prototheria",
  "n01871543": "prototherian",
  "n01871699": "Monotremata, order Monotremata",
  "n01871875": "monotreme, egg-laying mammal",
  "n01872094": "Tachyglossidae, family Tachyglossidae",
  "n01872244": "Tachyglossus, genus Tachyglossus",
  "n01872401": "echidna, spiny anteater, anteater",
  "n01872635": "Zaglossus, genus Zaglossus",
  "n01872772": "echidna, spiny anteater, anteater",
  "n01873007": "Ornithorhynchidae, family Ornithorhynchidae",
  "n01873144": "Ornithorhynchus, genus Ornithorhynchus",
  "n01873310": "platypus, duckbill, duckbilled platypus, duck-billed platypus, Ornithorhynchus anatinus",
  "n01873622": "Pantotheria, subclass Pantotheria",
  "n01873850": "Metatheria, subclass Metatheria",
  "n01873982": "metatherian",
  "n01874126": "Marsupialia, order Marsupialia",
  "n01874434": "marsupial, pouched mammal",
  "n01874784": "Didelphidae, family Didelphidae",
  "n01874928": "opossum, possum",
  "n01875165": "Didelphis, genus Didelphis",
  "n01875313": "common opossum, Didelphis virginiana, Didelphis marsupialis",
  "n01875610": "crab-eating opossum",
  "n01875717": "Caenolestidae, family Caenolestidae",
  "n01875880": "Caenolestes, genus Caenolestes",
  "n01876034": "opossum rat",
  "n01876180": "Peramelidae, family Peramelidae",
  "n01876326": "bandicoot",
  "n01876535": "Macrotis, genus Macrotis",
  "n01876667": "rabbit-eared bandicoot, rabbit bandicoot, bilby, Macrotis lagotis",
  "n01876843": "Macropodidae, family Macropodidae",
  "n01877134": "kangaroo",
  "n01877407": "Macropus, genus Macropus",
  "n01877606": "giant kangaroo, great grey kangaroo, Macropus giganteus",
  "n01877812": "wallaby, brush kangaroo",
  "n01878061": "common wallaby, Macropus agiles",
  "n01878203": "Lagorchestes, genus Lagorchestes",
  "n01878335": "hare wallaby, kangaroo hare",
  "n01878500": "Onychogalea, genus Onychogalea",
  "n01878639": "nail-tailed wallaby, nail-tailed kangaroo",
  "n01878803": "Petrogale, genus Petrogale",
  "n01878929": "rock wallaby, rock kangaroo",
  "n01879095": "Thylogale, genus Thylogale",
  "n01879217": "pademelon, paddymelon",
  "n01879379": "Dendrolagus, genus Dendrolagus",
  "n01879509": "tree wallaby, tree kangaroo",
  "n01879701": "Hypsiprymnodon, genus Hypsiprymnodon",
  "n01879837": "musk kangaroo, Hypsiprymnodon moschatus",
  "n01879983": "Potoroinae, subfamily Potoroinae",
  "n01880152": "rat kangaroo, kangaroo rat",
  "n01880355": "Potorous, genus Potorous",
  "n01880473": "potoroo",
  "n01880570": "Bettongia, genus Bettongia",
  "n01880716": "bettong",
  "n01880813": "jerboa kangaroo, kangaroo jerboa",
  "n01880937": "Phalangeridae, family Phalangeridae",
  "n01881171": "phalanger, opossum, possum",
  "n01881416": "genus Phalanger",
  "n01881564": "cuscus",
  "n01881717": "Trichosurus, genus Trichosurus",
  "n01881857": "brush-tailed phalanger, Trichosurus vulpecula",
  "n01881991": "Petaurus, genus Petaurus",
  "n01882125": "flying phalanger, flying opossum, flying squirrel",
  "n01882372": "Acrobates, genus Acrobates",
  "n01882508": "flying mouse",
  "n01882607": "Phascolarctos, genus Phascolarctos",
  "n01882714": "koala, koala bear, kangaroo bear, native bear, Phascolarctos cinereus",
  "n01882948": "Vombatidae, family Vombatidae",
  "n01883070": "wombat",
  "n01883212": "Dasyuridae, family Dasyuridae, family Dasyurinae",
  "n01883513": "dasyurid marsupial, dasyurid",
  "n01883762": "Dasyurus, genus Dasyurus",
  "n01883920": "dasyure",
  "n01884104": "eastern dasyure, Dasyurus quoll",
  "n01884203": "native cat, Dasyurus viverrinus",
  "n01884348": "Thylacinus, genus Thylacinus",
  "n01884476": "thylacine, Tasmanian wolf, Tasmanian tiger, Thylacinus cynocephalus",
  "n01884703": "Sarcophilus, genus Sarcophilus",
  "n01884834": "Tasmanian devil, ursine dasyure, Sarcophilus hariisi",
  "n01885032": "Phascogale, genus Phascogale",
  "n01885158": "pouched mouse, marsupial mouse, marsupial rat",
  "n01885367": "Myrmecobius, genus Myrmecobius",
  "n01885498": "numbat, banded anteater, anteater, Myrmecobius fasciatus",
  "n01885724": "Notoryctidae, family Notoryctidae",
  "n01885856": "Notoryctus, genus Notoryctus",
  "n01886045": "pouched mole, marsupial mole, Notoryctus typhlops",
  "n01886220": "Eutheria, subclass Eutheria",
  "n01886756": "placental, placental mammal, eutherian, eutherian mammal",
  "n01887474": "livestock, stock, farm animal",
  "n01887623": "bull",
  "n01887787": "cow",
  "n01887896": "calf",
  "n01888045": "calf",
  "n01888181": "yearling",
  "n01888264": "buck",
  "n01888411": "doe",
  "n01888520": "Insectivora, order Insectivora",
  "n01888821": "Lipotyphla, suborder Lipotyphla",
  "n01888948": "Menotyphla, suborder Menotyphla",
  "n01889074": "insectivore",
  "n01889328": "Talpidae, family Talpidae",
  "n01889520": "mole",
  "n01889740": "Condylura, genus Condylura",
  "n01889849": "starnose mole, star-nosed mole, Condylura cristata",
  "n01890033": "Parascalops, genus Parascalops",
  "n01890144": "brewer's mole, hair-tailed mole, Parascalops breweri",
  "n01890274": "Chrysochloridae, family Chrysochloridae",
  "n01890411": "Chrysochloris, genus Chrysochloris",
  "n01890564": "golden mole",
  "n01890718": "Uropsilus, genus Uropsilus",
  "n01890860": "shrew mole",
  "n01891013": "Asiatic shrew mole, Uropsilus soricipes",
  "n01891145": "Neurotrichus, genus Neurotrichus",
  "n01891274": "American shrew mole, Neurotrichus gibbsii",
  "n01891438": "Soricidae, family Soricidae",
  "n01891633": "shrew, shrewmouse",
  "n01891865": "Sorex, genus Sorex",
  "n01892030": "common shrew, Sorex araneus",
  "n01892145": "masked shrew, Sorex cinereus",
  "n01892271": "Blarina, genus Blarina",
  "n01892385": "short-tailed shrew, Blarina brevicauda",
  "n01892551": "water shrew",
  "n01892744": "American water shrew, Sorex palustris",
  "n01892876": "Neomys, genus Neomys",
  "n01893021": "European water shrew, Neomys fodiens",
  "n01893164": "Mediterranean water shrew, Neomys anomalus",
  "n01893294": "Cryptotis, genus Cryptotis",
  "n01893399": "least shrew, Cryptotis parva",
  "n01893535": "Erinaceidae, family Erinaceidae",
  "n01893666": "Erinaceus, genus Erinaceus",
  "n01893825": "hedgehog, Erinaceus europaeus, Erinaceus europeaeus",
  "n01894040": "Tenrecidae, family Tenrecidae",
  "n01894207": "tenrec, tendrac",
  "n01894381": "genus Tenrec",
  "n01894522": "tailless tenrec, Tenrec ecaudatus",
  "n01894670": "Potamogalidae, family Potamogalidae",
  "n01894803": "genus Potamogale",
  "n01894956": "otter shrew, potamogale, Potamogale velox",
  "n01895128": "chine",
  "n01895219": "saddle",
  "n01895355": "furcula",
  "n01895511": "wishbone, wishing bone",
  "n01895630": "cuticula",
  "n01895735": "hide, pelt, skin",
  "n01895850": "hypodermis",
  "n01896031": "feather, plume, plumage",
  "n01896561": "down, down feather",
  "n01896735": "duck down",
  "n01896844": "eiderdown",
  "n01896960": "goose down",
  "n01897053": "swan's down",
  "n01897146": "plumule",
  "n01897257": "aftershaft",
  "n01897426": "sickle feather",
  "n01897536": "contour feather",
  "n01897667": "bastard wing, alula, spurious wing",
  "n01897851": "marabou",
  "n01897991": "vane, web",
  "n01898185": "barb",
  "n01898328": "web",
  "n01898459": "hackle",
  "n01898593": "saddle hackle, saddle feather",
  "n01898731": "coat, pelage",
  "n01898906": "guard hair",
  "n01899062": "fur",
  "n01899238": "undercoat, underfur",
  "n01899367": "underpart",
  "n01899593": "wool, fleece",
  "n01899746": "mane",
  "n01899894": "encolure",
  "n01900006": "forelock, foretop",
  "n01900150": "hair",
  "n01900341": "cirrus",
  "n01900488": "spine",
  "n01900719": "ray",
  "n01900837": "quill",
  "n01900985": "aculea",
  "n01901091": "aculeus",
  "n01901213": "style",
  "n01901348": "stylet",
  "n01901492": "villus",
  "n01901610": "bristle",
  "n01901694": "chaeta",
  "n01901828": "whisker, vibrissa, sensory hair",
  "n01902022": "seta",
  "n01902132": "pilus",
  "n01902274": "horseback",
  "n01902368": "operculum",
  "n01902568": "protective covering",
  "n01902750": "armor, armour",
  "n01902877": "scale",
  "n01903110": "fish scale",
  "n01903234": "squama",
  "n01903346": "scute",
  "n01903498": "sclerite",
  "n01903631": "clypeus",
  "n01903756": "carapace, shell, cuticle, shield",
  "n01904029": "plastron",
  "n01904182": "shell",
  "n01904461": "valve",
  "n01904583": "valve",
  "n01904699": "test",
  "n01904806": "scallop shell",
  "n01904886": "oyster shell",
  "n01904965": "phragmocone, phragmacone",
  "n01905121": "apodeme",
  "n01905321": "theca",
  "n01905416": "lorica",
  "n01905543": "coelenteron",
  "n01905661": "invertebrate",
  "n01906178": "zoophyte",
  "n01906328": "Parazoa, subkingdom Parazoa",
  "n01906552": "Porifera, phylum Porifera",
  "n01906749": "sponge, poriferan, parazoan",
  "n01907034": "sponge genus",
  "n01907149": "flagellated cell",
  "n01907287": "choanocyte, collar cell",
  "n01907495": "Hyalospongiae, class Hyalospongiae",
  "n01907738": "glass sponge",
  "n01907902": "Euplectella, genus Euplectella",
  "n01908042": "Venus's flower basket",
  "n01908287": "coelenterate family",
  "n01908415": "coelenterate genus",
  "n01908703": "Metazoa, subkingdom Metazoa",
  "n01908958": "metazoan",
  "n01909111": "Cnidaria, phylum Cnidaria, Coelenterata, phylum Coelenterata",
  "n01909422": "coelenterate, cnidarian",
  "n01909788": "planula",
  "n01909906": "polyp",
  "n01910252": "medusa, medusoid, medusan",
  "n01910529": "Scyphozoa, class Scyphozoa",
  "n01910747": "jellyfish",
  "n01910998": "Aegina",
  "n01911063": "scyphozoan",
  "n01911271": "Chrysaora, genus Chrysaora",
  "n01911403": "Chrysaora quinquecirrha",
  "n01911511": "Hydrozoa, class Hydrozoa",
  "n01911839": "hydrozoan, hydroid",
  "n01912050": "genus Hydra",
  "n01912152": "hydra",
  "n01912272": "Siphonophora, order Siphonophora",
  "n01912454": "siphonophore",
  "n01912688": "genus Nanomia",
  "n01912809": "nanomia",
  "n01913035": "Physalia, genus Physalia",
  "n01913166": "Portuguese man-of-war, man-of-war, jellyfish",
  "n01913346": "praya",
  "n01913440": "apolemia",
  "n01913533": "Sertularia, genus Sertularia",
  "n01913691": "sertularian",
  "n01913838": "Anthozoa, class Anthozoa, Actinozoa, class Actinozoa",
  "n01914163": "anthozoan, actinozoan",
  "n01914415": "Actiniaria, order Actiniaria, Actinaria, order Actinaria",
  "n01914609": "sea anemone, anemone",
  "n01914830": "actinia, actinian, actiniarian",
  "n01914961": "Actinia, genus Actinia",
  "n01915093": "Alcyonaria, order Alcyonaria",
  "n01915290": "Alcyonacea, suborder Alcyonacea",
  "n01915414": "Pennatulidae, family Pennatulidae",
  "n01915541": "Pennatula, genus Pennatula",
  "n01915700": "sea pen",
  "n01915811": "coral",
  "n01916010": "Gorgonacea, suborder Gorgonacea, Gorgoniacea, suborder Gorgoniacea",
  "n01916187": "gorgonian, gorgonian coral",
  "n01916388": "sea feather",
  "n01916481": "sea fan",
  "n01916588": "red coral",
  "n01916738": "Madreporaria, order Madreporaria",
  "n01916925": "stony coral, madrepore, madriporian coral",
  "n01917167": "Maeandra, genus Maeandra",
  "n01917289": "brain coral",
  "n01917434": "Acropora, genus Acropora",
  "n01917611": "staghorn coral, stag's-horn coral",
  "n01917751": "Fungia, genus Fungia",
  "n01917882": "mushroom coral",
  "n01918010": "ctenophore family",
  "n01918152": "ctenophore genus",
  "n01918310": "Ctenophora, phylum Ctenophora",
  "n01918585": "ctene, comb-plate",
  "n01918744": "ctenophore, comb jelly",
  "n01919099": "Nuda, class Nuda",
  "n01919258": "genus Beroe",
  "n01919385": "beroe",
  "n01919504": "Tentaculata, class Tentaculata",
  "n01919714": "Cydippida, order Cydippida, Cydippidea, order Cydippidea, Cydippea, order Cydippea",
  "n01919912": "Platyctenea, order Platyctenea",
  "n01920051": "platyctenean",
  "n01920178": "Pleurobrachiidae, family Pleurobrachiidae",
  "n01920302": "Pleurobrachia, genus Pleurobrachia",
  "n01920438": "sea gooseberry",
  "n01920582": "Cestida, order Cestida",
  "n01920735": "Cestidae, family Cestidae",
  "n01920939": "Cestum, genus Cestum",
  "n01921059": "Venus's girdle, Cestum veneris",
  "n01921216": "Lobata, order Lobata",
  "n01921440": "comb",
  "n01921559": "worm family",
  "n01921887": "worm genus",
  "n01922303": "worm",
  "n01922717": "helminth, parasitic worm",
  "n01922948": "woodworm",
  "n01923025": "woodborer, borer",
  "n01923171": "Acanthocephala, phylum Acanthocephala",
  "n01923404": "acanthocephalan, spiny-headed worm",
  "n01923637": "Chaetognatha, phylum Chaetognatha",
  "n01923890": "arrowworm, chaetognath",
  "n01924151": "genus Sagitta",
  "n01924309": "sagitta",
  "n01924416": "genus Spadella",
  "n01924590": "Platyhelminthes, phylum Platyhelminthes",
  "n01924800": "bladder worm",
  "n01924916": "flatworm, platyhelminth",
  "n01925133": "Turbellaria, class Turbellaria",
  "n01925270": "planarian, planaria",
  "n01925469": "Trematoda, class Trematoda",
  "n01925695": "fluke, trematode, trematode worm",
  "n01925916": "cercaria",
  "n01926090": "Fasciolidae, family Fasciolidae",
  "n01926247": "Fasciola, genus Fasciola",
  "n01926379": "liver fluke, Fasciola hepatica",
  "n01926549": "Fasciolopsis, genus Fasciolopsis",
  "n01926689": "Fasciolopsis buski",
  "n01926840": "Schistosomatidae, family Schistosomatidae",
  "n01926988": "Schistosoma, genus Schistosoma",
  "n01927159": "schistosome, blood fluke",
  "n01927301": "Cestoda, class Cestoda",
  "n01927456": "tapeworm, cestode",
  "n01927665": "Taeniidae, family Taeniidae",
  "n01927816": "genus Echinococcus",
  "n01927928": "echinococcus",
  "n01928073": "genus Taenia",
  "n01928215": "taenia",
  "n01928360": "Nemertea, phylum Nemertea, Nemertina, phylum Nemertina",
  "n01928517": "ribbon worm, nemertean, nemertine, proboscis worm",
  "n01928737": "Pogonophora, phylum Pogonophora",
  "n01928865": "beard worm, pogonophoran",
  "n01929047": "Rotifera, phylum Rotifera",
  "n01929186": "rotifer",
  "n01929396": "Nematoda, phylum Nematoda, Aschelminthes, phylum Aschelminthes",
  "n01929788": "Aphasmidia, class Aphasmidia",
  "n01929951": "Phasmidia, class Phasmidia",
  "n01930112": "nematode, nematode worm, roundworm",
  "n01930485": "Ascaridae, family Ascaridae",
  "n01930672": "Ascaris, genus Ascaris",
  "n01930852": "common roundworm, Ascaris lumbricoides",
  "n01930995": "Ascaridia, genus Ascaridia",
  "n01931140": "chicken roundworm, Ascaridia galli",
  "n01931277": "Oxyuridae, family Oxyuridae",
  "n01931398": "Enterobius, genus Enterobius",
  "n01931520": "pinworm, threadworm, Enterobius vermicularis",
  "n01931714": "eelworm",
  "n01931845": "Cephalobidae, family Cephalobidae",
  "n01931984": "Anguillula, genus Anguillula, Turbatrix, genus Turbatrix",
  "n01932151": "vinegar eel, vinegar worm, Anguillula aceti, Turbatrix aceti",
  "n01932358": "Tylenchidae, family Tylenchidae",
  "n01932495": "Tylenchus, genus Tylenchus",
  "n01932643": "wheatworm, wheat eel, wheat eelworm, Tylenchus tritici",
  "n01932800": "Ancylostomatidae, family Ancylostomatidae",
  "n01932936": "trichina, Trichinella spiralis",
  "n01933151": "hookworm",
  "n01933342": "Filariidae, family Filariidae",
  "n01933478": "filaria",
  "n01933686": "Dracunculidae, family Dracunculidae",
  "n01933834": "Dracunculus, genus Dracunculus",
  "n01933988": "Guinea worm, Dracunculus medinensis",
  "n01934207": "Annelida, phylum Annelida",
  "n01934440": "annelid, annelid worm, segmented worm",
  "n01934705": "Archiannelida, class Archiannelida",
  "n01934844": "archiannelid",
  "n01935012": "Oligochaeta, class Oligochaeta",
  "n01935176": "oligochaete, oligochaete worm",
  "n01935395": "earthworm, angleworm, fishworm, fishing worm, wiggler, nightwalker, nightcrawler, crawler, dew worm, red worm",
  "n01935743": "Branchiobdellidae, family Branchiobdellidae",
  "n01935997": "Branchiobdella, genus Branchiobdella",
  "n01936219": "Polychaeta, class Polychaeta",
  "n01936391": "polychaete, polychete, polychaete worm, polychete worm",
  "n01936671": "lugworm, lug, lobworm",
  "n01936858": "sea mouse",
  "n01937015": "Terebellidae, family Terebellidae",
  "n01937234": "Terebella, genus Terebella",
  "n01937422": "Polycirrus, genus Polycirrus",
  "n01937579": "bloodworm",
  "n01937719": "Hirudinea, class Hirudinea",
  "n01937909": "leech, bloodsucker, hirudinean",
  "n01938155": "Hirudinidae, family Hirudinidae",
  "n01938312": "Hirudo, genus Hirudo",
  "n01938454": "medicinal leech, Hirudo medicinalis",
  "n01938618": "Haemopis, genus Haemopis",
  "n01938735": "horseleech",
  "n01938850": "mollusk family",
  "n01939598": "mollusk genus",
  "n01940488": "Mollusca, phylum Mollusca",
  "n01940736": "mollusk, mollusc, shellfish",
  "n01941030": "Scaphopoda, class Scaphopoda",
  "n01941223": "scaphopod",
  "n01941340": "tooth shell, tusk shell",
  "n01941538": "lip",
  "n01941670": "Gastropoda, class Gastropoda, Gasteropoda, class Gasteropoda",
  "n01942177": "gastropod, univalve",
  "n01942601": "Haliotidae, family Haliotidae",
  "n01942724": "Haliotis, genus Haliotis",
  "n01942869": "abalone, ear-shell",
  "n01943087": "ormer, sea-ear, Haliotis tuberculata",
  "n01943213": "Strombidae, family Strombidae",
  "n01943367": "Lambis, genus Lambis",
  "n01943541": "scorpion shell",
  "n01943754": "Strombus, genus Strombus",
  "n01943899": "conch",
  "n01944118": "giant conch, Strombus gigas",
  "n01944217": "Helicidae, family Helicidae",
  "n01944390": "snail",
  "n01944617": "Helix, genus Helix",
  "n01944812": "edible snail, Helix pomatia",
  "n01944955": "garden snail",
  "n01945143": "brown snail, Helix aspersa",
  "n01945340": "Helix hortensis",
  "n01945443": "Limacidae, family Limacidae",
  "n01945580": "Limax, genus Limax",
  "n01945685": "slug",
  "n01945845": "seasnail",
  "n01946118": "Neritidae, family Neritidae",
  "n01946277": "neritid, neritid gastropod",
  "n01946487": "genus Nerita",
  "n01946630": "nerita",
  "n01946827": "bleeding tooth, Nerita peloronta",
  "n01947020": "genus Neritina",
  "n01947139": "neritina",
  "n01947275": "Buccinidae, family Buccinidae",
  "n01947396": "whelk",
  "n01947613": "Cymatiidae, family Cymatiidae",
  "n01947735": "triton",
  "n01947874": "Naticidae, family Naticidae",
  "n01947997": "moon shell, moonshell",
  "n01948154": "Littorinidae, family Littorinidae",
  "n01948284": "Littorina, genus Littorina",
  "n01948446": "periwinkle, winkle",
  "n01948573": "limpet",
  "n01948788": "Patellidae, family Patellidae",
  "n01948917": "Patella, genus Patella",
  "n01949085": "common limpet, Patella vulgata",
  "n01949195": "Fissurellidae, family Fissurellidae",
  "n01949330": "Fissurella, genus Fissurella",
  "n01949499": "keyhole limpet, Fissurella apertura, Diodora apertura",
  "n01949684": "Ancylidae, family Ancylidae",
  "n01949817": "Ancylus, genus Ancylus",
  "n01949973": "river limpet, freshwater limpet, Ancylus fluviatilis",
  "n01950195": "Opisthobranchia, subclass Opisthobranchia",
  "n01950457": "Nudibranchia, order Nudibranchia",
  "n01950731": "sea slug, nudibranch",
  "n01950952": "Aplysiidae, family Aplysiidae, Tethyidae, family Tethyidae",
  "n01951107": "Aplysia, genus Aplysia, Tethys, genus Tethus",
  "n01951274": "sea hare, Aplysia punctata",
  "n01951472": "Hermissenda, genus Hermissenda",
  "n01951613": "Hermissenda crassicornis",
  "n01951721": "Akeridae, family Akeridae",
  "n01951845": "Haminoea, genus Haminoea",
  "n01952029": "bubble shell",
  "n01952162": "Pulmonata, order Pulmonata",
  "n01952429": "Physidae, family Physidae",
  "n01952557": "genus Physa",
  "n01952712": "physa",
  "n01952812": "Pectinibranchia, order Pectinibranchia",
  "n01953032": "Cypraeidae, family Cypraeidae",
  "n01953197": "Cypraea, genus Cypraea",
  "n01953361": "cowrie, cowry",
  "n01953594": "money cowrie, Cypraea moneta",
  "n01953762": "tiger cowrie, Cypraea tigris",
  "n01953877": "ctenidium",
  "n01954004": "ceras",
  "n01954202": "Amphineura, subclass Amphineura",
  "n01954340": "Solenogastres, order Solenogastres, Aplacophora, order Aplacophora",
  "n01954516": "solenogaster, aplacophoran",
  "n01954729": "Polyplacophora, class Polyplacophora",
  "n01954962": "genus Chiton",
  "n01955084": "chiton, coat-of-mail shell, sea cradle, polyplacophore",
  "n01955318": "byssus, beard",
  "n01955463": "Bivalvia, class Bivalvia, Lamellibranchia, class Lamellibranchia, class Pelecypoda",
  "n01955933": "bivalve, pelecypod, lamellibranch",
  "n01956344": "spat",
  "n01956481": "clam",
  "n01956764": "seashell",
  "n01956849": "clamshell",
  "n01956924": "Myaceae, order Myaceae",
  "n01957075": "Myacidae, family Myacidae",
  "n01957202": "Mya, genus Mya",
  "n01957335": "soft-shell clam, steamer, steamer clam, long-neck clam, Mya arenaria",
  "n01957591": "Veneridae, family Veneridae",
  "n01957739": "Venus, genus Venus",
  "n01957923": "Mercenaria, genus Mercenaria",
  "n01958038": "quahog, quahaug, hard-shell clam, hard clam, round clam, Venus mercenaria, Mercenaria mercenaria",
  "n01958346": "littleneck, littleneck clam",
  "n01958435": "cherrystone, cherrystone clam",
  "n01958531": "geoduck",
  "n01958790": "Solenidae, family Solenidae",
  "n01958914": "Ensis, genus Ensis",
  "n01959029": "razor clam, jackknife clam, knife-handle",
  "n01959187": "Tridacnidae, family Tridacnidae",
  "n01959333": "Tridacna, genus Tridacna",
  "n01959492": "giant clam, Tridacna gigas",
  "n01959668": "Cardiidae, family Cardiidae",
  "n01959834": "Cardium, genus Cardium",
  "n01959985": "cockle",
  "n01960177": "edible cockle, Cardium edule",
  "n01960301": "Ostreidae, family Ostreidae",
  "n01960459": "oyster",
  "n01960787": "seed oyster",
  "n01960900": "Ostrea, genus Ostrea",
  "n01961059": "bluepoint, blue point",
  "n01961234": "Japanese oyster, Ostrea gigas",
  "n01961468": "Crassostrea, genus Crassostrea",
  "n01961600": "Virginia oyster",
  "n01961736": "Pteriidae, family Pteriidae",
  "n01961862": "Pinctada, genus Pinctada",
  "n01961985": "pearl oyster, Pinctada margaritifera",
  "n01962223": "Anomiidae, family Anomiidae",
  "n01962350": "Anomia, genus Anomia",
  "n01962506": "saddle oyster, Anomia ephippium",
  "n01962662": "Placuna, genus Placuna",
  "n01962788": "window oyster, windowpane oyster, capiz, Placuna placenta",
  "n01963017": "Arcidae, family Arcidae",
  "n01963136": "Arca, genus Arca",
  "n01963317": "ark shell",
  "n01963479": "blood clam",
  "n01963571": "mussel",
  "n01963730": "Mytilidae, family Mytilidae",
  "n01963876": "Mytilus, genus Mytilus",
  "n01964049": "marine mussel, mytilid",
  "n01964271": "edible mussel, Mytilus edulis",
  "n01964441": "freshwater mussel, freshwater clam",
  "n01964636": "Unionidae, family Unionidae",
  "n01964821": "Unio, genus Unio",
  "n01964957": "pearly-shelled mussel",
  "n01965111": "Anodonta, genus Anodonta",
  "n01965252": "thin-shelled mussel",
  "n01965404": "Dreissena, genus Dreissena",
  "n01965529": "zebra mussel, Dreissena polymorpha",
  "n01965747": "Pectinidae, family Pectinidae",
  "n01965889": "scallop, scollop, escallop",
  "n01966204": "genus Pecten",
  "n01966377": "bay scallop, Pecten irradians",
  "n01966586": "sea scallop, giant scallop, Pecten magellanicus",
  "n01966797": "Teredinidae, family Teredinidae",
  "n01966961": "genus Teredo",
  "n01967094": "shipworm, teredinid",
  "n01967308": "teredo",
  "n01967396": "Bankia, genus Bankia",
  "n01967517": "giant northwest shipworm, Bankia setaceae",
  "n01967677": "Pholadidae, family Pholadidae",
  "n01967812": "Pholas, genus Pholas",
  "n01967963": "piddock",
  "n01968115": "Cephalopoda, class Cephalopoda",
  "n01968315": "cephalopod, cephalopod mollusk",
  "n01968591": "Nautilidae, family Nautilidae",
  "n01968732": "genus Nautilus",
  "n01968897": "chambered nautilus, pearly nautilus, nautilus",
  "n01969103": "Dibranchiata, subclass Dibranchiata, Dibranchia, subclass Dibranchia",
  "n01969429": "dibranchiate, dibranchiate mollusk, dibranch",
  "n01969550": "Octopoda, order Octopoda",
  "n01969726": "octopod",
  "n01969893": "Octopodidae, family Octopodidae",
  "n01970030": "genus Octopus",
  "n01970164": "octopus, devilfish",
  "n01970342": "Argonautidae, family Argonautidae",
  "n01970502": "Argonauta, genus Argonauta",
  "n01970667": "paper nautilus, nautilus, Argonaut, Argonauta argo",
  "n01970866": "Decapoda, order Decapoda",
  "n01971094": "decapod",
  "n01971280": "squid",
  "n01971517": "genus Loligo",
  "n01971620": "loligo",
  "n01971728": "genus Ommastrephes",
  "n01971850": "ommastrephes",
  "n01972017": "genus Architeuthis",
  "n01972131": "architeuthis, giant squid",
  "n01972283": "Sepiidae, family Sepiidae",
  "n01972411": "Sepia, genus Sepia",
  "n01972541": "cuttlefish, cuttle",
  "n01972733": "Spirulidae, family Spirulidae",
  "n01972947": "genus Spirula",
  "n01973148": "spirula, Spirula peronii",
  "n01973375": "Belemnoidea, order Belemnoidea",
  "n01973566": "Belemnitidae, family Belemnitidae",
  "n01973723": "belemnite",
  "n01974055": "craw, crop",
  "n01974229": "gizzard, ventriculus, gastric mill",
  "n01974399": "Crustacea, class Crustacea",
  "n01974773": "crustacean",
  "n01975117": "green gland",
  "n01975312": "Malacostraca, subclass Malacostraca",
  "n01975687": "malacostracan crustacean",
  "n01975880": "Decapoda, order Decapoda",
  "n01976146": "decapod crustacean, decapod",
  "n01976477": "Brachyura, suborder Brachyura",
  "n01976868": "brachyuran",
  "n01976957": "crab",
  "n01977366": "Menippe, genus Menippe",
  "n01977485": "stone crab, Menippe mercenaria",
  "n01977684": "Cancridae, family Cancridae",
  "n01977832": "Cancer, genus Cancer",
  "n01978010": "hard-shell crab",
  "n01978136": "soft-shell crab, soft-shelled crab",
  "n01978287": "Dungeness crab, Cancer magister",
  "n01978455": "rock crab, Cancer irroratus",
  "n01978587": "Jonah crab, Cancer borealis",
  "n01978744": "Portunidae, family Portunidae",
  "n01978930": "swimming crab",
  "n01979124": "Portunus, genus Portunus",
  "n01979269": "English lady crab, Portunus puber",
  "n01979395": "Ovalipes, genus Ovalipes",
  "n01979526": "American lady crab, lady crab, calico crab, Ovalipes ocellatus",
  "n01979738": "Callinectes, genus Callinectes",
  "n01979874": "blue crab, Callinectes sapidus",
  "n01980053": "Uca, genus Uca",
  "n01980166": "fiddler crab",
  "n01980328": "Pinnotheridae, family Pinnotheridae",
  "n01980471": "Pinnotheres, genus Pinnotheres",
  "n01980655": "pea crab",
  "n01980830": "oyster crab, Pinnotheres ostreum",
  "n01980993": "Lithodidae, family Lithodidae",
  "n01981137": "Paralithodes, genus Paralithodes",
  "n01981276": "king crab, Alaska crab, Alaskan king crab, Alaska king crab, Paralithodes camtschatica",
  "n01981543": "Majidae, family Majidae",
  "n01981702": "spider crab",
  "n01981884": "Maja, genus Maja, Maia, genus Maia",
  "n01982068": "European spider crab, king crab, Maja squinado",
  "n01982211": "Macrocheira, genus Macrocheira",
  "n01982347": "giant crab, Macrocheira kaempferi",
  "n01982482": "Reptantia, suborder Reptantia",
  "n01982650": "lobster",
  "n01982895": "Homaridae, family Homaridae",
  "n01983048": "true lobster",
  "n01983277": "Homarus, genus Homarus",
  "n01983481": "American lobster, Northern lobster, Maine lobster, Homarus americanus",
  "n01983674": "European lobster, Homarus vulgaris",
  "n01983829": "Cape lobster, Homarus capensis",
  "n01983958": "Nephropsidae, family Nephropsidae",
  "n01984131": "Nephrops, genus Nephrops",
  "n01984245": "Norway lobster, Nephrops norvegicus",
  "n01984416": "Palinuridae, family Palinuridae",
  "n01984547": "Palinurus, genus Palinurus",
  "n01984695": "spiny lobster, langouste, rock lobster, crawfish, crayfish, sea crawfish",
  "n01984958": "Astacidae, family Astacidae, Astacura",
  "n01985128": "crayfish, crawfish, crawdad, crawdaddy",
  "n01985331": "Astacus, genus Astacus",
  "n01985493": "Old World crayfish, ecrevisse",
  "n01985667": "Cambarus, genus Cambarus",
  "n01985797": "American crayfish",
  "n01985947": "Paguridae, family Paguridae",
  "n01986072": "Pagurus, genus Pagurus",
  "n01986214": "hermit crab",
  "n01986367": "Natantia, suborder Natantia",
  "n01986538": "Crangonidae, family Crangonidae",
  "n01986681": "Crangon, genus Crangon",
  "n01986806": "shrimp",
  "n01987076": "snapping shrimp, pistol shrimp",
  "n01987228": "Palaemonidae, family Palaemonidae",
  "n01987353": "Palaemon, genus Palaemon",
  "n01987545": "prawn",
  "n01987727": "long-clawed prawn, river prawn, Palaemon australis",
  "n01987938": "Peneidae, family Peneidae",
  "n01988064": "Peneus, genus Peneus",
  "n01988203": "tropical prawn",
  "n01988319": "Schizopoda",
  "n01988481": "Euphausiacea, order Euphausiacea",
  "n01988701": "krill",
  "n01988869": "Euphausia pacifica",
  "n01988971": "Mysidacea, order Mysidacea",
  "n01989097": "Mysidae, family Mysidae",
  "n01989254": "Mysis, genus Mysis",
  "n01989390": "Praunus, genus Praunus",
  "n01989516": "opossum shrimp",
  "n01989701": "Stomatopoda, order Stomatopoda",
  "n01989869": "stomatopod, stomatopod crustacean",
  "n01990007": "mantis shrimp, mantis crab",
  "n01990186": "Squillidae, family Squillidae",
  "n01990383": "genus Squilla",
  "n01990516": "squilla, mantis prawn",
  "n01990627": "Isopoda, order Isopoda",
  "n01990800": "isopod",
  "n01991028": "woodlouse, slater",
  "n01991233": "Armadillidiidae, family Armadillidiidae",
  "n01991367": "Armadillidium, genus Armadillidium",
  "n01991520": "pill bug",
  "n01991676": "Oniscidae, family Oniscidae",
  "n01991808": "Oniscus, genus Oniscus",
  "n01991982": "Porcellionidae, family Porcellionidae",
  "n01992113": "Porcellio, genus Porcellio",
  "n01992262": "sow bug",
  "n01992423": "sea louse, sea slater",
  "n01992516": "Amphipoda, order Amphipoda",
  "n01992773": "amphipod",
  "n01992935": "Orchestiidae, family Orchestiidae",
  "n01993065": "Orchestia, genus Orchestia",
  "n01993214": "beach flea, sand hopper, sandhopper, sand flea",
  "n01993400": "Caprella, genus Caprella",
  "n01993525": "skeleton shrimp",
  "n01993714": "Cyamus, genus Cyamus",
  "n01993830": "whale louse",
  "n01993949": "Entomostraca, subclass Entomostraca",
  "n01994176": "Branchiopoda, subclass Branchiopoda",
  "n01994492": "branchiopod crustacean, branchiopod, branchiopodan",
  "n01994801": "genus Daphnia",
  "n01994910": "daphnia, water flea",
  "n01995137": "Anostraca, order Anostraca",
  "n01995323": "Artemia, genus Artemia, Chirocephalus, genus Chirocephalus",
  "n01995514": "fairy shrimp",
  "n01995686": "brine shrimp, Artemia salina",
  "n01995803": "Notostraca, order Notostraca",
  "n01995975": "Triopidae, family Triopidae",
  "n01996091": "Triops, genus Triops",
  "n01996280": "tadpole shrimp",
  "n01996392": "Copepoda, subclass Copepoda",
  "n01996585": "copepod, copepod crustacean",
  "n01996895": "brit, britt",
  "n01997002": "genus Cyclops",
  "n01997119": "cyclops, water flea",
  "n01997436": "Branchiura, order Branchiura",
  "n01997605": "fish louse",
  "n01997698": "Ostracoda, subclass Ostracoda",
  "n01997825": "seed shrimp, mussel shrimp, ostracod",
  "n01998019": "Cirripedia, subclass Cirripedia",
  "n01998183": "barnacle, cirriped, cirripede",
  "n01998467": "Balanidae, family Balanidae",
  "n01998599": "Balanus, genus Balanus",
  "n01998741": "acorn barnacle, rock barnacle, Balanus balanoides",
  "n01998920": "Lepadidae, family Lepadidae",
  "n01999048": "Lepas, genus Lepas",
  "n01999186": "goose barnacle, gooseneck barnacle, Lepas fascicularis",
  "n01999374": "Onychophora, class Onychophora",
  "n01999767": "onychophoran, velvet worm, peripatus",
  "n02000036": "Peripatidae, family Peripatidae",
  "n02000195": "genus Peripatus",
  "n02000354": "Plicatoperipatus, genus Plicatoperipatus",
  "n02000502": "Plicatoperipatus jamaicensis",
  "n02000618": "Peripatopsidae, family Peripatopsidae",
  "n02000764": "Peripatopsis, genus Peripatopsis",
  "n02000954": "wading bird, wader",
  "n02001428": "Ciconiiformes, order Ciconiiformes",
  "n02001821": "Ciconiidae, family Ciconiidae",
  "n02002075": "stork",
  "n02002384": "Ciconia, genus Ciconia",
  "n02002556": "white stork, Ciconia ciconia",
  "n02002724": "black stork, Ciconia nigra",
  "n02002875": "Leptoptilus, genus Leptoptilus",
  "n02003037": "adjutant bird, adjutant, adjutant stork, Leptoptilus dubius",
  "n02003204": "marabou, marabout, marabou stork, Leptoptilus crumeniferus",
  "n02003456": "Anastomus, genus Anastomus",
  "n02003577": "openbill",
  "n02003735": "genus Jabiru",
  "n02003839": "jabiru, Jabiru mycteria",
  "n02003994": "Ephippiorhynchus, genus Ephippiorhynchus",
  "n02004131": "saddlebill, jabiru, Ephippiorhynchus senegalensis",
  "n02004343": "Xenorhyncus, genus Xenorhyncus",
  "n02004492": "policeman bird, black-necked stork, jabiru, Xenorhyncus asiaticus",
  "n02004661": "Mycteria, genus Mycteria",
  "n02004855": "wood ibis, wood stork, flinthead, Mycteria americana",
  "n02005102": "Balaenicipitidae, family Balaenicipitidae",
  "n02005238": "Balaeniceps, genus Balaeniceps",
  "n02005399": "shoebill, shoebird, Balaeniceps rex",
  "n02005598": "Threskiornithidae, family Threskiornithidae, family Ibidiidae",
  "n02005790": "ibis",
  "n02005962": "genus Ibis",
  "n02006063": "wood ibis, wood stork, Ibis ibis",
  "n02006211": "Threskiornis, genus Threskiornis",
  "n02006364": "sacred ibis, Threskiornis aethiopica",
  "n02006510": "Plataleidae, family Plataleidae",
  "n02006656": "spoonbill",
  "n02006827": "Platalea, genus Platalea",
  "n02006985": "common spoonbill, Platalea leucorodia",
  "n02007161": "Ajaia, genus Ajaia",
  "n02007284": "roseate spoonbill, Ajaia ajaja",
  "n02007422": "Phoenicopteridae, family Phoenicopteridae",
  "n02007558": "flamingo",
  "n02007721": "Ardeidae, family Ardeidae",
  "n02008041": "heron",
  "n02008316": "Ardea, genus Ardea",
  "n02008497": "great blue heron, Ardea herodius",
  "n02008643": "great white heron, Ardea occidentalis",
  "n02008796": "egret",
  "n02009015": "Egretta, genus Egretta",
  "n02009229": "little blue heron, Egretta caerulea",
  "n02009380": "snowy egret, snowy heron, Egretta thula",
  "n02009508": "little egret, Egretta garzetta",
  "n02009620": "Casmerodius, genus Casmerodius",
  "n02009750": "great white heron, Casmerodius albus",
  "n02009912": "American egret, great white heron, Egretta albus",
  "n02010144": "Bubulcus, genus Bubulcus",
  "n02010272": "cattle egret, Bubulcus ibis",
  "n02010453": "night heron, night raven",
  "n02010592": "Nycticorax, genus Nycticorax",
  "n02010728": "black-crowned night heron, Nycticorax nycticorax",
  "n02010881": "Nyctanassa, genus Nyctanassa",
  "n02011016": "yellow-crowned night heron, Nyctanassa violacea",
  "n02011156": "Cochlearius, genus Cochlearius",
  "n02011281": "boatbill, boat-billed heron, broadbill, Cochlearius cochlearius",
  "n02011460": "bittern",
  "n02011668": "Botaurus, genus Botaurus",
  "n02011805": "American bittern, stake driver, Botaurus lentiginosus",
  "n02011943": "European bittern, Botaurus stellaris",
  "n02012063": "Ixobrychus, genus Ixobrychus",
  "n02012185": "least bittern, Ixobrychus exilis",
  "n02012306": "Gruiformes, order Gruiformes",
  "n02012715": "Gruidae, family Gruidae",
  "n02012849": "crane",
  "n02013034": "Grus, genus Grus",
  "n02013177": "whooping crane, whooper, Grus americana",
  "n02013362": "Aramus, genus Aramus",
  "n02013567": "courlan, Aramus guarauna",
  "n02013706": "limpkin, Aramus pictus",
  "n02013889": "Cariamidae, family Cariamidae",
  "n02014061": "Cariama, genus Cariama",
  "n02014237": "crested cariama, seriema, Cariama cristata",
  "n02014406": "genus Chunga",
  "n02014524": "chunga, seriema, Chunga burmeisteri",
  "n02014646": "Rallidae, family Rallidae",
  "n02014941": "rail",
  "n02015221": "Gallirallus, genus Gallirallus",
  "n02015357": "weka, maori hen, wood hen",
  "n02015554": "crake",
  "n02015685": "Crex, genus Crex",
  "n02015797": "corncrake, land rail, Crex crex",
  "n02015944": "Porzana, genus Porzana",
  "n02016066": "spotted crake, Porzana porzana",
  "n02016198": "Gallinula, genus Gallinula",
  "n02016358": "gallinule, marsh hen, water hen, swamphen",
  "n02016659": "Florida gallinule, Gallinula chloropus cachinnans",
  "n02016816": "moorhen, Gallinula chloropus",
  "n02016956": "purple gallinule",
  "n02017093": "Porphyrio, genus Porphyrio",
  "n02017213": "European gallinule, Porphyrio porphyrio",
  "n02017335": "Porphyrula, genus Porphyrula",
  "n02017475": "American gallinule, Porphyrula martinica",
  "n02017607": "genus Notornis",
  "n02017725": "notornis, takahe, Notornis mantelli",
  "n02017878": "Fulica, genus Fulica",
  "n02018027": "coot",
  "n02018207": "American coot, marsh hen, mud hen, water hen, Fulica americana",
  "n02018368": "Old World coot, Fulica atra",
  "n02018485": "Otides, suborder Otides",
  "n02018638": "Otididae, family Otididae",
  "n02018795": "bustard",
  "n02019044": "Otis, genus Otis",
  "n02019190": "great bustard, Otis tarda",
  "n02019308": "Choriotis, genus Choriotis",
  "n02019438": "plain turkey, Choriotis australis",
  "n02019566": "Turnicidae, family Turnicidae",
  "n02019762": "Turnix, genus Turnix",
  "n02019929": "button quail, button-quail, bustard quail, hemipode",
  "n02020219": "striped button quail, Turnix sylvatica",
  "n02020345": "ortygan",
  "n02020450": "Pedionomus, genus Pedionomus",
  "n02020578": "plain wanderer, Pedionomus torquatus",
  "n02020777": "Psophiidae, family Psophiidae",
  "n02020902": "Psophia, genus Psophia",
  "n02021050": "trumpeter",
  "n02021281": "Brazilian trumpeter, Psophia crepitans",
  "n02021438": "Charadriiformes, order Charadriiformes",
  "n02021795": "seabird, sea bird, seafowl",
  "n02022135": "Charadrii, suborder Charadrii",
  "n02022498": "Limicolae, suborder Limicolae",
  "n02022684": "shorebird, shore bird, limicoline bird",
  "n02023133": "Charadriidae, family Charadriidae",
  "n02023341": "plover",
  "n02023664": "Charadrius, genus Charadrius",
  "n02023855": "piping plover, Charadrius melodus",
  "n02023992": "killdeer, kildeer, killdeer plover, Charadrius vociferus",
  "n02024185": "dotterel, dotrel, Charadrius morinellus, Eudromias morinellus",
  "n02024353": "Pluvialis, genus Pluvialis",
  "n02024479": "golden plover",
  "n02024636": "Vanellus, genus Vanellus",
  "n02024763": "lapwing, green plover, peewit, pewit",
  "n02024923": "Arenaria, genus Arenaria",
  "n02025043": "turnstone",
  "n02025239": "ruddy turnstone, Arenaria interpres",
  "n02025389": "black turnstone, Arenaria-Melanocephala",
  "n02025530": "Scolopacidae, family Scolopacidae",
  "n02026059": "sandpiper",
  "n02026498": "Aphriza, genus Aphriza",
  "n02026629": "surfbird, Aphriza virgata",
  "n02026798": "Actitis, genus Actitis",
  "n02026948": "European sandpiper, Actitis hypoleucos",
  "n02027075": "spotted sandpiper, Actitis macularia",
  "n02027209": "Erolia, genus Erolia",
  "n02027357": "least sandpiper, stint, Erolia minutilla",
  "n02027492": "red-backed sandpiper, dunlin, Erolia alpina",
  "n02027730": "Tringa, genus Tringa",
  "n02027897": "greenshank, Tringa nebularia",
  "n02028035": "redshank, Tringa totanus",
  "n02028175": "yellowlegs",
  "n02028342": "greater yellowlegs, Tringa melanoleuca",
  "n02028451": "lesser yellowlegs, Tringa flavipes",
  "n02028556": "Calidris, genus Calidris",
  "n02028727": "pectoral sandpiper, jacksnipe, Calidris melanotos",
  "n02028900": "knot, greyback, grayback, Calidris canutus",
  "n02029087": "curlew sandpiper, Calidris Ferruginea",
  "n02029243": "Crocethia, genus Crocethia",
  "n02029378": "sanderling, Crocethia alba",
  "n02029571": "Bartramia, genus Bartramia",
  "n02029706": "upland sandpiper, upland plover, Bartramian sandpiper, Bartramia longicauda",
  "n02029914": "Philomachus, genus Philomachus",
  "n02030035": "ruff, Philomachus pugnax",
  "n02030224": "reeve",
  "n02030287": "tattler",
  "n02030442": "Heteroscelus, genus Heteroscelus",
  "n02030568": "Polynesian tattler, Heteroscelus incanus",
  "n02030709": "Catoptrophorus, genus Catoptrophorus",
  "n02030837": "willet, Catoptrophorus semipalmatus",
  "n02030996": "woodcock",
  "n02031143": "Scolopax, genus Scolopax",
  "n02031298": "Eurasian woodcock, Scolopax rusticola",
  "n02031455": "Philohela, genus Philohela",
  "n02031585": "American woodcock, woodcock snipe, Philohela minor",
  "n02031752": "Gallinago, genus Gallinago, Capella, genus Capella",
  "n02031934": "snipe",
  "n02032222": "whole snipe, Gallinago gallinago",
  "n02032355": "Wilson's snipe, Gallinago gallinago delicata",
  "n02032480": "great snipe, woodcock snipe, Gallinago media",
  "n02032646": "Limnocryptes, genus Limnocryptes",
  "n02032769": "jacksnipe, half snipe, Limnocryptes minima",
  "n02032915": "Limnodromus, genus Limnodromus",
  "n02033041": "dowitcher",
  "n02033208": "greyback, grayback, Limnodromus griseus",
  "n02033324": "red-breasted snipe, Limnodromus scolopaceus",
  "n02033444": "Numenius, genus Numenius",
  "n02033561": "curlew",
  "n02033779": "European curlew, Numenius arquata",
  "n02033882": "Eskimo curlew, Numenius borealis",
  "n02034016": "Limosa, genus Limosa",
  "n02034129": "godwit",
  "n02034295": "Hudsonian godwit, Limosa haemastica",
  "n02034394": "Himantopus, genus Himantopus",
  "n02034661": "stilt, stiltbird, longlegs, long-legs, stilt plover, Himantopus stilt",
  "n02034971": "black-necked stilt, Himantopus mexicanus",
  "n02035210": "black-winged stilt, Himantopus himantopus",
  "n02035402": "white-headed stilt, Himantopus himantopus leucocephalus",
  "n02035656": "kaki, Himantopus novae-zelandiae",
  "n02035845": "Cladorhyncus, genus Cladorhyncus",
  "n02036053": "stilt, Australian stilt",
  "n02036228": "banded stilt, Cladorhyncus leucocephalum",
  "n02036399": "Recurvirostridae, family Recurvirostridae",
  "n02036548": "Recurvirostra, genus Recurvirostra",
  "n02036711": "avocet",
  "n02036864": "Haematopodidae, family Haematopodidae",
  "n02036982": "Haematopus, genus Haematopus",
  "n02037110": "oystercatcher, oyster catcher",
  "n02037278": "Phalaropidae, family Phalaropidae",
  "n02037464": "phalarope",
  "n02037713": "Phalaropus, genus Phalaropus",
  "n02037869": "red phalarope, Phalaropus fulicarius",
  "n02038010": "Lobipes, genus Lobipes",
  "n02038141": "northern phalarope, Lobipes lobatus",
  "n02038329": "Steganopus, genus Steganopus",
  "n02038466": "Wilson's phalarope, Steganopus tricolor",
  "n02038617": "Glareolidae, family Glareolidae",
  "n02038837": "Glareola, genus Glareola",
  "n02038993": "pratincole, glareole",
  "n02039171": "courser",
  "n02039377": "Cursorius, genus Cursorius",
  "n02039497": "cream-colored courser, Cursorius cursor",
  "n02039660": "Pluvianus, genus Pluvianus",
  "n02039780": "crocodile bird, Pluvianus aegyptius",
  "n02039942": "Burhinidae, family Burhinidae",
  "n02040113": "Burhinus, genus Burhinus",
  "n02040266": "stone curlew, thick-knee, Burhinus oedicnemus",
  "n02040505": "coastal diving bird",
  "n02040698": "Lari, suborder Lari",
  "n02040872": "Laridae, family Laridae",
  "n02041085": "larid",
  "n02041246": "gull, seagull, sea gull",
  "n02041492": "Larus, genus Larus",
  "n02041678": "mew, mew gull, sea mew, Larus canus",
  "n02041875": "black-backed gull, great black-backed gull, cob, Larus marinus",
  "n02042046": "herring gull, Larus argentatus",
  "n02042180": "laughing gull, blackcap, pewit, pewit gull, Larus ridibundus",
  "n02042342": "Pagophila, genus Pagophila",
  "n02042472": "ivory gull, Pagophila eburnea",
  "n02042637": "Rissa, genus Rissa",
  "n02042759": "kittiwake",
  "n02042923": "Sterninae, subfamily Sterninae",
  "n02043063": "tern",
  "n02043207": "Sterna, genus Sterna",
  "n02043333": "sea swallow, Sterna hirundo",
  "n02043497": "Rynchopidae, family Rynchopidae",
  "n02043659": "Rynchops, genus Rynchops",
  "n02043808": "skimmer",
  "n02043999": "Stercorariidae, family Stercorariidae",
  "n02044178": "jaeger",
  "n02044358": "Stercorarius, genus Stercorarius",
  "n02044517": "parasitic jaeger, arctic skua, Stercorarius parasiticus",
  "n02044659": "Catharacta, genus Catharacta",
  "n02044778": "skua, bonxie",
  "n02044908": "great skua, Catharacta skua",
  "n02045024": "Alcidae, family Alcidae",
  "n02045369": "auk",
  "n02045596": "auklet",
  "n02045705": "Alca, genus Alca",
  "n02045864": "razorbill, razor-billed auk, Alca torda",
  "n02046045": "Plautus, genus Plautus",
  "n02046171": "little auk, dovekie, Plautus alle",
  "n02046321": "Pinguinus, genus Pinguinus",
  "n02046442": "great auk, Pinguinus impennis",
  "n02046613": "Cepphus, genus Cepphus",
  "n02046759": "guillemot",
  "n02046939": "black guillemot, Cepphus grylle",
  "n02047045": "pigeon guillemot, Cepphus columba",
  "n02047152": "Uria, genus Uria",
  "n02047260": "murre",
  "n02047411": "common murre, Uria aalge",
  "n02047517": "thick-billed murre, Uria lomvia",
  "n02047614": "puffin",
  "n02047835": "Fratercula, genus Fratercula",
  "n02047975": "Atlantic puffin, Fratercula arctica",
  "n02048115": "horned puffin, Fratercula corniculata",
  "n02048242": "Lunda, genus Lunda",
  "n02048353": "tufted puffin, Lunda cirrhata",
  "n02048514": "Gaviiformes, order Gaviiformes",
  "n02048698": "gaviiform seabird",
  "n02048832": "Gavidae, family Gavidae",
  "n02048952": "Gavia, genus Gavia",
  "n02049088": "loon, diver",
  "n02049299": "Podicipitiformes, order Podicipitiformes, Podicipediformes, order Podicipediformes, Colymbiformes, order Colymbiformes",
  "n02049532": "podicipitiform seabird",
  "n02049672": "Podicipedidae, family Podicipedidae",
  "n02049855": "Podiceps, genus Podiceps",
  "n02050004": "grebe",
  "n02050313": "great crested grebe, Podiceps cristatus",
  "n02050442": "red-necked grebe, Podiceps grisegena",
  "n02050586": "black-necked grebe, eared grebe, Podiceps nigricollis",
  "n02050809": "dabchick, little grebe, Podiceps ruficollis",
  "n02050921": "Podilymbus, genus Podilymbus",
  "n02051059": "pied-billed grebe, Podilymbus podiceps",
  "n02051213": "Pelecaniformes, order Pelecaniformes",
  "n02051474": "pelecaniform seabird",
  "n02051701": "Pelecanidae, family Pelecanidae",
  "n02051845": "pelican",
  "n02052044": "Pelecanus, genus Pelecanus",
  "n02052204": "white pelican, Pelecanus erythrorhynchos",
  "n02052365": "Old world white pelican, Pelecanus onocrotalus",
  "n02052511": "Fregatidae, family Fregatidae",
  "n02052639": "Fregata, genus Fregata",
  "n02052775": "frigate bird, man-of-war bird",
  "n02052936": "Sulidae, family Sulidae",
  "n02053083": "gannet",
  "n02053279": "Sula, genus Sula",
  "n02053425": "solan, solan goose, solant goose, Sula bassana",
  "n02053584": "booby",
  "n02053720": "Phalacrocoracidae, family Phalacrocoracidae",
  "n02053859": "Phalacrocorax, genus Phalacrocorax",
  "n02054036": "cormorant, Phalacrocorax carbo",
  "n02054251": "Anhingidae, family Anhingidae",
  "n02054376": "genus Anhinga",
  "n02054502": "snakebird, anhinga, darter",
  "n02054711": "water turkey, Anhinga anhinga",
  "n02054834": "Phaethontidae, family Phaethontidae",
  "n02054966": "Phaethon, genus Phaethon",
  "n02055107": "tropic bird, tropicbird, boatswain bird",
  "n02055280": "Sphenisciformes, order Sphenisciformes",
  "n02055431": "Spheniscidae, family Spheniscidae",
  "n02055658": "sphenisciform seabird",
  "n02055803": "penguin",
  "n02056091": "Pygoscelis, genus Pygoscelis",
  "n02056228": "Adelie, Adelie penguin, Pygoscelis adeliae",
  "n02056421": "Aptenodytes, genus Aptenodytes",
  "n02056570": "king penguin, Aptenodytes patagonica",
  "n02056728": "emperor penguin, Aptenodytes forsteri",
  "n02056873": "Spheniscus, genus Spheniscus",
  "n02057035": "jackass penguin, Spheniscus demersus",
  "n02057208": "Eudyptes, genus Eudyptes",
  "n02057330": "rock hopper, crested penguin",
  "n02057478": "Procellariiformes, order Procellariiformes",
  "n02057731": "pelagic bird, oceanic bird",
  "n02057898": "procellariiform seabird",
  "n02058074": "Diomedeidae, family Diomedeidae",
  "n02058221": "albatross, mollymawk",
  "n02058453": "genus Diomedea",
  "n02058594": "wandering albatross, Diomedea exulans",
  "n02058747": "black-footed albatross, gooney, gooney bird, goonie, goony, Diomedea nigripes",
  "n02058933": "Procellariidae, family Procellariidae",
  "n02059162": "petrel",
  "n02059393": "Procellaria, genus Procellaria",
  "n02059541": "white-chinned petrel, Procellaria aequinoctialis",
  "n02059723": "Macronectes, genus Macronectes",
  "n02059852": "giant petrel, giant fulmar, Macronectes giganteus",
  "n02060016": "Fulmarus, genus Fulmarus",
  "n02060133": "fulmar, fulmar petrel, Fulmarus glacialis",
  "n02060290": "Puffinus, genus Puffinus",
  "n02060411": "shearwater",
  "n02060569": "Manx shearwater, Puffinus puffinus",
  "n02060719": "Hydrobatidae, family Hydrobatidae",
  "n02060889": "storm petrel",
  "n02061073": "Hydrobates, genus Hydrobates",
  "n02061217": "stormy petrel, northern storm petrel, Hydrobates pelagicus",
  "n02061425": "Oceanites, genus Oceanites",
  "n02061560": "Mother Carey's chicken, Mother Carey's hen, Oceanites oceanicus",
  "n02061716": "Pelecanoididae, family Pelecanoididae",
  "n02061853": "diving petrel",
  "n02062017": "aquatic mammal",
  "n02062209": "Cetacea, order Cetacea",
  "n02062430": "cetacean, cetacean mammal, blower",
  "n02062744": "whale",
  "n02062991": "Mysticeti, suborder Mysticeti",
  "n02063224": "baleen whale, whalebone whale",
  "n02063516": "Balaenidae, family Balaenidae",
  "n02063662": "right whale",
  "n02063846": "Balaena, genus Balaena",
  "n02064000": "bowhead, bowhead whale, Greenland whale, Balaena mysticetus",
  "n02064154": "Balaenopteridae, family Balaenopteridae",
  "n02064338": "rorqual, razorback",
  "n02064608": "Balaenoptera, genus Balaenoptera",
  "n02064816": "blue whale, sulfur bottom, Balaenoptera musculus",
  "n02065026": "finback, finback whale, fin whale, common rorqual, Balaenoptera physalus",
  "n02065263": "sei whale, Balaenoptera borealis",
  "n02065407": "lesser rorqual, piked whale, minke whale, Balaenoptera acutorostrata",
  "n02065599": "Megaptera, genus Megaptera",
  "n02065726": "humpback, humpback whale, Megaptera novaeangliae",
  "n02065932": "Eschrichtiidae, family Eschrichtiidae",
  "n02066086": "Eschrichtius, genus Eschrichtius",
  "n02066245": "grey whale, gray whale, devilfish, Eschrichtius gibbosus, Eschrichtius robustus",
  "n02066450": "Odontoceti, suborder Odontoceti",
  "n02066707": "toothed whale",
  "n02066950": "Physeteridae, family Physeteridae",
  "n02067100": "Physeter, genus Physeter",
  "n02067240": "sperm whale, cachalot, black whale, Physeter catodon",
  "n02067462": "Kogia, genus Kogia",
  "n02067603": "pygmy sperm whale, Kogia breviceps",
  "n02067768": "dwarf sperm whale, Kogia simus",
  "n02067941": "Ziphiidae, family Ziphiidae, Hyperodontidae, family Hyperodontidae",
  "n02068206": "beaked whale",
  "n02068408": "Hyperoodon, genus Hyperoodon",
  "n02068541": "bottle-nosed whale, bottlenose whale, bottlenose, Hyperoodon ampullatus",
  "n02068735": "Delphinidae, family Delphinidae",
  "n02068974": "dolphin",
  "n02069271": "Delphinus, genus Delphinus",
  "n02069412": "common dolphin, Delphinus delphis",
  "n02069569": "Tursiops, genus Tursiops",
  "n02069701": "bottlenose dolphin, bottle-nosed dolphin, bottlenose",
  "n02069974": "Atlantic bottlenose dolphin, Tursiops truncatus",
  "n02070174": "Pacific bottlenose dolphin, Tursiops gilli",
  "n02070311": "Phocoena, genus Phocoena",
  "n02070430": "porpoise",
  "n02070624": "harbor porpoise, herring hog, Phocoena phocoena",
  "n02070776": "vaquita, Phocoena sinus",
  "n02070923": "genus Grampus",
  "n02071028": "grampus, Grampus griseus",
  "n02071173": "Orcinus, genus Orcinus",
  "n02071294": "killer whale, killer, orca, grampus, sea wolf, Orcinus orca",
  "n02071506": "Globicephala, genus Globicephala",
  "n02071636": "pilot whale, black whale, common blackfish, blackfish, Globicephala melaena",
  "n02071905": "Platanistidae, family Platanistidae",
  "n02072040": "river dolphin",
  "n02072209": "Monodontidae, family Monodontidae",
  "n02072355": "Monodon, genus Monodon",
  "n02072493": "narwhal, narwal, narwhale, Monodon monoceros",
  "n02072665": "Delphinapterus, genus Delphinapterus",
  "n02072798": "white whale, beluga, Delphinapterus leucas",
  "n02072953": "spouter",
  "n02073041": "Sirenia, order Sirenia",
  "n02073250": "sea cow, sirenian mammal, sirenian",
  "n02073532": "Trichechidae, family Trichechidae",
  "n02073679": "Trichechus, genus Trichecus",
  "n02073831": "manatee, Trichechus manatus",
  "n02074004": "Dugongidae, family Dugongidae",
  "n02074214": "genus Dugong",
  "n02074367": "dugong, Dugong dugon",
  "n02074542": "Hydrodamalis, genus Hydrodamalis",
  "n02074726": "Steller's sea cow, Hydrodamalis gigas",
  "n02074915": "Carnivora, order Carnivora",
  "n02075296": "carnivore",
  "n02075612": "omnivore",
  "n02075727": "Pinnipedia, suborder Pinnipedia",
  "n02075927": "pinniped mammal, pinniped, pinnatiped",
  "n02076196": "seal",
  "n02076402": "crabeater seal, crab-eating seal",
  "n02076535": "Otariidae, family Otariidae",
  "n02076779": "eared seal",
  "n02077023": "Arctocephalus, genus Arctocephalus",
  "n02077152": "fur seal",
  "n02077384": "guadalupe fur seal, Arctocephalus philippi",
  "n02077533": "Callorhinus, genus Callorhinus",
  "n02077658": "fur seal",
  "n02077787": "Alaska fur seal, Callorhinus ursinus",
  "n02077923": "sea lion",
  "n02078159": "Otaria, genus Otaria",
  "n02078292": "South American sea lion, Otaria Byronia",
  "n02078436": "Zalophus, genus Zalophus",
  "n02078574": "California sea lion, Zalophus californianus, Zalophus californicus",
  "n02078738": "Australian sea lion, Zalophus lobatus",
  "n02078882": "Eumetopias, genus Eumetopias",
  "n02079005": "Steller sea lion, Steller's sea lion, Eumetopias jubatus",
  "n02079170": "Phocidae, family Phocidae",
  "n02079389": "earless seal, true seal, hair seal",
  "n02079706": "Phoca, genus Phoca",
  "n02079851": "harbor seal, common seal, Phoca vitulina",
  "n02080022": "Pagophilus, genus Pagophilus",
  "n02080146": "harp seal, Pagophilus groenlandicus",
  "n02080291": "Mirounga, genus Mirounga",
  "n02080415": "elephant seal, sea elephant",
  "n02080586": "Erignathus, genus Erignathus",
  "n02080713": "bearded seal, squareflipper square flipper, Erignathus barbatus",
  "n02080934": "Cystophora, genus Cystophora",
  "n02081060": "hooded seal, bladdernose, Cystophora cristata",
  "n02081282": "Odobenidae, family Odobenidae",
  "n02081423": "Odobenus, genus Odobenus",
  "n02081571": "walrus, seahorse, sea horse",
  "n02081798": "Atlantic walrus, Odobenus rosmarus",
  "n02081927": "Pacific walrus, Odobenus divergens",
  "n02082056": "Fissipedia",
  "n02082190": "fissiped mammal, fissiped",
  "n02082358": "Tubulidentata, order Tubulidentata",
  "n02082498": "Orycteropodidae, family Orycteropodidae",
  "n02082632": "Orycteropus, genus Orycteropus",
  "n02082791": "aardvark, ant bear, anteater, Orycteropus afer",
  "n02083038": "Canidae, family Canidae",
  "n02083346": "canine, canid",
  "n02083672": "bitch",
  "n02083780": "brood bitch",
  "n02083863": "Canis, genus Canis",
  "n02084071": "dog, domestic dog, Canis familiaris",
  "n02084732": "pooch, doggie, doggy, barker, bow-wow",
  "n02084861": "cur, mongrel, mutt",
  "n02085019": "feist, fice",
  "n02085118": "pariah dog, pye-dog, pie-dog",
  "n02085272": "lapdog",
  "n02085374": "toy dog, toy",
  "n02085620": "Chihuahua",
  "n02085782": "Japanese spaniel",
  "n02085936": "Maltese dog, Maltese terrier, Maltese",
  "n02086079": "Pekinese, Pekingese, Peke",
  "n02086240": "Shih-Tzu",
  "n02086346": "toy spaniel",
  "n02086478": "English toy spaniel",
  "n02086646": "Blenheim spaniel",
  "n02086753": "King Charles spaniel",
  "n02086910": "papillon",
  "n02087046": "toy terrier",
  "n02087122": "hunting dog",
  "n02087314": "courser",
  "n02087394": "Rhodesian ridgeback",
  "n02087551": "hound, hound dog",
  "n02088094": "Afghan hound, Afghan",
  "n02088238": "basset, basset hound",
  "n02088364": "beagle",
  "n02088466": "bloodhound, sleuthhound",
  "n02088632": "bluetick",
  "n02088745": "boarhound",
  "n02088839": "coonhound",
  "n02088992": "coondog",
  "n02089078": "black-and-tan coonhound",
  "n02089232": "dachshund, dachsie, badger dog",
  "n02089468": "sausage dog, sausage hound",
  "n02089555": "foxhound",
  "n02089725": "American foxhound",
  "n02089867": "Walker hound, Walker foxhound",
  "n02089973": "English foxhound",
  "n02090129": "harrier",
  "n02090253": "Plott hound",
  "n02090379": "redbone",
  "n02090475": "wolfhound",
  "n02090622": "borzoi, Russian wolfhound",
  "n02090721": "Irish wolfhound",
  "n02090827": "greyhound",
  "n02091032": "Italian greyhound",
  "n02091134": "whippet",
  "n02091244": "Ibizan hound, Ibizan Podenco",
  "n02091467": "Norwegian elkhound, elkhound",
  "n02091635": "otterhound, otter hound",
  "n02091831": "Saluki, gazelle hound",
  "n02092002": "Scottish deerhound, deerhound",
  "n02092173": "staghound",
  "n02092339": "Weimaraner",
  "n02092468": "terrier",
  "n02093056": "bullterrier, bull terrier",
  "n02093256": "Staffordshire bullterrier, Staffordshire bull terrier",
  "n02093428": "American Staffordshire terrier, Staffordshire terrier, American pit bull terrier, pit bull terrier",
  "n02093647": "Bedlington terrier",
  "n02093754": "Border terrier",
  "n02093859": "Kerry blue terrier",
  "n02093991": "Irish terrier",
  "n02094114": "Norfolk terrier",
  "n02094258": "Norwich terrier",
  "n02094433": "Yorkshire terrier",
  "n02094562": "rat terrier, ratter",
  "n02094721": "Manchester terrier, black-and-tan terrier",
  "n02094931": "toy Manchester, toy Manchester terrier",
  "n02095050": "fox terrier",
  "n02095212": "smooth-haired fox terrier",
  "n02095314": "wire-haired fox terrier",
  "n02095412": "wirehair, wirehaired terrier, wire-haired terrier",
  "n02095570": "Lakeland terrier",
  "n02095727": "Welsh terrier",
  "n02095889": "Sealyham terrier, Sealyham",
  "n02096051": "Airedale, Airedale terrier",
  "n02096177": "cairn, cairn terrier",
  "n02096294": "Australian terrier",
  "n02096437": "Dandie Dinmont, Dandie Dinmont terrier",
  "n02096585": "Boston bull, Boston terrier",
  "n02096756": "schnauzer",
  "n02097047": "miniature schnauzer",
  "n02097130": "giant schnauzer",
  "n02097209": "standard schnauzer",
  "n02097298": "Scotch terrier, Scottish terrier, Scottie",
  "n02097474": "Tibetan terrier, chrysanthemum dog",
  "n02097658": "silky terrier, Sydney silky",
  "n02097786": "Skye terrier",
  "n02097967": "Clydesdale terrier",
  "n02098105": "soft-coated wheaten terrier",
  "n02098286": "West Highland white terrier",
  "n02098413": "Lhasa, Lhasa apso",
  "n02098550": "sporting dog, gun dog",
  "n02098806": "bird dog",
  "n02098906": "water dog",
  "n02099029": "retriever",
  "n02099267": "flat-coated retriever",
  "n02099429": "curly-coated retriever",
  "n02099601": "golden retriever",
  "n02099712": "Labrador retriever",
  "n02099849": "Chesapeake Bay retriever",
  "n02099997": "pointer, Spanish pointer",
  "n02100236": "German short-haired pointer",
  "n02100399": "setter",
  "n02100583": "vizsla, Hungarian pointer",
  "n02100735": "English setter",
  "n02100877": "Irish setter, red setter",
  "n02101006": "Gordon setter",
  "n02101108": "spaniel",
  "n02101388": "Brittany spaniel",
  "n02101556": "clumber, clumber spaniel",
  "n02101670": "field spaniel",
  "n02101861": "springer spaniel, springer",
  "n02102040": "English springer, English springer spaniel",
  "n02102177": "Welsh springer spaniel",
  "n02102318": "cocker spaniel, English cocker spaniel, cocker",
  "n02102480": "Sussex spaniel",
  "n02102605": "water spaniel",
  "n02102806": "American water spaniel",
  "n02102973": "Irish water spaniel",
  "n02103181": "griffon, wire-haired pointing griffon",
  "n02103406": "working dog",
  "n02103841": "watchdog, guard dog",
  "n02104029": "kuvasz",
  "n02104184": "attack dog",
  "n02104280": "housedog",
  "n02104365": "schipperke",
  "n02104523": "shepherd dog, sheepdog, sheep dog",
  "n02104882": "Belgian sheepdog, Belgian shepherd",
  "n02105056": "groenendael",
  "n02105162": "malinois",
  "n02105251": "briard",
  "n02105412": "kelpie",
  "n02105505": "komondor",
  "n02105641": "Old English sheepdog, bobtail",
  "n02105855": "Shetland sheepdog, Shetland sheep dog, Shetland",
  "n02106030": "collie",
  "n02106166": "Border collie",
  "n02106382": "Bouvier des Flandres, Bouviers des Flandres",
  "n02106550": "Rottweiler",
  "n02106662": "German shepherd, German shepherd dog, German police dog, alsatian",
  "n02106854": "police dog",
  "n02106966": "pinscher",
  "n02107142": "Doberman, Doberman pinscher",
  "n02107312": "miniature pinscher",
  "n02107420": "Sennenhunde",
  "n02107574": "Greater Swiss Mountain dog",
  "n02107683": "Bernese mountain dog",
  "n02107908": "Appenzeller",
  "n02108000": "EntleBucher",
  "n02108089": "boxer",
  "n02108254": "mastiff",
  "n02108422": "bull mastiff",
  "n02108551": "Tibetan mastiff",
  "n02108672": "bulldog, English bulldog",
  "n02108915": "French bulldog",
  "n02109047": "Great Dane",
  "n02109150": "guide dog",
  "n02109256": "Seeing Eye dog",
  "n02109391": "hearing dog",
  "n02109525": "Saint Bernard, St Bernard",
  "n02109687": "seizure-alert dog",
  "n02109811": "sled dog, sledge dog",
  "n02109961": "Eskimo dog, husky",
  "n02110063": "malamute, malemute, Alaskan malamute",
  "n02110185": "Siberian husky",
  "n02110341": "dalmatian, coach dog, carriage dog",
  "n02110532": "liver-spotted dalmatian",
  "n02110627": "affenpinscher, monkey pinscher, monkey dog",
  "n02110806": "basenji",
  "n02110958": "pug, pug-dog",
  "n02111129": "Leonberg",
  "n02111277": "Newfoundland, Newfoundland dog",
  "n02111500": "Great Pyrenees",
  "n02111626": "spitz",
  "n02111889": "Samoyed, Samoyede",
  "n02112018": "Pomeranian",
  "n02112137": "chow, chow chow",
  "n02112350": "keeshond",
  "n02112497": "griffon, Brussels griffon, Belgian griffon",
  "n02112706": "Brabancon griffon",
  "n02112826": "corgi, Welsh corgi",
  "n02113023": "Pembroke, Pembroke Welsh corgi",
  "n02113186": "Cardigan, Cardigan Welsh corgi",
  "n02113335": "poodle, poodle dog",
  "n02113624": "toy poodle",
  "n02113712": "miniature poodle",
  "n02113799": "standard poodle",
  "n02113892": "large poodle",
  "n02113978": "Mexican hairless",
  "n02114100": "wolf",
  "n02114367": "timber wolf, grey wolf, gray wolf, Canis lupus",
  "n02114548": "white wolf, Arctic wolf, Canis lupus tundrarum",
  "n02114712": "red wolf, maned wolf, Canis rufus, Canis niger",
  "n02114855": "coyote, prairie wolf, brush wolf, Canis latrans",
  "n02115012": "coydog",
  "n02115096": "jackal, Canis aureus",
  "n02115335": "wild dog",
  "n02115641": "dingo, warrigal, warragal, Canis dingo",
  "n02115775": "Cuon, Cyon, genus Cuon, genus Cyon",
  "n02115913": "dhole, Cuon alpinus",
  "n02116079": "Dusicyon, genus Dusicyon",
  "n02116185": "crab-eating dog, crab-eating fox, Dusicyon cancrivorus",
  "n02116322": "Nyctereutes, genus Nyctereutes",
  "n02116450": "raccoon dog, Nyctereutes procyonides",
  "n02116630": "Lycaeon, genus Lycaeon",
  "n02116738": "African hunting dog, hyena dog, Cape hunting dog, Lycaon pictus",
  "n02116959": "Hyaenidae, family Hyaenidae",
  "n02117135": "hyena, hyaena",
  "n02117369": "genus Hyaena",
  "n02117512": "striped hyena, Hyaena hyaena",
  "n02117646": "brown hyena, strand wolf, Hyaena brunnea",
  "n02117772": "Crocuta, genus Crocuta",
  "n02117900": "spotted hyena, laughing hyena, Crocuta crocuta",
  "n02118058": "Proteles, genus Proteles",
  "n02118176": "aardwolf, Proteles cristata",
  "n02118333": "fox",
  "n02118643": "vixen",
  "n02118707": "Reynard",
  "n02118854": "Vulpes, genus Vulpes",
  "n02119022": "red fox, Vulpes vulpes",
  "n02119247": "black fox",
  "n02119359": "silver fox",
  "n02119477": "red fox, Vulpes fulva",
  "n02119634": "kit fox, prairie fox, Vulpes velox",
  "n02119789": "kit fox, Vulpes macrotis",
  "n02119961": "Alopex, genus Alopex",
  "n02120079": "Arctic fox, white fox, Alopex lagopus",
  "n02120278": "blue fox",
  "n02120387": "Urocyon, genus Urocyon",
  "n02120505": "grey fox, gray fox, Urocyon cinereoargenteus",
  "n02120692": "Felidae, family Felidae",
  "n02120997": "feline, felid",
  "n02121234": "Felis, genus Felis",
  "n02121620": "cat, true cat",
  "n02121808": "domestic cat, house cat, Felis domesticus, Felis catus",
  "n02122298": "kitty, kitty-cat, puss, pussy, pussycat",
  "n02122430": "mouser",
  "n02122510": "alley cat",
  "n02122580": "stray",
  "n02122725": "tom, tomcat",
  "n02122810": "gib",
  "n02122878": "tabby, queen",
  "n02122948": "kitten, kitty",
  "n02123045": "tabby, tabby cat",
  "n02123159": "tiger cat",
  "n02123242": "tortoiseshell, tortoiseshell-cat, calico cat",
  "n02123394": "Persian cat",
  "n02123478": "Angora, Angora cat",
  "n02123597": "Siamese cat, Siamese",
  "n02123785": "blue point Siamese",
  "n02123917": "Burmese cat",
  "n02124075": "Egyptian cat",
  "n02124157": "Maltese, Maltese cat",
  "n02124313": "Abyssinian, Abyssinian cat",
  "n02124484": "Manx, Manx cat",
  "n02124623": "wildcat",
  "n02125010": "sand cat",
  "n02125081": "European wildcat, catamountain, Felis silvestris",
  "n02125311": "cougar, puma, catamount, mountain lion, painter, panther, Felis concolor",
  "n02125494": "ocelot, panther cat, Felis pardalis",
  "n02125689": "jaguarundi, jaguarundi cat, jaguarondi, eyra, Felis yagouaroundi",
  "n02125872": "kaffir cat, caffer cat, Felis ocreata",
  "n02126028": "jungle cat, Felis chaus",
  "n02126139": "serval, Felis serval",
  "n02126317": "leopard cat, Felis bengalensis",
  "n02126465": "tiger cat, Felis tigrina",
  "n02126640": "margay, margay cat, Felis wiedi",
  "n02126787": "manul, Pallas's cat, Felis manul",
  "n02126951": "genus Lynx",
  "n02127052": "lynx, catamount",
  "n02127292": "common lynx, Lynx lynx",
  "n02127381": "Canada lynx, Lynx canadensis",
  "n02127482": "bobcat, bay lynx, Lynx rufus",
  "n02127586": "spotted lynx, Lynx pardina",
  "n02127678": "caracal, desert lynx, Lynx caracal",
  "n02127808": "big cat, cat",
  "n02128120": "Panthera, genus Panthera",
  "n02128385": "leopard, Panthera pardus",
  "n02128598": "leopardess",
  "n02128669": "panther",
  "n02128757": "snow leopard, ounce, Panthera uncia",
  "n02128925": "jaguar, panther, Panthera onca, Felis onca",
  "n02129165": "lion, king of beasts, Panthera leo",
  "n02129463": "lioness",
  "n02129530": "lionet",
  "n02129604": "tiger, Panthera tigris",
  "n02129837": "Bengal tiger",
  "n02129923": "tigress",
  "n02129991": "liger",
  "n02130086": "tiglon, tigon",
  "n02130190": "Acinonyx, genus Acinonyx",
  "n02130308": "cheetah, chetah, Acinonyx jubatus",
  "n02130545": "saber-toothed tiger, sabertooth",
  "n02130795": "Smiledon, genus Smiledon",
  "n02130925": "Smiledon californicus",
  "n02131076": "Nimravus, genus Nimravus",
  "n02131211": "false saber-toothed tiger",
  "n02131418": "Ursidae, family Ursidae",
  "n02131653": "bear",
  "n02131942": "Ursus, genus Ursus",
  "n02132136": "brown bear, bruin, Ursus arctos",
  "n02132320": "bruin",
  "n02132466": "Syrian bear, Ursus arctos syriacus",
  "n02132580": "grizzly, grizzly bear, silvertip, silver-tip, Ursus horribilis, Ursus arctos horribilis",
  "n02132788": "Alaskan brown bear, Kodiak bear, Kodiak, Ursus middendorffi, Ursus arctos middendorffi",
  "n02132974": "Euarctos, genus Euarctos",
  "n02133161": "American black bear, black bear, Ursus americanus, Euarctos americanus",
  "n02133400": "cinnamon bear",
  "n02133512": "Selenarctos, genus Selenarctos",
  "n02133704": "Asiatic black bear, black bear, Ursus thibetanus, Selenarctos thibetanus",
  "n02133902": "Thalarctos, genus Thalarctos",
  "n02134084": "ice bear, polar bear, Ursus Maritimus, Thalarctos maritimus",
  "n02134240": "Melursus, genus Melursus",
  "n02134418": "sloth bear, Melursus ursinus, Ursus ursinus",
  "n02134589": "Viverridae, family Viverridae, Viverrinae, family Viverrinae",
  "n02134971": "viverrine, viverrine mammal",
  "n02135220": "civet, civet cat",
  "n02135486": "Viverra, genus Viverra",
  "n02135610": "large civet, Viverra zibetha",
  "n02135726": "Viverricula, genus Viverricula",
  "n02135844": "small civet, Viverricula indica, Viverricula malaccensis",
  "n02135981": "Arctictis, genus Arctictis",
  "n02136103": "binturong, bearcat, Arctictis bintourong",
  "n02136285": "Cryptoprocta, genus Cryptoprocta",
  "n02136452": "fossa, fossa cat, Cryptoprocta ferox",
  "n02136623": "Fossa, genus Fossa",
  "n02136794": "fanaloka, Fossa fossa",
  "n02136901": "Genetta, genus Genetta",
  "n02137015": "genet, Genetta genetta",
  "n02137172": "Hemigalus, genus Hemigalus",
  "n02137302": "banded palm civet, Hemigalus hardwickii",
  "n02137428": "Herpestes, genus Herpestes",
  "n02137549": "mongoose",
  "n02137722": "Indian mongoose, Herpestes nyula",
  "n02137888": "ichneumon, Herpestes ichneumon",
  "n02138042": "Paradoxurus, genus Paradoxurus",
  "n02138169": "palm cat, palm civet",
  "n02138323": "Suricata, genus Suricata",
  "n02138441": "meerkat, mierkat",
  "n02138647": "slender-tailed meerkat, Suricata suricatta",
  "n02138777": "suricate, Suricata tetradactyla",
  "n02138921": "Chiroptera, order Chiroptera",
  "n02139199": "bat, chiropteran",
  "n02139479": "Megachiroptera, suborder Megachiroptera",
  "n02139671": "fruit bat, megabat",
  "n02139914": "Pteropus, genus Pteropus",
  "n02140049": "flying fox",
  "n02140179": "Pteropus capestratus",
  "n02140268": "Pteropus hypomelanus",
  "n02140357": "Nyctimene, genus Nyctimene",
  "n02140491": "harpy, harpy bat, tube-nosed bat, tube-nosed fruit bat",
  "n02140719": "Cynopterus, genus Cynopterus",
  "n02140858": "Cynopterus sphinx",
  "n02140970": "Microchiroptera, suborder Microchiroptera",
  "n02141306": "carnivorous bat, microbat",
  "n02141611": "mouse-eared bat",
  "n02141713": "leafnose bat, leaf-nosed bat",
  "n02142064": "Phyllostomidae, family Phyllostomidae, Phyllostomatidae, family Phyllostomatidae",
  "n02142295": "genus Macrotus",
  "n02142407": "macrotus, Macrotus californicus",
  "n02142575": "Phyllostomus, genus Phyllostomus",
  "n02142734": "spearnose bat",
  "n02142898": "Phyllostomus hastatus",
  "n02142993": "Choeronycteris, genus Choeronycteris",
  "n02143142": "hognose bat, Choeronycteris mexicana",
  "n02143293": "Rhinolophidae, family Rhinolophidae",
  "n02143439": "horseshoe bat",
  "n02143594": "Hipposideridae, family Hipposideridae",
  "n02143759": "Hipposideros, genus Hipposideros",
  "n02143891": "horseshoe bat",
  "n02144110": "Rhinonicteris, genus Rhinonicteris",
  "n02144251": "orange bat, orange horseshoe bat, Rhinonicteris aurantius",
  "n02144442": "Megadermatidae, family Megadermatidae",
  "n02144593": "false vampire, false vampire bat",
  "n02144792": "Megaderma, genus Megaderma",
  "n02144936": "big-eared bat, Megaderma lyra",
  "n02145084": "Vespertilionidae, family Vespertilionidae",
  "n02145424": "vespertilian bat, vespertilionid",
  "n02145767": "Vespertilio, genus Vespertilio",
  "n02145910": "frosted bat, Vespertilio murinus",
  "n02146064": "Lasiurus, genus Lasiurus",
  "n02146201": "red bat, Lasiurus borealis",
  "n02146371": "brown bat",
  "n02146526": "Myotis, genus Myotis",
  "n02146700": "little brown bat, little brown myotis, Myotis leucifugus",
  "n02146879": "cave myotis, Myotis velifer",
  "n02147034": "Eptesicus, genus Eptesicus",
  "n02147173": "big brown bat, Eptesicus fuscus",
  "n02147328": "serotine, European brown bat, Eptesicus serotinus",
  "n02147452": "Antrozous, genus Antrozous",
  "n02147591": "pallid bat, cave bat, Antrozous pallidus",
  "n02147747": "Pipistrellus, genus Pipistrellus",
  "n02147947": "pipistrelle, pipistrel, Pipistrellus pipistrellus",
  "n02148088": "eastern pipistrel, Pipistrellus subflavus",
  "n02148245": "western pipistrel, SPipistrellus hesperus",
  "n02148377": "Euderma, genus Euderma",
  "n02148512": "jackass bat, spotted bat, Euderma maculata",
  "n02148698": "Plecotus, genus Plecotus",
  "n02148835": "long-eared bat",
  "n02148991": "western big-eared bat, Plecotus townsendi",
  "n02149136": "Molossidae, family Molossidae",
  "n02149297": "Tadarida, genus Tadarida",
  "n02149420": "freetail, free-tailed bat, freetailed bat",
  "n02149653": "guano bat, Mexican freetail bat, Tadarida brasiliensis",
  "n02149861": "pocketed bat, pocketed freetail bat, Tadirida femorosacca",
  "n02150016": "Eumops, genus Eumops",
  "n02150134": "mastiff bat",
  "n02150306": "Desmodontidae, family Desmodontidae",
  "n02150482": "vampire bat, true vampire bat",
  "n02150730": "Desmodus, genus Desmodus",
  "n02150885": "Desmodus rotundus",
  "n02151108": "Diphylla, genus Diphylla",
  "n02151230": "hairy-legged vampire bat, Diphylla ecaudata",
  "n02151448": "water vascular system",
  "n02151625": "wing",
  "n02151940": "ala",
  "n02152064": "forewing, fore-wing, fore wing",
  "n02152212": "halter, haltere, balancer",
  "n02152446": "pennon, pinion",
  "n02152559": "wing case, elytron",
  "n02152740": "predator, predatory animal",
  "n02152881": "prey, quarry",
  "n02152991": "game",
  "n02153109": "big game",
  "n02153203": "game bird",
  "n02153445": "animal foot, foot",
  "n02153712": "fossorial foot",
  "n02153809": "fossorial mammal",
  "n02153959": "hoof",
  "n02154133": "hoof",
  "n02154264": "cloven foot, cloven hoof",
  "n02154416": "bird's foot",
  "n02154620": "webfoot",
  "n02154742": "claw",
  "n02154824": "zygodactyl foot",
  "n02155014": "heterodactyl foot",
  "n02155206": "webbed foot",
  "n02155313": "lobate foot",
  "n02155448": "calyculus, caliculus, calycle",
  "n02155678": "optic cup, eyecup",
  "n02155872": "tooth",
  "n02156031": "denticle",
  "n02156140": "claw",
  "n02156320": "bear claw",
  "n02156413": "talon",
  "n02156532": "claw, chela, nipper, pincer",
  "n02156732": "tetrapod",
  "n02156871": "quadruped",
  "n02157206": "hexapod",
  "n02157285": "biped",
  "n02157415": "belly",
  "n02157557": "tail",
  "n02157954": "brush",
  "n02158066": "bobtail, bob, dock",
  "n02158213": "caudal appendage",
  "n02158354": "uropygium",
  "n02158494": "oxtail",
  "n02158619": "fluke",
  "n02158739": "scut",
  "n02158846": "flag",
  "n02158972": "dock",
  "n02159117": "horse's foot",
  "n02159271": "Insecta, class Insecta, Hexapoda, class Hexapoda",
  "n02159955": "insect",
  "n02160947": "social insect",
  "n02161078": "ephemeron, ephemeral",
  "n02161225": "holometabola, metabola",
  "n02161338": "defoliator",
  "n02161457": "pollinator",
  "n02161588": "gallfly",
  "n02161737": "Mantophasmatodea, order mantophasmatodea",
  "n02161944": "Mecoptera, order Mecoptera",
  "n02162235": "mecopteran",
  "n02162404": "Panorpidae, family Panorpidae",
  "n02162561": "scorpion fly",
  "n02162831": "Bittacidae, family Bittacidae",
  "n02163008": "hanging fly",
  "n02163144": "Collembola, order Collembola",
  "n02163297": "collembolan, springtail",
  "n02163616": "Protura, order Protura",
  "n02163763": "proturan, telsontail",
  "n02163982": "Coleoptera, order Coleoptera",
  "n02164464": "beetle",
  "n02164973": "Cicindelidae, family Cicindelidae",
  "n02165105": "tiger beetle",
  "n02165247": "Coccinellidae, family Coccinellidae",
  "n02165456": "ladybug, ladybeetle, lady beetle, ladybird, ladybird beetle",
  "n02165754": "Adalia, genus Adalia",
  "n02165877": "two-spotted ladybug, Adalia bipunctata",
  "n02166024": "Epilachna, genus Epilachna",
  "n02166229": "Mexican bean beetle, bean beetle, Epilachna varivestis",
  "n02166436": "Hippodamia, genus Hippodamia",
  "n02166567": "Hippodamia convergens",
  "n02166674": "Rodolia, genus Rodolia, genus Vedalia",
  "n02166826": "vedalia, Rodolia cardinalis",
  "n02166986": "Carabidae, family Carabidae",
  "n02167151": "ground beetle, carabid beetle",
  "n02167375": "Brachinus, genus Brachinus",
  "n02167505": "bombardier beetle",
  "n02167645": "genus Calosoma",
  "n02167820": "calosoma",
  "n02167944": "searcher, searcher beetle, Calosoma scrutator",
  "n02168121": "Lampyridae, family Lampyridae",
  "n02168245": "firefly, lightning bug",
  "n02168427": "glowworm",
  "n02168542": "Cerambycidae, family Cerambycidae",
  "n02168699": "long-horned beetle, longicorn, longicorn beetle",
  "n02168876": "Monochamus, genus Monochamus",
  "n02169023": "sawyer, sawyer beetle",
  "n02169218": "pine sawyer",
  "n02169345": "Chrysomelidae, family Chrysomelidae",
  "n02169497": "leaf beetle, chrysomelid",
  "n02169705": "flea beetle",
  "n02169833": "Leptinotarsa, genus Leptinotarsa",
  "n02169974": "Colorado potato beetle, Colorado beetle, potato bug, potato beetle, Leptinotarsa decemlineata",
  "n02170269": "Dermestidae, family Dermestidae",
  "n02170400": "carpet beetle, carpet bug",
  "n02170599": "buffalo carpet beetle, Anthrenus scrophulariae",
  "n02170738": "black carpet beetle",
  "n02170848": "Cleridae, family Cleridae",
  "n02170993": "clerid beetle, clerid",
  "n02171164": "bee beetle",
  "n02171254": "Lamellicornia, superfamily Lamellicornia",
  "n02171453": "lamellicorn beetle",
  "n02171633": "Scarabaeidae, family Scarabaeidae",
  "n02171869": "scarabaeid beetle, scarabaeid, scarabaean",
  "n02172182": "dung beetle",
  "n02172387": "genus Scarabaeus",
  "n02172518": "scarab, scarabaeus, Scarabaeus sacer",
  "n02172678": "tumblebug",
  "n02172761": "dorbeetle",
  "n02172870": "June beetle, June bug, May bug, May beetle",
  "n02173113": "green June beetle, figeater",
  "n02173240": "Popillia, genus Popillia",
  "n02173373": "Japanese beetle, Popillia japonica",
  "n02173571": "Anomala, genus Anomala",
  "n02173784": "Oriental beetle, Asiatic beetle, Anomala orientalis",
  "n02174001": "rhinoceros beetle",
  "n02174153": "Melolonthidae, subfamily Melolonthidae",
  "n02174355": "melolonthid beetle",
  "n02174521": "Melolontha, genus Melolontha",
  "n02174659": "cockchafer, May bug, May beetle, Melolontha melolontha",
  "n02174870": "Macrodactylus, genus Macrodactylus",
  "n02175014": "rose chafer, rose bug, Macrodactylus subspinosus",
  "n02175263": "Cetoniidae, subfamily Cetoniidae",
  "n02175440": "Cetonia, genus Cetonia",
  "n02175569": "rose chafer, rose beetle, Cetonia aurata",
  "n02175791": "Lucanidae, family Lucanidae",
  "n02175916": "stag beetle",
  "n02176073": "Elateridae, family Elateridae",
  "n02176261": "elaterid beetle, elater, elaterid",
  "n02176439": "click beetle, skipjack, snapping beetle",
  "n02176611": "Pyrophorus, genus Pyrophorus",
  "n02176747": "firefly, fire beetle, Pyrophorus noctiluca",
  "n02176916": "wireworm",
  "n02177068": "Dytiscidae, family Dytiscidae",
  "n02177196": "water beetle",
  "n02177376": "Gyrinidae, family Gyrinidae",
  "n02177506": "whirligig beetle",
  "n02177644": "Anobiidae, family Anobiidae",
  "n02177775": "deathwatch beetle, deathwatch, Xestobium rufovillosum",
  "n02177972": "weevil",
  "n02178244": "Curculionidae, family Curculionidae",
  "n02178411": "snout beetle",
  "n02178563": "Anthonomus, genus Anthonomus",
  "n02178717": "boll weevil, Anthonomus grandis",
  "n02178886": "Meloidae, family Meloidae",
  "n02179012": "blister beetle, meloid",
  "n02179192": "oil beetle",
  "n02179340": "Spanish fly",
  "n02179429": "Scolytidae, family Scolytidae, Ipidae, family Ipidae",
  "n02179714": "Scolytus, genus Scolytus",
  "n02179891": "Dutch-elm beetle, Scolytus multistriatus",
  "n02180046": "Dendroctonus, genus Dendroctonus",
  "n02180233": "bark beetle",
  "n02180427": "spruce bark beetle, Dendroctonus rufipennis",
  "n02180742": "Staphylinidae, family Staphylinidae",
  "n02180875": "rove beetle",
  "n02181013": "Tenebrionidae, family Tenebrionidae",
  "n02181235": "darkling beetle, darkling groung beetle, tenebrionid",
  "n02181477": "mealworm",
  "n02181599": "Tribolium, genus Tribolium",
  "n02181724": "flour beetle, flour weevil",
  "n02181863": "Bruchidae, family Bruchidae",
  "n02182045": "seed beetle, seed weevil",
  "n02182220": "Bruchus, genus Bruchus",
  "n02182355": "pea weevil, Bruchus pisorum",
  "n02182498": "Acanthoscelides, genus Acanthoscelides",
  "n02182642": "bean weevil, Acanthoscelides obtectus",
  "n02182796": "Sitophylus, genus Sitophylus",
  "n02182930": "rice weevil, black weevil, Sitophylus oryzae",
  "n02183096": "Asian longhorned beetle, Anoplophora glabripennis",
  "n02183353": "Embioptera, order Embioptera, Embiodea, order Embiodea",
  "n02183507": "web spinner",
  "n02183697": "Anoplura, order Anoplura",
  "n02183857": "louse, sucking louse",
  "n02184114": "Pediculidae, family Pediculidae",
  "n02184270": "Pediculus, genus Pediculus",
  "n02184473": "common louse, Pediculus humanus",
  "n02184589": "head louse, Pediculus capitis",
  "n02184720": "body louse, cootie, Pediculus corporis",
  "n02184881": "Phthiriidae, family Phthiriidae",
  "n02185007": "Phthirius, genus Phthirius, Phthirus, genus Phthirus",
  "n02185167": "crab louse, pubic louse, crab, Phthirius pubis",
  "n02185337": "Mallophaga, order Mallophaga",
  "n02185481": "bird louse, biting louse, louse",
  "n02185694": "Menopon, genus Menopon",
  "n02185814": "chicken louse, shaft louse, Menopon palladum, Menopon gallinae",
  "n02185973": "Siphonaptera, order Siphonaptera",
  "n02186153": "flea",
  "n02186399": "Pulicidae, family Pulicidae",
  "n02186586": "Pulex, genus Pulex",
  "n02186717": "Pulex irritans",
  "n02186834": "Ctenocephalides, genus Ctenocephalides",
  "n02187022": "Ctenocephalus, genus Ctenocephalus",
  "n02187150": "dog flea, Ctenocephalides canis",
  "n02187279": "cat flea, Ctenocephalides felis",
  "n02187427": "Tunga, genus Tunga",
  "n02187554": "chigoe, chigger, chigoe flea, Tunga penetrans",
  "n02187759": "Echidnophaga, genus Echidnophaga",
  "n02187900": "sticktight, sticktight flea, Echidnophaga gallinacea",
  "n02188065": "Diptera, order Diptera",
  "n02188699": "dipterous insect, two-winged insects, dipteran, dipteron",
  "n02189214": "Cecidomyidae, family Cecidomyidae",
  "n02189363": "gall midge, gallfly, gall gnat",
  "n02189535": "Mayetiola, genus Mayetiola",
  "n02189670": "Hessian fly, Mayetiola destructor",
  "n02189822": "Muscoidea, superfamily Muscoidea",
  "n02190015": "Muscidae, family Muscidae",
  "n02190166": "fly",
  "n02190465": "alula, calypter",
  "n02190648": "Musca, genus Musca",
  "n02190790": "housefly, house fly, Musca domestica",
  "n02190963": "Glossinidae, family Glossinidae",
  "n02191131": "genus Glossina",
  "n02191273": "tsetse fly, tsetse, tzetze fly, tzetze, glossina",
  "n02191449": "Calliphoridae, family Calliphoridae",
  "n02191617": "Calliphora, genus Calliphora",
  "n02191773": "blowfly, blow fly",
  "n02191979": "bluebottle, Calliphora vicina",
  "n02192127": "Lucilia, genus Lucilia",
  "n02192252": "greenbottle, greenbottle fly",
  "n02192388": "Sarcophaga, genus Sarcophaga",
  "n02192513": "flesh fly, Sarcophaga carnaria",
  "n02192673": "Tachinidae, family Tachinidae",
  "n02192814": "tachina fly",
  "n02193009": "gadfly",
  "n02193163": "botfly",
  "n02193357": "Gasterophilidae, family Gasterophilidae",
  "n02193496": "Gasterophilus, genus Gasterophilus",
  "n02193665": "horse botfly, Gasterophilus intestinalis",
  "n02193799": "Cuterebridae, family Cuterebridae",
  "n02193955": "Cuterebra, genus Cuterebra",
  "n02194078": "Dermatobia, genus Dermatobia",
  "n02194249": "human botfly, Dermatobia hominis",
  "n02194414": "Oestridae, family Oestridae, Hypodermatidae, family Hypodermatidae",
  "n02194599": "Oestrus, genus Oestrus",
  "n02194750": "sheep botfly, sheep gadfly, Oestrus ovis",
  "n02194887": "Hypoderma, genus Hypoderma",
  "n02195091": "warble fly",
  "n02195257": "warble",
  "n02195403": "Tabanidae, family Tabanidae",
  "n02195526": "horsefly, cleg, clegg, horse fly",
  "n02195693": "Bombyliidae, family Bombyliidae",
  "n02195819": "bee fly",
  "n02195996": "Asilidae, family Asilidae",
  "n02196119": "robber fly, bee killer",
  "n02196344": "fruit fly, pomace fly",
  "n02196542": "Trypetidae, family Trypetidae, Trephritidae, family Trephritidae",
  "n02196761": "Rhagoletis, genus Rhagoletis",
  "n02196896": "apple maggot, railroad worm, Rhagoletis pomonella",
  "n02197048": "Ceratitis, genus Ceratitis",
  "n02197185": "Mediterranean fruit fly, medfly, Ceratitis capitata",
  "n02197413": "Drosophilidae, family Drosophilidae",
  "n02197545": "genus Drosophila",
  "n02197689": "drosophila, Drosophila melanogaster",
  "n02197877": "vinegar fly",
  "n02198021": "Philophylla, genus Philophylla",
  "n02198129": "leaf miner, leaf-miner",
  "n02198332": "Hippoboscidae, family Hippoboscidae",
  "n02198532": "louse fly, hippoboscid",
  "n02198714": "Hippobosca, genus Hippobosca",
  "n02198859": "horse tick, horsefly, Hippobosca equina",
  "n02198996": "Melophagus, genus Melophagus",
  "n02199170": "sheep ked, sheep-tick, sheep tick, Melophagus Ovinus",
  "n02199352": "Haematobia, genus Haematobia",
  "n02199502": "horn fly, Haematobia irritans",
  "n02199712": "Nematocera, suborder Nematocera",
  "n02199999": "Culicidae, family Culicidae",
  "n02200198": "mosquito",
  "n02200509": "wiggler, wriggler",
  "n02200630": "gnat",
  "n02200705": "Aedes, genus Aedes",
  "n02200850": "yellow-fever mosquito, Aedes aegypti",
  "n02201000": "Asian tiger mosquito, Aedes albopictus",
  "n02201252": "Anopheles, genus Anopheles",
  "n02201497": "anopheline",
  "n02201626": "malarial mosquito, malaria mosquito",
  "n02201758": "Culex, genus Culex",
  "n02202006": "common mosquito, Culex pipiens",
  "n02202124": "Culex quinquefasciatus, Culex fatigans",
  "n02202287": "gnat",
  "n02202509": "Ceratopogonidae, family Ceratopogonidae",
  "n02202678": "punkie, punky, punkey, no-see-um, biting midge",
  "n02202878": "Ceratopogon, genus Ceratopogon",
  "n02203008": "Chironomidae, family Chironomidae",
  "n02203152": "midge",
  "n02203332": "Chironomus, genus Chironomus",
  "n02203457": "Mycetophilidae, family Mycetophylidae",
  "n02203592": "fungus gnat",
  "n02203739": "Psychodidae, family Psychodidae",
  "n02203978": "psychodid",
  "n02204084": "Phlebotomus, genus Phlebotomus",
  "n02204249": "sand fly, sandfly, Phlebotomus papatasii",
  "n02204460": "Sciaridae, family Sciaridae",
  "n02204585": "genus Sciara",
  "n02204722": "fungus gnat, sciara, sciarid",
  "n02204907": "armyworm",
  "n02205095": "Tipulidae, family Tipulidae",
  "n02205219": "crane fly, daddy longlegs",
  "n02205383": "Simuliidae, family Simuliidae",
  "n02205523": "Simulium, genus Simulium",
  "n02205673": "blackfly, black fly, buffalo gnat",
  "n02205896": "Hymenoptera, order Hymenoptera",
  "n02206270": "hymenopterous insect, hymenopteran, hymenopteron, hymenopter",
  "n02206624": "Apoidea, superfamily Apoidea",
  "n02206856": "bee",
  "n02207179": "drone",
  "n02207345": "queen bee",
  "n02207449": "worker",
  "n02207647": "soldier",
  "n02207805": "worker bee",
  "n02207942": "Apidae, family Apidae",
  "n02208143": "Apis, genus Apis",
  "n02208280": "honeybee, Apis mellifera",
  "n02208498": "Africanized bee, Africanized honey bee, killer bee, Apis mellifera scutellata, Apis mellifera adansonii",
  "n02208848": "black bee, German bee",
  "n02208979": "Carniolan bee",
  "n02209111": "Italian bee",
  "n02209230": "Xylocopa, genus Xylocopa",
  "n02209354": "carpenter bee",
  "n02209508": "Bombus, genus Bombus",
  "n02209624": "bumblebee, humblebee",
  "n02209755": "Psithyrus, genus Psithyrus",
  "n02209964": "cuckoo-bumblebee",
  "n02210096": "Andrenidae, family Andrenidae",
  "n02210291": "genus Andrena",
  "n02210427": "andrena, andrenid, mining bee",
  "n02210567": "nomia, genus Nomia",
  "n02210728": "Halictidae, family Halictidae",
  "n02210921": "Nomia melanderi, alkali bee",
  "n02211099": "Megachilidae, family Megachilidae",
  "n02211283": "Megachile, genus Megachile",
  "n02211444": "leaf-cutting bee, leaf-cutter, leaf-cutter bee",
  "n02211627": "mason bee",
  "n02211773": "Anthidium, genus Anthidium",
  "n02211896": "potter bee",
  "n02212062": "wasp",
  "n02212323": "Vespidae, family Vespidae",
  "n02212602": "vespid, vespid wasp",
  "n02212811": "Vespa, genus Vespa",
  "n02212958": "paper wasp",
  "n02213107": "hornet",
  "n02213239": "giant hornet, Vespa crabro",
  "n02213362": "Vespula, genus Vespula",
  "n02213543": "common wasp, Vespula vulgaris",
  "n02213663": "bald-faced hornet, white-faced hornet, Vespula maculata",
  "n02213788": "yellow jacket, yellow hornet, Vespula maculifrons",
  "n02213967": "Polistes, genus Polistes",
  "n02214096": "Polistes annularis",
  "n02214203": "Eumenes, genus Eumenes",
  "n02214341": "mason wasp",
  "n02214499": "potter wasp",
  "n02214660": "Mutillidae, family Mutillidae",
  "n02214773": "velvet ant",
  "n02214972": "Sphecoidea, superfamily Sphecoidea",
  "n02215161": "sphecoid wasp, sphecoid",
  "n02215334": "Sphecidae, family Sphecidae",
  "n02215496": "Sceliphron, genus Sceliphron",
  "n02215621": "mason wasp",
  "n02215770": "digger wasp",
  "n02215941": "Stizidae, family Stizidae",
  "n02216066": "Sphecius, genus Sphecius",
  "n02216211": "cicada killer, Sphecius speciosis",
  "n02216365": "mud dauber",
  "n02216547": "Cynipidae, family Cynipidae",
  "n02216740": "gall wasp, gallfly, cynipid wasp, cynipid gall wasp",
  "n02216924": "Cynips, genus Cynips",
  "n02217050": "Amphibolips, genus Amphibolips",
  "n02217201": "Andricus, genus Andricus",
  "n02217334": "Chalcididae, family Chalcididae, Chalcidae, family Chalcidae",
  "n02217563": "chalcid fly, chalcidfly, chalcid, chalcid wasp",
  "n02217839": "strawworm, jointworm",
  "n02217997": "Chalcis, genus Chalcis",
  "n02218134": "chalcis fly",
  "n02218235": "Ichneumonidae, family Ichneumonidae",
  "n02218371": "ichneumon fly",
  "n02218563": "Tenthredinidae, family Tenthredinidae",
  "n02218713": "sawfly",
  "n02218912": "Fenusa, genus-Fenusa",
  "n02219015": "birch leaf miner, Fenusa pusilla",
  "n02219234": "Formicidae, family Formicidae",
  "n02219486": "ant, emmet, pismire",
  "n02219901": "Monomorium, genus Monomorium",
  "n02220055": "pharaoh ant, pharaoh's ant, Monomorium pharaonis",
  "n02220225": "little black ant, Monomorium minimum",
  "n02220393": "Dorylinae, subfamily Dorylinae",
  "n02220518": "army ant, driver ant, legionary ant",
  "n02220676": "Camponotus, genus Camponotus",
  "n02220804": "carpenter ant",
  "n02220960": "Solenopsis, genus Solenopsis",
  "n02221083": "fire ant",
  "n02221240": "Formica, genus Formica",
  "n02221414": "wood ant, Formica rufa",
  "n02221571": "slave ant",
  "n02221715": "Formica fusca",
  "n02221820": "slave-making ant, slave-maker",
  "n02222035": "sanguinary ant, Formica sanguinea",
  "n02222199": "Myrmecia, genus Myrmecia",
  "n02222321": "bulldog ant",
  "n02222459": "Polyergus, genus Polyergus",
  "n02222582": "Amazon ant, Polyergus rufescens",
  "n02222718": "Isoptera, order Isoptera",
  "n02223009": "Termitidae, family Termitidae",
  "n02223151": "Termes, genus Termes",
  "n02223266": "termite, white ant",
  "n02223520": "dry-wood termite",
  "n02223694": "Reticulitermes, genus Reticulitermes",
  "n02223901": "Reticulitermes flanipes",
  "n02224023": "Reticulitermes lucifugus",
  "n02224141": "Rhinotermitidae, family Rhinotermitidae",
  "n02224323": "Mastotermitidae, family Mastotermitidae",
  "n02224466": "Mastotermes, genus Mastotermes",
  "n02224713": "Mastotermes darwiniensis",
  "n02224937": "Mastotermes electromexicus",
  "n02225081": "Mastotermes electrodominicus",
  "n02225231": "Kalotermitidae, family Kalotermitidae",
  "n02225407": "Kalotermes, genus Kalotermes",
  "n02225577": "Cryptotermes, genus Cryptotermes",
  "n02225798": "powder-post termite, Cryptotermes brevis",
  "n02225959": "Orthoptera, order Orthoptera",
  "n02226183": "orthopterous insect, orthopteron, orthopteran",
  "n02226429": "grasshopper, hopper",
  "n02226598": "Acrididae, family Acrididae, Locustidae, family Locustidae",
  "n02226821": "short-horned grasshopper, acridid",
  "n02226970": "locust",
  "n02227119": "Locusta, genus Locusta",
  "n02227247": "migratory locust, Locusta migratoria",
  "n02227430": "Melanoplus, genus Melanoplus",
  "n02227604": "migratory grasshopper",
  "n02227773": "Tettigoniidae, family Tettigoniidae",
  "n02227966": "long-horned grasshopper, tettigoniid",
  "n02228215": "Microcentrum, genus Microcentrum",
  "n02228341": "katydid",
  "n02228565": "Anabrus, genus Anabrus",
  "n02228697": "mormon cricket, Anabrus simplex",
  "n02228874": "Stenopelmatidae, family Stenopelmatidae",
  "n02229023": "Stenopelmatus, genus Stenopelmatus",
  "n02229156": "sand cricket, Jerusalem cricket, Stenopelmatus fuscus",
  "n02229385": "Gryllidae, family Gryllidae",
  "n02229544": "cricket",
  "n02229765": "mole cricket",
  "n02229867": "Acheta, genus Acheta",
  "n02230023": "European house cricket, Acheta domestica",
  "n02230187": "field cricket, Acheta assimilis",
  "n02230355": "Oecanthus, genus Oecanthus",
  "n02230480": "tree cricket",
  "n02230634": "snowy tree cricket, Oecanthus fultoni",
  "n02230782": "Phasmida, order Phasmida, Phasmatodea, order Phasmatodea",
  "n02231052": "phasmid, phasmid insect",
  "n02231307": "Phasmidae, family Phasmidae, Phasmatidae, family Phasmatidae",
  "n02231487": "walking stick, walkingstick, stick insect",
  "n02231680": "genus Diapheromera",
  "n02231803": "diapheromera, Diapheromera femorata",
  "n02231930": "Phyllidae, family Phyllidae, Phillidae, family Phillidae",
  "n02232086": "Phyllium, genus Phyllium",
  "n02232223": "walking leaf, leaf insect",
  "n02232408": "Exopterygota, subclass Exopterygota, Hemimetabola",
  "n02232606": "Dictyoptera, order Dictyoptera",
  "n02232951": "dictyopterous insect",
  "n02233096": "Blattodea, suborder Blattodea, Blattaria, suborder Blattaria",
  "n02233338": "cockroach, roach",
  "n02233577": "Blattidae, family Blattidae",
  "n02233767": "Blatta, genus Blatta",
  "n02233943": "oriental cockroach, oriental roach, Asiatic cockroach, blackbeetle, Blatta orientalis",
  "n02234181": "Periplaneta, genus Periplaneta",
  "n02234355": "American cockroach, Periplaneta americana",
  "n02234570": "Australian cockroach, Periplaneta australasiae",
  "n02234719": "Blattella, genus Blattella",
  "n02234848": "German cockroach, Croton bug, crotonbug, water bug, Blattella germanica",
  "n02235078": "Blaberus, genus Blaberus",
  "n02235205": "giant cockroach",
  "n02235321": "Cryptocercidae, family Cryptocercidae",
  "n02235465": "Cryptocercus, genus Cryptocercus",
  "n02235575": "Manteodea, suborder Manteodea",
  "n02235761": "Mantidae, family Mantidae, Manteidae, family Manteidae",
  "n02235911": "genus Mantis",
  "n02236044": "mantis, mantid",
  "n02236241": "praying mantis, praying mantid, Mantis religioso",
  "n02236355": "bug",
  "n02236495": "Hemiptera, order Hemiptera",
  "n02236896": "hemipterous insect, bug, hemipteran, hemipteron",
  "n02237239": "Miridae, family Miridae, Capsidae, family Capsidae",
  "n02237424": "leaf bug, plant bug",
  "n02237581": "mirid bug, mirid, capsid",
  "n02237730": "Poecilocapsus, genus Poecilocapsus",
  "n02237868": "four-lined plant bug, four-lined leaf bug, Poecilocapsus lineatus",
  "n02238113": "Lygus, genus Lygus",
  "n02238235": "lygus bug",
  "n02238358": "tarnished plant bug, Lygus lineolaris",
  "n02238474": "Tingidae, family Tingidae",
  "n02238594": "lace bug",
  "n02238743": "Lygaeidae, family Lygaeidae",
  "n02238887": "lygaeid, lygaeid bug",
  "n02239073": "Blissus, genus Blissus",
  "n02239192": "chinch bug, Blissus leucopterus",
  "n02239347": "Coreidae, family Coreidae",
  "n02239528": "coreid bug, coreid",
  "n02239659": "Anasa, genus Anasa",
  "n02239774": "squash bug, Anasa tristis",
  "n02239934": "Leptoglossus, genus Leptoglossus",
  "n02240068": "leaf-footed bug, leaf-foot bug",
  "n02240223": "Cimicidae, family Cimicidae",
  "n02240377": "Cimex, genus Cimex",
  "n02240517": "bedbug, bed bug, chinch, Cimex lectularius",
  "n02240706": "Notonectidae, family Notonectidae",
  "n02240852": "Notonecta, genus Notonecta",
  "n02241008": "backswimmer, Notonecta undulata",
  "n02241184": "Heteroptera, suborder Heteroptera",
  "n02241426": "true bug",
  "n02241569": "heteropterous insect",
  "n02241799": "water bug",
  "n02242004": "Belostomatidae, family Belostomatidae",
  "n02242137": "giant water bug",
  "n02242293": "Nepidae, family Nepidae",
  "n02242455": "water scorpion",
  "n02242669": "Nepa, genus Nepa",
  "n02242816": "Ranatra, genus Ranatra",
  "n02242942": "Corixidae, family Corixidae",
  "n02243065": "Corixa, genus Corixa",
  "n02243209": "water boatman, boat bug",
  "n02243351": "Gerrididae, family Gerrididae, Gerridae, family Gerridae",
  "n02243562": "water strider, pond-skater, water skater",
  "n02243744": "Gerris, genus Gerris",
  "n02243878": "common pond-skater, Gerris lacustris",
  "n02244007": "Reduviidae, family Reduviidae",
  "n02244173": "assassin bug, reduviid",
  "n02244396": "Triatoma, genus Triatoma",
  "n02244515": "conenose, cone-nosed bug, conenose bug, big bedbug, kissing bug",
  "n02244670": "Arilus, genus Arilus",
  "n02244797": "wheel bug, Arilus cristatus",
  "n02244963": "Pyrrhocoridae, family Pyrrhocoridae",
  "n02245111": "firebug",
  "n02245239": "Dysdercus, genus Dysdercus",
  "n02245443": "cotton stainer",
  "n02245592": "Homoptera, suborder Homoptera",
  "n02246011": "homopterous insect, homopteran",
  "n02246284": "Aleyrodidae, family Aleyrodidae",
  "n02246487": "Aleyrodes, genus Aleyrodes",
  "n02246628": "whitefly",
  "n02246822": "Dialeurodes, genus Dialeurodes",
  "n02246941": "citrus whitefly, Dialeurodes citri",
  "n02247076": "Trialeurodes, genus Trialeurodes",
  "n02247216": "greenhouse whitefly, Trialeurodes vaporariorum",
  "n02247363": "Bemisia, genus Bemisia",
  "n02247511": "sweet-potato whitefly",
  "n02247655": "superbug, Bemisia tabaci, poinsettia strain",
  "n02247950": "superbug",
  "n02248062": "cotton strain",
  "n02248147": "Coccoidea, superfamily Coccoidea",
  "n02248368": "coccid insect",
  "n02248510": "scale insect",
  "n02248744": "Coccidae, family Coccidae",
  "n02248887": "soft scale",
  "n02249011": "genus Coccus",
  "n02249134": "brown soft scale, Coccus hesperidum",
  "n02249256": "wax insect",
  "n02249365": "Diaspididae, family Diaspididae",
  "n02249515": "armored scale",
  "n02249673": "Aspidiotus, genus Aspidiotus",
  "n02249809": "San Jose scale, Aspidiotus perniciosus",
  "n02249995": "Dactylopiidae, family Dactylopiidae",
  "n02250133": "Dactylopius, genus Dactylopius",
  "n02250280": "cochineal insect, cochineal, Dactylopius coccus",
  "n02250464": "Pseudococcidae, family Pseudococcidae",
  "n02250653": "Pseudococcus, genus Pseudococcus",
  "n02250822": "mealybug, mealy bug",
  "n02251067": "citrophilous mealybug, citrophilus mealybug, Pseudococcus fragilis",
  "n02251233": "Comstock mealybug, Comstock's mealybug, Pseudococcus comstocki",
  "n02251452": "Planococcus, genus Planococcus",
  "n02251593": "citrus mealybug, Planococcus citri",
  "n02251775": "plant louse, louse",
  "n02252039": "Aphidoidea, superfamily Aphidoidea",
  "n02252226": "aphid",
  "n02252429": "Aphididae, family Aphididae",
  "n02252608": "Aphis, genus Aphis",
  "n02252799": "apple aphid, green apple aphid, Aphis pomi",
  "n02252972": "blackfly, bean aphid, Aphis fabae",
  "n02253127": "greenfly",
  "n02253264": "green peach aphid",
  "n02253391": "pale chrysanthemum aphid",
  "n02253494": "ant cow",
  "n02253592": "Eriosoma, genus Eriosoma",
  "n02253715": "woolly aphid, woolly plant louse",
  "n02253913": "woolly apple aphid, American blight, Eriosoma lanigerum",
  "n02254110": "Prociphilus, genus Prociphilus",
  "n02254246": "woolly alder aphid, Prociphilus tessellatus",
  "n02254370": "Adelgidae, family Adelgidae",
  "n02254531": "Adelges, genus Adelges",
  "n02254697": "adelgid",
  "n02254901": "balsam woolly aphid, Adelges piceae",
  "n02255023": "spruce gall aphid, Adelges abietis",
  "n02255144": "Pineus, genus Pineus",
  "n02255270": "pine leaf aphid, Pineus pinifoliae",
  "n02255391": "woolly adelgid",
  "n02255567": "Phylloxeridae, family Phylloxeridae",
  "n02255698": "Phylloxera, genus Phylloxera",
  "n02255855": "grape louse, grape phylloxera, Phylloxera vitifoleae",
  "n02256010": "Psyllidae, family Psyllidae, Chermidae, family Chermidae",
  "n02256172": "jumping plant louse, psylla, psyllid",
  "n02256365": "Cicadidae, family Cicadidae",
  "n02256542": "genus Cicada",
  "n02256656": "cicada, cicala",
  "n02256882": "Tibicen, genus Tibicen",
  "n02257003": "dog-day cicada, harvest fly",
  "n02257149": "Magicicada, genus Magicicada",
  "n02257284": "seventeen-year locust, periodical cicada, Magicicada septendecim",
  "n02257536": "Cercopidae, family Cercopidae",
  "n02257715": "spittle insect, spittlebug",
  "n02257985": "froghopper",
  "n02258065": "Philaenus, genus Philaenus",
  "n02258198": "meadow spittlebug, Philaenus spumarius",
  "n02258354": "Aphrophora, genus Aphrophora",
  "n02258508": "pine spittlebug",
  "n02258629": "Saratoga spittlebug, Aphrophora saratogensis",
  "n02258780": "Cicadellidae, family Cicadellidae",
  "n02258910": "Jassidae, family Jassidae",
  "n02259136": "jassid",
  "n02259212": "leafhopper",
  "n02259377": "plant hopper, planthopper",
  "n02259565": "Membracidae, family Membracidae",
  "n02259708": "treehopper",
  "n02259844": "Fulgoridae, family Fulgoridae",
  "n02259987": "lantern fly, lantern-fly",
  "n02260183": "Psocoptera, order Psocoptera, Corrodentia, order Corrodentia",
  "n02260421": "psocopterous insect",
  "n02260623": "Psocidae, family Psocidae",
  "n02260863": "psocid",
  "n02261063": "bark-louse, bark louse",
  "n02261184": "Atropidae, family Atropidae",
  "n02261286": "Liposcelis, genus Liposcelis",
  "n02261419": "booklouse, book louse, deathwatch, Liposcelis divinatorius",
  "n02261630": "Trogium, genus Trogium",
  "n02261757": "common booklouse, Trogium pulsatorium",
  "n02261883": "Ephemeroptera, order Ephemeroptera, Ephemerida, order Ephemerida",
  "n02262062": "Plectophera",
  "n02262178": "ephemerid, ephemeropteran",
  "n02262324": "Ephemeridae, family Ephemeridae",
  "n02262449": "mayfly, dayfly, shadfly",
  "n02262679": "Plecoptera, order Plecoptera",
  "n02262803": "stonefly, stone fly, plecopteran",
  "n02263038": "Neuroptera, order Neuroptera",
  "n02263378": "neuropteron, neuropteran, neuropterous insect",
  "n02263717": "Myrmeleontidae, family Myrmeleontidae",
  "n02263848": "Myrmeleon, genus Myrmeleon",
  "n02264021": "ant lion, antlion, antlion fly",
  "n02264232": "doodlebug, ant lion, antlion",
  "n02264363": "lacewing, lacewing fly",
  "n02264591": "aphid lion, aphis lion",
  "n02264734": "Chrysopidae, family Chrysopidae",
  "n02264885": "green lacewing, chrysopid, stink fly",
  "n02265076": "goldeneye, golden-eyed fly",
  "n02265177": "Hemerobiidae, family Hemerobiidae",
  "n02265330": "brown lacewing, hemerobiid, hemerobiid fly",
  "n02265471": "Megaloptera, suborder Megaloptera",
  "n02265717": "Corydalidae, family Corydalidae",
  "n02265860": "Corydalus, genus Corydalus, Corydalis, genus Corydalis",
  "n02266050": "dobson, dobsonfly, dobson fly, Corydalus cornutus",
  "n02266269": "hellgrammiate, dobson",
  "n02266421": "fish fly, fish-fly",
  "n02266580": "Sialidae, family Sialidae",
  "n02266732": "Sialis, genus Sialis",
  "n02266864": "alderfly, alder fly, Sialis lutaria",
  "n02267019": "Raphidiidae, family Raphidiidae",
  "n02267208": "snakefly",
  "n02267356": "Mantispidae, family Mantispidae",
  "n02267483": "mantispid",
  "n02267644": "Sisyridae, family Sisyridae",
  "n02267826": "spongefly, spongillafly",
  "n02267975": "Odonata, order Odonata",
  "n02268148": "odonate",
  "n02268334": "Anisoptera, suborder Anisoptera",
  "n02268443": "dragonfly, darning needle, devil's darning needle, sewing needle, snake feeder, snake doctor, mosquito hawk, skeeter hawk",
  "n02268746": "Zygoptera, suborder Zygoptera",
  "n02268853": "damselfly",
  "n02269015": "Trichoptera, order Trichoptera",
  "n02269196": "trichopterous insect, trichopteran, trichopteron",
  "n02269340": "caddis fly, caddis-fly, caddice fly, caddice-fly",
  "n02269522": "caseworm",
  "n02269657": "caddisworm, strawworm",
  "n02269829": "Thysanura, order Thysanura",
  "n02270011": "thysanuran insect, thysanuron",
  "n02270200": "bristletail",
  "n02270326": "Lepismatidae, family Lepismatidae",
  "n02270473": "Lepisma, genus Lepisma",
  "n02270623": "silverfish, Lepisma saccharina",
  "n02270810": "Thermobia, genus Thermobia",
  "n02270945": "firebrat, Thermobia domestica",
  "n02271087": "Machilidae, family Machilidae",
  "n02271222": "jumping bristletail, machilid",
  "n02271427": "Thysanoptera, order Thysanoptera",
  "n02271570": "thysanopter, thysanopteron, thysanopterous insect",
  "n02271740": "Thripidae, family Thripidae",
  "n02271897": "thrips, thrip, thripid",
  "n02272152": "Frankliniella, genus Frankliniella",
  "n02272286": "tobacco thrips, Frankliniella fusca",
  "n02272428": "genus Thrips",
  "n02272552": "onion thrips, onion louse, Thrips tobaci",
  "n02272707": "Dermaptera, order Dermaptera",
  "n02272871": "earwig",
  "n02273120": "Forficulidae, family Forficulidae",
  "n02273254": "Forficula, genus Forficula",
  "n02273392": "common European earwig, Forficula auricularia",
  "n02273545": "Lepidoptera, order Lepidoptera",
  "n02274024": "lepidopterous insect, lepidopteron, lepidopteran",
  "n02274259": "butterfly",
  "n02274516": "Nymphalidae, family Nymphalidae",
  "n02274822": "nymphalid, nymphalid butterfly, brush-footed butterfly, four-footed butterfly",
  "n02275372": "Nymphalis, genus Nymphalis",
  "n02275560": "mourning cloak, mourning cloak butterfly, Camberwell beauty, Nymphalis antiopa",
  "n02275773": "tortoiseshell, tortoiseshell butterfly",
  "n02275921": "Vanessa, genus Vanessa",
  "n02276078": "painted beauty, Vanessa virginiensis",
  "n02276258": "admiral",
  "n02276355": "red admiral, Vanessa atalanta",
  "n02276527": "Limenitis, genus Limenitis",
  "n02276749": "white admiral, Limenitis camilla",
  "n02276902": "banded purple, white admiral, Limenitis arthemis",
  "n02277094": "red-spotted purple, Limenitis astyanax",
  "n02277268": "viceroy, Limenitis archippus",
  "n02277422": "anglewing",
  "n02277556": "Satyridae, family Satyridae",
  "n02277742": "ringlet, ringlet butterfly",
  "n02277895": "Polygonia, genus Polygonia",
  "n02278024": "comma, comma butterfly, Polygonia comma",
  "n02278210": "fritillary",
  "n02278343": "Spyeria, genus Spyeria",
  "n02278463": "silverspot",
  "n02278601": "Argynnis, genus Argynnis",
  "n02278704": "Apatura, genus Apatura",
  "n02278839": "emperor butterfly, emperor",
  "n02278980": "purple emperor, Apatura iris",
  "n02279127": "Inachis, genus Inachis",
  "n02279257": "peacock, peacock butterfly, Inachis io",
  "n02279442": "Danaidae, family Danaidae",
  "n02279637": "danaid, danaid butterfly",
  "n02279819": "Danaus, genus Danaus",
  "n02279972": "monarch, monarch butterfly, milkweed butterfly, Danaus plexippus",
  "n02280223": "Pieridae, family Pieridae",
  "n02280458": "pierid, pierid butterfly",
  "n02280649": "cabbage butterfly",
  "n02280845": "Pieris, genus Pieris",
  "n02281015": "small white, Pieris rapae",
  "n02281136": "large white, Pieris brassicae",
  "n02281267": "southern cabbage butterfly, Pieris protodice",
  "n02281406": "sulphur butterfly, sulfur butterfly",
  "n02281552": "Lycaenidae, family Lycaenidae",
  "n02281787": "lycaenid, lycaenid butterfly",
  "n02281987": "Lycaena, genus Lycaena",
  "n02282257": "blue",
  "n02282385": "copper",
  "n02282553": "American copper, Lycaena hypophlaeas",
  "n02282716": "Strymon, genus Strymon",
  "n02282903": "hairstreak, hairstreak butterfly",
  "n02283077": "Strymon melinus",
  "n02283201": "moth",
  "n02283617": "moth miller, miller",
  "n02283728": "Tortricidae, family Tortricidae",
  "n02283951": "tortricid, tortricid moth",
  "n02284224": "leaf roller, leaf-roller",
  "n02284367": "genus Tortrix",
  "n02284513": "Homona, genus Homona",
  "n02284611": "tea tortrix, tortrix, Homona coffearia",
  "n02284771": "Argyrotaenia, genus Argyrotaenia",
  "n02284884": "orange tortrix, tortrix, Argyrotaenia citrana",
  "n02285052": "Carpocapsa, genus Carpocapsa",
  "n02285179": "codling moth, codlin moth, Carpocapsa pomonella",
  "n02285359": "Lymantriidae, family Lymantriidae",
  "n02285548": "lymantriid, tussock moth",
  "n02285801": "tussock caterpillar",
  "n02285909": "Lymantria, genus Lymantria",
  "n02286089": "gypsy moth, gipsy moth, Lymantria dispar",
  "n02286271": "Euproctis, genus Euproctis",
  "n02286425": "browntail, brown-tail moth, Euproctis phaeorrhoea",
  "n02286654": "gold-tail moth, Euproctis chrysorrhoea",
  "n02286815": "Geometridae, family Geometridae",
  "n02287004": "geometrid, geometrid moth",
  "n02287204": "Paleacrita, genus Paleacrita",
  "n02287352": "Paleacrita vernata",
  "n02287476": "Alsophila, genus Alsophila",
  "n02287622": "Alsophila pometaria",
  "n02287799": "cankerworm",
  "n02287987": "spring cankerworm",
  "n02288122": "fall cankerworm",
  "n02288268": "measuring worm, inchworm, looper",
  "n02288473": "Pyralidae, family Pyralidae, Pyralididae, family Pyralididae",
  "n02288789": "pyralid, pyralid moth",
  "n02289061": "Pyralis, genus Pyralis",
  "n02289177": "Galleria, genus Galleria",
  "n02289307": "bee moth, wax moth, Galleria mellonella",
  "n02289466": "Pyrausta, genus Pyrausta",
  "n02289610": "corn borer, European corn borer moth, corn borer moth, Pyrausta nubilalis",
  "n02289845": "Anagasta, genus Anagasta",
  "n02289988": "Mediterranean flour moth, Anagasta kuehniella",
  "n02290153": "Ephestia, genus Ephestia",
  "n02290340": "tobacco moth, cacao moth, Ephestia elutella",
  "n02290521": "Cadra, genus Cadra",
  "n02290664": "almond moth, fig moth, Cadra cautella",
  "n02290870": "raisin moth, Cadra figulilella",
  "n02291024": "Tineoidea, superfamily Tineoidea",
  "n02291220": "tineoid, tineoid moth",
  "n02291391": "Tineidae, family Tineidae",
  "n02291572": "tineid, tineid moth",
  "n02291748": "clothes moth",
  "n02291940": "Tinea, genus Tinea",
  "n02292085": "casemaking clothes moth, Tinea pellionella",
  "n02292272": "Tineola, genus Tineola",
  "n02292401": "webbing clothes moth, webbing moth, Tineola bisselliella",
  "n02292564": "Trichophaga, genus Trichophaga",
  "n02292692": "carpet moth, tapestry moth, Trichophaga tapetzella",
  "n02292850": "Gracilariidae, Gracillariidae, family Gracilariidae",
  "n02292980": "gracilariid, gracilariid moth",
  "n02293135": "Gelechiidae, family Gelechiidae",
  "n02293352": "gelechiid, gelechiid moth",
  "n02293560": "Gelechia, genus Gelechia",
  "n02293715": "Gelechia gossypiella",
  "n02293868": "grain moth",
  "n02293974": "Sitotroga, genus Sitotroga",
  "n02294097": "angoumois moth, angoumois grain moth, Sitotroga cerealella",
  "n02294279": "Phthorimaea, genus Phthorimaea",
  "n02294407": "potato moth, potato tuber moth, splitworm, Phthorimaea operculella",
  "n02294577": "potato tuberworm, Phthorimaea operculella",
  "n02294761": "Noctuidae, family Noctuidae",
  "n02295064": "noctuid moth, noctuid, owlet moth",
  "n02295390": "cutworm",
  "n02295570": "Noctua, genus Noctua",
  "n02295717": "Catacala, genus Catacala",
  "n02295870": "underwing",
  "n02296021": "red underwing, Catocala nupta",
  "n02296150": "Cerapteryx, genus Cerapteryx",
  "n02296276": "antler moth, Cerapteryx graminis",
  "n02296480": "Heliothis, genus Heliothis",
  "n02296612": "heliothis moth, Heliothis zia",
  "n02296756": "Chorizagrotis, genus Chorizagrotis",
  "n02296912": "army cutworm, Chorizagrotis auxiliaris",
  "n02297127": "Pseudaletia, genus Pseudaletia",
  "n02297294": "armyworm, Pseudaletia unipuncta",
  "n02297442": "armyworm, army worm, Pseudaletia unipuncta",
  "n02297635": "Spodoptera, genus Spodoptera",
  "n02297819": "Spodoptera exigua",
  "n02297938": "beet armyworm, Spodoptera exigua",
  "n02298095": "Spodoptera frugiperda",
  "n02298218": "fall armyworm, Spodoptera frugiperda",
  "n02298379": "Sphingidae, family Sphingidae",
  "n02298541": "hawkmoth, hawk moth, sphingid, sphinx moth, hummingbird moth",
  "n02298833": "Manduca, genus Manduca",
  "n02299039": "Manduca sexta",
  "n02299157": "tobacco hornworm, tomato worm, Manduca sexta",
  "n02299378": "Manduca quinquemaculata",
  "n02299505": "tomato hornworm, potato worm, Manduca quinquemaculata",
  "n02299715": "Acherontia, genus Acherontia",
  "n02299846": "death's-head moth, Acherontia atropos",
  "n02300018": "Bombycidae, family Bombycidae",
  "n02300173": "bombycid, bombycid moth, silkworm moth",
  "n02300378": "Bombyx, genus Bombyx",
  "n02300554": "domestic silkworm moth, domesticated silkworm moth, Bombyx mori",
  "n02300797": "silkworm",
  "n02301072": "Saturniidae, family Saturniidae",
  "n02301452": "saturniid, saturniid moth",
  "n02301782": "Saturnia, genus Saturnia",
  "n02301935": "emperor, emperor moth, Saturnia pavonia",
  "n02302124": "Eacles, genus Eacles",
  "n02302244": "imperial moth, Eacles imperialis",
  "n02302459": "giant silkworm moth, silkworm moth",
  "n02302620": "silkworm, giant silkworm, wild wilkworm",
  "n02302853": "Actias, genus Actias",
  "n02302969": "luna moth, Actias luna",
  "n02303167": "Hyalophora, genus Hyalophora",
  "n02303284": "cecropia, cecropia moth, Hyalophora cecropia",
  "n02303448": "Samia, genus Samia",
  "n02303585": "cynthia moth, Samia cynthia, Samia walkeri",
  "n02303777": "ailanthus silkworm, Samia cynthia",
  "n02303917": "Automeris, genus Automeris",
  "n02304036": "io moth, Automeris io",
  "n02304229": "Antheraea, genus Antheraea",
  "n02304432": "polyphemus moth, Antheraea polyphemus",
  "n02304657": "pernyi moth, Antheraea pernyi",
  "n02304797": "tussah, tusseh, tussur, tussore, tusser, Antheraea mylitta",
  "n02304967": "Atticus, genus Atticus",
  "n02305085": "atlas moth, Atticus atlas",
  "n02305245": "Arctiidae, family Arctiidae",
  "n02305407": "arctiid, arctiid moth",
  "n02305636": "tiger moth",
  "n02305799": "Callimorpha, genus Callimorpha",
  "n02305929": "cinnabar, cinnabar moth, Callimorpha jacobeae",
  "n02306159": "Lasiocampidae, family Lasiocampidae",
  "n02306433": "lasiocampid, lasiocampid moth",
  "n02306672": "Lasiocampa, genus Lasiocampa",
  "n02306825": "eggar, egger",
  "n02307007": "Malacosoma, genus Malacosoma",
  "n02307176": "tent-caterpillar moth, Malacosoma americana",
  "n02307325": "tent caterpillar",
  "n02307515": "tent-caterpillar moth, Malacosoma disstria",
  "n02307681": "forest tent caterpillar, Malacosoma disstria",
  "n02307910": "lappet, lappet moth",
  "n02308033": "lappet caterpillar",
  "n02308139": "webworm",
  "n02308325": "Hyphantria, genus Hyphantria",
  "n02308471": "webworm moth",
  "n02308618": "Hyphantria cunea",
  "n02308735": "fall webworm, Hyphantria cunea",
  "n02308852": "Loxostege, genus Loxostege",
  "n02308998": "Loxostege similalis",
  "n02309120": "garden webworm, Loxostege similalis",
  "n02309242": "instar",
  "n02309337": "caterpillar",
  "n02309841": "corn borer, Pyrausta nubilalis",
  "n02310000": "bollworm",
  "n02310149": "pink bollworm, Gelechia gossypiella",
  "n02310334": "corn earworm, cotton bollworm, tomato fruitworm, tobacco budworm, vetchworm, Heliothis zia",
  "n02310585": "cabbageworm, Pieris rapae",
  "n02310717": "woolly bear, woolly bear caterpillar",
  "n02310941": "woolly bear moth",
  "n02311060": "larva",
  "n02311617": "nymph",
  "n02311748": "leptocephalus",
  "n02311879": "bot",
  "n02312006": "grub",
  "n02312175": "maggot",
  "n02312325": "leatherjacket",
  "n02312427": "pupa",
  "n02312640": "chrysalis",
  "n02312744": "cocoon",
  "n02312912": "imago",
  "n02313008": "queen",
  "n02313195": "Phoronida, Phoronidea, phylum Phoronida",
  "n02313360": "phoronid",
  "n02313495": "Bryozoa, phylum Bryozoa, polyzoa",
  "n02313709": "bryozoan, polyzoan, sea mat, sea moss, moss animal",
  "n02314001": "Ectoprocta, phylum Ectoprocta",
  "n02314158": "ectoproct",
  "n02314321": "Entoprocta, phylum Entoprocta, Endoprocta",
  "n02314492": "entoproct",
  "n02314717": "Cycliophora, phylum Cycliophora",
  "n02315024": "Symbion pandora",
  "n02315309": "Brachiopoda, phylum Brachiopoda",
  "n02315487": "brachiopod, lamp shell, lampshell",
  "n02315696": "Sipuncula, phylum Sipuncula",
  "n02315821": "peanut worm, sipunculid",
  "n02316038": "echinoderm family",
  "n02316180": "echinoderm genus",
  "n02316392": "Echinodermata, phylum Echinodermata",
  "n02316707": "echinoderm",
  "n02317025": "ambulacrum",
  "n02317212": "Asteroidea, class Asteroidea",
  "n02317335": "starfish, sea star",
  "n02317488": "Ophiuroidea, class Ophiuroidea",
  "n02317653": "Ophiurida, subclass Ophiurida",
  "n02317781": "brittle star, brittle-star, serpent star",
  "n02317983": "Euryalida, subclass Euryalida",
  "n02318167": "basket star, basket fish",
  "n02318437": "Euryale, genus Euryale",
  "n02318538": "Astrophyton, genus Astrophyton",
  "n02318687": "Astrophyton muricatum",
  "n02318798": "Gorgonocephalus, genus Gorgonocephalus",
  "n02318915": "Echinoidea, class Echinoidea",
  "n02319095": "sea urchin",
  "n02319308": "edible sea urchin, Echinus esculentus",
  "n02319423": "Exocycloida, order Exocycloida",
  "n02319555": "sand dollar",
  "n02319689": "Spatangoida, order Spatangoida",
  "n02319829": "heart urchin",
  "n02319967": "Crinoidea, class Crinoidea",
  "n02320127": "crinoid",
  "n02320339": "Ptilocrinus, genus Ptilocrinus",
  "n02320465": "sea lily",
  "n02320621": "Antedonidae, family Antedonidae",
  "n02320769": "Comatulidae, family Comatulidae",
  "n02320888": "Antedon, genus Antedon",
  "n02321062": "Comatula, genus Comatula",
  "n02321170": "feather star, comatulid",
  "n02321342": "Holothuroidea, class Holothuroidea",
  "n02321529": "sea cucumber, holothurian",
  "n02321759": "Holothuridae, family Holothuridae",
  "n02321903": "Holothuria, genus Holothuria",
  "n02322047": "trepang, Holothuria edulis",
  "n02322213": "foot, invertebrate foot",
  "n02322387": "tube foot",
  "n02322624": "roe",
  "n02322712": "milt",
  "n02322817": "splint bone",
  "n02322992": "Duplicidentata",
  "n02323186": "Lagomorpha, order Lagomorpha",
  "n02323449": "lagomorph, gnawing mammal",
  "n02323715": "Leporidae, family Leporidae",
  "n02323902": "leporid, leporid mammal",
  "n02324045": "rabbit, coney, cony",
  "n02324431": "rabbit ears",
  "n02324514": "lapin",
  "n02324587": "bunny, bunny rabbit",
  "n02324717": "Oryctolagus, genus Oryctolagus",
  "n02324850": "European rabbit, Old World rabbit, Oryctolagus cuniculus",
  "n02325211": "Sylvilagus, genus Sylvilagus",
  "n02325366": "wood rabbit, cottontail, cottontail rabbit",
  "n02325722": "eastern cottontail, Sylvilagus floridanus",
  "n02325884": "swamp rabbit, canecutter, swamp hare, Sylvilagus aquaticus",
  "n02326074": "marsh hare, swamp rabbit, Sylvilagus palustris",
  "n02326237": "Lepus, genus Lepus",
  "n02326432": "hare",
  "n02326763": "leveret",
  "n02326862": "European hare, Lepus europaeus",
  "n02327028": "jackrabbit",
  "n02327175": "white-tailed jackrabbit, whitetail jackrabbit, Lepus townsendi",
  "n02327435": "blacktail jackrabbit, Lepus californicus",
  "n02327656": "polar hare, Arctic hare, Lepus arcticus",
  "n02327842": "snowshoe hare, snowshoe rabbit, varying hare, Lepus americanus",
  "n02328009": "Belgian hare, leporide",
  "n02328150": "Angora, Angora rabbit",
  "n02328270": "Ochotonidae, family Ochotonidae",
  "n02328429": "pika, mouse hare, rock rabbit, coney, cony",
  "n02328662": "Ochotona, genus Ochotona",
  "n02328820": "little chief hare, Ochotona princeps",
  "n02328942": "collared pika, Ochotona collaris",
  "n02329093": "Rodentia, order Rodentia",
  "n02329401": "rodent, gnawer",
  "n02330245": "mouse",
  "n02330582": "Myomorpha, suborder Myomorpha",
  "n02330830": "Muroidea, superfamily Muroidea",
  "n02331046": "rat",
  "n02331309": "pocket rat",
  "n02331479": "Muridae, family Muridae",
  "n02331842": "murine",
  "n02331960": "Mus, genus Mus",
  "n02332156": "house mouse, Mus musculus",
  "n02332315": "Micromyx, genus Micromyx",
  "n02332447": "harvest mouse, Micromyx minutus",
  "n02332606": "Apodemus, genus Apodemus",
  "n02332755": "field mouse, fieldmouse",
  "n02332954": "nude mouse",
  "n02333190": "European wood mouse, Apodemus sylvaticus",
  "n02333368": "Rattus, genus Rattus",
  "n02333546": "brown rat, Norway rat, Rattus norvegicus",
  "n02333733": "wharf rat",
  "n02333819": "sewer rat",
  "n02333909": "black rat, roof rat, Rattus rattus",
  "n02334079": "Nesokia, genus Nesokia",
  "n02334201": "bandicoot rat, mole rat",
  "n02334337": "Conilurus, genus Conilurus",
  "n02334460": "jerboa rat",
  "n02334609": "Notomys, genus Notomys",
  "n02334728": "kangaroo mouse",
  "n02334849": "Hydromyinae, subfamily Hydromyinae",
  "n02335007": "Hydromys, genus Hydromys",
  "n02335127": "water rat",
  "n02335231": "beaver rat",
  "n02335349": "Cricetidae, family Cricetidae",
  "n02336011": "New World mouse",
  "n02336129": "Reithrodontomys, genus Reithrodontomys",
  "n02336275": "American harvest mouse, harvest mouse",
  "n02336451": "Peromyscus, genus Peromyscus",
  "n02336641": "wood mouse",
  "n02336826": "white-footed mouse, vesper mouse, Peromyscus leucopus",
  "n02337001": "deer mouse, Peromyscus maniculatus",
  "n02337171": "cactus mouse, Peromyscus eremicus",
  "n02337332": "cotton mouse, Peromyscus gossypinus",
  "n02337480": "Baiomys, genus Baiomys",
  "n02337598": "pygmy mouse, Baiomys taylori",
  "n02337774": "Onychomys, genus Onychomys",
  "n02337902": "grasshopper mouse",
  "n02338029": "Ondatra, genus Ondatra",
  "n02338145": "muskrat, musquash, Ondatra zibethica",
  "n02338319": "Neofiber, genus Neofiber",
  "n02338449": "round-tailed muskrat, Florida water rat, Neofiber alleni",
  "n02338592": "Sigmodon, genus Sigmodon",
  "n02338722": "cotton rat, Sigmodon hispidus",
  "n02338901": "wood rat, wood-rat",
  "n02339282": "dusky-footed wood rat",
  "n02339376": "vole, field mouse",
  "n02339768": "Neotoma, genus Neotoma",
  "n02339922": "packrat, pack rat, trade rat, bushytail woodrat, Neotoma cinerea",
  "n02340186": "dusky-footed woodrat, Neotoma fuscipes",
  "n02340358": "eastern woodrat, Neotoma floridana",
  "n02340521": "Oryzomys, genus Oryzomys",
  "n02340640": "rice rat, Oryzomys palustris",
  "n02340813": "Pitymys, genus Pitymys",
  "n02340930": "pine vole, pine mouse, Pitymys pinetorum",
  "n02341108": "Microtus, genus Microtus",
  "n02341288": "meadow vole, meadow mouse, Microtus pennsylvaticus",
  "n02341475": "water vole, Richardson vole, Microtus richardsoni",
  "n02341616": "prairie vole, Microtus ochrogaster",
  "n02341805": "Arvicola, genus Arvicola",
  "n02341974": "water vole, water rat, Arvicola amphibius",
  "n02342109": "Clethrionomys, genus Clethrionomys",
  "n02342250": "red-backed mouse, redback vole",
  "n02342413": "genus Phenacomys",
  "n02342534": "phenacomys",
  "n02342727": "Cricetus, genus Cricetus",
  "n02342885": "hamster",
  "n02343058": "Eurasian hamster, Cricetus cricetus",
  "n02343187": "Mesocricetus, genus Mesocricetus",
  "n02343320": "golden hamster, Syrian hamster, Mesocricetus auratus",
  "n02343487": "Gerbillinae, subfamily Gerbillinae",
  "n02343633": "Gerbillus, genus Gerbillus",
  "n02343772": "gerbil, gerbille",
  "n02344006": "Meriones, genus Meriones",
  "n02344175": "jird",
  "n02344270": "tamarisk gerbil, Meriones unguiculatus",
  "n02344408": "sand rat, Meriones longifrons",
  "n02344528": "lemming",
  "n02344785": "Lemmus, genus lemmus",
  "n02344918": "European lemming, Lemmus lemmus",
  "n02345078": "brown lemming, Lemmus trimucronatus",
  "n02345213": "Myopus, genus Myopus",
  "n02345340": "grey lemming, gray lemming, red-backed lemming",
  "n02345471": "Dicrostonyx, genus Dicrostonyx",
  "n02345600": "pied lemming",
  "n02345774": "Hudson bay collared lemming, Dicrostonyx hudsonius",
  "n02345890": "Synaptomys, genus Synaptomys",
  "n02345997": "southern bog lemming, Synaptomys cooperi",
  "n02346170": "northern bog lemming, Synaptomys borealis",
  "n02346315": "Hystricomorpha, suborder Hystricomorpha",
  "n02346627": "porcupine, hedgehog",
  "n02346823": "Hystricidae, family Hystricidae",
  "n02346998": "Old World porcupine",
  "n02347140": "Atherurus, genus Atherurus",
  "n02347274": "brush-tailed porcupine, brush-tail porcupine",
  "n02347443": "Trichys, genus Trichys",
  "n02347573": "long-tailed porcupine, Trichys lipura",
  "n02347744": "New World porcupine",
  "n02347865": "Erethizontidae, family Erethizontidae",
  "n02348036": "Erethizon, genus Erethizon",
  "n02348173": "Canada porcupine, Erethizon dorsatum",
  "n02348405": "Heteromyidae, family Heteromyidae",
  "n02348788": "pocket mouse",
  "n02349040": "Perognathus, genus Perognathus",
  "n02349205": "silky pocket mouse, Perognathus flavus",
  "n02349390": "plains pocket mouse, Perognathus flavescens",
  "n02349557": "hispid pocket mouse, Perognathus hispidus",
  "n02349730": "Liomys, genus Liomys",
  "n02349847": "Mexican pocket mouse, Liomys irroratus",
  "n02349980": "Dipodomys, genus Dipodomys",
  "n02350105": "kangaroo rat, desert rat, Dipodomys phillipsii",
  "n02350357": "Ord kangaroo rat, Dipodomys ordi",
  "n02350537": "Microdipodops, genus Microdipodops",
  "n02350670": "kangaroo mouse, dwarf pocket rat",
  "n02350845": "Zapodidae, family Zapodidae",
  "n02350989": "jumping mouse",
  "n02351212": "Zapus, genus Zapus",
  "n02351343": "meadow jumping mouse, Zapus hudsonius",
  "n02351518": "Dipodidae, family Dipodidae",
  "n02351686": "Dipus, genus Dipus",
  "n02351870": "jerboa",
  "n02352002": "typical jerboa",
  "n02352175": "Jaculus, genus Jaculus",
  "n02352290": "Jaculus jaculus",
  "n02352390": "Gliridae, family Gliridae",
  "n02352591": "dormouse",
  "n02352804": "Glis, genus Glis",
  "n02352932": "loir, Glis glis",
  "n02353037": "Muscardinus, genus Muscardinus",
  "n02353172": "hazel mouse, Muscardinus avellanarius",
  "n02353297": "Eliomys, genus Eliomys",
  "n02353411": "lerot",
  "n02353529": "Geomyidae, family Geomyidae",
  "n02353709": "Geomys, genus Geomys",
  "n02353861": "gopher, pocket gopher, pouched rat",
  "n02354162": "plains pocket gopher, Geomys bursarius",
  "n02354320": "southeastern pocket gopher, Geomys pinetis",
  "n02354470": "Thomomys, genus Thomomys",
  "n02354621": "valley pocket gopher, Thomomys bottae",
  "n02354781": "northern pocket gopher, Thomomys talpoides",
  "n02354950": "Sciuromorpha, suborder Sciuromorpha",
  "n02355227": "squirrel",
  "n02355477": "tree squirrel",
  "n02355711": "Sciuridae, family Sciuridae",
  "n02356108": "Sciurus, genus Sciurus",
  "n02356381": "eastern grey squirrel, eastern gray squirrel, cat squirrel, Sciurus carolinensis",
  "n02356612": "western grey squirrel, western gray squirrel, Sciurus griseus",
  "n02356798": "fox squirrel, eastern fox squirrel, Sciurus niger",
  "n02356977": "black squirrel",
  "n02357111": "red squirrel, cat squirrel, Sciurus vulgaris",
  "n02357280": "Tamiasciurus, genus Tamiasciurus",
  "n02357401": "American red squirrel, spruce squirrel, red squirrel, Sciurus hudsonicus, Tamiasciurus hudsonicus",
  "n02357585": "chickeree, Douglas squirrel, Tamiasciurus douglasi",
  "n02357741": "Citellus, genus Citellus, Spermophilus, genus Spermophilus",
  "n02357911": "antelope squirrel, whitetail antelope squirrel, antelope chipmunk, Citellus leucurus",
  "n02358091": "ground squirrel, gopher, spermophile",
  "n02358390": "mantled ground squirrel, Citellus lateralis",
  "n02358584": "suslik, souslik, Citellus citellus",
  "n02358712": "flickertail, Richardson ground squirrel, Citellus richardsoni",
  "n02358890": "rock squirrel, Citellus variegatus",
  "n02359047": "Arctic ground squirrel, parka squirrel, Citellus parryi",
  "n02359204": "Cynomys, genus Cynomys",
  "n02359324": "prairie dog, prairie marmot",
  "n02359556": "blacktail prairie dog, Cynomys ludovicianus",
  "n02359667": "whitetail prairie dog, Cynomys gunnisoni",
  "n02359775": "Tamias, genus Tamias",
  "n02359915": "eastern chipmunk, hackee, striped squirrel, ground squirrel, Tamias striatus",
  "n02360135": "Eutamias, genus Eutamias",
  "n02360282": "chipmunk",
  "n02360480": "baronduki, baranduki, barunduki, burunduki, Eutamius asiaticus, Eutamius sibiricus",
  "n02360643": "Glaucomys, genus Glaucomys",
  "n02360781": "American flying squirrel",
  "n02360933": "southern flying squirrel, Glaucomys volans",
  "n02361090": "northern flying squirrel, Glaucomys sabrinus",
  "n02361222": "Marmota, genus Marmota",
  "n02361337": "marmot",
  "n02361587": "groundhog, woodchuck, Marmota monax",
  "n02361706": "hoary marmot, whistler, whistling marmot, Marmota caligata",
  "n02361850": "yellowbelly marmot, rockchuck, Marmota flaviventris",
  "n02362025": "Petauristidae, subfamily Petauristidae",
  "n02362194": "Asiatic flying squirrel",
  "n02362420": "Petaurista, genus Petaurista",
  "n02362569": "taguan, flying marmot, flying cat, Petaurista petaurista",
  "n02362721": "Castoridae, family Castoridae",
  "n02362862": "Castor, genus Castor",
  "n02363005": "beaver",
  "n02363245": "Old World beaver, Castor fiber",
  "n02363351": "New World beaver, Castor canadensis",
  "n02363511": "Castoroides, genus Castoroides",
  "n02363681": "Aplodontiidae, family Aplodontiidae",
  "n02363818": "Aplodontia, genus Aplodontia",
  "n02363996": "mountain beaver, sewellel, Aplodontia rufa",
  "n02364221": "Caviidae, family Caviidae",
  "n02364377": "Cavia, genus Cavia",
  "n02364520": "cavy",
  "n02364673": "guinea pig, Cavia cobaya",
  "n02364840": "aperea, wild cavy, Cavia porcellus",
  "n02364989": "Dolichotis, genus Dolichotis",
  "n02365108": "mara, Dolichotis patagonum",
  "n02365244": "Hydrochoeridae, family Hydrochoeridae",
  "n02365356": "Hydrochoerus, genus Hydrochoerus",
  "n02365480": "capybara, capibara, Hydrochoerus hydrochaeris",
  "n02365672": "Dasyproctidae, family Dasyproctidae",
  "n02365848": "Dasyprocta, genus Dasyprocta",
  "n02366002": "agouti, Dasyprocta aguti",
  "n02366203": "Cuniculus, genus Cuniculus",
  "n02366301": "paca, Cuniculus paca",
  "n02366453": "Stictomys, genus Stictomys",
  "n02366579": "mountain paca",
  "n02366702": "Capromyidae, family Capromyidae",
  "n02366825": "Myocastor, genus Myocastor",
  "n02366959": "coypu, nutria, Myocastor coypus",
  "n02367131": "Chinchillidae, family Chinchillidae",
  "n02367360": "genus Chinchilla",
  "n02367492": "chinchilla, Chinchilla laniger",
  "n02367678": "Lagidium, genus Lagidium",
  "n02367812": "mountain chinchilla, mountain viscacha",
  "n02367993": "Lagostomus, genus Lagostomus",
  "n02368116": "viscacha, chinchillon, Lagostomus maximus",
  "n02368280": "Abrocoma, genus Abrocoma",
  "n02368399": "abrocome, chinchilla rat, rat chinchilla",
  "n02368563": "Spalacidae, family Spalacidae",
  "n02368687": "Spalax, genus Spalax",
  "n02368821": "mole rat",
  "n02369012": "Bathyergidae, family Bathyergidae",
  "n02369170": "Bathyergus, genus Bathyergus",
  "n02369293": "mole rat",
  "n02369424": "Heterocephalus, genus Heterocephalus",
  "n02369555": "sand rat",
  "n02369680": "naked mole rat",
  "n02369935": "queen, queen mole rat",
  "n02370137": "Damaraland mole rat",
  "n02370265": "dug",
  "n02370360": "udder, bag",
  "n02370525": "Ungulata",
  "n02370806": "ungulate, hoofed mammal",
  "n02371125": "Unguiculata",
  "n02371344": "unguiculate, unguiculate mammal",
  "n02371471": "Dinocerata, order Dinocerata",
  "n02371647": "Uintatheriidae, family Uintatheriidae",
  "n02371801": "Uintatherium, genus Uintatherium",
  "n02372046": "dinocerate",
  "n02372140": "dinoceras, uintathere",
  "n02372251": "Hyracoidea, order Hyracoidea",
  "n02372397": "Procaviidae, family Procaviidae",
  "n02372584": "hyrax, coney, cony, dassie, das",
  "n02372813": "Procavia, genus Procavia",
  "n02372952": "rock hyrax, rock rabbit, Procavia capensis",
  "n02373093": "Perissodactyla, order Perissodactyla",
  "n02373336": "odd-toed ungulate, perissodactyl, perissodactyl mammal",
  "n02373601": "Equidae, family Equidae",
  "n02373843": "Equus, genus Equus",
  "n02374149": "equine, equid",
  "n02374451": "horse, Equus caballus",
  "n02375302": "roan",
  "n02375438": "stablemate, stable companion",
  "n02375592": "Hyracotherium, genus Hyracotherium",
  "n02375757": "gee-gee",
  "n02375862": "eohippus, dawn horse",
  "n02376016": "genus Mesohippus",
  "n02376135": "mesohippus",
  "n02376303": "genus Protohippus",
  "n02376423": "protohippus",
  "n02376542": "foal",
  "n02376679": "filly",
  "n02376791": "colt",
  "n02376918": "male horse",
  "n02377063": "ridgeling, ridgling, ridgel, ridgil",
  "n02377181": "stallion, entire",
  "n02377291": "stud, studhorse",
  "n02377388": "gelding",
  "n02377480": "mare, female horse",
  "n02377603": "broodmare, stud mare",
  "n02377703": "saddle horse, riding horse, mount",
  "n02378149": "remount",
  "n02378299": "palfrey",
  "n02378415": "warhorse",
  "n02378541": "cavalry horse",
  "n02378625": "charger, courser",
  "n02378755": "steed",
  "n02378870": "prancer",
  "n02378969": "hack",
  "n02379081": "cow pony",
  "n02379183": "quarter horse",
  "n02379329": "Morgan",
  "n02379430": "Tennessee walker, Tennessee walking horse, Walking horse, Plantation walking horse",
  "n02379630": "American saddle horse",
  "n02379743": "Appaloosa",
  "n02379908": "Arabian, Arab",
  "n02380052": "Lippizan, Lipizzan, Lippizaner",
  "n02380335": "pony",
  "n02380464": "polo pony",
  "n02380583": "mustang",
  "n02380745": "bronco, bronc, broncho",
  "n02380875": "bucking bronco",
  "n02381004": "buckskin",
  "n02381119": "crowbait, crow-bait",
  "n02381261": "dun",
  "n02381364": "grey, gray",
  "n02381460": "wild horse",
  "n02381609": "tarpan, Equus caballus gomelini",
  "n02381744": "warrigal, warragal",
  "n02381831": "Przewalski's horse, Przevalski's horse, Equus caballus przewalskii, Equus caballus przevalskii",
  "n02382039": "cayuse, Indian pony",
  "n02382132": "hack",
  "n02382204": "hack, jade, nag, plug",
  "n02382338": "plow horse, plough horse",
  "n02382437": "pony",
  "n02382635": "Shetland pony",
  "n02382750": "Welsh pony",
  "n02382850": "Exmoor",
  "n02382948": "racehorse, race horse, bangtail",
  "n02383231": "thoroughbred",
  "n02383604": "Sir Barton",
  "n02383708": "Gallant Fox",
  "n02383813": "Omaha",
  "n02383912": "War Admiral",
  "n02384017": "Whirlaway",
  "n02384120": "Count Fleet",
  "n02384225": "Assault",
  "n02384326": "Citation",
  "n02384428": "Secretariat",
  "n02384533": "Seattle Slew",
  "n02384639": "Affirmed",
  "n02384741": "steeplechaser",
  "n02384858": "racer",
  "n02385002": "finisher",
  "n02385098": "pony",
  "n02385214": "yearling",
  "n02385348": "two-year-old horse, two year old",
  "n02385461": "three-year-old horse, three year old",
  "n02385580": "dark horse",
  "n02385676": "mudder",
  "n02385776": "nonstarter",
  "n02385898": "stalking-horse",
  "n02386014": "harness horse",
  "n02386141": "cob",
  "n02386224": "hackney",
  "n02386310": "workhorse",
  "n02386496": "draft horse, draught horse, dray horse",
  "n02386746": "packhorse",
  "n02386853": "carthorse, cart horse, drayhorse",
  "n02386968": "Clydesdale",
  "n02387093": "Percheron",
  "n02387254": "farm horse, dobbin",
  "n02387346": "shire, shire horse",
  "n02387452": "pole horse, poler",
  "n02387581": "wheel horse, wheeler",
  "n02387722": "post horse, post-horse, poster",
  "n02387887": "coach horse",
  "n02387983": "pacer",
  "n02388143": "pacer, pacemaker, pacesetter",
  "n02388276": "trotting horse, trotter",
  "n02388453": "pole horse",
  "n02388588": "stepper, high stepper",
  "n02388735": "chestnut",
  "n02388832": "liver chestnut",
  "n02388917": "bay",
  "n02389026": "sorrel",
  "n02389128": "palomino",
  "n02389261": "pinto",
  "n02389346": "ass",
  "n02389559": "domestic ass, donkey, Equus asinus",
  "n02389779": "burro",
  "n02389865": "moke",
  "n02389943": "jack, jackass",
  "n02390015": "jennet, jenny, jenny ass",
  "n02390101": "mule",
  "n02390258": "hinny",
  "n02390454": "wild ass",
  "n02390640": "African wild ass, Equus asinus",
  "n02390738": "kiang, Equus kiang",
  "n02390834": "onager, Equus hemionus",
  "n02390938": "chigetai, dziggetai, Equus hemionus hemionus",
  "n02391049": "zebra",
  "n02391234": "common zebra, Burchell's zebra, Equus Burchelli",
  "n02391373": "mountain zebra, Equus zebra zebra",
  "n02391508": "grevy's zebra, Equus grevyi",
  "n02391617": "quagga, Equus quagga",
  "n02391782": "Rhinocerotidae, family Rhinocerotidae, rhinoceros family",
  "n02391994": "rhinoceros, rhino",
  "n02392282": "genus Rhinoceros",
  "n02392434": "Indian rhinoceros, Rhinoceros unicornis",
  "n02392555": "woolly rhinoceros, Rhinoceros antiquitatis",
  "n02392710": "Ceratotherium, genus Ceratotherium",
  "n02392824": "white rhinoceros, Ceratotherium simum, Diceros simus",
  "n02393024": "Diceros, genus Diceros",
  "n02393161": "black rhinoceros, Diceros bicornis",
  "n02393300": "Tapiridae, family Tapiridae",
  "n02393445": "Tapirus, genus Tapirus",
  "n02393580": "tapir",
  "n02393807": "New World tapir, Tapirus terrestris",
  "n02393940": "Malayan tapir, Indian tapir, Tapirus indicus",
  "n02394068": "Artiodactyla, order Artiodactyla",
  "n02394477": "even-toed ungulate, artiodactyl, artiodactyl mammal",
  "n02394822": "Suidae, family Suidae",
  "n02395003": "swine",
  "n02395244": "Sus, genus Sus",
  "n02395406": "hog, pig, grunter, squealer, Sus scrofa",
  "n02395694": "piglet, piggy, shoat, shote",
  "n02395855": "sucking pig",
  "n02395931": "porker",
  "n02396014": "boar",
  "n02396088": "sow",
  "n02396157": "razorback, razorback hog, razorbacked hog",
  "n02396427": "wild boar, boar, Sus scrofa",
  "n02396667": "Babyrousa, genus Babyrousa",
  "n02396796": "babirusa, babiroussa, babirussa, Babyrousa Babyrussa",
  "n02396970": "Phacochoerus, genus Phacochoerus",
  "n02397096": "warthog",
  "n02397251": "Tayassuidae, family Tayassuidae",
  "n02397377": "Tayassu, genus Tayassu, genus Pecari",
  "n02397529": "peccary, musk hog",
  "n02397744": "collared peccary, javelina, Tayassu angulatus, Tayassu tajacu, Peccari angulatus",
  "n02397987": "white-lipped peccary, Tayassu pecari",
  "n02398141": "Chiacoan peccary",
  "n02398252": "Hippopotamidae, family Hippopotamidae",
  "n02398386": "genus Hippopotamus",
  "n02398521": "hippopotamus, hippo, river horse, Hippopotamus amphibius",
  "n02398732": "Ruminantia, suborder Ruminantia",
  "n02399000": "ruminant",
  "n02399424": "rumen, first stomach",
  "n02399648": "reticulum, second stomach",
  "n02399791": "psalterium, omasum, third stomach",
  "n02399942": "abomasum, fourth stomach",
  "n02400139": "Bovidae, family Bovidae",
  "n02401031": "bovid",
  "n02401305": "Bovinae, subfamily Bovinae",
  "n02401509": "Bovini, tribe Bovini",
  "n02401661": "Bos, genus Bos",
  "n02402010": "bovine",
  "n02402175": "ox, wild ox",
  "n02402425": "cattle, cows, kine, oxen, Bos taurus",
  "n02403003": "ox",
  "n02403153": "stirk",
  "n02403231": "bullock, steer",
  "n02403325": "bull",
  "n02403454": "cow, moo-cow",
  "n02403644": "springer, springing cow",
  "n02403740": "heifer",
  "n02403820": "bullock",
  "n02403920": "dogie, dogy, leppy",
  "n02404028": "maverick",
  "n02404186": "beef, beef cattle",
  "n02404432": "longhorn, Texas longhorn",
  "n02404573": "Brahman, Brahma, Brahmin, Bos indicus",
  "n02404906": "zebu",
  "n02405101": "aurochs, urus, Bos primigenius",
  "n02405302": "yak, Bos grunniens",
  "n02405440": "banteng, banting, tsine, Bos banteng",
  "n02405577": "Welsh, Welsh Black",
  "n02405692": "red poll",
  "n02405799": "Santa Gertrudis",
  "n02405929": "Aberdeen Angus, Angus, black Angus",
  "n02406046": "Africander",
  "n02406174": "dairy cattle, dairy cow, milch cow, milk cow, milcher, milker",
  "n02406432": "Ayrshire",
  "n02406533": "Brown Swiss",
  "n02406647": "Charolais",
  "n02406749": "Jersey",
  "n02406859": "Devon",
  "n02406952": "grade",
  "n02407071": "Durham, shorthorn",
  "n02407172": "milking shorthorn",
  "n02407276": "Galloway",
  "n02407390": "Friesian, Holstein, Holstein-Friesian",
  "n02407521": "Guernsey",
  "n02407625": "Hereford, whiteface",
  "n02407763": "cattalo, beefalo",
  "n02407959": "Old World buffalo, buffalo",
  "n02408217": "Bubalus, genus Bubalus, tribe Bubalus",
  "n02408429": "water buffalo, water ox, Asiatic buffalo, Bubalus bubalis",
  "n02408660": "Indian buffalo",
  "n02408817": "carabao",
  "n02408903": "genus Anoa",
  "n02409038": "anoa, dwarf buffalo, Anoa depressicornis",
  "n02409202": "tamarau, tamarao, Bubalus mindorensis, Anoa mindorensis",
  "n02409369": "Synercus, genus Synercus, tribe synercus",
  "n02409508": "Cape buffalo, Synercus caffer",
  "n02409702": "Bibos, genus Bibos",
  "n02409870": "Asian wild ox",
  "n02410011": "gaur, Bibos gaurus",
  "n02410141": "gayal, mithan, Bibos frontalis",
  "n02410313": "genus Bison",
  "n02410509": "bison",
  "n02410702": "American bison, American buffalo, buffalo, Bison bison",
  "n02410900": "wisent, aurochs, Bison bonasus",
  "n02411075": "Ovibos, genus Ovibos",
  "n02411206": "musk ox, musk sheep, Ovibos moschatus",
  "n02411427": "Ovis, genus Ovis",
  "n02411705": "sheep",
  "n02411999": "ewe",
  "n02412080": "ram, tup",
  "n02412210": "wether",
  "n02412337": "bellwether",
  "n02412440": "lamb",
  "n02412629": "lambkin",
  "n02412700": "baa-lamb",
  "n02412787": "hog, hogget, hogg",
  "n02412909": "teg",
  "n02412977": "Persian lamb",
  "n02413050": "black sheep",
  "n02413131": "domestic sheep, Ovis aries",
  "n02413484": "Cotswold",
  "n02413593": "Hampshire, Hampshire down",
  "n02413717": "Lincoln",
  "n02413824": "Exmoor",
  "n02413917": "Cheviot",
  "n02414043": "broadtail, caracul, karakul",
  "n02414209": "longwool",
  "n02414290": "merino, merino sheep",
  "n02414442": "Rambouillet",
  "n02414578": "wild sheep",
  "n02414763": "argali, argal, Ovis ammon",
  "n02414904": "Marco Polo sheep, Marco Polo's sheep, Ovis poli",
  "n02415130": "urial, Ovis vignei",
  "n02415253": "Dall sheep, Dall's sheep, white sheep, Ovis montana dalli",
  "n02415435": "mountain sheep",
  "n02415577": "bighorn, bighorn sheep, cimarron, Rocky Mountain bighorn, Rocky Mountain sheep, Ovis canadensis",
  "n02415829": "mouflon, moufflon, Ovis musimon",
  "n02415971": "Ammotragus, genus Ammotragus",
  "n02416104": "aoudad, arui, audad, Barbary sheep, maned sheep, Ammotragus lervia",
  "n02416270": "beard",
  "n02416410": "Capra, genus Capra",
  "n02416519": "goat, caprine animal",
  "n02416820": "kid",
  "n02416880": "billy, billy goat, he-goat",
  "n02416964": "nanny, nanny-goat, she-goat",
  "n02417070": "domestic goat, Capra hircus",
  "n02417242": "Cashmere goat, Kashmir goat",
  "n02417387": "Angora, Angora goat",
  "n02417534": "wild goat",
  "n02417663": "bezoar goat, pasang, Capra aegagrus",
  "n02417785": "markhor, markhoor, Capra falconeri",
  "n02417914": "ibex, Capra ibex",
  "n02418064": "goat antelope",
  "n02418341": "Oreamnos, genus Oreamnos",
  "n02418465": "mountain goat, Rocky Mountain goat, Oreamnos americanus",
  "n02418648": "Naemorhedus, genus Naemorhedus",
  "n02418770": "goral, Naemorhedus goral",
  "n02418934": "Capricornis, genus Capricornis",
  "n02419056": "serow",
  "n02419217": "Rupicapra, genus Rupicapra",
  "n02419336": "chamois, Rupicapra rupicapra",
  "n02419515": "Budorcas, genus Budorcas",
  "n02419634": "takin, gnu goat, Budorcas taxicolor",
  "n02419796": "antelope",
  "n02420389": "Antilope, genus Antilope",
  "n02420509": "blackbuck, black buck, Antilope cervicapra",
  "n02420675": "Litocranius, genus Litocranius",
  "n02420828": "gerenuk, Litocranius walleri",
  "n02420994": "genus Addax",
  "n02421136": "addax, Addax nasomaculatus",
  "n02421308": "Connochaetes, genus Connochaetes",
  "n02421449": "gnu, wildebeest",
  "n02421612": "Madoqua, genus Madoqua",
  "n02421792": "dik-dik",
  "n02421962": "Alcelaphus, genus Alcelaphus",
  "n02422106": "hartebeest",
  "n02422249": "Damaliscus, genus Damaliscus",
  "n02422391": "sassaby, topi, Damaliscus lunatus",
  "n02422561": "Aepyceros, genus Aepyceros",
  "n02422699": "impala, Aepyceros melampus",
  "n02422860": "Gazella, genus Gazella",
  "n02423022": "gazelle",
  "n02423218": "Thomson's gazelle, Gazella thomsoni",
  "n02423362": "Gazella subgutturosa",
  "n02423465": "Antidorcas, genus Antidorcas",
  "n02423589": "springbok, springbuck, Antidorcas marsupialis, Antidorcas euchore",
  "n02423787": "Tragelaphus, genus Tragelaphus, Strepsiceros, genus Strepsiceros",
  "n02424085": "bongo, Tragelaphus eurycerus, Boocercus eurycerus",
  "n02424305": "kudu, koodoo, koudou",
  "n02424486": "greater kudu, Tragelaphus strepsiceros",
  "n02424589": "lesser kudu, Tragelaphus imberbis",
  "n02424695": "harnessed antelope",
  "n02424909": "nyala, Tragelaphus angasi",
  "n02425086": "mountain nyala, Tragelaphus buxtoni",
  "n02425228": "bushbuck, guib, Tragelaphus scriptus",
  "n02425393": "Boselaphus, genus Boselaphus",
  "n02425532": "nilgai, nylghai, nylghau, blue bull, Boselaphus tragocamelus",
  "n02425756": "Hippotragus, genus Hippotragus",
  "n02425887": "sable antelope, Hippotragus niger",
  "n02426054": "genus Saiga",
  "n02426176": "saiga, Saiga tatarica",
  "n02426339": "Raphicerus, genus Raphicerus",
  "n02426481": "steenbok, steinbok, Raphicerus campestris",
  "n02426634": "Taurotragus, genus Taurotragus",
  "n02426813": "eland",
  "n02427032": "common eland, Taurotragus oryx",
  "n02427183": "giant eland, Taurotragus derbianus",
  "n02427337": "Kobus, genus Kobus",
  "n02427470": "kob, Kobus kob",
  "n02427576": "lechwe, Kobus leche",
  "n02427724": "waterbuck",
  "n02427958": "Adenota, genus Adenota",
  "n02428089": "puku, Adenota vardoni",
  "n02428229": "genus Oryx",
  "n02428349": "oryx, pasang",
  "n02428508": "gemsbok, gemsbuck, Oryx gazella",
  "n02428653": "Pseudoryx, genus Pseudoryx",
  "n02428842": "forest goat, spindle horn, Pseudoryx nghetinhensis",
  "n02429123": "Antilocapridae, family Antilocapridae",
  "n02429276": "Antilocapra, genus Antilocapra",
  "n02429456": "pronghorn, prongbuck, pronghorn antelope, American antelope, Antilocapra americana",
  "n02429695": "Cervidae, family Cervidae",
  "n02430045": "deer, cervid",
  "n02430559": "stag",
  "n02430643": "royal, royal stag",
  "n02430748": "pricket",
  "n02430830": "fawn",
  "n02430929": "Cervus, genus Cervus",
  "n02431122": "red deer, elk, American elk, wapiti, Cervus elaphus",
  "n02431337": "hart, stag",
  "n02431441": "hind",
  "n02431542": "brocket",
  "n02431628": "sambar, sambur, Cervus unicolor",
  "n02431785": "wapiti, elk, American elk, Cervus elaphus canadensis",
  "n02431976": "Japanese deer, sika, Cervus nipon, Cervus sika",
  "n02432139": "Odocoileus, genus Odocoileus",
  "n02432291": "Virginia deer, white tail, whitetail, white-tailed deer, whitetail deer, Odocoileus Virginianus",
  "n02432511": "mule deer, burro deer, Odocoileus hemionus",
  "n02432704": "black-tailed deer, blacktail deer, blacktail, Odocoileus hemionus columbianus",
  "n02432867": "Alces, genus Alces",
  "n02432983": "elk, European elk, moose, Alces alces",
  "n02433205": "Dama, genus Dama",
  "n02433318": "fallow deer, Dama dama",
  "n02433426": "Capreolus, genus Capreolus",
  "n02433546": "roe deer, Capreolus capreolus",
  "n02433729": "roebuck",
  "n02433796": "Rangifer, genus Rangifer",
  "n02433925": "caribou, reindeer, Greenland caribou, Rangifer tarandus",
  "n02434190": "woodland caribou, Rangifer caribou",
  "n02434415": "barren ground caribou, Rangifer arcticus",
  "n02434598": "Mazama, genus Mazama",
  "n02434712": "brocket",
  "n02434834": "Muntiacus, genus Muntiacus",
  "n02434954": "muntjac, barking deer",
  "n02435099": "Moschus, genus Moschus",
  "n02435216": "musk deer, Moschus moschiferus",
  "n02435386": "Elaphurus, genus Elaphurus",
  "n02435517": "pere david's deer, elaphure, Elaphurus davidianus",
  "n02435689": "Tragulidae, family Tragulidae",
  "n02435853": "chevrotain, mouse deer",
  "n02436067": "Tragulus, genus Tragulus",
  "n02436224": "kanchil, Tragulus kanchil",
  "n02436353": "napu, Tragulus Javanicus",
  "n02436514": "Hyemoschus, genus Hyemoschus",
  "n02436645": "water chevrotain, water deer, Hyemoschus aquaticus",
  "n02436813": "Camelidae, family Camelidae",
  "n02436993": "Camelus, genus Camelus",
  "n02437136": "camel",
  "n02437312": "Arabian camel, dromedary, Camelus dromedarius",
  "n02437482": "Bactrian camel, Camelus bactrianus",
  "n02437616": "llama",
  "n02437825": "Lama, genus Lama",
  "n02437971": "domestic llama, Lama peruana",
  "n02438173": "guanaco, Lama guanicoe",
  "n02438272": "alpaca, Lama pacos",
  "n02438452": "Vicugna, genus Vicugna",
  "n02438580": "vicuna, Vicugna vicugna",
  "n02438774": "Giraffidae, family Giraffidae",
  "n02438897": "Giraffa, genus Giraffa",
  "n02439033": "giraffe, camelopard, Giraffa camelopardalis",
  "n02439286": "Okapia, genus Okapia",
  "n02439398": "okapi, Okapia johnstoni",
  "n02439568": "trotter",
  "n02439728": "forefoot",
  "n02439829": "hindfoot",
  "n02439929": "paw",
  "n02440121": "forepaw",
  "n02440250": "hand",
  "n02440523": "pad",
  "n02440705": "Mustelidae, family Mustelidae",
  "n02441326": "musteline mammal, mustelid, musteline",
  "n02441723": "Mustela, genus Mustela",
  "n02441942": "weasel",
  "n02442172": "ermine, shorttail weasel, Mustela erminea",
  "n02442336": "stoat",
  "n02442446": "New World least weasel, Mustela rixosa",
  "n02442572": "Old World least weasel, Mustela nivalis",
  "n02442668": "longtail weasel, long-tailed weasel, Mustela frenata",
  "n02442845": "mink",
  "n02443015": "American mink, Mustela vison",
  "n02443114": "polecat, fitch, foulmart, foumart, Mustela putorius",
  "n02443346": "ferret",
  "n02443484": "black-footed ferret, ferret, Mustela nigripes",
  "n02443683": "Poecilogale, genus Poecilogale",
  "n02443808": "muishond",
  "n02443959": "snake muishond, Poecilogale albinucha",
  "n02444103": "Ictonyx, genus Ictonyx",
  "n02444251": "striped muishond, Ictonyx striata",
  "n02444384": "zoril, Ictonyx frenata",
  "n02444500": "Lutrinae, subfamily Lutrinae",
  "n02444647": "Lutra, genus Lutra",
  "n02444819": "otter",
  "n02445004": "river otter, Lutra canadensis",
  "n02445171": "Eurasian otter, Lutra lutra",
  "n02445276": "Enhydra, genus Enhydra",
  "n02445394": "sea otter, Enhydra lutris",
  "n02445564": "Mephitinae, subfamily Mephitinae",
  "n02445715": "skunk, polecat, wood pussy",
  "n02446014": "Mephitis, genus Mephitis",
  "n02446206": "striped skunk, Mephitis mephitis",
  "n02446352": "hooded skunk, Mephitis macroura",
  "n02446512": "Conepatus, genus Conepatus",
  "n02446645": "hog-nosed skunk, hognosed skunk, badger skunk, rooter skunk, Conepatus leuconotus",
  "n02446888": "Spilogale, genus Spilogale",
  "n02447021": "spotted skunk, little spotted skunk, Spilogale putorius",
  "n02447220": "Melinae, subfamily Melinae",
  "n02447366": "badger",
  "n02447591": "Taxidea, genus Taxidea",
  "n02447762": "American badger, Taxidea taxus",
  "n02447896": "Meles, genus Meles",
  "n02448060": "Eurasian badger, Meles meles",
  "n02448200": "Mellivora, genus Mellivora",
  "n02448318": "ratel, honey badger, Mellivora capensis",
  "n02448502": "Melogale, genus Melogale",
  "n02448633": "ferret badger",
  "n02448754": "Arctonyx, genus Arctonyx",
  "n02448885": "hog badger, hog-nosed badger, sand badger, Arctonyx collaris",
  "n02449060": "Gulo, genus Gulo",
  "n02449183": "wolverine, carcajou, skunk bear, Gulo luscus",
  "n02449350": "glutton, Gulo gulo, wolverine",
  "n02449464": "genus Grison",
  "n02449582": "genus Galictis",
  "n02449699": "grison, Grison vittatus, Galictis vittatus",
  "n02449921": "Martes, genus Martes",
  "n02450034": "marten, marten cat",
  "n02450295": "pine marten, Martes martes",
  "n02450426": "sable, Martes zibellina",
  "n02450561": "American marten, American sable, Martes americana",
  "n02450677": "stone marten, beech marten, Martes foina",
  "n02450829": "fisher, pekan, fisher cat, black cat, Martes pennanti",
  "n02450992": "Charronia, genus Charronia",
  "n02451125": "yellow-throated marten, Charronia flavigula",
  "n02451292": "Eira, genus Eira",
  "n02451415": "tayra, taira, Eira barbara",
  "n02451575": "fictional animal",
  "n02451818": "Easter bunny",
  "n02451912": "church mouse",
  "n02452014": "Mickey Mouse",
  "n02452138": "Minnie Mouse",
  "n02452225": "Donald Duck",
  "n02452347": "Mighty Mouse",
  "n02452464": "muzzle",
  "n02452637": "snout, neb",
  "n02452813": "snout, rostrum",
  "n02452967": "proboscis, trunk",
  "n02453108": "pachyderm",
  "n02453373": "Edentata, order Edentata",
  "n02453611": "edentate",
  "n02453890": "Xenarthra, suborder Xenarthra",
  "n02454119": "Dasypodidae, family Dasypodidae",
  "n02454379": "armadillo",
  "n02454657": "Dasypus, genus Dasypus",
  "n02454794": "peba, nine-banded armadillo, Texas armadillo, Dasypus novemcinctus",
  "n02454999": "Tolypeutes, genus Tolypeutes",
  "n02455135": "apar, three-banded armadillo, Tolypeutes tricinctus",
  "n02455310": "genus Cabassous",
  "n02455428": "tatouay, cabassous, Cabassous unicinctus",
  "n02455584": "Euphractus, genus Euphractus",
  "n02455720": "peludo, poyou, Euphractus sexcinctus",
  "n02455887": "Priodontes, genus Priodontes",
  "n02456008": "giant armadillo, tatou, tatu, Priodontes giganteus",
  "n02456147": "Chlamyphorus, genus Chlamyphorus",
  "n02456275": "pichiciago, pichiciego, fairy armadillo, chlamyphore, Chlamyphorus truncatus",
  "n02456505": "Burmeisteria, genus Burmeisteria",
  "n02456645": "greater pichiciego, Burmeisteria retusa",
  "n02456776": "Bradypodidae, family Bradypodidae",
  "n02456962": "sloth, tree sloth",
  "n02457249": "Bradypus, genus Bradypus",
  "n02457408": "three-toed sloth, ai, Bradypus tridactylus",
  "n02457586": "Megalonychidae, family Megalonychidae",
  "n02457756": "Choloepus, genus Choloepus",
  "n02457945": "two-toed sloth, unau, unai, Choloepus didactylus",
  "n02458135": "two-toed sloth, unau, unai, Choloepus hoffmanni",
  "n02458356": "Megatheriidae, family Megatheriidae",
  "n02458517": "megatherian, megatheriid, megatherian mammal",
  "n02458675": "Megatherium, genus Megatherium",
  "n02458822": "ground sloth, megathere",
  "n02459001": "Mylodontidae, family Mylodontidae",
  "n02459190": "mylodontid",
  "n02459313": "genus Mylodon",
  "n02459485": "mylodon",
  "n02459636": "mapinguari",
  "n02459808": "Myrmecophagidae, family Myrmecophagidae",
  "n02460009": "anteater, New World anteater",
  "n02460275": "Myrmecophaga, genus Myrmecophaga",
  "n02460451": "ant bear, giant anteater, great anteater, tamanoir, Myrmecophaga jubata",
  "n02460684": "Cyclopes, genus Cyclopes",
  "n02460817": "silky anteater, two-toed anteater, Cyclopes didactylus",
  "n02461014": "genus Tamandua",
  "n02461128": "tamandua, tamandu, lesser anteater, Tamandua tetradactyla",
  "n02461372": "Pholidota, order Pholidota",
  "n02461556": "Manidae, family Manidae",
  "n02461701": "Manis, genus Manis",
  "n02461830": "pangolin, scaly anteater, anteater",
  "n02462066": "pastern, fetter bone",
  "n02462213": "coronet",
  "n02462349": "fetlock",
  "n02462464": "fetlock, fetlock joint",
  "n02462602": "withers",
  "n02462828": "cannon, shank",
  "n02463024": "cannon bone",
  "n02463205": "hock, hock-joint",
  "n02463403": "loin, lumbus",
  "n02463611": "hindquarters, croup, croupe, rump",
  "n02463810": "haunch",
  "n02463913": "gaskin",
  "n02464065": "stifle, knee",
  "n02464223": "flank",
  "n02464327": "animal leg",
  "n02464461": "hind limb, hindlimb",
  "n02464626": "hind leg",
  "n02464785": "forelimb",
  "n02464965": "foreleg",
  "n02465084": "flipper",
  "n02465257": "parapodium",
  "n02465414": "sucker",
  "n02465585": "cupule",
  "n02465693": "stinger",
  "n02465929": "lateral line, lateral line organ",
  "n02466132": "fin",
  "n02466400": "dorsal fin",
  "n02466564": "pectoral fin",
  "n02466731": "pelvic fin, ventral fin",
  "n02466957": "tail fin, caudal fin",
  "n02467133": "heterocercal fin",
  "n02467320": "homocercal fin",
  "n02467491": "fishbone",
  "n02467581": "air bladder, swim bladder, float",
  "n02467746": "air sac",
  "n02467884": "air sac",
  "n02468017": "uropygial gland, preen gland",
  "n02468178": "silk gland, serictery, sericterium",
  "n02468368": "elbow",
  "n02468504": "chestnut",
  "n02468617": "quill, calamus, shaft",
  "n02468735": "vein, nervure",
  "n02468864": "flight feather, pinion, quill, quill feather",
  "n02469080": "primary, primary feather, primary quill",
  "n02469248": "scapular",
  "n02469344": "tail feather",
  "n02469472": "tadpole, polliwog, pollywog",
  "n02469588": "Primates, order Primates",
  "n02469914": "primate",
  "n02470238": "simian",
  "n02470325": "ape",
  "n02470451": "Anthropoidea, suborder Anthropoidea",
  "n02470709": "anthropoid",
  "n02470899": "anthropoid ape",
  "n02471072": "Hominoidea, superfamily Hominoidea",
  "n02471300": "hominoid",
  "n02471467": "Hominidae, family Hominidae",
  "n02471762": "hominid",
  "n02472012": "genus Homo",
  "n02472293": "homo, man, human being, human",
  "n02472987": "world, human race, humanity, humankind, human beings, humans, mankind, man",
  "n02473307": "Homo erectus",
  "n02473554": "Pithecanthropus, Pithecanthropus erectus, genus Pithecanthropus",
  "n02473720": "Java man, Trinil man",
  "n02473857": "Peking man",
  "n02473983": "Sinanthropus, genus Sinanthropus",
  "n02474110": "Homo soloensis",
  "n02474282": "Javanthropus, genus Javanthropus",
  "n02474431": "Solo man",
  "n02474605": "Homo habilis",
  "n02474777": "Homo sapiens",
  "n02475078": "Neandertal man, Neanderthal man, Neandertal, Neanderthal, Homo sapiens neanderthalensis",
  "n02475358": "Cro-magnon",
  "n02475478": "Boskop man",
  "n02475669": "Homo sapiens sapiens, modern man",
  "n02475821": "genus Australopithecus, Australopithecus",
  "n02476074": "Plesianthropus, genus Plesianthropus",
  "n02476219": "australopithecine",
  "n02476567": "Australopithecus afarensis",
  "n02476736": "Lucy",
  "n02476870": "Australopithecus africanus",
  "n02477028": "Australopithecus boisei",
  "n02477187": "Zinjanthropus, genus Zinjanthropus",
  "n02477329": "Australopithecus robustus",
  "n02477516": "Paranthropus, genus Paranthropus",
  "n02477647": "genus Sivapithecus",
  "n02477782": "Sivapithecus",
  "n02477890": "Dryopithecus, genus Dryopithecus",
  "n02478077": "dryopithecine",
  "n02478239": "rudapithecus, Dryopithecus Rudapithecus hungaricus",
  "n02478416": "Ouranopithecus, genus Ouranopithecus",
  "n02478539": "Lufengpithecus, genus Lufengpithecus",
  "n02478662": "genus Proconsul",
  "n02478875": "proconsul",
  "n02478990": "Kenyapithecus, genus Kenyapithecus",
  "n02479205": "genus Aegyptopithecus",
  "n02479332": "Aegyptopithecus",
  "n02479482": "Algeripithecus, genus Algeripithecus",
  "n02479634": "Algeripithecus minutus",
  "n02479896": "Pongidae, family Pongidae",
  "n02480153": "great ape, pongid",
  "n02480346": "Pongo, genus Pongo",
  "n02480495": "orangutan, orang, orangutang, Pongo pygmaeus",
  "n02480673": "genus Gorilla",
  "n02480855": "gorilla, Gorilla gorilla",
  "n02481103": "western lowland gorilla, Gorilla gorilla gorilla",
  "n02481235": "eastern lowland gorilla, Gorilla gorilla grauri",
  "n02481366": "mountain gorilla, Gorilla gorilla beringei",
  "n02481500": "silverback",
  "n02481629": "Pan, genus Pan",
  "n02481823": "chimpanzee, chimp, Pan troglodytes",
  "n02482060": "western chimpanzee, Pan troglodytes verus",
  "n02482286": "eastern chimpanzee, Pan troglodytes schweinfurthii",
  "n02482474": "central chimpanzee, Pan troglodytes troglodytes",
  "n02482650": "pygmy chimpanzee, bonobo, Pan paniscus",
  "n02482820": "Hylobatidae, family Hylobatidae",
  "n02483092": "lesser ape",
  "n02483224": "Hylobates, genus Hylobates",
  "n02483362": "gibbon, Hylobates lar",
  "n02483564": "Symphalangus, genus Symphalangus",
  "n02483708": "siamang, Hylobates syndactylus, Symphalangus syndactylus",
  "n02483915": "Cercopithecidae, family Cercopithecidae",
  "n02484322": "monkey",
  "n02484473": "Old World monkey, catarrhine",
  "n02484813": "Cercopithecus, genus Cercopithecus",
  "n02484975": "guenon, guenon monkey",
  "n02485225": "talapoin, Cercopithecus talapoin",
  "n02485371": "grivet, Cercopithecus aethiops",
  "n02485536": "vervet, vervet monkey, Cercopithecus aethiops pygerythrus",
  "n02485688": "green monkey, African green monkey, Cercopithecus aethiops sabaeus",
  "n02485865": "Cercocebus, genus Cercocebus",
  "n02485988": "mangabey",
  "n02486138": "Erythrocebus, genus Erythrocebus",
  "n02486261": "patas, hussar monkey, Erythrocebus patas",
  "n02486410": "baboon",
  "n02486565": "Papio, genus Papio",
  "n02486657": "chacma, chacma baboon, Papio ursinus",
  "n02486787": "Mandrillus, genus Mandrillus",
  "n02486908": "mandrill, Mandrillus sphinx",
  "n02487079": "drill, Mandrillus leucophaeus",
  "n02487217": "Macaca, genus Macaca",
  "n02487347": "macaque",
  "n02487547": "rhesus, rhesus monkey, Macaca mulatta",
  "n02487675": "bonnet macaque, bonnet monkey, capped macaque, crown monkey, Macaca radiata",
  "n02487847": "Barbary ape, Macaca sylvana",
  "n02488003": "crab-eating macaque, croo monkey, Macaca irus",
  "n02488149": "Presbytes, genus Presbytes, mammal Semnopithecus",
  "n02488291": "langur",
  "n02488415": "entellus, hanuman, Presbytes entellus, Semnopithecus entellus",
  "n02488578": "genus Colobus",
  "n02488702": "colobus, colobus monkey",
  "n02488894": "guereza, Colobus guereza",
  "n02489060": "Nasalis, genus Nasalis",
  "n02489166": "proboscis monkey, Nasalis larvatus",
  "n02489288": "Platyrrhini, superfamily Platyrrhini",
  "n02489589": "New World monkey, platyrrhine, platyrrhinian",
  "n02490030": "Callithricidae, family Callithricidae",
  "n02490219": "marmoset",
  "n02490435": "Callithrix, genus Callithrix",
  "n02490597": "true marmoset",
  "n02490686": "Cebuella, genus Cebuella",
  "n02490811": "pygmy marmoset, Cebuella pygmaea",
  "n02490964": "Leontocebus, genus Leontocebus, genus Leontideus",
  "n02491107": "tamarin, lion monkey, lion marmoset, leoncita",
  "n02491329": "silky tamarin, Leontocebus rosalia",
  "n02491474": "pinche, Leontocebus oedipus",
  "n02491590": "Cebidae, family Cebidae",
  "n02491906": "Cebus, genus Cebus",
  "n02492035": "capuchin, ringtail, Cebus capucinus",
  "n02492240": "Aotus, genus Aotus",
  "n02492356": "douroucouli, Aotus trivirgatus",
  "n02492536": "Alouatta, genus Alouatta",
  "n02492660": "howler monkey, howler",
  "n02492833": "Pithecia, genus Pithecia",
  "n02492948": "saki",
  "n02493109": "Cacajao, genus Cacajao",
  "n02493224": "uakari",
  "n02493390": "Callicebus, genus Callicebus",
  "n02493509": "titi, titi monkey",
  "n02493673": "Ateles, genus Ateles",
  "n02493793": "spider monkey, Ateles geoffroyi",
  "n02493974": "Saimiri, genus Saimiri",
  "n02494079": "squirrel monkey, Saimiri sciureus",
  "n02494257": "Lagothrix, genus Lagothrix",
  "n02494383": "woolly monkey",
  "n02494538": "Scandentia, order Scandentia",
  "n02494866": "Tupaiidae, family Tupaiidae",
  "n02495099": "Tupaia, genus Tupaia",
  "n02495242": "tree shrew",
  "n02495446": "Ptilocercus, genus Ptilocercus",
  "n02495570": "pentail, pen-tail, pen-tailed tree shrew",
  "n02495789": "Prosimii, suborder Prosimii",
  "n02496052": "prosimian",
  "n02496210": "Adapid, Adapid group",
  "n02496576": "Lemuroidea, suborder Lemuroidea",
  "n02496913": "lemur",
  "n02497214": "Strepsirhini, suborder Strepsirhini",
  "n02497408": "Lemuridae, family Lemuridae",
  "n02497550": "genus Lemur",
  "n02497673": "Madagascar cat, ring-tailed lemur, Lemur catta",
  "n02497832": "Daubentoniidae, family Daubentoniidae",
  "n02497983": "Daubentonia, genus Daubentonia",
  "n02498153": "aye-aye, Daubentonia madagascariensis",
  "n02498355": "Lorisidae, family Lorisidae",
  "n02498620": "genus Loris",
  "n02498743": "slender loris, Loris gracilis",
  "n02498888": "Nycticebus, genus Nycticebus",
  "n02499022": "slow loris, Nycticebus tardigradua, Nycticebus pygmaeus",
  "n02499178": "Perodicticus, genus Perodicticus",
  "n02499316": "potto, kinkajou, Perodicticus potto",
  "n02499434": "Arctocebus, genus Arctocebus",
  "n02499568": "angwantibo, golden potto, Arctocebus calabarensis",
  "n02499700": "genus Galago",
  "n02499808": "galago, bushbaby, bush baby",
  "n02499990": "Indriidae, family Indriidae",
  "n02500144": "genus Indri",
  "n02500267": "indri, indris, Indri indri, Indri brevicaudatus",
  "n02500472": "Avahi, genus Avahi",
  "n02500596": "woolly indris, Avahi laniger",
  "n02500749": "Omomyid, Omomyid group",
  "n02501101": "Tarsioidea, suborder Tarsioidea",
  "n02501275": "Tarsiidae, family Tarsiidae",
  "n02501432": "Tarsius, genus Tarsius",
  "n02501583": "tarsier",
  "n02501923": "Tarsius syrichta",
  "n02502006": "Tarsius glis",
  "n02502085": "Dermoptera, order Dermoptera",
  "n02502212": "Cynocephalidae, family Cynocephalidae",
  "n02502357": "Cynocephalus, genus Cynocephalus",
  "n02502514": "flying lemur, flying cat, colugo",
  "n02502807": "Cynocephalus variegatus",
  "n02502902": "Proboscidea, order Proboscidea",
  "n02503127": "proboscidean, proboscidian",
  "n02503313": "Elephantidae, family Elephantidae",
  "n02503517": "elephant",
  "n02503756": "rogue elephant",
  "n02503868": "Elephas, genus Elephas",
  "n02504013": "Indian elephant, Elephas maximus",
  "n02504196": "white elephant",
  "n02504323": "Loxodonta, genus Loxodonta",
  "n02504458": "African elephant, Loxodonta africana",
  "n02504635": "Mammuthus, genus Mammuthus",
  "n02504770": "mammoth",
  "n02505063": "woolly mammoth, northern mammoth, Mammuthus primigenius",
  "n02505238": "columbian mammoth, Mammuthus columbi",
  "n02505342": "Archidiskidon, genus Archidiskidon",
  "n02505485": "imperial mammoth, imperial elephant, Archidiskidon imperator",
  "n02505646": "Mammutidae, family Mammutidae, family Mastodontidae",
  "n02505809": "Mammut, genus Mammut, genus Mastodon",
  "n02505998": "mastodon, mastodont",
  "n02506248": "American mastodon, American mastodont, Mammut americanum",
  "n02506466": "Gomphotheriidae, family Gomphotheriidae",
  "n02506630": "Gomphotherium, genus Gomphotherium",
  "n02506783": "gomphothere",
  "n02506947": "plantigrade mammal, plantigrade",
  "n02507148": "digitigrade mammal, digitigrade",
  "n02507337": "Procyonidae, family Procyonidae",
  "n02507649": "procyonid",
  "n02507863": "Procyon, genus Procyon",
  "n02508021": "raccoon, racoon",
  "n02508213": "common raccoon, common racoon, coon, ringtail, Procyon lotor",
  "n02508346": "crab-eating raccoon, Procyon cancrivorus",
  "n02508458": "Bassariscidae, subfamily Bassariscidae",
  "n02508615": "Bassariscus, genus Bassariscus",
  "n02508742": "bassarisk, cacomistle, cacomixle, coon cat, raccoon fox, ringtail, ring-tailed cat, civet cat, miner's cat, Bassariscus astutus",
  "n02509071": "Potos, genus Potos",
  "n02509197": "kinkajou, honey bear, potto, Potos flavus, Potos caudivolvulus",
  "n02509405": "Nasua, genus Nasua",
  "n02509515": "coati, coati-mondi, coati-mundi, coon cat, Nasua narica",
  "n02509694": "Ailurus, genus Ailurus",
  "n02509815": "lesser panda, red panda, panda, bear cat, cat bear, Ailurus fulgens",
  "n02510065": "Ailuropodidae, family Ailuropodidae",
  "n02510240": "Ailuropoda, genus Ailuropoda",
  "n02510455": "giant panda, panda, panda bear, coon bear, Ailuropoda melanoleuca",
  "n02510769": "gill, branchia",
  "n02510978": "external gill",
  "n02511107": "gill slit, branchial cleft, gill cleft",
  "n02511303": "gill arch, branchial arch, gill bar",
  "n02511510": "peristome",
  "n02511633": "syrinx",
  "n02511730": "twitterer",
  "n02511824": "Pisces",
  "n02512053": "fish",
  "n02512752": "fingerling",
  "n02512830": "game fish, sport fish",
  "n02512938": "food fish",
  "n02513248": "rough fish",
  "n02513355": "groundfish, bottom fish",
  "n02513560": "young fish",
  "n02513727": "parr",
  "n02513805": "mouthbreeder",
  "n02513939": "spawner",
  "n02514041": "barracouta, snoek",
  "n02514198": "Channidae, class Channidae",
  "n02514320": "northern snakehead",
  "n02514575": "Osteichthyes, class Osteichthyes",
  "n02514825": "bony fish",
  "n02514988": "Crossopterygii, subclass Crossopterygii",
  "n02515214": "crossopterygian, lobefin, lobe-finned fish",
  "n02515410": "Latimeridae, family Latimeridae",
  "n02515560": "Latimeria, genus Latimeria",
  "n02515713": "coelacanth, Latimeria chalumnae",
  "n02515914": "Dipnoi, subclass Dipnoi",
  "n02516188": "lungfish",
  "n02516427": "Ceratodontidae, family Ceratodontidae",
  "n02516615": "genus Ceratodus",
  "n02516776": "ceratodus",
  "n02516867": "Neoceratodus, genus Neoceratodus",
  "n02516994": "Australian lungfish, Queensland lungfish, Neoceratodus forsteri",
  "n02517169": "Siluriformes, order Siluriformes",
  "n02517442": "catfish, siluriform fish",
  "n02517768": "Siluridae, family Siluridae",
  "n02517938": "silurid, silurid fish",
  "n02518178": "Silurus, genus Silurus",
  "n02518324": "European catfish, sheatfish, Silurus glanis",
  "n02518488": "Malopterurus, genus Malopterurus",
  "n02518622": "electric catfish, Malopterurus electricus",
  "n02518813": "Ameiuridae, family Ameiuridae",
  "n02518990": "Ameiurus, genus Ameiurus",
  "n02519148": "bullhead, bullhead catfish",
  "n02519340": "horned pout, hornpout, pout, Ameiurus Melas",
  "n02519472": "brown bullhead",
  "n02519576": "Ictalurus, genus Ictalurus",
  "n02519686": "channel catfish, channel cat, Ictalurus punctatus",
  "n02519862": "blue catfish, blue cat, blue channel catfish, blue channel cat",
  "n02520015": "Pylodictus, genus Pylodictus",
  "n02520147": "flathead catfish, mudcat, goujon, shovelnose catfish, spoonbill catfish, Pylodictus olivaris",
  "n02520391": "Laricariidae, family Laricariidae",
  "n02520525": "armored catfish",
  "n02520669": "Ariidae, family Ariidae",
  "n02520810": "sea catfish",
  "n02520985": "Arius, genus Arius",
  "n02521129": "crucifix fish",
  "n02521241": "Gadiformes, order Gadiformes",
  "n02521472": "Anacanthini, order Anacanthini",
  "n02521646": "gadoid, gadoid fish",
  "n02521916": "Gadidae, family Gadidae",
  "n02522247": "Gadus, genus Gadus",
  "n02522399": "cod, codfish",
  "n02522637": "codling",
  "n02522722": "Atlantic cod, Gadus morhua",
  "n02522866": "Pacific cod, Alaska cod, Gadus macrocephalus",
  "n02522990": "Merlangus, genus Merlangus",
  "n02523110": "whiting, Merlangus merlangus, Gadus merlangus",
  "n02523338": "Lota, genus Lota",
  "n02523427": "burbot, eelpout, ling, cusk, Lota lota",
  "n02523617": "scrod, schrod",
  "n02523750": "Melanogrammus, genus Melanogrammus",
  "n02523877": "haddock, Melanogrammus aeglefinus",
  "n02524081": "Pollachius, genus Pollachius",
  "n02524202": "pollack, pollock, Pollachius pollachius",
  "n02524424": "Merluccius, genus Merluccius",
  "n02524524": "hake",
  "n02524659": "silver hake, Merluccius bilinearis, whiting",
  "n02524811": "Urophycis, genus Urophycis",
  "n02524928": "ling",
  "n02525012": "Molva, genus Molva",
  "n02525120": "ling, Molva molva",
  "n02525287": "Brosmius, genus Browmius",
  "n02525382": "cusk, torsk, Brosme brosme",
  "n02525543": "Macrouridae, family Macrouridae, Macruridae, family Macruridae",
  "n02525703": "grenadier, rattail, rattail fish",
  "n02525866": "Anguilliformes, order Anguilliformes, order Apodes",
  "n02526121": "eel",
  "n02526425": "elver",
  "n02526486": "Anguillidae, family Anguillidae",
  "n02526673": "Anguilla, genus Anguilla",
  "n02526818": "common eel, freshwater eel",
  "n02527057": "tuna, Anguilla sucklandii",
  "n02527145": "Muraenidae, family Muraenidae",
  "n02527271": "moray, moray eel",
  "n02527498": "Congridae, family Congridae",
  "n02527622": "conger, conger eel",
  "n02527813": "Teleostei, subclass Teleostei",
  "n02528163": "teleost fish, teleost, teleostan",
  "n02528534": "Isospondyli, order Isospondyli",
  "n02528949": "Gonorhynchidae, family Gonorhynchidae",
  "n02529111": "Gonorhynchus, genus Gonorhynchus",
  "n02529293": "beaked salmon, sandfish, Gonorhynchus gonorhynchus",
  "n02529515": "Clupeidae, family Clupeidae",
  "n02529772": "clupeid fish, clupeid",
  "n02530052": "whitebait",
  "n02530188": "brit, britt",
  "n02530294": "Alosa, genus Alosa",
  "n02530421": "shad",
  "n02530637": "common American shad, Alosa sapidissima",
  "n02530831": "river shad, Alosa chrysocloris",
  "n02530999": "allice shad, allis shad, allice, allis, Alosa alosa",
  "n02531114": "alewife, Alosa pseudoharengus, Pomolobus pseudoharengus",
  "n02531362": "Pomolobus, genus Pomolobus",
  "n02531503": "Brevoortia, genus Brevoortia",
  "n02531625": "menhaden, Brevoortia tyrannis",
  "n02531820": "Clupea, genus Clupea",
  "n02532028": "herring, Clupea harangus",
  "n02532272": "Atlantic herring, Clupea harengus harengus",
  "n02532451": "Pacific herring, Clupea harengus pallasii",
  "n02532602": "sardine",
  "n02532786": "sild",
  "n02532918": "brisling, sprat, Clupea sprattus",
  "n02533075": "Sardina, genus Sardina, genus Sardinia",
  "n02533209": "pilchard, sardine, Sardina pilchardus",
  "n02533424": "Sardinops, genus Sardinops",
  "n02533545": "Pacific sardine, Sardinops caerulea",
  "n02533708": "Engraulidae, family Engraulidae",
  "n02533834": "anchovy",
  "n02534036": "Engraulis, genus Engraulis",
  "n02534165": "mediterranean anchovy, Engraulis encrasicholus",
  "n02534352": "Salmonidae, family Salmonidae",
  "n02534559": "salmonid",
  "n02534734": "salmon",
  "n02535080": "parr",
  "n02535163": "blackfish",
  "n02535258": "redfish",
  "n02535349": "Salmo, genus Salmo",
  "n02535537": "Atlantic salmon, Salmo salar",
  "n02535759": "landlocked salmon, lake salmon",
  "n02535909": "Oncorhynchus, genus Oncorhynchus",
  "n02536165": "sockeye, sockeye salmon, red salmon, blueback salmon, Oncorhynchus nerka",
  "n02536456": "chinook, chinook salmon, king salmon, quinnat salmon, Oncorhynchus tshawytscha",
  "n02536685": "chum salmon, chum, Oncorhynchus keta",
  "n02536864": "coho, cohoe, coho salmon, blue jack, silver salmon, Oncorhynchus kisutch",
  "n02537085": "trout",
  "n02537319": "brown trout, salmon trout, Salmo trutta",
  "n02537525": "rainbow trout, Salmo gairdneri",
  "n02537716": "sea trout",
  "n02537847": "Salvelinus, genus Salvelinus",
  "n02538010": "lake trout, salmon trout, Salvelinus namaycush",
  "n02538216": "brook trout, speckled trout, Salvelinus fontinalis",
  "n02538406": "char, charr",
  "n02538562": "Arctic char, Salvelinus alpinus",
  "n02538730": "Coregonidae, family Coregonidae",
  "n02538985": "whitefish",
  "n02539251": "Coregonus, genus Coregonus",
  "n02539424": "lake whitefish, Coregonus clupeaformis",
  "n02539573": "cisco, lake herring, Coregonus artedi",
  "n02539752": "Prosopium, genus Prosopium",
  "n02539894": "round whitefish, Menominee whitefish, Prosopium cylindraceum",
  "n02540091": "Rocky Mountain whitefish, Prosopium williamsonii",
  "n02540255": "Osmeridae, family Osmeridae",
  "n02540412": "smelt",
  "n02540637": "Osmerus, genus Osmerus",
  "n02540791": "rainbow smelt, Osmerus mordax",
  "n02540983": "sparling, European smelt, Osmerus eperlanus",
  "n02541139": "Mallotus, genus Mallotus",
  "n02541257": "capelin, capelan, caplin",
  "n02541431": "Elopidae, family Elopidae",
  "n02541583": "genus Tarpon",
  "n02541687": "tarpon, Tarpon atlanticus",
  "n02541875": "Elops, genus Elops",
  "n02542017": "ladyfish, tenpounder, Elops saurus",
  "n02542162": "Albulidae, family Albulidae",
  "n02542283": "Albula, genus Albula",
  "n02542432": "bonefish, Albula vulpes",
  "n02542598": "Argentinidae, family Argentinidae",
  "n02542804": "Argentina, genus Argentina",
  "n02542958": "argentine",
  "n02543093": "Myctophidae, family Myctophidae",
  "n02543255": "lanternfish",
  "n02543412": "Synodontidae, family Synodontidae",
  "n02543565": "lizardfish, snakefish, snake-fish",
  "n02543737": "Chlorophthalmidae, family Chlorophthalmidae",
  "n02543952": "greeneye",
  "n02544086": "Alepisaurus, genus Alepisaurus",
  "n02544274": "lancetfish, lancet fish, wolffish",
  "n02544475": "handsaw fish",
  "n02544596": "Osteoglossiformes, Order Osteoglossiformes",
  "n02544754": "Osteoglossidae, family Osteoglossidae",
  "n02544960": "Scleropages, genus Scleropages",
  "n02545153": "Australian arowana, Dawson River salmon, saratoga, spotted barramundi, spotted bonytongue, Scleropages leichardti",
  "n02545387": "Australian bonytongue, northern barramundi, Scleropages jardinii",
  "n02545569": "Lampridae, family Lampridae",
  "n02545687": "Lampris, genus Lampris",
  "n02545841": "opah, moonfish, Lampris regius",
  "n02546028": "New World opah, Lampris guttatus",
  "n02546177": "Trachipteridae, family Trachipteridae",
  "n02546331": "ribbonfish",
  "n02546477": "Trachipterus, genus Trachipterus",
  "n02546627": "dealfish, Trachipterus arcticus",
  "n02546744": "Regalecidae, family Regalecidae",
  "n02546873": "Reglaecus, genus Regalecus",
  "n02547014": "oarfish, king of the herring, ribbonfish, Regalecus glesne",
  "n02547213": "Pediculati, order Pediculati",
  "n02547562": "Ogcocephalidae, family Ogcocephalidae",
  "n02547733": "batfish",
  "n02547947": "Lophiidae, family Lophiidae",
  "n02548128": "Lophius, genus Lophius",
  "n02548247": "goosefish, angler, anglerfish, angler fish, monkfish, lotte, allmouth, Lophius Americanus",
  "n02548522": "Batrachoididae, family Batrachoididae",
  "n02548689": "toadfish, Opsanus tau",
  "n02548884": "oyster fish, oyster-fish, oysterfish",
  "n02548990": "Antennariidae, family Antennariidae",
  "n02549248": "frogfish",
  "n02549376": "sargassum fish",
  "n02549533": "Synentognathi, order Synentognathi",
  "n02549796": "Belonidae, family Belonidae",
  "n02549989": "needlefish, gar, billfish",
  "n02550203": "timucu",
  "n02550296": "Exocoetidae, family Exocoetidae",
  "n02550460": "flying fish",
  "n02550655": "monoplane flying fish, two-wing flying fish",
  "n02550780": "biplane flying fish, four-wing flying fish",
  "n02550915": "Hemiramphidae, family Hemiramphidae",
  "n02551134": "halfbeak",
  "n02551316": "Scomberesocidae, family Scomberesocidae, Scombresocidae, family Scombresocidae",
  "n02551494": "Scomberesox, genus Scomberesox, Scombresox, genus Scombresox",
  "n02551668": "saury, billfish, Scomberesox saurus",
  "n02551824": "Acanthopterygii, superorder Acanthopterygii",
  "n02552171": "spiny-finned fish, acanthopterygian",
  "n02552737": "Ophiodontidae, family Ophiodontidae",
  "n02552894": "Ophiodon, genus Ophiodon",
  "n02553028": "lingcod, Ophiodon elongatus",
  "n02553196": "Perciformes, order Perciformes, Percomorphi, order Percomorphi",
  "n02554512": "Percoidea, suborder Percoidea",
  "n02554730": "percoid fish, percoid, percoidean",
  "n02555863": "perch",
  "n02556014": "Anabantidae, family Anabantidae",
  "n02556195": "Anabas, genus Anabas",
  "n02556373": "climbing perch, Anabas testudineus, A. testudineus",
  "n02556623": "Percidae, family Percidae",
  "n02556846": "perch",
  "n02557033": "Perca, genus Perca",
  "n02557182": "yellow perch, Perca flavescens",
  "n02557318": "European perch, Perca fluviatilis",
  "n02557461": "Stizostedion, genus Stizostedion",
  "n02557591": "pike-perch, pike perch",
  "n02557749": "walleye, walleyed pike, jack salmon, dory, Stizostedion vitreum",
  "n02557909": "blue pike, blue pickerel, blue pikeperch, blue walleye, Strizostedion vitreum glaucum",
  "n02558079": "Percina, genus Percina",
  "n02558206": "snail darter, Percina tanasi",
  "n02558350": "Trichodontidae, family Trichodontidae",
  "n02558560": "sandfish",
  "n02558724": "Ophidiidae, family Ophidiidae",
  "n02558860": "cusk-eel",
  "n02558980": "Brotulidae, family Brotulidae",
  "n02559144": "brotula",
  "n02559232": "Carapidae, family Carapidae",
  "n02559383": "pearlfish, pearl-fish",
  "n02559606": "Centropomidae, family Centropomidae",
  "n02559862": "robalo",
  "n02559974": "Centropomus, genus Centropomus",
  "n02560110": "snook",
  "n02560271": "Latinae",
  "n02560383": "Lates, genus Lates",
  "n02560546": "barramundi, giant perch, giant seaperch, Asian seabass, white seabass, Lates calcarifer",
  "n02560823": "Esocidae, family Esocidae",
  "n02560964": "Esox, genus Esox",
  "n02561108": "pike",
  "n02561381": "northern pike, Esox lucius",
  "n02561514": "muskellunge, Esox masquinongy",
  "n02561661": "pickerel",
  "n02561803": "chain pickerel, chain pike, Esox niger",
  "n02561937": "redfin pickerel, barred pickerel, Esox americanus",
  "n02562085": "Centrarchidae, family Centrarchidae",
  "n02562315": "sunfish, centrarchid",
  "n02562680": "Pomoxis, genus Pomoxis",
  "n02562796": "crappie",
  "n02562971": "black crappie, Pomoxis nigromaculatus",
  "n02563079": "white crappie, Pomoxis annularis",
  "n02563182": "freshwater bream, bream",
  "n02563497": "Lepomis, genus Lepomis",
  "n02563648": "pumpkinseed, Lepomis gibbosus",
  "n02563792": "bluegill, Lepomis macrochirus",
  "n02563949": "spotted sunfish, stumpknocker, Lepomis punctatus",
  "n02564130": "Ambloplites, genus Ambloplites",
  "n02564270": "freshwater bass",
  "n02564403": "rock bass, rock sunfish, Ambloplites rupestris",
  "n02564572": "Micropterus, genus Micropterus",
  "n02564720": "black bass",
  "n02564935": "Kentucky black bass, spotted black bass, Micropterus pseudoplites",
  "n02565072": "smallmouth, smallmouth bass, smallmouthed bass, smallmouth black bass, smallmouthed black bass, Micropterus dolomieu",
  "n02565324": "largemouth, largemouth bass, largemouthed bass, largemouth black bass, largemouthed black bass, Micropterus salmoides",
  "n02565573": "bass",
  "n02565728": "Serranidae, family Serranidae",
  "n02566109": "serranid fish, serranid",
  "n02566325": "Morone, genus Morone",
  "n02566489": "white perch, silver perch, Morone americana",
  "n02566665": "yellow bass, Morone interrupta",
  "n02566834": "sea bass",
  "n02567201": "Synagrops, genus Synagrops",
  "n02567334": "blackmouth bass, Synagrops bellus",
  "n02567484": "Centropristis, genus Centropristis",
  "n02567633": "rock sea bass, rock bass, Centropristis philadelphica",
  "n02567772": "black sea bass, black bass, Centropistes striata",
  "n02567960": "Roccus, genus Roccus",
  "n02568087": "striped bass, striper, Roccus saxatilis, rockfish",
  "n02568326": "Polyprion, genus Polyprion",
  "n02568447": "stone bass, wreckfish, Polyprion americanus",
  "n02568636": "Serranus, genus Serranus",
  "n02568807": "belted sandfish, Serranus subligarius",
  "n02568959": "grouper",
  "n02569151": "Epinephelus, genus Epinephelus",
  "n02569334": "coney, Epinephelus fulvus",
  "n02569484": "hind",
  "n02569631": "rock hind, Epinephelus adscensionis",
  "n02569770": "Paranthias, genus Paranthias",
  "n02569905": "creole-fish, Paranthias furcifer",
  "n02570038": "Mycteroperca, genus Mycteroperca",
  "n02570164": "jewfish, Mycteroperca bonaci",
  "n02570312": "Rypticus, genus Rypticus",
  "n02570484": "soapfish",
  "n02570648": "Embiotocidae, family Embiotocidae",
  "n02570838": "surfperch, surffish, surf fish",
  "n02571034": "Hipsurus, genus Hipsurus",
  "n02571167": "rainbow seaperch, rainbow perch, Hipsurus caryi",
  "n02571300": "Priacanthidae, family Priacanthidae",
  "n02571486": "Priacanthus, genus Priacanthus",
  "n02571652": "bigeye",
  "n02571810": "catalufa, Priacanthus arenatus",
  "n02571983": "Apogonidae, family Apogonidae",
  "n02572196": "cardinalfish",
  "n02572369": "Apogon, genus Apogon",
  "n02572484": "flame fish, flamefish, Apogon maculatus",
  "n02572628": "Astropogon, genus Astropogon",
  "n02572763": "conchfish, Astropogon stellatus",
  "n02572904": "Malacanthidae, family Malacanthidae",
  "n02573075": "Lopholatilus, genus Lopholatilus",
  "n02573249": "tilefish, Lopholatilus chamaeleonticeps",
  "n02573406": "Pomatomidae, family Pomatomidae",
  "n02573563": "Pomatomus, genus Pomatomus",
  "n02573704": "bluefish, Pomatomus saltatrix",
  "n02573918": "Rachycentridae, family Rachycentridae",
  "n02574093": "Rachycentron, genus Rachycentron",
  "n02574271": "cobia, Rachycentron canadum, sergeant fish",
  "n02574489": "Discocephali, order Discocephali",
  "n02574651": "Echeneididae, family Echeneididae, family Echeneidae",
  "n02574910": "remora, suckerfish, sucking fish",
  "n02575168": "Echeneis, genus Echeneis",
  "n02575325": "sharksucker, Echeneis naucrates",
  "n02575455": "Remilegia, genus Remilegia",
  "n02575590": "whale sucker, whalesucker, Remilegia australis",
  "n02575766": "Carangidae, family Carangidae",
  "n02576223": "carangid fish, carangid",
  "n02576460": "Caranx, genus Caranx",
  "n02576575": "jack",
  "n02576906": "crevalle jack, jack crevalle, Caranx hippos",
  "n02577041": "yellow jack, Caranx bartholomaei",
  "n02577164": "runner, blue runner, Caranx crysos",
  "n02577291": "Elagatis, genus Elagatis",
  "n02577403": "rainbow runner, Elagatis bipinnulata",
  "n02577532": "Oligoplites, genus Oligoplites",
  "n02577662": "leatherjacket, leatherjack",
  "n02577823": "Alectis, genus Alectis",
  "n02577952": "threadfish, thread-fish, Alectis ciliaris",
  "n02578125": "Selene, genus Selene",
  "n02578233": "moonfish, Atlantic moonfish, horsefish, horsehead, horse-head, dollarfish, Selene setapinnis",
  "n02578454": "lookdown, lookdown fish, Selene vomer",
  "n02578604": "Seriola, genus Seriola",
  "n02578771": "amberjack, amberfish",
  "n02578928": "yellowtail, Seriola dorsalis",
  "n02579091": "rudderfish, banded rudderfish, Seriola zonata",
  "n02579303": "kingfish, Seriola grandis",
  "n02579420": "Trachinotus, genus Trachinotus",
  "n02579557": "pompano",
  "n02579762": "Florida pompano, Trachinotus carolinus",
  "n02579928": "permit, Trachinotus falcatus",
  "n02580055": "Naucrates, genus Naucrates",
  "n02580188": "pilotfish, Naucrates ductor",
  "n02580336": "scad",
  "n02580546": "Trachurus, genus Trachurus",
  "n02580679": "horse mackerel, jack mackerel, Spanish mackerel, saurel, Trachurus symmetricus",
  "n02580830": "horse mackerel, saurel, Trachurus trachurus",
  "n02580991": "Selar, genus Selar",
  "n02581108": "bigeye scad, big-eyed scad, goggle-eye, Selar crumenophthalmus",
  "n02581289": "Decapterus, genus Decapterus",
  "n02581482": "mackerel scad, mackerel shad, Decapterus macarellus",
  "n02581642": "round scad, cigarfish, quiaquia, Decapterus punctatus",
  "n02581803": "Coryphaenidae, family Coryphaenidae",
  "n02581957": "dolphinfish, dolphin, mahimahi",
  "n02582220": "Coryphaena hippurus",
  "n02582349": "Coryphaena equisetis",
  "n02582437": "Bramidae, family Bramidae",
  "n02582591": "Brama, genus Brama",
  "n02582721": "pomfret, Brama raii",
  "n02582919": "Branchiostegidae, family Branchiostegidae",
  "n02583096": "blanquillo, tilefish",
  "n02583211": "Characidae, family Characidae",
  "n02583457": "Characinidae, family Characinidae",
  "n02583567": "characin, characin fish, characid",
  "n02583768": "Hemigrammus, genus Hemigrammus",
  "n02583890": "tetra",
  "n02584004": "Paracheirodon, genus Paracheirodon",
  "n02584145": "cardinal tetra, Paracheirodon axelrodi",
  "n02584325": "Serrasalmus, genus Serrasalmus",
  "n02584449": "piranha, pirana, caribe",
  "n02584643": "tentacle",
  "n02584915": "antenna, feeler",
  "n02585168": "arista",
  "n02585285": "barbel, feeler",
  "n02585446": "swimmeret, pleopod",
  "n02585732": "Cichlidae, family Cichlidae",
  "n02585872": "cichlid, cichlid fish",
  "n02586129": "Tilapia, genus Tilapia",
  "n02586238": "bolti, Tilapia nilotica",
  "n02586382": "Lutjanidae, family Lutjanidae",
  "n02586543": "snapper",
  "n02586865": "Lutjanus, genus Lutjanus",
  "n02587051": "red snapper, Lutjanus blackfordi",
  "n02587300": "grey snapper, gray snapper, mangrove snapper, Lutjanus griseus",
  "n02587479": "mutton snapper, muttonfish, Lutjanus analis",
  "n02587618": "schoolmaster, Lutjanus apodus",
  "n02587761": "Ocyurus, genus Ocyurus",
  "n02587877": "yellowtail, yellowtail snapper, Ocyurus chrysurus",
  "n02588108": "Haemulidae, family Haemulidae",
  "n02588286": "grunt",
  "n02588580": "Haemulon, genus Haemulon",
  "n02588794": "margate, Haemulon album",
  "n02588945": "Spanish grunt, Haemulon macrostomum",
  "n02589062": "tomtate, Haemulon aurolineatum",
  "n02589196": "cottonwick, Haemulon malanurum",
  "n02589316": "sailor's-choice, sailors choice, Haemulon parra",
  "n02589486": "Anisotremus, genus Anisotremus",
  "n02589623": "porkfish, pork-fish, Anisotremus virginicus",
  "n02589796": "pompon, black margate, Anisotremus surinamensis",
  "n02589955": "Orthopristis, genus Orthopristis",
  "n02590094": "pigfish, hogfish, Orthopristis chrysopterus",
  "n02590237": "Sparidae, family Sparidae",
  "n02590495": "sparid, sparid fish",
  "n02590702": "sea bream, bream",
  "n02590987": "porgy",
  "n02591205": "Pagrus, genus Pagrus",
  "n02591330": "red porgy, Pagrus pagrus",
  "n02591493": "Pagellus, genus Pagellus",
  "n02591613": "European sea bream, Pagellus centrodontus",
  "n02591757": "Archosargus, genus Archosargus",
  "n02591911": "Atlantic sea bream, Archosargus rhomboidalis",
  "n02592055": "sheepshead, Archosargus probatocephalus",
  "n02592244": "Lagodon, genus Lagodon",
  "n02592371": "pinfish, sailor's-choice, squirrelfish, Lagodon rhomboides",
  "n02592607": "Calamus, genus Calamus",
  "n02592734": "sheepshead porgy, Calamus penna",
  "n02592866": "Chrysophrys, genus Chrysophrys",
  "n02593019": "snapper, Chrysophrys auratus",
  "n02593191": "black bream, Chrysophrys australis",
  "n02593353": "Stenotomus, genus Stenotomus",
  "n02593453": "scup, northern porgy, northern scup, Stenotomus chrysops",
  "n02593679": "scup, southern porgy, southern scup, Stenotomus aculeatus",
  "n02593863": "Sciaenidae, family Sciaenidae",
  "n02594250": "sciaenid fish, sciaenid",
  "n02594552": "drum, drumfish",
  "n02594807": "Equetus, genus Equetus",
  "n02594942": "striped drum, Equetus pulcher",
  "n02595056": "jackknife-fish, Equetus lanceolatus",
  "n02595217": "Bairdiella, genus Bairdiella",
  "n02595339": "silver perch, mademoiselle, Bairdiella chrysoura",
  "n02595569": "Sciaenops, genus Sciaenops",
  "n02595702": "red drum, channel bass, redfish, Sciaenops ocellatus",
  "n02595902": "Sciaena, genus Sciaena",
  "n02596067": "mulloway, jewfish, Sciaena antarctica",
  "n02596252": "maigre, maiger, Sciaena aquila",
  "n02596381": "croaker",
  "n02596592": "Micropogonias, genus Micropogonias",
  "n02596720": "Atlantic croaker, Micropogonias undulatus",
  "n02596888": "Umbrina, genus Umbrina",
  "n02597004": "yellowfin croaker, surffish, surf fish, Umbrina roncador",
  "n02597173": "Menticirrhus, genus Menticirrhus",
  "n02597367": "whiting",
  "n02597608": "kingfish",
  "n02597818": "king whiting, Menticirrhus americanus",
  "n02597972": "northern whiting, Menticirrhus saxatilis",
  "n02598134": "corbina, Menticirrhus undulatus",
  "n02598252": "silver whiting, Menticirrhus littoralis",
  "n02598438": "Genyonemus, genus Genyonemus",
  "n02598573": "white croaker, chenfish, kingfish, Genyonemus lineatus",
  "n02598747": "Seriphus, genus Seriphus",
  "n02598878": "white croaker, queenfish, Seriphus politus",
  "n02599052": "sea trout",
  "n02599207": "Cynoscion, genus Cynoscion",
  "n02599347": "weakfish, Cynoscion regalis",
  "n02599557": "spotted weakfish, spotted sea trout, spotted squeateague, Cynoscion nebulosus",
  "n02599784": "Mullidae, family Mullidae",
  "n02599958": "mullet",
  "n02600135": "Mullus, genus Mullus",
  "n02600298": "goatfish, red mullet, surmullet, Mullus surmuletus",
  "n02600503": "red goatfish, Mullus auratus",
  "n02600657": "Mulloidichthys, genus Mulloidichthys",
  "n02600798": "yellow goatfish, Mulloidichthys martinicus",
  "n02600953": "Mugiloidea, suborder Mugiloidea",
  "n02601200": "Mugilidae, family Mugilidae",
  "n02601344": "mullet, grey mullet, gray mullet",
  "n02601589": "Mugil, genus Mugil",
  "n02601767": "striped mullet, Mugil cephalus",
  "n02601921": "white mullet, Mugil curema",
  "n02602059": "liza, Mugil liza",
  "n02602215": "Atherinidae, family Atherinidae",
  "n02602405": "silversides, silverside",
  "n02602620": "Atherinopsis, genus Atherinopsis",
  "n02602760": "jacksmelt, Atherinopsis californiensis",
  "n02602970": "Sphyraenidae, family Sphyraenidae",
  "n02603174": "Sphyraena, genus Sphyraena",
  "n02603317": "barracuda",
  "n02603540": "great barracuda, Sphyraena barracuda",
  "n02603737": "Pempheridae, family Pempheridae",
  "n02603862": "sweeper",
  "n02604014": "Kyphosidae, family Kyphosidae",
  "n02604157": "sea chub",
  "n02604342": "Kyphosus, genus Kyphosus",
  "n02604480": "Bermuda chub, rudderfish, Kyphosus sectatrix",
  "n02604657": "Ephippidae, family Ephippidae",
  "n02604811": "Chaetodipterus, genus Chaetodipterus",
  "n02604954": "spadefish, angelfish, Chaetodipterus faber",
  "n02605139": "Chaetodontidae, family Chaetodontidae",
  "n02605316": "butterfly fish",
  "n02605571": "genus Chaetodon",
  "n02605703": "chaetodon",
  "n02605809": "Pomacanthus, genus Pomacanthus",
  "n02605936": "angelfish",
  "n02606052": "rock beauty, Holocanthus tricolor",
  "n02606194": "Pomacentridae, family Pomacentridae",
  "n02606384": "damselfish, demoiselle",
  "n02606590": "Pomacentrus, genus Pomacentrus",
  "n02606751": "beaugregory, Pomacentrus leucostictus",
  "n02606926": "Amphiprion, genus Amphiprion",
  "n02607072": "anemone fish",
  "n02607201": "clown anemone fish, Amphiprion percula",
  "n02607345": "Abudefduf, genus Abudefduf",
  "n02607470": "sergeant major, Abudefduf saxatilis",
  "n02607630": "Labridae, family Labridae",
  "n02607862": "wrasse",
  "n02608151": "Achoerodus, genus Achoerodus",
  "n02608284": "pigfish, giant pigfish, Achoerodus gouldii",
  "n02608429": "Lachnolaimus, genus Lachnolaimus",
  "n02608547": "hogfish, hog snapper, Lachnolaimus maximus",
  "n02608708": "Halicoeres, genus Halicoeres",
  "n02608860": "slippery dick, Halicoeres bivittatus",
  "n02608996": "puddingwife, pudding-wife, Halicoeres radiatus",
  "n02609169": "Thalassoma, genus Thalassoma",
  "n02609302": "bluehead, Thalassoma bifasciatum",
  "n02609466": "Hemipteronatus, genus Hemipteronatus",
  "n02609617": "razor fish, razor-fish",
  "n02609823": "pearly razorfish, Hemipteronatus novacula",
  "n02609951": "Tautoga, genus Tautoga",
  "n02610066": "tautog, blackfish, Tautoga onitis",
  "n02610234": "Tautogolabrus, genus Tautogolabrus",
  "n02610373": "cunner, bergall, Tautogolabrus adspersus",
  "n02610541": "Scaridae, family Scaridae",
  "n02610664": "parrotfish, polly fish, pollyfish",
  "n02610834": "Polynemidae, family Polynemidae",
  "n02610980": "threadfin",
  "n02611154": "Polydactylus, genus Polydactylus",
  "n02611294": "barbu, Polydactylus virginicus",
  "n02611425": "Opisthognathidae, family Opisthognathidae",
  "n02611561": "jawfish",
  "n02611767": "Uranoscopidae, family Uranoscopidae",
  "n02611898": "stargazer",
  "n02612046": "Dactyloscopidae, family Dactyloscopidae",
  "n02612167": "sand stargazer",
  "n02612393": "Blennioidea, suborder Blennioidea",
  "n02612657": "blennioid fish, blennioid",
  "n02612982": "Blenniidae, family Blenniidae",
  "n02613181": "blenny, combtooth blenny",
  "n02613434": "Blennius, genus Blennius",
  "n02613572": "shanny, Blennius pholis",
  "n02613687": "Scartella, genus Scartella",
  "n02613820": "Molly Miller, Scartella cristata",
  "n02613960": "Clinidae, family Clinidae",
  "n02614140": "clinid, clinid fish",
  "n02614288": "Chaenopsis, genus Chaenopsis",
  "n02614482": "pikeblenny",
  "n02614653": "bluethroat pikeblenny, Chaenopsis ocellata",
  "n02614788": "Pholidae, family Pholidae, family Pholididae",
  "n02614978": "gunnel, bracketed blenny",
  "n02615157": "Pholis, genus Pholis",
  "n02615298": "rock gunnel, butterfish, Pholis gunnellus",
  "n02615494": "Stichaeidae, family Stichaeidae",
  "n02615642": "prickleback",
  "n02615829": "Lumpenus, genus Lumpenus",
  "n02615980": "snakeblenny, Lumpenus lumpretaeformis",
  "n02616128": "eelblenny",
  "n02616251": "Cryptacanthodes, genus Cryptacanthodes",
  "n02616397": "wrymouth, ghostfish, Cryptacanthodes maculatus",
  "n02616572": "Anarhichadidae, family Anarhichadidae",
  "n02616705": "Anarhichas, genus Anarhichas",
  "n02616851": "wolffish, wolf fish, catfish",
  "n02617029": "Zoarcidae, family Zoarcidae",
  "n02617207": "eelpout, pout",
  "n02617402": "Zoarces, genus Zoarces",
  "n02617537": "viviparous eelpout, Zoarces viviparus",
  "n02617689": "Gymnelis, genus Gymnelis",
  "n02617819": "fish doctor, Gymnelis viridis",
  "n02617956": "Macrozoarces, genus Macrozoarces",
  "n02618094": "ocean pout, Macrozoarces americanus",
  "n02618244": "Ammodytidae, family Ammodytidae",
  "n02618372": "Ammodytes, genus Ammodytes",
  "n02618513": "sand lance, sand launce, sand eel, launce",
  "n02618697": "Callionymidae, family Callionymidae",
  "n02618827": "dragonet",
  "n02619029": "Gobiidae, family Gobiidae",
  "n02619165": "goby, gudgeon",
  "n02619409": "Periophthalmus, genus Periophthalmus",
  "n02619550": "mudskipper, mudspringer",
  "n02619738": "Eleotridae, family Eleotridae",
  "n02619861": "sleeper, sleeper goby",
  "n02620033": "Percophidae, family Percophidae",
  "n02620167": "flathead",
  "n02620318": "Toxotidae, family Toxotidae",
  "n02620443": "Toxotes, genus Toxotes",
  "n02620578": "archerfish, Toxotes jaculatrix",
  "n02620826": "Microdesmidae, family Microdesmidae",
  "n02620956": "worm fish",
  "n02621107": "Acanthuridae, family Acanthuridae",
  "n02621258": "surgeonfish",
  "n02621419": "Acanthurus, genus Acanthurus",
  "n02621577": "doctorfish, doctor-fish, Acanthurus chirurgus",
  "n02621721": "Gempylidae, family Gempylidae",
  "n02621908": "gempylid",
  "n02622130": "Gempylus, genus Gempylus",
  "n02622249": "snake mackerel, Gempylus serpens",
  "n02622408": "Lepidocybium, genus Lepidocybium",
  "n02622547": "escolar, Lepidocybium flavobrunneum",
  "n02622712": "oilfish, Ruvettus pretiosus",
  "n02622823": "Trichiuridae, family Trichiuridae",
  "n02622955": "cutlassfish, frost fish, hairtail",
  "n02623170": "Scombroidea, suborder Scombroidea",
  "n02623445": "scombroid, scombroid fish",
  "n02623868": "Scombridae, family Scombridae",
  "n02624167": "mackerel",
  "n02624377": "Scomber, genus Scomber",
  "n02624551": "common mackerel, shiner, Scomber scombrus",
  "n02624807": "Spanish mackerel, Scomber colias",
  "n02624987": "chub mackerel, tinker, Scomber japonicus",
  "n02625132": "Acanthocybium, genus Acanthocybium",
  "n02625258": "wahoo, Acanthocybium solandri",
  "n02625418": "Scomberomorus, genus Scomberomorus",
  "n02625612": "Spanish mackerel",
  "n02625851": "king mackerel, cavalla, cero, Scomberomorus cavalla",
  "n02626089": "Scomberomorus maculatus",
  "n02626265": "cero, pintado, kingfish, Scomberomorus regalis",
  "n02626471": "sierra, Scomberomorus sierra",
  "n02626590": "Thunnus, genus Thunnus",
  "n02626762": "tuna, tunny",
  "n02627037": "albacore, long-fin tunny, Thunnus alalunga",
  "n02627292": "bluefin, bluefin tuna, horse mackerel, Thunnus thynnus",
  "n02627532": "yellowfin, yellowfin tuna, Thunnus albacares",
  "n02627686": "Sarda, genus Sarda",
  "n02627835": "bonito",
  "n02628062": "skipjack, Atlantic bonito, Sarda sarda",
  "n02628259": "Chile bonito, Chilean bonito, Pacific bonito, Sarda chiliensis",
  "n02628467": "Euthynnus, genus Euthynnus",
  "n02628600": "skipjack, skipjack tuna, Euthynnus pelamis",
  "n02628856": "Katsuwonus, genus Katsuwonus",
  "n02629048": "Katsuwonidae, family Kasuwonidae",
  "n02629230": "bonito, oceanic bonito, Katsuwonus pelamis",
  "n02629435": "Xiphiidae, family Xiphiidae",
  "n02629581": "Xiphias, genus Xiphias",
  "n02629716": "swordfish, Xiphias gladius",
  "n02630052": "Istiophoridae, family Istiophoridae",
  "n02630281": "sailfish",
  "n02630468": "Istiophorus, genus Istiophorus",
  "n02630615": "Atlantic sailfish, Istiophorus albicans",
  "n02630739": "billfish",
  "n02630926": "Makaira, genus Makaira",
  "n02631041": "marlin",
  "n02631330": "blue marlin, Makaira nigricans",
  "n02631475": "black marlin, Makaira mazara, Makaira marlina",
  "n02631628": "striped marlin, Makaira mitsukurii",
  "n02631775": "white marlin, Makaira albida",
  "n02631899": "Tetrapturus, genus Tetrapturus",
  "n02632039": "spearfish",
  "n02632239": "Luvaridae, family Luvaridae",
  "n02632359": "Luvarus, genus Luvarus",
  "n02632494": "louvar, Luvarus imperialis",
  "n02632694": "Stromateidae, family Stromateidae",
  "n02632989": "butterfish, stromateid fish, stromateid",
  "n02633287": "Poronotus, genus Poronotus",
  "n02633422": "dollarfish, Poronotus triacanthus",
  "n02633555": "genus Palometa",
  "n02633677": "palometa, California pompano, Palometa simillima",
  "n02633844": "Paprilus, genus Paprilus",
  "n02633977": "harvestfish, Paprilus alepidotus",
  "n02634156": "Psenes, genus Psenes",
  "n02634285": "driftfish",
  "n02634414": "Ariomma, genus Ariomma",
  "n02634545": "driftfish",
  "n02634717": "Tetragonurus, genus Tetragonurus",
  "n02634858": "squaretail",
  "n02635013": "Hyperoglyphe, genus Hyperoglyphe",
  "n02635154": "barrelfish, black rudderfish, Hyperglyphe perciformis",
  "n02635310": "Gobiesocidae, family Gobiesocidae",
  "n02635459": "Gobiesox, genus Gobiesox",
  "n02635580": "clingfish",
  "n02635781": "skillet fish, skilletfish, Gobiesox strumosus",
  "n02635911": "Lobotidae, family Lobotidae",
  "n02636035": "Lobotes, genus Lobotes",
  "n02636170": "tripletail",
  "n02636405": "Atlantic tripletail, Lobotes surinamensis",
  "n02636550": "Pacific tripletail, Lobotes pacificus",
  "n02636666": "Gerreidae, family Gerreidae, Gerridae, family Gerridae",
  "n02636854": "mojarra",
  "n02637046": "Gerres, genus Gerres",
  "n02637179": "yellowfin mojarra, Gerres cinereus",
  "n02637337": "Eucinostomus, genus Eucinostomus",
  "n02637475": "silver jenny, Eucinostomus gula",
  "n02637637": "Sillaginidae, family Sillaginidae",
  "n02637839": "Sillago, genus Sillago",
  "n02637977": "whiting",
  "n02638104": "ganoin, ganoine",
  "n02638323": "Ganoidei, order Ganoidei",
  "n02638596": "ganoid, ganoid fish",
  "n02638835": "Amiidae, family Amiidae",
  "n02638960": "Amia, genus Amia",
  "n02639087": "bowfin, grindle, dogfish, Amia calva",
  "n02639312": "Polyodontidae, family Polyodontidae",
  "n02639464": "Polyodon, genus Polyodon",
  "n02639605": "paddlefish, duckbill, Polyodon spathula",
  "n02639786": "Psephurus, genus Psephurus",
  "n02639922": "Chinese paddlefish, Psephurus gladis",
  "n02640093": "Acipenseridae, family Acipenseridae",
  "n02640242": "sturgeon",
  "n02640453": "Acipenser, genus Acipenser",
  "n02640626": "Pacific sturgeon, white sturgeon, Sacramento sturgeon, Acipenser transmontanus",
  "n02640857": "beluga, hausen, white sturgeon, Acipenser huso",
  "n02641063": "Lepisosteidae, family Lepisosteidae",
  "n02641215": "Lepisosteus, genus Lepisosteus",
  "n02641379": "gar, garfish, garpike, billfish, Lepisosteus osseus",
  "n02641608": "Scleroparei, order Scleroparei",
  "n02641825": "Scorpaenoidea, suborder Scorpaenoidea",
  "n02642107": "scorpaenoid, scorpaenoid fish",
  "n02642430": "Scorpaenidae, family Scorpaenidae",
  "n02642644": "scorpaenid, scorpaenid fish",
  "n02642935": "Scorpaena, genus Scorpaena",
  "n02643112": "scorpionfish, scorpion fish, sea scorpion",
  "n02643316": "plumed scorpionfish, Scorpaena grandicornis",
  "n02643448": "Pterois, genus Pterois",
  "n02643566": "lionfish",
  "n02643713": "Synanceja, genus Synanceja",
  "n02643836": "stonefish, Synanceja verrucosa",
  "n02643989": "Sebastodes, genus Sebastodes",
  "n02644113": "rockfish",
  "n02644360": "copper rockfish, Sebastodes caurinus",
  "n02644501": "vermillion rockfish, rasher, Sebastodes miniatus",
  "n02644665": "red rockfish, Sebastodes ruberrimus",
  "n02644817": "rosefish, ocean perch, Sebastodes marinus",
  "n02644967": "Cottidae, family Cottidae",
  "n02645143": "Cottus, genus Cottus",
  "n02645304": "sculpin",
  "n02645538": "bullhead",
  "n02645691": "miller's-thumb",
  "n02645823": "Hemitripterus, genus Hemitripterus",
  "n02645953": "sea raven, Hemitripterus americanus",
  "n02646117": "Myxocephalus, genus Myxocephalus",
  "n02646241": "grubby, Myxocephalus aenaeus",
  "n02646377": "Cyclopteridae, family Cyclopteridae",
  "n02646508": "Cyclopterus, genus Cyclopterus",
  "n02646667": "lumpfish, Cyclopterus lumpus",
  "n02646892": "lumpsucker",
  "n02646985": "Liparididae, family Liparididae, Liparidae, family Liparidae",
  "n02647144": "Liparis, genus Liparis",
  "n02647294": "snailfish, seasnail, sea snail, Liparis liparis",
  "n02647503": "Agonidae, family Agonidae",
  "n02647660": "poacher, sea poacher, sea poker",
  "n02647903": "Agonus, genus Agonus",
  "n02648035": "pogge, armed bullhead, Agonus cataphractus",
  "n02648174": "Aspidophoroides, genus Aspidophoroides",
  "n02648313": "alligatorfish, Aspidophoroides monopterygius",
  "n02648456": "Hexagrammidae, family Hexagrammidae",
  "n02648625": "greenling",
  "n02648769": "Hexagrammos, genus Hexagrammos",
  "n02648916": "kelp greenling, Hexagrammos decagrammus",
  "n02649082": "Oxylebius, genus Oxylebius",
  "n02649218": "painted greenling, convict fish, convictfish, Oxylebius pictus",
  "n02649400": "Platycephalidae, family Platycephalidae",
  "n02649546": "flathead",
  "n02649689": "Triglidae, family Triglidae",
  "n02650050": "gurnard",
  "n02650282": "Triga, genus Triga",
  "n02650413": "tub gurnard, yellow gurnard, Trigla lucerna",
  "n02650541": "sea robin, searobin",
  "n02650696": "Triglinae, subfamily Triglinae",
  "n02650928": "Prionotus, genus Prionotus",
  "n02651060": "northern sea robin, Prionotus carolinus",
  "n02651215": "Peristediinae, subfamily Peristediinae",
  "n02651412": "Peristedion, genus Peristedion",
  "n02651617": "armored searobin, armored sea robin, Peristedion miniatum",
  "n02651846": "Dactylopteridae, family Dactylopteridae",
  "n02652005": "Dactylopterus, genus Dactylopterus",
  "n02652132": "flying gurnard, flying robin, butterflyfish",
  "n02652335": "Plectognathi, order Plectognathi, order Tetraodontiformes",
  "n02652668": "plectognath, plectognath fish",
  "n02652979": "Balistidae, family Balistidae",
  "n02653145": "triggerfish",
  "n02653359": "Balistes, genus Balistes",
  "n02653497": "queen triggerfish, Bessy cerca, oldwench, oldwife, Balistes vetula",
  "n02653655": "Monocanthidae, family Monocanthidae",
  "n02653786": "filefish",
  "n02653965": "Monocanthus, genus Monocanthus",
  "n02654112": "leatherjacket, leatherfish",
  "n02654256": "Ostraciidae, family Ostraciidae, family Ostraciontidae",
  "n02654425": "boxfish, trunkfish",
  "n02654609": "Lactophrys, genus Lactophrys",
  "n02654745": "cowfish, Lactophrys quadricornis",
  "n02654890": "Tetraodontidae, family Tetraodontidae",
  "n02655020": "puffer, pufferfish, blowfish, globefish",
  "n02655355": "Diodontidae, family Diodontidae",
  "n02655523": "spiny puffer",
  "n02655694": "Diodon, genus Diodon",
  "n02655848": "porcupinefish, porcupine fish, Diodon hystrix",
  "n02656032": "balloonfish, Diodon holocanthus",
  "n02656171": "Chilomycterus, genus Chilomycterus",
  "n02656301": "burrfish",
  "n02656426": "Molidae, family Molidae",
  "n02656550": "genus Mola",
  "n02656670": "ocean sunfish, sunfish, mola, headfish",
  "n02656969": "sharptail mola, Mola lanceolata",
  "n02657083": "Heterosomata, order Heterosomata, order Pleuronectiformes",
  "n02657368": "flatfish",
  "n02657694": "flounder",
  "n02657805": "Pleuronectidae, family Pleuronectidae",
  "n02658079": "righteye flounder, righteyed flounder",
  "n02658381": "Pleuronectes, genus Pleuronectes",
  "n02658531": "plaice, Pleuronectes platessa",
  "n02658670": "Platichthys, genus Platichthys",
  "n02658811": "European flatfish, Platichthys flesus",
  "n02658944": "Limanda, genus Limanda",
  "n02659176": "yellowtail flounder, Limanda ferruginea",
  "n02659342": "Pseudopleuronectes, genus Pseudopleuronectes",
  "n02659478": "winter flounder, blackback flounder, lemon sole, Pseudopleuronectes americanus",
  "n02659667": "Microstomus, genus Microstomus",
  "n02659808": "lemon sole, Microstomus kitt",
  "n02659961": "Hippoglossoides, genus Hippoglossoides",
  "n02660091": "American plaice, Hippoglossoides platessoides",
  "n02660208": "halibut, holibut",
  "n02660412": "Hippoglossus, genus Hippoglossus",
  "n02660519": "Atlantic halibut, Hippoglossus hippoglossus",
  "n02660640": "Pacific halibut, Hippoglossus stenolepsis",
  "n02660769": "Bothidae, family Bothidae",
  "n02661017": "lefteye flounder, lefteyed flounder",
  "n02661317": "Paralichthys, genus Paralichthys",
  "n02661473": "southern flounder, Paralichthys lethostigmus",
  "n02661618": "summer flounder, Paralichthys dentatus",
  "n02661765": "Etropus, genus Etropus",
  "n02661892": "grey flounder, gray flounder, Etropus rimosus",
  "n02662081": "Citharichthys, genus Citharichthys",
  "n02662239": "whiff",
  "n02662397": "horned whiff, Citharichthys cornutus",
  "n02662559": "sand dab",
  "n02662688": "Scophthalmus, genus Scophthalmus",
  "n02662825": "windowpane, Scophthalmus aquosus",
  "n02662993": "brill, Scophthalmus rhombus",
  "n02663086": "Psetta, genus Psetta",
  "n02663211": "turbot, Psetta maxima",
  "n02663352": "Cynoglossidae, family Cynoglossidae",
  "n02663485": "tonguefish, tongue-fish",
  "n02663657": "Soleidae, family Soleidae",
  "n02663849": "sole",
  "n02664136": "Solea, genus Solea",
  "n02664285": "European sole, Solea solea",
  "n02664399": "lemon sole, Solea lascaris",
  "n02664511": "Parophrys, genus Parophrys",
  "n02664642": "English sole, lemon sole, Parophrys vitulus",
  "n02664823": "Psettichthys, genus Psettichthys",
  "n02664960": "sand sole, Psettichthys melanostichus",
  "n02665119": "Trinectes, genus Trinectes",
  "n02665250": "hogchoker, Trinectes maculatus",
  "n02665413": "thick skin",
  "n02665543": "thorax",
  "n02665687": "prothorax",
  "n02665812": "metamere, somite",
  "n02665985": "aba",
  "n02666079": "aba",
  "n02666196": "abacus",
  "n02666347": "abacus",
  "n02666501": "abandoned ship, derelict",
  "n02666624": "A battery",
  "n02666735": "abattis, abatis",
  "n02666943": "abattoir, butchery, shambles, slaughterhouse",
  "n02667093": "abaya",
  "n02667244": "Abbe condenser",
  "n02667379": "abbey",
  "n02667478": "abbey",
  "n02667576": "abbey",
  "n02667693": "Abney level",
  "n02667906": "abortifacient, aborticide, abortion-inducing drug",
  "n02668093": "abortion pill, mifepristone, RU 486",
  "n02668393": "abrader, abradant",
  "n02668613": "abrading stone",
  "n02668747": "Abstract Expressionism, action painting",
  "n02669001": "abstraction",
  "n02669079": "abstractionism, abstract art",
  "n02669295": "abutment",
  "n02669442": "abutment arch",
  "n02669534": "academic costume",
  "n02669723": "academic gown, academic robe, judge's robe",
  "n02669885": "academy",
  "n02670049": "Acapulco gold, Mexican green",
  "n02670186": "accelerator, throttle, throttle valve",
  "n02670382": "accelerator, particle accelerator, atom smasher",
  "n02670683": "accelerator, accelerator pedal, gas pedal, gas, throttle, gun",
  "n02670935": "accelerometer",
  "n02671062": "access, approach",
  "n02671224": "access, memory access",
  "n02671421": "accessory, appurtenance, supplement, add-on",
  "n02671780": "accessory, accoutrement, accouterment",
  "n02671988": "access road, slip road",
  "n02672152": "accommodating lens implant, accommodating IOL",
  "n02672371": "accommodation",
  "n02672646": "accommodation ladder",
  "n02672831": "accordion, piano accordion, squeeze box",
  "n02673078": "accumulator, accumulator register",
  "n02673291": "ace",
  "n02673480": "acebutolol, Sectral",
  "n02673637": "ACE inhibitor, angiotensin-converting enzyme inhibitor",
  "n02674101": "ace of clubs",
  "n02674184": "ace of diamonds",
  "n02674273": "ace of hearts",
  "n02674358": "ace of spades",
  "n02674482": "acetaminophen, Datril, Tylenol, Panadol, Phenaphen, Tempra, Anacin III",
  "n02674912": "acetanilide, acetanilid, phenylacetamide",
  "n02675077": "acetate disk, phonograph recording disk",
  "n02675219": "acetate rayon, acetate",
  "n02675354": "acetophenetidin, acetphenetidin, phenacetin",
  "n02675522": "achromatic lens",
  "n02675657": "acid, back breaker, battery-acid, dose, dot, Elvis, loony toons, Lucy in the sky with diamonds, pane, superman, window pane, Zen",
  "n02675885": "acorn tube",
  "n02675987": "acoustic",
  "n02676097": "acoustic delay line, sonic delay line",
  "n02676261": "acoustic device",
  "n02676566": "acoustic guitar",
  "n02676670": "acoustic modem",
  "n02676805": "acoustic storage",
  "n02676938": "acropolis",
  "n02677028": "acrylic",
  "n02677136": "acrylic, acrylic paint",
  "n02677232": "Actifed",
  "n02677436": "actinometer",
  "n02677602": "actinomycin",
  "n02677718": "action, action mechanism",
  "n02678010": "active matrix screen",
  "n02678174": "active placebo",
  "n02678384": "actuator",
  "n02678528": "acyclovir, Zovirax",
  "n02678738": "Adam, ecstasy, XTC, go, disco biscuit, cristal, X, hug drug",
  "n02678897": "adapter, adaptor",
  "n02679142": "adder",
  "n02679257": "adding machine, totalizer, totaliser",
  "n02679415": "addition, add-on, improver",
  "n02679788": "additive",
  "n02679961": "addressing machine, Addressograph",
  "n02680110": "adhesive bandage",
  "n02680337": "adhesive tape",
  "n02680512": "adit",
  "n02680638": "adjoining room",
  "n02680754": "adjustable wrench, adjustable spanner",
  "n02680947": "adjuvant",
  "n02681084": "admixture, intermixture",
  "n02681392": "adobe, adobe brick",
  "n02681518": "adornment",
  "n02682038": "adrenergic, adrenergic drug",
  "n02682207": "adumbration",
  "n02682311": "adz, adze",
  "n02682407": "aeolian harp, aeolian lyre, wind harp",
  "n02682569": "aerator",
  "n02682699": "aerial ladder",
  "n02682811": "aerial torpedo",
  "n02682922": "aerosol, aerosol container, aerosol can, aerosol bomb, spray can",
  "n02683183": "Aertex",
  "n02683323": "afghan",
  "n02683454": "Afro-wig",
  "n02683558": "afterburner",
  "n02683692": "afterdeck",
  "n02683791": "after-shave, after-shave lotion",
  "n02683936": "afterthought",
  "n02684083": "agal",
  "n02684248": "agateware",
  "n02684356": "agglomerator",
  "n02684515": "aglet, aiglet, aiguilette",
  "n02684649": "aglet, aiglet",
  "n02684789": "agonist",
  "n02684962": "agora, public square",
  "n02685082": "aigrette, aigret",
  "n02685253": "aileron",
  "n02685365": "air bag",
  "n02685585": "air base, air station",
  "n02685701": "airbrake",
  "n02685855": "airbrake, dive brake",
  "n02685995": "airbrush",
  "n02686121": "airbus",
  "n02686227": "air compressor",
  "n02686379": "air conditioner, air conditioning",
  "n02686568": "aircraft",
  "n02687172": "aircraft carrier, carrier, flattop, attack aircraft carrier",
  "n02687423": "aircraft engine",
  "n02687539": "air cushion, inflatable cushion",
  "n02687682": "air cushion, air spring",
  "n02687821": "airdock, hangar, repair shed",
  "n02687992": "airfield, landing field, flying field, field",
  "n02688273": "air filter, air cleaner",
  "n02688443": "airfoil, aerofoil, control surface, surface",
  "n02688895": "Air Force Research Laboratory, AFRL",
  "n02689144": "airframe",
  "n02689274": "air gun, airgun, air rifle",
  "n02689434": "air hammer, jackhammer, pneumatic hammer",
  "n02689556": "air hole",
  "n02689648": "air horn",
  "n02689748": "air horn",
  "n02689819": "airing cupboard",
  "n02689961": "air-intake",
  "n02690081": "airline, airline business, airway",
  "n02690270": "airline, air hose",
  "n02690373": "airliner",
  "n02690583": "airlock, air lock",
  "n02690715": "airmailer",
  "n02690809": "air mattress",
  "n02690941": "air passage, air duct, airway",
  "n02691156": "airplane, aeroplane, plane",
  "n02692086": "airplane propeller, airscrew, prop",
  "n02692232": "airport, airdrome, aerodrome, drome",
  "n02692513": "air pump, vacuum pump",
  "n02692680": "air search radar",
  "n02692787": "air shaft, air well",
  "n02692877": "airship, dirigible",
  "n02693070": "airstrip, flight strip, landing strip, strip",
  "n02693246": "air terminal, airport terminal",
  "n02693413": "air-to-air missile",
  "n02693540": "air-to-ground missile, air-to-surface missile",
  "n02693709": "air transportation system",
  "n02693860": "aisle, gangway",
  "n02694045": "aisle",
  "n02694182": "aisle",
  "n02694279": "Aladdin's lamp",
  "n02694426": "alarm, warning device, alarm system",
  "n02694662": "alarm clock, alarm",
  "n02694776": "Alaskan pipeline, trans-Alaska pipeline",
  "n02694966": "alb",
  "n02695079": "album",
  "n02695321": "albuterol, Ventolin, Proventil",
  "n02695627": "alcazar",
  "n02695762": "alcohol thermometer, alcohol-in-glass thermometer",
  "n02696048": "alcove, bay",
  "n02696165": "alehouse",
  "n02696246": "alembic",
  "n02696384": "alendronate, Fosamax",
  "n02696569": "algometer",
  "n02696669": "Alhambra",
  "n02696843": "alidade, alidad",
  "n02697022": "alidade, alidad",
  "n02697221": "A-line",
  "n02697438": "alkylating agent",
  "n02697576": "Allen screw",
  "n02697675": "Allen wrench",
  "n02697759": "alley, alleyway, back street",
  "n02697876": "alligator wrench",
  "n02698036": "allopurinol, Zyloprim",
  "n02698244": "alms dish, alms tray",
  "n02698366": "aloes, bitter aloes",
  "n02698473": "alpaca",
  "n02698634": "alpenstock",
  "n02698769": "alpha blocker, alpha-blocker, alpha-adrenergic blocker, alpha-adrenergic blocking agent",
  "n02699096": "alpha-interferon",
  "n02699343": "alprazolam, Xanax",
  "n02699494": "altar",
  "n02699629": "altar, communion table, Lord's table",
  "n02699770": "altarpiece, reredos",
  "n02699915": "altazimuth",
  "n02700064": "alternator",
  "n02700258": "altimeter",
  "n02700422": "alto relievo, alto rilievo, high relief",
  "n02700622": "alum",
  "n02700767": "aluminum foil, aluminium foil, tin foil",
  "n02700895": "Amati",
  "n02701002": "ambulance",
  "n02701125": "ambulatory",
  "n02701260": "amen corner",
  "n02701393": "Americana",
  "n02701566": "American flag, Stars and Stripes, Star-Spangled Banner, Old Glory",
  "n02701730": "American organ",
  "n02701871": "American Stock Exchange, AMEX, Curb",
  "n02701984": "aminophylline",
  "n02702166": "aminopyrine, amidopyrine",
  "n02702304": "amiodarone, Cordarone",
  "n02702575": "amitriptyline, amitriptyline hydrochloride, Elavil",
  "n02702809": "amlodipine besylate, Norvasc",
  "n02702989": "ammeter",
  "n02703124": "ammonia clock",
  "n02703275": "ammunition, ammo",
  "n02703499": "amobarbital",
  "n02703692": "amobarbital sodium, blue, blue angel, blue devil, Amytal",
  "n02703894": "amoxicillin, Amoxil, Larotid, Polymox, Trimox, Augmentin",
  "n02704153": "amphetamine, pep pill, upper, speed",
  "n02704462": "amphetamine sulfate, amphetamine sulphate",
  "n02704645": "amphibian, amphibious aircraft",
  "n02704792": "amphibian, amphibious vehicle",
  "n02704949": "amphitheater, amphitheatre, coliseum",
  "n02705201": "amphitheater, amphitheatre",
  "n02705429": "amphora",
  "n02705558": "amphotericin",
  "n02705651": "ampicillin, Principen, Polycillin, SK-Ampicillin",
  "n02705944": "amplifier",
  "n02706221": "ampulla",
  "n02706373": "amrinone, Inocor",
  "n02706586": "amulet, talisman",
  "n02706806": "amusement arcade",
  "n02706916": "amyl nitrate",
  "n02707036": "anachronism",
  "n02707188": "anaglyph",
  "n02707344": "anaglyph",
  "n02707547": "analeptic",
  "n02707683": "analgesic, anodyne, painkiller, pain pill",
  "n02708093": "analog clock",
  "n02708224": "analog computer, analogue computer",
  "n02708433": "analog watch",
  "n02708555": "analytical balance, chemical balance",
  "n02708711": "analyzer, analyser",
  "n02708885": "anamorphosis, anamorphism",
  "n02709101": "anastigmat",
  "n02709265": "anastigmatic lens",
  "n02709367": "anchor, ground tackle",
  "n02709637": "anchor chain, anchor rope",
  "n02709763": "anchor light, riding light, riding lamp",
  "n02709908": "AND circuit, AND gate",
  "n02710044": "andiron, firedog, dog, dog-iron",
  "n02710201": "android, humanoid, mechanical man",
  "n02710324": "anechoic chamber",
  "n02710429": "anemometer, wind gauge, wind gage",
  "n02710600": "aneroid barometer, aneroid",
  "n02710766": "anesthetic, anaesthetic, anesthetic agent, anaesthetic agent",
  "n02711110": "anesthyl",
  "n02711237": "angiocardiogram",
  "n02711422": "angiogenesis inhibitor",
  "n02711573": "angiogram",
  "n02711780": "angioscope",
  "n02711890": "angiotensin, angiotonin, Hypertensin",
  "n02712134": "angiotensin I",
  "n02712280": "angiotensin II",
  "n02712393": "angiotensin II inhibitor",
  "n02712545": "angle bracket, angle iron",
  "n02712643": "angledozer",
  "n02712762": "Angostura Bridge",
  "n02712907": "animalization",
  "n02713003": "ankle brace",
  "n02713097": "anklet, ankle bracelet",
  "n02713218": "anklet, anklets, bobbysock, bobbysocks",
  "n02713364": "anklet",
  "n02713496": "ankus",
  "n02713594": "annex, annexe, extension, wing",
  "n02713769": "annulet",
  "n02713835": "annulet, bandelet, bandelette, bandlet, square and rabbet",
  "n02713992": "annulet, roundel",
  "n02714139": "annunciator",
  "n02714315": "anode",
  "n02714535": "anode",
  "n02714751": "answering machine",
  "n02714883": "antagonist",
  "n02715096": "antefix",
  "n02715229": "antenna, aerial, transmitting aerial",
  "n02715513": "anteroom, antechamber, entrance hall, hall, foyer, lobby, vestibule",
  "n02715712": "antiaircraft, antiaircraft gun, flak, flack, pom-pom, ack-ack, ack-ack gun",
  "n02715941": "antiarrhythmic, antiarrhythmic drug, antiarrhythmic medication",
  "n02716205": "antibacterial, antibacterial drug, bactericide",
  "n02716626": "antiballistic missile, ABM",
  "n02716866": "antibiotic, antibiotic drug",
  "n02717901": "anticholinergic, anticholinergic drug",
  "n02718084": "anticholinesterase",
  "n02718259": "anticoagulant, anticoagulant medication, decoagulant",
  "n02718469": "anticonvulsant, anticonvulsant drug, antiepileptic, antiepileptic drug",
  "n02718811": "antidepressant, antidepressant drug",
  "n02719105": "antidiabetic, antidiabetic drug",
  "n02719294": "antidiarrheal, antidiarrheal drug",
  "n02719450": "antidiuretic, antidiuretic drug",
  "n02719588": "antidote, counterpoison",
  "n02719750": "antiemetic, antiemetic drug",
  "n02719951": "antiflatulent",
  "n02720048": "antifouling paint",
  "n02720201": "antifungal, antifungal agent, fungicide, antimycotic, antimycotic agent",
  "n02720576": "anti-G suit, G suit",
  "n02720725": "antihistamine",
  "n02721160": "antihypertensive, antihypertensive drug",
  "n02721538": "anti-inflammatory, anti-inflammatory drug",
  "n02721813": "antimacassar",
  "n02721948": "antimalarial, antimalarial drug",
  "n02722166": "antimetabolite",
  "n02722349": "antimycin",
  "n02722458": "antineoplastic, antineoplastic drug, cancer drug",
  "n02722997": "antineoplastic antibiotic",
  "n02723165": "antiperspirant",
  "n02723292": "antiprotozoal, antiprotozoal drug",
  "n02723491": "antipruritic",
  "n02723595": "antipyretic, febrifuge",
  "n02723754": "antique",
  "n02724026": "antiquity",
  "n02724207": "antiseptic",
  "n02724533": "antispasmodic, spasmolytic, antispasmodic agent",
  "n02724722": "anti-submarine rocket",
  "n02724838": "antisyphilitic",
  "n02724966": "anti-TNF compound",
  "n02725242": "antitussive",
  "n02725367": "antiviral, antiviral agent, antiviral drug",
  "n02725596": "Antonine Wall",
  "n02725872": "anvil",
  "n02726017": "ao dai",
  "n02726210": "apadana",
  "n02726305": "apartment, flat",
  "n02726681": "apartment building, apartment house",
  "n02726861": "APC",
  "n02727016": "aperture",
  "n02727141": "aperture",
  "n02727281": "aphrodisiac",
  "n02727426": "apiary, bee house",
  "n02727566": "apishamore",
  "n02727660": "apomorphine",
  "n02727825": "apparatus, setup",
  "n02728440": "apparel, wearing apparel, dress, clothes",
  "n02728763": "appendage",
  "n02728968": "appendicle",
  "n02729060": "Appian Way",
  "n02729222": "applecart",
  "n02729345": "apple of discord",
  "n02729755": "apple orchard",
  "n02729837": "appliance",
  "n02729965": "appliance, contraption, contrivance, convenience, gadget, gizmo, gismo, widget",
  "n02730265": "applicator, applier",
  "n02730437": "applique",
  "n02730568": "appointment, fitting",
  "n02730748": "approach trench, communication trench",
  "n02730930": "apron",
  "n02731123": "apron",
  "n02731251": "apron string",
  "n02731398": "apse, apsis",
  "n02731629": "aqualung, Aqua-Lung, scuba",
  "n02731900": "aquaplane",
  "n02732072": "aquarium, fish tank, marine museum",
  "n02732295": "aquatint",
  "n02732433": "aqueduct",
  "n02732572": "arabesque",
  "n02732696": "araroba, Goa powder, chrysarobin",
  "n02732827": "arbor, arbour, bower, pergola",
  "n02733075": "arboretum, botanical garden",
  "n02733213": "arcade, colonnade",
  "n02733378": "arcade",
  "n02733524": "arch",
  "n02734217": "arch, archway",
  "n02734423": "architectural ornament",
  "n02734725": "architecture",
  "n02734835": "architrave",
  "n02734994": "architrave",
  "n02735086": "archive",
  "n02735268": "arch support",
  "n02735361": "arc lamp, arc light",
  "n02735538": "arctic, galosh, golosh, rubber, gumshoe",
  "n02735688": "area",
  "n02736396": "areaway",
  "n02736511": "arena, scene of action",
  "n02736645": "arena theater, theater in the round",
  "n02736798": "argyle, argyll",
  "n02736934": "argyle, argyll",
  "n02737166": "argyll, argyle",
  "n02737351": "ark",
  "n02737467": "Ark, Ark of the Covenant",
  "n02737660": "arm",
  "n02737833": "arm, branch, limb",
  "n02738031": "armament",
  "n02738271": "armature",
  "n02738449": "armband",
  "n02738535": "armchair",
  "n02738741": "armet",
  "n02738859": "arm guard, arm pad",
  "n02738978": "armhole",
  "n02739123": "armilla",
  "n02739248": "armillary sphere, armilla",
  "n02739427": "armlet, arm band",
  "n02739550": "armoire",
  "n02739668": "armor, armour",
  "n02739889": "armored car, armoured car",
  "n02740061": "armored car, armoured car",
  "n02740300": "armored personnel carrier, armoured personnel carrier, APC",
  "n02740533": "armored vehicle, armoured vehicle",
  "n02740764": "armor plate, armour plate, armor plating, plate armor, plate armour",
  "n02741367": "armory, armoury, arsenal",
  "n02741475": "armrest",
  "n02741587": "army base",
  "n02741681": "Army High Performance Computing Research Center, AHPCRC",
  "n02741981": "arnica",
  "n02742070": "arquebus, harquebus, hackbut, hagbut",
  "n02742194": "array",
  "n02742322": "array, raiment, regalia",
  "n02742468": "arrester, arrester hook",
  "n02742663": "arrival gate",
  "n02742753": "arrow",
  "n02743050": "arrowhead",
  "n02743207": "arsenal, armory, armoury",
  "n02743426": "arsenal, armory, armoury",
  "n02743547": "art, fine art",
  "n02744077": "Artemision at Ephesus",
  "n02744323": "arterial road",
  "n02744423": "arteriogram",
  "n02744532": "artery",
  "n02744634": "artesian well",
  "n02744844": "arthrogram",
  "n02744961": "arthroscope",
  "n02745091": "article of commerce",
  "n02745228": "articulated ladder",
  "n02745395": "artificial flower",
  "n02745492": "artificial heart",
  "n02745611": "artificial horizon, gyro horizon, flight indicator",
  "n02745816": "artificial joint",
  "n02746008": "artificial kidney, hemodialyzer",
  "n02746225": "artificial skin",
  "n02746365": "artillery, heavy weapon, gun, ordnance",
  "n02746595": "artillery shell",
  "n02746683": "artist's loft",
  "n02746841": "artist's workroom, atelier",
  "n02746978": "art school",
  "n02747063": "ascot",
  "n02747177": "ashcan, trash can, garbage can, wastebin, ash bin, ash-bin, ashbin, dustbin, trash barrel, trash bin",
  "n02747379": "Ash Can, Ashcan school",
  "n02747541": "ashlar",
  "n02747672": "ash-pan",
  "n02747802": "ashtray",
  "n02747915": "asparaginase, Elspar",
  "n02748088": "asparagus bed",
  "n02748183": "aspergill, aspersorium",
  "n02748359": "aspersorium",
  "n02748491": "aspirator",
  "n02748618": "aspirin, acetylsalicylic acid, Bayer, Empirin, St. Joseph",
  "n02749169": "aspirin powder, headache powder",
  "n02749292": "assault gun",
  "n02749479": "assault rifle, assault gun",
  "n02749670": "assegai, assagai",
  "n02749790": "assembly",
  "n02749953": "assembly",
  "n02750070": "assembly hall",
  "n02750169": "assembly plant",
  "n02750320": "astatic coils",
  "n02750652": "astatic galvanometer",
  "n02750835": "astringent, astringent drug, styptic",
  "n02751067": "astrodome",
  "n02751215": "astrolabe",
  "n02751295": "astronomical telescope",
  "n02751490": "astronomy satellite",
  "n02751623": "Aswan High Dam, High Dam",
  "n02751782": "atenolol, Tenormin",
  "n02752039": "athanor",
  "n02752199": "athenaeum, atheneum",
  "n02752311": "athletic facility",
  "n02752496": "athletic sock, sweat sock, varsity sock",
  "n02752615": "athletic supporter, supporter, suspensor, jockstrap, jock",
  "n02752810": "atlas, telamon",
  "n02752917": "atmometer, evaporometer",
  "n02753044": "atom bomb, atomic bomb, A-bomb, fission bomb, plutonium bomb",
  "n02753394": "atomic clock",
  "n02753569": "atomic cocktail",
  "n02753710": "atomic pile, atomic reactor, pile, chain reactor",
  "n02753881": "atomic warhead, nuclear warhead, thermonuclear warhead, nuke",
  "n02754103": "atomizer, atomiser, spray, sprayer, nebulizer, nebuliser",
  "n02754421": "atorvastatin, Lipitor",
  "n02754656": "atrium",
  "n02754756": "atropine",
  "n02755140": "attache case, attache",
  "n02755244": "attachment",
  "n02755352": "attachment, bond",
  "n02755529": "attack submarine",
  "n02755675": "attenuator",
  "n02755823": "attic",
  "n02755984": "attic fan",
  "n02756098": "attire, garb, dress",
  "n02756601": "auction block, block",
  "n02756751": "audio",
  "n02756854": "audio amplifier",
  "n02756977": "audiocassette",
  "n02757061": "audio CD, audio compact disc",
  "n02757211": "audiogram",
  "n02757337": "audiometer, sonometer",
  "n02757462": "audio system, sound system",
  "n02757714": "audiotape",
  "n02757810": "audiotape",
  "n02757927": "audiovisual, audiovisual aid",
  "n02758134": "auditorium",
  "n02758270": "Augean stables",
  "n02758490": "auger, gimlet, screw auger, wimble",
  "n02758600": "Auschwitz",
  "n02758753": "auto accessory",
  "n02758863": "autobahn",
  "n02758960": "autoclave, sterilizer, steriliser",
  "n02759257": "autofocus",
  "n02759387": "autogiro, autogyro, gyroplane",
  "n02759615": "autograph album",
  "n02759700": "autoinjector",
  "n02759963": "autoloader, self-loader",
  "n02760099": "automat",
  "n02760199": "automat",
  "n02760298": "automatic choke",
  "n02760429": "automatic firearm, automatic gun, automatic weapon",
  "n02760658": "automatic pistol, automatic",
  "n02760855": "automatic rifle, automatic, machine rifle",
  "n02761034": "automatic transmission, automatic drive",
  "n02761206": "automation",
  "n02761392": "automaton, robot, golem",
  "n02761557": "automobile engine",
  "n02761696": "automobile factory, auto factory, car factory",
  "n02761834": "automobile horn, car horn, motor horn, horn, hooter",
  "n02762044": "auto part, car part",
  "n02762169": "autopilot, automatic pilot, robot pilot",
  "n02762371": "autoradiograph",
  "n02762508": "autostrada",
  "n02762609": "auxiliary airfield",
  "n02762725": "auxiliary boiler, donkey boiler",
  "n02762909": "auxiliary engine, donkey engine",
  "n02763083": "auxiliary pump, donkey pump",
  "n02763198": "auxiliary research submarine",
  "n02763306": "auxiliary storage, external storage, secondary storage",
  "n02763472": "avenue, boulevard",
  "n02763604": "aviary, bird sanctuary, volary",
  "n02763714": "awl",
  "n02763901": "awning, sunshade, sunblind",
  "n02764044": "ax, axe",
  "n02764398": "ax handle, axe handle",
  "n02764505": "ax head, axe head",
  "n02764614": "axis, axis of rotation",
  "n02764779": "axle",
  "n02764935": "axle bar",
  "n02765028": "axletree",
  "n02765247": "azathioprine, Imuran",
  "n02765429": "zidovudine, Retrovir, ZDV, AZT",
  "n02765656": "azithromycin, Zithromax",
  "n02765868": "aztreonam, Azactam",
  "n02766044": "B-52",
  "n02766168": "babushka",
  "n02766320": "baby bed, baby's bed",
  "n02766534": "baby buggy, baby carriage, carriage, perambulator, pram, stroller, go-cart, pushchair, pusher",
  "n02766792": "baby grand, baby grand piano, parlor grand, parlor grand piano, parlour grand, parlour grand piano",
  "n02766961": "baby oil",
  "n02767038": "baby powder",
  "n02767147": "baby shoe",
  "n02767240": "bacitracin",
  "n02767433": "back, backrest",
  "n02767665": "back",
  "n02767815": "backband",
  "n02767956": "backbench",
  "n02768114": "backboard",
  "n02768226": "backboard, basketball backboard",
  "n02768433": "backbone",
  "n02768655": "back brace",
  "n02768744": "back door, backdoor, back entrance",
  "n02768864": "backdrop, background, backcloth",
  "n02768973": "backgammon board",
  "n02769075": "background, desktop, screen background",
  "n02769290": "backhoe",
  "n02769460": "backing, mount",
  "n02769669": "backlighting",
  "n02769748": "backpack, back pack, knapsack, packsack, rucksack, haversack",
  "n02769963": "backpacking tent, pack tent",
  "n02770078": "backplate",
  "n02770211": "back porch",
  "n02770293": "back room",
  "n02770441": "backroom",
  "n02770585": "backsaw, back saw",
  "n02770721": "backscratcher",
  "n02770830": "backseat",
  "n02771004": "backspace key, backspace, backspacer",
  "n02771166": "backstairs",
  "n02771286": "backstay",
  "n02771383": "backstitch",
  "n02771547": "backstop",
  "n02771750": "backsword",
  "n02771840": "backup, computer backup",
  "n02772101": "backup system",
  "n02772197": "backyard",
  "n02772282": "bacteria bed",
  "n02772435": "badminton court",
  "n02772554": "badminton equipment",
  "n02772700": "badminton racket, badminton racquet, battledore",
  "n02772868": "baffle, baffle board",
  "n02773037": "bag",
  "n02773838": "bag, traveling bag, travelling bag, grip, suitcase",
  "n02774152": "bag, handbag, pocketbook, purse",
  "n02774502": "bagatelle, fluff, frippery, frivolity",
  "n02774630": "baggage, luggage",
  "n02774921": "baggage",
  "n02775039": "baggage car, luggage van",
  "n02775178": "baggage claim",
  "n02775359": "bagger",
  "n02775483": "bagpipe",
  "n02775689": "bailey",
  "n02775813": "bailey",
  "n02775897": "Bailey bridge",
  "n02776007": "bain-marie",
  "n02776205": "bait, decoy, lure",
  "n02776505": "baize",
  "n02776631": "bakery, bakeshop, bakehouse",
  "n02776825": "balaclava, balaclava helmet",
  "n02776978": "balalaika",
  "n02777100": "balance",
  "n02777292": "balance beam, beam",
  "n02777402": "balance wheel, balance",
  "n02777638": "balbriggan",
  "n02777734": "balcony",
  "n02777927": "balcony",
  "n02778131": "baldachin",
  "n02778294": "baldric, baldrick",
  "n02778456": "bale",
  "n02778588": "baling wire",
  "n02778669": "ball",
  "n02779435": "ball",
  "n02779609": "ball and chain",
  "n02779719": "ball-and-socket joint",
  "n02779845": "ballast",
  "n02779971": "ballast, light ballast",
  "n02780121": "ballast resistor, ballast, barretter",
  "n02780315": "ball bearing, needle bearing, roller bearing",
  "n02780445": "ball cartridge",
  "n02780588": "ballcock, ball cock",
  "n02780704": "balldress",
  "n02780815": "ballet skirt, tutu",
  "n02780916": "ball field, baseball field, diamond",
  "n02781121": "ball gown",
  "n02781213": "ballistic galvanometer",
  "n02781338": "ballistic missile",
  "n02781517": "ballistic pendulum",
  "n02781764": "ballistocardiograph, cardiograph",
  "n02782093": "balloon",
  "n02782329": "balloon",
  "n02782432": "balloon bomb, Fugo",
  "n02782602": "balloon sail",
  "n02782681": "ballot box",
  "n02782778": "ballpark, park",
  "n02783035": "ball-peen hammer",
  "n02783161": "ballpoint, ballpoint pen, ballpen, Biro",
  "n02783324": "ballroom, dance hall, dance palace",
  "n02783459": "ball valve",
  "n02783567": "Balmoral, bluebonnet",
  "n02783708": "balmoral",
  "n02783790": "balsam",
  "n02783900": "balsa raft, Kon Tiki",
  "n02783994": "baluster",
  "n02784124": "banana boat",
  "n02784218": "band",
  "n02784732": "band, banding, stripe",
  "n02784998": "band",
  "n02785191": "band, ring",
  "n02785365": "band",
  "n02785570": "band",
  "n02785648": "bandage, patch",
  "n02786058": "Band Aid",
  "n02786198": "bandanna, bandana",
  "n02786331": "bandbox",
  "n02786463": "banderilla",
  "n02786611": "bandoleer, bandolier",
  "n02786736": "bandoneon",
  "n02786837": "bandsaw, band saw",
  "n02786984": "bandstand, outdoor stage, stand",
  "n02787120": "bandwagon",
  "n02787269": "bangalore torpedo",
  "n02787435": "bangle, bauble, gaud, gewgaw, novelty, fallal, trinket",
  "n02787622": "banjo",
  "n02787772": "bank, bank building",
  "n02788021": "banner, streamer",
  "n02788148": "bannister, banister, balustrade, balusters, handrail",
  "n02788386": "banquette",
  "n02788462": "banyan, banian",
  "n02788572": "baptismal font, baptistry, baptistery, font",
  "n02788689": "bar",
  "n02789271": "bar",
  "n02789487": "bar",
  "n02789770": "bar",
  "n02790012": "bar",
  "n02790154": "bar",
  "n02790322": "barb",
  "n02790550": "barb",
  "n02790669": "barbecue, barbeque",
  "n02790823": "barbed wire, barbwire",
  "n02790996": "barbell",
  "n02791124": "barber chair",
  "n02791270": "barbershop",
  "n02791385": "barbette",
  "n02791532": "barbette carriage",
  "n02791665": "barbican, barbacan",
  "n02791795": "bar bit",
  "n02791894": "barbital, veronal, barbitone, diethylbarbituric acid, diethylmalonylurea",
  "n02792049": "barbiturate",
  "n02792305": "bard",
  "n02792409": "bareboat",
  "n02792552": "barge, flatboat, hoy, lighter",
  "n02792822": "bargello, flame stitch",
  "n02792948": "barge pole",
  "n02793089": "baritone, baritone horn",
  "n02793199": "bark, barque",
  "n02793296": "bar magnet",
  "n02793414": "bar mask",
  "n02793495": "barn",
  "n02793684": "barndoor",
  "n02793842": "barn door",
  "n02793930": "barnyard",
  "n02794008": "barograph",
  "n02794156": "barometer",
  "n02794368": "barong",
  "n02794474": "barouche",
  "n02794664": "bar printer",
  "n02794779": "barrack",
  "n02794972": "barrage balloon",
  "n02795169": "barrel, cask",
  "n02795528": "barrel, gun barrel",
  "n02795670": "barrelhouse, honky-tonk",
  "n02795783": "barrel knot, blood knot",
  "n02795978": "barrel organ, grind organ, hand organ, hurdy gurdy, hurdy-gurdy, street organ",
  "n02796207": "barrel vault",
  "n02796318": "barrette",
  "n02796412": "barricade",
  "n02796623": "barrier",
  "n02796995": "barroom, bar, saloon, ginmill, taproom",
  "n02797295": "barrow, garden cart, lawn cart, wheelbarrow",
  "n02797455": "bar soap",
  "n02797535": "bascule",
  "n02797692": "base, pedestal, stand",
  "n02797881": "base, bag",
  "n02798117": "base",
  "n02798290": "base, base of operations",
  "n02798574": "base",
  "n02798769": "base",
  "n02799071": "baseball",
  "n02799175": "baseball bat, lumber",
  "n02799323": "baseball cap, jockey cap, golf cap",
  "n02799442": "baseball card",
  "n02799593": "baseball diamond, diamond, infield",
  "n02799897": "baseball equipment",
  "n02800213": "baseball glove, glove, baseball mitt, mitt",
  "n02800354": "baseboard, mopboard, skirting board",
  "n02800497": "basement, cellar",
  "n02800675": "basement",
  "n02800793": "basic, staple",
  "n02800940": "basic point defense missile system",
  "n02801047": "basilica, Roman basilica",
  "n02801184": "basilica",
  "n02801450": "basilisk",
  "n02801525": "basin",
  "n02801823": "basinet",
  "n02801938": "basket, handbasket",
  "n02802215": "basket, basketball hoop, hoop",
  "n02802426": "basketball",
  "n02802544": "basketball court",
  "n02802721": "basketball equipment",
  "n02802886": "basket hilt",
  "n02802990": "basket weave",
  "n02803129": "bas relief, low relief, basso relievo, basso rilievo",
  "n02803349": "bass",
  "n02803539": "bass clarinet",
  "n02803666": "bass drum, gran casa",
  "n02803809": "basset horn",
  "n02803934": "bass fiddle, bass viol, bull fiddle, double bass, contrabass, string bass",
  "n02804123": "bass guitar",
  "n02804252": "bass horn, sousaphone, tuba",
  "n02804414": "bassinet",
  "n02804515": "bassinet",
  "n02804610": "bassoon",
  "n02804772": "bastard, mongrel",
  "n02805111": "baste, basting, basting stitch, tacking",
  "n02805283": "baster",
  "n02805443": "bastille",
  "n02805584": "Bastille",
  "n02805845": "bastinado",
  "n02805983": "bastion",
  "n02806088": "bastion, citadel",
  "n02806274": "basuco",
  "n02806379": "bat",
  "n02806530": "bath",
  "n02806762": "bath chair",
  "n02806875": "bathhouse, bagnio",
  "n02806992": "bathhouse, bathing machine",
  "n02807133": "bathing cap, swimming cap",
  "n02807260": "bath linen",
  "n02807401": "bath mat",
  "n02807523": "bath oil",
  "n02807616": "bathrobe",
  "n02807731": "bathroom, bath",
  "n02807998": "bathroom cleaner",
  "n02808097": "bathroom fixture",
  "n02808185": "bath salts",
  "n02808304": "bath towel",
  "n02808440": "bathtub, bathing tub, bath, tub",
  "n02808695": "bathymeter, bathometer",
  "n02808829": "bathyscaphe, bathyscaph, bathyscape",
  "n02808968": "bathysphere",
  "n02809105": "batik",
  "n02809241": "batiste",
  "n02809364": "baton, wand",
  "n02809491": "baton",
  "n02809605": "baton",
  "n02809736": "baton",
  "n02809866": "Baton Rouge Bridge",
  "n02810008": "batten",
  "n02810139": "battering ram",
  "n02810270": "batter's box",
  "n02810471": "battery, electric battery",
  "n02810782": "battery, stamp battery",
  "n02810930": "batting, batten",
  "n02811059": "batting cage, cage",
  "n02811204": "batting glove",
  "n02811350": "batting helmet",
  "n02811468": "battle-ax, battle-axe",
  "n02811618": "battle cruiser",
  "n02811719": "battle dress",
  "n02811842": "battle flag",
  "n02811936": "battlement, crenelation, crenellation",
  "n02812201": "battleship, battlewagon",
  "n02812342": "battle sight, battlesight",
  "n02812505": "batwing",
  "n02812631": "bay",
  "n02812785": "bay",
  "n02812949": "bayonet",
  "n02813089": "Bayonne Bridge",
  "n02813252": "bay rum",
  "n02813399": "bay window, bow window",
  "n02813544": "bazaar, bazar",
  "n02813645": "bazaar, bazar",
  "n02813752": "bazooka",
  "n02813874": "BB, BB shot",
  "n02813981": "B battery",
  "n02814116": "BB gun",
  "n02814224": "beach ball, beachball",
  "n02814338": "beach house",
  "n02814428": "beach towel",
  "n02814533": "beach wagon, station wagon, wagon, estate car, beach waggon, station waggon, waggon",
  "n02814774": "beachwear",
  "n02814860": "beacon, lighthouse, beacon light, pharos",
  "n02815071": "bead",
  "n02815237": "beading, bead, beadwork, astragal",
  "n02815389": "beading, beadwork",
  "n02815478": "beading plane",
  "n02815600": "beads, string of beads",
  "n02815749": "beaker",
  "n02815834": "beaker",
  "n02815950": "beam",
  "n02816381": "beam",
  "n02816494": "beam balance",
  "n02816656": "beanbag",
  "n02816768": "beanie, beany",
  "n02816895": "bear claw",
  "n02817031": "bearing",
  "n02817251": "bearing rein, checkrein",
  "n02817386": "bearing wall",
  "n02817516": "bearskin, busby, shako",
  "n02817650": "beater",
  "n02817799": "beating-reed instrument, reed instrument, reed",
  "n02818027": "beauty spot",
  "n02818135": "beaver, castor",
  "n02818254": "beaver",
  "n02818402": "beaver board",
  "n02818507": "becket",
  "n02818687": "Beckman thermometer",
  "n02818832": "bed",
  "n02819474": "bed",
  "n02819697": "bed",
  "n02819881": "bed",
  "n02820085": "bed and breakfast, bed-and-breakfast",
  "n02820210": "bedclothes, bed clothing, bedding",
  "n02820432": "bedding material, bedding, litter",
  "n02820556": "Bedford cord",
  "n02820675": "bed jacket",
  "n02820798": "Bedlam, booby hatch, crazy house, cuckoo's nest, funny farm, funny house, loony bin, madhouse, nut house, nuthouse, sanatorium, snake pit",
  "n02821030": "bed linen",
  "n02821202": "bedpan",
  "n02821328": "bed pillow",
  "n02821415": "bedpost",
  "n02821543": "bedroll",
  "n02821627": "bedroom, sleeping room, sleeping accommodation, chamber, bedchamber",
  "n02821943": "bedroom furniture",
  "n02822064": "bedsitting room, bedsitter, bedsit",
  "n02822220": "bedspread, bedcover, bed cover, bed covering, counterpane, spread",
  "n02822399": "bedspring",
  "n02822579": "bedstead, bedframe",
  "n02822762": "beefcake",
  "n02822865": "beehive, hive",
  "n02823030": "beehive",
  "n02823124": "beeper, pager",
  "n02823335": "beer barrel, beer keg",
  "n02823428": "beer bottle",
  "n02823510": "beer can",
  "n02823586": "beer garden",
  "n02823750": "beer glass",
  "n02823848": "beer hall",
  "n02823964": "beer mat",
  "n02824058": "beer mug, stein",
  "n02824152": "belaying pin",
  "n02824319": "belfry",
  "n02824448": "bell",
  "n02824858": "bell",
  "n02825004": "belladonna",
  "n02825153": "bell arch",
  "n02825240": "bellarmine, longbeard, long-beard, greybeard",
  "n02825442": "bellbottom trousers, bell-bottoms, bellbottom pants",
  "n02825657": "bell cote, bell cot",
  "n02825775": "bell deck",
  "n02825872": "bell foundry",
  "n02825961": "bell gable",
  "n02826068": "bell jar, bell glass",
  "n02826259": "bellows",
  "n02826459": "bellpull",
  "n02826589": "bell push",
  "n02826683": "bell seat, balloon seat",
  "n02826812": "bell tent",
  "n02826886": "bell tower",
  "n02827020": "bellyband",
  "n02827148": "bellyband",
  "n02827289": "Belmont Park, Belmont",
  "n02827466": "Belsen",
  "n02827606": "belt",
  "n02827883": "belt",
  "n02828115": "belt, belt ammunition, belted ammunition",
  "n02828299": "belt buckle",
  "n02828427": "belting",
  "n02828555": "belvedere",
  "n02828648": "beltway, bypass, ring road, ringway",
  "n02828884": "bench",
  "n02829116": "bench",
  "n02829246": "bench clamp",
  "n02829353": "bench hook",
  "n02829510": "bench lathe",
  "n02829596": "bench press",
  "n02829696": "bend, curve",
  "n02830013": "bend, bend dexter",
  "n02830157": "bender",
  "n02830294": "Benjamin Franklin Bridge",
  "n02830430": "bentwood",
  "n02830596": "Benzedrine, bennie",
  "n02830721": "benzocaine, ethyl aminobenzoate",
  "n02830852": "benzodiazepine",
  "n02831237": "beret",
  "n02831335": "berlin",
  "n02831455": "Bermuda rig, Bermudan rig, Bermudian rig, Marconi rig",
  "n02831595": "Bermuda shorts, Jamaica shorts",
  "n02831724": "berth, bunk, built in bed",
  "n02831894": "besom",
  "n02831998": "Bessemer converter",
  "n02832168": "beta blocker, beta-blocking agent, beta-adrenergic blocker, beta-adrenergic blocking agent",
  "n02832652": "beta-interferon",
  "n02832838": "betatron, induction accelerator",
  "n02833040": "bethel",
  "n02833140": "betting shop",
  "n02833275": "bevatron",
  "n02833403": "bevel, bevel square",
  "n02833576": "bevel, cant, chamfer",
  "n02833793": "bevel gear, pinion and crown wheel, pinion and ring gear",
  "n02833943": "bezel",
  "n02834027": "B-flat clarinet, licorice stick",
  "n02834147": "bhang",
  "n02834295": "bib",
  "n02834397": "bib",
  "n02834506": "bib-and-tucker",
  "n02834642": "bicorn, bicorne",
  "n02834778": "bicycle, bike, wheel, cycle",
  "n02835271": "bicycle-built-for-two, tandem bicycle, tandem",
  "n02835412": "bicycle chain",
  "n02835551": "bicycle clip, trouser clip",
  "n02835724": "bicycle pump",
  "n02835829": "bicycle rack",
  "n02835915": "bicycle seat, saddle",
  "n02836035": "bicycle wheel",
  "n02836174": "bidet",
  "n02836268": "bier",
  "n02836392": "bier",
  "n02836513": "bi-fold door",
  "n02836607": "bifocals",
  "n02836766": "Big Ben",
  "n02836900": "Big Blue, BLU-82",
  "n02837134": "big board",
  "n02837309": "biggin",
  "n02837416": "big H, hell dust, nose drops, smack, thunder, skag, scag",
  "n02837567": "bight",
  "n02837702": "bijou",
  "n02837789": "bikini, two-piece",
  "n02837887": "bikini pants",
  "n02838014": "bilge",
  "n02838178": "bilge keel",
  "n02838345": "bilge pump",
  "n02838448": "bilges",
  "n02838577": "bilge well",
  "n02838728": "bill, peak, eyeshade, visor, vizor",
  "n02838958": "bill, billhook",
  "n02839110": "billboard, hoarding",
  "n02839200": "billet",
  "n02839351": "billiard ball",
  "n02839477": "billiard marker",
  "n02839592": "billiard room, billiard saloon, billiard parlor, billiard parlour, billiard hall",
  "n02839758": "bimetallic strip",
  "n02839910": "bin",
  "n02840134": "binder, ligature",
  "n02840245": "binder, ring-binder",
  "n02840361": "binder, reaper binder",
  "n02840515": "bindery",
  "n02840619": "binding, book binding, cover, back",
  "n02840927": "binding",
  "n02841063": "bin liner",
  "n02841187": "binnacle",
  "n02841315": "binoculars, field glasses, opera glasses",
  "n02841506": "binocular microscope",
  "n02841641": "biochip",
  "n02841847": "biohazard suit",
  "n02842008": "biology lab, biology laboratory, bio lab",
  "n02842133": "bioscope",
  "n02842219": "bioscope",
  "n02842303": "bioweapon, biological weapon, bioarm",
  "n02842573": "biplane",
  "n02842684": "biprism",
  "n02842809": "birch, birch rod",
  "n02843029": "birchbark canoe, birchbark, birch bark",
  "n02843158": "birdbath",
  "n02843276": "birdcage",
  "n02843365": "birdcage mask",
  "n02843465": "birdcall",
  "n02843553": "bird feeder, birdfeeder, feeder",
  "n02843684": "birdhouse",
  "n02843777": "bird shot, buckshot, duck shot",
  "n02843909": "biretta, berretta, birretta",
  "n02844056": "bishop",
  "n02844214": "bistro",
  "n02844307": "bit",
  "n02844714": "bit",
  "n02845002": "bit",
  "n02845130": "bite plate, biteplate",
  "n02845293": "bitewing",
  "n02845425": "bitmap, electronic image",
  "n02845576": "bitter end",
  "n02845763": "bitthead",
  "n02845860": "bitt pin",
  "n02845985": "bitumastic",
  "n02846141": "black",
  "n02846260": "black",
  "n02846399": "black and white, monochrome",
  "n02846511": "blackboard, chalkboard",
  "n02846619": "blackboard eraser",
  "n02846733": "black box",
  "n02846874": "blackface",
  "n02847009": "black flag, pirate flag, Jolly Roger, blackjack",
  "n02847223": "Black Hole of Calcutta",
  "n02847461": "blackjack, cosh, sap",
  "n02847631": "black tie",
  "n02847728": "Blackwall hitch",
  "n02847852": "blackwash",
  "n02847942": "blackwash, black lotion",
  "n02848118": "bladder",
  "n02848216": "blade",
  "n02848523": "blade, vane",
  "n02848806": "blade",
  "n02848921": "blank, dummy, blank shell",
  "n02849051": "blank",
  "n02849154": "blanket, cover",
  "n02849419": "blanket",
  "n02849565": "blanket stitch",
  "n02849729": "Blarney Stone",
  "n02849885": "blast furnace",
  "n02850060": "blasting cap",
  "n02850218": "blasting gelatin",
  "n02850358": "blazer, sport jacket, sport coat, sports jacket, sports coat",
  "n02850552": "bleachers",
  "n02850732": "blender, liquidizer, liquidiser",
  "n02850950": "blimp, sausage balloon, sausage",
  "n02851099": "blind, screen",
  "n02851384": "blind",
  "n02851540": "blind alley, cul de sac, dead-end street, impasse",
  "n02851675": "blind corner",
  "n02851795": "blind curve, blind bend",
  "n02851939": "blindfold",
  "n02852043": "bling, bling bling",
  "n02852173": "blinker, flasher",
  "n02852360": "blister pack, bubble pack",
  "n02852523": "block",
  "n02853016": "block",
  "n02853218": "blockade",
  "n02853336": "blockade-runner",
  "n02853449": "blockage, block, closure, occlusion, stop, stoppage",
  "n02853745": "block and tackle",
  "n02853870": "blockbuster",
  "n02853991": "block diagram",
  "n02854156": "blocker, blocking agent",
  "n02854378": "blockhouse",
  "n02854532": "block plane",
  "n02854630": "bloodmobile",
  "n02854739": "bloomers, pants, drawers, knickers",
  "n02854926": "blouse",
  "n02855089": "blower",
  "n02855253": "blowgun, blowpipe, blowtube, blow tube",
  "n02855390": "blowtorch, torch, blowlamp",
  "n02855560": "blowtube, blow tube, blowpipe",
  "n02855701": "blucher",
  "n02855793": "bludgeon",
  "n02855925": "blue",
  "n02856013": "blue chip",
  "n02856109": "blueprint",
  "n02856237": "blunderbuss",
  "n02856362": "blunt file",
  "n02856463": "board",
  "n02857023": "board, gameboard",
  "n02857365": "boarding",
  "n02857477": "boarding house, boardinghouse",
  "n02857644": "boardroom, council chamber",
  "n02857812": "board rule",
  "n02857907": "boards",
  "n02858044": "boards",
  "n02858179": "boardwalk",
  "n02858304": "boat",
  "n02859084": "boat deck",
  "n02859184": "boater, leghorn, Panama, Panama hat, sailor, skimmer, straw hat",
  "n02859343": "boat hook",
  "n02859443": "boathouse",
  "n02859557": "boatswain's chair, bosun's chair",
  "n02859729": "boat train",
  "n02859829": "boat whistle",
  "n02859955": "boatyard",
  "n02860063": "bob, bobber, cork, bobfloat",
  "n02860239": "bob",
  "n02860415": "bobbin, spool, reel",
  "n02860640": "bobby pin, hairgrip, grip",
  "n02860847": "bobsled, bobsleigh, bob",
  "n02861022": "bobsled, bobsleigh",
  "n02861147": "bocce ball, bocci ball, boccie ball",
  "n02861286": "bodega",
  "n02861387": "bodice",
  "n02861509": "bodkin, threader",
  "n02861658": "bodkin",
  "n02861777": "bodkin",
  "n02861886": "body",
  "n02862048": "body armor, body armour, suit of armor, suit of armour, coat of mail, cataphract",
  "n02862770": "body bag, personnel pouch, human remains pouch",
  "n02862916": "body lotion",
  "n02863014": "body stocking",
  "n02863176": "body plethysmograph",
  "n02863340": "body pad",
  "n02863426": "bodywork",
  "n02863536": "Bofors gun",
  "n02863638": "bogy, bogie, bogey",
  "n02863750": "boiler, steam boiler",
  "n02863995": "boilerplate",
  "n02864122": "boiling water reactor, BWR",
  "n02864342": "bola",
  "n02864504": "bolero",
  "n02864593": "bollard, bitt",
  "n02864878": "bollock, bullock block",
  "n02864987": "bolo, bolo knife",
  "n02865108": "bologram, bolograph",
  "n02865214": "bolometer",
  "n02865351": "bolo tie, bolo, bola tie, bola",
  "n02865509": "bolster, long pillow",
  "n02865665": "bolt",
  "n02865931": "bolt, deadbolt",
  "n02866106": "bolt",
  "n02866286": "bolt",
  "n02866386": "bolt cutter",
  "n02866474": "bolus",
  "n02866578": "bomb",
  "n02867267": "bombardon, bombard",
  "n02867401": "bombazine",
  "n02867592": "bomb calorimeter, bomb",
  "n02867715": "bomber",
  "n02867966": "bomber jacket",
  "n02868068": "bombie",
  "n02868240": "bomblet, cluster bomblet",
  "n02868429": "bomb rack",
  "n02868546": "bombshell",
  "n02868638": "bomb shelter, air-raid shelter, bombproof",
  "n02868870": "bombsight",
  "n02868975": "bone-ash cup, cupel, refractory pot",
  "n02869155": "bone china",
  "n02869249": "bones, castanets, clappers, finger cymbals",
  "n02869563": "boneshaker",
  "n02869737": "bongo, bongo drum",
  "n02869837": "bonnet, poke bonnet",
  "n02869965": "booby prize",
  "n02870092": "book, volume",
  "n02870526": "book",
  "n02870676": "book bag",
  "n02870772": "bookbindery",
  "n02870880": "bookcase",
  "n02871005": "bookend",
  "n02871147": "bookmark, bookmarker",
  "n02871314": "bookmobile",
  "n02871439": "bookshelf",
  "n02871525": "bookshop, bookstore, bookstall",
  "n02871631": "boom",
  "n02871824": "boom, microphone boom",
  "n02871963": "boomerang, throwing stick, throw stick",
  "n02872172": "booster, booster dose, booster shot, recall dose",
  "n02872333": "booster, booster rocket, booster unit, takeoff booster, takeoff rocket",
  "n02872529": "booster, booster amplifier, booster station, relay link, relay station, relay transmitter",
  "n02872752": "boot",
  "n02873244": "boot",
  "n02873363": "boot, the boot, iron boot, iron heel",
  "n02873520": "boot",
  "n02873623": "boot camp",
  "n02873733": "bootee, bootie",
  "n02873839": "booth, cubicle, stall, kiosk",
  "n02874086": "booth",
  "n02874214": "booth",
  "n02874336": "boothose",
  "n02874442": "bootjack",
  "n02874537": "bootlace",
  "n02874642": "bootleg",
  "n02874750": "bootstrap",
  "n02874876": "Bordeaux mixture",
  "n02875013": "border",
  "n02875233": "bore, bore-hole, drill hole",
  "n02875436": "bore bit, borer, rock drill, stone drill",
  "n02875626": "boron chamber",
  "n02875815": "boron counter tube",
  "n02875948": "borstal",
  "n02876084": "bosom",
  "n02876193": "Bosporus Bridge",
  "n02876326": "Boston rocker",
  "n02876457": "bota",
  "n02876537": "botanical",
  "n02876657": "bottle",
  "n02877266": "bottle, feeding bottle, nursing bottle",
  "n02877513": "bottle bank",
  "n02877642": "bottlebrush",
  "n02877765": "bottlecap",
  "n02877865": "bottleneck",
  "n02877962": "bottle opener",
  "n02878107": "bottling plant",
  "n02878222": "bottom, freighter, merchantman, merchant ship",
  "n02878425": "boucle",
  "n02878534": "boudoir",
  "n02878628": "boulle, boule, buhl",
  "n02878796": "bouncing betty",
  "n02878883": "Bounty, H.M.S. Bounty",
  "n02879087": "bouquet, corsage, posy, nosegay",
  "n02879228": "Bourse",
  "n02879309": "boutique, dress shop",
  "n02879422": "boutonniere",
  "n02879517": "bow",
  "n02879718": "bow",
  "n02880008": "bow, fore, prow, stem",
  "n02880189": "bow, bowknot",
  "n02880308": "bow",
  "n02880393": "bow and arrow",
  "n02880546": "bowed stringed instrument, string",
  "n02880842": "Bowie knife",
  "n02880940": "bowl",
  "n02881193": "bowl",
  "n02881397": "bowl, pipe bowl",
  "n02881546": "bowl",
  "n02881757": "bowler hat, bowler, derby hat, derby, plug hat",
  "n02881906": "bowline, bowline knot",
  "n02882014": "bowling alley, alley, skittle alley",
  "n02882190": "bowling alley",
  "n02882301": "bowling ball, bowl",
  "n02882483": "bowling equipment",
  "n02882647": "bowling pin, pin",
  "n02882894": "bowling shoe",
  "n02883004": "bowsprit",
  "n02883101": "bowstring",
  "n02883205": "bow tie, bow-tie, bowtie",
  "n02883344": "box",
  "n02884011": "box",
  "n02884225": "box, loge",
  "n02884450": "box, box seat",
  "n02884607": "box",
  "n02884859": "box beam, box girder",
  "n02884994": "box camera, box Kodak",
  "n02885108": "boxcar",
  "n02885233": "box coat",
  "n02885338": "boxing equipment",
  "n02885462": "boxing glove, glove",
  "n02885663": "boxing ring, prize ring",
  "n02885785": "box kite",
  "n02885882": "box office, ticket office, ticket booth",
  "n02886035": "box pleat",
  "n02886183": "box seat",
  "n02886321": "box spring",
  "n02886434": "box wrench, box end wrench",
  "n02886599": "brace, bracing",
  "n02886907": "brace, bitstock",
  "n02887079": "brace, braces, orthodontic braces",
  "n02887209": "brace",
  "n02887489": "brace, suspender, gallus",
  "n02887683": "brace",
  "n02887832": "brace and bit",
  "n02887970": "bracelet, bangle",
  "n02888133": "bracer, pick-me-up",
  "n02888270": "bracer, armguard",
  "n02888429": "brace wrench",
  "n02888569": "bracket, wall bracket",
  "n02888817": "brad",
  "n02888898": "bradawl, pricker",
  "n02889035": "braid, gold braid, braiding",
  "n02889228": "brail",
  "n02889332": "brail",
  "n02889425": "brake",
  "n02889646": "brake",
  "n02889856": "brake band",
  "n02889996": "brake cylinder, hydraulic brake cylinder, master cylinder",
  "n02890188": "brake disk",
  "n02890351": "brake drum, drum",
  "n02890513": "brake lining",
  "n02890662": "brake pad",
  "n02890804": "brake pedal",
  "n02890940": "brake shoe, shoe, skid",
  "n02891188": "brake system, brakes",
  "n02891430": "branch line, spur track, spur",
  "n02891566": "brand-name drug, proprietary drug",
  "n02891788": "brass, brass instrument",
  "n02892201": "brass, memorial tablet, plaque",
  "n02892304": "brass",
  "n02892392": "brassard",
  "n02892499": "brasserie",
  "n02892626": "brassie",
  "n02892767": "brassiere, bra, bandeau",
  "n02892948": "brass knucks, knucks, brass knuckles, knuckles, knuckle duster",
  "n02893148": "brass monkey",
  "n02893269": "brattice",
  "n02893418": "brazier, brasier",
  "n02893608": "breadbasket",
  "n02893692": "bread-bin, breadbox",
  "n02893800": "breadboard, bread board",
  "n02893941": "bread knife",
  "n02894024": "breakable",
  "n02894158": "breakfast area, breakfast nook",
  "n02894337": "breakfast table",
  "n02894431": "break seal",
  "n02894605": "breakwater, groin, groyne, mole, bulwark, seawall, jetty",
  "n02894847": "breast drill",
  "n02895008": "breast implant",
  "n02895154": "breastplate, aegis, egis",
  "n02895328": "breast pocket",
  "n02895438": "breathalyzer, breathalyser",
  "n02895606": "breathing device, breathing apparatus, breathing machine, ventilator",
  "n02895881": "breech, rear of barrel, rear of tube",
  "n02896074": "breechblock, breech closer",
  "n02896294": "breechcloth, breechclout, loincloth",
  "n02896442": "breeches, knee breeches, knee pants, knickerbockers, knickers",
  "n02896694": "breeches buoy",
  "n02896856": "breechloader",
  "n02896949": "breeder reactor",
  "n02897097": "Bren, Bren gun",
  "n02897237": "brewery",
  "n02897389": "brewpub",
  "n02897572": "briar, briar pipe",
  "n02897692": "bric-a-brac, knickknack, nicknack, knickknackery, whatnot",
  "n02897820": "brick",
  "n02898093": "brickkiln",
  "n02898173": "bricklayer's hammer",
  "n02898269": "brick trowel, mason's trowel",
  "n02898369": "brickwork",
  "n02898478": "brickyard, brickfield",
  "n02898585": "bridal gown, wedding gown, wedding dress",
  "n02898711": "bridge, span",
  "n02899257": "bridge, bridge deck",
  "n02899439": "bridge, nosepiece",
  "n02899569": "bridge, bridgework",
  "n02899694": "bridge",
  "n02899808": "bridge, bridge circuit",
  "n02900011": "bridged-T",
  "n02900160": "bridle",
  "n02900459": "bridle path, bridle road",
  "n02900594": "bridoon",
  "n02900705": "briefcase",
  "n02900857": "briefcase bomb",
  "n02900987": "briefcase computer",
  "n02901114": "briefs, Jockey shorts",
  "n02901259": "brig",
  "n02901377": "brig",
  "n02901481": "brigandine",
  "n02901620": "brigantine, hermaphrodite brig",
  "n02901793": "brilliantine",
  "n02901901": "brilliant pebble",
  "n02902079": "brim",
  "n02902250": "brim, rim, lip",
  "n02902411": "briquette, briquet",
  "n02902535": "bristle",
  "n02902687": "bristle brush",
  "n02902816": "britches",
  "n02902916": "broad arrow",
  "n02903006": "broadax, broadaxe",
  "n02903126": "brochette",
  "n02903204": "broadcaster, spreader",
  "n02903405": "broadcasting station, broadcast station",
  "n02903588": "broadcasting studio",
  "n02903727": "broadcloth",
  "n02903852": "broadcloth",
  "n02903964": "broad gauge",
  "n02904109": "broad hatchet",
  "n02904233": "broadloom",
  "n02904348": "broadside",
  "n02904505": "broadside",
  "n02904640": "broadsword",
  "n02904803": "brocade",
  "n02904927": "brogan, brogue, clodhopper, work shoe",
  "n02905036": "broiler",
  "n02905152": "broken arch",
  "n02905288": "brokerage house, brokerage",
  "n02905422": "brompheniramine maleate, Dimetane",
  "n02905612": "bronchodilator",
  "n02905886": "bronchoscope",
  "n02906027": "Bronx-Whitestone Bridge",
  "n02906175": "bronze",
  "n02906254": "bronze medal",
  "n02906438": "brooch, broach, breastpin",
  "n02906578": "Brooklyn Bridge",
  "n02906734": "broom",
  "n02906963": "broom closet",
  "n02907082": "broomstick, broom handle",
  "n02907194": "brougham",
  "n02907296": "brougham",
  "n02907391": "Browning automatic rifle, BAR",
  "n02907656": "Browning machine gun, Peacemaker",
  "n02907873": "brownstone",
  "n02907985": "Brown University, Brown",
  "n02908123": "brunch coat",
  "n02908217": "brush",
  "n02908605": "brush",
  "n02908773": "Brussels carpet",
  "n02908951": "Brussels lace",
  "n02909053": "bubble",
  "n02909165": "bubble chamber",
  "n02909285": "bubble jet printer, bubble-jet printer, bubblejet",
  "n02909408": "bubbler",
  "n02909543": "Buchenwald",
  "n02909706": "buckboard",
  "n02909870": "bucket, pail",
  "n02910145": "bucket seat",
  "n02910241": "bucket shop",
  "n02910353": "buckle",
  "n02910542": "buckram",
  "n02910701": "bucksaw",
  "n02910864": "buckskins",
  "n02910964": "buff, buffer",
  "n02911158": "buffer, fender",
  "n02911332": "buffer, polisher",
  "n02911485": "buffer, buffer storage, buffer store",
  "n02911890": "buffered aspirin, Bufferin",
  "n02912065": "buffet, counter, sideboard",
  "n02912319": "buffing wheel",
  "n02912440": "bug",
  "n02912557": "buggy, roadster",
  "n02912673": "buggy whip",
  "n02912894": "bugle",
  "n02913035": "bugle",
  "n02913152": "building, edifice",
  "n02914813": "building block",
  "n02914991": "building complex, complex",
  "n02915199": "building supply store, building supply house",
  "n02915379": "built-in bed",
  "n02915480": "bulb",
  "n02915675": "bulkhead",
  "n02915807": "bulla",
  "n02915904": "bulldog clip, alligator clip",
  "n02916065": "bulldog wrench",
  "n02916179": "bulldozer, dozer",
  "n02916350": "bullet, slug",
  "n02916538": "bulletin board, notice board",
  "n02916684": "bulletin board system, bulletin board, electronic bulletin board, bbs",
  "n02916936": "bulletproof vest",
  "n02917067": "bullet train, bullet",
  "n02917163": "bullfight, corrida",
  "n02917377": "bullhorn, loud hailer, loud-hailer",
  "n02917521": "bullion",
  "n02917607": "bullnose, bullnosed plane",
  "n02917742": "bullpen, detention cell, detention centre",
  "n02917964": "bullpen",
  "n02918112": "bullring",
  "n02918204": "bull tongue",
  "n02918330": "bulwark",
  "n02918455": "bumboat",
  "n02918595": "bumper",
  "n02918831": "bumper",
  "n02918964": "bumper car, Dodgem",
  "n02919148": "bumper guard",
  "n02919308": "bumper jack",
  "n02919414": "bundle, sheaf",
  "n02919648": "bung, spile",
  "n02919792": "bungalow, cottage",
  "n02919890": "bungee, bungee cord",
  "n02919976": "bunghole",
  "n02920083": "bunk",
  "n02920164": "bunk, feed bunk",
  "n02920259": "bunk bed, bunk",
  "n02920369": "bunker, sand trap, trap",
  "n02920503": "bunker, dugout",
  "n02920658": "bunker",
  "n02920785": "Bunker Buster, Guided Bomb Unit-28, GBU-28",
  "n02921029": "bunsen burner, bunsen, etna",
  "n02921195": "bunting",
  "n02921292": "bur, burr",
  "n02921406": "Burberry",
  "n02921592": "burette, buret",
  "n02921756": "burglar alarm",
  "n02921884": "burial chamber, sepulcher, sepulchre, sepulture",
  "n02922159": "burial garment",
  "n02922292": "burial mound, grave mound, barrow, tumulus",
  "n02922461": "burin",
  "n02922578": "burqa, burka",
  "n02922798": "burlap, gunny",
  "n02922877": "burn bag",
  "n02922996": "burn center",
  "n02923129": "burner",
  "n02923331": "burner",
  "n02923535": "burnous, burnoose, burnouse",
  "n02923682": "burp gun, machine pistol",
  "n02923802": "burr",
  "n02923915": "burr",
  "n02924023": "burthen",
  "n02924116": "bus, autobus, coach, charabanc, double-decker, jitney, motorbus, motorcoach, omnibus, passenger vehicle",
  "n02924554": "bus, jalopy, heap",
  "n02924713": "busbar, bus",
  "n02925009": "bushel basket",
  "n02925107": "bushing, cylindrical lining",
  "n02925250": "bushing",
  "n02925385": "bush jacket",
  "n02925519": "business suit",
  "n02925666": "buskin, combat boot, desert boot, half boot, top boot",
  "n02925808": "bus lane",
  "n02925901": "bus line",
  "n02926044": "buspirone, BuSpar",
  "n02926188": "bust",
  "n02926288": "bus terminal, bus depot, bus station, coach station",
  "n02926426": "bustier",
  "n02926591": "bustle",
  "n02926727": "butacaine, butacaine sulfate",
  "n02926886": "butcher board, butcher block",
  "n02927053": "butcher knife",
  "n02927161": "butcher shop, meat market",
  "n02927296": "butt, butt end",
  "n02927399": "butt, stub",
  "n02927608": "butt",
  "n02927764": "butter dish",
  "n02927887": "butterfly valve",
  "n02928049": "butter knife",
  "n02928172": "buttery",
  "n02928299": "butt hinge",
  "n02928413": "butt joint, butt",
  "n02928608": "button",
  "n02928841": "button",
  "n02928930": "buttonhole, button hole",
  "n02929059": "buttonhole stitch",
  "n02929184": "buttonhook",
  "n02929289": "buttress, buttressing",
  "n02929462": "butt shaft",
  "n02929582": "butt weld, butt-weld",
  "n02929749": "butyl nitrite, isobutyl nitrite",
  "n02929923": "buzz bomb, robot bomb, flying bomb, doodlebug, V-1",
  "n02930080": "buzzer",
  "n02930214": "BVD, BVD's",
  "n02930339": "bypass condenser, bypass capacitor",
  "n02930492": "by-product, byproduct, spin-off",
  "n02930645": "byway, bypath, byroad",
  "n02930766": "cab, hack, taxi, taxicab",
  "n02931013": "cab, cabriolet",
  "n02931148": "cab",
  "n02931294": "cabana",
  "n02931417": "cabaret, nightclub, night club, club, nightspot",
  "n02931836": "caber",
  "n02932019": "cabin",
  "n02932227": "cabin",
  "n02932400": "cabin",
  "n02932523": "cabin car, caboose",
  "n02932693": "cabin class, second class, economy class",
  "n02932891": "cabin cruiser, cruiser, pleasure boat, pleasure craft",
  "n02933112": "cabinet",
  "n02933340": "cabinet, console",
  "n02933462": "cabinet, locker, storage locker",
  "n02933649": "cabinetwork",
  "n02933750": "cabin liner",
  "n02933842": "cable",
  "n02933990": "cable, cable television, cable system, cable television service",
  "n02934168": "cable, line, transmission line",
  "n02934451": "cable car, car",
  "n02934641": "cable railway, funicular, funicular railway",
  "n02934888": "cache",
  "n02935017": "cache, memory cache",
  "n02935316": "cachet",
  "n02935387": "caddy, tea caddy",
  "n02935490": "caesium clock",
  "n02935658": "cafe, coffeehouse, coffee shop, coffee bar",
  "n02935891": "cafeteria",
  "n02936020": "cafeteria facility",
  "n02936176": "cafeteria tray",
  "n02936281": "caff",
  "n02936402": "caftan, kaftan",
  "n02936570": "caftan, kaftan",
  "n02936714": "cage, coop",
  "n02936921": "cage",
  "n02937010": "cagoule",
  "n02937093": "caisson, pneumatic caisson, cofferdam",
  "n02937237": "caisson, ammunition chest",
  "n02937336": "caisson",
  "n02937469": "cake, bar",
  "n02937614": "calabash, calabash pipe",
  "n02937766": "calamine lotion",
  "n02937958": "calash, caleche, calash top",
  "n02938095": "calash, caleche",
  "n02938218": "calceus",
  "n02938321": "calcimine",
  "n02938514": "calcium blocker, calcium-channel blocker",
  "n02938886": "calculator, calculating machine",
  "n02939185": "caldron, cauldron",
  "n02939291": "Caledonian Canal",
  "n02939565": "calender",
  "n02939763": "calico",
  "n02939866": "caliper, calliper",
  "n02940143": "calk, calkin",
  "n02940289": "call-board",
  "n02940385": "call center, call centre",
  "n02940570": "caller ID",
  "n02940706": "calliope, steam organ",
  "n02940878": "Caloosahatchee Canal",
  "n02941095": "calorimeter",
  "n02941228": "calpac, calpack, kalpac",
  "n02941408": "calumet, peace pipe, pipe of peace",
  "n02941605": "Calvary cross, cross of Calvary",
  "n02941716": "cam",
  "n02941845": "camail, aventail, ventail",
  "n02942015": "camber arch",
  "n02942147": "cambric",
  "n02942227": "Cambridge University, Cambridge",
  "n02942349": "camcorder",
  "n02942460": "camel's hair, camelhair",
  "n02942577": "cameo",
  "n02942699": "camera, photographic camera",
  "n02943241": "camera lens, optical lens",
  "n02943465": "camera lucida",
  "n02943686": "camera obscura",
  "n02943871": "camera tripod",
  "n02943964": "camise",
  "n02944075": "camisole",
  "n02944146": "camisole, underbodice",
  "n02944256": "camlet",
  "n02944375": "camlet",
  "n02944459": "camouflage",
  "n02944579": "camouflage, camo",
  "n02944826": "camp, encampment, cantonment, bivouac",
  "n02945161": "camp",
  "n02945379": "camp",
  "n02945594": "camp, summer camp",
  "n02945813": "camp, refugee camp",
  "n02945964": "campaign hat",
  "n02946127": "campanile, belfry",
  "n02946270": "camp chair",
  "n02946348": "camper, camping bus, motor home",
  "n02946509": "camper trailer",
  "n02946633": "camphor ice",
  "n02946753": "campstool",
  "n02946824": "camshaft",
  "n02946921": "can, tin, tin can",
  "n02947212": "canal",
  "n02947660": "canal boat, narrow boat, narrowboat",
  "n02947818": "candelabrum, candelabra",
  "n02947977": "candid camera",
  "n02948072": "candle, taper, wax light",
  "n02948293": "candlepin",
  "n02948403": "candlesnuffer",
  "n02948557": "candlestick, candle holder",
  "n02948719": "candlewick",
  "n02948834": "candlewick",
  "n02948942": "candy thermometer",
  "n02949084": "cane",
  "n02949202": "cane",
  "n02949356": "cangue",
  "n02949542": "canister, cannister, tin",
  "n02949691": "cannabis, marijuana, marihuana, ganja",
  "n02950018": "cannery",
  "n02950120": "cannikin",
  "n02950186": "cannikin",
  "n02950256": "cannon",
  "n02950482": "cannon",
  "n02950632": "cannon",
  "n02950826": "cannon",
  "n02950943": "cannonball, cannon ball, round shot",
  "n02951090": "cannon cracker",
  "n02951170": "cannula",
  "n02951358": "canoe",
  "n02951585": "can opener, tin opener",
  "n02951703": "canopic jar, canopic vase",
  "n02951843": "canopy",
  "n02952109": "canopy",
  "n02952237": "canopy",
  "n02952374": "canteen",
  "n02952485": "canteen",
  "n02952585": "canteen",
  "n02952674": "canteen, mobile canteen",
  "n02952798": "canteen",
  "n02952935": "cant hook",
  "n02953056": "cantilever",
  "n02953197": "cantilever bridge",
  "n02953455": "cantle",
  "n02953552": "Canton crepe",
  "n02953673": "canvas, canvass",
  "n02953850": "canvas, canvass",
  "n02954068": "canvas, canvass",
  "n02954163": "canvas tent, canvas, canvass",
  "n02954340": "cap",
  "n02954938": "cap",
  "n02955065": "cap",
  "n02955247": "capacitor, capacitance, condenser, electrical condenser",
  "n02955540": "caparison, trapping, housing",
  "n02955767": "cape, mantle",
  "n02955944": "capeline bandage",
  "n02956069": "capillary, capillary tube, capillary tubing",
  "n02956247": "capital, chapiter, cap",
  "n02956393": "capital ship",
  "n02956500": "Capitol, Capitol Building",
  "n02956699": "capitol",
  "n02956795": "cap opener",
  "n02956883": "capote, hooded cloak",
  "n02957008": "capote, hooded coat",
  "n02957135": "cap screw",
  "n02957252": "capstan",
  "n02957427": "capstone, copestone, coping stone, stretcher",
  "n02957586": "capsule",
  "n02957755": "capsule",
  "n02957862": "captain's chair",
  "n02958002": "captopril, Capoten",
  "n02958264": "capuchin",
  "n02958343": "car, auto, automobile, machine, motorcar",
  "n02959942": "car, railcar, railway car, railroad car",
  "n02960352": "car, elevator car",
  "n02960501": "car, gondola",
  "n02960690": "carabiner, karabiner, snap ring",
  "n02960903": "carafe, decanter",
  "n02961035": "caravansary, caravanserai, khan, caravan inn",
  "n02961225": "car battery, automobile battery",
  "n02961451": "carbine",
  "n02961544": "car bomb",
  "n02961714": "carbomycin",
  "n02961851": "carbon, carbon copy",
  "n02961947": "carbon arc lamp, carbon arc",
  "n02962061": "carboy",
  "n02962200": "carburetor, carburettor",
  "n02962414": "car carrier",
  "n02962545": "card",
  "n02962843": "cardcase",
  "n02962938": "cardiac monitor, heart monitor",
  "n02963159": "cardigan",
  "n02963302": "card index, card catalog, card catalogue",
  "n02963503": "cardiograph, electrocardiograph",
  "n02963692": "cardioid microphone",
  "n02963821": "car door",
  "n02963987": "cardroom",
  "n02964075": "card table",
  "n02964196": "card table",
  "n02964295": "car-ferry",
  "n02964389": "cargo, lading, freight, load, loading, payload, shipment, consignment",
  "n02964634": "cargo area, cargo deck, cargo hold, hold, storage area",
  "n02964843": "cargo container",
  "n02964934": "cargo door",
  "n02965024": "cargo hatch",
  "n02965122": "cargo helicopter",
  "n02965216": "cargo liner",
  "n02965300": "cargo ship, cargo vessel",
  "n02965529": "carillon",
  "n02965617": "carminative",
  "n02965783": "car mirror",
  "n02965936": "Carnegie Mellon University",
  "n02966068": "caroche",
  "n02966193": "carousel, carrousel, merry-go-round, roundabout, whirligig",
  "n02966372": "carousel, carrousel, luggage carousel, luggage carrousel",
  "n02966545": "carpenter's hammer, claw hammer, clawhammer",
  "n02966687": "carpenter's kit, tool kit",
  "n02966786": "carpenter's level",
  "n02966942": "carpenter's mallet",
  "n02967081": "carpenter's rule",
  "n02967170": "carpenter's square",
  "n02967294": "carpetbag",
  "n02967407": "carpet beater, rug beater",
  "n02967540": "carpet loom",
  "n02967626": "carpet pad, rug pad, underlay, underlayment",
  "n02967782": "carpet sweeper, sweeper",
  "n02967991": "carpet tack",
  "n02968074": "carport, car port",
  "n02968210": "carrack, carack",
  "n02968333": "carrel, carrell, cubicle, stall",
  "n02968473": "carriage, equipage, rig",
  "n02969010": "carriage",
  "n02969163": "carriage bolt",
  "n02969323": "carriageway",
  "n02969527": "carriage wrench",
  "n02969634": "carrick bend",
  "n02969754": "carrick bitt",
  "n02969886": "carrier",
  "n02970100": "carrier",
  "n02970311": "carron oil",
  "n02970408": "carryall, holdall, tote, tote bag",
  "n02970534": "carrycot",
  "n02970685": "car seat",
  "n02970849": "cart",
  "n02971167": "car tire, automobile tire, auto tire, rubber tire",
  "n02971356": "carton",
  "n02971473": "cartouche, cartouch",
  "n02971579": "car train",
  "n02971691": "cartridge",
  "n02971940": "cartridge, pickup",
  "n02972182": "cartridge",
  "n02972397": "cartridge belt",
  "n02972533": "cartridge ejector, ejector",
  "n02972714": "cartridge extractor, cartridge remover, extractor",
  "n02972934": "cartridge fuse",
  "n02973017": "cartridge holder, cartridge clip, clip, magazine",
  "n02973236": "cartwheel",
  "n02973395": "carvedilol",
  "n02973558": "carving",
  "n02973805": "carving fork",
  "n02973904": "carving knife",
  "n02974003": "car wheel",
  "n02974219": "car window",
  "n02974348": "caryatid",
  "n02974454": "cascade liquefier",
  "n02974565": "cascade transformer",
  "n02974697": "case",
  "n02975212": "case, display case, showcase, vitrine",
  "n02975412": "case, pillowcase, slip, pillow slip",
  "n02975589": "case, compositor's case, typesetter's case",
  "n02975994": "casein paint, casein",
  "n02976123": "case knife, sheath knife",
  "n02976249": "case knife",
  "n02976350": "casement",
  "n02976455": "casement window",
  "n02976552": "casern",
  "n02976641": "case shot, canister, canister shot",
  "n02976815": "cash bar",
  "n02976939": "cashbox, money box, till",
  "n02977058": "cash machine, cash dispenser, automated teller machine, automatic teller machine, automated teller, automatic teller, ATM",
  "n02977330": "cashmere",
  "n02977438": "cash register, register",
  "n02977619": "casing, case",
  "n02977822": "casing",
  "n02977936": "casino, gambling casino",
  "n02978055": "casket, jewel casket",
  "n02978205": "casque",
  "n02978367": "casquet, casquetel",
  "n02978478": "Cassegrainian telescope, Gregorian telescope",
  "n02978753": "casserole",
  "n02978881": "cassette",
  "n02979074": "cassette deck",
  "n02979186": "cassette player",
  "n02979290": "cassette recorder",
  "n02979399": "cassette tape",
  "n02979516": "cassock",
  "n02979662": "cast, casting",
  "n02979836": "cast, plaster cast, plaster bandage",
  "n02980036": "caster, castor",
  "n02980203": "caster, castor",
  "n02980325": "castile soap",
  "n02980441": "castle",
  "n02980625": "castle, rook",
  "n02980841": "castor oil",
  "n02981024": "catacomb",
  "n02981198": "catafalque",
  "n02981321": "catalytic converter",
  "n02981565": "catalytic cracker, cat cracker",
  "n02981792": "catamaran",
  "n02981911": "catapult, arbalest, arbalist, ballista, bricole, mangonel, onager, trebuchet, trebucket",
  "n02982232": "catapult, launcher",
  "n02982416": "catboat",
  "n02982515": "cat box",
  "n02982599": "catch",
  "n02982790": "catch, stop",
  "n02983072": "catchall",
  "n02983189": "catcher's mask",
  "n02983357": "catchment",
  "n02983507": "Caterpillar, cat",
  "n02983741": "catgut, gut",
  "n02983904": "cathedra, bishop's throne",
  "n02984061": "cathedral",
  "n02984203": "cathedral, duomo",
  "n02984328": "catherine wheel, pinwheel",
  "n02984469": "catheter",
  "n02984699": "cathode",
  "n02984937": "cathode",
  "n02985137": "cathode-ray tube, CRT",
  "n02985506": "catling",
  "n02985606": "cat-o'-nine-tails, cat",
  "n02985758": "cat rig",
  "n02985828": "cat's-paw",
  "n02985963": "catsup bottle, ketchup bottle",
  "n02986066": "cattle car",
  "n02986160": "cattle guard, cattle grid",
  "n02986348": "cattleship, cattle boat",
  "n02986480": "cattle trail",
  "n02986605": "catwalk",
  "n02986757": "catwalk",
  "n02986920": "causeway",
  "n02987047": "cautery, cauterant",
  "n02987379": "cavalier hat, slouch hat",
  "n02987492": "cavalry sword, saber, sabre",
  "n02987706": "cavetto",
  "n02987823": "cavity wall",
  "n02987950": "C battery",
  "n02988066": "C-clamp",
  "n02988156": "CD drive",
  "n02988304": "CD player",
  "n02988486": "CD-R, compact disc recordable, CD-WO, compact disc write-once",
  "n02988679": "CD-ROM, compact disc read-only memory",
  "n02988963": "CD-ROM drive",
  "n02989099": "cedar chest",
  "n02989178": "cefadroxil, Ultracef",
  "n02989313": "cefoperazone, Cefobid",
  "n02989475": "cefotaxime, Claforan",
  "n02989685": "ceftazidime, Fortaz, Tazicef",
  "n02989893": "ceftriaxone, Rocephin",
  "n02990103": "cefuroxime, Ceftin, Zinacef",
  "n02990373": "ceiling",
  "n02990561": "celecoxib, Celebrex",
  "n02990758": "celesta",
  "n02990920": "celestial globe",
  "n02991048": "cell, electric cell",
  "n02991302": "cell, jail cell, prison cell",
  "n02991555": "cell, cubicle",
  "n02991711": "cell",
  "n02991847": "cellar, wine cellar",
  "n02991950": "cellarage",
  "n02992032": "cellblock, ward",
  "n02992211": "cello, violoncello",
  "n02992368": "cellophane",
  "n02992529": "cellular telephone, cellular phone, cellphone, cell, mobile phone",
  "n02992795": "cellulose tape, Scotch tape, Sellotape",
  "n02993081": "Celtic cross",
  "n02993194": "cenotaph, empty tomb",
  "n02993368": "censer, thurible",
  "n02993546": "center, centre",
  "n02993877": "center bit, centre bit",
  "n02994012": "centerboard, centreboard, drop keel, sliding keel",
  "n02994219": "center field, centerfield, center",
  "n02994419": "centerpiece, centrepiece",
  "n02994573": "center punch",
  "n02994743": "Centigrade thermometer",
  "n02994858": "central, telephone exchange, exchange",
  "n02995151": "central heating",
  "n02995345": "central processing unit, CPU, C.P.U., central processor, processor, mainframe",
  "n02995770": "centrex",
  "n02995871": "centrifugal pump",
  "n02995998": "centrifuge, extractor, separator",
  "n02996249": "cephalexin, Keflex, Keflin, Keftab",
  "n02996570": "cephaloglycin, Kafocin",
  "n02996703": "cephaloridine",
  "n02996840": "cephalosporin, Mefoxin",
  "n02997292": "cephalothin",
  "n02997391": "ceramic",
  "n02997607": "ceramic ware",
  "n02997773": "cerate",
  "n02997910": "cereal bowl",
  "n02998003": "cereal box",
  "n02998107": "cerecloth",
  "n02998209": "cerivastatin, Baycol",
  "n02998363": "cervical cap",
  "n02998563": "cesspool, cesspit, sink, sump",
  "n02998696": "chachka, tsatske, tshatshke, tchotchke",
  "n02998841": "chador, chadar, chaddar, chuddar",
  "n02999001": "chaff",
  "n02999138": "chafing dish",
  "n02999272": "chafing gear",
  "n02999410": "chain",
  "n02999757": "chain, string, strand",
  "n02999936": "chain",
  "n03000021": "chain",
  "n03000134": "chainlink fence",
  "n03000247": "chain mail, ring mail, mail, chain armor, chain armour, ring armor, ring armour",
  "n03000530": "chain printer",
  "n03000684": "chain saw, chainsaw",
  "n03000808": "chain stitch",
  "n03000966": "chain stitch",
  "n03001115": "chain store",
  "n03001282": "chain tongs",
  "n03001540": "chain wrench",
  "n03001627": "chair",
  "n03002096": "chair",
  "n03002210": "chair of state",
  "n03002341": "chairlift, chair lift",
  "n03002555": "chaise, shay",
  "n03002711": "chaise longue, chaise, daybed",
  "n03002816": "chalet",
  "n03002948": "chalice, goblet",
  "n03003091": "chalk",
  "n03003378": "chalk line, snap line, snapline",
  "n03003548": "chalkpit, chalk pit",
  "n03003633": "challis",
  "n03003730": "chamber",
  "n03004146": "chamber",
  "n03004275": "chamberpot, potty, thunder mug",
  "n03004409": "chambray",
  "n03004531": "chamfer bit",
  "n03004620": "chamfer plane",
  "n03004713": "chamois cloth",
  "n03004824": "chancel, sanctuary, bema",
  "n03005033": "chancellery",
  "n03005147": "chancery",
  "n03005285": "chandelier, pendant, pendent",
  "n03005425": "chandlery",
  "n03005515": "chandlery",
  "n03005619": "chanfron, chamfron, testiere, frontstall, front-stall",
  "n03005769": "change",
  "n03005920": "change",
  "n03006105": "channel",
  "n03006398": "channel, television channel, TV channel",
  "n03006626": "chanter, melody pipe",
  "n03006788": "chantry",
  "n03006903": "chap",
  "n03007130": "chapel",
  "n03007297": "chapterhouse, fraternity house, frat house",
  "n03007444": "chapterhouse",
  "n03007591": "character printer, character-at-a-time printer, serial printer",
  "n03007824": "charcoal, fusain",
  "n03007955": "charcoal",
  "n03008080": "charcoal burner",
  "n03008177": "charcuterie",
  "n03008275": "charge, burster, bursting charge, explosive charge",
  "n03008565": "charge, bearing, heraldic bearing, armorial bearing",
  "n03008817": "charge-exchange accelerator",
  "n03008976": "charger, battery charger",
  "n03009111": "chariot",
  "n03009269": "chariot",
  "n03009477": "Charlestown Navy Yard",
  "n03009633": "charm, good luck charm",
  "n03009794": "charnel house, charnel",
  "n03009920": "chart",
  "n03010057": "charterhouse",
  "n03010138": "Chartres Cathedral",
  "n03010283": "chase",
  "n03010473": "chassis",
  "n03010656": "chassis",
  "n03010795": "chasuble",
  "n03010915": "chateau",
  "n03011018": "chatelaine",
  "n03011162": "check",
  "n03011355": "checker, chequer",
  "n03011521": "checkerboard, checker board",
  "n03011741": "checkout, checkout counter",
  "n03011892": "checkroom, left-luggage office",
  "n03012013": "cheekpiece",
  "n03012159": "cheeseboard, cheese tray",
  "n03012263": "cheesecake",
  "n03012373": "cheesecloth",
  "n03012499": "cheese cutter",
  "n03012644": "cheese press",
  "n03012734": "chemical bomb, gas bomb",
  "n03012897": "chemical plant",
  "n03013006": "chemical reactor",
  "n03013162": "chemical weapon",
  "n03013438": "chemise, sack, shift",
  "n03013580": "chemise, shimmy, shift, slip, teddy",
  "n03013718": "chemistry lab, chemistry laboratory, chem lab",
  "n03013850": "chenille",
  "n03013992": "chenille, chenille cord",
  "n03014119": "cheroot",
  "n03014204": "cherry bomb",
  "n03014317": "chessboard, chess board",
  "n03014440": "chessman, chess piece",
  "n03014705": "chest",
  "n03015052": "chesterfield",
  "n03015149": "chesterfield",
  "n03015254": "chest of drawers, chest, bureau, dresser",
  "n03015478": "chest protector",
  "n03015631": "cheval-de-frise, chevaux-de-frise",
  "n03015851": "cheval glass",
  "n03015975": "chevron",
  "n03016056": "chiaroscuro",
  "n03016209": "chicane",
  "n03016389": "chicken coop, coop, hencoop, henhouse",
  "n03016511": "chicken farm",
  "n03016609": "chicken wire",
  "n03016737": "chicken yard, hen yard, chicken run, fowl run",
  "n03016868": "chiffon",
  "n03016953": "chiffonier, commode",
  "n03017070": "child's room",
  "n03017168": "chime, bell, gong",
  "n03017428": "chimney",
  "n03017698": "chimney breast",
  "n03017835": "chimney corner, inglenook",
  "n03017931": "chimneypot",
  "n03018058": "chimneystack",
  "n03018209": "china",
  "n03018349": "china cabinet, china closet",
  "n03018493": "chinaware, china",
  "n03018614": "chinchilla",
  "n03018712": "Chinese lantern",
  "n03018848": "Chinese puzzle",
  "n03018971": "Chinese Wall, Great Wall, Great Wall of China",
  "n03019198": "chinning bar",
  "n03019304": "chino",
  "n03019434": "chino",
  "n03019535": "chinoiserie",
  "n03019685": "chin rest",
  "n03019806": "chin strap",
  "n03019938": "chintz",
  "n03020034": "chip, microchip, micro chip, silicon chip, microprocessor chip",
  "n03020416": "chip, poker chip",
  "n03020563": "chip",
  "n03020692": "chisel",
  "n03020927": "Chisholm Trail",
  "n03021121": "chiton",
  "n03021228": "chlamys",
  "n03021360": "chloral hydrate",
  "n03021531": "chlorambucil, Leukeran",
  "n03021696": "chloramine, chloramine-T",
  "n03021858": "chloramphenicol, Chloromycetin",
  "n03022041": "chlordiazepoxide, Librium, Libritabs",
  "n03022250": "chlorhexidine",
  "n03022406": "chloroform, trichloromethane",
  "n03022634": "chloroquine",
  "n03022788": "chlorothiazide, Diuril",
  "n03022978": "chlorpheniramine maleate, Coricidin, Chlor-Trimeton",
  "n03023175": "chlorpromazine, Thorazine",
  "n03023415": "chlortetracycline, Aureomycin",
  "n03023623": "chlorthalidone, Hygroton, Thalidone",
  "n03023878": "chock, wedge",
  "n03024064": "choir",
  "n03024233": "choir loft",
  "n03024333": "choke",
  "n03024518": "choke, choke coil, choking coil",
  "n03024746": "choker, ruff, ruffle, neck ruff",
  "n03024882": "choker, collar, dog collar, neckband",
  "n03025070": "chokey, choky",
  "n03025165": "choo-choo",
  "n03025250": "chopine, platform",
  "n03025357": "chopping block",
  "n03025513": "chopping board, cutting board",
  "n03025641": "chop shop",
  "n03025755": "chopstick",
  "n03025886": "chordophone",
  "n03026093": "choropleth map",
  "n03026350": "chrism, chrisom, sacramental oil, holy oil",
  "n03026506": "Christmas stocking",
  "n03026626": "Christmas tree",
  "n03026741": "chromatogram",
  "n03026907": "chronograph",
  "n03027001": "chronometer",
  "n03027108": "chronoscope",
  "n03027250": "chuck",
  "n03027505": "chuck wagon",
  "n03027625": "chukka, chukka boot",
  "n03027797": "chum",
  "n03027935": "chunnel, Channel Tunnel",
  "n03028079": "church, church building",
  "n03028596": "church bell",
  "n03028785": "church hat",
  "n03028907": "Churchill Downs",
  "n03029066": "church key",
  "n03029197": "church tower",
  "n03029296": "churidars",
  "n03029445": "churn, butter churn",
  "n03029603": "chute, slide, slideway, sloping trough",
  "n03029812": "cider mill",
  "n03029925": "ciderpress",
  "n03030035": "cigar",
  "n03030262": "cigar band",
  "n03030353": "cigar box",
  "n03030433": "cigar butt",
  "n03030557": "cigar cutter",
  "n03030663": "cigarette, cigaret, coffin nail, butt, fag",
  "n03030880": "cigarette butt",
  "n03031012": "cigarette case",
  "n03031152": "cigarette holder",
  "n03031306": "cigarillo",
  "n03031422": "cigar lighter, cigarette lighter, pocket lighter",
  "n03031553": "cimetidine, Tagamet",
  "n03031756": "cinch, girth",
  "n03031957": "cinder block, clinker block, breeze block",
  "n03032158": "cinder track",
  "n03032252": "cinema, movie theater, movie theatre, movie house, picture palace",
  "n03032453": "cinquefoil",
  "n03032576": "ciprofloxacin, Cipro",
  "n03032811": "circle, round",
  "n03033019": "circle, dress circle",
  "n03033267": "circlet",
  "n03033362": "circuit, electrical circuit, electric circuit",
  "n03033986": "circuit board, circuit card, board, card, plug-in, add-in",
  "n03034244": "circuit breaker, breaker",
  "n03034405": "circuitry",
  "n03034516": "circular plane, compass plane",
  "n03034663": "circular saw, buzz saw",
  "n03034860": "circus",
  "n03035089": "circus",
  "n03035252": "circus tent, big top, round top, top",
  "n03035510": "cistern",
  "n03035715": "cistern, water tank",
  "n03035832": "cittern, cithern, cither, citole, gittern",
  "n03036022": "city hall",
  "n03036149": "cityscape",
  "n03036244": "city university",
  "n03036341": "civies, civvies",
  "n03036469": "civilian clothing, civilian dress, civilian garb, plain clothes",
  "n03036701": "clack valve, clack, clapper valve",
  "n03036866": "clamp, clinch",
  "n03037108": "clamshell, grapple",
  "n03037228": "clapper, tongue",
  "n03037404": "clapperboard",
  "n03037590": "clarence",
  "n03037709": "clarinet",
  "n03037899": "clarion",
  "n03038041": "Clark cell, Clark standard cell",
  "n03038190": "claro",
  "n03038281": "clasp",
  "n03038480": "clasp knife, jackknife",
  "n03038595": "classic",
  "n03038685": "classroom, schoolroom",
  "n03038870": "clavichord",
  "n03039015": "clavier, Klavier",
  "n03039156": "claw hatchet",
  "n03039259": "clay pigeon",
  "n03039353": "claymore mine, claymore",
  "n03039493": "claymore",
  "n03039618": "clay pipe",
  "n03039711": "clean bomb",
  "n03039827": "cleaners, dry cleaners",
  "n03039947": "cleaning implement, cleaning device, cleaning equipment",
  "n03040229": "cleaning pad",
  "n03040376": "clean room, white room",
  "n03040587": "cleansing agent, cleanser, cleaner",
  "n03040836": "clearway",
  "n03040983": "cleat",
  "n03041114": "cleat",
  "n03041265": "cleat",
  "n03041449": "cleats",
  "n03041632": "cleaver, meat cleaver, chopper",
  "n03041810": "clerestory, clearstory",
  "n03041964": "clerical collar, Roman collar, dog collar",
  "n03042139": "clevis",
  "n03042384": "clews",
  "n03042490": "cliff dwelling",
  "n03042697": "climbing frame",
  "n03042829": "clinch",
  "n03042984": "clinch, clench",
  "n03043173": "clincher",
  "n03043274": "clinic",
  "n03043423": "clinical thermometer, mercury-in-glass clinical thermometer",
  "n03043693": "clinker, clinker brick",
  "n03043798": "clinometer, inclinometer",
  "n03043958": "clip",
  "n03044166": "clip",
  "n03044278": "clip art",
  "n03044418": "clipboard",
  "n03044537": "clip joint",
  "n03044671": "clip lead",
  "n03044801": "clip-on",
  "n03044934": "clipper",
  "n03045074": "clipper",
  "n03045228": "clipper, clipper ship",
  "n03045337": "cloak",
  "n03045698": "cloak",
  "n03045800": "cloakroom, coatroom",
  "n03045928": "cloakroom",
  "n03046029": "cloche",
  "n03046133": "cloche",
  "n03046257": "clock",
  "n03046657": "clock face, clock dial",
  "n03046802": "clock pendulum",
  "n03046921": "clock radio",
  "n03047052": "clock tower",
  "n03047171": "clockwork",
  "n03047353": "clofibrate, Atromid-S",
  "n03047553": "clog",
  "n03047690": "clog, geta, patten, sabot",
  "n03047799": "cloisonne",
  "n03047941": "cloister",
  "n03048094": "clomiphene, clomiphene citrate, Clomid",
  "n03048322": "clomipramine",
  "n03048412": "clonidine, Catapres",
  "n03048598": "clopidogrel bisulfate, Plavix",
  "n03048883": "closed circuit, loop",
  "n03049066": "closed-circuit television",
  "n03049326": "closed loop, closed-loop system",
  "n03049457": "closet",
  "n03049605": "closet auger",
  "n03049695": "closeup",
  "n03049782": "closeup lens",
  "n03049924": "cloth cap, flat cap",
  "n03050026": "cloth covering",
  "n03050453": "clothesbrush",
  "n03050546": "clothes closet, clothespress",
  "n03050655": "clothes dryer, clothes drier",
  "n03050864": "clothes hamper, laundry basket, clothes basket, voider",
  "n03051041": "clotheshorse",
  "n03051152": "clothesline",
  "n03051249": "clothespin, clothes pin, clothes peg",
  "n03051396": "clothes tree, coat tree, coat stand",
  "n03051540": "clothing, article of clothing, vesture, wear, wearable, habiliment",
  "n03052464": "clothing store, haberdashery, haberdashery store, mens store",
  "n03052628": "cloud chamber, Wilson cloud chamber",
  "n03052917": "clout nail, clout",
  "n03053047": "clove hitch",
  "n03053163": "cloverleaf",
  "n03053272": "clozapine, Clozaril",
  "n03053474": "club",
  "n03053788": "club",
  "n03053976": "club car, lounge car",
  "n03054098": "club drug",
  "n03054311": "clubhouse, club",
  "n03054491": "clubroom",
  "n03054605": "cluster bomb",
  "n03054901": "clutch",
  "n03055159": "clutch, clutch pedal",
  "n03055418": "clutch bag, clutch",
  "n03055537": "CN Tower",
  "n03055670": "coach, four-in-hand, coach-and-four",
  "n03055857": "coach house, carriage house, remise",
  "n03056010": "coalbin, coalhole",
  "n03056097": "coal car",
  "n03056215": "coal chute",
  "n03056288": "coal house",
  "n03056368": "coal mine, coalpit",
  "n03056493": "coal shovel",
  "n03056583": "coaming",
  "n03056701": "coaster",
  "n03056873": "coaster brake",
  "n03057021": "coat",
  "n03057541": "coat button",
  "n03057636": "coat closet",
  "n03057724": "coatdress",
  "n03057841": "coatee",
  "n03057920": "coat hanger, clothes hanger, dress hanger",
  "n03058107": "coating, coat",
  "n03058603": "coating",
  "n03058726": "coat of arms, arms, blazon, blazonry",
  "n03058949": "coat of paint",
  "n03059103": "coatrack, coat rack, hatrack",
  "n03059236": "coattail",
  "n03059366": "coaxial cable, coax, coax cable",
  "n03059528": "cobble, cobblestone, sett",
  "n03059685": "cobweb",
  "n03059806": "cobweb, gossamer",
  "n03059934": "cobweb",
  "n03060074": "coca",
  "n03060294": "cocaine, cocain",
  "n03060599": "cockade",
  "n03060728": "Cockcroft and Walton accelerator, Cockcroft-Walton accelerator, Cockcroft and Walton voltage multiplier, Cockcroft-Walton voltage multiplier",
  "n03061050": "cocked hat",
  "n03061211": "cockhorse",
  "n03061345": "cockleshell",
  "n03061428": "cockloft",
  "n03061505": "cockpit",
  "n03061674": "cockpit",
  "n03061819": "cockpit",
  "n03061893": "cockscomb, coxcomb",
  "n03062015": "cocktail dress, sheath",
  "n03062122": "cocktail lounge",
  "n03062245": "cocktail shaker",
  "n03062336": "cocotte",
  "n03062461": "codeine",
  "n03062651": "codpiece",
  "n03062798": "coelostat",
  "n03062985": "coffee can",
  "n03063073": "coffee cup",
  "n03063199": "coffee filter",
  "n03063338": "coffee maker",
  "n03063485": "coffee mill, coffee grinder",
  "n03063599": "coffee mug",
  "n03063689": "coffeepot",
  "n03063834": "coffee stall",
  "n03063968": "coffee table, cocktail table",
  "n03064118": "coffee-table book",
  "n03064250": "coffee urn",
  "n03064350": "coffer",
  "n03064443": "coffer, caisson, lacuna",
  "n03064562": "Coffey still",
  "n03064758": "coffin, casket",
  "n03064935": "cog, sprocket",
  "n03065063": "cog railway, rack railway",
  "n03065243": "coif",
  "n03065424": "coil, spiral, volute, whorl, helix",
  "n03065708": "coil",
  "n03066029": "coil",
  "n03066130": "coil",
  "n03066232": "coil",
  "n03066359": "coil spring, volute spring",
  "n03066464": "coin box",
  "n03066606": "coin slot",
  "n03066743": "coke, blow, nose candy, snow, C",
  "n03066849": "colander, cullender",
  "n03066965": "colchicine",
  "n03067093": "cold cathode",
  "n03067212": "cold chisel, set chisel",
  "n03067339": "cold cream, coldcream, face cream, vanishing cream",
  "n03067518": "cold frame",
  "n03067690": "cold medicine",
  "n03067810": "cold-water flat",
  "n03067912": "collage, montage",
  "n03068181": "collar, neckband",
  "n03068486": "collar",
  "n03068707": "collar, shoe collar",
  "n03068862": "collar",
  "n03068998": "collar",
  "n03069213": "collectible, collectable",
  "n03069381": "collector",
  "n03069567": "collector's item, showpiece, piece de resistance",
  "n03069752": "college",
  "n03069919": "collet",
  "n03070059": "collet, collet chuck",
  "n03070193": "collider",
  "n03070396": "colliery, pit",
  "n03070587": "collimator",
  "n03070854": "collimator",
  "n03071021": "cologne, cologne water, eau de cologne",
  "n03071160": "colonnade",
  "n03071288": "colonoscope",
  "n03071431": "colophon",
  "n03071552": "colorimeter, tintometer",
  "n03071782": "coloring book",
  "n03071923": "colors, colours",
  "n03072056": "colors, colours",
  "n03072201": "color television, colour television, color television system, colour television system, color TV, colour TV",
  "n03072440": "color tube, colour tube, color television tube, colour television tube, color TV tube, colour TV tube",
  "n03072682": "color wash, colour wash",
  "n03072828": "Colosseum, Amphitheatrum Flavium",
  "n03073016": "Colossus of Rhodes",
  "n03073296": "Colt",
  "n03073384": "colter, coulter",
  "n03073545": "columbarium",
  "n03073694": "columbarium, cinerarium",
  "n03073832": "Columbia University, Columbia",
  "n03073977": "column, pillar",
  "n03074380": "column, pillar",
  "n03074574": "column, chromatography column",
  "n03074855": "comb",
  "n03075097": "comb",
  "n03075248": "comber",
  "n03075370": "combination lock",
  "n03075500": "combination plane",
  "n03075634": "combine",
  "n03075768": "comforter, pacifier, baby's dummy, teething ring",
  "n03075946": "command module",
  "n03076104": "command post, general headquarters, GHQ",
  "n03076298": "commercial art",
  "n03076411": "commissary",
  "n03076623": "commissary",
  "n03076708": "commodity, trade good, good",
  "n03077074": "commodity exchange, commodities exchange, commodities market",
  "n03077304": "Commodore John Barry Bridge",
  "n03077442": "common ax, common axe, Dayton ax, Dayton axe",
  "n03077616": "common room",
  "n03077741": "communications satellite",
  "n03077958": "communication system, communication equipment",
  "n03078287": "communication system",
  "n03078506": "community center, civic center",
  "n03078670": "commutator",
  "n03078802": "commuter, commuter train",
  "n03078995": "compact, powder compact",
  "n03079136": "compact, compact car",
  "n03079230": "compact disk, compact disc, CD",
  "n03079494": "compact-disk burner, CD burner",
  "n03079616": "companionway",
  "n03079741": "compartment",
  "n03080309": "compartment",
  "n03080497": "compass",
  "n03080633": "compass",
  "n03080731": "compass card, mariner's compass",
  "n03080904": "compass saw",
  "n03081021": "component, constituent, element",
  "n03081660": "composition",
  "n03081859": "compound",
  "n03081986": "compound lens",
  "n03082127": "compound lever",
  "n03082280": "compound microscope",
  "n03082450": "compress",
  "n03082656": "compression bandage, tourniquet",
  "n03082807": "compressor",
  "n03082979": "computer, computing machine, computing device, data processor, electronic computer, information processing system",
  "n03084204": "computer accessory",
  "n03084420": "computer circuit",
  "n03084647": "computer graphics",
  "n03084834": "computerized axial tomography scanner, CAT scanner",
  "n03085013": "computer keyboard, keypad",
  "n03085219": "computer monitor",
  "n03085333": "computer network",
  "n03085602": "computer screen, computer display",
  "n03085781": "computer store",
  "n03085915": "computer system, computing system, automatic data processing system, ADP system, ADPS",
  "n03086183": "concentration camp, stockade",
  "n03086457": "concert grand, concert piano",
  "n03086580": "concert hall",
  "n03086670": "concertina",
  "n03086868": "concertina",
  "n03086963": "concourse",
  "n03087069": "concrete mixer, cement mixer",
  "n03087245": "condensation pump, diffusion pump",
  "n03087366": "condenser, optical condenser",
  "n03087521": "condenser",
  "n03087643": "condenser",
  "n03087816": "condenser microphone, capacitor microphone",
  "n03088020": "conditioner",
  "n03088164": "condom, rubber, safety, safe, prophylactic",
  "n03088389": "condominium",
  "n03088580": "condominium, condo",
  "n03088707": "conductor",
  "n03089014": "conduit",
  "n03089348": "cone",
  "n03089477": "cone clutch, cone friction clutch",
  "n03089624": "confectionery, confectionary, candy store",
  "n03089753": "conference center, conference house",
  "n03089879": "conference room",
  "n03090000": "conference table, council table, council board",
  "n03090172": "confessional",
  "n03090278": "confetti",
  "n03090437": "conformal projection, orthomorphic projection",
  "n03090598": "conge, congee",
  "n03090710": "congress boot, congress shoe, congress gaiter",
  "n03090856": "conic projection, conical projection",
  "n03091044": "connecting rod",
  "n03091223": "connecting room",
  "n03091374": "connection, connexion, connector, connecter, connective",
  "n03091907": "conning tower",
  "n03092053": "conning tower",
  "n03092166": "conservatory, hothouse, indoor garden",
  "n03092314": "conservatory, conservatoire",
  "n03092476": "console",
  "n03092656": "console",
  "n03092883": "console table, console",
  "n03093018": "Constitution, Old Ironsides",
  "n03093427": "consulate",
  "n03093574": "consumer goods",
  "n03093792": "contact, tangency",
  "n03094159": "contact, contact lens",
  "n03094347": "contact print",
  "n03094503": "container",
  "n03095699": "container ship, containership, container vessel",
  "n03095965": "containment",
  "n03096142": "contour map, relief map",
  "n03096273": "contraband",
  "n03096439": "contrabassoon, contrafagotto, double bassoon",
  "n03096593": "contraceptive, preventive, preventative, contraceptive device, prophylactic device, birth control device",
  "n03096960": "control, controller",
  "n03097362": "control center",
  "n03097535": "control circuit, negative feedback circuit",
  "n03097673": "control key, command key",
  "n03097890": "controlled substance",
  "n03098140": "control panel, instrument panel, control board, board, panel",
  "n03098515": "control rod",
  "n03098688": "control room",
  "n03098806": "control system",
  "n03098959": "control tower",
  "n03099147": "convector",
  "n03099274": "convenience store",
  "n03099454": "convent",
  "n03099622": "conventicle, meetinghouse",
  "n03099771": "converging lens, convex lens",
  "n03099945": "converter, convertor",
  "n03100240": "convertible",
  "n03100346": "convertible, sofa bed",
  "n03100490": "conveyance, transport",
  "n03100897": "conveyer belt, conveyor belt, conveyer, conveyor, transporter",
  "n03101156": "cooker",
  "n03101302": "cookfire",
  "n03101375": "cookhouse",
  "n03101517": "cookie cutter",
  "n03101664": "cookie jar, cooky jar",
  "n03101796": "cookie sheet, baking tray",
  "n03101986": "cooking utensil, cookware",
  "n03102371": "cookstove",
  "n03102516": "coolant system",
  "n03102654": "cooler, ice chest",
  "n03102771": "cooler, tank",
  "n03102859": "cooling system, cooling",
  "n03103128": "cooling system, engine cooling system",
  "n03103396": "cooling tower",
  "n03103563": "coonskin cap, coonskin",
  "n03103682": "Cooper Union, Cooper Union for the Advancement of Science and Art",
  "n03103904": "cope",
  "n03104019": "coping saw",
  "n03104147": "copper mine",
  "n03104247": "copperplate, copperplate engraving",
  "n03104413": "copperplate",
  "n03104512": "copperware",
  "n03104594": "copy",
  "n03105088": "copyholder",
  "n03105214": "coquille",
  "n03105306": "coracle",
  "n03105467": "corbel, truss",
  "n03105645": "corbel arch",
  "n03105810": "corbel step, corbie-step, corbiestep, crow step",
  "n03105974": "corbie gable",
  "n03106110": "cord",
  "n03106722": "cord, corduroy",
  "n03106898": "cord, electric cord",
  "n03107046": "cordage",
  "n03107152": "cordite",
  "n03107383": "cordon",
  "n03107488": "cords, corduroys",
  "n03107609": "corduroy",
  "n03107716": "core",
  "n03107904": "core",
  "n03108069": "core, magnetic core",
  "n03108455": "core bit",
  "n03108624": "core drill",
  "n03108759": "corer",
  "n03108853": "cork, bottle cork",
  "n03109033": "corker",
  "n03109150": "corkscrew, bottle screw",
  "n03109253": "corncrib",
  "n03109350": "Cornell University",
  "n03109486": "corner, street corner, turning point",
  "n03109693": "corner, quoin",
  "n03109881": "corner, nook",
  "n03110100": "corner pocket",
  "n03110202": "corner post",
  "n03110332": "cornerstone",
  "n03110470": "cornerstone",
  "n03110669": "cornet, horn, trumpet, trump",
  "n03110937": "corn exchange",
  "n03111041": "cornice",
  "n03111177": "cornice",
  "n03111296": "cornice, valance, valance board, pelmet",
  "n03111483": "corona",
  "n03111564": "coronet",
  "n03111690": "correctional institution",
  "n03111899": "corrective, restorative",
  "n03112099": "corridor",
  "n03112240": "corrugated fastener, wiggle nail",
  "n03112431": "corrugated iron",
  "n03112605": "corsair",
  "n03112719": "corselet, corslet",
  "n03112869": "corset, girdle, stays",
  "n03113035": "corvette",
  "n03113152": "cosmetic",
  "n03113337": "cosmography",
  "n03113505": "cosmotron",
  "n03113657": "costume",
  "n03113835": "costume",
  "n03114041": "costume",
  "n03114236": "costume",
  "n03114379": "cosy, tea cosy, cozy, tea cozy",
  "n03114504": "cot, camp bed",
  "n03114634": "cote",
  "n03114743": "cottage tent",
  "n03114839": "cotter, cottar",
  "n03115014": "cotter pin",
  "n03115180": "cotton",
  "n03115301": "cotton",
  "n03115400": "cotton flannel, Canton flannel",
  "n03115525": "cotton gin, gin",
  "n03115663": "cotton mill",
  "n03115762": "couch",
  "n03115897": "couch",
  "n03116008": "couchette",
  "n03116163": "coude telescope, coude system",
  "n03116333": "coulisse",
  "n03116435": "coulisse, wing flat",
  "n03116530": "counter",
  "n03116767": "counter, tabulator",
  "n03116942": "counter, heel counter",
  "n03117199": "counter",
  "n03117420": "counter",
  "n03117642": "counterbore, countersink, countersink bit",
  "n03117776": "counterirritant",
  "n03117939": "counterpart, similitude, twin",
  "n03118051": "countersink",
  "n03118245": "countertop",
  "n03118346": "counter tube",
  "n03118539": "counterweight, counterbalance, counterpoise, balance, equalizer, equaliser",
  "n03118846": "countinghouse",
  "n03118969": "country house",
  "n03119203": "country store, general store, trading post",
  "n03119396": "coupe",
  "n03119510": "coupling, coupler",
  "n03119790": "course",
  "n03120029": "course, row",
  "n03120198": "court, courtyard",
  "n03120491": "court",
  "n03120778": "court, courtroom",
  "n03121040": "court",
  "n03121190": "Courtelle",
  "n03121298": "courthouse",
  "n03121431": "courthouse",
  "n03121556": "court plaster",
  "n03121698": "cover, cover version, cover song",
  "n03121897": "coverall",
  "n03122073": "covered bridge",
  "n03122202": "covered couch",
  "n03122295": "covered wagon, Conestoga wagon, Conestoga, prairie wagon, prairie schooner",
  "n03122573": "cover glass, cover slip",
  "n03122748": "covering",
  "n03123553": "coverlet",
  "n03123666": "cover plate",
  "n03123809": "cowbarn, cowshed, cow barn, cowhouse, byre",
  "n03123917": "cowbell",
  "n03124043": "cowboy boot",
  "n03124170": "cowboy hat, ten-gallon hat",
  "n03124313": "cowhide",
  "n03124474": "cowl",
  "n03124590": "cow pen, cattle pen, corral",
  "n03124700": "Cox-2 inhibitor",
  "n03125057": "CPU board, mother board",
  "n03125184": "crack, crack cocaine, tornado",
  "n03125352": "cracker, snapper, cracker bonbon",
  "n03125588": "crackle, crackleware, crackle china",
  "n03125729": "cradle",
  "n03125870": "craft",
  "n03126090": "cramp, cramp iron",
  "n03126251": "cramp",
  "n03126385": "crampon, crampoon, climbing iron, climber",
  "n03126580": "crampon, crampoon",
  "n03126707": "crane",
  "n03126927": "craniometer",
  "n03127024": "crank, starter",
  "n03127203": "crankcase",
  "n03127302": "crank handle, starting handle",
  "n03127408": "crankshaft",
  "n03127531": "crash barrier",
  "n03127747": "crash helmet",
  "n03127925": "crate",
  "n03128085": "cravat",
  "n03128248": "crayon, wax crayon",
  "n03128427": "crazy quilt",
  "n03128519": "cream, ointment, emollient",
  "n03128868": "creamery",
  "n03129001": "cream pitcher, creamer",
  "n03129123": "creation",
  "n03129471": "creche, foundling hospital",
  "n03129636": "creche",
  "n03129753": "credenza, credence",
  "n03129848": "creel",
  "n03129944": "creep",
  "n03130066": "crematory, crematorium, cremation chamber",
  "n03130233": "crematory, crematorium",
  "n03130340": "crenel, crenelle",
  "n03130563": "crepe, crape",
  "n03130761": "crepe de Chine",
  "n03130866": "crescent wrench",
  "n03131038": "crest",
  "n03131193": "cretonne",
  "n03131325": "crewelwork",
  "n03131431": "crew neck, crew neckline",
  "n03131574": "crib, cot",
  "n03131669": "crib",
  "n03131791": "cribbage board",
  "n03131967": "cricket ball",
  "n03132076": "cricket bat, bat",
  "n03132261": "cricket equipment",
  "n03132438": "cringle, eyelet, loop, grommet, grummet",
  "n03132666": "crinoline",
  "n03132776": "crinoline",
  "n03132879": "crochet, crocheting",
  "n03133050": "crochet needle, crochet hook",
  "n03133177": "crochet stitch",
  "n03133415": "crock, earthenware jar",
  "n03133538": "crockery, dishware",
  "n03133744": "crocket",
  "n03133878": "Crock Pot",
  "n03134015": "croft",
  "n03134118": "crook, shepherd's crook",
  "n03134232": "Crookes radiometer",
  "n03134394": "Crookes tube",
  "n03134496": "crop",
  "n03134595": "crop",
  "n03134739": "croquet ball",
  "n03134853": "croquet equipment",
  "n03135030": "croquet mallet",
  "n03135152": "Cross",
  "n03135532": "cross",
  "n03135656": "crossbar",
  "n03135788": "crossbar",
  "n03135917": "crossbar",
  "n03136051": "crossbench",
  "n03136254": "cross bit",
  "n03136369": "crossbow",
  "n03136504": "crosscut saw, crosscut handsaw, cutoff saw",
  "n03136657": "crosse",
  "n03136773": "cross hair, cross wire",
  "n03137044": "crosshead",
  "n03137228": "crossing, crosswalk, crossover",
  "n03137473": "crossjack, mizzen course",
  "n03137579": "crosspiece",
  "n03137744": "cross-stitch",
  "n03137863": "cross-stitch",
  "n03137973": "cross street",
  "n03138128": "crotchet",
  "n03138217": "croupier's rake",
  "n03138344": "crowbar, wrecking bar, pry, pry bar",
  "n03138534": "crown",
  "n03138669": "crown, diadem",
  "n03138856": "crown",
  "n03138981": "crown, crest",
  "n03139089": "crown, crownwork, jacket, jacket crown, cap",
  "n03139341": "crown jewel",
  "n03139464": "crown jewels",
  "n03139640": "crown lens",
  "n03139731": "crown of thorns",
  "n03139887": "crown saw",
  "n03139998": "crow's nest",
  "n03140126": "crucible, melting pot",
  "n03140292": "crucifix, rood, rood-tree",
  "n03140431": "cruet, crewet",
  "n03140546": "cruet-stand",
  "n03140652": "cruise control",
  "n03140771": "cruise missile",
  "n03140900": "cruiser",
  "n03141065": "cruiser, police cruiser, patrol car, police car, prowl car, squad car",
  "n03141327": "cruise ship, cruise liner",
  "n03141455": "crupper",
  "n03141612": "cruse",
  "n03141702": "crusher",
  "n03141823": "crutch",
  "n03141991": "cryocautery",
  "n03142099": "cryometer",
  "n03142205": "cryoscope",
  "n03142325": "cryostat",
  "n03142431": "crypt",
  "n03142579": "cryptograph",
  "n03142679": "crystal, watch crystal, watch glass",
  "n03142834": "crystal",
  "n03142912": "crystal",
  "n03143131": "crystal ball",
  "n03143255": "crystal counter",
  "n03143400": "crystal detector",
  "n03143572": "crystal microphone",
  "n03143754": "crystal oscillator, quartz oscillator",
  "n03143982": "crystal pickup",
  "n03144156": "crystal set",
  "n03144262": "Cuban heel",
  "n03144365": "cubby, cubbyhole, snuggery, snug",
  "n03144486": "cubbyhole, pigeonhole",
  "n03144592": "cube, square block",
  "n03144756": "cubeb, cubeb cigarette",
  "n03144873": "cubitiere",
  "n03144982": "cucking stool, ducking stool",
  "n03145147": "cuckoo clock",
  "n03145277": "cuddy",
  "n03145384": "cudgel",
  "n03145522": "cue, cue stick, pool cue, pool stick",
  "n03145719": "cue ball",
  "n03145843": "cuff, turnup",
  "n03146075": "cufflink",
  "n03146219": "cuirass",
  "n03146342": "cuisse",
  "n03146449": "cul, cul de sac, dead end",
  "n03146560": "culdoscope",
  "n03146687": "cullis",
  "n03146777": "culotte",
  "n03146846": "cultivator, tiller",
  "n03147084": "culverin",
  "n03147156": "culverin",
  "n03147280": "culvert",
  "n03147397": "cummerbund",
  "n03147509": "cup",
  "n03147901": "cup, loving cup",
  "n03148130": "cup",
  "n03148324": "cupboard, closet",
  "n03148518": "cup hook",
  "n03148632": "Cupid's bow",
  "n03148727": "cupola",
  "n03148808": "cupola",
  "n03148920": "curb, curbing, kerb",
  "n03149135": "curb, curb bit",
  "n03149261": "curb market",
  "n03149401": "curb roof",
  "n03149531": "curbside",
  "n03149686": "curbstone, kerbstone",
  "n03149810": "curette, curet",
  "n03149951": "curio, curiosity, oddity, oddment, peculiarity, rarity",
  "n03150232": "curler, hair curler, roller, crimper",
  "n03150511": "curling iron",
  "n03150661": "currycomb",
  "n03150795": "cursor, pointer",
  "n03151077": "curtain, drape, drapery, mantle, pall",
  "n03151401": "curtain ring",
  "n03151500": "cushion",
  "n03151800": "cushion",
  "n03151916": "cusp",
  "n03152062": "cuspidation",
  "n03152144": "custard pie",
  "n03152303": "customhouse, customshouse",
  "n03152483": "custom-made, custom-built",
  "n03152619": "cut, gash",
  "n03152743": "cut",
  "n03152830": "cutaway",
  "n03152951": "cutaway, cutaway drawing, cutaway model",
  "n03153145": "cut glass",
  "n03153246": "cutlas, cutlass",
  "n03153375": "cutlery, eating utensil",
  "n03153585": "cutoff",
  "n03153681": "cutout",
  "n03153821": "cutout",
  "n03153948": "cutout",
  "n03154073": "cutter, cutlery, cutting tool",
  "n03154316": "cutter",
  "n03154446": "cutting implement",
  "n03154616": "cutting room",
  "n03154745": "cutty stool",
  "n03154895": "cutwork",
  "n03155055": "cyberart",
  "n03155178": "cybercafe",
  "n03155334": "cyclobenzaprine, Flexeril",
  "n03155502": "cyclopean masonry",
  "n03155661": "cyclopropane",
  "n03155788": "cycloserine",
  "n03155915": "cyclostyle",
  "n03156071": "cyclotron",
  "n03156279": "cylinder",
  "n03156405": "cylinder, piston chamber",
  "n03156606": "cylinder head",
  "n03156767": "cylinder lock",
  "n03156990": "cyma, cymatium",
  "n03157215": "cyma recta",
  "n03157348": "cymbal",
  "n03157582": "cyproheptadine, Periactin",
  "n03157751": "cytophotometer",
  "n03157987": "cytotoxic drug",
  "n03158186": "dacha",
  "n03158259": "Dachau",
  "n03158414": "Dacron, Terylene",
  "n03158542": "dado",
  "n03158668": "dado",
  "n03158796": "dado plane",
  "n03158885": "dagger, sticker",
  "n03159176": "daggerboard",
  "n03159331": "daguerreotype",
  "n03159535": "dairy, dairy farm",
  "n03159640": "dais, podium, pulpit, rostrum, ambo, stump, soapbox",
  "n03159828": "daisy chain",
  "n03160001": "daisy print wheel, daisy wheel",
  "n03160186": "daisywheel printer",
  "n03160309": "dam, dike, dyke",
  "n03160593": "damascene",
  "n03160740": "damask",
  "n03160894": "damask",
  "n03161016": "dampener, moistener",
  "n03161228": "damp-proof course, damp course",
  "n03161450": "damper, muffler",
  "n03161725": "damper",
  "n03161893": "damper block, piano damper",
  "n03162080": "dance floor",
  "n03162171": "dapsone",
  "n03162297": "dark lantern, bull's-eye",
  "n03162460": "darkroom",
  "n03162556": "darning needle, embroidery needle",
  "n03162714": "dart",
  "n03162818": "dart",
  "n03162940": "dartboard, dart board",
  "n03163081": "Dartmouth College, Dartmouth",
  "n03163222": "dashboard, fascia",
  "n03163381": "dashiki, daishiki",
  "n03163488": "dash-pot",
  "n03163649": "dasymeter",
  "n03163798": "data converter",
  "n03163973": "data input device, input device",
  "n03164192": "data multiplexer",
  "n03164344": "data system, information system",
  "n03164532": "daub",
  "n03164605": "davenport",
  "n03164722": "davenport",
  "n03164809": "Davis Cup",
  "n03164929": "davit",
  "n03165096": "daybed, divan bed",
  "n03165211": "daybook, ledger",
  "n03165343": "day camp",
  "n03165466": "day nursery, day care center",
  "n03165616": "day school",
  "n03165718": "dead-air space",
  "n03165823": "dead axle",
  "n03165955": "deadeye",
  "n03166120": "deadhead",
  "n03166213": "deadlight",
  "n03166360": "dead load",
  "n03166514": "deanery",
  "n03166600": "deathbed",
  "n03166685": "death camp",
  "n03166809": "death house, death row",
  "n03166951": "death knell, death bell",
  "n03167053": "death mask",
  "n03167153": "death seat",
  "n03167337": "deathtrap",
  "n03167464": "decal, decalcomania",
  "n03167666": "deck",
  "n03167978": "deck",
  "n03168107": "deck",
  "n03168217": "deck chair, beach chair",
  "n03168364": "decker",
  "n03168543": "deck-house",
  "n03168663": "deckle",
  "n03168774": "deckle edge, deckle",
  "n03168933": "declinometer, transit declinometer",
  "n03169063": "decoder",
  "n03169176": "decolletage",
  "n03169271": "decongestant",
  "n03169390": "decoration, ornament, ornamentation",
  "n03170292": "decoupage",
  "n03170459": "dedicated file server",
  "n03170635": "deep-freeze, Deepfreeze, deep freezer, freezer",
  "n03170872": "deerstalker",
  "n03171002": "deer trail",
  "n03171094": "defense laboratory",
  "n03171228": "defense system, defence system",
  "n03171356": "defensive structure, defense, defence",
  "n03171635": "defibrillator",
  "n03171910": "defilade",
  "n03172038": "deflector",
  "n03172211": "defroster, deicer",
  "n03172432": "delavirdine, Rescriptor",
  "n03172602": "Delaware Memorial Bridge",
  "n03172738": "delayed action",
  "n03172965": "delay line",
  "n03173142": "delf",
  "n03173270": "delft",
  "n03173387": "delicatessen, deli, food shop",
  "n03173524": "delineation, depiction, limning, line drawing",
  "n03173736": "deliverable",
  "n03173929": "delivery truck, delivery van, panel truck",
  "n03174079": "delta wing",
  "n03174211": "demeclocycline hydrochloride, Declomycin",
  "n03174450": "demijohn",
  "n03174605": "demister",
  "n03174731": "demitasse",
  "n03174829": "demulcent",
  "n03174991": "Demulen",
  "n03175081": "den",
  "n03175189": "denim, dungaree, jean",
  "n03175301": "densimeter, densitometer",
  "n03175457": "densitometer",
  "n03175604": "dental appliance",
  "n03175843": "dental floss, floss",
  "n03175983": "dental implant",
  "n03176084": "dentifrice",
  "n03176238": "dentist's drill, burr drill",
  "n03176386": "denture, dental plate, plate",
  "n03176594": "deodorant, deodourant",
  "n03176763": "department store, emporium",
  "n03176970": "departure gate",
  "n03177059": "departure lounge",
  "n03177165": "depilatory, depilator, epilator",
  "n03177349": "depository, deposit, depositary, repository",
  "n03177708": "depressor",
  "n03177856": "depth charge, depth bomb",
  "n03178000": "depth finder",
  "n03178173": "depth gauge, depth gage",
  "n03178316": "dermatome",
  "n03178430": "derrick",
  "n03178538": "derrick",
  "n03178674": "derringer",
  "n03178782": "design, pattern, figure",
  "n03179318": "design",
  "n03179489": "designer drug",
  "n03179701": "desk",
  "n03179910": "desk phone",
  "n03180011": "desktop computer",
  "n03180153": "desipramine",
  "n03180280": "dessert plate",
  "n03180384": "dessert spoon",
  "n03180504": "destroyer, guided missile destroyer",
  "n03180732": "destroyer escort",
  "n03180865": "detached house, single dwelling",
  "n03180969": "detector, sensor, sensing element",
  "n03181293": "detector",
  "n03181501": "detector, demodulator",
  "n03181667": "detention home, detention house, house of detention, detention camp",
  "n03181899": "detergent",
  "n03182140": "detonating fuse",
  "n03182232": "detonator, detonating device, cap",
  "n03182506": "detour, roundabout way",
  "n03182683": "detox",
  "n03182795": "deuce, two",
  "n03182912": "developer",
  "n03183080": "device",
  "n03185562": "device",
  "n03185746": "device",
  "n03185868": "Dewar flask, Dewar",
  "n03186005": "dextroamphetamine sulphate, Dexedrine",
  "n03186199": "dhoti",
  "n03186285": "dhow",
  "n03186399": "diagram",
  "n03186696": "dial",
  "n03186818": "dial, telephone dial",
  "n03187037": "dial",
  "n03187153": "dial",
  "n03187268": "dialog box, panel",
  "n03187595": "dial telephone, dial phone",
  "n03187751": "dialyzer, dialysis machine",
  "n03187972": "diamond",
  "n03188168": "diamond point",
  "n03188290": "diamante",
  "n03188416": "diapason, diapason stop",
  "n03188531": "diaper, nappy, napkin",
  "n03188725": "diaper",
  "n03188871": "diaphone",
  "n03188979": "diaphoretic",
  "n03189083": "diaphragm, stop",
  "n03189311": "diaphragm",
  "n03189461": "diaphragm, pessary, contraceptive diaphragm",
  "n03189707": "diary",
  "n03189818": "diathermy machine",
  "n03189995": "diazepam, Valium",
  "n03190303": "diazoxide, Hyperstat",
  "n03190458": "dibble, dibber",
  "n03190655": "dibucaine",
  "n03190763": "dideoxycytosine, ddC, DDC, zalcitabine",
  "n03190897": "dideoxyinosine, ddI, DDI, didanosine",
  "n03191029": "die, dice",
  "n03191286": "dice cup, dice box",
  "n03191451": "dicer",
  "n03191561": "dickey, dickie, dicky, shirtfront",
  "n03191776": "dickey, dickie, dicky, dickey-seat, dickie-seat, dicky-seat",
  "n03191967": "diclofenac potassium, Cataflam",
  "n03192142": "diclofenac sodium, Voltaren",
  "n03192347": "dicloxacillin, Dynapen",
  "n03192543": "Dictaphone",
  "n03192653": "dicumarol, dicoumarol",
  "n03192790": "die",
  "n03192907": "die",
  "n03193107": "diesel, diesel engine, diesel motor",
  "n03193260": "diesel-electric locomotive, diesel-electric",
  "n03193423": "diesel-hydraulic locomotive, diesel-hydraulic",
  "n03193597": "diesel locomotive",
  "n03193754": "diestock",
  "n03193882": "diethylstilbesterol, DES, stilbesterol",
  "n03194170": "differential analyzer",
  "n03194297": "differential gear, differential",
  "n03194538": "diffraction grating, grating",
  "n03194812": "diffuser, diffusor",
  "n03194992": "diffuser, diffusor",
  "n03195118": "diflunisal, Dolobid",
  "n03195332": "digester",
  "n03195485": "diggings, digs, domiciliation, lodgings, pad",
  "n03195659": "diggings, digs",
  "n03195799": "digital-analog converter, digital-to-analog converter",
  "n03195959": "digital audiotape, DAT",
  "n03196062": "digital camera",
  "n03196217": "digital clock",
  "n03196324": "digital computer",
  "n03196598": "digital display, alphanumeric display",
  "n03196841": "digital plethysmograph",
  "n03196990": "digital subscriber line, DSL",
  "n03197201": "digital voltmeter",
  "n03197337": "digital watch",
  "n03197446": "digitizer, digitiser, analog-digital converter, analog-to-digital converter",
  "n03197666": "digitoxin",
  "n03197804": "digoxin, Lanoxin",
  "n03198028": "dihydrostreptomycin",
  "n03198223": "dilator, dilater",
  "n03198383": "dilator",
  "n03198500": "dildo",
  "n03198637": "diltiazem, Cardizem",
  "n03198819": "dime bag, dime",
  "n03198951": "dimenhydrinate, Dramamine",
  "n03199142": "Dimetapp",
  "n03199358": "dimity",
  "n03199488": "dimmer",
  "n03199647": "diner",
  "n03199775": "dinette",
  "n03199901": "dinghy, dory, rowboat",
  "n03200152": "dinky, dinkey",
  "n03200231": "dining area",
  "n03200357": "dining car, diner, dining compartment, buffet car",
  "n03200539": "dining-hall",
  "n03200701": "dining room, dining-room",
  "n03200906": "dining-room furniture",
  "n03201035": "dining-room table",
  "n03201208": "dining table, board",
  "n03201529": "dinner bell",
  "n03201638": "dinner dress, dinner gown, formal, evening gown",
  "n03201776": "dinner jacket, tux, tuxedo, black tie",
  "n03201895": "dinner napkin",
  "n03201996": "dinner pail, dinner bucket",
  "n03202123": "dinner plate",
  "n03202246": "dinner service, dinner set",
  "n03202354": "dinner table",
  "n03202481": "dinner theater, dinner theatre",
  "n03202622": "dinnerware",
  "n03202760": "diode, semiconductor diode, junction rectifier, crystal rectifier",
  "n03202940": "diode, rectifying tube, rectifying valve",
  "n03203089": "dip",
  "n03203225": "diphenhydramine, Benadryl",
  "n03203441": "diphenylhydantoin, phenytoin, Dilantin",
  "n03203641": "diphenylbutyl piperidine",
  "n03203806": "diplomatic building",
  "n03203959": "diplomatic pouch",
  "n03204134": "dipole, dipole antenna",
  "n03204306": "dipper",
  "n03204436": "dipstick",
  "n03204558": "DIP switch, dual inline package switch",
  "n03204810": "diptych",
  "n03204955": "directional antenna",
  "n03205143": "directional microphone",
  "n03205304": "direction finder",
  "n03205458": "dirk",
  "n03205574": "dirndl",
  "n03205669": "dirndl",
  "n03205760": "dirt track",
  "n03205903": "dirty bomb",
  "n03206023": "discharge lamp",
  "n03206158": "discharge pipe",
  "n03206282": "disco, discotheque",
  "n03206405": "discount house, discount store, discounter, wholesale house",
  "n03206602": "discus, saucer",
  "n03206718": "disguise",
  "n03206908": "dish",
  "n03207305": "dish, dish aerial, dish antenna, saucer",
  "n03207548": "dishpan",
  "n03207630": "dish rack",
  "n03207743": "dishrag, dishcloth",
  "n03207835": "dishtowel, dish towel, tea towel",
  "n03207941": "dishwasher, dish washer, dishwashing machine",
  "n03208062": "dishwasher detergent, dishwashing detergent, dishwashing liquid",
  "n03208229": "disinfectant, germicide, antimicrobic, antimicrobial",
  "n03208556": "disk, disc",
  "n03208815": "disk access",
  "n03208938": "disk brake, disc brake",
  "n03209141": "disk cache",
  "n03209359": "disk clutch",
  "n03209477": "disk controller",
  "n03209666": "disk drive, disc drive, hard drive, Winchester drive",
  "n03209910": "diskette, floppy, floppy disk",
  "n03210245": "disk harrow, disc harrow",
  "n03210372": "dispatch case, dispatch box",
  "n03210552": "dispensary",
  "n03210683": "dispenser",
  "n03210940": "display, presentation",
  "n03211117": "display, video display",
  "n03211413": "display adapter, display adaptor",
  "n03211616": "display panel, display board, board",
  "n03211789": "display window, shop window, shopwindow, show window",
  "n03212003": "disposable",
  "n03212114": "disposal, electric pig, garbage disposal",
  "n03212247": "disrupting explosive, bursting explosive",
  "n03212406": "distaff",
  "n03212535": "distemper",
  "n03212698": "distemper",
  "n03212811": "distillery, still",
  "n03213014": "distributor, distributer, electrical distributor",
  "n03213361": "distributor cam",
  "n03213538": "distributor cap",
  "n03213715": "distributor housing",
  "n03213826": "distributor point, breaker point, point",
  "n03214051": "disulfiram, Antabuse",
  "n03214253": "ditch",
  "n03214450": "ditch spade, long-handled spade",
  "n03214582": "ditty bag",
  "n03214670": "diuretic drug, diuretic, water pill",
  "n03214966": "divan",
  "n03215076": "divan, diwan",
  "n03215191": "dive bomber",
  "n03215337": "diverging lens, concave lens",
  "n03215508": "divided highway, dual carriageway",
  "n03215749": "divider",
  "n03215930": "diving bell",
  "n03216080": "diving board",
  "n03216199": "divining rod, dowser, dowsing rod, waterfinder, water finder",
  "n03216402": "diving suit, diving dress",
  "n03216562": "dixie",
  "n03216710": "Dixie cup, paper cup",
  "n03216828": "dock, dockage, docking facility",
  "n03217191": "dock",
  "n03217333": "dock, loading dock",
  "n03217458": "document",
  "n03217653": "doeskin",
  "n03217739": "dogcart",
  "n03217814": "dog collar",
  "n03217889": "doggie bag, doggy bag",
  "n03218100": "dogleg",
  "n03218198": "dogsled, dog sled, dog sleigh",
  "n03218334": "dogtooth",
  "n03218446": "dog wrench",
  "n03218545": "doodad, doohickey, doojigger, gimmick, gizmo, gismo, gubbins, thingamabob, thingumabob, thingmabob, thingamajig, thingumajig, thingmajig, thingummy, whatchamacallit, whatchamacallum, whatsis, widget",
  "n03219010": "doily, doyley, doyly",
  "n03219135": "doll, dolly",
  "n03219362": "dollhouse, doll's house",
  "n03219483": "dollhouse, doll's house",
  "n03219612": "dolly",
  "n03219732": "dolly",
  "n03219859": "dolman",
  "n03219966": "dolman, dolman jacket",
  "n03220095": "dolman sleeve",
  "n03220237": "dolmen, cromlech, portal tomb",
  "n03220401": "dolphin striker, martingale",
  "n03220513": "dome",
  "n03220692": "dome, domed stadium, covered stadium",
  "n03220802": "domino",
  "n03221059": "domino, half mask, eye mask",
  "n03221205": "domino",
  "n03221351": "dongle",
  "n03221540": "donkey jacket",
  "n03221643": "doodlebug",
  "n03221720": "door",
  "n03222176": "door",
  "n03222318": "door",
  "n03222516": "doorbell, bell, buzzer",
  "n03222722": "doorframe, doorcase",
  "n03222857": "doorjamb, doorpost",
  "n03222959": "doorknob, doorhandle",
  "n03223162": "doorlock",
  "n03223299": "doormat, welcome mat",
  "n03223441": "doornail",
  "n03223553": "doorplate",
  "n03223686": "doorsill, doorstep, threshold",
  "n03223923": "doorstop, doorstopper",
  "n03224032": "doorway, door, room access, threshold",
  "n03224387": "dooryard",
  "n03224490": "Doppler radar",
  "n03224603": "dormer, dormer window",
  "n03224753": "dormer window",
  "n03224893": "dormitory, dorm, residence hall, hall, student residence",
  "n03225108": "dormitory, dormitory room, dorm room",
  "n03225238": "dose, dosage",
  "n03225458": "dosemeter, dosimeter",
  "n03225616": "dossal, dossel",
  "n03225777": "dot matrix printer, matrix printer, dot printer",
  "n03225988": "double bed",
  "n03226090": "double-bitted ax, double-bitted axe, Western ax, Western axe",
  "n03226254": "double boiler, double saucepan",
  "n03226375": "double-breasted jacket",
  "n03226538": "double-breasted suit",
  "n03226660": "double clinch",
  "n03226743": "double crochet, double stitch",
  "n03226880": "double door",
  "n03227010": "double glazing",
  "n03227184": "double-hung window",
  "n03227317": "double knit",
  "n03227505": "double-prop, double-propeller plane, twin-prop, twin-propeller-plane",
  "n03227721": "doubler",
  "n03227856": "double reed",
  "n03228016": "double-reed instrument, double reed",
  "n03228254": "doublet",
  "n03228365": "doubletree",
  "n03228533": "douche, douche bag",
  "n03228692": "dovecote, columbarium, columbary",
  "n03228796": "Dover's powder",
  "n03228967": "dovetail, dovetail joint",
  "n03229115": "dovetail plane",
  "n03229244": "dowel, dowel pin, joggle",
  "n03229420": "downcast",
  "n03229526": "downstage",
  "n03229656": "doxazosin, Cardura",
  "n03229905": "doxepin, doxepin hydrochloride, Adapin, Sinequan",
  "n03230149": "doxorubicin",
  "n03230247": "doxycycline, Vibramycin",
  "n03230446": "DPT vaccine",
  "n03230670": "draft, draught",
  "n03230785": "draft, rough drawing",
  "n03230914": "draft",
  "n03231024": "drafting board, drawing board",
  "n03231160": "drafting instrument",
  "n03231368": "drafting table, drawing table",
  "n03231476": "drag",
  "n03231723": "dragee",
  "n03231819": "Dragunov",
  "n03231912": "drain, drainpipe, waste pipe",
  "n03232158": "drain",
  "n03232309": "drainage ditch",
  "n03232417": "drainage system",
  "n03232543": "drain basket",
  "n03232676": "drainboard, draining board",
  "n03232815": "drainplug",
  "n03232923": "drape",
  "n03233123": "drapery",
  "n03233246": "draw",
  "n03233423": "draw, lot",
  "n03233624": "drawbar",
  "n03233744": "drawbridge, lift bridge",
  "n03233905": "drawer",
  "n03234164": "drawers, underdrawers, shorts, boxers, boxershorts",
  "n03234306": "drawing",
  "n03234795": "drawing card, loss leader, leader",
  "n03234952": "drawing chalk",
  "n03235042": "drawing room, withdrawing room",
  "n03235180": "drawing room",
  "n03235327": "drawknife, drawshave",
  "n03235433": "drawnwork",
  "n03235560": "drawstring, drawing string, string",
  "n03235796": "drawstring bag",
  "n03235979": "dray, camion",
  "n03236093": "dreadnought, dreadnaught",
  "n03236217": "dredge",
  "n03236423": "dredger",
  "n03236580": "dredging bucket",
  "n03236735": "dress, frock",
  "n03237212": "dress blues, dress whites",
  "n03237340": "dresser",
  "n03237416": "dress hat, high hat, opera hat, silk hat, stovepipe, top hat, topper, beaver",
  "n03237639": "dressing, medical dressing",
  "n03237839": "dressing case",
  "n03237992": "dressing gown, robe-de-chambre, lounging robe",
  "n03238131": "dressing room",
  "n03238286": "dressing sack, dressing sacque",
  "n03238407": "dressing station, aid station",
  "n03238586": "dressing table, dresser, vanity, toilet table",
  "n03238762": "dress rack",
  "n03238879": "dress shirt, evening shirt",
  "n03239054": "dress suit, full dress, tailcoat, tail coat, tails, white tie, white tie and tails",
  "n03239259": "dress uniform",
  "n03239399": "drift, heading, gallery",
  "n03239607": "drift net",
  "n03239726": "drill",
  "n03240140": "electric drill",
  "n03240327": "drilling bit, drill bit",
  "n03240482": "drilling pipe",
  "n03240683": "drilling platform, offshore rig",
  "n03240892": "drill press",
  "n03241093": "drill rig, drilling rig, oilrig, oil rig",
  "n03241236": "drill site",
  "n03241335": "drinking fountain, water fountain, bubbler",
  "n03241496": "drinking vessel",
  "n03241660": "drip, drip mold, drip mould",
  "n03241903": "drip loop",
  "n03242120": "drip mat",
  "n03242264": "drip pan",
  "n03242390": "dripping pan, drip pan",
  "n03242506": "drip pot",
  "n03242595": "dripstone, hoodmold, hoodmould",
  "n03242713": "drive, parkway",
  "n03242995": "drive",
  "n03243218": "drive",
  "n03243501": "drive-in",
  "n03243625": "drive line, drive line system",
  "n03243903": "driven well, tube well",
  "n03244047": "driver, number one wood",
  "n03244231": "driveshaft",
  "n03244388": "driveway, drive, private road",
  "n03244660": "driving belt",
  "n03244775": "driving iron, one iron",
  "n03244919": "driving wheel",
  "n03245075": "Drixoral",
  "n03245271": "drogue, drogue chute, drogue parachute",
  "n03245421": "drogue parachute",
  "n03245553": "dronabinol",
  "n03245724": "drone, drone pipe, bourdon",
  "n03245889": "drone, pilotless aircraft, radio-controlled aircraft",
  "n03246052": "drop",
  "n03246197": "drop arch",
  "n03246312": "drop cloth",
  "n03246454": "drop curtain, drop cloth, drop",
  "n03246653": "drop forge, drop hammer, drop press",
  "n03246788": "drop-leaf",
  "n03246933": "drop-leaf table",
  "n03247083": "dropper, eye dropper",
  "n03247351": "droshky, drosky",
  "n03247495": "drove, drove chisel",
  "n03247620": "drug",
  "n03248560": "drug cocktail, highly active antiretroviral therapy, HAART",
  "n03248835": "drugget",
  "n03248958": "drug of abuse, street drug",
  "n03249342": "drugstore, apothecary's shop, chemist's, chemist's shop, pharmacy",
  "n03249569": "drum, membranophone, tympan",
  "n03249956": "drum, metal drum",
  "n03250089": "drum brake",
  "n03250279": "drumhead, head",
  "n03250405": "drum printer",
  "n03250588": "drum sander, electric sander, sander, smoother",
  "n03250847": "drumstick",
  "n03250952": "dry battery",
  "n03251100": "dry-bulb thermometer",
  "n03251280": "dry cell",
  "n03251533": "dry dock, drydock, graving dock",
  "n03251766": "dryer, drier",
  "n03251932": "dry fly",
  "n03252064": "drygoods, soft goods",
  "n03252231": "dry kiln",
  "n03252324": "dry masonry",
  "n03252422": "dry point",
  "n03252542": "dry point",
  "n03252637": "dry wall, dry-stone wall",
  "n03252787": "dual scan display",
  "n03252959": "dubbing",
  "n03253071": "duck",
  "n03253187": "duckboard",
  "n03253279": "duckpin",
  "n03253398": "duct",
  "n03253516": "duct tape",
  "n03253714": "dudeen",
  "n03253796": "duffel, duffle",
  "n03253886": "duffel bag, duffle bag, duffel, duffle",
  "n03254046": "duffel coat, duffle coat",
  "n03254189": "dugout",
  "n03254374": "dugout canoe, dugout, pirogue",
  "n03254505": "Duke University",
  "n03254625": "dulciana",
  "n03254737": "dulcimer",
  "n03254862": "dulcimer",
  "n03255030": "dumbbell",
  "n03255167": "dumb bomb, gravity bomb",
  "n03255322": "dumbwaiter, food elevator",
  "n03255488": "dumdum, dumdum bullet",
  "n03255648": "dummy",
  "n03255790": "dump",
  "n03255899": "dumpcart",
  "n03256032": "Dumpster",
  "n03256166": "dump truck, dumper, tipper truck, tipper lorry, tip truck, tipper",
  "n03256472": "Dumpy level",
  "n03256631": "dunce cap, dunce's cap, fool's cap",
  "n03256788": "dune buggy, beach buggy",
  "n03256928": "dungeon",
  "n03257065": "duplex apartment, duplex",
  "n03257210": "duplex house, duplex, semidetached house",
  "n03257343": "duplicate, duplication",
  "n03257586": "duplicator, copier",
  "n03257877": "durables, durable goods, consumer durables",
  "n03258049": "durbar",
  "n03258192": "dust bag, vacuum bag",
  "n03258330": "dustcloth, dustrag, duster",
  "n03258456": "dust cover",
  "n03258577": "dust cover, dust sheet",
  "n03258730": "duster, gaberdine, gabardine, smock, dust coat",
  "n03258905": "dustmop, dust mop, dry mop",
  "n03259009": "dustpan",
  "n03259118": "Dutch door, half door",
  "n03259280": "Dutch oven",
  "n03259401": "Dutch oven",
  "n03259505": "dwelling, home, domicile, abode, habitation, dwelling house",
  "n03260206": "dye-works",
  "n03260293": "dynamite",
  "n03260504": "dynamo",
  "n03260733": "dynamometer, ergometer",
  "n03260849": "Eames chair",
  "n03261019": "earflap, earlap",
  "n03261157": "ear hole",
  "n03261263": "early warning radar",
  "n03261395": "early warning system",
  "n03261603": "earmuff",
  "n03261776": "earphone, earpiece, headphone, phone",
  "n03262072": "earplug",
  "n03262248": "earplug",
  "n03262349": "earring",
  "n03262519": "earthenware",
  "n03262717": "earthwork",
  "n03262809": "easel",
  "n03262932": "easy chair, lounge chair, overstuffed chair",
  "n03263076": "eaves",
  "n03263206": "ecce homo",
  "n03263338": "ecclesiastical attire, ecclesiastical robe",
  "n03263474": "echelon",
  "n03263640": "echinus",
  "n03263758": "echocardiograph",
  "n03263888": "echoencephalograph",
  "n03264021": "echo chamber",
  "n03264136": "edge",
  "n03264542": "edge",
  "n03264906": "edger",
  "n03265032": "edge tool",
  "n03265479": "edging",
  "n03265754": "efficiency apartment",
  "n03265874": "effigy, image, simulacrum",
  "n03266195": "egg-and-dart, egg-and-anchor, egg-and-tongue",
  "n03266371": "eggbeater, eggwhisk",
  "n03266498": "eggcup, egg cup",
  "n03266620": "egg timer",
  "n03266749": "eiderdown, duvet, continental quilt",
  "n03266906": "Eiffel Tower",
  "n03267113": "eight ball",
  "n03267256": "eightpenny nail",
  "n03267340": "eight-spot, eight",
  "n03267468": "ejection seat, ejector seat, capsule",
  "n03267696": "elastic",
  "n03267821": "elastic bandage",
  "n03267972": "elastic device",
  "n03268142": "Elastoplast",
  "n03268311": "elbow",
  "n03268470": "elbow",
  "n03268553": "elbow",
  "n03268645": "elbow pad",
  "n03268790": "electric, electric automobile, electric car",
  "n03268918": "electrical cable",
  "n03269073": "electrical contact",
  "n03269203": "electrical converter",
  "n03269401": "electrical device",
  "n03270165": "electrical system",
  "n03270579": "electrical system, electrical plant",
  "n03270695": "electric bell",
  "n03270854": "electric blanket",
  "n03271030": "electric chair, chair, death chair, hot seat",
  "n03271260": "electric clock",
  "n03271376": "electric-discharge lamp, gas-discharge lamp",
  "n03271574": "electric fan, blower",
  "n03271765": "electric frying pan",
  "n03271865": "electric furnace",
  "n03272010": "electric guitar",
  "n03272125": "electric hammer",
  "n03272239": "electric heater, electric fire",
  "n03272383": "electric lamp",
  "n03272562": "electric locomotive",
  "n03272696": "electric main",
  "n03272810": "electric meter, power meter",
  "n03272940": "electric mixer",
  "n03273061": "electric motor",
  "n03273551": "electric organ, electronic organ, Hammond organ, organ",
  "n03273740": "electric range",
  "n03273913": "electric refrigerator, fridge",
  "n03274137": "electric socket",
  "n03274265": "electric toothbrush",
  "n03274435": "electric typewriter",
  "n03274561": "electro-acoustic transducer",
  "n03274796": "electrode",
  "n03275125": "electrodynamometer",
  "n03275311": "electroencephalograph",
  "n03275451": "electrograph",
  "n03275566": "electrograph",
  "n03275681": "electrolytic, electrolytic capacitor, electrolytic condenser",
  "n03275864": "electrolytic cell",
  "n03276179": "electromagnet",
  "n03276431": "electromagnetic delay line",
  "n03276574": "electromechanical device",
  "n03276696": "electrometer",
  "n03276839": "electromyograph",
  "n03277004": "electron accelerator",
  "n03277149": "electron gun",
  "n03277459": "electronic balance",
  "n03277602": "electronic converter",
  "n03277771": "electronic device",
  "n03278248": "electronic equipment",
  "n03278914": "electronic fetal monitor, electronic foetal monitor, fetal monitor, foetal monitor",
  "n03279153": "electronic instrument, electronic musical instrument",
  "n03279364": "electronic voltmeter",
  "n03279508": "electron microscope",
  "n03279804": "electron multiplier",
  "n03279918": "electrophorus",
  "n03280064": "electroplate",
  "n03280216": "electroscope",
  "n03280394": "electrostatic generator, electrostatic machine, Wimshurst machine, Van de Graaff generator",
  "n03280644": "electrostatic printer",
  "n03280813": "elevated railway, elevated railroad, elevated, el, overhead railway",
  "n03281052": "elevation",
  "n03281145": "elevator, lift",
  "n03281524": "elevator",
  "n03281673": "elevator shaft",
  "n03281821": "ell",
  "n03281935": "elongation, extension",
  "n03282060": "embankment",
  "n03282295": "embassy",
  "n03282401": "embellishment",
  "n03282591": "emblem",
  "n03282933": "embroidery, fancywork",
  "n03283221": "emergency room, ER",
  "n03283413": "emesis basin",
  "n03283519": "emetic, vomit, vomitive, nauseant",
  "n03283717": "Emetrol",
  "n03283827": "emitter",
  "n03283973": "Empire State Building",
  "n03284120": "emplacement",
  "n03284308": "empty",
  "n03284482": "emulsion, photographic emulsion",
  "n03284743": "enamel",
  "n03284886": "enamel",
  "n03284981": "enamelware",
  "n03285106": "enalapril, Vasotec",
  "n03285348": "encainide, Enkaid",
  "n03285578": "encaustic",
  "n03285730": "encephalogram, pneumoencephalogram",
  "n03285912": "enclosure",
  "n03286383": "end, remainder, remnant, oddment",
  "n03286572": "endoscope",
  "n03286878": "endotracheal tube",
  "n03287178": "end product, output",
  "n03287351": "energizer, energiser",
  "n03287459": "enflurane, Ethrane",
  "n03287627": "engagement ring",
  "n03287733": "engine",
  "n03288003": "engine",
  "n03288225": "engine block, cylinder block, block",
  "n03288500": "engineering, engine room",
  "n03288643": "enginery",
  "n03288742": "English horn, cor anglais",
  "n03288886": "English saddle, English cavalry saddle",
  "n03289025": "engraving",
  "n03289268": "engraving",
  "n03289462": "enlargement, blowup, magnification",
  "n03289660": "enlarger",
  "n03289819": "Enovid",
  "n03289985": "ensemble",
  "n03290096": "ensign",
  "n03290195": "entablature",
  "n03290489": "enteric-coated aspirin",
  "n03290653": "entertainment center",
  "n03290771": "entrance, entranceway, entryway, entry, entree",
  "n03291243": "entrant",
  "n03291413": "entrenching tool, trenching spade",
  "n03291551": "entrenchment, intrenchment",
  "n03291741": "envelope",
  "n03291819": "envelope",
  "n03291963": "envelope, gasbag",
  "n03292085": "eolith",
  "n03292210": "epaulet, epaulette",
  "n03292362": "epauliere",
  "n03292475": "epee",
  "n03292603": "epergne",
  "n03292736": "epicyclic train, epicyclic gear train",
  "n03292960": "epidiascope",
  "n03293095": "epilating wax",
  "n03293321": "Epsom salts",
  "n03293471": "equal-area projection, equal-area map projection",
  "n03293741": "equalizer, equaliser",
  "n03293863": "equatorial",
  "n03294048": "equipment",
  "n03294604": "erasable programmable read-only memory, EPROM",
  "n03294833": "eraser",
  "n03295012": "erecting prism",
  "n03295140": "erection",
  "n03295246": "Erlenmeyer flask",
  "n03295357": "erythromycin, Erythrocin, E-Mycin, Ethril, Ilosone, Pediamycin",
  "n03295773": "escalator, moving staircase, moving stairway",
  "n03295928": "escape hatch",
  "n03296081": "escapement",
  "n03296217": "escape wheel",
  "n03296328": "escarpment, escarp, scarp, protective embankment",
  "n03296478": "escutcheon, scutcheon",
  "n03296597": "escutcheon",
  "n03296759": "esmolol, Brevibloc",
  "n03296963": "esophagoscope, oesophagoscope",
  "n03297103": "espadrille",
  "n03297226": "espalier",
  "n03297354": "esplanade",
  "n03297495": "espresso maker",
  "n03297644": "espresso shop",
  "n03297735": "establishment",
  "n03298089": "estaminet",
  "n03298211": "estazolam, ProSom",
  "n03298352": "estradiol patch",
  "n03298565": "estrogen antagonist, tamoxifen",
  "n03298716": "etagere",
  "n03298858": "etamine, etamin",
  "n03299006": "etanercept, Enbrel",
  "n03299261": "etcetera",
  "n03299406": "etching",
  "n03299519": "etching",
  "n03299648": "ethacrynic acid, Edecrin",
  "n03299788": "ethchlorvynol, Placidyl",
  "n03299929": "ether, ethoxyethane, divinyl ether, vinyl ether, diethyl ether, ethyl ether",
  "n03300216": "ethernet",
  "n03300443": "ethernet cable",
  "n03300578": "ethosuximide, Emeside, Zarontin",
  "n03300786": "ethyl chloride",
  "n03300907": "etodolac, Lodine",
  "n03301066": "Eton collar",
  "n03301175": "Eton jacket",
  "n03301291": "etui",
  "n03301389": "eudiometer",
  "n03301568": "euphonium",
  "n03301696": "euphoriant",
  "n03301833": "evaporative cooler",
  "n03301940": "evening bag",
  "n03302030": "Excalibur",
  "n03302121": "excavation",
  "n03302487": "exchange",
  "n03302671": "exercise bike, exercycle",
  "n03302790": "exercise device",
  "n03302938": "exhaust, exhaust system",
  "n03303217": "exhaust fan",
  "n03303333": "exhaust manifold",
  "n03303510": "exhaust pipe",
  "n03303669": "exhaust valve",
  "n03303831": "exhibition hall, exhibition area",
  "n03303965": "exit, issue, outlet, way out",
  "n03304197": "Exocet",
  "n03304323": "expansion bit, expansive bit",
  "n03304465": "expansion bolt",
  "n03304605": "expectorant, expectorator",
  "n03304730": "explosive",
  "n03305135": "explosive compound",
  "n03305300": "explosive detection system, EDS",
  "n03305522": "explosive device",
  "n03305844": "explosive mixture",
  "n03305953": "explosive trace detection, ETD",
  "n03306207": "export, exportation",
  "n03306385": "express, limited",
  "n03306610": "expressway, freeway, motorway, pike, state highway, superhighway, throughway, thruway",
  "n03306869": "extension, telephone extension, extension phone",
  "n03307037": "extension cord",
  "n03307156": "extension ladder",
  "n03307274": "exterior door, outside door",
  "n03307573": "external-combustion engine",
  "n03307792": "external drive",
  "n03307981": "extra, duplicate",
  "n03308152": "extractor",
  "n03308297": "eye",
  "n03308481": "eyebrow pencil",
  "n03308614": "eyecup, eyebath, eye cup",
  "n03308853": "eyelet, eyehole",
  "n03309110": "eyeliner",
  "n03309214": "eye-lotion, eyewash, collyrium",
  "n03309356": "eyepatch, patch",
  "n03309465": "eyepiece, ocular",
  "n03309687": "eyeshadow",
  "n03309808": "fabric, cloth, material, textile",
  "n03313333": "facade, frontage, frontal",
  "n03313456": "face",
  "n03313602": "face",
  "n03313873": "face",
  "n03314028": "face card, picture card, court card",
  "n03314227": "face guard",
  "n03314378": "face mask",
  "n03314608": "faceplate",
  "n03314780": "face powder",
  "n03314884": "face veil",
  "n03315023": "facility, installation",
  "n03315644": "facing, cladding",
  "n03315805": "facing",
  "n03315990": "facing, veneer",
  "n03316105": "facsimile, facsimile machine, fax",
  "n03316274": "facsimile, autotype",
  "n03316406": "factory, mill, manufacturing plant, manufactory",
  "n03316873": "factory ship",
  "n03316988": "factory whistle",
  "n03317136": "fag end",
  "n03317233": "fagot, faggot",
  "n03317340": "fagoting, faggoting",
  "n03317510": "fagot stitch, faggot stitch",
  "n03317673": "Fahrenheit thermometer",
  "n03317788": "faience",
  "n03317889": "faille",
  "n03317990": "fail-safe",
  "n03318136": "fairlead",
  "n03318294": "fairy light",
  "n03318438": "fake, sham, postiche",
  "n03318707": "fake book",
  "n03318865": "falchion",
  "n03318983": "fallboard, fall-board",
  "n03319167": "fallout shelter",
  "n03319296": "false bottom",
  "n03319457": "false face",
  "n03319576": "false teeth",
  "n03319653": "falsie",
  "n03319745": "family room",
  "n03319858": "famotidine, Pepcid",
  "n03320046": "fan",
  "n03320262": "fan belt",
  "n03320421": "fan blade",
  "n03320519": "fancy dress, masquerade, masquerade costume",
  "n03320735": "fancy goods",
  "n03320845": "fanion",
  "n03320959": "fanlight",
  "n03321103": "fanjet, fan-jet, fanjet engine, turbojet, turbojet engine, turbofan, turbofan engine",
  "n03321419": "fanjet, fan-jet, turbofan, turbojet",
  "n03321563": "fanny pack, butt pack",
  "n03321672": "fantail",
  "n03321843": "fan tracery",
  "n03321954": "fan vaulting",
  "n03322099": "farm",
  "n03322570": "farm building",
  "n03322704": "farmer's market, green market, greenmarket",
  "n03322836": "farmhouse",
  "n03322940": "farm machine",
  "n03323096": "farmplace, farm-place, farmstead",
  "n03323211": "farmyard",
  "n03323319": "farthingale",
  "n03323485": "fashion plate",
  "n03323593": "fashion",
  "n03323703": "fastener, fastening, holdfast, fixing",
  "n03324502": "fast lane",
  "n03324629": "fast reactor",
  "n03324814": "fat farm",
  "n03324928": "fatigues",
  "n03325088": "faucet, spigot",
  "n03325288": "fauld",
  "n03325403": "fauteuil",
  "n03325481": "feather bed, featherbed",
  "n03325584": "feather boa, boa",
  "n03325691": "featheredge",
  "n03325769": "feature",
  "n03325941": "fedora, felt hat, homburg, Stetson, trilby",
  "n03326073": "feedback circuit, feedback loop",
  "n03326239": "feeder line",
  "n03326371": "feedlot",
  "n03326475": "fell, felled seam",
  "n03326660": "felloe, felly",
  "n03326795": "felt",
  "n03326948": "felt-tip pen, felt-tipped pen, felt tip, Magic Marker",
  "n03327133": "felucca",
  "n03327234": "fence, fencing",
  "n03327553": "fencing mask, fencer's mask",
  "n03327691": "fencing sword",
  "n03327841": "fender, wing",
  "n03328076": "fender",
  "n03328201": "fender, buffer, cowcatcher, pilot",
  "n03328392": "fenoprofen, fenoprofen calcium, Nalfon",
  "n03328650": "Fentanyl, Sublimaze",
  "n03329058": "Feosol",
  "n03329180": "Fergon",
  "n03329302": "Ferris wheel",
  "n03329536": "ferrule, collet",
  "n03329663": "ferry, ferryboat",
  "n03329880": "fertility drug",
  "n03330002": "ferule",
  "n03330120": "fesse, fess",
  "n03330274": "festoon, festoonery",
  "n03330441": "festoon",
  "n03330665": "festoon",
  "n03330792": "fetoscope, foetoscope",
  "n03330947": "fetter, hobble",
  "n03331077": "fez, tarboosh",
  "n03331244": "fiber, fibre, vulcanized fiber",
  "n03331390": "fiberboard, fibreboard, particle board",
  "n03331599": "fiber optic cable, fibre optic cable",
  "n03331820": "fiber-optic transmission system, fibre-optic transmission system, FOTS",
  "n03332005": "fiberscope",
  "n03332173": "fichu",
  "n03332271": "fiddlestick, violin bow",
  "n03332393": "field artillery, field gun",
  "n03332591": "field coil, field winding",
  "n03332784": "field-effect transistor, FET",
  "n03332989": "field-emission microscope",
  "n03333129": "field glass, glass, spyglass",
  "n03333252": "field hockey ball",
  "n03333349": "field hospital",
  "n03333480": "field house",
  "n03333610": "field house, sports arena",
  "n03333711": "field lens",
  "n03333851": "field magnet",
  "n03334017": "field-sequential color television, field-sequential color TV, field-sequential color television system, field-sequential color TV system",
  "n03334291": "field tent",
  "n03334382": "fieldwork",
  "n03334492": "fife",
  "n03334667": "fife rail",
  "n03334775": "fifth wheel",
  "n03334912": "fifth wheel, spare",
  "n03335030": "fighter, fighter aircraft, attack aircraft",
  "n03335333": "fighting chair",
  "n03335461": "fig leaf",
  "n03335600": "figure",
  "n03335846": "figure eight, figure of eight",
  "n03336070": "figurehead",
  "n03336168": "figure loom, figured-fabric loom",
  "n03336282": "figure skate",
  "n03336459": "figurine, statuette",
  "n03336575": "filament",
  "n03336742": "filature",
  "n03336839": "file",
  "n03337140": "file, file cabinet, filing cabinet",
  "n03337383": "file folder",
  "n03337494": "file server",
  "n03337727": "filet",
  "n03337822": "filigree, filagree, fillagree",
  "n03338009": "filler",
  "n03338143": "fillet, stopping",
  "n03338287": "filling",
  "n03338648": "film",
  "n03338821": "film, photographic film",
  "n03339296": "film, plastic film",
  "n03339529": "film advance",
  "n03339643": "filter",
  "n03340009": "filter",
  "n03340183": "filter bed",
  "n03340306": "filter tip",
  "n03340463": "filter-tipped cigarette",
  "n03340581": "fin",
  "n03340723": "finder, viewfinder, view finder",
  "n03340923": "finery",
  "n03341035": "fine-tooth comb, fine-toothed comb",
  "n03341153": "finger",
  "n03341297": "fingerboard",
  "n03341606": "finger bowl",
  "n03341707": "finger hole",
  "n03341850": "finger hole",
  "n03342015": "finger paint, fingerpaint",
  "n03342127": "finger-painting",
  "n03342262": "finger plate, escutcheon, scutcheon",
  "n03342432": "fingerstall, cot",
  "n03342529": "finial",
  "n03342657": "finish coat, finishing coat",
  "n03342863": "finish coat, finishing coat",
  "n03342961": "finisher",
  "n03343047": "fin keel",
  "n03343234": "fipple",
  "n03343354": "fipple flute, fipple pipe, recorder, vertical flute",
  "n03343560": "fire",
  "n03343737": "fire alarm, smoke alarm",
  "n03343853": "firearm, piece, small-arm",
  "n03344205": "firebase",
  "n03344305": "fire bell",
  "n03344393": "fireboat",
  "n03344509": "firebox",
  "n03344642": "firebrick",
  "n03344784": "fire control radar",
  "n03344935": "fire control system",
  "n03345115": "firecracker, cracker, banger",
  "n03345362": "fire door",
  "n03345487": "fire engine, fire truck",
  "n03345658": "fire escape, emergency exit",
  "n03345837": "fire extinguisher, extinguisher, asphyxiator",
  "n03346004": "fire hose",
  "n03346135": "fire iron",
  "n03346289": "fireman's ax, fireman's axe",
  "n03346455": "fireplace, hearth, open fireplace",
  "n03346898": "fireplug, fire hydrant, plug",
  "n03347037": "fire screen, fireguard",
  "n03347191": "fire ship",
  "n03347338": "fire station, firehouse",
  "n03347472": "fire tongs, coal tongs",
  "n03347617": "fire tower",
  "n03347731": "firetrap",
  "n03347855": "fire trench",
  "n03347980": "firewall",
  "n03348142": "firewall",
  "n03348454": "firework, pyrotechnic",
  "n03348868": "firing chamber, gun chamber",
  "n03349020": "firing pin",
  "n03349150": "firing range, target range",
  "n03349296": "firkin",
  "n03349367": "firmer chisel",
  "n03349469": "first-aid kit",
  "n03349599": "first-aid station",
  "n03349771": "first base",
  "n03349892": "first class",
  "n03350011": "first gear, first, low gear, low",
  "n03350204": "fishbowl, fish bowl, goldfish bowl",
  "n03350352": "fisherman's bend",
  "n03350456": "fisherman's knot, true lover's knot, truelove knot",
  "n03350602": "fisherman's lure, fish lure",
  "n03350880": "fishery, piscary",
  "n03351036": "fish farm",
  "n03351151": "fishhook",
  "n03351262": "fishing boat, fishing smack, fishing vessel",
  "n03351434": "fishing gear, tackle, fishing tackle, fishing rig, rig",
  "n03351768": "fishing line",
  "n03351979": "fishing rod, fishing pole",
  "n03352232": "fish joint",
  "n03352366": "fish knife",
  "n03352484": "fish ladder",
  "n03352628": "fishnet, fishing net",
  "n03352853": "fishplate",
  "n03352961": "fish slice",
  "n03353147": "fishtail bit, blade bit",
  "n03353281": "fitment",
  "n03353467": "fitted sheet, contour sheet",
  "n03353616": "fitting",
  "n03353783": "five-spot, five",
  "n03353951": "fixative",
  "n03354082": "fixed-combination drug",
  "n03354207": "fixer-upper",
  "n03354350": "fixings, trimmings",
  "n03354613": "fixture",
  "n03354816": "fizgig",
  "n03354903": "flag",
  "n03355339": "flag, flagstone",
  "n03355468": "flageolet, treble recorder, shepherd's pipe",
  "n03355641": "flagging",
  "n03355768": "flagon",
  "n03355925": "flagpole, flagstaff",
  "n03356038": "flagship",
  "n03356158": "flagship",
  "n03356279": "flail",
  "n03356446": "flambeau",
  "n03356559": "flamethrower",
  "n03356670": "Flaminian Way",
  "n03356858": "flange, rim",
  "n03356982": "flannel",
  "n03357081": "flannel, gabardine, tweed, white",
  "n03357267": "flannelette",
  "n03357376": "flap",
  "n03357716": "flap, flaps",
  "n03357893": "flare",
  "n03358046": "flare path",
  "n03358172": "flash, photoflash, flash lamp, flashgun, flashbulb, flash bulb",
  "n03358380": "flash",
  "n03358593": "flashboard, flashboarding",
  "n03358726": "flash camera",
  "n03358841": "flasher",
  "n03359008": "flashing",
  "n03359137": "flashlight, torch",
  "n03359285": "flashlight battery",
  "n03359436": "flash memory",
  "n03359566": "flask",
  "n03359755": "flat",
  "n03359950": "flat, flat tire",
  "n03360038": "flat",
  "n03360133": "flat arch, straight arch",
  "n03360300": "flatbed",
  "n03360431": "flatbed press, cylinder press",
  "n03360622": "flat bench",
  "n03360731": "flatcar, flatbed, flat",
  "n03360845": "flat coat, ground, primer, priming, primer coat, priming coat, undercoat",
  "n03361109": "flat file",
  "n03361194": "flatiron",
  "n03361297": "flatlet",
  "n03361380": "flat panel display, FPD",
  "n03361550": "flats",
  "n03361683": "flat tip screwdriver",
  "n03361837": "flatware, silver",
  "n03361945": "flatware",
  "n03362119": "flatwork, flat wash",
  "n03362223": "fleabag",
  "n03362293": "fleapit",
  "n03362393": "flecainide, Tambocor",
  "n03362639": "fleece",
  "n03362771": "fleet ballistic missile submarine",
  "n03362890": "fleur-de-lis, fleur-de-lys",
  "n03363059": "flight, flight of stairs, flight of steps",
  "n03363216": "flight deck, landing deck",
  "n03363363": "flight simulator, trainer",
  "n03363549": "flintlock",
  "n03363749": "flintlock, firelock",
  "n03363887": "flip-flop",
  "n03364008": "flip-flop, thong",
  "n03364156": "flipper, fin",
  "n03364340": "float",
  "n03364599": "float, plasterer's float",
  "n03364775": "float",
  "n03364937": "floating dock, floating dry dock",
  "n03365078": "floating mine, marine mine",
  "n03365231": "floatplane, pontoon plane",
  "n03365374": "flood, floodlight, flood lamp, photoflood",
  "n03365592": "floor, flooring",
  "n03365991": "floor, level, storey, story",
  "n03366301": "floor, trading floor",
  "n03366464": "floor",
  "n03366630": "floorboard, floor board",
  "n03366721": "floorboard",
  "n03366823": "floor cover, floor covering",
  "n03366974": "floor joist",
  "n03367059": "floor lamp",
  "n03367147": "floor plan",
  "n03367321": "flophouse, dosshouse",
  "n03367410": "florist, florist shop, flower store",
  "n03367545": "floss",
  "n03367663": "flotilla",
  "n03367740": "flotilla",
  "n03367875": "flotsam, jetsam",
  "n03367969": "flour bin",
  "n03368048": "flour mill",
  "n03368141": "flower arrangement, floral arrangement",
  "n03368352": "flowerbed, flower bed, bed of flowers",
  "n03368508": "flower chain",
  "n03368637": "flower garden",
  "n03368750": "floxuridine",
  "n03368878": "flue",
  "n03369011": "flue pipe, flue, labial pipe",
  "n03369178": "flue stop",
  "n03369276": "flugelhorn, fluegelhorn",
  "n03369407": "fluid drive",
  "n03369512": "fluid flywheel",
  "n03369640": "fluke, flue",
  "n03369767": "fluke",
  "n03369866": "flume",
  "n03370020": "flunitrazepan, Rohypnol",
  "n03370260": "fluorescent, fluorescent fixture",
  "n03370387": "fluorescent lamp",
  "n03370646": "fluoroscope, roentgenoscope",
  "n03370821": "fluorouracil",
  "n03370927": "fluoxetine, fluoxetine hydrocholoride, Prozac, Sarafem",
  "n03371258": "fluphenazine",
  "n03371363": "flurazepam, flurazepam hydrochloride, Dalmane",
  "n03371532": "flurbiprofen, Ansaid",
  "n03371728": "flushless toilet",
  "n03371875": "flush toilet, lavatory",
  "n03372029": "flute, transverse flute",
  "n03372355": "flute, fluting",
  "n03372549": "flute, flute glass, champagne flute",
  "n03372656": "fluvastatin, Lescol",
  "n03372822": "flux applicator",
  "n03372933": "fluxmeter",
  "n03373060": "fly, fly front",
  "n03373237": "fly",
  "n03373415": "fly gallery, fly floor",
  "n03373611": "flying boat",
  "n03373752": "flying bridge, flybridge, fly bridge, monkey bridge",
  "n03373943": "flying buttress, arc-boutant",
  "n03374102": "flying carpet",
  "n03374282": "flying jib",
  "n03374372": "fly rod",
  "n03374473": "fly tent",
  "n03374570": "flytrap",
  "n03374649": "flywheel",
  "n03374838": "fob, watch chain, watch guard",
  "n03374978": "fob",
  "n03375070": "fob, watch pocket",
  "n03375171": "foghorn",
  "n03375329": "foglamp",
  "n03375443": "foible",
  "n03375575": "foil",
  "n03375694": "foil",
  "n03375956": "foil, transparency",
  "n03376159": "fold, sheepfold, sheep pen, sheepcote",
  "n03376279": "folder",
  "n03376438": "folderal, falderol, frill, gimcrackery, gimcrack, nonsense, trumpery",
  "n03376595": "folding chair",
  "n03376771": "folding door, accordion door",
  "n03376938": "folding saw",
  "n03377077": "foliation, foliage",
  "n03377245": "folio",
  "n03377451": "folk art",
  "n03377582": "follow-up, followup",
  "n03377845": "food additive, artificial additive",
  "n03378005": "food court",
  "n03378174": "food processor",
  "n03378342": "food hamper",
  "n03378442": "foot",
  "n03378593": "footage",
  "n03378765": "football",
  "n03378915": "football field, gridiron",
  "n03379051": "football helmet",
  "n03379204": "football stadium",
  "n03379343": "footbath",
  "n03379461": "footboard",
  "n03379592": "footboard",
  "n03379719": "foot brake",
  "n03379828": "footbridge, overcrossing, pedestrian bridge",
  "n03379989": "foothold, footing",
  "n03380134": "footlights",
  "n03380301": "footlocker, locker",
  "n03380461": "footplate",
  "n03380647": "foot rule",
  "n03380724": "footstool, footrest, ottoman, tuffet",
  "n03380867": "footwear, footgear",
  "n03381126": "footwear",
  "n03381231": "forceps",
  "n03381450": "force pump",
  "n03381565": "fore-and-after",
  "n03381664": "fore-and-aft rig",
  "n03381776": "fore-and-aft sail",
  "n03382104": "forecastle, fo'c'sle",
  "n03382292": "forecourt",
  "n03382413": "foredeck",
  "n03382533": "fore edge, foredge",
  "n03382708": "foreground",
  "n03382856": "foremast",
  "n03382969": "fore plane",
  "n03383099": "foresail",
  "n03383211": "forestay",
  "n03383378": "foretop",
  "n03383468": "fore-topmast",
  "n03383562": "fore-topsail",
  "n03383646": "forge, smithy",
  "n03383821": "forge",
  "n03383948": "fork",
  "n03384167": "fork",
  "n03384352": "forklift",
  "n03384535": "form",
  "n03384706": "formal garden",
  "n03384891": "formalwear, eveningwear, evening dress, evening clothes",
  "n03385117": "formation",
  "n03385295": "Formica",
  "n03385420": "forte",
  "n03385557": "fortification, munition",
  "n03386011": "fortress, fort",
  "n03386343": "forty-five",
  "n03386420": "forum, assembly, meeting place",
  "n03386544": "Foucault pendulum",
  "n03386726": "foulard",
  "n03386870": "foul-weather gear",
  "n03387016": "foundation, base, fundament, foot, groundwork, substructure, understructure",
  "n03387323": "foundation garment, foundation",
  "n03387506": "foundation stone",
  "n03387653": "foundry, metalworks",
  "n03387815": "fountain, fount",
  "n03387926": "fountain, jet",
  "n03388043": "fountain",
  "n03388183": "fountain pen",
  "n03388323": "four-in-hand",
  "n03388466": "fourpenny nail",
  "n03388549": "four-poster",
  "n03388711": "four-pounder",
  "n03388826": "four-spot, four",
  "n03388990": "four-stroke engine, four-stroke internal-combustion engine",
  "n03389353": "four-tailed bandage",
  "n03389611": "four-wheel drive, 4WD",
  "n03389761": "four-wheel drive, 4WD",
  "n03389889": "four-wheeler",
  "n03389983": "fowling piece",
  "n03390075": "foxhole, fox hole",
  "n03390207": "fraction",
  "n03390327": "fragmentation bomb, antipersonnel bomb, anti-personnel bomb, daisy cutter",
  "n03390673": "frail",
  "n03390786": "fraise",
  "n03390886": "fraise",
  "n03390983": "frame, framing",
  "n03391301": "frame",
  "n03391464": "frame",
  "n03391613": "frame buffer",
  "n03391770": "framework",
  "n03392648": "Francis turbine",
  "n03392741": "franking machine",
  "n03392908": "freeboard deck",
  "n03393017": "free house",
  "n03393199": "free-reed",
  "n03393324": "free-reed instrument",
  "n03393534": "free throw lane",
  "n03393761": "freewheel",
  "n03393912": "freight car",
  "n03394149": "freight elevator, service elevator",
  "n03394272": "freight liner, liner train",
  "n03394480": "freight train, rattler",
  "n03394649": "French door",
  "n03394816": "French heel",
  "n03394916": "French horn, horn",
  "n03395095": "French knot",
  "n03395256": "French polish, French polish shellac",
  "n03395401": "French roof",
  "n03395514": "French window",
  "n03395630": "fresco",
  "n03395745": "freshener",
  "n03395859": "Fresnel lens",
  "n03396074": "fret",
  "n03396311": "fret, Greek fret, Greek key, key pattern",
  "n03396580": "friary",
  "n03396654": "friction clutch",
  "n03396841": "friction tape, insulating tape",
  "n03396997": "frieze",
  "n03397087": "frieze",
  "n03397266": "frigate",
  "n03397412": "frigate",
  "n03397532": "frill, flounce, ruffle, furbelow",
  "n03397762": "fringe",
  "n03397947": "Frisbee",
  "n03398153": "frock",
  "n03398228": "frock coat",
  "n03398382": "frog",
  "n03398467": "front",
  "n03398775": "frontage road, service road",
  "n03398950": "frontal",
  "n03399047": "front bench",
  "n03399240": "front door, front entrance",
  "n03399391": "frontispiece",
  "n03399500": "frontispiece",
  "n03399579": "frontlet, frontal",
  "n03399677": "front porch",
  "n03399761": "front projector",
  "n03399852": "front yard",
  "n03399971": "fruit machine",
  "n03400231": "frying pan, frypan, skillet",
  "n03400389": "fuel-air explosive, FAE",
  "n03400798": "fuel cell",
  "n03400972": "fuel filter",
  "n03401129": "fuel gauge, fuel indicator",
  "n03401279": "fuel injection, fuel injection system",
  "n03401500": "fuel line, gas line, petrol line",
  "n03401721": "fuel system",
  "n03402080": "fulcrum",
  "n03402188": "full-dress uniform",
  "n03402369": "full metal jacket",
  "n03402511": "full skirt",
  "n03402621": "full-wave rectifier",
  "n03402785": "fumigator",
  "n03402941": "funeral home, funeral parlor, funeral parlour, funeral chapel, funeral church, funeral-residence",
  "n03403169": "fungible",
  "n03403325": "funk hole",
  "n03403431": "funnel",
  "n03403643": "funnel",
  "n03403873": "funnel web",
  "n03404012": "funny wagon",
  "n03404149": "fur",
  "n03404251": "fur coat",
  "n03404360": "fur hat",
  "n03404449": "furnace",
  "n03404900": "furnace lining, refractory",
  "n03405111": "furnace room",
  "n03405265": "furnishing",
  "n03405595": "furnishing, trappings",
  "n03405725": "furniture, piece of furniture, article of furniture",
  "n03406597": "furosemide, Lasix",
  "n03406759": "fur-piece",
  "n03406838": "furring strip, furring",
  "n03406966": "furrow",
  "n03407122": "fuse, fuze, fusee, fuzee, primer, priming",
  "n03407369": "fuse, electrical fuse, safety fuse",
  "n03407616": "fusee, fuzee",
  "n03407744": "fusee, fuzee",
  "n03407865": "fusee drive, fusee",
  "n03408054": "fuselage",
  "n03408264": "fusil",
  "n03408340": "fustian",
  "n03408444": "futon",
  "n03408621": "futtock shroud",
  "n03408721": "future",
  "n03408918": "futures exchange, futures market, forward market",
  "n03409070": "gabapentin, Neurontin",
  "n03409297": "gabardine",
  "n03409393": "gable, gable end, gable wall",
  "n03409591": "gable roof, saddle roof, saddleback, saddleback roof",
  "n03409806": "gaddi",
  "n03409920": "gadgetry",
  "n03410022": "gaff",
  "n03410147": "gaff",
  "n03410303": "gaff",
  "n03410423": "gaffsail, gaff-headed sail",
  "n03410571": "gaff topsail, fore-and-aft topsail",
  "n03410740": "gag, muzzle",
  "n03410938": "gaiter",
  "n03411079": "gaiter",
  "n03411208": "Galilean telescope",
  "n03411339": "galleon",
  "n03411544": "gallery",
  "n03411762": "gallery",
  "n03411927": "gallery",
  "n03412058": "gallery, art gallery, picture gallery",
  "n03412220": "galley, ship's galley, caboose, cookhouse",
  "n03412387": "galley",
  "n03412511": "galley",
  "n03412674": "galley",
  "n03412906": "gallows",
  "n03413124": "gallows tree, gallows-tree, gibbet, gallous",
  "n03413264": "galvanometer",
  "n03413428": "gambling house, gambling den, gambling hell, gaming house",
  "n03413684": "gambrel, gambrel roof",
  "n03413828": "game",
  "n03414029": "gamebag",
  "n03414162": "game equipment",
  "n03414574": "gaming card",
  "n03414676": "gaming table",
  "n03414814": "gamma hydroxybutyrate, GHB",
  "n03415083": "gamma-interferon",
  "n03415252": "gamp, brolly",
  "n03415363": "gang",
  "n03415486": "gangplank, gangboard, gangway",
  "n03415626": "gangsaw",
  "n03415749": "gangway",
  "n03415868": "gantlet",
  "n03416094": "gantry, gauntry",
  "n03416329": "gap, crack",
  "n03416489": "garage",
  "n03416640": "garage, service department",
  "n03416775": "Garand rifle, Garand, M-1, M-1 rifle",
  "n03416900": "garbage",
  "n03417042": "garbage truck, dustcart",
  "n03417202": "garboard, garboard plank, garboard strake",
  "n03417345": "garden",
  "n03417749": "garden",
  "n03417871": "garden hose",
  "n03417970": "garden rake",
  "n03418052": "garden roller",
  "n03418158": "garden spade",
  "n03418242": "garden tool, lawn tool",
  "n03418402": "garden trowel",
  "n03418488": "gargoyle",
  "n03418618": "gargoyle",
  "n03418749": "garibaldi",
  "n03418915": "garlic press",
  "n03419014": "garment",
  "n03420345": "garment bag",
  "n03420440": "garnish",
  "n03420559": "garrison, fort",
  "n03420801": "garrison cap, overseas cap",
  "n03420935": "garrote, garotte, garrotte, iron collar",
  "n03421117": "garter, supporter",
  "n03421324": "garter belt, suspender belt",
  "n03421485": "garter stitch",
  "n03421669": "gas guzzler",
  "n03421768": "gas shell",
  "n03421960": "gas bracket",
  "n03422072": "gas burner, gas jet",
  "n03422288": "gas chamber, death chamber",
  "n03422484": "gas-cooled reactor",
  "n03422589": "gas-discharge tube",
  "n03422771": "gas engine",
  "n03422934": "gas fitting",
  "n03423099": "gas fixture",
  "n03423224": "gas furnace",
  "n03423306": "gas gun",
  "n03423387": "gas heat",
  "n03423479": "gas heater",
  "n03423568": "gas holder, gasometer",
  "n03423719": "gasket",
  "n03423877": "gas lamp",
  "n03423966": "gas line",
  "n03424103": "gas main",
  "n03424204": "gas maser",
  "n03424325": "gasmask, respirator, gas helmet",
  "n03424489": "gas meter, gasometer",
  "n03424630": "gasoline engine, petrol engine",
  "n03424862": "gasoline gauge, gasoline gage, gas gauge, gas gage, petrol gauge, petrol gage",
  "n03425092": "gasoline station, gas station, filling station, petrol station",
  "n03425241": "gas oven",
  "n03425325": "gas oven",
  "n03425413": "gas pump, gasoline pump, petrol pump, island dispenser",
  "n03425595": "gas range, gas stove, gas cooker",
  "n03425769": "gas ring",
  "n03425956": "gas system",
  "n03426134": "gas tank, gasoline tank, petrol tank",
  "n03426285": "gas thermometer, air thermometer",
  "n03426462": "gastroscope",
  "n03426574": "gas turbine",
  "n03426871": "gas-turbine ship",
  "n03426984": "gas well",
  "n03427084": "gasworks",
  "n03427202": "gat, rod",
  "n03427296": "gate",
  "n03427656": "gate, logic gate",
  "n03427909": "gate",
  "n03428090": "gatehouse",
  "n03428226": "gateleg table",
  "n03428349": "gatepost",
  "n03428477": "gateway",
  "n03428571": "gateway drug",
  "n03428805": "gather, gathering",
  "n03429003": "gathered skirt",
  "n03429137": "Gatling gun",
  "n03429288": "gauge, gage",
  "n03429682": "gauntlet, gantlet",
  "n03429771": "gauntlet, gantlet, metal glove",
  "n03429914": "gauze, netting, veiling",
  "n03430091": "gauze, gauze bandage",
  "n03430313": "gavel",
  "n03430418": "gazebo, summerhouse",
  "n03430551": "gear, gear wheel, geared wheel, cogwheel",
  "n03430959": "gear, paraphernalia, appurtenance",
  "n03431243": "gear, gear mechanism",
  "n03431570": "gearbox, gear box, gear case",
  "n03431745": "gearing, gear, geartrain, power train, train",
  "n03432061": "gearset",
  "n03432129": "gearshift, gearstick, shifter, gear lever",
  "n03432360": "Geiger counter, Geiger-Muller counter",
  "n03432509": "Geiger tube, Geiger-Muller tube",
  "n03432668": "gelatin, gel",
  "n03432796": "gelignite, gelly",
  "n03432972": "gem, treasure",
  "n03433079": "gemfibrozil, Lopid",
  "n03433247": "gene chip, DNA chip",
  "n03433434": "general anesthetic, general anaesthetic",
  "n03433637": "general-purpose bomb, GP bomb",
  "n03433877": "generator",
  "n03434188": "generator",
  "n03434285": "generator",
  "n03434428": "generic",
  "n03434549": "generic drug",
  "n03434830": "Geneva gown",
  "n03434943": "genre",
  "n03435290": "genre painting",
  "n03435382": "gentamicin, Garamycin",
  "n03435593": "geodesic dome",
  "n03435743": "georgette",
  "n03435825": "George Washington Bridge",
  "n03435991": "gharry",
  "n03436075": "ghat",
  "n03436182": "ghetto blaster, boom box",
  "n03436290": "ghillie, gillie",
  "n03436417": "gift shop, novelty shop",
  "n03436549": "gift wrapping",
  "n03436656": "gig",
  "n03436772": "gig",
  "n03436891": "gig",
  "n03436990": "gig",
  "n03437184": "gildhall",
  "n03437295": "gill net",
  "n03437430": "gilt, gilding",
  "n03437581": "gimbal",
  "n03437741": "gingham",
  "n03437829": "girandole, girandola",
  "n03437941": "girder",
  "n03438071": "girdle, cincture, sash, waistband, waistcloth",
  "n03438257": "glass, drinking glass",
  "n03438661": "glass",
  "n03438780": "glass cutter",
  "n03438863": "glasses case",
  "n03438952": "glass eye",
  "n03439064": "glassware, glasswork",
  "n03439260": "glassworks",
  "n03439348": "glebe house",
  "n03439491": "Glen Canyon Dam",
  "n03439631": "Glengarry",
  "n03439814": "glider, sailplane",
  "n03440024": "glipizide, Glucotrol",
  "n03440216": "Global Positioning System, GPS",
  "n03440512": "globe",
  "n03440682": "glockenspiel, orchestral bells",
  "n03440876": "glory hole, lazaretto",
  "n03441009": "glossy",
  "n03441112": "glove",
  "n03441345": "glove compartment",
  "n03441465": "glow lamp",
  "n03441582": "glow tube",
  "n03441778": "glutethimide, Doriden",
  "n03441930": "glyburide, DiaBeta, Micronase",
  "n03442164": "glyph",
  "n03442288": "glyptic art, glyptography",
  "n03442487": "glyptics, lithoglyptics",
  "n03442597": "gnomon",
  "n03442756": "goal",
  "n03443005": "goalmouth",
  "n03443149": "goalpost",
  "n03443371": "goblet",
  "n03443461": "go board",
  "n03443543": "godown",
  "n03443669": "goffer, gauffer",
  "n03443775": "goffer, gauffer, goffering iron, gauffering iron",
  "n03443912": "goggles",
  "n03444034": "go-kart",
  "n03444169": "Golconda",
  "n03444268": "goldbrick",
  "n03444376": "golden calf",
  "n03444601": "Golden Gate Bridge",
  "n03444728": "gold foil",
  "n03444838": "gold leaf",
  "n03444942": "gold medal",
  "n03445120": "goldmine, gold mine",
  "n03445217": "goldmine, gold mine",
  "n03445326": "gold plate",
  "n03445472": "gold plate",
  "n03445617": "golf bag",
  "n03445777": "golf ball",
  "n03445924": "golfcart, golf cart",
  "n03446070": "golf club, golf-club, club",
  "n03446268": "golf-club head, club head, club-head, clubhead",
  "n03446528": "golf course, links course",
  "n03446832": "golf equipment",
  "n03447075": "golf glove",
  "n03447224": "golf range, driving range",
  "n03447358": "golliwog, golliwogg",
  "n03447447": "gondola",
  "n03447593": "gondola car, gondola",
  "n03447721": "gong, tam-tam",
  "n03447894": "goniometer",
  "n03448031": "Gordian knot",
  "n03448253": "gore, panel",
  "n03448491": "gorgerin, necking",
  "n03448590": "gorget",
  "n03448696": "gossamer",
  "n03448814": "Gota Canal",
  "n03448956": "Gothic arch",
  "n03449103": "gouache",
  "n03449217": "gouache",
  "n03449309": "gouge",
  "n03449451": "gourd, calabash",
  "n03449564": "government building",
  "n03449858": "government office",
  "n03450018": "governor, regulator",
  "n03450230": "gown",
  "n03450516": "gown, robe",
  "n03450734": "gown, surgical gown, scrubs",
  "n03450881": "grab",
  "n03450974": "grab bag",
  "n03451120": "grab bar",
  "n03451253": "grace cup",
  "n03451365": "grade separation",
  "n03451473": "graduate",
  "n03451711": "graduated cylinder",
  "n03451798": "graffito, graffiti",
  "n03451909": "grail, Holy Grail, Sangraal",
  "n03452055": "gramicidin",
  "n03452267": "gramophone, acoustic gramophone",
  "n03452449": "granary, garner",
  "n03452594": "grandfather clock, longcase clock",
  "n03452741": "grand piano, grand",
  "n03452953": "grandstand, covered stand",
  "n03453162": "grange",
  "n03453231": "graniteware",
  "n03453320": "granny knot, granny",
  "n03453443": "grape arbor, grape arbour",
  "n03453547": "grapeshot, grape",
  "n03453696": "graphic, computer graphic",
  "n03453809": "graphic art",
  "n03453985": "graphics",
  "n03454110": "grapnel, grapnel anchor",
  "n03454211": "grapnel, grapple, grappler, grappling hook, grappling iron",
  "n03454442": "grass skirt",
  "n03454536": "grate, grating",
  "n03454707": "grate, grating",
  "n03454885": "grater",
  "n03455033": "grave, tomb",
  "n03455279": "gravel pit",
  "n03455355": "graver, graving tool, pointel, pointrel",
  "n03455488": "gravestone, headstone, tombstone",
  "n03455642": "gravimeter, gravity meter",
  "n03455802": "gravure, photogravure, heliogravure",
  "n03455923": "gravure",
  "n03456024": "gravy boat, gravy holder, sauceboat, boat",
  "n03456186": "grey, gray",
  "n03456299": "grease-gun, gun",
  "n03456447": "greasepaint",
  "n03456548": "greasy spoon",
  "n03456665": "greatcoat, overcoat, topcoat",
  "n03456854": "Greater New Orleans Bridge",
  "n03457008": "great hall",
  "n03457184": "great seal",
  "n03457332": "Great Seal of the United States",
  "n03457451": "greave, jambeau",
  "n03457578": "Greek cross",
  "n03457686": "greengrocery",
  "n03457793": "greengrocery",
  "n03457902": "greenhouse, nursery, glasshouse",
  "n03458128": "greenroom",
  "n03458271": "grenade",
  "n03458422": "grid, gridiron",
  "n03458552": "grid, control grid",
  "n03458753": "grid, storage-battery grid",
  "n03458961": "reef",
  "n03459159": "reference grid",
  "n03459328": "griddle",
  "n03459498": "grigri, gres-gris, greegree",
  "n03459591": "grill, grille, grillwork",
  "n03459775": "grille, radiator grille",
  "n03459914": "grillroom, grill",
  "n03460040": "grinder",
  "n03460147": "grinding wheel, emery wheel",
  "n03460297": "grindstone",
  "n03460455": "gripsack",
  "n03460526": "grisaille",
  "n03460674": "griseofulvin, Fulvicin",
  "n03460899": "gristmill",
  "n03461037": "grizzle",
  "n03461119": "grocery, foodstuff",
  "n03461288": "grocery bag",
  "n03461385": "grocery store, grocery, food market, market",
  "n03461651": "grogram",
  "n03461783": "groin",
  "n03461882": "groined vault",
  "n03461988": "groover",
  "n03462110": "grosgrain",
  "n03462211": "gros point",
  "n03462315": "gros point",
  "n03462441": "grotesque",
  "n03462594": "ground",
  "n03462747": "ground, earth",
  "n03462972": "ground bait",
  "n03463073": "ground cable",
  "n03463185": "ground control",
  "n03463381": "ground floor, first floor, ground level",
  "n03463561": "ground plan",
  "n03463666": "groundsheet, ground cloth",
  "n03463832": "grove, woodlet, orchard, plantation",
  "n03464053": "G-string, thong",
  "n03464266": "guanabenz, Wytensin",
  "n03464467": "guard, safety, safety device",
  "n03464628": "guard boat",
  "n03464757": "guardhouse",
  "n03464952": "guardroom",
  "n03465040": "guardroom",
  "n03465151": "guard ship",
  "n03465320": "guard's van",
  "n03465426": "gueridon",
  "n03465500": "Guarnerius",
  "n03465605": "guesthouse",
  "n03465718": "guestroom",
  "n03465818": "guidance system, guidance device",
  "n03466027": "guide",
  "n03466162": "guided missile",
  "n03466493": "guided missile cruiser",
  "n03466600": "guided missile frigate",
  "n03466726": "guide rope",
  "n03466839": "guildhall",
  "n03466947": "guilloche",
  "n03467068": "guillotine",
  "n03467254": "guimpe",
  "n03467380": "guimpe",
  "n03467517": "guitar",
  "n03467796": "guitar pick",
  "n03467887": "gulag",
  "n03467984": "gun",
  "n03468570": "gunboat",
  "n03468696": "gun carriage",
  "n03468821": "gun case",
  "n03468900": "gun deck",
  "n03469031": "gun emplacement, weapons emplacement",
  "n03469175": "gun enclosure, gun turret, turret",
  "n03469346": "gunflint",
  "n03469493": "gunlock, firing mechanism",
  "n03469687": "gun muzzle, muzzle",
  "n03469832": "gunnery",
  "n03469903": "gunnysack, gunny sack, burlap bag",
  "n03470005": "gun pendulum",
  "n03470222": "gun room",
  "n03470387": "gunsight, gun-sight",
  "n03470629": "gun trigger, trigger",
  "n03470802": "gunwale, gunnel, gun rest",
  "n03470948": "gurney",
  "n03471030": "gusher",
  "n03471190": "gusset, inset",
  "n03471347": "gusset, gusset plate",
  "n03471473": "gutter, trough",
  "n03471685": "gutter",
  "n03471779": "guy, guy cable, guy wire, guy rope",
  "n03471974": "Guy",
  "n03472112": "gymnasium, gym",
  "n03472232": "gymnastic apparatus, exerciser",
  "n03472535": "gym shoe, sneaker, tennis shoe",
  "n03472672": "gym suit",
  "n03472796": "gymslip",
  "n03472937": "gypsy cab",
  "n03473078": "gyrocompass",
  "n03473227": "gyroscope, gyro",
  "n03473465": "gyrostabilizer, gyrostabiliser",
  "n03473704": "haberdashery, men's furnishings",
  "n03473817": "habergeon",
  "n03473966": "habit",
  "n03474167": "habit, riding habit",
  "n03474352": "hacienda",
  "n03474467": "hack",
  "n03474635": "hackney, hackney carriage, hackney coach",
  "n03474779": "hacksaw, hack saw, metal saw",
  "n03474896": "haft, helve",
  "n03475118": "Hagia Sophia, Hagia Sofia, Santa Sophia, Santa Sofia",
  "n03475421": "haik, haick",
  "n03475581": "hairbrush",
  "n03475674": "haircloth, hair",
  "n03475823": "hairdressing, hair tonic, hair oil, hair grease",
  "n03475961": "hairnet",
  "n03476083": "hairpiece, false hair, postiche",
  "n03476313": "hairpin",
  "n03476458": "hairpin bend",
  "n03476542": "hair shirt",
  "n03476684": "hair slide",
  "n03476843": "hair space",
  "n03476991": "hair spray",
  "n03477143": "hairspring",
  "n03477303": "hair trigger",
  "n03477410": "halberd",
  "n03477512": "half binding",
  "n03477674": "half cross stitch",
  "n03477773": "half hatchet",
  "n03477902": "half hitch",
  "n03478033": "half-length",
  "n03478141": "half sole",
  "n03478261": "halftone, halftone engraving, photoengraving",
  "n03478400": "halftone",
  "n03478491": "half track",
  "n03478589": "half track",
  "n03478756": "hall",
  "n03478907": "hall",
  "n03479121": "hall",
  "n03479266": "Hall of Fame",
  "n03479397": "hall of residence",
  "n03479502": "hallstand",
  "n03479647": "hallucinogen, hallucinogenic drug, psychedelic drug, psychodelic drug",
  "n03479952": "hallway, hall",
  "n03480186": "haloperidol, Haldol",
  "n03480367": "halothane",
  "n03480579": "halter",
  "n03480719": "halter, hackamore",
  "n03480863": "halyard, halliard",
  "n03480973": "hame",
  "n03481172": "hammer",
  "n03481521": "hammer, power hammer",
  "n03481674": "hammer",
  "n03481824": "hammer, cock",
  "n03482001": "hammer",
  "n03482128": "hammerhead",
  "n03482252": "hammock, sack",
  "n03482405": "hamper",
  "n03482523": "hand",
  "n03482727": "hand ax, hand axe",
  "n03482877": "handball",
  "n03482988": "handball court",
  "n03483086": "handbarrow",
  "n03483230": "handbell",
  "n03483316": "hand blower, blow dryer, blow drier, hair dryer, hair drier",
  "n03483531": "handbow",
  "n03483637": "hand brake, emergency, emergency brake, parking brake",
  "n03483823": "hand calculator, pocket calculator",
  "n03483971": "handcar",
  "n03484083": "handcart, pushcart, cart, go-cart",
  "n03484487": "hand cream",
  "n03484576": "handcuff, cuff, handlock, manacle",
  "n03484809": "hand drill, handheld drill",
  "n03484931": "hand glass, simple microscope, magnifying glass",
  "n03485198": "hand glass, hand mirror",
  "n03485309": "hand grenade",
  "n03485407": "hand-held computer, hand-held microcomputer",
  "n03485575": "handhold",
  "n03485655": "handicraft, handcraft, handiwork, handwork",
  "n03485794": "handkerchief, hankie, hanky, hankey",
  "n03485997": "handle, grip, handgrip, hold",
  "n03487090": "handlebar",
  "n03487222": "handline, hand line",
  "n03487331": "handloom",
  "n03487444": "hand lotion",
  "n03487533": "hand luggage",
  "n03487642": "hand-me-down",
  "n03487774": "hand mower",
  "n03487886": "hand pump",
  "n03487963": "hand puppet, glove puppet, glove doll",
  "n03488111": "handrest",
  "n03488188": "handsaw, hand saw, carpenter's saw",
  "n03488438": "handset, French telephone",
  "n03488603": "hand shovel",
  "n03488784": "handspike",
  "n03488887": "handstamp, rubber stamp",
  "n03489048": "hand throttle",
  "n03489162": "hand tool",
  "n03490006": "hand towel, face towel",
  "n03490119": "hand truck, truck",
  "n03490324": "handwear, hand wear",
  "n03490449": "handwheel",
  "n03490649": "handwheel",
  "n03490784": "hangar queen",
  "n03490884": "hanger",
  "n03491032": "hang glider",
  "n03491178": "hanging, wall hanging",
  "n03491491": "Hanging Gardens of Babylon",
  "n03491724": "hangman's rope, hangman's halter, halter, hemp, hempen necktie",
  "n03491988": "hank",
  "n03492087": "hansom, hansom cab",
  "n03492250": "harbor, harbour",
  "n03492391": "hardback, hardcover",
  "n03492542": "hard disc, hard disk, fixed disk",
  "n03492717": "hard drug",
  "n03492922": "hard hat, tin hat, safety hat",
  "n03493079": "hard shoulder",
  "n03493219": "hardtop",
  "n03493333": "hardware, computer hardware",
  "n03493664": "hardware",
  "n03493792": "hardware, ironware",
  "n03493911": "hardware store, ironmonger, ironmonger's shop",
  "n03494105": "harem, hareem, seraglio, serail",
  "n03494278": "harmonica, mouth organ, harp, mouth harp",
  "n03494537": "harmonium, organ, reed organ",
  "n03494706": "harness",
  "n03495039": "harness",
  "n03495258": "harp",
  "n03495570": "harp",
  "n03495671": "harpoon",
  "n03495941": "harpoon gun",
  "n03496042": "harpoon line",
  "n03496183": "harpoon log",
  "n03496296": "harpsichord, cembalo",
  "n03496486": "Harris Tweed",
  "n03496612": "harrow",
  "n03496749": "Harvard University, Harvard",
  "n03496892": "harvester, reaper",
  "n03497100": "hash house",
  "n03497182": "hashish, hasheesh, haschisch, hash",
  "n03497352": "hasp",
  "n03497531": "hassock",
  "n03497657": "hat, chapeau, lid",
  "n03498316": "hatband",
  "n03498441": "hatbox",
  "n03498536": "hatch",
  "n03498662": "hatchback, hatchback door",
  "n03498781": "hatchback",
  "n03498866": "hatchel, heckle",
  "n03498962": "hatchet",
  "n03499142": "hatchway, opening, scuttle",
  "n03499354": "hatpin",
  "n03499468": "hauberk, byrnie",
  "n03499611": "havelock",
  "n03499796": "haven, oasis",
  "n03499907": "Hawaiian guitar, steel guitar",
  "n03500090": "hawse, hawsehole, hawsepipe",
  "n03500209": "hawser",
  "n03500295": "hawser bend",
  "n03500389": "hay bale",
  "n03500457": "hayfork",
  "n03500557": "hayloft, haymow, mow",
  "n03500699": "haymaker, hay conditioner",
  "n03500838": "hayrack, hayrig",
  "n03500971": "hayrack",
  "n03501068": "haywire",
  "n03501152": "hazard",
  "n03501288": "head",
  "n03501520": "head",
  "n03501614": "head",
  "n03501811": "head",
  "n03502042": "headband",
  "n03502200": "headboard",
  "n03502331": "head covering, veil",
  "n03502509": "headdress, headgear",
  "n03502777": "header",
  "n03502897": "header",
  "n03503097": "header, coping, cope",
  "n03503233": "header, lintel",
  "n03503358": "headfast",
  "n03503477": "head gasket",
  "n03503567": "head gate",
  "n03503718": "headgear",
  "n03503908": "headgear",
  "n03503997": "headlight, headlamp",
  "n03504205": "headpiece",
  "n03504293": "headpin, kingpin",
  "n03504420": "headquarters, HQ, military headquarters",
  "n03504723": "headquarters, central office, main office, home office, home base",
  "n03505015": "headrace",
  "n03505133": "headrest",
  "n03505207": "headrest, head restraint",
  "n03505383": "headsail",
  "n03505504": "headscarf",
  "n03505667": "headset",
  "n03505764": "head shop",
  "n03505942": "headshot",
  "n03506028": "headstall, headpiece",
  "n03506184": "headstock",
  "n03506370": "health spa, spa, health club",
  "n03506560": "hearing aid, ear trumpet",
  "n03506727": "hearing aid, deaf-aid",
  "n03506880": "hearse",
  "n03507048": "heart",
  "n03507241": "hearth, fireside",
  "n03507458": "hearthrug",
  "n03507554": "hearthstone",
  "n03507658": "heart-lung machine",
  "n03507857": "heart valve",
  "n03507963": "heat engine",
  "n03508101": "heater, warmer",
  "n03508485": "heat exchanger",
  "n03508628": "heating element",
  "n03508881": "heating pad, hot pad",
  "n03509025": "heating system, heating plant, heating, heat",
  "n03509394": "heat lamp, infrared lamp",
  "n03509608": "heat pump",
  "n03509843": "heat-seeking missile",
  "n03510072": "heat shield",
  "n03510244": "heat sink",
  "n03510384": "heaume",
  "n03510487": "heaver",
  "n03510583": "heavier-than-air craft",
  "n03510866": "heckelphone, basset oboe",
  "n03510987": "hectograph, heliotype",
  "n03511175": "hedge, hedgerow",
  "n03511333": "hedge trimmer",
  "n03511426": "heel",
  "n03511786": "heel",
  "n03511949": "heel",
  "n03512030": "helicon, bombardon",
  "n03512147": "helicopter, chopper, whirlybird, eggbeater",
  "n03512452": "heliograph",
  "n03512624": "heliometer",
  "n03512830": "heliport",
  "n03512911": "helm",
  "n03513137": "helmet",
  "n03513376": "helmet",
  "n03513627": "hem",
  "n03513961": "hematinic, haematinic",
  "n03514129": "hematocrit, haematocrit",
  "n03514340": "hemming-stitch",
  "n03514451": "hemostat, haemostat",
  "n03514588": "hemstitch",
  "n03514693": "hemstitch, hemstitching",
  "n03514894": "henroost",
  "n03514974": "heparin, Lipo-Hepin, Liquaemin",
  "n03515338": "heraldry",
  "n03515500": "herbal medicine",
  "n03515644": "herb garden",
  "n03515728": "herm",
  "n03515934": "hermitage",
  "n03516011": "heroin, diacetylmorphine",
  "n03516266": "herringbone",
  "n03516367": "herringbone, herringbone pattern",
  "n03516647": "Herschelian telescope, off-axis reflector",
  "n03516844": "Hessian boot, hessian, jackboot, Wellington, Wellington boot",
  "n03516996": "heterodyne receiver, superheterodyne receiver, superhet",
  "n03517271": "hexachlorophene",
  "n03517427": "hex nut",
  "n03517509": "hibachi",
  "n03517647": "hideaway, retreat",
  "n03517760": "hi-fi, high fidelity sound system",
  "n03517899": "high altar",
  "n03517982": "high-angle gun",
  "n03518135": "highball glass",
  "n03518230": "highboard",
  "n03518305": "highboy, tallboy",
  "n03518445": "highchair, feeding chair",
  "n03518631": "high gear, high",
  "n03518829": "high-hat cymbal, high hat",
  "n03518943": "highlighter",
  "n03519081": "highlighter",
  "n03519226": "high-pass filter",
  "n03519387": "high-rise, tower block",
  "n03519578": "highroad, trunk road",
  "n03519674": "high table",
  "n03519848": "high-warp loom",
  "n03519981": "highway, main road",
  "n03520275": "highway system",
  "n03520399": "high wire",
  "n03520493": "hijab",
  "n03520654": "hilt",
  "n03520811": "hindrance, hinderance, hitch, preventive, preventative, encumbrance, incumbrance, interference",
  "n03521076": "hinge, flexible joint",
  "n03521431": "hinging post, swinging post",
  "n03521544": "hip boot, thigh boot",
  "n03521675": "hipflask, pocket flask",
  "n03521771": "hip pad",
  "n03521899": "hip pocket",
  "n03522003": "hippodrome",
  "n03522100": "hip roof, hipped roof",
  "n03522239": "histamine blocker",
  "n03522559": "hit",
  "n03522634": "hitch",
  "n03522863": "hitch",
  "n03522990": "hitching post",
  "n03523134": "hitchrack, hitching bar",
  "n03523288": "hob",
  "n03523398": "hob",
  "n03523506": "hobble skirt",
  "n03523633": "hobby, hobbyhorse, rocking horse",
  "n03523849": "hobnail",
  "n03523987": "hockey skate",
  "n03524150": "hockey stick",
  "n03524287": "hod",
  "n03524425": "hodoscope",
  "n03524574": "hoe",
  "n03524745": "hoe handle",
  "n03524840": "hogan",
  "n03524976": "hogshead",
  "n03525074": "hoist",
  "n03525252": "hold, keep",
  "n03525372": "hold",
  "n03525454": "holder",
  "n03525693": "holding cell",
  "n03525827": "holding device",
  "n03526062": "holding pen, holding paddock, holding yard",
  "n03526198": "hole",
  "n03526805": "hole, golf hole",
  "n03527000": "hole card",
  "n03527149": "hollowware, holloware",
  "n03527243": "hologram, holograph",
  "n03527444": "holster",
  "n03527565": "holster",
  "n03527675": "holy of holies, sanctum sanctorum",
  "n03527930": "Holy Sepulcher, Holy Sepulchre",
  "n03528100": "home, nursing home, rest home",
  "n03528263": "home appliance, household appliance",
  "n03528523": "home computer",
  "n03528622": "home court",
  "n03528761": "home-farm",
  "n03528901": "home plate, home base, home, plate",
  "n03529175": "home room, homeroom",
  "n03529444": "homespun",
  "n03529629": "homestead",
  "n03529740": "homestretch",
  "n03529860": "home theater, home theatre",
  "n03530041": "homing device",
  "n03530189": "homing torpedo",
  "n03530326": "homolosine projection",
  "n03530511": "hone",
  "n03530642": "honeycomb",
  "n03530803": "honkytonk, dive",
  "n03530910": "hood, bonnet, cowl, cowling",
  "n03531281": "hood",
  "n03531447": "hood",
  "n03531546": "hood, exhaust hood",
  "n03531691": "hood",
  "n03531808": "hood, lens hood",
  "n03531982": "hood latch",
  "n03532080": "hoodoo",
  "n03532187": "hood ornament",
  "n03532342": "hook",
  "n03532672": "hook, claw",
  "n03532919": "hook",
  "n03533014": "hookah, narghile, nargileh, sheesha, shisha, chicha, calean, kalian, water pipe, hubble-bubble, hubbly-bubbly",
  "n03533392": "hook and eye",
  "n03533486": "hookup, assemblage",
  "n03533654": "hookup",
  "n03533845": "hook wrench, hook spanner",
  "n03533972": "hoop, ring",
  "n03534429": "hoop",
  "n03534580": "hoopskirt, crinoline",
  "n03534695": "hoosegow, hoosgow",
  "n03534776": "Hoover",
  "n03534890": "Hoover Dam",
  "n03535024": "hope chest, wedding chest",
  "n03535186": "hop garden, hop field",
  "n03535284": "hopper",
  "n03535429": "hop-picker, hopper",
  "n03535526": "hop pole",
  "n03535647": "hopsacking, hopsack",
  "n03535780": "horizontal bar, high bar",
  "n03535970": "horizontal section",
  "n03536122": "horizontal stabilizer, horizontal stabiliser, tailplane",
  "n03536348": "horizontal surface, level",
  "n03536568": "horizontal tail",
  "n03536761": "horn",
  "n03536931": "horn, saddle horn",
  "n03537085": "horn",
  "n03537241": "horn",
  "n03537412": "horn button",
  "n03537550": "hornpipe, pibgorn, stockhorn",
  "n03537714": "horoscope",
  "n03537866": "horror",
  "n03538037": "horse, gymnastic horse",
  "n03538179": "horsebox",
  "n03538300": "horsecar",
  "n03538406": "horse cart, horse-cart",
  "n03538542": "horsecloth",
  "n03538634": "horse-drawn vehicle",
  "n03538817": "horsehair",
  "n03538957": "horsehair wig",
  "n03539103": "horseless carriage",
  "n03539293": "horse pistol, horse-pistol",
  "n03539433": "horseshoe, shoe",
  "n03539546": "horseshoe",
  "n03539678": "horse-trail",
  "n03539754": "horsewhip",
  "n03539875": "hose, hosepipe",
  "n03540090": "hose",
  "n03540267": "hosiery, hose",
  "n03540476": "hospice",
  "n03540595": "hospital, infirmary",
  "n03540914": "hospital bed",
  "n03541091": "hospital room",
  "n03541269": "hospital ship",
  "n03541393": "hospital train",
  "n03541537": "hostel, youth hostel, student lodging",
  "n03541696": "hostel, hostelry, inn, lodge, auberge",
  "n03541923": "hot-air balloon",
  "n03542073": "hotbed",
  "n03542220": "hotbox",
  "n03542333": "hotel",
  "n03542605": "hotel-casino, casino-hotel",
  "n03542727": "hotel-casino, casino-hotel",
  "n03542860": "hotel room",
  "n03543012": "hot line",
  "n03543112": "hot pants",
  "n03543254": "hot plate, hotplate",
  "n03543394": "hot rod, hot-rod",
  "n03543511": "hot spot, hotspot",
  "n03543603": "hot tub",
  "n03543735": "hot-water bottle, hot-water bag",
  "n03543945": "houndstooth check, hound's-tooth check, dogstooth check, dogs-tooth check, dog's-tooth check",
  "n03544143": "hourglass",
  "n03544238": "hour hand, little hand",
  "n03544360": "house",
  "n03545150": "house",
  "n03545470": "houseboat",
  "n03545585": "houselights",
  "n03545756": "house of cards, cardhouse, card-house, cardcastle",
  "n03545961": "house of correction",
  "n03546112": "house paint, housepaint",
  "n03546235": "housetop",
  "n03546340": "housing, lodging, living accommodations",
  "n03546766": "housing",
  "n03547054": "hovel, hut, hutch, shack, shanty",
  "n03547229": "hovercraft, ground-effect machine",
  "n03547397": "howdah, houdah",
  "n03547530": "huarache, huaraches",
  "n03547658": "hub",
  "n03547861": "hub-and-spoke, hub-and-spoke system",
  "n03548086": "hubcap",
  "n03548195": "huck, huckaback",
  "n03548320": "hug-me-tight",
  "n03548402": "hula-hoop",
  "n03548533": "hulk",
  "n03548626": "hull",
  "n03548797": "Humber Bridge",
  "n03548930": "humeral veil, veil",
  "n03549097": "humming top",
  "n03549199": "Humvee, Hum-Vee",
  "n03549350": "hunter, hunting watch",
  "n03549473": "hunting knife",
  "n03549589": "hurdle",
  "n03549732": "hurricane deck, hurricane roof, promenade deck, awning deck",
  "n03549897": "hurricane lamp, hurricane lantern, tornado lantern, storm lantern, storm lamp",
  "n03550153": "hut, army hut, field hut",
  "n03550289": "hutch",
  "n03550420": "hutment",
  "n03550533": "hydantoin",
  "n03550671": "hydralazine, Apresoline",
  "n03550916": "hydrant",
  "n03551084": "hydraulic brake, hydraulic brakes",
  "n03551395": "hydraulic press",
  "n03551582": "hydraulic pump, hydraulic ram",
  "n03551790": "hydraulic system",
  "n03552001": "hydraulic transmission, hydraulic transmission system",
  "n03552169": "hydrochlorothiazide, Microzide, Esidrix, HydroDIURIL",
  "n03552449": "hydroelectric turbine",
  "n03552639": "hydroflumethiazide",
  "n03552749": "hydrofoil, hydroplane",
  "n03553019": "hydrofoil, foil",
  "n03553248": "hydrogen bomb, H-bomb, fusion bomb, thermonuclear bomb",
  "n03553486": "hydrometer, gravimeter",
  "n03553708": "hydromorphone hydrochloride, hydromorphone, Dilaudid",
  "n03553908": "hydroxychloroquine, Plaquenil",
  "n03554131": "hydroxyzine hydrochloride, hydroxyzine, Atarax, Vistaril",
  "n03554375": "hygrodeik",
  "n03554460": "hygrometer",
  "n03554645": "hygroscope",
  "n03554795": "hyoscyamine",
  "n03555006": "hyperbaric chamber",
  "n03555217": "hypercoaster",
  "n03555426": "hypermarket",
  "n03555564": "hypodermic needle",
  "n03555662": "hypodermic syringe, hypodermic, hypo",
  "n03555862": "hypsometer",
  "n03555996": "hysterosalpingogram",
  "n03556173": "I-beam",
  "n03556281": "ibuprofen, isobutylphenyl propionic acid, Advil, Motrin, Nuprin",
  "n03556679": "ice ax, ice axe, piolet",
  "n03556811": "iceboat, ice yacht, scooter",
  "n03556992": "icebreaker, iceboat",
  "n03557141": "ice cube",
  "n03557270": "iced-tea spoon",
  "n03557360": "ice hockey rink, ice-hockey rink",
  "n03557512": "icehouse",
  "n03557590": "ice machine",
  "n03557692": "ice maker",
  "n03557840": "ice pack, ice bag",
  "n03558007": "icepick, ice pick",
  "n03558176": "ice rink, ice-skating rink, ice",
  "n03558404": "ice skate",
  "n03558633": "ice tongs",
  "n03558739": "icetray",
  "n03558841": "ice wagon, ice-wagon",
  "n03558971": "icon, ikon",
  "n03559144": "iconography",
  "n03559373": "iconoscope",
  "n03559531": "Identikit, Identikit picture",
  "n03559841": "Iditarod Trail",
  "n03559999": "idle pulley, idler pulley, idle wheel",
  "n03560161": "idol, graven image, god",
  "n03560430": "igloo, iglu",
  "n03560567": "ignition, ignition system",
  "n03560860": "ignition coil",
  "n03561047": "ignition key",
  "n03561169": "ignition switch",
  "n03561345": "illustration",
  "n03561573": "imaret",
  "n03561657": "imbrication, overlapping, lapping",
  "n03561889": "imipramine, impramine hydrochloride, Imavate, Tofranil",
  "n03562126": "imitation",
  "n03562262": "counterfeit, forgery",
  "n03562408": "immersion heater",
  "n03562565": "immovable bandage",
  "n03562739": "immunogen, immunizing agent",
  "n03562958": "immunosuppressant, immunosuppressor, immunosuppressive drug, immunosuppressive, immune suppressant drug",
  "n03563200": "impact printer",
  "n03563358": "impedimenta",
  "n03563460": "impeller",
  "n03563611": "imperial",
  "n03563710": "implant",
  "n03563967": "implement",
  "n03564667": "import, importation",
  "n03564849": "impression",
  "n03565051": "Impressionism",
  "n03565288": "imprint",
  "n03565402": "improvisation",
  "n03565565": "improvised explosive device, I.E.D., IED",
  "n03565710": "impulse turbine",
  "n03565830": "in-basket, in-tray",
  "n03565991": "incendiary bomb, incendiary, firebomb",
  "n03566193": "incinerator",
  "n03566329": "inclined plane",
  "n03566555": "inclinometer, dip circle",
  "n03566730": "inclinometer",
  "n03566860": "incrustation, encrustation",
  "n03567066": "incubator, brooder",
  "n03567325": "indapamide, Lozal",
  "n03567474": "Independence Hall",
  "n03567635": "index register",
  "n03567788": "Indiaman",
  "n03567912": "Indian club",
  "n03568008": "Indian trail",
  "n03568117": "indicator",
  "n03568430": "indinavir, Crixivan",
  "n03568561": "indirect lighting",
  "n03568653": "indomethacin, Indocin",
  "n03568818": "induction coil",
  "n03569014": "inductor, inductance",
  "n03569174": "industrial watercourse",
  "n03569293": "inertial guidance system, inertial navigation system",
  "n03569494": "inflater, inflator",
  "n03569657": "infliximab, Remicade",
  "n03569964": "infrastructure, base",
  "n03570372": "infrastructure, substructure",
  "n03570526": "ingot, metal bar, block of metal",
  "n03570709": "ingredient",
  "n03570838": "inhalation anesthetic, inhalation anaesthetic, inhalation general anesthetic, inhalation general anaesthetic",
  "n03571155": "inhalant, inhalation",
  "n03571280": "inhaler, inhalator",
  "n03571439": "injector",
  "n03571625": "ink bottle, inkpot",
  "n03571706": "ink cartridge",
  "n03571853": "ink eraser",
  "n03571942": "ink-jet printer",
  "n03572107": "inkle",
  "n03572205": "inkstand",
  "n03572321": "inkwell, inkstand",
  "n03572449": "inlay",
  "n03572631": "inlay",
  "n03572832": "inlet manifold",
  "n03573005": "inner tube",
  "n03573154": "input",
  "n03573282": "insert, inset",
  "n03573464": "inset",
  "n03573574": "inside caliper",
  "n03573739": "inside clinch",
  "n03573848": "insole, innersole",
  "n03574004": "inspiration, brainchild",
  "n03574243": "instep",
  "n03574416": "instillator",
  "n03574555": "institution",
  "n03574816": "instrument",
  "n03575240": "instrumentality, instrumentation",
  "n03575691": "instrument of execution",
  "n03575958": "instrument of punishment",
  "n03576215": "instrument of torture",
  "n03576443": "intaglio, diaglyph",
  "n03576617": "intake, inlet",
  "n03576779": "intake manifold",
  "n03576955": "intake valve",
  "n03577090": "integrated circuit, microcircuit",
  "n03577312": "integrator, planimeter",
  "n03577474": "Intelnet",
  "n03577672": "interceptor",
  "n03577818": "interchange",
  "n03578055": "intercommunication system, intercom",
  "n03578251": "intercontinental ballistic missile, ICBM",
  "n03578435": "interface",
  "n03578656": "interface, port",
  "n03578981": "interferometer",
  "n03579137": "interferon",
  "n03579355": "interior decoration, decor",
  "n03579538": "interior door",
  "n03579699": "interlayer",
  "n03579791": "interlock, ignition interlock",
  "n03579982": "internal-combustion engine, ICE",
  "n03580518": "internal drive",
  "n03580615": "internet, net, cyberspace",
  "n03580845": "interphone",
  "n03580990": "interrupter",
  "n03581125": "intersection, crossroad, crossway, crossing, carrefour",
  "n03581354": "interstate, interstate highway",
  "n03581531": "interstice",
  "n03581634": "intoxicant",
  "n03581756": "intranet",
  "n03581897": "intraocular lens",
  "n03582096": "intrauterine device, IUD",
  "n03582305": "intravenous anesthetic",
  "n03582508": "intravenous pyelogram, IVP",
  "n03582658": "invention, innovation",
  "n03582840": "inverted pleat",
  "n03582959": "inverter",
  "n03583109": "iodochlorhydroxyquin, Clioquinol",
  "n03583252": "iodoform, triiodomethane",
  "n03583419": "ion engine",
  "n03583621": "ionization chamber, ionization tube",
  "n03583809": "ion pump",
  "n03583967": "ipecac",
  "n03584111": "ipratropium bromide, Atrovent",
  "n03584254": "iPod",
  "n03584400": "video iPod",
  "n03584526": "iproclozide",
  "n03584649": "iris, iris diaphragm",
  "n03584829": "iron, smoothing iron",
  "n03585073": "iron",
  "n03585337": "iron, branding iron",
  "n03585438": "irons, chains",
  "n03585551": "ironclad",
  "n03585682": "iron foundry",
  "n03585778": "iron horse",
  "n03585875": "ironing",
  "n03586090": "ironing board",
  "n03586219": "iron lung",
  "n03586448": "iron maiden",
  "n03586631": "ironmongery",
  "n03586760": "ironwork",
  "n03586911": "ironworks",
  "n03587050": "irregular, second",
  "n03587205": "irrigation ditch",
  "n03587318": "island",
  "n03587442": "isocarboxazid, Marplan",
  "n03587623": "isoflurane",
  "n03587715": "isoniazid, INH, Nydrazid",
  "n03587874": "isoproterenol, Isuprel",
  "n03588046": "isosorbide, Isordil",
  "n03588216": "izar",
  "n03588414": "item",
  "n03588668": "itraconazole, Sporanox",
  "n03588841": "jabot",
  "n03588951": "jack",
  "n03589093": "jack, knave",
  "n03589220": "jack",
  "n03589313": "jack, jackstones",
  "n03589513": "jack",
  "n03589672": "jack",
  "n03589791": "jacket",
  "n03590306": "jacket",
  "n03590475": "jacket",
  "n03590588": "jack-in-the-box",
  "n03590732": "jacklight",
  "n03590841": "jack-o'-lantern",
  "n03590932": "jack plane",
  "n03591028": "jackscrew, screw jack",
  "n03591116": "Jacob's ladder, jack ladder, pilot ladder",
  "n03591313": "jaconet",
  "n03591465": "jackstraw, spillikin",
  "n03591592": "Jacquard loom, Jacquard",
  "n03591798": "jacquard",
  "n03591901": "jag, dag",
  "n03592055": "jag",
  "n03592245": "jail, jailhouse, gaol, clink, slammer, poky, pokey",
  "n03592669": "jalousie",
  "n03592773": "jamb",
  "n03592931": "jammer",
  "n03593034": "jampan",
  "n03593122": "jampot, jamjar",
  "n03593222": "japan",
  "n03593362": "japan",
  "n03593526": "jar",
  "n03593862": "Jarvik heart, Jarvik artificial heart",
  "n03594010": "jaunting car, jaunty car",
  "n03594148": "javelin",
  "n03594277": "jaw",
  "n03594523": "Jaws of Life",
  "n03594734": "jean, blue jean, denim",
  "n03594945": "jeep, landrover",
  "n03595055": "jellaba",
  "n03595179": "je ne sais quoi",
  "n03595264": "jerkin",
  "n03595409": "jeroboam, double-magnum",
  "n03595523": "jersey",
  "n03595614": "jersey, T-shirt, tee shirt",
  "n03595737": "Jerusalem cross",
  "n03595860": "jet, jet plane, jet-propelled plane",
  "n03596099": "jet bridge",
  "n03596285": "jet engine",
  "n03596543": "jetliner",
  "n03596639": "jetsam",
  "n03596787": "jewel, gem, precious stone",
  "n03597147": "jeweler's glass",
  "n03597317": "jewelled headdress, jeweled headdress",
  "n03597469": "jewelry, jewellery",
  "n03597916": "jew's harp, jews' harp, mouth bow",
  "n03598151": "jib",
  "n03598299": "jibboom",
  "n03598385": "jig",
  "n03598515": "jig",
  "n03598646": "jiggermast, jigger",
  "n03598783": "jigsaw, scroll saw, fretsaw",
  "n03598930": "jigsaw puzzle",
  "n03599111": "jim crow",
  "n03599212": "jimdandy, jimhickey, crackerjack",
  "n03599351": "jimmy, jemmy",
  "n03599486": "jinrikisha, ricksha, rickshaw",
  "n03599628": "job",
  "n03599761": "job",
  "n03599964": "jobcentre",
  "n03600169": "job-oriented terminal",
  "n03600285": "jodhpurs, jodhpur breeches, riding breeches",
  "n03600475": "jodhpur, jodhpur boot, jodhpur shoe",
  "n03600617": "Johns Hopkins",
  "n03600722": "joinery",
  "n03600806": "joint, marijuana cigarette, reefer, stick, spliff",
  "n03600977": "joint",
  "n03601335": "joint",
  "n03601442": "Joint Direct Attack Munition, JDAM",
  "n03601638": "jointer, jointer plane, jointing plane, long plane",
  "n03601840": "joist",
  "n03601964": "joker",
  "n03602081": "jolly boat, jolly",
  "n03602194": "jorum",
  "n03602267": "joss",
  "n03602365": "joss house",
  "n03602465": "journal",
  "n03602562": "journal",
  "n03602686": "journal bearing",
  "n03602790": "journal box",
  "n03602883": "joystick",
  "n03603119": "judas",
  "n03603199": "juke, jook, juke joint, jook joint, juke house, jook house",
  "n03603442": "jungle gym",
  "n03603594": "junk",
  "n03603722": "jug",
  "n03603878": "Juggernaut",
  "n03603958": "juju, voodoo, hoodoo, fetish, fetich",
  "n03604156": "jukebox, nickelodeon",
  "n03604311": "jumbojet, jumbo jet",
  "n03604400": "jumper, pinafore, pinny",
  "n03604536": "jumper",
  "n03604629": "jumper",
  "n03604763": "jumper",
  "n03604843": "jumper cable, jumper lead, lead, booster cable",
  "n03605073": "jumping jack",
  "n03605233": "jump rope, skip rope, skipping rope",
  "n03605417": "jump seat",
  "n03605504": "jump suit",
  "n03605598": "jump suit, jumpsuit",
  "n03605722": "junction",
  "n03605915": "junction, conjunction",
  "n03606106": "junction barrier, barrier strip",
  "n03606251": "junk shop",
  "n03606347": "jury box",
  "n03606465": "jury mast",
  "n03606572": "K, jet, super acid, special K, honey oil, green, cat valium, super C",
  "n03606719": "Kaaba, Caaba",
  "n03607029": "kachina",
  "n03607186": "kaffiyeh",
  "n03607358": "Kakemono",
  "n03607527": "kalansuwa",
  "n03607659": "Kalashnikov",
  "n03607756": "kaleidoscope",
  "n03607923": "kameez",
  "n03608074": "kamikaze",
  "n03608224": "Kammon Strait Bridge",
  "n03608356": "kanamycin, Kantrex",
  "n03608504": "kanzu",
  "n03608661": "Kaopectate",
  "n03608870": "kat, khat, qat, quat, cat, Arabian tea, African tea",
  "n03609147": "katharometer",
  "n03609235": "kayak",
  "n03609397": "kazoo",
  "n03609542": "keel",
  "n03609786": "keelboat",
  "n03609959": "keelson",
  "n03610098": "keep, donjon, dungeon",
  "n03610270": "keepsake, souvenir, token, relic",
  "n03610418": "keg",
  "n03610524": "kennel, doghouse, dog house",
  "n03610682": "kepi, peaked cap, service cap, yachting cap",
  "n03610836": "keratoscope",
  "n03610992": "kerchief",
  "n03611169": "kern",
  "n03611314": "Kerr cell",
  "n03611590": "ketamine, ketamine hydrochloride, Ketalar",
  "n03612010": "ketch",
  "n03612134": "ketoprofen, Orudis, Orudis KT, Oruvail",
  "n03612378": "ketorolac, Torodal",
  "n03612559": "ketorolac tromethamine, Acular, Toradol",
  "n03612814": "kettle, boiler",
  "n03612965": "kettle, kettledrum, tympanum, tympani, timpani",
  "n03613294": "key",
  "n03613592": "key",
  "n03613873": "keyboard",
  "n03614007": "keyboard",
  "n03614383": "keyboard buffer",
  "n03614532": "keyboard instrument",
  "n03614782": "keyhole",
  "n03614887": "keyhole saw",
  "n03615037": "key ring",
  "n03615133": "keystone, key, headstone",
  "n03615300": "khadi, khaddar",
  "n03615406": "khaki",
  "n03615563": "khakis",
  "n03615655": "khimar",
  "n03615790": "khukuri",
  "n03615987": "kibble",
  "n03616091": "kick pleat",
  "n03616225": "kicksorter, pulse height analyzer",
  "n03616428": "kickstand",
  "n03616763": "kick starter, kick start",
  "n03616979": "kid glove, suede glove",
  "n03617095": "kiln",
  "n03617312": "kilt",
  "n03617480": "kimono",
  "n03617594": "kinescope, picture tube, television tube",
  "n03617834": "Kinetoscope",
  "n03618101": "king",
  "n03618219": "king",
  "n03618339": "king",
  "n03618546": "kingbolt, kingpin, swivel pin",
  "n03618678": "king post",
  "n03618797": "Kipp's apparatus",
  "n03618982": "kirk",
  "n03619050": "kirpan",
  "n03619196": "kirtle",
  "n03619275": "kirtle",
  "n03619396": "kit, outfit",
  "n03619650": "kit",
  "n03619793": "kitbag, kit bag",
  "n03619890": "kitchen",
  "n03620052": "kitchen appliance",
  "n03620353": "kitchenette",
  "n03620443": "kitchen garden, vegetable garden, vegetable patch",
  "n03620600": "kitchen island",
  "n03620730": "kitchen match",
  "n03620889": "kitchen sink",
  "n03620967": "kitchen table",
  "n03621049": "kitchen utensil",
  "n03621377": "kitchenware",
  "n03621473": "kite",
  "n03621694": "kite balloon",
  "n03621826": "kite tail",
  "n03621935": "kitsch",
  "n03622058": "klaxon, claxon",
  "n03622207": "Klein bottle",
  "n03622401": "klieg light",
  "n03622526": "klystron",
  "n03622698": "knee",
  "n03622839": "knee brace",
  "n03622931": "knee-high, knee-hi",
  "n03623073": "kneeler",
  "n03623198": "knee pad",
  "n03623338": "knee piece",
  "n03623448": "knickknack, novelty",
  "n03623556": "knife",
  "n03624134": "knife",
  "n03624400": "knife blade",
  "n03624497": "knife edge, cutting edge",
  "n03624653": "knife pleat",
  "n03624767": "knight, horse",
  "n03624966": "knit, knitting, knitwork",
  "n03625226": "knit, knit stitch, plain, plain stitch",
  "n03625355": "knit",
  "n03625539": "knitting machine",
  "n03625646": "knitting needle",
  "n03625783": "knitting stitch",
  "n03625943": "knitwear",
  "n03626014": "knob",
  "n03626115": "knob, boss",
  "n03626272": "knob, pommel",
  "n03626418": "knobble",
  "n03626502": "knobkerrie, knobkerry",
  "n03626658": "knockabout",
  "n03626760": "knocker, doorknocker, rapper",
  "n03626925": "knockoff, clone",
  "n03627021": "knockout drops",
  "n03627232": "knot",
  "n03627819": "knout",
  "n03627954": "knuckle joint, hinge joint",
  "n03628071": "kohl",
  "n03628215": "koto",
  "n03628421": "kraal",
  "n03628511": "kremlin",
  "n03628590": "Kremlin",
  "n03628728": "kris, creese, crease",
  "n03628831": "krummhorn, crumhorn, cromorne",
  "n03628984": "Kundt's tube",
  "n03629100": "Kurdistan",
  "n03629231": "kurta",
  "n03629401": "kylie, kiley",
  "n03629520": "kylix, cylix",
  "n03629643": "kymograph, cymograph",
  "n03629857": "laager, lager",
  "n03629986": "lab, laboratory, research lab, research laboratory, science lab, science laboratory",
  "n03630262": "lab bench, laboratory bench",
  "n03630383": "lab coat, laboratory coat",
  "n03630544": "labetalol, labetalol hydrochloride, Trandate, Normodyne",
  "n03630849": "labor camp, labour camp",
  "n03630992": "Labyrinth of Minos",
  "n03631177": "lace",
  "n03631445": "lace, lacing",
  "n03631710": "lacework",
  "n03631811": "lacquer",
  "n03631922": "lacquerware",
  "n03632100": "lacrosse ball",
  "n03632189": "lacuna, blank",
  "n03632277": "ladder",
  "n03632577": "ladder-back",
  "n03632729": "ladder-back, ladder-back chair",
  "n03632852": "ladder truck, aerial ladder truck",
  "n03632963": "ladies' room, powder room",
  "n03633091": "ladle",
  "n03633341": "lady chapel",
  "n03633474": "lagan, lagend, ligan",
  "n03633632": "lagerphone",
  "n03633886": "lag screw, lag bolt",
  "n03634034": "lake dwelling, pile dwelling",
  "n03634189": "Lake Mead",
  "n03634469": "Lake Powell",
  "n03634723": "Lake Volta",
  "n03634899": "lally, lally column",
  "n03635032": "lamasery",
  "n03635108": "lambrequin",
  "n03635236": "lambrequin",
  "n03635330": "lame",
  "n03635452": "lamella",
  "n03635516": "laminar flow clean room",
  "n03635668": "laminate",
  "n03635932": "lamination",
  "n03636062": "lamivudine, 3TC",
  "n03636248": "lamp",
  "n03636649": "lamp",
  "n03636891": "lamp chimney, chimney",
  "n03637027": "lamp house, lamphouse, lamp housing",
  "n03637181": "lamppost",
  "n03637318": "lampshade, lamp shade",
  "n03637480": "lanai",
  "n03637618": "lancet, lance",
  "n03637787": "lancet arch, lancet",
  "n03637898": "lancet window",
  "n03638014": "landau",
  "n03638180": "lander",
  "n03638321": "landing, landing place",
  "n03638511": "landing",
  "n03638623": "landing craft",
  "n03638743": "landing flap",
  "n03638883": "landing gear",
  "n03639077": "landing net",
  "n03639230": "landing skid",
  "n03639360": "landing stage",
  "n03639497": "land line, landline",
  "n03639675": "land mine, ground-emplaced mine, booby trap",
  "n03639880": "land office",
  "n03640011": "landscape, landscape painting",
  "n03640166": "landscape",
  "n03640288": "landscaping",
  "n03640456": "landside",
  "n03640589": "lane",
  "n03640660": "lane",
  "n03640850": "lanolin",
  "n03640988": "lantern",
  "n03641134": "lantern pinion, lantern wheel",
  "n03641302": "lanyard, laniard",
  "n03641451": "lanyard, laniard",
  "n03641569": "lanyard, laniard",
  "n03641706": "lap, overlap",
  "n03641947": "lap, lap covering",
  "n03642144": "laparoscope",
  "n03642341": "lapboard",
  "n03642444": "lapel",
  "n03642573": "lap joint, splice",
  "n03642716": "lappet",
  "n03642806": "laptop, laptop computer",
  "n03642928": "larboard, port",
  "n03643149": "laryngoscope",
  "n03643253": "laser, optical maser",
  "n03643491": "laser-guided bomb, LGB",
  "n03643737": "laser printer",
  "n03643907": "lash, thong",
  "n03644073": "lashing",
  "n03644248": "lash-up, contrivance",
  "n03644378": "lasso, lariat, riata, reata",
  "n03644532": "last, shoemaker's last, cobbler's last",
  "n03644698": "Lastex",
  "n03644858": "latch",
  "n03645011": "latch, door latch",
  "n03645168": "latchet",
  "n03645290": "latchkey",
  "n03645411": "latchstring",
  "n03645577": "lateen, lateen sail",
  "n03645730": "lateen-rig",
  "n03645866": "Lateran Palace",
  "n03646020": "latex paint, latex, rubber-base paint",
  "n03646148": "lath",
  "n03646296": "lathe",
  "n03646546": "lathi, lathee",
  "n03646695": "Latin cross",
  "n03646809": "latrine",
  "n03646916": "lattice, latticework, fretwork",
  "n03647108": "laudanum, tincture of opium",
  "n03647291": "laugh track",
  "n03647423": "launch",
  "n03647520": "launcher, rocket launcher",
  "n03647691": "launching pad, launchpad, launch pad, launch area, pad",
  "n03647872": "launderette, Laundromat",
  "n03648066": "laundry",
  "n03648219": "laundry, wash, washing, washables",
  "n03648431": "laundry cart",
  "n03648527": "laundry detergent",
  "n03648667": "laundry truck",
  "n03648804": "laurel, laurel wreath, bay wreath",
  "n03649003": "lavalava",
  "n03649161": "lavaliere, lavalier, lavalliere",
  "n03649288": "laver",
  "n03649459": "court, lawcourt, court of law, court of justice",
  "n03649674": "lawn chair, garden chair",
  "n03649797": "lawn furniture",
  "n03649909": "lawn mower, mower",
  "n03650084": "laxative",
  "n03650173": "layer, bed",
  "n03650551": "layette",
  "n03650682": "lay figure",
  "n03650803": "lazaretto, lazaret, lazarette, lazar house, pesthouse",
  "n03650973": "lazy daisy stitch",
  "n03651084": "lead, pencil lead",
  "n03651265": "lead, leading",
  "n03651388": "lead-acid battery, lead-acid accumulator",
  "n03651605": "lead-in",
  "n03651739": "leading edge",
  "n03651843": "leading rein",
  "n03651947": "lead line, sounding line",
  "n03652100": "lead pencil",
  "n03652226": "leaf",
  "n03652389": "leaf spring",
  "n03652530": "Leaning Tower, Leaning Tower of Pisa",
  "n03652729": "lean-to",
  "n03652826": "lean-to tent",
  "n03652932": "leash, tether, lead",
  "n03653110": "leatherette, imitation leather",
  "n03653220": "leather strip",
  "n03653376": "leatherwork",
  "n03653454": "Leclanche cell",
  "n03653583": "lectern, reading desk",
  "n03653740": "lecture room",
  "n03653833": "lederhosen",
  "n03653975": "ledger board",
  "n03654086": "leflunomide, Arava",
  "n03654374": "left field, leftfield, left",
  "n03654576": "leg",
  "n03654826": "leg",
  "n03655072": "legging, leging, leg covering",
  "n03655295": "Lego, Lego set",
  "n03655470": "Leiden jar, Leyden jar",
  "n03655589": "leister",
  "n03655720": "leisure wear",
  "n03655838": "lemon, stinker",
  "n03656011": "lemon grove",
  "n03656091": "lending library, circulating library",
  "n03656231": "length",
  "n03656374": "lenitive",
  "n03656484": "lens, lense, lens system",
  "n03656957": "lens, electron lens",
  "n03657121": "lens cap, lens cover",
  "n03657239": "lens implant, interocular lens implant, IOL",
  "n03657511": "leotard, unitard, body suit, cat suit",
  "n03657825": "lethal dose",
  "n03657918": "letter bomb, parcel bomb, package bomb",
  "n03658102": "letter case",
  "n03658185": "letter opener, paper knife, paperknife",
  "n03658373": "levallorphan, Lorfan",
  "n03658635": "levee",
  "n03658759": "levee",
  "n03658858": "level, spirit level",
  "n03659122": "level crossing, grade crossing",
  "n03659292": "lever",
  "n03659686": "lever, lever tumbler",
  "n03659809": "lever",
  "n03659950": "lever lock",
  "n03660124": "Levi's, levis",
  "n03660228": "Liberty Bell",
  "n03660394": "liberty cap",
  "n03660562": "Liberty ship",
  "n03660664": "library, depository library",
  "n03660909": "library",
  "n03661043": "library",
  "n03661180": "license plate, numberplate",
  "n03661340": "lid",
  "n03661621": "lidar",
  "n03661861": "lido",
  "n03662016": "Lidocaine, Xylocaine",
  "n03662207": "lido deck",
  "n03662301": "Liebig condenser",
  "n03662452": "lie detector",
  "n03662601": "lifeboat",
  "n03662719": "life buoy, lifesaver, life belt, life ring",
  "n03662887": "life jacket, life vest, cork jacket",
  "n03663069": "lifeline",
  "n03663234": "lifeline",
  "n03663332": "life mask",
  "n03663433": "life office",
  "n03663531": "life preserver, preserver, flotation device",
  "n03663781": "life raft, Carling float",
  "n03663910": "life-support system, life support",
  "n03664159": "life-support system, life support",
  "n03664374": "lift",
  "n03664514": "lift",
  "n03664675": "lifting device",
  "n03664840": "lift pump",
  "n03664943": "ligament",
  "n03665083": "ligature",
  "n03665232": "ligature",
  "n03665366": "light, light source",
  "n03665851": "light arm",
  "n03665924": "light bulb, lightbulb, bulb, incandescent lamp, electric light, electric-light bulb",
  "n03666238": "light circuit, lighting circuit",
  "n03666362": "light-emitting diode, LED",
  "n03666591": "lighter, light, igniter, ignitor",
  "n03666917": "lighter-than-air craft",
  "n03667060": "light filter, diffusing screen",
  "n03667235": "lighting",
  "n03667380": "lighting fixture",
  "n03667552": "light machine gun",
  "n03667664": "light meter, exposure meter, photometer",
  "n03667829": "light microscope",
  "n03668067": "lightning rod, lightning conductor",
  "n03668279": "light pen, electronic stylus",
  "n03668488": "lightship",
  "n03668642": "likeness, semblance",
  "n03668803": "Lilo",
  "n03668906": "limb",
  "n03669094": "limb",
  "n03669245": "limber",
  "n03669367": "limbers",
  "n03669534": "limekiln",
  "n03669665": "limelight, calcium light",
  "n03669886": "limiter, clipper",
  "n03670208": "limousine, limo",
  "n03670339": "linchpin, lynchpin",
  "n03670456": "Lincoln Memorial",
  "n03670622": "lincomycin, Lincocin",
  "n03670849": "line",
  "n03671272": "line, railway line, rail line",
  "n03671473": "line",
  "n03671668": "line, product line, line of products, line of merchandise, business line, line of business",
  "n03671914": "linear accelerator, linac",
  "n03672097": "linecut, line block, line engraving",
  "n03672245": "linecut, line engraving",
  "n03672352": "linen",
  "n03672521": "linen",
  "n03672638": "line of defense, line of defence",
  "n03672827": "line printer, line-at-a-time printer",
  "n03673027": "liner, ocean liner",
  "n03673270": "liner, lining",
  "n03673450": "lingerie, intimate apparel",
  "n03673594": "liniment, embrocation",
  "n03673767": "lining, liner",
  "n03673971": "link, linkup, tie, tie-in",
  "n03674270": "link, data link",
  "n03674440": "linkage",
  "n03674591": "links, golf links",
  "n03674731": "Link trainer",
  "n03674842": "linocut",
  "n03674946": "linocut",
  "n03675076": "linoleum knife, linoleum cutter",
  "n03675235": "Linotype, Linotype machine",
  "n03675445": "linsey-woolsey",
  "n03675558": "linstock",
  "n03675780": "lint",
  "n03675907": "lion-jaw forceps",
  "n03676006": "lip balm",
  "n03676087": "lip-gloss",
  "n03676175": "lipid-lowering medicine, lipid-lowering medication, statin drug, statin",
  "n03676483": "lipstick, lip rouge",
  "n03676623": "liqueur glass",
  "n03676759": "liquid crystal display, LCD",
  "n03677026": "liquid detergent",
  "n03677115": "liquid metal reactor",
  "n03677231": "liquid soap",
  "n03677308": "lisinopril, Prinival, Zestril",
  "n03677540": "lisle, lisle thread",
  "n03677682": "lisle",
  "n03677766": "lister, lister plow, lister plough, middlebreaker, middle buster",
  "n03677976": "lithograph, lithograph machine",
  "n03678220": "lithograph",
  "n03678362": "litter",
  "n03678558": "litterbin, litter basket, litter-basket",
  "n03678729": "little theater, little theatre",
  "n03678879": "live axle, driving axle",
  "n03679037": "live load, superload",
  "n03679174": "livery",
  "n03679274": "livery stable",
  "n03679384": "living quarters, quarters",
  "n03679712": "living room, living-room, sitting room, front room, parlor, parlour",
  "n03679986": "load, loading, burden",
  "n03680248": "load",
  "n03680355": "Loafer",
  "n03680512": "loaner",
  "n03680629": "loan office",
  "n03680734": "lobe",
  "n03680858": "lobster pot",
  "n03680942": "local",
  "n03681148": "local anesthetic, local anaesthetic, local, topical anesthetic, topical anaesthetic",
  "n03681477": "local area network, LAN",
  "n03681813": "local oscillator, heterodyne oscillator",
  "n03682024": "local road, local street",
  "n03682189": "location",
  "n03682380": "Lochaber ax",
  "n03682487": "lock",
  "n03682877": "lock, ignition lock",
  "n03683079": "lock, lock chamber",
  "n03683341": "lock",
  "n03683457": "lockage",
  "n03683606": "locker",
  "n03683708": "locker room",
  "n03683995": "locket",
  "n03684143": "lock-gate",
  "n03684224": "locking pliers",
  "n03684338": "locknut, safety nut",
  "n03684489": "lockring, lock ring, lock washer",
  "n03684611": "lockstitch",
  "n03684740": "lockup",
  "n03684823": "locomotive, engine, locomotive engine, railway locomotive",
  "n03685307": "lodge, indian lodge",
  "n03685486": "lodge, hunting lodge",
  "n03685640": "lodge",
  "n03685820": "lodging house, rooming house",
  "n03685962": "Loestrin",
  "n03686130": "loft, attic, garret",
  "n03686363": "loft, pigeon loft",
  "n03686470": "loft",
  "n03686658": "log",
  "n03686924": "log cabin",
  "n03687003": "loge",
  "n03687137": "loggia",
  "n03687306": "logic element",
  "n03687448": "log line",
  "n03687604": "Lomotil",
  "n03687688": "lomustine",
  "n03687820": "longboat",
  "n03687928": "longbow",
  "n03688066": "long iron",
  "n03688192": "long johns",
  "n03688316": "longshot",
  "n03688405": "long sleeve",
  "n03688504": "long tom",
  "n03688605": "long trousers, long pants",
  "n03688707": "long underwear, union suit",
  "n03688832": "looking glass, glass",
  "n03688943": "lookout, observation tower, lookout station, observatory",
  "n03689157": "loom",
  "n03689347": "loop",
  "n03689443": "loophole",
  "n03689570": "loop knot",
  "n03689700": "loop-line",
  "n03689840": "Lo/Ovral",
  "n03690005": "lorazepam, Ativan",
  "n03690168": "lorgnette",
  "n03690279": "Lorraine cross, cross of Lorraine",
  "n03690473": "lorry, camion",
  "n03690600": "lorry",
  "n03690695": "lost-and-found",
  "n03690851": "lota",
  "n03690938": "lotion",
  "n03691128": "lotion, application",
  "n03691459": "loudspeaker, speaker, speaker unit, loudspeaker system, speaker system",
  "n03691817": "lounge, waiting room, waiting area",
  "n03692004": "lounger",
  "n03692136": "lounging jacket, smoking jacket",
  "n03692272": "lounging pajama, lounging pyjama",
  "n03692379": "loungewear",
  "n03692522": "loupe, jeweler's loupe",
  "n03692676": "louver, louvre, fin",
  "n03692842": "louvered window, jalousie",
  "n03692942": "Louvre, Louvre Museum",
  "n03693089": "lovastatin, Mevacor",
  "n03693293": "love knot, lovers' knot, lover's knot, true lovers' knot, true lover's knot",
  "n03693474": "love seat, loveseat, tete-a-tete, vis-a-vis",
  "n03693617": "love-token",
  "n03693707": "loving cup",
  "n03693860": "lowboy",
  "n03693973": "lower berth, lower",
  "n03694098": "lower deck, third deck",
  "n03694196": "low-pass filter",
  "n03694356": "low-warp-loom",
  "n03694490": "loxapine, Loxitane",
  "n03694639": "LP, L-P",
  "n03694761": "L-plate",
  "n03694949": "lubber's hole",
  "n03695122": "lubricating system, force-feed lubricating system, force feed, pressure-feed lubricating system, pressure feed",
  "n03695452": "luff",
  "n03695616": "lug",
  "n03695753": "luge",
  "n03695857": "Luger",
  "n03695957": "luggage carrier",
  "n03696065": "luggage compartment, automobile trunk, trunk",
  "n03696301": "luggage rack, roof rack",
  "n03696445": "lugger",
  "n03696568": "lugsail, lug",
  "n03696746": "lug wrench",
  "n03696909": "lumberjack, lumber jacket",
  "n03697007": "lumbermill, sawmill",
  "n03697109": "lumber room",
  "n03697265": "lumberyard",
  "n03697366": "lunar excursion module, lunar module, LEM",
  "n03697552": "lunchroom",
  "n03697665": "lunette, fenestella",
  "n03697812": "lunette",
  "n03697913": "lungi, lungyi, longyi",
  "n03698123": "lunula",
  "n03698226": "lusterware",
  "n03698360": "lute",
  "n03698604": "luxury liner, express luxury liner",
  "n03698723": "lyceum",
  "n03698815": "lychgate, lichgate",
  "n03698977": "lymphangiogram",
  "n03699138": "lypressin",
  "n03699280": "lyre",
  "n03699396": "lysergic acid diethylamide, LSD",
  "n03699591": "machete, matchet, panga",
  "n03699754": "machicolation",
  "n03699975": "machine",
  "n03700963": "machine, simple machine",
  "n03701191": "machine bolt",
  "n03701391": "machine gun",
  "n03701640": "machinery",
  "n03701790": "machine screw",
  "n03701947": "machine shop",
  "n03702067": "machine stitch, sewing-machine stitch",
  "n03702248": "machine tool",
  "n03702440": "machinist's vise, metalworking vise",
  "n03702582": "machmeter",
  "n03702719": "macintosh, mackintosh, mac, mack",
  "n03702904": "Mackinac Bridge",
  "n03703075": "mackinaw",
  "n03703203": "mackinaw, Mackinaw boat",
  "n03703320": "mackinaw, Mackinaw blanket",
  "n03703463": "mackinaw, Mackinaw coat",
  "n03703590": "mackintosh, macintosh",
  "n03703730": "macrame",
  "n03703862": "madras",
  "n03703945": "Mae West, air jacket",
  "n03704038": "magazine, powder store, powder magazine",
  "n03704210": "magazine",
  "n03704376": "magazine, cartridge",
  "n03704549": "magazine rack",
  "n03704640": "magic bullet",
  "n03704834": "magic lantern",
  "n03704926": "magic realism",
  "n03705134": "Maginot Line",
  "n03705379": "magnet",
  "n03705808": "magnetic bottle",
  "n03706016": "magnetic bubble memory",
  "n03706229": "magnetic compass",
  "n03706415": "magnetic core memory, core memory",
  "n03706653": "magnetic disk, magnetic disc, disk, disc",
  "n03706939": "magnetic head",
  "n03707171": "magnetic mine",
  "n03707372": "magnetic needle",
  "n03707597": "magnetic recorder",
  "n03707766": "magnetic stripe",
  "n03708036": "magnetic tape, mag tape, tape",
  "n03708425": "magneto, magnetoelectric machine",
  "n03708672": "magnetograph",
  "n03708843": "magnetometer, gaussmeter",
  "n03708962": "magnetron",
  "n03709206": "magnifier",
  "n03709363": "magnum",
  "n03709454": "magnum opus",
  "n03709545": "magnus hitch",
  "n03709644": "mail",
  "n03709823": "mailbag, postbag",
  "n03709960": "mailbag, mail pouch",
  "n03710079": "mailboat, mail boat, packet, packet boat",
  "n03710193": "mailbox, letter box",
  "n03710294": "mail car",
  "n03710421": "maildrop",
  "n03710528": "mailer",
  "n03710637": "maillot",
  "n03710721": "maillot, tank suit",
  "n03710819": "mail slot",
  "n03710937": "mailsorter",
  "n03711044": "mail train",
  "n03711145": "main",
  "n03711384": "main course",
  "n03711459": "main deck, second deck",
  "n03711603": "main drag",
  "n03711711": "mainframe, mainframe computer",
  "n03711897": "main line",
  "n03711999": "mainmast",
  "n03712111": "main rotor",
  "n03712337": "mainsail",
  "n03712444": "mainspring",
  "n03712637": "mainstay",
  "n03712729": "main street, high street",
  "n03712887": "main-topmast",
  "n03712981": "main-topsail",
  "n03713069": "main yard",
  "n03713151": "maisonette, maisonnette",
  "n03713254": "maisonette, maisonnette",
  "n03713436": "majolica, maiolica",
  "n03713556": "major suit",
  "n03713736": "major tranquilizer, major tranquillizer, major tranquilliser, antipsychotic drug, antipsychotic agent, antipsychotic, neuroleptic drug, neuroleptic agent, neuroleptic",
  "n03714235": "makeup, make-up, war paint",
  "n03714610": "makeweight",
  "n03714721": "makeweight, filler",
  "n03714899": "making",
  "n03715114": "Maksutov telescope",
  "n03715275": "malacca, malacca cane",
  "n03715386": "mallet, beetle",
  "n03715669": "mallet, hammer",
  "n03715892": "mallet",
  "n03716091": "Maltese cross",
  "n03716228": "mammogram",
  "n03716327": "man, piece",
  "n03716656": "mandala",
  "n03716887": "mandola",
  "n03716966": "mandolin",
  "n03717131": "manger, trough",
  "n03717285": "mangle",
  "n03717447": "manhole",
  "n03717622": "manhole cover",
  "n03717750": "manifold",
  "n03717921": "mannequin, manikin, mannikin, manakin, form",
  "n03718056": "mannitol, Osmitrol",
  "n03718212": "man-of-war, ship of the line",
  "n03718335": "manometer",
  "n03718458": "manor, manor house",
  "n03718581": "manor hall, hall",
  "n03718699": "MANPAD",
  "n03718789": "mansard, mansard roof",
  "n03718935": "manse",
  "n03719053": "mansion, mansion house, manse, hall, residence",
  "n03719249": "manta",
  "n03719343": "mantel, mantelpiece, mantle, mantlepiece, chimneypiece",
  "n03719560": "mantelet, mantilla",
  "n03719650": "mantelet, mantlet",
  "n03719743": "mantilla",
  "n03719826": "mantrap",
  "n03719911": "mantua",
  "n03720005": "Mao jacket",
  "n03720163": "map",
  "n03720443": "map projection",
  "n03720665": "maquiladora",
  "n03720891": "maraca",
  "n03721047": "marble",
  "n03721169": "marble",
  "n03721252": "marching order",
  "n03721384": "marimba, xylophone",
  "n03721590": "marina",
  "n03721691": "Marineland",
  "n03721797": "marker",
  "n03722007": "marker",
  "n03722169": "market garden",
  "n03722288": "marketplace, market place, mart, market",
  "n03722552": "marline",
  "n03722646": "marlinespike, marlinspike, marlingspike",
  "n03722827": "marmite",
  "n03722944": "marocain, crepe marocain",
  "n03723054": "maroon",
  "n03723153": "marquee, marquise",
  "n03723267": "marquetry, marqueterie",
  "n03723439": "marriage bed",
  "n03723534": "marseille",
  "n03723653": "marshalling yard",
  "n03723781": "martello tower",
  "n03723885": "martingale",
  "n03724066": "mascara",
  "n03724176": "maser",
  "n03724417": "masher",
  "n03724538": "mashie, five iron",
  "n03724623": "mashie niblick, seven iron",
  "n03724756": "masjid, musjid",
  "n03724870": "mask",
  "n03725035": "mask",
  "n03725198": "masking piece, masking",
  "n03725355": "masking tape, masking paper",
  "n03725506": "Masonite",
  "n03725600": "Mason jar",
  "n03725717": "masonry",
  "n03725869": "mason's level",
  "n03725968": "Massachusetts Institute of Technology, MIT",
  "n03726116": "massage parlor",
  "n03726233": "massage parlor",
  "n03726371": "mass spectrograph",
  "n03726516": "mass spectrometer, spectrometer",
  "n03726760": "mast",
  "n03726993": "mast",
  "n03727067": "mastaba, mastabah",
  "n03727274": "master, master copy, original",
  "n03727465": "master bedroom",
  "n03727605": "masterpiece, chef-d'oeuvre",
  "n03727738": "masthead",
  "n03727837": "mat",
  "n03727946": "mat, gym mat",
  "n03728131": "mat",
  "n03728318": "mat, matting",
  "n03728437": "match, lucifer, friction match",
  "n03728811": "match, mate",
  "n03728982": "match",
  "n03729131": "matchboard",
  "n03729308": "matchbook",
  "n03729402": "matchbox",
  "n03729482": "matchlock",
  "n03729647": "match plane, tonguing and grooving plane",
  "n03729826": "matchstick",
  "n03729951": "material",
  "n03730153": "materiel, equipage",
  "n03730334": "maternity hospital",
  "n03730494": "maternity ward",
  "n03730655": "matrix",
  "n03730788": "Matthew Walker, Matthew Walker knot",
  "n03730893": "matting",
  "n03731019": "mattock",
  "n03731164": "mattress",
  "n03731483": "mattress cover",
  "n03731590": "mattress pad",
  "n03731695": "maul, sledge, sledgehammer",
  "n03731882": "maulstick, mahlstick",
  "n03732020": "Mauser",
  "n03732114": "mausoleum",
  "n03732252": "Mausoleum at Halicarnasus",
  "n03732458": "maxi",
  "n03732543": "Maxim gun",
  "n03732658": "maximum and minimum thermometer",
  "n03732828": "Maxzide",
  "n03732992": "Mayflower",
  "n03733131": "maypole",
  "n03733281": "maze, labyrinth",
  "n03733465": "mazer",
  "n03733547": "means",
  "n03733644": "measure",
  "n03733805": "measuring cup",
  "n03733925": "measuring instrument, measuring system, measuring device",
  "n03735637": "measuring stick, measure, measuring rod",
  "n03735963": "meat counter",
  "n03736064": "meat grinder",
  "n03736147": "meat hook",
  "n03736269": "meat house",
  "n03736372": "meat safe",
  "n03736470": "meat thermometer",
  "n03736674": "mebendazole",
  "n03736809": "Meccano, Meccano set",
  "n03736970": "mechanical device",
  "n03737912": "mechanical drawing",
  "n03738066": "mechanical piano, Pianola, player piano",
  "n03738241": "mechanical system",
  "n03738472": "mechanism",
  "n03739136": "meclizine, meclizine hydrochloride, Antivert",
  "n03739327": "meclofenamate, meclofenamate sodium, Meclomen",
  "n03739518": "medical building, health facility, healthcare facility",
  "n03739693": "medical instrument",
  "n03740161": "medicine, medication, medicament, medicinal drug",
  "n03742019": "medicine ball",
  "n03742115": "medicine chest, medicine cabinet",
  "n03742238": "MEDLINE",
  "n03742416": "meerschaum",
  "n03742531": "mefenamic acid, Ponstel",
  "n03742728": "mefloquine, mefloquine hydrochloride, Larium, Mephaquine",
  "n03743016": "megalith, megalithic structure",
  "n03743279": "megaphone",
  "n03743422": "megaton bomb",
  "n03743577": "melphalan, Alkeran",
  "n03743761": "membrane",
  "n03743902": "memorial, monument",
  "n03744276": "memory, computer memory, storage, computer storage, store, memory board",
  "n03744684": "memory chip",
  "n03744840": "memory device, storage device",
  "n03745146": "menagerie, zoo, zoological garden",
  "n03745285": "mend, patch, darn",
  "n03745487": "mending",
  "n03745571": "menhir, standing stone",
  "n03745712": "meniscus",
  "n03745864": "meniscus",
  "n03746005": "menorah",
  "n03746155": "Menorah",
  "n03746330": "man's clothing",
  "n03746486": "men's room, men's",
  "n03746574": "mental hospital, psychiatric hospital, mental institution, institution, mental home, insane asylum, asylum",
  "n03746861": "menthol",
  "n03746994": "mentholated salve",
  "n03747103": "meperidine, meperidine hydrochloride, Demerol",
  "n03747281": "mephenytoin, Mesantoin",
  "n03747508": "mephobarbital, Mebaral",
  "n03747746": "meprobamate, Miltown, Equanil, Meprin",
  "n03748002": "merbromine, Mercurochrome",
  "n03748162": "mercantile establishment, retail store, sales outlet, outlet",
  "n03748456": "mercaptopurine, Purinethol",
  "n03748691": "Mercator projection, Mercator's projection",
  "n03748886": "merchandise, ware, product",
  "n03749409": "mercurial ointment",
  "n03749504": "mercury barometer",
  "n03749634": "mercury cell",
  "n03749807": "mercury thermometer, mercury-in-glass thermometer",
  "n03750206": "mercury-vapor lamp",
  "n03750437": "mercy seat",
  "n03750540": "mercy seat",
  "n03750614": "merlon",
  "n03750748": "Merrimac",
  "n03750912": "mescaline, peyote",
  "n03751065": "mess, mess hall",
  "n03751269": "mess jacket, monkey jacket, shell jacket",
  "n03751458": "mess kit",
  "n03751590": "messuage",
  "n03751757": "metal detector",
  "n03751977": "metallic",
  "n03752071": "metallic",
  "n03752185": "metal screw",
  "n03752262": "metalware",
  "n03752398": "metal wood",
  "n03752521": "metalwork",
  "n03752649": "metaproterenol, Alupent",
  "n03752922": "meteorological balloon",
  "n03753077": "meter",
  "n03753514": "meterstick, metrestick",
  "n03753657": "metformin, Glucophage",
  "n03753826": "methacholine, Mecholyl",
  "n03754014": "methadone, methadone hydrochloride, methadon, dolophine hydrochloride, fixer, synthetic heroin",
  "n03754295": "methamphetamine, methamphetamine hydrochloride, Methedrine, meth, deoxyephedrine, chalk, chicken feed, crank, glass, ice, shabu, trash",
  "n03754676": "methapyrilene",
  "n03754822": "methaqualone, Quaalude",
  "n03754979": "metharbital, Gemonil",
  "n03755140": "methenamine, Mandelamine, Urex",
  "n03755388": "methicillin",
  "n03755545": "methocarbamol, Robaxin",
  "n03755712": "methotrexate, methotrexate sodium, amethopterin",
  "n03755991": "methyldopa, alpha methyl dopa, Aldomet",
  "n03756184": "methylenedioxymethamphetamine, MDMA",
  "n03756624": "methylphenidate, Ritalin",
  "n03756857": "metoprolol, Lopressor",
  "n03757138": "metro, tube, underground, subway system, subway",
  "n03757428": "metronidazole, Flagyl",
  "n03757604": "metronome",
  "n03757723": "mews",
  "n03757925": "mexiletine, Mexitil",
  "n03758089": "mezzanine, mezzanine floor, entresol",
  "n03758220": "mezzanine, first balcony",
  "n03758334": "mezzo-relievo, mezzo-rilievo, half-relief",
  "n03758478": "mezzotint",
  "n03758614": "Mickey Finn",
  "n03758720": "miconazole, Monistat",
  "n03758894": "microbalance",
  "n03758992": "microbrewery",
  "n03759114": "microdot",
  "n03759243": "microfiche",
  "n03759432": "microfilm",
  "n03759661": "micrometer, micrometer gauge, micrometer caliper",
  "n03759795": "Micronor",
  "n03759954": "microphone, mike",
  "n03760164": "microphotometer",
  "n03760310": "microprocessor",
  "n03760671": "microscope",
  "n03760944": "microtome",
  "n03761084": "microwave, microwave oven",
  "n03761333": "microwave bomb, E-bomb",
  "n03761588": "microwave diathermy machine",
  "n03761731": "microwave linear accelerator",
  "n03761845": "midazolam, Versed",
  "n03762067": "middling",
  "n03762238": "middy, middy blouse",
  "n03762332": "midiron, two iron",
  "n03762434": "mihrab",
  "n03762602": "mihrab",
  "n03762809": "mild silver protein, Argyrol",
  "n03762982": "military hospital",
  "n03763133": "military installation",
  "n03763403": "military post, post",
  "n03763727": "military quarters",
  "n03763968": "military uniform",
  "n03764276": "military vehicle",
  "n03764606": "milk bar",
  "n03764736": "milk can",
  "n03764822": "milk float",
  "n03764995": "milking machine",
  "n03765128": "milking stool",
  "n03765269": "milk of magnesia",
  "n03765467": "milk wagon, milkwagon",
  "n03765561": "mill, grinder, milling machinery",
  "n03765934": "milldam",
  "n03766044": "miller, milling machine",
  "n03766218": "milliammeter",
  "n03766322": "millinery, woman's hat",
  "n03766508": "millinery, hat shop",
  "n03766600": "milling",
  "n03766697": "millivoltmeter",
  "n03766816": "millrace, millrun",
  "n03766935": "millstone",
  "n03767112": "millstone",
  "n03767203": "millwheel, mill wheel",
  "n03767363": "millwork",
  "n03767459": "mimeograph, mimeo, mimeograph machine, Roneo, Roneograph",
  "n03767745": "minaret",
  "n03767846": "Minato Ohashi Bridge",
  "n03767966": "mincer, mincing machine",
  "n03768132": "mine",
  "n03768346": "mine",
  "n03768683": "mine detector",
  "n03768823": "minelayer",
  "n03768916": "mineshaft",
  "n03769100": "minesweeper",
  "n03769235": "miniature, toy",
  "n03769397": "miniature, illumination",
  "n03769610": "minibar, cellaret",
  "n03769722": "minibike, motorbike",
  "n03769881": "minibus",
  "n03769967": "minicab",
  "n03770085": "minicar",
  "n03770224": "minicomputer",
  "n03770316": "ministry",
  "n03770439": "miniskirt, mini",
  "n03770520": "minisub, minisubmarine",
  "n03770679": "minivan",
  "n03770834": "miniver",
  "n03770954": "mink, mink coat",
  "n03771066": "minocycline, Minocin",
  "n03771261": "minor suit",
  "n03771443": "minor tranquilizer, minor tranquillizer, minor tranquilliser, antianxiety drug, anxiolytic, anxiolytic drug",
  "n03771820": "minoxidil, Loniten, Rogaine",
  "n03772077": "minster",
  "n03772269": "mint",
  "n03772417": "minute gun",
  "n03772584": "minute hand, big hand",
  "n03772674": "Minuteman",
  "n03772881": "miotic drug, myotic drug, miotic, myotic",
  "n03773035": "mirror",
  "n03773268": "mise en scene, stage setting, setting",
  "n03773504": "missile",
  "n03773835": "missile defense system, missile defence system",
  "n03773970": "miter, mitre",
  "n03774124": "miter, mitre",
  "n03774327": "miter box, mitre box",
  "n03774461": "miter joint, mitre joint, miter, mitre",
  "n03774673": "mithramycin, Mithracin",
  "n03774842": "mitomycin, Mutamycin",
  "n03775071": "mitten",
  "n03775199": "mixer",
  "n03775388": "mixer",
  "n03775546": "mixing bowl",
  "n03775636": "mixing faucet",
  "n03775747": "mizzen, mizen",
  "n03775847": "mizzenmast, mizenmast, mizzen, mizen",
  "n03776050": "moat, fosse",
  "n03776167": "mobcap",
  "n03776291": "mobile",
  "n03776460": "mobile home, manufactured home",
  "n03776673": "Mobius strip",
  "n03776877": "moccasin, mocassin",
  "n03776997": "mock-up",
  "n03777126": "mod con",
  "n03777283": "model, simulation",
  "n03777568": "Model T",
  "n03777754": "modem",
  "n03777961": "modernism",
  "n03778135": "Modicon",
  "n03778302": "modification",
  "n03778459": "modillion",
  "n03778600": "module",
  "n03778817": "module",
  "n03779000": "module",
  "n03779128": "mohair",
  "n03779246": "moire, watered-silk",
  "n03779370": "mold, mould, cast",
  "n03779621": "mold, mould, molding, moulding, modeling, clay sculpture",
  "n03779884": "moldboard, mouldboard",
  "n03780047": "moldboard plow, mouldboard plough",
  "n03780247": "molding, moulding, border",
  "n03780392": "molding, moulding",
  "n03780799": "moleskin",
  "n03780896": "molindone, Moban",
  "n03781055": "Molotov cocktail, petrol bomb, gasoline bomb",
  "n03781244": "monastery",
  "n03781467": "monastic habit",
  "n03781594": "moneybag",
  "n03781683": "money belt",
  "n03781787": "monitor",
  "n03782006": "monitor",
  "n03782190": "monitor, monitoring device",
  "n03782430": "Monitor",
  "n03782558": "monkey bridge",
  "n03782689": "monkey ladder",
  "n03782794": "monkey-wrench, monkey wrench",
  "n03782929": "monk's cloth",
  "n03783017": "monoamine oxidase inhibitor, MAOI",
  "n03783304": "monochrome",
  "n03783430": "monocle, eyeglass",
  "n03783575": "monofocal lens implant, monofocal IOL",
  "n03783755": "monolith",
  "n03783873": "monoplane",
  "n03783959": "monopoly board",
  "n03784053": "monorail",
  "n03784139": "monotype",
  "n03784270": "monstrance, ostensorium",
  "n03784475": "mooring, mooring line",
  "n03784666": "mooring anchor",
  "n03784793": "mooring tower, mooring mast",
  "n03784896": "Moorish arch, horseshoe arch",
  "n03785016": "moped",
  "n03785142": "mop handle",
  "n03785237": "moquette",
  "n03785361": "moreen",
  "n03785499": "morgue, mortuary, dead room",
  "n03785721": "morion, cabasset",
  "n03785843": "morning-after pill",
  "n03786096": "morning dress",
  "n03786194": "morning dress",
  "n03786313": "morning room",
  "n03786417": "morphine, morphia",
  "n03786621": "Morris chair",
  "n03786715": "mortar, howitzer, trench mortar",
  "n03786901": "mortar",
  "n03787032": "mortarboard",
  "n03787164": "mortarboard, hawk",
  "n03787308": "mortise, mortice",
  "n03787523": "mortise joint, mortise-and-tenon joint",
  "n03787759": "mosaic",
  "n03787904": "mosaic, arial mosaic, photomosaic",
  "n03788047": "mosaic",
  "n03788195": "mosque",
  "n03788365": "mosquito net",
  "n03788498": "motel",
  "n03788601": "motel room",
  "n03788703": "mothball, camphor ball",
  "n03788914": "Mother Hubbard, muumuu",
  "n03789014": "motif, motive",
  "n03789171": "motion-picture camera, movie camera, cine-camera",
  "n03789400": "motion-picture film, movie film, cine-film",
  "n03789603": "motley",
  "n03789794": "motley",
  "n03789946": "motor",
  "n03790230": "motorboat, powerboat",
  "n03790512": "motorcycle, bike",
  "n03790755": "motor hotel, motor inn, motor lodge, tourist court, court",
  "n03790953": "motorized wheelchair",
  "n03791053": "motor scooter, scooter",
  "n03791235": "motor vehicle, automotive vehicle",
  "n03792048": "mound, hill",
  "n03792334": "mound, hill, pitcher's mound",
  "n03792526": "mount, setting",
  "n03792782": "mountain bike, all-terrain bike, off-roader",
  "n03792972": "mountain tent",
  "n03793090": "mountain trail",
  "n03793186": "mounting",
  "n03793387": "mourning ring",
  "n03793489": "mouse, computer mouse",
  "n03793850": "mouse button",
  "n03793954": "mouse-tooth forceps",
  "n03794056": "mousetrap",
  "n03794136": "mousse, hair mousse, hair gel",
  "n03794291": "mousseline de sole",
  "n03794392": "mouth",
  "n03794540": "mouth hole",
  "n03794655": "mousepad, mouse mat",
  "n03794798": "mouthpiece, embouchure",
  "n03794957": "mouthpiece",
  "n03795123": "mouthpiece",
  "n03795269": "mouthpiece, gumshield",
  "n03795419": "mouthpiece",
  "n03795580": "movable barrier",
  "n03795758": "movement",
  "n03795976": "movie projector, cine projector, film projector",
  "n03796181": "moving-coil galvanometer",
  "n03796401": "moving van",
  "n03796522": "mud brick",
  "n03796605": "mudguard, splash guard, splash-guard",
  "n03796848": "mudhif",
  "n03796974": "muff",
  "n03797062": "muffle",
  "n03797182": "muffler",
  "n03797264": "mufti",
  "n03797390": "mug",
  "n03797548": "mug shot, mugshot",
  "n03797703": "mukataa",
  "n03797896": "mulch",
  "n03798061": "mule, scuff",
  "n03798165": "muller",
  "n03798268": "mullion",
  "n03798442": "multichannel recorder",
  "n03798610": "multiengine airplane, multiengine plane",
  "n03798729": "multifocal lens implant, multifocal IOL",
  "n03798982": "multiplex",
  "n03799113": "multiplexer",
  "n03799240": "multiprocessor",
  "n03799375": "multistage rocket, step rocket",
  "n03799610": "munition, ordnance, ordnance store",
  "n03799710": "mural, wall painting",
  "n03799876": "Murphy bed",
  "n03800001": "muscle relaxant",
  "n03800371": "musette, shepherd's pipe",
  "n03800485": "musette pipe",
  "n03800563": "museum",
  "n03800772": "mushroom anchor",
  "n03800933": "musical instrument, instrument",
  "n03801353": "music box, musical box",
  "n03801533": "music hall, vaudeville theater, vaudeville theatre",
  "n03801671": "music school",
  "n03801760": "music stand, music rack",
  "n03801880": "music stool, piano stool",
  "n03802007": "musket",
  "n03802228": "musket ball, ball",
  "n03802393": "muslin",
  "n03802507": "musnud",
  "n03802643": "mustache cup, moustache cup",
  "n03802800": "mustard plaster, sinapism",
  "n03802973": "mute",
  "n03803116": "muzzle loader",
  "n03803284": "muzzle",
  "n03803488": "mycomycin",
  "n03803610": "mydriatic, mydriatic drug",
  "n03803780": "myelogram",
  "n03803911": "mystification",
  "n03804048": "nabumetone, Relafen",
  "n03804211": "nacelle",
  "n03804311": "nadolol, Corgard",
  "n03804493": "nafcillin, Nafcil",
  "n03804744": "nail",
  "n03805180": "nailbrush",
  "n03805280": "nailfile",
  "n03805374": "nailhead",
  "n03805503": "nailhead",
  "n03805634": "nail hole",
  "n03805725": "nail polish, nail enamel, nail varnish",
  "n03805933": "nainsook",
  "n03806040": "nalidixic acid, NegGram",
  "n03806180": "nalorphine, Nalline",
  "n03806381": "naloxone, Narcan",
  "n03806547": "naltrexone",
  "n03806652": "nameplate",
  "n03806748": "NAND circuit, NAND gate",
  "n03806904": "nankeen",
  "n03807052": "naphazoline, Privine, Sudafed",
  "n03807334": "Napier's bones, Napier's rods",
  "n03807537": "napkin, table napkin, serviette",
  "n03807780": "napkin ring",
  "n03807895": "naproxen, Naprosyn",
  "n03808144": "naproxen sodium, Aleve, Anaprox, Aflaxen",
  "n03808424": "narcoleptic",
  "n03808564": "narcotic",
  "n03808977": "narcotic antagonist",
  "n03809211": "nard, spikenard",
  "n03809312": "narrowbody aircraft, narrow-body aircraft, narrow-body",
  "n03809456": "narrow gauge",
  "n03809603": "narrow wale",
  "n03809686": "narthex",
  "n03809802": "narthex",
  "n03809939": "nasal decongestant",
  "n03810186": "National Association of Securities Dealers Automated Quotations, NASDAQ",
  "n03810412": "nasotracheal tube",
  "n03810561": "National Baseball Hall of Fame",
  "n03810741": "National Library of Medicine, United States National Library of Medicine, U.S. National Library of Medicine",
  "n03810952": "national monument",
  "n03811163": "naumachy, naumachia",
  "n03811295": "nautilus, nuclear submarine, nuclear-powered submarine",
  "n03811444": "navigational system",
  "n03811648": "naval chart, navigational chart, pilot chart",
  "n03811847": "naval equipment",
  "n03811965": "naval gun",
  "n03812119": "naval installation, shore station",
  "n03812263": "naval missile",
  "n03812382": "naval radar",
  "n03812541": "Naval Research Laboratory, NRL",
  "n03812789": "naval tactical data system",
  "n03812924": "naval weaponry",
  "n03813078": "nave",
  "n03813176": "navigational instrument",
  "n03813369": "navigation light",
  "n03813612": "navy base",
  "n03813704": "navy yard, naval shipyard",
  "n03813834": "nearside",
  "n03813946": "nebuchadnezzar",
  "n03814112": "neck, neck opening",
  "n03814348": "neck",
  "n03814528": "neckband",
  "n03814639": "neck brace",
  "n03814727": "neckcloth, stock",
  "n03814817": "neckerchief",
  "n03814906": "necklace",
  "n03815149": "necklet",
  "n03815278": "neckline",
  "n03815482": "neckpiece",
  "n03815615": "necktie, tie",
  "n03816005": "neckwear",
  "n03816136": "needle",
  "n03816394": "needle",
  "n03816530": "needlenose pliers",
  "n03816659": "needlepoint, needlepoint embroidery",
  "n03816849": "needlework, needlecraft",
  "n03817062": "nefazodone, Serzone",
  "n03817191": "negative",
  "n03817331": "negative magnetic pole, negative pole, south-seeking pole",
  "n03817522": "negative pole",
  "n03817647": "negligee, neglige, peignoir, wrapper, housecoat",
  "n03817830": "nelfinavir, Viracept",
  "n03818001": "neolith",
  "n03818090": "neomycin, fradicin, Neobiotic",
  "n03818343": "neon lamp, neon induction lamp, neon tube",
  "n03818629": "Neosporin",
  "n03818843": "neostigmine, Prostigmin",
  "n03819047": "nephoscope",
  "n03819217": "nest",
  "n03819336": "nest",
  "n03819448": "nest egg",
  "n03819595": "net, network, mesh, meshing, meshwork",
  "n03819994": "net",
  "n03820154": "net",
  "n03820318": "net",
  "n03820474": "network",
  "n03820728": "network, electronic network",
  "n03820950": "network",
  "n03821145": "neutron bomb",
  "n03821259": "nevirapine, Viramune",
  "n03821424": "newel",
  "n03821518": "newel post, newel",
  "n03821660": "Newgate",
  "n03821898": "newmarket",
  "n03822015": "New River Gorge Bridge",
  "n03822171": "newspaper, paper",
  "n03822361": "newsroom",
  "n03822504": "newsroom",
  "n03822656": "newsstand",
  "n03822767": "Newtonian telescope, Newtonian reflector",
  "n03822951": "New York Stock Exchange, N. Y. Stock Exchange, NYSE, big board",
  "n03823111": "nib, pen nib",
  "n03823216": "niblick, nine iron",
  "n03823312": "nicad, nickel-cadmium accumulator",
  "n03823540": "nick",
  "n03823673": "nickel-iron battery, nickel-iron accumulator",
  "n03823906": "Nicol prism",
  "n03824014": "nifedipine, Procardia",
  "n03824197": "night bell",
  "n03824284": "nightcap",
  "n03824381": "nightgown, gown, nightie, night-robe, nightdress",
  "n03824589": "night latch",
  "n03824713": "night-light",
  "n03824867": "night-line",
  "n03824999": "nightshirt",
  "n03825080": "nightwear, sleepwear, nightclothes",
  "n03825271": "ninepin, skittle, skittle pin",
  "n03825442": "ninepin ball, skittle ball",
  "n03825548": "nine-spot, nine",
  "n03825673": "ninon",
  "n03825788": "nipple",
  "n03825913": "nipple shield",
  "n03826039": "niqab",
  "n03826186": "Nissen hut, Quonset hut",
  "n03826331": "nitrazepam",
  "n03826443": "nitrofurantoin, Macrodantin",
  "n03826762": "Nitrospan, Nitrostat",
  "n03826945": "nitrous oxide, laughing gas",
  "n03827107": "node, client, guest",
  "n03827284": "nog",
  "n03827420": "nogging",
  "n03827536": "noisemaker",
  "n03827716": "nomogram, nomograph",
  "n03827830": "non-dedicated file server",
  "n03828020": "nonsmoker, nonsmoking car",
  "n03828155": "non-nucleoside reverse transcriptase inhibitor, NNRTI",
  "n03828465": "nonsteroidal anti-inflammatory, nonsteroidal anti-inflammatory drug, NSAID",
  "n03829085": "nontricyclic, nontricyclic drug, nontricyclic antidepressant, nontricyclic antidepressant drug",
  "n03829340": "non-volatile storage, nonvolatile storage",
  "n03829563": "noose, running noose, slip noose",
  "n03829857": "Norfolk jacket",
  "n03829954": "noria",
  "n03830111": "Norinyl",
  "n03830278": "Norlestrin",
  "n03830448": "Nor-Q-D",
  "n03830582": "nortriptyline, Pamelor",
  "n03830835": "nose",
  "n03831062": "nose",
  "n03831203": "nosebag, feedbag",
  "n03831382": "noseband, nosepiece",
  "n03831537": "nose cone, ogive",
  "n03831757": "nose flute",
  "n03831899": "nosepiece",
  "n03832008": "nose ring",
  "n03832144": "nosewheel",
  "n03832305": "nostrum",
  "n03832405": "notch",
  "n03832673": "notebook, notebook computer",
  "n03832799": "notion",
  "n03832973": "notions counter",
  "n03833065": "novel",
  "n03833294": "novillada",
  "n03833409": "novobiocin",
  "n03833564": "nozzle, nose",
  "n03833750": "n-type semiconductor",
  "n03833907": "nuclear-powered ship",
  "n03834040": "nuclear reactor, reactor",
  "n03834472": "nuclear rocket",
  "n03834604": "nuclear weapon, atomic weapon",
  "n03834836": "nucleoside reverse transcriptase inhibitor, NRTI",
  "n03835197": "nude, nude painting",
  "n03835298": "nude, nude sculpture, nude statue",
  "n03835412": "number",
  "n03835582": "number cruncher",
  "n03835729": "numdah, numdah rug, nammad",
  "n03835853": "nunnery",
  "n03835941": "nun's habit",
  "n03836062": "nursery, baby's room",
  "n03836191": "nut",
  "n03836451": "nut and bolt",
  "n03836602": "nutcracker",
  "n03836699": "nux vomica",
  "n03836906": "nylon",
  "n03836976": "nylons, nylon stocking, rayons, rayon stocking, silk stocking",
  "n03837157": "nystatin, Mycostatin, Nystan",
  "n03837422": "oar",
  "n03837606": "oast",
  "n03837698": "oast house",
  "n03837869": "obelisk",
  "n03838024": "object ball",
  "n03838160": "objectification",
  "n03838298": "objective, objective lens, object lens, object glass",
  "n03838535": "objet d'art, art object, piece",
  "n03838748": "oblique bandage",
  "n03838899": "oboe, hautboy, hautbois",
  "n03839172": "oboe da caccia",
  "n03839276": "oboe d'amore",
  "n03839424": "observation dome",
  "n03839534": "observation station",
  "n03839671": "observatory",
  "n03839795": "obstacle",
  "n03839993": "obstruction, obstructor, obstructer, impediment, impedimenta",
  "n03840327": "obturator",
  "n03840507": "obverse",
  "n03840681": "ocarina, sweet potato",
  "n03840823": "octant",
  "n03841011": "odd-leg caliper",
  "n03841143": "odometer, hodometer, mileometer, milometer",
  "n03841290": "oeil de boeuf",
  "n03841417": "oeuvre, work, body of work",
  "n03841666": "office, business office",
  "n03842012": "office building, office block",
  "n03842156": "office furniture",
  "n03842276": "officer's mess",
  "n03842377": "off-line equipment, auxiliary equipment",
  "n03842585": "ogee, cyma reversa",
  "n03842754": "ogee arch, keel arch",
  "n03842868": "Ohio State University",
  "n03842986": "ohmmeter",
  "n03843092": "oil, oil color, oil colour",
  "n03843221": "oil burner, oil furnace",
  "n03843316": "oilcan",
  "n03843438": "oilcloth",
  "n03843555": "oil filter",
  "n03843724": "oil future, petroleum future",
  "n03843883": "oil heater, oilstove, kerosene heater, kerosine heater",
  "n03844045": "oil lamp, kerosene lamp, kerosine lamp",
  "n03844233": "oil paint",
  "n03844349": "oil painting",
  "n03844459": "oil pipeline",
  "n03844550": "oil pump",
  "n03844673": "oil refinery, petroleum refinery",
  "n03844815": "oilskin, slicker",
  "n03844965": "oil slick",
  "n03845107": "oilstone",
  "n03845190": "oil tanker, oiler, tanker, tank ship",
  "n03845360": "oil well, oiler",
  "n03845550": "ointment, unction, unguent, balm, salve",
  "n03845990": "old school tie",
  "n03846100": "olive drab",
  "n03846234": "olive drab, olive-drab uniform",
  "n03846431": "Olympian Zeus",
  "n03846677": "omelet pan, omelette pan",
  "n03846772": "omnidirectional antenna, nondirectional antenna",
  "n03846970": "omnirange, omnidirectional range, omnidirectional radio range",
  "n03847214": "one-spot",
  "n03847334": "one-way street",
  "n03847471": "onion dome",
  "n03847622": "op art",
  "n03847823": "open-air market, open-air marketplace, market square",
  "n03848033": "open circuit",
  "n03848168": "open-end wrench, tappet wrench",
  "n03848348": "opener",
  "n03848537": "open-hearth furnace",
  "n03848729": "opening",
  "n03849275": "openside plane, rabbet plane",
  "n03849412": "open sight",
  "n03849538": "open weave",
  "n03849679": "openwork",
  "n03849814": "opera, opera house",
  "n03849943": "opera cloak, opera hood",
  "n03850053": "operating microscope",
  "n03850245": "operating room, OR, operating theater, operating theatre, surgery",
  "n03850492": "operating table",
  "n03850613": "ophthalmoscope",
  "n03850746": "opiate",
  "n03850966": "opium",
  "n03851124": "opium den",
  "n03851219": "optical bench",
  "n03851341": "optical device",
  "n03851787": "optical disk, optical disc",
  "n03852031": "optical fiber, glass fiber, optical fibre, glass fibre",
  "n03852280": "optical instrument",
  "n03852544": "optical pyrometer, pyroscope",
  "n03852688": "optical telescope",
  "n03852908": "oracle",
  "n03853023": "orange grove",
  "n03853103": "orb web",
  "n03853178": "orchestra",
  "n03853291": "orchestra pit, pit",
  "n03853454": "OR circuit, OR gate",
  "n03853588": "order book",
  "n03853734": "ordinary",
  "n03853924": "ordinary, ordinary bicycle",
  "n03854065": "organ, pipe organ",
  "n03854421": "organdy, organdie",
  "n03854506": "organic light-emitting diode, OLED",
  "n03854722": "organ loft",
  "n03854815": "organ pipe, pipe, pipework",
  "n03854998": "organ stop",
  "n03855214": "organza",
  "n03855333": "oriel, oriel window",
  "n03855464": "oriflamme",
  "n03855604": "O ring",
  "n03855756": "Orlon",
  "n03855908": "orlop deck, orlop, fourth deck",
  "n03856012": "orphanage, orphans' asylum",
  "n03856148": "orphenadrine, Norflex",
  "n03856335": "orphrey",
  "n03856465": "orrery",
  "n03856728": "orthicon, image orthicon",
  "n03857026": "orthochromatic film",
  "n03857156": "orthopter, ornithopter",
  "n03857291": "orthoscope",
  "n03857430": "oscillator",
  "n03857588": "oscillogram",
  "n03857687": "oscillograph",
  "n03857828": "oscilloscope, scope, cathode-ray oscilloscope, CRO",
  "n03858085": "ossuary",
  "n03858183": "otoscope, auriscope, auroscope",
  "n03858418": "ottoman, pouf, pouffe, puff, hassock",
  "n03858533": "oubliette",
  "n03858662": "Ouija, Ouija board",
  "n03858837": "out-basket, out-tray",
  "n03859000": "outboard motor, outboard",
  "n03859170": "outboard motorboat, outboard",
  "n03859280": "outbuilding",
  "n03859495": "outerwear, overclothes",
  "n03859608": "outfall",
  "n03859717": "outfield",
  "n03859958": "outfit, getup, rig, turnout",
  "n03860234": "outfitter",
  "n03860404": "outhouse, privy, earth-closet, jakes",
  "n03860569": "outlet box",
  "n03860741": "outpost",
  "n03860882": "output, outturn, turnout",
  "n03861048": "output device",
  "n03861271": "outrigger",
  "n03861430": "outrigger canoe",
  "n03861596": "outside caliper",
  "n03861731": "outside clinch",
  "n03861842": "outside mirror",
  "n03861959": "outsider art, self-taught art, vernacular art, naive art, primitive art",
  "n03862190": "outsole",
  "n03862379": "outwork",
  "n03862531": "Oval Office",
  "n03862676": "oven",
  "n03862862": "oven thermometer",
  "n03862984": "ovenware",
  "n03863108": "overall",
  "n03863262": "overall, boilersuit, boilers suit",
  "n03863442": "overcast, overcasting",
  "n03863657": "overcoat, overcoating",
  "n03863783": "overdrive",
  "n03863923": "overgarment, outer garment",
  "n03864139": "overhand knot",
  "n03864254": "overhand stitch",
  "n03864356": "overhang",
  "n03864542": "overhead",
  "n03864692": "overhead projector",
  "n03864834": "overlay",
  "n03864994": "overload, overburden",
  "n03865171": "overload",
  "n03865288": "overmantel",
  "n03865371": "overnighter, overnight bag, overnight case",
  "n03865557": "overpass, flyover",
  "n03865704": "overprint, surprint",
  "n03865820": "override",
  "n03865949": "overshoe",
  "n03866082": "overskirt",
  "n03866176": "over-the-counter drug, over-the-counter medicine",
  "n03866350": "over-the-counter market, OTC market",
  "n03866555": "Ovocon",
  "n03866721": "ovolo, thumb, quarter round",
  "n03866908": "Ovral",
  "n03867070": "Ovrette",
  "n03867201": "Ovulen",
  "n03867355": "oxacillin",
  "n03867515": "oxaprozin, Daypro",
  "n03867675": "oxazepam, Serax",
  "n03867854": "oxbow",
  "n03868044": "Oxbridge",
  "n03868242": "oxcart",
  "n03868324": "oxeye",
  "n03868406": "oxford",
  "n03868509": "Oxford University, Oxford",
  "n03868643": "oximeter",
  "n03868763": "oxyacetylene torch",
  "n03868863": "oxygen mask",
  "n03869044": "oxyphenbutazone, Tandearil",
  "n03869222": "oxyphencyclimine, Daricon",
  "n03869389": "oxytetracycline, hydroxytetracycline, oxytetracycline hydrochloride, Terramycin",
  "n03869685": "oxytocic, oxytocic drug",
  "n03869838": "oyster bar",
  "n03869976": "oyster bed, oyster bank, oyster park",
  "n03870105": "pace car",
  "n03870290": "pacemaker, artificial pacemaker",
  "n03870461": "pacifier",
  "n03870546": "pack",
  "n03870672": "pack",
  "n03870822": "pack",
  "n03870980": "pack, face pack",
  "n03871083": "package, parcel",
  "n03871255": "packaged goods",
  "n03871371": "package store, liquor store, off-licence",
  "n03871524": "packaging",
  "n03871628": "packet",
  "n03871724": "packing box, packing case",
  "n03871860": "packinghouse, packing plant",
  "n03872016": "packinghouse",
  "n03872167": "packing needle",
  "n03872273": "packsaddle",
  "n03872386": "packthread",
  "n03872495": "pad",
  "n03872885": "pad, inkpad, inking pad, stamp pad",
  "n03873064": "padding, cushioning",
  "n03873416": "paddle, boat paddle",
  "n03873574": "paddle",
  "n03873699": "paddle",
  "n03873848": "paddle",
  "n03873996": "paddle box, paddle-box",
  "n03874138": "paddle steamer, paddle-wheeler",
  "n03874293": "paddlewheel, paddle wheel",
  "n03874487": "paddock",
  "n03874599": "padlock",
  "n03874823": "page printer, page-at-a-time printer",
  "n03874965": "pagoda",
  "n03875088": "paillasse, palliasse",
  "n03875218": "paint, pigment",
  "n03875806": "paintball",
  "n03875955": "paintball gun",
  "n03876111": "paintbox",
  "n03876231": "paintbrush",
  "n03876352": "painter",
  "n03876519": "painting, picture",
  "n03877229": "paint roller",
  "n03877351": "paisley",
  "n03877472": "pajama, pyjama, pj's, jammies",
  "n03877674": "pajama, pyjama",
  "n03877845": "palace",
  "n03878066": "palace, castle",
  "n03878211": "palace",
  "n03878294": "palanquin, palankeen",
  "n03878418": "paleolith",
  "n03878511": "palestra, palaestra",
  "n03878674": "palette, pallet",
  "n03878828": "palette knife",
  "n03878963": "palisade",
  "n03879116": "pall, shroud, cerement, winding-sheet, winding-clothes",
  "n03879336": "pallet",
  "n03879456": "pallet",
  "n03879582": "pallet",
  "n03879705": "pallette, palette",
  "n03879854": "palliative, alleviant, alleviator",
  "n03880032": "pallium",
  "n03880129": "pallium",
  "n03880323": "pan",
  "n03880531": "pan, cooking pan",
  "n03880770": "panacea, nostrum, catholicon, cure-all",
  "n03880951": "panache",
  "n03881034": "Panama Canal",
  "n03881207": "panatela, panetela, panetella",
  "n03881305": "pancake turner",
  "n03881404": "panchromatic film",
  "n03881534": "panda car",
  "n03881625": "Pandora's box",
  "n03881893": "pane, pane of glass, window glass",
  "n03882058": "panel",
  "n03882346": "panel",
  "n03882430": "panel heating",
  "n03882611": "paneling, panelling, pane",
  "n03882861": "panel light",
  "n03882960": "panhandle",
  "n03883054": "panic button",
  "n03883278": "pannier",
  "n03883385": "pannier",
  "n03883524": "pannier",
  "n03883664": "pannikin",
  "n03883773": "panopticon",
  "n03883944": "panopticon",
  "n03884072": "panorama, cyclorama, diorama",
  "n03884232": "panoramic sight",
  "n03884397": "panpipe, pandean pipe, syrinx",
  "n03884554": "pantaloon",
  "n03884639": "pantechnicon",
  "n03884778": "pantheon",
  "n03884926": "pantheon",
  "n03885028": "pantie, panty, scanty, step-in",
  "n03885194": "panting, trousering",
  "n03885293": "pant leg, trouser leg",
  "n03885410": "pantograph",
  "n03885535": "pantry, larder, buttery",
  "n03885669": "pants suit, pantsuit",
  "n03885788": "panty girdle",
  "n03885904": "pantyhose",
  "n03886053": "panzer",
  "n03886151": "papal cross",
  "n03886237": "papaverine, Kavrin",
  "n03886432": "paperback book, paper-back book, paperback, softback book, softback, soft-cover book, soft-cover",
  "n03886641": "paper chain",
  "n03886762": "paper clip, paperclip, gem clip",
  "n03886940": "paper cutter",
  "n03887066": "paper doll",
  "n03887185": "paper fastener",
  "n03887330": "paper feed",
  "n03887512": "paper mill",
  "n03887603": "paper plate",
  "n03887697": "paper towel",
  "n03887797": "paperweight",
  "n03887899": "parabolic mirror",
  "n03888022": "parabolic reflector, paraboloid reflector",
  "n03888257": "parachute, chute",
  "n03888605": "parallel bars, bars",
  "n03888808": "parallel circuit, shunt circuit",
  "n03888998": "parallel interface, parallel port",
  "n03889208": "paramagnet",
  "n03889397": "parang",
  "n03889503": "parapet, breastwork",
  "n03889626": "parapet",
  "n03889726": "parasail",
  "n03889871": "parasol, sunshade",
  "n03889974": "paregoric, camphorated tincture of opium",
  "n03890093": "parer, paring knife",
  "n03890233": "parfait glass",
  "n03890358": "pargeting, pargetting, pargetry",
  "n03890514": "pari-mutuel machine, totalizer, totaliser, totalizator, totalisator",
  "n03890713": "Paris University, University of Paris, Sorbonne",
  "n03890881": "park",
  "n03891051": "parka, windbreaker, windcheater, anorak",
  "n03891251": "park bench",
  "n03891332": "parking meter",
  "n03891538": "parlor, parlour",
  "n03891664": "parlor car, parlour car, drawing-room car, palace car, chair car",
  "n03891851": "paroxetime, Paxil",
  "n03892035": "parquet",
  "n03892178": "parquet, parquet floor",
  "n03892273": "parquet circle, parterre",
  "n03892425": "parquetry, parqueterie",
  "n03892557": "parsonage, vicarage, rectory",
  "n03892728": "Parsons table",
  "n03892891": "part, portion",
  "n03893603": "parterre",
  "n03893732": "Parthenon",
  "n03893935": "partial denture",
  "n03894051": "particle detector",
  "n03894217": "partisan, partizan",
  "n03894379": "partition, divider",
  "n03894677": "parts bin",
  "n03894762": "party favor, party favour, favor, favour",
  "n03894933": "party line",
  "n03895038": "party wall",
  "n03895170": "parvis",
  "n03895293": "passage",
  "n03895585": "passageway",
  "n03895866": "passenger car, coach, carriage",
  "n03896103": "passenger ship",
  "n03896233": "passenger train",
  "n03896419": "passenger van",
  "n03896526": "passe-partout",
  "n03896628": "passive matrix display",
  "n03896984": "passkey, passe-partout, master key, master",
  "n03897130": "pass-through",
  "n03897334": "paste-up",
  "n03897520": "pastiche",
  "n03897634": "pastry cart",
  "n03897759": "pasty",
  "n03897943": "patch",
  "n03898129": "patchcord",
  "n03898271": "patchouli, patchouly, pachouli",
  "n03898395": "patch pocket",
  "n03898500": "patchwork",
  "n03898633": "patchwork, patchwork quilt",
  "n03898787": "patent log, screw log, taffrail log",
  "n03898936": "patent medicine",
  "n03899100": "paternoster",
  "n03899328": "path",
  "n03899533": "pathway, footpath",
  "n03899612": "patina",
  "n03899768": "patio, terrace",
  "n03899933": "patisserie",
  "n03900028": "patka",
  "n03900104": "patriarchal cross",
  "n03900194": "patrol boat, patrol ship",
  "n03900301": "patty-pan",
  "n03900393": "pave",
  "n03900509": "paved surface",
  "n03900750": "pavement, paving",
  "n03900979": "pavilion, marquee",
  "n03901074": "paving stone",
  "n03901229": "pavior, paviour, paving machine",
  "n03901338": "pavis, pavise",
  "n03901548": "pawl, detent, click, dog",
  "n03901750": "pawn",
  "n03901974": "pawnbroker's shop, pawnshop, loan office",
  "n03902125": "pay-phone, pay-station",
  "n03902220": "PC board",
  "n03902381": "peacekeeper",
  "n03902482": "peach orchard",
  "n03902564": "peacock-throne",
  "n03902756": "pea jacket, peacoat",
  "n03902869": "pearl fishery",
  "n03902972": "pea shooter",
  "n03903133": "peavey, peavy, cant dog, dog hook",
  "n03903290": "pectoral, pectoral medallion",
  "n03903424": "pedal, treadle, foot pedal, foot lever",
  "n03903733": "pedal pusher, toreador pants",
  "n03903868": "pedestal, plinth, footstall",
  "n03904060": "pedestal table",
  "n03904183": "pedestrian crossing, zebra crossing",
  "n03904433": "pedicab, cycle rickshaw",
  "n03904657": "pediment",
  "n03904782": "pedometer",
  "n03904909": "peeler",
  "n03905053": "peen",
  "n03905208": "peephole, spyhole, eyehole",
  "n03905361": "peep sight",
  "n03905540": "peg, nog",
  "n03905730": "peg, pin, thole, tholepin, rowlock, oarlock",
  "n03905947": "peg",
  "n03906106": "peg, wooden leg, leg, pegleg",
  "n03906224": "pegboard",
  "n03906355": "peg top",
  "n03906463": "Pelham",
  "n03906590": "pelican crossing",
  "n03906789": "pelisse",
  "n03906894": "pelvimeter",
  "n03906997": "pen",
  "n03907227": "pen",
  "n03907475": "penal colony",
  "n03907654": "penal institution, penal facility",
  "n03907908": "penalty box",
  "n03908111": "pen-and-ink",
  "n03908204": "pencil",
  "n03908456": "pencil",
  "n03908618": "pencil box, pencil case",
  "n03908714": "pencil sharpener",
  "n03908831": "pendant, pendent",
  "n03909020": "pendant earring, drop earring, eardrop",
  "n03909160": "pendulum",
  "n03909406": "pendulum clock",
  "n03909516": "pendulum watch",
  "n03909658": "penetration bomb",
  "n03909835": "penicillamine, Cuprimine",
  "n03910033": "penicillin",
  "n03910417": "penicillinase-resistant antibiotic",
  "n03910601": "penicillin F",
  "n03910723": "penicillin G, benzylpenicillin",
  "n03910887": "penicillin O",
  "n03911039": "penicillin V, phenoxymethyl penicillin",
  "n03911251": "penicillin V potassium, Ledercillin VK",
  "n03911406": "penile implant",
  "n03911513": "penitentiary, pen",
  "n03911658": "penknife",
  "n03911767": "penlight",
  "n03911866": "pennant, pennon, streamer, waft",
  "n03911992": "pennoncel, penoncel, pennoncelle",
  "n03912105": "penny arcade",
  "n03912218": "pennywhistle, tin whistle, whistle",
  "n03912328": "pentaerythritol, Peritrate",
  "n03912496": "Pentagon",
  "n03912664": "pentazocine, Talwin",
  "n03912821": "penthouse",
  "n03912929": "pentimento",
  "n03913129": "pentobarbital sodium, pentobarbital, Nembutal, yellow jacket",
  "n03913343": "pentode",
  "n03913437": "pentoxifylline, Trental",
  "n03913702": "pentylenetetrazol, pentamethylenetetrazol, Metrazol",
  "n03913930": "peplos, peplus, peplum",
  "n03914106": "peplum",
  "n03914231": "pepper-and-salt",
  "n03914337": "pepper mill, pepper grinder",
  "n03914438": "pepper shaker, pepper box, pepper pot",
  "n03914583": "pepper spray",
  "n03914831": "percale",
  "n03914919": "perch",
  "n03915118": "percolator",
  "n03915320": "percussion cap",
  "n03915437": "percussion instrument, percussive instrument",
  "n03915900": "perforation",
  "n03916031": "perfume, essence",
  "n03916289": "perfumery",
  "n03916385": "perfumery",
  "n03916470": "perfumery",
  "n03916581": "period piece",
  "n03916720": "peripheral, computer peripheral, peripheral device",
  "n03917048": "periscope",
  "n03917198": "peristyle",
  "n03917327": "periwig, peruke",
  "n03917455": "periwinkle plant derivative",
  "n03917622": "permanent magnet, static magnet",
  "n03917814": "permanent press, durable press",
  "n03917973": "perpendicular",
  "n03918074": "perpetual motion machine",
  "n03918297": "perphenazine, Triavil",
  "n03918480": "personal computer, PC, microcomputer",
  "n03918737": "personal digital assistant, PDA, personal organizer, personal organiser, organizer, organiser",
  "n03919096": "personnel carrier",
  "n03919289": "pestle",
  "n03919430": "pestle, muller, pounder",
  "n03919702": "petard",
  "n03919808": "petcock",
  "n03919974": "Peter Pan collar",
  "n03920087": "petit point",
  "n03920181": "petit point, tent stitch",
  "n03920288": "Petri dish",
  "n03920384": "petrolatum gauze",
  "n03920497": "Petronas Towers",
  "n03920641": "pet shop",
  "n03920737": "petticoat, half-slip, underskirt",
  "n03920867": "pew, church bench",
  "n03920989": "pharmaceutical, pharmaceutic",
  "n03921209": "pharmacopoeia",
  "n03921337": "phenazopyridine, Pyridium",
  "n03921499": "phencyclidine, phencyclidine hydrochloride, PCP, angel dust",
  "n03921749": "phenelzine, Nardil",
  "n03921915": "pheniramine",
  "n03922109": "phenolphthalein",
  "n03922412": "phensuximide, Milontin",
  "n03922561": "phentolamine, Vasomax",
  "n03922722": "phenylbutazone, Butazolidin",
  "n03922863": "phenylephrine",
  "n03923004": "phenylpropanolamine",
  "n03923229": "phenyltoloxamine",
  "n03923379": "phial, vial, ampule, ampul, ampoule",
  "n03923564": "Phillips screw",
  "n03923692": "Phillips screwdriver",
  "n03923801": "phonograph album, record album",
  "n03923918": "phonograph needle, needle",
  "n03924069": "phonograph record, phonograph recording, record, disk, disc, platter",
  "n03924407": "photocathode",
  "n03924532": "photocoagulator",
  "n03924679": "photocopier",
  "n03924811": "photocopy",
  "n03924978": "photoelectric cell, photoconductive cell, photocell, electric eye, magic eye",
  "n03925226": "photograph, photo, exposure, picture, pic",
  "n03926061": "photograph album",
  "n03926148": "photographic equipment",
  "n03926412": "photographic paper, photographic material",
  "n03926575": "photographic print, print",
  "n03926757": "photolithograph",
  "n03926876": "photometer",
  "n03927091": "photomicrograph",
  "n03927201": "photomontage",
  "n03927299": "Photostat, Photostat machine",
  "n03927539": "photostat",
  "n03927652": "phrontistery",
  "n03927792": "physical pendulum, compound pendulum",
  "n03928001": "physics lab, physics laboratory",
  "n03928116": "piano, pianoforte, forte-piano",
  "n03928589": "piano action",
  "n03928814": "piano keyboard, fingerboard, clavier",
  "n03928994": "piano wire",
  "n03929091": "piccolo",
  "n03929202": "pick, pickax, pickaxe",
  "n03929443": "pick",
  "n03929660": "pick, plectrum, plectron",
  "n03929855": "pickelhaube",
  "n03929952": "picket, pale",
  "n03930087": "picket",
  "n03930229": "picket boat",
  "n03930313": "picket fence, paling",
  "n03930431": "picket ship",
  "n03930515": "pickle barrel",
  "n03930630": "pickup, pickup truck",
  "n03930777": "pickup, pick-me-up",
  "n03930946": "picot",
  "n03931044": "picture, image, icon, ikon",
  "n03931651": "picture book",
  "n03931765": "picture frame",
  "n03931885": "picture hat",
  "n03931980": "picture rail",
  "n03932080": "picture window",
  "n03932203": "piece",
  "n03932499": "piece",
  "n03932670": "piece of cloth, piece of material",
  "n03933183": "piece of leather",
  "n03933391": "pied-a-terre",
  "n03933529": "pier, wharf, wharfage, dock",
  "n03933933": "pier",
  "n03934042": "pier",
  "n03934229": "pier arch",
  "n03934311": "pier glass, pier mirror",
  "n03934416": "Pierre Laporte Bridge",
  "n03934565": "pier table",
  "n03934656": "pieta",
  "n03934780": "piezoelectric crystal",
  "n03934890": "piezometer",
  "n03934998": "pig",
  "n03935116": "pig bed, pig",
  "n03935234": "piggery, pig farm",
  "n03935335": "piggy bank, penny bank",
  "n03935450": "pike",
  "n03935698": "pike",
  "n03935789": "pikestaff",
  "n03935883": "pilaster",
  "n03936045": "pile, nap",
  "n03936269": "pile, spile, piling, stilt",
  "n03936466": "pile driver",
  "n03936568": "pill, lozenge, tablet, tab",
  "n03936764": "pill",
  "n03936895": "pill, birth control pill, contraceptive pill, oral contraceptive pill, oral contraceptive, anovulatory drug, anovulant",
  "n03937437": "pillar box",
  "n03937543": "pill bottle",
  "n03937633": "pillbox",
  "n03937751": "pillbox",
  "n03937835": "pillbox, toque, turban",
  "n03937931": "pillion",
  "n03938037": "pillory",
  "n03938244": "pillow",
  "n03938401": "pillow block",
  "n03938522": "pillow lace, bobbin lace",
  "n03938725": "pillow sham",
  "n03938838": "pilocarpine",
  "n03938951": "pilot balloon",
  "n03939062": "pilot bit",
  "n03939178": "pilot boat",
  "n03939281": "pilot burner, pilot light, pilot",
  "n03939440": "pilot cloth",
  "n03939565": "pilot engine",
  "n03939677": "pilothouse, wheelhouse",
  "n03939844": "pilot light, pilot lamp, indicator lamp",
  "n03940022": "Pimlico",
  "n03940154": "pimozide",
  "n03940256": "pin",
  "n03940713": "pin",
  "n03940894": "pin, flag",
  "n03941013": "pin, pin tumbler",
  "n03941231": "pinata",
  "n03941417": "pinball machine, pin table",
  "n03941586": "pince-nez",
  "n03941684": "pincer, pair of pincers, tweezer, pair of tweezers",
  "n03941887": "pinch bar",
  "n03942028": "pincurl clip",
  "n03942126": "pincushion",
  "n03942244": "pindolol, Visken",
  "n03942397": "pine-tar rag",
  "n03942600": "pinfold",
  "n03942692": "pinger",
  "n03942813": "ping-pong ball",
  "n03942920": "pinhead",
  "n03943010": "pinhole",
  "n03943115": "pinion",
  "n03943266": "pinnacle",
  "n03943448": "pinner",
  "n03943544": "pinpoint",
  "n03943623": "pinprick",
  "n03943714": "pinstripe",
  "n03943833": "pinstripe",
  "n03943920": "pinstripe",
  "n03944024": "pintle",
  "n03944138": "pinwheel, pinwheel wind collector",
  "n03944341": "pinwheel",
  "n03944530": "pin wrench",
  "n03944672": "pipe, pipage, piping",
  "n03945167": "pipe, tobacco pipe",
  "n03945459": "tabor pipe",
  "n03945615": "pipe",
  "n03945817": "pipe bomb",
  "n03945928": "pipe cleaner",
  "n03946076": "pipe cutter",
  "n03946162": "pipefitting, pipe fitting",
  "n03946325": "pipeline, line",
  "n03946532": "piperacillin, Pipracil",
  "n03946723": "pipe rack",
  "n03946814": "piperazine",
  "n03946933": "piperocaine, piperocaine hydrochloride, Metycaine",
  "n03947111": "pipet, pipette",
  "n03947343": "pipe vise, pipe clamp",
  "n03947466": "pipe wrench, tube wrench",
  "n03947681": "piping",
  "n03947798": "pique",
  "n03947888": "pirate, pirate ship",
  "n03948041": "piroxicam, Feldene",
  "n03948242": "piste",
  "n03948347": "piste",
  "n03948459": "pistol, handgun, side arm, shooting iron",
  "n03948830": "pistol grip",
  "n03948950": "piston, plunger",
  "n03949145": "piston ring",
  "n03949317": "piston rod",
  "n03949442": "pit, quarry, stone pit",
  "n03949663": "pit, pitfall",
  "n03949761": "pit",
  "n03949924": "pit",
  "n03950113": "pit",
  "n03950228": "pitcher, ewer",
  "n03950359": "pitchfork",
  "n03950537": "pitching wedge",
  "n03950647": "pitch pipe",
  "n03950818": "pithead",
  "n03950899": "pith hat, pith helmet, sun helmet, topee, topi",
  "n03951068": "piton",
  "n03951213": "Pitot-static tube, Pitot head, Pitot tube",
  "n03951453": "Pitot tube, Pitot",
  "n03951690": "pitprop, sprag",
  "n03951800": "pitsaw",
  "n03951971": "pivot, pin",
  "n03952150": "pivoting window",
  "n03952277": "pixel, pel, picture element",
  "n03952576": "pizzeria, pizza shop, pizza parlor",
  "n03952696": "placebo",
  "n03952886": "place mat",
  "n03953020": "place of business, business establishment",
  "n03953416": "place of worship, house of prayer, house of God, house of worship",
  "n03953743": "place setting, setting",
  "n03953901": "placket",
  "n03954012": "plain weave, taffeta weave",
  "n03954199": "plan, architectural plan",
  "n03954393": "planchet, coin blank",
  "n03954507": "planchette",
  "n03954731": "plane, carpenter's plane, woodworking plane",
  "n03955296": "plane, planer, planing machine",
  "n03955489": "plane seat",
  "n03955615": "plane table",
  "n03955809": "planetarium",
  "n03955941": "planetarium",
  "n03956157": "planetarium",
  "n03956331": "planetary gear, epicyclic gear, planet wheel, planet gear",
  "n03956531": "plank-bed",
  "n03956623": "planking",
  "n03956785": "planner",
  "n03956922": "plant, works, industrial plant",
  "n03957315": "planter",
  "n03957420": "plaster, adhesive plaster, sticking plaster",
  "n03957567": "plaster, plasterwork",
  "n03957762": "plasterboard, gypsum board",
  "n03957991": "plastering trowel",
  "n03958097": "plastic art",
  "n03958227": "plastic bag",
  "n03958338": "plastic bomb",
  "n03958448": "plastic explosive, plastique",
  "n03958630": "plastic laminate",
  "n03958752": "plastic wrap",
  "n03958910": "plastron",
  "n03959014": "plastron",
  "n03959123": "plastron",
  "n03959227": "plastron",
  "n03959350": "plat",
  "n03959485": "plate",
  "n03959701": "plate, scale, shell",
  "n03959936": "plate",
  "n03960374": "plate, collection plate",
  "n03960490": "plate",
  "n03960664": "plate",
  "n03960783": "plate, photographic plate",
  "n03960950": "plate",
  "n03961070": "plate glass, sheet glass",
  "n03961250": "plate iron",
  "n03961394": "platen",
  "n03961503": "platen",
  "n03961630": "platen",
  "n03961711": "plate rack",
  "n03961828": "plate rail",
  "n03961939": "platform",
  "n03962525": "platform, weapons platform",
  "n03962685": "platform",
  "n03962852": "platform bed",
  "n03962932": "platform rocker",
  "n03963028": "plating, metal plating",
  "n03963198": "platter",
  "n03963294": "playback",
  "n03963483": "playbox, play-box",
  "n03963645": "playground",
  "n03963813": "playhouse, wendy house",
  "n03963982": "playing card",
  "n03964495": "playpen, pen",
  "n03964611": "playsuit",
  "n03964744": "plaything, toy",
  "n03965456": "plaza, mall, center, shopping mall, shopping center, shopping centre",
  "n03965907": "pleat, plait",
  "n03966206": "plenum",
  "n03966325": "plethysmograph",
  "n03966582": "pleximeter, plessimeter",
  "n03966751": "plexor, plessor, percussor",
  "n03966976": "pliers, pair of pliers, plyers",
  "n03967270": "plimsoll",
  "n03967396": "plotter",
  "n03967562": "plow, plough",
  "n03967788": "plowshare, ploughshare, share",
  "n03967942": "plug, stopper, stopple",
  "n03968293": "plug, male plug",
  "n03968479": "plug fuse",
  "n03968581": "plughole",
  "n03968728": "plumb bob, plumb, plummet",
  "n03968886": "plumber's snake, auger",
  "n03969041": "plumbing, plumbing system",
  "n03969259": "plumbing fixture",
  "n03969510": "plumb level",
  "n03969627": "plumb line, perpendicular",
  "n03969902": "plumb rule",
  "n03969998": "plume",
  "n03970156": "plunger, plumber's helper",
  "n03970363": "plus fours",
  "n03970546": "plush",
  "n03970673": "plutonium trigger, plutonium pit",
  "n03970884": "ply",
  "n03971040": "ply",
  "n03971218": "plywood, plyboard",
  "n03971321": "pneumatic drill",
  "n03971422": "pneumatic tire, pneumatic tyre",
  "n03971771": "pneumococcal vaccine, Pneumovax",
  "n03971960": "p-n junction",
  "n03972146": "p-n-p transistor",
  "n03972372": "poacher",
  "n03972524": "pocket",
  "n03972799": "pocket",
  "n03973003": "pocket battleship",
  "n03973170": "pocketbook, pocket book, pocket edition",
  "n03973285": "pocketcomb, pocket comb",
  "n03973402": "pocket flap",
  "n03973520": "pocket-handkerchief",
  "n03973628": "pocketknife, pocket knife",
  "n03973839": "pocket watch",
  "n03973945": "pod, fuel pod",
  "n03974070": "pogo stick",
  "n03974215": "point",
  "n03974671": "point, power point",
  "n03974769": "point, gunpoint",
  "n03974915": "point-and-shoot camera",
  "n03975035": "pointed arch",
  "n03975232": "pointer",
  "n03975419": "pointillism",
  "n03975657": "pointing trowel",
  "n03975788": "point lace, needlepoint",
  "n03975926": "poker, stove poker, fire hook, salamander",
  "n03976105": "polarimeter, polariscope",
  "n03976268": "Polaroid",
  "n03976467": "Polaroid camera, Polaroid Land camera",
  "n03976657": "pole",
  "n03976960": "pole, magnetic pole",
  "n03977158": "pole",
  "n03977266": "poleax, poleaxe",
  "n03977430": "poleax, poleaxe",
  "n03977592": "police boat",
  "n03977678": "police station, police headquarters, station house",
  "n03977966": "police van, police wagon, paddy wagon, patrol wagon, wagon, black Maria",
  "n03978130": "poliovirus vaccine",
  "n03978300": "polka dot",
  "n03978421": "polling booth",
  "n03978575": "polo ball",
  "n03978686": "polo mallet, polo stick",
  "n03978815": "polonaise",
  "n03978966": "polo shirt, sport shirt",
  "n03979100": "polychrome",
  "n03979212": "polyconic projection",
  "n03979377": "polyester",
  "n03979492": "polygraph",
  "n03979717": "polymyxin",
  "n03979847": "polypropenonitrile, Acrilan",
  "n03980026": "pomade, pomatum",
  "n03980178": "pommel, saddlebow",
  "n03980332": "pommel",
  "n03980478": "pommel horse, side horse",
  "n03980734": "pompon, pom-pom",
  "n03980874": "poncho",
  "n03980986": "pongee",
  "n03981094": "poniard, bodkin",
  "n03981204": "Ponte 25 de Abril",
  "n03981340": "pontifical",
  "n03981464": "pontoon",
  "n03981566": "pontoon",
  "n03981760": "pontoon bridge, bateau bridge, floating bridge",
  "n03981924": "pony cart, ponycart, donkey cart, tub-cart",
  "n03982060": "pool",
  "n03982232": "pool ball",
  "n03982331": "poolroom",
  "n03982430": "pool table, billiard table, snooker table",
  "n03982642": "poop deck",
  "n03982767": "poor box, alms box, mite box",
  "n03982895": "poorhouse",
  "n03983068": "Pop Art",
  "n03983396": "pop bottle, soda bottle",
  "n03983499": "popgun",
  "n03983612": "poplin",
  "n03983712": "popper",
  "n03983814": "popper",
  "n03983928": "poppet, poppet valve",
  "n03984125": "pop tent",
  "n03984234": "porcelain",
  "n03984381": "porch",
  "n03984643": "porkpie, porkpie hat",
  "n03984759": "porringer",
  "n03984908": "port, embrasure, porthole",
  "n03985069": "portable",
  "n03985232": "portable computer",
  "n03985441": "portable circular saw, portable saw",
  "n03985588": "portage",
  "n03985684": "portal",
  "n03985881": "portcullis",
  "n03986071": "porte-cochere",
  "n03986224": "porte-cochere",
  "n03986355": "portfolio",
  "n03986562": "porthole",
  "n03986704": "portico",
  "n03986857": "portiere",
  "n03986949": "portmanteau, Gladstone, Gladstone bag",
  "n03987079": "portrait, portrayal",
  "n03987266": "portrait camera",
  "n03987376": "portrait lens",
  "n03987509": "positive",
  "n03987674": "positive pole, positive magnetic pole, north-seeking pole",
  "n03987865": "positive pole",
  "n03987990": "positron emission tomography scanner, PET scanner",
  "n03988170": "post",
  "n03988758": "postage meter",
  "n03988926": "post and lintel",
  "n03989074": "postbox, mailbox, letter box",
  "n03989199": "post chaise",
  "n03989349": "postern",
  "n03989447": "post exchange, PX",
  "n03989553": "posthole, post hole",
  "n03989665": "posthole digger, post-hole digger",
  "n03989777": "post horn",
  "n03989898": "posthouse, post house",
  "n03990025": "postmodernism",
  "n03990210": "Post-Office box, PO Box, POB, call box, letter box",
  "n03990385": "post road",
  "n03990474": "pot",
  "n03990834": "pot, grass, green goddess, dope, weed, gage, sess, sens, smoke, skunk, locoweed, Mary Jane",
  "n03991062": "pot, flowerpot",
  "n03991202": "potbelly, potbelly stove",
  "n03991321": "Potemkin village",
  "n03991443": "potential divider, voltage divider",
  "n03991646": "potentiometer, pot",
  "n03991837": "potentiometer",
  "n03991969": "pot farm",
  "n03992115": "potholder",
  "n03992208": "pothook",
  "n03992325": "potpourri",
  "n03992436": "potsherd",
  "n03992509": "potter's wheel",
  "n03992703": "pottery, clayware",
  "n03992888": "pottery",
  "n03992975": "pottle",
  "n03993053": "potty seat, potty chair",
  "n03993180": "pouch",
  "n03993403": "poultice, cataplasm, plaster",
  "n03993703": "pound, dog pound",
  "n03993878": "pound net",
  "n03994008": "powder",
  "n03994297": "powder and shot",
  "n03994417": "powdered mustard, dry mustard",
  "n03994614": "powder horn, powder flask",
  "n03994757": "powder keg",
  "n03994874": "powderpuff, puff",
  "n03995018": "power brake",
  "n03995265": "power cord",
  "n03995372": "power drill",
  "n03995535": "power line, power cable",
  "n03995661": "power loom",
  "n03995746": "power module",
  "n03995856": "power mower, motor mower",
  "n03996004": "power pack",
  "n03996145": "power saw, saw, sawing machine",
  "n03996416": "power shovel, excavator, digger, shovel",
  "n03996655": "power station, power plant, powerhouse",
  "n03996849": "power steering, power-assisted steering",
  "n03997027": "power system, power grid, grid",
  "n03997274": "power takeoff, PTO",
  "n03997484": "power tool",
  "n03997745": "practice range",
  "n03997875": "praetorium, pretorium",
  "n03997980": "pravastatin, Pravachol",
  "n03998194": "prayer rug, prayer mat",
  "n03998333": "prayer shawl, tallith, tallis",
  "n03998525": "prazosin, Minipress",
  "n03998673": "precipitator, electrostatic precipitator, Cottrell precipitator",
  "n03998867": "predictor",
  "n03999064": "prefab",
  "n03999160": "presbytery",
  "n03999280": "prescription drug, prescription, prescription medicine, ethical drug",
  "n03999621": "presence chamber",
  "n03999763": "presidio",
  "n03999992": "press, mechanical press",
  "n04000311": "press, printing press",
  "n04000480": "press",
  "n04000592": "press box",
  "n04000716": "press gallery",
  "n04000867": "pressing",
  "n04000998": "press of sail, press of canvas",
  "n04001132": "pressure cabin",
  "n04001265": "pressure cooker",
  "n04001397": "pressure dome",
  "n04001499": "pressure gauge, pressure gage",
  "n04001661": "pressurized water reactor, PWR",
  "n04001845": "pressure suit",
  "n04002026": "preventive, preventative, prophylactic",
  "n04002262": "pricket",
  "n04002371": "prie-dieu",
  "n04002452": "primaquine",
  "n04002536": "Primaxin",
  "n04002629": "primary coil, primary winding, primary",
  "n04002931": "primidone, Mysoline",
  "n04003110": "primitivism",
  "n04003241": "Primus stove, Primus",
  "n04003359": "Prince Albert",
  "n04003453": "Princeton University, Princeton",
  "n04003597": "print",
  "n04003856": "print",
  "n04003982": "print",
  "n04004099": "print buffer",
  "n04004210": "printed circuit",
  "n04004475": "printer, printing machine",
  "n04004767": "printer",
  "n04004990": "printer cable",
  "n04005090": "print shop, printing shop",
  "n04005197": "priory",
  "n04005340": "prism, optical prism",
  "n04005630": "prison, prison house",
  "n04005912": "prison camp, internment camp, prisoner of war camp, POW camp",
  "n04006067": "privateer",
  "n04006227": "private line",
  "n04006330": "privet hedge",
  "n04006411": "probe",
  "n04006584": "probenecid",
  "n04006727": "procaine, Ethocaine",
  "n04006953": "procaine hydrochloride, novocaine, Novocain",
  "n04007126": "procarbazine",
  "n04007239": "prochlorperazine",
  "n04007415": "proctoscope",
  "n04007510": "procyclidine, Kemadrin",
  "n04007664": "prod, goad",
  "n04007894": "product, production",
  "n04008385": "production line, assembly line, line",
  "n04008634": "projectile, missile",
  "n04008947": "projection",
  "n04009382": "projection",
  "n04009552": "projector",
  "n04009801": "projector",
  "n04009923": "prolonge",
  "n04010057": "prolonge knot, sailor's breastplate",
  "n04010205": "promenade, mall",
  "n04010348": "promethazine, Phenergan",
  "n04010566": "prompt box, prompter's box",
  "n04010779": "prompter, autocue",
  "n04010927": "prong",
  "n04011130": "proof",
  "n04011242": "prop",
  "n04011409": "propanolol, Inderal",
  "n04011609": "propellant explosive, impulse explosive",
  "n04011827": "propeller, propellor",
  "n04012084": "propeller plane",
  "n04012260": "property, prop",
  "n04012482": "propjet, turboprop, turbo-propeller plane",
  "n04012665": "proportional counter tube, proportional counter",
  "n04012852": "propoxyphene, propoxyphene hydrochloride, Darvon",
  "n04013060": "propulsion system",
  "n04013176": "proscenium, proscenium wall",
  "n04013362": "proscenium, apron, forestage",
  "n04013600": "proscenium arch",
  "n04013729": "prosthesis, prosthetic device",
  "n04013993": "protease inhibitor, PI",
  "n04014297": "protective covering, protective cover, protection",
  "n04015204": "protective garment",
  "n04015652": "proteosome vaccine, proteosome",
  "n04015786": "proton accelerator",
  "n04015908": "protractor",
  "n04016015": "protriptyline",
  "n04016133": "proving ground",
  "n04016240": "pruner, pruning hook, lopper",
  "n04016479": "pruning knife",
  "n04016576": "pruning saw",
  "n04016684": "pruning shears",
  "n04016846": "psaltery",
  "n04017019": "psilocybin, psilocin",
  "n04017137": "psychoactive drug, mind-altering drug, consciousness-altering drug, psychoactive substance",
  "n04017429": "psychotropic agent",
  "n04017571": "psychrometer",
  "n04017807": "PT boat, mosquito boat, mosquito craft, motor torpedo boat",
  "n04017993": "p-type semiconductor",
  "n04018155": "public address system, P.A. system, PA system, P.A., PA",
  "n04018399": "public house, pub, saloon, pothouse, gin mill, taphouse",
  "n04018667": "public toilet, comfort station, public convenience, convenience, public lavatory, restroom, toilet facility, wash room",
  "n04018951": "public transit",
  "n04019101": "public transport",
  "n04019335": "public works",
  "n04019541": "puck, hockey puck",
  "n04019696": "pull",
  "n04019881": "pullback, tieback",
  "n04020087": "pull chain",
  "n04020298": "pulley, pulley-block, pulley block, block",
  "n04020617": "pull-in, pull-up",
  "n04020744": "pull-off, rest area, rest stop, layby, lay-by",
  "n04020912": "Pullman, Pullman car",
  "n04021028": "pullover, slipover",
  "n04021164": "pull-through",
  "n04021362": "pulse counter",
  "n04021503": "pulse generator",
  "n04021704": "pulse timing circuit",
  "n04021798": "pump",
  "n04022332": "pump",
  "n04022434": "pump action, slide action",
  "n04022708": "pump house, pumping station",
  "n04022866": "pump room",
  "n04023021": "pump-type pliers",
  "n04023119": "pump well",
  "n04023249": "punch, puncher",
  "n04023422": "punchboard",
  "n04023695": "punch bowl",
  "n04023808": "punched card, punch card, Hollerith card",
  "n04023962": "punching bag, punch bag, punching ball, punchball",
  "n04024137": "punch pliers",
  "n04024274": "punch press",
  "n04024396": "puncture",
  "n04024576": "pung",
  "n04024676": "punkah",
  "n04024862": "punnet",
  "n04024983": "punt",
  "n04025130": "puppet",
  "n04025350": "puppet, marionette",
  "n04025508": "pup tent, shelter tent",
  "n04025633": "purdah",
  "n04025748": "purgative, cathartic, physic, aperient",
  "n04026053": "purifier",
  "n04026180": "purl, purl stitch",
  "n04026304": "purl",
  "n04026417": "purse",
  "n04026518": "purse seine",
  "n04026688": "purse string",
  "n04026813": "push-bike",
  "n04026918": "push broom",
  "n04027023": "push button, push, button",
  "n04027367": "push-button radio",
  "n04027504": "push-down storage, push-down store, stack",
  "n04027706": "pusher, zori",
  "n04027820": "put-put",
  "n04027935": "puttee",
  "n04028074": "putter, putting iron",
  "n04028221": "putty knife",
  "n04028315": "puzzle",
  "n04028472": "pyelogram",
  "n04028581": "pylon, power pylon",
  "n04028764": "pylon",
  "n04028881": "pyocyanase",
  "n04029015": "pyocyanin",
  "n04029125": "Pyramid, Great Pyramid, Pyramids of Egypt",
  "n04029416": "pyramidal tent",
  "n04029541": "pyrilamine",
  "n04029647": "pyrograph",
  "n04029734": "pyrometer",
  "n04029913": "pyrometric cone",
  "n04030054": "pyrostat",
  "n04030161": "pyx, pix",
  "n04030274": "pyx, pix, pyx chest, pix chest",
  "n04030414": "pyxis",
  "n04030518": "quad, quadrangle",
  "n04030655": "quad, space",
  "n04030846": "quadrant",
  "n04030965": "quadraphony, quadraphonic system, quadriphonic system",
  "n04031166": "quadruplicate",
  "n04031370": "Quaker gun",
  "n04031481": "quarrel",
  "n04031600": "quarter",
  "n04031770": "quarterdeck",
  "n04031884": "quartering",
  "n04032049": "quartering",
  "n04032242": "quarterlight",
  "n04032392": "quarter plate",
  "n04032509": "quarterstaff",
  "n04032603": "quartz battery, quartz mill",
  "n04032710": "quartz crystal",
  "n04032936": "quartz lamp",
  "n04033082": "quay",
  "n04033178": "Quebec Bridge",
  "n04033287": "queen",
  "n04033425": "queen",
  "n04033557": "queen post",
  "n04033647": "Queensboro Bridge",
  "n04033801": "quern",
  "n04033901": "quill, quill pen",
  "n04033995": "quilt, comforter, comfort, puff",
  "n04034262": "quilted bedspread",
  "n04034367": "quilting",
  "n04034477": "quilting",
  "n04034641": "quinacrine, quinacrine hydrochloride, mepacrine, Atabrine",
  "n04034884": "quinidine, Quinidex, Quinora",
  "n04035086": "quinine",
  "n04035231": "quipu",
  "n04035397": "quirk",
  "n04035481": "quirk bead, bead and quirk",
  "n04035634": "quirk molding, quirk moulding",
  "n04035748": "quirt",
  "n04035836": "quiver",
  "n04035912": "quoin, coign, coigne",
  "n04036004": "quoin, coign, coigne",
  "n04036155": "quoit",
  "n04036303": "QWERTY keyboard",
  "n04036494": "R-2, Mexican valium, rophy, rope, roofy, roach, forget me drug, circle",
  "n04036648": "rabato, rebato",
  "n04036776": "rabbet, rebate",
  "n04036963": "rabbet joint",
  "n04037076": "rabbit ears",
  "n04037220": "rabbit hutch",
  "n04037298": "raceabout",
  "n04037443": "racer, race car, racing car",
  "n04037625": "racetrack, racecourse, raceway, track",
  "n04037873": "raceway, race",
  "n04037964": "racing boat",
  "n04038109": "racing circuit, circuit",
  "n04038231": "racing gig",
  "n04038338": "racing skiff, single shell",
  "n04038440": "rack, stand",
  "n04038727": "rack",
  "n04039041": "rack, wheel",
  "n04039209": "rack and pinion",
  "n04039381": "racket, racquet",
  "n04039742": "racquetball",
  "n04039848": "radar, microwave radar, radio detection and ranging, radiolocation",
  "n04040247": "radial, radial tire, radial-ply tire",
  "n04040373": "radial engine, rotary engine",
  "n04040540": "radiation pyrometer",
  "n04040759": "radiator",
  "n04041069": "radiator",
  "n04041243": "radiator cap",
  "n04041408": "radiator hose",
  "n04041544": "radio, wireless",
  "n04041747": "radio antenna, radio aerial",
  "n04041930": "radio beacon, beacon",
  "n04042076": "radio chassis",
  "n04042204": "radio compass",
  "n04042358": "radiogram, radiograph, shadowgraph, skiagraph, skiagram",
  "n04042632": "radio interferometer",
  "n04042795": "radio link, link",
  "n04042985": "radiometer",
  "n04043168": "radiomicrometer",
  "n04043268": "radiopharmaceutical",
  "n04043411": "radio-phonograph, radio-gramophone",
  "n04043619": "radiophotograph, radiophoto",
  "n04043733": "radio receiver, receiving set, radio set, radio, tuner, wireless",
  "n04044119": "radio station",
  "n04044307": "radiotelegraph, radiotelegraphy, wireless telegraph, wireless telegraphy",
  "n04044498": "radiotelephone, radiophone, wireless telephone",
  "n04044716": "radio telescope, radio reflector",
  "n04044955": "radiotherapy equipment",
  "n04045085": "radio transmitter",
  "n04045255": "radome, radar dome",
  "n04045397": "raft",
  "n04045644": "rafter, balk, baulk",
  "n04045787": "raft foundation",
  "n04045941": "rag, shred, tag, tag end, tatter",
  "n04046091": "ragbag",
  "n04046172": "rag doll",
  "n04046277": "raglan",
  "n04046400": "raglan sleeve",
  "n04046590": "rail",
  "n04046810": "rail",
  "n04046974": "rail fence",
  "n04047139": "railhead",
  "n04047291": "railhead",
  "n04047401": "railing, rail",
  "n04047733": "railing",
  "n04047834": "railroad bed",
  "n04047949": "railroad flat",
  "n04048075": "railroad track, railroad, railway",
  "n04048441": "railroad tunnel",
  "n04048568": "railway, railroad, railroad line, railway line, railway system",
  "n04048979": "railway junction",
  "n04049098": "railway station, railroad station, railroad terminal, train station, train depot",
  "n04049303": "rain barrel",
  "n04049405": "raincoat, waterproof",
  "n04049585": "rain gauge, rain gage, pluviometer, udometer",
  "n04049753": "rain stick",
  "n04050066": "rake",
  "n04050313": "rake handle",
  "n04050410": "ram",
  "n04050600": "RAM disk",
  "n04050933": "ramekin, ramequin",
  "n04051068": "ramipril, Altace",
  "n04051269": "ramjet, ramjet engine, atherodyde, athodyd, flying drainpipe",
  "n04051439": "rammer",
  "n04051549": "ramp, incline",
  "n04051705": "rampant arch",
  "n04051825": "rampart, bulwark, wall",
  "n04052235": "ramrod",
  "n04052346": "ramrod",
  "n04052442": "ranch, spread, cattle ranch, cattle farm",
  "n04052658": "ranch house",
  "n04052757": "random-access memory, random access memory, random memory, RAM, read/write memory",
  "n04053218": "range",
  "n04053508": "rangefinder, range finder",
  "n04053677": "range hood",
  "n04053767": "range pole, ranging pole, flagpole",
  "n04053995": "ranitidine, Zantac",
  "n04054213": "rapid transit, mass rapid transit",
  "n04054361": "rapier, tuck",
  "n04054471": "rappee",
  "n04054566": "rariora",
  "n04054670": "rasp, wood file",
  "n04054795": "raster",
  "n04055030": "rat",
  "n04055180": "ratchet, rachet, ratch",
  "n04055447": "ratchet wheel",
  "n04055595": "rathole",
  "n04055700": "rathskeller",
  "n04055861": "ratline, ratlin",
  "n04056073": "rat-tail file",
  "n04056180": "rattan, ratan",
  "n04056289": "rattle",
  "n04056413": "rattrap",
  "n04056491": "rattrap",
  "n04056599": "ravehook",
  "n04056718": "Rayleigh disk",
  "n04056932": "rayon",
  "n04057047": "razor",
  "n04057215": "razorblade",
  "n04057323": "razor edge",
  "n04057435": "reaction-propulsion engine, reaction engine",
  "n04057673": "reaction turbine",
  "n04057846": "reactor",
  "n04057981": "reading lamp",
  "n04058096": "reading room",
  "n04058239": "read-only memory, ROM, read-only storage, fixed storage",
  "n04058486": "read-only memory chip",
  "n04058594": "readout, read-out",
  "n04058721": "read/write head, head",
  "n04058921": "ready-made",
  "n04059157": "ready-to-wear",
  "n04059298": "real storage",
  "n04059399": "reamer",
  "n04059516": "reamer, juicer, juice reamer",
  "n04059701": "rear, back",
  "n04059947": "rearview mirror",
  "n04060065": "rear window",
  "n04060198": "Reaumur thermometer",
  "n04060316": "reboxetine, Edronax",
  "n04060448": "rebozo",
  "n04060647": "receiver, receiving system",
  "n04060904": "receptacle",
  "n04061442": "receptacle",
  "n04061681": "reception desk",
  "n04061793": "reception room",
  "n04061969": "recess, niche",
  "n04062179": "reciprocating engine",
  "n04062428": "recliner, reclining chair, lounger",
  "n04062644": "reconnaissance plane",
  "n04062807": "reconnaissance vehicle, scout car",
  "n04062989": "restoration",
  "n04063154": "record changer, auto-changer, changer",
  "n04063373": "recorder, recording equipment, recording machine",
  "n04063661": "recording",
  "n04063868": "recording",
  "n04064107": "recording studio",
  "n04064213": "recording system",
  "n04064307": "record jacket",
  "n04064401": "record player, phonograph",
  "n04064747": "record sleeve, record cover",
  "n04064862": "recovery room",
  "n04064988": "recreational drug",
  "n04065132": "recreational facility, recreation facility",
  "n04065272": "recreational vehicle, RV, R.V.",
  "n04065464": "recreation room, rec room",
  "n04065632": "rectifier",
  "n04065789": "recycling bin",
  "n04065909": "recycling plant",
  "n04066023": "redbrick university",
  "n04066270": "red carpet",
  "n04066388": "redoubt",
  "n04066476": "redoubt",
  "n04066659": "reducer",
  "n04066767": "reduction gear",
  "n04066888": "reed, vibrating reed",
  "n04067143": "reed pipe",
  "n04067231": "reed stop",
  "n04067353": "reef knot, flat knot",
  "n04067472": "reel",
  "n04067658": "reel",
  "n04067818": "refectory",
  "n04067921": "refectory table",
  "n04068055": "refill",
  "n04068261": "refill",
  "n04068441": "refinery",
  "n04068601": "reflecting telescope, reflector",
  "n04068976": "reflection, reflexion",
  "n04069166": "reflectometer",
  "n04069276": "reflector",
  "n04069434": "reflex camera",
  "n04069582": "reflux condenser",
  "n04069777": "reformatory, reform school, training school",
  "n04070003": "reformer",
  "n04070207": "refracting telescope",
  "n04070415": "refractometer",
  "n04070545": "refrigeration system",
  "n04070727": "refrigerator, icebox",
  "n04070964": "refrigerator car",
  "n04071102": "refuge, sanctuary, asylum",
  "n04071263": "regalia",
  "n04071393": "regimentals",
  "n04071536": "register",
  "n04071712": "register",
  "n04071876": "register",
  "n04072193": "regulator",
  "n04072551": "rein",
  "n04072811": "relaxant",
  "n04072960": "relay, electrical relay",
  "n04073208": "release",
  "n04073425": "release, button",
  "n04073547": "relic",
  "n04073669": "relief, relievo, rilievo, embossment, sculptural relief",
  "n04073948": "religious residence, cloister",
  "n04074185": "reliquary",
  "n04074329": "remake, remaking",
  "n04074482": "remedy, curative, cure, therapeutic",
  "n04074876": "remise",
  "n04074963": "remote control, remote",
  "n04075160": "remote-control bomb",
  "n04075291": "remote terminal, link-attached terminal, remote station, link-attached station",
  "n04075468": "removable disk",
  "n04075615": "rendering",
  "n04075715": "rendering",
  "n04075813": "rep, repp",
  "n04075916": "repair shop, fix-it shop",
  "n04076052": "repeater",
  "n04076284": "repeating firearm, repeater",
  "n04076433": "repertory",
  "n04076533": "replica, replication, reproduction",
  "n04076713": "repository, monument",
  "n04076846": "representation",
  "n04077430": "reproducer",
  "n04077594": "rerebrace, upper cannon",
  "n04077734": "rescue equipment",
  "n04077889": "research center, research facility",
  "n04078002": "reseau",
  "n04078150": "reseau",
  "n04078236": "reserpine, Raudixin, Rau-Sed, Sandril, Serpasil",
  "n04078574": "reservoir",
  "n04078747": "reservoir, artificial lake, man-made lake",
  "n04078955": "reset",
  "n04079106": "reset button",
  "n04079244": "residence",
  "n04079603": "resistance pyrometer",
  "n04079746": "resistance thermometer, platinum thermometer",
  "n04079933": "resistor, resistance",
  "n04080138": "resonator",
  "n04080237": "resonator, resonant circuit",
  "n04080454": "resonator, cavity resonator, resonating chamber",
  "n04080705": "resort hotel, spa",
  "n04080833": "respirator, inhalator",
  "n04081044": "rest",
  "n04081281": "restaurant, eating house, eating place, eatery",
  "n04081699": "rest house",
  "n04081844": "restraint, constraint",
  "n04082344": "resuscitator",
  "n04082562": "retainer",
  "n04082710": "retaining wall",
  "n04082886": "reticle, reticule, graticule",
  "n04083113": "reticulation",
  "n04083309": "reticule",
  "n04083468": "restoration",
  "n04083649": "retort",
  "n04083800": "retractor",
  "n04083942": "retread, recap",
  "n04084089": "retrenchment",
  "n04084234": "retrofit",
  "n04084363": "retrorocket",
  "n04084517": "return key, return",
  "n04084682": "reverberatory furnace",
  "n04084889": "revers, revere",
  "n04085017": "reverse, reverse gear",
  "n04085181": "reverse, verso",
  "n04085365": "reverse transcriptase inhibitor",
  "n04085574": "reversible",
  "n04085746": "reversing thermometer",
  "n04085873": "revetment, revetement, stone facing",
  "n04086066": "revetment",
  "n04086150": "reviewing stand",
  "n04086273": "revolver, six-gun, six-shooter",
  "n04086446": "revolving door, revolver",
  "n04086663": "rheometer",
  "n04086794": "rheostat, variable resistor",
  "n04086937": "rhinoscope",
  "n04087126": "rib",
  "n04087290": "rib",
  "n04087432": "riband, ribband",
  "n04087524": "ribavirin, Virazole",
  "n04087709": "ribbed vault",
  "n04087826": "ribbing",
  "n04087899": "ribbon",
  "n04088058": "ribbon, typewriter ribbon",
  "n04088229": "ribbon development",
  "n04088343": "rib joint pliers",
  "n04088441": "ricer",
  "n04088593": "rickrack, ricrac",
  "n04088696": "riddle",
  "n04088797": "ride",
  "n04088982": "rider plate",
  "n04089152": "ridge, ridgepole, rooftree",
  "n04089376": "ridge rope",
  "n04089522": "riding bitt",
  "n04089666": "riding boot",
  "n04089836": "riding crop, hunting crop",
  "n04089976": "riding mower",
  "n04090064": "rifampin, Rifadin, Rimactane",
  "n04090263": "rifle",
  "n04090548": "rifle ball",
  "n04090682": "rifle butt",
  "n04090781": "rifle grenade",
  "n04090923": "rifle range",
  "n04091097": "rig",
  "n04091247": "rig, rigging",
  "n04091466": "rigger, rigger brush",
  "n04091584": "rigger",
  "n04091693": "rigging, tackle",
  "n04091839": "right field, rightfield, right",
  "n04092013": "right of way",
  "n04092168": "rigout",
  "n04092305": "rim",
  "n04092447": "rim",
  "n04092609": "ring, band",
  "n04092959": "ring",
  "n04093157": "ringlet",
  "n04093223": "rings",
  "n04093478": "ringside, ringside seat",
  "n04093625": "rink, skating rink",
  "n04093775": "riot gun",
  "n04093915": "ripcord",
  "n04094060": "ripcord",
  "n04094250": "ripping bar",
  "n04094438": "ripping chisel",
  "n04094608": "ripsaw, splitsaw",
  "n04094720": "riser",
  "n04094859": "riser, riser pipe, riser pipeline, riser main",
  "n04094983": "ritonavir, Norvir",
  "n04095109": "Ritz",
  "n04095210": "river boat",
  "n04095342": "rivet",
  "n04095577": "riveting machine, riveter, rivetter",
  "n04095725": "rivet line",
  "n04095853": "roach",
  "n04095938": "roach clip, roach holder",
  "n04096066": "road, route",
  "n04096733": "roadbed",
  "n04096848": "roadblock, barricade",
  "n04097085": "roadhouse",
  "n04097256": "road map",
  "n04097373": "roadster, runabout, two-seater",
  "n04097527": "road surface",
  "n04097622": "roadway",
  "n04097760": "roaster",
  "n04097866": "robe",
  "n04098032": "Robitussin",
  "n04098169": "robotics equipment",
  "n04098260": "Rochon prism, Wollaston prism",
  "n04098399": "rock bit, roller bit",
  "n04098513": "rocker",
  "n04098710": "rocker",
  "n04098795": "rocker, cradle",
  "n04099003": "rocker arm, valve rocker",
  "n04099175": "rocket, rocket engine",
  "n04099429": "rocket, projectile",
  "n04099649": "rocket base",
  "n04099761": "rocket range",
  "n04099854": "rock garden, rockery",
  "n04099969": "rocking chair, rocker",
  "n04100174": "rod",
  "n04100519": "rodeo",
  "n04100620": "roentgenogram, X ray, X-ray, X-ray picture, X-ray photograph",
  "n04100994": "rofecoxib, Vioxx",
  "n04101232": "roll",
  "n04101375": "roll",
  "n04101497": "roller",
  "n04101701": "roller",
  "n04101860": "roller bandage",
  "n04102037": "in-line skate",
  "n04102162": "Rollerblade",
  "n04102285": "roller blind",
  "n04102406": "roller coaster, big dipper, chute-the-chute",
  "n04102618": "roller skate",
  "n04102760": "roller towel",
  "n04102872": "roll film",
  "n04102962": "rolling hitch",
  "n04103094": "rolling mill",
  "n04103206": "rolling pin",
  "n04103364": "rolling stock",
  "n04103491": "roll of tobacco, smoke",
  "n04103665": "roll-on",
  "n04103769": "roll-on",
  "n04103918": "roll-on roll-off",
  "n04104147": "Rolodex",
  "n04104384": "Roman arch, semicircular arch",
  "n04104500": "Roman building",
  "n04104641": "Roman candle",
  "n04104770": "romper, romper suit",
  "n04104925": "rood screen",
  "n04105068": "roof",
  "n04105438": "roof",
  "n04105609": "roof garden",
  "n04105704": "roofing",
  "n04105791": "roof peak",
  "n04105893": "room",
  "n04107598": "roomette",
  "n04107743": "room light",
  "n04107870": "roost",
  "n04107984": "roost",
  "n04108137": "root cellar, cellar",
  "n04108268": "rope",
  "n04108822": "rope bridge",
  "n04108908": "rope ladder",
  "n04108999": "rope tow",
  "n04109104": "ropewalk, rope yard",
  "n04109240": "rope yarn",
  "n04109335": "rosary, prayer beads",
  "n04109471": "rose bed, bed of roses",
  "n04109599": "rose garden",
  "n04109702": "rosemaling",
  "n04109899": "rosette",
  "n04110068": "rose water",
  "n04110178": "rose window, rosette",
  "n04110281": "rosin bag",
  "n04110439": "rotary actuator, positioner",
  "n04110654": "rotary engine",
  "n04110841": "rotary press",
  "n04110955": "rotating mechanism",
  "n04111190": "rotating shaft, shaft",
  "n04111414": "rotisserie",
  "n04111531": "rotisserie",
  "n04111668": "rotor",
  "n04111962": "rotor, rotor coil",
  "n04112147": "rotor",
  "n04112252": "rotor blade, rotary wing",
  "n04112430": "rotor head, rotor shaft",
  "n04112579": "rotunda",
  "n04112654": "rotunda",
  "n04112752": "rouge, paint, blusher",
  "n04112921": "roughcast",
  "n04113038": "rouleau",
  "n04113124": "rouleau",
  "n04113194": "roulette, toothed wheel",
  "n04113316": "roulette ball",
  "n04113406": "roulette wheel, wheel",
  "n04113641": "round, unit of ammunition, one shot",
  "n04113765": "round arch",
  "n04113968": "round-bottom flask",
  "n04114069": "roundel",
  "n04114193": "rounder",
  "n04114301": "round file",
  "n04114428": "roundhouse",
  "n04114554": "Round Table, King Arthur's Round Table",
  "n04114719": "router",
  "n04114844": "router",
  "n04114996": "router plane",
  "n04115144": "rowel",
  "n04115256": "row house, town house",
  "n04115456": "rowing boat",
  "n04115542": "rowlock arch",
  "n04115670": "row of bricks",
  "n04115802": "royal",
  "n04115906": "royal brace",
  "n04115996": "royal mast",
  "n04116098": "rubber band, elastic band, elastic",
  "n04116294": "rubber boot, gum boot",
  "n04116389": "rubber bullet",
  "n04116512": "rubber eraser, rubber, pencil eraser",
  "n04116744": "rubbing",
  "n04116919": "rubbing alcohol",
  "n04117089": "rubefacient",
  "n04117216": "rudder",
  "n04117464": "rudder",
  "n04117639": "rudder blade",
  "n04117747": "rudderpost, rudderstock",
  "n04117917": "rue",
  "n04118021": "rug, carpet, carpeting",
  "n04118538": "rugby ball",
  "n04118635": "ruin",
  "n04118776": "rule, ruler",
  "n04119091": "rumble",
  "n04119230": "rumble seat",
  "n04119360": "rummer",
  "n04119478": "rumpus room, playroom, game room",
  "n04119630": "runcible spoon",
  "n04119751": "rundle, spoke, rung",
  "n04119892": "rung, round, stave",
  "n04120093": "runner",
  "n04120266": "runner",
  "n04120339": "running board",
  "n04120489": "running shoe",
  "n04120593": "running stitch",
  "n04120695": "running suit",
  "n04120842": "runway",
  "n04120998": "runway",
  "n04121142": "runway",
  "n04121228": "rushlight, rush candle",
  "n04121342": "russet",
  "n04121426": "rya, rya rug",
  "n04121511": "saber, sabre",
  "n04121728": "saber saw, jigsaw, reciprocating saw",
  "n04122011": "Sabin vaccine, oral poliovirus vaccine, OPV, trivalent live oral poliomyelitis vaccine, TOPV",
  "n04122262": "sable",
  "n04122349": "sable, sable brush, sable's hair pencil",
  "n04122492": "sable coat",
  "n04122578": "sabot, wooden shoe",
  "n04122685": "sachet",
  "n04122825": "sack, poke, paper bag, carrier bag",
  "n04123026": "sack, sacque",
  "n04123123": "sackbut",
  "n04123228": "sackcloth",
  "n04123317": "sackcloth",
  "n04123448": "sack coat",
  "n04123567": "sacking, bagging",
  "n04123740": "saddle",
  "n04123980": "saddle",
  "n04124098": "saddlebag",
  "n04124202": "saddle blanket, saddlecloth, horse blanket",
  "n04124370": "saddle oxford, saddle shoe",
  "n04124488": "saddlery",
  "n04124573": "saddle seat",
  "n04124764": "saddle soap, leather soap",
  "n04124887": "saddle stitch",
  "n04125021": "safe",
  "n04125116": "safe",
  "n04125257": "safe-deposit, safe-deposit box, safety-deposit, safety deposit box, deposit box, lockbox",
  "n04125466": "safehold",
  "n04125541": "safe house",
  "n04125692": "safety arch",
  "n04125853": "safety belt, life belt, safety harness",
  "n04126066": "safety bicycle, safety bike",
  "n04126244": "safety bolt, safety lock",
  "n04126373": "safety catch, safety lock",
  "n04126541": "safety curtain",
  "n04126659": "safety fuse",
  "n04126852": "safety lamp, Davy lamp",
  "n04126980": "safety match, book matches",
  "n04127117": "safety net",
  "n04127249": "safety pin",
  "n04127395": "safety rail, guardrail",
  "n04127521": "safety razor",
  "n04127633": "safety valve, relief valve, escape valve, escape cock, escape",
  "n04127904": "sail, canvas, canvass, sheet",
  "n04128413": "sail",
  "n04128499": "sailboat, sailing boat",
  "n04128710": "sailcloth",
  "n04128837": "sailing vessel, sailing ship",
  "n04129490": "sailing warship",
  "n04129688": "sailor cap",
  "n04129766": "sailor suit",
  "n04129872": "Saint Lawrence Seaway, St. Lawrence Seaway",
  "n04130143": "salad bar",
  "n04130257": "salad bowl",
  "n04130357": "salad fork",
  "n04130447": "salad plate, salad bowl",
  "n04130566": "salinometer",
  "n04130715": "Salk vaccine, IPV",
  "n04130907": "sallet, salade",
  "n04131015": "salon",
  "n04131113": "salon",
  "n04131208": "salon, beauty salon, beauty parlor, beauty parlour, beauty shop",
  "n04131368": "saltbox",
  "n04131499": "saltcellar",
  "n04131610": "salt mine",
  "n04131690": "saltshaker, salt shaker",
  "n04131811": "saltworks",
  "n04131929": "salver",
  "n04132056": "salvinorin",
  "n04132158": "salwar, shalwar",
  "n04132354": "Salyut",
  "n04132465": "Sam Browne belt",
  "n04132603": "samisen, shamisen",
  "n04132829": "samite",
  "n04132985": "samovar",
  "n04133114": "sampan",
  "n04133211": "sampler",
  "n04133346": "sampling station, sampler",
  "n04133497": "sanatorium, sanatarium, sanitarium",
  "n04133648": "sanctuary",
  "n04133789": "sandal",
  "n04134008": "sandbag",
  "n04134170": "sandblaster",
  "n04134339": "sandbox, sandpile, sandpit",
  "n04134523": "sandbox",
  "n04134632": "sandglass",
  "n04134862": "sand painting",
  "n04135024": "sand wedge",
  "n04135118": "sandwich board",
  "n04135315": "sanitary napkin, sanitary towel, Kotex",
  "n04135516": "Santa Fe Trail",
  "n04135710": "cling film, clingfilm, Saran Wrap",
  "n04135933": "sarcenet, sarsenet",
  "n04136045": "sarcophagus",
  "n04136161": "sari, saree",
  "n04136333": "sarong",
  "n04136510": "sash, window sash",
  "n04136676": "sash cord, sash line",
  "n04136800": "sash fastener, sash lock, window lock",
  "n04136997": "sash weight",
  "n04137089": "sash window",
  "n04137217": "satchel",
  "n04137355": "sateen",
  "n04137444": "satellite, artificial satellite, orbiter",
  "n04137773": "satellite receiver",
  "n04137897": "satellite television, satellite TV",
  "n04138131": "satellite transmitter",
  "n04138261": "satin",
  "n04138398": "satinet, satinette",
  "n04138560": "satin stitch",
  "n04138671": "satin weave",
  "n04138869": "Saturday night special",
  "n04138977": "saucepan",
  "n04139140": "saucepot",
  "n04139291": "saucer",
  "n04139395": "sauna, sweat room",
  "n04139534": "save-all",
  "n04139642": "save-all",
  "n04139748": "save-all",
  "n04139859": "savings bank, coin bank, money box, bank",
  "n04140064": "saw",
  "n04140340": "sawdust doll",
  "n04140434": "sawdust saloon",
  "n04140539": "sawed-off shotgun",
  "n04140631": "sawhorse, horse, sawbuck, buck",
  "n04140777": "sawmill",
  "n04140853": "saw set",
  "n04140976": "sawtooth",
  "n04141076": "sax, saxophone",
  "n04141198": "saxhorn",
  "n04141327": "scabbard",
  "n04141423": "scaffold",
  "n04141573": "scaffold",
  "n04141712": "scaffolding, staging",
  "n04141838": "scale",
  "n04141975": "scale, weighing machine",
  "n04142175": "scaler",
  "n04142327": "scaling ladder",
  "n04142434": "scalpel",
  "n04142549": "scan, CAT scan",
  "n04142731": "scanner, electronic scanner",
  "n04142999": "scanner",
  "n04143140": "scanner, digital scanner, image scanner",
  "n04143365": "scantling, stud",
  "n04143492": "scapular, scapulary",
  "n04143712": "scarecrow, straw man, strawman, bird-scarer, scarer",
  "n04143897": "scarf",
  "n04144241": "scarf joint, scarf",
  "n04144434": "scatter pin",
  "n04144539": "scatter rug, throw rug",
  "n04144651": "scauper, scorper",
  "n04144782": "scene, view",
  "n04145056": "scenery, scene",
  "n04145322": "scenic railway",
  "n04145417": "scheduler",
  "n04145578": "schematic, schematic drawing",
  "n04145735": "schlock, shlock, dreck",
  "n04145863": "Schmidt telescope, Schmidt camera",
  "n04146050": "school, schoolhouse",
  "n04146343": "schoolbag",
  "n04146504": "school bell",
  "n04146614": "school bus",
  "n04146721": "school crossing",
  "n04146862": "school ship, training ship",
  "n04146976": "school system",
  "n04147183": "schooner",
  "n04147291": "schooner",
  "n04147364": "science museum",
  "n04147495": "scientific instrument",
  "n04147793": "scimitar",
  "n04147916": "scintillation counter",
  "n04148054": "scissors, pair of scissors",
  "n04148285": "sclerometer",
  "n04148464": "scoinson arch, sconcheon arch",
  "n04148579": "sconce",
  "n04148703": "sconce",
  "n04148801": "sconce",
  "n04148936": "sconce",
  "n04149083": "scoop",
  "n04149208": "scoop, scoop shovel",
  "n04149374": "scooter",
  "n04149490": "scopolamine, hyoscine",
  "n04149813": "scoreboard",
  "n04149968": "scourge, flagellum",
  "n04150153": "scouring pad",
  "n04150273": "scow",
  "n04150371": "scow",
  "n04150474": "scrambler",
  "n04150668": "scrap",
  "n04150860": "scrapbook",
  "n04150980": "scraper",
  "n04151108": "scratcher",
  "n04151228": "scratchpad",
  "n04151405": "screed",
  "n04151581": "screen",
  "n04151940": "screen, cover, covert, concealment",
  "n04152387": "screen",
  "n04152593": "screen, CRT screen",
  "n04152829": "screen, silver screen, projection screen",
  "n04153025": "screen door, screen",
  "n04153330": "screening",
  "n04153436": "screen saver",
  "n04153751": "screw",
  "n04154152": "screw, screw propeller",
  "n04154340": "screw",
  "n04154565": "screwdriver",
  "n04154753": "screw eye",
  "n04154854": "screw key",
  "n04154938": "screw thread, thread",
  "n04155068": "screwtop",
  "n04155177": "screw wrench",
  "n04155310": "scribble, scrabble, doodle",
  "n04155457": "scriber, scribe, scratch awl",
  "n04155625": "scrim",
  "n04155735": "scrimshaw",
  "n04155889": "scriptorium",
  "n04156040": "scrubber",
  "n04156140": "scrub brush, scrubbing brush, scrubber",
  "n04156297": "scrub plane",
  "n04156411": "scuffer",
  "n04156591": "scuffle, scuffle hoe, Dutch hoe",
  "n04156720": "scull",
  "n04156814": "scull",
  "n04156946": "scull",
  "n04157099": "scullery",
  "n04157320": "sculpture",
  "n04157703": "scum",
  "n04157883": "scupper",
  "n04158002": "scuttle, coal scuttle",
  "n04158138": "scyphus",
  "n04158250": "scythe",
  "n04158457": "sea anchor, drogue",
  "n04158672": "seabag",
  "n04158807": "sea boat",
  "n04158956": "sea chest",
  "n04159058": "seal, stamp",
  "n04159354": "seal",
  "n04159545": "seal",
  "n04159676": "sea ladder, sea steps",
  "n04159850": "seal bomb",
  "n04160036": "sealing wax, seal",
  "n04160261": "sealskin",
  "n04160372": "seam",
  "n04160586": "seaplane, hydroplane",
  "n04160847": "searchlight",
  "n04161010": "searing iron",
  "n04161102": "Sears Tower",
  "n04161233": "seascape, waterscape",
  "n04161358": "seat",
  "n04161981": "seat",
  "n04162433": "seat",
  "n04162599": "seat",
  "n04162706": "seat belt, seatbelt",
  "n04162895": "seat cushion",
  "n04162998": "seating, seats, seating room, seating area",
  "n04163364": "seaway, sea lane, ship route, trade route",
  "n04163530": "secateurs",
  "n04163740": "secobarbital sodium, secobarbital, Seconal, red devil",
  "n04164002": "secondary coil, secondary winding, secondary",
  "n04164199": "second balcony, family circle, upper balcony, peanut gallery",
  "n04164406": "second base",
  "n04164529": "second gear, second",
  "n04164757": "second hand",
  "n04164868": "secretary, writing table, escritoire, secretaire",
  "n04164989": "section, segment",
  "n04165409": "sectional",
  "n04165551": "sector",
  "n04165675": "security blanket",
  "n04165811": "security blanket",
  "n04165945": "security system, security measure, security",
  "n04166111": "security system",
  "n04166281": "sedan, saloon",
  "n04166436": "sedan, sedan chair",
  "n04166553": "sedative, sedative drug, depressant, downer",
  "n04166841": "sedative-hypnotic, sedative-hypnotic drug",
  "n04167122": "seedbed",
  "n04167228": "seeder",
  "n04167346": "seeder",
  "n04167489": "seeker",
  "n04167661": "seersucker",
  "n04167759": "seesaw, teeter, teeter-totter, teetertotter, teeterboard, tilting board, dandle board",
  "n04168084": "segmental arch",
  "n04168199": "Segway, Segway Human Transporter, Segway HT",
  "n04168472": "seidel",
  "n04168541": "seine",
  "n04168716": "seismogram",
  "n04168840": "seismograph",
  "n04169035": "seizing",
  "n04169152": "selective-serotonin reuptake inhibitor, SSRI",
  "n04169437": "selector, selector switch",
  "n04169597": "selenium cell",
  "n04169707": "self-feeder, feeder",
  "n04169935": "self-portrait",
  "n04170037": "self-propelled vehicle",
  "n04170384": "self-registering thermometer",
  "n04170515": "self-starter",
  "n04170694": "selsyn, synchro",
  "n04170933": "selvage, selvedge",
  "n04171066": "selvage, selvedge",
  "n04171208": "semaphore",
  "n04171373": "semi-abstraction",
  "n04171459": "semiautomatic firearm",
  "n04171629": "semiautomatic pistol, semiautomatic",
  "n04171831": "semiconductor device, semiconductor unit, semiconductor",
  "n04172107": "semi-detached house",
  "n04172230": "semigloss",
  "n04172342": "semitrailer, semi",
  "n04172512": "sennit",
  "n04172607": "sensitometer",
  "n04172776": "sentry box",
  "n04172904": "separate",
  "n04173046": "septic tank",
  "n04173172": "sequence, episode",
  "n04173344": "sequencer",
  "n04173511": "sequencer, sequenator",
  "n04173698": "sequin, spangle, diamante",
  "n04173907": "serape, sarape",
  "n04174026": "serge",
  "n04174101": "serger",
  "n04174234": "serial port",
  "n04174394": "series circuit",
  "n04174500": "serpent",
  "n04174596": "serpent",
  "n04174705": "serration",
  "n04174853": "sertraline, Zoloft",
  "n04175039": "server",
  "n04175147": "server, host",
  "n04175380": "service, table service",
  "n04175574": "service club",
  "n04175669": "service door, service entrance, servant's entrance",
  "n04175859": "service station",
  "n04176068": "serving cart",
  "n04176190": "serving dish",
  "n04176295": "servo, servomechanism, servosystem",
  "n04176528": "set",
  "n04176747": "set-back, setoff, offset",
  "n04176889": "set decoration",
  "n04177041": "set gun, spring gun",
  "n04177202": "set piece",
  "n04177329": "setscrew",
  "n04177545": "setscrew",
  "n04177654": "set square",
  "n04177755": "settee",
  "n04177820": "settle, settee",
  "n04177931": "settlement house",
  "n04178062": "seven-spot, seven",
  "n04178190": "seventy-eight, 78",
  "n04178329": "Seven Wonders of the Ancient World, Seven Wonders of the World",
  "n04178668": "sewage disposal plant, disposal plant",
  "n04178784": "sewage farm",
  "n04178897": "sewage system, sewer system, sewage works",
  "n04179126": "sewer, sewerage, cloaca",
  "n04179271": "sewer main, sewer line",
  "n04179385": "sewing, stitchery",
  "n04179712": "sewing basket",
  "n04179824": "sewing kit",
  "n04179913": "sewing machine",
  "n04180063": "sewing needle",
  "n04180229": "sewing room",
  "n04180314": "sewing stitch, embroidery stitch",
  "n04180888": "sextant",
  "n04181083": "sgraffito",
  "n04181228": "shackle, bond, hamper, trammel",
  "n04181561": "shackle",
  "n04181718": "shade",
  "n04181983": "shade",
  "n04182152": "shadow box",
  "n04182322": "shaft",
  "n04182514": "shaft",
  "n04182708": "shaft, scape",
  "n04182890": "shaft",
  "n04183076": "shag",
  "n04183217": "shag rug",
  "n04183329": "shaker",
  "n04183516": "shampoo",
  "n04183663": "shank, waist",
  "n04183819": "shank",
  "n04183957": "shank",
  "n04184095": "shank, stem",
  "n04184316": "shantung",
  "n04184435": "shaper, shaping machine",
  "n04184600": "shaping tool",
  "n04184701": "shard, sherd, fragment",
  "n04184880": "sharkskin",
  "n04184957": "sharp",
  "n04185071": "sharpener",
  "n04185329": "sharpie",
  "n04185529": "Sharpie",
  "n04185656": "sharpshooter",
  "n04185804": "shaver, electric shaver, electric razor",
  "n04185946": "shaving brush",
  "n04186051": "shaving cream, shaving soap",
  "n04186268": "shaving foam",
  "n04186455": "shawl",
  "n04186624": "shawm",
  "n04186709": "shear",
  "n04186848": "shears",
  "n04187061": "sheath",
  "n04187233": "sheathing, overlay, overlayer",
  "n04187547": "shed",
  "n04187751": "sheep bell",
  "n04187885": "sheepshank",
  "n04187970": "sheepskin coat, afghan",
  "n04188064": "sheepwalk, sheeprun",
  "n04188179": "sheet, bed sheet",
  "n04188368": "sheet, tack, mainsheet, weather sheet, shroud",
  "n04188643": "sheet, flat solid",
  "n04188985": "sheet anchor, waist anchor",
  "n04189092": "sheet bend, becket bend, weaver's knot, weaver's hitch",
  "n04189282": "sheeting",
  "n04189392": "sheet iron",
  "n04189482": "sheet metal",
  "n04189651": "sheet pile, sheath pile, sheet piling",
  "n04189816": "Sheetrock",
  "n04189913": "sheet web",
  "n04190052": "shelf",
  "n04190376": "shelf bracket",
  "n04190464": "shell",
  "n04190747": "shell, case, casing",
  "n04190997": "shell, racing shell",
  "n04191150": "shellac, shellac varnish",
  "n04191346": "shell plating",
  "n04191502": "shell stitch",
  "n04191595": "shelter",
  "n04191943": "shelter",
  "n04192238": "shelter",
  "n04192361": "sheltered workshop",
  "n04192521": "Sheraton",
  "n04192698": "shield, buckler",
  "n04192858": "shield",
  "n04193035": "shielding",
  "n04193179": "shielding",
  "n04193377": "shift key, shift",
  "n04193552": "shift register",
  "n04193742": "shillelagh, shillalah",
  "n04193883": "shim",
  "n04194009": "shingle",
  "n04194127": "shin guard, shinpad",
  "n04194289": "ship",
  "n04196080": "shipboard system",
  "n04196337": "ship canal, shipway",
  "n04196502": "shipping, cargo ships, merchant marine, merchant vessels",
  "n04196677": "shipping office",
  "n04196803": "shipping room",
  "n04196925": "ship-towed long-range acoustic detection system",
  "n04197110": "shipwreck",
  "n04197235": "shipyard",
  "n04197391": "shirt",
  "n04197781": "shirt button",
  "n04197878": "shirtdress",
  "n04198015": "shirtfront",
  "n04198233": "shirting",
  "n04198355": "shirtsleeve",
  "n04198453": "shirttail",
  "n04198562": "shirtwaist, shirtwaister",
  "n04198722": "shiv",
  "n04198797": "shock absorber, shock, cushion",
  "n04199027": "shoe",
  "n04200000": "shoe",
  "n04200138": "shoe bomb",
  "n04200258": "shoebox",
  "n04200390": "shoebox",
  "n04200537": "shoehorn",
  "n04200637": "shoelace, shoe lace, shoestring, shoe string",
  "n04200800": "shoe shop, shoe-shop, shoe store",
  "n04200908": "shoetree",
  "n04201064": "shofar, shophar",
  "n04201297": "shoji",
  "n04201435": "shoofly",
  "n04201618": "shook",
  "n04201733": "shooting brake",
  "n04201845": "shooting gallery, shooting range",
  "n04201992": "shooting gallery",
  "n04202142": "shooting lodge, shooting box",
  "n04202282": "shooting stick",
  "n04202417": "shop, store",
  "n04203356": "shop bell",
  "n04203514": "shop floor",
  "n04203705": "shopfront, storefront",
  "n04203889": "shopping",
  "n04204081": "shopping bag",
  "n04204238": "shopping basket",
  "n04204347": "shopping cart",
  "n04204468": "shore, shoring",
  "n04204619": "short",
  "n04204755": "short circuit, short",
  "n04204953": "shortcut, cutoff, crosscut",
  "n04205062": "short iron",
  "n04205192": "short line",
  "n04205318": "short pants, shorts, trunks",
  "n04205505": "short sleeve",
  "n04205613": "shortwave diathermy machine",
  "n04205759": "shot, pellet",
  "n04205983": "shot",
  "n04206070": "shot",
  "n04206225": "shot glass, jigger, pony",
  "n04206356": "shotgun, scattergun",
  "n04206570": "shotgun shell",
  "n04206695": "shot hole",
  "n04206790": "shot tower",
  "n04206948": "shoulder, berm",
  "n04207151": "shoulder",
  "n04207343": "shoulder bag",
  "n04207473": "shoulder board, shoulder mark",
  "n04207596": "shouldered arch",
  "n04207763": "shoulder holster",
  "n04207903": "shoulder pad",
  "n04208065": "shoulder patch",
  "n04208210": "shovel",
  "n04208427": "shovel",
  "n04208582": "shovel hat",
  "n04208760": "showboat",
  "n04208936": "shower",
  "n04209133": "shower cap",
  "n04209239": "shower curtain",
  "n04209383": "showerhead",
  "n04209509": "shower room",
  "n04209613": "shower stall, shower bath",
  "n04209811": "showroom, salesroom, saleroom",
  "n04210012": "shrapnel",
  "n04210120": "shredder",
  "n04210288": "shrimper",
  "n04210390": "shrine",
  "n04210591": "shrink-wrap",
  "n04210723": "shroud",
  "n04210858": "shunt",
  "n04211001": "shunt, electrical shunt, bypass",
  "n04211219": "shunter",
  "n04211356": "shutter",
  "n04211528": "shutter",
  "n04211755": "shutting post",
  "n04211857": "shuttle",
  "n04211970": "shuttle",
  "n04212165": "shuttle bus",
  "n04212282": "shuttlecock, bird, birdie, shuttle",
  "n04212467": "shuttle helicopter",
  "n04212573": "siamese, siamese connection",
  "n04212810": "Sibley tent",
  "n04212943": "sick bag, sickbag",
  "n04213105": "sickbay, sick berth",
  "n04213264": "sickbed",
  "n04213353": "sickle, reaping hook, reap hook",
  "n04213530": "sickroom",
  "n04213626": "side",
  "n04214046": "sideboard",
  "n04214154": "sideboard",
  "n04214282": "sidecar",
  "n04214413": "side chapel",
  "n04214535": "side door, side entrance",
  "n04214649": "sidelight, running light",
  "n04214871": "sideline",
  "n04214958": "side pocket",
  "n04215056": "side road",
  "n04215153": "sidesaddle",
  "n04215287": "side street",
  "n04215402": "sidewalk, pavement",
  "n04215588": "sidewall",
  "n04215684": "sidewall",
  "n04215800": "side-wheeler",
  "n04215910": "sidewinder",
  "n04216013": "side yard",
  "n04216106": "siding, railroad siding, turnout, sidetrack",
  "n04216302": "Siege Perilous",
  "n04216508": "Siegfried line",
  "n04216634": "sieve, screen",
  "n04216860": "sifter",
  "n04216963": "sights",
  "n04217200": "sight setting",
  "n04217387": "sigmoidoscope, flexible sigmoidoscope",
  "n04217546": "signal box, signal tower",
  "n04217718": "signaling device",
  "n04217882": "signboard, sign",
  "n04218142": "signet",
  "n04218271": "signet ring, seal ring",
  "n04218383": "sildenafil, sildenafil citrate, Viagra",
  "n04218564": "silencer, muffler",
  "n04218773": "silencer",
  "n04218921": "silent butler",
  "n04219067": "silesia",
  "n04219185": "Silex",
  "n04219277": "silhouette",
  "n04219424": "silk",
  "n04219580": "silks",
  "n04219718": "silkscreen, silk screen print, serigraph",
  "n04220036": "sill",
  "n04220250": "silo",
  "n04220344": "silo",
  "n04220523": "silver medal, silver",
  "n04220717": "silver mine",
  "n04220805": "silver plate",
  "n04220945": "silver plate",
  "n04221076": "silverpoint",
  "n04221241": "silver protein",
  "n04221424": "silverware",
  "n04221586": "silverwork",
  "n04221673": "simple pendulum",
  "n04221823": "simulator",
  "n04221994": "simvastatin, Zocor",
  "n04222210": "single bed",
  "n04222307": "single-breasted jacket",
  "n04222470": "single-breasted suit",
  "n04222594": "single crochet, single stitch",
  "n04222723": "single prop, single-propeller plane",
  "n04222847": "single-reed instrument, single-reed woodwind",
  "n04223066": "single-rotor helicopter",
  "n04223170": "singlestick, fencing stick, backsword",
  "n04223299": "singlet, vest, undershirt",
  "n04223435": "singleton",
  "n04223580": "sink",
  "n04223778": "sinker",
  "n04223915": "sinusoidal projection, Sanson-Flamsteed projection",
  "n04224155": "siphon, syphon",
  "n04224395": "siren",
  "n04224543": "sister ship",
  "n04224671": "Sistine Chapel",
  "n04224842": "sitar",
  "n04225031": "sitz bath, hip bath",
  "n04225222": "six-pack, six pack, sixpack",
  "n04225337": "sixpenny nail",
  "n04225417": "six-spot, six",
  "n04225578": "size stick",
  "n04225729": "skate",
  "n04225987": "skateboard",
  "n04226172": "skeen arch, skene arch, scheme arch, diminished arch",
  "n04226322": "skeg",
  "n04226464": "skein",
  "n04226537": "skeleton, skeletal frame, frame, underframe",
  "n04226826": "skeleton key",
  "n04226962": "skep",
  "n04227050": "skep",
  "n04227144": "sketch, study",
  "n04227370": "sketchbook, sketch block, sketch pad",
  "n04227519": "sketcher",
  "n04227618": "sketch map",
  "n04227787": "skew arch",
  "n04227900": "skewer",
  "n04228054": "ski",
  "n04228215": "ski binding, binding",
  "n04228422": "skibob",
  "n04228581": "ski boot",
  "n04228693": "ski cap, stocking cap, toboggan cap",
  "n04228844": "skid",
  "n04229007": "skidder",
  "n04229107": "skid lid",
  "n04229195": "skidpan",
  "n04229363": "skid road",
  "n04229480": "skiff",
  "n04229620": "ski jump",
  "n04229737": "ski lodge",
  "n04229816": "ski mask",
  "n04229959": "skimmer",
  "n04230093": "skin",
  "n04230221": "skin",
  "n04230387": "ski parka, ski jacket",
  "n04230487": "ski-plane",
  "n04230603": "ski pole",
  "n04230707": "ski rack",
  "n04230808": "skirt",
  "n04231272": "skirt",
  "n04231444": "skirt of tasses",
  "n04231575": "ski run, ski trail",
  "n04231693": "ski tow, ski lift, lift",
  "n04231905": "Skivvies",
  "n04232034": "skull and crossbones",
  "n04232153": "skullcap",
  "n04232312": "skybox",
  "n04232437": "skyhook",
  "n04232543": "skyhook",
  "n04232691": "Skylab",
  "n04232800": "skylight, fanlight",
  "n04232902": "skyrocket, rocket",
  "n04233027": "skysail",
  "n04233124": "skyscraper",
  "n04233295": "skywalk",
  "n04233405": "slab",
  "n04233556": "slack",
  "n04233715": "slacks",
  "n04233832": "slack suit",
  "n04233960": "slapstick",
  "n04234160": "slasher",
  "n04234260": "slash pocket",
  "n04234455": "slat, spline",
  "n04234670": "slate",
  "n04234763": "slate pencil",
  "n04234887": "slate roof",
  "n04234969": "slave market",
  "n04235165": "slave ship",
  "n04235291": "sled, sledge, sleigh",
  "n04235646": "sleeper",
  "n04235771": "sleeper",
  "n04235860": "sleeping bag",
  "n04236001": "sleeping car, sleeper, wagon-lit",
  "n04236182": "sleeping pill, sleeping tablet, sleeping capsule, sleeping draught",
  "n04236377": "sleeve, arm",
  "n04236702": "sleeve",
  "n04236809": "sleigh bed",
  "n04236935": "sleigh bell, cascabel",
  "n04237086": "slice",
  "n04237174": "slice bar",
  "n04237287": "slicer",
  "n04237423": "slicer",
  "n04237565": "slick",
  "n04237654": "slick",
  "n04237773": "slide, lantern slide",
  "n04237924": "slide, microscope slide",
  "n04238128": "slide, playground slide, sliding board",
  "n04238321": "slide fastener, zip, zipper, zip fastener",
  "n04238617": "slide projector",
  "n04238763": "slide rule, slipstick",
  "n04238953": "slide valve",
  "n04239074": "sliding door",
  "n04239218": "sliding seat",
  "n04239333": "sliding window",
  "n04239436": "sling, scarf bandage, triangular bandage",
  "n04239639": "sling",
  "n04239786": "slingback, sling",
  "n04239900": "slinger ring",
  "n04240097": "slingshot, sling, catapult",
  "n04240327": "slip, slip of paper",
  "n04240434": "slip clutch, slip friction clutch",
  "n04240576": "slip coach, slip carriage",
  "n04240752": "slipcover",
  "n04240867": "slip-joint pliers",
  "n04241042": "slipknot",
  "n04241249": "slip-on",
  "n04241394": "slipper, carpet slipper",
  "n04241573": "slip ring",
  "n04241782": "slip stitch",
  "n04241940": "slit",
  "n04242084": "slit lamp",
  "n04242315": "slit trench",
  "n04242408": "sloop",
  "n04242587": "sloop of war",
  "n04242704": "slop basin, slop bowl",
  "n04242871": "slop chest",
  "n04243003": "slop pail, slop jar",
  "n04243142": "slops",
  "n04243251": "slopshop, slopseller's shop",
  "n04243370": "slot",
  "n04243546": "slot, one-armed bandit",
  "n04243727": "slot, expansion slot",
  "n04243941": "slot machine, coin machine",
  "n04244152": "slow lane",
  "n04244278": "slow match",
  "n04244379": "sluice, sluiceway, penstock",
  "n04244615": "sluicegate, sluice valve, floodgate, penstock, head gate, water gate",
  "n04244847": "smack",
  "n04244997": "small boat",
  "n04245218": "small computer system interface, SCSI",
  "n04245412": "small ship",
  "n04245508": "small stores",
  "n04245703": "small stuff",
  "n04245847": "smart bomb",
  "n04246060": "smelling bottle",
  "n04246156": "smelter, smeltery",
  "n04246271": "smocking",
  "n04246459": "smoke bomb, smoke grenade",
  "n04246633": "smoke hole",
  "n04246731": "smokehouse, meat house",
  "n04246855": "smoker, smoking car, smoking carriage, smoking compartment",
  "n04247011": "smoke screen, smokescreen",
  "n04247175": "smokestack, stack",
  "n04247336": "smoking mixture",
  "n04247440": "smoking room",
  "n04247544": "smoothbore",
  "n04247630": "smooth plane, smoothing plane",
  "n04247736": "snack bar, snack counter, buffet",
  "n04247876": "snaffle, snaffle bit",
  "n04248010": "snake",
  "n04248209": "snap, snap fastener, press stud",
  "n04248396": "snap brim",
  "n04248507": "snap-brim hat",
  "n04248607": "snapshot, snap, shot",
  "n04248851": "snare, gin, noose",
  "n04249029": "snare",
  "n04249204": "snare",
  "n04249415": "snare drum, snare, side drum",
  "n04249582": "snatch block",
  "n04249705": "Snellen chart",
  "n04249882": "snifter, brandy snifter, brandy glass",
  "n04250026": "snip, snippet, snipping",
  "n04250224": "sniper rifle, precision rifle",
  "n04250473": "snips, tinsnips",
  "n04250599": "Sno-cat",
  "n04250692": "snood",
  "n04250850": "snorkel, schnorkel, schnorchel, snorkel breather, breather",
  "n04251144": "snorkel",
  "n04251377": "snorter",
  "n04251560": "snowball",
  "n04251701": "snowbank, snow bank",
  "n04251791": "snowboard",
  "n04251983": "snowman",
  "n04252077": "snowmobile",
  "n04252225": "snowplow, snowplough",
  "n04252331": "snowshoe",
  "n04252560": "snowsuit",
  "n04252653": "snow thrower, snow blower",
  "n04252814": "snow tire",
  "n04252939": "snuff",
  "n04253057": "snuffbox",
  "n04253168": "snuffer",
  "n04253304": "snuffers",
  "n04253437": "soap",
  "n04253751": "soap, scoop, max, liquid ecstasy, grievous bodily harm, goop, Georgia home boy, easy lay",
  "n04253931": "soapbox",
  "n04254009": "soap dish",
  "n04254120": "soap dispenser",
  "n04254205": "soap film",
  "n04254319": "soap flakes",
  "n04254450": "soap pad",
  "n04254535": "soap powder, built-soap powder, washing powder",
  "n04254680": "soccer ball",
  "n04254777": "sock",
  "n04255034": "socket",
  "n04255163": "socket",
  "n04255346": "socket wrench",
  "n04255499": "socle",
  "n04255586": "soda can",
  "n04255670": "soda fountain",
  "n04255768": "soda fountain",
  "n04255899": "sod house, soddy, adobe house",
  "n04256033": "sodium salicylate",
  "n04256152": "sodium thiopental, phenobarbital, phenobarbitone, Luminal, purple heart",
  "n04256318": "sodium-vapor lamp, sodium-vapour lamp",
  "n04256520": "sofa, couch, lounge",
  "n04256758": "soffit",
  "n04256891": "softball, playground ball",
  "n04256993": "soft drug",
  "n04257223": "soft pedal",
  "n04257385": "soft soap, green soap",
  "n04257533": "software package, software product",
  "n04257684": "soil pipe",
  "n04257790": "solar array, solar battery, solar panel",
  "n04257986": "solar cell, photovoltaic cell",
  "n04258138": "solar dish, solar collector, solar furnace",
  "n04258333": "solar heater",
  "n04258438": "solar house",
  "n04258618": "solar telescope",
  "n04258732": "solar thermal system",
  "n04258859": "soldering iron",
  "n04258982": "sole",
  "n04259202": "solenoid",
  "n04259364": "solitaire",
  "n04259468": "solleret, sabaton",
  "n04259630": "sombrero",
  "n04259771": "sonar, echo sounder, asdic",
  "n04260192": "sonic depth finder, fathometer",
  "n04260364": "sonogram, echogram",
  "n04260589": "sonograph",
  "n04260816": "soothing syrup",
  "n04260934": "soporific, hypnotic",
  "n04261116": "sorter",
  "n04261281": "souk",
  "n04261369": "sound bow",
  "n04261506": "soundbox, body",
  "n04261638": "sound camera",
  "n04261767": "sounder",
  "n04261868": "sound film",
  "n04262010": "sound hole",
  "n04262161": "sounding board, soundboard",
  "n04262410": "sounding lead",
  "n04262530": "sounding rocket",
  "n04262678": "sound recording, audio recording, audio",
  "n04262869": "sound spectrograph",
  "n04262969": "soundtrack",
  "n04263139": "sound truck",
  "n04263257": "soup bowl",
  "n04263336": "soup ladle",
  "n04263417": "soup plate",
  "n04263502": "soupspoon, soup spoon",
  "n04263614": "source",
  "n04263760": "source of illumination",
  "n04263950": "sourdine",
  "n04264050": "sourdine, sordino",
  "n04264134": "soutache",
  "n04264233": "soutane",
  "n04264361": "sou'wester",
  "n04264485": "soybean future",
  "n04264628": "space bar",
  "n04264765": "space capsule, capsule",
  "n04264914": "spacecraft, ballistic capsule, space vehicle",
  "n04265275": "space heater",
  "n04265428": "space helmet",
  "n04265535": "Space Needle",
  "n04265668": "space probe",
  "n04265904": "space rocket",
  "n04266014": "space shuttle",
  "n04266162": "space station, space platform, space laboratory",
  "n04266375": "spacesuit",
  "n04266486": "spade",
  "n04266660": "spade",
  "n04266849": "spade bit",
  "n04266968": "spaghetti junction",
  "n04267091": "Spandau",
  "n04267165": "spandex",
  "n04267246": "spandrel, spandril",
  "n04267435": "spanker",
  "n04267577": "spar",
  "n04267870": "spare part, spare",
  "n04267985": "sparge pipe",
  "n04268142": "spark arrester, sparker",
  "n04268275": "spark arrester",
  "n04268418": "spark chamber, spark counter",
  "n04268565": "spark coil",
  "n04268680": "spark gap",
  "n04268799": "spark gap",
  "n04268969": "sparkler",
  "n04269086": "spark lever",
  "n04269270": "spark plug, sparking plug, plug",
  "n04269502": "sparkplug wrench",
  "n04269668": "spark transmitter",
  "n04269822": "spat, gaiter",
  "n04269944": "spatula",
  "n04270147": "spatula",
  "n04270276": "speakeasy",
  "n04270371": "speakerphone",
  "n04270576": "speaking trumpet",
  "n04270765": "speaking tube",
  "n04270891": "spear, lance, shaft",
  "n04271148": "spear, gig, fizgig, fishgig, lance",
  "n04271371": "spearhead, spearpoint, spear-point",
  "n04271531": "specialty store",
  "n04271640": "specific",
  "n04271793": "specimen bottle",
  "n04271891": "spectacle",
  "n04272054": "spectacles, specs, eyeglasses, glasses",
  "n04272389": "spectator pump, spectator",
  "n04272541": "spectinomycin",
  "n04272638": "spectrogram, spectrograph",
  "n04272782": "spectrograph",
  "n04272928": "spectrophotometer",
  "n04273064": "spectroscope, prism spectroscope",
  "n04273285": "speculum",
  "n04273433": "speculum",
  "n04273569": "speedboat",
  "n04273659": "speed bump",
  "n04273796": "speedometer, speed indicator",
  "n04273972": "speed skate, racing skate",
  "n04274110": "speedway",
  "n04274214": "speedway",
  "n04274309": "sperm bank",
  "n04274396": "spermicide, spermatocide",
  "n04274530": "sphere",
  "n04274686": "spherometer",
  "n04274807": "sphinx",
  "n04274985": "sphygmomanometer",
  "n04275093": "spicemill",
  "n04275175": "spice rack",
  "n04275283": "spider",
  "n04275363": "spider web, spider's web",
  "n04275548": "spider web, spider's web",
  "n04275661": "spike",
  "n04275904": "spike",
  "n04276050": "spike, spindle",
  "n04276249": "spike",
  "n04276404": "spike",
  "n04276531": "spike",
  "n04276808": "spike heel, spike, stiletto heel",
  "n04276929": "spike mike",
  "n04277034": "spillway, spill, wasteweir",
  "n04277204": "spinal anesthetic, spinal anaesthetic",
  "n04277352": "spindle",
  "n04277493": "spindle, mandrel, mandril, arbor",
  "n04277669": "spindle",
  "n04277826": "spin dryer, spin drier",
  "n04277980": "spine, backbone",
  "n04278247": "spinet",
  "n04278353": "spinet",
  "n04278447": "spinnaker",
  "n04278605": "spinner",
  "n04278751": "spinner",
  "n04278932": "spinning frame",
  "n04279063": "spinning jenny",
  "n04279172": "spinning machine",
  "n04279353": "spinning rod",
  "n04279462": "spinning wheel",
  "n04279666": "spiral, volute",
  "n04279858": "spiral bandage",
  "n04279987": "spiral ratchet screwdriver, ratchet screwdriver",
  "n04280259": "spiral spring",
  "n04280373": "spirit lamp",
  "n04280487": "spirit stove",
  "n04280603": "spirogram",
  "n04280706": "spirograph",
  "n04280845": "spirometer",
  "n04280970": "spit",
  "n04281131": "spitball",
  "n04281260": "spittoon, cuspidor",
  "n04281375": "splashboard, splasher, dashboard",
  "n04281571": "splasher",
  "n04281749": "splat",
  "n04281880": "splay",
  "n04281998": "splice, splicing",
  "n04282231": "splicer",
  "n04282379": "spline",
  "n04282494": "splint",
  "n04282664": "split",
  "n04282872": "split rail, fence rail",
  "n04282992": "Spode",
  "n04283096": "spoiler",
  "n04283255": "spoiler",
  "n04283378": "spoke, wheel spoke, radius",
  "n04283585": "spokeshave",
  "n04283784": "sponge cloth",
  "n04283905": "sponge mop",
  "n04284002": "spoon",
  "n04284341": "spoon",
  "n04284438": "Spork",
  "n04284572": "sporran",
  "n04284735": "sporting goods",
  "n04284869": "sport kite, stunt kite",
  "n04285008": "sports car, sport car",
  "n04285146": "sports equipment",
  "n04285622": "sports implement",
  "n04285803": "sportswear, athletic wear, activewear",
  "n04285965": "sport utility, sport utility vehicle, S.U.V., SUV",
  "n04286128": "spot",
  "n04286307": "spot",
  "n04286575": "spotlight, spot",
  "n04286796": "spot market",
  "n04286960": "spot weld, spot-weld",
  "n04287153": "spout",
  "n04287351": "spouter",
  "n04287451": "sprag",
  "n04287610": "spray",
  "n04287747": "spray gun",
  "n04287898": "spray paint",
  "n04287986": "spreader",
  "n04288165": "sprig",
  "n04288272": "spring",
  "n04288533": "spring balance, spring scale",
  "n04288673": "springboard",
  "n04288784": "springer, impost",
  "n04288919": "spring mattress",
  "n04289027": "sprinkler",
  "n04289195": "sprinkler system",
  "n04289449": "sprit",
  "n04289576": "spritsail",
  "n04289690": "sprocket, sprocket wheel",
  "n04289827": "sprocket",
  "n04289964": "spud, stump spud",
  "n04290079": "spun yarn",
  "n04290259": "spur, gad",
  "n04290507": "spur gear, spur wheel",
  "n04290615": "sputnik",
  "n04290762": "spy satellite",
  "n04290875": "squab",
  "n04290945": "squad room",
  "n04291069": "squad room",
  "n04291242": "square",
  "n04291511": "square",
  "n04291759": "square knot",
  "n04291912": "square nut",
  "n04291992": "square-rigger",
  "n04292080": "square sail",
  "n04292221": "squash ball",
  "n04292313": "squash court",
  "n04292414": "squash racket, squash racquet, bat",
  "n04292572": "squawk box, squawker, intercom speaker",
  "n04292733": "squeaker",
  "n04292921": "squeegee",
  "n04293119": "squeezer",
  "n04293258": "squelch circuit, squelch, squelcher",
  "n04293450": "squib",
  "n04293608": "saquinavir, Invirase",
  "n04293744": "squinch",
  "n04293902": "squirrel cage",
  "n04294041": "stabile",
  "n04294212": "stabilizer, stabiliser",
  "n04294426": "stabilizer",
  "n04294614": "stabilizer bar, anti-sway bar",
  "n04294879": "stable, stalls, horse barn",
  "n04295081": "stable gear, saddlery, tack",
  "n04295353": "stabling",
  "n04295475": "stacked heel",
  "n04295571": "stacks",
  "n04295777": "staddle",
  "n04295881": "stadium, bowl, arena, sports stadium",
  "n04296261": "staff",
  "n04296562": "stage",
  "n04296949": "stage, microscope stage",
  "n04297098": "stagecoach, stage",
  "n04297327": "stage door",
  "n04297476": "stage set, set",
  "n04297750": "stained-glass window",
  "n04297847": "stair-carpet",
  "n04297944": "stairhead",
  "n04298053": "stair-rod",
  "n04298171": "stairs, steps",
  "n04298308": "stairway, staircase",
  "n04298661": "stairwell",
  "n04298765": "stake",
  "n04298906": "stake",
  "n04299066": "stalking-horse",
  "n04299215": "stall, stand, sales booth",
  "n04299370": "stall",
  "n04299526": "stall bar",
  "n04299699": "stall",
  "n04299828": "stammel",
  "n04299963": "stamp",
  "n04300080": "stamp, pestle",
  "n04300281": "stamp album",
  "n04300358": "stamp mill, stamping mill",
  "n04300509": "stamping machine, stamper",
  "n04300643": "stanchion",
  "n04300741": "stand",
  "n04301000": "stand",
  "n04301140": "standard, banner",
  "n04301242": "standard",
  "n04301474": "standard cell",
  "n04301626": "standard gauge",
  "n04301760": "standard transmission, stick shift",
  "n04301919": "standby",
  "n04302034": "standee",
  "n04302200": "standing press",
  "n04302334": "standing room",
  "n04302527": "standpipe",
  "n04302598": "St. Andrew's cross, saltire",
  "n04302740": "Stanford University, Stanford",
  "n04302863": "stanhope",
  "n04302988": "Stanley Steamer",
  "n04303095": "staple",
  "n04303258": "staple",
  "n04303357": "staple gun, staplegun, tacker",
  "n04303497": "stapler, stapling machine",
  "n04303672": "starboard",
  "n04303852": "star drill",
  "n04304084": "Stars and Bars, Confederate flag",
  "n04304215": "starship, spaceship",
  "n04304375": "starter, starter motor, starting motor",
  "n04304557": "starting block",
  "n04304680": "starting gate, starting stall",
  "n04304812": "stash house",
  "n04305016": "Stassano furnace, electric-arc furnace",
  "n04305210": "Statehouse",
  "n04305323": "stately home",
  "n04305471": "state prison",
  "n04305572": "stateroom",
  "n04305641": "static line",
  "n04305947": "static tube",
  "n04306080": "station",
  "n04306454": "Station of the Cross",
  "n04306592": "stator, stator coil",
  "n04306847": "statue",
  "n04307106": "Statue of Liberty",
  "n04307269": "stave, lag",
  "n04307419": "stay",
  "n04307619": "stay",
  "n04307767": "staysail",
  "n04307878": "steakhouse, chophouse",
  "n04307986": "steak knife",
  "n04308084": "stealth aircraft",
  "n04308273": "stealth bomber",
  "n04308397": "stealth fighter",
  "n04308583": "steam bath, steam room, vapor bath, vapour bath",
  "n04308807": "steamboat",
  "n04308915": "steam chest",
  "n04309049": "steam engine",
  "n04309348": "steamer, steamship",
  "n04309548": "steamer",
  "n04309682": "steam heat, steam heating",
  "n04309833": "steam iron",
  "n04309925": "steam line, steam pipe",
  "n04310018": "steam locomotive",
  "n04310157": "steamroller, road roller",
  "n04310347": "steamship company, steamship line",
  "n04310507": "steam shovel",
  "n04310604": "steam turbine",
  "n04310721": "steam whistle",
  "n04310904": "steel",
  "n04311004": "steel arch bridge",
  "n04311174": "steel drum",
  "n04311397": "steel engraving",
  "n04311485": "steel engraving",
  "n04311595": "steel mill, steelworks, steel plant, steel factory",
  "n04311799": "steel plate",
  "n04311873": "steel trap",
  "n04312020": "steel-wool pad",
  "n04312154": "steelyard, lever scale, beam scale",
  "n04312311": "steeper",
  "n04312432": "steeple, spire",
  "n04312654": "steerage",
  "n04312756": "steering gear",
  "n04312916": "steering linkage",
  "n04313220": "steering system, steering mechanism",
  "n04313503": "steering wheel, wheel",
  "n04313628": "stele, stela",
  "n04313771": "stem",
  "n04313867": "stemmer",
  "n04313986": "stemmer",
  "n04314107": "stem-winder",
  "n04314216": "stencil",
  "n04314522": "Sten gun",
  "n04314632": "stenograph",
  "n04314735": "stent",
  "n04314914": "step, stair",
  "n04315185": "step",
  "n04315342": "step-down transformer",
  "n04315444": "stepper, stepping motor",
  "n04315599": "step ladder, stepladder",
  "n04315713": "step stool",
  "n04315828": "step-up transformer",
  "n04315948": "stereo, stereophony, stereo system, stereophonic system",
  "n04316275": "stereo, stereoscopic picture, stereoscopic photograph",
  "n04316498": "stereoscope",
  "n04316646": "stern, after part, quarter, poop, tail",
  "n04316815": "stern chaser",
  "n04316924": "sternpost",
  "n04317063": "sternwheeler",
  "n04317175": "stethoscope",
  "n04317325": "stewing pan, stewpan",
  "n04317420": "stick",
  "n04317833": "stick",
  "n04317976": "stick, control stick, joystick",
  "n04318131": "stick",
  "n04318384": "stick figure",
  "n04318551": "stick horse",
  "n04318690": "stickpin",
  "n04318787": "stile",
  "n04318892": "stiletto",
  "n04318982": "still",
  "n04319248": "still",
  "n04319432": "still life",
  "n04319545": "stillroom, still room",
  "n04319774": "Stillson wrench",
  "n04319937": "stilt",
  "n04320126": "stimulant, stimulant drug, excitant",
  "n04320405": "Stinger",
  "n04320598": "stink bomb, stench bomb",
  "n04320731": "stinker",
  "n04320871": "stirrer",
  "n04320973": "stirrup, stirrup iron",
  "n04321121": "stirrup pump",
  "n04321238": "stitch",
  "n04321453": "stob",
  "n04321534": "stock, inventory",
  "n04321804": "stock",
  "n04322026": "stock, gunstock",
  "n04322387": "stock",
  "n04322531": "stockade",
  "n04322692": "stockcar",
  "n04322801": "stock car",
  "n04322924": "stock car",
  "n04323026": "stock exchange, stock market, securities market",
  "n04323519": "stockinet, stockinette",
  "n04323645": "stockinette stitch",
  "n04323819": "stocking",
  "n04324120": "stock-in-trade",
  "n04324297": "stockpot",
  "n04324387": "stockroom, stock room",
  "n04324515": "stocks",
  "n04324741": "stocks",
  "n04324910": "stocks",
  "n04325041": "stock saddle, Western saddle",
  "n04325208": "stockyard",
  "n04325335": "stogy, stogie",
  "n04325409": "stokehold, stokehole, fireroom",
  "n04325592": "stoker",
  "n04325704": "stole",
  "n04325804": "stomacher",
  "n04325968": "stomach pump",
  "n04326084": "stone",
  "n04326547": "stone wall",
  "n04326676": "stoneware",
  "n04326799": "stonework",
  "n04326896": "stool",
  "n04327110": "stool pigeon",
  "n04327204": "stoop, stoep",
  "n04327324": "stop",
  "n04327544": "stop bath, short-stop, short-stop bath",
  "n04327682": "stopcock, cock, turncock",
  "n04327825": "stopper",
  "n04328054": "stopper knot",
  "n04328186": "stopwatch, stop watch",
  "n04328329": "storage battery, accumulator",
  "n04328580": "storage cell, secondary cell",
  "n04328703": "storage ring",
  "n04328946": "storage space",
  "n04329190": "storehouse, depot, entrepot, storage, store",
  "n04329477": "storeroom, storage room, stowage",
  "n04329681": "storm cellar, cyclone cellar, tornado cellar",
  "n04329834": "storm door",
  "n04329958": "storm window, storm sash",
  "n04330109": "stoup, stoop",
  "n04330189": "stoup",
  "n04330267": "stove",
  "n04330340": "stove, kitchen stove, range, kitchen range, cooking stove",
  "n04330669": "stove bolt",
  "n04330746": "stovepipe",
  "n04330896": "stovepipe iron",
  "n04330998": "Stradavarius, Strad",
  "n04331126": "straightaway, straight",
  "n04331277": "straight chair, side chair",
  "n04331443": "straightedge",
  "n04331639": "straightener",
  "n04331765": "straight flute, straight-fluted drill",
  "n04331892": "straight pin",
  "n04332074": "straight razor",
  "n04332243": "strainer",
  "n04332466": "strain gauge, strain gage",
  "n04332580": "straitjacket, straightjacket",
  "n04332783": "strand",
  "n04332987": "strap",
  "n04333129": "strap",
  "n04333500": "strap, shoulder strap",
  "n04333709": "strap",
  "n04333869": "strap hinge, joint hinge",
  "n04334105": "strapless",
  "n04334232": "straw, drinking straw",
  "n04334365": "streamer fly",
  "n04334504": "streamliner",
  "n04334599": "street",
  "n04335209": "street",
  "n04335435": "streetcar, tram, tramcar, trolley, trolley car",
  "n04335693": "street clothes",
  "n04335886": "streetlight, street lamp",
  "n04336034": "strengthener, reinforcement",
  "n04336380": "streptomycin",
  "n04336533": "streptothricin",
  "n04336645": "stretch",
  "n04336792": "stretcher",
  "n04336996": "stretcher",
  "n04337157": "stretcher",
  "n04337287": "stretch pants",
  "n04337400": "strickle",
  "n04337503": "strickle",
  "n04337650": "strickle",
  "n04337740": "striker",
  "n04337974": "string, twine",
  "n04338143": "string",
  "n04338359": "string",
  "n04338517": "stringed instrument",
  "n04338963": "stringer",
  "n04339062": "stringer",
  "n04339191": "string tie",
  "n04339291": "strip, slip",
  "n04339638": "strip",
  "n04339879": "strip lighting",
  "n04340019": "strip mall",
  "n04340231": "strip mine",
  "n04340378": "stripper well, stripper",
  "n04340521": "stroboscope, strobe, strobe light",
  "n04340750": "strongbox, deedbox",
  "n04340935": "stronghold, fastness",
  "n04341133": "strongroom",
  "n04341288": "strop",
  "n04341414": "structural member",
  "n04341686": "structure, construction",
  "n04343148": "strut",
  "n04343272": "stub nail",
  "n04343346": "stud, rivet",
  "n04343511": "student center",
  "n04343630": "student lamp",
  "n04343740": "student union",
  "n04343919": "stud farm",
  "n04344003": "stud finder",
  "n04344246": "studio",
  "n04344544": "studio",
  "n04344734": "studio apartment, studio",
  "n04344873": "studio couch, day bed",
  "n04345028": "study",
  "n04345201": "study hall",
  "n04345288": "stuff",
  "n04345458": "stuffing",
  "n04345595": "stuffing box, packing box",
  "n04345787": "stuffing nut, packing nut",
  "n04345915": "stumbling block",
  "n04346003": "stump",
  "n04346157": "stun gun, stun baton",
  "n04346328": "stupa, tope",
  "n04346428": "sty, pigsty, pigpen",
  "n04346511": "stylus, style",
  "n04346679": "stylus",
  "n04346855": "sub-assembly",
  "n04347004": "subbase",
  "n04347119": "subcompact, subcompact car",
  "n04347225": "subject, content, depicted object",
  "n04347519": "submachine gun",
  "n04347754": "submarine, pigboat, sub, U-boat",
  "n04348070": "submarine torpedo",
  "n04348184": "submersible, submersible warship",
  "n04348359": "submersible",
  "n04348548": "subsection, subdivision",
  "n04348702": "substation",
  "n04348854": "subtilin",
  "n04348988": "subtracter",
  "n04349077": "subway station",
  "n04349189": "subway token",
  "n04349306": "subway train",
  "n04349401": "subwoofer",
  "n04349526": "succinylcholine",
  "n04349701": "sucralfate, Carafate",
  "n04349913": "suction cup",
  "n04350104": "suction pump",
  "n04350235": "sudatorium, sudatory",
  "n04350347": "sudorific, sudatory",
  "n04350458": "suede cloth, suede",
  "n04350581": "sugar bowl",
  "n04350688": "sugar refinery",
  "n04350769": "sugar spoon, sugar shell",
  "n04350905": "suit, suit of clothes",
  "n04351233": "suit",
  "n04351550": "suite, rooms",
  "n04351699": "suiting",
  "n04351776": "sulfacetamide, Sulamyd",
  "n04351936": "sulfadiazine",
  "n04352070": "sulfa drug, sulfa, sulpha, sulfonamide",
  "n04352435": "sulfamethazine, sulfamezathine",
  "n04352580": "sulfamethoxazole, Gantanol",
  "n04352759": "sulfanilamide",
  "n04352906": "sulfapyridine",
  "n04353016": "sulfisoxazole, Gantrisin",
  "n04353189": "sulfonylurea",
  "n04353410": "sulindac, Clinoril",
  "n04353573": "sulky",
  "n04353687": "sulphur mine, sulfur mine",
  "n04353803": "sum, total, totality, aggregate",
  "n04354026": "summer house",
  "n04354182": "sumo ring",
  "n04354287": "sump",
  "n04354387": "sump",
  "n04354487": "sump pump",
  "n04354589": "sunbonnet",
  "n04354723": "sunburst",
  "n04354828": "sunburst",
  "n04354966": "sunburst pleat, sunray pleat",
  "n04355115": "Sunday best, Sunday clothes",
  "n04355267": "sun deck",
  "n04355338": "sundial",
  "n04355511": "sundress",
  "n04355684": "sundries",
  "n04355821": "sun gear",
  "n04355933": "sunglass",
  "n04356056": "sunglasses, dark glasses, shades",
  "n04356306": "sunken garden",
  "n04356423": "sunk fence, ha-ha, haw-haw",
  "n04356595": "sunhat, sun hat",
  "n04356772": "sunlamp, sun lamp, sunray lamp, sun-ray lamp",
  "n04356925": "sun parlor, sun parlour, sun porch, sunporch, sunroom, sun lounge, solarium",
  "n04357121": "sunroof, sunshine-roof",
  "n04357314": "sunscreen, sunblock, sun blocker",
  "n04357531": "sunsuit",
  "n04357639": "suntrap, solar trap",
  "n04357795": "sun visor",
  "n04357930": "supercharger",
  "n04358117": "supercomputer",
  "n04358256": "superconducting supercollider",
  "n04358381": "superficies",
  "n04358491": "superhighway, information superhighway",
  "n04358707": "supermarket",
  "n04358874": "superstructure",
  "n04359034": "supertanker",
  "n04359124": "supper club",
  "n04359217": "supplejack",
  "n04359335": "supply chamber",
  "n04359500": "supply closet",
  "n04359589": "support",
  "n04360501": "support",
  "n04360798": "support column",
  "n04360914": "support hose, support stocking",
  "n04361095": "supporting structure",
  "n04361260": "supporting tower",
  "n04361381": "suppository",
  "n04361529": "suppressant, appetite suppressant",
  "n04361641": "suppressor, suppresser",
  "n04361801": "surbase",
  "n04361937": "surcoat",
  "n04362025": "surface",
  "n04362624": "surface gauge, surface gage, scribing block",
  "n04362821": "surface lift",
  "n04362972": "surface search radar",
  "n04363082": "surface ship",
  "n04363210": "surface-to-air missile, SAM",
  "n04363412": "surface-to-air missile system",
  "n04363559": "surfboard",
  "n04363671": "surfboat",
  "n04363777": "surcoat",
  "n04363874": "surgeon's knot",
  "n04363991": "surgery",
  "n04364160": "surge suppressor, surge protector, spike suppressor, spike arrester, lightning arrester",
  "n04364397": "surgical dressing",
  "n04364545": "surgical instrument",
  "n04364827": "surgical knife",
  "n04364994": "surplice",
  "n04365112": "surrey",
  "n04365229": "surtout",
  "n04365328": "surveillance system",
  "n04365484": "surveying instrument, surveyor's instrument",
  "n04365751": "surveyor's level",
  "n04366033": "sushi bar",
  "n04366116": "suspension, suspension system",
  "n04366367": "suspension bridge",
  "n04366832": "suspensory, suspensory bandage",
  "n04367011": "sustaining pedal, loud pedal",
  "n04367205": "suture",
  "n04367371": "suture, surgical seam",
  "n04367480": "swab, swob, mop",
  "n04367746": "swab",
  "n04367950": "swaddling clothes, swaddling bands",
  "n04368109": "swag",
  "n04368235": "swage block",
  "n04368365": "swagger stick",
  "n04368496": "swallow-tailed coat, swallowtail, morning coat",
  "n04368695": "swamp buggy, marsh buggy",
  "n04368840": "swan's down",
  "n04368949": "swatch",
  "n04369025": "swathe, wrapping",
  "n04369146": "swathing",
  "n04369282": "swatter, flyswatter, flyswat",
  "n04369485": "sweat bag",
  "n04369618": "sweatband",
  "n04369732": "sweatband",
  "n04369856": "sweatbox",
  "n04369973": "sweatbox",
  "n04370048": "sweater, jumper",
  "n04370288": "sweat pants, sweatpants",
  "n04370456": "sweatshirt",
  "n04370600": "sweatshop",
  "n04370774": "sweat suit, sweatsuit, sweats, workout suit",
  "n04370955": "sweep, sweep oar",
  "n04371050": "sweep hand, sweep-second",
  "n04371225": "swimming pool, swimming bath, natatorium",
  "n04371430": "swimming trunks, bathing trunks",
  "n04371563": "swimsuit, swimwear, bathing suit, swimming costume, bathing costume",
  "n04371774": "swing",
  "n04371979": "swing door, swinging door",
  "n04372171": "switch",
  "n04372370": "switch, electric switch, electrical switch",
  "n04372756": "switch",
  "n04372948": "switch",
  "n04373089": "switchblade, switchblade knife, flick-knife, flick knife",
  "n04373264": "switchboard, patchboard, plugboard",
  "n04373428": "switch engine, donkey engine",
  "n04373563": "swivel",
  "n04373704": "swivel chair",
  "n04373795": "swizzle stick",
  "n04373894": "sword, blade, brand, steel",
  "n04374315": "sword cane, sword stick",
  "n04374421": "sword knot",
  "n04374521": "S wrench",
  "n04374608": "Sydney Harbor Bridge",
  "n04374735": "synagogue, temple, tabernacle",
  "n04374907": "synchrocyclotron",
  "n04375080": "synchroflash",
  "n04375241": "synchromesh",
  "n04375405": "synchronous converter, rotary, rotary converter",
  "n04375615": "synchronous motor",
  "n04375775": "synchrotron",
  "n04375926": "synchroscope, synchronoscope, synchronizer, synchroniser",
  "n04376259": "synergist",
  "n04376400": "synthesizer, synthesiser",
  "n04376715": "synthetism",
  "n04376876": "syringe",
  "n04377057": "system",
  "n04377842": "system clock",
  "n04378024": "system clock",
  "n04378123": "tab",
  "n04378489": "tabard",
  "n04378651": "Tabernacle",
  "n04378842": "Tabernacle, Mormon Tabernacle",
  "n04378956": "tabi, tabis",
  "n04379096": "tab key, tab",
  "n04379243": "table",
  "n04379964": "table",
  "n04380143": "tablecloth",
  "n04380255": "tablefork",
  "n04380346": "table knife",
  "n04380533": "table lamp",
  "n04380617": "table linen, napery",
  "n04380801": "table mat, hot pad",
  "n04380916": "table saw",
  "n04381073": "tablespoon",
  "n04381183": "tablet",
  "n04381302": "tablet",
  "n04381450": "tablet-armed chair",
  "n04381587": "table-tennis table, ping-pong table, pingpong table",
  "n04381724": "table-tennis racquet, table-tennis bat, pingpong paddle",
  "n04381860": "tabletop",
  "n04381994": "tableware",
  "n04382334": "tabor, tabour",
  "n04382438": "taboret, tabouret",
  "n04382537": "tachistoscope, t-scope",
  "n04382695": "tachograph",
  "n04382880": "tachometer, tach",
  "n04383015": "tachymeter, tacheometer",
  "n04383130": "tack",
  "n04383301": "tack hammer",
  "n04383401": "Tacoma Narrows Bridge",
  "n04383537": "tadalafil, Cialis",
  "n04383696": "taenia, tenia, fillet",
  "n04383839": "taffeta",
  "n04383923": "taffrail",
  "n04384016": "tail, tail assembly, empennage",
  "n04384199": "tail",
  "n04384406": "tail fin, tailfin, fin",
  "n04384593": "tailgate, tailboard",
  "n04384757": "tail gate",
  "n04384910": "taillight, tail lamp, rear light, rear lamp",
  "n04385079": "tailor-made",
  "n04385157": "tailor's chalk",
  "n04385272": "tailor's tack",
  "n04385431": "tailpiece",
  "n04385536": "tailpipe",
  "n04385669": "tailrace",
  "n04385799": "tail rotor, anti-torque rotor",
  "n04386051": "tailstock",
  "n04386283": "Taj Mahal",
  "n04386456": "take-up",
  "n04386664": "talaria",
  "n04386792": "talcum, talcum powder",
  "n04386962": "talking book",
  "n04387095": "tam, tam-o'-shanter, tammy",
  "n04387201": "tambour",
  "n04387261": "tambour, embroidery frame, embroidery hoop",
  "n04387400": "tambourine",
  "n04387531": "tammy",
  "n04387706": "tamp, tamper, tamping bar",
  "n04387932": "Tampax",
  "n04388040": "tampion, tompion",
  "n04388162": "tampon",
  "n04388372": "tandem trailer",
  "n04388473": "tandoor",
  "n04388574": "tangram",
  "n04388743": "tank, storage tank",
  "n04389033": "tank, army tank, armored combat vehicle, armoured combat vehicle",
  "n04389340": "tanka",
  "n04389430": "tankard",
  "n04389521": "tank car, tank",
  "n04389637": "tank circuit",
  "n04389718": "tank destroyer",
  "n04389854": "tank engine, tank locomotive",
  "n04389999": "tanker plane",
  "n04390156": "tank furnace",
  "n04390338": "tank iron",
  "n04390483": "tank shell",
  "n04390577": "tank top",
  "n04390759": "tannery",
  "n04390873": "tannoy",
  "n04390977": "tap, spigot",
  "n04391158": "tap",
  "n04391276": "tap",
  "n04391445": "tapa, tappa",
  "n04391569": "tape",
  "n04391838": "tape, tape recording, taping",
  "n04392113": "tape, tapeline, tape measure",
  "n04392370": "tape cartridge",
  "n04392526": "tape deck",
  "n04392764": "tape drive, tape transport, transport",
  "n04392985": "tape player",
  "n04393095": "tape recorder, tape machine",
  "n04393301": "taper file",
  "n04393404": "tapestry, arras",
  "n04393549": "tapestry, tapis",
  "n04393681": "Tappan Zee Bridge",
  "n04393808": "tappet",
  "n04393913": "tap wrench",
  "n04394031": "tare",
  "n04394261": "target, butt",
  "n04394421": "target acquisition system",
  "n04394630": "tarmacadam, tarmac, macadam",
  "n04394821": "tarot card, tarot",
  "n04395024": "tarpaulin, tarp",
  "n04395106": "tartan, plaid",
  "n04395201": "tassel",
  "n04395332": "tasset, tasse",
  "n04395495": "tatting",
  "n04395651": "tattoo",
  "n04395760": "tau cross, St. Anthony's cross",
  "n04395875": "tavern, tap house",
  "n04396093": "taw, shooter",
  "n04396226": "tawse",
  "n04396335": "taximeter",
  "n04396466": "taxiway, taxi strip",
  "n04396650": "T-bar lift, T-bar, Alpine lift",
  "n04396808": "tea bag",
  "n04396902": "tea ball",
  "n04397027": "tea cart, teacart, tea trolley, tea wagon",
  "n04397168": "tea chest",
  "n04397261": "teaching aid",
  "n04397378": "tea cloth",
  "n04397452": "teacup",
  "n04397553": "tea garden",
  "n04397645": "tea gown",
  "n04397768": "teakettle",
  "n04397860": "tea maker",
  "n04397949": "tea napkin",
  "n04398044": "teapot",
  "n04398167": "teaser",
  "n04398309": "tea service, tea set",
  "n04398497": "teashop, teahouse, tearoom, tea parlor, tea parlour",
  "n04398688": "teaspoon",
  "n04398834": "tea-strainer",
  "n04398951": "tea table",
  "n04399046": "tea tray",
  "n04399158": "tea urn",
  "n04399269": "technical",
  "n04399382": "teddy, teddy bear",
  "n04399537": "tee, golf tee",
  "n04399700": "tee, football tee",
  "n04399846": "tee hinge, T hinge",
  "n04400109": "telecom hotel, telco building",
  "n04400289": "telecommunication system, telecom system, telecommunication equipment, telecom equipment",
  "n04400499": "telegraph, telegraphy",
  "n04400737": "telegraph key",
  "n04400899": "telemeter",
  "n04401088": "telephone, phone, telephone set",
  "n04401578": "telephone bell",
  "n04401680": "telephone booth, phone booth, call box, telephone box, telephone kiosk",
  "n04401828": "telephone cord, phone cord",
  "n04401949": "telephone jack, phone jack",
  "n04402057": "telephone line, phone line, telephone circuit, subscriber line, line",
  "n04402342": "telephone plug, phone plug",
  "n04402449": "telephone pole, telegraph pole, telegraph post",
  "n04402580": "telephone receiver, receiver",
  "n04402746": "telephone system, phone system",
  "n04402984": "telephone wire, telephone line, telegraph wire, telegraph line",
  "n04403167": "telephotograph, telephoto",
  "n04403279": "telephotograph",
  "n04403413": "telephoto lens, zoom lens",
  "n04403524": "Teleprompter",
  "n04403638": "telescope, scope",
  "n04403925": "telescopic sight, telescope sight",
  "n04404072": "telethermometer",
  "n04404200": "teletypewriter, teleprinter, teletype machine, telex, telex machine",
  "n04404412": "television, television system",
  "n04404817": "television antenna, tv-antenna",
  "n04404997": "television camera, tv camera, camera",
  "n04405309": "television-camera tube, television pickup tube",
  "n04405540": "television equipment, video equipment",
  "n04405762": "television monitor, tv monitor",
  "n04405907": "television receiver, television, television set, tv, tv set, idiot box, boob tube, telly, goggle box",
  "n04406239": "television room, tv room",
  "n04406350": "television station, TV station",
  "n04406552": "television transmitter",
  "n04406687": "telpher, telfer",
  "n04406817": "telpherage, telferage",
  "n04407007": "temazepam, Restoril",
  "n04407257": "tempera, poster paint, poster color, poster colour",
  "n04407435": "temple",
  "n04407686": "temple",
  "n04407844": "Temple of Apollo, Oracle of Apollo, Delphic oracle, oracle of Delphi",
  "n04408174": "Temple of Artemis",
  "n04408330": "Temple of Jerusalem, Temple of Solomon",
  "n04408871": "temporary hookup, patch",
  "n04409011": "tender, supply ship",
  "n04409128": "tender, ship's boat, pinnace, cutter",
  "n04409279": "tender",
  "n04409384": "tenement, tenement house",
  "n04409515": "tennis ball",
  "n04409625": "tennis camp",
  "n04409712": "tennis court",
  "n04409806": "tennis racket, tennis racquet",
  "n04409911": "tenon",
  "n04410086": "tenor drum, tom-tom",
  "n04410190": "Tenoretic",
  "n04410365": "tenoroon",
  "n04410485": "tenpenny nail",
  "n04410565": "tenpin",
  "n04410663": "tensimeter",
  "n04410760": "tensiometer",
  "n04410886": "tensiometer",
  "n04411019": "tensiometer",
  "n04411142": "ten-spot, ten",
  "n04411264": "tent, collapsible shelter",
  "n04411835": "tenter",
  "n04411966": "tenterhook",
  "n04412097": "tent-fly, rainfly, fly sheet, fly, tent flap",
  "n04412300": "tent peg",
  "n04412416": "tepee, tipi, teepee",
  "n04412550": "terazosin, Hytrin",
  "n04412727": "terbinafine, Lamisil",
  "n04412901": "terminal, terminus, depot",
  "n04413151": "terminal, pole",
  "n04413419": "terminal",
  "n04413631": "terminus",
  "n04413723": "terminus, terminal figure, term",
  "n04413969": "terraced house",
  "n04414101": "terra cotta",
  "n04414199": "terrarium",
  "n04414319": "terra sigillata, Samian ware",
  "n04414476": "terry, terry cloth, terrycloth",
  "n04414675": "Tesla coil",
  "n04414821": "tessella",
  "n04414909": "tessera",
  "n04415078": "test bed",
  "n04415257": "test equipment",
  "n04415354": "tester",
  "n04415460": "test paper",
  "n04415578": "test range",
  "n04415663": "test rocket, research rocket, test instrument vehicle",
  "n04415815": "test room, testing room",
  "n04415921": "test tube",
  "n04416005": "testudo",
  "n04416201": "tetracaine",
  "n04416338": "tetrachlorethylene, tetrachloroethylene, ethylene tetrachloride, carbon dichloride",
  "n04416530": "tetracycline, Achromycin",
  "n04416785": "tetrahydrocannabinol, THC",
  "n04416901": "tetraskelion, tetraskele",
  "n04417086": "tetrode",
  "n04417180": "textile machine",
  "n04417361": "textile mill",
  "n04417467": "thalidomide",
  "n04417672": "thatch, thatched roof",
  "n04417809": "theater, theatre, house",
  "n04418357": "theater curtain, theatre curtain",
  "n04418644": "theater light",
  "n04418818": "theater stage, theatre stage",
  "n04419073": "theodolite, transit",
  "n04419315": "theophylline, Elixophyllin, Slo-Bid, Theobid",
  "n04419642": "theremin",
  "n04419868": "thermal printer",
  "n04420024": "thermal reactor",
  "n04420206": "thermistor, thermal resistor",
  "n04420461": "thermobaric bomb, fuel-air bomb, vacuum bomb, volume-detonation bomb, aerosol bomb",
  "n04420720": "thermocouple, thermocouple junction",
  "n04421083": "thermoelectric thermometer, thermel, electric thermometer",
  "n04421258": "thermograph, thermometrograph",
  "n04421417": "thermograph",
  "n04421582": "thermohydrometer, thermogravimeter",
  "n04421740": "thermojunction",
  "n04421872": "thermometer",
  "n04422409": "thermonuclear reactor, fusion reactor",
  "n04422566": "thermopile",
  "n04422727": "thermos, thermos bottle, thermos flask",
  "n04422875": "thermostat, thermoregulator",
  "n04423174": "thiabendazole",
  "n04423288": "thiazide",
  "n04423552": "thigh pad",
  "n04423687": "thill",
  "n04423845": "thimble",
  "n04424003": "thimerosal, sodium ethylmercurithiosalicylate, Merthiolate",
  "n04424218": "thing",
  "n04424418": "thing",
  "n04424692": "thinning shears",
  "n04424826": "thioguanine",
  "n04424936": "thiopental, thiopental sodium, thiopentobarbital sodium, Pentothal",
  "n04425262": "thioridazine, Mellaril",
  "n04425445": "Thiosulfil",
  "n04425544": "thiotepa",
  "n04425656": "thiothixene, Navane",
  "n04425804": "third base, third",
  "n04425977": "third gear, third",
  "n04426184": "third rail",
  "n04426316": "thong",
  "n04426427": "thong",
  "n04426618": "thoroughfare",
  "n04426788": "thread, yarn",
  "n04427216": "three-centered arch, basket-handle arch",
  "n04427376": "three-decker",
  "n04427473": "three-decker",
  "n04427559": "three-dimensional radar, 3d radar",
  "n04427715": "three-piece suit",
  "n04427857": "three-quarter binding",
  "n04428008": "three-way switch, three-point switch",
  "n04428191": "thresher, thrasher, threshing machine",
  "n04428382": "threshing floor",
  "n04428502": "threshold element, threshold gate",
  "n04428634": "thriftshop, second-hand store",
  "n04428763": "throat",
  "n04428920": "throat",
  "n04429038": "throat protector",
  "n04429169": "thrombolytic, thrombolytic agent, clot buster",
  "n04429376": "throne",
  "n04429613": "throstle",
  "n04429756": "throughput",
  "n04429971": "throw",
  "n04430158": "throwing stick, throwing board, spear thrower, dart thrower",
  "n04430364": "throw pillow",
  "n04430475": "thrust bearing",
  "n04430605": "thruster",
  "n04430738": "thrust stage",
  "n04430896": "thumb",
  "n04431025": "thumbhole",
  "n04431158": "thumbhole",
  "n04431296": "thumb index",
  "n04431436": "thumbscrew",
  "n04431547": "thumbscrew",
  "n04431648": "thumbstall",
  "n04431745": "thumbtack, drawing pin, pushpin",
  "n04431925": "thunderer",
  "n04432043": "thwart, cross thwart",
  "n04432203": "tiara",
  "n04432308": "tick",
  "n04432393": "ticker, stock ticker",
  "n04432538": "ticket window",
  "n04432662": "ticking",
  "n04432785": "tickler coil",
  "n04432942": "tidemark",
  "n04433077": "tidy",
  "n04433185": "tie",
  "n04433377": "tie, tie beam",
  "n04433585": "tie, railroad tie, crosstie, sleeper",
  "n04433800": "tie clip",
  "n04433905": "tier",
  "n04434059": "tier",
  "n04434207": "tie rack",
  "n04434285": "tiered seat",
  "n04434531": "tie rod",
  "n04434670": "tie tack, tiepin, scarfpin",
  "n04434780": "tightrope",
  "n04434932": "tights, leotards",
  "n04435180": "tile",
  "n04435378": "tile",
  "n04435552": "tile cutter",
  "n04435653": "tile roof",
  "n04435759": "tiller",
  "n04435870": "tilter",
  "n04436012": "tilt-top table, tip-top table, tip table",
  "n04436185": "timber",
  "n04436329": "timber",
  "n04436401": "timber hitch",
  "n04436542": "timbrel",
  "n04436675": "time-ball",
  "n04436832": "time bomb, infernal machine",
  "n04436992": "time capsule",
  "n04437131": "timecard",
  "n04437276": "time clock",
  "n04437380": "time-delay measuring instrument, time-delay measuring system",
  "n04437552": "time exposure",
  "n04437670": "time-fuse",
  "n04437793": "time machine",
  "n04437953": "timepiece, timekeeper, horologe",
  "n04438304": "timer",
  "n04438507": "timer",
  "n04438643": "time-switch",
  "n04438742": "timolol, Blocadren",
  "n04438897": "tin",
  "n04439039": "tin can",
  "n04439122": "tincture",
  "n04439305": "tincture of iodine, iodine",
  "n04439505": "tinderbox",
  "n04439585": "tine",
  "n04439712": "tinfoil, tin foil",
  "n04439840": "tin plate, tinplate",
  "n04440059": "tinsel",
  "n04440210": "tinsel",
  "n04440399": "tintack",
  "n04440486": "tinware",
  "n04440597": "tippet",
  "n04440749": "tire, tyre",
  "n04440963": "tire chain, snow chain",
  "n04441093": "tire iron, tire tool",
  "n04441282": "tissue plasminogen activator, Activase",
  "n04441528": "titfer",
  "n04441662": "tithe barn",
  "n04441790": "titrator",
  "n04441902": "T-junction",
  "n04442016": "T-network",
  "n04442143": "TNT, trinitrotoluene",
  "n04442312": "toaster",
  "n04442441": "toaster oven",
  "n04442582": "toasting fork",
  "n04442741": "toastrack",
  "n04442831": "tobacco, baccy",
  "n04443164": "tobacco pouch",
  "n04443257": "tobacco shop, tobacconist shop, tobacconist",
  "n04443433": "toboggan",
  "n04443588": "tobramycin, Nebcin",
  "n04443766": "toby, toby jug, toby fillpot jug",
  "n04443918": "tocainide, Tonocard",
  "n04444121": "tocsin, warning bell",
  "n04444218": "toe",
  "n04444345": "toe",
  "n04444522": "toe box",
  "n04444749": "toecap",
  "n04444953": "toehold",
  "n04445040": "toga",
  "n04445154": "toga virilis",
  "n04445327": "toggle",
  "n04445610": "toggle bolt",
  "n04445782": "toggle joint",
  "n04445952": "toggle switch, toggle, on-off switch, on/off switch",
  "n04446162": "togs, threads, duds",
  "n04446276": "toilet, lavatory, lav, can, john, privy, bathroom",
  "n04446521": "toilet, can, commode, crapper, pot, potty, stool, throne",
  "n04446844": "toilet bag, sponge bag",
  "n04447028": "toilet bowl",
  "n04447156": "toilet kit, travel kit",
  "n04447276": "toilet powder, bath powder, dusting powder",
  "n04447443": "toiletry, toilet articles",
  "n04447861": "toilet seat",
  "n04447965": "toilet soap, face soap, bath soap",
  "n04448070": "toilet water, eau de toilette",
  "n04448185": "tokamak",
  "n04448361": "token",
  "n04448511": "tolazamide, Tolinase",
  "n04448680": "tolazoline",
  "n04448826": "tolbutamide, Orinase",
  "n04449046": "tole",
  "n04449290": "tollbooth, tolbooth, tollhouse",
  "n04449449": "toll bridge",
  "n04449550": "tollgate, tollbar",
  "n04449700": "toll line",
  "n04449796": "tolmetin sodium, Tolectin",
  "n04449966": "tomahawk, hatchet",
  "n04450133": "Tommy gun, Thompson submachine gun",
  "n04450243": "tomograph",
  "n04450465": "tone arm, pickup, pickup arm",
  "n04450640": "toner",
  "n04450749": "tongs, pair of tongs",
  "n04450994": "tongue",
  "n04451139": "tongue and groove joint",
  "n04451318": "tongue depressor",
  "n04451473": "tonic, restorative",
  "n04451636": "tonometer",
  "n04451818": "tool",
  "n04452528": "tool bag",
  "n04452615": "toolbox, tool chest, tool cabinet, tool case",
  "n04452757": "toolshed, toolhouse",
  "n04452848": "tooth",
  "n04453037": "tooth",
  "n04453156": "toothbrush",
  "n04453298": "toothpaste",
  "n04453390": "toothpick",
  "n04453557": "tooth powder, toothpowder",
  "n04453666": "top",
  "n04453910": "top, cover",
  "n04454240": "top, whirligig, teetotum, spinning top",
  "n04454540": "top",
  "n04454654": "topgallant, topgallant mast",
  "n04454792": "topgallant, topgallant sail",
  "n04454908": "topiary",
  "n04455048": "topknot",
  "n04455169": "top lift",
  "n04455250": "topmast",
  "n04455442": "top of the line",
  "n04455579": "topper",
  "n04455652": "topsail",
  "n04455835": "topside",
  "n04456011": "toque",
  "n04456115": "torch",
  "n04456276": "tormenter, tormentor, teaser",
  "n04456472": "torpedo",
  "n04456734": "torpedo",
  "n04456964": "torpedo",
  "n04457157": "torpedo",
  "n04457326": "torpedo boat",
  "n04457474": "torpedo-boat destroyer",
  "n04457638": "torpedo tube",
  "n04457767": "torque converter",
  "n04457910": "torque wrench",
  "n04458045": "torsion balance",
  "n04458201": "torture chamber",
  "n04458299": "torus, tore",
  "n04458409": "totem",
  "n04458633": "totem pole",
  "n04458843": "touch screen, touchscreen",
  "n04459018": "toupee, toupe",
  "n04459122": "touring car, phaeton, tourer",
  "n04459243": "tourist class, third class",
  "n04459362": "towel",
  "n04459610": "toweling, towelling",
  "n04459773": "towel rack, towel horse",
  "n04459909": "towel rail, towel bar",
  "n04460038": "towel ring",
  "n04460130": "tower",
  "n04460634": "Tower of Babel, Babel",
  "n04460947": "Tower of London",
  "n04461148": "Tower of Pharos",
  "n04461294": "towline, towrope, towing line, towing rope",
  "n04461437": "town hall",
  "n04461570": "towpath, towing path",
  "n04461696": "tow truck, tow car, wrecker",
  "n04461879": "toy",
  "n04462011": "toy box, toy chest",
  "n04462102": "toy",
  "n04462240": "toyshop",
  "n04462318": "toy soldier",
  "n04462407": "trace",
  "n04462576": "trace detector",
  "n04462707": "tracer, tracer bullet",
  "n04462835": "tracer",
  "n04463017": "tracer",
  "n04463141": "tracery",
  "n04463273": "tracing, trace",
  "n04463510": "track, cart track, cartroad",
  "n04463679": "track, rail, rails, runway",
  "n04463983": "track",
  "n04464125": "track",
  "n04464211": "track, data track",
  "n04464418": "track, caterpillar track, caterpillar tread",
  "n04464615": "trackball",
  "n04464852": "tracked vehicle",
  "n04465050": "tract house",
  "n04465203": "tract housing",
  "n04465358": "traction engine",
  "n04465501": "tractor",
  "n04465666": "tractor",
  "n04465795": "trading card",
  "n04465933": "traffic circle, circle, rotary, roundabout",
  "n04466169": "traffic island, safety island, safety isle, safety zone",
  "n04466386": "traffic lane",
  "n04466613": "trail",
  "n04466871": "trail bike, dirt bike, scrambler",
  "n04467099": "trailer, house trailer",
  "n04467307": "trailer",
  "n04467506": "trailer camp, trailer park",
  "n04467665": "trailer truck, tractor trailer, trucking rig, rig, articulated lorry, semi",
  "n04467899": "trailing edge",
  "n04468005": "train, railroad train",
  "n04468476": "train",
  "n04468707": "train set",
  "n04468847": "tramcar, tram",
  "n04469003": "tramline, tramway, streetcar track",
  "n04469147": "trammel",
  "n04469251": "trammel",
  "n04469345": "trammel net, trammel",
  "n04469514": "trampoline",
  "n04469684": "tramp steamer, tramp",
  "n04469813": "tramway, tram, aerial tramway, cable tramway, ropeway",
  "n04470037": "trandolapril, Mavik",
  "n04470232": "tranquilizer, tranquillizer, tranquilliser, antianxiety agent, ataractic drug, ataractic agent, ataractic",
  "n04470605": "transcription",
  "n04470741": "transdermal patch, skin patch",
  "n04470953": "transducer",
  "n04471148": "transept",
  "n04471315": "transformer",
  "n04471632": "transistor, junction transistor, electronic transistor",
  "n04471912": "transit instrument",
  "n04472098": "transit line",
  "n04472243": "transmission, transmission system",
  "n04472563": "transmission shaft",
  "n04472726": "transmitter, sender",
  "n04472961": "transom, traverse",
  "n04473108": "transom, transom window, fanlight",
  "n04473275": "transponder",
  "n04473432": "transportation system, transportation, transit",
  "n04473884": "transporter",
  "n04474035": "transporter, car transporter",
  "n04474187": "transport ship",
  "n04474317": "tranylcypromine",
  "n04474466": "trap",
  "n04474843": "trap",
  "n04474922": "trap",
  "n04475032": "trap",
  "n04475195": "trap-and-drain auger",
  "n04475309": "trap door",
  "n04475411": "trapeze",
  "n04475496": "trave, traverse, crossbeam, crosspiece",
  "n04475631": "travel iron",
  "n04475749": "trawl, dragnet, trawl net",
  "n04475900": "trawl, trawl line, spiller, setline, trotline",
  "n04476116": "trawler, dragger",
  "n04476259": "tray",
  "n04476526": "tray cloth",
  "n04476633": "trazodone, trazodone hydrochloride, Desyrel",
  "n04476831": "tread",
  "n04476972": "tread",
  "n04477091": "tread",
  "n04477219": "treadmill, treadwheel, tread-wheel",
  "n04477387": "treadmill",
  "n04477548": "treasure chest",
  "n04477638": "treasure house",
  "n04477725": "treasure ship",
  "n04477825": "treasury",
  "n04477966": "tree house",
  "n04478066": "treenail, trenail, trunnel",
  "n04478257": "trefoil",
  "n04478383": "trefoil arch",
  "n04478512": "trellis, treillage",
  "n04478657": "trench",
  "n04478889": "trench",
  "n04479046": "trench coat",
  "n04479155": "trencher",
  "n04479287": "trench knife",
  "n04479405": "trepan",
  "n04479526": "trepan, trephine",
  "n04479694": "trestle",
  "n04479823": "trestle",
  "n04479939": "trestle bridge",
  "n04480033": "trestle table",
  "n04480141": "trestlework",
  "n04480303": "trews",
  "n04480415": "trey, three",
  "n04480527": "trial balloon",
  "n04480625": "triazolam, Halcion",
  "n04480853": "triangle",
  "n04480995": "triangle",
  "n04481144": "tribromoethanol, tribromoethyl alcohol",
  "n04481255": "tribune",
  "n04481373": "trichlormethiazide, Naqua",
  "n04481524": "triclinium",
  "n04481642": "triclinium",
  "n04481796": "tricolor, tricolour",
  "n04481946": "tricolor television tube, tricolour television tube, tricolor tube, tricolour tube",
  "n04482177": "tricorn, tricorne",
  "n04482297": "tricot",
  "n04482393": "tricycle, trike, velocipede",
  "n04482543": "tricyclic, tricyclic antidepressant, tricyclic antidepressant drug",
  "n04482975": "trident",
  "n04483073": "trigger",
  "n04483210": "trigon",
  "n04483307": "trimaran",
  "n04483399": "trimipramine, Surmontil",
  "n04483584": "trimmer, trimmer joist",
  "n04483779": "trimmer, trimming capacitor",
  "n04483925": "trimmer",
  "n04484024": "trimmer arch",
  "n04484160": "trimming, trim, passementerie",
  "n04484432": "triode",
  "n04484647": "triphammer",
  "n04484780": "triplicate",
  "n04484952": "trip line",
  "n04485082": "tripod",
  "n04485226": "tripper, trip",
  "n04485423": "triptych",
  "n04485586": "trip wire",
  "n04485750": "trireme",
  "n04485884": "triskelion, triskele",
  "n04486054": "triumphal arch",
  "n04486213": "trivet",
  "n04486322": "trivet",
  "n04486445": "triviality, trivia, trifle, small beer",
  "n04486616": "troika",
  "n04486718": "Trojan Horse, Wooden Horse",
  "n04486934": "troll",
  "n04487081": "trolleybus, trolley coach, trackless trolley",
  "n04487268": "trolley line",
  "n04487394": "trombone",
  "n04487576": "trompe l'oeil",
  "n04487724": "troop carrier, troop transport",
  "n04487894": "troopship",
  "n04487996": "trophy, prize",
  "n04488202": "trophy case",
  "n04488295": "trou-de-loup",
  "n04488427": "trough",
  "n04488530": "trouser",
  "n04488742": "trouser cuff",
  "n04488857": "trouser press, pants presser",
  "n04489008": "trouser, pant",
  "n04489695": "trousseau",
  "n04489817": "trowel",
  "n04490091": "truck, motortruck",
  "n04490490": "truck bed",
  "n04490614": "truck farm, truck garden",
  "n04490730": "truck stop",
  "n04490859": "trump",
  "n04491135": "trump, trump card",
  "n04491312": "trumpet arch",
  "n04491388": "truncheon, nightstick, baton, billy, billystick, billy club",
  "n04491545": "trundle",
  "n04491638": "trundle bed, trundle, truckle bed, truckle",
  "n04491769": "trunk",
  "n04491934": "trunk hose",
  "n04492060": "trunk lid",
  "n04492157": "trunk line",
  "n04492264": "trunk line, trunk route",
  "n04492375": "truss",
  "n04492585": "truss",
  "n04492749": "truss bridge",
  "n04492856": "truth serum, truth drug",
  "n04493109": "try square",
  "n04493259": "T-square",
  "n04493381": "tub, vat",
  "n04493505": "tube, tubing",
  "n04494204": "tube, vacuum tube, thermionic vacuum tube, thermionic tube, electron tube, thermionic valve",
  "n04494762": "tubeless, tubeless tire",
  "n04494906": "tuck",
  "n04495051": "tuck box",
  "n04495183": "tucker",
  "n04495310": "tucker-bag",
  "n04495450": "tuck shop",
  "n04495555": "Tudor arch, four-centered arch",
  "n04495698": "tudung",
  "n04495843": "tugboat, tug, towboat, tower",
  "n04496035": "Tuileries, Tuileries Gardens",
  "n04496173": "Tuileries, Tuileries Palace",
  "n04496404": "tuille",
  "n04496519": "tulip bed",
  "n04496614": "tulle",
  "n04496726": "tumble-dryer, tumble drier",
  "n04496872": "tumbler",
  "n04497005": "tumbler",
  "n04497249": "tumbrel, tumbril",
  "n04497442": "tun",
  "n04497570": "tunic",
  "n04497801": "tuning fork",
  "n04497962": "tunnel",
  "n04498275": "tupik, tupek, sealskin tent",
  "n04498389": "turban",
  "n04498523": "turbine",
  "n04498873": "turbogenerator",
  "n04499062": "tureen",
  "n04499180": "Turing machine",
  "n04499300": "Turkish bath",
  "n04499446": "Turkish towel, terry towel",
  "n04499554": "Turk's head",
  "n04499660": "turnaround",
  "n04499810": "turnbuckle",
  "n04500060": "turner, food turner",
  "n04500294": "turnery",
  "n04500390": "turnery",
  "n04500504": "turning",
  "n04500613": "turnip bed",
  "n04500704": "turnoff",
  "n04500866": "turnout, widening",
  "n04501018": "turnpike, toll road",
  "n04501127": "turnpike",
  "n04501281": "turnspit",
  "n04501370": "turnstile",
  "n04501550": "turntable",
  "n04501713": "turntable",
  "n04501837": "turntable, lazy Susan",
  "n04501947": "turret",
  "n04502059": "turret clock",
  "n04502197": "turtleneck, turtle, polo-neck",
  "n04502364": "turtleneck collar, polo-neck collar",
  "n04502502": "tweed",
  "n04502670": "tweeter",
  "n04502851": "twenty-two, .22",
  "n04502989": "twenty-two pistol",
  "n04503073": "twenty-two rifle",
  "n04503155": "twill",
  "n04503269": "twill, twill weave",
  "n04503413": "twin bed",
  "n04503499": "twinjet",
  "n04503593": "twist bit, twist drill",
  "n04503705": "two-by-four",
  "n04503836": "two-handed saw, whipsaw, two-man saw, lumberman's saw",
  "n04504038": "two-man tent",
  "n04504141": "two-piece, two-piece suit, lounge suit",
  "n04504297": "two-way street",
  "n04504486": "type",
  "n04504770": "typesetting machine",
  "n04504935": "type slug, slug",
  "n04505036": "typewriter",
  "n04505345": "typewriter carriage",
  "n04505470": "typewriter keyboard",
  "n04505706": "tyrocidine, tyrocidin",
  "n04505888": "tyrolean, tirolean",
  "n04506005": "tyrosine kinase inhibitor",
  "n04506125": "tyrothricin",
  "n04506289": "uke, ukulele",
  "n04506402": "ulster",
  "n04506506": "ultracentrifuge",
  "n04506688": "ultramicroscope, dark-field microscope",
  "n04506895": "Ultrasuede",
  "n04506994": "ultraviolet lamp, ultraviolet source",
  "n04507155": "umbrella",
  "n04507326": "umbrella tent",
  "n04507453": "undercarriage",
  "n04507609": "undercharge",
  "n04507689": "undercoat, underseal",
  "n04507891": "undercut",
  "n04508062": "underfelt",
  "n04508163": "undergarment, unmentionable",
  "n04508489": "underpants",
  "n04508804": "underpass, subway",
  "n04508949": "underwear, underclothes, underclothing",
  "n04509171": "undies",
  "n04509260": "uneven parallel bars, uneven bars",
  "n04509417": "unicycle, monocycle",
  "n04509592": "uniform",
  "n04509815": "union",
  "n04509985": "Union Jack, Union flag",
  "n04510090": "United States Army Criminal Investigation Laboratory, U.S. Army Criminal Investigation Laboratory, US Army Criminal Investigation Laboratory, USACIL",
  "n04510456": "United States Mint, U.S. Mint, US Mint",
  "n04510706": "universal joint, universal",
  "n04511002": "university",
  "n04511832": "University of California at Berkeley",
  "n04511971": "University of Chicago",
  "n04512092": "University of Michigan",
  "n04512216": "University of Nebraska",
  "n04512338": "University of North Carolina",
  "n04512476": "University of Pennsylvania, Pennsylvania, Penn",
  "n04512652": "University of Pittsburgh",
  "n04512783": "University of Sussex, Sussex University",
  "n04512933": "University of Texas",
  "n04513048": "University of Vermont",
  "n04513171": "University of Washington",
  "n04513297": "University of West Virginia",
  "n04513432": "University of Wisconsin",
  "n04513556": "upcast",
  "n04513686": "upgrade",
  "n04513827": "upholstery",
  "n04513998": "upholstery material",
  "n04514095": "upholstery needle",
  "n04514241": "uplift",
  "n04514359": "upper",
  "n04514648": "upper berth, upper",
  "n04514738": "upper deck",
  "n04514899": "upper surface",
  "n04515003": "upright, upright piano",
  "n04515129": "upright, vertical",
  "n04515444": "upset, swage",
  "n04515630": "upstage",
  "n04515729": "upstairs",
  "n04515890": "urceole",
  "n04515991": "urinal",
  "n04516116": "urn",
  "n04516214": "urn",
  "n04516354": "used-car, secondhand car",
  "n04516477": "USS Cole",
  "n04516672": "utensil",
  "n04516874": "utility",
  "n04517211": "Uzi",
  "n04517408": "vacation home",
  "n04517535": "vaccine, vaccinum",
  "n04517823": "vacuum, vacuum cleaner",
  "n04517999": "vacuum chamber",
  "n04518132": "vacuum flask, vacuum bottle",
  "n04518343": "vacuum gauge, vacuum gage",
  "n04518468": "valdecoxib, Bextra",
  "n04518643": "Valenciennes, Valenciennes lace",
  "n04518764": "valise",
  "n04518854": "valproic acid, Depokene",
  "n04519019": "valsartan, Diovan",
  "n04519153": "valve",
  "n04519536": "valve",
  "n04519728": "valve-in-head engine",
  "n04519887": "vambrace, lower cannon",
  "n04520020": "vamp",
  "n04520170": "van",
  "n04520382": "van, caravan",
  "n04520480": "van",
  "n04520618": "vancomycin, Vancocin",
  "n04520784": "vane",
  "n04520962": "vaporizer, vaporiser",
  "n04521125": "vapor lock, vapour lock",
  "n04521428": "vardenafil, Levitra",
  "n04521571": "variable-pitch propeller",
  "n04521699": "variation",
  "n04521863": "variometer",
  "n04521987": "varnish",
  "n04522168": "vase",
  "n04522309": "Vaseline",
  "n04522421": "vasoconstrictor, vasoconstrictive, pressor",
  "n04522904": "vasodilator, vasodilative",
  "n04523201": "vasopressor",
  "n04523380": "Vatican, Vatican Palace",
  "n04523525": "vault",
  "n04523690": "vault, burial vault",
  "n04523831": "vault, bank vault",
  "n04523993": "vaulting",
  "n04524142": "vaulting horse, long horse, buck",
  "n04524313": "vehicle",
  "n04524594": "Velcro",
  "n04524716": "velocipede",
  "n04524833": "velodrome",
  "n04524941": "velour, velours",
  "n04525038": "velvet",
  "n04525191": "velveteen",
  "n04525305": "vending machine",
  "n04525417": "veneer, veneering",
  "n04525584": "Venetian blind",
  "n04525729": "Venetian glass",
  "n04525821": "Venn diagram, Venn's diagram",
  "n04526112": "venogram, phlebogram",
  "n04526241": "vent, venthole, vent-hole, blowhole",
  "n04526416": "vent",
  "n04526520": "ventilation, ventilation system, ventilating system",
  "n04526800": "ventilation shaft",
  "n04526964": "ventilator",
  "n04527182": "ventriloquist's dummy",
  "n04527310": "venturi",
  "n04527475": "Venturi tube",
  "n04527648": "veranda, verandah, gallery",
  "n04527808": "verapamil, Calan, Isoptin",
  "n04528079": "verdigris",
  "n04528256": "verge",
  "n04528354": "vermicide",
  "n04528490": "vermiculation",
  "n04528630": "vermifuge, anthelmintic, anthelminthic, helminthic",
  "n04528968": "vernier caliper, vernier micrometer",
  "n04529108": "vernier scale, vernier",
  "n04529312": "Verrazano-Narrows Bridge",
  "n04529486": "Versailles, Palace of Versailles",
  "n04529681": "vertical file",
  "n04529795": "vertical section",
  "n04529962": "vertical stabilizer, vertical stabiliser, vertical fin, tail fin, tailfin",
  "n04530176": "vertical surface",
  "n04530283": "vertical tail",
  "n04530456": "Very pistol, Verey pistol",
  "n04530566": "vessel, watercraft",
  "n04531098": "vessel",
  "n04531873": "vest, waistcoat",
  "n04532022": "vestiture",
  "n04532106": "vestment",
  "n04532398": "vest pocket",
  "n04532504": "vestry, sacristy",
  "n04532670": "viaduct",
  "n04532831": "vibraphone, vibraharp, vibes",
  "n04533042": "vibrator",
  "n04533199": "vibrator",
  "n04533373": "victory garden",
  "n04533499": "Victrola",
  "n04533594": "vicuna",
  "n04533700": "videocassette",
  "n04533802": "videocassette recorder, VCR",
  "n04533946": "videodisk, videodisc, DVD",
  "n04534127": "video recording, video",
  "n04534359": "videotape",
  "n04534520": "videotape",
  "n04534651": "viewer",
  "n04534779": "viewgraph, overhead",
  "n04534895": "vigil light, vigil candle",
  "n04535011": "vignette",
  "n04535153": "vignette",
  "n04535252": "villa",
  "n04535370": "villa",
  "n04535524": "villa",
  "n04535634": "vinblastine, Velban",
  "n04535826": "vincristine, Oncovin",
  "n04536038": "vineyard, vinery",
  "n04536153": "viol",
  "n04536335": "viola",
  "n04536465": "viola da braccio",
  "n04536595": "viola da gamba, gamba, bass viol",
  "n04536765": "viola d'amore",
  "n04536866": "violin, fiddle",
  "n04537235": "viomycin, Viocin",
  "n04537436": "virginal, pair of virginals",
  "n04537602": "virility drug, anti-impotence drug",
  "n04537816": "virtu",
  "n04537919": "virtual memory, virtual storage",
  "n04538249": "viscometer, viscosimeter",
  "n04538403": "viscose rayon, viscose",
  "n04538552": "vise, bench vise",
  "n04538759": "visible speech",
  "n04538878": "visor, vizor",
  "n04539053": "visual display unit, VDU",
  "n04539203": "vivarium",
  "n04539407": "Viyella",
  "n04539535": "V neck",
  "n04539648": "voider, gusset",
  "n04539794": "voile",
  "n04539876": "volatile storage",
  "n04540053": "volleyball",
  "n04540153": "volleyball court",
  "n04540255": "volleyball net",
  "n04540397": "voltage regulator",
  "n04540547": "voltaic battery, galvanic battery",
  "n04540761": "voltaic cell, galvanic cell, primary cell",
  "n04541136": "voltaic pile, pile, galvanic pile",
  "n04541320": "voltmeter",
  "n04541475": "volumeter",
  "n04541662": "vomitory",
  "n04541777": "von Neumann machine",
  "n04541987": "voting booth",
  "n04542095": "voting machine",
  "n04542221": "vouge",
  "n04542329": "voussoir",
  "n04542474": "vox angelica, voix celeste",
  "n04542595": "vox humana",
  "n04542715": "waders",
  "n04542858": "wading pool",
  "n04542943": "waffle iron",
  "n04543158": "wagon, waggon",
  "n04543509": "wagon, coaster wagon",
  "n04543636": "wagon tire",
  "n04543772": "wagon wheel",
  "n04543924": "wain",
  "n04543996": "wainscot, wainscoting, wainscotting",
  "n04544138": "wainscot, dado",
  "n04544325": "wainscoting, wainscotting",
  "n04544450": "waist pack, belt bag",
  "n04544626": "wake board, wakeboard",
  "n04544805": "wale, strake",
  "n04544979": "walk, walkway, paseo",
  "n04545305": "walker, baby-walker, go-cart",
  "n04545471": "walker, Zimmer, Zimmer frame",
  "n04545748": "walker",
  "n04545858": "walkie-talkie, walky-talky",
  "n04545984": "walk-in",
  "n04546081": "walking shoe",
  "n04546194": "walking stick",
  "n04546340": "Walkman",
  "n04546481": "walk-up",
  "n04546595": "walk-up apartment, walk-up",
  "n04546734": "walk-through",
  "n04546855": "wall",
  "n04547592": "wall",
  "n04547821": "wall",
  "n04547991": "wallboard, drywall, dry wall",
  "n04548280": "wall clock",
  "n04548362": "wallet, billfold, notecase, pocketbook",
  "n04548503": "wall panel",
  "n04548613": "wall plate",
  "n04548771": "wall socket, wall plug, electric outlet, electrical outlet, outlet, electric receptacle",
  "n04549028": "wall tent",
  "n04549122": "wall unit",
  "n04549276": "Walt Whitman Bridge",
  "n04549407": "wampum, peag, wampumpeag",
  "n04549629": "wand",
  "n04549721": "Wankel engine, Wankel rotary engine, epitrochoidal engine",
  "n04549919": "ward, hospital ward",
  "n04550184": "wardrobe, closet, press",
  "n04550426": "wardrobe",
  "n04550546": "wardrobe",
  "n04550676": "wardroom",
  "n04550840": "ware",
  "n04551055": "warehouse, storage warehouse",
  "n04551205": "warfarin, Coumadin",
  "n04551375": "warhead, payload, load",
  "n04551638": "warhorse",
  "n04551833": "warming pan",
  "n04551950": "warp",
  "n04552097": "war paint",
  "n04552195": "war paint",
  "n04552348": "warplane, military plane",
  "n04552551": "war room",
  "n04552696": "warship, war vessel, combat ship",
  "n04553245": "wash, wash drawing",
  "n04553389": "wash",
  "n04553561": "wash-and-wear",
  "n04553703": "washbasin, handbasin, washbowl, lavabo, wash-hand basin",
  "n04553920": "washbasin, basin, washbowl, washstand, lavatory",
  "n04554211": "washboard, splashboard",
  "n04554406": "washboard",
  "n04554523": "washcloth, washrag, flannel, face cloth",
  "n04554684": "washer, automatic washer, washing machine",
  "n04554871": "washer",
  "n04554998": "washhouse",
  "n04555101": "Washington Monument",
  "n04555291": "washroom",
  "n04555400": "washstand, wash-hand stand",
  "n04555600": "washtub",
  "n04555700": "wastepaper basket, waste-paper basket, wastebasket, waste basket, circular file",
  "n04555897": "watch, ticker",
  "n04556204": "watchband, watchstrap, wristband, watch bracelet, bracelet",
  "n04556408": "watch cap",
  "n04556533": "watch case",
  "n04556664": "watch glass",
  "n04556816": "watch key",
  "n04556948": "watchtower",
  "n04557111": "water back",
  "n04557308": "water-base paint",
  "n04557522": "water bed",
  "n04557648": "water bottle",
  "n04557751": "water butt",
  "n04557872": "water cannon, watercannon",
  "n04558059": "water cart",
  "n04558199": "water chute",
  "n04558347": "water clock, clepsydra, water glass",
  "n04558478": "water closet, closet, W.C., loo",
  "n04558578": "watercolor, water-color, watercolour, water-colour",
  "n04558804": "watercolor, water-color, watercolour, water-colour",
  "n04559023": "water-cooled reactor",
  "n04559166": "water cooler",
  "n04559275": "watercourse, waterway",
  "n04559451": "water faucet, water tap, tap, hydrant",
  "n04559620": "water filter",
  "n04559730": "water gauge, water gage, water glass",
  "n04559910": "water glass",
  "n04559994": "water hazard",
  "n04560113": "water heater, hot-water heater, hot-water tank",
  "n04560292": "watering can, watering pot",
  "n04560502": "watering cart",
  "n04560619": "water jacket",
  "n04560804": "water jug",
  "n04560882": "water jump",
  "n04561010": "water level",
  "n04561167": "water main",
  "n04561287": "water meter",
  "n04561422": "water mill",
  "n04561548": "water pistol, water gun, squirt gun, squirter",
  "n04561734": "waterproof",
  "n04561857": "waterproofing",
  "n04561965": "water pump",
  "n04562122": "water scooter, sea scooter, scooter",
  "n04562262": "water ski",
  "n04562390": "waterskin, water skin",
  "n04562496": "waterspout",
  "n04562658": "water system, water supply, water",
  "n04562935": "water tower",
  "n04563020": "water wagon, water waggon",
  "n04563204": "waterwheel, water wheel",
  "n04563413": "waterwheel, water wheel",
  "n04563560": "water wings",
  "n04563790": "waterworks",
  "n04563942": "WATS, WATS line",
  "n04564118": "wattle",
  "n04564278": "wattmeter",
  "n04564413": "waveguide, wave guide",
  "n04564581": "waxwork, wax figure",
  "n04564698": "way",
  "n04565039": "ways, shipway, slipway",
  "n04565233": "wayside, roadside",
  "n04565375": "weapon, arm, weapon system",
  "n04565963": "weapon of mass destruction, WMD, W.M.D.",
  "n04566257": "weaponry, arms, implements of war, weapons system, munition",
  "n04566561": "weapons carrier",
  "n04566756": "weathercock",
  "n04566862": "weather deck, shelter deck",
  "n04567098": "weatherglass",
  "n04567222": "weather map, weather chart",
  "n04567480": "weather radar",
  "n04567593": "weather satellite, meteorological satellite",
  "n04567746": "weather ship",
  "n04567870": "weather strip, weatherstrip, weather stripping, weatherstripping",
  "n04568069": "weathervane, weather vane, vane, wind vane",
  "n04568298": "weave",
  "n04568557": "web, entanglement",
  "n04568713": "web",
  "n04568841": "webbing",
  "n04568944": "webbing",
  "n04569063": "webcam",
  "n04569205": "wedding picture",
  "n04569338": "wedding ring, wedding band",
  "n04569520": "wedge",
  "n04569822": "wedge",
  "n04569983": "wedge heel, wedge",
  "n04570118": "wedgie",
  "n04570214": "Wedgwood",
  "n04570416": "weeder, weed-whacker",
  "n04570532": "weeds, widow's weeds",
  "n04570680": "weed, mourning band",
  "n04570815": "weekender",
  "n04570958": "weighbridge",
  "n04571088": "weight",
  "n04571292": "weight, free weight, exercising weight",
  "n04571566": "weir",
  "n04571686": "weir",
  "n04571800": "welcome wagon",
  "n04571958": "weld",
  "n04572121": "welder's mask",
  "n04572235": "weldment",
  "n04572344": "well",
  "n04572559": "well",
  "n04572800": "well",
  "n04572935": "well",
  "n04573045": "wellhead",
  "n04573129": "well point, wellpoint",
  "n04573281": "welt",
  "n04573379": "Weston cell, cadmium cell",
  "n04573513": "wet bar",
  "n04573625": "wet-bulb thermometer",
  "n04573832": "wet cell",
  "n04573937": "wet fly",
  "n04574067": "wet suit",
  "n04574234": "whacker, whopper",
  "n04574348": "whaleboat",
  "n04574471": "whaler, whaling ship",
  "n04574606": "whaling gun",
  "n04574761": "Wheatstone bridge",
  "n04574861": "wheat future",
  "n04574999": "wheel",
  "n04575723": "wheel",
  "n04575824": "wheel and axle",
  "n04576002": "wheelchair",
  "n04576211": "wheeled vehicle",
  "n04576861": "wheel lock",
  "n04576971": "wheelwork",
  "n04577139": "wherry",
  "n04577293": "wherry, Norfolk wherry",
  "n04577426": "whetstone",
  "n04577567": "whiffletree, whippletree, swingletree",
  "n04577769": "whip",
  "n04578112": "whipcord",
  "n04578220": "whipcord",
  "n04578329": "whipping post",
  "n04578459": "whipping top, whip top",
  "n04578559": "whipstitch, whipping, whipstitching",
  "n04578708": "whirler",
  "n04578801": "whisk, whisk broom",
  "n04578934": "whisk",
  "n04579056": "whiskey bottle",
  "n04579145": "whiskey jug",
  "n04579230": "whispering gallery, whispering dome",
  "n04579432": "whistle",
  "n04579667": "whistle",
  "n04579795": "whistle stop, flag stop, way station",
  "n04579986": "white",
  "n04580126": "white flag, flag of truce",
  "n04580298": "white goods, household linen",
  "n04580493": "white goods",
  "n04580777": "White House",
  "n04580994": "white tie",
  "n04581102": "whitewash",
  "n04581262": "whizbang, whizzbang",
  "n04581425": "whizbang, whizzbang, whizbang shell",
  "n04581595": "whorehouse, brothel, bordello, bagnio, house of prostitution, house of ill repute, bawdyhouse, cathouse, sporting house",
  "n04581829": "wick, taper",
  "n04582044": "wick",
  "n04582205": "wicker, wickerwork, caning",
  "n04582349": "wicker basket",
  "n04582454": "wicket, lattice, grille",
  "n04582625": "wicket, wicket door, wicket gate",
  "n04582771": "wicket, hoop",
  "n04582869": "wicket",
  "n04583022": "wickiup, wikiup",
  "n04583212": "wide-angle lens, fisheye lens",
  "n04583477": "wide area network, WAN",
  "n04583620": "widebody aircraft, wide-body aircraft, wide-body, twin-aisle airplane",
  "n04583776": "wide screen",
  "n04583888": "wide wale",
  "n04583967": "widow's walk",
  "n04584056": "Wiffle, Wiffle Ball",
  "n04584207": "wig",
  "n04584373": "wigwam",
  "n04584512": "wild card",
  "n04584639": "wildcat well, wildcat",
  "n04584794": "willow",
  "n04584944": "willowware, willow-pattern",
  "n04585128": "Wilton, Wilton carpet",
  "n04585318": "wimple",
  "n04585456": "wincey",
  "n04585626": "winceyette",
  "n04585745": "winch, windlass",
  "n04585980": "Winchester",
  "n04586072": "windbreak, shelterbelt",
  "n04586225": "wind chime, wind bell",
  "n04586421": "winder",
  "n04586581": "winder, key",
  "n04586761": "wind farm, wind park, wind energy facility",
  "n04586932": "wind instrument, wind",
  "n04587327": "windjammer",
  "n04587404": "windmill, aerogenerator, wind generator",
  "n04587559": "windmill",
  "n04587648": "window",
  "n04588365": "window",
  "n04588587": "window",
  "n04588739": "window",
  "n04588986": "window",
  "n04589190": "window blind",
  "n04589325": "window box",
  "n04589434": "window envelope",
  "n04589593": "window frame",
  "n04589745": "windowpane, window",
  "n04589890": "window screen",
  "n04590021": "window seat",
  "n04590129": "window shade",
  "n04590263": "windowsill",
  "n04590416": "wind rose",
  "n04590553": "windshield, windscreen",
  "n04590746": "windshield wiper, windscreen wiper, wiper, wiper blade",
  "n04590933": "Windsor chair",
  "n04591056": "Windsor knot",
  "n04591157": "Windsor tie",
  "n04591249": "wind tee",
  "n04591359": "wind tunnel",
  "n04591517": "wind turbine",
  "n04591631": "wine bar",
  "n04591713": "wine bottle",
  "n04591887": "wine bucket, wine cooler",
  "n04592005": "wine cask, wine barrel",
  "n04592099": "wineglass",
  "n04592243": "wineglass heel",
  "n04592356": "winepress",
  "n04592465": "winery, wine maker",
  "n04592596": "wineskin",
  "n04592741": "wing",
  "n04592962": "wing, offstage, backstage",
  "n04593077": "wing chair",
  "n04593185": "wing nut, wing-nut, wing screw, butterfly nut, thumbnut",
  "n04593376": "wing tip",
  "n04593524": "wing tip",
  "n04593629": "winker, blinker, blinder",
  "n04593866": "wiper, wiper arm, contact arm",
  "n04594114": "wiper motor",
  "n04594218": "wire",
  "n04594489": "wire, conducting wire",
  "n04594742": "wire cloth",
  "n04594828": "wire cutter",
  "n04594919": "wire gauge, wire gage",
  "n04595028": "wireless local area network, WLAN, wireless fidelity, WiFi",
  "n04595285": "wire matrix printer, wire printer, stylus printer",
  "n04595501": "wire recorder",
  "n04595611": "wire stripper",
  "n04595762": "wirework, grillwork",
  "n04595855": "wiring",
  "n04595998": "wiring diagram",
  "n04596116": "wishing cap",
  "n04596224": "witch hazel, wych hazel",
  "n04596397": "withe",
  "n04596492": "witness box, witness stand",
  "n04596630": "wobbler",
  "n04596742": "wok",
  "n04596852": "woman's clothing",
  "n04597066": "wood",
  "n04597309": "woodcarving",
  "n04597400": "wood chisel",
  "n04597536": "woodcut, wood block, wood engraving",
  "n04597706": "woodcut, wood engraving",
  "n04597804": "woodenware",
  "n04597913": "wooden spoon",
  "n04598010": "wooden spoon",
  "n04598136": "woodscrew",
  "n04598318": "woodshed",
  "n04598416": "wood vise, woodworking vise, shoulder vise",
  "n04598582": "woodwind, woodwind instrument, wood",
  "n04598792": "woodwork",
  "n04598965": "woof, weft, filling, pick",
  "n04599124": "woofer",
  "n04599235": "wool, woolen, woollen",
  "n04599396": "work, piece of work",
  "n04600131": "worldly possession, worldly good",
  "n04600312": "workbasket, workbox, workbag",
  "n04600486": "workbench, work bench, bench",
  "n04600627": "workboard, work-board",
  "n04600764": "work camp, prison camp, prison farm",
  "n04600912": "work-clothing, work-clothes",
  "n04601041": "workhouse",
  "n04601159": "workhouse",
  "n04601291": "workhorse",
  "n04601473": "working, workings",
  "n04601587": "work in progress",
  "n04601690": "work of art",
  "n04601938": "workpiece",
  "n04602044": "workplace, work",
  "n04602762": "workroom",
  "n04602840": "works, workings",
  "n04602956": "work-shirt",
  "n04603081": "workshop, shop",
  "n04603399": "workstation",
  "n04603558": "work surface",
  "n04603729": "worktable, work table",
  "n04603872": "workwear",
  "n04604009": "World Trade Center, WTC, twin towers",
  "n04604276": "World Wide Web, WWW, web",
  "n04604513": "worm",
  "n04604644": "worm fence, snake fence, snake-rail fence, Virginia fence",
  "n04604806": "worm gear",
  "n04605057": "worm wheel",
  "n04605163": "worsted",
  "n04605321": "worsted, worsted yarn",
  "n04605446": "wrap, wrapper",
  "n04605572": "wraparound",
  "n04605726": "wrapping, wrap, wrapper",
  "n04606014": "wreath, garland, coronal, chaplet, lei",
  "n04606251": "wreck",
  "n04606358": "wreckage",
  "n04606574": "wrench, spanner",
  "n04607035": "wrestling mat",
  "n04607141": "wrestling ring",
  "n04607242": "wringer",
  "n04607398": "wristband",
  "n04607532": "wristlet, wrist band",
  "n04607640": "wrist pad",
  "n04607759": "wrist pin, gudgeon pin",
  "n04607869": "wristwatch, wrist watch",
  "n04607982": "writing arm",
  "n04608127": "writing board",
  "n04608329": "writing desk",
  "n04608435": "writing desk",
  "n04608567": "writing implement",
  "n04608809": "xerographic printer",
  "n04608923": "Xerox, xerographic copier, Xerox machine",
  "n04609235": "xerox, xerox copy",
  "n04609354": "X-OR circuit, XOR circuit, XOR gate",
  "n04609531": "X-ray film",
  "n04609651": "X-ray machine",
  "n04609811": "X-ray tube",
  "n04610013": "yacht, racing yacht",
  "n04610176": "yacht chair",
  "n04610274": "yagi, Yagi aerial",
  "n04610368": "Yale University, Yale",
  "n04610503": "yard",
  "n04610676": "yard",
  "n04610879": "yard, grounds, curtilage",
  "n04611154": "yard, railway yard, railyard",
  "n04611351": "yardarm",
  "n04611470": "yarder, yard donkey",
  "n04611654": "yard goods, piece goods",
  "n04611795": "yard marker",
  "n04611916": "yardstick, yard measure",
  "n04612026": "yarmulke, yarmulka, yarmelke",
  "n04612159": "yashmak, yashmac",
  "n04612257": "yataghan",
  "n04612373": "yawl, dandy",
  "n04612504": "yawl",
  "n04612623": "yellow jack",
  "n04612722": "yield, fruit",
  "n04612840": "yoke",
  "n04613015": "yoke",
  "n04613158": "yoke, coupling",
  "n04613350": "yoke",
  "n04613555": "yo-yo",
  "n04613696": "yurt",
  "n04613939": "Zamboni",
  "n04614083": "zapper",
  "n04614244": "zarf",
  "n04614372": "zeppelin, Graf Zeppelin",
  "n04614505": "zero",
  "n04614655": "ziggurat, zikkurat, zikurat",
  "n04614844": "zill",
  "n04615018": "zinc ointment",
  "n04615149": "zip gun",
  "n04615226": "zither, cither, zithern",
  "n04615456": "zodiac",
  "n04615644": "zoot suit",
  "n04615728": "ramp",
  "n04615866": "human nature",
  "n04616059": "trait",
  "n04616916": "character",
  "n04617106": "unit character",
  "n04617289": "thing",
  "n04617441": "common denominator",
  "n04617562": "personality",
  "n04618070": "identity, personal identity, individuality",
  "n04618359": "gender identity",
  "n04618581": "identification",
  "n04618781": "personhood",
  "n04618921": "personableness",
  "n04619083": "anal personality, anal retentive personality",
  "n04619420": "genital personality",
  "n04619596": "narcissistic personality",
  "n04619791": "obsessive-compulsive personality",
  "n04619946": "oral personality",
  "n04620216": "character, fiber, fibre",
  "n04620558": "spirit",
  "n04620741": "outwardness",
  "n04621010": "inwardness, internality",
  "n04621314": "spirituality, spiritualism, spiritism, otherworldliness",
  "n04621524": "worldliness",
  "n04621738": "extraversion, extroversion",
  "n04621963": "introversion",
  "n04622216": "ambiversion",
  "n04622415": "aloneness, loneliness, lonesomeness, solitariness",
  "n04622672": "friendlessness",
  "n04622772": "reclusiveness",
  "n04622932": "privacy, privateness, seclusion",
  "n04623113": "nature",
  "n04623443": "animality, animal nature",
  "n04623612": "disposition, temperament",
  "n04624289": "complexion",
  "n04624517": "animalism, physicality",
  "n04624687": "bloodiness, bloodthirstiness",
  "n04624826": "heart, spirit",
  "n04624959": "nervousness",
  "n04625129": "esprit de corps, morale, team spirit",
  "n04625284": "restlessness, uneasiness, queasiness",
  "n04625515": "jactitation, jactation",
  "n04625716": "skittishness, restiveness",
  "n04625882": "compulsiveness, compulsivity",
  "n04626062": "obsessiveness, obsessivity",
  "n04626194": "workaholism",
  "n04626280": "emotionality, emotionalism",
  "n04626705": "drama",
  "n04626879": "demonstrativeness",
  "n04627000": "affectionateness, fondness, lovingness, warmth",
  "n04627241": "tenderness",
  "n04627367": "uxoriousness",
  "n04627506": "mawkishness, sentimentality, drippiness, mushiness, soupiness, sloppiness",
  "n04627809": "corn",
  "n04627936": "schmaltz, shmaltz, schmalz",
  "n04628080": "sentimentalism",
  "n04628192": "heat, warmth, passion",
  "n04628336": "fieriness",
  "n04628466": "temperament",
  "n04628632": "moodiness",
  "n04628747": "blood",
  "n04628850": "excitability, excitableness, volatility",
  "n04629030": "boiling point",
  "n04629194": "unemotionality, emotionlessness",
  "n04629417": "blandness",
  "n04629604": "coldness, coolness, frigidity, frigidness, iciness, chilliness",
  "n04629958": "stone",
  "n04630137": "dispassion, dispassionateness, dryness",
  "n04630378": "stoicism, stolidity, stolidness",
  "n04630547": "tepidness, lukewarmness",
  "n04630689": "cheerfulness, cheer, sunniness, sunshine",
  "n04631067": "good-temperedness, good-humoredness, good-humouredness, good-naturedness",
  "n04631298": "uncheerfulness",
  "n04631470": "gloominess, lugubriousness, sadness",
  "n04631700": "animation, spiritedness, invigoration, brio, vivification",
  "n04632063": "chirpiness",
  "n04632157": "liveliness, life, spirit, sprightliness",
  "n04632703": "pertness",
  "n04632866": "airiness, delicacy",
  "n04632963": "alacrity, briskness, smartness",
  "n04633197": "energy, muscularity, vigor, vigour, vim",
  "n04633453": "vitality, verve",
  "n04633570": "elan",
  "n04633716": "esprit",
  "n04633797": "breeziness, jauntiness",
  "n04633959": "irrepressibility, buoyancy",
  "n04634161": "high-spiritedness",
  "n04634299": "vivacity",
  "n04634415": "mettlesomeness",
  "n04634540": "exuberance, enthusiasm, ebullience",
  "n04634739": "lyricism",
  "n04634833": "pep, peppiness, ginger",
  "n04635010": "inanition",
  "n04635104": "activeness, activity",
  "n04635482": "dynamism, pizzazz, pizzaz, oomph, zing",
  "n04635631": "inactiveness, inactivity, inertia",
  "n04635953": "languor, lethargy, sluggishness, phlegm, flatness",
  "n04636250": "restfulness",
  "n04636397": "passivity, passiveness",
  "n04636610": "apathy, indifference, numbness, spiritlessness",
  "n04636881": "listlessness, torpidity, torpidness, torpor",
  "n04637108": "indolence, laziness",
  "n04637290": "faineance, idleness",
  "n04637444": "sloth, slothfulness",
  "n04637571": "shiftlessness",
  "n04637722": "perfectionism",
  "n04637923": "permissiveness, tolerance",
  "n04638175": "toleration, acceptance, sufferance",
  "n04638472": "self acceptance",
  "n04638585": "indulgence, lenience, leniency",
  "n04638857": "softness",
  "n04639030": "overtolerance",
  "n04639113": "unpermissiveness, restrictiveness",
  "n04639371": "sternness, strictness",
  "n04639591": "Puritanism",
  "n04639732": "severity, severeness, harshness, rigor, rigour, rigorousness, rigourousness, inclemency, hardness, stiffness",
  "n04640176": "good nature",
  "n04640356": "grace, good will, goodwill",
  "n04640538": "patience, forbearance, longanimity",
  "n04640722": "easygoingness",
  "n04640830": "risibility",
  "n04640927": "agreeableness, agreeability",
  "n04641153": "complaisance, compliance, compliancy, obligingness, deference",
  "n04641447": "ill nature",
  "n04641700": "crabbiness, crabbedness, crossness",
  "n04641869": "crankiness, crotchetiness, contrariness, grumpiness",
  "n04642057": "sulkiness, sullenness, moroseness, sourness",
  "n04642258": "temper, biliousness, irritability, peevishness, pettishness, snappishness, surliness",
  "n04642596": "impatience",
  "n04642746": "intolerance",
  "n04642866": "shrewishness",
  "n04642980": "querulousness",
  "n04643099": "asperity, sharpness",
  "n04643221": "disagreeableness",
  "n04643397": "bitterness, acrimony, acerbity, jaundice, tartness, thorniness",
  "n04643662": "aggressiveness, belligerence, pugnacity",
  "n04643875": "bellicosity, bellicoseness",
  "n04643979": "quarrelsomeness, contentiousness",
  "n04644161": "truculence, truculency",
  "n04644306": "litigiousness",
  "n04644512": "willingness",
  "n04644719": "readiness, eagerness, zeal, forwardness",
  "n04645020": "receptiveness, receptivity, openness",
  "n04645473": "wholeheartedness",
  "n04645599": "unwillingness, involuntariness",
  "n04645943": "reluctance, hesitancy, hesitation, disinclination, indisposition",
  "n04646372": "resistance",
  "n04646548": "seriousness, earnestness, serious-mindedness, sincerity",
  "n04646990": "committedness, commitment",
  "n04647185": "investment",
  "n04647478": "graveness, gravity, sobriety, soberness, somberness, sombreness",
  "n04647826": "sedateness, staidness, solemnity, solemness",
  "n04648059": "stodginess, stuffiness",
  "n04648207": "frivolity, frivolousness",
  "n04648440": "giddiness, silliness",
  "n04648596": "lightsomeness, lightness",
  "n04648749": "levity",
  "n04648866": "flippancy, light-mindedness",
  "n04649051": "jocoseness, jocosity, merriness, humorousness",
  "n04649261": "playfulness, fun",
  "n04649560": "facetiousness",
  "n04649651": "impertinence, perkiness, pertness, sauciness, archness",
  "n04649835": "friskiness, frolicsomeness, sportiveness",
  "n04650010": "impishness, mischievousness, puckishness, whimsicality",
  "n04650201": "humor, humour, sense of humor, sense of humour",
  "n04650527": "communicativeness",
  "n04650731": "frankness, outspokenness",
  "n04650913": "bluffness",
  "n04651009": "effusiveness, expansiveness, expansivity",
  "n04651195": "fluency, volubility, articulateness",
  "n04651382": "garrulity, garrulousness, loquaciousness, loquacity, talkativeness",
  "n04651683": "leresis",
  "n04651784": "uncommunicativeness",
  "n04651974": "muteness, silence",
  "n04652177": "secrecy, secretiveness, silence",
  "n04652345": "mum",
  "n04652438": "reserve, reticence, taciturnity",
  "n04652635": "sociality",
  "n04652930": "sociability, sociableness",
  "n04653357": "conviviality, joviality",
  "n04653479": "companionability, companionableness",
  "n04653627": "chumminess, camaraderie, comradeliness, comradery, comradeship",
  "n04653869": "gregariousness",
  "n04654066": "openness, nakedness",
  "n04654337": "friendliness",
  "n04654652": "affability, affableness, amiability, amiableness, bonhomie, geniality",
  "n04654998": "amicability, amicableness",
  "n04655168": "condescension, condescendingness",
  "n04655442": "familiarity, intimacy, closeness",
  "n04655649": "approachability, accessibility",
  "n04655829": "congeniality",
  "n04655929": "amity, cordiality",
  "n04656051": "neighborliness, neighbourliness, good-neighborliness, good-neighbourliness",
  "n04656282": "hospitableness",
  "n04656448": "mellowness",
  "n04656598": "sweetness and light",
  "n04656748": "unsociability, unsociableness",
  "n04656996": "aloofness, remoteness, standoffishness, withdrawnness",
  "n04657244": "unapproachability",
  "n04657407": "closeness, secretiveness",
  "n04657631": "furtiveness, sneakiness, stealthiness",
  "n04657876": "unfriendliness",
  "n04658106": "hostility, ill will",
  "n04658268": "aggression",
  "n04658361": "virulence, virulency",
  "n04658524": "misanthropy",
  "n04658686": "uncongeniality",
  "n04658816": "unneighborliness",
  "n04658942": "inhospitableness",
  "n04659090": "adaptability",
  "n04659287": "flexibility, flexibleness",
  "n04659567": "wiggle room",
  "n04659730": "pliability, pliancy, pliantness, suppleness",
  "n04660080": "unadaptability",
  "n04660261": "inflexibility, rigidity, rigidness",
  "n04660536": "thoughtfulness",
  "n04660805": "pensiveness, meditativeness, contemplativeness",
  "n04660981": "introspectiveness",
  "n04661151": "deliberation, deliberateness",
  "n04661389": "intentionality",
  "n04661546": "reflectiveness, reflectivity",
  "n04661706": "unthoughtfulness, thoughtlessness",
  "n04661926": "recklessness, foolhardiness, rashness",
  "n04662182": "adventurism",
  "n04662283": "brashness",
  "n04662390": "desperation",
  "n04662504": "impulsiveness",
  "n04662730": "impetuousness, impetuosity",
  "n04662858": "hastiness",
  "n04662951": "attentiveness",
  "n04663204": "attentiveness",
  "n04663319": "inattentiveness",
  "n04663494": "carefulness",
  "n04663763": "mindfulness, heedfulness",
  "n04664058": "caution, cautiousness, carefulness",
  "n04664314": "precaution",
  "n04664413": "wariness, chariness",
  "n04664628": "alertness, sharp-sightedness, on the qui vive",
  "n04664778": "watchfulness, vigilance, weather eye",
  "n04664964": "carelessness, sloppiness",
  "n04665210": "incaution, incautiousness",
  "n04665406": "unwariness",
  "n04665543": "unmindfulness, heedlessness, inadvertence, inadvertency",
  "n04665813": "negligence, neglect, neglectfulness",
  "n04666083": "delinquency, dereliction, willful neglect",
  "n04666416": "laxness, laxity, remissness, slackness",
  "n04666615": "masculinity",
  "n04666837": "manfulness, manliness, virility",
  "n04667076": "boyishness",
  "n04667180": "machismo",
  "n04667258": "hoydenism, tomboyishness",
  "n04667406": "femininity, muliebrity",
  "n04667660": "womanliness, womanlike",
  "n04667821": "ladylikeness",
  "n04667923": "maidenliness",
  "n04668033": "girlishness",
  "n04668139": "effeminacy, effeminateness, sissiness, softness, womanishness, unmanliness",
  "n04668713": "emasculation",
  "n04668819": "trustworthiness, trustiness",
  "n04669063": "creditworthiness",
  "n04669247": "responsibility, responsibleness",
  "n04669692": "fault",
  "n04669828": "accountability, answerability, answerableness",
  "n04670022": "dependability, dependableness, reliability, reliableness",
  "n04670531": "untrustworthiness, untrustiness",
  "n04670746": "irresponsibility, irresponsibleness",
  "n04671075": "solidity, solidness",
  "n04671394": "undependability, undependableness, unreliability, unreliableness",
  "n04671841": "flightiness, arbitrariness, whimsicality, whimsy, whimsey, capriciousness",
  "n04672210": "carefreeness",
  "n04672355": "conscientiousness, painstakingness",
  "n04672605": "meticulousness, meticulosity, punctiliousness, scrupulousness",
  "n04672858": "thoroughness",
  "n04673006": "diligence",
  "n04673173": "strictness, stringency",
  "n04673341": "unconscientiousness",
  "n04673470": "nonchalance, unconcern, indifference",
  "n04673668": "recommendation, passport",
  "n04673965": "appearance, visual aspect",
  "n04674624": "agerasia",
  "n04674715": "look",
  "n04674861": "view",
  "n04674968": "color, colour",
  "n04675193": "complexion",
  "n04675314": "impression, effect",
  "n04675646": "figure",
  "n04675777": "image",
  "n04676064": "mark",
  "n04676308": "perspective, linear perspective",
  "n04676540": "phase",
  "n04676830": "tout ensemble",
  "n04676959": "vanishing point",
  "n04677113": "superficies",
  "n04677385": "format",
  "n04677514": "form, shape, cast",
  "n04677716": "persona, image",
  "n04677952": "semblance, gloss, color, colour",
  "n04678401": "color of law, colour of law",
  "n04678712": "simulacrum",
  "n04678804": "face value",
  "n04678908": "guise, pretense, pretence, pretext",
  "n04679074": "disguise, camouflage",
  "n04679300": "verisimilitude",
  "n04679419": "face",
  "n04679549": "countenance, visage",
  "n04679738": "expression, look, aspect, facial expression, face",
  "n04680027": "leer",
  "n04680133": "poker face",
  "n04680285": "marking",
  "n04680465": "band, banding, stria, striation",
  "n04680752": "collar",
  "n04680893": "stretch mark",
  "n04681058": "blaze",
  "n04681230": "speck, pinpoint",
  "n04681387": "crisscross, cross, mark",
  "n04681621": "eyespot, ocellus",
  "n04681797": "hatch, hatching, crosshatch, hachure",
  "n04682018": "shading",
  "n04682184": "nebula",
  "n04682319": "splash",
  "n04682462": "spot, speckle, dapple, patch, fleck, maculation",
  "n04683002": "worn spot, fret",
  "n04683136": "stripe, streak, bar",
  "n04683453": "hairiness, pilosity",
  "n04683600": "hirsuteness, hirsutism",
  "n04683707": "hairlessness",
  "n04683814": "beauty",
  "n04684255": "raw beauty",
  "n04684358": "glory, resplendence, resplendency",
  "n04684654": "exquisiteness",
  "n04684765": "picturesqueness",
  "n04684872": "pleasingness",
  "n04685062": "pulchritude",
  "n04685195": "glamor, glamour",
  "n04685396": "comeliness, fairness, loveliness, beauteousness",
  "n04685649": "prettiness, cuteness",
  "n04685840": "handsomeness, good looks",
  "n04686003": "attractiveness",
  "n04686257": "adorability, adorableness",
  "n04686388": "bewitchery, beguilement, animal magnetism",
  "n04686537": "charisma, personal appeal, personal magnetism",
  "n04686748": "curvaceousness, shapeliness, voluptuousness",
  "n04686935": "sex appeal, desirability, desirableness, oomph",
  "n04687119": "sultriness",
  "n04687333": "appeal, appealingness, charm",
  "n04687633": "siren call, siren song",
  "n04687822": "spiff",
  "n04687957": "winsomeness",
  "n04688058": "associability, associableness",
  "n04688246": "attraction, attractiveness",
  "n04688648": "affinity",
  "n04688842": "allure, allurement, temptingness",
  "n04689048": "invitation",
  "n04689198": "binding",
  "n04689330": "drawing power",
  "n04689450": "fascination",
  "n04689660": "lure, enticement, come-on",
  "n04689834": "sexual attraction",
  "n04689942": "show-stopper, showstopper",
  "n04690196": "ugliness",
  "n04690476": "unsightliness",
  "n04690592": "grotesqueness, grotesquery, grotesquerie",
  "n04690769": "garishness, gaudiness",
  "n04690933": "unpleasingness",
  "n04691061": "hideousness",
  "n04691178": "disfigurement, disfiguration, deformity",
  "n04691476": "unattractiveness",
  "n04691651": "homeliness, plainness",
  "n04691861": "shapelessness",
  "n04691992": "ballast",
  "n04692157": "blemish, defect, mar",
  "n04692638": "birthmark, nevus",
  "n04692787": "chatter mark",
  "n04692908": "check, chip",
  "n04693096": "crack",
  "n04693274": "craze",
  "n04693384": "dent, ding, gouge, nick",
  "n04693557": "dig",
  "n04693717": "eyesore",
  "n04693804": "mole",
  "n04693900": "scratch, scrape, scar, mark",
  "n04694090": "burn, burn mark",
  "n04694293": "cigarette burn",
  "n04694441": "smudge, spot, blot, daub, smear, smirch, slur",
  "n04694809": "blotch, splodge, splotch",
  "n04694980": "fingermark, fingerprint",
  "n04695102": "inkblot",
  "n04695176": "stain, discoloration, discolouration",
  "n04695504": "scorch",
  "n04695605": "bloodstain",
  "n04695693": "iron mold, iron mould",
  "n04695805": "mud stain",
  "n04695884": "oil stain",
  "n04695963": "tarnish",
  "n04696085": "stigma",
  "n04696193": "port-wine stain, nevus flammeus",
  "n04696316": "strawberry, strawberry mark, hemangioma simplex",
  "n04696432": "wart, verruca",
  "n04696695": "common wart",
  "n04696797": "genital wart, venereal wart, condyloma acuminatum, verruca acuminata",
  "n04696969": "juvenile wart",
  "n04697084": "plantar wart",
  "n04697267": "plainness",
  "n04697442": "chasteness, restraint, simplicity, simpleness",
  "n04697666": "austereness, severity, severeness",
  "n04697819": "bareness, starkness",
  "n04698112": "ornateness, elaborateness",
  "n04698307": "baroque, baroqueness",
  "n04698511": "classical style",
  "n04698656": "order",
  "n04698998": "Doric order, Dorian order",
  "n04699157": "Ionic order, Ionian order",
  "n04699298": "Corinthian order",
  "n04699474": "Composite order",
  "n04699637": "Tuscan order",
  "n04699767": "rococo",
  "n04699936": "flamboyance, floridness, floridity, showiness",
  "n04700199": "fussiness",
  "n04700327": "decorativeness",
  "n04700496": "etiolation",
  "n04700642": "coating, finish, finishing",
  "n04701039": "glaze",
  "n04701162": "luster, lustre",
  "n04701267": "shoeshine",
  "n04701460": "clearness, clarity, uncloudedness",
  "n04701943": "pellucidness, pellucidity, limpidity",
  "n04702127": "transparency, transparence, transparentness",
  "n04702351": "translucence, translucency, semitransparency",
  "n04702545": "visibility",
  "n04702688": "distinctness, sharpness",
  "n04702957": "definition",
  "n04703104": "discernability, legibility",
  "n04703235": "focus",
  "n04703424": "opacity, opaqueness",
  "n04703698": "cloudiness, murkiness, muddiness",
  "n04703932": "turbidity, turbidness",
  "n04704116": "haziness, mistiness, steaminess, vaporousness, vapourousness",
  "n04704346": "indistinctness, softness, blurriness, fogginess, fuzziness",
  "n04704675": "dimness, faintness",
  "n04704839": "vagueness",
  "n04705013": "divisibility",
  "n04705196": "fissiparity",
  "n04705324": "sharpness, keenness",
  "n04705536": "acuteness",
  "n04705671": "dullness, bluntness",
  "n04705950": "obtuseness",
  "n04706087": "conspicuousness",
  "n04706290": "obviousness, noticeability, noticeableness, patency",
  "n04706727": "apparentness, apparency",
  "n04706882": "blatancy",
  "n04707110": "obtrusiveness",
  "n04707252": "boldness, strikingness",
  "n04707409": "predomination, predominance",
  "n04707636": "inconspicuousness",
  "n04707808": "unnoticeableness",
  "n04707964": "unobtrusiveness",
  "n04708113": "ease, easiness, simplicity, simpleness",
  "n04708543": "effortlessness",
  "n04708796": "facility, readiness",
  "n04709011": "smoothness",
  "n04709253": "difficulty, difficultness",
  "n04709585": "effortfulness",
  "n04709759": "arduousness, strenuousness",
  "n04709944": "laboriousness, operoseness, toilsomeness",
  "n04710127": "asperity, grimness, hardship, rigor, rigour, severity, severeness, rigorousness, rigourousness",
  "n04710390": "sternness",
  "n04710588": "hardness, ruggedness",
  "n04710866": "formidability, toughness",
  "n04711031": "burdensomeness, heaviness, onerousness, oppressiveness",
  "n04711256": "subtlety, niceness",
  "n04711435": "troublesomeness, inconvenience, worriment",
  "n04711665": "awkwardness, cumbersomeness, unwieldiness",
  "n04711919": "flea bite",
  "n04712001": "fly in the ointment",
  "n04712130": "unwieldiness",
  "n04712405": "combustibility, combustibleness, burnability",
  "n04712568": "flammability, inflammability",
  "n04712735": "compatibility",
  "n04712978": "congenialness, congeniality",
  "n04713118": "harmony, harmoniousness",
  "n04713332": "accord",
  "n04713428": "agreement, correspondence",
  "n04713692": "conformity, conformance",
  "n04713853": "justness, rightness, nicety",
  "n04714058": "normality",
  "n04714156": "congruity, congruousness, congruence",
  "n04714440": "incompatibility",
  "n04714679": "conflict",
  "n04714847": "incongruity, incongruousness",
  "n04715075": "irony",
  "n04715308": "Socratic irony",
  "n04715487": "suitability, suitableness",
  "n04715818": "arability",
  "n04715947": "appropriateness",
  "n04716210": "felicity, felicitousness",
  "n04716491": "aptness, appositeness",
  "n04716703": "ticket, just the ticket",
  "n04716864": "fitness, fittingness",
  "n04717139": "qualification, making",
  "n04717552": "eligibility",
  "n04717787": "insurability",
  "n04718030": "marriageability",
  "n04718134": "ineligibility",
  "n04718306": "uninsurability",
  "n04718563": "convenience",
  "n04718826": "opportuneness, patness, timeliness",
  "n04718999": "handiness, accessibility, availability, availableness",
  "n04719376": "command",
  "n04719502": "impressiveness",
  "n04719661": "navigability",
  "n04719805": "neediness",
  "n04720024": "painfulness, distressingness",
  "n04720226": "sharpness",
  "n04720393": "piquancy, piquance, piquantness",
  "n04720536": "publicity",
  "n04720685": "spinnability",
  "n04720879": "spinnbarkeit",
  "n04721058": "unsuitability, unsuitableness, ineptness",
  "n04721428": "inaptness, inappositeness",
  "n04721650": "inappropriateness, unworthiness",
  "n04722051": "infelicity",
  "n04722231": "habitability, habitableness",
  "n04722373": "unfitness",
  "n04722574": "disqualification",
  "n04722715": "inconvenience",
  "n04722910": "inaccessibility, unavailability",
  "n04723107": "inopportuneness, untimeliness",
  "n04723286": "ethos",
  "n04723436": "eidos",
  "n04723622": "protectiveness",
  "n04723816": "quality",
  "n04726724": "nature",
  "n04726938": "humanness, humanity, manhood",
  "n04727214": "air, aura, atmosphere",
  "n04727559": "mystique",
  "n04727694": "note",
  "n04727883": "vibration, vibe",
  "n04728068": "quality, caliber, calibre",
  "n04728376": "superiority, high quality",
  "n04728604": "fineness, choiceness",
  "n04728786": "excellence",
  "n04728999": "ultimate",
  "n04729127": "admirability, admirableness, wonderfulness",
  "n04729328": "impressiveness, grandness, magnificence, richness",
  "n04729710": "expansiveness, expansivity",
  "n04729984": "stateliness, majesty, loftiness",
  "n04730191": "first class",
  "n04730285": "first water",
  "n04730367": "ingenuity, ingeniousness, cleverness",
  "n04730580": "inferiority, low quality",
  "n04730797": "poorness",
  "n04730985": "scrawniness, scrubbiness",
  "n04731207": "second class",
  "n04731309": "wretchedness",
  "n04731497": "characteristic",
  "n04731925": "point, spot",
  "n04732067": "point",
  "n04732237": "salability, salableness",
  "n04732386": "selling point",
  "n04732543": "hallmark, trademark, earmark, stylemark",
  "n04732709": "mold, mould",
  "n04732852": "saving grace",
  "n04733118": "aspect",
  "n04733204": "gaseousness",
  "n04733347": "bubbliness, effervescence, frothiness",
  "n04733539": "foaminess",
  "n04733640": "changeableness, changeability",
  "n04734145": "commutability, transmutability",
  "n04734272": "fluidity, fluidness",
  "n04734551": "reversibility",
  "n04734698": "shiftiness",
  "n04734885": "inconstancy, changefulness",
  "n04735075": "capriciousness, unpredictability",
  "n04735233": "variability, variableness, variance",
  "n04735556": "variedness",
  "n04735711": "diversity",
  "n04735828": "variegation",
  "n04735929": "exchangeability, interchangeability, interchangeableness, fungibility",
  "n04736337": "duality",
  "n04736516": "transferability",
  "n04736757": "convertibility",
  "n04737020": "inconvertibility",
  "n04737234": "replaceability, substitutability, commutability",
  "n04737430": "liquidity",
  "n04737568": "permutability, permutableness, transposability",
  "n04737743": "progressiveness, progressivity",
  "n04737934": "changelessness, unchangeability, unchangeableness, unchangingness",
  "n04738398": "absoluteness",
  "n04738641": "constancy, stability",
  "n04738995": "invariance",
  "n04739262": "metastability",
  "n04739535": "monotony",
  "n04739630": "innateness",
  "n04739768": "irreversibility",
  "n04739932": "invariability, invariableness, invariance",
  "n04740173": "unvariedness",
  "n04740326": "monotony, humdrum, sameness",
  "n04740655": "fixedness, unalterability",
  "n04740864": "unexchangeability",
  "n04741075": "incommutability",
  "n04741195": "irreplaceableness",
  "n04741311": "mutability, mutableness",
  "n04741530": "alterability",
  "n04741655": "vicissitude",
  "n04741807": "immutability, immutableness, fixity",
  "n04742084": "unalterability",
  "n04742251": "incurability",
  "n04742418": "agelessness",
  "n04742535": "sameness",
  "n04742766": "otherness, distinctness, separateness",
  "n04743024": "identity, identicalness, indistinguishability",
  "n04743370": "oneness, unity",
  "n04743486": "selfsameness",
  "n04743605": "similarity",
  "n04744032": "approximation",
  "n04744161": "homogeny",
  "n04744319": "homology",
  "n04744555": "homomorphism, homomorphy",
  "n04744645": "isomorphism, isomorphy",
  "n04744814": "likeness, alikeness, similitude",
  "n04745240": "parallelism, correspondence",
  "n04745370": "uniformity, uniformness",
  "n04745679": "homogeneity, homogeneousness",
  "n04745932": "consistency, consistence",
  "n04746134": "approach",
  "n04746262": "sort",
  "n04746430": "analogue, analog, parallel",
  "n04746607": "echo",
  "n04746842": "comparison, compare, equivalence, comparability",
  "n04747115": "mirror image, reflection, reflexion",
  "n04747246": "naturalness",
  "n04747445": "resemblance",
  "n04747616": "spitting image",
  "n04747710": "mutual resemblance",
  "n04747798": "affinity",
  "n04747899": "equality",
  "n04748187": "equatability",
  "n04748273": "equivalence",
  "n04748426": "parity",
  "n04748498": "evenness",
  "n04748654": "isometry",
  "n04748836": "difference",
  "n04749310": "differential",
  "n04749439": "differentia",
  "n04749572": "distinction",
  "n04749709": "discrepancy, disagreement, divergence, variance",
  "n04749991": "allowance, leeway, margin, tolerance",
  "n04750164": "dissimilarity, unsimilarity",
  "n04750414": "disparateness, distinctiveness",
  "n04750547": "heterology",
  "n04750764": "unlikeness, dissimilitude",
  "n04750940": "nonuniformity",
  "n04751098": "heterogeneity, heterogeneousness",
  "n04751305": "diverseness, diversity, multifariousness, variety",
  "n04751652": "biodiversity",
  "n04751847": "inconsistency",
  "n04752034": "variety, change",
  "n04752221": "inequality",
  "n04752414": "nonequivalence",
  "n04752530": "disparity",
  "n04752734": "far cry",
  "n04752859": "gap, spread",
  "n04753060": "gulf, disconnect, disconnection",
  "n04753331": "unevenness",
  "n04753455": "certainty, sure thing, foregone conclusion",
  "n04753799": "cert",
  "n04753929": "ineluctability, unavoidability",
  "n04754056": "inevitability, inevitableness",
  "n04754237": "determinateness, definiteness",
  "n04754440": "finality, conclusiveness, decisiveness",
  "n04754780": "surety",
  "n04754862": "indisputability, indubitability, unquestionability, unquestionableness",
  "n04755218": "incontrovertibility, incontrovertibleness, positivity, positiveness",
  "n04755466": "demonstrability, provability",
  "n04755630": "givenness",
  "n04755783": "moral certainty",
  "n04756025": "predictability",
  "n04756172": "probability",
  "n04756504": "odds",
  "n04756635": "likelihood, likeliness",
  "n04756887": "uncertainty, uncertainness, precariousness",
  "n04757350": "slam dunk",
  "n04757522": "doubt, dubiousness, doubtfulness, question",
  "n04757864": "indefiniteness, indeterminateness, indefinity, indetermination, indeterminacy",
  "n04758181": "inconclusiveness",
  "n04758313": "unpredictability",
  "n04758452": "improbability, improbableness",
  "n04758776": "unlikelihood, unlikeliness",
  "n04758980": "fortuitousness",
  "n04759134": "speculativeness",
  "n04759428": "factuality, factualness",
  "n04759712": "counterfactuality",
  "n04759849": "concreteness",
  "n04760024": "tangibility, tangibleness, palpability",
  "n04760296": "intangibility, intangibleness, impalpability",
  "n04760611": "literalness",
  "n04760771": "materiality, physicalness, corporeality, corporality",
  "n04761212": "substantiality, substantialness, solidness",
  "n04761517": "immateriality, incorporeality",
  "n04761815": "insubstantiality",
  "n04761960": "smoke",
  "n04762134": "abstractness",
  "n04762355": "reality",
  "n04762492": "unreality",
  "n04762633": "particularity, specialness",
  "n04762915": "specificity",
  "n04763147": "specificity",
  "n04763293": "individuality, individualism, individuation",
  "n04763650": "singularity, uniqueness",
  "n04763925": "peculiarity, specialness, specialty, speciality, distinctiveness",
  "n04764242": "idiosyncrasy, foible, mannerism",
  "n04764412": "generality",
  "n04764741": "commonality, commonness",
  "n04765038": "solidarity",
  "n04765166": "pervasiveness",
  "n04765355": "prevalence",
  "n04765586": "currency",
  "n04765691": "universality, catholicity",
  "n04765851": "totality",
  "n04766059": "simplicity, simpleness",
  "n04766275": "complexity, complexness",
  "n04766620": "complicatedness, complication, knottiness, tortuousness",
  "n04766852": "elaborateness, elaboration, intricacy, involution",
  "n04767023": "tapestry",
  "n04767182": "trickiness",
  "n04767347": "regularity",
  "n04767805": "cyclicity, periodicity",
  "n04768028": "rhythm, regular recurrence",
  "n04768186": "cardiac rhythm, heart rhythm",
  "n04768309": "atrioventricular nodal rhythm, nodal rhythm",
  "n04768483": "orderliness, methodicalness",
  "n04768657": "organization, organisation, system",
  "n04769049": "uniformity",
  "n04769234": "homogeneity",
  "n04769362": "inhomogeneity",
  "n04769456": "evenness, invariability",
  "n04769716": "smoothness",
  "n04769988": "even spacing",
  "n04770068": "steadiness",
  "n04770211": "irregularity, unregularity",
  "n04770535": "fitfulness, jerkiness",
  "n04770720": "intermittence, intermittency",
  "n04770911": "fluctuation, wavering",
  "n04771128": "scintillation",
  "n04771332": "randomness, haphazardness, stochasticity, noise",
  "n04771535": "ergodicity",
  "n04771738": "spasticity",
  "n04771890": "unevenness, variability",
  "n04772191": "rockiness",
  "n04772388": "ruggedness",
  "n04772557": "hilliness",
  "n04772691": "jaggedness",
  "n04772844": "patchiness",
  "n04772955": "waviness",
  "n04773068": "personal equation",
  "n04773182": "unsteadiness",
  "n04773351": "mobility",
  "n04773596": "locomotion, motive power, motivity",
  "n04773761": "motility",
  "n04773899": "movability, movableness",
  "n04774198": "maneuverability, manoeuvrability",
  "n04774365": "manipulability",
  "n04774511": "looseness, play",
  "n04774743": "restlessness",
  "n04774901": "weatherliness",
  "n04775185": "wiggliness",
  "n04775357": "slack, slackness",
  "n04775562": "unsteadiness, ricketiness",
  "n04775747": "instability, unstableness",
  "n04776021": "shakiness",
  "n04776176": "portability",
  "n04776299": "immobility",
  "n04776453": "immotility",
  "n04776572": "inertness",
  "n04776699": "immovability, immovableness",
  "n04776940": "tightness, tautness",
  "n04777098": "fastness, fixedness, fixity, fixture, secureness",
  "n04777421": "looseness",
  "n04777634": "lodgment, lodgement, lodging",
  "n04777852": "steadiness, firmness",
  "n04778114": "granite",
  "n04778267": "sureness",
  "n04778401": "stability, stableness",
  "n04778630": "pleasantness, sweetness",
  "n04779053": "agreeableness, amenity",
  "n04779336": "enjoyableness",
  "n04779521": "niceness",
  "n04779649": "unpleasantness",
  "n04779895": "disagreeableness",
  "n04780114": "abrasiveness",
  "n04780232": "acridity, acridness",
  "n04780396": "unpalatability, unpalatableness",
  "n04780605": "disgustingness, unsavoriness",
  "n04780755": "nastiness",
  "n04780958": "offensiveness, odiousness, distastefulness",
  "n04781349": "loathsomeness, repulsiveness, sliminess, vileness, lousiness, wickedness",
  "n04781755": "hatefulness, obnoxiousness, objectionableness",
  "n04781967": "beastliness",
  "n04782116": "awfulness, dreadfulness, horridness, terribleness",
  "n04782466": "frightfulness",
  "n04782610": "ghastliness, grimness, gruesomeness, luridness",
  "n04782878": "credibility, credibleness, believability",
  "n04783247": "authenticity, genuineness, legitimacy",
  "n04783445": "real McCoy, real thing, real stuff",
  "n04783567": "cogency, validity, rigor, rigour",
  "n04783724": "plausibility, plausibleness",
  "n04783888": "reasonableness, tenability, tenableness",
  "n04784142": "incredibility, incredibleness",
  "n04784322": "implausibility, implausibleness",
  "n04784525": "street credibility, street cred, cred",
  "n04784664": "logicality, logicalness",
  "n04784978": "rationality, rationalness",
  "n04785195": "consistency",
  "n04785414": "completeness",
  "n04785669": "illogicality, illogicalness, illogic, inconsequence",
  "n04785908": "naturalness",
  "n04786326": "unaffectedness",
  "n04786490": "simplicity, simmpleness",
  "n04786595": "sincerity, unassumingness",
  "n04786763": "spontaneity, spontaneousness",
  "n04786994": "ease, informality",
  "n04787154": "unpretentiousness",
  "n04787324": "naturalization, naturalisation",
  "n04787530": "unnaturalness",
  "n04787763": "affectedness",
  "n04788011": "airs, pose",
  "n04788159": "coyness, demureness",
  "n04788332": "preciosity, preciousness",
  "n04788494": "artificiality",
  "n04788693": "staginess, theatricality",
  "n04788853": "pretension, pretense, pretence",
  "n04788982": "pretentiousness, pretension, largeness",
  "n04789274": "ostentation",
  "n04789406": "supernaturalism, supernaturalness",
  "n04789613": "virtu, vertu",
  "n04789689": "wholesomeness",
  "n04789897": "nutritiousness, nutritiveness",
  "n04790070": "healthfulness",
  "n04790238": "salubrity, salubriousness",
  "n04790449": "unwholesomeness, morbidness, morbidity",
  "n04790774": "harmfulness, noisomeness, noxiousness",
  "n04790942": "perniciousness, toxicity",
  "n04791081": "deadliness, lethality",
  "n04791321": "fatality",
  "n04791453": "jejunity, jejuneness",
  "n04791591": "putrescence, rottenness",
  "n04791740": "unhealthfulness",
  "n04791928": "insalubrity, insalubriousness",
  "n04792127": "satisfactoriness",
  "n04792357": "adequacy, adequateness",
  "n04792679": "acceptability, acceptableness",
  "n04793016": "admissibility",
  "n04793209": "permissibility",
  "n04793355": "unsatisfactoriness",
  "n04793555": "inadequacy, inadequateness",
  "n04793731": "perishability, perishableness",
  "n04793925": "unacceptability, unacceptableness",
  "n04794236": "inadmissibility",
  "n04794408": "impermissibility",
  "n04794580": "palatability, palatableness",
  "n04794751": "ordinariness, mundaneness, mundanity",
  "n04795061": "averageness, mediocrity",
  "n04795252": "expectedness",
  "n04795413": "normality, normalcy",
  "n04795545": "commonness, commonplaceness, everydayness",
  "n04795878": "prosiness, prosaicness",
  "n04796086": "usualness",
  "n04796291": "familiarity",
  "n04796490": "extraordinariness",
  "n04796725": "unexpectedness, surprisingness",
  "n04796946": "uncommonness",
  "n04797130": "uncommonness",
  "n04797295": "unusualness",
  "n04797482": "unfamiliarity, strangeness",
  "n04797824": "oddity, queerness, quirk, quirkiness, crotchet",
  "n04798020": "eeriness, ghostliness",
  "n04798185": "abnormality, freakishness",
  "n04798365": "singularity",
  "n04798511": "outlandishness, bizarreness, weirdness",
  "n04798700": "quaintness",
  "n04798881": "eccentricity",
  "n04799011": "oddity, oddness",
  "n04799133": "ethnicity",
  "n04799344": "foreignness, strangeness, curiousness",
  "n04799612": "exoticism, exoticness, exotism",
  "n04799789": "alienage, alienism",
  "n04799881": "nativeness",
  "n04800152": "indigenousness, autochthony, endemism",
  "n04800359": "originality",
  "n04800596": "freshness, novelty",
  "n04800729": "unorthodoxy, heterodoxy",
  "n04800885": "unconventionality",
  "n04801030": "nonconformity",
  "n04801168": "unoriginality",
  "n04801313": "orthodoxy",
  "n04801532": "conventionality, convention, conventionalism",
  "n04801763": "ossification, conformity",
  "n04801877": "traditionalism, traditionality",
  "n04802079": "scholasticism, academicism, academism",
  "n04802198": "correctness, rightness",
  "n04802403": "incorrectness, wrongness",
  "n04802629": "erroneousness, error",
  "n04802776": "deviation",
  "n04802907": "accuracy, truth",
  "n04803209": "accuracy",
  "n04803430": "exactness, exactitude",
  "n04803702": "minuteness",
  "n04803880": "preciseness, precision",
  "n04804169": "trueness",
  "n04804306": "fidelity",
  "n04804451": "inaccuracy",
  "n04804611": "inexactness, inexactitude",
  "n04804787": "impreciseness, imprecision",
  "n04804947": "looseness",
  "n04805136": "infallibility",
  "n04805304": "inerrancy",
  "n04805472": "errancy",
  "n04805635": "papal infallibility",
  "n04805813": "errancy",
  "n04806067": "instability",
  "n04806169": "reproducibility, duplicability",
  "n04806316": "irreproducibility",
  "n04806512": "fallibility",
  "n04806655": "distinction",
  "n04806804": "worthiness",
  "n04807050": "deservingness, merit, meritoriousness",
  "n04807342": "praiseworthiness, laudability, laudableness",
  "n04807489": "quotability",
  "n04807626": "roadworthiness",
  "n04807776": "unworthiness",
  "n04807971": "baseness, sordidness, contemptibility, despicableness, despicability",
  "n04808281": "shamefulness, disgracefulness, ignominiousness",
  "n04808516": "scandalousness",
  "n04808639": "popularity",
  "n04808965": "hot stuff",
  "n04809089": "unpopularity",
  "n04809237": "legality",
  "n04809453": "validity, validness",
  "n04809642": "effect, force",
  "n04809784": "lawfulness",
  "n04810035": "legitimacy",
  "n04810194": "licitness",
  "n04810327": "illegality",
  "n04810510": "invalidity, invalidness",
  "n04810727": "fallaciousness",
  "n04810865": "unlawfulness",
  "n04811126": "lawlessness, outlawry",
  "n04811296": "infection",
  "n04811476": "illegitimacy",
  "n04811628": "illicitness",
  "n04811767": "shadiness",
  "n04811995": "refinement, civilization, civilisation",
  "n04812268": "elegance",
  "n04812636": "elegance",
  "n04812871": "dash, elan, flair, panache, style",
  "n04813066": "daintiness, delicacy, fineness",
  "n04813283": "courtliness",
  "n04813395": "tastefulness",
  "n04813540": "breeding, genteelness, gentility",
  "n04813712": "chic, chicness, chichi, modishness, smartness, stylishness, swank, last word",
  "n04814025": "jauntiness, nattiness, dapperness, rakishness",
  "n04814238": "magnificence, brilliance, splendor, splendour, grandeur, grandness",
  "n04814872": "eclat",
  "n04815002": "pomp, eclat",
  "n04815177": "class",
  "n04815321": "inelegance",
  "n04815624": "awkwardness, clumsiness, gracelessness, stiffness",
  "n04815916": "woodenness",
  "n04816125": "rusticity, gaucherie",
  "n04816302": "urbanity",
  "n04816528": "dowdiness, drabness, homeliness",
  "n04816761": "shabbiness, seediness, manginess, sleaziness",
  "n04817020": "tweediness",
  "n04817168": "raggedness",
  "n04817280": "coarseness, commonness, grossness, vulgarity, vulgarism, raunch",
  "n04817564": "crudeness, roughness",
  "n04817788": "boorishness, uncouthness",
  "n04817923": "ostentation, ostentatiousness, pomposity, pompousness, pretentiousness, puffiness, splashiness, inflation",
  "n04818284": "tastelessness",
  "n04818460": "cheapness, tackiness, tat, sleaze",
  "n04818700": "flashiness, garishness, gaudiness, loudness, brashness, meretriciousness, tawdriness, glitz",
  "n04819026": "comprehensibility, understandability",
  "n04819285": "legibility, readability",
  "n04819496": "intelligibility",
  "n04819694": "expressiveness",
  "n04819822": "picturesqueness",
  "n04819953": "readability",
  "n04820102": "speech intelligibility",
  "n04820258": "clarity, lucidity, lucidness, pellucidity, clearness, limpidity",
  "n04820908": "monosemy",
  "n04821084": "focus",
  "n04821277": "coherence, coherency",
  "n04821451": "preciseness, clearcutness",
  "n04821615": "perspicuity, perspicuousness, plainness",
  "n04821802": "unambiguity, unequivocalness",
  "n04822032": "explicitness",
  "n04822223": "incomprehensibility",
  "n04822524": "inscrutability",
  "n04822685": "illegibility",
  "n04822870": "impenetrability, impenetrableness",
  "n04823031": "noise",
  "n04823258": "opacity, opaqueness",
  "n04823416": "obscureness, obscurity, abstruseness, reconditeness",
  "n04823706": "unintelligibility",
  "n04823866": "unclearness",
  "n04824118": "elusiveness",
  "n04824350": "vagueness",
  "n04824695": "haziness",
  "n04824819": "inexplicitness",
  "n04824977": "implicitness",
  "n04825114": "ambiguity, equivocalness",
  "n04825383": "equivocation, prevarication, evasiveness",
  "n04825576": "polysemy, lexical ambiguity",
  "n04825815": "twilight zone, no man's land",
  "n04826235": "righteousness",
  "n04826485": "impeccability",
  "n04826623": "uprightness, rectitude",
  "n04826771": "piety, piousness",
  "n04826999": "devoutness, religiousness",
  "n04827175": "religiosity, religionism, religiousism, pietism",
  "n04827392": "dutifulness",
  "n04827503": "godliness",
  "n04827652": "unrighteousness",
  "n04827957": "sin, sinfulness, wickedness",
  "n04828126": "mark of Cain",
  "n04828255": "impiety, impiousness",
  "n04828485": "undutifulness",
  "n04828612": "irreligiousness, irreligion",
  "n04828754": "ungodliness, godlessness",
  "n04828925": "humaneness",
  "n04829182": "humanity",
  "n04829282": "mercifulness, mercy",
  "n04829550": "compassion, pity",
  "n04829764": "forgivingness, kindness",
  "n04829922": "lenience, leniency, mildness, lenity",
  "n04830102": "inhumaneness, inhumanity",
  "n04830343": "atrocity, atrociousness, barbarity, barbarousness, heinousness",
  "n04830578": "bestiality",
  "n04830689": "ferociousness, brutality, viciousness, savagery",
  "n04830904": "murderousness",
  "n04831031": "mercilessness, unmercifulness",
  "n04831264": "pitilessness, ruthlessness",
  "n04831437": "relentlessness, inexorability, inexorableness",
  "n04831727": "generosity, generousness",
  "n04832050": "charitableness",
  "n04832244": "bounty, bounteousness",
  "n04832401": "bigheartedness",
  "n04832518": "liberality, liberalness",
  "n04832716": "munificence, largess, largesse, magnanimity, openhandedness",
  "n04832951": "unselfishness",
  "n04833276": "altruism, selflessness",
  "n04833458": "stinginess",
  "n04833687": "meanness, minginess, niggardliness, niggardness, parsimony, parsimoniousness, tightness, tightfistedness, closeness",
  "n04834073": "pettiness, littleness, smallness",
  "n04834228": "miserliness",
  "n04834339": "penuriousness",
  "n04834457": "illiberality",
  "n04834605": "selfishness",
  "n04834817": "greediness, voraciousness, rapaciousness",
  "n04835028": "egoism, egocentrism, self-interest, self-concern, self-centeredness",
  "n04835260": "self-love, narcism, narcissism",
  "n04835488": "opportunism, self-interest, self-seeking, expedience",
  "n04835724": "drive",
  "n04835950": "action",
  "n04836074": "enterprise, enterprisingness, initiative, go-ahead",
  "n04836268": "ambition, ambitiousness",
  "n04836491": "aspiration",
  "n04836583": "power hunger, status seeking",
  "n04836683": "energy, push, get-up-and-go",
  "n04836960": "second wind",
  "n04837232": "aggressiveness",
  "n04837425": "competitiveness, fight",
  "n04837615": "combativeness, militance, militancy",
  "n04837814": "scrappiness",
  "n04837931": "intrusiveness, meddlesomeness, officiousness",
  "n04838210": "boldness, nerve, brass, face, cheek",
  "n04838510": "audacity, audaciousness",
  "n04838727": "presumption, presumptuousness, effrontery, assumption",
  "n04839009": "uppityness, uppishness",
  "n04839154": "fairness, equity",
  "n04839445": "non-discrimination",
  "n04839555": "sportsmanship",
  "n04839676": "unfairness, inequity",
  "n04839877": "gamesmanship",
  "n04840011": "kindness",
  "n04840285": "benevolence",
  "n04840405": "charity, brotherly love",
  "n04840537": "beneficence",
  "n04840715": "grace, grace of God, free grace",
  "n04840981": "benignity, benignancy, graciousness",
  "n04841245": "loving-kindness",
  "n04841358": "consideration, considerateness, thoughtfulness",
  "n04841664": "kindliness, helpfulness",
  "n04841810": "tact, tactfulness",
  "n04842029": "delicacy, diplomacy, discreetness, finesse",
  "n04842232": "savoir-faire, address",
  "n04842313": "malevolence, malevolency, malice",
  "n04842515": "cattiness, bitchiness, spite, spitefulness, nastiness",
  "n04842788": "malignity, malignancy, malignance",
  "n04842993": "sensitivity, sensitiveness",
  "n04843270": "antenna, feeler",
  "n04843429": "defensiveness",
  "n04843675": "bunker mentality",
  "n04843875": "perceptiveness",
  "n04844024": "insensitivity, insensitiveness",
  "n04844343": "crassness, crassitude",
  "n04844478": "tin ear",
  "n04844625": "unfeelingness, callousness, callosity, hardness, insensibility",
  "n04844891": "dullness",
  "n04845146": "unperceptiveness",
  "n04845312": "unkindness",
  "n04845475": "cruelty, cruelness, harshness",
  "n04845684": "beastliness, meanness",
  "n04845863": "unhelpfulness",
  "n04845967": "inconsideration, inconsiderateness, thoughtlessness",
  "n04846243": "tactlessness",
  "n04846383": "bluntness",
  "n04846533": "maleficence, mischief, balefulness",
  "n04846770": "morality",
  "n04847133": "rightness",
  "n04847298": "virtue, virtuousness, moral excellence",
  "n04847482": "virtue",
  "n04847600": "cardinal virtue",
  "n04847733": "natural virtue",
  "n04847991": "theological virtue, supernatural virtue",
  "n04848262": "hope",
  "n04848383": "saintliness",
  "n04848492": "conscience",
  "n04848686": "conscientiousness",
  "n04848866": "religiousness",
  "n04849074": "unconscientiousness",
  "n04849241": "good, goodness",
  "n04849614": "summum bonum",
  "n04849759": "virtue, chastity, sexual morality",
  "n04849972": "honor, honour, purity, pureness",
  "n04850117": "justice, justness",
  "n04850341": "right, rightfulness",
  "n04850589": "immorality",
  "n04850996": "corruption, degeneracy, depravation, depravity, putrefaction",
  "n04851421": "infection",
  "n04851585": "corruptibility",
  "n04851715": "licentiousness, wantonness",
  "n04851865": "anomie, anomy",
  "n04851962": "wrongness",
  "n04852088": "evil, evilness",
  "n04852534": "worst",
  "n04852750": "nefariousness, wickedness, vileness, ugliness",
  "n04852962": "filthiness",
  "n04853124": "enormity",
  "n04853212": "reprehensibility",
  "n04853361": "villainy, villainousness",
  "n04853503": "perversity, perverseness",
  "n04853765": "error, wrongdoing",
  "n04853873": "frailty, vice",
  "n04853948": "corruptness, corruption",
  "n04854259": "venality",
  "n04854389": "injustice, unjustness",
  "n04854604": "wrong, wrongfulness",
  "n04854893": "amorality",
  "n04854976": "divinity",
  "n04855138": "holiness, sanctity, sanctitude",
  "n04855332": "sacredness",
  "n04855524": "ideality",
  "n04855641": "holy of holies",
  "n04855840": "unholiness",
  "n04856014": "profaneness, unsanctification",
  "n04856182": "sacrilegiousness",
  "n04856308": "safeness",
  "n04856460": "dangerousness",
  "n04856621": "precariousness",
  "n04856721": "curability, curableness",
  "n04856899": "incurability, incurableness",
  "n04857083": "courage, courageousness, bravery, braveness",
  "n04857490": "heart, mettle, nerve, spunk",
  "n04857738": "heroism, gallantry, valor, valour, valorousness, valiance, valiancy",
  "n04858089": "dauntlessness, intrepidity",
  "n04858222": "Dutch courage",
  "n04858317": "stoutheartedness",
  "n04858455": "fearlessness",
  "n04858632": "coolness, nervelessness",
  "n04858785": "boldness, daring, hardiness, hardihood",
  "n04859177": "adventurousness, venturesomeness",
  "n04859323": "daredevilry, daredeviltry",
  "n04859449": "audacity, audaciousness, temerity",
  "n04859636": "shamelessness, brazenness",
  "n04859816": "gutsiness, pluck, pluckiness",
  "n04860065": "cowardice, cowardliness",
  "n04860272": "cravenness",
  "n04860369": "faintheartedness, faintness",
  "n04860586": "fearfulness",
  "n04860759": "timidity, timorousness",
  "n04860985": "pusillanimity, pusillanimousness",
  "n04861143": "poltroonery",
  "n04861221": "dastardliness",
  "n04861319": "gutlessness",
  "n04861486": "resoluteness, firmness, firmness of purpose, resolve, resolution",
  "n04862005": "self-control, self-possession, possession, willpower, will power, self-command, self-will",
  "n04862236": "nerves",
  "n04862382": "steadiness",
  "n04862592": "sturdiness",
  "n04862747": "presence of mind",
  "n04862888": "stiffness",
  "n04863074": "stubbornness, bullheadedness, obstinacy, obstinance, pigheadedness, self-will",
  "n04863358": "impenitence, impenitency",
  "n04863497": "intransigency, intransigence",
  "n04863675": "single-mindedness",
  "n04863793": "adamance, obduracy, unyieldingness",
  "n04863969": "decisiveness, decision",
  "n04864200": "determination, purpose",
  "n04864515": "doggedness, perseverance, persistence, persistency, tenacity, tenaciousness, pertinacity",
  "n04864824": "indefatigability, indefatigableness, tirelessness",
  "n04865016": "steadfastness",
  "n04865114": "diligence, industriousness, industry",
  "n04865502": "assiduity, assiduousness, concentration",
  "n04865722": "intentness, engrossment",
  "n04865921": "singleness",
  "n04866078": "sedulity, sedulousness",
  "n04866238": "studiousness",
  "n04866365": "bookishness",
  "n04866465": "irresoluteness, irresolution",
  "n04866671": "volatility, unpredictability",
  "n04866866": "indecisiveness, indecision",
  "n04867130": "sincerity",
  "n04867440": "sooth",
  "n04867539": "heartiness, wholeheartedness",
  "n04867700": "singleness, straightforwardness",
  "n04867871": "insincerity, falseness, hollowness",
  "n04868148": "hypocrisy",
  "n04868350": "sanctimoniousness, sanctimony",
  "n04868505": "fulsomeness, oiliness, oleaginousness, smarminess, unctuousness, unction",
  "n04868748": "honorableness, honourableness",
  "n04869106": "honor, honour",
  "n04869256": "scrupulousness",
  "n04869405": "venerability, venerableness",
  "n04869569": "integrity",
  "n04869811": "probity",
  "n04870040": "incorruptness, incorruption",
  "n04870206": "incorruptibility",
  "n04870340": "nobility, nobleness, magnanimousness, grandeur",
  "n04870643": "high-mindedness, idealism, noble-mindedness",
  "n04870889": "sublimity",
  "n04871002": "respectability, reputability",
  "n04871259": "decency",
  "n04871374": "honesty, honestness",
  "n04871720": "candor, candour, candidness, frankness, directness, forthrightness",
  "n04872016": "good faith, straightness",
  "n04872236": "truthfulness",
  "n04872414": "veracity",
  "n04872531": "ingenuousness",
  "n04872676": "artlessness",
  "n04872826": "parental quality",
  "n04872958": "motherliness, maternalism, maternal quality, maternity",
  "n04873305": "fatherliness, paternal quality",
  "n04873550": "dishonorableness, dishonourableness",
  "n04873804": "ignobleness, ignobility",
  "n04873939": "dishonor, dishonour",
  "n04874086": "unscrupulousness",
  "n04874223": "sleaziness",
  "n04874409": "unrespectability, disreputability, disreputableness",
  "n04874672": "dishonesty",
  "n04874939": "deceptiveness, obliquity",
  "n04875113": "speciousness, meretriciousness",
  "n04875352": "fraudulence, deceit",
  "n04875468": "jobbery",
  "n04875556": "crookedness, deviousness",
  "n04875728": "rascality, shiftiness, slipperiness, trickiness",
  "n04875935": "thievishness, larcenous",
  "n04876053": "untruthfulness",
  "n04876235": "mendacity",
  "n04876374": "disingenuousness",
  "n04876561": "craftiness, deceitfulness, guile",
  "n04876705": "artfulness",
  "n04876888": "cunning",
  "n04876985": "fidelity, faithfulness",
  "n04877264": "constancy",
  "n04877421": "dedication",
  "n04877530": "loyalty, trueness",
  "n04877783": "steadfastness, staunchness",
  "n04877938": "allegiance, fealty",
  "n04878101": "patriotism, nationalism",
  "n04878434": "regionalism",
  "n04878539": "Americanism",
  "n04878646": "chauvinism, jingoism, superpatriotism, ultranationalism",
  "n04878861": "infidelity, unfaithfulness",
  "n04879092": "faithlessness, falseness, fickleness, inconstancy",
  "n04879340": "disloyalty",
  "n04879498": "disaffection",
  "n04879658": "treason, subversiveness, traitorousness",
  "n04879879": "betrayal",
  "n04879964": "perfidy, perfidiousness, treachery",
  "n04880138": "insidiousness",
  "n04880273": "sophistication, worldliness, mundaneness, mundanity",
  "n04880573": "naivete, naivety, naiveness",
  "n04880830": "artlessness, innocence, ingenuousness, naturalness",
  "n04881033": "innocency",
  "n04881156": "credulousness, gullibility",
  "n04881369": "simplicity, simpleness, simple mindedness",
  "n04881623": "discipline",
  "n04881829": "self-discipline, self-denial",
  "n04881998": "austerity, asceticism, nonindulgence",
  "n04882214": "monasticism",
  "n04882438": "eremitism",
  "n04882622": "abstinence, abstention",
  "n04882813": "continence, continency",
  "n04882968": "restraint, control",
  "n04883243": "self-restraint, temperateness",
  "n04883436": "stiff upper lip",
  "n04883614": "temperance, moderation",
  "n04883804": "sobriety, dryness",
  "n04883942": "abstemiousness",
  "n04884072": "inhibition",
  "n04884177": "continence",
  "n04884302": "taboo, tabu",
  "n04884450": "indiscipline, undiscipline",
  "n04884627": "indulgence, self-indulgence",
  "n04884817": "dissoluteness, incontinence, self-gratification",
  "n04884995": "rakishness",
  "n04885091": "unrestraint",
  "n04885271": "intemperance",
  "n04885398": "abandon, wantonness, unconstraint",
  "n04885609": "looseness",
  "n04885784": "madness, rabidity, rabidness",
  "n04885990": "sottishness",
  "n04886101": "gluttony",
  "n04886235": "greediness, hoggishness, piggishness",
  "n04886402": "edacity, esurience, rapaciousness, rapacity, voracity, voraciousness",
  "n04886646": "pride",
  "n04886790": "civic pride, civic spirit",
  "n04886881": "dignity, self-respect, self-regard, self-worth",
  "n04887129": "conceit, conceitedness, vanity",
  "n04887373": "boastfulness, vainglory",
  "n04887497": "egotism, self-importance, swelled head",
  "n04887701": "posturing",
  "n04887790": "superiority complex",
  "n04887912": "arrogance, haughtiness, hauteur, high-handedness, lordliness",
  "n04888268": "condescension, superciliousness, disdainfulness",
  "n04888510": "contemptuousness",
  "n04888703": "hubris",
  "n04888788": "imperiousness, domineeringness, overbearingness",
  "n04889000": "superiority",
  "n04889162": "snobbery, snobbism, snobbishness",
  "n04889337": "clannishness, cliquishness, exclusiveness",
  "n04889527": "humility, humbleness",
  "n04889779": "meekness, subduedness",
  "n04889966": "spinelessness",
  "n04890112": "wisdom, wiseness",
  "n04890361": "judiciousness, sagacity, sagaciousness",
  "n04890546": "knowledgeability, knowledgeableness, initiation",
  "n04890865": "statesmanship, statecraft, diplomacy",
  "n04891010": "discretion, discernment",
  "n04891184": "circumspection, caution",
  "n04891333": "folly, foolishness, unwiseness",
  "n04891558": "indiscretion, injudiciousness",
  "n04891683": "absurdity, fatuity, fatuousness, silliness",
  "n04891944": "asininity",
  "n04892084": "judgment, judgement, sound judgment, sound judgement, perspicacity",
  "n04892344": "objectivity, objectiveness",
  "n04892544": "subjectivity, subjectiveness",
  "n04892794": "prudence",
  "n04892970": "providence",
  "n04893172": "foresight, foresightedness, foresightfulness",
  "n04893358": "frugality, frugalness",
  "n04893525": "parsimony, parsimoniousness, thrift, penny-pinching",
  "n04893787": "economy, thriftiness",
  "n04894037": "imprudence",
  "n04894204": "heedlessness, mindlessness, rashness",
  "n04894444": "lightheadedness",
  "n04894552": "improvidence, shortsightedness",
  "n04894807": "extravagance, prodigality, profligacy",
  "n04894964": "thriftlessness, waste, wastefulness",
  "n04895246": "trust, trustingness, trustfulness",
  "n04895558": "credulity",
  "n04895695": "overcredulity",
  "n04895773": "distrust, distrustfulness, mistrust",
  "n04895979": "suspicion, suspiciousness",
  "n04896161": "cleanliness",
  "n04896317": "fastidiousness",
  "n04896515": "tidiness, neatness",
  "n04896724": "uncleanliness",
  "n04896878": "slovenliness",
  "n04896995": "slatternliness, sluttishness",
  "n04897152": "squeamishness",
  "n04897428": "untidiness, messiness",
  "n04897604": "disarray, disorderliness",
  "n04897762": "demeanor, demeanour, behavior, behaviour, conduct, deportment",
  "n04898087": "manners",
  "n04898208": "citizenship",
  "n04898334": "swashbuckling",
  "n04898437": "propriety, properness, correctitude",
  "n04898804": "decorum, decorousness",
  "n04899031": "appropriateness, rightness",
  "n04899201": "correctness",
  "n04899416": "good form",
  "n04899573": "faultlessness, impeccability",
  "n04899713": "political correctness, political correctitude",
  "n04899980": "priggishness, primness",
  "n04900121": "modesty, reserve",
  "n04900236": "demureness",
  "n04900357": "seemliness, grace",
  "n04900597": "becomingness",
  "n04900739": "decency",
  "n04900947": "modesty, modestness",
  "n04901152": "primness, prudishness, prudery, Grundyism",
  "n04901326": "impropriety, improperness",
  "n04901599": "incorrectness",
  "n04901754": "political incorrectness",
  "n04901996": "inappropriateness, wrongness",
  "n04902165": "indelicacy",
  "n04902283": "gaminess, raciness, ribaldry, spiciness",
  "n04902470": "indecorum, indecorousness",
  "n04902668": "unseemliness",
  "n04902797": "unbecomingness",
  "n04902925": "indecency",
  "n04903136": "immodesty",
  "n04903247": "outrageousness, enormity",
  "n04903368": "obscenity, lewdness, bawdiness, salaciousness, salacity",
  "n04903678": "smuttiness, dirtiness",
  "n04903813": "composure, calm, calmness, equanimity",
  "n04904162": "aplomb, assuredness, cool, poise, sang-froid",
  "n04904352": "repose, quiet, placidity, serenity, tranquillity, tranquility",
  "n04904560": "ataraxia",
  "n04904664": "discomposure",
  "n04904851": "disquiet, unease, uneasiness",
  "n04904996": "perturbation, fluster",
  "n04905188": "tractability, tractableness, flexibility",
  "n04905546": "manageability, manageableness",
  "n04905697": "docility",
  "n04905842": "tameness, domestication",
  "n04906026": "amenability, amenableness, cooperativeness",
  "n04906273": "obedience",
  "n04906471": "submissiveness",
  "n04906712": "obsequiousness, servility, subservience",
  "n04906923": "sycophancy",
  "n04907020": "passivity, passiveness",
  "n04907153": "subordination",
  "n04907269": "intractability, intractableness",
  "n04907575": "refractoriness, unmanageableness, recalcitrance, recalcitrancy",
  "n04907826": "wildness",
  "n04907991": "defiance, rebelliousness",
  "n04908182": "insubordination",
  "n04908283": "obstreperousness",
  "n04908396": "unruliness, fractiousness, willfulness, wilfulness",
  "n04908721": "balkiness",
  "n04908835": "stubbornness, obstinacy, obstinance, mulishness",
  "n04909018": "contrariness, perversity, perverseness",
  "n04909270": "cussedness, orneriness",
  "n04909414": "disobedience",
  "n04909563": "naughtiness, mischievousness, badness",
  "n04909721": "prankishness, rascality, roguishness",
  "n04909887": "wildness",
  "n04910135": "manner, personal manner",
  "n04910377": "bearing, comportment, presence, mien",
  "n04910562": "bedside manner",
  "n04910684": "dignity, lordliness, gravitas",
  "n04910848": "foppishness, dandyism",
  "n04910973": "gentleness, softness, mildness",
  "n04911420": "formality, formalness",
  "n04911777": "ceremoniousness",
  "n04911894": "stateliness",
  "n04912052": "informality",
  "n04912240": "casualness, familiarity",
  "n04912506": "slanginess",
  "n04912610": "unceremoniousness",
  "n04912732": "courtesy, good manners",
  "n04912982": "politeness, niceness",
  "n04913212": "urbanity",
  "n04913322": "suavity, suaveness, blandness, smoothness",
  "n04913568": "graciousness",
  "n04913738": "chivalry, gallantry, politesse",
  "n04913839": "deference, respect, respectfulness",
  "n04914133": "civility",
  "n04914292": "discourtesy, rudeness",
  "n04914576": "boorishness",
  "n04914694": "impoliteness",
  "n04914881": "bad manners, ill-breeding",
  "n04914991": "ungraciousness",
  "n04915121": "crudeness, crudity, gaucheness",
  "n04915365": "incivility",
  "n04915462": "abruptness, brusqueness, curtness, gruffness, shortness",
  "n04915687": "contempt, disrespect",
  "n04915866": "crust, gall, impertinence, impudence, insolence, cheekiness, freshness",
  "n04916200": "chutzpa, chutzpah, hutzpah",
  "n04916342": "property",
  "n04917292": "actinism",
  "n04917439": "isotropy, symmetry",
  "n04917680": "anisotropy",
  "n04917870": "characteristic, device characteristic",
  "n04918053": "connectivity",
  "n04918210": "directness, straightness",
  "n04918498": "downrightness, straightforwardness",
  "n04918767": "immediacy, immediateness",
  "n04919011": "pointedness",
  "n04919209": "indirectness",
  "n04919458": "allusiveness",
  "n04919580": "mediacy, mediateness",
  "n04919712": "deviousness, obliqueness",
  "n04919872": "discursiveness",
  "n04920000": "robustness",
  "n04920237": "rurality, ruralism",
  "n04920435": "streak",
  "n04920568": "duality, wave-particle duality",
  "n04920867": "heredity, genetic endowment",
  "n04921011": "inheritance, heritage",
  "n04921308": "birthright",
  "n04921417": "background",
  "n04921576": "birthright",
  "n04921754": "upbringing",
  "n04921900": "education, training, breeding",
  "n04922113": "raising, rearing, nurture",
  "n04922338": "inheritance, hereditary pattern",
  "n04922597": "heterosis, hybrid vigor",
  "n04922787": "ancestry, lineage, derivation, filiation",
  "n04923024": "linkage, gene linkage",
  "n04923299": "X-linked dominant inheritance",
  "n04923496": "X-linked recessive inheritance",
  "n04923743": "origin, descent, extraction",
  "n04923907": "pedigree, bloodline",
  "n04924003": "full blood",
  "n04924103": "age",
  "n04924491": "chronological age",
  "n04924674": "bone age",
  "n04924878": "developmental age",
  "n04925064": "fetal age, fertilization age, gestational age",
  "n04925218": "mental age",
  "n04925348": "oldness",
  "n04925577": "obsoleteness, superannuation",
  "n04925756": "ancientness, antiquity",
  "n04925895": "old-fashionedness",
  "n04926038": "quaintness",
  "n04926217": "vintage, time of origin",
  "n04926308": "hoariness",
  "n04926427": "newness",
  "n04926620": "brand-newness",
  "n04926728": "freshness",
  "n04927098": "crispness",
  "n04927268": "recency, recentness",
  "n04927445": "oldness",
  "n04927632": "agedness, senescence",
  "n04927792": "senility",
  "n04927890": "longevity, seniority",
  "n04928008": "staleness",
  "n04928194": "mustiness, must, moldiness",
  "n04928416": "youngness",
  "n04928585": "youth, youthfulness, juvenility",
  "n04928760": "childishness, puerility",
  "n04928903": "manner, mode, style, way, fashion",
  "n04929422": "artistic style, idiom",
  "n04929742": "High Renaissance",
  "n04929976": "treatment",
  "n04930139": "drape",
  "n04930307": "fit",
  "n04930478": "form",
  "n04930632": "life style, life-style, lifestyle, modus vivendi",
  "n04930850": "fast lane",
  "n04931041": "free living",
  "n04931152": "vanity fair",
  "n04931267": "setup",
  "n04931428": "touch, signature",
  "n04931567": "common touch",
  "n04931733": "wise",
  "n04931837": "hang",
  "n04931965": "structure",
  "n04932278": "computer architecture, architecture",
  "n04932561": "complex instruction set computing, complex instruction set computer, CISC",
  "n04932875": "reduced instruction set computing, reduced instruction set computer, RISC",
  "n04933197": "cytoarchitecture, cytoarchitectonics",
  "n04933363": "framework, fabric",
  "n04933544": "constitution, composition, physical composition, makeup, make-up",
  "n04933852": "phenotype",
  "n04934043": "genotype, genetic constitution",
  "n04934220": "texture, grain",
  "n04934546": "consistency, consistence, eubstance, body",
  "n04935003": "viscosity, viscousness",
  "n04935239": "stickiness",
  "n04935405": "sliminess",
  "n04935528": "adhesiveness, adhesion, adherence, bond",
  "n04935904": "cohesiveness, glueyness, gluiness, gumminess, tackiness, ropiness, viscidity, viscidness",
  "n04936213": "gelatinousness, glutinosity, glutinousness",
  "n04936403": "thickness",
  "n04936585": "semifluidity",
  "n04936690": "creaminess, soupiness",
  "n04936846": "thinness",
  "n04937043": "fluidity, fluidness, liquidity, liquidness, runniness",
  "n04937427": "wateriness",
  "n04937587": "hardness",
  "n04937848": "hardness",
  "n04938110": "firmness",
  "n04938228": "softness",
  "n04938474": "compressibility, squeezability, sponginess",
  "n04938702": "incompressibility",
  "n04938838": "downiness, featheriness, fluffiness",
  "n04939046": "flabbiness, limpness, flaccidity",
  "n04939198": "mushiness, pulpiness",
  "n04939324": "breakableness",
  "n04939547": "brittleness, crispness, crispiness",
  "n04939742": "crumbliness, friability",
  "n04939872": "flakiness",
  "n04939987": "unbreakableness",
  "n04940146": "porosity, porousness",
  "n04940379": "sponginess",
  "n04940496": "permeability, permeableness",
  "n04940730": "penetrability, perviousness",
  "n04940964": "absorbency",
  "n04941124": "solidity, solidness",
  "n04941325": "compactness",
  "n04941453": "density, denseness",
  "n04941687": "specific gravity",
  "n04941809": "vapor density, vapour density",
  "n04941942": "impermeability, impermeableness",
  "n04942172": "retentiveness, retentivity, retention",
  "n04942373": "urinary retention",
  "n04942516": "impenetrability, imperviousness",
  "n04942738": "nonabsorbency",
  "n04942869": "disposition",
  "n04943154": "aptness, propensity",
  "n04943367": "mordacity",
  "n04943464": "predisposition",
  "n04943578": "proneness",
  "n04943704": "separatism",
  "n04944048": "tendency, inclination",
  "n04944412": "buoyancy",
  "n04944513": "electronegativity, negativity",
  "n04944710": "stainability",
  "n04944910": "basophilia",
  "n04945057": "desire",
  "n04945254": "hunger, hungriness, thirst, thirstiness",
  "n04945530": "greed",
  "n04945758": "avarice, avariciousness, covetousness, cupidity",
  "n04945942": "possessiveness",
  "n04946078": "acquisitiveness",
  "n04946266": "bibliomania",
  "n04946400": "retentiveness, retentivity",
  "n04946553": "tactile property, feel",
  "n04946760": "touch",
  "n04946877": "texture",
  "n04947087": "nap",
  "n04947186": "smoothness",
  "n04947494": "silkiness, sleekness",
  "n04947628": "slickness, slick, slipperiness, slip",
  "n04947888": "soapiness",
  "n04948069": "fineness, powderiness",
  "n04948241": "roughness, raggedness",
  "n04948603": "scaliness",
  "n04948722": "coarseness, nubbiness, tweediness",
  "n04948905": "slub, knot, burl",
  "n04949066": "harshness, abrasiveness, scratchiness",
  "n04949256": "coarseness, graininess, granularity",
  "n04949453": "sandiness",
  "n04949560": "shagginess",
  "n04949681": "bumpiness",
  "n04949799": "prickliness, bristliness, spininess, thorniness",
  "n04950026": "optics",
  "n04950126": "visual property",
  "n04950336": "sleekness",
  "n04950537": "texture",
  "n04950713": "grain",
  "n04950952": "wood grain, woodgrain, woodiness",
  "n04951071": "graining, woodgraining",
  "n04951186": "marbleization, marbleisation, marbleizing, marbleising",
  "n04951373": "light, lightness",
  "n04951716": "aura, aureole, halo, nimbus, glory, gloriole",
  "n04951875": "sunniness",
  "n04951978": "cloudlessness",
  "n04952120": "highlight, highlighting",
  "n04952242": "brightness",
  "n04952570": "glare, blaze, brilliance",
  "n04952821": "dazzle",
  "n04952944": "glitter, glister, glisten, scintillation, sparkle",
  "n04953186": "flash",
  "n04953296": "glint",
  "n04953380": "sparkle, twinkle, spark, light",
  "n04953678": "opalescence, iridescence",
  "n04953954": "radiance, radiancy, shine, effulgence, refulgence, refulgency",
  "n04954412": "radio brightness",
  "n04954534": "gleam, gleaming, glow, lambency",
  "n04954683": "shininess, sheen, luster, lustre",
  "n04954920": "luster, lustre, brilliancy, splendor, splendour",
  "n04955160": "polish, gloss, glossiness, burnish",
  "n04955426": "French polish",
  "n04955554": "glaze",
  "n04955633": "dullness",
  "n04955907": "dimness, subduedness",
  "n04956110": "flatness, lusterlessness, lustrelessness, mat, matt, matte",
  "n04956419": "softness",
  "n04956594": "color, colour, coloring, colouring",
  "n04957176": "primary color, primary colour",
  "n04957356": "primary color for pigments, primary colour for pigments",
  "n04957589": "primary color for light, primary colour for light",
  "n04957854": "primary subtractive color for light, primary subtractive colour for light",
  "n04958146": "heather mixture, heather",
  "n04958302": "mellowness",
  "n04958460": "richness",
  "n04958634": "colorlessness, colourlessness, achromatism, achromaticity",
  "n04958865": "mottle",
  "n04959061": "achromia",
  "n04959230": "shade, tint, tincture, tone",
  "n04959567": "undertone, tinge",
  "n04959672": "chromatic color, chromatic colour, spectral color, spectral colour",
  "n04960079": "achromatic color, achromatic colour",
  "n04960277": "black, blackness, inkiness",
  "n04960582": "coal black, ebony, jet black, pitch black, sable, soot black",
  "n04960729": "white, whiteness",
  "n04961062": "alabaster",
  "n04961136": "bleach",
  "n04961331": "bone, ivory, pearl, off-white",
  "n04961472": "chalk",
  "n04961583": "frostiness, hoariness",
  "n04961691": "gray, grayness, grey, greyness",
  "n04962062": "ash grey, ash gray, silver, silver grey, silver gray",
  "n04962240": "charcoal, charcoal grey, charcoal gray, oxford grey, oxford gray",
  "n04962395": "dapple-grey, dapple-gray, dappled-grey, dappled-gray",
  "n04962548": "iron-grey, iron-gray",
  "n04962689": "tattletale grey, tattletale gray",
  "n04962784": "red, redness",
  "n04963111": "sanguine",
  "n04963201": "chrome red",
  "n04963307": "Turkey red, alizarine red",
  "n04963449": "cardinal, carmine",
  "n04963588": "crimson, ruby, deep red",
  "n04963740": "dark red",
  "n04963905": "burgundy",
  "n04964001": "claret",
  "n04964079": "oxblood red",
  "n04964162": "wine, wine-colored, wine-coloured",
  "n04964287": "purplish red, purplish-red",
  "n04964443": "cerise, cherry, cherry red",
  "n04964586": "magenta",
  "n04964799": "fuschia",
  "n04964878": "maroon",
  "n04964977": "scarlet, vermilion, orange red",
  "n04965179": "orange, orangeness",
  "n04965371": "salmon",
  "n04965451": "reddish orange",
  "n04965574": "tangerine",
  "n04965661": "yellow, yellowness",
  "n04966017": "blond, blonde",
  "n04966114": "canary yellow, canary",
  "n04966240": "amber, gold",
  "n04966428": "brownish yellow",
  "n04966543": "gamboge, lemon, lemon yellow, maize",
  "n04966649": "old gold",
  "n04966717": "orange yellow, saffron",
  "n04966841": "ocher, ochre",
  "n04966941": "pale yellow, straw, wheat",
  "n04967094": "greenish yellow",
  "n04967191": "green, greenness, viridity",
  "n04967561": "greenishness",
  "n04967674": "sea green",
  "n04967801": "sage green",
  "n04967882": "bottle green",
  "n04967974": "chrome green",
  "n04968056": "emerald",
  "n04968139": "olive green, olive-green",
  "n04968257": "yellow green, yellowish green, chartreuse, Paris green, pea green",
  "n04968426": "bluish green, blue green, teal",
  "n04968619": "cyan",
  "n04968749": "jade green, jade",
  "n04968895": "blue, blueness",
  "n04969242": "azure, cerulean, sapphire, lazuline, sky-blue",
  "n04969431": "powder blue",
  "n04969540": "steel blue",
  "n04969617": "Prussian blue",
  "n04969703": "dark blue, navy, navy blue",
  "n04969798": "greenish blue, aqua, aquamarine, turquoise, cobalt blue, peacock blue",
  "n04969952": "purplish blue, royal blue",
  "n04970059": "purple, purpleness",
  "n04970312": "Tyrian purple",
  "n04970398": "indigo",
  "n04970470": "lavender",
  "n04970544": "mauve",
  "n04970631": "reddish purple, royal purple",
  "n04970758": "violet, reddish blue",
  "n04970916": "pink",
  "n04971131": "pinkness",
  "n04971211": "carnation",
  "n04971313": "rose, rosiness",
  "n04971447": "old rose",
  "n04971522": "solferino, purplish pink",
  "n04971675": "yellowish pink, apricot, peach, salmon pink",
  "n04971820": "coral",
  "n04971928": "brown, brownness",
  "n04972268": "Vandyke brown",
  "n04972350": "chestnut",
  "n04972451": "chocolate, coffee, deep brown, umber, burnt umber",
  "n04972603": "hazel",
  "n04972801": "light brown",
  "n04972950": "mocha",
  "n04973020": "tan, topaz",
  "n04973110": "dun, greyish brown, grayish brown, fawn",
  "n04973291": "beige, ecru",
  "n04973386": "reddish brown, sepia, burnt sienna, Venetian red, mahogany",
  "n04973585": "brick red",
  "n04973669": "copper, copper color",
  "n04973816": "Indian red",
  "n04973957": "yellowish brown, raw sienna, buff, caramel, caramel brown",
  "n04974145": "puce",
  "n04974248": "olive brown",
  "n04974340": "olive",
  "n04974463": "olive drab, drab",
  "n04974575": "pastel",
  "n04974681": "snuff-color, snuff-colour",
  "n04974774": "taupe",
  "n04974859": "ultramarine",
  "n04974968": "color property",
  "n04975122": "hue, chromaticity",
  "n04975340": "saturation, chroma, intensity, vividness",
  "n04975612": "paleness, pallidity",
  "n04975739": "complementary color, complementary",
  "n04975988": "coloration, colouration",
  "n04976188": "hair coloring",
  "n04976319": "pigmentation",
  "n04976489": "chromatism",
  "n04976567": "melanoderma",
  "n04976687": "depigmentation",
  "n04976870": "poliosis",
  "n04976952": "complexion, skin color, skin colour",
  "n04977247": "paleness, blondness, fairness",
  "n04977412": "ruddiness, rosiness",
  "n04977561": "lividness, lividity, luridness, paleness, pallidness, pallor, wanness, achromasia",
  "n04977946": "sallowness",
  "n04978050": "tawniness",
  "n04978216": "darkness, duskiness, swarthiness",
  "n04978371": "whiteness",
  "n04978561": "nonsolid color, nonsolid colour, dithered color, dithered colour",
  "n04978792": "protective coloration",
  "n04979002": "aposematic coloration, warning coloration",
  "n04979203": "apatetic coloration",
  "n04979307": "cryptic coloration",
  "n04979425": "value",
  "n04979758": "lightness",
  "n04979870": "darkness",
  "n04980008": "olfactory property, smell, aroma, odor, odour, scent",
  "n04980463": "bouquet, fragrance, fragrancy, redolence, sweetness",
  "n04980656": "malodorousness, stinkiness, foulness, rankness, fetidness",
  "n04980920": "body odor, body odour, B.O.",
  "n04981044": "muskiness",
  "n04981139": "sound",
  "n04981474": "noisiness, racketiness",
  "n04981658": "ring",
  "n04981781": "unison",
  "n04981941": "voice",
  "n04982113": "androglossia",
  "n04982207": "silence, quiet",
  "n04982478": "hush, stillness, still",
  "n04982745": "speechlessness",
  "n04982856": "quietness, soundlessness",
  "n04983013": "noiselessness",
  "n04983122": "sound property",
  "n04983402": "musicality, musicalness",
  "n04983688": "lyricality, lyricism, songfulness",
  "n04983848": "melodiousness, tunefulness",
  "n04984007": "texture",
  "n04984180": "harmony",
  "n04984351": "consonance, harmoniousness",
  "n04984514": "dissonance",
  "n04984698": "discordance, discord",
  "n04984809": "disharmony, inharmoniousness",
  "n04984938": "cacophony",
  "n04985064": "boisterousness",
  "n04985198": "pitch",
  "n04985580": "concert pitch, philharmonic pitch, international pitch",
  "n04985790": "high pitch, high frequency",
  "n04985930": "soprano, treble",
  "n04986054": "tenor",
  "n04986147": "key",
  "n04986258": "low pitch, low frequency",
  "n04986414": "deepness",
  "n04986526": "alto",
  "n04986637": "alto",
  "n04986796": "bass",
  "n04986883": "tone",
  "n04987169": "tune",
  "n04987356": "registration",
  "n04987620": "timbre, timber, quality, tone",
  "n04988078": "harmonic",
  "n04988258": "resonance",
  "n04988478": "color, colour, coloration, colouration",
  "n04988666": "harshness, roughness",
  "n04988861": "gruffness, hoarseness, huskiness",
  "n04989015": "fullness, mellowness, richness",
  "n04989362": "nasality",
  "n04989512": "twang, nasal twang",
  "n04989657": "plangency, resonance, reverberance, ringing, sonorousness, sonority, vibrancy",
  "n04990021": "shrillness, stridence, stridency",
  "n04990220": "volume, loudness, intensity",
  "n04990525": "crescendo",
  "n04990692": "swell",
  "n04990781": "forte, fortissimo",
  "n04990877": "softness",
  "n04991137": "faintness",
  "n04991225": "decrescendo, diminuendo",
  "n04991389": "piano, pianissimo",
  "n04991511": "rhythmicity",
  "n04991738": "meter, metre, time",
  "n04991879": "cadence, cadency",
  "n04992008": "lilt, swing",
  "n04992163": "taste property",
  "n04992431": "rancidness",
  "n04992570": "spiciness, spice, spicery",
  "n04992834": "pungency, bite, sharpness, raciness",
  "n04993108": "nip, piquance, piquancy, piquantness, tang, tanginess, zest",
  "n04993312": "hotness, pepperiness",
  "n04993413": "saltiness",
  "n04993604": "brininess, salinity",
  "n04993752": "brackishness",
  "n04993882": "sourness, sour, acidity",
  "n04994126": "acerbity, tartness",
  "n04994264": "vinegariness, vinegarishness",
  "n04994413": "sweetness, sweet",
  "n04994614": "saccharinity",
  "n04994727": "sugariness",
  "n04994824": "bitterness, bitter",
  "n04995009": "acerbity",
  "n04995100": "acridity, acridness",
  "n04995211": "palatability, palatableness",
  "n04995421": "pleasingness, tastiness",
  "n04995531": "appetizingness, appetisingness",
  "n04995793": "delectability, deliciousness, lusciousness, toothsomeness",
  "n04995940": "flavorsomeness, flavoursomeness, savoriness",
  "n04996113": "sapidity, sapidness",
  "n04996215": "succulence, succulency, juiciness",
  "n04996355": "unpalatability, unpalatableness",
  "n04996571": "disgustingness, distastefulness, nauseatingness, sickeningness, unsavoriness",
  "n04996823": "unappetizingness, unappetisingness",
  "n04997032": "flavorlessness, flavourlessness, savorlessness, savourlessness, tastelessness",
  "n04997282": "blandness, insipidity, insipidness",
  "n04997472": "edibility, edibleness",
  "n04997645": "digestibility, digestibleness",
  "n04997812": "indigestibility, indigestibleness",
  "n04997988": "bodily property",
  "n04998254": "bipedalism",
  "n04998417": "laterality, dominance",
  "n04998530": "physique, build, body-build, habitus",
  "n04998700": "lankiness",
  "n04998816": "dumpiness, squattiness",
  "n04998966": "body type, somatotype",
  "n04999111": "asthenic type, ectomorphy",
  "n04999214": "endomorphy, pyknic type",
  "n04999306": "athletic type, mesomorphy",
  "n04999401": "fatness, fat, blubber, avoirdupois",
  "n04999741": "adiposity, adiposeness, fattiness",
  "n04999964": "abdominousness, paunchiness",
  "n05000116": "greasiness, oiliness, oleaginousness",
  "n05000342": "fleshiness, obesity, corpulency",
  "n05000537": "corpulence, overweight, stoutness, adiposis",
  "n05000717": "exogenous obesity",
  "n05000809": "steatopygia",
  "n05000913": "plumpness, embonpoint, roundness",
  "n05001089": "chubbiness, pudginess, tubbiness, rolypoliness",
  "n05001302": "buxomness",
  "n05001482": "leanness, thinness, spareness",
  "n05001724": "skinniness, scrawniness",
  "n05001867": "bonyness, boniness, emaciation, gauntness, maceration",
  "n05002155": "slenderness, slightness, slimness",
  "n05002352": "stature, height",
  "n05002540": "tallness",
  "n05002680": "shortness",
  "n05002822": "carriage, bearing, posture",
  "n05003090": "walk, manner of walking",
  "n05003273": "slouch",
  "n05003423": "gracefulness",
  "n05003590": "grace, gracility",
  "n05003850": "agility, legerity, lightness, lightsomeness, nimbleness",
  "n05004091": "lissomeness, litheness, suppleness",
  "n05004294": "awkwardness, clumsiness",
  "n05004532": "gracelessness, ungracefulness",
  "n05004700": "gawkiness, ungainliness",
  "n05004895": "stiffness",
  "n05005064": "physiology",
  "n05005250": "physiological property",
  "n05005447": "animateness, aliveness, liveness",
  "n05005809": "animation, vitality",
  "n05006020": "sentience",
  "n05006285": "inanimateness, lifelessness",
  "n05006519": "deadness",
  "n05006749": "insentience",
  "n05006898": "sex, gender, sexuality",
  "n05007280": "sex characteristic, sexual characteristic, sex character",
  "n05007560": "primary sex characteristic, primary sexual characteristic, primary sex character",
  "n05007800": "secondary sex characteristic, secondary sexual characteristic, secondary sex character",
  "n05008085": "asexuality, sexlessness",
  "n05008227": "maleness, masculinity",
  "n05008449": "virility",
  "n05008591": "virilism",
  "n05008746": "androgyny, hermaphroditism, bisexuality",
  "n05008943": "femaleness, feminineness",
  "n05009170": "physical property",
  "n05009758": "chemical property",
  "n05009921": "volatility",
  "n05010062": "absorptivity, absorption factor",
  "n05010314": "dissolubility, solubleness",
  "n05010506": "drippiness",
  "n05010627": "reflection, reflexion, reflectivity",
  "n05010801": "echo, reverberation, sound reflection, replication",
  "n05011089": "re-echo",
  "n05011162": "echo",
  "n05011277": "deflection, deflexion, bending",
  "n05011431": "windage, wind deflection",
  "n05011568": "refractivity, refractiveness",
  "n05011790": "temperature",
  "n05012272": "heat content, total heat, enthalpy, H",
  "n05012585": "randomness, entropy, S",
  "n05012941": "conformational entropy",
  "n05013095": "absolute temperature",
  "n05013204": "absolute zero",
  "n05013461": "Curie temperature, Curie point",
  "n05013642": "dew point",
  "n05013809": "flash point, flashpoint",
  "n05013967": "freezing point, melting point",
  "n05014099": "boiling point, boil",
  "n05014308": "mercury",
  "n05014442": "room temperature",
  "n05014556": "simmer",
  "n05014710": "basal body temperature, basal temperature",
  "n05014879": "body temperature, blood heat",
  "n05015117": "coldness, cold, low temperature, frigidity, frigidness",
  "n05015463": "chill, iciness, gelidity",
  "n05015678": "chilliness, coolness, nip",
  "n05015878": "frostiness",
  "n05016001": "cool",
  "n05016171": "hotness, heat, high temperature",
  "n05016451": "calefaction, incalescence",
  "n05016553": "fieriness, red heat",
  "n05016667": "torridity",
  "n05016753": "warmth, warmness",
  "n05016936": "lukewarmness, tepidity, tepidness",
  "n05017121": "white heat",
  "n05017230": "perceptibility",
  "n05017458": "visibility, visibleness",
  "n05017757": "visual range",
  "n05017909": "invisibility, invisibleness",
  "n05018103": "luminosity, brightness, brightness level, luminance, luminousness, light",
  "n05018542": "illuminance, illumination",
  "n05018674": "incandescence",
  "n05018785": "luminescence, glow",
  "n05018934": "audibility, audibleness",
  "n05019163": "inaudibility, inaudibleness",
  "n05019339": "imperceptibility",
  "n05019499": "reluctivity",
  "n05019661": "sensitivity, sensitiveness",
  "n05020019": "frequency response",
  "n05020225": "magnetization, magnetisation",
  "n05020358": "elasticity, snap",
  "n05020697": "resilience, resiliency",
  "n05020981": "bounce, bounciness",
  "n05021151": "give, spring, springiness",
  "n05021345": "stretch, stretchiness, stretchability",
  "n05021535": "temper, toughness",
  "n05021740": "elasticity of shear",
  "n05021884": "malleability, plasticity",
  "n05022173": "ductility, ductileness",
  "n05022359": "fibrosity, fibrousness",
  "n05022457": "flexibility, flexibleness",
  "n05022709": "bendability, pliability",
  "n05022902": "whip",
  "n05023022": "pliancy, pliantness, suppleness",
  "n05023233": "inelasticity",
  "n05023404": "deadness",
  "n05023591": "stiffness",
  "n05023741": "rigidity, rigidness",
  "n05023974": "unmalleability",
  "n05024093": "inflexibility, inflexibleness",
  "n05024254": "mass",
  "n05024616": "body",
  "n05024691": "biomass",
  "n05024797": "critical mass",
  "n05024931": "rest mass",
  "n05025129": "relativistic mass",
  "n05025413": "bulk",
  "n05025520": "gravitational mass",
  "n05025693": "inertial mass",
  "n05025935": "atomic mass, atomic weight, relative atomic mass",
  "n05026171": "mass energy",
  "n05026312": "molecular weight, relative molecular mass",
  "n05026508": "equivalent, equivalent weight, combining weight, eq",
  "n05026744": "milliequivalent, meq",
  "n05026843": "weight",
  "n05027135": "body weight",
  "n05027276": "reporting weight",
  "n05027446": "dead weight",
  "n05027529": "heaviness, weightiness",
  "n05027837": "heft, heftiness, massiveness, ponderousness, ponderosity",
  "n05028159": "preponderance",
  "n05028348": "poundage",
  "n05028429": "tare",
  "n05028562": "throw-weight",
  "n05028700": "lightness, weightlessness",
  "n05028963": "airiness, buoyancy",
  "n05029137": "momentum",
  "n05029327": "angular momentum",
  "n05029594": "sustainability",
  "n05029706": "strength",
  "n05030149": "good part",
  "n05030251": "tensile strength",
  "n05030418": "brawn, brawniness, muscle, muscularity, sinew, heftiness",
  "n05030680": "might, mightiness, power",
  "n05030806": "vigor, vigour, dynamism, heartiness",
  "n05031012": "robustness, hardiness, lustiness, validity",
  "n05031214": "huskiness, ruggedness, toughness",
  "n05031367": "smallness, littleness",
  "n05031560": "stoutness, stalwartness",
  "n05031726": "sturdiness",
  "n05031849": "firmness, soundness",
  "n05032028": "indomitability, invincibility",
  "n05032193": "fortitude",
  "n05032351": "backbone, grit, guts, moxie, sand, gumption",
  "n05032565": "endurance",
  "n05032795": "sufferance",
  "n05032918": "stamina, staying power, toughness",
  "n05033046": "legs",
  "n05033171": "wiriness",
  "n05033291": "long-sufferance, long-suffering",
  "n05033410": "tolerance",
  "n05033569": "capacity",
  "n05033681": "invulnerability",
  "n05033906": "immunity",
  "n05034048": "power of appointment",
  "n05034225": "potency, effectiveness, strength",
  "n05034473": "valence, valency",
  "n05034761": "covalence, covalency",
  "n05034989": "valence, valency",
  "n05035264": "sea power",
  "n05035353": "force, forcefulness, strength",
  "n05035712": "brunt",
  "n05035820": "momentum, impulse",
  "n05035961": "energy, vigor, vigour, zip",
  "n05036237": "athleticism, strenuosity",
  "n05036394": "intensity, intensiveness",
  "n05036715": "badness, severity, severeness",
  "n05037037": "foulness, raininess",
  "n05037197": "seriousness, distressfulness",
  "n05037394": "vehemence, emphasis",
  "n05037617": "top",
  "n05037737": "overemphasis",
  "n05037813": "ferocity, fierceness, furiousness, fury, vehemence, violence, wildness",
  "n05038251": "savageness, savagery",
  "n05038593": "concentration",
  "n05038823": "titer, titre",
  "n05038959": "hydrogen ion concentration",
  "n05039106": "pH, pH scale",
  "n05039526": "acidity",
  "n05039633": "hyperacidity",
  "n05039709": "alkalinity",
  "n05039819": "neutrality",
  "n05039907": "molality, molal concentration",
  "n05040081": "molarity, molar concentration, M",
  "n05040275": "weakness",
  "n05040833": "adynamia",
  "n05040939": "feebleness, tenuity",
  "n05041165": "faintness",
  "n05041320": "flimsiness, shoddiness",
  "n05041503": "fragility, delicacy",
  "n05041595": "insubstantiality",
  "n05041712": "attenuation",
  "n05041871": "enervation",
  "n05042012": "fatigability",
  "n05042138": "inanition, lassitude, lethargy, slackness",
  "n05042283": "weak part, weak spot, soft spot",
  "n05042468": "Achilles' heel",
  "n05042573": "jugular",
  "n05042720": "underbelly",
  "n05042871": "vulnerability",
  "n05043091": "defenselessness, defencelessness, unprotectedness",
  "n05043349": "assailability",
  "n05043459": "destructibility",
  "n05043621": "indestructibility",
  "n05043755": "fragility, breakability, frangibleness, frangibility",
  "n05043973": "exposure",
  "n05044215": "windage, wind exposure",
  "n05044387": "solarization, solarisation",
  "n05044528": "temporal property",
  "n05044673": "temporal arrangement, temporal order",
  "n05044822": "sequence, chronological sequence, succession, successiveness, chronological succession",
  "n05045208": "rain, pelting",
  "n05045381": "rotation",
  "n05045680": "row",
  "n05045841": "run",
  "n05046009": "timing",
  "n05046471": "approach, approaching, coming",
  "n05046659": "earliness",
  "n05046865": "forwardness",
  "n05047059": "lateness",
  "n05047279": "priority, antecedence, antecedency, anteriority, precedence, precedency",
  "n05047617": "posteriority, subsequentness, subsequence",
  "n05047778": "punctuality, promptness",
  "n05047956": "tardiness",
  "n05048123": "simultaneity, simultaneousness",
  "n05048301": "concurrence, coincidence, conjunction, co-occurrence",
  "n05048690": "concomitance",
  "n05048833": "overlap",
  "n05048948": "contemporaneity, contemporaneousness",
  "n05049137": "unison",
  "n05049253": "seasonableness, timeliness",
  "n05049458": "unseasonableness, untimeliness",
  "n05049656": "pastness",
  "n05049808": "recency, recentness",
  "n05049930": "futurity",
  "n05050115": "presentness, nowness",
  "n05050379": "currentness, currency, up-to-dateness",
  "n05050668": "modernity, modernness, modernism, contemporaneity, contemporaneousness",
  "n05051088": "spark advance, lead",
  "n05051249": "duration, length",
  "n05051601": "longness",
  "n05051734": "longevity, length of service",
  "n05051896": "lengthiness, prolongation, continuation, protraction",
  "n05052243": "fermata",
  "n05052387": "endlessness",
  "n05052587": "continuousness, ceaselessness, incessancy, incessantness",
  "n05052832": "shortness",
  "n05053042": "brevity, briefness, transience",
  "n05053215": "permanence, permanency",
  "n05053527": "perpetuity, sempiternity",
  "n05053688": "lastingness, durability, enduringness, strength",
  "n05054130": "continuity, persistence",
  "n05054312": "changelessness",
  "n05054426": "everlastingness",
  "n05054537": "imperishability, imperishableness, imperishingness",
  "n05054746": "perdurability",
  "n05054863": "impermanence, impermanency",
  "n05055141": "temporariness",
  "n05055278": "transience, transiency, transitoriness",
  "n05055503": "fugacity, fugaciousness",
  "n05055689": "ephemerality, ephemeralness, fleetingness",
  "n05055878": "fugacity",
  "n05055974": "mortality",
  "n05056102": "immortality",
  "n05056234": "viability",
  "n05056386": "audio, audio frequency",
  "n05056490": "radio frequency",
  "n05056811": "infrared, infrared frequency",
  "n05057061": "station",
  "n05057163": "extremely low frequency, ELF",
  "n05057275": "very low frequency, VLF",
  "n05057382": "low frequency, LF",
  "n05057485": "medium frequency, MF",
  "n05057593": "high frequency, HF",
  "n05057695": "very high frequency, VHF",
  "n05057805": "ultrahigh frequency, UHF",
  "n05057917": "superhigh frequency, SHF",
  "n05058025": "extremely high frequency, EHF",
  "n05058140": "speed, swiftness, fastness",
  "n05058580": "pace, rate",
  "n05058893": "beat",
  "n05059017": "fleetness",
  "n05059132": "celerity, quickness, rapidity, rapidness, speediness",
  "n05059525": "immediacy, immediateness, instantaneousness, instancy",
  "n05059830": "dispatch, despatch, expedition, expeditiousness",
  "n05060052": "promptness, promptitude",
  "n05060189": "haste, hastiness, hurry, hurriedness, precipitation",
  "n05060476": "abruptness, precipitateness, precipitousness, precipitance, precipitancy, suddenness",
  "n05060783": "acceleration",
  "n05061003": "pickup, getaway",
  "n05061149": "precipitation",
  "n05061345": "deceleration, slowing, retardation",
  "n05061619": "execution speed",
  "n05061805": "graduality, gradualness",
  "n05061977": "slowness, deliberation, deliberateness, unhurriedness",
  "n05062255": "leisureliness",
  "n05062370": "dilatoriness, procrastination",
  "n05062518": "sluggishness",
  "n05062748": "spatial property, spatiality",
  "n05062993": "dimensionality",
  "n05063218": "one-dimensionality, linearity",
  "n05063349": "two-dimensionality, flatness, planeness",
  "n05063564": "three-dimensionality, third-dimensionality",
  "n05063729": "cubicity",
  "n05063853": "directionality",
  "n05064037": "shape, form, configuration, contour, conformation",
  "n05064541": "topography",
  "n05064722": "lobularity",
  "n05064827": "symmetry, symmetricalness, correspondence, balance",
  "n05065211": "regularity, geometrical regularity",
  "n05065386": "bilaterality, bilateralism, bilateral symmetry",
  "n05065555": "radial symmetry",
  "n05065717": "asymmetry, dissymmetry, imbalance",
  "n05066012": "irregularity, geometrical irregularity",
  "n05066195": "lopsidedness, skewness",
  "n05066333": "obliqueness",
  "n05066490": "radial asymmetry",
  "n05066626": "directivity, directionality",
  "n05066867": "directivity, directiveness",
  "n05067007": "handedness, laterality",
  "n05067199": "ambidexterity, ambidextrousness",
  "n05067367": "left-handedness, sinistrality",
  "n05067514": "right-handedness, dextrality",
  "n05067679": "footedness",
  "n05067807": "eyedness",
  "n05067928": "occlusion",
  "n05068080": "tilt, list, inclination, lean, leaning",
  "n05068461": "gradient, slope",
  "n05068716": "grade",
  "n05068918": "upgrade, rise, rising slope",
  "n05069085": "downgrade",
  "n05069199": "pitch, rake, slant",
  "n05069447": "loft",
  "n05069624": "abruptness, precipitousness, steepness",
  "n05069853": "gradualness, gentleness",
  "n05070032": "concavity, concaveness",
  "n05070197": "hollowness",
  "n05070290": "convexity, convexness",
  "n05070453": "roundedness, bulginess",
  "n05070622": "oblateness, ellipticity",
  "n05070849": "angularity",
  "n05071027": "narrowing",
  "n05071185": "coarctation",
  "n05071368": "taper",
  "n05071556": "point, pointedness",
  "n05071726": "unpointedness",
  "n05071869": "rectangularity, oblongness",
  "n05072078": "orthogonality",
  "n05072208": "perpendicularity",
  "n05072406": "squareness",
  "n05072523": "triangularity",
  "n05072663": "curvature, curve",
  "n05072911": "roundness",
  "n05073131": "sphericity, sphericalness, globosity, globularness, rotundity, rotundness",
  "n05073403": "cylindricality, cylindricalness",
  "n05073559": "circularity, disk shape",
  "n05073723": "concentricity",
  "n05073888": "eccentricity",
  "n05074057": "straightness",
  "n05074218": "crookedness",
  "n05074374": "curliness, waviness",
  "n05074559": "straightness",
  "n05074690": "stratification",
  "n05074774": "position, spatial relation",
  "n05075602": "placement, arrangement",
  "n05076069": "columniation",
  "n05076237": "point of view",
  "n05076384": "camera angle",
  "n05076472": "composition, composing",
  "n05076709": "fenestration",
  "n05076827": "proportion, proportionality, balance",
  "n05077146": "alignment",
  "n05077348": "true",
  "n05077524": "misalignment",
  "n05077661": "coincidence",
  "n05077863": "dead center, dead centre",
  "n05078025": "centrality",
  "n05078147": "marginality",
  "n05078289": "anteriority",
  "n05078450": "posteriority",
  "n05078623": "outwardness, externality",
  "n05078894": "inwardness",
  "n05079074": "malposition, misplacement",
  "n05079180": "northernness",
  "n05079309": "southernness",
  "n05079474": "horizontality",
  "n05079638": "verticality, verticalness, erectness, uprightness",
  "n05079866": "position, posture, attitude",
  "n05080382": "ballet position",
  "n05080526": "decubitus",
  "n05080616": "eversion",
  "n05080753": "lithotomy position",
  "n05080923": "lotus position",
  "n05081057": "missionary position",
  "n05081300": "pose",
  "n05081434": "presentation",
  "n05081660": "ectopia",
  "n05081802": "arabesque",
  "n05081957": "asana",
  "n05082116": "matsyendra",
  "n05082222": "guard",
  "n05082337": "sprawl, sprawling",
  "n05082507": "stance",
  "n05082648": "address",
  "n05082790": "attention",
  "n05083054": "erectness, uprightness",
  "n05083200": "ramification",
  "n05083328": "spacing, spatial arrangement",
  "n05083567": "tandem",
  "n05083687": "tuck",
  "n05083905": "openness",
  "n05084067": "patency",
  "n05084201": "distance",
  "n05084733": "way",
  "n05084889": "piece",
  "n05084982": "mean distance",
  "n05085165": "farness, remoteness, farawayness",
  "n05085436": "far cry",
  "n05085572": "nearness, closeness",
  "n05085867": "proximity, propinquity",
  "n05085991": "adjacency, contiguity, contiguousness",
  "n05086269": "wavelength",
  "n05086450": "focal distance, focal length",
  "n05086563": "hyperfocal distance",
  "n05086740": "leap",
  "n05086903": "elevation",
  "n05087061": "span",
  "n05087173": "wheelbase",
  "n05087297": "distribution, dispersion",
  "n05087664": "complementary distribution, complementation",
  "n05087894": "diaspora",
  "n05088056": "dissemination, diffusion",
  "n05088189": "innervation",
  "n05088324": "scatter, spread",
  "n05088645": "diffuseness",
  "n05088804": "concentration, density, denseness, tightness, compactness",
  "n05089048": "bits per inch, bpi",
  "n05089199": "flux density, flux",
  "n05089367": "optical density, transmission density, photographic density, absorbance",
  "n05089612": "rarity, tenuity, low density",
  "n05089782": "relative density",
  "n05089947": "interval, separation",
  "n05090111": "clearance",
  "n05090255": "remove",
  "n05090441": "magnitude",
  "n05090979": "absolute magnitude",
  "n05091194": "proportion, dimension",
  "n05091316": "order, order of magnitude",
  "n05091527": "information, selective information, entropy",
  "n05091770": "probability, chance",
  "n05092236": "conditional probability, contingent probability",
  "n05092421": "cross section",
  "n05092635": "exceedance",
  "n05092969": "fair chance, sporting chance",
  "n05093080": "fat chance, slim chance",
  "n05093181": "joint probability",
  "n05093293": "risk, risk of exposure",
  "n05093418": "risk, risk of infection",
  "n05093581": "dimension",
  "n05093890": "degree, grade, level",
  "n05094424": "grind",
  "n05094565": "degree",
  "n05094725": "depth",
  "n05094863": "profundity, profoundness",
  "n05095111": "superficiality, shallowness",
  "n05095324": "glibness, slickness",
  "n05095561": "sciolism",
  "n05095691": "size",
  "n05096095": "extra large",
  "n05096191": "large",
  "n05096294": "number",
  "n05096408": "octavo, eightvo, 8vo",
  "n05096577": "outsize",
  "n05096708": "petite",
  "n05096820": "quarto, 4to",
  "n05096972": "regular",
  "n05097081": "small",
  "n05097166": "stout",
  "n05097278": "tall",
  "n05097361": "highness",
  "n05097536": "high",
  "n05097706": "low",
  "n05097845": "lowness",
  "n05098099": "extreme",
  "n05098225": "extremeness",
  "n05098311": "amplitude",
  "n05098425": "amplitude level",
  "n05098520": "signal level",
  "n05098620": "noise level, background level",
  "n05098750": "multiplicity",
  "n05098857": "triplicity",
  "n05098942": "size",
  "n05099231": "size",
  "n05099389": "bulk, mass, volume",
  "n05099662": "muchness",
  "n05099796": "intensity, strength, intensity level",
  "n05100156": "threshold level",
  "n05100269": "field strength, field intensity",
  "n05100525": "magnetic field strength, magnetic intensity, magnetic induction, magnetic flux density",
  "n05100751": "candlepower, light intensity",
  "n05100866": "acoustic power, sound pressure level",
  "n05100981": "acoustic radiation pressure",
  "n05101174": "half-intensity",
  "n05101261": "circumference, perimeter",
  "n05101407": "girth",
  "n05101512": "spread",
  "n05101675": "circumference",
  "n05101815": "diameter, diam",
  "n05102101": "radius, r",
  "n05102300": "semidiameter",
  "n05102435": "curvature",
  "n05102578": "radius of curvature",
  "n05102764": "center of curvature, centre of curvature",
  "n05102889": "circle of curvature, osculating circle",
  "n05103072": "thickness",
  "n05103283": "bore, gauge, caliber, calibre",
  "n05103430": "gauge",
  "n05103503": "windage",
  "n05103648": "thinness, tenuity, slenderness",
  "n05103946": "largeness, bigness",
  "n05104256": "ampleness",
  "n05104421": "bulkiness, massiveness",
  "n05104548": "enormousness, grandness, greatness, immenseness, immensity, sizeableness, vastness, wideness",
  "n05105009": "enormity",
  "n05105265": "capaciousness, roominess, spaciousness, commodiousness",
  "n05105745": "airiness",
  "n05105879": "seating capacity",
  "n05106024": "fullness, voluminosity, voluminousness",
  "n05106220": "gigantism, giantism",
  "n05106317": "largeness, extensiveness",
  "n05106633": "smallness, littleness",
  "n05106928": "diminutiveness, minuteness, petiteness, tininess, weeness",
  "n05107216": "delicacy, slightness",
  "n05107322": "grain",
  "n05107495": "puniness, runtiness, stuntedness",
  "n05107668": "dwarfishness",
  "n05107765": "amount",
  "n05108109": "positivity, positiveness",
  "n05108262": "negativity, negativeness",
  "n05108412": "critical mass",
  "n05108740": "quantity",
  "n05108947": "increase, increment",
  "n05109324": "amplification, gain",
  "n05109511": "complement, accompaniment",
  "n05109808": "decrease, decrement",
  "n05109986": "fare increase",
  "n05110107": "price increase",
  "n05110185": "raise, rise, wage hike, hike, wage increase, salary increase",
  "n05110408": "rise, boost, hike, cost increase",
  "n05110583": "smallness",
  "n05110772": "supplement, supplementation",
  "n05110988": "tax-increase, tax boost, tax hike",
  "n05111141": "up-tick",
  "n05111248": "loop gain",
  "n05111511": "correction",
  "n05111704": "voltage drop",
  "n05111835": "drop, dip, fall, free fall",
  "n05112215": "shrinkage",
  "n05112308": "dollar volume, turnover",
  "n05112474": "stuffiness, closeness",
  "n05112609": "sufficiency, adequacy",
  "n05112910": "ampleness",
  "n05113133": "insufficiency, inadequacy, deficiency",
  "n05113462": "meagerness, meagreness, leanness, poorness, scantiness, scantness, exiguity",
  "n05113929": "wateriness",
  "n05114262": "abstemiousness",
  "n05114371": "deficit, shortage, shortfall",
  "n05114652": "oxygen deficit",
  "n05114781": "sparseness, spareness, sparsity, thinness",
  "n05115040": "abundance, copiousness, teemingness",
  "n05115418": "amplitude, bountifulness, bounty",
  "n05115568": "plenty, plentifulness, plenteousness, plenitude, plentitude",
  "n05115804": "profusion, profuseness, richness, cornucopia",
  "n05116128": "wealth",
  "n05116243": "luxuriance, lushness, voluptuousness",
  "n05116486": "overgrowth",
  "n05116590": "greenness, verdancy, verdure",
  "n05116730": "wilderness",
  "n05116953": "scarcity, scarceness",
  "n05117140": "dearth, paucity",
  "n05117237": "rarity, rareness, infrequency",
  "n05117406": "slenderness",
  "n05117660": "moderation, moderateness",
  "n05117977": "golden mean",
  "n05118062": "reasonableness",
  "n05118251": "immoderation, immoderateness",
  "n05118437": "excess, excessiveness, inordinateness",
  "n05118707": "sun protection factor, SPF",
  "n05118862": "extravagance, extravagancy",
  "n05119096": "exorbitance, outrageousness",
  "n05119223": "luxury",
  "n05119367": "overabundance, overmuch, overmuchness, superabundance",
  "n05119714": "excess, surplus, surplusage, nimiety",
  "n05119837": "glut, oversupply, surfeit",
  "n05120000": "bellyful",
  "n05120116": "overplus, plethora, superfluity, embarrassment",
  "n05120310": "redundancy, redundance",
  "n05120564": "fifth wheel, deadwood",
  "n05120683": "margin",
  "n05120835": "margin of safety, safety margin, margin of error",
  "n05121095": "narrow margin, narrowness, slimness",
  "n05121418": "number, figure",
  "n05121908": "numerousness, numerosity, multiplicity",
  "n05122099": "preponderance, prevalence",
  "n05122295": "multitudinousness",
  "n05122419": "innumerableness, countlessness",
  "n05122557": "majority, bulk",
  "n05122850": "minority",
  "n05123098": "fewness",
  "n05123206": "roundness",
  "n05123416": "extent",
  "n05123760": "coverage",
  "n05123935": "frontage",
  "n05124057": "limit, bound, boundary",
  "n05124404": "knife-edge",
  "n05124534": "starkness, absoluteness, utterness",
  "n05124792": "thermal barrier, heat barrier",
  "n05124928": "utmost, uttermost, maximum, level best",
  "n05125193": "verge, brink",
  "n05125377": "scope, range, reach, orbit, compass, ambit",
  "n05126066": "approximate range, ballpark",
  "n05126228": "confines",
  "n05126362": "contrast",
  "n05126611": "internationality, internationalism",
  "n05126849": "register",
  "n05127150": "head register, head voice, head tone",
  "n05127357": "falsetto",
  "n05127500": "chest register, chest voice, chest tone",
  "n05127640": "latitude",
  "n05127782": "horizon, view, purview",
  "n05127959": "sweep, expanse",
  "n05128096": "gamut",
  "n05128219": "spectrum",
  "n05128370": "palette, pallet",
  "n05128519": "area, expanse, surface area",
  "n05128870": "acreage, land area",
  "n05129054": "footprint",
  "n05129201": "length",
  "n05129565": "distance, length",
  "n05129928": "arm's length",
  "n05130028": "gauge",
  "n05130210": "light time",
  "n05130402": "skip distance",
  "n05130614": "wingspan, wingspread",
  "n05130735": "wingspread",
  "n05130875": "yardage",
  "n05131023": "hour, minute",
  "n05131194": "mileage, milage",
  "n05131283": "elevation",
  "n05131537": "isometry",
  "n05131647": "altitude, height",
  "n05131902": "level",
  "n05132045": "grade, ground level",
  "n05132221": "water level",
  "n05132340": "sea level",
  "n05132542": "ceiling",
  "n05132667": "ceiling",
  "n05132827": "absolute ceiling",
  "n05132962": "combat ceiling, service ceiling",
  "n05133107": "length",
  "n05133287": "longness",
  "n05133535": "extension, lengthiness, prolongation",
  "n05133748": "coextension",
  "n05133839": "elongation",
  "n05133944": "shortness",
  "n05134122": "curtailment",
  "n05134238": "briefness",
  "n05134353": "depth",
  "n05134547": "depth, deepness",
  "n05134880": "deepness, profundity, profoundness",
  "n05135155": "draft, draught",
  "n05135288": "penetration",
  "n05135451": "sounding",
  "n05135582": "bottomlessness",
  "n05135725": "shallowness",
  "n05135951": "superficiality",
  "n05136150": "width, breadth",
  "n05136343": "wideness, broadness",
  "n05136546": "beam",
  "n05136662": "thickness, heaviness",
  "n05136788": "narrowness",
  "n05136978": "fineness, thinness",
  "n05137165": "height, tallness",
  "n05137402": "highness, loftiness",
  "n05137557": "lowness",
  "n05137778": "squatness, stubbiness",
  "n05137938": "shortness, truncation",
  "n05138065": "third dimension",
  "n05138208": "worth",
  "n05138488": "value",
  "n05138958": "merit, virtue",
  "n05139094": "demerit, fault",
  "n05139342": "praisworthiness",
  "n05139436": "worthwhileness",
  "n05139561": "worthlessness, ineptitude",
  "n05139942": "fecklessness",
  "n05140086": "groundlessness, idleness",
  "n05140278": "paltriness, sorriness",
  "n05140454": "valuelessness",
  "n05140593": "shoddiness, trashiness",
  "n05140793": "damn, darn, hoot, red cent, shit, shucks, tinker's damn, tinker's dam",
  "n05141040": "vanity, emptiness",
  "n05141222": "invaluableness, preciousness, pricelessness, valuableness",
  "n05141492": "gold",
  "n05141683": "price",
  "n05141840": "desirability, desirableness",
  "n05142008": "undesirability",
  "n05142180": "good, goodness",
  "n05142641": "benefit, welfare",
  "n05142863": "advantage, reward",
  "n05143077": "sake, interest",
  "n05143300": "behalf",
  "n05143558": "better",
  "n05143690": "better",
  "n05143864": "optimum",
  "n05144079": "bad, badness",
  "n05144453": "worse",
  "n05144663": "evil",
  "n05144880": "Four Horsemen",
  "n05145118": "monetary value, price, cost",
  "n05145573": "average cost",
  "n05145708": "marginal cost, incremental cost, differential cost",
  "n05145891": "expensiveness",
  "n05146055": "assessment",
  "n05146178": "tax assessment",
  "n05146272": "costliness, dearness, preciousness",
  "n05146471": "lavishness, luxury, sumptuosity, sumptuousness",
  "n05146739": "inexpensiveness",
  "n05146904": "reasonableness, moderateness, modestness",
  "n05147237": "bargain rate, cheapness, cut rate, cut price",
  "n05147381": "fruitfulness, fecundity",
  "n05147586": "richness, rankness, prolificacy, fertility",
  "n05147940": "productiveness, productivity",
  "n05148186": "fruitlessness, aridity, barrenness",
  "n05148399": "poorness",
  "n05148539": "unproductiveness",
  "n05148699": "utility, usefulness",
  "n05149127": "detergency, detergence",
  "n05149325": "function, purpose, role, use",
  "n05149589": "raison d'etre",
  "n05149695": "helpfulness",
  "n05149832": "avail, help, service",
  "n05149978": "use",
  "n05150129": "serviceability, serviceableness, usableness, useableness, usability",
  "n05150458": "instrumentality",
  "n05150588": "inutility, uselessness, unusefulness",
  "n05150855": "futility",
  "n05150986": "worthlessness",
  "n05151088": "practicality",
  "n05151372": "functionality",
  "n05151576": "viability",
  "n05151701": "sensibleness",
  "n05151869": "realism, pragmatism",
  "n05152150": "practicability, practicableness",
  "n05152364": "feasibility, feasibleness",
  "n05152518": "impracticality",
  "n05152696": "idealism",
  "n05152902": "romanticism",
  "n05153037": "knight errantry, quixotism",
  "n05153155": "impracticability, impracticableness",
  "n05153359": "infeasibility, unfeasibility",
  "n05153520": "competence, competency",
  "n05153795": "fitness",
  "n05153897": "linguistic competence",
  "n05154114": "proficiency",
  "n05154241": "incompetence, incompetency",
  "n05154517": "asset, plus",
  "n05154676": "resource",
  "n05154908": "aid, assistance, help",
  "n05155123": "recourse, refuge, resort",
  "n05155349": "shadow",
  "n05155476": "resourcefulness",
  "n05155650": "inner resource",
  "n05155821": "advantage, vantage",
  "n05156319": "favor, favour",
  "n05156481": "leverage",
  "n05156673": "bargaining chip",
  "n05156822": "handicap",
  "n05156993": "homecourt advantage",
  "n05157143": "lead",
  "n05157274": "pull, clout",
  "n05157406": "start, head start",
  "n05157574": "profit, gain",
  "n05157732": "account",
  "n05157866": "profitableness, profitability, gainfulness, lucrativeness",
  "n05158095": "preference",
  "n05158296": "privilege",
  "n05158431": "expedience, expediency",
  "n05158619": "superiority, favorable position, favourable position",
  "n05158857": "edge",
  "n05158975": "inside track",
  "n05159123": "upper hand, whip hand",
  "n05159225": "forte, strong suit, long suit, metier, specialty, speciality, strong point, strength",
  "n05159495": "green thumb, green fingers",
  "n05159606": "weak point",
  "n05159725": "good",
  "n05159854": "common good, commonweal",
  "n05159948": "wisdom, wiseness, soundness",
  "n05160173": "unsoundness",
  "n05160399": "advisability",
  "n05160574": "reasonableness",
  "n05160796": "favorableness, favourableness, advantageousness, positivity, positiveness, profitableness",
  "n05161150": "auspiciousness, propitiousness",
  "n05161436": "liability",
  "n05161614": "disadvantage",
  "n05161967": "unfavorableness, unfavourableness",
  "n05162217": "inauspiciousness, unpropitiousness",
  "n05162455": "limitation",
  "n05162642": "defect, shortcoming",
  "n05162807": "awkwardness, nuisance value",
  "n05162985": "loss, deprivation",
  "n05163189": "penalty",
  "n05163401": "scratch",
  "n05163582": "game misconduct",
  "n05163807": "price, cost, toll",
  "n05164101": "richness",
  "n05164353": "death toll",
  "n05164521": "drawback",
  "n05164673": "catch, gimmick",
  "n05164845": "penalty",
  "n05165028": "inadvisability",
  "n05165175": "inferiority, unfavorable position",
  "n05165303": "inexpedience, inexpediency",
  "n05165533": "unprofitableness, unprofitability",
  "n05165745": "constructiveness",
  "n05165904": "destructiveness",
  "n05166072": "harmfulness, injuriousness",
  "n05166260": "insidiousness",
  "n05166397": "poison",
  "n05166560": "virulence, virulency",
  "n05166805": "positivity, positiveness, positivism",
  "n05167117": "affirmativeness",
  "n05167237": "assertiveness, self-assertiveness",
  "n05167412": "bumptiousness, cockiness, pushiness, forwardness",
  "n05167618": "negativity, negativeness, negativism",
  "n05167927": "occidentalism",
  "n05168063": "orientalism",
  "n05168261": "importance",
  "n05168697": "big deal",
  "n05168795": "face",
  "n05168890": "magnitude",
  "n05169037": "account",
  "n05169242": "matter",
  "n05169412": "momentousness",
  "n05169507": "prominence",
  "n05169601": "greatness, illustriousness",
  "n05169813": "significance",
  "n05170088": "historicalness",
  "n05170236": "meaningfulness",
  "n05170400": "purposefulness, sense of purpose",
  "n05170574": "consequence, import, moment",
  "n05170923": "hell to pay",
  "n05171045": "essentiality, essentialness",
  "n05171352": "vitalness",
  "n05171491": "indispensability, indispensableness, vitalness",
  "n05171800": "urgency",
  "n05171978": "edge, sharpness",
  "n05172131": "imperativeness, instancy",
  "n05172322": "weight, weightiness",
  "n05172596": "unimportance",
  "n05172815": "inessentiality",
  "n05172953": "dispensability, dispensableness",
  "n05173205": "pettiness, triviality, slightness, puniness",
  "n05173443": "joke",
  "n05173579": "insignificance",
  "n05173795": "meaninglessness",
  "n05174023": "inanity, senselessness, mindlessness, vacuity, pointlessness",
  "n05174326": "purposelessness, aimlessness",
  "n05174517": "inconsequence",
  "n05174653": "right",
  "n05175467": "access",
  "n05175613": "advowson",
  "n05175771": "cabotage",
  "n05175904": "claim, title",
  "n05176082": "due",
  "n05176188": "entree, access, accession, admission, admittance",
  "n05176341": "floor",
  "n05176477": "grant",
  "n05176607": "authority, authorization, authorisation, sanction",
  "n05176846": "human right",
  "n05177285": "legal right",
  "n05177705": "compulsory process",
  "n05177897": "conjugal right",
  "n05178220": "conjugal visitation right, conjugal visitation",
  "n05178394": "preemption, pre-emption",
  "n05178571": "preemption, pre-emption",
  "n05178715": "prerogative, privilege, perquisite, exclusive right",
  "n05179027": "public easement",
  "n05179180": "easement",
  "n05179410": "privilege of the floor",
  "n05179567": "privilege",
  "n05179838": "attorney-client privilege",
  "n05179993": "informer's privilege",
  "n05180135": "journalist's privilege",
  "n05180286": "marital communications privilege, husband-wife privilege",
  "n05180483": "physician-patient privilege",
  "n05180677": "priest-penitent privilege",
  "n05180881": "door",
  "n05181078": "open door",
  "n05181199": "title, claim",
  "n05181513": "own right",
  "n05181754": "entitlement",
  "n05181975": "right to privacy",
  "n05182080": "right to life",
  "n05182157": "right to liberty",
  "n05182240": "right to the pursuit of happiness",
  "n05182354": "freedom of thought",
  "n05182452": "equality before the law",
  "n05182563": "civil right",
  "n05183242": "civil liberty",
  "n05183453": "habeas corpus",
  "n05183624": "freedom of religion",
  "n05183779": "freedom of speech",
  "n05183932": "freedom of the press",
  "n05184082": "freedom of assembly",
  "n05184313": "freedom to bear arms",
  "n05184461": "freedom from search and seizure",
  "n05184620": "right to due process",
  "n05184810": "freedom from self-incrimination, privilege against self incrimination",
  "n05185106": "freedom from double jeopardy",
  "n05185270": "right to speedy and public trial by jury",
  "n05185444": "right to an attorney",
  "n05185598": "right to confront accusors",
  "n05185752": "freedom from cruel and unusual punishment",
  "n05185921": "freedom from involuntary servitude",
  "n05186090": "equal protection of the laws",
  "n05186306": "right to vote, vote, suffrage",
  "n05186626": "universal suffrage",
  "n05186766": "freedom from discrimination",
  "n05186998": "equal opportunity",
  "n05187187": "eminent domain",
  "n05187446": "franchise, enfranchisement",
  "n05187711": "representation",
  "n05187857": "right of action",
  "n05187941": "right of search",
  "n05188095": "right of way",
  "n05188220": "states' rights",
  "n05188367": "voting right",
  "n05188547": "water right, riparian right",
  "n05188646": "patent right",
  "n05188801": "right of election",
  "n05189057": "right of entry",
  "n05189208": "right of re-entry",
  "n05189396": "right of offset",
  "n05189591": "right of privacy",
  "n05189819": "right of way",
  "n05189946": "seat",
  "n05190106": "use, enjoyment",
  "n05190356": "usufruct",
  "n05190586": "visitation right",
  "n05190804": "power, powerfulness",
  "n05191486": "preponderance",
  "n05191695": "puissance",
  "n05191832": "persuasiveness, strength",
  "n05192130": "convincingness",
  "n05192240": "irresistibility, irresistibleness",
  "n05192451": "interest, interestingness",
  "n05192897": "newsworthiness, news",
  "n05193160": "topicality",
  "n05193338": "color, colour, vividness",
  "n05193621": "shrillness",
  "n05193781": "stranglehold, chokehold, throttlehold",
  "n05194043": "sway",
  "n05194151": "influence",
  "n05194435": "dead hand, dead hand of the past, mortmain",
  "n05194578": "force",
  "n05194874": "grip, grasp",
  "n05195175": "tentacle",
  "n05195362": "pressure",
  "n05195548": "duress",
  "n05195653": "heartbeat",
  "n05195796": "lifeblood",
  "n05195920": "wheel",
  "n05196054": "repellent, repellant",
  "n05196220": "hydrophobicity",
  "n05196375": "control",
  "n05196582": "authority, authorization, authorisation, potency, dominance, say-so",
  "n05197043": "corporatism",
  "n05197232": "hold",
  "n05197388": "iron fist",
  "n05197574": "rein",
  "n05197701": "carte blanche",
  "n05197797": "command",
  "n05197945": "imperium",
  "n05198036": "lordship",
  "n05198132": "muscle",
  "n05198321": "sovereignty",
  "n05198427": "legal power, jurisdiction",
  "n05198622": "disposal",
  "n05198756": "free will, discretion",
  "n05198928": "veto",
  "n05199151": "self-determination",
  "n05199286": "effectiveness, effectivity, effectualness, effectuality",
  "n05199680": "incisiveness, trenchancy",
  "n05199869": "efficacy, efficaciousness",
  "n05200169": "ability",
  "n05200670": "form",
  "n05200816": "interoperability",
  "n05201060": "magical ability, magical power",
  "n05201183": "lycanthropy",
  "n05201342": "Midas touch",
  "n05201452": "penetration",
  "n05201659": "physical ability",
  "n05201813": "contractility",
  "n05202034": "astringency, stypsis",
  "n05202284": "voice",
  "n05202413": "lung-power",
  "n05202497": "capability, capableness",
  "n05202954": "defensibility",
  "n05203207": "executability",
  "n05203397": "capacity",
  "n05203649": "military capability, military strength, strength, military posture, posture",
  "n05204004": "firepower",
  "n05204143": "operating capability, performance capability",
  "n05204316": "overkill",
  "n05204473": "envelope",
  "n05204637": "powerlessness, impotence, impotency",
  "n05204982": "helplessness, weakness, impuissance",
  "n05205220": "unpersuasiveness",
  "n05205340": "uninterestingness",
  "n05205537": "voicelessness",
  "n05205739": "dullness",
  "n05206006": "boringness, dreariness, insipidness, insipidity",
  "n05206237": "tediousness, tedium, tiresomeness",
  "n05206445": "drag",
  "n05206573": "jejunity, jejuneness, tameness, vapidity, vapidness",
  "n05206829": "ponderousness, heaviness",
  "n05207130": "inability, unfitness",
  "n05207352": "paper tiger",
  "n05207570": "incapability, incapableness",
  "n05207846": "incapacity",
  "n05207963": "ineffectiveness, ineffectualness, ineffectuality",
  "n05208291": "inefficacy, inefficaciousness",
  "n05208422": "romanticism, romance",
  "n05208595": "stardust",
  "n05208690": "analyticity",
  "n05208796": "compositeness",
  "n05208896": "primality",
  "n05208988": "selectivity",
  "n05209113": "domesticity",
  "n05209324": "infiniteness, infinitude, unboundedness, boundlessness, limitlessness",
  "n05209659": "finiteness, finitude, boundedness",
  "n05209822": "quantifiability, measurability",
  "n05210003": "ratability",
  "n05210116": "scalability",
  "n05210221": "solubility",
  "n05210373": "insolubility",
  "n05210534": "stuff",
  "n05210717": "comicality",
  "n05210820": "hot stuff, voluptuousness",
  "n05211044": "humor, humour",
  "n05211254": "pathos, poignancy",
  "n05211451": "tone",
  "n05211793": "optimism",
  "n05211974": "pessimism",
  "n05212168": "epicurism",
  "n05212264": "gourmandism",
  "n05212362": "brachycephaly, brachycephalism",
  "n05212493": "dolichocephaly, dolichocephalism",
  "n05212609": "relativity",
  "n05212808": "response",
  "n05212973": "responsiveness",
  "n05213201": "unresponsiveness, deadness",
  "n05213693": "frigidity, frigidness",
  "n05213894": "resistance",
  "n05214121": "subjectivism",
  "n05214211": "fair use",
  "n05214374": "fruition",
  "n05214469": "vascularity",
  "n05214599": "extension",
  "n05214851": "snootiness",
  "n05214976": "totipotency, totipotence",
  "n05215257": "ulteriority",
  "n05215483": "solvability, solubility",
  "n05215699": "unsolvability, insolubility",
  "n05215921": "memorability",
  "n05216102": "woodiness, woodsiness",
  "n05216240": "waxiness",
  "n05216365": "body, organic structure, physical structure",
  "n05217061": "life form",
  "n05217168": "human body, physical body, material body, soma, build, figure, physique, anatomy, shape, bod, chassis, frame, form, flesh",
  "n05217688": "person",
  "n05217859": "body, dead body",
  "n05218119": "cadaver, corpse, stiff, clay, remains",
  "n05218533": "cremains",
  "n05218630": "mummy",
  "n05218788": "live body",
  "n05218899": "apparatus",
  "n05219097": "system",
  "n05219297": "juvenile body",
  "n05219420": "child's body",
  "n05219561": "adult body",
  "n05219724": "male body",
  "n05219923": "female body",
  "n05220126": "adult female body, woman's body",
  "n05220306": "adult male body, man's body",
  "n05220461": "body part",
  "n05221526": "corpus",
  "n05221649": "adnexa, annexa",
  "n05221895": "area, region",
  "n05222467": "dilator",
  "n05222591": "groove, vallecula",
  "n05222790": "partition",
  "n05222940": "septum",
  "n05223110": "nasal septum",
  "n05223248": "costal groove",
  "n05223370": "fissure",
  "n05223550": "sulcus",
  "n05223823": "fissure of Rolando, Rolando's fissure, central sulcus, sulcus centralis",
  "n05224080": "fissure of Sylvius, Sylvian fissure, lateral cerebral sulcus, sulcus lateralis cerebri",
  "n05224358": "parieto-occipital sulcus, parieto-occipital fissure",
  "n05224585": "calcarine sulcus, calcarine fissure",
  "n05224755": "hilus, hilum",
  "n05224966": "erogenous zone",
  "n05225090": "external body part",
  "n05225439": "arthromere",
  "n05225602": "structure, anatomical structure, complex body part, bodily structure, body structure",
  "n05226937": "birth canal",
  "n05227079": "bulb",
  "n05227209": "carina",
  "n05227438": "carina fornicis",
  "n05227572": "fornix, trigonum cerebrale",
  "n05227868": "fornix",
  "n05228020": "mamillary body, mammillary body, corpus mamillare",
  "n05228264": "cauda",
  "n05228374": "keel",
  "n05228496": "chiasma, chiasm, decussation",
  "n05228732": "cingulum",
  "n05228881": "optic chiasma, optic chiasm, chiasma opticum",
  "n05229049": "optic radiation, radiatio optica",
  "n05229198": "concha",
  "n05229341": "nasal concha",
  "n05229468": "filament, filum",
  "n05229622": "fiber, fibre",
  "n05229805": "germ",
  "n05229990": "infundibulum",
  "n05230171": "interstice",
  "n05230357": "landmark",
  "n05230603": "craniometric point",
  "n05231264": "acanthion",
  "n05231397": "asterion",
  "n05231592": "auriculare, auricular point",
  "n05231769": "bregma",
  "n05231940": "condylion",
  "n05232074": "coronion",
  "n05232221": "crotaphion",
  "n05232345": "dacryon",
  "n05232503": "entomion",
  "n05232691": "glabella, mesophyron",
  "n05232972": "gnathion",
  "n05233100": "gonion",
  "n05233238": "inion",
  "n05233420": "jugale, jugal point",
  "n05233601": "lambda",
  "n05233741": "mandibular notch",
  "n05233875": "mastoidale",
  "n05234016": "metopion",
  "n05234162": "nasion",
  "n05234315": "obelion",
  "n05234438": "ophryon",
  "n05234593": "orbitale, orbital point",
  "n05234737": "pogonion",
  "n05234911": "prosthion, prostheon, alveolar point",
  "n05235100": "pterion",
  "n05235217": "rhinion",
  "n05235350": "sphenion",
  "n05235461": "stephanion",
  "n05235607": "symphysion",
  "n05235745": "limbus",
  "n05235879": "rib",
  "n05236029": "blade",
  "n05236152": "radicle",
  "n05236322": "plexus, rete",
  "n05236582": "aortic plexus",
  "n05236709": "autonomic plexus, plexus autonomici",
  "n05236848": "nerve plexus",
  "n05237227": "system",
  "n05237755": "body covering",
  "n05238036": "sheath, case",
  "n05238282": "skin, tegument, cutis",
  "n05239039": "pressure point",
  "n05239243": "integument",
  "n05239437": "skin graft",
  "n05239589": "buff",
  "n05239680": "dewlap",
  "n05239808": "epithelium, epithelial tissue",
  "n05240076": "exuviae",
  "n05240211": "epidermis, cuticle",
  "n05240522": "endothelium",
  "n05240710": "mesothelium",
  "n05240850": "neuroepithelium",
  "n05241072": "skin cell",
  "n05241218": "epidermal cell",
  "n05241374": "melanoblast",
  "n05241485": "melanocyte",
  "n05241662": "prickle cell",
  "n05241827": "epithelial cell",
  "n05242070": "columnar cell, columnar epithelial cell",
  "n05242239": "spongioblast",
  "n05242396": "cuboidal cell, cuboidal epithelial cell",
  "n05242525": "goblet cell",
  "n05242622": "hair cell",
  "n05242752": "Kupffer's cell",
  "n05242928": "squamous cell",
  "n05243077": "stratum corneum, corneum, horny layer",
  "n05243259": "stratum lucidum",
  "n05243435": "stratum granulosum",
  "n05243704": "stratum germinativum, stratum basale, malpighian layer, rete Malpighii",
  "n05243879": "dermis, corium, derma",
  "n05244045": "mantle, pallium",
  "n05244239": "plaque",
  "n05244421": "amyloid plaque, amyloid protein plaque",
  "n05244619": "arterial plaque",
  "n05244755": "dental plaque, bacterial plaque",
  "n05244934": "macule, macula",
  "n05245192": "freckle, lentigo",
  "n05245387": "liver spot",
  "n05245521": "plague spot",
  "n05245626": "whitehead, milium",
  "n05245775": "blackhead, comedo",
  "n05245906": "pore",
  "n05246089": "aortic orifice",
  "n05246215": "stoma",
  "n05246511": "tube, tube-shaped structure",
  "n05246796": "tubule",
  "n05246969": "microtubule",
  "n05247057": "salpinx",
  "n05247178": "nephron, uriniferous tubule",
  "n05247369": "malpighian body, malpighian corpuscle, renal corpuscle",
  "n05247621": "Bowman's capsule, glomerular capsule, capsula glomeruli",
  "n05247804": "glomerulus",
  "n05248006": "tomentum, tomentum cerebri",
  "n05248181": "passage, passageway",
  "n05248553": "meatus",
  "n05248667": "auditory meatus, acoustic meatus, ear canal, auditory canal, external auditory canal",
  "n05248921": "deltoid tuberosity, deltoid eminence",
  "n05249094": "nasal meatus",
  "n05249232": "spinal canal, vertebral canal, canalis vertebralis",
  "n05249420": "anastomosis, inosculation",
  "n05249636": "orifice, opening, porta",
  "n05250190": "porta hepatis",
  "n05250310": "spiracle",
  "n05250420": "blowhole",
  "n05250550": "stigma",
  "n05250659": "duct, epithelial duct, canal, channel",
  "n05251537": "ductule, ductulus",
  "n05251655": "canaliculus",
  "n05251789": "canal of Schlemm, Schlemm's canal, sinus venosus sclerae",
  "n05252016": "venous sinus, sinus",
  "n05252259": "cavernous sinus, sinus cavernosus",
  "n05252402": "coronary sinus, sinus coronarius",
  "n05252567": "sigmoid sinus, sinus sigmoideus",
  "n05252705": "straight sinus, tentorial sinus, sinus rectus",
  "n05252834": "transverse sinus, sinus transversus",
  "n05252970": "sinus",
  "n05253165": "ethmoid sinus, ethmoidal sinus, sinus ethmoidales",
  "n05253338": "frontal sinus",
  "n05253443": "maxillary sinus",
  "n05253561": "paranasal sinus, sinus paranasales, nasal sinus",
  "n05253829": "sinusoid",
  "n05253951": "locule, loculus",
  "n05254077": "lumen",
  "n05254197": "ampulla",
  "n05254393": "hair, pilus",
  "n05254627": "ingrown hair",
  "n05254795": "hair",
  "n05255435": "headful",
  "n05255578": "body hair",
  "n05255692": "down, pile",
  "n05255939": "lanugo",
  "n05256085": "mane, head of hair",
  "n05256220": "hairline",
  "n05256358": "part, parting",
  "n05256562": "widow's peak",
  "n05256702": "cowlick",
  "n05256862": "hairdo, hairstyle, hair style, coiffure, coif",
  "n05257393": "beehive",
  "n05257476": "bouffant",
  "n05257593": "haircut",
  "n05257737": "lock, curl, ringlet, whorl",
  "n05257967": "sausage curl",
  "n05258051": "forelock",
  "n05258177": "quiff",
  "n05258299": "crimp",
  "n05258462": "pin curl",
  "n05258627": "spit curl, kiss curl",
  "n05258743": "dreadlock",
  "n05258889": "Afro, Afro hairdo",
  "n05258985": "bang, fringe",
  "n05259109": "bob",
  "n05259240": "wave",
  "n05259426": "finger wave",
  "n05259512": "braid, plait, tress, twist",
  "n05259726": "chignon",
  "n05259823": "queue",
  "n05259914": "pigtail",
  "n05259991": "marcel",
  "n05260127": "pageboy",
  "n05260240": "pompadour",
  "n05260380": "ponytail",
  "n05260533": "permanent wave, permanent, perm",
  "n05260714": "brush cut",
  "n05260821": "crew cut, flattop",
  "n05260928": "mohawk, mohawk haircut",
  "n05261088": "roach",
  "n05261203": "scalp lock",
  "n05261310": "thatch",
  "n05261404": "facial hair",
  "n05261566": "beard, face fungus, whiskers",
  "n05261894": "fuzz",
  "n05261999": "imperial, imperial beard",
  "n05262120": "beaver",
  "n05262185": "mustache, moustache",
  "n05262422": "soup-strainer, toothbrush",
  "n05262534": "mustachio, moustachio, handle-bars",
  "n05262698": "walrus mustache, walrus moustache",
  "n05262802": "sideburn, burnside, mutton chop, side-whiskers",
  "n05263029": "goatee",
  "n05263183": "stubble",
  "n05263316": "vandyke beard, vandyke",
  "n05263448": "soul patch, Attilio",
  "n05263587": "pubic hair, bush, crotch hair",
  "n05263732": "minge",
  "n05263850": "body substance",
  "n05264081": "solid body substance",
  "n05264247": "scab",
  "n05264417": "eschar",
  "n05264545": "fundus",
  "n05264756": "funiculus",
  "n05264913": "node",
  "n05265038": "nodule",
  "n05265139": "smear, cytologic smear, cytosmear",
  "n05265417": "alimentary tract smear",
  "n05265736": "esophageal smear",
  "n05265861": "gastric smear",
  "n05265981": "oral smear",
  "n05266096": "paraduodenal smear, duodenal smear",
  "n05266239": "cervical smear, Pap smear, Papanicolaou smear",
  "n05266486": "lower respiratory tract smear, bronchoscopic smear, sputum smear",
  "n05266759": "vaginal smear",
  "n05266879": "specimen",
  "n05267073": "cytologic specimen",
  "n05267211": "isthmus, band",
  "n05267345": "tissue",
  "n05267548": "animal tissue",
  "n05268112": "flesh",
  "n05268255": "areolar tissue",
  "n05268383": "beta cell",
  "n05268510": "capillary bed",
  "n05268619": "parenchyma",
  "n05268797": "interstitial tissue",
  "n05268965": "adipose tissue, fat, fatty tissue",
  "n05269401": "flab",
  "n05269495": "atheroma",
  "n05269668": "cellulite",
  "n05269784": "puppy fat",
  "n05269901": "bone, os",
  "n05271383": "anklebone, astragal, astragalus, talus",
  "n05271607": "bare bone",
  "n05271685": "cuboid bone",
  "n05271814": "carpal bone, carpal, wrist bone",
  "n05272110": "carpal tunnel",
  "n05272276": "scaphoid bone, os scaphoideum, navicular",
  "n05272423": "lunate bone, semilunar bone, os lunatum",
  "n05272545": "triquetral, triquetral bone, os triquetrum, cuneiform bone, pyramidal bone",
  "n05272745": "pisiform, pisiform bone, os pisiforme",
  "n05272891": "trapezium, trapezium bone, os trapezium",
  "n05273073": "trapezoid, trapezoid bone, os trapezoideum",
  "n05273223": "capitate, capitate bone, os capitatum",
  "n05273408": "hamate, hamate bone, unciform bone, os hamatum",
  "n05273555": "cartilage bone",
  "n05273684": "centrum",
  "n05273822": "cheekbone, zygomatic bone, zygomatic, malar, malar bone, jugal bone, os zygomaticum",
  "n05274105": "clavicle, collarbone",
  "n05274247": "coccyx, tail bone",
  "n05274446": "dentine, dentin",
  "n05274590": "ethmoid, ethmoid bone",
  "n05274808": "heelbone, calcaneus, os tarsi fibulare",
  "n05274959": "hipbone, innominate bone",
  "n05275162": "hyoid, hyoid bone, os hyoideum",
  "n05275315": "ilium",
  "n05275466": "ischium, ischial bone, os ischii",
  "n05275651": "long bone, os longum",
  "n05275905": "lower jaw, mandible, mandibula, mandibular bone, submaxilla, lower jawbone, jawbone, jowl",
  "n05276290": "ramus",
  "n05276425": "raphe, rhaphe",
  "n05276566": "palatine raphe",
  "n05276668": "mandibular joint, temporomandibular joint, articulatio temporomandibularis",
  "n05276860": "membrane bone",
  "n05277100": "mentum",
  "n05277261": "metacarpal, metacarpal bone",
  "n05277405": "metatarsal",
  "n05277532": "nasal, nasal bone, os nasale",
  "n05277728": "ossicle, bonelet, ossiculum",
  "n05277941": "auditory ossicle",
  "n05278152": "palatine, palatine bone, os palatinum",
  "n05278395": "patella, kneecap, kneepan",
  "n05278584": "phalanx",
  "n05278714": "pubis, pubic bone, os pubis",
  "n05278922": "punctum",
  "n05279026": "rib, costa",
  "n05279321": "round bone",
  "n05279407": "sacrum",
  "n05279688": "scapula, shoulder blade, shoulder bone",
  "n05279953": "glenoid fossa, glenoid cavity",
  "n05280154": "glenoid fossa, mandibular fossa",
  "n05280365": "acromion, acromial process",
  "n05280512": "sesamoid bone, sesamoid, os sesamoideum",
  "n05280698": "short bone, os breve",
  "n05280831": "socket",
  "n05280998": "sphenoid bone, sphenoid, os sphenoidale",
  "n05281189": "sternum, breastbone",
  "n05281452": "gladiolus, corpus sternum",
  "n05281584": "manubrium",
  "n05281691": "xiphoid process",
  "n05281874": "tarsal, tarsal bone",
  "n05282000": "temporal bone, os temporale",
  "n05282247": "primary dentition",
  "n05282339": "secondary dentition",
  "n05282433": "dentition, teeth",
  "n05282652": "diastema",
  "n05282746": "tooth",
  "n05283326": "pulp cavity",
  "n05283498": "chopper, pearly",
  "n05283649": "carnassial tooth",
  "n05283816": "turbinate bone, turbinate, turbinal",
  "n05284020": "tympanic bone",
  "n05284132": "upper jaw, upper jawbone, maxilla, maxillary",
  "n05284333": "vertebra",
  "n05284617": "intervertebral disc, intervertebral disk",
  "n05284851": "zygoma, zygomatic arch, arcus zygomaticus",
  "n05285101": "hip socket",
  "n05285275": "eye socket, orbit, cranial orbit, orbital cavity",
  "n05285480": "tooth socket, alveolus",
  "n05285623": "marrow, bone marrow",
  "n05285835": "red marrow, red bone marrow",
  "n05286008": "yellow marrow, yellow bone marrow",
  "n05286167": "axolemma",
  "n05286272": "basilar membrane",
  "n05286412": "cambium",
  "n05286536": "connective tissue",
  "n05287090": "collagen",
  "n05287368": "elastic tissue",
  "n05287607": "endoneurium",
  "n05287731": "elastin",
  "n05287882": "lymphatic tissue, lymphoid tissue",
  "n05288091": "cartilage, gristle",
  "n05288396": "meniscus, semilunar cartilage",
  "n05288593": "fibrocartilage",
  "n05288739": "hyaline cartilage",
  "n05288912": "erectile tissue",
  "n05289057": "muscle, muscular tissue",
  "n05289297": "muscle, musculus",
  "n05289601": "contractile organ, contractor",
  "n05289743": "striated muscle tissue",
  "n05289861": "skeletal muscle, striated muscle",
  "n05290756": "head",
  "n05290897": "voluntary muscle",
  "n05291010": "abductor, abductor muscle",
  "n05291230": "musculus abductor digiti minimi manus",
  "n05291373": "musculus abductor digiti minimi pedis",
  "n05291495": "musculus abductor hallucis",
  "n05291604": "musculus abductor pollicis",
  "n05291728": "adductor, adductor muscle",
  "n05291945": "musculus adductor brevis",
  "n05292073": "musculus adductor longus",
  "n05292200": "musculus adductor magnus, great adductor muscle",
  "n05292359": "musculus adductor hallucis",
  "n05292468": "pronator",
  "n05292587": "supinator",
  "n05292736": "levator",
  "n05292856": "anconeous muscle, musculus anconeus",
  "n05293040": "antagonistic muscle",
  "n05293268": "agonist",
  "n05293420": "antagonist",
  "n05293597": "articular muscle",
  "n05293773": "musculus articularis cubiti",
  "n05293944": "musculus articularis genus",
  "n05294068": "cheek muscle, buccinator muscle, musculus buccinator",
  "n05294256": "masseter",
  "n05294374": "platysma",
  "n05294606": "extensor muscle, extensor",
  "n05294819": "quadriceps, quadriceps femoris, musculus quadriceps femoris, quad",
  "n05294995": "fibrous tissue",
  "n05295160": "trabecula",
  "n05295381": "ligament",
  "n05295593": "falciform ligament",
  "n05295740": "round ligament of the uterus, ligamentum teres uteri",
  "n05296001": "perineurium",
  "n05296127": "perimysium",
  "n05296253": "tendon, sinew",
  "n05296503": "flexor muscle, flexor",
  "n05296639": "articulatory system",
  "n05296775": "nervous tissue, nerve tissue",
  "n05296968": "ganglion",
  "n05297163": "autonomic ganglion",
  "n05297363": "otic ganglion, otoganglion",
  "n05297523": "organ",
  "n05298159": "primordium, anlage",
  "n05298313": "vital organ, vitals",
  "n05298421": "effector",
  "n05298572": "external organ",
  "n05298729": "internal organ, viscus",
  "n05298988": "viscera, entrails, innards",
  "n05299178": "sense organ, sensory receptor, receptor",
  "n05299687": "interoceptor, enteroceptor",
  "n05299814": "exteroceptor",
  "n05299927": "third eye, pineal eye",
  "n05300128": "baroreceptor",
  "n05300231": "chemoreceptor",
  "n05300397": "thermoreceptor",
  "n05300507": "auditory system",
  "n05300675": "auditory apparatus",
  "n05300926": "visual system",
  "n05301072": "tongue, lingua, glossa, clapper",
  "n05301392": "articulator",
  "n05301526": "glottis",
  "n05301752": "epiglottis",
  "n05301908": "mouth",
  "n05302307": "trap, cakehole, hole, maw, yap, gob",
  "n05302422": "os",
  "n05302499": "mouth, oral cavity, oral fissure, rima oris",
  "n05302899": "buccal cavity",
  "n05303020": "incompetent cervix",
  "n05303232": "cervix, uterine cervix, cervix uteri",
  "n05303402": "cavity, bodily cavity, cavum",
  "n05304252": "antrum",
  "n05304341": "cloaca",
  "n05304603": "vestibule",
  "n05304795": "vestibule of the ear",
  "n05304932": "gingiva, gum",
  "n05305136": "tastebud, taste bud, gustatory organ",
  "n05305389": "taste cell, gustatory cell",
  "n05305614": "speech organ, vocal organ, organ of speech",
  "n05305806": "lip",
  "n05306055": "overlip",
  "n05306122": "underlip",
  "n05306190": "front tooth, anterior",
  "n05306390": "bucktooth",
  "n05306476": "back tooth, posterior",
  "n05306604": "malposed tooth",
  "n05306708": "permanent tooth, adult tooth",
  "n05306894": "primary tooth, deciduous tooth, baby tooth, milk tooth",
  "n05307091": "canine, canine tooth, eyetooth, eye tooth, dogtooth, cuspid",
  "n05307358": "premolar, bicuspid",
  "n05307521": "cusp",
  "n05307641": "incisor",
  "n05307773": "molar, grinder",
  "n05307952": "wisdom tooth",
  "n05308141": "crown",
  "n05308310": "root, tooth root",
  "n05308481": "root canal",
  "n05308655": "enamel, tooth enamel",
  "n05308810": "cementum, cement",
  "n05308950": "pulp",
  "n05309050": "tonsil, palatine tonsil, faucial tonsil, tonsilla",
  "n05309245": "uvula",
  "n05309392": "soft palate, velum",
  "n05309591": "hard palate",
  "n05309725": "palate, roof of the mouth",
  "n05309965": "ala",
  "n05310206": "alveolar arch",
  "n05310351": "alveolar ridge, gum ridge, alveolar process",
  "n05310596": "caul, veil, embryonic membrane",
  "n05310790": "fetal membrane",
  "n05311054": "eye, oculus, optic",
  "n05311795": "naked eye",
  "n05312040": "peeper",
  "n05312149": "oculus dexter, OD",
  "n05312227": "oculus sinister, OS",
  "n05312306": "simple eye, stemma, ocellus",
  "n05312427": "compound eye",
  "n05312614": "ommatidium",
  "n05312782": "cell membrane, cytomembrane, plasma membrane",
  "n05313115": "choroid, choroid coat",
  "n05313344": "ciliary body",
  "n05313535": "eyebrow, brow, supercilium",
  "n05313679": "protective fold",
  "n05313822": "eyelid, lid, palpebra",
  "n05314075": "canthus",
  "n05314255": "epicanthus, epicanthic fold",
  "n05314462": "nasal canthus",
  "n05314549": "temporal canthus",
  "n05314639": "nictitating membrane, third eyelid",
  "n05314834": "haw",
  "n05314919": "eyelash, lash, cilium",
  "n05315095": "conjunctiva",
  "n05315382": "bulbar conjunctiva, conjunctival layer of bulb, tunica conjunctiva bulbi",
  "n05315612": "palpebra conjunctiva, conjunctival layer of eyelids, tunica conjunctiva palpebrarum",
  "n05315855": "pinguecula",
  "n05316025": "eyeball, orb",
  "n05316175": "ocular muscle, eye muscle",
  "n05316407": "abducens muscle, lateral rectus muscle, lateral rectus, rectus lateralis",
  "n05316590": "rectus",
  "n05316674": "inferior rectus muscle, inferior rectus, rectus inferior",
  "n05316850": "medial rectus muscle, medial rectus, rectus medialis",
  "n05317013": "superior rectus muscle, superior rectus, rectus superior",
  "n05317191": "capsule",
  "n05317354": "cornea",
  "n05317603": "pterygium",
  "n05317815": "arcus, arcus senilis",
  "n05317960": "uvea",
  "n05318137": "uveoscleral pathway",
  "n05318263": "aqueous humor, aqueous humour",
  "n05318407": "vitreous humor, vitreous humour, vitreous body",
  "n05318606": "diaphragm, midriff",
  "n05318831": "eardrum, tympanum, tympanic membrane, myringa",
  "n05319028": "endocranium",
  "n05319144": "endosteum",
  "n05319279": "ependyma",
  "n05319419": "fertilization membrane",
  "n05319588": "hyaloid membrane, hyaloid",
  "n05319760": "intima",
  "n05319936": "iris",
  "n05320183": "pupil",
  "n05320362": "lens, crystalline lens, lens of the eye",
  "n05320636": "lens cortex, cortex",
  "n05320764": "lens nucleus, nucleus",
  "n05320899": "ear",
  "n05321144": "organ of hearing",
  "n05321307": "inner ear, internal ear, labyrinth",
  "n05321664": "membranous labyrinth",
  "n05321917": "bony labyrinth, osseous labyrinth",
  "n05322103": "endolymph",
  "n05322247": "perilymph",
  "n05322432": "utricle, utriculus",
  "n05322570": "saccule, sacculus",
  "n05322735": "modiolus",
  "n05322855": "organ of Corti",
  "n05323036": "vestibular apparatus, vestibular system",
  "n05323228": "semicircular canal",
  "n05323454": "stretch receptor",
  "n05323588": "earlobe, ear lobe",
  "n05323723": "external ear, outer ear",
  "n05323889": "auricle, pinna, ear",
  "n05324134": "tragus",
  "n05324276": "cauliflower ear",
  "n05324388": "perforated eardrum",
  "n05324553": "umbo",
  "n05324691": "mediastinum",
  "n05324888": "middle ear, tympanic cavity, tympanum",
  "n05325153": "Eustachian tube, auditory tube",
  "n05325378": "fenestra",
  "n05325606": "fenestra ovalis, fenestra vestibuli, oval window, fenestra of the vestibule",
  "n05325786": "fenestra rotunda, fenestra cochleae, round window, fenestra of the cochlea",
  "n05325943": "malleus, hammer",
  "n05326060": "lamella",
  "n05326200": "lens capsule",
  "n05326340": "incus, anvil",
  "n05326465": "stapes, stirrup",
  "n05326624": "cochlea",
  "n05326900": "meninx, meninges",
  "n05327134": "mucous membrane, mucosa",
  "n05327373": "periosteum",
  "n05327652": "perithelium",
  "n05327767": "gland, secretory organ, secretor, secreter",
  "n05328115": "oil gland",
  "n05328232": "sebaceous gland, sebaceous follicle, glandulae sebaceae",
  "n05328508": "Meibomian gland, tarsal gland",
  "n05328695": "Montgomery's tubercle",
  "n05328867": "exocrine gland, exocrine, duct gland",
  "n05329215": "digestive system, gastrointestinal system, systema alimentarium, systema digestorium",
  "n05329533": "endocrine system",
  "n05329735": "endocrine gland, endocrine, ductless gland",
  "n05330244": "thyroid gland, thyroid",
  "n05330365": "parathyroid gland, parathyroid",
  "n05330518": "sweat duct",
  "n05330659": "sweat gland, sudoriferous gland",
  "n05330848": "apocrine gland",
  "n05331035": "eccrine gland",
  "n05331171": "adrenal gland, adrenal, suprarenal gland",
  "n05331404": "prostate gland, prostate",
  "n05331653": "lacrimal gland, lachrymal gland, tear gland",
  "n05331812": "lacrimal duct, lachrymal duct, tear duct",
  "n05331988": "lacrimal sac, tear sac, dacryocyst",
  "n05332225": "lacrimal bone",
  "n05332438": "nasolacrimal duct",
  "n05332569": "thymus gland, thymus",
  "n05332802": "kidney",
  "n05333259": "excretory organ, urinary organ",
  "n05333467": "spleen, lien",
  "n05333777": "artery, arteria, arterial blood vessel",
  "n05335310": "alveolar artery, arteria alveolaris",
  "n05335515": "inferior alveolar artery, arteria alveolaris inferior",
  "n05335700": "superior alveolar artery, arteria alveolaris superior",
  "n05335850": "angular artery, arteria angularis",
  "n05335971": "aorta",
  "n05336246": "ascending aorta",
  "n05336377": "aortic arch",
  "n05336487": "descending aorta",
  "n05336635": "abdominal aorta",
  "n05336748": "thoracic aorta",
  "n05336893": "appendicular artery, arteria appendicularis",
  "n05337055": "arcuate artery, arteria arcuata",
  "n05337178": "arcuate artery of the kidney",
  "n05337301": "arteriole, arteriola, capillary artery",
  "n05337485": "artery of the penis bulb, arteria bulbi penis",
  "n05337663": "artery of the vestibule bulb, arteria bulbi vestibuli",
  "n05337855": "ascending artery, arteria ascendens",
  "n05338025": "auricular artery, arteria auricularis",
  "n05338166": "axillary artery, arteria axillaris",
  "n05338410": "basilar artery, arteria basilaris",
  "n05338614": "brachial artery, arteria brachialis",
  "n05338847": "radial artery, arteria radialis",
  "n05339047": "bronchial artery",
  "n05339168": "buccal artery, arteria buccalis",
  "n05339357": "carotid artery, arteria carotis",
  "n05339583": "common carotid artery, common carotid",
  "n05339751": "external carotid artery, external carotid",
  "n05339946": "internal carotid artery",
  "n05340121": "carotid body",
  "n05340317": "celiac trunk, celiac artery, truncus celiacus, arteria celiaca",
  "n05340599": "central artery of the retina, arteria centralis retinae",
  "n05340795": "cerebellar artery, arteria cerebelli",
  "n05340972": "inferior cerebellar artery",
  "n05341089": "superior cerebellar artery",
  "n05341206": "cerebral artery, arteria cerebri",
  "n05341416": "anterior cerebral artery",
  "n05341641": "middle cerebral artery",
  "n05341789": "posterior cerebral artery",
  "n05341920": "cervical artery, areteria cervicalis",
  "n05342070": "choroidal artery, arteria choroidea",
  "n05342214": "ciliary artery, arteria ciliaris",
  "n05342374": "circle of Willis",
  "n05342499": "circumflex artery",
  "n05342673": "circumflex artery of the thigh, arteria circumflexa femoris",
  "n05342854": "circumflex humeral artery, arteria circumflexa humeri",
  "n05343037": "circumflex iliac artery, arteria circumflexa ilium",
  "n05343218": "circumflex scapular artery, arteria circumflexa scapulae",
  "n05343408": "colic artery, arteria colica",
  "n05343542": "communicating artery, arteria communicans",
  "n05343718": "coronary artery, arteria coronaria",
  "n05343941": "atrial artery",
  "n05344073": "right coronary artery",
  "n05344213": "left coronary artery",
  "n05344350": "cystic artery, arteria cystica",
  "n05344514": "digital arteries, arteria digitalis",
  "n05344697": "epigastric artery, arteria epigastrica",
  "n05344848": "ethmoidal artery, arteria ethmoidalis",
  "n05345038": "facial artery, arteria facialis, external maxillary artery",
  "n05345247": "femoral artery, arteria femoralis",
  "n05345421": "popliteal artery, arteria poplitea",
  "n05345581": "gastric artery, arteria gastrica",
  "n05345783": "right gastric artery, ateria gastrica dextra",
  "n05345978": "left gastric artery, arteria gastrica sinistra",
  "n05346191": "short gastric artery, arteria gastrica breves, vasa brevis",
  "n05346406": "gluteal artery, arteria glutes",
  "n05346585": "hepatic artery, arteria hepatica",
  "n05346714": "ileal artery, intestinal artery, arteria ileum",
  "n05346892": "ileocolic artery, arteria ileocolica",
  "n05347146": "iliac artery, arteria iliaca",
  "n05347359": "common iliac artery",
  "n05347465": "external iliac artery",
  "n05347631": "internal iliac artery, hypogastric artery",
  "n05347871": "iliolumbar artery, arteria iliolumbalis",
  "n05348054": "infraorbital artery, arteria infraorbitalis",
  "n05348269": "innominate artery",
  "n05348540": "intercostal artery, arteria intercostalis",
  "n05348698": "jejunal artery, intestinal artery",
  "n05348884": "labial artery, arteria labialis",
  "n05349101": "inferior labial artery, arteria labialis inferior",
  "n05349273": "superior labial artery, arteria labialis superior",
  "n05349445": "labyrinthine artery, artery of the labyrinth, internal auditory artery",
  "n05349659": "lacrimal artery, arteria lacrimalis",
  "n05349906": "laryngeal artery, arteria laryngea",
  "n05350061": "lienal artery, splenic artery, arteria lienalis",
  "n05350256": "lingual artery, arteria lingualis",
  "n05350453": "lumbar artery, arteria lumbalis",
  "n05350679": "maxillary artery, arteria maxillaris",
  "n05350900": "internal maxillary artery",
  "n05351058": "meningeal artery, arteria meningea",
  "n05351278": "anterior meningeal artery",
  "n05351442": "middle meningeal artery",
  "n05351572": "posterior meningeal artery",
  "n05351746": "mesenteric artery, arteria mesenterica",
  "n05351968": "inferior mesenteric artery",
  "n05352112": "superior mesenteric artery",
  "n05352291": "metacarpal artery, arteria metacarpea",
  "n05352433": "metatarsal artery, arteria metatarsea",
  "n05352601": "musculophrenic artery, arteria musculophrenica",
  "n05352754": "nutrient artery, arteria nutricia",
  "n05352896": "ophthalmic artery, arteria ophthalmica",
  "n05353070": "ovarian artery, arteria ovarica",
  "n05353241": "palatine artery, arteria palatina",
  "n05353364": "pancreatic artery, arteria pancreatica",
  "n05353507": "perineal artery, arteria perinealis",
  "n05353683": "pudendal artery, arteria pudenda",
  "n05353819": "pulmonary artery, arteria pulmonalis",
  "n05354026": "pulmonary trunk, truncus pulmonalis",
  "n05354238": "rectal artery, arteria rectalis",
  "n05354381": "renal artery, arteria renalis",
  "n05354580": "subclavian artery, arteria subclavia",
  "n05354744": "temporal artery",
  "n05354955": "anterior temporal artery, arteria temporalis anterior",
  "n05355146": "intermediate temporal artery, arteria temporalis intermedia",
  "n05355341": "posterior temporal artery, arteria temporalis posterior",
  "n05355527": "testicular artery, internal spermatic artery, arteria testicularis",
  "n05355706": "ulnar artery, arteria ulnaris",
  "n05355890": "uterine artery, arteria uterina",
  "n05356083": "vaginal artery, arteria vaginalis",
  "n05356291": "vertebral artery, arteria vertebralis",
  "n05356442": "accessory cephalic vein, vena cephalica accessoria",
  "n05356657": "accessory hemiazygos vein, accessory hemiazygous vein, vena hemiazygos accessoria",
  "n05356889": "accessory vertebral vein, vena vertebralis accessoria",
  "n05357160": "accompanying vein, vena comitans",
  "n05357366": "anastomotic vein, vena anastomotica",
  "n05357500": "angular vein, vena angularis",
  "n05357682": "anterior vertebral vein, vena vertebralis anterior",
  "n05357867": "appendicular vein, vena appendicularis",
  "n05358036": "arcuate vein of the kidney, vena arcuata renis",
  "n05358226": "auricular vein, vena auricularis",
  "n05358345": "axillary vein, vena axillaris",
  "n05358507": "azygos vein, azygous vein, vena azygos",
  "n05358764": "basal vein, vena basalis",
  "n05358944": "basilic vein, vena basilica",
  "n05359124": "basivertebral vein, vena basivertebralis",
  "n05359347": "brachial vein, vena brachialis",
  "n05359516": "brachiocephalic vein, innominate vein, vena brachiocephalica",
  "n05359695": "bronchial vein, vena bronchialis",
  "n05359828": "cardinal vein",
  "n05360042": "anterior cardinal vein",
  "n05360170": "posterior cardinal vein",
  "n05360297": "common cardinal vein",
  "n05360470": "central veins of liver, venae centrales hepatis",
  "n05360678": "central vein of retina, vena centrales retinae",
  "n05360880": "central vein of suprarenal gland, vena centralis glandulae suprarenalis",
  "n05361123": "cephalic vein, vena cephalica",
  "n05361278": "cerebellar vein, vena cerebellum",
  "n05361391": "cerebral vein, vena cerebri",
  "n05361640": "anterior cerebral vein, vena cerebri anterior",
  "n05361805": "anterior facial vein, vena facialis anterior",
  "n05362016": "great cerebral vein, vena cerebri magna",
  "n05362197": "inferior cerebral vein, venae cerebrum inferior",
  "n05362405": "internal cerebral vein, vena cerebrum internus",
  "n05362593": "middle cerebral vein, vena cerebri media",
  "n05362745": "deep middle cerebral vein",
  "n05362912": "superficial middle cerebral vein",
  "n05363064": "superior cerebral vein, vena cerebrum superior",
  "n05363270": "cervical vein, deep cervical vein, vena cervicalis profunda",
  "n05363428": "choroid vein, vena choroidea",
  "n05363559": "ciliary veins, venae ciliares",
  "n05363676": "circumflex vein, vena circumflexa",
  "n05363868": "circumflex iliac vein, vena circumflexa ilium",
  "n05364000": "circumflex femoral vein, vena circumflexus femoris",
  "n05364184": "clitoral vein, vena clitoridis",
  "n05364345": "colic vein, vena colica",
  "n05364448": "common facial vein",
  "n05364612": "conjunctival veins, venae conjunctivales",
  "n05364730": "costoaxillary vein",
  "n05364905": "cutaneous vein, vena cutanea",
  "n05365059": "cystic vein, vena cystica",
  "n05365164": "digital vein, vena digitalis",
  "n05365284": "diploic vein, vena diploica",
  "n05365420": "dorsal scapular vein, vena scapularis dorsalis",
  "n05365633": "dorsal root, dorsal horn",
  "n05365838": "emissary vein, vena emissaria",
  "n05366043": "epigastric vein",
  "n05366206": "inferior epigastric vein, vena epigastrica inferior",
  "n05366341": "superficial epigastric vein, vena epigastrica superficialis",
  "n05366485": "superior epigastric veins, venae epigastricae superiores",
  "n05366640": "episcleral veins, venae episclerales",
  "n05366812": "esophageal veins, oesophageal veins, venae esophageae",
  "n05367003": "ethmoidal vein, vena ethmoidalis",
  "n05367165": "external nasal vein, vena nasalis externa",
  "n05367341": "facial vein, vena facialis",
  "n05367508": "femoral vein, vena femoralis",
  "n05367735": "gastric vein, vena gastrica",
  "n05367912": "gastroomental vein, gastroepiploic vein, vena gastroomentalis",
  "n05368100": "genicular vein, vena genus",
  "n05368278": "glans",
  "n05368444": "glans clitoridis",
  "n05368594": "glans penis",
  "n05368739": "gluteal vein, vena gluteus",
  "n05368874": "hemizygos vein, hemizygous vein, vena hemizygos",
  "n05369092": "hemorrhoidal vein, rectal vein, vena rectalis",
  "n05369254": "hepatic vein, vena hepatica",
  "n05369404": "hypogastric vein, internal iliac vein",
  "n05369566": "ileocolic vein, vena ileocolica",
  "n05369757": "external iliac vein",
  "n05369924": "common iliac vein",
  "n05370125": "iliac vein, vena iliaca",
  "n05370293": "iliolumbar vein, vena iliolumbalis",
  "n05370410": "intercapitular vein, vena intercapitalis",
  "n05370636": "intercostal vein, vena intercostalis",
  "n05370781": "intervertebral vein, vena intervertebralis",
  "n05370918": "jugular vein, vena jugularis, jugular",
  "n05371144": "anterior jugular vein",
  "n05371301": "external jugular vein",
  "n05371482": "internal jugular vein",
  "n05371663": "labial vein, vena labialis",
  "n05371829": "inferior labial vein, vena labialis inferior",
  "n05371977": "superior labial vein, vena labialis superior",
  "n05372125": "labial vein, vena labialis",
  "n05372290": "labyrinthine vein, internal auditory vein",
  "n05372428": "lacrimal vein, vena lacrimalis",
  "n05372593": "laryngeal vein, vena laryngea",
  "n05372725": "left gastric vein, vena gastrica sinistra",
  "n05372924": "lingual vein, vena lingualis",
  "n05373145": "lumbar vein, vena lumbalis",
  "n05373300": "maxillary vein, vena maxillaris",
  "n05373495": "meningeal veins, venae meningeae",
  "n05373616": "mesenteric vein, vena mesenterica",
  "n05373790": "metacarpal vein, vena metacarpus",
  "n05373924": "metatarsal vein, vena metatarsus",
  "n05374101": "musculophrenic vein, vena musculophrenica",
  "n05374280": "nasofrontal vein, vena nasofrontalis",
  "n05374475": "oblique vein of the left atrium, vena obliqua atrii sinistri",
  "n05374658": "obturator vein, vena obturatoria",
  "n05374861": "occipital vein, vena occipitalis",
  "n05374980": "ophthalmic vein, vena ophthalmica",
  "n05375169": "inferior ophthalmic vein",
  "n05375322": "superior ophthalmic vein",
  "n05375530": "ovarian vein, vena ovarica",
  "n05375729": "palatine vein, vena palatina",
  "n05375881": "pancreatic vein, venae pancreatica",
  "n05376050": "paraumbilical vein, vena paraumbilicalis",
  "n05376219": "parotid vein",
  "n05376391": "pectoral vein, vena pectoralis",
  "n05376541": "perforating vein, vena perforantis",
  "n05376721": "pericardial vein, vena pericardiaca",
  "n05376844": "peroneal vein, fibular vein, vena peroneus",
  "n05377098": "pharyngeal vein, vena pharyngeus",
  "n05377252": "phrenic vein, vena phrenica",
  "n05377371": "popliteal vein, vena poplitea",
  "n05377515": "portal system",
  "n05377665": "portal vein, hepatic portal vein, portal, vena portae",
  "n05377833": "posterior vein of the left ventricle, vena posterior ventriculi sinistri",
  "n05378022": "prepyloric vein, vena pylorica",
  "n05378137": "pudendal vein, venae pudendum",
  "n05378234": "pulmonary vein, vena pulmonalis",
  "n05378458": "inferior pulmonary vein, vena pulmanalis inferior",
  "n05378654": "superior pulmonary vein, vena pulmonalis superior",
  "n05378850": "pyloric vein, right gastric vein, vena gastrica-dextra",
  "n05379039": "radial vein, vena radialis",
  "n05379247": "renal vein, vena renalis",
  "n05379438": "retromandibular vein, vena retromandibularis, posterior facial vein",
  "n05379635": "sacral vein, vena sacralis",
  "n05379734": "saphenous vein, vena saphena",
  "n05379944": "long saphenous vein, great saphenous vein",
  "n05380122": "short saphenous vein",
  "n05380252": "scleral veins, venae sclerales",
  "n05380404": "scrotal vein, vena scrotalis",
  "n05380532": "sigmoid vein, vena sigmoideus",
  "n05380697": "spinal vein, vena spinalis",
  "n05380822": "splenic vein, vena lienalis",
  "n05381036": "stellate venule",
  "n05381148": "sternocleidomastoid vein, vena sternocleidomastoidea",
  "n05381325": "stylomastoid vein, vena stylomastoidea",
  "n05381491": "subclavian vein, vena subclavia",
  "n05381666": "sublingual vein, vena sublingualis",
  "n05381779": "supraorbital vein, vena supraorbitalis",
  "n05381956": "supratrochlear vein, vena supratrochlearis",
  "n05382135": "temporal vein, vena temporalis",
  "n05382316": "deep temporal vein",
  "n05382432": "middle temporal vein",
  "n05382589": "superficial temporal vein",
  "n05382729": "testicular vein, vena testicularis",
  "n05382855": "thalamostriate vein",
  "n05383004": "thoracoepigastric vein, vena thoracoepigastrica",
  "n05383216": "superior thalamostriate vein",
  "n05383348": "inferior thalamostriate vein, striate vein",
  "n05383467": "thoracic vein, vena thoracica",
  "n05383598": "thyroid vein, vena thyroidea",
  "n05383791": "inferior thyroid vein",
  "n05383956": "middle thyroid vein",
  "n05384128": "superior thyroid vein",
  "n05384300": "tibial vein, vena tibialis",
  "n05384446": "tracheal vein, vena trachealis",
  "n05384560": "tympanic vein",
  "n05384691": "ulnar vein, vena ulnaris",
  "n05384817": "umbilical vein, vena umbilicalis",
  "n05384999": "uterine vein",
  "n05385161": "gallbladder, gall bladder",
  "n05385363": "hypochondrium",
  "n05385534": "liver",
  "n05386139": "Haversian canal",
  "n05386309": "hepatic lobe",
  "n05386426": "hepatic duct",
  "n05386544": "inguinal canal, canalis inguinalis",
  "n05386845": "common bile duct, bile duct",
  "n05387028": "biliary ductule",
  "n05387167": "pancreas",
  "n05387395": "pancreatic duct",
  "n05387544": "lung",
  "n05387842": "alveolar bed",
  "n05387959": "lobe of the lung",
  "n05388115": "pleura",
  "n05388311": "parietal pleura",
  "n05388437": "visceral pleura",
  "n05388527": "pleural cavity",
  "n05388665": "pleural space",
  "n05388805": "heart, pump, ticker",
  "n05389182": "athlete's heart",
  "n05389310": "biauriculate heart",
  "n05389460": "pacemaker, cardiac pacemaker, sinoatrial node, SA node",
  "n05389625": "cusp, leaflet",
  "n05389762": "flap",
  "n05389939": "cardiac muscle, heart muscle",
  "n05390233": "papillary muscle",
  "n05390479": "atrioventricular bundle, bundle of His, atrioventricular trunk, truncus atrioventricularis",
  "n05390761": "atrioventricular node",
  "n05391000": "myocardium",
  "n05391118": "Purkinje fiber",
  "n05391269": "Purkinje network, Purkinje's tissue, Purkinje's system",
  "n05391540": "area of cardiac dullness",
  "n05391763": "ventricle, heart ventricle",
  "n05391977": "left ventricle",
  "n05392157": "right ventricle",
  "n05392348": "auricle, atrial auricle, auricula atrii",
  "n05392562": "auricula, auricular appendage, auricular appendix",
  "n05392744": "chamber",
  "n05392906": "cranial cavity, intracranial cavity",
  "n05393023": "atrium",
  "n05393230": "atrium cordis, atrium of the heart",
  "n05393429": "right atrium, right atrium of the heart, atrium dextrum",
  "n05393629": "left atrium, left atrium of the heart, atrium sinistrum",
  "n05393813": "mitral valve, bicuspid valve, left atrioventricular valve",
  "n05393996": "tricuspid valve, right atrioventricular valve",
  "n05394277": "atrioventricular valve",
  "n05394526": "aortic valve",
  "n05394707": "pulmonary valve",
  "n05394904": "semilunar valve",
  "n05395098": "heart valve, cardiac valve",
  "n05395286": "valve",
  "n05395548": "valvule, valvelet, valvula",
  "n05395690": "stomach, tummy, tum, breadbasket",
  "n05396071": "epigastrium",
  "n05396229": "cardia",
  "n05396366": "lymphatic system, systema lymphaticum",
  "n05396690": "thoracic duct",
  "n05396807": "lymph vessel, lymphatic vessel",
  "n05397032": "lacteal",
  "n05397178": "vascular structure",
  "n05397333": "vessel, vas",
  "n05397468": "liquid body substance, bodily fluid, body fluid, humor, humour",
  "n05398023": "extracellular fluid, ECF",
  "n05398298": "interstitial fluid",
  "n05398462": "intracellular fluid",
  "n05398609": "juice, succus",
  "n05398760": "cancer juice",
  "n05398871": "karyolymph",
  "n05399034": "milk",
  "n05399243": "mother's milk",
  "n05399356": "colostrum, foremilk",
  "n05399486": "amniotic cavity",
  "n05399627": "amniotic fluid, amnionic fluid, waters",
  "n05399847": "blood",
  "n05400445": "arterial blood",
  "n05400601": "blood group, blood type",
  "n05400860": "A, type A, group A",
  "n05400978": "B, type B, group B",
  "n05401096": "AB, type AB, group AB",
  "n05401229": "O, type O, group O",
  "n05401411": "Rh-positive blood type, Rh positive",
  "n05401587": "Rh-negative blood type, Rh-negative blood, Rh negative",
  "n05401753": "gore",
  "n05401851": "lifeblood",
  "n05401951": "bloodstream, blood stream",
  "n05402091": "clot, coagulum",
  "n05402333": "blood clot, grume",
  "n05402472": "cord blood",
  "n05402576": "menorrhea, menstrual blood, menstrual flow",
  "n05402788": "venous blood",
  "n05402961": "whole blood",
  "n05403149": "serum, blood serum",
  "n05403427": "plasma, plasm, blood plasma",
  "n05403702": "antiserum",
  "n05403849": "chyle",
  "n05404074": "lymph",
  "n05404336": "semen, seed, seminal fluid, ejaculate, cum, come",
  "n05404598": "ink",
  "n05404728": "secretion",
  "n05405139": "lacrimal secretion, lachrymal secretion",
  "n05405324": "tear, teardrop",
  "n05405554": "lacrimal apparatus",
  "n05405751": "perspiration, sweat, sudor",
  "n05405946": "digestive juice, digestive fluid",
  "n05406128": "gastric juice, gastric acid",
  "n05406397": "pancreatic juice",
  "n05406570": "bile, gall",
  "n05406782": "black bile, melancholy",
  "n05406958": "yellow bile, choler",
  "n05407119": "hormone, endocrine, internal secretion",
  "n05407772": "intestinal juice",
  "n05407890": "noradrenaline, norepinephrine",
  "n05408113": "adrenocorticotropic hormone, adrenocorticotrophic hormone, ACTH, adrenocorticotropin, adrenocorticotrophin, corticotropin, corticotrophin",
  "n05408388": "epinephrine, epinephrin, adrenaline, Adrenalin",
  "n05408684": "gastrointestinal hormone, GI hormones",
  "n05408908": "gastrin",
  "n05409136": "cholecystokinin",
  "n05409426": "secretin",
  "n05409663": "ghrelin",
  "n05409945": "motilin",
  "n05410131": "glucagon",
  "n05410315": "gonadotropin, gonadotrophin, gonadotropic hormone, gonadotrophic hormone",
  "n05410646": "insulin",
  "n05410901": "Lente Insulin, Lente Iletin",
  "n05411049": "recombinant human insulin, Humulin",
  "n05411338": "melatonin",
  "n05411430": "neurohormone",
  "n05411571": "oxytocin, Pitocin",
  "n05411785": "parathyroid hormone, parathormone",
  "n05412053": "relaxin",
  "n05412242": "releasing hormone, RH, releasing factor, hypothalamic releasing hormone, hypothalamic releasing factor",
  "n05412649": "somatotropin, somatotrophin, somatotropic hormone, somatotrophic hormone, STH, human growth hormone, growth hormone",
  "n05412912": "Protropin",
  "n05413102": "thymosin",
  "n05413241": "thyroid hormone",
  "n05413465": "calcitonin, thyrocalcitonin",
  "n05413647": "thyroxine, thyroxin, tetraiodothyronine, T",
  "n05413873": "triiodothyronine, liothyronine, T",
  "n05414147": "vasopressin, antidiuretic hormone, ADH, Pitressin",
  "n05414534": "autacoid, autocoid",
  "n05414724": "histamine",
  "n05414931": "prostaglandin",
  "n05415200": "synovia, synovial fluid",
  "n05415395": "mucus, mucous secretion",
  "n05415815": "phlegm, sputum",
  "n05416048": "snot",
  "n05416128": "booger",
  "n05416198": "saliva, spit, spittle",
  "n05416545": "salivary duct",
  "n05416678": "drool, dribble, drivel, slobber",
  "n05416861": "tobacco juice",
  "n05416979": "sebum",
  "n05417162": "smegma",
  "n05417272": "lochia",
  "n05417472": "pus, purulence, suppuration, ichor, sanies, festering",
  "n05417698": "gleet",
  "n05417821": "leukorrhea, leucorrhea",
  "n05417975": "blood vessel",
  "n05418177": "ductus arteriosus",
  "n05418432": "patent ductus arteriosus",
  "n05418585": "vasa vasorum",
  "n05418717": "vein, vena, venous blood vessel",
  "n05421414": "venation, venous blood system",
  "n05421586": "varicose vein",
  "n05421723": "vena bulbi penis",
  "n05421900": "vena canaliculi cochleae",
  "n05421997": "vein of penis",
  "n05422135": "venae dorsales penis superficiales",
  "n05422317": "venae dorsales penis profunda",
  "n05422457": "vena profunda penis",
  "n05422574": "vena bulbi vestibuli",
  "n05422668": "vena cava",
  "n05422852": "inferior vena cava, postcava",
  "n05423095": "superior vena cava, precava",
  "n05423319": "venae profundae clitoridis",
  "n05423443": "vena dorsalis clitoridis profunda",
  "n05423595": "venae dorsales clitoridis superficiales",
  "n05423779": "venae palpebrales",
  "n05423882": "venae interlobulares renis",
  "n05424074": "venae interlobulares hepatis",
  "n05424199": "venae renis",
  "n05424351": "venae labiales anteriores, anterior labial veins",
  "n05424504": "venae labiales posteriores, posterior labial veins",
  "n05424679": "venter",
  "n05424796": "venter",
  "n05424963": "ventral root, ventral horn, anterior root, anterior horn",
  "n05425207": "vertebral vein, vena vertebralis",
  "n05425436": "vesical vein, vena vesicalis",
  "n05425581": "vestibular vein, vena vestibularis",
  "n05425703": "vortex vein, vorticose vein, vena vorticosum",
  "n05425910": "capillary, capillary vessel",
  "n05426087": "venule, venula, capillary vein",
  "n05426243": "membrane, tissue layer",
  "n05426989": "retina",
  "n05427346": "ganglion cell, gangliocyte",
  "n05427570": "sarcolemma",
  "n05427739": "peritoneum",
  "n05427946": "peritoneal cavity, greater peritoneal sac",
  "n05428136": "bursa omentalis, omental bursa, lesser peritoneal cavity",
  "n05428331": "endocardium",
  "n05428473": "pericardium",
  "n05428645": "epicardium, visceral pericardium",
  "n05428794": "parietal pericardium",
  "n05428974": "pericardial cavity, pericardial space",
  "n05429195": "mesentery",
  "n05429404": "mesocolon",
  "n05429525": "omentum",
  "n05429658": "greater omentum, gastrocolic omentum, caul",
  "n05429840": "lesser omentum",
  "n05429994": "submucosa",
  "n05430095": "lymph node, lymph gland, node",
  "n05430300": "axillary node",
  "n05430454": "Peyer's patch, Peter's gland",
  "n05430628": "somatic cell, vegetative cell",
  "n05431177": "neoplastic cell",
  "n05431285": "cancer cell",
  "n05431402": "blastema",
  "n05431585": "energid, protoplast",
  "n05431762": "pronucleus",
  "n05431926": "zygote, fertilized ovum",
  "n05432230": "heterozygote",
  "n05432420": "homozygote",
  "n05432623": "parthenote",
  "n05432736": "protoplasm, living substance",
  "n05432948": "cytoplasm, cytol",
  "n05433400": "cytoplast",
  "n05433496": "cytoskeleton",
  "n05433709": "cytosol",
  "n05433856": "ectoplasm",
  "n05433953": "endoplasm",
  "n05434053": "hyaloplasm, ground substance",
  "n05434203": "lysosome",
  "n05434361": "microsome",
  "n05434557": "Golgi body, Golgi apparatus, Golgi complex, dictyosome",
  "n05434784": "nucleoplasm, karyoplasm",
  "n05434927": "nucleus, cell nucleus, karyon",
  "n05435277": "nucleolus, nucleole",
  "n05435477": "nucleolus organizer, nucleolus organiser, nucleolar organizer, nucleolar organiser",
  "n05435722": "germ plasm, plasm",
  "n05435855": "sex chromatin",
  "n05436080": "chromatin, chromatin granule",
  "n05436381": "achromatin",
  "n05436548": "linin",
  "n05436752": "gene, cistron, factor",
  "n05437600": "dominant gene",
  "n05437785": "allele, allelomorph",
  "n05438130": "dominant allele, dominant",
  "n05438315": "recessive allele, recessive",
  "n05438483": "genetic marker",
  "n05438642": "homeotic gene",
  "n05438778": "homeobox, homeobox gene",
  "n05438956": "lethal gene",
  "n05439100": "linkage group, linked genes",
  "n05439286": "modifier, modifier gene",
  "n05439413": "mutant gene",
  "n05439582": "haplotype",
  "n05439810": "cystic fibrosis transport regulator, CFTR",
  "n05439942": "nonallele",
  "n05440046": "operator gene",
  "n05440207": "operon",
  "n05440410": "oncogene, transforming gene",
  "n05440555": "polygene",
  "n05440756": "proto-oncogene",
  "n05440875": "recessive gene",
  "n05441053": "regulatory gene, regulator gene",
  "n05441224": "repressor gene",
  "n05441338": "structural gene",
  "n05441468": "suppressor, suppresser, suppressor gene, suppresser gene",
  "n05441686": "transgene",
  "n05441806": "tumor suppressor gene",
  "n05441929": "X-linked gene",
  "n05442022": "Y-linked gene, holandric gene",
  "n05442131": "chromosome",
  "n05442594": "X chromosome",
  "n05442789": "XX",
  "n05442916": "XXX",
  "n05443050": "XXY",
  "n05443211": "XY",
  "n05443333": "XYY",
  "n05443484": "Y chromosome",
  "n05443651": "sex chromosome",
  "n05443966": "autosome, somatic chromosome",
  "n05444175": "chromatid",
  "n05444324": "centromere, kinetochore",
  "n05444610": "acentric chromosome",
  "n05444709": "acrocentric chromosome",
  "n05444876": "karyotype",
  "n05445099": "metacentric chromosome",
  "n05445247": "telocentric chromosome",
  "n05445389": "mitochondrion, chondriosome",
  "n05445546": "sarcosome",
  "n05445668": "organelle, cell organelle, cell organ",
  "n05446006": "aster",
  "n05446181": "centriole",
  "n05446394": "ribosome",
  "n05446767": "centrosome, central body",
  "n05446971": "sarcoplasm",
  "n05447087": "vacuole",
  "n05447218": "sclera, sclerotic coat",
  "n05447423": "semipermeable membrane",
  "n05447599": "bone cell",
  "n05447757": "embryonic cell, formative cell",
  "n05447962": "blastocyte",
  "n05448053": "ameloblast",
  "n05448149": "osteoblast, bone-forming cell",
  "n05448257": "erythroblast",
  "n05448400": "fibroblast",
  "n05448501": "neuroblast",
  "n05448597": "myelocyte",
  "n05448704": "myeloblast",
  "n05448827": "sideroblast",
  "n05448928": "megakaryocyte",
  "n05449073": "osteoclast",
  "n05449196": "osteocyte",
  "n05449268": "blood cell, blood corpuscle, corpuscle",
  "n05449538": "akaryocyte, akaryote, acaryote",
  "n05449661": "megalocyte, macrocyte",
  "n05449797": "megaloblast",
  "n05449959": "leukocyte, leucocyte, white blood cell, white cell, white blood corpuscle, white corpuscle, WBC",
  "n05450370": "packed cells",
  "n05450617": "histiocyte",
  "n05450740": "macrophage",
  "n05450888": "phagocyte, scavenger cell",
  "n05451099": "fixed phagocyte",
  "n05451265": "free phagocyte",
  "n05451384": "lymphocyte, lymph cell",
  "n05451695": "B cell, B lymphocyte",
  "n05451981": "T cell, T lymphocyte",
  "n05452204": "helper T cell, helper cell, CD4 T cell, CD4 cell",
  "n05452516": "killer T cell, killer cell, cytotoxic T cell, CD8 T cell, CD8 cell",
  "n05452765": "lymphoblast",
  "n05452845": "plasma cell, plasmacyte",
  "n05453057": "plasmablast",
  "n05453145": "granulocyte",
  "n05453267": "monocyte",
  "n05453412": "monoblast",
  "n05453523": "basophil, basophile",
  "n05453657": "neutrophil, neutrophile",
  "n05453815": "microphage",
  "n05453943": "eosinophil, eosinophile",
  "n05454070": "red blood cell, RBC, erythrocyte",
  "n05454452": "acanthocyte",
  "n05454578": "microcyte",
  "n05454702": "reticulocyte",
  "n05454833": "sickle cell",
  "n05454978": "siderocyte",
  "n05455113": "spherocyte",
  "n05455206": "target cell",
  "n05455375": "fovea, fovea centralis",
  "n05455563": "parafovea",
  "n05455690": "macula, macula lutea, macular area, yellow spot",
  "n05455912": "visual cell",
  "n05456082": "blind spot, optic disc, optic disk",
  "n05456257": "cone, cone cell, retinal cone",
  "n05456456": "rod, rod cell, retinal rod",
  "n05456622": "fat cell, adipose cell",
  "n05456732": "reproductive cell, germ cell, sex cell",
  "n05456945": "gamete",
  "n05457201": "anisogamete",
  "n05457343": "isogamete",
  "n05457469": "sperm, sperm cell, spermatozoon, spermatozoan",
  "n05457795": "acrosome",
  "n05457973": "ovum, egg cell",
  "n05458173": "ootid",
  "n05458300": "ovule",
  "n05458412": "gametocyte",
  "n05458576": "oocyte",
  "n05458703": "polar body",
  "n05458859": "spermatocele",
  "n05458990": "spermatocyte",
  "n05459101": "spermatid",
  "n05459232": "muscle cell, muscle fiber, muscle fibre",
  "n05459457": "Leydig cell, Leydig's cell",
  "n05459590": "Sertoli cell, Sertoli's cell",
  "n05459769": "striated muscle cell, striated muscle fiber",
  "n05459953": "myofibril, myofibrilla, sarcostyle",
  "n05460163": "sarcomere",
  "n05460291": "smooth muscle",
  "n05460473": "smooth muscle, involuntary muscle",
  "n05460759": "smooth muscle cell",
  "n05460870": "immune system",
  "n05461179": "integumentary system",
  "n05461349": "reticuloendothelial system, RES",
  "n05461610": "mononuclear phagocyte system, MPS, system of macrophages",
  "n05461816": "muscular structure, musculature, muscle system",
  "n05462057": "musculoskeletal system",
  "n05462315": "nervous system, systema nervosum",
  "n05462674": "neural structure",
  "n05463449": "reflex arc",
  "n05463533": "center, centre, nerve center, nerve centre",
  "n05463873": "auditory center",
  "n05464104": "nerve fiber, nerve fibre",
  "n05464378": "medullated nerve fiber, myelinated nerve fiber",
  "n05464534": "Ranvier's nodes, nodes of Ranvier",
  "n05464685": "medullary sheath, myelin sheath",
  "n05464939": "neurolemma, neurilemma",
  "n05465071": "Schwann cell",
  "n05465228": "effector",
  "n05465392": "end organ",
  "n05465567": "nerve cell, neuron",
  "n05465868": "brain cell",
  "n05466005": "Golgi's cell, Golgi cell",
  "n05466244": "Purkinje cell",
  "n05466393": "end-plate, endplate, motor end plate",
  "n05466575": "osmoreceptor",
  "n05466696": "motor neuron, efferent neuron, motor nerve fiber, motoneuron",
  "n05466892": "sensory neuron, afferent neuron",
  "n05467054": "neuroglia, glia",
  "n05467432": "neurogliacyte, neuroglial cell, glial cell",
  "n05467619": "astroglia, macroglia",
  "n05467758": "astrocyte",
  "n05467922": "fibrous astrocyte",
  "n05468098": "protoplasmic astrocyte",
  "n05468211": "microglia",
  "n05468351": "microgliacyte",
  "n05468523": "oligodendroglia, oligodendria",
  "n05468739": "oligodendrocyte",
  "n05468849": "axon, axone",
  "n05469032": "nerve ending, nerve end",
  "n05469240": "free nerve ending",
  "n05469424": "Pacinian corpuscle",
  "n05469664": "proprioceptor",
  "n05469861": "dendrite",
  "n05470013": "hybridoma",
  "n05470189": "process, outgrowth, appendage",
  "n05471181": "caruncle, caruncula",
  "n05471427": "wattle, lappet",
  "n05471629": "condyle",
  "n05471837": "condylar process, condyloid process, mandibular condyle",
  "n05472032": "coronoid process, processus coronoideus",
  "n05472205": "coronoid process of the mandible",
  "n05472394": "lateral condyle",
  "n05472538": "medial condyle",
  "n05472681": "epicondyle",
  "n05472846": "lateral epicondyle",
  "n05472959": "fimbria",
  "n05473104": "apophysis",
  "n05473318": "spicule, spiculum",
  "n05473507": "osteophyte",
  "n05473593": "papilla",
  "n05473735": "papilla",
  "n05473928": "synapse",
  "n05474195": "neuromuscular junction, myoneural junction",
  "n05474346": "nerve, nervus",
  "n05474738": "motor nerve, efferent nerve, efferent",
  "n05474976": "motor fiber, efferent fiber",
  "n05475134": "sensory nerve, afferent nerve, afferent",
  "n05475397": "sensory fiber, afferent fiber",
  "n05475562": "lemniscus, fillet",
  "n05475681": "fiber bundle, fibre bundle, fascicle, fasciculus",
  "n05475878": "nerve pathway, tract, nerve tract, pathway",
  "n05476094": "commissure",
  "n05476256": "cranial nerve",
  "n05476592": "depressor, depressor nerve",
  "n05476754": "peduncle, cerebral peduncle",
  "n05476915": "hemisphere, cerebral hemisphere",
  "n05477112": "left hemisphere, left brain",
  "n05477305": "pyriform area, piriform area, pyriform lobe, piriform lobe",
  "n05477510": "right hemisphere, right brain",
  "n05477686": "rhinencephalon, olfactory brain",
  "n05477946": "olfactory nerve, nervii olfactorii, first cranial nerve",
  "n05478139": "olfactory bulb",
  "n05478336": "optic nerve, nervus opticus, second cranial nerve, optic tract",
  "n05478526": "oculomotor, oculomotor nerve, nervus oculomotorius, third cranial nerve",
  "n05478684": "trochlear, trochlear nerve, trochlearis, fourth cranial nerve",
  "n05478896": "trigeminal, trigeminal nerve, trigeminus, nervus trigeminus, fifth cranial nerve",
  "n05479108": "abducent, abducent nerve, abducens, abducens nerve, nervus abducens, sixth cranial nerve",
  "n05479314": "facial, facial nerve, nervus facialis, seventh cranial nerve",
  "n05479503": "acoustic nerve, auditory nerve, vestibulocochlear nerve, nervus vestibulocochlearis, eighth cranial nerve",
  "n05479786": "glossopharyngeal nerve, nervus glossopharyngeus, ninth cranial nerve",
  "n05480076": "vagus, vagus nerve, nervus vagus, pneumogastric, pneumogastric nerve, tenth cranial nerve, wandering nerve",
  "n05480401": "accessory nerve, spinal accessory, nervus accessorius, eleventh cranial nerve",
  "n05480607": "hypoglossal, hypoglossal nerve, nervus hypoglosus, twelfth cranial nerve",
  "n05480794": "central nervous system, CNS, systema nervosum centrale",
  "n05481095": "brain, encephalon",
  "n05481549": "neencephalon, neoencephalon",
  "n05481746": "neopallium, neocortex",
  "n05481870": "archipallium, paleocortex",
  "n05481997": "metencephalon",
  "n05482129": "paleencephalon, paleoencephalon, palaeencephalon",
  "n05482328": "leptomeninges",
  "n05482529": "dura mater, dura",
  "n05482676": "arachnoid, arachnoid membrane",
  "n05482800": "pia mater",
  "n05482922": "subarachnoid space",
  "n05483122": "neuropil, neuropile",
  "n05483388": "grey matter, gray matter, grey substance, gray substance, substantia grisea",
  "n05483677": "white matter, substantia alba",
  "n05483890": "pituitary, pituitary gland, pituitary body, hypophysis",
  "n05484198": "hypophyseal stalk",
  "n05484355": "anterior pituitary, anterior pituitary gland, adenohypophysis",
  "n05484573": "pars distilis, pars anterior",
  "n05484711": "pars intermedia",
  "n05484862": "posterior pituitary, posterior pituitary gland, neurohypophysis, pars nervosa",
  "n05485098": "pineal gland, pineal body, epiphysis cerebri, epiphysis",
  "n05485314": "islands of Langerhans, isles of Langerhans, islets of Langerhans",
  "n05485554": "cerebellum",
  "n05485856": "cerebellar hemisphere",
  "n05485988": "dentate nucleus",
  "n05486159": "vermis, vermis cerebelli",
  "n05486319": "paleocerebellum",
  "n05486510": "cerebral cortex, cerebral mantle, pallium, cortex",
  "n05486920": "cortical area, cortical region",
  "n05487224": "association area, association cortex",
  "n05487423": "geniculate body",
  "n05487694": "lateral geniculate body, corpus geniculatum laterale, lateral geniculate",
  "n05487941": "medial geniculate body, corpus geniculatum mediale, medial geniculate",
  "n05488223": "auditory area, auditory cortex",
  "n05488385": "Broca's area, Broca's center, Broca's gyrus, Broca's convolution, convolution of Broca",
  "n05488615": "Brodmann's area",
  "n05488750": "frontal gyrus",
  "n05488909": "temporal gyrus",
  "n05489070": "parietal gyrus",
  "n05489231": "occipital gyrus",
  "n05489394": "language area, language zone",
  "n05489640": "motor area, motor region, motor cortex, Rolando's area, excitable area",
  "n05489810": "sensorium",
  "n05489998": "sensorimotor area, sensorimotor region",
  "n05490204": "visual area, visual cortex",
  "n05490370": "Wernicke's area, Wernicke's center",
  "n05490578": "cortex",
  "n05490799": "medulla",
  "n05490983": "adrenal cortex",
  "n05491154": "renal cortex",
  "n05491308": "adrenal medulla",
  "n05491461": "corpus callosum",
  "n05491612": "pyramidal tract, pyramidal motor system, corticospinal tract",
  "n05491993": "cerebrum",
  "n05492259": "fold, plica",
  "n05492426": "gyrus, convolution",
  "n05492655": "central gyrus",
  "n05492806": "precentral gyrus",
  "n05493002": "precordium",
  "n05493141": "postcentral gyrus",
  "n05493303": "lobe",
  "n05493650": "lobule",
  "n05493758": "frontal lobe, frontal cortex",
  "n05493992": "prefrontal lobe, prefrontal cortex",
  "n05494130": "parietal lobe, parietal cortex",
  "n05494365": "occipital lobe, occipital cortex",
  "n05494617": "striate cortex, striate area, first visual area, area 17 of Brodmann, Brodmann's area 17",
  "n05494933": "temporal lobe, temporal ccortex",
  "n05495172": "medulla oblongata, medulla, bulb",
  "n05495571": "amygdala, amygdaloid nucleus, corpus amygdaloideum",
  "n05495981": "forebrain, prosencephalon",
  "n05496261": "hippocampus",
  "n05496592": "cingulate gyrus, gyrus cinguli",
  "n05496802": "telencephalon",
  "n05496990": "diencephalon, interbrain, betweenbrain, thalmencephalon",
  "n05497363": "basal ganglion",
  "n05497741": "caudate nucleus, caudate",
  "n05497922": "claustrum",
  "n05498048": "lenticular nucleus, lentiform nucleus",
  "n05498300": "pallidum, globus pallidus, paleostriatum",
  "n05498460": "putamen",
  "n05498581": "subthalamic nucleus",
  "n05498773": "limbic system, visceral brain, limbic brain",
  "n05499044": "subthalamus",
  "n05499172": "thalamus",
  "n05499379": "hypothalamus",
  "n05499542": "corpus striatum, striatum, striate body",
  "n05499828": "midbrain, mesencephalon",
  "n05500006": "substantia nigra, nucleus niger, locus niger",
  "n05500312": "superior colliculus",
  "n05500465": "inferior colliculus",
  "n05500594": "hindbrain, rhombencephalon",
  "n05500812": "myelencephalon",
  "n05500992": "pons, pons Varolii",
  "n05501185": "brainstem, brain-stem, brain stem",
  "n05501485": "reticulum",
  "n05501711": "neural network, neural net",
  "n05501932": "nucleus",
  "n05502090": "reticular formation, RF",
  "n05502375": "reticular activating system, RAS",
  "n05502556": "ventricle",
  "n05502855": "fourth ventricle",
  "n05503009": "third ventricle",
  "n05503196": "lateral ventricle",
  "n05503401": "cerebral aqueduct, Sylvian aqueduct, aqueductus cerebri",
  "n05503555": "radiation",
  "n05503705": "spinal cord, medulla spinalis",
  "n05504107": "spinal fluid, cerebrospinal fluid",
  "n05504336": "peripheral nervous system, systema nervosum periphericum",
  "n05504532": "autonomic nervous system, ANS",
  "n05504807": "radial nerve, nervus radialis, musculospiral nerve",
  "n05505131": "sympathetic nervous system",
  "n05505479": "splanchnic nerve",
  "n05505679": "parasympathetic nervous system, parasympathetic",
  "n05506034": "brachial plexus, plexus brachialis",
  "n05506229": "cardiac plexus, plexus cardiacus",
  "n05506369": "carotid plexus, plexus caroticus",
  "n05506507": "cervical plexus, plexus cervicalis",
  "n05506648": "choroid plexus, plexus choroideus",
  "n05506832": "coccygeal plexus, plexus coccygeus",
  "n05506976": "hypogastric plexus, plexus hypogastricus",
  "n05507109": "lumbar plexus, plexus lumbalis",
  "n05507268": "lumbar plexus, plexus lumbalis",
  "n05507427": "lumbosacral plexus",
  "n05507642": "mesenteric plexus, plexus mesentericus",
  "n05507756": "myenteric plexus, plexus myentericus",
  "n05507977": "periarterial plexus, plexus periarterialis",
  "n05508113": "plexus dentalis",
  "n05508211": "pterygoid plexus",
  "n05508403": "pulmonary plexis, plexus pulmonalis",
  "n05508534": "sacral plexus, plexus sacralis",
  "n05508735": "solar plexus, coeliac plexus, plexus celiacus, abdominal nerve plexus",
  "n05508943": "pit of the stomach, epigastric fossa",
  "n05509146": "reproductive system, genital system",
  "n05509452": "urogenital system, urogenital apparatus, urinary system, urinary apparatus, genitourinary system, genitourinary apparatus, systema urogenitale, apparatus urogenitalis",
  "n05509889": "respiratory system, systema respiratorium",
  "n05510173": "respiratory tract, airway",
  "n05510358": "lower respiratory tract",
  "n05510506": "upper respiratory tract",
  "n05510702": "sensory system",
  "n05510907": "tract",
  "n05511061": "urinary tract",
  "n05511286": "vascular system",
  "n05511618": "circulatory system, cardiovascular system",
  "n05511975": "fetal circulation, foetal circulation",
  "n05512139": "bladder, vesica",
  "n05512337": "urinary bladder",
  "n05512505": "introitus",
  "n05512670": "urethral orifice, external orifice",
  "n05512835": "ureter",
  "n05513020": "urethra",
  "n05513302": "reproductive organ, sex organ",
  "n05513529": "female reproductive system",
  "n05513807": "male reproductive system",
  "n05514081": "genitalia, genital organ, genitals, private parts, privates, crotch",
  "n05514272": "pudendum",
  "n05514410": "female genitalia, female genitals, female genital organ, fanny",
  "n05514717": "female internal reproductive organ",
  "n05514905": "male genitalia, male genitals, male genital organ, family jewels",
  "n05515157": "male internal reproductive organ",
  "n05515287": "ovary",
  "n05515545": "ovotestis",
  "n05515670": "sac",
  "n05516067": "target organ",
  "n05516230": "taret organ",
  "n05516366": "acinus",
  "n05516554": "bursa",
  "n05516711": "cisterna, cistern",
  "n05516848": "pouch, pocket",
  "n05517145": "cheek pouch",
  "n05517268": "marsupium",
  "n05517406": "scrotum",
  "n05517578": "vesicle, cyst",
  "n05517837": "blister, bulla, bleb",
  "n05518094": "follicle",
  "n05518257": "hair follicle",
  "n05518449": "Graafian follicle",
  "n05518614": "corpus luteum",
  "n05518870": "Fallopian tube, uterine tube, oviduct",
  "n05519085": "uterus, womb",
  "n05519401": "uterine cavity",
  "n05519559": "cervical canal, canalis cervicis uteri",
  "n05519707": "decidua",
  "n05519820": "endometrium",
  "n05520168": "myometrium",
  "n05520292": "liposome",
  "n05520479": "umbilical cord, umbilical",
  "n05520699": "placenta",
  "n05520965": "afterbirth",
  "n05521111": "vagina",
  "n05521514": "cunt, puss, pussy, slit, snatch, twat",
  "n05521636": "vulva",
  "n05521934": "hymen, maidenhead, virginal membrane",
  "n05522148": "imperforate hymen",
  "n05522283": "mons, mons veneris, mons pubis",
  "n05522456": "labium",
  "n05522674": "labia majora",
  "n05522784": "pudendal cleft, urogenital cleft, rima pudendi, rima vulvae, pudendal cleavage, pudendal slit, vulvar slit",
  "n05522998": "labia minora",
  "n05523108": "vestibule of the vagina",
  "n05523269": "erectile organ",
  "n05523420": "clitoris, clit, button",
  "n05523629": "Cowper's gland, bulbourethral gland",
  "n05523859": "Bartholin's gland",
  "n05524062": "cervical glands, cervical glands of the uterus, glandulae cervicales uteri",
  "n05524243": "seminiferous tubule",
  "n05524430": "gonad, sex gland",
  "n05524615": "testis, testicle, orchis, ball, ballock, bollock, nut, egg",
  "n05525100": "cobblers",
  "n05525252": "male reproductive gland",
  "n05525391": "undescended testis, undescended testicle",
  "n05525628": "epididymis",
  "n05525807": "rete testis",
  "n05525970": "vasa efferentia",
  "n05526175": "vas deferens, ductus deferens",
  "n05526384": "penis, phallus, member",
  "n05526713": "cock, prick, dick, shaft, pecker, peter, tool, putz",
  "n05526860": "micropenis, microphallus",
  "n05526957": "prepuce, foreskin",
  "n05527085": "prepuce, foreskin",
  "n05527216": "seminal duct",
  "n05527389": "ejaculatory duct",
  "n05527597": "seminal vesicle",
  "n05527848": "spermatic cord",
  "n05528060": "respiratory organ",
  "n05528245": "book lung",
  "n05528395": "alveolus, air sac, air cell",
  "n05528604": "nasal cavity",
  "n05528854": "nasopharynx",
  "n05529012": "oropharynx",
  "n05529159": "laryngopharynx",
  "n05529286": "pharyngeal tonsil, adenoid, Luschka's tonsil, third tonsil, tonsilla pharyngealis, tonsilla adenoidea",
  "n05529729": "larynx, voice box",
  "n05530092": "arytenoid, arytaenoid, arytenoid cartilage",
  "n05530296": "thyroid cartilage, Adam's apple",
  "n05530429": "vocal cord, vocal fold, vocal band, plica vocalis",
  "n05530657": "false vocal cord, false vocal fold, superior vocal cord, ventricular fold, vestibular fold",
  "n05530871": "true vocal cord, true vocal fold, inferior vocal cord, inferior vocal fold",
  "n05531161": "cartilaginous structure",
  "n05531379": "cartilaginous tube",
  "n05531511": "bronchus, bronchial tube",
  "n05531666": "bronchiole",
  "n05531814": "trachea, windpipe",
  "n05532050": "trachea",
  "n05532225": "alimentary canal, alimentary tract, digestive tube, digestive tract, gastrointestinal tract, GI tract",
  "n05532641": "enteron",
  "n05532795": "digestive gland",
  "n05532944": "salivary gland",
  "n05533212": "parotid gland",
  "n05533392": "sublingual gland, sublingual salivary gland",
  "n05533635": "submaxillary gland, submaxillary salivary gland, submandibular gland, submandibular salivary gland, mandibular gland",
  "n05533948": "esophagus, oesophagus, gorge, gullet",
  "n05534174": "epicardia",
  "n05534333": "intestine, bowel, gut",
  "n05534578": "hindgut",
  "n05534712": "small intestine",
  "n05534955": "duodenum",
  "n05535095": "pylorus",
  "n05535247": "jejunum",
  "n05535367": "ileum",
  "n05535484": "large intestine",
  "n05535869": "colon",
  "n05536178": "megacolon",
  "n05536370": "cecum, caecum, blind gut",
  "n05536638": "ileocecal valve",
  "n05536858": "transverse colon",
  "n05537060": "ascending colon",
  "n05537243": "descending colon",
  "n05537417": "sigmoid colon, sigmoid flexure",
  "n05537576": "appendix, vermiform appendix, vermiform process, cecal appendage",
  "n05537806": "rectum",
  "n05538016": "anus",
  "n05538215": "arse, arsehole, asshole, bunghole",
  "n05538338": "imperforate anus",
  "n05538494": "perineum",
  "n05538625": "head, caput",
  "n05539012": "poll",
  "n05539138": "human head",
  "n05539370": "bullethead",
  "n05539454": "attic, bean, bonce, noodle, noggin, dome",
  "n05539595": "pate, poll, crown",
  "n05539717": "tonsure",
  "n05539834": "epicranium",
  "n05539947": "scalp",
  "n05540121": "skull",
  "n05540407": "calvaria, skullcap",
  "n05540513": "cranium, braincase, brainpan",
  "n05540976": "occiput",
  "n05541097": "sinciput",
  "n05541231": "frontal bone, os frontale, forehead",
  "n05541509": "frontal eminence",
  "n05541645": "parietal bone",
  "n05541872": "occipital bone",
  "n05542052": "occipital protuberance",
  "n05542193": "mastoid, mastoid process, mastoid bone, mastoidal",
  "n05542416": "styloid process",
  "n05542539": "pterygoid process",
  "n05542686": "tuberosity, tubercle, eminence",
  "n05542893": "suture, sutura, fibrous joint",
  "n05543177": "synovial joint, articulatio synovialis, diarthrosis",
  "n05543425": "anterior fontanelle",
  "n05543541": "sphenoid fontanelle, sphenoid fontanel, sphenoidal fontanelle, sphenoidal fontanel",
  "n05543917": "coronal suture, sutura coronalis",
  "n05544078": "frontal suture, sutura frontalis",
  "n05544264": "intermaxillary suture, sutura intermaxillaris",
  "n05544432": "internasal suture, sutura internasalis",
  "n05544575": "lamboid suture, sutura lamboidea",
  "n05544725": "occipitomastoid suture",
  "n05544906": "parietomastoid suture",
  "n05545047": "sagittal suture, interparietal suture, sutura sagittalis",
  "n05545212": "fontanelle, fontanel, soft spot",
  "n05545439": "foramen, hiatus",
  "n05545611": "interventricular foramen, foramen of Monro, Monro's foramen",
  "n05545879": "foramen magnum",
  "n05546040": "jaw",
  "n05546298": "chop",
  "n05546383": "zygomatic process",
  "n05546540": "neck, cervix",
  "n05546997": "frill, ruff",
  "n05547149": "frill",
  "n05547313": "bull neck",
  "n05547396": "nape, scruff, nucha",
  "n05547508": "throat, pharynx",
  "n05547904": "fauces",
  "n05548032": "fistula, sinus",
  "n05548203": "bypass",
  "n05548314": "portacaval shunt",
  "n05548521": "shunt",
  "n05548726": "tubular cavity",
  "n05548840": "shoulder",
  "n05549061": "shoulder, shoulder joint, articulatio humeri",
  "n05549350": "deltoid, deltoid muscle, musculus deltoideus",
  "n05549576": "armpit, axilla, axillary cavity, axillary fossa",
  "n05549830": "torso, trunk, body",
  "n05550330": "serratus, serratus muscles",
  "n05550494": "anterior serratus muscle, serratus anterior, musculus serratus anterior, serratus magnus",
  "n05550688": "posterior serratus muscle, serratus posterior, musculus serratus posterior",
  "n05550908": "serratus posterior inferior",
  "n05551142": "serratus posterior superior",
  "n05551318": "side",
  "n05551494": "female chest, bust",
  "n05551617": "male chest",
  "n05551711": "pectoral, pectoral muscle, pectoralis, musculus pectoralis, pecs",
  "n05551939": "pectoralis major, musculus pectoralis major, greater pectoral muscle",
  "n05552106": "pectoralis minor, musculus pectoralis minor, smaller pectoral muscle",
  "n05552287": "intercostal, intercostal muscle, musculus intercostalis",
  "n05552467": "depressor, depressor muscle",
  "n05552607": "thorax, chest, pectus",
  "n05553049": "chest cavity, thoracic cavity",
  "n05553288": "breast, chest",
  "n05553486": "bosom",
  "n05553618": "thorax",
  "n05553768": "rib cage",
  "n05553897": "cleavage",
  "n05554051": "lactiferous duct",
  "n05554189": "mammary gland, mamma",
  "n05554405": "breast, bosom, knocker, boob, tit, titty",
  "n05554653": "nipple, mammilla, mamilla, pap, teat, tit",
  "n05554804": "areola, ring of color",
  "n05555017": "areola",
  "n05555167": "nabothian gland",
  "n05555294": "vestibular gland",
  "n05555473": "middle, midriff, midsection",
  "n05555688": "waist, waistline",
  "n05555840": "wasp waist",
  "n05555917": "belly, paunch",
  "n05556071": "pot, potbelly, bay window, corporation, tummy",
  "n05556204": "spare tire, love handle",
  "n05556325": "hip",
  "n05556472": "haunch",
  "n05556595": "navel, umbilicus, bellybutton, belly button, omphalos, omphalus",
  "n05556943": "abdomen, venter, stomach, belly",
  "n05557339": "abdominal, abdominal muscle, ab",
  "n05557500": "dorsum",
  "n05557723": "underbelly, underbody",
  "n05557839": "external oblique muscle, musculus obliquus externus abdominis, abdominal external oblique muscle, oblique",
  "n05558078": "transversus abdominis muscle, transverse muscle of abdomen, musculus transversalis abdominis, transversus abdominis",
  "n05558345": "abdominal cavity, abdomen",
  "n05558555": "pubes, pubic region, loins",
  "n05558717": "back, dorsum",
  "n05559023": "small",
  "n05559122": "latissimus dorsi, lat",
  "n05559256": "buttocks, nates, arse, butt, backside, bum, buns, can, fundament, hindquarters, hind end, keister, posterior, prat, rear, rear end, rump, stern, seat, tail, tail end, tooshie, tush, bottom, behind, derriere, fanny, ass",
  "n05559727": "buttock, cheek",
  "n05559908": "extremity, appendage, member",
  "n05560244": "limb",
  "n05560631": "stump",
  "n05560787": "leg",
  "n05561390": "crus",
  "n05561507": "leg",
  "n05561707": "pin, peg, stick",
  "n05561834": "bowleg, bow leg, bandyleg, bandy leg, genu varum, tibia vara",
  "n05562015": "shank's mare, shanks' mare, shank's pony, shanks' pony",
  "n05562161": "spindlelegs, spindleshanks",
  "n05562249": "thigh",
  "n05562595": "lap",
  "n05562756": "shank",
  "n05562902": "shin",
  "n05563034": "vertebrate foot, pedal extremity",
  "n05563266": "foot, human foot, pes",
  "n05563770": "arm",
  "n05564229": "cubitus",
  "n05564323": "forearm",
  "n05564590": "hand, manus, mitt, paw",
  "n05565064": "fist, clenched fist",
  "n05565192": "hooks, meat hooks, maulers",
  "n05565337": "right, right hand",
  "n05565548": "left, left hand",
  "n05565696": "palm, thenar",
  "n05565937": "thenar",
  "n05566097": "digit, dactyl",
  "n05566366": "minimus",
  "n05566504": "finger",
  "n05566919": "extremity",
  "n05567117": "fingertip",
  "n05567217": "thumb, pollex",
  "n05567381": "index, index finger, forefinger",
  "n05567489": "ring finger, annualry",
  "n05567604": "middle finger",
  "n05567727": "little finger, pinkie, pinky",
  "n05567875": "sciatic nerve, nervus ischiadicus",
  "n05568104": "femoral nerve, nervus femoralis, anterior crural nerve",
  "n05568349": "saphenous nerve, nervus saphenus",
  "n05568550": "phrenic nerve, nervus phrenicus",
  "n05568767": "ulnar nerve, cubital nerve, nervus ulnaris",
  "n05569053": "spinal nerve, nervus spinalis",
  "n05569427": "cervical nerve",
  "n05569577": "coccygeal nerve, nervus coccygeus",
  "n05569690": "lumbar nerve",
  "n05569835": "sacral nerve",
  "n05569979": "thoracic nerve",
  "n05570129": "gluteus, gluteus muscle, gluteal muscle, glute",
  "n05570396": "gluteus maximus",
  "n05570500": "gluteus medius",
  "n05570600": "gluteus minimus",
  "n05570704": "hamstring, hamstring tendon",
  "n05570839": "sphincter, anatomical sphincter, sphincter muscle",
  "n05571132": "cardiac sphincter",
  "n05571341": "esophagogastric junction, oesophagogastric junction",
  "n05571501": "physiological sphincter",
  "n05571713": "anal sphincter, sphincter ani, musculus sphincter ani",
  "n05571904": "musculus sphincter ani externus",
  "n05572057": "musculus sphincter ani internus",
  "n05572227": "urethral sphincter, musculus sphincter urethrae",
  "n05572396": "bladder sphincter, musculus sphincter vesicae",
  "n05572629": "musculus sphincter ductus choledochi",
  "n05572782": "musculus sphincter ductus pancreatici",
  "n05572940": "pupillary sphincter, musculus sphincter pupillae",
  "n05573099": "pyloric sphincter, pyloric valve, musculus sphincter pylori",
  "n05573306": "tensor",
  "n05573474": "tensor tympani",
  "n05573602": "knee, knee joint, human knee, articulatio genus, genu",
  "n05573895": "femur, thighbone, femoris",
  "n05574151": "trochanter",
  "n05574332": "calf, sura",
  "n05574519": "mid-calf",
  "n05574633": "gastrocnemius, gastrocnemius muscle",
  "n05574862": "psoas",
  "n05575002": "rhomboid, rhomboid muscle",
  "n05575185": "rhomboideus major muscle, greater rhomboid muscle, musculus rhomboideus major",
  "n05575374": "rhomboid minor muscle, lesser rhomboid muscle, musculus rhomboideus minor",
  "n05575582": "soleus, soleus muscle",
  "n05575743": "splenius, splenius muscle",
  "n05575960": "peroneus",
  "n05576074": "pterygoid muscle",
  "n05576194": "ball",
  "n05576421": "flatfoot, splayfoot, pes planus",
  "n05576573": "arch",
  "n05576827": "metatarsal arch",
  "n05576950": "instep",
  "n05577060": "sunken arch, fallen arch",
  "n05577190": "sole",
  "n05577304": "tiptoe",
  "n05577410": "toe",
  "n05577654": "toe",
  "n05577741": "big toe, great toe, hallux",
  "n05577866": "hammertoe",
  "n05577969": "little toe",
  "n05578095": "heel",
  "n05578251": "gliding joint, articulatio plana",
  "n05578442": "ankle, ankle joint, mortise joint, articulatio talocruralis",
  "n05578740": "Achilles tendon, tendon of Achilles",
  "n05578911": "girdle",
  "n05579053": "musculus biceps femoris, femoral biceps",
  "n05579239": "biceps",
  "n05579436": "biceps brachii, musculus biceps brachii, biceps humeri",
  "n05579604": "triceps",
  "n05579753": "triceps brachii, musculus triceps brachii",
  "n05579944": "elbow, elbow joint, human elbow, cubitus, cubital joint, articulatio cubiti",
  "n05580286": "interphalangeal joint",
  "n05580416": "hinge joint, ginglymus, ginglymoid joint",
  "n05580662": "funny bone, crazy bone",
  "n05580929": "lamina",
  "n05581161": "lamina arcus vertebrae",
  "n05581349": "plate",
  "n05581514": "horny structure, unguis",
  "n05581693": "nail",
  "n05581932": "cuticle",
  "n05582038": "half-moon, lunula, lunule",
  "n05582191": "matrix",
  "n05582305": "matrix, intercellular substance, ground substance",
  "n05582513": "fascia, facia",
  "n05582690": "aponeurosis",
  "n05582859": "graft, transplant",
  "n05583158": "autograft, autoplasty",
  "n05583386": "homograft, allograft",
  "n05583624": "heterograft, xenograft",
  "n05583822": "scar tissue",
  "n05584008": "adhesion",
  "n05584152": "stroma",
  "n05584265": "fingernail",
  "n05584390": "thumbnail",
  "n05584486": "toenail",
  "n05584605": "ingrown toenail, onyxis",
  "n05584746": "hangnail, agnail",
  "n05584928": "wrist, carpus, wrist joint, radiocarpal joint, articulatio radiocarpea",
  "n05585205": "knuckle, knuckle joint, metacarpophalangeal joint",
  "n05585383": "skeletal system, skeleton, frame, systema skeletale",
  "n05585665": "skeletal structure",
  "n05585999": "column",
  "n05586111": "pectoral girdle",
  "n05586280": "shoulder girdle, pectoral arch",
  "n05586446": "endoskeleton",
  "n05586759": "exoskeleton",
  "n05587034": "appendicular skeleton",
  "n05587288": "axial skeleton",
  "n05587531": "axial muscle",
  "n05587628": "transverse process",
  "n05587814": "hemal arch, haemal arch",
  "n05587997": "neural arch, vertebral arch",
  "n05588174": "spinal column, vertebral column, spine, backbone, back, rachis",
  "n05588551": "cervical vertebra, neck bone",
  "n05588746": "atlas, atlas vertebra",
  "n05588840": "axis, axis vertebra",
  "n05588991": "odontoid process",
  "n05589132": "thoracic vertebra, dorsal vertebra",
  "n05589378": "lumbar vertebra",
  "n05589596": "sacral vertebra",
  "n05589756": "coccygeal vertebra, caudal vertebra",
  "n05589896": "sartorius, sartorius muscle, musculus sartorius",
  "n05590144": "scalenus, scalene muscle, musculus scalenus",
  "n05590366": "sternocleidomastoid, sternocleidomastoid muscle, sternocleido mastoideus, musculus sternocleidomastoideus",
  "n05590740": "teres, teres muscle",
  "n05590939": "teres major, teres major muscle, musculus teres major",
  "n05591096": "teres minor, teres minor muscle, musculus teres minor",
  "n05591256": "tibialis, tibialis muscle, musculus tibialis",
  "n05591503": "tibialis anticus, tibialis anterior",
  "n05591661": "tibialis posticus, tibialis posterior",
  "n05591770": "trapezius, trapezius muscle, cowl muscle, musculus trapezius",
  "n05591999": "true rib",
  "n05592126": "costa",
  "n05592302": "costal cartilage",
  "n05592504": "epiphysis",
  "n05592733": "diaphysis, shaft",
  "n05592870": "metaphysis",
  "n05593017": "arm bone",
  "n05593181": "humerus",
  "n05593318": "radius",
  "n05593476": "ulna, elbow bone",
  "n05593654": "olecranon, olecranon process",
  "n05593871": "metacarpus",
  "n05594037": "leg bone",
  "n05594201": "fibula, calf bone",
  "n05594367": "tibia, shinbone, shin bone, shin",
  "n05594568": "metatarsus",
  "n05594822": "tarsus",
  "n05595083": "joint, articulation, articulatio",
  "n05595531": "ball-and-socket joint, spheroid joint, cotyloid joint, enarthrodial joint, enarthrosis, articulatio spheroidea",
  "n05595837": "head",
  "n05596004": "hip, hip joint, coxa, articulatio coxae",
  "n05596224": "acetabulum, cotyloid cavity",
  "n05596442": "pelvis, renal pelvis",
  "n05596651": "pelvis, pelvic girdle, pelvic arch, hip",
  "n05597050": "pelvic cavity",
  "n05597188": "pivot joint, rotary joint, rotatory joint, articulatio trochoidea",
  "n05597436": "crotch, fork",
  "n05597594": "loins",
  "n05597734": "groin, inguen",
  "n05597980": "quick",
  "n05598147": "nose, olfactory organ",
  "n05598707": "beak, honker, hooter, nozzle, snoot, snout, schnozzle, schnoz",
  "n05598868": "conk",
  "n05598982": "hawk nose",
  "n05599084": "proboscis",
  "n05599203": "bridge",
  "n05599398": "pug nose",
  "n05599501": "Roman nose, hooknose",
  "n05599617": "chin, mentum",
  "n05599769": "double chin, buccula",
  "n05599874": "dimple",
  "n05600030": "lantern jaw",
  "n05600109": "nostril, anterior naris",
  "n05600270": "posterior naris",
  "n05600431": "naris",
  "n05600637": "face, human face",
  "n05601198": "face",
  "n05601357": "countenance, physiognomy, phiz, visage, kisser, smiler, mug",
  "n05601662": "pudding face, pudding-face",
  "n05601758": "feature, lineament",
  "n05602132": "facial muscle",
  "n05602304": "temporalis muscle, temporal muscle, temporalis, musculus temporalis",
  "n05602548": "brow, forehead",
  "n05602683": "temple",
  "n05602835": "cheek",
  "n05602982": "jowl",
  "n05603160": "jaw",
  "n05603342": "ridge",
  "n05603475": "supraorbital ridge, supraorbital torus, superciliary ridge, superciliary arch",
  "n05603650": "excrescence",
  "n05603854": "vegetation",
  "n05604022": "rudiment",
  "n05604254": "wall, paries",
  "n05604434": "abdominal wall",
  "n05604535": "humor, humour",
  "n05604854": "pericardial sac",
  "n05604950": "rotator cuff",
  "n05605192": "respiratory center",
  "n05605498": "spindle",
  "n05605762": "syncytium",
  "n05605944": "serous membrane, serosa",
  "n05606247": "synovial membrane, synovium",
  "n05606429": "tunica albuginea testes",
  "n05606528": "albuginea",
  "n05606633": "tunic, tunica, adventitia",
  "n05606801": "celom, coelom, celoma",
  "n05607001": "cornu",
  "n05607126": "corona",
  "n05607271": "ruga",
  "n05607402": "tentorium",
  "n05607602": "mast cell, mastocyte, labrocyte",
  "n05607863": "stem cell",
  "n05608044": "hematopoeitic stem cell",
  "n05608206": "target cell",
  "n05608419": "McBurney's point",
  "n05608615": "zona pellucida",
  "n05608868": "receptor",
  "n05609111": "alpha receptor, alpha-adrenergic receptor, alpha-adrenoceptor",
  "n05609524": "beta receptor, beta-adrenergic receptor, beta-adrenoceptor",
  "n05609884": "pharyngeal recess",
  "n05610008": "rima",
  "n05610198": "rima glottidis, rima vocalis, true glottis, glottis vera",
  "n05610365": "rima vestibuli, rima respiratoria, false glottis, glottis spuria",
  "n05610520": "telomere",
  "n05610734": "vomer",
  "n05610919": "Wormian bone, sutural bone",
  "n05611062": "zone, zona",
  "n05611221": "zonule, zonula",
  "n05611302": "mind, head, brain, psyche, nous",
  "n05611684": "noddle",
  "n05611822": "place",
  "n05612067": "public knowledge, general knowledge",
  "n05612241": "common knowledge",
  "n05612358": "episteme",
  "n05612531": "ancient history",
  "n05612706": "light",
  "n05612809": "open, surface",
  "n05613043": "tabula rasa",
  "n05613170": "ego",
  "n05613274": "unconscious mind, unconscious",
  "n05613478": "subconscious mind, subconscious",
  "n05613625": "superego",
  "n05613794": "id",
  "n05613962": "astuteness, profundity, profoundness, depth, deepness",
  "n05614175": "sagacity, sagaciousness, judgment, judgement, discernment",
  "n05614476": "eye",
  "n05614657": "common sense, good sense, gumption, horse sense, sense, mother wit",
  "n05615028": "logic",
  "n05615147": "nous",
  "n05615258": "road sense",
  "n05615373": "judiciousness",
  "n05615500": "discretion, discreetness, circumspection, prudence",
  "n05615749": "confidentiality",
  "n05615869": "caution, precaution, care, forethought",
  "n05616092": "injudiciousness, indiscreetness",
  "n05616246": "ability, power",
  "n05616786": "know-how",
  "n05616955": "bag of tricks",
  "n05617107": "wisdom, sapience",
  "n05617310": "leadership",
  "n05617467": "generalship",
  "n05617606": "intelligence",
  "n05618056": "brain, brainpower, learning ability, mental capacity, mentality, wit",
  "n05618293": "breadth, comprehensiveness, largeness",
  "n05618607": "capaciousness, roominess",
  "n05618849": "mind, intellect",
  "n05619059": "nonverbal intelligence",
  "n05619226": "verbal intelligence",
  "n05619345": "mental quickness, quickness, quick-wittedness",
  "n05619553": "nimbleness, mental dexterity",
  "n05619743": "brilliance, genius",
  "n05619903": "coruscation",
  "n05620050": "pyrotechnics",
  "n05620190": "scintillation",
  "n05620294": "precociousness, precocity",
  "n05620468": "acuteness, acuity, sharpness, keenness",
  "n05620754": "steel trap",
  "n05620955": "brightness, cleverness, smartness",
  "n05621178": "craft, craftiness, cunning, foxiness, guile, slyness, wiliness",
  "n05621439": "shrewdness, astuteness, perspicacity, perspicaciousness",
  "n05621808": "insightfulness, acumen",
  "n05621946": "knowingness",
  "n05622076": "street smarts",
  "n05622196": "wits, marbles",
  "n05622456": "aptitude",
  "n05622617": "bilingualism",
  "n05622723": "instinct, inherent aptitude",
  "n05622956": "capacity, mental ability",
  "n05623181": "capability, capableness, potentiality",
  "n05623444": "perfectibility",
  "n05623628": "compass, range, reach, grasp",
  "n05623818": "sight, ken",
  "n05623937": "natural ability",
  "n05624042": "endowment, gift, talent, natural endowment",
  "n05624254": "bent, knack, hang",
  "n05624461": "flair, genius",
  "n05624615": "raw talent",
  "n05624700": "creativity, creativeness, creative thinking",
  "n05625066": "fecundity, fruitfulness",
  "n05625210": "flight",
  "n05625373": "genius, wizardry",
  "n05625465": "imagination, imaginativeness, vision",
  "n05625879": "imaginary place, mythical place, fictitious place",
  "n05626524": "afterworld",
  "n05626618": "Annwfn, Annwn",
  "n05626767": "Asgard",
  "n05626929": "Atlantis",
  "n05627083": "Brobdingnag",
  "n05627222": "cloud-cuckoo-land",
  "n05627385": "Cockaigne",
  "n05627514": "El Dorado, eldorado",
  "n05627682": "fairyland, faerie, faery",
  "n05627785": "Heaven",
  "n05628031": "Abraham's bosom, bosom of Abraham",
  "n05628193": "Celestial City, City of God, Heavenly City, Holy City",
  "n05628403": "Elysium, Elysian Fields",
  "n05628565": "Elysium",
  "n05628658": "Eden, Garden of Eden",
  "n05628939": "Paradise",
  "n05629104": "Promised Land",
  "n05629204": "Valhalla, Walhalla",
  "n05629381": "Hell, Hades, infernal region, netherworld, Scheol, underworld",
  "n05629682": "Hell, perdition, Inferno, infernal region, nether region, pit",
  "n05630145": "Houyhnhnms",
  "n05630277": "Gehenna, Tartarus",
  "n05630409": "hellfire, red region",
  "n05630539": "Laputa",
  "n05630734": "Lilliput",
  "n05630874": "limbo",
  "n05631076": "limbo",
  "n05631175": "Midgard",
  "n05631304": "never-never land, dreamland, dreamworld",
  "n05631449": "purgatory",
  "n05631681": "Ruritania",
  "n05631841": "spirit world",
  "n05632056": "Utopia, Zion, Sion",
  "n05632175": "wonderland",
  "n05632272": "fancy",
  "n05632446": "fantasy, phantasy",
  "n05632732": "pipe dream, dream",
  "n05632927": "fantasy life, phantasy life",
  "n05633044": "fantasy world, phantasy world, fairyland",
  "n05633228": "paracosm",
  "n05633385": "invention, innovation, excogitation, conception, design",
  "n05633672": "inventiveness, ingeniousness, ingenuity, cleverness",
  "n05633860": "resource, resourcefulness, imagination",
  "n05634059": "armory, armoury, inventory",
  "n05634219": "concoction",
  "n05634457": "contrivance",
  "n05634613": "originality",
  "n05634767": "innovativeness",
  "n05634910": "unconventionality",
  "n05635055": "novelty, freshness",
  "n05635188": "aviation, airmanship",
  "n05635321": "eristic",
  "n05635448": "falconry",
  "n05635624": "fortification",
  "n05635764": "homiletics",
  "n05635841": "horology",
  "n05635970": "minstrelsy",
  "n05636048": "musicianship",
  "n05636171": "enology, oenology",
  "n05636294": "puppetry",
  "n05636402": "taxidermy",
  "n05636554": "telescopy",
  "n05636666": "ventriloquism, ventriloquy",
  "n05636887": "skill, science",
  "n05637106": "nose",
  "n05637222": "virtuosity",
  "n05637356": "bravura",
  "n05637558": "skill, accomplishment, acquirement, acquisition, attainment",
  "n05637965": "hand",
  "n05638063": "craft, craftsmanship, workmanship",
  "n05638374": "horsemanship",
  "n05638486": "literacy",
  "n05638606": "marksmanship",
  "n05638700": "mastership",
  "n05638778": "mixology",
  "n05638882": "superior skill",
  "n05638987": "art, artistry, prowess",
  "n05639431": "numeracy",
  "n05639556": "oarsmanship",
  "n05639651": "salesmanship",
  "n05639832": "seamanship",
  "n05639941": "boatmanship",
  "n05640040": "showmanship",
  "n05640184": "soldiering, soldiership",
  "n05640339": "swordsmanship",
  "n05640433": "skillfulness",
  "n05640729": "expertness, expertise",
  "n05640924": "handiness",
  "n05641089": "professionalism",
  "n05641205": "sophistication",
  "n05641432": "coordination",
  "n05641556": "coordination",
  "n05641720": "incoordination",
  "n05641834": "versatility",
  "n05641959": "command, control, mastery",
  "n05642175": "adeptness, adroitness, deftness, facility, quickness",
  "n05642553": "touch",
  "n05642678": "finishing touch, capstone, copestone",
  "n05642815": "dexterity, manual dexterity, sleight",
  "n05642947": "fluency",
  "n05643072": "disfluency",
  "n05643190": "proficiency, technique",
  "n05643491": "brushwork",
  "n05643611": "musketry",
  "n05643722": "housecraft",
  "n05643807": "priestcraft",
  "n05643908": "stagecraft",
  "n05644016": "tradecraft",
  "n05644334": "watercraft",
  "n05644423": "woodcraft",
  "n05644527": "efficiency",
  "n05644727": "economy",
  "n05644922": "inability",
  "n05645199": "block, mental block",
  "n05645465": "writer's block",
  "n05645597": "stupidity",
  "n05645854": "denseness, dumbness, slow-wittedness",
  "n05646039": "dullness, obtuseness",
  "n05646218": "retardation, mental retardation, backwardness, slowness, subnormality",
  "n05646535": "abnormality, mental defectiveness",
  "n05646723": "feeblemindedness",
  "n05646828": "moronity, mental deficiency",
  "n05646926": "idiocy, amentia",
  "n05647015": "imbecility",
  "n05647156": "folly, foolishness, craziness, madness",
  "n05647451": "vacuousness",
  "n05647643": "inaptitude",
  "n05647772": "talentlessness",
  "n05647867": "incapability, incapableness",
  "n05648011": "imperfectibility",
  "n05648146": "incapacity",
  "n05648247": "unskillfulness",
  "n05648459": "awkwardness, clumsiness, ineptness, ineptitude, maladroitness, slowness",
  "n05648756": "rustiness",
  "n05648953": "inefficiency",
  "n05649116": "amateurishness",
  "n05649256": "illiteracy, analphabetism",
  "n05649385": "uncreativeness",
  "n05649538": "fruitlessness",
  "n05649628": "unoriginality",
  "n05649785": "triteness, staleness",
  "n05649960": "camp",
  "n05650180": "conventionality",
  "n05650329": "faculty, mental faculty, module",
  "n05650579": "attention",
  "n05650820": "language, speech",
  "n05651068": "lexis",
  "n05651242": "vocabulary, lexicon, mental lexicon",
  "n05651399": "memory, retention, retentiveness, retentivity",
  "n05651680": "reason, understanding, intellect",
  "n05651971": "sense, sensation, sentience, sentiency, sensory faculty",
  "n05652396": "modality, sense modality, sensory system",
  "n05652593": "volition, will",
  "n05652843": "velleity",
  "n05652926": "sensitivity, sensitiveness, sensibility",
  "n05653327": "acuteness",
  "n05653475": "hypersensitivity",
  "n05653575": "responsiveness, reactivity",
  "n05653710": "excitability, irritability",
  "n05653848": "exteroception",
  "n05654052": "interoception",
  "n05654201": "photosensitivity, radiosensitivity",
  "n05654362": "sight, vision, visual sense, visual modality",
  "n05654783": "stigmatism",
  "n05654873": "somatosense",
  "n05655119": "touch, sense of touch, skin senses, touch modality, cutaneous senses",
  "n05655482": "achromatic vision",
  "n05655567": "acuity, visual acuity, sharp-sightedness",
  "n05655810": "twenty-twenty, 20/20",
  "n05655966": "oxyopia",
  "n05656042": "binocular vision",
  "n05656160": "central vision",
  "n05656294": "color vision, chromatic vision, trichromacy",
  "n05656418": "distance vision",
  "n05656537": "eyesight, seeing, sightedness",
  "n05656722": "foveal vision",
  "n05656803": "monocular vision",
  "n05656890": "near vision",
  "n05656997": "night vision, night-sight, scotopic vision, twilight vision",
  "n05657166": "daylight vision, photopic vision",
  "n05657373": "peripheral vision",
  "n05657515": "stereoscopic vision, stereoscopy",
  "n05657718": "hearing, audition, auditory sense, sense of hearing, auditory modality",
  "n05657999": "ear",
  "n05658106": "absolute pitch, perfect pitch",
  "n05658226": "taste, gustation, sense of taste, gustatory modality",
  "n05658603": "smell, sense of smell, olfaction, olfactory modality",
  "n05658826": "nose",
  "n05658985": "kinesthesis, kinaesthesis, kinesthesia, kinaesthesia, kinesthetics, muscle sense, sense of movement",
  "n05659260": "kinanesthesia",
  "n05659365": "equilibrium, labyrinthine sense, vestibular sense, sense of balance, sense of equilibrium",
  "n05659621": "proprioception",
  "n05659856": "somesthesia, somesthesis, somaesthesia, somaesthesis, somatesthesia, somataesthesis, somatosensory system, somatic sensory system, somatic sense",
  "n05660268": "method",
  "n05660631": "scientific method",
  "n05660801": "experimental method",
  "n05660937": "teaching method, pedagogics, pedagogy",
  "n05661118": "Socratic method, maieutic method",
  "n05661294": "method of choice",
  "n05661400": "methodology",
  "n05661551": "mnemonics",
  "n05661668": "solution",
  "n05661825": "silver bullet",
  "n05661996": "system, system of rules",
  "n05662532": "accounting",
  "n05662876": "discipline",
  "n05663104": "frame of reference, frame",
  "n05663275": "vocabulary",
  "n05663491": "gambling system",
  "n05663671": "government",
  "n05663878": "honor system",
  "n05664069": "logic, logical system, system of logic",
  "n05664344": "Aristotelian logic",
  "n05664487": "merit system",
  "n05664640": "point system",
  "n05664803": "spoils system",
  "n05664986": "organon",
  "n05665146": "technique",
  "n05665421": "antialiasing",
  "n05665625": "Benday process",
  "n05665769": "bonding",
  "n05665984": "emulation",
  "n05666176": "terminal emulation",
  "n05666324": "immunofluorescence",
  "n05666530": "photomechanics",
  "n05666700": "simulation, computer simulation",
  "n05667086": "technicolor",
  "n05667196": "practice",
  "n05667404": "custom, tradition",
  "n05667613": "convention, normal, pattern, rule, formula",
  "n05667951": "mores",
  "n05668095": "code of conduct, code of behavior",
  "n05668359": "universal",
  "n05668581": "courtly love",
  "n05668725": "knight errantry",
  "n05668915": "protocol",
  "n05669034": "habit, wont",
  "n05669181": "Hadith",
  "n05669350": "institution",
  "n05669618": "levirate",
  "n05669797": "heritage",
  "n05669934": "cognitive state, state of mind",
  "n05670343": "enthusiasm",
  "n05670512": "Anglomania",
  "n05670615": "balletomania",
  "n05670710": "concern",
  "n05670972": "worldly concern, earthly concern, world, earth",
  "n05671217": "interestedness",
  "n05671325": "matter, affair, thing",
  "n05671526": "least",
  "n05671676": "personal business, personal matters, affairs",
  "n05671842": "dirty linen, dirty laundry",
  "n05671974": "part",
  "n05672178": "point of honor",
  "n05672286": "cult of personality",
  "n05672391": "amnesia, memory loss, blackout",
  "n05672698": "paramnesia",
  "n05672862": "anterograde amnesia, posttraumatic amnesia",
  "n05673088": "retrograde amnesia",
  "n05673209": "forgetfulness",
  "n05673323": "senior moment",
  "n05673439": "selective amnesia",
  "n05673710": "posthypnotic amnesia",
  "n05673908": "forgetfulness",
  "n05674134": "obliviousness, oblivion",
  "n05674319": "transient global amnesia",
  "n05674584": "set, readiness",
  "n05674927": "ivory tower",
  "n05675130": "consciousness",
  "n05675437": "stream of consciousness",
  "n05675601": "self, ego",
  "n05675715": "anima",
  "n05675905": "awareness, consciousness, cognizance, cognisance, knowingness",
  "n05676476": "incognizance",
  "n05676605": "self-awareness",
  "n05676755": "orientation",
  "n05676911": "self-consciousness",
  "n05677097": "unselfconsciousness",
  "n05677340": "feel",
  "n05677504": "sense",
  "n05677744": "sense of direction",
  "n05677850": "sense of responsibility",
  "n05677952": "awareness, sentience",
  "n05678148": "sensibility, esthesia, aesthesia",
  "n05678300": "waking",
  "n05678474": "wakefulness",
  "n05678745": "arousal",
  "n05678855": "vigil",
  "n05678932": "unconsciousness",
  "n05679305": "automatic pilot, autopilot",
  "n05679611": "unknowingness, unawareness",
  "n05679800": "blackout",
  "n05679906": "grogginess, stupor, stupefaction, semiconsciousness",
  "n05680193": "coma, comatoseness",
  "n05680423": "diabetic coma, Kussmaul's coma",
  "n05680573": "hepatic coma",
  "n05680684": "electrosleep",
  "n05680839": "semicoma",
  "n05680982": "insensibility",
  "n05681117": "sleeping",
  "n05681252": "hebetude",
  "n05681334": "trance",
  "n05681543": "semitrance",
  "n05681701": "hypnotic trance",
  "n05681855": "religious trance, ecstatic state",
  "n05682055": "narcosis",
  "n05682196": "nitrogen narcosis",
  "n05682431": "subconsciousness",
  "n05682570": "curiosity, wonder",
  "n05682798": "desire to know, lust for learning, thirst for knowledge",
  "n05682950": "interest, involvement",
  "n05683197": "curiousness, inquisitiveness",
  "n05683390": "nosiness, prying, snoopiness",
  "n05683582": "confusion, mental confusion, confusedness, muddiness, disarray",
  "n05684003": "disorientation",
  "n05684249": "culture shock",
  "n05684440": "distraction",
  "n05684561": "daze, fog, haze",
  "n05684739": "half-cock",
  "n05684839": "jamais vu",
  "n05685030": "bewilderment, obfuscation, puzzlement, befuddlement, mystification, bafflement, bemusement",
  "n05685363": "perplexity",
  "n05685538": "mystery, enigma, secret, closed book",
  "n05685879": "tangle, snarl, maze",
  "n05686086": "dilemma, quandary",
  "n05686272": "double bind",
  "n05686481": "cognitive factor",
  "n05686690": "divine guidance, inspiration",
  "n05686955": "difficulty",
  "n05687338": "trouble, problem",
  "n05687654": "pressure point",
  "n05687832": "can of worms",
  "n05687958": "deep water",
  "n05688030": "growing pains",
  "n05688169": "hydra",
  "n05688397": "matter",
  "n05688486": "facer",
  "n05688630": "killer",
  "n05688770": "kink",
  "n05688907": "pisser",
  "n05688990": "pitfall, booby trap",
  "n05689109": "snorter",
  "n05689249": "hindrance, hinderance, deterrent, impediment, balk, baulk, check, handicap",
  "n05689645": "albatross, millstone",
  "n05689801": "bind",
  "n05689909": "diriment impediment",
  "n05690091": "drag",
  "n05690269": "obstacle, obstruction",
  "n05690684": "straitjacket",
  "n05690916": "barrier, roadblock",
  "n05691144": "hang-up, hitch, rub, snag",
  "n05691241": "hurdle",
  "n05691376": "stymie, stymy",
  "n05691492": "ideological barrier",
  "n05691689": "iron curtain",
  "n05691963": "language barrier",
  "n05692094": "bamboo curtain",
  "n05692234": "color bar, colour bar, color line, colour line, Jim Crow",
  "n05692419": "determinant, determiner, determinative, determining factor, causal factor",
  "n05692758": "clincher, decisive factor",
  "n05692910": "influence",
  "n05693206": "imponderable",
  "n05693383": "imprint",
  "n05693537": "morale builder",
  "n05693669": "pestilence, canker",
  "n05693919": "support",
  "n05694232": "anchor, mainstay, keystone, backbone, linchpin, lynchpin",
  "n05694527": "lifeline",
  "n05694791": "temptation, enticement",
  "n05695002": "forbidden fruit",
  "n05695232": "bait, come-on, hook, lure, sweetener",
  "n05695446": "allurement",
  "n05695554": "equivalent",
  "n05695806": "counterpart, opposite number, vis-a-vis",
  "n05696020": "match",
  "n05696199": "mismatch",
  "n05696297": "complement",
  "n05696425": "substitute, replacement",
  "n05696701": "ersatz",
  "n05696803": "successor",
  "n05696941": "succedaneum",
  "n05697135": "certainty",
  "n05697363": "assurance, self-assurance, confidence, self-confidence, authority, sureness",
  "n05697789": "certitude, cocksureness, overconfidence",
  "n05697976": "reliance, trust",
  "n05698247": "doubt, uncertainty, incertitude, dubiety, doubtfulness, dubiousness",
  "n05698620": "mental reservation, reservation, arriere pensee",
  "n05698791": "misgiving, mistrust, distrust, suspicion",
  "n05698982": "incredulity, disbelief, skepticism, mental rejection",
  "n05699172": "indecision, indecisiveness, irresolution",
  "n05699434": "hesitation, vacillation, wavering",
  "n05699600": "peradventure",
  "n05699770": "suspense",
  "n05699906": "morbidity, morbidness",
  "n05700087": "preoccupation, preoccupancy, absorption, engrossment",
  "n05700401": "obsession, fixation",
  "n05700625": "abstractedness, abstraction",
  "n05700793": "reverie, revery",
  "n05700925": "dream",
  "n05701107": "brown study",
  "n05701209": "absentmindedness",
  "n05701363": "process, cognitive process, mental process, operation, cognitive operation",
  "n05701738": "process, unconscious process",
  "n05701944": "basic cognitive process",
  "n05702275": "attention, attending",
  "n05702602": "attention span",
  "n05702726": "attentiveness, heed, regard, paying attention",
  "n05703070": "clock-watching",
  "n05703205": "ear",
  "n05703307": "eye",
  "n05703429": "notice, observation, observance",
  "n05703803": "notice",
  "n05703956": "mind",
  "n05704096": "advertence, advertency",
  "n05704266": "concentration, engrossment, absorption, immersion",
  "n05704550": "mental note",
  "n05704694": "focus, focusing, focussing, focal point, direction, centering",
  "n05705075": "particularism",
  "n05705184": "specialism",
  "n05705355": "study",
  "n05705484": "hang-up",
  "n05705583": "hobbyhorse",
  "n05705722": "watchfulness, wakefulness, vigilance, alertness",
  "n05706076": "jealousy",
  "n05706228": "inattention",
  "n05706375": "inattentiveness, heedlessness",
  "n05706547": "distraction",
  "n05706629": "disregard, neglect",
  "n05706815": "oversight, inadvertence",
  "n05706954": "omission",
  "n05707146": "pretermission",
  "n05707269": "exception, exclusion, elision",
  "n05707495": "intuition",
  "n05707718": "feeling, intuitive feeling",
  "n05707885": "sprachgefuhl",
  "n05708030": "gnosis",
  "n05708207": "insight, sixth sense",
  "n05708322": "immediacy, immediate apprehension",
  "n05708432": "perception",
  "n05708658": "apperception",
  "n05708818": "constancy, perceptual constancy",
  "n05709148": "brightness constancy",
  "n05709343": "color constancy, colour constancy",
  "n05709507": "shape constancy",
  "n05709766": "size constancy",
  "n05710020": "perception",
  "n05710210": "discernment, perceptiveness",
  "n05710356": "penetration, insight",
  "n05710481": "cognizance",
  "n05710573": "remark",
  "n05710687": "detection, sensing",
  "n05710860": "visual perception, beholding, seeing",
  "n05711084": "contrast",
  "n05711206": "face recognition",
  "n05711308": "object recognition",
  "n05711414": "visual space",
  "n05711503": "auditory perception, sound perception",
  "n05711674": "speech perception",
  "n05711791": "musical perception",
  "n05711915": "melody, tonal pattern",
  "n05712076": "sensation, esthesis, aesthesis, sense experience, sense impression, sense datum",
  "n05712426": "threshold, limen",
  "n05712559": "absolute threshold",
  "n05712698": "pain threshold",
  "n05712892": "difference threshold, differential threshold, difference limen, differential limen",
  "n05713101": "just-noticeable difference, jnd",
  "n05713347": "masking",
  "n05713524": "vision, visual sensation",
  "n05713737": "smell, odor, odour, olfactory sensation, olfactory perception",
  "n05714161": "scent",
  "n05714332": "musk",
  "n05714466": "aroma, fragrance, perfume, scent",
  "n05714745": "incense",
  "n05714894": "malodor, malodour, stench, stink, reek, fetor, foetor, mephitis",
  "n05715150": "niff, pong",
  "n05715283": "taste, taste sensation, gustatory sensation, taste perception, gustatory perception",
  "n05715864": "relish, flavor, flavour, sapidity, savor, savour, smack, nip, tang",
  "n05716342": "lemon",
  "n05716462": "vanilla",
  "n05716577": "sweet, sweetness, sugariness",
  "n05716744": "sour, sourness, tartness",
  "n05716961": "acidity, acidulousness",
  "n05717115": "bitter, bitterness",
  "n05717266": "acridity",
  "n05717342": "salt, saltiness, salinity",
  "n05717549": "astringency, astringence",
  "n05717747": "finish",
  "n05717953": "flatness",
  "n05718118": "mellowness",
  "n05718254": "sound, auditory sensation",
  "n05718556": "music, euphony",
  "n05718935": "music",
  "n05719325": "piano music",
  "n05719466": "music of the spheres",
  "n05719605": "tone, pure tone",
  "n05719804": "harmonic",
  "n05719958": "fundamental, fundamental frequency, first harmonic",
  "n05720093": "overtone, partial, partial tone",
  "n05720248": "noise, dissonance, racket",
  "n05720521": "dub",
  "n05720602": "synesthesia, synaesthesia",
  "n05720826": "chromesthesia, chromaesthesia",
  "n05721019": "colored hearing, colored audition",
  "n05721180": "somesthesia, somaesthesia, somatesthesia, somatic sensation",
  "n05721500": "feeling",
  "n05721728": "constriction, tightness",
  "n05721990": "tactual sensation, tactility, touch perception, skin perceptiveness",
  "n05722208": "kinesthesia, kinaesthesia, feeling of movement",
  "n05722427": "touch, touch sensation, tactual sensation, tactile sensation, feeling",
  "n05722754": "pins and needles",
  "n05722868": "prickling, tingle, tingling",
  "n05723080": "creepiness",
  "n05723210": "cutaneous sensation, haptic sensation, skin sensation",
  "n05723417": "tickle",
  "n05723563": "itch, itchiness, itching",
  "n05723811": "pruritus",
  "n05724019": "pruritus ani",
  "n05724121": "pruritus vulvae",
  "n05724234": "topognosia, topognosis",
  "n05724357": "urtication",
  "n05724475": "pressure, pressure sensation",
  "n05724694": "pain, pain sensation, painful sensation",
  "n05724965": "mittelschmerz",
  "n05725137": "phantom limb pain",
  "n05725269": "twinge",
  "n05725378": "temperature",
  "n05725527": "heat, warmth",
  "n05725676": "cold, coldness",
  "n05725879": "comfort zone",
  "n05726093": "believing",
  "n05726237": "doublethink",
  "n05726345": "structure",
  "n05726596": "arrangement, organization, organisation, system",
  "n05727220": "classification system",
  "n05727427": "Dewey decimal classification, Dewey decimal system, decimal system of classification",
  "n05727808": "contrivance",
  "n05728024": "coordinate system, frame of reference, reference system, reference frame",
  "n05728271": "Cartesian coordinate system",
  "n05728493": "data structure",
  "n05728678": "design, plan",
  "n05729036": "distribution, statistical distribution",
  "n05729362": "equidistribution",
  "n05729483": "genetic map",
  "n05729609": "kinship system",
  "n05729875": "lattice",
  "n05730046": "living arrangement",
  "n05730163": "redundancy",
  "n05730365": "topology, network topology",
  "n05730591": "bus topology, bus",
  "n05730724": "loop topology, loop",
  "n05730930": "star topology, star",
  "n05731062": "mesh topology, mesh",
  "n05731223": "physical topology",
  "n05731405": "logical topology",
  "n05731568": "unitization, unitisation, chunking",
  "n05731779": "configuration, constellation",
  "n05732086": "space lattice, crystal lattice, Bravais lattice",
  "n05732273": "hierarchical structure, hierarchical data structure",
  "n05732446": "hierarchical classification system",
  "n05732614": "file system, filing system",
  "n05732756": "classification, categorization, categorisation, sorting",
  "n05733090": "grouping, pigeonholing",
  "n05733221": "rating system, scoring system",
  "n05733361": "ABO blood group system, ABO system, ABO group",
  "n05733583": "appraisal, assessment",
  "n05733864": "critical appraisal, critical analysis",
  "n05734018": "criticism, critique",
  "n05734288": "examen",
  "n05734381": "knock, roast",
  "n05734477": "self-criticism",
  "n05734559": "attribution, ascription",
  "n05734909": "attribution, ascription",
  "n05735145": "zoomorphism",
  "n05735256": "animatism",
  "n05735478": "imputation",
  "n05735680": "externalization, externalisation",
  "n05735789": "cross-classification, cross-division",
  "n05736002": "subsumption",
  "n05736149": "evaluation, valuation, rating",
  "n05736468": "overvaluation",
  "n05736593": "undervaluation",
  "n05736736": "pricing",
  "n05736893": "price gouging",
  "n05737025": "reevaluation",
  "n05737153": "mark, grade, score",
  "n05737532": "grade point",
  "n05737752": "percentile, centile",
  "n05737982": "decile",
  "n05738204": "quartile",
  "n05738404": "bond rating",
  "n05738625": "assay, check",
  "n05738894": "countercheck, double check",
  "n05739043": "diagnostic test, diagnostic assay",
  "n05739400": "Apgar score",
  "n05739613": "agglutination test",
  "n05739965": "heterophil test",
  "n05740162": "Widal test, Widal's test",
  "n05740300": "bioassay, bio-assay",
  "n05740560": "immunoassay, immunochemical assay",
  "n05740804": "radioimmunoassay",
  "n05740929": "biopsy",
  "n05741139": "cloze procedure, cloze test",
  "n05741340": "fecal occult test, faecal occult test, stool test",
  "n05741578": "GI series",
  "n05741765": "glucose tolerance test",
  "n05741949": "complement fixation test",
  "n05742241": "Wassermann test, Wasserman reaction, Wassermann",
  "n05742551": "blood test",
  "n05742703": "PSA blood test",
  "n05742962": "chorionic villus sampling, chorionic villus biopsy",
  "n05743174": "needle biopsy",
  "n05743296": "Pap test, Papanicolaou test, smear test",
  "n05743496": "paternity test",
  "n05743770": "PKU test",
  "n05743870": "pregnancy test",
  "n05744010": "Friedman test, rabbit test",
  "n05744288": "Queckenstedt's test",
  "n05744421": "radioactive iodine test",
  "n05744620": "radioactive iodine excretion test",
  "n05744792": "radioactive iodine uptake test, RAIU",
  "n05744976": "Rubin test",
  "n05745098": "skin test",
  "n05745369": "Dick test",
  "n05745486": "patch test",
  "n05745648": "Schick test",
  "n05745744": "scratch test",
  "n05745907": "tuberculin test, tuberculin skin test",
  "n05746163": "Mantoux test",
  "n05746424": "tine test",
  "n05746708": "intradermal test, subcutaneous test",
  "n05746868": "tissue typing",
  "n05747056": "Snellen test",
  "n05747160": "stress test",
  "n05747322": "treadmill test",
  "n05747495": "acid test",
  "n05747582": "reappraisal, revaluation, review, reassessment",
  "n05747815": "stocktaking, stock-taking",
  "n05747937": "underevaluation",
  "n05748054": "discrimination, secernment",
  "n05748285": "differentiation, distinction",
  "n05748614": "contradistinction",
  "n05748786": "line, dividing line, demarcation, contrast",
  "n05749042": "Rubicon, point of no return",
  "n05749210": "hairsplitting, word-splitting",
  "n05749402": "individualization, individualisation, individuation",
  "n05749619": "taste, appreciation, discernment, perceptiveness",
  "n05750027": "virtu, vertu, connoisseurship",
  "n05750163": "vogue, trend, style",
  "n05750471": "New Look",
  "n05750657": "fashion",
  "n05750948": "cut",
  "n05751063": "haute couture, high fashion, high style",
  "n05751173": "fad, craze, furor, furore, cult, rage",
  "n05751430": "retro",
  "n05751533": "bandwagon",
  "n05751707": "delicacy, discretion",
  "n05751794": "culture",
  "n05752020": "counterculture",
  "n05752179": "mass culture",
  "n05752296": "flower power",
  "n05752422": "letters",
  "n05752544": "learning, acquisition",
  "n05752921": "conditioning",
  "n05753207": "developmental learning",
  "n05753379": "digestion",
  "n05753564": "education",
  "n05753954": "internalization, internalisation, incorporation",
  "n05754197": "introjection",
  "n05754519": "introjection",
  "n05754730": "imprinting",
  "n05754899": "language learning",
  "n05755007": "audio lingual acquisition",
  "n05755156": "memorization, memorisation, committal to memory",
  "n05755393": "rote, rote learning",
  "n05755486": "accommodation",
  "n05755714": "assimilation",
  "n05755883": "study, work",
  "n05756203": "transfer, transfer of training, carry-over",
  "n05756414": "generalization, generalisation, stimulus generalization, stimulus generalisation",
  "n05756698": "irradiation",
  "n05756888": "physical education",
  "n05757049": "acculturation, assimilation",
  "n05757234": "mastering",
  "n05757423": "self-education, self-cultivation",
  "n05757536": "school, schooling",
  "n05757731": "special education",
  "n05757902": "vocational training, vocational education",
  "n05758059": "experience",
  "n05758382": "familiarization, familiarisation",
  "n05758548": "woodcraft",
  "n05758692": "extinction, experimental extinction",
  "n05758907": "aversive conditioning",
  "n05759017": "conditioned emotional response, CER, conditioned emotion",
  "n05759182": "classical conditioning",
  "n05759487": "instrumental conditioning",
  "n05759696": "operant conditioning",
  "n05759949": "counter conditioning",
  "n05760202": "memory, remembering",
  "n05760611": "short-term memory, STM, immediate memory",
  "n05760751": "working memory",
  "n05760877": "long-term memory, LTM",
  "n05761044": "episodic memory, personal memory",
  "n05761160": "semantic memory",
  "n05761277": "motor memory, muscle memory",
  "n05761380": "retrieval",
  "n05761559": "recall, recollection, reminiscence",
  "n05761918": "remembrance, recollection, anamnesis",
  "n05762149": "mind",
  "n05762258": "reconstruction, reconstructive memory",
  "n05762483": "reproduction, reproductive memory",
  "n05762671": "regurgitation",
  "n05762848": "reminiscence",
  "n05762998": "recognition, identification",
  "n05763412": "identity",
  "n05763767": "speaker identification, talker identification",
  "n05763916": "association, connection, connexion",
  "n05764197": "colligation",
  "n05764365": "overlap, convergence, intersection",
  "n05764613": "crossroads",
  "n05764779": "interface",
  "n05764973": "retrospection",
  "n05765159": "representational process",
  "n05765415": "symbol, symbolization, symbolisation, symbolic representation",
  "n05765901": "typification, exemplification",
  "n05766077": "picture",
  "n05766247": "interpretation, interpreting, rendition, rendering",
  "n05766698": "broad interpretation, judicial activism",
  "n05766984": "depicting, depiction, portraying, portrayal",
  "n05767245": "mirror",
  "n05767386": "anthropomorphism, theanthropism",
  "n05767580": "theanthropism",
  "n05767733": "imagination, imaging, imagery, mental imagery",
  "n05768107": "mind's eye",
  "n05768252": "vision",
  "n05768415": "picturing, envisioning",
  "n05768553": "dream, dreaming",
  "n05768806": "dream, dreaming",
  "n05769062": "nightmare",
  "n05769156": "wet dream",
  "n05769314": "chimera, chimaera",
  "n05769471": "reverie, revery, daydream, daydreaming, oneirism, air castle, castle in the air, castle in Spain",
  "n05769726": "woolgathering",
  "n05769833": "evocation",
  "n05769930": "pretense, pretence, make-believe",
  "n05770058": "search",
  "n05770257": "hunt",
  "n05770391": "pursuit, pursuance, quest",
  "n05770664": "higher cognitive process",
  "n05770926": "thinking, thought, thought process, cerebration, intellection, mentation",
  "n05771532": "free association",
  "n05771679": "suggestion",
  "n05771836": "construction, mental synthesis",
  "n05772044": "crystallization",
  "n05772215": "gestation",
  "n05772356": "reasoning, logical thinking, abstract thought",
  "n05772667": "analysis, analytic thinking",
  "n05773049": "argumentation, logical argument, argument, line of reasoning, line",
  "n05773407": "line of thought",
  "n05773548": "train of thought, thread",
  "n05773776": "line of inquiry, line of questioning",
  "n05773923": "conjecture",
  "n05774129": "deduction, deductive reasoning, synthesis",
  "n05774415": "generalization, generalisation, induction, inductive reasoning",
  "n05774614": "inference, illation",
  "n05775081": "prediction, anticipation, prevision",
  "n05775293": "projection",
  "n05775407": "prophecy, prognostication, vaticination",
  "n05775695": "crystal gazing",
  "n05775829": "prevision",
  "n05775919": "retrovision",
  "n05776015": "prefiguration, foreshadowing, adumbration",
  "n05776212": "divination, foretelling, soothsaying, fortune telling",
  "n05776592": "arithmancy",
  "n05776679": "dowse, dowsing, rhabdomancy",
  "n05776875": "geomancy",
  "n05777149": "hydromancy",
  "n05777298": "lithomancy",
  "n05777439": "necromancy",
  "n05777599": "oneiromancy",
  "n05777722": "onomancy",
  "n05777830": "palmistry, palm reading, chiromancy, chirology",
  "n05778029": "pyromancy",
  "n05778131": "astrology, star divination",
  "n05778646": "horoscopy",
  "n05778749": "alchemy",
  "n05778954": "pseudoscience",
  "n05779116": "syllogism",
  "n05779371": "theorization, theorisation",
  "n05779568": "ideology",
  "n05779712": "supposition, supposal",
  "n05779923": "presupposition",
  "n05780104": "abstraction, generalization, generalisation",
  "n05780339": "analogy",
  "n05780563": "corollary",
  "n05780718": "derivation",
  "n05780885": "deduction, entailment, implication",
  "n05781145": "extrapolation",
  "n05781347": "presumption",
  "n05781541": "conclusion",
  "n05781663": "non sequitur",
  "n05781800": "breakdown, partitioning",
  "n05781953": "cost-benefit analysis",
  "n05782140": "dissection",
  "n05782245": "elimination, reasoning by elimination",
  "n05782448": "reductionism",
  "n05782563": "reductionism",
  "n05782713": "systems analysis",
  "n05782884": "resolution, resolving",
  "n05783041": "factorization, factorisation, factoring",
  "n05783357": "diagonalization, diagonalisation",
  "n05783658": "ratiocination",
  "n05783768": "regress, reasoning backward",
  "n05783940": "synthesis, synthetic thinking",
  "n05784152": "trend analysis",
  "n05784242": "cogitation, study",
  "n05784482": "lucubration",
  "n05784560": "mysticism",
  "n05784699": "ideation",
  "n05784831": "consideration",
  "n05785067": "deliberation, weighing, advisement",
  "n05785311": "exploration",
  "n05785508": "contemplation, reflection, reflexion, rumination, musing, thoughtfulness",
  "n05785885": "meditation, speculation",
  "n05786184": "meditation",
  "n05786372": "think",
  "n05786655": "introspection, self-contemplation, self-examination",
  "n05786871": "soul-searching, self-analysis",
  "n05787005": "examen, examination",
  "n05787175": "inwardness",
  "n05787498": "outwardness",
  "n05787848": "omphaloskepsis, navel-gazing",
  "n05788029": "retrospect",
  "n05788149": "decision making, deciding",
  "n05788552": "determination",
  "n05788713": "eclecticism, eclectic method",
  "n05788924": "groupthink",
  "n05789089": "settlement, resolution, closure",
  "n05789432": "judgment, judgement, judging",
  "n05789666": "prejudgment, prejudgement",
  "n05789808": "reversal, change of mind, flip-flop, turnabout, turnaround",
  "n05790012": "reconsideration, second thought, afterthought, rethink",
  "n05790242": "choice, pick, selection",
  "n05790452": "pleasure",
  "n05790572": "cull, reject",
  "n05790758": "favorite, favourite",
  "n05790944": "option, alternative, choice",
  "n05791311": "obverse",
  "n05791452": "preference, druthers",
  "n05791602": "wish",
  "n05791764": "way",
  "n05791864": "default option, default",
  "n05792010": "possibility, possible action, opening",
  "n05792192": "possible",
  "n05792312": "impossibility, impossible action",
  "n05792446": "impossible",
  "n05792575": "Hobson's choice",
  "n05792691": "soft option",
  "n05792842": "excogitation",
  "n05793000": "explanation",
  "n05793210": "rationale, principle",
  "n05793554": "basis, base, foundation, fundament, groundwork, cornerstone",
  "n05793907": "meat and potatoes",
  "n05794057": "key",
  "n05794189": "natural history",
  "n05794403": "rationalization, rationalisation",
  "n05794618": "raison d'etre",
  "n05794694": "planning, preparation, provision",
  "n05795044": "agreement, arrangement",
  "n05795244": "collusion",
  "n05795334": "prearrangement",
  "n05795460": "reservation",
  "n05795654": "upgrade",
  "n05795835": "applecart",
  "n05795957": "mens rea, malice aforethought",
  "n05796222": "premeditation",
  "n05796423": "calculation, deliberation",
  "n05796617": "premeditation, forethought",
  "n05796750": "problem solving",
  "n05796937": "convergent thinking",
  "n05797177": "divergent thinking, out-of-the-box thinking",
  "n05797473": "inspiration",
  "n05797597": "inquiry, enquiry, research",
  "n05797899": "nature study",
  "n05798043": "experiment, experimentation",
  "n05798315": "empirical research",
  "n05798413": "control experiment",
  "n05798569": "control condition, control",
  "n05798863": "condition, experimental condition",
  "n05799071": "pilot experiment",
  "n05799212": "trial, trial run, test, tryout",
  "n05799581": "field trial, field test",
  "n05799761": "alpha test",
  "n05799952": "beta test",
  "n05800153": "road test",
  "n05800272": "testament",
  "n05800387": "test drive",
  "n05800527": "trial balloon",
  "n05800611": "probe, investigation",
  "n05800838": "fishing expedition",
  "n05800998": "poll, opinion poll, public opinion poll, canvass",
  "n05801286": "exit poll",
  "n05801465": "straw vote, straw poll",
  "n05801594": "heraldry",
  "n05802185": "calculation, computation, figuring, reckoning",
  "n05802547": "extrapolation",
  "n05802730": "interpolation",
  "n05802912": "conversion",
  "n05803095": "data conversion",
  "n05803212": "digitization, digitisation",
  "n05803379": "estimate, estimation, approximation, idea",
  "n05803747": "credit rating, credit",
  "n05803938": "guess, guesswork, guessing, shot, dead reckoning",
  "n05804136": "guesstimate, guestimate",
  "n05804274": "overestimate, overestimation, overrating, overreckoning",
  "n05804491": "underestimate, underestimation, underrating, underreckoning",
  "n05804793": "knowing",
  "n05805012": "know",
  "n05805157": "cognizance, ken",
  "n05805277": "prevision, foresight, farsightedness, prospicience",
  "n05805475": "understanding, apprehension, discernment, savvy",
  "n05805902": "comprehension",
  "n05806231": "incomprehension",
  "n05806380": "self-knowledge",
  "n05806498": "smattering",
  "n05806623": "appreciation, grasp, hold",
  "n05806855": "grasping",
  "n05807012": "sense",
  "n05807147": "hindsight",
  "n05807306": "insight, brainstorm, brainwave",
  "n05807540": "realization, realisation, recognition",
  "n05807933": "light",
  "n05808102": "revelation",
  "n05808218": "discovery, breakthrough, find",
  "n05808442": "flash",
  "n05808557": "linguistic process, language",
  "n05808794": "reading",
  "n05809082": "speed-reading",
  "n05809192": "content, cognitive content, mental object",
  "n05809745": "tradition",
  "n05809878": "world, reality",
  "n05810143": "otherworld",
  "n05810250": "real world, real life",
  "n05810440": "deja vu",
  "n05810561": "life, living",
  "n05810768": "reliving, re-experiencing",
  "n05810948": "object",
  "n05811214": "food, food for thought, intellectual nourishment",
  "n05811380": "pabulum",
  "n05811466": "antipathy",
  "n05811662": "bugbear, hobgoblin",
  "n05811884": "execration",
  "n05812038": "center, centre, center of attention, centre of attention",
  "n05812370": "conversation piece",
  "n05812485": "crosshairs",
  "n05812646": "cynosure",
  "n05812813": "eye-catcher",
  "n05812921": "hallucination",
  "n05813109": "infatuation",
  "n05813229": "love, passion",
  "n05813457": "noumenon, thing-in-itself",
  "n05813626": "reminder",
  "n05813822": "memento, souvenir",
  "n05813912": "memento mori",
  "n05814019": "shades of",
  "n05814162": "universe, universe of discourse",
  "n05814291": "topic, subject, issue, matter",
  "n05814650": "issue",
  "n05814952": "gut issue, hot-button issue",
  "n05815075": "paramount issue",
  "n05815267": "pocketbook issue, bread-and-butter issue",
  "n05815412": "quodlibet",
  "n05815517": "area",
  "n05815691": "blind spot",
  "n05815890": "remit",
  "n05816154": "res judicata, res adjudicata",
  "n05816287": "information",
  "n05816622": "datum, data point",
  "n05816790": "reading, meter reading, indication",
  "n05817145": "acquaintance, familiarity, conversance, conversancy",
  "n05817396": "fact",
  "n05817743": "case",
  "n05817845": "detail, item, point",
  "n05818182": "particular, specific",
  "n05818388": "general",
  "n05818620": "matter of fact",
  "n05818741": "observation",
  "n05818978": "scientific fact",
  "n05819149": "reason",
  "n05819338": "score",
  "n05819453": "truth",
  "n05819644": "home truth",
  "n05819773": "verity",
  "n05819890": "minutia",
  "n05820023": "nook and cranny, nooks and crannies",
  "n05820170": "respect, regard",
  "n05820325": "sticking point",
  "n05820462": "technicality, trifle, triviality",
  "n05820620": "example, illustration, instance, representative",
  "n05821102": "apology, excuse",
  "n05821246": "exception",
  "n05821486": "precedent, case in point",
  "n05821646": "quintessence",
  "n05821775": "sample",
  "n05822001": "coupon",
  "n05822085": "cross section",
  "n05822239": "specimen",
  "n05822337": "grab sample",
  "n05822507": "random sample",
  "n05822612": "tasting",
  "n05822746": "circumstance, condition, consideration",
  "n05823054": "justification",
  "n05823317": "mitigating circumstance",
  "n05823517": "background, background knowledge",
  "n05823747": "descriptor",
  "n05823932": "evidence, grounds",
  "n05824313": "predictor",
  "n05824514": "probable cause",
  "n05824739": "proof, cogent evidence",
  "n05824985": "reductio ad absurdum, reductio",
  "n05825245": "confirmation, verification, check, substantiation",
  "n05825688": "bed check",
  "n05825802": "crosscheck",
  "n05825942": "parity check, redundancy check, odd-even check",
  "n05826092": "checksum",
  "n05826291": "establishment, validation",
  "n05826469": "disproof, falsification, refutation",
  "n05826722": "confutation",
  "n05826832": "counterexample",
  "n05826914": "lead, track, trail",
  "n05827129": "tip-off",
  "n05827253": "evocation, induction, elicitation",
  "n05827509": "kick",
  "n05827684": "stimulation, stimulus, stimulant, input",
  "n05828102": "turn-on",
  "n05828263": "turnoff, negative stimulation",
  "n05828425": "conditioned stimulus",
  "n05828552": "reinforcing stimulus, reinforcer, reinforcement",
  "n05828820": "positive reinforcing stimulus, positive reinforcer",
  "n05829012": "negative reinforcing stimulus, negative reinforcer",
  "n05829213": "discriminative stimulus, cue",
  "n05829342": "positive stimulus",
  "n05829480": "negative stimulus",
  "n05829656": "bonus, fillip",
  "n05829782": "joy, delight, pleasure",
  "n05830059": "annoyance, bother, botheration, pain, infliction, pain in the neck, pain in the ass",
  "n05830527": "nuisance",
  "n05830832": "abatable nuisance",
  "n05831057": "attractive nuisance",
  "n05831270": "mixed nuisance",
  "n05831412": "private nuisance",
  "n05831566": "public nuisance, common nuisance",
  "n05831784": "irritant, thorn",
  "n05831939": "plague",
  "n05832113": "aversive stimulus",
  "n05832264": "concern, worry, headache, vexation",
  "n05832621": "bugaboo",
  "n05832745": "burden, load, encumbrance, incumbrance, onus",
  "n05833022": "business",
  "n05833169": "dead weight",
  "n05833252": "fardel",
  "n05833371": "imposition",
  "n05833511": "pill",
  "n05833683": "grief, sorrow",
  "n05833840": "idea, thought",
  "n05834567": "inspiration",
  "n05834758": "source, seed, germ",
  "n05834946": "taproot",
  "n05835162": "muse",
  "n05835292": "mother",
  "n05835449": "afflatus",
  "n05835568": "cogitation",
  "n05835747": "concept, conception, construct",
  "n05836275": "conceptualization, conceptualisation, conceptuality",
  "n05836468": "perception",
  "n05836598": "notion",
  "n05836714": "mumpsimus",
  "n05836921": "preoccupation",
  "n05837128": "self-absorption",
  "n05837271": "layout",
  "n05837370": "trap, snare",
  "n05837651": "iron trap",
  "n05837743": "speed trap",
  "n05837850": "idea",
  "n05837957": "judgment, judgement, mind",
  "n05838176": "decision, determination, conclusion",
  "n05838530": "predetermination",
  "n05838765": "category",
  "n05839024": "kind, sort, form, variety",
  "n05839547": "breed",
  "n05839663": "pigeonhole",
  "n05839776": "rubric",
  "n05839910": "way",
  "n05840076": "description",
  "n05840188": "type",
  "n05840431": "nature",
  "n05840650": "version, variant, variation, edition",
  "n05841029": "antitype",
  "n05841151": "art form",
  "n05841351": "architectural style, style of architecture, type of architecture",
  "n05841624": "Bauhaus",
  "n05841740": "Byzantine architecture",
  "n05841985": "classical architecture, Greco-Roman architecture",
  "n05842191": "Greek architecture",
  "n05842290": "Roman architecture",
  "n05842387": "Gothic, Gothic architecture",
  "n05842706": "Romanesque, Romanesque architecture",
  "n05843042": "Norman architecture",
  "n05843236": "perpendicular, perpendicular style, English-Gothic, English-Gothic architecture",
  "n05843513": "Tudor architecture",
  "n05843687": "Moorish, Moorish architecture",
  "n05843895": "Victorian architecture",
  "n05844105": "style",
  "n05844282": "flavor, flavour",
  "n05844433": "charm",
  "n05844545": "strangeness",
  "n05844663": "color, colour",
  "n05844881": "species",
  "n05845013": "genus",
  "n05845140": "brand, make",
  "n05845332": "genre",
  "n05845419": "like, ilk",
  "n05845562": "manner",
  "n05845652": "model",
  "n05845749": "stripe",
  "n05845888": "like, the like, the likes of",
  "n05846054": "rule, regulation",
  "n05846355": "restriction, limitation",
  "n05846626": "narrowness",
  "n05846932": "rule, formula",
  "n05847320": "metarule",
  "n05847438": "algorithm, algorithmic rule, algorithmic program",
  "n05847658": "sorting algorithm",
  "n05847753": "stemmer, stemming algorithm",
  "n05847956": "heuristic, heuristic rule, heuristic program",
  "n05848182": "lateral thinking",
  "n05848357": "recursion",
  "n05848541": "guidepost, guideline, rule of thumb",
  "n05848691": "cy pres, rule of cy pres, cy pres doctrine",
  "n05848912": "working principle, working rule",
  "n05849040": "property, attribute, dimension",
  "n05849284": "quality, character, lineament",
  "n05849667": "texture",
  "n05849789": "feature, characteristic",
  "n05850212": "feature of speech, feature",
  "n05850432": "invariant",
  "n05850624": "aspect, facet",
  "n05850823": "attraction, attractor, attracter, attractive feature, magnet",
  "n05851131": "badge",
  "n05851330": "centerpiece, centrepiece",
  "n05851517": "contour",
  "n05851744": "excellence, excellency",
  "n05852007": "external",
  "n05852125": "peculiarity, distinctive feature, distinguishing characteristic",
  "n05852287": "calling card",
  "n05852430": "safety feature",
  "n05852553": "side",
  "n05852819": "downside",
  "n05852973": "hand",
  "n05853100": "sector, sphere",
  "n05853273": "department",
  "n05853449": "surface",
  "n05853636": "attention",
  "n05853820": "tourist attraction",
  "n05853924": "foil, enhancer",
  "n05854150": "abstraction, abstract",
  "n05854474": "absolute",
  "n05854812": "teacher",
  "n05855004": "thing",
  "n05855125": "quantity",
  "n05855517": "quantum",
  "n05855748": "quantum",
  "n05855904": "term",
  "n05856066": "numerical quantity",
  "n05856217": "zero, zero point",
  "n05856388": "value",
  "n05856589": "eigenvalue, eigenvalue of a matrix, eigenvalue of a square matrix, characteristic root of a square matrix",
  "n05856867": "scale value",
  "n05856979": "average",
  "n05857165": "vote, voter turnout",
  "n05857346": "operand",
  "n05857459": "variable, variable quantity",
  "n05857781": "argument",
  "n05857974": "arity",
  "n05858093": "independent variable, experimental variable",
  "n05858317": "factor",
  "n05858407": "correlate, correlative",
  "n05858551": "degree of freedom",
  "n05858699": "dependent variable",
  "n05858936": "constant, constant quantity, invariable",
  "n05859071": "parameter, parametric quantity",
  "n05859277": "parameter",
  "n05859477": "degree of freedom",
  "n05859630": "product, mathematical product",
  "n05859811": "factorial",
  "n05859991": "multiple",
  "n05860200": "double",
  "n05860377": "triple",
  "n05860498": "quadruple",
  "n05860639": "lowest common multiple, least common multiple, lcm",
  "n05860869": "grand total",
  "n05860975": "subtotal",
  "n05861067": "sum, amount, total",
  "n05861317": "degree",
  "n05861463": "degree of a term",
  "n05861579": "degree of a polynomial",
  "n05861716": "first degree",
  "n05861855": "polynomial, multinomial",
  "n05862113": "biquadratic, biquadratic polynomial, quartic polynomial",
  "n05862268": "homogeneous polynomial",
  "n05862410": "monic polynomial",
  "n05862501": "quadratic, quadratic polynomial",
  "n05862613": "quantic",
  "n05862721": "series",
  "n05862970": "power series",
  "n05863106": "convergence, convergency",
  "n05863302": "divergence, divergency",
  "n05863480": "geometric series",
  "n05863583": "Fourier series",
  "n05863736": "predictor variable",
  "n05863919": "proportional",
  "n05864047": "infinitesimal",
  "n05864177": "random variable, variate, variant, stochastic variable, chance variable",
  "n05864351": "scalar",
  "n05864481": "tensor",
  "n05864577": "vector",
  "n05864758": "vector product, cross product",
  "n05864884": "scalar product, inner product, dot product",
  "n05865034": "vector sum, resultant",
  "n05865174": "radius vector",
  "n05865311": "radius vector",
  "n05865454": "be-all and end-all, be all and end all",
  "n05865652": "plot element",
  "n05865774": "McGuffin, MacGuffin",
  "n05865998": "point",
  "n05866199": "attractor, attracter",
  "n05866489": "strange attractor, chaotic attractor",
  "n05866653": "intersection, intersection point, point of intersection",
  "n05866822": "metacenter, metacentre",
  "n05867283": "vertex",
  "n05867413": "part, section, division",
  "n05868051": "beginning",
  "n05868272": "middle",
  "n05868477": "end",
  "n05868779": "high point",
  "n05868954": "component, constituent, element, factor, ingredient",
  "n05869407": "leaven, leavening",
  "n05869584": "whole",
  "n05869857": "unit",
  "n05870055": "one",
  "n05870180": "compound",
  "n05870365": "complex, composite",
  "n05870615": "hybrid",
  "n05870790": "syndrome",
  "n05870916": "law, natural law",
  "n05871140": "divine law",
  "n05871245": "dictate",
  "n05871362": "fundamentals, basics, fundamental principle, basic principle, bedrock",
  "n05871618": "logic",
  "n05871792": "pleasure principle, pleasure-pain principle, pleasure-unpleasure principle",
  "n05872098": "reality principle",
  "n05872362": "insurrectionism",
  "n05872477": "principle",
  "n05872742": "rudiment, first rudiment, first principle, alphabet, ABC, ABC's, ABCs",
  "n05872982": "law, law of nature",
  "n05873884": "lexicalized concept",
  "n05874017": "all-or-none law",
  "n05874232": "principle, rule",
  "n05874716": "Archimedes' principle, law of Archimedes",
  "n05874941": "Avogadro's law, Avogadro's hypothesis",
  "n05875148": "Bernoulli's law, law of large numbers",
  "n05875392": "Benford's law",
  "n05875723": "Bose-Einstein statistics",
  "n05875988": "Boyle's law, Mariotte's law",
  "n05876148": "Coulomb's Law",
  "n05876469": "Dalton's law, Dalton's law of partial pressures, law of partial pressures",
  "n05876912": "distribution law",
  "n05877178": "Maxwell-Boltzmann distribution law, Boltzmann distribution law",
  "n05877412": "equilibrium law, law of chemical equilibrium",
  "n05877718": "Fechner's law, Weber-Fechner law",
  "n05877991": "Fermi-Dirac statistics",
  "n05878229": "Gay-Lussac's law, Charles's law, law of volumes",
  "n05878440": "Gestalt law of organization, Gestalt principle of organization",
  "n05878771": "Henry's law",
  "n05879003": "Hooke's law",
  "n05879204": "Hubble's law, Hubble law",
  "n05879441": "Kepler's law, Kepler's law of planetary motion",
  "n05879693": "Kepler's first law",
  "n05879856": "Kepler's second law, law of areas, law of equal areas",
  "n05880187": "Kepler's third law, harmonic law",
  "n05880432": "Kirchhoff's laws",
  "n05880715": "law of averages",
  "n05880854": "law of constant proportion, law of definite proportions",
  "n05881102": "law of diminishing returns",
  "n05881303": "law of effect",
  "n05881578": "law of equivalent proportions, law of reciprocal proportions",
  "n05881867": "law of gravitation, Newton's law of gravitation",
  "n05882226": "law of multiple proportions, Dalton's law",
  "n05882537": "law of mass action",
  "n05882793": "law of thermodynamics",
  "n05883035": "second law of thermodynamics",
  "n05883296": "third law of thermodynamics",
  "n05883473": "zeroth law of thermodynamics",
  "n05883688": "Le Chatelier's principle, Le Chatelier's law, Le Chatelier principle, Le Chatelier-Braun principle",
  "n05883992": "Gresham's Law",
  "n05884433": "Mendel's law",
  "n05884736": "law of segregation",
  "n05884984": "law of independent assortment",
  "n05885187": "mass-energy equivalence",
  "n05885412": "Naegele's rule",
  "n05885622": "Newton's law of motion, Newton's law, law of motion",
  "n05885822": "first law of motion, Newton's first law of motion, Newton's first law",
  "n05886039": "second law of motion, Newton's second law of motion, Newton's second law",
  "n05886266": "third law of motion, Newton's third law of motion, Newton's third law, law of action and reaction",
  "n05886455": "Ohm's law",
  "n05886613": "Pascal's law, Pascal's law of fluid pressures",
  "n05886939": "Pauli exclusion principle, exclusion principle",
  "n05887156": "periodic law, Mendeleev's law",
  "n05887365": "Planck's law",
  "n05887712": "Planck's radiation law",
  "n05887911": "big-bang theory, big bang theory",
  "n05888215": "nebular hypothesis",
  "n05888377": "planetesimal hypothesis",
  "n05888572": "steady state theory, continuous creation theory",
  "n05888929": "hypothesis, possibility, theory",
  "n05889575": "hypothetical",
  "n05889762": "gemmule",
  "n05889896": "fact",
  "n05890021": "mean sun",
  "n05890249": "model, theoretical account, framework",
  "n05890642": "Copernican system",
  "n05890809": "Ptolemaic system",
  "n05890963": "M-theory",
  "n05891232": "string theory",
  "n05891426": "audit program, audit programme",
  "n05891572": "outline, schema, scheme",
  "n05891783": "speculation, conjecture",
  "n05892096": "assumption, supposition, supposal",
  "n05892427": "prerequisite, requirement",
  "n05892651": "requirement, demand",
  "n05892880": "precondition",
  "n05892991": "academic requirement",
  "n05893143": "language requirement",
  "n05893261": "essential condition, sine qua non",
  "n05893356": "given, presumption, precondition",
  "n05893512": "basic assumption, constatation, self-evident truth",
  "n05893653": "misconception",
  "n05893916": "fallacy, false belief",
  "n05894143": "logical fallacy",
  "n05894311": "hysteron proteron",
  "n05894460": "ignoratio elenchi",
  "n05894631": "pathetic fallacy",
  "n05894814": "petitio principii, petitio",
  "n05894972": "post hoc, post hoc ergo propter hoc",
  "n05895138": "sophism, sophistry, sophistication",
  "n05895373": "paralogism",
  "n05895465": "error, erroneous belief",
  "n05895588": "self-deception, self-deceit",
  "n05895723": "mistake, misunderstanding, misapprehension",
  "n05896059": "illusion, fantasy, phantasy, fancy",
  "n05896379": "bubble",
  "n05896515": "will-o'-the-wisp, ignis fatuus",
  "n05896618": "wishful thinking",
  "n05896733": "delusion, hallucination",
  "n05896998": "autism",
  "n05897255": "infantile autism",
  "n05897553": "apparition, phantom, phantasm, phantasma, fantasm, shadow",
  "n05897825": "unidentified flying object, UFO, flying saucer",
  "n05898035": "Flying Dutchman",
  "n05898171": "ghost, shade, spook, wraith, specter, spectre",
  "n05898430": "disorientation, freak out",
  "n05898568": "plan, program, programme",
  "n05899087": "program, programme",
  "n05899442": "master plan",
  "n05899621": "Apollo program",
  "n05899831": "Gemini program",
  "n05900010": "Mercury program",
  "n05900263": "defense program, defense policy, defence program, defence policy",
  "n05900430": "educational program",
  "n05900583": "rehabilitation program",
  "n05900716": "space program",
  "n05900888": "Superfund program, Superfund",
  "n05901248": "vocational rehabilitation program",
  "n05901410": "tax program, tax policy",
  "n05901508": "policy",
  "n05901840": "activism",
  "n05902014": "beggar-my-neighbor policy, beggar-my-neighbour policy, beggar-my-neighbor strategy, beggar-my-neighbour strategy",
  "n05902327": "blueprint, design, pattern",
  "n05902545": "plan of action",
  "n05902786": "battle plan",
  "n05902872": "system",
  "n05903112": "credit system",
  "n05903229": "legal system",
  "n05903445": "bail",
  "n05903738": "jury system",
  "n05903878": "patent system",
  "n05904010": "tax system",
  "n05904135": "voting system, electoral system",
  "n05904313": "uninominal system, uninominal voting system, single-member system, scrutin uninomial system, scrutin uninominal voting system",
  "n05904616": "list system, scrutin de liste, scrutin de liste system",
  "n05904918": "pricing system",
  "n05905027": "promotion system",
  "n05905152": "tactic, tactics, maneuver, manoeuvre",
  "n05905348": "scheme, strategy",
  "n05905706": "travel plan, itinerary",
  "n05905802": "contrivance, stratagem, dodge",
  "n05906080": "plant",
  "n05906305": "pump-and-dump scheme",
  "n05906554": "wangle, wangling",
  "n05906734": "counterterrorism",
  "n05906867": "game plan",
  "n05906998": "game plan",
  "n05907251": "house of cards, bubble",
  "n05907479": "playbook",
  "n05907682": "plot, secret plan, game",
  "n05908000": "pyramid scheme",
  "n05908391": "counterplot, counterplan",
  "n05908520": "intrigue, machination",
  "n05908727": "priestcraft",
  "n05908882": "conspiracy, cabal",
  "n05909097": "Gunpowder Plot",
  "n05909306": "waiting game",
  "n05909384": "wheeze",
  "n05909585": "regimen, regime",
  "n05909730": "academic program",
  "n05909921": "training program",
  "n05910070": "biofeedback",
  "n05910328": "preemployment training program",
  "n05910453": "project, projection",
  "n05910577": "moneymaker, money-spinner, cash cow",
  "n05910712": "vocational program",
  "n05910810": "works program",
  "n05910940": "agenda, docket, schedule",
  "n05911124": "menu, fare",
  "n05911255": "pension plan, pension account, retirement plan, retirement savings plan, retirement savings account, retirement account, retirement program",
  "n05911560": "401-k plan, 401-k",
  "n05912012": "individual retirement account, IRA",
  "n05912243": "Keogh plan",
  "n05912399": "employee savings plan",
  "n05912552": "road map, guideline",
  "n05912814": "stock purchase plan",
  "n05912969": "employee stock ownership plan, ESOP",
  "n05913160": "figment",
  "n05913275": "generalization, generalisation, generality",
  "n05913538": "principle, rule",
  "n05913842": "pillar",
  "n05913994": "pillar of Islam",
  "n05914234": "shahadah",
  "n05914359": "salat, salaat, salah, salaah",
  "n05914591": "sawm",
  "n05914851": "zakat",
  "n05915034": "hajj, haj, hadj",
  "n05915356": "yang",
  "n05915584": "yin",
  "n05915811": "feng shui",
  "n05916155": "suggestion",
  "n05916306": "inkling, intimation, glimmering, glimmer",
  "n05916520": "posthypnotic suggestion",
  "n05916739": "impression, feeling, belief, notion, opinion",
  "n05917174": "presence",
  "n05917328": "reaction",
  "n05917477": "effect",
  "n05917675": "first blush",
  "n05917869": "sound effect",
  "n05917994": "special effect",
  "n05918171": "stage effect",
  "n05918267": "theorem",
  "n05918379": "Bayes' theorem",
  "n05918704": "Bayes' postulate",
  "n05919034": "intuition, hunch, suspicion",
  "n05919263": "heart, bosom",
  "n05919431": "prescience, prevision",
  "n05919549": "notion, whim, whimsy, whimsey",
  "n05919866": "meaning, substance",
  "n05920159": "burden",
  "n05920272": "theme, motif",
  "n05920482": "topos",
  "n05920651": "semantics",
  "n05920791": "significance, import, implication",
  "n05921123": "kernel, substance, core, center, centre, essence, gist, heart, heart and soul, inwardness, marrow, meat, nub, pith, sum, nitty-gritty",
  "n05921685": "bare bones",
  "n05921868": "hypostasis",
  "n05922014": "quiddity, haecceity",
  "n05922175": "quintessence",
  "n05922305": "stuff",
  "n05922450": "tenor, strain",
  "n05922651": "drift, purport",
  "n05922809": "undertone, undercurrent",
  "n05922949": "reference, denotation, extension",
  "n05923314": "reference",
  "n05923566": "connotation",
  "n05923696": "ideal",
  "n05923983": "value",
  "n05924165": "introject",
  "n05924408": "idealization, idealisation",
  "n05924519": "paragon, idol, perfection, beau ideal",
  "n05924730": "gold standard",
  "n05924920": "criterion, standard",
  "n05925177": "design criteria",
  "n05925366": "exemplar, example, model, good example",
  "n05925702": "beauty, beaut",
  "n05925862": "ego ideal",
  "n05926049": "keynote",
  "n05926150": "kink",
  "n05926236": "wisdom",
  "n05926358": "reconditeness, abstruseness, abstrusity, profoundness, profundity",
  "n05926676": "representation, mental representation, internal representation",
  "n05927169": "instantiation",
  "n05927364": "antitype",
  "n05927586": "stereotype",
  "n05927813": "schema, scheme",
  "n05928118": "image, mental image",
  "n05928391": "imagination image, thought-image",
  "n05928513": "interpretation, reading, version",
  "n05928733": "reinterpretation",
  "n05928840": "phantasmagoria",
  "n05929008": "character, role, theatrical role, part, persona",
  "n05929363": "bit part, minor role",
  "n05929461": "soubrette",
  "n05929582": "heavy",
  "n05929670": "hero",
  "n05929780": "ingenue",
  "n05929887": "title role, name part",
  "n05930010": "psychosexuality",
  "n05930136": "percept, perception, perceptual experience",
  "n05930386": "figure",
  "n05930574": "ground",
  "n05930736": "form, shape, pattern",
  "n05931152": "fractal",
  "n05931341": "gestalt",
  "n05931512": "grid",
  "n05931658": "Amsler grid",
  "n05931827": "kaleidoscope",
  "n05931980": "mosaic",
  "n05932062": "strand",
  "n05932282": "sonata form",
  "n05932477": "visual percept, visual image",
  "n05932709": "eye candy",
  "n05932891": "field, field of view",
  "n05933054": "sight",
  "n05933246": "view, aspect, prospect, scene, vista, panorama",
  "n05933638": "visual field, field of vision, field of regard",
  "n05933834": "background, ground",
  "n05934029": "coast",
  "n05934123": "exposure",
  "n05934278": "foreground",
  "n05934396": "glimpse",
  "n05934550": "middle distance",
  "n05934673": "side view",
  "n05934780": "tableau",
  "n05934852": "microscopic field",
  "n05934962": "operative field",
  "n05935060": "memory",
  "n05935292": "recollection",
  "n05935381": "engram, memory trace",
  "n05935535": "confabulation",
  "n05935717": "screen memory",
  "n05935871": "memory image",
  "n05936016": "memory picture",
  "n05936130": "afterimage, aftersensation",
  "n05936300": "aftertaste",
  "n05936381": "visual image, visualization, visualisation",
  "n05936561": "fusion, optical fusion",
  "n05936704": "mental picture, picture, impression",
  "n05936995": "auditory image",
  "n05937112": "model, example",
  "n05937417": "lodestar, loadstar",
  "n05937524": "prototype, paradigm, epitome, image",
  "n05937878": "concentrate",
  "n05938014": "imago",
  "n05938170": "type specimen, holotype",
  "n05938314": "microcosm",
  "n05938400": "original, archetype, pilot",
  "n05938633": "pacesetter, pacemaker",
  "n05938795": "pattern",
  "n05938976": "template, templet, guide",
  "n05939109": "prefiguration",
  "n05939244": "prodigy",
  "n05939432": "appearance",
  "n05939636": "illusion, semblance",
  "n05939809": "irradiation",
  "n05939948": "three-D, 3-D, 3D",
  "n05940090": "phantom limb",
  "n05940214": "mirage",
  "n05940302": "front",
  "n05940414": "blur, fuzz",
  "n05940688": "unsoundness",
  "n05940869": "abstractionism, unrealism",
  "n05941037": "concretism, concrete representation",
  "n05941210": "shape, embodiment",
  "n05941423": "belief",
  "n05942104": "apophatism",
  "n05942338": "cataphatism",
  "n05942579": "doctrine of analogy, analogy",
  "n05942888": "conviction, strong belief, article of faith",
  "n05943066": "faith, trust",
  "n05943300": "doctrine, philosophy, philosophical system, school of thought, ism",
  "n05944686": "philosophy",
  "n05944958": "expectation, outlook, prospect",
  "n05945227": "fetishism, fetichism",
  "n05945383": "geneticism",
  "n05945508": "meliorism",
  "n05945642": "opinion, sentiment, persuasion, view, thought",
  "n05946089": "autotelism",
  "n05946234": "originalism",
  "n05946406": "pacifism, pacificism",
  "n05946582": "predestinarianism",
  "n05946687": "religion, faith, religious belief",
  "n05948264": "cult, cultus, religious cult",
  "n05948537": "cult",
  "n05948716": "ecclesiasticism",
  "n05948857": "mysticism, religious mysticism",
  "n05949088": "quietism",
  "n05949266": "Sufism",
  "n05949336": "nature worship",
  "n05949472": "revealed religion",
  "n05949603": "eyes",
  "n05949726": "public opinion, popular opinion, opinion, vox populi",
  "n05949937": "preconception, prepossession, parti pris, preconceived opinion, preconceived idea, preconceived notion",
  "n05950234": "taboo, tabu",
  "n05950475": "irrational hostility",
  "n05950559": "pole",
  "n05950733": "promise, hope",
  "n05950982": "rainbow",
  "n05951072": "foretaste",
  "n05951180": "possibility",
  "n05951323": "anticipation, expectancy",
  "n05951566": "apprehension, misgiving",
  "n05951692": "revolutionism",
  "n05951820": "sacerdotalism",
  "n05951969": "spiritualism",
  "n05952199": "spiritual world, spiritual domain, unseen",
  "n05952367": "suffragism",
  "n05952490": "supernaturalism",
  "n05952678": "superstition, superstitious notion",
  "n05952829": "supremacism",
  "n05952979": "theory",
  "n05953263": "egoism",
  "n05953416": "patchwork, hodgepodge, jumble",
  "n05953614": "theosophy",
  "n05953804": "theosophism",
  "n05953881": "anthroposophy",
  "n05954100": "Kabbalah, Kabbala, Kabala, Cabbalah, Cabbala, Cabala, Qabbalah, Qabbala",
  "n05954366": "Kabbalism, Cabalism",
  "n05954481": "thought",
  "n05954672": "totemism",
  "n05954807": "tribalism",
  "n05954894": "values",
  "n05955095": "vampirism",
  "n05955186": "mainstream",
  "n05955323": "principle",
  "n05955593": "accounting principle, accounting standard",
  "n05955848": "chivalry, knightliness",
  "n05956019": "ethic, moral principle, value-system, value orientation",
  "n05956287": "Chartism",
  "n05956489": "Hellenism",
  "n05956651": "legal principle, judicial principle, judicial doctrine",
  "n05956906": "jus sanguinis",
  "n05957078": "jus soli",
  "n05957238": "preemption, pre-emption",
  "n05957428": "relation back, relation",
  "n05957737": "scruple",
  "n05957913": "Golden Rule",
  "n05958114": "Athanasian Creed",
  "n05958208": "abolitionism",
  "n05958337": "absolutism",
  "n05958427": "amoralism",
  "n05958549": "animalism",
  "n05958712": "animism",
  "n05958919": "antiestablishmentarianism, antiestablishmentism",
  "n05959082": "asceticism",
  "n05959263": "British empiricism",
  "n05959407": "contextualism",
  "n05959578": "creationism",
  "n05959785": "creation science",
  "n05959954": "creed, credo",
  "n05960121": "divine right, divine right of kings",
  "n05960464": "dogma",
  "n05960698": "dualism",
  "n05960925": "dynamism",
  "n05961141": "epicureanism",
  "n05961278": "establishmentarianism, establishmentism",
  "n05961429": "ethicism",
  "n05961608": "expansionism",
  "n05961745": "experimentalism",
  "n05961867": "formalism",
  "n05962043": "functionalism",
  "n05962166": "Girondism",
  "n05962252": "gospel",
  "n05962414": "gymnosophy",
  "n05962602": "imitation",
  "n05962785": "mimesis",
  "n05962936": "individualism, laissez faire",
  "n05963101": "individualism",
  "n05963299": "rugged individualism",
  "n05963494": "internationalism",
  "n05963744": "unilateralism",
  "n05963973": "one-way street",
  "n05964098": "irredentism, irridentism",
  "n05964322": "literalism",
  "n05964445": "majority rule, democracy",
  "n05964643": "monism",
  "n05964805": "multiculturalism",
  "n05965022": "nationalism",
  "n05965195": "nationalism",
  "n05965388": "nihilism",
  "n05965586": "pacifism, pacificism, passivism",
  "n05965749": "pluralism",
  "n05965933": "populism",
  "n05966129": "predestination, foreordination, preordination, predetermination",
  "n05966602": "presentism",
  "n05966798": "election",
  "n05966958": "rationalism, freethinking",
  "n05967097": "reformism",
  "n05967191": "humanism, secular humanism",
  "n05967402": "humanitarianism, humanism",
  "n05967588": "egalitarianism, equalitarianism",
  "n05967773": "feminism",
  "n05967894": "juju",
  "n05967977": "magic, thaumaturgy",
  "n05968209": "mojo",
  "n05968288": "occultism",
  "n05968450": "occultism",
  "n05968553": "reincarnationism",
  "n05968710": "secessionism",
  "n05968835": "secularism",
  "n05968971": "aesthetic, esthetic",
  "n05969194": "Aristotelianism, peripateticism",
  "n05969537": "conceptualism",
  "n05969758": "Confucianism",
  "n05970012": "deconstruction, deconstructionism",
  "n05970311": "empiricism, empiricist philosophy, sensationalism",
  "n05970564": "environmentalism",
  "n05970755": "existentialism, existential philosophy, existentialist philosophy",
  "n05971086": "determinism",
  "n05971394": "fatalism",
  "n05971621": "formalism",
  "n05971913": "hereditarianism",
  "n05972103": "idealism",
  "n05972264": "intuitionism",
  "n05972417": "logicism",
  "n05972585": "materialism, physicalism",
  "n05972781": "mechanism",
  "n05973007": "mentalism",
  "n05973198": "nativism",
  "n05973372": "naturalism",
  "n05973603": "Neoplatonism",
  "n05974156": "nominalism",
  "n05974363": "operationalism",
  "n05974564": "Platonism, realism",
  "n05974798": "pragmatism",
  "n05975051": "instrumentalism",
  "n05975258": "probabilism",
  "n05975473": "rationalism",
  "n05975659": "realism, naive realism",
  "n05975846": "relativism",
  "n05976065": "Scholasticism",
  "n05976257": "semiotics, semiology",
  "n05976471": "sensualism, sensationalism",
  "n05976649": "solipsism",
  "n05976805": "spiritualism",
  "n05976948": "Stoicism",
  "n05977135": "subjectivism",
  "n05977340": "Taoism, Daoism",
  "n05977619": "teleology",
  "n05977800": "traditionalism",
  "n05977977": "vitalism",
  "n05978159": "conjuring, conjuration, conjury, invocation",
  "n05978370": "old wives' tale",
  "n05978472": "exorcism, dispossession",
  "n05978623": "evocation, summoning",
  "n05978812": "sorcery, black magic, black art, necromancy",
  "n05979198": "theurgy",
  "n05979350": "witchcraft, witchery",
  "n05979454": "enchantment, bewitchment",
  "n05979595": "diabolism, demonism, Satanism",
  "n05979800": "white magic",
  "n05979909": "unbelief, disbelief",
  "n05980051": "agnosticism, skepticism, scepticism",
  "n05980256": "atheism",
  "n05980412": "heresy, unorthodoxy",
  "n05980750": "iconoclasm",
  "n05980875": "goal, end",
  "n05981230": "aim, object, objective, target",
  "n05981546": "bourn, bourne",
  "n05981648": "end-all",
  "n05981768": "destination, terminus",
  "n05981936": "grail",
  "n05982024": "no-goal",
  "n05982152": "purpose, intent, intention, aim, design",
  "n05982720": "intention",
  "n05982915": "mind, idea",
  "n05983122": "cross-purpose",
  "n05983217": "final cause",
  "n05983347": "sake",
  "n05983463": "view",
  "n05983654": "will",
  "n05983801": "business",
  "n05983943": "occasions",
  "n05984066": "point",
  "n05984182": "thing",
  "n05984287": "education",
  "n05984584": "experience",
  "n05984936": "acculturation, culture",
  "n05985126": "meme",
  "n05985381": "lore, traditional knowledge",
  "n05985602": "folklore",
  "n05985999": "eruditeness, erudition, learnedness, learning, scholarship, encyclopedism, encyclopaedism",
  "n05986295": "letters",
  "n05986395": "enlightenment",
  "n05986594": "foundation, grounding",
  "n05986822": "centralism",
  "n05986948": "containment",
  "n05987283": "moderationism",
  "n05987397": "obscurantism",
  "n05987522": "Thatcherism",
  "n05987650": "ultramontanism",
  "n05987835": "edification, sophistication",
  "n05987969": "satori",
  "n05988097": "disenchantment, disillusion, disillusionment",
  "n05988282": "ignorance",
  "n05988498": "dark, darkness",
  "n05988743": "ignorantness, nescience, unknowing, unknowingness",
  "n05988956": "inexperience, rawness",
  "n05989265": "unenlightenment",
  "n05989386": "illiteracy",
  "n05989479": "theory",
  "n05990089": "theory of gravitation, theory of gravity, gravitational theory, Newton's theory of gravitation",
  "n05990480": "principle of relativity",
  "n05990738": "Occam's Razor, Ockham's Razor, principle of parsimony, law of parsimony",
  "n05990981": "principle of equivalence",
  "n05991242": "principle of liquid displacement",
  "n05991441": "principle of superposition, Huygens' principle of superposition",
  "n05991776": "principle of superposition, superposition principle, superposition",
  "n05992015": "mass-action principle, mass action",
  "n05992274": "localization of function, localisation of function, localization principle, localisation principle, localization, localisation",
  "n05992624": "lateralization, lateralisation, laterality",
  "n05992804": "blastogenesis",
  "n05992949": "preformation, theory of preformation",
  "n05993235": "dialectical materialism",
  "n05993367": "positivism, logical positivism",
  "n05993622": "Comtism",
  "n05993844": "scientific theory",
  "n05994484": "field theory",
  "n05994706": "organicism",
  "n05994935": "economic theory",
  "n05995220": "consumerism",
  "n05995355": "Keynesianism",
  "n05995576": "liberalism",
  "n05995726": "Malthusianism, Malthusian theory",
  "n05995898": "monetarism",
  "n05996110": "Stevens' law, power law, Stevens' power law",
  "n05996353": "Weber's law",
  "n05996646": "discipline, subject, subject area, subject field, field, field of study, study, bailiwick",
  "n05997361": "communications, communication theory",
  "n05997659": "major",
  "n05997814": "region, realm",
  "n05998052": "frontier",
  "n05998225": "genealogy",
  "n05998356": "allometry",
  "n05998526": "bibliotics",
  "n05998724": "ology",
  "n05998893": "symbology",
  "n05998994": "grey area, gray area",
  "n05999134": "territory",
  "n05999266": "knowledge domain, knowledge base, domain",
  "n05999455": "metaknowledge",
  "n05999540": "scientific knowledge",
  "n05999797": "science, scientific discipline",
  "n06000400": "natural science",
  "n06000644": "mathematics, math, maths",
  "n06003682": "pure mathematics",
  "n06004067": "arithmetic",
  "n06004599": "algorism",
  "n06004685": "geometry",
  "n06005399": "affine geometry",
  "n06005518": "elementary geometry, parabolic geometry, Euclidean geometry",
  "n06005692": "Euclid's axiom, Euclid's postulate, Euclidean axiom",
  "n06005999": "Euclid's first axiom",
  "n06006117": "Euclid's second axiom",
  "n06006243": "Euclid's third axiom",
  "n06006364": "Euclid's fourth axiom",
  "n06006458": "Euclid's fifth axiom, parallel axiom",
  "n06006609": "fractal geometry",
  "n06006777": "non-Euclidean geometry",
  "n06007046": "hyperbolic geometry",
  "n06007381": "elliptic geometry, Riemannian geometry",
  "n06007642": "numerical analysis",
  "n06007859": "spherical geometry",
  "n06008007": "spherical trigonometry",
  "n06008148": "triangulation",
  "n06008382": "analytic geometry, analytical geometry, coordinate geometry",
  "n06008609": "axis",
  "n06008896": "coordinate axis",
  "n06009086": "origin",
  "n06009233": "x-axis",
  "n06009334": "y-axis",
  "n06009433": "z-axis",
  "n06009537": "major axis",
  "n06009685": "semimajor axis",
  "n06009857": "minor axis",
  "n06009988": "semiminor axis",
  "n06010105": "principal axis, optic axis",
  "n06010381": "optic axis",
  "n06010583": "inertial reference frame, inertial frame",
  "n06010737": "space-time, space-time continuum",
  "n06010930": "coordinate, co-ordinate",
  "n06011092": "Cartesian coordinate",
  "n06011446": "dimension",
  "n06011589": "abscissa",
  "n06011692": "ordinate",
  "n06011793": "intercept",
  "n06011903": "polar coordinate",
  "n06012108": "plane geometry",
  "n06012225": "solid geometry",
  "n06012340": "projective geometry, descriptive geometry",
  "n06012513": "trigonometry, trig",
  "n06012726": "algebra",
  "n06012998": "quadratics",
  "n06013126": "linear algebra",
  "n06013298": "vector algebra",
  "n06013471": "decomposition, vector decomposition",
  "n06013584": "matrix algebra",
  "n06013741": "calculus, infinitesimal calculus",
  "n06014043": "analysis",
  "n06014278": "Fourier analysis, harmonic analysis",
  "n06014435": "differential calculus, method of fluxions",
  "n06014730": "derived function, derivative, differential coefficient, differential, first derivative",
  "n06015053": "partial derivative, partial",
  "n06015271": "integral calculus",
  "n06015505": "integral",
  "n06015700": "indefinite integral",
  "n06015863": "definite integral",
  "n06015978": "calculus of variations",
  "n06016121": "set theory",
  "n06016276": "interval",
  "n06016462": "closed interval, bounded interval",
  "n06016600": "open interval, unbounded interval",
  "n06016746": "sub-interval",
  "n06016853": "group, mathematical group",
  "n06017053": "subgroup",
  "n06017193": "group theory",
  "n06017334": "Galois theory",
  "n06017472": "Abelian group, commutative group",
  "n06017594": "topology, analysis situs",
  "n06017895": "metamathematics",
  "n06018022": "applied mathematics, applied math",
  "n06018267": "linear programming",
  "n06018465": "statistics",
  "n06020737": "statistical method, statistical procedure",
  "n06021013": "least squares, method of least squares",
  "n06021247": "multivariate analysis",
  "n06021499": "statistic",
  "n06021761": "average, norm",
  "n06022076": "demographic",
  "n06022291": "deviation",
  "n06022509": "moment",
  "n06022727": "nonparametric statistic, distribution free statistic",
  "n06023022": "parametric statistic",
  "n06023243": "age norm",
  "n06023369": "outlier",
  "n06023476": "mean deviation, mean deviation from the mean",
  "n06023675": "mode, modal value",
  "n06023821": "median, median value",
  "n06023969": "mean, mean value",
  "n06024230": "arithmetic mean, first moment, expectation, expected value",
  "n06024431": "geometric mean",
  "n06024576": "harmonic mean",
  "n06024761": "second moment",
  "n06024936": "variance",
  "n06025172": "standard deviation",
  "n06025287": "covariance",
  "n06025521": "frequency distribution",
  "n06025689": "normal distribution, Gaussian distribution",
  "n06025854": "Poisson distribution",
  "n06026088": "normal curve, bell-shaped curve, Gaussian curve, Gaussian shape",
  "n06026276": "population, universe",
  "n06026508": "subpopulation",
  "n06026635": "sample distribution, sample, sampling",
  "n06026885": "random sample",
  "n06027051": "stratified sample, representative sample, proportional sample",
  "n06027264": "regression, simple regression, regression toward the mean, statistical regression",
  "n06027613": "multiple regression, multiple correlation",
  "n06027843": "multicollinearity",
  "n06028021": "regression analysis",
  "n06028260": "regression equation, regression of y on x",
  "n06028566": "regression coefficient",
  "n06028904": "linear regression, rectilinear regression",
  "n06029118": "curvilinear regression",
  "n06029306": "regression line, regression curve",
  "n06029547": "time series",
  "n06029676": "vital statistics",
  "n06029830": "correlational analysis",
  "n06030049": "correlation matrix",
  "n06030196": "factor analysis",
  "n06030601": "analysis of variance, ANOVA",
  "n06030906": "correlation table",
  "n06031248": "correlation, correlational statistics",
  "n06031657": "curvilinear correlation, nonlinear correlation, skew correlation",
  "n06031866": "partial correlation",
  "n06032066": "first-order correlation",
  "n06032246": "correlation coefficient, coefficient of correlation, correlation",
  "n06032752": "covariation",
  "n06032898": "positive correlation, direct correlation",
  "n06033172": "negative correlation, indirect correlation",
  "n06033427": "product-moment correlation coefficient, Pearson product-moment correlation coefficient",
  "n06033692": "multiple correlation coefficient",
  "n06033889": "biserial correlation coefficient, biserial correlation",
  "n06034101": "nonparametric statistics",
  "n06034301": "rank-order correlation coefficient, rank-order correlation, rank-difference correlation coefficient, rank-difference correlation",
  "n06034611": "Kendall test",
  "n06034870": "Kendall partial rank correlation",
  "n06035014": "coefficient of concordance",
  "n06035213": "tau coefficient of correlation, Kendall's tau, Kendall rank correlation",
  "n06035413": "phi coefficient, phi correlation, fourfold point correlation",
  "n06035674": "split-half correlation, chance-half correlation",
  "n06035918": "tetrachoric correlation coefficient, tetrachoric correlation",
  "n06036157": "spurious correlation",
  "n06036506": "binomial",
  "n06036697": "binomial distribution, Bernoulli distribution",
  "n06036939": "binomial theorem",
  "n06037108": "probability theory, theory of probability",
  "n06037298": "life science, bioscience",
  "n06037666": "biology, biological science",
  "n06041466": "biomedical science",
  "n06041600": "biometrics, biometry, biostatistics",
  "n06041805": "craniology",
  "n06041955": "dermatoglyphics",
  "n06042187": "dietetics",
  "n06042312": "macrobiotics",
  "n06042486": "eugenics",
  "n06042690": "dysgenics, cacogenics",
  "n06042905": "euthenics",
  "n06043075": "medicine, medical specialty",
  "n06045562": "medical science",
  "n06045847": "phrenology",
  "n06046037": "aeromedicine, aerospace medicine, aviation medicine",
  "n06046245": "allergology",
  "n06046383": "anesthesiology",
  "n06046528": "angiology",
  "n06046692": "bacteriology",
  "n06046898": "biomedicine",
  "n06047096": "biomedicine",
  "n06047275": "cardiology",
  "n06047430": "dentistry, dental medicine, odontology",
  "n06047923": "cosmetic dentistry",
  "n06048052": "dental surgery",
  "n06048184": "endodontics, endodontia",
  "n06048373": "exodontics, exodontia",
  "n06048552": "orthodontics, orthodontia, orthodonture, dental orthopedics, dental orthopaedics",
  "n06048851": "periodontics, periodontia",
  "n06049091": "prosthetics",
  "n06049250": "prosthodontics, prosthodontia",
  "n06049500": "dermatology",
  "n06049673": "emergency medicine",
  "n06049850": "endocrinology",
  "n06050024": "epidemiology",
  "n06050257": "forensic medicine, forensic pathology",
  "n06050490": "gastroenterology",
  "n06050650": "geriatrics, gerontology",
  "n06050901": "gynecology, gynaecology",
  "n06051134": "hematology, haematology",
  "n06051380": "hygiene, hygienics",
  "n06051542": "immunology",
  "n06051925": "immunochemistry, chemoimmunology",
  "n06052165": "immunopathology",
  "n06052300": "internal medicine, general medicine",
  "n06052521": "nephrology",
  "n06052685": "nuclear medicine",
  "n06052864": "neurology, clinical neurology",
  "n06053101": "neuropsychiatry",
  "n06053280": "nosology, diagnostics",
  "n06053439": "obstetrics, OB, tocology, midwifery",
  "n06053854": "fetology, foetology",
  "n06053982": "perinatology",
  "n06054266": "oncology",
  "n06054446": "ophthalmology",
  "n06054700": "otology",
  "n06054892": "pharmacology, pharmacological medicine, materia medica",
  "n06055300": "pharmacy, pharmaceutics",
  "n06055529": "pharmacokinetics",
  "n06055692": "posology",
  "n06055824": "psychopharmacology",
  "n06055946": "psychiatry, psychopathology, psychological medicine",
  "n06056787": "alienism",
  "n06056923": "psychotherapy, psychotherapeutics, mental hygiene",
  "n06057172": "clinical psychology",
  "n06057326": "Freudian psychology",
  "n06057435": "Jungian psychology",
  "n06057539": "anatomy, general anatomy",
  "n06059125": "clinical anatomy, applied anatomy",
  "n06059282": "comparative anatomy",
  "n06059412": "dental anatomy",
  "n06059540": "developmental anatomy",
  "n06059709": "functional anatomy, physiological anatomy, morphophysiology",
  "n06059865": "gross anatomy, macroscopic anatomy",
  "n06060049": "microscopic anatomy",
  "n06060208": "neuroanatomy",
  "n06060336": "osteology",
  "n06060463": "regional anatomy, topographic anatomy, topology",
  "n06060732": "audiology, audiometry",
  "n06060845": "pathology",
  "n06061631": "pediatrics, paediatrics, pediatric medicine, pedology",
  "n06061917": "neonatology",
  "n06062076": "podiatry, chiropody",
  "n06062225": "proctology",
  "n06062407": "radiology",
  "n06062655": "rheumatology",
  "n06062842": "rhinolaryngology, otorhinolaryngology, otolaryngology",
  "n06063072": "serology",
  "n06063275": "space medicine",
  "n06063417": "sports medicine",
  "n06063588": "surgery",
  "n06064106": "orthopedics, orthopaedics",
  "n06064345": "therapeutics",
  "n06064462": "toxicology",
  "n06064683": "thoracic medicine",
  "n06064838": "traumatology, accident surgery",
  "n06065020": "tropical medicine",
  "n06065208": "urology, urogenital medicine",
  "n06065423": "veterinary medicine",
  "n06065640": "virology",
  "n06065819": "agronomy, scientific agriculture",
  "n06066072": "agrobiology",
  "n06066267": "agrology",
  "n06066414": "biogeography",
  "n06066555": "botany, phytology",
  "n06069747": "mycology",
  "n06069886": "pomology",
  "n06069996": "cryobiology",
  "n06070179": "cryonics",
  "n06070503": "cytology",
  "n06070738": "cytogenetics",
  "n06070929": "ecology, bionomics, environmental science",
  "n06071426": "embryology",
  "n06071722": "exobiology, space biology, astrobiology",
  "n06071934": "forestry",
  "n06072145": "silviculture",
  "n06072275": "entomology, bugology",
  "n06072476": "lepidopterology, lepidoptery",
  "n06072619": "ethology",
  "n06072776": "herpetology",
  "n06072912": "ichthyology",
  "n06073067": "malacology",
  "n06073215": "mammalogy",
  "n06073331": "oology",
  "n06073494": "ornithology",
  "n06073647": "primatology",
  "n06073748": "protozoology",
  "n06073888": "paleontology, palaeontology, fossilology",
  "n06074189": "paleoanthropology, palaeoanthropology, human paleontology, human palaeontology",
  "n06074372": "paleobotany, palaeobotany",
  "n06074507": "phycology, algology",
  "n06074613": "pteridology",
  "n06074728": "paleodendrology, palaeodendrology",
  "n06074860": "paleozoology, palaeozoology",
  "n06075016": "paleomammalogy",
  "n06075112": "paleornithology, palaeornithology",
  "n06075218": "functional genomics",
  "n06075378": "structural genomics",
  "n06075527": "genetics, genetic science",
  "n06077087": "genomics",
  "n06077276": "proteomics",
  "n06077413": "histology",
  "n06077648": "microbiology",
  "n06077916": "molecular genetics",
  "n06078088": "molecular biology",
  "n06078327": "morphology",
  "n06078511": "neurobiology",
  "n06078724": "paleobiology, palaeobiology",
  "n06078978": "neurology",
  "n06079247": "pharmacogenetics",
  "n06079439": "teratology",
  "n06079620": "biochemistry",
  "n06080192": "enzymology",
  "n06080361": "zymology, zymurgy",
  "n06080522": "physiology",
  "n06081602": "neurophysiology",
  "n06081833": "neuroscience",
  "n06082025": "brain science",
  "n06082136": "cognitive neuroscience",
  "n06082310": "hemodynamics",
  "n06082459": "kinesiology",
  "n06082610": "myology",
  "n06082709": "paleoecology, palaeoecology",
  "n06082834": "radiobiology",
  "n06082972": "sociobiology",
  "n06083243": "zoology, zoological science",
  "n06084469": "chemistry, chemical science",
  "n06088995": "organic chemistry",
  "n06089314": "inorganic chemistry",
  "n06089447": "physical chemistry",
  "n06089728": "phytochemistry",
  "n06089857": "electrochemistry",
  "n06090064": "femtochemistry",
  "n06090304": "geochemistry",
  "n06090397": "photochemistry",
  "n06090540": "radiochemistry, nuclear chemistry",
  "n06090678": "surface chemistry",
  "n06090869": "physics, natural philosophy",
  "n06094587": "physics, physical science",
  "n06094774": "acoustics",
  "n06095022": "astronomy, uranology",
  "n06096451": "astrodynamics",
  "n06096600": "astrometry",
  "n06096759": "radio astronomy",
  "n06096913": "aeronautics, astronautics",
  "n06097231": "avionics",
  "n06097478": "biophysics",
  "n06097594": "celestial mechanics",
  "n06097775": "astrophysics",
  "n06097983": "selenology",
  "n06098088": "solar physics",
  "n06098195": "cosmology, cosmogony, cosmogeny",
  "n06098687": "cryogenics, cryogeny",
  "n06098876": "crystallography",
  "n06099107": "electromagnetism, electromagnetics",
  "n06099269": "electronics",
  "n06100101": "electrostatics",
  "n06100236": "mechanics",
  "n06100555": "nuclear physics, atomic physics, nucleonics",
  "n06100778": "optics",
  "n06101180": "catoptrics",
  "n06101333": "holography",
  "n06101551": "particle physics, high-energy physics, high energy physics",
  "n06101849": "plasma physics",
  "n06101973": "quantum physics",
  "n06102116": "quasiparticle",
  "n06102316": "rheology",
  "n06102476": "atomism, atomic theory, atomist theory, atomistic theory",
  "n06102865": "holism, holistic theory",
  "n06103270": "atomic theory",
  "n06103422": "Bohr theory",
  "n06103746": "Rutherford atom",
  "n06104073": "conservation",
  "n06104372": "conservation of charge, conservation of electricity",
  "n06104578": "conservation of energy, law of conservation of energy, first law of thermodynamics",
  "n06104844": "conservation of mass, conservation of matter, law of conservation of mass, law of conservation of matter",
  "n06105107": "conservation of momentum",
  "n06105314": "parity, conservation of parity, space-reflection symmetry, mirror symmetry",
  "n06105609": "cell theory, cell doctrine",
  "n06105873": "wave theory, undulatory theory, wave theory of light",
  "n06106084": "corpuscular theory, corpuscular theory of light",
  "n06106305": "kinetic theory, kinetic theory of gases",
  "n06106502": "relativity, theory of relativity, relativity theory, Einstein's theory of relativity",
  "n06106820": "general relativity, general theory of relativity, general relativity theory, Einstein's general theory of relativity",
  "n06107083": "special relativity, special theory of relativity, special relativity theory, Einstein's special theory of relativity",
  "n06107436": "supersymmetry",
  "n06107660": "quantum theory",
  "n06107850": "wave mechanics",
  "n06108079": "uncertainty principle, indeterminacy principle",
  "n06108339": "kinetic theory of heat",
  "n06108487": "germ theory",
  "n06108643": "information theory",
  "n06108850": "theory of dissociation, theory of electrolytic dissociation, Arrhenius theory of dissociation",
  "n06109227": "theory of evolution, theory of organic evolution, evolutionism",
  "n06109487": "theory of indicators, Ostwald's theory of indicators",
  "n06109771": "theory of inheritance",
  "n06109972": "Mendelism, Mendelianism",
  "n06110091": "Darwinism",
  "n06110267": "neo-Darwinism",
  "n06110426": "Lamarckism",
  "n06110598": "Neo-Lamarckism",
  "n06110823": "thermochemistry",
  "n06111003": "punctuated equilibrium, theory of punctuated equilibrium",
  "n06111253": "harmonics",
  "n06111335": "classical mechanics, Newtonian mechanics",
  "n06111479": "solid-state physics",
  "n06111728": "statistical mechanics",
  "n06111961": "quantum mechanics",
  "n06112223": "quantum field theory",
  "n06112505": "quantum electrodynamics, QED",
  "n06112681": "quantum chromodynamics, QCD",
  "n06113009": "fluid mechanics, hydraulics",
  "n06113170": "pneumatics",
  "n06113301": "statics",
  "n06113415": "hydrostatics",
  "n06113597": "dynamics, kinetics",
  "n06113777": "kinematics",
  "n06113914": "hydrodynamics, hydrokinetics",
  "n06114069": "magnetohydrodynamics",
  "n06114246": "ballistics",
  "n06114351": "aeromechanics, aerodynamics",
  "n06114578": "thermodynamics",
  "n06114849": "thermostatics, thermodynamics of equilibrium",
  "n06115004": "electron optics",
  "n06115179": "microelectronics",
  "n06115322": "thermionics",
  "n06115476": "earth science",
  "n06115701": "geology",
  "n06117202": "hypsography",
  "n06117395": "paleogeology, palaeogeology",
  "n06117562": "geophysics, geophysical science",
  "n06117855": "morphology, geomorphology",
  "n06118146": "orology, orography",
  "n06118236": "stratigraphy",
  "n06118370": "tectonics, plate tectonics, plate tectonic theory",
  "n06118563": "meteorology",
  "n06119226": "aerology",
  "n06119377": "climatology",
  "n06119496": "bioclimatology",
  "n06119608": "nephology",
  "n06119729": "hydrology",
  "n06119904": "oceanography, oceanology",
  "n06120111": "hydrography",
  "n06120317": "limnology",
  "n06120496": "seismology",
  "n06120653": "volcanology, vulcanology",
  "n06120769": "magnetism, magnetics",
  "n06120881": "geodesy",
  "n06121113": "mineralogy",
  "n06121375": "petrology, lithology",
  "n06121554": "speleology, spelaeology",
  "n06121690": "petroleum geology",
  "n06121854": "economic geology",
  "n06122033": "mining geology",
  "n06122178": "geography, geographics",
  "n06122445": "physical geography, physiography",
  "n06122578": "topography",
  "n06122747": "topology",
  "n06122976": "economic geography",
  "n06123126": "cosmography",
  "n06123363": "architecture",
  "n06124395": "architectonics, tectonics",
  "n06124532": "landscape architecture",
  "n06124706": "urban planning",
  "n06124864": "interior design",
  "n06125041": "engineering, engineering science, applied science, technology",
  "n06125698": "metallurgy",
  "n06125996": "powder metallurgy",
  "n06126177": "aeronautical engineering",
  "n06126336": "bionics",
  "n06126523": "biotechnology, bioengineering, ergonomics",
  "n06126761": "biotechnology, biotech",
  "n06127057": "bioremediation",
  "n06127209": "genetic engineering, gene-splicing, recombinant DNA technology",
  "n06127460": "chemical engineering",
  "n06127683": "civil engineering",
  "n06127873": "hydraulic engineering",
  "n06128024": "electrical engineering, EE",
  "n06128307": "telecommunication",
  "n06128570": "computer science, computing",
  "n06132724": "object",
  "n06133018": "logic",
  "n06133203": "artificial intelligence, AI",
  "n06133503": "machine translation, MT",
  "n06133654": "robotics",
  "n06133802": "animatronics",
  "n06133935": "telerobotics",
  "n06134075": "architectural engineering",
  "n06134271": "industrial engineering, industrial management",
  "n06134510": "information technology, IT",
  "n06134716": "mechanical engineering",
  "n06134899": "nanotechnology",
  "n06135095": "tribology",
  "n06135335": "nuclear engineering",
  "n06135503": "naval engineering",
  "n06135659": "rocketry",
  "n06135806": "metrology",
  "n06135915": "nutrition",
  "n06136105": "futurology, futuristics",
  "n06136258": "psychology, psychological science",
  "n06137759": "abnormal psychology, psychopathology",
  "n06137918": "associationism, association theory",
  "n06138099": "atomism",
  "n06138347": "applied psychology, industrial psychology",
  "n06138582": "cognitive psychology",
  "n06138789": "comparative psychology, animal psychology",
  "n06138941": "developmental psychology, genetic psychology, child psychology",
  "n06139135": "differential psychology",
  "n06139285": "experimental psychology, psychonomics",
  "n06139491": "psychophysics",
  "n06139764": "behaviorism, behaviourism, behavioristic psychology, behaviouristic psychology",
  "n06140054": "functionalism",
  "n06140268": "memory",
  "n06140437": "problem solving",
  "n06140587": "psycholinguistics",
  "n06140799": "physiological psychology, neuropsychology, psychophysiology",
  "n06141023": "psychometry, psychometrics, psychometrika",
  "n06141197": "reflexology",
  "n06141324": "Gestalt psychology, configurationism",
  "n06141561": "social psychology",
  "n06141768": "psychodynamics",
  "n06141962": "group dynamics",
  "n06142118": "information science, informatics, information processing, IP",
  "n06142412": "natural language processing, NLP, human language technology",
  "n06142598": "cybernetics",
  "n06142861": "cognitive science",
  "n06143154": "social science",
  "n06143454": "civics",
  "n06143546": "anthropology",
  "n06144081": "archeology, archaeology",
  "n06144762": "Assyriology",
  "n06144855": "Egyptology",
  "n06144970": "Sumerology",
  "n06145062": "micropaleontology",
  "n06145158": "marine archeology, marine archaeology, underwater archeology, underwater archaeology",
  "n06145326": "paleoclimatology, palaeoclimatology",
  "n06145446": "paleogeography, palaeogeography",
  "n06145586": "paleography",
  "n06145729": "paleopathology, palaeopathology",
  "n06145897": "paletiology, palaetiology",
  "n06146053": "epigraphy",
  "n06146142": "paleology, palaeology",
  "n06146260": "protohistory, protoanthropology",
  "n06146407": "protoarcheology, protoarchaeology",
  "n06146546": "ethnography, descriptive anthropology",
  "n06146760": "paleoethnography, palaeoethnography",
  "n06146880": "ethnology",
  "n06147141": "physical anthropology",
  "n06147308": "craniometry",
  "n06147522": "social anthropology, cultural anthropology",
  "n06147770": "garbology",
  "n06147873": "mythology",
  "n06148002": "ritualism",
  "n06148148": "politics, political science, government",
  "n06148748": "geopolitics",
  "n06148915": "geostrategy",
  "n06149038": "realpolitik, practical politics",
  "n06149192": "politics, political sympathies",
  "n06149344": "home economics, home ec, domestic science, household arts",
  "n06149484": "economics, economic science, political economy",
  "n06150222": "game theory, theory of games",
  "n06150449": "econometrics",
  "n06150633": "finance",
  "n06150933": "macroeconomics",
  "n06151108": "microeconomics",
  "n06151282": "supply-side economics",
  "n06151550": "proxemics",
  "n06151693": "sociology",
  "n06151942": "criminology",
  "n06152125": "demography, human ecology",
  "n06152311": "psephology",
  "n06152460": "penology, poenology",
  "n06152631": "sociometry",
  "n06152734": "strategics",
  "n06152821": "systematics",
  "n06152973": "biosystematics, biosystematy",
  "n06153186": "taxonomy",
  "n06153402": "cladistics, cladistic analysis",
  "n06153698": "thanatology",
  "n06153846": "humanistic discipline, humanities, liberal arts, arts",
  "n06154464": "neoclassicism",
  "n06154724": "classicism, classicalism",
  "n06155075": "Romanticism, Romantic Movement",
  "n06155432": "English",
  "n06155567": "history",
  "n06155891": "historicism",
  "n06156015": "art history",
  "n06156169": "iconology",
  "n06156346": "chronology",
  "n06156521": "glottochronology",
  "n06156752": "history",
  "n06156968": "fine arts, beaux arts",
  "n06157326": "performing arts",
  "n06158057": "Occidentalism",
  "n06158185": "Orientalism, Oriental Studies",
  "n06158346": "philosophy",
  "n06159473": "ethics, moral philosophy",
  "n06159777": "bioethics",
  "n06159924": "neuroethics",
  "n06160055": "casuistry",
  "n06160244": "casuistry",
  "n06160418": "eudemonism, endaemonism",
  "n06160596": "hedonism",
  "n06160781": "probabilism",
  "n06161048": "etiology, aetiology",
  "n06161223": "aesthetics, esthetics",
  "n06161606": "axiology",
  "n06161718": "jurisprudence, law, legal philosophy",
  "n06162042": "contract law",
  "n06162223": "corporation law",
  "n06162377": "matrimonial law",
  "n06162528": "patent law",
  "n06162653": "metaphysics",
  "n06162847": "ontology",
  "n06162979": "ontology",
  "n06163223": "cosmology",
  "n06163396": "dialectics",
  "n06163548": "dialectic",
  "n06163751": "logic",
  "n06164665": "symbolic logic, mathematical logic, formal logic",
  "n06165000": "Boolean logic, Boolean algebra",
  "n06165147": "propositional logic, propositional calculus",
  "n06165364": "predicate calculus, functional calculus",
  "n06165623": "quantification",
  "n06165823": "modal logic",
  "n06166046": "alethic logic",
  "n06166166": "deontic logic",
  "n06166274": "epistemic logic",
  "n06166394": "doxastic logic",
  "n06166494": "fuzzy logic",
  "n06166644": "modal logic",
  "n06166748": "epistemology",
  "n06166898": "methodology, methodological analysis",
  "n06167106": "phenomenology",
  "n06167328": "philosophical doctrine, philosophical theory",
  "n06168152": "structuralism, structural sociology",
  "n06168334": "structuralism, structural anthropology",
  "n06168552": "computational linguistics",
  "n06168703": "dialect geography, linguistic geography",
  "n06168855": "etymology",
  "n06169050": "historical linguistics, diachronic linguistics, diachrony",
  "n06169285": "literary study",
  "n06169488": "literature, lit",
  "n06169667": "comparative literature",
  "n06169806": "literary criticism, lit crit",
  "n06169932": "poetics",
  "n06170025": "prosody, metrics",
  "n06170371": "classics",
  "n06170498": "rhetoric",
  "n06170892": "library science",
  "n06171040": "linguistics, philology",
  "n06171265": "dialectology",
  "n06171388": "musicology",
  "n06171524": "Sinology",
  "n06171650": "stemmatology, stemmatics",
  "n06172071": "trivium",
  "n06172294": "quadrivium",
  "n06172502": "cryptanalysis, cryptanalytics, cryptography, cryptology",
  "n06172789": "linguistics",
  "n06174404": "grammar",
  "n06175829": "descriptive grammar",
  "n06175967": "prescriptive grammar",
  "n06176107": "syntax, sentence structure, phrase structure",
  "n06176322": "syntax",
  "n06176519": "generative grammar",
  "n06176837": "orthoepy",
  "n06177033": "phonetics",
  "n06177450": "phonology, phonemics",
  "n06177729": "morphology, sound structure, syllable structure, word structure",
  "n06177923": "affixation",
  "n06178042": "morphology",
  "n06178238": "inflectional morphology, accidence",
  "n06178380": "derivational morphology",
  "n06178510": "compound morphology",
  "n06178660": "morphophonemics",
  "n06178812": "lexicology",
  "n06178995": "onomastics",
  "n06179145": "toponymy, toponomy",
  "n06179290": "neurolinguistics",
  "n06179492": "pragmatics",
  "n06179574": "lexicostatistics",
  "n06179792": "semantics",
  "n06179985": "deixis",
  "n06180282": "formal semantics",
  "n06180412": "lexical semantics",
  "n06180548": "cognitive semantics, conceptual semantics, semasiology",
  "n06180720": "sound law",
  "n06180890": "Grimm's law",
  "n06181034": "Verner's law",
  "n06181123": "sociolinguistics",
  "n06181284": "structuralism, structural linguistics",
  "n06181448": "synchronic linguistics",
  "n06181584": "descriptive linguistics",
  "n06181893": "prescriptive linguistics",
  "n06182144": "theology, divinity",
  "n06182757": "angelology",
  "n06182866": "apologetics",
  "n06183004": "ecclesiology",
  "n06183162": "eschatology",
  "n06183398": "hermeneutics",
  "n06183518": "homiletics",
  "n06183652": "liturgics, liturgiology",
  "n06183745": "theodicy",
  "n06183899": "theology, theological system",
  "n06184270": "Christian theology",
  "n06184574": "Christology",
  "n06184733": "liberation theology",
  "n06184963": "natural theology",
  "n06185138": "Jesuitism, Jesuitry",
  "n06185302": "patristics, patrology",
  "n06185458": "polemics",
  "n06185581": "states' rights",
  "n06185748": "nullification",
  "n06185955": "teaching, precept, commandment",
  "n06186166": "mitzvah, mitsvah",
  "n06186301": "theological doctrine",
  "n06187013": "Christology",
  "n06187154": "antinomianism",
  "n06187350": "Thomism",
  "n06187529": "utilitarianism",
  "n06187756": "Arianism",
  "n06187960": "Athanasianism",
  "n06188146": "Boehmenism, Behmenism",
  "n06188292": "consubstantiation",
  "n06188603": "Episcopalianism",
  "n06188721": "Erastianism, Byzantinism, Caesaropapism",
  "n06188889": "Hinayanism",
  "n06188989": "Jansenism",
  "n06189241": "Mahayanism",
  "n06189341": "Marcionism",
  "n06189551": "millenarianism, millenarism, millenniumism, chiliasm",
  "n06189776": "Monophysitism",
  "n06190060": "Monothelitism",
  "n06190305": "Nestorianism",
  "n06190554": "Pelagianism",
  "n06190809": "Quakerism",
  "n06191018": "rationalism",
  "n06191212": "reincarnation",
  "n06191483": "Rosicrucianism",
  "n06191687": "soteriology",
  "n06191854": "synergism",
  "n06192027": "total depravity",
  "n06192186": "transcendentalism, transcendental philosophy",
  "n06192411": "transubstantiation",
  "n06192665": "universalism",
  "n06192789": "vertebrate paleontology",
  "n06192908": "Virgin Birth, Nativity",
  "n06193203": "attitude, mental attitude",
  "n06193727": "credence, acceptance",
  "n06194030": "fatalism",
  "n06194216": "recognition",
  "n06194409": "culture",
  "n06194736": "cyberculture",
  "n06194894": "Kalashnikov culture",
  "n06195096": "mosaic culture",
  "n06195249": "defensive, defensive attitude",
  "n06195418": "hardball",
  "n06195555": "high horse",
  "n06195698": "southernism",
  "n06195839": "mentality, outlook, mindset, mind-set",
  "n06196071": "paternalism",
  "n06196284": "position, stance, posture",
  "n06196442": "hard line",
  "n06196584": "inclination, disposition, tendency",
  "n06197215": "direction",
  "n06197503": "tenor",
  "n06197664": "drift, trend, movement",
  "n06197958": "evolutionary trend",
  "n06198083": "neoteny",
  "n06198313": "gravitation",
  "n06198505": "Call",
  "n06198708": "denominationalism",
  "n06198876": "devices",
  "n06199142": "sympathy, understanding",
  "n06199446": "favoritism, favouritism",
  "n06199561": "proclivity, propensity, leaning",
  "n06199702": "bent, set",
  "n06199893": "literalism",
  "n06200010": "perseveration",
  "n06200178": "predisposition",
  "n06200344": "predilection, preference, orientation",
  "n06200617": "favor, favour",
  "n06200741": "disfavor, disfavour, dislike, disapproval",
  "n06200929": "doghouse",
  "n06201042": "reprobation",
  "n06201136": "partiality, partisanship",
  "n06201476": "anthropocentrism, anthropocentricity",
  "n06201667": "ethnocentrism",
  "n06201778": "Eurocentrism",
  "n06201908": "bias, prejudice, preconception",
  "n06202296": "tilt",
  "n06202429": "sectionalism, provincialism, localism",
  "n06202553": "unfairness",
  "n06202686": "impartiality, nonpartisanship",
  "n06202907": "disinterestedness",
  "n06203030": "fairness, fair-mindedness, candor, candour",
  "n06203222": "experimenter bias",
  "n06203472": "homophobia",
  "n06203603": "Islamophobia",
  "n06203758": "racism",
  "n06203956": "anti-Semitism, antisemitism",
  "n06204126": "white supremacy",
  "n06204289": "tendentiousness",
  "n06204406": "tolerance",
  "n06204623": "broad-mindedness",
  "n06204820": "liberality, liberalness",
  "n06205018": "disinterest, neutrality",
  "n06205154": "intolerance",
  "n06205411": "narrow-mindedness, narrowness",
  "n06205698": "parochialism",
  "n06205827": "pettiness",
  "n06205937": "provincialism",
  "n06206021": "sectarianism, denominationalism",
  "n06206210": "bigotry, dogmatism",
  "n06206334": "fanaticism, fanatism, zealotry",
  "n06206470": "religionism",
  "n06206576": "zero tolerance",
  "n06206800": "respect, esteem, regard",
  "n06207029": "estimate, estimation",
  "n06207199": "reputation, report",
  "n06207437": "disrespect",
  "n06207561": "reverence",
  "n06207733": "irreverence",
  "n06207874": "profaneness",
  "n06208021": "orientation",
  "n06208265": "wavelength",
  "n06208409": "experimentalism",
  "n06208614": "reorientation",
  "n06208751": "position, view, perspective",
  "n06209101": "bird's eye view, panoramic view",
  "n06209242": "futurism",
  "n06209419": "vanguard, forefront, cutting edge",
  "n06209770": "cityscape",
  "n06209940": "landscape",
  "n06210171": "paradigm",
  "n06210363": "point of view, viewpoint, stand, standpoint",
  "n06210791": "light",
  "n06210964": "sight",
  "n06211078": "slant, angle",
  "n06211206": "complexion",
  "n06211389": "Weltanschauung, world view",
  "n06211529": "clockwork universe",
  "n06211702": "straddle",
  "n06211811": "orthodoxy",
  "n06211963": "conformity, conformism",
  "n06212155": "conventionality",
  "n06212302": "legalism",
  "n06212422": "unorthodoxy, heterodoxy, heresy",
  "n06212650": "nonconformity, nonconformism, nonconformance",
  "n06212839": "political orientation, ideology, political theory",
  "n06213688": "absolutism, totalitarianism, totalism",
  "n06213890": "anarchism",
  "n06214020": "autocracy",
  "n06214164": "Machiavellianism",
  "n06214379": "centrism, moderatism",
  "n06214580": "collectivism",
  "n06214744": "communism",
  "n06215043": "Castroism",
  "n06215152": "Leninism, Marxism-Leninism",
  "n06215495": "Maoism",
  "n06215618": "Marxism",
  "n06215977": "Trotskyism",
  "n06216160": "conservatism, conservativism",
  "n06216408": "neoconservatism",
  "n06216634": "reaction",
  "n06216805": "segregationism",
  "n06216948": "constitutionalism",
  "n06217103": "democracy",
  "n06217318": "social democracy",
  "n06217464": "domino theory",
  "n06217657": "elitism",
  "n06217806": "extremism",
  "n06217944": "fascism",
  "n06218162": "federalism",
  "n06218308": "imperialism",
  "n06218459": "leftism",
  "n06218623": "liberalism",
  "n06218824": "meritocracy",
  "n06219009": "neoliberalism",
  "n06219185": "libertarianism",
  "n06219300": "monarchism",
  "n06219415": "Negritude",
  "n06219613": "Orleanism",
  "n06219711": "progressivism",
  "n06219875": "radicalism",
  "n06220041": "Jacobinism",
  "n06220199": "reactionism",
  "n06220299": "republicanism",
  "n06220449": "rightism",
  "n06220616": "socialism",
  "n06220819": "Fabianism",
  "n06220955": "guild socialism",
  "n06221119": "utopian socialism",
  "n06221224": "theocracy",
  "n06221323": "Utopianism",
  "n06221485": "dovishness",
  "n06221659": "peace advocacy",
  "n06221790": "hawkishness",
  "n06221974": "militarism",
  "n06222236": "warmongering, war advocacy",
  "n06222336": "religious orientation",
  "n06222508": "agnosticism",
  "n06222731": "Docetism",
  "n06222959": "Gnosticism",
  "n06223204": "Mandaeanism, Mandeanism",
  "n06223468": "atheism, godlessness",
  "n06223669": "theism",
  "n06223922": "deism, free thought",
  "n06224136": "monotheism",
  "n06224305": "polytheism",
  "n06224439": "tritheism",
  "n06224657": "paganism, pagan religion, heathenism",
  "n06224831": "druidism",
  "n06224975": "pantheism",
  "n06225235": "pantheism",
  "n06225357": "cargo cult",
  "n06225505": "macumba",
  "n06225667": "obeah, obi",
  "n06225855": "Rastafarianism",
  "n06226057": "Christianity, Christian religion",
  "n06226934": "Adventism, Second Adventism",
  "n06227121": "Seventh-Day Adventism",
  "n06227263": "Catholicism, Catholicity",
  "n06227453": "anti-Catholicism",
  "n06227562": "Romanism, Roman Catholicism, papism",
  "n06227708": "Albigensianism, Catharism",
  "n06228086": "Donatism",
  "n06228346": "Eastern Catholicism",
  "n06228549": "Protestantism",
  "n06229106": "Anglicanism",
  "n06229240": "Anglo-Catholicism, High Anglicanism",
  "n06229410": "Tractarianism, Puseyism",
  "n06229586": "Arminianism",
  "n06229853": "Calvinism",
  "n06230060": "Christian Science",
  "n06230208": "Lutheranism",
  "n06230358": "Unitarianism",
  "n06230502": "Trinitarianism",
  "n06230613": "Congregationalism",
  "n06230809": "Mennonitism",
  "n06231030": "evangelicalism",
  "n06231191": "revivalism",
  "n06231328": "fundamentalism",
  "n06231494": "Methodism",
  "n06231680": "Wesleyanism, Wesleyism",
  "n06231794": "Anabaptism",
  "n06232047": "Baptistic doctrine",
  "n06232166": "Mormonism",
  "n06232298": "pentecostalism",
  "n06232499": "Presbyterianism",
  "n06232635": "Puritanism",
  "n06232880": "Judaism",
  "n06234104": "Orthodox Judaism",
  "n06234257": "Hasidism, Hassidism, Chasidism, Chassidism",
  "n06234415": "Chabad, Chabad Hasidism",
  "n06234619": "Conservative Judaism",
  "n06234728": "Reform Judaism",
  "n06234825": "Islam, Islamism, Mohammedanism, Muhammadanism, Muslimism",
  "n06235829": "Mahdism",
  "n06235977": "Salafism, Salafi movement",
  "n06236309": "Shiism",
  "n06236492": "Ismailism",
  "n06236602": "Wahhabism, Wahabism",
  "n06236802": "Hinduism, Hindooism",
  "n06237865": "Brahmanism, Brahminism",
  "n06238036": "Darsana",
  "n06238265": "Mimamsa",
  "n06238574": "Vedanta",
  "n06238831": "Krishnaism",
  "n06238931": "Shivaism, Sivaism",
  "n06239076": "Shaktism, Saktism",
  "n06239215": "Vaishnavism, Vaisnavism, Vishnuism",
  "n06239361": "yoga",
  "n06239655": "Jainism",
  "n06239931": "Sikhism",
  "n06240244": "Buddhism",
  "n06240865": "Mahayana, Mahayana Buddhism",
  "n06241156": "Theravada, Theravada Buddhism",
  "n06241576": "Hinayana, Hinayana Buddhism",
  "n06241825": "Lamaism, Tibetan Buddhism",
  "n06242048": "Zen, Zen Buddhism",
  "n06242237": "Shingon",
  "n06242427": "Tantra, Tantrism",
  "n06242780": "Yogacara",
  "n06242970": "Tao",
  "n06243096": "Taoism, Hsuan Chiao",
  "n06243347": "Shinto, Shintoism",
  "n06243575": "Manichaeism, Manichaeanism",
  "n06243963": "Mithraism, Mithraicism",
  "n06244149": "Zoroastrianism, Mazdaism",
  "n06244445": "Parsiism, Parseeism",
  "n06244552": "Bahaism",
  "n06244852": "shamanism, Asian shamanism",
  "n06245084": "shamanism",
  "n06245286": "Vedism",
  "n06245462": "Wicca",
  "n06245674": "obiism",
  "n06245816": "voodoo, vodoun, voodooism, hoodooism",
  "n06246076": "amateurism",
  "n06246241": "anagoge",
  "n06246361": "dynamical system",
  "n06246529": "chaos",
  "n06246700": "condensation",
  "n06246896": "level, layer, stratum",
  "n06247181": "transference",
  "n06247484": "countertransference",
  "n06247701": "restraint",
  "n06247867": "floodgate",
  "n06248043": "military science",
  "n06248214": "escapology",
  "n06248361": "graphology",
  "n06248530": "numerology",
  "n06248693": "protology",
  "n06248863": "theogony",
  "n06248968": "tactics",
  "n06249177": "strategy",
  "n06249421": "closure, law of closure",
  "n06249685": "common fate, law of common fate",
  "n06249910": "descriptivism",
  "n06250061": "descriptivism",
  "n06250208": "good continuation, continuation, law of continuation",
  "n06250444": "prescriptivism",
  "n06250597": "prescriptivism",
  "n06250771": "proximity, law of proximity",
  "n06251033": "similarity, law of similarity",
  "n06251295": "wrinkle",
  "n06251424": "wrinkle",
  "n06251553": "Zurvanism",
  "n06251781": "transmission",
  "n06252138": "communication, communicating",
  "n06252743": "intercommunication",
  "n06252954": "conveyance, imparting, impartation",
  "n06253140": "dissemination, airing, public exposure, spreading",
  "n06253371": "circulation",
  "n06253518": "propagation, extension",
  "n06253690": "message",
  "n06254007": "broadcast",
  "n06254239": "cipher, cypher",
  "n06254371": "heliogram",
  "n06254475": "medium",
  "n06254669": "medium",
  "n06254915": "ether, aether",
  "n06255081": "vehicle",
  "n06255354": "air, airwave",
  "n06255613": "paper",
  "n06255777": "sheet, piece of paper, sheet of paper",
  "n06256041": "signature",
  "n06256229": "leaf, folio",
  "n06256494": "flyleaf",
  "n06256591": "interleaf",
  "n06256697": "page",
  "n06257141": "tear sheet",
  "n06257250": "full page",
  "n06257372": "half page",
  "n06257484": "recto",
  "n06257569": "verso",
  "n06257653": "title page",
  "n06257772": "half title, bastard title",
  "n06257909": "sports page",
  "n06258031": "spread, spread head, spreadhead, facing pages",
  "n06258228": "center spread, centre spread",
  "n06258361": "centerfold, centrefold",
  "n06258541": "foldout, gatefold",
  "n06258680": "pagination, folio, page number, paging",
  "n06258852": "stationery, letter paper",
  "n06259027": "letterhead",
  "n06259166": "notepaper",
  "n06259297": "Post-It",
  "n06259487": "foolscap",
  "n06259584": "style sheet",
  "n06259736": "worksheet",
  "n06259898": "channel, transmission channel",
  "n06260121": "channel, communication channel, line",
  "n06260518": "band",
  "n06260628": "frequency band, waveband",
  "n06260817": "back channel",
  "n06261060": "lens",
  "n06261260": "liaison, link, contact, inter-group communication",
  "n06261464": "channels",
  "n06261586": "medium, mass medium",
  "n06261744": "multimedia, multimedia system",
  "n06261922": "hypermedia, hypermedia system, interactive multimedia, interactive multimedia system",
  "n06262176": "hypertext",
  "n06262567": "film, cinema, celluloid",
  "n06262943": "silver screen",
  "n06263020": "gutter press",
  "n06263202": "free press",
  "n06263369": "press, public press",
  "n06263609": "print media",
  "n06263762": "storage medium, data-storage medium",
  "n06263895": "magnetic storage medium, magnetic medium, magnetic storage",
  "n06264176": "broadcast medium, broadcasting",
  "n06264398": "mail, mail service, postal service, post",
  "n06264812": "airmail, airpost",
  "n06264932": "snail mail",
  "n06265142": "rural free delivery, RFD",
  "n06265272": "first class, 1st class, first-class mail, 1st-class mail",
  "n06265475": "express, express mail",
  "n06265636": "poste restante",
  "n06265815": "pony express",
  "n06265983": "parcel post",
  "n06266077": "bulk mail",
  "n06266296": "direct mail",
  "n06266417": "journalism, news media",
  "n06266633": "Fleet Street",
  "n06266710": "photojournalism",
  "n06266878": "news photography",
  "n06266973": "rotogravure",
  "n06267145": "newspaper, paper",
  "n06267564": "daily",
  "n06267655": "gazette",
  "n06267758": "school newspaper, school paper",
  "n06267893": "tabloid, rag, sheet",
  "n06267991": "yellow journalism, tabloid, tab",
  "n06268096": "article",
  "n06268357": "column",
  "n06268567": "column, editorial, newspaper column",
  "n06268784": "feature, feature article",
  "n06269014": "magazine article",
  "n06269130": "news article, news story, newspaper article",
  "n06269396": "piece",
  "n06269674": "morceau",
  "n06269785": "offprint, reprint, separate",
  "n06269956": "paper",
  "n06270129": "think piece",
  "n06270308": "reissue, reprint, reprinting",
  "n06270541": "new edition",
  "n06270690": "article of faith, credendum",
  "n06270879": "lead, lead-in, lede",
  "n06271055": "opening line",
  "n06271192": "lead, lead story",
  "n06271288": "personal",
  "n06271403": "sidebar",
  "n06271514": "agony column",
  "n06271620": "samizdat, underground press",
  "n06271778": "telecommunication, telecom",
  "n06272290": "telephone, telephony",
  "n06272612": "voice mail, voicemail",
  "n06272803": "call, phone call, telephone call",
  "n06273207": "call-back",
  "n06273294": "collect call",
  "n06273414": "call forwarding",
  "n06273555": "call-in",
  "n06273743": "call waiting",
  "n06273890": "crank call",
  "n06273986": "local call",
  "n06274092": "long distance, long-distance call, trunk call",
  "n06274292": "toll call",
  "n06274408": "conference call",
  "n06274546": "wake-up call",
  "n06274760": "three-way calling",
  "n06274921": "telegraphy",
  "n06275095": "cable, cablegram, overseas telegram",
  "n06275219": "letter telegram",
  "n06275353": "wireless",
  "n06275471": "radiotelegraph, radiotelegraphy, wireless telegraphy",
  "n06275634": "mail",
  "n06275977": "third-class mail, third class",
  "n06276141": "junk mail",
  "n06276287": "phone message, telephone message",
  "n06276401": "radiogram",
  "n06276501": "radiotelephone, radiotelephony, wireless telephone",
  "n06276697": "broadcasting",
  "n06276902": "Rediffusion",
  "n06277025": "multiplex",
  "n06277135": "radio, radiocommunication, wireless",
  "n06277280": "television, telecasting, TV, video",
  "n06277803": "video, picture",
  "n06277992": "video",
  "n06278136": "audio, sound",
  "n06278338": "cable television, cable",
  "n06278475": "high-definition television, HDTV",
  "n06278662": "electronic communication",
  "n06278830": "digital communication, data communication",
  "n06279071": "asynchronous transfer mode, ATM",
  "n06279326": "electronic mail, e-mail, email",
  "n06279805": "freemail",
  "n06279939": "emoticon",
  "n06280162": "smiley",
  "n06280244": "smoking gun",
  "n06280347": "spam, junk e-mail",
  "n06280477": "messaging, electronic messaging",
  "n06280604": "prompt, command prompt",
  "n06280816": "fiber optics, fiberoptics, fibre optics, fibreoptics",
  "n06281040": "reception",
  "n06281175": "signal detection, detection",
  "n06281295": "modulation",
  "n06281594": "amplitude modulation, AM",
  "n06281721": "frequency modulation, FM",
  "n06281848": "phase modulation, PM",
  "n06281959": "pulse modulation",
  "n06282093": "pulse-time modulation",
  "n06282209": "demodulation",
  "n06282383": "contagion, infection",
  "n06282651": "language, linguistic communication",
  "n06283459": "usage",
  "n06283647": "dead language",
  "n06283764": "words",
  "n06283912": "source language",
  "n06284067": "object language, target language",
  "n06284225": "language unit, linguistic unit",
  "n06284573": "slot",
  "n06284777": "discourse",
  "n06284898": "context, linguistic context, context of use",
  "n06285090": "sentence",
  "n06285435": "simple sentence",
  "n06285559": "complex sentence",
  "n06285789": "loose sentence",
  "n06285940": "periodic sentence",
  "n06286100": "compound sentence",
  "n06286230": "sentential function",
  "n06286395": "word",
  "n06287620": "anagram",
  "n06287859": "anaphor",
  "n06288024": "antonym, opposite word, opposite",
  "n06288350": "back-formation",
  "n06288527": "blend, portmanteau word, portmanteau",
  "n06288947": "charade",
  "n06289074": "cognate, cognate word",
  "n06289250": "content word, open-class word",
  "n06289472": "contraction",
  "n06289693": "deictic, deictic word",
  "n06290051": "derivative",
  "n06290246": "diminutive",
  "n06290401": "dirty word",
  "n06290539": "disyllable, dissyllable",
  "n06290637": "form, word form, signifier, descriptor",
  "n06291122": "four-letter word, four-letter Anglo-Saxon word",
  "n06291318": "function word, closed-class word",
  "n06291603": "guide word, guideword, catchword",
  "n06291809": "head, head word",
  "n06292000": "headword",
  "n06292154": "headword, head word",
  "n06292290": "heteronym",
  "n06292478": "holonym, whole name",
  "n06292649": "homonym",
  "n06292836": "hypernym, superordinate, superordinate word",
  "n06292973": "hyponym, subordinate, subordinate word",
  "n06293106": "key word",
  "n06293229": "loanblend, loan-blend, hybrid",
  "n06293460": "loanword, loan",
  "n06293655": "Latinism",
  "n06293746": "meronym, part name",
  "n06293898": "metonym",
  "n06294147": "antigram",
  "n06294309": "monosyllable, monosyllabic word",
  "n06294441": "neologism, neology, coinage",
  "n06294584": "nonce word, hapax legomenon",
  "n06294716": "oxytone",
  "n06294828": "palindrome",
  "n06294941": "primitive",
  "n06295113": "primitive",
  "n06295235": "plural, plural form",
  "n06299916": "singular, singular form",
  "n06300058": "ghost word",
  "n06300193": "root, root word, base, stem, theme, radical",
  "n06300445": "etymon, root",
  "n06300632": "citation form, main entry word, entry word",
  "n06300823": "lexical entry, dictionary entry",
  "n06301012": "Beatitude",
  "n06301302": "logion",
  "n06301445": "calque, calque formation, loan translation",
  "n06301672": "paroxytone",
  "n06301792": "partitive",
  "n06301953": "polysemant, polysemantic word, polysemous word",
  "n06302083": "polysyllable, polysyllabic word",
  "n06302269": "proparoxytone",
  "n06302381": "quantifier",
  "n06302542": "quantifier, logical quantifier",
  "n06302755": "existential quantifier, existential operator",
  "n06302967": "universal quantifier",
  "n06303150": "reduplication",
  "n06303315": "retronym",
  "n06303540": "substantive",
  "n06303682": "synonym, equivalent word",
  "n06303888": "term",
  "n06304059": "terminology, nomenclature, language",
  "n06304338": "trisyllable",
  "n06304425": "troponym, manner name",
  "n06304576": "vocable, spoken word",
  "n06304671": "syllable",
  "n06305029": "ultima",
  "n06305109": "penult, penultima, penultimate",
  "n06305223": "antepenult, antepenultima, antepenultimate",
  "n06305385": "jawbreaker",
  "n06305474": "sesquipedalian, sesquipedalia",
  "n06305610": "reduplication",
  "n06305716": "direct antonym",
  "n06305838": "indirect antonym",
  "n06306034": "lexeme",
  "n06306233": "morpheme",
  "n06306480": "formative",
  "n06306606": "allomorph",
  "n06306842": "free morpheme, free form",
  "n06306945": "bound morpheme, bound form",
  "n06307152": "combining form",
  "n06308049": "affix",
  "n06308304": "prefix",
  "n06308436": "classifier",
  "n06308674": "alpha privative",
  "n06308765": "ending, termination",
  "n06308987": "suffix, postfix",
  "n06309131": "inflectional ending, inflectional suffix",
  "n06309272": "infix",
  "n06309383": "grammatical category, syntactic category",
  "n06309684": "substitution class, paradigm",
  "n06309931": "subject",
  "n06310125": "subject",
  "n06310237": "object",
  "n06310449": "prepositional object, object of a preposition",
  "n06310578": "direct object, object of the verb",
  "n06310713": "indirect object",
  "n06310848": "retained object",
  "n06310945": "case, grammatical case",
  "n06311160": "nominative, nominative case, subject case",
  "n06311334": "oblique, oblique case",
  "n06311557": "accusative, accusative case, objective case",
  "n06311723": "dative, dative case",
  "n06311852": "genitive, genitive case, possessive, possessive case",
  "n06312001": "attributive genitive, attributive genitive case",
  "n06312237": "vocative, vocative case",
  "n06312418": "ablative, ablative case",
  "n06312612": "ablative absolute",
  "n06312767": "adjunct",
  "n06312966": "constituent, grammatical constituent",
  "n06313300": "immediate constituent",
  "n06313457": "syntagma, syntagm",
  "n06313651": "construction, grammatical construction, expression",
  "n06314036": "misconstruction",
  "n06314144": "clause",
  "n06314423": "main clause, independent clause",
  "n06314595": "coordinate clause",
  "n06314808": "subordinate clause, dependent clause",
  "n06315096": "relative clause",
  "n06315301": "restrictive clause",
  "n06315454": "nonrestrictive clause, descriptive clause",
  "n06315638": "complement",
  "n06315755": "involution",
  "n06315873": "parenthetical expression, parenthetical",
  "n06316048": "phrase",
  "n06316376": "predicator",
  "n06316480": "noun phrase, nominal phrase, nominal",
  "n06316626": "predicate, verb phrase",
  "n06316813": "predicate",
  "n06317107": "split infinitive",
  "n06317247": "prepositional phrase",
  "n06317351": "pronominal phrase, pronominal",
  "n06317464": "part of speech, form class, word class",
  "n06317672": "major form class",
  "n06317862": "noun",
  "n06318062": "verb",
  "n06318352": "gerund",
  "n06318513": "auxiliary verb",
  "n06318722": "modal auxiliary verb, modal auxiliary, modal verb, modal",
  "n06318904": "infinitive",
  "n06319029": "adjective",
  "n06319157": "adverb",
  "n06319293": "noun",
  "n06319575": "collective noun",
  "n06319710": "mass noun",
  "n06319799": "count noun",
  "n06319881": "generic noun",
  "n06320004": "proper noun, proper name",
  "n06320153": "common noun",
  "n06320314": "verbal noun, deverbal noun",
  "n06320440": "adnoun",
  "n06320569": "verb",
  "n06320801": "modifier, qualifier",
  "n06321054": "intensifier, intensive",
  "n06321702": "adjective",
  "n06321932": "descriptive adjective, qualifying adjective",
  "n06322157": "relational adjective, classifying adjective",
  "n06322357": "pertainym",
  "n06322468": "positive, positive degree",
  "n06322693": "comparative, comparative degree",
  "n06323284": "superlative, superlative degree",
  "n06323612": "adverb",
  "n06323822": "dangling modifier, misplaced modifier",
  "n06324086": "dangling participle",
  "n06324355": "adverbial",
  "n06324475": "determiner, determinative",
  "n06324669": "article",
  "n06324864": "definite article",
  "n06325001": "indefinite article",
  "n06325145": "preposition",
  "n06325370": "pronoun",
  "n06325576": "anaphoric pronoun",
  "n06325678": "demonstrative pronoun, demonstrative",
  "n06325826": "conjunction, conjunctive, connective, continuative",
  "n06326054": "coordinating conjunction",
  "n06326226": "subordinating conjunction, subordinate conjunction",
  "n06326406": "particle",
  "n06326526": "number",
  "n06326797": "person",
  "n06327136": "personal pronoun",
  "n06327261": "reciprocal pronoun",
  "n06327551": "relative pronoun",
  "n06327718": "first person",
  "n06327870": "second person",
  "n06328022": "third person",
  "n06328207": "reflexive pronoun, reflexive",
  "n06328386": "reflexive verb",
  "n06328643": "gender, grammatical gender",
  "n06328996": "feminine",
  "n06329165": "masculine",
  "n06329313": "neuter",
  "n06329506": "tense",
  "n06329734": "present, present tense",
  "n06329909": "historical present",
  "n06330037": "aorist",
  "n06330262": "past, past tense",
  "n06330401": "future, future tense",
  "n06330528": "participle, participial",
  "n06330764": "phrasal verb",
  "n06331014": "present participle",
  "n06331154": "past participle, perfect participle",
  "n06331281": "transitive verb, transitive verb form, transitive",
  "n06331477": "doubly transitive verb, doubly transitive verb form",
  "n06331641": "intransitive verb, intransitive verb form, intransitive",
  "n06331803": "semantic role, participant role",
  "n06332136": "affected role, patient role, patient",
  "n06332364": "agentive role, agent",
  "n06332545": "benefactive role, beneficiary",
  "n06332731": "instrumental role, instrument",
  "n06332919": "locative role, locative",
  "n06333095": "recipient role, recipient",
  "n06333285": "resultant role, result",
  "n06333478": "temporal role, temporal",
  "n06333653": "name",
  "n06334377": "agnomen",
  "n06334512": "assumed name, fictitious name, Doing Business As, DBA",
  "n06334778": "eponym",
  "n06334985": "eponym",
  "n06335162": "extension, filename extension, file name extension",
  "n06335532": "filename, file name, computer filename, computer file name",
  "n06335832": "patronymic, patronym",
  "n06336149": "matronymic, metronymic",
  "n06336285": "street name",
  "n06336363": "street name",
  "n06336537": "street name",
  "n06336723": "street name",
  "n06336904": "surname, family name, cognomen, last name",
  "n06337111": "maiden name",
  "n06337202": "middle name",
  "n06337307": "first name, given name, forename",
  "n06337458": "Christian name, baptismal name",
  "n06337594": "praenomen",
  "n06337693": "nickname, moniker, cognomen, sobriquet, soubriquet, byname",
  "n06338003": "nickname",
  "n06338158": "alias, assumed name, false name",
  "n06338278": "pseudonym, anonym, nom de guerre",
  "n06338485": "misnomer",
  "n06338571": "stage name",
  "n06338653": "pen name, nom de plume",
  "n06338744": "writer's name, author's name",
  "n06338908": "appellation, denomination, designation, appellative",
  "n06339244": "pet name, hypocorism",
  "n06339416": "title, title of respect, form of address",
  "n06340047": "Aga, Agha",
  "n06340182": "Defender of the Faith",
  "n06340395": "Don",
  "n06340563": "Dona",
  "n06340707": "Frau",
  "n06340838": "Fraulein",
  "n06340977": "Hakham",
  "n06341127": "Herr",
  "n06341249": "Miss",
  "n06341340": "Mister, Mr, Mr.",
  "n06341431": "Mrs, Mrs.",
  "n06341525": "Ms, Ms.",
  "n06341609": "Rabbi",
  "n06341773": "Reverend",
  "n06341862": "Senor",
  "n06342015": "Senora",
  "n06342182": "Senorita",
  "n06342362": "Signora",
  "n06342490": "Signorina",
  "n06342623": "Very Reverend",
  "n06342781": "Lordship",
  "n06342977": "Ladyship",
  "n06343117": "title",
  "n06343364": "baronetcy",
  "n06343440": "viscountcy",
  "n06343520": "title",
  "n06343838": "place name, toponym",
  "n06343971": "heading, header, head",
  "n06344329": "crossheading, crosshead",
  "n06344461": "headline, newspaper headline",
  "n06344691": "lemma",
  "n06344846": "masthead",
  "n06344998": "rubric",
  "n06345131": "running head, running headline",
  "n06345320": "subheading, subhead",
  "n06345441": "running title",
  "n06345566": "dropline, drop line, stepped line, stagger head, staggered head, stephead",
  "n06345773": "screamer",
  "n06345878": "streamer, banner",
  "n06345993": "title, statute title, rubric",
  "n06346220": "title",
  "n06346461": "title",
  "n06346681": "credit",
  "n06346891": "caption, legend",
  "n06347037": "subtitle",
  "n06347122": "mistranslation",
  "n06347225": "pony, trot, crib",
  "n06347388": "retroversion",
  "n06347588": "subtitle, caption",
  "n06347811": "supertitle, surtitle",
  "n06347996": "line of poetry, line of verse",
  "n06348215": "acatalectic",
  "n06348373": "Alexandrine",
  "n06348500": "catalectic",
  "n06348685": "hypercatalectic",
  "n06348885": "by-line, credit line",
  "n06349030": "dateline",
  "n06349220": "written communication, written language, black and white",
  "n06349597": "transcription, written text",
  "n06349828": "transliteration",
  "n06349952": "phonetic transcription",
  "n06350127": "shorthand, stenography, tachygraphy",
  "n06350274": "longhand, running hand, cursive, cursive script",
  "n06350592": "minuscule",
  "n06350777": "copperplate",
  "n06350918": "italic",
  "n06351086": "round hand",
  "n06351202": "orthography, writing system",
  "n06351613": "script",
  "n06351888": "Aramaic, Aramaic script",
  "n06352117": "Armenian, Armenian alphabet",
  "n06352301": "Avestan",
  "n06352446": "Babylonian",
  "n06352618": "Brahmi",
  "n06352782": "Devanagari, Devanagari script, Nagari, Nagari script",
  "n06352954": "Pahlavi",
  "n06353089": "Uighur, Uigur, Uygur",
  "n06353225": "uncial",
  "n06353445": "spelling",
  "n06353653": "misspelling",
  "n06353757": "coding system",
  "n06353934": "code",
  "n06354204": "access, access code",
  "n06354481": "back door, backdoor",
  "n06354626": "area code",
  "n06354774": "bar code, Universal Product Code",
  "n06355078": "color code",
  "n06355183": "cryptogram, cryptograph, secret writing",
  "n06355307": "cipher, cypher, cryptograph, secret code",
  "n06355459": "Morse, Morse code, international Morse code",
  "n06355705": "ZIP code, ZIP, postcode, postal code",
  "n06355894": "code, computer code",
  "n06356299": "argument, parameter",
  "n06356515": "address, computer address, reference",
  "n06356755": "American Standard Code for Information Interchange, ASCII",
  "n06357078": "ASCII character set",
  "n06357304": "binary code",
  "n06357424": "error correction code, ECC",
  "n06357632": "cyclic redundancy check",
  "n06357814": "firmware, microcode",
  "n06357985": "machine code, machine language",
  "n06358159": "object code",
  "n06358311": "operation code, order code",
  "n06358513": "source code",
  "n06358747": "URL, uniform resource locator, universal resource locator",
  "n06358900": "web page, webpage",
  "n06359096": "home page, homepage",
  "n06359193": "web site, website, internet site, site",
  "n06359467": "chat room, chatroom",
  "n06359657": "portal site, portal",
  "n06359877": "writing",
  "n06360314": "written word",
  "n06360599": "bigram",
  "n06360723": "trigram",
  "n06360850": "tetragram",
  "n06360996": "Tetragrammaton",
  "n06361218": "picture writing",
  "n06361314": "alphabetic writing, alphabetic script",
  "n06361446": "boustrophedon",
  "n06361635": "cuneiform",
  "n06361770": "syllabary, syllabic script",
  "n06361934": "Linear A",
  "n06362059": "Linear B",
  "n06362172": "ideography",
  "n06362260": "hieratic, hieratic script",
  "n06362441": "hieroglyph, hieroglyphic",
  "n06362628": "point system",
  "n06362792": "braille",
  "n06362953": "writing, written material, piece of writing",
  "n06363778": "writing",
  "n06364004": "patristics, patrology",
  "n06364149": "rewrite, revision, rescript",
  "n06364329": "literary composition, literary work",
  "n06364641": "literature",
  "n06365102": "literature",
  "n06365357": "historiography",
  "n06365467": "matter",
  "n06365808": "prescription",
  "n06366002": "prescription",
  "n06366130": "acrostic",
  "n06366269": "belles-lettres, belles lettres",
  "n06366391": "dialogue, dialog",
  "n06366581": "allegory",
  "n06366879": "euphuism",
  "n06367107": "fiction",
  "n06367373": "fictionalization, fictionalisation",
  "n06367571": "nonfiction, nonfictional prose",
  "n06367702": "dystopia",
  "n06367879": "novel",
  "n06368142": "detective novel, mystery novel",
  "n06368321": "dime novel, penny dreadful",
  "n06368425": "fantasy, phantasy",
  "n06368628": "science fiction",
  "n06368814": "cyberpunk",
  "n06368962": "novelette, novella",
  "n06369077": "roman a clef",
  "n06369216": "romance",
  "n06369405": "Gothic romance",
  "n06369554": "bodice ripper",
  "n06369690": "roman fleuve",
  "n06369829": "story",
  "n06370178": "utopia",
  "n06370268": "adventure story, heroic tale",
  "n06370403": "thriller",
  "n06370522": "saga",
  "n06370792": "mystery, mystery story, whodunit",
  "n06370985": "detective story",
  "n06371145": "murder mystery",
  "n06371267": "love story, romance",
  "n06371413": "legend, fable",
  "n06371734": "Arthurian legend",
  "n06371999": "short story",
  "n06372095": "fable, parable, allegory, apologue",
  "n06372428": "Aesop's fables",
  "n06372572": "Pilgrim's Progress",
  "n06372680": "myth",
  "n06372925": "Gotterdammerung, Ragnarok, Twilight of the Gods",
  "n06373090": "parable",
  "n06373314": "plot",
  "n06373513": "action",
  "n06373645": "storyline, plot line",
  "n06373747": "climax, culmination",
  "n06373991": "anticlimax, bathos",
  "n06374148": "tearjerker",
  "n06374241": "interior monologue",
  "n06374408": "stream of consciousness",
  "n06374587": "criticism, literary criticism",
  "n06374834": "explication de texte",
  "n06375008": "textual criticism",
  "n06375215": "new criticism",
  "n06375329": "higher criticism",
  "n06375471": "lower criticism",
  "n06375624": "Masorah, Masora",
  "n06376014": "analysis",
  "n06376154": "drama",
  "n06376572": "prose",
  "n06376776": "prose poem",
  "n06376860": "polyphonic prose",
  "n06377000": "hagiology",
  "n06377133": "lucubration",
  "n06377275": "pastoral",
  "n06377442": "poem, verse form",
  "n06377971": "abecedarius",
  "n06378104": "Alcaic, Alcaic verse",
  "n06378298": "ballad, lay",
  "n06378427": "ballade",
  "n06378524": "blank verse",
  "n06378627": "clerihew",
  "n06378917": "couplet",
  "n06379094": "dithyramb",
  "n06379253": "doggerel, doggerel verse, jingle",
  "n06379439": "eclogue, bucolic, idyll, idyl",
  "n06379568": "elegy, lament",
  "n06379721": "epic poem, heroic poem, epic, epos",
  "n06380055": "Aeneid",
  "n06380256": "Divine Comedy, Divina Commedia",
  "n06380373": "free verse, vers libre",
  "n06380495": "haiku",
  "n06380603": "limerick",
  "n06380726": "lyric, lyric poem",
  "n06380879": "rondeau, rondel",
  "n06381134": "roundel",
  "n06381274": "rondelet",
  "n06381372": "sonnet",
  "n06381594": "tanka",
  "n06381761": "terza rima",
  "n06381869": "verse, rhyme",
  "n06382072": "Iliad",
  "n06382193": "Odyssey",
  "n06382345": "Nibelungenlied",
  "n06382506": "chanson de geste",
  "n06382590": "rhapsody",
  "n06382716": "Petrarchan sonnet, Italian sonnet",
  "n06382923": "octave",
  "n06383016": "sestet",
  "n06383107": "Shakespearean sonnet, Elizabethan sonnet, English sonnet",
  "n06383334": "Spenserian sonnet",
  "n06383523": "epos",
  "n06383659": "ode",
  "n06383819": "epithalamium",
  "n06383934": "Horatian ode, Sapphic ode",
  "n06384034": "Pindaric ode, Pindaric",
  "n06384165": "choral ode",
  "n06384269": "canto",
  "n06384371": "envoy, envoi",
  "n06384480": "quatrain",
  "n06384593": "elegiac stanza",
  "n06384708": "verse, verse line",
  "n06384995": "iambic",
  "n06385080": "Adonic, Adonic line",
  "n06385244": "versicle",
  "n06385434": "sursum corda",
  "n06385582": "response",
  "n06385728": "closed couplet",
  "n06385842": "heroic couplet",
  "n06385997": "heroic stanza",
  "n06386156": "heroic verse, heroic meter, heroic",
  "n06386345": "mock-heroic",
  "n06386440": "Spenserian stanza",
  "n06386693": "strophe",
  "n06386832": "antistrophe",
  "n06387053": "potboiler",
  "n06387204": "tushery",
  "n06387332": "dictation",
  "n06387538": "cookie",
  "n06387727": "session cookie",
  "n06387864": "precision cookie",
  "n06387980": "text, textual matter",
  "n06388579": "text",
  "n06388777": "machine-displayable text",
  "n06388918": "machine-readable text",
  "n06389109": "typescript",
  "n06389230": "erasure",
  "n06389398": "margin",
  "n06389553": "space, blank space, place",
  "n06389753": "indentation, indention, indent, indenture",
  "n06389947": "word order",
  "n06390051": "core dump",
  "n06390227": "dump",
  "n06390423": "fair copy",
  "n06390512": "copy, written matter",
  "n06390688": "front matter, prelims",
  "n06390805": "back matter, end matter",
  "n06390962": "draft, draft copy",
  "n06391171": "electronic text",
  "n06391347": "soft copy",
  "n06391566": "hard copy",
  "n06391766": "installment, instalment",
  "n06391902": "fascicle, fascicule",
  "n06392001": "section, subdivision",
  "n06392660": "above",
  "n06392787": "sports section",
  "n06392935": "article, clause",
  "n06393253": "arbitration clause",
  "n06393424": "deductible",
  "n06393620": "double indemnity",
  "n06393800": "escalator clause, escalator",
  "n06394051": "joker",
  "n06394282": "reserve clause",
  "n06394564": "rider",
  "n06394701": "body",
  "n06394865": "book",
  "n06396142": "chapter",
  "n06396330": "episode",
  "n06396468": "spot",
  "n06396644": "spot",
  "n06396812": "insert",
  "n06396930": "introduction",
  "n06397171": "exordium",
  "n06397307": "narration",
  "n06397476": "opening",
  "n06397645": "teaser",
  "n06397789": "salutation",
  "n06397903": "foreword, preface, prolusion",
  "n06398090": "preamble",
  "n06398270": "prolegomenon",
  "n06398401": "conclusion, end, close, closing, ending",
  "n06398760": "epilogue, epilog",
  "n06398963": "epilogue, epilog",
  "n06399126": "peroration",
  "n06399337": "appendix",
  "n06399503": "sequel, continuation",
  "n06399631": "addendum, supplement, postscript",
  "n06399870": "shirttail",
  "n06399995": "paragraph",
  "n06400271": "passage",
  "n06400510": "excerpt, excerption, extract, selection",
  "n06400849": "chrestomathy",
  "n06401004": "locus classicus",
  "n06401107": "place",
  "n06401222": "purple passage",
  "n06401328": "transition",
  "n06401526": "flashback",
  "n06401787": "flash-forward",
  "n06402031": "diary, journal",
  "n06402202": "web log, blog",
  "n06402425": "capitalization, capitalisation",
  "n06402565": "typing, typewriting",
  "n06402770": "double-spacing",
  "n06402871": "single-spacing",
  "n06402970": "triple-spacing",
  "n06403107": "touch typing, touch system",
  "n06403291": "printing",
  "n06403393": "handwriting, hand, script",
  "n06403679": "hieroglyph, hieroglyphic",
  "n06403852": "skywriting",
  "n06403969": "calligraphy, penmanship, chirography",
  "n06404147": "scribble, scratch, scrawl, cacography",
  "n06404322": "chicken scratch",
  "n06404416": "squiggle",
  "n06404582": "signature",
  "n06404806": "allograph",
  "n06404907": "autograph, John Hancock",
  "n06405020": "countersignature, countersign",
  "n06405198": "endorsement, indorsement",
  "n06405423": "blank endorsement, endorsement in blank",
  "n06405589": "sign manual",
  "n06405699": "inscription, lettering",
  "n06405891": "Rosetta Stone",
  "n06406177": "superscription",
  "n06406317": "dedication, inscription",
  "n06406547": "epigraph",
  "n06406625": "epitaph",
  "n06406757": "epitaph",
  "n06406865": "festschrift",
  "n06406979": "manuscript, ms",
  "n06407094": "autograph",
  "n06407221": "manuscript, holograph",
  "n06407372": "codex, leaf-book",
  "n06407514": "palimpsest",
  "n06407733": "scroll, roll",
  "n06407915": "Dead Sea scrolls",
  "n06408239": "Megillah",
  "n06408442": "Torah",
  "n06408651": "treatise",
  "n06408779": "adaptation, version",
  "n06408990": "modernization",
  "n06409085": "dissertation, thesis",
  "n06409290": "tract, pamphlet",
  "n06409448": "monograph",
  "n06409562": "essay",
  "n06409752": "composition, paper, report, theme",
  "n06409937": "term paper",
  "n06410070": "disquisition",
  "n06410187": "memoir",
  "n06410283": "thanatopsis",
  "n06410391": "review, critique, critical review, review article",
  "n06410655": "book review",
  "n06410776": "notice",
  "n06410904": "book",
  "n06411592": "authority",
  "n06411753": "curiosa",
  "n06411883": "formulary, pharmacopeia",
  "n06412213": "last word",
  "n06412350": "trade book, trade edition",
  "n06412497": "best seller, bestseller",
  "n06412610": "bestiary",
  "n06412771": "catechism",
  "n06413020": "cookbook, cookery book",
  "n06413130": "instruction book",
  "n06413260": "pop-up book, pop-up",
  "n06413459": "storybook",
  "n06413579": "tome",
  "n06413666": "volume",
  "n06413889": "booklet, brochure, folder, leaflet, pamphlet",
  "n06414151": "blue book",
  "n06414267": "ticket book",
  "n06414372": "textbook, text, text edition, schoolbook, school text",
  "n06414727": "crammer",
  "n06414849": "introduction",
  "n06414948": "primer",
  "n06415061": "reader",
  "n06415186": "McGuffey Eclectic Readers",
  "n06415323": "speller",
  "n06415419": "notebook",
  "n06415584": "commonplace book",
  "n06415688": "jotter",
  "n06415791": "workbook",
  "n06415922": "copybook",
  "n06416049": "appointment book, appointment calendar",
  "n06416206": "catalog, catalogue",
  "n06416452": "phrase book",
  "n06416598": "playbook",
  "n06416766": "playbook",
  "n06416946": "prayer book, prayerbook",
  "n06417096": "breviary",
  "n06417279": "missal",
  "n06417467": "Psalter, Book of Psalms",
  "n06417598": "reference book, reference, reference work, book of facts",
  "n06418006": "review copy",
  "n06418146": "songbook",
  "n06418258": "hymnal, hymnbook, hymnary",
  "n06418375": "prayer wheel",
  "n06418558": "source book",
  "n06418693": "wordbook",
  "n06418901": "dictionary, lexicon",
  "n06419234": "bilingual dictionary",
  "n06419354": "desk dictionary, collegiate dictionary",
  "n06419503": "etymological dictionary",
  "n06419628": "gazetteer",
  "n06419757": "learner's dictionary, school dictionary",
  "n06419912": "pocket dictionary, little dictionary",
  "n06420053": "spell-checker, spelling checker",
  "n06420219": "unabridged dictionary, unabridged",
  "n06420424": "Oxford English Dictionary, O.E.D., OED",
  "n06420573": "onomasticon",
  "n06420678": "vocabulary",
  "n06420781": "glossary, gloss",
  "n06421016": "thesaurus, synonym finder",
  "n06421154": "word finder, wordfinder",
  "n06421301": "handbook, enchiridion, vade mecum",
  "n06421538": "hornbook",
  "n06421685": "manual",
  "n06421844": "consuetudinary, consuetudinal",
  "n06422032": "grimoire",
  "n06422144": "instruction manual, instructions, book of instructions, operating instructions",
  "n06422364": "reference manual",
  "n06422488": "sex manual",
  "n06422643": "bible",
  "n06422740": "guidebook, guide",
  "n06422912": "field guide",
  "n06423090": "roadbook",
  "n06423233": "baedeker",
  "n06423378": "travel guidebook, itinerary",
  "n06423496": "reckoner, ready reckoner",
  "n06423619": "directory",
  "n06423754": "phonebook, phone book, telephone book, telephone directory",
  "n06423994": "ballistic identification, ballistic fingerprinting, bullet fingerprinting",
  "n06424275": "biometric identification, biometric authentication, identity verification",
  "n06424869": "key",
  "n06425065": "number, identification number",
  "n06425404": "business card",
  "n06425534": "bank identification number, BIN, ABA transit number",
  "n06425808": "license number, registration number",
  "n06425960": "Social Security number",
  "n06426111": "phone number, telephone number, number",
  "n06426279": "almanac",
  "n06426468": "annual, yearly, yearbook",
  "n06426655": "almanac, farmer's calendar",
  "n06426865": "ephemeris",
  "n06427086": "atlas, book of maps, map collection",
  "n06427240": "dialect atlas, linguistic atlas",
  "n06427387": "encyclopedia, cyclopedia, encyclopaedia, cyclopaedia",
  "n06427710": "book of knowledge",
  "n06427831": "editing, redaction",
  "n06428105": "copy editing",
  "n06428216": "deletion, excision, cut",
  "n06428537": "correction",
  "n06428646": "erasure",
  "n06428792": "rewriting, revising",
  "n06428976": "revision, revisal, revise, rescript",
  "n06429145": "rewording, recasting, rephrasing",
  "n06429316": "paraphrase, paraphrasis",
  "n06429483": "translation",
  "n06429590": "sacred text, sacred writing, religious writing, religious text",
  "n06430309": "screed",
  "n06430385": "scripture, sacred scripture",
  "n06430537": "canon",
  "n06430784": "Adi Granth, Granth, Granth Sahib",
  "n06430996": "Avesta, Zend-Avesta",
  "n06431156": "Bhagavad-Gita, Bhagavadgita, Gita",
  "n06431496": "Mahabharata, Mahabharatam, Mahabharatum",
  "n06431740": "Bible, Christian Bible, Book, Good Book, Holy Scripture, Holy Writ, Scripture, Word of God, Word",
  "n06432376": "Genesis, Book of Genesis",
  "n06432715": "Exodus, Book of Exodus",
  "n06433035": "Leviticus, Book of Leviticus",
  "n06433249": "Numbers, Book of Numbers",
  "n06433475": "Deuteronomy, Book of Deuteronomy",
  "n06433672": "mezuzah, mezuza",
  "n06433923": "Joshua, Josue, Book of Joshua",
  "n06434165": "Judges, Book of Judges",
  "n06434368": "Ruth, Book of Ruth",
  "n06434650": "I Samuel, 1 Samuel",
  "n06434826": "II Samuel, 2 Samuel",
  "n06435004": "I Kings, 1 Kings",
  "n06435198": "II Kings, 2 Kings",
  "n06435394": "Paralipomenon",
  "n06435651": "I Chronicles, 1 Chronicles",
  "n06435916": "II Chronicles, 2 Chronicles",
  "n06436183": "Ezra, Book of Ezra",
  "n06436443": "Nehemiah, Book of Nehemiah",
  "n06436717": "Esther, Book of Esther",
  "n06436939": "Job, Book of Job",
  "n06437137": "Psalms, Book of Psalms",
  "n06437308": "Proverbs, Book of Proverbs",
  "n06437531": "Ecclesiastes, Book of Ecclesiastes",
  "n06437824": "Song of Songs, Song of Solomon, Canticle of Canticles, Canticles",
  "n06438126": "Isaiah, Book of Isaiah",
  "n06438290": "Jeremiah, Book of Jeremiah",
  "n06438477": "Lamentations, Book of Lamentations",
  "n06438748": "Ezekiel, Ezechiel, Book of Ezekiel",
  "n06438995": "Daniel, Book of Daniel, Book of the Prophet Daniel",
  "n06439253": "Hosea, Book of Hosea",
  "n06439408": "Joel, Book of Joel",
  "n06439560": "Amos, Book of Amos",
  "n06439712": "Obadiah, Abdias, Book of Obadiah",
  "n06439924": "Jonah, Book of Jonah",
  "n06440102": "Micah, Micheas, Book of Micah",
  "n06440313": "Nahum, Book of Nahum",
  "n06440489": "Habakkuk, Habacuc, Book of Habakkuk",
  "n06440663": "Zephaniah, Sophonias, Book of Zephaniah",
  "n06440937": "Haggai, Aggeus, Book of Haggai",
  "n06441195": "Zechariah, Zacharias, Book of Zachariah",
  "n06441464": "Malachi, Malachias, Book of Malachi",
  "n06441607": "Matthew, Gospel According to Matthew",
  "n06441803": "Mark, Gospel According to Mark",
  "n06441973": "Luke, Gospel of Luke, Gospel According to Luke",
  "n06442239": "John, Gospel According to John",
  "n06442405": "Acts of the Apostles, Acts",
  "n06442616": "Epistle",
  "n06443163": "Epistle of Paul the Apostle to the Romans, Epistle to the Romans, Romans",
  "n06443398": "First Epistle of Paul the Apostle to the Corinthians, First Epistle to the Corinthians, I Corinthians",
  "n06443658": "Second Epistle of Paul the Apostle to the Corinthians, Second Epistle to the Corinthians, II Corinthians",
  "n06443922": "Epistle of Paul the Apostle to the Galatians, Epistle to the Galatians, Galatians",
  "n06444148": "Epistle of Paul the Apostle to the Ephesians, Epistle to the Ephesians, Ephesians",
  "n06444458": "Epistle of Paul the Apostle to the Philippians, Epistle to the Philippians, Philippians",
  "n06444711": "Epistle of Paul the Apostle to the Colossians, Epistle to the Colossians, Colossians",
  "n06444959": "First Epistle of Paul the Apostle to the Thessalonians, First Epistle to the Thessalonians, I Thessalonians",
  "n06445214": "Second Epistle of Paul the Apostle to the Thessalonians, Second Epistle to the Thessalonians, II Thessalonians",
  "n06445473": "First Epistle of Paul the Apostle to Timothy, First Epistle to Timothy, I Timothy",
  "n06445729": "Second Epistle of Paul the Apostle to Timothy, Second Epistle to Timothy, II Timothy",
  "n06445989": "Epistle of Paul the Apostle to Titus, Epistle to Titus, Titus",
  "n06446217": "Epistle of Paul the Apostle to Philemon, Epistle to Philemon, Philemon",
  "n06446476": "Epistle to the Hebrews, Hebrews",
  "n06446711": "Epistle of James, James",
  "n06446860": "First Epistle of Peter, I Peter",
  "n06447039": "Second Epistle of Peter, II Peter",
  "n06447221": "First Epistle of John, I John",
  "n06447400": "Second Epistel of John, II John",
  "n06447582": "Third Epistel of John, III John",
  "n06447763": "Epistle of Jude, Jude",
  "n06447897": "Revelation, Revelation of Saint John the Divine, Apocalypse, Book of Revelation",
  "n06448283": "family Bible",
  "n06448397": "Septuagint",
  "n06448594": "Vulgate",
  "n06448868": "Douay Bible, Douay Version, Douay-Rheims Bible, Douay-Rheims Version, Rheims-Douay Bible, Rheims-Douay Version",
  "n06449095": "Authorized Version, King James Version, King James Bible",
  "n06449254": "Revised Version",
  "n06449361": "New English Bible",
  "n06449477": "American Standard Version, American Revised Version",
  "n06449620": "Revised Standard Version",
  "n06449735": "Old Testament",
  "n06451891": "Torah, Pentateuch, Laws",
  "n06452223": "Torah",
  "n06452363": "Tanakh, Tanach, Hebrew Scripture",
  "n06452601": "Prophets, Nebiim",
  "n06453119": "Haftorah, Haftarah, Haphtorah, Haphtarah",
  "n06453324": "Hagiographa, Ketubim, Writings",
  "n06453723": "Testament",
  "n06453849": "New Testament",
  "n06455138": "Gospel, Gospels, evangel",
  "n06455497": "Synoptic Gospels, Synoptics",
  "n06455682": "Word of God",
  "n06455775": "Book of Mormon",
  "n06455990": "prayer",
  "n06456276": "Agnus Dei",
  "n06456384": "Angelus",
  "n06456515": "Ave Maria, Hail Mary",
  "n06456642": "Canticle of Simeon, Nunc dimittis",
  "n06456759": "Evening Prayer, evensong",
  "n06456945": "Kol Nidre",
  "n06457045": "service book",
  "n06457171": "Book of Common Prayer",
  "n06457442": "Litany",
  "n06457612": "Lord's Prayer",
  "n06457796": "Paternoster",
  "n06457952": "Apocrypha",
  "n06458685": "Additions to Esther",
  "n06458836": "Prayer of Azariah and Song of the Three Children",
  "n06459016": "Susanna, Book of Susanna",
  "n06459173": "Bel and the Dragon",
  "n06459323": "Baruch, Book of Baruch",
  "n06459450": "Letter of Jeremiah, Epistle of Jeremiah",
  "n06459681": "Tobit, Book of Tobit",
  "n06459834": "Judith, Book of Judith",
  "n06459978": "I Esdra, 1 Esdras",
  "n06460167": "II Esdras, 2 Esdras",
  "n06460295": "Ben Sira, Sirach, Ecclesiasticus, Wisdom of Jesus the Son of Sirach",
  "n06460524": "Wisdom of Solomon, Wisdom",
  "n06460776": "I Maccabees, 1 Maccabees",
  "n06460926": "II Maccabees, 2 Maccabees",
  "n06461077": "sapiential book, wisdom book, wisdom literature",
  "n06461406": "Pseudepigrapha",
  "n06461609": "Koran, Quran, al-Qur'an, Book",
  "n06461830": "sura",
  "n06462002": "Fatiha, Fatihah",
  "n06462219": "Talmudic literature",
  "n06462396": "Talmud",
  "n06462656": "Gemara",
  "n06462807": "Mishna, Mishnah",
  "n06463011": "Haggadah, Haggada, Hagada",
  "n06463170": "Halakah, Halaka, Halacha",
  "n06463347": "Sanskrit literature",
  "n06463559": "Hastinapura",
  "n06463786": "Purana",
  "n06464024": "Ramayana",
  "n06464241": "tantra",
  "n06464419": "Vedic literature, Veda",
  "n06464838": "Samhita",
  "n06465008": "Rig-Veda",
  "n06465139": "Sama-Veda",
  "n06465255": "Atharva-Veda",
  "n06465351": "Yajur-Veda",
  "n06465457": "Brahmana",
  "n06465623": "Aranyaka",
  "n06465780": "Vedanga",
  "n06466030": "Ayurveda",
  "n06466253": "Upanishad",
  "n06466479": "mantra",
  "n06466677": "psalm",
  "n06466787": "Psalm",
  "n06467007": "summary, sum-up",
  "n06467445": "summarization, summarisation",
  "n06467680": "argument, literary argument",
  "n06467868": "capitulation",
  "n06467996": "compendium",
  "n06468123": "condensation, abridgement, abridgment, capsule",
  "n06468328": "conspectus",
  "n06468403": "curriculum vitae, CV, resume",
  "n06468523": "line score",
  "n06468640": "brief",
  "n06468749": "apercu",
  "n06468818": "epitome",
  "n06468951": "outline, synopsis, abstract, precis",
  "n06469223": "overview",
  "n06469377": "recapitulation, recap, review",
  "n06469597": "roundup",
  "n06469694": "sketch, survey, resume",
  "n06469874": "summation, summing up, rundown",
  "n06470073": "document, written document, papers",
  "n06470688": "articles of incorporation",
  "n06470922": "ballot",
  "n06471069": "brevet",
  "n06471242": "capitulation",
  "n06471345": "certificate, certification, credential, credentials",
  "n06471737": "charter",
  "n06472025": "commercial document, commercial instrument",
  "n06472242": "confession",
  "n06472409": "confession",
  "n06472607": "Augsburg Confession",
  "n06472824": "copula, copulative, linking verb",
  "n06473041": "frequentative",
  "n06473168": "copyright, right of first publication",
  "n06473381": "enclosure, inclosure",
  "n06473563": "form",
  "n06473837": "application form",
  "n06473940": "claim form",
  "n06474030": "order form",
  "n06474122": "questionnaire",
  "n06474289": "personality inventory, personality assessment",
  "n06474603": "self-report personality inventory, self-report inventory",
  "n06474972": "California Personality Inventory, CPI",
  "n06475307": "Eysenck Personality Inventory, EPI",
  "n06475589": "Minnesota Multiphasic Personality Inventory, MMPI",
  "n06475933": "Sixteen Personality Factor Questionnaire, 16 PF",
  "n06476200": "requisition, requisition form",
  "n06476384": "tax form",
  "n06476475": "telegraph form",
  "n06476573": "absentee ballot",
  "n06476741": "certificate of incorporation",
  "n06476899": "bank charter",
  "n06477003": "Magna Carta, Magna Charta, The Great Charter",
  "n06477209": "royal charter",
  "n06477371": "card, identity card",
  "n06477645": "donor card",
  "n06477818": "keycard",
  "n06477970": "membership card",
  "n06478097": "union card",
  "n06478199": "library card, borrower's card",
  "n06478331": "ration card",
  "n06478452": "birth certificate",
  "n06478582": "diploma, sheepskin",
  "n06478734": "Higher National Diploma, HND",
  "n06478988": "commission, military commission",
  "n06479204": "bill of health",
  "n06479374": "registration",
  "n06479494": "teaching certificate, teacher's certificate",
  "n06479665": "legal document, legal instrument, official document, instrument",
  "n06480506": "derivative instrument, derivative",
  "n06480723": "futures contract",
  "n06480991": "stock-index futures",
  "n06481156": "negotiable instrument",
  "n06481320": "list, listing",
  "n06482401": "item, point",
  "n06482868": "agenda item",
  "n06482978": "incidental",
  "n06483061": "inventory item",
  "n06483171": "line item",
  "n06483339": "news item",
  "n06483454": "place, position",
  "n06483702": "postposition",
  "n06483992": "preposition",
  "n06484279": "topicalization",
  "n06484696": "ammunition",
  "n06484887": "factoid",
  "n06485001": "factoid",
  "n06485178": "papyrus",
  "n06485261": "agenda, agendum, order of business",
  "n06485431": "A-list",
  "n06485593": "docket",
  "n06485800": "order of the day",
  "n06485933": "order paper, order book",
  "n06486042": "network programming",
  "n06486161": "batting order, card, lineup",
  "n06486405": "cleanup, cleanup position, cleanup spot",
  "n06486630": "bibliography",
  "n06486874": "bill",
  "n06487050": "bill of entry",
  "n06487172": "bill of goods",
  "n06487260": "blacklist, black book, shitlist",
  "n06487395": "calendar",
  "n06487610": "calorie chart",
  "n06487729": "canon",
  "n06487897": "catalog, catalogue",
  "n06488224": "discography",
  "n06488325": "library catalog, library catalogue",
  "n06488473": "card catalog, card catalogue",
  "n06488629": "parts catalog, parts catalogue",
  "n06488763": "seed catalog, seed catalogue",
  "n06488880": "character set",
  "n06489070": "checklist",
  "n06489190": "class list, honours list",
  "n06489410": "clericalism",
  "n06489560": "codex",
  "n06489659": "contents, table of contents",
  "n06489847": "corrigenda",
  "n06489968": "credits",
  "n06490173": "criminal record, record",
  "n06490451": "directory",
  "n06490627": "distribution list",
  "n06490743": "subdirectory",
  "n06490887": "enumeration, numbering",
  "n06491026": "FAQ",
  "n06491168": "free list",
  "n06491277": "grocery list",
  "n06491371": "grocery list, shopping list",
  "n06491561": "hit list",
  "n06491665": "hit parade",
  "n06491786": "index",
  "n06492040": "concordance",
  "n06492188": "key",
  "n06492320": "key word",
  "n06492438": "key",
  "n06492569": "parts inventory",
  "n06492664": "inventory, stock list",
  "n06492811": "mailing list",
  "n06492939": "menu, bill of fare, card, carte du jour, carte",
  "n06493158": "masthead, flag",
  "n06493392": "menu, computer menu",
  "n06493579": "drop-down menu",
  "n06493721": "hierarchical menu, cascading menu, submenu",
  "n06493926": "necrology",
  "n06494016": "playlist, play list",
  "n06494153": "portfolio",
  "n06494373": "posting",
  "n06494538": "price list",
  "n06494646": "push-down list, push-down stack, stack",
  "n06494816": "queue",
  "n06495000": "roll, roster",
  "n06495220": "death-roll",
  "n06495328": "schedule",
  "n06495516": "shopping list",
  "n06495625": "short list, shortlist",
  "n06495824": "sick list",
  "n06495948": "slate, ticket",
  "n06496116": "standing",
  "n06496299": "wish list",
  "n06496397": "timetable",
  "n06496499": "timetable",
  "n06496624": "muster roll",
  "n06496773": "church roll",
  "n06496862": "rota",
  "n06497013": "waiting list",
  "n06497117": "a la carte",
  "n06497233": "prix fixe",
  "n06497331": "table d'hote",
  "n06497459": "alphabet",
  "n06497872": "Roman alphabet, Latin alphabet",
  "n06498569": "Hebrew alphabet, Hebraic alphabet, Hebrew script",
  "n06499244": "Greek alphabet",
  "n06499796": "Cyrillic alphabet, Cyrillic",
  "n06500062": "Arabic alphabet",
  "n06500262": "alphanumerics, alphanumeric characters",
  "n06500448": "phonetic alphabet, sound alphabet",
  "n06500639": "visible speech",
  "n06500765": "manual alphabet, finger alphabet",
  "n06500937": "passport",
  "n06501141": "patent, patent of invention",
  "n06501311": "platform, political platform, political program, program",
  "n06501622": "plank",
  "n06501748": "ship's papers",
  "n06501918": "manifest",
  "n06502054": "push-down queue",
  "n06502192": "cadaster, cadastre",
  "n06502378": "written record, written account",
  "n06502858": "blotter, day book, police blotter, rap sheet, charge sheet",
  "n06503034": "casebook",
  "n06503224": "chronology",
  "n06503386": "Domesday Book, Doomsday Book",
  "n06503551": "dossier",
  "n06503724": "entry",
  "n06503884": "log",
  "n06504049": "logbook",
  "n06504155": "log",
  "n06504326": "bell book",
  "n06504462": "note",
  "n06504662": "paper trail",
  "n06504838": "timecard",
  "n06504965": "timeline",
  "n06505154": "time sheet",
  "n06505252": "nolle prosequi, nol pros",
  "n06505434": "notebook entry",
  "n06505517": "transcript, copy",
  "n06505705": "memorabilia",
  "n06505799": "jotting, jot",
  "n06505935": "marginalia",
  "n06506037": "scholium, scholia",
  "n06506191": "memo, memorandum, memoranda",
  "n06506315": "minute",
  "n06506430": "aide-memoire, position paper",
  "n06506603": "corker",
  "n06506757": "reminder",
  "n06506926": "check register",
  "n06507041": "register, registry",
  "n06507330": "studbook",
  "n06507454": "rent-roll",
  "n06507592": "won-lost record",
  "n06507711": "blue book",
  "n06507815": "stub, check stub, counterfoil",
  "n06507941": "card, scorecard",
  "n06508112": "minutes, proceedings, transactions",
  "n06508299": "minute book",
  "n06508417": "Congressional Record",
  "n06508579": "Hansard",
  "n06508816": "file, data file",
  "n06509075": "Combined DNA Index System",
  "n06509210": "computer file",
  "n06509513": "backup file",
  "n06509675": "binary file",
  "n06509902": "master file, main file",
  "n06510103": "disk file",
  "n06510271": "transaction file, detail file",
  "n06510478": "input file, input data",
  "n06510663": "output file",
  "n06510836": "read-only file",
  "n06510977": "text file, document",
  "n06511242": "ASCII text file",
  "n06511396": "mug file, mug book",
  "n06511560": "resignation",
  "n06511762": "abdication, stepping down",
  "n06511874": "resolution, declaration, resolve",
  "n06512099": "Declaration of Independence",
  "n06512324": "joint resolution",
  "n06512580": "application",
  "n06512857": "job application",
  "n06512943": "credit application",
  "n06513043": "loan application",
  "n06513154": "mortgage application",
  "n06513255": "patent application",
  "n06513366": "request, petition, postulation",
  "n06513624": "memorial",
  "n06513764": "solicitation, appeal, collection, ingathering",
  "n06513953": "whip-round",
  "n06514093": "history, account, chronicle, story",
  "n06514529": "ancient history",
  "n06514621": "etymology",
  "n06514786": "folk etymology",
  "n06514880": "case history",
  "n06515054": "family history",
  "n06515297": "medical history, medical record, anamnesis",
  "n06515489": "historical document, historical paper, historical record",
  "n06515662": "annals, chronological record",
  "n06515827": "biography, life, life story, life history",
  "n06516087": "autobiography",
  "n06516242": "hagiography",
  "n06516404": "profile",
  "n06516495": "memoir",
  "n06516595": "statement, financial statement",
  "n06516782": "bank statement",
  "n06516955": "bill, account, invoice",
  "n06517345": "electric bill",
  "n06517447": "hotel bill",
  "n06517547": "medical bill, doctor's bill",
  "n06517663": "phone bill, telephone bill",
  "n06517779": "reckoning, tally",
  "n06517867": "tax bill",
  "n06517942": "check, chit, tab",
  "n06518068": "coupon, voucher",
  "n06518253": "book token",
  "n06518420": "meal ticket, luncheon voucher",
  "n06518565": "twofer",
  "n06518719": "ticket",
  "n06519077": "commutation ticket, season ticket",
  "n06519253": "plane ticket, airplane ticket",
  "n06519369": "pass",
  "n06519480": "transfer",
  "n06519607": "railroad ticket, train ticket",
  "n06519728": "theater ticket, theatre ticket",
  "n06519846": "bus ticket",
  "n06519936": "round-trip ticket, return ticket",
  "n06520092": "day return",
  "n06520222": "receipt",
  "n06520429": "stub, ticket stub",
  "n06520571": "rain check",
  "n06520742": "bill of lading, waybill",
  "n06520944": "contract",
  "n06521878": "adhesion contract, contract of adhesion",
  "n06522119": "aleatory contract",
  "n06522357": "bilateral contract",
  "n06522501": "charter",
  "n06522633": "conditional contract",
  "n06522784": "cost-plus contract",
  "n06522941": "gambling contract",
  "n06523132": "lease",
  "n06523353": "marriage contract, marriage settlement",
  "n06523473": "output contract",
  "n06523641": "policy, insurance policy, insurance",
  "n06523842": "purchase contract, purchase agreement",
  "n06523969": "quasi contract",
  "n06524111": "requirements contract",
  "n06524278": "sealed instrument, contract under seal, special contract",
  "n06524454": "service contract",
  "n06524552": "severable contract",
  "n06524767": "subcontract",
  "n06524935": "conspiracy, confederacy",
  "n06525132": "fair-trade agreement",
  "n06525406": "conspiracy of silence",
  "n06525588": "covenant",
  "n06525813": "unilateral contract",
  "n06526004": "debenture",
  "n06526124": "floater, floating policy",
  "n06526291": "partnership",
  "n06526486": "articles of agreement, shipping articles",
  "n06526619": "concession, grant",
  "n06526811": "franchise",
  "n06526961": "labor contract, labor agreement, collective agreement",
  "n06527170": "yellow-dog contract",
  "n06527320": "employment contract, employment agreement",
  "n06527447": "distribution agreement",
  "n06527576": "licensing agreement",
  "n06527710": "merger agreement, acquisition agreement",
  "n06527851": "sale, sales agreement",
  "n06528191": "conditional sale",
  "n06528403": "sale in gross, contract of hazard",
  "n06528557": "sheriff's sale, execution sale, judicial sale, forced sale",
  "n06528783": "appraisal, estimate, estimation",
  "n06528992": "overestimate, overestimation, overvaluation, overappraisal",
  "n06529219": "order, purchase order",
  "n06529630": "credit order, bill-me order",
  "n06529798": "open account",
  "n06529879": "indent",
  "n06529996": "market order",
  "n06530143": "production order",
  "n06530255": "reorder",
  "n06530412": "stop order, stop-loss order",
  "n06530586": "stop payment",
  "n06530703": "mail order",
  "n06530789": "power of attorney",
  "n06530976": "stock power",
  "n06531141": "proxy",
  "n06531327": "stock symbol",
  "n06531481": "letters of administration",
  "n06531657": "letters testamentary",
  "n06531908": "working papers, work papers, work permit",
  "n06532095": "act, enactment",
  "n06532330": "law",
  "n06532763": "nullity",
  "n06532910": "anti-drug law",
  "n06533039": "anti-racketeering law, Racketeer Influenced and Corrupt Organizations Act, RICO Act, RICO",
  "n06533297": "antitrust legislation, antitrust law",
  "n06533484": "statute of limitations",
  "n06533648": "fundamental law, organic law, constitution",
  "n06533863": "Articles of Confederation",
  "n06534132": "United States Constitution, U.S. Constitution, US Constitution, Constitution, Constitution of the United States",
  "n06534548": "public law",
  "n06534659": "Roman law, Justinian code, civil law, jus civile",
  "n06534918": "Salic law",
  "n06535035": "case law, precedent, common law",
  "n06535222": "legislation, statute law",
  "n06535476": "enabling legislation",
  "n06535652": "occupational safety and health act, federal job safety law",
  "n06535980": "advice and consent",
  "n06536227": "statute book",
  "n06536389": "translation, interlingual rendition, rendering, version",
  "n06536737": "worksheet",
  "n06536853": "bill, measure",
  "n06537124": "appropriation bill",
  "n06537323": "bill of attainder",
  "n06537562": "bottle bill",
  "n06537701": "farm bill",
  "n06537832": "trade bill",
  "n06537951": "bylaw",
  "n06538105": "blue law",
  "n06538215": "blue sky law",
  "n06538412": "gag law",
  "n06538525": "game law",
  "n06538655": "homestead law",
  "n06538785": "poor law",
  "n06538895": "Riot Act",
  "n06539076": "riot act",
  "n06539178": "criminal law",
  "n06539502": "court order",
  "n06539770": "decree, edict, fiat, order, rescript",
  "n06540284": "consent decree",
  "n06540527": "curfew",
  "n06540702": "decree nisi",
  "n06540863": "divestiture",
  "n06541167": "imperial decree",
  "n06541301": "ukase",
  "n06541381": "legal separation, judicial separation",
  "n06541578": "pragmatic sanction, pragmatic",
  "n06541726": "programma",
  "n06541820": "prohibition",
  "n06542047": "prohibition, ban, proscription",
  "n06542267": "stay",
  "n06542569": "stay of execution",
  "n06542742": "banning-order",
  "n06542830": "injunction, enjoining, enjoinment, cease and desist order",
  "n06543246": "mandatory injunction",
  "n06543389": "permanent injunction, final injunction",
  "n06543536": "temporary injunction, interlocutory injunction",
  "n06543781": "brief, legal brief",
  "n06543949": "amicus curiae brief",
  "n06544142": "will, testament",
  "n06544432": "probate, probate will",
  "n06544657": "codicil",
  "n06544841": "living will",
  "n06545137": "deed, deed of conveyance, title",
  "n06545528": "assignment",
  "n06545728": "bill of sale",
  "n06545843": "deed poll",
  "n06545960": "enfeoffment",
  "n06546157": "mortgage deed",
  "n06546261": "title deed",
  "n06546408": "trust deed, deed of trust",
  "n06546633": "conveyance",
  "n06546783": "quitclaim, quitclaim deed",
  "n06546931": "muniments",
  "n06547059": "warrant",
  "n06547321": "search warrant",
  "n06547615": "bench warrant, arrest warrant",
  "n06547832": "pickup",
  "n06547992": "death warrant",
  "n06548110": "cachet, lettre de cachet",
  "n06548317": "reprieve",
  "n06548498": "commutation",
  "n06548671": "tax return, income tax return, return",
  "n06548991": "amended return",
  "n06549115": "declaration of estimated tax, estimated tax return",
  "n06549318": "false return",
  "n06549407": "information return",
  "n06549566": "joint return",
  "n06549661": "license, licence, permit",
  "n06550046": "building permit",
  "n06550206": "driver's license, driver's licence, driving license, driving licence",
  "n06550381": "fishing license, fishing licence, fishing permit",
  "n06550552": "hunting license, hunting licence, hunting permit, game license",
  "n06550763": "learner's permit",
  "n06550891": "letter of marque, letters of marque, letter of mark and reprisal",
  "n06551070": "liquor license, liquor licence",
  "n06551226": "on-license",
  "n06551339": "marriage license, marriage licence, wedding license, wedding licence",
  "n06551498": "occupation license, occupation licence",
  "n06551627": "patent, letters patent",
  "n06551784": "opinion, legal opinion, judgment, judgement",
  "n06552116": "concurring opinion",
  "n06552320": "dissenting opinion",
  "n06552470": "majority opinion",
  "n06552639": "pardon, amnesty",
  "n06552814": "acquittance, release",
  "n06552984": "writ, judicial writ",
  "n06553472": "assize",
  "n06553631": "certiorari, writ of certiorari",
  "n06553846": "execution, writ of execution",
  "n06554078": "execution, execution of instrument",
  "n06554373": "habeas corpus, writ of habeas corpus",
  "n06554531": "venire facias",
  "n06554675": "mandamus, writ of mandamus",
  "n06554981": "attachment",
  "n06555191": "fieri facias",
  "n06555341": "scire facias",
  "n06555570": "sequestration",
  "n06555695": "writ of detinue",
  "n06555847": "writ of election",
  "n06555971": "writ of error",
  "n06556150": "writ of prohibition",
  "n06556343": "writ of right",
  "n06556481": "mandate, authorization, authorisation",
  "n06556692": "summons, process",
  "n06557047": "subpoena, subpoena ad testificandum",
  "n06557317": "subpoena duces tecum",
  "n06557585": "gag order",
  "n06557827": "garnishment",
  "n06558088": "interdict, interdiction",
  "n06558277": "citation",
  "n06558434": "monition, process of monition",
  "n06558678": "ticket",
  "n06558893": "speeding ticket",
  "n06559004": "parking ticket",
  "n06559114": "bill of Particulars",
  "n06559365": "pleading",
  "n06559832": "affirmative pleading",
  "n06560021": "alternative pleading, pleading in the alternative",
  "n06560254": "answer",
  "n06560758": "evasive answer",
  "n06560939": "nolo contendere, non vult",
  "n06561138": "plea",
  "n06561343": "counterplea",
  "n06561461": "dilatory plea",
  "n06561744": "insanity plea, plea of insanity",
  "n06561942": "charge, complaint",
  "n06562217": "complaint",
  "n06562436": "libel",
  "n06562615": "defective pleading",
  "n06562802": "demurrer",
  "n06562993": "rebutter, rebuttal",
  "n06563168": "replication",
  "n06563324": "rejoinder",
  "n06563478": "special pleading",
  "n06563640": "surrebutter, surrebuttal",
  "n06563801": "surrejoinder",
  "n06563950": "plea bargain, plea bargaining",
  "n06564387": "legislative act, statute",
  "n06564640": "fair-trade act",
  "n06564887": "Stamp Act",
  "n06565201": "enabling act, enabling clause",
  "n06565397": "Foreign Intelligence Surveillance Act, FISA",
  "n06565788": "ordinance",
  "n06565919": "special act",
  "n06566077": "software, software program, computer software, software system, software package, package",
  "n06566805": "alpha software",
  "n06566949": "authoring language",
  "n06567143": "beta software",
  "n06567400": "compatible software",
  "n06567531": "compatible software",
  "n06567689": "computer-aided design, CAD",
  "n06567865": "freeware",
  "n06567960": "groupware",
  "n06568134": "operating system, OS",
  "n06568422": "DOS, disk operating system",
  "n06568552": "MS-DOS, Microsoft disk operating system",
  "n06568706": "UNIX, UNIX system, UNIX operating system",
  "n06568855": "Linux",
  "n06568978": "program, programme, computer program, computer programme",
  "n06569966": "anti-virus program",
  "n06570110": "application, application program, applications programme",
  "n06570520": "active application",
  "n06570647": "applet",
  "n06570805": "frame",
  "n06570963": "binary, binary program",
  "n06571301": "browser, web browser",
  "n06571538": "Internet Explorer, Explorer, IE",
  "n06571639": "Konqueror",
  "n06571724": "lynx",
  "n06571790": "Mosaic",
  "n06571862": "Netscape",
  "n06571938": "Opera",
  "n06572011": "natural language processor, natural language processing application",
  "n06572204": "disambiguator",
  "n06572482": "job",
  "n06572654": "word processor, word processing system",
  "n06572857": "loop",
  "n06573020": "malevolent program",
  "n06573223": "patch",
  "n06573337": "assembler, assembly program",
  "n06573472": "checking program",
  "n06573600": "compiler, compiling program",
  "n06573934": "C compiler",
  "n06574027": "Fortran compiler",
  "n06574132": "LISP compiler",
  "n06574231": "Pascal compiler",
  "n06574334": "debugger",
  "n06574473": "driver, device driver",
  "n06574680": "diagnostic program",
  "n06574841": "editor program, editor",
  "n06575092": "input program",
  "n06575227": "interface, user interface",
  "n06575505": "command line interface, CLI",
  "n06575681": "graphical user interface, GUI",
  "n06575932": "interpreter, interpretive program",
  "n06576153": "job control",
  "n06576265": "library program",
  "n06576376": "linkage editor",
  "n06576539": "monitor program, monitoring program",
  "n06576727": "text editor",
  "n06576893": "object program, target program",
  "n06577063": "source program",
  "n06577232": "output program",
  "n06577369": "parser",
  "n06577585": "tagger, tagging program",
  "n06577781": "sense tagger",
  "n06577916": "part-of-speech tagger, pos tagger",
  "n06578060": "relocatable program",
  "n06578203": "reusable program",
  "n06578323": "Web Map Service, Web Map Server",
  "n06578527": "MapQuest",
  "n06578654": "search engine",
  "n06578905": "Google",
  "n06579127": "Yahoo",
  "n06579294": "Ask Jeeves",
  "n06579452": "self-adapting program",
  "n06579592": "snapshot program",
  "n06579715": "spider, wanderer",
  "n06579952": "spreadsheet",
  "n06580103": "sort program, sorting program",
  "n06580219": "stored program",
  "n06580351": "supervisory program, supervisor, executive program",
  "n06580543": "syntax checker",
  "n06580646": "system program, systems program, systems software",
  "n06580866": "trace program",
  "n06581068": "text-matching",
  "n06581268": "translator, translating program",
  "n06581410": "utility program, utility, service program",
  "n06581823": "Windows",
  "n06581959": "decision table",
  "n06582085": "flow chart, flowchart, flow diagram, flow sheet",
  "n06582267": "logic diagram, logical diagram",
  "n06582403": "routine, subroutine, subprogram, procedure, function",
  "n06582761": "call",
  "n06582986": "function call",
  "n06583178": "cataloged procedure",
  "n06583354": "contingency procedure",
  "n06583518": "dump routine",
  "n06583645": "input routine",
  "n06583790": "library routine",
  "n06583929": "output routine",
  "n06584049": "random number generator",
  "n06584162": "recursive routine",
  "n06584256": "reusable routine",
  "n06584376": "supervisory routine, executive routine",
  "n06584536": "tracing routine",
  "n06584702": "utility routine, service routine",
  "n06584891": "instruction, command, statement, program line",
  "n06585203": "logic bomb, slag code",
  "n06585552": "trojan, trojan horse",
  "n06585816": "virus, computer virus",
  "n06586098": "worm",
  "n06586355": "command line",
  "n06586471": "link",
  "n06586672": "hyperlink",
  "n06586886": "macro, macro instruction",
  "n06587048": "system error",
  "n06587217": "system call, supervisor call instruction",
  "n06587399": "toggle",
  "n06587596": "shareware",
  "n06587790": "shrink-wrapped software",
  "n06587980": "spyware",
  "n06588139": "supervisory software",
  "n06588326": "software documentation, documentation",
  "n06588511": "electronic database, on-line database, computer database, electronic information service",
  "n06588785": "database management system, DBMS",
  "n06588998": "relational database management system",
  "n06589151": "object-oriented database management system",
  "n06589352": "hypertext system",
  "n06589574": "publication",
  "n06590065": "read",
  "n06590210": "impression, printing",
  "n06590446": "edition",
  "n06590601": "limited edition",
  "n06590723": "variorum, variorum edition",
  "n06590885": "proof, test copy, trial impression",
  "n06591108": "galley proof",
  "n06591224": "foundry proof",
  "n06591342": "mackle",
  "n06591442": "collection, compendium",
  "n06591609": "anthology",
  "n06591815": "album, record album",
  "n06592078": "concept album",
  "n06592281": "rock opera",
  "n06592421": "tribute album, benefit album",
  "n06592576": "divan, diwan",
  "n06592699": "florilegium, garland, miscellany",
  "n06592844": "omnibus",
  "n06592991": "archives",
  "n06593099": "compilation, digest",
  "n06593296": "periodical",
  "n06593542": "digest",
  "n06593668": "pictorial",
  "n06593803": "series, serial, serial publication",
  "n06594130": "semiweekly",
  "n06594258": "weekly",
  "n06594376": "semimonthly",
  "n06594505": "monthly",
  "n06594625": "quarterly",
  "n06594751": "bimonthly",
  "n06594904": "biweekly",
  "n06595056": "organ",
  "n06595220": "house organ",
  "n06595351": "magazine, mag",
  "n06595797": "tip sheet",
  "n06596025": "dope sheet, scratch sheet",
  "n06596179": "colour supplement",
  "n06596364": "comic book",
  "n06596474": "news magazine",
  "n06596607": "pulp, pulp magazine",
  "n06596727": "slick, slick magazine, glossy",
  "n06596845": "trade magazine",
  "n06596978": "issue, number",
  "n06597204": "edition",
  "n06597349": "extra",
  "n06597466": "journal",
  "n06597617": "annals",
  "n06597758": "review",
  "n06597930": "literary review",
  "n06598030": "reading, reading material",
  "n06598244": "bumf, bumph",
  "n06598445": "perusal, perusing, poring over, studying",
  "n06598648": "browse, browsing",
  "n06598746": "skim, skimming",
  "n06598915": "message, content, subject matter, substance",
  "n06599655": "latent content",
  "n06599788": "subject, topic, theme",
  "n06600139": "bone of contention",
  "n06600282": "precedent",
  "n06600421": "didacticism",
  "n06600684": "digression, aside, excursus, divagation, parenthesis",
  "n06600903": "declarative sentence, declaratory sentence",
  "n06601053": "run-on sentence",
  "n06601217": "topic sentence",
  "n06601327": "meaning, significance, signification, import",
  "n06601973": "lexical meaning",
  "n06602148": "grammatical meaning",
  "n06602324": "symbolization, symbolisation",
  "n06602472": "sense, signified",
  "n06602801": "word meaning, word sense, acceptation",
  "n06602935": "intension, connotation",
  "n06603130": "referent",
  "n06603242": "referent",
  "n06603366": "relatum",
  "n06603494": "referent",
  "n06603651": "antecedent",
  "n06603816": "denotatum",
  "n06603927": "designatum",
  "n06604066": "effect, essence, burden, core, gist",
  "n06604209": "alpha and omega",
  "n06604319": "ambiguity",
  "n06604548": "loophole",
  "n06604743": "amphibology, amphiboly",
  "n06604963": "parisology",
  "n06605046": "euphemism",
  "n06605396": "dysphemism",
  "n06605682": "shucks",
  "n06605780": "double entendre",
  "n06605897": "intent, purport, spirit",
  "n06606044": "moral, lesson",
  "n06606191": "nuance, nicety, shade, subtlety, refinement",
  "n06606464": "overtone",
  "n06606618": "bottom line",
  "n06606694": "crux, crux of the matter",
  "n06606808": "point",
  "n06607067": "rallying point",
  "n06607204": "talking point",
  "n06607339": "nonsense, bunk, nonsensicality, meaninglessness, hokum",
  "n06607809": "absurdity, absurdness, ridiculousness",
  "n06608035": "amphigory, nonsense verse",
  "n06608143": "balderdash, fiddle-faddle, piffle",
  "n06608277": "buzzword, cant",
  "n06608405": "cobblers",
  "n06608525": "crock",
  "n06608617": "fa la, fal la",
  "n06608728": "gibberish, gibber",
  "n06608977": "incoherence, incoherency, unintelligibility",
  "n06609176": "word salad",
  "n06609296": "jabberwocky",
  "n06609403": "mummery, flummery",
  "n06609503": "palaver, hot air, empty words, empty talk, rhetoric",
  "n06609672": "rigmarole, rigamarole",
  "n06609785": "shmegegge, schmegegge",
  "n06609909": "stuff, stuff and nonsense, hooey, poppycock",
  "n06610063": "abracadabra",
  "n06610143": "babble, babbling, lallation",
  "n06610332": "blather, blatherskite",
  "n06610436": "double Dutch",
  "n06610557": "bill of goods",
  "n06610779": "humbug, snake oil",
  "n06610897": "double talk",
  "n06610992": "jabber, jabbering, gabble",
  "n06611147": "baloney, boloney, bilgewater, bosh, drool, humbug, taradiddle, tarradiddle, tommyrot, tosh, twaddle",
  "n06611376": "bullshit, bull, Irish bull, horseshit, shit, crap, dogshit",
  "n06611681": "bunk, bunkum, buncombe, guff, rot, hogwash",
  "n06611856": "chickenshit",
  "n06611998": "folderol, rubbish, tripe, trumpery, trash, wish-wash, applesauce, codswallop",
  "n06612174": "pap, pablum",
  "n06612266": "drivel, garbage",
  "n06612420": "mumbo jumbo",
  "n06612538": "analects, analecta",
  "n06612649": "clipping, newspaper clipping, press clipping, cutting, press cutting",
  "n06612865": "cut, track",
  "n06613056": "quotation, quote, citation",
  "n06613340": "epigraph",
  "n06613448": "mimesis",
  "n06613576": "misquotation, misquote",
  "n06613686": "movie, film, picture, moving picture, moving-picture show, motion picture, motion-picture show, picture show, pic, flick",
  "n06614628": "telefilm",
  "n06614729": "scene, shot",
  "n06614901": "outtake",
  "n06615026": "feature, feature film",
  "n06615216": "final cut",
  "n06615348": "travelogue, travelog",
  "n06615458": "home movie",
  "n06615561": "attraction",
  "n06615736": "counterattraction",
  "n06615818": "collage film",
  "n06615927": "coming attraction",
  "n06616035": "Western, horse opera",
  "n06616216": "shoot-'em-up",
  "n06616314": "short subject",
  "n06616464": "cartoon, animated cartoon, toon",
  "n06616703": "newsreel",
  "n06616806": "documentary, docudrama, documentary film, infotainment",
  "n06617011": "cinema verite",
  "n06617165": "film noir",
  "n06617413": "skin flick",
  "n06617527": "peepshow",
  "n06617644": "rough cut",
  "n06617752": "silent movie, silent picture, silents",
  "n06617866": "slow motion",
  "n06618035": "dissolve",
  "n06618234": "cut",
  "n06618427": "jump",
  "n06618549": "jump cut",
  "n06618653": "spaghetti Western",
  "n06618822": "talking picture, talkie",
  "n06618937": "three-D, 3-D, 3D",
  "n06619065": "show",
  "n06619428": "broadcast, program, programme",
  "n06619751": "rebroadcast",
  "n06619850": "news program, news show, news",
  "n06620063": "rerun",
  "n06620227": "talk show, chat show",
  "n06620467": "phone-in",
  "n06620579": "television program, TV program, television show, TV show",
  "n06620790": "colorcast, colourcast",
  "n06620906": "pilot program, pilot film, pilot",
  "n06621061": "game show, giveaway",
  "n06621215": "quiz program",
  "n06621323": "film clip",
  "n06621447": "serial, series",
  "n06621681": "cliffhanger",
  "n06621771": "episode, installment, instalment",
  "n06621917": "sustaining program",
  "n06622020": "soap opera",
  "n06622252": "tetralogy",
  "n06622366": "radio broadcast",
  "n06622449": "simulcast",
  "n06622595": "telecast",
  "n06622709": "telegram, wire",
  "n06622877": "night letter",
  "n06622993": "airmail",
  "n06623100": "air mail",
  "n06623207": "surface mail",
  "n06623316": "registered mail, registered post",
  "n06623481": "special delivery",
  "n06623614": "correspondence",
  "n06623788": "Kamasutra",
  "n06623997": "sutra",
  "n06624161": "letter, missive",
  "n06624724": "business letter",
  "n06624816": "covering letter, cover letter",
  "n06624967": "crank letter",
  "n06625062": "encyclical, encyclical letter",
  "n06625217": "fan letter",
  "n06625329": "personal letter",
  "n06625465": "form letter",
  "n06625601": "open letter",
  "n06625739": "chain letter",
  "n06625850": "pastoral",
  "n06625965": "round robin",
  "n06626060": "airmail letter, air letter, aerogram, aerogramme",
  "n06626183": "epistle",
  "n06626286": "note, short letter, line, billet",
  "n06626446": "excuse",
  "n06626618": "love letter, billet doux",
  "n06626743": "dead letter, dead mail",
  "n06626860": "letter of intent",
  "n06627006": "card",
  "n06627260": "birthday card",
  "n06627357": "get-well card",
  "n06627450": "greeting card",
  "n06627623": "Christmas card",
  "n06627722": "Easter card",
  "n06627816": "Valentine",
  "n06627938": "postcard, post card, postal card, mailing-card",
  "n06628125": "lettercard",
  "n06628228": "picture postcard",
  "n06628328": "sympathy card",
  "n06628450": "Mass card",
  "n06628663": "spiritual bouquet",
  "n06628861": "acknowledgment, acknowledgement",
  "n06629392": "farewell, word of farewell",
  "n06629610": "adieu, adios, arrivederci, auf wiedersehen, au revoir, bye, bye-bye, cheerio, good-by, goodby, good-bye, goodbye, good day, sayonara, so long",
  "n06629858": "bon voyage, send-off",
  "n06630017": "greeting, salutation",
  "n06630459": "well-wishing",
  "n06630627": "regard, wish, compliments",
  "n06630852": "reception, response",
  "n06631050": "hail",
  "n06631140": "pax, kiss of peace",
  "n06631322": "welcome",
  "n06631506": "cordial reception, hospitality",
  "n06631661": "inhospitality",
  "n06631833": "glad hand",
  "n06631921": "aloha, ciao",
  "n06632097": "handshake, shake, handshaking, handclasp",
  "n06632358": "salute",
  "n06632511": "hello, hullo, hi, howdy, how-do-you-do",
  "n06632671": "good morning, morning",
  "n06632807": "good afternoon, afternoon",
  "n06632947": "good night",
  "n06633041": "salute, military greeting",
  "n06633205": "calling card, visiting card, card",
  "n06633363": "apology",
  "n06633596": "mea culpa",
  "n06633692": "condolence, commiseration",
  "n06633896": "congratulation, felicitation",
  "n06634095": "refusal",
  "n06634239": "declination, regrets",
  "n06634376": "information, info",
  "n06634960": "misinformation",
  "n06635104": "blowback",
  "n06635313": "disinformation",
  "n06635509": "material",
  "n06635759": "rehash",
  "n06635944": "details, inside information",
  "n06636113": "dope, poop, the skinny, low-down",
  "n06636259": "fact",
  "n06636524": "record, record book, book",
  "n06636806": "format, formatting, data format, data formatting",
  "n06637149": "high-level formatting",
  "n06637350": "low-level formatting, initialization, initialisation",
  "n06637677": "gen",
  "n06637824": "database",
  "n06637973": "relational database",
  "n06638254": "Medical Literature Analysis and Retrieval System, MEDLARS",
  "n06638527": "object-oriented database",
  "n06638739": "subdata base",
  "n06638868": "lexical database",
  "n06639023": "machine readable dictionary, MRD, electronic dictionary",
  "n06639204": "WordNet, Princeton WordNet",
  "n06639428": "wordnet",
  "n06639563": "basics, rudiments",
  "n06639674": "index, index number, indicant, indicator",
  "n06640016": "body mass index, BMI",
  "n06640317": "business index",
  "n06640533": "Dow Jones, Dow-Jones Industrial Average",
  "n06640848": "Standard and Poor's, Standard and Poor's Index",
  "n06640977": "leading indicator",
  "n06641181": "price index, price level",
  "n06641423": "retail price index",
  "n06641524": "producer price index, wholesale price index",
  "n06641654": "consumer price index, CPI, cost-of-living index",
  "n06641818": "short account",
  "n06641924": "stock index, stock market index",
  "n06642138": "news, intelligence, tidings, word",
  "n06642356": "news",
  "n06642518": "nuts and bolts",
  "n06642672": "intelligence, intelligence information",
  "n06642899": "military intelligence",
  "n06643120": "good word",
  "n06643185": "latest",
  "n06643303": "update",
  "n06643408": "evidence",
  "n06643763": "clue, clew, cue",
  "n06643883": "DNA fingerprint, genetic fingerprint",
  "n06644105": "face recognition, facial recognition, automatic face recognition",
  "n06644393": "fingerprint",
  "n06644658": "finger scan, finger scanning",
  "n06644823": "loop",
  "n06644916": "thumbprint",
  "n06645039": "footprint, footmark, step",
  "n06645266": "footprint evidence",
  "n06645458": "iris scanning",
  "n06645623": "signature recognition",
  "n06645831": "retinal scanning",
  "n06646022": "voiceprint",
  "n06646243": "sign, mark",
  "n06646531": "token",
  "n06646628": "trace, vestige, tincture, shadow",
  "n06646854": "footprint",
  "n06647036": "trace",
  "n06647206": "record",
  "n06647614": "proof",
  "n06647864": "mathematical proof",
  "n06647960": "logical proof",
  "n06648046": "demonstration, monstrance",
  "n06648207": "testimony, testimonial",
  "n06648462": "good authority",
  "n06648590": "testament",
  "n06648724": "argument, statement",
  "n06649108": "counterargument",
  "n06649223": "pro",
  "n06649325": "con",
  "n06649426": "case",
  "n06649567": "clincher, determiner, determining factor",
  "n06649704": "adducing",
  "n06649786": "last word",
  "n06649915": "attestation",
  "n06650070": "confirmation",
  "n06650238": "reinforcement, reenforcement",
  "n06650431": "documentation, certification, corroboration",
  "n06650701": "guidance, counsel, counseling, counselling, direction",
  "n06651047": "career counseling",
  "n06651145": "cynosure",
  "n06651302": "genetic counseling",
  "n06651463": "marriage counseling",
  "n06651577": "tip, lead, steer, confidential information, wind, hint",
  "n06651854": "insider information",
  "n06652064": "rule",
  "n06652242": "rule, prescript",
  "n06652614": "rubric",
  "n06652712": "rubric",
  "n06652878": "order, rules of order, parliamentary law, parliamentary procedure",
  "n06653160": "interpellation",
  "n06653363": "rule of evidence",
  "n06653727": "best evidence rule",
  "n06653911": "estoppel",
  "n06654074": "exclusionary rule",
  "n06654275": "fruit of the poisonous tree",
  "n06654490": "hearsay rule",
  "n06654639": "parol evidence rule",
  "n06654791": "res ipsa loquitur",
  "n06654977": "standing order",
  "n06655074": "Miranda rule",
  "n06655388": "principle, precept",
  "n06655569": "higher law",
  "n06655683": "moral principle",
  "n06655805": "golden rule",
  "n06655927": "GIGO",
  "n06656143": "categorical imperative",
  "n06656274": "hypothetical imperative",
  "n06656408": "policy",
  "n06656741": "economic policy",
  "n06656961": "fiscal policy",
  "n06657109": "New Deal",
  "n06657202": "control",
  "n06657464": "price control",
  "n06657646": "ceiling, roof, cap",
  "n06657913": "glass ceiling",
  "n06658118": "floor, base",
  "n06658269": "price floor",
  "n06658441": "wage floor",
  "n06658545": "perestroika",
  "n06658786": "protectionism",
  "n06658994": "social policy",
  "n06659168": "apartheid",
  "n06659397": "glasnost",
  "n06659530": "social action",
  "n06659685": "affirmative action",
  "n06660009": "fence mending",
  "n06660224": "trade barrier, import barrier",
  "n06660396": "quota",
  "n06660520": "embargo, trade embargo, trade stoppage",
  "n06660668": "nativism",
  "n06660816": "party line",
  "n06660942": "foreign policy",
  "n06661243": "brinkmanship",
  "n06661396": "imperialism",
  "n06661562": "intervention, interference",
  "n06661747": "nonintervention, noninterference",
  "n06661923": "nonaggression",
  "n06662057": "manifest destiny",
  "n06662193": "isolationism",
  "n06662366": "Monroe Doctrine",
  "n06662522": "Truman doctrine",
  "n06662712": "neutralism",
  "n06662850": "regionalism",
  "n06663018": "trade policy, national trading policy",
  "n06663168": "open-door policy, open door",
  "n06663308": "zero-tolerance policy",
  "n06663463": "Zionism",
  "n06663617": "ethic, ethical code",
  "n06663785": "caveat emptor",
  "n06663940": "dictate",
  "n06664051": "regulation, ordinance",
  "n06664213": "age limit",
  "n06664353": "assize",
  "n06664473": "speed limit",
  "n06664594": "canon",
  "n06664845": "etiquette",
  "n06664981": "protocol",
  "n06665108": "protocol, communications protocol",
  "n06665370": "file transfer protocol, FTP",
  "n06665612": "anonymous ftp, anonymous file transfer protocol",
  "n06665847": "hypertext transfer protocol, HTTP",
  "n06666030": "musical instrument digital interface, MIDI",
  "n06666210": "transmission control protocol, TCP",
  "n06666486": "transmission control protocol/internet protocol, TCP/IP",
  "n06666729": "punctilio",
  "n06666829": "closure, cloture, gag rule, gag law",
  "n06667027": "closure by compartment, guillotine",
  "n06667169": "point of order",
  "n06667317": "code, codification",
  "n06667625": "Bushido",
  "n06667792": "legal code",
  "n06668019": "penal code",
  "n06668147": "United States Code, U. S. Code",
  "n06668432": "building code",
  "n06668611": "dress code",
  "n06668813": "fire code",
  "n06669013": "omerta",
  "n06669193": "sanitary code, health code",
  "n06669384": "Highway Code",
  "n06669513": "double standard",
  "n06669673": "double standard of sexual behavior",
  "n06669864": "equation",
  "n06670137": "linear equation",
  "n06670240": "quadratic equation, quadratic",
  "n06670390": "biquadratic equation, biquadratic",
  "n06670521": "differential equation",
  "n06670711": "Maxwell's equations",
  "n06670866": "partial differential equation",
  "n06671013": "Schrodinger equation, Schrodinger wave equation",
  "n06671150": "simultaneous equations",
  "n06671339": "wave equation",
  "n06671484": "advice",
  "n06671637": "recommendation",
  "n06671818": "indication",
  "n06672040": "referral",
  "n06672297": "admonition, monition, warning, word of advice",
  "n06672752": "example, deterrent example, lesson, object lesson",
  "n06672953": "secret, arcanum",
  "n06673142": "secret",
  "n06673435": "confidence",
  "n06673671": "esoterica",
  "n06673770": "cabala, cabbala, cabbalah, kabala, kabbala, kabbalah, qabala, qabalah",
  "n06674028": "open secret",
  "n06674188": "password, watchword, word, parole, countersign",
  "n06674385": "trade secret",
  "n06674542": "propaganda",
  "n06674791": "agitprop",
  "n06674947": "course catalog, course catalogue, prospectus",
  "n06675122": "source",
  "n06675338": "specification",
  "n06675581": "source materials",
  "n06675691": "voucher",
  "n06675799": "working papers",
  "n06675979": "well, wellspring, fountainhead",
  "n06676109": "copy",
  "n06676254": "filler",
  "n06676416": "course of study, program, programme, curriculum, syllabus",
  "n06676709": "crash course, crash program, crash programme",
  "n06677089": "reading program",
  "n06677194": "degree program",
  "n06677302": "printing, printing process",
  "n06677974": "typography, composition",
  "n06678146": "print",
  "n06678302": "print",
  "n06678506": "small print, fine print",
  "n06678664": "relief printing, letterpress",
  "n06678784": "intaglio printing, intaglio, gravure",
  "n06679041": "process printing",
  "n06679167": "photogravure",
  "n06679308": "rotogravure",
  "n06679457": "planographic printing, planography",
  "n06679726": "collotype, collotype printing, photogelatin process",
  "n06680002": "lithography",
  "n06680195": "photolithography",
  "n06680333": "chromolithography",
  "n06680431": "photo-offset printing, photo-offset",
  "n06680570": "offset, offset printing",
  "n06680818": "offset lithography",
  "n06680913": "letterset printing",
  "n06681035": "carbon process",
  "n06681177": "news",
  "n06681454": "business news",
  "n06681551": "report, news report, story, account, write up",
  "n06681976": "newsletter, newssheet",
  "n06682152": "market letter",
  "n06682290": "bulletin",
  "n06682494": "news bulletin, newsflash, flash, newsbreak",
  "n06682683": "information bulletin",
  "n06682794": "dispatch, despatch, communique",
  "n06682952": "urban legend",
  "n06683183": "exclusive, scoop",
  "n06683359": "newscast",
  "n06683513": "radio news",
  "n06683595": "sportscast",
  "n06683692": "television news",
  "n06683784": "coverage, reporting, reportage",
  "n06684052": "hard news",
  "n06684153": "soft news",
  "n06684262": "stop press",
  "n06684383": "commitment, dedication",
  "n06684572": "oath, swearing",
  "n06684798": "affirmation",
  "n06685040": "profession",
  "n06685198": "giving",
  "n06685456": "guarantee, warrant, warrantee, warranty",
  "n06685754": "security, surety",
  "n06685860": "deposit",
  "n06685975": "stock warrant",
  "n06686174": "guarantee",
  "n06686467": "safety net",
  "n06686573": "full faith and credit",
  "n06686736": "approval, commendation",
  "n06687178": "approbation",
  "n06687358": "sanction, countenance, endorsement, indorsement, warrant, imprimatur",
  "n06687701": "O.K., OK, okay, okey, okeh",
  "n06687883": "visa",
  "n06688059": "nihil obstat",
  "n06688274": "recognition, credit",
  "n06688522": "memorial, commemoration, remembrance",
  "n06688751": "ovation, standing ovation",
  "n06688913": "salute, salutation",
  "n06689125": "connivance, secret approval, tacit consent",
  "n06689297": "permission",
  "n06689554": "all clear",
  "n06689667": "consent",
  "n06689829": "dismissal",
  "n06689948": "green light",
  "n06690114": "leave",
  "n06690226": "pass, liberty chit",
  "n06690408": "pass, laissez passer",
  "n06690647": "boarding card, boarding pass",
  "n06690770": "hall pass",
  "n06690928": "ticket-of-leave",
  "n06691083": "pass, passport",
  "n06691267": "safe-conduct, safeguard",
  "n06691442": "encouragement",
  "n06691684": "acclaim, acclamation, plaudits, plaudit, eclat",
  "n06691989": "applause, hand clapping, clapping",
  "n06692238": "hand",
  "n06692369": "handclap",
  "n06692464": "round",
  "n06692572": "cheer",
  "n06692740": "banzai",
  "n06692834": "bravo",
  "n06692973": "hurrah, hooray",
  "n06693086": "salvo",
  "n06693198": "praise, congratulations, kudos, extolment",
  "n06693502": "praise",
  "n06693655": "hallelujah",
  "n06693744": "rave",
  "n06693870": "superlative",
  "n06694149": "encomium, eulogy, panegyric, paean, pean",
  "n06694359": "eulogy, eulogium",
  "n06694540": "recommendation, testimonial, good word",
  "n06694796": "character, reference, character reference",
  "n06695110": "puff",
  "n06695227": "compliment",
  "n06695424": "trade-last",
  "n06695579": "flattery",
  "n06695753": "adulation",
  "n06695862": "blandishment, cajolery, palaver",
  "n06696025": "blarney, coaxing, soft soap, sweet talk",
  "n06696181": "puffery",
  "n06696308": "unction, smarm, fulsomeness",
  "n06696483": "award, accolade, honor, honour, laurels",
  "n06696991": "aliyah",
  "n06697195": "tribute, testimonial",
  "n06697331": "academic degree, degree",
  "n06697703": "associate degree, associate",
  "n06697935": "Associate in Arts, AA",
  "n06698031": "Associate in Applied Science, AAS",
  "n06698150": "Associate in Nursing, AN",
  "n06698252": "bachelor's degree, baccalaureate",
  "n06698640": "Bachelor of Arts, BA, Artium Baccalaurens, AB",
  "n06698775": "Bachelor of Arts in Library Science, ABLS",
  "n06698902": "Bachelor of Arts in Nursing, BAN",
  "n06699012": "Bachelor of Divinity, BD",
  "n06699115": "Bachelor of Literature, BLitt",
  "n06699225": "Bachelor of Medicine, MB",
  "n06699366": "Bachelor of Music, BMus",
  "n06699465": "Bachelor of Naval Science, BNS",
  "n06699579": "Bachelor of Science, BS, SB",
  "n06699685": "Bachelor of Science in Architecture, BSArch",
  "n06699811": "Bachelor of Science in Engineering",
  "n06699926": "Bachelor of Theology, ThB",
  "n06700030": "honours, honours degree",
  "n06700169": "first, first-class honours degree",
  "n06700325": "double first",
  "n06700447": "master's degree",
  "n06700844": "Master of Architecture, MArch",
  "n06701001": "Master of Arts, MA, Artium Magister, AM",
  "n06701128": "Master of Arts in Library Science, MALS",
  "n06701251": "Master of Arts in Teaching, MAT",
  "n06701359": "Master in Business, Master in Business Administration, MBA",
  "n06701495": "Master of Divinity, MDiv",
  "n06701596": "Master of Education, MEd",
  "n06701698": "Master of Fine Arts, MFA",
  "n06701800": "Master of Literature, MLitt",
  "n06701906": "Master of Library Science, MLS",
  "n06702020": "Master in Public Affairs",
  "n06702139": "Master of Science, MS, SM, MSc",
  "n06702247": "Master of Science in Engineering",
  "n06702358": "Master of Theology, ThM",
  "n06702458": "doctor's degree, doctorate",
  "n06702888": "Doctor of Dental Medicine, DMD",
  "n06703002": "Doctor of Dental Surgery, DDS",
  "n06703114": "Doctor of Divinity, DD",
  "n06703213": "Doctor of Education, EdD, DEd",
  "n06703321": "Doctor of Medicine, MD",
  "n06703420": "Doctor of Music, DMus, MusD",
  "n06703522": "Doctor of Musical Arts, AMusD",
  "n06703632": "Doctor of Optometry, OD",
  "n06703733": "Doctor of Osteopathy, DO",
  "n06703834": "Doctor of Arts, D.A.",
  "n06703953": "Doctor of Philosophy",
  "n06704115": "Ph.D., PhD",
  "n06704295": "DPhil",
  "n06704366": "Doctor of Public Health, DPH",
  "n06704482": "Doctor of Theology, ThD",
  "n06704582": "Doctor of Sacred Theology, STD",
  "n06704740": "law degree",
  "n06704898": "Bachelor of Laws, LLB",
  "n06704990": "Master of Laws, LLM",
  "n06705079": "honorary degree, honoris causa",
  "n06705306": "Doctor of Arts, ArtsD",
  "n06705398": "Doctor of Fine Arts",
  "n06705495": "Doctor of Humane Letters",
  "n06705595": "Doctor of Humanities",
  "n06705698": "Doctor of Laws, LLD",
  "n06705787": "Doctor of Science, DS, ScD",
  "n06705891": "pennant, crown",
  "n06705984": "cachet, seal, seal of approval",
  "n06706125": "citation, commendation",
  "n06706317": "mention, honorable mention",
  "n06706504": "letter, varsity letter",
  "n06706676": "decoration, laurel wreath, medal, medallion, palm, ribbon",
  "n06707178": "Medal of Honor, Congressional Medal of Honor",
  "n06707382": "Distinguished Service Medal",
  "n06707555": "Distinguished Service Cross",
  "n06707709": "Navy Cross",
  "n06707846": "Distinguished Flying Cross",
  "n06708007": "Air Medal",
  "n06708167": "Silver Star Medal, Silver Star",
  "n06708304": "Bronze Star Medal, Bronze Star",
  "n06708475": "Order of the Purple Heart, Purple Heart",
  "n06708664": "Oak Leaf Cluster",
  "n06708866": "Victoria Cross",
  "n06708970": "Distinguished Conduct Medal",
  "n06709112": "Distinguished Service Order",
  "n06709245": "Croix de Guerre",
  "n06709349": "Medaille Militaire",
  "n06709442": "trophy",
  "n06709533": "disapproval",
  "n06709692": "disapprobation, condemnation",
  "n06709998": "censure, animadversion",
  "n06710152": "demonization, demonisation",
  "n06710330": "interdict",
  "n06710546": "criticism, unfavorable judgment",
  "n06710960": "brickbat",
  "n06711030": "faultfinding, carping",
  "n06711159": "fire, attack, flak, flack, blast",
  "n06711404": "thrust",
  "n06711539": "potshot",
  "n06711705": "counterblast",
  "n06711855": "rebuke, reproof, reproval, reprehension, reprimand",
  "n06712325": "sermon, preaching",
  "n06712498": "slating",
  "n06712625": "static",
  "n06712761": "stricture",
  "n06712833": "chiding, scolding, objurgation, tongue-lashing",
  "n06713026": "what for",
  "n06713099": "wig, wigging",
  "n06713187": "castigation, earful, bawling out, chewing out, upbraiding, going-over, dressing down",
  "n06713377": "berating, blowing up",
  "n06713512": "reproach",
  "n06713650": "self-reproach, self-reproof",
  "n06713752": "blame, rap",
  "n06713930": "lecture, speech, talking to",
  "n06714184": "curtain lecture",
  "n06714288": "correction, chastening, chastisement",
  "n06714420": "admonition, admonishment, monition",
  "n06714552": "respects",
  "n06714697": "ad-lib",
  "n06714874": "courtesy",
  "n06714976": "disrespect, discourtesy",
  "n06715223": "abuse, insult, revilement, contumely, vilification",
  "n06715638": "derision",
  "n06715786": "ridicule",
  "n06715927": "contempt, scorn",
  "n06716117": "fleer",
  "n06716234": "jeer, jeering, mockery, scoff, scoffing",
  "n06716483": "sneer, leer",
  "n06716675": "sneer",
  "n06716796": "put-down, squelch, squelcher, takedown",
  "n06716992": "stultification, befooling",
  "n06717170": "disparagement, depreciation, derogation",
  "n06718269": "cold water",
  "n06718434": "denigration, belittling",
  "n06718543": "aspersion, slur",
  "n06718862": "ethnic slur",
  "n06719203": "detraction, petty criticism",
  "n06719299": "sour grapes",
  "n06719404": "condescension, disdain, patronage",
  "n06719579": "defamation, calumny, calumniation, obloquy, traducement, hatchet job",
  "n06719974": "character assassination, assassination, blackwash",
  "n06720137": "mud",
  "n06720216": "smear, vilification, malignment",
  "n06720371": "libel",
  "n06720600": "slander",
  "n06720784": "name calling, names",
  "n06720964": "name, epithet",
  "n06721081": "smear word",
  "n06721269": "low blow",
  "n06721342": "scurrility, billingsgate",
  "n06721461": "stinger, cut",
  "n06721604": "vituperation, invective, vitriol",
  "n06721813": "impudence, cheek, impertinence",
  "n06721949": "sass, sassing, backtalk, back talk, lip, mouth",
  "n06722186": "interpolation, insertion",
  "n06722453": "statement",
  "n06723452": "statement",
  "n06723635": "amendment",
  "n06723908": "thing",
  "n06724066": "truth, true statement",
  "n06724323": "gospel, gospel truth",
  "n06724437": "antinomy",
  "n06724559": "paradox",
  "n06724763": "description, verbal description",
  "n06724942": "job description",
  "n06725067": "specification, spec",
  "n06725249": "computer architecture",
  "n06725467": "neural network, neural net",
  "n06725680": "network architecture",
  "n06725877": "declaration",
  "n06726158": "announcement, proclamation, annunciation, declaration",
  "n06726478": "bastardization",
  "n06726671": "edict",
  "n06726761": "bull, papal bull",
  "n06726939": "promulgation",
  "n06727133": "confession",
  "n06727224": "manifesto, pronunciamento",
  "n06727416": "Communist Manifesto",
  "n06727616": "pronouncement, dictum, say-so",
  "n06727758": "Bill of Rights",
  "n06728034": "First Amendment",
  "n06728331": "Fifth Amendment",
  "n06728726": "Fourteenth Amendment",
  "n06728998": "Eighteenth Amendment",
  "n06729251": "Nineteenth Amendment",
  "n06729499": "assertion, averment, asseveration",
  "n06729864": "claim",
  "n06730068": "cause of action",
  "n06730241": "dibs",
  "n06730371": "pretension",
  "n06730563": "claim",
  "n06730780": "accusation, charge",
  "n06731069": "countercharge",
  "n06731186": "allegation, allegement",
  "n06731378": "contention",
  "n06731510": "submission",
  "n06731697": "ipse dixit, ipsedixitism",
  "n06731802": "formula, expression",
  "n06732013": "formula",
  "n06732169": "mathematical statement",
  "n06732350": "avowal, avouchment, affirmation",
  "n06732581": "reassertion, reaffirmation",
  "n06732710": "testimony",
  "n06732925": "profession, professing",
  "n06733119": "protestation",
  "n06733227": "postulation, predication",
  "n06733476": "threat",
  "n06733682": "commination",
  "n06733782": "menace",
  "n06733939": "evidence",
  "n06734322": "exhibit",
  "n06734467": "testimony",
  "n06734702": "witness",
  "n06734823": "corpus delicti",
  "n06735077": "direct evidence",
  "n06735261": "res gestae",
  "n06735533": "circumstantial evidence, indirect evidence",
  "n06735731": "corroborating evidence",
  "n06735922": "hearsay evidence",
  "n06736083": "state's evidence",
  "n06736217": "declaration",
  "n06736405": "attestation",
  "n06736529": "affidavit",
  "n06736769": "verification",
  "n06736946": "subornation",
  "n06737112": "bid, bidding",
  "n06737394": "contract, declaration",
  "n06737628": "takeout",
  "n06737758": "overbid, overcall",
  "n06738008": "preemptive bid, pre-empt, preempt",
  "n06738162": "word",
  "n06738281": "explanation, account",
  "n06738823": "explicandum, explanandum",
  "n06739013": "explanans",
  "n06739180": "simplification",
  "n06739345": "oversimplification, simplism",
  "n06739509": "accounting",
  "n06739676": "value statement",
  "n06739782": "representation",
  "n06739990": "reason",
  "n06740183": "justification",
  "n06740402": "cause, reason, grounds",
  "n06740644": "defense, defence, vindication",
  "n06740919": "apology, apologia",
  "n06741099": "alibi",
  "n06741305": "excuse, alibi, exculpation, self-justification",
  "n06741728": "extenuation, mitigation",
  "n06741993": "exposition",
  "n06742173": "exposition",
  "n06742426": "exposition, expounding",
  "n06742630": "construal",
  "n06742772": "philosophizing",
  "n06742932": "moralizing, moralization, moralisation",
  "n06743230": "preachification",
  "n06743362": "explication",
  "n06743506": "solution, answer, result, resolution, solvent",
  "n06743867": "denouement",
  "n06744000": "gloss, rubric",
  "n06744154": "deriving, derivation, etymologizing",
  "n06744396": "definition",
  "n06744663": "contextual definition",
  "n06745015": "dictionary definition",
  "n06745156": "explicit definition",
  "n06745285": "ostensive definition",
  "n06745422": "recursive definition",
  "n06745628": "redefinition",
  "n06745861": "stipulative definition",
  "n06746005": "answer, reply, response",
  "n06746347": "rescript",
  "n06746471": "feedback",
  "n06746580": "announcement, promulgation",
  "n06747008": "advisory",
  "n06747190": "Annunciation",
  "n06747369": "banns",
  "n06747484": "handout, press release, release",
  "n06747670": "notice",
  "n06747907": "caveat",
  "n06748133": "obituary, obit, necrology",
  "n06748270": "Parallel Lives",
  "n06748466": "program, programme",
  "n06748703": "playbill",
  "n06748841": "racecard",
  "n06748969": "prediction, foretelling, forecasting, prognostication",
  "n06749267": "extropy",
  "n06749468": "fortunetelling",
  "n06749594": "horoscope",
  "n06749729": "meteorology, weather forecasting",
  "n06749881": "prognosis, forecast",
  "n06750154": "prophecy, divination",
  "n06750339": "oracle",
  "n06750514": "financial forecast",
  "n06750698": "weather forecast, weather outlook",
  "n06750804": "proposition",
  "n06751142": "particular, particular proposition",
  "n06751367": "universal, universal proposition",
  "n06751572": "negation",
  "n06751742": "converse",
  "n06751833": "lemma",
  "n06751974": "term",
  "n06752293": "theorem",
  "n06752410": "categorem, categoreme",
  "n06752695": "syncategorem, syncategoreme",
  "n06753030": "conclusion, ratiocination",
  "n06753299": "postulate, posit",
  "n06753550": "axiom",
  "n06753800": "premise, premiss, assumption",
  "n06754184": "major premise, major premiss",
  "n06754415": "minor premise, minor premiss, subsumption",
  "n06754658": "major term",
  "n06754816": "minor term",
  "n06754972": "middle term",
  "n06755156": "specious argument",
  "n06755342": "vicious circle",
  "n06755454": "thesis",
  "n06755568": "condition, precondition, stipulation",
  "n06755776": "boundary condition",
  "n06755947": "provision, proviso",
  "n06756111": "scenario",
  "n06756267": "quotation",
  "n06756407": "falsehood, falsity, untruth",
  "n06756680": "dodge, dodging, scheme",
  "n06756831": "lie, prevarication",
  "n06757057": "fib, story, tale, tarradiddle, taradiddle",
  "n06757289": "fairytale, fairy tale, fairy story, cock-and-bull story, song and dance",
  "n06757479": "jactitation",
  "n06757676": "whopper, walloper",
  "n06757771": "white lie",
  "n06757891": "fabrication, fiction, fable",
  "n06758135": "canard",
  "n06758225": "misrepresentation, deceit, deception",
  "n06758584": "half-truth",
  "n06758698": "facade, window dressing",
  "n06758835": "exaggeration, overstatement, magnification",
  "n06759063": "understatement",
  "n06759256": "snow job",
  "n06759349": "pretense, pretence, feigning, dissembling",
  "n06759602": "bluff",
  "n06759776": "pretext, stalking-horse",
  "n06759974": "putoff",
  "n06760076": "hypocrisy, lip service",
  "n06760249": "crocodile tears",
  "n06760508": "subterfuge, blind",
  "n06760722": "trickery, hocus-pocus, slickness, hanky panky, jiggery-pokery, skulduggery, skullduggery",
  "n06760969": "fraudulence, duplicity",
  "n06761099": "evasion, equivocation",
  "n06761342": "circumlocution, indirect expression",
  "n06761482": "doublespeak",
  "n06761603": "hedge, hedging",
  "n06761798": "quibble, quiddity, cavil",
  "n06761994": "fine print, small print",
  "n06762245": "weasel word",
  "n06762380": "reservation, qualification",
  "n06762602": "cautious statement",
  "n06762711": "comment, commentary",
  "n06763052": "Midrash",
  "n06763273": "note, annotation, notation",
  "n06763681": "citation, cite, acknowledgment, credit, reference, mention, quotation",
  "n06764244": "footnote, footer",
  "n06764380": "nota bene, NB, N.B.",
  "n06764623": "postscript, PS",
  "n06764751": "photo credit",
  "n06764867": "cross-reference, cross-index",
  "n06765044": "remark, comment, input",
  "n06765531": "gambit, ploy",
  "n06765656": "fatwah",
  "n06765887": "obiter dictum, dictum",
  "n06766092": "obiter dictum, passing comment",
  "n06766190": "mention, reference",
  "n06766544": "allusion",
  "n06766654": "retrospection",
  "n06766812": "name-dropping",
  "n06767035": "observation, reflection, reflexion",
  "n06767287": "Parkinson's law",
  "n06767512": "Parkinson's law",
  "n06767693": "rib",
  "n06767777": "wisecrack, crack, sally, quip",
  "n06767922": "shot, shaft, slam, dig, barb, jibe, gibe",
  "n06768259": "cheap shot",
  "n06768394": "conversation stopper, stopper",
  "n06768528": "rhetorical question",
  "n06768735": "misstatement",
  "n06768901": "restatement",
  "n06769032": "demythologization, demythologisation",
  "n06769238": "mythologization, mythologisation",
  "n06769392": "error, mistake",
  "n06769578": "corrigendum",
  "n06769670": "misprint, erratum, typographical error, typo, literal error, literal",
  "n06769883": "malapropism, malaprop",
  "n06770028": "slip of the tongue",
  "n06770164": "spoonerism",
  "n06770275": "agreement, understanding",
  "n06770875": "condition, term",
  "n06771159": "bargain, deal",
  "n06771424": "working agreement",
  "n06771526": "gentlemen's agreement",
  "n06771653": "written agreement",
  "n06771957": "submission",
  "n06772094": "submission",
  "n06772260": "covenant, compact, concordat",
  "n06772461": "entente, entente cordiale",
  "n06772583": "oral contract",
  "n06772837": "indenture",
  "n06773006": "indenture",
  "n06773150": "obligation",
  "n06773347": "debt",
  "n06773434": "treaty, pact, accord",
  "n06773684": "alliance",
  "n06773857": "commercial treaty",
  "n06773976": "peace, peace treaty, pacification",
  "n06774192": "Peace of Westphalia",
  "n06774316": "convention",
  "n06774468": "Chemical Weapons Convention",
  "n06774650": "Geneva Convention",
  "n06774871": "Lateran Treaty",
  "n06775086": "North Atlantic Treaty",
  "n06775218": "SALT I",
  "n06775409": "SALT II",
  "n06775602": "Treaty of Versailles",
  "n06775812": "sentimentalism",
  "n06775969": "treacle, mush, slop, glop",
  "n06776138": "wit, humor, humour, witticism, wittiness",
  "n06776679": "jeu d'esprit",
  "n06776783": "bon mot, mot",
  "n06776877": "esprit de l'escalier",
  "n06776986": "pungency, bite",
  "n06777164": "sarcasm, irony, satire, caustic remark",
  "n06777687": "repartee",
  "n06777794": "banter, raillery, give-and-take, backchat",
  "n06777961": "badinage",
  "n06778032": "persiflage",
  "n06778102": "joke, gag, laugh, jest, jape",
  "n06778777": "punch line, laugh line, gag line, tag line",
  "n06778925": "belly laugh, sidesplitter, howler, thigh-slapper, scream, wow, riot",
  "n06779096": "dirty joke, dirty story, blue joke, blue story",
  "n06779210": "ethnic joke",
  "n06779310": "funny story, good story, funny remark, funny",
  "n06779511": "in-joke",
  "n06779642": "one-liner",
  "n06779713": "shaggy dog story",
  "n06779839": "sick joke",
  "n06779914": "sight gag, visual joke",
  "n06780069": "caricature, imitation, impersonation",
  "n06780309": "parody, lampoon, spoof, sendup, mockery, takeoff, burlesque, travesty, charade, pasquinade, put-on",
  "n06780678": "cartoon, sketch",
  "n06780882": "fun, play, sport",
  "n06781151": "jocosity, jocularity",
  "n06781281": "waggery, waggishness",
  "n06781383": "drollery, clowning, comedy, funniness",
  "n06781581": "pun, punning, wordplay, paronomasia",
  "n06781811": "ribaldry",
  "n06781878": "topper",
  "n06782019": "opinion, view",
  "n06782383": "adverse opinion",
  "n06782680": "guess, conjecture, supposition, surmise, surmisal, speculation, hypothesis",
  "n06783155": "divination",
  "n06783265": "side, position",
  "n06783442": "approximation",
  "n06783598": "estimate",
  "n06783768": "question, head",
  "n06784003": "problem",
  "n06784317": "question of fact, matter of fact",
  "n06784470": "question of law, matter of law",
  "n06784639": "puzzle, puzzler, mystifier, teaser",
  "n06784966": "case",
  "n06785101": "homework problem",
  "n06785223": "riddle, conundrum, enigma, brain-teaser",
  "n06785367": "poser, stumper, toughie, sticker",
  "n06785541": "Gordian knot",
  "n06785654": "crossword puzzle, crossword",
  "n06785840": "koan",
  "n06786012": "sudoku",
  "n06786209": "word square, acrostic",
  "n06786358": "pons asinorum",
  "n06786486": "rebus",
  "n06786629": "direction, instruction",
  "n06787037": "misdirection",
  "n06787150": "address, destination, name and address",
  "n06787429": "return address",
  "n06787602": "markup",
  "n06787835": "markup language",
  "n06788007": "standard generalized markup language, SGML",
  "n06788262": "hypertext markup language, hypertext mark-up language, HTML",
  "n06788463": "toponymy, toponomy",
  "n06788565": "prescription",
  "n06788785": "recipe, formula",
  "n06788897": "rule",
  "n06789080": "stage direction",
  "n06789215": "style",
  "n06789411": "religious doctrine, church doctrine, gospel, creed",
  "n06789801": "ahimsa",
  "n06790042": "dogma, tenet",
  "n06790235": "ecumenism, ecumenicism, ecumenicalism",
  "n06790557": "Immaculate Conception, Immaculate Conception of the Virgin Mary",
  "n06790845": "Incarnation",
  "n06791017": "Nicene Creed",
  "n06791195": "real presence",
  "n06791372": "signal, signaling, sign",
  "n06792188": "starting signal, start",
  "n06792364": "storm signal",
  "n06792526": "storm cone",
  "n06792645": "radio beam, beam",
  "n06792818": "tickler, tickler file",
  "n06792950": "ticktack",
  "n06793091": "time signal",
  "n06793231": "sign",
  "n06793426": "poster, posting, placard, notice, bill, card",
  "n06793817": "show bill, show card, theatrical poster",
  "n06793959": "flash card, flashcard",
  "n06794110": "street sign",
  "n06794216": "address",
  "n06794374": "signpost, guidepost",
  "n06794537": "fingerpost, fingerboard",
  "n06794666": "mark, stigma, brand, stain",
  "n06794980": "demerit",
  "n06795168": "dog-ear",
  "n06795290": "bar sinister, bend sinister",
  "n06795438": "earmark",
  "n06795543": "brand",
  "n06795657": "cloven hoof, cloven foot",
  "n06795746": "token, item",
  "n06795967": "type",
  "n06796119": "postage, postage stamp, stamp",
  "n06796333": "trading stamp",
  "n06796506": "animal communication",
  "n06796642": "birdcall, call, birdsong, song",
  "n06796942": "bell-like call",
  "n06797047": "two-note call",
  "n06797169": "indication, indicant",
  "n06797671": "indication",
  "n06797947": "contraindication",
  "n06798187": "symptom",
  "n06798336": "signalization, signalisation",
  "n06798472": "pointing out",
  "n06798558": "manifestation",
  "n06798750": "mark, print",
  "n06799142": "mintmark",
  "n06799260": "stroke",
  "n06799485": "downstroke",
  "n06799588": "upstroke",
  "n06799688": "flick",
  "n06799754": "hoofprint, hoof mark, hoof-mark",
  "n06799897": "line",
  "n06800344": "dotted line",
  "n06800532": "ascender",
  "n06800698": "bar line",
  "n06800871": "descender",
  "n06801033": "squiggle, curlicue",
  "n06801138": "spectrum line",
  "n06801313": "trend line",
  "n06801418": "underscore, underline",
  "n06801580": "contour, contour line",
  "n06801740": "isometric line, isometric",
  "n06801865": "thalweg",
  "n06801965": "graduation",
  "n06802138": "guideline",
  "n06802258": "hairline",
  "n06802347": "hair stroke",
  "n06802444": "glimpse",
  "n06802571": "harbinger, forerunner, predecessor, herald, precursor",
  "n06802785": "hint, clue",
  "n06802880": "smoke",
  "n06803023": "air alert",
  "n06803157": "alarm, alert, warning signal, alarum",
  "n06803478": "burglar alarm",
  "n06803636": "distress signal, distress call",
  "n06803845": "SOS",
  "n06803954": "Mayday",
  "n06804097": "all clear",
  "n06804199": "bugle call",
  "n06804388": "recall",
  "n06804483": "taps, lights-out",
  "n06804606": "curfew",
  "n06804728": "reveille, wake-up signal",
  "n06804847": "retreat",
  "n06804988": "retreat",
  "n06805128": "drumbeat",
  "n06805297": "tattle, singing, telling",
  "n06805497": "tattoo",
  "n06805665": "telegraphic signal, radiotelegraphic signal",
  "n06805826": "dot, dit",
  "n06805962": "dash, dah",
  "n06806098": "whistle, whistling",
  "n06806318": "high sign",
  "n06806469": "symbol",
  "n06807056": "nose",
  "n06807198": "numeral, number",
  "n06807501": "Arabic numeral, Hindu numeral, Hindu-Arabic numeral",
  "n06807657": "Roman numeral",
  "n06807849": "symbolism",
  "n06807971": "crossbones",
  "n06808121": "horn of plenty, cornucopia",
  "n06808271": "death's head",
  "n06808408": "lingam",
  "n06808493": "notation, notational system",
  "n06808720": "mathematical notation",
  "n06809074": "numeration system, number system, number representation system, system of numeration",
  "n06809291": "oriflamme",
  "n06809421": "positional notation, positional representation system",
  "n06809756": "pound, pound sign",
  "n06809905": "binary notation",
  "n06810020": "binary numeration system, pure binary numeration system, binary number system, binary system",
  "n06810238": "octal numeration system, octal number system",
  "n06810407": "decimal notation",
  "n06810544": "octal notation",
  "n06810691": "algorism",
  "n06810790": "decimal numeration system, decimal number system, decimal system",
  "n06810997": "duodecimal notation",
  "n06811109": "duodecimal number system, duodecimal system",
  "n06811283": "hexadecimal notation, sexadecimal notation",
  "n06811419": "hexadecimal number system, sexadecimal number system, hexadecimal system",
  "n06811625": "sign",
  "n06811867": "equal sign",
  "n06811986": "plus sign",
  "n06812085": "minus sign",
  "n06812188": "radical sign",
  "n06812289": "decimal point, percentage point, point",
  "n06812417": "exponent, power, index",
  "n06812631": "logarithm, log",
  "n06812794": "antilogarithm, antilog",
  "n06812915": "common logarithm",
  "n06813004": "natural logarithm, Napierian logarithm",
  "n06813115": "mantissa, fixed-point part",
  "n06813310": "characteristic",
  "n06813506": "fixed-point notation, fixed-point representation system",
  "n06813700": "floating-point notation, floating-point representation system",
  "n06814021": "infix notation",
  "n06814236": "parenthesis-free notation",
  "n06814443": "prefix notation, Lukasiewicz notation, Polish notation",
  "n06814655": "postfix notation, suffix notation, reverse Polish notation",
  "n06814870": "musical notation",
  "n06815458": "lead sheet",
  "n06815621": "piano music",
  "n06815714": "score, musical score",
  "n06815970": "obbligato, obligato",
  "n06816106": "sheet music",
  "n06816335": "tablature",
  "n06816445": "choreography",
  "n06816555": "Labanotation",
  "n06816785": "chemical notation",
  "n06816935": "formula, chemical formula",
  "n06817173": "molecular formula",
  "n06817312": "structural formula",
  "n06817459": "empirical formula",
  "n06817623": "written symbol, printed symbol",
  "n06817782": "mark",
  "n06818121": "arrow, pointer",
  "n06818317": "broad arrow",
  "n06818439": "call mark, call number, pressmark",
  "n06818603": "caret",
  "n06818747": "check mark, check, tick",
  "n06818970": "character, grapheme, graphic symbol",
  "n06819557": "allograph",
  "n06819698": "readout, read-out",
  "n06819824": "check character",
  "n06820023": "superscript, superior",
  "n06820212": "subscript, inferior",
  "n06820425": "ASCII character",
  "n06820601": "control character, ASCII control character",
  "n06820837": "backspace character",
  "n06820964": "diacritical mark, diacritic",
  "n06821279": "ditto mark, ditto",
  "n06821422": "dollar mark, dollar sign",
  "n06821581": "dollar, dollar mark, dollar sign",
  "n06821770": "shaft",
  "n06821889": "phonogram",
  "n06822034": "point, head",
  "n06822198": "accent, accent mark",
  "n06822476": "stress mark",
  "n06822576": "acute accent, acute, ague",
  "n06822707": "grave accent, grave",
  "n06822831": "breve",
  "n06822958": "cedilla",
  "n06823102": "circumflex",
  "n06823259": "hacek, wedge",
  "n06823440": "macron",
  "n06823561": "tilde",
  "n06823760": "umlaut, dieresis, diaeresis",
  "n06823925": "ligature",
  "n06824041": "monogram",
  "n06824227": "capital, capital letter, uppercase, upper-case letter, majuscule",
  "n06824757": "small letter, lowercase, lower-case letter, minuscule",
  "n06824955": "small capital, small cap",
  "n06825120": "type",
  "n06825273": "type family",
  "n06825399": "font, fount, typeface, face, case",
  "n06825736": "unicameral script",
  "n06825863": "bicameral script",
  "n06825996": "typewriter font, constant-width font, fixed-width font, monospaced font",
  "n06826214": "proportional font",
  "n06826407": "font cartridge, cartridge font",
  "n06826589": "Gothic, black letter",
  "n06826726": "modern, modern font, Bodoni, Bodoni font",
  "n06826974": "old style, old style font",
  "n06827219": "boldface, bold face, bold",
  "n06827344": "italic",
  "n06827503": "roman, roman type, roman letters, roman print",
  "n06827679": "screen font, raster font",
  "n06827947": "sans serif, Helvetica",
  "n06828061": "serif, seriph",
  "n06828199": "percent sign, percentage sign",
  "n06828389": "asterisk, star",
  "n06828529": "dagger, obelisk",
  "n06828662": "double dagger, double obelisk, diesis",
  "n06828818": "letter, letter of the alphabet, alphabetic character",
  "n06830561": "ascender",
  "n06830708": "descender",
  "n06830838": "digraph, digram",
  "n06830997": "initial",
  "n06831177": "A, a",
  "n06831284": "B, b",
  "n06831391": "C, c",
  "n06831498": "D, d",
  "n06831605": "E, e",
  "n06831712": "F, f",
  "n06831819": "G, g",
  "n06831926": "H, h",
  "n06832033": "I, i",
  "n06832140": "J, j",
  "n06832248": "K, k",
  "n06832356": "L, l",
  "n06832464": "M, m",
  "n06832572": "N, n",
  "n06832680": "O, o",
  "n06832788": "P, p",
  "n06832896": "Q, q",
  "n06833004": "R, r",
  "n06833112": "S, s",
  "n06833220": "T, t",
  "n06833328": "U, u",
  "n06833436": "V, v",
  "n06833544": "W, w, double-u",
  "n06833663": "X, x, ex",
  "n06833776": "Y, y, wye",
  "n06833890": "Z, z, zee, zed, ezed, izzard",
  "n06834138": "alpha",
  "n06834245": "beta",
  "n06834351": "gamma",
  "n06834458": "delta",
  "n06834565": "epsilon",
  "n06834674": "zeta",
  "n06834780": "eta",
  "n06834885": "theta",
  "n06834992": "iota",
  "n06835098": "kappa",
  "n06835206": "lambda",
  "n06835315": "mu",
  "n06835420": "nu",
  "n06835525": "xi",
  "n06835630": "omicron",
  "n06835740": "pi",
  "n06835845": "rho",
  "n06835951": "sigma",
  "n06836059": "tau",
  "n06836165": "upsilon",
  "n06836275": "phi",
  "n06836381": "chi, khi",
  "n06836493": "psi",
  "n06836599": "omega",
  "n06836714": "aleph",
  "n06836822": "beth",
  "n06836929": "gimel",
  "n06837037": "daleth",
  "n06837146": "he",
  "n06837251": "waw",
  "n06837357": "zayin",
  "n06837465": "heth",
  "n06837572": "teth",
  "n06837679": "yodh",
  "n06837787": "kaph",
  "n06837895": "lamedh",
  "n06838005": "mem",
  "n06838112": "nun",
  "n06838219": "samekh",
  "n06838329": "ayin",
  "n06838437": "pe",
  "n06838543": "sadhe",
  "n06838652": "qoph",
  "n06838760": "resh",
  "n06838868": "sin",
  "n06838975": "shin",
  "n06839083": "taw",
  "n06839190": "space, blank",
  "n06839411": "polyphone, polyphonic letter",
  "n06839637": "block letter, block capital",
  "n06839737": "scarlet letter",
  "n06839858": "phonetic symbol",
  "n06840047": "mathematical symbol",
  "n06840187": "rune, runic letter",
  "n06840442": "thorn",
  "n06840530": "pictograph",
  "n06840648": "ideogram, ideograph",
  "n06840890": "logogram, logograph",
  "n06841159": "radical",
  "n06841287": "stenograph",
  "n06841365": "punctuation, punctuation mark",
  "n06841873": "ampersand",
  "n06841987": "apostrophe",
  "n06842165": "brace",
  "n06842290": "bracket, square bracket",
  "n06842452": "bracket, angle bracket",
  "n06842660": "colon",
  "n06842852": "comma",
  "n06843017": "exclamation mark, exclamation point",
  "n06843148": "hyphen, dash",
  "n06843393": "parenthesis",
  "n06843520": "period, point, full stop, stop, full point",
  "n06843838": "suspension point",
  "n06844040": "question mark, interrogation point",
  "n06844199": "quotation mark, quote, inverted comma",
  "n06844427": "single quote",
  "n06844509": "double quotes",
  "n06844594": "scare quote",
  "n06844739": "semicolon",
  "n06844903": "solidus, slash, virgule, diagonal, stroke, separatrix",
  "n06845076": "swung dash",
  "n06845199": "company name",
  "n06845303": "domain name",
  "n06845599": "trade name, brand name, brand, marque",
  "n06851516": "label, recording label",
  "n06851742": "trademark",
  "n06855035": "authentication, hallmark, assay-mark",
  "n06855207": "stamp, impression",
  "n06855432": "imprint, embossment",
  "n06855563": "imprint",
  "n06855817": "revenue stamp, stamp",
  "n06855985": "seal",
  "n06856189": "phylactery, tefillin",
  "n06856487": "white feather",
  "n06856568": "scale, musical scale",
  "n06856884": "flourish, fanfare, tucket",
  "n06857122": "glissando",
  "n06857264": "swoop, slide",
  "n06857487": "gamut",
  "n06857591": "roulade",
  "n06857726": "tonic, keynote",
  "n06857867": "supertonic",
  "n06857986": "mediant",
  "n06858144": "subdominant",
  "n06858266": "dominant",
  "n06858384": "submediant",
  "n06858539": "subtonic, leading tone",
  "n06858674": "pedal point, pedal",
  "n06858779": "interval, musical interval",
  "n06859056": "tone, whole tone, step, whole step",
  "n06859175": "semitone, half step",
  "n06859309": "quarter tone, quarter-tone",
  "n06859401": "octave, musical octave",
  "n06859504": "third",
  "n06859674": "fourth",
  "n06859800": "fifth",
  "n06859925": "sixth",
  "n06860049": "seventh",
  "n06860177": "trill, shake",
  "n06860323": "diatonic scale",
  "n06860481": "ecclesiastical mode, Gregorian mode, church mode, medieval mode",
  "n06860699": "Greek mode",
  "n06860826": "major scale, major diatonic scale",
  "n06861020": "minor scale, minor diatonic scale",
  "n06861196": "chromatic scale",
  "n06861315": "gapped scale",
  "n06861435": "pentatonic scale, pentatone",
  "n06861630": "mode, musical mode",
  "n06861860": "staff, stave",
  "n06862059": "staff line",
  "n06862202": "space",
  "n06862399": "ledger line, leger line",
  "n06862562": "clef",
  "n06862805": "treble clef, treble staff, G clef",
  "n06862954": "bass clef, F clef",
  "n06863086": "alto clef, viola clef",
  "n06863209": "C clef",
  "n06863326": "soprano clef",
  "n06863442": "tenor clef",
  "n06863618": "key signature, signature",
  "n06863751": "key, tonality",
  "n06864014": "atonality, atonalism",
  "n06864192": "major key, major mode",
  "n06864308": "minor key, minor mode",
  "n06864407": "tonic key, home key",
  "n06864524": "time signature, musical time signature",
  "n06864725": "measure, bar",
  "n06864900": "double bar",
  "n06865049": "alla breve",
  "n06865177": "rest",
  "n06865345": "note, musical note, tone",
  "n06865953": "slur",
  "n06866124": "tie",
  "n06866332": "C",
  "n06866442": "C major, C major scale, scale of C major",
  "n06866599": "sharp",
  "n06866757": "double sharp",
  "n06866919": "flat",
  "n06867057": "double flat",
  "n06867218": "natural, cancel",
  "n06867345": "accidental",
  "n06867510": "fermata",
  "n06867675": "solmization, solmisation",
  "n06867880": "tonic solfa, solfa",
  "n06868043": "solfa syllable",
  "n06868309": "do, doh, ut",
  "n06868445": "re, ray",
  "n06868582": "mi",
  "n06868709": "fa",
  "n06868844": "sol, soh, so",
  "n06868986": "la, lah",
  "n06869129": "ti, te, si",
  "n06869271": "segno",
  "n06869428": "sforzando",
  "n06869610": "arpeggio",
  "n06869742": "sforzando",
  "n06869815": "middle C",
  "n06869951": "chord",
  "n06870211": "common chord, triad",
  "n06870349": "seventh chord",
  "n06870437": "passing note, passing tone",
  "n06870576": "whole note, semibreve",
  "n06870726": "whole rest",
  "n06870844": "half note, minim",
  "n06870965": "half rest",
  "n06871127": "quarter note, crotchet",
  "n06871262": "quarter rest",
  "n06871384": "eighth note, quaver",
  "n06871534": "sixteenth note, semiquaver",
  "n06871675": "thirty-second note, demisemiquaver",
  "n06871828": "sixty-fourth note, hemidemisemiquaver",
  "n06871983": "grace note, appoggiatura, acciaccatura",
  "n06872122": "singing voice",
  "n06872354": "bass, bass voice, basso",
  "n06872479": "basso profundo",
  "n06872562": "baritone, baritone voice",
  "n06872676": "tenor, tenor voice",
  "n06872785": "countertenor, alto",
  "n06872905": "contralto, alto",
  "n06873017": "mezzo-soprano, mezzo",
  "n06873139": "soprano",
  "n06873252": "visual communication",
  "n06873571": "visual signal",
  "n06873927": "watch fire",
  "n06874019": "light",
  "n06874185": "traffic light, traffic signal, stoplight",
  "n06874391": "green light, go-ahead",
  "n06874497": "red light",
  "n06874571": "red light, warning light",
  "n06874688": "idiot light",
  "n06874837": "yellow light",
  "n06874930": "flare, flash",
  "n06875094": "flag, signal flag",
  "n06875288": "pennant",
  "n06875392": "code flag, nautical signal flag",
  "n06875552": "blue peter",
  "n06875697": "sign language, signing",
  "n06875883": "finger spelling, fingerspelling",
  "n06876007": "ASL, American sign language",
  "n06876144": "sign",
  "n06876309": "gesture, motion",
  "n06876771": "gesticulation",
  "n06876892": "body language",
  "n06877008": "beck",
  "n06877078": "facial expression, facial gesture",
  "n06877381": "gape",
  "n06877509": "rictus",
  "n06877578": "grimace, face",
  "n06877742": "pout, moue, wry face",
  "n06877849": "frown, scowl",
  "n06877990": "simper",
  "n06878071": "smile, smiling, grin, grinning",
  "n06878404": "laugh",
  "n06878580": "smirk",
  "n06878706": "snarl",
  "n06878803": "straight face",
  "n06878934": "wink",
  "n06879056": "wince",
  "n06879180": "demonstration, demo",
  "n06879521": "display, show",
  "n06879766": "eye contact",
  "n06880013": "big stick",
  "n06880134": "gaudery, pomp",
  "n06880249": "expression, manifestation, reflection, reflexion",
  "n06880533": "exemplification, illustration",
  "n06880664": "emblem, allegory",
  "n06881224": "cupid",
  "n06881360": "donkey",
  "n06881494": "dove",
  "n06881563": "eagle",
  "n06881662": "elephant",
  "n06881798": "fasces",
  "n06882009": "national flag, ensign",
  "n06882138": "hammer and sickle",
  "n06882244": "red flag",
  "n06882333": "Star of David, Shield of David, Magen David, Mogen David, Solomon's seal",
  "n06882561": "badge",
  "n06882953": "merit badge",
  "n06883073": "insignia",
  "n06883274": "Agnus Dei, Paschal Lamb",
  "n06883383": "maple-leaf",
  "n06883460": "medallion",
  "n06883565": "spread eagle",
  "n06883725": "swastika, Hakenkreuz",
  "n06883956": "mantle",
  "n06884097": "Crown",
  "n06884337": "British Crown",
  "n06884520": "caduceus",
  "n06884670": "insignia of rank",
  "n06884790": "shoulder flash",
  "n06884954": "service stripe, hashmark, hash mark",
  "n06885083": "identification",
  "n06885389": "positive identification",
  "n06885906": "negative identification",
  "n06886273": "facial profiling",
  "n06886488": "fingerprint",
  "n06886655": "linguistic profiling",
  "n06886818": "profiling",
  "n06887055": "green card",
  "n06887278": "ID, I.D.",
  "n06887441": "personal identification number, PIN, PIN number",
  "n06887599": "projection",
  "n06887726": "display",
  "n06887962": "acting out",
  "n06888174": "array",
  "n06888345": "screening, showing, viewing",
  "n06888506": "preview",
  "n06888674": "preview, prevue, trailer",
  "n06888848": "sneak preview",
  "n06888944": "sight",
  "n06889138": "spectacle",
  "n06889330": "ostentation, fanfare, flash",
  "n06889591": "bravado, bluster",
  "n06889701": "exhibitionism",
  "n06889875": "ritz",
  "n06890000": "splurge",
  "n06890128": "pedantry",
  "n06890254": "flourish, brandish",
  "n06890373": "flourish",
  "n06890470": "flourish",
  "n06890577": "flourish",
  "n06890688": "paraph",
  "n06890846": "flaunt",
  "n06891022": "presentation",
  "n06891339": "unveiling",
  "n06891493": "performance, public presentation",
  "n06892016": "act, routine, number, turn, bit",
  "n06892358": "show-stopper, showstopper, stopper",
  "n06892534": "benefit",
  "n06892657": "benefit concert",
  "n06892775": "concert",
  "n06892979": "rock concert",
  "n06893065": "pianism",
  "n06893183": "play reading",
  "n06893285": "premiere",
  "n06893441": "recital",
  "n06893565": "rendition, rendering",
  "n06893772": "song and dance",
  "n06893885": "theatrical performance, theatrical, representation, histrionics",
  "n06894111": "matinee",
  "n06894243": "spectacular",
  "n06894371": "world premiere",
  "n06894544": "artificial language",
  "n06895176": "Antido",
  "n06895266": "Arulo",
  "n06895396": "Basic English",
  "n06895601": "Blaia Zimondal",
  "n06895684": "Esperantido",
  "n06895791": "Esperanto",
  "n06895940": "Europan",
  "n06896059": "Idiom Neutral",
  "n06896304": "Interlingua",
  "n06896500": "Ido",
  "n06896623": "Latinesce",
  "n06896716": "Latino",
  "n06896856": "Latino sine flexione",
  "n06896961": "Lingualumina",
  "n06897042": "Lingvo Kosmopolita",
  "n06897129": "Monario",
  "n06897205": "Nov-Esperanto",
  "n06897306": "Novial",
  "n06897381": "Nov-Latin",
  "n06897474": "Occidental",
  "n06897553": "Optez",
  "n06897627": "Pasigraphy",
  "n06897797": "Ro",
  "n06897976": "Romanal",
  "n06898052": "Solresol",
  "n06898129": "Volapuk",
  "n06898352": "programming language, programing language",
  "n06898794": "algebraic language",
  "n06898972": "algorithmic language",
  "n06899110": "application-oriented language, problem-oriented language",
  "n06899310": "assembly language",
  "n06899446": "command language, query language, search language",
  "n06899633": "computer language, computer-oriented language, machine language, machine-oriented language",
  "n06899845": "high-level language",
  "n06900005": "job-control language",
  "n06900156": "metalanguage",
  "n06900282": "multidimensional language",
  "n06900507": "object language, target language",
  "n06900684": "object-oriented programming language, object-oriented programing language",
  "n06901053": "Java",
  "n06901163": "one-dimensional language",
  "n06901333": "stratified language",
  "n06901471": "syntax language",
  "n06901591": "unstratified language",
  "n06901764": "ALGOL",
  "n06901936": "LISP, list-processing language",
  "n06902109": "LISP program",
  "n06902193": "Prolog, logic programing, logic programming",
  "n06902362": "FORTRAN",
  "n06902521": "FORTRAN program",
  "n06902611": "COBOL",
  "n06902696": "C",
  "n06902831": "C program",
  "n06902909": "BASIC",
  "n06903115": "Pascal",
  "n06903255": "upgrade",
  "n06903396": "native language",
  "n06903519": "indigenous language",
  "n06903699": "substrate, substratum",
  "n06903972": "superstrate, superstratum",
  "n06904171": "natural language, tongue",
  "n06904748": "mother tongue, maternal language, first language",
  "n06904943": "tone language, tonal language",
  "n06905141": "contour language",
  "n06905243": "register language",
  "n06905358": "creole",
  "n06905513": "Haitian Creole",
  "n06905680": "pidgin",
  "n06905828": "Chinook Jargon, Oregon Jargon",
  "n06906028": "Sango",
  "n06906116": "lingua franca, interlanguage, koine",
  "n06906439": "Amerind, Amerindian language, American-Indian language, American Indian, Indian",
  "n06906971": "Algonquian, Algonquin, Algonquian language",
  "n06907567": "Atakapa, Atakapan, Attacapa, Attacapan",
  "n06907728": "Athapaskan, Athapascan, Athabaskan, Athabascan, Athapaskan language",
  "n06908048": "Abnaki",
  "n06908159": "Algonkian, Algonkin",
  "n06908291": "Arapaho, Arapahoe",
  "n06908401": "Biloxi",
  "n06908494": "Blackfoot",
  "n06908605": "Catawba",
  "n06908700": "Cheyenne",
  "n06908801": "Chiwere",
  "n06908968": "Iowa, Ioway",
  "n06909079": "Missouri",
  "n06909190": "Oto, Otoe",
  "n06909298": "Cree",
  "n06909391": "Crow",
  "n06909478": "Dakota",
  "n06909571": "Delaware",
  "n06909672": "Dhegiha",
  "n06909848": "Fox",
  "n06909932": "Hidatsa, Gros Ventre",
  "n06910039": "Hunkpapa",
  "n06910134": "Illinois",
  "n06910238": "Haida",
  "n06910323": "Kansa, Kansas",
  "n06910423": "Kickapoo",
  "n06910517": "Malecite, Maleseet",
  "n06910639": "Massachuset, Massachusetts",
  "n06910755": "Menomini, Menominee",
  "n06910868": "Micmac",
  "n06910958": "Mohican, Mahican",
  "n06911067": "Nanticoke",
  "n06911180": "Ofo",
  "n06911265": "Oglala, Ogalala",
  "n06911366": "Ojibwa, Ojibway, Chippewa",
  "n06911484": "Omaha",
  "n06911575": "Osage",
  "n06911666": "Pamlico",
  "n06911758": "Ponca, Ponka",
  "n06911857": "Potawatomi",
  "n06911962": "Powhatan",
  "n06912056": "Quapaw",
  "n06912149": "Shawnee",
  "n06912248": "Alabama",
  "n06912340": "Chickasaw",
  "n06912436": "Choctaw, Chahta",
  "n06912537": "Hitchiti",
  "n06912638": "Koasati",
  "n06912737": "Muskogee",
  "n06912838": "Santee",
  "n06912931": "Seminole",
  "n06913025": "Tlingit",
  "n06913121": "Tutelo",
  "n06913214": "Winnebago",
  "n06913313": "Muskhogean, Muskhogean language, Muskogean, Muskogean language",
  "n06913635": "Na-Dene",
  "n06913768": "Mosan",
  "n06913948": "Chemakuan",
  "n06914069": "Chemakum",
  "n06914167": "Salish, Salishan",
  "n06914331": "Skagit",
  "n06914423": "Wakashan, Wakashan language",
  "n06914612": "Kwakiutl",
  "n06914709": "Nootka",
  "n06914802": "Shoshone",
  "n06914930": "Comanche",
  "n06915031": "Hopi",
  "n06915124": "Paiute",
  "n06915221": "Ute",
  "n06915313": "Shoshonean, Shoshonean language, Shoshonian, Shoshonian language",
  "n06915601": "Caddo, Caddoan, Caddoan language",
  "n06915821": "Arikara, Aricara",
  "n06915927": "Pawnee",
  "n06916021": "Wichita",
  "n06916117": "Cherokee",
  "n06916217": "Cayuga",
  "n06916313": "Mohawk",
  "n06916409": "Seneca",
  "n06916505": "Oneida",
  "n06916601": "Onondaga",
  "n06916701": "Tuscarora",
  "n06916803": "Iroquoian, Iroquois, Iroquoian language",
  "n06917083": "Quechua, Quechuan, Quechuan language, Kechua, Kechuan",
  "n06917280": "Guarani",
  "n06917392": "Maraco, Maracan language",
  "n06917497": "Tupi",
  "n06917602": "Tupi-Guarani, Tupi-Guarani language",
  "n06917764": "Arawak, Arawakan",
  "n06917926": "Carib, Caribbean language",
  "n06918042": "Eskimo-Aleut, Eskimo-Aleut language",
  "n06918215": "Eskimo, Esquimau",
  "n06918312": "Aleut",
  "n06918396": "Uto-Aztecan, Uto-Aztecan language",
  "n06918640": "Pima",
  "n06918734": "Aztecan",
  "n06918832": "Nahuatl",
  "n06918932": "Cahita",
  "n06919023": "Tatahumara",
  "n06919122": "Zapotec, Zapotecan",
  "n06919215": "Maya, Mayan, Mayan language",
  "n06919433": "Apache",
  "n06919548": "Chiricahua Apache",
  "n06919630": "San Carlos Apache",
  "n06919712": "Navaho, Navajo",
  "n06919818": "Hupa",
  "n06919911": "Mattole",
  "n06920010": "Chipewyan, Chippewyan, Chippewaian",
  "n06920129": "Siouan, Siouan language",
  "n06920497": "Tanoan, Tanoan language",
  "n06920665": "Kiowa",
  "n06920756": "Hokan, Hoka",
  "n06920994": "Chimariko",
  "n06921092": "Esselen",
  "n06921186": "Kulanapan",
  "n06921300": "Pomo",
  "n06921392": "Quoratean",
  "n06921506": "Karok",
  "n06921593": "Shastan",
  "n06921755": "Achomawi",
  "n06921853": "Atsugewi",
  "n06921951": "Shasta",
  "n06922045": "Yuman",
  "n06922389": "Akwa'ala",
  "n06922485": "Cochimi",
  "n06922579": "Cocopa, Cocopah",
  "n06922681": "Diegueno",
  "n06922777": "Havasupai",
  "n06922875": "Kamia",
  "n06922965": "Kiliwa, Kiliwi",
  "n06923066": "Maricopa",
  "n06923182": "Mohave, Mojave",
  "n06923283": "Walapai, Hualapai, Hualpai",
  "n06923398": "Yavapai",
  "n06923492": "Yuma",
  "n06923580": "Yanan",
  "n06923704": "Yahi",
  "n06923792": "Yana",
  "n06923880": "Penutian",
  "n06924214": "Copehan",
  "n06924376": "Patwin",
  "n06924468": "Wintun",
  "n06924560": "Costanoan",
  "n06924659": "Mariposan, Yokuts",
  "n06924790": "Moquelumnan, Miwok",
  "n06924895": "Pujunan, Maidu",
  "n06924996": "Chinookan, Chinook",
  "n06925103": "Kalapooian, Kalapuyan",
  "n06925214": "Kusan",
  "n06925305": "Shahaptian, Sahaptin",
  "n06925435": "Nez Perce",
  "n06925538": "Takilman, Takelma",
  "n06925644": "Tsimshian",
  "n06925743": "Kekchi",
  "n06925833": "Mam",
  "n06925917": "Yucatec, Yucateco",
  "n06926020": "Quiche",
  "n06926112": "Cakchiquel",
  "n06926212": "Altaic, Altaic language",
  "n06926458": "Turki, Turkic, Turko-Tatar, Turkic language",
  "n06926798": "Turkish",
  "n06926889": "Turkmen, Turkoman, Turcoman",
  "n06927007": "Azerbaijani",
  "n06927128": "Kazak, Kazakh",
  "n06927228": "Tatar",
  "n06927363": "Uzbek, Uzbeg, Uzbak, Usbek, Usbeg",
  "n06927486": "Uighur, Uigur, Uygur",
  "n06927645": "Yakut",
  "n06927736": "Kirghiz, Kirgiz, Khirghiz",
  "n06927851": "Karakalpak",
  "n06927952": "Chuvash",
  "n06928047": "Chagatai, Jagatai, Jaghatai, Eastern Turki",
  "n06928234": "Chukchi, Chukchi language",
  "n06928430": "Tungusic, Tungusic language",
  "n06928610": "Tungus, Tunguz, Evenki, Ewenki",
  "n06928744": "Manchu",
  "n06928839": "Mongolian, Mongolic, Mongolic language",
  "n06929008": "Khalkha, Khalka, Kalka",
  "n06929171": "Korean",
  "n06929279": "Japanese",
  "n06929459": "Ryukyuan",
  "n06929618": "Sinitic, Sinitic language",
  "n06929742": "Chinese",
  "n06930093": "Mandarin, Mandarin Chinese, Mandarin dialect, Beijing dialect",
  "n06930298": "Wu, Wu dialect, Shanghai dialect",
  "n06930427": "Yue, Yue dialect, Cantonese, Cantonese dialect",
  "n06930633": "Min, Min dialect, Fukien, Fukkianese, Hokkianese, Amoy, Taiwanese",
  "n06930804": "Hakka, Hakka dialect",
  "n06930934": "Sino-Tibetan, Sino-Tibetan language",
  "n06931199": "Tibeto-Burman, Tibeto-Burman language",
  "n06931499": "Qiang, Qiangic",
  "n06931605": "Bai, Baic",
  "n06931724": "Himalayish",
  "n06931891": "Kamarupan",
  "n06932130": "Karen, Karenic",
  "n06932265": "Lolo-Burmese, Burmese-Yi",
  "n06932435": "Burmese",
  "n06932537": "Loloish",
  "n06932734": "Lisu",
  "n06932803": "Hani, Akha",
  "n06932879": "Lahu",
  "n06932948": "Lolo, Yi",
  "n06933022": "Kachin, Kachinic",
  "n06933185": "Jinghpo, Jinghpaw, Chingpo",
  "n06933279": "Kuki, Chin, Kuki-Chin",
  "n06933429": "Naga",
  "n06933546": "Mikir-Meithei",
  "n06933689": "Bodo-Garo, Barish",
  "n06933824": "Miri, Mirish, Abor, Dafla",
  "n06933931": "Tibetan",
  "n06934020": "Newari",
  "n06934132": "Kadai, Kam-Tai, Kadai language",
  "n06934309": "Kam-Sui",
  "n06934389": "Tai",
  "n06934848": "White Tai",
  "n06934933": "Red Tai",
  "n06935016": "Tai Dam, Black Tai",
  "n06935111": "Tai Nuea, Chinese Shan, Dehong Dai",
  "n06935223": "Tai Long, Shan",
  "n06935314": "Tai Lue, Xishuangbanna Dai",
  "n06935417": "Tai Yuan, Kam Muang",
  "n06935513": "Khuen",
  "n06935594": "Lao",
  "n06935741": "Khamti",
  "n06935823": "Southern Tai",
  "n06935911": "Tay",
  "n06935990": "Nung",
  "n06936070": "Tho",
  "n06936149": "Thai, Siamese, Central Thai",
  "n06936290": "Bouyei, Buyi",
  "n06936379": "Zhuang",
  "n06936461": "Yay",
  "n06936540": "Saek",
  "n06936620": "Austro-Asiatic, Austro-Asiatic language, Munda-Mon-Khmer",
  "n06936823": "Munda",
  "n06936948": "Mon-Khmer",
  "n06937098": "Hmong, Hmong language, Miao",
  "n06937229": "Vietnamese, Annamese, Annamite",
  "n06937348": "Khmer",
  "n06937441": "Mon",
  "n06937531": "Austronesian, Austronesian language",
  "n06937768": "Malayo-Polynesian, Polynesian",
  "n06937985": "Oceanic, Eastern Malayo-Polynesian",
  "n06938190": "Tongan",
  "n06938294": "Tahitian",
  "n06938386": "Maori",
  "n06938493": "Hawaiian",
  "n06938623": "Fijian",
  "n06938729": "Western Malayo-Polynesian",
  "n06938887": "Malay",
  "n06939051": "Malaysian, Bahasa Malaysia, Bahasa Melayu, Bahasa Kebangsaan",
  "n06939198": "Indonesian, Bahasa Indonesia, Bahasa",
  "n06939431": "Javanese",
  "n06939542": "Sundanese",
  "n06939656": "Balinese",
  "n06939756": "Philippine, Filipino",
  "n06940001": "Tagalog",
  "n06940109": "Cebuan, Cebuano",
  "n06940290": "Australian, Aboriginal Australian",
  "n06940502": "Dyirbal, Jirrbal",
  "n06940601": "Walbiri, Warlpiri",
  "n06940701": "Formosan",
  "n06940854": "Tayalic, Atayalic",
  "n06940969": "Tsouic",
  "n06941041": "Paiwanic",
  "n06941115": "Papuan, Papuan language",
  "n06941341": "Khoisan, Khoisan language",
  "n06941479": "Khoikhoin, Khoikhoi, Hottentot",
  "n06941644": "Indo-European, Indo-European language, Indo-Hittite",
  "n06942094": "Proto-Indo European, PIE",
  "n06942252": "Albanian",
  "n06942419": "Gheg, Gheg dialect",
  "n06942550": "Tosk, Tosk dialect",
  "n06942699": "Armenian, Armenian language",
  "n06942874": "Illyrian",
  "n06943027": "Thraco-Phrygian",
  "n06943221": "Thracian",
  "n06943378": "Phrygian",
  "n06943558": "Balto-Slavic, Balto-Slavic language, Balto-Slavonic",
  "n06943771": "Slavic, Slavic language, Slavonic, Slavonic language",
  "n06944156": "Old Church Slavonic, Old Church Slavic, Church Slavic, Old Bulgarian",
  "n06944348": "Russian",
  "n06944480": "Belarusian, Byelorussian, White Russian",
  "n06944623": "Ukrainian",
  "n06944720": "Polish",
  "n06944820": "Slovak",
  "n06944911": "Czech",
  "n06945010": "Slovene",
  "n06945095": "Serbo-Croat, Serbo-Croatian",
  "n06945344": "Sorbian, Lusatian",
  "n06945473": "Macedonian",
  "n06945569": "Bulgarian",
  "n06945679": "Baltic, Baltic language",
  "n06946017": "Old Prussian",
  "n06946199": "Lithuanian",
  "n06946355": "Latvian, Lettish",
  "n06946497": "Germanic, Germanic language",
  "n06946823": "West Germanic, West Germanic language",
  "n06947032": "English, English language",
  "n06947479": "American English, American language, American",
  "n06947658": "African American Vernacular English, AAVE, African American English, Black English, Black English Vernacular, Black Vernacular, Black Vernacular English, Ebonics",
  "n06948017": "cockney",
  "n06948149": "geordie",
  "n06948260": "King's English, Queen's English",
  "n06948396": "Received Pronunciation",
  "n06948761": "Middle English",
  "n06948943": "East Midland",
  "n06949121": "West Midland",
  "n06949207": "Northern",
  "n06949326": "Kentish",
  "n06949407": "Southwestern, West Saxon",
  "n06949506": "Modern English",
  "n06949591": "Old English, Anglo-Saxon",
  "n06949762": "West Saxon",
  "n06949852": "Anglian",
  "n06949946": "Kentish, Jutish",
  "n06950049": "Oxford English",
  "n06950209": "Scottish, Scots, Scots English",
  "n06950400": "Lallans, Scottish Lallans",
  "n06950528": "German, High German, German language",
  "n06950886": "Old High German",
  "n06950973": "Middle High German",
  "n06951067": "Yiddish",
  "n06952115": "Pennsylvania Dutch",
  "n06952249": "Low German, Plattdeutsch",
  "n06952400": "Old Saxon",
  "n06952480": "Middle Low German",
  "n06952572": "Dutch",
  "n06952705": "Flemish, Flemish dialect",
  "n06952861": "Afrikaans, Taal, South African Dutch",
  "n06953054": "Proto-Norse",
  "n06953167": "Old Norse",
  "n06953332": "Old Icelandic",
  "n06953471": "Edda",
  "n06953731": "Scandinavian, Scandinavian language, Nordic, Norse, North Germanic, North Germanic language",
  "n06954048": "Danish",
  "n06954183": "Icelandic",
  "n06954303": "Norwegian",
  "n06954461": "Bokmal, Bokmaal, Dano-Norwegian",
  "n06954640": "Riksmal, Riksmaal",
  "n06954768": "Nynorsk, New Norwegian, Landsmal, Landsmaal",
  "n06954925": "Swedish",
  "n06955087": "Faroese, Faeroese",
  "n06955242": "Frisian",
  "n06955418": "Old Frisian",
  "n06955560": "East Germanic, East Germanic language",
  "n06955706": "Gothic",
  "n06955931": "Ural-Altaic",
  "n06956129": "Uralic, Uralic language",
  "n06956287": "Finno-Ugric, Finno-Ugrian",
  "n06956544": "Fennic, Finnic, Non-Ugric",
  "n06956794": "Udmurt, Votyak",
  "n06956896": "Permic",
  "n06957042": "Komi, Zyrian",
  "n06957140": "Volgaic",
  "n06957287": "Cheremis, Cheremiss, Mari",
  "n06957403": "Mordva, Mordvin, Mordvinian",
  "n06957524": "Baltic-Finnic",
  "n06957768": "Livonian",
  "n06957896": "Estonian, Esthonian",
  "n06958063": "Karelian, Carelian",
  "n06958178": "Ludian",
  "n06958255": "Finnish, Suomi",
  "n06958417": "Veps, Vepse, Vepsian",
  "n06958522": "Ingrian",
  "n06958615": "Ugric, Ugrian",
  "n06958836": "Hungarian, Magyar",
  "n06959018": "Khanty, Ostyak",
  "n06959140": "Mansi, Vogul",
  "n06959261": "Lappic, Lappish",
  "n06959427": "Lapp, Sami, Saami, Same, Saame",
  "n06959584": "Samoyedic, Samoyed",
  "n06959788": "Nenets, Nentsi, Nentsy, Yurak-Samoyed",
  "n06959932": "Enets, Entsi, Entsy, Yenisei, Yenisei-Samoyed, Yeniseian",
  "n06960083": "Nganasan",
  "n06960180": "Selkup, Ostyak-Samoyed",
  "n06960298": "Celtic, Celtic language",
  "n06960566": "Gaelic, Goidelic, Erse",
  "n06960778": "Irish, Irish Gaelic",
  "n06960948": "Old Irish",
  "n06961033": "Middle Irish",
  "n06961122": "Scottish Gaelic, Scots Gaelic",
  "n06961221": "Manx",
  "n06961399": "Brythonic, Brittanic",
  "n06961557": "Welsh, Cymric",
  "n06961681": "Cornish",
  "n06961771": "Breton",
  "n06961853": "Italic, Italic language",
  "n06962060": "Osco-Umbrian",
  "n06962244": "Umbrian",
  "n06962350": "Oscan",
  "n06962454": "Sabellian",
  "n06962600": "Latin",
  "n06962954": "Old Latin",
  "n06963082": "classical Latin",
  "n06963292": "Low Latin",
  "n06963429": "Vulgar Latin",
  "n06963571": "Late Latin, Biblical Latin",
  "n06963704": "Medieval Latin",
  "n06963822": "Neo-Latin, New Latin",
  "n06963951": "Romance, Romance language, Latinian language",
  "n06964247": "Italian",
  "n06964485": "Old Italian",
  "n06964600": "Sardinian",
  "n06964788": "Tuscan",
  "n06964901": "French",
  "n06965249": "Langue d'oil, Langue d'oil French",
  "n06965406": "Langue d'oc, Langue d'oc French",
  "n06965580": "Old French",
  "n06965698": "Norman-French, Norman French, Old North French",
  "n06965834": "Anglo-French, Anglo-Norman",
  "n06965961": "Canadian French",
  "n06966070": "Walloon",
  "n06966190": "Provencal, Occitan",
  "n06966310": "Portuguese",
  "n06966454": "Galician",
  "n06966695": "Basque",
  "n06966825": "Spanish",
  "n06967185": "Castilian",
  "n06967282": "Judeo-Spanish, Ladino",
  "n06967428": "Mexican Spanish",
  "n06967529": "Catalan",
  "n06967710": "Rhaeto-Romance, Rhaeto-Romanic",
  "n06967935": "Friulian, Friuli",
  "n06968052": "Ladin",
  "n06968174": "Romansh, Rumansh",
  "n06968343": "Romanian, Rumanian",
  "n06968454": "Elamitic, Elamite, Susian",
  "n06968601": "Kassite, Cassite",
  "n06968707": "Tocharian",
  "n06968909": "Turfan, East Tocharian, Turfan dialect",
  "n06969018": "Kuchean, West Tocharian, Kuchean dialect",
  "n06969129": "Sanskrit, Sanskritic language",
  "n06969705": "Sindhi",
  "n06969822": "Romany, Gypsy",
  "n06969916": "Urdu",
  "n06970103": "Hindi",
  "n06970373": "Hindustani, Hindoostani, Hindostani",
  "n06970510": "Bihari",
  "n06970645": "Magadhan",
  "n06970784": "Assamese, Asamiya",
  "n06970946": "Bengali, Bangla",
  "n06971123": "Oriya",
  "n06971281": "Marathi, Mahratti",
  "n06971454": "Gujarati, Gujerati",
  "n06971605": "Punjabi, Panjabi",
  "n06971741": "Sinhalese, Singhalese, Sinhala",
  "n06971872": "Indo-Iranian, Indo-Iranian language",
  "n06972090": "Indic, Indo-Aryan",
  "n06972311": "Dard, Dardic, Dardic language",
  "n06972558": "Shina",
  "n06972654": "Khowar",
  "n06972756": "Kafiri",
  "n06972874": "Kashmiri",
  "n06972985": "Nepali",
  "n06973092": "Prakrit",
  "n06973334": "Pali",
  "n06973505": "Prakrit",
  "n06973610": "Iranian, Iranian language",
  "n06973941": "Avestan, Zend",
  "n06974047": "Gathic",
  "n06974127": "Persian, Farsi",
  "n06974283": "Dari, Dari Persian",
  "n06974390": "Tajiki, Tajik, Tadzhik",
  "n06974558": "Kurdish",
  "n06974687": "Balochi, Baluchi",
  "n06974846": "Pahlavi, Pehlevi",
  "n06975007": "Parthian",
  "n06975132": "Pashto, Pashtu, Paxto, Afghani, Afghan",
  "n06975368": "Ossete",
  "n06975469": "Scythian",
  "n06975594": "Anatolian, Anatolian language",
  "n06975902": "Hittite",
  "n06976087": "Lycian",
  "n06976161": "Luwian, Luvian",
  "n06976244": "Lydian",
  "n06976318": "Palaic",
  "n06976392": "Greek, Hellenic, Hellenic language",
  "n06976680": "Modern Greek, New Greek",
  "n06976833": "Romaic, Demotic",
  "n06976959": "Katharevusa",
  "n06977087": "Late Greek",
  "n06977190": "Medieval Greek, Middle Greek, Byzantine Greek",
  "n06977328": "Koine",
  "n06977434": "Ancient Greek",
  "n06977610": "Attic, Ionic, Ionic dialect, Classical Greek",
  "n06977800": "Aeolic, Aeolic dialect, Eolic",
  "n06977948": "Arcadic, Arcadic dialect",
  "n06978068": "Doric, Doric dialect",
  "n06978180": "Caucasian, Caucasian language",
  "n06978422": "Chechen",
  "n06978527": "Circassian",
  "n06978656": "Abkhazian, Abkhasian",
  "n06978766": "Georgian",
  "n06978904": "Ubykh",
  "n06979014": "Dravidian, Dravidic, Dravidian language",
  "n06979249": "South Dravidian",
  "n06979492": "Irula",
  "n06979645": "Kota, Kotar",
  "n06979743": "Toda",
  "n06979853": "Badaga",
  "n06979957": "Kannada, Kanarese",
  "n06980085": "Tulu",
  "n06980175": "Malayalam",
  "n06980315": "Tamil",
  "n06980465": "South-Central Dravidian",
  "n06980721": "Telugu",
  "n06980837": "Savara",
  "n06980971": "Gondi",
  "n06981085": "Pengo",
  "n06981187": "Manda",
  "n06981289": "Kui",
  "n06981401": "Kuvi",
  "n06981498": "Central Dravidian",
  "n06981706": "Kolami",
  "n06981818": "Naiki",
  "n06981920": "Parji",
  "n06982022": "Ollari",
  "n06982125": "Gadaba",
  "n06982221": "North Dravidian",
  "n06982391": "Kurux",
  "n06982487": "Malto",
  "n06982581": "Brahui",
  "n06982697": "Hausa, Haussa",
  "n06982855": "Bole, Bolanci",
  "n06982989": "Angas",
  "n06983114": "Ron, Bokkos, Daffo",
  "n06983225": "Bade",
  "n06983320": "Warji",
  "n06983416": "Zaar, Sayanci",
  "n06983521": "West Chadic",
  "n06983795": "Tera, Pidlimdi, Yamaltu",
  "n06983895": "Bura, Pabir",
  "n06983997": "Higi, Kapsiki",
  "n06984101": "Mandara, Wandala",
  "n06984247": "Matakam, Mafa",
  "n06984351": "Sukur",
  "n06984446": "Daba, Kola, Musgoi",
  "n06984556": "Bata",
  "n06984650": "Kotoko",
  "n06984746": "Musgu, Munjuk, Mulwi",
  "n06984858": "Gidar",
  "n06984953": "Biu-Mandara",
  "n06985310": "Somrai, Sibine",
  "n06985404": "Nancere",
  "n06985490": "Kera",
  "n06985573": "Dangla, Dangaleat",
  "n06985722": "Mokulu",
  "n06985807": "Sokoro",
  "n06985892": "East Chadic",
  "n06986100": "Masa",
  "n06986276": "Chad, Chadic, Chadic language",
  "n06986558": "Afroasiatic, Afro-Asiatic, Afroasiatic language, Afrasian, Afrasian language, Hamito-Semitic",
  "n06986894": "Semitic",
  "n06987124": "Hebrew",
  "n06987409": "Modern Hebrew",
  "n06987525": "Akkadian",
  "n06987659": "Assyrian Akkadian, Assyrian",
  "n06987812": "Amharic, Ethiopian language",
  "n06988057": "Arabic, Arabic language",
  "n06988307": "Aramaic",
  "n06988540": "Biblical Aramaic",
  "n06988684": "Assyrian Neo-Aramaic, Assyrian",
  "n06988808": "Mandaean, Mandean",
  "n06988913": "Maltese, Maltese language, Malti",
  "n06989146": "Canaanitic, Canaanitic language",
  "n06989325": "Canaanite",
  "n06989473": "Phoenician",
  "n06989657": "Punic",
  "n06989769": "Ugaritic",
  "n06989869": "Hamitic, Hamitic language",
  "n06990000": "Egyptian",
  "n06990193": "Demotic, Demotic script",
  "n06990371": "Coptic",
  "n06990544": "Berber",
  "n06990741": "Tuareg",
  "n06990836": "Cushitic",
  "n06991004": "Somali",
  "n06991117": "Omotic",
  "n06991277": "Niger-Kordofanian, Niger-Kordofanian language",
  "n06991622": "Kordofanian",
  "n06991764": "Niger-Congo",
  "n06991980": "Bantu, Bantoid language",
  "n06992600": "Chichewa",
  "n06992709": "ChiMwini",
  "n06992807": "Chishona",
  "n06992929": "Fang",
  "n06993015": "Gikuyu",
  "n06993108": "Giriama",
  "n06993225": "Herero",
  "n06993348": "Kamba",
  "n06993445": "Kichaga, Chaga, Chagga",
  "n06993573": "Kinyarwanda",
  "n06993647": "Kiswahili",
  "n06993719": "Kongo",
  "n06993870": "Luba, Tshiluba",
  "n06993977": "LuGanda",
  "n06994089": "Luyia",
  "n06994157": "Mashi",
  "n06994225": "Mwera",
  "n06994329": "Nguni",
  "n06994488": "Ndebele, Matabele",
  "n06994608": "Swazi",
  "n06994700": "Xhosa",
  "n06994792": "Zulu",
  "n06994918": "Nyamwezi",
  "n06995016": "Pokomo",
  "n06995135": "Shona",
  "n06995222": "Sotho",
  "n06995439": "Umbundu",
  "n06995526": "Sesotho, Basuto",
  "n06995664": "Tswana, Setswana, Sechuana",
  "n06995792": "Swahili",
  "n06996029": "Tonga",
  "n06996160": "Gur, Voltaic",
  "n06996309": "West African",
  "n06996510": "Fula, Ful, Fulani, Peul",
  "n06996755": "Serer",
  "n06996894": "Wolof",
  "n06997012": "Mande",
  "n06997168": "Kwa",
  "n06997368": "Yoruba, Aku",
  "n06997486": "Akan",
  "n06997587": "Ewe",
  "n06997697": "Nilo-Saharan, Nilo-Saharan language",
  "n06997938": "Chari-Nile",
  "n06998108": "Nilotic, Nilotic language",
  "n06998322": "Dinka",
  "n06998392": "Luo",
  "n06998460": "Masai",
  "n06998530": "Saharan",
  "n06998631": "Songhai",
  "n06998748": "artwork, art, graphics, nontextual matter",
  "n06999045": "graphic design",
  "n06999233": "illustration",
  "n06999436": "picture, pictorial matter",
  "n06999647": "figure, fig",
  "n06999802": "chart",
  "n07000065": "plot",
  "n07000195": "graph, graphical record",
  "n07000716": "frequency-response curve, frequency-response characteristic",
  "n07000941": "curve",
  "n07001065": "characteristic curve, characterisic function",
  "n07001294": "organization chart",
  "n07001446": "color chart",
  "n07001547": "color circle, color wheel",
  "n07001717": "bar chart, bar graph",
  "n07001865": "histogram",
  "n07002022": "eye chart",
  "n07002146": "flip chart",
  "n07002312": "pie chart",
  "n07002459": "star chart",
  "n07002599": "profile",
  "n07002863": "population profile",
  "n07002992": "tabulation, tabular matter",
  "n07003119": "drawing",
  "n07003352": "comic strip, cartoon strip, strip, funnies",
  "n07003568": "frame",
  "n07003672": "ballistocardiogram",
  "n07003807": "echoencephalogram",
  "n07003935": "echocardiogram",
  "n07004057": "electrocardiogram, cardiogram, EKG, ECG",
  "n07004241": "electroencephalogram, encephalogram, EEG",
  "n07004422": "electromyogram, EMG",
  "n07004564": "electroretinogram",
  "n07004738": "Laffer curve",
  "n07004962": "learning curve",
  "n07005158": "myogram",
  "n07005270": "radiation pattern, radiation diagram, pattern",
  "n07005523": "lobe",
  "n07005715": "major lobe",
  "n07005897": "tachogram",
  "n07006018": "thermogram",
  "n07006119": "dramaturgy, dramatic art, dramatics, theater, theatre",
  "n07006712": "stage",
  "n07006951": "production",
  "n07007171": "theatrical production, staging",
  "n07007341": "coup de theatre",
  "n07007444": "coup de theatre",
  "n07007556": "summer stock",
  "n07007684": "dramatic composition, dramatic work",
  "n07007945": "play, drama, dramatic play",
  "n07008544": "afterpiece",
  "n07008680": "fragment",
  "n07008849": "Grand Guignol",
  "n07008947": "hiatus",
  "n07009042": "snatch, bit",
  "n07009161": "theater of the absurd",
  "n07009421": "prologue",
  "n07009538": "playlet",
  "n07009640": "act",
  "n07009808": "scene",
  "n07009946": "script, book, playscript",
  "n07010275": "promptbook, prompt copy",
  "n07010393": "continuity",
  "n07010541": "dialogue, dialog",
  "n07010692": "duologue",
  "n07010821": "actor's line, speech, words",
  "n07011075": "aside",
  "n07011209": "cue",
  "n07011387": "monologue",
  "n07011529": "soliloquy",
  "n07011689": "throwaway",
  "n07011803": "prompt, prompting",
  "n07012025": "libretto",
  "n07012135": "scenario",
  "n07012279": "screenplay",
  "n07012414": "shooting script",
  "n07012534": "line",
  "n07012979": "line",
  "n07013269": "orphan",
  "n07013400": "spiel, patter, line of gab",
  "n07013549": "string",
  "n07013736": "string of words, word string, linguistic string",
  "n07013933": "substring",
  "n07014029": "act",
  "n07014187": "lipogram",
  "n07014320": "effusion, gush, outburst, blowup, ebullition",
  "n07014596": "acting out",
  "n07014752": "cry",
  "n07014854": "explosion",
  "n07014997": "flare",
  "n07015151": "collocation",
  "n07015242": "high-five",
  "n07015400": "closet drama",
  "n07015510": "comedy",
  "n07015841": "black comedy",
  "n07015928": "commedia dell'arte",
  "n07016098": "dark comedy",
  "n07016250": "farce, farce comedy, travesty",
  "n07016428": "high comedy",
  "n07016542": "low comedy",
  "n07016648": "melodrama",
  "n07016797": "seriocomedy, tragicomedy",
  "n07016948": "tragedy",
  "n07017161": "tragicomedy",
  "n07017379": "situation comedy, sitcom",
  "n07017534": "special",
  "n07017733": "situation comedy, sitcom",
  "n07017877": "slapstick",
  "n07017999": "burlesque",
  "n07018211": "exode",
  "n07018313": "miracle play",
  "n07018444": "morality play",
  "n07018604": "mystery play",
  "n07018724": "Passion play",
  "n07018824": "satyr play",
  "n07018931": "play",
  "n07019172": "musical, musical comedy, musical theater",
  "n07019396": "curtain raiser",
  "n07019500": "galanty show, shadow show, shadow play",
  "n07019633": "puppet show, puppet play",
  "n07019743": "minstrel show",
  "n07019866": "revue, review",
  "n07020017": "follies",
  "n07020121": "Ziegfeld Follies",
  "n07020239": "variety show, variety",
  "n07020423": "vaudeville, music hall",
  "n07020538": "dance",
  "n07020763": "choreography",
  "n07020895": "music",
  "n07024607": "pizzicato",
  "n07024705": "monophony, monophonic music, monody",
  "n07024929": "polyphony, polyphonic music, concerted music",
  "n07025151": "polytonality, polytonalism",
  "n07025302": "popularism",
  "n07025419": "counterpoint",
  "n07025604": "black music, African-American music",
  "n07025900": "classical music, classical, serious music",
  "n07026239": "chamber music",
  "n07026352": "opera",
  "n07026646": "comic opera, opera bouffe, bouffe, opera comique",
  "n07026827": "grand opera",
  "n07026920": "musical drama",
  "n07027090": "operetta, light opera",
  "n07027180": "harmony, musical harmony",
  "n07027458": "harmonization, harmonisation",
  "n07027651": "reharmonization, reharmonisation",
  "n07027805": "four-part harmony",
  "n07027942": "preparation",
  "n07028221": "resolution",
  "n07028373": "tune, melody, air, strain, melodic line, line, melodic phrase",
  "n07028797": "leitmotiv, leitmotif",
  "n07028964": "theme song",
  "n07029088": "signature, signature tune, theme song",
  "n07029247": "theme, melodic theme, musical theme, idea",
  "n07029581": "obbligato, obligato",
  "n07029682": "motif, motive",
  "n07029819": "statement",
  "n07030012": "variation",
  "n07030174": "inversion",
  "n07030376": "augmentation",
  "n07030549": "diminution",
  "n07030718": "part, voice",
  "n07030976": "part music",
  "n07031144": "homophony",
  "n07031276": "primo",
  "n07031401": "secondo",
  "n07031534": "voice part",
  "n07031636": "canto",
  "n07031752": "accompaniment, musical accompaniment, backup, support",
  "n07032026": "descant, discant",
  "n07032206": "vamp",
  "n07032292": "bass, bass part",
  "n07032426": "ground bass",
  "n07032556": "figured bass, basso continuo, continuo, thorough bass",
  "n07032753": "crossover",
  "n07033007": "religious music, church music",
  "n07033245": "antiphon, antiphony",
  "n07033433": "gradual",
  "n07033613": "Mass",
  "n07033753": "Mass",
  "n07033913": "Requiem",
  "n07034035": "Shema",
  "n07034409": "processional, prosodion",
  "n07034516": "antiphonary, antiphonal",
  "n07034634": "chant",
  "n07034865": "Hallel",
  "n07035061": "Hare Krishna",
  "n07035153": "plainsong, plainchant, Gregorian chant",
  "n07035305": "cantus firmus",
  "n07035420": "religious song",
  "n07035598": "spiritual, Negro spiritual",
  "n07035747": "carol, Christmas carol",
  "n07035870": "hymn, anthem",
  "n07036203": "doxology",
  "n07036328": "chorale, choral",
  "n07036443": "canticle",
  "n07036546": "Dies Irae",
  "n07036699": "hymeneal",
  "n07036768": "Internationale",
  "n07036862": "paean, pean",
  "n07037059": "Magnificat",
  "n07037232": "recessional",
  "n07037384": "Te Deum",
  "n07037465": "musical composition, opus, composition, piece, piece of music",
  "n07038400": "musical arrangement, arrangement",
  "n07038615": "orchestration",
  "n07038767": "instrumental music",
  "n07038910": "instrumentation",
  "n07039056": "realization, realisation",
  "n07039238": "recapitulation",
  "n07039478": "finale, coda",
  "n07039620": "intermezzo",
  "n07039770": "allegro",
  "n07039949": "allegretto",
  "n07040148": "andante",
  "n07040292": "intermezzo",
  "n07040413": "introit",
  "n07040543": "prelude",
  "n07040693": "chorale prelude",
  "n07040820": "overture",
  "n07040939": "solo",
  "n07041125": "voluntary",
  "n07041344": "postlude",
  "n07041451": "duet, duette, duo",
  "n07041595": "trio",
  "n07041688": "quartet, quartette",
  "n07041795": "quintet, quintette",
  "n07041902": "sextet, sextette, sestet",
  "n07042023": "septet, septette",
  "n07042137": "octet, octette",
  "n07042249": "cantata, oratorio",
  "n07042405": "Messiah",
  "n07042498": "bagatelle",
  "n07042586": "divertimento, serenade",
  "n07042735": "keen",
  "n07042862": "canon",
  "n07043026": "enigma canon, enigmatic canon, enigmatical canon, riddle canon",
  "n07043275": "concerto",
  "n07043389": "concerto grosso",
  "n07043518": "etude",
  "n07043675": "fugue",
  "n07043824": "pastorale, pastoral, idyll, idyl",
  "n07043968": "rondo, rondeau",
  "n07044088": "sonata",
  "n07044276": "piano sonata",
  "n07044353": "toccata",
  "n07044543": "fantasia",
  "n07044680": "sonatina",
  "n07044760": "symphony, symphonic music",
  "n07044917": "passage, musical passage",
  "n07045245": "intro",
  "n07045353": "phrase, musical phrase",
  "n07045519": "ligature",
  "n07045637": "ostinato",
  "n07045788": "riff",
  "n07045928": "cadence",
  "n07046030": "plagal cadence, amen cadence",
  "n07046218": "cadenza",
  "n07046339": "movement",
  "n07046543": "largo",
  "n07046737": "larghetto",
  "n07046917": "scherzo",
  "n07047011": "suite",
  "n07047165": "partita",
  "n07047275": "partita",
  "n07047373": "symphonic poem, tone poem",
  "n07047505": "medley, potpourri, pastiche",
  "n07047679": "nocturne, notturno",
  "n07047804": "adagio",
  "n07048000": "song, vocal",
  "n07048627": "study",
  "n07048782": "antiphony",
  "n07048928": "anthem",
  "n07049055": "national anthem",
  "n07049205": "Marseillaise",
  "n07049291": "The Star-Spangled Banner",
  "n07049514": "aria",
  "n07049634": "arietta, short aria",
  "n07049713": "ballad, lay",
  "n07049849": "minstrelsy",
  "n07049931": "barcarole, barcarolle",
  "n07050042": "chantey, chanty, sea chantey, shanty",
  "n07050177": "refrain, chorus",
  "n07050379": "tra-la, tra-la-la",
  "n07050503": "ditty",
  "n07050619": "dirge, coronach, lament, requiem, threnody",
  "n07050827": "drinking song",
  "n07050952": "folk song, folksong, folk ballad",
  "n07051185": "blues",
  "n07051441": "fado",
  "n07051517": "blue note",
  "n07051620": "lied",
  "n07051728": "love song, love-song",
  "n07051851": "lullaby, cradlesong, berceuse",
  "n07051975": "lyric, words, language",
  "n07052291": "stanza",
  "n07052619": "love lyric",
  "n07052700": "oldie, golden oldie",
  "n07052799": "partsong",
  "n07052925": "madrigal",
  "n07053089": "round, troll",
  "n07053364": "prothalamion, prothalamium",
  "n07053491": "roundelay",
  "n07053606": "scolion, banquet song",
  "n07053732": "serenade",
  "n07053884": "torch song",
  "n07053993": "work song",
  "n07054122": "shivaree, chivaree, charivari, callithump, callathump, belling",
  "n07054336": "ballet",
  "n07054433": "dance music",
  "n07054994": "beguine",
  "n07055103": "bolero",
  "n07055203": "carioca",
  "n07055295": "conga",
  "n07055401": "flamenco",
  "n07055502": "gavotte",
  "n07055612": "habanera",
  "n07055720": "hornpipe",
  "n07055805": "jig, gigue",
  "n07055923": "landler",
  "n07056021": "mazurka",
  "n07056113": "minuet",
  "n07056265": "paso doble",
  "n07056406": "pavane, pavan",
  "n07056504": "polka",
  "n07056593": "quadrille",
  "n07056680": "reel",
  "n07056764": "rumba, rhumba",
  "n07056895": "samba",
  "n07057001": "saraband",
  "n07057095": "schottische",
  "n07057196": "serialism, serial music",
  "n07057385": "syncopation",
  "n07057539": "twelve-tone music, 12-tone music, twelve-tone system, 12-tone system",
  "n07057829": "tango",
  "n07057948": "tarantella",
  "n07058064": "techno",
  "n07058184": "waltz",
  "n07058296": "marching music, march",
  "n07058468": "military march, military music, martial music",
  "n07058668": "quickstep",
  "n07058762": "pibroch",
  "n07058871": "processional march, recessional march",
  "n07059028": "funeral march, dead march",
  "n07059150": "wedding march",
  "n07059255": "popular music, popular music genre",
  "n07059626": "disco, disco music",
  "n07059821": "macumba",
  "n07059962": "pop music, pop",
  "n07060167": "folk music, ethnic music, folk",
  "n07060440": "country music, country and western, C and W",
  "n07060697": "dance music, danceroom music, ballroom music",
  "n07060844": "ragtime, rag",
  "n07060976": "jazz",
  "n07061180": "kwela",
  "n07061334": "gospel, gospel singing",
  "n07061677": "doo-wop",
  "n07061968": "soul",
  "n07062190": "bluegrass",
  "n07062315": "hillbilly music",
  "n07062451": "square-dance music",
  "n07062550": "zydeco",
  "n07062697": "jazz",
  "n07063101": "bop, bebop",
  "n07063249": "boogie, boogie-woogie",
  "n07063413": "cool jazz",
  "n07063585": "funk",
  "n07063775": "hot jazz",
  "n07063921": "modern jazz, new jazz, neo jazz",
  "n07064055": "rap, rap music, hip-hop",
  "n07064315": "rhythm and blues, R and B",
  "n07064537": "rockabilly",
  "n07064715": "rock 'n' roll, rock'n'roll, rock-and-roll, rock and roll, rock, rock music",
  "n07065149": "heavy metal, heavy metal music",
  "n07065333": "progressive rock, art rock",
  "n07065562": "psychedelic rock, acid rock",
  "n07065740": "punk rock, punk",
  "n07065932": "trad",
  "n07066042": "swing, swing music, jive",
  "n07066285": "reggae",
  "n07066459": "skiffle",
  "n07066659": "expressive style, style",
  "n07067591": "address",
  "n07067760": "catch",
  "n07067876": "analysis",
  "n07068048": "bathos",
  "n07068133": "black humor, black humour",
  "n07068324": "Gongorism",
  "n07068473": "conceit",
  "n07068631": "development",
  "n07068844": "device",
  "n07069002": "doctorspeak",
  "n07069074": "ecobabble",
  "n07069210": "eloquence, fluency, smoothness",
  "n07069517": "euphuism",
  "n07069632": "Eurobabble",
  "n07069747": "flatness",
  "n07069948": "formulation, expression",
  "n07070321": "gobbledygook",
  "n07070429": "grandiosity, magniloquence, ornateness, grandiloquence, rhetoric",
  "n07070779": "headlinese",
  "n07070883": "honorific",
  "n07071017": "jargon",
  "n07071250": "journalese",
  "n07071366": "legalese",
  "n07071483": "manner of speaking, speech, delivery",
  "n07071942": "music genre, musical genre, genre, musical style",
  "n07072195": "officialese",
  "n07072337": "pathos",
  "n07072434": "prose",
  "n07072587": "psychobabble",
  "n07072698": "rhetoric",
  "n07072838": "saltiness, coarseness",
  "n07073071": "self-expression",
  "n07073208": "articulation, voice",
  "n07073447": "archaism, archaicism",
  "n07075006": "boilerplate",
  "n07075172": "colloquialism",
  "n07080778": "mot juste",
  "n07080868": "verbalization, verbalisation",
  "n07081043": "verbalization, verbalisation",
  "n07081177": "parlance, idiom",
  "n07081345": "Americanism",
  "n07081473": "Anglicism, Briticism, Britishism",
  "n07081646": "Gallicism",
  "n07081739": "wording, diction, phrasing, phraseology, choice of words, verbiage",
  "n07082025": "paralanguage, paralinguistic communication",
  "n07082198": "tongue",
  "n07082346": "sharp tongue",
  "n07082444": "shibboleth",
  "n07082573": "tone, tone of voice",
  "n07082819": "note",
  "n07082972": "roundness, rotundity",
  "n07083136": "undertone",
  "n07083246": "elocution",
  "n07083441": "barrage, bombardment, outpouring, onslaught",
  "n07083732": "prosody, inflection",
  "n07083958": "modulation, inflection",
  "n07084166": "intonation, modulation, pitch contour",
  "n07084428": "intonation pattern",
  "n07084560": "monotone, drone, droning",
  "n07084747": "monotone",
  "n07084927": "singsong",
  "n07085072": "caesura",
  "n07085210": "enjambment, enjambement",
  "n07085375": "stress, emphasis, accent",
  "n07085786": "accentuation",
  "n07085961": "tonic accent, pitch accent",
  "n07086088": "word stress, word accent",
  "n07086215": "sentence stress",
  "n07086323": "rhythm, speech rhythm",
  "n07086518": "rhythm, beat, musical rhythm",
  "n07086861": "backbeat",
  "n07086972": "downbeat",
  "n07087102": "upbeat, offbeat",
  "n07087223": "syncopation",
  "n07087346": "recitative",
  "n07087508": "arioso",
  "n07087641": "transition, modulation",
  "n07087777": "bombast, fustian, rant, claptrap, blah",
  "n07087940": "sesquipedality",
  "n07088053": "sensationalism, luridness",
  "n07088319": "technobabble",
  "n07088438": "terseness",
  "n07088645": "turn of phrase, turn of expression",
  "n07088868": "conceit",
  "n07089024": "conciseness, concision, pithiness, succinctness",
  "n07089276": "crispness",
  "n07089431": "brevity",
  "n07089531": "laconism, laconicism",
  "n07089622": "vein",
  "n07089751": "verboseness, verbosity",
  "n07090019": "verbiage, verbalism",
  "n07090108": "prolixity, prolixness, windiness, long-windedness, wordiness",
  "n07090379": "circumlocution, periphrasis, ambage",
  "n07090573": "turgidity, turgidness, flatulence",
  "n07090721": "repetitiveness, repetitiousness",
  "n07090938": "redundancy",
  "n07091087": "pleonasm",
  "n07091214": "tautology",
  "n07091385": "tautology",
  "n07091587": "abbreviation",
  "n07091737": "apocope",
  "n07091902": "acronym",
  "n07092158": "writing style, literary genre, genre",
  "n07092356": "form",
  "n07092592": "poetry, poesy, verse",
  "n07093158": "heroic poetry, epic poetry",
  "n07093273": "poetry",
  "n07093489": "versification",
  "n07093603": "versification",
  "n07093759": "versification",
  "n07093895": "poetic rhythm, rhythmic pattern, prosody",
  "n07094093": "meter, metre, measure, beat, cadence",
  "n07094355": "catalexis",
  "n07094508": "scansion",
  "n07094621": "sprung rhythm",
  "n07094731": "common measure, common meter",
  "n07094843": "metrical foot, foot, metrical unit",
  "n07095148": "dactyl",
  "n07095280": "iamb, iambus",
  "n07095390": "anapest, anapaest",
  "n07095516": "amphibrach",
  "n07095653": "trochee",
  "n07095775": "spondee",
  "n07095895": "pyrrhic, dibrach",
  "n07096029": "tetrameter",
  "n07096142": "pentameter",
  "n07096237": "hexameter",
  "n07096330": "octameter",
  "n07096425": "octosyllable",
  "n07096569": "decasyllable",
  "n07096661": "rhyme, rime",
  "n07096995": "internal rhyme",
  "n07097094": "alliteration, initial rhyme, beginning rhyme, head rhyme",
  "n07097346": "assonance, vowel rhyme",
  "n07097548": "consonance, consonant rhyme",
  "n07097707": "double rhyme",
  "n07097831": "rhyme royal",
  "n07097965": "ottava rima",
  "n07098093": "eye rhyme",
  "n07098193": "rhetorical device",
  "n07098989": "anacoluthia, anacoluthon",
  "n07099135": "asyndeton",
  "n07099271": "repetition",
  "n07099598": "anadiplosis, reduplication",
  "n07099752": "epanalepsis",
  "n07099844": "epanodos",
  "n07099965": "epanodos",
  "n07100113": "epiphora, epistrophe",
  "n07100253": "gemination",
  "n07100387": "ploce",
  "n07100522": "polyptoton",
  "n07100678": "epanaphora, anaphora",
  "n07100815": "anaphora",
  "n07100942": "symploce",
  "n07101140": "anastrophe, inversion",
  "n07101250": "antiphrasis",
  "n07101406": "antithesis",
  "n07101557": "antinomasia",
  "n07101649": "apophasis",
  "n07101778": "aposiopesis",
  "n07101937": "apostrophe",
  "n07102034": "catachresis",
  "n07102271": "chiasmus",
  "n07102373": "climax",
  "n07102483": "conversion",
  "n07102593": "dramatic irony",
  "n07102802": "ecphonesis, exclamation",
  "n07102945": "emphasis",
  "n07103106": "enallage",
  "n07103258": "epanorthosis",
  "n07103402": "epiplexis",
  "n07103558": "hendiadys",
  "n07103671": "hypallage",
  "n07103802": "hyperbaton",
  "n07103912": "hypozeugma",
  "n07104020": "hypozeuxis",
  "n07104149": "hysteron proteron",
  "n07104292": "litotes, meiosis",
  "n07104574": "onomatopoeia",
  "n07104733": "paralepsis, paraleipsis, paralipsis, preterition",
  "n07104912": "paregmenon",
  "n07105045": "polysyndeton",
  "n07105238": "prolepsis",
  "n07105342": "wellerism",
  "n07105475": "trope, figure of speech, figure, image",
  "n07106113": "conceit",
  "n07106246": "irony",
  "n07106502": "hyperbole, exaggeration",
  "n07106651": "kenning",
  "n07106800": "metaphor",
  "n07107083": "dead metaphor, frozen metaphor",
  "n07107381": "mixed metaphor",
  "n07107522": "synesthetic metaphor",
  "n07107676": "metonymy",
  "n07107896": "metalepsis",
  "n07108010": "oxymoron",
  "n07108123": "personification, prosopopoeia",
  "n07108282": "simile",
  "n07108453": "synecdoche",
  "n07108657": "syllepsis",
  "n07108807": "zeugma",
  "n07109019": "auditory communication",
  "n07109196": "speech, speech communication, spoken communication, spoken language, language, voice communication, oral communication",
  "n07109730": "words",
  "n07109847": "utterance, vocalization",
  "n07110457": "speech",
  "n07110615": "voice, vocalization, vocalisation, vocalism, phonation, vox",
  "n07111047": "phone, speech sound, sound",
  "n07111510": "morphophoneme",
  "n07111711": "phoneme",
  "n07111933": "allophone",
  "n07112103": "ablaut",
  "n07112282": "grade, gradation",
  "n07112364": "diphthong",
  "n07112550": "vowel, vowel sound",
  "n07112805": "accentual system, prosodic system",
  "n07112942": "consonant system, consonantal system",
  "n07113080": "morphophonemic system",
  "n07113192": "phonemic system",
  "n07113337": "phonological system, phonologic system",
  "n07113491": "syllabicity",
  "n07113625": "tense system",
  "n07113732": "tone system, tonal system",
  "n07113903": "vowel system, vocalism",
  "n07114023": "schwa, shwa",
  "n07114154": "murmur vowel, murmur",
  "n07114284": "stem vowel, thematic vowel",
  "n07114409": "semivowel, glide",
  "n07114535": "palatal",
  "n07114712": "vowel",
  "n07114870": "vowel point",
  "n07115021": "consonant",
  "n07115381": "consonant",
  "n07115493": "alveolar consonant, dental consonant, alveolar, dental",
  "n07115684": "obstruent",
  "n07115914": "stop consonant, stop, occlusive, plosive consonant, plosive speech sound, plosive",
  "n07116304": "implosion",
  "n07116443": "plosion, explosion",
  "n07116641": "affrication",
  "n07116758": "aspirate",
  "n07116869": "aspiration",
  "n07117007": "labial consonant, labial",
  "n07117193": "labiodental consonant, labiodental",
  "n07117333": "bilabial",
  "n07117472": "labial stop",
  "n07117595": "glottal stop, glottal plosive, glottal catch",
  "n07117788": "epenthesis",
  "n07118002": "nasalization, nasalisation",
  "n07118210": "suction stop, click",
  "n07118367": "continuant consonant, continuant",
  "n07118554": "fricative consonant, fricative, spirant",
  "n07118747": "sibilant, sibilant consonant",
  "n07118990": "affricate, affricate consonant, affricative",
  "n07119211": "nasal consonant, nasal",
  "n07119358": "orinasal phone, orinasal",
  "n07119521": "lingual",
  "n07119643": "liquid",
  "n07119776": "geminate",
  "n07119897": "surd, voiceless consonant",
  "n07120025": "velar, velar consonant",
  "n07120176": "guttural, guttural consonant, pharyngeal, pharyngeal consonant",
  "n07120364": "sonant, voiced sound",
  "n07120524": "cry, outcry, call, yell, shout, vociferation",
  "n07121157": "cry, yell",
  "n07121361": "bellow, bellowing, holla, holler, hollering, hollo, holloa, roar, roaring, yowl",
  "n07121904": "blue murder",
  "n07122018": "catcall",
  "n07122118": "clamor, clamoring, clamour, clamouring, hue and cry",
  "n07122409": "halloo",
  "n07122555": "hoot",
  "n07122639": "hosanna",
  "n07122730": "noise",
  "n07123012": "scream, screaming, shriek, shrieking, screech, screeching",
  "n07123288": "whoop",
  "n07123404": "war cry, war whoop, rallying cry, battle cry",
  "n07123552": "yelling, shouting",
  "n07123710": "yodel",
  "n07123870": "boo, hoot, Bronx cheer, hiss, raspberry, razzing, razz, snort, bird",
  "n07124172": "blasphemy",
  "n07124340": "obscenity, smut, vulgarism, filth, dirty word",
  "n07124736": "bawdry, bawdy",
  "n07124928": "scatology",
  "n07125096": "curse, curse word, expletive, oath, swearing, swearword, cuss",
  "n07125367": "croak, croaking",
  "n07125523": "exclamation, exclaiming",
  "n07125786": "devil, deuce, dickens",
  "n07125958": "ejaculation, interjection",
  "n07126116": "expostulation",
  "n07126228": "expletive",
  "n07126383": "groan, moan",
  "n07126521": "hem, ahem",
  "n07126734": "howl, howling, ululation",
  "n07127006": "laugh, laughter",
  "n07127252": "mumble",
  "n07127350": "cachinnation",
  "n07127451": "cackle",
  "n07127563": "chortle, chuckle",
  "n07127693": "giggle",
  "n07127790": "guffaw, belly laugh",
  "n07127911": "hee-haw, horselaugh, ha-ha, haw-haw",
  "n07128060": "snicker, snort, snigger",
  "n07128225": "titter",
  "n07128322": "paging",
  "n07128527": "profanity",
  "n07128692": "pronunciation, orthoepy",
  "n07128946": "pronunciation",
  "n07129202": "sibilation, assibilation",
  "n07129422": "exultation, rejoicing, jubilation",
  "n07129602": "sigh, suspiration",
  "n07129758": "snarl",
  "n07129867": "speaking, speech production",
  "n07130050": "speech",
  "n07130183": "sputter, splutter",
  "n07130341": "whisper, whispering, susurration, voicelessness",
  "n07130580": "stage whisper",
  "n07130774": "rasp, rasping",
  "n07130918": "mispronunciation",
  "n07131022": "homograph",
  "n07131169": "homophone",
  "n07131366": "homophony",
  "n07131511": "accent, speech pattern",
  "n07131741": "drawl",
  "n07131854": "articulation",
  "n07132226": "retroflection, retroflexion",
  "n07132415": "enunciation, diction",
  "n07132634": "mumbling",
  "n07132729": "syncope, syncopation",
  "n07132937": "sandhi",
  "n07133143": "thickness",
  "n07133313": "tongue twister",
  "n07133477": "trill",
  "n07133701": "conversation",
  "n07134179": "crossfire",
  "n07134287": "phatic speech, phatic communication",
  "n07134445": "intercourse, social intercourse",
  "n07134575": "communion, sharing",
  "n07134706": "exchange",
  "n07134850": "chat, confab, confabulation, schmooze, schmoose",
  "n07135080": "chitchat, chit-chat, chit chat, small talk, gab, gabfest, gossip, tittle-tattle, chin wag, chin-wag, chin wagging, chin-wagging, causerie",
  "n07135450": "gossiping, gossipmongering",
  "n07135627": "scandalmongering",
  "n07135734": "talk, talking",
  "n07136098": "cant, pious platitude",
  "n07136206": "dialogue, dialog, duologue",
  "n07136315": "heart-to-heart",
  "n07136469": "shmooze",
  "n07136578": "shop talk",
  "n07136711": "wind, malarkey, malarky, idle words, jazz, nothingness",
  "n07136940": "yak, yack, yakety-yak, chatter, cackle",
  "n07137129": "prate, prattle, idle talk, blether, chin music",
  "n07137315": "nothings",
  "n07137461": "sweet nothings, honeyed words",
  "n07137622": "commerce",
  "n07137733": "colloquy",
  "n07137807": "detail",
  "n07137950": "dilation",
  "n07138085": "discussion, treatment, discourse",
  "n07138504": "indirect discourse",
  "n07138736": "direct discourse, direct quotation",
  "n07138915": "consideration",
  "n07139151": "expatiation",
  "n07139316": "talk",
  "n07139532": "reconsideration",
  "n07139700": "exhortation",
  "n07139873": "expression, verbal expression, verbalism",
  "n07140237": "cold turkey",
  "n07140348": "congratulation, felicitation",
  "n07140659": "discussion, give-and-take, word",
  "n07140978": "argument, argumentation, debate",
  "n07141319": "logomachy",
  "n07141437": "parley",
  "n07141537": "rap",
  "n07141644": "rap session",
  "n07141789": "second-hand speech",
  "n07141922": "table talk",
  "n07142005": "telephone conversation",
  "n07142107": "tete-a-tete",
  "n07142242": "pillow talk",
  "n07142365": "deliberation",
  "n07142566": "conference, group discussion",
  "n07142924": "bull session",
  "n07143044": "colloquy",
  "n07143137": "consultation",
  "n07143450": "sidebar",
  "n07143624": "consultation, audience, interview",
  "n07143869": "panel discussion",
  "n07144039": "postmortem, post-mortem",
  "n07144190": "public discussion, ventilation",
  "n07144416": "huddle, powwow",
  "n07144571": "backgrounder",
  "n07144834": "press conference, news conference",
  "n07145026": "pretrial, pretrial conference",
  "n07145314": "round table, roundtable, round-table conference",
  "n07145508": "session",
  "n07145783": "teach-in",
  "n07145958": "teleconference, teleconferencing",
  "n07146190": "sitting",
  "n07146300": "clinic",
  "n07146534": "reading clinic",
  "n07146640": "basketball clinic",
  "n07146784": "baseball clinic",
  "n07146980": "hockey clinic",
  "n07147116": "executive session, closed session",
  "n07147267": "hearing",
  "n07147511": "confirmation hearing",
  "n07147733": "skull session",
  "n07147902": "special session",
  "n07148022": "tutorial",
  "n07148192": "negotiation, dialogue, talks",
  "n07148573": "diplomacy, diplomatic negotiations",
  "n07148828": "dollar diplomacy",
  "n07148938": "power politics, gunboat diplomacy",
  "n07149107": "recognition",
  "n07149348": "shuttle diplomacy",
  "n07149582": "Strategic Arms Limitation Talks, SALT",
  "n07149836": "bargaining",
  "n07150023": "collective bargaining",
  "n07150138": "haggle, haggling, wrangle, wrangling",
  "n07150328": "holdout",
  "n07150499": "horse trading",
  "n07150644": "mediation",
  "n07150850": "arbitration",
  "n07151122": "conciliation",
  "n07151285": "umpirage",
  "n07151380": "saying, expression, locution",
  "n07151892": "anatomical reference, anatomical",
  "n07152027": "southernism",
  "n07152151": "sound bite",
  "n07152259": "motto, slogan, catchword, shibboleth",
  "n07152463": "catchphrase, catch phrase",
  "n07152590": "mantra",
  "n07152752": "war cry, rallying cry, battle cry, cry, watchword",
  "n07152948": "maxim, axiom",
  "n07153130": "aphorism, apothegm, apophthegm",
  "n07153385": "gnome",
  "n07153502": "Murphy's Law, Sod's Law",
  "n07153641": "moralism",
  "n07153727": "epigram, quip",
  "n07153838": "proverb, adage, saw, byword",
  "n07154046": "platitude, cliche, banality, commonplace, bromide",
  "n07154243": "truism",
  "n07154330": "idiom, idiomatic expression, phrasal idiom, set phrase, phrase",
  "n07154666": "ruralism, rusticism",
  "n07154760": "agrapha",
  "n07154868": "sumpsimus",
  "n07155081": "non-standard speech",
  "n07155426": "baby talk, babytalk",
  "n07155531": "baby talk, babytalk, motherese",
  "n07155661": "dialect, idiom, accent",
  "n07156091": "eye dialect",
  "n07156219": "patois",
  "n07156375": "localism",
  "n07156497": "regionalism",
  "n07156693": "idiolect",
  "n07156819": "monologue",
  "n07157000": "telegraphese",
  "n07157123": "vernacular",
  "n07157273": "slang, cant, jargon, lingo, argot, patois, vernacular",
  "n07159276": "rhyming slang",
  "n07159467": "slang, slang expression, slang term",
  "n07159791": "spell, magic spell, magical spell, charm",
  "n07160116": "incantation, conjuration",
  "n07160296": "invocation",
  "n07160424": "hex, jinx, curse, whammy",
  "n07160635": "dictation",
  "n07160752": "soliloquy, monologue",
  "n07160883": "speech act",
  "n07161429": "proposal, proposition",
  "n07161613": "contract offer",
  "n07161741": "marriage proposal, proposal of marriage, marriage offer, proposal",
  "n07161912": "proposition",
  "n07162059": "question",
  "n07162194": "proposal",
  "n07162424": "counterproposal",
  "n07162545": "hypothesis",
  "n07162680": "suggestion, proposition, proffer",
  "n07162975": "introduction",
  "n07163110": "re-introduction",
  "n07163272": "first reading",
  "n07163381": "second reading",
  "n07163593": "motion, question",
  "n07163803": "previous question",
  "n07163988": "hint, intimation, breath",
  "n07164163": "touch, trace, ghost",
  "n07164349": "overture, advance, approach, feeler",
  "n07164546": "offer, offering",
  "n07164972": "counteroffer",
  "n07165086": "bid, tender",
  "n07165298": "overbid",
  "n07165410": "buyout bid",
  "n07165506": "prospectus",
  "n07165778": "preliminary prospectus, red herring",
  "n07166101": "tender offer",
  "n07166281": "reward",
  "n07166432": "rights offering, rights issue",
  "n07166851": "special",
  "n07167041": "price",
  "n07167186": "peace offering, olive branch",
  "n07167326": "twofer",
  "n07167415": "presentation",
  "n07167578": "submission, entry",
  "n07167954": "filing",
  "n07168131": "command, bid, bidding, dictation",
  "n07168486": "countermand",
  "n07168623": "order",
  "n07168983": "marching orders",
  "n07169098": "summons",
  "n07169242": "word",
  "n07169353": "call up",
  "n07169480": "commission, charge, direction",
  "n07169743": "misdirection",
  "n07169848": "commandment",
  "n07169970": "Decalogue, Ten Commandments",
  "n07170080": "directive",
  "n07170282": "Presidential Directive",
  "n07170467": "injunction",
  "n07170572": "behest",
  "n07170660": "open sesame",
  "n07170753": "interpretation",
  "n07171206": "clarification, elucidation, illumination",
  "n07171513": "disambiguation",
  "n07171666": "lexical disambiguation",
  "n07171785": "eisegesis",
  "n07171940": "exegesis",
  "n07172117": "ijtihad",
  "n07172557": "text",
  "n07172756": "expansion, enlargement, elaboration",
  "n07172979": "embellishment, embroidery",
  "n07173258": "literal interpretation",
  "n07173389": "letter",
  "n07173585": "version",
  "n07173766": "reading",
  "n07173959": "construction, twist",
  "n07174138": "reconstruction",
  "n07174260": "popularization, popularisation",
  "n07174433": "misinterpretation, misunderstanding, mistaking",
  "n07174785": "imbroglio",
  "n07174877": "misconstrual, misconstruction",
  "n07175102": "misreading",
  "n07175241": "agreement",
  "n07175575": "assent, acquiescence",
  "n07175863": "informed consent",
  "n07176073": "acceptance",
  "n07176243": "concession, conceding, yielding",
  "n07176499": "bye, pass",
  "n07176682": "concurrence, concurrency",
  "n07176804": "accord, conformity, accordance",
  "n07176962": "connivance, collusion",
  "n07177111": "cahoot",
  "n07177192": "accession, assenting",
  "n07177437": "accommodation",
  "n07177622": "conclusion",
  "n07177772": "reservation",
  "n07177924": "settlement",
  "n07178182": "out-of-court settlement",
  "n07178337": "property settlement",
  "n07178525": "accord and satisfaction",
  "n07178698": "severance agreement",
  "n07178839": "golden handshake",
  "n07178992": "suicide pact",
  "n07179192": "modus vivendi",
  "n07179342": "compromise",
  "n07179609": "Missouri Compromise",
  "n07179820": "subscription",
  "n07179943": "ratification, confirmation",
  "n07180183": "harmony, concord, concordance",
  "n07180372": "second, secondment, endorsement, indorsement",
  "n07180570": "citation",
  "n07180787": "disagreement",
  "n07181043": "confrontation, encounter, showdown, face-off",
  "n07181208": "dissidence",
  "n07181358": "dissent",
  "n07181453": "nonconformity",
  "n07181546": "discord, dissension",
  "n07181713": "confrontation",
  "n07181842": "division, variance",
  "n07181935": "dispute, difference, difference of opinion, conflict",
  "n07182367": "straw man, strawman",
  "n07182485": "argy-bargy, argle-bargle",
  "n07182614": "firestorm",
  "n07182744": "sparring",
  "n07182888": "special pleading",
  "n07183000": "collision",
  "n07183151": "controversy, contention, contestation, disputation, disceptation, tilt, argument, arguing",
  "n07183660": "polemic",
  "n07183853": "gap",
  "n07184024": "generation gap",
  "n07184149": "quarrel, wrangle, row, words, run-in, dustup",
  "n07184391": "fight",
  "n07184545": "affray, altercation, fracas",
  "n07184652": "batrachomyomachia",
  "n07184735": "bicker, bickering, spat, tiff, squabble, pettifoggery, fuss",
  "n07184965": "bust-up",
  "n07185076": "offer, offering",
  "n07185325": "request, asking",
  "n07185668": "notification, notice",
  "n07185870": "wish, indirect request",
  "n07186148": "invitation",
  "n07186528": "bidding, summons",
  "n07186661": "invite",
  "n07186828": "entreaty, prayer, appeal",
  "n07187150": "adjuration",
  "n07187297": "demagoguery, demagogy",
  "n07187486": "flag waving, jingoism",
  "n07187638": "supplication, plea",
  "n07187773": "solicitation",
  "n07187996": "beggary, begging, mendicancy",
  "n07188238": "touch",
  "n07188385": "importunity, urgency, urging",
  "n07188570": "suit",
  "n07188685": "courtship, wooing, courting, suit",
  "n07188979": "bundling",
  "n07189130": "prayer, petition, orison",
  "n07189411": "benediction, blessing",
  "n07189562": "benison",
  "n07189633": "collect",
  "n07189779": "commination",
  "n07189932": "deprecation",
  "n07190039": "grace, blessing, thanksgiving",
  "n07190188": "intercession",
  "n07190290": "invocation, supplication",
  "n07190474": "rogation",
  "n07190584": "requiescat",
  "n07190693": "call",
  "n07190840": "recall",
  "n07190941": "charge, billing",
  "n07191162": "presentment",
  "n07191279": "demand",
  "n07191569": "challenge",
  "n07191696": "ultimatum",
  "n07191777": "insistence, insisting",
  "n07191950": "purism",
  "n07192129": "call, claim",
  "n07192240": "requisition",
  "n07192511": "call",
  "n07192661": "margin call, call",
  "n07192845": "wage claim, pay claim",
  "n07192990": "trick or treat",
  "n07193184": "questioning, inquiring",
  "n07193405": "challenge",
  "n07193596": "question, inquiry, enquiry, query, interrogation",
  "n07193958": "interrogation, examination, interrogatory",
  "n07194293": "catechism",
  "n07194499": "deposition",
  "n07194676": "inquisition",
  "n07194811": "third degree",
  "n07194950": "cross-examination",
  "n07195241": "direct examination",
  "n07195404": "redirect examination, reexamination",
  "n07195630": "cross-question",
  "n07195765": "leading question",
  "n07195969": "yes-no question",
  "n07196075": "interview",
  "n07196405": "job interview, employment interview",
  "n07196575": "telephone interview",
  "n07196682": "question, interrogation, interrogative, interrogative sentence",
  "n07197021": "examination, exam, test",
  "n07197537": "bar examination, bar exam",
  "n07197889": "comprehensive examination, comprehensive, comp",
  "n07198119": "entrance examination, entrance exam",
  "n07198276": "final examination, final exam, final",
  "n07198437": "litmus test",
  "n07198605": "midterm examination, midterm exam, midterm",
  "n07198757": "pop quiz",
  "n07198846": "oral, oral exam, oral examination, viva voce, viva",
  "n07199013": "preliminary examination, preliminary exam, prelim",
  "n07199191": "quiz",
  "n07199328": "test paper, examination paper, exam paper, question sheet",
  "n07199456": "tripos",
  "n07199565": "reply, response",
  "n07199814": "non sequitur",
  "n07199922": "rejoinder, retort, return, riposte, replication, comeback, counter",
  "n07200290": "echo",
  "n07200421": "echolalia",
  "n07200527": "answer",
  "n07200692": "Urim and Thummim",
  "n07200813": "refutation, defense, defence",
  "n07201105": "confutation",
  "n07201220": "rebuttal",
  "n07201365": "description",
  "n07201562": "characterization, characterisation",
  "n07201804": "word picture, word-painting, delineation, depiction, picture, characterization, characterisation",
  "n07202311": "epithet",
  "n07202391": "portrayal, portraiture, portrait",
  "n07202579": "label",
  "n07202812": "particularization, particularisation, detailing",
  "n07203016": "sketch, vignette",
  "n07203126": "affirmation, assertion, statement",
  "n07203388": "representation",
  "n07203613": "say-so",
  "n07203696": "affirmative",
  "n07203900": "yes",
  "n07204008": "yea",
  "n07204110": "declaration",
  "n07204240": "denial, disaffirmation",
  "n07204401": "denial",
  "n07204665": "abnegation",
  "n07204822": "naysaying",
  "n07204911": "negative",
  "n07205104": "no",
  "n07205210": "nay",
  "n07205308": "double negative",
  "n07205439": "double negative",
  "n07205573": "refusal",
  "n07205718": "repudiation",
  "n07205946": "disavowal, disclaimer",
  "n07206096": "retraction, abjuration, recantation",
  "n07206302": "withdrawal, backdown, climb-down",
  "n07206461": "negation",
  "n07206596": "contradiction",
  "n07206800": "self-contradiction",
  "n07206887": "contradiction, contradiction in terms",
  "n07207142": "cancellation",
  "n07207273": "rejection",
  "n07207410": "repudiation, renunciation",
  "n07207680": "disclaimer",
  "n07207860": "disownment, disowning",
  "n07208000": "rebuff, snub, repulse",
  "n07208184": "short shrift, summary treatment",
  "n07208338": "objection",
  "n07208565": "challenge",
  "n07208708": "complaint",
  "n07208930": "complaint",
  "n07209089": "demur, demurral, demurrer",
  "n07209305": "dissent",
  "n07209533": "exception",
  "n07209691": "caption",
  "n07209868": "exclamation",
  "n07209965": "gripe, kick, beef, bitch, squawk",
  "n07210225": "protest, protestation",
  "n07210553": "protest",
  "n07210801": "grievance",
  "n07210951": "growling",
  "n07211092": "grumble, grumbling, murmur, murmuring, mutter, muttering",
  "n07211374": "jeremiad",
  "n07211503": "kvetch",
  "n07211604": "pet peeve",
  "n07211752": "whimper, whine",
  "n07211950": "lament, lamentation, plaint, wail",
  "n07212190": "informing, making known",
  "n07212424": "telling, apprisal, notification",
  "n07212612": "notice",
  "n07212874": "warning",
  "n07213079": "dismissal, dismission, pink slip",
  "n07213232": "revelation, divine revelation",
  "n07213395": "disclosure, revelation, revealing",
  "n07213717": "display",
  "n07213889": "histrionics",
  "n07213989": "production",
  "n07214150": "sackcloth and ashes",
  "n07214267": "divulgence, divulgement",
  "n07214432": "discovery",
  "n07214642": "discovery",
  "n07214894": "giveaway",
  "n07214994": "informing, ratting",
  "n07215185": "leak, news leak",
  "n07215377": "exposure",
  "n07215568": "expose, unmasking",
  "n07215816": "muckraking",
  "n07215948": "admission",
  "n07216083": "confession",
  "n07216228": "self-accusation, self-condemnation",
  "n07216412": "concession",
  "n07216615": "sop",
  "n07216761": "stipulation, judicial admission",
  "n07217117": "takeaway",
  "n07217260": "wage concession",
  "n07217349": "presentation, introduction, intro",
  "n07217579": "debut",
  "n07217673": "reintroduction",
  "n07217782": "briefing",
  "n07217924": "report, account",
  "n07218299": "megillah",
  "n07218470": "report, study, written report",
  "n07218853": "skinny",
  "n07219004": "stuff",
  "n07219162": "assay",
  "n07219297": "case study",
  "n07219530": "white book, white paper",
  "n07219636": "blue book",
  "n07219751": "green paper",
  "n07219923": "progress report",
  "n07220045": "position paper",
  "n07220178": "medical report",
  "n07220300": "report card, report",
  "n07220466": "debriefing",
  "n07220586": "anecdote",
  "n07220773": "narration, recital, yarn",
  "n07221094": "narrative, narration, story, tale",
  "n07221623": "Canterbury Tales",
  "n07221756": "recital",
  "n07221939": "tall tale",
  "n07222050": "folktale, folk tale",
  "n07222230": "Arabian Nights' Entertainment, Arabian Nights, Thousand and One Nights",
  "n07222433": "sob story, sob stuff",
  "n07222581": "fairytale, fairy tale, fairy story",
  "n07222728": "nursery rhyme",
  "n07222823": "relation, telling, recounting",
  "n07223094": "earful",
  "n07223170": "gossip, comment, scuttlebutt",
  "n07223450": "rumor, rumour, hearsay",
  "n07223635": "grapevine, pipeline, word of mouth",
  "n07223811": "scandal, dirt, malicious gossip",
  "n07223985": "talk, talk of the town",
  "n07224151": "warning",
  "n07224481": "wake-up call",
  "n07224684": "alarmism",
  "n07224774": "alert, alerting",
  "n07224923": "Emergency Alert System, EAS",
  "n07225167": "caution, caveat",
  "n07225333": "false alarm",
  "n07225450": "forewarning, premonition",
  "n07225577": "heads-up",
  "n07225696": "strategic warning",
  "n07225857": "tactical warning",
  "n07226015": "threat",
  "n07226151": "warning of attack",
  "n07226330": "warning of war",
  "n07226545": "promise",
  "n07226841": "oath",
  "n07227055": "bayat",
  "n07227158": "Hippocratic oath",
  "n07227301": "parole, word, word of honor",
  "n07227406": "assurance",
  "n07227589": "clean bill of health",
  "n07227772": "assurance, pledge",
  "n07228105": "plight, troth",
  "n07228211": "betrothal, troth, engagement",
  "n07228349": "pinning",
  "n07228531": "ringing",
  "n07228630": "rain check",
  "n07228751": "vow",
  "n07228971": "thanks",
  "n07229111": "appreciation",
  "n07229245": "thank you",
  "n07229341": "bow, curtain call",
  "n07229530": "boast, boasting, self-praise, jactitation",
  "n07229747": "brag, bragging, crow, crowing, vaporing, line-shooting, gasconade",
  "n07230089": "braggadocio, bluster, rodomontade, rhodomontade",
  "n07230227": "vaunt",
  "n07230320": "self-assertion",
  "n07230502": "naming",
  "n07230796": "acrophony",
  "n07230959": "numeration",
  "n07231048": "indication, denotation",
  "n07231199": "specification",
  "n07231294": "challenge",
  "n07231532": "dare, daring",
  "n07231728": "confrontation",
  "n07231840": "call-out",
  "n07231943": "defiance",
  "n07232053": "calling into question, demand for explanation",
  "n07232189": "demand for identification",
  "n07232275": "gauntlet, gantlet",
  "n07232421": "explanation",
  "n07232655": "elucidation",
  "n07232811": "explication",
  "n07232988": "denunciation, denouncement",
  "n07233214": "excoriation",
  "n07233304": "fulmination, diatribe",
  "n07233415": "tirade, philippic, broadside",
  "n07233542": "damnation",
  "n07233634": "execration, condemnation, curse",
  "n07233863": "anathema",
  "n07233996": "imprecation, malediction",
  "n07234230": "accusation, accusal",
  "n07234657": "recrimination",
  "n07234735": "recital",
  "n07234881": "recitation, recital, reading",
  "n07235218": "recitation",
  "n07235335": "indictment, bill of indictment",
  "n07235574": "murder charge, murder indictment",
  "n07235696": "true bill",
  "n07235790": "impeachment",
  "n07235936": "arraignment",
  "n07236077": "allegation",
  "n07236307": "blame game",
  "n07236466": "grievance",
  "n07236627": "lodgment, lodgement",
  "n07236759": "plaint",
  "n07236957": "imprecation",
  "n07237038": "imputation",
  "n07237234": "finger-pointing, fingerpointing",
  "n07237409": "indictment",
  "n07237561": "information",
  "n07237647": "preferment",
  "n07237758": "incrimination, inculpation, blame",
  "n07238102": "implication",
  "n07238308": "unspoken accusation, veiled accusation",
  "n07238455": "insinuation, innuendo",
  "n07238589": "self-incrimination",
  "n07238694": "address, speech",
  "n07239183": "allocution",
  "n07239327": "colloquium",
  "n07239428": "dithyramb",
  "n07239550": "Gettysburg Address",
  "n07239788": "impromptu",
  "n07239927": "impromptu",
  "n07240077": "inaugural address, inaugural",
  "n07240278": "keynote",
  "n07240417": "keynote speech, keynote address",
  "n07240549": "lecture, public lecture, talk",
  "n07240763": "litany",
  "n07240925": "nominating speech, nominating address, nomination",
  "n07241205": "oratory",
  "n07241551": "oration",
  "n07241723": "peroration",
  "n07241837": "public speaking, speechmaking, speaking, oral presentation",
  "n07242104": "debate, disputation, public debate",
  "n07242324": "declamation",
  "n07242470": "declamation",
  "n07242657": "epideictic oratory",
  "n07242912": "harangue, rant, ranting",
  "n07243114": "screed",
  "n07243193": "raving",
  "n07243306": "stump speech",
  "n07243382": "salutatory address, salutatory oration, salutatory",
  "n07243565": "valediction, valedictory address, valedictory oration, valedictory",
  "n07243837": "sermon, discourse, preaching",
  "n07244154": "baccalaureate",
  "n07244305": "kerygma, kerugma",
  "n07244433": "Sermon on the Mount",
  "n07244613": "evangelism",
  "n07244822": "televangelism",
  "n07244949": "homily, preachment",
  "n07245125": "persuasion, suasion",
  "n07245472": "arm-twisting",
  "n07245686": "dissuasion",
  "n07245885": "electioneering, bell ringing, canvassing",
  "n07246036": "exhortation, incitement",
  "n07246215": "pep talk",
  "n07246382": "proselytism",
  "n07246469": "sloganeering",
  "n07246582": "suggestion, prompting",
  "n07246742": "expostulation, remonstrance, remonstration, objection",
  "n07246932": "weapon, artillery",
  "n07247071": "promotion, publicity, promotional material, packaging",
  "n07247451": "buildup",
  "n07247602": "sensationalism",
  "n07247803": "shocker",
  "n07247925": "public relations, PR",
  "n07248060": "endorsement, indorsement, blurb",
  "n07248320": "book jacket, dust cover, dust jacket, dust wrapper",
  "n07248507": "ballyhoo, hoopla, hype, plug",
  "n07248653": "sales talk, sales pitch, pitch",
  "n07248801": "ad, advertisement, advertizement, advertising, advertizing, advert",
  "n07249180": "advertorial",
  "n07249336": "mailer",
  "n07249426": "newspaper ad, newspaper advertisement",
  "n07249585": "classified ad, classified advertisement, classified",
  "n07249817": "sales promotion",
  "n07249932": "want ad",
  "n07250034": "commercial, commercial message",
  "n07250198": "infomercial, informercial",
  "n07250339": "circular, handbill, bill, broadside, broadsheet, flier, flyer, throwaway",
  "n07250727": "stuffer",
  "n07250868": "teaser",
  "n07251003": "top billing",
  "n07251148": "white pages",
  "n07251373": "yellow pages",
  "n07251619": "abetment, abettal, instigation",
  "n07251779": "cheering, shouting",
  "n07251984": "promotion, furtherance, advancement",
  "n07252206": "fostering, fosterage",
  "n07252378": "goad, goading, prod, prodding, urging, spur, spurring",
  "n07252764": "provocation, incitement",
  "n07252967": "subornation",
  "n07253142": "subornation of perjury",
  "n07253482": "vote of confidence",
  "n07253637": "discouragement",
  "n07253814": "disheartenment",
  "n07253948": "dissuasion",
  "n07254057": "determent, deterrence, intimidation",
  "n07254267": "resignation",
  "n07254456": "abdication, stepping down",
  "n07254594": "renunciation, renouncement",
  "n07254836": "relinquishment, relinquishing",
  "n07255027": "giving up, yielding, surrender",
  "n07255174": "prohibition",
  "n07255299": "interdiction",
  "n07255401": "ban, banning, forbiddance, forbidding",
  "n07255628": "test ban",
  "n07255791": "psychic communication, psychical communication, anomalous communication",
  "n07255998": "telepathy, thought transference",
  "n07256230": "telegnosis",
  "n07256375": "psychic phenomena, psychic phenomenon, parapsychology",
  "n07256695": "clairvoyance, second sight, extrasensory perception, E.S.P., ESP",
  "n07256932": "precognition, foreknowledge",
  "n07257045": "telekinesis, psychokinesis",
  "n07257227": "table rapping, table tapping, spirit rapping",
  "n07257393": "table tipping, table tilting, table turning, table lifting",
  "n07257582": "windsock, wind sock, sock, air sock, air-sleeve, wind sleeve, wind cone, drogue",
  "n07257815": "post, stake",
  "n07258109": "starting post",
  "n07258238": "winning post",
  "n07258332": "reference point, point of reference, reference",
  "n07258664": "reference, source",
  "n07258964": "republication",
  "n07259145": "benchmark, bench mark",
  "n07259319": "landmark",
  "n07259438": "merestone, meerestone, mearstone",
  "n07259610": "lubber's line, lubber line, lubber's mark, lubber's point",
  "n07259772": "rule, linguistic rule",
  "n07259984": "universal, linguistic universal",
  "n07260175": "grammatical rule, rule of grammar",
  "n07260332": "transformation",
  "n07260493": "morphological rule, rule of morphology",
  "n07260623": "standard, criterion, measure, touchstone",
  "n07261143": "benchmark",
  "n07261300": "earned run average, ERA",
  "n07261542": "grade point average, GPA",
  "n07261782": "procrustean standard, procrustean rule, procrustean bed",
  "n07261955": "yardstick",
  "n07262108": "target, mark",
  "n07262278": "clout",
  "n07262354": "drogue",
  "n07262462": "white line",
  "n07262579": "indicator",
  "n07262704": "blinker, turn signal, turn indicator, trafficator",
  "n07262942": "armband",
  "n07263053": "rocket, skyrocket",
  "n07263220": "electronic signal",
  "n07263358": "blip, pip, radar target",
  "n07263503": "radar echo",
  "n07263745": "clutter",
  "n07263897": "radar beacon, racon",
  "n07264086": "radio beacon",
  "n07264213": "beacon, beacon fire",
  "n07264384": "star shell",
  "n07264484": "Bengal light",
  "n07264615": "Very light, Very-light",
  "n07264725": "signal fire, signal light",
  "n07264820": "input signal, input",
  "n07264925": "output signal, output",
  "n07265075": "printout",
  "n07265170": "readout, read-out",
  "n07265276": "fire alarm",
  "n07265381": "foghorn, fogsignal",
  "n07265508": "horn",
  "n07265619": "red flag",
  "n07265789": "siren",
  "n07265886": "tocsin, alarm bell",
  "n07265990": "stoplight, brake light",
  "n07266178": "buoy",
  "n07266442": "acoustic buoy",
  "n07266573": "bell buoy, gong buoy",
  "n07266665": "whistle buoy, whistling buoy",
  "n07266776": "can, can buoy",
  "n07266879": "conical buoy, nun, nun buoy",
  "n07266979": "spar buoy",
  "n07267067": "barber's pole",
  "n07267160": "staff",
  "n07267309": "crosier, crozier",
  "n07267461": "mace",
  "n07267573": "scepter, sceptre, verge, wand",
  "n07267702": "bauble",
  "n07267795": "tipstaff",
  "n07267931": "cordon",
  "n07268035": "wings",
  "n07268166": "black belt",
  "n07268302": "blue ribbon, cordon bleu",
  "n07268413": "button",
  "n07268603": "Emmy",
  "n07268759": "Nobel prize",
  "n07268967": "Academy Award, Oscar",
  "n07269163": "Prix de Rome",
  "n07269430": "Prix Goncourt",
  "n07269552": "chevron, stripe, stripes, grade insignia",
  "n07269758": "stripe",
  "n07269916": "icon",
  "n07270179": "marker, marking, mark",
  "n07270601": "identifier",
  "n07270718": "postmark",
  "n07270893": "watermark",
  "n07271052": "broad arrow",
  "n07271167": "milestone, milepost",
  "n07271280": "variable",
  "n07271478": "placeholder",
  "n07271648": "unknown, unknown quantity",
  "n07271791": "peg, pin",
  "n07271942": "spot, pip",
  "n07272084": "logo, logotype",
  "n07272172": "label",
  "n07272416": "bookplate, ex libris",
  "n07272545": "gummed label, sticker, paster",
  "n07272694": "dog tag",
  "n07272807": "dog tag",
  "n07272920": "name tag",
  "n07273024": "price tag",
  "n07273136": "tag, ticket",
  "n07273416": "tag",
  "n07273616": "title bar",
  "n07273802": "cairn",
  "n07273924": "shrug",
  "n07274027": "wave, waving, wafture",
  "n07274197": "V sign",
  "n07274316": "nod",
  "n07274425": "bow, bowing, obeisance",
  "n07274730": "sign of the cross",
  "n07274890": "curtsy, curtsey",
  "n07275078": "genuflection, genuflexion",
  "n07275275": "kowtow, kotow",
  "n07275489": "scrape, scraping",
  "n07275713": "salaam",
  "n07275823": "ground rule",
  "n07276018": "sign",
  "n07276242": "system command",
  "n07276414": "walking papers, marching orders",
  "n07276538": "wanted notice, wanted poster",
  "n07276732": "International Wanted Notice, Red Notice",
  "n07277158": "plagiarism",
  "n07277384": "transcript",
  "n07277697": "voice",
  "n07277915": "Bach",
  "n07278014": "Beethoven",
  "n07278146": "Brahms",
  "n07278256": "Chopin",
  "n07278365": "Gilbert and Sullivan",
  "n07278510": "Handel",
  "n07278582": "Haydn",
  "n07278652": "Mozart",
  "n07278775": "Stravinsky",
  "n07278925": "Wagner",
  "n07279045": "language system",
  "n07279285": "contact, touch",
  "n07279453": "traffic",
  "n07279687": "order",
  "n07279968": "short order",
  "n07280072": "recall, callback",
  "n07280295": "uplink",
  "n07280424": "capriccio",
  "n07280599": "interrogation",
  "n07280754": "motet",
  "n07280941": "negation",
  "n07281099": "packet",
  "n07281219": "program music, programme music",
  "n07281375": "incidental music",
  "n07281523": "slanguage",
  "n07281635": "Ta'ziyeh",
  "n07281871": "sprechgesang, sprechstimme",
  "n07282006": "vocal music, vocal",
  "n07282166": "voice over",
  "n07282289": "walk-through",
  "n07282497": "yearbook",
  "n07282695": "zinger",
  "n07282929": "Das Kapital, Capital",
  "n07283065": "Erewhon",
  "n07283198": "Utopia",
  "n07283364": "might-have-been",
  "n07283473": "nonevent",
  "n07283608": "happening, occurrence, occurrent, natural event",
  "n07284554": "accompaniment, concomitant, attendant, co-occurrence",
  "n07284848": "associate",
  "n07285036": "avalanche",
  "n07285191": "background",
  "n07285403": "experience",
  "n07285872": "appalling",
  "n07286014": "augury, sign, foretoken, preindication",
  "n07286278": "war cloud",
  "n07286368": "omen, portent, presage, prognostic, prognostication, prodigy",
  "n07286799": "auspice",
  "n07286905": "foreboding",
  "n07286999": "death knell",
  "n07287088": "flash, flashing",
  "n07287288": "good time, blast",
  "n07287472": "loss",
  "n07287610": "near-death experience",
  "n07287730": "ordeal",
  "n07287812": "out-of-body experience",
  "n07288024": "taste",
  "n07288215": "time",
  "n07288385": "trip, head trip",
  "n07288485": "vision",
  "n07288639": "social event",
  "n07288801": "miracle",
  "n07289014": "trouble",
  "n07289358": "treat",
  "n07289481": "miracle",
  "n07289588": "wonder, marvel",
  "n07289831": "thing",
  "n07289956": "episode",
  "n07290144": "feast",
  "n07290278": "drama, dramatic event",
  "n07290503": "night terror",
  "n07290761": "eventuality, contingency, contingence",
  "n07290905": "beginning",
  "n07291215": "casus belli",
  "n07291312": "ending, conclusion, finish",
  "n07291794": "end, last, final stage",
  "n07292118": "endgame, end game",
  "n07292273": "endgame, end game",
  "n07292418": "homestretch",
  "n07292577": "passing",
  "n07292694": "result, resultant, final result, outcome, termination",
  "n07293080": "denouement",
  "n07293180": "deal",
  "n07293391": "fair deal, square deal",
  "n07293475": "raw deal",
  "n07293546": "decision",
  "n07293678": "decision",
  "n07293903": "split decision",
  "n07294019": "consequence, aftermath",
  "n07294260": "corollary",
  "n07294423": "deserts, comeuppance, comeupance",
  "n07294550": "fruit",
  "n07294699": "sequella",
  "n07294777": "train",
  "n07294907": "poetic justice, just deserts",
  "n07295047": "offspring, materialization, materialisation",
  "n07295391": "separation",
  "n07295507": "sequel, subsequence",
  "n07295629": "wages, reward, payoff",
  "n07295850": "foregone conclusion, matter of course",
  "n07295955": "worst",
  "n07296083": "one-off",
  "n07296190": "periodic event, recurrent event",
  "n07296428": "change, alteration, modification",
  "n07297376": "avulsion",
  "n07297633": "break",
  "n07297811": "mutation",
  "n07297927": "sublimation",
  "n07298154": "surprise",
  "n07298396": "bombshell, thunderbolt, thunderclap",
  "n07298543": "coup de theatre",
  "n07298624": "eye opener",
  "n07298715": "peripeteia, peripetia, peripety",
  "n07298982": "shock, blow",
  "n07299203": "blip",
  "n07299439": "stunner",
  "n07299569": "error, computer error",
  "n07299790": "hardware error",
  "n07299965": "disk error",
  "n07300092": "software error, programming error",
  "n07300316": "semantic error, run-time error, runtime error",
  "n07300494": "syntax error",
  "n07300781": "algorithm error",
  "n07300960": "accident, stroke, fortuity, chance event",
  "n07301336": "accident",
  "n07301543": "collision",
  "n07301846": "near miss",
  "n07301950": "crash, wreck",
  "n07302164": "prang",
  "n07302267": "derailment",
  "n07302407": "ground loop",
  "n07302542": "collision, hit",
  "n07302836": "fire",
  "n07303153": "backfire",
  "n07303335": "bonfire, balefire",
  "n07303466": "brush fire",
  "n07303585": "campfire",
  "n07303697": "conflagration, inferno",
  "n07303839": "forest fire",
  "n07303988": "grassfire, prairie fire",
  "n07304096": "smoulder, smolder",
  "n07304266": "smudge",
  "n07304353": "crown fire",
  "n07304507": "ground fire",
  "n07304630": "surface fire",
  "n07304753": "wildfire",
  "n07304852": "misfortune, bad luck",
  "n07305098": "pity, shame",
  "n07305234": "affliction",
  "n07305438": "convulsion",
  "n07305551": "embarrassment",
  "n07305760": "disembarrassment",
  "n07305907": "hell, blaze",
  "n07306032": "calvary, martyrdom",
  "n07306160": "onslaught",
  "n07306252": "scandal, outrage",
  "n07306481": "skeleton, skeleton in the closet, skeleton in the cupboard",
  "n07306680": "Teapot Dome, Teapot Dome scandal",
  "n07306963": "Watergate, Watergate scandal",
  "n07307172": "chapter",
  "n07307297": "idyll",
  "n07307477": "incident",
  "n07307646": "cause celebre",
  "n07307754": "discharge",
  "n07307895": "electrical discharge",
  "n07308042": "nerve impulse, nervous impulse, neural impulse, impulse",
  "n07308308": "action potential",
  "n07308450": "spike",
  "n07308563": "explosion, detonation, blowup",
  "n07308889": "case, instance, example",
  "n07309223": "humiliation, mortification",
  "n07309457": "piece, bit",
  "n07309599": "time, clip",
  "n07309781": "movement, motion",
  "n07310338": "crustal movement, tectonic movement",
  "n07310507": "approach, approaching",
  "n07310642": "passing, passage",
  "n07310839": "deflection, deflexion",
  "n07310991": "bending, bend",
  "n07311115": "change of location, travel",
  "n07311540": "fender-bender",
  "n07311661": "ascension",
  "n07311822": "Ascension, Ascension of Christ",
  "n07312018": "Resurrection, Christ's Resurrection, Resurrection of Christ",
  "n07312221": "circulation",
  "n07312503": "creep",
  "n07312616": "migration",
  "n07312829": "migration",
  "n07313004": "shrinking, shrinkage",
  "n07313241": "compression, condensation, contraction",
  "n07313518": "constriction, coarctation",
  "n07313636": "injury, accidental injury",
  "n07313814": "rupture, breach, break, severance, rift, falling out",
  "n07314078": "schism",
  "n07314277": "hap",
  "n07314427": "mishap, misadventure, mischance",
  "n07314658": "puncture",
  "n07314838": "calamity, catastrophe, disaster, tragedy, cataclysm",
  "n07315350": "act of God, force majeure, vis major, inevitable accident, unavoidable casualty",
  "n07315631": "apocalypse",
  "n07315790": "famine",
  "n07315965": "the Irish Famine, the Great Hunger, the Great Starvation, the Great Calamity",
  "n07316242": "kiss of death",
  "n07316403": "meltdown",
  "n07316603": "plague",
  "n07316724": "visitation",
  "n07316856": "break, good luck, happy chance",
  "n07316999": "coincidence, happenstance",
  "n07317144": "lottery",
  "n07317285": "pileup",
  "n07317369": "smash, smash-up",
  "n07317519": "slip, trip",
  "n07317764": "failure",
  "n07318133": "downfall, ruin, ruination",
  "n07318299": "flame-out",
  "n07318476": "malfunction",
  "n07318618": "blowout",
  "n07318824": "stall",
  "n07319103": "success",
  "n07319399": "barnburner",
  "n07319549": "Godspeed",
  "n07319652": "miscarriage, abortion",
  "n07319774": "miss, misfire",
  "n07319909": "emergence, egress, issue",
  "n07320176": "eruption",
  "n07320302": "birth, nativity, nascency, nascence",
  "n07320622": "delivery",
  "n07320734": "live birth",
  "n07320894": "blessed event, happy event",
  "n07321012": "posthumous birth",
  "n07321142": "posthumous birth",
  "n07321247": "reincarnation, rebirth, renascence",
  "n07321387": "transmigration",
  "n07321517": "cycle of rebirth",
  "n07321657": "moksa",
  "n07321772": "appearance",
  "n07321967": "reappearance",
  "n07322138": "egress, emersion",
  "n07322341": "ingress, immersion",
  "n07322550": "Second Coming, Second Coming of Christ, Second Advent, Advent, Parousia",
  "n07322769": "makeup, make-up",
  "n07323024": "materialization, materialisation, manifestation",
  "n07323231": "manifestation",
  "n07323436": "apparition",
  "n07323605": "epiphany",
  "n07323682": "theophany",
  "n07323821": "Word of God",
  "n07323922": "origin, origination, inception",
  "n07324235": "germination",
  "n07324380": "genesis, generation",
  "n07324502": "ground floor",
  "n07324673": "emergence, outgrowth, growth",
  "n07324917": "rise",
  "n07325102": "crime wave",
  "n07325190": "start",
  "n07325639": "adrenarche",
  "n07325762": "menarche",
  "n07325864": "thelarche",
  "n07325990": "onset, oncoming",
  "n07326108": "dawn, morning",
  "n07326262": "flying start, running start",
  "n07326369": "opener",
  "n07326557": "cause",
  "n07326880": "antecedent",
  "n07327013": "preliminary, overture, prelude",
  "n07327288": "emanation, rise, procession",
  "n07327608": "etiology, aetiology",
  "n07327805": "factor",
  "n07328058": "fundamental",
  "n07328305": "parameter",
  "n07328436": "unknown quantity",
  "n07328646": "wild card",
  "n07328756": "producer",
  "n07328942": "creation, conception",
  "n07329169": "alpha",
  "n07329363": "opening, opening night, curtain raising",
  "n07329568": "kickoff, send-off, start-off",
  "n07329737": "racing start",
  "n07329833": "flying start, running start",
  "n07330007": "destiny, fate",
  "n07330250": "inevitable",
  "n07330362": "karma",
  "n07330560": "kismet, kismat",
  "n07330666": "predestination",
  "n07330828": "annihilation, disintegration",
  "n07331013": "eradication, obliteration",
  "n07331210": "debilitation, enervation, enfeeblement, exhaustion",
  "n07331400": "separation, breakup, detachment",
  "n07331600": "diffusion",
  "n07331759": "dispersion, scattering",
  "n07331932": "Diaspora",
  "n07332148": "dissipation",
  "n07332313": "invasion",
  "n07332550": "irradiation",
  "n07332691": "extinction, extermination",
  "n07332864": "Crucifixion",
  "n07332956": "fatality, human death",
  "n07333162": "finish",
  "n07333347": "martyrdom",
  "n07333506": "megadeath",
  "n07333649": "passing, loss, departure, exit, expiration, going, release",
  "n07333971": "wrongful death",
  "n07334206": "doom, doomsday, day of reckoning, end of the world",
  "n07334490": "destruction, demolition, wipeout",
  "n07334876": "ravage, depredation",
  "n07335097": "razing, wrecking",
  "n07335243": "ruin, ruination",
  "n07335414": "devastation, desolation",
  "n07335581": "wrack, rack",
  "n07335716": "disappearance",
  "n07335917": "evanescence",
  "n07336104": "vanishing",
  "n07336214": "receding, fadeout",
  "n07336346": "disappearance",
  "n07336612": "adversity",
  "n07336763": "hardship",
  "n07336999": "knock",
  "n07337096": "vagary",
  "n07337390": "variation, fluctuation",
  "n07337694": "vicissitude",
  "n07337935": "allomerism",
  "n07338114": "engagement, mesh, meshing, interlocking",
  "n07338358": "flick",
  "n07338552": "impact",
  "n07338681": "blow, bump",
  "n07338970": "slam",
  "n07339098": "jolt, jar, jounce, shock",
  "n07339329": "contact, impinging, striking",
  "n07339653": "damage, equipment casualty",
  "n07339808": "battle damage, combat casualty",
  "n07339941": "operational damage, operational casualty",
  "n07340094": "casualty",
  "n07340249": "wound, injury, combat injury",
  "n07340453": "blighty wound",
  "n07340592": "flesh wound",
  "n07340725": "personnel casualty, loss",
  "n07340895": "sacrifice",
  "n07341038": "cycle, oscillation",
  "n07341304": "cardiac cycle",
  "n07341665": "Carnot cycle, Carnot's ideal cycle",
  "n07341860": "pass",
  "n07342049": "repeat, repetition",
  "n07342383": "sequence",
  "n07342495": "cycle",
  "n07342772": "merry-go-round",
  "n07343017": "samsara",
  "n07343195": "replay, rematch",
  "n07343363": "recurrence, return",
  "n07343574": "atavism, reversion, throwback",
  "n07343713": "flashback",
  "n07343910": "sunrise",
  "n07344015": "sunset",
  "n07344120": "ground swell, heavy swell",
  "n07344233": "surf, breaker, breakers",
  "n07344368": "wake, backwash",
  "n07344528": "swash",
  "n07344663": "ripple, rippling, riffle, wavelet",
  "n07344875": "gravity wave, gravitation wave",
  "n07345066": "sine wave",
  "n07345166": "oscillation, vibration",
  "n07345441": "ripple",
  "n07345593": "wave, undulation",
  "n07345960": "jitter",
  "n07346057": "fluctuation",
  "n07346195": "seiche",
  "n07346344": "soliton, soliton wave, solitary wave",
  "n07346772": "standing wave, stationary wave",
  "n07347051": "traveling wave, travelling wave",
  "n07347224": "sound wave, acoustic wave",
  "n07347373": "air wave",
  "n07347468": "transient",
  "n07347664": "wave form, waveform, wave shape",
  "n07347846": "shock wave, blast wave",
  "n07348041": "sonic boom",
  "n07348258": "swell, crestless wave",
  "n07348399": "lift, rise",
  "n07348545": "billow, surge",
  "n07348694": "tidal wave",
  "n07348870": "tidal wave",
  "n07349069": "tidal wave",
  "n07349299": "tsunami",
  "n07349532": "roller, roll, rolling wave",
  "n07349679": "periodic motion, periodic movement",
  "n07349880": "harmonic motion",
  "n07350069": "heave",
  "n07350192": "recoil, repercussion, rebound, backlash",
  "n07350401": "bounce, bouncing",
  "n07350567": "resilience, resiliency",
  "n07350754": "recoil, kick",
  "n07350893": "seek",
  "n07351031": "squeeze, wring",
  "n07351195": "throw, stroke, cam stroke",
  "n07351380": "instroke",
  "n07351497": "outstroke",
  "n07351612": "turning, turn",
  "n07351909": "twist, wrench",
  "n07352048": "undulation",
  "n07352190": "wave, moving ridge",
  "n07352617": "comber",
  "n07352693": "whitecap, white horse",
  "n07352835": "wave",
  "n07353075": "shipwreck, wreck",
  "n07353232": "capsizing",
  "n07353376": "finish",
  "n07353716": "draw, standoff, tie",
  "n07354009": "dead heat",
  "n07354080": "stalemate",
  "n07354243": "photo finish",
  "n07354504": "second-place finish, runner-up finish",
  "n07354628": "third-place finish",
  "n07354731": "win",
  "n07354911": "first-place finish",
  "n07355014": "omega, Z",
  "n07355194": "conversion",
  "n07355347": "Christianization, Christianisation",
  "n07355491": "death, decease, expiry",
  "n07355887": "decrease, lessening, drop-off",
  "n07356266": "sinking",
  "n07356489": "destabilization",
  "n07356676": "increase",
  "n07356970": "attrition",
  "n07357101": "easing, moderation, relief",
  "n07357253": "breath of fresh air",
  "n07357388": "improvement, betterment, advance",
  "n07357679": "refinement, elaboration",
  "n07357866": "Assumption",
  "n07358060": "deformation",
  "n07358377": "Transfiguration, Transfiguration of Jesus",
  "n07358576": "transition",
  "n07358768": "ground swell",
  "n07358985": "leap, jump, saltation",
  "n07359184": "quantum leap, quantum jump",
  "n07359377": "quantum jump",
  "n07359599": "transformation, transmutation, shift",
  "n07359920": "population shift",
  "n07360077": "pyrolysis",
  "n07360211": "sea change",
  "n07360293": "sublimation",
  "n07360477": "tin pest, tin disease, tin plague",
  "n07360647": "infection, contagion, transmission",
  "n07360841": "scene",
  "n07360957": "sideshow",
  "n07361128": "collapse",
  "n07361416": "cave in, subsidence",
  "n07361576": "killing, violent death",
  "n07361717": "fatal accident, casualty",
  "n07361863": "collateral damage",
  "n07362075": "cessation, surcease",
  "n07362218": "settling, subsiding, subsidence",
  "n07362386": "drop, fall",
  "n07362699": "free fall",
  "n07362830": "gravitation",
  "n07363067": "levitation",
  "n07363180": "lightening",
  "n07363346": "descent",
  "n07363545": "set",
  "n07363668": "shower, cascade",
  "n07363883": "sinking",
  "n07364115": "submergence, submerging, submersion, immersion",
  "n07364434": "dip",
  "n07364573": "foundering, going under",
  "n07364700": "wobble",
  "n07364851": "shimmy",
  "n07365024": "flop, bust, fizzle",
  "n07365193": "turkey, bomb, dud",
  "n07365432": "debacle, fiasco",
  "n07365524": "implosion",
  "n07365673": "gravitational collapse",
  "n07365849": "stop, halt",
  "n07366145": "stand, standstill, tie-up",
  "n07366289": "deviation, divergence, departure, difference",
  "n07366627": "discrepancy, variance, variant",
  "n07366799": "driftage",
  "n07366971": "inflection, flection, flexion",
  "n07367091": "malformation, miscreation",
  "n07367231": "monstrosity",
  "n07367385": "dislocation, disruption",
  "n07367548": "break",
  "n07367708": "snap",
  "n07367812": "interruption, break",
  "n07368130": "punctuation",
  "n07368256": "suspension, respite, reprieve, hiatus, abatement",
  "n07368482": "defervescence",
  "n07368646": "eclipse, occultation",
  "n07368877": "solar eclipse",
  "n07368993": "lunar eclipse",
  "n07369099": "annular eclipse",
  "n07369206": "total eclipse",
  "n07369344": "partial eclipse",
  "n07369470": "augmentation",
  "n07369604": "adjustment, accommodation, fitting",
  "n07369922": "shakedown",
  "n07370125": "entrance, entering",
  "n07370270": "fall",
  "n07370410": "climb, climbing, mounting",
  "n07370671": "elevation, lift, raising",
  "n07370968": "heave, heaving",
  "n07371168": "liftoff",
  "n07371293": "sound",
  "n07372565": "fuss, trouble, bother, hassle",
  "n07372779": "headway, head",
  "n07372959": "trial, tribulation, visitation",
  "n07373277": "union",
  "n07373481": "amphimixis",
  "n07373602": "fusion, merger, unification",
  "n07373803": "combining, combine",
  "n07374152": "recombination",
  "n07374357": "recombination",
  "n07374633": "consolidation",
  "n07374756": "mix, mixture",
  "n07375053": "concoction",
  "n07375214": "conglomeration, conglobation",
  "n07375405": "blend",
  "n07375525": "rapid climb, rapid growth, zoom",
  "n07375635": "takeoff",
  "n07375781": "upheaval, uplift, upthrow, upthrust",
  "n07376008": "uplifting",
  "n07376168": "baa",
  "n07376257": "bang, clap, eruption, blast, bam",
  "n07376454": "bong",
  "n07376539": "banging",
  "n07376621": "bark",
  "n07376731": "bark",
  "n07376836": "bay",
  "n07376937": "beat",
  "n07377082": "beep, bleep",
  "n07377244": "bell, toll",
  "n07377473": "blare, blaring, cacophony, clamor, din",
  "n07377682": "boom, roar, roaring, thunder",
  "n07377931": "bleat",
  "n07378059": "bray",
  "n07378163": "bow-wow",
  "n07378234": "buzz, bombilation, bombination",
  "n07378437": "cackle",
  "n07378569": "caterwaul",
  "n07378682": "caw",
  "n07378781": "chatter, chattering",
  "n07378952": "chatter, chattering",
  "n07379094": "cheep, peep",
  "n07379223": "chink, click, clink",
  "n07379409": "chirp",
  "n07379577": "chirrup, twitter",
  "n07379695": "chorus",
  "n07379852": "chug",
  "n07379963": "clack, clap",
  "n07380144": "clang, clangor, clangour, clangoring, clank, clash, crash",
  "n07380473": "clatter",
  "n07380686": "click-clack",
  "n07380784": "clickety-clack, clickety-click",
  "n07380934": "clip-clop, clippety-clop, clop, clopping, clunking, clumping",
  "n07381100": "cluck, clucking",
  "n07381231": "cock-a-doodle-doo",
  "n07381329": "coo",
  "n07381423": "crack, cracking, snap",
  "n07381678": "crackle, crackling, crepitation",
  "n07381864": "creak, creaking",
  "n07382044": "crepitation rale",
  "n07382286": "crow",
  "n07382414": "crunch",
  "n07382572": "cry",
  "n07383092": "decrepitation",
  "n07383257": "ding",
  "n07383323": "drip, dripping",
  "n07383475": "drum",
  "n07383616": "ding-dong",
  "n07383696": "explosion",
  "n07383823": "footfall, footstep, step",
  "n07384023": "gargle",
  "n07384127": "gobble",
  "n07384244": "grate",
  "n07384344": "grinding",
  "n07384473": "growl, growling",
  "n07384614": "grunt, oink",
  "n07384741": "gurgle",
  "n07384898": "hiss, hissing, hushing, fizzle, sibilation",
  "n07385249": "honk",
  "n07385367": "howl",
  "n07385548": "howl",
  "n07385660": "hubbub, uproar, brouhaha, katzenjammer",
  "n07385803": "hum, humming",
  "n07385998": "jingle, jangle",
  "n07386194": "knell",
  "n07386370": "knock, knocking",
  "n07386614": "meow, mew, miaou, miaow, miaul",
  "n07386821": "moo",
  "n07386920": "mutter, muttering, murmur, murmuring, murmuration, mussitation",
  "n07387316": "neigh, nicker, whicker, whinny",
  "n07387509": "noise",
  "n07388706": "pant",
  "n07388816": "paradiddle, roll, drum roll",
  "n07388987": "pat, rap, tap",
  "n07389170": "patter",
  "n07389330": "peal, pealing, roll, rolling",
  "n07389569": "ping",
  "n07389757": "pitter-patter",
  "n07389931": "plonk",
  "n07390049": "plop",
  "n07390205": "plump",
  "n07390307": "plunk",
  "n07390400": "pop, popping",
  "n07390645": "purr",
  "n07390762": "quack",
  "n07390857": "quaver",
  "n07390945": "racket",
  "n07391079": "rat-a-tat-tat, rat-a-tat, rat-tat",
  "n07391240": "rattle, rattling, rale",
  "n07391516": "report",
  "n07391682": "rhonchus",
  "n07391863": "ring, ringing, tintinnabulation",
  "n07392280": "roar",
  "n07392373": "rub-a-dub, rataplan, drumbeat",
  "n07392483": "rumble, rumbling, grumble, grumbling",
  "n07392783": "rustle, rustling, whisper, whispering",
  "n07392982": "scrape, scraping, scratch, scratching",
  "n07393161": "screech, screeching, shriek, shrieking, scream, screaming",
  "n07393500": "scrunch",
  "n07393589": "shrilling",
  "n07393756": "sigh",
  "n07393919": "sizzle",
  "n07393988": "skirl",
  "n07394115": "slam",
  "n07394236": "snap",
  "n07394478": "snore",
  "n07394588": "song",
  "n07394814": "spatter, spattering, splatter, splattering, sputter, splutter, sputtering",
  "n07395104": "splash, plash",
  "n07395303": "splat",
  "n07395446": "squawk",
  "n07395623": "squeak",
  "n07395777": "squeal",
  "n07395867": "squish",
  "n07395957": "stridulation",
  "n07396120": "strum",
  "n07396233": "susurration, susurrus",
  "n07396414": "swish",
  "n07396530": "swoosh, whoosh",
  "n07396658": "tapping",
  "n07396822": "throbbing",
  "n07396945": "thump, thumping, clump, clunk, thud",
  "n07397230": "thrum",
  "n07397355": "thunder",
  "n07397552": "thunderclap",
  "n07397641": "thunk",
  "n07397761": "tick, ticking",
  "n07397955": "ticktock, tocktact, tictac",
  "n07398097": "ting, tinkle",
  "n07398276": "toot",
  "n07398362": "tootle",
  "n07398560": "tramp",
  "n07398659": "trample, trampling",
  "n07398873": "twang",
  "n07399027": "tweet",
  "n07399137": "vibrato",
  "n07399336": "tremolo",
  "n07399452": "voice",
  "n07399620": "vroom",
  "n07399735": "water hammer",
  "n07399829": "whack",
  "n07399917": "whir, whirr, whirring, birr",
  "n07400156": "whistle, whistling",
  "n07400361": "whiz",
  "n07400552": "yip, yelp, yelping",
  "n07400701": "zing",
  "n07400831": "news event",
  "n07400906": "pulse, pulsation, heartbeat, beat",
  "n07401236": "diastole",
  "n07401409": "systole",
  "n07401604": "extrasystole",
  "n07401726": "throb, throbbing, pounding",
  "n07401960": "high tide, high water, highwater",
  "n07402147": "ebb, reflux",
  "n07402271": "low tide, low water",
  "n07402393": "ebbtide",
  "n07402519": "tide",
  "n07402873": "direct tide",
  "n07403030": "flood tide, flood, rising tide",
  "n07403300": "neap tide, neap",
  "n07403459": "springtide",
  "n07403601": "leeward tide, lee tide",
  "n07403779": "slack water, slack tide",
  "n07403920": "tidal bore, bore, eagre, aegir, eager",
  "n07404114": "tidal flow, tidal current",
  "n07404261": "undertow, sea puss, sea-puss, sea purse, sea-purse, sea-poose",
  "n07404443": "riptide, rip current",
  "n07404584": "rip, riptide, tide rip, crosscurrent, countercurrent",
  "n07404798": "undertide, undercurrent",
  "n07404944": "slide",
  "n07405137": "avalanche",
  "n07405292": "lahar",
  "n07405411": "landslide, landslip",
  "n07405579": "mudslide",
  "n07405652": "Plinian eruption",
  "n07405817": "rockslide",
  "n07405893": "flow, flowing",
  "n07406350": "backflow, backflowing",
  "n07406474": "regurgitation",
  "n07406601": "airflow, air flow, flow of air",
  "n07406765": "current, stream",
  "n07407137": "freshet, spate",
  "n07407272": "overflow, runoff, overspill",
  "n07407459": "dripping, drippage",
  "n07407593": "torrent, violent stream",
  "n07407777": "discharge, outpouring, run",
  "n07407970": "flux, fluxion",
  "n07408086": "airburst",
  "n07408171": "blast",
  "n07408288": "bomb blast",
  "n07408386": "nuclear explosion, atomic explosion",
  "n07408519": "backblast, back-blast",
  "n07408621": "backfire",
  "n07408796": "big bang",
  "n07408965": "blowback, backfire",
  "n07409121": "fragmentation",
  "n07409255": "inflation",
  "n07409475": "ricochet, carom",
  "n07409592": "touch, touching",
  "n07409930": "concussion",
  "n07410021": "rap, strike, tap",
  "n07410207": "knock, bash, bang, smash, belt",
  "n07410526": "pounding, buffeting",
  "n07410613": "sideswipe",
  "n07410745": "slap, smack",
  "n07410883": "deflection, deflexion, refraction",
  "n07411011": "simple harmonic motion",
  "n07411160": "reversal, turn around",
  "n07411350": "yaw, swerve",
  "n07411490": "concussion",
  "n07411645": "twinkle, scintillation, sparkling",
  "n07411851": "shimmer, play",
  "n07412092": "flash",
  "n07412310": "flicker, spark, glint",
  "n07412478": "gleam, gleaming, glimmer",
  "n07412668": "glitter, sparkle, coruscation",
  "n07412876": "heat flash",
  "n07412993": "lightning",
  "n07413237": "heat lightning",
  "n07413452": "sheet lighting",
  "n07413629": "streak",
  "n07413714": "brush, light touch",
  "n07413815": "stroke",
  "n07413899": "concentration",
  "n07414068": "explosion",
  "n07414222": "jump, leap",
  "n07414370": "runup, run-up",
  "n07414566": "waxing",
  "n07414740": "convergence",
  "n07414922": "meeting, encounter",
  "n07415167": "conjunction, alignment",
  "n07415396": "inferior conjunction",
  "n07415561": "superior conjunction",
  "n07415730": "conversion, transition, changeover",
  "n07415962": "glycogenesis",
  "n07416107": "isomerization, isomerisation",
  "n07416308": "rectification",
  "n07416441": "transmutation",
  "n07416714": "juncture, occasion",
  "n07417043": "climax, flood tide",
  "n07417298": "conjuncture",
  "n07417405": "emergency, exigency, pinch",
  "n07417644": "crisis",
  "n07417851": "landmark, turning point, watershed",
  "n07418146": "Fall of Man",
  "n07418497": "road to Damascus",
  "n07418716": "milestone",
  "n07418822": "pass, head, straits",
  "n07418963": "reality check",
  "n07419233": "compaction, compression, concretion, densification",
  "n07419408": "rarefaction",
  "n07419599": "conservation, preservation",
  "n07419792": "recovery",
  "n07419960": "remission, remittal, subsidence",
  "n07420217": "resolution",
  "n07420354": "curse, torment",
  "n07420435": "fire",
  "n07420538": "detriment, hurt",
  "n07420671": "expense",
  "n07420770": "damage, harm, impairment",
  "n07420991": "pulsation, pulsing, pulse, impulse",
  "n07421316": "breakdown, equipment failure",
  "n07421580": "brake failure",
  "n07421669": "engine failure",
  "n07421749": "misfire, dud",
  "n07421859": "outage",
  "n07422036": "power outage, power failure",
  "n07422244": "fault",
  "n07422506": "blackout",
  "n07422629": "flame-out",
  "n07422800": "dwindling, dwindling away",
  "n07423001": "waning",
  "n07423248": "fading away",
  "n07423365": "turn, turn of events, twist",
  "n07423560": "development",
  "n07423818": "phenomenon",
  "n07423899": "complication, ramification",
  "n07424109": "revolution",
  "n07424436": "Cultural Revolution, Great Proletarian Cultural Revolution",
  "n07424797": "green revolution",
  "n07425011": "mutation, genetic mutation, chromosomal mutation",
  "n07425427": "sex change",
  "n07425577": "deletion",
  "n07425726": "inversion",
  "n07425903": "transposition",
  "n07426108": "mutagenesis",
  "n07426241": "insertional mutagenesis",
  "n07426406": "point mutation, gene mutation",
  "n07426573": "reversion",
  "n07426732": "saltation",
  "n07426893": "degeneration, retrogression",
  "n07427060": "atrophy, withering",
  "n07427224": "strengthening",
  "n07427337": "weakening",
  "n07427534": "attenuation, fading",
  "n07427728": "fall, downfall",
  "n07427951": "anticlimax",
  "n07428111": "abiotrophy",
  "n07428257": "cataplasia",
  "n07428450": "perturbation, disturbance",
  "n07428797": "magnetic storm",
  "n07428954": "earthquake, quake, temblor, seism",
  "n07429276": "shock, seismic disturbance",
  "n07429484": "tremor, earth tremor, microseism",
  "n07429637": "aftershock",
  "n07429782": "foreshock",
  "n07429870": "seaquake, submarine earthquake",
  "n07429976": "invasion, encroachment, intrusion",
  "n07430211": "noise, interference, disturbance",
  "n07430480": "background, background signal",
  "n07430770": "background noise, ground noise",
  "n07430964": "surface noise",
  "n07431119": "background radiation",
  "n07431247": "crosstalk, XT",
  "n07431369": "fadeout",
  "n07431502": "jitter",
  "n07431683": "static, atmospherics, atmospheric static",
  "n07431852": "white noise",
  "n07432034": "radio noise",
  "n07432119": "seepage, ooze, oozing",
  "n07432337": "exudation, transudation",
  "n07432559": "drip, trickle, dribble",
  "n07432835": "intravenous drip",
  "n07432973": "eddy, twist",
  "n07433145": "whirlpool, vortex, maelstrom",
  "n07433336": "Charybdis",
  "n07433510": "dismemberment, taking apart",
  "n07433662": "mutilation",
  "n07433868": "emission",
  "n07433973": "distortion, deformation",
  "n07434102": "warp, warping",
  "n07434209": "plunge",
  "n07434320": "precipitation",
  "n07434473": "fertilization, fertilisation, fecundation, dressing",
  "n07434678": "top dressing",
  "n07434782": "dissilience",
  "n07434942": "outburst, burst, flare-up",
  "n07435149": "salvo",
  "n07435273": "outbreak, eruption, irruption",
  "n07435533": "epidemic",
  "n07435713": "pandemic",
  "n07435891": "recrudescence",
  "n07436100": "jet, squirt, spurt, spirt",
  "n07436352": "rush",
  "n07436475": "volcanic eruption, eruption",
  "n07436661": "escape, leak, leakage, outflow",
  "n07436986": "fertilization, fertilisation, fecundation, impregnation",
  "n07437372": "pollination, pollenation",
  "n07437575": "cross-fertilization, cross-fertilisation",
  "n07437870": "allogamy",
  "n07437990": "self-fertilization, self-fertilisation",
  "n07438190": "superfecundation",
  "n07438396": "superfetation",
  "n07438655": "autogamy",
  "n07438792": "cross-pollination",
  "n07438983": "self-pollination",
  "n07439167": "cleistogamy",
  "n07439284": "flap, flapping, flutter, fluttering",
  "n07439570": "flush, gush, outpouring",
  "n07439883": "radiation",
  "n07440045": "adaptive radiation",
  "n07440240": "rush, spate, surge, upsurge",
  "n07440411": "debris surge, debris storm",
  "n07440617": "onrush",
  "n07440852": "springtide",
  "n07440979": "rotation, revolution, gyration",
  "n07441373": "dextrorotation, clockwise rotation",
  "n07441494": "levorotation, counterclockwise rotation",
  "n07441619": "axial rotation, axial motion, roll",
  "n07441902": "orbital rotation, orbital motion",
  "n07442068": "whirl, commotion",
  "n07442288": "spin",
  "n07442439": "backspin",
  "n07442569": "English, side",
  "n07442744": "topspin",
  "n07442874": "wallow",
  "n07443010": "run, ladder, ravel",
  "n07443210": "relaxation, loosening, slackening",
  "n07443539": "thaw",
  "n07443761": "substitution, permutation, transposition, replacement, switch",
  "n07444100": "business cycle, trade cycle",
  "n07444278": "daily variation",
  "n07444392": "diurnal variation",
  "n07444495": "tide",
  "n07444668": "shift, displacement",
  "n07444882": "amplitude",
  "n07445010": "luxation",
  "n07445149": "subluxation",
  "n07445265": "progress, progression, advance",
  "n07445480": "rise, rising, ascent, ascension",
  "n07445896": "spread, spreading",
  "n07446270": "stampede",
  "n07446404": "translation",
  "n07446533": "spray",
  "n07446599": "spritz",
  "n07446744": "angelus bell, angelus",
  "n07446920": "bell ringing",
  "n07447022": "return, coming back",
  "n07447177": "volution",
  "n07447261": "affair, occasion, social occasion, function, social function",
  "n07447641": "party",
  "n07448038": "bash, do, brawl",
  "n07448121": "birthday party",
  "n07448232": "bunfight, bun-fight",
  "n07448394": "ceilidh",
  "n07448608": "cocktail party",
  "n07448717": "dance",
  "n07448885": "ball, formal",
  "n07449037": "cotillion, cotilion",
  "n07449157": "masked ball, masquerade ball, fancy-dress ball",
  "n07449316": "promenade, prom",
  "n07449452": "barn dance",
  "n07449548": "hop, record hop",
  "n07449676": "rave",
  "n07449862": "fete, feast, fiesta",
  "n07450055": "luau",
  "n07450229": "house party",
  "n07450343": "jolly",
  "n07450463": "tea party",
  "n07450549": "whist drive",
  "n07450651": "celebration, jubilation",
  "n07450842": "ceremony, ceremonial, ceremonial occasion, observance",
  "n07451338": "circumstance",
  "n07451463": "funeral",
  "n07451687": "burial, entombment, inhumation, interment, sepulture",
  "n07451903": "sky burial",
  "n07452074": "wedding, wedding ceremony, nuptials, hymeneals",
  "n07452251": "pageant, pageantry",
  "n07452348": "dedication",
  "n07452559": "rededication",
  "n07452699": "opening",
  "n07452841": "commemoration, memorialization, memorialisation",
  "n07453063": "military ceremony",
  "n07453195": "initiation, induction, installation",
  "n07453638": "coronation, enthronement, enthronization, enthronisation, investiture",
  "n07453924": "bar mitzvah",
  "n07454196": "bat mitzvah, bath mitzvah, bas mitzvah",
  "n07454452": "exercise",
  "n07454632": "fire walking",
  "n07454758": "commencement, commencement exercise, commencement ceremony, graduation, graduation exercise",
  "n07455007": "formality, formalities",
  "n07455151": "Maundy",
  "n07455301": "potlatch",
  "n07455640": "fundraiser",
  "n07455760": "photo opportunity, photo op",
  "n07455984": "sleepover",
  "n07456188": "contest, competition",
  "n07456638": "athletic contest, athletic competition, athletics",
  "n07456906": "bout",
  "n07457016": "decathlon",
  "n07457126": "Olympic Games, Olympics, Olympiad",
  "n07457311": "Special Olympics",
  "n07457485": "Winter Olympic Games, Winter Olympics",
  "n07457599": "preliminary, prelim",
  "n07457722": "pentathlon",
  "n07457834": "championship",
  "n07457936": "chicken",
  "n07458099": "cliffhanger",
  "n07458212": "dogfight",
  "n07458453": "race",
  "n07458892": "automobile race, auto race, car race",
  "n07459066": "Grand Prix",
  "n07459157": "rally",
  "n07459249": "bicycle race",
  "n07459363": "Tour de France",
  "n07459523": "boat race",
  "n07459642": "burnup",
  "n07459778": "chariot race",
  "n07459868": "dog racing",
  "n07459992": "sailing-race, yacht race",
  "n07460104": "footrace, foot race, run",
  "n07460403": "funrun, fun run",
  "n07460546": "marathon",
  "n07460651": "freestyle",
  "n07460793": "cross country",
  "n07460904": "Iditarod, Iditarod Trail Dog Sled Race",
  "n07461050": "three-day event",
  "n07461288": "heat",
  "n07461411": "horse race",
  "n07461650": "claiming race",
  "n07461831": "selling race",
  "n07461956": "harness race, harness racing",
  "n07462119": "Kentucky Derby",
  "n07462269": "Preakness",
  "n07462405": "Belmont Stakes",
  "n07462547": "stake race",
  "n07462693": "steeplechase",
  "n07462808": "Grand National",
  "n07462917": "obstacle race",
  "n07463029": "steeplechase",
  "n07463171": "thoroughbred race",
  "n07463323": "potato race",
  "n07463470": "sack race",
  "n07463607": "scratch race",
  "n07463733": "ski race, skiing race",
  "n07463872": "downhill",
  "n07463950": "slalom",
  "n07464083": "relay, relay race",
  "n07464234": "repechage",
  "n07464402": "torch race",
  "n07464568": "World Cup",
  "n07464725": "tournament, tourney",
  "n07464969": "elimination tournament",
  "n07465159": "open",
  "n07465290": "playoff",
  "n07465448": "series",
  "n07465657": "home stand",
  "n07465781": "World Series",
  "n07465960": "boxing match",
  "n07466108": "chess match",
  "n07466195": "cockfight",
  "n07466322": "cricket match",
  "n07466415": "diving, diving event",
  "n07466557": "field event",
  "n07466832": "final",
  "n07467027": "cup final",
  "n07467212": "quarterfinal",
  "n07467393": "semifinal, semi",
  "n07467579": "round robin",
  "n07467704": "field trial",
  "n07467846": "meet, sports meeting",
  "n07468116": "gymkhana",
  "n07468244": "race meeting",
  "n07468456": "regatta",
  "n07468534": "swimming meet, swim meet",
  "n07468692": "track meet",
  "n07468861": "track event",
  "n07469043": "dash",
  "n07469162": "hurdles, hurdling, hurdle race",
  "n07469325": "mile",
  "n07469476": "high jump",
  "n07469609": "long jump, broad jump",
  "n07469772": "pole vault, pole vaulting, pole jump, pole jumping",
  "n07469975": "shot put",
  "n07470110": "hammer throw",
  "n07470285": "discus",
  "n07470420": "javelin",
  "n07470545": "swimming event",
  "n07470671": "match",
  "n07471037": "tennis match",
  "n07471126": "test match, match game, matched game",
  "n07471246": "wrestling match",
  "n07471371": "fall, pin",
  "n07471514": "takedown",
  "n07471708": "sparring match",
  "n07471806": "prizefight, prize fight",
  "n07471972": "triple jump, hop-step-and-jump",
  "n07472168": "tug-of-war",
  "n07472327": "tournament",
  "n07472460": "joust, tilt",
  "n07472657": "race",
  "n07472808": "arms race",
  "n07472929": "political campaign, campaign, run",
  "n07473207": "governor's race, campaign for governor",
  "n07473332": "senate campaign, senate race",
  "n07473441": "victory, triumph",
  "n07474006": "independence",
  "n07474185": "landslide",
  "n07474318": "last laugh",
  "n07474533": "Pyrrhic victory",
  "n07474645": "slam, sweep",
  "n07474809": "grand slam",
  "n07474911": "little slam, small slam",
  "n07475035": "checkmate",
  "n07475107": "runaway, blowout, romp, laugher, shoo-in, walkaway",
  "n07475242": "service break",
  "n07475364": "defeat, licking",
  "n07475762": "walk-in, waltz",
  "n07475870": "reverse, reversal, setback, blow, black eye",
  "n07476092": "whammy",
  "n07476177": "heartbreaker",
  "n07476282": "lurch",
  "n07476404": "rout",
  "n07476495": "shutout, skunk",
  "n07476623": "thrashing, walloping, debacle, drubbing, slaughter, trouncing, whipping",
  "n07476851": "waterloo",
  "n07476952": "whitewash",
  "n07477066": "spelling bee, spelldown, spelling contest",
  "n07477231": "trial",
  "n07477413": "bite",
  "n07477587": "boom, bonanza, gold rush, gravy, godsend, manna from heaven, windfall, bunce",
  "n07477945": "crash, collapse",
  "n07478169": "loss of consciousness",
  "n07478318": "faint, swoon, syncope, deliquium",
  "n07478531": "Fall",
  "n07478718": "shipwreck",
  "n07478874": "crash",
  "n07479144": "head crash",
  "n07479366": "spike",
  "n07479525": "supervention",
  "n07479628": "zap",
  "n07479799": "zizz",
  "n07479926": "affect",
  "n07480068": "emotion",
  "n07480356": "thing",
  "n07480521": "glow",
  "n07480666": "faintness",
  "n07480790": "soul, soulfulness",
  "n07480896": "passion, passionateness",
  "n07481084": "infatuation",
  "n07481223": "wildness, abandon",
  "n07481375": "ardor, ardour, fervor, fervour, fervency, fire, fervidness",
  "n07481625": "storminess",
  "n07481785": "zeal",
  "n07481951": "sentiment",
  "n07482128": "sentimentality",
  "n07482267": "mawkishness, bathos",
  "n07482368": "razbliuto",
  "n07482521": "complex",
  "n07482782": "Oedipus complex, Oedipal complex",
  "n07483005": "Electra complex",
  "n07483120": "inferiority complex",
  "n07483305": "ambivalence, ambivalency",
  "n07483439": "conflict",
  "n07483622": "apathy",
  "n07483782": "emotionlessness, impassivity, impassiveness, phlegm, indifference, stolidity, unemotionality",
  "n07484109": "languor, lassitude, listlessness",
  "n07484265": "desire",
  "n07484547": "ambition, aspiration, dream",
  "n07484792": "American Dream",
  "n07484929": "emulation",
  "n07485047": "nationalism",
  "n07485203": "bloodlust",
  "n07485281": "temptation",
  "n07485475": "craving",
  "n07485626": "appetite, appetency, appetence",
  "n07485916": "stomach",
  "n07486055": "sweet tooth",
  "n07486145": "addiction",
  "n07486229": "wish, wishing, want",
  "n07486527": "velleity",
  "n07486628": "longing, yearning, hungriness",
  "n07486922": "hankering, yen",
  "n07487063": "pining",
  "n07487177": "wishfulness",
  "n07487276": "wistfulness",
  "n07487375": "nostalgia",
  "n07487493": "lovesickness",
  "n07487594": "homesickness",
  "n07487695": "sex, sexual urge",
  "n07487955": "sexual desire, eros, concupiscence, physical attraction",
  "n07488340": "love, sexual love, erotic love",
  "n07488567": "aphrodisia",
  "n07488676": "anaphrodisia",
  "n07488788": "passion",
  "n07488875": "sensuality, sensualness, sensualism",
  "n07489059": "amorousness, eroticism, erotism, sexiness, amativeness",
  "n07489294": "fetish",
  "n07489548": "libido",
  "n07489714": "lecherousness, lust, lustfulness",
  "n07489925": "nymphomania",
  "n07490042": "satyriasis",
  "n07490138": "the hots",
  "n07490214": "prurience, pruriency, lasciviousness, carnality, lubricity",
  "n07490451": "urge, itch",
  "n07490579": "caprice, impulse, whim",
  "n07490713": "pleasure, pleasance",
  "n07491038": "delight, delectation",
  "n07491286": "entrancement, ravishment",
  "n07491476": "amusement",
  "n07491591": "Schadenfreude",
  "n07491708": "enjoyment",
  "n07491895": "joie de vivre",
  "n07491981": "gusto, relish, zest, zestfulness",
  "n07492170": "pleasantness",
  "n07492368": "afterglow",
  "n07492516": "comfort",
  "n07492655": "consolation, solace, solacement",
  "n07492928": "cold comfort",
  "n07493095": "silver lining, bright side",
  "n07493280": "relief, alleviation, assuagement",
  "n07493527": "sexual pleasure",
  "n07493682": "algolagnia, algophilia",
  "n07493830": "sadism",
  "n07494018": "sadomasochism",
  "n07494175": "masochism",
  "n07494363": "pain, painfulness",
  "n07494682": "growing pains",
  "n07494787": "unpleasantness",
  "n07494972": "pang, stab, twinge",
  "n07495156": "guilt pang",
  "n07495236": "mental anguish",
  "n07495327": "suffering, hurt",
  "n07495551": "agony, torment, torture",
  "n07495878": "throes",
  "n07495973": "discomfort, soreness, irritation",
  "n07496166": "chafing",
  "n07496291": "intertrigo",
  "n07496463": "distress, hurt, suffering",
  "n07496755": "anguish, torment, torture",
  "n07496924": "self-torture, self-torment",
  "n07497019": "tsoris",
  "n07497122": "wound",
  "n07497473": "liking",
  "n07497797": "fondness, fancy, partiality",
  "n07497976": "captivation, enchantment, enthrallment, fascination",
  "n07498210": "preference, penchant, predilection, taste",
  "n07498455": "acquired taste",
  "n07498614": "weakness",
  "n07498769": "mysophilia",
  "n07498854": "inclination",
  "n07499113": "leaning, propensity, tendency",
  "n07499276": "stomach",
  "n07499449": "undertow",
  "n07499615": "friendliness",
  "n07499830": "amicability, amicableness",
  "n07499930": "good will, goodwill",
  "n07500042": "brotherhood",
  "n07500159": "approval",
  "n07500414": "favor, favour",
  "n07500612": "approbation",
  "n07500741": "admiration, esteem",
  "n07500957": "Anglophilia",
  "n07501093": "hero worship",
  "n07501194": "philhellenism",
  "n07501326": "philogyny",
  "n07501420": "worship, adoration",
  "n07501545": "dislike",
  "n07501922": "disinclination",
  "n07502099": "Anglophobia",
  "n07502241": "unfriendliness",
  "n07502387": "alienation, disaffection, estrangement",
  "n07502576": "isolation",
  "n07502669": "antipathy, aversion, distaste",
  "n07502829": "disapproval",
  "n07502980": "contempt, disdain, scorn, despite",
  "n07503260": "disgust",
  "n07503430": "abhorrence, abomination, detestation, execration, loathing, odium",
  "n07503716": "creepy-crawlies",
  "n07503849": "scunner",
  "n07503987": "repugnance, repulsion, revulsion, horror",
  "n07504111": "nausea",
  "n07504258": "technophobia",
  "n07504343": "gratitude",
  "n07504529": "gratefulness, thankfulness, appreciativeness",
  "n07504711": "ingratitude, ungratefulness",
  "n07504841": "concern",
  "n07505047": "care",
  "n07505207": "solicitude, solicitousness",
  "n07505347": "softheartedness, tenderness",
  "n07505538": "unconcern",
  "n07505676": "indifference",
  "n07505871": "distance, aloofness",
  "n07506031": "withdrawal, detachment",
  "n07506149": "heartlessness, coldheartedness, hardheartedness",
  "n07506382": "cruelty, mercilessness, pitilessness, ruthlessness",
  "n07506569": "shame",
  "n07506797": "conscience",
  "n07506962": "self-disgust, self-hatred",
  "n07507098": "embarrassment",
  "n07507329": "self-consciousness, uneasiness, uncomfortableness",
  "n07507560": "shamefacedness, sheepishness",
  "n07507742": "chagrin, humiliation, mortification",
  "n07507912": "confusion, discombobulation",
  "n07508092": "abashment, bashfulness",
  "n07508232": "discomfiture, discomposure, disconcertion, disconcertment",
  "n07508486": "pride, pridefulness",
  "n07508705": "self-esteem, self-pride",
  "n07508806": "ego, egotism, self-importance",
  "n07508996": "amour propre, conceit, self-love, vanity",
  "n07509131": "humility, humbleness",
  "n07509325": "meekness, submission",
  "n07509474": "self-depreciation",
  "n07509572": "astonishment, amazement",
  "n07509827": "devastation",
  "n07509996": "wonder, wonderment, admiration",
  "n07510184": "awe",
  "n07510348": "surprise",
  "n07510495": "stupefaction",
  "n07510625": "daze, shock, stupor",
  "n07510923": "expectation",
  "n07511080": "anticipation, expectancy",
  "n07511238": "suspense",
  "n07511380": "fever",
  "n07511524": "buck fever",
  "n07511626": "gold fever",
  "n07511733": "hope",
  "n07511906": "levity",
  "n07512039": "gaiety, playfulness",
  "n07512147": "gravity, solemnity",
  "n07512315": "earnestness, seriousness, sincerity",
  "n07512465": "sensitivity, sensitiveness",
  "n07512719": "oversensitiveness",
  "n07512848": "sensibility",
  "n07513035": "feelings",
  "n07513247": "insight, perceptiveness, perceptivity",
  "n07513413": "sensuousness",
  "n07513508": "agitation",
  "n07513711": "unrest",
  "n07513795": "fidget, fidgetiness, restlessness",
  "n07514071": "impatience",
  "n07514189": "stewing",
  "n07514345": "stir",
  "n07514520": "tumult, turmoil",
  "n07514600": "electricity",
  "n07514782": "sensation",
  "n07514968": "calmness",
  "n07515169": "placidity, placidness",
  "n07515328": "coolness, imperturbability, imperturbableness",
  "n07515560": "tranquillity, tranquility, quietness, quietude",
  "n07515790": "peace, peacefulness, peace of mind, repose, serenity, heartsease, ataraxis",
  "n07515974": "easiness, relaxation",
  "n07516222": "languor, dreaminess",
  "n07516354": "anger, choler, ire",
  "n07516756": "dudgeon, high dudgeon",
  "n07516905": "wrath",
  "n07516997": "fury, rage, madness",
  "n07517292": "lividity",
  "n07517417": "infuriation, enragement",
  "n07517550": "umbrage, offense, offence",
  "n07517737": "indignation, outrage",
  "n07517869": "huffiness",
  "n07518000": "dander, hackles",
  "n07518132": "bad temper, ill temper",
  "n07518261": "annoyance, chafe, vexation",
  "n07518468": "pique, temper, irritation",
  "n07518663": "frustration",
  "n07518878": "aggravation, exasperation",
  "n07519040": "harassment, torment",
  "n07519253": "fear, fearfulness, fright",
  "n07519773": "alarm, dismay, consternation",
  "n07519983": "creeps",
  "n07520112": "frisson, shiver, chill, quiver, shudder, thrill, tingle",
  "n07520411": "horror",
  "n07520507": "hysteria",
  "n07520612": "panic, terror, affright",
  "n07520925": "swivet",
  "n07521039": "fear, reverence, awe, veneration",
  "n07521437": "scare, panic attack",
  "n07521563": "stage fright",
  "n07521674": "apprehension, apprehensiveness, dread",
  "n07522043": "trepidation",
  "n07522128": "foreboding, premonition, presentiment, boding",
  "n07522418": "shadow",
  "n07522536": "presage",
  "n07522632": "suspense",
  "n07522729": "timidity, timidness, timorousness",
  "n07523016": "cold feet",
  "n07523180": "shyness",
  "n07523286": "diffidence, self-doubt, self-distrust",
  "n07523485": "hesitance, hesitancy",
  "n07523649": "unassertiveness",
  "n07523760": "intimidation",
  "n07523905": "anxiety",
  "n07524242": "worry, trouble",
  "n07524529": "concern, care, fear",
  "n07524760": "anxiousness, disquiet",
  "n07524918": "insecurity",
  "n07525057": "edginess, uneasiness, inquietude, disquietude",
  "n07525259": "willies",
  "n07525367": "sinking, sinking feeling",
  "n07525555": "scruple, qualm, misgiving",
  "n07525760": "jitteriness, jumpiness, nervousness, restiveness",
  "n07526002": "angst",
  "n07526182": "fearlessness, bravery",
  "n07526338": "security",
  "n07526505": "confidence",
  "n07526757": "happiness",
  "n07527061": "bonheur",
  "n07527167": "gladness, gladfulness, gladsomeness",
  "n07527352": "joy, joyousness, joyfulness",
  "n07527656": "elation, high spirits, lightness",
  "n07527817": "exultation, jubilance, jubilancy, jubilation",
  "n07528097": "triumph",
  "n07528212": "exhilaration, excitement",
  "n07528470": "bang, boot, charge, rush, flush, thrill, kick",
  "n07528807": "intoxication",
  "n07528976": "titillation",
  "n07529096": "euphoria, euphory",
  "n07529245": "gaiety, merriment",
  "n07529377": "hilarity, mirth, mirthfulness, glee, gleefulness",
  "n07529563": "rejoicing",
  "n07529683": "jocundity, jocularity",
  "n07529817": "belonging",
  "n07530000": "comfortableness",
  "n07530124": "closeness, intimacy",
  "n07530322": "togetherness",
  "n07530478": "cheerfulness, blitheness",
  "n07530723": "buoyancy, perkiness",
  "n07530866": "carefreeness, insouciance, lightheartedness, lightsomeness",
  "n07531105": "contentment",
  "n07531255": "satisfaction",
  "n07531536": "pride",
  "n07531713": "complacency, complacence, self-complacency, self-satisfaction",
  "n07531998": "smugness",
  "n07532112": "fulfillment, fulfilment",
  "n07532276": "gloat, gloating, glee",
  "n07532440": "sadness, unhappiness",
  "n07532832": "dolefulness",
  "n07532945": "heaviness",
  "n07533097": "melancholy",
  "n07533257": "gloom, gloominess, somberness, sombreness",
  "n07533492": "heavyheartedness",
  "n07533607": "pensiveness, brooding",
  "n07533735": "world-weariness, Weltschmerz",
  "n07533877": "woe, woefulness",
  "n07533978": "misery",
  "n07534108": "forlornness, loneliness, desolation",
  "n07534278": "weepiness, tearfulness",
  "n07534430": "sorrow",
  "n07534700": "attrition, contrition, contriteness",
  "n07534847": "broken heart",
  "n07535010": "grief, heartache, heartbreak, brokenheartedness",
  "n07535209": "mournfulness, sorrowfulness, ruthfulness",
  "n07535430": "plaintiveness",
  "n07535532": "dolor, dolour",
  "n07535670": "sorrow, regret, rue, ruefulness",
  "n07536074": "compunction, remorse, self-reproach",
  "n07536245": "guilt, guilty conscience, guilt feelings, guilt trip",
  "n07536437": "survivor guilt",
  "n07536870": "repentance, penitence, penance",
  "n07537068": "cheerlessness, uncheerfulness",
  "n07537259": "chill, pall",
  "n07537376": "joylessness",
  "n07537485": "depression",
  "n07537668": "downheartedness, dejectedness, low-spiritedness, lowness, dispiritedness",
  "n07537973": "demoralization, demoralisation",
  "n07538142": "helplessness",
  "n07538272": "self-pity",
  "n07538395": "despondency, despondence, heartsickness, disconsolateness",
  "n07538674": "oppression, oppressiveness",
  "n07538812": "weight",
  "n07538965": "discontentment, discontent, discontentedness",
  "n07539259": "disgruntlement",
  "n07539367": "dysphoria",
  "n07539511": "dissatisfaction",
  "n07539790": "boredom, ennui, tedium",
  "n07539962": "blahs",
  "n07540081": "fatigue",
  "n07540424": "displeasure",
  "n07540602": "disappointment, letdown",
  "n07540866": "frustration, defeat",
  "n07541053": "hope",
  "n07541316": "hopefulness",
  "n07541449": "encouragement",
  "n07541558": "optimism",
  "n07541760": "sanguinity, sanguineness",
  "n07541923": "despair",
  "n07542249": "hopelessness",
  "n07542433": "resignation, surrender",
  "n07542560": "defeatism",
  "n07542675": "discouragement, disheartenment, dismay",
  "n07542881": "intimidation",
  "n07543033": "pessimism",
  "n07543204": "cynicism",
  "n07543288": "love",
  "n07543732": "agape, agape love",
  "n07543910": "agape",
  "n07544039": "filial love",
  "n07544129": "ardor, ardour",
  "n07544213": "amorousness, enamoredness",
  "n07544351": "puppy love, calf love, crush, infatuation",
  "n07544491": "devotion, devotedness",
  "n07544647": "affection, affectionateness, fondness, tenderness, heart, warmness, warmheartedness, philia",
  "n07545161": "attachment, fond regard",
  "n07545303": "protectiveness",
  "n07545415": "regard, respect",
  "n07545594": "soft spot",
  "n07545717": "benevolence",
  "n07545833": "beneficence",
  "n07545957": "heartstrings",
  "n07546125": "lovingness, caring",
  "n07546279": "warmheartedness, warmth",
  "n07546389": "loyalty",
  "n07546465": "hate, hatred",
  "n07546844": "misanthropy",
  "n07546973": "misogamy",
  "n07547064": "misogyny, misogynism",
  "n07547219": "misology",
  "n07547293": "misoneism",
  "n07547397": "misocainea",
  "n07547473": "misopedia",
  "n07547547": "murderousness",
  "n07547674": "despisal, despising",
  "n07547805": "hostility, enmity, ill will",
  "n07548100": "animosity, animus, bad blood",
  "n07548224": "class feeling",
  "n07548366": "antagonism",
  "n07548567": "aggression, aggressiveness",
  "n07548695": "belligerence, belligerency",
  "n07548860": "warpath",
  "n07548978": "resentment, bitterness, gall, rancor, rancour",
  "n07549265": "heartburning",
  "n07549401": "sulkiness, huffishness",
  "n07549536": "grudge, score, grievance",
  "n07549716": "envy, enviousness",
  "n07549979": "covetousness",
  "n07550079": "jealousy, green-eyed monster",
  "n07550204": "penis envy",
  "n07550369": "malevolence, malignity",
  "n07550551": "maleficence",
  "n07550666": "malice, maliciousness, spite, spitefulness, venom",
  "n07550891": "vindictiveness, vengefulness",
  "n07551052": "temper, mood, humor, humour",
  "n07551418": "peeve",
  "n07551498": "sulk, sulkiness",
  "n07551691": "good humor, good humour, good temper, amiability",
  "n07551890": "jollity, jolliness, joviality",
  "n07552087": "ill humor, ill humour, distemper",
  "n07552252": "moodiness",
  "n07552367": "moroseness, glumness, sullenness",
  "n07552549": "irascibility, short temper, spleen, quick temper",
  "n07552729": "irritability, crossness, fretfulness, fussiness, peevishness, petulance, choler",
  "n07553016": "testiness, touchiness, tetchiness",
  "n07553176": "pet",
  "n07553301": "sympathy, fellow feeling",
  "n07553590": "kindheartedness, kind-heartedness",
  "n07553741": "compassion, compassionateness",
  "n07553964": "commiseration, pity, ruth, pathos",
  "n07554224": "mellowness",
  "n07554342": "tenderness, tenderheartedness",
  "n07554500": "mercifulness, mercy",
  "n07554640": "forgiveness",
  "n07554758": "compatibility",
  "n07554856": "empathy",
  "n07555014": "enthusiasm",
  "n07555184": "eagerness, avidity, avidness, keenness",
  "n07555402": "ardor, ardour, elan, zeal",
  "n07555647": "exuberance",
  "n07555775": "technophilia",
  "n07555863": "food, solid food",
  "n07556406": "comfort food",
  "n07556637": "comestible, edible, eatable, pabulum, victual, victuals",
  "n07556872": "tuck",
  "n07556970": "course",
  "n07557165": "dainty, delicacy, goody, kickshaw, treat",
  "n07557434": "dish",
  "n07560193": "fast food",
  "n07560331": "finger food",
  "n07560422": "ingesta",
  "n07560542": "kosher",
  "n07560652": "fare",
  "n07560903": "diet",
  "n07561112": "diet",
  "n07561590": "dietary",
  "n07561730": "allergy diet",
  "n07561848": "balanced diet",
  "n07562017": "bland diet, ulcer diet",
  "n07562172": "clear liquid diet",
  "n07562379": "diabetic diet",
  "n07562495": "dietary supplement",
  "n07562651": "carbohydrate loading, carbo loading",
  "n07562881": "fad diet",
  "n07562984": "gluten-free diet",
  "n07563207": "high-protein diet",
  "n07563366": "high-vitamin diet, vitamin-deficiency diet",
  "n07563509": "leftovers",
  "n07563642": "light diet",
  "n07563800": "liquid diet",
  "n07564008": "low-calorie diet",
  "n07564101": "low-fat diet",
  "n07564292": "low-sodium diet, low-salt diet, salt-free diet",
  "n07564515": "macrobiotic diet",
  "n07564629": "reducing diet, obesity diet",
  "n07564796": "soft diet, pap, spoon food",
  "n07564971": "vegetarianism",
  "n07565083": "menu",
  "n07565161": "chow, chuck, eats, grub",
  "n07565259": "board, table",
  "n07565458": "training table",
  "n07565608": "mess",
  "n07565725": "ration",
  "n07565945": "field ration",
  "n07566092": "K ration",
  "n07566231": "C-ration",
  "n07566340": "foodstuff, food product",
  "n07566863": "starches",
  "n07567039": "breadstuff",
  "n07567139": "coloring, colouring, food coloring, food colouring, food color, food colour",
  "n07567390": "concentrate",
  "n07567611": "tomato concentrate",
  "n07567707": "meal",
  "n07567980": "kibble",
  "n07568095": "cornmeal, Indian meal",
  "n07568241": "farina",
  "n07568389": "matzo meal, matzoh meal, matzah meal",
  "n07568502": "oatmeal, rolled oats",
  "n07568625": "pea flour",
  "n07568706": "pinole",
  "n07568818": "roughage, fiber",
  "n07568991": "bran",
  "n07569106": "flour",
  "n07569423": "plain flour",
  "n07569543": "wheat flour",
  "n07569644": "whole wheat flour, graham flour, graham, whole meal flour",
  "n07569873": "soybean meal, soybean flour, soy flour",
  "n07570021": "semolina",
  "n07570141": "blood meal",
  "n07570237": "gluten",
  "n07570447": "corn gluten",
  "n07570530": "corn gluten feed",
  "n07570635": "wheat gluten",
  "n07570720": "nutriment, nourishment, nutrition, sustenance, aliment, alimentation, victuals",
  "n07571324": "cuisine, culinary art",
  "n07571547": "dim sum",
  "n07571765": "haute cuisine",
  "n07571903": "nouvelle cuisine",
  "n07572134": "rechauffe",
  "n07572206": "gastronomy",
  "n07572353": "commissariat, provisions, provender, viands, victuals",
  "n07572519": "food cache",
  "n07572616": "larder",
  "n07572712": "fresh food, fresh foods",
  "n07572858": "frozen food, frozen foods",
  "n07572957": "canned food, canned foods, canned goods, tinned goods",
  "n07573103": "canned meat, tinned meat",
  "n07573241": "Fanny Adams",
  "n07573347": "Spam",
  "n07573453": "dehydrated food, dehydrated foods",
  "n07573563": "square meal",
  "n07573696": "meal, repast",
  "n07574176": "potluck",
  "n07574426": "refection",
  "n07574504": "refreshment",
  "n07574602": "breakfast",
  "n07574780": "continental breakfast, petit dejeuner",
  "n07574923": "brunch",
  "n07575076": "lunch, luncheon, tiffin, dejeuner",
  "n07575226": "business lunch",
  "n07575392": "high tea",
  "n07575510": "tea, afternoon tea, teatime",
  "n07575726": "dinner",
  "n07575984": "supper",
  "n07576182": "buffet",
  "n07576311": "TV dinner",
  "n07576438": "picnic",
  "n07576577": "cookout",
  "n07576781": "barbecue, barbeque",
  "n07576969": "clambake",
  "n07577144": "fish fry",
  "n07577244": "wiener roast, weenie roast",
  "n07577374": "bite, collation, snack",
  "n07577538": "nosh",
  "n07577657": "nosh-up",
  "n07577772": "ploughman's lunch",
  "n07577918": "coffee break, tea break",
  "n07578093": "banquet, feast, spread",
  "n07578363": "helping, portion, serving",
  "n07578879": "taste, mouthful",
  "n07579076": "morsel, bit, bite",
  "n07579276": "swallow, sup",
  "n07579399": "chew, chaw, cud, quid, plug, wad",
  "n07579575": "entree, main course",
  "n07579688": "piece de resistance",
  "n07579787": "plate",
  "n07579917": "adobo",
  "n07580053": "side dish, side order, entremets",
  "n07580253": "special",
  "n07580359": "casserole",
  "n07580470": "chicken casserole",
  "n07580592": "chicken cacciatore, chicken cacciatora, hunter's chicken",
  "n07580782": "roast, joint",
  "n07581132": "confit",
  "n07581249": "antipasto",
  "n07581346": "appetizer, appetiser, starter",
  "n07581607": "canape",
  "n07581775": "cocktail",
  "n07581931": "fruit cocktail",
  "n07582027": "crab cocktail",
  "n07582152": "shrimp cocktail",
  "n07582277": "hors d'oeuvre",
  "n07582441": "relish",
  "n07582609": "dip",
  "n07582811": "bean dip",
  "n07582892": "cheese dip",
  "n07582970": "clam dip",
  "n07583066": "guacamole",
  "n07583197": "soup",
  "n07583865": "soup du jour",
  "n07583978": "alphabet soup",
  "n07584110": "consomme",
  "n07584228": "madrilene",
  "n07584332": "bisque",
  "n07584423": "borsch, borsh, borscht, borsht, borshch, bortsch",
  "n07584593": "broth",
  "n07584727": "liquor, pot liquor, pot likker",
  "n07584859": "barley water",
  "n07584938": "bouillon",
  "n07585015": "beef broth, beef stock",
  "n07585107": "chicken broth, chicken stock",
  "n07585208": "broth, stock",
  "n07585474": "stock cube",
  "n07585557": "chicken soup",
  "n07585644": "cock-a-leekie, cocky-leeky",
  "n07585758": "gazpacho",
  "n07585906": "gumbo",
  "n07585997": "julienne",
  "n07586099": "marmite",
  "n07586179": "mock turtle soup",
  "n07586318": "mulligatawny",
  "n07586485": "oxtail soup",
  "n07586604": "pea soup",
  "n07586718": "pepper pot, Philadelphia pepper pot",
  "n07586894": "petite marmite, minestrone, vegetable soup",
  "n07587023": "potage, pottage",
  "n07587111": "pottage",
  "n07587206": "turtle soup, green turtle soup",
  "n07587331": "eggdrop soup",
  "n07587441": "chowder",
  "n07587618": "corn chowder",
  "n07587700": "clam chowder",
  "n07587819": "Manhattan clam chowder",
  "n07587962": "New England clam chowder",
  "n07588111": "fish chowder",
  "n07588193": "won ton, wonton, wonton soup",
  "n07588299": "split-pea soup",
  "n07588419": "green pea soup, potage St. Germain",
  "n07588574": "lentil soup",
  "n07588688": "Scotch broth",
  "n07588817": "vichyssoise",
  "n07588947": "stew",
  "n07589458": "bigos",
  "n07589543": "Brunswick stew",
  "n07589724": "burgoo",
  "n07589872": "burgoo",
  "n07589967": "olla podrida, Spanish burgoo",
  "n07590068": "mulligan stew, mulligan, Irish burgoo",
  "n07590177": "purloo, chicken purloo, poilu",
  "n07590320": "goulash, Hungarian goulash, gulyas",
  "n07590502": "hotchpotch",
  "n07590611": "hot pot, hotpot",
  "n07590752": "beef goulash",
  "n07590841": "pork-and-veal goulash",
  "n07590974": "porkholt",
  "n07591049": "Irish stew",
  "n07591162": "oyster stew",
  "n07591236": "lobster stew",
  "n07591330": "lobscouse, lobscuse, scouse",
  "n07591473": "fish stew",
  "n07591586": "bouillabaisse",
  "n07591813": "matelote",
  "n07591961": "paella",
  "n07592094": "fricassee",
  "n07592317": "chicken stew",
  "n07592400": "turkey stew",
  "n07592481": "beef stew",
  "n07592558": "stew meat",
  "n07592656": "ragout",
  "n07592768": "ratatouille",
  "n07592922": "salmi",
  "n07593004": "pot-au-feu",
  "n07593107": "slumgullion",
  "n07593199": "smorgasbord",
  "n07593471": "viand",
  "n07593549": "convenience food",
  "n07593774": "ready-mix",
  "n07593972": "brownie mix",
  "n07594066": "cake mix",
  "n07594155": "lemonade mix",
  "n07594250": "self-rising flour, self-raising flour",
  "n07594406": "delicatessen, delicatessen food",
  "n07594511": "takeout, takeout food, takeaway",
  "n07594737": "choice morsel, tidbit, titbit",
  "n07594840": "savory, savoury",
  "n07595051": "calf's-foot jelly",
  "n07595180": "caramel, caramelized sugar",
  "n07595368": "lump sugar",
  "n07595499": "sugarloaf, sugar loaf, loaf sugar",
  "n07595649": "cane sugar",
  "n07595751": "castor sugar, caster sugar",
  "n07595914": "powdered sugar",
  "n07596046": "granulated sugar",
  "n07596160": "icing sugar",
  "n07596258": "beet sugar",
  "n07596362": "corn sugar",
  "n07596452": "brown sugar",
  "n07596566": "demerara, demerara sugar",
  "n07596684": "sweet, confection",
  "n07596967": "confectionery",
  "n07597145": "confiture",
  "n07597263": "sweetmeat",
  "n07597365": "candy, confect",
  "n07598256": "candy bar",
  "n07598335": "carob, carob powder, Saint-John's-bread",
  "n07598529": "carob bar",
  "n07598622": "hardbake",
  "n07598734": "hard candy",
  "n07598928": "barley-sugar, barley candy",
  "n07599068": "brandyball",
  "n07599161": "jawbreaker",
  "n07599242": "lemon drop",
  "n07599383": "sourball",
  "n07599468": "patty",
  "n07599554": "peppermint patty",
  "n07599649": "bonbon",
  "n07599783": "brittle, toffee, toffy",
  "n07599911": "peanut brittle",
  "n07599998": "chewing gum, gum",
  "n07600177": "gum ball",
  "n07600285": "bubble gum",
  "n07600394": "butterscotch",
  "n07600506": "candied fruit, succade, crystallized fruit",
  "n07600696": "candied apple, candy apple, taffy apple, caramel apple, toffee apple",
  "n07600895": "crystallized ginger",
  "n07601025": "grapefruit peel",
  "n07601175": "lemon peel",
  "n07601290": "orange peel",
  "n07601407": "candied citrus peel",
  "n07601572": "candy cane",
  "n07601686": "candy corn",
  "n07601809": "caramel",
  "n07601999": "chocolate",
  "n07602279": "bitter chocolate, baking chocolate, cooking chocolate",
  "n07602454": "chocolate candy",
  "n07602650": "center, centre",
  "n07602829": "chocolate liquor",
  "n07602996": "cocoa butter",
  "n07603177": "cocoa powder",
  "n07603411": "choc",
  "n07603511": "chocolate bar",
  "n07603613": "Hershey bar",
  "n07603722": "bittersweet chocolate, semi-sweet chocolate, dark chocolate",
  "n07603930": "couverture",
  "n07604043": "Dutch-processed cocoa",
  "n07604182": "jimmies, sprinkles",
  "n07604307": "milk chocolate",
  "n07604587": "white chocolate",
  "n07604840": "nonpareil",
  "n07604956": "comfit",
  "n07605040": "cotton candy, spun sugar, candyfloss",
  "n07605198": "dragee",
  "n07605282": "dragee",
  "n07605380": "fondant",
  "n07605474": "fudge",
  "n07605597": "chocolate fudge",
  "n07605693": "divinity, divinity fudge",
  "n07605804": "penuche, penoche, panoche, panocha",
  "n07605944": "gumdrop",
  "n07606058": "jujube",
  "n07606191": "honey crisp",
  "n07606278": "mint, mint candy",
  "n07606419": "horehound",
  "n07606538": "peppermint, peppermint candy",
  "n07606669": "jelly bean, jelly egg",
  "n07606764": "kiss, candy kiss",
  "n07606933": "molasses kiss",
  "n07607027": "meringue kiss",
  "n07607138": "chocolate kiss",
  "n07607280": "Scotch kiss",
  "n07607361": "licorice, liquorice",
  "n07607492": "Life Saver",
  "n07607605": "lollipop, sucker, all-day sucker",
  "n07607707": "lozenge",
  "n07607832": "cachou",
  "n07607967": "cough drop, troche, pastille, pastil",
  "n07608098": "marshmallow",
  "n07608245": "marzipan, marchpane",
  "n07608339": "nougat",
  "n07608429": "nougat bar",
  "n07608533": "nut bar",
  "n07608641": "peanut bar",
  "n07608721": "popcorn ball",
  "n07608866": "praline",
  "n07608980": "rock candy",
  "n07609083": "rock candy, rock",
  "n07609215": "sugar candy",
  "n07609316": "sugarplum",
  "n07609407": "taffy",
  "n07609549": "molasses taffy",
  "n07609632": "truffle, chocolate truffle",
  "n07609728": "Turkish Delight",
  "n07609840": "dessert, sweet, afters",
  "n07610295": "ambrosia, nectar",
  "n07610502": "ambrosia",
  "n07610620": "baked Alaska",
  "n07610746": "blancmange",
  "n07610890": "charlotte",
  "n07611046": "compote, fruit compote",
  "n07611148": "dumpling",
  "n07611267": "flan",
  "n07611358": "frozen dessert",
  "n07611733": "junket",
  "n07611839": "mousse",
  "n07611991": "mousse",
  "n07612137": "pavlova",
  "n07612273": "peach melba",
  "n07612367": "whip",
  "n07612530": "prune whip",
  "n07612632": "pudding",
  "n07612996": "pudding, pud",
  "n07613158": "syllabub, sillabub",
  "n07613266": "tiramisu",
  "n07613480": "trifle",
  "n07613671": "tipsy cake",
  "n07613815": "jello, Jell-O",
  "n07614004": "charlotte russe",
  "n07614103": "apple dumpling",
  "n07614198": "ice, frappe",
  "n07614348": "water ice, sorbet",
  "n07614500": "ice cream, icecream",
  "n07614730": "ice-cream cone",
  "n07614825": "chocolate ice cream",
  "n07614942": "choc-ice",
  "n07615052": "Neapolitan ice cream",
  "n07615190": "peach ice cream",
  "n07615289": "sherbert, sherbet",
  "n07615460": "strawberry ice cream",
  "n07615569": "tutti-frutti",
  "n07615671": "vanilla ice cream",
  "n07615774": "ice lolly, lolly, lollipop, popsicle",
  "n07615954": "ice milk",
  "n07616046": "frozen yogurt",
  "n07616174": "snowball",
  "n07616265": "snowball",
  "n07616386": "parfait",
  "n07616487": "ice-cream sundae, sundae",
  "n07616590": "split",
  "n07616748": "banana split",
  "n07616906": "frozen pudding",
  "n07617051": "frozen custard, soft ice cream",
  "n07617188": "pudding",
  "n07617344": "flummery",
  "n07617447": "fish mousse",
  "n07617526": "chicken mousse",
  "n07617611": "chocolate mousse",
  "n07617708": "plum pudding, Christmas pudding",
  "n07617839": "carrot pudding",
  "n07617932": "corn pudding",
  "n07618029": "steamed pudding",
  "n07618119": "duff, plum duff",
  "n07618281": "vanilla pudding",
  "n07618432": "chocolate pudding",
  "n07618587": "brown Betty",
  "n07618684": "Nesselrode, Nesselrode pudding",
  "n07618871": "pease pudding",
  "n07619004": "custard",
  "n07619208": "creme caramel",
  "n07619301": "creme anglais",
  "n07619409": "creme brulee",
  "n07619508": "fruit custard",
  "n07619594": "quiche",
  "n07619786": "quiche Lorraine",
  "n07619881": "tapioca",
  "n07620047": "tapioca pudding",
  "n07620145": "roly-poly, roly-poly pudding",
  "n07620327": "suet pudding",
  "n07620485": "spotted dick",
  "n07620597": "Bavarian cream",
  "n07620689": "maraschino, maraschino cherry",
  "n07620822": "frosting, icing, ice",
  "n07621001": "glaze",
  "n07621140": "meringue",
  "n07621264": "nonpareil",
  "n07621388": "whipped cream",
  "n07621497": "zabaglione, sabayon",
  "n07621618": "garnish",
  "n07621776": "topping",
  "n07621976": "streusel",
  "n07622061": "baked goods",
  "n07622261": "crumb",
  "n07622473": "breadcrumb",
  "n07622587": "cracker crumbs",
  "n07622708": "pastry",
  "n07623136": "pastry, pastry dough",
  "n07623263": "pie crust, pie shell",
  "n07623363": "dowdy, pandowdy",
  "n07623475": "frangipane",
  "n07623576": "streusel",
  "n07623664": "tart",
  "n07623834": "apple tart",
  "n07623933": "tart",
  "n07624130": "apple tart",
  "n07624239": "lobster tart",
  "n07624339": "tartlet",
  "n07624466": "turnover",
  "n07624666": "apple turnover",
  "n07624757": "knish",
  "n07624924": "pirogi, piroshki, pirozhki",
  "n07625061": "samosa",
  "n07625201": "timbale, timbale case",
  "n07625324": "timbale",
  "n07625493": "pie",
  "n07625831": "deep-dish pie, cobbler",
  "n07625977": "shoofly pie",
  "n07626094": "mince pie",
  "n07626174": "apple pie",
  "n07626287": "lemon meringue pie",
  "n07626405": "blueberry pie",
  "n07626501": "rhubarb pie",
  "n07626602": "pecan pie",
  "n07626721": "pumpkin pie",
  "n07626833": "squash pie",
  "n07626959": "French pastry",
  "n07627086": "napoleon",
  "n07627223": "patty shell, bouchee",
  "n07627310": "patty",
  "n07627380": "sausage roll",
  "n07627478": "toad-in-the-hole",
  "n07627564": "vol-au-vent",
  "n07627693": "strudel",
  "n07627790": "baklava",
  "n07627931": "puff paste, pate feuillete",
  "n07628068": "phyllo",
  "n07628181": "puff batter, pouf paste, pate a choux",
  "n07628328": "profiterole",
  "n07628470": "puff",
  "n07628576": "cream puff, chou",
  "n07628691": "eclair",
  "n07628779": "chocolate eclair",
  "n07628870": "cake",
  "n07629697": "applesauce cake",
  "n07629797": "baba",
  "n07629898": "baba au rhum, rum baba",
  "n07629988": "birthday cake",
  "n07630089": "cheesecake",
  "n07630220": "chiffon cake",
  "n07630294": "chocolate cake",
  "n07630398": "coconut cake",
  "n07630512": "coffeecake, coffee cake",
  "n07630667": "babka",
  "n07630782": "crumb cake",
  "n07630909": "crumpet",
  "n07631023": "cupcake",
  "n07631109": "devil's food, devil's food cake",
  "n07631212": "Eccles cake",
  "n07631350": "fruitcake",
  "n07631511": "Christmas cake",
  "n07631672": "simnel",
  "n07631834": "gateau",
  "n07631926": "ice-cream cake, icebox cake",
  "n07632037": "sponge cake",
  "n07632240": "angel cake, angel food cake",
  "n07632357": "jellyroll, Swiss roll",
  "n07632515": "Madeira cake, Madeira sponge",
  "n07632679": "Twinkie",
  "n07632802": "wedding cake, bridecake",
  "n07632980": "white cake",
  "n07633064": "spice cake",
  "n07633146": "gingerbread",
  "n07633229": "pound cake",
  "n07633369": "layer cake",
  "n07633511": "torte",
  "n07633640": "petit four",
  "n07633743": "prune cake",
  "n07633861": "jumble, jumbal",
  "n07633960": "savarin",
  "n07634048": "Boston cream pie",
  "n07634141": "upside-down cake, skillet cake",
  "n07634306": "honey cake",
  "n07634406": "marble cake",
  "n07634514": "genoise",
  "n07634605": "seedcake, seed cake",
  "n07634751": "teacake",
  "n07634901": "teacake, tea biscuit",
  "n07635044": "Sally Lunn",
  "n07635155": "cookie, cooky, biscuit",
  "n07635746": "dog biscuit",
  "n07635827": "butter cookie",
  "n07635934": "spice cookie",
  "n07636020": "shortbread, shortbread cookie",
  "n07636126": "almond cookie, almond crescent",
  "n07636271": "brownie",
  "n07636384": "gingersnap, ginger snap, snap, ginger nut",
  "n07636534": "macaroon",
  "n07636669": "ratafia, ratafia biscuit",
  "n07636779": "coconut macaroon",
  "n07636869": "kiss",
  "n07636957": "ladyfinger",
  "n07637045": "anise cookie",
  "n07637178": "molasses cookie",
  "n07637292": "oreo, oreo cookie",
  "n07637398": "raisin-nut cookie",
  "n07637508": "refrigerator cookie",
  "n07637652": "raisin cookie",
  "n07637737": "fruit bar",
  "n07637949": "apricot bar",
  "n07638039": "date bar",
  "n07638128": "sugar cookie",
  "n07638226": "oatmeal cookie",
  "n07638317": "chocolate chip cookie, Toll House cookie",
  "n07638439": "fortune cookie",
  "n07638574": "gingerbread man",
  "n07638676": "friedcake",
  "n07638898": "doughboy",
  "n07639069": "doughnut, donut, sinker",
  "n07639188": "raised doughnut",
  "n07639344": "Berlin doughnut, bismark, jelly doughnut",
  "n07639475": "fastnacht",
  "n07639577": "cruller, twister",
  "n07639716": "French fritter, beignet",
  "n07639855": "fritter",
  "n07640014": "apple fritter",
  "n07640105": "corn fritter",
  "n07640203": "pancake, battercake, flannel cake, flannel-cake, flapcake, flapjack, griddlecake, hotcake, hot cake",
  "n07640560": "yeast cake",
  "n07640653": "buckwheat cake",
  "n07640749": "buttermilk pancake",
  "n07640844": "blini, bliny",
  "n07640991": "blintz, blintze",
  "n07641138": "crape, crepe, French pancake",
  "n07641256": "crepe Suzette",
  "n07641380": "pfannkuchen, german pancake",
  "n07641581": "potato pancake, latke",
  "n07641699": "waffle",
  "n07641807": "Belgian waffle",
  "n07641928": "fish cake, fish ball",
  "n07642070": "rock cake",
  "n07642182": "Victoria sandwich, Victoria sponge",
  "n07642361": "fish stick, fish finger",
  "n07642471": "conserve, preserve, conserves, preserves",
  "n07642742": "apple butter",
  "n07642833": "chowchow",
  "n07642933": "jam",
  "n07643026": "lemon curd, lemon cheese",
  "n07643200": "strawberry jam, strawberry preserves",
  "n07643306": "jelly",
  "n07643474": "apple jelly",
  "n07643577": "crabapple jelly",
  "n07643679": "grape jelly",
  "n07643764": "marmalade",
  "n07643891": "orange marmalade",
  "n07643981": "gelatin, jelly",
  "n07644244": "gelatin dessert",
  "n07644382": "bird, fowl",
  "n07644706": "poultry",
  "n07644967": "chicken, poulet, volaille",
  "n07645242": "broiler",
  "n07645384": "capon",
  "n07645469": "fryer, frier, pullet",
  "n07645631": "roaster",
  "n07645769": "Oven Stuffer, Oven Stuffer Roaster",
  "n07645886": "spatchcock",
  "n07646067": "hen",
  "n07646163": "Rock Cornish hen",
  "n07646284": "guinea hen",
  "n07646403": "squab, dove",
  "n07646578": "duck",
  "n07646718": "duckling",
  "n07646821": "goose",
  "n07646927": "wildfowl",
  "n07647115": "grouse",
  "n07647321": "quail",
  "n07647496": "partridge",
  "n07647602": "pheasant",
  "n07647731": "turkey",
  "n07647870": "drumstick",
  "n07648036": "turkey leg, turkey drumstick",
  "n07648150": "chicken leg, chicken drumstick",
  "n07648267": "second joint, thigh",
  "n07648408": "breast, white meat",
  "n07648549": "wing",
  "n07648717": "turkey wing",
  "n07648814": "chicken wing",
  "n07648913": "buffalo wing",
  "n07648997": "barbecued wing",
  "n07649096": "giblet, giblets",
  "n07649202": "medallion",
  "n07649342": "oyster",
  "n07649463": "parson's nose, pope's nose",
  "n07649582": "loaf",
  "n07649854": "meat",
  "n07650449": "game",
  "n07650637": "dark meat",
  "n07650792": "mess",
  "n07650903": "mince",
  "n07651025": "puree",
  "n07651159": "raw meat",
  "n07651245": "gobbet",
  "n07651325": "red meat",
  "n07651454": "variety meat, organs",
  "n07651774": "offal",
  "n07651905": "heart",
  "n07652052": "liver",
  "n07652207": "calves' liver, calf's liver",
  "n07652310": "chicken liver",
  "n07652401": "goose liver",
  "n07652488": "sweetbread, sweetbreads",
  "n07652585": "brain",
  "n07652696": "calf's brain",
  "n07652788": "stomach sweetbread",
  "n07652881": "neck sweetbread, throat sweetbread",
  "n07652995": "tongue",
  "n07653126": "beef tongue",
  "n07653217": "calf's tongue",
  "n07653311": "venison",
  "n07653394": "cut, cut of meat",
  "n07653982": "chop",
  "n07654148": "barbecue, barbeque",
  "n07654298": "biryani, biriani",
  "n07654438": "cold cuts",
  "n07654538": "chine",
  "n07654667": "piece, slice",
  "n07654886": "cutlet, scallop, scollop, escallop",
  "n07655067": "escalope de veau Orloff",
  "n07655263": "saute",
  "n07655337": "fillet, filet, fish fillet, fish filet",
  "n07655505": "leg",
  "n07655626": "side, side of meat",
  "n07655790": "side of beef",
  "n07655898": "forequarter",
  "n07655988": "hindquarter",
  "n07656077": "cut of beef",
  "n07656452": "chuck",
  "n07656645": "chuck short ribs",
  "n07656741": "rib",
  "n07656866": "entrecote",
  "n07656961": "sparerib",
  "n07657068": "shank",
  "n07657233": "foreshank",
  "n07657337": "hindshank",
  "n07657440": "shin, shin bone",
  "n07657547": "brisket",
  "n07657664": "plate",
  "n07657757": "flank",
  "n07657903": "steak",
  "n07658058": "fish steak",
  "n07658168": "beefsteak",
  "n07658461": "flank steak",
  "n07658580": "minute steak",
  "n07658678": "loin",
  "n07658814": "beef loin",
  "n07658958": "sirloin",
  "n07659196": "wedge bone",
  "n07659308": "flat bone",
  "n07659425": "pin bone",
  "n07659560": "sirloin tip",
  "n07659684": "sirloin steak",
  "n07659793": "tenderloin, undercut",
  "n07659968": "beef tenderloin",
  "n07660065": "fillet, filet",
  "n07660283": "pork tenderloin",
  "n07660361": "Chateaubriand",
  "n07660463": "Delmonico steak, club steak",
  "n07660590": "tournedos",
  "n07660686": "filet mignon",
  "n07660800": "porterhouse, porterhouse steak",
  "n07661002": "T-bone steak",
  "n07661165": "blade",
  "n07661273": "blade roast",
  "n07661357": "neck",
  "n07661486": "beef neck",
  "n07661583": "shoulder",
  "n07661711": "pot roast",
  "n07661845": "short ribs",
  "n07661950": "rib roast, standing rib roast",
  "n07662156": "round",
  "n07662275": "round steak",
  "n07662406": "top round",
  "n07662517": "bottom round",
  "n07662618": "rump steak",
  "n07662719": "strip steak, New York strip",
  "n07662833": "rump",
  "n07662978": "rump roast",
  "n07663118": "aitchbone",
  "n07663220": "tripe",
  "n07663360": "honeycomb tripe",
  "n07663492": "buffalo",
  "n07663592": "beef, boeuf",
  "n07663796": "beef roast, roast beef",
  "n07663899": "patty, cake",
  "n07664007": "ground beef, hamburger",
  "n07664121": "chopped steak, chop steak, chopsteak, hamburger steak, beef patty",
  "n07664266": "bully beef, corned beef, corn beef",
  "n07664379": "pastrami",
  "n07664468": "carbonado",
  "n07664582": "halal",
  "n07664770": "jerky, jerked meat, jerk",
  "n07664936": "beef jerky",
  "n07665013": "biltong",
  "n07665146": "pemmican, pemican",
  "n07665308": "veal, veau",
  "n07665438": "veal parmesan, veal parmigiana",
  "n07665595": "cut of veal",
  "n07665713": "scrag, scrag end",
  "n07665844": "veal roast, roast veal",
  "n07665983": "breast of veal",
  "n07666071": "fricandeau",
  "n07666176": "veal cordon bleu",
  "n07666303": "calves' feet",
  "n07666406": "horsemeat, horseflesh",
  "n07666521": "rabbit, hare",
  "n07666733": "mouton, mutton",
  "n07666847": "mutton chop",
  "n07666933": "scrag",
  "n07667042": "cut of mutton",
  "n07667151": "lamb",
  "n07667326": "cut of lamb",
  "n07667480": "breast of lamb, poitrine d'agneau",
  "n07667636": "saddle",
  "n07667778": "saddle of lamb",
  "n07667872": "loin of lamb",
  "n07667974": "lamb chop, lamb-chop, lambchop",
  "n07668073": "rack",
  "n07668215": "lamb roast, roast lamb",
  "n07668356": "rack of lamb, crown roast",
  "n07668481": "ham hock",
  "n07668588": "leg of lamb, gigot",
  "n07668702": "pork, porc",
  "n07668902": "cut of pork",
  "n07669098": "cochon de lait, suckling pig",
  "n07669211": "flitch, side of bacon",
  "n07669367": "gammon",
  "n07669470": "pork loin",
  "n07669588": "side of pork",
  "n07669713": "pork belly",
  "n07669788": "pork roast, roast pork",
  "n07669891": "ham, jambon, gammon",
  "n07670044": "Virginia ham",
  "n07670147": "picnic ham, picnic shoulder",
  "n07670254": "porkchop",
  "n07670328": "prosciutto",
  "n07670433": "bacon",
  "n07670657": "bacon strip",
  "n07670731": "rind",
  "n07670902": "bacon rind",
  "n07670995": "Canadian bacon",
  "n07671107": "salt pork",
  "n07671269": "fatback",
  "n07671363": "sowbelly",
  "n07671459": "spareribs",
  "n07671581": "pigs' feet, pigs' knuckles",
  "n07671722": "chitterlings, chitlins, chitlings",
  "n07671845": "cracklings",
  "n07671953": "haslet",
  "n07672135": "edible fat",
  "n07672421": "lard",
  "n07672583": "marbling",
  "n07672687": "shortening",
  "n07672790": "suet",
  "n07672914": "margarine, margarin, oleo, oleomargarine, marge",
  "n07673145": "cooking oil",
  "n07673249": "drippings",
  "n07673397": "vegetable oil, oil",
  "n07673872": "sweet oil",
  "n07674001": "canola oil, canola",
  "n07674161": "coconut oil, copra oil",
  "n07674267": "corn oil",
  "n07674393": "cottonseed oil",
  "n07674508": "olive oil",
  "n07674617": "palm oil",
  "n07674749": "peanut oil, groundnut oil",
  "n07674912": "salad oil",
  "n07675040": "safflower oil",
  "n07675156": "sesame oil",
  "n07675262": "soybean oil, soyabean oil",
  "n07675411": "sunflower oil, sunflower-seed oil",
  "n07675535": "walnut oil",
  "n07675627": "sausage",
  "n07675962": "sausage meat",
  "n07676121": "blood sausage, blood pudding, black pudding",
  "n07676273": "bologna, Bologna sausage",
  "n07676425": "chipolata",
  "n07676520": "chorizo",
  "n07676602": "frank, frankfurter, hotdog, hot dog, dog, wiener, wienerwurst, weenie",
  "n07676855": "Vienna sausage",
  "n07676967": "polony",
  "n07677071": "headcheese",
  "n07677255": "knackwurst, knockwurst",
  "n07677360": "liver pudding, liver sausage, liverwurst",
  "n07677480": "pepperoni",
  "n07677593": "pork sausage",
  "n07677747": "salami",
  "n07677860": "banger",
  "n07677982": "bratwurst, brat",
  "n07678065": "linguica",
  "n07678193": "saveloy",
  "n07678313": "souse",
  "n07678428": "lunch meat, luncheon meat",
  "n07678586": "mincemeat",
  "n07678729": "stuffing, dressing",
  "n07678953": "turkey stuffing",
  "n07679034": "oyster stuffing, oyster dressing",
  "n07679140": "forcemeat, farce",
  "n07679356": "bread, breadstuff, staff of life",
  "n07680168": "anadama bread",
  "n07680313": "bap",
  "n07680416": "barmbrack",
  "n07680517": "breadstick, bread-stick",
  "n07680655": "grissino",
  "n07680761": "brown bread, Boston brown bread",
  "n07680932": "bun, roll",
  "n07681264": "tea bread",
  "n07681355": "caraway seed bread",
  "n07681450": "challah, hallah",
  "n07681691": "cinnamon bread",
  "n07681805": "cracked-wheat bread",
  "n07681926": "cracker",
  "n07682197": "crouton",
  "n07682316": "dark bread, whole wheat bread, whole meal bread, brown bread",
  "n07682477": "English muffin",
  "n07682624": "flatbread",
  "n07682808": "garlic bread",
  "n07682952": "gluten bread",
  "n07683039": "graham bread",
  "n07683138": "Host",
  "n07683265": "flatbrod",
  "n07683360": "bannock",
  "n07683490": "chapatti, chapati",
  "n07683617": "pita, pocket bread",
  "n07683786": "loaf of bread, loaf",
  "n07683973": "heel",
  "n07684084": "French loaf",
  "n07684164": "matzo, matzoh, matzah, unleavened bread",
  "n07684289": "nan, naan",
  "n07684422": "onion bread",
  "n07684517": "raisin bread",
  "n07684600": "quick bread",
  "n07684938": "banana bread",
  "n07685031": "date bread",
  "n07685118": "date-nut bread",
  "n07685218": "nut bread",
  "n07685303": "oatcake",
  "n07685399": "Irish soda bread",
  "n07685546": "skillet bread, fry bread",
  "n07685730": "rye bread",
  "n07685918": "black bread, pumpernickel",
  "n07686021": "Jewish rye bread, Jewish rye",
  "n07686202": "limpa",
  "n07686299": "Swedish rye bread, Swedish rye",
  "n07686461": "salt-rising bread",
  "n07686634": "simnel",
  "n07686720": "sour bread, sourdough bread",
  "n07686873": "toast",
  "n07687053": "wafer",
  "n07687211": "white bread, light bread",
  "n07687381": "baguet, baguette",
  "n07687469": "French bread",
  "n07687626": "Italian bread",
  "n07687789": "cornbread",
  "n07688021": "corn cake",
  "n07688130": "skillet corn bread",
  "n07688265": "ashcake, ash cake, corn tash",
  "n07688412": "hoecake",
  "n07688624": "cornpone, pone",
  "n07688757": "corn dab, corn dodger, dodger",
  "n07688898": "hush puppy, hushpuppy",
  "n07689003": "johnnycake, johnny cake, journey cake",
  "n07689217": "Shawnee cake",
  "n07689313": "spoon bread, batter bread",
  "n07689490": "cinnamon toast",
  "n07689624": "orange toast",
  "n07689757": "Melba toast",
  "n07689842": "zwieback, rusk, Brussels biscuit, twice-baked bread",
  "n07690019": "frankfurter bun, hotdog bun",
  "n07690152": "hamburger bun, hamburger roll",
  "n07690273": "muffin, gem",
  "n07690431": "bran muffin",
  "n07690511": "corn muffin",
  "n07690585": "Yorkshire pudding",
  "n07690739": "popover",
  "n07690892": "scone",
  "n07691091": "drop scone, griddlecake, Scotch pancake",
  "n07691237": "cross bun, hot cross bun",
  "n07691424": "coffee ring",
  "n07691539": "brioche",
  "n07691650": "crescent roll, croissant",
  "n07691758": "hard roll, Vienna roll",
  "n07691863": "soft roll",
  "n07691954": "kaiser roll",
  "n07692114": "Parker House roll",
  "n07692248": "clover-leaf roll",
  "n07692405": "onion roll",
  "n07692517": "bialy, bialystoker",
  "n07692614": "sweet roll, coffee roll",
  "n07692887": "bear claw, bear paw",
  "n07693048": "cinnamon roll, cinnamon bun, cinnamon snail",
  "n07693223": "honey bun, sticky bun, caramel bun, schnecken",
  "n07693439": "pinwheel roll",
  "n07693590": "danish, danish pastry",
  "n07693725": "bagel, beigel",
  "n07693889": "onion bagel",
  "n07693972": "biscuit",
  "n07694169": "rolled biscuit",
  "n07694268": "drop biscuit",
  "n07694403": "baking-powder biscuit",
  "n07694516": "buttermilk biscuit, soda biscuit",
  "n07694659": "shortcake",
  "n07694839": "hardtack, pilot biscuit, pilot bread, sea biscuit, ship biscuit",
  "n07695012": "wafer",
  "n07695097": "brandysnap",
  "n07695187": "saltine",
  "n07695284": "soda cracker",
  "n07695410": "oyster cracker",
  "n07695504": "water biscuit",
  "n07695652": "graham cracker",
  "n07695742": "pretzel",
  "n07695878": "soft pretzel",
  "n07695965": "sandwich",
  "n07696403": "sandwich plate",
  "n07696527": "butty",
  "n07696625": "ham sandwich",
  "n07696728": "chicken sandwich",
  "n07696839": "club sandwich, three-decker, triple-decker",
  "n07696977": "open-face sandwich, open sandwich",
  "n07697100": "hamburger, beefburger, burger",
  "n07697313": "cheeseburger",
  "n07697408": "tunaburger",
  "n07697537": "hotdog, hot dog, red hot",
  "n07697699": "Sloppy Joe",
  "n07697825": "bomber, grinder, hero, hero sandwich, hoagie, hoagy, Cuban sandwich, Italian sandwich, poor boy, sub, submarine, submarine sandwich, torpedo, wedge, zep",
  "n07698250": "gyro",
  "n07698401": "bacon-lettuce-tomato sandwich, BLT",
  "n07698543": "Reuben",
  "n07698672": "western, western sandwich",
  "n07698782": "wrap",
  "n07698915": "pasta, alimentary paste",
  "n07699430": "farfalle, bowtie pasta",
  "n07699584": "noodle",
  "n07699682": "orzo",
  "n07699815": "egg noodle",
  "n07699914": "spaghetti",
  "n07700003": "spaghetti",
  "n07700095": "spaghettini",
  "n07700167": "tortellini",
  "n07700255": "ziti",
  "n07700348": "rigatoni",
  "n07700439": "fedelline",
  "n07700539": "linguine, linguini",
  "n07700638": "fettuccine, fettuccini",
  "n07700766": "fettuccine Alfredo",
  "n07700868": "vermicelli",
  "n07700964": "macaroni",
  "n07701053": "lasagna, lasagne",
  "n07701147": "penne",
  "n07701244": "ravioli, cappelletti",
  "n07701372": "tagliatelle",
  "n07701457": "manicotti",
  "n07701597": "couscous",
  "n07701715": "gnocchi",
  "n07701912": "matzo ball, matzoh ball, matzah ball",
  "n07702057": "won ton, wonton",
  "n07702193": "dumpling, dumplings",
  "n07702362": "health food",
  "n07702509": "junk food",
  "n07702642": "breakfast food",
  "n07702796": "cereal",
  "n07702957": "muesli",
  "n07703053": "Pablum",
  "n07703177": "hot cereal",
  "n07703333": "mush, cornmeal mush",
  "n07703515": "atole",
  "n07703599": "hasty pudding",
  "n07703743": "polenta",
  "n07703889": "hasty pudding",
  "n07704054": "gruel",
  "n07704205": "congee, jook",
  "n07704305": "skilly",
  "n07704428": "grits, hominy grits",
  "n07704579": "kasha",
  "n07704656": "frumenty",
  "n07704755": "cold cereal, dry cereal",
  "n07704994": "granola",
  "n07705137": "granola bar",
  "n07705221": "raisin bran",
  "n07705303": "corn flake",
  "n07705386": "bran flake",
  "n07705473": "wheatflake",
  "n07705557": "puffed rice",
  "n07705633": "puffed wheat",
  "n07705711": "produce, green goods, green groceries, garden truck",
  "n07705931": "edible fruit",
  "n07707451": "vegetable, veggie, veg",
  "n07708124": "julienne, julienne vegetable",
  "n07708260": "eater",
  "n07708398": "raw vegetable, rabbit food",
  "n07708512": "crudites",
  "n07708685": "celery stick",
  "n07708798": "legume",
  "n07709046": "pulse",
  "n07709172": "potherb",
  "n07709333": "greens, green, leafy vegetable",
  "n07709701": "chop-suey greens",
  "n07709881": "bean curd, tofu",
  "n07710007": "solanaceous vegetable",
  "n07710283": "root vegetable",
  "n07710616": "potato, white potato, Irish potato, murphy, spud, tater",
  "n07710952": "baked potato",
  "n07711080": "french fries, french-fried potatoes, fries, chips",
  "n07711232": "home fries, home-fried potatoes",
  "n07711371": "jacket potato",
  "n07711471": "jacket",
  "n07711569": "mashed potato",
  "n07711683": "potato skin, potato peel, potato peelings",
  "n07711799": "Uruguay potato",
  "n07711907": "yam",
  "n07712063": "sweet potato",
  "n07712267": "yam",
  "n07712382": "snack food",
  "n07712559": "chip, crisp, potato chip, Saratoga chip",
  "n07712748": "corn chip",
  "n07712856": "tortilla chip",
  "n07712959": "nacho",
  "n07713074": "eggplant, aubergine, mad apple",
  "n07713267": "pieplant, rhubarb",
  "n07713395": "cruciferous vegetable",
  "n07713763": "mustard, mustard greens, leaf mustard, Indian mustard",
  "n07713895": "cabbage, chou",
  "n07714078": "kale, kail, cole",
  "n07714188": "collards, collard greens",
  "n07714287": "Chinese cabbage, celery cabbage, Chinese celery",
  "n07714448": "bok choy, bok choi",
  "n07714571": "head cabbage",
  "n07714802": "red cabbage",
  "n07714895": "savoy cabbage, savoy",
  "n07714990": "broccoli",
  "n07715103": "cauliflower",
  "n07715221": "brussels sprouts",
  "n07715407": "broccoli rabe, broccoli raab",
  "n07715561": "squash",
  "n07715721": "summer squash",
  "n07716034": "yellow squash",
  "n07716203": "crookneck, crookneck squash, summer crookneck",
  "n07716358": "zucchini, courgette",
  "n07716504": "marrow, vegetable marrow",
  "n07716649": "cocozelle",
  "n07716750": "pattypan squash",
  "n07716906": "spaghetti squash",
  "n07717070": "winter squash",
  "n07717410": "acorn squash",
  "n07717556": "butternut squash",
  "n07717714": "hubbard squash",
  "n07717858": "turban squash",
  "n07718068": "buttercup squash",
  "n07718195": "cushaw",
  "n07718329": "winter crookneck squash",
  "n07718472": "cucumber, cuke",
  "n07718671": "gherkin",
  "n07718747": "artichoke, globe artichoke",
  "n07718920": "artichoke heart",
  "n07719058": "Jerusalem artichoke, sunchoke",
  "n07719213": "asparagus",
  "n07719330": "bamboo shoot",
  "n07719437": "sprout",
  "n07719616": "bean sprout",
  "n07719756": "alfalfa sprout",
  "n07719839": "beet, beetroot",
  "n07719980": "beet green",
  "n07720084": "sugar beet",
  "n07720185": "mangel-wurzel",
  "n07720277": "chard, Swiss chard, spinach beet, leaf beet",
  "n07720442": "pepper",
  "n07720615": "sweet pepper",
  "n07720875": "bell pepper",
  "n07721018": "green pepper",
  "n07721118": "globe pepper",
  "n07721195": "pimento, pimiento",
  "n07721325": "hot pepper",
  "n07721456": "chili, chili pepper, chilli, chilly, chile",
  "n07721678": "jalapeno, jalapeno pepper",
  "n07721833": "chipotle",
  "n07721942": "cayenne, cayenne pepper",
  "n07722052": "tabasco, red pepper",
  "n07722217": "onion",
  "n07722390": "Bermuda onion",
  "n07722485": "green onion, spring onion, scallion",
  "n07722666": "Vidalia onion",
  "n07722763": "Spanish onion",
  "n07722888": "purple onion, red onion",
  "n07723039": "leek",
  "n07723177": "shallot",
  "n07723330": "salad green, salad greens",
  "n07723559": "lettuce",
  "n07723753": "butterhead lettuce",
  "n07723968": "buttercrunch",
  "n07724078": "Bibb lettuce",
  "n07724173": "Boston lettuce",
  "n07724269": "crisphead lettuce, iceberg lettuce, iceberg",
  "n07724492": "cos, cos lettuce, romaine, romaine lettuce",
  "n07724654": "leaf lettuce, loose-leaf lettuce",
  "n07724819": "celtuce",
  "n07724943": "bean, edible bean",
  "n07725158": "goa bean",
  "n07725255": "lentil",
  "n07725376": "pea",
  "n07725531": "green pea, garden pea",
  "n07725663": "marrowfat pea",
  "n07725789": "snow pea, sugar pea",
  "n07725888": "sugar snap pea",
  "n07726009": "split-pea",
  "n07726095": "chickpea, garbanzo",
  "n07726230": "cajan pea, pigeon pea, dahl",
  "n07726386": "field pea",
  "n07726525": "mushy peas",
  "n07726672": "black-eyed pea, cowpea",
  "n07726796": "common bean",
  "n07727048": "kidney bean",
  "n07727140": "navy bean, pea bean, white bean",
  "n07727252": "pinto bean",
  "n07727377": "frijole",
  "n07727458": "black bean, turtle bean",
  "n07727578": "fresh bean",
  "n07727741": "flageolet, haricot",
  "n07727868": "green bean",
  "n07728053": "snap bean, snap",
  "n07728181": "string bean",
  "n07728284": "Kentucky wonder, Kentucky wonder bean",
  "n07728391": "scarlet runner, scarlet runner bean, runner bean, English runner bean",
  "n07728585": "haricot vert, haricots verts, French bean",
  "n07728708": "wax bean, yellow bean",
  "n07728804": "shell bean",
  "n07729000": "lima bean",
  "n07729142": "Fordhooks",
  "n07729225": "sieva bean, butter bean, butterbean, civet bean",
  "n07729384": "fava bean, broad bean",
  "n07729485": "soy, soybean, soya, soya bean",
  "n07729828": "green soybean",
  "n07729926": "field soybean",
  "n07730033": "cardoon",
  "n07730207": "carrot",
  "n07730320": "carrot stick",
  "n07730406": "celery",
  "n07730562": "pascal celery, Paschal celery",
  "n07730708": "celeriac, celery root",
  "n07730855": "chicory, curly endive",
  "n07731006": "radicchio",
  "n07731122": "coffee substitute",
  "n07731284": "chicory, chicory root",
  "n07731436": "Postum",
  "n07731587": "chicory escarole, endive, escarole",
  "n07731767": "Belgian endive, French endive, witloof",
  "n07731952": "corn, edible corn",
  "n07732168": "sweet corn, green corn",
  "n07732302": "hominy",
  "n07732433": "lye hominy",
  "n07732525": "pearl hominy",
  "n07732636": "popcorn",
  "n07732747": "cress",
  "n07732904": "watercress",
  "n07733005": "garden cress",
  "n07733124": "winter cress",
  "n07733217": "dandelion green",
  "n07733394": "gumbo, okra",
  "n07733567": "kohlrabi, turnip cabbage",
  "n07733712": "lamb's-quarter, pigweed, wild spinach",
  "n07733847": "wild spinach",
  "n07734017": "tomato",
  "n07734183": "beefsteak tomato",
  "n07734292": "cherry tomato",
  "n07734417": "plum tomato",
  "n07734555": "tomatillo, husk tomato, Mexican husk tomato",
  "n07734744": "mushroom",
  "n07734879": "stuffed mushroom",
  "n07735052": "salsify",
  "n07735179": "oyster plant, vegetable oyster",
  "n07735294": "scorzonera, black salsify",
  "n07735404": "parsnip",
  "n07735510": "pumpkin",
  "n07735687": "radish",
  "n07735803": "turnip",
  "n07735981": "white turnip",
  "n07736087": "rutabaga, swede, swedish turnip, yellow turnip",
  "n07736256": "turnip greens",
  "n07736371": "sorrel, common sorrel",
  "n07736527": "French sorrel",
  "n07736692": "spinach",
  "n07736813": "taro, taro root, cocoyam, dasheen, edda",
  "n07736971": "truffle, earthnut",
  "n07737081": "edible nut",
  "n07737594": "bunya bunya",
  "n07737745": "peanut, earthnut, goober, goober pea, groundnut, monkey nut",
  "n07737980": "water chestnut",
  "n07738105": "freestone",
  "n07738224": "cling, clingstone",
  "n07738353": "peel, skin",
  "n07738570": "banana peel, banana skin",
  "n07738760": "lemon peel, lemon rind",
  "n07738905": "orange peel, orange rind",
  "n07739035": "windfall",
  "n07739125": "apple",
  "n07739344": "crab apple, crabapple",
  "n07739506": "eating apple, dessert apple",
  "n07739923": "Baldwin",
  "n07740033": "Cortland",
  "n07740115": "Cox's Orange Pippin",
  "n07740220": "Delicious",
  "n07740342": "Golden Delicious, Yellow Delicious",
  "n07740461": "Red Delicious",
  "n07740597": "Empire",
  "n07740744": "Grimes' golden",
  "n07740855": "Jonathan",
  "n07740954": "McIntosh",
  "n07741138": "Macoun",
  "n07741235": "Northern Spy",
  "n07741357": "Pearmain",
  "n07741461": "Pippin",
  "n07741623": "Prima",
  "n07741706": "Stayman",
  "n07741804": "Winesap",
  "n07741888": "Stayman Winesap",
  "n07742012": "cooking apple",
  "n07742224": "Bramley's Seedling",
  "n07742313": "Granny Smith",
  "n07742415": "Lane's Prince Albert",
  "n07742513": "Newtown Wonder",
  "n07742605": "Rome Beauty",
  "n07742704": "berry",
  "n07743224": "bilberry, whortleberry, European blueberry",
  "n07743384": "huckleberry",
  "n07743544": "blueberry",
  "n07743723": "wintergreen, boxberry, checkerberry, teaberry, spiceberry",
  "n07743902": "cranberry",
  "n07744057": "lingonberry, mountain cranberry, cowberry, lowbush cranberry",
  "n07744246": "currant",
  "n07744430": "gooseberry",
  "n07744559": "black currant",
  "n07744682": "red currant",
  "n07744811": "blackberry",
  "n07745046": "boysenberry",
  "n07745197": "dewberry",
  "n07745357": "loganberry",
  "n07745466": "raspberry",
  "n07745661": "saskatoon, serviceberry, shadberry, juneberry",
  "n07745803": "lanseh, lansa, lansat, lanset",
  "n07745940": "strawberry",
  "n07746038": "sugarberry, hackberry",
  "n07746186": "persimmon",
  "n07746334": "acerola, barbados cherry, surinam cherry, West Indian cherry",
  "n07746551": "carambola, star fruit",
  "n07746749": "ceriman, monstera",
  "n07746910": "carissa plum, natal plum",
  "n07747055": "citrus, citrus fruit, citrous fruit",
  "n07747455": "section",
  "n07747607": "orange",
  "n07747811": "temple orange",
  "n07747951": "mandarin, mandarin orange",
  "n07748157": "clementine",
  "n07748276": "satsuma",
  "n07748416": "tangerine",
  "n07748574": "tangelo, ugli, ugli fruit",
  "n07748753": "bitter orange, Seville orange, sour orange",
  "n07748912": "sweet orange",
  "n07749095": "Jaffa orange",
  "n07749192": "navel orange",
  "n07749312": "Valencia orange",
  "n07749446": "kumquat",
  "n07749582": "lemon",
  "n07749731": "lime",
  "n07749870": "key lime",
  "n07749969": "grapefruit",
  "n07750146": "pomelo, shaddock",
  "n07750299": "citrange",
  "n07750449": "citron",
  "n07750586": "almond",
  "n07750736": "Jordan almond",
  "n07750872": "apricot",
  "n07751004": "peach",
  "n07751148": "nectarine",
  "n07751280": "pitahaya",
  "n07751451": "plum",
  "n07751737": "damson, damson plum",
  "n07751858": "greengage, greengage plum",
  "n07751977": "beach plum",
  "n07752109": "sloe",
  "n07752264": "Victoria plum",
  "n07752377": "dried fruit",
  "n07752514": "dried apricot",
  "n07752602": "prune",
  "n07752664": "raisin",
  "n07752782": "seedless raisin, sultana",
  "n07752874": "seeded raisin",
  "n07752966": "currant",
  "n07753113": "fig",
  "n07753275": "pineapple, ananas",
  "n07753448": "anchovy pear, river pear",
  "n07753592": "banana",
  "n07753743": "passion fruit",
  "n07753980": "granadilla",
  "n07754155": "sweet calabash",
  "n07754279": "bell apple, sweet cup, water lemon, yellow granadilla",
  "n07754451": "breadfruit",
  "n07754684": "jackfruit, jak, jack",
  "n07754894": "cacao bean, cocoa bean",
  "n07755089": "cocoa",
  "n07755262": "canistel, eggfruit",
  "n07755411": "melon",
  "n07755619": "melon ball",
  "n07755707": "muskmelon, sweet melon",
  "n07755929": "cantaloup, cantaloupe",
  "n07756096": "winter melon",
  "n07756325": "honeydew, honeydew melon",
  "n07756499": "Persian melon",
  "n07756641": "net melon, netted melon, nutmeg melon",
  "n07756838": "casaba, casaba melon",
  "n07756951": "watermelon",
  "n07757132": "cherry",
  "n07757312": "sweet cherry, black cherry",
  "n07757511": "bing cherry",
  "n07757602": "heart cherry, oxheart, oxheart cherry",
  "n07757753": "blackheart, blackheart cherry",
  "n07757874": "capulin, Mexican black cherry",
  "n07757990": "sour cherry",
  "n07758125": "amarelle",
  "n07758260": "morello",
  "n07758407": "cocoa plum, coco plum, icaco",
  "n07758582": "gherkin",
  "n07758680": "grape",
  "n07758950": "fox grape",
  "n07759194": "Concord grape",
  "n07759324": "Catawba",
  "n07759424": "muscadine, bullace grape",
  "n07759576": "scuppernong",
  "n07759691": "slipskin grape",
  "n07759816": "vinifera grape",
  "n07760070": "emperor",
  "n07760153": "muscat, muscatel, muscat grape",
  "n07760297": "ribier",
  "n07760395": "sultana",
  "n07760501": "Tokay",
  "n07760673": "flame tokay",
  "n07760755": "Thompson Seedless",
  "n07760859": "custard apple",
  "n07761141": "cherimoya, cherimolla",
  "n07761309": "soursop, guanabana",
  "n07761461": "bullock's heart, Jamaica apple",
  "n07761611": "sweetsop, annon, sugar apple",
  "n07761777": "ilama",
  "n07761954": "pond apple",
  "n07762114": "papaw, pawpaw",
  "n07762244": "papaya",
  "n07762373": "kai apple",
  "n07762534": "ketembilla, kitembilla, kitambilla",
  "n07762740": "ackee, akee",
  "n07762913": "durian",
  "n07763107": "feijoa, pineapple guava",
  "n07763290": "genip, Spanish lime",
  "n07763483": "genipap, genipap fruit",
  "n07763629": "kiwi, kiwi fruit, Chinese gooseberry",
  "n07763792": "loquat, Japanese plum",
  "n07763987": "mangosteen",
  "n07764155": "mango",
  "n07764315": "sapodilla, sapodilla plum, sapota",
  "n07764486": "sapote, mammee, marmalade plum",
  "n07764630": "tamarind, tamarindo",
  "n07764847": "avocado, alligator pear, avocado pear, aguacate",
  "n07765073": "date",
  "n07765208": "elderberry",
  "n07765361": "guava",
  "n07765517": "mombin",
  "n07765612": "hog plum, yellow mombin",
  "n07765728": "hog plum, wild plum",
  "n07765862": "jaboticaba",
  "n07765999": "jujube, Chinese date, Chinese jujube",
  "n07766173": "litchi, litchi nut, litchee, lichi, leechee, lichee, lychee",
  "n07766409": "longanberry, dragon's eye",
  "n07766530": "mamey, mammee, mammee apple",
  "n07766723": "marang",
  "n07766891": "medlar",
  "n07767002": "medlar",
  "n07767171": "mulberry",
  "n07767344": "olive",
  "n07767549": "black olive, ripe olive",
  "n07767709": "green olive",
  "n07767847": "pear",
  "n07768068": "bosc",
  "n07768139": "anjou",
  "n07768230": "bartlett, bartlett pear",
  "n07768318": "seckel, seckel pear",
  "n07768423": "plantain",
  "n07768590": "plumcot",
  "n07768694": "pomegranate",
  "n07768858": "prickly pear",
  "n07769005": "garambulla",
  "n07769102": "Barbados gooseberry, blade apple",
  "n07769306": "quandong, quandang, quantong, native peach",
  "n07769465": "quandong nut",
  "n07769584": "quince",
  "n07769731": "rambutan, rambotan",
  "n07769886": "pulasan, pulassan",
  "n07770034": "rose apple",
  "n07770180": "sorb, sorb apple",
  "n07770289": "sour gourd",
  "n07770439": "sour gourd, monkey bread",
  "n07770571": "edible seed",
  "n07770763": "pumpkin seed",
  "n07770869": "betel nut, areca nut",
  "n07771082": "beechnut",
  "n07771212": "walnut",
  "n07771405": "black walnut",
  "n07771539": "English walnut",
  "n07771731": "brazil nut, brazil",
  "n07771891": "butternut",
  "n07772026": "souari nut",
  "n07772147": "cashew, cashew nut",
  "n07772274": "chestnut",
  "n07772413": "chincapin, chinkapin, chinquapin",
  "n07772631": "water chinquapin",
  "n07772788": "hazelnut, filbert, cobnut, cob",
  "n07772935": "coconut, cocoanut",
  "n07773238": "coconut, coconut meat",
  "n07773428": "coconut milk, coconut water",
  "n07773572": "copra",
  "n07773700": "dika nut",
  "n07773827": "dika bread",
  "n07774032": "groundnut, potato bean, wild bean",
  "n07774182": "grugru nut",
  "n07774295": "hickory nut",
  "n07774479": "cola extract",
  "n07774596": "macadamia nut",
  "n07774719": "pecan",
  "n07774842": "pine nut, pignolia, pinon nut",
  "n07775050": "pistachio, pistachio nut",
  "n07775197": "sunflower seed",
  "n07775375": "fish",
  "n07775905": "saltwater fish",
  "n07776545": "freshwater fish",
  "n07776866": "seafood",
  "n07777189": "bream, sea bream",
  "n07777358": "bream, freshwater bream",
  "n07777512": "freshwater bass, bass",
  "n07777735": "largemouth bass",
  "n07777840": "smallmouth bass",
  "n07777945": "sea bass, bass",
  "n07778103": "striped bass, striper",
  "n07778224": "grouper",
  "n07778342": "croaker",
  "n07778494": "whiting",
  "n07778680": "whiting",
  "n07778810": "cusk",
  "n07778938": "dolphinfish, mahimahi",
  "n07779123": "carp",
  "n07779263": "buffalofish",
  "n07779375": "pike",
  "n07779535": "muskellunge",
  "n07779664": "pickerel",
  "n07779747": "monkfish",
  "n07779895": "sucker",
  "n07780038": "catfish, mudcat",
  "n07780173": "perch",
  "n07780307": "sunfish",
  "n07780486": "crappie",
  "n07780627": "tuna, tuna fish, tunny",
  "n07780874": "albacore",
  "n07781022": "bonito",
  "n07781207": "bluefin, bluefin tuna",
  "n07781319": "mackerel",
  "n07781497": "Spanish mackerel",
  "n07781689": "pompano",
  "n07781801": "squid, calamari, calamary",
  "n07781972": "blowfish, sea squab, puffer, pufferfish",
  "n07782204": "fugu",
  "n07782365": "octopus",
  "n07782475": "escargot, snail",
  "n07782668": "periwinkle, winkle",
  "n07782804": "whelk",
  "n07782939": "panfish",
  "n07783102": "stockfish",
  "n07783210": "shellfish",
  "n07783550": "mussel",
  "n07783667": "anchovy",
  "n07783827": "anchovy paste",
  "n07783967": "eel",
  "n07784197": "smoked eel",
  "n07784274": "elver",
  "n07784367": "mullet, grey mullet, gray mullet",
  "n07784522": "herring",
  "n07784810": "kingfish",
  "n07784991": "lingcod",
  "n07785138": "kipper, kippered herring",
  "n07785235": "bloater",
  "n07785342": "pickled herring",
  "n07785487": "rollmops",
  "n07785631": "alewife",
  "n07785783": "bluefish",
  "n07785887": "swordfish",
  "n07786005": "butterfish",
  "n07786164": "huitre, oyster",
  "n07786337": "oysters Rockefeller",
  "n07786484": "bluepoint, blue point",
  "n07786686": "clam",
  "n07786856": "quahaug, quahog, hard-shell clam, round clam",
  "n07787108": "littleneck, littleneck clam",
  "n07787270": "cherrystone, cherrystone clam",
  "n07787429": "soft-shell clam, steamer, steamer clam, long-neck clam",
  "n07787613": "cockle",
  "n07787715": "crab, crabmeat",
  "n07787980": "blue crab",
  "n07788106": "crab legs",
  "n07788198": "soft-shell crab, soft-shelled crab",
  "n07788340": "Japanese crab",
  "n07788435": "Alaska king crab, Alaskan king crab, king crab, Alaska crab",
  "n07788609": "Dungeness crab",
  "n07788730": "stone crab",
  "n07788885": "crayfish, crawfish, crawdad, ecrevisse",
  "n07789063": "cod, codfish",
  "n07789240": "pollack, pollock",
  "n07789382": "schrod, scrod",
  "n07789541": "haddock",
  "n07789745": "finnan haddie, finnan haddock, finnan, smoked haddock",
  "n07789911": "salt cod",
  "n07790081": "porgy, scup",
  "n07790246": "scup",
  "n07790400": "flatfish",
  "n07790601": "flounder",
  "n07790800": "yellowtail flounder",
  "n07790935": "plaice",
  "n07791039": "turbot",
  "n07791145": "sand dab",
  "n07791274": "sole, fillet of sole",
  "n07791434": "grey sole, gray sole",
  "n07791535": "lemon sole, English sole",
  "n07791663": "lemon sole, winter flounder",
  "n07791808": "halibut",
  "n07791937": "flitch",
  "n07792027": "hake",
  "n07792117": "redfish, rosefish, ocean perch",
  "n07792219": "rockfish",
  "n07792383": "sailfish",
  "n07792470": "weakfish",
  "n07792640": "limpet",
  "n07792725": "lobster",
  "n07792926": "American lobster, Northern lobster, Maine lobster",
  "n07793133": "European lobster",
  "n07793260": "spiny lobster, langouste, rock lobster, crayfish",
  "n07793521": "Norwegian lobster, langoustine, scampo",
  "n07793685": "lobster tail",
  "n07793795": "coral",
  "n07793946": "tomalley",
  "n07794063": "sardine, pilchard",
  "n07794159": "prawn, shrimp",
  "n07794353": "river prawn",
  "n07794452": "trout",
  "n07794605": "rainbow trout",
  "n07794744": "sea trout, salmon trout",
  "n07794893": "brook trout, speckled trout",
  "n07795019": "lake trout",
  "n07795133": "whitefish",
  "n07795317": "whitefish",
  "n07795459": "lake herring, cisco",
  "n07795598": "rock salmon",
  "n07795751": "salmon",
  "n07796005": "Atlantic salmon",
  "n07796165": "red salmon, sockeye, sockeye salmon",
  "n07796321": "chinook salmon, chinook, king salmon",
  "n07796468": "silver salmon, coho salmon, coho, cohoe",
  "n07796649": "smoked salmon",
  "n07796750": "lox",
  "n07796877": "Scandinavian lox",
  "n07796970": "Nova Scotia lox, Nova lox, Nova Scotia salmon, Nova salmon, Nova style salmon",
  "n07797189": "snapper",
  "n07797357": "red snapper",
  "n07797518": "red rockfish",
  "n07797641": "scallop, scollop, escallop",
  "n07797913": "sea scallop",
  "n07798025": "bay scallop",
  "n07798147": "kippered salmon",
  "n07798233": "red herring, smoked herring",
  "n07798357": "shad",
  "n07798554": "smelt",
  "n07798728": "American smelt, rainbow smelt",
  "n07798872": "European smelt, sparling",
  "n07798985": "sprat, brisling",
  "n07799132": "whitebait",
  "n07799278": "roe, hard roe",
  "n07799447": "milt, soft roe",
  "n07799579": "caviar, caviare",
  "n07799738": "beluga caviar",
  "n07799874": "shad roe",
  "n07800004": "smoked mackerel",
  "n07800091": "feed, provender",
  "n07800487": "cattle cake",
  "n07800636": "creep feed",
  "n07800740": "fodder",
  "n07801007": "feed grain",
  "n07801091": "eatage, forage, pasture, pasturage, grass",
  "n07801342": "silage, ensilage",
  "n07801508": "oil cake",
  "n07801709": "oil meal",
  "n07801779": "alfalfa",
  "n07801892": "broad bean, horse bean",
  "n07802026": "hay",
  "n07802152": "timothy",
  "n07802246": "stover",
  "n07802417": "grain, food grain, cereal",
  "n07802767": "grist",
  "n07802863": "groats",
  "n07802963": "millet",
  "n07803093": "barley, barleycorn",
  "n07803213": "pearl barley",
  "n07803310": "buckwheat",
  "n07803408": "bulgur, bulghur, bulgur wheat",
  "n07803545": "wheat, wheat berry",
  "n07803779": "cracked wheat",
  "n07803895": "stodge",
  "n07803992": "wheat germ",
  "n07804152": "oat",
  "n07804323": "rice",
  "n07804543": "brown rice",
  "n07804657": "white rice, polished rice",
  "n07804771": "wild rice, Indian rice",
  "n07804900": "paddy",
  "n07805006": "slop, slops, swill, pigswill, pigwash",
  "n07805254": "mash",
  "n07805389": "chicken feed, scratch",
  "n07805478": "cud, rechewed food",
  "n07805594": "bird feed, bird food, birdseed",
  "n07805731": "petfood, pet-food, pet food",
  "n07805872": "mast",
  "n07805966": "dog food",
  "n07806043": "cat food",
  "n07806120": "canary seed",
  "n07806221": "salad",
  "n07806633": "tossed salad",
  "n07806774": "green salad",
  "n07806879": "Caesar salad",
  "n07807002": "salmagundi",
  "n07807171": "salad nicoise",
  "n07807317": "combination salad",
  "n07807472": "chef's salad",
  "n07807594": "potato salad",
  "n07807710": "pasta salad",
  "n07807834": "macaroni salad",
  "n07807922": "fruit salad",
  "n07808022": "Waldorf salad",
  "n07808166": "crab Louis",
  "n07808268": "herring salad",
  "n07808352": "tuna fish salad, tuna salad",
  "n07808479": "chicken salad",
  "n07808587": "coleslaw, slaw",
  "n07808675": "aspic",
  "n07808806": "molded salad",
  "n07808904": "tabbouleh, tabooli",
  "n07809096": "ingredient, fixings",
  "n07809368": "flavorer, flavourer, flavoring, flavouring, seasoner, seasoning",
  "n07810531": "bouillon cube",
  "n07810633": "beef tea, Bovril",
  "n07810745": "lemon zest",
  "n07810825": "orange zest",
  "n07810907": "condiment",
  "n07811416": "herb",
  "n07812046": "fines herbes",
  "n07812184": "spice",
  "n07812497": "peppermint oil",
  "n07812662": "spearmint oil",
  "n07812790": "lemon oil",
  "n07812913": "wintergreen oil, oil of wintergreen",
  "n07813107": "salt, table salt, common salt",
  "n07813324": "celery salt",
  "n07813409": "garlic salt",
  "n07813495": "onion salt",
  "n07813579": "seasoned salt",
  "n07813717": "sour salt",
  "n07813833": "five spice powder",
  "n07814007": "allspice",
  "n07814203": "cinnamon",
  "n07814390": "stick cinnamon",
  "n07814487": "clove",
  "n07814634": "cumin, cumin seed",
  "n07814790": "fennel",
  "n07814925": "ginger, gingerroot",
  "n07815163": "ginger, powdered ginger",
  "n07815294": "mace",
  "n07815424": "nutmeg",
  "n07815588": "pepper, peppercorn",
  "n07815839": "black pepper",
  "n07815956": "white pepper",
  "n07816052": "sassafras",
  "n07816164": "basil, sweet basil",
  "n07816296": "bay leaf",
  "n07816398": "borage",
  "n07816575": "hyssop",
  "n07816726": "caraway",
  "n07816839": "chervil",
  "n07817024": "chives",
  "n07817160": "comfrey, healing herb",
  "n07817315": "coriander, Chinese parsley, cilantro",
  "n07817465": "coriander, coriander seed",
  "n07817599": "costmary",
  "n07817758": "fennel, common fennel",
  "n07817871": "fennel, Florence fennel, finocchio",
  "n07818029": "fennel seed",
  "n07818133": "fenugreek, fenugreek seed",
  "n07818277": "garlic, ail",
  "n07818422": "clove, garlic clove",
  "n07818572": "garlic chive",
  "n07818689": "lemon balm",
  "n07818825": "lovage",
  "n07818995": "marjoram, oregano",
  "n07819166": "mint",
  "n07819303": "mustard seed",
  "n07819480": "mustard, table mustard",
  "n07819682": "Chinese mustard",
  "n07819769": "nasturtium",
  "n07819896": "parsley",
  "n07820036": "salad burnet",
  "n07820145": "rosemary",
  "n07820297": "rue",
  "n07820497": "sage",
  "n07820683": "clary sage",
  "n07820814": "savory, savoury",
  "n07820960": "summer savory, summer savoury",
  "n07821107": "winter savory, winter savoury",
  "n07821260": "sweet woodruff, waldmeister",
  "n07821404": "sweet cicely",
  "n07821610": "tarragon, estragon",
  "n07821758": "thyme",
  "n07821919": "turmeric",
  "n07822053": "caper",
  "n07822197": "catsup, ketchup, cetchup, tomato ketchup",
  "n07822323": "cardamom, cardamon, cardamum",
  "n07822518": "cayenne, cayenne pepper, red pepper",
  "n07822687": "chili powder",
  "n07822845": "chili sauce",
  "n07823004": "chili vinegar",
  "n07823105": "chutney, Indian relish",
  "n07823280": "steak sauce",
  "n07823369": "taco sauce",
  "n07823460": "salsa",
  "n07823591": "mint sauce",
  "n07823698": "cranberry sauce",
  "n07823814": "curry powder",
  "n07823951": "curry",
  "n07824191": "lamb curry",
  "n07824268": "duck sauce, hoisin sauce",
  "n07824383": "horseradish",
  "n07824502": "marinade",
  "n07824702": "paprika",
  "n07824863": "Spanish paprika",
  "n07824988": "pickle",
  "n07825194": "dill pickle",
  "n07825312": "chowchow",
  "n07825399": "bread and butter pickle",
  "n07825496": "pickle relish",
  "n07825597": "piccalilli",
  "n07825717": "sweet pickle",
  "n07825850": "applesauce, apple sauce",
  "n07825972": "soy sauce, soy",
  "n07826091": "Tabasco, Tabasco sauce",
  "n07826250": "tomato paste",
  "n07826340": "angelica",
  "n07826453": "angelica",
  "n07826544": "almond extract",
  "n07826653": "anise, aniseed, anise seed",
  "n07826930": "Chinese anise, star anise, star aniseed",
  "n07827130": "juniper berries",
  "n07827284": "saffron",
  "n07827410": "sesame seed, benniseed",
  "n07827554": "caraway seed",
  "n07827750": "poppy seed",
  "n07827896": "dill, dill weed",
  "n07828041": "dill seed",
  "n07828156": "celery seed",
  "n07828275": "lemon extract",
  "n07828378": "monosodium glutamate, MSG",
  "n07828642": "vanilla bean",
  "n07828807": "vanilla, vanilla extract",
  "n07828987": "vinegar, acetum",
  "n07829248": "cider vinegar",
  "n07829331": "wine vinegar",
  "n07829412": "sauce",
  "n07830493": "anchovy sauce",
  "n07830593": "hot sauce",
  "n07830690": "hard sauce",
  "n07830841": "horseradish sauce, sauce Albert",
  "n07830986": "bolognese pasta sauce",
  "n07831146": "carbonara",
  "n07831267": "tomato sauce",
  "n07831450": "tartare sauce, tartar sauce",
  "n07831663": "wine sauce",
  "n07831821": "marchand de vin, mushroom wine sauce",
  "n07831955": "bread sauce",
  "n07832099": "plum sauce",
  "n07832202": "peach sauce",
  "n07832307": "apricot sauce",
  "n07832416": "pesto",
  "n07832592": "ravigote, ravigotte",
  "n07832741": "remoulade sauce",
  "n07832902": "dressing, salad dressing",
  "n07833333": "sauce Louis",
  "n07833535": "bleu cheese dressing, blue cheese dressing",
  "n07833672": "blue cheese dressing, Roquefort dressing",
  "n07833816": "French dressing, vinaigrette, sauce vinaigrette",
  "n07833951": "Lorenzo dressing",
  "n07834065": "anchovy dressing",
  "n07834160": "Italian dressing",
  "n07834286": "half-and-half dressing",
  "n07834507": "mayonnaise, mayo",
  "n07834618": "green mayonnaise, sauce verte",
  "n07834774": "aioli, aioli sauce, garlic sauce",
  "n07834872": "Russian dressing, Russian mayonnaise",
  "n07835051": "salad cream",
  "n07835173": "Thousand Island dressing",
  "n07835331": "barbecue sauce",
  "n07835457": "hollandaise",
  "n07835547": "bearnaise",
  "n07835701": "Bercy, Bercy butter",
  "n07835823": "bordelaise",
  "n07835921": "bourguignon, bourguignon sauce, Burgundy sauce",
  "n07836077": "brown sauce, sauce Espagnole",
  "n07836269": "Espagnole, sauce Espagnole",
  "n07836456": "Chinese brown sauce, brown sauce",
  "n07836600": "blanc",
  "n07836731": "cheese sauce",
  "n07836838": "chocolate sauce, chocolate syrup",
  "n07837002": "hot-fudge sauce, fudge sauce",
  "n07837110": "cocktail sauce, seafood sauce",
  "n07837234": "Colbert, Colbert butter",
  "n07837362": "white sauce, bechamel sauce, bechamel",
  "n07837545": "cream sauce",
  "n07837630": "Mornay sauce",
  "n07837755": "demiglace, demi-glaze",
  "n07837912": "gravy, pan gravy",
  "n07838073": "gravy",
  "n07838233": "spaghetti sauce, pasta sauce",
  "n07838441": "marinara",
  "n07838551": "mole",
  "n07838659": "hunter's sauce, sauce chausseur",
  "n07838811": "mushroom sauce",
  "n07838905": "mustard sauce",
  "n07839055": "Nantua, shrimp sauce",
  "n07839172": "Hungarian sauce, paprika sauce",
  "n07839312": "pepper sauce, Poivrade",
  "n07839478": "roux",
  "n07839593": "Smitane",
  "n07839730": "Soubise, white onion sauce",
  "n07839864": "Lyonnaise sauce, brown onion sauce",
  "n07840027": "veloute",
  "n07840124": "allemande, allemande sauce",
  "n07840219": "caper sauce",
  "n07840304": "poulette",
  "n07840395": "curry sauce",
  "n07840520": "Worcester sauce, Worcestershire, Worcestershire sauce",
  "n07840672": "coconut milk, coconut cream",
  "n07840804": "egg, eggs",
  "n07841037": "egg white, white, albumen, ovalbumin",
  "n07841345": "egg yolk, yolk",
  "n07841495": "boiled egg, coddled egg",
  "n07841639": "hard-boiled egg, hard-cooked egg",
  "n07841800": "Easter egg",
  "n07841907": "Easter egg",
  "n07842044": "chocolate egg",
  "n07842130": "candy egg",
  "n07842202": "poached egg, dropped egg",
  "n07842308": "scrambled eggs",
  "n07842433": "deviled egg, stuffed egg",
  "n07842605": "shirred egg, baked egg, egg en cocotte",
  "n07842753": "omelet, omelette",
  "n07842972": "firm omelet",
  "n07843117": "French omelet",
  "n07843220": "fluffy omelet",
  "n07843348": "western omelet",
  "n07843464": "souffle",
  "n07843636": "fried egg",
  "n07843775": "dairy product",
  "n07844042": "milk",
  "n07844604": "milk",
  "n07844786": "sour milk",
  "n07844867": "soya milk, soybean milk, soymilk",
  "n07845087": "formula",
  "n07845166": "pasteurized milk",
  "n07845335": "cows' milk",
  "n07845421": "yak's milk",
  "n07845495": "goats' milk",
  "n07845571": "acidophilus milk",
  "n07845702": "raw milk",
  "n07845775": "scalded milk",
  "n07845863": "homogenized milk",
  "n07846014": "certified milk",
  "n07846143": "powdered milk, dry milk, dried milk, milk powder",
  "n07846274": "nonfat dry milk",
  "n07846359": "evaporated milk",
  "n07846471": "condensed milk",
  "n07846557": "skim milk, skimmed milk",
  "n07846688": "semi-skimmed milk",
  "n07846802": "whole milk",
  "n07846938": "low-fat milk",
  "n07847047": "buttermilk",
  "n07847198": "cream",
  "n07847453": "clotted cream, Devonshire cream",
  "n07847585": "double creme, heavy whipping cream",
  "n07847706": "half-and-half",
  "n07847827": "heavy cream",
  "n07847917": "light cream, coffee cream, single cream",
  "n07848093": "sour cream, soured cream",
  "n07848196": "whipping cream, light whipping cream",
  "n07848338": "butter",
  "n07848645": "stick",
  "n07848771": "clarified butter, drawn butter",
  "n07848936": "ghee",
  "n07849026": "brown butter, beurre noisette",
  "n07849186": "Meuniere butter, lemon butter",
  "n07849336": "yogurt, yoghurt, yoghourt",
  "n07849506": "blueberry yogurt",
  "n07849619": "raita",
  "n07849733": "whey",
  "n07849912": "curd",
  "n07850083": "curd",
  "n07850219": "clabber",
  "n07850329": "cheese",
  "n07850957": "cheese rind",
  "n07851054": "paring",
  "n07851298": "cream cheese",
  "n07851443": "double cream",
  "n07851554": "mascarpone",
  "n07851641": "triple cream, triple creme",
  "n07851767": "cottage cheese, pot cheese, farm cheese, farmer's cheese",
  "n07851926": "process cheese, processed cheese",
  "n07852045": "bleu, blue cheese",
  "n07852229": "Stilton",
  "n07852302": "Roquefort",
  "n07852376": "gorgonzola",
  "n07852452": "Danish blue",
  "n07852532": "Bavarian blue",
  "n07852614": "Brie",
  "n07852712": "brick cheese",
  "n07852833": "Camembert",
  "n07852919": "cheddar, cheddar cheese, Armerican cheddar, American cheese",
  "n07853125": "rat cheese, store cheese",
  "n07853232": "Cheshire cheese",
  "n07853345": "double Gloucester",
  "n07853445": "Edam",
  "n07853560": "goat cheese, chevre",
  "n07853648": "Gouda, Gouda cheese",
  "n07853762": "grated cheese",
  "n07853852": "hand cheese",
  "n07853946": "Liederkranz",
  "n07854066": "Limburger",
  "n07854184": "mozzarella",
  "n07854266": "Muenster",
  "n07854348": "Parmesan",
  "n07854455": "quark cheese, quark",
  "n07854614": "ricotta",
  "n07854707": "string cheese",
  "n07854813": "Swiss cheese",
  "n07854982": "Emmenthal, Emmental, Emmenthaler, Emmentaler",
  "n07855105": "Gruyere",
  "n07855188": "sapsago",
  "n07855317": "Velveeta",
  "n07855413": "nut butter",
  "n07855510": "peanut butter",
  "n07855603": "marshmallow fluff",
  "n07855721": "onion butter",
  "n07855812": "pimento butter",
  "n07855907": "shrimp butter",
  "n07856045": "lobster butter",
  "n07856186": "yak butter",
  "n07856270": "spread, paste",
  "n07856756": "cheese spread",
  "n07856895": "anchovy butter",
  "n07856992": "fishpaste",
  "n07857076": "garlic butter",
  "n07857170": "miso",
  "n07857356": "wasabi",
  "n07857598": "snail butter",
  "n07857731": "hummus, humus, hommos, hoummos, humous",
  "n07857959": "pate",
  "n07858114": "duck pate",
  "n07858197": "foie gras, pate de foie gras",
  "n07858336": "tapenade",
  "n07858484": "tahini",
  "n07858595": "sweetening, sweetener",
  "n07858841": "aspartame",
  "n07858978": "honey",
  "n07859142": "saccharin",
  "n07859284": "sugar, refined sugar",
  "n07859583": "syrup, sirup",
  "n07859796": "sugar syrup",
  "n07859951": "molasses",
  "n07860103": "sorghum, sorghum molasses",
  "n07860208": "treacle, golden syrup",
  "n07860331": "grenadine",
  "n07860447": "maple syrup",
  "n07860548": "corn syrup",
  "n07860629": "miraculous food, manna, manna from heaven",
  "n07860805": "batter",
  "n07860988": "dough",
  "n07861158": "bread dough",
  "n07861247": "pancake batter",
  "n07861334": "fritter batter",
  "n07861421": "sop, sops",
  "n07861557": "coq au vin",
  "n07861681": "chicken provencale",
  "n07861813": "chicken and rice",
  "n07861983": "moo goo gai pan",
  "n07862095": "arroz con pollo",
  "n07862244": "bacon and eggs",
  "n07862348": "barbecued spareribs, spareribs",
  "n07862461": "beef Bourguignonne, boeuf Bourguignonne",
  "n07862611": "beef Wellington, filet de boeuf en croute",
  "n07862770": "bitok",
  "n07862946": "boiled dinner, New England boiled dinner",
  "n07863107": "Boston baked beans",
  "n07863229": "bubble and squeak",
  "n07863374": "pasta",
  "n07863547": "cannelloni",
  "n07863644": "carbonnade flamande, Belgian beef stew",
  "n07863802": "cheese souffle",
  "n07863935": "chicken Marengo",
  "n07864065": "chicken cordon bleu",
  "n07864198": "Maryland chicken",
  "n07864317": "chicken paprika, chicken paprikash",
  "n07864475": "chicken Tetrazzini",
  "n07864638": "Tetrazzini",
  "n07864756": "chicken Kiev",
  "n07864934": "chili, chili con carne",
  "n07865105": "chili dog",
  "n07865196": "chop suey",
  "n07865484": "chow mein",
  "n07865575": "codfish ball, codfish cake",
  "n07865700": "coquille",
  "n07865788": "coquilles Saint-Jacques",
  "n07865911": "Cornish pasty",
  "n07866015": "croquette",
  "n07866151": "cottage pie",
  "n07866277": "rissole",
  "n07866409": "dolmas, stuffed grape leaves",
  "n07866571": "egg foo yong, egg fu yung",
  "n07866723": "egg roll, spring roll",
  "n07866868": "eggs Benedict",
  "n07867021": "enchilada",
  "n07867164": "falafel, felafel",
  "n07867324": "fish and chips",
  "n07867421": "fondue, fondu",
  "n07867616": "cheese fondue",
  "n07867751": "chocolate fondue",
  "n07867883": "fondue, fondu",
  "n07868045": "beef fondue, boeuf fondu bourguignon",
  "n07868200": "French toast",
  "n07868340": "fried rice, Chinese fried rice",
  "n07868508": "frittata",
  "n07868684": "frog legs",
  "n07868830": "galantine",
  "n07868955": "gefilte fish, fish ball",
  "n07869111": "haggis",
  "n07869291": "ham and eggs",
  "n07869391": "hash",
  "n07869522": "corned beef hash",
  "n07869611": "jambalaya",
  "n07869775": "kabob, kebab, shish kebab",
  "n07869937": "kedgeree",
  "n07870069": "souvlaki, souvlakia",
  "n07870167": "lasagna, lasagne",
  "n07870313": "seafood Newburg",
  "n07870478": "lobster Newburg, lobster a la Newburg",
  "n07870620": "shrimp Newburg",
  "n07870734": "Newburg sauce",
  "n07870894": "lobster thermidor",
  "n07871065": "lutefisk, lutfisk",
  "n07871234": "macaroni and cheese",
  "n07871335": "macedoine",
  "n07871436": "meatball",
  "n07871588": "porcupine ball, porcupines",
  "n07871720": "Swedish meatball",
  "n07871810": "meat loaf, meatloaf",
  "n07871940": "meat pie",
  "n07872170": "pasty",
  "n07872284": "pork pie",
  "n07872381": "tourtiere",
  "n07872492": "mostaccioli",
  "n07872593": "moussaka",
  "n07872748": "osso buco",
  "n07872915": "marrowbone",
  "n07873057": "marrow, bone marrow",
  "n07873198": "pheasant under glass",
  "n07873348": "pigs in blankets",
  "n07873464": "pilaf, pilaff, pilau, pilaw",
  "n07873679": "bulgur pilaf",
  "n07873807": "pizza, pizza pie",
  "n07874063": "sausage pizza",
  "n07874159": "pepperoni pizza",
  "n07874259": "cheese pizza",
  "n07874343": "anchovy pizza",
  "n07874441": "Sicilian pizza",
  "n07874531": "poi",
  "n07874674": "pork and beans",
  "n07874780": "porridge",
  "n07874995": "oatmeal, burgoo",
  "n07875086": "loblolly",
  "n07875152": "potpie",
  "n07875267": "rijsttaffel, rijstaffel, rijstafel",
  "n07875436": "risotto, Italian rice",
  "n07875560": "roulade",
  "n07875693": "fish loaf",
  "n07875835": "salmon loaf",
  "n07875926": "Salisbury steak",
  "n07876026": "sauerbraten",
  "n07876189": "sauerkraut",
  "n07876281": "scallopine, scallopini",
  "n07876460": "veal scallopini",
  "n07876550": "scampi",
  "n07876651": "Scotch egg",
  "n07876775": "Scotch woodcock",
  "n07876893": "scrapple",
  "n07877062": "shepherd's pie",
  "n07877187": "spaghetti and meatballs",
  "n07877299": "Spanish rice",
  "n07877411": "steak and kidney pie",
  "n07877575": "kidney pie",
  "n07877675": "steak tartare, tartar steak, cannibal mound",
  "n07877849": "pepper steak",
  "n07877961": "steak au poivre, peppered steak, pepper steak",
  "n07878145": "beef Stroganoff",
  "n07878283": "stuffed cabbage",
  "n07878479": "kishke, stuffed derma",
  "n07878647": "stuffed peppers",
  "n07878785": "stuffed tomato, hot stuffed tomato",
  "n07878926": "stuffed tomato, cold stuffed tomato",
  "n07879072": "succotash",
  "n07879174": "sukiyaki",
  "n07879350": "sashimi",
  "n07879450": "sushi",
  "n07879560": "Swiss steak",
  "n07879659": "tamale",
  "n07879821": "tamale pie",
  "n07879953": "tempura",
  "n07880080": "teriyaki",
  "n07880213": "terrine",
  "n07880325": "Welsh rarebit, Welsh rabbit, rarebit",
  "n07880458": "schnitzel, Wiener schnitzel",
  "n07880583": "tortilla",
  "n07880751": "taco",
  "n07880880": "chicken taco",
  "n07880968": "burrito",
  "n07881117": "beef burrito",
  "n07881205": "quesadilla",
  "n07881310": "tostada",
  "n07881404": "tostada",
  "n07881525": "bean tostada",
  "n07881625": "refried beans, frijoles refritos",
  "n07881800": "beverage, drink, drinkable, potable",
  "n07882420": "wish-wash",
  "n07882497": "concoction, mixture, intermixture",
  "n07882886": "mix, premix",
  "n07883031": "filling",
  "n07883156": "lekvar",
  "n07883251": "potion",
  "n07883384": "elixir",
  "n07883510": "elixir of life",
  "n07883661": "philter, philtre, love-potion, love-philter, love-philtre",
  "n07883860": "chaser",
  "n07883980": "draft, draught, potation, tipple",
  "n07884182": "quaff",
  "n07884266": "round, round of drinks",
  "n07884413": "pledge, toast",
  "n07884567": "alcohol, alcoholic drink, alcoholic beverage, intoxicant, inebriant",
  "n07885223": "drink",
  "n07885705": "proof spirit",
  "n07885832": "libation",
  "n07885937": "libation",
  "n07886057": "home brew, homebrew",
  "n07886176": "hooch, hootch",
  "n07886317": "kava, kavakava",
  "n07886463": "aperitif",
  "n07886572": "brew, brewage",
  "n07886849": "beer",
  "n07887099": "draft beer, draught beer",
  "n07887192": "suds",
  "n07887304": "Munich beer, Munchener",
  "n07887461": "bock, bock beer",
  "n07887634": "lager, lager beer",
  "n07887967": "light beer",
  "n07888058": "Oktoberfest, Octoberfest",
  "n07888229": "Pilsner, Pilsener",
  "n07888378": "shebeen",
  "n07888465": "Weissbier, white beer, wheat beer",
  "n07888709": "Weizenbier",
  "n07888816": "Weizenbock",
  "n07888909": "malt",
  "n07889193": "wort",
  "n07889274": "malt, malt liquor",
  "n07889510": "ale",
  "n07889814": "bitter",
  "n07889990": "Burton",
  "n07890068": "pale ale",
  "n07890226": "porter, porter's beer",
  "n07890352": "stout",
  "n07890540": "Guinness",
  "n07890617": "kvass",
  "n07890750": "mead",
  "n07890890": "metheglin",
  "n07890970": "hydromel",
  "n07891095": "oenomel",
  "n07891189": "near beer",
  "n07891309": "ginger beer",
  "n07891433": "sake, saki, rice beer",
  "n07891613": "nipa",
  "n07891726": "wine, vino",
  "n07892418": "vintage",
  "n07892512": "red wine",
  "n07892813": "white wine",
  "n07893253": "blush wine, pink wine, rose, rose wine",
  "n07893425": "altar wine, sacramental wine",
  "n07893528": "sparkling wine",
  "n07893642": "champagne, bubbly",
  "n07893792": "cold duck",
  "n07893891": "Burgundy, Burgundy wine",
  "n07894102": "Beaujolais",
  "n07894298": "Medoc",
  "n07894451": "Canary wine",
  "n07894551": "Chablis, white Burgundy",
  "n07894703": "Montrachet",
  "n07894799": "Chardonnay, Pinot Chardonnay",
  "n07894965": "Pinot noir",
  "n07895100": "Pinot blanc",
  "n07895237": "Bordeaux, Bordeaux wine",
  "n07895435": "claret, red Bordeaux",
  "n07895595": "Chianti",
  "n07895710": "Cabernet, Cabernet Sauvignon",
  "n07895839": "Merlot",
  "n07895962": "Sauvignon blanc",
  "n07896060": "California wine",
  "n07896165": "Cotes de Provence",
  "n07896287": "dessert wine",
  "n07896422": "Dubonnet",
  "n07896560": "jug wine",
  "n07896661": "macon, maconnais",
  "n07896765": "Moselle",
  "n07896893": "Muscadet",
  "n07896994": "plonk",
  "n07897116": "retsina",
  "n07897200": "Rhine wine, Rhenish, hock",
  "n07897438": "Riesling",
  "n07897600": "liebfraumilch",
  "n07897750": "Rhone wine",
  "n07897865": "Rioja",
  "n07897975": "sack",
  "n07898117": "Saint Emilion",
  "n07898247": "Soave",
  "n07898333": "zinfandel",
  "n07898443": "Sauterne, Sauternes",
  "n07898617": "straw wine",
  "n07898745": "table wine",
  "n07898895": "Tokay",
  "n07899003": "vin ordinaire",
  "n07899108": "vermouth",
  "n07899292": "sweet vermouth, Italian vermouth",
  "n07899434": "dry vermouth, French vermouth",
  "n07899533": "Chenin blanc",
  "n07899660": "Verdicchio",
  "n07899769": "Vouvray",
  "n07899899": "Yquem",
  "n07899976": "generic, generic wine",
  "n07900225": "varietal, varietal wine",
  "n07900406": "fortified wine",
  "n07900616": "Madeira",
  "n07900734": "malmsey",
  "n07900825": "port, port wine",
  "n07900958": "sherry",
  "n07901176": "Manzanilla",
  "n07901264": "Amontillado",
  "n07901355": "Marsala",
  "n07901457": "muscat, muscatel, muscadel, muscadelle",
  "n07901587": "liquor, spirits, booze, hard drink, hard liquor, John Barleycorn, strong drink",
  "n07902121": "neutral spirits, ethyl alcohol",
  "n07902336": "aqua vitae, ardent spirits",
  "n07902443": "eau de vie",
  "n07902520": "moonshine, bootleg, corn liquor",
  "n07902698": "bathtub gin",
  "n07902799": "aquavit, akvavit",
  "n07902937": "arrack, arak",
  "n07903101": "bitters",
  "n07903208": "brandy",
  "n07903543": "applejack",
  "n07903643": "Calvados",
  "n07903731": "Armagnac",
  "n07903841": "Cognac",
  "n07903962": "grappa",
  "n07904072": "kirsch",
  "n07904190": "marc",
  "n07904293": "slivovitz",
  "n07904395": "gin",
  "n07904637": "sloe gin",
  "n07904760": "geneva, Holland gin, Hollands",
  "n07904865": "grog",
  "n07904934": "ouzo",
  "n07905038": "rum",
  "n07905296": "demerara, demerara rum",
  "n07905386": "Jamaica rum",
  "n07905474": "schnapps, schnaps",
  "n07905618": "pulque",
  "n07905770": "mescal",
  "n07905979": "tequila",
  "n07906111": "vodka",
  "n07906284": "whiskey, whisky",
  "n07906572": "blended whiskey, blended whisky",
  "n07906718": "bourbon",
  "n07906877": "corn whiskey, corn whisky, corn",
  "n07907037": "firewater",
  "n07907161": "Irish, Irish whiskey, Irish whisky",
  "n07907342": "poteen",
  "n07907429": "rye, rye whiskey, rye whisky",
  "n07907548": "Scotch, Scotch whiskey, Scotch whisky, malt whiskey, malt whisky, Scotch malt whiskey, Scotch malt whisky",
  "n07907831": "sour mash, sour mash whiskey",
  "n07907943": "liqueur, cordial",
  "n07908411": "absinth, absinthe",
  "n07908567": "amaretto",
  "n07908647": "anisette, anisette de Bordeaux",
  "n07908812": "benedictine",
  "n07908923": "Chartreuse",
  "n07909129": "coffee liqueur",
  "n07909231": "creme de cacao",
  "n07909362": "creme de menthe",
  "n07909504": "creme de fraise",
  "n07909593": "Drambuie",
  "n07909714": "Galliano",
  "n07909811": "orange liqueur",
  "n07909954": "curacao, curacoa",
  "n07910048": "triple sec",
  "n07910152": "Grand Marnier",
  "n07910245": "kummel",
  "n07910379": "maraschino, maraschino liqueur",
  "n07910538": "pastis",
  "n07910656": "Pernod",
  "n07910799": "pousse-cafe",
  "n07910970": "Kahlua",
  "n07911061": "ratafia, ratafee",
  "n07911249": "sambuca",
  "n07911371": "mixed drink",
  "n07911677": "cocktail",
  "n07912093": "Dom Pedro",
  "n07912211": "highball",
  "n07912499": "eye opener",
  "n07912619": "nightcap",
  "n07912726": "hair of the dog",
  "n07912834": "shandygaff, shandy",
  "n07912933": "stirrup cup",
  "n07913081": "sundowner",
  "n07913180": "mixer",
  "n07913300": "bishop",
  "n07913393": "Bloody Mary",
  "n07913537": "Virgin Mary, bloody shame",
  "n07913644": "bullshot",
  "n07913774": "cobbler",
  "n07913882": "collins, Tom Collins",
  "n07914006": "cooler",
  "n07914128": "refresher",
  "n07914271": "smoothie",
  "n07914413": "daiquiri, rum cocktail",
  "n07914586": "strawberry daiquiri",
  "n07914686": "NADA daiquiri",
  "n07914777": "spritzer",
  "n07914887": "flip",
  "n07914995": "gimlet",
  "n07915094": "gin and tonic",
  "n07915213": "grasshopper",
  "n07915366": "Harvey Wallbanger",
  "n07915491": "julep, mint julep",
  "n07915618": "manhattan",
  "n07915800": "Rob Roy",
  "n07915918": "margarita",
  "n07916041": "martini",
  "n07916183": "gin and it",
  "n07916319": "vodka martini",
  "n07916437": "old fashioned",
  "n07916582": "pink lady",
  "n07916773": "posset",
  "n07916872": "syllabub, sillabub",
  "n07916970": "sangaree, sangria",
  "n07917133": "Sazerac",
  "n07917272": "screwdriver",
  "n07917392": "sidecar",
  "n07917507": "Scotch and soda",
  "n07917618": "sling",
  "n07917791": "brandy sling",
  "n07917874": "gin sling",
  "n07917951": "rum sling",
  "n07918028": "sour",
  "n07918193": "whiskey sour, whisky sour",
  "n07918309": "stinger",
  "n07918454": "whiskey neat, whisky neat",
  "n07918601": "whiskey on the rocks, whisky on the rocks",
  "n07918706": "swizzle",
  "n07918879": "hot toddy, toddy",
  "n07919055": "Tom and Jerry",
  "n07919165": "zombie, zombi",
  "n07919310": "fizz",
  "n07919441": "Irish coffee",
  "n07919572": "cafe au lait",
  "n07919665": "cafe noir, demitasse",
  "n07919787": "decaffeinated coffee, decaf",
  "n07919894": "drip coffee",
  "n07920052": "espresso",
  "n07920222": "caffe latte, latte",
  "n07920349": "cappuccino, cappuccino coffee, coffee cappuccino",
  "n07920540": "iced coffee, ice coffee",
  "n07920663": "instant coffee",
  "n07920872": "mocha, mocha coffee",
  "n07920989": "mocha",
  "n07921090": "cassareep",
  "n07921239": "Turkish coffee",
  "n07921360": "chocolate milk",
  "n07921455": "cider, cyder",
  "n07921615": "hard cider",
  "n07921834": "scrumpy",
  "n07921948": "sweet cider",
  "n07922041": "mulled cider",
  "n07922147": "perry",
  "n07922299": "pruno",
  "n07922512": "rotgut",
  "n07922607": "slug",
  "n07922764": "cocoa, chocolate, hot chocolate, drinking chocolate",
  "n07922955": "criollo",
  "n07923034": "ice-cream soda, ice-cream float, float",
  "n07923176": "root beer float",
  "n07923297": "milkshake, milk shake, shake",
  "n07923495": "eggshake",
  "n07923576": "frappe",
  "n07923665": "frappe",
  "n07923748": "juice",
  "n07924033": "fruit juice, fruit crush",
  "n07924276": "nectar",
  "n07924366": "apple juice",
  "n07924443": "cranberry juice",
  "n07924560": "grape juice",
  "n07924655": "must",
  "n07924747": "grapefruit juice",
  "n07924834": "orange juice",
  "n07924955": "frozen orange juice, orange-juice concentrate",
  "n07925116": "pineapple juice",
  "n07925229": "lemon juice",
  "n07925327": "lime juice",
  "n07925423": "papaya juice",
  "n07925500": "tomato juice",
  "n07925608": "carrot juice",
  "n07925708": "V-8 juice",
  "n07925808": "koumiss, kumis",
  "n07925966": "fruit drink, ade",
  "n07926127": "lacing",
  "n07926250": "lemonade",
  "n07926346": "limeade",
  "n07926442": "orangeade",
  "n07926540": "malted milk",
  "n07926642": "malted, malt, malted milk",
  "n07926785": "mate",
  "n07926920": "mulled wine",
  "n07927070": "negus",
  "n07927197": "soft drink",
  "n07927512": "pop, soda, soda pop, soda water, tonic",
  "n07927716": "birch beer",
  "n07927836": "bitter lemon",
  "n07927931": "cola, dope",
  "n07928163": "cream soda",
  "n07928264": "egg cream",
  "n07928367": "ginger ale, ginger pop",
  "n07928488": "orange soda",
  "n07928578": "phosphate",
  "n07928696": "Coca Cola, Coke",
  "n07928790": "Pepsi, Pepsi Cola",
  "n07928887": "root beer",
  "n07928998": "sarsaparilla",
  "n07929172": "tonic, tonic water, quinine water",
  "n07929351": "coffee bean, coffee berry, coffee",
  "n07929519": "coffee, java",
  "n07929940": "cafe royale, coffee royal",
  "n07930062": "fruit punch",
  "n07930205": "milk punch",
  "n07930315": "mimosa, buck's fizz",
  "n07930433": "pina colada",
  "n07930554": "punch",
  "n07930864": "cup",
  "n07931001": "champagne cup",
  "n07931096": "claret cup",
  "n07931280": "wassail",
  "n07931452": "planter's punch",
  "n07931612": "White Russian",
  "n07931733": "fish house punch",
  "n07931870": "May wine",
  "n07932039": "eggnog",
  "n07932176": "glogg",
  "n07932323": "cassiri",
  "n07932454": "spruce beer",
  "n07932614": "rickey",
  "n07932762": "gin rickey",
  "n07932841": "tea, tea leaf",
  "n07933154": "tea bag",
  "n07933274": "tea",
  "n07933530": "tea-like drink",
  "n07933652": "cambric tea",
  "n07933799": "cuppa, cupper",
  "n07933891": "herb tea, herbal tea, herbal",
  "n07934032": "tisane",
  "n07934152": "camomile tea",
  "n07934282": "ice tea, iced tea",
  "n07934373": "sun tea",
  "n07934530": "black tea",
  "n07934678": "congou, congo, congou tea, English breakfast tea",
  "n07934800": "Darjeeling",
  "n07934908": "orange pekoe, pekoe",
  "n07935043": "souchong, soochong",
  "n07935152": "green tea",
  "n07935288": "hyson",
  "n07935379": "oolong",
  "n07935504": "water",
  "n07935737": "bottled water",
  "n07935878": "branch water",
  "n07936015": "spring water",
  "n07936093": "sugar water",
  "n07936177": "tap water",
  "n07936263": "drinking water",
  "n07936459": "ice water",
  "n07936548": "soda water, carbonated water, club soda, seltzer, sparkling water",
  "n07936745": "mineral water",
  "n07936979": "seltzer",
  "n07937069": "Vichy water",
  "n07937203": "brine",
  "n07937344": "perishable, spoilable",
  "n07937461": "couscous",
  "n07937621": "ramekin, ramequin",
  "n07937786": "rugulah, rugelach, ruggelach",
  "n07938007": "multivitamin, multivitamin pill",
  "n07938149": "vitamin pill",
  "n07938313": "soul food",
  "n07938427": "slop",
  "n07938594": "mold, mould",
  "n07938773": "arrangement",
  "n07939159": "straggle",
  "n07939382": "array",
  "n07939638": "classification, categorization, categorisation",
  "n07939880": "dichotomy, duality",
  "n07940121": "trichotomy",
  "n07940242": "clone, clon",
  "n07940448": "kingdom",
  "n07940552": "kingdom",
  "n07940865": "subkingdom",
  "n07941045": "mineral kingdom",
  "n07941170": "biological group",
  "n07941405": "genotype",
  "n07941574": "biotype",
  "n07941729": "community, biotic community",
  "n07941945": "biome",
  "n07942152": "people",
  "n07943300": "peoples",
  "n07943480": "age group, age bracket, cohort",
  "n07943646": "ancients",
  "n07943870": "aged, elderly",
  "n07944050": "young, youth",
  "n07944242": "baffled",
  "n07944408": "blind",
  "n07944618": "blood",
  "n07944754": "brave",
  "n07944900": "timid, cautious",
  "n07945077": "business people, businesspeople",
  "n07945236": "country people, countryfolk",
  "n07945369": "country people, countryfolk",
  "n07945490": "damned",
  "n07945657": "dead",
  "n07945818": "living",
  "n07945949": "deaf",
  "n07946135": "defeated, discomfited",
  "n07946288": "disabled, handicapped",
  "n07946516": "the halt",
  "n07946694": "doomed, lost",
  "n07946851": "enemy",
  "n07946969": "episcopacy, episcopate",
  "n07947069": "estivation, aestivation",
  "n07947255": "folk, folks, common people",
  "n07947675": "gentlefolk",
  "n07947789": "grass roots",
  "n07947958": "free, free people",
  "n07948098": "home folk",
  "n07948183": "homebound",
  "n07948314": "homeless",
  "n07948518": "initiate, enlightened",
  "n07948754": "uninitiate",
  "n07948971": "mentally retarded, retarded, developmentally challenged",
  "n07949160": "network army",
  "n07949463": "nationality",
  "n07949796": "peanut gallery",
  "n07950090": "pocket",
  "n07950295": "retreated",
  "n07950418": "sick",
  "n07950592": "slain",
  "n07950685": "tradespeople",
  "n07950786": "wounded, maimed",
  "n07950920": "social group",
  "n07951464": "collection, aggregation, accumulation, assemblage",
  "n07953367": "armamentarium",
  "n07953499": "art collection",
  "n07953603": "backlog",
  "n07953827": "battery",
  "n07954014": "block",
  "n07954211": "book, rule book",
  "n07954441": "book",
  "n07954576": "bottle collection",
  "n07954731": "bunch, lot, caboodle",
  "n07954863": "coin collection",
  "n07954946": "collage",
  "n07955057": "content",
  "n07955280": "ensemble, tout ensemble",
  "n07955455": "corpus",
  "n07955566": "crop",
  "n07955726": "tenantry",
  "n07955841": "loan collection",
  "n07955961": "findings",
  "n07956112": "flagging",
  "n07956250": "flinders",
  "n07956426": "pack",
  "n07956552": "disk pack, disc pack",
  "n07956721": "pack of cards, deck of cards, deck",
  "n07956887": "hand, deal",
  "n07957193": "long suit",
  "n07957327": "bridge hand",
  "n07957456": "chicane",
  "n07957546": "strong suit",
  "n07957655": "poker hand",
  "n07957855": "royal flush",
  "n07957986": "straight flush",
  "n07958099": "full house",
  "n07958196": "flush",
  "n07958294": "straight",
  "n07958407": "pair",
  "n07958501": "herbarium",
  "n07958642": "stamp collection",
  "n07958727": "statuary",
  "n07958820": "Elgin Marbles",
  "n07959016": "sum, summation, sum total",
  "n07959269": "agglomeration",
  "n07959393": "edition",
  "n07959659": "electron shell",
  "n07959863": "gimmickry",
  "n07959943": "bunch, clump, cluster, clustering",
  "n07960384": "knot",
  "n07960569": "nuclear club",
  "n07960666": "swad",
  "n07960769": "tuft, tussock",
  "n07960928": "wisp",
  "n07961016": "ball, clod, glob, lump, clump, chunk",
  "n07961270": "gob",
  "n07961379": "clew",
  "n07961480": "pile, heap, mound, agglomerate, cumulation, cumulus",
  "n07961956": "compost heap, compost pile",
  "n07962124": "mass",
  "n07962295": "dunghill, midden, muckheap, muckhill",
  "n07962405": "logjam",
  "n07962500": "shock",
  "n07962628": "scrapheap",
  "n07962707": "shock",
  "n07962991": "slagheap",
  "n07963087": "stack",
  "n07963208": "haystack, hayrick, rick",
  "n07963330": "haycock",
  "n07963494": "pyre, funeral pyre",
  "n07963613": "woodpile",
  "n07963711": "combination",
  "n07963987": "amalgam",
  "n07964144": "color scheme, colour scheme",
  "n07964324": "complexion",
  "n07964495": "combination",
  "n07964809": "combination in restraint of trade",
  "n07965085": "body",
  "n07965817": "public",
  "n07965937": "world, domain",
  "n07966140": "society",
  "n07966421": "migration",
  "n07966570": "minority",
  "n07966719": "sector",
  "n07966927": "business, business sector",
  "n07967129": "big business",
  "n07967382": "ethnic group, ethnos",
  "n07967736": "ethnic minority",
  "n07967982": "race",
  "n07968354": "color, colour, people of color, people of colour",
  "n07968550": "master race, Herrenvolk",
  "n07968702": "interest, interest group",
  "n07968974": "special interest",
  "n07969192": "vested interest",
  "n07969366": "military-industrial complex",
  "n07969695": "kin, kin group, kinship group, kindred, clan, tribe",
  "n07970079": "mishpocha, mishpachah",
  "n07970301": "kith",
  "n07970406": "family, family unit",
  "n07970721": "family, family line, folk, kinfolk, kinsfolk, sept, phratry",
  "n07971023": "folks",
  "n07971141": "people",
  "n07971298": "homefolk",
  "n07971449": "house",
  "n07971582": "dynasty",
  "n07972279": "name, gens",
  "n07972425": "feudalism, feudal system",
  "n07972674": "patriarchy, patriarchate",
  "n07972888": "matriarchy, matriarchate",
  "n07973088": "meritocracy",
  "n07973241": "building",
  "n07973373": "broken home",
  "n07973487": "nuclear family, conjugal family",
  "n07973653": "extended family",
  "n07973782": "foster family",
  "n07973868": "foster home",
  "n07974025": "class, stratum, social class, socio-economic class",
  "n07974766": "age class",
  "n07974850": "fringe",
  "n07975026": "gathering, assemblage",
  "n07975909": "bee",
  "n07976069": "carload",
  "n07976181": "congregation",
  "n07976420": "contingent",
  "n07976596": "floor",
  "n07976725": "love feast",
  "n07976851": "quilting bee",
  "n07976936": "pair",
  "n07977108": "hit parade",
  "n07977237": "Judaica",
  "n07977344": "kludge",
  "n07977592": "library, program library, subroutine library",
  "n07977870": "library",
  "n07978055": "bibliotheca",
  "n07978170": "public library",
  "n07978276": "rental collection",
  "n07978423": "mythology",
  "n07978924": "classical mythology",
  "n07979425": "Greek mythology",
  "n07983170": "Roman mythology",
  "n07983856": "Norse mythology",
  "n07984805": "Nag Hammadi, Nag Hammadi Library",
  "n07985135": "singleton",
  "n07985223": "pair, brace",
  "n07985384": "team",
  "n07985533": "relay",
  "n07985628": "couple, twosome, duo, duet",
  "n07985825": "trilogy",
  "n07985948": "room",
  "n07986066": "trio, threesome, triad, trinity",
  "n07986198": "trio, triad, triplet, triple",
  "n07986381": "trip wire",
  "n07986617": "Trimurti",
  "n07986771": "triplicity, trigon",
  "n07986982": "triumvirate",
  "n07987135": "troika",
  "n07987216": "turnout",
  "n07987380": "quartet, quartette, foursome",
  "n07987580": "quintet, quintette, fivesome",
  "n07987689": "sextet, sextette, sixsome",
  "n07987794": "septet, septette, sevensome",
  "n07987903": "octet, octette, eightsome",
  "n07988010": "quadrumvirate",
  "n07988089": "quartet, quartette, quadruplet, quadruple",
  "n07988229": "quintet, quintette, quintuplet, quintuple",
  "n07988369": "sextet, sextette, sestet",
  "n07988490": "septet, septette",
  "n07988604": "octet, octette",
  "n07988716": "Tweedledum and Tweedledee, Tweedledee and Tweedledum",
  "n07988857": "couple, mates, match",
  "n07989083": "power couple",
  "n07989220": "DINK",
  "n07989373": "marriage, married couple, man and wife",
  "n07989620": "Bronte sisters",
  "n07989741": "Marx Brothers",
  "n07989973": "same-sex marriage",
  "n07990158": "mixed marriage",
  "n07990377": "association",
  "n07990648": "antibiosis",
  "n07990824": "brood",
  "n07990956": "flock",
  "n07991169": "flock, fold",
  "n07991272": "flock",
  "n07991364": "congregation, fold, faithful",
  "n07991579": "bevy",
  "n07991780": "covert",
  "n07991868": "covey",
  "n07991994": "exaltation",
  "n07992116": "gaggle",
  "n07992222": "wisp",
  "n07992308": "clade",
  "n07992450": "taxonomic group, taxonomic category, taxon",
  "n07993109": "biota, biology",
  "n07993279": "fauna, zoology",
  "n07993526": "petting zoo",
  "n07993643": "avifauna",
  "n07993776": "wildlife",
  "n07993929": "animal group",
  "n07994331": "herd",
  "n07994555": "herd",
  "n07994752": "gam",
  "n07994818": "remuda",
  "n07994941": "pack",
  "n07995074": "wolf pack",
  "n07995164": "pod",
  "n07995278": "pride",
  "n07995365": "clowder",
  "n07995453": "school, shoal",
  "n07995617": "caste",
  "n07995856": "colony",
  "n07996010": "colony",
  "n07996149": "swarm, cloud",
  "n07996412": "infestation, plague",
  "n07996598": "warren",
  "n07996689": "set",
  "n07997338": "chess set",
  "n07997486": "manicure set",
  "n07997581": "Victoriana",
  "n07997703": "class, category, family",
  "n07998206": "brass family",
  "n07998323": "violin family",
  "n07998450": "woodwind family",
  "n07998573": "stamp",
  "n07998712": "union, sum, join",
  "n07998904": "direct sum",
  "n07999068": "intersection, product, Cartesian product",
  "n07999291": "sex",
  "n07999471": "field",
  "n07999584": "field",
  "n07999699": "set",
  "n08000118": "domain, domain of a function",
  "n08000304": "image, range, range of a function",
  "n08000644": "universal set",
  "n08000881": "locus",
  "n08001083": "subgroup",
  "n08001209": "subset",
  "n08001344": "null set",
  "n08001441": "Mandelbrot set",
  "n08001685": "mathematical space, topological space",
  "n08002015": "broadcasting company",
  "n08002125": "bureau de change",
  "n08002269": "car company, auto company",
  "n08002384": "dot-com, dot com, dot com company",
  "n08002578": "drug company, pharmaceutical company, pharma",
  "n08002717": "East India Company",
  "n08003035": "electronics company",
  "n08003173": "film company",
  "n08003306": "indie",
  "n08003427": "food company",
  "n08003525": "furniture company",
  "n08003619": "mining company",
  "n08003717": "shipping company",
  "n08003839": "steel company",
  "n08003935": "subsidiary company, subsidiary",
  "n08004089": "transportation company",
  "n08004210": "trucking company",
  "n08004342": "subspace",
  "n08004443": "null space",
  "n08004548": "manifold",
  "n08004695": "metric space",
  "n08004953": "Euclidean space",
  "n08005123": "Hilbert space",
  "n08005260": "field",
  "n08005580": "field",
  "n08005761": "bit field",
  "n08005877": "scalar field",
  "n08005954": "solution, root",
  "n08006094": "bracket",
  "n08006250": "income bracket, tax bracket, income tax bracket",
  "n08006405": "price bracket",
  "n08006511": "declension",
  "n08006741": "conjugation",
  "n08006868": "conjugation",
  "n08006989": "denomination",
  "n08007204": "histocompatibility complex",
  "n08007430": "job lot",
  "n08007534": "suite",
  "n08007667": "bedroom suite, bedroom set",
  "n08007777": "diningroom suite, diningroom set",
  "n08007897": "livingroom suite, livingroom set",
  "n08008017": "package, bundle, packet, parcel",
  "n08008254": "wisp",
  "n08008335": "organization, organisation",
  "n08009239": "adhocracy",
  "n08009478": "affiliate",
  "n08009659": "bureaucracy",
  "n08009834": "nongovernmental organization, NGO",
  "n08010191": "Alcoholics Anonymous, AA",
  "n08010364": "Abu Hafs al-Masri Brigades",
  "n08010559": "Abu Sayyaf, Bearer of the Sword",
  "n08010942": "Aksa Martyrs Brigades, al-Aksa Martyrs Brigades, Martyrs of al-Aqsa",
  "n08011266": "Alex Boncayao Brigade, ABB, Revolutionary Proletarian Army, RPA-ABB",
  "n08011523": "al-Fatah, Fatah, al-Asifa",
  "n08012028": "al-Gama'a al-Islamiyya, Islamic Group",
  "n08012384": "al Itihaad al Islamiya, al-Itihaad al-Islamiya, Islamic Unity, AIAI",
  "n08012765": "al-Jihad, Egyptian Islamic Jihad, Islamic Jihad, Vanguards of Conquest",
  "n08013176": "al-Ma'unah",
  "n08013453": "al-Muhajiroun",
  "n08013653": "Al Nathir",
  "n08013845": "al-Qaeda, Qaeda, al-Qa'ida, al-Qaida, Base",
  "n08014202": "al-Rashid Trust",
  "n08014615": "al Sunna Wal Jamma, Followers of the Phrophet",
  "n08014860": "al-Tawhid, Al Tawhid, Divine Unity",
  "n08015116": "al-Ummah",
  "n08015321": "Ansar al Islam, Ansar al-Islam, Supporters of Islam",
  "n08015731": "Armata Corsa, Corsican Army",
  "n08016035": "Armed Islamic Group, GIA",
  "n08016385": "Armenian Secret Army for the Liberation of Armenia, ASALA, Orly Group, 3rd October Organization",
  "n08016900": "Army for the Liberation of Rwanda, ALIR, Former Armed Forces, FAR, Interahamwe",
  "n08017257": "Asbat al-Ansar, Band of Partisans",
  "n08017614": "Aum Shinrikyo, Aum, Supreme Truth",
  "n08017974": "Baader Meinhof Gang, Baader-Meinhof Gang",
  "n08018189": "Basque Homeland and Freedom, Basque Fatherland and Liberty, Euskadi ta Askatasuna, ETA",
  "n08018666": "Black September Movement",
  "n08018983": "Chukaku-Ha",
  "n08019281": "Continuity Irish Republican Army, CIRA, Continuity Army Council",
  "n08019523": "Democratic Front for the Liberation of Palestine, DFLP, Popular Democratic Front for the Liberation of Palestine, PDFLP",
  "n08019913": "East Turkistan Islamic Movement, East Turkestan Islamic Movement",
  "n08020242": "Fatah Revolutionary Council, Fatah-RC, Abu Nidal Organization, ANO, Arab Revolutionary Brigades, Black September, Revolutionary Organization of Socialist Muslims",
  "n08020785": "Fatah Tanzim, Tanzim",
  "n08021129": "First of October Antifascist Resistance Group, GRAPO",
  "n08021464": "Force 17",
  "n08021785": "Forces of Umar Al-Mukhtar, Umar al-Mukhtar Forces",
  "n08022087": "Greenpeace",
  "n08022259": "Hamas, Islamic Resistance Movement",
  "n08022666": "Harkat-ul-Jihad-e-Islami, Harakat ul-Jihad-I-Islami, HUJI",
  "n08022972": "Harkat-ul-Mujahidin, HUM, Harkat ul-Ansar, HUA, Harkat ul-Mujahedeen, Al Faran, Movement of Holy Warriors",
  "n08023374": "Hizballah, Hezbollah, Hizbollah, Hizbullah, Lebanese Hizballah, Party of God, Islamic Jihad, Islamic Jihad for the Liberation of Palestine, Revolutionary Justice Organization, Organization of the Oppressed on Earth",
  "n08023843": "Hizb ut-Tahrir, Freedom Party",
  "n08024096": "International Islamic Front for Jihad against Jews and Crusaders",
  "n08024408": "Irish National Liberation Army, INLA, People's Liberation Army, People's Republican Army, Catholic Reaction Force",
  "n08024732": "Irish Republican Army, IRA, Provisional Irish Republican Army, Provisional IRA, Provos",
  "n08025112": "Islamic Army of Aden, IAA, Islamic Army of Aden-Abyan, Aden-Abyan Islamic Army",
  "n08025497": "Islamic Great Eastern Raiders-Front, IBDA-C",
  "n08025835": "Islamic Group of Uzbekistan, IMU, Islamic Party of Turkestan",
  "n08026197": "Jaish-i-Mohammed, Jaish-e-Muhammad, JEM, Army of Muhammad",
  "n08026539": "Jamaat ul-Fuqra, Fuqra, Tanzimul Fuqra",
  "n08026904": "Japanese Red Army, JRA, Anti-Imperialist International Brigade",
  "n08027314": "Jayshullah",
  "n08027518": "Jemaah Islamiyah, JI, Islamic Group, Islamic Community, Malaysian Mujahidin Group, Malaysia Militant Group",
  "n08027920": "Jerusalem Warriors",
  "n08028148": "Jund-ul-Islam, Soldiers of God",
  "n08028397": "Kahane Chai, Kach",
  "n08028623": "Kaplan Group, Association of Islamic Groups and Communities, Caliphate State",
  "n08028999": "Khmer Rouge, KR, Party of Democratic Kampuchea, Communist Party of Kampuchea",
  "n08029421": "Ku Klux Klan, Klan, KKK",
  "n08029784": "klavern",
  "n08029908": "Kurdistan Workers Party, Kurdistan Labor Pary, Partiya Karkeran Kurdistan, PPK",
  "n08030185": "Contras",
  "n08030379": "Pesh Merga",
  "n08030481": "Lashkar-e-Jhangvi",
  "n08030711": "Lashkar-e-Omar, Al Qanoon",
  "n08031020": "Lashkar-e-Taiba, Lashkar-e-Toiba, Lashkar-e-Tayyiba, LET, Army of the Pure, Army of the Righteous",
  "n08031386": "Laskar Jihad, Holy War Warriors",
  "n08031663": "Lautaro Youth Movement, Lautaro Faction of the United Popular Action Movement, Lautaro Popular Rebel Forces",
  "n08032023": "Liberation Tigers of Tamil Eelam, LTTE, Tamil Tigers, Tigers, World Tamil Association, World Tamil Movement",
  "n08032594": "Libyan Islamic Fighting Group, FIG, Al-Jama'a al-Islamiyyah al-Muqatilah bi-Libya, Libyan Fighting Group, Libyan Islamic Group",
  "n08032955": "Lord's Resistance Army",
  "n08033194": "Loyalist Volunteer Force",
  "n08033454": "Maktab al-Khidmat, MAK",
  "n08033829": "Manuel Rodriquez Patriotic Front",
  "n08034028": "Moranzanist Patriotic Front",
  "n08034299": "Moro Islamic Liberation Front",
  "n08034579": "Mujahedeen Kompak",
  "n08034778": "Mujahidin-e Khalq Organization, MKO, MEK, People's Mujahidin of Iran",
  "n08035233": "National Liberation Army, ELN, Nestor Paz Zamora Commission, CNPZ",
  "n08035601": "National Liberation Army, ELN",
  "n08036005": "National Liberation Front of Corsica, FLNC",
  "n08036293": "New People's Army, NPA",
  "n08036669": "Orange Order, Association of Orangemen",
  "n08036849": "Orange Group, OV",
  "n08037118": "Palestine Islamic Jihad, Palestinian Islamic Jihad, PIJ, Harakat al-Jihad al-Islami al-Filastini",
  "n08037503": "Palestine Liberation Front, PLF, Jabat al-Tahrir al-Filistiniyyah",
  "n08037861": "Palestinian Hizballah",
  "n08038131": "Pentagon Gang",
  "n08038379": "Popular Front for the Liberation of Palestine, PFLP",
  "n08038748": "Popular Front for the Liberation of Palestine-General Command, PFLP-GC",
  "n08038995": "Popular Struggle Front, PSF",
  "n08039312": "15 May Organization",
  "n08039601": "People against Gangsterism and Drugs, PAGAD",
  "n08040008": "Puka Inti, Sol Rojo, Red Sun",
  "n08040257": "Qassam Brigades, Salah al-Din Battalions, Iz Al-Din Al-Qassam Battalions",
  "n08040522": "Qibla",
  "n08040762": "Real IRA, Real Irish Republican Army, RIRA, Dissident Irish Republican Army",
  "n08041106": "Red Army Faction, RAF",
  "n08041484": "Red Brigades, Brigate Rosse, BR",
  "n08041840": "Red Hand Defenders, RHD",
  "n08042183": "Revolutionary Armed Forces of Colombia, Fuerzas Armadas Revolucionarios de Colombia, FARC",
  "n08042536": "Revolutionary Organization 17 November, 17 November",
  "n08042856": "Revolutionary People's Liberation Party, Revolutionary People's Liberation Front",
  "n08043169": "Revolutionary People's Struggle, ELA",
  "n08043499": "Revolutionary Proletarian Nucleus, Revolutionary Proletarian Initiative Nuclei, NIPR",
  "n08043848": "Revolutionary United Front, RUF",
  "n08044265": "Salafist Group, Salafast Group for Call and Combat, GSPC",
  "n08044676": "Shining Path, Sendero Luminoso, SL",
  "n08045140": "Sipah-e-Sahaba",
  "n08045428": "Tareekh e Kasas, Movement for Revenge",
  "n08045681": "Tupac Amaru Revolutionary Movement, Movimiento Revolucionario Tupac Anaru, MRTA",
  "n08046032": "Tupac Katari Guerrilla Army, EGTK",
  "n08046346": "Turkish Hizballah",
  "n08046759": "Ulster Defence Association, UDA",
  "n08047032": "United Self-Defense Force of Colombia, United Self-Defense Group of Colombia, Autodefensas Unidas de Colombia, AUC",
  "n08047501": "Markaz-ud-Dawa-wal-Irshad, MDI",
  "n08047747": "Red Cross",
  "n08047890": "Salvation Army",
  "n08048042": "Tammany Hall, Tammany Society, Tammany",
  "n08048300": "Umma Tameer-e-Nau, UTN",
  "n08048625": "fiefdom",
  "n08048743": "line of defense, line of defence",
  "n08048948": "line organization, line organisation",
  "n08049125": "National Trust, NT",
  "n08049401": "association",
  "n08049989": "British Commonwealth, Commonwealth of Nations",
  "n08050385": "polity",
  "n08050484": "quango, quasi-NGO",
  "n08050678": "government, authorities, regime",
  "n08051384": "authoritarian state, authoritarian regime",
  "n08051565": "bureaucracy",
  "n08051739": "ancien regime",
  "n08051946": "court, royal court",
  "n08052135": "Court of Saint James's",
  "n08052227": "Porte, Sublime Porte",
  "n08052330": "Downing Street",
  "n08052413": "empire",
  "n08052549": "federal government",
  "n08052690": "government-in-exile",
  "n08052874": "local government",
  "n08053003": "military government, stratocracy",
  "n08053121": "palace",
  "n08053260": "papacy, pontificate",
  "n08053407": "Soviets",
  "n08053576": "institution, establishment",
  "n08053905": "medical institution",
  "n08054076": "clinic",
  "n08054226": "extended care facility",
  "n08054417": "hospital",
  "n08054613": "eye clinic",
  "n08054721": "financial institution, financial organization, financial organisation",
  "n08055150": "issuer",
  "n08055321": "giro",
  "n08055516": "clearing house",
  "n08055718": "lending institution",
  "n08055824": "charity",
  "n08055964": "community chest",
  "n08056128": "soup kitchen",
  "n08056231": "enterprise",
  "n08056471": "giant",
  "n08056601": "collective",
  "n08056747": "collective farm",
  "n08056873": "kibbutz",
  "n08057068": "kolkhoz",
  "n08057206": "agency",
  "n08057460": "brokerage",
  "n08057633": "carrier, common carrier",
  "n08057816": "chain",
  "n08058098": "company",
  "n08058937": "conglomerate, empire",
  "n08059125": "large cap",
  "n08059250": "small cap",
  "n08059412": "corporation, corp",
  "n08059870": "firm, house, business firm",
  "n08060193": "franchise, dealership",
  "n08060446": "manufacturer, maker, manufacturing business",
  "n08060694": "partnership",
  "n08060878": "copartnership",
  "n08061042": "business, concern, business concern, business organization, business organisation",
  "n08061609": "apparel chain",
  "n08061695": "discount chain",
  "n08061801": "restaurant chain",
  "n08061905": "distributor",
  "n08062092": "direct mailer",
  "n08062206": "retail chain",
  "n08062326": "accounting firm",
  "n08062464": "consulting firm, consulting company",
  "n08062623": "publisher, publishing house, publishing firm, publishing company",
  "n08062842": "publishing conglomerate, publishing empire",
  "n08062970": "newspaper, paper, newspaper publisher",
  "n08063129": "newsroom",
  "n08063303": "magazine, magazine publisher",
  "n08063446": "dealer",
  "n08063650": "car dealer",
  "n08063738": "computer dealer",
  "n08063836": "jewelry dealer, jewelry store",
  "n08063947": "truck dealer",
  "n08064039": "law firm",
  "n08064130": "defense, defence, defense force, defence force",
  "n08064372": "bastion",
  "n08064523": "defense, defence, defense team, defense lawyers",
  "n08064742": "prosecution",
  "n08064888": "planting",
  "n08065093": "commercial enterprise",
  "n08065234": "industry",
  "n08065937": "processor",
  "n08066095": "armorer, armourer",
  "n08066186": "aluminum business, aluminum industry",
  "n08066317": "apparel industry, garment industry, fashion industry, fashion business, rag trade",
  "n08066491": "banking industry, banking system",
  "n08066644": "bottler",
  "n08066763": "car manufacturer, car maker, carmaker, auto manufacturer, auto maker, automaker",
  "n08066965": "computer business",
  "n08067077": "automobile industry",
  "n08067218": "aviation",
  "n08067342": "chemical industry",
  "n08067460": "coal industry",
  "n08067565": "computer industry",
  "n08067683": "construction industry, housing industry",
  "n08067801": "electronics industry",
  "n08067951": "entertainment industry, show business, show biz",
  "n08068151": "film industry, movie industry",
  "n08068375": "Bollywood",
  "n08068457": "filmdom, screenland, screen",
  "n08068597": "Hollywood",
  "n08068691": "growth industry",
  "n08068788": "lighting industry",
  "n08068924": "munitions industry, arms industry",
  "n08069050": "oil industry, refining industry, oil business",
  "n08069241": "oil company",
  "n08069342": "packaging company, packaging concern",
  "n08069487": "pipeline company",
  "n08069627": "printing concern, printing business, printing company",
  "n08069768": "plastics industry",
  "n08069878": "brokerage, brokerage firm, securities firm",
  "n08070130": "bucket shop",
  "n08070236": "commodity brokerage",
  "n08070341": "marriage brokerage, marriage mart",
  "n08070465": "insurance company, insurance firm, insurer, insurance underwriter, underwriter",
  "n08070674": "pension fund",
  "n08070850": "investment company, investment trust, investment firm, fund",
  "n08071229": "hedge fund, hedgefund",
  "n08071516": "mutual fund, mutual fund company, open-end fund, open-end investment company",
  "n08071756": "index fund",
  "n08071908": "closed-end fund, closed-end investment company",
  "n08072104": "face-amount certificate company",
  "n08072287": "Real Estate Investment Trust, REIT",
  "n08072536": "unit investment trust, unit trust",
  "n08072837": "market, securities industry",
  "n08073130": "bear market",
  "n08073243": "bull market",
  "n08073355": "the City",
  "n08073468": "Wall Street, the Street",
  "n08073601": "money market",
  "n08073700": "service industry",
  "n08073844": "management consulting",
  "n08073992": "shipbuilder",
  "n08074109": "shipbuilding industry",
  "n08074225": "shoe industry",
  "n08074330": "sign industry",
  "n08074421": "signage",
  "n08074582": "steel industry",
  "n08074691": "sunrise industry",
  "n08074879": "tobacco industry",
  "n08075009": "toy industry, toy business",
  "n08075140": "trucking industry",
  "n08075287": "agriculture",
  "n08075388": "brotherhood, fraternity, sodality",
  "n08075647": "sisterhood",
  "n08075847": "establishment",
  "n08075929": "corporate investor",
  "n08076075": "target company, takeover target",
  "n08076253": "raider",
  "n08076455": "sleeping beauty",
  "n08076578": "underperformer",
  "n08076705": "white knight",
  "n08076833": "white squire",
  "n08076946": "auction house",
  "n08077035": "A-team",
  "n08077177": "battery",
  "n08077292": "administrative unit, administrative body",
  "n08077711": "company",
  "n08077842": "coronary care unit",
  "n08078020": "family, household, house, home, menage",
  "n08078453": "menage a trois",
  "n08078644": "flying squad",
  "n08078819": "major-league team, major-league club",
  "n08078976": "minor-league team, minor-league club",
  "n08079151": "farm team, farm club",
  "n08079319": "baseball team",
  "n08079613": "baseball club, ball club, club, nine",
  "n08079852": "basketball team, five",
  "n08080025": "football team, eleven",
  "n08080386": "hockey team",
  "n08080510": "junior varsity, JV",
  "n08080652": "varsity, first team",
  "n08080762": "second string",
  "n08080947": "police squad, squad",
  "n08081142": "powerhouse",
  "n08081244": "offense, offence",
  "n08081403": "defense, defence, defending team",
  "n08081668": "religion, faith, organized religion",
  "n08082236": "Christendom, Christianity",
  "n08082602": "church, Christian church",
  "n08082899": "church",
  "n08083083": "Armenian Church, Armenian Apostolic Orthodox Church",
  "n08083320": "Catholic Church",
  "n08083599": "Roman Catholic, Western Church, Roman Catholic Church, Church of Rome, Roman Church",
  "n08085159": "Albigenses, Cathars, Cathari",
  "n08085359": "Nestorian Church",
  "n08085535": "Rome",
  "n08085648": "Curia",
  "n08085824": "Sacred College, College of Cardinals",
  "n08086050": "Old Catholic Church",
  "n08086219": "Eastern Church, Byzantine Church",
  "n08086356": "Orthodox Church, Orthodox Catholic Church, Eastern Orthodox Church, Eastern Church, Eastern Orthodox",
  "n08086646": "Greek Orthodox Church, Greek Church",
  "n08086821": "Russian Orthodox Church",
  "n08086987": "Uniat Church, Uniate Church",
  "n08087203": "Coptic Church",
  "n08087319": "Pentecostal religion",
  "n08087570": "Protestant Church, Protestant",
  "n08087776": "Christian Church, Disciples of Christ",
  "n08087981": "Anglican Church, Anglican Communion, Church of England",
  "n08088472": "Episcopal Church, Protestant Episcopal Church",
  "n08088675": "Church of Ireland",
  "n08088792": "Episcopal Church, Episcopal Church of Scotland",
  "n08088963": "High Church, High Anglican Church",
  "n08089205": "Church of Jesus Christ of Latter-Day Saints, Mormon Church, Mormons",
  "n08089420": "Baptist Church, Baptists",
  "n08089627": "Baptist denomination",
  "n08089797": "American Baptist Convention, Northern Baptist Convention",
  "n08089955": "Southern Baptist Convention",
  "n08090083": "Arminian Baptist, General Baptist",
  "n08090329": "Calvinistic Baptist, Particular Baptist",
  "n08090547": "Church of the Brethren, Dunkers, Dippers",
  "n08090803": "Christian Science, Church of Christ Scientist",
  "n08090973": "Congregational Church",
  "n08091152": "Congregational Christian Church",
  "n08091290": "Evangelical and Reformed Church",
  "n08091412": "United Church of Christ",
  "n08091575": "Jehovah's Witnesses",
  "n08091744": "Lutheran Church",
  "n08091891": "Presbyterian Church",
  "n08092040": "Unitarian Church",
  "n08092190": "Arminian Church",
  "n08092340": "Methodist Church, Methodists",
  "n08092539": "Methodist denomination",
  "n08092713": "Wesleyan Methodist Church, Wesleyan Methodists",
  "n08092855": "Evangelical United Brethren Church",
  "n08092960": "United Methodist Church",
  "n08093111": "Anabaptist denomination",
  "n08093269": "Mennonite Church",
  "n08093421": "Unification Church",
  "n08093653": "Abecedarian",
  "n08093813": "Amish sect",
  "n08094013": "Judaism, Hebraism, Jewish religion",
  "n08094262": "Sanhedrin",
  "n08094386": "Karaites",
  "n08094659": "Orthodox Judaism, Jewish Orthodoxy",
  "n08094866": "Hasidim, Hassidim, Hasidism, Chasidim, Chassidim",
  "n08095160": "Conservative Judaism",
  "n08095426": "Reform Judaism",
  "n08095647": "Islam, Muslimism",
  "n08096027": "Islamism",
  "n08096301": "Shiah, Shia, Shiah Islam",
  "n08096474": "Sunni, Sunni Islam",
  "n08096624": "Hinduism, Hindooism",
  "n08096950": "Brahmanism, Brahminism",
  "n08097072": "Shivaism, Sivaism",
  "n08097222": "Shaktism, Saktism",
  "n08097391": "Vaishnavism, Vaisnavism",
  "n08097531": "Haredi",
  "n08097766": "Hare Krishna, International Society for Krishna Consciousness, ISKCON",
  "n08098192": "Jainism",
  "n08098346": "Taoism",
  "n08098499": "Taoism",
  "n08098708": "Buddhism",
  "n08099029": "Zen, Zen Buddhism",
  "n08099264": "Mahayana",
  "n08099475": "Hinayana",
  "n08099647": "Tantrism",
  "n08099825": "Khalsa",
  "n08100033": "Scientology, Church of Scientology",
  "n08100320": "Shinto",
  "n08100481": "Kokka Shinto, Kokka",
  "n08100656": "Shuha Shinto, Shua",
  "n08100778": "established church",
  "n08100907": "vicariate, vicarship",
  "n08101085": "variety",
  "n08101410": "breed, strain, stock",
  "n08101755": "bloodstock",
  "n08101846": "pedigree",
  "n08101937": "lineage, line, line of descent, descent, bloodline, blood line, blood, pedigree, ancestry, origin, parentage, stemma, stock",
  "n08102282": "side",
  "n08102402": "genealogy, family tree",
  "n08102555": "phylum",
  "n08103299": "subphylum",
  "n08103457": "superphylum",
  "n08103635": "phylum",
  "n08103777": "class",
  "n08106661": "subclass",
  "n08106798": "superclass",
  "n08106934": "order",
  "n08107191": "suborder",
  "n08107343": "superorder",
  "n08107499": "family",
  "n08108304": "superfamily",
  "n08108450": "form family",
  "n08108627": "subfamily",
  "n08108784": "tribe",
  "n08108972": "genus",
  "n08109624": "subgenus",
  "n08109772": "monotype",
  "n08109940": "type genus",
  "n08110197": "form genus",
  "n08110373": "species",
  "n08110648": "subspecies, race",
  "n08110866": "endangered species",
  "n08111027": "fish species",
  "n08111157": "form, variant, strain, var.",
  "n08111419": "type",
  "n08111599": "type species",
  "n08111783": "civilization, civilisation",
  "n08112096": "profession",
  "n08112402": "legal profession, bar, legal community",
  "n08112630": "health profession",
  "n08112829": "medical profession, medical community",
  "n08112976": "nursing",
  "n08113073": "businessmen, business community",
  "n08113197": "community of scholars",
  "n08113322": "economics profession",
  "n08113443": "priesthood",
  "n08113634": "pastorate",
  "n08113710": "prelacy, prelature",
  "n08113797": "ministry",
  "n08113910": "rabbinate",
  "n08114004": "ministry",
  "n08114152": "Foreign Office",
  "n08114287": "Home Office",
  "n08114400": "French Foreign Office, Quai d'Orsay",
  "n08114581": "Free French, Fighting French",
  "n08114861": "department, section",
  "n08115204": "academic department",
  "n08115602": "anthropology department, department of anthropology",
  "n08115778": "art department",
  "n08115912": "biology department, department of biology",
  "n08116073": "chemistry department, department of chemistry",
  "n08116240": "department of computer science",
  "n08116398": "economics department, department of economics",
  "n08116565": "English department, department of English",
  "n08116734": "history department, department of history",
  "n08116879": "linguistics department, department of linguistics",
  "n08117052": "mathematics department, department of mathematics",
  "n08117225": "philosophy department, department of philosophy",
  "n08117379": "physics department, department of physics",
  "n08117540": "music department, department of music",
  "n08117702": "psychology department, department of psychology",
  "n08117872": "sociology department, department of sociology",
  "n08118039": "business department",
  "n08118260": "advertising department, advertising division",
  "n08118414": "editorial department",
  "n08118592": "city desk, city room",
  "n08118725": "sports desk",
  "n08118849": "parts department",
  "n08118991": "personnel department, personnel office, personnel, staff office",
  "n08119226": "plant department, building department",
  "n08119397": "purchasing department",
  "n08119525": "sales department, sales division, sales force",
  "n08119698": "service department",
  "n08119821": "government department",
  "n08120079": "payroll, payroll department",
  "n08120235": "treasury",
  "n08120384": "local department, department of local government",
  "n08120624": "corrections, department of corrections",
  "n08120910": "security, security department",
  "n08121117": "fire department",
  "n08121301": "fire brigade",
  "n08121394": "fire brigade, fire company",
  "n08121542": "police department",
  "n08121694": "sanitation department",
  "n08121867": "Special Branch",
  "n08122009": "State Department",
  "n08122141": "federal department, federal office, department of the federal government",
  "n08122358": "Atomic Energy Commission, AEC",
  "n08122596": "Nuclear Regulatory Commission, NRC",
  "n08122768": "Manhattan Project",
  "n08122960": "Environmental Protection Agency, EPA",
  "n08123167": "executive department",
  "n08123696": "executive agency",
  "n08123970": "Federal Emergency Management Agency, FEMA",
  "n08124256": "Food and Drug Administration, FDA",
  "n08124496": "Council of Economic Advisors",
  "n08124649": "Center for Disease Control and Prevention, CDC",
  "n08124971": "Central Intelligence Agency, CIA",
  "n08125420": "Counterterrorist Center, CTC",
  "n08125722": "Nonproliferation Center, NPC",
  "n08125993": "Interstate Commerce Commission, ICC",
  "n08126290": "National Aeronautics and Space Administration, NASA",
  "n08126483": "National Archives and Records Administration, NARA",
  "n08126716": "National Labor Relations Board, NLRB",
  "n08126921": "National Science Foundation, NSF",
  "n08127166": "Postal Rate Commission",
  "n08127304": "United States Postal Service, US Postal Service, USPS",
  "n08127557": "United States Postal Inspection Service, US Postal Inspection Service",
  "n08127760": "National Security Council, NSC",
  "n08128004": "Council on Environmental Policy",
  "n08128159": "Joint Chiefs of Staff, Joint Chiefs",
  "n08128492": "Office of Management and Budget, OMB",
  "n08128645": "United States Trade Representative, US Trade Representative",
  "n08128837": "White House, EXEC",
  "n08128964": "Department of Agriculture, Agriculture Department, Agriculture, USDA",
  "n08129268": "Department of Commerce, Commerce Department, Commerce, DoC",
  "n08129621": "Bureau of the Census, Census Bureau",
  "n08129883": "National Oceanic and Atmospheric Administration, NOAA",
  "n08130292": "National Climatic Data Center, NCDC",
  "n08130476": "National Weather Service",
  "n08130712": "Technology Administration",
  "n08131005": "National Institute of Standards and Technology, NIST",
  "n08131254": "National Technical Information Service, NTIS",
  "n08131530": "Department of Defense, Defense Department, United States Department of Defense, Defense, DoD",
  "n08132046": "Defense Advanced Research Projects Agency, DARPA",
  "n08132323": "Department of Defense Laboratory System, LABLINK",
  "n08132637": "Department of Education, Education Department, Education",
  "n08132955": "Department of Energy, Energy Department, Energy, DOE",
  "n08133189": "Department of Energy Intelligence, DOEI",
  "n08133536": "Department of Health and Human Services, Health and Human Services, HHS",
  "n08133855": "United States Public Health Service, PHS",
  "n08134081": "National Institutes of Health, NIH",
  "n08134415": "Federal Communications Commission, FCC",
  "n08134649": "Social Security Administration, SSA",
  "n08134807": "Department of Homeland Security, Homeland Security",
  "n08135062": "Department of Housing and Urban Development, Housing and Urban Development, HUD",
  "n08135342": "Department of Justice, Justice Department, Justice, DoJ",
  "n08135770": "Bureau of Justice Assistance, BJA",
  "n08136027": "Bureau of Justice Statistics, BJS",
  "n08136260": "Federal Bureau of Investigation, FBI",
  "n08136502": "Immigration and Naturalization Service, INS",
  "n08136767": "United States Border Patrol, US Border Patrol",
  "n08137028": "Federal Law Enforcement Training Center, FLETC",
  "n08137251": "Financial Crimes Enforcement Network, FinCEN",
  "n08137495": "Department of Labor, Labor Department, Labor, DoL",
  "n08137738": "Department of State, United States Department of State, State Department, State, DoS",
  "n08138079": "Foggy Bottom",
  "n08138259": "Bureau of Diplomatic Security, DS",
  "n08138466": "Foreign Service",
  "n08138686": "Bureau of Intelligence and Research, INR",
  "n08139000": "Department of the Interior, Interior Department, Interior, DoI",
  "n08139270": "United States Fish and Wildlife Service, US Fish and Wildlife Service, FWS",
  "n08139637": "National Park Service",
  "n08139795": "Department of the Treasury, Treasury Department, Treasury, United States Treasury",
  "n08140219": "Bureau of Alcohol Tobacco and Firearms, ATF",
  "n08140506": "Financial Management Service",
  "n08140767": "Office of Intelligence Support, OIS",
  "n08141092": "Criminal Investigation Command, CID",
  "n08141374": "Drug Enforcement Administration, Drug Enforcement Agency, DEA",
  "n08141664": "Federal Bureau of Prisons, BoP",
  "n08141951": "Federal Judiciary",
  "n08142170": "National Institute of Justice, NIJ",
  "n08142370": "United States Marshals Service, US Marshals Service, Marshals",
  "n08142801": "Comptroller of the Currency",
  "n08142972": "Bureau of Customs, Customs Bureau, Customs Service, USCB",
  "n08143163": "Bureau of Engraving and Printing",
  "n08143321": "Internal Revenue Service, IRS",
  "n08143486": "Inland Revenue, IR",
  "n08143653": "Department of Transportation, Transportation, DoT",
  "n08143926": "Federal Aviation Agency, FAA",
  "n08144122": "Department of Veterans Affairs, VA",
  "n08144308": "Transportation Security Administration, TSA",
  "n08144524": "Department of Commerce and Labor",
  "n08144720": "Department of Health Education and Welfare",
  "n08144907": "Navy Department",
  "n08145092": "War Department",
  "n08145277": "United States Post Office, US Post Office, Post Office, PO",
  "n08145553": "post office, local post office",
  "n08145701": "general delivery, poste restante",
  "n08145871": "generally accepted accounting principles, GAAP",
  "n08146135": "instrumentality",
  "n08146410": "neonatal intensive care unit, NICU",
  "n08146593": "intensive care unit, ICU",
  "n08146782": "denomination",
  "n08147019": "communion",
  "n08147188": "Protestant denomination",
  "n08147670": "brethren",
  "n08147794": "order, monastic order",
  "n08148067": "Augustinian order",
  "n08148294": "Augustinian Canons",
  "n08148388": "Augustinian Hermits",
  "n08148512": "Austin Friars",
  "n08148601": "Benedictine order, order of Saint Benedict",
  "n08148827": "Carmelite order, Order of Our Lady of Mount Carmel",
  "n08149004": "Carthusian order",
  "n08149160": "Dominican order",
  "n08149314": "Franciscan order",
  "n08149473": "Society of Jesus, Jesuit order",
  "n08149781": "sect, religious sect, religious order",
  "n08150377": "Religious Society of Friends, Society of Friends, Quakers",
  "n08150576": "Shakers, United Society of Believers in Christ's Second Appearing",
  "n08150770": "Assemblies of God",
  "n08150892": "Waldenses, Vaudois",
  "n08151096": "Zurvanism",
  "n08151229": "cult",
  "n08151490": "cult",
  "n08151753": "cargo cult",
  "n08152069": "macumba",
  "n08152198": "obeah, obi",
  "n08152353": "Rastafarian",
  "n08152507": "voodoo",
  "n08152657": "sainthood",
  "n08152787": "clergy",
  "n08153022": "cardinalate",
  "n08153102": "laity, temporalty",
  "n08153337": "pantheon",
  "n08153437": "royalty, royal family, royal line, royal house",
  "n08153874": "Ordnance Survey",
  "n08154012": "Bourbon, Bourbon dynasty",
  "n08154213": "Capetian dynasty",
  "n08154363": "Carolingian dynasty, Carlovingian dynasty",
  "n08154548": "Flavian dynasty",
  "n08154703": "Han, Han dynasty",
  "n08154960": "Hanover, House of Hanover, Hanoverian line",
  "n08155302": "Habsburg, Hapsburg",
  "n08155518": "Hohenzollern",
  "n08155765": "Lancaster, House of Lancaster, Lancastrian line",
  "n08156054": "Liao, Liao dynasty",
  "n08156200": "Merovingian, Merovingian dynasty",
  "n08156392": "Ming, Ming dynasty",
  "n08156505": "Ottoman, Ottoman dynasty",
  "n08156685": "Plantagenet, Plantagenet line",
  "n08156948": "Ptolemy, Ptolemaic dynasty",
  "n08157182": "Qin, Qin dynasty, Ch'in, Ch'in dynasty",
  "n08157405": "Qing, Qing dynasty, Ch'ing, Ch'ing dynasty, Manchu, Manchu dynasty",
  "n08157672": "Romanov, Romanoff",
  "n08157809": "Saxe-Coburg-Gotha",
  "n08158089": "Seljuk",
  "n08158314": "Shang, Shang dynasty",
  "n08158460": "Stuart",
  "n08158741": "Sung, Sung dynasty, Song, Song dynasty",
  "n08158920": "Tang, Tang dynasty",
  "n08159031": "Tudor, House of Tudor",
  "n08159331": "Umayyad, Ommiad, Omayyad",
  "n08159464": "Valois",
  "n08159591": "Wei, Wei dynasty",
  "n08159740": "Windsor, House of Windsor",
  "n08159924": "York, House of York",
  "n08160146": "Yuan, Yuan dynasty, Mongol dynasty",
  "n08160276": "citizenry, people",
  "n08160586": "Achaean, Arcado-Cyprians",
  "n08160697": "Aeolian",
  "n08160808": "Dorian",
  "n08160947": "Ionian",
  "n08161068": "electorate",
  "n08161258": "governed",
  "n08161477": "senate",
  "n08161591": "United States Senate, U.S. Senate, US Senate, Senate",
  "n08161757": "Congress, United States Congress, U.S. Congress, US Congress",
  "n08161971": "United States House of Representatives, U.S. House of Representatives, US House of Representatives, House of Representatives, U.S. House, US House",
  "n08162245": "Government Accounting Office, GAO, United States Government Accounting Office",
  "n08162575": "House of Burgesses",
  "n08162691": "House of Commons, British House of Commons",
  "n08162860": "House of Lords, British House of Lords",
  "n08163025": "house",
  "n08163273": "legislature, legislative assembly, legislative body, general assembly, law-makers",
  "n08163702": "legislative council",
  "n08163792": "assembly",
  "n08164110": "Areopagus",
  "n08164261": "States General",
  "n08164464": "Estates General",
  "n08164585": "administration, governance, governing body, establishment, brass, organization, organisation",
  "n08165353": "top brass",
  "n08165455": "executive",
  "n08165650": "Bush administration",
  "n08165760": "Clinton administration",
  "n08165866": "Bush administration",
  "n08165979": "Reagan administration",
  "n08166083": "Carter administration",
  "n08166187": "judiciary, bench",
  "n08166318": "judiciary, judicature, judicatory, judicial system",
  "n08166552": "nation, land, country",
  "n08166931": "commonwealth country",
  "n08167046": "developing country",
  "n08167249": "Dominion",
  "n08167365": "estate of the realm, estate, the three estates",
  "n08167779": "first estate, Lords Spiritual",
  "n08167953": "second estate, Lords Temporal",
  "n08168117": "third estate, Commons",
  "n08168241": "fourth estate",
  "n08168367": "foreign country",
  "n08168531": "tribe, federation of tribes",
  "n08168737": "Free World",
  "n08168831": "Third World",
  "n08168978": "state, nation, country, land, commonwealth, res publica, body politic",
  "n08169573": "Reich",
  "n08169677": "Holy Roman Empire",
  "n08169919": "Hohenzollern empire, Second Reich",
  "n08170070": "Weimar Republic",
  "n08170255": "Third Reich, Nazi Germany",
  "n08170374": "rogue state, renegade state, rogue nation",
  "n08170535": "suzerain",
  "n08170686": "member",
  "n08170978": "allies",
  "n08171094": "bloc, axis",
  "n08171210": "Allies",
  "n08171415": "Central Powers",
  "n08171592": "Allies",
  "n08171792": "Axis",
  "n08171992": "entente, entente cordiale",
  "n08172103": "Arab League",
  "n08172695": "Europe",
  "n08172877": "Asia",
  "n08173030": "North America",
  "n08173165": "Central America",
  "n08173289": "South America",
  "n08173515": "European Union, EU, European Community, EC, European Economic Community, EEC, Common Market, Europe",
  "n08174167": "Supreme Headquarters Allied Powers Europe, SHAPE",
  "n08174398": "North Atlantic Treaty Organization, NATO",
  "n08174995": "Allied Command Atlantic, ACLANT",
  "n08175233": "Supreme Allied Commander Atlantic, SACLANT",
  "n08175498": "Allied Command Europe, ACE",
  "n08175700": "Supreme Allied Commander Europe, SACEUR",
  "n08175875": "Organization for the Prohibition of Chemical Weapons, OPCW",
  "n08176077": "Organization of American States, OAS",
  "n08176901": "Pan American Union",
  "n08177030": "Organization of Petroleum-Exporting Countries, OPEC",
  "n08177487": "sea power",
  "n08177592": "world power, major power, great power, power, superpower",
  "n08177781": "hegemon",
  "n08177863": "church-state",
  "n08177958": "city state, city-state",
  "n08178085": "welfare state",
  "n08178321": "puppet government, puppet state, pupet regime",
  "n08178547": "state",
  "n08178741": "population",
  "n08178935": "overpopulation",
  "n08179033": "overspill",
  "n08179205": "poor people, poor",
  "n08179455": "rich people, rich",
  "n08179689": "populace, public, world",
  "n08179879": "population",
  "n08180067": "home front",
  "n08180190": "multitude, masses, mass, hoi polloi, people, the great unwashed",
  "n08180484": "admass",
  "n08180639": "labor, labour, working class, proletariat",
  "n08180944": "labor force, labor pool",
  "n08181075": "lumpenproletariat",
  "n08181267": "organized labor",
  "n08181375": "Laurel and Hardy",
  "n08181540": "lower class, underclass",
  "n08181658": "middle class, bourgeoisie",
  "n08181820": "booboisie",
  "n08181930": "commonalty, commonality, commons",
  "n08182105": "petit bourgeois, petite bourgeoisie, petty bourgeoisie",
  "n08182283": "peasantry",
  "n08182379": "crowd",
  "n08182716": "multitude, throng, concourse",
  "n08182892": "hive",
  "n08182962": "horde, host, legion",
  "n08183046": "ruck, herd",
  "n08183290": "army",
  "n08183398": "crush, jam, press",
  "n08183541": "traffic jam, snarl-up",
  "n08183698": "gridlock",
  "n08183802": "host, legion",
  "n08183920": "Roman Legion",
  "n08184052": "Sabaoth",
  "n08184217": "drove, horde, swarm",
  "n08184335": "drove",
  "n08184439": "huddle",
  "n08184600": "mob, rabble, rout",
  "n08184730": "lynch mob",
  "n08184861": "company",
  "n08185063": "attendance",
  "n08185211": "limited company, Ltd., Ld.",
  "n08185369": "holding company",
  "n08185501": "bank holding company",
  "n08185643": "multibank holding company",
  "n08185758": "utility, public utility, public utility company, public-service corporation",
  "n08186047": "service",
  "n08186221": "telephone company, telephone service, phone company, phone service, telco",
  "n08186393": "power company, power service, electric company, light company",
  "n08186546": "water company, waterworks",
  "n08186655": "gas company, gas service",
  "n08186761": "bus company, bus service",
  "n08186898": "livery company",
  "n08187033": "company, troupe",
  "n08187370": "opera company",
  "n08187460": "theater company",
  "n08187569": "stock company, repertory company",
  "n08187707": "ballet company",
  "n08187837": "chorus",
  "n08187988": "chorus, Greek chorus",
  "n08188235": "ensemble",
  "n08188449": "chorus, chorus line",
  "n08188638": "choir",
  "n08188814": "choir, consort",
  "n08188967": "husking bee, cornhusking",
  "n08189089": "corps de ballet, ensemble",
  "n08189211": "circus",
  "n08189371": "minstrel show",
  "n08189581": "minstrelsy",
  "n08189659": "unit, social unit",
  "n08190292": "command",
  "n08190482": "enemy",
  "n08190609": "task force",
  "n08190754": "army unit",
  "n08191230": "army, regular army, ground forces",
  "n08191532": "naval unit",
  "n08191701": "navy, naval forces",
  "n08191987": "United States Navy, US Navy, USN, Navy",
  "n08192361": "coastguard",
  "n08192557": "United States Coast Guard, U. S. Coast Guard, US Coast Guard",
  "n08192817": "Marines",
  "n08192970": "United States Marine Corps, United States Marines, Marine Corps, US Marine Corps, USMC",
  "n08193212": "Naval Air Warfare Center Weapons Division, NAWCWPNS",
  "n08193448": "Naval Special Warfare, NSW",
  "n08193645": "Naval Surface Warfare Center, NSWC",
  "n08193854": "Naval Underwater Warfare Center, NUWC",
  "n08194074": "United States Naval Academy, US Naval Academy",
  "n08194266": "Office of Naval Intelligence, ONI",
  "n08194546": "Marine Corps Intelligence Activity, MCIA",
  "n08194815": "Air Corps",
  "n08194927": "United States Air Force Academy, US Air Force Academy",
  "n08195132": "Royal Air Force, RAF",
  "n08195229": "Luftwaffe, German Luftwaffe",
  "n08195323": "League of Nations",
  "n08195636": "Peace Corps",
  "n08195797": "air unit",
  "n08196024": "air force, airforce",
  "n08196230": "United States Air Force, U. S. Air Force, US Air Force, Air Force, USAF",
  "n08196622": "Air Combat Command, ACC",
  "n08196892": "Air Force Space Command, AFSPC",
  "n08197149": "Air National Guard, ANG",
  "n08197386": "Air Force Intelligence Surveillance and Reconnaissance, Air Force ISR, AFISR",
  "n08197742": "armor, armour",
  "n08197895": "guerrilla force, guerilla force",
  "n08198137": "military service, armed service, service",
  "n08198398": "military unit, military force, military group, force",
  "n08199025": "military, armed forces, armed services, military machine, war machine",
  "n08206460": "military reserve, reserve",
  "n08206663": "mujahidin, mujahedin, mujahedeen, mujahadeen, mujahadin, mujahideen, mujahadein",
  "n08207001": "Mujahedeen Khalq",
  "n08207095": "Pentagon",
  "n08207209": "paramilitary, paramilitary force, paramilitary unit, paramilitary organization, paramilitary organisation",
  "n08207540": "fedayeen",
  "n08207672": "Fedayeen Saddam, Saddam's Martyrs",
  "n08207863": "force",
  "n08208016": "force, personnel",
  "n08208444": "task force",
  "n08208560": "team, squad",
  "n08209076": "hit squad",
  "n08209187": "death squad",
  "n08209386": "Sparrow Unit",
  "n08209519": "bench",
  "n08209687": "police, police force, constabulary, law",
  "n08210042": "Europol, European Law Enforcement Organisation",
  "n08210254": "gendarmerie, gendarmery",
  "n08210411": "Mutawa'een, Mutawa",
  "n08210670": "Royal Canadian Mounted Police, RCMP, Mounties",
  "n08210835": "Scotland Yard, New Scotland Yard",
  "n08210982": "security force, private security force",
  "n08211169": "vice squad",
  "n08211290": "military police, MP",
  "n08211475": "shore patrol",
  "n08211584": "secret police",
  "n08211760": "Gestapo",
  "n08211924": "Schutzstaffel, SS",
  "n08212146": "SA, Sturmabteilung, Storm Troops",
  "n08212347": "work force, workforce, manpower, hands, men",
  "n08212527": "corps, army corps",
  "n08212740": "Women's Army Corps, WAC",
  "n08212920": "Reserve Officers Training Corps, ROTC",
  "n08213079": "corps",
  "n08213205": "division",
  "n08213424": "Special Forces, U. S. Army Special Forces, United States Army Special Forces",
  "n08213671": "battle group",
  "n08213817": "regiment",
  "n08213978": "brigade",
  "n08214083": "battalion",
  "n08214272": "company",
  "n08214470": "platoon",
  "n08214698": "platoon",
  "n08214832": "section",
  "n08214966": "den",
  "n08215044": "platoon",
  "n08215248": "detachment",
  "n08215495": "vanguard, van",
  "n08215603": "guard",
  "n08215801": "bodyguard",
  "n08215989": "yeomanry",
  "n08216176": "patrol",
  "n08216295": "picket",
  "n08216408": "press gang",
  "n08216535": "provost guard",
  "n08216647": "rearguard",
  "n08216795": "section",
  "n08216900": "section",
  "n08217135": "brass section, brass",
  "n08217302": "string section, strings",
  "n08217450": "violin section",
  "n08217557": "percussion section, percussion, rhythm section",
  "n08217738": "trumpet section",
  "n08217865": "reed section",
  "n08218004": "clarinet section",
  "n08218122": "squad",
  "n08218212": "complement, full complement",
  "n08218393": "shift",
  "n08218592": "day shift, day watch",
  "n08218713": "evening shift",
  "n08218832": "night shift, graveyard shift",
  "n08218965": "relay",
  "n08219059": "ship's company, company",
  "n08219226": "division, naval division",
  "n08219330": "division, air division",
  "n08219493": "wing",
  "n08219628": "air group",
  "n08219768": "squadron",
  "n08219923": "escadrille",
  "n08220089": "squadron",
  "n08220228": "squadron",
  "n08220367": "escadrille",
  "n08220440": "flight",
  "n08220534": "flight",
  "n08220620": "flight",
  "n08220714": "division",
  "n08220891": "division",
  "n08221348": "division",
  "n08221747": "form division",
  "n08221897": "audience",
  "n08222203": "gallery",
  "n08222293": "audience",
  "n08222571": "readership",
  "n08222750": "viewing audience, TV audience, viewers",
  "n08222871": "grandstand",
  "n08222966": "house",
  "n08223137": "claque",
  "n08223263": "following, followers",
  "n08223475": "faithful",
  "n08223581": "fandom",
  "n08223688": "parish",
  "n08223802": "community",
  "n08224274": "community",
  "n08224413": "convent",
  "n08224580": "house",
  "n08224684": "Ummah, Umma, Muslim Ummah, Islamic Ummah, Islam Nation",
  "n08224985": "speech community",
  "n08225090": "neighborhood, neighbourhood",
  "n08225334": "hood",
  "n08225426": "street",
  "n08225581": "municipality",
  "n08225736": "municipal government",
  "n08225853": "commission plan",
  "n08226018": "state government",
  "n08226127": "totalitarian state, totalitation regime",
  "n08226335": "city, metropolis",
  "n08226514": "town, townspeople, townsfolk",
  "n08226699": "village, small town, settlement",
  "n08226838": "moshav",
  "n08226978": "hamlet, crossroads",
  "n08227088": "cooperative",
  "n08227214": "club, social club, society, guild, gild, lodge, order",
  "n08227916": "family, fellowship",
  "n08228229": "koinonia",
  "n08228405": "athenaeum, atheneum",
  "n08228538": "bookclub",
  "n08228665": "chapter",
  "n08228838": "chapter",
  "n08228980": "American Legion",
  "n08229134": "Veterans of Foreign Wars, VFW",
  "n08229275": "chess club",
  "n08229362": "country club",
  "n08229467": "fraternity, frat",
  "n08229605": "glee club",
  "n08229694": "golf club",
  "n08229779": "hunt, hunt club",
  "n08229887": "investors club",
  "n08230009": "jockey club",
  "n08230110": "racket club",
  "n08230219": "rowing club",
  "n08230294": "slate club",
  "n08230477": "sorority",
  "n08230590": "tennis club",
  "n08230679": "turnverein",
  "n08230785": "yacht club, boat club",
  "n08230906": "yakuza",
  "n08231065": "yoke",
  "n08231184": "league, conference",
  "n08231499": "major league, big league, majors",
  "n08231678": "minor league, minors, bush league",
  "n08231874": "baseball league",
  "n08231999": "little league",
  "n08232160": "little-league team",
  "n08232299": "basketball league",
  "n08232410": "bowling league",
  "n08232496": "football league",
  "n08232603": "hockey league",
  "n08232706": "Ivy League",
  "n08233056": "union, labor union, trade union, trades union, brotherhood",
  "n08233426": "industrial union, vertical union",
  "n08233627": "Teamsters Union",
  "n08233762": "United Mine Workers of America, United Mine Workers",
  "n08233913": "American Federation of Labor, AFL",
  "n08234103": "American Federation of Labor and Congress of Industrial Organizations, AFL-CIO",
  "n08234298": "Congress of Industrial Organizations, CIO",
  "n08234493": "craft union",
  "n08234628": "credit union",
  "n08234792": "company union",
  "n08234935": "open shop",
  "n08235076": "closed shop",
  "n08235173": "union shop",
  "n08235343": "secret society",
  "n08235513": "Freemasonry, Masonry",
  "n08235623": "Rashtriya Swayamsevak Sangh, National Volunteers Association",
  "n08235828": "service club",
  "n08236027": "Lions Club, International Association of Lions clubs",
  "n08236251": "Rotary Club, Rotary International",
  "n08236438": "consortium, pool, syndicate",
  "n08236621": "trust, corporate trust, combine, cartel",
  "n08236963": "drug cartel",
  "n08237202": "Medellin cartel",
  "n08237400": "Cali cartel",
  "n08237699": "oil cartel",
  "n08237863": "cast, cast of characters, dramatis personae",
  "n08238048": "ensemble, supporting players",
  "n08238156": "constituency",
  "n08238313": "electoral college",
  "n08238463": "class, form, grade, course",
  "n08238660": "class, year",
  "n08238909": "graduating class",
  "n08239024": "master class",
  "n08239152": "section, discussion section",
  "n08239383": "senior class",
  "n08239488": "junior class",
  "n08239590": "sophomore class",
  "n08239706": "freshman class",
  "n08239808": "class, division",
  "n08240022": "revolving door",
  "n08240169": "set, circle, band, lot",
  "n08240484": "car pool",
  "n08240633": "clique, coterie, ingroup, inner circle, pack, camp",
  "n08240966": "Bloomsbury Group",
  "n08241309": "bohemia",
  "n08241512": "kitchen cabinet, brain trust",
  "n08241654": "loop",
  "n08241798": "cabal, faction, junto, camarilla",
  "n08241964": "military junta, junta",
  "n08242100": "cadre",
  "n08242223": "core, nucleus, core group",
  "n08242428": "portfolio",
  "n08242675": "professional association",
  "n08242799": "gang, crew, work party",
  "n08243081": "detail",
  "n08243248": "chain gang",
  "n08243340": "ground crew, ground-service crew",
  "n08243492": "road gang",
  "n08243570": "section gang",
  "n08243695": "stage crew",
  "n08243851": "Fabian Society",
  "n08244062": "gang, pack, ring, mob",
  "n08244346": "nest",
  "n08244532": "sleeper nest",
  "n08244747": "youth gang",
  "n08244895": "demimonde",
  "n08245059": "underworld",
  "n08245172": "organized crime, gangland, gangdom",
  "n08245425": "mafia, maffia",
  "n08245549": "Mafia, Maffia, Sicilian Mafia",
  "n08245802": "Mafia, Maffia, Cosa Nostra",
  "n08246036": "Black Hand",
  "n08246185": "Camorra",
  "n08246302": "syndicate, crime syndicate, mob, family",
  "n08246502": "yeomanry",
  "n08246613": "musical organization, musical organisation, musical group",
  "n08247021": "duet, duette, duo",
  "n08247152": "trio",
  "n08247251": "quartet, quartette",
  "n08247400": "barbershop quartet",
  "n08247561": "string quartet, string quartette",
  "n08247703": "quintet, quintette",
  "n08247816": "sextet, sextette, sestet",
  "n08247935": "septet, septette",
  "n08248047": "octet, octette",
  "n08248157": "orchestra",
  "n08248411": "chamber orchestra",
  "n08248521": "gamelan",
  "n08248747": "string orchestra",
  "n08248856": "symphony orchestra, symphony, philharmonic",
  "n08249038": "band",
  "n08249207": "marching band",
  "n08249336": "brass band",
  "n08249459": "concert band, military band",
  "n08249608": "jug band",
  "n08249752": "pop group",
  "n08249854": "indie",
  "n08249960": "dance band, band, dance orchestra",
  "n08250168": "big band",
  "n08250302": "jazz band, jazz group, combo",
  "n08250409": "mariachi",
  "n08250501": "rock group, rock band",
  "n08250635": "skiffle group",
  "n08250750": "steel band",
  "n08250889": "horde",
  "n08250978": "Golden Horde",
  "n08251104": "cohort",
  "n08251213": "cohort",
  "n08251303": "conspiracy, confederacy",
  "n08251493": "Four Hundred",
  "n08251605": "horsy set, horsey set",
  "n08251756": "jet set",
  "n08251877": "faction, sect",
  "n08252083": "splinter group",
  "n08252211": "social gathering, social affair",
  "n08252467": "function",
  "n08252602": "party",
  "n08253045": "shindig, shindy",
  "n08253141": "dance",
  "n08253268": "ball",
  "n08253450": "masquerade, masquerade party, masque, mask",
  "n08253640": "banquet, feast",
  "n08253815": "dinner, dinner party",
  "n08254055": "gaudy",
  "n08254195": "beanfeast",
  "n08254331": "reception",
  "n08254540": "at home",
  "n08254627": "levee",
  "n08254741": "tea",
  "n08254876": "wedding reception",
  "n08254993": "open house",
  "n08255109": "housewarming",
  "n08255231": "soiree",
  "n08255373": "musical soiree, soiree musicale",
  "n08255508": "garden party, lawn party, fete champetre",
  "n08255660": "bachelor party",
  "n08255795": "shower",
  "n08256005": "stag party, smoker",
  "n08256151": "hen party",
  "n08256229": "slumber party",
  "n08256369": "sociable, social, mixer",
  "n08256567": "supper",
  "n08256735": "wedding, wedding party",
  "n08256968": "party, political party",
  "n08257779": "American Labor Party",
  "n08257971": "American Party, Know-Nothing Party",
  "n08258184": "Anti-Masonic Party",
  "n08258358": "Black Panthers",
  "n08258523": "Communist Party",
  "n08258744": "Conservative Party",
  "n08258974": "Constitutional Union Party",
  "n08259156": "Democratic Party",
  "n08259318": "Democratic-Republican Party",
  "n08259611": "Farmer-Labor Party",
  "n08259753": "Federalist Party, American Federalist Party, Federal Party",
  "n08260002": "Free Soil Party",
  "n08260220": "Gironde",
  "n08260386": "Green Party",
  "n08260498": "Greenback Party",
  "n08260691": "Kuomintang, Guomindang",
  "n08260961": "labor party, labour party",
  "n08261162": "Australian Labor Party",
  "n08261320": "British Labour Party, Labour Party, Labour, Labor",
  "n08261589": "Liberal Democrat Party",
  "n08261810": "Liberal Party",
  "n08261958": "Liberty Party",
  "n08262154": "Militant Tendency",
  "n08262309": "National Socialist German Workers' Party, Nazi Party",
  "n08262494": "People's Party, Populist Party",
  "n08262695": "Progressive Party, Bull Moose Party",
  "n08262937": "Prohibition Party",
  "n08263113": "Republican Party, GOP",
  "n08263303": "Social Democratic Party",
  "n08263560": "Socialist Labor Party",
  "n08263733": "Socialist Party",
  "n08263870": "States' Rights Democratic Party, Dixiecrats",
  "n08264110": "war party",
  "n08264203": "Whig Party",
  "n08264441": "third party",
  "n08264583": "machine, political machine",
  "n08264759": "machine",
  "n08264897": "party, company",
  "n08265252": "fatigue party",
  "n08265347": "landing party",
  "n08265469": "party to the action, party to the transaction",
  "n08265614": "rescue party",
  "n08265692": "search party",
  "n08265790": "stretcher party",
  "n08265912": "war party",
  "n08266070": "professional organization, professional organisation",
  "n08266235": "table, tabular array",
  "n08266542": "actuarial table, statistical table",
  "n08266669": "mortality table",
  "n08266849": "calendar",
  "n08267008": "perpetual calendar",
  "n08267197": "file allocation table",
  "n08267446": "periodic table",
  "n08267640": "matrix",
  "n08267956": "dot matrix",
  "n08268085": "square matrix",
  "n08268321": "diagonal",
  "n08268636": "main diagonal, principal diagonal",
  "n08268807": "secondary diagonal",
  "n08268962": "diagonal matrix",
  "n08269114": "scalar matrix",
  "n08269257": "identity matrix, unit matrix",
  "n08269396": "determinant",
  "n08269506": "Latin square",
  "n08269707": "magic square",
  "n08269939": "nonsingular matrix",
  "n08270067": "real matrix",
  "n08270169": "singular matrix",
  "n08270290": "transpose",
  "n08270417": "diagonal",
  "n08270662": "Oort cloud",
  "n08270938": "galaxy",
  "n08271042": "galaxy, extragalactic nebula",
  "n08271457": "spiral galaxy, spiral nebula",
  "n08271656": "Andromeda galaxy",
  "n08271801": "legion",
  "n08271941": "foreign legion",
  "n08272086": "French Foreign Legion",
  "n08272248": "legion",
  "n08272352": "echelon",
  "n08272460": "phalanx",
  "n08272564": "phalanx",
  "n08272652": "score",
  "n08272774": "threescore",
  "n08272860": "synset",
  "n08272961": "combination",
  "n08273167": "crew",
  "n08273406": "aircrew, air crew",
  "n08273531": "bomber crew, bomber aircrew",
  "n08273645": "merchant marine",
  "n08273736": "crew",
  "n08273843": "crowd, crew, gang, bunch",
  "n08274004": "shock troops",
  "n08274126": "SWAT team, SWAT squad, Special Weapons and Tactics team, Special Weapons and Tactics squad",
  "n08274354": "troop",
  "n08274443": "troop",
  "n08274565": "troop, flock",
  "n08274718": "troop, scout troop, scout group",
  "n08274826": "outfit",
  "n08274923": "academia, academe",
  "n08275081": "Grub Street",
  "n08275185": "school",
  "n08275497": "Ashcan School, Eight",
  "n08275704": "deconstructivism",
  "n08275843": "historical school",
  "n08276053": "pointillism",
  "n08276342": "educational institution",
  "n08276539": "preschool",
  "n08276720": "school",
  "n08277393": "school",
  "n08277612": "junior school",
  "n08277709": "infant school",
  "n08277805": "academy",
  "n08278032": "yeshiva, yeshivah",
  "n08278169": "college",
  "n08278324": "college",
  "n08278589": "correspondence school",
  "n08278707": "crammer",
  "n08278822": "dancing school",
  "n08278924": "direct-grant school",
  "n08279184": "driving school",
  "n08279298": "academy",
  "n08279424": "police academy",
  "n08279524": "military academy",
  "n08279665": "naval academy",
  "n08279800": "air force academy",
  "n08279943": "Plato's Academy",
  "n08280124": "academy, honorary society",
  "n08280462": "Academy of Motion Picture Arts and Sciences",
  "n08280649": "Academy of Television Arts and Sciences",
  "n08280813": "French Academy",
  "n08281029": "National Academy of Sciences",
  "n08281205": "Royal Academy, Royal Academy of Arts",
  "n08281409": "Royal Society, Royal Society of London for Improving Natural Knowledge",
  "n08281682": "business college",
  "n08281812": "business school",
  "n08281963": "dental school, school of dentistry",
  "n08282109": "finishing school",
  "n08282257": "flying school",
  "n08282364": "junior college",
  "n08282542": "community college",
  "n08282696": "graduate school, grad school",
  "n08282950": "language school",
  "n08283051": "law school, school of law",
  "n08283180": "madrasa, madrasah",
  "n08283507": "medical school, school of medicine",
  "n08283649": "music school, school of music",
  "n08283757": "nursing school, school of nursing",
  "n08283866": "pesantran, pesantren",
  "n08284054": "religious school",
  "n08284221": "church school, parochial school",
  "n08284370": "riding school",
  "n08284481": "secondary school, lyceum, lycee, Gymnasium, middle school",
  "n08284847": "secretarial school",
  "n08284994": "seminary",
  "n08285109": "seminary",
  "n08285246": "technical school, tech",
  "n08285407": "polytechnic institute, polytechnic, engineering school",
  "n08285594": "trade school, vocational school",
  "n08285719": "trainband",
  "n08285896": "training college",
  "n08286039": "training school",
  "n08286163": "university",
  "n08286342": "gown",
  "n08286569": "university",
  "n08286801": "multiversity",
  "n08286946": "Open University",
  "n08287182": "varsity",
  "n08287335": "veterinary school",
  "n08287436": "conservatory",
  "n08287586": "staff, faculty",
  "n08287844": "culture, civilization, civilisation",
  "n08288189": "open society",
  "n08288390": "tribal society",
  "n08288518": "hunting and gathering tribe, hunting and gathering society",
  "n08288753": "subculture",
  "n08288977": "suburbia",
  "n08289089": "youth culture",
  "n08289282": "hip-hop",
  "n08289449": "youth subculture",
  "n08289841": "flower people, hippies, hipsters",
  "n08290156": "Aegean civilization, Aegean civilisation, Aegean culture",
  "n08290435": "Helladic civilization, Helladic civilisation, Helladic culture",
  "n08290616": "Indus civilization",
  "n08290763": "Minoan civilization, Minoan civilisation, Minoan culture",
  "n08290928": "Cycladic civilization, Cycladic civilisation, Cycladic culture, Cyclades",
  "n08291157": "Mycenaean civilization, Mycenaean civilisation, Mycenaean culture",
  "n08291338": "Paleo-American culture, Paleo-Amerind culture, Paleo-Indian culture",
  "n08291582": "Clovis culture",
  "n08291813": "Folsom culture",
  "n08292051": "Western culture, Western civilization",
  "n08292298": "psychedelia",
  "n08292418": "Rastafari, Rastas",
  "n08292756": "fleet",
  "n08293003": "armada",
  "n08293088": "Spanish Armada, Invincible Armada",
  "n08293238": "battle fleet",
  "n08293336": "fleet",
  "n08293490": "fleet",
  "n08293684": "motor pool",
  "n08293831": "fleet",
  "n08293982": "alliance, coalition, alignment, alinement",
  "n08294395": "nonalignment, nonalinement",
  "n08294581": "popular front",
  "n08294696": "world organization, world organisation, international organization, international organisation, global organization",
  "n08295138": "Commonwealth of Independent States, CIS",
  "n08295580": "United Nations, UN",
  "n08295894": "deliberative assembly",
  "n08296059": "General Assembly",
  "n08296196": "United Nations Secretariat",
  "n08296332": "Security Council, SC",
  "n08296500": "Trusteeship Council, TC",
  "n08296720": "Economic and Social Council, ECOSOC",
  "n08296911": "Economic and Social Council commission, ECOSOC commission",
  "n08297284": "Commission on Human Rights",
  "n08297464": "Commission on Narcotic Drugs",
  "n08297646": "Commission on the Status of Women",
  "n08297863": "Economic Commission for Africa",
  "n08298074": "Economic Commission for Asia and the Far East",
  "n08298319": "Economic Commission for Europe",
  "n08298521": "Economic Commission for Latin America",
  "n08298737": "Population Commission",
  "n08298918": "Social Development Commission",
  "n08299107": "Statistical Commission",
  "n08299307": "International Court of Justice, World Court",
  "n08299493": "United Nations agency, UN agency",
  "n08299911": "United Nations Children's Fund, United Nations International Children's Emergency Fund, UNICEF",
  "n08300190": "Food and Agriculture Organization, Food and Agriculture Organization of the United Nations, FAO",
  "n08300429": "General Agreement on Tariffs and Trade, GATT",
  "n08300641": "International Atomic Energy Agency, IAEA",
  "n08300783": "International Bank for Reconstruction and Development, World Bank, IBRD",
  "n08301005": "International Civil Aviation Organization, ICAO",
  "n08301155": "International Development Association, IDA",
  "n08301307": "International Finance Corporation, IFC",
  "n08301525": "International Labor Organization, International Labour Organization, ILO",
  "n08301709": "International Maritime Organization, IMO",
  "n08301871": "International Monetary Fund, IMF",
  "n08302052": "United Nations Educational Scientific and Cultural Organization, UNESCO",
  "n08302257": "United Nations Office for Drug Control and Crime Prevention, DCCP",
  "n08302468": "United Nations Crime Prevention and Criminal Justice, Centre for International Crime Prevention",
  "n08302724": "World Health Organization, WHO",
  "n08302919": "World Meteorological Organization, WMO",
  "n08303097": "sterling area, sterling bloc, scheduled territories",
  "n08303275": "confederation, confederacy, federation",
  "n08303504": "federation",
  "n08303692": "nation",
  "n08303862": "Creek Confederacy",
  "n08304135": "Hanseatic League",
  "n08304744": "enosis",
  "n08304895": "union",
  "n08305114": "league",
  "n08305277": "Iroquois League, League of Iroquois, Five Nations, Six Nations",
  "n08305568": "customs union",
  "n08305766": "Benelux",
  "n08305942": "ally",
  "n08306047": "caste",
  "n08306194": "caste",
  "n08306371": "jati",
  "n08306665": "varna",
  "n08306959": "brahman, brahmin",
  "n08307145": "rajanya",
  "n08307267": "vaisya",
  "n08307440": "sudra, shudra",
  "n08307589": "meeting, group meeting",
  "n08308039": "board meeting, committee meeting",
  "n08308156": "camp meeting",
  "n08308313": "caucus",
  "n08308410": "conclave",
  "n08308497": "conference",
  "n08308800": "congress",
  "n08308922": "Congress of Racial Equality, CORE",
  "n08309086": "convention",
  "n08309226": "Constitutional Convention",
  "n08309409": "council",
  "n08309611": "encounter group",
  "n08309754": "forum",
  "n08309854": "plenum",
  "n08310034": "psychotherapy group",
  "n08310168": "stockholders meeting",
  "n08310309": "covey",
  "n08310389": "meeting, get together",
  "n08310722": "North Atlantic Council, NAC",
  "n08310949": "council",
  "n08311282": "city council",
  "n08311409": "executive council",
  "n08311522": "panchayat, panchayet, punchayet",
  "n08311687": "privy council",
  "n08311848": "divan, diwan",
  "n08311933": "works council",
  "n08312241": "town meeting",
  "n08312359": "summit, summit meeting",
  "n08312462": "town meeting",
  "n08312559": "council",
  "n08312988": "ecumenical council",
  "n08313592": "Nicaea, First Council of Nicaea",
  "n08313790": "Constantinople, First Council of Constantinople",
  "n08313983": "Ephesus, Council of Ephesus",
  "n08314153": "Chalcedon, Council of Chalcedon",
  "n08314327": "Constantinople, Second Council of Constantinople",
  "n08314501": "Constantinople, Third Council of Constantinople",
  "n08314715": "Nicaea, Second Council of Nicaea",
  "n08314901": "Constantinople, Fourth Council of Constantinople",
  "n08315194": "Lateran Council",
  "n08315442": "First Lateran Council",
  "n08315682": "Second Lateran Council",
  "n08315860": "Third Lateran Council",
  "n08316019": "Fourth Lateran Council",
  "n08316346": "Lyons, First Council of Lyons",
  "n08316564": "Lyons, Second Council of Lyons",
  "n08316748": "Vienne, Council of Vienne",
  "n08316965": "Constance, Council of Constance",
  "n08317139": "Council of Basel-Ferrara-Florence",
  "n08317340": "Fifth Lateran Council",
  "n08317529": "Council of Trent",
  "n08317885": "Vatican Council",
  "n08318032": "First Vatican Council, Vatican I",
  "n08318216": "Second Vatican Council, Vatican II",
  "n08318423": "Continental Congress",
  "n08318691": "congress",
  "n08318777": "diet",
  "n08318904": "chamber",
  "n08319061": "chamber of commerce",
  "n08319198": "parliament",
  "n08319408": "British Parliament",
  "n08319539": "Dail Eireann, Dail",
  "n08319679": "Knesset, Knesseth",
  "n08319777": "Oireachtas",
  "n08319908": "Seanad Eireann, Seanad",
  "n08320052": "Duma",
  "n08320201": "soviet",
  "n08320385": "Palestine Liberation Organization, PLO",
  "n08320923": "Palestine National Authority, Palestinian National Authority, Palestine Authority",
  "n08321218": "Sinn Fein",
  "n08321469": "Red Guard",
  "n08321621": "syndicalism",
  "n08321807": "indaba",
  "n08321956": "Jirga",
  "n08322187": "Loya Jirga",
  "n08322430": "powwow",
  "n08322520": "synod",
  "n08322625": "world council",
  "n08322756": "blue ribbon commission, blue ribbon committee",
  "n08322981": "board",
  "n08323301": "appeal board, appeals board, board of appeals",
  "n08323471": "board of selectmen",
  "n08323658": "board of regents",
  "n08323825": "board of trustees",
  "n08323980": "Federal Reserve Board",
  "n08324107": "governing board",
  "n08324274": "secretariat, secretariate",
  "n08324514": "committee, commission",
  "n08325124": "election commission",
  "n08325237": "fairness commission",
  "n08325386": "planning commission",
  "n08325530": "conservancy",
  "n08325686": "committee, citizens committee",
  "n08325851": "select committee",
  "n08325993": "subcommittee",
  "n08326114": "vigilance committee",
  "n08326266": "welcoming committee",
  "n08326383": "standing committee",
  "n08326487": "Ways and Means Committee",
  "n08326706": "steering committee",
  "n08326850": "ethics committee, ethics panel",
  "n08326976": "finance committee",
  "n08327090": "politburo",
  "n08327212": "political action committee, PAC",
  "n08327390": "presidium, praesidium",
  "n08327616": "symposium",
  "n08327816": "seminar",
  "n08327906": "colloquium",
  "n08328088": "Potsdam Conference",
  "n08328327": "Yalta Conference",
  "n08328563": "research colloquium",
  "n08328700": "Bench",
  "n08328876": "border patrol",
  "n08329010": "harbor patrol",
  "n08329113": "patrol",
  "n08329322": "court, royal court",
  "n08329453": "court, tribunal, judicature",
  "n08330106": "appellate court, appeals court, court of appeals",
  "n08330298": "circuit court of appeals",
  "n08330514": "circuit",
  "n08330843": "assizes, court of assize, court of assize and nisi prius",
  "n08331011": "chancery, court of chancery",
  "n08331121": "consistory",
  "n08331213": "criminal court",
  "n08331357": "drumhead court-martial",
  "n08331525": "court-martial",
  "n08331762": "special court-martial",
  "n08331960": "divorce court",
  "n08332090": "family court, domestic relations court, court of domestic relations",
  "n08332330": "federal court",
  "n08332485": "Foreign Intelligence Surveillance Court, F.I.S.C.",
  "n08332896": "inferior court, lower court",
  "n08333030": "Inquisition",
  "n08333224": "Spanish Inquisition",
  "n08333639": "Roman Inquisition, Congregation of the Inquisition",
  "n08333868": "juvenile court",
  "n08333995": "kangaroo court",
  "n08334087": "military court",
  "n08334319": "moot court",
  "n08334451": "night court",
  "n08334581": "Old Bailey",
  "n08334693": "provost court",
  "n08334873": "police court",
  "n08335087": "probate court",
  "n08335253": "quarter sessions",
  "n08335414": "Rota",
  "n08335599": "Star Chamber",
  "n08335751": "superior court",
  "n08335886": "Supreme Court, Supreme Court of the United States, United States Supreme Court",
  "n08336188": "supreme court, state supreme court, high court",
  "n08336355": "traffic court",
  "n08336490": "trial court",
  "n08336627": "repertoire, repertory",
  "n08336844": "repertory, repertoire",
  "n08337108": "representation",
  "n08337324": "agency, federal agency, government agency, bureau, office, authority",
  "n08338847": "independent agency",
  "n08339454": "intelligence, intelligence service, intelligence agency",
  "n08339706": "military intelligence, military intelligence agency",
  "n08339939": "United States intelligence agency",
  "n08340153": "Intelligence Community, National Intelligence Community, United States Intelligence Community, IC",
  "n08340753": "Advanced Research and Development Activity, ARDA",
  "n08340989": "Defense Intelligence Agency, DIA",
  "n08341330": "Defense Logistics Agency",
  "n08341551": "Defense Reutilization and Marketing Service, DRMS",
  "n08341798": "Defense Technical Information Center, DTIC",
  "n08342039": "international intelligence agency",
  "n08342419": "Canadian Security Intelligence Service, CSIS",
  "n08342670": "Central Intelligence Machinery, CIM",
  "n08342888": "Communications Security Establishment, CSE",
  "n08343102": "Criminal Intelligence Services of Canada, CISC",
  "n08343324": "Department of Justice Canada, DoJC",
  "n08343534": "Directorate for Inter-Services Intelligence, Inter-Services Intelligence, ISI",
  "n08343905": "Foreign Intelligence Service, Sluzhba Vneshney Razvedki, SVR",
  "n08344301": "International Relations and Security Network, ISN",
  "n08344551": "international law enforcement agency",
  "n08344756": "Interpol",
  "n08344917": "Iraqi Intelligence Service, IIS, Iraqi Mukhabarat",
  "n08345189": "Republican Guard",
  "n08345366": "Haganah",
  "n08345613": "Israeli Defense Force, IDF",
  "n08345770": "Sayeret Matkal, Sayeret Mat'kal, sayeret",
  "n08346031": "Special Air Service, SAS",
  "n08346286": "A'man",
  "n08346490": "Mossad",
  "n08346655": "Secret Intelligence Service, MI, Military Intelligence Section 6",
  "n08346905": "Security Intelligence Review Committee, SIRC",
  "n08347206": "Security Service, MI, Military Intelligence Section 5",
  "n08347457": "Shin Bet, General Security Services",
  "n08347704": "National Reconnaissance Office, NRO",
  "n08348091": "National Security Agency, NSA",
  "n08348400": "United States Secret Service, US Secret Service, USSS, Secret Service, SS",
  "n08348815": "law enforcement agency",
  "n08349138": "Occupational Safety and Health Administration, OSHA",
  "n08349350": "organ",
  "n08349548": "admiralty",
  "n08349681": "Patent and Trademark Office Database, Patent Office",
  "n08349916": "central bank",
  "n08350244": "European Central Bank",
  "n08350470": "Federal Reserve System, Federal Reserve, Fed, FRS",
  "n08350919": "Federal Reserve Bank, reserve bank",
  "n08351107": "Federal Trade Commission, FTC",
  "n08351380": "Office of Inspector General, OIG",
  "n08351532": "General Services Administration, GSA",
  "n08351777": "Federal Protective Service, FPS",
  "n08352035": "Bank of England",
  "n08352134": "Bundesbank",
  "n08352218": "Bank of Japan",
  "n08352303": "office, office staff",
  "n08352473": "research staff",
  "n08352613": "sales staff",
  "n08352720": "security staff",
  "n08352848": "service staff, maintenance staff",
  "n08352994": "Small Business Administration, SBA",
  "n08353244": "redevelopment authority",
  "n08353406": "regulatory agency, regulatory authority",
  "n08353563": "Selective Service, Selective Service System, SSS",
  "n08353734": "weather bureau",
  "n08353912": "advertising agency, ad agency",
  "n08354065": "credit bureau",
  "n08354243": "detective agency",
  "n08354352": "employment agency, employment office",
  "n08354523": "placement office, placement center",
  "n08354678": "hiring hall",
  "n08354842": "mercantile agency, commercial agency",
  "n08355075": "news agency, press agency, wire service, press association, news organization, news organisation",
  "n08355324": "syndicate",
  "n08355506": "service agency, service bureau, service firm",
  "n08355692": "travel agency",
  "n08355791": "United States government, United States, U.S. government, US Government, U.S.",
  "n08356074": "executive branch, Executive Office of the President",
  "n08356375": "legislative branch",
  "n08356573": "United States Government Printing Office, US Government Printing Office, Government Printing Office, GPO",
  "n08356903": "judicial branch",
  "n08357129": "Capital, Washington",
  "n08357258": "civil service",
  "n08357448": "Whitehall",
  "n08357529": "county council",
  "n08357647": "diplomatic service, diplomatic corps, corps diplomatique",
  "n08357784": "government officials, officialdom",
  "n08357992": "quorum",
  "n08358165": "minyan",
  "n08358332": "rally, mass meeting",
  "n08358492": "pep rally",
  "n08358594": "cell, cadre",
  "n08358824": "sleeper cell",
  "n08358963": "terrorist cell, radical cell",
  "n08359316": "operational cell",
  "n08359432": "intelligence cell",
  "n08359581": "auxiliary cell",
  "n08359753": "fifth column, Trojan horse",
  "n08359949": "political unit, political entity",
  "n08360248": "amphictyony",
  "n08360430": "lunatic fringe",
  "n08360540": "revolutionary group",
  "n08360672": "underground, resistance",
  "n08360849": "Maquis",
  "n08361001": "autocracy, autarchy",
  "n08361172": "constitutionalism",
  "n08361329": "democracy, republic, commonwealth",
  "n08361612": "diarchy, dyarchy",
  "n08361720": "gerontocracy",
  "n08361817": "gynecocracy, gynarchy",
  "n08361924": "hegemony",
  "n08362302": "mobocracy, ochlocracy",
  "n08362455": "oligarchy",
  "n08362775": "plutocracy",
  "n08362917": "republic",
  "n08363106": "technocracy",
  "n08363428": "theocracy",
  "n08363622": "hierocracy",
  "n08363711": "parliamentary democracy",
  "n08363812": "monarchy",
  "n08364044": "parliamentary monarchy",
  "n08364143": "capitalism, capitalist economy",
  "n08364367": "venture capitalism",
  "n08364548": "black economy",
  "n08364757": "industrialism",
  "n08364959": "market economy, free enterprise, private enterprise, laissez-faire economy",
  "n08365222": "mixed economy",
  "n08365344": "non-market economy",
  "n08365484": "state capitalism",
  "n08365672": "state socialism",
  "n08365855": "communism",
  "n08366071": "International",
  "n08366202": "socialism, socialist economy",
  "n08366440": "Nazism, Naziism, national socialism",
  "n08366664": "Falange",
  "n08366753": "economy, economic system",
  "n08367100": "managed economy",
  "n08367339": "mercantilism, mercantile system",
  "n08367579": "communist economy",
  "n08367683": "pluralism",
  "n08367880": "political system, form of government",
  "n08368308": "Bolshevism, collectivism, sovietism",
  "n08368516": "revisionism",
  "n08368631": "revisionism",
  "n08368757": "ecosystem",
  "n08368907": "generation",
  "n08369142": "posterity",
  "n08369220": "descendants, posterity",
  "n08369406": "coevals, contemporaries, generation",
  "n08369615": "beat generation, beats, beatniks",
  "n08369920": "Beatles",
  "n08370204": "teddy boys",
  "n08370505": "punks",
  "n08370878": "rockers, bikers",
  "n08371200": "skinheads, bootboys",
  "n08371574": "mods",
  "n08371838": "baby boom, baby-boom generation",
  "n08372020": "generation X, gen X",
  "n08372190": "peer group",
  "n08372299": "moiety",
  "n08372411": "tribe, folk",
  "n08372574": "totem",
  "n08372715": "tableau, tableau vivant",
  "n08372847": "Tribes of Israel, Twelve Tribes of Israel",
  "n08373055": "Lost Tribes",
  "n08373244": "venation",
  "n08373380": "vernation",
  "n08373544": "combination",
  "n08373723": "combination",
  "n08373818": "Fibonacci sequence",
  "n08373969": "phyle",
  "n08374049": "colony, settlement",
  "n08374653": "frontier settlement, outpost",
  "n08374773": "Plantation",
  "n08375031": "proprietary colony",
  "n08375154": "commonwealth",
  "n08375369": "commune",
  "n08375526": "lobby, pressure group, third house",
  "n08375722": "National Rifle Association, NRA",
  "n08375912": "lobby",
  "n08376051": "hierarchy, power structure, pecking order",
  "n08376250": "chain, concatenation",
  "n08376526": "catena",
  "n08376823": "daisy chain",
  "n08376948": "cordon",
  "n08377085": "course, line",
  "n08377332": "cycle",
  "n08377454": "electromotive series, electromotive force series, electrochemical series",
  "n08377806": "hierarchy",
  "n08378066": "celestial hierarchy",
  "n08378180": "data hierarchy",
  "n08378356": "taxonomy",
  "n08378555": "class structure",
  "n08378698": "caste system",
  "n08378819": "social organization, social organisation, social structure, social system, structure",
  "n08379353": "racial segregation",
  "n08379455": "petty apartheid",
  "n08379630": "de facto segregation",
  "n08379782": "de jure segregation",
  "n08379882": "purdah, sex segregation",
  "n08380017": "ulema, ulama",
  "n08380340": "segregation, separatism",
  "n08380606": "white separatism",
  "n08380768": "directorate, board of directors",
  "n08380975": "staggered board of directors",
  "n08381165": "management",
  "n08381296": "house",
  "n08381436": "leadership, leaders",
  "n08381636": "advisory board, planning board",
  "n08381820": "cabinet",
  "n08382056": "British Cabinet",
  "n08382297": "shadow cabinet",
  "n08382570": "United States Cabinet, US Cabinet",
  "n08383067": "draft board",
  "n08383185": "Kashag",
  "n08383310": "stock company",
  "n08383417": "joint-stock company",
  "n08383690": "closed corporation, close corporation, private corporation, privately held corporation",
  "n08383909": "family business",
  "n08384041": "closely held corporation",
  "n08384201": "shell corporation, shell entity",
  "n08384342": "Federal Deposit Insurance Corporation, FDIC",
  "n08384539": "Federal Home Loan Mortgage Corporation, Freddie Mac, FHLMC",
  "n08384738": "Federal National Mortgage Association, Fannie Mae, FNMA",
  "n08384900": "conventicle",
  "n08385009": "date, appointment, engagement",
  "n08385280": "visit",
  "n08385490": "blind date",
  "n08385602": "double date",
  "n08385717": "tryst, rendezvous",
  "n08385849": "luncheon meeting, lunch meeting",
  "n08385989": "power breakfast",
  "n08386143": "revival, revival meeting",
  "n08386280": "argosy",
  "n08386365": "upper class, upper crust",
  "n08386555": "elite, elite group",
  "n08386853": "chosen, elect",
  "n08387035": "cream, pick",
  "n08387213": "gentry, aristocracy",
  "n08387354": "intelligentsia, clerisy",
  "n08387495": "culturati",
  "n08387603": "literati",
  "n08387685": "landed gentry, squirearchy",
  "n08387806": "ruling class, people in power",
  "n08387930": "society, high society, beau monde, smart set, bon ton",
  "n08388074": "few",
  "n08388207": "nobility, aristocracy",
  "n08388503": "noblesse",
  "n08388636": "peerage, baronage",
  "n08388783": "baronetage",
  "n08388871": "knighthood",
  "n08388966": "samurai",
  "n08389094": "ninja",
  "n08389297": "artillery, artillery unit",
  "n08389438": "musketry",
  "n08389572": "battery",
  "n08389710": "cavalry",
  "n08389900": "horse cavalry",
  "n08390012": "mechanized cavalry",
  "n08390157": "infantry, foot",
  "n08390374": "paratroops",
  "n08390511": "militia, reserves",
  "n08390731": "militia",
  "n08391021": "home guard",
  "n08391206": "territorial, territorial reserve",
  "n08391387": "National Guard, home reserve",
  "n08391696": "National Guard Bureau, NGB",
  "n08391953": "Territorial Army",
  "n08392137": "terrorist organization, terrorist group, foreign terrorist organization, FTO",
  "n08394423": "standing army",
  "n08394535": "Union Army",
  "n08394657": "Confederate Army, Army of the Confederacy",
  "n08394811": "Continental Army",
  "n08394922": "United States Army, US Army, U. S. Army, Army, USA",
  "n08395298": "United States Army Rangers",
  "n08395465": "United States Military Academy, US Military Academy",
  "n08395682": "Army Intelligence, AI",
  "n08395991": "Ballistic Missile Defense Organization, BMDO",
  "n08396207": "Defense Information Systems Agency, DISA",
  "n08396537": "National Geospatial-Intelligence Agency, NGA",
  "n08396760": "Casualty Care Research Center, CCRC",
  "n08396990": "Army National Guard, ARNG",
  "n08397255": "military personnel, soldiery, troops",
  "n08397489": "friendly",
  "n08397675": "hostile",
  "n08397856": "cavalry, horse cavalry, horse",
  "n08398036": "garrison",
  "n08398179": "rank and file, rank",
  "n08398467": "coven",
  "n08398580": "sabbat, witches' Sabbath",
  "n08398773": "assortment, mixture, mixed bag, miscellany, miscellanea, variety, salmagundi, smorgasbord, potpourri, motley",
  "n08399287": "grab bag",
  "n08399378": "witches' brew, witches' broth, witch's brew",
  "n08399586": "range",
  "n08399818": "selection",
  "n08399977": "odds and ends, oddments, melange, farrago, ragbag, mishmash, mingle-mangle, hodgepodge, hotchpotch, gallimaufry, omnium-gatherum",
  "n08400191": "alphabet soup",
  "n08400331": "litter",
  "n08400452": "batch, clutch",
  "n08400605": "schmeer, schmear, shmear",
  "n08400772": "batch",
  "n08400870": "clutch",
  "n08400965": "membership, rank",
  "n08401248": "branch, subdivision, arm",
  "n08401554": "clientele, patronage, business",
  "n08401711": "rank and file",
  "n08401819": "rabble, riffraff, ragtag, ragtag and bobtail",
  "n08401970": "smart money",
  "n08402147": "trash, scum",
  "n08402222": "convocation",
  "n08402339": "alma mater",
  "n08402442": "deputation, commission, delegation, delegacy, mission",
  "n08402693": "diplomatic mission",
  "n08402828": "embassy",
  "n08402944": "High Commission",
  "n08403082": "legation, foreign mission",
  "n08403225": "mission, missionary post, missionary station, foreign mission",
  "n08403435": "press corps",
  "n08403631": "occupational group, vocation",
  "n08403787": "opposition",
  "n08403907": "Iraqi National Congress, INC",
  "n08404188": "Opposition",
  "n08404373": "commando",
  "n08404549": "contingent, detail",
  "n08404735": "general staff",
  "n08404895": "headquarters",
  "n08405124": "headquarters staff",
  "n08405267": "high command, supreme headquarters",
  "n08405490": "posse, posse comitatus",
  "n08405603": "kingdom",
  "n08405723": "empire",
  "n08405873": "Mogul empire",
  "n08406036": "Second Empire",
  "n08406160": "rogue's gallery",
  "n08406259": "galere, rogue's gallery",
  "n08406361": "hard core",
  "n08406486": "foundation",
  "n08406619": "charity",
  "n08406825": "philanthropic foundation",
  "n08407000": "private foundation",
  "n08407140": "public charity",
  "n08407330": "institute",
  "n08407449": "sisterhood, sistership",
  "n08407619": "exhibition, exposition, expo",
  "n08407839": "art exhibition",
  "n08407969": "retrospective",
  "n08408115": "peepshow, raree-show",
  "n08408267": "fair",
  "n08408418": "book fair, bookfair",
  "n08408557": "fair",
  "n08408709": "side",
  "n08408900": "working group, working party",
  "n08409130": "expedition",
  "n08409323": "Lewis and Clark Expedition",
  "n08409617": "senior high school, senior high, high, highschool, high school",
  "n08409835": "junior high school, junior high",
  "n08409969": "preparatory school, prep school",
  "n08410092": "choir school, schola cantorum",
  "n08410282": "public school",
  "n08410454": "charter school",
  "n08410688": "public school",
  "n08410891": "Eton College",
  "n08411031": "Winchester College",
  "n08411170": "private school",
  "n08411369": "Catholic school",
  "n08411483": "dance school",
  "n08411585": "day school",
  "n08411701": "boarding school",
  "n08411849": "day school",
  "n08411970": "night school",
  "n08412136": "kindergarten",
  "n08412265": "nursery school",
  "n08412362": "playschool, play group",
  "n08412492": "Sunday school, Sabbath school",
  "n08412620": "normal school, teachers college",
  "n08412749": "grade school, grammar school, elementary school, primary school",
  "n08412958": "grammar school",
  "n08413092": "secondary modern school",
  "n08413248": "comprehensive school, composite school",
  "n08413408": "school board, board of education",
  "n08413529": "zoning board",
  "n08413681": "zoning commission",
  "n08413834": "immigration",
  "n08414040": "inspectorate",
  "n08414119": "jury",
  "n08414381": "panel",
  "n08414608": "panel, venire",
  "n08414807": "jury, panel",
  "n08414964": "grand jury",
  "n08415127": "hung jury",
  "n08415272": "petit jury, petty jury",
  "n08415469": "special jury, blue ribbon jury",
  "n08415661": "spearhead",
  "n08415774": "bevy",
  "n08415983": "firing line",
  "n08416137": "immigrant class",
  "n08416328": "left, left wing",
  "n08416523": "center",
  "n08416652": "right, right wing",
  "n08416890": "religious right",
  "n08417099": "hard right",
  "n08417178": "old guard",
  "n08417281": "pro-choice faction",
  "n08417436": "pro-life faction",
  "n08417572": "old school",
  "n08417673": "convoy",
  "n08417801": "convoy",
  "n08417920": "seance, sitting, session",
  "n08418103": "aggregate, congeries, conglomeration",
  "n08418316": "agent bank",
  "n08418420": "commercial bank, full service bank",
  "n08418631": "national bank",
  "n08418763": "state bank",
  "n08418885": "lead bank, agent bank",
  "n08419033": "member bank",
  "n08419163": "merchant bank, acquirer",
  "n08419354": "acquirer",
  "n08419562": "acquirer",
  "n08419774": "transfer agent",
  "n08419984": "nondepository financial institution",
  "n08420278": "depository financial institution, bank, banking concern, banking company",
  "n08420839": "finance company",
  "n08421100": "consumer finance company, small loan company",
  "n08421291": "industrial bank, industrial loan company",
  "n08421441": "captive finance company",
  "n08421644": "sales finance company",
  "n08421807": "commercial finance company, commercial credit company",
  "n08421975": "Farm Credit System, FCS",
  "n08422247": "hawala",
  "n08422524": "thrift institution",
  "n08422714": "savings and loan, savings and loan association",
  "n08422930": "building society",
  "n08423057": "savings bank",
  "n08423298": "Home Loan Bank",
  "n08423490": "Federal Home Loan Bank System",
  "n08423634": "Federal Housing Administration, FHA",
  "n08423840": "child welfare agency, child welfare service",
  "n08424044": "Securities and Exchange Commission, SEC",
  "n08424222": "trust company, trust corporation",
  "n08424501": "mutual savings bank, MSB",
  "n08424662": "federal savings bank, FSB",
  "n08424769": "firing squad, firing party",
  "n08424951": "market",
  "n08425173": "black market",
  "n08425303": "traffic",
  "n08425556": "air traffic",
  "n08425657": "commuter traffic",
  "n08425777": "pedestrian traffic, foot traffic",
  "n08425888": "vehicular traffic, vehicle traffic",
  "n08426111": "automobile traffic, car traffic",
  "n08426211": "bicycle traffic",
  "n08426298": "bus traffic",
  "n08426378": "truck traffic",
  "n08426461": "formation",
  "n08426816": "military formation",
  "n08426993": "open order",
  "n08427163": "close order",
  "n08427282": "extended order",
  "n08427453": "sick call, sick parade",
  "n08427629": "caravan, train, wagon train",
  "n08427918": "cavalcade",
  "n08428019": "march",
  "n08428252": "hunger march",
  "n08428383": "motorcade",
  "n08428485": "parade",
  "n08428623": "callithump, callathump, callithump parade",
  "n08428756": "file, single file, Indian file",
  "n08428944": "snake dance",
  "n08429052": "column",
  "n08429167": "cortege, retinue, suite, entourage",
  "n08429346": "Praetorian Guard",
  "n08429482": "cortege",
  "n08429556": "recession, recessional",
  "n08429744": "backfield",
  "n08429899": "secondary",
  "n08430054": "linemen",
  "n08430203": "line",
  "n08430568": "line",
  "n08430940": "line of march",
  "n08431048": "line of succession",
  "n08431205": "lineup",
  "n08431345": "picket line",
  "n08431437": "row",
  "n08431600": "serration",
  "n08431721": "terrace",
  "n08431942": "rank",
  "n08432149": "conga line",
  "n08432264": "trap line",
  "n08432345": "queue, waiting line",
  "n08432606": "breadline, bread line",
  "n08432714": "checkout line",
  "n08432820": "chow line",
  "n08432974": "gas line",
  "n08433077": "reception line",
  "n08433245": "ticket line",
  "n08433343": "unemployment line",
  "n08433447": "row",
  "n08433575": "column",
  "n08433727": "aviation, air power",
  "n08433861": "dragnet",
  "n08434029": "machinery",
  "n08434259": "network, web",
  "n08434622": "espionage network",
  "n08434704": "old boy network",
  "n08435024": "support system",
  "n08435246": "nonlinear system",
  "n08435388": "system, scheme",
  "n08435937": "subsystem",
  "n08436036": "organism",
  "n08436203": "syntax",
  "n08436288": "body",
  "n08436452": "shebang",
  "n08436562": "craft, trade",
  "n08436759": "vegetation, flora, botany",
  "n08437317": "browse",
  "n08437515": "brush, brushwood, coppice, copse, thicket",
  "n08437721": "brake",
  "n08437847": "canebrake",
  "n08437968": "spinney",
  "n08438067": "growth",
  "n08438223": "scrub, chaparral, bush",
  "n08438384": "stand",
  "n08438533": "forest, wood, woods",
  "n08438840": "bosk",
  "n08438928": "grove",
  "n08439022": "jungle",
  "n08439126": "rain forest, rainforest",
  "n08439268": "temperate rain forest",
  "n08439369": "tropical rain forest, selva",
  "n08439476": "underbrush, undergrowth, underwood",
  "n08439694": "shrubbery",
  "n08439808": "garden",
  "n08439955": "staff",
  "n08440382": "line personnel",
  "n08440499": "management personnel",
  "n08440630": "dictatorship, absolutism, authoritarianism, Caesarism, despotism, monocracy, one-man rule, shogunate, Stalinism, totalitarianism, tyranny",
  "n08441039": "police state",
  "n08441203": "law, jurisprudence",
  "n08453108": "administrative law",
  "n08453299": "canon law, ecclesiastical law",
  "n08453464": "civil law",
  "n08453722": "common law, case law, precedent",
  "n08454003": "international law, law of nations",
  "n08454191": "maritime law, marine law, admiralty law",
  "n08454445": "law of the land",
  "n08454818": "martial law",
  "n08455037": "mercantile law, commercial law, law merchant",
  "n08455271": "military law",
  "n08455520": "Mosaic law, Law of Moses",
  "n08455829": "shariah, shariah law, sharia, sharia law, Islamic law",
  "n08456178": "hudud, hudood",
  "n08456347": "statutory law",
  "n08456474": "securities law",
  "n08456619": "tax law",
  "n08456727": "bureaucracy, bureaucratism",
  "n08456888": "menagerie",
  "n08456993": "ordering, order, ordination",
  "n08457369": "genetic code",
  "n08457543": "genome",
  "n08457789": "triplet code",
  "n08457976": "series",
  "n08458504": "series",
  "n08458833": "nexus",
  "n08458912": "progression, patterned advance",
  "n08459087": "rash, blizzard",
  "n08459252": "sequence",
  "n08459648": "string, train",
  "n08459908": "succession",
  "n08460152": "cascade",
  "n08460395": "parade",
  "n08460585": "streak, run",
  "n08460769": "losing streak",
  "n08460847": "winning streak",
  "n08460924": "arithmetic progression",
  "n08461172": "geometric progression",
  "n08461424": "harmonic progression",
  "n08461595": "stream, flow, current",
  "n08461875": "wave train",
  "n08461981": "panoply",
  "n08462066": "bank",
  "n08462205": "stockpile",
  "n08462320": "data, information",
  "n08462534": "accounting data",
  "n08462721": "metadata",
  "n08462858": "raw data",
  "n08462964": "ana",
  "n08463063": "mail, post",
  "n08463345": "fan mail",
  "n08463509": "hate mail",
  "n08463647": "mailing",
  "n08463817": "sampler",
  "n08463969": "treasure",
  "n08464098": "treasure trove",
  "n08464324": "trinketry",
  "n08464449": "troponymy, troponomy",
  "n08464601": "movement, social movement, front",
  "n08465118": "deco, art deco",
  "n08465312": "art nouveau",
  "n08465558": "avant-garde, vanguard, van, new wave",
  "n08465776": "constructivism",
  "n08466010": "suprematism",
  "n08466175": "cubism",
  "n08466412": "dada, dadaism",
  "n08466643": "artistic movement, art movement",
  "n08467258": "expressionism",
  "n08467576": "neoexpressionism",
  "n08467757": "supra expressionism",
  "n08467871": "fauvism",
  "n08468084": "futurism",
  "n08468261": "Hudson River school, romantic realism",
  "n08468499": "imagism",
  "n08468721": "lake poets",
  "n08468958": "luminism",
  "n08469233": "minimalism, minimal art, reductivism",
  "n08469457": "naturalism, realism",
  "n08469678": "needy",
  "n08469785": "neoromanticism",
  "n08469917": "New Wave, Nouvelle Vague",
  "n08470034": "secession, sezession",
  "n08470210": "surrealism",
  "n08470507": "symbolism",
  "n08470710": "Boy Scouts",
  "n08470928": "Boy Scouts of America",
  "n08471185": "Girl Scouts",
  "n08471369": "Civil Rights movement",
  "n08471589": "common front",
  "n08471799": "cultural movement",
  "n08471949": "ecumenism, oecumenism",
  "n08472120": "falun gong",
  "n08472335": "political movement",
  "n08472590": "Enlightenment, Age of Reason",
  "n08472890": "labor movement, trade union movement, labor",
  "n08473173": "Industrial Workers of the World, IWW, I.W.W.",
  "n08473482": "unionism, trade unionism",
  "n08473623": "reform movement",
  "n08473787": "religious movement",
  "n08474120": "Akhbari",
  "n08474341": "Usuli",
  "n08474623": "Counter Reformation",
  "n08474875": "ecumenical movement",
  "n08475070": "Gallicanism",
  "n08475332": "Lubavitch, Lubavitch movement, Chabad-Lubavitch, Chabad",
  "n08475586": "Oxford movement",
  "n08475722": "Pietism",
  "n08475929": "Reformation, Protestant Reformation",
  "n08476165": "Taliban, Taleban",
  "n08476425": "Northern Alliance, United Front",
  "n08476629": "Nation of Islam",
  "n08476829": "humanism",
  "n08476988": "analytical cubism",
  "n08477077": "synthetic cubism",
  "n08477164": "unconfessed",
  "n08477307": "unemployed people, unemployed",
  "n08477511": "wolf pack",
  "n08477634": "womanhood, woman, fair sex",
  "n08477912": "womankind",
  "n08478018": "camp",
  "n08478171": "hobo camp, jungle",
  "n08478261": "record company",
  "n08478371": "reunion",
  "n08478482": "mover, public mover, moving company, removal firm, removal company",
  "n08478702": "think tank, think factory",
  "n08478851": "vestry",
  "n08479095": "Jewry",
  "n08479202": "Zionism, Zionist movement",
  "n08479407": "Zhou, Zhou dynasty, Chou, Chou dynasty, Chow, Chow dynasty",
  "n08479615": "muster",
  "n08479795": "rap group",
  "n08479894": "rave-up",
  "n08479986": "registration, enrollment",
  "n08480135": "table",
  "n08480303": "World Council of Churches",
  "n08480512": "number",
  "n08480662": "vote",
  "n08480847": "blue",
  "n08481009": "grey, gray",
  "n08481184": "host",
  "n08481369": "pool",
  "n08481604": "typing pool",
  "n08481715": "shipper",
  "n08481832": "center",
  "n08481983": "diaspora",
  "n08482113": "flank, wing",
  "n08482271": "head",
  "n08482577": "local authority",
  "n08482700": "rear",
  "n08482866": "smithereens",
  "n08483136": "chosen people",
  "n08483265": "Azeri",
  "n08483353": "Bengali",
  "n08483500": "Berbers, Arab-Berbers",
  "n08483650": "Dagestani",
  "n08483788": "Flemish",
  "n08483935": "Hebrews, Israelites",
  "n08484239": "Maori",
  "n08484386": "Mayas",
  "n08484522": "Mbundu, Ovimbundu",
  "n08484648": "Pathan, Pashtun",
  "n08484810": "Tajik, Tadzhik",
  "n08484986": "Walloons",
  "n08485160": "Ferdinand and Isabella",
  "n08485281": "Medici",
  "n08485434": "Committee for State Security, KGB, Soviet KGB",
  "n08485598": "Federal Security Bureau, FSB, Federal Security Service",
  "n08485830": "Russian agency",
  "n08485977": "Wicca",
  "n08486075": "William and Mary",
  "n08486189": "wine tasting",
  "n08486306": "wing",
  "n08486538": "Wise Men, Magi",
  "n08486971": "World Trade Organization, WTO",
  "n08487149": "Association for the Advancement of Retired Persons, AARP",
  "n08487319": "National Association of Realtors",
  "n08487504": "Association of Southeast Asian Nations, ASEAN",
  "n08487953": "Abkhaz, Abkhas",
  "n08488081": "Achomawi",
  "n08488250": "Akwa'ala",
  "n08488411": "Aleut",
  "n08488675": "Circassian",
  "n08488833": "Inca, Inka",
  "n08489121": "Quechua, Kechua",
  "n08489285": "Xhosa",
  "n08489392": "Zulu",
  "n08489497": "here",
  "n08489627": "there",
  "n08489765": "somewhere",
  "n08489890": "bilocation",
  "n08490039": "seat",
  "n08490199": "home",
  "n08490402": "base, home",
  "n08490531": "aclinic line, magnetic equator",
  "n08490685": "agonic line",
  "n08490840": "isogonic line, isogonal line, isogone",
  "n08491027": "address",
  "n08491245": "mailing address",
  "n08491410": "box number",
  "n08491531": "post-office box number, PO box number, PO Box No, box number",
  "n08491710": "street address",
  "n08491826": "administrative district, administrative division, territorial division",
  "n08492354": "aerie, aery, eyrie, eyry",
  "n08492461": "agora",
  "n08492546": "air lane, flight path, airway, skyway",
  "n08492747": "traffic pattern, approach pattern, pattern",
  "n08493064": "territory, soil",
  "n08493261": "Andalusia, Andalucia",
  "n08493493": "Appalachia",
  "n08493705": "flight path",
  "n08493825": "wing",
  "n08493961": "approach path, approach, glide path, glide slope",
  "n08494133": "ambiance, ambience",
  "n08494231": "amusement park, funfair, pleasure ground",
  "n08494459": "Antarctic, Antarctic Zone, South Frigid Zone",
  "n08494676": "Antarctic Circle",
  "n08494782": "Adelie Land, Terre Adelie, Adelie Coast",
  "n08494987": "apex, solar apex, apex of the sun's way",
  "n08495249": "antapex",
  "n08495420": "apogee",
  "n08495617": "apoapsis, point of apoapsis",
  "n08495908": "aphelion",
  "n08496095": "apojove",
  "n08496199": "aposelene, apolune",
  "n08496334": "apron",
  "n08496477": "Arctic, Arctic Zone, North Frigid Zone",
  "n08496655": "polar circle",
  "n08496798": "Arctic Circle",
  "n08497107": "arena",
  "n08497294": "area, country",
  "n08498163": "high country",
  "n08498279": "ascending node",
  "n08498424": "node",
  "n08498580": "node",
  "n08498732": "antinode",
  "n08498888": "asteroid belt",
  "n08499057": "atmosphere, air",
  "n08499357": "bed ground, bed-ground, bedground",
  "n08499504": "biosphere",
  "n08499680": "back of beyond",
  "n08499840": "colony, dependency",
  "n08500079": "Crown Colony",
  "n08500213": "depth",
  "n08500433": "outer space, space",
  "n08500819": "interplanetary space",
  "n08500989": "interstellar space",
  "n08501114": "frontier",
  "n08501275": "heliopause",
  "n08501565": "heliosphere",
  "n08501729": "holding pattern",
  "n08501887": "intergalactic space",
  "n08502051": "deep space",
  "n08502171": "aerospace",
  "n08502317": "airspace",
  "n08502507": "backwater",
  "n08502672": "backwoods, back country, boondocks, hinterland",
  "n08502797": "Bad Lands, Badlands",
  "n08503004": "banana republic",
  "n08503238": "Barbary",
  "n08503477": "Barbary Coast",
  "n08503639": "Barbary Coast",
  "n08503921": "Bithynia",
  "n08504151": "Nicaea",
  "n08504375": "Nubia",
  "n08504594": "barren, waste, wasteland",
  "n08504851": "heath, heathland",
  "n08505018": "bush",
  "n08505110": "outback",
  "n08505265": "Never-Never",
  "n08505402": "frontier",
  "n08505573": "desert",
  "n08506347": "semidesert",
  "n08506496": "oasis",
  "n08506641": "battlefield, battleground, field of battle, field of honor, field",
  "n08506932": "Armageddon",
  "n08507109": "Camlan",
  "n08507255": "minefield",
  "n08507381": "beat, round",
  "n08507558": "beginning, origin, root, rootage, source",
  "n08508105": "derivation",
  "n08508361": "spring",
  "n08508449": "fountainhead, headspring, head",
  "n08508629": "headwater",
  "n08508736": "wellhead, wellspring",
  "n08508834": "jumping-off place, point of departure",
  "n08509111": "jungle",
  "n08509251": "concrete jungle",
  "n08509442": "zone",
  "n08509786": "belt",
  "n08510030": "Bible Belt",
  "n08510169": "fatherland, homeland, motherland, mother country, country of origin, native land",
  "n08510350": "birthplace, place of birth",
  "n08510456": "birthplace, cradle, place of origin, provenance, provenience",
  "n08510666": "side, face",
  "n08511017": "beam-ends",
  "n08511241": "bottom, underside, undersurface",
  "n08511473": "underbelly",
  "n08511570": "foot",
  "n08511777": "base",
  "n08511970": "bottom",
  "n08512182": "rock bottom",
  "n08512259": "boundary, bound, bounds",
  "n08512736": "boundary line, border, borderline, delimitation, mete",
  "n08513072": "bourn, bourne",
  "n08513163": "borderland, border district, march, marchland",
  "n08513417": "narco-state",
  "n08513718": "place, property",
  "n08514034": "center, centre",
  "n08514233": "colony",
  "n08514412": "nerve center, nerve centre",
  "n08514592": "circumference, circuit",
  "n08514865": "fence line",
  "n08514975": "Green Line",
  "n08515126": "Line of Control",
  "n08515349": "property line",
  "n08515457": "state line, state boundary",
  "n08515581": "Mason-Dixon line, Mason and Dixon line, Mason and Dixon's line",
  "n08515817": "district line",
  "n08515911": "county line",
  "n08516002": "city line",
  "n08516080": "balk, baulk",
  "n08516260": "balkline, baulk-line, string line",
  "n08516431": "bomb site",
  "n08516584": "bowels",
  "n08516660": "bowling green",
  "n08516767": "breadbasket",
  "n08516885": "breeding ground",
  "n08517010": "bridgehead",
  "n08517127": "brink",
  "n08517204": "broadcast area",
  "n08517327": "buffer state, buffer country",
  "n08517449": "bull's eye, bull",
  "n08517554": "bus route",
  "n08517676": "bus stop",
  "n08517825": "checkpoint",
  "n08517966": "cabstand, taxistand, taxi rank",
  "n08518171": "campsite, campground, camping site, camping ground, bivouac, encampment, camping area",
  "n08518374": "campus",
  "n08518505": "capital",
  "n08518747": "capital",
  "n08518940": "river basin, basin, watershed, drainage basin, catchment area, catchment basin, drainage area",
  "n08519299": "detention basin",
  "n08519444": "retention basin",
  "n08519624": "Caucasia, Caucasus",
  "n08519916": "Transcaucasia",
  "n08520190": "celestial equator, equinoctial circle, equinoctial line, equinoctial",
  "n08520401": "celestial point",
  "n08520728": "equinoctial point, equinox",
  "n08520958": "vernal equinox",
  "n08521112": "autumnal equinox",
  "n08521267": "celestial sphere, sphere, empyrean, firmament, heavens, vault of heaven, welkin",
  "n08521623": "cemetery, graveyard, burial site, burial ground, burying ground, memorial park, necropolis",
  "n08521816": "center, centre, midpoint",
  "n08522287": "center of buoyancy, centre of buoyancy, center of immersion, centre of immersion",
  "n08522518": "center of gravity, centre of gravity",
  "n08522742": "center of flotation, centre of flotation",
  "n08522872": "center of mass, centre of mass",
  "n08523064": "barycenter",
  "n08523217": "centroid",
  "n08523340": "trichion, crinion",
  "n08523483": "center, centre, middle, heart, eye",
  "n08524021": "center stage, centre stage",
  "n08524130": "city center, city centre, central city",
  "n08524262": "core",
  "n08524404": "navel, navel point",
  "n08524572": "storm center, storm centre",
  "n08524735": "city, metropolis, urban center",
  "n08537708": "megalopolis",
  "n08537837": "city district",
  "n08538426": "precinct",
  "n08538584": "police precinct",
  "n08538730": "voting precinct, election district",
  "n08538917": "polling place, polling station",
  "n08539072": "business district, downtown",
  "n08539276": "outskirts",
  "n08539457": "environs, purlieu",
  "n08539557": "Tin Pan Alley",
  "n08539717": "conurbation, urban sprawl, sprawl",
  "n08539893": "subtopia",
  "n08540016": "borough",
  "n08540193": "burgh",
  "n08540266": "pocket borough",
  "n08540416": "rotten borough",
  "n08540532": "borough",
  "n08540770": "canton",
  "n08540903": "city",
  "n08541130": "city limit, city limits",
  "n08541288": "clearing, glade",
  "n08541454": "Coats Land",
  "n08541609": "commune",
  "n08541841": "zone, geographical zone",
  "n08542081": "climatic zone",
  "n08542304": "commons, common land",
  "n08542403": "commonwealth",
  "n08542634": "confluence, meeting",
  "n08542884": "congressional district",
  "n08543081": "financial center",
  "n08543223": "hub",
  "n08543496": "civic center, municipal center, down town",
  "n08543625": "inner city",
  "n08543776": "chokepoint",
  "n08543916": "Corn Belt",
  "n08544125": "corncob, corn cob",
  "n08544275": "corner",
  "n08544419": "corner",
  "n08544593": "corner",
  "n08544719": "cornfield, corn field",
  "n08544813": "country, state, land",
  "n08546183": "county",
  "n08546870": "county",
  "n08547048": "county palatine",
  "n08547143": "county seat, county courthouse",
  "n08547300": "county town, shire town",
  "n08547468": "cow pasture",
  "n08547544": "crest",
  "n08547655": "timber line, timberline, tree line",
  "n08547816": "snow line",
  "n08547938": "crossing",
  "n08548065": "cross section",
  "n08548239": "culmination",
  "n08548415": "profile",
  "n08548605": "soil profile",
  "n08548733": "department",
  "n08548924": "descending node",
  "n08549070": "development",
  "n08549292": "ghetto",
  "n08549480": "housing development",
  "n08549733": "housing estate",
  "n08549911": "housing project, public housing",
  "n08550076": "dig, excavation, archeological site",
  "n08550364": "abbacy",
  "n08550455": "archbishopric",
  "n08550560": "archdeaconry",
  "n08550664": "bailiwick",
  "n08550766": "caliphate",
  "n08550862": "archdiocese",
  "n08550966": "diocese, bishopric, episcopate",
  "n08551177": "disaster area",
  "n08551296": "eparchy, exarchate",
  "n08551420": "theater of war, theatre of war",
  "n08551628": "field, field of operations, theater, theater of operations, theatre, theatre of operations",
  "n08551984": "zone of interior",
  "n08552138": "district, territory, territorial dominion, dominion",
  "n08553132": "enclave",
  "n08553280": "federal district",
  "n08553424": "palatinate",
  "n08553535": "residential district, residential area, community",
  "n08553887": "planned community",
  "n08554040": "retirement community, retirement complex",
  "n08554206": "uptown",
  "n08554346": "red-light district",
  "n08554440": "suburb, suburbia, suburban area",
  "n08554762": "exurbia",
  "n08554872": "addition",
  "n08555001": "bedroom community",
  "n08555102": "faubourg",
  "n08555333": "stockbroker belt",
  "n08555443": "tenement district",
  "n08555569": "airspace, air space",
  "n08555710": "crawlspace, crawl space",
  "n08555883": "disk space, disc space",
  "n08556065": "disk overhead, overhead",
  "n08556266": "swap space, swap file",
  "n08556386": "distance",
  "n08556491": "domain, demesne, land",
  "n08556950": "archduchy",
  "n08557057": "barony",
  "n08557131": "duchy, dukedom",
  "n08557271": "earldom",
  "n08557396": "emirate",
  "n08557482": "empire, imperium",
  "n08557754": "fiefdom",
  "n08557864": "grand duchy",
  "n08557976": "viscounty",
  "n08558082": "khanate",
  "n08558155": "kingdom, realm",
  "n08558289": "Camelot",
  "n08558488": "principality, princedom",
  "n08558661": "Kingdom of God",
  "n08558770": "sheikdom, sheikhdom",
  "n08558882": "suzerainty",
  "n08558963": "residence, abode",
  "n08559155": "domicile, legal residence",
  "n08559508": "home, place",
  "n08559766": "home away from home, home from home",
  "n08559922": "business address",
  "n08560027": "dump, garbage dump, trash dump, rubbish dump, wasteyard, waste-yard, dumpsite",
  "n08560295": "dude ranch",
  "n08560415": "honeymoon resort",
  "n08560560": "eitchen midden, midden, kitchen midden",
  "n08560785": "earshot, earreach, hearing",
  "n08560952": "view, eyeshot",
  "n08561081": "north",
  "n08561230": "northeast",
  "n08561351": "east",
  "n08561462": "southeast",
  "n08561583": "south",
  "n08561714": "southwest",
  "n08561835": "west",
  "n08561946": "northwest",
  "n08562067": "Earth, earth",
  "n08562243": "eastern hemisphere, orient",
  "n08562454": "Old World",
  "n08562620": "East, Orient",
  "n08562757": "Far East",
  "n08562990": "northland",
  "n08563085": "southland",
  "n08563180": "East, eastern United States",
  "n08563478": "Southeast, southeastern United States",
  "n08563627": "Southwest, southwestern United States",
  "n08563990": "Northeast, northeastern United States",
  "n08564139": "Northwest, northwestern United States",
  "n08564307": "Midwest, middle west, midwestern United States",
  "n08564739": "Pacific Northwest",
  "n08565006": "Rustbelt",
  "n08565214": "ecliptic",
  "n08565506": "Eden, paradise, nirvana, heaven, promised land, Shangri-la",
  "n08565701": "edge, border",
  "n08565894": "end",
  "n08566028": "end, terminal",
  "n08566554": "end point, endpoint, termination, terminus",
  "n08566707": "end",
  "n08566884": "end",
  "n08567072": "Enderby Land",
  "n08567235": "environment, environs, surroundings, surround",
  "n08567600": "Finger Lakes",
  "n08567877": "finish, destination, goal",
  "n08568142": "medium",
  "n08568256": "setting, scene",
  "n08568430": "scenario",
  "n08568579": "element",
  "n08568719": "equator",
  "n08568978": "extremity",
  "n08569165": "extreme point, extreme, extremum",
  "n08569319": "fairway",
  "n08569482": "farmland, farming area",
  "n08569591": "fault line",
  "n08569777": "field",
  "n08569998": "field",
  "n08570242": "field of fire",
  "n08570402": "grounds",
  "n08570533": "bent",
  "n08570634": "hayfield, meadow",
  "n08570758": "playing field, athletic field, playing area, field",
  "n08571139": "medical center",
  "n08571275": "midfield",
  "n08571459": "finishing line, finish line",
  "n08571642": "firebreak, fireguard",
  "n08571799": "firing line",
  "n08571898": "flea market",
  "n08572020": "Fleet Street",
  "n08572162": "flies",
  "n08572335": "focus",
  "n08572467": "forefront, head",
  "n08572726": "foul line",
  "n08572877": "foul line",
  "n08573010": "foul line",
  "n08573140": "baseline",
  "n08573258": "Frigid Zone, polar zone, polar region",
  "n08573472": "front, front end, forepart",
  "n08573674": "battlefront, front, front line",
  "n08573842": "garbage heap, junk heap, rubbish heap, scrapheap, trash heap, junk pile, trash pile, refuse heap",
  "n08574038": "toxic waste dump, toxic waste site, toxic dumpsite",
  "n08574216": "gathering place",
  "n08574314": "geographical area, geographic area, geographical region, geographic region",
  "n08578032": "epicenter, epicentre",
  "n08578174": "dust bowl",
  "n08578364": "biogeographical region",
  "n08578517": "benthos, benthic division, benthonic zone",
  "n08578706": "geographic point, geographical point",
  "n08578951": "ghetto",
  "n08579134": "goal line",
  "n08579266": "goldfield",
  "n08579352": "grainfield, grain field",
  "n08579487": "great circle",
  "n08579780": "green, putting green, putting surface",
  "n08580011": "greenbelt, greenway",
  "n08580134": "ground",
  "n08580326": "ground zero",
  "n08580447": "ground zero",
  "n08580583": "habitat, home ground",
  "n08580803": "habitation",
  "n08580944": "half-mast, half-staff",
  "n08581122": "Harley Street",
  "n08581299": "hatchery",
  "n08581503": "haunt, hangout, resort, repair, stamping ground",
  "n08581699": "hearth, fireside",
  "n08581897": "heartland",
  "n08582065": "hunting ground",
  "n08582157": "D-layer, D region",
  "n08582337": "Appleton layer, F layer, F region",
  "n08582613": "Heaviside layer, Kennelly-Heaviside layer, E layer, E region",
  "n08582837": "hell, hell on earth, hellhole, snake pit, the pits, inferno",
  "n08583095": "hemisphere",
  "n08583292": "hemline",
  "n08583455": "heronry",
  "n08583554": "hipline",
  "n08583682": "hipline",
  "n08583793": "drop",
  "n08583994": "dead drop",
  "n08584218": "hideout, hideaway, den",
  "n08584361": "lurking place",
  "n08584449": "hiding place",
  "n08584618": "high, heights",
  "n08584787": "hilltop, brow",
  "n08584914": "hole-in-the-wall",
  "n08585056": "holy place, sanctum, holy",
  "n08585158": "home",
  "n08585298": "point source",
  "n08585447": "trail head, trailhead",
  "n08585540": "home range, home territory",
  "n08585657": "horizon, apparent horizon, visible horizon, sensible horizon, skyline",
  "n08585846": "horizon, celestial horizon",
  "n08586036": "horse latitude",
  "n08586218": "hot spot, hotspot",
  "n08586330": "hot spot, hotspot",
  "n08586509": "hour angle",
  "n08586688": "hour circle",
  "n08586825": "see",
  "n08586978": "junkyard",
  "n08587086": "justiciary",
  "n08587174": "reservation, reserve",
  "n08587328": "Indian reservation",
  "n08587439": "preserve",
  "n08587571": "shooting preserve",
  "n08587709": "school district",
  "n08587828": "shire",
  "n08587985": "industrial park",
  "n08588152": "inside, interior",
  "n08588294": "inside, interior",
  "n08588486": "belly",
  "n08588596": "midland",
  "n08588699": "midst, thick",
  "n08588840": "penetralia",
  "n08588916": "ionosphere",
  "n08589140": "irredenta, irridenta",
  "n08589351": "isobar",
  "n08589532": "isochrone",
  "n08589670": "isoclinic line, isoclinal",
  "n08589801": "isogram, isopleth, isarithm",
  "n08590055": "isohel",
  "n08590172": "isotherm",
  "n08590369": "jurisdiction",
  "n08590719": "turf",
  "n08590909": "key, paint",
  "n08591269": "kingdom",
  "n08591399": "lair, den",
  "n08591486": "launching site",
  "n08591592": "lawn",
  "n08591680": "layer",
  "n08592165": "lead",
  "n08592352": "lee, lee side, leeward",
  "n08592494": "limb",
  "n08592656": "limit, demarcation, demarcation line",
  "n08592898": "upper limit",
  "n08593014": "lower limit",
  "n08593130": "limit",
  "n08593262": "line",
  "n08593924": "line",
  "n08594286": "line",
  "n08594543": "flight line",
  "n08594714": "line of battle",
  "n08594886": "salient",
  "n08595054": "battle line",
  "n08595171": "line of flight",
  "n08595299": "line of march",
  "n08595398": "line of sight, line of vision",
  "n08595531": "latitude",
  "n08595720": "latitude, line of latitude, parallel of latitude, parallel",
  "n08595958": "lunar latitude",
  "n08596076": "littoral, litoral, littoral zone, sands",
  "n08596214": "loading zone, loading area",
  "n08596336": "load line, Plimsoll line, Plimsoll mark, Plimsoll",
  "n08596519": "Lombard Street",
  "n08596668": "longitude",
  "n08596830": "Whitehall",
  "n08597023": "Trafalgar Square",
  "n08597176": "lookout, observation post",
  "n08597323": "Maghreb, Mahgrib",
  "n08597579": "magnetic pole",
  "n08597727": "mandate, mandatory",
  "n08597973": "market cross",
  "n08598132": "maximum",
  "n08598301": "grassland",
  "n08598568": "mecca",
  "n08598696": "melting pot",
  "n08598823": "meridian, line of longitude",
  "n08599174": "observer's meridian",
  "n08599292": "prime meridian",
  "n08599488": "Greenwich Meridian",
  "n08599662": "magnetic meridian",
  "n08599792": "dateline, date line, International Date Line",
  "n08599976": "meteorological observation post, weather station",
  "n08600147": "midair",
  "n08600274": "minimum",
  "n08600443": "monument",
  "n08600618": "mud flat",
  "n08600760": "nadir",
  "n08600992": "national park",
  "n08602037": "Acadia National Park",
  "n08602254": "Arches National Park",
  "n08602456": "Badlands National Park",
  "n08602650": "Big Bend",
  "n08602822": "Big Bend National Park",
  "n08602997": "Biscayne National Park",
  "n08603158": "Bryce Canyon National Park",
  "n08603308": "Canyonlands National Park",
  "n08603525": "Capitol Reef National Park",
  "n08603704": "Carlsbad Caverns National Park",
  "n08603924": "Channel Islands National Park",
  "n08604085": "Crater Lake National Park",
  "n08604283": "Denali National Park",
  "n08604487": "Everglades National Park",
  "n08604721": "Gates of the Arctic National Park",
  "n08604891": "Grand Canyon National Park",
  "n08605123": "Grand Teton National Park",
  "n08605261": "Great Smoky Mountains National Park",
  "n08605493": "Guadalupe Mountains National Park",
  "n08605720": "Haleakala National Park",
  "n08605863": "Hawaii Volcanoes National Park",
  "n08606012": "Hot Springs National Park",
  "n08606206": "Isle Royal National Park",
  "n08606370": "Katmai National Park",
  "n08606502": "Kenai Fjords National Park",
  "n08606688": "Kings Canyon National Park",
  "n08606869": "Kobuk Valley National Park",
  "n08607066": "Lake Clark National Park",
  "n08607231": "Lassen Volcanic National Park",
  "n08607408": "Mammoth Cave National Park",
  "n08607575": "Mesa Verde National Park",
  "n08607752": "Mount Ranier National Park",
  "n08607957": "North Cascades National Park",
  "n08608157": "Olympic National Park",
  "n08608314": "Petrified Forest National Park",
  "n08608505": "Platt National Park",
  "n08608641": "Redwood National Park",
  "n08608813": "Rocky Mountain National Park",
  "n08608986": "Sequoia National Park",
  "n08609131": "Shenandoah National Park",
  "n08609278": "Theodore Roosevelt Memorial National Park",
  "n08609487": "Virgin Islands National Park",
  "n08609685": "Voyageurs National Park",
  "n08609858": "Wind Cave National Park",
  "n08610023": "windward",
  "n08610141": "Wrangell-St. Elias National Park",
  "n08610305": "Yellowstone National Park",
  "n08610646": "Yosemite National Park",
  "n08610810": "Zion National Park",
  "n08610975": "nesting place",
  "n08611063": "no-go area",
  "n08611218": "no man's land",
  "n08611339": "nombril",
  "n08611421": "no-parking zone",
  "n08611532": "north celestial pole",
  "n08611662": "northern hemisphere",
  "n08611836": "North Pole",
  "n08611954": "old country",
  "n08612049": "orbit, celestial orbit",
  "n08612340": "orbit, electron orbit",
  "n08612498": "geosynchronous orbit",
  "n08612644": "geostationary orbit",
  "n08612786": "outline, lineation",
  "n08613000": "coastline",
  "n08613078": "paper route, paper round",
  "n08613202": "profile",
  "n08613345": "silhouette",
  "n08613472": "outside, exterior",
  "n08613593": "outside, exterior",
  "n08613733": "outdoors, out-of-doors, open air, open",
  "n08613985": "outstation, outpost",
  "n08614104": "overlook",
  "n08614229": "paddy, paddy field, rice paddy",
  "n08614357": "panhandle",
  "n08614545": "parade ground",
  "n08614632": "fairground",
  "n08614746": "midway",
  "n08614900": "fairway",
  "n08615001": "parish",
  "n08615149": "park, parkland",
  "n08615374": "park, commons, common, green",
  "n08615638": "parking lot, car park, park, parking area",
  "n08615810": "parking space, parking zone",
  "n08615926": "parts",
  "n08616050": "pasture, pastureland, grazing land, lea, ley",
  "n08616311": "path, route, itinerary",
  "n08616872": "beeline",
  "n08616985": "circuit",
  "n08617311": "crosscut",
  "n08617399": "supply line, supply route",
  "n08617516": "line of fire",
  "n08617622": "migration route, flyway",
  "n08617751": "fairway",
  "n08617873": "patriarchate",
  "n08617963": "peak, crown, crest, top, tip, summit",
  "n08618379": "periapsis, point of periapsis",
  "n08618669": "perigee",
  "n08618831": "perihelion",
  "n08619006": "perijove",
  "n08619112": "periselene, perilune",
  "n08619250": "pesthole",
  "n08619343": "picnic area, picnic ground",
  "n08619457": "pinnacle",
  "n08619524": "prairie",
  "n08619620": "public square, square",
  "n08619795": "plaza, place, piazza",
  "n08619949": "toll plaza",
  "n08620061": "point",
  "n08620763": "abutment",
  "n08620881": "pole",
  "n08621099": "pole, celestial pole",
  "n08621283": "pole position",
  "n08621393": "polls",
  "n08621472": "pride of place",
  "n08621598": "position, place",
  "n08622201": "position",
  "n08622340": "pressure point",
  "n08622586": "military position, position",
  "n08622763": "anomaly",
  "n08622950": "site, situation",
  "n08623260": "active site",
  "n08623424": "close quarters",
  "n08623568": "locus",
  "n08623676": "locus of infection",
  "n08623800": "restriction site",
  "n08623927": "setting",
  "n08624095": "juxtaposition",
  "n08624196": "lie",
  "n08624385": "post, station",
  "n08624656": "pitch",
  "n08624891": "landmark",
  "n08625073": "right",
  "n08625297": "stage right, right stage",
  "n08625462": "left",
  "n08625684": "stage left, left stage",
  "n08625846": "back, rear",
  "n08626080": "front",
  "n08626283": "municipality",
  "n08626522": "new town",
  "n08626758": "perch",
  "n08626845": "potter's field",
  "n08626947": "prefecture",
  "n08627161": "premises",
  "n08627316": "protectorate, associated state",
  "n08627554": "quadrant",
  "n08627665": "quadrant",
  "n08627805": "quadrant, quarter-circle",
  "n08627919": "quarter",
  "n08628141": "kasbah, casbah",
  "n08628308": "medina",
  "n08628414": "Queen Maud Land",
  "n08628578": "radius",
  "n08628746": "rain shadow",
  "n08628921": "range, reach",
  "n08629199": "range",
  "n08629508": "rear, backside, back end",
  "n08629746": "rearward",
  "n08629922": "red line",
  "n08630039": "region, part",
  "n08630985": "region",
  "n08631531": "possession",
  "n08631750": "antipodes",
  "n08631945": "rifle range, rifle shot",
  "n08632096": "unknown, unknown region, terra incognita",
  "n08632258": "staging area",
  "n08632423": "open, clear",
  "n08632678": "rhumb line, rhumb, loxodrome",
  "n08632894": "declination, celestial latitude, dec",
  "n08633191": "right ascension, RA, celestial longitude",
  "n08633683": "waterfront",
  "n08633851": "seafront",
  "n08633957": "port",
  "n08637771": "entrepot, transshipment center",
  "n08638033": "free port",
  "n08638141": "home port",
  "n08638260": "outport",
  "n08638442": "port of entry, point of entry",
  "n08639058": "seaport, haven, harbor, harbour",
  "n08639367": "coaling station",
  "n08639479": "port of call",
  "n08639586": "free port, free zone",
  "n08639776": "anchorage, anchorage ground",
  "n08639951": "treaty port",
  "n08640111": "mooring, moorage, berth, slip",
  "n08640300": "roads, roadstead",
  "n08640392": "dockyard",
  "n08640531": "resort, resort hotel, holiday resort",
  "n08640739": "resort area, playground, vacation spot",
  "n08640962": "rough",
  "n08641113": "vicinity, locality, neighborhood, neighbourhood, neck of the woods",
  "n08641744": "gold coast",
  "n08641944": "'hood",
  "n08642037": "place",
  "n08642145": "block, city block",
  "n08642331": "neighborhood",
  "n08642517": "proximity",
  "n08642632": "presence, front",
  "n08642870": "rendezvous",
  "n08643015": "retreat",
  "n08643267": "ashram",
  "n08643386": "ashram",
  "n08643499": "Camp David",
  "n08643648": "nook",
  "n08643729": "nest",
  "n08643806": "pleasance",
  "n08643933": "safety, refuge",
  "n08644045": "harborage, harbourage",
  "n08644213": "danger",
  "n08644327": "danger line",
  "n08644426": "rookery",
  "n08644552": "Rubicon",
  "n08644722": "country, rural area",
  "n08645033": "countryside",
  "n08645104": "scrubland",
  "n08645212": "weald",
  "n08645318": "wold",
  "n08645420": "safari park",
  "n08645571": "sanctum, sanctum sanctorum",
  "n08645749": "sandlot",
  "n08645847": "savanna, savannah",
  "n08645963": "scene",
  "n08646188": "light",
  "n08646306": "darkness, dark, shadow",
  "n08646486": "field of honor",
  "n08646566": "stage",
  "n08646787": "scenery",
  "n08646902": "landscape",
  "n08647049": "seascape",
  "n08647121": "separation",
  "n08647264": "schoolyard",
  "n08647354": "churchyard, God's acre",
  "n08647457": "scour",
  "n08647616": "seat, place",
  "n08647945": "seat",
  "n08648153": "section, plane section",
  "n08648322": "section",
  "n08648658": "sector",
  "n08648781": "service area",
  "n08648917": "showplace",
  "n08649067": "shrubbery",
  "n08649167": "side",
  "n08649345": "side",
  "n08649711": "bedside",
  "n08649873": "blind side",
  "n08649984": "dockside",
  "n08650073": "east side",
  "n08650157": "hand",
  "n08650305": "north side",
  "n08650391": "shipside",
  "n08650507": "south side",
  "n08650593": "west side",
  "n08650677": "scrimmage line, line of scrimmage",
  "n08650919": "service line, baseline",
  "n08651104": "sideline, out of bounds",
  "n08651247": "site, land site",
  "n08651735": "skyline",
  "n08651832": "slum, slum area",
  "n08652036": "shantytown",
  "n08652156": "skid road",
  "n08652256": "skid row",
  "n08652376": "ski resort",
  "n08652480": "solitude",
  "n08652551": "southern hemisphere",
  "n08652717": "South Pole",
  "n08652854": "south celestial pole",
  "n08652970": "space",
  "n08653314": "air",
  "n08653474": "vacuum, vacuity",
  "n08653570": "sphere, sphere of influence",
  "n08653706": "stand",
  "n08653873": "start, starting line, scratch, scratch line",
  "n08654029": "touchline",
  "n08654127": "yard line",
  "n08654260": "eparchy",
  "n08654360": "state, province",
  "n08655464": "American state",
  "n08656590": "station",
  "n08656750": "Stonehenge",
  "n08656893": "stop",
  "n08657100": "stopover, way station",
  "n08657249": "stratum",
  "n08657748": "Strand",
  "n08657887": "substrate, substratum",
  "n08658001": "superstrate, superstratum",
  "n08658118": "horizon",
  "n08658309": "soil horizon",
  "n08658469": "A-horizon, A horizon",
  "n08658612": "B-horizon, B horizon",
  "n08658776": "C-horizon, C horizon",
  "n08658918": "geological horizon",
  "n08659076": "seam, bed",
  "n08659242": "coal seam",
  "n08659331": "coalface",
  "n08659446": "field",
  "n08659663": "coalfield",
  "n08659759": "gasfield",
  "n08659861": "oilfield",
  "n08659993": "corner",
  "n08660146": "substrate, substratum",
  "n08660339": "surface",
  "n08660817": "tank farm",
  "n08660934": "target, target area",
  "n08661063": "ground zero",
  "n08661170": "tax haven",
  "n08661277": "tee, teeing ground",
  "n08661467": "toxic site, toxic waste area, Superfund site",
  "n08661697": "orphan site",
  "n08661878": "Temperate Zone",
  "n08662155": "North Temperate Zone",
  "n08662288": "South Temperate Zone",
  "n08662427": "terreplein",
  "n08662570": "testing ground, laboratory",
  "n08662914": "theme park",
  "n08663051": "three-mile limit",
  "n08663156": "tip",
  "n08663354": "top, top side, upper side, upside",
  "n08663703": "desktop",
  "n08663774": "rooftop",
  "n08663860": "top",
  "n08664184": "head",
  "n08664365": "tiptop",
  "n08664443": "topographic point, place, spot",
  "n08665101": "pool, puddle",
  "n08665281": "Torrid Zone, tropical zone, tropics",
  "n08665504": "town",
  "n08671168": "burg",
  "n08671293": "boom town",
  "n08671382": "cow town, cowtown",
  "n08671509": "ghost town",
  "n08671644": "hometown",
  "n08671822": "Main Street",
  "n08672073": "market town",
  "n08672199": "township, town",
  "n08672397": "ward",
  "n08672562": "settlement",
  "n08672738": "village, hamlet",
  "n08673039": "kampong, campong",
  "n08673131": "kraal",
  "n08673273": "pueblo",
  "n08673395": "tract, piece of land, piece of ground, parcel of land, parcel",
  "n08674251": "subdivision",
  "n08674344": "subtropics, semitropics",
  "n08674464": "mine field",
  "n08674563": "terrain",
  "n08674739": "plot, plot of land, plot of ground, patch",
  "n08674970": "lot",
  "n08675145": "tropic",
  "n08675538": "Tropic of Cancer",
  "n08675665": "Tropic of Capricorn",
  "n08675795": "trust territory, trusteeship",
  "n08675967": "urban area, populated area",
  "n08676253": "barrio",
  "n08676349": "barrio",
  "n08676499": "used-car lot",
  "n08676622": "vacant lot, building site",
  "n08676760": "Van Allen belt",
  "n08676951": "vanishing point",
  "n08677075": "vantage",
  "n08677250": "vantage point, viewpoint",
  "n08677424": "veld, veldt",
  "n08677525": "venue",
  "n08677628": "venue, locale, locus",
  "n08677801": "vertex, peak, apex, acme",
  "n08677991": "vertical circle",
  "n08678151": "viceroyalty",
  "n08678253": "Victoria Land",
  "n08678405": "village green",
  "n08678515": "warren, rabbit warren",
  "n08678615": "watering place, watering hole, spa",
  "n08678783": "waterline, water line, water level",
  "n08679011": "water line, watermark",
  "n08679167": "high-water mark",
  "n08679269": "low-water mark",
  "n08679369": "watershed, water parting, divide",
  "n08679562": "continental divide",
  "n08679807": "Great Divide",
  "n08679972": "direction, way",
  "n08680237": "trade route",
  "n08680363": "Silk Road",
  "n08680571": "Northwest Passage",
  "n08680888": "bearing, heading, aim",
  "n08681084": "tack",
  "n08681222": "course, trend",
  "n08681422": "east-west direction",
  "n08681534": "north-south direction",
  "n08681649": "qibla",
  "n08681777": "tendency, trend",
  "n08681966": "wave front, wavefront",
  "n08682188": "Wilkes Land",
  "n08682389": "western hemisphere, occident, New World",
  "n08682575": "West, Occident",
  "n08682819": "West, western United States",
  "n08683177": "Wild West",
  "n08683286": "wheatfield, wheat field",
  "n08683383": "whereabouts",
  "n08683548": "wilderness, wild",
  "n08683841": "winner's circle",
  "n08683986": "tape",
  "n08684107": "wire",
  "n08684191": "workspace",
  "n08684294": "yard",
  "n08684572": "tiltyard",
  "n08684676": "yard",
  "n08684769": "zenith",
  "n08685019": "exaltation",
  "n08685188": "zodiac",
  "n08685677": "sign of the zodiac, star sign, sign, mansion, house, planetary house",
  "n08686129": "Aries, Aries the Ram, Ram",
  "n08686332": "Taurus, Taurus the Bull, Bull",
  "n08686495": "Gemini, Gemini the Twins, Twins",
  "n08686658": "Cancer, Cancer the Crab, Crab",
  "n08686821": "Leo, Leo the Lion, Lion",
  "n08686979": "Virgo, Virgo the Virgin, Virgin",
  "n08687150": "Libra, Libra the Balance, Balance, Libra the Scales",
  "n08687345": "Scorpio, Scorpio the Scorpion, Scorpion",
  "n08687525": "Sagittarius, Sagittarius the Archer, Archer",
  "n08687709": "Capricorn, Capricorn the Goat, Goat",
  "n08687884": "Aquarius, Aquarius the Water Bearer, Water Bearer",
  "n08688076": "Pisces, Pisces the Fishes, Fish",
  "n08688247": "zone",
  "n08688424": "buffer zone, buffer",
  "n08688590": "combat zone, combat area",
  "n08688779": "war zone",
  "n08689028": "bridgehead, foothold",
  "n08689357": "airhead",
  "n08689449": "beachhead",
  "n08689646": "combat zone, tenderloin",
  "n08689771": "turf",
  "n08689873": "danger zone",
  "n08689947": "demilitarized zone, DMZ",
  "n08690194": "drop zone, dropping zone",
  "n08690352": "kill zone, killing zone",
  "n08690479": "killing field",
  "n08690665": "enterprise zone",
  "n08690792": "outskirt, fringe",
  "n08690974": "strike zone",
  "n08691188": "tidal zone",
  "n08691276": "time zone",
  "n08691445": "transit zone",
  "n08691669": "national capital",
  "n08695198": "provincial capital",
  "n08695539": "state capital",
  "n08696737": "Continent",
  "n08696931": "European country, European nation",
  "n08697827": "Scandinavian country, Scandinavian nation",
  "n08698038": "Balkans",
  "n08698126": "Balkan country, Balkan nation, Balkan state",
  "n08698379": "African country, African nation",
  "n08699426": "East Africa",
  "n08699654": "Namibia, Republic of Namibia, South West Africa",
  "n08700133": "Windhoek",
  "n08700255": "Asian country, Asian nation",
  "n08701161": "Cappadocia",
  "n08701296": "Galatia",
  "n08701410": "Phrygia",
  "n08701555": "Colossae",
  "n08701719": "Pontus",
  "n08701942": "Asia Minor, Anatolia",
  "n08702402": "South American country, South American nation",
  "n08702805": "North American country, North American nation",
  "n08703035": "Central American country, Central American nation",
  "n08703454": "Afghanistan, Islamic State of Afghanistan",
  "n08703972": "Herat",
  "n08704116": "Jalalabad",
  "n08704237": "Kabul, capital of Afghanistan",
  "n08704409": "Kandahar, Qandahar",
  "n08704554": "Mazar-i-Sharif",
  "n08704665": "Illyria",
  "n08704822": "Albania, Republic of Albania",
  "n08705091": "Tirana, Albanian capital",
  "n08705251": "Durres, Durazzo",
  "n08705397": "Algeria, Algerie, Democratic and Popular Republic of Algeria",
  "n08706058": "Algiers, Algerian capital",
  "n08706247": "Annaba",
  "n08706399": "Batna",
  "n08706502": "Blida",
  "n08706663": "Oran",
  "n08706823": "Constantine",
  "n08707035": "Djanet",
  "n08707145": "Hippo, Hippo Regius",
  "n08707332": "Reggane",
  "n08707431": "Timgad",
  "n08707587": "Timimoun",
  "n08707710": "Numidia",
  "n08707917": "Angola, Republic of Angola",
  "n08708304": "Luanda, Angolan capital",
  "n08708481": "Huambo, Nova Lisboa",
  "n08708609": "Lobito",
  "n08708742": "Anguilla",
  "n08708870": "Aran Islands",
  "n08709038": "Caribbean",
  "n08709399": "Cayman Islands",
  "n08709593": "George Town",
  "n08709704": "Antigua and Barbuda",
  "n08709945": "Antigua",
  "n08710113": "Barbuda",
  "n08710219": "Redonda",
  "n08710325": "St. John's, Saint John's, capital of Antigua and Barbuda",
  "n08710535": "Bengal",
  "n08710678": "Bermuda, Bermudas",
  "n08710873": "Hamilton",
  "n08710951": "Bermuda Triangle",
  "n08711143": "Bouvet Island",
  "n08711300": "Montserrat",
  "n08711468": "Patagonia",
  "n08711655": "Triple Frontier",
  "n08711974": "Argentina, Argentine Republic",
  "n08712698": "Bahia Blanca",
  "n08712876": "Buenos Aires, capital of Argentina",
  "n08713136": "Cordoba, Cordova",
  "n08713285": "Moron",
  "n08713405": "Rosario",
  "n08713539": "Vicente Lopez",
  "n08713655": "pampas",
  "n08713772": "Balkan Peninsula, Balkans",
  "n08714132": "Bulgaria, Republic of Bulgaria",
  "n08714458": "Sofia, Serdica, Bulgarian capital",
  "n08714624": "Dobrich, Tolbukhin",
  "n08714795": "Plovdiv, Philippopolis",
  "n08714966": "Varna",
  "n08715110": "Southeast Asia",
  "n08715390": "Myanmar, Union of Burma, Burma",
  "n08715777": "Yangon, Rangoon",
  "n08715952": "Mandalay",
  "n08716076": "Moulmein, Mawlamyine",
  "n08716219": "Burundi, Republic of Burundi",
  "n08716517": "Bujumbura, Usumbura, capital of Burundi",
  "n08716738": "Cambodia, Kingdom of Cambodia, Kampuchea",
  "n08717059": "Phnom Penh, Pnom Penh, Cambodian capital",
  "n08717209": "Cameroon, Republic of Cameroon, Cameroun",
  "n08717510": "Yaounde, capital of Cameroon",
  "n08717629": "Douala",
  "n08717730": "Cape Verde Islands",
  "n08717915": "Cape Verde, Republic of Cape Verde",
  "n08718094": "Praia, Cidade de Praia, capital of Cape Verde",
  "n08718253": "Sao Tiago Island",
  "n08718391": "Falkland Islands",
  "n08718577": "Central African Republic, Central Africa",
  "n08718801": "Bangui, capital of Central Africa",
  "n08718962": "Ceylon",
  "n08719100": "Sri Lanka, Democratic Socialist Republic of Sri Lanka, Ceylon",
  "n08719465": "Colombo, capital of Sri Lanka",
  "n08719705": "Kandy",
  "n08719892": "Eelam, Tamil Eelam",
  "n08720037": "Chad, Republic of Chad, Tchad",
  "n08720280": "N'Djamena, Ndjamena, Fort-Lamy, capital of Chad",
  "n08720481": "Chile, Republic of Chile",
  "n08721145": "Antofagasta",
  "n08721286": "Chiloe",
  "n08721449": "Concepcion",
  "n08721559": "Gran Santiago, Santiago, Santiago de Chile, capital of Chile",
  "n08721796": "Punta Arenas",
  "n08721961": "Temuco",
  "n08722084": "Valparaiso",
  "n08722270": "Vina del Mar",
  "n08722394": "Tierra del Fuego",
  "n08722645": "Cape Horn",
  "n08722844": "Manchuria",
  "n08723006": "China, People's Republic of China, mainland China, Communist China, Red China, PRC, Cathay",
  "n08724545": "Turkistan, Turkestan",
  "n08724726": "Beijing, Peking, Peiping, capital of Red China",
  "n08724972": "Forbidden City",
  "n08725161": "Chongqing, Chungking",
  "n08725336": "Guangdong, Kwangtung, Guangdong province",
  "n08725454": "Guangzhou, Kuangchou, Kwangchow, Canton",
  "n08725692": "Gansu, Kansu, Gansu province",
  "n08725926": "Hebei, Hopei, Hopeh, Hebei province",
  "n08726072": "Hunan, Hunan province",
  "n08726305": "Szechwan, Sichuan, Szechuan, Szechwan province",
  "n08726463": "Yunnan, Yunnan province",
  "n08726582": "Lanzhou, Lanchou, Lanchow",
  "n08726745": "Luda, Luta",
  "n08727003": "Dalian, Talien, Dairen",
  "n08727230": "Luoyang, Loyang",
  "n08727396": "Lushun, Port Arthur",
  "n08727606": "Hangzhou, Hangchow",
  "n08727806": "Nanchang, Nan-chang",
  "n08727945": "Nanning, Nan-ning",
  "n08728066": "Nanjing, Nanking",
  "n08728268": "Shanghai",
  "n08728462": "Shenyang, Mukden, Moukden, Fengtien",
  "n08728595": "Taiyuan",
  "n08728749": "Tangshan",
  "n08728882": "Tianjin, Tientsin, T'ien-ching",
  "n08729094": "Grand Canal",
  "n08729283": "Wuhan",
  "n08729452": "Xian, Sian, Singan, Changan, Hsian",
  "n08729626": "Xinjiang, Sinkiang, Xinjiang Uighur Autonomous Region",
  "n08729971": "Inner Mongolia, Nei Monggol",
  "n08730218": "Hohhot",
  "n08730354": "Taiwan, Formosa",
  "n08730550": "Taiwan, China, Nationalist China, Republic of China",
  "n08730895": "Taipei, Taipeh, capital of Taiwan",
  "n08731057": "Taichung",
  "n08731148": "Hong Kong",
  "n08731423": "Macao, Macau",
  "n08731606": "Indochina, Indochinese peninsula",
  "n08731953": "French Indochina",
  "n08732116": "Colombia, Republic of Colombia",
  "n08732807": "Barranquilla",
  "n08732979": "Bogota, capital of Colombia",
  "n08733163": "Cali",
  "n08733291": "Medellin",
  "n08733415": "Cartagena",
  "n08733563": "Soledad",
  "n08733690": "Comoro Islands, Iles Comores",
  "n08733897": "Comoros, Federal Islamic Republic of the Comoros",
  "n08734044": "Congo, Republic of the Congo, French Congo",
  "n08734250": "Brazzaville",
  "n08734385": "Congo, Democratic Republic of the Congo, Zaire, Belgian Congo",
  "n08734853": "Goma",
  "n08735008": "Kananga, Luluabourg",
  "n08735164": "Kinshasa, Leopoldville",
  "n08735345": "Lubumbashi, Elisabethville",
  "n08735564": "Mesoamerica",
  "n08735705": "Central America",
  "n08736107": "Costa Rica, Republic of Costa Rica",
  "n08736376": "San Jose, capital of Costa Rica",
  "n08736517": "Ivory Coast, Cote d'Ivoire, Republic of Cote d'Ivoire",
  "n08736779": "Abidjan",
  "n08736956": "Yamoussukro",
  "n08737041": "Guatemala, Republic of Guatemala",
  "n08737376": "Guatemala City, capital of Guatemala",
  "n08737521": "Belize, British Honduras",
  "n08737716": "Honduras, Republic of Honduras",
  "n08738014": "Tegucigalpa, Honduran capital",
  "n08738151": "San Pedro Sula",
  "n08738272": "El Salvador, Republic of El Salvador, Salvador",
  "n08738531": "San Salvador, Salvadoran capital",
  "n08738715": "Santa Ana",
  "n08738820": "Nicaragua, Republic of Nicaragua",
  "n08739047": "Managua, capital of Nicaragua, Nicaraguan capital",
  "n08739206": "Panama, Republic of Panama",
  "n08739512": "Panama City, capital of Panama, Panamanian capital",
  "n08739669": "Colon, Aspinwall",
  "n08739829": "Panama Canal Zone, Canal Zone",
  "n08740022": "Yucatan, Yucatan Peninsula",
  "n08740230": "Yucatan",
  "n08740367": "Merida",
  "n08740483": "Campeche",
  "n08740617": "Campeche",
  "n08740729": "Cancun",
  "n08740875": "Mexico, United Mexican States",
  "n08742205": "Acapulco, Acapulco de Juarez",
  "n08742455": "Chihuahua",
  "n08742578": "Chihuahua",
  "n08742743": "Ciudad Juarez, Juarez",
  "n08742892": "Ciudad Victoria",
  "n08743003": "Coahuila",
  "n08743125": "Culiacan",
  "n08743229": "Durango, Victoria de Durango",
  "n08743370": "Guadalajara",
  "n08743557": "Hermosillo",
  "n08743691": "Leon",
  "n08743786": "Matamoros",
  "n08743945": "Mazatlan",
  "n08744105": "Mexicali",
  "n08744236": "Mexico City, Ciudad de Mexico, Mexican capital, capital of Mexico",
  "n08744509": "Monterrey",
  "n08744626": "Nogales",
  "n08744750": "Oaxaca, Oaxaca de Juarez",
  "n08744871": "Orizaba",
  "n08745011": "Puebla, Puebla de Zaragoza, Heroica Puebla de Zaragoza",
  "n08745249": "Quintana Roo",
  "n08745407": "San Luis Potosi",
  "n08745513": "Santa Maria del Tule",
  "n08745687": "Tabasco",
  "n08745800": "Tepic",
  "n08745901": "Tampico",
  "n08746023": "Torreon",
  "n08746147": "Tijuana",
  "n08746363": "Tuxtla Gutierrez",
  "n08746475": "Veracruz",
  "n08746636": "Villahermosa, Villa Hermosa",
  "n08746797": "Guadalupe Island",
  "n08746942": "Caribbean Island",
  "n08747054": "West Indies, the Indies",
  "n08747494": "British West Indies",
  "n08747737": "Antilles",
  "n08747887": "French West Indies",
  "n08748076": "Greater Antilles",
  "n08748280": "Lesser Antilles, Caribees",
  "n08748499": "Netherlands Antilles",
  "n08748794": "Aruba",
  "n08748917": "Bonaire",
  "n08749042": "Curacao",
  "n08749167": "Saba",
  "n08749312": "Saint Eustatius, St. Eustatius",
  "n08749447": "Leeward Islands",
  "n08749650": "Saint Martin, St. Martin, Saint Maarten, St. Maarten",
  "n08749864": "Windward Islands, Windward Isles",
  "n08750151": "Cuba",
  "n08750334": "Cuba, Republic of Cuba",
  "n08750612": "Havana, capital of Cuba, Cuban capital",
  "n08750822": "Santiago de Cuba, Santiago",
  "n08750986": "Guantanamo",
  "n08751126": "Guadeloupe",
  "n08751317": "Hispaniola, Haiti, Hayti",
  "n08751494": "Haiti, Republic of Haiti",
  "n08751885": "Port-au-Prince, Haitian capital",
  "n08752021": "Dominican Republic",
  "n08752293": "Santo Domingo, Ciudad Trujillo, capital of the Dominican Republic",
  "n08752671": "Santiago de los Caballeros, Santiago",
  "n08752814": "Puerto Rico, Porto Rico",
  "n08752974": "Puerto Rico, Porto Rico, Commonwealth of Puerto Rico, PR",
  "n08753294": "San Juan",
  "n08753412": "Culebra",
  "n08753561": "Vieques",
  "n08753729": "Jamaica",
  "n08753933": "Jamaica",
  "n08754238": "Kingston, capital of Jamaica, Jamaican capital",
  "n08754388": "Montego Bay",
  "n08754529": "Virgin Islands",
  "n08754814": "British Virgin Islands",
  "n08755003": "United States Virgin Islands, American Virgin Islands, VI",
  "n08755214": "Barbados",
  "n08755436": "Barbados",
  "n08755664": "Bridgetown, capital of Barbados",
  "n08755852": "Trinidad",
  "n08756052": "Tobago",
  "n08756202": "Trinidad and Tobago, Republic of Trinidad and Tobago",
  "n08756511": "Port of Spain, Port-of-Spain, capital of Trinidad and Tobago",
  "n08756735": "Cyprus",
  "n08756884": "Cyprus, Republic of Cyprus",
  "n08757132": "Nicosia, capital of Cyprus",
  "n08757264": "Czech Republic",
  "n08757569": "Czechoslovakia",
  "n08757791": "Pilsen, Plzen",
  "n08757926": "Prague, Praha, Prag, Czech capital",
  "n08758172": "Austerlitz",
  "n08758334": "Brno, Brunn",
  "n08758487": "Ostrava",
  "n08758679": "Moravia",
  "n08758882": "Bohemia",
  "n08759013": "Slovakia, Slovak Republic",
  "n08759263": "Bratislava, capital of Slovakia, Pressburg, Pozsony",
  "n08759420": "Benin, Republic of Benin, Dahomey",
  "n08759684": "Porto Novo, capital of Benin",
  "n08759852": "Cotonou",
  "n08759986": "Togo, Togolese Republic",
  "n08760228": "Lome, capital of Togo",
  "n08760393": "northern Europe",
  "n08760510": "Scandinavia",
  "n08760856": "Scandinavia, Scandinavian Peninsula",
  "n08761039": "Jutland, Jylland",
  "n08761244": "Denmark, Kingdom of Denmark, Danmark",
  "n08761697": "Zealand, Seeland, Sjaelland",
  "n08761868": "Copenhagen, Kobenhavn, Danish capital",
  "n08762104": "Arhus, Aarhus",
  "n08762243": "Aalborg, Alborg",
  "n08762380": "Viborg",
  "n08762495": "Djibouti, Republic of Djibouti, Afars and Issas",
  "n08762823": "Djibouti, capital of Djibouti",
  "n08763010": "Dominica",
  "n08763193": "Dominica, Commonwealth of Dominica",
  "n08763387": "Roseau",
  "n08763500": "Equatorial Guinea, Republic of Equatorial Guinea, Spanish Guinea",
  "n08763765": "Malabo",
  "n08763932": "Bioko",
  "n08764107": "Norway, Kingdom of Norway, Norge, Noreg",
  "n08764561": "Svalbard",
  "n08764717": "Spitsbergen, Spitzbergen",
  "n08764899": "Lofoten",
  "n08765069": "Oslo, Christiania, capital of Norway",
  "n08765315": "Bergen",
  "n08765460": "Stavanger",
  "n08765623": "Trondheim, Nidaros",
  "n08765771": "Lindesnes, Naze",
  "n08765890": "Sweden, Kingdom of Sweden, Sverige",
  "n08766236": "Stockholm, capital of Sweden",
  "n08766455": "Malmo",
  "n08766571": "Lund",
  "n08766667": "Goteborg, Goeteborg, Gothenburg",
  "n08766846": "Uppsala, Upsala",
  "n08766988": "Germany, Federal Republic of Germany, Deutschland, FRG",
  "n08768647": "East Germany, German Democratic Republic",
  "n08768881": "West Germany, Federal Republic of Germany",
  "n08769179": "Saxony, Sachsen, Saxe",
  "n08769329": "Lower Saxony",
  "n08769439": "Aachen, Aken, Aix-la-Chapelle",
  "n08769645": "Berlin, German capital",
  "n08769836": "West Berlin",
  "n08770013": "Bremen",
  "n08770274": "Bremerhaven",
  "n08770518": "Chemnitz, Karl-Marx-Stadt",
  "n08770718": "Dortmund",
  "n08770932": "Dresden",
  "n08771116": "Leipzig",
  "n08771277": "Solingen",
  "n08771400": "Weimar",
  "n08771596": "Bavaria",
  "n08771841": "Hameln, Hamelin",
  "n08772028": "Hohenlinden",
  "n08772137": "Bonn",
  "n08772307": "Cologne, Koln",
  "n08772551": "Braunschweig, Brunswick",
  "n08772667": "Dusseldorf",
  "n08772794": "Essen",
  "n08772922": "Frankfurt on the Main, Frankfurt, Frankfort",
  "n08773098": "Halle, Halle-an-der-Saale",
  "n08773336": "Hamburg",
  "n08773679": "Hannover, Hanover",
  "n08773880": "Lubeck",
  "n08774073": "Mannheim",
  "n08774227": "Munich, Muenchen",
  "n08774374": "Nuremberg, Nurnberg",
  "n08774546": "Potsdam",
  "n08774704": "Rostock",
  "n08774912": "Stuttgart",
  "n08775053": "Wiesbaden",
  "n08775179": "Wurzburg, Wuerzburg",
  "n08775297": "Rhineland, Rheinland",
  "n08775439": "Palatinate, Pfalz",
  "n08775597": "Brandenburg",
  "n08775784": "Prussia, Preussen",
  "n08776138": "Ruhr, Ruhr Valley",
  "n08776320": "Thuringia",
  "n08776435": "East Timor",
  "n08776687": "Ecuador, Republic of Ecuador",
  "n08777130": "Guayaquil",
  "n08777233": "Quito, capital of Ecuador",
  "n08777348": "Galapagos Islands, Galapagos",
  "n08777544": "Eritrea, State of Eritrea",
  "n08777815": "Asmara, Asmera",
  "n08777919": "Massawa",
  "n08778061": "Ethiopia, Federal Democratic Republic of Ethiopia, Yaltopya, Abyssinia",
  "n08778401": "Addis Ababa, New Flower, capital of Ethiopia",
  "n08778597": "Fiji Islands, Fijis",
  "n08778936": "Viti Levu",
  "n08779042": "Vanua Levu",
  "n08779149": "Fiji, Republic of Fiji",
  "n08779375": "Suva",
  "n08779504": "Finland, Republic of Finland, Suomi",
  "n08779830": "Karelia",
  "n08780018": "Helsinki, Helsingfors, capital of Finland, Finnish capital",
  "n08780282": "Espoo",
  "n08780380": "Tampere, Tammerfors",
  "n08780510": "Aland islands, Aaland islands, Ahvenanmaa",
  "n08780720": "Mariehamn, Maarianhamina",
  "n08780881": "Greece, Hellenic Republic, Ellas",
  "n08782319": "Greece",
  "n08782490": "Achaea",
  "n08782627": "Aegean island",
  "n08782847": "Aegina, Aigina",
  "n08782976": "Chios, Khios",
  "n08783149": "Cyclades, Kikladhes",
  "n08783286": "Dodecanese, Dhodhekanisos",
  "n08783444": "Doris",
  "n08783583": "Lesbos, Lesvos, Mytilene",
  "n08783812": "Rhodes, Rodhos",
  "n08784104": "Aeolis, Aeolia",
  "n08784333": "Crete, Kriti",
  "n08784581": "Knossos, Cnossos, Cnossus",
  "n08784746": "Ithaca, Ithaki",
  "n08784905": "Egadi Islands, Aegadean Isles, Aegadean Islands, Isole Egadi, Aegates",
  "n08785132": "Athos, Mount Athos",
  "n08785343": "Athens, Athinai, capital of Greece, Greek capital",
  "n08785743": "Areopagus",
  "n08785958": "Dipylon gate, Dipylon",
  "n08786161": "Actium",
  "n08786283": "Attica",
  "n08786432": "Corinth, Korinthos",
  "n08786660": "Argos",
  "n08786855": "Delphi",
  "n08787049": "Mycenae",
  "n08787240": "Sparta",
  "n08787466": "Epirus",
  "n08787695": "Laconia",
  "n08787878": "Lycia",
  "n08788004": "Lydia",
  "n08788190": "Nemea",
  "n08788326": "Ephesus",
  "n08788588": "Patras, Patrai",
  "n08788887": "Troy, Ilion, Ilium",
  "n08789076": "Thebes",
  "n08789243": "Boeotia",
  "n08789447": "Plataea",
  "n08789605": "Thessaloniki, Salonika, Salonica, Thessalonica",
  "n08789835": "Stagira, Stagirus",
  "n08789970": "Thessalia, Thessaly",
  "n08790199": "Cynoscephalae",
  "n08790353": "Arcadia",
  "n08790495": "Peloponnese, Peloponnesus, Peloponnesian Peninsula",
  "n08790748": "Lemnos, Limnos",
  "n08790953": "Olympia",
  "n08791167": "Middle East, Mideast, Near East",
  "n08791978": "Mashriq",
  "n08792083": "Fertile Crescent",
  "n08792295": "Israel",
  "n08792548": "Israel, State of Israel, Yisrael, Zion, Sion",
  "n08793310": "Acre, Akko, Akka, Accho",
  "n08793489": "West Bank",
  "n08793746": "Nablus",
  "n08793914": "Galilee",
  "n08794193": "Nazareth",
  "n08794366": "Gaza Strip, Gaza",
  "n08794574": "Golan Heights, Golan",
  "n08794798": "Jerusalem, capital of Israel",
  "n08795232": "Bethlehem, Bayt Lahm, Bethlehem Ephrathah, Bethlehem-Judah",
  "n08795492": "Caesarea",
  "n08795654": "Sodom",
  "n08795880": "sodom",
  "n08795974": "Gomorrah, Gomorrha",
  "n08796219": "Calvary, Golgotha",
  "n08796351": "Zion, Sion",
  "n08796707": "Cotswolds, Cotswold Hills",
  "n08796844": "Cheviots, Cheviot Hills",
  "n08797013": "Pennines, Pennine Chain",
  "n08797254": "Seven Hills of Rome",
  "n08797412": "Palatine",
  "n08797619": "Wailing Wall",
  "n08797840": "Tel Aviv, Tel Aviv-Yalo, Tel Aviv-Jaffa",
  "n08798062": "Hefa, Haifa",
  "n08798195": "Jaffa, Joppa, Yafo",
  "n08798382": "Palestine, Canaan, Holy Land, Promised Land",
  "n08798771": "Palestine",
  "n08799123": "Judah, Juda",
  "n08799271": "Judea, Judaea",
  "n08799462": "Samaria",
  "n08799706": "Philistia",
  "n08799958": "Roman Republic",
  "n08800258": "Roman Empire",
  "n08800676": "Byzantine Empire, Byzantium, Eastern Roman Empire",
  "n08800911": "Western Roman Empire, Western Empire",
  "n08801099": "Byzantium",
  "n08801364": "Italian Peninsula",
  "n08801546": "Ticino, Tessin",
  "n08801678": "Italy, Italian Republic, Italia",
  "n08803382": "Italian region",
  "n08803883": "Pompeii",
  "n08804049": "Herculaneum",
  "n08804154": "Abruzzi, Abruzzi e Molise",
  "n08804319": "Aquila, L'Aquila, Aquila degli Abruzzi",
  "n08804487": "Basilicata, Lucania",
  "n08804662": "Bolzano",
  "n08804845": "Brescia",
  "n08804962": "Calabria",
  "n08805122": "Campania",
  "n08805386": "Ferrara",
  "n08805565": "Naples, Napoli",
  "n08805801": "Messina",
  "n08805953": "Capri",
  "n08806142": "Ischia",
  "n08806311": "Emilia-Romagna",
  "n08806458": "Bologna",
  "n08806617": "Friuli-Venezia Giulia",
  "n08806735": "Latium, Lazio",
  "n08806897": "Rome, Roma, Eternal City, Italian capital, capital of Italy",
  "n08807554": "Lateran",
  "n08807716": "Anzio",
  "n08807894": "Brindisi",
  "n08808077": "Tivoli, Tibur",
  "n08808292": "Liguria",
  "n08808452": "Genoa, Genova",
  "n08808614": "Lombardy, Lombardia",
  "n08808792": "Cremona",
  "n08808979": "La Spezia",
  "n08809165": "Milan, Milano",
  "n08809384": "Marche, Marches",
  "n08809492": "Molise",
  "n08809596": "Papal States",
  "n08809749": "Piedmont, Piemonte",
  "n08809910": "Pisa",
  "n08810051": "Syracuse, Siracusa",
  "n08810220": "Turin, Torino",
  "n08810358": "Puglia, Apulia",
  "n08810505": "Bari",
  "n08810631": "Sardinia, Sardegna",
  "n08810786": "Sardinia, Sardegna",
  "n08810999": "Sicily, Sicilia",
  "n08811215": "Sicily, Sicilia",
  "n08811473": "Palermo",
  "n08811653": "Cape Passero, Passero Cape",
  "n08811812": "Agrigento, Acragas",
  "n08811982": "Tuscany, Toscana",
  "n08812166": "Firenze, Florence",
  "n08812399": "Trentino-Alto Adige",
  "n08812552": "Trento, Trent",
  "n08812732": "Umbria",
  "n08812842": "Valle D'Aosta",
  "n08812952": "Veneto, Venezia-Euganea, Venetia",
  "n08813156": "Padua, Padova, Patavium",
  "n08813264": "Venice, Venezia",
  "n08813586": "Grand Canal",
  "n08813699": "Verona",
  "n08813807": "Etruria",
  "n08813978": "Romania, Roumania, Rumania",
  "n08814333": "Brasov",
  "n08814474": "Bucharest, Bucharesti, Bucuresti, capital of Romania",
  "n08814664": "Constantina",
  "n08814781": "Transylvania",
  "n08815046": "Rwanda, Rwandese Republic, Ruanda",
  "n08815346": "Kigali, capital of Rwanda",
  "n08815513": "Yugoslavia",
  "n08815858": "Croatia, Republic of Croatia, Hrvatska",
  "n08816236": "Serbia and Montenegro, Union of Serbia and Montenegro, Yugoslavia, Federal Republic of Yugoslavia, Jugoslavija",
  "n08816807": "Kosovo",
  "n08816969": "Serbia, Srbija",
  "n08817235": "Montenegro, Crna Gora",
  "n08817418": "Belgrade, Beograd, capital of Serbia and Montenegro",
  "n08817630": "Bosnia and Herzegovina, Republic of Bosnia and Herzegovina, Bosna i Hercegovina, Bosnia-Herzegovina, Bosnia",
  "n08818135": "Bosnia",
  "n08818247": "Sarajevo",
  "n08818444": "Slovenia, Republic of Slovenia, Slovenija",
  "n08818736": "Ljubljana",
  "n08818835": "Dubrovnik, Ragusa",
  "n08819016": "Split",
  "n08819128": "Zagreb",
  "n08819223": "Dalmatia",
  "n08819397": "Greenland, Gronland, Kalaallit Nunaat",
  "n08819683": "Baffin Island",
  "n08819883": "Labrador",
  "n08820121": "Canada",
  "n08821187": "Acadia",
  "n08821319": "Laurentian Plateau, Laurentian Highlands, Canadian Shield",
  "n08821578": "Maritime Provinces, Maritimes, Canadian Maritime Provinces",
  "n08821885": "Canadian province",
  "n08822202": "Alberta",
  "n08822427": "Banff",
  "n08822546": "Calgary",
  "n08822742": "Edmonton",
  "n08822855": "British Columbia",
  "n08823156": "Nanaimo",
  "n08823314": "Victoria",
  "n08823480": "Vancouver",
  "n08823728": "Vancouver Island",
  "n08823968": "Manitoba",
  "n08824152": "Winnipeg",
  "n08824323": "Churchill",
  "n08824484": "New Brunswick",
  "n08824654": "Fredericton",
  "n08824771": "Saint John, St. John",
  "n08824937": "Newfoundland and Labrador",
  "n08825211": "Newfoundland",
  "n08825321": "Saint John's, St. John's",
  "n08825477": "Northwest Territories",
  "n08825664": "Nunavut",
  "n08825987": "Arctic Archipelago",
  "n08826150": "Yellowknife",
  "n08826306": "Nova Scotia",
  "n08826699": "Cape Breton Island",
  "n08826841": "Nova Scotia",
  "n08827002": "Halifax",
  "n08827126": "Ontario",
  "n08827486": "Ottawa, Canadian capital, capital of Canada",
  "n08827689": "Hamilton",
  "n08827853": "Kingston",
  "n08828013": "Sault Sainte Marie",
  "n08828151": "Sudbury",
  "n08828296": "Thunder Bay",
  "n08828432": "Toronto",
  "n08828627": "Windsor",
  "n08828750": "Prince Edward Island",
  "n08828928": "Charlottetown",
  "n08829071": "Quebec",
  "n08829317": "Quebec, Quebec City",
  "n08829533": "Montreal",
  "n08829775": "Saskatchewan",
  "n08830005": "Regina",
  "n08830116": "Saskatoon",
  "n08830256": "Dawson",
  "n08830456": "Yukon, Yukon Territory",
  "n08830720": "Klondike",
  "n08830882": "Whitehorse",
  "n08831004": "Australia, Commonwealth of Australia",
  "n08832269": "Canberra, Australian capital, capital of Australia",
  "n08832447": "Australian state",
  "n08832691": "Queensland",
  "n08832877": "Brisbane",
  "n08833130": "New South Wales",
  "n08833295": "Sydney",
  "n08833524": "Wagga Wagga",
  "n08833682": "Victoria",
  "n08833809": "Melbourne",
  "n08833981": "Tasmania",
  "n08834123": "Tasmania",
  "n08834280": "Hobart",
  "n08834408": "South Australia",
  "n08834543": "Adelaide",
  "n08834654": "Western Australia",
  "n08834806": "Perth",
  "n08834916": "Northern Territory",
  "n08835058": "Darwin",
  "n08835188": "Norfolk Island",
  "n08835390": "Nullarbor Plain",
  "n08835675": "Aleutian Islands, Aleutians",
  "n08835875": "Oceania, Oceanica",
  "n08836165": "Australasia",
  "n08836329": "Austronesia",
  "n08836630": "Melanesia",
  "n08836886": "Micronesia",
  "n08837048": "Micronesia, Federated States of Micronesia, TT",
  "n08837457": "Kolonia",
  "n08837552": "Mariana Islands, Marianas, Ladrone Islands",
  "n08837864": "Northern Marianas, Northern Mariana Islands",
  "n08838070": "Saipan",
  "n08838319": "Guam, GU",
  "n08838556": "Wake Island, Wake",
  "n08838716": "Caroline Islands",
  "n08838887": "Marshall Islands",
  "n08839092": "Marshall Islands, Republic of the Marshall Islands",
  "n08839296": "Bikini",
  "n08839475": "Eniwetok",
  "n08839714": "Kwajalein",
  "n08839916": "Gilbert Islands",
  "n08840200": "Tuvalu, Ellice Islands",
  "n08840374": "Tuvalu",
  "n08840657": "Funafuti",
  "n08840749": "Kiribati, Republic of Kiribati",
  "n08840964": "Tarawa, Bairiki",
  "n08841075": "Gilbert and Ellice Islands",
  "n08841209": "Nauru, Nauru Island, Pleasant Island",
  "n08841483": "Nauru, Republic of Nauru",
  "n08841667": "Polynesia",
  "n08841956": "Malay Archipelago, East Indies, East India",
  "n08842258": "Sunda Islands",
  "n08842427": "Greater Sunda Islands",
  "n08842583": "Lesser Sunda Islands, Nusa Tenggara",
  "n08842819": "Bismarck Archipelago",
  "n08843085": "Admiralty Islands",
  "n08843215": "Borneo, Kalimantan",
  "n08843571": "Bougainville",
  "n08843735": "Guadalcanal",
  "n08843958": "New Britain",
  "n08844108": "New Caledonia",
  "n08844279": "New Guinea",
  "n08844557": "Papua New Guinea, Independent State of Papua New Guinea",
  "n08844923": "Papua",
  "n08845053": "Port Moresby, capital of Papua New Guinea",
  "n08845225": "New Ireland",
  "n08845366": "Austria-Hungary",
  "n08845555": "Austria, Republic of Austria, Oesterreich",
  "n08846135": "Tyrol, Tirol",
  "n08846324": "Vienna, Austrian capital, capital of Austria",
  "n08846626": "Graz",
  "n08846739": "Linz, Lentia",
  "n08846885": "Salzburg",
  "n08847024": "Innsbruck",
  "n08847165": "Wagram",
  "n08847268": "Bahamas, Commonwealth of the Bahamas, Bahama Islands",
  "n08847570": "Nassau, capital of the Bahamas",
  "n08847694": "Arabian Peninsula, Arabia",
  "n08848094": "Bahrain, State of Bahrain, Bahrein",
  "n08848421": "Bahrain, Bahrain Island, Bahrein, Bahrein Island",
  "n08848568": "Manama, capital of Bahrain",
  "n08848731": "Bangladesh, People's Republic of Bangladesh, Bangla Desh, East Pakistan",
  "n08849226": "Dhaka, Dacca, capital of Bangladesh",
  "n08849372": "Chittagong",
  "n08849549": "Flanders",
  "n08849753": "Belgium, Kingdom of Belgium, Belgique",
  "n08850450": "Bruxelles, Brussels, Belgian capital, capital of Belgium",
  "n08850663": "Aalst",
  "n08850741": "Antwerpen, Antwerp, Anvers",
  "n08851034": "Bruges, City of Bridges",
  "n08851364": "Charleroi",
  "n08851500": "Gent, Gand, Ghent",
  "n08851687": "Liege, Luik",
  "n08851830": "Namur",
  "n08852065": "Waterloo",
  "n08852209": "Bhutan, Kingdom of Bhutan",
  "n08852389": "Botswana, Republic of Botswana",
  "n08852685": "Gaborone, capital of Botswana",
  "n08852843": "Bolivia, Republic of Bolivia",
  "n08853308": "La Paz, capital of Bolivia",
  "n08853510": "Santa Cruz",
  "n08853612": "Sucre",
  "n08853741": "Brazil, Federative Republic of Brazil, Brasil",
  "n08854725": "Acre",
  "n08854855": "Belem, Para, Feliz Lusitania, Santa Maria de Belem, St. Mary of Bethlehem",
  "n08855126": "Belo Horizonte",
  "n08855308": "Brasilia, Brazilian capital, capital of Brazil",
  "n08855505": "Curitiba",
  "n08855609": "Joao Pessoa",
  "n08855763": "Governador Valadares",
  "n08855909": "Limeira",
  "n08856037": "Natal",
  "n08856162": "Osasco",
  "n08856266": "Rio de Janeiro, Rio",
  "n08856475": "Recife, Pernambuco",
  "n08856630": "Santos",
  "n08856793": "Sao Bernardo do Campo",
  "n08856945": "Sao Goncalo",
  "n08857099": "Sao Joao de Meriti",
  "n08857260": "Sao Jose dos Campos",
  "n08857405": "Sao Louis",
  "n08857529": "Sao Paulo",
  "n08857682": "British Empire",
  "n08858248": "British Isles",
  "n08858529": "British East Africa",
  "n08858713": "British West Africa",
  "n08858942": "Great Britain, GB",
  "n08859173": "Ireland, Hibernia, Emerald Isle",
  "n08860001": "Erin",
  "n08860123": "United Kingdom, UK, U.K., Britain, United Kingdom of Great Britain and Northern Ireland, Great Britain",
  "n08871007": "England",
  "n08872936": "Albion",
  "n08873067": "Anglia",
  "n08873147": "Blighty",
  "n08873269": "Lancaster",
  "n08873412": "Lake District, Lakeland",
  "n08873622": "London, Greater London, British capital, capital of the United Kingdom",
  "n08874273": "City of London, the City",
  "n08874469": "Home Counties",
  "n08874703": "Greenwich",
  "n08874920": "Bloomsbury",
  "n08875057": "Soho",
  "n08875202": "Wembley",
  "n08875369": "West End",
  "n08875547": "Westminster, City of Westminster",
  "n08875843": "Buckingham Palace",
  "n08875972": "Downing Street",
  "n08876139": "Pall Mall",
  "n08876278": "Houses of Parliament",
  "n08876435": "Westminster Abbey",
  "n08876773": "Wimbledon",
  "n08876975": "Manchester",
  "n08877208": "Hull, Kingston-upon Hull",
  "n08877382": "Liverpool",
  "n08877613": "Birmingham, Brummagem",
  "n08877807": "Oxford",
  "n08878016": "Cambridge",
  "n08878202": "Bath",
  "n08878367": "Blackpool",
  "n08878533": "Brighton",
  "n08878708": "Bristol",
  "n08878885": "Cheddar",
  "n08879028": "Leeds",
  "n08879197": "Leicester",
  "n08879388": "Newcastle, Newcastle-upon-Tyne",
  "n08879680": "Portsmouth, Pompey",
  "n08879867": "Coventry",
  "n08880083": "Gloucester",
  "n08880223": "Reading",
  "n08880375": "Sheffield",
  "n08880529": "Stratford-on-Avon, Stratford-upon-Avon",
  "n08880713": "Sunderland",
  "n08880859": "Winchester",
  "n08881017": "Worcester",
  "n08881153": "Avon",
  "n08881256": "Berkshire",
  "n08881398": "Cornwall",
  "n08881549": "Cumbria",
  "n08881674": "Cumbria",
  "n08881944": "Devon, Devonshire",
  "n08882061": "Essex",
  "n08882224": "Gloucestershire",
  "n08882365": "Hampshire",
  "n08882530": "New Forest",
  "n08882807": "Hertfordshire",
  "n08882934": "Kent",
  "n08883179": "Somerset",
  "n08883309": "East Sussex",
  "n08883476": "Hastings",
  "n08883643": "West Sussex",
  "n08883772": "Canterbury",
  "n08884012": "Leicestershire, Leicester",
  "n08884191": "Lincolnshire",
  "n08884328": "Northumberland",
  "n08884513": "Flodden",
  "n08884673": "East Anglia",
  "n08884806": "Lancashire",
  "n08884961": "Surrey",
  "n08885099": "Marston Moor",
  "n08885211": "Yorkshire",
  "n08885380": "North Yorkshire",
  "n08885490": "West Yorkshire",
  "n08885631": "South Yorkshire",
  "n08885773": "Northamptonshire",
  "n08885921": "Northampton",
  "n08886037": "Naseby",
  "n08886147": "Northumbria",
  "n08886277": "West Country",
  "n08886432": "Sussex",
  "n08886636": "Wessex",
  "n08886814": "Hadrian's Wall",
  "n08887013": "Channel Island",
  "n08887238": "Jersey, island of Jersey",
  "n08887344": "Guernsey, island of Guernsey",
  "n08887463": "Scilly Islands, Isles of Scilly",
  "n08887716": "Man, Isle of Man",
  "n08887841": "Northern Ireland",
  "n08888181": "Ulster",
  "n08888367": "Bangor",
  "n08888479": "Belfast, capital of Northern Ireland",
  "n08888676": "Ireland, Republic of Ireland, Irish Republic, Eire",
  "n08889191": "Dublin, Irish capital, capital of Ireland",
  "n08889400": "Cork",
  "n08889521": "Galway",
  "n08889657": "Limerick",
  "n08889784": "Tara",
  "n08889944": "Waterford",
  "n08890097": "Scotland",
  "n08891415": "Caledonia",
  "n08891595": "Highlands, Highlands of Scotland",
  "n08891889": "Lowlands, Lowlands of Scotland",
  "n08892058": "Galloway",
  "n08892186": "Aberdeen",
  "n08892327": "Ayr",
  "n08892428": "Balmoral Castle",
  "n08892596": "Edinburgh",
  "n08892766": "Lothian Region",
  "n08892971": "Glasgow",
  "n08893223": "Hebrides, Hebridean Islands, Hebridean Isles, Western Islands, Western Isles",
  "n08893492": "Inner Hebrides",
  "n08893717": "Isle of Skye",
  "n08893868": "Islay",
  "n08894011": "Mull",
  "n08894133": "Staffa",
  "n08894319": "Outer Hebrides",
  "n08894456": "Wales, Cymru, Cambria",
  "n08895148": "Aberdare",
  "n08895254": "Bangor",
  "n08895386": "Cardiff",
  "n08895497": "Newport",
  "n08895623": "Sealyham",
  "n08895771": "Swansea",
  "n08895928": "Anglesey, Anglesey Island, Anglesea, Anglesea Island, Mona",
  "n08896092": "Brunei, Negara Brunei Darussalam",
  "n08896327": "sultanate",
  "n08896440": "Burkina Faso, Upper Volta",
  "n08896645": "Sinai, Sinai Peninsula",
  "n08896831": "Egyptian Empire, Egypt",
  "n08897065": "Egypt, Arab Republic of Egypt, United Arab Republic",
  "n08897843": "Lower Egypt",
  "n08898002": "Upper Egypt",
  "n08898187": "Alexandria, El Iskandriyah",
  "n08898457": "Aswan, Assuan, Assouan",
  "n08898633": "Cairo, Al Qahira, El Qahira, Egyptian capital, capital of Egypt",
  "n08898941": "El Alamein",
  "n08899149": "Giza, El Giza, Gizeh",
  "n08899351": "Memphis",
  "n08899478": "Nag Hammadi",
  "n08899577": "Luxor, El-Aksur",
  "n08899776": "Thebes",
  "n08900047": "Saqqara, Saqqarah, Sakkara",
  "n08900204": "Suez",
  "n08900377": "Suez Canal",
  "n08900535": "India, Republic of India, Bharat",
  "n08902097": "Assam",
  "n08902196": "Karnataka, Mysore",
  "n08902321": "Manipur",
  "n08902422": "Hindustan",
  "n08902569": "Sikkim",
  "n08902753": "Kanara, Canara",
  "n08902894": "Punjab",
  "n08903049": "New Delhi, Indian capital, capital of India",
  "n08903220": "Delhi, Old Delhi",
  "n08903352": "Bangalore",
  "n08903487": "Jabalpur, Jubbulpore",
  "n08903636": "Kolkata, Calcutta",
  "n08903872": "Mumbai, Bombay",
  "n08904115": "Agra",
  "n08904269": "Hyderabad",
  "n08904392": "Chennai, Madras",
  "n08904533": "Lucknow",
  "n08904731": "Mysore",
  "n08904858": "Salem",
  "n08904954": "Andhra Pradesh",
  "n08905085": "Bihar",
  "n08905186": "Goa",
  "n08905313": "Gujarat, Gujerat",
  "n08905467": "Tamil Nadu, Madras",
  "n08905646": "Uttar Pradesh",
  "n08905751": "Gujarat, Gujerat",
  "n08905936": "Maharashtra",
  "n08906053": "Orissa",
  "n08906169": "Nilgiri Hills",
  "n08906272": "West Bengal",
  "n08906374": "Nepal, Kingdom of Nepal",
  "n08906809": "Kathmandu, Katmandu, capital of Nepal",
  "n08906952": "Tibet, Thibet, Xizang, Sitsang",
  "n08907377": "Lhasa, Lassa, capital of Tibet, Forbidden City",
  "n08907606": "Indonesia, Republic of Indonesia, Dutch East Indies",
  "n08908248": "Java",
  "n08908509": "Bali",
  "n08908739": "Timor",
  "n08908954": "Sumatra",
  "n08909107": "Celebes, Sulawesi",
  "n08909233": "Moluccas, Spice Islands",
  "n08909537": "Indonesian Borneo, Kalimantan",
  "n08909719": "Jakarta, Djakarta, capital of Indonesia",
  "n08909933": "Bandung",
  "n08910106": "Medan",
  "n08910230": "Semarang, Samarang",
  "n08910394": "Gulf States",
  "n08910668": "Iran, Islamic Republic of Iran, Persia",
  "n08911421": "Teheran, Tehran, capital of Iran, Iranian capital",
  "n08911602": "Abadan",
  "n08911726": "Bam",
  "n08911868": "Mashhad, Meshed",
  "n08912012": "Isfahan, Esfahan, Aspadana",
  "n08912153": "Rasht, Resht",
  "n08912279": "Shiraz",
  "n08912427": "Tabriz",
  "n08912559": "Urmia, Orumiyeh",
  "n08912703": "Qum",
  "n08912842": "Persia, Persian Empire",
  "n08913085": "Persepolis",
  "n08913242": "Elam, Susiana",
  "n08913434": "Iraq, Republic of Iraq, Al-Iraq, Irak",
  "n08914193": "Baghdad, Bagdad, capital of Iraq",
  "n08914413": "Basra, Basia",
  "n08914573": "Kerbala, Karbala, Kerbela",
  "n08914850": "Kirkuk",
  "n08915017": "Mosul",
  "n08915159": "Levant",
  "n08915372": "Macedon, Macedonia, Makedonija",
  "n08915660": "Philippi",
  "n08915784": "Thrace",
  "n08916111": "Edirne, Adrianople, Adrianopolis",
  "n08916316": "Mesopotamia",
  "n08916832": "Babylon",
  "n08917503": "Babylonia, Chaldaea, Chaldea",
  "n08917881": "Chaldea, Chaldaea",
  "n08918248": "Sumer",
  "n08918800": "Ur",
  "n08918944": "Assyria",
  "n08919241": "Assur, Asur, Ashur",
  "n08919475": "Nineveh",
  "n08919693": "Phoenicia, Phenicia",
  "n08919949": "Carthage",
  "n08920207": "Utica",
  "n08920381": "Japan, Japanese Islands, Japanese Archipelago",
  "n08920722": "Hokkaido, Ezo, Yezo",
  "n08920924": "Honshu, Hondo",
  "n08921392": "Kyushu",
  "n08921613": "Shikoku",
  "n08921850": "Japan, Nippon, Nihon",
  "n08923177": "Asahikawa",
  "n08923348": "Tokyo, Tokio, Yeddo, Yedo, Edo, Japanese capital, capital of Japan",
  "n08923586": "Nagano",
  "n08923755": "Nagoya",
  "n08923884": "Omiya",
  "n08924023": "Osaka",
  "n08924238": "Yokohama",
  "n08924400": "Okinawa",
  "n08924560": "Naha City",
  "n08924691": "Ryukyu Islands",
  "n08924913": "Kobe",
  "n08925093": "Kyoto",
  "n08925287": "Hiroshima",
  "n08925552": "Sapporo",
  "n08925700": "Kitakyushu",
  "n08925830": "Fukuoka",
  "n08925957": "Nagasaki",
  "n08926231": "Toyohashi",
  "n08926381": "Toyonaki",
  "n08926543": "Toyota",
  "n08926681": "Asama, Mount Asama",
  "n08926877": "Volcano Islands",
  "n08927068": "Iwo Jima",
  "n08927186": "Jordan, Hashemite Kingdom of Jordan",
  "n08927548": "Amman, capital of Jordan",
  "n08927678": "Al Aqabah, Aqaba, Akaba",
  "n08927836": "Jericho",
  "n08928083": "Az Zarqa, Zarqa",
  "n08928193": "Kenya, Republic of Kenya",
  "n08928582": "Nairobi, capital of Kenya",
  "n08928742": "Kisumu",
  "n08928933": "Mombasa",
  "n08929102": "Nakuru",
  "n08929243": "Kuwait, State of Kuwait, Koweit",
  "n08929555": "Kuwait, Kuwait City, Koweit, capital of Kuwait",
  "n08929722": "Gaul, Gallia",
  "n08929922": "France, French Republic",
  "n08932568": "Paris, City of Light, French capital, capital of France",
  "n08933084": "Left Bank, Latin Quarter",
  "n08933287": "Montmartre",
  "n08933437": "Clichy, Clichy-la-Garenne",
  "n08933621": "Orly",
  "n08933770": "Quai d'Orsay",
  "n08933940": "Right Bank",
  "n08934067": "Ile-St-Louis",
  "n08934174": "Champs Elysees",
  "n08934313": "Avignon",
  "n08934532": "Bordeaux",
  "n08934694": "Brest",
  "n08934868": "Calais",
  "n08935212": "Cannes",
  "n08935381": "Chablis",
  "n08935516": "Chartres",
  "n08935674": "Cherbourg",
  "n08935848": "Dijon",
  "n08935978": "Dunkirk, Dunkerque",
  "n08936180": "Grenoble",
  "n08936303": "Le Havre",
  "n08936476": "Lille",
  "n08936647": "Lyon, Lyons",
  "n08936833": "Marseille, Marseilles",
  "n08936996": "Nancy",
  "n08937109": "Nantes",
  "n08937251": "Nice",
  "n08937414": "Orleans",
  "n08937594": "Rheims, Reims",
  "n08937850": "Strasbourg, Strassburg",
  "n08937995": "Toulon",
  "n08938163": "Toulouse",
  "n08938351": "Tours",
  "n08938478": "Valenciennes",
  "n08938619": "Versailles",
  "n08938819": "Vichy",
  "n08939028": "Vienne",
  "n08939201": "Riviera",
  "n08939437": "French Riviera, Cote d'Azur",
  "n08939562": "French region",
  "n08940209": "Alsace, Alsatia, Elsass",
  "n08940371": "Anjou",
  "n08940517": "Aquitaine, Aquitania",
  "n08940670": "Artois",
  "n08940835": "Auvergne",
  "n08940936": "Basse-Normandie, Lower-Normandy",
  "n08941057": "Bourgogne, Burgundy",
  "n08941208": "Bretagne, Brittany, Breiz",
  "n08941426": "Centre",
  "n08941535": "Champagne, Champagne-Ardenne",
  "n08941681": "Ardennes",
  "n08941895": "Corse, Corsica",
  "n08942091": "Corse, Corsica",
  "n08942277": "Franche-Comte",
  "n08942392": "Gascogne, Gascony",
  "n08942508": "Haute-Normandie, Upper-Normandy",
  "n08942629": "Ile-de-France",
  "n08942780": "Languedoc-Roussillon",
  "n08942965": "Limousin",
  "n08943104": "Lorraine, Lothringen",
  "n08943242": "Martinique",
  "n08943461": "Mayenne",
  "n08943601": "Midi",
  "n08943699": "Midi-Pyrenees",
  "n08943810": "Nord-Pas-de-Calais",
  "n08943926": "Pays de la Loire",
  "n08944089": "Picardie, Picardy",
  "n08944224": "Poitou-Charentes, Poitou",
  "n08944378": "Rhone-Alpes",
  "n08944561": "Normandie, Normandy",
  "n08944818": "Orleanais",
  "n08944960": "Provence",
  "n08945110": "Lyonnais",
  "n08945277": "Savoy",
  "n08945529": "Gabon, Gabonese Republic, Gabun",
  "n08945705": "Libreville, capital of Gabon",
  "n08945821": "Gambia, The Gambia, Republic of The Gambia",
  "n08946042": "Banjul, capital of Gambia",
  "n08946187": "Ghana, Republic of Ghana, Gold Coast",
  "n08946564": "Accra, capital of Ghana",
  "n08946715": "Kumasi",
  "n08946812": "Tamale",
  "n08946909": "Grenada",
  "n08947180": "St. George's, capital of Grenada",
  "n08947319": "Guinea, Republic of Guinea, French Guinea",
  "n08947617": "Conakry, Konakri, capital of Guinea",
  "n08947772": "Guinea-Bissau, Republic of Guinea-Bissau, Guine-Bissau, Portuguese Guinea",
  "n08948027": "Bissau, capital of Guinea-Bissau",
  "n08948155": "Guiana",
  "n08948346": "Guyana, Co-operative Republic of Guyana, British Guiana",
  "n08948704": "Georgetown, Stabroek",
  "n08948958": "Demerara",
  "n08949093": "Netherlands, The Netherlands, Kingdom of The Netherlands, Nederland, Holland",
  "n08949737": "Amsterdam, Dutch capital, capital of The Netherlands",
  "n08950035": "Apeldoorn",
  "n08950230": "Arnhem",
  "n08950407": "The Hague, 's Gravenhage, Den Haag",
  "n08950649": "Eindhoven",
  "n08950787": "Nijmegen",
  "n08950907": "Rotterdam",
  "n08951077": "Leiden, Leyden",
  "n08951278": "Utrecht",
  "n08951385": "Friesland",
  "n08951513": "Friesland",
  "n08951777": "Frisia",
  "n08951957": "Frisian Islands",
  "n08952190": "Hungary, Republic of Hungary, Magyarorszag",
  "n08952423": "Budapest, Hungarian capital, capital of Hungary",
  "n08952628": "Faroe Islands, Faeroe Islands, Faroes, Faeroes",
  "n08952856": "Faroe Islands, Faeroe Islands, Faroes, Faeroes",
  "n08953029": "Thorshavn",
  "n08953151": "Iceland",
  "n08953324": "Iceland, Republic of Iceland",
  "n08953596": "Reykjavik, capital of Iceland",
  "n08953829": "Orkney Islands",
  "n08954057": "Shetland, Shetland Islands, Zetland",
  "n08954269": "Thule, ultima Thule",
  "n08954443": "Thule",
  "n08954611": "Korea, Korean Peninsula, Dae-Han-Min-Gook, Han-Gook",
  "n08954975": "Chosen",
  "n08955082": "North Korea, Democratic People's Republic of Korea, D.P.R.K., DPRK",
  "n08955397": "Pyongyang, capital of North Korea",
  "n08955626": "South Korea, Republic of Korea",
  "n08955951": "Seoul, capital of South Korea",
  "n08956140": "Inchon, Incheon, Chemulpo",
  "n08956305": "Kwangju",
  "n08956461": "Taegu, Tegu",
  "n08956574": "Pusan",
  "n08956760": "Laos, Lao People's Democratic Republic",
  "n08957064": "Vientiane, Laotian capital, capital of Laos",
  "n08957212": "Lappland, Lapland",
  "n08957381": "Lebanon, Lebanese Republic",
  "n08957806": "Bayrut, Beirut, capital of Lebanon",
  "n08957993": "Tarabulus, Tripoli, Tarabulus Ash-Sham, Trablous",
  "n08958212": "Sayda, Saida, Sidon",
  "n08958334": "Sur, Tyre",
  "n08958535": "Byblos",
  "n08958830": "Lesotho, Kingdom of Lesotho, Basutoland",
  "n08959105": "Maseru, capital of Lesotho",
  "n08959254": "Liberia, Republic of Liberia",
  "n08959495": "Monrovia, Liberian capital, capital of Liberia",
  "n08959683": "Libya, Socialist People's Libyan Arab Jamahiriya",
  "n08960099": "Tripoli, Tarabulus Al-Gharb, capital of Libya",
  "n08960363": "Benghazi",
  "n08960548": "Liechtenstein, Principality of Liechtenstein",
  "n08960843": "Vaduz, capital of Liechtenstein",
  "n08960987": "Luxembourg, Grand Duchy of Luxembourg, Luxemburg",
  "n08961402": "Luxembourg-Ville, Luxembourg, Luxemburg, Luxembourg City, capital of Luxembourg",
  "n08961630": "Macedonia",
  "n08961852": "Skopje, Skoplje, Uskub",
  "n08961970": "Madagascar",
  "n08962187": "Madagascar, Republic of Madagascar, Malagasy Republic",
  "n08962465": "Antananarivo, capital of Madagascar",
  "n08962610": "Malawi, Republic of Malawi, Nyasaland",
  "n08962951": "Blantyre",
  "n08963095": "Lilongwe, capital of Malawi",
  "n08963244": "Zomba",
  "n08963369": "Malaysia, Malaya",
  "n08963813": "Kuala Lumpur",
  "n08963969": "Putrajaya, capital of Malaysia",
  "n08964099": "East Malaysia",
  "n08964288": "Sabah, North Borneo",
  "n08964474": "Sarawak",
  "n08964647": "West Malaysia",
  "n08964810": "Malay Peninsula",
  "n08965055": "Maldives, Maldive Islands",
  "n08965251": "Maldives, Republic of Maldives",
  "n08965475": "Male",
  "n08965598": "Mali, Republic of Mali, French Sudan",
  "n08965958": "Bamako",
  "n08966085": "Timbuktu",
  "n08966239": "Malta",
  "n08966408": "Malta, Republic of Malta",
  "n08966647": "Valletta, Valetta, capital of Malta",
  "n08966820": "Mauritania, Islamic Republic of Mauritania, Mauritanie, Muritaniya",
  "n08967176": "Nouakchott",
  "n08967329": "Mauritius",
  "n08967484": "Mauritius, Republic of Mauritius",
  "n08967688": "Port Louis",
  "n08967868": "Monaco, Principality of Monaco",
  "n08968125": "Monaco-Ville",
  "n08968225": "Monte Carlo",
  "n08968390": "Tartary, Tatary",
  "n08968677": "Mongolia",
  "n08968879": "Mongolia, Mongolian People's Republic, Outer Mongolia",
  "n08969123": "Ulan Bator, Ulaanbaatar, Urga, Kulun, capital of Mongolia",
  "n08969291": "Morocco, Kingdom of Morocco, Maroc, Marruecos, Al-Magrib",
  "n08969798": "Casablanca",
  "n08969948": "El Aaium",
  "n08970064": "Fez, Fes",
  "n08970189": "Marrakesh, Marrakech",
  "n08970318": "Oujda",
  "n08970445": "Rabat, capital of Morocco",
  "n08970611": "Tangier, Tangiers",
  "n08970833": "Western Sahara, Spanish Sahara",
  "n08971025": "Mozambique, Republic of Mozambique, Mocambique",
  "n08971404": "Beira",
  "n08971554": "Maputo, capital of Mozambique",
  "n08971693": "Natal, KwaZulu-Natal",
  "n08971914": "New Zealand, New Zealand Islands",
  "n08972148": "North Island",
  "n08972335": "South Island",
  "n08972521": "New Zealand",
  "n08972920": "Auckland",
  "n08973064": "Christchurch",
  "n08973202": "Wellington, capital of New Zealand",
  "n08973330": "Niger, Republic of Niger",
  "n08973647": "Niamey, capital of Niger",
  "n08973776": "Nigeria, Federal Republic of Nigeria",
  "n08974171": "Abuja, capital of Nigeria, Nigerian capital",
  "n08974330": "Ibadan",
  "n08974468": "Katsina",
  "n08974604": "Lagos",
  "n08974818": "Maiduguri, Yerwa-Maiduguri",
  "n08974974": "Zaria",
  "n08975106": "Oman, Sultanate of Oman, Muscat and Oman",
  "n08975435": "Muscat, Masqat, capital of Oman",
  "n08975617": "Kashmir, Cashmere, Jammu and Kashmir",
  "n08975902": "Pakistan, Islamic Republic of Pakistan, West Pakistan",
  "n08976799": "Faisalabad, Lyallpur",
  "n08976913": "Hyderabad",
  "n08977035": "Islamabad, capital of Pakistan",
  "n08977211": "Karachi",
  "n08977428": "Lahore",
  "n08977527": "Peshawar",
  "n08977665": "Rawalpindi",
  "n08977845": "Sind",
  "n08977948": "Palau, Palau Islands, Belau, Pelew",
  "n08978161": "Palau, Republic of Palau, TT",
  "n08978343": "Paraguay, Republic of Paraguay",
  "n08978667": "Asuncion, capital of Paraguay",
  "n08978821": "Parthia",
  "n08979054": "Peru, Republic of Peru",
  "n08979590": "Arequipa",
  "n08979740": "Cuzco, Cusco",
  "n08979878": "Lima, capital of Peru",
  "n08980119": "Machu Picchu",
  "n08980300": "Philippines, Philippine Islands",
  "n08980640": "Cebu",
  "n08980814": "Luzon",
  "n08980920": "Mindanao",
  "n08981123": "Mindoro",
  "n08981244": "Philippines, Republic of the Philippines",
  "n08981746": "Manila, capital of the Philippines",
  "n08981922": "Caloocan",
  "n08982037": "Cebu, Cebu City",
  "n08982181": "Quezon City",
  "n08982289": "Pinatubo, Mount Pinatubo",
  "n08982452": "Visayan Islands, Bisayas",
  "n08982587": "Poland, Republic of Poland, Polska",
  "n08983105": "Warszawa, Warsaw, capital of Poland",
  "n08983274": "Bydgoszcz, Bromberg",
  "n08983413": "Cracow, Krakow, Krakau",
  "n08983556": "Czestochowa",
  "n08983742": "Gdansk, Danzig",
  "n08984010": "Katowice",
  "n08984122": "Lodz",
  "n08984223": "Lublin",
  "n08984332": "Wroclaw, Breslau",
  "n08984457": "Zabrze",
  "n08984567": "Iberian Peninsula, Iberia",
  "n08984788": "Portugal, Portuguese Republic",
  "n08985345": "Azores, Acores",
  "n08985559": "Madeira",
  "n08985724": "Madeira Islands, Madeiras",
  "n08985958": "Braga",
  "n08986066": "Lisbon, Lisboa, capital of Portugal",
  "n08986374": "Porto, Oporto",
  "n08986526": "Setubal",
  "n08986691": "Qatar, Qatar Peninsula, Katar, Katar Peninsula",
  "n08986905": "Qatar, State of Qatar, Katar, State of Katar",
  "n08987262": "Doha, Bida, El Beda, capital of Qatar",
  "n08987423": "Saint Kitts and Nevis, Federation of Saint Kitts and Nevis, Saint Christopher-Nevis, St. Christopher-Nevis, St. Kitts and Nevis",
  "n08987879": "Saint Christopher, St. Christopher, Saint Kitts, St. Kitts",
  "n08988068": "Basseterre",
  "n08988216": "Nevis",
  "n08988333": "Sombrero",
  "n08988453": "Saint Lucia, St. Lucia",
  "n08988609": "Saint Lucia, St. Lucia",
  "n08988861": "Castries",
  "n08989031": "Saint Vincent and the Grenadines, St. Vincent and the Grenadines",
  "n08989324": "Saint Vincent, St. Vincent",
  "n08989556": "Kingstown",
  "n08989697": "French Polynesia, French Oceania",
  "n08989941": "Tahiti",
  "n08990203": "Papeete",
  "n08990344": "Society Islands",
  "n08990522": "Tuamotu Archipelago, Paumotu Archipelago, Low Archipelago",
  "n08990701": "Tubuai Islands, Austral Islands",
  "n08990844": "Gambier Islands",
  "n08990999": "Marquesas Islands, Iles Marquises",
  "n08991182": "Samoa, Samoan Islands",
  "n08991491": "Samoa, Independent State of Samoa, Western Samoa, Samoa i Sisifo",
  "n08991752": "Apia, capital of Western Samoa",
  "n08991878": "American Samoa, Eastern Samoa, AS",
  "n08992067": "Pago Pago, Pango Pango",
  "n08992181": "San Marino, Republic of San Marino",
  "n08992508": "San Marino, capital of San Marino",
  "n08992648": "Sao Tome and Principe, Democratic Republic of Sao Tome and Principe, Sao Tome e Principe, Sao Thome e Principe, St. Thomas and Principe",
  "n08993037": "Sao Tome",
  "n08993144": "Principe",
  "n08993288": "Saudi Arabia, Kingdom of Saudi Arabia",
  "n08993871": "Riyadh, capital of Saudi Arabia",
  "n08994090": "Mecca",
  "n08994339": "Medina, Al Madinah",
  "n08994540": "Dhahran",
  "n08994834": "Jeddah, Jed'dah, Jiddah, Jidda",
  "n08995013": "Tabuk",
  "n08995120": "Taif",
  "n08995242": "Nejd, Najd",
  "n08995515": "Hejaz, Hedjaz, Hijaz",
  "n08995862": "Senegal, Republic of Senegal",
  "n08996118": "Dakar, capital of Senegal",
  "n08996284": "Seychelles, Seychelles islands",
  "n08996483": "Seychelles, Republic of Seychelles",
  "n08996714": "Victoria, capital of Seychelles",
  "n08996871": "Sierra Leone, Republic of Sierra Leone",
  "n08997132": "Freetown, capital of Sierra Leone",
  "n08997310": "Singapore, Singapore Island",
  "n08997487": "Singapore, Republic of Singapore",
  "n08997801": "Singapore, capital of Singapore",
  "n08997976": "Solomons, Solomon Islands",
  "n08998233": "Solomon Islands",
  "n08998451": "Honiara",
  "n08998560": "Somalia",
  "n08998853": "Mogadishu, Mogadiscio, capital of Somalia",
  "n08999049": "Hargeisa",
  "n08999154": "Somali peninsula, Horn of Africa",
  "n08999482": "South Africa, Republic of South Africa",
  "n09000272": "Pretoria, capital of South Africa",
  "n09000462": "Cape Town",
  "n09000665": "Johannesburg",
  "n09000854": "Kimberley",
  "n09001007": "Durban",
  "n09001184": "Free State, Orange Free State",
  "n09001373": "Transvaal",
  "n09001580": "Cape Province, Cape of Good Hope Province, Cape Colony",
  "n09001881": "Witwatersrand, Rand, Reef",
  "n09002096": "Cape of Good Hope",
  "n09002195": "Cape of Good Hope",
  "n09002325": "Bloemfontein",
  "n09002433": "Soweto",
  "n09002630": "Rus",
  "n09002814": "Russia",
  "n09003284": "Soviet Union, Russia, Union of Soviet Socialist Republics, USSR",
  "n09003918": "Muscovy",
  "n09004068": "Moscow, capital of the Russian Federation, Russian capital",
  "n09004358": "Astrakhan",
  "n09004495": "Cherepovets",
  "n09004625": "Chechnya, Chechenia, Chechen Republic",
  "n09004992": "Grozny, Groznyy",
  "n09005153": "Kaluga",
  "n09005273": "Khabarovsk",
  "n09005457": "Khabarovsk",
  "n09005611": "Kursk",
  "n09005712": "Siberia",
  "n09006205": "Soviet Socialist Republic",
  "n09006413": "Russia, Russian Federation",
  "n09007471": "European Russia",
  "n09007613": "Asian Russia",
  "n09007723": "Soviet Russia, Russia, Russian Soviet Federated Socialist Republic",
  "n09008130": "Nizhnyi Novgorod, Nizhni Novgorod, Gorki, Gorky, Gorkiy",
  "n09008333": "Kazan",
  "n09008454": "St. Petersburg, Leningrad, Peterburg, Petrograd, Saint Petersburg",
  "n09008723": "Murmansk",
  "n09008993": "Nalchik",
  "n09009174": "Novgorod",
  "n09009372": "Perm, Molotov",
  "n09009490": "Rostov, Rostov on Don, Rostov na Donu",
  "n09009693": "Saratov",
  "n09009816": "Smolensk",
  "n09009978": "Ufa",
  "n09010085": "Volgograd, Stalingrad, Tsaritsyn",
  "n09010300": "Novosibirsk",
  "n09010453": "Chelyabinsk",
  "n09010565": "Omsk",
  "n09010670": "Vladivostok",
  "n09010785": "Novaya Zemlya, Nova Zembla",
  "n09010974": "Kola Peninsula",
  "n09011151": "Belarus, Republic of Belarus, Byelarus, Byelorussia, Belorussia, White Russia",
  "n09011518": "Minsk, capital of Belarus",
  "n09011679": "Homyel, Homel, Gomel",
  "n09011820": "Pinsk",
  "n09011922": "Lubavitch",
  "n09012101": "Baltic State, Baltic Republic",
  "n09012297": "Estonia, Esthonia, Republic of Estonia",
  "n09012530": "Tallinn, Tallin, capital of Estonia",
  "n09012735": "Tartu",
  "n09012898": "Livonia",
  "n09013074": "Latvia, Republic of Latvia",
  "n09013353": "Riga, capital of Latvia",
  "n09013603": "Liepaja",
  "n09013724": "Daugavpils",
  "n09013830": "Lithuania, Republic of Lithuania, Lietuva",
  "n09014066": "Klaipeda, Memel",
  "n09014273": "Vilnius, Vilna, Vilno, Wilno, capital of Lithuania",
  "n09014470": "Kaunas, Kovna, Kovno",
  "n09014586": "Moldova, Republic of Moldova, Moldavia",
  "n09014850": "Kishinev, Chisinau, capital of Moldova",
  "n09014979": "Ukraine, Ukrayina",
  "n09015460": "Crimea",
  "n09015653": "Colchis",
  "n09015907": "Kyyiv, Kiev, capital of the Ukraine",
  "n09016099": "Donetsk, Donetske, Stalino",
  "n09016232": "Donets Basin, Donbass, Donbas",
  "n09016365": "Chernobyl",
  "n09016539": "Dneprodzerzhinsk",
  "n09016698": "Dnipropetrovsk, Yekaterinoslav",
  "n09016860": "Kharkov, Kharkiv",
  "n09017005": "Odessa, Odesa",
  "n09017168": "Sebastopol, Sevastopol",
  "n09017301": "Yalta",
  "n09017526": "Armenia, Republic of Armenia, Hayastan",
  "n09018030": "Yerevan, Jerevan, Erivan, capital of Armenia",
  "n09018162": "Azerbaijan, Azerbaijani Republic, Azerbajdzhan, Azerbajdzhan Republic",
  "n09018426": "Baku, capital of Azerbaijan",
  "n09018647": "Iberia",
  "n09018848": "Georgia, Sakartvelo",
  "n09019194": "Tbilisi, Tiflis, capital of Georgia",
  "n09019355": "Abkhaz, Abkhazia",
  "n09019592": "Adzhar, Adzharia",
  "n09019726": "Kazakhstan, Republic of Kazakhstan, Kazakstan, Kazakh, Kazak",
  "n09020119": "Astana, Akmola, capital of Kazakhstan",
  "n09020299": "Almaty, Alma-Ata",
  "n09020440": "Kyrgyzstan, Kyrgyz Republic, Kirghizia, Kirgizia, Kirghiz, Kirgiz, Kirghizstan, Kirgizstan",
  "n09020792": "Bishkek, Biskek, Frunze, capital of Kyrgyzstan",
  "n09020961": "Tajikistan, Republic of Tajikistan, Tadzhikistan, Tadzhik, Tadjik, Tajik",
  "n09021313": "Dushanbe, Dusanbe, Dyushambe, Stalinabad, capital of Tajikistan",
  "n09021503": "Turkmenistan, Turkomen, Turkmen, Turkmenia",
  "n09021812": "Ashkhabad, capital of Turkmenistan",
  "n09021958": "Kamchatka Peninsula",
  "n09022132": "Taimyr Peninsula, Taymyr Peninsula",
  "n09022265": "Uzbekistan, Republic of Uzbekistan, Uzbek",
  "n09022538": "Tashkent, Taskent, capital of Uzbek",
  "n09022667": "Samarkand, Samarcand",
  "n09022831": "Latin America",
  "n09023118": "Andorra, Principality of Andorra",
  "n09023321": "Spain, Kingdom of Spain, Espana",
  "n09024467": "Madrid, capital of Spain, Spanish capital",
  "n09024668": "Balearic Islands",
  "n09024844": "Majorca",
  "n09024972": "Canary Islands, Canaries",
  "n09025189": "Barcelona",
  "n09025451": "Cadiz",
  "n09025584": "Cartagena",
  "n09025728": "Cordoba, Cordova",
  "n09025863": "Granada",
  "n09026204": "Jerez, Jerez de la Frontera",
  "n09026360": "Leon",
  "n09026499": "Logrono",
  "n09026614": "Malaga",
  "n09026780": "Oviedo",
  "n09026911": "San Sebastian",
  "n09027089": "Sevilla, Seville",
  "n09027292": "Toledo",
  "n09027460": "Aragon",
  "n09027679": "Zaragoza, Saragossa",
  "n09027853": "Castile, Castilla",
  "n09028062": "Catalonia",
  "n09028204": "Galicia",
  "n09028367": "Leon",
  "n09028477": "Valencia",
  "n09028643": "Tenerife",
  "n09028841": "Gibraltar, Rock of Gibraltar, Calpe",
  "n09029242": "Sudan, Soudan",
  "n09029457": "Sudan, Republic of the Sudan, Soudan",
  "n09029884": "Darfur",
  "n09030096": "Kordofan",
  "n09030210": "Khartoum, capital of Sudan",
  "n09030382": "Nyala",
  "n09030467": "Port Sudan",
  "n09030596": "Omdurman",
  "n09030752": "Suriname, Republic of Suriname, Surinam, Dutch Guiana, Netherlands Guiana",
  "n09031061": "Paramaribo, capital of Suriname",
  "n09031233": "Swaziland, Kingdom of Swaziland",
  "n09031501": "Mbabane, capital of Swaziland",
  "n09031653": "Switzerland, Swiss Confederation, Suisse, Schweiz, Svizzera",
  "n09032191": "Swiss canton",
  "n09032321": "Bern, Berne, capital of Switzerland",
  "n09032483": "Basel, Basle, Bale",
  "n09032604": "Geneva, Geneve, Genf",
  "n09032843": "Interlaken",
  "n09032981": "Lausanne",
  "n09033117": "Zurich",
  "n09033333": "Syria, Syrian Arab Republic",
  "n09033813": "Aram",
  "n09033936": "Dimash, Damascus, capital of Syria",
  "n09034286": "Halab, Aleppo, Alep",
  "n09034402": "Al Ladhiqiyah, Latakia",
  "n09034550": "Tanzania, United Republic of Tanzania",
  "n09034967": "Dar es Salaam, capital of Tanzania",
  "n09035153": "Dodoma",
  "n09035305": "Tanganyika",
  "n09035458": "Zanzibar",
  "n09035632": "Mbeya",
  "n09035735": "Mwanza",
  "n09035852": "Tabora",
  "n09035951": "Tanga",
  "n09036098": "Serengeti, Serengeti Plain",
  "n09036295": "Serengeti National Park",
  "n09036452": "Thailand, Kingdom of Thailand, Siam",
  "n09036880": "Bangkok, capital of Thailand, Krung Thep",
  "n09037133": "Tonga, Kingdom of Tonga, Friendly Islands",
  "n09037394": "Tunisia, Republic of Tunisia",
  "n09037838": "Tunis, capital of Tunisia",
  "n09037991": "Ariana",
  "n09038079": "Ehadhamen",
  "n09038170": "Gafsa",
  "n09038272": "Sfax, Safaqis",
  "n09038439": "Sousse, Susa, Susah",
  "n09038597": "Ottoman Empire, Turkish Empire",
  "n09038990": "Kurdistan",
  "n09039260": "Iraqi Kurdistan",
  "n09039411": "Turkey, Republic of Turkey",
  "n09040299": "Abydos",
  "n09040475": "Adana, Seyhan",
  "n09040601": "Ankara, Turkish capital, capital of Turkey, Angora",
  "n09040839": "Antalya, Adalia",
  "n09040998": "Antioch, Antakya, Antakiya",
  "n09041199": "Chalcedon, Kadikoy",
  "n09041371": "Dardanelles, Canakkale Bogazi, Hellespont",
  "n09041582": "Halicarnassus",
  "n09041785": "Istanbul, Stambul, Stamboul, Constantinople",
  "n09042213": "Bursa, Brusa",
  "n09042322": "Izmir, Smyrna",
  "n09042451": "Pergamum",
  "n09042675": "Sardis",
  "n09042924": "Ionia",
  "n09043052": "Uganda, Republic of Uganda",
  "n09043411": "Buganda",
  "n09043556": "Entebbe",
  "n09043802": "Jinja",
  "n09043926": "Kampala, capital of Uganda",
  "n09044094": "Gulu",
  "n09044190": "United Arab Emirates",
  "n09044536": "Abu Dhabi, United Arab Emirates's capital",
  "n09044714": "Dubai",
  "n09044862": "United States, United States of America, America, the States, US, U.S., USA, U.S.A.",
  "n09048127": "East Coast",
  "n09048303": "West Coast",
  "n09048460": "Colony",
  "n09048880": "New England",
  "n09049303": "Mid-Atlantic states",
  "n09049599": "Gulf States",
  "n09049909": "slave state",
  "n09050089": "free state",
  "n09050244": "Confederacy, Confederate States, Confederate States of America, South, Dixie, Dixieland",
  "n09050730": "South",
  "n09051235": "Deep South",
  "n09051609": "Old South",
  "n09051726": "Sunbelt",
  "n09051898": "Tidewater, Tidewater region",
  "n09052100": "Piedmont",
  "n09052314": "Union, North",
  "n09052652": "North",
  "n09052835": "Carolina, Carolinas",
  "n09053019": "Dakota",
  "n09053185": "Alabama, Heart of Dixie, Camellia State, AL",
  "n09053801": "Montgomery, capital of Alabama",
  "n09053947": "Birmingham, Pittsburgh of the South",
  "n09054110": "Decatur",
  "n09054233": "Gadsden",
  "n09054350": "Huntsville",
  "n09054480": "Mobile",
  "n09054616": "Selma",
  "n09054797": "Tuscaloosa",
  "n09054915": "Tuskegee",
  "n09055015": "Alaska, Last Frontier, AK",
  "n09055786": "Juneau, capital of Alaska",
  "n09055906": "Anchorage",
  "n09056055": "Nome",
  "n09056278": "Sitka",
  "n09056476": "Skagway",
  "n09056677": "Valdez",
  "n09056837": "Seward Peninsula",
  "n09057021": "Alexander Archipelago",
  "n09057170": "Admiralty Island",
  "n09057311": "Arizona, Grand Canyon State, AZ",
  "n09057930": "Flagstaff",
  "n09058071": "Mesa",
  "n09058219": "Nogales",
  "n09058376": "Phoenix, capital of Arizona",
  "n09058635": "Prescott",
  "n09058735": "Sun City",
  "n09058841": "Tucson",
  "n09059125": "Yuma",
  "n09059274": "Arkansas, Land of Opportunity, AR",
  "n09059741": "Fayetteville",
  "n09059876": "Fort Smith",
  "n09060024": "Hot Springs",
  "n09060176": "Jonesboro",
  "n09060280": "Little Rock, capital of Arkansas",
  "n09060480": "Pine Bluff",
  "n09060615": "Texarkana",
  "n09060768": "California, Golden State, CA, Calif.",
  "n09062015": "Anaheim",
  "n09062184": "Disneyland",
  "n09062320": "Bakersfield",
  "n09062478": "Barstow",
  "n09062585": "Berkeley",
  "n09062791": "Beverly Hills",
  "n09062961": "Chula Vista",
  "n09063125": "Eureka",
  "n09063259": "Fresno",
  "n09063477": "Long Beach",
  "n09063673": "Los Angeles, City of the Angels",
  "n09063950": "Hollywood",
  "n09064080": "Monterey",
  "n09064264": "Oakland",
  "n09064468": "Palo Alto",
  "n09064594": "Pasadena",
  "n09064729": "Redding",
  "n09064861": "Riverside",
  "n09064966": "Sacramento, capital of California",
  "n09065191": "San Bernardino",
  "n09065328": "San Diego",
  "n09065557": "San Francisco",
  "n09065899": "Nob Hill",
  "n09066017": "San Jose",
  "n09066249": "San Mateo",
  "n09066375": "San Pablo",
  "n09066534": "Santa Ana",
  "n09066665": "Santa Barbara",
  "n09066799": "Santa Clara",
  "n09066948": "Santa Catalina, Catalina Island",
  "n09067138": "Santa Cruz",
  "n09067277": "Colorado, Centennial State, CO",
  "n09067721": "Boulder",
  "n09067878": "Colorado Springs",
  "n09068107": "Denver, Mile-High City, capital of Colorado",
  "n09068320": "Pueblo",
  "n09068444": "Connecticut, Nutmeg State, Constitution State, CT",
  "n09068805": "Connecticut",
  "n09068921": "Bridgeport",
  "n09069072": "Farmington",
  "n09069190": "Hartford, capital of Connecticut",
  "n09069415": "New Haven",
  "n09069569": "New London",
  "n09069752": "Waterbury",
  "n09069862": "Delaware, Diamond State, First State, DE",
  "n09070120": "Delaware",
  "n09070233": "Dover, capital of Delaware",
  "n09070363": "Wilmington",
  "n09070487": "District of Columbia, D.C., DC",
  "n09070793": "Washington, Washington D.C., American capital, capital of the United States",
  "n09071246": "Potomac",
  "n09071349": "Capitol Hill, the Hill",
  "n09071571": "Georgetown",
  "n09071690": "Florida, Sunshine State, Everglade State, FL",
  "n09072611": "Daytona Beach",
  "n09072810": "Fort Lauderdale",
  "n09073034": "Fort Myers",
  "n09073138": "Gainesville",
  "n09073258": "Jacksonville",
  "n09073442": "Key West",
  "n09073584": "Melbourne",
  "n09073697": "Miami",
  "n09073938": "Miami Beach",
  "n09074140": "Orlando",
  "n09074285": "Palm Beach",
  "n09074431": "Panama City",
  "n09074596": "Pensacola",
  "n09074707": "Sarasota",
  "n09074834": "St. Augustine, Saint Augustine",
  "n09075007": "St. Petersburg, Saint Petersburg",
  "n09075170": "Tallahassee, capital of Florida",
  "n09075329": "Tampa",
  "n09075496": "West Palm Beach",
  "n09075688": "Walt Disney World",
  "n09075842": "Georgia, Empire State of the South, Peach State, GA",
  "n09076421": "Georgia",
  "n09076533": "Albany",
  "n09076675": "Atlanta, capital of Georgia",
  "n09076982": "Athens",
  "n09077111": "Augusta",
  "n09077266": "Brunswick",
  "n09077410": "Columbus",
  "n09077556": "Macon",
  "n09077681": "Oxford",
  "n09077821": "Savannah",
  "n09077977": "Valdosta",
  "n09078102": "Vidalia",
  "n09078231": "Hawaii, Hawai'i, Aloha State, HI",
  "n09078654": "Hilo",
  "n09078784": "Honolulu, capital of Hawaii, Hawaiian capital",
  "n09079001": "Waikiki",
  "n09079153": "Hawaiian Islands, Sandwich Islands",
  "n09079505": "Hawaii, Hawaii Island",
  "n09079747": "Kahoolawe, Kahoolawe Island",
  "n09079875": "Kauai, Kauai Island",
  "n09080006": "Lanai, Lanai Island",
  "n09080146": "Maui, Maui Island",
  "n09080273": "Molokai, Molokai Island",
  "n09080415": "Nihau, Nihau Island",
  "n09080554": "Oahu, Oahu Island",
  "n09080782": "Pearl Harbor",
  "n09080989": "Midway Islands",
  "n09081213": "Idaho, Gem State, ID",
  "n09081560": "Boise, capital of Idaho",
  "n09081688": "Coeur d'Alene",
  "n09081830": "Idaho Falls",
  "n09081955": "Lewiston",
  "n09082058": "Nampa",
  "n09082158": "Pocatello",
  "n09082273": "Sun Valley",
  "n09082395": "Twin Falls",
  "n09082540": "Illinois, Prairie State, Land of Lincoln, IL",
  "n09082988": "Cairo",
  "n09083151": "Carbondale",
  "n09083255": "Champaign",
  "n09083390": "Chicago, Windy City",
  "n09083659": "Decatur",
  "n09083795": "East Saint Louis",
  "n09083949": "Moline",
  "n09084075": "Peoria",
  "n09084196": "Rockford",
  "n09084298": "Rock Island",
  "n09084483": "Springfield, capital of Illinois",
  "n09084615": "Urbana",
  "n09084750": "Indiana, Hoosier State, IN",
  "n09085089": "Bloomington",
  "n09085209": "Evansville",
  "n09085334": "Fort Wayne",
  "n09085441": "Gary",
  "n09085593": "Indianapolis, capital of Indiana",
  "n09085830": "Lafayette",
  "n09085967": "Muncie",
  "n09086070": "South Bend",
  "n09086173": "Iowa, Hawkeye State, IA",
  "n09086509": "Council Bluffs",
  "n09086635": "Davenport",
  "n09086793": "Cedar Rapids",
  "n09086894": "Clinton",
  "n09086995": "Des Moines, capital of Iowa",
  "n09087126": "Dubuque",
  "n09087247": "Mason City",
  "n09087352": "Ottumwa",
  "n09087450": "Sioux City",
  "n09087599": "Kansas, Sunflower State, KS",
  "n09087996": "Dodge City",
  "n09088151": "Abilene",
  "n09088301": "Hays",
  "n09088396": "Kansas City",
  "n09088556": "Lawrence",
  "n09088718": "Salina",
  "n09088815": "Topeka, capital of Kansas",
  "n09088989": "Wichita",
  "n09089139": "Kentucky, Bluegrass State, KY",
  "n09089524": "Bowling Green",
  "n09089631": "Frankfort, capital of Kentucky",
  "n09089782": "Lexington",
  "n09089923": "Louisville",
  "n09090128": "Owensboro",
  "n09090271": "Paducah",
  "n09090389": "Bluegrass, Bluegrass Country, Bluegrass Region",
  "n09090559": "Louisiana Purchase",
  "n09090825": "Louisiana, Pelican State, LA",
  "n09091285": "Alexandria",
  "n09091398": "Baton Rouge, capital of Louisiana",
  "n09091538": "Lafayette",
  "n09091668": "Monroe",
  "n09091774": "Morgan City",
  "n09091909": "New Orleans",
  "n09092352": "Shreveport",
  "n09092497": "Maine, Pine Tree State, ME",
  "n09092822": "Augusta, capital of Maine",
  "n09092948": "Bangor",
  "n09093072": "Brunswick",
  "n09093187": "Lewiston",
  "n09093315": "Orono",
  "n09093472": "Portland",
  "n09093608": "Maryland, Old Line State, Free State, MD",
  "n09093980": "Maryland",
  "n09094093": "Aberdeen",
  "n09094217": "Annapolis, capital of Maryland",
  "n09094381": "Baltimore",
  "n09094581": "Fort Meade, Fort George Gordon Meade, Fort George G. Meade",
  "n09094791": "Frederick",
  "n09094919": "Hagerstown",
  "n09095023": "Massachusetts, Bay State, Old Colony, MA",
  "n09095606": "Massachusetts, Massachusetts Bay Colony",
  "n09095751": "Boston, Hub of the Universe, Bean Town, Beantown, capital of Massachusetts",
  "n09096089": "Boston Harbor",
  "n09096190": "Beacon Hill",
  "n09096343": "Breed's Hill",
  "n09096498": "Charlestown",
  "n09096664": "Cambridge",
  "n09096903": "Concord",
  "n09097079": "Gloucester",
  "n09097283": "Lexington",
  "n09097461": "Medford",
  "n09097599": "Pittsfield",
  "n09097707": "Springfield",
  "n09097871": "Worcester",
  "n09098027": "Cape Ann",
  "n09098184": "Cape Cod",
  "n09098358": "Cape Cod Canal",
  "n09098488": "Martha's Vineyard",
  "n09098592": "Nantucket",
  "n09098721": "Plymouth",
  "n09098885": "Plymouth Colony",
  "n09099098": "Plymouth Rock",
  "n09099264": "Salem",
  "n09099411": "Williamstown",
  "n09099526": "Michigan, Wolverine State, Great Lakes State, MI",
  "n09100080": "Alpena",
  "n09100223": "Ann Arbor",
  "n09100394": "Detroit, Motor City, Motown",
  "n09100690": "Flint",
  "n09100837": "Grand Rapids",
  "n09100982": "Houghton",
  "n09101127": "Jackson",
  "n09101233": "Kalamazoo",
  "n09101318": "Lansing, capital of Michigan",
  "n09101495": "Marquette",
  "n09101639": "Monroe",
  "n09101753": "Saginaw",
  "n09101882": "Traverse City",
  "n09102016": "Minnesota, Gopher State, North Star State, MN",
  "n09102415": "Bemidji",
  "n09102517": "Duluth",
  "n09102655": "Hibbing",
  "n09102781": "Mankato",
  "n09102883": "Minneapolis",
  "n09103112": "Rochester",
  "n09103217": "Saint Cloud, St. Cloud",
  "n09103377": "Saint Paul, St. Paul, capital of Minnesota",
  "n09103648": "Twin Cities",
  "n09103803": "Virginia",
  "n09103943": "Mississippi, Magnolia State, MS",
  "n09104474": "Biloxi",
  "n09104604": "Columbus",
  "n09104737": "Greenville",
  "n09104894": "Hattiesburg",
  "n09105003": "Jackson, capital of Mississippi",
  "n09105156": "Meridian",
  "n09105260": "Natchez",
  "n09105390": "Tupelo",
  "n09105494": "Vicksburg",
  "n09105821": "Missouri, Show Me State, MO",
  "n09106393": "Cape Girardeau",
  "n09106502": "Columbia",
  "n09106614": "Hannibal",
  "n09106770": "Independence",
  "n09106912": "Jefferson City, capital of Missouri",
  "n09107098": "Kansas City",
  "n09107305": "Poplar Bluff",
  "n09107412": "Saint Joseph, St. Joseph",
  "n09107626": "Saint Louis, St. Louis, Gateway to the West",
  "n09107950": "Sedalia",
  "n09108055": "Springfield",
  "n09108164": "Montana, Treasure State, MT",
  "n09108586": "Bozeman",
  "n09108728": "Billings",
  "n09108884": "Butte",
  "n09109012": "Great Falls",
  "n09109180": "Helena, capital of Montana",
  "n09109333": "Missoula",
  "n09109444": "Nebraska, Cornhusker State, NE",
  "n09109771": "Grand Island",
  "n09109882": "Lincoln, capital of Nebraska",
  "n09110099": "North Platte",
  "n09110229": "Omaha",
  "n09110422": "Nevada, Silver State, Battle Born State, Sagebrush State, NV",
  "n09110784": "Carson City, capital of Nevada",
  "n09110939": "Las Vegas",
  "n09111168": "Reno",
  "n09111366": "New Hampshire, Granite State, NH",
  "n09111636": "New Hampshire",
  "n09111754": "Concord, capital of New Hampshire",
  "n09111955": "Manchester",
  "n09112123": "Portsmouth",
  "n09112282": "New Jersey, Jersey, Garden State, NJ",
  "n09112742": "New Jersey",
  "n09112857": "Atlantic City",
  "n09113022": "Trenton, capital of New Jersey",
  "n09113207": "Bayonne",
  "n09113333": "Camden",
  "n09113479": "Jersey City",
  "n09113611": "Morristown",
  "n09113762": "Newark",
  "n09113901": "New Brunswick",
  "n09114020": "Paterson",
  "n09114128": "Princeton",
  "n09114262": "Cape May",
  "n09114401": "Liberty Island, Bedloe's Island",
  "n09114696": "New Mexico, Land of Enchantment, NM",
  "n09115315": "Albuquerque",
  "n09115478": "Carlsbad",
  "n09115646": "Farmington",
  "n09115756": "Gallup",
  "n09115886": "Las Cruces",
  "n09116010": "Los Alamos",
  "n09116214": "Roswell",
  "n09116318": "Santa Fe, capital of New Mexico",
  "n09116488": "Silver City",
  "n09116599": "Taos",
  "n09116709": "Manhattan Island",
  "n09116876": "New Amsterdam",
  "n09117118": "New Netherland",
  "n09117351": "New York, New York State, Empire State, NY",
  "n09118181": "New York",
  "n09118313": "Albany, capital of New York",
  "n09118505": "Buffalo",
  "n09118639": "Cooperstown",
  "n09118817": "Erie Canal",
  "n09119070": "New York State Barge Canal",
  "n09119277": "New York, New York City, Greater New York",
  "n09119989": "Bronx",
  "n09120087": "Brooklyn",
  "n09120207": "Coney Island",
  "n09120353": "Ellis Island",
  "n09120594": "Manhattan",
  "n09120939": "Fifth Avenue",
  "n09121101": "Seventh Avenue",
  "n09121230": "Central Park",
  "n09121334": "Harlem",
  "n09121458": "Hell's Kitchen, Hell's Half Acre",
  "n09121619": "SoHo, South of Houston",
  "n09121814": "Ithaca",
  "n09121936": "Bowery",
  "n09122086": "Broadway, Great White Way",
  "n09122258": "Park Avenue, Park Ave.",
  "n09122397": "off-Broadway",
  "n09122542": "Times Square",
  "n09122779": "Wall Street, Wall St.",
  "n09122968": "Greenwich Village, Village",
  "n09123182": "Queens",
  "n09123281": "Staten Island",
  "n09123387": "East River",
  "n09123538": "Harlem River",
  "n09123662": "Verrazano Narrows",
  "n09123809": "West Point",
  "n09124039": "Long Island",
  "n09124252": "Elmont",
  "n09124399": "Kennedy, Kennedy Interrnational, Kennedy International Airport",
  "n09124589": "Binghamton",
  "n09124732": "Kingston",
  "n09124845": "Newburgh",
  "n09125016": "Niagara Falls",
  "n09125203": "Rochester",
  "n09125354": "Schenectady",
  "n09125528": "Syracuse",
  "n09125629": "Utica",
  "n09125727": "Saratoga Springs",
  "n09125881": "Watertown",
  "n09125984": "borscht circuit, borsht circuit, borscht belt, borsht belt",
  "n09126305": "North Carolina, Old North State, Tar Heel State, NC",
  "n09126895": "North Carolina",
  "n09127014": "Cape Fear",
  "n09127161": "Cape Flattery",
  "n09127274": "Cape Froward",
  "n09127461": "Cape Hatteras",
  "n09127701": "Hatteras Island",
  "n09127844": "Raleigh, capital of North Carolina",
  "n09128040": "Asheville",
  "n09128201": "Chapel Hill",
  "n09128372": "Charlotte, Queen City",
  "n09128536": "Durham",
  "n09128691": "Fayetteville",
  "n09128808": "Goldsboro",
  "n09128947": "Greensboro",
  "n09129062": "Greenville",
  "n09129187": "Wilmington",
  "n09129324": "Winston-Salem",
  "n09129442": "North Dakota, Peace Garden State, ND",
  "n09129719": "Bismarck, capital of North Dakota",
  "n09129926": "Fargo",
  "n09130076": "Ohio, Buckeye State, OH",
  "n09130452": "Akron",
  "n09130599": "Athens",
  "n09130714": "Cleveland",
  "n09130883": "Cincinnati",
  "n09131001": "Columbus, capital of Ohio",
  "n09131205": "Dayton",
  "n09131324": "Mansfield",
  "n09131428": "Toledo",
  "n09131553": "Youngstown",
  "n09131654": "Oklahoma, Sooner State, OK",
  "n09132053": "Bartlesville",
  "n09132163": "Enid",
  "n09132266": "Lawton",
  "n09132367": "McAlester",
  "n09132474": "Muskogee",
  "n09132597": "Oklahoma City, capital of Oklahoma",
  "n09132778": "Tulsa",
  "n09133010": "Oregon, Beaver State, OR",
  "n09133364": "Bend",
  "n09133500": "Eugene",
  "n09133643": "Klamath Falls",
  "n09133775": "Medford",
  "n09133895": "Portland",
  "n09134202": "Salem, capital of Oregon",
  "n09134386": "Pennsylvania, Keystone State, PA",
  "n09134882": "Pennsylvania",
  "n09134999": "Allentown",
  "n09135142": "Altoona",
  "n09135246": "Bethlehem",
  "n09135447": "Erie",
  "n09135590": "Gettysburg",
  "n09135733": "Harrisburg, capital of Pennsylvania",
  "n09135894": "Hershey",
  "n09135993": "Chester",
  "n09136182": "Philadelphia, City of Brotherly Love",
  "n09136582": "Pittsburgh",
  "n09136929": "Scranton",
  "n09137032": "Rhode Island, Little Rhody, Ocean State, RI",
  "n09137334": "Rhode Island",
  "n09137451": "Providence, capital of Rhode Island",
  "n09137682": "Newport",
  "n09137869": "South Carolina, Palmetto State, SC",
  "n09138239": "South Carolina",
  "n09138358": "Columbia, capital of South Carolina",
  "n09138538": "Charleston",
  "n09138676": "Florence",
  "n09138808": "Greenville",
  "n09138935": "South Dakota, Coyote State, Mount Rushmore State, SD",
  "n09139380": "Aberdeen",
  "n09139508": "Pierre, capital of South Dakota",
  "n09139698": "Rapid City",
  "n09139849": "Sioux Falls",
  "n09139993": "Black Hills",
  "n09140148": "Tennessee, Volunteer State, TN",
  "n09140569": "Chattanooga",
  "n09140674": "Columbia",
  "n09140781": "Jackson",
  "n09140882": "Johnson City",
  "n09140993": "Knoxville",
  "n09141119": "Memphis",
  "n09141297": "Nashville, capital of Tennessee",
  "n09141526": "Texas, Lone-Star State, TX",
  "n09142674": "Abilene",
  "n09142771": "Amarillo",
  "n09142887": "Arlington",
  "n09143017": "Austin, capital of Texas",
  "n09143205": "Beaumont",
  "n09143321": "Brownsville",
  "n09143549": "Bryan",
  "n09143649": "Corpus Christi",
  "n09143786": "Dallas",
  "n09143973": "Del Rio",
  "n09144117": "El Paso",
  "n09144323": "Fort Worth",
  "n09144484": "Galveston",
  "n09144605": "Galveston Island",
  "n09144730": "Garland",
  "n09144851": "Houston",
  "n09145083": "Laredo",
  "n09145217": "Lubbock",
  "n09145341": "Lufkin",
  "n09145437": "McAllen",
  "n09145553": "Midland",
  "n09145655": "Odessa",
  "n09145751": "Paris",
  "n09145851": "Plano",
  "n09145970": "San Angelo",
  "n09146111": "San Antonio",
  "n09146324": "Sherman",
  "n09146451": "Texarkana",
  "n09146584": "Tyler",
  "n09146681": "Victoria",
  "n09146813": "Waco",
  "n09146912": "Wichita Falls",
  "n09147046": "Utah, Beehive State, Mormon State, UT",
  "n09147504": "Ogden",
  "n09147618": "Provo",
  "n09147737": "Salt Lake City, capital of Utah",
  "n09147964": "Vermont, Green Mountain State, VT",
  "n09148259": "Montpelier, capital of Vermont",
  "n09148422": "Bennington",
  "n09148529": "Brattleboro",
  "n09148662": "Burlington",
  "n09148871": "Rutland",
  "n09148970": "Virginia, Old Dominion, Old Dominion State, VA",
  "n09149934": "Virginia",
  "n09150047": "Richmond, capital of Virginia",
  "n09150284": "Blacksburg",
  "n09150448": "Jamestown",
  "n09150662": "Newport News",
  "n09150863": "Norfolk",
  "n09151114": "Lynchburg",
  "n09151216": "Portsmouth",
  "n09151411": "Roanoke",
  "n09151516": "Virginia Beach",
  "n09151800": "Bull Run",
  "n09151963": "Chancellorsville",
  "n09152080": "Fredericksburg",
  "n09152218": "Petersburg",
  "n09152401": "Spotsylvania",
  "n09152570": "Yorktown",
  "n09152769": "Mount Vernon",
  "n09152944": "Washington, Evergreen State, WA",
  "n09153570": "Aberdeen",
  "n09153710": "Bellingham",
  "n09153873": "Kennewick",
  "n09154000": "Olympia, capital of Washington",
  "n09154178": "Seattle",
  "n09154607": "Spokane",
  "n09154731": "Tacoma",
  "n09154905": "Vancouver",
  "n09155065": "Walla Walla",
  "n09155199": "Yakima",
  "n09155306": "West Virginia, Mountain State, WV",
  "n09155692": "Beckley",
  "n09155798": "Charleston, capital of West Virginia",
  "n09155986": "Clarksburg",
  "n09156095": "Fayetteville",
  "n09156241": "Huntington",
  "n09156395": "Harpers Ferry, Harper's Ferry",
  "n09156666": "Morgantown",
  "n09156889": "Parkersburg",
  "n09157021": "Wheeling",
  "n09157163": "Wisconsin, Badger State, WI",
  "n09157555": "Appleton",
  "n09157657": "Eau Claire",
  "n09157766": "Green Bay",
  "n09157896": "La Crosse",
  "n09158024": "Madison, capital of Wisconsin",
  "n09158268": "Milwaukee",
  "n09158501": "Racine",
  "n09158649": "Superior",
  "n09158789": "Watertown",
  "n09158897": "Wausau",
  "n09159003": "Wyoming, Equality State, WY",
  "n09159546": "Casper",
  "n09159675": "Cheyenne, capital of Wyoming",
  "n09159859": "Jackson",
  "n09159958": "Lander",
  "n09160056": "Laramie",
  "n09160168": "Rock Springs",
  "n09160295": "Uruguay",
  "n09160571": "Montevideo, capital of Uruguay",
  "n09160775": "Vanuatu, Republic of Vanuatu, New Hebrides",
  "n09160968": "Port Vila, Vila, capital of Vanuatu",
  "n09161090": "Holy See, The Holy See, State of the Vatican City",
  "n09161452": "Vatican City, Citta del Vaticano",
  "n09161615": "Guiana Highlands",
  "n09161803": "Venezuela, Republic of Venezuela",
  "n09162276": "Caracas, capital of Venezuela",
  "n09162414": "Ciudad Bolivar",
  "n09162581": "Cumana",
  "n09162803": "Maracaibo",
  "n09162955": "Maracay",
  "n09163077": "Valencia",
  "n09163192": "Vietnam, Socialist Republic of Vietnam, Viet Nam, Annam",
  "n09163584": "North Vietnam",
  "n09163844": "South Vietnam",
  "n09164095": "Hanoi, capital of Vietnam",
  "n09164241": "Ho Chi Minh City, Saigon",
  "n09164417": "Haiphong",
  "n09164561": "Yemen, Republic of Yemen",
  "n09164903": "Aden",
  "n09165146": "Hodeida, Al-Hudaydah",
  "n09165294": "Mukalla, Al-Mukalla",
  "n09165464": "Sana, Sanaa, Sana'a",
  "n09165613": "Zambia, Republic of Zambia, Northern Rhodesia",
  "n09165996": "Lusaka, capital of Zambia",
  "n09166127": "Low Countries",
  "n09166304": "Lusitania",
  "n09166534": "Silesia, Slask, Slezsko, Schlesien",
  "n09166756": "Big Sur",
  "n09166902": "Silicon Valley",
  "n09167101": "Zimbabwe, Republic of Zimbabwe, Rhodesia, Southern Rhodesia",
  "n09167505": "Harare, Salisbury, capital of Zimbabwe",
  "n09167652": "Bulawayo",
  "n09167767": "Arabian Desert, Great Arabian Desert",
  "n09168020": "Arabian Desert, Eastern Desert",
  "n09168193": "Atacama Desert",
  "n09168336": "Australian Desert, Great Australian Desert",
  "n09168592": "Black Rock Desert",
  "n09168707": "Chihuahuan Desert",
  "n09168915": "Colorado Desert",
  "n09169038": "Dasht-e-Kavir, Kavir Desert, Great Salt Desert",
  "n09169188": "Dasht-e-Lut, Lut Desert",
  "n09169303": "Death Valley",
  "n09169557": "Gibson Desert",
  "n09169690": "Gila Desert",
  "n09169801": "Gobi, Gobi Desert",
  "n09169930": "Great Sandy Desert",
  "n09170109": "Great Victoria Desert",
  "n09170294": "Kalahari, Kalahari Desert",
  "n09170475": "Kara Kum, Qara Qum, Turkestan Desert",
  "n09170633": "Kyzyl Kum, Kizil Kum, Qizil Qum",
  "n09170788": "Libyan Desert",
  "n09170996": "Mojave, Mojave Desert, Mohave, Mohave Desert",
  "n09171204": "Namib Desert",
  "n09171376": "Nefud, An Nefud, Nafud, An Nafud",
  "n09171560": "Negev, Negev Desert",
  "n09171674": "Nubian Desert",
  "n09171853": "Painted Desert",
  "n09171984": "Patagonian Desert",
  "n09172111": "Rub al-Khali, Ar Rimsal, Dahna, Great Sandy Desert",
  "n09172283": "Sahara, Sahara Desert",
  "n09172480": "Sub-Saharan Africa, Black Africa",
  "n09172634": "Simpson Desert",
  "n09172751": "Sinai, Sinai Desert",
  "n09172910": "Sonoran Desert",
  "n09173023": "Syrian Desert",
  "n09173288": "Taklimakan Desert, Taklamakan Desert",
  "n09173417": "Thar Desert, Great Indian Desert",
  "n09173623": "Cameroon",
  "n09173777": "Citlaltepetl, Mount Orizaba, Mt Orizaba, Pico de Orizaba",
  "n09174015": "Colima, Nevado de Colima, Volcan de Colima",
  "n09174166": "Cotacachi",
  "n09174301": "Cotopaxi",
  "n09174457": "Demavend",
  "n09174566": "El Misti",
  "n09174718": "Etna, Mount Etna, Mt Etna",
  "n09174908": "Fuego",
  "n09175016": "Fuji, Mount Fuji, Fujiyama, Fujinoyama, Fuji-san",
  "n09175322": "Galeras, Pasto",
  "n09175459": "Guallatiri",
  "n09175617": "Huainaputina",
  "n09175767": "Klyuchevskaya",
  "n09175915": "Krakatau, Krakatao, Krakatoa",
  "n09176159": "New Siberian Islands",
  "n09176342": "Lascar",
  "n09176446": "Mauna Kea",
  "n09176608": "Mauna Loa",
  "n09176732": "Nyamuragira",
  "n09176844": "Nyiragongo",
  "n09176955": "Purace",
  "n09177103": "Sangay",
  "n09177249": "Tupungatito",
  "n09177385": "Mount Saint Helens, Mount St. Helens, Mt. St. Helens",
  "n09177647": "Scythia",
  "n09177883": "Vesuvius, Mount Vesuvius, Mt. Vesuvius",
  "n09178141": "North Africa",
  "n09178310": "West Africa",
  "n09178481": "Dar al-Islam, House of Islam",
  "n09178596": "Dar al-harb, House of War",
  "n09178727": "life",
  "n09178821": "rational motive",
  "n09178999": "reason, ground",
  "n09179264": "occasion",
  "n09179382": "score, account",
  "n09179606": "why, wherefore",
  "n09179776": "incentive, inducement, motivator",
  "n09179962": "moral force, dynamic",
  "n09180118": "disincentive, deterrence",
  "n09180259": "irrational motive",
  "n09180431": "urge, impulse",
  "n09180616": "abience",
  "n09180791": "adience",
  "n09180967": "death instinct, death wish, Thanatos",
  "n09181114": "irrational impulse",
  "n09181330": "compulsion, irresistible impulse",
  "n09181557": "mania, passion, cacoethes",
  "n09181889": "agromania",
  "n09181993": "dipsomania, alcoholism, potomania",
  "n09182142": "egomania",
  "n09182273": "kleptomania",
  "n09182401": "logorrhea, logomania",
  "n09182524": "monomania, possession",
  "n09182650": "necrophilia, necrophilism, necromania",
  "n09182784": "phaneromania",
  "n09182929": "pyromania",
  "n09183031": "trichotillomania",
  "n09183140": "wanderlust, itchy feet",
  "n09183255": "compulsion, obsession",
  "n09183532": "onomatomania",
  "n09183693": "ethical motive, ethics, morals, morality",
  "n09183971": "hedonism",
  "n09184136": "conscience, scruples, moral sense, sense of right and wrong",
  "n09184405": "wee small voice, small voice, voice of conscience",
  "n09184543": "sense of shame, sense of duty",
  "n09184668": "Inner Light, Light, Light Within, Christ Within",
  "n09184834": "psychic energy, mental energy",
  "n09184975": "incitement, incitation, provocation",
  "n09185280": "signal",
  "n09185440": "libidinal energy",
  "n09185612": "cathexis, charge",
  "n09185865": "acathexis",
  "n09186064": "Aare, Aar, Aare River",
  "n09186225": "Abukir, Abukir Bay",
  "n09186359": "abyss, abysm",
  "n09186592": "abyssal zone",
  "n09186709": "Acheron, River Acheron",
  "n09186928": "achondrite",
  "n09187036": "acicula",
  "n09187204": "Aconcagua",
  "n09187407": "Adams, Mount Adams",
  "n09187594": "Adam's Peak, Samanala",
  "n09187743": "Adige, River Adige",
  "n09187923": "Adirondacks, Adirondack Mountains",
  "n09188094": "Admiralty Range",
  "n09188229": "adjunct",
  "n09188383": "Adriatic, Adriatic Sea",
  "n09188609": "Aegean, Aegean Sea",
  "n09188894": "Aegospotami, Aegospotamos",
  "n09189157": "aerie, aery, eyrie, eyry",
  "n09189289": "aerolite",
  "n09189411": "Africa",
  "n09190918": "agent",
  "n09191427": "airborne transmission",
  "n09191635": "air bubble",
  "n09191707": "Aire, River Aire, Aire River",
  "n09191875": "Alabama, Alabama River",
  "n09192122": "Alaska Peninsula",
  "n09192280": "Alaska Range",
  "n09192462": "Aldebaran",
  "n09192566": "Algol",
  "n09192708": "Alleghenies, Allegheny Mountains",
  "n09192973": "Allegheny, Allegheny River",
  "n09193282": "alluvial sediment, alluvial deposit, alluvium, alluvion",
  "n09193551": "alluvial flat, alluvial plain",
  "n09193705": "alp",
  "n09193772": "Alpha Centauri, Rigil Kent, Rigil",
  "n09193933": "Alpha Crucis",
  "n09194052": "alpha particle",
  "n09194227": "Alpine glacier, Alpine type of glacier",
  "n09194357": "Alps, the Alps",
  "n09194710": "Altai Mountains, Altay Mountains",
  "n09194938": "Altair",
  "n09195098": "altocumulus, altocumulus cloud",
  "n09195235": "altostratus, altostratus cloud",
  "n09195372": "Amazon, Amazon River",
  "n09195615": "America",
  "n09195796": "American Falls",
  "n09195958": "ammonite, ammonoid",
  "n09196103": "Amur, Amur River, Heilong Jiang, Heilong",
  "n09196302": "Ancohuma",
  "n09196454": "Andaman Sea",
  "n09196611": "Andes",
  "n09197271": "Andromeda",
  "n09197432": "Angara, Angara River, Tunguska, Upper Tunguska",
  "n09197660": "Angel, Angel Falls",
  "n09197815": "anion",
  "n09197945": "Annapurna, Anapurna",
  "n09198106": "Antarctica, Antarctic continent",
  "n09198574": "Antarctic Ocean",
  "n09198755": "Antarctic Peninsula, Palmer Peninsula",
  "n09198997": "Antares",
  "n09199101": "anthill, formicary",
  "n09199219": "antibaryon",
  "n09199341": "antilepton",
  "n09199498": "antimeson",
  "n09199581": "antimuon, positive muon",
  "n09199728": "antineutrino",
  "n09199817": "antineutron",
  "n09199904": "antiparticle",
  "n09200193": "antiproton",
  "n09200317": "antiquark",
  "n09200419": "antitauon, tau-plus particle",
  "n09200527": "Antlia",
  "n09200649": "Apalachicola, Apalachicola River",
  "n09200874": "Apennines",
  "n09201031": "aperture",
  "n09201189": "Apollo asteroid",
  "n09201301": "Appalachians, Appalachian Mountains",
  "n09201714": "Apus",
  "n09201840": "Aquarius",
  "n09201998": "aquifer",
  "n09202142": "Aquila",
  "n09202285": "Ara",
  "n09202405": "Arabian Sea",
  "n09202603": "Arafura Sea",
  "n09202810": "Araguaia, Araguaia River, Araguaya, Araguaya River",
  "n09203032": "Ararat, Mount Ararat, Mt. Ararat",
  "n09203217": "Aras, Araxes",
  "n09203481": "Arauca",
  "n09203677": "archeological remains",
  "n09203827": "archipelago",
  "n09204977": "Arctic Ocean",
  "n09205334": "Arcturus",
  "n09205509": "arete",
  "n09205607": "Argo",
  "n09205890": "Argun, Argun River, Ergun He",
  "n09206133": "Aries",
  "n09206289": "Aristarchus",
  "n09206375": "Arkansas, Arkansas River",
  "n09206693": "Arno, Arno River, River Arno",
  "n09206896": "arroyo",
  "n09206985": "ascent, acclivity, rise, raise, climb, upgrade",
  "n09207288": "Asia",
  "n09208496": "asterism",
  "n09208702": "asteroid",
  "n09209025": "asthenosphere",
  "n09209132": "Atacama Trench",
  "n09209263": "Atlantic, Atlantic Ocean",
  "n09210236": "Atlantic Coast",
  "n09210346": "Atlas Mountains",
  "n09210604": "atmosphere",
  "n09210862": "atoll",
  "n09211056": "Auriga, Charioteer",
  "n09211266": "Australia",
  "n09211735": "Australian Alps",
  "n09211944": "Avon, River Avon, Upper Avon, Upper Avon River",
  "n09212151": "Avon, River Avon",
  "n09212360": "backwater",
  "n09212572": "badlands",
  "n09212690": "Baffin Bay",
  "n09212935": "Balaton, Lake Balaton, Plattensee",
  "n09213076": "Balkans, Balkan Mountains, Balkan Mountain Range",
  "n09213254": "Baltic, Baltic Sea",
  "n09213434": "bank",
  "n09213565": "bank",
  "n09213828": "bank, cant, camber",
  "n09214060": "bar",
  "n09214269": "barbecue pit",
  "n09214422": "Barents Sea",
  "n09214581": "barrier",
  "n09214760": "barrier island",
  "n09214916": "barrier reef",
  "n09215023": "baryon, heavy particle",
  "n09215315": "base",
  "n09215437": "basin",
  "n09215664": "bay, embayment",
  "n09216442": "Bay of Bengal",
  "n09216588": "Bay of Biscay",
  "n09216781": "Bay of Fundy",
  "n09216968": "Bay of Naples",
  "n09217086": "bayou",
  "n09217230": "beach",
  "n09217414": "beachfront",
  "n09217508": "Beaufort Sea",
  "n09217638": "bed, bottom",
  "n09217867": "bed",
  "n09218044": "bedrock",
  "n09218159": "beehive, hive",
  "n09218315": "honeycomb",
  "n09218494": "belay",
  "n09218641": "ben",
  "n09218782": "Bering Sea",
  "n09218963": "Bering Strait",
  "n09219078": "Berkshires, Berkshire Hills",
  "n09219233": "berm",
  "n09219349": "Beta Centauri",
  "n09219467": "Beta Crucis",
  "n09219592": "beta particle",
  "n09219731": "Betelgeuse, Alpha Orionis",
  "n09219858": "Big Dipper, Dipper, Plough, Charles's Wain, Wain, Wagon",
  "n09220046": "Bighorn, Bighorn River",
  "n09220239": "bight",
  "n09220434": "Bight of Benin",
  "n09220574": "Big Sioux River",
  "n09220770": "billabong",
  "n09220922": "billabong",
  "n09221070": "binary star, binary, double star",
  "n09221424": "biological agent, biohazard",
  "n09221571": "bird's nest, bird nest, birdnest",
  "n09221723": "Biscayne Bay",
  "n09221900": "Bismarck Sea",
  "n09222051": "bit, chip, flake, fleck, scrap",
  "n09222406": "black body, blackbody, full radiator",
  "n09222742": "Black Forest, Schwarzwald",
  "n09222880": "Black Hills",
  "n09223177": "black hole",
  "n09223325": "Black Sea, Euxine Sea",
  "n09223487": "bladder stone, cystolith",
  "n09223591": "blade",
  "n09223725": "blanket, mantle",
  "n09223911": "blood-brain barrier",
  "n09224193": "Blue Nile",
  "n09224325": "Blue Ridge Mountains, Blue Ridge",
  "n09224566": "blue sky, blue, blue air, wild blue yonder",
  "n09224725": "bluff",
  "n09224828": "b-meson",
  "n09224911": "body",
  "n09225146": "body of water, water",
  "n09225943": "bog, peat bog",
  "n09226209": "Bo Hai, Po Hai",
  "n09226367": "bolt-hole",
  "n09226498": "bonanza",
  "n09226591": "Bonete",
  "n09226738": "Bootes",
  "n09226869": "borrow pit",
  "n09226997": "boson",
  "n09227219": "Bosporus",
  "n09227428": "bottomland, bottom",
  "n09227530": "bottom quark, beauty quark",
  "n09227683": "Bougainville Trench",
  "n09227839": "boulder, bowlder",
  "n09228055": "brae",
  "n09228144": "Brahmaputra, Brahmaputra River",
  "n09228324": "branch",
  "n09228492": "branched chain",
  "n09228619": "Brazos, Brazos River",
  "n09228801": "breach",
  "n09228928": "Brenner Pass",
  "n09229158": "brickbat",
  "n09229249": "Bristol Channel",
  "n09229409": "brook, creek",
  "n09229641": "brooklet",
  "n09229709": "bubble",
  "n09229941": "bullet hole",
  "n09230041": "burrow, tunnel",
  "n09230202": "butte",
  "n09230361": "Buzzards Bay",
  "n09230500": "Cachi",
  "n09230646": "Caelum",
  "n09230768": "calculus, concretion",
  "n09231117": "caldera",
  "n09231263": "Callisto",
  "n09231361": "Caloosahatchee, Caloosahatchee River",
  "n09231587": "Cam, River Cam, Cam River",
  "n09231761": "Cambrian Mountains",
  "n09231890": "Canadian, Canadian River",
  "n09232165": "Canadian Falls, Horseshoe Falls",
  "n09232317": "canal",
  "n09232526": "Canandaigua Lake, Lake Canandaigua",
  "n09232687": "Cancer",
  "n09232841": "Canis Major, Great Dog",
  "n09232989": "Canis Minor, Little Dog",
  "n09233134": "Canopus",
  "n09233284": "Cantabrian Mountains",
  "n09233446": "canyon, canon",
  "n09233603": "canyonside",
  "n09233715": "cape, ness",
  "n09234104": "Cape Canaveral, Cape Kennedy",
  "n09234374": "Cape Cod Bay",
  "n09234491": "Cape Fear River",
  "n09234654": "Capella",
  "n09234756": "Cape Sable",
  "n09234885": "Cape Sable",
  "n09235053": "Cape Trafalgar",
  "n09235244": "Cape York",
  "n09235469": "Cape York Peninsula",
  "n09235713": "Capricornus, Capricorn",
  "n09235894": "Caribbean, Caribbean Sea",
  "n09236077": "Carina",
  "n09236251": "Carlsbad Caverns",
  "n09236423": "Carpathians, Carpathian Mountains",
  "n09236766": "carpet",
  "n09236957": "cascade",
  "n09237076": "Cascades, Cascade Range, Cascade Mountains",
  "n09237404": "Caspian, Caspian Sea",
  "n09237627": "Cassiopeia",
  "n09237749": "Castor, Alpha Geminorum",
  "n09237918": "cataract",
  "n09238030": "Cataract Canyon",
  "n09238143": "catch",
  "n09238294": "cation",
  "n09238425": "Catskills, Catskill Mountains",
  "n09238674": "Caucasus, Caucasus Mountains",
  "n09238926": "cave",
  "n09239302": "cavern",
  "n09239451": "cavern",
  "n09239581": "Cayuga Lake, Lake Cayuga",
  "n09239740": "celestial body, heavenly body",
  "n09240051": "Centaurus, Centaur",
  "n09240272": "Cepheus",
  "n09240409": "Ceres",
  "n09240507": "Cetus",
  "n09240621": "chain, chemical chain",
  "n09240889": "Chamaeleon, Chameleon",
  "n09241047": "Changtzu",
  "n09241247": "channel",
  "n09241712": "Chao Phraya",
  "n09241929": "chap",
  "n09242037": "Charles, Charles River",
  "n09242232": "charm quark",
  "n09242389": "chasm",
  "n09242514": "Chattahoochee, Chattahoochee River",
  "n09242767": "Baikal, Lake Baikal, Baykal, Lake Baykal",
  "n09242956": "Lake Chelan",
  "n09243100": "Coeur d'Alene Lake",
  "n09243209": "Lake Tahoe",
  "n09243405": "Chesapeake Bay",
  "n09243615": "Chimborazo",
  "n09243769": "chink",
  "n09243906": "chip, cow chip, cow dung, buffalo chip",
  "n09244022": "Chiron",
  "n09244191": "chondrite",
  "n09244331": "chondrule",
  "n09244469": "chromosphere",
  "n09244683": "Chukchi Peninsula",
  "n09244831": "Chukchi Sea",
  "n09244972": "Cimarron, Cimarron River",
  "n09245212": "cinder, clinker",
  "n09245370": "Circinus",
  "n09245515": "cirque, corrie, cwm",
  "n09245650": "cirrocumulus, cirrocumulus cloud",
  "n09245833": "cirrostratus, cirrostratus cloud",
  "n09245964": "cirrus, cirrus cloud",
  "n09246134": "clast",
  "n09246292": "clastic rock",
  "n09246464": "cliff, drop, drop-off",
  "n09246660": "Clinch River",
  "n09246883": "closed chain, ring",
  "n09247071": "closed universe",
  "n09247410": "cloud",
  "n09247847": "cloud bank",
  "n09247942": "Clyde",
  "n09248153": "coast",
  "n09248294": "coastal plain",
  "n09248399": "coastland",
  "n09248477": "Coast Range, Coast Mountains",
  "n09248724": "Cocytus, River Cocytus",
  "n09248914": "coffee grounds",
  "n09249034": "col, gap",
  "n09249155": "collector",
  "n09249272": "collision course",
  "n09249418": "Colorado, Colorado River",
  "n09250016": "Colorado, Colorado River",
  "n09250165": "Colorado Plateau",
  "n09250551": "Columba, Dove",
  "n09250678": "Columbia, Columbia River",
  "n09251002": "coma",
  "n09251229": "Coma Berenices",
  "n09251407": "comet",
  "n09251689": "commemorative",
  "n09251832": "Communism Peak, Mount Communism, Stalin Peak, Mount Garmo",
  "n09252078": "Congo, Congo River, Zaire River",
  "n09252273": "Connecticut, Connecticut River",
  "n09252586": "consolidation",
  "n09252766": "Constance, Lake Constance, Bodensee",
  "n09252970": "constellation",
  "n09254614": "continent",
  "n09255070": "continental glacier",
  "n09255207": "continental shelf",
  "n09255343": "continental slope, bathyal zone, bathyal district",
  "n09255519": "contrail, condensation trail",
  "n09255768": "Cook Strait",
  "n09255921": "Coosa, Coosa River",
  "n09256182": "Copernicus",
  "n09256272": "coprolite",
  "n09256360": "coprolith, fecalith, faecalith, stercolith",
  "n09256479": "coral reef",
  "n09256663": "Coral Sea",
  "n09256815": "core",
  "n09256895": "core",
  "n09257011": "corner",
  "n09257141": "Corona Borealis",
  "n09257280": "Coropuna",
  "n09257429": "Corvus, Crow",
  "n09257563": "couple",
  "n09257761": "cove",
  "n09257843": "cove",
  "n09257949": "covering, natural covering, cover",
  "n09258587": "Crab Nebula",
  "n09258715": "crack, cleft, crevice, fissure, scissure",
  "n09259025": "crag",
  "n09259104": "cranny",
  "n09259219": "crater",
  "n09259376": "Crater",
  "n09259500": "craton",
  "n09259677": "crevasse",
  "n09259746": "Cross-Florida Waterway, Okeechobee Waterway",
  "n09260010": "crust, Earth's crust",
  "n09260218": "crust, incrustation, encrustation",
  "n09260466": "crystal, crystallization",
  "n09260744": "crystallite",
  "n09260907": "cultivated land, farmland, plowland, ploughland, tilled land, tillage, tilth",
  "n09261138": "Cumberland, Cumberland River",
  "n09261407": "Cumberland Gap",
  "n09261604": "Cumberland Mountains, Cumberland Plateau",
  "n09261772": "cumulonimbus, cumulonimbus cloud, thundercloud",
  "n09261960": "cumulus, cumulus cloud",
  "n09262082": "Cuquenan, Cuquenan Falls, Kukenaam, Kukenaam Falls",
  "n09262233": "curtain",
  "n09262371": "cutting",
  "n09262490": "Cygnus",
  "n09262690": "dale",
  "n09262798": "dander",
  "n09262955": "dandruff",
  "n09263087": "Danube, Danube River, Danau",
  "n09263479": "Darling, Darling River",
  "n09263619": "Dead Sea",
  "n09263811": "deep",
  "n09263912": "defile, gorge",
  "n09264021": "Deimos",
  "n09264116": "Delaware, Delaware River",
  "n09264425": "Delaware Bay",
  "n09264599": "dell, dingle",
  "n09264680": "Delphinus",
  "n09264803": "delta",
  "n09265025": "delta ray",
  "n09265134": "Demerara",
  "n09265274": "Denali Fault",
  "n09265392": "Deneb",
  "n09265492": "Denebola",
  "n09265620": "descent, declivity, fall, decline, declination, declension, downslope",
  "n09265910": "desideratum",
  "n09266052": "Detroit River",
  "n09266287": "deuteron",
  "n09266453": "Dhaulagiri",
  "n09266604": "diapir",
  "n09266790": "diffuse nebula, gaseous nebula",
  "n09266946": "dipole",
  "n09267128": "dipole molecule",
  "n09267227": "direct transmission",
  "n09267490": "discard",
  "n09267602": "distributary",
  "n09267758": "ditch",
  "n09267854": "divot",
  "n09268007": "divot",
  "n09268236": "Dnieper, Dnieper River",
  "n09268480": "dog shit, dog do, doggy do, dog turd",
  "n09268592": "Dolomite Alps",
  "n09268778": "Don, Don River",
  "n09268927": "Donner Pass",
  "n09269174": "Dorado",
  "n09269341": "down",
  "n09269472": "downhill",
  "n09269573": "down quark",
  "n09269717": "Draco, Dragon",
  "n09269882": "draw",
  "n09269972": "dregs, settlings",
  "n09270160": "drey",
  "n09270233": "drift",
  "n09270414": "drift ice",
  "n09270508": "drink",
  "n09270657": "drumlin",
  "n09270735": "dune, sand dune",
  "n09270894": "Earth, earth, world, globe",
  "n09271291": "East China Sea",
  "n09271415": "Ebro, Ebro River",
  "n09271558": "Elbe, Elbe River",
  "n09271774": "electric dipole, electric doublet",
  "n09271904": "electron, negatron",
  "n09272085": "elementary particle, fundamental particle",
  "n09272468": "eliminator",
  "n09272595": "Elizabeth River",
  "n09272773": "El Libertador",
  "n09272927": "El Muerto",
  "n09273130": "ember, coal",
  "n09273291": "enclosure, natural enclosure",
  "n09273447": "English Channel",
  "n09273681": "enterolith",
  "n09273776": "envelope",
  "n09273928": "Epsilon Aurigae",
  "n09274032": "Eridanus",
  "n09274152": "escarpment, scarp",
  "n09274305": "esker",
  "n09274500": "estuary",
  "n09274739": "Euphrates, Euphrates River",
  "n09275016": "Eurasia",
  "n09275335": "Europa",
  "n09275473": "Europe",
  "n09276872": "evening star, Hesperus, Vesper",
  "n09277010": "Everest, Mount Everest, Mt. Everest",
  "n09277279": "Everglades",
  "n09277432": "exosphere",
  "n09277538": "expanse",
  "n09277686": "extraterrestrial object, estraterrestrial body",
  "n09277913": "Eyre, Lake Eyre",
  "n09278162": "Eyre Peninsula",
  "n09278295": "fallow",
  "n09278432": "fatigue crack",
  "n09278537": "fault, faulting, geological fault, shift, fracture, break",
  "n09278997": "feeder, tributary, confluent, affluent",
  "n09279161": "fermion",
  "n09279345": "filing",
  "n09279458": "finding",
  "n09279727": "Fingal's Cave",
  "n09279870": "fireball",
  "n09279986": "fireball",
  "n09280113": "fire pit",
  "n09280236": "firestone",
  "n09280380": "firth",
  "n09280573": "Firth of Clyde",
  "n09280731": "Firth of Forth",
  "n09280913": "fishpond",
  "n09280995": "fixed star",
  "n09281104": "fjord, fiord",
  "n09281252": "flare star",
  "n09281411": "flat",
  "n09281545": "Flint, Flint River",
  "n09281777": "floater",
  "n09281914": "floodplain, flood plain",
  "n09282084": "floor",
  "n09282208": "floor",
  "n09282363": "floor",
  "n09282534": "flowage",
  "n09282724": "foam, froth",
  "n09283066": "folium",
  "n09283193": "fomite, vehicle",
  "n09283405": "foothill",
  "n09283514": "footwall",
  "n09283623": "ford, crossing",
  "n09283767": "foreland",
  "n09283866": "foreshore",
  "n09284015": "forest, woodland, timberland, timber",
  "n09284308": "Fornax",
  "n09284433": "Forth, Forth River",
  "n09284589": "fossil",
  "n09284917": "Fountain of Youth",
  "n09285128": "Fox River",
  "n09285254": "fragment",
  "n09285648": "free electron",
  "n09285828": "Galan",
  "n09285979": "Galilean satellite, Galilean",
  "n09286200": "gallstone, bilestone",
  "n09286318": "Galveston Bay",
  "n09286478": "Galway Bay",
  "n09286630": "Ganges, Ganges River",
  "n09286843": "Gan Jiang, Kan River",
  "n09287033": "Ganymede",
  "n09287124": "Garonne, Garonne River",
  "n09287289": "Gasherbrum",
  "n09287415": "gauge boson",
  "n09287613": "Gemini",
  "n09287817": "geode",
  "n09287968": "geological formation, formation",
  "n09288635": "geyser",
  "n09288769": "giant star, giant",
  "n09288946": "Gila, Gila River",
  "n09289177": "glacial boulder",
  "n09289331": "glacier",
  "n09289596": "glen",
  "n09289709": "globule",
  "n09289802": "gluon",
  "n09289913": "Golden Gate",
  "n09290121": "Gondwanaland",
  "n09290350": "gopher hole",
  "n09290444": "gorge",
  "n09290626": "Gosainthan",
  "n09290777": "grain",
  "n09291027": "Grand Canyon",
  "n09291185": "Grand River",
  "n09291340": "Grand Teton",
  "n09291513": "granule",
  "n09291633": "graviton",
  "n09291781": "Great Attractor",
  "n09292007": "Great Australian Bight",
  "n09292189": "Great Bear, Ursa Major",
  "n09292348": "Great Barrier Reef",
  "n09292545": "Great Dividing Range, Eastern Highlands",
  "n09292751": "Great Lakes",
  "n09293011": "Great Plains, Great Plains of North America",
  "n09293340": "Great Rift Valley",
  "n09293613": "Great Salt Lake",
  "n09293744": "Great Slave Lake",
  "n09293917": "Great Smoky Mountains",
  "n09294066": "Green, Green River",
  "n09294285": "Greenland Sea",
  "n09294413": "Green Mountains",
  "n09294599": "greenwood",
  "n09294716": "grinding",
  "n09294877": "grotto, grot",
  "n09294984": "grounds",
  "n09295210": "growler",
  "n09295338": "growth",
  "n09295455": "Grus, Crane",
  "n09295576": "Guadalupe Mountains",
  "n09295790": "Guantanamo Bay",
  "n09295946": "gulch, flume",
  "n09296053": "gulf",
  "n09296121": "gulf",
  "n09296695": "Gulf Coast",
  "n09296804": "Gulf of Aden",
  "n09296937": "Gulf of Alaska",
  "n09297104": "Gulf of Antalya",
  "n09297240": "Gulf of Aqaba, Gulf of Akaba",
  "n09297423": "Gulf of Bothnia",
  "n09297584": "Gulf of California, Sea of Cortes",
  "n09297729": "Gulf of Campeche, Golfo de Campeche, Bay of Campeche",
  "n09297920": "Gulf of Carpentaria, Carpentaria",
  "n09298100": "Gulf of Corinth, Gulf of Lepanto",
  "n09298267": "Gulf of Finland",
  "n09298410": "Gulf of Guinea",
  "n09298569": "Gulf of Martaban",
  "n09298698": "Gulf of Mexico, Golfo de Mexico",
  "n09298974": "Gulf of Ob, Bay of Ob",
  "n09299125": "Gulf of Oman",
  "n09299265": "Gulf of Riga",
  "n09299397": "Gulf of Saint Lawrence, Gulf of St. Lawrence",
  "n09299584": "Gulf of Sidra",
  "n09299727": "Gulf of Suez",
  "n09299885": "Gulf of Tehuantepec",
  "n09300030": "Gulf of Thailand, Gulf of Siam",
  "n09300199": "Gulf of Venice",
  "n09300306": "gully",
  "n09300483": "gut",
  "n09300559": "guyot",
  "n09300674": "hadron",
  "n09300905": "hail",
  "n09301044": "hairball, hair ball, trichobezoar",
  "n09301249": "Hampton Roads",
  "n09301461": "Handies Peak",
  "n09301625": "hanging wall",
  "n09301738": "Hangzhou Bay",
  "n09301844": "head",
  "n09302031": "head",
  "n09302127": "headstream",
  "n09302263": "Hercules",
  "n09302400": "heterocyclic ring, heterocycle",
  "n09302616": "highland, upland",
  "n09302804": "high sea, international waters",
  "n09303008": "hill",
  "n09303528": "hillside",
  "n09303647": "Himalayas, Himalaya Mountains, Himalaya",
  "n09304164": "Hindu Kush, Hindu Kush Mountains",
  "n09304376": "hogback, horseback",
  "n09304465": "hole",
  "n09304750": "hole, hollow",
  "n09305031": "hollow, holler",
  "n09305229": "holystone",
  "n09305358": "hood, cap",
  "n09305479": "Hook of Holland, Hoek van Holland",
  "n09305646": "horsepond",
  "n09305728": "horst",
  "n09305898": "hot spring, thermal spring",
  "n09306031": "Housatonic, Housatonic River",
  "n09306257": "Huang He, Hwang Ho, Yellow River",
  "n09306497": "Huascaran",
  "n09306642": "Hubbard, Mount Hubbard",
  "n09306840": "Hudson, Hudson River",
  "n09307031": "Hudson Bay",
  "n09307140": "Humber",
  "n09307300": "hunk, lump",
  "n09307552": "Hydra, Snake",
  "n09307719": "hydrogen ion",
  "n09307902": "hydrosphere",
  "n09308116": "Hydrus",
  "n09308242": "hyperon",
  "n09308398": "ice",
  "n09308572": "iceberg, berg",
  "n09308743": "icecap, ice cap",
  "n09308918": "icefall",
  "n09309046": "ice field",
  "n09309168": "ice floe, floe",
  "n09309292": "ice mass",
  "n09309456": "Iguazu, Iguazu Falls, Iguassu, Iguassu Falls, Victoria Falls",
  "n09309666": "IJssel, IJssel river",
  "n09309820": "IJsselmeer",
  "n09310011": "Illampu",
  "n09310162": "Illimani",
  "n09310314": "Illinois River",
  "n09310460": "impairer",
  "n09310616": "inclined fault",
  "n09310806": "inclusion body, cellular inclusion, inclusion",
  "n09311067": "index fossil, guide fossil",
  "n09311259": "Indian Ocean",
  "n09311710": "Indigirka, Indigirka River",
  "n09311885": "indirect transmission",
  "n09312094": "indumentum, indument",
  "n09312231": "Indus, Indus River",
  "n09312516": "Indus",
  "n09312645": "inessential, nonessential",
  "n09312843": "infectious agent, infective agent",
  "n09312999": "inferior planet",
  "n09313159": "ingrowth",
  "n09313241": "Inland Passage, Inside Passage",
  "n09313454": "Inland Sea",
  "n09313716": "inlet, recess",
  "n09314013": "inside track",
  "n09314128": "intermediate vector boson",
  "n09314263": "interplanetary dust",
  "n09314398": "interplanetary gas",
  "n09314603": "interplanetary medium",
  "n09314829": "interstellar medium",
  "n09314964": "intrusion",
  "n09315057": "Io",
  "n09315159": "ion",
  "n09315455": "Ionian Sea",
  "n09315624": "Irish Sea",
  "n09315762": "iron filing",
  "n09315870": "Irrawaddy, Irrawaddy River",
  "n09316066": "Irtish, Irtish River, Irtysh, Irtysh River",
  "n09316312": "Isere, Isere River",
  "n09316454": "island",
  "n09319456": "isle, islet",
  "n09319604": "isthmus",
  "n09319902": "Isthmus of Corinth",
  "n09320111": "Isthmus of Kra",
  "n09320239": "Isthmus of Panama, Isthmus of Darien",
  "n09320485": "Isthmus of Suez",
  "n09320611": "Isthmus of Tehuantepec",
  "n09320826": "jag",
  "n09320985": "James, James River",
  "n09321180": "James, James River",
  "n09321338": "James Bay",
  "n09321527": "Japan Trench",
  "n09321694": "Jebel Musa, Abila, Abyla",
  "n09321901": "Jordan, Jordan River",
  "n09322087": "Jovian planet, gas giant",
  "n09322349": "J particle, psi particle",
  "n09322454": "Jupiter",
  "n09322701": "K2, Godwin Austen, Mount Godwin Austen, Dapsang",
  "n09322930": "Kamet",
  "n09323085": "Kanawha, Kanawha River",
  "n09323221": "Kanchenjunga, Mount Kanchenjunga, Kanchanjanga, Kinchinjunga",
  "n09323470": "Kansas, Kansas River, Kaw River",
  "n09323660": "kaon, kappa-meson, k-meson, K particle",
  "n09323824": "Karakoram, Karakoram Range, Karakorum Range, Mustagh, Mustagh Range",
  "n09324118": "Kara Sea",
  "n09324322": "Karelian Isthmus",
  "n09324474": "Kasai, Kasai River, River Kasai",
  "n09324859": "Kattegatt",
  "n09325038": "kettle hole, kettle",
  "n09325246": "Keuka Lake, Lake Keuka",
  "n09325395": "key, cay, Florida key",
  "n09325530": "Khyber Pass",
  "n09325824": "kidney stone, urinary calculus, nephrolith, renal calculus",
  "n09325963": "Kilimanjaro, Mount Kilimanjaro",
  "n09326139": "Kissimmee, Kissimmee River",
  "n09326299": "Kivu, Lake Kivu",
  "n09326467": "Klamath, Klamath River",
  "n09326662": "knoll, mound, hillock, hummock, hammock",
  "n09326844": "Kodiak, Kodiak Island",
  "n09327077": "kopje, koppie",
  "n09327201": "Korea Bay",
  "n09327361": "Korean Strait, Korea Strait",
  "n09327538": "Kuiper belt, Edgeworth-Kuiper belt",
  "n09327705": "Kuiper belt object, KBO",
  "n09327881": "Kunlun, Kunlan Shan, Kunlun Mountains, Kuenlun, Kuenlun Mountains",
  "n09328148": "Kura, Kura River",
  "n09328311": "Labrador-Ungava Peninsula, Labrador Peninsula",
  "n09328596": "Labrador Sea",
  "n09328746": "lagoon, laguna, lagune",
  "n09328904": "lake",
  "n09330061": "Lake Albert, Lake Albert Nyanza, Mobuto Lake",
  "n09330233": "Lake Aral, Aral Sea",
  "n09330378": "lake bed, lake bottom",
  "n09330467": "Lake Chad, Chad",
  "n09330604": "Lake Champlain, Champlain",
  "n09330913": "Lake Edward",
  "n09331066": "Lake Erie, Erie",
  "n09331251": "lakefront",
  "n09331328": "Lake Geneva, Lake Leman",
  "n09331535": "Lake Huron, Huron",
  "n09331654": "Lake Ilmen, Ilmen",
  "n09331819": "Lake Ladoga, Ladoga",
  "n09332050": "Lake Michigan, Michigan",
  "n09332246": "Lake Nasser, Nasser",
  "n09332394": "Lake Nyasa, Lake Malawi",
  "n09332598": "Lake Onega, Onega",
  "n09332770": "Lake Ontario, Ontario",
  "n09332890": "lakeside, lakeshore",
  "n09332976": "Lake St. Clair, Lake Saint Clair",
  "n09333171": "Lake Superior, Superior",
  "n09333334": "Lake Tana, Lake Tsana",
  "n09333512": "Lake Tanganyika, Tanganyika",
  "n09333706": "Lake Urmia, Urmia, Daryacheh-ye Orumiyeh",
  "n09333905": "Lake Vanern, Vanern",
  "n09334049": "Lake Victoria, Victoria Nyanza",
  "n09334266": "lambda particle, lambda hyperon",
  "n09334396": "land, dry land, earth, ground, solid ground, terra firma",
  "n09335240": "land, ground, soil",
  "n09335693": "landfall",
  "n09335809": "landfill",
  "n09335916": "landmass, land mass",
  "n09336052": "Laptev Sea",
  "n09336271": "Large Magellanic Cloud",
  "n09336416": "Lascaux",
  "n09336555": "lather",
  "n09336702": "Laudo",
  "n09336853": "Laurasia",
  "n09337048": "leak",
  "n09337253": "ledge, shelf",
  "n09337406": "lees",
  "n09337531": "Lehigh River",
  "n09337686": "Lena, Lena River",
  "n09337833": "Leo",
  "n09338013": "lepton",
  "n09338243": "Lepus",
  "n09338361": "lethal agent",
  "n09338453": "Lethe, River Lethe",
  "n09338712": "Lhotse",
  "n09338910": "Liaodong Peninsula, Liaodong Bandao",
  "n09339109": "Libra",
  "n09339272": "Ligurian Sea",
  "n09339414": "liman",
  "n09339512": "Limpopo, Crocodile River",
  "n09339668": "liposomal delivery vector",
  "n09339810": "lithosphere, geosphere",
  "n09340024": "Little Bear, Ursa Minor",
  "n09340203": "Little Bighorn, Little Bighorn River, Little Horn",
  "n09340452": "Little Dipper, Dipper",
  "n09340644": "Little Missouri, Little Missouri River",
  "n09340935": "Little Sioux River",
  "n09341145": "Little Wabash, Little Wabash River",
  "n09341319": "llano",
  "n09341465": "Llano Estacado",
  "n09341673": "Llullaillaco",
  "n09341874": "loch",
  "n09341987": "loch",
  "n09342141": "Loch Achray",
  "n09342245": "Loch Linnhe",
  "n09342386": "Loch Ness",
  "n09342563": "lodestar, loadstar",
  "n09342729": "Logan, Mount Logan",
  "n09342937": "Loire, Loire River",
  "n09343123": "Loire Valley",
  "n09343266": "long chain, long-chain molecule",
  "n09343422": "Long Island Sound",
  "n09343587": "lough",
  "n09343660": "lough",
  "n09343761": "Lower California, Baja California",
  "n09343943": "lower mantle",
  "n09344050": "Lower Peninsula",
  "n09344198": "lowland",
  "n09344324": "lunar crater",
  "n09344449": "Lupus",
  "n09344559": "Lyra",
  "n09344724": "maar",
  "n09344863": "Mackenzie, Mackenzie River",
  "n09345002": "mackerel sky",
  "n09345127": "Madeira, Madeira River",
  "n09345265": "Magdalena, Magdalena River",
  "n09345503": "Magellanic Cloud",
  "n09345693": "magnetic dipole",
  "n09345792": "magnetic monopole",
  "n09345932": "main, briny",
  "n09346120": "mainland",
  "n09346284": "Makalu",
  "n09346450": "mantle",
  "n09346613": "mare, maria",
  "n09346735": "mare clausum",
  "n09346874": "mare liberum",
  "n09347008": "mare nostrum",
  "n09347116": "mare's tail",
  "n09347208": "Marmara, Sea of Marmara, Marmara Denizi, Marmora, Sea of Marmora",
  "n09347445": "Mars, Red Planet",
  "n09347779": "marsh, marshland, fen, fenland",
  "n09348055": "mass",
  "n09348236": "Massachusetts Bay",
  "n09348460": "massif",
  "n09348643": "Massif Central",
  "n09348827": "mat",
  "n09348949": "matchwood",
  "n09349055": "matrix",
  "n09349192": "Matterhorn",
  "n09349425": "McKinley, Mount McKinley, Mt. McKinley, Denali",
  "n09349648": "meander",
  "n09349797": "mechanism",
  "n09350045": "Mediterranean, Mediterranean Sea",
  "n09350524": "Mekong, Mekong River",
  "n09350776": "Menai Strait",
  "n09350922": "Mendenhall Glacier, Great Mendenhall Glacier",
  "n09351086": "Mensa",
  "n09351257": "Mercedario",
  "n09351408": "Mercury",
  "n09351547": "mere",
  "n09351647": "Merrimack, Merrimack River",
  "n09351905": "mesa, table",
  "n09352108": "Mesabi Range",
  "n09352282": "meson, mesotron",
  "n09352578": "mesosphere",
  "n09352721": "metal filing",
  "n09352849": "meteorite",
  "n09353109": "meteoroid, meteor",
  "n09353338": "meteor swarm",
  "n09353437": "Meuse, Meuse River",
  "n09353603": "micelle",
  "n09353815": "microfossil",
  "n09353918": "micrometeorite, micrometeoroid, micrometeor",
  "n09354142": "Microscopium",
  "n09354283": "Mid-Atlantic Ridge",
  "n09354511": "midstream",
  "n09354608": "mid-water",
  "n09354780": "Milk, Milk River",
  "n09354984": "Milky Way, Milky Way Galaxy, Milky Way System",
  "n09355261": "millpond",
  "n09355397": "Minamata Bay",
  "n09355623": "minor planet, planetoid",
  "n09355850": "mire, quagmire, quag, morass, slack",
  "n09356080": "Mississippi, Mississippi River",
  "n09356320": "Missouri, Missouri River",
  "n09356639": "Mobile, Mobile River",
  "n09356781": "Mobile Bay",
  "n09356929": "Mohawk River",
  "n09357080": "Mohorovicic discontinuity, Moho",
  "n09357346": "molehill",
  "n09357447": "monocline",
  "n09357580": "Monongahela, Monongahela River",
  "n09357847": "Mont Blanc, Monte Bianco",
  "n09358096": "Monterey Bay",
  "n09358226": "moon",
  "n09358358": "Moon, moon",
  "n09358550": "moon",
  "n09358751": "moor, moorland",
  "n09358907": "moraine",
  "n09359031": "Moray Firth",
  "n09359150": "Moreau River",
  "n09359312": "Moreton Bay",
  "n09359471": "morning star, daystar, Phosphorus, Lucifer",
  "n09359631": "motor",
  "n09359803": "mountain, mount",
  "n09360122": "mountain peak",
  "n09361517": "mountainside, versant",
  "n09361681": "Mount Bartle Frere",
  "n09361816": "Mount Carmel",
  "n09362050": "Mount Elbert",
  "n09362207": "mouse nest, mouse's nest",
  "n09362316": "mouth",
  "n09362469": "mouth",
  "n09362666": "Mozambique Channel",
  "n09362820": "mud puddle",
  "n09362945": "mull",
  "n09363064": "multiple star",
  "n09363214": "muon, negative muon, mu-meson",
  "n09363420": "Murray, Murray River",
  "n09363620": "Murrumbidgee, Murrumbidgee River",
  "n09363810": "Musca",
  "n09363970": "must",
  "n09364086": "mutagen",
  "n09364249": "Muztag, Muztagh",
  "n09364426": "Nacimiento",
  "n09364582": "nacreous cloud, mother-of-pearl cloud",
  "n09364778": "Namoi, Namoi River",
  "n09364954": "Nan, Nan River",
  "n09365128": "Nanda Devi",
  "n09365288": "Nanga Parbat",
  "n09365443": "Nan Ling",
  "n09365587": "Nares Deep",
  "n09365730": "Narragansett Bay",
  "n09365863": "narrow",
  "n09366017": "natural depression, depression",
  "n09366317": "natural elevation, elevation",
  "n09366597": "natural order",
  "n09366762": "nature",
  "n09366940": "nebula",
  "n09367137": "nebule",
  "n09367203": "necessity, essential, requirement, requisite, necessary",
  "n09367733": "neck",
  "n09367827": "Neckar, Neckar River",
  "n09367991": "need, want",
  "n09368224": "neighbor, neighbour",
  "n09368479": "Neosho, Neosho River",
  "n09368699": "Neptune",
  "n09369039": "neritic zone",
  "n09369169": "nest",
  "n09369412": "neutrino",
  "n09369520": "neutron",
  "n09369692": "neutron star",
  "n09369844": "Neva, Neva River",
  "n09370006": "neve",
  "n09370168": "New River",
  "n09370383": "New York Bay",
  "n09370552": "Niagara, Niagara River",
  "n09370773": "Niagara, Niagara Falls",
  "n09371028": "nidus",
  "n09371151": "Niger, Niger River",
  "n09371360": "Nile, Nile River",
  "n09371686": "nimbus, nimbus cloud, rain cloud",
  "n09371816": "Niobrara, Niobrara River",
  "n09371941": "nodule",
  "n09372176": "Norma",
  "n09372313": "normal fault, gravity fault, common fault",
  "n09372504": "North America",
  "n09373016": "North Atlantic",
  "n09373175": "North Channel",
  "n09373333": "Northern Cross",
  "n09373485": "North Pacific",
  "n09373623": "North Peak",
  "n09373716": "North Platte, North Platte River",
  "n09374036": "North Sea",
  "n09374306": "Norwegian Sea",
  "n09374493": "nova",
  "n09374646": "nub, stub",
  "n09374773": "nubbin",
  "n09374886": "nucleon",
  "n09375085": "nucleus",
  "n09375223": "nucleus",
  "n09375481": "nugget",
  "n09375606": "nullah",
  "n09375693": "Nuptse",
  "n09375891": "Ob, Ob River",
  "n09376078": "obliterator",
  "n09376198": "ocean",
  "n09376526": "ocean floor, sea floor, ocean bottom, seabed, sea bottom, Davy Jones's locker, Davy Jones",
  "n09376786": "oceanfront",
  "n09376866": "Octans",
  "n09376979": "Oder, Oder River",
  "n09377125": "offing",
  "n09377315": "Ohio, Ohio River",
  "n09377511": "oil-water interface",
  "n09377657": "Ojos del Salado",
  "n09377861": "Okeechobee, Lake Okeechobee",
  "n09378014": "Okefenokee Swamp",
  "n09378178": "Old Faithful",
  "n09378349": "Olduvai Gorge",
  "n09378529": "Olympus, Mount Olympus, Mt. Olympus, Olimbos",
  "n09378801": "Omega Centauri",
  "n09378929": "open chain",
  "n09379111": "opening, gap",
  "n09379572": "Ophiuchus",
  "n09379705": "Orange, Orange River",
  "n09379868": "ore bed",
  "n09379938": "Orinoco, Orinoco River",
  "n09380117": "Orion, Hunter",
  "n09380299": "Osage, Osage River",
  "n09380446": "Osaka Bay",
  "n09380588": "Outaouais, Ottawa, Ottawa river",
  "n09380817": "Ouachita, Ouachita River",
  "n09381048": "Ouse, Ouse River",
  "n09381242": "outcrop, outcropping, rock outcrop",
  "n09381480": "outer planet",
  "n09381781": "outthrust",
  "n09381880": "overburden",
  "n09382019": "oxbow",
  "n09382099": "oxbow",
  "n09382191": "oxbow lake",
  "n09382362": "Ozarks, Ozark Mountains, Ozark Plateau",
  "n09382570": "ozone hole",
  "n09382726": "ozone layer, ozonosphere",
  "n09382990": "Pacific, Pacific Ocean",
  "n09383793": "Pacific Coast",
  "n09383901": "pack ice, ice pack",
  "n09383998": "Pallas",
  "n09384106": "pallasite",
  "n09384223": "Pamir Mountains, the Pamirs",
  "n09384532": "Pangaea, Pangea",
  "n09384771": "Para, Para River",
  "n09384921": "Parana, Parana River",
  "n09385137": "paring, sliver, shaving",
  "n09385416": "Parnaiba, Parnahiba",
  "n09385586": "Parnassus, Mount Parnassus, Liakoura",
  "n09385911": "part, piece",
  "n09386422": "particle, subatomic particle",
  "n09386842": "pass, mountain pass, notch",
  "n09387222": "path, track, course",
  "n09387624": "Paulo Afonso, Paulo Afonso Falls",
  "n09387764": "Pavo",
  "n09387880": "Pearl River",
  "n09388023": "pebble",
  "n09388121": "Pecos, Pecos River",
  "n09388318": "Pee Dee, Pee Dee River",
  "n09388530": "Pegasus",
  "n09388653": "peneplain, peneplane",
  "n09388848": "peninsula",
  "n09389601": "Penobscot, Penobscot River",
  "n09389746": "Penobscot Bay",
  "n09389867": "perforation",
  "n09389999": "Perejil",
  "n09390146": "permafrost",
  "n09390236": "Perseus",
  "n09390424": "Persian Gulf, Arabian Gulf",
  "n09390680": "petrifaction",
  "n09390873": "Phobos",
  "n09390967": "Phoenix",
  "n09391089": "photoelectron",
  "n09391223": "photon",
  "n09391386": "photosphere",
  "n09391524": "Pictor",
  "n09391644": "piedmont",
  "n09391774": "Piedmont glacier, Piedmont type of glacier",
  "n09391996": "Pike's Peak",
  "n09392162": "Pillars of Hercules",
  "n09392402": "pinetum",
  "n09392507": "Ping, Ping River",
  "n09392656": "pion, pi-meson",
  "n09392824": "Pisces",
  "n09392961": "Pissis",
  "n09393108": "pit, cavity",
  "n09393403": "placer",
  "n09393524": "plage",
  "n09393605": "plain, field, champaign",
  "n09394007": "planet, major planet",
  "n09394646": "planet",
  "n09394797": "planetary nebula",
  "n09395086": "planetesimal",
  "n09395266": "plasmid, plasmid DNA",
  "n09395457": "plate, crustal plate",
  "n09395592": "Platte, Platte River",
  "n09395763": "Pleiades",
  "n09395899": "Pluto",
  "n09396150": "Po, Po River",
  "n09396275": "Pobeda Peak, Pobedy Peak",
  "n09396465": "point",
  "n09396608": "polar glacier",
  "n09396712": "Polaris, North Star, pole star, polar star, polestar",
  "n09397001": "polder",
  "n09397150": "Pollux",
  "n09397268": "polynya",
  "n09397391": "pond, pool",
  "n09397607": "pool, puddle",
  "n09397896": "positron, antielectron",
  "n09398076": "pothole, chuckhole",
  "n09398217": "Potomac, Potomac River",
  "n09398533": "Poyang",
  "n09398677": "precipice",
  "n09398769": "primary",
  "n09398935": "prion",
  "n09399198": "virino",
  "n09399485": "Procyon",
  "n09399592": "promontory, headland, head, foreland",
  "n09399899": "protein molecule",
  "n09400037": "proton",
  "n09400190": "Proxima, Proxima Centauri",
  "n09400337": "Prudhoe Bay",
  "n09400485": "pruning",
  "n09400584": "ptyalith",
  "n09400667": "Puget Sound",
  "n09400826": "pulp, mush",
  "n09400987": "pulsar",
  "n09401159": "Puppis",
  "n09401340": "Purus, Purus River",
  "n09401474": "Pyrenees",
  "n09401613": "Pyxis",
  "n09401750": "Quaoar",
  "n09401834": "quark",
  "n09402370": "quasar, quasi-stellar radio source",
  "n09402553": "Queen Charlotte Sound",
  "n09402704": "quickener, invigorator, enlivener",
  "n09402944": "quicksand",
  "n09403086": "rabbit burrow, rabbit hole",
  "n09403211": "radiator",
  "n09403333": "radio source",
  "n09403427": "rainbow",
  "n09403581": "Rakaposhi",
  "n09403734": "range, mountain range, range of mountains, chain, mountain chain, chain of mountains",
  "n09405078": "rangeland",
  "n09405169": "Ranier, Mount Ranier, Mt. Ranier, Mount Tacoma",
  "n09405396": "rapid",
  "n09405515": "Rappahannock, Rappahannock River",
  "n09405683": "rathole",
  "n09405787": "ravine",
  "n09405949": "Red, Red River",
  "n09406198": "red dwarf, red dwarf star",
  "n09406368": "red giant, red giant star",
  "n09406551": "Red Sea",
  "n09406793": "reef",
  "n09406944": "Regulus",
  "n09407043": "relaxer",
  "n09407194": "relict",
  "n09407346": "remains",
  "n09407541": "repressor, represser",
  "n09407632": "Republican, Republican River",
  "n09407867": "reservoir, source",
  "n09408105": "restriction fragment",
  "n09408250": "retardant, retardent, retardation",
  "n09408412": "Reticulum",
  "n09408540": "Rhine, Rhine River, Rhein",
  "n09408795": "Rhodope Mountains",
  "n09408977": "Rhone, Rhone River",
  "n09409203": "ribbon, thread",
  "n09409512": "ridge",
  "n09409752": "ridge, ridgeline",
  "n09409898": "ridge",
  "n09410026": "rift",
  "n09410101": "rift",
  "n09410224": "rift valley",
  "n09410365": "Rigel, Beta Orionis",
  "n09410460": "rill",
  "n09410558": "Rio de la Plata, La Plata, Plata River",
  "n09410724": "Rio Grande, Rio Bravo",
  "n09410928": "rip, rent, snag, split, tear",
  "n09411189": "riparian forest",
  "n09411295": "ripple mark",
  "n09411430": "river",
  "n09415584": "riverbank, riverside",
  "n09415671": "riverbed, river bottom",
  "n09415793": "river boulder",
  "n09415938": "rivulet, rill, run, runnel, streamlet",
  "n09416076": "rock, stone",
  "n09416570": "Rockies, Rocky Mountains",
  "n09416890": "roof",
  "n09417097": "round",
  "n09417240": "Ross Sea",
  "n09417365": "row",
  "n09417560": "Ruhr, Ruhr River",
  "n09417668": "Rushmore, Mount Rushmore, Mt. Rushmore",
  "n09417926": "Russell's body, cancer body",
  "n09418059": "Russian River",
  "n09418169": "Saale, Saale River",
  "n09418331": "Sabine, Sabine River",
  "n09418484": "Sacramento Mountains",
  "n09418629": "Sacramento River",
  "n09418810": "saddleback, saddle",
  "n09418954": "Sagitta",
  "n09419112": "Sagittarius",
  "n09419281": "Saint Francis, Saint Francis River, St. Francis, St. Francis River",
  "n09419536": "Saint John, Saint John River, St. John, St. John River",
  "n09419783": "Saint Johns, Saint Johns River, St. Johns, St. Johns River",
  "n09420030": "Saint Lawrence, Saint Lawrence River, St. Lawrence, St. Lawrence River",
  "n09420273": "Sajama",
  "n09420423": "Salmon, Salmon River",
  "n09420550": "salt flat, salt plain",
  "n09420689": "salt lick, lick",
  "n09420794": "salt marsh",
  "n09420911": "Salton Sea",
  "n09421031": "saltpan",
  "n09421191": "Sambre, Sambre River",
  "n09421425": "sample",
  "n09421604": "San Andreas Fault",
  "n09421799": "sandbank",
  "n09421951": "sandbar, sand bar",
  "n09422048": "San Diego Bay",
  "n09422190": "sandpit",
  "n09422294": "San Fernando Valley",
  "n09422486": "San Francisco Bay",
  "n09422631": "sanitary landfill",
  "n09422751": "San Joaquin River",
  "n09422964": "San Joaquin Valley",
  "n09423112": "San Juan Hill",
  "n09423379": "San Juan Mountains",
  "n09423617": "Sao Francisco",
  "n09423754": "Saone, Saone River",
  "n09423943": "Sargasso Sea",
  "n09424118": "Saronic Gulf, Gulf of Aegina",
  "n09424270": "satellite",
  "n09424489": "satisfier",
  "n09424642": "Saturn",
  "n09424865": "Savannah, Savannah River",
  "n09425019": "sawpit",
  "n09425159": "Sayan Mountains",
  "n09425344": "scablands",
  "n09425607": "scale, scurf, exfoliation",
  "n09425835": "Scheldt, Scheldt River",
  "n09426038": "scintilla",
  "n09426143": "Scorpius, Scorpio",
  "n09426307": "scraping",
  "n09426494": "Sculptor",
  "n09426621": "scurf",
  "n09426788": "sea",
  "n09427752": "seamount",
  "n09427876": "Sea of Azov, Sea of Azof, Sea of Azoff",
  "n09428036": "Sea of Japan, East Sea",
  "n09428171": "Sea of Okhotsk",
  "n09428293": "seashore, coast, seacoast, sea-coast",
  "n09428628": "seaside, seaboard",
  "n09428741": "section",
  "n09428967": "sediment, deposit",
  "n09429194": "Sedna",
  "n09429387": "segment",
  "n09429630": "seif dune",
  "n09429752": "Seine, Seine River",
  "n09429934": "Selkirk Mountains",
  "n09430100": "Seneca Lake, Lake Seneca",
  "n09430259": "Serpens",
  "n09430416": "Sete Quedas, Guaira, Guaira Falls",
  "n09430615": "seven seas",
  "n09430771": "Severn, River Severn, Severn River",
  "n09430990": "Severn, Severn River",
  "n09431133": "Seyhan, Seyhan River",
  "n09431283": "shag",
  "n09431409": "Shari, Shari River, Chari, Chari River",
  "n09431569": "Shasta, Mount Shasta",
  "n09431744": "Shenandoah River",
  "n09431902": "Sherman, Mount Sherman",
  "n09432060": "sheet",
  "n09432163": "shelf ice, ice shelf",
  "n09432283": "shell",
  "n09432430": "shell, eggshell",
  "n09432549": "Shenandoah Valley",
  "n09432785": "Sherwood Forest",
  "n09432990": "shiner",
  "n09433134": "shoal, shallow",
  "n09433312": "shoal",
  "n09433442": "shore",
  "n09433721": "shore boulder",
  "n09433839": "shoreline",
  "n09433952": "shortener",
  "n09434125": "sialolith, salivary calculus",
  "n09434237": "siderite",
  "n09434345": "sierra",
  "n09434469": "Sierra Madre Occidental",
  "n09434661": "Sierra Madre Oriental",
  "n09434845": "Sierra Nevada, Sierra Nevada Mountains, High Sierra",
  "n09435065": "Sierra Nevada",
  "n09435232": "sill",
  "n09435405": "silva, sylva",
  "n09435512": "Sinai, Mount Sinai",
  "n09435739": "sinkhole, sink, swallow hole",
  "n09435965": "Sirius, Dog Star, Canicula, Sothis",
  "n09436132": "Skagens Odde, Skaw",
  "n09436299": "Skagerrak, Skagerak",
  "n09436444": "ski slope",
  "n09436531": "skim",
  "n09436708": "sky",
  "n09436906": "slack, slack water",
  "n09437098": "slash",
  "n09437241": "slice",
  "n09437369": "slit",
  "n09437454": "slope, incline, side",
  "n09437887": "slot",
  "n09438055": "slough",
  "n09438212": "slough",
  "n09438313": "slough",
  "n09438408": "Small Magellanic Cloud",
  "n09438554": "Snake, Snake River",
  "n09438844": "snowcap",
  "n09438940": "snowdrift",
  "n09439032": "snowfield",
  "n09439120": "soap bubble",
  "n09439213": "soapsuds, suds, lather",
  "n09439433": "solar system",
  "n09439879": "Solent",
  "n09440036": "Solway Firth",
  "n09440186": "sound",
  "n09440400": "South America",
  "n09440948": "South Atlantic",
  "n09441107": "South China Sea",
  "n09441352": "Southern Cross, Crux, Crux Australis",
  "n09441587": "South Pacific",
  "n09441725": "South Platte, South Platte River",
  "n09441875": "South Sea",
  "n09442001": "South Sea Islands",
  "n09442146": "spall, spawl",
  "n09442341": "spark",
  "n09442496": "Spica",
  "n09442595": "spit, tongue",
  "n09442723": "splint",
  "n09442838": "splinter, sliver",
  "n09443136": "split",
  "n09443281": "spoor",
  "n09443453": "spring, fountain, outflow, outpouring, natural spring",
  "n09443641": "spume",
  "n09443753": "stalactite",
  "n09443921": "stalagmite",
  "n09444100": "star",
  "n09444783": "star",
  "n09444942": "starlet",
  "n09445008": "steep",
  "n09445088": "St. Elias Range, St. Elias Mountains",
  "n09445289": "steppe",
  "n09445435": "stepping stone",
  "n09445566": "steps",
  "n09445780": "Sterope, Asterope",
  "n09445910": "storm cloud",
  "n09446012": "straight chain",
  "n09446115": "strait, sound",
  "n09446617": "Strait of Georgia",
  "n09446747": "Strait of Gibraltar",
  "n09446868": "Strait of Hormuz, Strait of Ormuz",
  "n09447047": "Strait of Magellan",
  "n09447335": "Strait of Messina",
  "n09447450": "Strait of Dover, Strait of Calais, Pas de Calais",
  "n09447666": "strand",
  "n09447806": "strange particle",
  "n09447917": "strange quark, squark",
  "n09448090": "stratosphere",
  "n09448251": "stratus, stratus cloud",
  "n09448361": "stream, watercourse",
  "n09448690": "streambed, creek bed",
  "n09448829": "stressor",
  "n09448945": "stretch",
  "n09449127": "strike-slip fault",
  "n09449282": "string, cosmic string",
  "n09449510": "strip",
  "n09449657": "stub",
  "n09449773": "Styx, River Styx",
  "n09449949": "subcontinent",
  "n09450163": "sun, Sun",
  "n09450454": "sun",
  "n09450553": "Sun River",
  "n09450708": "supergiant",
  "n09450866": "superior planet",
  "n09451103": "supernatant",
  "n09451237": "supernova",
  "n09451378": "superstring",
  "n09451517": "surface, Earth's surface",
  "n09451864": "Suriname River, Surinam River",
  "n09452017": "Susquehanna, Susquehanna River",
  "n09452291": "swale",
  "n09452395": "swamp, swampland",
  "n09452653": "swath, belt",
  "n09452760": "swell",
  "n09452866": "swimming hole",
  "n09453008": "tableland, plateau",
  "n09453288": "Taconic Mountains",
  "n09453566": "Tagus, Tagus River",
  "n09453718": "Takakkaw",
  "n09453887": "Tallapoosa, Tallapoosa River",
  "n09454153": "talus, scree",
  "n09454265": "Tampa Bay",
  "n09454412": "tangle",
  "n09454642": "tarn",
  "n09454744": "tar pit",
  "n09454925": "tartar, calculus, tophus",
  "n09455046": "Tasman Sea",
  "n09455206": "tauon, tau-minus particle",
  "n09455306": "Taurus",
  "n09455524": "Telescopium",
  "n09455640": "Tennessee, Tennessee River",
  "n09455910": "tent",
  "n09455998": "teratogen",
  "n09456207": "terrace, bench",
  "n09456369": "terrestrial planet",
  "n09456614": "territorial waters",
  "n09456860": "Teton Range",
  "n09457020": "Thames, River Thames, Thames River",
  "n09457199": "thermion",
  "n09457367": "thermosphere",
  "n09457507": "thrust fault, overthrust fault, reverse fault",
  "n09457693": "thunderhead",
  "n09457851": "Tiber, Tevere",
  "n09457979": "tidal basin",
  "n09458079": "tidal river, tidewater river, tidal stream, tidewater stream",
  "n09458269": "tideland",
  "n09458372": "tidewater",
  "n09458494": "tideway",
  "n09458587": "Tien Shan, Tyan Shan",
  "n09458791": "Tigris, Tigris River",
  "n09458967": "Timor Sea",
  "n09459114": "Tirich Mir",
  "n09459269": "Titan",
  "n09459393": "Tocantins, Tocantins River",
  "n09459557": "Tombigbee, Tombigbee River",
  "n09459812": "top quark, truth quark",
  "n09459979": "tor",
  "n09460046": "tor",
  "n09460139": "Torres Strait",
  "n09460312": "trail",
  "n09460516": "transducing vector, gene delivery vector",
  "n09460680": "transmission mechanism",
  "n09460888": "Transylvanian Alps",
  "n09461069": "Trapezium",
  "n09461170": "tree farm",
  "n09461315": "trench, deep, oceanic abyss",
  "n09461515": "Trent, River Trent, Trent River",
  "n09461724": "Triangulum, Triangle",
  "n09461864": "Triangulum Australe, Southern Triangle",
  "n09462049": "Trinity River",
  "n09462231": "Triton",
  "n09462312": "Trondheim Fjord, Trondheim Fiord",
  "n09462452": "tropopause",
  "n09462600": "troposphere",
  "n09462773": "trough",
  "n09462924": "Tucana",
  "n09463078": "Tugela, Tugela Falls",
  "n09463226": "tundra",
  "n09463362": "Tunguska, Lower Tunguska",
  "n09463547": "Tunguska, Stony Tunguska",
  "n09463721": "Tupungato",
  "n09463919": "turf, sod, sward, greensward",
  "n09464098": "turning",
  "n09464221": "twilight zone",
  "n09464335": "Twin, Twin Falls",
  "n09464486": "twinkler",
  "n09464652": "Tyrolean Alps",
  "n09464805": "Tyne, River Tyne, Tyne River",
  "n09464962": "Tyrrhenian Sea",
  "n09465135": "Ulugh Muztagh, Ulugh Muz Tagh",
  "n09465290": "Uncompahgre Peak",
  "n09465459": "unit, building block",
  "n09465795": "unit cell",
  "n09465984": "United States waters, U.S. waters",
  "n09466280": "universe, existence, creation, world, cosmos, macrocosm",
  "n09466678": "uphill",
  "n09466757": "upper mantle",
  "n09466863": "Upper Peninsula",
  "n09467043": "up quark",
  "n09467185": "Urals, Ural Mountains",
  "n09467417": "Uranus",
  "n09467696": "urolith",
  "n09467765": "Urubupunga, Urubupunga Falls",
  "n09467921": "Uruguay River",
  "n09468237": "vagabond",
  "n09468447": "valence electron",
  "n09468604": "valley, vale",
  "n09468959": "variable",
  "n09469152": "variable star, variable",
  "n09469285": "vector, transmitter",
  "n09469744": "vector-borne transmission",
  "n09469912": "Vega",
  "n09470027": "vehicle-borne transmission",
  "n09470222": "vein, mineral vein",
  "n09470343": "Vela",
  "n09470550": "vent, volcano",
  "n09470762": "Venus",
  "n09471224": "Vesta",
  "n09471349": "vesture",
  "n09471481": "Vetluga, Vetluga River",
  "n09471638": "Victoria, Victoria Falls",
  "n09471865": "viral delivery vector",
  "n09471976": "Virgo",
  "n09472135": "Vistula, Vistula River",
  "n09472268": "Volans",
  "n09472413": "volcanic crater, crater",
  "n09472597": "volcano",
  "n09473239": "Volga, Volga River",
  "n09473397": "Volkhov, Volkhov River",
  "n09473558": "Volta",
  "n09473685": "Vulpecula",
  "n09473808": "Wabash, Wabash River",
  "n09474010": "wadi",
  "n09474162": "wall",
  "n09474412": "wall",
  "n09474557": "wallow",
  "n09474663": "wall rock",
  "n09474765": "warren, rabbit warren",
  "n09474895": "wash, dry wash",
  "n09475044": "wasp's nest, wasps' nest, hornet's nest, hornets' nest",
  "n09475179": "watercourse",
  "n09475292": "waterfall, falls",
  "n09475716": "water gap",
  "n09475827": "water hole",
  "n09475925": "waterside",
  "n09476011": "water system",
  "n09476123": "water table, water level, groundwater level",
  "n09476331": "waterway",
  "n09476521": "weakener",
  "n09476717": "weakly interacting massive particle, WIMP",
  "n09477037": "web",
  "n09477319": "webbing",
  "n09477427": "Weddell Sea",
  "n09477567": "Weisshorn",
  "n09477718": "Weser, Weser River",
  "n09477890": "wetland",
  "n09478047": "Wheeler Peak",
  "n09478210": "whinstone, whin",
  "n09478355": "White, White River",
  "n09478569": "white dwarf, white dwarf star",
  "n09478678": "White Nile",
  "n09478810": "White Sea",
  "n09478962": "white water, whitewater",
  "n09479072": "Whitney, Mount Whitney",
  "n09479238": "Wight, Isle of Wight",
  "n09479424": "Wilderness",
  "n09479635": "Willamette, Willamette River",
  "n09479811": "Wilson, Mount Wilson",
  "n09479962": "wind gap",
  "n09480077": "window",
  "n09480241": "Windward Passage",
  "n09480428": "Winnipeg, Lake Winnipeg",
  "n09480556": "Wisconsin, Wisconsin River",
  "n09480699": "wonderland",
  "n09480809": "world",
  "n09480959": "wormcast",
  "n09481036": "wormhole",
  "n09481120": "xenolith",
  "n09481285": "Yalu, Yalu River",
  "n09481523": "Chang Jiang, Changjiang, Chang, Yangtze, Yangtze River, Yangtze Kiang",
  "n09481754": "Yazoo, Yazoo River",
  "n09481958": "Yellow Sea, Huang Hai",
  "n09482131": "Yellowstone, Yellowstone River",
  "n09482330": "Yenisei, Yenisei River, Yenisey, Yenisey River",
  "n09482566": "Yerupaja",
  "n09482715": "Yosemite, Yosemite Falls",
  "n09482916": "Yukon, Yukon River",
  "n09483129": "Zambezi, Zambezi River",
  "n09483340": "Zhu Jiang, Canton River, Chu Kiang, Pearl River",
  "n09483519": "Zuider Zee",
  "n09483738": "imaginary being, imaginary creature",
  "n09484313": "hypothetical creature",
  "n09484465": "extraterrestrial being, extraterrestrial, alien",
  "n09484664": "mythical being",
  "n09486424": "Augeas",
  "n09486639": "Alcyone, Halcyon",
  "n09486781": "Arjuna",
  "n09487022": "legendary creature",
  "n09487252": "abominable snowman, yeti",
  "n09487385": "Bigfoot, Sasquatch",
  "n09487546": "Demogorgon",
  "n09487692": "doppelganger",
  "n09487821": "Loch Ness monster, Nessie",
  "n09488006": "sea serpent",
  "n09488117": "bogeyman, bugbear, bugaboo, boogeyman, booger",
  "n09488259": "Death",
  "n09488448": "Gargantua",
  "n09488584": "Grim Reaper, Reaper",
  "n09488711": "giant",
  "n09488995": "hobbit",
  "n09489146": "Maxwell's demon",
  "n09489601": "mermaid",
  "n09489697": "merman",
  "n09489790": "Martian",
  "n09489889": "Argus",
  "n09490054": "Cadmus",
  "n09490210": "Calypso",
  "n09490352": "sea nymph",
  "n09490572": "Cyclops",
  "n09490756": "giantess",
  "n09490825": "ogre",
  "n09490961": "ogress",
  "n09491038": "Humpty Dumpty",
  "n09491222": "Jack Frost",
  "n09491324": "Mammon",
  "n09491504": "Scylla",
  "n09491784": "Stentor",
  "n09491966": "monster",
  "n09492123": "mythical monster, mythical creature",
  "n09492733": "amphisbaena",
  "n09492877": "basilisk",
  "n09493043": "centaur",
  "n09493204": "Cerberus, hellhound",
  "n09493374": "Charon",
  "n09493562": "Chimera, Chimaera",
  "n09493807": "Chiron",
  "n09493983": "Circe",
  "n09494149": "cockatrice",
  "n09494280": "Dardanus",
  "n09494388": "dragon, firedrake",
  "n09494609": "Fafnir",
  "n09494764": "Ganymede",
  "n09494951": "Geryon",
  "n09495103": "Gorgon",
  "n09495382": "Grace",
  "n09495619": "Aglaia",
  "n09495732": "Euphrosyne",
  "n09495849": "Thalia",
  "n09495962": "gryphon, griffin, griffon",
  "n09496099": "Harpy",
  "n09496261": "Hydra",
  "n09496460": "Hyperborean",
  "n09496673": "Hypnos",
  "n09496802": "leviathan",
  "n09496918": "Niobe",
  "n09497163": "Perseus",
  "n09497364": "Andromeda",
  "n09497641": "Cepheus",
  "n09497775": "Cassiopeia",
  "n09497913": "Medusa",
  "n09498072": "Stheno",
  "n09498186": "Euryale",
  "n09498301": "manticore, mantichora, manticora, mantiger",
  "n09498497": "Midas",
  "n09498697": "Sisyphus",
  "n09499056": "Minotaur",
  "n09499230": "Morpheus",
  "n09499338": "Narcissus",
  "n09499494": "Nemean lion",
  "n09499657": "Nibelung",
  "n09499854": "Nibelung",
  "n09500003": "Bellerophon",
  "n09500217": "Paris",
  "n09500398": "Patroclus",
  "n09500625": "Pegasus",
  "n09500936": "phoenix",
  "n09501198": "Python",
  "n09501322": "roc",
  "n09501427": "salamander",
  "n09501527": "Sarpedon",
  "n09501737": "Siegfried",
  "n09502080": "Sigurd",
  "n09502298": "Sphinx",
  "n09502481": "troll",
  "n09502700": "Typhoeus",
  "n09502906": "Typhon",
  "n09503121": "werewolf, wolfman, lycanthrope, loup-garou",
  "n09503282": "witch",
  "n09503486": "wyvern, wivern",
  "n09503682": "nature",
  "n09503877": "supernatural, occult",
  "n09504135": "spiritual being, supernatural being",
  "n09504464": "theurgy",
  "n09504603": "first cause, prime mover, primum mobile",
  "n09504790": "control",
  "n09504915": "destiny, fate",
  "n09505153": "spiritual leader",
  "n09505418": "deity, divinity, god, immortal",
  "n09506216": "daemon, demigod",
  "n09506337": "Fury, Eumenides, Erinyes",
  "n09506598": "Alecto",
  "n09506674": "Megaera",
  "n09506751": "Tisiphone",
  "n09506830": "sea god",
  "n09506973": "sun god",
  "n09507097": "Celtic deity",
  "n09507639": "Amaethon",
  "n09507756": "Ana",
  "n09507909": "Angus Og, Aengus, Oengus, Angus",
  "n09508076": "Arawn",
  "n09508228": "Arianrhod, Arianrod",
  "n09508368": "Boann",
  "n09508474": "Brigit",
  "n09508666": "Dagda",
  "n09508809": "Danu, Dana",
  "n09508975": "Don",
  "n09509119": "Dylan",
  "n09509232": "Epona",
  "n09509375": "Fomor, Fomorian",
  "n09509546": "Gwydion",
  "n09509678": "Gwyn",
  "n09509770": "Lir, Ler",
  "n09509911": "Llew Llaw Gyffes",
  "n09510073": "LLud",
  "n09510164": "Llyr",
  "n09510305": "Lug, Lugh",
  "n09510419": "Manannan",
  "n09510527": "Manawydan, Manawyddan",
  "n09510643": "Morrigan, Morrigu",
  "n09510746": "Tuatha De Danann, Tuatha De",
  "n09510904": "Egyptian deity",
  "n09511368": "Amen, Amon, Amun",
  "n09511519": "Amen-Ra, Amon-Ra, Amun Ra",
  "n09511712": "Anubis, Anpu",
  "n09511875": "Aten, Aton",
  "n09512028": "Bast",
  "n09512157": "Geb, Keb",
  "n09512266": "Horus",
  "n09512391": "Isis",
  "n09512516": "Khepera",
  "n09512611": "Min",
  "n09512692": "Nephthys",
  "n09512835": "Nut",
  "n09512913": "Osiris",
  "n09513065": "Ptah",
  "n09513216": "Ra, Re",
  "n09513430": "Sekhet, Eye of Ra",
  "n09513569": "Set, Seth",
  "n09513744": "Thoth",
  "n09513902": "Semitic deity",
  "n09515027": "Adad",
  "n09515131": "Adapa",
  "n09515269": "Anshar",
  "n09515473": "Antum",
  "n09515570": "Anu",
  "n09515712": "Anunnaki, Enuki",
  "n09515877": "Apsu",
  "n09515988": "Aruru",
  "n09516146": "Ashur, Ashir",
  "n09516323": "Astarte, Ashtoreth",
  "n09516498": "Ishtar, Mylitta",
  "n09516700": "Baal",
  "n09516881": "Bel",
  "n09517057": "Dagon",
  "n09517209": "Dagan",
  "n09517342": "Damkina, Damgalnunna",
  "n09517492": "Dumuzi, Tammuz",
  "n09517628": "Ea",
  "n09517890": "Enki",
  "n09518020": "Enlil, En-lil",
  "n09518145": "Ereshkigal, Eresh-kigal, Ereshkigel",
  "n09518306": "Girru",
  "n09518451": "Gula",
  "n09518578": "Igigi",
  "n09518683": "Inanna",
  "n09518782": "Ki",
  "n09518908": "Kishar",
  "n09519075": "Lilith",
  "n09519192": "Mama",
  "n09519288": "Marduk, Merodach, Baal Merodach, Bel-Merodach",
  "n09519454": "Moloch, Molech",
  "n09519596": "Nabu, Nebo",
  "n09519748": "Nammu",
  "n09519901": "Namtar, Namtaru",
  "n09520103": "Nanna",
  "n09520223": "Nergal",
  "n09520367": "Nina",
  "n09520498": "Ningal",
  "n09520617": "Ningirsu",
  "n09520752": "Ningishzida",
  "n09520880": "Ninkhursag, Ninhursag, Ninkharsag",
  "n09521051": "Nintu, Nintoo",
  "n09521157": "Ninurta, Ninib",
  "n09521395": "Nusku",
  "n09521521": "Ramman",
  "n09521648": "Sarpanitu, Zirbanit, Zarpanit",
  "n09521763": "Shamash",
  "n09521994": "Sin",
  "n09522121": "Tashmit, Tashmitum",
  "n09522222": "Tiamat",
  "n09522344": "Utnapishtim",
  "n09522498": "Utu, Utug",
  "n09522615": "Zu, Zubird",
  "n09522735": "Enkidu",
  "n09522837": "Gilgamish",
  "n09522978": "Hindu deity",
  "n09524027": "Aditi",
  "n09524151": "Aditya",
  "n09524263": "Agni",
  "n09524440": "Asura",
  "n09524555": "Ahura",
  "n09524671": "Asvins",
  "n09524840": "Bhaga",
  "n09524921": "Brahma",
  "n09525065": "Brihaspati",
  "n09525170": "Bhumi Devi",
  "n09525279": "Devi",
  "n09525486": "Chandi",
  "n09525579": "Dharma",
  "n09525746": "Durga",
  "n09525874": "Dyaus, Dyaus-pitar",
  "n09525961": "Ganesh, Ganesa, Ganesha, Ganapati",
  "n09526107": "Garuda",
  "n09526221": "Gauri",
  "n09526367": "Hanuman",
  "n09526497": "Indra",
  "n09526606": "Ka",
  "n09526703": "Kali",
  "n09526808": "Kama",
  "n09526913": "Mara",
  "n09527001": "Kartikeya, Karttikeya",
  "n09527091": "Lakshmi",
  "n09527185": "Marut",
  "n09527291": "Mitra",
  "n09527454": "Parjanya",
  "n09527560": "Parvati, Anapurna, Annapurna",
  "n09527707": "Prajapati",
  "n09527825": "Praxiteles",
  "n09527924": "Pushan",
  "n09528032": "Rahu",
  "n09528135": "Ribhus, Rhibhus",
  "n09528237": "Rudra",
  "n09528380": "Sarasvati",
  "n09528475": "Savitar",
  "n09528587": "Shakti, Sakti",
  "n09528727": "Siva, Shiva",
  "n09528919": "Bairava",
  "n09529013": "Skanda",
  "n09529083": "Soma",
  "n09529201": "Surya",
  "n09529364": "Uma",
  "n09529454": "Ushas",
  "n09529573": "Vajra",
  "n09529644": "Varuna",
  "n09529933": "Vayu",
  "n09529999": "Vishnu",
  "n09530141": "Yama",
  "n09530238": "avatar",
  "n09530515": "Jagannath, Jagannatha, Jagganath, Juggernaut",
  "n09530629": "Kalki",
  "n09530721": "Krishna",
  "n09530867": "Rama",
  "n09531186": "Ramachandra",
  "n09531296": "Sita",
  "n09531409": "Balarama",
  "n09531515": "Parashurama",
  "n09531630": "Persian deity",
  "n09531826": "Mithras, Mithra",
  "n09531955": "Ormazd, Ormuzd, Ahura Mazda",
  "n09532101": "Ahriman",
  "n09532214": "Buddha, Siddhartha, Gautama, Gautama Siddhartha, Gautama Buddha",
  "n09532384": "Bodhisattva, Boddhisatva",
  "n09532553": "Maitreya",
  "n09532691": "Avalokitesvara, Avalokiteshvara",
  "n09532837": "Arhat, Arhant, lohan",
  "n09532942": "Buddha",
  "n09533048": "Chinese deity",
  "n09533301": "Chang Kuo, Chang Kuo-lao",
  "n09533406": "Wen Ch'ang, Wen-Ti",
  "n09533498": "Taoist Trinity",
  "n09533668": "Tien-pao, Heavenly Jewel",
  "n09533790": "Ling-pao, Mystic Jewel",
  "n09533910": "Shen-pao, Spiritual Jewel",
  "n09534058": "Chuang-tzu",
  "n09534187": "Kwan-yin, Kuan Yin",
  "n09534428": "Japanese deity",
  "n09534675": "Amaterasu, Amaterasu Omikami",
  "n09534843": "Hachiman",
  "n09534918": "Hotei",
  "n09535001": "Izanagi",
  "n09535130": "Izanami",
  "n09535255": "Kami",
  "n09535409": "Kwannon",
  "n09535504": "Ninigi, Ninigino-Mikoto",
  "n09535622": "goddess",
  "n09535708": "earth-god, earth god",
  "n09535809": "earth-goddess, earth goddess",
  "n09535940": "earth mother",
  "n09536058": "God, Supreme Being",
  "n09536363": "Godhead, Lord, Creator, Maker, Divine, God Almighty, Almighty, Jehovah",
  "n09536584": "eon, aeon",
  "n09536789": "Trinity, Holy Trinity, Blessed Trinity, Sacred Trinity",
  "n09536973": "Father, Father-God, Fatherhood",
  "n09537144": "Son, Word, Logos",
  "n09537325": "Messiah",
  "n09537502": "Messiah",
  "n09537660": "messiah, christ",
  "n09537781": "Holy Ghost, Holy Spirit, Paraclete",
  "n09538021": "hypostasis, hypostasis of Christ",
  "n09538318": "Yahweh, YHWH, Yahwe, Yahveh, YHVH, Yahve, Wahvey, Jahvey, Jahweh, Jehovah, JHVH",
  "n09538544": "Allah",
  "n09538633": "demiurge",
  "n09538757": "faun",
  "n09538915": "angel",
  "n09539183": "archangel",
  "n09539394": "Gabriel",
  "n09539517": "Michael",
  "n09539640": "Raphael",
  "n09539730": "cherub",
  "n09539872": "seraph",
  "n09540055": "guardian spirit, guardian angel",
  "n09540241": "genius loci",
  "n09540329": "divine messenger",
  "n09540430": "fairy, faery, faerie, fay, sprite",
  "n09540739": "elf, hob, gremlin, pixie, pixy, brownie, imp",
  "n09540939": "fairy godmother",
  "n09541125": "gnome, dwarf",
  "n09541346": "undine",
  "n09541434": "leprechaun",
  "n09541526": "sandman",
  "n09541661": "Morgan le Fay",
  "n09541809": "Puck, Robin Goodfellow",
  "n09541919": "evil spirit",
  "n09542170": "bad fairy",
  "n09542258": "bogey, bogy, bogie",
  "n09542339": "devil, fiend, demon, daemon, daimon",
  "n09542541": "cacodemon, cacodaemon",
  "n09542697": "eudemon, eudaemon, good spirit",
  "n09542868": "incubus",
  "n09543021": "succubus, succuba",
  "n09543154": "dybbuk, dibbuk",
  "n09543353": "Satan, Old Nick, Devil, Lucifer, Beelzebub, the Tempter, Prince of Darkness",
  "n09543673": "ghoul",
  "n09543748": "goblin, hob, hobgoblin",
  "n09543924": "kelpy, kelpie",
  "n09544109": "vampire, lamia",
  "n09544262": "banshee, banshie",
  "n09544433": "genie, jinni, jinnee, djinni, djinny, djinn",
  "n09544746": "shaitan, shaytan",
  "n09544876": "eblis",
  "n09545000": "houri",
  "n09545171": "familiar, familiar spirit",
  "n09545324": "spirit, disembodied spirit",
  "n09545755": "trickster",
  "n09545976": "ghost",
  "n09546145": "poltergeist",
  "n09546280": "Oberson",
  "n09546453": "Titania",
  "n09546604": "tooth fairy",
  "n09546772": "water sprite, water nymph, water spirit",
  "n09546905": "peri",
  "n09547111": "apparition, phantom, phantasm, phantasma, fantasm, specter, spectre",
  "n09547353": "Flying Dutchman",
  "n09547528": "presence",
  "n09547629": "Adonis",
  "n09547903": "Greco-Roman deity, Graeco-Roman deity",
  "n09548111": "satyr, forest god",
  "n09548316": "Silenus",
  "n09548489": "silenus",
  "n09548632": "nymph",
  "n09549047": "Echo",
  "n09549216": "Hesperides, Atlantides",
  "n09549416": "Hyades",
  "n09549643": "Oread",
  "n09549758": "Pleiades",
  "n09549983": "Sterope, Asterope",
  "n09550125": "water nymph",
  "n09550278": "Daphne",
  "n09550443": "naiad",
  "n09550585": "Nereid",
  "n09550758": "Thetis",
  "n09550899": "Oceanid",
  "n09551040": "dryad, wood nymph",
  "n09551152": "Salmacis",
  "n09551260": "hamadryad",
  "n09551356": "Greek deity",
  "n09552681": "Roman deity",
  "n09553347": "Olympian, Olympic god",
  "n09553490": "Aeolus",
  "n09553599": "Aether",
  "n09553767": "Apollo, Phoebus, Phoebus Apollo",
  "n09554019": "Pythius",
  "n09554132": "Aphrodite, Cytherea",
  "n09554297": "Hero",
  "n09554505": "Leander",
  "n09554701": "Pygmalion",
  "n09554910": "Galatea",
  "n09555127": "Venus, Urania",
  "n09555236": "Ares",
  "n09555391": "Eris",
  "n09555514": "Thanatos",
  "n09555653": "Mors",
  "n09555785": "Mars",
  "n09555962": "Nyx",
  "n09556121": "Rhea Silvia, Rea Silvia",
  "n09556305": "Romulus",
  "n09556580": "Remus",
  "n09556697": "Artemis, Cynthia",
  "n09556915": "Boreas",
  "n09557130": "Diana",
  "n09557289": "Ate",
  "n09557387": "Athena, Athene, Pallas, Pallas Athena, Pallas Athene",
  "n09557632": "Minerva",
  "n09557770": "Chaos",
  "n09557965": "Cronus",
  "n09558177": "Dido",
  "n09558388": "Saturn",
  "n09558569": "Demeter",
  "n09558756": "Ceres",
  "n09558898": "Dionysus",
  "n09559071": "Doris",
  "n09559201": "Aesculapius, Asclepius, Asklepios",
  "n09559404": "Bacchus",
  "n09559573": "Erebus",
  "n09559769": "Nox, Night",
  "n09559896": "Eros",
  "n09560061": "Cupid, Amor",
  "n09560196": "Daedalus, Daedal",
  "n09560426": "Damon and Pythias",
  "n09560762": "Gaea, Gaia, Ge",
  "n09560940": "Hebe",
  "n09561137": "Helios",
  "n09561320": "Icarus",
  "n09561593": "Sol",
  "n09561755": "Hecate",
  "n09561970": "Hephaestus, Hephaistos",
  "n09562166": "Vulcan",
  "n09562319": "Hermes",
  "n09562526": "Hermaphroditus",
  "n09562704": "Mercury",
  "n09562883": "Hygeia",
  "n09563042": "Panacea",
  "n09563183": "Hera, Here",
  "n09563425": "Io",
  "n09563619": "Janus",
  "n09563805": "Juno",
  "n09563998": "Hestia",
  "n09564177": "Vesta",
  "n09564371": "Hymen",
  "n09564480": "Hyperion",
  "n09564680": "Minos",
  "n09564889": "Ariadne",
  "n09565099": "Moirai, Moirae",
  "n09565334": "Parcae",
  "n09565503": "Clotho, Klotho",
  "n09565639": "Lachesis",
  "n09565787": "Atropos",
  "n09565914": "Momus, Momos",
  "n09565999": "Muse",
  "n09566320": "Calliope",
  "n09566436": "Clio",
  "n09566544": "Erato",
  "n09566667": "Euterpe",
  "n09566791": "Melpomene",
  "n09566904": "Polyhymnia",
  "n09567044": "Terpsichore",
  "n09567180": "Thalia",
  "n09567309": "Urania",
  "n09567421": "Nemesis",
  "n09567560": "Nereus",
  "n09567767": "Nike",
  "n09567913": "Victoria",
  "n09568051": "Ouranos, Uranus",
  "n09568241": "Pan, goat god",
  "n09568488": "Faunus",
  "n09568643": "Pasiphae",
  "n09568776": "Pontus, Pontos",
  "n09568927": "Poseidon",
  "n09569142": "Proteus",
  "n09569330": "Neptune",
  "n09569467": "Persephone, Despoina, Kore, Cora",
  "n09569709": "Procrustes",
  "n09569985": "Proserpina, Proserpine",
  "n09570114": "Phaethon",
  "n09570298": "Pluto, Hades, Aides, Aidoneus",
  "n09570522": "Dis, Orcus",
  "n09570630": "Pythia, Pythoness",
  "n09570794": "Priapus",
  "n09570957": "Rhadamanthus",
  "n09571090": "Selene",
  "n09571249": "Luna",
  "n09571390": "Eos",
  "n09571581": "Eurydice",
  "n09571693": "Orion",
  "n09571914": "Orpheus",
  "n09572113": "Aurora",
  "n09572249": "Tellus",
  "n09572425": "Titan",
  "n09572825": "Titaness",
  "n09573145": "Triton",
  "n09573262": "Tyche",
  "n09573405": "Fortuna",
  "n09573561": "Zephyr",
  "n09573682": "Zeus",
  "n09573966": "Jupiter, Jove",
  "n09574226": "Jupiter Fulgur, Jupiter Fulminator, Lightning Hurler",
  "n09574350": "Jupiter Tonans, Thunderer",
  "n09574446": "Jupiter Pluvius, Rain-giver",
  "n09574544": "Jupiter Optimus Maximus, Best and Greatest",
  "n09574657": "Jupiter Fidius, Protector of Boundaries",
  "n09574767": "Oceanus",
  "n09574926": "Cocus",
  "n09575033": "Crius",
  "n09575140": "Iapetus",
  "n09575316": "vestal virgin",
  "n09575548": "Atlas",
  "n09575701": "Epimetheus",
  "n09575902": "Prometheus",
  "n09576164": "Thea, Theia",
  "n09576338": "Rhea",
  "n09576548": "Ops",
  "n09576746": "Sylvanus, Silvanus",
  "n09576914": "Agdistis",
  "n09577004": "Themis",
  "n09577155": "Mnemosyne",
  "n09577308": "Phoebe",
  "n09577467": "Tethys",
  "n09577601": "Psyche",
  "n09577833": "Leto, Latona",
  "n09578005": "Hercules, Heracles, Herakles, Alcides",
  "n09578248": "Pandora",
  "n09578465": "Norse deity",
  "n09579070": "Aesir",
  "n09579198": "Andvari",
  "n09579345": "Vanir",
  "n09579491": "Balder, Baldr",
  "n09579714": "Bragi, Brage",
  "n09579848": "Elli",
  "n09579994": "Forseti",
  "n09580125": "Frey, Freyr",
  "n09580354": "Freya, Freyja",
  "n09580521": "Frigg, Frigga",
  "n09580673": "Heimdall, Heimdal, Heimdallr",
  "n09580829": "Hel, Hela",
  "n09580971": "Hoenir",
  "n09581129": "Hoth, Hothr, Hoder, Hodr, Hodur",
  "n09581338": "Idun, Ithunn",
  "n09581526": "Jotun, Jotunn",
  "n09581680": "Loki",
  "n09581859": "Mimir",
  "n09582019": "Nanna",
  "n09582122": "Njord, Njorth",
  "n09582343": "Norn, weird sister",
  "n09582616": "Urd, Urth",
  "n09582726": "Verdandi, Verthandi",
  "n09582845": "Skuld",
  "n09582949": "Odin",
  "n09583196": "Sif",
  "n09583321": "Sigyn",
  "n09583506": "Thor",
  "n09583723": "Tyr, Tyrr",
  "n09583888": "Ull, Ullr",
  "n09584079": "Vali",
  "n09584218": "Vitharr, Vithar, Vidar",
  "n09584405": "Fenrir",
  "n09584559": "Volund",
  "n09584725": "Yggdrasil, Ygdrasil",
  "n09584933": "Ymir",
  "n09585218": "Wayland, Wayland the Smith, Wieland",
  "n09585434": "Teutonic deity",
  "n09585650": "Donar",
  "n09585757": "Nerthus, Hertha",
  "n09585889": "Wotan",
  "n09586011": "Anglo-Saxon deity",
  "n09586231": "Tiu",
  "n09586326": "Woden, Wodan",
  "n09586442": "Wyrd, Weird",
  "n09586553": "Adam",
  "n09586743": "Eve",
  "n09586994": "Cain",
  "n09587217": "Abel",
  "n09587409": "Seth",
  "n09587565": "fictional character, fictitious character, character",
  "n09589323": "Ajax",
  "n09589444": "Aladdin",
  "n09589585": "Argonaut",
  "n09589748": "Babar",
  "n09589876": "Beatrice",
  "n09589996": "Beowulf",
  "n09590205": "Bluebeard",
  "n09590377": "Bond, James Bond",
  "n09590495": "Brunhild, Brunnhilde, Brynhild",
  "n09590772": "Valkyrie",
  "n09590980": "Brer Rabbit",
  "n09591155": "Bunyan, Paul Bunyan",
  "n09591450": "John Henry",
  "n09591676": "Cheshire cat",
  "n09591814": "Chicken Little",
  "n09591973": "Cinderella",
  "n09592147": "Colonel Blimp",
  "n09592272": "Dracula",
  "n09592384": "Jason",
  "n09592563": "Medea",
  "n09592734": "Laertes",
  "n09592848": "Odysseus",
  "n09593044": "Ulysses",
  "n09593144": "Penelope",
  "n09593416": "Theseus",
  "n09593651": "Tantalus",
  "n09593937": "Phrygian deity",
  "n09594093": "Cybele, Dindymene, Great Mother, Magna Mater, Mater Turrita",
  "n09594298": "Achilles",
  "n09594604": "Aeneas",
  "n09594756": "Atreus",
  "n09594906": "Agamemnon",
  "n09595059": "Menelaus",
  "n09595240": "Iphigenia",
  "n09595545": "Clytemnestra",
  "n09595717": "Aegisthus",
  "n09595957": "Orestes",
  "n09596186": "Cassandra",
  "n09596374": "Antigone",
  "n09596547": "Creon",
  "n09596733": "Jocasta",
  "n09596884": "Electra",
  "n09597141": "Laocoon",
  "n09597478": "Laius",
  "n09597628": "Myrmidon",
  "n09597798": "Oedipus, King Oedipus, Oedipus Rex",
  "n09598066": "Tiresias",
  "n09598270": "Peleus",
  "n09598370": "Don Quixote",
  "n09598491": "El Cid",
  "n09598598": "Fagin",
  "n09598750": "Falstaff, Sir John Falstaff",
  "n09598888": "Father Brown",
  "n09599023": "Faust, Faustus",
  "n09599195": "Frankenstein",
  "n09599401": "Frankenstein, Frankenstein's monster",
  "n09599633": "Goofy",
  "n09599728": "Gulliver",
  "n09599891": "Hamlet",
  "n09600031": "Hector",
  "n09600188": "Helen, Helen of Troy",
  "n09600439": "Horatio Hornblower, Captain Horatio Hornblower",
  "n09600624": "Iago",
  "n09600764": "Inspector Maigret, Commissaire Maigret",
  "n09600901": "Kilroy",
  "n09601049": "Lear, King Lear",
  "n09601219": "Leda",
  "n09601429": "Lilliputian",
  "n09601571": "Marlowe, Philip Marlowe",
  "n09601769": "Mephistopheles",
  "n09601906": "Micawber, Wilkins Micawber",
  "n09602048": "Mother Goose",
  "n09602162": "Mr. Moto",
  "n09602258": "Othello",
  "n09602383": "Pangloss",
  "n09602484": "Pantaloon",
  "n09602610": "Pantaloon",
  "n09602716": "Perry Mason",
  "n09602828": "Peter Pan",
  "n09602964": "Pied Piper, Pied Piper of Hamelin",
  "n09603120": "Pierrot",
  "n09603258": "Pluto",
  "n09603353": "Huckleberry Finn, Huck Finn",
  "n09603471": "Rip van Winkle",
  "n09603667": "Ruritanian",
  "n09603761": "Tarzan, Tarzan of the Apes",
  "n09603919": "Tom Sawyer",
  "n09604014": "Uncle Remus",
  "n09604227": "Uncle Tom",
  "n09604345": "Uncle Sam",
  "n09604451": "Sherlock Holmes, Holmes",
  "n09604574": "Simon Legree",
  "n09604706": "Sinbad the Sailor, Sinbad",
  "n09604868": "Snoopy",
  "n09604981": "self",
  "n09605110": "number one",
  "n09605289": "adult, grownup",
  "n09606009": "adventurer, venturer",
  "n09606380": "anomalist",
  "n09606527": "anomaly, unusual person",
  "n09606873": "anachronism",
  "n09607055": "Ananias",
  "n09607208": "apache",
  "n09607280": "applicant, applier",
  "n09607630": "appointee, appointment",
  "n09607782": "argonaut",
  "n09607903": "Ashkenazi",
  "n09608002": "attendant, attender, attendee, meeter",
  "n09608377": "auctioneer",
  "n09608520": "behaviorist, behaviourist",
  "n09608709": "benefactor, helper",
  "n09609155": "benefactress",
  "n09609232": "capitalist",
  "n09609561": "captor, capturer",
  "n09609746": "caster",
  "n09609871": "changer, modifier",
  "n09610093": "coadjutor",
  "n09610173": "cofounder",
  "n09610255": "color-blind person",
  "n09610405": "commoner, common man, common person",
  "n09610660": "communicator",
  "n09611548": "Conservative Jew",
  "n09611722": "conservator",
  "n09611884": "constituent",
  "n09612131": "contestee",
  "n09612291": "contester",
  "n09612447": "Contra",
  "n09612580": "contrapuntist",
  "n09612700": "contrarian",
  "n09612848": "consumer",
  "n09613118": "contadino",
  "n09613191": "contestant",
  "n09613559": "coon",
  "n09613690": "cosigner, cosignatory",
  "n09613853": "cosigner",
  "n09614047": "coward",
  "n09614315": "creator",
  "n09614684": "defender, guardian, protector, shielder",
  "n09615211": "defender, withstander",
  "n09615336": "discussant",
  "n09615465": "disputant, controversialist, eristic",
  "n09615807": "engineer, applied scientist, technologist",
  "n09616573": "enologist, oenologist, fermentologist",
  "n09616722": "ensign",
  "n09616922": "entertainer",
  "n09617161": "eulogist, panegyrist",
  "n09617292": "excavator",
  "n09617435": "ex-gambler",
  "n09617508": "ex-mayor",
  "n09617577": "experimenter",
  "n09617696": "experimenter",
  "n09617867": "expert",
  "n09618760": "exponent",
  "n09618880": "ex-president",
  "n09618957": "face",
  "n09619168": "female, female person",
  "n09619452": "finisher",
  "n09619605": "finisher",
  "n09619734": "finisher",
  "n09619824": "individualist",
  "n09620078": "inhabitant, habitant, dweller, denizen, indweller",
  "n09620794": "native, indigen, indigene, aborigine, aboriginal",
  "n09621232": "native",
  "n09621359": "innocent, inexperienced person",
  "n09621545": "intellectual, intellect",
  "n09622049": "juvenile, juvenile person",
  "n09622302": "lover",
  "n09622745": "lover",
  "n09622928": "loved one",
  "n09623038": "leader",
  "n09624168": "male, male person",
  "n09624559": "mediator, go-between, intermediator, intermediary, intercessor",
  "n09624899": "mediatrix",
  "n09624980": "money handler, money dealer",
  "n09625213": "monochromat",
  "n09625309": "naprapath",
  "n09625401": "national, subject",
  "n09625676": "nativist",
  "n09625789": "nonreligious person",
  "n09626031": "nonworker",
  "n09626238": "peer, equal, match, compeer",
  "n09626589": "perceiver, percipient, observer, beholder",
  "n09627017": "percher",
  "n09627117": "precursor, forerunner",
  "n09627263": "preteen, preteenager",
  "n09627462": "primitive, primitive person",
  "n09627807": "prize winner, lottery winner",
  "n09627906": "recipient, receiver",
  "n09628382": "religious person",
  "n09629065": "religionist",
  "n09629246": "sensualist",
  "n09629477": "ticket agent, booking clerk",
  "n09629643": "ticket holder",
  "n09629752": "traveler, traveller",
  "n09630641": "unfortunate, unfortunate person",
  "n09631129": "unwelcome person, persona non grata",
  "n09631463": "unpleasant person, disagreeable person",
  "n09632274": "unskilled person",
  "n09632518": "worker",
  "n09633969": "wrongdoer, offender",
  "n09634494": "African",
  "n09635534": "Black African",
  "n09635635": "Afrikaner, Afrikander, Boer",
  "n09635823": "Aryan, Indo-European",
  "n09635973": "Aryan",
  "n09636106": "person of color, person of colour",
  "n09636339": "Black, Black person, blackamoor, Negro, Negroid",
  "n09636796": "Negress",
  "n09636890": "Black race, Negroid race, Negro race",
  "n09637013": "African-American, African American, Afro-American, Black American",
  "n09637211": "Black man",
  "n09637339": "Black woman",
  "n09637435": "soul brother",
  "n09637512": "colored person, colored",
  "n09637684": "darky, darkie, darkey",
  "n09637837": "boy",
  "n09638009": "nigger, nigga, spade, coon, jigaboo, nigra",
  "n09638245": "Tom, Uncle Tom",
  "n09638454": "mulatto",
  "n09638569": "quadroon",
  "n09638722": "octoroon",
  "n09638875": "White, White person, Caucasian",
  "n09639237": "White race, White people, Caucasoid race, Caucasian race",
  "n09639382": "Circassian",
  "n09639543": "Abkhaz, Abkhazian, Abkhas, Abkhasian",
  "n09639719": "paleface",
  "n09639919": "Semite",
  "n09640220": "Babylonian",
  "n09640327": "Chaldean, Chaldaean, Chaldee",
  "n09640490": "Assyrian",
  "n09640577": "Kassite, Cassite",
  "n09640715": "Elamite",
  "n09640859": "Phoenician",
  "n09641002": "white man",
  "n09641130": "white woman",
  "n09641226": "white trash, poor white trash",
  "n09641422": "whitey, honky, honkey, honkie",
  "n09641578": "WASP, white Anglo-Saxon Protestant",
  "n09641757": "Asian, Asiatic",
  "n09642821": "Asian American",
  "n09642917": "coolie, cooly",
  "n09643078": "Oriental, oriental person",
  "n09643421": "Yellow race, Mongoloid race, Mongolian race",
  "n09643545": "yellow man",
  "n09643670": "yellow woman",
  "n09643799": "gook, slant-eye",
  "n09644008": "Evenki, Ewenki",
  "n09644152": "Mongol, Mongolian",
  "n09644332": "Tatar",
  "n09644518": "Udmurt, Votyak",
  "n09644657": "Tatar, Tartar, Mongol Tatar",
  "n09644820": "Amerindian, Native American",
  "n09645091": "Indian, American Indian, Red Indian",
  "n09645788": "brave",
  "n09645871": "Abnaki, Abenaki",
  "n09645996": "Achomawi",
  "n09646112": "Akwa'ala",
  "n09646220": "Alabama",
  "n09646432": "Algonkian, Algonkin",
  "n09646608": "Algonquian, Algonquin",
  "n09647473": "Anasazi",
  "n09647680": "Atakapa, Attacapan",
  "n09647834": "Athapaskan, Athapascan, Athabaskan, Athabascan",
  "n09648176": "Indian race, Amerindian race",
  "n09648309": "Mayan, Maya",
  "n09648743": "Nahuatl",
  "n09648911": "Aztec",
  "n09649067": "Olmec",
  "n09649249": "Toltec",
  "n09649372": "Zapotec, Zapotecan",
  "n09649554": "Plains Indian, Buffalo Indian",
  "n09649926": "Apache",
  "n09650250": "Arapaho, Arapahoe",
  "n09650456": "Arikara, Aricara",
  "n09650609": "Atsugewi",
  "n09650729": "Biloxi",
  "n09650839": "Blackfoot",
  "n09650989": "Brule",
  "n09651123": "Caddo",
  "n09651418": "Cakchiquel",
  "n09651537": "Catawba",
  "n09651653": "Cayuga",
  "n09651790": "Cherokee",
  "n09651968": "Cheyenne",
  "n09652149": "Chickasaw",
  "n09652278": "Chimakum",
  "n09652398": "Chimariko",
  "n09652535": "Chinook",
  "n09652746": "Chipewyan",
  "n09652900": "Choctaw",
  "n09653014": "Cochimi",
  "n09653144": "Cocopa, Cocopah",
  "n09653295": "Coeur d'Alene",
  "n09653438": "Comanche",
  "n09653616": "Conoy",
  "n09653805": "Costanoan",
  "n09653971": "Cree",
  "n09654079": "Creek",
  "n09654259": "Crow",
  "n09654374": "Dakota",
  "n09654518": "Delaware",
  "n09654687": "Dhegiha",
  "n09654898": "Diegueno",
  "n09655018": "Erie",
  "n09655213": "Esselen",
  "n09655354": "Essene",
  "n09655466": "Eyeish",
  "n09655569": "Fox",
  "n09655709": "Haida",
  "n09655891": "Halchidhoma",
  "n09656077": "Havasupai",
  "n09656205": "Hidatsa, Gros Ventre",
  "n09656378": "Hitchiti",
  "n09656547": "Hopi",
  "n09656673": "Hokan, Hoka",
  "n09657206": "Hunkpapa",
  "n09657489": "Hupa",
  "n09657615": "Illinois",
  "n09657748": "Iowa, Ioway",
  "n09657887": "Iroquois",
  "n09658254": "Kalapooia, Kalapuya, Calapooya, Calapuya",
  "n09658398": "Kamia",
  "n09658543": "Kansa, Kansas",
  "n09658670": "Karok",
  "n09658815": "Kekchi",
  "n09658921": "Kichai",
  "n09659039": "Kickapoo",
  "n09659188": "Kiliwa, Kiliwi",
  "n09659327": "Kiowa",
  "n09659447": "Koasati",
  "n09659605": "Kusan",
  "n09659711": "Kwakiutl",
  "n09659866": "Maidu",
  "n09660010": "Malecite",
  "n09660138": "Mam",
  "n09660240": "Maricopa",
  "n09660373": "Massachuset, Massachusetts",
  "n09660524": "Mattole",
  "n09660645": "Menomini, Menominee",
  "n09660820": "Miniconju",
  "n09660958": "Missouri",
  "n09661107": "Miami",
  "n09661258": "Micmac, Mikmaq",
  "n09661397": "Miwok",
  "n09661543": "Mohave, Mojave",
  "n09661731": "Mohawk",
  "n09661873": "Mohican, Mahican",
  "n09662038": "Muskhogean, Muskogean",
  "n09662325": "Muskogee",
  "n09662530": "Nanticoke",
  "n09662661": "Navaho, Navajo",
  "n09662804": "Nez Perce",
  "n09662951": "Nootka",
  "n09663121": "Ofo",
  "n09663248": "Oglala, Ogalala",
  "n09663472": "Ojibwa, Ojibway, Chippewa",
  "n09663628": "Omaha, Maha",
  "n09663786": "Osage",
  "n09663999": "Oneida",
  "n09664121": "Onondaga",
  "n09664295": "Oto, Otoe",
  "n09664450": "Ottawa",
  "n09664556": "Paiute, Piute",
  "n09664766": "Pamlico",
  "n09664908": "Passamaquody",
  "n09665075": "Patwin",
  "n09665218": "Pawnee",
  "n09665367": "Penobscot",
  "n09665545": "Penutian",
  "n09665897": "Pima",
  "n09666039": "Pomo",
  "n09666206": "Ponca, Ponka",
  "n09666349": "Potawatomi",
  "n09666476": "Powhatan",
  "n09666622": "Pueblo",
  "n09666883": "kachina",
  "n09666974": "Quapaw",
  "n09667097": "Quiche",
  "n09667205": "Redskin, Injun, red man",
  "n09667358": "Salish",
  "n09667572": "Santee, Santee Sioux, Santee Dakota, Eastern Sioux",
  "n09667715": "Sauk, Sac",
  "n09667927": "Seminole",
  "n09668058": "Seneca",
  "n09668199": "Shahaptian, Sahaptin, Sahaptino",
  "n09668437": "Shasta",
  "n09668562": "Shawnee",
  "n09668729": "Shoshone, Shoshoni",
  "n09668988": "Sihasapa",
  "n09669125": "Sioux, Siouan",
  "n09669631": "Teton, Lakota, Teton Sioux, Teton Dakota",
  "n09669942": "Skagit",
  "n09670051": "Takelma",
  "n09670170": "Taos",
  "n09670280": "Taracahitian",
  "n09670415": "Cahita",
  "n09670521": "Tarahumara",
  "n09670637": "Tlingit",
  "n09670772": "Tsimshian",
  "n09670909": "Tuscarora",
  "n09671089": "Tutelo",
  "n09671202": "Two Kettle",
  "n09671334": "Ute",
  "n09671453": "Wakashan",
  "n09671641": "Wampanoag",
  "n09671808": "Walapai, Hualapai, Hualpai",
  "n09671974": "Wichita",
  "n09672105": "Winnebago",
  "n09672314": "Wintun",
  "n09672455": "Yahi",
  "n09672590": "Yana",
  "n09672725": "Yavapai",
  "n09672840": "Yokuts",
  "n09672963": "Yucatec, Yucateco",
  "n09673091": "Yuma",
  "n09673232": "Zuni",
  "n09673341": "Indian race",
  "n09673495": "Indian",
  "n09673779": "Assamese",
  "n09673916": "Dravidian",
  "n09674307": "Badaga",
  "n09674412": "Gadaba",
  "n09674521": "Gond",
  "n09674631": "Kanarese, Canarese",
  "n09674786": "Kolam",
  "n09674908": "Kota, Kotar",
  "n09675045": "Kui",
  "n09675156": "Malto",
  "n09675283": "Savara",
  "n09675393": "Tamil",
  "n09675529": "Telugu",
  "n09675673": "Toda",
  "n09675799": "Tulu",
  "n09675922": "Gujarati, Gujerati",
  "n09676021": "Kashmiri",
  "n09676127": "Oriya",
  "n09676247": "Punjabi, Panjabi",
  "n09676374": "Maratha, Mahratta",
  "n09676490": "Aborigine, Abo, Aboriginal, native Australian, Australian Aborigine",
  "n09676746": "Slavic people, Slavic race",
  "n09676884": "Slav",
  "n09677168": "Acadian",
  "n09677280": "Cajun",
  "n09677427": "Anabaptist",
  "n09677561": "Mennonite",
  "n09677713": "Amish",
  "n09677830": "Dunker, Dunkard, Tunker",
  "n09678009": "Christian",
  "n09678581": "Christian Scientist",
  "n09678747": "Adventist, Second Adventist",
  "n09678917": "gentile",
  "n09679028": "gentile, non-Jew, goy",
  "n09679170": "gentile",
  "n09679316": "Protestant",
  "n09679708": "Friend, Quaker",
  "n09679925": "Catholic",
  "n09680136": "non-Catholic",
  "n09680235": "Anglican Catholic",
  "n09680387": "Greek Catholic",
  "n09680504": "Roman Catholic",
  "n09680657": "papist",
  "n09680908": "Old Catholic",
  "n09681107": "Uniat, Uniate, Uniate Christian",
  "n09681234": "Copt",
  "n09681351": "Jew, Hebrew, Israelite",
  "n09681973": "Jewess",
  "n09682122": "kike, hymie, sheeny, yid",
  "n09682291": "Muslim, Moslem",
  "n09682803": "Islamist",
  "n09682967": "Almoravid",
  "n09683180": "Jihadist",
  "n09683306": "Shiite, Shi'ite, Shiite Muslim, Shi'ite Muslim, Shia Muslim",
  "n09683559": "Sunnite, Sunni, Sunni Muslim",
  "n09683757": "Buddhist",
  "n09683924": "Zen Buddhist",
  "n09684082": "Mahayanist",
  "n09684226": "Hinayanist",
  "n09684352": "Lamaist",
  "n09684476": "Tantrist",
  "n09684609": "Hindu, Hindoo",
  "n09684901": "swami",
  "n09685006": "chela",
  "n09685085": "Jainist",
  "n09685233": "Hare Krishna",
  "n09685398": "Shaktist",
  "n09685564": "Shivaist",
  "n09685711": "Vaishnava",
  "n09685806": "Shintoist",
  "n09685922": "Rastafarian, Rasta",
  "n09686051": "Mithraist",
  "n09686146": "Zoroastrian",
  "n09686262": "Eurafrican",
  "n09686401": "Eurasian",
  "n09686536": "European",
  "n09687775": "sahib",
  "n09687940": "memsahib",
  "n09688008": "Celt, Kelt",
  "n09688233": "Gael",
  "n09688366": "Briton",
  "n09688487": "Gaul",
  "n09688596": "Galatian",
  "n09688804": "Frank",
  "n09689000": "Salian Frank, Salian",
  "n09689152": "Teuton",
  "n09689435": "Afghan, Afghanistani",
  "n09689631": "Kafir",
  "n09689739": "Pathan, Pashtun, Pushtun, Pashtoon",
  "n09689958": "Albanian",
  "n09690083": "Algerian",
  "n09690208": "Altaic",
  "n09690371": "Armenian",
  "n09690496": "Andorran",
  "n09690621": "Angolan",
  "n09690762": "Angolese",
  "n09690864": "Anguillan",
  "n09691024": "Antiguan",
  "n09691149": "Argentinian",
  "n09691279": "Australian, Aussie",
  "n09691435": "Austronesian",
  "n09691604": "Austrian",
  "n09691729": "Bahamian",
  "n09691858": "Bahraini, Bahreini",
  "n09691994": "Bangladeshi",
  "n09692125": "Basotho",
  "n09692250": "Basque",
  "n09692430": "Bengali",
  "n09692624": "Bantu",
  "n09692915": "Herero",
  "n09693100": "Hutu",
  "n09693244": "Luba, Chiluba",
  "n09693372": "Sotho",
  "n09693618": "Tswana, Bechuana, Batswana",
  "n09693809": "Tutsi, Watutsi, Watusi",
  "n09693982": "Barbadian",
  "n09694109": "Belgian",
  "n09694269": "Beninese",
  "n09694392": "Bermudan, Bermudian",
  "n09694529": "Bhutanese, Bhutani",
  "n09694664": "Bolivian",
  "n09694771": "Bornean",
  "n09694894": "Brazilian",
  "n09695019": "Carioca",
  "n09695132": "Tupi",
  "n09695257": "Guarani",
  "n09695379": "Maraco",
  "n09695514": "Bruneian",
  "n09695620": "Bulgarian",
  "n09695747": "Burmese",
  "n09695871": "Burundian",
  "n09695979": "Byelorussian, Belorussian, White Russian",
  "n09696124": "Byzantine",
  "n09696280": "Cambodian, Kampuchean",
  "n09696456": "Cameroonian",
  "n09696585": "Canadian",
  "n09696763": "French Canadian",
  "n09696944": "Canuck",
  "n09697070": "Carthaginian",
  "n09697246": "Cebuan",
  "n09697401": "Central American",
  "n09697650": "Chadian",
  "n09697771": "Chewa, Cewa, Chichewa",
  "n09697986": "Chilean",
  "n09698108": "Chinese",
  "n09698337": "chink, Chinaman",
  "n09698517": "Colombian",
  "n09698644": "Congolese",
  "n09698788": "Costa Rican",
  "n09698901": "Cuban",
  "n09699020": "Cypriot, Cypriote, Cyprian",
  "n09699200": "Czechoslovakian, Czechoslovak, Czech",
  "n09699403": "Czech",
  "n09699536": "Slovak",
  "n09699642": "Dane",
  "n09699763": "Dutch, Dutch people",
  "n09699918": "Frisian",
  "n09700017": "Zealander",
  "n09700125": "Djiboutian",
  "n09700253": "East Indian",
  "n09700371": "Ecuadorian, Ecuadoran",
  "n09700492": "Egyptian",
  "n09700706": "Copt",
  "n09700823": "Salvadoran, Salvadorian, Salvadorean",
  "n09700964": "Britisher, Briton, Brit",
  "n09701148": "English person",
  "n09701603": "Englishman",
  "n09701833": "Englishwoman",
  "n09701977": "Anglo-Saxon",
  "n09702134": "Anglo-Saxon",
  "n09702541": "Anglo-Indian",
  "n09702673": "Angle",
  "n09702836": "Saxon",
  "n09703101": "West Saxon",
  "n09703181": "Jute",
  "n09703344": "Lombard, Langobard",
  "n09703485": "limey, John Bull",
  "n09703573": "pommy, pom",
  "n09703708": "Cantabrigian",
  "n09703809": "Cornishman",
  "n09703932": "Cornishwoman",
  "n09704057": "Lancastrian",
  "n09704157": "Lancastrian",
  "n09704283": "Geordie",
  "n09704368": "Hanoverian",
  "n09704509": "Liverpudlian, Scouser",
  "n09704630": "Londoner",
  "n09704770": "Cockney",
  "n09704876": "Mancunian",
  "n09705003": "Oxonian",
  "n09705124": "Ethiopian",
  "n09705287": "Ewe",
  "n09705471": "Fulani, Fula, Fulah, Fellata, Fulbe",
  "n09705671": "Amhara",
  "n09705784": "Eritrean",
  "n09705909": "Fijian",
  "n09706029": "Finn",
  "n09706132": "Fleming",
  "n09706255": "Komi",
  "n09706396": "Cheremis, Cheremiss, Mari",
  "n09706548": "Ingrian, Inger, Ingerman",
  "n09706746": "Karelian, Carelian",
  "n09706911": "Ostyak, Khanty",
  "n09707061": "Livonian",
  "n09707166": "Latvian",
  "n09707289": "Lithuanian",
  "n09707400": "Mordva, Mordvin, Mordvinian",
  "n09707587": "Nganasan",
  "n09707735": "Selkup, Ostyak-Samoyed",
  "n09707887": "Samoyed",
  "n09707992": "Veps, Vepse, Vepsian",
  "n09708118": "Vogul, Mansi",
  "n09708247": "Yeniseian",
  "n09708405": "Frenchman, Frenchwoman, French person",
  "n09708648": "frog, Gaul",
  "n09708750": "Parisian",
  "n09708889": "Parisienne",
  "n09709001": "Breton",
  "n09709155": "Savoyard",
  "n09709229": "Angevin, Angevine",
  "n09709332": "Balkan",
  "n09709441": "Castillian",
  "n09709531": "Creole",
  "n09709673": "Creole",
  "n09709795": "Cretan",
  "n09709916": "Minoan",
  "n09710041": "Gabonese",
  "n09710164": "Greek, Hellene",
  "n09710619": "Achaean, Achaian",
  "n09710753": "Aeolian, Eolian",
  "n09710886": "Dorian",
  "n09711009": "Ionian",
  "n09711132": "Athenian",
  "n09711320": "Corinthian",
  "n09711435": "Laconian",
  "n09711530": "Lesbian",
  "n09711661": "Spartan",
  "n09711754": "Arcadian",
  "n09711870": "Theban",
  "n09711978": "Theban",
  "n09712090": "Thracian",
  "n09712195": "Guatemalan",
  "n09712324": "Guyanese",
  "n09712448": "Haitian",
  "n09712570": "Honduran",
  "n09712696": "Malay, Malayan",
  "n09712967": "Moro",
  "n09713108": "Netherlander, Dutchman, Hollander",
  "n09713260": "Norman",
  "n09713357": "Palestinian, Palestinian Arab",
  "n09713501": "Hindu, Hindoo, Hindustani",
  "n09713764": "Hmong, Miao",
  "n09713985": "Hungarian, Magyar",
  "n09714120": "Icelander",
  "n09714264": "Indonesian",
  "n09714429": "Irani, Iranian, Persian",
  "n09714694": "Iraqi, Iraki",
  "n09714952": "Irish person, Irelander",
  "n09715165": "Irishman",
  "n09715303": "Irishwoman",
  "n09715427": "Dubliner",
  "n09715521": "Paddy, Mick, Mickey",
  "n09715685": "Israelite",
  "n09715833": "Israeli",
  "n09715974": "sabra",
  "n09716047": "Italian",
  "n09716439": "wop, dago, ginzo, Guinea, greaseball",
  "n09716643": "Etruscan",
  "n09716837": "Neopolitan",
  "n09716933": "Roman",
  "n09717047": "Roman",
  "n09717233": "Sabine",
  "n09717454": "Venetian",
  "n09717566": "Sicilian",
  "n09717696": "Tuscan",
  "n09717789": "Oscan",
  "n09717917": "Samnite",
  "n09718092": "Jamaican",
  "n09718217": "Japanese, Nipponese",
  "n09718518": "Ryukyuan",
  "n09718652": "Jap, Nip",
  "n09718811": "Jordanian",
  "n09718936": "Korean",
  "n09719105": "North Korean",
  "n09719207": "South Korean",
  "n09719309": "Kenyan",
  "n09719430": "Kurd",
  "n09719653": "Kuwaiti",
  "n09719794": "Lao, Laotian",
  "n09720033": "Lapp, Lapplander, Sami, Saami, Same, Saame",
  "n09720256": "Latin American, Latino",
  "n09720406": "spic, spik, spick",
  "n09720595": "Lebanese",
  "n09720702": "Levantine",
  "n09720842": "Liberian",
  "n09720967": "Libyan",
  "n09721088": "Liechtensteiner",
  "n09721244": "Luxemburger, Luxembourger",
  "n09721444": "Macedonian",
  "n09721647": "Madagascan",
  "n09721759": "Malawian",
  "n09721883": "Malaysian",
  "n09722064": "Sabahan",
  "n09722159": "Maldivian, Maldivan",
  "n09722279": "Malian",
  "n09722399": "Mauritanian",
  "n09722530": "Mauritian",
  "n09722658": "Mexican",
  "n09722817": "Chicano",
  "n09722898": "greaser, wetback, taco",
  "n09723067": "Mexican-American, Mexicano",
  "n09723228": "Montserratian",
  "n09723361": "Moor",
  "n09723564": "Moroccan",
  "n09723689": "Mozambican",
  "n09723819": "Namibian",
  "n09723944": "Nauruan",
  "n09724066": "Nepalese, Nepali",
  "n09724234": "Gurkha",
  "n09724365": "Gurkha",
  "n09724533": "New Zealander, Kiwi",
  "n09724656": "Nicaraguan",
  "n09724785": "Nigerian",
  "n09725000": "Hausa, Haussa",
  "n09725124": "Nigerien",
  "n09725229": "North American",
  "n09725402": "Norwegian, Norseman, Norse",
  "n09725546": "Nova Scotian, bluenose",
  "n09725653": "Omani",
  "n09725772": "Pakistani",
  "n09725935": "Brahui",
  "n09726037": "Sindhi",
  "n09726120": "Panamanian",
  "n09726246": "Paraguayan",
  "n09726374": "Parthian",
  "n09726499": "Peruvian",
  "n09726621": "South American Indian",
  "n09726811": "Carib, Carib Indian",
  "n09726970": "Quechua, Kechua",
  "n09727191": "Inca, Inka, Incan",
  "n09727323": "Inca",
  "n09727440": "Filipino",
  "n09727627": "Pole",
  "n09727747": "polack",
  "n09727826": "Polynesian",
  "n09728009": "Portuguese",
  "n09728137": "Qatari, Katari",
  "n09728285": "Romanian, Rumanian",
  "n09728403": "Russian",
  "n09728850": "Great Russian",
  "n09729062": "Muscovite",
  "n09729156": "Georgian",
  "n09729271": "Samoan",
  "n09729387": "Saudi, Saudi Arabian",
  "n09729530": "Arab, Arabian",
  "n09729945": "San Marinese",
  "n09730077": "Sarawakian",
  "n09730204": "Scandinavian, Norse, Northman",
  "n09730383": "Viking",
  "n09730533": "Scot, Scotsman, Scotchman",
  "n09730732": "Scotswoman, Scotchwoman",
  "n09730824": "Senegalese",
  "n09730951": "Seychellois",
  "n09731082": "Siberian",
  "n09731225": "Sierra Leonean",
  "n09731343": "Slovene",
  "n09731436": "South African",
  "n09731571": "South American",
  "n09731906": "Spaniard",
  "n09732047": "Sinhalese, Singhalese",
  "n09732170": "Sudanese",
  "n09732293": "Swazi",
  "n09732441": "Swede",
  "n09732544": "British, British people, Brits",
  "n09732668": "English, English people",
  "n09732778": "Irish, Irish people",
  "n09732903": "French, French people",
  "n09733028": "Sherpa",
  "n09733243": "Spanish, Spanish people",
  "n09733333": "Swiss, Swiss people",
  "n09733459": "Syrian",
  "n09733580": "Damascene",
  "n09733707": "Khmer",
  "n09733793": "Tahitian",
  "n09733899": "Taiwanese",
  "n09734006": "Tajik, Tadzhik",
  "n09734185": "Tanzanian",
  "n09734294": "Thai, Tai, Siamese",
  "n09734450": "Tibetan",
  "n09734535": "Togolese",
  "n09734639": "Tuareg",
  "n09734760": "Tunisian",
  "n09734885": "Turk",
  "n09735023": "Tyrolean",
  "n09735113": "Ottoman, Ottoman Turk, Osmanli",
  "n09735258": "Turki",
  "n09735542": "Azerbaijani",
  "n09735654": "Chuvash",
  "n09735790": "effendi",
  "n09735913": "Karakalpak",
  "n09736035": "Kazak, Kazakh",
  "n09736181": "Kazakhstani",
  "n09736312": "Kirghiz, Kirgiz, Khirghiz",
  "n09736485": "Turkoman, Turkmen, Turcoman",
  "n09736633": "Uighur, Uigur, Uygur",
  "n09736798": "Uzbek, Uzbeg, Uzbak, Usbek, Usbeg",
  "n09736945": "Ugandan",
  "n09737050": "Ukranian",
  "n09737161": "Yakut",
  "n09737297": "Tungusic",
  "n09737453": "Tungus, Evenk",
  "n09737653": "Manchu",
  "n09737835": "Khalkha, Khalka, Kalka",
  "n09737981": "Edo",
  "n09738121": "Igbo",
  "n09738275": "Yoruba",
  "n09738400": "American",
  "n09738708": "American",
  "n09740085": "American Revolutionary leader",
  "n09740724": "Anglo-American",
  "n09740856": "Alabaman, Alabamian",
  "n09740954": "Alaskan",
  "n09741074": "Alaska Native, Alaskan Native, Native Alaskan",
  "n09741233": "Arizonan, Arizonian",
  "n09741331": "Arkansan, Arkansawyer",
  "n09741432": "Bay Stater",
  "n09741526": "Bostonian",
  "n09741612": "Californian",
  "n09741722": "Carolinian",
  "n09741816": "Coloradan",
  "n09741904": "Connecticuter",
  "n09741999": "Delawarean, Delawarian",
  "n09742101": "Floridian",
  "n09742188": "Franco-American",
  "n09742315": "German American",
  "n09742443": "Georgian",
  "n09742569": "Hawaiian",
  "n09742690": "Native Hawaiian",
  "n09742844": "Idahoan",
  "n09742927": "Illinoisan",
  "n09743016": "Indianan, Hoosier",
  "n09743112": "Iowan",
  "n09743192": "Kansan",
  "n09743275": "Kentuckian, Bluegrass Stater",
  "n09743383": "Louisianan, Louisianian",
  "n09743487": "Mainer, Down Easter",
  "n09743601": "Marylander",
  "n09743690": "Michigander, Wolverine",
  "n09743792": "Minnesotan, Gopher",
  "n09743891": "Mississippian",
  "n09743986": "Missourian",
  "n09744075": "Montanan",
  "n09744161": "Nebraskan, Cornhusker",
  "n09744262": "Nevadan",
  "n09744346": "New Hampshirite, Granite Stater",
  "n09744462": "New Jerseyan, New Jerseyite, Garden Stater",
  "n09744587": "New Mexican",
  "n09744679": "New Yorker",
  "n09744834": "North Carolinian, Tarheel",
  "n09744945": "North Dakotan",
  "n09745041": "Ohioan, Buckeye",
  "n09745132": "Oklahoman, Sooner",
  "n09745229": "Oregonian, Beaver",
  "n09745324": "Pennsylvanian, Keystone Stater",
  "n09745438": "Rhode Islander",
  "n09745535": "South Carolinian",
  "n09745636": "South Dakotan",
  "n09745732": "Tennessean, Volunteer",
  "n09745834": "Texan",
  "n09745933": "Utahan",
  "n09746014": "Vermonter",
  "n09746101": "Virginian",
  "n09746189": "Washingtonian",
  "n09746314": "Washingtonian",
  "n09746438": "West Virginian",
  "n09746536": "Wisconsinite, Badger",
  "n09746637": "Wyomingite",
  "n09746725": "Puerto Rican",
  "n09746819": "Yankee, Yank, Yankee-Doodle",
  "n09746936": "Uruguayan",
  "n09747062": "Venezuelan",
  "n09747191": "Vietnamese, Annamese",
  "n09747329": "Welshman, Welsh, Cambrian, Cymry",
  "n09747495": "Gambian",
  "n09747618": "Maltese",
  "n09747722": "German",
  "n09747969": "Teuton",
  "n09748101": "East German",
  "n09748239": "Kraut, Krauthead, Boche, Jerry, Hun",
  "n09748408": "Berliner",
  "n09748541": "West Berliner",
  "n09748648": "Prussian",
  "n09748770": "Junker",
  "n09748889": "Ghanian",
  "n09749011": "Gibraltarian",
  "n09749142": "Glaswegian",
  "n09749260": "Grenadian",
  "n09749386": "Guinean",
  "n09749509": "Rwandan",
  "n09749614": "Singaporean",
  "n09749753": "Slovenian",
  "n09749880": "Somalian, Somali",
  "n09750046": "Sri Lankan",
  "n09750157": "Sumatran",
  "n09750282": "Papuan",
  "n09750410": "Tongan",
  "n09750524": "Trojan, Dardan, Dardanian",
  "n09750641": "Walloon",
  "n09750770": "Yemeni",
  "n09750891": "Yugoslav, Jugoslav, Yugoslavian, Jugoslavian",
  "n09751076": "Serbian, Serb",
  "n09751256": "Croatian, Croat",
  "n09751403": "Sorbian",
  "n09751496": "Xhosa",
  "n09751622": "Zairese, Zairean",
  "n09751772": "Zambian",
  "n09751895": "Zimbabwean",
  "n09752023": "Zulu",
  "n09752246": "Aries, Ram",
  "n09752381": "Taurus, Bull",
  "n09752519": "Gemini, Twin",
  "n09752657": "Cancer, Crab",
  "n09752795": "Leo, Lion",
  "n09752927": "Virgo, Virgin",
  "n09753065": "Libra, Balance",
  "n09753204": "Scorpio, Scorpion",
  "n09753348": "Sagittarius, Archer",
  "n09753498": "Capricorn, Goat",
  "n09753642": "Aquarius, Water Bearer",
  "n09753792": "Pisces, Fish",
  "n09753930": "abandoned person",
  "n09754051": "abator",
  "n09754152": "abbe",
  "n09754217": "abbess, mother superior, prioress",
  "n09754404": "abbot, archimandrite",
  "n09754541": "abjurer",
  "n09754633": "abnegator",
  "n09754780": "abominator, loather",
  "n09754907": "abridger, abbreviator",
  "n09755086": "abstractor, abstracter",
  "n09755241": "absconder",
  "n09755398": "absolutist",
  "n09755555": "absolver",
  "n09755657": "abdicator",
  "n09755788": "abecedarian",
  "n09755893": "aberrant",
  "n09756049": "abettor, abetter",
  "n09756195": "abhorrer",
  "n09756400": "abiogenist",
  "n09756500": "able seaman, able-bodied seaman",
  "n09756637": "abolitionist, emancipationist",
  "n09756961": "abomination",
  "n09757076": "autochthon",
  "n09757175": "abortionist",
  "n09757311": "abrogator",
  "n09757449": "abseiler, rappeller",
  "n09757653": "absentee",
  "n09757836": "AWOL",
  "n09757944": "abstainer, abstinent, nondrinker",
  "n09758173": "abstainer, ascetic",
  "n09758424": "abstractionist, abstract artist",
  "n09758643": "abuser, maltreater",
  "n09758781": "abutter",
  "n09758885": "academic administrator",
  "n09759069": "academician, academic, faculty member",
  "n09759311": "academician, schoolman",
  "n09759501": "academician",
  "n09759684": "acceptor",
  "n09759875": "accessory, accessary",
  "n09760080": "accessory after the fact",
  "n09760290": "accessory before the fact",
  "n09760476": "accessory during the fact",
  "n09760609": "companion",
  "n09760735": "accommodation endorser",
  "n09760913": "accompanist, accompanyist",
  "n09761068": "accomplice, confederate",
  "n09761310": "accordionist",
  "n09761403": "accountant, comptroller, controller",
  "n09761753": "account executive, account representative, registered representative, customer's broker, customer's man",
  "n09762011": "accused",
  "n09762101": "defendant, suspect",
  "n09762385": "accuser",
  "n09762509": "ace, adept, champion, sensation, maven, mavin, virtuoso, genius, hotshot, star, superstar, whiz, whizz, wizard, wiz",
  "n09762821": "achiever, winner, success, succeeder",
  "n09763272": "acid head",
  "n09763349": "acolyte",
  "n09763668": "acoustician",
  "n09763784": "acquaintance, friend",
  "n09764201": "acquirer",
  "n09764381": "acrobat",
  "n09764598": "aerialist",
  "n09764732": "action officer",
  "n09764900": "active",
  "n09765118": "active citizen",
  "n09765278": "actor, histrion, player, thespian, role player",
  "n09767197": "actor, doer, worker",
  "n09767592": "actor's agent, theatrical agent",
  "n09767700": "actress",
  "n09768753": "adder",
  "n09768830": "addict",
  "n09769076": "addict, nut, freak, junkie, junky",
  "n09769345": "addressee",
  "n09769525": "adducer",
  "n09769636": "adjudicator",
  "n09769803": "adjunct",
  "n09769929": "adjuster, adjustor, claims adjuster, claims adjustor, claim agent",
  "n09770179": "adjutant, aide, aide-de-camp",
  "n09770359": "adjutant general",
  "n09770472": "administrator, executive",
  "n09770743": "administrator",
  "n09770949": "administrator, decision maker",
  "n09771204": "admiral, full admiral",
  "n09771435": "admirer, adorer",
  "n09771664": "admirer",
  "n09771855": "admonisher, monitor, reminder",
  "n09772029": "adolescent, stripling, teenager, teen",
  "n09772330": "adoptee",
  "n09772448": "adoptive parent, adopter",
  "n09772606": "adulterator",
  "n09772746": "adulterer, fornicator",
  "n09772930": "adulteress, fornicatress, hussy, jade, loose woman, slut, strumpet, trollop",
  "n09773076": "advancer",
  "n09773169": "adventuress",
  "n09773245": "adversary, antagonist, opponent, opposer, resister",
  "n09773682": "adverse witness, hostile witness",
  "n09773962": "advertiser, advertizer, adman",
  "n09774167": "advisee",
  "n09774266": "adviser, advisor, consultant",
  "n09774783": "advocate, advocator, proponent, exponent",
  "n09775663": "advocate, counsel, counselor, counsellor, counselor-at-law, pleader",
  "n09775907": "aeronautical engineer",
  "n09776079": "aerospace engineer",
  "n09776186": "aerophile",
  "n09776261": "affiant",
  "n09776346": "affiliate",
  "n09776522": "affine",
  "n09776642": "affluent",
  "n09776807": "aficionado",
  "n09776933": "aficionado",
  "n09777012": "agent, factor, broker",
  "n09777353": "agent",
  "n09777870": "buck sergeant",
  "n09777975": "business agent",
  "n09778144": "literary agent",
  "n09778266": "agent-in-place",
  "n09778392": "agent provocateur, provocateur",
  "n09778537": "aggravator, annoyance",
  "n09778676": "aggressor",
  "n09778783": "agitator, fomenter",
  "n09778927": "agnostic",
  "n09779124": "agnostic, doubter",
  "n09779280": "agonist",
  "n09779461": "agony aunt",
  "n09779623": "agricultural laborer, agricultural labourer",
  "n09779790": "agriculturist, agriculturalist, cultivator, grower, raiser",
  "n09780120": "agronomist",
  "n09780249": "aide, auxiliary",
  "n09780395": "air attache",
  "n09780547": "aircraftsman, aircraftman",
  "n09780676": "aircrewman",
  "n09780828": "air force officer, commander",
  "n09780984": "airhead",
  "n09781171": "air marshal, sky marshal",
  "n09781398": "air traveler, air traveller",
  "n09781504": "alarmist",
  "n09781650": "albino",
  "n09781804": "alcalde",
  "n09781921": "alchemist",
  "n09782167": "alcoholic, alky, dipsomaniac, boozer, lush, soaker, souse",
  "n09782397": "alderman",
  "n09782589": "Aleut, Aleutian",
  "n09782730": "Alexandrian",
  "n09782855": "alexic",
  "n09782946": "Ali Baba",
  "n09783130": "alien absconder",
  "n09783369": "alienator",
  "n09783537": "alienee, grantee",
  "n09783653": "alienist",
  "n09783776": "alienor",
  "n09783884": "aliterate, aliterate person",
  "n09784043": "algebraist",
  "n09784160": "allegorizer, allegoriser",
  "n09784306": "allergist",
  "n09784443": "alleviator",
  "n09784564": "alliterator",
  "n09784707": "allocator, distributor",
  "n09784870": "all-rounder, all arounder",
  "n09785042": "ally, friend",
  "n09785236": "almoner, medical social worker",
  "n09785409": "alphabetizer, alphabetiser",
  "n09785580": "almsgiver",
  "n09785659": "alpinist",
  "n09785786": "Alsatian",
  "n09785891": "altar boy",
  "n09785992": "alter ego",
  "n09786115": "alto",
  "n09786226": "alto saxophonist, altoist",
  "n09786338": "alumnus, alumna, alum, graduate, grad",
  "n09786585": "amateur",
  "n09786760": "amateur",
  "n09786922": "amalgamator",
  "n09787076": "Amazon",
  "n09787293": "amazon, virago",
  "n09787390": "maenad",
  "n09787534": "ambassador, embassador",
  "n09787765": "ambassador",
  "n09787878": "ambassadress",
  "n09787955": "ambulance chaser",
  "n09788073": "ambusher",
  "n09788237": "amicus curiae, friend of the court",
  "n09788521": "amigo",
  "n09788611": "amnesic, amnesiac",
  "n09788761": "amora",
  "n09789150": "amoralist",
  "n09789304": "amorist",
  "n09789449": "amputator",
  "n09789566": "amputee",
  "n09789668": "anagnost",
  "n09789898": "analogist",
  "n09790047": "analphabet, analphabetic",
  "n09790188": "analysand",
  "n09790278": "analyst, psychoanalyst",
  "n09790482": "analyst",
  "n09790666": "credit analyst",
  "n09790865": "financial analyst, securities analyst",
  "n09791014": "industry analyst",
  "n09791151": "oil-industry analyst",
  "n09791248": "market analyst",
  "n09791419": "market strategist",
  "n09791530": "analyst",
  "n09791816": "anarchist, nihilist, syndicalist",
  "n09792125": "anathema, bete noire",
  "n09792237": "anatomist",
  "n09792555": "ancestor, ascendant, ascendent, antecedent, root",
  "n09792896": "ancestress",
  "n09792969": "anchor, anchorman, anchorperson",
  "n09793141": "ancient",
  "n09793248": "ancient, antediluvian",
  "n09793352": "anecdotist, raconteur",
  "n09793495": "anesthesiologist, anesthetist, anaesthetist",
  "n09793717": "angel, backer",
  "n09793830": "angiologist",
  "n09793946": "angler, troller",
  "n09794081": "angler",
  "n09794211": "anglophile, anglophil",
  "n09794320": "anglophobe",
  "n09794426": "animal fancier",
  "n09794550": "animator",
  "n09794668": "animist",
  "n09794797": "annalist",
  "n09794917": "annihilator",
  "n09795010": "annotator",
  "n09795124": "announcer",
  "n09795334": "announcer",
  "n09795556": "annuitant",
  "n09795639": "anointer",
  "n09795751": "anorexic, anorectic",
  "n09795894": "antediluvian, antediluvian patriarch",
  "n09796062": "anthologist",
  "n09796185": "anthropoid, ape",
  "n09796323": "anthropologist",
  "n09796809": "anti",
  "n09796974": "anti-American",
  "n09797113": "anticipator, anticipant",
  "n09797257": "antinomian",
  "n09797375": "antipope",
  "n09797606": "antiquary, antiquarian, archaist",
  "n09797742": "anti-Semite, Jew-baiter",
  "n09797873": "Anzac",
  "n09797998": "ape-man",
  "n09798096": "aphakic",
  "n09798244": "aphasic",
  "n09798404": "aphorist",
  "n09798534": "apologist, vindicator, justifier",
  "n09798811": "Apostle",
  "n09799213": "Apostle, Apostelic Father",
  "n09799461": "apostle",
  "n09799607": "apostolic delegate",
  "n09799808": "Appalachian",
  "n09799920": "apparatchik",
  "n09800056": "apparatchik",
  "n09800249": "appeaser",
  "n09800469": "appellant, plaintiff in error",
  "n09800631": "apple polisher, bootlicker, fawner, groveller, groveler, truckler",
  "n09800964": "appointee",
  "n09801102": "apprehender",
  "n09801275": "April fool",
  "n09801371": "aquanaut, oceanaut",
  "n09801533": "aspirant, aspirer, hopeful, wannabe, wannabee",
  "n09801864": "apprentice, learner, prentice",
  "n09802050": "appraiser, valuator",
  "n09802239": "appraiser, authenticator",
  "n09802445": "appreciator",
  "n09802641": "appropriator",
  "n09802844": "approver",
  "n09802951": "Arabist",
  "n09803060": "Aramean, Aramaean",
  "n09803247": "Arawak, Arawakan",
  "n09803429": "arbiter, arbitrator, umpire",
  "n09803800": "arbitrageur, arbitrager, arb",
  "n09804053": "arbiter, supreme authority",
  "n09804230": "archaist",
  "n09804343": "archdeacon",
  "n09804518": "archduchess",
  "n09804658": "archduke",
  "n09804806": "archeologist, archaeologist",
  "n09805151": "archbishop",
  "n09805324": "archer, bowman",
  "n09805475": "architect, designer",
  "n09806944": "archivist",
  "n09807075": "archpriest, hierarch, high priest, prelate, primate",
  "n09807379": "Areopagite",
  "n09807495": "Argive",
  "n09807609": "arianist",
  "n09807754": "aristocrat, blue blood, patrician",
  "n09808080": "Aristotelian, Aristotelean, Peripatetic",
  "n09808351": "arithmetician",
  "n09808466": "armchair liberal",
  "n09808591": "armiger",
  "n09808686": "armiger, armor-bearer",
  "n09808794": "armorer, armourer, artificer",
  "n09808949": "armorer, armourer",
  "n09809134": "arms manufacturer",
  "n09809279": "army attache",
  "n09809427": "army brat",
  "n09809538": "army engineer, military engineer",
  "n09809749": "army officer",
  "n09809925": "arranger, adapter, transcriber",
  "n09810166": "arrival, arriver, comer",
  "n09810364": "arrogator",
  "n09810633": "arrowsmith",
  "n09810707": "arsonist, incendiary, firebug",
  "n09810867": "art critic",
  "n09810983": "art dealer",
  "n09811112": "art director",
  "n09811281": "art editor",
  "n09811414": "art historian",
  "n09811568": "arthritic",
  "n09811712": "articulator",
  "n09811852": "artilleryman, cannoneer, gunner, machine gunner",
  "n09812068": "illustrator",
  "n09812338": "artist, creative person",
  "n09813126": "artiste",
  "n09813219": "artist's model, sitter",
  "n09813351": "art student",
  "n09813441": "art teacher",
  "n09813522": "ascender",
  "n09813632": "ass",
  "n09813696": "assassin, assassinator, bravo",
  "n09814084": "assassin",
  "n09814252": "assayer",
  "n09814381": "assemblyman",
  "n09814488": "assemblywoman",
  "n09814567": "assenter",
  "n09814660": "asserter, declarer, affirmer, asseverator, avower",
  "n09815076": "assessee",
  "n09815188": "asshole, bastard, cocksucker, dickhead, shit, mother fucker, motherfucker, prick, whoreson, son of a bitch, SOB",
  "n09815455": "assignee",
  "n09815677": "assignor",
  "n09815790": "assistant, helper, help, supporter",
  "n09816654": "assistant professor",
  "n09816771": "associate",
  "n09817174": "associate",
  "n09817386": "associate professor",
  "n09817536": "asthmatic",
  "n09817694": "astrogator",
  "n09817816": "astrologer, astrologist",
  "n09818022": "astronaut, spaceman, cosmonaut",
  "n09818343": "astronomer, uranologist, stargazer",
  "n09819291": "astrophysicist",
  "n09819477": "cosmographer, cosmographist",
  "n09819667": "cosmologist",
  "n09819860": "atavist, throwback",
  "n09820044": "atheist",
  "n09820263": "athlete, jock",
  "n09821086": "attache",
  "n09821253": "attacker, aggressor, assailant, assaulter",
  "n09821831": "attendant, attender, tender",
  "n09822640": "attester, attestant",
  "n09822830": "attorney general",
  "n09822955": "auditor",
  "n09823153": "auditor",
  "n09823287": "augur, auspex",
  "n09823502": "aunt, auntie, aunty",
  "n09823683": "au pair",
  "n09823832": "au pair girl",
  "n09823927": "auteur",
  "n09824065": "authoress",
  "n09824135": "authoritarian, dictator",
  "n09824361": "authority",
  "n09824609": "authority",
  "n09824978": "authority figure",
  "n09825096": "authorizer, authoriser",
  "n09825296": "autobiographer",
  "n09825413": "autodidact",
  "n09825519": "automaton, zombi, zombie",
  "n09825750": "automobile mechanic, auto-mechanic, car-mechanic, mechanic, grease monkey",
  "n09825939": "automotive engineer",
  "n09826074": "avenger, retaliator",
  "n09826204": "aviator, aeronaut, airman, flier, flyer",
  "n09826605": "aviatrix, airwoman, aviatress",
  "n09826698": "avower",
  "n09826821": "ayah",
  "n09826945": "ayatollah",
  "n09827166": "baas",
  "n09827246": "babu, baboo",
  "n09827363": "baby, babe, sister",
  "n09827519": "baby",
  "n09827683": "baby, babe, infant",
  "n09828216": "baby",
  "n09828403": "baby boomer, boomer",
  "n09828600": "baby buster, buster",
  "n09828760": "baby doctor, pediatrician, pediatrist, paediatrician",
  "n09828988": "baby farmer",
  "n09829122": "babyminder, baby minder, minder",
  "n09829305": "babysitter, baby-sitter, sitter",
  "n09829506": "bacchant",
  "n09829650": "bacchante",
  "n09829798": "bacchant, bacchanal",
  "n09829923": "bachelor, unmarried man",
  "n09830080": "bachelor girl, bachelorette",
  "n09830194": "back",
  "n09830400": "backbencher",
  "n09830553": "back judge",
  "n09830629": "backpacker, packer",
  "n09830759": "backroom boy, brain truster",
  "n09830926": "backscratcher",
  "n09831057": "backseat driver",
  "n09831166": "backslapper",
  "n09831303": "backstroker",
  "n09831411": "bacteriologist",
  "n09831731": "bad egg",
  "n09831856": "bad guy",
  "n09831962": "bad person",
  "n09832321": "bag, old bag",
  "n09832456": "baggage",
  "n09832538": "baggageman",
  "n09832633": "bag lady",
  "n09832762": "bagman",
  "n09832873": "Bahai",
  "n09832978": "bailee",
  "n09833111": "bailiff",
  "n09833275": "bailor",
  "n09833441": "bairn",
  "n09833536": "baker, bread maker",
  "n09833651": "baker",
  "n09833751": "balancer",
  "n09833896": "baldhead, baldpate, baldy",
  "n09833997": "balker, baulker, noncompliant",
  "n09834159": "ball boy",
  "n09834258": "ball-buster, ball-breaker",
  "n09834378": "ball carrier, runner",
  "n09834592": "ballerina, danseuse",
  "n09834699": "ballet dancer",
  "n09834885": "ballet master",
  "n09835017": "ballet mistress",
  "n09835153": "balletomane",
  "n09835230": "ball hawk",
  "n09835348": "balloonist",
  "n09835506": "ballplayer, baseball player",
  "n09836057": "bulimic",
  "n09836160": "bullfighter, toreador",
  "n09836343": "banderillero",
  "n09836519": "matador",
  "n09836658": "novillero",
  "n09836786": "picador",
  "n09836976": "torero",
  "n09837088": "bandit, brigand",
  "n09837201": "bandleader",
  "n09837360": "bandmaster",
  "n09837459": "bandsman",
  "n09837580": "bank commissioner",
  "n09837720": "banker",
  "n09837824": "banker",
  "n09837981": "bank examiner",
  "n09838117": "bank guard",
  "n09838200": "bank manager",
  "n09838295": "bank robber",
  "n09838370": "bankrupt, insolvent",
  "n09838511": "bantamweight",
  "n09838621": "bantamweight",
  "n09838701": "Baptist",
  "n09838895": "barber",
  "n09839022": "bard",
  "n09839085": "bar fly",
  "n09839167": "bargainer",
  "n09839343": "bargain hunter",
  "n09839436": "baritone, barytone",
  "n09839515": "barker",
  "n09839702": "barmaid",
  "n09839774": "barnburner",
  "n09839860": "barnstormer, stunt flier, stunt pilot",
  "n09840050": "barnstormer, playactor, play-actor, trouper",
  "n09840217": "baron, big businessman, business leader, king, magnate, mogul, power, top executive, tycoon",
  "n09840435": "baron",
  "n09840520": "baron",
  "n09840639": "baronet, Bart",
  "n09840871": "barrator, barrater",
  "n09840963": "barrister",
  "n09841188": "bartender, barman, barkeep, barkeeper, mixologist",
  "n09841400": "barterer",
  "n09841515": "baseball coach, baseball manager",
  "n09841696": "base runner, runner",
  "n09841865": "basileus",
  "n09841955": "basketball coach",
  "n09842047": "basketball player, basketeer, cager",
  "n09842288": "basketweaver, basketmaker",
  "n09842395": "Basket Maker",
  "n09842528": "bass, basso",
  "n09842629": "bassist",
  "n09842716": "bassoonist",
  "n09842823": "bastard, by-blow, love child, illegitimate child, illegitimate, whoreson",
  "n09843048": "baster, tacker",
  "n09843200": "baster",
  "n09843309": "baroness",
  "n09843443": "bat boy",
  "n09843602": "bather",
  "n09843716": "batman",
  "n09843824": "baton twirler, twirler",
  "n09843956": "batter, hitter, slugger, batsman",
  "n09844221": "batting coach",
  "n09844356": "battle-ax, battle-axe",
  "n09844457": "Bavarian",
  "n09844566": "bawler",
  "n09844685": "beachcomber",
  "n09844770": "beadle",
  "n09844898": "beadsman, bedesman",
  "n09845016": "bean counter",
  "n09845191": "bear",
  "n09845401": "beard",
  "n09845589": "beast, wolf, savage, brute, wildcat",
  "n09845737": "beater",
  "n09845849": "beatnik, beat",
  "n09845999": "beautician, cosmetician",
  "n09846142": "beauty consultant",
  "n09846265": "bedfellow",
  "n09846355": "bedfellow",
  "n09846469": "Bedouin, Beduin",
  "n09846586": "bedwetter, bed wetter, wetter",
  "n09846755": "beekeeper, apiarist, apiculturist",
  "n09846894": "beer drinker, ale drinker",
  "n09847010": "beggar, mendicant",
  "n09847267": "beggarman",
  "n09847344": "beggarwoman",
  "n09847425": "begum",
  "n09847543": "beldam, beldame",
  "n09847629": "bel esprit",
  "n09847727": "believer, worshiper, worshipper",
  "n09848110": "theist",
  "n09848285": "Taoist, Tao",
  "n09848489": "believer, truster",
  "n09848775": "bellboy, bellman, bellhop",
  "n09848916": "bell captain",
  "n09849012": "belle",
  "n09849167": "bell founder",
  "n09849256": "bell ringer",
  "n09849349": "bellwether",
  "n09849462": "belly dancer, exotic belly dancer, exotic dancer",
  "n09849598": "beloved, dear, dearest, honey, love",
  "n09849801": "belt maker",
  "n09849874": "bench warmer",
  "n09849990": "benedick, benedict",
  "n09850121": "beneficiary, donee",
  "n09850317": "Berber",
  "n09850457": "bereaved, bereaved person",
  "n09850642": "berk",
  "n09850760": "berserker, berserk",
  "n09850974": "besieger",
  "n09851087": "besieger",
  "n09851165": "best, topper",
  "n09851371": "best friend",
  "n09851465": "best man",
  "n09851575": "betrothed",
  "n09851719": "better",
  "n09851876": "bettor, better, wagerer, punter",
  "n09852081": "taker",
  "n09852179": "bey",
  "n09852289": "bey",
  "n09852430": "B-girl, bar girl",
  "n09852558": "bibliographer",
  "n09852679": "bibliophile, booklover, book lover",
  "n09852826": "bibliopole, bibliopolist",
  "n09852979": "bibliotist",
  "n09853087": "bidder",
  "n09853184": "bidder",
  "n09853305": "bigamist",
  "n09853467": "big brother",
  "n09853541": "Big Brother",
  "n09853645": "bigot",
  "n09853881": "big shot, big gun, big wheel, big cheese, big deal, big enchilada, big fish, head honcho",
  "n09854218": "big sister",
  "n09854290": "bilingual, bilingualist",
  "n09854421": "billiard player",
  "n09854510": "bill poster, poster, bill sticker",
  "n09854708": "bimbo",
  "n09854813": "bimetallist",
  "n09854915": "biochemist",
  "n09855433": "biographer",
  "n09855630": "biologist, life scientist",
  "n09856267": "biophysicist",
  "n09856401": "bird fancier",
  "n09856500": "bird watcher, birder",
  "n09856671": "birth",
  "n09856827": "birth-control campaigner, birth-control reformer",
  "n09857007": "bisexual, bisexual person",
  "n09857200": "bishop",
  "n09857852": "biter",
  "n09857957": "Black and Tan",
  "n09858165": "black belt",
  "n09858299": "blackmailer, extortioner, extortionist",
  "n09858560": "black marketeer",
  "n09858733": "Black Muslim",
  "n09858913": "Black Panther",
  "n09859038": "Blackshirt",
  "n09859152": "blacksmith",
  "n09859285": "blade",
  "n09859412": "blasphemer",
  "n09859557": "blaster, chargeman",
  "n09859684": "bleacher",
  "n09859818": "bleeding heart",
  "n09859975": "blind date",
  "n09860130": "blind person",
  "n09860248": "blocker",
  "n09860415": "blogger",
  "n09860506": "blond, blonde",
  "n09860653": "blood brother",
  "n09860813": "blood donor",
  "n09860940": "blubberer",
  "n09861059": "bludgeoner",
  "n09861166": "blue baby",
  "n09861287": "bluecoat",
  "n09861395": "bluejacket, navy man, sailor, sailor boy",
  "n09861599": "bluestocking, bas bleu",
  "n09861718": "bluffer, four-flusher",
  "n09861863": "boatbuilder",
  "n09861946": "boatman, boater, waterman",
  "n09862183": "boatswain, bos'n, bo's'n, bosun, bo'sun",
  "n09862345": "boarder",
  "n09862479": "boarder",
  "n09862621": "bobby",
  "n09862713": "bobbysoxer, bobby-socker",
  "n09862845": "bodybuilder, muscle builder, muscle-builder, musclebuilder, muscleman",
  "n09863031": "bodyguard, escort",
  "n09863238": "body servant",
  "n09863339": "boffin",
  "n09863500": "bohemian",
  "n09863620": "Bohemian",
  "n09863749": "Bolshevik, Marxist, red, bolshie, bolshy",
  "n09863936": "Bolshevik, Bolshevist",
  "n09864252": "bombardier",
  "n09864429": "bombardier",
  "n09864536": "bomber",
  "n09864632": "bombshell",
  "n09864761": "bondholder",
  "n09864891": "bondman, bondsman",
  "n09864968": "bondman, bondsman",
  "n09865068": "bondwoman, bondswoman, bondmaid",
  "n09865162": "bondwoman, bondswoman, bondmaid",
  "n09865279": "bondsman, bondswoman",
  "n09865398": "bond servant",
  "n09865547": "bonesetter",
  "n09865672": "book agent",
  "n09865744": "bookbinder",
  "n09865838": "bookdealer, book seller",
  "n09865954": "booker, booking agent",
  "n09866115": "bookkeeper",
  "n09866222": "bookmaker, bookie",
  "n09866354": "bookmaker",
  "n09866473": "bookseller",
  "n09866559": "bookworm",
  "n09866661": "booster, shoplifter, lifter",
  "n09866817": "bootblack, shoeblack",
  "n09866922": "bootlegger, moonshiner",
  "n09867069": "bootmaker, boot maker",
  "n09867154": "borderer",
  "n09867311": "border patrolman",
  "n09867437": "bore, dullard",
  "n09867633": "borrower",
  "n09867818": "born-again Christian",
  "n09867956": "boss, hirer",
  "n09868157": "Boswell",
  "n09868270": "botanist, phytologist, plant scientist",
  "n09868703": "bottom dog",
  "n09868782": "bottom feeder",
  "n09868899": "boulevardier",
  "n09869009": "bouncer, chucker-out",
  "n09869171": "bounder, leaper",
  "n09869317": "bounty hunter",
  "n09869447": "bounty hunter",
  "n09869578": "Bourbon",
  "n09869706": "Bourbon",
  "n09869830": "bourgeois, burgher",
  "n09869961": "bowler",
  "n09870096": "bowler",
  "n09870208": "boxer, pugilist",
  "n09870746": "Boxer",
  "n09870926": "boy",
  "n09871095": "slugger, slogger",
  "n09871229": "cub, lad, laddie, sonny, sonny boy",
  "n09871364": "boyfriend, fellow, beau, swain, young man",
  "n09871580": "ex-boyfriend",
  "n09871681": "Boy Scout",
  "n09871867": "boy scout",
  "n09871952": "boy wonder",
  "n09872066": "bragger, braggart, boaster, blowhard, line-shooter, vaunter",
  "n09872285": "bracero",
  "n09872464": "brachycephalic",
  "n09872557": "brahman, brahmin",
  "n09872782": "brahman, brahmin",
  "n09872996": "brainworker, brain-worker",
  "n09873135": "brakeman",
  "n09873242": "brass hat",
  "n09873348": "brawler",
  "n09873473": "breadwinner",
  "n09873604": "breaker, ledgeman",
  "n09873769": "breaststroker",
  "n09873899": "breeder, stock breeder",
  "n09874014": "brewer",
  "n09874118": "brewer, beer maker",
  "n09874260": "briber, suborner",
  "n09874428": "brick",
  "n09874518": "bricklayer",
  "n09874618": "bride",
  "n09874725": "bride",
  "n09874862": "bridesmaid, maid of honor",
  "n09875025": "bridge agent",
  "n09875188": "bridge partner",
  "n09875353": "bridge player, hand",
  "n09875540": "brigadier, brigadier general",
  "n09875663": "broad",
  "n09875786": "broadcaster",
  "n09875979": "broadcast journalist",
  "n09876152": "broker-dealer",
  "n09876308": "broth of a boy, broth of a man",
  "n09876454": "brother, blood brother",
  "n09876701": "Brother",
  "n09876892": "brother",
  "n09877124": "brother, comrade",
  "n09877288": "brother-in-law",
  "n09877370": "Brownie",
  "n09877443": "Brownshirt",
  "n09877587": "browser",
  "n09877750": "Brummie, Brummy",
  "n09877856": "brunet, brunette",
  "n09877951": "buddy, brother, chum, crony, pal, sidekick",
  "n09878179": "bugler",
  "n09878275": "builder, constructor",
  "n09878559": "builder",
  "n09878702": "bull, bruiser, strapper, Samson",
  "n09878921": "bull",
  "n09879144": "bull, cop, copper, fuzz, pig",
  "n09879297": "bully, tough, hooligan, ruffian, roughneck, rowdy, yob, yobo, yobbo",
  "n09879552": "bully",
  "n09879616": "bullyboy",
  "n09879744": "bungler, blunderer, fumbler, bumbler, stumbler, sad sack, botcher, butcher, fuckup",
  "n09880075": "bunkmate",
  "n09880189": "bunny, bunny girl",
  "n09880338": "bunter",
  "n09880427": "bureaucrat, administrative official",
  "n09880646": "burgess, burgher",
  "n09880741": "burglar",
  "n09880881": "burgomaster",
  "n09881009": "burgrave",
  "n09881133": "burgrave",
  "n09881265": "bursar",
  "n09881358": "busboy, waiter's assistant",
  "n09881519": "bushman",
  "n09881613": "Bushman",
  "n09881748": "bushwhacker",
  "n09881895": "business editor",
  "n09882007": "businessman, man of affairs",
  "n09882615": "businesswoman",
  "n09882716": "businessperson, bourgeois",
  "n09883047": "business traveler",
  "n09883174": "busker",
  "n09883362": "busman, bus driver",
  "n09883452": "buster",
  "n09883630": "buster, bronco buster, broncobuster",
  "n09883740": "buster",
  "n09883807": "busybody, nosy-parker, nosey-parker, quidnunc",
  "n09883947": "butch, dike, dyke",
  "n09884133": "butcher, slaughterer",
  "n09884305": "butcher",
  "n09884391": "butcher, meatman",
  "n09884509": "butler, pantryman",
  "n09884666": "butt, goat, laughingstock, stooge",
  "n09884815": "butter",
  "n09884934": "butterfingers",
  "n09885059": "buttinsky",
  "n09885145": "buyer, purchaser, emptor, vendee",
  "n09885334": "bystander",
  "n09885416": "Cabalist, Kabbalist",
  "n09885534": "cabalist, kabbalist",
  "n09885676": "cabalist",
  "n09885769": "cabin boy",
  "n09885866": "cabinetmaker, furniture maker",
  "n09886010": "cabinet minister",
  "n09886220": "cad, bounder, blackguard, dog, hound, heel",
  "n09886403": "caddie, golf caddie",
  "n09886540": "cadet, plebe",
  "n09886700": "caffeine addict, caffein addict",
  "n09886807": "Cairene",
  "n09886911": "caitiff",
  "n09887034": "calculator, reckoner, figurer, estimator, computer",
  "n09887378": "number cruncher",
  "n09887496": "caliph, calif, kaliph, kalif, khalif, khalifah",
  "n09887850": "caller, company",
  "n09888017": "caller, caller-up, phoner, telephoner",
  "n09888269": "caller",
  "n09888429": "caller",
  "n09888635": "caller, caller-out",
  "n09888832": "caller",
  "n09888978": "caller",
  "n09889065": "call girl",
  "n09889170": "calligrapher, calligraphist",
  "n09889346": "Calvinist, Genevan",
  "n09889539": "cameraman, camera operator, cinematographer",
  "n09889691": "campaigner, candidate, nominee",
  "n09889941": "camper",
  "n09890072": "Campfire Girl",
  "n09890192": "camp follower",
  "n09890296": "camp follower",
  "n09890411": "campmate",
  "n09890507": "Canaanite",
  "n09890662": "canary",
  "n09890749": "candidate, prospect",
  "n09890894": "candlemaker",
  "n09890987": "candy striper",
  "n09891079": "cannibal, man-eater, anthropophagus, anthropophagite",
  "n09891300": "cannon fodder, fodder, fresh fish",
  "n09891470": "canoeist, paddler",
  "n09891613": "canon",
  "n09891730": "canonist",
  "n09891864": "cantor, hazan",
  "n09892066": "canvasser",
  "n09892156": "Capetian",
  "n09892262": "capitalist",
  "n09892410": "capo",
  "n09892513": "captain, headwaiter, maitre d'hotel, maitre d'",
  "n09892693": "captain, senior pilot",
  "n09892831": "captain, skipper",
  "n09893015": "captain, police captain, police chief",
  "n09893191": "captain",
  "n09893344": "captain, chieftain",
  "n09893502": "captive",
  "n09893600": "captive",
  "n09893746": "carbineer, carabineer, carabinier",
  "n09893916": "cardholder",
  "n09894040": "cardholder",
  "n09894143": "cardinal",
  "n09894445": "cardiologist, heart specialist, heart surgeon",
  "n09894654": "card player",
  "n09894909": "cardsharp, card sharp, cardsharper, card sharper, sharper, sharpie, sharpy, card shark",
  "n09895138": "career girl",
  "n09895222": "careerist",
  "n09895480": "career man",
  "n09895561": "caregiver",
  "n09895701": "caretaker",
  "n09895902": "caretaker",
  "n09896084": "carhop",
  "n09896170": "caricaturist",
  "n09896311": "carillonneur",
  "n09896401": "caroler, caroller",
  "n09896520": "Carolingian, Carlovingian",
  "n09896685": "carpenter",
  "n09896826": "carper, niggler",
  "n09896973": "carpetbagger",
  "n09897170": "carpet knight",
  "n09897350": "carrier, immune carrier",
  "n09897603": "carrier, newsboy",
  "n09897696": "carrier, bearer, toter",
  "n09897914": "carter",
  "n09898020": "Cartesian",
  "n09898107": "Carthusian",
  "n09898215": "cartographer, map maker",
  "n09898346": "cartoonist",
  "n09898692": "cartwright",
  "n09898797": "Casanova",
  "n09898892": "case",
  "n09899134": "case officer",
  "n09899289": "cashier",
  "n09899438": "castaway, shipwreck survivor",
  "n09899534": "castrato",
  "n09899671": "casualty, injured party",
  "n09899782": "casualty",
  "n09899929": "casuist, sophist",
  "n09900153": "cat",
  "n09900284": "Catalan",
  "n09900391": "cataleptic",
  "n09900499": "cataloger, cataloguer",
  "n09900711": "catalyst",
  "n09900873": "catamite",
  "n09900981": "catch, match",
  "n09901143": "catcher, backstop",
  "n09901337": "catechist",
  "n09901502": "catechumen, neophyte",
  "n09901642": "caterer",
  "n09901786": "Catholicos",
  "n09901921": "cat fancier",
  "n09902017": "cattleman, cow man, beef man",
  "n09902128": "Cavalier, Royalist",
  "n09902256": "cavalier, chevalier",
  "n09902353": "cavalryman, trooper",
  "n09902611": "cavalryman, trooper",
  "n09902731": "caveman, cave man, cave dweller, troglodyte",
  "n09902851": "celebrant",
  "n09902954": "celebrant, celebrator, celebrater",
  "n09903153": "celebrity, famous person",
  "n09903367": "celibate",
  "n09903501": "cellist, violoncellist",
  "n09903639": "censor",
  "n09903936": "censor",
  "n09904057": "census taker, enumerator",
  "n09904208": "centenarian",
  "n09904321": "center, snapper",
  "n09904556": "center",
  "n09904708": "center",
  "n09904837": "centrist, middle of the roader, moderate, moderationist",
  "n09905050": "centurion",
  "n09905185": "certified public accountant, CPA",
  "n09905530": "chachka, tsatske, tshatshke, tchotchke, tchotchkeleh",
  "n09905697": "chain-smoker",
  "n09905842": "chairman of the board",
  "n09905965": "Chaldean, Chaldaean, Chaldee",
  "n09906079": "chamberlain",
  "n09906195": "chamberlain",
  "n09906293": "chambermaid, fille de chambre",
  "n09906449": "chameleon",
  "n09906538": "champion, fighter, hero, paladin",
  "n09906704": "champion, champ, title-holder",
  "n09906848": "chancellor",
  "n09906986": "chancellor, premier, prime minister",
  "n09907196": "Prime Minister, PM, premier",
  "n09907408": "Chancellor of the Exchequer, Chancellor",
  "n09907566": "chandler",
  "n09907717": "wax-chandler",
  "n09907804": "chandler",
  "n09907919": "changeling",
  "n09908025": "chap, fellow, feller, fella, lad, gent, blighter, cuss, bloke",
  "n09908273": "chaperon, chaperone",
  "n09908508": "chaplain",
  "n09908678": "chapman",
  "n09908769": "prison chaplain",
  "n09908853": "chauffeur",
  "n09908986": "chauffeuse",
  "n09909060": "character, eccentric, type, case",
  "n09909363": "character actor",
  "n09909477": "character witness",
  "n09909660": "charcoal burner",
  "n09909760": "charge",
  "n09909929": "charge d'affaires",
  "n09910084": "charge of quarters",
  "n09910222": "charioteer",
  "n09910374": "charmer, beguiler",
  "n09910556": "chartered accountant",
  "n09910719": "charter member",
  "n09910840": "chartist, technical analyst",
  "n09911051": "Chartist",
  "n09911226": "charwoman, char, cleaning woman, cleaning lady, woman",
  "n09911465": "chatelaine",
  "n09911570": "chatterer, babbler, prater, chatterbox, magpie, spouter",
  "n09911849": "chauvinist, jingoist, jingo, flag-waver, hundred-percenter, patrioteer",
  "n09912075": "chauvinist",
  "n09912243": "antifeminist",
  "n09912431": "male chauvinist, sexist",
  "n09912598": "cheapjack",
  "n09912681": "cheapskate, tightwad",
  "n09912765": "chebab",
  "n09912907": "Chechen",
  "n09912995": "checker",
  "n09913110": "checker",
  "n09913240": "check girl, hatcheck girl",
  "n09913329": "cheerer",
  "n09913455": "cheerleader",
  "n09913593": "cheerleader",
  "n09913741": "cheesemonger",
  "n09913824": "chemist",
  "n09915434": "Cheops, Khufu",
  "n09915577": "cherub",
  "n09915651": "chess master",
  "n09915834": "chess player",
  "n09915964": "chewer",
  "n09916109": "chichi",
  "n09916209": "Chief Constable",
  "n09916348": "chief executive officer, CEO, chief operating officer",
  "n09916601": "chief financial officer, CFO",
  "n09916788": "chief justice",
  "n09917214": "chief of staff",
  "n09917345": "chief petty officer",
  "n09917481": "Chief Secretary",
  "n09917593": "child, kid, youngster, minor, shaver, nipper, small fry, tiddler, tike, tyke, fry, nestling",
  "n09918248": "child, kid",
  "n09918554": "child, baby",
  "n09918762": "child",
  "n09918867": "child prodigy, infant prodigy, wonder child",
  "n09919061": "chimneysweeper, chimneysweep, sweep",
  "n09919200": "chiropractor",
  "n09919297": "chiropodist, foot doctor, podiatrist",
  "n09919451": "chit",
  "n09919690": "choirboy",
  "n09919771": "choirmaster, precentor, cantor",
  "n09919899": "choker",
  "n09920106": "choragus",
  "n09920283": "choreographer",
  "n09920771": "chorister",
  "n09920901": "chorus girl, showgirl, chorine",
  "n09921034": "chosen",
  "n09921168": "chronicler",
  "n09921291": "Chukchi",
  "n09921409": "chump, fool, gull, mark, patsy, fall guy, sucker, soft touch, mug",
  "n09921673": "chutzpanik",
  "n09921792": "Church Father, Father of the Church, Father",
  "n09922485": "churchgoer, church member",
  "n09922659": "churchwarden",
  "n09922799": "church officer",
  "n09923003": "cicerone",
  "n09923101": "cigarette smoker",
  "n09923186": "cigar smoker",
  "n09923263": "Cinderella",
  "n09923418": "cipher, cypher, nobody, nonentity",
  "n09923561": "circus acrobat",
  "n09923673": "citizen",
  "n09923996": "city editor",
  "n09924106": "city father",
  "n09924195": "city man",
  "n09924313": "city slicker, city boy",
  "n09924437": "civic leader, civil leader",
  "n09924540": "civil engineer",
  "n09924742": "civilian",
  "n09924854": "civil libertarian",
  "n09924996": "civil rights leader, civil rights worker, civil rights activist",
  "n09925459": "civil servant",
  "n09925592": "claimant",
  "n09925824": "claim jumper",
  "n09925953": "clairvoyant",
  "n09926088": "clapper, applauder",
  "n09926246": "clarinetist, clarinettist",
  "n09926426": "classic",
  "n09926519": "classicist",
  "n09926656": "classicist, classical scholar",
  "n09926862": "classifier",
  "n09926991": "claustrophobe",
  "n09927089": "cleaner",
  "n09927305": "cleaner, dry cleaner",
  "n09927451": "clergyman, reverend, man of the cloth",
  "n09928136": "cleric, churchman, divine, ecclesiastic",
  "n09928364": "clericalist",
  "n09928451": "clerk",
  "n09928845": "clever Dick, clever clogs",
  "n09929005": "cliff dweller",
  "n09929202": "climatologist",
  "n09929298": "climber",
  "n09929577": "clinician",
  "n09929770": "clip artist",
  "n09929861": "cloakmaker, furrier",
  "n09929988": "clock watcher",
  "n09930102": "clocksmith, clockmaker",
  "n09930257": "closer, finisher",
  "n09930464": "closer",
  "n09930628": "closet queen",
  "n09930772": "clothier, haberdasher",
  "n09930876": "clown, buffoon, goof, goofball, merry andrew",
  "n09931165": "clown, buffoon",
  "n09931267": "clumsy person",
  "n09931418": "coach, private instructor, tutor",
  "n09931640": "coach, manager, handler",
  "n09931989": "line coach",
  "n09932098": "pitching coach",
  "n09932227": "coachbuilder",
  "n09932336": "coachman",
  "n09932429": "coalman",
  "n09932508": "coal miner, collier, pitman",
  "n09932616": "coaster",
  "n09932706": "coaster",
  "n09932788": "coastguardsman",
  "n09932892": "coauthor, joint author",
  "n09933020": "cobber",
  "n09933098": "cobbler, shoemaker",
  "n09933235": "cocaine addict",
  "n09933324": "cocksucker",
  "n09933411": "codefendant, co-defendant",
  "n09933613": "codetalker, windtalker",
  "n09933842": "codger, old codger",
  "n09933972": "co-beneficiary",
  "n09934085": "co-discoverer",
  "n09934213": "co-ed, college girl",
  "n09934337": "cog",
  "n09934488": "cognitive neuroscientist",
  "n09934647": "cognitive scientist",
  "n09934774": "coiffeur",
  "n09934846": "coiffeuse",
  "n09934921": "coiner, minter, moneyer",
  "n09935107": "coiner",
  "n09935233": "coiner",
  "n09935351": "cold fish",
  "n09935434": "collaborator, cooperator, partner, pardner",
  "n09935793": "collaborator, collaborationist, quisling",
  "n09935990": "colleague, confrere, fellow",
  "n09936215": "colleague, co-worker, fellow worker, workfellow",
  "n09936362": "collector, gatherer, accumulator",
  "n09936620": "collector, aggregator",
  "n09936825": "colleen",
  "n09936892": "college student, university student",
  "n09937056": "collegian, college man, college boy",
  "n09937250": "colonel",
  "n09937489": "Colonel Blimp, Blimp",
  "n09937688": "colonial",
  "n09937802": "colonialist",
  "n09937903": "colonizer, coloniser",
  "n09938080": "coloratura, coloratura soprano",
  "n09938272": "color bearer, standard-bearer",
  "n09938449": "color guard",
  "n09938554": "color sergeant",
  "n09938672": "colorist",
  "n09938851": "Colossian",
  "n09938991": "colossus, behemoth, giant, heavyweight, titan",
  "n09939154": "columnist, editorialist",
  "n09939313": "combatant, battler, belligerent, fighter, scrapper",
  "n09939827": "combat pilot",
  "n09940026": "comber",
  "n09940146": "comedian, comic",
  "n09940725": "comedian",
  "n09940818": "comedienne",
  "n09940987": "comedienne",
  "n09941089": "comer",
  "n09941172": "comfort woman, ianfu",
  "n09941383": "commander",
  "n09941571": "commander",
  "n09941787": "commander in chief, generalissimo",
  "n09941964": "commanding officer, commandant, commander",
  "n09942275": "commando, ranger",
  "n09942431": "commentator, reviewer",
  "n09942587": "commercial artist",
  "n09942697": "commissar, political commissar",
  "n09942871": "commissionaire",
  "n09942970": "commissioned officer",
  "n09943239": "commissioned military officer",
  "n09943541": "commissioned naval officer",
  "n09943811": "commissioner",
  "n09944022": "commissioner",
  "n09944160": "committee member",
  "n09944337": "committeeman",
  "n09944430": "committeewoman",
  "n09944529": "couch potato",
  "n09944677": "councilman",
  "n09944763": "council member, councillor",
  "n09944931": "councilwoman",
  "n09945021": "commodore",
  "n09945223": "communicant",
  "n09945319": "communist, commie",
  "n09945603": "Communist",
  "n09945745": "commuter",
  "n09945905": "companion, comrade, fellow, familiar, associate",
  "n09946278": "companion, fellow traveler, fellow traveller",
  "n09946437": "company man",
  "n09946574": "company operator",
  "n09946672": "comparative anatomist",
  "n09946814": "compere",
  "n09946957": "compiler",
  "n09947127": "complexifier",
  "n09947232": "composer",
  "n09949946": "compositor, typesetter, setter, typographer",
  "n09950150": "Comptroller General",
  "n09950318": "Comptroller of the Currency",
  "n09950457": "compulsive",
  "n09950728": "computational linguist",
  "n09950917": "computer expert, computer guru",
  "n09951070": "computer scientist",
  "n09951274": "computer user",
  "n09951524": "Comrade",
  "n09951616": "concert-goer, music lover",
  "n09951717": "concessionaire, concessioner",
  "n09951835": "conchologist",
  "n09951953": "concierge",
  "n09952163": "conciliator, make-peace, pacifier, peacemaker, reconciler",
  "n09952393": "concubine, courtesan, doxy, paramour",
  "n09952539": "conductor, music director, director",
  "n09953052": "conductor",
  "n09953178": "conditioner",
  "n09953275": "conductress",
  "n09953350": "confectioner, candymaker",
  "n09953483": "confederate, collaborator, henchman, partner in crime",
  "n09953615": "Confederate",
  "n09953775": "Confederate soldier",
  "n09953965": "conferee",
  "n09954081": "conferee",
  "n09954246": "conferrer",
  "n09954355": "confessor",
  "n09954479": "confessor",
  "n09954639": "confidant, intimate",
  "n09954804": "confidante",
  "n09954879": "confidence man, con man, con artist",
  "n09955015": "swindler, defrauder, chiseller, chiseler, gouger, scammer, grifter",
  "n09955406": "Confucian, Confucianist",
  "n09955517": "congregant",
  "n09955643": "Congregationalist",
  "n09955781": "congressman, congresswoman, representative",
  "n09955944": "rep",
  "n09956035": "connection",
  "n09956147": "connection",
  "n09956387": "connoisseur, cognoscente",
  "n09956578": "conqueror, vanquisher",
  "n09956780": "conquistador",
  "n09957013": "conscientious objector, CO",
  "n09957156": "conservative, conservativist",
  "n09957523": "Conservative",
  "n09957614": "conformist",
  "n09957834": "nonconformist, recusant",
  "n09958133": "Nonconformist, chapelgoer",
  "n09958292": "Anglican",
  "n09958447": "consignee",
  "n09958569": "consigner, consignor",
  "n09958724": "consort",
  "n09958892": "conspirator, coconspirator, plotter, machinator",
  "n09959142": "constable",
  "n09959258": "constable, police constable",
  "n09959387": "constitutionalist",
  "n09959527": "construction worker, hard hat",
  "n09959658": "constructivist",
  "n09959797": "consul",
  "n09960001": "consumptive, lunger, tubercular",
  "n09960117": "contact, middleman",
  "n09960315": "contemplative",
  "n09960417": "contemporary, coeval",
  "n09960545": "contortionist",
  "n09960688": "contractor",
  "n09960891": "contractor",
  "n09961012": "contractor, declarer",
  "n09961198": "contralto",
  "n09961331": "contributor",
  "n09961469": "control freak",
  "n09961605": "convalescent",
  "n09961739": "convener",
  "n09961889": "conventioneer",
  "n09961999": "conversationalist, conversationist, schmoozer",
  "n09962214": "Converso",
  "n09962414": "convert",
  "n09962612": "conveyancer",
  "n09962789": "conveyer, conveyor",
  "n09962966": "convict, con, inmate, yard bird, yardbird",
  "n09963159": "convict",
  "n09963320": "cook",
  "n09963574": "chef",
  "n09963680": "cookie, cooky",
  "n09963773": "cooper, barrel maker",
  "n09963914": "coordinator",
  "n09964064": "copartner",
  "n09964202": "copilot, co-pilot",
  "n09964315": "coppersmith",
  "n09964411": "copycat, imitator, emulator, ape, aper",
  "n09964659": "copy editor, copyreader, text editor",
  "n09964805": "copyist, scribe, scrivener",
  "n09965021": "copywriter",
  "n09965134": "coquette, flirt, vamp, vamper, minx, tease, prickteaser",
  "n09965424": "cordon bleu",
  "n09965515": "coreligionist",
  "n09965625": "corespondent, co-respondent",
  "n09965787": "cornerback",
  "n09965905": "cornhusker",
  "n09965985": "coroner, medical examiner",
  "n09966139": "corporal",
  "n09966255": "corporate executive, business executive",
  "n09966470": "corporatist",
  "n09966554": "correspondent, letter writer",
  "n09966710": "correspondent, newspaperman, newspaperwoman, newswriter, pressman",
  "n09966941": "corsair, Barbary pirate",
  "n09967063": "cosmetician",
  "n09967159": "cosmetologist",
  "n09967270": "cosmetic surgeon, plastic surgeon",
  "n09967406": "cosmopolitan, cosmopolite",
  "n09967555": "Cossack",
  "n09967816": "cost accountant",
  "n09967967": "co-star",
  "n09968128": "costermonger, barrow-man, barrow-boy",
  "n09968259": "costumier, costumer, costume designer",
  "n09968433": "cotenant",
  "n09968549": "cottager, cottage dweller",
  "n09968652": "cotter, cottier",
  "n09968741": "cotter, cottar",
  "n09968845": "counselor, counsellor",
  "n09969062": "counselor, counsellor",
  "n09969218": "count",
  "n09969375": "count palatine",
  "n09969491": "counter",
  "n09969589": "counterdemonstrator",
  "n09969718": "counterperson, counterwoman, counterman",
  "n09969869": "counterrevolutionist, counter-revolutionist, counterrevolutionary",
  "n09970088": "counterterrorist",
  "n09970192": "counterspy, mole",
  "n09970295": "countertenor",
  "n09970402": "countess",
  "n09970493": "country doctor",
  "n09970699": "compatriot",
  "n09970822": "compromiser",
  "n09970963": "countryman",
  "n09971047": "countrywoman",
  "n09971135": "countryman, ruralist",
  "n09971273": "countrywoman",
  "n09971385": "county agent, agricultural agent, extension agent",
  "n09971595": "coureur de bois",
  "n09971682": "courser",
  "n09971839": "courtier",
  "n09972010": "cousin, first cousin, cousin-german, full cousin",
  "n09972157": "couturier, fashion designer, clothes designer, designer",
  "n09972458": "cover girl, pin-up, lovely",
  "n09972587": "cow",
  "n09972661": "cowboy, cowpuncher, puncher, cowman, cattleman, cowpoke, cowhand, cowherd",
  "n09972946": "cowboy",
  "n09973072": "cowboy, rodeo rider",
  "n09973209": "vaquero, buckaroo, buckeroo",
  "n09973422": "cowgirl",
  "n09973490": "coxcomb, cockscomb",
  "n09973624": "coxswain, cox",
  "n09973749": "coyote",
  "n09973903": "coyote",
  "n09974054": "crab, crabby person",
  "n09974177": "crack addict, binger",
  "n09974278": "cracker",
  "n09974496": "crackpot, crank, nut, nut case, fruitcake, screwball",
  "n09974648": "craftsman, artisan, journeyman, artificer",
  "n09975425": "craftsman, crafter",
  "n09975630": "craftsman",
  "n09975806": "crammer",
  "n09975933": "crammer",
  "n09976024": "crapshooter",
  "n09976119": "crawler, creeper",
  "n09976283": "crazy, loony, looney, nutcase, weirdo",
  "n09976429": "creature, wight",
  "n09976551": "creature, tool, puppet",
  "n09976728": "creditor",
  "n09976917": "creep, weirdo, weirdie, weirdy, spook",
  "n09977082": "crewman",
  "n09977178": "crewman, crew member",
  "n09977326": "cricketer",
  "n09977520": "crier",
  "n09977660": "criminal, felon, crook, outlaw, malefactor",
  "n09978442": "criminologist",
  "n09978566": "crimp, crimper",
  "n09978697": "criollo",
  "n09978889": "cripple",
  "n09979072": "critic",
  "n09979321": "critic",
  "n09979589": "critic",
  "n09979913": "Croesus",
  "n09979985": "crofter",
  "n09980090": "crooner, balladeer",
  "n09980275": "crossbencher",
  "n09980458": "cross-examiner, cross-questioner",
  "n09980658": "crossing guard",
  "n09980805": "crossover voter, crossover",
  "n09980985": "croupier",
  "n09981092": "crown prince",
  "n09981183": "crown princess",
  "n09981278": "crown princess",
  "n09981365": "Crusader",
  "n09981540": "cryptanalyst, cryptographer, cryptologist",
  "n09981740": "crystallographer",
  "n09981834": "cub, greenhorn, rookie",
  "n09981939": "Cub Scout",
  "n09982013": "cubist",
  "n09982152": "cuckold",
  "n09982277": "cuirassier",
  "n09982370": "cultist",
  "n09982525": "cultist",
  "n09982760": "cultural attache",
  "n09982873": "cunt, bitch",
  "n09983053": "cupbearer",
  "n09983214": "cur",
  "n09983314": "curandera",
  "n09983444": "curandero",
  "n09983572": "curate, minister of religion, minister, parson, pastor, rector",
  "n09983889": "curator, conservator",
  "n09984047": "curmudgeon",
  "n09984187": "currier",
  "n09984298": "custodian, keeper, steward",
  "n09984659": "customer, client",
  "n09984960": "customer agent",
  "n09985075": "client",
  "n09985207": "cutler",
  "n09985279": "cutter",
  "n09985470": "cutter, carver",
  "n09985577": "cutthroat",
  "n09985683": "cybernaut",
  "n09985809": "cyberpunk",
  "n09985978": "cyborg, bionic man, bionic woman",
  "n09986189": "cyclist, bicyclist, bicycler, wheeler",
  "n09986450": "cymbalist",
  "n09986532": "cynic, faultfinder",
  "n09986700": "Cynic",
  "n09986904": "cytogeneticist",
  "n09987045": "cytologist",
  "n09987161": "czar",
  "n09987239": "czar, tsar, tzar",
  "n09987573": "czarina, tsarina, tzarina, czaritza, tsaritsa",
  "n09987696": "dabbler, dilettante, sciolist",
  "n09987927": "dacoit, dakoit",
  "n09988063": "dad, dada, daddy, pa, papa, pappa, pop",
  "n09988216": "dairymaid, milkmaid",
  "n09988311": "dairyman",
  "n09988392": "dairyman, dairy farmer",
  "n09988493": "Dalai Lama, Grand Lama",
  "n09988597": "dalesman",
  "n09988703": "dallier, dillydallier, dilly-dallier, mope, lounger",
  "n09988918": "Dalmatian",
  "n09989045": "dame, doll, wench, skirt, chick, bird",
  "n09989168": "damsel, demoiselle, damoiselle, damosel, damozel",
  "n09989290": "dame, madam, ma'am, lady, gentlewoman",
  "n09989502": "dancer, professional dancer, terpsichorean",
  "n09990415": "dancer, social dancer",
  "n09990690": "clog dancer",
  "n09990777": "dancing-master, dance master",
  "n09990904": "dancing partner",
  "n09991026": "dandy, dude, fop, gallant, sheik, beau, swell, fashion plate, clotheshorse",
  "n09991285": "Daniel",
  "n09991406": "danseur, danseur noble",
  "n09991530": "daredevil, madcap, hothead, swashbuckler, lunatic, harum-scarum",
  "n09991740": "dark horse",
  "n09991867": "darling, favorite, favourite, pet, dearie, deary, ducky",
  "n09992138": "darner",
  "n09992238": "dart player",
  "n09992331": "Darwinian",
  "n09992429": "dastard",
  "n09992538": "date, escort",
  "n09992746": "dauber",
  "n09992837": "daughter, girl",
  "n09993040": "daughter-in-law",
  "n09993122": "dauphin",
  "n09993252": "dawdler, drone, laggard, lagger, trailer, poke",
  "n09993651": "day boarder",
  "n09993822": "dayboy",
  "n09993901": "daydreamer, woolgatherer",
  "n09994119": "lotus-eater, stargazer",
  "n09994319": "daygirl",
  "n09994400": "day laborer, day labourer",
  "n09994520": "deacon",
  "n09994673": "deacon, Protestant deacon",
  "n09994808": "deaconess",
  "n09994878": "deadeye",
  "n09994943": "dead person, dead soul, deceased person, deceased, decedent, departed",
  "n09995253": "decipherer",
  "n09995398": "decoder, decipherer",
  "n09995573": "decoy, steerer",
  "n09995757": "deer hunter",
  "n09995829": "deipnosophist",
  "n09995925": "dropout",
  "n09996039": "dropout",
  "n09996149": "deadbeat dad",
  "n09996304": "deadhead",
  "n09996481": "deaf person",
  "n09996636": "dealer",
  "n09996784": "dean",
  "n09996920": "dean",
  "n09997068": "dean, doyen",
  "n09997212": "debaser, degrader",
  "n09997404": "debater, arguer",
  "n09997622": "debtor, debitor",
  "n09997834": "debutante, deb",
  "n09997939": "decadent",
  "n09998101": "deceiver, cheat, cheater, trickster, beguiler, slicker",
  "n09998788": "deckhand, roustabout",
  "n09998907": "decorator, ornamentalist",
  "n09999036": "deep-sea diver",
  "n09999135": "defamer, maligner, slanderer, vilifier, libeler, backbiter, traducer",
  "n09999443": "defaulter",
  "n09999532": "defaulter, deadbeat",
  "n09999683": "defaulter",
  "n09999795": "defeatist, negativist",
  "n10000007": "defecator, voider, shitter",
  "n10000158": "defense attorney, defense lawyer",
  "n10000294": "defense contractor",
  "n10000459": "deist, freethinker",
  "n10000616": "delayer",
  "n10000787": "delegate",
  "n10000945": "delinquent, juvenile delinquent",
  "n10001058": "deliverer",
  "n10001217": "deliveryman, delivery boy, deliverer",
  "n10001481": "demagogue, demagog, rabble-rouser",
  "n10001647": "demander",
  "n10001764": "demigod, superman, Ubermensch",
  "n10001882": "demimondaine",
  "n10002031": "democrat, populist",
  "n10002151": "Democrat",
  "n10002257": "demographer, demographist, population scientist",
  "n10002477": "demon",
  "n10002644": "demoniac",
  "n10002760": "demonstrator, protester",
  "n10002982": "demonstrator, sales demonstrator",
  "n10003120": "demonstrator",
  "n10003283": "denier",
  "n10003476": "den mother",
  "n10003575": "den mother",
  "n10003782": "dental assistant",
  "n10003870": "dental hygienist",
  "n10004019": "dental technician, denturist",
  "n10004171": "dental surgeon",
  "n10004282": "dentist, tooth doctor, dental practitioner",
  "n10004539": "departer, leaver, goer",
  "n10004718": "department head",
  "n10004804": "dependant, dependent",
  "n10005006": "depositor",
  "n10005163": "depressive",
  "n10005280": "deputy, lieutenant",
  "n10005548": "deputy, deputy sheriff",
  "n10005721": "deputy, surrogate",
  "n10005934": "deputy",
  "n10006081": "derelict",
  "n10006177": "dermatologist, skin doctor",
  "n10006337": "dervish",
  "n10006511": "descendant, descendent",
  "n10006748": "descender",
  "n10006842": "deserter, defector",
  "n10007109": "deserter, apostate, renegade, turncoat, recreant, ratter",
  "n10007511": "designated driver",
  "n10007684": "designated hitter",
  "n10007809": "designer, intriguer",
  "n10007995": "desk clerk, hotel desk clerk, hotel clerk",
  "n10008123": "desk officer",
  "n10008254": "desk sergeant, deskman, station keeper",
  "n10008388": "desperado, desperate criminal",
  "n10008535": "desperate",
  "n10008716": "destroyer, ruiner, undoer, waster, uprooter",
  "n10009162": "detainee, political detainee",
  "n10009276": "detective, investigator, tec, police detective",
  "n10009484": "detective",
  "n10009671": "detractor, disparager, depreciator, knocker",
  "n10009926": "deus ex machina",
  "n10010062": "developer",
  "n10010243": "deviationist",
  "n10010400": "devil's advocate",
  "n10010525": "devil worshiper",
  "n10010632": "devisee",
  "n10010767": "devisor",
  "n10010864": "devourer",
  "n10010977": "diabetic",
  "n10011074": "diagnostician, pathologist",
  "n10011360": "dialectician",
  "n10011486": "diarist, diary keeper, journalist",
  "n10011659": "dichromat",
  "n10011785": "dick, gumshoe, hawkshaw",
  "n10011902": "dictator, potentate",
  "n10012244": "dictator",
  "n10012377": "dieter",
  "n10012484": "dietician, dietitian, nutritionist",
  "n10012713": "diemaker, diesinker, die-sinker",
  "n10012815": "differentiator, discriminator",
  "n10012989": "digger",
  "n10013114": "dimwit, nitwit, half-wit, doofus",
  "n10013242": "diner",
  "n10013399": "dingbat",
  "n10013614": "dining-room attendant, restaurant attendant",
  "n10013811": "diocesan",
  "n10013927": "diplomat, diplomatist",
  "n10014658": "diplomat",
  "n10014771": "diplomate",
  "n10014939": "director, manager, managing director",
  "n10015215": "director, theater director, theatre director",
  "n10015485": "director",
  "n10015607": "Director of Central Intelligence, DCI",
  "n10015792": "dirty old man",
  "n10015897": "disbeliever, nonbeliever, unbeliever",
  "n10016103": "disciple, adherent",
  "n10016954": "disentangler, unraveler, unraveller",
  "n10017190": "dishwasher",
  "n10017272": "disk jockey, disc jockey, dj",
  "n10017422": "dispatcher",
  "n10017664": "dispatch rider",
  "n10017794": "dispenser",
  "n10017890": "displaced person, DP, stateless person",
  "n10018021": "dissenter, dissident, protester, objector, contestant",
  "n10018373": "disturber",
  "n10018532": "political dissident",
  "n10018653": "distiller",
  "n10018747": "distortionist",
  "n10018861": "distributor, distributer",
  "n10019072": "district attorney, DA",
  "n10019187": "district manager",
  "n10019308": "ditch digger, mud digger",
  "n10019406": "diver, plunger",
  "n10019552": "diver, frogman, underwater diver",
  "n10019733": "divergent thinker",
  "n10019888": "divider",
  "n10020031": "diviner",
  "n10020366": "divorcee, grass widow",
  "n10020533": "ex-wife, ex",
  "n10020670": "divorce lawyer",
  "n10020807": "docent",
  "n10020890": "doctor, doc, physician, MD, Dr., medico",
  "n10021892": "doctor, Dr.",
  "n10022111": "Doctor of the Church, Doctor",
  "n10022645": "dodderer",
  "n10022759": "dodger, fox, slyboots",
  "n10022908": "dodo, fogy, fogey, fossil",
  "n10023039": "dog",
  "n10023129": "dog catcher",
  "n10023264": "doge",
  "n10023381": "dogfighter",
  "n10023506": "dog in the manger",
  "n10023656": "dogmatist, doctrinaire",
  "n10023885": "dogsbody",
  "n10024025": "dolichocephalic",
  "n10024119": "domestic, domestic help, house servant",
  "n10024362": "domestic partner, significant other, spousal equivalent, spouse equivalent",
  "n10024621": "domestic prelate",
  "n10024784": "dominatrix",
  "n10024937": "Dominican",
  "n10025060": "dominus, dominie, domine, dominee",
  "n10025195": "Don",
  "n10025295": "don, father",
  "n10025391": "Donatist",
  "n10025487": "Don Juan",
  "n10025635": "donna",
  "n10025730": "donor, giver, presenter, bestower, conferrer",
  "n10026058": "donor",
  "n10026261": "Don Quixote",
  "n10026367": "don't-know",
  "n10026553": "doorkeeper, doorman, door guard, hall porter, porter, gatekeeper, ostiary",
  "n10026763": "doorkeeper, ostiary, ostiarius",
  "n10026976": "dosser, street person",
  "n10027105": "dotard",
  "n10027246": "double, image, look-alike",
  "n10027476": "double agent",
  "n10027590": "double-crosser, double-dealer, two-timer, betrayer, traitor",
  "n10027798": "double dipper",
  "n10027953": "doubting Thomas",
  "n10028123": "dove, peacenik",
  "n10028289": "dowager",
  "n10028402": "down-and-out",
  "n10028541": "doyenne",
  "n10028638": "draft dodger, draft evader",
  "n10028765": "draftee, conscript, inductee",
  "n10028977": "drafter",
  "n10029068": "draftsman, drawer",
  "n10029269": "draftsman, draughtsman, draftsperson",
  "n10029429": "dragoman",
  "n10029729": "dragon, tartar",
  "n10029831": "dragoon",
  "n10029985": "redcoat, lobsterback",
  "n10030147": "drama critic, theater critic",
  "n10030277": "dramatist, playwright",
  "n10032190": "draper",
  "n10032342": "drawee",
  "n10032524": "drawer",
  "n10032676": "drawing card, draw, attraction, attractor, attracter",
  "n10032884": "drawler",
  "n10032987": "dreamer",
  "n10033082": "dresser, actor's assistant",
  "n10033225": "dresser",
  "n10033412": "dressmaker, modiste, needlewoman, seamstress, sempstress",
  "n10033572": "dressmaker's model",
  "n10033663": "dribbler, driveller, slobberer, drooler",
  "n10033888": "dribbler",
  "n10034020": "drill master, drill instructor",
  "n10034201": "drinker, imbiber, toper, juicer",
  "n10034614": "drinker",
  "n10034785": "driveller, jabberer",
  "n10034906": "driver",
  "n10035314": "driver",
  "n10035430": "driver",
  "n10035655": "dropkicker",
  "n10035809": "drudge, peon, navvy, galley slave",
  "n10035952": "drug addict, junkie, junky",
  "n10036135": "drug baron, drug lord",
  "n10036266": "drug user, substance abuser, user",
  "n10036444": "Druid",
  "n10036574": "drum major",
  "n10036692": "drum majorette, majorette",
  "n10036802": "drum majorette, majorette",
  "n10036929": "drummer",
  "n10037080": "drunk",
  "n10037194": "drunk-and-disorderly",
  "n10037385": "drunkard, drunk, rummy, sot, inebriate, wino",
  "n10037588": "Druze, Druse",
  "n10037922": "dry, prohibitionist",
  "n10038119": "dry nurse",
  "n10038226": "dualist",
  "n10038320": "duce",
  "n10038409": "duchess",
  "n10038547": "duck hunter",
  "n10038620": "duke",
  "n10038778": "duke",
  "n10038929": "dueler, dueller, duelist, duellist",
  "n10039164": "duenna",
  "n10039271": "duffer",
  "n10039391": "dumbbell, dummy, dope, boob, booby, pinhead",
  "n10039569": "dummy, silent person",
  "n10039663": "dunce, dunderhead, numskull, blockhead, bonehead, lunkhead, hammerhead, knucklehead, loggerhead, muttonhead, shithead, dumbass, fuckhead",
  "n10039946": "dunker",
  "n10040049": "dunker",
  "n10040240": "Dutch uncle",
  "n10040344": "dwarf, midget, nanus",
  "n10040515": "dyer",
  "n10040617": "dyslectic",
  "n10040698": "dyspeptic",
  "n10040789": "dynamiter, dynamitist",
  "n10040945": "eager beaver, busy bee, live wire, sharpie, sharpy",
  "n10041093": "Eagle Scout",
  "n10041195": "ear doctor, ear specialist, otologist",
  "n10041373": "earl",
  "n10041539": "Earl Marshal",
  "n10041684": "early bird",
  "n10041787": "early bird",
  "n10041887": "earner, wage earner",
  "n10042074": "easterner",
  "n10042186": "East-sider",
  "n10042300": "eater, feeder",
  "n10042690": "eavesdropper",
  "n10042845": "eccentric, eccentric person, flake, oddball, geek",
  "n10043024": "eclectic, eclecticist",
  "n10043163": "ecologist",
  "n10043331": "economic libertarian",
  "n10043491": "econometrician, econometrist",
  "n10043643": "economist, economic expert",
  "n10044470": "economizer, economiser",
  "n10044682": "ectomorph",
  "n10044763": "edger",
  "n10044879": "editor, editor in chief",
  "n10045361": "subeditor",
  "n10045454": "educationist, educationalist",
  "n10045713": "educator, pedagogue, pedagog",
  "n10046387": "Edwardian",
  "n10046527": "effecter, effector",
  "n10046717": "efficiency expert, efficiency engineer",
  "n10046870": "egalitarian, equalitarian",
  "n10047030": "egghead",
  "n10047199": "egocentric, egoist",
  "n10047371": "egomaniac",
  "n10047459": "egotist, egoist, swellhead",
  "n10047822": "Egyptologist",
  "n10048001": "ejaculator",
  "n10048117": "ejaculator",
  "n10048218": "elder, senior",
  "n10048367": "elder",
  "n10048485": "elder statesman",
  "n10048612": "elder statesman",
  "n10048729": "eldest hand, elder hand",
  "n10048836": "elected official",
  "n10049017": "electrical engineer",
  "n10049363": "electrician, lineman, linesman",
  "n10049522": "electrocutioner",
  "n10049648": "electrologist",
  "n10049788": "electroplater",
  "n10049896": "electrotherapist",
  "n10050043": "elegist",
  "n10050163": "elevator girl",
  "n10050261": "elevator man, elevator boy, liftman",
  "n10050432": "elevator operator",
  "n10050558": "elitist",
  "n10050712": "Elizabethan",
  "n10050880": "elocutionist",
  "n10051026": "emancipator, manumitter",
  "n10051215": "embalmer",
  "n10051337": "embezzler, defalcator, peculator",
  "n10051552": "embroiderer",
  "n10051683": "embroideress",
  "n10051761": "embryologist",
  "n10051861": "emeritus",
  "n10051975": "emigrant, emigre, emigree, outgoer",
  "n10052128": "Emile",
  "n10052244": "eminence grise",
  "n10052497": "emir, amir, emeer, ameer",
  "n10052694": "emissary, envoy",
  "n10052843": "emotional person",
  "n10053004": "emperor",
  "n10053439": "empress",
  "n10053591": "empiricist",
  "n10053708": "employable",
  "n10053808": "employee",
  "n10054657": "employer",
  "n10054875": "employment agent",
  "n10054975": "empty nester",
  "n10055085": "enchanter",
  "n10055181": "conjurer, conjuror, conjure man",
  "n10055297": "enchantress, witch",
  "n10055410": "enchantress, temptress, siren, Delilah, femme fatale",
  "n10055566": "encyclopedist, encyclopaedist",
  "n10055730": "endomorph",
  "n10055847": "enemy, foe, foeman, opposition",
  "n10056103": "energizer, energiser, vitalizer, vitaliser, animator",
  "n10056398": "end",
  "n10056611": "end man",
  "n10056719": "end man, corner man",
  "n10056914": "endocrinologist",
  "n10057114": "endodontist",
  "n10057271": "endorser, indorser",
  "n10057491": "end user",
  "n10057595": "enfant terrible",
  "n10057714": "engineer, locomotive engineer, railroad engineer, engine driver",
  "n10057918": "English teacher, English professor",
  "n10058027": "engraver",
  "n10058155": "engraver",
  "n10058411": "enjoyer",
  "n10058585": "enlisted man",
  "n10058777": "enlisted person",
  "n10058962": "enlisted woman",
  "n10059067": "enophile, oenophile",
  "n10059162": "enrollee",
  "n10059323": "ENT man, ear-nose-and-throat doctor, otolaryngologist, otorhinolaryngologist, rhinolaryngologist",
  "n10059582": "enthusiast, partisan, partizan",
  "n10059904": "entomologist, bugologist, bug-hunter",
  "n10060075": "entrant",
  "n10060175": "entrant",
  "n10060352": "entrepreneur, enterpriser",
  "n10060621": "environmentalist, conservationist",
  "n10060904": "Green",
  "n10061043": "envoy, envoy extraordinary, minister plenipotentiary",
  "n10061195": "enzymologist",
  "n10061323": "eparch",
  "n10061431": "eparch",
  "n10061554": "Ephesian",
  "n10061656": "epicure, gourmet, gastronome, bon vivant, epicurean, foodie",
  "n10061882": "epidemiologist",
  "n10062042": "epigone, epigon",
  "n10062176": "epileptic",
  "n10062275": "Episcopalian",
  "n10062385": "epistemologist",
  "n10062492": "equerry",
  "n10062594": "equerry",
  "n10062716": "erotic",
  "n10062785": "escalader",
  "n10062905": "escapee",
  "n10062996": "escapist, dreamer, wishful thinker",
  "n10063177": "escapologist, escape expert",
  "n10063340": "eschatologist",
  "n10063461": "escort",
  "n10063635": "Eskimo, Esquimau, Inuit",
  "n10063919": "espionage agent",
  "n10064046": "Esquire, Esq",
  "n10064229": "esquire",
  "n10064405": "essayist, litterateur",
  "n10064537": "esthete, aesthete",
  "n10064669": "esthetician, aesthetician",
  "n10064831": "esthetician, aesthetician",
  "n10064977": "etcher",
  "n10065066": "ethicist, ethician",
  "n10065261": "ethnarch",
  "n10065547": "ethnic",
  "n10065643": "ethnographer",
  "n10065758": "ethnologist",
  "n10065911": "ethologist",
  "n10066059": "etiologist, aetiologist",
  "n10066206": "Etonian",
  "n10066314": "etymologist",
  "n10066452": "eunuch, castrate",
  "n10066624": "evacuee",
  "n10066732": "evaluator, judge",
  "n10067011": "evangelist, revivalist, gospeler, gospeller",
  "n10067305": "Evangelist",
  "n10067600": "event planner",
  "n10067793": "everyman",
  "n10067867": "evolutionist",
  "n10067968": "examiner, inspector",
  "n10068234": "examiner, tester, quizzer",
  "n10068425": "exarch",
  "n10068537": "exarch",
  "n10068682": "exarch",
  "n10068804": "excogitator",
  "n10068928": "Excellency",
  "n10069120": "exchanger, money changer",
  "n10069296": "executant",
  "n10069427": "executioner, public executioner",
  "n10069645": "executive, executive director",
  "n10069869": "executive officer",
  "n10069981": "executive secretary",
  "n10070108": "executive vice president",
  "n10070219": "executor",
  "n10070377": "executrix",
  "n10070449": "exegete",
  "n10070563": "exhibitor, exhibitioner, shower",
  "n10070711": "exhibitionist, show-off",
  "n10070942": "exhibitionist, flasher",
  "n10071139": "exile, deportee",
  "n10071332": "exile, expatriate, expat",
  "n10071557": "existentialist, existentialist philosopher, existential philosopher",
  "n10071912": "exodontist",
  "n10072054": "exorcist, exorciser",
  "n10072187": "exorcist",
  "n10072346": "expert witness",
  "n10072546": "exploiter, user",
  "n10072708": "explorer, adventurer",
  "n10073634": "exporter",
  "n10073762": "expositor, expounder",
  "n10073851": "expressionist",
  "n10073992": "expurgator, bowdlerizer, bowdleriser",
  "n10074249": "ex-spouse",
  "n10074339": "exterminator, terminator, eradicator",
  "n10074578": "extern, medical extern",
  "n10074735": "extremist",
  "n10074841": "extrovert, extravert",
  "n10075063": "eyeful",
  "n10075185": "eyeglass wearer",
  "n10075299": "eyewitness",
  "n10075416": "Fabian",
  "n10075529": "fabulist",
  "n10075693": "facilitator",
  "n10075802": "factotum",
  "n10075899": "faddist",
  "n10076033": "fagot, faggot, fag, fairy, nance, pansy, queen, queer, poof, poove, pouf",
  "n10076224": "fairy godmother",
  "n10076307": "fakir, fakeer, faqir, faquir",
  "n10076483": "falangist, phalangist",
  "n10076604": "falconer, hawker",
  "n10076778": "faller",
  "n10076957": "falsifier",
  "n10077106": "familiar",
  "n10077278": "family doctor",
  "n10077394": "family man",
  "n10077504": "famulus",
  "n10077593": "fan, buff, devotee, lover",
  "n10077879": "fanatic, fiend",
  "n10078131": "fancier, enthusiast",
  "n10078333": "fancy man, paramour",
  "n10078415": "fantasist",
  "n10078529": "fantast, futurist",
  "n10078643": "fare",
  "n10078719": "farm boy",
  "n10078806": "farmer, husbandman, granger, sodbuster",
  "n10079210": "farmerette",
  "n10079310": "farm girl",
  "n10079399": "farmhand, fieldhand, field hand, farm worker",
  "n10079677": "farrier, horseshoer",
  "n10079769": "Farsi",
  "n10079893": "fascist",
  "n10080117": "fascista",
  "n10080206": "fashion consultant, fashionmonger",
  "n10080337": "fastener",
  "n10080508": "fatalist, determinist, predestinarian, predestinationist",
  "n10080784": "fat cat",
  "n10080869": "father, male parent, begetter",
  "n10081204": "Father, Padre",
  "n10081456": "father",
  "n10081670": "father figure, father surrogate",
  "n10081842": "father-figure",
  "n10082043": "father-in-law",
  "n10082146": "fatso, fatty, fat person, roly-poly, butterball",
  "n10082299": "Fauntleroy, Little Lord Fauntleroy",
  "n10082423": "Fauve, fauvist",
  "n10082562": "favorite son",
  "n10082687": "featherweight",
  "n10082805": "featherweight",
  "n10082886": "featherweight",
  "n10082997": "federalist",
  "n10083097": "Federalist",
  "n10083264": "fellah",
  "n10083358": "fellow, dude, buster",
  "n10083526": "fellow",
  "n10083677": "fellow traveler, fellow traveller",
  "n10083823": "female aristocrat",
  "n10084043": "female offspring",
  "n10084181": "female sibling",
  "n10084295": "female child, girl, little girl",
  "n10084635": "feminist, women's rightist, women's liberationist, libber",
  "n10085101": "fence",
  "n10085217": "fencer, swordsman",
  "n10085344": "fence-sitter",
  "n10085449": "ferryman",
  "n10085548": "fetishist",
  "n10085736": "feudal lord, seigneur, seignior",
  "n10085869": "fiance, groom-to-be",
  "n10085970": "fiancee, bride-to-be",
  "n10086074": "fiduciary",
  "n10086383": "fielder, fieldsman",
  "n10086568": "fielder",
  "n10086744": "field judge",
  "n10086821": "field marshal",
  "n10087080": "field-grade officer, field officer, FO",
  "n10087255": "fifth columnist, saboteur",
  "n10087434": "fighter pilot",
  "n10087574": "file clerk, filing clerk, filer",
  "n10087736": "filer",
  "n10087868": "filibuster, filibusterer",
  "n10088101": "filicide",
  "n10088200": "film director, director",
  "n10088390": "film maker, filmmaker, film producer, movie maker",
  "n10089484": "film star, movie star",
  "n10089615": "finagler, wangler",
  "n10089779": "finalist",
  "n10089892": "finance minister, minister of finance",
  "n10090020": "financier, moneyman",
  "n10090498": "finder, discoverer, spotter",
  "n10090745": "finder",
  "n10090864": "fingerprint expert, fingerprint specialist, fingerprint man",
  "n10091012": "fink, snitch, snitcher, stoolpigeon, stool pigeon, stoolie, sneak, sneaker, canary",
  "n10091256": "fieldworker",
  "n10091349": "fire chief, fire marshal",
  "n10091450": "fire-eater, fire-swallower",
  "n10091564": "fire-eater, hothead",
  "n10091651": "fireman, firefighter, fire fighter, fire-eater",
  "n10091861": "fire marshall",
  "n10091997": "fire walker",
  "n10092098": "fire warden, forest fire fighter, ranger",
  "n10092299": "fire watcher",
  "n10092488": "first baseman, first sacker",
  "n10092643": "firstborn, eldest",
  "n10092794": "first lady",
  "n10092880": "first lady",
  "n10092978": "first lieutenant, 1st lieutenant",
  "n10093167": "first offender",
  "n10093264": "first-nighter",
  "n10093396": "first-rater",
  "n10093475": "first sergeant, sergeant first class",
  "n10093658": "fisherman, fisher",
  "n10093818": "fishmonger, fishwife",
  "n10093908": "fitter",
  "n10094046": "fixer, influence peddler",
  "n10094236": "flag captain",
  "n10094320": "flagellant",
  "n10094444": "flagellant",
  "n10094584": "flag officer",
  "n10094782": "flak catcher, flak, flack catcher, flack",
  "n10094954": "flamen",
  "n10095061": "flanker",
  "n10095265": "flanker back, flanker",
  "n10095420": "flapper",
  "n10095550": "flash in the pan",
  "n10095664": "flatfoot, patrolman",
  "n10095769": "flatmate",
  "n10095869": "flatterer, adulator",
  "n10096016": "fleet admiral, five-star admiral",
  "n10096126": "flibbertigibbet, foolish woman",
  "n10096217": "flier, flyer",
  "n10096375": "flight engineer",
  "n10096508": "flight surgeon",
  "n10096620": "floater",
  "n10096725": "floater",
  "n10096855": "floater",
  "n10096964": "flogger, scourger",
  "n10097154": "floor leader",
  "n10097262": "floorwalker, shopwalker",
  "n10097477": "flop, dud, washout",
  "n10097590": "Florentine",
  "n10097705": "florist",
  "n10097842": "flower girl",
  "n10097995": "flower girl",
  "n10098092": "flunky, flunkey, stooge, yes-man",
  "n10098245": "flutist, flautist, flute player",
  "n10098388": "fly-by-night",
  "n10098517": "flyweight",
  "n10098624": "flyweight",
  "n10098710": "foe, enemy",
  "n10098862": "folk dancer",
  "n10099002": "folk poet",
  "n10099093": "folk singer, jongleur, minstrel, poet-singer, troubadour",
  "n10099278": "folk writer",
  "n10099375": "follower",
  "n10100124": "follower",
  "n10100314": "fondler",
  "n10100514": "food faddist",
  "n10100620": "food manufacturer",
  "n10100761": "fool, sap, saphead, muggins, tomfool",
  "n10101078": "foot",
  "n10101202": "football coach",
  "n10101308": "football hero",
  "n10101427": "football official",
  "n10101634": "football player, footballer",
  "n10101981": "footman",
  "n10102130": "footpad, padder",
  "n10102222": "forager",
  "n10102369": "forebear, forbear",
  "n10102506": "forecaster, predictor, prognosticator, soothsayer",
  "n10102800": "forefather, father, sire",
  "n10102969": "forefather",
  "n10103155": "foremother",
  "n10103228": "foreign agent",
  "n10103315": "foreign correspondent",
  "n10103485": "foreigner, alien, noncitizen, outlander",
  "n10103794": "foreign minister, secretary of state",
  "n10103921": "foreigner, outsider",
  "n10104064": "boss",
  "n10104209": "foreman, chief, gaffer, honcho, boss",
  "n10104487": "foreman",
  "n10104592": "foreperson",
  "n10104756": "forester, tree farmer, arboriculturist",
  "n10104888": "forewoman",
  "n10104983": "forewoman, forelady",
  "n10105085": "forger, counterfeiter",
  "n10105260": "forger",
  "n10105359": "fortune hunter",
  "n10105462": "fortuneteller, fortune teller",
  "n10105618": "forty-niner",
  "n10105733": "forward",
  "n10105906": "foster-brother, foster brother",
  "n10106080": "foster-child, foster child, fosterling",
  "n10106242": "foster-daughter, foster daughter",
  "n10106387": "foster-father, foster father",
  "n10106509": "foster-mother, foster mother",
  "n10106642": "foster-nurse",
  "n10106752": "foster-parent, foster parent",
  "n10106995": "foster-sister, foster sister",
  "n10107173": "foster-son, foster son",
  "n10107303": "founder, beginner, founding father, father",
  "n10107604": "Founding Father",
  "n10107778": "founder",
  "n10107883": "foundling, abandoned infant",
  "n10108018": "foundress",
  "n10108089": "four-minute man",
  "n10108240": "fowler",
  "n10108348": "fox hunter",
  "n10108464": "framer",
  "n10108606": "framer",
  "n10108719": "Francophile, Francophil",
  "n10108832": "Francophobe",
  "n10108937": "franc-tireur",
  "n10109050": "franklin",
  "n10109197": "fraternal twin, dizygotic twin",
  "n10109342": "fratricide",
  "n10109443": "freak, monster, monstrosity, lusus naturae",
  "n10109662": "free agent, free spirit, freewheeler",
  "n10109826": "free agent",
  "n10109985": "freedman, freedwoman",
  "n10110093": "freedom rider",
  "n10110287": "freeholder",
  "n10110421": "freelancer, freelance, free-lance, free lance, independent, self-employed person",
  "n10110731": "free-liver",
  "n10110893": "freeloader",
  "n10111023": "freeman, freewoman",
  "n10111144": "Freemason, Mason",
  "n10111358": "free trader",
  "n10111463": "freight agent",
  "n10111601": "French teacher",
  "n10111688": "freshman, fresher",
  "n10111779": "Freudian",
  "n10111903": "friar, mendicant",
  "n10112129": "monk, monastic",
  "n10112434": "Benedictine",
  "n10112591": "friend",
  "n10113072": "frontiersman, backwoodsman, mountain man",
  "n10113249": "frontierswoman",
  "n10113362": "frontbencher",
  "n10113583": "front man, front, figurehead, nominal head, straw man, strawman",
  "n10113753": "front-runner, favorite, favourite",
  "n10113869": "frotteur",
  "n10113997": "fruiterer",
  "n10114114": "fruit grower",
  "n10114209": "frump, dog",
  "n10114384": "fry cook",
  "n10114476": "fucker",
  "n10114550": "fucker",
  "n10114662": "fuddy-duddy",
  "n10114897": "fugitive, fugitive from justice",
  "n10115082": "fugitive, runaway, fleer",
  "n10115271": "fugleman",
  "n10115430": "fullback",
  "n10115603": "fuller",
  "n10115748": "full professor",
  "n10115840": "fumigator",
  "n10115946": "funambulist, tightrope walker",
  "n10116093": "functional illiterate",
  "n10116246": "functionalist",
  "n10116370": "fundamentalist",
  "n10116478": "fundraiser",
  "n10116579": "fusilier",
  "n10116702": "futurist",
  "n10116905": "gadabout",
  "n10117017": "gadgeteer",
  "n10117154": "gaffer",
  "n10117267": "gagman, gagster, gagwriter",
  "n10117415": "gagman, standup comedian",
  "n10117511": "gainer",
  "n10117739": "gainer, weight gainer",
  "n10117851": "gal",
  "n10117957": "Galilean, Galilaean",
  "n10118113": "galley slave",
  "n10118208": "gallows bird",
  "n10118301": "galoot",
  "n10118382": "galvanizer, galvaniser, inspirer",
  "n10118587": "galvanizer, galvaniser",
  "n10118743": "gambist",
  "n10118844": "gambler",
  "n10119200": "gambler, risk taker",
  "n10119351": "gamekeeper, game warden",
  "n10119491": "games-master, games-mistress",
  "n10119609": "gamine",
  "n10119685": "gamine",
  "n10119775": "gandy dancer",
  "n10119874": "ganger",
  "n10119953": "gangsta",
  "n10120085": "gangster, mobster",
  "n10120330": "garbage man, garbageman, garbage collector, garbage carter, garbage hauler, refuse collector, dustman",
  "n10120533": "gardener, nurseryman",
  "n10120671": "gardener",
  "n10120816": "garmentmaker, garment-worker, garment worker",
  "n10121026": "garment cutter",
  "n10121144": "garnishee",
  "n10121246": "garroter, garrotter, strangler, throttler, choker",
  "n10121467": "gasbag, windbag",
  "n10121595": "gas fitter",
  "n10121714": "gasman",
  "n10121800": "gastroenterologist",
  "n10121952": "gatecrasher, crasher, unwelcome guest",
  "n10122128": "gatekeeper",
  "n10122300": "gatherer",
  "n10122441": "gaucho",
  "n10122531": "gawker",
  "n10122645": "gay man, shirtlifter",
  "n10122765": "gazetteer",
  "n10122858": "geisha, geisha girl",
  "n10123026": "gem cutter",
  "n10123122": "gendarme",
  "n10123214": "genealogist",
  "n10123312": "Genevan",
  "n10123415": "Genoese",
  "n10123517": "genre painter",
  "n10123615": "geek",
  "n10123711": "geezer",
  "n10123844": "general, full general",
  "n10125561": "general, superior general",
  "n10125697": "general manager",
  "n10125786": "general officer",
  "n10126009": "general practitioner, GP",
  "n10126177": "generator, source, author",
  "n10126424": "geneticist",
  "n10126708": "genitor",
  "n10126806": "progenitor, primogenitor",
  "n10126926": "genius, mastermind, brain, brainiac, Einstein",
  "n10127186": "gent",
  "n10127273": "gentleman",
  "n10127420": "gentleman-at-arms",
  "n10127555": "geographer",
  "n10127689": "geologist",
  "n10127916": "geomancer",
  "n10128016": "geometer, geometrician",
  "n10128163": "geometry teacher",
  "n10128254": "Germanist",
  "n10128381": "gerontologist, geriatrician",
  "n10128519": "geophysicist",
  "n10128748": "ghostwriter, ghost",
  "n10128909": "giant, goliath, behemoth, monster, colossus",
  "n10129133": "giant, hulk, heavyweight, whale",
  "n10129338": "Gibson girl",
  "n10129464": "gigolo",
  "n10129585": "gilder",
  "n10129719": "gillie",
  "n10129825": "girl, miss, missy, young lady, young woman, fille",
  "n10130447": "girl",
  "n10130584": "girl Friday",
  "n10130686": "girlfriend, girl, lady friend",
  "n10130877": "girlfriend",
  "n10131016": "Girl Scout",
  "n10131151": "girl wonder",
  "n10131268": "Girondist, Girondin",
  "n10131515": "gitana",
  "n10131590": "gitano",
  "n10131663": "giver",
  "n10131815": "gladiator",
  "n10132035": "glassblower",
  "n10132145": "glass cutter, glass-cutter, glassworker, glazier, glazer",
  "n10132305": "glass cutter, glass-cutter",
  "n10132422": "glassmaker",
  "n10132502": "gleaner",
  "n10132641": "gleaner",
  "n10132775": "globetrotter, world traveler",
  "n10132887": "glossarist",
  "n10132988": "glutton, gourmand, gourmandizer, trencherman",
  "n10133210": "Gnostic",
  "n10133307": "god",
  "n10133458": "gonif, goniff, ganef, ganof",
  "n10133644": "government agent",
  "n10133850": "G-man, FBI agent, government man",
  "n10134001": "goalkeeper, goalie, goaltender, netkeeper, netminder",
  "n10134178": "goat herder, goatherd",
  "n10134282": "gobbler",
  "n10134396": "godchild",
  "n10134552": "goddaughter",
  "n10134627": "godfather",
  "n10134760": "godfather",
  "n10134870": "godmother",
  "n10134982": "godparent",
  "n10135129": "godson",
  "n10135197": "gofer",
  "n10135297": "goffer, gopher",
  "n10135411": "Gog and Magog",
  "n10135709": "go-getter, whizz-kid, whiz-kid, ball of fire",
  "n10135842": "goldbeater, gold-beater",
  "n10135953": "goldbrick, goof-off, ne'er-do-well, good-for-nothing, no-account, good-for-naught",
  "n10136146": "goldbrick",
  "n10136283": "gold digger",
  "n10136463": "gold miner, gold digger, gold panner",
  "n10136615": "goldsmith, goldworker, gold-worker",
  "n10136775": "golem",
  "n10136959": "golfer, golf player, linksman",
  "n10137367": "golf pro, professional golfer",
  "n10137498": "golf widow",
  "n10137632": "goliard",
  "n10137825": "gondolier, gondoliere",
  "n10137936": "goner, toast",
  "n10138114": "Gongorist",
  "n10138242": "good egg",
  "n10138369": "good guy",
  "n10138472": "good old boy, good ole boy, good ol' boy",
  "n10138767": "good person",
  "n10139077": "good Samaritan",
  "n10139206": "goody-goody",
  "n10139347": "gossip, gossiper, gossipmonger, rumormonger, rumourmonger, newsmonger",
  "n10139651": "gossip columnist",
  "n10139774": "Goth",
  "n10139944": "Gothic romancer",
  "n10140051": "gouger",
  "n10140169": "governess",
  "n10140314": "governor",
  "n10140597": "governor general",
  "n10140683": "grabber",
  "n10140783": "grader",
  "n10140929": "graduate nurse, trained nurse",
  "n10141109": "graduate student, grad student, postgraduate",
  "n10141268": "grain merchant",
  "n10141364": "grammarian, syntactician",
  "n10141590": "grandchild",
  "n10141732": "granddaughter",
  "n10141811": "grand dragon",
  "n10141930": "grand duchess",
  "n10142060": "grand duke",
  "n10142166": "grande dame",
  "n10142290": "grandee",
  "n10142391": "grandfather, gramps, granddad, grandad, granddaddy, grandpa",
  "n10142537": "Grand Inquisitor",
  "n10142747": "grandma, grandmother, granny, grannie, gran, nan, nanna",
  "n10142946": "grandmaster",
  "n10143085": "grand mufti",
  "n10143172": "grandparent",
  "n10143299": "grandson",
  "n10143371": "grandstander",
  "n10143530": "granny",
  "n10143595": "grantee",
  "n10143725": "granter",
  "n10143889": "grantor",
  "n10144055": "graphic designer, designer",
  "n10144188": "graphologist, handwriting expert",
  "n10144338": "grass widower, divorced man",
  "n10144468": "gravedigger",
  "n10144571": "graverobber, ghoul, body snatcher",
  "n10144730": "graverobber",
  "n10144838": "gravida",
  "n10144962": "grazier",
  "n10145081": "great",
  "n10145239": "great-aunt, grandaunt",
  "n10145340": "great grandchild",
  "n10145480": "great granddaughter",
  "n10145590": "great grandmother",
  "n10145682": "great grandfather",
  "n10145774": "great grandparent",
  "n10145902": "great grandson",
  "n10146002": "great-nephew, grandnephew",
  "n10146104": "great-niece, grandniece",
  "n10146209": "great-uncle, granduncle",
  "n10146313": "Grecian",
  "n10146416": "Green Beret",
  "n10146559": "greengrocer",
  "n10146682": "greenskeeper",
  "n10146816": "grenadier, grenade thrower",
  "n10146927": "greeter, saluter, welcomer",
  "n10147121": "gringo",
  "n10147262": "grinner",
  "n10147352": "griot",
  "n10147488": "grip",
  "n10147619": "groaner",
  "n10147710": "grocer",
  "n10147849": "grocery boy",
  "n10147935": "groom, bridegroom",
  "n10148035": "groom, bridegroom",
  "n10148165": "groomsman",
  "n10148305": "grouch, grump, crank, churl, crosspatch",
  "n10148542": "groundling",
  "n10148663": "groundsman, groundskeeper, groundkeeper",
  "n10148825": "group captain",
  "n10148991": "groupie",
  "n10149128": "growler",
  "n10149241": "grunt",
  "n10149436": "grunter",
  "n10149527": "guarantor, surety, warrantor, warranter",
  "n10149720": "guard",
  "n10149867": "prison guard, jailer, jailor, gaoler, screw, turnkey",
  "n10150071": "guard",
  "n10150281": "guard",
  "n10150415": "guardsman",
  "n10150556": "guerrilla, guerilla, irregular, insurgent",
  "n10150794": "guesser",
  "n10150940": "guest, invitee",
  "n10151133": "guest",
  "n10151261": "guest of honor",
  "n10151367": "guest worker, guestworker",
  "n10151570": "guide",
  "n10151760": "guitarist, guitar player",
  "n10151957": "gulper, guzzler",
  "n10152083": "gunman, gunslinger, hired gun, gun, gun for hire, triggerman, hit man, hitman, torpedo, shooter",
  "n10152306": "gunnery sergeant",
  "n10152440": "gunrunner, arms-runner",
  "n10152528": "gunsmith",
  "n10152616": "guru",
  "n10152763": "guru",
  "n10152889": "Guru",
  "n10153009": "gutter",
  "n10153155": "guvnor",
  "n10153266": "guzzler",
  "n10153414": "guy, cat, hombre, bozo",
  "n10153594": "gymnast",
  "n10153724": "gymnosophist",
  "n10153865": "gym rat",
  "n10154013": "gynecologist, gynaecologist, woman's doctor",
  "n10154186": "Gypsy, Gipsy, Romany, Rommany, Romani, Roma, Bohemian",
  "n10154601": "hack, drudge, hacker",
  "n10154740": "hack, hack writer, literary hack",
  "n10154871": "hacker",
  "n10155140": "hacker",
  "n10155222": "hacker, cyber-terrorist, cyberpunk",
  "n10155485": "hag, beldam, beldame, witch, crone",
  "n10155600": "haggler",
  "n10155692": "hagiographer, hagiographist, hagiologist",
  "n10155849": "hairdresser, hairstylist, stylist, styler",
  "n10156173": "hairsplitter",
  "n10156284": "hajji, hadji, haji",
  "n10156425": "hajji",
  "n10156629": "hakim, hakeem",
  "n10156726": "hakim",
  "n10156831": "Hakka",
  "n10157016": "halberdier",
  "n10157128": "halfback",
  "n10157271": "half blood",
  "n10157378": "half-caste",
  "n10157549": "half-breed",
  "n10157744": "fathead, goof, goofball, bozo, jackass, goose, cuckoo, twat, zany",
  "n10158010": "ham, ham actor",
  "n10158139": "ham",
  "n10158222": "Ham",
  "n10158319": "Haman",
  "n10158506": "hand",
  "n10158608": "handler",
  "n10158756": "handicapped person",
  "n10159045": "animal trainer, handler",
  "n10159186": "handmaid, handmaiden",
  "n10159289": "handyman, jack of all trades, odd-job man",
  "n10159434": "hanger",
  "n10159533": "hang glider",
  "n10159615": "hangman",
  "n10159714": "haranguer",
  "n10159852": "Hanoverian",
  "n10160012": "harasser, harrier",
  "n10160188": "hardliner",
  "n10160280": "harlequin",
  "n10160412": "harmonizer, harmoniser",
  "n10160624": "harmonizer, harmoniser",
  "n10160770": "harpist, harper",
  "n10160913": "harpooner, harpooneer",
  "n10161047": "harpsichordist",
  "n10161178": "harasser",
  "n10161273": "harridan",
  "n10161363": "harvester, reaper",
  "n10161521": "has-been, back-number",
  "n10161622": "hash head",
  "n10161695": "Hasid, Hassid, Chasid, Chassid",
  "n10161867": "hatchet man, enforcer",
  "n10162016": "hatchet man, iceman",
  "n10162104": "hatemonger",
  "n10162194": "hater",
  "n10162354": "hatmaker, hatter, milliner, modiste",
  "n10162507": "hauler, haulier",
  "n10162644": "hawk, war hawk",
  "n10162780": "head",
  "n10162991": "head, chief, top dog",
  "n10163354": "head",
  "n10163452": "headhunter, head-shrinker",
  "n10163593": "headhunter",
  "n10163723": "headliner, star",
  "n10163900": "head linesman",
  "n10164025": "headman, tribal chief, chieftain, chief",
  "n10164233": "headmaster, schoolmaster, master",
  "n10164397": "headmistress",
  "n10164492": "head nurse",
  "n10164605": "head of household",
  "n10164747": "head of state, chief of state",
  "n10164997": "headsman, headman",
  "n10165109": "health professional, primary care provider, PCP, health care provider, caregiver",
  "n10165448": "hearer, listener, auditor, attender",
  "n10165673": "audile",
  "n10165808": "motile",
  "n10165957": "hearing examiner, hearing officer",
  "n10166189": "heartbreaker",
  "n10166313": "heartthrob",
  "n10166394": "heathen, pagan, gentile, infidel",
  "n10166626": "paynim",
  "n10166762": "heaver",
  "n10166902": "heavy hitter",
  "n10167042": "heavyweight",
  "n10167152": "heavyweight",
  "n10167252": "heavyweight",
  "n10167361": "heavy",
  "n10167448": "Hebraist",
  "n10167565": "heckler, badgerer",
  "n10167740": "hedger",
  "n10167838": "hedger",
  "n10168012": "hedger, equivocator, tergiversator",
  "n10168183": "hedonist, pagan, pleasure seeker",
  "n10168368": "Hegelian",
  "n10168457": "Heidelberg man, Homo heidelbergensis",
  "n10168584": "heir, inheritor, heritor",
  "n10168837": "heir apparent",
  "n10169008": "heir-at-law",
  "n10169147": "heiress, inheritress, inheritrix",
  "n10169241": "heir presumptive",
  "n10169419": "hellion, heller, devil",
  "n10169607": "hellhound",
  "n10169678": "hell-kite, hell-rooster, gamecock",
  "n10169796": "helmsman, steersman, steerer",
  "n10169937": "hierarch",
  "n10170060": "hire",
  "n10170172": "hired help",
  "n10170359": "histologist",
  "n10170598": "helpmate, helpmeet",
  "n10170681": "hematologist, haematologist",
  "n10170866": "hemiplegic",
  "n10170989": "hemophiliac, haemophiliac, bleeder, hemophile, haemophile",
  "n10171219": "herald, trumpeter",
  "n10171456": "herbalist, herb doctor",
  "n10171567": "herder, herdsman, drover",
  "n10171755": "heretic, misbeliever, religious outcast",
  "n10171953": "heretic",
  "n10172080": "hermaphrodite, intersex, gynandromorph, androgyne, epicene, epicene person",
  "n10172448": "hermit, recluse, solitary, solitudinarian, troglodyte",
  "n10172668": "herpetologist",
  "n10172793": "protagonist, agonist",
  "n10172942": "antihero",
  "n10173086": "hero",
  "n10173305": "heroine",
  "n10173410": "heroine",
  "n10173579": "heroin addict",
  "n10173665": "hero worshiper, hero worshipper",
  "n10173771": "Herr",
  "n10173895": "heterosexual, heterosexual person, straight person, straight",
  "n10174148": "hewer",
  "n10174253": "highbinder",
  "n10174330": "highbrow",
  "n10174445": "high commissioner",
  "n10174589": "highflier, highflyer",
  "n10174695": "Highlander, Scottish Highlander, Highland Scot",
  "n10174845": "Highlander",
  "n10174971": "high-muck-a-muck, pooh-bah",
  "n10175090": "Highness",
  "n10175248": "high priest",
  "n10175418": "high roller",
  "n10175507": "highjacker, highwayman, hijacker, road agent",
  "n10175725": "highjacker, hijacker",
  "n10175963": "highway engineer",
  "n10176111": "hiker, tramp, tramper",
  "n10176357": "hillbilly, bushwhacker",
  "n10176475": "hippie, hippy, hipster, flower child",
  "n10176679": "hired hand, hand, hired man",
  "n10176913": "hireling, pensionary",
  "n10177014": "hisser",
  "n10177150": "historian, historiographer",
  "n10178077": "hitchhiker",
  "n10178216": "hitter, striker",
  "n10178464": "Hittite",
  "n10178611": "hoarder",
  "n10178821": "hobbledehoy",
  "n10178917": "hobbler, limper",
  "n10179069": "hobbyist",
  "n10179207": "hockey coach",
  "n10179291": "hockey player, ice-hockey player",
  "n10179527": "hod carrier, hodman",
  "n10179649": "hog, pig",
  "n10179817": "hoister",
  "n10179911": "holder, bearer",
  "n10180178": "holder",
  "n10180580": "holdout",
  "n10180791": "holdover, hangover",
  "n10180923": "holdup man, stickup man",
  "n10181026": "Holy Roller",
  "n10181137": "Holy Roman Emperor",
  "n10181445": "homeboy",
  "n10181547": "homeboy",
  "n10181656": "homebuilder, home-builder, housebuilder, house-builder",
  "n10181799": "home buyer",
  "n10181878": "homegirl",
  "n10181990": "home help",
  "n10182190": "homeless, homeless person",
  "n10182402": "homeopath, homoeopath",
  "n10182499": "homeowner, householder",
  "n10182628": "Home Secretary, Secretary of State for the Home Department",
  "n10182812": "homophobe",
  "n10182913": "homosexual, homophile, homo, gay",
  "n10183157": "homunculus",
  "n10183347": "honest woman",
  "n10183556": "honker",
  "n10183757": "honoree",
  "n10183931": "honor guard, guard of honor",
  "n10184081": "hood, hoodlum, goon, punk, thug, tough, toughie, strong-armer",
  "n10184290": "hoodoo",
  "n10184403": "hoofer, stepper",
  "n10184505": "hooker",
  "n10184683": "hooker",
  "n10184822": "Hooray Henry",
  "n10184946": "hope",
  "n10185148": "hoper",
  "n10185327": "hopper",
  "n10185483": "hornist",
  "n10185591": "horse doctor",
  "n10185682": "horseman, horse fancier",
  "n10185793": "horseman, equestrian, horseback rider",
  "n10186068": "horse trader",
  "n10186143": "horsewoman",
  "n10186216": "horse wrangler, wrangler",
  "n10186350": "horticulturist, plantsman",
  "n10186578": "hosier",
  "n10186686": "hospital chaplain",
  "n10186774": "host, innkeeper, boniface",
  "n10186950": "hosteller",
  "n10187059": "hostess",
  "n10187130": "host",
  "n10187356": "host",
  "n10187491": "hostess",
  "n10187557": "hostage, surety",
  "n10187710": "hotdog, hot dog",
  "n10187842": "hotel detective, house detective, house dick",
  "n10187990": "hotelier, hotelkeeper, hotel manager, hotelman, hosteller",
  "n10188146": "hotspur",
  "n10188226": "housebreaker, cat burglar",
  "n10188369": "housefather",
  "n10188472": "house guest, houseguest",
  "n10188576": "house husband, househusband",
  "n10188715": "housekeeper",
  "n10188856": "housemaster",
  "n10188957": "housemate",
  "n10189059": "housemother",
  "n10189179": "house painter",
  "n10189278": "house physician, resident, resident physician",
  "n10189597": "house sitter",
  "n10189776": "housewife, homemaker, lady of the house, woman of the house",
  "n10189975": "housewrecker, housebreaker",
  "n10190122": "housing commissioner",
  "n10190231": "Houyhnhnm",
  "n10190373": "Huayna Capac",
  "n10190516": "huckster, cheap-jack",
  "n10190644": "huckster",
  "n10190745": "huddler",
  "n10190871": "huddler",
  "n10191001": "hugger",
  "n10191089": "Huguenot",
  "n10191192": "humanist",
  "n10191388": "humanist, humanitarian",
  "n10191613": "humanitarian, do-gooder, improver",
  "n10191802": "hummer",
  "n10191943": "humorist, humourist",
  "n10192412": "humpback, hunchback, crookback",
  "n10192588": "Hun",
  "n10192704": "hunger marcher",
  "n10192839": "hunk",
  "n10192926": "hunted person",
  "n10193026": "hunter, huntsman",
  "n10193419": "hunter-gatherer",
  "n10193543": "hunting guide",
  "n10193650": "huntress",
  "n10193719": "hunter",
  "n10193864": "hurdler",
  "n10193967": "husband, hubby, married man",
  "n10194231": "ex-husband, ex",
  "n10194341": "hussar",
  "n10194463": "Hussite",
  "n10194566": "hustler, wheeler dealer, operator",
  "n10194775": "hydrologist",
  "n10194865": "hydromancer",
  "n10194969": "hygienist",
  "n10195056": "hyperope",
  "n10195155": "hypertensive",
  "n10195261": "hypnotist, hypnotizer, hypnotiser, mesmerist, mesmerizer",
  "n10195487": "hypochondriac",
  "n10195593": "hypocrite, dissembler, dissimulator, phony, phoney, pretender",
  "n10195942": "hypotensive",
  "n10196046": "hysteric",
  "n10196133": "Iberian",
  "n10196283": "Iberian",
  "n10196404": "iceman",
  "n10196490": "ice-skater",
  "n10196617": "ichthyologist",
  "n10196725": "iconoclast",
  "n10196845": "iconoclast, image breaker",
  "n10196965": "idealist, dreamer",
  "n10197182": "identical twin, monozygotic twin, monozygous twin",
  "n10197392": "ideologist, ideologue",
  "n10197525": "idiot, imbecile, cretin, moron, changeling, half-wit, retard",
  "n10197780": "idiot savant",
  "n10197967": "idler, loafer, do-nothing, layabout, bum",
  "n10198437": "idol, matinee idol",
  "n10198602": "idolater, idolizer, idoliser, idol worshiper",
  "n10198759": "idolatress",
  "n10198832": "idolizer, idoliser",
  "n10198958": "ignoramus, know nothing, uneducated person",
  "n10199103": "illiterate, illiterate person, nonreader",
  "n10199251": "imam, imaum",
  "n10199489": "immigrant",
  "n10199644": "immortal",
  "n10199783": "immune",
  "n10199902": "immunologist",
  "n10200047": "imp, scamp, monkey, rascal, rapscallion, scalawag, scallywag",
  "n10200246": "imperialist",
  "n10200365": "impersonator, imitator",
  "n10200531": "import, importee",
  "n10200781": "important person, influential person, personage",
  "n10201366": "importer",
  "n10201535": "imposter, impostor, pretender, fake, faker, fraud, sham, shammer, pseudo, pseud, role player",
  "n10201956": "impressionist",
  "n10202085": "inamorata",
  "n10202225": "inamorato",
  "n10202363": "incompetent, incompetent person",
  "n10202531": "incubus",
  "n10202624": "incumbent, officeholder",
  "n10202763": "incurable",
  "n10202854": "index case",
  "n10203016": "indexer",
  "n10203135": "Indian agent",
  "n10203298": "Indian chief, Indian chieftain",
  "n10203682": "Indian giver",
  "n10203839": "individual",
  "n10203949": "inductee",
  "n10204177": "industrialist",
  "n10204833": "infanticide",
  "n10204921": "infantryman, marcher, foot soldier, footslogger",
  "n10205138": "doughboy",
  "n10205231": "inferior",
  "n10205344": "infernal",
  "n10205457": "infielder",
  "n10205714": "infiltrator",
  "n10205833": "infiltrator",
  "n10205985": "informant, source",
  "n10206173": "informer, betrayer, rat, squealer, blabber",
  "n10206506": "ingenue",
  "n10206629": "ingenue",
  "n10206763": "ingrate, thankless wretch, ungrateful person",
  "n10206887": "initiate, learned person, pundit, savant",
  "n10207077": "polymath",
  "n10207169": "in-law, relative-in-law",
  "n10207370": "inmate",
  "n10207514": "inoculator, vaccinator",
  "n10207681": "inpatient, inmate",
  "n10207831": "inquirer, enquirer, questioner, querier, asker",
  "n10208189": "inquiry agent",
  "n10208287": "inquisitor, interrogator",
  "n10208432": "Inquisitor",
  "n10208583": "insider",
  "n10208748": "insomniac, sleepless person",
  "n10208847": "inspector",
  "n10208950": "inspector general",
  "n10209082": "instigator, initiator",
  "n10209246": "instigator, provoker, inciter, instigant, firebrand",
  "n10209539": "instructress",
  "n10209616": "instrument, pawn, cat's-paw",
  "n10209731": "insurance broker, insurance agent, general agent, underwriter",
  "n10209888": "insured, insured person",
  "n10210137": "insurgent, insurrectionist, freedom fighter, rebel",
  "n10210512": "intelligence analyst",
  "n10210648": "interior designer, designer, interior decorator, house decorator, room decorator, decorator",
  "n10210911": "interlocutor, conversational partner",
  "n10211036": "interlocutor, middleman",
  "n10211203": "intern, interne, houseman, medical intern",
  "n10211463": "internal auditor",
  "n10211666": "International Grandmaster",
  "n10211830": "internationalist",
  "n10211962": "internationalist",
  "n10212074": "internee",
  "n10212231": "internist",
  "n10212338": "internuncio",
  "n10212501": "interpreter, translator",
  "n10212780": "interpreter",
  "n10213034": "intervenor",
  "n10213180": "interviewee",
  "n10213319": "interviewer",
  "n10213429": "introvert",
  "n10213652": "intruder, interloper, trespasser",
  "n10214062": "invader, encroacher",
  "n10214230": "invalid, shut-in",
  "n10214390": "invalidator, voider, nullifier",
  "n10214637": "inventor, discoverer, artificer",
  "n10215623": "investigator",
  "n10215815": "investment adviser, investment advisor",
  "n10215953": "investment banker, underwriter",
  "n10216106": "investor",
  "n10216403": "invigilator",
  "n10216560": "iron man, ironman",
  "n10216690": "ironmonger, hardwareman",
  "n10216839": "ironside",
  "n10216928": "ironworker",
  "n10217038": "irredentist, irridentist",
  "n10217208": "irreligionist",
  "n10217335": "Islamist",
  "n10217436": "islander, island-dweller",
  "n10217552": "Ismaili, Ismailian",
  "n10217684": "isolationist",
  "n10217831": "itinerant, gypsy, gipsy",
  "n10218043": "Ivy Leaguer",
  "n10218164": "Jack of all trades",
  "n10218292": "Jacksonian",
  "n10218390": "Jacob",
  "n10218692": "Jacobean",
  "n10218802": "Jacobin",
  "n10218989": "Jacobite",
  "n10219121": "jail bird, jailbird, gaolbird",
  "n10219240": "Jane Doe",
  "n10219380": "Janissary",
  "n10219453": "janissary",
  "n10219577": "janitor",
  "n10219680": "Jansenist",
  "n10219778": "Japheth",
  "n10219879": "Jat",
  "n10220080": "Javanese, Javan",
  "n10220228": "jawan",
  "n10220360": "jaywalker",
  "n10220486": "jazz musician, jazzman",
  "n10220807": "Jeffersonian",
  "n10220924": "Jekyll and Hyde",
  "n10221040": "jerk, dork",
  "n10221162": "jerry-builder",
  "n10221312": "jester, fool, motley fool",
  "n10221520": "Jesuit",
  "n10221656": "jewel, gem",
  "n10221777": "jeweler, jeweller",
  "n10221956": "jewelry maker, jeweler, jeweller",
  "n10222170": "jezebel",
  "n10222259": "jilt",
  "n10222353": "jimdandy, jimhickey, crackerjack",
  "n10222497": "jobber, middleman, wholesaler",
  "n10222716": "job candidate",
  "n10222822": "jobholder",
  "n10222949": "Job",
  "n10223069": "Job's comforter",
  "n10223177": "jockey",
  "n10223294": "jockey",
  "n10223459": "jogger",
  "n10223606": "John Doe",
  "n10223744": "John Doe, Joe Blow, Joe Bloggs, man in the street",
  "n10223869": "joiner",
  "n10223994": "joiner",
  "n10224098": "joker, jokester",
  "n10224295": "joker, turkey",
  "n10224446": "jonah, jinx",
  "n10224578": "journalist",
  "n10225118": "Judas",
  "n10225219": "judge, justice, jurist",
  "n10225787": "judge advocate",
  "n10225931": "judge advocate",
  "n10226060": "judge advocate general",
  "n10226219": "Judith",
  "n10226413": "juggler",
  "n10226556": "juggernaut, steamroller",
  "n10226803": "jumper",
  "n10226993": "jumper",
  "n10227166": "Jungian",
  "n10227266": "junior",
  "n10227393": "junior",
  "n10227490": "Junior, Jr, Jnr",
  "n10227601": "junior featherweight",
  "n10227698": "junior lightweight",
  "n10227793": "junior middleweight",
  "n10227889": "junior welterweight",
  "n10227985": "jurist, legal expert",
  "n10228278": "juror, juryman, jurywoman",
  "n10228468": "justice of the peace",
  "n10228592": "justiciar, justiciary",
  "n10228712": "kachina",
  "n10228864": "kaffir, kafir, caffer, caffre",
  "n10229034": "Kalon Tripa",
  "n10229193": "kamikaze",
  "n10229338": "Kaiser",
  "n10229498": "keeper",
  "n10229721": "kerb crawler",
  "n10229883": "keyboardist",
  "n10229983": "Keynesian",
  "n10230097": "khan",
  "n10230216": "Khedive",
  "n10230335": "kibbutznik",
  "n10230432": "kibitzer",
  "n10230580": "kicker",
  "n10230736": "kiddy",
  "n10230801": "kidnapper, kidnaper, abductor, snatcher",
  "n10231087": "killer, slayer",
  "n10231361": "killer bee",
  "n10231515": "king, male monarch, Rex",
  "n10233248": "kingmaker",
  "n10233445": "King of England, King of Great Britain",
  "n10234340": "King of France",
  "n10234867": "King of the Germans",
  "n10235024": "king, queen, world-beater",
  "n10235142": "kingpin, top banana, bigwig",
  "n10235269": "King's Counsel",
  "n10235385": "Counsel to the Crown",
  "n10235549": "relative, relation",
  "n10236114": "blood relation, blood relative, cognate, sib",
  "n10236304": "kin, kinsperson, family",
  "n10236521": "enate, matrikin, matrilineal kin, matrisib, matrilineal sib",
  "n10236663": "agnate, patrikin, patrilineal kin, patrisib, patrilineal sib",
  "n10236842": "kink",
  "n10236946": "kinsman",
  "n10237069": "kinswoman",
  "n10237196": "kisser, osculator",
  "n10237315": "kissing cousin, kissing kin",
  "n10237464": "kitchen help",
  "n10237556": "kitchen police, KP",
  "n10237676": "Klansman, Ku Kluxer, Kluxer",
  "n10237799": "kleptomaniac",
  "n10237935": "klutz",
  "n10238029": "knacker",
  "n10238127": "knacker",
  "n10238272": "kneeler",
  "n10238375": "knight",
  "n10238749": "knight bachelor, bachelor-at-arms, bachelor",
  "n10238898": "knight banneret, knight of the square flag, banneret",
  "n10239089": "Knight of the Round Table",
  "n10239217": "knight-errant",
  "n10239329": "Knight Templar, Templar",
  "n10239501": "Knight Templar",
  "n10239619": "knitter",
  "n10239761": "knocker",
  "n10239928": "knocker",
  "n10240082": "knower, apprehender",
  "n10240235": "know-it-all, know-all",
  "n10240417": "kolkhoznik",
  "n10240514": "kook, odd fellow, odd fish, queer bird, queer duck, odd man out",
  "n10240715": "koto player",
  "n10240821": "Kshatriya",
  "n10240921": "kvetch",
  "n10241024": "labor coach, birthing coach, doula, monitrice",
  "n10241300": "laborer, manual laborer, labourer, jack",
  "n10242032": "labor leader",
  "n10242328": "Labourite",
  "n10242439": "lacer",
  "n10242573": "lackey, flunky, flunkey",
  "n10242682": "lacrosse player",
  "n10242791": "Lady, noblewoman, peeress",
  "n10243137": "lady",
  "n10243273": "lady-in-waiting",
  "n10243384": "ladylove, dulcinea",
  "n10243483": "lady's maid",
  "n10243582": "laird",
  "n10243664": "lama",
  "n10243791": "Lamarckian",
  "n10243872": "lamb, dear",
  "n10243988": "lamb",
  "n10244108": "lame duck",
  "n10244226": "laminator",
  "n10244359": "lamplighter",
  "n10244503": "lampoon artist",
  "n10244665": "lance corporal",
  "n10244818": "lancer",
  "n10244913": "land agent",
  "n10245029": "landgrave",
  "n10245156": "landlady",
  "n10245236": "landlord",
  "n10245341": "landlubber, lubber, landsman",
  "n10245507": "landlubber, landsman, landman",
  "n10245639": "landowner, landholder, property owner",
  "n10245863": "landscape architect, landscape gardener, landscaper, landscapist",
  "n10246176": "landscapist",
  "n10246317": "langlaufer",
  "n10246395": "languisher",
  "n10246511": "lapidary, lapidist",
  "n10246703": "lapidary, lapidarist",
  "n10246913": "larcenist, larcener",
  "n10247044": "large person",
  "n10247176": "lascar",
  "n10247250": "lasher",
  "n10247358": "lass, lassie, young girl, jeune fille",
  "n10247519": "latchkey child",
  "n10247720": "latecomer",
  "n10247800": "lather",
  "n10247880": "Latin",
  "n10248008": "Latin",
  "n10248091": "Latinist",
  "n10248198": "latitudinarian",
  "n10248377": "Jehovah's Witness",
  "n10248542": "Latter-Day Saint, Mormon",
  "n10248711": "laudator, lauder, extoller",
  "n10248876": "laugher",
  "n10249011": "laureate",
  "n10249191": "law agent",
  "n10249270": "lawgiver, lawmaker",
  "n10249459": "lawman, law officer, peace officer",
  "n10249869": "law student",
  "n10249950": "lawyer, attorney",
  "n10250527": "layman, layperson, secular",
  "n10250712": "lay reader",
  "n10250873": "lay witness",
  "n10251002": "Lazarus",
  "n10251125": "Lazarus",
  "n10251329": "lazybones",
  "n10251417": "leading lady",
  "n10251517": "leading man",
  "n10251612": "leaker",
  "n10251779": "learner, scholar, assimilator",
  "n10252075": "leaseholder, lessee",
  "n10252222": "lector, lecturer, reader",
  "n10252354": "lector, reader",
  "n10252547": "lecturer",
  "n10252674": "leech, parasite, sponge, sponger",
  "n10252921": "left-handed pitcher, left-hander, left hander, lefthander, lefty, southpaw",
  "n10253122": "left-hander, lefty, southpaw",
  "n10253296": "legal representative",
  "n10253479": "legate, official emissary",
  "n10253611": "legatee",
  "n10253703": "legionnaire, legionary",
  "n10253887": "Legionnaire",
  "n10253995": "legislator",
  "n10254392": "lender, loaner",
  "n10254585": "leper",
  "n10254670": "leper, lazar",
  "n10254761": "lepidopterist, lepidopterologist, butterfly collector",
  "n10254965": "lesbian, tribade, gay woman",
  "n10255096": "lessor, lease giver",
  "n10255207": "letter",
  "n10255348": "letterer",
  "n10255459": "letterman",
  "n10255567": "leveler, leveller",
  "n10255728": "leviathan",
  "n10255915": "Levite",
  "n10256080": "lexicographer, lexicologist",
  "n10256537": "liar, prevaricator",
  "n10256756": "liberal, liberalist, progressive",
  "n10257084": "liberal",
  "n10257221": "liberator",
  "n10257402": "libertarian",
  "n10257524": "libertarian",
  "n10257647": "libertine, debauchee, rounder",
  "n10257948": "librarian, bibliothec",
  "n10258152": "librettist",
  "n10258304": "licensed practical nurse, LPN, practical nurse",
  "n10258493": "licensee",
  "n10258602": "licenser",
  "n10258786": "licentiate",
  "n10258896": "lie-abed, slugabed",
  "n10259016": "lieder singer",
  "n10259094": "liege, liege lord",
  "n10259225": "lieutenant, police lieutenant",
  "n10259348": "lieutenant",
  "n10259527": "lieutenant",
  "n10259780": "lieutenant colonel, light colonel",
  "n10259997": "lieutenant commander",
  "n10260166": "lieutenant general",
  "n10260322": "lieutenant governor",
  "n10260473": "lieutenant junior grade, lieutenant JG",
  "n10260706": "life",
  "n10260800": "lifeguard, lifesaver",
  "n10260943": "life peer",
  "n10261041": "lifer",
  "n10261211": "life tenant",
  "n10261388": "lighterman, bargeman, bargee",
  "n10261511": "light flyweight",
  "n10261624": "light heavyweight, cruiserweight",
  "n10261762": "light heavyweight",
  "n10261862": "light heavyweight",
  "n10261977": "light middleweight",
  "n10262093": "lighthouse keeper",
  "n10262183": "lightning rod",
  "n10262343": "light-o'-love, light-of-love",
  "n10262445": "lightweight",
  "n10262561": "lightweight",
  "n10262655": "lightweight",
  "n10262764": "light welterweight",
  "n10262880": "lilliputian",
  "n10263002": "linebacker, line backer",
  "n10263146": "limnologist",
  "n10263259": "line judge",
  "n10263411": "lineman",
  "n10263580": "lineman",
  "n10263684": "linendraper",
  "n10263790": "line officer",
  "n10263931": "linesman",
  "n10264120": "line worker",
  "n10264219": "linguist, polyglot",
  "n10264437": "linguist, linguistic scientist",
  "n10264928": "linkboy, linkman",
  "n10265070": "lion, social lion",
  "n10265200": "lion-hunter",
  "n10265281": "lion-hunter",
  "n10265390": "lip reader",
  "n10265532": "liquidator, receiver",
  "n10265801": "lisper",
  "n10265891": "lister",
  "n10266016": "literary critic",
  "n10266328": "literate, literate person",
  "n10266486": "lithographer",
  "n10266744": "lithomancer",
  "n10266848": "litigant, litigator",
  "n10267166": "litterer, litterbug, litter lout",
  "n10267311": "little brother",
  "n10267437": "Little John",
  "n10267561": "little leaguer",
  "n10267719": "Little Red Riding Hood",
  "n10267865": "little sister",
  "n10267941": "liturgist",
  "n10268058": "liveborn infant",
  "n10268180": "liver",
  "n10268299": "liver",
  "n10268422": "liveryman",
  "n10268505": "loader",
  "n10268629": "lobbyist",
  "n10268827": "lobsterman",
  "n10268930": "locator, locater",
  "n10269078": "lockmaster, lockman, lockkeeper",
  "n10269199": "locksmith",
  "n10269289": "locum tenens, locum",
  "n10269458": "lodger, boarder, roomer",
  "n10269611": "logical positivist",
  "n10269785": "logician, logistician",
  "n10270109": "logomach, logomachist",
  "n10270232": "loiterer, lingerer",
  "n10270383": "Lolita",
  "n10270468": "lollipop lady, lollipop woman",
  "n10270628": "loner, lone wolf, lone hand",
  "n10270777": "longbowman",
  "n10270878": "longer, thirster, yearner",
  "n10271124": "long shot",
  "n10271216": "lookout, lookout man, sentinel, sentry, watch, spotter, scout, picket",
  "n10271451": "loon",
  "n10271525": "loose cannon",
  "n10271677": "Lord, noble, nobleman",
  "n10272171": "Lord Chancellor, Lord High Chancellor",
  "n10272375": "Lord of Misrule",
  "n10272499": "Lord Privy Seal",
  "n10272657": "Lorelei",
  "n10272782": "loser",
  "n10272913": "loser, also-ran",
  "n10273064": "failure, loser, nonstarter, unsuccessful person",
  "n10273352": "old maid",
  "n10273438": "underboss",
  "n10273572": "underdog",
  "n10273669": "Lot",
  "n10273906": "Lot's wife",
  "n10274173": "Lothario",
  "n10274318": "loudmouth, blusterer",
  "n10274474": "lounge lizard, lizard",
  "n10274639": "lout, clod, stumblebum, goon, oaf, lubber, lummox, lump, gawk",
  "n10274815": "lowerclassman, underclassman",
  "n10274985": "low-birth-weight baby, low-birth-weight infant",
  "n10275249": "Lowlander, Scottish Lowlander, Lowland Scot",
  "n10275395": "loyalist, stalwart",
  "n10275534": "Lubavitcher",
  "n10275676": "Luddite",
  "n10275848": "Luddite",
  "n10275940": "luger, slider",
  "n10276045": "lumberman, lumberjack, logger, feller, faller",
  "n10276238": "luminary, leading light, guiding light, notable, notability",
  "n10276477": "lumper",
  "n10276659": "light",
  "n10276764": "lunatic, madman, maniac",
  "n10276942": "bedlamite",
  "n10277027": "pyromaniac",
  "n10277132": "luncher",
  "n10277231": "lunger",
  "n10277352": "lurker, skulker, lurcher",
  "n10277509": "luthier",
  "n10277638": "lutist, lutanist, lutenist",
  "n10277815": "Lutheran",
  "n10277912": "lyricist, lyrist",
  "n10278128": "ma, mama, mamma, mom, momma, mommy, mammy, mum, mummy",
  "n10278263": "macaroni",
  "n10278456": "macebearer, mace, macer",
  "n10278568": "Machiavellian",
  "n10278666": "machine",
  "n10278805": "machine politician, ward-heeler, political hack, hack",
  "n10279018": "machinist, mechanic, shop mechanic",
  "n10279219": "macho",
  "n10279317": "Mackem",
  "n10279392": "macroeconomist, macroeconomic expert",
  "n10279540": "macushla",
  "n10279669": "madam, brothel keeper",
  "n10279778": "madame",
  "n10279867": "madrigalist",
  "n10279964": "madwoman",
  "n10280034": "maenad",
  "n10280130": "maestro, master",
  "n10280364": "mafioso",
  "n10280495": "mafioso",
  "n10280598": "magdalen",
  "n10280674": "magician, prestidigitator, conjurer, conjuror, illusionist",
  "n10280945": "magistrate",
  "n10281176": "magnifico",
  "n10281276": "magpie, scavenger, pack rat",
  "n10281432": "magus",
  "n10281546": "magus",
  "n10281637": "maharaja, maharajah",
  "n10281770": "maharani, maharanee",
  "n10281896": "mahatma",
  "n10282014": "Mahdi",
  "n10282262": "Mahdist",
  "n10282374": "mahout",
  "n10282482": "maid, maiden",
  "n10282672": "maid, maidservant, housemaid, amah",
  "n10282845": "maiden aunt",
  "n10282920": "mailer",
  "n10283037": "mailman, postman, mail carrier, letter carrier, carrier",
  "n10283170": "major",
  "n10283366": "major",
  "n10283546": "major-domo, seneschal",
  "n10283663": "major-general",
  "n10283824": "majority leader",
  "n10283931": "major leaguer, big leaguer",
  "n10284064": "maker, shaper",
  "n10284763": "malacologist",
  "n10284871": "malahini",
  "n10284965": "malcontent",
  "n10285135": "male aristocrat",
  "n10285313": "male child, boy",
  "n10285762": "transgressor",
  "n10285938": "male offspring, man-child",
  "n10286084": "male sibling",
  "n10286200": "malfeasant",
  "n10286282": "malik",
  "n10286539": "malingerer, skulker, shammer",
  "n10286749": "Malthusian",
  "n10286855": "maltster, maltman",
  "n10286989": "mammalogist",
  "n10287082": "mammy",
  "n10287213": "man, adult male",
  "n10288516": "man",
  "n10288763": "man",
  "n10288964": "adonis",
  "n10289039": "man",
  "n10289176": "man",
  "n10289338": "management consultant",
  "n10289462": "manageress",
  "n10289534": "managing editor",
  "n10289671": "mandarin",
  "n10289766": "mandarin",
  "n10289863": "mandarin",
  "n10289969": "mandatary, mandatory",
  "n10290099": "mandator",
  "n10290223": "Mandaean, Mandean",
  "n10290422": "maneuverer, manoeuvrer",
  "n10290541": "maniac",
  "n10290684": "manic-depressive",
  "n10290813": "Manichaean, Manichean, Manichee",
  "n10290919": "manicurist",
  "n10291110": "manipulator",
  "n10291240": "mannequin, manikin, mannikin, manakin, fashion model, model",
  "n10291469": "man-at-arms",
  "n10291580": "manikin, mannikin, homunculus",
  "n10291730": "man jack",
  "n10291822": "man of action, man of deeds",
  "n10291942": "man of letters",
  "n10292052": "man of means, rich man, wealthy man",
  "n10292192": "manservant",
  "n10292316": "manufacturer, producer",
  "n10292622": "Maoist",
  "n10292714": "map-reader",
  "n10292824": "Maquis, Maquisard",
  "n10292969": "marathoner, marathon runner, road runner, long-distance runner",
  "n10293172": "marauder, predator, vulture, piranha",
  "n10293332": "marcher, parader",
  "n10293502": "marcher",
  "n10293590": "marchioness, marquise",
  "n10293773": "marchioness",
  "n10293861": "margrave",
  "n10294020": "margrave",
  "n10294139": "Marine, devil dog, leatherneck, shipboard soldier",
  "n10294318": "marine",
  "n10294421": "marine engineer, naval engineer",
  "n10294602": "mariner, seaman, tar, Jack-tar, Jack, old salt, seafarer, gob, sea dog",
  "n10294953": "marksman, sharpshooter, crack shot",
  "n10295190": "maroon",
  "n10295371": "marquess",
  "n10295479": "marquis, marquess",
  "n10295597": "Marrano",
  "n10295819": "married",
  "n10295951": "marshal, marshall",
  "n10296176": "marshal, marshall",
  "n10296444": "martinet, disciplinarian, moralist",
  "n10296618": "martyr, sufferer",
  "n10296832": "martyr",
  "n10297140": "Marxist",
  "n10297234": "mascot",
  "n10297367": "masochist",
  "n10297531": "mason, stonemason",
  "n10297655": "Masorete, Massorete, Masorite",
  "n10297841": "masquerader, masker, masquer",
  "n10297983": "massager",
  "n10298202": "masseur",
  "n10298271": "masseuse",
  "n10298343": "mass murderer",
  "n10298482": "master, professional",
  "n10298647": "master",
  "n10298798": "master",
  "n10298912": "master, captain, sea captain, skipper",
  "n10299125": "master-at-arms",
  "n10299250": "master of ceremonies, emcee, host",
  "n10299583": "master sergeant",
  "n10299700": "masturbator, onanist",
  "n10299875": "matchmaker, matcher, marriage broker",
  "n10300041": "mate, first mate",
  "n10300154": "mate",
  "n10300303": "mate",
  "n10300500": "mater",
  "n10300654": "material",
  "n10300829": "materialist",
  "n10300963": "materialist",
  "n10301089": "material witness",
  "n10301261": "mathematician",
  "n10302446": "math teacher, mathematics teacher",
  "n10302576": "matriarch, materfamilias",
  "n10302700": "matriarch",
  "n10302816": "matricide",
  "n10302905": "matriculate",
  "n10303037": "matron",
  "n10303186": "matron",
  "n10303310": "matron",
  "n10303385": "matron of honor",
  "n10303513": "mauler",
  "n10303654": "maverick, rebel",
  "n10303814": "mayor, city manager",
  "n10304018": "mayoress",
  "n10304086": "mayoress",
  "n10304160": "May queen, queen of the May",
  "n10304278": "meanie, meany, unkind person",
  "n10304383": "measurer",
  "n10304505": "meat packer, packer",
  "n10304650": "mechanical engineer",
  "n10304783": "mechanist",
  "n10304914": "medalist, medallist, medal winner",
  "n10305062": "medalist, medallist",
  "n10305192": "meddler",
  "n10305391": "media consultant, media guru",
  "n10305523": "medical assistant",
  "n10305635": "medical officer, medic",
  "n10305802": "medical practitioner, medical man",
  "n10306004": "medical scientist",
  "n10306181": "medical student, medico",
  "n10306279": "medium, spiritualist, sensitive",
  "n10306496": "megalomaniac",
  "n10306595": "melancholic, melancholiac",
  "n10306716": "Melkite, Melchite",
  "n10306890": "Melkite, Melchite",
  "n10307114": "melter",
  "n10307234": "member, fellow member",
  "n10308066": "nonmember",
  "n10308168": "board member",
  "n10308275": "clansman, clanswoman, clan member",
  "n10308394": "club member",
  "n10308504": "memorizer, memoriser",
  "n10308653": "Mendelian",
  "n10308732": "mender, repairer, fixer",
  "n10308938": "menial",
  "n10309009": "mensch, mensh",
  "n10309147": "Menshevik",
  "n10309347": "mentioner",
  "n10309496": "mentor, wise man",
  "n10309614": "mercenary, soldier of fortune",
  "n10309785": "mercer",
  "n10309896": "merchant, merchandiser",
  "n10310404": "Merovingian",
  "n10310516": "meshuggeneh, meshuggener",
  "n10310647": "mesne lord",
  "n10310783": "Mesoamerican",
  "n10310903": "mesomorph",
  "n10311021": "messenger, courier",
  "n10311243": "bearer",
  "n10311375": "messenger boy, errand boy",
  "n10311506": "messmate",
  "n10311661": "mestiza",
  "n10311823": "mestizo, ladino",
  "n10311995": "metalhead",
  "n10312077": "metallurgist, metallurgical engineer",
  "n10312287": "meteorologist",
  "n10312491": "meter maid",
  "n10312600": "Methodist",
  "n10312773": "Wesleyan",
  "n10312890": "metic",
  "n10313000": "Metis",
  "n10313146": "metropolitan",
  "n10313239": "metropolitan",
  "n10313441": "mezzo-soprano, mezzo",
  "n10313580": "microbiologist",
  "n10313724": "microeconomist, microeconomic expert",
  "n10313872": "microscopist",
  "n10314054": "middle-aged man",
  "n10314182": "middlebrow",
  "n10314305": "middleweight",
  "n10314422": "middleweight",
  "n10314517": "middleweight",
  "n10314627": "midinette",
  "n10314703": "midshipman",
  "n10314836": "midwife, accoucheuse",
  "n10314952": "migrant, migrator",
  "n10315217": "mikado, tenno",
  "n10315382": "Milady",
  "n10315456": "Milanese",
  "n10315561": "miler",
  "n10315730": "miles gloriosus",
  "n10315837": "militant, activist",
  "n10316013": "militarist, warmonger",
  "n10316164": "military adviser, military advisor",
  "n10316360": "military attache",
  "n10316527": "military chaplain, padre, Holy Joe, sky pilot",
  "n10316683": "military governor",
  "n10316862": "military leader",
  "n10317007": "military officer, officer",
  "n10317500": "military policeman, MP",
  "n10317717": "militiaman",
  "n10317884": "milkman",
  "n10317963": "mill agent",
  "n10318087": "miller",
  "n10318193": "mill-girl",
  "n10318293": "mill-hand, factory worker",
  "n10318414": "millenarian, millenarist, chiliast",
  "n10318607": "millionairess",
  "n10318686": "millwright",
  "n10318802": "milord",
  "n10318892": "mime, mimer, mummer, pantomimer, pantomimist",
  "n10319163": "mimic, mimicker",
  "n10319313": "minder",
  "n10319580": "mind reader, telepathist, thought-reader",
  "n10319796": "miner, mineworker",
  "n10319996": "mineralogist",
  "n10320106": "miniaturist",
  "n10320230": "minimalist",
  "n10320354": "minimalist",
  "n10320484": "mining engineer",
  "n10320612": "minion",
  "n10320695": "minister, diplomatic minister",
  "n10320863": "minister, government minister",
  "n10321126": "ministrant",
  "n10321233": "minority leader",
  "n10321340": "minor leaguer, bush leaguer",
  "n10321474": "minstrel",
  "n10321632": "Minuteman",
  "n10321754": "miracle man, miracle worker",
  "n10321882": "misanthrope, misanthropist",
  "n10322084": "miser",
  "n10322238": "misfit",
  "n10322391": "misleader",
  "n10322546": "misogamist",
  "n10322648": "misogynist, woman hater",
  "n10322801": "missing link, ape-man",
  "n10322957": "missionary",
  "n10323182": "missionary, missioner",
  "n10323529": "missus, missis",
  "n10323634": "mistress",
  "n10323752": "mistress, kept woman, fancy woman",
  "n10323999": "mixed-blood",
  "n10324201": "mnemonist",
  "n10324357": "mod",
  "n10324560": "model, poser",
  "n10324851": "model, role model",
  "n10325013": "hero",
  "n10325243": "ideal, paragon, nonpareil, saint, apotheosis, nonesuch, nonsuch",
  "n10325549": "class act",
  "n10325656": "humdinger",
  "n10325774": "modeler, modeller",
  "n10325957": "moderationist",
  "n10326087": "moderator",
  "n10326257": "moderator",
  "n10326392": "moderator",
  "n10326551": "modern",
  "n10326643": "modernist",
  "n10326776": "modifier",
  "n10326883": "Mogul, Moghul",
  "n10327002": "Mohammedan, Muhammedan, Muhammadan",
  "n10327143": "molecular biologist",
  "n10327333": "molester",
  "n10327475": "moll, gun moll, gangster's moll",
  "n10327583": "mollycoddle",
  "n10327696": "Mon",
  "n10327824": "monarchist, royalist",
  "n10327987": "Monegasque, Monacan",
  "n10328123": "monetarist",
  "n10328328": "moneygrubber",
  "n10328437": "moneymaker",
  "n10328560": "mongoloid",
  "n10328696": "Mongoloid",
  "n10328782": "monogamist, monogynist",
  "n10328941": "monolingual",
  "n10329035": "monologist",
  "n10329143": "monomaniac",
  "n10329251": "Monophysite",
  "n10329337": "monopolist, monopolizer, monopoliser",
  "n10329579": "monotheist",
  "n10329675": "Monsieur",
  "n10329789": "Monsignor",
  "n10329945": "monster, fiend, devil, demon, ogre",
  "n10330189": "moocher, mooch, cadger, scrounger",
  "n10330439": "Moonie",
  "n10330593": "moonlighter",
  "n10330722": "mopper",
  "n10330835": "moppet",
  "n10330931": "moralist",
  "n10331098": "morosoph",
  "n10331167": "morris dancer",
  "n10331258": "mortal enemy",
  "n10331347": "mortgagee, mortgage holder",
  "n10331569": "mortgagor, mortgager",
  "n10331841": "mortician, undertaker, funeral undertaker, funeral director",
  "n10332016": "mossback",
  "n10332110": "moss-trooper",
  "n10332257": "most valuable player, MVP",
  "n10332385": "mother, female parent",
  "n10332861": "mother",
  "n10332953": "mother",
  "n10333044": "mother figure",
  "n10333165": "mother hen",
  "n10333317": "mother-in-law",
  "n10333439": "mother's boy, mamma's boy, mama's boy",
  "n10333601": "mother's daughter",
  "n10333719": "mother's son",
  "n10333838": "motorcycle cop, motorcycle policeman, speed cop",
  "n10334009": "motorcyclist",
  "n10334101": "motorist, automobilist",
  "n10334255": "motorman",
  "n10334335": "motormouth",
  "n10334461": "Mound Builder",
  "n10334567": "mountaineer, mountain climber",
  "n10334782": "mountebank, charlatan",
  "n10334957": "mounter, climber",
  "n10335123": "mounter",
  "n10335246": "mourner, griever, sorrower, lamenter",
  "n10335563": "mouse",
  "n10335679": "mouth",
  "n10335801": "mouthpiece, mouth",
  "n10335931": "mover",
  "n10336234": "mover",
  "n10336411": "moviegoer, motion-picture fan",
  "n10336537": "muckraker, mudslinger",
  "n10336730": "muezzin, muazzin, muadhdhin",
  "n10336904": "muffin man",
  "n10337020": "mufti",
  "n10337134": "muggee",
  "n10337300": "mugger",
  "n10337488": "mugwump, independent, fencesitter",
  "n10337645": "Mugwump",
  "n10337789": "mujahid",
  "n10337913": "mujtihad",
  "n10338094": "muleteer, mule skinner, mule driver, skinner",
  "n10338231": "Mullah, Mollah, Mulla",
  "n10338391": "muncher",
  "n10338498": "muralist",
  "n10338628": "murderee",
  "n10338707": "murderer, liquidator, manslayer",
  "n10339179": "murderess",
  "n10339251": "murder suspect",
  "n10339350": "muscleman, muscle",
  "n10339504": "muser, muller, ponderer, ruminator",
  "n10339717": "musher",
  "n10339856": "music critic",
  "n10339966": "musician",
  "n10340312": "musician, instrumentalist, player",
  "n10341243": "musicologist",
  "n10341343": "music teacher",
  "n10341446": "musketeer",
  "n10341573": "Muslimah",
  "n10341660": "mutant, mutation, variation, sport",
  "n10341955": "mutilator, maimer, mangler",
  "n10342180": "mutineer",
  "n10342367": "mute, deaf-mute, deaf-and-dumb person",
  "n10342543": "mutterer, mumbler, murmurer",
  "n10342770": "muzhik, moujik, mujik, muzjik",
  "n10342893": "muzzler",
  "n10342992": "Mycenaen",
  "n10343088": "mycologist",
  "n10343211": "mycophagist, mycophage",
  "n10343355": "myope",
  "n10343449": "myrmidon",
  "n10343554": "mystic, religious mystic",
  "n10343869": "mythologist",
  "n10344004": "nabob",
  "n10344121": "naif",
  "n10344203": "nailer",
  "n10344319": "namby-pamby",
  "n10344443": "name, figure, public figure",
  "n10344656": "name dropper",
  "n10344774": "namer",
  "n10344922": "namesake",
  "n10345015": "nan",
  "n10345100": "nanny, nursemaid, nurse",
  "n10345302": "narc, nark, narcotics agent",
  "n10345422": "narcissist, narcist",
  "n10345556": "narcoleptic",
  "n10345659": "nark, copper's nark",
  "n10345804": "narrator, storyteller, teller",
  "n10346015": "nationalist",
  "n10346198": "nationalist leader",
  "n10346392": "natural",
  "n10346514": "naturalist, natural scientist",
  "n10346955": "naturalist",
  "n10347110": "naturopath",
  "n10347204": "nautch girl",
  "n10347298": "naval attache",
  "n10347446": "naval commander",
  "n10347593": "naval officer",
  "n10347883": "navigator",
  "n10348365": "navigator",
  "n10348526": "Navy SEAL, SEAL",
  "n10348752": "nawab, nabob",
  "n10348874": "oblate",
  "n10348989": "obsessive",
  "n10349090": "obsessive-compulsive",
  "n10349243": "obstructionist, obstructor, obstructer, resister, thwarter",
  "n10349551": "naysayer",
  "n10349670": "Nazarene",
  "n10349750": "Nazarene",
  "n10349836": "Nazarene, Ebionite",
  "n10350220": "Nazi, German Nazi",
  "n10350484": "nazi",
  "n10350666": "Neapolitan",
  "n10350774": "nebbish, nebbech",
  "n10350896": "necessitarian",
  "n10351064": "necker",
  "n10351152": "necromancer",
  "n10351281": "needleworker",
  "n10351491": "negativist",
  "n10351625": "neglecter",
  "n10351874": "negotiator, negotiant, treater",
  "n10352208": "negotiatress, negotiatrix",
  "n10352299": "neighbor, neighbour",
  "n10352557": "neoclassicist",
  "n10352663": "neoconservative, neocon",
  "n10352782": "neoliberal",
  "n10352898": "neologist",
  "n10353016": "neonate, newborn, newborn infant, newborn baby",
  "n10353269": "Neoplatonist",
  "n10353355": "nephew",
  "n10353475": "nepotist",
  "n10353598": "nerd",
  "n10353734": "Nestorian",
  "n10353813": "neurasthenic",
  "n10353928": "neurobiologist",
  "n10354053": "neurolinguist",
  "n10354265": "neurologist, brain doctor",
  "n10354580": "neuroscientist",
  "n10354754": "neurosurgeon, brain surgeon",
  "n10354898": "neurotic, psychoneurotic, mental case",
  "n10355142": "neutral",
  "n10355306": "neutralist",
  "n10355449": "newcomer, fledgling, fledgeling, starter, neophyte, freshman, newbie, entrant",
  "n10355688": "newcomer",
  "n10355806": "New Dealer",
  "n10355942": "New Englander, Yankee",
  "n10356066": "newlywed, honeymooner",
  "n10356213": "newsagent, newsdealer, newsvendor, newsstand operator",
  "n10356344": "newscaster",
  "n10356450": "newspaper editor",
  "n10356592": "newspaper columnist",
  "n10356749": "newspaper critic",
  "n10356877": "newsreader, news reader",
  "n10357012": "Newtonian",
  "n10357112": "New Waver",
  "n10357207": "next friend",
  "n10357349": "next of kin",
  "n10357484": "nibbler",
  "n10357613": "niece",
  "n10357737": "niggard, skinflint, scrooge, churl",
  "n10357907": "night owl, nighthawk, nightbird",
  "n10358032": "night porter",
  "n10358124": "night rider, nightrider",
  "n10358322": "night watchman",
  "n10358420": "nihilist",
  "n10358575": "NIMBY",
  "n10358794": "nincompoop, poop, ninny",
  "n10358889": "ninja",
  "n10359117": "niqaabi",
  "n10359300": "Nisei",
  "n10359422": "nitpicker",
  "n10359546": "Nobelist, Nobel Laureate",
  "n10359659": "NOC",
  "n10359759": "nomad",
  "n10359988": "nominalist",
  "n10360101": "nominator",
  "n10360253": "nonagenarian",
  "n10360366": "noncandidate",
  "n10360578": "noncombatant",
  "n10360747": "noncommissioned officer, noncom, enlisted officer",
  "n10361060": "nondescript",
  "n10361194": "nondriver",
  "n10361296": "nonparticipant",
  "n10361390": "nonpartisan, nonpartizan",
  "n10361525": "nonperson, unperson",
  "n10361901": "nonreader",
  "n10362003": "nonresident",
  "n10362195": "non-resistant, passive resister",
  "n10362319": "nonsmoker",
  "n10362428": "normalizer, normaliser",
  "n10362557": "Northern Baptist",
  "n10362682": "Northerner",
  "n10362765": "nosher, snacker",
  "n10362917": "no-show, nonattender, truant",
  "n10363017": "no-show",
  "n10363149": "notary, notary public",
  "n10363360": "noticer",
  "n10363445": "noticer",
  "n10363573": "novelist",
  "n10363913": "novice, beginner, tyro, tiro, initiate",
  "n10364198": "novitiate, novice",
  "n10364333": "Nubian",
  "n10364502": "nuclear chemist, radiochemist",
  "n10364643": "nuclear physicist",
  "n10365399": "nude, nude person",
  "n10365514": "nudger",
  "n10365701": "nudist, naturist",
  "n10365846": "nudnik, nudnick",
  "n10365984": "nullifier",
  "n10366145": "nullipara",
  "n10366276": "number theorist",
  "n10366383": "numerologist",
  "n10366484": "numen",
  "n10366692": "Numidian",
  "n10366779": "numismatist, numismatologist, coin collector",
  "n10366966": "nurse",
  "n10367409": "nurser",
  "n10367548": "nursing aide, nurse's aide",
  "n10367694": "nurse-midwife",
  "n10367819": "nurse practitioner, NP, nurse clinician",
  "n10368009": "nun",
  "n10368113": "nuncio, papal nuncio",
  "n10368291": "nursling, nurseling, suckling",
  "n10368414": "nutter, wacko, whacko",
  "n10368528": "nymph, houri",
  "n10368624": "nymphet",
  "n10368711": "nympholept",
  "n10368798": "nymphomaniac, nympho",
  "n10368920": "oarsman, rower",
  "n10369095": "oarswoman",
  "n10369166": "obliger, accommodator",
  "n10369317": "oboist",
  "n10369417": "obscurantist",
  "n10369528": "observer, commentator",
  "n10369699": "obstetrician, accoucheur",
  "n10369845": "Occidental",
  "n10369955": "occupier",
  "n10370122": "oceanographer",
  "n10370286": "octogenarian",
  "n10370381": "occultist",
  "n10370603": "odalisque",
  "n10370683": "odds-maker, handicapper",
  "n10370881": "odist",
  "n10370955": "wine lover",
  "n10371052": "offerer, offeror",
  "n10371221": "office-bearer",
  "n10371330": "office boy",
  "n10371450": "officeholder, officer",
  "n10371741": "officer, ship's officer",
  "n10372076": "official",
  "n10372373": "official, functionary",
  "n10373390": "officiant",
  "n10373525": "Federal, Fed, federal official",
  "n10373639": "Federal, Federal soldier, Union soldier",
  "n10373801": "agent, federal agent",
  "n10373998": "offspring, progeny, issue",
  "n10374282": "ogler",
  "n10374415": "oiler",
  "n10374541": "oilman",
  "n10374652": "oilman",
  "n10374762": "oil painter",
  "n10374849": "oil tycoon",
  "n10374943": "old-age pensioner",
  "n10375052": "old boy",
  "n10375129": "old boy",
  "n10375214": "old boy, old man",
  "n10375314": "old lady",
  "n10375402": "old man",
  "n10375506": "old man, greybeard, graybeard, Methuselah",
  "n10375690": "old man",
  "n10375794": "old master",
  "n10376523": "oldster, old person, senior citizen, golden ager",
  "n10376890": "old-timer, oldtimer, gaffer, old geezer, antique",
  "n10377021": "old woman",
  "n10377185": "oligarch",
  "n10377291": "Olympian",
  "n10377394": "ombudsman",
  "n10377542": "omnivore",
  "n10377633": "oncologist",
  "n10377732": "oneiromancer",
  "n10377865": "one of the boys",
  "n10378026": "onlooker, looker-on",
  "n10378113": "onomancer",
  "n10378213": "operagoer",
  "n10378290": "opera star, operatic star",
  "n10378412": "operator, manipulator",
  "n10378780": "operator",
  "n10378952": "operator",
  "n10379073": "ophthalmologist, eye doctor, oculist",
  "n10379278": "opium addict, opium taker",
  "n10379376": "opportunist, self-seeker",
  "n10379620": "opposition, opponent, opposite",
  "n10379758": "oppressor",
  "n10380000": "optician, lens maker",
  "n10380126": "optimist",
  "n10380305": "optometrist, oculist",
  "n10380499": "Orangeman",
  "n10380672": "orator, speechmaker, rhetorician, public speaker, speechifier",
  "n10381100": "orchestrator",
  "n10381214": "ordainer",
  "n10381369": "orderer, systematizer, systematiser, systemizer, systemiser, systematist",
  "n10381698": "orderer",
  "n10381804": "orderly, hospital attendant",
  "n10381981": "orderly",
  "n10382157": "orderly sergeant",
  "n10382302": "ordinand",
  "n10382380": "ordinary",
  "n10382480": "ordinary",
  "n10382597": "organ donor",
  "n10382710": "organ-grinder",
  "n10382825": "organist",
  "n10383094": "organization man",
  "n10383237": "organizer, organiser, arranger",
  "n10383505": "organizer, organiser, labor organizer",
  "n10383689": "orientalist",
  "n10383816": "originator, conceiver, mastermind",
  "n10384054": "Orleanist",
  "n10384214": "ornithologist, bird watcher",
  "n10384392": "orphan",
  "n10384496": "orphan",
  "n10384610": "orthodontist",
  "n10384772": "Orthodox Jew",
  "n10384935": "orthoepist",
  "n10385159": "orthopedist, orthopaedist",
  "n10385354": "orthoptist",
  "n10385455": "osteologist, osteologer",
  "n10385566": "osteopath, osteopathist",
  "n10385707": "ostrich",
  "n10385923": "Ostrogoth",
  "n10386071": "ouster, ejector",
  "n10386196": "out-and-outer",
  "n10386312": "outcast, castaway, pariah, Ishmael",
  "n10386515": "outcaste",
  "n10386618": "outdoorsman",
  "n10386754": "outdoorswoman",
  "n10386874": "outfielder",
  "n10386984": "outfielder",
  "n10387196": "right fielder",
  "n10387324": "right-handed pitcher, right-hander",
  "n10387476": "center fielder, centerfielder",
  "n10387586": "left fielder",
  "n10387712": "outfitter",
  "n10387836": "outlier",
  "n10387936": "outpatient",
  "n10388082": "outrider",
  "n10388192": "outsider",
  "n10388321": "overachiever",
  "n10388440": "overlord, master, lord",
  "n10388647": "overnighter",
  "n10388732": "overseer, superintendent",
  "n10388924": "owner, proprietor",
  "n10389398": "owner, possessor",
  "n10389760": "owner-driver",
  "n10389865": "owner-occupier",
  "n10389976": "oyabun",
  "n10390050": "pachuco",
  "n10390199": "pacifist, pacificist, disarmer",
  "n10390427": "packer, bagger, boxer",
  "n10390600": "packrat",
  "n10390698": "padrone",
  "n10390807": "padrone",
  "n10390902": "pagan",
  "n10391086": "page, varlet",
  "n10391248": "page",
  "n10391416": "page, pageboy",
  "n10391513": "pain, pain in the neck, nuisance",
  "n10391653": "painter",
  "n10393909": "painter",
  "n10394141": "palatine, palsgrave",
  "n10394329": "palatine",
  "n10394434": "Paleo-American, Paleo-Amerind, Paleo-Indian",
  "n10394673": "paleographer, paleographist",
  "n10394786": "paleontologist, palaeontologist, fossilist",
  "n10395073": "pallbearer, bearer",
  "n10395209": "palmist, palmister, chiromancer",
  "n10395390": "pamperer, spoiler, coddler, mollycoddler",
  "n10395605": "pamphleteer",
  "n10395828": "Panchen Lama",
  "n10395926": "panderer",
  "n10396106": "panelist, panellist",
  "n10396337": "panhandler",
  "n10396462": "pansexual",
  "n10396594": "pantheist",
  "n10396727": "paparazzo",
  "n10396908": "paperboy",
  "n10397001": "paperhanger, paperer",
  "n10397142": "paperhanger",
  "n10397275": "paper-pusher",
  "n10397392": "papoose, pappoose",
  "n10397482": "parachutist, parachuter, parachute jumper",
  "n10397694": "paragrapher",
  "n10397886": "paralegal, legal assistant",
  "n10398033": "paralytic",
  "n10398176": "paramedic, paramedical",
  "n10398370": "paranoid, paranoiac",
  "n10398487": "paraplegic",
  "n10398624": "paraprofessional",
  "n10398806": "parapsychologist",
  "n10399019": "paratrooper, para",
  "n10399130": "pardoner",
  "n10399299": "pardoner, forgiver, excuser",
  "n10399491": "parent",
  "n10399895": "parer",
  "n10400003": "paretic",
  "n10400108": "parishioner",
  "n10400205": "park commissioner",
  "n10400309": "parliamentarian",
  "n10400437": "Parliamentarian, Member of Parliament",
  "n10400618": "parliamentary agent",
  "n10400826": "parlormaid, parlourmaid",
  "n10400998": "parodist, lampooner",
  "n10401204": "parricide",
  "n10401331": "parrot",
  "n10401468": "Parsee, Parsi",
  "n10401639": "partaker, sharer",
  "n10401829": "participant",
  "n10402086": "partisan, zealot, drumbeater",
  "n10402285": "partitionist",
  "n10402417": "partner",
  "n10402603": "part-owner",
  "n10402709": "part-timer",
  "n10402824": "party",
  "n10403162": "party boss, political boss, boss",
  "n10403366": "party girl",
  "n10403490": "partygoer",
  "n10403633": "party man, party liner",
  "n10403768": "pasha, pacha",
  "n10403876": "passenger, rider",
  "n10404242": "passer, forward passer",
  "n10404426": "passer",
  "n10404550": "passer",
  "n10404672": "passerby, passer-by, passer",
  "n10404810": "passive source",
  "n10404998": "paster",
  "n10405107": "past master",
  "n10405229": "past master",
  "n10405320": "pastry cook",
  "n10405410": "patentee",
  "n10405540": "pater",
  "n10405694": "patient",
  "n10406072": "patrial",
  "n10406266": "patriarch",
  "n10406391": "patriarch",
  "n10406765": "patriarch, paterfamilias",
  "n10406905": "patriarch",
  "n10407105": "patrician",
  "n10407221": "patricide",
  "n10407310": "patriot, nationalist",
  "n10407552": "patroller",
  "n10407726": "patron, frequenter",
  "n10407954": "patron, sponsor, supporter",
  "n10408324": "patron",
  "n10408438": "patroness, patronne",
  "n10408552": "patron saint",
  "n10408809": "patternmaker",
  "n10408939": "patzer",
  "n10409011": "pauper",
  "n10409161": "pavement artist",
  "n10409322": "pawer",
  "n10409459": "pawnbroker",
  "n10409634": "payee",
  "n10409752": "payer, remunerator",
  "n10410035": "paymaster",
  "n10410125": "peacekeeper",
  "n10410246": "peacekeeper",
  "n10410440": "peanut",
  "n10410531": "pearl diver, pearler",
  "n10410668": "peasant, provincial, bucolic",
  "n10410815": "peasant, barbarian, boor, churl, Goth, tyke, tike",
  "n10410996": "peasant",
  "n10411163": "pedaler, pedaller",
  "n10411356": "pedant, bookworm, scholastic",
  "n10411551": "peddler, pedlar, packman, hawker, pitchman",
  "n10411867": "pederast, paederast, child molester",
  "n10412055": "pedestrian, walker, footer",
  "n10412554": "pedodontist",
  "n10412669": "pedophile, paedophile",
  "n10412784": "peeler",
  "n10412910": "peer",
  "n10413276": "peer of the realm",
  "n10413429": "pelter",
  "n10413588": "pendragon",
  "n10413688": "penetrator",
  "n10413834": "penitent",
  "n10414084": "penny pincher",
  "n10414239": "penologist",
  "n10414379": "pen pal, pen-friend",
  "n10414507": "penpusher, pencil pusher",
  "n10414612": "pensioner, pensionary",
  "n10414768": "pentathlete",
  "n10414865": "Pentecostal, Pentecostalist",
  "n10415037": "percussionist",
  "n10415230": "perfectionist",
  "n10415439": "perfecter",
  "n10415638": "performer, performing artist",
  "n10416270": "perfumer",
  "n10416364": "peri",
  "n10416444": "perinatologist",
  "n10416567": "periodontist",
  "n10416732": "peripatetic",
  "n10416828": "perisher",
  "n10416909": "perjurer, false witness",
  "n10417045": "peroxide blond, peroxide blonde",
  "n10417168": "perpetrator, culprit",
  "n10417288": "peshmerga",
  "n10417424": "personality",
  "n10417551": "personal representative",
  "n10417682": "personage",
  "n10417843": "persona grata",
  "n10417969": "persona non grata",
  "n10418101": "personification",
  "n10418302": "embodiment, incarnation, avatar",
  "n10418577": "deification",
  "n10418735": "perspirer, sweater",
  "n10418841": "persuader, inducer",
  "n10419047": "pervert, deviant, deviate, degenerate",
  "n10419472": "pessimist",
  "n10419630": "pest, blighter, cuss, pesterer, gadfly",
  "n10419785": "Peter Pan",
  "n10419916": "petit bourgeois",
  "n10420031": "petitioner, suppliant, supplicant, requester",
  "n10420277": "petit juror, petty juror",
  "n10420392": "petroleum geologist, oil geologist",
  "n10420507": "pet sitter, critter sitter",
  "n10420649": "petter, fondler",
  "n10420809": "petty officer, PO, P.O.",
  "n10421016": "Pharaoh, Pharaoh of Egypt",
  "n10421183": "Pharisee",
  "n10421339": "pharisee",
  "n10421470": "pharmacist, druggist, chemist, apothecary, pill pusher, pill roller",
  "n10421753": "pharmacologist, pharmaceutical chemist",
  "n10421956": "philanthropist, altruist",
  "n10422405": "philatelist, stamp collector",
  "n10422540": "philhellene, philhellenist, Graecophile",
  "n10422759": "Philippian",
  "n10422871": "Philistine",
  "n10423031": "philistine, anti-intellectual, lowbrow",
  "n10423225": "philologist, philologue",
  "n10423514": "philomath",
  "n10423589": "philosopher",
  "n10425946": "philosopher",
  "n10426184": "philosophizer, philosophiser",
  "n10426363": "phlebotomist",
  "n10426454": "phonetician",
  "n10426630": "phonologist",
  "n10426749": "photographer, lensman",
  "n10427103": "photographer's model",
  "n10427223": "photojournalist",
  "n10427359": "photometrist, photometrician",
  "n10427467": "phrenologist, craniologist",
  "n10427658": "Phrygian",
  "n10427764": "physical therapist, physiotherapist",
  "n10428004": "physicist",
  "n10429965": "physiologist",
  "n10430554": "phytochemist",
  "n10430665": "pianist, piano player",
  "n10431122": "piano maker",
  "n10431224": "piano teacher",
  "n10431330": "pickaninny, piccaninny, picaninny",
  "n10431514": "picker",
  "n10431625": "picker, chooser, selector",
  "n10431770": "picket",
  "n10431907": "pickpocket, cutpurse, dip",
  "n10432053": "pickup",
  "n10432189": "picnicker, picknicker",
  "n10432320": "pied piper",
  "n10432441": "pilgrim",
  "n10432532": "pilgrim",
  "n10432674": "Pilgrim, Pilgrim Father",
  "n10432875": "pill",
  "n10432957": "pillar, mainstay",
  "n10433077": "pill head",
  "n10433164": "pilot, airplane pilot",
  "n10433452": "pilot",
  "n10433610": "Piltdown man, Piltdown hoax",
  "n10433737": "pimp, procurer, panderer, pander, pandar, fancy man, ponce",
  "n10434054": "pinchgut",
  "n10434160": "pinch hitter",
  "n10434321": "pinko, pink",
  "n10434424": "pioneer",
  "n10434725": "pioneer, innovator, trailblazer, groundbreaker",
  "n10434947": "pipe major",
  "n10435041": "piper, bagpiper",
  "n10435169": "pipe smoker",
  "n10435251": "pip-squeak, squirt, small fry",
  "n10435367": "pirate, buccaneer, sea robber, sea rover",
  "n10435716": "pisser, urinator",
  "n10435855": "pistoleer",
  "n10435988": "pitcher, hurler, twirler",
  "n10436334": "pitchman",
  "n10436458": "pituitary dwarf, hypophysial dwarf, Levi-Lorrain dwarf",
  "n10436707": "pivot, pivot man",
  "n10436851": "place-kicker, placekicker",
  "n10437014": "placeman, placeseeker",
  "n10437137": "placer miner",
  "n10437262": "plagiarist, plagiarizer, plagiariser, literary pirate, pirate",
  "n10437590": "plainclothesman",
  "n10437698": "plainsman",
  "n10437852": "plaintiff, complainant",
  "n10438042": "plaiter",
  "n10438172": "planner, contriver, deviser",
  "n10438470": "plant",
  "n10438619": "planter, plantation owner",
  "n10438728": "planter",
  "n10438842": "plasterer",
  "n10438952": "plaster saint",
  "n10439087": "platelayer, tracklayer",
  "n10439203": "plater",
  "n10439373": "platinum blond, platinum blonde",
  "n10439523": "platitudinarian",
  "n10439629": "Platonist",
  "n10439727": "playboy, man-about-town, Corinthian",
  "n10439851": "player, participant",
  "n10440387": "player",
  "n10440580": "player",
  "n10440717": "playgoer, theatergoer, theatregoer",
  "n10440886": "playmaker",
  "n10441037": "playmate, playfellow",
  "n10441124": "pleaser",
  "n10441251": "plebeian, pleb",
  "n10441410": "pledge",
  "n10441534": "pledgee",
  "n10441694": "pledger",
  "n10441819": "pledge taker",
  "n10441962": "plenipotentiary",
  "n10442093": "plier, plyer",
  "n10442232": "plodder, slowpoke, stick-in-the-mud, slowcoach",
  "n10442417": "plodder, slogger",
  "n10442573": "plotter, mapper",
  "n10442705": "plowboy, ploughboy",
  "n10442815": "plowman, ploughman, plower",
  "n10442923": "plowwright, ploughwright",
  "n10443032": "plumber, pipe fitter",
  "n10443170": "plunderer, pillager, looter, spoiler, despoiler, raider, freebooter",
  "n10443482": "pluralist",
  "n10443659": "pluralist",
  "n10443830": "pluralist",
  "n10443955": "plutocrat",
  "n10444058": "poacher",
  "n10444194": "poet",
  "n10447255": "poetess",
  "n10447359": "poet laureate",
  "n10447585": "poet laureate",
  "n10447828": "poilu",
  "n10447924": "pointillist",
  "n10448065": "point man",
  "n10448157": "point man",
  "n10448322": "pointsman",
  "n10448455": "point woman",
  "n10448568": "poisoner",
  "n10448670": "polemicist, polemist, polemic",
  "n10448834": "police commissioner",
  "n10448983": "policeman, police officer, officer",
  "n10449412": "police matron, policewoman",
  "n10449521": "police sergeant, sergeant",
  "n10449664": "policyholder",
  "n10449845": "policy maker",
  "n10450038": "political prisoner",
  "n10450161": "political scientist",
  "n10450303": "politician, politico, pol, political leader",
  "n10451263": "politician",
  "n10451450": "politician",
  "n10451590": "pollster, poll taker, headcounter, canvasser",
  "n10451858": "polluter, defiler",
  "n10452024": "poltroon, craven, recreant",
  "n10452151": "polyandrist",
  "n10452260": "polygamist",
  "n10452432": "polygynist",
  "n10452535": "polytheist",
  "n10452631": "pomologist",
  "n10452752": "ponce",
  "n10452892": "pontifex",
  "n10453061": "pooler",
  "n10453184": "pool player",
  "n10453265": "poor devil, wretch",
  "n10453357": "poor person, have-not",
  "n10453533": "pope, Catholic Pope, Roman Catholic Pope, pontiff, Holy Father, Vicar of Christ, Bishop of Rome",
  "n10454645": "popinjay",
  "n10454752": "popularizer, populariser, vulgarizer, vulgariser",
  "n10454972": "pork butcher",
  "n10455094": "pornographer, porn merchant",
  "n10455305": "porter, Pullman porter",
  "n10455447": "porter",
  "n10455619": "portraitist, portrait painter, portrayer, limner",
  "n10455821": "portwatcher, port watcher",
  "n10455915": "poseur, poser",
  "n10456070": "poseuse",
  "n10456138": "positivist, rationalist",
  "n10456391": "posseman",
  "n10456514": "possible",
  "n10456603": "postdiluvian",
  "n10456696": "postdoc, post doc",
  "n10456874": "poster boy",
  "n10456950": "poster child",
  "n10457214": "poster girl",
  "n10457293": "postmature infant",
  "n10457444": "postulator",
  "n10457597": "postulator",
  "n10457777": "posturer",
  "n10457903": "private citizen",
  "n10458024": "probable",
  "n10458111": "problem solver, solver, convergent thinker",
  "n10458356": "pro-lifer",
  "n10458519": "proprietress",
  "n10458596": "prosthetist",
  "n10458696": "prosthodontist",
  "n10458834": "provincial",
  "n10459116": "postal clerk, mail clerk",
  "n10459212": "postilion, postillion",
  "n10459414": "Postimpressionist, Post-impressionist",
  "n10459575": "postmaster",
  "n10459687": "postmistress",
  "n10459764": "postmaster general",
  "n10459882": "postulant",
  "n10460033": "potboy, potman",
  "n10460193": "pothead",
  "n10460286": "potholer, spelunker, speleologist, spelaeologist",
  "n10460501": "pothunter",
  "n10460599": "pothunter",
  "n10460720": "pothunter",
  "n10460806": "potter, thrower, ceramicist, ceramist",
  "n10461060": "poultryman, poulterer",
  "n10461169": "powderer",
  "n10461280": "powder monkey",
  "n10461424": "power, force",
  "n10461747": "influence",
  "n10462005": "Moloch",
  "n10462217": "power broker, powerbroker",
  "n10462429": "powerhouse, human dynamo, ball of fire, fireball",
  "n10462588": "power user",
  "n10462751": "power worker, power-station worker",
  "n10462860": "practitioner, practician",
  "n10463028": "praetor, pretor",
  "n10463259": "Praetorian Guard, Praetorian",
  "n10463386": "pragmatist",
  "n10463582": "pragmatist",
  "n10463714": "prankster, cut-up, trickster, tricker, hoaxer, practical joker",
  "n10463943": "prattler",
  "n10464052": "prayer, supplicant",
  "n10464178": "preacher, preacher man, sermonizer, sermoniser",
  "n10464432": "prebendary",
  "n10464542": "preceptor, don",
  "n10464711": "predecessor",
  "n10464870": "preemptor, pre-emptor",
  "n10465002": "preemptor, pre-emptor",
  "n10465128": "prefect",
  "n10465248": "Pre-Raphaelite",
  "n10465451": "premature baby, preterm baby, premature infant, preterm infant, preemie, premie",
  "n10465635": "presbyope",
  "n10465831": "presbyter",
  "n10465922": "Presbyterian",
  "n10466060": "preschooler, kindergartner, kindergartener",
  "n10466198": "presenter, sponsor",
  "n10466387": "presenter",
  "n10466564": "presentist",
  "n10466759": "preservationist",
  "n10466918": "preserver",
  "n10467052": "preserver",
  "n10467179": "president",
  "n10467395": "President of the United States, United States President, President, Chief Executive",
  "n10468559": "president",
  "n10468750": "president, prexy",
  "n10468962": "president, chairman, chairwoman, chair, chairperson",
  "n10469346": "presiding officer",
  "n10469511": "pre-Socratic",
  "n10469611": "press agent, publicity man, public relations man, PR man",
  "n10469786": "press lord",
  "n10469874": "press photographer",
  "n10469979": "Pretender",
  "n10470132": "preterist",
  "n10470314": "prevailing party",
  "n10470460": "prey, quarry, target, fair game",
  "n10470779": "priest",
  "n10471250": "priest, non-Christian priest",
  "n10471570": "priestess",
  "n10471640": "prima ballerina",
  "n10471732": "prima donna, diva",
  "n10471859": "prima donna",
  "n10471948": "primary care physician",
  "n10472129": "primigravida, gravida I",
  "n10472274": "primipara, para I",
  "n10472447": "primordial dwarf, hypoplastic dwarf, true dwarf, normal dwarf",
  "n10472690": "primus",
  "n10472799": "prince",
  "n10473273": "Elector",
  "n10473453": "prince charming",
  "n10473562": "prince consort",
  "n10473718": "princeling",
  "n10473789": "princeling",
  "n10473917": "Prince of Wales",
  "n10474064": "princess",
  "n10474343": "princess royal",
  "n10474446": "principal, dealer",
  "n10474645": "principal, school principal, head teacher, head",
  "n10474950": "principal",
  "n10475163": "principal investigator, PI",
  "n10475297": "printer, pressman",
  "n10475584": "printer's devil",
  "n10475687": "printmaker, graphic artist",
  "n10475835": "print seller",
  "n10475940": "prior",
  "n10476086": "prisoner, captive",
  "n10476331": "prisoner of war, POW",
  "n10476467": "private, buck private, common soldier",
  "n10476671": "private detective, PI, private eye, private investigator, operative, shamus, sherlock",
  "n10476928": "privateer, privateersman",
  "n10477077": "prizefighter, gladiator",
  "n10477465": "probability theorist",
  "n10477585": "probationer, parolee",
  "n10477713": "probationer, student nurse",
  "n10477839": "probation officer",
  "n10477955": "processor",
  "n10478118": "process-server",
  "n10478293": "proconsul",
  "n10478462": "proconsul",
  "n10478626": "procrastinator, postponer, cunctator",
  "n10478827": "proctologist",
  "n10478960": "proctor, monitor",
  "n10479135": "procurator",
  "n10479328": "procurer, securer",
  "n10479493": "procuress",
  "n10479561": "prodigal, profligate, squanderer",
  "n10479783": "prodigy",
  "n10480018": "producer",
  "n10480253": "professional, professional person",
  "n10480583": "professional, pro",
  "n10480730": "professor, prof",
  "n10481003": "profiteer",
  "n10481167": "profit taker",
  "n10481268": "programmer, computer programmer, coder, software engineer",
  "n10481561": "projectionist",
  "n10481711": "proletarian, prole, worker",
  "n10481929": "promisee",
  "n10482054": "promiser, promisor",
  "n10482220": "promoter, booster, plugger",
  "n10482414": "prompter, theater prompter",
  "n10482587": "promulgator",
  "n10482768": "proofreader, reader",
  "n10482921": "propagandist",
  "n10483138": "propagator, disseminator",
  "n10483274": "propagator",
  "n10483395": "property man, propman, property master",
  "n10483530": "prophet, prophesier, oracle, seer, vaticinator",
  "n10483799": "prophetess",
  "n10483890": "prophet",
  "n10484526": "proposer, mover",
  "n10484739": "propositus",
  "n10484858": "prosecutor, public prosecutor, prosecuting officer, prosecuting attorney",
  "n10485168": "proselyte",
  "n10485298": "prospector",
  "n10485440": "prostitute, cocotte, whore, harlot, bawd, tart, cyprian, fancy woman, working girl, sporting lady, lady of pleasure, woman of the street",
  "n10485883": "protectionist",
  "n10485989": "protege",
  "n10486166": "protegee",
  "n10486236": "protozoologist",
  "n10486349": "provider",
  "n10486468": "provost",
  "n10486561": "provost marshal",
  "n10486679": "prowler, sneak, stalker",
  "n10486909": "proxy, placeholder, procurator",
  "n10487026": "prude, puritan",
  "n10487182": "pruner, trimmer",
  "n10487363": "psalmist",
  "n10487592": "psephologist",
  "n10487710": "pseudohermaphrodite",
  "n10488016": "psychiatrist, head-shrinker, shrink",
  "n10488309": "psychic",
  "n10488513": "spirit rapper",
  "n10488656": "psycholinguist",
  "n10488865": "psychologist",
  "n10489426": "psychophysicist",
  "n10489564": "sociopath, psychopath",
  "n10489818": "psychopomp",
  "n10489944": "psychotherapist, clinical psychologist",
  "n10490141": "psychotic, psychotic person, psycho",
  "n10490330": "pteridologist",
  "n10490421": "publican, tavern keeper",
  "n10490557": "public defender",
  "n10490699": "publicist, publicizer, publiciser",
  "n10490965": "public relations person",
  "n10491136": "public servant",
  "n10491309": "publisher",
  "n10491575": "publisher, newspaper publisher",
  "n10491869": "puddler",
  "n10491998": "pudge",
  "n10492086": "puerpera",
  "n10492202": "puller, tugger, dragger",
  "n10492447": "puller",
  "n10492627": "puncher",
  "n10492727": "punching bag",
  "n10492894": "punk rocker, punk",
  "n10493093": "punster",
  "n10493199": "punter",
  "n10493419": "punter",
  "n10493528": "puppet ruler, puppet leader",
  "n10493685": "puppeteer",
  "n10493835": "puppy, pup",
  "n10493922": "purchasing agent",
  "n10494037": "purist",
  "n10494195": "puritan",
  "n10494373": "Puritan",
  "n10494643": "purser",
  "n10494778": "pursued, chased",
  "n10494935": "pursuer, chaser",
  "n10495167": "pursuer",
  "n10495299": "purveyor",
  "n10495421": "pusher, shover",
  "n10495555": "pusher, drug peddler, peddler, drug dealer, drug trafficker",
  "n10495756": "pusher, thruster",
  "n10495880": "pushover",
  "n10495975": "pussycat",
  "n10496081": "putter",
  "n10496193": "putterer, potterer",
  "n10496393": "putz",
  "n10496489": "Pygmy, Pigmy",
  "n10496626": "pygmy, pigmy",
  "n10496715": "pyrographer",
  "n10496825": "pyromancer",
  "n10496927": "python",
  "n10497046": "pythoness",
  "n10497135": "qadi",
  "n10497202": "quack",
  "n10497373": "quadripara",
  "n10497534": "quadriplegic",
  "n10497645": "quadruplet, quad",
  "n10497775": "quaestor",
  "n10497946": "quaffer",
  "n10498046": "quaker, trembler",
  "n10498196": "qualifier",
  "n10498422": "quarreler, quarreller",
  "n10498551": "quarryman, quarrier",
  "n10498699": "quarter",
  "n10498816": "quarterback, signal caller, field general",
  "n10498986": "quartermaster",
  "n10499110": "quartermaster general",
  "n10499232": "Quebecois",
  "n10499355": "queen, queen regnant, female monarch",
  "n10499631": "Queen of England",
  "n10499857": "queen",
  "n10500217": "queen",
  "n10500419": "queen consort",
  "n10500506": "queen dowager",
  "n10500603": "queen mother",
  "n10500717": "queen regent",
  "n10500824": "Queen's Counsel",
  "n10500942": "question master, quizmaster",
  "n10501097": "Quetzalcoatl",
  "n10501203": "quibbler, caviller, caviler, pettifogger",
  "n10501453": "quick study, sponge",
  "n10501635": "quietist",
  "n10501747": "quintipara",
  "n10501908": "quintuplet, quint, quin",
  "n10502046": "quitter",
  "n10502204": "quoter",
  "n10502329": "rabbi",
  "n10502576": "racer, race driver, automobile driver",
  "n10502762": "racetrack tout",
  "n10502950": "racist, racialist",
  "n10503138": "racker",
  "n10503247": "racketeer",
  "n10503452": "radical",
  "n10503710": "radio announcer",
  "n10503818": "radiobiologist",
  "n10503965": "radiographer",
  "n10504090": "radiologic technologist",
  "n10504206": "radiologist, radiotherapist",
  "n10504426": "radio operator",
  "n10504545": "raftsman, raftman, rafter",
  "n10504664": "ragamuffin, tatterdemalion",
  "n10504805": "ragpicker",
  "n10504956": "ragsorter",
  "n10505085": "railbird",
  "n10505206": "rail-splitter, splitter",
  "n10505347": "rainmaker",
  "n10505459": "rainmaker",
  "n10505613": "raiser",
  "n10505732": "raja, rajah",
  "n10505816": "Rajput, Rajpoot",
  "n10505942": "rake, rakehell, profligate, rip, blood, roue",
  "n10506094": "rambler",
  "n10506220": "rambler",
  "n10506336": "ramrod",
  "n10506417": "rancher",
  "n10506544": "ranch hand",
  "n10506642": "rani, ranee",
  "n10506762": "ranker",
  "n10506915": "ranker",
  "n10507070": "ranter, raver",
  "n10507230": "raper, rapist",
  "n10507380": "rape suspect",
  "n10507482": "rapper",
  "n10507565": "rapporteur",
  "n10507692": "rare bird, rara avis",
  "n10507783": "Raskolnikov, Rodya Raskolnikov",
  "n10508008": "rat-catcher, disinfestation officer",
  "n10508141": "ratepayer",
  "n10508272": "raver",
  "n10508379": "raw recruit",
  "n10508475": "reactionary, ultraconservative, extreme right-winger",
  "n10508710": "reader",
  "n10508862": "reader",
  "n10509063": "reading teacher",
  "n10509161": "realist",
  "n10509389": "realist",
  "n10509605": "realist",
  "n10509810": "real estate broker, real estate agent, estate agent, land agent, house agent",
  "n10510078": "Realtor",
  "n10510245": "rear admiral",
  "n10510339": "reasoner, ratiocinator",
  "n10510546": "rebutter, disprover, refuter, confuter",
  "n10510818": "receiver, pass receiver, pass catcher",
  "n10510974": "receiver",
  "n10511069": "receptionist",
  "n10511239": "recidivist, backslider, reversionist",
  "n10511425": "recidivist, repeater, habitual criminal",
  "n10511649": "recitalist",
  "n10511771": "reciter",
  "n10511856": "record-breaker, record-holder",
  "n10511960": "recorder",
  "n10512109": "recorder player",
  "n10512201": "recruit, enlistee",
  "n10512372": "recruit, military recruit",
  "n10512562": "recruiter",
  "n10512708": "recruiter",
  "n10512859": "recruiting-sergeant",
  "n10512982": "rectifier",
  "n10513120": "redact, redactor, reviser, rewriter, rewrite man",
  "n10513386": "redcap",
  "n10513509": "redcap",
  "n10513623": "redeemer",
  "n10513823": "redhead, redheader, red-header, carrottop",
  "n10513938": "redneck, cracker",
  "n10514051": "reeler",
  "n10514121": "reenactor",
  "n10514255": "referral",
  "n10514429": "referee, ref",
  "n10514643": "referee",
  "n10514784": "refiner",
  "n10514962": "refinisher, renovator, restorer, preserver",
  "n10515194": "reformer, reformist, crusader, social reformer, meliorist",
  "n10515863": "Reform Jew",
  "n10516016": "refugee",
  "n10516117": "regent",
  "n10516294": "regent, trustee",
  "n10516422": "regicide",
  "n10516527": "registered nurse, RN",
  "n10516692": "registrant",
  "n10516874": "registrar, record-keeper, recorder",
  "n10517032": "registrar",
  "n10517137": "registrar",
  "n10517283": "Regius professor",
  "n10517405": "regular, habitue, fixture",
  "n10517583": "regular",
  "n10517743": "regular",
  "n10517826": "regulator",
  "n10518003": "reincarnation",
  "n10518194": "reliever, relief pitcher, fireman",
  "n10518349": "reliever, allayer, comforter",
  "n10518602": "religious",
  "n10518945": "eremite",
  "n10519126": "anchorite, hermit",
  "n10519291": "cenobite, coenobite",
  "n10519494": "religious leader",
  "n10519884": "remittance man",
  "n10519984": "remover",
  "n10520109": "Renaissance man",
  "n10520286": "Renaissance man, generalist",
  "n10520544": "renegade",
  "n10520675": "rent collector",
  "n10520804": "renter",
  "n10520964": "rentier",
  "n10521100": "repairman, maintenance man, service man",
  "n10521288": "repatriate",
  "n10521470": "repeater",
  "n10521662": "reporter, newsman, newsperson",
  "n10521853": "newswoman",
  "n10521928": "repository, secretary",
  "n10522035": "representative",
  "n10522324": "reprobate, miscreant",
  "n10522495": "republican",
  "n10522633": "Republican",
  "n10522759": "rescuer, recoverer, saver",
  "n10522956": "research director, director of research",
  "n10523076": "research worker, researcher, investigator",
  "n10523341": "reservist",
  "n10523519": "resident, occupant, occupier",
  "n10524076": "resident commissioner",
  "n10524223": "respecter",
  "n10524413": "respondent, responder, answerer",
  "n10524711": "respondent",
  "n10524869": "restaurateur, restauranter",
  "n10524973": "rester",
  "n10525134": "restrainer, controller",
  "n10525436": "retailer, retail merchant",
  "n10525617": "retiree, retired person",
  "n10525785": "retreatant",
  "n10525878": "returning officer",
  "n10526096": "reveler, reveller, merrymaker",
  "n10526300": "drunken reveler, drunken reveller, bacchanal, bacchant",
  "n10526441": "revenant",
  "n10526534": "revenant",
  "n10526633": "revenuer",
  "n10526795": "reversioner",
  "n10526927": "reviewer, referee, reader",
  "n10527147": "revisionist",
  "n10527334": "revolutionist, revolutionary, subversive, subverter",
  "n10527915": "rheumatic",
  "n10528023": "rheumatologist",
  "n10528148": "Rhodesian man, Homo rhodesiensis",
  "n10528295": "Rhodes scholar",
  "n10528493": "rhymer, rhymester, versifier, poetizer, poetiser",
  "n10528816": "rhythm and blues musician",
  "n10528969": "ribald",
  "n10529103": "Richard Roe",
  "n10529231": "rich person, wealthy person, have",
  "n10529555": "millionaire",
  "n10529684": "billionaire",
  "n10529823": "multi-billionaire",
  "n10529965": "rider",
  "n10530150": "rider",
  "n10530288": "disreputable person",
  "n10530383": "riding master",
  "n10530475": "Riff, Riffian",
  "n10530571": "rifleman",
  "n10530677": "rifleman",
  "n10530769": "rigger",
  "n10530862": "rigger, oil rigger",
  "n10530959": "right-hander, right hander, righthander",
  "n10531109": "right-hand man, chief assistant, man Friday",
  "n10531227": "rightist, right-winger",
  "n10531445": "ringer",
  "n10531557": "ringer, dead ringer, clone",
  "n10531694": "ring girl",
  "n10531838": "ringleader",
  "n10531948": "ringmaster",
  "n10532058": "rioter",
  "n10532252": "ripper",
  "n10532393": "Rip van Winkle",
  "n10532479": "Rip van Winkle",
  "n10532576": "riser",
  "n10532751": "ritualist",
  "n10532886": "ritualist",
  "n10533013": "rival, challenger, competitor, competition, contender",
  "n10533541": "riveter, rivetter",
  "n10533681": "road builder",
  "n10533780": "road hog, roadhog",
  "n10533874": "roadman, road mender",
  "n10533983": "roarer, bawler, bellower, screamer, screecher, shouter, yeller",
  "n10534277": "roaster",
  "n10534389": "roaster",
  "n10534586": "robber",
  "n10534748": "robbery suspect",
  "n10534849": "Robert's Rules of Order",
  "n10535047": "Robin Hood",
  "n10535196": "Robinson Crusoe",
  "n10535366": "rock",
  "n10535604": "rock climber, cragsman",
  "n10535706": "rocker",
  "n10535881": "rocker, rock 'n' roll musician",
  "n10536021": "rocker",
  "n10536134": "rocket engineer, rocket scientist",
  "n10536274": "rocket scientist",
  "n10536416": "rock star",
  "n10536728": "rogue, knave, rascal, rapscallion, scalawag, scallywag, varlet",
  "n10536897": "roisterer",
  "n10537018": "rollerblader",
  "n10537107": "roller-skater",
  "n10537240": "Roman Emperor, Emperor of Rome",
  "n10537708": "Romanov, Romanoff",
  "n10537822": "romantic",
  "n10537906": "romanticist, romantic",
  "n10538082": "Romeo",
  "n10538154": "romper",
  "n10538272": "roofer",
  "n10538398": "room clerk",
  "n10538518": "roommate, roomie, roomy",
  "n10538629": "ropemaker, rope-maker, roper",
  "n10538733": "roper",
  "n10538853": "roper",
  "n10539015": "ropewalker, ropedancer",
  "n10539160": "rosebud",
  "n10539278": "Rosicrucian",
  "n10539462": "Rosicrucian",
  "n10539616": "Rotarian",
  "n10539715": "rotter, dirty dog, rat, skunk, stinker, stinkpot, bum, puke, crumb, lowlife, scum bag, so-and-so, git",
  "n10540114": "Mountie",
  "n10540252": "Rough Rider",
  "n10540413": "roughrider",
  "n10540526": "Roundhead",
  "n10540656": "roundhead",
  "n10540735": "roundsman",
  "n10540872": "router",
  "n10540984": "rover, scouter",
  "n10541106": "rubberneck, rubbernecker",
  "n10541229": "ruler, swayer",
  "n10541833": "civil authority, civil officer",
  "n10541983": "dynast",
  "n10542054": "rug merchant",
  "n10542138": "Rumpelstiltskin",
  "n10542401": "Shylock",
  "n10542499": "rumrunner",
  "n10542608": "runner",
  "n10542761": "runner",
  "n10542888": "runner",
  "n10543057": "runner-up, second best",
  "n10543161": "running back",
  "n10543420": "running mate",
  "n10543544": "runt, shrimp, peewee, half-pint",
  "n10543713": "ruralist",
  "n10543795": "rusher",
  "n10543937": "rusher",
  "n10544067": "rusher",
  "n10544232": "rustic",
  "n10544480": "rustler, cattle thief",
  "n10544615": "Sabbatarian",
  "n10544748": "saboteur, wrecker, diversionist",
  "n10544940": "sachem, sagamore",
  "n10545087": "sachem",
  "n10545187": "sacred cow",
  "n10545296": "sacrificer",
  "n10545415": "saddler",
  "n10545524": "Sadducee",
  "n10545682": "sadhu, saddhu",
  "n10545792": "sadist",
  "n10545940": "sadomasochist",
  "n10546062": "safebreaker, safecracker, cracksman",
  "n10546202": "sage",
  "n10546428": "sailing master, navigator",
  "n10546561": "sailmaker",
  "n10546633": "sailor, crewman",
  "n10546850": "saint, holy man, holy person, angel",
  "n10547145": "saint",
  "n10548227": "salesclerk, shop clerk, clerk, shop assistant",
  "n10548419": "salesgirl, saleswoman, saleslady",
  "n10548537": "salesman",
  "n10548681": "salesperson, sales representative, sales rep",
  "n10548976": "saloon keeper",
  "n10549062": "salter, salt merchant",
  "n10549165": "salter",
  "n10549315": "salutatorian, salutatory speaker",
  "n10549510": "salvager, salvor",
  "n10549648": "Samaritan",
  "n10549763": "samurai",
  "n10549925": "sandbagger",
  "n10550090": "sandboy",
  "n10550252": "sandwichman",
  "n10550369": "sangoma",
  "n10550468": "sannup",
  "n10550551": "sannyasi, sannyasin, sanyasi",
  "n10550673": "Santa Claus, Santa, Kriss Kringle, Father Christmas, Saint Nicholas, Saint Nick, St. Nick",
  "n10550951": "Tristan, Tristram",
  "n10551265": "Iseult, Isolde",
  "n10551576": "sapper",
  "n10551751": "sapper",
  "n10551898": "Saracen",
  "n10552019": "Saracen",
  "n10552203": "Saracen",
  "n10552285": "Sardinian",
  "n10552393": "Sassenach",
  "n10552486": "Satanist, diabolist",
  "n10552639": "satellite, planet",
  "n10552742": "satirist, ironist, ridiculer",
  "n10552980": "satyr, lecher, lech, letch",
  "n10553140": "satrap",
  "n10553235": "saunterer, stroller, ambler",
  "n10553402": "savage, barbarian",
  "n10553627": "saver",
  "n10553805": "savior, saviour, rescuer, deliverer",
  "n10554024": "Savoyard",
  "n10554141": "sawyer",
  "n10554243": "saxophonist, saxist",
  "n10554455": "scab, strikebreaker, blackleg, rat",
  "n10554666": "scalawag, scallywag",
  "n10554846": "scalper",
  "n10555059": "scandalmonger",
  "n10555156": "scanner",
  "n10555311": "scapegoat, whipping boy",
  "n10555430": "scapegrace, black sheep",
  "n10555538": "Scaramouch, Scaramouche",
  "n10555679": "scaremonger, stirrer",
  "n10555825": "scatterbrain, forgetful person",
  "n10555936": "scenarist",
  "n10556033": "scene painter",
  "n10556124": "sceneshifter, shifter",
  "n10556235": "scene-stealer",
  "n10556401": "scenic artist, scene painter",
  "n10556518": "schemer, plotter",
  "n10556704": "schizophrenic",
  "n10556825": "schlemiel, shlemiel",
  "n10556953": "schlepper, shlepper, schlep, shlep",
  "n10557094": "schlimazel, shlimazel",
  "n10557246": "schlockmeister, shlockmeister",
  "n10557404": "schmuck, shmuck, schmo, shmo",
  "n10557517": "schnook, shnook",
  "n10557699": "schnorrer, shnorrer",
  "n10557854": "scholar, scholarly person, bookman, student",
  "n10558773": "scholar",
  "n10558896": "Scholastic",
  "n10559009": "scholiast",
  "n10559192": "schoolboy",
  "n10559288": "schoolchild, school-age child, pupil",
  "n10559508": "schoolfriend",
  "n10559603": "schoolgirl",
  "n10559683": "Schoolman, medieval Schoolman",
  "n10559840": "schoolmarm, schoolma'am, schoolmistress, mistress",
  "n10559996": "schoolmaster",
  "n10560106": "schoolmate, classmate, schoolfellow, class fellow",
  "n10560247": "school superintendent",
  "n10560352": "schoolteacher, school teacher",
  "n10560548": "science teacher",
  "n10560637": "scientist",
  "n10561222": "scion",
  "n10561320": "scoffer, flouter, mocker, jeerer",
  "n10561613": "scoffer, gorger",
  "n10561736": "scofflaw",
  "n10561861": "scold, scolder, nag, nagger, common scold",
  "n10562135": "scorekeeper, scorer",
  "n10562283": "scorer",
  "n10562391": "scorer",
  "n10562509": "scourer",
  "n10562645": "scourer",
  "n10562749": "scout, pathfinder, guide",
  "n10562968": "scout, talent scout",
  "n10563183": "Scout",
  "n10563314": "scoutmaster",
  "n10563403": "scrambler",
  "n10563610": "scratch",
  "n10563711": "scratcher",
  "n10563826": "scratcher",
  "n10563940": "scrawler, scribbler",
  "n10564098": "screen actor, movie actor",
  "n10564224": "screener",
  "n10564400": "screenwriter, film writer",
  "n10564541": "screwballer",
  "n10564660": "scribe, scribbler, penman",
  "n10564800": "scrimshanker",
  "n10564905": "scriptwriter",
  "n10565048": "scrubber",
  "n10565197": "scrub nurse",
  "n10565302": "scrutinizer, scrutiniser",
  "n10565502": "scrutineer, canvasser",
  "n10565667": "scuba diver",
  "n10565764": "sculler",
  "n10565951": "scullion",
  "n10566072": "sculptor, sculpturer, carver, statue maker",
  "n10566893": "sculptress",
  "n10566966": "Scythian",
  "n10567096": "sea king",
  "n10567172": "sea lawyer",
  "n10567268": "sealer",
  "n10567401": "searcher",
  "n10567613": "Sea Scout",
  "n10567722": "seasonal worker, seasonal",
  "n10567848": "seasoner",
  "n10567979": "secessionist",
  "n10568083": "second",
  "n10568200": "second baseman, second sacker",
  "n10568358": "second cousin",
  "n10568443": "seconder",
  "n10568608": "second fiddle, second banana",
  "n10568754": "second hand",
  "n10568915": "second-in-command",
  "n10569011": "second lieutenant, 2nd lieutenant",
  "n10569179": "second-rater, mediocrity",
  "n10569411": "secret agent, intelligence officer, intelligence agent, operative",
  "n10569744": "secretary, secretarial assistant",
  "n10570019": "secretary",
  "n10570429": "Attorney General, United States Attorney General, US Attorney General",
  "n10570704": "Secretary of Agriculture, Agriculture Secretary",
  "n10570961": "Secretary of Commerce, Commerce Secretary",
  "n10571202": "Secretary of Defense, Defense Secretary",
  "n10571435": "Secretary of Education, Education Secretary",
  "n10571670": "Secretary of Energy, Energy Secretary",
  "n10571907": "Secretary of Health and Human Services",
  "n10572185": "Secretary of Housing and Urban Development",
  "n10572469": "Secretary of Labor, Labor Secretary",
  "n10572706": "Secretary of State",
  "n10572889": "Secretary of the Interior, Interior Secretary",
  "n10573129": "Secretary of the Treasury, Treasury Secretary",
  "n10573349": "Secretary of Transportation, Transportation Secretary",
  "n10573596": "Secretary of Veterans Affairs",
  "n10573829": "Secretary General",
  "n10573957": "sectarian, sectary, sectarist",
  "n10574154": "Section Eight",
  "n10574311": "section hand",
  "n10574425": "section man",
  "n10574538": "secularist",
  "n10574723": "secundigravida, gravida II",
  "n10574840": "security consultant",
  "n10574958": "security director",
  "n10575089": "seducer, ladies' man, lady killer",
  "n10575241": "seducer",
  "n10575387": "seductress",
  "n10575463": "seeded player, seed",
  "n10575594": "seeder, cloud seeder",
  "n10575705": "seedsman, seedman",
  "n10575787": "seeker, searcher, quester",
  "n10576071": "seer",
  "n10576223": "segregate",
  "n10576316": "segregator, segregationist",
  "n10576513": "seismologist",
  "n10576676": "selectman",
  "n10576818": "selectwoman",
  "n10576962": "selfish person",
  "n10577182": "self-starter",
  "n10577284": "seller, marketer, vender, vendor, trafficker",
  "n10577710": "selling agent",
  "n10577820": "semanticist, semiotician",
  "n10578021": "semifinalist",
  "n10578162": "seminarian, seminarist",
  "n10578349": "semiprofessional, semipro",
  "n10578471": "senator",
  "n10578656": "sendee",
  "n10578762": "sender, transmitter",
  "n10578952": "Senhor",
  "n10579062": "senior",
  "n10579176": "senior chief petty officer, SCPO",
  "n10579369": "senior master sergeant, SMSgt",
  "n10579549": "senior vice president",
  "n10579676": "sentimentalist, romanticist",
  "n10579835": "sensationalist, ballyhoo artist",
  "n10580030": "separatist, separationist",
  "n10580304": "Sephardi, Sephardic Jew",
  "n10580437": "septuagenarian",
  "n10580535": "serf, helot, villein",
  "n10580772": "sergeant",
  "n10581094": "sergeant at arms, serjeant-at-arms",
  "n10581278": "sergeant major, command sergeant major",
  "n10581464": "serial killer, serial murderer",
  "n10581648": "spree killer",
  "n10581772": "sericulturist",
  "n10581890": "serjeant-at-law, serjeant, sergeant-at-law, sergeant",
  "n10582032": "serologist",
  "n10582154": "servant, retainer",
  "n10582507": "servant girl, serving girl",
  "n10582604": "server",
  "n10582746": "serviceman, military man, man, military personnel",
  "n10583250": "servitor",
  "n10583387": "settler, colonist",
  "n10583790": "settler",
  "n10583916": "settler",
  "n10584021": "settlor, trustor",
  "n10584318": "sewer",
  "n10584501": "sewing-machine operator",
  "n10584617": "sexagenarian",
  "n10584729": "sex kitten, sexpot, sex bomb",
  "n10584853": "sex object",
  "n10584973": "sex offender",
  "n10585077": "sex symbol",
  "n10585217": "sexton, sacristan",
  "n10585359": "shadow",
  "n10585496": "Shah, Shah of Iran",
  "n10585628": "shaheed",
  "n10585773": "Shaker",
  "n10585976": "shaker, mover and shaker",
  "n10586166": "Shakespearian, Shakespearean",
  "n10586265": "shanghaier, seizer",
  "n10586444": "sharecropper, cropper, sharecrop farmer",
  "n10586557": "shark",
  "n10586674": "shark",
  "n10586793": "sharpshooter",
  "n10586903": "shaver",
  "n10586998": "Shavian",
  "n10587089": "shearer",
  "n10587227": "shearer",
  "n10587378": "shedder, spiller",
  "n10587536": "she-devil",
  "n10587605": "sheepherder, shepherd, sheepman",
  "n10587806": "sheepman",
  "n10587894": "sheep",
  "n10588074": "sheep",
  "n10588182": "shegetz",
  "n10588357": "sheik, tribal sheik, sheikh, tribal sheikh, Arab chief",
  "n10588519": "sheika, sheikha",
  "n10588601": "sheller",
  "n10588724": "shelver",
  "n10588860": "Shem",
  "n10588965": "shepherd",
  "n10589066": "shepherdess",
  "n10589140": "sheriff",
  "n10589243": "shiksa, shikse",
  "n10589402": "shill",
  "n10589566": "shingler",
  "n10589666": "ship-breaker",
  "n10589785": "ship broker",
  "n10589930": "shipbuilder, ship builder",
  "n10590042": "ship chandler",
  "n10590146": "shipmate",
  "n10590239": "shipowner",
  "n10590339": "shipper",
  "n10590452": "shipping agent",
  "n10590537": "shipping clerk",
  "n10590638": "ship's chandler",
  "n10590764": "shipwright, shipbuilder, ship builder",
  "n10590903": "shirtmaker",
  "n10590977": "shocker",
  "n10591072": "shogun",
  "n10591239": "Shona",
  "n10591347": "shoofly",
  "n10591481": "shooter",
  "n10591678": "shooter, crap-shooter",
  "n10591811": "shopaholic",
  "n10591949": "shop boy",
  "n10592049": "shop girl",
  "n10592152": "shopkeeper, tradesman, storekeeper, market keeper",
  "n10592397": "shopper",
  "n10592595": "shopper",
  "n10592811": "shop steward, steward",
  "n10592986": "shortstop",
  "n10593115": "shot, shooter",
  "n10593392": "gunman, gun",
  "n10593521": "shot putter",
  "n10593618": "shoveler, shoveller",
  "n10593745": "showman, promoter, impresario",
  "n10594043": "showman",
  "n10594147": "shrew, termagant",
  "n10594284": "Shudra, Sudra",
  "n10594408": "shuffler",
  "n10594523": "shuffler",
  "n10594634": "shutterbug",
  "n10594715": "shy person, shrinking violet",
  "n10594857": "shyster, pettifogger",
  "n10595012": "Siamese twin, conjoined twin",
  "n10595164": "sibling, sib",
  "n10595361": "sibyl",
  "n10595551": "sibyl",
  "n10595647": "sick person, diseased person, sufferer",
  "n10596272": "side judge",
  "n10596348": "sidesman",
  "n10596517": "sightreader",
  "n10596689": "sightseer, excursionist, tripper, rubberneck",
  "n10596899": "signaler, signaller",
  "n10597091": "signalman",
  "n10597234": "signer, signatory",
  "n10597505": "signer",
  "n10597642": "sign painter",
  "n10597745": "signor, signior",
  "n10597889": "signora",
  "n10598013": "signore",
  "n10598181": "signorina",
  "n10598349": "Sikh",
  "n10598459": "silent partner, sleeping partner",
  "n10598641": "silly",
  "n10598749": "silversmith, silverworker, silver-worker",
  "n10598904": "addle-head, addlehead, loon, birdbrain",
  "n10599068": "Simeon",
  "n10599215": "simperer",
  "n10599354": "simpleton, simple",
  "n10599806": "singer, vocalist, vocalizer, vocaliser",
  "n10601078": "sinner, evildoer",
  "n10601234": "Sinologist",
  "n10601362": "sipper",
  "n10601451": "sir",
  "n10601526": "Sir",
  "n10601644": "sirdar",
  "n10601725": "sire",
  "n10601840": "Siren",
  "n10602119": "sirrah",
  "n10602258": "sister",
  "n10602470": "Sister",
  "n10602695": "Beguine",
  "n10602985": "sister, sis",
  "n10603242": "half sister, half-sister, stepsister",
  "n10603378": "sissy, pantywaist, pansy, milksop, Milquetoast",
  "n10603528": "waverer, vacillator, hesitator, hesitater",
  "n10603766": "sister-in-law",
  "n10603851": "sitar player",
  "n10603959": "sitter",
  "n10604089": "sitting duck, easy mark",
  "n10604180": "six-footer",
  "n10604275": "sixth-former",
  "n10604380": "skateboarder",
  "n10604491": "skater",
  "n10604634": "skeptic, sceptic, doubter",
  "n10604880": "sketcher",
  "n10604979": "skidder",
  "n10605088": "skidder, slider, slipper",
  "n10605253": "skier",
  "n10605375": "ski jumper",
  "n10605510": "skimmer",
  "n10605608": "Skinnerian",
  "n10605737": "skinny-dipper",
  "n10605848": "skirmisher",
  "n10605985": "skilled worker, trained worker, skilled workman",
  "n10607291": "skin-diver, aquanaut",
  "n10607478": "skinhead",
  "n10607706": "skinner",
  "n10607824": "skipper",
  "n10607933": "skivvy, slavey",
  "n10608073": "skycap",
  "n10608188": "skydiver",
  "n10608385": "slacker, shirker",
  "n10608658": "slammer",
  "n10608803": "slapper, spanker",
  "n10609092": "slasher",
  "n10609198": "slattern, slut, slovenly woman, trollop",
  "n10609325": "slave",
  "n10609556": "slave, striver, hard worker",
  "n10609686": "slave",
  "n10609871": "slave driver",
  "n10609960": "slave driver, Simon Legree",
  "n10610096": "slaveholder, slave owner, slaver",
  "n10610201": "slaver, slave dealer, slave trader",
  "n10610333": "sledder",
  "n10610465": "sleeper, slumberer",
  "n10610699": "sleeper",
  "n10610850": "sleeper",
  "n10611117": "Sleeping Beauty",
  "n10611267": "sleeping beauty",
  "n10611361": "sleepwalker, somnambulist, noctambulist",
  "n10611541": "sleepyhead",
  "n10611613": "sleuth, sleuthhound",
  "n10611729": "slicer",
  "n10611869": "slicker",
  "n10611988": "slinger",
  "n10612104": "slip",
  "n10612210": "slob, sloven, pig, slovenly person",
  "n10612373": "sloganeer",
  "n10612518": "slopseller, slop-seller",
  "n10612645": "slouch",
  "n10612803": "sloucher",
  "n10612931": "sluggard, slug",
  "n10613052": "small businessman",
  "n10613198": "small-for-gestational-age infant, SGA infant",
  "n10613387": "smallholder",
  "n10613505": "small person",
  "n10613738": "small farmer",
  "n10613839": "smarta",
  "n10613996": "smasher, stunner, knockout, beauty, ravisher, sweetheart, peach, lulu, looker, mantrap, dish",
  "n10614225": "smasher",
  "n10614363": "smiler",
  "n10614507": "smirker",
  "n10614629": "smith, metalworker",
  "n10614812": "smith",
  "n10614976": "smoker, tobacco user",
  "n10615179": "smoothie, smoothy, sweet talker, charmer",
  "n10615334": "smuggler, runner, contrabandist, moon curser, moon-curser",
  "n10615584": "snake charmer",
  "n10615702": "snake, snake in the grass",
  "n10615808": "snarer",
  "n10615929": "snatcher",
  "n10616048": "sneak",
  "n10616204": "sneak thief, pilferer, snitcher",
  "n10616379": "sneerer, scorner",
  "n10616578": "sneezer",
  "n10616670": "sniffer",
  "n10616779": "sniffler, sniveler",
  "n10616899": "sniper",
  "n10617024": "snob, prig, snot, snoot",
  "n10617193": "snoop, snooper",
  "n10617397": "snorer",
  "n10617501": "snorter",
  "n10617665": "snowboarder",
  "n10617814": "snuffer, snuff user",
  "n10617904": "snuffer",
  "n10618007": "snuffler",
  "n10618146": "sobersides",
  "n10618234": "sob sister",
  "n10618342": "soccer player",
  "n10618465": "social anthropologist, cultural anthropologist",
  "n10618685": "social climber, climber",
  "n10618848": "socialist",
  "n10619176": "collectivist, leftist, left-winger",
  "n10619409": "socialite",
  "n10619492": "socializer, socialiser",
  "n10619642": "social scientist",
  "n10619888": "social secretary",
  "n10620027": "social worker, caseworker, welfare worker",
  "n10620212": "Socinian",
  "n10620437": "sociobiologist",
  "n10620586": "sociolinguist",
  "n10620758": "sociologist",
  "n10621140": "sod",
  "n10621294": "soda jerk, soda jerker",
  "n10621400": "sodalist",
  "n10621514": "sodomite, sodomist, sod, bugger",
  "n10621738": "softy, softie",
  "n10621847": "sojourner",
  "n10621941": "solderer",
  "n10622053": "soldier",
  "n10623175": "solicitor",
  "n10623354": "solicitor, canvasser",
  "n10623533": "solicitor general",
  "n10623650": "soloist",
  "n10623806": "sommelier, wine waiter, wine steward",
  "n10623949": "somniloquist",
  "n10624074": "son, boy",
  "n10624310": "songster",
  "n10624437": "songstress",
  "n10624540": "songwriter, songster, ballad maker",
  "n10624915": "son-in-law",
  "n10625000": "sonneteer",
  "n10625099": "Sophist",
  "n10625285": "sophisticate, man of the world",
  "n10625438": "sophomore, soph",
  "n10625546": "soprano",
  "n10625860": "sorcerer, magician, wizard, necromancer, thaumaturge, thaumaturgist",
  "n10626194": "shaman, priest-doctor",
  "n10626439": "medicine man",
  "n10626540": "sorceress",
  "n10626630": "sorehead",
  "n10626722": "sort",
  "n10626867": "sorter",
  "n10626994": "soubrette",
  "n10627082": "soul, psyche",
  "n10627252": "soul mate",
  "n10627349": "sounding board",
  "n10627557": "soundman",
  "n10627714": "sourdough",
  "n10627899": "sourpuss, picklepuss, gloomy Gus, pouter",
  "n10628097": "Southern Baptist",
  "n10628222": "Southerner",
  "n10628368": "Rebel, Reb, Johnny Reb, Johnny, greyback",
  "n10628644": "sovereign, crowned head, monarch",
  "n10629020": "sower",
  "n10629142": "space cadet",
  "n10629329": "spacewalker",
  "n10629449": "space writer",
  "n10629545": "spammer",
  "n10629647": "Spanish American, Hispanic American, Hispanic",
  "n10629820": "sparer",
  "n10629939": "sparring partner, sparring mate",
  "n10630093": "spastic",
  "n10630188": "speaker, talker, utterer, verbalizer, verbaliser",
  "n10631131": "native speaker",
  "n10631309": "Speaker",
  "n10631512": "spearhead",
  "n10631654": "speechwriter",
  "n10631765": "special agent",
  "n10631941": "specialist, specializer, specialiser",
  "n10632576": "specialist, medical specialist",
  "n10633298": "specifier",
  "n10633450": "spectator, witness, viewer, watcher, looker",
  "n10634075": "speculator, plunger",
  "n10634316": "speculator",
  "n10634464": "speech therapist",
  "n10634596": "speeder, speed demon",
  "n10634723": "speed freak",
  "n10634849": "speedskater, speed skater",
  "n10634990": "spellbinder",
  "n10635149": "speller, good speller, poor speller",
  "n10635275": "spender, disburser, expender",
  "n10635460": "spendthrift, spend-all, spender, scattergood",
  "n10635625": "big spender, high roller",
  "n10635788": "sphinx",
  "n10635907": "spindlelegs, spindleshanks",
  "n10636014": "spin doctor, spinmeister",
  "n10636340": "spinner, spinster, thread maker",
  "n10636488": "spinster, old maid",
  "n10636598": "spirit",
  "n10636746": "spitfire",
  "n10636874": "spitter, expectorator",
  "n10637038": "spiv",
  "n10637229": "splicer",
  "n10637360": "splicer",
  "n10637483": "split end",
  "n10637635": "splitter, divider",
  "n10637839": "splitter",
  "n10637961": "spoiler",
  "n10638136": "spoilsport, killjoy, wet blanket, party pooper",
  "n10638310": "spokesman",
  "n10638385": "spokesperson, interpreter, representative, voice",
  "n10638734": "spokeswoman",
  "n10638813": "sponger",
  "n10638922": "sport, sportsman, sportswoman",
  "n10639047": "sport",
  "n10639238": "sport, summercater",
  "n10639359": "sporting man, outdoor man",
  "n10639469": "sporting man",
  "n10639637": "sports announcer, sportscaster, sports commentator",
  "n10639817": "sports editor",
  "n10639925": "sports fan, fan, rooter",
  "n10640084": "sports writer, sportswriter",
  "n10640195": "spotter",
  "n10640327": "spotter",
  "n10640446": "spot-welder, spot welder",
  "n10640620": "spouse, partner, married person, mate, better half",
  "n10640968": "sprawler",
  "n10641112": "sprayer",
  "n10641223": "sprog",
  "n10641301": "sprog",
  "n10641413": "sprinter",
  "n10641551": "spurner",
  "n10641755": "spy, undercover agent",
  "n10642151": "spy",
  "n10642388": "spymaster",
  "n10642499": "squabbler",
  "n10642596": "square dancer",
  "n10642705": "square shooter, straight shooter, straight arrow",
  "n10642845": "square, square toes",
  "n10642988": "square, lame",
  "n10643095": "squatter",
  "n10643218": "squatter, homesteader, nester",
  "n10643400": "squaw",
  "n10643476": "squaw man",
  "n10643584": "squeeze",
  "n10643727": "squinter, squint-eye",
  "n10643837": "squire",
  "n10643937": "squire",
  "n10644062": "squire, gallant",
  "n10644179": "stabber",
  "n10644301": "stableman, stableboy, groom, hostler, ostler",
  "n10644469": "stacker",
  "n10644598": "staff member, staffer",
  "n10644839": "staff officer",
  "n10645017": "staff sergeant",
  "n10645223": "stage director",
  "n10645392": "stagehand, stage technician",
  "n10645611": "stage manager, stager",
  "n10645854": "staggerer, totterer, reeler",
  "n10646032": "stainer",
  "n10646140": "stakeholder",
  "n10646325": "Stalinist",
  "n10646433": "stalker",
  "n10646528": "stalker",
  "n10646641": "stalking-horse",
  "n10646780": "stammerer, stutterer",
  "n10646942": "stamper, stomper, tramper, trampler",
  "n10647168": "stamper",
  "n10647359": "stamp dealer",
  "n10647475": "standard-bearer",
  "n10647582": "standardizer, standardiser",
  "n10647745": "standee",
  "n10648033": "stander",
  "n10648237": "stand-in, substitute, relief, reliever, backup, backup man, fill-in",
  "n10648696": "star, principal, lead",
  "n10648909": "starer",
  "n10649047": "starets",
  "n10649197": "starlet",
  "n10649308": "starter, dispatcher",
  "n10649438": "starter",
  "n10649574": "starting pitcher",
  "n10649724": "starveling",
  "n10649823": "stater",
  "n10649962": "state's attorney, state attorney",
  "n10650076": "state senator",
  "n10650162": "statesman, solon, national leader",
  "n10652511": "stateswoman",
  "n10652605": "state treasurer",
  "n10652703": "stationer, stationery seller",
  "n10652837": "stationmaster, station agent",
  "n10652954": "statistician, actuary",
  "n10653238": "statistician, mathematical statistician",
  "n10653388": "stay-at-home, homebody",
  "n10653529": "steamfitter",
  "n10653679": "steelmaker, steelworker, steelman",
  "n10653793": "steeplejack",
  "n10653903": "stemmer",
  "n10654015": "stenographer, amanuensis, shorthand typist",
  "n10654211": "stentor",
  "n10654321": "stepbrother, half-brother, half brother",
  "n10654461": "stepchild",
  "n10654596": "stepdaughter",
  "n10654701": "stepfather",
  "n10654827": "stepmother",
  "n10654932": "stepparent",
  "n10655075": "stepson",
  "n10655169": "stevedore, loader, longshoreman, docker, dockhand, dock worker, dockworker, dock-walloper, lumper",
  "n10655442": "steward",
  "n10655594": "steward, flight attendant",
  "n10655730": "steward",
  "n10655875": "stewardess, air hostess, hostess",
  "n10655986": "stickler",
  "n10656120": "stiff",
  "n10656223": "stifler, smotherer",
  "n10656488": "stigmatic, stigmatist",
  "n10656706": "stillborn infant",
  "n10656832": "stinter",
  "n10656969": "stipendiary, stipendiary magistrate",
  "n10657165": "stippler",
  "n10657306": "stitcher",
  "n10657444": "stockbroker",
  "n10657556": "stockjobber",
  "n10657681": "stocktaker, stock-taker",
  "n10657835": "stock trader",
  "n10657969": "stockholder, shareholder, shareowner",
  "n10658122": "stockholder of record",
  "n10658304": "stockist",
  "n10658501": "stockman, stock raiser, stock farmer",
  "n10658676": "Stoic",
  "n10658867": "stoic, unemotional person",
  "n10659042": "stoker, fireman",
  "n10659188": "stone breaker",
  "n10659294": "stonecutter, cutter",
  "n10659393": "stoner, lapidator",
  "n10659571": "stonewaller",
  "n10659762": "stooper",
  "n10659961": "stooper",
  "n10660128": "store detective",
  "n10660250": "storm trooper",
  "n10660333": "storyteller, fibber, fabricator",
  "n10660471": "stowaway",
  "n10660621": "strafer",
  "n10660729": "straggler, strayer",
  "n10660883": "straight man, second banana",
  "n10661002": "stranger, alien, unknown",
  "n10661216": "stranger",
  "n10661334": "straphanger",
  "n10661464": "straphanger",
  "n10661563": "strategist, strategian",
  "n10661732": "straw boss, assistant foreman",
  "n10661865": "strider",
  "n10662046": "stringer",
  "n10662162": "stringer",
  "n10662306": "streaker",
  "n10662474": "street cleaner, street sweeper",
  "n10662649": "street fighter, tough",
  "n10662819": "street fighter",
  "n10662952": "street urchin, guttersnipe",
  "n10663137": "street arab, gamin, throwaway",
  "n10663315": "streetwalker, street girl, hooker, hustler, floozy, floozie, slattern",
  "n10663549": "stretcher-bearer, litter-bearer",
  "n10663659": "strike leader",
  "n10663745": "striker",
  "n10663858": "striker",
  "n10663996": "striker",
  "n10664076": "striper",
  "n10664235": "strip miner",
  "n10664340": "stripper, striptease artist, striptease, stripteaser, exotic dancer, ecdysiast, peeler",
  "n10664656": "stripper, stemmer, sprigger",
  "n10664850": "stroke",
  "n10665006": "strongman",
  "n10665190": "strongman",
  "n10665302": "struggler",
  "n10665455": "Stuart",
  "n10665587": "stud, he-man, macho-man",
  "n10665698": "student, pupil, educatee",
  "n10666259": "student teacher, practice teacher",
  "n10666464": "study",
  "n10666615": "stuffed shirt",
  "n10666752": "stumblebum, palooka",
  "n10666846": "double, stunt man, stunt woman",
  "n10667041": "stumbler, tripper",
  "n10667187": "stupid, stupid person, stupe, dullard, dolt, pudding head, pudden-head, poor fish, pillock",
  "n10667477": "stylist",
  "n10667596": "stylite",
  "n10667709": "subaltern",
  "n10667863": "subcontractor",
  "n10668024": "subdeacon",
  "n10668280": "subdivider",
  "n10668450": "subduer, surmounter, overcomer",
  "n10668666": "subject, case, guinea pig",
  "n10669009": "subjectivist",
  "n10669125": "subjugator",
  "n10669236": "sublieutenant",
  "n10669357": "submariner",
  "n10669486": "submitter",
  "n10669727": "submitter",
  "n10669876": "subnormal",
  "n10669991": "subordinate, subsidiary, underling, foot soldier",
  "n10670310": "subscriber, contributor",
  "n10670483": "subscriber, reader",
  "n10670668": "subscriber, endorser, indorser, ratifier",
  "n10670885": "subsidizer, subsidiser",
  "n10671042": "substitute, reserve, second-stringer",
  "n10671300": "subtracter",
  "n10671387": "suburbanite",
  "n10671467": "subvocalizer, subvocaliser",
  "n10671613": "successor, heir",
  "n10671736": "successor, replacement",
  "n10671898": "succorer, succourer",
  "n10672066": "sucker",
  "n10672192": "suer, petitioner",
  "n10672371": "Sufi",
  "n10672540": "suffragan, suffragan bishop",
  "n10672662": "suffragette",
  "n10672908": "suffragist",
  "n10673296": "sugar daddy",
  "n10673451": "suggester, proposer",
  "n10673669": "suicide, felo-de-se",
  "n10673776": "suicide bomber",
  "n10673946": "suit",
  "n10674130": "suitor, suer, wooer",
  "n10674315": "sultan, grand Turk",
  "n10674471": "Sumerian",
  "n10674592": "summercaters",
  "n10674713": "sumo wrestler",
  "n10674896": "sun",
  "n10675010": "sunbather",
  "n10675142": "sundowner",
  "n10675258": "sun worshiper",
  "n10675346": "supercargo",
  "n10675481": "supergrass, grass",
  "n10675609": "super heavyweight",
  "n10675721": "superintendent, super",
  "n10675876": "superior",
  "n10676018": "superior, higher-up, superordinate",
  "n10676214": "supermarketer, supermarketeer",
  "n10676319": "supermodel",
  "n10676434": "supermom",
  "n10676569": "supernumerary, spear carrier, extra",
  "n10676682": "supernumerary",
  "n10676877": "supervisor",
  "n10677271": "supplier, provider",
  "n10677604": "supply officer",
  "n10677713": "supporter, protagonist, champion, admirer, booster, friend",
  "n10678472": "suppressor, suppresser",
  "n10678662": "supremacist",
  "n10678841": "suprematist",
  "n10678937": "supremo",
  "n10679054": "surfer, surfboarder",
  "n10679174": "surgeon, operating surgeon, sawbones",
  "n10679503": "Surgeon General",
  "n10679610": "Surgeon General",
  "n10679723": "surpriser",
  "n10679845": "surrealist",
  "n10679998": "surrenderer, yielder",
  "n10680153": "surrogate, alternate, replacement",
  "n10680370": "surrogate mother",
  "n10680609": "surveyor",
  "n10680796": "surveyor",
  "n10680910": "survivalist",
  "n10681060": "survivor",
  "n10681194": "survivor, subsister",
  "n10681383": "suspect",
  "n10681557": "sutler, victualer, victualler, provisioner",
  "n10681748": "Svengali",
  "n10681891": "Svengali",
  "n10682038": "swaggerer",
  "n10682169": "swagman, swagger, swaggie",
  "n10682380": "swearer",
  "n10682501": "swearer",
  "n10682599": "sweater girl",
  "n10682713": "sweeper",
  "n10682850": "sweetheart",
  "n10682953": "sweetheart, sweetie, steady, truelove",
  "n10683126": "swimmer, natator, bather",
  "n10683349": "swimmer",
  "n10683593": "swineherd, pigman",
  "n10683675": "swinger, tramp",
  "n10683801": "swinger",
  "n10683927": "swing voter, floating voter",
  "n10684146": "switcher, whipper",
  "n10684311": "switch-hitter",
  "n10684446": "switch-hitter",
  "n10684538": "switchman",
  "n10684630": "swot, grind, nerd, wonk, dweeb",
  "n10684827": "sycophant, toady, crawler, lackey, ass-kisser",
  "n10685123": "syllogist, syllogizer, syllogiser",
  "n10685300": "sylph",
  "n10685398": "sylph",
  "n10685480": "sylvan, silvan",
  "n10685587": "symbolic logician",
  "n10685685": "symbolist",
  "n10685853": "symbolist, symbolizer, symboliser",
  "n10686073": "sympathizer, sympathiser, well-wisher",
  "n10686313": "sympathizer, sympathiser, comforter",
  "n10686517": "symphonist",
  "n10686598": "symposiast",
  "n10686694": "syncopator",
  "n10686885": "syndic",
  "n10687025": "syndicator",
  "n10687135": "synonymist",
  "n10687231": "synthesist, synthesizer, synthesiser",
  "n10687427": "syphilitic",
  "n10687516": "system administrator",
  "n10687728": "systems analyst",
  "n10687826": "tablemate, dining companion",
  "n10687922": "tacker",
  "n10688070": "tackle",
  "n10688238": "tackler",
  "n10688356": "tactician",
  "n10688491": "Tagalog",
  "n10688671": "tagalong, hanger-on",
  "n10688811": "tagger",
  "n10688975": "tagger",
  "n10689104": "tail, shadow, shadower",
  "n10689306": "tailback",
  "n10689429": "tailgater",
  "n10689564": "tailor, seamster, sartor",
  "n10689784": "taker",
  "n10689878": "talent",
  "n10690002": "talent agent",
  "n10690095": "talking head",
  "n10690268": "tallyman, tally clerk",
  "n10690421": "tallyman",
  "n10690538": "tamer",
  "n10690648": "tanker, tank driver",
  "n10690742": "tanner",
  "n10690849": "tantalizer, tantaliser",
  "n10691052": "taoiseach",
  "n10691148": "tap dancer, tapper",
  "n10691318": "tapper, wiretapper, phone tapper",
  "n10691485": "tapper",
  "n10691600": "tapper",
  "n10691764": "tapster, tapper",
  "n10691937": "Tartuffe, Tartufe",
  "n10692090": "Tarzan",
  "n10692269": "taskmaster",
  "n10692405": "taskmistress",
  "n10692482": "taster, taste tester, taste-tester, sampler",
  "n10692696": "tattletale, tattler, taleteller, talebearer, telltale, blabbermouth",
  "n10692883": "tax assessor, assessor",
  "n10693052": "tax collector, taxman, exciseman, collector of internal revenue, internal revenue agent",
  "n10693235": "taxer",
  "n10693334": "taxi dancer",
  "n10693459": "taxidermist, animal stuffer, stuffer",
  "n10693646": "taxidriver, taximan, cabdriver, cabman, cabby, hack driver, hack-driver, livery driver",
  "n10693824": "taxonomist, taxonomer, systematist",
  "n10694163": "taxpayer",
  "n10694258": "teacher, instructor",
  "n10694849": "teacher's pet",
  "n10694939": "teaching fellow",
  "n10695050": "teammate, mate",
  "n10695192": "teamster, trucker, truck driver",
  "n10695333": "teamster",
  "n10695450": "tearaway",
  "n10695555": "tease, teaser, annoyer, vexer",
  "n10695822": "teaser",
  "n10695917": "techie, tekki",
  "n10696101": "technical sergeant",
  "n10696251": "technician",
  "n10696508": "technician",
  "n10696672": "technocrat",
  "n10696755": "technocrat",
  "n10696888": "technophile",
  "n10697013": "technophobe",
  "n10697135": "Ted, Teddy boy",
  "n10697282": "teetotaler, teetotaller, teetotalist",
  "n10697420": "telecaster",
  "n10697519": "telegrapher, telegraphist, telegraph operator",
  "n10697782": "teleologist",
  "n10697879": "telepathist, thought-reader, mental telepathist, mind reader",
  "n10698064": "telephone operator, telephonist, switchboard operator",
  "n10698241": "televangelist",
  "n10698368": "television reporter, television newscaster, TV reporter, TV newsman",
  "n10698550": "television star, TV star",
  "n10698649": "Tell, William Tell",
  "n10698970": "teller, cashier, bank clerk",
  "n10699099": "teller, vote counter",
  "n10699262": "tellurian, earthling, earthman, worldling",
  "n10699415": "temp, temporary, temporary worker",
  "n10699558": "temporizer, temporiser",
  "n10699752": "tempter",
  "n10699981": "term infant",
  "n10700105": "toiler",
  "n10700201": "tenant, renter",
  "n10700517": "tenant",
  "n10700640": "tenant",
  "n10700840": "tenant farmer",
  "n10700963": "tenderfoot",
  "n10701096": "tennis coach",
  "n10701180": "tennis player",
  "n10701644": "tennis pro, professional tennis player",
  "n10701783": "tenor",
  "n10701962": "tenor saxophonist, tenorist",
  "n10702077": "tentmaker",
  "n10702167": "termer",
  "n10702307": "territorial",
  "n10702483": "terror, brat, little terror, holy terror",
  "n10702615": "terror, scourge, threat",
  "n10702781": "terrorist",
  "n10703221": "tertigravida, gravida III",
  "n10703336": "testator, testate",
  "n10703480": "testatrix",
  "n10703553": "test driver",
  "n10703692": "testee, examinee",
  "n10703905": "testifier, deponent, deposer",
  "n10704098": "test pilot",
  "n10704238": "test-tube baby",
  "n10704516": "Teutonist",
  "n10704712": "Texas Ranger, Ranger",
  "n10704886": "thane",
  "n10705100": "thane",
  "n10705211": "thatcher",
  "n10705345": "Thatcherite",
  "n10705448": "theatrical producer",
  "n10705615": "theologian, theologist, theologizer, theologiser",
  "n10706812": "theorist, theoretician, theorizer, theoriser, idealogue",
  "n10707134": "theosophist",
  "n10707233": "therapist, healer",
  "n10707616": "Thessalian",
  "n10707707": "Thessalonian",
  "n10707804": "thief, stealer",
  "n10708292": "thinker, creative thinker, mind",
  "n10708454": "thinker",
  "n10708797": "thin person, skin and bones, scrag",
  "n10708976": "third baseman, third sacker",
  "n10709131": "third party",
  "n10709256": "third-rater",
  "n10709358": "thoroughbred",
  "n10709435": "thrall",
  "n10709529": "thrower",
  "n10709745": "throwster, thrower",
  "n10709876": "thrush",
  "n10709979": "thunderbird",
  "n10710171": "thurifer",
  "n10710259": "ticket collector, ticket taker",
  "n10710403": "tier, tier up",
  "n10710509": "tier",
  "n10710632": "tiger",
  "n10710778": "tight end",
  "n10710913": "tiler",
  "n10711005": "tiller",
  "n10711144": "tilter",
  "n10711253": "timberman",
  "n10711370": "timekeeper",
  "n10711483": "timekeeper, timer",
  "n10711641": "timeserver",
  "n10711766": "Timorese",
  "n10711852": "tinker",
  "n10712055": "tinker, tinkerer",
  "n10712229": "tinkerer, fiddler",
  "n10712374": "tinsmith, tinner",
  "n10712474": "tinter",
  "n10712573": "tipper",
  "n10712690": "tippler, social drinker",
  "n10712835": "tipster, tout",
  "n10713012": "tither",
  "n10713124": "titterer, giggler",
  "n10713254": "T-man",
  "n10713367": "toast",
  "n10713502": "toaster, wassailer",
  "n10713686": "toastmaster, symposiarch",
  "n10713843": "toast mistress",
  "n10713923": "tobacconist",
  "n10714054": "Tobagonian",
  "n10714195": "tobogganist",
  "n10714317": "Todd, Sweeney Todd",
  "n10714465": "toddler, yearling, tot, bambino",
  "n10714577": "toff, nob",
  "n10714684": "tollkeeper, tollman, tollgatherer, toll collector, toll taker, toll agent, toller",
  "n10714851": "toller, bell ringer, ringer",
  "n10715030": "tomboy, romp, hoyden",
  "n10715136": "Tom Thumb",
  "n10715211": "Tom Thumb",
  "n10715347": "toolmaker",
  "n10715447": "top banana",
  "n10715544": "topper",
  "n10715671": "topper",
  "n10715789": "torchbearer",
  "n10715881": "torch singer",
  "n10716005": "tormentor, tormenter, persecutor",
  "n10716286": "tort-feasor, tortfeasor",
  "n10716389": "torturer",
  "n10716576": "Tory",
  "n10716698": "Tory",
  "n10716864": "Tory",
  "n10717055": "tosser",
  "n10717196": "tosser, jerk-off, wanker",
  "n10717337": "totalitarian",
  "n10717461": "totemist",
  "n10717589": "toucher",
  "n10717772": "touch-typist",
  "n10717921": "tough guy, plug-ugly",
  "n10718040": "tour guide",
  "n10718131": "tourist, tourer, holidaymaker",
  "n10718349": "tout, touter",
  "n10718509": "tout, ticket tout",
  "n10718665": "tovarich, tovarisch",
  "n10718794": "tower of strength, pillar of strength",
  "n10718952": "towhead",
  "n10719036": "town clerk",
  "n10719132": "town crier, crier",
  "n10719267": "townsman, towner",
  "n10719395": "townee",
  "n10719573": "townie, towny",
  "n10719692": "townsman",
  "n10719807": "toxicologist",
  "n10719972": "tracer",
  "n10720097": "tracker",
  "n10720197": "track star",
  "n10720267": "Tractarian",
  "n10720453": "trader, bargainer, dealer, monger",
  "n10720964": "trade unionist, unionist, union member",
  "n10721124": "traditionalist, diehard",
  "n10721321": "traffic cop",
  "n10721470": "dealer",
  "n10721612": "tragedian",
  "n10721708": "tragedian",
  "n10721819": "tragedienne",
  "n10721919": "trailblazer",
  "n10722029": "trail boss",
  "n10722137": "trainbandsman",
  "n10722239": "trainbearer",
  "n10722385": "trainee",
  "n10722575": "trainer",
  "n10722758": "trainman, railroader, railroad man, railwayman, railway man",
  "n10722965": "traitor, treasonist",
  "n10723230": "traitress",
  "n10723300": "tramp, hobo, bum",
  "n10723474": "trampler",
  "n10723597": "transactor",
  "n10723731": "transalpine",
  "n10723873": "transcendentalist",
  "n10723984": "transcriber",
  "n10724132": "transcriber",
  "n10724272": "transcriber",
  "n10724372": "transfer, transferee",
  "n10724570": "transferee",
  "n10724699": "transferer, transferrer",
  "n10724999": "transferor",
  "n10725149": "transient",
  "n10725280": "translator, transcriber",
  "n10725438": "transmigrante",
  "n10725613": "transplanter",
  "n10725734": "transsexual, transexual",
  "n10725893": "transsexual, transexual",
  "n10726031": "transvestite, cross-dresser",
  "n10726233": "trapper",
  "n10726388": "Trappist, Cistercian",
  "n10726524": "trapshooter",
  "n10726669": "travel agent",
  "n10726786": "traveling salesman, travelling salesman, commercial traveler, commercial traveller, roadman, bagman",
  "n10727016": "traverser",
  "n10727171": "trawler",
  "n10727256": "treasurer, financial officer",
  "n10727458": "Treasury, First Lord of the Treasury",
  "n10727623": "tree hugger",
  "n10727818": "tree surgeon, arborist",
  "n10727944": "trekker",
  "n10728117": "trencher",
  "n10728233": "trend-setter, taste-maker, fashion arbiter",
  "n10728361": "trial attorney, trial lawyer",
  "n10728523": "trial judge",
  "n10728624": "tribesman",
  "n10728727": "tribologist",
  "n10728828": "tribune",
  "n10728998": "trier, attempter, essayer",
  "n10729175": "trier",
  "n10729330": "trifler",
  "n10729442": "trigonometrician",
  "n10729567": "Trilby",
  "n10729711": "Trinidadian",
  "n10729801": "triplet",
  "n10729923": "tripper",
  "n10730077": "tritheist",
  "n10730262": "triumvir",
  "n10730416": "trombonist, trombone player",
  "n10730542": "trooper",
  "n10730615": "trouper",
  "n10730728": "trooper, state trooper",
  "n10730820": "trophy wife",
  "n10731013": "Trotskyite, Trotskyist, Trot",
  "n10731244": "troublemaker, trouble maker, troubler, mischief-maker, bad hat",
  "n10731560": "troubleshooter, trouble shooter",
  "n10731732": "truant, hooky player",
  "n10731848": "trudger, plodder, slogger",
  "n10732010": "trumpeter, cornetist",
  "n10732212": "trustbuster",
  "n10732314": "trustee, legal guardian",
  "n10732521": "trusty",
  "n10732644": "tub-thumper",
  "n10732748": "tucker",
  "n10732854": "Tudor",
  "n10732967": "tumbler",
  "n10733117": "tuner, piano tuner",
  "n10733225": "turncock",
  "n10733350": "turner",
  "n10733487": "turner",
  "n10733574": "turner",
  "n10733705": "turtler",
  "n10733820": "tutee",
  "n10733891": "tv announcer, television announcer",
  "n10733999": "twaddler",
  "n10734129": "twerp, twirp, twit",
  "n10734235": "twiddler, fiddler",
  "n10734394": "twin",
  "n10734568": "twiner",
  "n10734741": "two-timer",
  "n10734891": "Tyke",
  "n10734963": "tympanist, timpanist",
  "n10735173": "typist",
  "n10735298": "tyrant, autocrat, despot",
  "n10735564": "tyrant",
  "n10735707": "tyrant",
  "n10735852": "ugly duckling",
  "n10735984": "umpire, ump",
  "n10736091": "uncle",
  "n10736259": "uncle",
  "n10736394": "underachiever, underperformer, nonachiever",
  "n10736602": "undergraduate, undergrad",
  "n10736788": "undersecretary",
  "n10736926": "underseller",
  "n10737103": "understudy, standby",
  "n10737264": "undesirable",
  "n10737431": "undoer, opener, unfastener, untier",
  "n10737717": "undoer",
  "n10737860": "unemployed person",
  "n10737964": "unicorn",
  "n10738111": "unicyclist",
  "n10738215": "unilateralist",
  "n10738321": "uniocular dichromat",
  "n10738515": "union representative",
  "n10738670": "Unitarian",
  "n10738787": "Trinitarian",
  "n10738871": "Arminian",
  "n10738968": "universal agent, general agent",
  "n10739135": "universal donor",
  "n10739297": "UNIX guru",
  "n10739391": "Unknown Soldier",
  "n10739512": "unmarried woman",
  "n10739636": "unpleasant woman, disagreeable woman",
  "n10739881": "untouchable, Harijan",
  "n10740017": "upbraider, reprover, reproacher, rebuker",
  "n10740219": "upholder, maintainer, sustainer",
  "n10740482": "upholsterer",
  "n10740594": "upsetter",
  "n10740732": "upstager",
  "n10740868": "upstart, parvenu, nouveau-riche, arriviste",
  "n10741152": "upstart",
  "n10741258": "urban guerrilla",
  "n10741367": "urchin",
  "n10741493": "urologist",
  "n10741590": "user",
  "n10741821": "usher, guide",
  "n10742005": "usherette",
  "n10742111": "usher, doorkeeper",
  "n10742269": "usufructuary",
  "n10742384": "usurer, loan shark, moneylender, shylock",
  "n10742546": "usurper, supplanter",
  "n10742736": "utilitarian",
  "n10742881": "utility man",
  "n10742997": "utility man",
  "n10743124": "utilizer, utiliser",
  "n10743356": "Utopian",
  "n10743543": "utterer",
  "n10743675": "utterer, vocalizer, vocaliser",
  "n10743941": "uxor, ux.",
  "n10744078": "uxoricide",
  "n10744164": "vacationer, vacationist",
  "n10744456": "vaccinee",
  "n10744544": "vagrant, drifter, floater, vagabond",
  "n10744844": "Vaisya",
  "n10745006": "valedictorian, valedictory speaker",
  "n10745181": "valentine",
  "n10745332": "valet, valet de chambre, gentleman, gentleman's gentleman, man",
  "n10745613": "valetudinarian",
  "n10745770": "valley girl",
  "n10745894": "valuer",
  "n10746056": "vandal",
  "n10746195": "Vandal",
  "n10746346": "vanisher",
  "n10746460": "varnisher",
  "n10746581": "vassal, liege, liegeman, liege subject, feudatory",
  "n10746799": "vaudevillian",
  "n10746931": "vaulter, pole vaulter, pole jumper",
  "n10747119": "vegetarian",
  "n10747294": "Vedist",
  "n10747424": "vegan",
  "n10747548": "venerator",
  "n10747672": "venter",
  "n10747815": "ventriloquist",
  "n10747965": "venture capitalist",
  "n10748142": "venturer, merchant-venturer",
  "n10748309": "verger",
  "n10748506": "vermin, varmint",
  "n10748620": "very important person, VIP, high-up, dignitary, panjandrum, high muckamuck",
  "n10748804": "vestal",
  "n10748889": "vestryman",
  "n10749004": "vestrywoman",
  "n10749123": "veteran, old-timer, oldtimer, old hand, warhorse, old stager, stager",
  "n10749353": "veteran, vet, ex-serviceman",
  "n10749528": "veteran, veteran soldier",
  "n10749715": "veterinarian, veterinary, veterinary surgeon, vet",
  "n10749928": "vibist, vibraphonist",
  "n10750031": "vicar",
  "n10750188": "vicar",
  "n10750365": "vicar",
  "n10750524": "vicar apostolic",
  "n10750640": "vicar-general",
  "n10750788": "vice admiral",
  "n10750911": "vice chairman",
  "n10751026": "vice chancellor",
  "n10751152": "vicegerent",
  "n10751265": "vice president, V.P.",
  "n10751527": "Vice President of the United States",
  "n10751710": "vice-regent",
  "n10751785": "viceroy, vicereine",
  "n10752020": "vicereine",
  "n10752093": "victim",
  "n10752480": "victim, dupe",
  "n10752719": "victimizer, victimiser",
  "n10752930": "victor, master, superior",
  "n10753061": "Victorian",
  "n10753182": "victualer, victualler",
  "n10753339": "vigilante, vigilance man",
  "n10753442": "villager",
  "n10753546": "villain, scoundrel",
  "n10753779": "villain, baddie",
  "n10753917": "villainess",
  "n10753989": "vintager",
  "n10754088": "vintner, winemaker, wine maker",
  "n10754189": "vintner, wine merchant",
  "n10754281": "violator, debaucher, ravisher",
  "n10754449": "violator, lawbreaker, law offender",
  "n10754578": "violinist, fiddler",
  "n10754920": "violin maker",
  "n10755080": "violist",
  "n10755164": "virago",
  "n10755257": "virgin",
  "n10755394": "virologist",
  "n10755512": "virtuoso",
  "n10755648": "Visayan, Bisayan",
  "n10755815": "viscount",
  "n10755926": "viscountess",
  "n10756061": "viscountess",
  "n10756148": "viscount",
  "n10756261": "Visigoth",
  "n10756433": "visionary, illusionist, seer",
  "n10756641": "visionary",
  "n10756837": "visiting fireman",
  "n10756937": "visiting nurse",
  "n10757050": "visiting professor",
  "n10757193": "visitor, visitant",
  "n10757492": "visualizer, visualiser",
  "n10757625": "visually impaired person",
  "n10757817": "vitalist",
  "n10757918": "vital principle, life principle",
  "n10758121": "viticulturist",
  "n10758207": "vivisectionist",
  "n10758337": "vixen, harpy, hellcat",
  "n10758445": "vizier",
  "n10758589": "vociferator",
  "n10758713": "voice",
  "n10758847": "voicer",
  "n10758949": "voicer",
  "n10759047": "volleyball player",
  "n10759151": "volunteer, unpaid worker",
  "n10759331": "volunteer, military volunteer, voluntary",
  "n10759543": "voluptuary, sybarite",
  "n10759702": "vomiter, spewer",
  "n10759820": "votary",
  "n10759982": "votary",
  "n10760199": "votary",
  "n10760340": "voter, elector",
  "n10760622": "vouchee",
  "n10760763": "voucher, verifier",
  "n10760951": "vower",
  "n10761190": "voyager",
  "n10761326": "voyeur, Peeping Tom, peeper",
  "n10761519": "vulcanizer, vulcaniser",
  "n10761693": "vulgarian",
  "n10761858": "vulgarizer, vulgariser",
  "n10761962": "Wac",
  "n10762064": "waddler",
  "n10762212": "waffler",
  "n10762342": "wag, wit, card",
  "n10762480": "Wagnerian",
  "n10762607": "wagoner, waggoner",
  "n10762693": "wagonwright, waggonwright, wainwright",
  "n10762792": "Wahhabi, Wahabi",
  "n10763075": "waif, street child",
  "n10763245": "wailer",
  "n10763383": "waiter, server",
  "n10763620": "waitress",
  "n10763725": "waiter",
  "n10763878": "waker",
  "n10763985": "waker, rouser, arouser",
  "n10764128": "walk-in",
  "n10764296": "walk-in",
  "n10764465": "walking delegate",
  "n10764622": "walk-on",
  "n10764719": "wallah",
  "n10764907": "wallflower",
  "n10765000": "walloper",
  "n10765098": "walloper",
  "n10765189": "wallpaperer, wall-paperer",
  "n10765305": "wally",
  "n10765435": "Walter Mitty",
  "n10765587": "waltzer",
  "n10765679": "wanderer, roamer, rover, bird of passage",
  "n10765885": "Wandering Jew",
  "n10766025": "wanter, needer",
  "n10766260": "wanton",
  "n10766409": "war baby",
  "n10766492": "warbler",
  "n10766626": "war bride",
  "n10766718": "war correspondent",
  "n10766899": "war criminal",
  "n10767020": "ward",
  "n10767154": "warden",
  "n10767265": "warder",
  "n10767432": "wardress",
  "n10767519": "warehouser, warehouseman",
  "n10767654": "war god, god of war",
  "n10767762": "warlock",
  "n10767837": "warlord",
  "n10768022": "warner",
  "n10768148": "warrantee",
  "n10768272": "warrantee",
  "n10768391": "warrant officer",
  "n10768505": "warrener",
  "n10768585": "warrior",
  "n10768810": "war widow",
  "n10768903": "washer",
  "n10769084": "washerman, laundryman",
  "n10769188": "washwoman, washerwoman, laundrywoman, laundress",
  "n10769321": "wassailer, carouser",
  "n10769459": "wastrel, waster",
  "n10769607": "watchdog",
  "n10769782": "watcher",
  "n10769905": "watchmaker, horologist, horologer",
  "n10770059": "watchman, watcher, security guard",
  "n10770309": "water boy, waterer",
  "n10770433": "water dog, water rat",
  "n10770545": "watercolorist, watercolourist",
  "n10770767": "waterer",
  "n10770891": "water witch, dowser, rhabdomancer",
  "n10771066": "Wave",
  "n10771270": "waver",
  "n10771392": "wayfarer, journeyer",
  "n10771539": "wayfarer",
  "n10771636": "weakling, doormat, wuss",
  "n10771809": "wearer",
  "n10771990": "weasel",
  "n10772092": "weatherman, weather forecaster",
  "n10772190": "weaver",
  "n10772289": "webmaster",
  "n10772392": "wedding guest",
  "n10772472": "weekender",
  "n10772580": "weekend warrior",
  "n10772721": "weekend warrior",
  "n10772937": "weeder",
  "n10773040": "weeper",
  "n10773126": "weeper, crier",
  "n10773277": "weigher",
  "n10773394": "weightlifter, lifter",
  "n10773527": "welcher, welsher",
  "n10773665": "welder",
  "n10773800": "welfare case, charity case",
  "n10773901": "welterweight",
  "n10774018": "welterweight",
  "n10774113": "welterweight",
  "n10774223": "wencher",
  "n10774329": "westerner",
  "n10774440": "West Indian",
  "n10774756": "West-sider",
  "n10774870": "wet nurse, wet-nurse, wetnurse, amah",
  "n10775003": "wetter",
  "n10775128": "whaler",
  "n10775245": "wharf rat",
  "n10775379": "wheedler, coaxer",
  "n10775536": "wheeler",
  "n10775660": "wheelwright, wheeler",
  "n10775771": "whiffer",
  "n10775911": "Whig",
  "n10776052": "Whig",
  "n10776141": "Whig",
  "n10776339": "whiner, complainer, moaner, sniveller, crybaby, bellyacher, grumbler, squawker",
  "n10776766": "whip, party whip",
  "n10776887": "whipper-in",
  "n10776987": "whippersnapper, jackanapes, lightweight",
  "n10777147": "whirling dervish, whirler",
  "n10777299": "whisperer",
  "n10777400": "whistle blower, whistle-blower, whistleblower",
  "n10777768": "whistler",
  "n10777894": "whited sepulcher, whited sepulchre",
  "n10778044": "whiteface",
  "n10778148": "Carmelite, White Friar",
  "n10778345": "Dominican, Black Friar, Blackfriar, friar preacher",
  "n10778553": "Franciscan, Grey Friar",
  "n10778711": "Augustinian",
  "n10778888": "Austin Friar",
  "n10778999": "white hope, great white hope",
  "n10779238": "white separatist",
  "n10779416": "white slave",
  "n10779504": "white slaver",
  "n10779610": "white supremacist",
  "n10779775": "whittler",
  "n10779897": "whoremaster, whoremonger",
  "n10779995": "whoremaster, whoremonger, john, trick",
  "n10780105": "Wiccan, witch",
  "n10780185": "wicket-keeper",
  "n10780284": "widow, widow woman",
  "n10780506": "widower, widowman",
  "n10780632": "wife, married woman",
  "n10781236": "wiggler, wriggler, squirmer",
  "n10781460": "wigmaker",
  "n10781547": "wildcatter",
  "n10781684": "wild man, feral man",
  "n10781817": "wimp, chicken, crybaby",
  "n10781984": "winder",
  "n10782135": "wing",
  "n10782248": "wingback",
  "n10782362": "wing commander",
  "n10782471": "winger",
  "n10782632": "wingman",
  "n10782791": "winner",
  "n10782940": "winner, victor",
  "n10783145": "window cleaner",
  "n10783240": "window dresser, window trimmer",
  "n10783352": "window washer",
  "n10783438": "wine taster",
  "n10783539": "winker",
  "n10783646": "wiper",
  "n10783734": "wireman, wirer",
  "n10783881": "wire-puller",
  "n10784015": "wirer",
  "n10784113": "wise guy, smart aleck, wiseacre, wisenheimer, weisenheimer",
  "n10784281": "junior",
  "n10784436": "wisp",
  "n10784544": "witch doctor",
  "n10784675": "witch-hunter",
  "n10784795": "withdrawer",
  "n10784922": "withdrawer",
  "n10785085": "withdrawer",
  "n10785203": "withdrawer",
  "n10785333": "withdrawer",
  "n10785480": "withdrawer",
  "n10785574": "withholder",
  "n10785695": "withholder",
  "n10785869": "withstander",
  "n10786033": "witness, witnesser, informant",
  "n10786270": "witness, attestant, attestor, attestator",
  "n10786517": "witness",
  "n10786755": "wittol",
  "n10786892": "Wobbly",
  "n10786992": "wog",
  "n10787197": "wolf, woman chaser, skirt chaser, masher",
  "n10787364": "wolf boy",
  "n10787470": "woman, adult female",
  "n10788852": "woman",
  "n10789118": "womanizer, womaniser, philanderer",
  "n10789415": "wonder boy, golden boy",
  "n10789534": "wonderer, marveller",
  "n10789709": "wonderer",
  "n10789820": "wonder woman",
  "n10789963": "woodcarver, carver",
  "n10790077": "woodcutter",
  "n10790192": "woodworker, woodsman, woodman",
  "n10790384": "woodsman, woodman",
  "n10790479": "wool stapler, woolsorter",
  "n10790596": "wool stapler",
  "n10790671": "wordmonger",
  "n10790813": "word-painter",
  "n10790918": "wordsmith",
  "n10791002": "workaholic",
  "n10791115": "working girl",
  "n10791221": "workman, workingman, working man, working person",
  "n10791820": "workmate",
  "n10791890": "worldling",
  "n10792028": "worm, louse, insect, dirt ball",
  "n10792178": "worrier, fuss-budget, fusspot, worrywart",
  "n10792335": "worshiper, worshipper",
  "n10792506": "worthy",
  "n10792746": "wrangler",
  "n10792856": "wrecker",
  "n10793019": "wrester",
  "n10793168": "wrestler, grappler, matman",
  "n10793492": "wretch",
  "n10793570": "wright",
  "n10793799": "write-in candidate, write-in",
  "n10794014": "writer, author",
  "n10801291": "writer",
  "n10801561": "Wykehamist",
  "n10801697": "xylophonist",
  "n10801802": "yakuza",
  "n10801893": "Yahoo",
  "n10802043": "yachtsman, yachtswoman",
  "n10802147": "yanker, jerker",
  "n10802283": "Yankee, Yank, Northerner",
  "n10802507": "yard bird, yardbird",
  "n10802621": "yardie",
  "n10802842": "yardman",
  "n10802953": "yardman",
  "n10803031": "yardmaster, trainmaster, train dispatcher",
  "n10803193": "yawner",
  "n10803282": "yenta",
  "n10803463": "yenta",
  "n10803586": "yeoman",
  "n10803691": "yeoman, yeoman of the guard, beefeater",
  "n10803838": "yodeller",
  "n10803978": "yogi",
  "n10804102": "yokel, rube, hick, yahoo, hayseed, bumpkin, chawbacon",
  "n10804287": "young buck, young man",
  "n10804406": "young person, youth, younker, spring chicken",
  "n10804636": "young Turk",
  "n10804732": "Young Turk",
  "n10804923": "yuppie",
  "n10805113": "zany",
  "n10805274": "Zealot",
  "n10805501": "Zionist",
  "n10805638": "zombi, zombie, living dead",
  "n10805783": "zombi, zombie, snake god",
  "n10805932": "zombi, zombie, zombi spirit, zombie spirit",
  "n10806113": "zoo keeper",
  "n10806222": "zoologist, animal scientist",
  "n10806612": "Zurvan",
  "n10806693": "Aalto, Alvar Aalto, Hugo Alvar Henrik Aalto",
  "n10806841": "Aaron",
  "n10807016": "Aaron, Henry Louis Aaron, Hank Aaron",
  "n10807197": "Abel, Niels Abel, Niels Henrik Abel",
  "n10807317": "Abelard, Peter Abelard, Pierre Abelard",
  "n10807487": "Abraham, Ibrahim",
  "n10807858": "Acheson, Dean Acheson, Dean Gooderham Acheson",
  "n10808045": "Adam, Robert Adam",
  "n10808200": "Adams, John Adams, President Adams, President John Adams",
  "n10808353": "Adams, John Quincy Adams, President Adams, President John Quincy Adams",
  "n10808539": "Adams, Sam Adams, Samuel Adams",
  "n10808756": "Adenauer, Konrad Adenauer",
  "n10808886": "Adrian, Edgar Douglas Adrian, Baron Adrian",
  "n10809086": "Aeschylus",
  "n10809223": "Aesop",
  "n10809317": "Agassiz, Louis Agassiz, Jean Louis Rodolphe Agassiz",
  "n10809576": "Agee, James Agee",
  "n10809675": "Agricola, Gnaeus Julius Agricola",
  "n10809857": "Agrippa, Marcus Vipsanius Agrippa",
  "n10810046": "Agrippina, Agrippina the Elder",
  "n10810215": "Agrippina, Agrippina the Younger",
  "n10810397": "Ahab",
  "n10810549": "Aiken, Conrad Aiken, Conrad Potter Aiken",
  "n10810671": "Ailey, Alvin Ailey",
  "n10810818": "a Kempis, Thomas a Kempis",
  "n10810923": "Akhenaton, Akhenaten, Ikhanaton, Amenhotep IV",
  "n10811118": "Alaric",
  "n10811228": "Albee, Edward Albee, Edward Franklin Albeen",
  "n10811352": "Albers, Josef Albers",
  "n10811540": "Albert, Prince Albert, Albert Francis Charles Augustus Emmanuel",
  "n10811708": "Alberti, Leon Battista Alberti",
  "n10811900": "Alcaeus",
  "n10812047": "Alcibiades",
  "n10812225": "Alcott, Louisa May Alcott",
  "n10812360": "Alexander, Alexander the Great",
  "n10812550": "Alexander I, Czar Alexander I, Aleksandr Pavlovich",
  "n10812800": "Alexander II, Czar Alexander II, Alexander the Liberator",
  "n10813049": "Alexander III, Czar Alexander III",
  "n10813204": "Alexander VI, Pope Alexander VI, Borgia, Rodrigo Borgia",
  "n10813374": "Alfred, Alfred the Great",
  "n10813527": "Alger, Horatio Alger",
  "n10813711": "Algren, Nelson Algren",
  "n10813813": "Al-hakim",
  "n10813986": "Alhazen, Alhacen, al-Haytham, Ibn al-Haytham, Al-Hasan ibn al-Haytham",
  "n10814328": "Ali",
  "n10814574": "Ali, Muhammad Ali, Cassius Clay, Cassius Marcellus Clay",
  "n10814776": "Allen, Ethan Allen",
  "n10814953": "Allen, Woody Allen, Allen Stewart Konigsberg",
  "n10815113": "Allen, Gracie Allen, Grace Ethel Cecile Rosalie Allen, Gracie",
  "n10815343": "Alonso, Alicia Alonso",
  "n10815477": "Amati, Nicolo Amati, Nicola Amati",
  "n10815648": "Ambrose, Saint Ambrose, St. Ambrose",
  "n10816136": "Amos",
  "n10816223": "Amundsen, Roald Amundsen",
  "n10816424": "Anaxagoras",
  "n10816644": "Anaximander",
  "n10816890": "Anaximenes",
  "n10817102": "Andersen, Hans Christian Andersen",
  "n10817244": "Anderson, Carl Anderson, Carl David Anderson",
  "n10817458": "Anderson, Marian Anderson",
  "n10817607": "Anderson, Maxwell Anderson",
  "n10817717": "Anderson, Philip Anderson, Philip Warren Anderson, Phil Anderson",
  "n10817935": "Anderson, Sherwood Anderson",
  "n10818088": "Andrew, Saint Andrew, St. Andrew, Saint Andrew the Apostle",
  "n10818312": "Andrews, Roy Chapman Andrews",
  "n10818469": "Anne",
  "n10818715": "Anouilh, Jean Anouilh",
  "n10818860": "Anselm, Saint Anselm, St. Anselm",
  "n10819134": "Anthony, Susan Anthony, Susan B. Anthony, Susan Brownell Anthony",
  "n10819285": "Antichrist",
  "n10819533": "Antigonus, Antigonus Cyclops, Monophthalmos",
  "n10819755": "Antoninus, Aurelius, Marcus Aurelius, Marcus Aurelius Antoninus, Marcus Annius Verus",
  "n10820053": "Antonius Pius",
  "n10820163": "Antony, Anthony, Mark Antony, Mark Anthony, Antonius, Marcus Antonius",
  "n10820444": "Apollinaire, Guillaume Apollinaire, Wilhelm Apollinaris de Kostrowitzki",
  "n10820613": "Appleton, Edward Appleton, Sir Edward Victor Appleton",
  "n10820790": "Aquinas, Thomas Aquinas, Saint Thomas, St. Thomas, Saint Thomas Aquinas, St. Thomas Aquinas",
  "n10821218": "Arafat, Yasser Arafat",
  "n10821379": "Aragon, Louis Aragon",
  "n10821514": "Archimedes",
  "n10821699": "Arendt, Hannah Arendt",
  "n10821867": "Aristarchus",
  "n10822027": "Aristarchus of Samos",
  "n10822211": "Aristophanes",
  "n10822338": "Aristotle",
  "n10822567": "Arius",
  "n10822786": "Arminius, Armin, Hermann",
  "n10822962": "Arminius, Jacobus Arminius, Jacob Harmensen, Jakob Hermandszoon",
  "n10823199": "Armstrong, Louis Armstrong, Satchmo",
  "n10823369": "Armstrong, Neil Armstrong",
  "n10823529": "Arnold, Benedict Arnold",
  "n10823760": "Arnold, Matthew Arnold",
  "n10823894": "Arnold of Brescia",
  "n10824146": "Arp, Jean Arp, Hans Arp",
  "n10824352": "Arrhenius, Svante August Arrhenius",
  "n10824541": "Artaxerxes I, Artaxerxes",
  "n10824710": "Artaxerxes II, Artaxerxes",
  "n10824888": "Arthur, King Arthur",
  "n10825180": "Arthur, Chester A. Arthur, Chester Alan Arthur, President Arthur",
  "n10825407": "Asanga",
  "n10825554": "Asch, Sholem Asch, Shalom Asch, Sholom Asch",
  "n10825718": "Ashe, Arthur Ashe, Arthur Robert Ashe",
  "n10825923": "Ashton, Sir Frederick Ashton",
  "n10826033": "Ashurbanipal, Assurbanipal, Asurbanipal",
  "n10826204": "Asimov, Isaac Asimov",
  "n10826352": "Astaire, Fred Astaire",
  "n10826557": "Astor, John Jacob Astor",
  "n10826717": "Astor, Nancy Witcher Astor, Viscountess Astor",
  "n10826952": "Ataturk, Kemal Ataturk, Kemal Pasha, Mustafa Kemal",
  "n10827155": "Athanasius, Saint Athanasius, St. Athanasius, Athanasius the Great",
  "n10827513": "Athelstan",
  "n10827678": "Attila, Attila the Hun, Scourge of God, Scourge of the Gods",
  "n10827873": "Attlee, Clement Attlee, Clement Richard Attlee, 1st Earl Attlee",
  "n10828091": "Auchincloss, Louis Auchincloss, Louis Stanton Auchincloss",
  "n10828233": "Auden, W. H. Auden, Wystan Hugh Auden",
  "n10828368": "Audubon, John James Audubon",
  "n10828573": "Augustine, Saint Augustine, St. Augustine, Augustine of Hippo",
  "n10828990": "Augustus, Gaius Octavianus, Gaius Julius Caesar Octavianus, Octavian",
  "n10829293": "Austen, Jane Austen",
  "n10829450": "Averroes, ibn-Roshd, Abul-Walid Mohammed ibn-Ahmad Ibn-Mohammed ibn-Roshd",
  "n10829733": "Avicenna, ibn-Sina, Abu Ali al-Husain ibn Abdallah ibn Sina",
  "n10830046": "Avogadro, Amedeo Avogadro",
  "n10830229": "Bach, Johann Sebastian Bach",
  "n10830456": "Bacon, Francis Bacon, Sir Francis Bacon, Baron Verulam, 1st Baron Verulam, Viscount St. Albans",
  "n10830731": "Bacon, Roger Bacon",
  "n10831003": "Baedeker, Karl Baedeker",
  "n10831136": "Bailey, Nathan Bailey, Nathaniel Bailey",
  "n10831363": "Bailey, Pearl Bailey, Pearl Mae Bailey",
  "n10831483": "Bakunin, Mikhail Bakunin, Mikhail Aleksandrovich Bakunin",
  "n10831656": "Balanchine, George Balanchine",
  "n10831860": "Balboa, Vasco Nunez de Balboa",
  "n10832085": "Baldwin, Stanley Baldwin, 1st Earl Baldwin of Bewdley",
  "n10832251": "Baldwin, James Baldwin, James Arthur Baldwin",
  "n10832415": "Balenciaga, Cristobal Balenciaga",
  "n10832568": "Balfour, Arthur James Balfour, 1st Earl of Balfour",
  "n10832731": "Ball, Lucille Ball",
  "n10832908": "Balthazar, Balthasar",
  "n10833111": "Balzac, Honore Balzac, Honore de Balzac",
  "n10833304": "Bankhead, Tallulah Bankhead",
  "n10833425": "Banks, Sir Joseph Banks",
  "n10833595": "Bannister, Roger Bannister, Sir Roger Gilbert Bannister",
  "n10833805": "Banting, F. G. Banting, Sir Frederick Grant Banting",
  "n10834011": "Baraka, Imamu Amiri Baraka, LeRoi Jones",
  "n10834176": "Barany, Robert Barany",
  "n10834337": "Barbarossa, Khayr ad-Din",
  "n10834439": "Barber, Samuel Barber",
  "n10834543": "Bardeen, John Bardeen",
  "n10834690": "Barkley, Alben Barkley, Alben William Barkley",
  "n10834869": "Barnum, P. T. Barnum, Phineas Taylor Barnum",
  "n10835022": "Barrie, James Barrie, J. M. Barrie, James Matthew Barrie, Sir James Matthew Barrie",
  "n10835218": "Barrymore, Maurice Barrymore, Herbert Blythe",
  "n10835450": "Barrymore, Georgiana Barrymore, Georgiana Emma Barrymore",
  "n10835709": "Barrymore, Lionel Barrymore",
  "n10835866": "Barrymore, Ethel Barrymore",
  "n10836029": "Barrymore, John Barrymore",
  "n10836184": "Barth, John Barth, John Simmons Barth",
  "n10836308": "Barth, Karl Barth",
  "n10836413": "Barthelme, Donald Barthelme",
  "n10836555": "Bartholdi, Frederic Auguste Bartholdi",
  "n10836725": "Bartholin, Caspar Bartholin",
  "n10836862": "Bartlett, John Bartlett",
  "n10837023": "Bartlett, Robert Bartlett, Robert Abram Bartlett, Captain Bob",
  "n10837258": "Bartok, Bela Bartok",
  "n10837461": "Baruch",
  "n10837567": "Baruch, Bernard Baruch, Bernard Mannes Baruch",
  "n10837737": "Baryshnikov, Mikhail Baryshnikov",
  "n10837918": "Basil, St. Basil, Basil of Caesarea, Basil the Great, St. Basil the Great",
  "n10838288": "Bathsheba",
  "n10838484": "Baudelaire, Charles Baudelaire, Charles Pierre Baudelaire",
  "n10838665": "Baum, Frank Baum, Lyman Frank Brown",
  "n10838802": "Bayard, Seigneur de Bayard, Chevalier de Bayard, Pierre Terrail, Pierre de Terrail",
  "n10838997": "Bayes, Thomas Bayes",
  "n10839131": "Beadle, George Beadle, George Wells Beadle",
  "n10839329": "Beaumont, Francis Beaumont",
  "n10839469": "Beaumont, William Beaumont",
  "n10839617": "Beauvoir, Simone de Beauvoir",
  "n10839791": "Beaverbrook, 1st Baron Beaverbrook, William Maxwell Aitken",
  "n10840021": "Becket, Thomas a Becket, Saint Thomas a Becket, St. Thomas a Becket",
  "n10840354": "Beckett, Samuel Beckett",
  "n10840563": "Becquerel, Henri Becquerel, Antoine Henri Becquerel",
  "n10840769": "Bede, Saint Bede, St. Bede, Baeda, Saint Baeda, St. Baeda, Beda, Saint Beda, St. Beda, the Venerable Bede",
  "n10841065": "Beecher, Henry Ward Beecher",
  "n10841241": "Beerbohm, Max Beerbohm, Sir Henry Maxmilian Beerbohm",
  "n10841405": "Beethoven, van Beethoven, Ludwig van Beethoven",
  "n10841657": "Begin, Menachem Begin",
  "n10841880": "Behrens, Peter Behrens",
  "n10842030": "Belisarius",
  "n10842213": "Bell, Alexander Bell, Alexander Graham Bell",
  "n10842376": "Bell, Vanessa Bell, Vanessa Stephen",
  "n10842575": "Bell, Melville Bell, Alexander Melville Bell",
  "n10842730": "Bellarmine, Bellarmino, Cardinal Bellarmine, Roberto Francesco Romolo Bellarmine",
  "n10842923": "Bellini, Vincenzo Bellini",
  "n10843035": "Belloc, Hilaire Belloc, Joseph Hilaire Peter Belloc",
  "n10843228": "Bellow, Saul Bellow, Solomon Bellow",
  "n10843425": "Belshazzar",
  "n10843705": "Benchley, Robert Benchley, Robert Charles Benchley",
  "n10843858": "Benedict, Saint Benedict, St. Benedict",
  "n10844031": "Benedict XIV, Prospero Lambertini",
  "n10844231": "Benedict XV, Giacomo della Chiesa",
  "n10844401": "Benedict, Ruth Benedict, Ruth Fulton",
  "n10844527": "Benet, William Rose Benet",
  "n10844667": "Benet, Stephen Vincent Benet",
  "n10844805": "Ben Gurion, David Ben Gurion, David Grun",
  "n10845050": "Benjamin",
  "n10845248": "Bennett, Floyd Bennett",
  "n10845424": "Benny, Jack Benny, Benjamin Kubelsky",
  "n10845603": "Bentham, Jeremy Bentham",
  "n10845763": "Benton, Thomas Hart Benton, Old Bullion",
  "n10845926": "Benton, Thomas Hart Benton",
  "n10846089": "Berg, Alban Berg",
  "n10846224": "Bergman, Ingmar Bergman",
  "n10846400": "Bergman, Ingrid Bergman",
  "n10846504": "Bergson, Henri Bergson, Henri Louis Bergson",
  "n10846693": "Beria, Lavrenti Pavlovich Beria",
  "n10846913": "Bering, Vitus Bering, Behring, Vitus Behring",
  "n10847125": "Berkeley, Bishop Berkeley, George Berkeley",
  "n10847330": "Berlage, Hendrik Petrus Berlage",
  "n10847454": "Berlin, Irving Berlin, Israel Baline",
  "n10847653": "Berlioz, Hector Berlioz, Louis-Hector Berlioz",
  "n10847793": "Bernard, Claude Bernard",
  "n10847994": "Bernhardt, Sarah Bernhardt, Henriette Rosine Bernard",
  "n10848122": "Bernini, Giovanni Lorenzo Bernini",
  "n10848356": "Bernoulli, Jakob Bernoulli, Jacques Bernoulli, James Bernoulli",
  "n10848500": "Bernoulli, Johann Bernoulli, Jean Bernoulli, John Bernoulli",
  "n10848641": "Bernoulli, Daniel Bernoulli",
  "n10848802": "Bernstein, Leonard Bernstein",
  "n10848946": "Berra, Lawrence Peter Berra, Yogi, Yogi Berra",
  "n10849083": "Berry, Chuck Berry, Charles Edward Berry",
  "n10849213": "Bertillon, Alphonse Bertillon",
  "n10849323": "Bertolucci, Bernardo Bertolucci",
  "n10849435": "Berzelius, Jons Jakob Berzelius",
  "n10849625": "Bessel, Friedrich Wilhelm Bessel",
  "n10849873": "Bessemer, Sir Henry Bessemer",
  "n10850049": "Best, C. H. Best, Charles Herbert Best",
  "n10850273": "Bethe, Hans Bethe, Hans Albrecht Bethe",
  "n10850469": "Bethune, Mary McLeod Bethune",
  "n10850667": "Beveridge, William Henry Beveridge, First Baron Beveridge",
  "n10850964": "Bevin, Ernest Bevin",
  "n10851139": "Bierce, Ambrose Bierce, Ambrose Gwinett Bierce",
  "n10851282": "Binet, Alfred Binet",
  "n10851452": "bin Laden, Osama bin Laden, Usama bin Laden",
  "n10851599": "Bismarck, von Bismarck, Otto von Bismarck, Prince Otto von Bismarck, Prince Otto Eduard Leopold von Bismarck, Iron Chancellor",
  "n10851865": "Bizet, Georges Bizet",
  "n10851987": "Black, Shirley Temple Black, Shirley Temple",
  "n10852130": "Black, Joseph Black",
  "n10852320": "Black Hawk, Makataimeshekiakiak",
  "n10852506": "Blair, Tony Blair, Anthony Charles Lynton Blair",
  "n10852669": "Blake, William Blake",
  "n10852803": "Bleriot, Louis Bleriot",
  "n10852961": "Bligh, William Bligh, Captain Bligh",
  "n10853244": "Blitzstein, Marc Blitzstein",
  "n10853413": "Bloch, Ernest Bloch",
  "n10853628": "Blok, Alexander Alexandrovich Blok, Aleksandr Aleksandrovich Blok",
  "n10853767": "Bloomfield, Leonard Bloomfield",
  "n10853932": "Blucher, von Blucher, G. L. von Blucher, Gebhard Leberecht von Blucher",
  "n10854146": "Boccaccio, Giovanni Boccaccio",
  "n10854265": "Bodoni, Gianbattista Bodoni",
  "n10854397": "Boehme, Jakob Boehme, Bohme, Jakob Bohme, Boehm, Jakob Boehm, Behmen, Jakob Behmen",
  "n10854627": "Boell, Heinrich Boell, Heinrich Theodor Boell",
  "n10854777": "Boethius, Anicius Manlius Severinus Boethius",
  "n10855047": "Bogart, Humphrey Bogart, Humphrey DeForest Bogart",
  "n10855200": "Bohr, Niels Bohr, Niels Henrik David Bohr",
  "n10855431": "Boleyn, Anne Boleyn",
  "n10855604": "Bolivar, Simon Bolivar, El Libertador",
  "n10855834": "Boltzmann, Ludwig Boltzmann",
  "n10855987": "Bond, Julian Bond",
  "n10856215": "Bonhoeffer, Dietrich Bonhoeffer",
  "n10856486": "Boniface, Saint Boniface, St. Boniface, Winfred, Wynfrith, Apostle of Germany",
  "n10856799": "Boniface VIII, Benedetto Caetani",
  "n10857001": "Bonney, William H. Bonney, Billie the Kid",
  "n10857159": "Bontemps, Arna Wendell Bontemps",
  "n10857271": "Boole, George Boole",
  "n10857418": "Boone, Daniel Boone",
  "n10857540": "Booth, John Wilkes Booth",
  "n10857697": "Borges, Jorge Borges, Jorge Luis Borges",
  "n10857849": "Borgia, Cesare Borgia",
  "n10858018": "Borgia, Lucrezia Borgia, Duchess of Ferrara",
  "n10858164": "Born, Max Born",
  "n10858333": "Borodin, Aleksandr Borodin, Aleksandr Porfirevich Borodin",
  "n10858468": "Bosch, Hieronymus Bosch, Jerom Bos",
  "n10858577": "Bose, Satyendra N. Bose, Satyendra Nath Bose",
  "n10858872": "Boswell, James Boswell",
  "n10859012": "Botticelli, Sandro Botticelli, Alessandro di Mariano dei Filipepi",
  "n10859194": "Bougainville, Louis Antoine de Bougainville",
  "n10859369": "Boulez, Pierre Boulez",
  "n10859485": "Bowditch, Nathaniel Bowditch",
  "n10859669": "Bowdler, Thomas Bowdler",
  "n10859857": "Bowie, Jim Bowie, James Bowie",
  "n10860108": "Boyle, Robert Boyle",
  "n10860347": "Boyle, Kay Boyle",
  "n10860444": "Bradbury, Ray Bradbury, Ray Douglas Bradbury",
  "n10860589": "Bradford, William Bradford",
  "n10860807": "Bradley, Omar Bradley, Omar Nelson Bradley",
  "n10860999": "Bradley, Thomas Bradley, Tom Bradley",
  "n10861174": "Bradstreet, Anne Bradstreet, Anne Dudley Bradstreet",
  "n10861329": "Brady, James Buchanan Brady, Diamond Jim Brady, Diamond Jim",
  "n10861535": "Brady, Mathew B. Brady",
  "n10861741": "Bragg, Braxton Bragg",
  "n10861926": "Brahe, Tycho Brahe",
  "n10862113": "Brahms, Johannes Brahms",
  "n10862281": "Braille, Louis Braille",
  "n10862507": "Bramante, Donato Bramante, Donato d'Agnolo Bramante",
  "n10862676": "Brancusi, Constantin Brancusi",
  "n10862822": "Brandt, Willy Brandt",
  "n10862999": "Braque, Georges Braque",
  "n10863124": "Braun, von Braun, Wernher von Braun, Wernher Magnus Maximilian von Braun",
  "n10863440": "Braun, Eva Braun",
  "n10863552": "Brecht, Bertolt Brecht",
  "n10863717": "Breuer, Marcel Lajos Breuer",
  "n10863896": "Brezhnev, Leonid Brezhnev, Leonid Ilyich Brezhnev",
  "n10864064": "Bridges, Harry Bridges",
  "n10864204": "Bridget, Saint Bridget, St. Bridget, Brigid, Saint Brigid, St. Brigid, Bride, Saint Bride, St. Bride",
  "n10864428": "Brinton, Daniel Garrison Brinton",
  "n10864635": "Britten, Benjamin Britten, Edward Benjamin Britten, Lord Britten of Aldeburgh",
  "n10864858": "Broca, Pierre-Paul Broca",
  "n10865140": "Brockhouse, Bertram Brockhouse",
  "n10865329": "Broglie, de Broglie, Louis Victor de Broglie",
  "n10865567": "Bronte, Charlotte Bronte",
  "n10865700": "Bronte, Emily Bronte, Emily Jane Bronte, Currer Bell",
  "n10865860": "Bronte, Anne Bronte",
  "n10865990": "Brooke, Rupert Brooke",
  "n10866090": "Brooks, Van Wyck Brooks",
  "n10866217": "Brown, John Brown",
  "n10866385": "Brown, Robert Brown",
  "n10866571": "Browne, Charles Farrar Browne, Artemus Ward",
  "n10866738": "Browne, Hablot Knight Browne, Phiz",
  "n10866883": "Browning, Elizabeth Barrett Browning",
  "n10867064": "Browning, Robert Browning",
  "n10867238": "Browning, John M. Browning, John Moses Browning",
  "n10867473": "Bruce, Robert the Bruce, Robert I",
  "n10867708": "Bruce, David Bruce, Sir David Bruce",
  "n10867933": "Bruch, Max Bruch",
  "n10868025": "Bruckner, Anton Bruckner",
  "n10868177": "Brueghel, Breughel, Bruegel, Pieter Brueghel, Pieter Breughel, Pieter Bruegel, Breughel the Elder, Pieter Brueghel the Elder",
  "n10868397": "Brummell, George Bryan Brummell, Beau Brummell",
  "n10868562": "Brunelleschi, Filippo Brunelleschi",
  "n10868738": "Bruno, Giordano Bruno",
  "n10868980": "Bruno, Saint Bruno, St. Bruno",
  "n10869207": "Brutus, Marcus Junius Brutus",
  "n10869385": "Bryan, William Jennings Bryan, Great Commoner, Boy Orator of the Platte",
  "n10869683": "Buber, Martin Buber",
  "n10869931": "Buchanan, James Buchanan, President Buchanan",
  "n10870072": "Buchner, Eduard Buchner",
  "n10870235": "Buck, Pearl Buck, Pearl Sydenstricker Buck",
  "n10870440": "Budge, Don Budge, John Donald Budge",
  "n10870691": "Bukharin, Nikolai Ivanovich Bukharin",
  "n10870897": "Bullfinch, Charles Bullfinch",
  "n10871129": "Bultmann, Rudolf Bultmann, Rudolf Karl Bultmann",
  "n10871270": "Bunche, Ralph Bunche, Ralph Johnson Bunche",
  "n10871424": "Bunsen, Robert Bunsen, Robert Wilhelm Bunsen",
  "n10871655": "Bunuel, Luis Bunuel",
  "n10871756": "Bunyan, John Bunyan",
  "n10871926": "Burbage, Richard Burbage",
  "n10872105": "Burbank, Luther Burbank",
  "n10872287": "Burger, Warren Burger, Warren E. Burger, Warren Earl Burger",
  "n10872505": "Burgess, Anthony Burgess",
  "n10872624": "Burgoyne, John Burgoyne, Gentleman Johnny",
  "n10872839": "Burk, Martha Jane Burk, Burke, Martha Jane Burke, Calamity Jane",
  "n10873059": "Burke, Edmund Burke",
  "n10873303": "Burnett, Frances Hodgson Burnett, Frances Eliza Hodgson Burnett",
  "n10873505": "Burnham, Daniel Hudson Burnham",
  "n10873679": "Burns, Robert Burns",
  "n10873783": "Burns, George Burns, Nathan Birnbaum",
  "n10873937": "Burnside, A. E. Burnside, Ambrose Everett Burnside",
  "n10874162": "Burr, Aaron Burr",
  "n10874393": "Burroughs, Edgar Rice Burroughs",
  "n10874540": "Burroughs, William Seward Burroughs",
  "n10874706": "Burroughs, William Burroughs, William S. Burroughs, William Seward Burroughs",
  "n10874921": "Burt, Cyril Burt, Cyril Lodowic Burt",
  "n10875107": "Burton, Richard Burton",
  "n10875249": "Burton, Richard Burton, Sir Richard Burton, Sir Richard Francis Burton",
  "n10875468": "Bush, George Bush, George H.W. Bush, George Herbert Walker Bush, President Bush",
  "n10875681": "Bush, Vannevar Bush",
  "n10875910": "Bush, George Bush, George W. Bush, George Walker Bush, President Bush, President George W. Bush, Dubyuh, Dubya",
  "n10876160": "Bushnell, David Bushnell, Father of the Submarine",
  "n10876419": "Butler, Samuel Butler",
  "n10876513": "Butler, Samuel Butler",
  "n10876661": "Butterfield, William Butterfield",
  "n10876798": "Byrd, Richard E. Byrd, Richard Evelyn Byrd, Admiral Byrd",
  "n10877015": "Byrd, William Byrd",
  "n10877253": "Byron, Lord George Gordon Byron, Sixth Baron Byron of Rochdale",
  "n10877456": "Cabell, James Branch Cabell",
  "n10877584": "Cabot, John Cabot, Giovanni Cabato",
  "n10877841": "Cabot, Sebastian Cabot",
  "n10878161": "Caesar, Julius Caesar, Gaius Julius Caesar",
  "n10878375": "Caesar, Sid Caesar, Sidney Caesar",
  "n10878530": "Cage, John Cage, John Milton Cage Jr.",
  "n10878672": "Cagliostro, Count Alessandro di Cagliostro, Giuseppe Balsamo",
  "n10878844": "Cagney, Jimmy Cagney, James Cagney",
  "n10879009": "Calder, Alexander Calder",
  "n10879155": "Calderon, Calderon de la Barca, Pedro Calderon de la Barca",
  "n10879364": "Caldwell, Erskine Caldwell, Erskine Preston Caldwell",
  "n10879551": "Caligula, Gaius, Gaius Caesar",
  "n10879789": "Calixtus II, Guy of Burgundy",
  "n10880024": "Calixtus III, Borgia, Alfonso Borgia",
  "n10880189": "Callas, Maria Callas, Maria Meneghini Callas",
  "n10880398": "Calvin, John Calvin, Jean Cauvin, Jean Caulvin, Jean Chauvin",
  "n10880669": "Calvin, Melvin Calvin",
  "n10880841": "Calvino, Italo Calvino",
  "n10880981": "Campbell, Joseph Campbell",
  "n10881092": "Camus, Albert Camus",
  "n10881269": "Canetti, Elias Canetti",
  "n10881382": "Canute, Cnut, Knut, Canute the Great",
  "n10881616": "Capek, Karel Capek",
  "n10881784": "Capone, Al Capone, Alphonse Capone, Scarface",
  "n10881986": "Capra, Frank Capra",
  "n10882089": "Caravaggio, Michelangelo Merisi da Caravaggio",
  "n10882293": "Carducci, Giosue Carducci",
  "n10882436": "Carew, Thomas Carew",
  "n10882588": "Carl XVI Gustav, Carl XVI Gustaf",
  "n10882706": "Carlyle, Thomas Carlyle",
  "n10882846": "Carmichael, Hoagy Carmichael, Hoagland Howard Carmichael",
  "n10882988": "Carnegie, Andrew Carnegie",
  "n10883202": "Carnegie, Dale Carnegie",
  "n10883380": "Carnot, Sadi Carnot, Nicolas Leonard Sadi Carnot",
  "n10883533": "Carothers, Wallace Carothers, Wallace Hume Carothers",
  "n10883688": "Carrel, Alexis Carrel",
  "n10883871": "Carrere, John Merven Carrere",
  "n10884061": "Carroll, Lewis Carroll, Dodgson, Reverend Dodgson, Charles Dodgson, Charles Lutwidge Dodgson",
  "n10884369": "Carson, Kit Carson, Christopher Carson",
  "n10884597": "Carson, Rachel Carson, Rachel Louise Carson",
  "n10884831": "Carter, Jimmy Carter, James Earl Carter, James Earl Carter Jr., President Carter",
  "n10885006": "Carter, Howard Carter",
  "n10885176": "Cartier, Jacques Cartier",
  "n10885352": "Cartwright, Edmund Cartwright",
  "n10885487": "Caruso, Enrico Caruso",
  "n10885603": "Carver, George Washington Carver",
  "n10885831": "Casals, Pablo Casals",
  "n10885994": "Casanova, Giovanni Jacopo Casanova, Casanova de Seingalt, Giovanni Jacopo Casanova de Seingalt",
  "n10886222": "Cash, Johnny Cash, John Cash",
  "n10886361": "Caspar, Gaspar",
  "n10886558": "Cassirer, Ernst Cassirer",
  "n10886763": "Cassius, Cassius Longinus, Gaius Cassius Longinus",
  "n10886929": "Castro, Fidel Castro, Fidel Castro Ruz",
  "n10887137": "Cather, Willa Cather, Willa Sibert Cather",
  "n10887290": "Catherine I",
  "n10887420": "Catherine II, Catherine, Catherine the Great",
  "n10887593": "Catherine of Aragon, Catherine",
  "n10887790": "Catherine de Medicis",
  "n10887981": "Catullus, Gaius Valerius Catullus",
  "n10888151": "Cavell, Edith Cavell, Edith Louisa Cavell",
  "n10888401": "Cavendish, Henry Cavendish",
  "n10888644": "Caxton, William Caxton",
  "n10888788": "Cellini, Benvenuto Cellini",
  "n10888891": "Celsius, Anders Celsius",
  "n10889032": "Cervantes, Miguel de Cervantes, Cervantes Saavedra, Miguel de Cervantes Saavedra",
  "n10889316": "Cezanne, Paul Cezanne",
  "n10889533": "Chagall, Marc Chagall",
  "n10889689": "Chamberlain, Neville Chamberlain, Arthur Neville Chamberlain",
  "n10889905": "Chambers, William Chambers, Sir William Chambers",
  "n10890032": "Champlain, Samuel de Champlain",
  "n10890211": "Champollion, Jean Francois Champollion",
  "n10890437": "Chandler, Raymond Chandler, Raymond Thornton Chandler",
  "n10890637": "Chaplin, Charlie Chaplin, Sir Charles Spencer Chaplin",
  "n10890868": "Chapman, John Chapman, Johnny Appleseed",
  "n10891029": "Chain, Ernst Boris Chain, Sir Ernst Boris Chain",
  "n10891271": "Capet, Hugh Capet",
  "n10891428": "Cattell, James McKeen Cattell",
  "n10891569": "Cattell, Ray Cattell, R. B. Cattell, Raymond B. Cattell, Raymond Bernard Cattell",
  "n10891830": "Charcot, Jean Martin Charcot",
  "n10891981": "Charlemagne, Carolus, Charles, Charles I, Charles the Great",
  "n10892218": "Charles, Jacques Charles, Jacques Alexandre Cesar Charles",
  "n10892416": "Charles, Prince Charles",
  "n10892564": "Charles, Charles I, Charles Stuart",
  "n10892786": "Charles, Charles II",
  "n10892947": "Charles, Charles II, Charles I, Charles the Bald",
  "n10893153": "Charles, Charles VII",
  "n10893433": "Charles, Charles IX",
  "n10893606": "Chase, Salmon P. Chase, Salmon Portland Chase",
  "n10893830": "Chateaubriand, Francois Rene Chateaubriand, Vicomte de Chateaubriand",
  "n10894065": "Chaucer, Geoffrey Chaucer",
  "n10894208": "Chavez, Cesar Chavez, Cesar Estrada Chavez",
  "n10894365": "Chavez, Carlos Chavez",
  "n10894522": "Cheever, John Cheever",
  "n10894652": "Chekhov, Chekov, Anton Chekhov, Anton Chekov, Anton Pavlovich Chekhov, Anton Pavlovich Chekov",
  "n10894905": "Cherubini, Luigi Cherubini, Maria Luigi Carlo Zenobio Cherubini",
  "n10895073": "Chesterfield, Fourth Earl of Chesterfield, Philip Dormer Stanhope",
  "n10895274": "Chesterton, G. K. Chesterton, Gilbert Keith Chesterton",
  "n10895549": "Chevalier, Maurice Chevalier",
  "n10895688": "Chiang Kai-shek, Chiang Chung-cheng",
  "n10896046": "Chippendale, Thomas Chippendale",
  "n10896255": "Chirico, Giorgio de Chirico",
  "n10896452": "Chomsky, Noam Chomsky, A. Noam Chomsky",
  "n10896644": "Chopin, Frederic Francois Chopin",
  "n10896823": "Chopin, Kate Chopin, Kate O'Flaherty Chopin",
  "n10896987": "Christie, Agatha Christie, Dame Agatha Mary Clarissa Christie",
  "n10897154": "Christopher, Saint Christopher, St. Christopher",
  "n10897312": "Churchill, Winston Churchill, Winston S. Churchill, Sir Winston Leonard Spenser Churchill",
  "n10897594": "Churchill, John Churchill, Duke of Marlborough, First Duke of Marlborough",
  "n10897796": "Ciardi, John Ciardi, John Anthony Ciardi",
  "n10897946": "Cicero, Marcus Tullius Cicero, Tully",
  "n10898133": "Cimabue, Giovanni Cimabue",
  "n10898308": "Cincinnatus, Lucius Quinctius Cincinnatus",
  "n10898549": "Clark, Joe Clark, Charles Joseph Clark",
  "n10898693": "Clark, Kenneth Clark, Kenneth Bancroft Clark",
  "n10898929": "Clark, Mark Clark, Mark Wayne Clark",
  "n10899164": "Clark, William Clark",
  "n10899414": "Claudius, Claudius I, Tiberius Claudius Drusus Nero Germanicus",
  "n10899747": "Clausewitz, Karl von Clausewitz",
  "n10899951": "Clay, Henry Clay, the Great Compromiser",
  "n10900146": "Clay, Lucius Clay, Lucius DuBignon Clay",
  "n10900366": "Cleanthes",
  "n10900524": "Clemenceau, Georges Clemenceau, Georges Eugene Benjamin Clemenceau",
  "n10900730": "Clemens, Samuel Langhorne Clemens, Mark Twain",
  "n10900953": "Clement III, Guibert of Ravenna",
  "n10901192": "Clement VII, Giulio de' Medici",
  "n10901420": "Clement XI, Giovanni Francesco Albani",
  "n10901589": "Clement XIV, Lorenzo Ganganelli",
  "n10901827": "Cleopatra",
  "n10902051": "Cleveland, Grover Cleveland, Stephen Grover Cleveland, President Cleveland",
  "n10902232": "Cline, Martin Cline",
  "n10902409": "Clinton, DeWitt Clinton",
  "n10902591": "Clinton, Bill Clinton, William Jefferson Clinton, President Clinton",
  "n10902752": "Clinton, Hilary Clinton, Hilary Rodham Clinton",
  "n10902934": "Clive, Robert Clive, Baron Clive, Baron Clive of Plassey",
  "n10903172": "Clovis, Clovis I",
  "n10903413": "Coca, Imogene Coca",
  "n10903570": "Cochise",
  "n10903722": "Cochran, Jacqueline Cochran",
  "n10903918": "Cockcroft, Sir John Cockcroft, Sir John Douglas Cockcroft",
  "n10904107": "Cocteau, Jean Cocteau",
  "n10904270": "Cody, William F. Cody, William Frederick Cody, Buffalo Bill, Buffalo Bill Cody",
  "n10904463": "Cohan, George M. Cohan, George Michael Cohan",
  "n10904639": "Cohn, Ferdinand Julius Cohn",
  "n10904821": "Coleridge, Samuel Taylor Coleridge",
  "n10904992": "Colette, Sidonie-Gabrielle Colette, Sidonie-Gabrielle Claudine Colette",
  "n10905159": "Collins, Wilkie Collins, William Wilkie Collins",
  "n10905315": "Columbus, Christopher Columbus, Cristoforo Colombo, Cristobal Colon",
  "n10905568": "Comenius, John Amos Comenius, Jan Amos Komensky",
  "n10905703": "Compton, Arthur Compton, Arthur Holly Compton",
  "n10906048": "Comstock, Anthony Comstock",
  "n10906234": "Comte, Auguste Comte, Isidore Auguste Marie Francois Comte",
  "n10906462": "Conan Doyle, A. Conan Doyle, Arthur Conan Doyle, Sir Arthur Conan Doyle",
  "n10906638": "Condorcet, Marquis de Condorcet, Marie Jean Antoine Nicolas Caritat",
  "n10906822": "Confucius, Kongfuze, K'ung Futzu, Kong the Master",
  "n10907103": "Congreve, William Congreve",
  "n10907236": "Connolly, Maureen Catherine Connolly, Little Mo Connolly",
  "n10907501": "Connors, Jimmy Conors, James Scott Connors",
  "n10907647": "Conrad, Joseph Conrad, Teodor Josef Konrad Korzeniowski",
  "n10907851": "Constable, John Constable",
  "n10907962": "Constantine, Constantine I, Constantine the Great, Flavius Valerius Constantinus",
  "n10908313": "Cook, James Cook, Captain Cook, Captain James Cook",
  "n10908534": "Cooke, Jay Cooke",
  "n10908756": "Cooke, Alistair Cooke, Alfred Alistair Cooke",
  "n10908919": "Coolidge, Calvin Coolidge, President Coolidge",
  "n10909127": "Cooper, James Fenimore Cooper",
  "n10909303": "Cooper, Gary Cooper, Frank Cooper",
  "n10909471": "Cooper, Peter Cooper",
  "n10909724": "Copernicus, Nicolaus Copernicus, Mikolaj Kopernik",
  "n10909929": "Copland, Aaron Copland",
  "n10910076": "Copley, John Copley, John Singleton Copley",
  "n10910305": "Coppola, Francis Ford Coppola",
  "n10910421": "Corbett, Jim Corbett, James John Corbett, Gentleman Jim",
  "n10910580": "Corday, Charlotte Corday, Marie Anne Charlotte Corday d'Armont",
  "n10910769": "Cordoba, Francisco Fernandez Cordoba, Cordova, Francisco Fernandez de Cordova",
  "n10910948": "Corelli, Arcangelo Corelli",
  "n10911104": "Corneille, Pierre Corneille",
  "n10911288": "Cornell, Ezra Cornell",
  "n10911534": "Cornell, Katherine Cornell",
  "n10911687": "Cornwallis, Charles Cornwallis, First Marquess Cornwallis",
  "n10911948": "Corot, Jean Baptiste Camille Corot",
  "n10912079": "Correggio, Antonio Allegri da Correggio",
  "n10912243": "Cortes, Cortez, Hernando Cortes, Hernando Cortez, Hernan Cortes, Hernan Cortez",
  "n10912451": "Cosimo de Medici, Cosimo the Elder",
  "n10912626": "Coue, Emile Coue",
  "n10912802": "Coulomb, Charles Augustin de Coulomb",
  "n10913010": "Couperin, Francois Couperin",
  "n10913203": "Courbet, Gustave Courbet",
  "n10913355": "Court, Margaret Court",
  "n10913503": "Cousteau, Jacques Costeau, Jacques Yves Costeau",
  "n10913641": "Coward, Noel Coward, Sir Noel Pierce Coward",
  "n10913871": "Cowper, William Cowper",
  "n10914006": "Cowper, William Cowper",
  "n10914134": "Craigie, William A. Craigie, Sir William Alexander Craigie",
  "n10914331": "Crane, Hart Crane, Harold Hart Crane",
  "n10914447": "Crane, Stephen Crane",
  "n10914548": "Crawford, Joan Crawford",
  "n10914658": "Crawford, Thomas Crawford",
  "n10914779": "Crazy Horse, Tashunca-Uitco",
  "n10915025": "Crichton, James Crichton, The Admirable Crichton",
  "n10915173": "Crick, Francis Crick, Francis Henry Compton Crick",
  "n10915373": "Crispin, Saint Crispin, St. Crispin",
  "n10915566": "Crockett, Davy Crockett, David Crockett",
  "n10915772": "Croesus",
  "n10915862": "Crohn, Burrill Bernard Crohn",
  "n10916105": "Cromwell, Oliver Cromwell, Ironsides",
  "n10916325": "Cronyn, Hume Cronyn, Hume Blake Cronyn",
  "n10916505": "Crookes, William Crookes, Sir William Crookes",
  "n10916731": "Crosby, Bing Crosby, Harry Lillis Crosby",
  "n10916887": "Crouse, Russel Crouse",
  "n10916993": "Culbertson, Ely Culbertson",
  "n10917164": "Cumberland, William Augustus, Duke of Cumberland, Butcher Cumberland",
  "n10917377": "cummings, e. e. cummings, Edward Estlin Cummings",
  "n10917554": "Cunningham, Merce Cunningham",
  "n10917703": "Curie, Marie Curie, Madame Curie, Marya Sklodowska",
  "n10917999": "Curie, Pierre Curie",
  "n10918119": "Curl, Robert Curl, Robert F. Curl, Robert Floyd Curl Jr.",
  "n10918358": "Currier, Nathaniel Currier",
  "n10918558": "Curtis, William Curtis",
  "n10918679": "Curtiss, Glenn Curtiss, Glenn Hammond Curtiss",
  "n10918834": "Cushing, Harvey Cushing, Harvery Williams Cushing",
  "n10919061": "Custer, George Armstrong Custer, General Custer",
  "n10919278": "Cuvier, Georges Cuvier, Baron Georges Cuvier, Georges Leopold Chretien Frederic Dagobert Cuvier",
  "n10919496": "Cynewulf, Cynwulf",
  "n10919598": "Cyrano de Bergerac, Savinien Cyrano de Bergerac",
  "n10919886": "Cyril, Saint Cyril, St. Cyril",
  "n10920051": "Cyrus, Cyrus the Younger",
  "n10920207": "Cyrus II, Cyrus the Elder, Cyrus the Great",
  "n10920366": "Czerny, Karl Czerny",
  "n10920582": "da Gamma, Vasco da Gamma, Gamma",
  "n10920832": "Daguerre, Louis Jacques Mande Daguerre",
  "n10921009": "Daimler, Gottlieb Daimler",
  "n10921218": "Dali, Salvador Dali",
  "n10921324": "Dalton, John Dalton",
  "n10921571": "Damocles",
  "n10921879": "Damon",
  "n10922019": "Daniel",
  "n10922239": "Dante, Dante Alighieri",
  "n10922501": "Danton, Georges Jacques Danton",
  "n10922772": "Darius I, Darius the Great",
  "n10922968": "Darius III",
  "n10923146": "Darrow, Clarence Darrow, Clarence Seward Darrow",
  "n10923313": "Darwin, Charles Darwin, Charles Robert Darwin",
  "n10923521": "Daumier, Honore Daumier",
  "n10923700": "David",
  "n10924072": "David, Jacques Louis David",
  "n10924231": "David, Saint David, St. David",
  "n10924347": "Davis, Bette Davis",
  "n10924452": "Davis, Dwight Davis, Dwight Filley Davis",
  "n10924649": "Davis, Jefferson Davis",
  "n10924828": "Davis, Miles Davis, Miles Dewey Davis Jr.",
  "n10924987": "Davis, Stuart Davis",
  "n10925132": "Davy, Humphrey Davy, Sir Humphrey Davy",
  "n10925402": "Davys, John Davys, Davis, John Davis",
  "n10925584": "Dawes, William Dawes",
  "n10925772": "Day, Clarence Day, Clarence Shepard Day Jr.",
  "n10925939": "Dayan, Moshe Dayan",
  "n10926066": "Dean, James Dean, James Byron Dean",
  "n10926238": "De Bakey, Michael Ellis De Bakey",
  "n10926429": "Debs, Eugene V. Debs, Eugene Victor Debs",
  "n10926597": "Debussy, Claude Debussey, Claude Achille Debussy",
  "n10926773": "Decatur, Stephen Decatur",
  "n10926917": "Decius",
  "n10927104": "Deere, John Deere",
  "n10927270": "Defoe, Daniel Defoe",
  "n10927424": "De Forest, Lee De Forest, Father of Radio",
  "n10927687": "Degas, Edgar Degas, Hilaire Germain Edgar Degas",
  "n10927824": "de Gaulle, General de Gaulle, Charles de Gaulle, General Charles de Gaulle, Charles Andre Joseph Marie de Gaulle",
  "n10928140": "Dekker, Decker, Thomas Dekker, Thomas Decker",
  "n10928299": "de Kooning, Willem de Kooning",
  "n10928498": "Delacroix, Eugene Delacroix, Ferdinand Victor Eugene Delacroix",
  "n10928645": "de la Mare, Walter de la Mare, Walter John de la Mare",
  "n10928810": "Delbruck, Max Delbruck",
  "n10928978": "Delibes, Leo Delibes, Clement Philibert Leo Delibes",
  "n10929116": "Delilah",
  "n10929316": "Delius, Frederick Delius",
  "n10929437": "Delorme, Philibert Delorme, de l'Orme, Philibert de l'Orme",
  "n10929657": "Demetrius, Demetrius I, Demetrius Poliorcetes",
  "n10929886": "de Mille, Agnes de Mille, Agnes George de Mille",
  "n10930099": "DeMille, Cecil B. DeMille, Cecil Blount DeMille",
  "n10930296": "Democritus",
  "n10930428": "Demosthenes",
  "n10930591": "Dempsey, Jack Dempsey, William Harrison Dempsey, Manassa Mauler",
  "n10930778": "Deng Xiaoping, Teng Hsiao-ping, Teng Hsiaoping",
  "n10930913": "De Niro, Robert De Niro",
  "n10931059": "Depardieu, Gerard Depardieu",
  "n10931167": "De Quincey, Thomas De Quincey",
  "n10931333": "Derain, Andre Derain",
  "n10931452": "Derrida, Jacques Derrida",
  "n10931634": "de Saussure, Ferdinand de Saussure, Saussure",
  "n10931854": "Descartes, Rene Descartes",
  "n10932140": "De Sica, Vittorio De Sica",
  "n10932244": "de Valera, Eamon de Valera",
  "n10932495": "deVries, De Vries, Hugo deVries, Hugo De Vries",
  "n10932696": "Dewar, Sir James Dewar",
  "n10932898": "Dewey, John Dewey",
  "n10933084": "Dewey, George Dewey, Admiral Dewey",
  "n10933266": "Dewey, Melvil Dewey, Melville Louis Kossuth Dewey",
  "n10933449": "Diaghilev, Sergei Diaghilev, Sergei Pavlovich Diaghilev",
  "n10933658": "Diana, Princess Diana, Princess of Wales, Lady Diana Frances Spencer",
  "n10933929": "Diane de Poitiers, Duchesse de Valentinois",
  "n10934154": "Dias, Diaz, Bartholomeu Dias, Bartholomeu Diaz",
  "n10934410": "Dickens, Charles Dickens, Charles John Huffam Dickens",
  "n10934611": "Dickinson, Emily Dickinson",
  "n10934758": "Diderot, Denis Diderot",
  "n10935025": "Didion, Joan Didion",
  "n10935128": "Diesel, Rudolf Diesel, Rudolf Christian Karl Diesel",
  "n10935304": "Dietrich, Marlene Dietrich, Maria Magdalene von Losch",
  "n10935567": "DiMaggio, Joe DiMaggio, Joseph Paul DiMaggio",
  "n10935745": "Dinesen, Isak Dinesen, Blixen, Karen Blixen, Baroness Karen Blixen",
  "n10935968": "Diocletian, Gaius Aurelius Valerius Diocletian",
  "n10936279": "Diogenes",
  "n10936424": "Dionysius, Dionysius the Elder",
  "n10936567": "Diophantus",
  "n10936716": "Dior, Christian Dior",
  "n10936894": "Dirac, Paul Adrien Maurice Dirac",
  "n10937126": "Disney, Walt Disney, Walter Elias Disney",
  "n10937364": "Disraeli, Benjamin Disraeli, First Earl of Beaconsfield",
  "n10937611": "Dix, Dorothea Dix, Dorothea Lynde Dix",
  "n10937882": "Doctorow, E. L. Doctorow, Edgard Lawrence Doctorow",
  "n10938019": "Dolby, Ray M. Dolby",
  "n10938199": "Domingo, Placido Domingo",
  "n10938363": "Dominic, Saint Dominic, St. Dominic, Domingo de Guzman",
  "n10938640": "Domino, Fats Domino, Antoine Domino",
  "n10938821": "Domitian, Titus Flavius Domitianus",
  "n10939040": "Donatello, Donato di Betto Bardi",
  "n10939187": "Donatus, Aelius Donatus",
  "n10939360": "Donizetti, Gaetano Donizetti",
  "n10939475": "Don Juan",
  "n10939630": "Donkin, Bryan Donkin",
  "n10939856": "Donne, John Donne",
  "n10940053": "Doolittle, Jimmy Doolittle, James Harold Doolittle",
  "n10940315": "Doppler, Christian Johann Doppler",
  "n10940474": "Dos Passos, John Dos Passos, John Roderigo Dos Passos",
  "n10940669": "Dostoyevsky, Dostoevski, Dostoevsky, Feodor Dostoyevsky, Fyodor Dostoyevsky, Feodor Dostoevski, Fyodor Dostoevski, Feodor Dostoevsky, Fyodor Dostoevsky, Feodor Mikhailovich Dostoyevsky, Fyodor Mikhailovich Dostoyevsky, Feodor Mikhailovich Dostoevski, Fyodor Mikhailovich Dostoevski, Feodor Mikhailovich Dostoevsky, Fyodor Mikhailovich Dostoevsky",
  "n10941206": "Douglas, Stephen A. Douglas, Stephen Arnold Douglas, Little Giant",
  "n10941515": "Douglass, Frederick Douglass",
  "n10941714": "Dowding, Hugh Dowding, Baron Hugh Caswall Tremenheere Dowding, Dowdy",
  "n10941992": "Dowland, John Dowland",
  "n10942144": "Down, John L. H. Down",
  "n10942279": "Downing, Andrew Jackson Downing",
  "n10942473": "D'Oyly Carte, Richard D'Oyly Carte",
  "n10942675": "Draco",
  "n10942866": "Drake, Francis Drake, Sir Francis Drake",
  "n10943115": "Dreiser, Theodore Dreiser, Theodore Herman Albert Dreiser",
  "n10943256": "Drew, John Drew",
  "n10943405": "Dreyfus, Alfred Dreyfus",
  "n10943659": "Dryden, John Dryden",
  "n10943811": "Du Barry, Comtesse Du Barry, Marie Jeanne Becu",
  "n10944013": "Du Bois, W. E. B. Du Bois, William Edward Burghardt Du Bois",
  "n10944238": "Duchamp, Marcel Duchamp",
  "n10944468": "Dufy, Raoul Dufy",
  "n10944593": "Dukas, Paul Dukas",
  "n10944686": "Dulles, John Foster Dulles",
  "n10944902": "Dumas, Alexandre Dumas",
  "n10945048": "du Maurier, George du Maurier, George Louis Palmella Busson du Maurier",
  "n10945263": "du Maurier, Daphne du Maurier, Dame Daphne du Maurier",
  "n10945415": "Duncan, Isadora Duncan",
  "n10945546": "Duns Scotus, John Duns Scotus",
  "n10945699": "Durant, Will Durant, William James Durant",
  "n10945825": "Durante, Jimmy Durante",
  "n10945977": "Durer, Albrecht Durer",
  "n10946134": "Durkheim, Emile Durkheim",
  "n10946286": "Durrell, Lawrence Durrell, Lawrence George Durrell",
  "n10946481": "Duse, Eleonora Duse",
  "n10946624": "Duvalier, Francois Duvalier, Papa Doc",
  "n10946750": "Duvalier, Jean-Claude Duvalier, Baby Doc",
  "n10946961": "Dvorak, Antonin Dvorak",
  "n10947108": "Dylan, Bob Dylan",
  "n10947259": "Eames, Charles Eames",
  "n10947403": "Earhart, Amelia Earhart",
  "n10947628": "Eastman, George Eastman",
  "n10947922": "Eccles, John Eccles, Sir John Carew Eccles",
  "n10948117": "Eck, Johann Eck, Johann Maier Eck, Johann Maier",
  "n10948312": "Eckhart, Johannes Eckhart, Meister Eckhart",
  "n10948478": "Eddington, Sir Arthur Stanley Eddington",
  "n10948656": "Eddy, Mary Baker Eddy, Mary Morse Baker Eddy",
  "n10948798": "Ederle, Gertrude Ederle, Gertrude Caroline Ederle",
  "n10948993": "Edgar",
  "n10949192": "Edison, Thomas Edison, Thomas Alva Edison",
  "n10949424": "Edmund I",
  "n10949586": "Edmund II, Edmund Ironside",
  "n10949782": "Edward, Black Prince",
  "n10949952": "Edward, Prince Edward, Edward Antony Richard Louis",
  "n10950092": "Edward, Edward I",
  "n10950219": "Edward, Edward II",
  "n10950448": "Edward, Edward III",
  "n10950786": "Edward, Edward IV",
  "n10951017": "Edward, Edward V",
  "n10951278": "Edward, Edward VI",
  "n10951459": "Edward, Edward VII, Albert Edward",
  "n10951697": "Edward, Edward VIII, Duke of Windsor",
  "n10951948": "Edwards, Jonathan Edwards",
  "n10952143": "Edward the Confessor, Saint Edward the Confessor, St. Edward the Confessor",
  "n10952421": "Edward the Elder",
  "n10952629": "Edward the Martyr, Saint Edward the Martyr, St. Edward the Martyr",
  "n10952919": "Edwin",
  "n10953035": "Edwy, Eadwig",
  "n10953188": "Egbert",
  "n10953344": "Eglevsky, Andre Eglevsky",
  "n10953473": "Ehrenberg, Ilya Ehrenberg, Ilya Grigorievich Ehrenberg",
  "n10953605": "Ehrlich, Paul Ehrlich",
  "n10953797": "Eichmann, Adolf Eichmann, Karl Adolf Eichmann",
  "n10954039": "Eiffel, Alexandre Gustave Eiffel",
  "n10954180": "Eigen, Manfred Eigen",
  "n10954328": "Eijkman, Christiaan Eijkman",
  "n10954498": "Einstein, Albert Einstein",
  "n10954819": "Einthoven, Willem Einthoven",
  "n10954966": "Eisenhower, Dwight Eisenhower, Dwight D. Eisenhower, Dwight David Eisenhower, Ike, President Eisenhower",
  "n10955282": "Eisenstaedt, Alfred Eisenstaedt",
  "n10955483": "Eisenstein, Sergei Eisenstein, Sergei Mikhailovich Eisenstein",
  "n10955748": "Ekman, Vagn Walfrid Ekman",
  "n10955920": "Eleanor of Aquitaine",
  "n10956134": "Elgar, Sir Edward Elgar, Sir Edward William Elgar",
  "n10956377": "El Greco, Greco, Domenikos Theotocopoulos",
  "n10956612": "Elijah",
  "n10956883": "Eliot, George Eliot, Mary Ann Evans",
  "n10957072": "Eliot, T. S. Eliot, Thomas Stearns Eliot",
  "n10957330": "Elizabeth, Elizabeth I",
  "n10957748": "Elizabeth, Elizabeth II",
  "n10958010": "Ellington, Duke Ellington, Edward Kennedy Ellington",
  "n10958182": "Ellison, Ralph Ellison, Ralph Waldo Ellison",
  "n10958381": "Ellsworth, Oliver Ellsworth",
  "n10958552": "Emerson, Ralph Waldo Emerson",
  "n10958703": "Empedocles",
  "n10958885": "Endecott, Endicott, John Endecott, John Endicott",
  "n10959074": "Enesco, Georges Enesco, George Enescu",
  "n10959223": "Engels, Friedrich Engels",
  "n10959374": "Epictetus",
  "n10959479": "Epicurus",
  "n10959664": "Epstein, Jacob Epstein, Sir Jacob Epstein",
  "n10959857": "Erasmus, Desiderius Erasmus, Gerhard Gerhards, Geert Geerts",
  "n10960230": "Eratosthenes",
  "n10960439": "Erlenmeyer, Richard August Carl Emil Erlenmeyer",
  "n10960561": "Ernst, Max Ernst",
  "n10960772": "Erving, Julius Erving, Julius Winfield Erving, Dr. J",
  "n10960922": "Esaki, Leo Esaki",
  "n10961087": "Esau",
  "n10961372": "Esther",
  "n10961650": "Ethelbert",
  "n10961816": "Ethelred, Ethelred I",
  "n10962041": "Ethelred, Ethelred II, Ethelred the Unready",
  "n10962302": "Euclid",
  "n10962423": "Eugene, Prince Eugene of Savoy",
  "n10962611": "Euler, Leonhard Euler",
  "n10962712": "Euripides",
  "n10962837": "Eusebius, Eusebius of Caesarea",
  "n10963050": "Eustachio, Bartolommeo Eustachio",
  "n10963254": "Evans, Arthur Evans, Sir Arthur John Evans",
  "n10963467": "Evans, Herbert McLean Evans",
  "n10963642": "Evers, Medgar Evers, Medgar Wiley Evers",
  "n10963815": "Evert, Chris Evert, Chrissie Evert, Christine Marie Evert",
  "n10964052": "Eyck, van Eyck, Jan van Eyck",
  "n10964261": "Eysenck, Hans Eysenck, H. J. Eysenck, Hans Jurgen Eysenck",
  "n10964520": "Ezekiel, Ezechiel",
  "n10964660": "Ezra",
  "n10964829": "Faberge, Peter Carl Faberge",
  "n10965019": "Fahd, Fahd ibn Abdel Aziz al-Saud",
  "n10965151": "Fahrenheit, Gabriel Daniel Fahrenheit",
  "n10965361": "Fairbanks, Douglas Fairbanks, Douglas Elton Fairbanks, Julius Ullman",
  "n10965550": "Fairbanks, Douglas Fairbanks Jr.",
  "n10965700": "Faisal, Faisal ibn Abdel Aziz al-Saud",
  "n10965836": "Falla, Manuel de Falla",
  "n10965966": "Fallopius, Gabriele Fallopius, Fallopio, Gabriello Fallopio",
  "n10966145": "Fallot, Etienne-Louis Arthur Fallot",
  "n10966318": "Faraday, Michael Faraday",
  "n10966496": "Farmer, Fannie Farmer, Fannie Merritt Farmer",
  "n10966665": "Farmer, James Leonard Farmer",
  "n10966842": "Farouk I, Faruk I",
  "n10966982": "Farragut, David Glasgow Farragut",
  "n10967152": "Farrell, Eileen Farrell",
  "n10967311": "Farrell, James Thomas Farrell",
  "n10967447": "Fatima, Fatimah",
  "n10967633": "Faulkner, William Faulkner, William Cuthbert Faulkner, Falkner, William Falkner",
  "n10967872": "Fawkes, Guy Fawkes",
  "n10968058": "Fechner, Gustav Theodor Fechner",
  "n10968257": "Feifer, Jules Feifer",
  "n10968401": "Fellini, Federico Fellini",
  "n10968504": "Ferber, Edna Ferber",
  "n10968640": "Ferdinand I, Ferdinand the Great",
  "n10968835": "Ferdinand I",
  "n10968956": "Ferdinand II",
  "n10969118": "Ferdinand III",
  "n10969305": "Ferdinand, King Ferdinand, Ferdinand of Aragon, Ferdinand V, Ferdinand the Catholic",
  "n10969799": "Fermat, Pierre de Fermat",
  "n10969986": "Fermi, Enrico Fermi",
  "n10970279": "Feynman, Richard Feynman, Richard Phillips Feynman",
  "n10970488": "Fiedler, Arthur Fiedler",
  "n10970603": "Fielding, Henry Fielding",
  "n10970718": "Fields, W. C. Fields, William Claude Dukenfield",
  "n10970864": "Fillmore, Millard Fillmore, President Fillmore",
  "n10971080": "Finnbogadottir, Vigdis Finnbogadottir",
  "n10971264": "Firth, J. R. Firth, John Rupert Firth",
  "n10971528": "Fischer, Bobby Fischer, Robert James Fischer",
  "n10971697": "Fischer, Emil Hermann Fischer",
  "n10971852": "Fischer, Hans Fischer",
  "n10971981": "Fitzgerald, Ella Fitzgerald",
  "n10972094": "Fitzgerald, F. Scott Fitzgerald, Francis Scott Key Fitzgerald",
  "n10972298": "Fitzgerald, Edward Fitzgerald",
  "n10972495": "Flaminius, Gaius Flaminius",
  "n10972697": "Flaubert, Gustave Flaubert",
  "n10972825": "Fleming, Alexander Fleming, Sir Alexander Fleming",
  "n10972985": "Fleming, Ian Fleming, Ian Lancaster Fleming",
  "n10973164": "Fletcher, John Fletcher",
  "n10973339": "Flinders, Matthew Flinders, Sir Matthew Flinders",
  "n10973497": "Florey, Howard Florey, Sir Howard Walter Florey",
  "n10973722": "Florio, John Florio",
  "n10973873": "Flory, Paul John Flory",
  "n10974033": "Fonda, Henry Fonda",
  "n10974136": "Fonda, Jane Fonda",
  "n10974271": "Fontanne, Lynn Fontanne",
  "n10974455": "Fonteyn, Dame Margot Fonteyn",
  "n10974592": "Ford, Henry Ford",
  "n10974740": "Ford, Gerald Ford, Gerald R. Ford, Gerald Rudolph Ford, President Ford",
  "n10974971": "Ford, Ford Madox Ford, Ford Hermann Hueffer",
  "n10975101": "Ford, Edsel Bryant Ford",
  "n10975202": "Ford, Henry Ford II",
  "n10975304": "Ford, John Ford",
  "n10975404": "Forester, C. S. Forester, Cecil Scott Forester",
  "n10975583": "Fosbury, Dick Fosbury, Richard D. Fosbury",
  "n10975796": "Foster, Stephen Foster, Stephen Collins Foster",
  "n10976004": "Foucault, Jean Bernard Leon Foucault",
  "n10976256": "Fourier, Charles Fourier, Francois Marie Charles Fourier",
  "n10976468": "Fourier, Jean Baptiste Joseph Fourier, Baron Jean Baptiste Joseph Fourier",
  "n10976708": "Fowler, Henry Watson Fowler",
  "n10976862": "Fox, George Fox",
  "n10976997": "Fox, Charles James Fox",
  "n10977159": "Fragonard, Jean Honore Fragonard",
  "n10977368": "France, Anatole France, Jacques Anatole Francois Thibault",
  "n10977542": "Francis II, Emperor Francis II",
  "n10977660": "Francis Ferdinand, Franz Ferdinand",
  "n10977881": "Francis Joseph, Franz Joseph, Francis Joseph I, Franz Josef I",
  "n10978098": "Francis of Assisi, Saint Francis of Assisi, St. Francis of Assisi, Saint Francis, St. Francis, Giovanni di Bernardone",
  "n10978422": "Franck, James Franck",
  "n10978693": "Franck, Cesar Franck",
  "n10978842": "Franco, Francisco Franco, El Caudillo, General Franco",
  "n10979079": "Franklin, Benjamin Franklin",
  "n10979535": "Franklin, John Hope Franklin",
  "n10979694": "Frazer, James George Frazer, Sir James George Frazer",
  "n10979887": "Frederick I, Frederick Barbarossa, Barbarossa",
  "n10980097": "Frederick I",
  "n10980256": "Frederick II, Holy Roman Emperor Frederick II",
  "n10980448": "Frederick II, Frederick the Great",
  "n10980681": "Frederick William, Great Elector",
  "n10980893": "Frederick William I",
  "n10981089": "Frederick William II",
  "n10981251": "Frederick William III",
  "n10981409": "Frederick William IV",
  "n10981569": "Fremont, John C. Fremont, John Charles Fremont",
  "n10981750": "French, Daniel Chester French",
  "n10981961": "Fresnel, Augustin Jean Fresnel",
  "n10982127": "Freud, Sigmund Freud",
  "n10982295": "Frick, Henry Clay Frick",
  "n10982450": "Friedan, Betty Friedan, Betty Naomi Friedan, Betty Naomi Goldstein Friedan",
  "n10982658": "Friedman, Milton Friedman",
  "n10982870": "Frisch, Karl von Frisch",
  "n10983007": "Frisch, Ragnar Frisch, Ragnar Anton Kittil Frisch",
  "n10983172": "Frisch, Otto Frisch, Otto Robert Frisch",
  "n10983503": "Frobisher, Sir Martin Frobisher",
  "n10983758": "Froebel, Friedrich Froebel, Friedrich Wilhelm August Froebel",
  "n10983931": "Frost, Robert Frost, Robert Lee Frost",
  "n10984126": "Fry, Christopher Fry",
  "n10984257": "Fry, Roger Fry, Roger Eliot Fry",
  "n10984418": "Frye, Northrop Frye, Herman Northrop Frye",
  "n10984589": "Fuchs, Klaus Fuchs, Emil Klaus Julius Fuchs",
  "n10984876": "Fuentes, Carlos Fuentes",
  "n10984979": "Fugard, Athol Fugard",
  "n10985160": "Fulbright, William Fulbright, James William Fulbright",
  "n10985440": "Fuller, Buckminster Fuller, R. Buckminster Fuller, Richard Buckminster Fuller",
  "n10985653": "Fuller, Melville W. Fuller, Melville Weston Fuller",
  "n10985838": "Fulton, Robert Fulton",
  "n10986022": "Funk, Casimir Funk",
  "n10986272": "Furnivall, Frederick James Furnivall",
  "n10986437": "Gable, Clark Gable, William Clark Gable",
  "n10986562": "Gabor, Dennis Gabor",
  "n10986710": "Gaboriau, Emile Gaboriau",
  "n10986866": "Gagarin, Yuri Gagarin, Yuri Alekseyevich Gagarin",
  "n10987044": "Gainsborough, Thomas Gainsborough",
  "n10987176": "Galahad, Sir Galahad",
  "n10987358": "Galbraith, John Galbraith, John Kenneth Galbraith",
  "n10987565": "Galen",
  "n10987724": "Galileo, Galileo Galilei",
  "n10988059": "Gallaudet, Thomas Hopkins Gallaudet",
  "n10988261": "Galois, Evariste Galois",
  "n10988466": "Galsworthy, John Galsworthy",
  "n10988570": "Galton, Francis Galton, Sir Francis Galton",
  "n10988887": "Galvani, Luigi Galvani",
  "n10989099": "Gamow, George Gamow",
  "n10989339": "Gandhi, Mahatma Gandhi, Mohandas Karamchand Gandhi",
  "n10989610": "Gandhi, Indira Gandhi, Indira Nehru Gandhi, Mrs. Gandhi",
  "n10989801": "Garbo, Greta Garbo, Greta Louisa Gustafsson",
  "n10989977": "Garcia Lorca, Frederico Garcia Lorca, Lorca",
  "n10990212": "Gardiner, Samuel Rawson Gardiner",
  "n10990371": "Gardner, Erle Stanley Gardner",
  "n10990509": "Gardner, Isabella Stewart Gardner",
  "n10990733": "Garfield, James Garfield, James A. Garfield, James Abraham Garfield, President Garfield",
  "n10990963": "Garibaldi, Giuseppe Garibaldi",
  "n10991165": "Garland, Judy Garland",
  "n10991303": "Garnier, Jean Louis Charles Garnier",
  "n10991415": "Garrick, David Garrick",
  "n10991583": "Garrison, William Lloyd Garrison",
  "n10991740": "Gaskell, Elizabeth Gaskell, Elizabeth Cleghorn Stevenson Gaskell",
  "n10991936": "Gates, Bill Gates, William Henry Gates",
  "n10992192": "Gatling, Richard Jordan Gatling",
  "n10992336": "Gaudi, Antonio Gaudi, Gaudi i Cornet, Antonio Gaudi i Cornet",
  "n10992528": "Gauguin, Paul Gauguin",
  "n10992675": "Gauss, Karl Gauss, Karl Friedrich Gauss",
  "n10992933": "Gawain, Sir Gawain",
  "n10993098": "Gay-Lussac, Joseph Louis Gay-Lussac",
  "n10993350": "Gehrig, Lou Gehrig, Henry Louis Gehrig",
  "n10993507": "Geiger, Hans Geiger",
  "n10993636": "Geisel, Theodor Seuss Geisel, Dr. Seuss",
  "n10993777": "Gell-Mann, Murray Gell-Mann",
  "n10993936": "Genet, Jean Genet",
  "n10994097": "Genet, Edmund Charles Edouard Genet, Citizen Genet",
  "n10994308": "Genghis Khan, Jinghis Khan, Jenghiz Khan, Temujin",
  "n10994500": "Genseric, Gaiseric",
  "n10994660": "Geoffrey of Monmouth",
  "n10994906": "George, George I",
  "n10995115": "George, George II",
  "n10995292": "George, George III",
  "n10995592": "George, George IV",
  "n10995850": "George, George V",
  "n10996090": "George, George VI",
  "n10996285": "George, Saint George, St. George",
  "n10996533": "Geraint, Sir Geraint",
  "n10996677": "Geronimo",
  "n10996876": "Gershwin, George Gershwin",
  "n10997068": "Gershwin, Ira Gershwin",
  "n10997234": "Gesell, Arnold Gesell, Arnold Lucius Gesell",
  "n10997405": "Gesner, Konrad von Gesner",
  "n10997553": "Giacometti, Alberto Giacometti",
  "n10997742": "Gibbon, Edward Gibbon",
  "n10997888": "Gibbs, Josiah Willard Gibbs",
  "n10997997": "Gibran, Kahlil Gibran",
  "n10998117": "Gibson, Althea Gibson",
  "n10998305": "Gibson, Mel Gibson, Mel Columcille Gerard Gibson",
  "n10998474": "Gibson, C. D. Gibson, Charles Dana Gibson",
  "n10998651": "Gide, Andre Gide, Andre Paul Guillaume Gide",
  "n10998860": "Gielgud, Sir John Gielgud, Arthur John Gielgud",
  "n10999048": "Gilbert, Cass Gilbert",
  "n10999202": "Gilbert, Humphrey Gilbert, Sir Humphrey Gilbert",
  "n10999410": "Gilbert, William Gilbert",
  "n10999584": "Gilbert, William Gilbert, William S. Gilbert, William Schwenk Gilbert, Sir William Gilbert",
  "n10999873": "Gilgamesh",
  "n11000012": "Gillespie, Dizzy Gillespie, John Birks Gillespie",
  "n11000172": "Gillette, King Camp Gilette",
  "n11000349": "Gilman, Charlotte Anna Perkins Gilman",
  "n11000469": "Gilmer, Elizabeth Merriwether Gilmer, Dorothy Dix",
  "n11000660": "Ginsberg, Allen Ginsberg",
  "n11000786": "Giotto, Giotto di Bondone",
  "n11001053": "Girard, Stephen Girard",
  "n11001211": "Giraudoux, Jean Giraudoux, Hippolyte Jean Giraudoux",
  "n11001422": "Gish, Lillian Gish",
  "n11001567": "Gjellerup, Karl Gjellerup",
  "n11001668": "Gladstone, William Gladstone, William Ewart Gladstone",
  "n11001848": "Glaser, Donald Glaser, Donald Arthur Glaser",
  "n11002040": "Glendower, Owen Glendower",
  "n11002191": "Glenn, John Glenn, John Herschel Glenn Jr.",
  "n11002422": "Glinka, Mikhail Glinka, Mikhail Ivanovich Glinka",
  "n11002548": "Gluck, Christoph Willibald von Gluck",
  "n11002684": "Godard, Jean Luc Godard",
  "n11002895": "Goddard, Robert Hutchings Goddard",
  "n11003068": "Godel, Kurt Godel",
  "n11003276": "Godiva, Lady Godiva",
  "n11003599": "Godunov, Boris Godunov, Boris Fyodorovich Godunov",
  "n11003724": "Goebbels, Joseph Goebbels, Paul Joseph Goebbels",
  "n11003918": "Goethals, George Washington Goethals",
  "n11004106": "Goethe, Johann Wolfgang von Goethe",
  "n11004333": "Gogol, Nikolai Vasilievich Gogol",
  "n11004485": "Goldberg, Rube Goldberg, Reuben Lucius Goldberg",
  "n11004731": "Golding, William Golding, Sir William Gerald Golding",
  "n11004861": "Goldman, Emma Goldman",
  "n11005050": "Goldmark, Peter Goldmark, Peter Carl Goldmark",
  "n11005320": "Goldoni, Carlo Goldoni",
  "n11005429": "Goldsmith, Oliver Goldsmith",
  "n11005571": "Goldwyn, Sam Goldwyn, Samuel Goldwyn",
  "n11005780": "Golgi, Camillo Golgi",
  "n11005972": "Goliath",
  "n11006128": "Gombrowicz, Witold Gombrowicz",
  "n11006231": "Gompers, Samuel Gompers",
  "n11006431": "Goncourt, Edmond de Goncourt, Edmond Louis Antoine Huot de Goncourt",
  "n11006689": "Goncourt, Jules de Goncourt, Jules Alfred Huot de Goncourt",
  "n11006889": "Gongora, Luis de Gongora y Argote",
  "n11007059": "Gonne, Maud Gonne",
  "n11007181": "Goodall, Jane Goodall",
  "n11007332": "Goodman, Benny Goodman, Benjamin David Goodman, King of Swing",
  "n11007620": "Goodyear, Charles Goodyear",
  "n11007750": "Gorbachev, Mikhail Gorbachev, Mikhail Sergeyevich Gorbachev",
  "n11007993": "Gordimer, Nadine Gordimer",
  "n11008173": "Gordius",
  "n11008313": "Gore, Al Gore, Albert Gore Jr.",
  "n11008462": "Gorgas, William Crawford Gorgas",
  "n11008647": "Goring, Goering, Hermann Goring, Hermann Goering, Hermann Wilhelm Goring",
  "n11008870": "Gorky, Maksim Gorky, Gorki, Maxim Gorki, Aleksey Maksimovich Peshkov, Aleksey Maximovich Peshkov",
  "n11009115": "Goudy, Frederic Goudy, Frederic William Goudy",
  "n11009273": "Gould, Jay Gould",
  "n11009495": "Gould, Stephen Jay Gould",
  "n11009635": "Gounod, Charles Francois Gounod",
  "n11009773": "Goya, Goya y Lucientes, Francisco Goya, Francisco de Goya, Francisco Jose de Goya, Francisco Jose de Goya y Lucientes",
  "n11010019": "Graf, Steffi Graf, Stephanie Graf",
  "n11010187": "Graham, Martha Graham",
  "n11010385": "Graham, Billy Graham, William Franklin Graham",
  "n11010557": "Grahame, Kenneth Grahame",
  "n11010697": "Grainger, Percy Grainger, Percy Aldridge Grainger, George Percy Aldridge Grainger",
  "n11010936": "Gram, Hans C. J. Gram",
  "n11011123": "Grant, Ulysses Grant, Ulysses S. Grant, Ulysses Simpson Grant, Hiram Ulysses Grant, President Grant",
  "n11011398": "Grant, Cary Grant",
  "n11011559": "Grant, Duncan Grant, Duncan James Corrow Grant",
  "n11011764": "Granville-Barker, Harley Granville-Barker",
  "n11012042": "Grappelli, Stephane Grappelli",
  "n11012153": "Grass, Gunter Grass, Gunter Wilhelm Grass",
  "n11012300": "Graves, Robert Graves, Robert Ranke Graves",
  "n11012474": "Gray, Asa Gray",
  "n11012676": "Gray, Robert Gray",
  "n11012846": "Gray, Thomas Gray",
  "n11012993": "Gray, Louis Harold Gray",
  "n11013191": "Greeley, Horace Greeley",
  "n11013324": "Green, William Green",
  "n11013574": "Greenberg, Joseph Greenberg",
  "n11013743": "Greene, Graham Greene, Henry Graham Greene",
  "n11013876": "Gregory, Gregory I, Saint Gregory I, St. Gregory I, Gregory the Great",
  "n11014212": "Gregory, Gregory VII, Hildebrand",
  "n11014450": "Gregory, Gregory XII, Angelo Correr",
  "n11014652": "Gregory, Gregory XIII, Ugo Buoncompagni",
  "n11014833": "Gregory, Gregory XVI, Bartolomeo Alberto Capillari",
  "n11015080": "Gregory, Gregory Nazianzen, Gregory of Nazianzen, St. Gregory of Nazianzen",
  "n11015420": "Gresham, Sir Thomas Gresham",
  "n11015525": "Gretzky, Wayne Gretzky",
  "n11015650": "Grey, Charles Grey, Second Earl Grey",
  "n11015872": "Grey, Lady Jane Grey",
  "n11016075": "Grey, Zane Grey",
  "n11016199": "Grieg, Edvard Grieg, Edvard Hagerup Grieg",
  "n11016374": "Griffith, D. W. Griffith, David Lewelyn Wark Griffith",
  "n11016563": "Grimm, Jakob Grimm, Jakob Ludwig Karl Grimm",
  "n11016841": "Grimm, Wilhelm Grimm, Wilhelm Karl Grimm",
  "n11017020": "Gris, Jaun Gris",
  "n11017118": "Gromyko, Andrei Gromyko, Andrei Andreyevich Gromyko",
  "n11017295": "Gropius, Walter Gropius",
  "n11017454": "Grotius, Hugo Grotius, Huig de Groot",
  "n11017661": "Groves, Leslie Richard Groves",
  "n11017831": "Guarneri, Guarnieri, Guarnerius, Andrea Guarneri",
  "n11017987": "Guarneri, Guarnieri, Guarnerius, Guiseppe Guarneri",
  "n11018153": "Guevara, Ernesto Guevara, Che Guevara",
  "n11018439": "Guinevere, Guenevere",
  "n11018683": "Guest, Edgar Guest, Edgar Albert Guest",
  "n11018862": "Guggenheim, Meyer Guggenheim",
  "n11019073": "Guggenheim, Solomon Guggenheim",
  "n11019269": "Guinness, Alec Guinness, Sir Alec Guinness",
  "n11019425": "Gustavus, Gustavus I",
  "n11019570": "Gustavus, Gustavus II, Gustavus Adolphus",
  "n11019900": "Gustavus, Gustavus III",
  "n11020069": "Gustavus, Gustavus IV",
  "n11020225": "Gustavus, Gustavus V",
  "n11020375": "Gustavus, Gustavus VI",
  "n11020513": "Gutenberg, Johann Gutenberg, Johannes Gutenberg",
  "n11020721": "Guthrie, Woody Guthrie, Woodrow Wilson Guthrie",
  "n11020888": "Gwynn, Gywn, Gynne, Nell Gwynn, Nell Gywn, Nell Gwynne, Eleanor Gwynn, Eleanor Gwyn, Eleanor Gwynne",
  "n11021100": "Habakkuk",
  "n11021178": "Haber, Fritz Haber",
  "n11021342": "Hadrian, Publius Aelius Hadrianus, Adrian",
  "n11021667": "Haeckel, Ernst Heinrich Haeckel",
  "n11021916": "Haggai, Aggeus",
  "n11022001": "Haggard, Rider Haggard, Sir Henry Rider Haggard",
  "n11022160": "Hahn, Otto Hahn",
  "n11022309": "Haile Selassie, Ras Tafari Makonnen, Ras Tafari",
  "n11022465": "Haldane, Richard Haldane, Richard Burdon Haldane, First Viscount Haldane of Cloan",
  "n11022669": "Haldane, Elizabeth Haldane, Elizabeth Sanderson Haldane",
  "n11022848": "Haldane, John Haldane, John Scott Haldane",
  "n11023070": "Haldane, J. B. S. Haldane, John Burdon Sanderson Haldane",
  "n11023327": "Hale, Edward Everett Hale",
  "n11023442": "Hale, George Ellery Hale",
  "n11023623": "Hale, Nathan Hale",
  "n11023883": "Halevy, Fromental Halevy, Jacques Francois Fromental Elie Halevy",
  "n11024033": "Haley, Alex Haley",
  "n11024226": "Haley, Bill Haley, William John Clifton Haley Jr.",
  "n11024419": "Hall, Asaph Hall",
  "n11024582": "Hall, Charles Francis Hall",
  "n11024731": "Hall, Charles Martin Hall",
  "n11024908": "Hall, G. Stanley Hall, Granville Stanley Hall",
  "n11025125": "Hall, Radclyffe Hall, Marguerite Radclyffe Hall",
  "n11025326": "Halley, Edmond Halley, Edmund Halley",
  "n11025508": "Hals, Frans Hals",
  "n11025668": "Hamilton, Alexander Hamilton",
  "n11025926": "Hamilton, Alice Hamilton",
  "n11026078": "Hamilton, Lady Emma Hamilton, Amy Lyon",
  "n11026231": "Hamilton, William Rowan Hamilton, Sir William Rowan Hamilton",
  "n11026372": "Hammarskjold, Dag Hammarskjold, Dag Hjalmar Agne Carl Hammarskjold",
  "n11026597": "Hammerstein, Oscar Hammerstein, Oscar Hammerstein II",
  "n11026816": "Hammett, Dashiell Hammett, Samuel Dashiell Hammett",
  "n11026981": "Hammurabi, Hammurapi",
  "n11027127": "Hampton, Lionel Hampton",
  "n11027294": "Hamsun, Knut Hamsun, Knut Pedersen",
  "n11027416": "Hancock, John Hancock",
  "n11027631": "Handel, George Frideric Handel, George Frederick Handel, Georg Friedrich Handel",
  "n11027885": "Handy, W. C. Handy, William Christopher Handy",
  "n11028074": "Hanks, Tom Hanks, Thomas J. Hanks",
  "n11028196": "Hannibal",
  "n11028446": "Harding, Warren Harding, Warren Gamaliel Harding, President Harding",
  "n11028675": "Hardy, Thomas Hardy",
  "n11028780": "Hardy, Oliver Hardy",
  "n11029005": "Hargreaves, James Hargreaves",
  "n11029132": "Harlow, Jean Harlow, Harlean Carpenter",
  "n11029298": "Harmsworth, Alfred Charles William Harmsworth, Viscount Northcliffe",
  "n11029454": "Harold I, King Harold I, Harold Harefoot, Harefoot",
  "n11029637": "Harold II, King Harold II",
  "n11029888": "Harriman, E. H. Harriman, Edward Henry Harriman",
  "n11030025": "Harriman, Averell Harriman, William Averell Harriman",
  "n11030260": "Harris, Benjamin Harris",
  "n11030395": "Harris, Bomber Harris, Sir Arthur Travers Harris",
  "n11030679": "Harris, Frank Harris, James Thomas Harris",
  "n11030855": "Harris, Townsend Harris",
  "n11031016": "Harris, Zellig Harris, Zellig Sabbatai Harris",
  "n11031257": "Harris, Joel Harris, Joel Chandler Harris",
  "n11031420": "Harrison, William Henry Harrison, President Harrison, President William Henry Harrison",
  "n11031668": "Harrison, Benjamin Harrison, President Harrison, President Benjamin Harrison",
  "n11031842": "Harrison, George Harrison",
  "n11031995": "Harrison, Rex Harrison, Sir Rex Harrison, Reginald Carey Harrison",
  "n11032158": "Harrod, Charles Henry Harrod",
  "n11032359": "Harrod, Charles Digby Harrod",
  "n11032541": "Hart, Lorenz Hart, Lorenz Milton Hart",
  "n11032700": "Hart, Moss Hart",
  "n11032840": "Harte, Bret Harte",
  "n11033003": "Hartley, David Hartley",
  "n11033159": "Harvard, John Harvard",
  "n11033358": "Harvey, William Harvey",
  "n11033631": "Hasdrubal",
  "n11033870": "Hasek, Jaroslav Hasek",
  "n11033992": "Hassam, Childe Hassam, Frederick Childe Hassam",
  "n11034167": "Hassel, Odd Hassel",
  "n11034307": "Hastings, Thomas Hastings",
  "n11034485": "Hathaway, Anne Hathaway",
  "n11034596": "Havel, Vaclav Havel",
  "n11034874": "Hawking, Stephen Hawking, Stephen William Hawking",
  "n11035017": "Hawkins, Coleman Hawkins",
  "n11035132": "Hawkins, Hawkyns, Sir John Hawkins, Sir John Hawkyns",
  "n11035363": "Haworth, Sir Walter Norman Haworth",
  "n11035618": "Hawthorne, Nathaniel Hawthorne",
  "n11035780": "Haydn, Joseph Haydn, Franz Joseph Haydn",
  "n11035957": "Hayek, Friedrich August von Hayek",
  "n11036140": "Hayes, Rutherford B. Hayes, Rutherford Birchard Hayes, President Hayes",
  "n11036405": "Hayes, Helen Hayes",
  "n11036521": "Hays, Arthur Garfield Hays",
  "n11036668": "Hays, Will Hays, William Harrison Hays",
  "n11036911": "Haywood, Big Bill Haywood, William Dudley Haywood",
  "n11037157": "Hazlitt, William Hazlitt",
  "n11037278": "Hearst, William Randolph Hearst",
  "n11037495": "Heaviside, Oliver Heaviside",
  "n11037836": "Hebbel, Friedrich Hebbel, Christian Friedrich Hebbel",
  "n11037966": "Hecht, Ben Hecht",
  "n11038084": "Hegel, Georg Wilhelm Friedrich Hegel",
  "n11038293": "Heidegger, Martin Heidegger",
  "n11038505": "Heinlein, Robert A. Heinlein, Robert Anson Heinlein",
  "n11038657": "Heinz, Henry John Heinz",
  "n11038810": "Heisenberg, Werner Karl Heisenberg",
  "n11038978": "Heller, Joseph Heller",
  "n11039188": "Hellman, Lillian Hellman",
  "n11039344": "Helmholtz, Hermann von Helmholtz, Hermann Ludwig Ferdinand von Helmholtz, Baron Hermann Ludwig Ferdinand von Helmholtz",
  "n11039577": "Heloise",
  "n11039690": "Heming, Hemminge, John Heming, John Hemminge",
  "n11039860": "Hemingway, Ernest Hemingway",
  "n11040024": "Hendrix, Jimi Hendrix, James Marshall Hendrix",
  "n11040240": "Henry, Joseph Henry",
  "n11040381": "Henry, Patrick Henry",
  "n11040596": "Henry, William Henry",
  "n11040800": "Henry I, Henry Beauclerc",
  "n11040985": "Henry II",
  "n11041236": "Henry II",
  "n11041447": "Henry III",
  "n11041660": "Henry III",
  "n11041814": "Henry IV, Bolingbroke, Henry Bolingbroke",
  "n11042036": "Henry IV",
  "n11042165": "Henry IV, Henry of Navarre, Henry the Great",
  "n11042482": "Henry V",
  "n11042691": "Henry VI",
  "n11043040": "Henry VII, Henry Tudor",
  "n11043389": "Henry VII",
  "n11043519": "Henry VIII",
  "n11043836": "Henson, Jim Henson",
  "n11043980": "Hepburn, Katharine Hepburn, Katharine Houghton Hepburn",
  "n11044168": "Hepworth, Barbara Hepworth, Dame Barbara Hepworth",
  "n11044295": "Heraclitus",
  "n11044517": "Herbart, Johann Friedrich Herbart",
  "n11044629": "Herbert, Victor Herbert",
  "n11044789": "Herder, Johann Gottfried von Herder",
  "n11044939": "Herman, Woody Herman, Woodrow Charles Herman",
  "n11045106": "Hero, Heron, Hero of Alexandria",
  "n11045353": "Herod, Herod the Great",
  "n11045569": "Herodotus",
  "n11045796": "Herrick, Robert Herrick",
  "n11045898": "Herschel, William Herschel, Sir William Herschel, Sir Frederick William Herschel",
  "n11046169": "Herschel, John Herschel, Sir John Herschel, Sir John Frederick William Herschel",
  "n11046457": "Hershey, Milton Snavely Hershey",
  "n11046722": "Hertz, Gustav Hertz, Gustav Ludwig Hertz",
  "n11046934": "Hertz, Heinrich Hertz, Heinrich Rudolph Hertz",
  "n11047139": "Herzberg, Gerhard Herzberg",
  "n11047333": "Hesiod",
  "n11047521": "Hess, Victor Hess, Victor Franz Hess",
  "n11047701": "Hess, Rudolf Hess, Walther Richard Rudolf Hess",
  "n11047961": "Hess, Walter Hess, Walter Rudolf Hess",
  "n11048109": "Hess, Dame Myra Hess",
  "n11048205": "Hesse, Hermann Hesse",
  "n11048389": "Hevesy, George Charles Hevesy de Hevesy",
  "n11048586": "Heyerdahl, Thor Hyerdahl",
  "n11048739": "Heyrovsky, Joroslav Heyrovsky",
  "n11048879": "Heyse, Paul Heyse, Paul Johann Ludwig von Heyse",
  "n11049001": "Heyward, DuBois Heyward, Edwin DuBois Hayward",
  "n11049128": "Hezekiah, Ezekias",
  "n11049278": "Hiawatha",
  "n11049443": "Hickock, Wild Bill Hickock, James Butler Hickock",
  "n11049608": "Higginson, Thomas Higginson, Thomas Wentworth Storrow Higginson",
  "n11049835": "Hilbert, David Hilbert",
  "n11049938": "Hill, Benny Hill, Alfred Hawthorne",
  "n11050057": "Hill, J. J. Hill, James Jerome Hill",
  "n11050183": "Hillary, Edmund Hillary, Sir Edmund Hillary, Sir Edmund Percival Hillary",
  "n11050436": "Hillel",
  "n11050537": "Himmler, Heinrich Himmler",
  "n11050723": "Hinault, Bernard Hinault",
  "n11050870": "Hindemith, Paul Hindemith",
  "n11051070": "Hindenburg, Paul von Hindenburg, Paul Ludwig von Beneckendorff und von Hindenburg",
  "n11051376": "Hipparchus",
  "n11051632": "Hippocrates",
  "n11051822": "Hirohito, Michinomiya Hirohito",
  "n11052043": "Hirschfeld, Al Hirschfeld",
  "n11052186": "Hirschsprung, Harold Hirschsprung",
  "n11052299": "Hitchcock, Alfred Hitchcock, Sir Alfred Hitchcock, Alfred Joseph Hitchcock",
  "n11052498": "Hitchings, George Herbert Hitchings",
  "n11052672": "Hitler, Adolf Hitler, Der Fuhrer",
  "n11052843": "Hoagland, Hudson Hoagland",
  "n11052955": "Hobbes, Thomas Hobbes",
  "n11053218": "Hobbs, Sir Jack Hobbs, John Berry Hobbs",
  "n11053344": "Ho Chi Minh, Nguyen Tat Thanh",
  "n11053559": "Hodgkin, Alan Hodgkin, Sir Alan Hodgkin, Alan Lloyd Hodgkin",
  "n11053817": "Hodgkin, Dorothy Hodgkin, Dorothy Mary Crowfoot Hodgkin",
  "n11054034": "Hodgkin, Thomas Hodgkin",
  "n11054173": "Hoffa, Jimmy Hoffa, James Riddle Hoffa",
  "n11054442": "Hoffman, Dustin Hoffman",
  "n11054563": "Hoffman, Malvina Hoffman",
  "n11054670": "Hoffmann, E. T. A. Hoffmann, Ernst Theodor Amadeus Hoffmann, Ernst Theodor Wilhelm Hoffmann",
  "n11054856": "Hoffmann, Roald Hoffmann",
  "n11055042": "Hoffmann, August Wilhelm von Hoffmann",
  "n11055154": "Hoffmann, Josef Hoffmann",
  "n11055296": "Hoffmannsthal, Hugo von Hoffmannsthal",
  "n11055454": "Hogan, Ben Hogan, William Benjamin Hogan",
  "n11055612": "Hogarth, William Hogarth",
  "n11055807": "Hogg, James Hogg",
  "n11055915": "Hokusai, Katsushika Hokusai",
  "n11056060": "Holbein, Hans Holbein, Holbein the Elder",
  "n11056195": "Holbein, Hans Holbein, Holbein the Younger",
  "n11056466": "Hollerith, Herman Hollerith",
  "n11056654": "Holly, Buddy Holly, Charles Hardin Holley",
  "n11056799": "Holmes, Arthur Holmes",
  "n11056947": "Holmes, Oliver Wendell Holmes",
  "n11057076": "Holmes, Oliver Wendell Holmes Jr.",
  "n11057221": "Holofernes",
  "n11057381": "Homer",
  "n11057548": "Homer, Winslow Homer",
  "n11057679": "Honegger, Arthur Honegger",
  "n11057925": "Hooke, Robert Hooke",
  "n11058335": "Hooker, Richard Hooker",
  "n11058436": "Hooker, Joseph Hooker, Fighting Joe Hooker",
  "n11058633": "Hoover, Herbert Hoover, Herbert Clark Hoover, President Hoover",
  "n11058914": "Hoover, J. Edgar Hoover, John Edgar Hoover",
  "n11059079": "Hoover, William Hoover, William Henry Hoover",
  "n11059263": "Hope, Bob Hope, Leslie Townes Hope",
  "n11059438": "Hopkins, Anthony Hopkins, Sir Anthony Hopkins, Sir Anthony Philip Hopkins",
  "n11059593": "Hopkins, Sir Frederick Gowland Hopkins",
  "n11059772": "Hopkins, Gerard Manley Hopkins",
  "n11059875": "Hopkins, Johns Hopkins",
  "n11060103": "Hopkins, Mark Hopkins",
  "n11060241": "Hopkinson, Francis Hopkinson",
  "n11060416": "Horace",
  "n11060535": "Horne, Lena Horne, Lena Calhoun Horne",
  "n11060688": "Horne, Marilyn Horne",
  "n11060805": "Horney, Karen Horney, Karen Danielsen Horney",
  "n11060937": "Horowitz, Vladimir Horowitz",
  "n11061089": "Horta, Victor Horta",
  "n11061225": "Hosea",
  "n11061317": "Houdini, Harry Houdini, Erik Weisz",
  "n11061552": "Houghton, Henry Oscar Houghton",
  "n11061734": "Housman, A. E. Housman, Alfred Edward Housman",
  "n11061853": "Houston, Sam Houston, Samuel Houston",
  "n11062111": "Howard, Catherine Howard",
  "n11062285": "Howard, Leslie Howard, Leslie Howard Stainer",
  "n11062424": "Howe, Elias Howe",
  "n11062649": "Howe, Julia Ward Howe",
  "n11062801": "Howe, Gordie Howe, Gordon Howe",
  "n11062963": "Howe, Irving Howe",
  "n11063061": "Howells, William Dean Howells",
  "n11063182": "Hoyle, Edmond Hoyle",
  "n11063309": "Hoyle, Fred Hoyle, Sir Fred Hoyle",
  "n11063535": "Hubbard, L. Ron Hubbard",
  "n11063687": "Hubble, Edwin Hubble, Edwin Powell Hubble",
  "n11063944": "Hubel, David Hubel",
  "n11064106": "Hudson, Henry Hudson",
  "n11064330": "Hudson, W. H. Hudson, William Henry Hudson",
  "n11064472": "Huggins, Sir William Huggins",
  "n11064662": "Hughes, Charles Evans Hughes",
  "n11064834": "Hughes, Howard Hughes, Howard Robard Hughes",
  "n11065101": "Hughes, Langston Hughes, James Langston Hughes",
  "n11065229": "Hughes, Ted Hughes, Edward James Hughes",
  "n11065345": "Hugo, Victor Hugo, Victor-Marie Hugo",
  "n11065562": "Hull, Cordell Hull",
  "n11065718": "Hull, Isaac Hull",
  "n11065938": "Humboldt, Baron Alexander von Humboldt, Baron Friedrich Heinrich Alexander von Humboldt",
  "n11066209": "Humboldt, Baron Wilhelm von Humboldt, Baron Karl Wilhelm von Humboldt",
  "n11066425": "Hume, David Hume",
  "n11066621": "Humperdinck, Engelbert Humperdinck",
  "n11066772": "Hunt, Leigh Hunt, James Henry Leigh Hunt",
  "n11066938": "Hunt, Richard Morris Hunt",
  "n11067047": "Hunt, Holman Hunt, William Holman Hunt",
  "n11067184": "Huntington, Collis Potter Huntington",
  "n11067396": "Huntington, Samuel Huntington",
  "n11067604": "Huntington, George Huntington",
  "n11067745": "Hurok, Sol Hurok, Solomon Hurok",
  "n11067885": "Huss, John Huss, Hus, Jan Hus",
  "n11068196": "Hussein, Husain, Husayn, ibn Talal Hussein, King Hussein",
  "n11068401": "Hussein, Husain, Husayn, Saddam Hussein, Saddam, Saddam bin Hussein at-Takriti",
  "n11068630": "Husserl, Edmund Husserl",
  "n11068760": "Huston, John Huston",
  "n11068922": "Hutchins, Robert Maynard Hutchins",
  "n11069085": "Hutchinson, Anne Hutchinson",
  "n11069261": "Hutton, James Hutton",
  "n11069430": "Hutton, Sir Leonard Hutton",
  "n11069534": "Huxley, Thomas Huxley, Thomas Henry Huxley",
  "n11069746": "Huxley, Aldous Huxley, Aldous Leonard Huxley",
  "n11069974": "Huxley, Andrew Huxley, Andrew Fielding Huxley",
  "n11070218": "Huygens, Christiaan Huygens, Christian Huygens",
  "n11070387": "Hypatia",
  "n11070531": "Ibert, Jacques Francois Antoine Ibert",
  "n11070644": "Ibsen, Henrik Ibsen, Henrik Johan Ibsen",
  "n11070855": "Iglesias, Julio Iglesias",
  "n11070994": "Ignatius, Saint Ignatius, St. Ignatius",
  "n11071177": "Ignatius of Loyola, Saint Ignatius of Loyola, St. Ignatius of Loyola, Loyola",
  "n11071467": "Indiana, Robert Indiana",
  "n11071574": "Inge, William Inge",
  "n11071677": "Inge, William Ralph Inge, Gloomy Dean",
  "n11071838": "Ingres, Jean Auguste Dominique Ingres",
  "n11071960": "Innocent III, Lotario di Segni",
  "n11072189": "Innocent VIII, Giovanni Battista Cibo",
  "n11072396": "Innocent XI, Benedetto Odescalchi",
  "n11072673": "Innocent XII, Antonio Pignatelli",
  "n11072887": "Ionesco, Eugene Ionesco",
  "n11073061": "Irenaeus, Saint Irenaeus, St. Irenaeus",
  "n11073324": "Irving, John Irving",
  "n11073453": "Irving, Washington Irving",
  "n11073586": "Isaac",
  "n11073795": "Isabella, Queen Isabella, Isabella I, Isabella the Catholic",
  "n11074140": "Isaiah",
  "n11074284": "Isherwood, Christopher Isherwood, Christopher William Bradshaw Isherwood",
  "n11074573": "Ishmael",
  "n11074769": "Isocrates",
  "n11074870": "Issachar",
  "n11075021": "Ivan III, Ivan III Vasilievich, Ivan the Great",
  "n11075219": "Ivan IV, Ivan Iv Vasilievich, Ivan the Terrible",
  "n11075352": "Ivanov, Lev Ivanov",
  "n11075452": "Ives, James Ives, James Merritt Ives",
  "n11075670": "Ives, Charles Edward Ives",
  "n11075823": "Jackson, Andrew Jackson, Old Hickory",
  "n11076079": "Jackson, Thomas Jackson, Thomas J. Jackson, Thomas Jonathan Jackson, Stonewall Jackson",
  "n11076359": "Jackson, Helen Hunt Jackson, Helen Maria Fiske Hunt Jackson",
  "n11076566": "Jackson, Jesse Jackson, Jesse Louis Jackson",
  "n11076820": "Jackson, Mahalia Jackson",
  "n11076965": "Jackson, Michael Jackson, Michael Joe Jackson",
  "n11077195": "Jackson, Glenda Jackson",
  "n11077369": "Jack the Ripper",
  "n11077484": "Jacob, Francois Jacob",
  "n11077649": "Jacobi, Karl Gustav Jacob Jacobi",
  "n11077762": "Jacobs, Aletta Jacobs",
  "n11077927": "Jacobs, Jane Jacobs",
  "n11078059": "Jacobs, W. W. Jacobs, William Wymark Jacobs",
  "n11078203": "Jacquard, Joseph M. Jacquard, Joseph Marie Jacquard",
  "n11078404": "Jaffar, Jafar, Jaffar Umar Thalib, Jafar Umar Thalib",
  "n11078650": "Jagger, Mick Jagger, Michael Philip Jagger",
  "n11078774": "Jakobson, Roman Jakobson, Roman Osipovich Jakobson",
  "n11078982": "James, Saint James, St. James, Saint James the Apostle, St. James the Apostle",
  "n11079252": "James, Henry James",
  "n11079392": "James, William James",
  "n11079544": "James, Jesse James",
  "n11079802": "James, James I, King James, King James I",
  "n11080174": "James, James II",
  "n11080351": "James, James IV",
  "n11080601": "Jamison, Judith Jamison",
  "n11080745": "Jansen, Cornelis Jansen, Cornelius Jansenius",
  "n11080884": "Jarrell, Randall Jarrell",
  "n11080987": "Jaspers, Karl Jaspers, Karl Theodor Jaspers",
  "n11081111": "Jay, John Jay",
  "n11081353": "Jeanne d'Arc, Joan of Arc, Saint Joan",
  "n11081673": "Jeffers, Robinson Jeffers, John Robinson Jeffers",
  "n11081828": "Jefferson, Thomas Jefferson, President Jefferson",
  "n11082135": "Jenner, Edward Jenner",
  "n11082353": "Jenny, William Le Baron Jenny",
  "n11082535": "Jensen, Johannes Vilhelm Jensen",
  "n11082652": "Jeroboam, Jeroboam I",
  "n11082842": "Jeremiah",
  "n11083064": "Jerome, Saint Jerome, St. Jerome, Hieronymus, Eusebius Hieronymus, Eusebius Sophronius Hieronymus",
  "n11083527": "Jespersen, Otto Jespersen, Jens Otto Harry Jespersen",
  "n11083656": "Jesus, Jesus of Nazareth, the Nazarene, Jesus Christ, Christ, Savior, Saviour, Good Shepherd, Redeemer, Deliverer",
  "n11084110": "El Nino",
  "n11084200": "Jevons, William Stanley Jevons",
  "n11084409": "Jewison, Norman Jewison",
  "n11084514": "Jezebel",
  "n11084789": "Jimenez, Juan Ramon Jimenez",
  "n11084895": "Jimenez de Cisneros, Francisco Jimenez de Cisneros",
  "n11085113": "Jinnah, Muhammad Ali Jinnah",
  "n11085267": "Joachim, Joseph Joachim",
  "n11085402": "Job",
  "n11085559": "Joel",
  "n11085633": "Joffre, Joseph Jacques Cesaire Joffre",
  "n11085813": "Joffrey, Robert Joffrey",
  "n11085924": "John, Saint John, St. John, Saint John the Apostle, St. John the Apostle, John the Evangelist, John the Divine",
  "n11086279": "John, King John, John Lackland",
  "n11086607": "John XXIII, Angelo Guiseppe Roncalli",
  "n11086774": "John Chrysostom, St. John Chrysostom",
  "n11087091": "John of Gaunt, Duke of Lancaster",
  "n11087359": "John the Baptist, St. John the Baptist",
  "n11087612": "John Paul I, Albino Luciano",
  "n11087767": "John Paul II, Karol Wojtyla",
  "n11087931": "Johns, Jasper Johns",
  "n11088059": "Johnson, Andrew Johnson, President Johnson, President Andrew Johnson",
  "n11088346": "Johnson, Lyndon Johnson, Lyndon Baines Johnson, LBJ, President Johnson, President Lyndon Johnson",
  "n11088622": "Johnson, Samuel Johnson, Dr. Johnson",
  "n11088771": "Johnston, J. E. Johnston, Joseph Eggleston Johnston",
  "n11088969": "Joliot, Jean-Frederic Joliot, Joliot-Curie, Jean-Frederic Joliot-Curie",
  "n11089318": "Joliot-Curie, Irene Joliot-Curie",
  "n11089484": "Jolliet, Louis Jolliet, Joliet, Louis Joliet",
  "n11089669": "Jolson, Al Jolson, Asa Yoelson",
  "n11089868": "Jonah",
  "n11090136": "Jones, Daniel Jones",
  "n11090235": "Jones, Inigo Jones",
  "n11090378": "Jones, John Paul Jones",
  "n11090512": "Jones, Bobby Jones, Robert Tyre Jones",
  "n11090631": "Jones, Casey Jones, John Luther Jones",
  "n11090884": "Jones, Mother Jones, Mary Harris Jones",
  "n11091084": "Jong, Erica Jong",
  "n11091184": "Jonson, Ben Jonson, Benjamin Jonson",
  "n11091374": "Joplin, Scott Joplin",
  "n11091545": "Joplin, Janis Joplin",
  "n11091706": "Joseph",
  "n11091863": "Joseph",
  "n11092126": "Joseph, Chief Joseph",
  "n11092292": "Josephus, Flavius Josephus, Joseph ben Matthias",
  "n11092541": "Joshua",
  "n11092740": "Joule, James Prescott Joule",
  "n11092938": "Jowett, Benjamin Jowett",
  "n11093116": "Joyce, James Joyce, James Augustine Aloysius Joyce",
  "n11093325": "Juan Carlos, Juan Carlos Victor Maria de Borbon y Borbon",
  "n11093469": "Judah",
  "n11093674": "Judas, Judas Iscariot",
  "n11093848": "Judas Maccabaeus",
  "n11094055": "Jude, Saint Jude, St. Jude, Judas, Thaddaeus",
  "n11094312": "Julian, Julian the Apostate, Flavius Claudius Julianus",
  "n11094611": "Jung, Carl Jung, Carl Gustav Jung",
  "n11094780": "Junkers, Hugo Junkers",
  "n11094928": "Jussieu, Antoine Laurent de Jussieu",
  "n11095123": "Justinian, Justinian I, Justinian the Great",
  "n11095391": "Juvenal, Decimus Junius Juvenalis",
  "n11095587": "Kachaturian, Aram Kachaturian",
  "n11095731": "Kafka, Franz Kafka",
  "n11095907": "Kahn, Louis Isadore Kahn",
  "n11096033": "Kalinin, Mikhail Kalinin, Mikhail Ivanovich Kalinin",
  "n11096192": "Kamehameha I, Kamehameha the Great",
  "n11096338": "Kandinsky, Wassily Kandinsky, Kandinski, Wassily Kandinski",
  "n11096508": "Kant, Immanuel Kant",
  "n11096645": "Karlfeldt, Erik Axel Karlfeldt",
  "n11096801": "Karloff, Boris Karloff, William Henry Pratt",
  "n11096991": "Karpov, Anatoli Karpov, Anatoli Yevgenevich Karpov",
  "n11097208": "Karsavina, Tamara Karsavina",
  "n11097335": "Kasparov, Gary Kasparov, Gary Weinstein",
  "n11097525": "Kastler, Alfred Kastler",
  "n11097625": "Kaufman, George S. Kaufman, George Simon Kaufman",
  "n11097820": "Kaunda, Kenneth Kaunda, Kenneth David Kaunda",
  "n11098039": "Kazan, Elia Kazan, Elia Kazanjoglous",
  "n11098223": "Kean, Edmund Kean",
  "n11098380": "Keaton, Buster Keaton, Joseph Francis Keaton",
  "n11098601": "Keats, John Keats",
  "n11098707": "Keble, John Keble",
  "n11098876": "Kekule, Friedrich August Kekule, Friedrich August Kekule von Stradonitz",
  "n11099085": "Keller, Helen Keller, Helen Adams Keller",
  "n11099438": "Kellogg, W. K. Kellogg, Will Keith Kellog",
  "n11099729": "Kelly, Gene Kelly, Eugene Curran Kelly",
  "n11099923": "Kelly, Grace Kelly, Grace Patricia Kelly, Princess Grace of Monaco",
  "n11100139": "Kelly, Emmett Kelly, Weary Willie",
  "n11100260": "Kelvin, First Baron Kelvin, William Thompson",
  "n11100462": "Kendall, Edward Kendall, Edward Calvin Kendall",
  "n11100619": "Kendrew, Sir John Cowdery Kendrew",
  "n11100798": "Kennan, George F. Kennan, George Frost Kennan",
  "n11101000": "Kennedy, Jack Kennedy, John Fitzgerald Kennedy, JFK, President Kennedy, President John F. Kennedy",
  "n11101250": "Kennelly, A. E. Kennelly, Arthur Edwin Kennelly",
  "n11101576": "Kent, Rockwell Kent",
  "n11101700": "Kenyata, Jomo Kenyata",
  "n11101843": "Keokuk",
  "n11101986": "Kepler, Johannes Kepler, Johan Kepler",
  "n11102144": "Kerensky, Aleksandr Feodorovich Kerensky",
  "n11102353": "Kern, Jerome Kern, Jerome David Kern",
  "n11102493": "Kerouac, Jack Kerouac, Jean-Louis Lebris de Kerouac",
  "n11102674": "Kesey, Ken Kesey, Ken Elton Kesey",
  "n11102878": "Kettering, Charles Kettering, Charles Franklin Kettering",
  "n11103104": "Key, Francis Scott Key",
  "n11103397": "Keynes, John Maynard Keynes",
  "n11103648": "Khachaturian, Aram Khachaturian, Aram Ilich Khachaturian",
  "n11103864": "Khama, Sir Seretse Khama",
  "n11104008": "Khomeini, Ruholla Khomeini, Ayatollah Khomeini, Ayatollah Ruholla Khomeini",
  "n11104287": "Khrushchev, Nikita Khrushchev, Nikita Sergeyevich Khrushchev",
  "n11104458": "Kidd, William Kidd, Captain Kidd",
  "n11104676": "Kierkegaard, Soren Kierkegaard, Soren Aabye Kierkegaard",
  "n11104896": "Kieslowski, Krzysztof Kieslowski",
  "n11105054": "King, Martin Luther King, Martin Luther King Jr.",
  "n11105298": "King, B. B. King, Riley B King",
  "n11105463": "King, Billie Jean King, Billie Jean Moffitt King",
  "n11105609": "Kinsey, Alfred Charles Kinsey",
  "n11105778": "Kipling, Rudyard Kipling, Joseph Rudyard Kipling",
  "n11105945": "Kirchhoff, G. R. Kirchhoff, Gustav Robert Kirchhoff",
  "n11106170": "Kirchner, Ernst Ludwig Kirchner",
  "n11106290": "Kissinger, Henry Kissinger, Henry Alfred Kissinger",
  "n11106479": "Kitchener, Herbert Kitchener, Horatio Herbert Kitchener, First Earl Kitchener of Khartoum",
  "n11106652": "Klaproth, Martin Heinrich Klaproth",
  "n11106830": "Klee, Paul Klee",
  "n11106943": "Klein, Calvin Klein, Calvin Richard Klein",
  "n11107110": "Klein, Melanie Klein",
  "n11107308": "Klein, Felix Klein",
  "n11107436": "Kleist, Heinrich von Kleist, Bernd Heinrich Wilhelm von Kleist",
  "n11107635": "Klimt, Gustav Klimt",
  "n11107757": "Kline, Franz Kline, Franz Joseph Kline",
  "n11107901": "Klinefelter, Harry F. Klinefelter, Harry Fitch Kleinfelter",
  "n11108084": "Klopstock, Friedrich Gottlieb Klopstock",
  "n11108195": "Knox, John Knox",
  "n11108400": "Koch, Robert Koch",
  "n11108584": "Koestler, Arthur Koestler",
  "n11108767": "Konoe, Fumimaro Konoe, Prince Fumimaro Konoe, Konoye, Fumimaro Konoye, Prince Fumimaro Konoye",
  "n11109027": "Koopmans, Tjalling Koopmans, Tjalling Charles Koopmans",
  "n11109192": "Korbut, Olga Korbut",
  "n11109289": "Korchnoi, Viktor Korchnoi, Viktor Lvovich Korchnoi",
  "n11109424": "Korda, Sir Alexander Korda, Sandor Kellner",
  "n11109563": "Korzybski, Alfred Korzybski, Alfred Habdank Skarbek Korzybski",
  "n11109728": "Kosciusko, Thaddeus Kosciusko, Kosciuszko, Tadeusz Andrzej Bonawentura Kosciuszko",
  "n11109970": "Koussevitzky, Serge Koussevitzky, Sergei Aleksandrovich Koussevitzky",
  "n11110205": "Krafft-Ebing, Richard von Krafft-Ebing, Baron Richard von Krafft-Ebing",
  "n11110396": "Krasner, Lee Krasner",
  "n11110617": "Krebs, Hans Adolf Krebs, Sir Hans Adolf Krebs",
  "n11110791": "Kreisler, Fritz Kreisler",
  "n11110917": "Kroeber, Alfred Kroeber, Alfred Louis Kroeber",
  "n11111085": "Kronecker, Leopold Kronecker",
  "n11111194": "Kropotkin, Prince Peter Kropotkin, Pyotr Alexeyevich Kropotkin",
  "n11111335": "Kroto, Harold Kroto, Harold W. Kroto, Sir Harold Walter Kroto",
  "n11111577": "Kruger, Oom Paul Kruger, Stephanus Johannes Paulus Kruger",
  "n11111710": "Krupp, Friedrich Krupp",
  "n11111845": "Krupp, Alfred Krupp",
  "n11112058": "Kublai Khan, Kubla Khan, Kublai Kaan",
  "n11112377": "Kubrick, Stanley Kubrick",
  "n11112488": "Kuhn, Richard Kuhn",
  "n11112628": "Kuiper, Gerard Kuiper, Gerard Peter Kuiper",
  "n11112907": "Kurosawa, Akira Kurosawa",
  "n11113077": "Kutuzov, Mikhail Ilarionovich Kutuzov, Prince of Smolensk",
  "n11113266": "Kuznets, Simon Kuznets",
  "n11113489": "Kyd, Kid, Thomas Kyd, Thomas Kid",
  "n11113601": "Laban, Rudolph Laban",
  "n11113732": "Labrouste, Henri Labrouste",
  "n11113898": "Lachaise, Gaston Lachaise",
  "n11114056": "Lafayette, La Fayette, Marie Joseph Paul Yves Roch Gilbert du Motier, Marquis de Lafayette",
  "n11114285": "Laffer, Arthur Laffer",
  "n11114423": "Laffite, Lafitte, Jean Laffite, Jean Lafitte",
  "n11114637": "La Fontaine, Jean de La Fontaine",
  "n11114791": "Lamarck, Jean Baptiste de Lamarck, Chevalier de Lamarck",
  "n11115029": "Lamb, Charles Lamb, Elia",
  "n11115131": "Lambert, Constant Lambert, Leonard Constant Lambert",
  "n11115293": "Lancelot, Sir Lancelot",
  "n11115558": "Land, Din Land, Edwin Herbert Land",
  "n11115785": "Landau, Lev Davidovich Landau",
  "n11115929": "Landowska, Wanda Landowska",
  "n11116117": "Landsteiner, Karl Landsteiner",
  "n11116284": "Laney, Lucy Craft Laney",
  "n11116466": "Lange, Dorothea Lange",
  "n11116642": "Langley, Samuel Pierpoint Langley",
  "n11116875": "Langmuir, Irving Langmuir",
  "n11117108": "Langtry, Lillie Langtry, Jersey Lillie, Emilie Charlotte le Breton",
  "n11117307": "Lao-tzu, Lao-tse, Lao-zi",
  "n11117451": "Laplace, Marquis de Laplace, Pierre Simon de Laplace",
  "n11117744": "Lardner, Ring Lardner, Ringgold Wilmer Lardner",
  "n11117931": "La Rochefoucauld, Francois de La Rochefoucauld",
  "n11118072": "Larousse, Pierre Larousse, Pierre Athanase Larousse",
  "n11118205": "LaSalle, Sieur de LaSalle, Rene-Robert Cavelier",
  "n11118362": "Lasso, Orlando di Lasso, Roland de Lassus",
  "n11118481": "La Tour, Georges de La Tour",
  "n11118602": "Latrobe, Benjamin Henry Latrobe",
  "n11118886": "Lauder, Harry Lauder, Sir Harry MacLennan Lauder",
  "n11119061": "Laughton, Charles Laughton",
  "n11119190": "Laurel, Stan Laurel, Arthur Stanley Jefferson Laurel",
  "n11119476": "Laurens, Henry Laurens",
  "n11119634": "Laver, Rod Laver, Rodney George Laver",
  "n11119917": "Lavoisier, Antoine Lavoisier, Antoine Laurent Lavoisier",
  "n11120146": "Lawrence, D. H. Lawrence, David Herbert Lawrence",
  "n11120368": "Lawrence, E. O. Lawrence, Ernest Orlando Lawrence",
  "n11120530": "Lawrence, Gertrude Lawrence",
  "n11120633": "Lawrence, Sir Thomas Lawrence",
  "n11120834": "Lawrence, T. E. Lawrence, Thomas Edward Lawrence, Lawrence of Arabia",
  "n11121108": "Lawrence, Saint Lawrence, St. Lawrence, Laurentius",
  "n11121451": "Leacock, Stephen Leacock, Stephen Butler Leacock",
  "n11121640": "Leakey, Louis Leakey, Louis Seymour Bazett Leakey",
  "n11121876": "Leakey, Mary Leakey, Mary Douglas Leakey",
  "n11122114": "Leakey, Richard Leakey, Richard Erskine Leakey",
  "n11122439": "Lear, Edward Lear",
  "n11122579": "Leary, Tim Leary, Timothy Leary, Timothy Francis Leary",
  "n11122825": "le Carre, John le Carre, David John Moore Cornwell",
  "n11122977": "le Chatelier, Henry le Chatelier",
  "n11123124": "Le Corbusier, Charles Edouard Jeanneret",
  "n11123262": "Ledbetter, Huddie Leadbetter, Leadbelly",
  "n11123439": "Le Duc Tho",
  "n11123590": "Lee, Robert E. Lee, Robert Edward Lee",
  "n11123762": "Lee, Henry Lee, Lighthorse Harry Lee",
  "n11123894": "Lee, Richard Henry Lee",
  "n11124088": "Lee, Tsung Dao Lee",
  "n11124300": "Lee, Bruce Lee, Lee Yuen Kam",
  "n11124472": "Lee, Gypsy Rose Lee, Rose Louise Hovick",
  "n11124647": "Lee, Spike Lee, Shelton Jackson Lee",
  "n11124831": "Le Gallienne, Eva Le Gallienne",
  "n11124961": "Leger, Fernand Leger",
  "n11125080": "Lehar, Franz Lehar",
  "n11125193": "Leibniz, Leibnitz, Gottfried Wilhelm Leibniz, Gottfried Wilhelm Leibnitz",
  "n11125546": "Leigh, Vivien Leigh",
  "n11125646": "Lemaitre, Georges Henri Lemaitre, Edouard Lemaitre",
  "n11125840": "Lemmon, Jack Lemmon, John Uhler",
  "n11125957": "Lenard, Philipp Lenard",
  "n11126081": "Lendl, Ivan Lendl",
  "n11126295": "L'Enfant, Charles L'Enfant, Pierre Charles L'Enfant",
  "n11126490": "Lenin, Vladimir Lenin, Nikolai Lenin, Vladimir Ilyich Lenin, Vladimir Ilich Lenin, Vladimir Ilyich Ulyanov, Vladimir Ilich Ulyanov",
  "n11126783": "Lennon, John Lennon",
  "n11127011": "Le Notre, Andre Le Notre",
  "n11127188": "Leo I, St. Leo I, Leo the Great",
  "n11127419": "Leo III",
  "n11127565": "Leo IX, Bruno, Bruno of Toul",
  "n11127752": "Leo X, Giovanni de'Medici",
  "n11127996": "Leo XIII, Gioacchino Pecci, Giovanni Vincenzo Pecci",
  "n11128236": "Leonard, Elmore Leonard, Elmore John Leonard, Dutch Leonard",
  "n11128394": "Leonardo, Leonardo da Vinci, da Vinci",
  "n11128673": "Leonidas",
  "n11128834": "Leontief, Wassily Leontief",
  "n11129017": "Lermontov, Mikhail Yurievich Lermontov",
  "n11129130": "Lerner, Alan Jay Lerner",
  "n11129286": "Lesseps, Ferdinand de Lesseps, Vicomte Ferdinand Marie de Lesseps",
  "n11129478": "Lessing, Doris Lessing, Doris May Lessing",
  "n11129665": "Lessing, Gotthold Ephraim Lessing",
  "n11129808": "Leuwenhoek, Leeuwenhoek, Anton van Leuwenhoek, Anton van Leeuwenhoek",
  "n11130113": "Levi-Strauss, Claude Levi-Strauss",
  "n11130291": "Lewis, C. S. Lewis, Clive Staples Lewis",
  "n11130474": "Lewis, Sinclair Lewis, Harry Sinclair Lewis",
  "n11130661": "Lewis, John L. Lewis, John Llewelly Lewis",
  "n11130940": "Lewis, Meriwether Lewis",
  "n11131135": "Lewis, Carl Lewis, Frederick Carleton Lewis",
  "n11131358": "Lewis, Jerry Lee Lewis",
  "n11131505": "Libby, Willard Frank Libby",
  "n11131658": "Lichtenstein, Roy Lichtenstein",
  "n11131808": "Lie, Trygve Lie, Trygve Halvden Lie",
  "n11131981": "Liliuokalani, Lydia Kamekeha Paki Liliuokalani",
  "n11132117": "Lillie, Beatrice Lillie, Lady Peel",
  "n11132245": "Lin, Maya Lin",
  "n11132462": "Lincoln, Abraham Lincoln, President Lincoln, President Abraham Lincoln",
  "n11132768": "Lind, Jenny Lind, Swedish Nightingale",
  "n11132948": "Lindbergh, Charles Lindbergh, Charles A. Lindbergh, Charles Augustus Lindbergh, Lucky Lindy",
  "n11133197": "Lindsay, Vachel Lindsay, Nicholas Vachel Lindsay",
  "n11133387": "Lindsay, Howard Lindsay",
  "n11133551": "Linnaeus, Carolus Linnaeus, Carl von Linne, Karl Linne",
  "n11133778": "Lipchitz, Jacques Lipchitz",
  "n11133938": "Lipmann, Fritz Albert Lipmann",
  "n11134115": "Li Po",
  "n11134196": "Lippi, Fra Filippo Lippi",
  "n11134339": "Lippi, Filippino Lippi",
  "n11134466": "Lippmann, Gabriel Lippmann",
  "n11134620": "Lippmann, Walter Lippmann",
  "n11134730": "Lipscomb, William Nunn Lipscom Jr.",
  "n11134895": "Lister, Joseph Lister, Baron Lister",
  "n11135044": "Liston, Sonny Liston, Charles Liston",
  "n11135236": "Liszt, Franz Liszt",
  "n11135371": "Littre, Maximilien Paul Emile Littre",
  "n11135488": "Livermore, Mary Ashton Rice Livermore",
  "n11135610": "Livingston, Robert R. Livingston",
  "n11135797": "Livingstone, David Livingstone",
  "n11135991": "Livy, Titus Livius",
  "n11136214": "Lloyd, Harold Lloyd, Harold Clayton Lloyd",
  "n11136405": "Lloyd Webber, Andrew Lloyd Webber, Baron Lloyd Webber of Sydmonton",
  "n11136622": "Lobachevsky, Nikolai Ivanovich Lobachevsky",
  "n11136798": "Locke, John Locke",
  "n11136973": "Lodge, Sir Oliver Lodge, Sir Oliver Joseph Lodge",
  "n11137175": "Loeb, Jacques Loeb",
  "n11137334": "Loewe, Frederick Loewe",
  "n11137506": "Loewi, Otto Loewi",
  "n11137748": "London, Jack London, John Griffith Chaney",
  "n11137928": "Longfellow, Henry Wadsworth Longfellow",
  "n11138085": "Loos, Adolf Loos",
  "n11138180": "Loren, Sophia Loren, Sofia Scicolone",
  "n11138301": "Lorentz, Hendrik Antoon Lorentz",
  "n11138449": "Lorenz, Konrad Lorenz, Konrad Zacharias Lorenz",
  "n11138681": "Lorenzo de'Medici, Lorenzo the Magnificent",
  "n11138924": "Lorre, Peter Lorre, Laszlo Lowestein",
  "n11139092": "Louis I, Louis the Pious",
  "n11139253": "Louis II, Louis le Begue, Louis the Stammerer, Louis the German",
  "n11139403": "Louis III",
  "n11139520": "Louis IV, Louis d'Outremer",
  "n11139619": "Louis V, Louis le Faineant",
  "n11139739": "Louis VI, Louis the Far, Louis the Wideawake, Louis the Bruiser",
  "n11139927": "Louis VII",
  "n11140100": "Louis VIII",
  "n11140243": "Louis IX, Saint Louis, St. Louis",
  "n11140461": "Louis X, Louis le Hutin, Louis the Quarreller",
  "n11140582": "Louis XI",
  "n11140745": "Louis XII",
  "n11140862": "Louis XIII",
  "n11141019": "Louis XIV, Sun King, Louis the Great",
  "n11141289": "Louis XV",
  "n11141495": "Louis XVI",
  "n11141709": "Louis, Joe Louis, Joseph Louis Barrow",
  "n11141882": "Lovelace, Richard Lovelace",
  "n11141981": "Lovell, Sir Bernard Lovell, Sir Alfred Charles Bernard Lovell",
  "n11142155": "Low, David Low, Sir David Low, Sir David Alexander Cecil Low",
  "n11142368": "Lowell, Abbott Lawrence Lowell",
  "n11142517": "Lowell, Amy Lowell",
  "n11142614": "Lowell, Percival Lowell",
  "n11142788": "Lowell, Robert Lowell, Robert Traill Spence Lowell Jr.",
  "n11142922": "Lowry, Malcolm Lowry, Clarence Malcolm Lowry",
  "n11143044": "Lowry, L. S. Lowry, Laurence Stephen Lowry",
  "n11143163": "Lozier, Clemence Sophia Harned Lozier",
  "n11143331": "Lubitsch, Ernst Lubitsch",
  "n11143458": "Lucas, George Lucas",
  "n11143600": "Lucullus, Licinius Lucullus, Lucius Licinius Lucullus",
  "n11143806": "Luce, Clare Booth Luce",
  "n11143933": "Luce, Henry Luce, Henry Robinson Luce",
  "n11144068": "Lucretius, Titus Lucretius Carus",
  "n11144291": "Luculus, Lucius Licinius Luculus",
  "n11144431": "Lugosi, Bela Lugosi, Bela Ferenc Blasko",
  "n11144604": "Luke, Saint Luke, St. Luke",
  "n11144860": "Lully, Jean Baptiste Lully, Lulli, Giambattista Lulli",
  "n11145085": "Lully, Raymond Lully, Ramon Lully",
  "n11145199": "Lunt, Alfred Lunt",
  "n11145364": "Luther, Martin Luther",
  "n11145563": "Lutyens, Sir Edwin Lutyens, Sir Edwin Landseer Luytens",
  "n11145730": "Lyly, John Lyly",
  "n11145850": "Lysander",
  "n11146010": "Lysenko, Trofim Denisovich Lysenko",
  "n11146193": "Lysimachus",
  "n11146407": "Lysippus",
  "n11146494": "Lytton, First Baron Lytton, Bulwer-Lytton, Edward George Earle Bulwer-Lytton",
  "n11146670": "MacArthur, Douglas MacArthur",
  "n11146914": "Macaulay, Thomas Babington Macaulay, First Baron Macaulay, Lord Macaulay",
  "n11147099": "Macbeth",
  "n11147185": "MacDowell, Edward MacDowell",
  "n11147348": "MacGregor, Robert MacGregor, Rob Roy",
  "n11147533": "Mach, Ernst Mach",
  "n11147729": "Machiavelli, Niccolo Machiavelli",
  "n11147924": "Mackenzie, Sir Alexander Mackenzie",
  "n11148152": "MacLeish, Archibald MacLeish",
  "n11148259": "Macleod, John Macleod, John James Rickard Macleod",
  "n11148486": "Madison, James Madison, President Madison",
  "n11148748": "Madonna, Madonna Louise Ciccone",
  "n11148899": "Maeterlinck, Count Maurice Maeterlinck",
  "n11149016": "Magellan, Ferdinand Magellan, Fernao Magalhaes",
  "n11149243": "Maginot, Andre Maginot",
  "n11149373": "Magritte, Rene Magritte",
  "n11149483": "Mahan, Alfred Thayer Mahan",
  "n11149630": "Mahler, Gustav Mahler",
  "n11149762": "Mailer, Norman Mailer",
  "n11149867": "Maillol, Aristide Maillol",
  "n11149995": "Maimonides, Moses Maimonides, Rabbi Moses Ben Maimon",
  "n11150224": "Maintenon, Marquise de Maintenon, Francoise d'Aubigne, Madame de Maintenon",
  "n11150471": "Maitland, Frederic William Maitland",
  "n11150634": "Major, John Major, John R. Major, John Roy Major",
  "n11150809": "Makarios III",
  "n11150973": "Malachi, Malachias",
  "n11151084": "Malamud, Bernard Malamud",
  "n11151189": "Malcolm X, Malcolm Little",
  "n11151322": "Malebranche, Nicolas de Malebranche",
  "n11151436": "Malevich, Kazimir Malevich, Kazimir Severinovich Malevich",
  "n11151579": "Malinowski, Bronislaw Malinowski, Bronislaw Kasper Malinowski",
  "n11151798": "Mallarme, Stephane Mallarme",
  "n11151932": "Mallon, Mary Mallon, Typhoid Mary",
  "n11152122": "Malone, Edmund Malone, Edmond Malone",
  "n11152331": "Malory, Thomas Malory, Sir Thomas Malory",
  "n11152544": "Malpighi, Marcello Malpighi",
  "n11152758": "Malraux, Andre Malraux",
  "n11152856": "Malthus, Thomas Malthus, Thomas Robert Malthus",
  "n11153094": "Mamet, David Mamet",
  "n11153200": "Mandela, Nelson Mandela, Nelson Rolihlahla Mandela",
  "n11153456": "Mandelbrot, Benoit Mandelbrot",
  "n11153615": "Mandelstam, Osip Mandelstam, Osip Emilevich Mandelstam, Mandelshtam",
  "n11153783": "Manes",
  "n11153887": "Manet, Edouard Manet",
  "n11154023": "Mann, Thomas Mann",
  "n11154174": "Mann, Horace Mann",
  "n11154355": "Mansart, Francois Mansart",
  "n11154489": "Mansfield, Katherine Mansfield, Kathleen Mansfield Beauchamp",
  "n11154646": "Manson, Sir Patrick Manson",
  "n11154857": "Mantegna, Andrea Mantegna",
  "n11155013": "Mantell, Gideon Algernon Mantell",
  "n11155196": "Mantle, Mickey Mantle, Mickey Charles Mantle",
  "n11155331": "Manzoni, Alessandro Manzoni",
  "n11155444": "Mao, Mao Zedong, Mao Tsetung",
  "n11155577": "Marat, Jean Paul Marat",
  "n11155814": "Marceau, Marcel Marceau",
  "n11155943": "Marciano, Rocco Marciano, Rocky Marciano",
  "n11156122": "Marconi, Guglielmo Marconi",
  "n11156333": "Marcuse, Herbert Marcuse",
  "n11156548": "Marie Antoinette",
  "n11156812": "Marini, Giambattista Marini, Marino, Giambattista Marino",
  "n11156943": "Mark, Saint Mark, St. Mark",
  "n11157140": "Markova, Dame Alicia Markova, Lilian Alicia Marks",
  "n11157275": "Markov, Andrei Markov, Markoff, Andre Markoff",
  "n11157422": "Marks, Simon Marks, First Baron Marks of Broughton",
  "n11157580": "Marley, Robert Nesta Marley, Bob Marley",
  "n11157719": "Marlowe, Christopher Marlowe",
  "n11157954": "Marquand, John Marquand, John Philip Marquand",
  "n11158156": "Marquette, Jacques Marquette, Pere Jacques Marquette",
  "n11158364": "Marquis, Don Marquis, Donald Robert Perry Marquis",
  "n11158533": "Marsh, Ngaio Marsh",
  "n11158651": "Marsh, Reginald Marsh",
  "n11158754": "Marshall, John Marshall",
  "n11158982": "Marshall, George Marshall, George Catlett Marshall",
  "n11159214": "Marshall, E. G. Marshall",
  "n11159318": "Marstan, John Marstan",
  "n11159418": "Marti, Jose Julian Marti",
  "n11159595": "Martial",
  "n11159698": "Martin, Dean Martin, Dino Paul Crocetti",
  "n11159819": "Martin, Mary Martin",
  "n11159920": "Martin, Steve Martin",
  "n11160055": "Martin, St. Martin",
  "n11160200": "Martin V, Oddone Colonna",
  "n11160361": "Marvell, Andrew Marvell",
  "n11160457": "Marx, Karl Marx",
  "n11160676": "Marx, Julius Marx, Groucho",
  "n11160861": "Marx, Leonard Marx, Chico",
  "n11161045": "Marx, Arthur Marx, Harpo",
  "n11161228": "Marx, Herbert Marx, Zeppo",
  "n11161412": "Mary, Virgin Mary, The Virgin, Blessed Virgin, Madonna",
  "n11161664": "Mary I, Mary Tudor, Bloody Mary",
  "n11162010": "Mary II",
  "n11162206": "Mary Queen of Scots, Mary Stuart",
  "n11162582": "Mary Magdalene, St. Mary Magdalene, Mary Magdalen, St. Mary Magdalen",
  "n11162793": "Masefield, John Masefield, John Edward Masefield",
  "n11162915": "Mason, A. E. W. Mason, Alfred Edward Woodley Mason",
  "n11163041": "Mason, James Mason, James Neville Mason",
  "n11163160": "Mason, George Mason",
  "n11163342": "Masoud, Ahmad Shah Masoud",
  "n11163582": "Massasoit",
  "n11163709": "Massenet, Jules Emile Frederic Massenet",
  "n11163859": "Massine, Leonide Fedorovitch Massine, Leonid Fyodorovich Myasin",
  "n11164058": "Masters, Edgar Lee Masters",
  "n11164163": "Mata Hari, Margarete Gertrud Zelle",
  "n11164332": "Mathias, Bob Mathias, Robert Bruce Mathias",
  "n11164505": "Matisse, Henri Matisse, Henri Emile Benoit Matisse",
  "n11164671": "Matthew, Saint Matthew, St. Matthew, Saint Matthew the Apostle, St. Matthew the Apostle, Levi",
  "n11164970": "Maugham, Somerset Maugham, W. Somerset Maugham, William Somerset Maugham",
  "n11165164": "Mauldin, Bill Mauldin, William Henry Mauldin",
  "n11165339": "Maupassant, Guy de Maupassant, Henri Rene Albert Guy de Maupassant",
  "n11165519": "Mauriac, Francois Mauriac, Francois Charles Mauriac",
  "n11165712": "Maurois, Andre Maurois, Emile Herzog",
  "n11165854": "Mauser, von Mauser, P. P. von Mauser, Peter Paul Mauser",
  "n11166060": "Maxim, Sir Hiram Stevens Maxim",
  "n11166251": "Maximian, Marcus Aurelius Valerius Maximianus, Herculius",
  "n11166504": "Maxwell, J. C. Maxwell, James Clerk Maxwell",
  "n11166732": "Mayakovski, Vladimir Vladimirovich Mayakovski",
  "n11166877": "Mayer, Louis B. Mayer, Louis Burt Mayer",
  "n11167088": "Mayer, Marie Goeppert Mayer",
  "n11167269": "Mays, Willie Mays, Willie Howard Mays Jr., Say Hey Kid",
  "n11167418": "Mazzini, Giuseppe Mazzini",
  "n11167595": "McCarthy, Joseph McCarthy, Joseph Raymond McCarthy",
  "n11167792": "McCarthy, Mary McCarthy, Mary Therese McCarthy",
  "n11167952": "McCartney, Paul McCartney, Sir James Paul McCartney",
  "n11168218": "McCauley, Mary McCauley, Mary Ludwig Hays McCauley, Molly Pitcher",
  "n11168513": "McCormick, John McCormick",
  "n11168645": "McCormick, Cyrus McCormick, Cyrus Hall McCormick",
  "n11168839": "McCullers, Carson McCullers, Carson Smith McCullers",
  "n11168974": "McGraw, John McGraw, John Joseph McGraw",
  "n11169135": "McGuffey, William Holmes McGuffey",
  "n11169294": "McKim, Charles Follen McKim",
  "n11169418": "McKinley, William McKinley, President McKinley",
  "n11169595": "McLuhan, Marshall McLuhan, Herbert Marshall McLuhan",
  "n11169764": "McMaster, John Bach McMaster",
  "n11169943": "McPherson, Aimee Semple McPherson",
  "n11170123": "Mead, George Herbert Mead",
  "n11170248": "Mead, Margaret Mead",
  "n11170438": "Meade, George Gordon Meade",
  "n11170604": "Meade, James Edward Meade",
  "n11170764": "Meany, George Meany",
  "n11170913": "Medawar, Peter Medawar, Sir Peter Brian Medawar",
  "n11171157": "Meiji Tenno, Mutsuhito",
  "n11171298": "Meir, Golda Meir",
  "n11171409": "Meissner, Fritz W. Meissner",
  "n11171513": "Meissner, Georg Meissner",
  "n11171614": "Meitner, Lise Meitner",
  "n11171851": "Melanchthon, Philipp Melanchthon, Philipp Schwarzerd",
  "n11172045": "Melba, Dame Nellie Melba, Helen Porter Mitchell",
  "n11172181": "Melchior",
  "n11172411": "Melchior, Lauritz Melchior, Lauritz Lebrecht Hommel Melchior",
  "n11172609": "Mellon, Andrew Mellon, Andrew W. Mellon, Andrew William Mellon",
  "n11172795": "Melville, Herman Melville",
  "n11172929": "Menander",
  "n11173031": "Mencken, H. L. Mencken, Henry Louis Mencken",
  "n11173199": "Mendel, Gregor Mendel, Johann Mendel",
  "n11173475": "Mendeleyev, Mendeleev, Dmitri Mendeleyev, Dmitri Mendeleev, Dmitri Ivanovich Mendeleyev, Dmitri Ivanovich Mendeleev",
  "n11173778": "Mendelsohn, Erich Mendelsohn",
  "n11173917": "Mendelssohn, Felix Mendelssohn, Jakob Ludwig Felix Mendelssohn-Bartholdy",
  "n11174119": "Meniere, Prosper Meniere",
  "n11174354": "Menninger, Charles Menninger, Charles Frederick Menninger",
  "n11174563": "Menninger, Karl Menninger, Karl Augustus Menninger",
  "n11174730": "Menninger, William Menninger, William Claire Menninger",
  "n11174901": "Menotti, Gian Carlo Menotti",
  "n11175040": "Menuhin, Yehudi Menuhin, Sir Yehudi Menuhin",
  "n11175243": "Mercator, Gerardus Mercator, Gerhard Kremer",
  "n11175445": "Mercer, John Mercer",
  "n11175605": "Merckx, Eddy Merckx",
  "n11175748": "Mercouri, Melina Mercouri, Anna Amalia Mercouri",
  "n11175875": "Meredith, George Meredith",
  "n11176005": "Meredith, James Meredith, James Howard Meredith",
  "n11176230": "Mergenthaler, Ottmar Mergenthaler",
  "n11176388": "Merlin",
  "n11176527": "Merman, Ethel Merman",
  "n11176669": "Merton, Robert Merton, Robert King Merton",
  "n11176797": "Merton, Thomas Merton",
  "n11176932": "Mesmer, Franz Anton Mesmer, Friedrich Anton Mesmer",
  "n11177151": "Metchnikoff, Elie Metchnikoff, Metchnikov, Elie Metchnikov, Ilya Ilich Metchnikov",
  "n11177370": "Methuselah",
  "n11177532": "Metternich, Klemens Metternich, Prince Klemens Wenzel Nepomuk Lothar von Metternich",
  "n11177695": "Meyerbeer, Giacomo Meyerbeer, Jakob Liebmann Beer",
  "n11177873": "Meyerhof, Otto Meyerhof, Otto Fritz Meyerhof",
  "n11178059": "Micah, Micheas",
  "n11178161": "Michelangelo, Michelangelo Buonarroti",
  "n11178393": "Michelson, A. A. Michelson, Albert Michelson, Albert Abraham Michelson",
  "n11178631": "Michener, James Michener, James Albert Michener",
  "n11178781": "Middleton, Thomas Middleton",
  "n11178922": "Mies Van Der Rohe, Ludwig Mies Van Der Rohe",
  "n11179124": "Milhaud, Darius Milhaud",
  "n11179287": "Mill, John Mill, John Stuart Mill",
  "n11179502": "Mill, James Mill",
  "n11179666": "Millais, Sir John Everett Millais",
  "n11179797": "Millay, Edna Millay, Edna Saint Vincent Millay",
  "n11179923": "Miller, Arthur Miller",
  "n11180029": "Miller, Henry Miller, Henry Valentine Miller",
  "n11180209": "Miller, Glenn Miller, Alton Glenn Miller",
  "n11180357": "Millet, Jean Francois Millet",
  "n11180476": "Millikan, Robert Andrews Millikan",
  "n11180643": "Mills, Robert Mills",
  "n11180812": "Milne, A. A. Milne, Alan Alexander Milne",
  "n11180952": "Miltiades",
  "n11181073": "Milton, John Milton",
  "n11181255": "Minkowski, Hermann Minkowski",
  "n11181438": "Minuit, Peter Minuit, Minnewit, Peter Minnewit",
  "n11181634": "Mirabeau, Comte de Mirabeau, Honore-Gabriel Victor Riqueti",
  "n11181835": "Miro, Joan Miro",
  "n11181937": "Mitchell, Arthur Mitchell",
  "n11182098": "Mitchell, John Mitchell",
  "n11182275": "Mitchell, Margaret Mitchell, Margaret Munnerlyn Mitchell",
  "n11182479": "Mitchell, Maria Mitchell",
  "n11182621": "Mitchell, William Mitchell, Billy Mitchell",
  "n11182825": "Mitchell, R. J. Mitchell, Reginald Joseph Mitchell",
  "n11182966": "Mitchum, Robert Mitchum",
  "n11183074": "Mitford, Nancy Mitford, Nancy Freeman Mitford",
  "n11183211": "Mitford, Jessica Mitford, Jessica Lucy Mitford",
  "n11183387": "Mithridates, Mithridates VI, Mithridates the Great",
  "n11183605": "Mitterrand, Francois Mitterrand, Francois Maurice Marie Mitterrand",
  "n11183791": "Mobius, August F. Mobius, August Ferdinand Mobius",
  "n11183955": "Modigliani, Amedeo Modigliano",
  "n11184092": "Mohammed, Mohammad, Muhammad, Mahomet, Mahound",
  "n11184273": "Mohammed Ali, Mehemet Ali, Muhammad Ali",
  "n11184531": "Mohorovicic, Andrija Mohorovicic",
  "n11184694": "Moliere, Jean-Baptiste Poquelin",
  "n11184825": "Molnar, Ferenc Molnar",
  "n11184927": "Molotov, Vyacheslav Mikhailovich Molotov",
  "n11185044": "Mommsen, Theodor Mommsen",
  "n11185175": "Mondrian, Piet Mondrian",
  "n11185392": "Monet, Claude Monet",
  "n11185500": "Monk, Thelonious Monk, Thelonious Sphere Monk",
  "n11185680": "Monnet, Jean Monnet",
  "n11185816": "Monod, Jacques Monod, Jacques Lucien Monod",
  "n11186042": "Monroe, James Monroe, President Monroe",
  "n11186207": "Monroe, Marilyn Monroe, Norma Jean Baker",
  "n11186356": "Montagu, Ashley Montagu",
  "n11186511": "Montaigne, Michel Montaigne, Michel Eyquem Montaigne",
  "n11186685": "Montespan, Marquise de Montespan, Francoise-Athenais de Rochechouart",
  "n11186911": "Montesquieu, Baron de la Brede et de Montesquieu, Charles Louis de Secondat",
  "n11187154": "Montessori, Maria Montesorri",
  "n11187364": "Monteverdi, Claudio Monteverdi",
  "n11187471": "Montez, Lola Montez, Marie Dolores Eliza Rosanna Gilbert",
  "n11187601": "Montezuma II",
  "n11187754": "Montfort, Simon de Montfort, Earl of Leicester",
  "n11187930": "Montgolfier, Josef Michel Montgolfier",
  "n11188123": "Montgolfier, Jacques Etienne Montgolfier",
  "n11188316": "Montgomery, Bernard Law Montgomery, Sir Bernard Law Montgomery, 1st Viscount Montgomery of Alamein",
  "n11188613": "Montgomery, L. M. Montgomery, Lucy Maud Montgomery",
  "n11188742": "Moody, Dwight Lyman Moody",
  "n11188852": "Moody, Helen Wills Moody, Helen Wills, Helen Newington Wills",
  "n11189054": "Moon, Sun Myung Moon",
  "n11189274": "Moore, Henry Moore, Henry Spencer Moore",
  "n11189432": "Moore, Marianne Moore, Marianne Craig Moore",
  "n11189579": "Moore, Thomas Moore",
  "n11189709": "Moore, G. E. Moore, George Edward Moore",
  "n11189829": "Moore, Dudley Moore, Dudley Stuart John Moore",
  "n11190024": "Moore, Douglas Moore",
  "n11190183": "More, Thomas More, Sir Thomas More",
  "n11190446": "Morgan, J. P. Morgan, John Pierpont Morgan",
  "n11190592": "Morgan, Daniel Morgan",
  "n11190774": "Morgan, Henry Morgan, Sir Henry Morgan",
  "n11190954": "Morgan, Thomas Hunt Morgan",
  "n11191113": "Morgan, Lewis Henry Morgan",
  "n11191251": "Morley, E. W. Morley, Edward Morley, Edward Williams Morley",
  "n11191475": "Mormon",
  "n11191653": "Morris, Gouverneur Morris",
  "n11191848": "Morris, Robert Morris",
  "n11192067": "Morris, William Morris",
  "n11192195": "Morris, Esther Morris, Esther Hobart McQuigg Slack Morris",
  "n11192349": "Morrison, Toni Morrison, Chloe Anthony Wofford",
  "n11192533": "Morrison, Jim Morrison, James Douglas Morrison",
  "n11192666": "Morse, Samuel Morse, Samuel F. B. Morse, Samuel Finley Breese Morse",
  "n11192901": "Mortimer, Roger de Mortimer",
  "n11193058": "Morton, Jelly Roll Morton, Ferdinand Joseph La Menthe Morton",
  "n11193250": "Mosander, Carl Gustaf Mossander",
  "n11193392": "Moses",
  "n11193645": "Moses, Grandma Moses, Anna Mary Robertson Moses",
  "n11193814": "Mossbauer, Rudolf Ludwig Mossbauer",
  "n11193928": "Motherwell, Robert Motherwell",
  "n11194062": "Mott, Lucretia Coffin Mott",
  "n11194205": "Moynihan, Daniel Patrick Moynihan",
  "n11194355": "Mozart, Wolfgang Amadeus Mozart",
  "n11194587": "Mubarak, Hosni Mubarak",
  "n11194749": "Muhammad, Elijah Muhammad",
  "n11194910": "Muir, John Muir",
  "n11195073": "Mullah Omar, Mullah Mohammed Omar",
  "n11195295": "Muller, Hermann Joseph Muller",
  "n11195452": "Muller, Max Muller, Friedrich Max Muller",
  "n11195619": "Muller, Johann Muller, Regiomontanus",
  "n11195771": "Muller, Johannes Peter Muller",
  "n11195913": "Muller, Karl Alex Muller",
  "n11196046": "Muller, Paul Hermann Muller",
  "n11196208": "Munch, Edvard Munch",
  "n11196305": "Munchhausen, Karl Friedrich Hieronymus von Munchhausen, Munchausen, Baron Munchausen",
  "n11196627": "Munro, H. H. Munro, Hector Hugh Munro, Saki",
  "n11196764": "Murdoch, Iris Murdoch, Dame Jean Iris Murdoch",
  "n11196934": "Murdoch, Rupert Murdoch, Keith Rupert Murdoch",
  "n11197099": "Murray, James Murray, James Augustus Murray, James Augustus Henry Murray, Sir James Murray, Sir James Augustus Murray, Sir James Augustus Henry Murray",
  "n11197417": "Murray, Gilbert Murray, George Gilbert Aime Murphy",
  "n11197633": "Murillo, Bartolome Esteban Murillo",
  "n11197743": "Murrow, Edward R. Murrow, Edward Roscoe Murrow",
  "n11197944": "Musial, Stan Musial, Stanley Frank Musial, Stan the Man",
  "n11198094": "Musset, Alfred de Musset, Louis Charles Alfred de Musset",
  "n11198253": "Mussolini, Benito Mussolini, Il Duce",
  "n11198375": "Mussorgsky, Moussorgsky, Modest Mussorgsky, Modest Moussorgsky, Modest Petrovich Mussorgsky, Modest Petrovich Moussorgsky",
  "n11198608": "Muybridge, Eadweard Muybridge, Edward James Muggeridge",
  "n11198861": "Myrdal, Gunnar Myrdal, Karl Gunnar Myrdal",
  "n11198981": "Nabokov, Vladimir Nabokov, Vladimir vladimirovich Nabokov",
  "n11199137": "Nahum",
  "n11199234": "Naismith, James Naismith",
  "n11199394": "Nanak, Guru Nanak",
  "n11199727": "Nansen, Fridtjof Nansen",
  "n11199943": "Naomi, Noemi",
  "n11200090": "Napier, John Napier",
  "n11200276": "Napoleon, Napoleon I, Napoleon Bonaparte, Bonaparte, Little Corporal",
  "n11200492": "Napoleon III, Emperor Napoleon III, Charles Louis Napoleon Bonaparte",
  "n11200686": "Nash, Ogden Nash",
  "n11200812": "Nasser, Gamal Abdel Nasser",
  "n11200949": "Nast, Thomas Nast",
  "n11201061": "Nation, Carry Nation, Carry Amelia Moore Nation",
  "n11201264": "Natta, Giulio Natta",
  "n11201386": "Sanchez, Ilich Sanchez, Ilich Ramirez Sanchez, Carlos, Carlos the Jackal, Salim, Andres Martinez, Taurus, Glen Gebhard, Hector Hevodidbon, Michael Assat",
  "n11201740": "Navratilova, Martina Navratilova",
  "n11201941": "Nazimova, Alla Nazimova",
  "n11202063": "Nebuchadnezzar, Nebuchadnezzar II, Nebuchadrezzar, Nebuchadrezzar II",
  "n11202322": "Nicholas V, Tomasso Parentucelli",
  "n11202477": "Nimrod",
  "n11202581": "Neel, Louis Eugene Felix Neel",
  "n11202722": "Nefertiti",
  "n11202833": "Nehru, Jawaharlal Nehru",
  "n11203059": "Nelson, Horatio Nelson, Viscount Nelson, Admiral Nelson, Lord Nelson",
  "n11203287": "Nernst, Walther Hermann Nernst",
  "n11203472": "Nero, Nero Claudius Caesar Drusus Germanicus, Lucius Domitius Ahenobarbus",
  "n11203795": "Neruda, Pablo Neruda, Reyes, Neftali Ricardo Reyes",
  "n11203920": "Nervi, Pier Luigi Nervi",
  "n11204069": "Nerva, Marcus Cocceius Nerva",
  "n11204276": "Nestor",
  "n11204409": "Nestorius",
  "n11204689": "Nevelson, Louise Nevelson",
  "n11204855": "Newcomb, Simon Newcomb",
  "n11204962": "Newman, John Henry Newman, Cardinal Newman",
  "n11205246": "Newman, Paul Newman, Paul Leonard Newman",
  "n11205375": "Newton, Isaac Newton, Sir Isaac Newton",
  "n11205647": "Ney, Michel Ney, Duc d'Elchingen",
  "n11205797": "Nicholas, Saint Nicholas, St. Nicholas",
  "n11205975": "Nicholas I, Czar Nicholas I",
  "n11206150": "Nicholas II",
  "n11206356": "Nicklaus, Jack Nicklaus, Jack William Nicklaus",
  "n11206544": "Nicolson, Harold Nicolson, Sir Harold George Nicolson",
  "n11206705": "Niebuhr, Barthold George Niebuhr",
  "n11206885": "Niebuhr, Reinhold Niebuhr",
  "n11207006": "Nielsen, Carl Nielsen, Carl August Nielsen",
  "n11207125": "Nietzsche, Friedrich Wilhelm Nietzsche",
  "n11207410": "Nightingale, Florence Nightingale, Lady with the Lamp",
  "n11207585": "Nijinsky, Vaslav Nijinsky, Waslaw Nijinsky",
  "n11207768": "Nilsson, Brigit Nilsson, Marta Brigit Nilsson",
  "n11207929": "Nimitz, Chester Nimitz, Chester William Nimitz, Admiral Nimitz",
  "n11208172": "Nixon, Richard Nixon, Richard M. Nixon, Richard Milhous Nixon, President Nixon",
  "n11208431": "Noah",
  "n11208688": "Nobel, Alfred Nobel, Alfred Bernhard Nobel",
  "n11208917": "Noether, Emmy Noether",
  "n11209019": "Noguchi, Hideyo Noguchi",
  "n11209201": "Noguchi, Isamu Noguchi",
  "n11209306": "Norman, Greg Norman, Gregory John Norman",
  "n11209428": "Norman, Jessye Norman",
  "n11209543": "Norris, Frank Norris, Benjamin Franklin Norris Jr.",
  "n11209675": "Norrish, Ronald George Wreyford Norrish",
  "n11209790": "North, Frederick North, Second Earl of Guilford",
  "n11209990": "Northrop, John Howard Northrop",
  "n11210105": "Nostradamus, Michel de Notredame",
  "n11210291": "Noyes, Alfred Noyes",
  "n11210383": "Nuffield, William Richard Morris, First Viscount Nuffield",
  "n11210610": "Nureyev, Rudolf Nureyev",
  "n11210806": "Oakley, Annie Oakley",
  "n11210963": "Oates, Joyce Carol Oates",
  "n11211071": "Oates, Titus Oates",
  "n11211236": "Obadiah, Abdias",
  "n11211322": "O'Brien, Edna O'Brien",
  "n11211419": "O'Casey, Sean O'Casey",
  "n11211517": "Occam, William of Occam, Ockham, William of Ockham",
  "n11211696": "Ochoa, Severo Ochoa",
  "n11211870": "Ochs, Adolph Simon Ochs",
  "n11211987": "O'Connor, Flannery O'Connor, Mary Flannery O'Connor",
  "n11212120": "Odets, Clifford Odets",
  "n11212226": "Odoacer, Odovacar, Odovakar",
  "n11212426": "Oersted, Hans Christian Oersted",
  "n11212534": "Offenbach, Jacques Offenbach",
  "n11212669": "O'Flaherty, Liam O'Flaherty",
  "n11212786": "Ogden, C. K. Ogden, Charles Kay Ogden",
  "n11212988": "O'Hara, John Henry O'Hara",
  "n11213094": "Ohm, Georg Simon Ohm",
  "n11213216": "O'Keeffe, Georgia Okeeffe",
  "n11213323": "Oken, Lorenz Oken, Okenfuss, Lorenz Okenfuss",
  "n11213552": "Olaf II, Olav II, Saint Olaf, Saint Olav, St. Olaf, St. Olav",
  "n11213726": "Oldenburg, Claes Oldenburg, Claes Thure Oldenburg",
  "n11213966": "Oldfield, Barney Oldfield, Berna Eli Oldfield",
  "n11214153": "Oliver, Joseph Oliver, King Oliver",
  "n11214320": "Olivier, Laurence Olivier, Sir Laurence Kerr Olivier, Baron Olivier of Birghton",
  "n11214513": "Olmsted, Frederick Law Olmsted",
  "n11214707": "Omar Khayyam",
  "n11214926": "Ondaatje, Michael Ondaatje, Philip Michael Ondaatje",
  "n11215070": "O'Neill, Eugene O'Neill, Eugene Gladstone O'Neill",
  "n11215205": "Ono, Yoko Ono",
  "n11215384": "Onsager, Lars Onsager",
  "n11215541": "Oort, Jan Hendrix Oort",
  "n11215724": "Opel, Wilhelm von Opel",
  "n11215909": "Oppenheimer, Robert Oppenheimer",
  "n11216100": "Orbison, Roy Orbison",
  "n11216264": "Orczy, Baroness Emmusca Orczy",
  "n11216386": "Orff, Carl Orff",
  "n11216562": "Origen",
  "n11216797": "Ormandy, Eugene Ormandy",
  "n11216922": "Orozco, Jose Orozco, Jose Clemente Orozco",
  "n11217072": "Orr, Bobby Orr, Robert Orr",
  "n11217182": "Ortega, Daniel Ortega, Daniel Ortega Saavedra",
  "n11217312": "Ortega y Gasset, Jose Ortega y Gasset",
  "n11217479": "Orwell, George Orwell, Eric Blair, Eric Arthur Blair",
  "n11217668": "Osborne, John Osborne, John James Osborne",
  "n11217789": "Osman I, Othman I",
  "n11218054": "Ostwald, Wilhelm Ostwald",
  "n11218153": "Oswald, Lee Harvey Oswald",
  "n11218290": "Otis, Elisha Graves Otis",
  "n11218473": "O'Toole, Peter O'Toole, Peter Seamus O'Toole",
  "n11218625": "Otto I, Otho I, Otto the Great",
  "n11218776": "Ovid, Publius Ovidius Naso",
  "n11218938": "Owen, Sir Richard Owen",
  "n11219121": "Owen, Robert Owen",
  "n11219293": "Owens, Jesse Owens, James Cleveland Owens",
  "n11219502": "Ozawa, Seiji Ozawa",
  "n11219635": "Paderewski, Ignace Paderewski, Ignace Jan Paderewski",
  "n11219851": "Paganini, Niccolo Paganini",
  "n11219992": "Page, Thomas Nelson Page",
  "n11220149": "Page, Sir Frederick Handley Page",
  "n11220319": "Paget, Sir James Paget",
  "n11220461": "Pahlavi, Mohammed Reza Pahlavi, Shah Pahlavi, Pahlevi, Mohammed Reza Pahlevi",
  "n11220664": "Paige, Satchel Paige, Leroy Robert Paige",
  "n11220836": "Paine, Tom Paine, Thomas Paine",
  "n11221107": "Paine, Robert Treat Paine",
  "n11221268": "Palestrina, Giovanni Pierluigi da Palestrina",
  "n11221389": "Palgrave, Francis Turner Palgrave",
  "n11221495": "Palladio, Andrea Palladio",
  "n11221650": "Palmer, Arnold Palmer, Arnold Daniel Palmer",
  "n11221778": "Panini",
  "n11221956": "Panofsky, Erwin Panofsky",
  "n11222054": "Paracelsus, Philippus Aureolus Paracelsus, Theophrastus Philippus Aureolus Bombastus von Hohenheim",
  "n11222457": "Pareto, Vilfredo Pareto",
  "n11222655": "Park, Mungo Park",
  "n11222759": "Parker, Dorothy Parker, Dorothy Rothschild Parker",
  "n11222914": "Parker, Charlie Parker, Yardbird Parker, Bird Parker, Charles Christopher Parker",
  "n11223119": "Parkinson, C. Northcote Parkinson, Cyril Northcote Parkinson",
  "n11223294": "Parkinson, James Parkinson",
  "n11223396": "Parks, Rosa Parks",
  "n11223642": "Parmenides",
  "n11223929": "Parnell, Charles Stewart Parnell",
  "n11224046": "Parr, Catherine Parr",
  "n11224173": "Parrish, Maxfield Parrish, Maxfield Frederick Parrish",
  "n11224309": "Parsons, Talcott Parsons",
  "n11224419": "Pascal, Blaise Pascal",
  "n11224654": "Pasternak, Boris Pasternak, Boris Leonidovich Pasternak",
  "n11224877": "Pasteur, Louis Pasteur",
  "n11225165": "Paterson, William Patterson",
  "n11225350": "Paton, Alan Paton, Alan Stewart Paton",
  "n11225469": "Patrick, Saint Patrick, St. Patrick",
  "n11225661": "Paul, Saint Paul, St. Paul, Apostle Paul, Paul the Apostle, Apostle of the Gentiles, Saul, Saul of Tarsus",
  "n11226126": "Paul III, Alessandro Farnese",
  "n11226427": "Paul VI, Giovanni Battista Montini",
  "n11226614": "Paul, Alice Paul",
  "n11226713": "Pauli, Wolfgang Pauli",
  "n11226933": "Pauling, Linus Pauling, Linus Carl Pauling",
  "n11227101": "Pavarotti, Luciano Pavarotti",
  "n11227206": "Pavlov, Ivan Pavlov, Ivan Petrovich Pavlov",
  "n11227400": "Pavlova, Anna Pavlova",
  "n11227499": "Paxton, Joseph Paxton, Sir Joseph Paxton",
  "n11227618": "Peabody, Elizabeth Peabody, Elizabeth Palmer Peabody",
  "n11227796": "Peary, Robert Peary, Robert E. Peary, Robert Edwin Peary",
  "n11228039": "Peel, Robert Peel, Sir Robert Peel",
  "n11228153": "Pei, I. M. Pei, Ieoh Ming Pei",
  "n11228298": "Peirce, Charles Peirce, Charles Sanders Peirce",
  "n11228486": "Peirce, Benjamin Peirce",
  "n11228691": "Pelagius",
  "n11228956": "Penn, William Penn",
  "n11229095": "Pepin, Pepin III, Pepin the Short",
  "n11229323": "Pepys, Samuel Pepys",
  "n11229499": "Percy, Sir Henry Percy, Hotspur, Harry Hotspur",
  "n11229662": "Percy, Walker Percy",
  "n11229801": "Pericles",
  "n11230021": "Peron, Juan Domingo Peron",
  "n11230158": "Perry, Oliver Hazard Perry, Commodore Perry",
  "n11230402": "Perry, Matthew Calbraith Perry",
  "n11230669": "Perry, Ralph Barton Perry",
  "n11230780": "Pershing, John Joseph Pershing, Black Jack Pershing",
  "n11230977": "Perutz, Max Perutz, Max Ferdinand Perutz",
  "n11231157": "Peter, Simon Peter, Saint Peter, St. Peter, Saint Peter the Apostle, St. Peter the Apostle",
  "n11231433": "Peter I, Czar Peter I, Peter the Great",
  "n11231683": "Petrarch, Petrarca, Francesco Petrarca",
  "n11231821": "Petronius, Gaius Petronius, Petronius Arbiter",
  "n11231943": "Phidias, Pheidias",
  "n11232050": "Philemon",
  "n11232229": "Philemon",
  "n11232475": "Philip, Prince Philip, Duke of Edinburgh",
  "n11232615": "Philip II, Philip II of Spain",
  "n11232837": "Philip II, Philip II of Macedon",
  "n11232989": "Philip II, Philip Augustus",
  "n11233211": "Philip V",
  "n11233383": "Philip VI, Philip of Valois",
  "n11233595": "Phintias, Pythias",
  "n11233934": "Photius",
  "n11234152": "Piaf, Edith Piaf, Edith Giovanna Gassion, Little Sparrow",
  "n11234292": "Piaget, Jean Piaget",
  "n11234472": "Pickett, George Edward Pickett",
  "n11234643": "Pickford, Mary Pickford, Gladys Smith",
  "n11234813": "Pierce, Franklin Pierce, President Pierce",
  "n11234951": "Picasso, Pablo Picasso",
  "n11235112": "Pilate, Pontius Pilate",
  "n11235263": "Pincus, Gregory Pincus, Gregory Goodwin Pincus",
  "n11235510": "Pindar",
  "n11235619": "Pinter, Harold Pinter",
  "n11235787": "Pirandello, Luigi Pirandello",
  "n11235926": "Piston, Walter Piston",
  "n11236043": "Pitman, Sir Isaac Pitman",
  "n11236188": "Pitot, Henri Pitot",
  "n11236317": "Pitt, William Pitt, First Earl of Chatham, Pitt the Elder",
  "n11236497": "Pitt, William Pitt, Second Earl of Chatham, Pitt the Younger",
  "n11236663": "Pitt, George Pitt, George Dibdin Pitt, George Dibdin-Pitt",
  "n11236852": "Pius II, Aeneas Silvius, Enea Silvio Piccolomini",
  "n11237075": "Pius V, Antonio Ghislieri",
  "n11237275": "Pius VI, Giovanni Angelo Braschi, Giannangelo Braschi",
  "n11237550": "Pius VII, Barnaba Chiaramonti, Luigi Barnaba Gregorio Chiaramonti",
  "n11237868": "Pius IX, Giovanni Mastai-Ferretti, Giovanni Maria Mastai-Ferretti",
  "n11238092": "Pius X, Giuseppe Sarto, Giuseppe Melchiorre Sarto",
  "n11238303": "Pius XI, Achille Ratti, Ambrogio Damiano Achille Ratti",
  "n11238511": "Pius XII, Eugenio Pacelli",
  "n11238726": "Pizarro, Francisco Pizarro",
  "n11238906": "Planck, Max Planck, Max Karl Ernst Ludwig Planck",
  "n11239143": "Plath, Sylvia Plath",
  "n11239271": "Plato",
  "n11239443": "Plautus, Titus Maccius Plautus",
  "n11239567": "Pliny, Pliny the Elder, Gaius Plinius Secundus",
  "n11239765": "Pliny, Pliny the Younger, Gaius Plinius Caecilius Secundus",
  "n11239989": "Plotinus",
  "n11240139": "Plutarch",
  "n11240249": "Pocahontas, Matoaka, Rebecca Rolfe",
  "n11240480": "Poe, Edgar Allan Poe",
  "n11240609": "Poitier, Sidney Poitier",
  "n11240733": "Polk, James Polk, James K. Polk, James Knox Polk, President Polk",
  "n11240996": "Pollack, Sydney Pollack",
  "n11241106": "Pollock, Jackson Pollock",
  "n11241300": "Polo, Marco Polo",
  "n11241455": "Polycarp, Saint Polycarp, St. Polycarp",
  "n11241655": "Pompadour, Marquise de Pompadour, Jeanne Antoinette Poisson",
  "n11241854": "Pompey, Gnaeus Pompeius Magnus, Pompey the Great",
  "n11242077": "Ponce de Leon, Juan Ponce de Leon",
  "n11242321": "Pons, Lily Pons, Alice-Josephine Pons",
  "n11242469": "Ponselle, Rosa Ponselle, Rosa Melba Ponselle",
  "n11242596": "Pontiac",
  "n11242743": "Pope, Alexander Pope",
  "n11242849": "Popper, Karl Popper, Sir Karl Raimund Popper",
  "n11243102": "Porter, William Sydney Porter, O. Henry",
  "n11243268": "Porter, Cole Porter, Cole Albert Porter",
  "n11243424": "Porter, Katherine Anne Porter",
  "n11243562": "Post, C. W. Post, Charles William Post",
  "n11243720": "Post, Emily Post, Emily Price Post",
  "n11243907": "Post, Wiley Post",
  "n11244061": "Potemkin, Potyokin, Grigori Potemkin, Grigori Potyokin, Grigori Aleksandrovich Potemkin",
  "n11244419": "Poulenc, Francis Poulenc",
  "n11244550": "Pound, Ezra Pound, Ezra Loomis Pound",
  "n11244765": "Poussin, Nicolas Poussin",
  "n11244887": "Powell, Cecil Frank Powell",
  "n11245110": "Powell, Colin Powell, Colin luther Powell",
  "n11245382": "Powhatan, Wahunsonacock",
  "n11245590": "Powys, John Cowper Powys",
  "n11245744": "Powys, Theodore Francis Powys",
  "n11245902": "Powys, Llewelyn Powys",
  "n11246040": "Presley, Elvis Presley, Elvis Aron Presley",
  "n11246255": "Priam",
  "n11246408": "Price, Leontyne Price, Mary Leontyne Price",
  "n11246542": "Priestley, Joseph Priestley",
  "n11246718": "Prokhorov, Aleksandr Prokhorov, Aleksandr Mikjailovich Prokhorov",
  "n11247002": "Prokofiev, Sergei Sergeyevich Prokofiev",
  "n11247155": "Proudhon, Pierre Joseph Proudhon",
  "n11247298": "Proust, Marcel Proust",
  "n11247413": "Ptolemy, Claudius Ptolemaeus",
  "n11247644": "Ptolemy I",
  "n11247897": "Ptolemy II",
  "n11248077": "Puccini, Giacomo Puccini",
  "n11248232": "Pugin, Augustus Welby Northmore Pugin",
  "n11248426": "Pulitzer, Joseph Pulitzer",
  "n11248599": "Purcell, Henry Purcell",
  "n11248777": "Purkinje, Jan Evangelista Purkinje, Johannes Evangelista Purkinje",
  "n11248997": "Pusey, Edward Pusey, Edward Bouverie Pusey",
  "n11249191": "Pushkin, Alexander Pushkin, Aleksandr Sergeyevich Pushkin",
  "n11249322": "Putin, Vladimir Putin, Vladimir Vladimirovich Putin",
  "n11249559": "Pyle, Howard Pyle",
  "n11249712": "Pynchon, Thomas Pynchon",
  "n11249890": "Pyrrhus",
  "n11250056": "Pythagoras",
  "n11250287": "Qaddafi, Qadhafi, Khadafy, Gaddafi, Muammar al-Qaddafi, Muammar el-Qaddafi",
  "n11250581": "Qin Shi Huang Ti, Ch'in Shih Huang Ti",
  "n11250833": "Quincy, Josiah Quincy",
  "n11250991": "Quine, W. V. Quine, Willard Van Orman Quine",
  "n11251225": "Rabelais, Francois Rabelais",
  "n11251384": "Rachel",
  "n11251531": "Rachmaninoff, Sergei Rachmaninoff, Sergei Vasilievich Rachmaninoff, Rachmaninov, Sergei Rachmaninov, Sergei Vasilievich Rachmaninov",
  "n11251788": "Racine, Jean Racine, Jean Baptiste Racine",
  "n11251995": "Radhakrishnan, Sarvepalli Radhakrishnan, Sir Sarvepalli Radhakrishnan",
  "n11252222": "Raffles, Sir Thomas Raffles, Sir Thomas Stamford Raffles",
  "n11252392": "Rain-in-the-Face",
  "n11252627": "Raleigh, Walter Raleigh, Sir Walter Raleigh, Ralegh, Walter Ralegh, Sir Walter Ralegh",
  "n11252915": "Rameau, Jean-Philippe Rameau",
  "n11253097": "Rameses, Ramesses, Ramses",
  "n11253248": "Rameses II, Ramesses II, Ramses II, Rameses the Great, Ramesses the Great, Ramses the Great",
  "n11253455": "Ramon y Cajal, Santiago Ramon y Cajal",
  "n11253630": "Rand, Ayn Rand",
  "n11253802": "Rankin, Jeannette Rankin",
  "n11254028": "Raphael, Raffaello Santi, Raffaello Sanzio",
  "n11254213": "Rask, Rasmus Christian Rask",
  "n11254393": "Rasmussen, Kund Johan Victor Rasmussen",
  "n11254683": "Rasputin, Grigori Efimovich Rasputin",
  "n11254950": "Rattigan, Terence Rattigan, Sir Terence Mervyn Rattigan",
  "n11255085": "Ravel, Maurice Ravel",
  "n11255211": "Rayleigh, Third Baron Rayleigh, Lord Rayleigh, John William Strutt",
  "n11255460": "Reagan, Ronald Reagan, Ronald Wilson Reagan, President Reagan",
  "n11255619": "Reaumur, Rene Antoine Ferchault de Reaumur",
  "n11255775": "Rebecca, Rebekah",
  "n11255917": "Red Cloud",
  "n11256125": "Redford, Robert Redford, Charles Robert Redford",
  "n11256335": "Reed, Walter Reed",
  "n11256494": "Reed, John Reed",
  "n11256765": "Rehnquist, William Rehnquist, William Hubbs Rehnquist",
  "n11257031": "Reich, Steve Reich, Stephen Michael Reich",
  "n11257159": "Reich, Wilhelm Reich",
  "n11257395": "Reichstein, Tadeus Reichstein",
  "n11257547": "Reid, Thomas Reid",
  "n11257697": "Reiter, Hans Conrad Julius Reiter",
  "n11257932": "Rembrandt, Rembrandt van Rijn, Rembrandt van Ryn, Rembrandt Harmensz van Rijn",
  "n11258096": "Renoir, Pierre Auguste Renoir",
  "n11258214": "Respighi, Ottorino Respighi",
  "n11258353": "Reuben",
  "n11258501": "Revere, Paul Revere",
  "n11258769": "Reynolds, Sir Joshua Reynolds",
  "n11258924": "Rhine, J. B. Rhine, Joseph Banks Rhine",
  "n11259054": "Rhodes, Cecil Rhodes, Cecil J. Rhodes, Cecil John Rhodes",
  "n11259457": "Ricardo, David Ricardo",
  "n11259635": "Rice, Elmer Rice, Elmer Leopold Rice, Elmer Reizenstein",
  "n11259777": "Rice, Sir Tim Rice, Timothy Miles Bindon Rice",
  "n11259950": "Richard I, Richard Coeur de Lion, Richard the Lionheart, Richard the Lion-Hearted",
  "n11260325": "Richard II",
  "n11260622": "Richard III",
  "n11260945": "Richards, I. A. Richards, Ivor Armstrong Richards",
  "n11261184": "Richardson, Ralph Richardson, Sir Ralph David Richardson",
  "n11261364": "Richardson, Henry Hobson Richardson",
  "n11261483": "Richelieu, Duc de Richelieu, Armand Jean du Plessis, Cardinal Richelieu",
  "n11261698": "Richler, Mordecai Richler",
  "n11261804": "Rickenbacker, Eddie Rickenbacker, Edward Vernon Rickenbacker",
  "n11261986": "Rickover, Hyman Rickover, Hyman George Rickover",
  "n11262168": "Riemann, Bernhard Riemann, Georg Friedrich Bernhard Riemann",
  "n11262340": "Riesman, David Riesman, David Riesman Jr.",
  "n11262468": "Riley, James Whitcomb Riley",
  "n11262574": "Rilke, Rainer Maria Rilke",
  "n11262765": "Rimbaud, Arthur Rimbaud, Jean Nicholas Arthur Rimbaud",
  "n11262929": "Rimsky-Korsakov, Nikolai Andreyevich Rimsky-Korsakov, Rimski-Korsakov, Nikolai Andreyevich Rimski-Korsakov",
  "n11263180": "Ringling, Charles Ringling",
  "n11263337": "Rittenhouse, David Rittenhouse",
  "n11263558": "Ritz, Cesar Ritz",
  "n11263687": "Rivera, Diego Rivera",
  "n11263803": "Robbins, Jerome Robbins",
  "n11264002": "Robert, Henry M. Robert, Henry Martyn Robert",
  "n11264193": "Roberts, Bartholomew Roberts",
  "n11264343": "Roberts, Kenneth Roberts",
  "n11264508": "Roberts, Oral Roberts",
  "n11264614": "Roberts, Richard J. Roberts, Richard John Roberts",
  "n11264828": "Robertson, Oscar Robertson, Oscar Palmer Robertson",
  "n11264973": "Robeson, Paul Robeson, Paul Bustill Robeson",
  "n11265183": "Robespierre, Maxmillien Marie Isidore de Robespierre",
  "n11265416": "Robinson, Edward G. Robinson, Edward Goldenberg Robinson",
  "n11265591": "Robinson, Edwin Arlington Robinson",
  "n11265731": "Robinson, Jackie Robinson, Jack Roosevelt Robinson",
  "n11265914": "Robinson, James Harvey Robinson",
  "n11266117": "Robinson, Lennox Robinson, Esme Stuart Lennox Robinson",
  "n11266279": "Robinson, Ray Robinson, Sugar Ray Robinson, Walker Smith",
  "n11266524": "Robinson, Robert Robinson, Sir Robert Robinson",
  "n11266703": "Rochambeau, Comte de Rochambeau, Jean Baptiste Donatien de Vimeur",
  "n11266920": "Rock, John Rock",
  "n11267113": "Rockefeller, John D. Rockefeller, John Davison Rockefeller",
  "n11267343": "Rockingham, Second Marquis of Rockingham, Charles Watson-Wentworth",
  "n11267568": "Rockwell, Norman Rockwell",
  "n11267745": "Rodgers, Richard Rodgers",
  "n11267949": "Rodin, Auguste Rodin, Francois Auguste Rene Rodin",
  "n11268118": "Roebling, John Roebling, John Augustus Roebling",
  "n11268326": "Roentgen, Wilhelm Konrad Roentgen, Rontgen, Wilhelm Konrad Rontgen",
  "n11268523": "Rogers, Carl Rogers",
  "n11268667": "Rogers, Ginger Rogers, Virginia McMath, Virginia Katherine McMath",
  "n11268883": "Rogers, Will Rogers, William Penn Adair Rogers",
  "n11269085": "Roget, Peter Mark Roget",
  "n11269236": "Rollo, Rolf, Hrolf",
  "n11269367": "Romberg, Sigmund Romberg",
  "n11269515": "Rommel, Erwin Rommel, Desert Fox",
  "n11269697": "Roosevelt, Theodore Roosevelt, President Roosevelt, President Theodore Roosevelt",
  "n11270023": "Roosevelt, Franklin Roosevelt, Franklin Delano Roosevelt, F. D. Roosevelt, President Roosevelt, President Franklin Roosevelt, FDR",
  "n11270380": "Roosevelt, Eleanor Roosevelt, Anna Eleanor Roosevelt",
  "n11270577": "Ross, Betsy Ross, Betsy Griscom Ross",
  "n11270772": "Ross, Nellie Ross, Nellie Tayloe Ross",
  "n11270948": "Ross, Sir Ronald Ross",
  "n11271094": "Ross, James Clark Ross, Sir James Clark Ross",
  "n11271349": "Ross, John Ross, Sir John Ross",
  "n11271563": "Rossetti, Dante Gabriel Rossetti",
  "n11271720": "Rossini, Giloacchino Antonio Rossini",
  "n11271859": "Rostand, Edmond Rostand",
  "n11272030": "Roth, Philip Roth, Philip Milton Roth",
  "n11272198": "Rothko, Mark Rothko",
  "n11272415": "Rothschild",
  "n11272523": "Rous, Peyton Rous, Francis Peyton Rous",
  "n11272689": "Rousseau, Jean-Jacques Rousseau",
  "n11272972": "Rousseau, Henri Rousseau, Le Douanier Rousseau",
  "n11273104": "Rubens, Peter Paul Rubens, Sir Peter Paul Rubens",
  "n11273286": "Rubinstein, Anton Rubenstein, Anton Gregor Rubinstein, Anton Grigorevich Rubinstein",
  "n11273479": "Rubinstein, Arthur Rubinstein, Artur Rubinstein",
  "n11273679": "Rundstedt, von Rundstedt, Karl Rudolf Gerd von Rundstedt",
  "n11273907": "Runyon, Damon Runyon, Alfred Damon Runyon",
  "n11274103": "Rupert, Prince Rupert",
  "n11274269": "Rush, Benjamin Rush",
  "n11274454": "Rushdie, Salman Rushdie, Ahmed Salman Rushdie",
  "n11274714": "Ruskin, John Ruskin",
  "n11274812": "Russell, Bertrand Russell, Bertrand Arthur William Russell, Earl Russell",
  "n11275035": "Russell, George William Russell, A.E.",
  "n11275170": "Russell, Henry Russell, Henry Norris Russell",
  "n11275344": "Russell, Lillian Russell",
  "n11275495": "Russell, Bill Russell, William Felton Russell",
  "n11275636": "Russell, Ken Russell, Henry Kenneth Alfred Russell",
  "n11275772": "Russell, Charles Taze Russell",
  "n11275952": "Ruth",
  "n11276100": "Ruth, Babe Ruth, George Herman Ruth, Sultan of Swat",
  "n11276285": "Rutherford, Ernest Rutherford, First Baron Rutherford, First Baron Rutherford of Nelson",
  "n11276549": "Rutherford, Daniel Rutherford",
  "n11276676": "Rutledge, John Rutledge",
  "n11276971": "Saarinen, Eero Saarinen",
  "n11277096": "Saarinen, Eliel Saarinen",
  "n11277279": "Sabin, Albert Sabin, Albert Bruce Sabin",
  "n11277500": "Sacagawea, Sacajawea",
  "n11277680": "Sacco, Nicola Sacco",
  "n11277900": "Sadat, Anwar Sadat, Anwar el-Sadat",
  "n11278120": "Sade, de Sade, Comte Donatien Alphonse Francois de Sade, Marquis de Sade",
  "n11278351": "Saint-Saens, Charles Camille Saint-Saens",
  "n11278498": "Sakharov, Andrei Sakharov, Andrei Dimitrievich Sakharov",
  "n11278756": "Saladin, Salah-ad-Din Yusuf ibn-Ayyub",
  "n11278980": "Salinger, J. D. Salinger, Jerome David Salinger",
  "n11279109": "Salk, Jonas Salk, Jonas Edward Salk",
  "n11279300": "Salome",
  "n11279433": "Salomon, Haym Salomon",
  "n11279648": "Samson",
  "n11279874": "Samuel",
  "n11280013": "Sand, George Sand, Amandine Aurore Lucie Dupin, Baroness Dudevant",
  "n11280213": "Sandburg, Carl Sandburg",
  "n11280405": "Sanger, Margaret Sanger, Margaret Higgins Sanger",
  "n11280653": "Sanger, Frederick Sanger, Fred Sanger",
  "n11280905": "Santa Anna, Santa Ana, Antonio Lopez de Santa Anna, Antonio Lopez de Santa Ana",
  "n11281174": "Sapir, Edward Sapir",
  "n11281345": "Sappho",
  "n11281555": "Sarah",
  "n11281682": "Sarazen, Gene Sarazen",
  "n11281837": "Sargent, John Singer Sargent",
  "n11281995": "Sarnoff, David Sarnoff",
  "n11282154": "Saroyan, William Saroyan",
  "n11282286": "Sartre, Jean-Paul Sartre",
  "n11282434": "Satie, Erik Satie, Erik Alfred Leslie Satie",
  "n11282613": "Saul",
  "n11282802": "Savonarola, Girolamo Savonarola",
  "n11283158": "Sax, Adolphe Sax",
  "n11283300": "Saxe, Hermann Maurice Saxe, comte de Saxe, Marshal Saxe",
  "n11283517": "Saxo Grammaticus",
  "n11283682": "Sayers, Dorothy Sayers, Dorothy L. Sayers, Dorothy Leigh Sayers",
  "n11283843": "Scheele, Karl Scheele, Karl Wilhelm Scheele",
  "n11284024": "Schiaparelli, Elsa Schiaparelli",
  "n11284216": "Schiaparelli, Giovanni Virginio Schiaparelli",
  "n11284409": "Schiller, Johann Christoph Friedrich von Schiller",
  "n11284541": "Schleiden, Matthias Schleiden, M. J. Schleiden",
  "n11284742": "Schlesinger, Arthur Schlesinger, Arthur Meier Schlesinger",
  "n11284884": "Schlesinger, Arthur Schlesinger, Arthur Schlesinger Jr., Arthur Meier Schlesinger Jr.",
  "n11285091": "Schliemann, Heinrich Schliemann",
  "n11285282": "Schmidt, Helmut Schmidt, Helmut Heinrich Waldemar Schmidt",
  "n11285456": "Schnabel, Artur Schnabel",
  "n11285692": "Schonbein, Christian Schonbein, Christian Friedrich Schonbein",
  "n11285902": "Schonberg, Arnold Schonberg, Schoenberg, Arnold Schoenberg",
  "n11286117": "Schoolcraft, Henry Rowe Schoolcraft",
  "n11286354": "Schopenhauer, Arthur Schopenhauer",
  "n11286476": "Schrodinger, Erwin Schrodinger",
  "n11286618": "Schubert, Franz Schubert, Franz Peter Schubert, Franz Seraph Peter Schubert",
  "n11286820": "Schulz, Charles Schulz, Charles M. Schulz, Charles Munroe Schulz",
  "n11287016": "Schumann, Robert Schumann, Robert Alexander Schumann",
  "n11287186": "Schumann, Clara Josephine Schumann",
  "n11287437": "Schumann-Heink, Ernestine Schumann-Heink",
  "n11287570": "Schumpeter, Joseph Schumpeter, Joseph Alois Schumpeter",
  "n11287734": "Schwann, Theodor Schwann",
  "n11287964": "Schweitzer, Albert Schweitzer",
  "n11288216": "Scipio, Scipio Africanus, Scipio Africanus Major, Publius Cornelius Scipio, Publius Cornelius Scipio Africanus Major, Scipio the Elder",
  "n11288528": "Scopes, John Scopes, John Thomas Scopes",
  "n11288818": "Scorsese, Martin Scorsese",
  "n11288930": "Scott, Dred Scott",
  "n11289161": "Scott, Walter Scott, Sir Walter Scott",
  "n11289307": "Scott, Winfield Scott",
  "n11289491": "Scott, Robert Scott, Robert Falcon Scott",
  "n11289709": "Scott, George C. Scott",
  "n11289830": "Scriabin, Aleksandr Scriabin, Aleksandr Nikolayevich Scriabin",
  "n11289999": "Scribe, Augustin Eugene Scribe",
  "n11290107": "Scripps, James Edmund Scripps",
  "n11290272": "Scripps, Edward Wyllis Scripps",
  "n11290477": "Seaborg, Glenn T. Seaborg, Glenn Theodore Seaborg",
  "n11290653": "Seaman, Elizabeth Seaman, Elizabeth Cochrane Seaman, Nellie Bly",
  "n11290864": "Seeger, Alan Seeger",
  "n11290984": "Seeger, Pete Seeger, Peter Seeger",
  "n11291179": "Segal, George Segal",
  "n11291284": "Segovia, Andres Segovia",
  "n11291451": "Seles, Monica Seles",
  "n11291575": "Seleucus, Seleucus I, Seleucus I Nicator",
  "n11291824": "Selkirk, Selcraig, Alexander Selkirk, Alexander Selcraig",
  "n11292105": "Sellers, Peter Sellers",
  "n11292207": "Selznick, David O. Selznick, David Oliver Selznick",
  "n11292391": "Seneca, Lucius Annaeus Seneca",
  "n11292662": "Senefelder, Alois Senefelder, Aloys Senefelder",
  "n11292809": "Sennacherib",
  "n11293008": "Sennett, Mack Sennett",
  "n11293157": "Sequoya, Sequoyah, George Guess",
  "n11293334": "Serkin, Rudolf Serkin",
  "n11293470": "Serra, Junipero Serra, Miguel Jose Serra",
  "n11293636": "Service, Robert William Service",
  "n11293805": "Sessions, Roger Sessions, Roger Huntington Sessions",
  "n11293972": "Seton, Elizabeth Seton, Saint Elizabeth Ann Bayley Seton, Mother Seton",
  "n11294202": "Seurat, Georges Seurat, Georges Pierre Seurat",
  "n11294349": "Seward, William Henry Seward",
  "n11294582": "Sexton, Anne Sexton",
  "n11294680": "Seymour, Jane Seymour",
  "n11294834": "Shah Jahan",
  "n11295033": "Shahn, Ben Shahn, Benjamin Shahn",
  "n11295196": "Shakespeare, William Shakespeare, Shakspere, William Shakspere, Bard of Avon",
  "n11295464": "Shankar, Ravi Shankar",
  "n11295619": "Shannon, Claude Shannon, Claude E. Shannon, Claude Elwood Shannon",
  "n11295828": "Shapley, Harlow Shapley",
  "n11295936": "Shaw, G. B. Shaw, George Bernard Shaw",
  "n11296139": "Shaw, Anna Howard Shaw",
  "n11296279": "Shaw, Henry Wheeler Shaw, Josh Billings",
  "n11296429": "Shaw, Artie Shaw, Arthur Jacob Arshawsky",
  "n11296602": "Shawn, Ted Shawn",
  "n11296775": "Shearer, Moira Shearer",
  "n11296914": "Shelley, Percy Bysshe Shelley",
  "n11297032": "Shelley, Mary Shelley, Mary Wollstonecraft Shelley, Mary Godwin Wollstonecraft Shelley",
  "n11297263": "Shepard, Alan Shepard, Alan Bartlett Shepard Jr.",
  "n11297457": "Shepard, Sam Shepard",
  "n11297595": "Sheridan, Richard Brinsley Sheridan",
  "n11297756": "Sherman, Roger Sherman",
  "n11297983": "Sherman, William Tecumseh Sherman",
  "n11298240": "Sherrington, Sir Charles Scott Sherrington",
  "n11298403": "Sherwood, Robert Emmet Sherwood",
  "n11298519": "Shevchenko, Taras Grigoryevich Shevchenko",
  "n11298634": "Shirer, William Lawrence Shirer",
  "n11298810": "Shockley, William Shockley, William Bradford Shockley",
  "n11299030": "Shostakovich, Dmitri Shostakovich, Dmitri Dmitrievich Shostakovich",
  "n11299212": "Shute, Nevil Shute, Nevil Shute Norway",
  "n11299367": "Sibelius, Jean Sibelius, Johan Julius Christian Sibelius",
  "n11299501": "Siddons, Sarah Siddons, Sarah Kemble Siddons",
  "n11299672": "Sidney, Sir Philip Sidney",
  "n11299770": "Siemens, Ernst Werner von Siemens",
  "n11299890": "Siemens, Karl Wilhelm Siemens, Sir Charles William Siemens",
  "n11300089": "Sikorsky, Igor Sikorsky, Igor Ivanovich Sikorsky",
  "n11300326": "Sills, Beverly Sills, Belle Miriam Silverman",
  "n11300465": "Silverstein, Shel Silverstein, Shelby Silverstein",
  "n11300678": "Simenon, Georges Simenon, Georges Joseph Christian Simenon",
  "n11300893": "Simon, Herb Simon, Herbert A. Simon, Herbert Alexander Simon",
  "n11301129": "Simon, Neil Simon, Marvin Neil Simon",
  "n11301279": "Simon, Paul Simon",
  "n11301414": "Simon, St. Simon, Simon Zelotes, Simon the Zealot, Simon the Canaanite",
  "n11301597": "Simpson, Sir James Young Simpson",
  "n11301809": "Simpson, Mrs. Simpson, Wallis Warfield Simpson, Wallis Warfield Windsor, Duchess of Windsor",
  "n11302062": "Sinatra, Frank Sinatra, Francis Albert Sinatra",
  "n11302224": "Sinclair, Clive Sinclair, Sir Clive Marles Sinclair",
  "n11302449": "Sinclair, Upton Sinclair, Upton Beall Sinclair",
  "n11302615": "Singer, Isaac Bashevis Singer",
  "n11302772": "Singer, Isaac M. Singer, Isaac Merrit Singer",
  "n11302962": "Siqueiros, David Siqueiros, David Alfaro Siqueiros",
  "n11303132": "Siraj-ud-daula",
  "n11303522": "Sitter, Willem de Sitter",
  "n11303698": "Sitting Bull",
  "n11304011": "Sitwell, Dame Edith Sitwell, Dame Edith Louisa Sitwell",
  "n11304139": "Sixtus IV, Francesco della Rovere",
  "n11304354": "Skeat, Walter William Skeat",
  "n11304461": "Skinner, Fred Skinner, B. F. Skinner, Burrhus Frederic Skinner",
  "n11304669": "Skinner, Cornelia Otis Skinner",
  "n11304811": "Skinner, Otis Skinner",
  "n11304912": "Smalley, Richard Smalley, Richard E. Smalley, Richard Errett Smalley",
  "n11305159": "Smetana, Bedrich Smetana",
  "n11305258": "Smith, Adam Smith",
  "n11305402": "Smith, John Smith, Captain John Smith",
  "n11305611": "Smith, Joseph Smith",
  "n11305745": "Smith, Bessie Smith",
  "n11305851": "Smith, Julia Evelina Smith",
  "n11306008": "Smith, Kate Smith, Kathryn Elizabeth Smith",
  "n11306175": "Smith, David Smith, David Roland Smith",
  "n11306297": "Smith, Ian Smith, Ian Douglas Smith",
  "n11306473": "Smollett, Tobias Smollett, Tobias George Smollett",
  "n11306619": "Smuts, Jan Christian Smuts",
  "n11306760": "Snead, Sam Snead, Samuel Jackson Snead",
  "n11306920": "Snellen, Hermann Snellen",
  "n11307082": "Snow, C. P. Snow, Charles Percy Snow, Baron Snow of Leicester",
  "n11307262": "Socinus, Faustus Socinus, Fausto Paolo Sozzini",
  "n11307422": "Socrates",
  "n11307587": "Soddy, Frederick Soddy",
  "n11307759": "Solomon",
  "n11307937": "Solvay, Ernest Solvay",
  "n11308120": "Solzhenitsyn, Alexander Isayevich Solzhenitsyn, Aleksandr Solzhenitsyn, Aleksandr I. Solzhenitsyn",
  "n11308396": "Sondheim, Stephen Sondheim",
  "n11308520": "Sontag, Susan Sontag",
  "n11308624": "Sophocles",
  "n11308739": "Sorensen, Soren Peter Lauritz Sorensen",
  "n11308877": "Soufflot, Jacques Germain Soufflot",
  "n11308988": "Sousa, John Philip Sousa, March King",
  "n11309164": "Southey, Robert Southey",
  "n11309318": "Soutine, Chaim Soutine",
  "n11309449": "Spallanzani, Lazzaro Spallanzani",
  "n11309613": "Spark, Muriel Spark, Dame Muriel Spark, Muriel Sarah Spark",
  "n11309772": "Spassky, Boris Spassky, Boris Vasilevich Spassky",
  "n11309958": "Speer, Albert Speer",
  "n11310081": "Speke, John Speke, John Hanning Speke",
  "n11310321": "Spencer, Herbert Spencer",
  "n11310523": "Spender, Stephen Spender, Sir Stephen Harold Spender",
  "n11310679": "Spengler, Oswald Spengler",
  "n11310833": "Spenser, Edmund Spenser",
  "n11311011": "Sperry, Elmer Ambrose Sperry",
  "n11311173": "Spielberg, Steven Spielberg",
  "n11311287": "Spillane, Mickey Spillane, Frank Morrison Spillane",
  "n11311450": "Spinoza, de Spinoza, Baruch de Spinoza, Benedict de Spinoza",
  "n11311623": "Spock, Benjamin Spock",
  "n11311817": "Spode, Josiah Spode",
  "n11311959": "Stael, Madame de Stael, Baronne Anne Louise Germaine Necker de Steal-Holstein",
  "n11312120": "Stalin, Joseph Stalin, Iosif Vissarionovich Dzhugashvili",
  "n11312362": "Standish, Miles Standish, Myles Standish",
  "n11312556": "Stanford, Leland Stanford",
  "n11312709": "Stanislavsky, Konstantin Stanislavsky, Konstantin Sergeyevich Stanislavsky, Konstantin Sergeevich Alekseev",
  "n11313011": "Stanley, Henry M. Stanley, Sir Henry Morton Stanley, John Rowlands",
  "n11313357": "Stanley, Francis Edgar Stanley",
  "n11313507": "Stanton, Elizabeth Cady Stanton",
  "n11313726": "Starr, Ringo Starr, Starkey, Richard Starkey",
  "n11313911": "St. Denis, Saint Denis, Ruth Saint Denis, Ruth St. Denis",
  "n11314119": "Steele, Sir Richrd Steele",
  "n11314219": "Steen, Jan Steen",
  "n11314315": "Steffens, Lincoln Steffens, Joseph Lincoln Steffens",
  "n11314514": "Steichen, Edward Jean Steichen",
  "n11314666": "Stein, Gertrude Stein",
  "n11314792": "Steinbeck, John Steinbeck, John Ernst Steinbeck",
  "n11314969": "Steinberg, Saul Steinberg",
  "n11315140": "Steinem, Gloria Steinem",
  "n11315249": "Steiner, Rudolf Steiner",
  "n11315379": "Steinman, David Barnard Steinman",
  "n11315580": "Steinmetz, Charles Proteus Steinmetz",
  "n11315760": "Steinway, Henry Steinway, Henry Engelhard Steinway, Heinrich Engelhard Steinway",
  "n11316003": "Stella, Frank Stella, Frank Philip Stella",
  "n11316141": "Steller, Georg Wilhelm Steller",
  "n11316249": "Stendhal, Marie Henri Beyle",
  "n11316429": "Stengel, Casey Stengel, Charles Dillon Stengel",
  "n11316567": "Stephen, Sir Leslie Stephen",
  "n11316669": "Stephenson, George Stephenson",
  "n11316828": "Stern, Isaac Stern",
  "n11316970": "Sterne, Laurence Sterne",
  "n11317086": "Steuben, Baron Friedrich Wilhelm Ludolf Gerhard Augustin von Steuben",
  "n11317309": "Stevens, George Stevens",
  "n11317416": "Stevens, Wallace Stevens",
  "n11317519": "Stevens, Smitty Stevens, S. Smith Stevens, Stanley Smith Stevens",
  "n11317749": "Stevenson, Adlai Stevenson, Adlai Ewing Stevenson",
  "n11317897": "Stevenson, Robert Louis Stevenson, Robert Louis Balfour Stevenson",
  "n11318039": "Stewart, Dugald Stewart",
  "n11318171": "Stewart, Jimmy Stewart, James Maitland Stewart",
  "n11318348": "Stieglitz, Alfred Stieglitz",
  "n11318462": "Stilwell, Joseph Warren Stilwell, Vinegar Joe Stilwell, Uncle Joe",
  "n11318692": "Stockton, Frank Stockton, Francis Richard Stockton",
  "n11318824": "Stoker, Bram Stoker, Abraham Stoker",
  "n11318967": "Stokowski, Leopold Stokowski, Leopold Antoni Stanislaw Stokowski",
  "n11319134": "Stone, Edward Durell Stone",
  "n11319244": "Stone, Harlan Fiske Stone",
  "n11319413": "Stone, I. F. Stone, Isidor Feinstein Stone",
  "n11319570": "Stone, Lucy Stone",
  "n11319704": "Stone, Oliver Stone",
  "n11319810": "Stone, Harlan Stone, Harlan F. Stone, Harlan Fisk Stone",
  "n11320044": "Stopes, Marie Stopes, Marie Charlotte Carmichael Stopes",
  "n11320245": "Stoppard, Tom Stoppard, Sir Tom Stoppard, Thomas Straussler",
  "n11320405": "Stowe, Harriet Beecher Stowe, Harriet Elizabeth Beecher Stowe",
  "n11320631": "Strachey, Lytton Strachey, Giles Lytton Strachey",
  "n11320821": "Stradivari, Antonio Stradivari, Stradivarius, Antonius Stradivarius",
  "n11321051": "Strasberg, Lee Strasberg, Israel Strassberg",
  "n11321296": "Strauss, Johann Strauss, Strauss the Elder",
  "n11321428": "Strauss, Johann Strauss, Strauss the Younger",
  "n11321647": "Strauss, Richard Strauss",
  "n11321841": "Stravinsky, Igor Stravinsky, Igor Fyodorovich Stravinsky",
  "n11322068": "Streep, Meryl Streep",
  "n11322178": "Streisand, Barbra Streisand, Barbra Joan Streisand",
  "n11322344": "Strickland, William Strickland",
  "n11322481": "Strindberg, August Strindberg, Johan August Strindberg",
  "n11322627": "Stroheim, Erich von Stroheim",
  "n11322758": "Stuart, Gilbert Stuart, Gilbert Charles Stuart",
  "n11322937": "Stubbs, William Stubbs",
  "n11323094": "Stuyvesant, Peter Stuyvesant, Petrus Stuyvesant",
  "n11323316": "Styron, William Styron",
  "n11323448": "Suckling, Sir John Suckling",
  "n11323580": "Sue, Eugene Sue",
  "n11323721": "Suharto",
  "n11323849": "Sukarno, Achmad Sukarno",
  "n11324090": "Sulla, Lucius Cornelius Sulla Felix",
  "n11324212": "Sullivan, Arthur Sullivan, Arthur Seymour Sullivan, Sir Arthur Sullivan",
  "n11324428": "Sullivan, Anne Sullivan, Anne Mansfield Sullivan",
  "n11324619": "Sullivan, Ed Sullivan, Edward Vincent Sullivan",
  "n11324785": "Sullivan, Harry Stack Sullivan",
  "n11324902": "Sullivan, Louis Sullivan, Louis Henry Sullivan, Louis Henri Sullivan",
  "n11325146": "Sully, Duc de Sully, Maxmilien de Bethune",
  "n11325265": "Sully, Thomas Sully",
  "n11325419": "Sumner, William Graham Sumner",
  "n11325534": "Sunday, Billy Sunday, William Ashley Sunday",
  "n11325663": "Sun Yat-sen, Sun Yixian",
  "n11325867": "Sutherland, Joan Sutherland, Dame Joan Sutherland",
  "n11326008": "Sverdrup, Otto Neumann Sverdrup",
  "n11326154": "Swammerdam, Jan Swammerdam",
  "n11326433": "Swanson, Gloria Swanson, Gloria May Josephine Svensson",
  "n11326591": "Swedenborg, Svedberg, Emanuel Swedenborg, Emanuel Svedberg",
  "n11326730": "Sweet, Henry Sweet",
  "n11326869": "Swift, Jonathan Swift, Dean Swift",
  "n11326999": "Swift, Gustavus Franklin Swift",
  "n11327163": "Swinburne, Algernon Charles Swinburne",
  "n11327273": "Sydenham, Thomas Sydenham, English Hippocrates",
  "n11327398": "Sylvester II, Gerbert",
  "n11327544": "Symonds, John Addington Symonds",
  "n11327650": "Symons, Arthur Symons",
  "n11327744": "Synge, J. M. Synge, John Millington Synge, Edmund John Millington Synge",
  "n11327964": "Szell, George Szell",
  "n11328085": "Szent-Gyorgyi, Albert Szent-Gyorgyi, Albert von Szent-Gyorgyi",
  "n11328289": "Szilard, Leo Szilard",
  "n11328524": "Tacitus, Publius Cornelius Tacitus, Gaius Cornelius Tacitus",
  "n11328714": "Taft, William Howard Taft, President Taft",
  "n11328930": "Taft, Lorado Taft",
  "n11329030": "Tagore, Rabindranath Tagore, Sir Rabindranath Tagore",
  "n11329281": "Talbot, Fox Talbot, William Henry Fox Talbot",
  "n11329507": "Tallchief, Maria Tallchief",
  "n11329690": "Talleyrand, Charles Maurice de Talleyrand",
  "n11329808": "Tallis, Thomas Tallis",
  "n11330026": "Tamerlane, Tamburlaine, Timur, Timur Lenk",
  "n11330228": "Tamm, Igor Tamm, Igor Yevgeneevich Tamm",
  "n11330346": "Tancred",
  "n11330504": "Tandy, Jessica Tandy",
  "n11330692": "Taney, Roger Taney, Roger Brooke Taney",
  "n11330963": "Tange, Kenzo Tange",
  "n11331063": "Tanguy, Yves Tanguy",
  "n11331192": "Tappan, Arthur Tappan",
  "n11331300": "Tarantino, Quentin Tarantino, Quentin Jerome Tarantino",
  "n11331442": "Tarbell, Ida Tarbell, Ida M. Tarbell, Ida Minerva Tarbell",
  "n11331669": "Tarkovsky, Andrei Tarkovsky, Andrei Arsenevich Tarkovsky",
  "n11331804": "Tarquin, Tarquin the Proud, Tarquinius, Tarquinius Superbus, Lucius Tarquinius Superbus",
  "n11332068": "Tasman, Abel Tasman, Abel Janszoon Tasman",
  "n11332250": "Tasso, Torquato Tasso",
  "n11332423": "Tate, Allen Tate, John Orley Allen Tate",
  "n11332572": "Tati, Jacques Tati, Jacques Tatischeff",
  "n11332688": "Tatum, Art Tatum, Arthur Tatum",
  "n11332892": "Tatum, Edward Lawrie Tatum",
  "n11333071": "Tawney, Richard Henry Tawney",
  "n11333237": "Taylor, Zachary Taylor, President Taylor",
  "n11333390": "Taylor, Elizabeth Taylor",
  "n11333601": "Taylor, Deems Taylor, Joseph Deems Taylor",
  "n11333762": "Tchaikovsky, Peter Tchaikovsky, Peter Ilich Tchaikovsky, Pyotr Tchaikovsky, Pyotr Ilych Tchaikovsky",
  "n11334003": "Teach, Edward Teach, Thatch, Edward Thatch, Blackbeard",
  "n11334215": "Teasdale, Sara Teasdale",
  "n11334317": "Tebaldi, Renata Tebaldi",
  "n11334428": "Tecumseh, Tecumtha",
  "n11334609": "Teilhard de Chardin, Pierre Teilhard de Chardin",
  "n11334773": "Te Kanawa, Dame Kiri Te Kanawa, Dame Kiri Janette Te Kanawa",
  "n11334925": "Telemann, Georg Philipp Telemann",
  "n11335041": "Teller, Edward Teller",
  "n11335226": "Tenniel, Sir John Tenniel",
  "n11335330": "Tennyson, Alfred Tennyson, First Baron Tennyson, Alfred Lord Tennyson",
  "n11335491": "Tenzing Norgay",
  "n11335695": "Terence, Publius Terentius Afer",
  "n11335878": "Teresa, Mother Teresa, Theresa, Mother Theresa, Agnes Gonxha Bojaxhiu",
  "n11336176": "Teresa of Avila, Saint Teresa of Avila",
  "n11336364": "Tereshkova, Valentina Tereshkova, Valentina Vladmirovna Tereshkova",
  "n11336544": "Terry, Dame Ellen Terry, Dame Alice Ellen Terry",
  "n11336668": "Tertullian, Quintus Septimius Florens Tertullianus",
  "n11336850": "Tesla, Nikola Tesla",
  "n11337202": "Thackeray, William Makepeace Thackeray",
  "n11337331": "Thales, Thales of Miletus",
  "n11337629": "Tharp, Twyla Tharp",
  "n11337779": "Thatcher, Margaret Thatcher, Margaret Hilda Thatcher, Baroness Thatcher of Kesteven, Iron Lady",
  "n11337999": "Themistocles",
  "n11338172": "Theodosius, Theodosius I, Theodosius the Great, Flavius Theodosius",
  "n11338484": "Theophrastus",
  "n11338667": "Thespis",
  "n11338796": "Thomas, Saint Thomas, St. Thomas, doubting Thomas, Thomas the doubting Apostle",
  "n11339041": "Thomas, Dylan Thomas, Dylan Marlais Thomas",
  "n11339155": "Thomas, Lowell Thomas, Lowell Jackson Thomas",
  "n11339361": "Thomas, Norman Thomas, Norman Mattoon Thomas",
  "n11339534": "Thomas, Seth Thomas",
  "n11339669": "Thompson, Benjamin Thompson, Count Rumford",
  "n11339905": "Thompson, Homer Thompson, Homer A. Thompson, Homer Armstrong Thompson",
  "n11340146": "Thomson, Joseph John Thomson, Sir Joseph John Thomson",
  "n11340411": "Thomson, George Paget Thomson, Sir George Paget Thomson",
  "n11340642": "Thomson, Elihu Thomson",
  "n11340824": "Thomson, Virgil Thomson, Virgil Garnett Thomson",
  "n11340992": "Thoreau, Henry David Thoreau",
  "n11341137": "Thorndike, Edward Lee Thorndike",
  "n11341267": "Thorndike, Dame Sybil Thorndike",
  "n11341374": "Thornton, William Thornton",
  "n11341479": "Thorpe, Jim Thorpe, James Francis Thorpe",
  "n11341614": "Thucydides",
  "n11341760": "Thurber, James Thurber, James Grover Thurber",
  "n11341970": "Tiberius, Tiberius Claudius Nero Caesar Augustus",
  "n11342191": "Tiepolo, Giovanni Battista Tiepolo",
  "n11342301": "Tiffany, Louis Comfort Tiffany",
  "n11342440": "Tilden, Big Bill Tilden, William Tatem Tilden Jr.",
  "n11342618": "Tillich, Paul Tillich, Paul Johannes Tillich",
  "n11342766": "Timothy",
  "n11342905": "Tinbergen, Jan Tinbergen",
  "n11343040": "Tinbergen, Nikolaas Tinbergen",
  "n11343208": "Tintoretto, Jacopo Robusti",
  "n11343333": "Tirso de Molina, Gabriel Tellez",
  "n11343507": "Titian, Tiziano Vecellio",
  "n11343625": "Tito, Marshal Tito, Josip Broz",
  "n11343847": "Titus, Titus Vespasianus Augustus, Titus Flavius Vespasianus",
  "n11343998": "Titus",
  "n11344092": "Tobey, Mark Tobey",
  "n11344235": "Tobin, James Tobin",
  "n11344337": "Tocqueville, Alexis de Tocqueville, Alexis Charles Henri Maurice de Tocqueville",
  "n11344549": "Todd, Sir Alexander Robertus Todd, Lord Todd",
  "n11344733": "Tojo, Tojo Hideki, Tojo Eiki",
  "n11345017": "Toklas, Alice B. Toklas",
  "n11345181": "Tolkien, J.R.R. Tolkien, John Ronald Reuel Tolkien",
  "n11345378": "Tolstoy, Leo Tolstoy, Count Lev Nikolayevitch Tolstoy",
  "n11345539": "Tombaugh, Clyde Tombaugh, Clyde William Tombaugh",
  "n11345705": "Tonegawa Susumu",
  "n11345878": "Torquemada, Tomas de Torquemada",
  "n11346110": "Torricelli, Evangelista Torricelli",
  "n11346257": "Toscanini, Arturo Toscanini",
  "n11346391": "Toulouse-Lautrec, Henri Toulouse-Lautrec",
  "n11346568": "Tourette, Gilles de la Tourette, Georges Gilles de la Tourette",
  "n11346710": "Town, Ithiel Town",
  "n11346873": "Townes, Charles Townes, Charles Hard Townes",
  "n11347080": "Townsend, Francis Everett Townsend",
  "n11347317": "Toynbee, Arnold Toynbee, Arnold Joseph Toynbee",
  "n11347519": "Tracy, Spencer Tracy",
  "n11347674": "Tradescant, John Tradescant",
  "n11347834": "Trajan, Marcus Ulpius Traianus",
  "n11348047": "Traubel, Helen Traubel",
  "n11348160": "Tree, Sir Herbert Beerbohm Tree",
  "n11348356": "Trevelyan, George Otto Trevelyan, Sir George Otto Trevelyan",
  "n11348584": "Trevelyan, George Macaulay Trevelyan",
  "n11348812": "Trevino, Lee Trevino, Lee Buck Trevino, Supermex",
  "n11348946": "Trevithick, Richard Trevithick",
  "n11349092": "Trilling, Lionel Trilling",
  "n11349207": "Trollope, Anthony Trollope",
  "n11349318": "Trotsky, Leon Trotsky, Lev Davidovich Bronstein",
  "n11349635": "Truffaut, Francois Truffaut",
  "n11349739": "Truman, Harry Truman, Harry S Truman, President Truman",
  "n11350059": "Trumbo, Dalton Trumbo",
  "n11350286": "Trumbull, John Trumbull",
  "n11350393": "Trumbull, John Trumbull",
  "n11350514": "Trumbull, Jonathan Trumbull",
  "n11350705": "Truth, Sojourner Truth",
  "n11350959": "Tubman, Harriet Tubman",
  "n11351212": "Tuchman, Barbara Tuchman, Barbara Wertheim Tuchman",
  "n11351347": "Tucker, Sophie Tucker",
  "n11351529": "Tucker, Benjamin Ricketson Tucker",
  "n11351677": "Tudor, Antony Tudor",
  "n11351832": "Tunney, Gene Tunney, James Joseph Tunney",
  "n11352035": "Turgenev, Ivan Turgenev, Ivan Sergeevich Turgenev",
  "n11352192": "Turgot, Anne Robert Jacques Turgot",
  "n11352498": "Turing, Alan Turing, Alan Mathison Turing",
  "n11352701": "Turner, Frederick Jackson Turner",
  "n11352883": "Turner, Joseph Mallord William Turner",
  "n11353078": "Turner, Henry Hubert Turner",
  "n11353195": "Turner, Nat Turner",
  "n11353412": "Turpin, Dick Turpin",
  "n11353510": "Tussaud, Marie Tussaud, Madame Tussaud, Marie Grosholtz",
  "n11353847": "Tutankhamen",
  "n11354001": "Tutu, Desmond Tutu",
  "n11354145": "Tyler, John Tyler, President Tyler",
  "n11354333": "Tyndale, William Tyndale, Tindale, William Tindale, Tindal, William Tindal",
  "n11354743": "Tyndall, John Tyndall",
  "n11355082": "Tyson, Mike Tyson, Michael Gerald Tyson",
  "n11355247": "Tzara, Tristan Tzara, Samuel Rosenstock",
  "n11355428": "Uhland, Johann Ludwig Uhland",
  "n11355537": "Ulanova, Galina Ulanova, Galina Sergeevna Ulanova",
  "n11355669": "Ulfilas, Bishop Ulfilas, Ulfila, Bishop Ulfila, Wulfila, Bishop Wulfila",
  "n11356018": "ultramontane",
  "n11356183": "Undset, Sigrid Undset",
  "n11356283": "Untermeyer, Louis Untermeyer",
  "n11356392": "Updike, John Updike, John Hoyer Updike",
  "n11356512": "Upjohn, Richard Upjohn",
  "n11356636": "Urban II, Odo, Odo of Lagery, Otho, Otho of Lagery",
  "n11356822": "Urban V, Guillaume de Grimoard",
  "n11357086": "Urban VI, Bartolomeo Prignano",
  "n11357332": "Urban VIII, Maffeo Barberini",
  "n11357514": "Urey, Harold Urey, Harold Clayton Urey",
  "n11357660": "Uriah",
  "n11357879": "Ussher, James Ussher, Usher, James Usher",
  "n11358065": "Ustinov, Sir Peter Ustinov, Peter Alexander Ustinov",
  "n11358225": "Utrillo, Maurice Utrillo",
  "n11358374": "Van Allen, James Alfred Van Allen",
  "n11358598": "Vanbrugh, John Vanbrugh, Sir John Vanbrigh",
  "n11358719": "Van Buren, Martin Van Buren, President Van Buren",
  "n11358863": "Vancouver, George Vancouver",
  "n11359037": "Van de Graaff, Robert Van de Graaff, Robert Jemison Van de Graaff",
  "n11359187": "Vanderbilt, Cornelius Vanderbilt, Commodore Vanderbilt",
  "n11359412": "van der Waals, Johannes van der Waals, Johannes Diderik van der Waals",
  "n11359558": "van de Velde, Henri van de Velde, Henri Clemens van de Velde",
  "n11359697": "Van Doren, Carl Van Doren, Carl Clinton Van Doren",
  "n11359867": "Vandyke, Van Dyck, Anthony Vandyke, Sir Anthony Vandyke",
  "n11360022": "van Gogh, Vincent van Gogh, Gogh",
  "n11360175": "Van Vleck, John Van Vleck, John Hasbrouck Van Vleck",
  "n11360311": "Vanzetti, Bartolomeo Vanzetti",
  "n11360534": "Varese, Edgar Varese",
  "n11360744": "Vargas, Getulio Dornelles Vargas",
  "n11360895": "Vargas Llosa, Mario Vargas Llosa, Jorge Mario Pedro Vargas Llosa",
  "n11361039": "Varro, Marcus Terentius Varro",
  "n11361142": "Vasarely, Viktor Vasarely",
  "n11361288": "Vasari, Giorgio Vasari",
  "n11361423": "Vaughan, Sarah Vaughan",
  "n11361585": "Vaughan Williams, Ralph Vaughan Williams",
  "n11361757": "Vaux, Calvert Vaux",
  "n11361913": "Veblen, Oswald Veblen",
  "n11362022": "Veblen, Thorstein Veblen, Thorstein Bunde Veblen",
  "n11362195": "Vega, Lope de Vega, Lope Felix de Vega Carpio",
  "n11362329": "Velazquez, Diego Rodriguez de Silva y Velazquez",
  "n11362452": "Venn, John Venn",
  "n11362573": "Ventner, Craig Ventner, J. Craig Ventner",
  "n11362885": "Venturi, Robert Venturi, Robert Charles Venturi",
  "n11363020": "Verdi, Giuseppe Verdi, Guiseppe Fortunino Francesco Verdi",
  "n11363164": "Verlaine, Paul Verlaine",
  "n11363269": "Vermeer, Jan Vermeer, Jan van der Meer",
  "n11363412": "Verne, Jules Verne",
  "n11363552": "Verner, Karl Adolph Verner",
  "n11363657": "Vernier, Paul Vernier",
  "n11363791": "Veronese, Paolo Veronese, Paola Caliari",
  "n11363930": "Verrazano, Giovanni da Verrazano, Verrazzano, Giovanni da Verrazzano",
  "n11364135": "Versace, Gianni Versace",
  "n11364243": "Verwoerd, Hendrik Verwoerd, Hendrik Frensch Verwoerd",
  "n11364419": "Vesalius, Andreas Vesalius",
  "n11364570": "Vesey, Denmark Vesey",
  "n11364799": "Vespasian, Titus Flavius Sabinus Vespasianus",
  "n11365100": "Vespucci, Amerigo Vespucci, Americus Vespucius",
  "n11365300": "Vestris, Gaetan Vestris",
  "n11365516": "Victor Emanuel II",
  "n11365674": "Victor Emanuel III",
  "n11365857": "Victoria, Queen Victoria",
  "n11366109": "Vidal, Gore Vidal, Eugene Luther Vidal",
  "n11366232": "Vigee-Lebrun, Elisabeth Vigee-Lebrun, Marie Louise Elisabeth Vigee-Lebrun",
  "n11366405": "Villa, Pancho Villa, Francisco Villa, Doroteo Arango",
  "n11366548": "Villa-Lobos, Heitor Villa-Lobos",
  "n11366658": "Villard, Henry Villard",
  "n11366787": "Villon, Francois Villon",
  "n11366895": "Vinogradoff, Sir Paul Gavrilovich Vinogradoff",
  "n11367035": "Vinson, Frederick Moore Vinson",
  "n11367195": "Virchow, Rudolf Virchow, Rudolf Karl Virchow",
  "n11367436": "Virgil, Vergil, Publius Vergilius Maro",
  "n11367581": "Visconti, Luchino Visconti, Don Luchino Visconti Conte di Modrone",
  "n11367725": "Vitus, St. Vitus",
  "n11367910": "Vivaldi, Antonio Vivaldi, Antonio Lucio Vivaldi",
  "n11368076": "Vizcaino, Sebastian Vizcaino",
  "n11368240": "Vlaminck, Maurice de Vlaminck",
  "n11368368": "Volta, Count Alessandro Volta, Conte Alessandro Volta, Conte Alessandro Giuseppe Antonio Anastasio Volta",
  "n11368638": "Voltaire, Arouet, Francois-Marie Arouet",
  "n11368841": "Vonnegut, Kurt Vonnegut",
  "n11369035": "von Neumann, Neumann, John von Neumann",
  "n11369251": "von Sternberg, Josef von Sternberg",
  "n11369444": "Voznesenski, Andrei Voznesenski",
  "n11369551": "Vuillard, Edouard Vuillard, Jean Edouard Vuillard",
  "n11369676": "Wade, Virginia Wade",
  "n11369834": "Wagner, Richard Wagner, Wilhelm Richard Wagner",
  "n11370068": "Wagner, Otto Wagner",
  "n11370201": "Wain, John Wain, John Barrington Wain",
  "n11370314": "Waite, Morrison Waite, Morrison R. Waite, Morrison Remick Waite",
  "n11370554": "Wajda, Andrzej Wajda",
  "n11370654": "Waldheim, Kurt Waldheim",
  "n11370990": "Walesa, Lech Walesa",
  "n11371125": "Walker, Alice Walker, Alice Malsenior Walker",
  "n11371254": "Walker, John Walker",
  "n11371443": "Wallace, Alfred Russel Wallace",
  "n11371622": "Wallace, Edgar Wallace, Richard Horatio Edgar Wallace",
  "n11371778": "Wallace, Sir William Wallace",
  "n11372054": "Wallenstein, Albrecht Eusebius Wenzel von Wallenstein",
  "n11372242": "Waller, Fats Waller, Thomas Wright Waller",
  "n11372372": "Walpole, Robert Walpole, Sir Robert Walpole, First Earl of Orford",
  "n11372599": "Walpole, Horace Walpole, Horatio Walpole, Fourth Earl of Orford",
  "n11372799": "Walter, Bruno Walter",
  "n11372896": "Walton, E. T. S. Walton, Ernest Walton, Ernest Thomas Sinton Walton",
  "n11373099": "Walton, Izaak Walton",
  "n11373231": "Walton, William Walton, Sir William Walton, Sir William Turner Walton",
  "n11373379": "Wanamaker, John Wanamaker",
  "n11373550": "Warburg, Aby Warburg, Aby Moritz Warburg",
  "n11373672": "Warburg, Otto Heinrich Warburg",
  "n11373897": "Ward, Montgomery Ward, Aaron Montgomery Ward",
  "n11374085": "Ward, Mrs. Humphrey Ward, Mary Augusta Arnold Ward",
  "n11374281": "Ward, Barbara Ward, Baroness Jackson of Lodsworth",
  "n11374448": "Warhol, Andy Warhol",
  "n11374589": "Warner, Charles Dudley Warner",
  "n11374789": "Warren, Earl Warren",
  "n11374952": "Warren, Robert Penn Warren",
  "n11375087": "Warwick, Earl of Warwick, Richard Neville, Kingmaker",
  "n11375418": "Washington, George Washington, President Washington",
  "n11375677": "Washington, Booker T. Washington, Booker Taliaferro Washington",
  "n11375909": "Wassermann, August von Wassermann",
  "n11376069": "Waters, Ethel Waters",
  "n11376201": "Watson, James Watson, James Dewey Watson",
  "n11376400": "Watson, John Broadus Watson",
  "n11376565": "Watson, Thomas Augustus Watson",
  "n11376742": "Watt, James Watt",
  "n11376939": "Watteau, Jean Antoine Watteau",
  "n11377043": "Watts, Isaac Watts",
  "n11377168": "Waugh, Evelyn Waugh, Evelyn Arthur Saint John Waugh",
  "n11377315": "Wavell, Archibald Percival Wavell, First Earl Wavell",
  "n11377564": "Wayne, Anthony Wayne, Mad Anthony Wayne",
  "n11377712": "Wayne, John Wayne, Duke Wayne",
  "n11377851": "Webb, Sidney Webb, Sidney James Webb, First Baron Passfield",
  "n11378087": "Webb, Beatrice Webb, Martha Beatrice Potter Webb",
  "n11378254": "Weber, E. H. Weber, Ernst Heinrich Weber",
  "n11378462": "Weber, Carl Maria von Weber, Baron Karl Maria Friedrich Ernst von Weber",
  "n11378662": "Weber, Max Weber",
  "n11378805": "Weber, Max Weber",
  "n11378929": "Weber, Wilhelm Eduard Weber",
  "n11379108": "Webster, Noah Webster",
  "n11379217": "Webster, Daniel Webster",
  "n11379336": "Webster, John Webster",
  "n11379436": "Wedgwood, Josiah Wedgwood",
  "n11379536": "Wegener, Alfred Lothar Wegener",
  "n11379691": "Weil, Andre Weil",
  "n11379812": "Weil, Simone Weil",
  "n11379908": "Weill, Kurt Weill",
  "n11380035": "Weinberg, Steven Weinberg",
  "n11380159": "Weismann, August Friedrich Leopold Weismann",
  "n11380429": "Weizmann, Chaim Weizmann, Chaim Azriel Weizmann",
  "n11380655": "Weld, Theodore Dwight Weld",
  "n11380768": "Welles, Orson Welles, George Orson Welles",
  "n11380923": "Wellington, Duke of Wellington, First Duke of Wellington, Arthur Wellesley, Iron Duke",
  "n11381193": "Wells, H. G. Wells, Herbert George Wells",
  "n11381457": "Welty, Eudora Welty",
  "n11381583": "Werfel, Franz Werfel",
  "n11381684": "Wernicke, Karl Wernicke",
  "n11381824": "Wesley, John Wesley",
  "n11381964": "Wesley, Charles Wesley",
  "n11382112": "West, Benjamin West",
  "n11382278": "West, Mae West",
  "n11382398": "West, Rebecca West, Dame Rebecca West, Cicily Isabel Fairfield",
  "n11382555": "Westinghouse, George Westinghouse",
  "n11382688": "Weston, Edward Weston",
  "n11382795": "Wharton, Edith Wharton, Edith Newbold Jones Wharton",
  "n11382930": "Wheatley, Phillis Wheatley",
  "n11383100": "Wheatstone, Sir Charles Wheatstone",
  "n11383278": "Wheeler, Sir Mortimer Wheeler, Sir Robert Eric Mortimer Wheeler",
  "n11383425": "Whistler, James Abbott McNeill Whistler",
  "n11383546": "White, Andrew D. White, Andrew Dickson White",
  "n11383767": "White, E. B. White, Elwyn Brooks White",
  "n11383917": "White, Stanford White",
  "n11384022": "White, T. H. White, Theodore Harold White",
  "n11384159": "White, Patrick White, Patrick Victor Martindale White",
  "n11384291": "White, Edward White, Edward D. White, Edward Douglas White Jr.",
  "n11384566": "Whitehead, Alfred North Whitehead",
  "n11384755": "Whitman, Marcus Whitman",
  "n11384986": "Whitman, Walt Whitman",
  "n11385126": "Whitney, Eli Whitney",
  "n11385277": "Whittier, John Greenleaf Whittier",
  "n11385442": "Whittle, Frank Whittle, Sir Frank Whittle",
  "n11385611": "Wiener, Norbert Wiener",
  "n11385748": "Wiesel, Elie Wiesel, Eliezer Wiesel",
  "n11386005": "Wiesenthal, Samuel Wiesenthal",
  "n11386138": "Wigner, Eugene Wigner, Eugene Paul Wigner",
  "n11386346": "Wilde, Oscar Wilde, Oscar Fingal O'Flahertie Wills Wilde",
  "n11386503": "Wilder, Billy Wilder, Samuel Wilder",
  "n11386692": "Wilder, Thornton Wilder, Thornton Niven Wilder",
  "n11386853": "Wilhelm II, Kaiser Wilhelm, Kaiser Bill",
  "n11387060": "Wilkes, Charles Wilkes",
  "n11387179": "Wilkes, John Wilkes",
  "n11387362": "Wilkins, Maurice Wilkins, Maurice Hugh Frederick Wilkins",
  "n11387539": "Wilkins, George Hubert Wilkins",
  "n11387692": "Wilkins, Roy Wilkins",
  "n11387806": "Wilkinson, Sir Geoffrey Wilkinson",
  "n11387973": "Willard, Emma Hart Willard",
  "n11388141": "Willard, Frances Elizabeth Caroline Willard",
  "n11388321": "Willebrand, von Willebrand, E. A. von Willebrand, Erik von Willebrand, Erik Adolf von Willebrand",
  "n11388538": "William I, William the Conqueror",
  "n11388841": "William II, William Rufus",
  "n11389003": "William III, William of Orange",
  "n11389301": "William IV, Sailor King",
  "n11389481": "Williams, Tennessee Williams, Thomas Lanier Williams",
  "n11389619": "Williams, Roger Williams",
  "n11389901": "Williams, Ted Williams, Theodore Samuel Williams",
  "n11390058": "Williams, William Carlos Williams",
  "n11390170": "Williams, Sir Bernard Williams, Bernard Arthur Owen Williams",
  "n11390364": "Williams, Hank Williams, Hiram Williams, Hiram King Williams",
  "n11390549": "Willis, Thomas Willis",
  "n11390692": "Wilmut, Ian Wilmut",
  "n11390855": "Wilson, Woodrow Wilson, Thomas Woodrow Wilson, President Wilson",
  "n11391123": "Wilson, Edmund Wilson",
  "n11391234": "Wilson, Charles Thomson Rees Wilson",
  "n11391379": "Wilson, E. O. Wilson, Edward Osborne Wilson",
  "n11391587": "Wilson, James Wilson",
  "n11391759": "Wilson, John Tuzo Wilson",
  "n11391915": "Wilson, Robert Woodrow Wilson",
  "n11392082": "Wilson, Alexander Wilson",
  "n11392210": "Wilson, Sir Angus Wilson, Angus Frank Johnstone Wilson",
  "n11392368": "Wilson, Harriet Wilson",
  "n11392539": "Winckelmann, Johann Winckelmann, Johann Joachim Winckelmann",
  "n11392754": "Windaus, Adolf Windaus",
  "n11392913": "Winslow, Edward Winslow",
  "n11393124": "Wise, Isaac Mayer Wise",
  "n11393315": "Wise, Stephen Samuel Wise",
  "n11393446": "Wister, Owen Wister",
  "n11393546": "Witherspoon, John Witherspoon",
  "n11393828": "Wittgenstein, Ludwig Wittgenstein, Ludwig Josef Johan Wittgenstein",
  "n11394042": "Wodehouse, P. G. Wodehouse, Pelham Grenville Wodehouse",
  "n11394214": "Wolf, Friedrich August Wolf",
  "n11394398": "Wolf, Hugo Wolf",
  "n11394491": "Wolfe, Thomas Wolfe, Thomas Clayton Wolfe",
  "n11394657": "Wolfe, Tom Wolfe, Thomas Wolfe, Thomas Kennerly Wolfe Jr.",
  "n11394848": "Wolff, Kaspar Friedrich Wolff",
  "n11394954": "Wollaston, William Hyde Wollaston",
  "n11395199": "Wollstonecraft, Mary Wollstonecraft, Mary Wollstonecraft Godwin",
  "n11395466": "Wood, Grant Wood",
  "n11395609": "Wood, Mrs. Henry Wood, Ellen Price Wood",
  "n11395773": "Wood, Sir Henry Wood, Sir Henry Joseph Wood",
  "n11395895": "Wood, Natalie Wood",
  "n11396000": "Woodbury, Helen Laura Sumner Woodbury",
  "n11396128": "Woodhull, Victoria Clafin Woodhull",
  "n11396338": "Woodward, Bob Woodward, Robert Woodward, Robert Burns Woodward",
  "n11396535": "Woodward, C. Vann Woodward, Comer Vann Woodward",
  "n11396667": "Woolf, Virginia Woolf, Adeline Virginia Stephen Woolf",
  "n11396943": "Woollcott, Alexander Woollcott",
  "n11397094": "Woolley, Sir Leonard Woolley, Sir Charles Leonard Woolley",
  "n11397271": "Woolworth, Frank Winfield Woolworth",
  "n11397488": "Worcester, Joseph Emerson Worcester",
  "n11397657": "Wordsworth, William Wordsworth",
  "n11397885": "Worth, Charles Frederick Worth",
  "n11398094": "Wouk, Herman Wouk",
  "n11398195": "Wren, Sir Christopher Wren",
  "n11398344": "Wright, Frances Wright, Fanny Wright",
  "n11398489": "Wright, Frank Lloyd Wright",
  "n11398611": "Wright, Orville Wright",
  "n11398783": "Wright, Wilbur Wright",
  "n11398955": "Wright, Richard Wright",
  "n11399123": "Wright, Willard Huntington Wright, S. S. Van Dine",
  "n11399274": "Wurlitzer, Rudolf Wurlitzer",
  "n11399446": "Wyatt, Sir Thomas Wyatt, Wyat, Sir Thomas Wyat",
  "n11399620": "Wyatt, James Wyatt",
  "n11399716": "Wycherley, William Wycherley",
  "n11399866": "Wycliffe, John Wycliffe, Wickliffe, John Wickliffe, Wyclif, John Wyclif, Wiclif, John Wiclif",
  "n11400126": "Wyeth, Andrew Wyeth",
  "n11400230": "Wykeham, William of Wykeham",
  "n11400490": "Wyler, William Wyler",
  "n11400594": "Wylie, Elinor Morton Hoyt Wylie",
  "n11400704": "Wynette, Tammy Wynette, Tammy Wynetter Pugh",
  "n11400837": "Wyszynski, Stefan Wyszynski",
  "n11401009": "Xavier, Saint Francis Xavier",
  "n11401194": "Xenophanes",
  "n11401282": "Xenophon",
  "n11401418": "Xerxes I, Xerxes the Great",
  "n11401617": "Yale, Elihu Yale",
  "n11401797": "Yamamoto, Isoroku Yamamoto",
  "n11401947": "Yamani, Ahmed Zoki Yamani",
  "n11402120": "Yang Chen Ning, Chen N. Yang",
  "n11402341": "Yastrzemski, Carl Yastrzemski",
  "n11402463": "Yeats, William Butler Yeats, W. B. Yeats",
  "n11402626": "Yerkes, Robert M. Yerkes, Robert Mearns Yerkes",
  "n11402801": "Yersin, Alexandre Yersin, Alexandre Emile Jean Yersin",
  "n11403015": "Yevtushenko, Yevgeni Yevtushenko, Yevgeni Aleksandrovich Yevtushenko",
  "n11403220": "Young, Brigham Young",
  "n11403456": "Young, Cy Young, Danton True Young",
  "n11403600": "Young, Edward Young",
  "n11403692": "Young, Pres Young, Lester Willis Young",
  "n11403828": "Young, Thomas Young",
  "n11404140": "Young, Whitney Young, Whitney Moore Young Jr.",
  "n11404280": "Young, Loretta Young",
  "n11404402": "Yukawa, Hideki Yukawa",
  "n11404666": "Zaharias, Babe Zaharias, Didrikson, Babe Didrikson, Mildred Ella Didrikson, Mildred Ella Didrikson Zaharias",
  "n11404871": "Zangwill, Israel Zangwill",
  "n11404971": "Zanuck, Darryl Zanuck, Darryl Francis Zanuck",
  "n11405176": "Zapata, Emiliano Zapata",
  "n11405319": "Zechariah, Zacharias",
  "n11405437": "Zeeman, Pieter Zeeman",
  "n11405684": "Zeno, Zeno of Citium",
  "n11405826": "Zeno, Zeno of Elea",
  "n11406023": "Zephaniah, Sophonias",
  "n11406141": "Zeppelin, Count Ferdinand von Zeppelin",
  "n11406314": "Zhou En-lai, Chou En-lai",
  "n11406460": "Zhukov, Georgi Zhukov, Georgi Konstantinovich Zhukov",
  "n11406700": "Ziegfeld, Flo Ziegfeld, Florenz Ziegfeld",
  "n11406906": "Ziegler, Karl Waldemar Ziegler",
  "n11407048": "Zimbalist, Efrem Zimbalist",
  "n11407175": "Zinnemann, Fred Zinnemann",
  "n11407302": "Zinsser, Hans Zinsser",
  "n11407465": "Zinzendorf, Count Nikolaus Ludwig von Zinzendorf",
  "n11407591": "Zola, Emile Zola",
  "n11407715": "Zoroaster, Zarathustra",
  "n11407865": "Zsigmondy, Richard Adolph Zsigmondy",
  "n11408041": "Zukerman, Pinchas Zukerman",
  "n11408148": "Zweig, Stefan Zweig",
  "n11408243": "Zwingli, Ulrich Zwingli, Huldreich Zwingli",
  "n11408414": "Zworykin, Vladimir Kosma Zworykin",
  "n11408559": "natural phenomenon",
  "n11408733": "levitation",
  "n11408914": "metempsychosis, rebirth",
  "n11409059": "chemical phenomenon",
  "n11409329": "allotropy, allotropism",
  "n11409538": "exchange",
  "n11409689": "photoconductivity, photoconduction",
  "n11409892": "superconductivity",
  "n11410023": "photochemical exchange",
  "n11410172": "photoemission",
  "n11410298": "crystallization, crystallisation, crystallizing",
  "n11410527": "efflorescence, bloom",
  "n11410625": "consequence, effect, outcome, result, event, issue, upshot",
  "n11411501": "aftereffect",
  "n11411610": "aftermath, wake, backwash",
  "n11411839": "bandwagon effect",
  "n11412179": "brisance",
  "n11412334": "butterfly effect",
  "n11412592": "by-product, byproduct",
  "n11412727": "change",
  "n11412993": "coattails effect",
  "n11413263": "Coriolis effect",
  "n11413661": "dent",
  "n11413800": "dominance",
  "n11414041": "domino effect",
  "n11414257": "harvest",
  "n11414411": "impact, wallop",
  "n11414608": "influence",
  "n11414874": "perturbation",
  "n11415084": "variation",
  "n11415342": "purchase",
  "n11415492": "wind",
  "n11415608": "knock-on effect",
  "n11415721": "outgrowth, branch, offshoot, offset",
  "n11415842": "product",
  "n11416087": "placebo effect",
  "n11416534": "position effect",
  "n11416722": "repercussion, reverberation",
  "n11416988": "response",
  "n11417129": "reaction",
  "n11417387": "side effect, fallout",
  "n11417561": "engine",
  "n11417672": "geological phenomenon",
  "n11418011": "endogeny",
  "n11418138": "luck, fortune, chance, hazard",
  "n11418460": "luck, fortune",
  "n11418750": "organic phenomenon",
  "n11419404": "physical phenomenon",
  "n11420139": "aberration, distortion, optical aberration",
  "n11420376": "abiogenesis, autogenesis, autogeny, spontaneous generation",
  "n11420631": "absorption band",
  "n11420831": "spectrum",
  "n11421214": "absorption spectrum",
  "n11421401": "actinic radiation, actinic ray",
  "n11421618": "action spectrum",
  "n11421822": "activation energy, energy of activation",
  "n11422114": "aerodynamic force",
  "n11422277": "aerodynamic lift, lift",
  "n11422446": "ground effect",
  "n11422597": "aerosol",
  "n11422822": "affinity, chemical attraction",
  "n11423028": "air pocket, pocket, air hole",
  "n11423197": "slipstream, airstream, race, backwash, wash",
  "n11423356": "airstream",
  "n11423471": "alluvial fan, alluvial cone",
  "n11423607": "alpha radiation, alpha ray",
  "n11423756": "alpha rhythm, alpha wave",
  "n11423952": "alternating current, AC, alternating electric current",
  "n11424194": "alternation of generations, heterogenesis, xenogenesis",
  "n11424400": "alternative energy",
  "n11424589": "metagenesis, digenesis",
  "n11424704": "amperage",
  "n11424816": "annual ring, growth ring",
  "n11424938": "antitrades",
  "n11425088": "apparent motion, motion, apparent movement, movement",
  "n11425401": "acoustic phenomenon",
  "n11425580": "atmospheric phenomenon",
  "n11425989": "atomic energy, nuclear energy",
  "n11426125": "atomic power, nuclear power",
  "n11426288": "atomic spectrum",
  "n11426530": "attraction, attractive force",
  "n11426778": "affinity",
  "n11426914": "repulsion, repulsive force",
  "n11427067": "aureole, corona",
  "n11427241": "aurora",
  "n11427501": "aurora australis, southern lights",
  "n11427619": "aurora borealis, northern lights",
  "n11427736": "autofluorescence",
  "n11427842": "bad luck, mischance, mishap",
  "n11428023": "beam, beam of light, light beam, ray, ray of light, shaft, shaft of light, irradiation",
  "n11428379": "beam, ray, electron beam",
  "n11428567": "cathode ray",
  "n11428699": "beta radiation, beta ray, electron radiation",
  "n11428862": "beta rhythm, beta wave",
  "n11429057": "binding energy, separation energy",
  "n11429238": "bioelectricity",
  "n11429356": "bise, bize",
  "n11429458": "atmospheric pressure, air pressure, pressure",
  "n11429661": "black-body radiation, blackbody radiation",
  "n11429968": "blood pressure",
  "n11430483": "systolic pressure",
  "n11430659": "diastolic pressure",
  "n11430863": "arterial pressure",
  "n11431058": "venous pressure",
  "n11431191": "boundary layer",
  "n11431302": "brainwave, brain wave, cortical potential",
  "n11431617": "calm air, calm",
  "n11431754": "breeze, zephyr, gentle wind, air",
  "n11432155": "sea breeze",
  "n11432262": "breath",
  "n11432387": "light air",
  "n11432508": "light breeze",
  "n11432632": "gentle breeze",
  "n11432758": "moderate breeze",
  "n11432887": "fresh breeze",
  "n11433013": "strong breeze",
  "n11433140": "Brownian movement, Brownian motion, pedesis",
  "n11433297": "brush discharge",
  "n11433435": "candlelight, candle flame",
  "n11433546": "capacitance, electrical capacity, capacity",
  "n11433698": "elastance, electrical elastance",
  "n11433806": "capillarity, capillary action",
  "n11434016": "catastrophe, cataclysm",
  "n11434150": "nuclear winter",
  "n11434448": "continental drift",
  "n11434594": "centrifugal force",
  "n11434747": "centripetal force",
  "n11434899": "chaos",
  "n11435028": "charge, electric charge",
  "n11435358": "electrostatic charge",
  "n11435541": "electrostatic field",
  "n11435661": "pyroelectricity",
  "n11435871": "positive charge",
  "n11436019": "negative charge",
  "n11436163": "chatter mark",
  "n11436283": "chemical bond, bond",
  "n11436491": "valency",
  "n11436585": "cohesion",
  "n11436748": "covalent bond",
  "n11436929": "cross-link, cross-linkage",
  "n11437111": "hydrogen bond",
  "n11437344": "ionic bond, electrovalent bond, electrostatic bond",
  "n11437577": "ionizing radiation",
  "n11437823": "double bond",
  "n11437957": "coordinate bond, dative bond",
  "n11438105": "metallic bond",
  "n11438260": "peptide bond, peptide linkage",
  "n11438468": "chemical energy",
  "n11438612": "chinook, chinook wind, snow eater",
  "n11438756": "harmattan",
  "n11438904": "chromatic aberration",
  "n11439031": "circulation",
  "n11439312": "systemic circulation",
  "n11439446": "pulmonary circulation",
  "n11439566": "vitelline circulation",
  "n11439690": "cloud",
  "n11439924": "cold wave",
  "n11440012": "cold weather",
  "n11440123": "Coriolis force",
  "n11440521": "freeze, frost",
  "n11440691": "corona",
  "n11440802": "corona discharge, corona, corposant, St. Elmo's fire, Saint Elmo's fire, Saint Elmo's light, Saint Ulmo's fire, Saint Ulmo's light, electric glow",
  "n11441077": "cosmic background radiation, CBR, cosmic microwave background radiation, CMBR, cosmic microwave background, CMB",
  "n11441416": "cosmic dust",
  "n11441561": "cosmic radiation",
  "n11441707": "cosmic ray",
  "n11441980": "dust cloud",
  "n11442073": "mushroom, mushroom cloud, mushroom-shaped cloud",
  "n11442298": "counterglow, gegenschein",
  "n11442524": "crosswind",
  "n11442630": "fohn, foehn",
  "n11442752": "khamsin",
  "n11442894": "Santa Ana",
  "n11443057": "high wind",
  "n11443173": "headwind",
  "n11443283": "katabatic wind, catabatic wind",
  "n11443409": "tailwind",
  "n11443532": "current, electric current",
  "n11443721": "cyclone",
  "n11443929": "cyclosis, streaming",
  "n11444038": "daylight",
  "n11444117": "death",
  "n11444371": "decalescence",
  "n11444643": "decay, decomposition",
  "n11444816": "dehiscence",
  "n11445187": "delta rhythm, delta wave",
  "n11445395": "deposit, sedimentation, alluviation",
  "n11445564": "desquamation, peeling, shedding",
  "n11445753": "exfoliation",
  "n11445960": "mother lode, champion lode",
  "n11446067": "lode, load",
  "n11446242": "condensation, condensate",
  "n11446459": "sweat",
  "n11446598": "diapedesis",
  "n11446771": "dichroism",
  "n11446934": "diffraction",
  "n11447153": "direct current, DC, direct electric current",
  "n11447319": "signal",
  "n11447532": "interrupt",
  "n11447691": "doldrums",
  "n11447851": "drift, impetus, impulsion",
  "n11448013": "dust devil",
  "n11448153": "dust storm, duster, sandstorm, sirocco",
  "n11448343": "east wind, easter, easterly",
  "n11448475": "northwest wind, northwester",
  "n11448593": "southwester, sou'wester",
  "n11448714": "southeaster, sou'easter",
  "n11448835": "elastic energy, elastic potential energy",
  "n11449002": "electrical phenomenon",
  "n11449419": "electrical power, electric power, wattage",
  "n11449561": "load",
  "n11449658": "electric field",
  "n11449784": "dielectric heating",
  "n11449907": "electricity",
  "n11450453": "galvanism",
  "n11450566": "electricity, electrical energy",
  "n11450869": "electromagnetic radiation, electromagnetic wave, nonparticulate radiation",
  "n11451310": "Hertzian wave",
  "n11451442": "electromagnetic spectrum",
  "n11451944": "emission spectrum",
  "n11452079": "energy",
  "n11452218": "energy, free energy",
  "n11452750": "energy level, energy state",
  "n11453016": "power",
  "n11453210": "rest energy",
  "n11453420": "work",
  "n11453735": "epiphenomenon",
  "n11453860": "event",
  "n11454042": "facilitation",
  "n11454310": "flare",
  "n11454470": "flashover",
  "n11454591": "flood, inundation, deluge, alluvion",
  "n11454953": "debacle",
  "n11455092": "flash flood, flashflood",
  "n11455236": "floodhead",
  "n11455386": "Noah's flood, Noachian deluge, Noah and the Flood, the Flood",
  "n11455695": "focus, focal point",
  "n11455901": "food chain",
  "n11456083": "food pyramid",
  "n11456273": "food web, food cycle",
  "n11456462": "fair weather, sunshine, temperateness",
  "n11456615": "fata morgana",
  "n11456760": "field, field of force, force field",
  "n11457057": "line of force, field line",
  "n11457283": "electrical line of force",
  "n11457392": "magnetic line of force",
  "n11457496": "fine spray",
  "n11457586": "fine structure",
  "n11457841": "firelight",
  "n11457944": "firestorm",
  "n11458102": "fluorescence",
  "n11458314": "fog",
  "n11458514": "fogbank",
  "n11458624": "force",
  "n11459200": "forked lightning, chain lightning",
  "n11459369": "friar's lantern, ignis fatuus, jack-o'-lantern, will-o'-the-wisp",
  "n11459538": "friction, rubbing",
  "n11459748": "fringe, interference fringe",
  "n11459907": "gene expression",
  "n11460063": "grinding, abrasion, attrition, detrition",
  "n11460281": "grip, traction, adhesive friction",
  "n11460488": "front",
  "n11460728": "warm front",
  "n11460829": "cold front, polar front",
  "n11460962": "squall line",
  "n11461080": "occluded front, occlusion",
  "n11461268": "greenhouse effect, greenhouse warming",
  "n11461563": "inversion",
  "n11461684": "frost heave, frost heaving",
  "n11461825": "gale",
  "n11462013": "moderate gale, near gale",
  "n11462152": "fresh gale",
  "n11462276": "strong gale",
  "n11462401": "whole gale",
  "n11462526": "storm, violent storm",
  "n11462951": "northeaster, noreaster",
  "n11463073": "gamma radiation, gamma ray",
  "n11463265": "gaslight",
  "n11463371": "radiance, glow, glowing",
  "n11463544": "glow",
  "n11463647": "sky glow",
  "n11463746": "good luck, fluke, good fortune",
  "n11463895": "serendipity",
  "n11464027": "gravitational field",
  "n11464143": "gravity, gravitation, gravitational attraction, gravitational force",
  "n11464784": "solar gravity",
  "n11464926": "gun smoke",
  "n11465017": "gust, blast, blow",
  "n11465297": "bluster",
  "n11465425": "sandblast",
  "n11465530": "hail",
  "n11465688": "hailstorm",
  "n11465775": "half-light",
  "n11465888": "haze",
  "n11466043": "heat, heat energy",
  "n11466337": "geothermal energy",
  "n11466458": "histocompatibility",
  "n11466701": "hot weather",
  "n11466834": "scorcher",
  "n11466909": "sultriness",
  "n11467018": "hurricane",
  "n11467202": "hydroelectricity",
  "n11467318": "hysteresis",
  "n11467543": "ice fog, pogonip",
  "n11467650": "ice storm, silver storm",
  "n11467786": "incandescence, glow",
  "n11467996": "incidence",
  "n11468172": "induction, inductance",
  "n11468442": "mutual induction",
  "n11468578": "self-induction",
  "n11468763": "inertia",
  "n11468973": "moment of inertia",
  "n11469108": "angular acceleration",
  "n11469265": "angular velocity",
  "n11469481": "infrared, infrared light, infrared radiation, infrared emission",
  "n11469691": "infrared spectrum",
  "n11469827": "interreflection",
  "n11470008": "jet propulsion",
  "n11470139": "jet stream",
  "n11470348": "juice",
  "n11470492": "kinetic energy, K.E.",
  "n11470621": "heat of dissociation",
  "n11470765": "heat of formation",
  "n11470933": "heat of solution",
  "n11471097": "latent heat, heat of transformation",
  "n11471359": "heat of condensation",
  "n11471578": "heat of fusion",
  "n11471828": "heat of solidification",
  "n11471977": "heat of sublimation",
  "n11472131": "heat of vaporization, heat of vaporisation",
  "n11472352": "specific heat",
  "n11472503": "heat ray",
  "n11472612": "heat wave",
  "n11472699": "high beam",
  "n11472820": "infrared ray",
  "n11472960": "lamplight",
  "n11473033": "levanter",
  "n11473138": "leverage, purchase",
  "n11473291": "life",
  "n11473488": "biology",
  "n11473685": "aerobiosis",
  "n11473807": "life cycle",
  "n11473954": "light, visible light, visible radiation",
  "n11474637": "line",
  "n11474774": "elves",
  "n11475067": "jet, blue jet, reverse lightning",
  "n11475279": "lightning",
  "n11475476": "line spectrum",
  "n11475682": "Lorentz force",
  "n11475992": "sprites, red sprites",
  "n11476231": "atmospheric electricity",
  "n11476430": "luminescence",
  "n11476619": "bioluminescence",
  "n11476767": "chemiluminescence",
  "n11476939": "luminous energy",
  "n11477041": "magnetosphere",
  "n11477269": "solar magnetic field",
  "n11477384": "magnetic field, magnetic flux, flux",
  "n11477582": "radiation field",
  "n11477710": "beat",
  "n11477921": "resonance",
  "n11478171": "sympathetic vibration",
  "n11478299": "resonance",
  "n11478514": "nuclear resonance",
  "n11478682": "magnetic resonance",
  "n11478898": "nuclear magnetic resonance, NMR, proton magnetic resonance",
  "n11479058": "magnetism, magnetic attraction, magnetic force",
  "n11479368": "electromagnetism",
  "n11479640": "antiferromagnetism",
  "n11479816": "diamagnetism",
  "n11480091": "ferrimagnetism",
  "n11480284": "ferromagnetism",
  "n11480521": "paramagnetism",
  "n11480698": "mechanical phenomenon",
  "n11480930": "sound",
  "n11481209": "ultrasound",
  "n11481334": "trajectory, flight",
  "n11481487": "ballistics, ballistic trajectory",
  "n11481627": "gravity-assist",
  "n11481824": "mass defect, mass deficiency",
  "n11482013": "mechanical energy",
  "n11482140": "thaw, thawing, warming",
  "n11482312": "microwave",
  "n11482579": "midnight sun",
  "n11482706": "mist",
  "n11482873": "mistral",
  "n11482985": "moment",
  "n11483179": "moment of a couple",
  "n11483354": "dipole moment",
  "n11483472": "electric dipole moment",
  "n11483580": "magnetic dipole moment",
  "n11483829": "magnetic moment, moment of a magnet",
  "n11483990": "meteor, shooting star",
  "n11484260": "bolide, fireball",
  "n11484375": "mirage",
  "n11484570": "monsoon",
  "n11484741": "monsoon",
  "n11484861": "moonbeam, moon ray, moon-ray",
  "n11484975": "moonlight, moonshine, Moon",
  "n11485186": "starlight",
  "n11485264": "sunburst",
  "n11485367": "sunlight, sunshine, sun",
  "n11485582": "sunbeam, sunray",
  "n11485681": "laser beam",
  "n11485774": "low beam",
  "n11485907": "particle beam",
  "n11486049": "ion beam, ionic beam",
  "n11486178": "necrobiosis, cell death",
  "n11486381": "apoptosis, programmed cell death, caspase-mediated cell death",
  "n11486708": "necrosis, mortification, gangrene, sphacelus",
  "n11486983": "myonecrosis",
  "n11487078": "brain death, cerebral death",
  "n11487298": "neutron radiation",
  "n11487424": "halo",
  "n11487533": "solar halo, parhelic circle, parhelic ring",
  "n11487732": "parhelion, mock sun, sundog",
  "n11487950": "north wind, northerly, norther, boreas",
  "n11488160": "nuclear propulsion",
  "n11488387": "ocean current",
  "n11488601": "El Nino",
  "n11488828": "El Nino southern oscillation",
  "n11489070": "equatorial current",
  "n11489230": "North Equatorial Current",
  "n11489383": "South Equatorial Current",
  "n11489536": "Gulf stream",
  "n11489686": "Japan current, Kuroshio current, Kuroshio",
  "n11489878": "Peruvian current, Humboldt current",
  "n11490058": "opacity",
  "n11490257": "optical opacity",
  "n11490335": "radiopacity, radio-opacity",
  "n11490463": "optical illusion",
  "n11490638": "optical phenomenon",
  "n11490934": "pea soup, pea-souper",
  "n11491026": "phosphorescence",
  "n11491194": "photoelectricity",
  "n11491308": "piezoelectricity, piezoelectric effect, piezo effect",
  "n11491619": "pleochroism",
  "n11491816": "pleomorphism",
  "n11492014": "polarization, polarisation",
  "n11492240": "depolarization, depolarisation",
  "n11492388": "polymorphism",
  "n11492643": "dimorphism",
  "n11492833": "polymorphism, pleomorphism",
  "n11493083": "dimorphism",
  "n11493266": "polymorphism",
  "n11493452": "single nucleotide polymorphism, SNP",
  "n11493827": "electric potential, potential, potential difference, potential drop, voltage",
  "n11494076": "evoked potential",
  "n11494287": "resting potential",
  "n11494472": "potential energy, P.E.",
  "n11494638": "precipitation, downfall",
  "n11494935": "gas pressure",
  "n11495041": "pressure, pressure level, force per unit area",
  "n11495607": "head",
  "n11495708": "barometric pressure",
  "n11495822": "compartment pressure",
  "n11495964": "overpressure",
  "n11496157": "sea-level pressure",
  "n11496296": "hydrostatic head",
  "n11496503": "intraocular pressure, IOP",
  "n11496746": "oil pressure",
  "n11496881": "osmotic pressure",
  "n11497173": "radiation pressure, corpuscular-radiation pressure",
  "n11497378": "sound pressure, instantaneous sound pressure",
  "n11497586": "prevailing wind",
  "n11497777": "propulsion",
  "n11497888": "puff, puff of air, whiff",
  "n11498040": "pull",
  "n11498203": "push, thrust",
  "n11498461": "reaction",
  "n11498679": "rocket propulsion",
  "n11498850": "reaction propulsion",
  "n11499091": "radiant energy",
  "n11499284": "radiation",
  "n11499510": "corpuscular radiation, particulate radiation",
  "n11499817": "radio wave, radio emission, radio radiation",
  "n11500122": "sky wave",
  "n11500349": "ionospheric wave",
  "n11500458": "ground wave",
  "n11500570": "radio signal",
  "n11500679": "mass spectrum",
  "n11500816": "microwave spectrum",
  "n11500968": "radio spectrum, radio-frequency spectrum",
  "n11501230": "carrier wave, carrier",
  "n11501381": "rain, rainfall",
  "n11501649": "raindrop",
  "n11501737": "rainstorm",
  "n11501864": "line storm, equinoctial storm",
  "n11501988": "thundershower",
  "n11502102": "downpour, cloudburst, deluge, waterspout, torrent, pelter, soaker",
  "n11502322": "drizzle, mizzle",
  "n11502497": "shower, rain shower",
  "n11502695": "recognition",
  "n11503060": "reflection, reflexion",
  "n11503287": "refraction",
  "n11503482": "double refraction, birefringence",
  "n11503644": "resistance",
  "n11503813": "conductance",
  "n11503968": "electric resistance, electrical resistance, impedance, resistance, resistivity, ohmic resistance",
  "n11504225": "ohmage",
  "n11504313": "reactance",
  "n11504478": "acoustic resistance, acoustic impedance, acoustic reactance",
  "n11504750": "reluctance",
  "n11504898": "drag, retarding force",
  "n11505057": "sonic barrier, sound barrier",
  "n11505210": "windage",
  "n11505318": "rejection",
  "n11505546": "rejuvenation, greening",
  "n11505769": "resolving power, resolution",
  "n11505952": "resolution",
  "n11506167": "scattering, sprinkle, sprinkling",
  "n11506349": "scintillation",
  "n11506549": "sex linkage",
  "n11506738": "shear",
  "n11507000": "meteor shower, meteor stream",
  "n11507174": "short wave",
  "n11507321": "medium wave",
  "n11507511": "long wave",
  "n11507655": "simoom, simoon, samiel",
  "n11507797": "skin effect",
  "n11507951": "sleet",
  "n11508092": "smoke, fume",
  "n11508303": "smother",
  "n11508382": "snow, snowfall",
  "n11508578": "flurry, snow flurry",
  "n11508808": "whiteout",
  "n11509066": "snowflake, flake",
  "n11509204": "virga",
  "n11509377": "ice crystal, snow mist, diamond dust, poudrin, ice needle, frost snow, frost mist",
  "n11509570": "blizzard, snowstorm",
  "n11509697": "solar energy, solar power",
  "n11509968": "insolation",
  "n11510067": "solar radiation",
  "n11510223": "solar flare, flare",
  "n11510411": "solar prominence",
  "n11510571": "solar wind",
  "n11510714": "sound spectrum, acoustic spectrum",
  "n11510896": "speech spectrum",
  "n11511004": "sunspot, macula",
  "n11511176": "facula",
  "n11511327": "facula",
  "n11511405": "south wind, souther, southerly",
  "n11511523": "discharge, spark, arc, electric arc, electric discharge",
  "n11511765": "distortion",
  "n11512144": "nonlinear distortion, amplitude distortion",
  "n11512331": "projection, acoustic projection, sound projection",
  "n11512650": "electrical conduction",
  "n11512818": "conduction, conductivity",
  "n11512992": "propagation",
  "n11513179": "Doppler effect, Doppler shift",
  "n11513357": "red shift, redshift",
  "n11513631": "wave front",
  "n11513761": "spherical aberration",
  "n11513880": "spillover",
  "n11514008": "squall",
  "n11514171": "line squall",
  "n11514288": "electrical disturbance",
  "n11514462": "static electricity",
  "n11514559": "dynamic electricity, current electricity",
  "n11514672": "thermoelectricity",
  "n11514805": "stress",
  "n11515051": "tension",
  "n11515325": "strain",
  "n11515526": "breaking point",
  "n11515644": "overstrain",
  "n11515734": "streamer",
  "n11515849": "torchlight",
  "n11515935": "twilight",
  "n11516113": "interaction, fundamental interaction",
  "n11516439": "electromagnetic interaction",
  "n11516659": "gravitational interaction",
  "n11516819": "strong interaction, strong force, color force",
  "n11517041": "supertwister",
  "n11517210": "weak interaction, weak force",
  "n11517494": "suction",
  "n11517656": "sunrise",
  "n11517776": "sunset",
  "n11517898": "afterglow",
  "n11517999": "surface tension",
  "n11518168": "interfacial tension, interfacial surface tension",
  "n11518330": "syzygy",
  "n11518494": "tempest",
  "n11518645": "thermal",
  "n11518743": "thermionic current",
  "n11518915": "theta rhythm, theta wave",
  "n11519121": "thunderbolt, bolt, bolt of lightning",
  "n11519253": "thunderstorm, electrical storm, electric storm",
  "n11519450": "tornado, twister",
  "n11519702": "torsion, torque",
  "n11519799": "tossup, toss-up, even chance",
  "n11519949": "trade wind, trade",
  "n11520120": "antitrade wind, antitrade",
  "n11520271": "tramontane, tramontana",
  "n11520435": "transgression",
  "n11520619": "transparency, transparence",
  "n11520821": "trichroism",
  "n11520989": "turbulence, turbulency",
  "n11521145": "typhoon",
  "n11521267": "turbulent flow",
  "n11521404": "sea",
  "n11521534": "head sea",
  "n11521665": "streamline flow",
  "n11521824": "laminar flow",
  "n11521940": "ultraviolet, ultraviolet radiation, ultraviolet light, ultraviolet illumination, UV",
  "n11522206": "sunray, sun-ray",
  "n11522325": "ultraviolet spectrum",
  "n11522448": "draft, draught",
  "n11522649": "updraft",
  "n11522730": "downdraft",
  "n11522815": "van der Waal's forces",
  "n11523031": "vapor pressure, vapour pressure",
  "n11523256": "virtual image",
  "n11523369": "visible spectrum, color spectrum",
  "n11523538": "voltage, electromotive force, emf",
  "n11523736": "magnetomotive force",
  "n11523839": "life force, vital force, vitality, elan vital",
  "n11524110": "volcanism",
  "n11524213": "waterpower",
  "n11524324": "waterspout",
  "n11524451": "wave",
  "n11524662": "weather, weather condition, conditions, atmospheric condition",
  "n11525303": "elements",
  "n11525480": "west wind, wester",
  "n11525614": "prevailing westerly, westerly",
  "n11525779": "whirlwind",
  "n11525955": "wind, air current, current of air",
  "n11526894": "wind generation, wind power",
  "n11527014": "windstorm",
  "n11527177": "X ray, X-ray, X-radiation, roentgen ray",
  "n11527386": "X-ray diffraction",
  "n11527556": "zodiacal light",
  "n11527767": "chop",
  "n11527967": "flotation, floatation",
  "n11528225": "parallax",
  "n11528427": "Tyndall effect",
  "n11528636": "heliocentric parallax, annual parallax",
  "n11528841": "stellar parallax",
  "n11528939": "geocentric parallax, diurnal parallax",
  "n11529158": "horizontal parallax",
  "n11529295": "pulsation",
  "n11529441": "solar parallax",
  "n11529603": "Plantae, kingdom Plantae, plant kingdom",
  "n11530008": "microflora",
  "n11530149": "plant cell",
  "n11530283": "cell wall",
  "n11530512": "crop",
  "n11530715": "endemic",
  "n11530860": "holophyte",
  "n11530990": "non-flowering plant",
  "n11531090": "plantlet",
  "n11531193": "wilding",
  "n11531334": "semi-climber",
  "n11531457": "Thallophyta",
  "n11531701": "thallophyte",
  "n11531916": "button",
  "n11532017": "thallus",
  "n11532194": "crustose thallus",
  "n11532351": "cap, pileus",
  "n11532547": "calyptra",
  "n11532682": "volva",
  "n11532816": "ascocarp",
  "n11533026": "acervulus",
  "n11533212": "basidiocarp",
  "n11533365": "peridium",
  "n11533472": "ascoma",
  "n11533622": "apothecium",
  "n11533772": "cleistothecium, cleistocarp",
  "n11533999": "domatium",
  "n11534161": "podetium",
  "n11534360": "seta",
  "n11534434": "Tracheophyta, division Tracheophyta",
  "n11534677": "plant order",
  "n11536087": "ornamental",
  "n11536230": "pot plant",
  "n11536369": "acrogen",
  "n11536567": "apomict",
  "n11536673": "aquatic",
  "n11536778": "Bryophyta, division Bryophyta",
  "n11537327": "bryophyte, nonvascular plant",
  "n11537506": "moss",
  "n11537665": "moss family",
  "n11537886": "moss genus",
  "n11538123": "Anthoceropsida, class Anthoceropsida",
  "n11538341": "Anthocerotales, order Anthocerotales",
  "n11538582": "Anthocerotaceae, family Anthocerotaceae",
  "n11538716": "Anthoceros, genus Anthoceros",
  "n11538820": "hornwort",
  "n11538935": "Bryopsida, class Bryopsida, Musci, class Musci",
  "n11539289": "acrocarp, acrocarpous moss",
  "n11539467": "pleurocarp, pleurocarpous moss",
  "n11539675": "Andreaeales, order Andreaeales",
  "n11539825": "Andreaea, genus Andreaea",
  "n11540000": "Bryales, order Bryales",
  "n11540230": "Dicranales, order Dicranales",
  "n11540439": "Dicranaceae, family Dicranaceae",
  "n11540631": "Dicranum, genus Dicranum",
  "n11540747": "Eubryales, order Eubryales",
  "n11540970": "Bryaceae, family Bryaceae",
  "n11541111": "Bryum, genus Bryum",
  "n11541322": "Mniaceae, family Mniaceae",
  "n11541579": "Mnium, genus Mnium",
  "n11541713": "Sphagnales, order Sphagnales",
  "n11541919": "genus Sphagnum",
  "n11542137": "sphagnum, sphagnum moss, peat moss, bog moss",
  "n11542341": "Hepaticopsida, class Hepaticopsida, Hepaticae, class Hepaticae",
  "n11542640": "liverwort, hepatic",
  "n11542920": "Jungermanniales, order Jungermanniales",
  "n11543105": "leafy liverwort, scale moss",
  "n11543264": "Jungermanniaceae, family Jungermanniaceae",
  "n11543429": "Marchantiales, order Marchantiales",
  "n11543602": "Marchantiaceae, family Marchantiaceae",
  "n11543792": "Marchantia, genus Marchantia",
  "n11544015": "hepatica, Marchantia polymorpha",
  "n11544131": "Sphaerocarpales, order Sphaerocarpales",
  "n11544314": "Sphaerocarpaceae, family Sphaerocarpaceae",
  "n11544540": "Sphaerocarpus, genus Sphaerocarpus, Sphaerocarpos, genus Sphaerocarpos",
  "n11544769": "Pteridophyta, division Pteridophyta",
  "n11545214": "genus Pecopteris",
  "n11545350": "pecopteris",
  "n11545524": "pteridophyte, nonflowering plant",
  "n11545714": "fern",
  "n11547562": "fern ally",
  "n11547737": "agamete",
  "n11547855": "spore",
  "n11548465": "basidiospore",
  "n11548594": "endospore",
  "n11548728": "carpospore",
  "n11548870": "chlamydospore",
  "n11549009": "conidium, conidiospore",
  "n11549138": "conidiophore",
  "n11549245": "oospore",
  "n11549391": "oosphere",
  "n11549487": "resting spore",
  "n11549638": "teliospore",
  "n11549779": "tetraspore",
  "n11549895": "zoospore",
  "n11550022": "fern seed",
  "n11550205": "fructification",
  "n11550340": "gleba",
  "n11550474": "hymenium",
  "n11550620": "pycnidium",
  "n11550725": "sporocarp, spore case",
  "n11550890": "stipule",
  "n11551044": "tepal",
  "n11551211": "Spermatophyta, division Spermatophyta",
  "n11551659": "Phanerogamae",
  "n11551898": "Cryptogamia",
  "n11552133": "cryptogam",
  "n11552386": "spermatophyte, phanerogam, seed plant",
  "n11552594": "seedling",
  "n11552686": "balsam",
  "n11552806": "annual",
  "n11552976": "biennial",
  "n11553240": "perennial",
  "n11553419": "escape",
  "n11553522": "hygrophyte",
  "n11553634": "neophyte",
  "n11553763": "gymnosperm family",
  "n11554175": "gymnosperm genus",
  "n11555413": "monocot family, liliopsid family",
  "n11556187": "liliid monocot family",
  "n11556857": "monocot genus, liliopsid genus",
  "n11561228": "liliid monocot genus",
  "n11562747": "dicot family, magnoliopsid family",
  "n11564258": "magnoliid dicot family",
  "n11564734": "hamamelid dicot family",
  "n11565040": "caryophylloid dicot family",
  "n11565385": "dilleniid dicot family",
  "n11566230": "asterid dicot family",
  "n11566682": "rosid dicot family",
  "n11567411": "dicot genus, magnoliopsid genus",
  "n11571907": "magnoliid dicot genus",
  "n11573173": "hamamelid dicot genus",
  "n11573660": "caryophylloid dicot genus",
  "n11575425": "dilleniid dicot genus",
  "n11579418": "asterid dicot genus",
  "n11585340": "rosid dicot genus",
  "n11590783": "fungus family",
  "n11592146": "fungus genus",
  "n11594676": "fungus order",
  "n11595312": "Gymnospermae, class Gymnospermae, Gymnospermophyta, division Gymnospermophyta",
  "n11596108": "gymnosperm",
  "n11596344": "progymnosperm",
  "n11596486": "Gnetopsida, class Gnetopsida, Gnetophytina, subdivision Gnetophytina, Gnetophyta",
  "n11596845": "Gnetales, order Gnetales",
  "n11597126": "Gnetaceae, family Gnetaceae",
  "n11597396": "genus Gnetum",
  "n11597657": "gnetum, Gnetum gnemon",
  "n11597924": "Ephedraceae, family Ephedraceae",
  "n11598100": "Catha, genus Catha",
  "n11598287": "Catha edulis",
  "n11598452": "genus Ephedra",
  "n11598686": "ephedra, joint fir",
  "n11598886": "mahuang, Ephedra sinica",
  "n11598991": "Welwitschiaceae, family Welwitschiaceae",
  "n11599165": "genus Welwitschia, genus Welwitchia",
  "n11599324": "welwitschia, Welwitschia mirabilis",
  "n11599694": "Cycadopsida, class Cycadopsida, Cycadophytina, subdivision Cycadophytina, Cycadophyta, subdivision Cycadophyta",
  "n11600139": "Cycadales, order Cycadales",
  "n11600372": "cycad",
  "n11600671": "Cycadaceae, family Cycadaceae, cycad family",
  "n11600900": "Cycas, genus Cycas",
  "n11601177": "sago palm, Cycas revoluta",
  "n11601333": "false sago, fern palm, Cycas circinalis",
  "n11601487": "Zamiaceae, family Zamiaceae, zamia family",
  "n11601757": "genus Zamia",
  "n11601918": "zamia",
  "n11602091": "coontie, Florida arrowroot, Seminole bread, Zamia pumila",
  "n11602304": "genus Ceratozamia",
  "n11602478": "ceratozamia",
  "n11602671": "genus Dioon",
  "n11602873": "dioon",
  "n11603045": "genus Encephalartos",
  "n11603246": "encephalartos",
  "n11603462": "kaffir bread, Encephalartos caffer",
  "n11603630": "genus Macrozamia",
  "n11603835": "macrozamia",
  "n11604046": "burrawong, Macrozamia communis, Macrozamia spiralis",
  "n11604225": "Bennettitales, order Bennettitales",
  "n11604393": "Bennettitaceae, family Bennettitaceae",
  "n11604576": "Bennettitis, genus Bennettitis",
  "n11604698": "Pteridospermopsida, class Pteridospermopsida",
  "n11604904": "Cycadofilicales, order Cycadofilicales, Lyginopteridales, order Lyginopteridales",
  "n11605147": "Pteridospermae, group Pteridospermae, Pteridospermaphyta, group Pteridospermaphyta",
  "n11605396": "Lyginopteris, genus Lyginopteris",
  "n11605542": "seed fern, pteridosperm",
  "n11605708": "Coniferopsida, class Coniferopsida, Coniferophytina, subdivision Coniferophytina, Coniferophyta",
  "n11606379": "Cordaitales, order Cordaitales",
  "n11606661": "Cordaitaceae, family Cordaitaceae",
  "n11606846": "Cordaites, genus Cordaites",
  "n11607071": "Pinopsida, class Pinopsida, Pinophytina, subdivision Pinophytina",
  "n11607392": "Coniferales, order Coniferales",
  "n11607739": "Pinaceae, family Pinaceae, pine family",
  "n11608055": "Pinus, genus Pinus",
  "n11608250": "pine, pine tree, true pine",
  "n11608885": "pine",
  "n11609122": "knotty pine",
  "n11609251": "white pine",
  "n11609362": "yellow pine",
  "n11609475": "pinon, pinyon",
  "n11609684": "nut pine",
  "n11609862": "pinon pine, Mexican nut pine, Pinus cembroides",
  "n11610047": "Rocky mountain pinon, Pinus edulis",
  "n11610215": "single-leaf, single-leaf pine, single-leaf pinyon, Pinus monophylla",
  "n11610437": "bishop pine, bishop's pine, Pinus muricata",
  "n11610602": "California single-leaf pinyon, Pinus californiarum",
  "n11610823": "Parry's pinyon, Pinus quadrifolia, Pinus parryana",
  "n11611087": "spruce pine, Pinus glabra",
  "n11611233": "black pine, Pinus nigra",
  "n11611356": "pitch pine, northern pitch pine, Pinus rigida",
  "n11611561": "pond pine, Pinus serotina",
  "n11611758": "stone pine, umbrella pine, European nut pine, Pinus pinea",
  "n11612018": "Swiss pine, Swiss stone pine, arolla pine, cembra nut tree, Pinus cembra",
  "n11612235": "cembra nut, cedar nut",
  "n11612349": "Swiss mountain pine, mountain pine, dwarf mountain pine, mugho pine, mugo pine, Pinus mugo",
  "n11612575": "ancient pine, Pinus longaeva",
  "n11612923": "white pine",
  "n11613219": "American white pine, eastern white pine, weymouth pine, Pinus strobus",
  "n11613459": "western white pine, silver pine, mountain pine, Pinus monticola",
  "n11613692": "southwestern white pine, Pinus strobiformis",
  "n11613867": "limber pine, Pinus flexilis",
  "n11614039": "whitebark pine, whitebarked pine, Pinus albicaulis",
  "n11614250": "yellow pine",
  "n11614420": "ponderosa, ponderosa pine, western yellow pine, bull pine, Pinus ponderosa",
  "n11614713": "Jeffrey pine, Jeffrey's pine, black pine, Pinus jeffreyi",
  "n11615026": "shore pine, lodgepole, lodgepole pine, spruce pine, Pinus contorta",
  "n11615259": "Sierra lodgepole pine, Pinus contorta murrayana",
  "n11615387": "loblolly pine, frankincense pine, Pinus taeda",
  "n11615607": "jack pine, Pinus banksiana",
  "n11615812": "swamp pine",
  "n11615967": "longleaf pine, pitch pine, southern yellow pine, Georgia pine, Pinus palustris",
  "n11616260": "shortleaf pine, short-leaf pine, shortleaf yellow pine, Pinus echinata",
  "n11616486": "red pine, Canadian red pine, Pinus resinosa",
  "n11616662": "Scotch pine, Scots pine, Scotch fir, Pinus sylvestris",
  "n11616852": "scrub pine, Virginia pine, Jersey pine, Pinus virginiana",
  "n11617090": "Monterey pine, Pinus radiata",
  "n11617272": "bristlecone pine, Rocky Mountain bristlecone pine, Pinus aristata",
  "n11617631": "table-mountain pine, prickly pine, hickory pine, Pinus pungens",
  "n11617878": "knobcone pine, Pinus attenuata",
  "n11618079": "Japanese red pine, Japanese table pine, Pinus densiflora",
  "n11618290": "Japanese black pine, black pine, Pinus thunbergii",
  "n11618525": "Torrey pine, Torrey's pine, soledad pine, grey-leaf pine, sabine pine, Pinus torreyana",
  "n11618750": "Larix, genus Larix",
  "n11618861": "larch, larch tree",
  "n11619136": "larch",
  "n11619227": "American larch, tamarack, black larch, Larix laricina",
  "n11619455": "western larch, western tamarack, Oregon larch, Larix occidentalis",
  "n11619687": "subalpine larch, Larix lyallii",
  "n11619845": "European larch, Larix decidua",
  "n11620016": "Siberian larch, Larix siberica, Larix russica",
  "n11620248": "Pseudolarix, genus Pseudolarix",
  "n11620389": "golden larch, Pseudolarix amabilis",
  "n11620560": "Abies, genus Abies",
  "n11620673": "fir, fir tree, true fir",
  "n11620912": "fir",
  "n11621029": "silver fir",
  "n11621281": "amabilis fir, white fir, Pacific silver fir, red silver fir, Christmas tree, Abies amabilis",
  "n11621547": "European silver fir, Christmas tree, Abies alba",
  "n11621727": "white fir, Colorado fir, California white fir, Abies concolor, Abies lowiana",
  "n11621950": "balsam fir, balm of Gilead, Canada balsam, Abies balsamea",
  "n11622184": "Fraser fir, Abies fraseri",
  "n11622368": "lowland fir, lowland white fir, giant fir, grand fir, Abies grandis",
  "n11622591": "Alpine fir, subalpine fir, Abies lasiocarpa",
  "n11622771": "Santa Lucia fir, bristlecone fir, Abies bracteata, Abies venusta",
  "n11622988": "Cedrus, genus Cedrus",
  "n11623105": "cedar, cedar tree, true cedar",
  "n11623304": "cedar, cedarwood",
  "n11623556": "red cedar",
  "n11623685": "pencil cedar",
  "n11623815": "cedar of Lebanon, Cedrus libani",
  "n11623967": "deodar, deodar cedar, Himalayan cedar, Cedrus deodara",
  "n11624192": "Atlas cedar, Cedrus atlantica",
  "n11624367": "Picea, genus Picea",
  "n11624531": "spruce",
  "n11624840": "spruce",
  "n11625003": "Norway spruce, Picea abies",
  "n11625223": "weeping spruce, Brewer's spruce, Picea breweriana",
  "n11625391": "Engelmann spruce, Engelmann's spruce, Picea engelmannii",
  "n11625632": "white spruce, Picea glauca",
  "n11625804": "black spruce, Picea mariana, spruce pine",
  "n11626010": "Siberian spruce, Picea obovata",
  "n11626152": "Sitka spruce, Picea sitchensis",
  "n11626409": "oriental spruce, Picea orientalis",
  "n11626585": "Colorado spruce, Colorado blue spruce, silver spruce, Picea pungens",
  "n11626826": "red spruce, eastern spruce, yellow spruce, Picea rubens",
  "n11627028": "Tsuga, genus Tsuga",
  "n11627168": "hemlock, hemlock tree",
  "n11627364": "hemlock",
  "n11627512": "eastern hemlock, Canadian hemlock, spruce pine, Tsuga canadensis",
  "n11627714": "Carolina hemlock, Tsuga caroliniana",
  "n11627908": "mountain hemlock, black hemlock, Tsuga mertensiana",
  "n11628087": "western hemlock, Pacific hemlock, west coast hemlock, Tsuga heterophylla",
  "n11628284": "Pseudotsuga, genus Pseudotsuga",
  "n11628456": "douglas fir",
  "n11628678": "douglas fir",
  "n11628793": "green douglas fir, douglas spruce, douglas pine, douglas hemlock, Oregon fir, Oregon pine, Pseudotsuga menziesii",
  "n11629047": "big-cone spruce, big-cone douglas fir, Pseudotsuga macrocarpa",
  "n11629211": "genus Cathaya",
  "n11629354": "Cathaya",
  "n11629501": "Cupressaceae, family Cupressaceae, cypress family",
  "n11630017": "cedar, cedar tree",
  "n11630351": "Cupressus, genus Cupressus",
  "n11630489": "cypress, cypress tree",
  "n11630890": "cypress",
  "n11631159": "gowen cypress, Cupressus goveniana",
  "n11631405": "pygmy cypress, Cupressus pigmaea, Cupressus goveniana pigmaea",
  "n11631619": "Santa Cruz cypress, Cupressus abramsiana, Cupressus goveniana abramsiana",
  "n11631854": "Arizona cypress, Cupressus arizonica",
  "n11631985": "Guadalupe cypress, Cupressus guadalupensis",
  "n11632167": "Monterey cypress, Cupressus macrocarpa",
  "n11632376": "Mexican cypress, cedar of Goa, Portuguese cypress, Cupressus lusitanica",
  "n11632619": "Italian cypress, Mediterranean cypress, Cupressus sempervirens",
  "n11632794": "Athrotaxis, genus Athrotaxis",
  "n11632929": "King William pine, Athrotaxis selaginoides",
  "n11633116": "Austrocedrus, genus Austrocedrus",
  "n11633284": "Chilean cedar, Austrocedrus chilensis",
  "n11633459": "Callitris, genus Callitris",
  "n11633633": "cypress pine",
  "n11633863": "Port Jackson pine, Callitris cupressiformis",
  "n11633999": "black cypress pine, red cypress pine, Callitris endlicheri, Callitris calcarata",
  "n11634243": "white cypress pine, Callitris glaucophylla, Callitris glauca",
  "n11634393": "stringybark pine, Callitris parlatorei",
  "n11634526": "Calocedrus, genus Calocedrus",
  "n11634736": "incense cedar, red cedar, Calocedrus decurrens, Libocedrus decurrens",
  "n11634970": "Chamaecyparis, genus Chamaecyparis",
  "n11635152": "southern white cedar, coast white cedar, Atlantic white cedar, white cypress, white cedar, Chamaecyparis thyoides",
  "n11635433": "Oregon cedar, Port Orford cedar, Lawson's cypress, Lawson's cedar, Chamaecyparis lawsoniana",
  "n11635709": "Port Orford cedar",
  "n11635830": "yellow cypress, yellow cedar, Nootka cypress, Alaska cedar, Chamaecyparis nootkatensis",
  "n11636068": "Cryptomeria, genus Cryptomeria",
  "n11636204": "Japanese cedar, Japan cedar, sugi, Cryptomeria japonica",
  "n11636389": "Juniperus, genus Juniperus",
  "n11636566": "juniper",
  "n11636835": "juniper berry",
  "n11637015": "pencil cedar, pencil cedar tree",
  "n11637247": "eastern red cedar, red cedar, red juniper, Juniperus virginiana",
  "n11637482": "Bermuda cedar, Juniperus bermudiana",
  "n11637659": "east African cedar, Juniperus procera",
  "n11637810": "southern red cedar, Juniperus silicicola",
  "n11637991": "dwarf juniper, savin, Juniperus sabina",
  "n11638109": "common juniper, Juniperus communis",
  "n11638378": "ground cedar, dwarf juniper, Juniperus communis depressa",
  "n11638525": "creeping juniper, Juniperus horizontalis",
  "n11638698": "Mexican juniper, drooping juniper, Juniperus flaccida",
  "n11638902": "Libocedrus, genus Libocedrus",
  "n11639084": "incense cedar",
  "n11639306": "kawaka, Libocedrus plumosa",
  "n11639445": "pahautea, Libocedrus bidwillii, mountain pine",
  "n11639609": "Taxodiaceae, subfamily Taxodiaceae, redwood family",
  "n11639863": "genus Metasequoia",
  "n11640132": "metasequoia, dawn redwood, Metasequoia glyptostrodoides",
  "n11640471": "genus Sequoia",
  "n11640645": "sequoia, redwood",
  "n11640898": "redwood",
  "n11641034": "California redwood, coast redwood, Sequoia sempervirens",
  "n11641275": "Sequoiadendron, genus Sequoiadendron",
  "n11641494": "giant sequoia, big tree, Sierra redwood, Sequoiadendron giganteum, Sequoia gigantea, Sequoia Wellingtonia",
  "n11641788": "Taxodium, genus Taxodium",
  "n11641963": "bald cypress, swamp cypress, pond bald cypress, southern cypress, Taxodium distichum",
  "n11642243": "pond cypress, bald cypress, Taxodium ascendens",
  "n11642430": "Montezuma cypress, Mexican swamp cypress, Taxodium mucronatum",
  "n11642622": "Ahuehuete, Tule tree",
  "n11642912": "Tetraclinis, genus Tetraclinis",
  "n11643022": "sandarac, sandarac tree, Tetraclinis articulata, Callitris quadrivalvis",
  "n11643354": "sandarac, sandarach",
  "n11643506": "sandarac, citronwood",
  "n11643684": "Thuja, genus Thuja",
  "n11643835": "arborvitae",
  "n11644046": "western red cedar, red cedar, canoe cedar, Thuja plicata",
  "n11644226": "American arborvitae, northern white cedar, white cedar, Thuja occidentalis",
  "n11644462": "Oriental arborvitae, Thuja orientalis, Platycladus orientalis",
  "n11644712": "Thujopsis, genus Thujopsis",
  "n11644872": "hiba arborvitae, Thujopsis dolobrata",
  "n11645041": "genus Keteleeria",
  "n11645163": "keteleeria",
  "n11645271": "Araucariaceae, family Araucariaceae, araucaria family",
  "n11645590": "Wollemi pine",
  "n11645783": "genus Araucaria",
  "n11645914": "araucaria",
  "n11646167": "monkey puzzle, chile pine, Araucaria araucana",
  "n11646344": "norfolk island pine, Araucaria heterophylla, Araucaria excelsa",
  "n11646517": "new caledonian pine, Araucaria columnaris",
  "n11646694": "bunya bunya, bunya bunya tree, Araucaria bidwillii",
  "n11646955": "hoop pine, Moreton Bay pine, Araucaria cunninghamii",
  "n11647131": "Agathis, genus Agathis",
  "n11647306": "kauri pine, dammar pine",
  "n11647548": "kauri",
  "n11647703": "kauri, kaury, Agathis australis",
  "n11647868": "amboina pine, amboyna pine, Agathis dammara, Agathis alba",
  "n11648039": "dundathu pine, queensland kauri, smooth bark kauri, Agathis robusta",
  "n11648268": "red kauri, Agathis lanceolata",
  "n11648428": "Cephalotaxaceae, family Cephalotaxaceae, plum-yew family",
  "n11648617": "Cephalotaxus, genus Cephalotaxus",
  "n11648776": "plum-yew",
  "n11649012": "Torreya, genus Torreya",
  "n11649150": "California nutmeg, nutmeg-yew, Torreya californica",
  "n11649359": "stinking cedar, stinking yew, Torrey tree, Torreya taxifolia",
  "n11649597": "Phyllocladaceae, family Phyllocladaceae",
  "n11649749": "Phyllocladus, genus Phyllocladus",
  "n11649878": "celery pine",
  "n11650160": "celery top pine, celery-topped pine, Phyllocladus asplenifolius",
  "n11650307": "tanekaha, Phyllocladus trichomanoides",
  "n11650430": "Alpine celery pine, Phyllocladus alpinus",
  "n11650558": "yellowwood, yellowwood tree",
  "n11650759": "gymnospermous yellowwood",
  "n11650919": "angiospermous yellowwood",
  "n11651133": "yellowwood",
  "n11651259": "Podocarpaceae, family Podocarpaceae, podocarpus family",
  "n11651731": "Podocarpus, genus Podocarpus",
  "n11652039": "podocarp",
  "n11652217": "yacca, yacca podocarp, Podocarpus coriaceus",
  "n11652376": "brown pine, Rockingham podocarp, Podocarpus elatus",
  "n11652578": "cape yellowwood, African yellowwood, Podocarpus elongatus",
  "n11652753": "South-African yellowwood, Podocarpus latifolius",
  "n11652966": "alpine totara, Podocarpus nivalis",
  "n11653126": "totara, Podocarpus totara",
  "n11653323": "Afrocarpus, genus Afrocarpus",
  "n11653570": "common yellowwood, bastard yellowwood, Afrocarpus falcata",
  "n11653728": "Dacrycarpus, genus Dacrycarpus",
  "n11653904": "kahikatea, New Zealand Dacryberry, New Zealand white pine, Dacrycarpus dacrydioides, Podocarpus dacrydioides",
  "n11654124": "Dacrydium, genus Dacrydium",
  "n11654293": "rimu, imou pine, red pine, Dacrydium cupressinum",
  "n11654438": "tarwood, tar-wood, Dacrydium colensoi",
  "n11654667": "Falcatifolium, genus Falcatifolium",
  "n11654984": "common sickle pine, Falcatifolium falciforme",
  "n11655152": "yellow-leaf sickle pine, Falcatifolium taxoides",
  "n11655407": "Halocarpus, genus Halocarpus",
  "n11655592": "tarwood, tar-wood, New Zealand mountain pine, Halocarpus bidwilli, Dacrydium bidwilli",
  "n11655764": "Lagarostrobus, genus Lagarostrobus",
  "n11655974": "westland pine, silver pine, Lagarostrobus colensoi",
  "n11656123": "huon pine, Lagarostrobus franklinii, Dacrydium franklinii",
  "n11656380": "Lepidothamnus, genus Lepidothamnus",
  "n11656549": "Chilean rimu, Lepidothamnus fonkii",
  "n11656771": "mountain rimu, Lepidothamnus laxifolius, Dacridium laxifolius",
  "n11656974": "Microstrobos, genus Microstrobos",
  "n11657153": "Tasman dwarf pine, Microstrobos niphophilus",
  "n11657314": "Nageia, genus Nageia",
  "n11657585": "nagi, Nageia nagi",
  "n11657763": "Parasitaxus, genus Parasitaxus",
  "n11657904": "parasite yew, Parasitaxus ustus",
  "n11658104": "Prumnopitys, genus Prumnopitys",
  "n11658331": "miro, black pine, Prumnopitys ferruginea, Podocarpus ferruginea",
  "n11658544": "matai, black pine, Prumnopitys taxifolia, Podocarpus spicata",
  "n11658709": "plum-fruited yew, Prumnopitys andina, Prumnopitys elegans",
  "n11658872": "Retrophyllum, genus Retrophyllum",
  "n11659068": "Saxe-gothea, Saxegothea, genus Saxe-gothea, genus Saxegothea",
  "n11659248": "Prince Albert yew, Prince Albert's yew, Saxe-gothea conspicua",
  "n11659500": "Sundacarpus, genus Sundacarpus",
  "n11659627": "Sundacarpus amara, Prumnopitys amara, Podocarpus amara",
  "n11659909": "Sciadopityaceae, family Sciadopityaceae",
  "n11660121": "Sciadopitys, genus Sciadopitys",
  "n11660300": "Japanese umbrella pine, Sciadopitys verticillata",
  "n11660537": "Taxopsida, class Taxopsida, Taxophytina, subdivision Taxophytina",
  "n11660848": "Taxales, order Taxales",
  "n11660979": "Taxaceae, family Taxaceae, yew family",
  "n11661207": "Taxus, genus Taxus",
  "n11661372": "yew",
  "n11661707": "yew",
  "n11661909": "Old World yew, English yew, Taxus baccata",
  "n11662128": "Pacific yew, California yew, western yew, Taxus brevifolia",
  "n11662371": "Japanese yew, Taxus cuspidata",
  "n11662585": "Florida yew, Taxus floridana",
  "n11662764": "Austrotaxus, genus Austrotaxus",
  "n11662937": "New Caledonian yew, Austrotaxus spicata",
  "n11663136": "Pseudotaxus, genus Pseudotaxus",
  "n11663263": "white-berry yew, Pseudotaxus chienii",
  "n11663449": "Ginkgopsida, class Ginkgopsida, Ginkgophytina, class Ginkgophytina, subdivision Ginkgophytina, subdivision Ginkgophyta",
  "n11663813": "Ginkgoales, order Ginkgoales",
  "n11664090": "Ginkgoaceae, family Ginkgoaceae, ginkgo family",
  "n11664301": "genus Ginkgo",
  "n11664418": "ginkgo, gingko, maidenhair tree, Ginkgo biloba",
  "n11664677": "Pteropsida, subdivision Pteropsida",
  "n11664929": "Angiospermae, class Angiospermae, Magnoliophyta, division Magnoliophyta, Anthophyta, division Anthophyta",
  "n11665372": "angiosperm, flowering plant",
  "n11665648": "angiocarp",
  "n11665781": "Dicotyledones, class Dicotyledones, Dicotyledonae, class Dicotyledonae, Magnoliopsida, class Magnoliopsida",
  "n11666854": "dicot, dicotyledon, magnoliopsid, exogen",
  "n11667112": "Magnoliidae, subclass Magnoliidae, ranalian complex",
  "n11667562": "Monocotyledones, class Monocotyledones, Monocotyledonae, class Monocotyledonae, Liliopsida, class Liliopsida",
  "n11668117": "monocot, monocotyledon, liliopsid, endogen",
  "n11668340": "Alismatidae, subclass Alismatidae",
  "n11668573": "Arecidae, subclass Arecidae",
  "n11668952": "Commelinidae, subclass Commelinidae",
  "n11669335": "flower, bloom, blossom",
  "n11669786": "floret, floweret",
  "n11669921": "flower",
  "n11672269": "bloomer",
  "n11672400": "wildflower, wild flower",
  "n11674019": "apetalous flower",
  "n11674105": "flower head",
  "n11674332": "inflorescence",
  "n11674603": "ray flower, ray floret",
  "n11674798": "catkin, ament",
  "n11674914": "bud",
  "n11675025": "rosebud",
  "n11675096": "stamen",
  "n11675245": "anther",
  "n11675404": "gynostegium",
  "n11675537": "pollen",
  "n11675738": "pollinium",
  "n11675842": "reproductive structure",
  "n11676500": "pistil",
  "n11676743": "gynobase",
  "n11676850": "gynophore",
  "n11676963": "simple pistil",
  "n11677045": "compound pistil",
  "n11677144": "pistillode",
  "n11677259": "style",
  "n11677485": "stylopodium",
  "n11677603": "stigma",
  "n11677743": "carpel",
  "n11677902": "carpophore",
  "n11678010": "cornstalk, corn stalk",
  "n11678123": "filament",
  "n11678199": "funicle, funiculus",
  "n11678299": "petiolule",
  "n11678377": "mericarp",
  "n11678493": "hilum",
  "n11678635": "ovary",
  "n11678768": "ovule",
  "n11679085": "chalaza",
  "n11679250": "nucellus",
  "n11679378": "micropyle",
  "n11679528": "amphitropous ovule",
  "n11679652": "anatropous ovule",
  "n11679779": "campylotropous ovule",
  "n11679909": "orthotropous ovule",
  "n11680032": "stoma, stomate, pore",
  "n11680277": "germ pore",
  "n11680457": "germ tube",
  "n11680596": "pollen tube",
  "n11680838": "placenta",
  "n11680995": "placentation",
  "n11681255": "apical placentation",
  "n11681393": "axile placentation",
  "n11681554": "basal placentation",
  "n11681692": "free central placentation",
  "n11681865": "lamellate placentation",
  "n11682003": "marginal placentation, ventral placentation",
  "n11682166": "parietal placentation",
  "n11682349": "testa, episperm, seed coat",
  "n11682512": "endosperm",
  "n11682659": "gemma",
  "n11682842": "cone, strobilus, strobile",
  "n11683105": "fir cone",
  "n11683216": "galbulus",
  "n11683331": "pinecone",
  "n11683443": "septum",
  "n11683556": "shell",
  "n11683747": "nutshell",
  "n11683838": "nectary, honey gland",
  "n11683989": "seed",
  "n11684264": "pericarp, seed vessel",
  "n11684499": "epicarp, exocarp",
  "n11684654": "mesocarp",
  "n11684739": "stone, pit, endocarp",
  "n11685091": "pip",
  "n11685179": "capsule",
  "n11685400": "bilocular capsule",
  "n11685512": "boll",
  "n11685621": "silique, siliqua",
  "n11685766": "silicle",
  "n11685876": "peristome",
  "n11686049": "haustorium",
  "n11686195": "cataphyll",
  "n11686398": "cotyledon, seed leaf",
  "n11686503": "embryo",
  "n11686652": "perisperm",
  "n11686780": "monocarp, monocarpic plant, monocarpous plant",
  "n11686912": "sporophyte",
  "n11687071": "gametophyte",
  "n11687266": "megagametophyte",
  "n11687432": "megasporangium, macrosporangium",
  "n11687553": "megasporophyll",
  "n11687681": "microgametophyte",
  "n11687789": "microspore",
  "n11687964": "microsporangium",
  "n11688069": "microsporophyll",
  "n11688199": "megaspore, macrospore",
  "n11688378": "archespore, archesporium",
  "n11688552": "daughter cell",
  "n11688750": "mother cell",
  "n11688958": "spore mother cell",
  "n11689054": "archegonium",
  "n11689197": "bonduc nut, nicker nut, nicker seed",
  "n11689367": "Job's tears",
  "n11689483": "oilseed, oil-rich seed",
  "n11689678": "castor bean",
  "n11689815": "cottonseed",
  "n11689957": "candlenut",
  "n11690088": "peach pit",
  "n11690169": "cherry stone",
  "n11690254": "hypanthium, floral cup, calyx tube",
  "n11690455": "petal, flower petal",
  "n11690612": "sepal",
  "n11690737": "mentum",
  "n11690893": "floral leaf",
  "n11691046": "corolla",
  "n11691332": "corona",
  "n11691523": "calyx",
  "n11691857": "lip",
  "n11691990": "hull",
  "n11692108": "epicalyx, false calyx, calycle, calyculus",
  "n11692265": "perianth, chlamys, floral envelope, perigone, perigonium",
  "n11692604": "pappus",
  "n11692792": "thistledown",
  "n11692952": "Ranales, order Ranales, Ranunculales, order Ranunculales",
  "n11693566": "Annonaceae, family Annonaceae, custard-apple family",
  "n11693812": "Annona, genus Annona",
  "n11693981": "custard apple, custard apple tree",
  "n11694300": "cherimoya, cherimoya tree, Annona cherimola",
  "n11694469": "ilama, ilama tree, Annona diversifolia",
  "n11694664": "soursop, prickly custard apple, soursop tree, Annona muricata",
  "n11694866": "bullock's heart, bullock's heart tree, bullock heart, Annona reticulata",
  "n11695085": "sweetsop, sweetsop tree, Annona squamosa",
  "n11695285": "pond apple, pond-apple tree, Annona glabra",
  "n11695485": "Asimina, genus Asimina",
  "n11695599": "pawpaw, papaw, papaw tree, Asimina triloba",
  "n11695813": "Cananga, genus Cananga, Canangium, genus Canangium",
  "n11695974": "ilang-ilang, ylang-ylang, Cananga odorata",
  "n11696211": "ilang-ilang",
  "n11696338": "Oxandra, genus Oxandra",
  "n11696450": "lancewood, lancewood tree, Oxandra lanceolata",
  "n11696606": "lancewood",
  "n11696776": "Xylopia, genus Xylopia",
  "n11696935": "Guinea pepper, negro pepper, Xylopia aethiopica",
  "n11697158": "Berberidaceae, family Berberidaceae, barberry family",
  "n11697388": "Berberis, genus Berberis",
  "n11697560": "barberry",
  "n11697802": "American barberry, Berberis canadensis",
  "n11698042": "common barberry, European barberry, Berberis vulgaris",
  "n11698245": "Japanese barberry, Berberis thunbergii",
  "n11698433": "Caulophyllum, genus Caulophyllum",
  "n11698562": "blue cohosh, blueberry root, papooseroot, papoose root, squawroot, squaw root, Caulophyllum thalictrioides, Caulophyllum thalictroides",
  "n11698895": "Epimedium, genus Epimedium",
  "n11699071": "barrenwort, bishop's hat, Epimedium grandiflorum",
  "n11699283": "Mahonia, genus Mahonia",
  "n11699442": "Oregon grape, Oregon holly grape, hollygrape, mountain grape, holly-leaves barberry, Mahonia aquifolium",
  "n11699751": "Oregon grape, Mahonia nervosa",
  "n11699915": "Podophyllum, genus Podophyllum",
  "n11700058": "mayapple, May apple, wild mandrake, Podophyllum peltatum",
  "n11700279": "May apple",
  "n11700401": "Calycanthaceae, family Calycanthaceae, calycanthus family, strawberry-shrub family",
  "n11700676": "Calycanthus, genus Calycanthus",
  "n11700864": "allspice",
  "n11701066": "Carolina allspice, strawberry shrub, strawberry bush, sweet shrub, Calycanthus floridus",
  "n11701302": "spicebush, California allspice, Calycanthus occidentalis",
  "n11701492": "Chimonanthus, genus Chimonanthus",
  "n11701698": "Japan allspice, Japanese allspice, winter sweet, Chimonanthus praecox",
  "n11701903": "Ceratophyllaceae, family Ceratophyllaceae",
  "n11702081": "Ceratophyllum, genus Ceratophyllum",
  "n11702252": "hornwort",
  "n11702428": "Cercidiphyllaceae, family Cercidiphyllaceae",
  "n11702566": "Cercidiphyllum, genus Cercidiphyllum",
  "n11702713": "katsura tree, Cercidiphyllum japonicum",
  "n11702999": "Lardizabalaceae, family Lardizabalaceae, lardizabala family",
  "n11703205": "Lardizabala, genus Lardizabala",
  "n11703386": "Lauraceae, family Lauraceae, laurel family",
  "n11703669": "laurel",
  "n11703935": "Laurus, genus Laurus",
  "n11704093": "true laurel, bay, bay laurel, bay tree, Laurus nobilis",
  "n11704401": "Cinnamomum, genus Cinnamomum",
  "n11704620": "camphor tree, Cinnamomum camphora",
  "n11704791": "cinnamon, Ceylon cinnamon, Ceylon cinnamon tree, Cinnamomum zeylanicum",
  "n11705052": "cinnamon, cinnamon bark",
  "n11705171": "cassia, cassia-bark tree, Cinnamomum cassia",
  "n11705387": "cassia bark, Chinese cinnamon",
  "n11705573": "Saigon cinnamon, Cinnamomum loureirii",
  "n11705776": "cinnamon bark",
  "n11705921": "Lindera, genus Lindera",
  "n11706120": "Benzoin, genus Benzoin",
  "n11706325": "spicebush, spice bush, American spicebush, Benjamin bush, Lindera benzoin, Benzoin odoriferum",
  "n11706629": "Persea, genus Persea",
  "n11706761": "avocado, avocado tree, Persea Americana",
  "n11706942": "laurel-tree, red bay, Persea borbonia",
  "n11707109": "genus Sassafras",
  "n11707229": "sassafras, sassafras tree, Sassafras albidum",
  "n11707511": "sassafras oil",
  "n11707668": "Umbellularia, genus Umbellularia",
  "n11707827": "California laurel, California bay tree, Oregon myrtle, pepperwood, spice tree, sassafras laurel, California olive, mountain laurel, Umbellularia californica",
  "n11708181": "Magnoliaceae, family Magnoliaceae, magnolia family",
  "n11708442": "Illicium, genus Illicium",
  "n11708658": "anise tree",
  "n11708857": "purple anise, Illicium floridanum",
  "n11709045": "star anise, Illicium anisatum",
  "n11709205": "star anise, Chinese anise, Illicium verum",
  "n11709450": "genus Magnolia",
  "n11709674": "magnolia",
  "n11710008": "magnolia",
  "n11710136": "southern magnolia, evergreen magnolia, large-flowering magnolia, bull bay, Magnolia grandiflora",
  "n11710393": "umbrella tree, umbrella magnolia, elkwood, elk-wood, Magnolia tripetala",
  "n11710658": "earleaved umbrella tree, Magnolia fraseri",
  "n11710827": "cucumber tree, Magnolia acuminata",
  "n11710987": "large-leaved magnolia, large-leaved cucumber tree, great-leaved macrophylla, Magnolia macrophylla",
  "n11711289": "saucer magnolia, Chinese magnolia, Magnolia soulangiana",
  "n11711537": "star magnolia, Magnolia stellata",
  "n11711764": "sweet bay, swamp bay, swamp laurel, Magnolia virginiana",
  "n11711971": "manglietia, genus Manglietia",
  "n11712153": "Liriodendron, genus Liriodendron",
  "n11712282": "tulip tree, tulip poplar, yellow poplar, canary whitewood, Liriodendron tulipifera",
  "n11712621": "tulipwood, true tulipwood, whitewood, white poplar, yellow poplar",
  "n11712827": "Menispermaceae, family Menispermaceae, moonseed family",
  "n11713034": "Menispermum, genus Menispermum",
  "n11713164": "moonseed",
  "n11713370": "common moonseed, Canada moonseed, yellow parilla, Menispermum canadense",
  "n11713628": "Cocculus, genus Cocculus",
  "n11713763": "Carolina moonseed, Cocculus carolinus",
  "n11713960": "Myristicaceae, family Myristicaceae, nutmeg family",
  "n11714150": "Myristica, genus Myristica",
  "n11714382": "nutmeg, nutmeg tree, Myristica fragrans",
  "n11714618": "Nymphaeaceae, family Nymphaeaceae, water-lily family",
  "n11714853": "water lily",
  "n11715207": "Nymphaea, genus Nymphaea",
  "n11715430": "water nymph, fragrant water lily, pond lily, Nymphaea odorata",
  "n11715678": "European white lily, Nymphaea alba",
  "n11715810": "lotus, white lotus, Egyptian water lily, white lily, Nymphaea lotus",
  "n11716041": "blue lotus, Nymphaea caerulea",
  "n11716167": "blue lotus, Nymphaea stellata",
  "n11716285": "Nuphar, genus Nuphar",
  "n11716422": "spatterdock, cow lily, yellow pond lily, Nuphar advena",
  "n11716698": "southern spatterdock, Nuphar sagittifolium",
  "n11716877": "yellow water lily, Nuphar lutea",
  "n11717007": "Nelumbonaceae, subfamily Nelumbonaceae",
  "n11717239": "Nelumbo, genus Nelumbo",
  "n11717399": "lotus, Indian lotus, sacred lotus, Nelumbo nucifera",
  "n11717577": "water chinquapin, American lotus, yanquapin, Nelumbo lutea",
  "n11717820": "Cabombaceae, subfamily Cabombaceae, water-shield family",
  "n11718096": "Cabomba, genus Cabomba",
  "n11718296": "water-shield, fanwort, Cabomba caroliniana",
  "n11718521": "Brasenia, genus Brasenia",
  "n11718681": "water-shield, Brasenia schreberi, water-target",
  "n11718911": "Paeoniaceae, family Paeoniaceae, peony family",
  "n11719120": "Paeonia, genus Paeonia",
  "n11719286": "peony, paeony",
  "n11719468": "Ranunculaceae, family Ranunculaceae, buttercup family, crowfoot family",
  "n11720088": "Ranunculus, genus Ranunculus",
  "n11720353": "buttercup, butterflower, butter-flower, crowfoot, goldcup, kingcup",
  "n11720643": "meadow buttercup, tall buttercup, tall crowfoot, tall field buttercup, Ranunculus acris",
  "n11720891": "water crowfoot, water buttercup, Ranunculus aquatilis",
  "n11721124": "common buttercup, Ranunculus bulbosus",
  "n11721337": "lesser celandine, pilewort, Ranunculus ficaria",
  "n11721642": "lesser spearwort, Ranunculus flammula",
  "n11721844": "sagebrush buttercup, Ranunculus glaberrimus",
  "n11722036": "greater spearwort, Ranunculus lingua",
  "n11722199": "mountain lily, Mount Cook lily, Ranunculus lyalii",
  "n11722342": "western buttercup, Ranunculus occidentalis",
  "n11722466": "creeping buttercup, creeping crowfoot, Ranunculus repens",
  "n11722621": "cursed crowfoot, celery-leaved buttercup, Ranunculus sceleratus",
  "n11722769": "Aconitum, genus Aconitum",
  "n11722982": "aconite",
  "n11723227": "monkshood, helmetflower, helmet flower, Aconitum napellus",
  "n11723452": "wolfsbane, wolfbane, wolf's bane, Aconitum lycoctonum",
  "n11723655": "Actaea, genus Actaea",
  "n11723770": "baneberry, cohosh, herb Christopher",
  "n11723986": "baneberry",
  "n11724109": "red baneberry, redberry, red-berry, snakeberry, Actaea rubra",
  "n11724363": "white baneberry, white cohosh, white bead, doll's eyes, Actaea alba",
  "n11724529": "Adonis, genus Adonis",
  "n11724660": "pheasant's-eye, Adonis annua",
  "n11724822": "genus Anemone",
  "n11725015": "anemone, windflower",
  "n11725311": "Alpine anemone, mountain anemone, Anemone tetonensis",
  "n11725480": "Canada anemone, Anemone Canadensis",
  "n11725623": "thimbleweed, Anemone cylindrica",
  "n11725821": "wood anemone, Anemone nemorosa",
  "n11725973": "wood anemone, snowdrop, Anemone quinquefolia",
  "n11726145": "longheaded thimbleweed, Anemone riparia",
  "n11726269": "snowdrop anemone, snowdrop windflower, Anemone sylvestris",
  "n11726433": "Virginia thimbleweed, Anemone virginiana",
  "n11726569": "Anemonella, genus Anemonella",
  "n11726707": "rue anemone, Anemonella thalictroides",
  "n11726925": "genus Aquilegia",
  "n11727091": "columbine, aquilegia, aquilege",
  "n11727358": "meeting house, honeysuckle, Aquilegia canadensis",
  "n11727540": "blue columbine, Aquilegia caerulea, Aquilegia scopulorum calcarea",
  "n11727738": "granny's bonnets, Aquilegia vulgaris",
  "n11727976": "Caltha, genus Caltha",
  "n11728099": "marsh marigold, kingcup, meadow bright, May blob, cowslip, water dragon, Caltha palustris",
  "n11728350": "Cimicifuga, genus Cimicifuga",
  "n11728530": "bugbane",
  "n11728769": "American bugbane, summer cohosh, Cimicifuga americana",
  "n11728945": "black cohosh, black snakeroot, rattle-top, Cimicifuga racemosa",
  "n11729142": "fetid bugbane, foetid bugbane, Cimicifuga foetida",
  "n11729315": "genus Clematis",
  "n11729478": "clematis",
  "n11729860": "pine hyacinth, Clematis baldwinii, Viorna baldwinii",
  "n11730015": "blue jasmine, blue jessamine, curly clematis, marsh clematis, Clematis crispa",
  "n11730205": "pipestem clematis, Clematis lasiantha",
  "n11730312": "curly-heads, Clematis ochreleuca",
  "n11730458": "golden clematis, Clematis tangutica",
  "n11730602": "scarlet clematis, Clematis texensis",
  "n11730750": "leather flower, Clematis versicolor",
  "n11730933": "leather flower, vase-fine, vase vine, Clematis viorna",
  "n11731157": "virgin's bower, old man's beard, devil's darning needle, Clematis virginiana",
  "n11731411": "traveler's joy, traveller's joy, old man's beard, Clematis vitalba",
  "n11731659": "purple clematis, purple virgin's bower, mountain clematis, Clematis verticillaris",
  "n11731861": "Coptis, genus Coptis",
  "n11732052": "goldthread, golden thread, Coptis groenlandica, Coptis trifolia groenlandica",
  "n11732309": "Consolida, genus Consolida",
  "n11732567": "rocket larkspur, Consolida ambigua, Delphinium ajacis",
  "n11732857": "genus Delphinium",
  "n11733054": "delphinium",
  "n11733312": "larkspur",
  "n11733424": "Eranthis, genus Eranthis",
  "n11733548": "winter aconite, Eranthis hyemalis",
  "n11733769": "Helleborus, genus Helleborus",
  "n11733904": "hellebore",
  "n11734093": "stinking hellebore, bear's foot, setterwort, Helleborus foetidus",
  "n11734300": "Christmas rose, winter rose, black hellebore, Helleborus niger",
  "n11734493": "lenten rose, black hellebore, Helleborus orientalis",
  "n11734698": "green hellebore, Helleborus viridis",
  "n11734872": "genus Hepatica",
  "n11735053": "hepatica, liverleaf",
  "n11735325": "Hydrastis, genus Hydrastis",
  "n11735570": "goldenseal, golden seal, yellow root, turmeric root, Hydrastis Canadensis",
  "n11735822": "Isopyrum, genus Isopyrum",
  "n11735977": "false rue anemone, false rue, Isopyrum biternatum",
  "n11736216": "Laccopetalum, genus Laccopetalum",
  "n11736362": "giant buttercup, Laccopetalum giganteum",
  "n11736569": "genus Nigella",
  "n11736694": "nigella",
  "n11736851": "love-in-a-mist, Nigella damascena",
  "n11737009": "fennel flower, Nigella hispanica",
  "n11737125": "black caraway, nutmeg flower, Roman coriander, Nigella sativa",
  "n11737316": "Pulsatilla, genus Pulsatilla",
  "n11737534": "pasqueflower, pasque flower",
  "n11737752": "American pasqueflower, Eastern pasque flower, wild crocus, lion's beard, prairie anemone, blue tulip, American pulsatilla, Pulsatilla patens, Anemone ludoviciana",
  "n11738063": "Western pasqueflower, Pulsatilla occidentalis, Anemone occidentalis",
  "n11738203": "European pasqueflower, Pulsatilla vulgaris, Anemone pulsatilla",
  "n11738378": "Thalictrum, genus Thalictrum",
  "n11738547": "meadow rue",
  "n11738832": "Trautvetteria, genus Trautvetteria",
  "n11738997": "false bugbane, Trautvetteria carolinensis",
  "n11739199": "Trollius, genus Trollius",
  "n11739365": "globeflower, globe flower",
  "n11739530": "Winteraceae, family Winteraceae, winter's bark family",
  "n11739809": "Drimys, genus Drimys",
  "n11739978": "winter's bark, winter's bark tree, Drimys winteri",
  "n11740208": "Pseudowintera, genus Pseudowintera, Wintera, genus Wintera",
  "n11740414": "pepper shrub, Pseudowintera colorata, Wintera colorata",
  "n11740655": "Myricales, order Myricales",
  "n11740824": "Myricaceae, family Myricaceae, wax-myrtle family",
  "n11741010": "Myrica, genus Myrica",
  "n11741175": "sweet gale, Scotch gale, Myrica gale",
  "n11741350": "wax myrtle",
  "n11741575": "bay myrtle, puckerbush, Myrica cerifera",
  "n11741797": "bayberry, candleberry, swamp candleberry, waxberry, Myrica pensylvanica",
  "n11742003": "bayberry wax, bayberry tallow",
  "n11742175": "Comptonia, genus Comptonia",
  "n11742310": "sweet fern, Comptonia peregrina, Comptonia asplenifolia",
  "n11742531": "Leitneriaceae, family Leitneriaceae, corkwood family",
  "n11742745": "Leitneria, genus Leitneria",
  "n11742878": "corkwood, corkwood tree, Leitneria floridana",
  "n11743109": "Juncaceae, family Juncaceae, rush family",
  "n11743294": "rush",
  "n11743570": "Juncus, genus Juncus",
  "n11743772": "bulrush, bullrush, common rush, soft rush, Juncus effusus",
  "n11744011": "jointed rush, Juncus articulatus",
  "n11744108": "toad rush, Juncus bufonius",
  "n11744251": "hard rush, Juncus inflexus",
  "n11744355": "salt rush, Juncus leseurii",
  "n11744471": "slender rush, Juncus tenuis",
  "n11744583": "plant family",
  "n11744859": "plant genus",
  "n11745817": "zebrawood, zebrawood tree",
  "n11746060": "zebrawood",
  "n11746224": "Connaraceae, family Connaraceae, zebrawood family",
  "n11746419": "Connarus, genus Connarus",
  "n11746600": "Connarus guianensis",
  "n11746776": "Leguminosae, family Leguminosae, Fabaceae, family Fabaceae, legume family, pea family",
  "n11747468": "legume, leguminous plant",
  "n11748002": "legume",
  "n11748330": "Arachis, genus Arachis",
  "n11748501": "peanut, peanut vine, Arachis hypogaea",
  "n11748811": "peanut",
  "n11748936": "Brya, genus Brya",
  "n11749112": "granadilla tree, granadillo, Brya ebenus",
  "n11749273": "cocuswood, cocoswood, granadilla wood",
  "n11749462": "Centrolobium, genus Centrolobium",
  "n11749603": "arariba, Centrolobium robustum",
  "n11749742": "Coumarouna, genus Coumarouna, Dipteryx, genus Dipteryx",
  "n11749920": "tonka bean, tonka bean tree, Coumarouna odorata, Dipteryx odorata",
  "n11750173": "tonka bean, coumara nut",
  "n11750359": "Hymenaea, genus Hymenaea",
  "n11750508": "courbaril, Hymenaea courbaril",
  "n11750745": "courbaril copal",
  "n11750855": "genus Melilotus",
  "n11750989": "melilotus, melilot, sweet clover",
  "n11751213": "white sweet clover, white melilot, Melilotus alba",
  "n11751347": "yellow sweet clover, Melilotus officinalis",
  "n11751598": "Swainsona, genus Swainsona",
  "n11751765": "darling pea, poison bush",
  "n11751974": "smooth darling pea, Swainsona galegifolia",
  "n11752168": "hairy darling pea, Swainsona greyana, Swainsona grandiflora",
  "n11752404": "Trifolium, genus Trifolium",
  "n11752578": "clover, trefoil",
  "n11752798": "alpine clover, Trifolium alpinum",
  "n11752937": "hop clover, shamrock, lesser yellow trefoil, Trifolium dubium",
  "n11753143": "crimson clover, Italian clover, Trifolium incarnatum",
  "n11753355": "red clover, purple clover, Trifolium pratense",
  "n11753562": "buffalo clover, Trifolium reflexum, Trifolium stoloniferum",
  "n11753700": "white clover, dutch clover, shamrock, Trifolium repens",
  "n11753936": "Mimosaceae, family Mimosaceae",
  "n11754188": "Mimosoideae, subfamily Mimosoideae",
  "n11754633": "genus Mimosa",
  "n11754893": "mimosa",
  "n11755110": "sensitive plant, Mimosa sensitiva",
  "n11755319": "sensitive plant, touch-me-not, shame plant, live-and-die, humble plant, action plant, Mimosa pudica",
  "n11755694": "genus Acacia",
  "n11756092": "acacia",
  "n11756329": "shittah, shittah tree",
  "n11756522": "shittimwood",
  "n11756669": "wattle",
  "n11756870": "black wattle, Acacia auriculiformis",
  "n11757017": "gidgee, stinking wattle, Acacia cambegei",
  "n11757190": "catechu, Jerusalem thorn, Acacia catechu",
  "n11757433": "black catechu, catechu",
  "n11757653": "silver wattle, mimosa, Acacia dealbata",
  "n11757851": "huisache, cassie, mimosa bush, sweet wattle, sweet acacia, scented wattle, flame tree, Acacia farnesiana",
  "n11758122": "lightwood, Acacia melanoxylon",
  "n11758276": "golden wattle, Acacia pycnantha",
  "n11758483": "fever tree, Acacia xanthophloea",
  "n11758628": "Adenanthera, genus Adenanthera",
  "n11758799": "coralwood, coral-wood, red sandalwood, Barbados pride, peacock flower fence, Adenanthera pavonina",
  "n11759049": "genus Albizia, genus Albizzia",
  "n11759224": "albizzia, albizia",
  "n11759404": "silk tree, Albizia julibrissin, Albizzia julibrissin",
  "n11759609": "siris, siris tree, Albizia lebbeck, Albizzia lebbeck",
  "n11759853": "rain tree, saman, monkeypod, monkey pod, zaman, zamang, Albizia saman",
  "n11760128": "Anadenanthera, genus Anadenanthera",
  "n11760294": "Anadenanthera colubrina, Piptadenia macrocarpa",
  "n11760560": "genus Calliandra",
  "n11760785": "calliandra",
  "n11761007": "Enterolobium, genus Enterolobium",
  "n11761202": "conacaste, elephant's ear, Enterolobium cyclocarpa",
  "n11761484": "genus Inga",
  "n11761650": "inga",
  "n11761836": "ice-cream bean, Inga edulis",
  "n11762018": "guama, Inga laurina",
  "n11762237": "Leucaena, genus Leucaena",
  "n11762433": "lead tree, white popinac, Leucaena glauca, Leucaena leucocephala",
  "n11762706": "Lysiloma, genus Lysiloma",
  "n11762927": "wild tamarind, Lysiloma latisiliqua, Lysiloma bahamensis",
  "n11763142": "sabicu, Lysiloma sabicu",
  "n11763340": "sabicu, sabicu wood",
  "n11763473": "Parkia, genus Parkia",
  "n11763625": "nitta tree",
  "n11763874": "Parkia javanica",
  "n11764072": "Piptadenia, genus Piptadenia",
  "n11764231": "Pithecellobium, genus Pithecellobium, Pithecolobium, genus Pithecolobium",
  "n11764478": "manila tamarind, camachile, huamachil, wild tamarind, Pithecellobium dulce",
  "n11764814": "cat's-claw, catclaw, black bead, Pithecellodium unguis-cati",
  "n11765099": "Prosopis, genus Prosopis",
  "n11765277": "mesquite, mesquit",
  "n11765568": "honey mesquite, Western honey mesquite, Prosopis glandulosa",
  "n11765859": "algarroba, Prosopis juliflora, Prosopis juliiflora",
  "n11766046": "algarroba, algarrobilla, algarobilla",
  "n11766189": "screw bean, screwbean, tornillo, screwbean mesquite, Prosopis pubescens",
  "n11766432": "screw bean",
  "n11766609": "Apocynaceae, family Apocynaceae, dogbane family",
  "n11767196": "Apocynum, genus Apocynum",
  "n11767354": "dogbane",
  "n11767630": "common dogbane, spreading dogbane, rheumatism weed, Apocynum androsaemifolium",
  "n11767877": "Indian hemp, rheumatism weed, Apocynum cannabinum",
  "n11768108": "Rocky Mountain dogbane, Apocynum pumilum",
  "n11768242": "Acocanthera, genus Acocanthera, Acokanthera, genus Acokanthera",
  "n11768505": "winter sweet, poison arrow plant, Acocanthera oblongifolia, Acocanthera spectabilis",
  "n11768816": "bushman's poison, ordeal tree, Acocanthera oppositifolia, Acocanthera venenata",
  "n11769002": "Adenium, genus Adenium",
  "n11769176": "impala lily, mock azalia, desert rose, kudu lily, Adenium obesum, Adenium multiflorum",
  "n11769483": "genus Allamanda",
  "n11769621": "allamanda",
  "n11769803": "common allamanda, golden trumpet, Allamanda cathartica",
  "n11770013": "Alstonia, genus Alstonia",
  "n11770256": "dita, dita bark, devil tree, Alstonia scholaris",
  "n11770526": "Amsonia, genus Amsonia",
  "n11770753": "blue star, Amsonia tabernaemontana",
  "n11770969": "Beaumontia, genus Beaumontia",
  "n11771147": "Nepal trumpet flower, Easter lily vine, Beaumontia grandiflora",
  "n11771383": "genus Carissa",
  "n11771539": "carissa",
  "n11771746": "hedge thorn, natal plum, Carissa bispinosa",
  "n11771924": "natal plum, amatungulu, Carissa macrocarpa, Carissa grandiflora",
  "n11772154": "Catharanthus, genus Catharanthus",
  "n11772408": "periwinkle, rose periwinkle, Madagascar periwinkle, old maid, Cape periwinkle, red periwinkle, cayenne jasmine, Catharanthus roseus, Vinca rosea",
  "n11772702": "Holarrhena, genus Holarrhena",
  "n11772879": "ivory tree, conessi, kurchi, kurchee, Holarrhena pubescens, Holarrhena antidysenterica",
  "n11773138": "Mandevilla, genus Mandevilla, Dipladenia, genus Dipladenia",
  "n11773408": "white dipladenia, Mandevilla boliviensis, Dipladenia boliviensis",
  "n11773628": "Chilean jasmine, Mandevilla laxa",
  "n11773860": "Nerium, genus Nerium",
  "n11773987": "oleander, rose bay, Nerium oleander",
  "n11774279": "Plumeria, genus Plumeria, Plumiera",
  "n11774513": "frangipani, frangipanni",
  "n11774795": "pagoda tree, temple tree, Plumeria acutifolia",
  "n11774972": "West Indian jasmine, pagoda tree, Plumeria alba",
  "n11775160": "genus Rauwolfia, genus Rauvolfia",
  "n11775340": "rauwolfia, rauvolfia",
  "n11775626": "snakewood, Rauwolfia serpentina",
  "n11775780": "genus Strophanthus",
  "n11775959": "strophanthus",
  "n11776234": "Strophanthus kombe",
  "n11776337": "Tabernaemontana, genus Tabernaemontana",
  "n11776511": "crape jasmine, crepe jasmine, crepe gardenia, pinwheel flower, East Indian rosebay, Adam's apple, Nero's crown, coffee rose, Tabernaemontana divaricate",
  "n11776861": "Thevetia, genus Thevetia",
  "n11777080": "yellow oleander, Thevetia peruviana, Thevetia neriifolia",
  "n11777365": "Trachelospermum, genus Trachelospermum",
  "n11777552": "star jasmine, confederate jasmine, Trachelospermum jasminoides",
  "n11777779": "Vinca, genus Vinca",
  "n11777929": "periwinkle",
  "n11778092": "myrtle, Vinca minor",
  "n11778257": "large periwinkle, Vinca major",
  "n11778391": "Arales, order Arales",
  "n11778534": "Araceae, family Araceae, arum family",
  "n11779300": "arum, aroid",
  "n11779801": "genus Arum",
  "n11780018": "arum",
  "n11780148": "cuckoopint, lords-and-ladies, jack-in-the-pulpit, Arum maculatum",
  "n11780424": "black calla, Arum palaestinum",
  "n11780589": "Acorus, genus Acorus",
  "n11780747": "Acoraceae, subfamily Acoraceae",
  "n11780930": "sweet flag, calamus, sweet calamus, myrtle flag, flagroot, Acorus calamus",
  "n11781176": "calamus",
  "n11781301": "calamus oil",
  "n11781430": "Aglaonema, genus Aglaonema",
  "n11781658": "Chinese evergreen, Japanese leaf, Aglaonema modestum",
  "n11781850": "genus Alocasia",
  "n11782036": "alocasia, elephant's ear, elephant ear",
  "n11782266": "giant taro, Alocasia macrorrhiza",
  "n11782522": "genus Amorphophallus",
  "n11782761": "amorphophallus",
  "n11782878": "pungapung, telingo potato, elephant yam, Amorphophallus paeonifolius, Amorphophallus campanulatus",
  "n11783162": "devil's tongue, snake palm, umbrella arum, Amorphophallus rivieri",
  "n11783488": "krubi, titan arum, Amorphophallus titanum",
  "n11783723": "genus Anthurium",
  "n11783920": "anthurium, tailflower, tail-flower",
  "n11784126": "flamingo flower, flamingo plant, Anthurium andraeanum, Anthurium scherzerianum",
  "n11784323": "Arisaema, genus Arisaema",
  "n11784497": "jack-in-the-pulpit, Indian turnip, wake-robin, Arisaema triphyllum, Arisaema atrorubens",
  "n11784825": "green dragon, Arisaema dracontium",
  "n11785100": "Arisarum, genus Arisarum",
  "n11785276": "friar's-cowl, Arisarum vulgare",
  "n11785475": "genus Caladium",
  "n11785668": "caladium",
  "n11785875": "Caladium bicolor",
  "n11786017": "Calla, genus Calla",
  "n11786131": "wild calla, water arum, Calla palustris",
  "n11786365": "Colocasia, genus Colocasia",
  "n11786539": "taro, taro plant, dalo, dasheen, Colocasia esculenta",
  "n11786843": "taro, cocoyam, dasheen, eddo",
  "n11786983": "genus Cryptocoryne",
  "n11787190": "cryptocoryne, water trumpet",
  "n11787391": "Dieffenbachia, genus Dieffenbachia",
  "n11787625": "dumb cane, mother-in-law plant, mother-in-law's tongue, Dieffenbachia sequine",
  "n11787892": "genus Dracontium",
  "n11788039": "dracontium",
  "n11788223": "Dracunculus, genus Dracunculus",
  "n11788382": "dragon arum, green dragon, Dracunculus vulgaris",
  "n11788536": "Epipremnum, genus Epipremnum",
  "n11788727": "golden pothos, pothos, ivy arum, Epipremnum aureum, Scindapsus aureus",
  "n11788926": "Lysichiton, genus Lysichiton, Lysichitum, genus Lysichitum",
  "n11789066": "skunk cabbage, Lysichiton americanum",
  "n11789280": "genus Monstera",
  "n11789438": "monstera",
  "n11789589": "ceriman, Monstera deliciosa",
  "n11789796": "genus Nephthytis",
  "n11789962": "nephthytis",
  "n11790089": "Nephthytis afzelii",
  "n11790239": "Orontium, genus Orontium",
  "n11790390": "golden club, Orontium aquaticum",
  "n11790624": "Peltandra, genus Peltandra",
  "n11790788": "arrow arum",
  "n11790936": "green arrow arum, tuckahoe, Peltandra virginica",
  "n11791155": "genus Philodendron",
  "n11791341": "philodendron",
  "n11791446": "genus Pistia",
  "n11791569": "pistia, water lettuce, water cabbage, Pistia stratiotes, Pistia stratoites",
  "n11791819": "Scindapsus, genus Scindapsus, genus Pothos",
  "n11792029": "pothos",
  "n11792155": "genus Spathiphyllum",
  "n11792341": "spathiphyllum, peace lily, spathe flower",
  "n11792598": "Symplocarpus, genus Symplocarpus",
  "n11792742": "skunk cabbage, polecat weed, foetid pothos, Symplocarpus foetidus",
  "n11793032": "Syngonium, genus Syngonium",
  "n11793252": "Xanthosoma, genus Xanthosoma",
  "n11793403": "yautia, tannia, spoonflower, malanga, Xanthosoma sagittifolium, Xanthosoma atrovirens",
  "n11793651": "Zantedeschia, genus Zantedeschia",
  "n11793779": "calla lily, calla, arum lily, Zantedeschia aethiopica",
  "n11794024": "pink calla, Zantedeschia rehmanii",
  "n11794139": "golden calla",
  "n11794267": "Lemnaceae, family Lemnaceae, duckweed family",
  "n11794519": "duckweed",
  "n11794791": "Lemna, genus Lemna",
  "n11795049": "common duckweed, lesser duckweed, Lemna minor",
  "n11795216": "star-duckweed, Lemna trisulca",
  "n11795366": "Spirodela, genus Spirodela",
  "n11795580": "great duckweed, water flaxseed, Spirodela polyrrhiza",
  "n11795774": "Wolffia, genus Wolffia",
  "n11796005": "watermeal",
  "n11796188": "common wolffia, Wolffia columbiana",
  "n11796318": "Wolffiella, genus Wolffiella",
  "n11796573": "mud midget, bogmat, Wolffiella gladiata",
  "n11796744": "Araliaceae, family Araliaceae, ivy family",
  "n11797016": "genus Aralia",
  "n11797321": "aralia",
  "n11797508": "American angelica tree, devil's walking stick, Hercules'-club, Aralia spinosa",
  "n11797722": "wild sarsaparilla, false sarsaparilla, wild sarsparilla, Aralia nudicaulis",
  "n11797981": "American spikenard, petty morel, life-of-man, Aralia racemosa",
  "n11798270": "bristly sarsaparilla, bristly sarsparilla, dwarf elder, Aralia hispida",
  "n11798496": "Japanese angelica tree, Aralia elata",
  "n11798688": "Chinese angelica, Chinese angelica tree, Aralia stipulata",
  "n11798851": "Hedera, genus Hedera",
  "n11798978": "ivy, common ivy, English ivy, Hedera helix",
  "n11799158": "Meryta, genus Meryta",
  "n11799331": "puka, Meryta sinclairii",
  "n11799520": "Panax, genus Panax",
  "n11799732": "ginseng, nin-sin, Panax ginseng, Panax schinseng, Panax pseudoginseng",
  "n11800020": "American ginseng, sang, Panax quinquefolius",
  "n11800236": "ginseng",
  "n11800359": "Schefflera, genus Schefflera",
  "n11800565": "umbrella tree, Schefflera actinophylla, Brassaia actinophylla",
  "n11800799": "Aristolochiales, order Aristolochiales",
  "n11801038": "Aristolochiaceae, family Aristolochiaceae, birthwort family",
  "n11801247": "Aristolochia, genus Aristolochia",
  "n11801392": "birthwort, Aristolochia clematitis",
  "n11801665": "Dutchman's-pipe, pipe vine, Aristolochia macrophylla, Aristolochia durior",
  "n11801891": "Virginia snakeroot, Virginia serpentaria, Virginia serpentary, Aristolochia serpentaria",
  "n11802076": "Asarum, genus Asarum",
  "n11802212": "wild ginger",
  "n11802410": "Canada ginger, black snakeroot, Asarum canadense",
  "n11802586": "heartleaf, heart-leaf, Asarum virginicum",
  "n11802800": "heartleaf, heart-leaf, Asarum shuttleworthii",
  "n11802995": "asarabacca, Asarum europaeum",
  "n11803118": "Rafflesiaceae, family Rafflesiaceae",
  "n11803277": "Hydnoraceae, family Hydnoraceae",
  "n11803475": "Caryophyllidae, subclass Caryophyllidae",
  "n11804082": "Caryophyllales, order Caryophyllales, Chenopodiales, order-Chenopodiales",
  "n11804433": "Centrospermae, group Centrospermae",
  "n11804604": "Caryophyllaceae, family Caryophyllaceae, carnation family, pink family",
  "n11805255": "caryophyllaceous plant",
  "n11805380": "Agrostemma, genus Agrostemma",
  "n11805544": "corn cockle, corn campion, crown-of-the-field, Agrostemma githago",
  "n11805837": "Arenaria, genus Arenaria",
  "n11805956": "sandwort",
  "n11806219": "mountain sandwort, mountain starwort, mountain daisy, Arenaria groenlandica",
  "n11806369": "pine-barren sandwort, longroot, Arenaria caroliniana",
  "n11806521": "seabeach sandwort, Arenaria peploides",
  "n11806679": "rock sandwort, Arenaria stricta",
  "n11806814": "thyme-leaved sandwort, Arenaria serpyllifolia",
  "n11806975": "Cerastium, genus Cerastium",
  "n11807108": "mouse-ear chickweed, mouse eared chickweed, mouse ear, clammy chickweed, chickweed",
  "n11807367": "field chickweed, field mouse-ear, Cerastium arvense",
  "n11807525": "snow-in-summer, love-in-a-mist, Cerastium tomentosum",
  "n11807696": "Alpine mouse-ear, Arctic mouse-ear, Cerastium alpinum",
  "n11807849": "Dianthus, genus Dianthus",
  "n11807979": "pink, garden pink",
  "n11808299": "sweet William, Dianthus barbatus",
  "n11808468": "carnation, clove pink, gillyflower, Dianthus caryophyllus",
  "n11808721": "china pink, rainbow pink, Dianthus chinensis",
  "n11808932": "Japanese pink, Dianthus chinensis heddewigii",
  "n11809094": "maiden pink, Dianthus deltoides",
  "n11809271": "cheddar pink, Diangus gratianopolitanus",
  "n11809437": "button pink, Dianthus latifolius",
  "n11809594": "cottage pink, grass pink, Dianthus plumarius",
  "n11809754": "fringed pink, Dianthus supurbus",
  "n11809922": "genus Drypis",
  "n11810030": "drypis",
  "n11810190": "Gypsophila, genus Gypsophila",
  "n11810358": "baby's breath, babies'-breath, Gypsophila paniculata",
  "n11810559": "Hernaria, genus Hernaria",
  "n11810728": "rupturewort, Hernaria glabra",
  "n11810918": "Illecebrum, genus Illecebrum",
  "n11811059": "coral necklace, Illecebrum verticullatum",
  "n11811308": "genus Lychnis",
  "n11811473": "lychnis, catchfly",
  "n11811706": "ragged robin, cuckoo flower, Lychnis flos-cuculi, Lychins floscuculi",
  "n11811921": "scarlet lychnis, maltese cross, Lychins chalcedonica",
  "n11812094": "mullein pink, rose campion, gardener's delight, dusty miller, Lychnis coronaria",
  "n11812358": "Minuartia, genus Minuartia",
  "n11812573": "Moehringia, genus Moehringia",
  "n11812910": "sandwort, Moehringia lateriflora",
  "n11813077": "sandwort, Moehringia mucosa",
  "n11813309": "Paronychia, genus Paronychia",
  "n11813490": "whitlowwort",
  "n11813830": "Petrocoptis, genus Petrocoptis",
  "n11814059": "Sagina, genus Sagina",
  "n11814238": "pearlwort, pearlweed, pearl-weed",
  "n11814440": "Saponaria, genus Saponaria",
  "n11814584": "soapwort, hedge pink, bouncing Bet, bouncing Bess, Saponaria officinalis",
  "n11814824": "Scleranthus, genus Scleranthus",
  "n11814996": "knawel, knawe, Scleranthus annuus",
  "n11815194": "genus Silene",
  "n11815491": "silene, campion, catchfly",
  "n11815721": "moss campion, Silene acaulis",
  "n11815918": "wild pink, Silene caroliniana",
  "n11816121": "red campion, red bird's eye, Silene dioica, Lychnis dioica",
  "n11816336": "white campion, evening lychnis, white cockle, bladder campion, Silene latifolia, Lychnis alba",
  "n11816649": "fire pink, Silene virginica",
  "n11816829": "bladder campion, Silene uniflora, Silene vulgaris",
  "n11817000": "Spergula, genus Spergula",
  "n11817160": "corn spurry, corn spurrey, Spergula arvensis",
  "n11817329": "Spergularia, genus Spergularia",
  "n11817501": "sand spurry, sea spurry, Spergularia rubra",
  "n11817774": "Stellaria, genus Stellaria",
  "n11817914": "chickweed",
  "n11818069": "common chickweed, Stellaria media",
  "n11818271": "stitchwort, greater stitchwort, starwort, Stellaria holostea",
  "n11818515": "Vaccaria, genus Vaccaria",
  "n11818636": "cowherb, cow cockle, Vaccaria hispanica, Vaccaria pyramidata, Saponaria vaccaria",
  "n11818945": "Aizoaceae, family Aizoaceae, Tetragoniaceae, family Tetragoniaceae, carpetweed family",
  "n11819354": "Carpobrotus, genus Carpobrotus",
  "n11819509": "Hottentot fig, Hottentot's fig, sour fig, Carpobrotus edulis, Mesembryanthemum edule",
  "n11819751": "Dorotheanthus, genus Dorotheanthus",
  "n11819912": "livingstone daisy, Dorotheanthus bellidiformis",
  "n11820191": "papilla",
  "n11820323": "genus Lithops",
  "n11820463": "lithops, living stone, stoneface, stone-face, stone plant, stone life face, flowering stone",
  "n11820751": "Mesembryanthemum, genus Mesembryanthemum",
  "n11820965": "fig marigold, pebble plant",
  "n11821184": "ice plant, icicle plant, Mesembryanthemum crystallinum",
  "n11821415": "Molluga, genus Molluga",
  "n11821534": "carpetweed, Indian chickweed, Molluga verticillata",
  "n11821777": "Pleiospilos, genus Pleiospilos",
  "n11821929": "living granite, living rock, stone mimicry plant",
  "n11822167": "Tetragonia, genus Tetragonia",
  "n11822300": "New Zealand spinach, Tetragonia tetragonioides, Tetragonia expansa",
  "n11822557": "Amaranthaceae, family Amaranthaceae, amaranth family",
  "n11822849": "Amaranthus, genus Amaranthus",
  "n11823043": "amaranth",
  "n11823305": "amaranth",
  "n11823436": "tumbleweed, Amaranthus albus, Amaranthus graecizans",
  "n11823572": "love-lies-bleeding, velvet flower, tassel flower, Amaranthus caudatus",
  "n11823756": "prince's-feather, gentleman's-cane, prince's-plume, red amaranth, purple amaranth, Amaranthus cruentus, Amaranthus hybridus hypochondriacus, Amaranthus hybridus erythrostachys",
  "n11824146": "pigweed, Amaranthus hypochondriacus",
  "n11824344": "thorny amaranth, Amaranthus spinosus",
  "n11824548": "Alternanthera, genus Alternanthera",
  "n11824747": "alligator weed, alligator grass, Alternanthera philoxeroides",
  "n11825013": "Celosia, genus Celosia",
  "n11825211": "red fox, Celosia argentea",
  "n11825351": "cockscomb, common cockscomb, Celosia cristata, Celosia argentea cristata",
  "n11825535": "Froelichia, genus Froelichia",
  "n11825749": "cottonweed",
  "n11825988": "Gomphrena, genus Gomphrena",
  "n11826198": "globe amaranth, bachelor's button, Gomphrena globosa",
  "n11826416": "Iresine, genus Iresine",
  "n11826569": "bloodleaf",
  "n11826715": "beefsteak plant, beef plant, Iresine herbstii, Iresine reticulata",
  "n11826999": "Telanthera, genus Telanthera",
  "n11827169": "Batidaceae, family Batidaceae, saltwort family",
  "n11827348": "Batis, genus Batis",
  "n11827541": "saltwort, Batis maritima",
  "n11827775": "Chenopodiaceae, family Chenopodiaceae, goosefoot family",
  "n11828113": "Chenopodium, genus Chenopodium",
  "n11828247": "goosefoot",
  "n11828577": "lamb's-quarters, pigweed, wild spinach, Chenopodium album",
  "n11828804": "American wormseed, Mexican tea, Spanish tea, wormseed, Chenopodium ambrosioides",
  "n11828973": "good-king-henry, allgood, fat hen, wild spinach, Chenopodium bonus-henricus",
  "n11829205": "Jerusalem oak, feather geranium, Mexican tea, Chenopodium botrys, Atriplex mexicana",
  "n11829435": "strawberry blite, strawberry pigweed, Indian paint, Chenopodium capitatum",
  "n11829672": "oak-leaved goosefoot, oakleaf goosefoot, Chenopodium glaucum",
  "n11829922": "sowbane, red goosefoot, Chenopodium hybridum",
  "n11830045": "nettle-leaved goosefoot, nettleleaf goosefoot, Chenopodium murale",
  "n11830252": "red goosefoot, French spinach, Chenopodium rubrum",
  "n11830400": "stinking goosefoot, Chenopodium vulvaria",
  "n11830570": "Atriplex, genus Atriplex",
  "n11830714": "orach, orache",
  "n11830906": "saltbush",
  "n11831100": "garden orache, mountain spinach, Atriplex hortensis",
  "n11831297": "desert holly, Atriplex hymenelytra",
  "n11831521": "quail bush, quail brush, white thistle, Atriplex lentiformis",
  "n11831730": "Bassia, genus Bassia, Kochia, genus Kochia",
  "n11831874": "summer cypress, burning bush, fire bush, fire-bush, belvedere, Bassia scoparia, Kochia scoparia",
  "n11832108": "Beta, genus Beta",
  "n11832214": "beet, common beet, Beta vulgaris",
  "n11832480": "beetroot, Beta vulgaris rubra",
  "n11832671": "chard, Swiss chard, spinach beet, leaf beet, chard plant, Beta vulgaris cicla",
  "n11832899": "mangel-wurzel, mangold-wurzel, mangold, Beta vulgaris vulgaris",
  "n11833073": "sugar beet",
  "n11833208": "Cycloloma, genus Cycloloma",
  "n11833373": "winged pigweed, tumbleweed, Cycloloma atriplicifolium",
  "n11833577": "genus Halogeton",
  "n11833749": "halogeton, Halogeton glomeratus",
  "n11833999": "barilla, Halogeton souda",
  "n11834148": "Salicornia, genus Salicornia",
  "n11834272": "glasswort, samphire, Salicornia europaea",
  "n11834521": "Salsola, genus Salsola",
  "n11834654": "saltwort, barilla, glasswort, kali, kelpwort, Salsola kali, Salsola soda",
  "n11834890": "Russian thistle, Russian tumbleweed, Russian cactus, tumbleweed, Salsola kali tenuifolia",
  "n11835114": "Sarcobatus, genus Sarcobatus",
  "n11835251": "greasewood, black greasewood, Sarcobatus vermiculatus",
  "n11835451": "Spinacia, genus Spinacia",
  "n11835568": "spinach, spinach plant, prickly-seeded spinach, Spinacia oleracea",
  "n11835806": "Nyctaginaceae, family Nyctaginaceae, Allioniaceae, family Allioniaceae, four-o'clock family",
  "n11836137": "Nyctaginia, genus Nyctaginia",
  "n11836327": "scarlet musk flower, Nyctaginia capitata",
  "n11836556": "Abronia, genus Abronia",
  "n11836722": "sand verbena",
  "n11837020": "snowball, sweet sand verbena, Abronia elliptica",
  "n11837204": "sweet sand verbena, Abronia fragrans",
  "n11837351": "yellow sand verbena, Abronia latifolia",
  "n11837562": "beach pancake, Abronia maritima",
  "n11837743": "beach sand verbena, pink sand verbena, Abronia umbellata",
  "n11837970": "desert sand verbena, Abronia villosa",
  "n11838266": "Allionia, genus Allionia",
  "n11838413": "trailing four o'clock, trailing windmills, Allionia incarnata",
  "n11838741": "genus Bougainvillea, Bougainvillaea, genus Bougainvillaea",
  "n11838916": "bougainvillea",
  "n11839167": "paper flower, Bougainvillea glabra",
  "n11839297": "Mirabilis, genus Mirabilis",
  "n11839460": "umbrellawort",
  "n11839568": "four o'clock",
  "n11839823": "common four-o'clock, marvel-of-Peru, Mirabilis jalapa, Mirabilis uniflora",
  "n11840067": "California four o'clock, Mirabilis laevis, Mirabilis californica",
  "n11840246": "sweet four o'clock, maravilla, Mirabilis longiflora",
  "n11840476": "desert four o'clock, Colorado four o'clock, maravilla, Mirabilis multiflora",
  "n11840764": "mountain four o'clock, Mirabilis oblongifolia",
  "n11841061": "Pisonia, genus Pisonia",
  "n11841247": "cockspur, Pisonia aculeata",
  "n11841368": "Opuntiales, order Opuntiales",
  "n11841529": "Cactaceae, family Cactaceae, cactus family",
  "n11842204": "cactus",
  "n11842861": "Acanthocereus, genus Acanthocereus",
  "n11843053": "pitahaya cactus, pitahaya, Acanthocereus tetragonus, Acanthocereus pentagonus",
  "n11843285": "Aporocactus, genus Aporocactus",
  "n11843441": "rattail cactus, rat's-tail cactus, Aporocactus flagelliformis",
  "n11843709": "Ariocarpus, genus Ariocarpus",
  "n11843896": "living rock, Ariocarpus fissuratus",
  "n11844203": "Carnegiea, genus Carnegiea",
  "n11844371": "saguaro, sahuaro, Carnegiea gigantea",
  "n11844651": "Cereus, genus Cereus",
  "n11844892": "night-blooming cereus",
  "n11845019": "genus Coryphantha",
  "n11845277": "coryphantha",
  "n11845387": "genus Echinocactus",
  "n11845557": "echinocactus, barrel cactus",
  "n11845793": "hedgehog cactus",
  "n11845913": "golden barrel cactus, Echinocactus grusonii",
  "n11846087": "Echinocereus, genus Echinocereus",
  "n11846312": "hedgehog cereus",
  "n11846425": "rainbow cactus",
  "n11846582": "genus Epiphyllum",
  "n11846765": "epiphyllum, orchid cactus",
  "n11846970": "Ferocactus, genus Ferocactus",
  "n11847169": "barrel cactus",
  "n11847414": "Gymnocalycium, genus Gymnocalycium",
  "n11847615": "Harrisia, genus Harrisia",
  "n11847841": "Hatiora, genus Hatiora",
  "n11848009": "Easter cactus, Hatiora gaertneri, Schlumbergera gaertneri",
  "n11848253": "Hylocereus, genus Hylocereus",
  "n11848479": "night-blooming cereus",
  "n11848610": "Lemaireocereus, genus Lemaireocereus",
  "n11848867": "chichipe, Lemaireocereus chichipe",
  "n11849017": "Lophophora, genus Lophophora",
  "n11849271": "mescal, mezcal, peyote, Lophophora williamsii",
  "n11849467": "mescal button, sacred mushroom, magic mushroom",
  "n11849666": "genus Mammillaria",
  "n11849871": "mammillaria",
  "n11849983": "feather ball, Mammillaria plumosa",
  "n11850136": "Melocactus, genus Melocactus",
  "n11850337": "Myrtillocactus, genus Myrtillocactus",
  "n11850521": "garambulla, garambulla cactus, Myrtillocactus geometrizans",
  "n11850748": "Pediocactus, genus Pediocactus",
  "n11850918": "Knowlton's cactus, Pediocactus knowltonii",
  "n11851101": "Nopalea, genus Nopalea",
  "n11851258": "nopal",
  "n11851395": "Opuntia, genus Opuntia",
  "n11851578": "prickly pear, prickly pear cactus",
  "n11851839": "cholla, Opuntia cholla",
  "n11852028": "nopal, Opuntia lindheimeri",
  "n11852148": "tuna, Opuntia tuna",
  "n11852255": "Pereskia, genus Pereskia, Peireskia, genus Peireskia",
  "n11852531": "Barbados gooseberry, Barbados-gooseberry vine, Pereskia aculeata",
  "n11852814": "Rhipsalis, genus Rhipsalis",
  "n11853079": "mistletoe cactus",
  "n11853191": "Schlumbergera, genus Schlumbergera",
  "n11853356": "Christmas cactus, Schlumbergera buckleyi, Schlumbergera baridgesii",
  "n11853644": "Selenicereus, genus Selenicereus",
  "n11853813": "night-blooming cereus",
  "n11853979": "queen of the night, Selenicereus grandiflorus",
  "n11854232": "Zygocactus, genus Zygocactus",
  "n11854479": "crab cactus, Thanksgiving cactus, Zygocactus truncatus, Schlumbergera truncatus",
  "n11854760": "Phytolaccaceae, family Phytolaccaceae, pokeweed family",
  "n11855122": "Phytolacca, genus Phytolacca",
  "n11855274": "pokeweed",
  "n11855435": "Indian poke, Phytolacca acinosa",
  "n11855553": "poke, pigeon berry, garget, scoke, Phytolacca americana",
  "n11855842": "ombu, bella sombra, Phytolacca dioica",
  "n11856055": "Agdestis, genus Agdestis",
  "n11856271": "Ercilla, genus Ercilla",
  "n11856389": "Rivina, genus Rivina",
  "n11856573": "bloodberry, blood berry, rougeberry, rouge plant, Rivina humilis",
  "n11856815": "Trichostigma, genus Trichostigma",
  "n11856981": "Portulacaceae, family Portulacaceae, purslane family",
  "n11857320": "purslane",
  "n11857528": "genus Portulaca",
  "n11857696": "portulaca",
  "n11857875": "rose moss, sun plant, Portulaca grandiflora",
  "n11858077": "common purslane, pussley, pussly, verdolagas, Portulaca oleracea",
  "n11858406": "Calandrinia, genus Calandrinia",
  "n11858703": "rock purslane",
  "n11858814": "red maids, redmaids, Calandrinia ciliata",
  "n11859024": "Claytonia, genus Claytonia",
  "n11859275": "Carolina spring beauty, Claytonia caroliniana",
  "n11859472": "spring beauty, Clatonia lanceolata",
  "n11859737": "Virginia spring beauty, Claytonia virginica",
  "n11859981": "Lewisia, genus Lewisia",
  "n11860208": "siskiyou lewisia, Lewisia cotyledon",
  "n11860555": "bitterroot, Lewisia rediviva",
  "n11860801": "Montia, genus Montia",
  "n11861021": "Indian lettuce",
  "n11861238": "broad-leaved montia, Montia cordifolia",
  "n11861487": "blinks, blinking chickweed, water chickweed, Montia lamprosperma",
  "n11861641": "toad lily, Montia chamissoi",
  "n11861853": "winter purslane, miner's lettuce, Cuban spinach, Montia perfoliata",
  "n11862089": "Spraguea, genus Spraguea",
  "n11862300": "pussy-paw, pussy-paws, pussy's-paw, Spraguea umbellatum, Calyptridium umbellatum",
  "n11862598": "Talinum, genus Talinum",
  "n11862835": "flame flower, flame-flower, flameflower, Talinum aurantiacum",
  "n11863242": "narrow-leaved flame flower, Talinum augustissimum",
  "n11863467": "pigmy talinum, Talinum brevifolium",
  "n11863717": "rock pink, Talinum calycinum",
  "n11863877": "jewels-of-opar, Talinum paniculatum",
  "n11864114": "spiny talinum, Talinum spinescens",
  "n11864364": "Rhoeadales, order Rhoeadales, Papaverales, order Papaverales",
  "n11864602": "Capparidaceae, family Capparidaceae, caper family",
  "n11864906": "Capparis, genus Capparis",
  "n11865071": "caper",
  "n11865276": "native pomegranate, Capparis arborea",
  "n11865429": "caper tree, Jamaica caper tree, Capparis cynophallophora",
  "n11865574": "caper tree, bay-leaved caper, Capparis flexuosa",
  "n11865738": "native orange, Capparis mitchellii",
  "n11865874": "common caper, Capparis spinosa",
  "n11866078": "genus Cleome, Cleome",
  "n11866248": "spiderflower, cleome",
  "n11866469": "spider flower, spider plant, Cleome hassleriana",
  "n11866706": "Rocky Mountain bee plant, stinking clover, Cleome serrulata",
  "n11866942": "Crateva, genus Crateva",
  "n11867070": "Polanisia, genus Polanisia",
  "n11867311": "clammyweed, Polanisia graveolens, Polanisia dodecandra",
  "n11867525": "Cruciferae, family Cruciferae, Brassicaceae, family Brassicaceae, mustard family",
  "n11868814": "crucifer, cruciferous plant",
  "n11869351": "cress, cress plant",
  "n11869689": "watercress",
  "n11869890": "Aethionema, genus Aethionema",
  "n11870044": "stonecress, stone cress",
  "n11870212": "Alliaria, genus Alliaria",
  "n11870418": "garlic mustard, hedge garlic, sauce-alone, jack-by-the-hedge, Alliaria officinalis",
  "n11870607": "Alyssum, genus Alyssum",
  "n11870747": "alyssum, madwort",
  "n11870916": "Anastatica, genus Anastatica",
  "n11871059": "rose of Jericho, resurrection plant, Anastatica hierochuntica",
  "n11871294": "Arabidopsis, genus Arabidopsis",
  "n11871496": "Arabidopsis thaliana, mouse-ear cress",
  "n11871748": "Arabidopsis lyrata",
  "n11871916": "Arabis, genus Arabis",
  "n11872146": "rock cress, rockcress",
  "n11872324": "sicklepod, Arabis Canadensis",
  "n11872473": "tower cress, tower mustard, Arabis turrita",
  "n11872658": "tower mustard, tower cress, Turritis glabra, Arabis glabra",
  "n11872850": "Armoracia, genus Armoracia",
  "n11872973": "horseradish, horse radish, red cole, Armoracia rusticana",
  "n11873182": "horseradish, horseradish root",
  "n11873396": "Barbarea, genus Barbarea",
  "n11873612": "winter cress, St. Barbara's herb, scurvy grass",
  "n11873845": "Belle Isle cress, early winter cress, land cress, American cress, American watercress, Barbarea verna, Barbarea praecox",
  "n11874081": "yellow rocket, rockcress, rocket cress, Barbarea vulgaris, Sisymbrium barbarea",
  "n11874300": "Berteroa, genus Berteroa",
  "n11874423": "hoary alison, hoary alyssum, Berteroa incana",
  "n11874707": "Biscutella, genus Biscutella",
  "n11874878": "buckler mustard, Biscutalla laevigata",
  "n11875100": "Brassica, genus Brassica",
  "n11875523": "wild cabbage, Brassica oleracea",
  "n11875691": "cabbage, cultivated cabbage, Brassica oleracea",
  "n11875938": "head cabbage, head cabbage plant, Brassica oleracea capitata",
  "n11876204": "savoy cabbage",
  "n11876316": "red cabbage",
  "n11876432": "brussels sprout, Brassica oleracea gemmifera",
  "n11876634": "cauliflower, Brassica oleracea botrytis",
  "n11876803": "broccoli, Brassica oleracea italica",
  "n11876976": "kale, kail, cole, borecole, colewort, Brassica oleracea acephala",
  "n11877193": "collard",
  "n11877283": "kohlrabi, Brassica oleracea gongylodes",
  "n11877473": "turnip plant",
  "n11877646": "turnip, white turnip, Brassica rapa",
  "n11877860": "rutabaga, turnip cabbage, swede, Swedish turnip, rutabaga plant, Brassica napus napobrassica",
  "n11878101": "broccoli raab, broccoli rabe, Brassica rapa ruvo",
  "n11878283": "mustard",
  "n11878520": "mustard oil",
  "n11878633": "chinese mustard, indian mustard, leaf mustard, gai choi, Brassica juncea",
  "n11878808": "Chinese cabbage, celery cabbage, napa, pe-tsai, Brassica rapa pekinensis",
  "n11879054": "bok choy, bok choi, pakchoi, pak choi, Chinese white cabbage, Brassica rapa chinensis",
  "n11879291": "tendergreen, spinach mustard, Brassica perviridis, Brassica rapa perviridis",
  "n11879505": "black mustard, Brassica nigra",
  "n11879722": "rape, colza, Brassica napus",
  "n11879895": "rapeseed",
  "n11880032": "rape oil, rapeseed oil, colza oil",
  "n11880218": "Cakile, genus Cakile",
  "n11880411": "sea-rocket, Cakile maritima",
  "n11880610": "Camelina, genus Camelina, false flax",
  "n11880791": "gold of pleasure, Camelina sativa",
  "n11881063": "Capsella, genus Capsella",
  "n11881189": "shepherd's purse, shepherd's pouch, Capsella bursa-pastoris",
  "n11881426": "Cardamine, genus Cardamine",
  "n11881563": "Dentaria, genus Dentaria",
  "n11881742": "bittercress, bitter cress",
  "n11882074": "lady's smock, cuckooflower, cuckoo flower, meadow cress, Cardamine pratensis",
  "n11882237": "coral-root bittercress, coralroot, coralwort, Cardamine bulbifera, Dentaria bulbifera",
  "n11882426": "crinkleroot, crinkle-root, crinkle root, pepper root, toothwort, Cardamine diphylla, Dentaria diphylla",
  "n11882636": "American watercress, mountain watercress, Cardamine rotundifolia",
  "n11882821": "spring cress, Cardamine bulbosa",
  "n11882972": "purple cress, Cardamine douglasii",
  "n11883137": "Cheiranthus, genus Cheiranthus",
  "n11883328": "wallflower, Cheiranthus cheiri, Erysimum cheiri",
  "n11883628": "prairie rocket",
  "n11883799": "Cochlearia, genus Cochlearia",
  "n11883945": "scurvy grass, common scurvy grass, Cochlearia officinalis",
  "n11884198": "Crambe, genus Crambe",
  "n11884384": "sea kale, sea cole, Crambe maritima",
  "n11884667": "Descurainia, genus Descurainia",
  "n11884967": "tansy mustard, Descurainia pinnata",
  "n11885148": "Diplotaxis, genus Diplotaxis",
  "n11885292": "wall rocket, Diplotaxis muralis, Diplotaxis tenuifolia",
  "n11885524": "white rocket, Diplotaxis erucoides",
  "n11885697": "genus Draba",
  "n11885856": "draba",
  "n11886157": "whitlow grass, shadflower, shad-flower, Draba verna",
  "n11886380": "Eruca, genus Eruca",
  "n11886537": "rocket, roquette, garden rocket, rocket salad, arugula, Eruca sativa, Eruca vesicaria sativa",
  "n11886788": "Erysimum, genus Erysimum",
  "n11887119": "wallflower",
  "n11887310": "prairie rocket",
  "n11887476": "Siberian wall flower, Erysimum allionii, Cheiranthus allionii",
  "n11887750": "western wall flower, Erysimum asperum, Cheiranthus asperus, Erysimum arkansanum",
  "n11888061": "wormseed mustard, Erysimum cheiranthoides",
  "n11888271": "genus Heliophila",
  "n11888424": "heliophila",
  "n11888621": "Hesperis, genus Hesperis",
  "n11888800": "damask violet, Dame's violet, sweet rocket, Hesperis matronalis",
  "n11889078": "Hugueninia, genus Hugueninia",
  "n11889205": "tansy-leaved rocket, Hugueninia tanacetifolia, Sisymbrium tanacetifolia",
  "n11889473": "Iberis, genus Iberis",
  "n11889619": "candytuft",
  "n11889847": "Isatis, genus Isatis",
  "n11890022": "woad",
  "n11890150": "dyer's woad, Isatis tinctoria",
  "n11890329": "Lepidium, genus Lepidium",
  "n11890507": "common garden cress, garden pepper cress, pepper grass, pepperwort, Lepidium sativum",
  "n11890723": "Lesquerella, genus Lesquerella",
  "n11890884": "bladderpod",
  "n11891050": "Lobularia, genus Lobularia",
  "n11891175": "sweet alyssum, sweet alison, Lobularia maritima",
  "n11891395": "Lunaria, genus Lunaria",
  "n11891541": "honesty, silver dollar, money plant, satin flower, satinpod, Lunaria annua",
  "n11891838": "Malcolmia, genus Malcolmia",
  "n11892029": "Malcolm stock, stock",
  "n11892181": "Virginian stock, Virginia stock, Malcolmia maritima",
  "n11892460": "Matthiola, genus Matthiola",
  "n11892637": "stock, gillyflower",
  "n11892817": "brompton stock, Matthiola incana",
  "n11893004": "Nasturtium, genus Nasturtium",
  "n11893131": "common watercress, Rorippa nasturtium-aquaticum, Nasturtium officinale",
  "n11893451": "Physaria, genus Physaria",
  "n11893640": "bladderpod",
  "n11893808": "Pritzelago, genus Pritzelago",
  "n11893916": "chamois cress, Pritzelago alpina, Lepidium alpina",
  "n11894173": "Raphanus, genus Raphanus",
  "n11894327": "radish plant, radish",
  "n11894558": "jointed charlock, wild radish, wild rape, runch, Raphanus raphanistrum",
  "n11894770": "radish, Raphanus sativus",
  "n11894958": "radish",
  "n11895092": "radish, daikon, Japanese radish, Raphanus sativus longipinnatus",
  "n11895270": "Rorippa, genus Rorippa",
  "n11895472": "marsh cress, yellow watercress, Rorippa islandica",
  "n11895714": "great yellowcress, Rorippa amphibia, Nasturtium amphibium",
  "n11895980": "genus Schizopetalon",
  "n11896141": "schizopetalon, Schizopetalon walkeri",
  "n11896365": "Sinapis, genus Sinapis",
  "n11896519": "white mustard, Brassica hirta, Sinapis alba",
  "n11896722": "field mustard, wild mustard, charlock, chadlock, Brassica kaber, Sinapis arvensis",
  "n11896904": "genus Sisymbrium",
  "n11897116": "hedge mustard, Sisymbrium officinale",
  "n11897342": "Stanleya, genus Stanleya",
  "n11897466": "desert plume, prince's-plume, Stanleya pinnata, Cleome pinnata",
  "n11897760": "Stephanomeria, genus Stephanomeria",
  "n11897900": "malheur wire lettuce, Stephanomeria malheurensis",
  "n11898079": "Subularia, genus Subularia",
  "n11898271": "awlwort, Subularia aquatica",
  "n11898474": "Thlaspi, genus Thlaspi",
  "n11898639": "pennycress",
  "n11898775": "field pennycress, French weed, fanweed, penny grass, stinkweed, mithridate mustard, Thlaspi arvense",
  "n11899027": "Thysanocarpus, genus Thysanocarpus",
  "n11899223": "fringepod, lacepod",
  "n11899432": "Turritis, genus Turritis",
  "n11899595": "Vesicaria, genus Vesicaria",
  "n11899762": "bladderpod",
  "n11899921": "wasabi",
  "n11900058": "Papaveraceae, family Papaveraceae, poppy family",
  "n11900569": "poppy",
  "n11900986": "Papaver, genus Papaver",
  "n11901294": "Iceland poppy, Papaver alpinum",
  "n11901452": "western poppy, Papaver californicum",
  "n11901597": "prickly poppy, Papaver argemone",
  "n11901759": "Iceland poppy, arctic poppy, Papaver nudicaule",
  "n11901977": "oriental poppy, Papaver orientale",
  "n11902200": "corn poppy, field poppy, Flanders poppy, Papaver rhoeas",
  "n11902389": "opium poppy, Papaver somniferum",
  "n11902595": "genus Argemone",
  "n11902709": "prickly poppy, argemone, white thistle, devil's fig",
  "n11902982": "Mexican poppy, Argemone mexicana",
  "n11903167": "genus Bocconia",
  "n11903333": "bocconia, tree celandine, Bocconia frutescens",
  "n11903525": "Chelidonium, genus Chelidonium",
  "n11903671": "celandine, greater celandine, swallowwort, swallow wort, Chelidonium majus",
  "n11903881": "Corydalis, genus Corydalis",
  "n11904109": "corydalis",
  "n11904274": "climbing corydalis, Corydalis claviculata, Fumaria claviculata",
  "n11904477": "Roman wormwood, rock harlequin, Corydalis sempervirens, Fumaria sempervirens",
  "n11904743": "fumewort, fumeroot, Corydalis solida",
  "n11904896": "Dendromecon, genus Dendromecon",
  "n11905035": "bush poppy, tree poppy",
  "n11905236": "Eschscholtzia, genus Eschscholtzia",
  "n11905392": "California poppy, Eschscholtzia californica",
  "n11905584": "Glaucium, genus Glaucium",
  "n11905749": "horn poppy, horned poppy, yellow horned poppy, sea poppy, Glaucium flavum",
  "n11905989": "Hunnemannia, genus Hunnemania",
  "n11906127": "golden cup, Mexican tulip poppy, Hunnemania fumariifolia",
  "n11906359": "Macleaya, genus Macleaya",
  "n11906514": "plume poppy, bocconia, Macleaya cordata",
  "n11906713": "Meconopsis, genus Meconopsis",
  "n11906917": "blue poppy, Meconopsis betonicifolia",
  "n11907100": "Welsh poppy, Meconopsis cambrica",
  "n11907267": "Platystemon, genus Platystemon",
  "n11907405": "creamcups, Platystemon californicus",
  "n11907554": "Romneya, genus Romneya",
  "n11907689": "matilija poppy, California tree poppy, Romneya coulteri",
  "n11907939": "Sanguinaria, genus Sanguinaria",
  "n11908077": "bloodroot, puccoon, redroot, tetterwort, Sanguinaria canadensis",
  "n11908431": "Stylomecon, genus Stylomecon",
  "n11908549": "wind poppy, flaming poppy, Stylomecon heterophyllum, Papaver heterophyllum",
  "n11908718": "Stylophorum, genus Stylophorum",
  "n11908846": "celandine poppy, wood poppy, Stylophorum diphyllum",
  "n11909048": "Fumariaceae, family Fumariaceae, fumitory family",
  "n11909353": "Fumaria, genus Fumaria",
  "n11909527": "fumitory, fumewort, fumeroot, Fumaria officinalis",
  "n11909745": "Adlumia, genus Adlumia",
  "n11909864": "climbing fumitory, Allegheny vine, Adlumia fungosa, Fumaria fungosa",
  "n11910070": "Dicentra, genus Dicentra",
  "n11910271": "bleeding heart, lyreflower, lyre-flower, Dicentra spectabilis",
  "n11910460": "Dutchman's breeches, Dicentra cucullaria",
  "n11910666": "squirrel corn, Dicentra canadensis",
  "n11910835": "Asteridae, subclass Asteridae",
  "n11911274": "Campanulales, order Campanulales",
  "n11911591": "Compositae, family Compositae, Asteraceae, family Asteraceae, aster family",
  "n11915214": "composite, composite plant",
  "n11915658": "compass plant, compass flower",
  "n11915899": "everlasting, everlasting flower",
  "n11916268": "genus Achillea",
  "n11916467": "achillea",
  "n11916696": "yarrow, milfoil, Achillea millefolium",
  "n11916965": "sneezeweed yarrow, sneezewort, Achillea ptarmica",
  "n11917186": "Acroclinium, genus Acroclinium",
  "n11917407": "pink-and-white everlasting, pink paper daisy, Acroclinium roseum",
  "n11917633": "Ageratina, genus Ageratina",
  "n11917835": "white snakeroot, white sanicle, Ageratina altissima, Eupatorium rugosum",
  "n11918131": "genus Ageratum",
  "n11918286": "ageratum",
  "n11918473": "common ageratum, Ageratum houstonianum",
  "n11918631": "Amberboa, genus Amberboa",
  "n11918808": "sweet sultan, Amberboa moschata, Centaurea moschata",
  "n11919026": "genus Ambrosia",
  "n11919232": "Ambrosiaceae, family Ambrosiaceae",
  "n11919447": "ragweed, ambrosia, bitterweed",
  "n11919761": "common ragweed, Ambrosia artemisiifolia",
  "n11919975": "great ragweed, Ambrosia trifida",
  "n11920133": "western ragweed, perennial ragweed, Ambrosia psilostachya",
  "n11920344": "genus Ammobium",
  "n11920498": "ammobium",
  "n11920663": "winged everlasting, Ammobium alatum",
  "n11920867": "Anacyclus, genus Anacyclus",
  "n11920998": "pellitory, pellitory-of-Spain, Anacyclus pyrethrum",
  "n11921200": "Anaphalis, genus Anaphalis",
  "n11921395": "pearly everlasting, cottonweed, Anaphalis margaritacea",
  "n11921622": "genus Andryala",
  "n11921792": "andryala",
  "n11921949": "Antennaria, genus Antennaria",
  "n11922192": "ladies' tobacco, lady's tobacco, Antennaria plantaginifolia",
  "n11922374": "cat's foot, cat's feet, pussytoes, Antennaria dioica",
  "n11922661": "plantain-leaved pussytoes",
  "n11922755": "field pussytoes",
  "n11922839": "solitary pussytoes",
  "n11922926": "mountain everlasting",
  "n11923016": "Anthemis, genus Anthemis",
  "n11923174": "mayweed, dog fennel, stinking mayweed, stinking chamomile, Anthemis cotula",
  "n11923397": "yellow chamomile, golden marguerite, dyers' chamomile, Anthemis tinctoria",
  "n11923637": "corn chamomile, field chamomile, corn mayweed, Anthemis arvensis",
  "n11923827": "Antheropeas, genus Antheropeas",
  "n11924014": "woolly daisy, dwarf daisy, Antheropeas wallacei, Eriophyllum wallacei",
  "n11924330": "Arctium, genus Arctium",
  "n11924445": "burdock, clotbur",
  "n11924661": "common burdock, lesser burdock, Arctium minus",
  "n11924849": "great burdock, greater burdock, cocklebur, Arctium lappa",
  "n11925140": "Arctotis, genus Arctotis",
  "n11925303": "African daisy",
  "n11925450": "blue-eyed African daisy, Arctotis stoechadifolia, Arctotis venusta",
  "n11925720": "Argyranthemum, genus Argyranthemum",
  "n11925898": "marguerite, marguerite daisy, Paris daisy, Chrysanthemum frutescens, Argyranthemum frutescens",
  "n11926185": "Argyroxiphium, genus Argyroxiphium",
  "n11926365": "silversword, Argyroxiphium sandwicense",
  "n11926640": "genus Arnica",
  "n11926833": "arnica",
  "n11926976": "heartleaf arnica, Arnica cordifolia",
  "n11927215": "Arnica montana",
  "n11927509": "arnica",
  "n11927616": "Arnoseris, genus Arnoseris",
  "n11927740": "lamb succory, dwarf nipplewort, Arnoseris minima",
  "n11927901": "genus Artemisia",
  "n11928352": "artemisia",
  "n11928549": "wormwood",
  "n11928858": "mugwort",
  "n11929027": "sagebrush, sage brush",
  "n11929291": "southernwood, Artemisia abrotanum",
  "n11929477": "common wormwood, absinthe, old man, lad's love, Artemisia absinthium",
  "n11929743": "sweet wormwood, Artemisia annua",
  "n11929880": "California sagebrush, California sage, Artemisia californica",
  "n11930038": "field wormwood, Artemisia campestris",
  "n11930203": "tarragon, estragon, Artemisia dracunculus",
  "n11930353": "sand sage, silvery wormwood, Artemisia filifolia",
  "n11930571": "wormwood sage, prairie sagewort, Artemisia frigida",
  "n11930788": "western mugwort, white sage, cudweed, prairie sage, Artemisia ludoviciana, Artemisia gnaphalodes",
  "n11930994": "Roman wormwood, Artemis pontica",
  "n11931135": "bud brush, bud sagebrush, Artemis spinescens",
  "n11931312": "dusty miller, beach wormwood, old woman, Artemisia stelleriana",
  "n11931540": "common mugwort, Artemisia vulgaris",
  "n11931756": "genus Aster",
  "n11931918": "aster",
  "n11932745": "wood aster",
  "n11932927": "whorled aster, Aster acuminatus",
  "n11933099": "heath aster, Aster arenosus",
  "n11933257": "heart-leaved aster, Aster cordifolius",
  "n11933387": "white wood aster, Aster divaricatus",
  "n11933546": "bushy aster, Aster dumosus",
  "n11933728": "heath aster, Aster ericoides",
  "n11933903": "white prairie aster, Aster falcatus",
  "n11934041": "stiff aster, Aster linarifolius",
  "n11934239": "goldilocks, goldilocks aster, Aster linosyris, Linosyris vulgaris",
  "n11934463": "large-leaved aster, Aster macrophyllus",
  "n11934616": "New England aster, Aster novae-angliae",
  "n11934807": "Michaelmas daisy, New York aster, Aster novi-belgii",
  "n11935027": "upland white aster, Aster ptarmicoides",
  "n11935187": "Short's aster, Aster shortii",
  "n11935330": "sea aster, sea starwort, Aster tripolium",
  "n11935469": "prairie aster, Aster turbinellis",
  "n11935627": "annual salt-marsh aster",
  "n11935715": "aromatic aster",
  "n11935794": "arrow leaved aster",
  "n11935877": "azure aster",
  "n11935953": "bog aster",
  "n11936027": "crooked-stemmed aster",
  "n11936113": "Eastern silvery aster",
  "n11936199": "flat-topped white aster",
  "n11936287": "late purple aster",
  "n11936369": "panicled aster",
  "n11936448": "perennial salt marsh aster",
  "n11936539": "purple-stemmed aster",
  "n11936624": "rough-leaved aster",
  "n11936707": "rush aster",
  "n11936782": "Schreiber's aster",
  "n11936864": "small white aster",
  "n11936946": "smooth aster",
  "n11937023": "southern aster",
  "n11937102": "starved aster, calico aster",
  "n11937195": "tradescant's aster",
  "n11937278": "wavy-leaved aster",
  "n11937360": "Western silvery aster",
  "n11937446": "willow aster",
  "n11937523": "genus Ayapana",
  "n11937692": "ayapana, Ayapana triplinervis, Eupatorium aya-pana",
  "n11937965": "Baccharis, genus Baccharis",
  "n11938261": "groundsel tree, groundsel bush, consumption weed, cotton-seed tree, Baccharis halimifolia",
  "n11938556": "mule fat, Baccharis viminea",
  "n11938732": "coyote brush, coyote bush, chaparral broom, kidney wort, Baccharis pilularis",
  "n11938977": "Balsamorhiza, genus Balsamorhiza",
  "n11939180": "balsamroot",
  "n11939380": "Bellis, genus Bellis",
  "n11939491": "daisy",
  "n11939699": "common daisy, English daisy, Bellis perennis",
  "n11939887": "Bidens, genus Bidens",
  "n11940006": "bur marigold, burr marigold, beggar-ticks, beggar's-ticks, sticktight",
  "n11940349": "Spanish needles, Bidens bipinnata",
  "n11940478": "Spanish needles, beggar-ticks",
  "n11940599": "tickseed sunflower, Bidens coronata, Bidens trichosperma",
  "n11940750": "European beggar-ticks, trifid beggar-ticks, trifid bur marigold, Bidens tripartita",
  "n11940915": "swampy beggar-ticks, Bidens connata",
  "n11941094": "slender knapweed",
  "n11941178": "Jersey knapweed",
  "n11941261": "Boltonia, genus Boltonia",
  "n11941478": "false chamomile",
  "n11941719": "Brachycome, genus Brachycome",
  "n11941924": "Swan River daisy, Brachycome Iberidifolia",
  "n11942144": "Brickellia, genus Brickelia",
  "n11942366": "Buphthalmum, genus Buphthalmum",
  "n11942487": "oxeye",
  "n11942659": "woodland oxeye, Buphthalmum salicifolium",
  "n11942875": "Cacalia, genus Cacalia",
  "n11943133": "Indian plantain",
  "n11943299": "genus Calendula",
  "n11943407": "calendula",
  "n11943660": "common marigold, pot marigold, ruddles, Scotch marigold, Calendula officinalis",
  "n11943824": "Callistephus, genus Callistephus",
  "n11943992": "China aster, Callistephus chinensis",
  "n11944196": "thistle",
  "n11944569": "Carduus, genus Carduus",
  "n11944751": "welted thistle, Carduus crispus",
  "n11944954": "musk thistle, nodding thistle, Carduus nutans",
  "n11945228": "Carlina, genus Carlina",
  "n11945367": "carline thistle",
  "n11945514": "stemless carline thistle, Carlina acaulis",
  "n11945783": "common carline thistle, Carlina vulgaris",
  "n11945930": "Carthamus, genus Carthamus",
  "n11946051": "safflower, false saffron, Carthamus tinctorius",
  "n11946313": "safflower seed",
  "n11946433": "safflower oil",
  "n11946584": "genus Catananche",
  "n11946727": "catananche",
  "n11946918": "blue succory, cupid's dart, Catananche caerulea",
  "n11947079": "Centaurea, genus Centaurea",
  "n11947251": "centaury",
  "n11947429": "basket flower, Centaurea americana",
  "n11947629": "dusty miller, Centaurea cineraria, Centaurea gymnocarpa",
  "n11947802": "cornflower, bachelor's button, bluebottle, Centaurea cyanus",
  "n11948044": "star-thistle, caltrop, Centauria calcitrapa",
  "n11948264": "knapweed",
  "n11948469": "sweet sultan, Centaurea imperialis",
  "n11948656": "lesser knapweed, black knapweed, hardheads, Centaurea nigra",
  "n11948864": "great knapweed, greater knapweed, Centaurea scabiosa",
  "n11949015": "Barnaby's thistle, yellow star-thistle, Centaurea solstitialis",
  "n11949217": "Chamaemelum, genus Chamaemelum",
  "n11949402": "chamomile, camomile, Chamaemelum nobilis, Anthemis nobilis",
  "n11949707": "genus Chaenactis",
  "n11949857": "chaenactis",
  "n11950028": "genus Chrysanthemum",
  "n11950345": "chrysanthemum",
  "n11950686": "corn marigold, field marigold, Chrysanthemum segetum",
  "n11950877": "crown daisy, Chrysanthemum coronarium",
  "n11951052": "chop-suey greens, tong ho, shun giku, Chrysanthemum coronarium spatiosum",
  "n11951271": "chrysanthemum",
  "n11951385": "Chrysopsis, genus Chrysopsis",
  "n11951511": "golden aster",
  "n11951820": "Maryland golden aster, Chrysopsis mariana",
  "n11951961": "grass-leaved golden aster",
  "n11952058": "sickleweed golden aster",
  "n11952153": "Chrysothamnus, genus Chrysothamnus",
  "n11952346": "goldenbush",
  "n11952541": "rabbit brush, rabbit bush, Chrysothamnus nauseosus",
  "n11952900": "Cichorium, genus Cichorium",
  "n11953038": "chicory, succory, chicory plant, Cichorium intybus",
  "n11953339": "endive, witloof, Cichorium endivia",
  "n11953610": "chicory, chicory root",
  "n11953762": "Cirsium, genus Cirsium",
  "n11953884": "plume thistle, plumed thistle",
  "n11954161": "Canada thistle, creeping thistle, Cirsium arvense",
  "n11954345": "field thistle, Cirsium discolor",
  "n11954484": "woolly thistle, Cirsium flodmanii",
  "n11954642": "European woolly thistle, Cirsium eriophorum",
  "n11954798": "melancholy thistle, Cirsium heterophylum, Cirsium helenioides",
  "n11955040": "brook thistle, Cirsium rivulare",
  "n11955153": "bull thistle, boar thistle, spear thistle, Cirsium vulgare, Cirsium lanceolatum",
  "n11955398": "Cnicus, genus Cnicus",
  "n11955532": "blessed thistle, sweet sultan, Cnicus benedictus",
  "n11955770": "Conoclinium, genus Conoclinium",
  "n11955896": "mistflower, mist-flower, ageratum, Conoclinium coelestinum, Eupatorium coelestinum",
  "n11956208": "Conyza, genus Conyza",
  "n11956348": "horseweed, Canadian fleabane, fleabane, Conyza canadensis, Erigeron canadensis",
  "n11956671": "genus Coreopsis",
  "n11956850": "coreopsis, tickseed, tickweed, tick-weed",
  "n11957162": "subgenus Calliopsis",
  "n11957317": "giant coreopsis, Coreopsis gigantea",
  "n11957514": "sea dahlia, Coreopsis maritima",
  "n11957678": "calliopsis, Coreopsis tinctoria",
  "n11957912": "genus Cosmos",
  "n11958080": "cosmos, cosmea",
  "n11958316": "Cotula, genus Cotula",
  "n11958499": "brass buttons, Cotula coronopifolia",
  "n11958742": "Craspedia, genus Craspedia",
  "n11958888": "billy buttons",
  "n11959104": "Crepis, genus Crepis",
  "n11959259": "hawk's-beard, hawk's-beards",
  "n11959489": "Cynara, genus Cynara",
  "n11959632": "artichoke, globe artichoke, artichoke plant, Cynara scolymus",
  "n11959862": "cardoon, Cynara cardunculus",
  "n11960084": "genus Dahlia",
  "n11960245": "dahlia, Dahlia pinnata",
  "n11960540": "Delairea, genus Delairea",
  "n11960673": "German ivy, Delairea odorata, Senecio milkanioides",
  "n11960943": "Dendranthema, genus Dendranthema",
  "n11961100": "florist's chrysanthemum, florists' chrysanthemum, mum, Dendranthema grandifloruom, Chrysanthemum morifolium",
  "n11961266": "Dimorphotheca, genus Dimorphotheca",
  "n11961446": "cape marigold, sun marigold, star of the veldt",
  "n11961686": "Doronicum, genus Doronicum",
  "n11961871": "leopard's-bane, leopardbane",
  "n11962108": "Echinacea, genus Echinacea",
  "n11962272": "coneflower",
  "n11962500": "Echinops, genus Echinops",
  "n11962667": "globe thistle",
  "n11962853": "Elephantopus, genus Elephantopus",
  "n11962994": "elephant's-foot",
  "n11963158": "Emilia, genus Emilia",
  "n11963305": "tassel flower, Emilia coccinea, Emilia javanica, Emilia flammea, Cacalia javanica, Cacalia lutea",
  "n11963572": "tassel flower, Emilia sagitta",
  "n11963755": "Encelia, genus Encelia",
  "n11963932": "brittlebush, brittle bush, incienso, Encelia farinosa",
  "n11964269": "Enceliopsis, genus Enceliopsis",
  "n11964446": "sunray, Enceliopsis nudicaulis",
  "n11964688": "genus Engelmannia",
  "n11964848": "engelmannia",
  "n11965054": "genus Erechtites",
  "n11965218": "fireweed, Erechtites hieracifolia",
  "n11965378": "Erigeron, genus Erigeron",
  "n11965627": "fleabane",
  "n11965962": "blue fleabane, Erigeron acer",
  "n11966083": "daisy fleabane, Erigeron annuus",
  "n11966215": "orange daisy, orange fleabane, Erigeron aurantiacus",
  "n11966385": "spreading fleabane, Erigeron divergens",
  "n11966617": "seaside daisy, beach aster, Erigeron glaucous",
  "n11966896": "Philadelphia fleabane, Erigeron philadelphicus",
  "n11967142": "robin's plantain, Erigeron pulchellus",
  "n11967315": "showy daisy, Erigeron speciosus",
  "n11967572": "Eriophyllum, genus Eriophyllum",
  "n11967744": "woolly sunflower",
  "n11967878": "golden yarrow, Eriophyllum lanatum",
  "n11968104": "Eupatorium, genus Eupatorium",
  "n11968335": "hemp agrimony, Eupatorium cannabinum",
  "n11968519": "dog fennel, Eupatorium capillifolium",
  "n11968704": "Joe-Pye weed, spotted Joe-Pye weed, Eupatorium maculatum",
  "n11968931": "boneset, agueweed, thoroughwort, Eupatorium perfoliatum",
  "n11969166": "Joe-Pye weed, purple boneset, trumpet weed, marsh milkweed, Eupatorium purpureum",
  "n11969410": "Felicia, genus Felicia",
  "n11969607": "blue daisy, blue marguerite, Felicia amelloides",
  "n11969806": "kingfisher daisy, Felicia bergeriana",
  "n11969977": "genus Filago",
  "n11970101": "cotton rose, cudweed, filago",
  "n11970298": "herba impia, Filago germanica",
  "n11970429": "genus Gaillardia",
  "n11970586": "gaillardia",
  "n11970846": "blanket flower, Indian blanket, fire wheel, fire-wheel, Gaillardia pulchella",
  "n11971094": "genus Gazania",
  "n11971248": "gazania",
  "n11971406": "treasure flower, Gazania rigens",
  "n11971600": "Gerbera, genus Gerbera",
  "n11971783": "African daisy",
  "n11971927": "Barberton daisy, Transvaal daisy, Gerbera jamesonii",
  "n11972141": "Gerea, genus Gerea",
  "n11972291": "desert sunflower, Gerea canescens",
  "n11972569": "Gnaphalium, genus Gnaphalium",
  "n11972759": "cudweed",
  "n11972959": "chafeweed, wood cudweed, Gnaphalium sylvaticum",
  "n11973159": "Grindelia, genus Grindelia",
  "n11973341": "gumweed, gum plant, tarweed, rosinweed",
  "n11973634": "Grindelia robusta",
  "n11973749": "curlycup gumweed, Grindelia squarrosa",
  "n11973888": "Gutierrezia, genus Gutierrezia",
  "n11974126": "matchweed, matchbush",
  "n11974373": "little-head snakeweed, Gutierrezia microcephala",
  "n11974557": "rabbitweed, rabbit-weed, snakeweed, broom snakeweed, broom snakeroot, turpentine weed, Gutierrezia sarothrae",
  "n11974888": "broomweed, broom-weed, Gutierrezia texana",
  "n11975100": "Gynura, genus Gynura",
  "n11975254": "velvet plant, purple velvet plant, royal velvet plant, Gynura aurantiaca",
  "n11975482": "Haastia, genus Haastia",
  "n11975658": "vegetable sheep, sheep plant, Haastia pulvinaris",
  "n11975853": "Haplopappus, genus Haplopappus",
  "n11976170": "goldenbush",
  "n11976314": "camphor daisy, Haplopappus phyllocephalus",
  "n11976511": "yellow spiny daisy, Haplopappus spinulosus",
  "n11976715": "Hazardia, genus Hazardia",
  "n11976933": "hoary golden bush, Hazardia cana",
  "n11977125": "Helenium, genus Helenium",
  "n11977303": "sneezeweed",
  "n11977512": "autumn sneezeweed, Helenium autumnale",
  "n11977660": "orange sneezeweed, owlclaws, Helenium hoopesii",
  "n11977887": "rosilla, Helenium puberulum",
  "n11978035": "genus Helianthus",
  "n11978233": "sunflower, helianthus",
  "n11978551": "swamp sunflower, Helianthus angustifolius",
  "n11978713": "common sunflower, mirasol, Helianthus annuus",
  "n11978961": "giant sunflower, tall sunflower, Indian potato, Helianthus giganteus",
  "n11979187": "showy sunflower, Helianthus laetiflorus",
  "n11979354": "Maximilian's sunflower, Helianthus maximilianii",
  "n11979527": "prairie sunflower, Helianthus petiolaris",
  "n11979715": "Jerusalem artichoke, girasol, Jerusalem artichoke sunflower, Helianthus tuberosus",
  "n11979964": "Jerusalem artichoke",
  "n11980088": "Helichrysum, genus Helichrysum",
  "n11980318": "strawflower, golden everlasting, yellow paper daisy, Helichrysum bracteatum",
  "n11980577": "genus Heliopsis",
  "n11980682": "heliopsis, oxeye",
  "n11980867": "Helipterum, genus Helipterum",
  "n11981192": "strawflower",
  "n11981314": "Heterotheca, genus Heterotheca",
  "n11981475": "hairy golden aster, prairie golden aster, Heterotheca villosa, Chrysopsis villosa",
  "n11981817": "Hieracium, genus Hieracium",
  "n11982115": "hawkweed",
  "n11982342": "king devil, yellow hawkweed, Hieracium praealtum",
  "n11982545": "rattlesnake weed, Hieracium venosum",
  "n11982724": "Homogyne, genus Homogyne",
  "n11982939": "alpine coltsfoot, Homogyne alpina, Tussilago alpina",
  "n11983160": "Hulsea, genus Hulsea",
  "n11983375": "alpine gold, alpine hulsea, Hulsea algida",
  "n11983606": "dwarf hulsea, Hulsea nana",
  "n11983739": "Hyalosperma, genus Hyalosperma",
  "n11983910": "Hypochaeris, genus Hypochaeris, Hypochoeris, genus Hypochoeris",
  "n11984144": "cat's-ear, California dandelion, capeweed, gosmore, Hypochaeris radicata",
  "n11984397": "genus Inula",
  "n11984542": "inula",
  "n11984659": "elecampane, Inula helenium",
  "n11984854": "genus Iva",
  "n11985053": "marsh elder, iva",
  "n11985321": "burweed marsh elder, false ragweed, Iva xanthifolia",
  "n11985586": "genus Krigia",
  "n11985739": "krigia",
  "n11985903": "dwarf dandelion, Krigia dandelion, Krigia bulbosa",
  "n11986091": "Lactuca, genus Lactuca",
  "n11986306": "lettuce",
  "n11986511": "garden lettuce, common lettuce, Lactuca sativa",
  "n11986729": "cos lettuce, romaine lettuce, Lactuca sativa longifolia",
  "n11986900": "head lettuce, Lactuca sativa capitata",
  "n11987126": "leaf lettuce, Lactuca sativa crispa",
  "n11987349": "celtuce, stem lettuce, Lactuca sativa asparagina",
  "n11987511": "prickly lettuce, horse thistle, Lactuca serriola, Lactuca scariola",
  "n11987722": "Lagenophera, genus Lagenophera",
  "n11987956": "Lasthenia, genus Lasthenia",
  "n11988132": "goldfields, Lasthenia chrysostoma",
  "n11988419": "Layia, genus Layia",
  "n11988596": "tidytips, tidy tips, Layia platyglossa",
  "n11988774": "Leontodon, genus Leontodon",
  "n11988893": "hawkbit",
  "n11989087": "fall dandelion, arnica bud, Leontodon autumnalis",
  "n11989266": "Leontopodium, genus Leontopodium",
  "n11989393": "edelweiss, Leontopodium alpinum",
  "n11989636": "Leucanthemum, genus Leucanthemum",
  "n11989869": "oxeye daisy, ox-eyed daisy, marguerite, moon daisy, white daisy, Leucanthemum vulgare, Chrysanthemum leucanthemum",
  "n11990167": "oxeye daisy, Leucanthemum maximum, Chrysanthemum maximum",
  "n11990313": "shasta daisy, Leucanthemum superbum, Chrysanthemum maximum maximum",
  "n11990627": "Pyrenees daisy, Leucanthemum lacustre, Chrysanthemum lacustre",
  "n11990804": "Leucogenes, genus Leucogenes",
  "n11990920": "north island edelweiss, Leucogenes leontopodium",
  "n11991080": "Liatris, genus Liatris",
  "n11991263": "blazing star, button snakeroot, gayfeather, gay-feather, snakeroot",
  "n11991549": "dotted gayfeather, Liatris punctata",
  "n11991777": "dense blazing star, Liatris pycnostachya",
  "n11991993": "Ligularia, genus Ligularia",
  "n11992166": "leopard plant",
  "n11992340": "Lindheimera, genus Lindheimera",
  "n11992479": "Texas star, Lindheimera texana",
  "n11992674": "Lonas, genus Lonas",
  "n11992806": "African daisy, yellow ageratum, Lonas inodora, Lonas annua",
  "n11993007": "Machaeranthera, genus Machaeranthera",
  "n11993203": "tahoka daisy, tansy leaf aster, Machaeranthera tanacetifolia",
  "n11993444": "sticky aster, Machaeranthera bigelovii",
  "n11993675": "Mojave aster, Machaeranthera tortifoloia",
  "n11993932": "Madia, genus Madia",
  "n11994150": "tarweed",
  "n11994336": "common madia, common tarweed, Madia elegans",
  "n11994527": "melosa, Chile tarweed, madia oil plant, Madia sativa",
  "n11994718": "madia oil",
  "n11994827": "Matricaria, genus Matricaria",
  "n11995092": "sweet false chamomile, wild chamomile, German chamomile, Matricaria recutita, Matricaria chamomilla",
  "n11995396": "pineapple weed, rayless chamomile, Matricaria matricarioides",
  "n11995683": "Melampodium, genus Melampodium",
  "n11995840": "blackfoot daisy, Melampodium leucanthum",
  "n11996092": "Mikania, genus Mikania",
  "n11996251": "climbing hempweed, climbing boneset, wild climbing hempweed, climbing hemp-vine, Mikania scandens",
  "n11996490": "genus Mutisia",
  "n11996677": "mutisia",
  "n11996792": "Nabalus, genus Nabalus",
  "n11997032": "rattlesnake root",
  "n11997160": "white lettuce, cankerweed, Nabalus alba, Prenanthes alba",
  "n11997409": "lion's foot, gall of the earth, Nabalus serpentarius, Prenanthes serpentaria",
  "n11997775": "Olearia, genus Olearia",
  "n11997969": "daisybush, daisy-bush, daisy bush",
  "n11998317": "muskwood, Olearia argophylla",
  "n11998492": "New Zealand daisybush, Olearia haastii",
  "n11998648": "Onopordum, genus Onopordum, Onopordon, genus Onopordon",
  "n11998888": "cotton thistle, woolly thistle, Scotch thistle, Onopordum acanthium, Onopordon acanthium",
  "n11999140": "genus Othonna",
  "n11999278": "othonna",
  "n11999455": "Ozothamnus, genus Ozothamnus",
  "n11999656": "cascade everlasting, Ozothamnus secundiflorus, Helichrysum secundiflorum",
  "n11999958": "Packera, genus Packera",
  "n12000191": "butterweed",
  "n12000356": "golden groundsel, golden ragwort, Packera aurea, Senecio aureus",
  "n12000609": "Parthenium, genus Parthenium",
  "n12000851": "guayule, Parthenium argentatum",
  "n12001077": "bastard feverfew, Parthenium hysterophorus",
  "n12001294": "American feverfew, wild quinine, prairie dock, Parthenium integrifolium",
  "n12001565": "Pericallis, genus Pericallis",
  "n12001707": "cineraria, Pericallis cruenta, Senecio cruentus",
  "n12001924": "florest's cineraria, Pericallis hybrida",
  "n12002197": "Petasites, genus Petasites",
  "n12002428": "butterbur, bog rhubarb, Petasites hybridus, Petasites vulgaris",
  "n12002651": "winter heliotrope, sweet coltsfoot, Petasites fragrans",
  "n12002826": "sweet coltsfoot, Petasites sagitattus",
  "n12002957": "Picris, genus Picris",
  "n12003167": "oxtongue, bristly oxtongue, bitterweed, bugloss, Picris echioides",
  "n12003407": "Pilosella, genus Pilosella",
  "n12003696": "hawkweed",
  "n12003814": "orange hawkweed, Pilosella aurantiaca, Hieracium aurantiacum",
  "n12004120": "mouse-ear hawkweed, Pilosella officinarum, Hieracium pilocella",
  "n12004310": "Piqueria, genus Piqueria",
  "n12004547": "stevia",
  "n12004686": "Prenanthes, genus Prenanthes",
  "n12004987": "rattlesnake root, Prenanthes purpurea",
  "n12005148": "genus Pteropogon",
  "n12005329": "pteropogon, Pteropogon humboltianum",
  "n12005500": "Pulicaria, genus Pulicaria",
  "n12005656": "fleabane, feabane mullet, Pulicaria dysenterica",
  "n12005869": "Pyrethrum, genus Pyrethrum",
  "n12006081": "Raoulia, genus Raoulia",
  "n12006306": "sheep plant, vegetable sheep, Raoulia lutescens, Raoulia australis",
  "n12006503": "Ratibida, genus Ratibida",
  "n12006766": "coneflower",
  "n12006930": "Mexican hat, Ratibida columnaris",
  "n12007196": "long-head coneflower, prairie coneflower, Ratibida columnifera",
  "n12007406": "prairie coneflower, Ratibida tagetes",
  "n12007560": "genus Rhodanthe",
  "n12007766": "Swan River everlasting, rhodanthe, Rhodanthe manglesii, Helipterum manglesii",
  "n12008017": "Rudbeckia, genus Rudbeckia",
  "n12008252": "coneflower",
  "n12008487": "black-eyed Susan, Rudbeckia hirta, Rudbeckia serotina",
  "n12008749": "cutleaved coneflower, Rudbeckia laciniata",
  "n12009047": "golden glow, double gold, hortensia, Rudbeckia laciniata hortensia",
  "n12009250": "Santolina, genus Santolina",
  "n12009420": "lavender cotton, Santolina chamaecyparissus",
  "n12009616": "Sanvitalia, genus Sanvitalia",
  "n12009792": "creeping zinnia, Sanvitalia procumbens",
  "n12010021": "Saussurea, genus Saussurea",
  "n12010188": "costusroot, Saussurea costus, Saussurea lappa",
  "n12010458": "Scolymus, genus Scolymus",
  "n12010628": "golden thistle",
  "n12010815": "Spanish oyster plant, Scolymus hispanicus",
  "n12011067": "Senecio, genus Senecio",
  "n12011370": "nodding groundsel, Senecio bigelovii",
  "n12011620": "dusty miller, Senecio cineraria, Cineraria maritima",
  "n12011838": "threadleaf groundsel, Senecio doublasii",
  "n12012111": "butterweed, ragwort, Senecio glabellus",
  "n12012253": "ragwort, tansy ragwort, ragweed, benweed, Senecio jacobaea",
  "n12012510": "arrowleaf groundsel, Senecio triangularis",
  "n12012755": "groundsel, Senecio vulgaris",
  "n12012897": "genus Scorzonera",
  "n12013035": "black salsify, viper's grass, scorzonera, Scorzonera hispanica",
  "n12013323": "Sericocarpus, genus Sericocarpus",
  "n12013511": "white-topped aster",
  "n12013701": "narrow-leaved white-topped aster",
  "n12013811": "Seriphidium, genus Seriphidium",
  "n12014085": "silver sage, silver sagebrush, grey sage, gray sage, Seriphidium canum, Artemisia cana",
  "n12014355": "sea wormwood, Seriphidium maritimum, Artemisia maritima",
  "n12014524": "big sagebrush, blue sage, Seriphidium tridentatum, Artemisia tridentata",
  "n12014739": "Serratula, genus Serratula",
  "n12014923": "sawwort, Serratula tinctoria",
  "n12015076": "Silphium, genus Silphium",
  "n12015221": "rosinweed, Silphium laciniatum",
  "n12015384": "Silybum, genus Silybum",
  "n12015525": "milk thistle, lady's thistle, Our Lady's mild thistle, holy thistle, blessed thistle, Silybum marianum",
  "n12015840": "Solidago, genus Solidago",
  "n12015959": "goldenrod",
  "n12016434": "silverrod, Solidago bicolor",
  "n12016567": "meadow goldenrod, Canadian goldenrod, Solidago canadensis",
  "n12016777": "Missouri goldenrod, Solidago missouriensis",
  "n12016914": "alpine goldenrod, Solidago multiradiata",
  "n12017127": "grey goldenrod, gray goldenrod, Solidago nemoralis",
  "n12017326": "Blue Mountain tea, sweet goldenrod, Solidago odora",
  "n12017511": "dyer's weed, Solidago rugosa",
  "n12017664": "seaside goldenrod, beach goldenrod, Solidago sempervirens",
  "n12017853": "narrow goldenrod, Solidago spathulata",
  "n12018014": "Boott's goldenrod",
  "n12018100": "Elliott's goldenrod",
  "n12018188": "Ohio goldenrod",
  "n12018271": "rough-stemmed goldenrod",
  "n12018363": "showy goldenrod",
  "n12018447": "tall goldenrod",
  "n12018530": "zigzag goldenrod, broad leaved goldenrod",
  "n12018640": "Sonchus, genus Sonchus",
  "n12018760": "sow thistle, milk thistle",
  "n12019035": "milkweed, Sonchus oleraceus",
  "n12019190": "Stenotus, genus Stenotus",
  "n12019375": "stemless golden weed, Stenotus acaulis, Haplopappus acaulis",
  "n12019675": "genus Stevia",
  "n12019827": "stevia",
  "n12020048": "Stokesia, genus Stokesia",
  "n12020184": "stokes' aster, cornflower aster, Stokesia laevis",
  "n12020388": "Tageteste, genus Tagetes",
  "n12020507": "marigold",
  "n12020736": "African marigold, big marigold, Aztec marigold, Tagetes erecta",
  "n12020941": "French marigold, Tagetes patula",
  "n12021120": "Tanacetum, genus Tanacetum",
  "n12021499": "costmary, alecost, bible leaf, mint geranium, balsam herb, Tanacetum balsamita, Chrysanthemum balsamita",
  "n12021882": "camphor dune tansy, Tanacetum camphoratum",
  "n12022054": "painted daisy, pyrethrum, Tanacetum coccineum, Chrysanthemum coccineum",
  "n12022382": "pyrethrum, Dalmatian pyrethrum, Dalmatia pyrethrum, Tanacetum cinerariifolium, Chrysanthemum cinerariifolium",
  "n12022719": "pyrethrum",
  "n12022821": "northern dune tansy, Tanacetum douglasii",
  "n12023108": "feverfew, Tanacetum parthenium, Chrysanthemum parthenium",
  "n12023407": "dusty miller, silver-lace, silver lace, Tanacetum ptarmiciflorum, Chrysanthemum ptarmiciflorum",
  "n12023726": "tansy, golden buttons, scented fern, Tanacetum vulgare",
  "n12023996": "Taraxacum, genus Taraxacum",
  "n12024176": "dandelion, blowball",
  "n12024445": "common dandelion, Taraxacum ruderalia, Taraxacum officinale",
  "n12024690": "dandelion green",
  "n12024805": "Russian dandelion, kok-saghyz, kok-sagyz, Taraxacum kok-saghyz",
  "n12025019": "Tetraneuris, genus Tetraneuris",
  "n12025220": "stemless hymenoxys, Tetraneuris acaulis, Hymenoxys acaulis",
  "n12025507": "old man of the mountain, alpine sunflower, Tetraneuris grandiflora, Hymenoxys grandiflora",
  "n12025849": "genus Tithonia",
  "n12026018": "Mexican sunflower, tithonia",
  "n12026306": "Townsendia, genus Townsendia",
  "n12026476": "Easter daisy, stemless daisy, Townsendia Exscapa",
  "n12026764": "Tragopogon, genus Tragopogon",
  "n12026981": "yellow salsify, Tragopogon dubius",
  "n12027222": "salsify, oyster plant, vegetable oyster, Tragopogon porrifolius",
  "n12027538": "salsify, oyster plant",
  "n12027658": "meadow salsify, goatsbeard, shepherd's clock, Tragopogon pratensis",
  "n12027864": "Trilisa, genus Trilisa",
  "n12028012": "wild vanilla, Trilisa odoratissima",
  "n12028196": "Tripleurospermum, genus Tripleurospermum",
  "n12028424": "scentless camomile, scentless false camomile, scentless mayweed, scentless hayweed, corn mayweed, Tripleurospermum inodorum, Matricaria inodorum",
  "n12028818": "turfing daisy, Tripleurospermum oreades tchihatchewii, Matricaria oreades",
  "n12029039": "turfing daisy, Tripleurospermum tchihatchewii, Matricaria tchihatchewii",
  "n12029326": "Tussilago, genus Tussilago",
  "n12029635": "coltsfoot, Tussilago farfara",
  "n12029929": "genus Ursinia",
  "n12030092": "ursinia",
  "n12030265": "Verbesina, genus Verbesina",
  "n12030479": "Actinomeris, genus Actinomeris",
  "n12030654": "crownbeard, crown-beard, crown beard",
  "n12030908": "wingstem, golden ironweed, yellow ironweed, golden honey plant, Verbesina alternifolia, Actinomeris alternifolia",
  "n12031139": "cowpen daisy, golden crownbeard, golden crown beard, butter daisy, Verbesina encelioides, Ximenesia encelioides",
  "n12031388": "gravelweed, Verbesina helianthoides",
  "n12031547": "Virginia crownbeard, frostweed, frost-weed, Verbesina virginica",
  "n12031739": "genus Vernonia",
  "n12031927": "ironweed, vernonia",
  "n12032215": "genus Wyethia",
  "n12032429": "mule's ears, Wyethia amplexicaulis",
  "n12032686": "white-rayed mule's ears, Wyethia helianthoides",
  "n12032939": "Xanthium, genus Xanthium",
  "n12033139": "cocklebur, cockle-bur, cockleburr, cockle-burr",
  "n12033310": "genus Xeranthemum",
  "n12033504": "xeranthemum",
  "n12033709": "immortelle, Xeranthemum annuum",
  "n12033939": "genus Zinnia",
  "n12034141": "zinnia, old maid, old maid flower",
  "n12034384": "white zinnia, Zinnia acerosa",
  "n12034594": "little golden zinnia, Zinnia grandiflora",
  "n12034828": "Loasaceae, family Loasaceae, loasa family",
  "n12035064": "genus Loasa",
  "n12035213": "loasa",
  "n12035423": "Mentzelia, genus Mentzelia",
  "n12035631": "blazing star, Mentzelia livicaulis, Mentzelia laevicaulis",
  "n12035907": "bartonia, Mentzelia lindleyi",
  "n12036067": "achene",
  "n12036226": "samara, key fruit, key",
  "n12036368": "bur, burr",
  "n12036533": "Campanulaceae, family Campanulaceae, bellflower family",
  "n12036781": "genus Campanula",
  "n12036939": "campanula, bellflower",
  "n12037328": "harebell, bluebell, Campanula rotundifolia",
  "n12037499": "creeping bellflower, Campanula rapunculoides",
  "n12037691": "Canterbury bell, cup and saucer, Campanula medium",
  "n12037864": "southern harebell, Campanula divaricata",
  "n12038038": "tall bellflower, Campanula americana",
  "n12038208": "marsh bellflower, Campanula aparinoides",
  "n12038406": "clustered bellflower, Campanula glomerata",
  "n12038585": "peach bells, peach bell, willow bell, Campanula persicifolia",
  "n12038760": "chimney plant, chimney bellflower, Campanula pyramidalis",
  "n12038898": "rampion, rampion bellflower, Campanula rapunculus",
  "n12039122": "throatwort, nettle-leaved bellflower, Campanula trachelium",
  "n12039317": "tussock bellflower, spreading bellflower, Campanula carpatica",
  "n12039524": "Orchidales, order Orchidales",
  "n12039743": "Orchidaceae, family Orchidaceae, orchid family",
  "n12041446": "orchid, orchidaceous plant",
  "n12043248": "genus Orchis",
  "n12043444": "orchis",
  "n12043673": "male orchis, early purple orchid, Orchis mascula",
  "n12043836": "butterfly orchid, butterfly orchis, Orchis papilionaceae",
  "n12044041": "showy orchis, purple orchis, purple-hooded orchis, Orchis spectabilis",
  "n12044269": "genus Aerides",
  "n12044467": "aerides",
  "n12044571": "genus Angrecum, Angraecum, genus Angraecum",
  "n12044784": "angrecum",
  "n12045004": "Anoectochilus, genus Anoectochilus",
  "n12045157": "jewel orchid",
  "n12045352": "Aplectrum, genus Aplectrum",
  "n12045514": "puttyroot, adam-and-eve, Aplectrum hyemale",
  "n12045695": "genus Arethusa",
  "n12045860": "arethusa",
  "n12046028": "bog rose, wild pink, dragon's mouth, Arethusa bulbosa",
  "n12046251": "genus Bletia",
  "n12046428": "bletia",
  "n12046620": "Bletilla, genus Bletilla",
  "n12046815": "Bletilla striata, Bletia striata",
  "n12047060": "pseudobulb",
  "n12047173": "genus Brassavola",
  "n12047345": "brassavola",
  "n12047586": "Brassia, genus Brassia",
  "n12047884": "spider orchid, Brassia lawrenceana",
  "n12048056": "spider orchid, Brassia verrucosa",
  "n12048231": "genus Caladenia",
  "n12048399": "caladenia",
  "n12048537": "zebra orchid, Caladenia cairnsiana",
  "n12048772": "genus Calanthe",
  "n12048928": "calanthe",
  "n12049134": "Calopogon, genus Calopogon",
  "n12049282": "grass pink, Calopogon pulchellum, Calopogon tuberosum",
  "n12049412": "genus Calypso",
  "n12049562": "calypso, fairy-slipper, Calypso bulbosa",
  "n12049796": "Catasetum, genus Catasetum",
  "n12050014": "jumping orchid, Catasetum macrocarpum",
  "n12050295": "genus Cattleya",
  "n12050533": "cattleya",
  "n12050766": "Cephalanthera, genus Cephalanthera",
  "n12050959": "helleborine",
  "n12051103": "red helleborine, Cephalanthera rubra",
  "n12051285": "Cleistes, genus Cleistes",
  "n12051514": "spreading pogonia, funnel-crest rosebud orchid, Cleistes divaricata, Pogonia divaricata",
  "n12051792": "rosebud orchid, Cleistes rosea, Pogonia rosea",
  "n12052053": "Coeloglossum, genus Coeloglossum",
  "n12052267": "satyr orchid, Coeloglossum bracteatum",
  "n12052447": "frog orchid, Coeloglossum viride",
  "n12052630": "genus Coelogyne",
  "n12052787": "coelogyne",
  "n12053138": "Corallorhiza, genus Corallorhiza",
  "n12053405": "coral root",
  "n12053690": "spotted coral root, Corallorhiza maculata",
  "n12053962": "striped coral root, Corallorhiza striata",
  "n12054195": "early coral root, pale coral root, Corallorhiza trifida",
  "n12054499": "Coryanthes, genus Coryanthes",
  "n12054678": "helmetflower, helmet orchid",
  "n12054902": "Cycnoches, genus Cycnoches",
  "n12055073": "swan orchid, swanflower, swan-flower, swanneck, swan-neck",
  "n12055317": "genus Cymbidium",
  "n12055516": "cymbid, cymbidium",
  "n12055839": "Cypripedium, genus Cypripedium",
  "n12056099": "cypripedia",
  "n12056217": "lady's slipper, lady-slipper, ladies' slipper, slipper orchid",
  "n12056601": "moccasin flower, nerveroot, Cypripedium acaule",
  "n12056758": "common lady's-slipper, showy lady's-slipper, showy lady slipper, Cypripedium reginae, Cypripedium album",
  "n12056990": "ram's-head, ram's-head lady's slipper, Cypripedium arietinum",
  "n12057211": "yellow lady's slipper, yellow lady-slipper, Cypripedium calceolus, Cypripedium parviflorum",
  "n12057447": "large yellow lady's slipper, Cypripedium calceolus pubescens",
  "n12057660": "California lady's slipper, Cypripedium californicum",
  "n12057895": "clustered lady's slipper, Cypripedium fasciculatum",
  "n12058192": "mountain lady's slipper, Cypripedium montanum",
  "n12058429": "Dactylorhiza, genus Dactylorhiza",
  "n12058630": "marsh orchid",
  "n12058822": "common spotted orchid, Dactylorhiza fuchsii, Dactylorhiza maculata fuchsii",
  "n12059090": "genus Dendrobium",
  "n12059314": "dendrobium",
  "n12059479": "genus Disa",
  "n12059625": "disa",
  "n12059851": "Dracula, genus Dracula",
  "n12060118": "Dryadella, genus Dryadella",
  "n12060380": "Eburophyton, genus Eburophyton",
  "n12060546": "phantom orchid, snow orchid, Eburophyton austinae",
  "n12060816": "Encyclia, genus Encyclia",
  "n12061104": "tulip orchid, Encyclia citrina, Cattleya citrina",
  "n12061380": "butterfly orchid, Encyclia tampensis, Epidendrum tampense",
  "n12061614": "butterfly orchid, butterfly orchis, Epidendrum venosum, Encyclia venosa",
  "n12061849": "Epidendrum, genus Epidendrum",
  "n12062105": "epidendron",
  "n12062227": "Epipactis, genus Epipactis",
  "n12062468": "helleborine",
  "n12062626": "Epipactis helleborine",
  "n12062781": "stream orchid, chatterbox, giant helleborine, Epipactis gigantea",
  "n12063066": "Glossodia, genus Glossodia",
  "n12063211": "tongueflower, tongue-flower",
  "n12063414": "Goodyera, genus Goodyera",
  "n12063639": "rattlesnake plantain, helleborine",
  "n12063887": "Grammatophyllum, genus Grammatophyllum",
  "n12064183": "Gymnadenia, genus Gymnadenia",
  "n12064389": "fragrant orchid, Gymnadenia conopsea",
  "n12064591": "short-spurred fragrant orchid, Gymnadenia odoratissima",
  "n12064814": "Gymnadeniopsis, genus Gymnadeniopsis",
  "n12064996": "Habenaria, genus Habenaria",
  "n12065316": "fringed orchis, fringed orchid",
  "n12065649": "frog orchid",
  "n12065777": "rein orchid, rein orchis",
  "n12066018": "bog rein orchid, bog candles, Habenaria dilatata",
  "n12066261": "white fringed orchis, white fringed orchid, Habenaria albiflora",
  "n12066451": "elegant Habenaria, Habenaria elegans",
  "n12066630": "purple-fringed orchid, purple-fringed orchis, Habenaria fimbriata",
  "n12066821": "coastal rein orchid, Habenaria greenei",
  "n12067029": "Hooker's orchid, Habenaria hookeri",
  "n12067193": "ragged orchid, ragged orchis, ragged-fringed orchid, green fringed orchis, Habenaria lacera",
  "n12067433": "prairie orchid, prairie white-fringed orchis, Habenaria leucophaea",
  "n12067672": "snowy orchid, Habenaria nivea",
  "n12067817": "round-leaved rein orchid, Habenaria orbiculata",
  "n12068138": "purple fringeless orchid, purple fringeless orchis, Habenaria peramoena",
  "n12068432": "purple-fringed orchid, purple-fringed orchis, Habenaria psycodes",
  "n12068615": "Alaska rein orchid, Habenaria unalascensis",
  "n12068824": "Hexalectris, genus Hexalectris",
  "n12069009": "crested coral root, Hexalectris spicata",
  "n12069217": "Texas purple spike, Hexalectris warnockii",
  "n12069488": "Himantoglossum, genus Himantoglossum",
  "n12069679": "lizard orchid, Himantoglossum hircinum",
  "n12069821": "genus Laelia",
  "n12070016": "laelia",
  "n12070177": "genus Liparis",
  "n12070381": "liparis",
  "n12070583": "twayblade",
  "n12070712": "fen orchid, fen orchis, Liparis loeselii",
  "n12070950": "Listera, genus Listera",
  "n12071259": "broad-leaved twayblade, Listera convallarioides",
  "n12071477": "lesser twayblade, Listera cordata",
  "n12071744": "twayblade, Listera ovata",
  "n12071965": "Malaxis, genus Malaxis",
  "n12072210": "green adder's mouth, Malaxis-unifolia, Malaxis ophioglossoides",
  "n12072419": "genus Masdevallia",
  "n12072722": "masdevallia",
  "n12073007": "genus Maxillaria",
  "n12073217": "maxillaria",
  "n12073410": "Miltonia, genus Miltonia",
  "n12073554": "pansy orchid",
  "n12073744": "genus Odontoglossum",
  "n12073991": "odontoglossum",
  "n12074205": "genus Oncidium",
  "n12074408": "oncidium, dancing lady orchid, butterfly plant, butterfly orchid",
  "n12074678": "Ophrys, genus Ophrys",
  "n12074867": "bee orchid, Ophrys apifera",
  "n12075010": "fly orchid, Ophrys insectifera, Ophrys muscifera",
  "n12075151": "spider orchid",
  "n12075299": "early spider orchid, Ophrys sphegodes",
  "n12075495": "Paphiopedilum, genus Paphiopedilum",
  "n12075830": "Venus' slipper, Venus's slipper, Venus's shoe",
  "n12076075": "genus Phaius",
  "n12076223": "phaius",
  "n12076381": "Phalaenopsis, genus Phalaenopsis",
  "n12076577": "moth orchid, moth plant",
  "n12076852": "butterfly plant, Phalaenopsis amabilis",
  "n12077062": "Pholidota, genus Pholidota",
  "n12077244": "rattlesnake orchid",
  "n12077505": "Phragmipedium, genus Phragmipedium",
  "n12077732": "Platanthera, genus Platanthera",
  "n12077944": "lesser butterfly orchid, Platanthera bifolia, Habenaria bifolia",
  "n12078172": "greater butterfly orchid, Platanthera chlorantha, Habenaria chlorantha",
  "n12078451": "prairie white-fringed orchid, Platanthera leucophea",
  "n12078596": "Plectorrhiza, genus Plectorrhiza",
  "n12078747": "tangle orchid",
  "n12078954": "Pleione, genus Pleione",
  "n12079120": "Indian crocus",
  "n12079352": "genus Pleurothallis",
  "n12079523": "pleurothallis",
  "n12079737": "genus Pogonia",
  "n12079963": "pogonia",
  "n12080199": "Psychopsis, genus Psychopsis",
  "n12080395": "butterfly orchid",
  "n12080588": "Psychopsis krameriana, Oncidium papilio kramerianum",
  "n12080820": "Psychopsis papilio, Oncidium papilio",
  "n12081022": "Pterostylis, genus Pterostylis",
  "n12081215": "helmet orchid, greenhood",
  "n12081488": "Rhyncostylis, genus Rhyncostylis",
  "n12081649": "foxtail orchid",
  "n12081851": "Sarcochilus, genus Sarcochilus",
  "n12082131": "orange-blossom orchid, Sarcochilus falcatus",
  "n12082357": "Scaphosepalum, genus Scaphosepalum",
  "n12082593": "Schomburgkia, genus Schomburgkia",
  "n12082764": "Selenipedium, genus Selenipedium",
  "n12082980": "genus Sobralia",
  "n12083113": "sobralia",
  "n12083339": "Spiranthes, genus Spiranthes",
  "n12083591": "ladies' tresses, lady's tresses",
  "n12083847": "screw augur, Spiranthes cernua",
  "n12084158": "hooded ladies' tresses, Spiranthes romanzoffiana",
  "n12084400": "western ladies' tresses, Spiranthes porrifolia",
  "n12084555": "European ladies' tresses, Spiranthes spiralis",
  "n12084746": "genus Stanhopea",
  "n12084890": "stanhopea",
  "n12085117": "genus Stelis",
  "n12085267": "stelis",
  "n12085469": "Trichoceros, genus Trichoceros",
  "n12085664": "fly orchid",
  "n12085840": "genus Vanda",
  "n12086012": "vanda",
  "n12086192": "blue orchid, Vanda caerulea",
  "n12086362": "genus Vanilla",
  "n12086539": "vanilla",
  "n12086778": "vanilla orchid, Vanilla planifolia",
  "n12087032": "vanillin",
  "n12087207": "Burmanniaceae, family Burmanniaceae",
  "n12087408": "Burmannia, genus Burmannia",
  "n12087650": "Dioscoreaceae, family Dioscoreaceae, yam family",
  "n12087807": "Dioscorea, genus Dioscorea",
  "n12087961": "yam, yam plant",
  "n12088223": "yam",
  "n12088327": "white yam, water yam, Dioscorea alata",
  "n12088495": "cinnamon vine, Chinese yam, Dioscorea batata",
  "n12088768": "air potato, Dioscorea bulbifera",
  "n12088909": "elephant's-foot, tortoise plant, Hottentot bread vine, Hottentot's bread vine, Dioscorea elephantipes",
  "n12089178": "Hottentot bread, Hottentot's bread",
  "n12089320": "wild yam, Dioscorea paniculata",
  "n12089496": "cush-cush, Dioscorea trifida",
  "n12089625": "Tamus, genus Tamus",
  "n12089846": "black bryony, black bindweed, Tamus communis",
  "n12090041": "Primulales, order Primulales",
  "n12090318": "Primulaceae, family Primulaceae, primrose family",
  "n12090702": "genus Primula",
  "n12090890": "primrose, primula",
  "n12091213": "English primrose, Primula vulgaris",
  "n12091377": "cowslip, paigle, Primula veris",
  "n12091550": "oxlip, paigle, Primula elatior",
  "n12091697": "Chinese primrose, Primula sinensis",
  "n12091806": "auricula, bear's ear, Primula auricula",
  "n12091953": "polyanthus, Primula polyantha",
  "n12092127": "Anagallis, genus Anagallis",
  "n12092262": "pimpernel",
  "n12092417": "scarlet pimpernel, red pimpernel, poor man's weatherglass, Anagallis arvensis",
  "n12092629": "bog pimpernel, Anagallis tenella",
  "n12092766": "Centunculus, genus Centunculus",
  "n12092930": "chaffweed, bastard pimpernel, false pimpernel",
  "n12093088": "genus Cyclamen",
  "n12093329": "cyclamen, Cyclamen purpurascens",
  "n12093600": "sowbread, Cyclamen hederifolium, Cyclamen neopolitanum",
  "n12093769": "Glaux, genus Glaux",
  "n12093885": "sea milkwort, sea trifoly, black saltwort, Glaux maritima",
  "n12094121": "Hottonia, genus Hottonia",
  "n12094244": "featherfoil, feather-foil",
  "n12094401": "water gillyflower, American featherfoil, Hottonia inflata",
  "n12094612": "water violet, Hottonia palustris",
  "n12094786": "Lysimachia, genus Lysimachia",
  "n12095020": "loosestrife",
  "n12095281": "gooseneck loosestrife, Lysimachia clethroides Duby",
  "n12095412": "yellow pimpernel, Lysimachia nemorum",
  "n12095543": "fringed loosestrife, Lysimachia ciliatum",
  "n12095647": "moneywort, creeping Jenny, creeping Charlie, Lysimachia nummularia",
  "n12095781": "yellow loosestrife, garden loosestrife, Lysimachia vulgaris",
  "n12095934": "swamp candles, Lysimachia terrestris",
  "n12096089": "whorled loosestrife, Lysimachia quadrifolia",
  "n12096223": "Samolus, genus Samolus",
  "n12096395": "water pimpernel",
  "n12096563": "brookweed, Samolus valerandii",
  "n12096674": "brookweed, Samolus parviflorus, Samolus floribundus",
  "n12096798": "Myrsinaceae, family Myrsinaceae, myrsine family",
  "n12097013": "Myrsine, genus Myrsine",
  "n12097180": "Ardisia, genus Ardisia",
  "n12097396": "coralberry, spiceberry, Ardisia crenata",
  "n12097556": "marlberry, Ardisia escallonoides, Ardisia paniculata",
  "n12097746": "Plumbaginales, order Plumbaginales",
  "n12097927": "Plumbaginaceae, family Plumbaginaceae, leadwort family, sea-lavender family",
  "n12098227": "genus Plumbago",
  "n12098403": "plumbago",
  "n12098524": "leadwort, Plumbago europaea",
  "n12098665": "Armeria, genus Armeria",
  "n12098827": "thrift",
  "n12099031": "cliff rose, sea pink, Armeria maritima",
  "n12099220": "Limonium, genus Limonium",
  "n12099342": "sea lavender, marsh rosemary, statice",
  "n12099556": "Theophrastaceae, family Theophrastaceae",
  "n12099803": "Jacquinia, genus Jacquinia",
  "n12099972": "bracelet wood, Jacquinia armillaris",
  "n12100187": "barbasco, joewood, Jacquinia keyensis",
  "n12100382": "Graminales, order Graminales",
  "n12100538": "Gramineae, family Gramineae, Graminaceae, family Graminaceae, Poaceae, family Poaceae, grass family",
  "n12101870": "gramineous plant, graminaceous plant",
  "n12102133": "grass",
  "n12103349": "beach grass",
  "n12103476": "bunchgrass, bunch grass",
  "n12103680": "midgrass",
  "n12103894": "shortgrass, short-grass",
  "n12104104": "sword grass",
  "n12104238": "tallgrass, tall-grass",
  "n12104384": "lemongrass, lemon grass",
  "n12104501": "herbage, pasturage",
  "n12104614": "Aegilops, genus Aegilops",
  "n12104734": "goat grass, Aegilops triuncalis",
  "n12104943": "Agropyron, genus Agropyron",
  "n12105125": "wheatgrass, wheat-grass",
  "n12105353": "crested wheatgrass, crested wheat grass, fairway crested wheat grass, Agropyron cristatum",
  "n12105578": "dog grass, couch grass, quackgrass, quack grass, quick grass, witch grass, witchgrass, Agropyron repens",
  "n12105828": "bearded wheatgrass, Agropyron subsecundum",
  "n12105981": "western wheatgrass, bluestem wheatgrass, Agropyron smithii",
  "n12106134": "intermediate wheatgrass, Agropyron intermedium, Elymus hispidus",
  "n12106323": "slender wheatgrass, Agropyron trachycaulum, Agropyron pauciflorum, Elymus trachycaulos",
  "n12106540": "Agrostis, genus Agrostis",
  "n12106786": "bent, bent grass, bent-grass",
  "n12107002": "velvet bent, velvet bent grass, brown bent, Rhode Island bent, dog bent, Agrostis canina",
  "n12107191": "cloud grass, Agrostis nebulosa",
  "n12107336": "creeping bent, creeping bentgrass, Agrostis palustris",
  "n12107489": "Alopecurus, genus Alopecurus",
  "n12107710": "meadow foxtail, Alopecurus pratensis",
  "n12107970": "foxtail, foxtail grass",
  "n12108249": "Andropogon, genus Andropogon",
  "n12108432": "broom grass",
  "n12108613": "broom sedge, Andropogon virginicus",
  "n12108742": "Arrhenatherum, genus Arrhenatherum",
  "n12108871": "tall oat grass, tall meadow grass, evergreen grass, false oat, French rye, Arrhenatherum elatius",
  "n12109189": "Arundo, genus Arundo",
  "n12109365": "toetoe, toitoi, Arundo conspicua, Chionochloa conspicua",
  "n12109498": "giant reed, Arundo donax",
  "n12109719": "Avena, genus Avena",
  "n12109827": "oat",
  "n12110085": "cereal oat, Avena sativa",
  "n12110236": "wild oat, wild oat grass, Avena fatua",
  "n12110352": "slender wild oat, Avena barbata",
  "n12110475": "wild red oat, animated oat, Avene sterilis",
  "n12110630": "Bromus, genus Bromus",
  "n12110778": "brome, bromegrass",
  "n12111043": "awnless bromegrass, Bromus inermis",
  "n12111238": "chess, cheat, Bromus secalinus",
  "n12111399": "downy brome, downy bromegrass, downy cheat, downy chess, cheatgrass, drooping brome, Bromus tectorum",
  "n12111627": "field brome, Bromus arvensis",
  "n12111744": "Japanese brome, Japanese chess, Bromus japonicus",
  "n12111882": "Bouteloua, genus Bouteloua",
  "n12112008": "grama, grama grass, gramma, gramma grass",
  "n12112219": "blue grama, Bouteloua gracilis",
  "n12112337": "black grama, Bouteloua eriopoda",
  "n12112488": "Buchloe, genus Buchloe",
  "n12112609": "buffalo grass, Buchloe dactyloides",
  "n12112789": "Calamagrostis, genus Calamagrostis",
  "n12112918": "reed grass",
  "n12113195": "feather reed grass, feathertop, Calamagrostis acutiflora",
  "n12113323": "Australian reed grass, Calamagrostic quadriseta",
  "n12113471": "Cenchrus, genus Cenchrus",
  "n12113657": "burgrass, bur grass",
  "n12113790": "sandbur, sandspur, field sandbur, Cenchrus tribuloides",
  "n12114010": "buffel grass, Cenchrus ciliaris, Pennisetum cenchroides",
  "n12114226": "Chloris, genus Chloris",
  "n12114397": "finger grass",
  "n12114590": "Rhodes grass, Chloris gayana",
  "n12114770": "windmill grass, creeping windmill grass, star grass, Chloris truncata",
  "n12114981": "Cortaderia, genus Cortaderia",
  "n12115180": "pampas grass, Cortaderia selloana",
  "n12115383": "plumed tussock, toe toe, toetoe, Cortaderia richardii, Arundo richardii",
  "n12115563": "Cynodon, genus Cynodon",
  "n12115748": "Bermuda grass, devil grass, Bahama grass, kweek, doob, scutch grass, star grass, Cynodon dactylon",
  "n12116058": "giant star grass, Cynodon plectostachyum",
  "n12116267": "Dactylis, genus Dactylis",
  "n12116429": "orchard grass, cocksfoot, cockspur, Dactylis glomerata",
  "n12116583": "Dactyloctenium, genus Dactyloctenium",
  "n12116734": "Egyptian grass, crowfoot grass, Dactyloctenium aegypticum",
  "n12116881": "Digitaria, genus Digitaria",
  "n12117017": "crabgrass, crab grass, finger grass",
  "n12117235": "smooth crabgrass, Digitaria ischaemum",
  "n12117326": "large crabgrass, hairy finger grass, Digitaria sanguinalis",
  "n12117507": "Echinochloa, genus Echinochloa",
  "n12117695": "barnyard grass, barn grass, barn millet, Echinochloa crusgalli",
  "n12117912": "Japanese millet, billion-dollar grass, Japanese barnyard millet, sanwa millet, Echinochloa frumentacea",
  "n12118223": "Eleusine, genus Eleusine",
  "n12118414": "yardgrass, yard grass, wire grass, goose grass, Eleusine indica",
  "n12118661": "finger millet, ragi, ragee, African millet, coracan, corakan, kurakkan, Eleusine coracana",
  "n12118912": "Elymus, genus Elymus",
  "n12119099": "lyme grass",
  "n12119238": "wild rye",
  "n12119390": "giant ryegrass, Elymus condensatus, Leymus condensatus",
  "n12119539": "sea lyme grass, European dune grass, Elymus arenarius, Leymus arenaria",
  "n12119717": "Canada wild rye, Elymus canadensis",
  "n12119822": "medusa's head, Elymus caput-medusae",
  "n12119947": "Eragrostis, genus Eragrostis",
  "n12120114": "love grass, bay grass",
  "n12120347": "teff, teff grass, Eragrostis tef, Eragrostic abyssinica",
  "n12120578": "weeping love grass, African love grass, Eragrostis curvula",
  "n12120812": "Erianthus, genus Erianthus",
  "n12121033": "plume grass",
  "n12121187": "Ravenna grass, wool grass, Erianthus ravennae",
  "n12121405": "Festuca, genus Festuca",
  "n12121610": "fescue, fescue grass, meadow fescue, Festuca elatior",
  "n12121835": "sheep fescue, sheep's fescue, Festuca ovina",
  "n12122018": "silver grass",
  "n12122124": "Glyceria, genus Glyceria",
  "n12122245": "manna grass, sweet grass",
  "n12122442": "reed meadow grass, Glyceria grandis",
  "n12122581": "Holcus, genus Holcus",
  "n12122725": "velvet grass, Yorkshire fog, Holcus lanatus",
  "n12122918": "creeping soft grass, Holcus mollis",
  "n12123050": "Hordeum, genus Hordeum",
  "n12123244": "barley",
  "n12123450": "common barley, Hordeum vulgare",
  "n12123648": "barleycorn",
  "n12123741": "barley grass, wall barley, Hordeum murinum",
  "n12123932": "squirreltail barley, foxtail barley, squirreltail grass, Hordeum jubatum",
  "n12124172": "little barley, Hordeum pusillum",
  "n12124358": "Leymus, genus Leymus",
  "n12124505": "Lolium, genus Lolium",
  "n12124627": "rye grass, ryegrass",
  "n12124818": "perennial ryegrass, English ryegrass, Lolium perenne",
  "n12125001": "Italian ryegrass, Italian rye, Lolium multiflorum",
  "n12125183": "darnel, tare, bearded darnel, cheat, Lolium temulentum",
  "n12125398": "Muhlenbergia, genus Muhlenbergia",
  "n12125584": "nimblewill, nimble Will, Muhlenbergia schreberi",
  "n12125782": "Oryza, genus Oryza",
  "n12125890": "rice",
  "n12126084": "cultivated rice, Oryza sativa",
  "n12126238": "Oryzopsis, genus Oryzopsis",
  "n12126360": "ricegrass, rice grass",
  "n12126516": "mountain rice, silkgrass, silk grass, Indian millet, Oryzopsis hymenoides",
  "n12126736": "smilo, smilo grass, Oryzopsis miliacea",
  "n12126911": "Panicum, genus Panicum",
  "n12127030": "panic grass",
  "n12127237": "witchgrass, witch grass, old witchgrass, old witch grass, tumble grass, Panicum capillare",
  "n12127460": "switch grass, Panicum virgatum",
  "n12127575": "broomcorn millet, hog millet, Panicum miliaceum",
  "n12127768": "goose grass, Texas millet, Panicum Texanum",
  "n12127890": "genus Paspalum",
  "n12128071": "dallisgrass, dallis grass, paspalum, Paspalum dilatatum",
  "n12128306": "Bahia grass, Paspalum notatum",
  "n12128490": "knotgrass, Paspalum distichum",
  "n12128645": "Pennisetum, genus Pennisetum",
  "n12128825": "pearl millet, bulrush millet, cattail millet, Pennisetum glaucum, Pennisetum Americanum",
  "n12129134": "fountain grass, Pennisetum ruppelii, Pennisetum setaceum",
  "n12129349": "feathertop, feathertop grass, Pennistum villosum",
  "n12129525": "Phalaris, genus Phalaris",
  "n12129738": "reed canary grass, gardener's garters, lady's laces, ribbon grass, Phalaris arundinacea",
  "n12129986": "canary grass, birdseed grass, Phalaris canariensis",
  "n12130160": "hardinggrass, Harding grass, toowomba canary grass, Phalaris aquatica, Phalaris tuberosa",
  "n12130408": "Phleum, genus Phleum",
  "n12130549": "timothy, herd's grass, Phleum pratense",
  "n12130759": "Phragmites, genus Phragmites",
  "n12130937": "ditch reed, common reed, carrizo, Phragmites communis",
  "n12131216": "Poa, genus Poa",
  "n12131405": "bluegrass, blue grass",
  "n12131550": "meadowgrass, meadow grass",
  "n12131767": "Kentucky bluegrass, Kentucky blue, Kentucy blue grass, June grass, Poa pratensis",
  "n12132092": "wood meadowgrass, Poa nemoralis, Agrostis alba",
  "n12132299": "Saccharum, genus Saccharum",
  "n12132502": "sugarcane, sugar cane, Saccharum officinarum",
  "n12132754": "sugarcane, sugar cane",
  "n12132956": "noble cane",
  "n12133151": "munj, munja, Saccharum bengalense, Saccharum munja",
  "n12133332": "Schizachyrium, genus Schizachyrium",
  "n12133462": "broom beard grass, prairie grass, wire grass, Andropogon scoparius, Schizachyrium scoparium",
  "n12133682": "bluestem, blue stem, Andropogon furcatus, Andropogon gerardii",
  "n12133870": "Secale, genus Secale",
  "n12134025": "rye, Secale cereale",
  "n12134300": "Setaria, genus Setaria",
  "n12134486": "bristlegrass, bristle grass",
  "n12134695": "giant foxtail",
  "n12134836": "yellow bristlegrass, yellow bristle grass, yellow foxtail, glaucous bristlegrass, Setaria glauca",
  "n12135049": "green bristlegrass, green foxtail, rough bristlegrass, bottle-grass, bottle grass, Setaria viridis",
  "n12135270": "foxtail millet, Italian millet, Hungarian grass, Setaria italica",
  "n12135576": "Siberian millet, Setaria italica rubrofructa",
  "n12135729": "German millet, golden wonder millet, Setaria italica stramineofructa",
  "n12135898": "millet",
  "n12136206": "cane",
  "n12136392": "rattan, rattan cane",
  "n12136581": "malacca",
  "n12136720": "reed",
  "n12136944": "genus Sorghum, Sorghum",
  "n12137120": "sorghum",
  "n12137337": "great millet, kaffir, kafir corn, kaffir corn, Sorghum bicolor",
  "n12137569": "grain sorghum",
  "n12137791": "durra, doura, dourah, Egyptian corn, Indian millet, Guinea corn",
  "n12137954": "feterita, federita, Sorghum vulgare caudatum",
  "n12138110": "hegari",
  "n12138248": "kaoliang",
  "n12138444": "milo, milo maize",
  "n12138578": "shallu, Sorghum vulgare rosburghii",
  "n12138757": "sorgo, sorgho, sweet sorghum, sugar sorghum",
  "n12138905": "Johnson grass, Aleppo grass, means grass, evergreen millet, Sorghum halepense",
  "n12139196": "broomcorn, Sorghum vulgare technicum",
  "n12139367": "Spartina, genus Spartina",
  "n12139575": "cordgrass, cord grass",
  "n12139793": "salt reed grass, Spartina cynosuroides",
  "n12139921": "prairie cordgrass, freshwater cordgrass, slough grass, Spartina pectinmata",
  "n12140137": "Sporobolus, genus Sporobolus",
  "n12140358": "dropseed, drop-seed",
  "n12140511": "smut grass, blackseed, carpet grass, Sporobolus poiretii",
  "n12140759": "sand dropseed, Sporobolus cryptandrus",
  "n12140903": "rush grass, rush-grass",
  "n12141037": "Stenotaphrum, genus Stenotaphrum",
  "n12141167": "St. Augustine grass, Stenotaphrum secundatum, buffalo grass",
  "n12141385": "grain",
  "n12141495": "cereal, cereal grass",
  "n12141890": "Triticum, genus Triticum",
  "n12142085": "wheat",
  "n12142357": "wheat berry",
  "n12142450": "durum, durum wheat, hard wheat, Triticum durum, Triticum turgidum, macaroni wheat",
  "n12142750": "soft wheat",
  "n12142874": "common wheat, Triticum aestivum",
  "n12143065": "spelt, Triticum spelta, Triticum aestivum spelta",
  "n12143215": "emmer, starch wheat, two-grain spelt, Triticum dicoccum",
  "n12143405": "wild wheat, wild emmer, Triticum dicoccum dicoccoides",
  "n12143572": "Zea, genus Zea",
  "n12143676": "corn, maize, Indian corn, Zea mays",
  "n12144117": "corn",
  "n12144313": "mealie",
  "n12144399": "field corn",
  "n12144580": "corn",
  "n12144742": "sweet corn, sugar corn, green corn, sweet corn plant, Zea mays rugosa, Zea saccharata",
  "n12144987": "dent corn, Zea mays indentata",
  "n12145148": "flint corn, flint maize, Yankee corn, Zea mays indurata",
  "n12145325": "soft corn, flour corn, squaw corn, Zea mays amylacea",
  "n12145477": "popcorn, Zea mays everta",
  "n12145638": "cornsilk, corn silk",
  "n12145802": "Zizania, genus Zizania",
  "n12145919": "wild rice, Zizania aquatica",
  "n12146100": "genus Zoysia, Zoisia, genus Zoisia",
  "n12146311": "zoysia",
  "n12146488": "Manila grass, Japanese carpet grass, Zoysia matrella",
  "n12146654": "Korean lawn grass, Japanese lawn grass, Zoysia japonica",
  "n12146823": "mascarene grass, Korean velvet grass, Zoysia tenuifolia",
  "n12147031": "Bambuseae, tribe Bambuseae",
  "n12147226": "bamboo",
  "n12147539": "bamboo",
  "n12147699": "Bambusa, genus Bambusa",
  "n12147835": "common bamboo, Bambusa vulgaris",
  "n12148079": "Arundinaria, genus Arundinaria",
  "n12148253": "giant cane, cane reed, Arundinaria gigantea",
  "n12148439": "small cane, switch cane, Arundinaria tecta",
  "n12148610": "Dendrocalamus, genus Dendrocalamus",
  "n12148757": "giant bamboo, kyo-chiku, Dendrocalamus giganteus",
  "n12148962": "Phyllostachys, genus Phyllostachys",
  "n12149144": "fishpole bamboo, gosan-chiku, hotei-chiku, Phyllostachys aurea",
  "n12149350": "black bamboo, kuri-chiku, Phyllostachys nigra",
  "n12149521": "giant timber bamboo, madake, ku-chiku, Phyllostachys bambusoides",
  "n12149751": "Cyperaceae, family Cyperaceae, sedge family",
  "n12150028": "sedge",
  "n12150447": "Cyperus, genus Cyperus",
  "n12150722": "umbrella plant, umbrella sedge, Cyperus alternifolius",
  "n12150969": "chufa, yellow nutgrass, earth almond, ground almond, rush nut, Cyperus esculentus",
  "n12151170": "galingale, galangal, Cyperus longus",
  "n12151365": "papyrus, Egyptian paper reed, Egyptian paper rush, paper rush, paper plant, Cyperus papyrus",
  "n12151615": "nutgrass, nut grass, nutsedge, nut sedge, Cyperus rotundus",
  "n12151814": "Carex, genus Carex",
  "n12152031": "sand sedge, sand reed, Carex arenaria",
  "n12152251": "cypress sedge, Carex pseudocyperus",
  "n12152406": "Eriophorum, genus Eriophorum",
  "n12152532": "cotton grass, cotton rush",
  "n12152722": "common cotton grass, Eriophorum angustifolium",
  "n12152869": "Scirpus, genus Scirpus",
  "n12153033": "hardstem bulrush, hardstemmed bulrush, Scirpus acutus",
  "n12153224": "wool grass, Scirpus cyperinus",
  "n12153393": "Eleocharis, genus Eleocharis",
  "n12153580": "spike rush",
  "n12153741": "water chestnut, Chinese water chestnut, Eleocharis dulcis",
  "n12153914": "needle spike rush, needle rush, slender spike rush, hair grass, Eleocharis acicularis",
  "n12154114": "creeping spike rush, Eleocharis palustris",
  "n12154228": "Pandanales, order Pandanales",
  "n12154426": "Pandanaceae, family Pandanaceae, screw-pine family",
  "n12154628": "genus Pandanus",
  "n12154773": "pandanus, screw pine",
  "n12155009": "textile screw pine, lauhala, Pandanus tectorius",
  "n12155126": "pandanus",
  "n12155259": "Typhaceae, family Typhaceae, cattail family",
  "n12155459": "Typha, genus Typha",
  "n12155583": "cattail",
  "n12155773": "cat's-tail, bullrush, bulrush, nailrod, reed mace, reedmace, Typha latifolia",
  "n12156117": "lesser bullrush, narrow-leaf cattail, narrow-leaved reedmace, soft flag, Typha angustifolia",
  "n12156308": "Sparganiaceae, family Sparganiaceae, bur-reed family",
  "n12156484": "Sparganium, genus Sparganium",
  "n12156679": "bur reed",
  "n12156819": "grain, caryopsis",
  "n12157056": "kernel",
  "n12157179": "rye",
  "n12157276": "Cucurbitaceae, family Cucurbitaceae, gourd family",
  "n12157677": "cucurbit",
  "n12157769": "gourd, gourd vine",
  "n12158031": "gourd",
  "n12158148": "Cucurbita, genus Cucurbita",
  "n12158443": "pumpkin, pumpkin vine, autumn pumpkin, Cucurbita pepo",
  "n12158798": "squash, squash vine",
  "n12159055": "summer squash, summer squash vine, Cucurbita pepo melopepo",
  "n12159388": "yellow squash",
  "n12159555": "marrow, marrow squash, vegetable marrow",
  "n12159804": "zucchini, courgette",
  "n12159942": "cocozelle, Italian vegetable marrow",
  "n12160125": "cymling, pattypan squash",
  "n12160303": "spaghetti squash",
  "n12160490": "winter squash, winter squash plant",
  "n12160857": "acorn squash",
  "n12161056": "hubbard squash, Cucurbita maxima",
  "n12161285": "turban squash, Cucurbita maxima turbaniformis",
  "n12161577": "buttercup squash",
  "n12161744": "butternut squash, Cucurbita maxima",
  "n12161969": "winter crookneck, winter crookneck squash, Cucurbita moschata",
  "n12162181": "cushaw, Cucurbita mixta, Cucurbita argyrosperma",
  "n12162425": "prairie gourd, prairie gourd vine, Missouri gourd, wild pumpkin, buffalo gourd, calabazilla, Cucurbita foetidissima",
  "n12162758": "prairie gourd",
  "n12162905": "genus Bryonia",
  "n12163035": "bryony, briony",
  "n12163279": "white bryony, devil's turnip, Bryonia alba",
  "n12163456": "red bryony, wild hop, Bryonia dioica",
  "n12163649": "Citrullus, genus Citrullus",
  "n12163824": "melon, melon vine",
  "n12164065": "watermelon, watermelon vine, Citrullus vulgaris",
  "n12164215": "Cucumis, genus Cucumis",
  "n12164363": "sweet melon, muskmelon, sweet melon vine, Cucumis melo",
  "n12164656": "cantaloupe, cantaloup, cantaloupe vine, cantaloup vine, Cucumis melo cantalupensis",
  "n12164881": "winter melon, Persian melon, honeydew melon, winter melon vine, Cucumis melo inodorus",
  "n12165170": "net melon, netted melon, nutmeg melon, Cucumis melo reticulatus",
  "n12165384": "cucumber, cucumber vine, Cucumis sativus",
  "n12165608": "Ecballium, genus Ecballium",
  "n12165758": "squirting cucumber, exploding cucumber, touch-me-not, Ecballium elaterium",
  "n12166003": "Lagenaria, genus Lagenaria",
  "n12166128": "bottle gourd, calabash, Lagenaria siceraria",
  "n12166312": "genus Luffa",
  "n12166424": "luffa, dishcloth gourd, sponge gourd, rag gourd, strainer vine",
  "n12166793": "loofah, vegetable sponge, Luffa cylindrica",
  "n12166929": "angled loofah, sing-kwa, Luffa acutangula",
  "n12167075": "loofa, loofah, luffa, loufah sponge",
  "n12167282": "Momordica, genus Momordica",
  "n12167436": "balsam apple, Momordica balsamina",
  "n12167602": "balsam pear, Momordica charantia",
  "n12167749": "Goodeniaceae, family Goodeniaceae, Goodenia family",
  "n12167955": "Goodenia",
  "n12168126": "Lobeliaceae, family Lobeliaceae, lobelia family",
  "n12168385": "genus Lobelia",
  "n12168565": "lobelia",
  "n12168750": "cardinal flower, Indian pink, Lobelia cardinalis",
  "n12168898": "Indian tobacco, bladderpod, Lobelia inflata",
  "n12169099": "water lobelia, Lobelia dortmanna",
  "n12169320": "great lobelia, blue cardinal flower, Lobelia siphilitica",
  "n12169526": "Malvales, order Malvales",
  "n12169776": "Malvaceae, family Malvaceae, mallow family",
  "n12170415": "Malva, genus Malva",
  "n12170585": "mallow",
  "n12171098": "musk mallow, mus rose, Malva moschata",
  "n12171316": "common mallow, Malva neglecta",
  "n12171503": "tall mallow, high mallow, cheese, cheeseflower, Malva sylvestris",
  "n12171750": "Abelmoschus, genus Abelmoschus",
  "n12171966": "okra, gumbo, okra plant, lady's-finger, Abelmoschus esculentus, Hibiscus esculentus",
  "n12172364": "okra",
  "n12172481": "abelmosk, musk mallow, Abelmoschus moschatus, Hibiscus moschatus",
  "n12172715": "Abutilon, genus Abutilon",
  "n12172906": "flowering maple",
  "n12173069": "velvetleaf, velvet-leaf, velvetweed, Indian mallow, butter-print, China jute, Abutilon theophrasti",
  "n12173407": "Alcea, genus Alcea",
  "n12173664": "hollyhock",
  "n12173912": "rose mallow, Alcea rosea, Althea rosea",
  "n12174124": "genus Althaea",
  "n12174311": "althea, althaea, hollyhock",
  "n12174521": "marsh mallow, white mallow, Althea officinalis",
  "n12174742": "Callirhoe, genus Callirhoe",
  "n12174926": "poppy mallow",
  "n12175181": "fringed poppy mallow, Callirhoe digitata",
  "n12175370": "purple poppy mallow, Callirhoe involucrata",
  "n12175598": "clustered poppy mallow, Callirhoe triangulata",
  "n12175797": "Gossypium, genus Gossypium",
  "n12175949": "cotton, cotton plant",
  "n12176278": "tree cotton, Gossypium arboreum",
  "n12176453": "sea island cotton, tree cotton, Gossypium barbadense",
  "n12176709": "Levant cotton, Gossypium herbaceum",
  "n12176953": "upland cotton, Gossypium hirsutum",
  "n12177129": "Peruvian cotton, Gossypium peruvianum",
  "n12177249": "Egyptian cotton",
  "n12177455": "wild cotton, Arizona wild cotton, Gossypium thurberi",
  "n12177592": "genus Hibiscus",
  "n12177844": "hibiscus",
  "n12178129": "kenaf, kanaf, deccan hemp, bimli, bimli hemp, Indian hemp, Bombay hemp, Hibiscus cannabinus",
  "n12178358": "kenaf, deccan hemp",
  "n12178494": "Cuban bast, blue mahoe, mahoe, majagua, mahagua, Hibiscus elatus",
  "n12178780": "sorrel tree, Hibiscus heterophyllus",
  "n12178896": "rose mallow, swamp mallow, common rose mallow, swamp rose mallow, Hibiscus moscheutos",
  "n12179122": "cotton rose, Confederate rose, Confederate rose mallow, Hibiscus mutabilis",
  "n12179391": "China rose, Chinese hibiscus, Rose of China, shoeblack plant, shoe black, Hibiscus rosa-sinensis",
  "n12179632": "roselle, rozelle, sorrel, red sorrel, Jamaica sorrel, Hibiscus sabdariffa",
  "n12179907": "rose of Sharon, Hibiscus syriacus",
  "n12180168": "mahoe, majagua, mahagua, balibago, purau, Hibiscus tiliaceus",
  "n12180456": "flower-of-an-hour, flowers-of-an-hour, bladder ketmia, black-eyed Susan, Hibiscus trionum",
  "n12180714": "Hoheria, genus Hoheria",
  "n12180885": "lacebark, ribbonwood, houhere, Hoheria populnea",
  "n12181147": "Iliamna, genus Iliamna",
  "n12181352": "wild hollyhock, Iliamna remota, Sphaeralcea remota",
  "n12181612": "mountain hollyhock, Iliamna ruvularis, Iliamna acerifolia",
  "n12181851": "Kosteletzya, genus Kosteletzya",
  "n12182049": "seashore mallow",
  "n12182276": "salt marsh mallow, Kosteletzya virginica",
  "n12182414": "Lavatera, genus Lavatera",
  "n12182615": "tree mallow, velvetleaf, velvet-leaf, Lavatera arborea",
  "n12182858": "Malacothamnus, genus Malacothamnus",
  "n12183026": "chaparral mallow, Malacothamnus fasciculatus, Sphaeralcea fasciculata",
  "n12183318": "genus Malope",
  "n12183452": "malope, Malope trifida",
  "n12183636": "Malvastrum, genus Malvastrum",
  "n12183816": "false mallow",
  "n12183916": "Malvaviscus, genus Malvaviscus",
  "n12184095": "waxmallow, wax mallow, sleeping hibiscus",
  "n12184337": "Napaea, genus Napaea",
  "n12184468": "glade mallow, Napaea dioica",
  "n12184724": "genus Pavonia",
  "n12184912": "pavonia",
  "n12185078": "Plagianthus, genus Plagianthus",
  "n12185254": "ribbon tree, ribbonwood, Plagianthus regius, Plagianthus betulinus",
  "n12185526": "New Zealand cotton",
  "n12185687": "Radyera, genus Radyera",
  "n12185859": "bush hibiscus, Radyera farragei, Hibiscus farragei",
  "n12186116": "Sida, genus Sida",
  "n12186352": "Virginia mallow, Sida hermaphrodita",
  "n12186554": "Queensland hemp, jellyleaf, Sida rhombifolia",
  "n12186839": "Indian mallow, Sida spinosa",
  "n12187030": "Sidalcea, genus Sidalcea",
  "n12187247": "checkerbloom, wild hollyhock, Sidalcea malviflora",
  "n12187450": "Sphaeralcea, genus Sphaeralcea",
  "n12187663": "globe mallow, false mallow",
  "n12187891": "prairie mallow, red false mallow, Sphaeralcea coccinea, Malvastrum coccineum",
  "n12188120": "Thespesia, genus Thespesia",
  "n12188289": "tulipwood tree",
  "n12188484": "tulipwood",
  "n12188635": "portia tree, bendy tree, seaside mahoe, Thespesia populnea",
  "n12188985": "Bombacaceae, family Bombacaceae",
  "n12189293": "Bombax, genus Bombax",
  "n12189429": "red silk-cotton tree, simal, Bombax ceiba, Bombax malabarica",
  "n12189620": "Adansonia, genus Adansonia",
  "n12189779": "cream-of-tartar tree, sour gourd, Adansonia gregorii",
  "n12189987": "baobab, monkey-bread tree, Adansonia digitata",
  "n12190241": "Ceiba, genus Ceiba",
  "n12190410": "kapok, ceiba tree, silk-cotton tree, white silk-cotton tree, Bombay ceiba, God tree, Ceiba pentandra",
  "n12190712": "Durio, genus Durio",
  "n12190869": "durian, durion, durian tree, Durio zibethinus",
  "n12191075": "genus Montezuma",
  "n12191240": "Montezuma",
  "n12191461": "Ochroma, genus Ochroma",
  "n12191587": "balsa, Ochroma lagopus",
  "n12191813": "balsa, balsa wood",
  "n12191965": "Pseudobombax, genus Pseudobombax",
  "n12192132": "shaving-brush tree, Pseudobombax ellipticum",
  "n12192373": "Elaeocarpaceae, family Elaeocarpaceae, elaeocarpus family",
  "n12192722": "Elaeocarpus, genus Elaeocarpus",
  "n12192877": "quandong, quandong tree, Brisbane quandong, silver quandong tree, blue fig, Elaeocarpus grandis",
  "n12193205": "silver quandong",
  "n12193334": "quandong, blue fig",
  "n12193458": "Aristotelia, genus Aristotelia",
  "n12193665": "makomako, New Zealand wine berry, wineberry, Aristotelia serrata, Aristotelia racemosa",
  "n12193964": "Muntingia, genus Muntingia",
  "n12194147": "Jamaican cherry, calabur tree, calabura, silk wood, silkwood, Muntingia calabura",
  "n12194466": "Sloanea, genus Sloanea",
  "n12194613": "breakax, breakaxe, break-axe, Sloanea jamaicensis",
  "n12194776": "Sterculiaceae, family Sterculiaceae, sterculia family",
  "n12195186": "genus Sterculia",
  "n12195391": "sterculia",
  "n12195533": "Panama tree, Sterculia apetala",
  "n12195734": "kalumpang, Java olives, Sterculia foetida",
  "n12195965": "Brachychiton, genus Brachychiton",
  "n12196129": "bottle-tree, bottle tree",
  "n12196336": "flame tree, flame durrajong, Brachychiton acerifolius, Sterculia acerifolia",
  "n12196527": "flame tree, broad-leaved bottletree, Brachychiton australis",
  "n12196694": "kurrajong, currajong, Brachychiton populneus",
  "n12196954": "Queensland bottletree, narrow-leaved bottletree, Brachychiton rupestris, Sterculia rupestris",
  "n12197211": "Cola, genus Cola",
  "n12197359": "kola, kola nut, kola nut tree, goora nut, Cola acuminata",
  "n12197601": "kola nut, cola nut",
  "n12197765": "genus Dombeya",
  "n12197901": "dombeya",
  "n12198140": "Firmiana, genus Firmiana",
  "n12198286": "Chinese parasol tree, Chinese parasol, Japanese varnish tree, phoenix tree, Firmiana simplex",
  "n12198628": "Fremontodendron, genus Fremontodendron, Fremontia, genus Fremontia",
  "n12198793": "flannelbush, flannel bush, California beauty",
  "n12199030": "Helicteres, genus Helicteres",
  "n12199266": "screw tree",
  "n12199399": "nut-leaved screw tree, Helicteres isora",
  "n12199564": "Heritiera, genus Heritiera, Terrietia, genus Terrietia",
  "n12199790": "red beech, brown oak, booyong, crow's foot, stave wood, silky elm, Heritiera trifoliolata, Terrietia trifoliolata",
  "n12199982": "looking glass tree, Heritiera macrophylla",
  "n12200143": "looking-glass plant, Heritiera littoralis",
  "n12200315": "Hermannia, genus Hermannia",
  "n12200504": "honey bell, honeybells, Hermannia verticillata, Mahernia verticillata",
  "n12200747": "Pterospermum, genus Pterospermum",
  "n12200905": "mayeng, maple-leaved bayur, Pterospermum acerifolium",
  "n12201166": "Tarrietia, genus Tarrietia",
  "n12201331": "silver tree, Tarrietia argyrodendron",
  "n12201456": "Theobroma, genus Theobroma",
  "n12201580": "cacao, cacao tree, chocolate tree, Theobroma cacao",
  "n12201761": "Triplochiton, genus Triplochiton",
  "n12201938": "obeche, obechi, arere, samba, Triplochiton scleroxcylon",
  "n12202234": "obeche",
  "n12202352": "Tiliaceae, family Tiliaceae, linden family",
  "n12202712": "Tilia, genus Tilia",
  "n12202936": "linden, linden tree, basswood, lime, lime tree",
  "n12203331": "basswood, linden",
  "n12203529": "American basswood, American lime, Tilia americana",
  "n12203699": "small-leaved linden, small-leaved lime, Tilia cordata",
  "n12203896": "white basswood, cottonwood, Tilia heterophylla",
  "n12204032": "Japanese linden, Japanese lime, Tilia japonica",
  "n12204175": "silver lime, silver linden, Tilia tomentosa",
  "n12204405": "Entelea, genus Entelea",
  "n12204546": "Corchorus, genus Corchorus",
  "n12204730": "corchorus",
  "n12204925": "Grewia, genus Grewia",
  "n12205104": "phalsa, Grewia asiatica",
  "n12205308": "Sparmannia, genus Sparmannia",
  "n12205460": "African hemp, Sparmannia africana",
  "n12205694": "herb, herbaceous plant",
  "n12212361": "vegetable",
  "n12212690": "simple",
  "n12212810": "Rosidae, subclass Rosidae",
  "n12213197": "Umbellales, order Umbellales",
  "n12213485": "Proteales, order Proteales",
  "n12213635": "Proteaceae, family Proteaceae, protea family",
  "n12214245": "Bartle Frere, genus Bartle-Frere, green dinosaur",
  "n12214605": "genus Protea",
  "n12214789": "protea",
  "n12215022": "honeypot, king protea, Protea cynaroides",
  "n12215210": "honeyflower, honey-flower, Protea mellifera",
  "n12215373": "genus Banksia",
  "n12215579": "banksia",
  "n12215824": "honeysuckle, Australian honeysuckle, coast banksia, Banksia integrifolia",
  "n12216028": "Conospermum, genus Conospermum",
  "n12216215": "smoke bush",
  "n12216382": "Embothrium, genus Embothrium",
  "n12216628": "Chilean firebush, Chilean flameflower, Embothrium coccineum",
  "n12216836": "Guevina, genus Guevina",
  "n12216968": "Chilean nut, Chile nut, Chile hazel, Chilean hazelnut, Guevina heterophylla, Guevina avellana",
  "n12217211": "genus Grevillea",
  "n12217453": "grevillea",
  "n12217586": "silk oak",
  "n12217851": "red-flowered silky oak, Grevillea banksii",
  "n12218054": "silver oak, Grevillela parallela",
  "n12218274": "silky oak, Grevillea robusta",
  "n12218490": "beefwood, Grevillea striata",
  "n12218621": "Hakea, genus Hakea",
  "n12218868": "cushion flower, pincushion hakea, Hakea laurina",
  "n12219065": "needlewood, needle-wood, needle wood, Hakea leucoptera",
  "n12219289": "needlebush, needle-bush, needle bush, Hakea lissosperma",
  "n12219495": "Knightia, genus Knightia",
  "n12219668": "rewa-rewa, New Zealand honeysuckle",
  "n12219875": "Lambertia, genus Lambertia",
  "n12220019": "honeyflower, honey-flower, mountain devil, Lambertia formosa",
  "n12220247": "Leucadendron, genus Leucadendron",
  "n12220496": "silver tree, Leucadendron argenteum",
  "n12220654": "genus Lomatia",
  "n12220829": "lomatia",
  "n12220994": "genus Macadamia",
  "n12221191": "macadamia, macadamia tree",
  "n12221368": "Macadamia integrifolia",
  "n12221522": "macadamia nut, macadamia nut tree, Macadamia ternifolia",
  "n12221801": "Queensland nut, Macadamia tetraphylla",
  "n12221943": "Orites, genus Orites",
  "n12222090": "prickly ash, Orites excelsa",
  "n12222334": "Persoonia, genus Persoonia",
  "n12222493": "geebung",
  "n12222715": "Stenocarpus, genus Stenocarpus",
  "n12222900": "wheel tree, firewheel tree, Stenocarpus sinuatus",
  "n12223160": "scrub beefwood, beefwood, Stenocarpus salignus",
  "n12223405": "Telopea, genus Telopea",
  "n12223569": "waratah, Telopea Oreades",
  "n12223764": "waratah, Telopea speciosissima",
  "n12223950": "Xylomelum, genus Xylomelum",
  "n12224140": "native pear, woody pear, Xylomelum pyriforme",
  "n12224309": "Casuarinales, order Casuarinales",
  "n12224522": "Casuarinaceae, family Casuarinaceae",
  "n12224669": "genus Casuarina",
  "n12224978": "casuarina",
  "n12225222": "she-oak",
  "n12225349": "beefwood",
  "n12225563": "Australian pine, Casuarina equisetfolia",
  "n12225769": "beefwood",
  "n12226009": "Ericales, order Ericales",
  "n12226322": "Ericaceae, family Ericaceae, heath family",
  "n12226932": "heath",
  "n12227220": "genus Erica",
  "n12227420": "erica, true heath",
  "n12227658": "tree heath, briar, brier, Erica arborea",
  "n12227909": "briarroot",
  "n12228045": "briarwood, brierwood, brier-wood",
  "n12228229": "winter heath, spring heath, Erica carnea",
  "n12228387": "bell heather, heather bell, fine-leaved heath, Erica cinerea",
  "n12228546": "cross-leaved heath, bell heather, Erica tetralix",
  "n12228689": "Cornish heath, Erica vagans",
  "n12228886": "Spanish heath, Portuguese heath, Erica lusitanica",
  "n12229111": "Prince-of-Wales'-heath, Prince of Wales heath, Erica perspicua",
  "n12229282": "genus Andromeda",
  "n12229443": "andromeda",
  "n12229651": "bog rosemary, moorwort, Andromeda glaucophylla",
  "n12229887": "marsh andromeda, common bog rosemary, Andromeda polifolia",
  "n12230146": "genus Arbutus",
  "n12230347": "arbutus",
  "n12230540": "madrona, madrono, manzanita, Arbutus menziesii",
  "n12230794": "strawberry tree, Irish strawberry, Arbutus unedo",
  "n12231031": "Arctostaphylos, genus Arctostaphylos",
  "n12231192": "bearberry",
  "n12231358": "common bearberry, red bearberry, wild cranberry, mealberry, hog cranberry, sand berry, sandberry, mountain box, bear's grape, creashak, Arctostaphylos uva-ursi",
  "n12231709": "alpine bearberry, black bearberry, Arctostaphylos alpina",
  "n12231918": "manzanita",
  "n12232114": "heartleaf manzanita, Arctostaphylos andersonii",
  "n12232280": "Parry manzanita, Arctostaphylos manzanita",
  "n12232503": "downy manzanita, woolly manzanita, Arctostaphylos tomentosa",
  "n12232683": "Bruckenthalia, genus Bruckenthalia",
  "n12232851": "spike heath, Bruckenthalia spiculifolia",
  "n12233094": "genus Bryanthus",
  "n12233249": "bryanthus",
  "n12233410": "Calluna, genus Calluna",
  "n12233529": "heather, ling, Scots heather, broom, Calluna vulgaris",
  "n12233759": "Cassiope, genus Cassiope",
  "n12233998": "white heather, Cassiope mertensiana",
  "n12234176": "Chamaedaphne, genus Chamaedaphne",
  "n12234318": "leatherleaf, Chamaedaphne calyculata",
  "n12234513": "Daboecia, genus Daboecia",
  "n12234669": "Connemara heath, St. Dabeoc's heath, Daboecia cantabrica",
  "n12234913": "Epigaea, genus Epigaea",
  "n12235051": "trailing arbutus, mayflower, Epigaea repens",
  "n12235263": "Gaultheria, genus Gaultheria",
  "n12235479": "creeping snowberry, moxie plum, maidenhair berry, Gaultheria hispidula",
  "n12235765": "teaberry, wintergreen, checkerberry, mountain tea, groundberry, ground-berry, creeping wintergreen, Gaultheria procumbens",
  "n12236160": "salal, shallon, Gaultheria shallon",
  "n12236363": "Gaylussacia, genus Gaylussacia",
  "n12236546": "huckleberry",
  "n12236768": "black huckleberry, Gaylussacia baccata",
  "n12236977": "dangleberry, dangle-berry, Gaylussacia frondosa",
  "n12237152": "box huckleberry, Gaylussacia brachycera",
  "n12237350": "genus Kalmia",
  "n12237486": "kalmia",
  "n12237641": "mountain laurel, wood laurel, American laurel, calico bush, Kalmia latifolia",
  "n12237855": "swamp laurel, bog laurel, bog kalmia, Kalmia polifolia",
  "n12238073": "sheep laurel, pig laurel, lambkill, Kalmia angustifolia",
  "n12238306": "Ledum, genus Ledum",
  "n12238491": "Labrador tea, crystal tea, Ledum groenlandicum",
  "n12238756": "trapper's tea, glandular Labrador tea",
  "n12238913": "wild rosemary, marsh tea, Ledum palustre",
  "n12239100": "Leiophyllum, genus Leiophyllum",
  "n12239240": "sand myrtle, Leiophyllum buxifolium",
  "n12239458": "genus Leucothoe",
  "n12239647": "leucothoe",
  "n12239880": "dog laurel, dog hobble, switch-ivy, Leucothoe fontanesiana, Leucothoe editorum",
  "n12240150": "sweet bells, Leucothoe racemosa",
  "n12240335": "Loiseleuria, genus Loiseleuria",
  "n12240477": "alpine azalea, mountain azalea, Loiseleuria procumbens",
  "n12240715": "Lyonia, genus Lyonia",
  "n12240965": "staggerbush, stagger bush, Lyonia mariana",
  "n12241192": "maleberry, male berry, privet andromeda, he-huckleberry, Lyonia ligustrina",
  "n12241426": "fetterbush, fetter bush, shiny lyonia, Lyonia lucida",
  "n12241699": "Menziesia, genus Menziesia",
  "n12241880": "false azalea, fool's huckleberry, Menziesia ferruginea",
  "n12242123": "minniebush, minnie bush, Menziesia pilosa",
  "n12242287": "Oxydendrum, genus Oxydendrum",
  "n12242409": "sorrel tree, sourwood, titi, Oxydendrum arboreum",
  "n12242668": "Phyllodoce, genus Phyllodoce",
  "n12242850": "mountain heath, Phyllodoce caerulea, Bryanthus taxifolius",
  "n12243109": "purple heather, Brewer's mountain heather, Phyllodoce breweri",
  "n12243292": "Pieris, genus Pieris",
  "n12243459": "andromeda, Japanese andromeda, lily-of-the-valley tree, Pieris japonica",
  "n12243693": "fetterbush, mountain fetterbush, mountain andromeda, Pieris floribunda",
  "n12243927": "genus Rhododendron",
  "n12244153": "rhododendron",
  "n12244458": "coast rhododendron, Rhododendron californicum",
  "n12244650": "rosebay, Rhododendron maxima",
  "n12244819": "swamp azalea, swamp honeysuckle, white honeysuckle, Rhododendron viscosum",
  "n12245067": "subgenus Azalea, Azaleastrum, subgenus Azaleastrum",
  "n12245319": "azalea",
  "n12245472": "Vaccinium, genus Vaccinium",
  "n12245695": "cranberry",
  "n12245885": "American cranberry, large cranberry, Vaccinium macrocarpon",
  "n12246037": "European cranberry, small cranberry, Vaccinium oxycoccus",
  "n12246232": "blueberry, blueberry bush",
  "n12246641": "huckleberry",
  "n12246773": "farkleberry, sparkleberry, Vaccinium arboreum",
  "n12246941": "low-bush blueberry, low blueberry, Vaccinium angustifolium, Vaccinium pennsylvanicum",
  "n12247202": "rabbiteye blueberry, rabbit-eye blueberry, rabbiteye, Vaccinium ashei",
  "n12247407": "dwarf bilberry, dwarf blueberry, Vaccinium caespitosum",
  "n12247664": "high-bush blueberry, tall bilberry, swamp blueberry, Vaccinium corymbosum",
  "n12247963": "evergreen blueberry, Vaccinium myrsinites",
  "n12248141": "evergreen huckleberry, Vaccinium ovatum",
  "n12248359": "bilberry, thin-leaved bilberry, mountain blue berry, Viccinium membranaceum",
  "n12248574": "bilberry, whortleberry, whinberry, blaeberry, Viccinium myrtillus",
  "n12248780": "bog bilberry, bog whortleberry, moor berry, Vaccinium uliginosum alpinum",
  "n12248941": "dryland blueberry, dryland berry, Vaccinium pallidum",
  "n12249122": "grouseberry, grouse-berry, grouse whortleberry, Vaccinium scoparium",
  "n12249294": "deerberry, squaw huckleberry, Vaccinium stamineum",
  "n12249542": "cowberry, mountain cranberry, lingonberry, lingenberry, lingberry, foxberry, Vaccinium vitis-idaea",
  "n12249821": "Clethraceae, family Clethraceae, white-alder family",
  "n12249993": "Clethra, genus Clethra",
  "n12250180": "sweet pepperbush, pepper bush, summer sweet, white alder, Clethra alnifolia",
  "n12250413": "Diapensiaceae, family Diapensiaceae, diapensia family",
  "n12250708": "Diapensiales, order Diapensiales",
  "n12250874": "genus Diapensia",
  "n12251001": "diapensia",
  "n12251137": "genus Galax",
  "n12251278": "galax, galaxy, wandflower, beetleweed, coltsfoot, Galax urceolata",
  "n12251577": "Pyxidanthera, genus Pyxidanthera",
  "n12251740": "pyxie, pixie, pixy, Pyxidanthera barbulata",
  "n12251997": "genus Shortia",
  "n12252168": "shortia",
  "n12252383": "oconee bells, Shortia galacifolia",
  "n12252620": "Epacridaceae, family Epacridaceae, epacris family",
  "n12252866": "Australian heath",
  "n12253083": "genus Epacris",
  "n12253229": "epacris",
  "n12253487": "common heath, Epacris impressa",
  "n12253664": "common heath, blunt-leaf heath, Epacris obtusifolia",
  "n12253835": "Port Jackson heath, Epacris purpurascens",
  "n12254014": "Astroloma, genus Astroloma",
  "n12254168": "native cranberry, groundberry, ground-berry, cranberry heath, Astroloma humifusum, Styphelia humifusum",
  "n12254478": "Richea, genus Richea",
  "n12254667": "Australian grass tree, Richea dracophylla",
  "n12254891": "tree heath, grass tree, Richea pandanifolia",
  "n12255086": "Styphelia, genus Styphelia",
  "n12255225": "pink fivecorner, Styphelia triflora",
  "n12255452": "Lennoaceae, family Lennoaceae",
  "n12255659": "Pyrolaceae, family Pyrolaceae, wintergreen family",
  "n12255934": "genus Pyrola",
  "n12256112": "wintergreen, pyrola",
  "n12256325": "false wintergreen, Pyrola americana, Pyrola rotundifolia americana",
  "n12256522": "lesser wintergreen, Pyrola minor",
  "n12256708": "wild lily of the valley, shinleaf, Pyrola elliptica",
  "n12256920": "wild lily of the valley, Pyrola rotundifolia",
  "n12257140": "Orthilia, genus Orthilia",
  "n12257343": "Chimaphila, genus Chimaphila",
  "n12257570": "pipsissewa, prince's pine",
  "n12257725": "love-in-winter, western prince's pine, Chimaphila umbellata, Chimaphila corymbosa",
  "n12257920": "Moneses, genus Moneses",
  "n12258101": "one-flowered wintergreen, one-flowered pyrola, Moneses uniflora, Pyrola uniflora",
  "n12258380": "Monotropaceae, family Monotropaceae",
  "n12258663": "Monotropa, genus Monotropa",
  "n12258885": "Indian pipe, waxflower, Monotropa uniflora",
  "n12259122": "Hypopitys, genus Hypopitys",
  "n12259316": "pinesap, false beachdrops, Monotropa hypopithys",
  "n12259615": "Sarcodes, genus Sarcodes",
  "n12259788": "snow plant, Sarcodes sanguinea",
  "n12260021": "Fagales, order Fagales",
  "n12260208": "Fagaceae, family Fagaceae, beech family",
  "n12260593": "Fagus, genus Fagus",
  "n12260799": "beech, beech tree",
  "n12261179": "beech, beechwood",
  "n12261359": "common beech, European beech, Fagus sylvatica",
  "n12261571": "copper beech, purple beech, Fagus sylvatica atropunicea, Fagus purpurea, Fagus sylvatica purpurea",
  "n12261808": "American beech, white beech, red beech, Fagus grandifolia, Fagus americana",
  "n12262018": "weeping beech, Fagus pendula, Fagus sylvatica pendula",
  "n12262185": "Japanese beech",
  "n12262327": "Castanea, genus Castanea",
  "n12262553": "chestnut, chestnut tree",
  "n12262905": "chestnut",
  "n12263038": "American chestnut, American sweet chestnut, Castanea dentata",
  "n12263204": "European chestnut, sweet chestnut, Spanish chestnut, Castanea sativa",
  "n12263410": "Chinese chestnut, Castanea mollissima",
  "n12263588": "Japanese chestnut, Castanea crenata",
  "n12263738": "Allegheny chinkapin, eastern chinquapin, chinquapin, dwarf chestnut, Castanea pumila",
  "n12263987": "Ozark chinkapin, Ozark chinquapin, chinquapin, Castanea ozarkensis",
  "n12264254": "Castanopsis, genus Castanopsis",
  "n12264512": "oak chestnut",
  "n12264621": "Chrysolepis, genus Chrysolepis",
  "n12264786": "giant chinkapin, golden chinkapin, Chrysolepis chrysophylla, Castanea chrysophylla, Castanopsis chrysophylla",
  "n12265083": "dwarf golden chinkapin, Chrysolepis sempervirens",
  "n12265266": "Lithocarpus, genus Lithocarpus",
  "n12265394": "tanbark oak, Lithocarpus densiflorus",
  "n12265600": "Japanese oak, Lithocarpus glabra, Lithocarpus glaber",
  "n12265740": "tanbark",
  "n12265900": "Nothofagus, genus Nothofagus",
  "n12266217": "southern beech, evergreen beech",
  "n12266528": "myrtle beech, Nothofagus cuninghamii",
  "n12266644": "Coigue, Nothofagus dombeyi",
  "n12266796": "New Zealand beech",
  "n12266984": "silver beech, Nothofagus menziesii",
  "n12267133": "roble beech, Nothofagus obliqua",
  "n12267265": "rauli beech, Nothofagus procera",
  "n12267411": "black beech, Nothofagus solanderi",
  "n12267534": "hard beech, Nothofagus truncata",
  "n12267677": "acorn",
  "n12267841": "cup",
  "n12267931": "cupule, acorn cup",
  "n12268096": "Quercus, genus Quercus",
  "n12268246": "oak, oak tree",
  "n12268918": "oak",
  "n12269099": "fumed oak",
  "n12269241": "live oak",
  "n12269406": "coast live oak, California live oak, Quercus agrifolia",
  "n12269652": "white oak",
  "n12270027": "American white oak, Quercus alba",
  "n12270278": "Arizona white oak, Quercus arizonica",
  "n12270460": "swamp white oak, swamp oak, Quercus bicolor",
  "n12270741": "European turkey oak, turkey oak, Quercus cerris",
  "n12270946": "canyon oak, canyon live oak, maul oak, iron oak, Quercus chrysolepis",
  "n12271187": "scarlet oak, Quercus coccinea",
  "n12271451": "jack oak, northern pin oak, Quercus ellipsoidalis",
  "n12271643": "red oak",
  "n12271933": "southern red oak, swamp red oak, turkey oak, Quercus falcata",
  "n12272239": "Oregon white oak, Oregon oak, Garry oak, Quercus garryana",
  "n12272432": "holm oak, holm tree, holly-leaved oak, evergreen oak, Quercus ilex",
  "n12272650": "holm oak",
  "n12272735": "bear oak, Quercus ilicifolia",
  "n12272883": "shingle oak, laurel oak, Quercus imbricaria",
  "n12273114": "bluejack oak, turkey oak, Quercus incana",
  "n12273344": "California black oak, Quercus kelloggii",
  "n12273515": "American turkey oak, turkey oak, Quercus laevis",
  "n12273768": "laurel oak, pin oak, Quercus laurifolia",
  "n12273939": "California white oak, valley oak, valley white oak, roble, Quercus lobata",
  "n12274151": "overcup oak, Quercus lyrata",
  "n12274358": "bur oak, burr oak, mossy-cup oak, mossycup oak, Quercus macrocarpa",
  "n12274630": "scrub oak",
  "n12274863": "blackjack oak, blackjack, jack oak, Quercus marilandica",
  "n12275131": "swamp chestnut oak, Quercus michauxii",
  "n12275317": "Japanese oak, Quercus mongolica, Quercus grosseserrata",
  "n12275489": "chestnut oak",
  "n12275675": "chinquapin oak, chinkapin oak, yellow chestnut oak, Quercus muehlenbergii",
  "n12275888": "myrtle oak, seaside scrub oak, Quercus myrtifolia",
  "n12276110": "water oak, possum oak, Quercus nigra",
  "n12276314": "Nuttall oak, Nuttall's oak, Quercus nuttalli",
  "n12276477": "durmast, Quercus petraea, Quercus sessiliflora",
  "n12276628": "basket oak, cow oak, Quercus prinus, Quercus montana",
  "n12276872": "pin oak, swamp oak, Quercus palustris",
  "n12277150": "willow oak, Quercus phellos",
  "n12277334": "dwarf chinkapin oak, dwarf chinquapin oak, dwarf oak, Quercus prinoides",
  "n12277578": "common oak, English oak, pedunculate oak, Quercus robur",
  "n12277800": "northern red oak, Quercus rubra, Quercus borealis",
  "n12278107": "Shumard oak, Shumard red oak, Quercus shumardii",
  "n12278371": "post oak, box white oak, brash oak, iron oak, Quercus stellata",
  "n12278650": "cork oak, Quercus suber",
  "n12278865": "Spanish oak, Quercus texana",
  "n12279060": "huckleberry oak, Quercus vaccinifolia",
  "n12279293": "Chinese cork oak, Quercus variabilis",
  "n12279458": "black oak, yellow oak, quercitron, quercitron oak, Quercus velutina",
  "n12279772": "southern live oak, Quercus virginiana",
  "n12280060": "interior live oak, Quercus wislizenii, Quercus wizlizenii",
  "n12280364": "mast",
  "n12280487": "Betulaceae, family Betulaceae, birch family",
  "n12280886": "Betula, genus Betula",
  "n12281241": "birch, birch tree",
  "n12281600": "birch",
  "n12281788": "yellow birch, Betula alleghaniensis, Betula leutea",
  "n12281974": "American white birch, paper birch, paperbark birch, canoe birch, Betula cordifolia, Betula papyrifera",
  "n12282235": "grey birch, gray birch, American grey birch, American gray birch, Betula populifolia",
  "n12282527": "silver birch, common birch, European white birch, Betula pendula",
  "n12282737": "downy birch, white birch, Betula pubescens",
  "n12282933": "black birch, river birch, red birch, Betula nigra",
  "n12283147": "sweet birch, cherry birch, black birch, Betula lenta",
  "n12283395": "Yukon white birch, Betula neoalaskana",
  "n12283542": "swamp birch, water birch, mountain birch, Western paper birch, Western birch, Betula fontinalis",
  "n12283790": "Newfoundland dwarf birch, American dwarf birch, Betula glandulosa",
  "n12283981": "Alnus, genus Alnus",
  "n12284262": "alder, alder tree",
  "n12284665": "alder",
  "n12284821": "common alder, European black alder, Alnus glutinosa, Alnus vulgaris",
  "n12285049": "grey alder, gray alder, Alnus incana",
  "n12285195": "seaside alder, Alnus maritima",
  "n12285369": "white alder, mountain alder, Alnus rhombifolia",
  "n12285512": "red alder, Oregon alder, Alnus rubra",
  "n12285705": "speckled alder, Alnus rugosa",
  "n12285900": "smooth alder, hazel alder, Alnus serrulata",
  "n12286068": "green alder, Alnus veridis",
  "n12286197": "green alder, Alnus veridis crispa, Alnus crispa",
  "n12286372": "Carpinaceae, subfamily Carpinaceae, family Carpinaceae",
  "n12286581": "Carpinus, genus Carpinus",
  "n12286826": "hornbeam",
  "n12286988": "European hornbeam, Carpinus betulus",
  "n12287195": "American hornbeam, Carpinus caroliniana",
  "n12287388": "Ostrya, genus Ostrya",
  "n12287642": "hop hornbeam",
  "n12287836": "Old World hop hornbeam, Ostrya carpinifolia",
  "n12288005": "Eastern hop hornbeam, ironwood, ironwood tree, Ostrya virginiana",
  "n12288188": "Ostryopsis, genus Ostryopsis",
  "n12288422": "Corylaceae, subfamily Corylaceae, family Corylaceae",
  "n12288598": "Corylus, genus Corylus",
  "n12288823": "hazelnut, hazel, hazelnut tree",
  "n12289115": "hazel",
  "n12289310": "American hazel, Corylus americana",
  "n12289433": "cobnut, filbert, Corylus avellana, Corylus avellana grandis",
  "n12289585": "beaked hazelnut, Corylus cornuta",
  "n12289744": "Gentianales, order Gentianales",
  "n12290116": "Gentianaceae, family Gentianaceae, gentian family",
  "n12290522": "Centaurium, genus Centaurium",
  "n12290748": "centaury",
  "n12290975": "rosita, Centaurium calycosum",
  "n12291143": "lesser centaury, Centaurium minus",
  "n12291292": "tufted centaury, Centaurium scilloides",
  "n12291459": "seaside centaury",
  "n12291564": "broad leaved centaury",
  "n12291671": "slender centaury",
  "n12291763": "Eustoma, genus Eustoma",
  "n12291959": "prairie gentian, tulip gentian, bluebell, Eustoma grandiflorum",
  "n12292285": "Exacum, genus Exacum",
  "n12292463": "Persian violet, Exacum affine",
  "n12292655": "Frasera, genus Frasera",
  "n12292877": "columbo, American columbo, deer's-ear, deer's-ears, pyramid plant, American gentian",
  "n12293180": "green gentian, Frasera speciosa, Swertia speciosa",
  "n12293419": "Gentiana, genus Gentiana",
  "n12293723": "gentian",
  "n12294124": "gentianella, Gentiana acaulis",
  "n12294331": "closed gentian, blind gentian, bottle gentian, Gentiana andrewsii",
  "n12294542": "explorer's gentian, Gentiana calycosa",
  "n12294723": "closed gentian, blind gentian, Gentiana clausa",
  "n12294871": "great yellow gentian, Gentiana lutea",
  "n12295033": "marsh gentian, calathian violet, Gentiana pneumonanthe",
  "n12295237": "soapwort gentian, Gentiana saponaria",
  "n12295429": "striped gentian, Gentiana villosa",
  "n12295560": "Gentianella, genus Gentianella",
  "n12295796": "agueweed, ague weed, five-flowered gentian, stiff gentian, Gentianella quinquefolia, Gentiana quinquefolia",
  "n12296045": "felwort, gentianella amarella",
  "n12296218": "Gentianopsis, genus Gentianopsis",
  "n12296432": "fringed gentian",
  "n12296735": "Gentianopsis crinita, Gentiana crinita",
  "n12296929": "Gentianopsis detonsa, Gentiana detonsa",
  "n12297110": "Gentianopsid procera, Gentiana procera",
  "n12297280": "Gentianopsis thermalis, Gentiana thermalis",
  "n12297507": "tufted gentian, Gentianopsis holopetala, Gentiana holopetala",
  "n12297678": "Halenia, genus Halenia",
  "n12297846": "spurred gentian",
  "n12298003": "genus Sabbatia",
  "n12298165": "sabbatia",
  "n12298395": "marsh pink, rose pink, bitter floom, American centaury, Sabbatia stellaris, Sabbatia Angularis",
  "n12298637": "prairia Sabbatia, Texas star, Sabbatia campestris",
  "n12298783": "Swertia, genus Swertia",
  "n12298958": "marsh felwort, Swertia perennia",
  "n12299165": "Salvadoraceae, family Salvadoraceae, Salvadora family",
  "n12299425": "Salvadora, genus Salvadora",
  "n12299640": "toothbrush tree, mustard tree, Salvadora persica",
  "n12299988": "Oleaceae, family Oleaceae, olive family",
  "n12300441": "Oleales, order Oleales",
  "n12300625": "Olea, genus Olea",
  "n12300840": "olive tree",
  "n12301038": "olive",
  "n12301180": "olive, European olive tree, Olea europaea",
  "n12301445": "olive",
  "n12301613": "black maire, Olea cunninghamii",
  "n12301766": "white maire, Olea lanceolata",
  "n12301917": "Chionanthus, genus Chionanthus",
  "n12302071": "fringe tree",
  "n12302248": "fringe bush, Chionanthus virginicus",
  "n12302418": "genus Forestiera",
  "n12302565": "forestiera",
  "n12302692": "tanglebush, desert olive, Forestiera neomexicana",
  "n12302974": "genus Forsythia",
  "n12303083": "forsythia",
  "n12303349": "Fraxinus, genus Fraxinus",
  "n12303462": "ash, ash tree",
  "n12303921": "ash",
  "n12304115": "white ash, Fraxinus Americana",
  "n12304286": "swamp ash, Fraxinus caroliniana",
  "n12304420": "flowering ash, Fraxinus cuspidata",
  "n12304572": "flowering ash, Fraxinus dipetala",
  "n12304703": "European ash, common European ash, Fraxinus excelsior",
  "n12304899": "Oregon ash, Fraxinus latifolia, Fraxinus oregona",
  "n12305089": "black ash, basket ash, brown ash, hoop ash, Fraxinus nigra",
  "n12305293": "manna ash, flowering ash, Fraxinus ornus",
  "n12305475": "red ash, downy ash, Fraxinus pennsylvanica",
  "n12305654": "green ash, Fraxinus pennsylvanica subintegerrima",
  "n12305819": "blue ash, Fraxinus quadrangulata",
  "n12305986": "mountain ash, Fraxinus texensis",
  "n12306089": "pumpkin ash, Fraxinus tomentosa",
  "n12306270": "Arizona ash, Fraxinus velutina",
  "n12306419": "ash-key",
  "n12306519": "Jasminum, genus Jasminum",
  "n12306717": "jasmine",
  "n12306938": "primrose jasmine, Jasminum mesnyi",
  "n12307076": "winter jasmine, Jasminum nudiflorum",
  "n12307240": "common jasmine, true jasmine, jessamine, Jasminum officinale",
  "n12307455": "Arabian jasmine, Jasminum sambac",
  "n12307611": "Ligustrum, genus Ligustrum",
  "n12307756": "privet",
  "n12308112": "Amur privet, Ligustrum amurense",
  "n12308259": "ibolium privet, ibota privet, Ligustrum ibolium",
  "n12308447": "Japanese privet, Ligustrum japonicum",
  "n12308664": "Chinese privet, white wax tree, Ligustrum lucidum",
  "n12308907": "Ligustrum obtusifolium",
  "n12309052": "California privet, Ligustrum ovalifolium",
  "n12309277": "common privet, Ligustrum vulgare",
  "n12309403": "Osmanthus, genus Osmanthus",
  "n12309630": "devilwood, American olive, Osmanthus americanus",
  "n12309850": "Phillyrea, genus Phillyrea",
  "n12310021": "mock privet",
  "n12310153": "Syringa, genus Syringa",
  "n12310349": "lilac",
  "n12310638": "Himalayan lilac, Syringa emodi",
  "n12310840": "Hungarian lilac, Syringa josikaea, Syringa josikea",
  "n12311045": "Persian lilac, Syringa persica",
  "n12311224": "Japanese tree lilac, Syringa reticulata, Syringa amurensis japonica",
  "n12311413": "Japanese lilac, Syringa villosa",
  "n12311579": "common lilac, Syringa vulgaris",
  "n12311800": "manna",
  "n12311894": "Haemodoraceae, family Haemodoraceae, bloodwort family",
  "n12312110": "bloodwort",
  "n12312276": "Haemodorum, genus Haemodorum",
  "n12312405": "Anigozanthus, genus Anigozanthus",
  "n12312728": "kangaroo paw, kangaroo's paw, kangaroo's-foot, kangaroo-foot plant, Australian sword lily, Anigozanthus manglesii",
  "n12313005": "Hamamelidae, subclass Hamamelidae",
  "n12313574": "Amentiferae, group Amentiferae",
  "n12313735": "Hamamelidanthum, genus Hamamelidanthum",
  "n12313954": "Hamamelidoxylon, genus Hamamelidoxylon",
  "n12314146": "Hamamelites, genus Hamamelites",
  "n12314315": "Hamamelidaceae, family Hamamelidaceae, witch-hazel family",
  "n12314652": "Hamamelis, genus Hamamelis",
  "n12314808": "witch hazel, witch hazel plant, wych hazel, wych hazel plant",
  "n12315060": "Virginian witch hazel, Hamamelis virginiana",
  "n12315245": "vernal witch hazel, Hamamelis vernalis",
  "n12315424": "Corylopsis, genus Corylopsis",
  "n12315598": "winter hazel, flowering hazel",
  "n12315818": "genus Fothergilla, Fothergilla",
  "n12315999": "fothergilla, witch alder",
  "n12316300": "genus Liquidambar, Liquidambar",
  "n12316444": "liquidambar",
  "n12316572": "sweet gum, sweet gum tree, bilsted, red gum, American sweet gum, Liquidambar styraciflua",
  "n12316853": "sweet gum, liquidambar",
  "n12316982": "sweet gum, satin walnut, hazelwood, red gum",
  "n12317164": "Parrotia, genus Parrotia",
  "n12317296": "iron tree, iron-tree, ironwood, ironwood tree",
  "n12317487": "ironwood",
  "n12317611": "Parrotiopsis, genus Parrotiopsis",
  "n12317763": "Juglandales, order Juglandales",
  "n12317919": "Juglandaceae, family Juglandaceae, walnut family",
  "n12318164": "Juglans, genus Juglans",
  "n12318378": "walnut, walnut tree",
  "n12318615": "walnut",
  "n12318782": "California black walnut, Juglans californica",
  "n12318965": "butternut, butternut tree, white walnut, Juglans cinerea",
  "n12319204": "black walnut, black walnut tree, black hickory, Juglans nigra",
  "n12319414": "English walnut, English walnut tree, Circassian walnut, Persian walnut, Juglans regia",
  "n12319687": "Carya, genus Carya",
  "n12320010": "hickory, hickory tree",
  "n12320285": "hickory",
  "n12320414": "water hickory, bitter pecan, water bitternut, Carya aquatica",
  "n12320627": "pignut, pignut hickory, brown hickory, black hickory, Carya glabra",
  "n12320806": "bitternut, bitternut hickory, bitter hickory, bitter pignut, swamp hickory, Carya cordiformis",
  "n12321077": "pecan, pecan tree, Carya illinoensis, Carya illinoinsis",
  "n12321304": "pecan",
  "n12321395": "big shellbark, big shellbark hickory, big shagbark, king nut, king nut hickory, Carya laciniosa",
  "n12321669": "nutmeg hickory, Carya myristicaeformis, Carya myristiciformis",
  "n12321873": "shagbark, shagbark hickory, shellbark, shellbark hickory, Carya ovata",
  "n12322099": "mockernut, mockernut hickory, black hickory, white-heart hickory, big-bud hickory, Carya tomentosa",
  "n12322359": "Pterocarya, genus Pterocarya",
  "n12322501": "wing nut, wing-nut",
  "n12322699": "Caucasian walnut, Pterocarya fraxinifolia",
  "n12322887": "Myrtales, order Myrtales, Thymelaeales, order Thymelaeales",
  "n12323411": "Combretaceae, family Combretaceae, combretum family",
  "n12323665": "dhawa, dhava",
  "n12323820": "genus Combretum",
  "n12324056": "combretum",
  "n12324222": "hiccup nut, hiccough nut, Combretum bracteosum",
  "n12324388": "bush willow, Combretum appiculatum",
  "n12324558": "bush willow, Combretum erythrophyllum",
  "n12324756": "Conocarpus, genus Conocarpus",
  "n12324906": "button tree, button mangrove, Conocarpus erectus",
  "n12325093": "Laguncularia, genus Laguncularia",
  "n12325234": "white mangrove, Laguncularia racemosa",
  "n12325497": "Elaeagnaceae, family Elaeagnaceae, oleaster family",
  "n12325667": "Elaeagnus, genus Elaeagnus",
  "n12325787": "oleaster",
  "n12326033": "wild olive, Elaeagnus latifolia",
  "n12326178": "silverberry, silver berry, silverbush, silver-bush, Elaeagnus commutata",
  "n12326369": "Russian olive, silver berry, Elaeagnus augustifolia",
  "n12326604": "Haloragidaceae, family Haloragidaceae, Haloragaceae, family Haloragaceae, water-milfoil family",
  "n12326842": "Myriophyllum, genus Myriophyllum",
  "n12327022": "water milfoil",
  "n12327209": "Lecythidaceae, family Lecythidaceae",
  "n12327407": "Grias, genus Grias",
  "n12327528": "anchovy pear, anchovy pear tree, Grias cauliflora",
  "n12327718": "Bertholletia, genus Bertholletia",
  "n12327846": "brazil nut, brazil-nut tree, Bertholletia excelsa",
  "n12328026": "Lythraceae, family Lythraceae, loosestrife family",
  "n12328241": "Lythrum, genus Lythrum",
  "n12328398": "loosestrife",
  "n12328567": "purple loosestrife, spiked loosestrife, Lythrum salicaria",
  "n12328801": "grass poly, hyssop loosestrife, Lythrum hyssopifolia",
  "n12329020": "Lagerstroemia, genus Lagerstroemia",
  "n12329260": "crape myrtle, crepe myrtle, crepe flower, Lagerstroemia indica",
  "n12329473": "Queen's crape myrtle, pride-of-India, Lagerstroemia speciosa",
  "n12329744": "pyinma",
  "n12329899": "Myrtaceae, family Myrtaceae, myrtle family",
  "n12330239": "myrtaceous tree",
  "n12330336": "Myrtus, genus Myrtus",
  "n12330469": "myrtle",
  "n12330587": "common myrtle, Myrtus communis",
  "n12330751": "Pimenta, genus Pimenta",
  "n12330891": "bayberry, bay-rum tree, Jamaica bayberry, wild cinnamon, Pimenta acris",
  "n12331066": "allspice, allspice tree, pimento tree, Pimenta dioica",
  "n12331263": "allspice tree, Pimenta officinalis",
  "n12331415": "Eugenia, genus Eugenia",
  "n12331655": "sour cherry, Eugenia corynantha",
  "n12331788": "nakedwood, Eugenia dicrana",
  "n12332030": "Surinam cherry, pitanga, Eugenia uniflora",
  "n12332218": "rose apple, rose-apple tree, jambosa, Eugenia jambos",
  "n12332422": "genus Feijoa",
  "n12332555": "feijoa, feijoa bush",
  "n12332718": "Jambos, genus Jambos",
  "n12332866": "Myrciaria, genus Myrciaria, Myrcia",
  "n12333053": "jaboticaba, jaboticaba tree, Myrciaria cauliflora",
  "n12333397": "Psidium, genus Psidium",
  "n12333530": "guava, true guava, guava bush, Psidium guajava",
  "n12333771": "guava, strawberry guava, yellow cattley guava, Psidium littorale",
  "n12333961": "cattley guava, purple strawberry guava, Psidium cattleianum, Psidium littorale longipes",
  "n12334153": "Brazilian guava, Psidium guineense",
  "n12334293": "gum tree, gum",
  "n12334520": "gumwood, gum",
  "n12334686": "genus Eucalyptus",
  "n12334891": "eucalyptus, eucalypt, eucalyptus tree",
  "n12335351": "eucalyptus",
  "n12335483": "flooded gum",
  "n12335664": "mallee",
  "n12335800": "stringybark",
  "n12335937": "smoothbark",
  "n12336092": "red gum, peppermint, peppermint gum, Eucalyptus amygdalina",
  "n12336224": "red gum, marri, Eucalyptus calophylla",
  "n12336333": "river red gum, river gum, Eucalyptus camaldulensis, Eucalyptus rostrata",
  "n12336586": "mountain swamp gum, Eucalyptus camphora",
  "n12336727": "snow gum, ghost gum, white ash, Eucalyptus coriacea, Eucalyptus pauciflora",
  "n12336973": "alpine ash, mountain oak, Eucalyptus delegatensis",
  "n12337131": "white mallee, congoo mallee, Eucalyptus dumosa",
  "n12337246": "white stringybark, thin-leaved stringybark, Eucalyptusd eugenioides",
  "n12337391": "white mountain ash, Eucalyptus fraxinoides",
  "n12337617": "blue gum, fever tree, Eucalyptus globulus",
  "n12337800": "rose gum, Eucalypt grandis",
  "n12337922": "cider gum, Eucalypt gunnii",
  "n12338034": "swamp gum, Eucalypt ovata",
  "n12338146": "spotted gum, Eucalyptus maculata",
  "n12338258": "lemon-scented gum, Eucalyptus citriodora, Eucalyptus maculata citriodora",
  "n12338454": "black mallee, black sally, black gum, Eucalytus stellulata",
  "n12338655": "forest red gum, Eucalypt tereticornis",
  "n12338796": "mountain ash, Eucalyptus regnans",
  "n12338979": "manna gum, Eucalyptus viminalis",
  "n12339090": "eucalyptus gum, eucalyptus kino, red gum",
  "n12339319": "Syzygium, genus Syzygium",
  "n12339526": "clove, clove tree, Syzygium aromaticum, Eugenia aromaticum, Eugenia caryophyllatum",
  "n12339831": "clove",
  "n12339972": "Nyssaceae, family Nyssaceae, sour-gum family, tupelo family",
  "n12340202": "Nyssa, genus Nyssa",
  "n12340383": "tupelo, tupelo tree",
  "n12340581": "water gum, Nyssa aquatica",
  "n12340755": "sour gum, black gum, pepperidge, Nyssa sylvatica",
  "n12340998": "tupelo",
  "n12341126": "Onagraceae, family Onagraceae, evening-primrose family",
  "n12341412": "Circaea, genus Circaea",
  "n12341542": "enchanter's nightshade",
  "n12341795": "Alpine enchanter's nightshade, Circaea alpina",
  "n12341931": "Circaea lutetiana",
  "n12342043": "Epilobium, genus Epilobium",
  "n12342299": "willowherb",
  "n12342498": "fireweed, giant willowherb, rosebay willowherb, wickup, Epilobium angustifolium",
  "n12342852": "California fuchsia, humming bird's trumpet, Epilobium canum canum, Zauschneria californica",
  "n12343092": "hairy willowherb, codlins-and-cream, Epilobium hirsutum",
  "n12343306": "genus Fuchsia",
  "n12343480": "fuchsia",
  "n12343753": "lady's-eardrop, ladies'-eardrop, lady's-eardrops, ladies'-eardrops, Fuchsia coccinea",
  "n12343951": "konini, tree fuchsia, native fuchsia, Fuchsia excorticata",
  "n12344131": "Oenothera, genus Oenothera",
  "n12344283": "evening primrose",
  "n12344483": "common evening primrose, German rampion, Oenothera biennis",
  "n12344700": "sundrops, Oenothera fruticosa",
  "n12344837": "Missouri primrose, Ozark sundrops, Oenothera macrocarpa",
  "n12344996": "Punicaceae, family Punicaceae",
  "n12345136": "Punica, genus Punica",
  "n12345280": "pomegranate, pomegranate tree, Punica granatum",
  "n12345495": "Rhizophoraceae, family Rhizophoraceae, mangrove family",
  "n12345709": "Rhizophora, genus Rhizophora",
  "n12345899": "mangrove, Rhizophora mangle",
  "n12346179": "Thymelaeaceae, family Thymelaeaceae, daphne family",
  "n12346448": "genus Daphne",
  "n12346578": "daphne",
  "n12346813": "garland flower, Daphne cneorum",
  "n12346986": "spurge laurel, wood laurel, Daphne laureola",
  "n12347158": "mezereon, February daphne, Daphne mezereum",
  "n12347380": "mezereum",
  "n12347490": "Dirca, genus Dirca",
  "n12347639": "leatherwood, moosewood, moose-wood, wicopy, ropebark, Dirca palustris",
  "n12347892": "Trapaceae, family Trapaceae",
  "n12348127": "Trapa, genus Trapa",
  "n12348294": "water chestnut, water chestnut plant, caltrop",
  "n12348518": "water caltrop, Jesuits' nut, Trapa natans",
  "n12348635": "ling, ling ko, Trapa bicornis",
  "n12348774": "Melastomataceae, family Melastomataceae, Melastomaceae, family Melastomaceae, meadow-beauty family",
  "n12349091": "Melastoma, genus Melastoma",
  "n12349315": "Indian rhododendron, Melastoma malabathricum",
  "n12349491": "Medinilla, genus Medinilla",
  "n12349711": "Medinilla magnifica",
  "n12349916": "Rhexia, genus Rhexia",
  "n12350032": "deer grass, meadow beauty",
  "n12350234": "Musales, order Musales",
  "n12350433": "Cannaceae, family Cannaceae",
  "n12350578": "genus Canna",
  "n12350758": "canna",
  "n12350959": "canna lily, Canna generalis",
  "n12351091": "achira, indian shot, arrowroot, Canna indica, Canna edulis",
  "n12351287": "Marantaceae, family Marantaceae, arrowroot family",
  "n12351477": "genus Maranta",
  "n12351600": "maranta",
  "n12351790": "arrowroot, American arrowroot, obedience plant, Maranta arundinaceae",
  "n12351975": "Musaceae, family Musaceae, banana family",
  "n12352150": "Musa, genus Musa",
  "n12352287": "banana, banana tree",
  "n12352639": "dwarf banana, Musa acuminata",
  "n12352844": "Japanese banana, Musa basjoo",
  "n12352990": "plantain, plantain tree, Musa paradisiaca",
  "n12353203": "edible banana, Musa paradisiaca sapientum",
  "n12353431": "abaca, Manila hemp, Musa textilis",
  "n12353604": "Ensete, genus Ensete",
  "n12353754": "Abyssinian banana, Ethiopian banana, Ensete ventricosum, Musa ensete",
  "n12354068": "Strelitziaceae, family Strelitziaceae, strelitzia family",
  "n12354374": "Strelitzia, genus Strelitzia",
  "n12354619": "bird of paradise, Strelitzia reginae",
  "n12354849": "genus Ravenala",
  "n12355023": "traveler's tree, traveller's tree, ravenala, Ravenala madagascariensis",
  "n12355320": "Zingiberaceae, family Zingiberaceae, ginger family",
  "n12355594": "Zingiber, genus Zingiber",
  "n12355760": "ginger",
  "n12356023": "common ginger, Canton ginger, stem ginger, Zingiber officinale",
  "n12356255": "Curcuma, genus Curcuma",
  "n12356395": "turmeric, Curcuma longa, Curcuma domestica",
  "n12356668": "Alpinia, genus Alpinia, genus Zerumbet, genus Languas",
  "n12356960": "galangal, Alpinia galanga",
  "n12357100": "lesser galangal, Alpinia officinarum, Alpinia officinalis",
  "n12357343": "red ginger, Alpinia purpurata",
  "n12357485": "shellflower, shall-flower, shell ginger, Alpinia Zerumbet, Alpinia speciosa, Languas speciosa",
  "n12357802": "Aframomum, genus Aframomum",
  "n12357968": "grains of paradise, Guinea grains, Guinea pepper, melagueta pepper, Aframomum melegueta",
  "n12358173": "Elettaria, genus Elettaria",
  "n12358293": "cardamom, cardamon, Elettaria cardamomum",
  "n12358485": "Dilleniidae, subclass Dilleniidae",
  "n12359026": "Parietales, order Parietales, Hypericales, order Hypericales",
  "n12359578": "Guttiferales, order Guttiferales",
  "n12359734": "Begoniaceae, family Begoniaceae, begonia family",
  "n12359952": "genus Begonia",
  "n12360108": "begonia",
  "n12360534": "fibrous-rooted begonia",
  "n12360684": "tuberous begonia",
  "n12360817": "rhizomatous begonia",
  "n12360958": "Christmas begonia, blooming-fool begonia, Begonia cheimantha",
  "n12361135": "angel-wing begonia, Begonia cocchinea",
  "n12361350": "grape-leaf begonia, maple-leaf begonia, Begonia dregei",
  "n12361560": "beefsteak begonia, kidney begonia, Begonia erythrophylla, Begonia feastii",
  "n12361754": "star begonia, star-leaf begonia, Begonia heracleifolia",
  "n12361946": "rex begonia, king begonia, painted-leaf begonia, beefsteak geranium, Begonia rex",
  "n12362274": "wax begonia, Begonia semperflorens",
  "n12362514": "Socotra begonia, Begonia socotrana",
  "n12362668": "hybrid tuberous begonia, Begonia tuberhybrida",
  "n12362844": "Dilleniaceae, family Dilleniaceae",
  "n12363110": "genus Dillenia",
  "n12363301": "dillenia",
  "n12363580": "Hibbertia, genus Hibbertia",
  "n12363768": "guinea gold vine, guinea flower",
  "n12363988": "Guttiferae, family Guttiferae, Clusiaceae, family Clusiaceae, St John's wort family",
  "n12364379": "Calophyllum, genus Calophyllum",
  "n12364604": "poon",
  "n12364823": "poon",
  "n12364940": "calaba, Santa Maria tree, Calophyllum calaba",
  "n12365158": "Maria, Calophyllum longifolium",
  "n12365285": "laurelwood, lancewood tree, Calophyllum candidissimum",
  "n12365462": "Alexandrian laurel, Calophyllum inophyllum",
  "n12365670": "genus Clusia",
  "n12365900": "clusia",
  "n12366053": "wild fig, Clusia flava",
  "n12366186": "waxflower, Clusia insignis",
  "n12366313": "pitch apple, strangler fig, Clusia rosea, Clusia major",
  "n12366507": "Garcinia, genus Garcinia",
  "n12366675": "mangosteen, mangosteen tree, Garcinia mangostana",
  "n12366870": "gamboge tree, Garcinia hanburyi, Garcinia cambogia, Garcinia gummi-gutta",
  "n12367122": "Hypericaceae, family Hypericaceae",
  "n12367306": "Hypericum, genus Hypericum",
  "n12367611": "St John's wort",
  "n12368028": "common St John's wort, tutsan, Hypericum androsaemum",
  "n12368257": "great St John's wort, Hypericum ascyron, Hypericum pyramidatum",
  "n12368451": "creeping St John's wort, Hypericum calycinum",
  "n12368637": "orange grass, nitweed, pineweed, pine-weed, Hypericum gentianoides",
  "n12368847": "St Andrews's cross, Hypericum crux andrae",
  "n12369066": "low St Andrew's cross, Hypericum hypericoides",
  "n12369309": "klammath weed, Hypericum perforatum",
  "n12369476": "shrubby St John's wort, Hypericum prolificum, Hypericum spathulatum",
  "n12369665": "St Peter's wort, Hypericum tetrapterum, Hypericum maculatum",
  "n12369845": "marsh St-John's wort, Hypericum virginianum",
  "n12370011": "Mammea, genus Mammea",
  "n12370174": "mammee apple, mammee, mamey, mammee tree, Mammea americana",
  "n12370384": "Mesua, genus Mesua",
  "n12370549": "rose chestnut, ironwood, ironwood tree, Mesua ferrea",
  "n12370842": "Actinidiaceae, family Actinidiaceae",
  "n12371002": "Actinidia, genus Actinidia",
  "n12371202": "bower actinidia, tara vine, Actinidia arguta",
  "n12371439": "Chinese gooseberry, kiwi, kiwi vine, Actinidia chinensis, Actinidia deliciosa",
  "n12371704": "silvervine, silver vine, Actinidia polygama",
  "n12371911": "Canellaceae, family Canellaceae, canella family",
  "n12372124": "genus Canella",
  "n12372233": "wild cinnamon, white cinnamon tree, Canella winterana, Canella-alba",
  "n12372520": "canella, canella bark, white cinnamon",
  "n12372708": "Caricaceae, family Caricaceae, papaya family",
  "n12372932": "Carica, genus Carica",
  "n12373100": "papaya, papaia, pawpaw, papaya tree, melon tree, Carica papaya",
  "n12373361": "Caryocaraceae, family Caryocaraceae",
  "n12373526": "Caryocar, genus Caryocar",
  "n12373739": "souari, souari nut, souari tree, Caryocar nuciferum",
  "n12374002": "Cistaceae, family Cistaceae, rockrose family",
  "n12374238": "Cistus, genus Cistus",
  "n12374418": "rockrose, rock rose",
  "n12374705": "white-leaved rockrose, Cistus albidus",
  "n12374862": "common gum cistus, Cistus ladanifer, Cistus ladanum",
  "n12375107": "labdanum, ladanum",
  "n12375294": "genus Helianthemum",
  "n12375518": "helianthemum, sunrose, sun rose",
  "n12375769": "frostweed, frost-weed, frostwort, Helianthemum canadense, Crocanthemum canadense",
  "n12376073": "rockrose, rock rose",
  "n12376240": "rush rose, Helianthemum scoparium",
  "n12376382": "Hudsonia, genus Hudsonia",
  "n12376553": "false heather, golden heather, Hudsonia ericoides",
  "n12376740": "beach heather, poverty grass, Hudsonia tomentosa",
  "n12376950": "Dipterocarpaceae, family Dipterocarpaceae",
  "n12377198": "dipterocarp",
  "n12377328": "Shorea, genus Shorea",
  "n12377494": "red lauan, red lauan tree, Shorea teysmanniana",
  "n12377658": "red lauan",
  "n12377809": "Flacourtiaceae, family Flacourtiaceae, flacourtia family",
  "n12378080": "Flacourtia, genus Flacourtia",
  "n12378249": "governor's plum, governor plum, Madagascar plum, ramontchi, batoko palm, Flacourtia indica",
  "n12378546": "Dovyalis, genus Dovyalis",
  "n12378753": "kei apple, kei apple bush, Dovyalis caffra",
  "n12378963": "ketembilla, kitembilla, kitambilla, ketembilla tree, Ceylon gooseberry, Dovyalis hebecarpa",
  "n12379278": "Hydnocarpus, genus Hydnocarpus, Taraktagenos, genus Taraktagenos, Taraktogenos, genus Taraktogenos",
  "n12379531": "chaulmoogra, chaulmoogra tree, chaulmugra, Hydnocarpus kurzii, Taraktagenos kurzii, Taraktogenos kurzii",
  "n12379781": "Hydnocarpus laurifolia, Hydnocarpus wightiana",
  "n12380029": "hydnocarpus oil",
  "n12380197": "genus Idesia",
  "n12380305": "idesia, Idesia polycarpa",
  "n12380597": "Kiggelaria, genus Kiggelaria",
  "n12380761": "wild peach, Kiggelaria africana",
  "n12380926": "genus Xylosma",
  "n12381095": "xylosma, Xylosma congestum",
  "n12381321": "Fouquieriaceae, family Fouquieriaceae",
  "n12381511": "candlewood",
  "n12381666": "Fouquieria, genus Fouquieria",
  "n12381931": "ocotillo, coachwhip, Jacob's staff, vine cactus, Fouquieria splendens",
  "n12382233": "boojum tree, cirio, Fouquieria columnaris, Idria columnaris",
  "n12382484": "Ochnaceae, family Ochnaceae, ochna family",
  "n12382699": "Ochna, genus Ochna",
  "n12382875": "bird's-eye bush, Ochna serrulata",
  "n12383073": "Passifloraceae, family Passifloraceae, passionflower family",
  "n12383256": "Passiflora, genus Passiflora",
  "n12383402": "passionflower, passionflower vine",
  "n12383737": "granadilla, purple granadillo, Passiflora edulis",
  "n12383894": "granadilla, sweet granadilla, Passiflora ligularis",
  "n12384037": "granadilla, giant granadilla, Passiflora quadrangularis",
  "n12384227": "maypop, Passiflora incarnata",
  "n12384375": "Jamaica honeysuckle, yellow granadilla, Passiflora laurifolia",
  "n12384569": "banana passion fruit, Passiflora mollissima",
  "n12384680": "sweet calabash, Passiflora maliformis",
  "n12384839": "love-in-a-mist, running pop, wild water lemon, Passiflora foetida",
  "n12385046": "Resedaceae, family Resedaceae, mignonette family",
  "n12385219": "genus Reseda",
  "n12385429": "reseda",
  "n12385566": "mignonette, sweet reseda, Reseda odorata",
  "n12385830": "dyer's rocket, dyer's mignonette, weld, Reseda luteola",
  "n12386039": "Tamaricaceae, family Tamaricaceae, tamarisk family",
  "n12386263": "Tamarix, genus Tamarix",
  "n12386462": "tamarisk",
  "n12386724": "Myricaria, genus Myricaria",
  "n12386945": "false tamarisk, German tamarisk, Myricaria germanica",
  "n12387103": "halophyte",
  "n12387201": "Violaceae, family Violaceae, violet family",
  "n12387478": "Viola, genus Viola",
  "n12387633": "viola",
  "n12387839": "violet",
  "n12388143": "field pansy, heartsease, Viola arvensis",
  "n12388293": "American dog violet, Viola conspersa",
  "n12388444": "sweet white violet, white violet, woodland white violet, Viola blanda",
  "n12388652": "Canada violet, tall white violet, white violet, Viola canadensis",
  "n12388858": "dog violet, heath violet, Viola canina",
  "n12388989": "horned violet, tufted pansy, Viola cornuta",
  "n12389130": "two-eyed violet, heartsease, Viola ocellata",
  "n12389317": "sweet violet, garden violet, English violet, Viola odorata",
  "n12389501": "bird's-foot violet, pansy violet, Johnny-jump-up, wood violet, Viola pedata",
  "n12389727": "downy yellow violet, Viola pubescens",
  "n12389932": "long-spurred violet, Viola rostrata",
  "n12390099": "pale violet, striped violet, cream violet, Viola striata",
  "n12390314": "hedge violet, wood violet, Viola sylvatica, Viola reichenbachiana",
  "n12390485": "pansy, Viola tricolor hortensis",
  "n12390681": "wild pansy, Johnny-jump-up, heartsease, love-in-idleness, pink of my John, Viola tricolor",
  "n12390914": "Hybanthus, genus Hybanthus",
  "n12391111": "Hymenanthera, genus Hymenanthera",
  "n12391280": "Melicytus, genus Melicytus",
  "n12391477": "Urticales, order Urticales",
  "n12391745": "Urticaceae, family Urticaceae, nettle family",
  "n12392070": "nettle",
  "n12392385": "Urtica, genus Urtica",
  "n12392549": "stinging nettle, Urtica dioica",
  "n12392765": "Roman nettle, Urtica pipulifera",
  "n12392943": "Boehmeria, genus Boehmeria",
  "n12393086": "false nettle, bog hemp",
  "n12393269": "ramie, ramee, Chinese silk plant, China grass, Boehmeria nivea",
  "n12393527": "Helxine, genus Helxine, Soleirolia, genus Soleirolia",
  "n12393723": "baby's tears, baby tears, Helxine soleirolia, Soleirolia soleirolii",
  "n12393942": "Laportea, genus Laportea",
  "n12394118": "wood nettle, Laportea canadensis",
  "n12394328": "Australian nettle, Australian nettle tree",
  "n12394494": "Parietaria, genus Parietaria",
  "n12394638": "pellitory-of-the-wall, wall pellitory, pellitory, Parietaria difussa",
  "n12394861": "Pilea, genus Pilea",
  "n12395068": "richweed, clearweed, dead nettle, Pilea pumilla",
  "n12395289": "artillery plant, Pilea microphylla",
  "n12395463": "friendship plant, panamica, panamiga, Pilea involucrata",
  "n12395717": "Pipturus, genus Pipturus",
  "n12395906": "Queensland grass-cloth plant, Pipturus argenteus",
  "n12396091": "Pipturus albidus",
  "n12396255": "Cannabidaceae, family Cannabidaceae, hemp family",
  "n12396666": "genus Cannabis",
  "n12396924": "cannabis, hemp",
  "n12397210": "marijuana, marihuana, ganja, Cannabis sativa",
  "n12397431": "Indian hemp, Cannabis indica",
  "n12397594": "Humulus, genus Humulus",
  "n12397864": "hop, hops",
  "n12398174": "common hop, common hops, bine, European hop, Humulus lupulus",
  "n12398384": "American hop, Humulus americanus",
  "n12398526": "Japanese hop, Humulus japonicus",
  "n12398682": "Moraceae, family Moraceae, mulberry family",
  "n12398990": "Morus, genus Morus",
  "n12399132": "mulberry, mulberry tree",
  "n12399384": "white mulberry, Morus alba",
  "n12399534": "black mulberry, Morus nigra",
  "n12399656": "red mulberry, Morus rubra",
  "n12399784": "Maclura, genus Maclura",
  "n12399899": "osage orange, bow wood, mock orange, Maclura pomifera",
  "n12400261": "Artocarpus, genus Artocarpus",
  "n12400489": "breadfruit, breadfruit tree, Artocarpus communis, Artocarpus altilis",
  "n12400720": "jackfruit, jackfruit tree, Artocarpus heterophyllus",
  "n12400924": "marang, marang tree, Artocarpus odoratissima",
  "n12401122": "Ficus, genus Ficus",
  "n12401335": "fig tree",
  "n12401684": "fig, common fig, common fig tree, Ficus carica",
  "n12401893": "caprifig, Ficus carica sylvestris",
  "n12402051": "golden fig, Florida strangler fig, strangler fig, wild fig, Ficus aurea",
  "n12402348": "banyan, banyan tree, banian, banian tree, Indian banyan, East Indian fig tree, Ficus bengalensis",
  "n12402596": "pipal, pipal tree, pipul, peepul, sacred fig, bo tree, Ficus religiosa",
  "n12402840": "India-rubber tree, India-rubber plant, India-rubber fig, rubber plant, Assam rubber, Ficus elastica",
  "n12403075": "mistletoe fig, mistletoe rubber plant, Ficus diversifolia, Ficus deltoidea",
  "n12403276": "Port Jackson fig, rusty rig, little-leaf fig, Botany Bay fig, Ficus rubiginosa",
  "n12403513": "sycamore, sycamore fig, mulberry fig, Ficus sycomorus",
  "n12403862": "Broussonetia, genus Broussonetia",
  "n12403994": "paper mulberry, Broussonetia papyrifera",
  "n12404314": "Cecropiaceae, family Cecropiaceae",
  "n12404484": "Cecropia, genus Cecropia",
  "n12404729": "trumpetwood, trumpet-wood, trumpet tree, snake wood, imbauba, Cecropia peltata",
  "n12404943": "Ulmaceae, family Ulmaceae, elm family",
  "n12405209": "Ulmus, genus Ulmus",
  "n12405714": "elm, elm tree",
  "n12406155": "elm, elmwood",
  "n12406304": "winged elm, wing elm, Ulmus alata",
  "n12406488": "American elm, white elm, water elm, rock elm, Ulmus americana",
  "n12406715": "smooth-leaved elm, European field elm, Ulmus carpinifolia",
  "n12406902": "cedar elm, Ulmus crassifolia",
  "n12407079": "witch elm, wych elm, Ulmus glabra",
  "n12407222": "Dutch elm, Ulmus hollandica",
  "n12407396": "Huntingdon elm, Ulmus hollandica vegetata",
  "n12407545": "water elm, Ulmus laevis",
  "n12407715": "Chinese elm, Ulmus parvifolia",
  "n12407890": "English elm, European elm, Ulmus procera",
  "n12408077": "Siberian elm, Chinese elm, dwarf elm, Ulmus pumila",
  "n12408280": "slippery elm, red elm, Ulmus rubra",
  "n12408466": "Jersey elm, guernsey elm, wheately elm, Ulmus sarniensis, Ulmus campestris sarniensis, Ulmus campestris wheatleyi",
  "n12408717": "September elm, red elm, Ulmus serotina",
  "n12408873": "rock elm, Ulmus thomasii",
  "n12409016": "Celtis, genus Celtis",
  "n12409231": "hackberry, nettle tree",
  "n12409470": "European hackberry, Mediterranean hackberry, Celtis australis",
  "n12409651": "American hackberry, Celtis occidentalis",
  "n12409840": "sugarberry, Celtis laevigata",
  "n12410032": "Planera, genus Planera",
  "n12410205": "Trema, genus Trema",
  "n12410381": "Liliidae, subclass Liliidae",
  "n12410715": "Liliales, order Liliales",
  "n12411084": "Iridaceae, family Iridaceae, iris family",
  "n12411461": "iridaceous plant",
  "n12411710": "genus Iris",
  "n12411922": "iris, flag, fleur-de-lis, sword lily",
  "n12412355": "bearded iris",
  "n12412606": "beardless iris",
  "n12412850": "bulbous iris",
  "n12412987": "orrisroot, orris",
  "n12413165": "dwarf iris, Iris cristata",
  "n12413301": "Dutch iris, Iris filifolia",
  "n12413419": "Florentine iris, orris, Iris germanica florentina, Iris florentina",
  "n12413642": "stinking iris, gladdon, gladdon iris, stinking gladwyn, roast beef plant, Iris foetidissima",
  "n12413880": "German iris, Iris germanica",
  "n12414035": "Japanese iris, Iris kaempferi",
  "n12414159": "German iris, Iris kochii",
  "n12414329": "Dalmatian iris, Iris pallida",
  "n12414449": "Persian iris, Iris persica",
  "n12414602": "yellow iris, yellow flag, yellow water flag, Iris pseudacorus",
  "n12414818": "Dutch iris, Iris tingitana",
  "n12414932": "dwarf iris, vernal iris, Iris verna",
  "n12415089": "blue flag, Iris versicolor",
  "n12415272": "southern blue flag, Iris virginica",
  "n12415401": "English iris, Iris xiphioides",
  "n12415595": "Spanish iris, xiphium iris, Iris xiphium",
  "n12415765": "falls",
  "n12415911": "Belamcanda, genus Belamcanda",
  "n12416073": "blackberry-lily, leopard lily, Belamcanda chinensis",
  "n12416278": "genus Crocus",
  "n12416423": "crocus",
  "n12416703": "saffron, saffron crocus, Crocus sativus",
  "n12416917": "genus Freesia",
  "n12417062": "freesia",
  "n12417273": "genus Gladiolus",
  "n12417382": "gladiolus, gladiola, glad, sword lily",
  "n12417686": "Ixia, genus Ixia",
  "n12417836": "corn lily",
  "n12418065": "Sisyrinchium, genus Sisyrinchium",
  "n12418221": "blue-eyed grass",
  "n12418356": "Sparaxis, genus Sparaxis",
  "n12418507": "wandflower, Sparaxis tricolor",
  "n12418680": "Amaryllidaceae, family Amaryllidaceae, amaryllis family",
  "n12419037": "amaryllis",
  "n12419217": "genus Amaryllis",
  "n12419394": "belladonna lily, naked lady, Amaryllis belladonna",
  "n12419592": "Bomarea, genus Bomarea",
  "n12419878": "salsilla, Bomarea edulis",
  "n12420124": "salsilla, Bomarea salsilla",
  "n12420335": "Haemanthus, genus Haemanthus",
  "n12420535": "blood lily",
  "n12420722": "Cape tulip, Haemanthus coccineus",
  "n12420991": "genus Hippeastrum",
  "n12421137": "hippeastrum, Hippeastrum puniceum",
  "n12421334": "genus Narcissus",
  "n12421467": "narcissus",
  "n12421683": "daffodil, Narcissus pseudonarcissus",
  "n12421917": "jonquil, Narcissus jonquilla",
  "n12422129": "jonquil",
  "n12422230": "paper white, Narcissus papyraceus",
  "n12422399": "Strekelia, genus Strekelia",
  "n12422559": "Jacobean lily, Aztec lily, Strekelia formosissima",
  "n12422751": "Hypoxidaceae, family Hypoxidaceae",
  "n12422931": "Hypoxis, genus Hypoxis",
  "n12423211": "star grass",
  "n12423444": "American star grass, Hypoxis hirsuta",
  "n12423565": "Liliaceae, family Liliaceae, lily family",
  "n12425281": "liliaceous plant",
  "n12426100": "Lilium, genus Lilium",
  "n12426248": "lily",
  "n12426623": "mountain lily, Lilium auratum",
  "n12426749": "Canada lily, wild yellow lily, meadow lily, wild meadow lily, Lilium canadense",
  "n12426978": "Madonna lily, white lily, Annunciation lily, Lent lily, Lilium candidum",
  "n12427184": "tiger lily, leopard lily, pine lily, Lilium catesbaei",
  "n12427391": "Columbia tiger lily, Oregon lily, Lilium columbianum",
  "n12427566": "tiger lily, devil lily, kentan, Lilium lancifolium",
  "n12427757": "Easter lily, Bermuda lily, white trumpet lily, Lilium longiflorum",
  "n12427946": "coast lily, Lilium maritinum",
  "n12428076": "Turk's-cap, martagon, Lilium martagon",
  "n12428242": "Michigan lily, Lilium michiganense",
  "n12428412": "leopard lily, panther lily, Lilium pardalinum",
  "n12428587": "wood lily, Lilium philadelphicum",
  "n12428747": "Turk's-cap, Turk's cap-lily, Lilium superbum",
  "n12428915": "genus Agapanthus",
  "n12429148": "agapanthus, lily of the Nile",
  "n12429352": "African lily, African tulip, blue African lily, Agapanthus africanus",
  "n12429589": "genus Albuca",
  "n12429770": "albuca",
  "n12429942": "Aletris, genus Aletris",
  "n12430198": "colicroot, colic root, crow corn, star grass, unicorn root",
  "n12430471": "ague root, ague grass, Aletris farinosa",
  "n12430675": "yellow colicroot, Aletris aurea",
  "n12430878": "Alliaceae, family Alliaceae",
  "n12431128": "Allium, genus Allium",
  "n12431434": "alliaceous plant",
  "n12431861": "wild onion",
  "n12432069": "Hooker's onion, Allium acuminatum",
  "n12432356": "wild leek, Levant garlic, kurrat, Allium ampeloprasum",
  "n12432574": "Canada garlic, meadow leek, rose leek, Allium canadense",
  "n12432707": "keeled garlic, Allium carinatum",
  "n12432808": "onion, onion plant, Allium cepa",
  "n12433081": "onion",
  "n12433178": "shallot, eschalot, multiplier onion, Allium cepa aggregatum, Allium ascalonicum",
  "n12433429": "shallot",
  "n12433540": "tree onion, Egyptian onion, top onion, Allium cepa viviparum",
  "n12433769": "nodding onion, nodding wild onion, lady's leek, Allium cernuum",
  "n12433952": "Welsh onion, Japanese leek, Allium fistulosum",
  "n12434106": "red-skinned onion, Allium haematochiton",
  "n12434239": "leek, scallion, Allium porrum",
  "n12434483": "daffodil garlic, flowering onion, Naples garlic, Allium neopolitanum",
  "n12434634": "few-flowered leek, Allium paradoxum",
  "n12434775": "garlic, Allium sativum",
  "n12434985": "sand leek, giant garlic, Spanish garlic, rocambole, Allium scorodoprasum",
  "n12435152": "chives, chive, cive, schnittlaugh, Allium schoenoprasum",
  "n12435338": "ramp, wild leek, Allium tricoccum",
  "n12435486": "crow garlic, false garlic, field garlic, stag's garlic, wild garlic, Allium vineale",
  "n12435649": "wild garlic, wood garlic, Ramsons, Allium ursinum",
  "n12435777": "garlic chive, Chinese chive, Oriental garlic, Allium tuberosum",
  "n12435965": "round-headed leek, Allium sphaerocephalum",
  "n12436090": "three-cornered leek, triquetrous leek, Allium triquetrum",
  "n12436260": "Aloeaceae, family Aloeaceae, aloe family",
  "n12436490": "genus Aloe",
  "n12436677": "aloe",
  "n12436907": "cape aloe, Aloe ferox",
  "n12437047": "burn plant, Aloe vera",
  "n12437311": "genus Kniphofia",
  "n12437513": "kniphofia, tritoma, flame flower, flame-flower, flameflower",
  "n12437769": "poker plant, Kniphofia uvaria",
  "n12437930": "red-hot poker, Kniphofia praecox",
  "n12438046": "Alstroemeriaceae, family Alstroemeriaceae",
  "n12438324": "genus Alstroemeria",
  "n12438571": "alstroemeria",
  "n12438783": "Peruvian lily, lily of the Incas, Alstroemeria pelegrina",
  "n12438977": "Amianthum, genus Amianthum",
  "n12439154": "fly poison, Amianthum muscaetoxicum, Amianthum muscitoxicum",
  "n12439400": "Anthericum, genus Anthericum",
  "n12439626": "Saint-Bernard's-lily, Anthericum liliago",
  "n12439830": "amber lily, Anthericum torreyi",
  "n12440128": "Aphyllanthaceae, family Aphyllanthaceae",
  "n12440385": "Aphyllanthes, genus Aphyllanthes",
  "n12440623": "Asparagaceae, family Asparagaceae",
  "n12440869": "genus Asparagus",
  "n12441183": "asparagus, edible asparagus, Asparagus officinales",
  "n12441390": "asparagus fern, Asparagus setaceous, Asparagus plumosus",
  "n12441552": "smilax, Asparagus asparagoides",
  "n12441770": "Asphodelaceae, family Asphodelaceae",
  "n12441958": "asphodel",
  "n12442220": "Asphodeline, genus Asphodeline",
  "n12442548": "Jacob's rod",
  "n12442697": "king's spear, yellow asphodel, Asphodeline lutea",
  "n12442865": "Asphodelus, genus Asphodelus",
  "n12443144": "genus Aspidistra",
  "n12443323": "aspidistra, cast-iron plant, bar-room plant, Aspidistra elatio",
  "n12443547": "Bessera, genus Bessera",
  "n12443736": "coral drops, Bessera elegans",
  "n12443929": "Blandfordia, genus Blandfordia",
  "n12444095": "Christmas bells",
  "n12444261": "Bloomeria, genus Bloomeria",
  "n12444490": "golden star, golden stars, Bloomeria crocea",
  "n12444666": "Bowiea, genus Bowiea",
  "n12444898": "climbing onion, Bowiea volubilis",
  "n12445138": "genus Brodiaea",
  "n12445387": "brodiaea",
  "n12445628": "elegant brodiaea, Brodiaea elegans",
  "n12445848": "Calochortus, genus Calochortus",
  "n12446200": "mariposa, mariposa tulip, mariposa lily",
  "n12446519": "globe lily, fairy lantern",
  "n12446737": "cat's-ear",
  "n12446908": "white globe lily, white fairy lantern, Calochortus albus",
  "n12447121": "yellow globe lily, golden fairy lantern, Calochortus amabilis",
  "n12447346": "rose globe lily, Calochortus amoenus",
  "n12447581": "star tulip, elegant cat's ears, Calochortus elegans",
  "n12447891": "desert mariposa tulip, Calochortus kennedyi",
  "n12448136": "yellow mariposa tulip, Calochortus luteus",
  "n12448361": "sagebrush mariposa tulip, Calochortus macrocarpus",
  "n12448700": "sego lily, Calochortus nuttallii",
  "n12449024": "Camassia, genus Camassia, Quamassia, genus Quamassia",
  "n12449296": "camas, camass, quamash, camosh, camash",
  "n12449526": "common camas, Camassia quamash",
  "n12449784": "Leichtlin's camas, Camassia leichtlinii",
  "n12449934": "wild hyacinth, indigo squill, Camassia scilloides",
  "n12450099": "Erythronium, genus Erythronium",
  "n12450344": "dogtooth violet, dogtooth, dog's-tooth violet",
  "n12450607": "white dogtooth violet, white dog's-tooth violet, blonde lilian, Erythronium albidum",
  "n12450840": "yellow adder's tongue, trout lily, amberbell, Erythronium americanum",
  "n12451070": "European dogtooth, Erythronium dens-canis",
  "n12451240": "fawn lily, Erythronium californicum",
  "n12451399": "glacier lily, snow lily, Erythronium grandiflorum",
  "n12451566": "avalanche lily, Erythronium montanum",
  "n12451789": "Fritillaria, genus Fritillaria",
  "n12451915": "fritillary, checkered lily",
  "n12452256": "mission bells, rice-grain fritillary, Fritillaria affinis, Fritillaria lanceolata, Fritillaria mutica",
  "n12452480": "mission bells, black fritillary, Fritillaria biflora",
  "n12452673": "stink bell, Fritillaria agrestis",
  "n12452836": "crown imperial, Fritillaria imperialis",
  "n12453018": "white fritillary, Fritillaria liliaceae",
  "n12453186": "snake's head fritillary, guinea-hen flower, checkered daffodil, leper lily, Fritillaria meleagris",
  "n12453539": "brown bells, Fritillaria micrantha, Fritillaria parviflora",
  "n12453714": "adobe lily, pink fritillary, Fritillaria pluriflora",
  "n12453857": "scarlet fritillary, Fritillaria recurva",
  "n12454021": "Tulipa, genus Tulipa",
  "n12454159": "tulip",
  "n12454436": "dwarf tulip, Tulipa armena, Tulipa suaveolens",
  "n12454556": "lady tulip, candlestick tulip, Tulipa clusiana",
  "n12454705": "Tulipa gesneriana",
  "n12454793": "cottage tulip",
  "n12454949": "Darwin tulip",
  "n12455101": "Colchicaceae, family Colchicaceae",
  "n12455342": "Colchicum, genus Colchicum",
  "n12455540": "autumn crocus, meadow saffron, naked lady, Colchicum autumnale",
  "n12455787": "genus Gloriosa",
  "n12455950": "gloriosa, glory lily, climbing lily, creeping lily, Gloriosa superba",
  "n12456278": "Hemerocallidaceae, family Hemerocallidaceae",
  "n12456527": "Hemerocallis, genus Hemerocallis",
  "n12456845": "day lily, daylily",
  "n12457091": "lemon lily, Hemerocallis lilio-asphodelus, Hemerocallis flava",
  "n12457250": "Hostaceae, family Hostaceae, Funkaceae, family Funkaceae",
  "n12457519": "Hosta, genus Hosta, Funka, genus Funka",
  "n12457771": "plantain lily, day lily",
  "n12458002": "Hyacinthaceae, family Hyacinthaceae",
  "n12458224": "genus Hyacinthus",
  "n12458383": "hyacinth",
  "n12458550": "common hyacinth, Hyacinthus orientalis",
  "n12458713": "Roman hyacinth, Hyacinthus orientalis albulus",
  "n12458874": "summer hyacinth, cape hyacinth, Hyacinthus candicans, Galtonia candicans",
  "n12459048": "Hyacinthoides, genus Hyacinthoides",
  "n12459275": "wild hyacinth, wood hyacinth, bluebell, harebell, Hyacinthoides nonscripta, Scilla nonscripta",
  "n12459471": "Ornithogalum, genus Ornithogalum",
  "n12459629": "star-of-Bethlehem",
  "n12459882": "starflower, sleepy dick, summer snowflake, Ornithogalum umbellatum",
  "n12460146": "bath asparagus, Prussian asparagus, Ornithogalum pyrenaicum",
  "n12460308": "chincherinchee, wonder flower, Ornithogalum thyrsoides",
  "n12460549": "Muscari, genus Muscari",
  "n12460697": "grape hyacinth",
  "n12460957": "common grape hyacinth, Muscari neglectum",
  "n12461109": "tassel hyacinth, Muscari comosum",
  "n12461326": "genus Scilla",
  "n12461466": "scilla, squill",
  "n12461673": "spring squill, Scilla verna, sea onion",
  "n12461809": "Tofieldia, genus Tofieldia",
  "n12462032": "false asphodel",
  "n12462221": "Scotch asphodel, Tofieldia pusilla",
  "n12462401": "Urginea, genus Urginea",
  "n12462582": "sea squill, sea onion, squill, Urginea maritima",
  "n12462805": "squill",
  "n12462951": "Ruscus, genus Ruscus",
  "n12463134": "butcher's broom, Ruscus aculeatus",
  "n12463322": "Melanthiaceae, family Melanthiaceae",
  "n12463574": "Narthecium, genus Narthecium",
  "n12463743": "bog asphodel",
  "n12463975": "European bog asphodel, Narthecium ossifragum",
  "n12464128": "American bog asphodel, Narthecium americanum",
  "n12464278": "Veratrum, genus Veratrum",
  "n12464476": "hellebore, false hellebore",
  "n12464649": "white hellebore, American hellebore, Indian poke, bugbane, Veratrum viride",
  "n12464903": "Ruscaceae, family Ruscaceae",
  "n12465107": "Tecophilaeacea, family Tecophilaeacea",
  "n12465321": "Xerophyllum, genus Xerophyllum",
  "n12465557": "squaw grass, bear grass, Xerophyllum tenax",
  "n12465796": "Xanthorrhoeaceae, family Xanthorrhoeaceae, grass tree family",
  "n12466034": "Xanthorroea, genus Xanthorroea",
  "n12466206": "grass tree, Australian grass tree",
  "n12466450": "Zigadenus, genus Zigadenus",
  "n12466727": "death camas, zigadene",
  "n12467018": "alkali grass, Zigadenus elegans",
  "n12467197": "white camas, Zigadenus glaucus",
  "n12467433": "poison camas, Zigadenus nuttalli",
  "n12467592": "grassy death camas, Zigadenus venenosus, Zigadenus venenosus gramineus",
  "n12467811": "Trilliaceae, family Trilliaceae, trillium family",
  "n12468081": "genus Trillium",
  "n12468243": "trillium, wood lily, wake-robin",
  "n12468545": "prairie wake-robin, prairie trillium, Trillium recurvatum",
  "n12468719": "dwarf-white trillium, snow trillium, early wake-robin",
  "n12468900": "purple trillium, red trillium, birthroot, Trillium erectum",
  "n12469157": "red trillium, toadshade, sessile trillium, Trillium sessile",
  "n12469372": "Paris, genus Paris",
  "n12469517": "herb Paris, Paris quadrifolia",
  "n12469725": "Smilacaceae, subfamily Smilacaceae",
  "n12469936": "Smilax, genus Smilax",
  "n12470092": "sarsaparilla",
  "n12470329": "sarsaparilla root",
  "n12470512": "bullbrier, greenbrier, catbrier, horse brier, horse-brier, brier, briar, Smilax rotundifolia",
  "n12470907": "rough bindweed, Smilax aspera",
  "n12471150": "Convallariaceae, family Convallariaceae",
  "n12471366": "Convallaria, genus Convallaria",
  "n12471544": "lily of the valley, May lily, Convallaria majalis",
  "n12471825": "genus Clintonia",
  "n12472024": "clintonia, Clinton's lily",
  "n12472326": "red Clintonia, Andrew's clintonia, Clintonia andrewsiana",
  "n12472559": "yellow clintonia, heal all, Clintonia borealis",
  "n12472778": "queen's cup, bride's bonnet, Clintonia uniflora",
  "n12473011": "Liriope, genus Liriope",
  "n12473171": "lilyturf, lily turf, Liriope muscari",
  "n12473405": "Maianthemum, genus Maianthemum",
  "n12473608": "false lily of the valley, Maianthemum canadense",
  "n12473840": "false lily of the valley, Maianthemum bifolium",
  "n12474006": "Polygonatum, genus Polygonatum",
  "n12474167": "Solomon's-seal",
  "n12474418": "great Solomon's-seal, Polygonatum biflorum, Polygonatum commutatum",
  "n12474620": "Uvulariaceae, subfamily Uvulariaceae",
  "n12474828": "Uvularia, genus Uvularia",
  "n12475035": "bellwort, merry bells, wild oats",
  "n12475242": "strawflower, cornflower, Uvularia grandiflora",
  "n12475450": "Taccaceae, family Taccaceae",
  "n12475593": "Tacca, genus Tacca",
  "n12475774": "pia, Indian arrowroot, Tacca leontopetaloides, Tacca pinnatifida",
  "n12476036": "Agavaceae, family Agavaceae, agave family, sisal family",
  "n12476510": "agave, century plant, American aloe",
  "n12476902": "genus Agave",
  "n12477163": "American agave, Agave americana",
  "n12477401": "sisal, Agave sisalana",
  "n12477583": "maguey, cantala, Agave cantala",
  "n12477747": "maguey, Agave atrovirens",
  "n12477983": "Agave tequilana",
  "n12478131": "cantala, Cebu maguey, manila maguey",
  "n12478283": "Cordyline, genus Cordyline",
  "n12478506": "ti, Cordyline terminalis",
  "n12478768": "cabbage tree, grass tree, Cordyline australis",
  "n12479066": "Dracenaceae, subfamily Dracenaceae, Dracaenaceae, subfamily Dracaenaceae",
  "n12479303": "genus Dracaena",
  "n12479537": "dracaena",
  "n12479689": "dragon tree, Dracaena draco",
  "n12479821": "Nolina, genus Nolina",
  "n12480004": "bear grass, Nolina microcarpa",
  "n12480233": "Polianthes, genus Polianthes",
  "n12480456": "tuberose, Polianthes tuberosa",
  "n12480677": "genus Sansevieria",
  "n12480895": "sansevieria, bowstring hemp",
  "n12481150": "African bowstring hemp, African hemp, Sansevieria guineensis",
  "n12481289": "Ceylon bowstring hemp, Sansevieria zeylanica",
  "n12481458": "mother-in-law's tongue, snake plant, Sansevieria trifasciata",
  "n12481641": "bowstring hemp",
  "n12481806": "genus Yucca",
  "n12482031": "yucca",
  "n12482437": "Spanish bayonet, Yucca aloifolia",
  "n12482668": "Spanish bayonet, Yucca baccata",
  "n12482893": "Joshua tree, Yucca brevifolia",
  "n12483091": "Spanish dagger, Yucca carnerosana",
  "n12483282": "soapweed, soap-weed, soap tree, Yucca elata",
  "n12483427": "Adam's needle, Adam's needle-and-thread, spoonleaf yucca, needle palm, Yucca filamentosa",
  "n12483625": "bear grass, Yucca glauca",
  "n12483841": "Spanish dagger, Yucca gloriosa",
  "n12484029": "bear grass, Yucca smalliana",
  "n12484244": "Our Lord's candle, Yucca whipplei",
  "n12484413": "Menyanthaceae, family Menyanthaceae, buckbean family",
  "n12484612": "Menyanthes, genus Menyanthes",
  "n12484784": "water shamrock, buckbean, bogbean, bog myrtle, marsh trefoil, Menyanthes trifoliata",
  "n12485122": "Loganiaceae, family Loganiaceae",
  "n12485331": "Logania, genus Logania",
  "n12485523": "genus Buddleia",
  "n12485653": "butterfly bush, buddleia",
  "n12485811": "Gelsemium, genus Gelsemium",
  "n12485981": "yellow jasmine, yellow jessamine, Carolina jasmine, evening trumpet flower, Gelsemium sempervirens",
  "n12486254": "Linaceae, family Linaceae, flax family",
  "n12486397": "Linum, genus Linum",
  "n12486574": "flax",
  "n12486732": "Physostigma, genus Physostigma",
  "n12486882": "calabar-bean vine, Physostigma venenosum",
  "n12487058": "calabar bean, ordeal bean",
  "n12487262": "physostigmine",
  "n12487394": "Caesalpiniaceae, family Caesalpiniaceae",
  "n12487647": "Caesalpinioideae, subfamily Caesalpinioideae",
  "n12488121": "Caesalpinia, genus Caesalpinia",
  "n12488454": "bonduc, bonduc tree, Caesalpinia bonduc, Caesalpinia bonducella",
  "n12488709": "divi-divi, Caesalpinia coriaria",
  "n12488914": "divi-divi",
  "n12489046": "Mysore thorn, Caesalpinia decapetala, Caesalpinia sepiaria",
  "n12489268": "brazilwood, peachwood, peach-wood, pernambuco wood, Caesalpinia echinata",
  "n12489524": "brazilwood",
  "n12489676": "brazilian ironwood, Caesalpinia ferrea",
  "n12489815": "bird of paradise, poinciana, Caesalpinia gilliesii, Poinciana gilliesii",
  "n12490054": "pride of barbados, paradise flower, flamboyant tree, Caesalpinia pulcherrima, Poinciana pulcherrima",
  "n12490330": "Acrocarpus, genus Acrocarpus",
  "n12490490": "shingle tree, Acrocarpus fraxinifolius",
  "n12490671": "Bauhinia, genus Bauhinia",
  "n12490827": "butterfly flower, Bauhinia monandra",
  "n12491017": "mountain ebony, orchid tree, Bauhinia variegata",
  "n12491200": "Brachystegia, genus Brachystegia",
  "n12491435": "msasa, Brachystegia speciformis",
  "n12491626": "genus Cassia, Cassia",
  "n12491826": "cassia",
  "n12492106": "golden shower tree, drumstick tree, purging cassia, pudding pipe tree, canafistola, canafistula, Cassia fistula",
  "n12492460": "pink shower, pink shower tree, horse cassia, Cassia grandis",
  "n12492682": "rainbow shower, Cassia javonica",
  "n12492900": "horse cassia, Cassia roxburghii, Cassia marginata",
  "n12493090": "Ceratonia, genus Ceratonia",
  "n12493208": "carob, carob tree, carob bean tree, algarroba, Ceratonia siliqua",
  "n12493426": "carob, carob bean, algarroba bean, algarroba, locust bean, locust pod",
  "n12493699": "Cercidium, genus Cercidium",
  "n12493868": "paloverde",
  "n12494115": "Chamaecrista, genus Chamaecrista",
  "n12494358": "partridge pea, sensitive pea, wild sensitive plant, Chamaecrista fasciculata, Cassia fasciculata",
  "n12494629": "Delonix, genus Delonix",
  "n12494794": "royal poinciana, flamboyant, flame tree, peacock flower, Delonix regia, Poinciana regia",
  "n12495146": "locust tree, locust",
  "n12495396": "locust",
  "n12495509": "Gleditsia, genus Gleditsia",
  "n12495670": "water locust, swamp locust, Gleditsia aquatica",
  "n12495895": "honey locust, Gleditsia triacanthos",
  "n12496207": "Gymnocladus, genus Gymnocladus",
  "n12496427": "Kentucky coffee tree, bonduc, chicot, Gymnocladus dioica",
  "n12496735": "Haematoxylum, genus Haematoxylum, Haematoxylon, genus Haematoxylon",
  "n12496949": "logwood, logwood tree, campeachy, bloodwood tree, Haematoxylum campechianum",
  "n12497322": "logwood",
  "n12497492": "Parkinsonia, genus Parkinsonia",
  "n12497669": "Jerusalem thorn, horsebean, Parkinsonia aculeata",
  "n12498055": "palo verde, Parkinsonia florida, Cercidium floridum",
  "n12498316": "Petteria, genus Petteria",
  "n12498457": "Dalmatian laburnum, Petteria ramentacea, Cytisus ramentaceus",
  "n12498739": "Poinciana, subgenus Poinciana",
  "n12498928": "genus Senna",
  "n12499163": "senna",
  "n12499439": "ringworm bush, ringworm shrub, ringworm cassia, Senna alata, Cassia alata",
  "n12499757": "avaram, tanner's cassia, Senna auriculata, Cassia auriculata",
  "n12499979": "Alexandria senna, Alexandrian senna, true senna, tinnevelly senna, Indian senna, Senna alexandrina, Cassia acutifolia, Cassia augustifolia",
  "n12500309": "wild senna, Senna marilandica, Cassia marilandica",
  "n12500518": "sicklepod, Senna obtusifolia, Cassia tora",
  "n12500751": "coffee senna, mogdad coffee, styptic weed, stinking weed, Senna occidentalis, Cassia occidentalis",
  "n12501035": "Tamarindus, genus Tamarindus",
  "n12501202": "tamarind, tamarind tree, tamarindo, Tamarindus indica",
  "n12501537": "Papilionaceae, family Papilionacea",
  "n12501745": "Papilionoideae, subfamily Papilionoideae",
  "n12503908": "genus Amorpha",
  "n12504094": "amorpha",
  "n12504306": "leadplant, lead plant, Amorpha canescens",
  "n12504570": "false indigo, bastard indigo, Amorpha californica",
  "n12504783": "false indigo, bastard indigo, Amorpha fruticosa",
  "n12505032": "Amphicarpaea, genus Amphicarpaea, Amphicarpa, genus Amphicarpa",
  "n12505253": "hog peanut, wild peanut, Amphicarpaea bracteata, Amphicarpa bracteata",
  "n12505563": "Anagyris, genus Anagyris",
  "n12505752": "bean trefoil, stinking bean trefoil, Anagyris foetida",
  "n12505987": "Andira, genus Andira",
  "n12506181": "angelim, andelmin",
  "n12506341": "cabbage bark, cabbage-bark tree, cabbage tree, Andira inermis",
  "n12506614": "Anthyllis, genus Anthyllis",
  "n12506784": "Jupiter's beard, silverbush, silver-bush, Anthyllis barba-jovis",
  "n12506991": "kidney vetch, Anthyllis vulneraria",
  "n12507236": "Apios, genus Apios",
  "n12507379": "groundnut, groundnut vine, Indian potato, potato bean, wild bean, Apios americana, Apios tuberosa",
  "n12507670": "Aspalathus, genus Aspalathus",
  "n12507823": "rooibos, Aspalathus linearis, Aspalathus cedcarbergensis",
  "n12508077": "Astragalus, genus Astragalus",
  "n12508309": "milk vetch, milk-vetch",
  "n12508497": "wild licorice, wild liquorice, Astragalus glycyphyllos",
  "n12508618": "alpine milk vetch, Astragalus alpinus",
  "n12508762": "purple milk vetch, Astragalus danicus",
  "n12508936": "Baphia, genus Baphia",
  "n12509109": "camwood, African sandalwood, Baphia nitida",
  "n12509297": "Baptisia, genus Baptisia",
  "n12509476": "wild indigo, false indigo",
  "n12509665": "blue false indigo, Baptisia australis",
  "n12509821": "white false indigo, Baptisia lactea",
  "n12509993": "indigo broom, horsefly weed, rattle weed, Baptisia tinctoria",
  "n12510197": "Butea, genus Butea",
  "n12510343": "dhak, dak, palas, Butea frondosa, Butea monosperma",
  "n12510569": "Cajanus, genus Cajanus",
  "n12510774": "pigeon pea, pigeon-pea plant, cajan pea, catjang pea, red gram, dhal, dahl, Cajanus cajan",
  "n12511046": "Canavalia, genus Canavalia",
  "n12511239": "jack bean, wonder bean, giant stock bean, Canavalia ensiformis",
  "n12511488": "sword bean, Canavalia gladiata",
  "n12511703": "genus Caragana",
  "n12511856": "pea tree, caragana",
  "n12512095": "Siberian pea tree, Caragana arborescens",
  "n12512294": "Chinese pea tree, Caragana sinica",
  "n12512460": "Castanospermum, genus Castanospermum",
  "n12512674": "Moreton Bay chestnut, Australian chestnut",
  "n12512947": "Centrosema, genus Centrosema",
  "n12513172": "butterfly pea, Centrosema virginianum",
  "n12513426": "Cercis, genus Cercis",
  "n12513613": "Judas tree, love tree, Circis siliquastrum",
  "n12513933": "redbud, Cercis canadensis",
  "n12514138": "western redbud, California redbud, Cercis occidentalis",
  "n12514324": "Chamaecytisus, genus Chamaecytisus",
  "n12514592": "tagasaste, Chamaecytisus palmensis, Cytesis proliferus",
  "n12514802": "Chordospartium, genus Chordospartium",
  "n12514992": "weeping tree broom",
  "n12515219": "Chorizema, genus Chorizema",
  "n12515393": "flame pea",
  "n12515597": "Cicer, genus Cicer",
  "n12515711": "chickpea, chickpea plant, Egyptian pea, Cicer arietinum",
  "n12515925": "chickpea, garbanzo",
  "n12516040": "Cladrastis, genus Cladrastis",
  "n12516165": "Kentucky yellowwood, gopherwood, Cladrastis lutea, Cladrastis kentukea",
  "n12516432": "genus Clianthus",
  "n12516584": "glory pea, clianthus",
  "n12516828": "desert pea, Sturt pea, Sturt's desert pea, Clianthus formosus, Clianthus speciosus",
  "n12517077": "parrot's beak, parrot's bill, Clianthus puniceus",
  "n12517253": "Clitoria, genus Clitoria",
  "n12517445": "butterfly pea, Clitoria mariana",
  "n12517642": "blue pea, butterfly pea, Clitoria turnatea",
  "n12517820": "Codariocalyx, genus Codariocalyx",
  "n12518013": "telegraph plant, semaphore plant, Codariocalyx motorius, Desmodium motorium, Desmodium gyrans",
  "n12518305": "Colutea, genus Colutea",
  "n12518481": "bladder senna, Colutea arborescens",
  "n12518725": "genus Coronilla",
  "n12518879": "coronilla",
  "n12519089": "axseed, crown vetch, Coronilla varia",
  "n12519328": "genus Crotalaria",
  "n12519563": "crotalaria, rattlebox",
  "n12519824": "American rattlebox, Crotalaria sagitallis",
  "n12520015": "Indian rattlebox, Crotalaria spectabilis",
  "n12520223": "Cyamopsis, genus Cyamopsis",
  "n12520406": "guar, cluster bean, Cyamopsis tetragonolobus, Cyamopsis psoraloides",
  "n12520661": "Cytisus, genus Cytisus",
  "n12520864": "broom",
  "n12521186": "white broom, white Spanish broom, Cytisus albus, Cytisus multiflorus",
  "n12521394": "common broom, Scotch broom, green broom, Cytisus scoparius",
  "n12521624": "witches' broom, witch broom, hexenbesen, staghead",
  "n12521847": "Dalbergia, genus Dalbergia",
  "n12522188": "rosewood, rosewood tree",
  "n12522493": "rosewood",
  "n12522678": "Indian blackwood, East Indian rosewood, East India rosewood, Indian rosewood, Dalbergia latifolia",
  "n12522894": "sissoo, sissu, sisham, Dalbergia sissoo",
  "n12523141": "kingwood, kingwood tree, Dalbergia cearensis",
  "n12523319": "kingwood",
  "n12523475": "Brazilian rosewood, caviuna wood, jacaranda, Dalbergia nigra",
  "n12523698": "Honduras rosewood, Dalbergia stevensonii",
  "n12523850": "cocobolo, Dalbergia retusa",
  "n12524010": "granadilla wood",
  "n12524188": "blackwood, blackwood tree",
  "n12524395": "blackwood",
  "n12524518": "Dalea, genus Dalea",
  "n12524633": "smoke tree, Dalea spinosa",
  "n12524944": "Daviesia, genus Daviesia",
  "n12525168": "bitter pea",
  "n12525347": "genus Derris",
  "n12525513": "derris",
  "n12525753": "derris root, tuba root, Derris elliptica",
  "n12525975": "Desmanthus, genus Desmanthus",
  "n12526178": "prairie mimosa, prickle-weed, Desmanthus ilinoensis",
  "n12526380": "Desmodium, genus Desmodium",
  "n12526516": "tick trefoil, beggar lice, beggar's lice",
  "n12526754": "beggarweed, Desmodium tortuosum, Desmodium purpureum",
  "n12526946": "Dipogon, genus Dipogon",
  "n12527081": "Australian pea, Dipogon lignosus, Dolichos lignosus",
  "n12527391": "Dolichos, genus Dolichos",
  "n12527569": "genus Erythrina",
  "n12527738": "coral tree, erythrina",
  "n12528109": "kaffir boom, Cape kafferboom, Erythrina caffra",
  "n12528382": "coral bean tree, Erythrina corallodendrum",
  "n12528549": "ceibo, crybaby tree, cry-baby tree, common coral tree, Erythrina crista-galli",
  "n12528768": "kaffir boom, Transvaal kafferboom, Erythrina lysistemon",
  "n12528974": "Indian coral tree, Erythrina variegata, Erythrina Indica",
  "n12529220": "cork tree, Erythrina vespertilio",
  "n12529353": "Galega, genus Galega",
  "n12529500": "goat's rue, goat rue, Galega officinalis",
  "n12529730": "genus Gastrolobium",
  "n12529905": "poison bush, poison pea, gastrolobium",
  "n12530208": "Genista, genus Genista",
  "n12530439": "broom tree, needle furze, petty whin, Genista anglica",
  "n12530629": "Spanish broom, Spanish gorse, Genista hispanica",
  "n12530818": "woodwaxen, dyer's greenweed, dyer's-broom, dyeweed, greenweed, whin, woadwaxen, Genista tinctoria",
  "n12531144": "Geoffroea, genus Geoffroea",
  "n12531328": "chanar, chanal, Geoffroea decorticans",
  "n12531552": "genus Gliricidia",
  "n12531727": "gliricidia",
  "n12532008": "Glycine, genus Glycine",
  "n12532168": "soy, soya, soybean, soya bean, soybean plant, soja, soja bean, Glycine max",
  "n12532564": "soy, soybean, soya bean",
  "n12532720": "Glycyrrhiza, genus Glycyrrhiza",
  "n12532886": "licorice, liquorice, Glycyrrhiza glabra",
  "n12533190": "wild licorice, wild liquorice, American licorice, American liquorice, Glycyrrhiza lepidota",
  "n12533437": "licorice root",
  "n12533588": "Halimodendron, genus Halimodendron",
  "n12533730": "salt tree, Halimodendron halodendron, Halimodendron argenteum",
  "n12533992": "Hardenbergia, genus Hardenbergia",
  "n12534208": "Western Australia coral pea, Hardenbergia comnptoniana",
  "n12534453": "Hedysarum, genus Hedysarum",
  "n12534625": "sweet vetch, Hedysarum boreale",
  "n12534862": "French honeysuckle, sulla, Hedysarum coronarium",
  "n12535101": "Hippocrepis, genus Hippocrepis",
  "n12535254": "horseshoe vetch, Hippocrepis comosa",
  "n12535461": "genus Hovea",
  "n12535593": "hovea, purple pea",
  "n12535820": "Indigofera, genus Indigofera",
  "n12536040": "indigo, indigo plant, Indigofera tinctoria",
  "n12536291": "anil, Indigofera suffruticosa, Indigofera anil",
  "n12536455": "Jacksonia, genus Jacksonia",
  "n12536665": "Kennedia, genus Kennedia, Kennedya, genus Kennedya",
  "n12536871": "coral pea",
  "n12537068": "coral vine, Kennedia coccinea",
  "n12537253": "scarlet runner, running postman, Kennedia prostrata",
  "n12537437": "Lablab, genus Lablab",
  "n12537569": "hyacinth bean, bonavist, Indian bean, Egyptian bean, Lablab purpureus, Dolichos lablab",
  "n12537988": "Laburnum, genus Laburnum",
  "n12538209": "Scotch laburnum, Alpine golden chain, Laburnum alpinum",
  "n12538380": "common laburnum, golden chain, golden rain, Laburnum anagyroides",
  "n12538603": "Lathyrus, genus Lathyrus",
  "n12539074": "vetchling",
  "n12539306": "wild pea",
  "n12539564": "singletary pea, Caley pea, rough pea, wild winterpea, Lathyrus hirsutus",
  "n12539832": "everlasting pea",
  "n12540001": "broad-leaved everlasting pea, perennial pea, Lathyrus latifolius",
  "n12540250": "beach pea, sea pea, Lathyrus maritimus, Lathyrus japonicus",
  "n12540488": "black pea, Lathyrus niger",
  "n12540647": "grass vetch, grass vetchling, Lathyrus nissolia",
  "n12540800": "sweet pea, sweetpea, Lathyrus odoratus",
  "n12540966": "marsh pea, Lathyrus palustris",
  "n12541157": "common vetchling, meadow pea, yellow vetchling, Lathyrus pratensis",
  "n12541403": "grass pea, Indian pea, khesari, Lathyrus sativus",
  "n12541606": "pride of California, Lathyrus splendens",
  "n12541805": "flat pea, narrow-leaved everlasting pea, Lathyrus sylvestris",
  "n12542043": "Tangier pea, Tangier peavine, Lalthyrus tingitanus",
  "n12542240": "heath pea, earth-nut pea, earthnut pea, tuberous vetch, Lathyrus tuberosus",
  "n12542466": "spring vetchling, spring vetch, Lathyrus vernus",
  "n12542649": "genus Lespedeza",
  "n12542910": "bush clover, lespedeza",
  "n12543186": "bicolor lespediza, ezo-yama-hagi, Lespedeza bicolor",
  "n12543455": "japanese clover, japan clover, jap clover, Lespedeza striata",
  "n12543639": "Korean lespedeza, Lespedeza stipulacea",
  "n12543826": "sericea lespedeza, Lespedeza sericea, Lespedeza cuneata",
  "n12544027": "Lens, genus Lens",
  "n12544240": "lentil, lentil plant, Lens culinaris",
  "n12544539": "lentil",
  "n12544646": "Lonchocarpus, genus Lonchocarpus",
  "n12544862": "cube",
  "n12545090": "Lotus, genus Lotus",
  "n12545232": "prairie bird's-foot trefoil, compass plant, prairie lotus, prairie trefoil, Lotus americanus",
  "n12545429": "coral gem, Lotus berthelotii",
  "n12545635": "bird's foot trefoil, bird's foot clover, babies' slippers, bacon and eggs, Lotus corniculatus",
  "n12545865": "winged pea, asparagus pea, Lotus tetragonolobus",
  "n12546015": "Lupinus, genus Lupinus",
  "n12546183": "lupine, lupin",
  "n12546420": "white lupine, field lupine, wolf bean, Egyptian lupine, Lupinus albus",
  "n12546617": "tree lupine, Lupinus arboreus",
  "n12546832": "yellow lupine, Lupinus luteus",
  "n12546962": "wild lupine, sundial lupine, Indian beet, old-maid's bonnet, Lupinus perennis",
  "n12547215": "bluebonnet, buffalo clover, Texas bluebonnet, Lupinus subcarnosus",
  "n12547503": "Texas bluebonnet, Lupinus texensis",
  "n12547658": "Macrotyloma, genus Macrotyloma",
  "n12547872": "horse gram, horse grain, poor man's pulse, Macrotyloma uniflorum, Dolichos biflorus",
  "n12548134": "Medicago, genus Medicago",
  "n12548280": "medic, medick, trefoil",
  "n12548564": "moon trefoil, Medicago arborea",
  "n12548804": "sickle alfalfa, sickle lucerne, sickle medick, Medicago falcata",
  "n12549005": "Calvary clover, Medicago intertexta, Medicago echinus",
  "n12549192": "black medick, hop clover, yellow trefoil, nonesuch clover, Medicago lupulina",
  "n12549420": "alfalfa, lucerne, Medicago sativa",
  "n12549649": "genus Millettia",
  "n12549799": "millettia",
  "n12549976": "genus Mucuna, Stizolobium, genus Stizolobium",
  "n12550210": "mucuna",
  "n12550408": "cowage, velvet bean, Bengal bean, Benghal bean, Florida bean, Mucuna pruriens utilis, Mucuna deeringiana, Mucuna aterrima, Stizolobium deeringiana",
  "n12550788": "cowage",
  "n12550968": "Myroxylon, genus Myroxylon",
  "n12551173": "tolu tree, tolu balsam tree, Myroxylon balsamum, Myroxylon toluiferum",
  "n12551457": "Peruvian balsam, Myroxylon pereirae, Myroxylon balsamum pereirae",
  "n12551669": "tolu, balsam of tolu, tolu balsam",
  "n12551877": "balsam of Peru",
  "n12552081": "Onobrychis, genus Onobrychis",
  "n12552309": "sainfoin, sanfoin, holy clover, esparcet, Onobrychis viciifolia, Onobrychis viciaefolia",
  "n12552658": "Ononis, genus Ononis",
  "n12552893": "restharrow, rest-harrow, Ononis repens",
  "n12553114": "restharrow, rest-harrow, Ononis spinosa",
  "n12553314": "Ormosia, genus Ormosia",
  "n12553573": "necklace tree",
  "n12553742": "bead tree, jumby bean, jumby tree, Ormosia monosperma",
  "n12554029": "jumby bead, jumbie bead, Ormosia coarctata",
  "n12554242": "Oxytropis, genus Oxytropis",
  "n12554526": "locoweed, crazyweed, crazy weed",
  "n12554729": "purple locoweed, purple loco, Oxytropis lambertii",
  "n12554911": "tumbleweed",
  "n12555069": "Pachyrhizus, genus Pachyrhizus",
  "n12555255": "yam bean, Pachyrhizus erosus",
  "n12555553": "yam bean, potato bean, Pachyrhizus tuberosus",
  "n12555720": "Parochetus, genus Parochetus",
  "n12555859": "shamrock pea, Parochetus communis",
  "n12556030": "Phaseolus, genus Phaseolus",
  "n12556307": "bean, bean plant",
  "n12556533": "bush bean",
  "n12556656": "pole bean",
  "n12556793": "common bean, common bean plant, Phaseolus vulgaris",
  "n12557064": "kidney bean, frijol, frijole",
  "n12557280": "green bean",
  "n12557438": "haricot",
  "n12557556": "wax bean",
  "n12557681": "scarlet runner, scarlet runner bean, Dutch case-knife bean, runner bean, Phaseolus coccineus, Phaseolus multiflorus",
  "n12557995": "shell bean, shell bean plant",
  "n12558230": "lima bean, lima bean plant, Phaseolus limensis",
  "n12558425": "sieva bean, butter bean, butter-bean plant, lima bean, Phaseolus lunatus",
  "n12558680": "tepary bean, Phaseolus acutifolius latifolius",
  "n12558902": "Pickeringia, genus Pickeringia",
  "n12559044": "chaparral pea, stingaree-bush, Pickeringia montana",
  "n12559302": "Piscidia, genus Piscidia",
  "n12559518": "Jamaica dogwood, fish fuddle, Piscidia piscipula, Piscidia erythrina",
  "n12559842": "Pisum, genus Pisum",
  "n12560016": "pea, pea plant",
  "n12560282": "pea",
  "n12560420": "garden pea, garden pea plant, common pea, Pisum sativum",
  "n12560621": "garden pea",
  "n12560775": "edible-pod pea, edible-podded pea, Pisum sativum macrocarpon",
  "n12561038": "snow pea, sugar pea",
  "n12561169": "sugar snap pea, snap pea",
  "n12561309": "field pea, field-pea plant, Austrian winter pea, Pisum sativum arvense, Pisum arvense",
  "n12561594": "field pea",
  "n12561696": "Platylobium, genus Platylobium",
  "n12561897": "flat pea, Platylobium formosum",
  "n12562141": "common flat pea, native holly, Playlobium obtusangulum",
  "n12562420": "Platymiscium, genus Platymiscium",
  "n12562577": "quira",
  "n12562785": "roble, Platymiscium trinitatis",
  "n12563045": "Panama redwood tree, Panama redwood, Platymiscium pinnatum",
  "n12563281": "Panama redwood, quira",
  "n12563404": "Podalyria, genus Podalyria",
  "n12563567": "Pongamia, genus Pongamia",
  "n12563702": "Indian beech, Pongamia glabra",
  "n12563913": "Psophocarpus, genus Psophocarpus",
  "n12564083": "winged bean, winged pea, goa bean, goa bean vine, Manila bean, Psophocarpus tetragonolobus",
  "n12564381": "Psoralea, genus Psoralea",
  "n12564613": "breadroot, Indian breadroot, pomme blanche, pomme de prairie, Psoralea esculenta",
  "n12564840": "Pterocarpus, genus Pterocarpus",
  "n12565102": "bloodwood tree, kiaat, Pterocarpus angolensis",
  "n12565368": "padauk, padouk, amboyna, Pterocarpus indicus",
  "n12565590": "amboyna, Andaman redwood",
  "n12565730": "Burma padauk, Burmese rosewood, Pterocarpus macrocarpus",
  "n12565912": "kino, Pterocarpus marsupium",
  "n12566112": "East India kino, Malabar kino, kino gum",
  "n12566331": "red sandalwood, red sanders, red sanderswood, red saunders, Pterocarpus santalinus",
  "n12566627": "ruby wood, red sandalwood",
  "n12566809": "Pueraria, genus Pueraria",
  "n12566954": "kudzu, kudzu vine, Pueraria lobata",
  "n12567316": "Retama, genus Retama",
  "n12567490": "retem, raetam, juniper bush, juniper, Retama raetam, Genista raetam",
  "n12567768": "Robinia, genus Robinia",
  "n12567950": "bristly locust, rose acacia, moss locust, Robinia hispida",
  "n12568186": "black locust, yellow locust, Robinia pseudoacacia",
  "n12568506": "black locust",
  "n12568649": "clammy locust, Robinia viscosa",
  "n12568865": "Sabinea, genus Sabinea",
  "n12569037": "carib wood, Sabinea carinalis",
  "n12569233": "genus Sesbania",
  "n12569426": "sesbania",
  "n12569616": "Colorado River hemp, Sesbania exaltata",
  "n12569851": "scarlet wisteria tree, vegetable hummingbird, Sesbania grandiflora",
  "n12570126": "Sophora, genus Sophora",
  "n12570394": "Japanese pagoda tree, Chinese scholartree, Chinese scholar tree, Sophora japonica, Sophora sinensis",
  "n12570703": "mescal bean, coral bean, frijolito, frijolillo, Sophora secundiflora",
  "n12570972": "kowhai, Sophora tetraptera",
  "n12571194": "Spartium, genus Spartium",
  "n12571330": "Spanish broom, weaver's broom, Spartium junceum",
  "n12571606": "Strongylodon, genus Strongylodon",
  "n12571781": "jade vine, emerald creeper, Strongylodon macrobotrys",
  "n12572021": "Templetonia, genus Templetonia",
  "n12572188": "coral bush, flame bush, Templetonia retusa",
  "n12572373": "Tephrosia, genus Tephrosia",
  "n12572546": "hoary pea",
  "n12572759": "bastard indigo, Tephrosia purpurea",
  "n12572858": "catgut, goat's rue, wild sweet pea, Tephrosia virginiana",
  "n12573078": "Thermopsis, genus Thermopsis",
  "n12573256": "bush pea",
  "n12573474": "false lupine, golden pea, yellow pea, Thermopsis macrophylla",
  "n12573647": "Carolina lupine, Thermopsis villosa",
  "n12573760": "Tipuana, genus Tipuana",
  "n12573911": "tipu, tipu tree, yellow jacaranda, pride of Bolivia",
  "n12574143": "Trigonella, genus Trigonella",
  "n12574320": "bird's foot trefoil, Trigonella ornithopodioides",
  "n12574470": "fenugreek, Greek clover, Trigonella foenumgraecum",
  "n12574727": "Ulex, genus Ulex",
  "n12574866": "gorse, furze, whin, Irish gorse, Ulex europaeus",
  "n12575089": "Vicia, genus Vicia",
  "n12575322": "vetch",
  "n12575679": "tare",
  "n12575812": "tufted vetch, bird vetch, Calnada pea, Vicia cracca",
  "n12576029": "broad bean, broad-bean, broad-bean plant, English bean, European bean, field bean, Vicia faba",
  "n12576323": "broad bean, fava bean, horsebean",
  "n12576451": "bitter betch, Vicia orobus",
  "n12576555": "spring vetch, Vicia sativa",
  "n12576695": "bush vetch, Vicia sepium",
  "n12576838": "hairy vetch, hairy tare, Vicia villosa",
  "n12577000": "Vigna, genus Vigna",
  "n12577362": "moth bean, Vigna aconitifolia, Phaseolus aconitifolius",
  "n12577686": "adzuki bean, adsuki bean, Vigna angularis, Phaseolus angularis",
  "n12577895": "snailflower, snail-flower, snail flower, snail bean, corkscrew flower, Vigna caracalla, Phaseolus caracalla",
  "n12578255": "mung, mung bean, green gram, golden gram, Vigna radiata, Phaseolus aureus",
  "n12578626": "cowpea, cowpea plant, black-eyed pea, Vigna unguiculata, Vigna sinensis",
  "n12578916": "cowpea, black-eyed pea",
  "n12579038": "asparagus bean, yard-long bean, Vigna unguiculata sesquipedalis, Vigna sesquipedalis",
  "n12579242": "Viminaria, genus Viminaria",
  "n12579404": "swamp oak, Viminaria juncea, Viminaria denudata",
  "n12579593": "Virgilia, genus Virgilia",
  "n12579822": "keurboom, Virgilia capensis, Virgilia oroboides",
  "n12580012": "keurboom, Virgilia divaricata",
  "n12580204": "genus Wisteria",
  "n12580457": "wisteria, wistaria",
  "n12580654": "Japanese wistaria, Wisteria floribunda",
  "n12580786": "Chinese wistaria, Wisteria chinensis",
  "n12580896": "American wistaria, American wisteria, Wisteria frutescens",
  "n12581110": "silky wisteria, Wisteria venusta",
  "n12581230": "Palmales, order Palmales",
  "n12581381": "Palmae, family Palmae, Palmaceae, family Palmaceae, Arecaceae, family Arecaceae, palm family",
  "n12582231": "palm, palm tree",
  "n12582665": "sago palm",
  "n12582846": "feather palm",
  "n12583126": "fan palm",
  "n12583401": "palmetto",
  "n12583529": "Acrocomia, genus Acrocomia",
  "n12583681": "coyol, coyol palm, Acrocomia vinifera",
  "n12583855": "grugru, gri-gri, grugru palm, macamba, Acrocomia aculeata",
  "n12584057": "genus Areca",
  "n12584191": "areca",
  "n12584365": "betel palm, Areca catechu",
  "n12584559": "Arenga, genus Arenga",
  "n12584715": "sugar palm, gomuti, gomuti palm, Arenga pinnata",
  "n12584970": "Attalea, genus Attalea",
  "n12585137": "piassava palm, pissaba palm, Bahia piassava, bahia coquilla, Attalea funifera",
  "n12585373": "coquilla nut",
  "n12585512": "Borassus, genus Borassus",
  "n12585629": "palmyra, palmyra palm, toddy palm, wine palm, lontar, longar palm, Borassus flabellifer",
  "n12585967": "bassine",
  "n12586110": "genus Calamus",
  "n12586298": "calamus",
  "n12586499": "rattan, rattan palm, Calamus rotang",
  "n12586725": "lawyer cane, Calamus australis",
  "n12586867": "Caryota, genus Caryota",
  "n12586989": "fishtail palm",
  "n12587132": "wine palm, jaggery palm, kitul, kittul, kitul tree, toddy palm, Caryota urens",
  "n12587366": "Ceroxylon, genus Ceroxylon",
  "n12587487": "wax palm, Ceroxylon andicola, Ceroxylon alpinum",
  "n12587686": "Cocos, genus Cocos",
  "n12587803": "coconut, coconut palm, coco palm, coco, cocoa palm, coconut tree, Cocos nucifera",
  "n12588054": "coir",
  "n12588156": "Copernicia, genus Copernicia",
  "n12588320": "carnauba, carnauba palm, wax palm, Copernicia prunifera, Copernicia cerifera",
  "n12588584": "carnauba wax, carnauba",
  "n12588780": "caranday, caranda, caranda palm, wax palm, Copernicia australis, Copernicia alba",
  "n12588989": "genus Corozo",
  "n12589142": "corozo, corozo palm",
  "n12589286": "Corypha, genus Corypha",
  "n12589458": "gebang palm, Corypha utan, Corypha gebanga",
  "n12589687": "latanier, latanier palm",
  "n12589841": "talipot, talipot palm, Corypha umbraculifera",
  "n12590117": "Elaeis, genus Elaeis",
  "n12590232": "oil palm",
  "n12590499": "African oil palm, Elaeis guineensis",
  "n12590600": "American oil palm, Elaeis oleifera",
  "n12590715": "palm nut, palm kernel",
  "n12590842": "Euterpe, genus Euterpe",
  "n12591017": "cabbage palm, Euterpe oleracea",
  "n12591195": "Livistona, genus Livistona",
  "n12591351": "cabbage palm, cabbage tree, Livistona australis",
  "n12591523": "Metroxylon, genus Metroxylon",
  "n12591702": "true sago palm, Metroxylon sagu",
  "n12591897": "Nipa, genus Nipa, Nypa, genus Nypa",
  "n12592058": "nipa palm, Nipa fruticans",
  "n12592351": "Orbignya, genus Orbignya",
  "n12592544": "babassu, babassu palm, coco de macao, Orbignya phalerata, Orbignya spesiosa, Orbignya martiana",
  "n12592839": "babassu nut",
  "n12592971": "babassu oil, babacu oil",
  "n12593122": "cohune palm, Orbignya cohune, cohune",
  "n12593341": "cohune nut",
  "n12593508": "cohune-nut oil, cohune oil, cohune fat",
  "n12593689": "phoenicophorium, genus Phoenicophorium",
  "n12593826": "phoenix, genus Phoenix",
  "n12593994": "date palm, Phoenix dactylifera",
  "n12594165": "phytelephas, genus Phytelephas",
  "n12594324": "ivory palm, ivory-nut palm, ivory plant, Phytelephas macrocarpa",
  "n12594533": "ivory nut, vegetable ivory, apple nut",
  "n12594746": "Raffia, genus Raffia, Raphia, genus Raphia",
  "n12594989": "raffia palm, Raffia farinifera, Raffia ruffia",
  "n12595305": "raffia",
  "n12595452": "jupati, jupaty, jupati palm, Raffia taedigera",
  "n12595699": "bamboo palm, Raffia vinifera",
  "n12595801": "Rhapis, genus Rhapis",
  "n12595964": "lady palm",
  "n12596148": "miniature fan palm, bamboo palm, fern rhapis, Rhapis excelsa",
  "n12596345": "reed rhapis, slender lady palm, Rhapis humilis",
  "n12596525": "Roystonea, genus Roystonea",
  "n12596709": "royal palm, Roystonea regia",
  "n12596849": "cabbage palm, Roystonea oleracea",
  "n12597006": "Sabal, genus Sabal",
  "n12597134": "cabbage palmetto, cabbage palm, Sabal palmetto",
  "n12597333": "Serenoa, genus Serenoa",
  "n12597466": "saw palmetto, scrub palmetto, Serenoa repens",
  "n12597640": "Thrinax, genus Thrinax",
  "n12597798": "thatch palm, thatch tree, silver thatch, broom palm, Thrinax parviflora",
  "n12598027": "key palm, silvertop palmetto, silver thatch, Thrinax microcarpa, Thrinax morrisii, Thrinax keyensis",
  "n12598247": "Plantaginales, order Plantaginales",
  "n12598409": "Plantaginaceae, family Plantaginaceae, plantain family",
  "n12598629": "Plantago, genus Plantago",
  "n12598826": "plantain",
  "n12599185": "English plantain, narrow-leaved plantain, ribgrass, ribwort, ripple-grass, buckthorn, Plantago lanceolata",
  "n12599435": "broad-leaved plantain, common plantain, white-man's foot, whiteman's foot, cart-track plant, Plantago major",
  "n12599661": "hoary plantain, Plantago media",
  "n12599874": "fleawort, psyllium, Spanish psyllium, Plantago psyllium",
  "n12600095": "rugel's plantain, broad-leaved plantain, Plantago rugelii",
  "n12600267": "hoary plantain, Plantago virginica",
  "n12600417": "Polygonales, order Polygonales",
  "n12600574": "Polygonaceae, family Polygonaceae, buckwheat family",
  "n12600888": "Polygonum, genus Polygonum",
  "n12601106": "silver lace vine, China fleece vine, Russian vine, Polygonum aubertii",
  "n12601335": "Fagopyrum, genus Fagopyrum",
  "n12601494": "buckwheat, Polygonum fagopyrum, Fagopyrum esculentum",
  "n12601805": "prince's-feather, princess feather, kiss-me-over-the-garden-gate, prince's-plume, Polygonum orientale",
  "n12602118": "genus Eriogonum",
  "n12602262": "eriogonum",
  "n12602434": "umbrella plant, Eriogonum allenii",
  "n12602612": "wild buckwheat, California buckwheat, Erigonum fasciculatum",
  "n12602850": "Rheum, genus Rheum",
  "n12602980": "rhubarb, rhubarb plant",
  "n12603273": "Himalayan rhubarb, Indian rhubarb, red-veined pie plant, Rheum australe, Rheum emodi",
  "n12603449": "pie plant, garden rhubarb, Rheum cultorum, Rheum rhabarbarum, Rheum rhaponticum",
  "n12603672": "Chinese rhubarb, Rheum palmatum",
  "n12603784": "Rumex, genus Rumex",
  "n12603959": "dock, sorrel, sour grass",
  "n12604228": "sour dock, garden sorrel, Rumex acetosa",
  "n12604460": "sheep sorrel, sheep's sorrel, Rumex acetosella",
  "n12604639": "bitter dock, broad-leaved dock, yellow dock, Rumex obtusifolius",
  "n12604845": "French sorrel, garden sorrel, Rumex scutatus",
  "n12605019": "Xyridales, order Xyridales, Commelinales, order Commelinales",
  "n12605315": "Xyridaceae, family Xyridaceae, yellow-eyed grass family",
  "n12605519": "Xyris, genus Xyris",
  "n12605683": "yellow-eyed grass",
  "n12605872": "tall yellow-eye, Xyris operculata",
  "n12605965": "Commelinaceae, family Commelinaceae, spiderwort family",
  "n12606227": "genus Commelina",
  "n12606438": "commelina",
  "n12606545": "spiderwort, dayflower",
  "n12606688": "St.-Bruno's-lily, Paradisea liliastrum",
  "n12606797": "Tradescantia, genus Tradescantia",
  "n12606907": "Bromeliaceae, family Bromeliaceae, pineapple family",
  "n12607198": "Ananas, genus Ananas",
  "n12607456": "pineapple, pineapple plant, Ananas comosus",
  "n12607717": "Bromelia",
  "n12607896": "Tillandsia, genus Tillandsia",
  "n12608127": "Spanish moss, old man's beard, black moss, long moss, Tillandsia usneoides",
  "n12608447": "Mayacaceae, family Mayacaceae",
  "n12608620": "Mayaca, genus Mayaca",
  "n12608778": "Rapateaceae, family Rapateaceae",
  "n12608941": "Eriocaulaceae, family Eriocaulaceae, pipewort family",
  "n12609128": "Eriocaulon, genus Eriocaulon",
  "n12609379": "pipewort, Eriocaulon aquaticum",
  "n12609638": "Pontederiaceae, family Pontederiaceae, pickerelweed family",
  "n12609842": "Pontederia, genus Pontederia",
  "n12609968": "pickerelweed, pickerel weed, wampee, Pontederia cordata",
  "n12610186": "Eichhornia, genus Eichhornia",
  "n12610328": "water hyacinth, water orchid, Eichhornia crassipes, Eichhornia spesiosa",
  "n12610609": "Heteranthera, genus Heteranthera",
  "n12610740": "water star grass, mud plantain, Heteranthera dubia",
  "n12610933": "Naiadales, order Naiadales, Alismales, order Alismales",
  "n12611243": "Naiadaceae, family Naiadaceae, Najadaceae, family Najadaceae, naiad family",
  "n12611479": "Naias, genus Naias, Najas, genus Najas",
  "n12611640": "naiad, water nymph",
  "n12611815": "Alismataceae, family Alismataceae, water-plantain family",
  "n12612020": "Alisma, genus Alisma",
  "n12612170": "water plantain, Alisma plantago-aquatica",
  "n12612410": "Sagittaria, genus Sagittaria",
  "n12612640": "common arrowhead",
  "n12612709": "ribbon-leaved water plantain",
  "n12612811": "narrow-leaved water plantain",
  "n12612913": "Hydrocharitaceae, family Hydrocharitaceae, Hydrocharidaceae, family Hydrocharidaceae, frogbit family, frog's-bit family",
  "n12613285": "Hydrocharis, genus Hydrocharis",
  "n12613408": "frogbit, frog's-bit, Hydrocharis morsus-ranae",
  "n12613596": "genus Hydrilla",
  "n12613706": "hydrilla, Hydrilla verticillata",
  "n12613968": "Limnobium, genus Limnobium",
  "n12614096": "American frogbit, Limnodium spongia",
  "n12614317": "Elodea, genus Elodea, pondweed, ditchmoss",
  "n12614477": "waterweed",
  "n12614625": "Canadian pondweed, Elodea canadensis",
  "n12614763": "dense-leaved elodea, Elodea densa, Egeria densa",
  "n12614962": "Egeria, genus Egeria",
  "n12615097": "Vallisneria, genus Vallisneria",
  "n12615232": "tape grass, eelgrass, wild celery, Vallisneria spiralis",
  "n12615427": "Potamogetonaceae, family Potamogetonaceae, pondweed family",
  "n12615710": "pondweed",
  "n12615986": "Potamogeton, genus Potamogeton",
  "n12616248": "curled leaf pondweed, curly pondweed, Potamogeton crispus",
  "n12616442": "variously-leaved pondweed, Potamogeton gramineous",
  "n12616630": "loddon pondweed, Potamogeton nodosus, Potamogeton americanus",
  "n12616825": "Groenlandia, genus Groenlandia",
  "n12616996": "frog's lettuce",
  "n12617140": "Scheuchzeriaceae, family Scheuchzeriaceae, Juncaginaceae, family Juncaginaceae, arrow-grass family",
  "n12617384": "Triglochin, genus Triglochin",
  "n12617559": "arrow grass, Triglochin maritima",
  "n12617739": "Zannichelliaceae, family Zannichelliaceae",
  "n12617950": "Zannichellia, genus Zannichellia",
  "n12618146": "horned pondweed, Zannichellia palustris",
  "n12618336": "Zosteraceae, family Zosteraceae, eelgrass family",
  "n12618524": "Zostera, genus Zostera",
  "n12618727": "eelgrass, grass wrack, sea wrack, Zostera marina",
  "n12618942": "Rosales, order Rosales",
  "n12619306": "Rosaceae, family Rosaceae, rose family",
  "n12620031": "Rosa, genus Rosa",
  "n12620196": "rose, rosebush",
  "n12620546": "hip, rose hip, rosehip",
  "n12620661": "mountain rose, Rosa pendulina",
  "n12620779": "ground rose, Rosa spithamaea",
  "n12620969": "banksia rose, Rosa banksia",
  "n12621110": "dog rose, Rosa canina",
  "n12621260": "China rose, Bengal rose, Rosa chinensis",
  "n12621410": "damask rose, summer damask rose, Rosa damascena",
  "n12621619": "sweetbrier, sweetbriar, brier, briar, eglantine, Rosa eglanteria",
  "n12621836": "brier, brierpatch, brier patch",
  "n12621945": "Cherokee rose, Rosa laevigata",
  "n12622072": "multiflora, multiflora rose, Japanese rose, baby rose, Rosa multiflora",
  "n12622297": "musk rose, Rosa moschata",
  "n12622483": "tea rose, Rosa odorata",
  "n12622653": "genus Agrimonia",
  "n12622875": "agrimonia, agrimony",
  "n12623077": "harvest-lice, Agrimonia eupatoria",
  "n12623211": "fragrant agrimony, Agrimonia procera",
  "n12623368": "Amelanchier, genus Amelanchier",
  "n12623524": "Juneberry, serviceberry, service tree, shadbush, shadblow",
  "n12623818": "alderleaf Juneberry, alder-leaved serviceberry, Amelanchier alnifolia",
  "n12624055": "Bartram Juneberry, Amelanchier bartramiana",
  "n12624249": "Chaenomeles, genus Chaenomeles",
  "n12624381": "flowering quince",
  "n12624568": "japonica, maule's quince, Chaenomeles japonica",
  "n12624721": "Japanese quince, Chaenomeles speciosa",
  "n12624873": "Chrysobalanus, genus Chrysobalanus",
  "n12625003": "coco plum, coco plum tree, cocoa plum, icaco, Chrysobalanus icaco",
  "n12625215": "genus Cotoneaster",
  "n12625383": "cotoneaster",
  "n12625670": "Cotoneaster dammeri",
  "n12625823": "Cotoneaster horizontalis",
  "n12626030": "Crataegus, genus Crataegus",
  "n12626353": "hawthorn, haw",
  "n12626674": "parsley haw, parsley-leaved thorn, Crataegus apiifolia, Crataegus marshallii",
  "n12626878": "scarlet haw, Crataegus biltmoreana",
  "n12627119": "blackthorn, pear haw, pear hawthorn, Crataegus calpodendron, Crataegus tomentosa",
  "n12627347": "cockspur thorn, cockspur hawthorn, Crataegus crus-galli",
  "n12627526": "mayhaw, summer haw, Crataegus aestivalis",
  "n12627750": "whitethorn, English hawthorn, may, Crataegus laevigata, Crataegus oxycantha",
  "n12628060": "English hawthorn, Crataegus monogyna",
  "n12628356": "red haw, downy haw, Crataegus mollis, Crataegus coccinea mollis",
  "n12628579": "evergreen thorn, Crataegus oxyacantha",
  "n12628705": "red haw, Crataegus pedicellata, Crataegus coccinea",
  "n12628872": "Cydonia, genus Cydonia",
  "n12628986": "quince, quince bush, Cydonia oblonga",
  "n12629187": "Dryas, genus Dryas",
  "n12629305": "mountain avens, Dryas octopetala",
  "n12629523": "Eriobotrya, genus Eriobotrya",
  "n12629666": "loquat, loquat tree, Japanese medlar, Japanese plum, Eriobotrya japonica",
  "n12629946": "Fragaria, genus Fragaria",
  "n12630144": "strawberry",
  "n12630478": "garden strawberry, cultivated strawberry, Fragaria ananassa",
  "n12630641": "wild strawberry, wood strawberry, Fragaria vesca",
  "n12630763": "beach strawberry, Chilean strawberry, Fragaria chiloensis",
  "n12630999": "Virginia strawberry, scarlet strawberry, Fragaria virginiana",
  "n12631224": "Geum, genus Geum",
  "n12631331": "avens",
  "n12631637": "yellow avens, Geum alleppicum strictum, Geum strictum",
  "n12631813": "bennet, white avens, Geum canadense",
  "n12631932": "yellow avens, Geum macrophyllum",
  "n12632072": "water avens, Indian chocolate, purple avens, chocolate root, Geum rivale",
  "n12632335": "prairie smoke, purple avens, Geum triflorum",
  "n12632526": "herb bennet, cloveroot, clover-root, wood avens, Geum urbanum",
  "n12632733": "bennet, white avens, Geum virginianum",
  "n12632875": "Heteromeles, genus Heteromeles",
  "n12633061": "toyon, tollon, Christmasberry, Christmas berry, Heteromeles arbutifolia, Photinia arbutifolia",
  "n12633386": "Malus, genus Malus",
  "n12633638": "apple tree",
  "n12633866": "applewood",
  "n12633994": "apple, orchard apple tree, Malus pumila",
  "n12634211": "wild apple, crab apple, crabapple",
  "n12634429": "crab apple, crabapple, cultivated crab apple",
  "n12634734": "Siberian crab, Siberian crab apple, cherry apple, cherry crab, Malus baccata",
  "n12634986": "wild crab, Malus sylvestris",
  "n12635151": "American crab apple, garland crab, Malus coronaria",
  "n12635359": "Oregon crab apple, Malus fusca",
  "n12635532": "Southern crab apple, flowering crab, Malus angustifolia",
  "n12635744": "Iowa crab, Iowa crab apple, prairie crab, western crab apple, Malus ioensis",
  "n12635955": "Bechtel crab, flowering crab",
  "n12636107": "Mespilus, genus Mespilus",
  "n12636224": "medlar, medlar tree, Mespilus germanica",
  "n12636430": "Photinia, genus Photinia",
  "n12636705": "Potentilla, genus Potentilla",
  "n12636885": "cinquefoil, five-finger",
  "n12637123": "silverweed, goose-tansy, goose grass, Potentilla anserina",
  "n12637319": "Poterium, genus Poterium",
  "n12637485": "salad burnet, burnet bloodwort, pimpernel, Poterium sanguisorba",
  "n12637729": "Prunus, genus Prunus",
  "n12638218": "plum, plum tree",
  "n12638556": "wild plum, wild plum tree",
  "n12638753": "Allegheny plum, Alleghany plum, sloe, Prunus alleghaniensis",
  "n12638964": "American red plum, August plum, goose plum, Prunus americana",
  "n12639168": "chickasaw plum, hog plum, hog plum bush, Prunus angustifolia",
  "n12639376": "beach plum, beach plum bush, Prunus maritima",
  "n12639584": "common plum, Prunus domestica",
  "n12639736": "bullace, Prunus insititia",
  "n12639910": "damson plum, damson plum tree, Prunus domestica insititia",
  "n12640081": "big-tree plum, Prunus mexicana",
  "n12640284": "Canada plum, Prunus nigra",
  "n12640435": "plumcot, plumcot tree",
  "n12640607": "apricot, apricot tree",
  "n12640839": "Japanese apricot, mei, Prunus mume",
  "n12641007": "common apricot, Prunus armeniaca",
  "n12641180": "purple apricot, black apricot, Prunus dasycarpa",
  "n12641413": "cherry, cherry tree",
  "n12641796": "cherry",
  "n12641931": "wild cherry, wild cherry tree",
  "n12642090": "wild cherry",
  "n12642200": "sweet cherry, Prunus avium",
  "n12642435": "heart cherry, oxheart, oxheart cherry",
  "n12642600": "gean, mazzard, mazzard cherry",
  "n12642734": "Western sand cherry, Rocky Mountains cherry, Prunus besseyi",
  "n12642964": "capulin, capulin tree, Prunus capuli",
  "n12643113": "cherry laurel, laurel cherry, mock orange, wild orange, Prunus caroliniana",
  "n12643313": "cherry plum, myrobalan, myrobalan plum, Prunus cerasifera",
  "n12643473": "sour cherry, sour cherry tree, Prunus cerasus",
  "n12643688": "amarelle, Prunus cerasus caproniana",
  "n12643877": "morello, Prunus cerasus austera",
  "n12644057": "marasca, marasca cherry, maraschino cherry, Prunus cerasus marasca",
  "n12644283": "marasca",
  "n12644464": "Amygdalaceae, family Amygdalaceae",
  "n12644713": "Amygdalus, genus Amygdalus",
  "n12644902": "almond tree",
  "n12645174": "almond, sweet almond, Prunus dulcis, Prunus amygdalus, Amygdalus communis",
  "n12645530": "bitter almond, Prunus dulcis amara, Amygdalus communis amara",
  "n12645754": "almond oil, expressed almond oil, sweet almond oil",
  "n12645913": "bitter almond oil",
  "n12646072": "jordan almond",
  "n12646197": "dwarf flowering almond, Prunus glandulosa",
  "n12646397": "holly-leaved cherry, holly-leaf cherry, evergreen cherry, islay, Prunus ilicifolia",
  "n12646605": "fuji, fuji cherry, Prunus incisa",
  "n12646740": "flowering almond, oriental bush cherry, Prunus japonica",
  "n12646950": "cherry laurel, laurel cherry, Prunus laurocerasus",
  "n12647231": "Catalina cherry, Prunus lyonii",
  "n12647376": "bird cherry, bird cherry tree",
  "n12647560": "hagberry tree, European bird cherry, common bird cherry, Prunus padus",
  "n12647787": "hagberry",
  "n12647893": "pin cherry, Prunus pensylvanica",
  "n12648045": "peach, peach tree, Prunus persica",
  "n12648196": "nectarine, nectarine tree, Prunus persica nectarina",
  "n12648424": "sand cherry, Prunus pumila, Prunus pumilla susquehanae, Prunus susquehanae, Prunus cuneata",
  "n12648693": "Japanese plum, Prunus salicina",
  "n12648888": "black cherry, black cherry tree, rum cherry, Prunus serotina",
  "n12649065": "flowering cherry",
  "n12649317": "oriental cherry, Japanese cherry, Japanese flowering cherry, Prunus serrulata",
  "n12649539": "Japanese flowering cherry, Prunus sieboldii",
  "n12649723": "blackthorn, sloe, Prunus spinosa",
  "n12649866": "Sierra plum, Pacific plum, Prunus subcordata",
  "n12650038": "rosebud cherry, winter flowering cherry, Prunus subhirtella",
  "n12650229": "Russian almond, dwarf Russian almond, Prunus tenella",
  "n12650379": "flowering almond, Prunus triloba",
  "n12650556": "chokecherry, chokecherry tree, Prunus virginiana",
  "n12650805": "chokecherry",
  "n12650915": "western chokecherry, Prunus virginiana demissa, Prunus demissa",
  "n12651062": "genus Pyracantha",
  "n12651229": "Pyracantha, pyracanth, fire thorn, firethorn",
  "n12651465": "Pyrus, genus Pyrus",
  "n12651611": "pear, pear tree, Pyrus communis",
  "n12651821": "fruit tree",
  "n12652876": "fruitwood",
  "n12653056": "Rubus, genus Rubus",
  "n12653218": "bramble bush",
  "n12653436": "lawyerbush, lawyer bush, bush lawyer, Rubus cissoides, Rubus australis",
  "n12653633": "stone bramble, Rubus saxatilis",
  "n12653762": "blackberry, blackberry bush",
  "n12654012": "true blackberry, Rubus fruticosus",
  "n12654227": "sand blackberry, Rubus cuneifolius",
  "n12654387": "dewberry, dewberry bush, running blackberry",
  "n12654659": "western blackberry, western dewberry, Rubus ursinus",
  "n12654857": "boysenberry, boysenberry bush",
  "n12655062": "loganberry, Rubus loganobaccus, Rubus ursinus loganobaccus",
  "n12655245": "American dewberry, Rubus canadensis",
  "n12655351": "Northern dewberry, American dewberry, Rubus flagellaris",
  "n12655498": "Southern dewberry, Rubus trivialis",
  "n12655605": "swamp dewberry, swamp blackberry, Rubus hispidus",
  "n12655726": "European dewberry, Rubus caesius",
  "n12655869": "raspberry, raspberry bush",
  "n12656229": "red raspberry",
  "n12656369": "wild raspberry, European raspberry, framboise, Rubus idaeus",
  "n12656528": "American raspberry, Rubus strigosus, Rubus idaeus strigosus",
  "n12656685": "black raspberry, blackcap, blackcap raspberry, thimbleberry, Rubus occidentalis",
  "n12656909": "salmonberry, Rubus spectabilis",
  "n12657082": "salmonberry, salmon berry, thimbleberry, Rubus parviflorus",
  "n12657294": "cloudberry, dwarf mulberry, bakeapple, baked-apple berry, salmonberry, Rubus chamaemorus",
  "n12657509": "flowering raspberry, purple-flowering raspberry, Rubus odoratus, thimbleberry",
  "n12657755": "wineberry, Rubus phoenicolasius",
  "n12657940": "Sorbus, genus Sorbus",
  "n12658118": "mountain ash",
  "n12658308": "rowan, rowan tree, European mountain ash, Sorbus aucuparia",
  "n12658481": "rowanberry",
  "n12658603": "American mountain ash, Sorbus americana",
  "n12658715": "Western mountain ash, Sorbus sitchensis",
  "n12658846": "service tree, sorb apple, sorb apple tree, Sorbus domestica",
  "n12659064": "wild service tree, Sorbus torminalis",
  "n12659203": "Spiraea, genus Spiraea",
  "n12659356": "spirea, spiraea",
  "n12659539": "bridal wreath, bridal-wreath, Saint Peter's wreath, St. Peter's wreath, Spiraea prunifolia",
  "n12659730": "Rubiales, order Rubiales",
  "n12660009": "Rubiaceae, family Rubiaceae, madder family",
  "n12660601": "madderwort, rubiaceous plant",
  "n12660796": "Rubia, genus Rubia",
  "n12661045": "Indian madder, munjeet, Rubia cordifolia",
  "n12661227": "madder, Rubia tinctorum",
  "n12661420": "Asperula, genus Asperula",
  "n12661538": "woodruff",
  "n12661661": "dyer's woodruff, Asperula tinctoria",
  "n12661873": "Calycophyllum, genus Calycophyllum",
  "n12662074": "dagame, lemonwood tree, Calycophyllum candidissimum",
  "n12662223": "Chiococca, genus Chiococca",
  "n12662379": "blolly, West Indian snowberry, Chiococca alba",
  "n12662654": "Coffea, genus Coffea",
  "n12662772": "coffee, coffee tree",
  "n12663023": "Arabian coffee, Coffea arabica",
  "n12663254": "Liberian coffee, Coffea liberica",
  "n12663359": "robusta coffee, Rio Nunez coffee, Coffea robusta, Coffea canephora",
  "n12663554": "genus Cinchona, genus Chinchona",
  "n12663804": "cinchona, chinchona",
  "n12664005": "Cartagena bark, Cinchona cordifolia, Cinchona lancifolia",
  "n12664187": "calisaya, Cinchona officinalis, Cinchona ledgeriana, Cinchona calisaya",
  "n12664469": "cinchona tree, Cinchona pubescens",
  "n12664710": "cinchona, cinchona bark, Peruvian bark, Jesuit's bark",
  "n12664897": "Galium, genus Galium",
  "n12665048": "bedstraw",
  "n12665271": "sweet woodruff, waldmeister, woodruff, fragrant bedstraw, Galium odoratum, Asperula odorata",
  "n12665659": "Northern bedstraw, Northern snow bedstraw, Galium boreale",
  "n12665857": "yellow bedstraw, yellow cleavers, Our Lady's bedstraw, Galium verum",
  "n12666050": "wild licorice, Galium lanceolatum",
  "n12666159": "cleavers, clivers, goose grass, catchweed, spring cleavers, Galium aparine",
  "n12666369": "wild madder, white madder, white bedstraw, infant's-breath, false baby's breath, Galium mollugo",
  "n12666602": "genus Gardenia",
  "n12666768": "gardenia",
  "n12666965": "cape jasmine, cape jessamine, Gardenia jasminoides, Gardenia augusta",
  "n12667179": "genus Genipa",
  "n12667406": "genipa",
  "n12667582": "genipap fruit, jagua, marmalade box, Genipa Americana",
  "n12667817": "genus Hamelia",
  "n12667964": "hamelia",
  "n12668131": "scarlet bush, scarlet hamelia, coloradillo, Hamelia patens, Hamelia erecta",
  "n12668364": "Mitchella, genus Mitchella",
  "n12668517": "partridgeberry, boxberry, twinberry, Mitchella repens",
  "n12668732": "Nauclea, genus Nauclea",
  "n12668917": "opepe, Nauclea diderrichii, Sarcocephalus diderrichii",
  "n12669157": "Pinckneya, genus Pinckneya",
  "n12669362": "fever tree, Georgia bark, bitter-bark, Pinckneya pubens",
  "n12669641": "Psychotria, genus Psychotria",
  "n12669803": "lemonwood, lemon-wood, lemonwood tree, lemon-wood tree, Psychotria capensis",
  "n12670013": "lemonwood",
  "n12670172": "Sarcocephalus, genus Sarcocephalus",
  "n12670334": "negro peach, Sarcocephalus latifolius, Sarcocephalus esculentus",
  "n12670558": "Vangueria, genus Vangueria",
  "n12670758": "wild medlar, wild medlar tree, medlar, Vangueria infausta",
  "n12670962": "Spanish tamarind, Vangueria madagascariensis",
  "n12671157": "Caprifoliaceae, family Caprifoliaceae, honeysuckle family",
  "n12671529": "genus Abelia",
  "n12671651": "abelia",
  "n12671898": "Diervilla, genus Diervilla",
  "n12672083": "bush honeysuckle, Diervilla lonicera",
  "n12672289": "bush honeysuckle, Diervilla sessilifolia",
  "n12672497": "Kolkwitzia, genus Kolkwitzia",
  "n12672631": "beauty bush, Kolkwitzia amabilis",
  "n12672843": "Leycesteria, genus Leycesteria",
  "n12673012": "Himalaya honeysuckle, Leycesteria formosa",
  "n12673178": "Linnaea, genus Linnaea",
  "n12673328": "twinflower, Linnaea borealis",
  "n12673588": "American twinflower, Linnaea borealis americana",
  "n12673755": "Lonicera, genus Lonicera",
  "n12674120": "honeysuckle",
  "n12674484": "white honeysuckle, Lonicera albiflora",
  "n12674685": "American fly honeysuckle, fly honeysuckle, Lonicera canadensis",
  "n12674895": "Italian honeysuckle, Italian woodbine, Lonicera caprifolium",
  "n12675100": "yellow honeysuckle, Lonicera dioica",
  "n12675299": "yellow honeysuckle, Lonicera flava",
  "n12675515": "hairy honeysuckle, Lonicera hirsuta",
  "n12675716": "twinberry, Lonicera involucrata",
  "n12675876": "Japanese honeysuckle, Lonicera japonica",
  "n12676134": "Hall's honeysuckle, Lonicera japonica halliana",
  "n12676370": "Morrow's honeysuckle, Lonicera morrowii",
  "n12676534": "woodbine, Lonicera periclymenum",
  "n12676703": "trumpet honeysuckle, coral honeysuckle, trumpet flower, trumpet vine, Lonicera sempervirens",
  "n12676940": "bush honeysuckle, Tartarian honeysuckle, Lonicera tatarica",
  "n12677120": "European fly honeysuckle, European honeysuckle, Lonicera xylosteum",
  "n12677331": "swamp fly honeysuckle",
  "n12677427": "Symphoricarpos, genus Symphoricarpos",
  "n12677612": "snowberry, common snowberry, waxberry, Symphoricarpos alba",
  "n12677841": "coralberry, Indian currant, Symphoricarpos orbiculatus",
  "n12678059": "Sambucus, genus Sambucus",
  "n12678224": "elder, elderberry bush",
  "n12678548": "American elder, black elderberry, sweet elder, Sambucus canadensis",
  "n12678794": "blue elder, blue elderberry, Sambucus caerulea",
  "n12679023": "dwarf elder, danewort, Sambucus ebulus",
  "n12679201": "bourtree, black elder, common elder, elderberry, European elder, Sambucus nigra",
  "n12679432": "American red elder, red-berried elder, stinking elder, Sambucus pubens",
  "n12679593": "European red elder, red-berried elder, Sambucus racemosa",
  "n12679712": "Triostium, genus Triostium",
  "n12679876": "feverroot, horse gentian, tinker's root, wild coffee, Triostium perfoliatum",
  "n12680125": "Viburnum, genus Viburnum",
  "n12680402": "cranberry bush, cranberry tree, American cranberry bush, highbush cranberry, Viburnum trilobum",
  "n12680652": "wayfaring tree, twist wood, twistwood, Viburnum lantana",
  "n12680864": "guelder rose, European cranberrybush, European cranberry bush, crampbark, cranberry tree, Viburnum opulus",
  "n12681141": "arrow wood, southern arrow wood, Viburnum dentatum",
  "n12681376": "arrow wood, Viburnum recognitum",
  "n12681579": "black haw, Viburnum prunifolium",
  "n12681768": "genus Weigela",
  "n12681893": "weigela, Weigela florida",
  "n12682054": "Dipsacaceae, family Dipsacaceae",
  "n12682264": "Dipsacus, genus Dipsacus",
  "n12682411": "teasel, teazel, teasle",
  "n12682668": "common teasel, Dipsacus fullonum",
  "n12682882": "fuller's teasel, Dipsacus sativus",
  "n12683096": "wild teasel, Dipsacus sylvestris",
  "n12683248": "genus Scabiosa",
  "n12683407": "scabious, scabiosa",
  "n12683571": "sweet scabious, pincushion flower, mournful widow, Scabiosa atropurpurea",
  "n12683791": "field scabious, Scabiosa arvensis",
  "n12683950": "Balsaminaceae, family Balsaminaceae, balsam family",
  "n12684153": "genus Impatiens",
  "n12684379": "jewelweed, lady's earrings, orange balsam, celandine, touch-me-not, Impatiens capensis",
  "n12684640": "Geraniales, order Geraniales",
  "n12685214": "Geraniaceae, family Geraniaceae, geranium family",
  "n12685431": "geranium",
  "n12685679": "genus Geranium",
  "n12685831": "cranesbill, crane's bill",
  "n12686077": "wild geranium, spotted cranesbill, Geranium maculatum",
  "n12686274": "meadow cranesbill, Geranium pratense",
  "n12686496": "Richardson's geranium, Geranium richardsonii",
  "n12686676": "herb robert, herbs robert, herb roberts, Geranium robertianum",
  "n12686877": "sticky geranium, Geranium viscosissimum",
  "n12687044": "dove's foot geranium, Geranium molle",
  "n12687211": "Pelargonium, genus Pelargonium",
  "n12687462": "rose geranium, sweet-scented geranium, Pelargonium graveolens",
  "n12687698": "fish geranium, bedding geranium, zonal pelargonium, Pelargonium hortorum",
  "n12687957": "ivy geranium, ivy-leaved geranium, hanging geranium, Pelargonium peltatum",
  "n12688187": "apple geranium, nutmeg geranium, Pelargonium odoratissimum",
  "n12688372": "lemon geranium, Pelargonium limoneum",
  "n12688526": "Erodium, genus Erodium",
  "n12688716": "storksbill, heron's bill",
  "n12688903": "redstem storksbill, alfilaria, alfileria, filaree, filaria, clocks, pin grass, pin clover, Erodium cicutarium",
  "n12689305": "musk clover, muskus grass, white-stemmed filaree, Erodium moschatum",
  "n12689491": "Texas storksbill, Erodium texanum",
  "n12689641": "Erythroxylaceae, family Erythroxylaceae",
  "n12689808": "Erythroxylon, genus Erythroxylon, Erythroxylum, genus Erythroxylum",
  "n12690046": "Erythroxylon coca, coca, coca plant",
  "n12690240": "Erythroxylon truxiuense",
  "n12690388": "Burseraceae, family Burseraceae, torchwood family",
  "n12690653": "incense tree",
  "n12691028": "elemi, gum elemi",
  "n12691189": "Bursera, genus Bursera",
  "n12691428": "elephant tree, Bursera microphylla",
  "n12691661": "gumbo-limbo, Bursera simaruba",
  "n12691834": "Boswellia, genus Boswellia",
  "n12692024": "Boswellia carteri",
  "n12692160": "salai, Boswellia serrata",
  "n12692323": "Commiphora, genus Commiphora",
  "n12692521": "balm of gilead, Commiphora meccanensis",
  "n12692714": "myrrh tree, Commiphora myrrha",
  "n12692875": "myrrh, gum myrrh, sweet cicely",
  "n12693033": "Protium, genus Protium",
  "n12693244": "Protium heptaphyllum",
  "n12693352": "Protium guianense",
  "n12693457": "incense wood",
  "n12693590": "Callitrichaceae, family Callitrichaceae",
  "n12693734": "Callitriche, genus Callitriche",
  "n12693865": "water starwort",
  "n12694048": "Malpighiaceae, family Malpighiaceae",
  "n12694193": "Malpighia, genus Malpighia",
  "n12694336": "jiqui, Malpighia obovata",
  "n12694486": "barbados cherry, acerola, Surinam cherry, West Indian cherry, Malpighia glabra",
  "n12694707": "Meliaceae, family Meliaceae, mahogany family",
  "n12695144": "mahogany, mahogany tree",
  "n12695572": "mahogany",
  "n12695760": "Melia, genus Melia",
  "n12695975": "chinaberry, chinaberry tree, China tree, Persian lilac, pride-of-India, azederach, azedarach, Melia azederach, Melia azedarach",
  "n12696322": "Azadirachta, genus Azadirachta",
  "n12696492": "neem, neem tree, nim tree, margosa, arishth, Azadirachta indica, Melia Azadirachta",
  "n12696830": "neem seed",
  "n12697021": "Cedrela, genus Cedrela",
  "n12697152": "Spanish cedar, Spanish cedar tree, Cedrela odorata",
  "n12697360": "Chloroxylon, genus Chloroxylon",
  "n12697514": "satinwood, satinwood tree, Chloroxylon swietenia",
  "n12697709": "satinwood",
  "n12697883": "Entandrophragma, genus Entandrophragma",
  "n12698027": "African scented mahogany, cedar mahogany, sapele mahogany, Entandrophragma cylindricum",
  "n12698283": "Flindersia, genus Flindersia",
  "n12698435": "silver ash",
  "n12698598": "native beech, flindosa, flindosy, Flindersia australis",
  "n12698774": "bunji-bunji, Flindersia schottiana",
  "n12698905": "Khaya, genus Khaya",
  "n12699031": "African mahogany",
  "n12699157": "genus Lansium",
  "n12699301": "lanseh tree, langsat, langset, Lansium domesticum",
  "n12699485": "Lovoa, genus Lovoa",
  "n12699618": "African walnut, Lovoa klaineana",
  "n12699778": "Swietinia, genus Swietinia",
  "n12699922": "true mahogany, Cuban mahogany, Dominican mahogany, Swietinia mahogani",
  "n12700088": "Honduras mahogany, Swietinia macrophylla",
  "n12700219": "Toona, genus Toona",
  "n12700357": "Philippine mahogany, Philippine cedar, kalantas, Toona calantas, Cedrela calantas",
  "n12700560": "Philippine mahogany",
  "n12700711": "cigar-box cedar",
  "n12700831": "genus Turreae",
  "n12700996": "turreae",
  "n12701178": "Lepidobotryaceae, family Lepidobotryaceae",
  "n12701491": "genus Lepidobotrys",
  "n12701666": "lepidobotrys",
  "n12701901": "Ruptiliocarpon, genus Ruptiliocarpon",
  "n12702124": "caracolito, Ruptiliocarpon caracolito",
  "n12702443": "Oxalidaceae, family Oxalidaceae, wood-sorrel family",
  "n12702706": "genus Oxalis",
  "n12702948": "oxalis, sorrel, wood sorrel",
  "n12703190": "common wood sorrel, cuckoo bread, shamrock, Oxalis acetosella",
  "n12703383": "Bermuda buttercup, English-weed, Oxalis pes-caprae, Oxalis cernua",
  "n12703557": "creeping oxalis, creeping wood sorrel, Oxalis corniculata",
  "n12703716": "goatsfoot, goat's foot, Oxalis caprina",
  "n12703856": "violet wood sorrel, Oxalis violacea",
  "n12704041": "oca, oka, Oxalis tuberosa, Oxalis crenata",
  "n12704191": "Averrhoa, genus Averrhoa",
  "n12704343": "carambola, carambola tree, Averrhoa carambola",
  "n12704513": "bilimbi, Averrhoa bilimbi",
  "n12704636": "Polygalaceae, family Polygalaceae, milkwort family",
  "n12704844": "Polygala, genus Polygala",
  "n12705013": "milkwort",
  "n12705220": "senega, Polygala alba",
  "n12705458": "orange milkwort, yellow milkwort, candyweed, yellow bachelor's button, Polygala lutea",
  "n12705698": "flowering wintergreen, gaywings, bird-on-the-wing, fringed polygala, Polygala paucifolia",
  "n12705978": "Seneca snakeroot, Seneka snakeroot, senga root, senega root, senega snakeroot, Polygala senega",
  "n12706240": "senega",
  "n12706410": "common milkwort, gand flower, Polygala vulgaris",
  "n12706644": "Rutaceae, family Rutaceae, rue family",
  "n12707040": "Ruta, genus Ruta",
  "n12707199": "rue, herb of grace, Ruta graveolens",
  "n12707432": "genus Citrus",
  "n12707781": "citrus, citrus tree",
  "n12708293": "orange, orange tree",
  "n12708520": "orangewood",
  "n12708654": "sour orange, Seville orange, bitter orange, bitter orange tree, bigarade, marmalade orange, Citrus aurantium",
  "n12708941": "bergamot, bergamot orange, Citrus bergamia",
  "n12709103": "pomelo, pomelo tree, pummelo, shaddock, Citrus maxima, Citrus grandis, Citrus decumana",
  "n12709349": "citron, citron tree, Citrus medica",
  "n12709591": "citronwood",
  "n12709688": "grapefruit, Citrus paradisi",
  "n12709901": "mandarin, mandarin orange, mandarin orange tree, Citrus reticulata",
  "n12710295": "tangerine, tangerine tree",
  "n12710415": "clementine, clementine tree",
  "n12710577": "satsuma, satsuma tree",
  "n12710693": "sweet orange, sweet orange tree, Citrus sinensis",
  "n12710917": "temple orange, temple orange tree, tangor, king orange, Citrus nobilis",
  "n12711182": "tangelo, tangelo tree, ugli fruit, Citrus tangelo",
  "n12711398": "rangpur, rangpur lime, lemanderin, Citrus limonia",
  "n12711596": "lemon, lemon tree, Citrus limon",
  "n12711817": "sweet lemon, sweet lime, Citrus limetta",
  "n12711984": "lime, lime tree, Citrus aurantifolia",
  "n12712149": "Citroncirus, genus Citroncirus",
  "n12712320": "citrange, citrange tree, Citroncirus webberi",
  "n12712488": "Dictamnus, genus Dictamnus",
  "n12712626": "fraxinella, dittany, burning bush, gas plant, Dictamnus alba",
  "n12712820": "Fortunella, genus Fortunella",
  "n12713063": "kumquat, cumquat, kumquat tree",
  "n12713358": "marumi, marumi kumquat, round kumquat, Fortunella japonica",
  "n12713521": "nagami, nagami kumquat, oval kumquat, Fortunella margarita",
  "n12713664": "Phellodendron, genus Phellodendron",
  "n12713866": "cork tree, Phellodendron amurense",
  "n12714114": "Poncirus, genus Poncirus",
  "n12714254": "trifoliate orange, trifoliata, wild orange, Poncirus trifoliata",
  "n12714550": "Zanthoxylum, genus Zanthoxylum",
  "n12714755": "prickly ash",
  "n12714949": "toothache tree, sea ash, Zanthoxylum americanum, Zanthoxylum fraxineum",
  "n12715195": "Hercules'-club, Hercules'-clubs, Hercules-club, Zanthoxylum clava-herculis",
  "n12715408": "satinwood, West Indian satinwood, Zanthoxylum flavum",
  "n12715569": "Simaroubaceae, family Simaroubaceae, quassia family",
  "n12715914": "bitterwood tree",
  "n12716166": "Simarouba, genus Simarouba",
  "n12716400": "marupa, Simarouba amara",
  "n12716594": "paradise tree, bitterwood, Simarouba glauca",
  "n12716861": "genus Ailanthus",
  "n12717072": "ailanthus",
  "n12717224": "tree of heaven, tree of the gods, Ailanthus altissima",
  "n12717524": "Irvingia, genus Irvingia",
  "n12717644": "wild mango, dika, wild mango tree, Irvingia gabonensis",
  "n12717914": "Kirkia, genus Kirkia",
  "n12718074": "pepper tree, Kirkia wilmsii",
  "n12718314": "Picrasma, genus Picrasma",
  "n12718483": "Jamaica quassia, bitterwood, Picrasma excelsa, Picrasma excelsum",
  "n12718685": "Jamaica quassia",
  "n12718807": "genus Quassia",
  "n12718995": "quassia, bitterwood, Quassia amara",
  "n12719277": "Tropaeolaceae, family Tropaeolaceae, nasturtium family",
  "n12719455": "Tropaeolum, genus Tropaeolum",
  "n12719684": "nasturtium",
  "n12719944": "garden nasturtium, Indian cress, Tropaeolum majus",
  "n12720200": "bush nasturtium, Tropaeolum minus",
  "n12720354": "canarybird flower, canarybird vine, canary creeper, Tropaeolum peregrinum",
  "n12720532": "Zygophyllaceae, family Zygophyllaceae, bean-caper family",
  "n12720893": "Zygophyllum, genus Zygophyllum",
  "n12721122": "bean caper, Syrian bean caper, Zygophyllum fabago",
  "n12721357": "Bulnesia, genus Bulnesia",
  "n12721477": "palo santo, Bulnesia sarmienti",
  "n12721705": "guaiac wood, guaiacum wood",
  "n12721864": "Guaiacum, genus Guaiacum",
  "n12722071": "lignum vitae, Guaiacum officinale",
  "n12722382": "lignum vitae, guaiac, guaiacum",
  "n12722567": "bastard lignum vitae, Guaiacum sanctum",
  "n12722768": "guaiacum",
  "n12722884": "Larrea, genus Larrea",
  "n12723062": "creosote bush, coville, hediondilla, Larrea tridentata",
  "n12723330": "Sonora gum",
  "n12723446": "Tribulus, genus Tribulus",
  "n12723610": "caltrop, devil's weed, Tribulus terestris",
  "n12723835": "Salicales, order Salicales",
  "n12723985": "Salicaceae, family Salicaceae, willow family",
  "n12724201": "Salix, genus Salix",
  "n12724942": "willow, willow tree",
  "n12725521": "osier",
  "n12725738": "white willow, Huntingdon willow, Salix alba",
  "n12725940": "silver willow, silky willow, Salix alba sericea, Salix sericea",
  "n12726159": "golden willow, Salix alba vitellina, Salix vitellina",
  "n12726357": "cricket-bat willow, Salix alba caerulea",
  "n12726528": "arctic willow, Salix arctica",
  "n12726670": "weeping willow, Babylonian weeping willow, Salix babylonica",
  "n12726902": "Wisconsin weeping willow, Salix pendulina, Salix blanda, Salix pendulina blanda",
  "n12727101": "pussy willow, Salix discolor",
  "n12727301": "sallow",
  "n12727518": "goat willow, florist's willow, pussy willow, Salix caprea",
  "n12727729": "peachleaf willow, peach-leaved willow, almond-leaves willow, Salix amygdaloides",
  "n12727960": "almond willow, black Hollander, Salix triandra, Salix amygdalina",
  "n12728164": "hoary willow, sage willow, Salix candida",
  "n12728322": "crack willow, brittle willow, snap willow, Salix fragilis",
  "n12728508": "prairie willow, Salix humilis",
  "n12728656": "dwarf willow, Salix herbacea",
  "n12728864": "grey willow, gray willow, Salix cinerea",
  "n12729023": "arroyo willow, Salix lasiolepis",
  "n12729164": "shining willow, Salix lucida",
  "n12729315": "swamp willow, black willow, Salix nigra",
  "n12729521": "bay willow, laurel willow, Salix pentandra",
  "n12729729": "purple willow, red willow, red osier, basket willow, purple osier, Salix purpurea",
  "n12729950": "balsam willow, Salix pyrifolia",
  "n12730143": "creeping willow, Salix repens",
  "n12730370": "Sitka willow, silky willow, Salix sitchensis",
  "n12730544": "dwarf grey willow, dwarf gray willow, sage willow, Salix tristis",
  "n12730776": "bearberry willow, Salix uva-ursi",
  "n12731029": "common osier, hemp willow, velvet osier, Salix viminalis",
  "n12731202": "Populus, genus Populus",
  "n12731401": "poplar, poplar tree",
  "n12731714": "poplar",
  "n12731835": "balsam poplar, hackmatack, tacamahac, Populus balsamifera",
  "n12732009": "white poplar, white aspen, abele, aspen poplar, silver-leaved poplar, Populus alba",
  "n12732252": "grey poplar, gray poplar, Populus canescens",
  "n12732491": "black poplar, Populus nigra",
  "n12732605": "Lombardy poplar, Populus nigra italica",
  "n12732756": "cottonwood",
  "n12732966": "Eastern cottonwood, necklace poplar, Populus deltoides",
  "n12733218": "black cottonwood, Western balsam poplar, Populus trichocarpa",
  "n12733428": "swamp cottonwood, black cottonwood, downy poplar, swamp poplar, Populus heterophylla",
  "n12733647": "aspen",
  "n12733870": "quaking aspen, European quaking aspen, Populus tremula",
  "n12734070": "American quaking aspen, American aspen, Populus tremuloides",
  "n12734215": "Canadian aspen, bigtooth aspen, bigtoothed aspen, big-toothed aspen, large-toothed aspen, large tooth aspen, Populus grandidentata",
  "n12734446": "Santalales, order Santalales",
  "n12734722": "Santalaceae, family Santalaceae, sandalwood family",
  "n12735009": "Santalum, genus Santalum",
  "n12735160": "sandalwood tree, true sandalwood, Santalum album",
  "n12735452": "sandalwood",
  "n12735666": "genus Buckleya",
  "n12735817": "buckleya, Buckleya distichophylla",
  "n12736064": "Comandra, genus Comandra",
  "n12736228": "bastard toadflax, Comandra pallida",
  "n12736455": "Eucarya, genus Eucarya, Fusanus, genus Fusanus",
  "n12736603": "quandong, quandang, quandong tree, Eucarya acuminata, Fusanus acuminatus",
  "n12736840": "Pyrularia, genus Pyrularia",
  "n12736999": "rabbitwood, buffalo nut, Pyrularia pubera",
  "n12737251": "buffalo nut, elk nut, oil nut",
  "n12737383": "Loranthaceae, family Loranthaceae, mistletoe family",
  "n12737745": "Loranthus, genus Loranthus",
  "n12737898": "mistletoe, Loranthus europaeus",
  "n12738087": "Arceuthobium, genus Arceuthobium",
  "n12738259": "American mistletoe, Arceuthobium pusillum",
  "n12738480": "Nuytsia, genus Nuytsia",
  "n12738599": "flame tree, fire tree, Christmas tree, Nuytsia floribunda",
  "n12738859": "Viscaceae, family Viscaceae, mistletoe family",
  "n12739072": "Viscum, genus Viscum",
  "n12739332": "mistletoe, Viscum album, Old World mistletoe",
  "n12739595": "Phoradendron, genus Phoradendron",
  "n12739801": "mistletoe, false mistletoe",
  "n12739966": "American mistletoe, Phoradendron serotinum, Phoradendron flavescens",
  "n12740196": "Sapindales, order Sapindales",
  "n12740514": "Sapindaceae, family Sapindaceae, soapberry family",
  "n12740967": "aalii",
  "n12741079": "Dodonaea, genus Dodonaea",
  "n12741222": "soapberry, soapberry tree",
  "n12741409": "Sapindus, genus Sapindus",
  "n12741586": "wild China tree, Sapindus drumondii, Sapindus marginatus",
  "n12741792": "China tree, false dogwood, jaboncillo, chinaberry, Sapindus saponaria",
  "n12742041": "Blighia, genus Blighia",
  "n12742290": "akee, akee tree, Blighia sapida",
  "n12742546": "Cardiospermum, genus Cardiospermum",
  "n12742741": "soapberry vine",
  "n12742878": "heartseed, Cardiospermum grandiflorum",
  "n12743009": "balloon vine, heart pea, Cardiospermum halicacabum",
  "n12743232": "Dimocarpus, genus Dimocarpus",
  "n12743352": "longan, lungen, longanberry, Dimocarpus longan, Euphorbia litchi, Nephelium longana",
  "n12743680": "genus Harpullia",
  "n12743823": "harpullia",
  "n12743976": "harpulla, Harpullia cupanioides",
  "n12744142": "Moreton Bay tulipwood, Harpullia pendula",
  "n12744277": "genus Litchi",
  "n12744387": "litchi, lichee, litchi tree, Litchi chinensis, Nephelium litchi",
  "n12744656": "Melicoccus, genus Melicoccus, Melicocca, genus Melicocca",
  "n12744850": "Spanish lime, Spanish lime tree, honey berry, mamoncillo, genip, ginep, Melicocca bijuga, Melicocca bijugatus",
  "n12745160": "Nephelium, genus Nephelium",
  "n12745386": "rambutan, rambotan, rambutan tree, Nephelium lappaceum",
  "n12745564": "pulasan, pulassan, pulasan tree, Nephelium mutabile",
  "n12745788": "Buxaceae, family Buxaceae, box family",
  "n12745976": "Buxus, genus Buxus",
  "n12746106": "box, boxwood",
  "n12746253": "common box, European box, Buxus sempervirens",
  "n12746474": "boxwood, Turkish boxwood",
  "n12746733": "genus Pachysandra",
  "n12746884": "pachysandra",
  "n12747120": "Allegheny spurge, Allegheny mountain spurge, Pachysandra procumbens",
  "n12747371": "Japanese spurge, Pachysandra terminalis",
  "n12747563": "Celastraceae, family Celastraceae, spindle-tree family, staff-tree family",
  "n12747831": "staff tree",
  "n12747961": "Celastrus, genus Celastrus",
  "n12748248": "bittersweet, American bittersweet, climbing bittersweet, false bittersweet, staff vine, waxwork, shrubby bittersweet, Celastrus scandens",
  "n12748534": "Japanese bittersweet, Japan bittersweet, oriental bittersweet, Celastrus orbiculatus, Celastric articulatus",
  "n12748815": "Euonymus, genus Euonymus",
  "n12749049": "spindle tree, spindleberry, spindleberry tree",
  "n12749289": "common spindle tree, Euonymus europaeus",
  "n12749456": "winged spindle tree, Euonymous alatus",
  "n12749679": "wahoo, burning bush, Euonymus atropurpureus",
  "n12749852": "strawberry bush, wahoo, Euonymus americanus",
  "n12750076": "evergreen bittersweet, Euonymus fortunei radicans, Euonymus radicans vegetus",
  "n12750306": "Cyrilliaceae, family Cyrilliaceae, cyrilla family, titi family",
  "n12750577": "genus Cyrilla",
  "n12750767": "cyrilla, leatherwood, white titi, Cyrilla racemiflora",
  "n12751043": "Cliftonia, genus Cliftonia",
  "n12751172": "titi, buckwheat tree, Cliftonia monophylla",
  "n12751402": "Empetraceae, family Empetraceae, crowberry family",
  "n12751554": "Empetrum, genus Empetrum",
  "n12751675": "crowberry",
  "n12751823": "Aceraceae, family Aceraceae, maple family",
  "n12752039": "Acer, genus Acer",
  "n12752205": "maple",
  "n12752666": "maple",
  "n12752892": "bird's-eye maple",
  "n12753007": "silver maple, Acer saccharinum",
  "n12753245": "sugar maple, rock maple, Acer saccharum",
  "n12753573": "red maple, scarlet maple, swamp maple, Acer rubrum",
  "n12753762": "moosewood, moose-wood, striped maple, striped dogwood, goosefoot maple, Acer pennsylvanicum",
  "n12754003": "Oregon maple, big-leaf maple, Acer macrophyllum",
  "n12754174": "dwarf maple, Rocky-mountain maple, Acer glabrum",
  "n12754311": "mountain maple, mountain alder, Acer spicatum",
  "n12754468": "vine maple, Acer circinatum",
  "n12754648": "hedge maple, field maple, Acer campestre",
  "n12754781": "Norway maple, Acer platanoides",
  "n12754981": "sycamore, great maple, scottish maple, Acer pseudoplatanus",
  "n12755225": "box elder, ash-leaved maple, Acer negundo",
  "n12755387": "California box elder, Acer negundo Californicum",
  "n12755559": "pointed-leaf maple, Acer argutum",
  "n12755727": "Japanese maple, full moon maple, Acer japonicum",
  "n12755876": "Japanese maple, Acer palmatum",
  "n12756059": "Dipteronia, genus Dipteronia",
  "n12756286": "Aquifoliaceae, family Aquifoliaceae, holly family",
  "n12756457": "holly",
  "n12756862": "Ilex, genus Ilex",
  "n12757115": "Chinese holly, Ilex cornuta",
  "n12757303": "bearberry, possum haw, winterberry, Ilex decidua",
  "n12757458": "inkberry, gallberry, gall-berry, evergreen winterberry, Ilex glabra",
  "n12757668": "mate, Paraguay tea, Ilex paraguariensis",
  "n12757816": "American holly, Christmas holly",
  "n12757930": "low gallberry holly",
  "n12758014": "tall gallberry holly",
  "n12758099": "yaupon holly",
  "n12758176": "deciduous holly",
  "n12758250": "juneberry holly",
  "n12758325": "largeleaf holly",
  "n12758399": "Geogia holly",
  "n12758471": "common winterberry holly",
  "n12758555": "smooth winterberry holly",
  "n12758639": "Anacardiaceae, family Anacardiaceae, sumac family",
  "n12759120": "Anacardium, genus Anacardium",
  "n12759273": "cashew, cashew tree, Anacardium occidentale",
  "n12759496": "Astronium, genus Astronium",
  "n12759668": "goncalo alves, Astronium fraxinifolium",
  "n12760013": "Cotinus, genus Cotinus",
  "n12760132": "smoke tree, smoke bush",
  "n12760316": "American smokewood, chittamwood, Cotinus americanus, Cotinus obovatus",
  "n12760539": "Venetian sumac, wig tree, Cotinus coggygria",
  "n12760722": "Malosma, genus Malosma",
  "n12760875": "laurel sumac, Malosma laurina, Rhus laurina",
  "n12761123": "Mangifera, genus Mangifera",
  "n12761284": "mango, mango tree, Mangifera indica",
  "n12761471": "Pistacia, genus Pistacia",
  "n12761702": "pistachio, Pistacia vera, pistachio tree",
  "n12761905": "terebinth, Pistacia terebinthus",
  "n12762049": "mastic, mastic tree, lentisk, Pistacia lentiscus",
  "n12762245": "Rhodosphaera, genus Rhodosphaera",
  "n12762405": "Australian sumac, Rhodosphaera rhodanthema, Rhus rhodanthema",
  "n12762583": "Rhus, genus Rhus",
  "n12762896": "sumac, sumach, shumac",
  "n12763205": "sumac",
  "n12763291": "fragrant sumac, lemon sumac, Rhus aromatica",
  "n12763529": "smooth sumac, scarlet sumac, vinegar tree, Rhus glabra",
  "n12763762": "dwarf sumac, mountain sumac, black sumac, shining sumac, Rhus copallina",
  "n12764008": "sugar-bush, sugar sumac, Rhus ovata",
  "n12764202": "staghorn sumac, velvet sumac, Virginian sumac, vinegar tree, Rhus typhina",
  "n12764507": "squawbush, squaw-bush, skunkbush, Rhus trilobata",
  "n12764703": "Schinus, genus Schinus",
  "n12764978": "aroeira blanca, Schinus chichita",
  "n12765115": "pepper tree, molle, Peruvian mastic tree, Schinus molle",
  "n12765402": "Brazilian pepper tree, Schinus terebinthifolius",
  "n12765679": "Spondias, genus Spondias",
  "n12765846": "hog plum, yellow mombin, yellow mombin tree, Spondias mombin",
  "n12766043": "mombin, mombin tree, jocote, Spondias purpurea",
  "n12766241": "Toxicodendron, genus Toxicodendron",
  "n12766595": "poison ash, poison dogwood, poison sumac, Toxicodendron vernix, Rhus vernix",
  "n12766869": "poison ivy, markweed, poison mercury, poison oak, Toxicodendron radicans, Rhus radicans",
  "n12767208": "western poison oak, Toxicodendron diversilobum, Rhus diversiloba",
  "n12767423": "eastern poison oak, Toxicodendron quercifolium, Rhus quercifolia, Rhus toxicodenedron",
  "n12767648": "varnish tree, lacquer tree, Chinese lacquer tree, Japanese lacquer tree, Japanese varnish tree, Japanese sumac, Toxicodendron vernicifluum, Rhus verniciflua",
  "n12767951": "Hippocastanaceae, family Hippocastanaceae, horse-chestnut family",
  "n12768177": "Aesculus, genus Aesculus",
  "n12768369": "horse chestnut, buckeye, Aesculus hippocastanum",
  "n12768682": "buckeye, horse chestnut, conker",
  "n12768809": "sweet buckeye",
  "n12768933": "Ohio buckeye",
  "n12769065": "dwarf buckeye, bottlebrush buckeye",
  "n12769219": "red buckeye",
  "n12769318": "particolored buckeye",
  "n12769430": "Staphylaceae, family Staphylaceae, bladdernut family",
  "n12769663": "Staphylea, genus Staphylea",
  "n12769815": "Ebenales, order Ebenales",
  "n12770068": "Ebenaceae, family Ebenaceae, ebony family",
  "n12770277": "Diospyros, genus Diospyros",
  "n12770529": "ebony, ebony tree, Diospyros ebenum",
  "n12770736": "ebony",
  "n12770892": "marblewood, marble-wood, Andaman marble, Diospyros kurzii",
  "n12771085": "marblewood, marble-wood",
  "n12771192": "persimmon, persimmon tree",
  "n12771390": "Japanese persimmon, kaki, Diospyros kaki",
  "n12771597": "American persimmon, possumwood, Diospyros virginiana",
  "n12771890": "date plum, Diospyros lotus",
  "n12772081": "Sapotaceae, family Sapotaceae, sapodilla family",
  "n12772419": "Achras, genus Achras",
  "n12772557": "Bumelia, genus Bumelia",
  "n12772753": "buckthorn",
  "n12772908": "southern buckthorn, shittimwood, shittim, mock orange, Bumelia lycioides",
  "n12773142": "false buckthorn, chittamwood, chittimwood, shittimwood, black haw, Bumelia lanuginosa",
  "n12773334": "Calocarpum, genus Calocarpum",
  "n12773488": "Chrysophyllum, genus Chrysophyllum",
  "n12773651": "star apple, caimito, Chrysophyllum cainito",
  "n12773917": "satinleaf, satin leaf, caimitillo, damson plum, Chrysophyllum oliviforme",
  "n12774127": "Manilkara, genus Manilkara",
  "n12774299": "balata, balata tree, beefwood, bully tree, Manilkara bidentata",
  "n12774496": "balata, gutta balata",
  "n12774641": "sapodilla, sapodilla tree, Manilkara zapota, Achras zapota",
  "n12774891": "Palaquium, genus Palaquium",
  "n12775070": "gutta-percha tree, Palaquium gutta",
  "n12775225": "Payena, genus Payena",
  "n12775393": "gutta-percha tree",
  "n12775530": "Pouteria, genus Pouteria",
  "n12775717": "canistel, canistel tree, Pouteria campechiana nervosa",
  "n12775919": "marmalade tree, mammee, sapote, Pouteria zapota, Calocarpum zapota",
  "n12776212": "Symplocaceae, family Symplocaceae, sweetleaf family",
  "n12776391": "Symplocus, genus Symplocus",
  "n12776558": "sweetleaf, Symplocus tinctoria",
  "n12776774": "Asiatic sweetleaf, sapphire berry, Symplocus paniculata",
  "n12776946": "Styracaceae, family Styracaceae, storax family, styrax family",
  "n12777179": "storax",
  "n12777294": "genus Styrax",
  "n12777436": "styrax",
  "n12777680": "snowbell, Styrax obassia",
  "n12777778": "Japanese snowbell, Styrax japonicum",
  "n12777892": "Texas snowbell, Texas snowbells, Styrax texana",
  "n12778045": "Halesia, genus Halesia",
  "n12778219": "silver bell",
  "n12778398": "silver-bell tree, silverbell tree, snowdrop tree, opossum wood, Halesia carolina, Halesia tetraptera",
  "n12778605": "carnivorous plant",
  "n12778926": "Sarraceniales, order Sarraceniales",
  "n12779233": "Sarraceniaceae, family Sarraceniaceae, pitcher-plant family",
  "n12779437": "Sarracenia, genus Sarracenia",
  "n12779603": "pitcher plant",
  "n12779851": "common pitcher plant, huntsman's cup, huntsman's cups, Sarracenia purpurea",
  "n12780168": "pitcher",
  "n12780325": "hooded pitcher plant, Sarracenia minor",
  "n12780563": "huntsman's horn, huntsman's horns, yellow trumpet, yellow pitcher plant, trumpets, Sarracenia flava",
  "n12780852": "Darlingtonia, genus Darlingtonia",
  "n12781007": "California pitcher plant, Darlingtonia californica",
  "n12781241": "Heliamphora, genus Heliamphora",
  "n12781421": "sun pitcher",
  "n12781659": "Nepenthaceae, family Nepenthaceae",
  "n12781814": "Nepenthes, genus Nepenthes",
  "n12781940": "tropical pitcher plant",
  "n12782108": "Droseraceae, family Droseraceae, sundew family",
  "n12782338": "Drosera, genus Drosera",
  "n12782530": "sundew, sundew plant, daily dew",
  "n12782774": "Dionaea, genus Dionaea",
  "n12782915": "Venus's flytrap, Venus's flytraps, Dionaea muscipula",
  "n12783173": "Aldrovanda, genus Aldrovanda",
  "n12783316": "waterwheel plant, Aldrovanda vesiculosa",
  "n12783601": "Drosophyllum, genus Drosophyllum",
  "n12783730": "Drosophyllum lusitanicum",
  "n12783996": "Roridulaceae, family Roridulaceae",
  "n12784173": "genus Roridula",
  "n12784371": "roridula",
  "n12784543": "Cephalotaceae, family Cephalotaceae",
  "n12784738": "Cephalotus, genus Cephalotus",
  "n12784889": "Australian pitcher plant, Cephalotus follicularis",
  "n12785110": "Crassulaceae, family Crassulaceae, stonecrop family",
  "n12785312": "Crassula, genus Crassula",
  "n12785499": "genus Sedum",
  "n12785724": "sedum",
  "n12785889": "stonecrop",
  "n12786097": "wall pepper, Sedum acre",
  "n12786273": "rose-root, midsummer-men, Sedum rosea",
  "n12786464": "orpine, orpin, livelong, live-forever, Sedum telephium",
  "n12786684": "Aeonium, genus Aeonium",
  "n12786836": "pinwheel, Aeonium haworthii",
  "n12787007": "Cunoniaceae, family Cunoniaceae, cunonia family",
  "n12787196": "Ceratopetalum, genus Ceratopetalum",
  "n12787364": "Christmas bush, Christmas tree, Ceratopetalum gummiferum",
  "n12787565": "Hydrangeaceae, family Hydrangeaceae, hydrangea family",
  "n12787846": "genus Hydrangea",
  "n12788201": "hydrangea",
  "n12788487": "climbing hydrangea, Hydrangea anomala",
  "n12788678": "wild hydrangea, Hydrangea arborescens",
  "n12788854": "hortensia, Hydrangea macrophylla hortensis",
  "n12789054": "fall-blooming hydrangea, Hydrangea paniculata",
  "n12789226": "climbing hydrangea, Hydrangea petiolaris",
  "n12789399": "genus Carpenteria",
  "n12789554": "carpenteria, Carpenteria californica",
  "n12789767": "Decumaria, genus Decumaria",
  "n12789977": "decumary, Decumaria barbata, Decumaria barbara",
  "n12790185": "genus Deutzia",
  "n12790430": "deutzia",
  "n12790656": "Philadelphaceae, subfamily Philadelphaceae",
  "n12790835": "genus Philadelphus",
  "n12791064": "philadelphus",
  "n12791329": "mock orange, syringa, Philadelphus coronarius",
  "n12791539": "Schizophragma, genus Schizophragma",
  "n12791790": "climbing hydrangea, Schizophragma hydrangeoides",
  "n12792041": "Saxifragaceae, family Saxifragaceae, saxifrage family",
  "n12792638": "Saxifraga, genus Saxifraga",
  "n12793015": "saxifrage, breakstone, rockfoil",
  "n12793284": "yellow mountain saxifrage, Saxifraga aizoides",
  "n12793494": "meadow saxifrage, fair-maids-of-France, Saxifraga granulata",
  "n12793695": "mossy saxifrage, Saxifraga hypnoides",
  "n12793886": "western saxifrage, Saxifraga occidentalis",
  "n12794135": "purple saxifrage, Saxifraga oppositifolia",
  "n12794367": "star saxifrage, starry saxifrage, Saxifraga stellaris",
  "n12794568": "strawberry geranium, strawberry saxifrage, mother-of-thousands, Saxifraga stolonifera, Saxifraga sarmentosam",
  "n12794853": "genus Astilbe",
  "n12794985": "astilbe",
  "n12795209": "false goatsbeard, Astilbe biternata",
  "n12795352": "dwarf astilbe, Astilbe chinensis pumila",
  "n12795555": "spirea, spiraea, Astilbe japonica",
  "n12795829": "genus Bergenia",
  "n12796022": "bergenia",
  "n12796192": "Boykinia, genus Boykinia",
  "n12796385": "coast boykinia, Boykinia elata, Boykinia occidentalis",
  "n12796617": "Chrysosplenium, genus Chrysosplenium",
  "n12796849": "golden saxifrage, golden spleen",
  "n12797025": "water carpet, water mat, Chrysosplenium americanum",
  "n12797213": "Darmera, genus Darmera, Peltiphyllum, genus Peltiphyllum",
  "n12797368": "umbrella plant, Indian rhubarb, Darmera peltata, Peltiphyllum peltatum",
  "n12797693": "Francoa, genus Francoa",
  "n12797860": "bridal wreath, bridal-wreath, Francoa ramosa",
  "n12798041": "Heuchera, genus Heuchera",
  "n12798284": "alumroot, alumbloom",
  "n12798466": "rock geranium, Heuchera americana",
  "n12798632": "poker alumroot, poker heuchera, Heuchera cylindrica",
  "n12798910": "coralbells, Heuchera sanguinea",
  "n12799119": "Leptarrhena, genus Leptarrhena",
  "n12799269": "leatherleaf saxifrage, Leptarrhena pyrolifolia",
  "n12799580": "Lithophragma, genus Lithophragma",
  "n12799776": "woodland star, Lithophragma affine, Lithophragma affinis, Tellima affinis",
  "n12800049": "prairie star, Lithophragma parviflorum",
  "n12800327": "Mitella, genus Mitella",
  "n12800586": "miterwort, mitrewort, bishop's cap",
  "n12800832": "fairy cup, Mitella diphylla",
  "n12801072": "five-point bishop's cap, Mitella pentandra",
  "n12801323": "genus Parnassia",
  "n12801520": "parnassia, grass-of-Parnassus",
  "n12801781": "bog star, Parnassia palustris",
  "n12801966": "fringed grass of Parnassus, Parnassia fimbriata",
  "n12802248": "genus Suksdorfia",
  "n12802442": "suksdorfia",
  "n12802707": "violet suksdorfia, Suksdorfia violaceae",
  "n12802987": "Tellima, genus Tellima",
  "n12803226": "false alumroot, fringe cups, Tellima grandiflora",
  "n12803517": "Tiarella, genus Tiarella",
  "n12803754": "foamflower, coolwart, false miterwort, false mitrewort, Tiarella cordifolia",
  "n12803958": "false miterwort, false mitrewort, Tiarella unifoliata",
  "n12804216": "Tolmiea, genus Tolmiea",
  "n12804352": "pickaback plant, piggyback plant, youth-on-age, Tolmiea menziesii",
  "n12804621": "Grossulariaceae, family Grossulariaceae, gooseberry family",
  "n12804866": "Ribes, genus Ribes",
  "n12805146": "currant, currant bush",
  "n12805373": "red currant, garden current, Ribes rubrum",
  "n12805561": "black currant, European black currant, Ribes nigrum",
  "n12805762": "white currant, Ribes sativum",
  "n12805899": "winter currant, Ribes sanguineum",
  "n12806015": "gooseberry, gooseberry bush, Ribes uva-crispa, Ribes grossularia",
  "n12806270": "Platanaceae, family Platanaceae, plane-tree family",
  "n12806455": "Platanus, genus Platanus",
  "n12806732": "plane tree, sycamore, platan",
  "n12807082": "sycamore, lacewood",
  "n12807251": "London plane, Platanus acerifolia",
  "n12807409": "American sycamore, American plane, buttonwood, Platanus occidentalis",
  "n12807624": "oriental plane, Platanus orientalis",
  "n12807773": "California sycamore, Platanus racemosa",
  "n12808007": "Arizona sycamore, Platanus wrightii",
  "n12808227": "Polemoniales, order Polemoniales",
  "n12808751": "Scrophulariales, order Scrophulariales",
  "n12808933": "Polemoniaceae, family Polemoniaceae, phlox family",
  "n12809233": "genus Polemonium",
  "n12809365": "polemonium",
  "n12809626": "Jacob's ladder, Greek valerian, charity, Polemonium caeruleum, Polemonium van-bruntiae, Polymonium caeruleum van-bruntiae",
  "n12809868": "Greek valerian, Polemonium reptans",
  "n12810007": "northern Jacob's ladder, Polemonium boreale",
  "n12810151": "skunkweed, skunk-weed, Polemonium viscosum",
  "n12810318": "genus Phlox",
  "n12810595": "phlox",
  "n12810847": "chickweed phlox, sand phlox, Phlox bifida, Phlox stellaria",
  "n12811027": "moss pink, mountain phlox, moss phlox, dwarf phlox, Phlox subulata",
  "n12811294": "Linanthus, genus Linanthus",
  "n12811501": "ground pink, fringed pink, moss pink, Linanthus dianthiflorus",
  "n12811713": "evening-snow, Linanthus dichotomus",
  "n12811856": "Acanthaceae, family Acanthaceae, acanthus family",
  "n12812121": "genus Acanthus",
  "n12812235": "acanthus",
  "n12812478": "bear's breech, bear's breeches, sea holly, Acanthus mollis",
  "n12812665": "Graptophyllum, genus Graptophyllum",
  "n12812801": "caricature plant, Graptophyllum pictum",
  "n12813024": "Thunbergia, genus Thunbergia",
  "n12813189": "black-eyed Susan, black-eyed Susan vine, Thunbergia alata",
  "n12813393": "Bignoniaceae, family Bignoniaceae",
  "n12813753": "bignoniad",
  "n12813870": "Bignonia, genus Bignonia",
  "n12814003": "cross vine, trumpet flower, quartervine, quarter-vine, Bignonia capreolata",
  "n12814235": "trumpet creeper, trumpet vine, Campsis radicans",
  "n12814417": "genus Catalpa",
  "n12814643": "catalpa, Indian bean",
  "n12814857": "Catalpa bignioides",
  "n12814960": "Catalpa speciosa",
  "n12815060": "Chilopsis, genus Chilopsis",
  "n12815198": "desert willow, Chilopsis linearis",
  "n12815434": "Crescentia, genus Crescentia",
  "n12815668": "calabash, calabash tree, Crescentia cujete",
  "n12815838": "calabash",
  "n12815925": "Boraginaceae, family Boraginaceae, borage family",
  "n12816359": "Borago, genus Borago",
  "n12816508": "borage, tailwort, Borago officinalis",
  "n12816753": "Amsinckia, genus Amsinckia",
  "n12816942": "common amsinckia, Amsinckia intermedia",
  "n12817129": "large-flowered fiddleneck, Amsinckia grandiflora",
  "n12817335": "genus Anchusa",
  "n12817464": "anchusa",
  "n12817694": "bugloss, alkanet, Anchusa officinalis",
  "n12817855": "cape forget-me-not, Anchusa capensis",
  "n12818004": "cape forget-me-not, Anchusa riparia",
  "n12818147": "Cordia, genus Cordia",
  "n12818346": "Spanish elm, Equador laurel, salmwood, cypre, princewood, Cordia alliodora",
  "n12818601": "princewood, Spanish elm, Cordia gerascanthus",
  "n12818742": "Cynoglossum, genus Cynoglossum",
  "n12818966": "Chinese forget-me-not, Cynoglossum amabile",
  "n12819141": "hound's-tongue, Cynoglossum officinale",
  "n12819354": "hound's-tongue, Cynoglossum virginaticum",
  "n12819560": "Echium, genus Echium",
  "n12819728": "blueweed, blue devil, blue thistle, viper's bugloss, Echium vulgare",
  "n12819953": "Hackelia, genus Hackelia, Lappula, genus Lappula",
  "n12820113": "beggar's lice, beggar lice",
  "n12820292": "stickweed",
  "n12820434": "Lithospermum, genus Lithospermum",
  "n12820669": "gromwell, Lithospermum officinale",
  "n12820853": "puccoon, Lithospermum caroliniense",
  "n12821048": "hoary puccoon, Indian paint, Lithospermum canescens",
  "n12821257": "Mertensia, genus Mertensia",
  "n12821505": "Virginia bluebell, Virginia cowslip, Mertensia virginica",
  "n12821736": "Myosotis, genus Myosotis",
  "n12821895": "garden forget-me-not, Myosotis sylvatica",
  "n12822115": "forget-me-not, mouse ear, Myosotis scorpiodes",
  "n12822284": "Onosmodium, genus Onosmodium",
  "n12822466": "false gromwell",
  "n12822650": "Symphytum, genus Symphytum",
  "n12822769": "comfrey, cumfrey",
  "n12822955": "common comfrey, boneset, Symphytum officinale",
  "n12823164": "Convolvulaceae, family Convolvulaceae, morning-glory family",
  "n12823531": "genus Convolvulus",
  "n12823717": "convolvulus",
  "n12823859": "bindweed",
  "n12824053": "field bindweed, wild morning-glory, Convolvulus arvensis",
  "n12824289": "scammony, Convolvulus scammonia",
  "n12824482": "scammony",
  "n12824581": "Argyreia, genus Argyreia",
  "n12824735": "silverweed",
  "n12824909": "Calystegia, genus Calystegia",
  "n12825061": "hedge bindweed, wild morning-glory, Calystegia sepium, Convolvulus sepium",
  "n12825301": "Cuscuta, genus Cuscuta",
  "n12825497": "dodder",
  "n12825721": "love vine, Cuscuta gronovii",
  "n12825949": "genus Dichondra",
  "n12826143": "dichondra, Dichondra micrantha",
  "n12826395": "Ipomoea, genus Ipomoea",
  "n12826516": "morning glory",
  "n12826895": "common morning glory, Ipomoea purpurea",
  "n12827068": "common morning glory, Ipomoea tricolor",
  "n12827270": "cypress vine, star-glory, Indian pink, Ipomoea quamoclit, Quamoclit pennata",
  "n12827537": "moonflower, belle de nuit, Ipomoea alba",
  "n12827684": "sweet potato, sweet potato vine, Ipomoea batatas",
  "n12827907": "wild potato vine, wild sweet potato vine, man-of-the-earth, manroot, scammonyroot, Ipomoea panurata, Ipomoea fastigiata",
  "n12828220": "red morning-glory, star ipomoea, Ipomoea coccinea",
  "n12828379": "man-of-the-earth, Ipomoea leptophylla",
  "n12828520": "scammony, Ipomoea orizabensis",
  "n12828628": "railroad vine, beach morning glory, Ipomoea pes-caprae",
  "n12828791": "Japanese morning glory, Ipomoea nil",
  "n12828977": "imperial Japanese morning glory, Ipomoea imperialis",
  "n12829099": "Gesneriaceae, family Gesneriaceae, gesneria family",
  "n12829582": "gesneriad",
  "n12829809": "genus Gesneria",
  "n12829975": "gesneria",
  "n12830080": "genus Achimenes",
  "n12830222": "achimenes, hot water plant",
  "n12830404": "genus Aeschynanthus",
  "n12830568": "aeschynanthus",
  "n12830789": "lipstick plant, Aeschynanthus radicans",
  "n12830974": "Alsobia, genus Alsobia",
  "n12831141": "lace-flower vine, Alsobia dianthiflora, Episcia dianthiflora",
  "n12831389": "genus Columnea",
  "n12831535": "columnea",
  "n12831776": "genus Episcia",
  "n12831932": "episcia",
  "n12832140": "genus Gloxinia",
  "n12832315": "gloxinia",
  "n12832538": "Canterbury bell, Gloxinia perennis",
  "n12832690": "genus Kohleria",
  "n12832822": "kohleria",
  "n12832976": "Saintpaulia, genus Saintpaulia",
  "n12833149": "African violet, Saintpaulia ionantha",
  "n12833341": "Sinningia, genus Sinningia",
  "n12833526": "florist's gloxinia, Sinningia speciosa, Gloxinia spesiosa",
  "n12833793": "genus Streptocarpus",
  "n12833985": "streptocarpus",
  "n12834190": "Cape primrose",
  "n12834408": "Hydrophyllaceae, family Hydrophyllaceae, waterleaf family",
  "n12834671": "Hydrophyllum, genus Hydrophyllum",
  "n12834798": "waterleaf",
  "n12834938": "Virginia waterleaf, Shawnee salad, shawny, Indian salad, John's cabbage, Hydrophyllum virginianum",
  "n12835196": "Emmanthe, genus Emmanthe",
  "n12835331": "yellow bells, California yellow bells, whispering bells, Emmanthe penduliflora",
  "n12835578": "Eriodictyon, genus Eriodictyon",
  "n12835766": "yerba santa, Eriodictyon californicum",
  "n12836033": "genus Nemophila",
  "n12836212": "nemophila",
  "n12836337": "baby blue-eyes, Nemophila menziesii",
  "n12836508": "five-spot, Nemophila maculata",
  "n12836663": "genus Phacelia",
  "n12836862": "scorpionweed, scorpion weed, phacelia",
  "n12837052": "California bluebell, Phacelia campanularia",
  "n12837259": "California bluebell, whitlavia, Phacelia minor, Phacelia whitlavia",
  "n12837466": "fiddleneck, Phacelia tanacetifolia",
  "n12837643": "Pholistoma, genus Pholistoma",
  "n12837803": "fiesta flower, Pholistoma auritum, Nemophila aurita",
  "n12838027": "Labiatae, family Labiatae, Lamiaceae, family Lamiaceae, mint family",
  "n12839319": "mint",
  "n12839409": "Acinos, genus Acinos",
  "n12839574": "basil thyme, basil balm, mother of thyme, Acinos arvensis, Satureja acinos",
  "n12839839": "Agastache, genus Agastache",
  "n12839979": "giant hyssop",
  "n12840168": "yellow giant hyssop, Agastache nepetoides",
  "n12840362": "anise hyssop, Agastache foeniculum",
  "n12840502": "Mexican hyssop, Agastache mexicana",
  "n12840640": "Ajuga, genus Ajuga",
  "n12840749": "bugle, bugleweed",
  "n12841007": "creeping bugle, Ajuga reptans",
  "n12841193": "erect bugle, blue bugle, Ajuga genevensis",
  "n12841354": "pyramid bugle, Ajuga pyramidalis",
  "n12841472": "ground pine, yellow bugle, Ajuga chamaepitys",
  "n12841686": "Ballota, genus Ballota",
  "n12841872": "black horehound, black archangel, fetid horehound, stinking horehound, Ballota nigra",
  "n12842105": "Blephilia, genus Blephilia",
  "n12842302": "wood mint",
  "n12842519": "hairy wood mint, Blephilia hirsuta",
  "n12842642": "downy wood mint, Blephilia celiata",
  "n12842765": "Calamintha, genus Calamintha",
  "n12842887": "calamint",
  "n12843144": "common calamint, Calamintha sylvatica, Satureja calamintha officinalis",
  "n12843316": "large-flowered calamint, Calamintha grandiflora, Clinopodium grandiflorum, Satureja grandiflora",
  "n12843557": "lesser calamint, field balm, Calamintha nepeta, Calamintha nepeta glantulosa, Satureja nepeta, Satureja calamintha glandulosa",
  "n12843844": "Clinopodium, genus Clinopodium",
  "n12843970": "wild basil, cushion calamint, Clinopodium vulgare, Satureja vulgaris",
  "n12844220": "Collinsonia, genus Collinsonia",
  "n12844409": "horse balm, horseweed, stoneroot, stone-root, richweed, stone root, Collinsonia canadensis",
  "n12844697": "genus Coleus",
  "n12844939": "coleus, flame nettle",
  "n12845187": "country borage, Coleus aromaticus, Coleus amboinicus, Plectranthus amboinicus",
  "n12845413": "painted nettle, Joseph's coat, Coleus blumei, Solenostemon blumei, Solenostemon scutellarioides",
  "n12845732": "Conradina, genus Conradina",
  "n12845908": "Apalachicola rosemary, Conradina glabra",
  "n12846143": "Dracocephalum, genus Dracocephalum",
  "n12846335": "dragonhead, dragon's head, Dracocephalum parviflorum",
  "n12846546": "genus Elsholtzia",
  "n12846690": "elsholtzia",
  "n12846869": "Galeopsis, genus Galeopsis",
  "n12847008": "hemp nettle, dead nettle, Galeopsis tetrahit",
  "n12847254": "Glechoma, genus Glechoma",
  "n12847374": "ground ivy, alehoof, field balm, gill-over-the-ground, runaway robin, Glechoma hederaceae, Nepeta hederaceae",
  "n12847749": "Hedeoma, genus Hedeoma",
  "n12847927": "pennyroyal, American pennyroyal, Hedeoma pulegioides",
  "n12848212": "pennyroyal oil, hedeoma oil",
  "n12848343": "Hyssopus, genus Hyssopus",
  "n12848499": "hyssop, Hyssopus officinalis",
  "n12848770": "hyssop oil",
  "n12848870": "Lamium, genus Lamium",
  "n12849061": "dead nettle",
  "n12849279": "white dead nettle, Lamium album",
  "n12849416": "henbit, Lamium amplexicaule",
  "n12849597": "Lavandula, genus Lavandula",
  "n12849717": "lavender",
  "n12849952": "English lavender, Lavandula angustifolia, Lavandula officinalis",
  "n12850168": "French lavender, Lavandula stoechas",
  "n12850336": "spike lavender, French lavender, Lavandula latifolia",
  "n12850531": "spike lavender oil, spike oil",
  "n12850718": "Leonotis, genus Leonotis",
  "n12850906": "dagga, Cape dagga, red dagga, wilde dagga, Leonotis leonurus",
  "n12851094": "lion's-ear, Leonotis nepetaefolia, Leonotis nepetifolia",
  "n12851304": "Leonurus, genus Leonurus",
  "n12851469": "motherwort, Leonurus cardiaca",
  "n12851673": "Lepechinia, genus Lepechinia, Sphacele, genus Sphacele",
  "n12851860": "pitcher sage, Lepechinia calycina, Sphacele calycina",
  "n12852049": "Lycopus, genus Lycopus",
  "n12852234": "bugleweed, Lycopus virginicus",
  "n12852428": "water horehound, Lycopus americanus",
  "n12852570": "gipsywort, gypsywort, Lycopus europaeus",
  "n12852726": "genus Origanum",
  "n12852930": "Majorana, genus Majorana",
  "n12853080": "origanum",
  "n12853287": "oregano, marjoram, pot marjoram, wild marjoram, winter sweet, Origanum vulgare",
  "n12853482": "sweet marjoram, knotted marjoram, Origanum majorana, Majorana hortensis",
  "n12853706": "dittany of crete, cretan dittany, crete dittany, hop marjoram, winter sweet, Origanum dictamnus",
  "n12853901": "Marrubium, genus Marrubium",
  "n12854048": "horehound",
  "n12854193": "common horehound, white horehound, Marrubium vulgare",
  "n12854443": "Melissa, genus Melissa",
  "n12854600": "lemon balm, garden balm, sweet balm, bee balm, beebalm, Melissa officinalis",
  "n12854925": "Mentha, genus Mentha",
  "n12855042": "mint",
  "n12855365": "corn mint, field mint, Mentha arvensis",
  "n12855494": "water-mint, water mint, Mentha aquatica",
  "n12855710": "bergamot mint, lemon mint, eau de cologne mint, Mentha citrata",
  "n12855886": "horsemint, Mentha longifolia",
  "n12856091": "peppermint, Mentha piperita",
  "n12856287": "spearmint, Mentha spicata",
  "n12856479": "apple mint, applemint, Mentha rotundifolia, Mentha suaveolens",
  "n12856680": "pennyroyal, Mentha pulegium",
  "n12856868": "pennyroyal oil",
  "n12857024": "Micromeria, genus Micromeria",
  "n12857204": "yerba buena, Micromeria chamissonis, Micromeria douglasii, Satureja douglasii",
  "n12857456": "savory, Micromeria juliana",
  "n12857594": "Molucella, genus Molucella",
  "n12857779": "molucca balm, bells of Ireland, Molucella laevis",
  "n12858019": "genus Monarda",
  "n12858150": "monarda, wild bergamot",
  "n12858397": "bee balm, beebalm, bergamot mint, oswego tea, Monarda didyma",
  "n12858618": "horsemint, Monarda punctata",
  "n12858871": "bee balm, beebalm, Monarda fistulosa",
  "n12858987": "lemon mint, horsemint, Monarda citriodora",
  "n12859153": "plains lemon monarda, Monarda pectinata",
  "n12859272": "basil balm, Monarda clinopodia",
  "n12859488": "Monardella, genus Monardella",
  "n12859679": "mustang mint, Monardella lanceolata",
  "n12859873": "Nepeta, genus Nepeta",
  "n12859986": "catmint, catnip, Nepeta cataria",
  "n12860254": "Ocimum, genus Ocimum",
  "n12860365": "basil",
  "n12860542": "common basil, sweet basil, Ocimum basilicum",
  "n12860842": "Perilla, genus Perilla",
  "n12860978": "beefsteak plant, Perilla frutescens crispa",
  "n12861139": "genus Phlomis",
  "n12861345": "phlomis",
  "n12861541": "Jerusalem sage, Phlomis fruticosa",
  "n12861751": "genus Physostegia",
  "n12861892": "physostegia",
  "n12862116": "false dragonhead, false dragon head, obedient plant, Physostegia virginiana",
  "n12862312": "genus Plectranthus",
  "n12862512": "plectranthus",
  "n12862648": "Pogostemon, genus Pogostemon",
  "n12862828": "patchouli, patchouly, pachouli, Pogostemon cablin",
  "n12863026": "Prunella, genus Prunella",
  "n12863234": "self-heal, heal all, Prunella vulgaris",
  "n12863458": "Pycnanthemum, genus Pycnanthemum, Koellia, genus Koellia",
  "n12863624": "mountain mint",
  "n12863819": "basil mint, Pycnanthemum virginianum",
  "n12864038": "Rosmarinus, genus Rosmarinus",
  "n12864160": "rosemary, Rosmarinus officinalis",
  "n12864363": "genus Salvia",
  "n12864545": "sage, salvia",
  "n12864902": "blue sage, Salvia azurea",
  "n12865037": "clary sage, Salvia clarea",
  "n12865239": "blue sage, mealy sage, Salvia farinacea",
  "n12865367": "blue sage, Salvia reflexa, Salvia lancifolia",
  "n12865562": "purple sage, chaparral sage, Salvia leucophylla",
  "n12865708": "cancerweed, cancer weed, Salvia lyrata",
  "n12865824": "common sage, ramona, Salvia officinalis",
  "n12866002": "meadow clary, Salvia pratensis",
  "n12866162": "clary, Salvia sclarea",
  "n12866333": "pitcher sage, Salvia spathacea",
  "n12866459": "Mexican mint, Salvia divinorum",
  "n12866635": "wild sage, wild clary, vervain sage, Salvia verbenaca",
  "n12866824": "Satureja, genus Satureja, Satureia, genus Satureia",
  "n12866968": "savory",
  "n12867184": "summer savory, Satureja hortensis, Satureia hortensis",
  "n12867449": "winter savory, Satureja montana, Satureia montana",
  "n12867679": "Scutellaria, genus Scutellaria",
  "n12867826": "skullcap, helmetflower",
  "n12868019": "blue pimpernel, blue skullcap, mad-dog skullcap, mad-dog weed, Scutellaria lateriflora",
  "n12868248": "Sideritis, genus Sideritis",
  "n12868418": "Solenostemon, genus Solenostemon",
  "n12868634": "Stachys, genus Stachys",
  "n12868880": "hedge nettle, dead nettle, Stachys sylvatica",
  "n12869061": "hedge nettle, Stachys palustris",
  "n12869248": "Teucrium, genus Teucrium",
  "n12869478": "germander",
  "n12869668": "American germander, wood sage, Teucrium canadense",
  "n12869874": "wall germander, Teucrium chamaedrys",
  "n12870048": "cat thyme, marum, Teucrium marum",
  "n12870225": "wood sage, Teucrium scorodonia",
  "n12870392": "Thymus, genus Thymus",
  "n12870535": "thyme",
  "n12870682": "common thyme, Thymus vulgaris",
  "n12870891": "wild thyme, creeping thyme, Thymus serpyllum",
  "n12871074": "Trichostema, genus Trichostema",
  "n12871272": "blue curls",
  "n12871484": "black sage, wooly blue curls, California romero, Trichostema lanatum",
  "n12871696": "turpentine camphor weed, camphorweed, vinegarweed, Trichostema lanceolatum",
  "n12871859": "bastard pennyroyal, Trichostema dichotomum",
  "n12871992": "Lentibulariaceae, family Lentibulariaceae, bladderwort family",
  "n12872257": "Utricularia, genus Utricularia",
  "n12872458": "bladderwort",
  "n12872698": "Pinguicula, genus Pinguicula",
  "n12872914": "butterwort",
  "n12873182": "genus Genlisea",
  "n12873341": "genlisea",
  "n12873646": "Martyniaceae, family Martyniaceae",
  "n12873834": "genus Martynia",
  "n12873984": "martynia, Martynia annua",
  "n12874231": "Orobanchaceae, family Orobanchaceae, broomrape family",
  "n12874429": "Pedaliaceae, family Pedaliaceae, sesame family",
  "n12874642": "Sesamum, genus Sesamum",
  "n12874783": "sesame, benne, benni, benny, Sesamum indicum",
  "n12874996": "Proboscidea, genus Proboscidea",
  "n12875269": "common unicorn plant, devil's claw, common devil's claw, elephant-tusk, proboscis flower, ram's horn, Proboscidea louisianica",
  "n12875594": "beak",
  "n12875697": "sand devil's claw, Proboscidea arenaria, Martynia arenaria",
  "n12875861": "sweet unicorn plant, Proboscidea fragrans, Martynia fragrans",
  "n12876032": "Scrophulariaceae, family Scrophulariaceae, figwort family, foxglove family",
  "n12876684": "Scrophularia, genus Scrophularia",
  "n12876899": "figwort",
  "n12877041": "Antirrhinum, genus Antirrhinum",
  "n12877244": "snapdragon",
  "n12877493": "white snapdragon, Antirrhinum coulterianum",
  "n12877637": "yellow twining snapdragon, Antirrhinum filipes",
  "n12877838": "Mediterranean snapdragon, Antirrhinum majus",
  "n12878019": "Besseya, genus Besseya",
  "n12878169": "kitten-tails",
  "n12878325": "Alpine besseya, Besseya alpina",
  "n12878525": "Aureolaria, genus Aureolaria",
  "n12878784": "false foxglove, Aureolaria pedicularia, Gerardia pedicularia",
  "n12879068": "false foxglove, Aureolaria virginica, Gerardia virginica",
  "n12879350": "genus Calceolaria",
  "n12879527": "calceolaria, slipperwort",
  "n12879719": "Castilleja, genus Castilleja, Castilleia, genus Castilleia",
  "n12879963": "Indian paintbrush, painted cup",
  "n12880244": "desert paintbrush, Castilleja chromosa",
  "n12880462": "giant red paintbrush, Castilleja miniata",
  "n12880638": "great plains paintbrush, Castilleja sessiliflora",
  "n12880799": "sulfur paintbrush, Castilleja sulphurea",
  "n12880963": "Chelone, genus Chelone",
  "n12881105": "shellflower, shell-flower, turtlehead, snakehead, snake-head, Chelone glabra",
  "n12881429": "Collinsia, genus Collinsia",
  "n12881631": "purple chinese houses, innocense, Collinsia bicolor, Collinsia heterophylla",
  "n12881913": "maiden blue-eyed Mary, Collinsia parviflora",
  "n12882158": "blue-eyed Mary, Collinsia verna",
  "n12882321": "Culver's root, Culvers root, Culver's physic, Culvers physic, whorlywort, Veronicastrum virginicum",
  "n12882591": "genus Digitalis",
  "n12882779": "foxglove, digitalis",
  "n12882945": "common foxglove, fairy bell, fingerflower, finger-flower, fingerroot, finger-root, Digitalis purpurea",
  "n12883265": "yellow foxglove, straw foxglove, Digitalis lutea",
  "n12883395": "genus Gerardia",
  "n12883628": "gerardia",
  "n12883733": "Agalinis, genus Agalinis",
  "n12883923": "Linaria, genus Linaria",
  "n12884100": "blue toadflax, old-field toadflax, Linaria canadensis",
  "n12884260": "toadflax, butter-and-eggs, wild snapdragon, devil's flax, Linaria vulgaris",
  "n12884523": "Penstemon, genus Penstemon",
  "n12885045": "golden-beard penstemon, Penstemon barbatus",
  "n12885265": "scarlet bugler, Penstemon centranthifolius",
  "n12885510": "red shrubby penstemon, redwood penstemon",
  "n12885754": "Platte River penstemon, Penstemon cyananthus",
  "n12885956": "Davidson's penstemon, Penstemon davidsonii",
  "n12886185": "hot-rock penstemon, Penstemon deustus",
  "n12886402": "Jones' penstemon, Penstemon dolius",
  "n12886600": "shrubby penstemon, lowbush penstemon, Penstemon fruticosus",
  "n12886831": "narrow-leaf penstemon, Penstemon linarioides",
  "n12887065": "mountain pride, Penstemon newberryi",
  "n12887293": "balloon flower, scented penstemon, Penstemon palmeri",
  "n12887532": "Parry's penstemon, Penstemon parryi",
  "n12887713": "rock penstemon, cliff penstemon, Penstemon rupicola",
  "n12888016": "Rydberg's penstemon, Penstemon rydbergii",
  "n12888234": "cascade penstemon, Penstemon serrulatus",
  "n12888457": "Whipple's penstemon, Penstemon whippleanus",
  "n12888733": "Verbascum, genus Verbascum",
  "n12888906": "mullein, flannel leaf, velvet plant",
  "n12889219": "moth mullein, Verbascum blattaria",
  "n12889412": "white mullein, Verbascum lychnitis",
  "n12889579": "purple mullein, Verbascum phoeniceum",
  "n12889713": "common mullein, great mullein, Aaron's rod, flannel mullein, woolly mullein, torch, Verbascum thapsus",
  "n12890009": "genus Veronica",
  "n12890265": "veronica, speedwell",
  "n12890490": "field speedwell, Veronica agrestis",
  "n12890685": "brooklime, American brooklime, Veronica americana",
  "n12890928": "corn speedwell, Veronica arvensis",
  "n12891093": "brooklime, European brooklime, Veronica beccabunga",
  "n12891305": "germander speedwell, bird's eye, Veronica chamaedrys",
  "n12891469": "water speedwell, Veronica michauxii, Veronica anagallis-aquatica",
  "n12891643": "common speedwell, gypsyweed, Veronica officinalis",
  "n12891824": "purslane speedwell, Veronica peregrina",
  "n12892013": "thyme-leaved speedwell, Veronica serpyllifolia",
  "n12892226": "Solanaceae, family Solanaceae, potato family",
  "n12893094": "Solanum, genus Solanum",
  "n12893463": "nightshade",
  "n12893794": "kangaroo apple, poroporo, Solanum aviculare",
  "n12893993": "horse nettle, ball nettle, bull nettle, ball nightshade, Solanum carolinense",
  "n12894247": "potato tree, Solanum crispum",
  "n12894438": "Uruguay potato, Uruguay potato vine, Solanum commersonii",
  "n12894607": "bittersweet, bittersweet nightshade, climbing nightshade, deadly nightshade, poisonous nightshade, woody nightshade, Solanum dulcamara",
  "n12894930": "trompillo, white horse nettle, prairie berry, purple nightshade, silverleaf nightshade, silver-leaved nightshade, silver-leaved nettle, Solanum elaeagnifolium",
  "n12895298": "African holly, Solanum giganteum",
  "n12895578": "wild potato, Solanum jamesii",
  "n12895811": "potato vine, Solanum jasmoides",
  "n12896000": "eggplant, aubergine, brinjal, eggplant bush, garden egg, mad apple, Solanum melongena",
  "n12896307": "black nightshade, common nightshade, poisonberry, poison-berry, Solanum nigrum",
  "n12896615": "garden huckleberry, wonderberry, sunberry, Solanum nigrum guineese, Solanum melanocerasum, Solanum burbankii",
  "n12896862": "Jerusalem cherry, winter cherry, Madeira winter cherry, Solanum pseudocapsicum",
  "n12897118": "naranjilla, Solanum quitoense",
  "n12897341": "buffalo bur, Solanum rostratum",
  "n12897493": "potato, white potato, white potato vine, Solanum tuberosum",
  "n12897788": "potato vine, giant potato creeper, Solanum wendlandii",
  "n12897999": "potato tree, Brazilian potato tree, Solanum wrightii, Solanum macranthum",
  "n12898226": "Atropa, genus Atropa",
  "n12898342": "belladonna, belladonna plant, deadly nightshade, Atropa belladonna",
  "n12898628": "genus Browallia",
  "n12898774": "bush violet, browallia",
  "n12898959": "Brunfelsia, genus Brunfelsia",
  "n12899166": "lady-of-the-night, Brunfelsia americana",
  "n12899333": "Brugmansia, genus Brugmansia",
  "n12899537": "angel's trumpet, maikoa, Brugmansia arborea, Datura arborea",
  "n12899752": "angel's trumpet, Brugmansia suaveolens, Datura suaveolens",
  "n12899971": "red angel's trumpet, Brugmansia sanguinea, Datura sanguinea",
  "n12900148": "genus Capsicum, Capsicum",
  "n12900462": "capsicum, pepper, capsicum pepper plant",
  "n12900783": "cone pepper, Capsicum annuum conoides",
  "n12900987": "cayenne, cayenne pepper, chili pepper, chilli pepper, long pepper, jalapeno, Capsicum annuum longum",
  "n12901264": "sweet pepper, bell pepper, pimento, pimiento, paprika, sweet pepper plant, Capsicum annuum grossum",
  "n12901565": "cherry pepper, Capsicum annuum cerasiforme",
  "n12901724": "bird pepper, Capsicum frutescens baccatum, Capsicum baccatum",
  "n12902021": "tabasco pepper, hot pepper, tabasco plant, Capsicum frutescens",
  "n12902297": "Cestrum, genus Cestrum",
  "n12902466": "day jessamine, Cestrum diurnum",
  "n12902662": "night jasmine, night jessamine, Cestrum nocturnum",
  "n12902887": "Cyphomandra, genus Cyphomandra",
  "n12903014": "tree tomato, tamarillo",
  "n12903250": "Datura, genus Datura",
  "n12903367": "thorn apple",
  "n12903503": "jimsonweed, jimson weed, Jamestown weed, common thorn apple, apple of Peru, Datura stramonium",
  "n12903794": "Fabiana, genus Fabiana",
  "n12903964": "pichi, Fabiana imbricata",
  "n12904148": "Hyoscyamus, genus Hyoscyamus",
  "n12904314": "henbane, black henbane, stinking nightshade, Hyoscyamus niger",
  "n12904562": "Egyptian henbane, Hyoscyamus muticus",
  "n12904720": "Lycium, genus Lycium",
  "n12904938": "matrimony vine, boxthorn",
  "n12905135": "common matrimony vine, Duke of Argyll's tea tree, Lycium barbarum, Lycium halimifolium",
  "n12905412": "Christmasberry, Christmas berry, Lycium carolinianum",
  "n12905655": "Lycopersicon, genus Lycopersicon, Lycopersicum, genus Lycopersicum",
  "n12905817": "tomato, love apple, tomato plant, Lycopersicon esculentum",
  "n12906021": "cherry tomato, Lycopersicon esculentum cerasiforme",
  "n12906214": "plum tomato",
  "n12906334": "Mandragora, genus Mandragora",
  "n12906498": "mandrake, devil's apples, Mandragora officinarum",
  "n12906771": "mandrake root, mandrake",
  "n12906926": "Nicandra, genus Nicandra",
  "n12907057": "apple of Peru, shoo fly, Nicandra physaloides",
  "n12907287": "Nicotiana, genus Nicotiana",
  "n12907465": "tobacco, tobacco plant",
  "n12907671": "flowering tobacco, Jasmine tobacco, Nicotiana alata",
  "n12907857": "common tobacco, Nicotiana tabacum",
  "n12908093": "wild tobacco, Indian tobacco, Nicotiana rustica",
  "n12908230": "tree tobacco, mustard tree, Nicotiana glauca",
  "n12908432": "genus Nierembergia",
  "n12908645": "cupflower, nierembergia",
  "n12908854": "whitecup, Nierembergia repens, Nierembergia rivularis",
  "n12909068": "tall cupflower, Nierembergia frutescens",
  "n12909252": "genus Petunia, Petunia",
  "n12909421": "petunia",
  "n12909614": "large white petunia, Petunia axillaris",
  "n12909759": "violet-flowered petunia, Petunia integrifolia",
  "n12909917": "hybrid petunia, Petunia hybrida",
  "n12910141": "Physalis, genus Physalis",
  "n12910285": "ground cherry, husk tomato",
  "n12910676": "downy ground cherry, strawberry tomato, Physalis pubescens",
  "n12910875": "Chinese lantern plant, winter cherry, bladder cherry, Physalis alkekengi",
  "n12911079": "cape gooseberry, purple ground cherry, Physalis peruviana",
  "n12911264": "strawberry tomato, dwarf cape gooseberry, Physalis pruinosa",
  "n12911440": "tomatillo, jamberry, Mexican husk tomato, Physalis ixocarpa",
  "n12911673": "tomatillo, miltomate, purple ground cherry, jamberry, Physalis philadelphica",
  "n12911914": "yellow henbane, Physalis viscosa",
  "n12912105": "Salpichroa, genus Salpichroa",
  "n12912274": "cock's eggs, Salpichroa organifolia, Salpichroa rhomboidea",
  "n12912498": "genus Salpiglossis",
  "n12912670": "salpiglossis",
  "n12912801": "painted tongue, Salpiglossis sinuata",
  "n12913004": "genus Schizanthus",
  "n12913144": "butterfly flower, poor man's orchid, schizanthus",
  "n12913352": "Scopolia, genus Scopolia",
  "n12913524": "Scopolia carniolica",
  "n12913645": "Solandra, genus Solandra",
  "n12913791": "chalice vine, trumpet flower, cupflower, Solandra guttata",
  "n12914048": "Streptosolen, genus Streptosolen",
  "n12914193": "marmalade bush, fire bush, fire-bush, Streptosolen jamesonii",
  "n12914433": "Verbenaceae, family Verbenaceae, verbena family, vervain family",
  "n12914731": "genus Verbena",
  "n12914923": "verbena, vervain",
  "n12915140": "lantana",
  "n12915230": "Avicennia, genus Avicennia",
  "n12915400": "Avicenniaceae, family Avicenniaceae",
  "n12915568": "black mangrove, Avicennia marina",
  "n12915811": "white mangrove, Avicennia officinalis",
  "n12916025": "Aegiceras, genus Aegiceras",
  "n12916179": "black mangrove, Aegiceras majus",
  "n12916356": "Tectona, genus Tectona",
  "n12916511": "teak, Tectona grandis",
  "n12916723": "teak, teakwood",
  "n12916935": "Euphorbiaceae, family Euphorbiaceae, spurge family",
  "n12917338": "Euphorbia, genus Euphorbia",
  "n12917901": "spurge",
  "n12918404": "caper spurge, myrtle spurge, mole plant, Euphorbia lathyris",
  "n12918609": "sun spurge, wartweed, wartwort, devil's milk, Euphorbia helioscopia",
  "n12918810": "petty spurge, devil's milk, Euphorbia peplus",
  "n12918991": "medusa's head, Euphorbia medusae, Euphorbia caput-medusae",
  "n12919195": "wild spurge, flowering spurge, tramp's spurge, Euphorbia corollata",
  "n12919403": "snow-on-the-mountain, snow-in-summer, ghost weed, Euphorbia marginata",
  "n12919646": "cypress spurge, Euphorbia cyparissias",
  "n12919847": "leafy spurge, wolf's milk, Euphorbia esula",
  "n12920043": "hairy spurge, Euphorbia hirsuta",
  "n12920204": "poinsettia, Christmas star, Christmas flower, lobster plant, Mexican flameleaf, painted leaf, Euphorbia pulcherrima",
  "n12920521": "Japanese poinsettia, mole plant, paint leaf, Euphorbia heterophylla",
  "n12920719": "fire-on-the-mountain, painted leaf, Mexican fire plant, Euphorbia cyathophora",
  "n12920955": "wood spurge, Euphorbia amygdaloides",
  "n12921126": "candelilla, Euphorbia antisyphilitica",
  "n12921315": "dwarf spurge, Euphorbia exigua",
  "n12921499": "scarlet plume, Euphorbia fulgens",
  "n12921660": "naboom, cactus euphorbia, Euphorbia ingens",
  "n12921868": "crown of thorns, Christ thorn, Christ plant, Euphorbia milii",
  "n12922119": "toothed spurge, Euphorbia dentata",
  "n12922283": "Acalypha, genus Acalypha",
  "n12922458": "three-seeded mercury, Acalypha virginica",
  "n12922600": "genus Croton",
  "n12922763": "croton, Croton tiglium",
  "n12922933": "croton oil",
  "n12923108": "cascarilla, Croton eluteria",
  "n12923257": "cascarilla bark, eleuthera bark, sweetwood bark",
  "n12923439": "Codiaeum, genus Codiaeum",
  "n12923652": "croton, Codiaeum variegatum",
  "n12923839": "Mercurialis, genus Mercurialis",
  "n12924036": "herb mercury, herbs mercury, boys-and-girls, Mercurialis annua",
  "n12924284": "dog's mercury, dog mercury, Mercurialis perennis",
  "n12924452": "Ricinus, genus Ricinus",
  "n12924623": "castor-oil plant, castor bean plant, palma christi, palma christ, Ricinus communis",
  "n12924984": "Cnidoscolus, genus Cnidoscolus",
  "n12925179": "spurge nettle, tread-softly, devil nettle, pica-pica, Cnidoscolus urens, Jatropha urens, Jatropha stimulosus",
  "n12925394": "Jatropha, genus Jatropha",
  "n12925583": "physic nut, Jatropha curcus",
  "n12925836": "Hevea, rubber tree, genus Hevea",
  "n12926039": "Para rubber tree, caoutchouc tree, Hevea brasiliensis",
  "n12926316": "Manihot, genus Manihot",
  "n12926480": "cassava, casava",
  "n12926689": "bitter cassava, manioc, mandioc, mandioca, tapioca plant, gari, Manihot esculenta, Manihot utilissima",
  "n12927013": "cassava, manioc",
  "n12927194": "sweet cassava, Manihot dulcis",
  "n12927354": "Aleurites, genus Aleurites",
  "n12927494": "candlenut, varnish tree, Aleurites moluccana",
  "n12927758": "tung tree, tung, tung-oil tree, Aleurites fordii",
  "n12927921": "Pedilanthus, genus Pedilanthus",
  "n12928071": "slipper spurge, slipper plant",
  "n12928307": "candelilla, Pedilanthus bracteatus, Pedilanthus pavonis",
  "n12928491": "Jewbush, Jew-bush, Jew bush, redbird cactus, redbird flower, Pedilanthus tithymaloides",
  "n12928690": "Sebastiana, genus Sebastiana",
  "n12928819": "jumping bean, jumping seed, Mexican jumping bean",
  "n12929061": "Theaceae, family Theaceae, tea family",
  "n12929237": "genus Camellia",
  "n12929403": "camellia, camelia",
  "n12929600": "japonica, Camellia japonica",
  "n12929783": "tea, Camellia sinensis",
  "n12930044": "Umbelliferae, family Umbelliferae, Apiaceae, family Apiaceae, carrot family",
  "n12930778": "umbellifer, umbelliferous plant",
  "n12930951": "wild parsley",
  "n12931109": "Aethusa, genus Aethusa",
  "n12931231": "fool's parsley, lesser hemlock, Aethusa cynapium",
  "n12931449": "Anethum, genus Anethum",
  "n12931542": "dill, Anethum graveolens",
  "n12931738": "genus Angelica",
  "n12931906": "angelica, angelique",
  "n12932173": "garden angelica, archangel, Angelica Archangelica",
  "n12932365": "wild angelica, Angelica sylvestris",
  "n12932532": "Anthriscus, genus Anthriscus",
  "n12932706": "chervil, beaked parsley, Anthriscus cereifolium",
  "n12932966": "cow parsley, wild chervil, Anthriscus sylvestris",
  "n12933164": "Apium, genus Apium",
  "n12933274": "wild celery, Apium graveolens",
  "n12933403": "celery, cultivated celery, Apium graveolens dulce",
  "n12933616": "celeriac, celery root, knob celery, root celery, turnip-rooted celery, Apium graveolens rapaceum",
  "n12933827": "genus Astrantia",
  "n12934036": "astrantia, masterwort",
  "n12934174": "greater masterwort, Astrantia major",
  "n12934368": "Carum, genus Carum",
  "n12934479": "caraway, Carum carvi",
  "n12934685": "whorled caraway",
  "n12934776": "Cicuta, genus Cicuta",
  "n12934985": "water hemlock, Cicuta verosa",
  "n12935166": "spotted cowbane, spotted hemlock, spotted water hemlock",
  "n12935457": "Conium, genus Conium",
  "n12935609": "hemlock, poison hemlock, poison parsley, California fern, Nebraska fern, winter fern, Conium maculatum",
  "n12935982": "Conopodium, genus Conopodium",
  "n12936155": "earthnut, Conopodium denudatum",
  "n12936333": "Coriandrum, genus Coriandrum",
  "n12936469": "coriander, coriander plant, Chinese parsley, cilantro, Coriandrum sativum",
  "n12936713": "Cuminum, genus Cuminum",
  "n12936826": "cumin, Cuminum cyminum",
  "n12936999": "Daucus, genus Daucus",
  "n12937130": "wild carrot, Queen Anne's lace, Daucus carota",
  "n12937388": "carrot, cultivated carrot, Daucus carota sativa",
  "n12937678": "carrot",
  "n12937822": "Eryngium, genus Eryngium",
  "n12938081": "eryngo, eringo",
  "n12938193": "sea holly, sea holm, sea eryngium, Eryngium maritimum",
  "n12938445": "button snakeroot, Eryngium aquaticum",
  "n12938667": "rattlesnake master, rattlesnake's master, button snakeroot, Eryngium yuccifolium",
  "n12938897": "Foeniculum, genus Foeniculum",
  "n12939104": "fennel",
  "n12939282": "common fennel, Foeniculum vulgare",
  "n12939479": "Florence fennel, Foeniculum dulce, Foeniculum vulgare dulce",
  "n12939664": "Heracleum, genus Heracleum",
  "n12939874": "cow parsnip, hogweed, Heracleum sphondylium",
  "n12940060": "Levisticum, genus Levisticum",
  "n12940226": "lovage, Levisticum officinale",
  "n12940427": "Myrrhis, genus Myrrhis",
  "n12940609": "sweet cicely, Myrrhis odorata",
  "n12940778": "Oenanthe, genus Oenanthe",
  "n12940939": "water dropwort, hemlock water dropwort, Oenanthe crocata",
  "n12941220": "water fennel, Oenanthe aquatica",
  "n12941360": "Pastinaca, genus Pastinaca",
  "n12941536": "parsnip, Pastinaca sativa",
  "n12941717": "cultivated parsnip",
  "n12941914": "parsnip",
  "n12942025": "wild parsnip, madnep",
  "n12942270": "Petroselinum, genus Petroselinum",
  "n12942395": "parsley, Petroselinum crispum",
  "n12942572": "Italian parsley, flat-leaf parsley, Petroselinum crispum neapolitanum",
  "n12942729": "Hamburg parsley, turnip-rooted parsley, Petroselinum crispum tuberosum",
  "n12942930": "Pimpinella, genus Pimpinella",
  "n12943049": "anise, anise plant, Pimpinella anisum",
  "n12943302": "Sanicula, genus Sanicula",
  "n12943443": "sanicle, snakeroot",
  "n12943743": "footsteps-of-spring, Sanicula arctopoides",
  "n12943912": "purple sanicle, Sanicula bipinnatifida",
  "n12944095": "European sanicle, Sanicula Europaea",
  "n12944238": "Seseli, genus Seseli",
  "n12944390": "moon carrot, stone parsley",
  "n12944590": "Sison, genus Sison",
  "n12944723": "stone parsley, Sison amomum",
  "n12944960": "Sium, genus Sium",
  "n12945177": "water parsnip, Sium suave",
  "n12945366": "greater water parsnip, Sium latifolium",
  "n12945549": "skirret, Sium sisarum",
  "n12945708": "Smyrnium, genus Smyrnium",
  "n12945828": "Alexander, Alexanders, black lovage, horse parsley, Smyrnium olusatrum",
  "n12946088": "Cornaceae, family Cornaceae, dogwood family",
  "n12946432": "Aucuba, genus Aucuba",
  "n12946578": "Cornus, genus Cornus",
  "n12946849": "dogwood, dogwood tree, cornel",
  "n12947171": "dogwood",
  "n12947313": "common white dogwood, eastern flowering dogwood, Cornus florida",
  "n12947544": "red osier, red osier dogwood, red dogwood, American dogwood, redbrush, Cornus stolonifera",
  "n12947756": "silky dogwood, Cornus obliqua",
  "n12947895": "silky cornel, silky dogwood, Cornus amomum",
  "n12948053": "common European dogwood, red dogwood, blood-twig, pedwood, Cornus sanguinea",
  "n12948251": "bunchberry, dwarf cornel, crackerberry, pudding berry, Cornus canadensis",
  "n12948495": "cornelian cherry, Cornus mas",
  "n12948633": "Corokia, genus Corokia",
  "n12948849": "Curtisia, genus Curtisia",
  "n12948978": "Griselinia, genus Griselinia",
  "n12949160": "puka, Griselinia lucida",
  "n12949361": "kapuka, Griselinia littoralis",
  "n12949549": "Helwingia, genus Helwingia",
  "n12949722": "Valerianaceae, family Valerianaceae, valerian family",
  "n12949955": "Valeriana, genus Valeriana",
  "n12950126": "valerian",
  "n12950314": "common valerian, garden heliotrope, Valeriana officinalis",
  "n12950501": "Valerianella, genus Valerianella",
  "n12950669": "corn salad",
  "n12950796": "common corn salad, lamb's lettuce, Valerianella olitoria, Valerianella locusta",
  "n12950984": "Centranthus, genus Centranthus",
  "n12951146": "red valerian, French honeysuckle, Centranthus ruber",
  "n12951331": "cutch, kutch",
  "n12951465": "Hymenophyllaceae, family Hymenophyllaceae",
  "n12951668": "Hymenophyllum, genus Hymenophyllum",
  "n12951835": "filmy fern, film fern",
  "n12952022": "Trichomanes, genus Trichomanes",
  "n12952165": "bristle fern, filmy fern",
  "n12952469": "hare's-foot bristle fern, Trichomanes boschianum",
  "n12952590": "Killarney fern, Trichomanes speciosum",
  "n12952717": "kidney fern, Trichomanes reniforme",
  "n12952852": "Osmundaceae, family Osmundaceae",
  "n12953079": "genus Osmunda",
  "n12953206": "flowering fern, osmund",
  "n12953484": "royal fern, royal osmund, king fern, ditch fern, French bracken, Osmunda regalis",
  "n12953712": "interrupted fern, Osmunda clatonia",
  "n12953919": "cinnamon fern, fiddlehead, fiddlehead fern, Osmunda cinnamonea",
  "n12954185": "Leptopteris, genus Leptopteris",
  "n12954353": "crape fern, Prince-of-Wales fern, Prince-of-Wales feather, Prince-of-Wales plume, Leptopteris superba, Todea superba",
  "n12954634": "Todea, genus Todea",
  "n12954799": "crepe fern, king fern, Todea barbara",
  "n12954978": "Schizaeaceae, family Schizaeaceae",
  "n12955191": "Schizaea, genus Schizaea",
  "n12955414": "curly grass, curly grass fern, Schizaea pusilla",
  "n12955639": "Anemia, genus Anemia",
  "n12955840": "pine fern, Anemia adiantifolia",
  "n12956029": "Lygodium, genus Lygodium",
  "n12956170": "climbing fern",
  "n12956367": "creeping fern, Hartford fern, Lygodium palmatum",
  "n12956588": "climbing maidenhair, climbing maidenhair fern, snake fern, Lygodium microphyllum",
  "n12956791": "Mohria, genus Mohria",
  "n12956922": "scented fern, Mohria caffrorum",
  "n12957076": "aquatic fern, water fern",
  "n12957298": "Marsileaceae, family Marsileaceae",
  "n12957467": "Marsilea, genus Marsilea",
  "n12957608": "clover fern, pepperwort",
  "n12957803": "nardoo, nardo, common nardoo, Marsilea drummondii",
  "n12957924": "water clover, Marsilea quadrifolia",
  "n12958140": "Pilularia, genus Pilularia",
  "n12958261": "pillwort, Pilularia globulifera",
  "n12958470": "genus Regnellidium",
  "n12958615": "regnellidium, Regnellidium diphyllum",
  "n12958772": "Salviniaceae, family Salviniaceae",
  "n12958921": "Salvinia, genus Salvinia",
  "n12959074": "floating-moss, Salvinia rotundifolia, Salvinia auriculata",
  "n12959226": "Azollaceae, family Azollaceae",
  "n12959371": "Azolla, genus Azolla",
  "n12959538": "mosquito fern, floating fern, Carolina pond fern, Azolla caroliniana",
  "n12959802": "Ophioglossales, order Ophioglossales",
  "n12959967": "Ophioglossaceae, family Ophioglossaceae",
  "n12960211": "Ophioglossum, genus Ophioglossum",
  "n12960378": "adder's tongue, adder's tongue fern",
  "n12960552": "ribbon fern, Ophioglossum pendulum",
  "n12960729": "Botrychium, genus Botrychium",
  "n12960863": "grape fern",
  "n12961112": "moonwort, common moonwort, Botrychium lunaria",
  "n12961242": "daisyleaf grape fern, daisy-leaved grape fern, Botrychium matricariifolium",
  "n12961393": "leathery grape fern, Botrychium multifidum",
  "n12961536": "rattlesnake fern, Botrychium virginianum",
  "n12961689": "Helminthostachys, genus Helminthostachys",
  "n12961879": "flowering fern, Helminthostachys zeylanica",
  "n12962059": "soldier grainy club",
  "n12962149": "ostiole",
  "n12962310": "perithecium",
  "n12962389": "stroma",
  "n12962488": "stroma",
  "n12962622": "plastid",
  "n12962847": "chromoplast",
  "n12962992": "chloroplast",
  "n12963140": "Erysiphales, order Erysiphales",
  "n12963307": "Erysiphaceae, family Erysiphaceae",
  "n12963494": "Erysiphe, genus Erysiphe",
  "n12963628": "powdery mildew",
  "n12963796": "Sphaeriales, order Sphaeriales",
  "n12964130": "Sphaeriaceae, family Sphaeriaceae",
  "n12964321": "Neurospora, genus Neurospora",
  "n12964572": "Ceratostomataceae, family Ceratostomataceae",
  "n12964750": "Ceratostomella, genus Ceratostomella",
  "n12964920": "Dutch elm fungus, Ceratostomella ulmi",
  "n12965056": "Hypocreales, order Hypocreales",
  "n12965209": "Hypocreaceae, family Hypocreaceae",
  "n12965463": "Claviceps, genus Claviceps",
  "n12965626": "ergot, Claviceps purpurea",
  "n12965951": "rye ergot",
  "n12966048": "mushroom pimple",
  "n12966193": "orange mushroom pimple",
  "n12966290": "green mushroom pimple",
  "n12966386": "Xylariaceae, family Xylariaceae",
  "n12966581": "Xylaria, genus Xylaria",
  "n12966804": "black root rot fungus, Xylaria mali",
  "n12966945": "dead-man's-fingers, dead-men's-fingers, Xylaria polymorpha",
  "n12967124": "Rosellinia, genus Rosellinia",
  "n12967281": "Helotiales, order Helotiales",
  "n12967504": "Helotiaceae, family Helotiaceae",
  "n12967656": "Helotium, genus Helotium",
  "n12967776": "Sclerotiniaceae, family Sclerotiniaceae",
  "n12967955": "genus Sclerotinia",
  "n12968136": "sclerotinia",
  "n12968309": "brown cup",
  "n12968408": "Sclerodermatales, order Sclerodermatales",
  "n12968658": "Sclerodermataceae, family Sclerodermataceae",
  "n12968882": "Scleroderma, genus Scleroderma",
  "n12969131": "earthball, false truffle, puffball, hard-skinned puffball",
  "n12969425": "Scleroderma citrinum, Scleroderma aurantium",
  "n12969670": "Scleroderma flavidium, star earthball",
  "n12969927": "Scleroderma bovista, smooth earthball",
  "n12970193": "Podaxaceae",
  "n12970293": "stalked puffball",
  "n12970379": "Tulostomaceae, family Tulostomaceae, Tulostomataceae, family Tulostomataceae",
  "n12970560": "Tulostoma, genus Tulostoma, Tulestoma, genus Tulestoma",
  "n12970733": "stalked puffball",
  "n12970872": "Hymenogastrales, order Hymenogastrales",
  "n12971157": "Rhizopogonaceae, family Rhizopogonaceae",
  "n12971400": "false truffle",
  "n12971624": "Rhizopogon, genus Rhizopogon",
  "n12971804": "Rhizopogon idahoensis",
  "n12971956": "Truncocolumella, genus Truncocolumella",
  "n12972136": "Truncocolumella citrina",
  "n12972414": "Zygomycota, subdivision Zygomycota, Zygomycotina, subdivision Zygomycotina",
  "n12972629": "Zygomycetes, class Zygomycetes",
  "n12972818": "Mucorales, order Mucorales",
  "n12972966": "Mucoraceae, family Mucoraceae",
  "n12973202": "genus Mucor",
  "n12973443": "mucor",
  "n12973541": "genus Rhizopus",
  "n12973791": "rhizopus",
  "n12973937": "bread mold, Rhizopus nigricans",
  "n12974062": "leak fungus, ring rot fungus, Rhizopus stolonifer",
  "n12974286": "Entomophthorales, order Entomophthorales",
  "n12974457": "Entomophthoraceae, family Entomophthoraceae",
  "n12974662": "Entomophthora, genus Entomophthora",
  "n12974826": "rhizoid",
  "n12974987": "slime mold, slime mould",
  "n12975207": "Myxomycota, division Myxomycota, Gymnomycota, division Gymnomycota",
  "n12975608": "Myxomycetes, class Myxomycetes",
  "n12975804": "true slime mold, acellular slime mold, plasmodial slime mold, myxomycete",
  "n12975982": "Acrasiomycetes, class Acrasiomycetes",
  "n12976198": "cellular slime mold",
  "n12976389": "genus Dictostylium",
  "n12976554": "dictostylium",
  "n12976672": "Phycomycetes, Phycomycetes group",
  "n12976985": "Mastigomycota, subdivision Mastigomycota, Mastigomycotina, subdivision Mastigomycotina",
  "n12977296": "Oomycetes, class Oomycetes",
  "n12977565": "Chytridiomycetes, class Chytridiomycetes",
  "n12977795": "Chytridiales, order Chytridiales",
  "n12978076": "pond-scum parasite",
  "n12978232": "Chytridiaceae, family Chytridiaceae",
  "n12978381": "Blastocladiales, order Blastocladiales",
  "n12978654": "Blastodiaceae, family Blastodiaceae",
  "n12978826": "Blastocladia, genus Blastocladia",
  "n12978969": "Synchytriaceae, family Synchytriaceae",
  "n12979129": "Synchytrium, genus Synchytrium",
  "n12979316": "potato wart fungus, Synchytrium endobioticum",
  "n12979478": "Saprolegniales, order Saprolegniales",
  "n12979630": "Saprolegnia, genus Saprolegnia",
  "n12979829": "white fungus, Saprolegnia ferax",
  "n12980080": "water mold",
  "n12980231": "Peronosporales, order Peronosporales",
  "n12980478": "Peronosporaceae, family Peronosporaceae",
  "n12980652": "Peronospora, genus Peronospora",
  "n12980840": "downy mildew, false mildew",
  "n12981086": "blue mold fungus, Peronospora tabacina",
  "n12981301": "onion mildew, Peronospora destructor",
  "n12981443": "tobacco mildew, Peronospora hyoscyami",
  "n12981595": "Albuginaceae, family Albuginaceae",
  "n12981791": "Albugo, genus Albugo",
  "n12981954": "white rust",
  "n12982103": "Pythiaceae, family Pythiaceae",
  "n12982338": "genus Pythium",
  "n12982468": "pythium",
  "n12982590": "damping off fungus, Pythium debaryanum",
  "n12982723": "Phytophthora, genus Phytophthora",
  "n12982915": "Phytophthora citrophthora",
  "n12983048": "Phytophthora infestans",
  "n12983217": "Plasmodiophoraceae, family Plasmodiophoraceae",
  "n12983404": "Plasmodiophora, genus Plasmodiophora",
  "n12983654": "clubroot fungus, Plasmodiophora brassicae",
  "n12983873": "Geglossaceae",
  "n12983961": "Sarcosomataceae",
  "n12984160": "black felt cup",
  "n12984267": "Rufous rubber cup",
  "n12984377": "charred pancake cup",
  "n12984489": "devil's cigar",
  "n12984595": "devil's urn",
  "n12984699": "winter urn",
  "n12984802": "Tuberales, order Tuberales",
  "n12985010": "Tuberaceae, family Tuberaceae",
  "n12985236": "Tuber, genus Tuber",
  "n12985420": "truffle, earthnut, earth-ball",
  "n12985629": "Clavariaceae, family Clavariaceae",
  "n12985773": "club fungus",
  "n12985857": "coral fungus",
  "n12986084": "Hydnaceae, family Hydnaceae",
  "n12986227": "tooth fungus",
  "n12986337": "Hydnum, genus Hydnum",
  "n12986447": "Lichenes, division Lichenes",
  "n12986858": "Lichenales, order Lichenales",
  "n12987056": "lichen",
  "n12987423": "ascolichen",
  "n12987535": "basidiolichen",
  "n12987652": "Lechanorales, order Lechanorales",
  "n12987834": "Lecanoraceae, family Lecanoraceae",
  "n12987993": "genus Lecanora",
  "n12988158": "lecanora",
  "n12988341": "manna lichen",
  "n12988572": "archil, orchil",
  "n12988703": "Roccellaceae, family Roccellaceae",
  "n12988858": "genus Roccella",
  "n12989007": "roccella, Roccella tinctoria",
  "n12989142": "Pertusariaceae, family Pertusariaceae",
  "n12989301": "Pertusaria, genus Pertusaria",
  "n12989462": "Usneaceae, family Usneaceae",
  "n12989739": "Usnea, genus Usnea",
  "n12989938": "beard lichen, beard moss, Usnea barbata",
  "n12990092": "Evernia, genus Evernia",
  "n12990250": "Ramalina, genus Ramalina",
  "n12990407": "Alectoria, genus Alectoria",
  "n12990597": "horsehair lichen, horsetail lichen",
  "n12990800": "Cladoniaceae, family Cladoniaceae",
  "n12990938": "Cladonia, genus Cladonia",
  "n12991184": "reindeer moss, reindeer lichen, arctic moss, Cladonia rangiferina",
  "n12991488": "Parmeliaceae, family Parmeliaceae",
  "n12991645": "Parmelia, genus Parmelia",
  "n12991837": "crottle, crottal, crotal",
  "n12992022": "Cetraria, genus Cetraria",
  "n12992177": "Iceland moss, Iceland lichen, Cetraria islandica",
  "n12992464": "Fungi, kingdom Fungi, fungus kingdom",
  "n12992868": "fungus",
  "n12994475": "basidium",
  "n12994711": "hypobasidium",
  "n12994892": "promycelium",
  "n12994979": "Eumycota, division Eumycota",
  "n12995435": "Eumycetes, class Eumycetes",
  "n12995601": "true fungus",
  "n12995724": "Deuteromycota, subdivision Deuteromycota, Deuteromycotina, Fungi imperfecti, subdivision Deuteromycotina",
  "n12996068": "Deuteromycetes, class Deuteromycetes",
  "n12996225": "Basidiomycota, subdivision Basidiomycota, Basidiomycotina, subdivision Basidiomycotina",
  "n12996841": "Basidiomycetes, class Basidiomycetes",
  "n12997128": "Homobasidiomycetes, subclass Homobasidiomycetes",
  "n12997432": "Heterobasidiomycetes, subclass Heterobasidiomycetes",
  "n12997654": "basidiomycete, basidiomycetous fungi",
  "n12997919": "mushroom",
  "n12998130": "Hymenomycetes, class Hymenomycetes",
  "n12998349": "Agaricales, order Agaricales",
  "n12998815": "agaric",
  "n13000372": "Agaricaceae, family Agaricaceae",
  "n13000668": "Agaricus, genus Agaricus",
  "n13000891": "mushroom",
  "n13001041": "mushroom",
  "n13001206": "toadstool",
  "n13001366": "horse mushroom, Agaricus arvensis",
  "n13001529": "meadow mushroom, field mushroom, Agaricus campestris",
  "n13001743": "Lentinus, genus Lentinus",
  "n13001930": "shiitake, shiitake mushroom, Chinese black mushroom, golden oak mushroom, Oriental black mushroom, Lentinus edodes",
  "n13002209": "scaly lentinus, Lentinus lepideus",
  "n13002433": "Amanita, genus Amanita",
  "n13002750": "royal agaric, Caesar's agaric, Amanita caesarea",
  "n13002925": "false deathcap, Amanita mappa",
  "n13003061": "fly agaric, Amanita muscaria",
  "n13003254": "death cap, death cup, death angel, destroying angel, Amanita phalloides",
  "n13003522": "blushing mushroom, blusher, Amanita rubescens",
  "n13003712": "destroying angel, Amanita verna",
  "n13003846": "slime mushroom",
  "n13003974": "white slime mushroom",
  "n13004065": "Fischer's slime mushroom",
  "n13004160": "Cantharellus, genus Cantharellus",
  "n13004423": "chanterelle, chantarelle, Cantharellus cibarius",
  "n13004640": "floccose chanterelle, Cantharellus floccosus",
  "n13004826": "pig's ears, Cantharellus clavatus",
  "n13004992": "cinnabar chanterelle, Cantharellus cinnabarinus",
  "n13005166": "Omphalotus, genus Omphalotus",
  "n13005329": "jack-o-lantern fungus, jack-o-lantern, jack-a-lantern, Omphalotus illudens",
  "n13005568": "Coprinus, genus Coprinus",
  "n13005835": "Coprinaceae, family Coprinaceae",
  "n13005984": "inky cap, inky-cap mushroom, Coprinus atramentarius",
  "n13006171": "shaggymane, shaggy cap, shaggymane mushroom, Coprinus comatus",
  "n13006377": "Lactarius, genus Lactarius",
  "n13006631": "milkcap, Lactarius delicioso",
  "n13006741": "Marasmius, genus Marasmius",
  "n13006894": "fairy-ring mushroom, Marasmius oreades",
  "n13007034": "fairy ring, fairy circle",
  "n13007195": "Pleurotus, genus Pleurotus",
  "n13007417": "oyster mushroom, oyster fungus, oyster agaric, Pleurotus ostreatus",
  "n13007629": "olive-tree agaric, Pleurotus phosphoreus",
  "n13007770": "Pholiota, genus Pholiota",
  "n13008157": "Pholiota astragalina",
  "n13008315": "Pholiota aurea, golden pholiota",
  "n13008485": "Pholiota destruens",
  "n13008689": "Pholiota flammans",
  "n13008839": "Pholiota flavida",
  "n13009085": "nameko, viscid mushroom, Pholiota nameko",
  "n13009244": "Pholiota squarrosa-adiposa",
  "n13009429": "Pholiota squarrosa, scaly pholiota",
  "n13009656": "Pholiota squarrosoides",
  "n13009780": "Russula, genus Russula",
  "n13010064": "Russulaceae, family Russulaceae",
  "n13010219": "Strophariaceae, family Strophariaceae",
  "n13010401": "Stropharia, genus Stropharia, ring-stalked fungus",
  "n13010694": "Stropharia ambigua",
  "n13010951": "Stropharia hornemannii",
  "n13011221": "Stropharia rugoso-annulata",
  "n13011461": "galea",
  "n13011595": "gill fungus",
  "n13011679": "gill, lamella",
  "n13011856": "Entolomataceae, family Entolomataceae",
  "n13012030": "Entoloma, genus Entoloma",
  "n13012253": "Entoloma lividum, Entoloma sinuatum",
  "n13012469": "Entoloma aprile",
  "n13012613": "Lepiotaceae, family Lepiotaceae",
  "n13012835": "genus Chlorophyllum",
  "n13012973": "Chlorophyllum molybdites",
  "n13013187": "genus Lepiota",
  "n13013534": "lepiota",
  "n13013764": "parasol mushroom, Lepiota procera",
  "n13013965": "poisonous parasol, Lepiota morgani",
  "n13014097": "Lepiota naucina",
  "n13014265": "Lepiota rhacodes",
  "n13014409": "American parasol, Lepiota americana",
  "n13014581": "Lepiota rubrotincta",
  "n13014741": "Lepiota clypeolaria",
  "n13014879": "onion stem, Lepiota cepaestipes",
  "n13015040": "Thelephoraceae, family Thelephoraceae",
  "n13015229": "Corticium, genus Corticium",
  "n13015509": "pink disease fungus, Corticium salmonicolor",
  "n13015688": "bottom rot fungus, Corticium solani",
  "n13015826": "Pellicularia, genus Pellicularia",
  "n13016076": "potato fungus, Pellicularia filamentosa, Rhizoctinia solani",
  "n13016289": "coffee fungus, Pellicularia koleroga",
  "n13016457": "Tricholomataceae, family Tricholomataceae",
  "n13016749": "Tricholoma, genus Tricholoma",
  "n13017102": "blewits, Clitocybe nuda",
  "n13017240": "sandy mushroom, Tricholoma populinum",
  "n13017439": "Tricholoma pessundatum",
  "n13017610": "Tricholoma sejunctum",
  "n13017789": "man-on-a-horse, Tricholoma flavovirens",
  "n13017979": "Tricholoma venenata",
  "n13018088": "Tricholoma pardinum",
  "n13018232": "Tricholoma vaccinum",
  "n13018407": "Tricholoma aurantium",
  "n13018579": "Volvariaceae, family Volvariaceae",
  "n13018749": "Volvaria, genus Volvaria",
  "n13018906": "Volvaria bombycina",
  "n13019017": "Pluteaceae, family Pluteaceae",
  "n13019202": "Pluteus, genus Pluteus, roof mushroom",
  "n13019496": "Pluteus aurantiorugosus",
  "n13019643": "Pluteus magnus, sawdust mushroom",
  "n13019835": "deer mushroom, Pluteus cervinus",
  "n13020011": "Volvariella, genus Volvariella",
  "n13020191": "straw mushroom, Chinese mushroom, Volvariella volvacea",
  "n13020481": "Volvariella bombycina",
  "n13020623": "Clitocybe, genus Clitocybe",
  "n13020964": "Clitocybe clavipes",
  "n13021166": "Clitocybe dealbata",
  "n13021332": "Clitocybe inornata",
  "n13021543": "Clitocybe robusta, Clytocybe alba",
  "n13021689": "Clitocybe irina, Tricholoma irinum, Lepista irina",
  "n13021867": "Clitocybe subconnexa",
  "n13022078": "Flammulina, genus Flammulina",
  "n13022210": "winter mushroom, Flammulina velutipes",
  "n13022538": "hypha",
  "n13022709": "mycelium",
  "n13022903": "sclerotium",
  "n13023134": "sac fungus",
  "n13023292": "Ascomycota, subdivision Ascomycota, Ascomycotina, subdivision Ascomycotina",
  "n13023783": "Ascomycetes, class Ascomycetes",
  "n13024012": "ascomycete, ascomycetous fungus",
  "n13024348": "Euascomycetes, subclass Euascomycetes",
  "n13024500": "Clavicipitaceae, grainy club mushrooms",
  "n13024653": "grainy club",
  "n13024763": "Hemiascomycetes, class Hemiascomycetes",
  "n13024967": "Endomycetales, order Endomycetales",
  "n13025197": "Saccharomycetaceae, family Saccharomycetaceae",
  "n13025421": "Saccharomyces, genus Saccharomyces",
  "n13025647": "yeast",
  "n13025854": "baker's yeast, brewer's yeast, Saccharomyces cerevisiae",
  "n13026015": "wine-maker's yeast, Saccharomyces ellipsoides",
  "n13026146": "Schizosaccharomycetaceae, family Schizosaccharomycetaceae",
  "n13026339": "Schizosaccharomyces, genus Schizosaccharomyces",
  "n13026529": "Plectomycetes, class Plectomycetes",
  "n13026763": "Eurotiales, order Eurotiales, Aspergillales, order Aspergillales",
  "n13027049": "Eurotium, genus Eurotium",
  "n13027190": "Aspergillaceae, family Aspergillaceae",
  "n13027375": "Aspergillus, genus Aspergillus",
  "n13027557": "Aspergillus fumigatus",
  "n13027670": "Thielavia, genus Thielavia",
  "n13027879": "brown root rot fungus, Thielavia basicola",
  "n13028070": "Pyrenomycetes, class Pyrenomycetes",
  "n13028337": "Discomycetes, subclass Discomycetes",
  "n13028611": "discomycete, cup fungus",
  "n13028937": "Leotia lubrica",
  "n13029122": "Mitrula elegans",
  "n13029326": "Sarcoscypha coccinea, scarlet cup",
  "n13029610": "Caloscypha fulgens",
  "n13029760": "Aleuria aurantia, orange peel fungus",
  "n13029946": "Pezizales, order Pezizales",
  "n13030157": "Pezizaceae, family Pezizaceae",
  "n13030337": "elf cup",
  "n13030438": "Peziza, genus Peziza",
  "n13030616": "Peziza domicilina",
  "n13030852": "blood cup, fairy cup, Peziza coccinea",
  "n13031007": "Plectania, genus Plectania",
  "n13031193": "Urnula craterium, urn fungus",
  "n13031323": "Galiella rufa",
  "n13031474": "Jafnea semitosta",
  "n13031690": "Morchellaceae, family Morchellaceae",
  "n13031956": "Morchella, genus Morchella",
  "n13032115": "morel",
  "n13032381": "common morel, Morchella esculenta, sponge mushroom, sponge morel",
  "n13032618": "Disciotis venosa, cup morel",
  "n13032923": "Verpa, bell morel",
  "n13033134": "Verpa bohemica, early morel",
  "n13033396": "Verpa conica, conic Verpa",
  "n13033577": "black morel, Morchella conica, conic morel, Morchella angusticeps, narrowhead morel",
  "n13033879": "Morchella crassipes, thick-footed morel",
  "n13034062": "Morchella semilibera, half-free morel, cow's head",
  "n13034277": "Sarcoscyphaceae, family Sarcoscyphaceae",
  "n13034431": "Wynnea, genus Wynnea",
  "n13034555": "Wynnea americana",
  "n13034788": "Wynnea sparassoides",
  "n13034953": "Helvellaceae, family Helvellaceae",
  "n13035241": "false morel",
  "n13035389": "lorchel",
  "n13035521": "genus Helvella",
  "n13035707": "helvella",
  "n13035925": "Helvella crispa, miter mushroom",
  "n13036116": "Helvella acetabulum",
  "n13036312": "Helvella sulcata",
  "n13036592": "genus Discina",
  "n13036804": "discina",
  "n13036907": "Discina macrospora",
  "n13037124": "genus Gyromitra",
  "n13037406": "gyromitra",
  "n13037585": "Gyromitra californica, California false morel",
  "n13037805": "Gyromitra sphaerospora, round-spored gyromitra",
  "n13038068": "Gyromitra esculenta, brain mushroom, beefsteak morel",
  "n13038376": "Gyromitra infula, saddled-shaped false morel",
  "n13038577": "Gyromitra fastigiata, Gyromitra brunnea",
  "n13038744": "Gyromitra gigas",
  "n13038944": "Gasteromycetes, class Gasteromycetes, Gastromycetes, class Gastromycetes",
  "n13039349": "gasteromycete, gastromycete",
  "n13039553": "Phallales, order Phallales",
  "n13039870": "Phallaceae, family Phallaceae",
  "n13040108": "Phallus, genus Phallus",
  "n13040303": "stinkhorn, carrion fungus",
  "n13040629": "common stinkhorn, Phallus impudicus",
  "n13040796": "Phallus ravenelii",
  "n13040971": "Dictyophera, genus Dictyophera",
  "n13041172": "Mutinus, genus Mutinus",
  "n13041312": "dog stinkhorn, Mutinus caninus",
  "n13041548": "Tulostomatales, order Tulostomatales",
  "n13041725": "Calostomataceae, family Calostomataceae",
  "n13041943": "Calostoma lutescens",
  "n13042134": "Calostoma cinnabarina",
  "n13042316": "Calostoma ravenelii",
  "n13042514": "Clathraceae, family Clathraceae",
  "n13042694": "Clathrus, genus Clathrus",
  "n13042814": "Pseudocolus, genus Pseudocolus",
  "n13042982": "stinky squid, Pseudocolus fusiformis",
  "n13043264": "Lycoperdales, order Lycoperdales",
  "n13043516": "Lycoperdaceae, family Lycoperdaceae",
  "n13043746": "Lycoperdon, genus Lycoperdon",
  "n13043926": "puffball, true puffball",
  "n13044149": "Calvatia, genus Calvatia",
  "n13044375": "giant puffball, Calvatia gigantea",
  "n13044541": "Geastraceae, family Geastraceae",
  "n13044778": "earthstar",
  "n13045027": "Geastrum, genus Geastrum",
  "n13045210": "Geastrum coronatum",
  "n13045429": "Radiigera, genus Radiigera",
  "n13045594": "Radiigera fuscogleba",
  "n13045834": "Astreus, genus Astreus",
  "n13045975": "Astreus pteridis",
  "n13046130": "Astreus hygrometricus",
  "n13046285": "Nidulariales, order Nidulariales",
  "n13046512": "Nidulariaceae, family Nidulariaceae",
  "n13046669": "bird's-nest fungus",
  "n13046887": "Nidularia, genus Nidularia",
  "n13047011": "Sphaerobolaceae, family Sphaerobolaceae",
  "n13047216": "Secotiales, order Secotiales",
  "n13047385": "Secotiaceae, family Secotiaceae",
  "n13047706": "Gastrocybe, genus Gastrocybe",
  "n13047862": "Gastrocybe lateritia",
  "n13048212": "Macowanites, genus Macowanites",
  "n13048447": "Macowanites americanus",
  "n13048666": "Gastroboletus, genus Gastroboletus",
  "n13048932": "Gastroboletus scabrosus",
  "n13049105": "Gastroboletus turbinatus",
  "n13049285": "Aphyllophorales, order Aphyllophorales",
  "n13049561": "Polyporaceae, family Polyporaceae",
  "n13049953": "polypore, pore fungus, pore mushroom",
  "n13050397": "bracket fungus, shelf fungus",
  "n13050555": "Albatrellus, genus Albatrellus",
  "n13050705": "Albatrellus dispansus",
  "n13050940": "Albatrellus ovinus, sheep polypore",
  "n13051196": "Neolentinus, genus Neolentinus",
  "n13051346": "Neolentinus ponderosus",
  "n13051546": "Nigroporus, genus Nigroporus",
  "n13051694": "Nigroporus vinosus",
  "n13051866": "Oligoporus, genus Oligoporus",
  "n13052014": "Oligoporus leucospongia",
  "n13052248": "Polyporus tenuiculus",
  "n13052431": "Polyporus, genus Polyporus",
  "n13052670": "hen-of-the-woods, hen of the woods, Polyporus frondosus, Grifola frondosa",
  "n13052931": "Polyporus squamosus, scaly polypore",
  "n13053187": "Fistulinaceae, family Fistulinaceae",
  "n13053450": "Fistulina, genus Fistulina",
  "n13053608": "beefsteak fungus, Fistulina hepatica",
  "n13053816": "Fomes, genus Fomes",
  "n13054073": "agaric, Fomes igniarius",
  "n13054211": "Boletaceae, family Boletaceae",
  "n13054560": "bolete",
  "n13055009": "Boletus, genus Boletus",
  "n13055423": "Boletus chrysenteron",
  "n13055577": "Boletus edulis",
  "n13055792": "Frost's bolete, Boletus frostii",
  "n13055949": "Boletus luridus",
  "n13056135": "Boletus mirabilis",
  "n13056349": "Boletus pallidus",
  "n13056607": "Boletus pulcherrimus",
  "n13056799": "Boletus pulverulentus",
  "n13057054": "Boletus roxanae",
  "n13057242": "Boletus subvelutipes",
  "n13057422": "Boletus variipes",
  "n13057639": "Boletus zelleri",
  "n13057845": "Fuscoboletinus, genus Fuscoboletinus",
  "n13058037": "Fuscoboletinus paluster",
  "n13058272": "Fuscoboletinus serotinus",
  "n13058447": "Leccinum, genus Leccinum",
  "n13058608": "Leccinum fibrillosum",
  "n13058796": "Phylloporus, genus Phylloporus",
  "n13058963": "Phylloporus boletinoides",
  "n13059139": "Suillus, genus Suillus",
  "n13059298": "Suillus albivelatus",
  "n13059485": "Strobilomyces, genus Strobilomyces",
  "n13059657": "old-man-of-the-woods, Strobilomyces floccopus",
  "n13059852": "Boletellus, genus Boletellus",
  "n13060017": "Boletellus russellii",
  "n13060190": "jelly fungus",
  "n13060451": "Tremellales, order Tremellales",
  "n13060689": "Tremellaceae, family Tremellaceae",
  "n13060912": "Tremella, genus Tremella",
  "n13061172": "snow mushroom, Tremella fuciformis",
  "n13061348": "witches' butter, Tremella lutescens",
  "n13061471": "Tremella foliacea",
  "n13061704": "Tremella reticulata",
  "n13061921": "Auriculariales, order Auriculariales",
  "n13062112": "Auriculariaceae, family Auriculariaceae",
  "n13062272": "Auricularia, genus Auricularia",
  "n13062421": "Jew's-ear, Jew's-ears, ear fungus, Auricularia auricula",
  "n13062630": "Dacrymycetaceae, family Dacrymycetaceae",
  "n13062868": "Dacrymyces, genus Dacrymyces",
  "n13063046": "Uredinales, order Uredinales",
  "n13063269": "rust, rust fungus",
  "n13063514": "aecium",
  "n13063666": "aeciospore",
  "n13063784": "Melampsoraceae, family Melampsoraceae",
  "n13063936": "Melampsora, genus Melampsora",
  "n13064111": "flax rust, flax rust fungus, Melampsora lini",
  "n13064247": "Cronartium, genus Cronartium",
  "n13064457": "blister rust, Cronartium ribicola",
  "n13064678": "Pucciniaceae, family Pucciniaceae",
  "n13064852": "Puccinia, genus Puccinia",
  "n13065089": "wheat rust, Puccinia graminis",
  "n13065215": "Gymnosporangium, genus Gymnosporangium",
  "n13065514": "apple rust, cedar-apple rust, Gymnosporangium juniperi-virginianae",
  "n13065702": "Tiliomycetes, class Tiliomycetes",
  "n13065902": "Ustilaginales, order Ustilaginales",
  "n13066129": "smut, smut fungus",
  "n13066448": "covered smut",
  "n13066631": "Ustilaginaceae, family Ustilaginaceae",
  "n13066803": "Ustilago, genus Ustilago",
  "n13066979": "loose smut",
  "n13067191": "cornsmut, corn smut",
  "n13067330": "boil smut, Ustilago maydis",
  "n13067532": "Sphacelotheca, genus Sphacelotheca",
  "n13067672": "head smut, Sphacelotheca reiliana",
  "n13067845": "Tilletiaceae, family Tilletiaceae",
  "n13068073": "Tilletia, genus Tilletia",
  "n13068255": "bunt, Tilletia caries",
  "n13068434": "bunt, stinking smut, Tilletia foetida",
  "n13068565": "Urocystis, genus Urocystis",
  "n13068735": "onion smut, Urocystis cepulae",
  "n13068917": "flag smut fungus",
  "n13069224": "wheat flag smut, Urocystis tritici",
  "n13069348": "Septobasidiaceae, family Septobasidiaceae",
  "n13069535": "Septobasidium, genus Septobasidium",
  "n13069773": "felt fungus, Septobasidium pseudopedicellatum",
  "n13070003": "Hygrophoraceae, family Hygrophoraceae",
  "n13070308": "waxycap",
  "n13070708": "Hygrocybe, genus Hygrocybe",
  "n13070875": "Hygrocybe acutoconica, conic waxycap",
  "n13071029": "Hygrophorus, genus Hygrophorus",
  "n13071371": "Hygrophorus borealis",
  "n13071553": "Hygrophorus caeruleus",
  "n13071815": "Hygrophorus inocybiformis",
  "n13072031": "Hygrophorus kauffmanii",
  "n13072209": "Hygrophorus marzuolus",
  "n13072350": "Hygrophorus purpurascens",
  "n13072528": "Hygrophorus russula",
  "n13072706": "Hygrophorus sordidus",
  "n13072863": "Hygrophorus tennesseensis",
  "n13073055": "Hygrophorus turundus",
  "n13073211": "Hygrotrama, genus Hygrotrama",
  "n13073380": "Hygrotrama foetens",
  "n13073526": "Neohygrophorus, genus Neohygrophorus",
  "n13073703": "Neohygrophorus angelesianus",
  "n13073979": "cortina",
  "n13074084": "Cortinariaceae, family Cortinariaceae",
  "n13074277": "Cortinarius, genus Cortinarius",
  "n13074619": "Cortinarius armillatus",
  "n13074814": "Cortinarius atkinsonianus",
  "n13075020": "Cortinarius corrugatus",
  "n13075272": "Cortinarius gentilis",
  "n13075441": "Cortinarius mutabilis, purple-staining Cortinarius",
  "n13075684": "Cortinarius semisanguineus",
  "n13075847": "Cortinarius subfoetidus",
  "n13076041": "Cortinarius violaceus",
  "n13076181": "Gymnopilus, genus Gymnopilus",
  "n13076405": "Gymnopilus spectabilis",
  "n13076643": "Gymnopilus validipes",
  "n13076831": "Gymnopilus ventricosus",
  "n13077033": "mold, mould",
  "n13077295": "mildew",
  "n13077479": "Moniliales, order Moniliales",
  "n13077811": "genus Verticillium",
  "n13078021": "verticillium",
  "n13078133": "Moniliaceae, family Moniliaceae",
  "n13078483": "Trichophyton, genus Trichophyton",
  "n13078652": "Microsporum, genus Microsporum",
  "n13078809": "genus Monilia",
  "n13079073": "monilia",
  "n13079203": "genus Candida",
  "n13079419": "candida",
  "n13079567": "Candida albicans, Monilia albicans",
  "n13079775": "Cercosporella, genus Cercosporella",
  "n13079953": "Penicillium, genus Penicillium",
  "n13080174": "Blastomyces, genus Blastomyces",
  "n13080306": "blastomycete",
  "n13080471": "Dematiaceae, family Dematiaceae",
  "n13080674": "Cercospora, genus Cercospora",
  "n13080866": "yellow spot fungus, Cercospora kopkei",
  "n13081050": "Ustilaginoidea, genus Ustilaginoidea",
  "n13081229": "green smut fungus, Ustilaginoidea virens",
  "n13081369": "Tuberculariaceae, family Tuberculariaceae",
  "n13081565": "Tubercularia, genus Tubercularia",
  "n13081778": "genus Fusarium",
  "n13081999": "dry rot",
  "n13082077": "Mycelia Sterilia, order Mycelia Sterilia",
  "n13082293": "genus Rhizoctinia, form genus Rhizoctinia",
  "n13082568": "rhizoctinia",
  "n13082711": "Ozonium, genus Ozonium",
  "n13082829": "Sclerotium, genus Sclerotium",
  "n13083023": "houseplant",
  "n13083306": "garden plant",
  "n13083461": "bedder, bedding plant",
  "n13083586": "vascular plant, tracheophyte",
  "n13084184": "succulent",
  "n13084479": "nonvascular organism",
  "n13084633": "relict",
  "n13084834": "cultivar",
  "n13084993": "cultivated plant",
  "n13085113": "weed",
  "n13085747": "wort",
  "n13085864": "crop, harvest",
  "n13086063": "cash crop",
  "n13086220": "catch crop",
  "n13086438": "cover crop",
  "n13086556": "field crop",
  "n13086753": "fruitage",
  "n13086908": "plant part, plant structure",
  "n13087625": "plant organ",
  "n13088096": "plant process, enation",
  "n13088460": "apophysis",
  "n13088688": "callus",
  "n13088858": "blister",
  "n13088989": "nodule, tubercle",
  "n13089135": "spur",
  "n13089246": "fruiting body",
  "n13089419": "aculeus",
  "n13089526": "acumen",
  "n13089631": "spine, thorn, prickle, pricker, sticker, spikelet",
  "n13089902": "glochidium, glochid",
  "n13090018": "brier",
  "n13090091": "hair, fuzz, tomentum",
  "n13090248": "stinging hair",
  "n13090395": "coma",
  "n13090594": "beard",
  "n13090743": "awn",
  "n13090871": "aril",
  "n13091057": "duct",
  "n13091227": "laticifer",
  "n13091312": "antheridium",
  "n13091492": "antheridiophore",
  "n13091620": "sporophyll, sporophyl",
  "n13091774": "sporangium, spore case, spore sac",
  "n13091982": "sporangiophore",
  "n13092078": "ascus",
  "n13092240": "ascospore",
  "n13092385": "arthrospore",
  "n13092548": "arthrospore",
  "n13092722": "theca, sac",
  "n13092836": "paraphysis",
  "n13092987": "eusporangium",
  "n13093102": "leptosporangium",
  "n13093275": "tetrasporangium",
  "n13093380": "sporophore",
  "n13093629": "gametangium",
  "n13093725": "gametoecium",
  "n13093852": "gynoecium",
  "n13093928": "androecium",
  "n13094003": "gametophore",
  "n13094145": "sorus",
  "n13094273": "sorus",
  "n13094381": "indusium",
  "n13094556": "veil, velum",
  "n13094732": "universal veil",
  "n13095013": "partial veil",
  "n13095348": "annulus, skirt",
  "n13095543": "antherozoid, spermatozoid",
  "n13095685": "plant tissue",
  "n13096035": "pulp, flesh",
  "n13096159": "pith",
  "n13096317": "parenchyma",
  "n13096677": "chlorenchyma",
  "n13096779": "lignum",
  "n13096863": "vascular tissue",
  "n13097187": "stele",
  "n13097338": "cambium",
  "n13097536": "sapwood",
  "n13097752": "heartwood, duramen",
  "n13097949": "vascular bundle, vascular strand, fibrovascular bundle",
  "n13098186": "vein, nervure",
  "n13098420": "midrib, midvein",
  "n13098515": "vascular ray, medullary ray",
  "n13098648": "xylem",
  "n13098853": "tracheid",
  "n13098962": "phloem, bast",
  "n13099194": "sieve tube",
  "n13099373": "pseudophloem",
  "n13099444": "bast, bast fiber",
  "n13099586": "gall",
  "n13099734": "oak apple",
  "n13099833": "evergreen, evergreen plant",
  "n13099999": "deciduous plant",
  "n13100156": "poisonous plant",
  "n13100677": "vine",
  "n13102409": "climber",
  "n13102648": "creeper",
  "n13102775": "tendril",
  "n13102946": "cirrus, cirrhus",
  "n13103023": "root climber",
  "n13103136": "woody plant, ligneous plant",
  "n13103660": "lignosae",
  "n13103750": "arborescent plant",
  "n13103877": "snag",
  "n13104059": "tree",
  "n13107694": "timber tree",
  "n13107807": "treelet",
  "n13107891": "arbor",
  "n13108131": "bean tree",
  "n13108323": "pollard",
  "n13108481": "sapling",
  "n13108545": "shade tree",
  "n13108662": "gymnospermous tree",
  "n13108841": "conifer, coniferous tree",
  "n13109733": "angiospermous tree, flowering tree",
  "n13110915": "nut tree",
  "n13111174": "spice tree",
  "n13111340": "fever tree",
  "n13111504": "stump, tree stump",
  "n13111700": "stool",
  "n13111881": "bonsai",
  "n13112035": "ming tree",
  "n13112201": "ming tree",
  "n13112296": "groundcover, ground cover",
  "n13112427": "groundcover, ground cover",
  "n13112664": "shrub, bush",
  "n13118330": "undershrub",
  "n13118398": "burning bush",
  "n13118569": "shrublet",
  "n13118707": "subshrub, suffrutex",
  "n13119870": "bramble",
  "n13120003": "flowering shrub",
  "n13120211": "liana",
  "n13120446": "parasitic plant",
  "n13120775": "hemiparasite, semiparasite",
  "n13120958": "geophyte",
  "n13121104": "desert plant, xerophyte, xerophytic plant, xerophile, xerophilous plant",
  "n13121349": "mesophyte, mesophytic plant",
  "n13121544": "aquatic plant, water plant, hydrophyte, hydrophytic plant",
  "n13122364": "marsh plant, bog plant, swamp plant",
  "n13122985": "air plant, epiphyte, aerophyte, epiphytic plant",
  "n13123309": "hemiepiphyte, semiepiphyte",
  "n13123431": "strangler, strangler tree",
  "n13123681": "rock plant",
  "n13123841": "lithophyte, lithophytic plant",
  "n13124019": "rupestral plant, rupestrine plant, rupicolous plant, saxicolous plant",
  "n13124164": "saprophyte, saprophytic organism",
  "n13124358": "saprobe",
  "n13124529": "katharobe",
  "n13124654": "autophyte, autophytic plant, autotroph, autotrophic organism",
  "n13124855": "butt",
  "n13125003": "rootage, root system",
  "n13125117": "root",
  "n13125882": "pneumatophore",
  "n13126050": "taproot",
  "n13126192": "adventitious root",
  "n13126308": "root crop",
  "n13126428": "root cap",
  "n13126580": "rootlet",
  "n13126684": "root hair",
  "n13126856": "prop root",
  "n13127001": "prophyll",
  "n13127091": "stock",
  "n13127303": "rootstock",
  "n13127473": "cutting, slip",
  "n13127666": "quickset",
  "n13127843": "stolon, runner, offset",
  "n13128003": "crown, treetop",
  "n13128156": "capitulum",
  "n13128278": "tuberous plant",
  "n13128365": "tuber",
  "n13128582": "rhizome, rootstock, rootstalk",
  "n13128771": "axis",
  "n13128976": "rachis",
  "n13129078": "caudex",
  "n13129165": "stalk, stem",
  "n13129826": "internode",
  "n13129938": "beanstalk",
  "n13130014": "cladode, cladophyll, phylloclad, phylloclade",
  "n13130161": "receptacle",
  "n13130305": "stock, caudex",
  "n13130423": "axil",
  "n13130569": "stipe",
  "n13130726": "scape, flower stalk",
  "n13130885": "meristem",
  "n13131028": "umbel",
  "n13131282": "corymb",
  "n13131516": "ray",
  "n13131618": "petiole, leafstalk",
  "n13131752": "phyllode",
  "n13131883": "blade, leaf blade",
  "n13132034": "peduncle",
  "n13132156": "pedicel, pedicle",
  "n13132338": "flower cluster",
  "n13132486": "raceme",
  "n13132656": "panicle",
  "n13132756": "thyrse, thyrsus",
  "n13132940": "cyme",
  "n13133140": "cymule",
  "n13133233": "glomerule",
  "n13133316": "scorpioid cyme",
  "n13133438": "spike",
  "n13133613": "ear, spike, capitulum",
  "n13133786": "capitulum, head",
  "n13133932": "spadix",
  "n13134059": "bulb",
  "n13134302": "bulbous plant",
  "n13134531": "bulbil, bulblet",
  "n13134680": "corm",
  "n13134844": "cormous plant",
  "n13134947": "fruit",
  "n13135692": "fruitlet",
  "n13135832": "seed",
  "n13136316": "bean",
  "n13136556": "nut",
  "n13136781": "nutlet",
  "n13136863": "pyrene",
  "n13137010": "kernel, meat",
  "n13137225": "syconium",
  "n13137409": "berry",
  "n13137672": "aggregate fruit, multiple fruit, syncarp",
  "n13137951": "simple fruit, bacca",
  "n13138155": "acinus",
  "n13138308": "drupe, stone fruit",
  "n13138658": "drupelet",
  "n13138842": "pome, false fruit",
  "n13139055": "pod, seedpod",
  "n13139321": "loment",
  "n13139482": "pyxidium, pyxis",
  "n13139647": "husk",
  "n13139837": "cornhusk",
  "n13139918": "hull",
  "n13140049": "pod, cod, seedcase",
  "n13140242": "pea pod, peasecod",
  "n13140367": "accessory fruit, pseudocarp",
  "n13140535": "Rhamnales, order Rhamnales",
  "n13140699": "Rhamnaceae, family Rhamnaceae, buckthorn family",
  "n13140993": "Rhamnus, genus Rhamnus",
  "n13141141": "buckthorn",
  "n13141415": "buckthorn berry, yellow berry",
  "n13141564": "cascara buckthorn, bearberry, bearwood, chittamwood, chittimwood, Rhamnus purshianus",
  "n13141797": "cascara, cascara sagrada, chittam bark, chittem bark",
  "n13141972": "Carolina buckthorn, indian cherry, Rhamnus carolinianus",
  "n13142182": "coffeeberry, California buckthorn, California coffee, Rhamnus californicus",
  "n13142380": "alder buckthorn, alder dogwood, Rhamnus frangula",
  "n13142504": "redberry, red-berry, Rhamnus croceus",
  "n13142695": "Colubrina, genus Colubrina",
  "n13142907": "nakedwood",
  "n13143097": "Ziziphus, genus Ziziphus",
  "n13143285": "jujube, jujube bush, Christ's-thorn, Jerusalem thorn, Ziziphus jujuba",
  "n13143483": "lotus tree, Ziziphus lotus",
  "n13143626": "Paliurus, genus Paliurus",
  "n13143758": "Christ's-thorn, Jerusalem thorn, Paliurus spina-christi",
  "n13143930": "Pomaderris, genus Pomaderris",
  "n13144084": "hazel, hazel tree, Pomaderris apetala",
  "n13144303": "Vitaceae, family Vitaceae, Vitidaceae, grapevine family",
  "n13144511": "Vitis, genus Vitis",
  "n13144794": "grape, grapevine, grape vine",
  "n13145040": "fox grape, Vitis labrusca",
  "n13145250": "muscadine, Vitis rotundifolia",
  "n13145444": "vinifera, vinifera grape, common grape vine, Vitis vinifera",
  "n13145924": "Chardonnay, chardonnay grape",
  "n13146035": "Pinot, Pinot grape",
  "n13146225": "Pinot noir",
  "n13146403": "Pinot blanc",
  "n13146583": "Sauvignon grape",
  "n13146740": "Cabernet Sauvignon grape",
  "n13146928": "Sauvignon blanc",
  "n13147045": "Merlot",
  "n13147153": "Muscadet",
  "n13147270": "Riesling",
  "n13147386": "Zinfandel",
  "n13147532": "Chenin blanc",
  "n13147689": "malvasia",
  "n13147788": "muscat, muskat",
  "n13147918": "Verdicchio",
  "n13148019": "Parthenocissus, genus Parthenocissus",
  "n13148208": "Boston ivy, Japanese ivy, Parthenocissus tricuspidata",
  "n13148384": "Virginia creeper, American ivy, woodbine, Parthenocissus quinquefolia",
  "n13148602": "Piperales, order Piperales",
  "n13148791": "Piperaceae, family Piperaceae, pepper family",
  "n13149039": "Piper, genus Piper",
  "n13149296": "true pepper, pepper vine",
  "n13149506": "pepper, common pepper, black pepper, white pepper, Madagascar pepper, Piper nigrum",
  "n13149829": "long pepper, Piper longum",
  "n13149970": "betel, betel pepper, Piper betel",
  "n13150178": "cubeb, cubeb vine, Java pepper, Piper cubeba",
  "n13150378": "cubeb",
  "n13150592": "schizocarp",
  "n13150741": "genus Peperomia",
  "n13150894": "peperomia",
  "n13151082": "watermelon begonia, Peperomia argyreia, Peperomia sandersii",
  "n13151265": "Chloranthaceae, family Chloranthaceae",
  "n13151439": "Chloranthus, genus Chloranthus",
  "n13151568": "Saururaceae, family Saururaceae, lizard's-tail family",
  "n13151820": "Saururus, genus Saururus",
  "n13151975": "lizard's-tail, swamp lily, water dragon, Saururus cernuus",
  "n13152203": "Anemopsis, genus Anemopsis",
  "n13152339": "yerba mansa, Anemopsis californica",
  "n13152592": "Houttuynia, genus Houttuynia",
  "n13152742": "leaf, leafage, foliage",
  "n13153532": "amplexicaul leaf",
  "n13153633": "greenery, verdure",
  "n13153729": "hydathode, water pore, water stoma",
  "n13153888": "lenticel",
  "n13154077": "leaflet",
  "n13154190": "node, leaf node",
  "n13154388": "pinna, pinnule",
  "n13154494": "frond",
  "n13154586": "pad",
  "n13154736": "lily pad",
  "n13154841": "bract",
  "n13155095": "bracteole, bractlet",
  "n13155175": "spathe",
  "n13155305": "involucre",
  "n13155451": "lemma, flowering glume",
  "n13155611": "glume",
  "n13155759": "scale, scale leaf",
  "n13155899": "squamule",
  "n13156006": "fig leaf",
  "n13156083": "simple leaf",
  "n13156592": "compound leaf",
  "n13156799": "trifoliolate leaf",
  "n13156884": "quinquefoliate leaf",
  "n13156986": "palmate leaf",
  "n13157137": "pinnate leaf",
  "n13157346": "bijugate leaf, bijugous leaf, twice-pinnate",
  "n13157481": "decompound leaf",
  "n13157595": "acerate leaf, needle",
  "n13157684": "acuminate leaf",
  "n13157780": "cordate leaf",
  "n13157858": "cuneate leaf",
  "n13157971": "deltoid leaf",
  "n13158071": "elliptic leaf",
  "n13158167": "ensiform leaf",
  "n13158258": "hastate leaf",
  "n13158386": "lanceolate leaf",
  "n13158512": "linear leaf, elongate leaf",
  "n13158605": "lyrate leaf",
  "n13158714": "obtuse leaf",
  "n13158815": "oblanceolate leaf",
  "n13158925": "oblong leaf",
  "n13159044": "obovate leaf",
  "n13159155": "ovate leaf",
  "n13159263": "orbiculate leaf",
  "n13159357": "pandurate leaf, panduriform leaf",
  "n13159457": "peltate leaf",
  "n13159558": "perfoliate leaf",
  "n13159691": "reniform leaf",
  "n13159778": "sagittate-leaf, sagittiform leaf",
  "n13159890": "spatulate leaf",
  "n13160015": "bipinnate leaf",
  "n13160116": "even-pinnate leaf, abruptly-pinnate leaf",
  "n13160254": "odd-pinnate leaf",
  "n13160365": "pedate leaf",
  "n13160485": "entire leaf",
  "n13160604": "crenate leaf",
  "n13160695": "serrate leaf",
  "n13160831": "dentate leaf",
  "n13160938": "denticulate leaf",
  "n13161056": "emarginate leaf",
  "n13161151": "erose leaf",
  "n13161254": "runcinate leaf",
  "n13161412": "lobed leaf",
  "n13161506": "lobe",
  "n13161639": "parallel-veined leaf",
  "n13161754": "parted leaf",
  "n13161904": "prickly-edged leaf",
  "n13161998": "rosette",
  "n13162164": "ligule",
  "n13162297": "bark",
  "n13162815": "winter's bark",
  "n13162945": "tapa, tapa bark, tappa, tappa bark",
  "n13163093": "angostura bark, angostura",
  "n13163250": "branch",
  "n13163471": "culm",
  "n13163553": "deadwood",
  "n13163649": "haulm, halm",
  "n13163803": "limb, tree branch",
  "n13163991": "branchlet, twig, sprig",
  "n13164285": "wand",
  "n13164403": "withe, withy",
  "n13164501": "osier",
  "n13164583": "sprout",
  "n13164763": "shoot",
  "n13164881": "sucker",
  "n13164970": "tiller",
  "n13165086": "bud",
  "n13165286": "leaf bud",
  "n13165409": "flower bud",
  "n13165552": "mixed bud",
  "n13165646": "stick",
  "n13165727": "bough",
  "n13165815": "trunk, tree trunk, bole",
  "n13166044": "burl",
  "n13166191": "burl",
  "n13166338": "fern family",
  "n13167078": "fern genus",
  "n13169219": "Filicopsida, class Filicopsida, Filicinae, class Filicinae",
  "n13169674": "Filicales, order Filicales, Polypodiales, order Polypodiales",
  "n13170060": "Gleicheniaceae, family Gleicheniaceae",
  "n13170286": "Gleichenia, genus Gleichenia",
  "n13170498": "Dicranopteris, genus Dicranopteris",
  "n13170661": "Diplopterygium, genus Diplopterygium",
  "n13170840": "giant scrambling fern, Diplopterygium longissimum",
  "n13171041": "Sticherus, genus Sticherus",
  "n13171210": "umbrella fern, fan fern, Sticherus flabellatus, Gleichenia flabellata",
  "n13171447": "Parkeriaceae, family Parkeriaceae",
  "n13171649": "Ceratopteris, genus Ceratopteris",
  "n13171797": "floating fern, water sprite, Ceratopteris pteridioides",
  "n13171975": "floating fern, Ceratopteris thalictroides",
  "n13172107": "Polypodiaceae, family Polypodiaceae",
  "n13172725": "Polypodium, genus Polypodium",
  "n13172923": "polypody",
  "n13173132": "licorice fern, Polypodium glycyrrhiza",
  "n13173259": "grey polypody, gray polypody, resurrection fern, Polypodium polypodioides",
  "n13173488": "leatherleaf, leathery polypody, coast polypody, Polypodium scouleri",
  "n13173697": "rock polypody, rock brake, American wall fern, Polypodium virgianum",
  "n13173882": "common polypody, adder's fern, wall fern, golden maidenhair, golden polypody, sweet fern, Polypodium vulgare",
  "n13174206": "Aglaomorpha, genus Aglaomorpha",
  "n13174354": "bear's-paw fern, Aglaomorpha meyeniana",
  "n13174515": "Campyloneurum, genus Campyloneurum",
  "n13174670": "strap fern",
  "n13174823": "Florida strap fern, cow-tongue fern, hart's-tongue fern",
  "n13175074": "Central American strap fern, narrow-leaved strap fern, Campyloneurum augustifolium",
  "n13175324": "Drymoglossum, genus Drymoglossum",
  "n13175484": "Drynaria, genus Drynaria",
  "n13175682": "basket fern, Drynaria rigidula",
  "n13175847": "genus Lecanopteris",
  "n13176000": "lecanopteris",
  "n13176201": "Microgramma, genus Microgramma",
  "n13176363": "snake polypody, Microgramma-piloselloides",
  "n13176523": "Microsorium, genus Microsorium",
  "n13176714": "climbing bird's nest fern, Microsorium punctatum",
  "n13176873": "Phlebodium, genus Phlebodium",
  "n13177048": "golden polypody, serpent fern, rabbit's-foot fern, Phlebodium aureum, Polypodium aureum",
  "n13177354": "Platycerium, genus Platycerium",
  "n13177529": "staghorn fern",
  "n13177768": "South American staghorn, Platycerium andinum",
  "n13177884": "common staghorn fern, elkhorn fern, Platycerium bifurcatum, Platycerium alcicorne",
  "n13178107": "Pyrrosia, genus Pyrrosia",
  "n13178284": "felt fern, tongue fern, Pyrrosia lingua, Cyclophorus lingua",
  "n13178500": "Solanopteris, genus Solanopteris",
  "n13178707": "potato fern, Solanopteris bifrons",
  "n13178883": "Cyclophorus, genus Cyclophorus",
  "n13179056": "myrmecophyte",
  "n13179216": "Adiantaceae, family Adiantaceae",
  "n13179410": "Vittariaceae, family Vittariaceae",
  "n13179648": "Vittaria, genus Vittaria",
  "n13179804": "grass fern, ribbon fern, Vittaria lineata",
  "n13179972": "Aspleniaceae, family Aspleniaceae",
  "n13180304": "Asplenium, genus Asplenium",
  "n13180534": "spleenwort",
  "n13180875": "black spleenwort, Asplenium adiantum-nigrum",
  "n13181055": "bird's nest fern, Asplenium nidus",
  "n13181244": "ebony spleenwort, Scott's Spleenwort, Asplenium platyneuron",
  "n13181406": "black-stem spleenwort, black-stemmed spleenwort, little ebony spleenwort",
  "n13181616": "Camptosorus, genus Camptosorus",
  "n13181811": "walking fern, walking leaf, Asplenium rhizophyllum, Camptosorus rhizophyllus",
  "n13181988": "maidenhair spleenwort, Asplenium trichomanes",
  "n13182164": "green spleenwort, Asplenium viride",
  "n13182338": "mountain spleenwort, Asplenium montanum",
  "n13182462": "wall rue, wall rue spleenwort, Asplenium ruta-muraria",
  "n13182662": "Bradley's spleenwort, Asplenium bradleyi",
  "n13182799": "lobed spleenwort, Asplenium pinnatifidum",
  "n13182937": "lanceolate spleenwort, Asplenium billotii",
  "n13183056": "hart's-tongue, hart's-tongue fern, Asplenium scolopendrium, Phyllitis scolopendrium",
  "n13183251": "Ceterach, genus Ceterach",
  "n13183489": "scale fern, scaly fern, Asplenium ceterach, Ceterach officinarum",
  "n13183669": "Pleurosorus, genus Pleurosorus",
  "n13183874": "Schaffneria, genus Schaffneria",
  "n13184001": "Schaffneria nigripes, Asplenium nigripes, Scolopendrium nigripes",
  "n13184164": "Phyllitis, genus Phyllitis, genus Scolopendrium",
  "n13184394": "scolopendrium",
  "n13184492": "Blechnaceae, family Blechnaceae",
  "n13184873": "Blechnum, genus Blechnum",
  "n13185134": "hard fern",
  "n13185269": "deer fern, Blechnum spicant",
  "n13185436": "genus Doodia, Doodia",
  "n13185658": "doodia, rasp fern",
  "n13185820": "Sadleria, genus Sadleria",
  "n13185998": "Stenochlaena, genus Stenochlaena",
  "n13186200": "Woodwardia, genus Woodwardia",
  "n13186388": "chain fern",
  "n13186546": "Virginia chain fern, Woodwardia virginica",
  "n13186654": "tree fern",
  "n13187031": "Cyatheaceae, family Cyatheaceae",
  "n13187167": "Cyathea, genus Cyathea",
  "n13187367": "silver tree fern, sago fern, black tree fern, Cyathea medullaris",
  "n13187604": "Davalliaceae, family Davalliaceae",
  "n13187826": "genus Davallia",
  "n13188096": "davallia",
  "n13188268": "hare's-foot fern",
  "n13188462": "Canary Island hare's foot fern, Davallia canariensis",
  "n13188619": "Australian hare's foot, Davallia pyxidata",
  "n13188767": "squirrel's-foot fern, ball fern, Davalia bullata, Davalia bullata mariesii, Davallia Mariesii",
  "n13188973": "Dennstaedtiaceae, family Dennstaedtiaceae",
  "n13189222": "Dennstaedtia, genus Dennstaedtia",
  "n13189428": "hay-scented, hay-scented fern, scented fern, boulder fern, Dennstaedtia punctilobula",
  "n13189656": "Pteridium, genus Pteridium",
  "n13189844": "bracken, pasture brake, brake, Pteridium aquilinum",
  "n13190060": "bracken, Pteridium esculentum",
  "n13190218": "Dicksoniaceae, family Dicksoniaceae",
  "n13190469": "Dicksonia, genus Dicksonia",
  "n13190747": "soft tree fern, Dicksonia antarctica",
  "n13190917": "Cibotium, genus Cibotium",
  "n13191148": "Scythian lamb, Cibotium barometz",
  "n13191318": "Culcita, genus Culcita",
  "n13191620": "false bracken, Culcita dubia",
  "n13191770": "genus Thyrsopteris",
  "n13191884": "thyrsopteris, Thyrsopteris elegans",
  "n13192025": "Dryopteridaceae, family Dryopteridaceae, Athyriaceae, family Athyriaceae",
  "n13192625": "shield fern, buckler fern",
  "n13192898": "Dryopteris, genus Dryopteris",
  "n13193143": "broad buckler-fern, Dryopteris dilatata",
  "n13193269": "fragrant cliff fern, fragrant shield fern, fragrant wood fern, Dryopteris fragrans",
  "n13193466": "Goldie's fern, Goldie's shield fern, goldie's wood fern, Dryopteris goldiana",
  "n13193642": "wood fern, wood-fern, woodfern",
  "n13193856": "male fern, Dryopteris filix-mas",
  "n13194036": "marginal wood fern, evergreen wood fern, leatherleaf wood fern, Dryopteris marginalis",
  "n13194212": "mountain male fern, Dryopteris oreades",
  "n13194328": "Athyrium, genus Athyrium",
  "n13194572": "lady fern, Athyrium filix-femina",
  "n13194758": "Alpine lady fern, Athyrium distentifolium",
  "n13194918": "silvery spleenwort, glade fern, narrow-leaved spleenwort, Athyrium pycnocarpon, Diplazium pycnocarpon",
  "n13195151": "Cyrtomium, genus Cyrtomium",
  "n13195341": "holly fern, Cyrtomium aculeatum, Polystichum aculeatum",
  "n13195547": "Cystopteris, genus Cystopteris",
  "n13195761": "bladder fern",
  "n13196003": "brittle bladder fern, brittle fern, fragile fern, Cystopteris fragilis",
  "n13196234": "mountain bladder fern, Cystopteris montana",
  "n13196369": "bulblet fern, bulblet bladder fern, berry fern, Cystopteris bulbifera",
  "n13196545": "Deparia, genus Deparia",
  "n13196738": "silvery spleenwort, Deparia acrostichoides, Athyrium thelypteroides",
  "n13196942": "Diacalpa, genus Diacalpa",
  "n13197085": "Gymnocarpium, genus Gymnocarpium",
  "n13197274": "oak fern, Gymnocarpium dryopteris, Thelypteris dryopteris",
  "n13197507": "limestone fern, northern oak fern, Gymnocarpium robertianum",
  "n13197670": "Lastreopsis, genus Lastreopsis",
  "n13197800": "Matteuccia, genus Matteuccia, Pteretis, genus Pteretis",
  "n13198054": "ostrich fern, shuttlecock fern, fiddlehead, Matteuccia struthiopteris, Pteretis struthiopteris, Onoclea struthiopteris",
  "n13198354": "Olfersia, genus Olfersia",
  "n13198482": "hart's-tongue, hart's-tongue fern, Olfersia cervina, Polybotrya cervina, Polybotria cervina",
  "n13198728": "Onoclea, genus Onoclea",
  "n13198914": "sensitive fern, bead fern, Onoclea sensibilis",
  "n13199244": "Polybotrya, genus Polybotrya, Polybotria, genus Polybotria",
  "n13199445": "Polystichum, genus Polystichum",
  "n13199717": "Christmas fern, canker brake, dagger fern, evergreen wood fern, Polystichum acrostichoides",
  "n13199970": "holly fern",
  "n13200193": "Braun's holly fern, prickly shield fern, Polystichum braunii",
  "n13200411": "northern holly fern, Polystichum lonchitis",
  "n13200542": "western holly fern, Polystichum scopulinum",
  "n13200651": "soft shield fern, Polystichum setiferum",
  "n13200806": "Rumohra, genus Rumohra",
  "n13200986": "leather fern, leatherleaf fern, ten-day fern, Rumohra adiantiformis, Polystichum adiantiformis",
  "n13201239": "Tectaria, genus Tectaria",
  "n13201423": "button fern, Tectaria cicutaria",
  "n13201566": "Indian button fern, Tectaria macrodonta",
  "n13201725": "genus Woodsia",
  "n13201969": "woodsia",
  "n13202125": "rusty woodsia, fragrant woodsia, oblong woodsia, Woodsia ilvensis",
  "n13202355": "Alpine woodsia, northern woodsia, flower-cup fern, Woodsia alpina",
  "n13202602": "smooth woodsia, Woodsia glabella",
  "n13202749": "Lomariopsidaceae, family Lomariopsidaceae",
  "n13202933": "Bolbitis, genus Bolbitis",
  "n13203115": "Lomogramma, genus Lomogramma",
  "n13203251": "Lophosoriaceae, family Lophosoriaceae",
  "n13203405": "Lophosoria, genus Lophosoria",
  "n13203551": "Loxomataceae, family Loxomataceae",
  "n13203708": "Loxoma, genus Loxoma",
  "n13203842": "Oleandraceae, family Oleandraceae",
  "n13204102": "Oleandra, genus Oleandra",
  "n13204276": "oleander fern, Oleandra neriiformis, Oleandra mollis",
  "n13204482": "Arthropteris, genus Arthropteris",
  "n13204646": "Nephrolepis, genus Nephrolepis",
  "n13204826": "sword fern",
  "n13205058": "Boston fern, Nephrolepis exaltata, Nephrolepis exaltata bostoniensis",
  "n13205249": "basket fern, toothed sword fern, Nephrolepis pectinata",
  "n13205482": "Pteridaceae, family Pteridaceae",
  "n13206001": "Acrostichum, genus Acrostichum",
  "n13206178": "golden fern, leather fern, Acrostichum aureum",
  "n13206438": "Actiniopteris, genus Actiniopteris",
  "n13206584": "Adiantum, genus Adiantum",
  "n13206817": "maidenhair, maidenhair fern",
  "n13207094": "common maidenhair, Venushair, Venus'-hair fern, southern maidenhair, Venus maidenhair, Adiantum capillus-veneris",
  "n13207335": "American maidenhair fern, five-fingered maidenhair fern, Adiantum pedatum",
  "n13207572": "Bermuda maidenhair, Bermuda maidenhair fern, Adiantum bellum",
  "n13207736": "brittle maidenhair, brittle maidenhair fern, Adiantum tenerum",
  "n13207923": "Farley maidenhair, Farley maidenhair fern, Barbados maidenhair, glory fern, Adiantum tenerum farleyense",
  "n13208138": "Anogramma, genus Anogramma",
  "n13208302": "annual fern, Jersey fern, Anogramma leptophylla",
  "n13208468": "Cheilanthes, genus Cheilanthes",
  "n13208705": "lip fern, lipfern",
  "n13208965": "smooth lip fern, Alabama lip fern, Cheilanthes alabamensis",
  "n13209129": "lace fern, Cheilanthes gracillima",
  "n13209270": "wooly lip fern, hairy lip fern, Cheilanthes lanosa",
  "n13209460": "southwestern lip fern, Cheilanthes eatonii",
  "n13209647": "Coniogramme, genus Coniogramme",
  "n13209808": "bamboo fern, Coniogramme japonica",
  "n13210006": "Cryptogramma, genus Cryptogramma",
  "n13210205": "rock brake",
  "n13210350": "American rock brake, American parsley fern, Cryptogramma acrostichoides",
  "n13210597": "European parsley fern, mountain parsley fern, Cryptogramma crispa",
  "n13210827": "Doryopteris, genus Doryopteris",
  "n13211020": "hand fern, Doryopteris pedata",
  "n13211179": "Jamesonia, genus Jamesonia",
  "n13211305": "Onychium, genus Onychium",
  "n13211516": "Pellaea, genus Pellaea",
  "n13211790": "cliff brake, cliff-brake, rock brake",
  "n13212025": "coffee fern, Pellaea andromedifolia",
  "n13212175": "purple rock brake, Pellaea atropurpurea",
  "n13212379": "bird's-foot fern, Pellaea mucronata, Pellaea ornithopus",
  "n13212559": "button fern, Pellaea rotundifolia",
  "n13212751": "Pityrogramma, genus Pityrogramma",
  "n13213066": "silver fern, Pityrogramma argentea",
  "n13213235": "silver fern, Pityrogramma calomelanos",
  "n13213397": "golden fern, Pityrogramma calomelanos aureoflava",
  "n13213577": "gold fern, Pityrogramma chrysophylla",
  "n13213768": "Pteris, genus Pteris",
  "n13214031": "brake",
  "n13214217": "Pteris cretica",
  "n13214340": "spider brake, spider fern, Pteris multifida",
  "n13214485": "ribbon fern, spider fern, Pteris serrulata",
  "n13214645": "Marattiales, order Marattiales",
  "n13214813": "Marattiaceae, family Marattiaceae",
  "n13215063": "Marattia, genus Marattia",
  "n13215258": "potato fern, Marattia salicina",
  "n13215462": "genus Angiopteris",
  "n13215586": "angiopteris, giant fern, Angiopteris evecta",
  "n13215799": "Danaea, genus Danaea",
  "n13215936": "Psilopsida, class Psilopsida, Psilotatae, class Psilotatae",
  "n13216238": "Psilotales, order Psilotales",
  "n13216475": "Psilotaceae, family Psilotaceae",
  "n13216673": "Psilotum, genus Psilotum",
  "n13216812": "whisk fern",
  "n13217005": "skeleton fork fern, Psilotum nudum",
  "n13217213": "Psilophytales, order Psilophytales",
  "n13217494": "psilophyte",
  "n13217624": "Psilophytaceae, family Psilophytaceae",
  "n13217763": "genus Psilophyton",
  "n13217993": "psilophyton",
  "n13218114": "Rhyniaceae, family Rhyniaceae",
  "n13218281": "Rhynia, genus Rhynia",
  "n13218504": "Horneophyton, genus Horneophyton",
  "n13218722": "Sphenopsida, class Sphenopsida, Equisetatae, class Equisetatae",
  "n13218900": "Equisetales, order Equisetales",
  "n13219067": "Equisetaceae, family Equisetaceae, horsetail family",
  "n13219258": "Equisetum, genus Equisetum",
  "n13219422": "horsetail",
  "n13219833": "common horsetail, field horsetail, Equisetum arvense",
  "n13219976": "swamp horsetail, water horsetail, Equisetum fluviatile",
  "n13220122": "scouring rush, rough horsetail, Equisetum hyemale, Equisetum hyemale robustum, Equisetum robustum",
  "n13220355": "marsh horsetail, Equisetum palustre",
  "n13220525": "wood horsetail, Equisetum Sylvaticum",
  "n13220663": "variegated horsetail, variegated scouring rush, Equisetum variegatum",
  "n13220842": "Lycopsida, class Lycopsida, Lycopodiate, class Lycopodiate",
  "n13221243": "Lycophyta",
  "n13221383": "Lycopodineae, class Lycopodineae",
  "n13221529": "club moss, club-moss, lycopod",
  "n13221807": "Lepidodendrales, order Lepidodendrales",
  "n13222035": "Lepidodendraceae, family Lepidodendraceae",
  "n13222227": "Lycopodiales, order Lycopodiales",
  "n13222477": "Lycopodiaceae, family Lycopodiaceae, clubmoss family",
  "n13222669": "Lycopodium, genus Lycopodium",
  "n13222877": "shining clubmoss, Lycopodium lucidulum",
  "n13222985": "alpine clubmoss, Lycopodium alpinum",
  "n13223090": "fir clubmoss, mountain clubmoss, little clubmoss, Lycopodium selago",
  "n13223265": "ground pine, Christmas green",
  "n13223485": "running pine, Lycopodium clavitum",
  "n13223588": "ground cedar, staghorn moss, Lycopodium complanatum",
  "n13223710": "ground fir, princess pine, tree clubmoss, Lycopodium obscurum",
  "n13223843": "foxtail grass, Lycopodium alopecuroides",
  "n13224086": "Selaginellales, order Selaginellales",
  "n13224256": "Selaginellaceae, family Selaginellaceae",
  "n13224454": "Selaginella, genus Selaginella",
  "n13224673": "spikemoss, spike moss, little club moss",
  "n13224922": "meadow spikemoss, basket spikemoss, Selaginella apoda",
  "n13225075": "rock spikemoss, dwarf lycopod, Selaginella rupestris",
  "n13225244": "desert selaginella, Selaginella eremophila",
  "n13225365": "resurrection plant, rose of Jericho, Selaginella lepidophylla",
  "n13225617": "florida selaginella, Selaginella eatonii",
  "n13225729": "Isoetales, order Isoetales",
  "n13225955": "Isoetaceae, family Isoetaceae, quillwort family",
  "n13226135": "Isoetes, genus Isoetes",
  "n13226320": "quillwort",
  "n13226526": "Geoglossaceae, family Geoglossaceae",
  "n13226698": "Geoglossum, genus Geoglossum",
  "n13226871": "earthtongue, earth-tongue",
  "n13227009": "Cryptogrammataceae, family Cryptogrammataceae",
  "n13227235": "Thelypteridaceae, family Thelypteridaceae",
  "n13227557": "Thelypteris, genus Thelypteris",
  "n13227778": "marsh fern, Thelypteris palustris, Dryopteris thelypteris",
  "n13228017": "snuffbox fern, meadow fern, Thelypteris palustris pubescens, Dryopteris thelypteris pubescens",
  "n13228194": "Amauropelta, genus Amauropelta",
  "n13228357": "genus Christella",
  "n13228536": "christella",
  "n13228692": "Cyclosorus, genus Cyclosorus",
  "n13228867": "Goniopteris, genus Goniopteris",
  "n13229018": "Macrothelypteris, genus Macrothelypteris",
  "n13229227": "Meniscium, genus Meniscium",
  "n13229358": "Oreopteris, genus Oreopteris",
  "n13229543": "mountain fern, Oreopteris limbosperma, Dryopteris oreopteris",
  "n13229747": "Parathelypteris, genus Parathelypteris",
  "n13229951": "New York fern, Parathelypteris novae-boracensis, Dryopteris noveboracensis",
  "n13230190": "Massachusetts fern, Parathelypteris simulata, Thelypteris simulata",
  "n13230421": "Phegopteris, genus Phegopteris",
  "n13230662": "beech fern",
  "n13230843": "broad beech fern, southern beech fern, Phegopteris hexagonoptera, Dryopteris hexagonoptera, Thelypteris hexagonoptera",
  "n13231078": "long beech fern, narrow beech fern, northern beech fern, Phegopteris connectilis, Dryopteris phegopteris, Thelypteris phegopteris",
  "n13231298": "rhizomorph",
  "n13231436": "Armillaria, genus Armillaria",
  "n13231678": "shoestring fungus",
  "n13231919": "Armillaria caligata, booted armillaria",
  "n13232106": "Armillaria ponderosa, white matsutake",
  "n13232363": "Armillaria zelleri",
  "n13232515": "Armillariella, genus Armillariella",
  "n13232779": "honey mushroom, honey fungus, Armillariella mellea",
  "n13233012": "Asclepiadaceae, family Asclepiadaceae, milkweed family",
  "n13233435": "asclepiad",
  "n13233548": "Asclepias, genus Asclepias",
  "n13233727": "milkweed, silkweed",
  "n13234114": "white milkweed, Asclepias albicans",
  "n13234293": "blood flower, swallowwort, Asclepias curassavica",
  "n13234519": "poke milkweed, Asclepias exaltata",
  "n13234678": "swamp milkweed, Asclepias incarnata",
  "n13234857": "Mead's milkweed, Asclepias meadii, Asclepia meadii",
  "n13235011": "purple silkweed, Asclepias purpurascens",
  "n13235159": "showy milkweed, Asclepias speciosa",
  "n13235319": "poison milkweed, horsetail milkweed, Asclepias subverticillata",
  "n13235503": "butterfly weed, orange milkweed, chigger flower, chiggerflower, pleurisy root, tuber root, Indian paintbrush, Asclepias tuberosa",
  "n13235766": "whorled milkweed, Asclepias verticillata",
  "n13235947": "Araujia, genus Araujia",
  "n13236100": "cruel plant, Araujia sericofera",
  "n13236354": "genus Cynancum",
  "n13236495": "cynancum",
  "n13236726": "genus Hoya",
  "n13236887": "hoya",
  "n13237075": "honey plant",
  "n13237188": "wax plant, Hoya carnosa",
  "n13237343": "Periploca, genus Periploca",
  "n13237508": "silk vine, Periploca graeca",
  "n13237788": "Sarcostemma, genus Sarcostemma",
  "n13237988": "soma, haoma, Sarcostemma acidum",
  "n13238178": "genus Stapelia",
  "n13238375": "stapelia, carrion flower, starfish flower",
  "n13238654": "Stapelias asterias",
  "n13238828": "genus Stephanotis",
  "n13238988": "stephanotis",
  "n13239177": "Madagascar jasmine, waxflower, Stephanotis floribunda",
  "n13239471": "Vincetoxicum, genus Vincetoxicum",
  "n13239736": "negro vine, Vincetoxicum hirsutum, Vincetoxicum negrum",
  "n13239921": "zygospore",
  "n13240025": "old growth, virgin forest",
  "n13240203": "second growth",
  "n13240362": "tree of knowledge",
  "n13240514": "ownership",
  "n13240839": "community",
  "n13240971": "severalty",
  "n13241057": "property right",
  "n13241182": "preemptive right",
  "n13241407": "subscription right",
  "n13241600": "option",
  "n13241940": "stock option",
  "n13242239": "stock option",
  "n13242410": "call option",
  "n13242484": "put option",
  "n13242558": "lock-up option",
  "n13242719": "tenure, land tenure",
  "n13242923": "copyhold",
  "n13243134": "freehold",
  "n13243261": "freehold",
  "n13243374": "villeinage",
  "n13243466": "stock buyback",
  "n13243668": "public domain",
  "n13243780": "proprietorship, proprietary",
  "n13243991": "employee ownership",
  "n13244109": "property, belongings, holding",
  "n13244753": "tangible possession, material possession",
  "n13244881": "worldly possessions, worldly belongings, worldly goods",
  "n13245076": "ratables, rateables",
  "n13245217": "hereditament",
  "n13245338": "intellectual property",
  "n13245504": "community property",
  "n13245626": "personal property, personal estate, personalty, private property",
  "n13245846": "chattel, personal chattel, movable",
  "n13246079": "effects, personal effects",
  "n13246332": "things",
  "n13246475": "real property, real estate, realty, immovable",
  "n13246662": "estate, land, landed estate, acres, demesne",
  "n13247149": "commonage",
  "n13247228": "glebe",
  "n13247358": "landholding",
  "n13247445": "landholding",
  "n13247554": "salvage",
  "n13247712": "shareholding",
  "n13247818": "spiritualty, spirituality, church property",
  "n13247981": "temporalty, temporality",
  "n13248087": "benefice, ecclesiastical benefice",
  "n13248274": "sinecure",
  "n13248393": "lease, rental, letting",
  "n13248598": "car rental, hire car, rent-a-car, self-drive, u-drive, you-drive",
  "n13248792": "trade-in",
  "n13248928": "sublease, sublet",
  "n13249062": "public property",
  "n13249154": "leasehold",
  "n13249245": "smallholding",
  "n13249400": "homestead",
  "n13249599": "farmstead",
  "n13249699": "homestead",
  "n13249805": "no man's land",
  "n13249927": "fief, feoff",
  "n13250048": "land",
  "n13250244": "mortmain, dead hand",
  "n13250398": "wealth",
  "n13250542": "money",
  "n13250680": "pile, bundle, big bucks, megabucks, big money",
  "n13250930": "estate",
  "n13251154": "stuff, clobber",
  "n13251289": "gross estate",
  "n13251424": "net estate",
  "n13251715": "life estate, estate for life",
  "n13251906": "barony",
  "n13251980": "countryseat",
  "n13252062": "Crown land",
  "n13252168": "manor",
  "n13252293": "seigneury, seigniory, signory",
  "n13252395": "hacienda",
  "n13252513": "plantation",
  "n13252672": "orangery",
  "n13252853": "white elephant",
  "n13252973": "transferred property, transferred possession",
  "n13253255": "acquisition",
  "n13253423": "accession, addition",
  "n13253612": "purchase",
  "n13253751": "bargain, buy, steal",
  "n13254011": "song",
  "n13254105": "travel bargain",
  "n13254237": "grant, assignment",
  "n13254443": "appanage, apanage",
  "n13254697": "land grant",
  "n13254805": "gain",
  "n13254985": "financial gain",
  "n13255145": "income",
  "n13255562": "disposable income",
  "n13255713": "double dipping",
  "n13255883": "easy money, gravy train",
  "n13255994": "EBITDA, Earnings Before Interest Taxes Depreciation and Amortization",
  "n13256303": "easy money",
  "n13256434": "tight money",
  "n13256599": "escheat",
  "n13256691": "gross, revenue, receipts",
  "n13256894": "national income",
  "n13257098": "gross national product, GNP",
  "n13257354": "real gross national product, real GNP",
  "n13257511": "gross domestic product, GDP",
  "n13257792": "deflator",
  "n13257982": "royalty",
  "n13258167": "box office",
  "n13258269": "gate",
  "n13258362": "net income, net, net profit, lucre, profit, profits, earnings",
  "n13258825": "paper profit",
  "n13259013": "paper loss",
  "n13259199": "cash flow",
  "n13259359": "personal income",
  "n13259481": "earning per share",
  "n13259630": "windfall profit",
  "n13259797": "killing, cleanup",
  "n13259917": "winnings, win, profits",
  "n13260092": "rental income",
  "n13260190": "return, issue, take, takings, proceeds, yield, payoff",
  "n13260510": "fast buck, quick buck",
  "n13260645": "filthy lucre",
  "n13260762": "gross profit, gross profit margin, margin",
  "n13260936": "gross sales, gross revenue, sales",
  "n13261107": "net sales",
  "n13261242": "margin of profit, profit margin, gross margin",
  "n13261380": "unearned income, unearned revenue",
  "n13261597": "unearned income, unearned revenue",
  "n13261779": "government income, government revenue",
  "n13261916": "tax income, taxation, tax revenue, revenue",
  "n13262077": "internal revenue",
  "n13262200": "per capita income",
  "n13262335": "stolen property",
  "n13262462": "spoil",
  "n13262663": "loot, booty, pillage, plunder, prize, swag, dirty money",
  "n13262913": "inheritance, heritage",
  "n13263257": "primogeniture",
  "n13263375": "borough English",
  "n13263540": "accretion",
  "n13263779": "bequest, legacy",
  "n13263920": "birthright, patrimony",
  "n13264076": "devise",
  "n13264203": "dower",
  "n13264342": "jointure, legal jointure",
  "n13264522": "free lunch",
  "n13264688": "heirloom",
  "n13264794": "heirloom",
  "n13265011": "gift",
  "n13265425": "dowry, dowery, dower, portion",
  "n13265603": "bride price",
  "n13265749": "largess, largesse",
  "n13265904": "aid, economic aid, financial aid",
  "n13266170": "scholarship",
  "n13266348": "fellowship",
  "n13266515": "foreign aid",
  "n13266690": "Marshall Plan, European Recovery Program",
  "n13266892": "grant",
  "n13267014": "subsidy",
  "n13267263": "subvention",
  "n13267411": "price support",
  "n13267534": "grant-in-aid",
  "n13267687": "postdoctoral, postdoc, post doc",
  "n13267816": "traineeship",
  "n13268020": "gratuity",
  "n13268146": "prize, award",
  "n13268484": "door prize",
  "n13268683": "jackpot",
  "n13268758": "prize money",
  "n13268842": "present",
  "n13269054": "birthday present, birthday gift",
  "n13269186": "Christmas present, Christmas gift",
  "n13269318": "stocking filler, stocking stuffer",
  "n13269459": "wedding present, wedding gift",
  "n13269596": "bride-gift",
  "n13269683": "cash surrender value",
  "n13269890": "contribution",
  "n13270038": "contribution, donation",
  "n13270260": "benefaction",
  "n13270373": "offering",
  "n13270545": "tithe",
  "n13270648": "offertory",
  "n13270760": "hearth money, Peter's pence",
  "n13270906": "political contribution, political donation",
  "n13271094": "soft money",
  "n13271320": "endowment, endowment fund",
  "n13271498": "enrichment",
  "n13271631": "patrimony",
  "n13271705": "chantry",
  "n13271797": "lagniappe",
  "n13271937": "bestowal, bestowment",
  "n13272059": "bounty, premium",
  "n13272283": "premium",
  "n13272545": "freebie, freebee",
  "n13272712": "giveaway",
  "n13272860": "gift horse",
  "n13273050": "thank offering",
  "n13273154": "bonus, incentive",
  "n13273381": "incentive program, incentive scheme",
  "n13273550": "deductible",
  "n13273738": "defalcation",
  "n13273836": "dividend",
  "n13273949": "sales incentive",
  "n13274092": "allowance, adjustment",
  "n13274364": "cost-of-living allowance",
  "n13274487": "depreciation allowance",
  "n13274597": "deduction, discount",
  "n13274732": "trade discount",
  "n13274892": "seasonal adjustment",
  "n13275137": "tare",
  "n13275288": "outgo, spending, expenditure, outlay",
  "n13275495": "expense, disbursal, disbursement",
  "n13275847": "cost",
  "n13276330": "business expense, trade expense",
  "n13276519": "interest expense",
  "n13276604": "lobbying expense",
  "n13276778": "medical expense",
  "n13276913": "non-cash expense",
  "n13277056": "moving expense",
  "n13277179": "operating expense, operating cost, overhead, budget items",
  "n13277492": "organization expense",
  "n13277644": "personal expense",
  "n13277789": "promotional expense",
  "n13277886": "expense",
  "n13278100": "transfer payment",
  "n13278275": "capital expenditure",
  "n13278375": "payment",
  "n13279046": "overpayment",
  "n13279162": "underpayment",
  "n13279262": "wage, pay, earnings, remuneration, salary",
  "n13279809": "combat pay",
  "n13279913": "double time",
  "n13280008": "found",
  "n13280139": "half-pay",
  "n13280251": "living wage",
  "n13280373": "merit pay",
  "n13280515": "minimum wage",
  "n13280658": "pay envelope, pay packet",
  "n13280795": "sick pay",
  "n13280896": "strike pay",
  "n13280992": "take-home pay",
  "n13281130": "subscription",
  "n13281275": "regular payment",
  "n13281476": "pay rate, rate of pay",
  "n13281641": "time and a half",
  "n13281770": "payment rate, rate of payment, repayment rate, installment rate",
  "n13281925": "blood money",
  "n13282007": "recompense",
  "n13282161": "refund",
  "n13282275": "rebate, discount",
  "n13282419": "rent-rebate",
  "n13282550": "compensation",
  "n13282948": "overcompensation",
  "n13283033": "workmen's compensation",
  "n13283187": "conscience money",
  "n13283314": "support payment",
  "n13283485": "palimony",
  "n13283620": "alimony, maintenance",
  "n13283764": "reward",
  "n13283952": "honorarium",
  "n13284048": "ransom, ransom money",
  "n13284184": "blood money",
  "n13284283": "guerdon",
  "n13284356": "meed",
  "n13284442": "hush money",
  "n13284562": "bribe, payoff",
  "n13284784": "kickback",
  "n13284973": "payola",
  "n13285101": "soap",
  "n13285176": "share, portion, part, percentage",
  "n13285619": "tranche",
  "n13285714": "dispensation",
  "n13285855": "dole",
  "n13285973": "way",
  "n13286099": "ration",
  "n13286254": "allowance",
  "n13286524": "slice, piece",
  "n13286640": "split",
  "n13286801": "interest, stake",
  "n13287239": "grubstake",
  "n13287414": "controlling interest",
  "n13287540": "insurable interest",
  "n13287787": "vested interest",
  "n13287984": "security interest",
  "n13288206": "terminable interest",
  "n13288337": "undivided interest, undivided right",
  "n13288529": "fee",
  "n13288661": "fee simple",
  "n13288798": "fee tail",
  "n13288942": "entail",
  "n13289020": "profit sharing",
  "n13289159": "cut",
  "n13289306": "rake-off, vigorish",
  "n13289467": "allotment, allocation",
  "n13289630": "reallocation",
  "n13289744": "quota",
  "n13289845": "appropriation",
  "n13290002": "reimbursement",
  "n13290285": "emolument",
  "n13290566": "blood money",
  "n13290676": "damages, amends, indemnity, indemnification, restitution, redress",
  "n13290991": "relief",
  "n13291189": "counterbalance, offset",
  "n13291356": "actual damages, compensatory damages, general damages",
  "n13291614": "nominal damages",
  "n13291831": "punitive damages, exemplary damages, smart money",
  "n13292104": "double damages",
  "n13292244": "treble damages",
  "n13292390": "reparation",
  "n13292613": "reparation",
  "n13292787": "atonement, expiation, satisfaction",
  "n13292989": "residual",
  "n13293246": "poverty line, poverty level",
  "n13293377": "allowance",
  "n13293512": "breakage",
  "n13293625": "costs",
  "n13293752": "per diem",
  "n13293901": "travel allowance, travel reimbursement",
  "n13294029": "mileage",
  "n13294135": "stipend",
  "n13294302": "privy purse",
  "n13294403": "prebend",
  "n13294503": "annuity, rente",
  "n13294763": "annuity in advance",
  "n13295025": "ordinary annuity",
  "n13295183": "reversionary annuity, survivorship annuity",
  "n13295360": "tontine",
  "n13295657": "rent",
  "n13295972": "ground rent",
  "n13296089": "peppercorn rent",
  "n13296194": "rack rent",
  "n13296270": "economic rent, rent",
  "n13296460": "payback",
  "n13296593": "installment plan, installment buying, time plan",
  "n13296752": "hire-purchase, never-never",
  "n13296899": "benefit",
  "n13297097": "cost-of-living benefit",
  "n13297254": "death benefit",
  "n13297397": "advance death benefit",
  "n13297595": "viatical settlement",
  "n13297740": "disability benefit",
  "n13297850": "sick benefit, sickness benefit",
  "n13298011": "fringe benefit, perquisite, perk",
  "n13298287": "appanage, apanage",
  "n13298537": "tax benefit, tax break",
  "n13298701": "gratuity, tip, pourboire, baksheesh, bakshish, bakshis, backsheesh",
  "n13298935": "Christmas box",
  "n13299071": "child support",
  "n13299248": "lump sum",
  "n13299357": "payoff, final payment",
  "n13299453": "remittance, remittal, remission, remitment",
  "n13299651": "repayment, quittance",
  "n13299804": "redemption",
  "n13300025": "token payment",
  "n13300141": "nonpayment, default, nonremittal",
  "n13300324": "delinquency",
  "n13300411": "default, nonpayment, nonremittal",
  "n13300555": "penalty",
  "n13300711": "pittance",
  "n13300828": "retribution, requital",
  "n13300922": "forfeit, forfeiture",
  "n13301174": "forfeit, forfeiture",
  "n13301328": "fine, mulct, amercement",
  "n13301500": "library fine",
  "n13301620": "premium, insurance premium",
  "n13301715": "installment",
  "n13301835": "cost overrun",
  "n13302001": "cost of living",
  "n13302201": "borrowing cost",
  "n13302293": "distribution cost",
  "n13302467": "handling cost, handling charge",
  "n13302621": "marketing cost",
  "n13302775": "production cost",
  "n13302905": "replacement cost",
  "n13303045": "reproduction cost, physical value",
  "n13303214": "unit cost",
  "n13303315": "price, terms, damage",
  "n13303759": "price",
  "n13303880": "markup",
  "n13304009": "asking price, selling price",
  "n13304186": "bid price",
  "n13304340": "closing price",
  "n13304508": "offer price",
  "n13304665": "upset price",
  "n13304819": "factory price",
  "n13304927": "highway robbery",
  "n13305070": "list price",
  "n13305262": "purchase price",
  "n13305373": "spot price, cash price",
  "n13305510": "support level",
  "n13305673": "valuation",
  "n13305794": "opportunity cost",
  "n13305932": "cost of capital, capital cost",
  "n13306190": "carrying cost, carrying charge",
  "n13306346": "portage",
  "n13306436": "incidental expense, incidental, minor expense",
  "n13306669": "travel expense",
  "n13306870": "charge",
  "n13307370": "carrying charge",
  "n13307514": "depreciation charge",
  "n13307685": "undercharge",
  "n13307784": "overcharge",
  "n13307901": "extortion",
  "n13307995": "corkage",
  "n13308147": "fare, transportation",
  "n13308336": "airfare",
  "n13308432": "bus fare, carfare",
  "n13308543": "cab fare, taxi fare",
  "n13308650": "subway fare",
  "n13308750": "train fare",
  "n13308864": "levy",
  "n13308999": "tax, taxation, revenue enhancement",
  "n13309776": "tax base",
  "n13309865": "tax rate",
  "n13309956": "tax liability",
  "n13310091": "single tax",
  "n13310230": "income tax",
  "n13310415": "bracket creep",
  "n13310539": "estimated tax",
  "n13310727": "FICA",
  "n13310858": "business deduction",
  "n13310985": "exemption",
  "n13311217": "entertainment deduction",
  "n13311368": "withholding tax, withholding",
  "n13311573": "PAYE, pay as you earn",
  "n13311679": "unearned increment",
  "n13311830": "capital gain",
  "n13312010": "capital loss",
  "n13312190": "capital gains tax",
  "n13312329": "capital levy",
  "n13312434": "departure tax",
  "n13312569": "property tax, land tax",
  "n13312754": "council tax",
  "n13312962": "franchise tax",
  "n13313188": "gift tax",
  "n13313322": "inheritance tax, estate tax, death tax, death duty",
  "n13313464": "direct tax",
  "n13313591": "tax advantage",
  "n13313733": "tax shelter, shelter",
  "n13313899": "indirect tax",
  "n13314085": "hidden tax",
  "n13314226": "capitation",
  "n13314355": "poll tax",
  "n13314495": "progressive tax, graduated tax",
  "n13314652": "proportional tax",
  "n13314796": "degressive tax",
  "n13314936": "rates",
  "n13315077": "poor rates",
  "n13315191": "stamp tax, stamp duty",
  "n13315364": "surtax, supertax",
  "n13315518": "pavage",
  "n13315616": "transfer tax",
  "n13315743": "tithe",
  "n13315827": "special assessment",
  "n13315999": "duty, tariff",
  "n13316332": "excise, excise tax",
  "n13316568": "sales tax, nuisance tax",
  "n13316722": "VAT, value-added tax, ad valorem tax",
  "n13316905": "gasoline tax",
  "n13317002": "customs, customs duty, custom, impost",
  "n13317137": "ship money",
  "n13317269": "tonnage, tunnage, tonnage duty",
  "n13317422": "octroi",
  "n13317517": "revenue tariff",
  "n13317611": "protective tariff",
  "n13317759": "anti-dumping duty",
  "n13317858": "import duty",
  "n13317941": "export duty",
  "n13318024": "countervailing duty",
  "n13318147": "fixed charge, fixed cost, fixed costs",
  "n13318411": "cover charge, cover",
  "n13318584": "interest",
  "n13318804": "compound interest",
  "n13318934": "simple interest",
  "n13319032": "interest rate, rate of interest",
  "n13319253": "discount rate, discount, bank discount",
  "n13319415": "bank rate",
  "n13319512": "discount rate",
  "n13319726": "base rate",
  "n13319872": "prime interest rate",
  "n13320045": "usury, vigorish",
  "n13320168": "fee",
  "n13320691": "anchorage",
  "n13320766": "cellarage",
  "n13320860": "commission",
  "n13321081": "contingency fee",
  "n13321230": "dockage, docking fee",
  "n13321338": "drop-off charge",
  "n13321495": "entrance fee, admission, admission charge, admission fee, admission price, price of admission, entrance money",
  "n13321722": "finder's fee",
  "n13321992": "legal fee",
  "n13322113": "licensing fee, license fee, license tax",
  "n13322343": "refresher",
  "n13322520": "lighterage",
  "n13322641": "lockage",
  "n13322758": "mintage",
  "n13322867": "moorage",
  "n13322938": "origination fee",
  "n13323102": "pipage",
  "n13323207": "poundage",
  "n13323313": "retainer, consideration",
  "n13323460": "quid pro quo, quid",
  "n13323648": "seigniorage",
  "n13323749": "toll",
  "n13323885": "truckage",
  "n13323988": "tuition, tuition fee",
  "n13324188": "wharfage, quayage",
  "n13324297": "agio, agiotage, premium, exchange premium",
  "n13324427": "demurrage",
  "n13324609": "installation charge",
  "n13324710": "porterage",
  "n13324826": "postage",
  "n13324912": "poundage",
  "n13325010": "rate, charge per unit",
  "n13325382": "water-rate",
  "n13325505": "surcharge",
  "n13325722": "single supplement",
  "n13325847": "service charge, service fee",
  "n13326021": "stowage",
  "n13326103": "tankage",
  "n13326198": "freight, freightage, freight rate",
  "n13326450": "rate of depreciation, depreciation rate",
  "n13326620": "rate of exchange, exchange rate",
  "n13326772": "excursion rate",
  "n13326871": "footage",
  "n13326975": "linage, lineage",
  "n13327136": "room rate",
  "n13327231": "loss, red ink, red",
  "n13327503": "squeeze",
  "n13327676": "loss",
  "n13327896": "financial loss",
  "n13328073": "sacrifice",
  "n13328273": "wastage",
  "n13328357": "depreciation, wear and tear",
  "n13328578": "straight-line method, straight-line method of depreciation",
  "n13328853": "write-off, write-down",
  "n13329047": "tax write-off, tax deduction, deduction",
  "n13329322": "losings, losses",
  "n13329489": "circumstances",
  "n13329641": "assets",
  "n13330168": "payables",
  "n13330283": "receivables",
  "n13330425": "crown jewel",
  "n13330591": "deep pocket",
  "n13330764": "reserve assets",
  "n13330941": "special drawing rights, paper gold",
  "n13331198": "sum, sum of money, amount, amount of money",
  "n13331634": "figure",
  "n13331778": "resource",
  "n13332009": "natural resource, natural resources",
  "n13332197": "labor resources",
  "n13332290": "land resources",
  "n13332395": "mineral resources",
  "n13332500": "renewable resource",
  "n13332670": "intangible, intangible asset",
  "n13332820": "good will, goodwill",
  "n13333047": "liquid assets, current assets, quick assets",
  "n13333237": "investment, investment funds",
  "n13333420": "equity",
  "n13333546": "sweat equity",
  "n13333696": "equity",
  "n13333833": "stock",
  "n13334307": "stockholding",
  "n13334492": "stockholding, stockholdings",
  "n13334650": "capital stock",
  "n13334767": "blue chip, blue-chip stock",
  "n13335008": "classified stock",
  "n13335172": "common stock, common shares, ordinary shares",
  "n13335521": "stock of record",
  "n13335635": "par value, face value, nominal value",
  "n13335808": "no-par-value stock, no-par stock",
  "n13335974": "preferred stock, preferred shares, preference shares",
  "n13336204": "cumulative preferred, cumulative preferred stock",
  "n13336368": "float",
  "n13336492": "common stock equivalent",
  "n13336650": "control stock",
  "n13336770": "growth stock",
  "n13336933": "hot stock, hot issue",
  "n13337050": "penny stock",
  "n13337146": "book value",
  "n13337322": "market value, market price",
  "n13337471": "bond issue",
  "n13337634": "convertible bond",
  "n13337770": "corporate bond",
  "n13338066": "coupon bond, bearer bond",
  "n13338234": "government bond",
  "n13338439": "junk bond, high-yield bond",
  "n13338657": "municipal bond",
  "n13338780": "noncallable bond",
  "n13338960": "performance bond, surety bond",
  "n13339189": "post-obit bond",
  "n13339325": "registered bond",
  "n13339542": "revenue bond",
  "n13339734": "secured bond",
  "n13339844": "unsecured bond, debenture, debenture bond",
  "n13340080": "government security, agency security",
  "n13340244": "mortgage-backed security",
  "n13340579": "registered security",
  "n13340751": "savings bond",
  "n13340902": "utility bond, utility revenue bond",
  "n13341052": "zero coupon bond, zero-coupon bond",
  "n13341350": "reversion",
  "n13341593": "escheat",
  "n13341756": "right",
  "n13341962": "accession",
  "n13342135": "share",
  "n13342398": "authorized shares, authorized stock, capital stock",
  "n13342597": "quarter stock",
  "n13342692": "speculation, venture",
  "n13342987": "gamble",
  "n13343123": "smart money",
  "n13343284": "pyramid",
  "n13343526": "stake, stakes, bet, wager",
  "n13343774": "pot, jackpot, kitty",
  "n13343917": "ante",
  "n13344071": "security, protection",
  "n13344386": "easy street",
  "n13344462": "hedge, hedging",
  "n13344664": "coverage, insurance coverage",
  "n13344804": "insurance",
  "n13345286": "assurance",
  "n13345403": "automobile insurance, car insurance",
  "n13345560": "no fault insurance, no fault automobile insurance",
  "n13345962": "business interruption insurance",
  "n13346209": "coinsurance",
  "n13346337": "fire insurance",
  "n13346432": "group insurance",
  "n13346630": "hazard insurance",
  "n13346773": "health insurance",
  "n13346912": "hospitalization insurance, hospitalization",
  "n13347065": "liability insurance",
  "n13347237": "life insurance, life assurance",
  "n13347489": "endowment insurance",
  "n13347750": "tontine, tontine insurance",
  "n13347947": "whole life insurance, ordinary life insurance, straight life insurance",
  "n13348227": "malpractice insurance",
  "n13348456": "reinsurance",
  "n13348792": "self-insurance",
  "n13348964": "term insurance",
  "n13349208": "health maintenance organization, HMO",
  "n13349395": "security, surety",
  "n13349662": "deposit",
  "n13349834": "down payment, deposit",
  "n13349997": "satisfaction",
  "n13350182": "earnest",
  "n13350322": "earnest money, arles",
  "n13350443": "recognizance, recognisance",
  "n13350702": "pledge",
  "n13350875": "pawn",
  "n13350976": "bail, bail bond, bond",
  "n13351278": "margin, security deposit",
  "n13351457": "brokerage account",
  "n13351680": "cash account",
  "n13351860": "custodial account",
  "n13351986": "margin account",
  "n13352138": "mortgage",
  "n13352341": "conditional sale",
  "n13352464": "first mortgage",
  "n13352610": "second mortgage",
  "n13352722": "chattel mortgage",
  "n13352865": "collateral",
  "n13353004": "guarantee, guaranty",
  "n13353156": "material resource",
  "n13353280": "wealth, riches",
  "n13353446": "gold",
  "n13353607": "capital",
  "n13353858": "means, substance",
  "n13354021": "pocketbook",
  "n13354154": "wherewithal",
  "n13354260": "venture capital, risk capital",
  "n13354420": "capital, working capital",
  "n13354769": "operating capital",
  "n13354985": "account, accounting, account statement",
  "n13355301": "income statement, earnings report, operating statement, profit-and-loss statement",
  "n13355504": "capital account",
  "n13355656": "capital account",
  "n13355868": "principal, corpus, principal sum",
  "n13356002": "seed money",
  "n13356112": "funds, finances, monetary resource, cash in hand, pecuniary resource",
  "n13356402": "bank",
  "n13356569": "bankroll, roll",
  "n13356774": "pocket",
  "n13356887": "Medicaid funds",
  "n13356985": "treasury, exchequer",
  "n13357178": "money supply",
  "n13357410": "M1",
  "n13357572": "M2",
  "n13357720": "M3",
  "n13357891": "public treasury, trough, till",
  "n13358000": "pork barrel, pork",
  "n13358151": "bursary",
  "n13358260": "subtreasury",
  "n13358360": "fisc",
  "n13358549": "fund, monetary fund",
  "n13358895": "mutual fund",
  "n13359032": "exchange traded fund, ETF",
  "n13359153": "index fund",
  "n13359263": "revolving fund",
  "n13359428": "sinking fund",
  "n13359572": "savings, nest egg",
  "n13359690": "bank account",
  "n13359941": "giro account",
  "n13360103": "pension fund, superannuation fund",
  "n13360254": "war chest",
  "n13360390": "slush fund",
  "n13360498": "trust",
  "n13360950": "active trust",
  "n13361083": "blind trust",
  "n13361333": "passive trust",
  "n13361465": "charitable trust, public trust",
  "n13361624": "Clifford trust, grantor trust",
  "n13361830": "implied trust",
  "n13361962": "constructive trust, involuntary trust",
  "n13362182": "resulting trust",
  "n13362347": "direct trust, express trust",
  "n13362540": "discretionary trust",
  "n13362734": "nondiscretionary trust, fixed investment trust",
  "n13362918": "living trust, inter vivos trust",
  "n13363056": "spendthrift trust",
  "n13363217": "testamentary trust",
  "n13363365": "savings account trust, savings bank trust, trust account, trustee account, Totten trust",
  "n13363704": "voting trust",
  "n13363893": "trust fund",
  "n13363970": "checking account, chequing account, current account",
  "n13364212": "savings account",
  "n13364368": "time deposit account, deposit account",
  "n13364602": "dormant account",
  "n13364821": "passbook savings account",
  "n13364984": "cash equivalent",
  "n13365137": "certificate of deposit, CD",
  "n13365286": "support, keep, livelihood, living, bread and butter, sustenance",
  "n13365698": "support, financial support, funding, backing, financial backing",
  "n13365978": "ways and means",
  "n13366137": "comforts, creature comforts, amenities, conveniences",
  "n13366311": "maintenance",
  "n13366428": "meal ticket",
  "n13366537": "subsistence",
  "n13366693": "accumulation",
  "n13366912": "hoard, cache, stash",
  "n13367070": "store, stock, fund",
  "n13367448": "provision",
  "n13367593": "issue, military issue, government issue",
  "n13367787": "seed stock",
  "n13367913": "seed corn, seed grain",
  "n13368052": "reserve, backlog, stockpile",
  "n13368318": "bank",
  "n13368517": "blood bank",
  "n13368675": "eye bank",
  "n13368900": "food bank",
  "n13369074": "fuel level",
  "n13369159": "hole card",
  "n13369282": "soil bank",
  "n13369474": "pool, kitty",
  "n13369567": "hidden reserve",
  "n13369723": "cookie jar reserve",
  "n13369857": "pool",
  "n13370014": "reserve account, reserve fund",
  "n13370215": "valuation reserve, valuation account, allowance, allowance account",
  "n13370448": "treasure",
  "n13370669": "treasure, hoarded wealth",
  "n13370938": "fortune",
  "n13371030": "valuable",
  "n13371190": "swag",
  "n13371274": "king's ransom",
  "n13371355": "treasure trove, trove",
  "n13371489": "precious metal",
  "n13371682": "bullion",
  "n13371760": "gold",
  "n13371866": "silver",
  "n13371958": "diamond",
  "n13372123": "ice, sparkler",
  "n13372262": "ruby",
  "n13372403": "pearl",
  "n13372585": "seed pearl",
  "n13372665": "emerald",
  "n13372812": "sapphire",
  "n13372961": "medium of exchange, monetary system",
  "n13373214": "standard, monetary standard",
  "n13373426": "gold standard",
  "n13373584": "silver standard",
  "n13373746": "bimetallism",
  "n13374008": "tender, legal tender, stamp",
  "n13374161": "food stamp",
  "n13374281": "credit, deferred payment",
  "n13374426": "consumer credit",
  "n13374597": "bank loan",
  "n13374764": "business loan, commercial loan",
  "n13374887": "interbank loan",
  "n13374979": "home loan, home equity credit, home equity loan, equity credit line",
  "n13375167": "installment credit, installment loan",
  "n13375323": "open-end credit, revolving credit, charge account credit",
  "n13375604": "credit account, charge account, open account",
  "n13375758": "revolving charge account",
  "n13375891": "advance, cash advance",
  "n13376012": "credit card, charge card, charge plate, plastic",
  "n13376426": "bank card",
  "n13376512": "calling card, phone card",
  "n13376643": "cash card, cashcard",
  "n13376764": "debit card",
  "n13377003": "smart card",
  "n13377268": "draft, bill of exchange, order of payment",
  "n13377654": "overdraft",
  "n13377749": "foreign bill, foreign draft",
  "n13377899": "inland bill",
  "n13378031": "redraft",
  "n13378174": "trade acceptance",
  "n13378348": "foreign exchange",
  "n13378518": "credit",
  "n13378717": "cheap money",
  "n13378816": "overage",
  "n13378968": "tax credit",
  "n13379106": "export credit",
  "n13379264": "import credit",
  "n13379413": "credit line, line of credit, bank line, line, personal credit line, personal line of credit",
  "n13379619": "commercial credit",
  "n13379753": "letter of credit",
  "n13380004": "commercial letter of credit",
  "n13380148": "traveler's letter of credit, traveller's letter of credit",
  "n13380309": "traveler's check, traveller's check, banker's check",
  "n13380530": "bank draft, banker's draft",
  "n13380667": "dividend warrant",
  "n13380820": "money order, postal order",
  "n13380996": "overdraft credit, check overdraft credit",
  "n13381145": "deposit, bank deposit",
  "n13381303": "demand deposit",
  "n13381428": "time deposit",
  "n13381602": "acceptance, banker's acceptance",
  "n13381734": "check, bank check, cheque",
  "n13382121": "bad check, bad cheque",
  "n13382306": "kite",
  "n13382471": "kite",
  "n13382614": "counter check",
  "n13382766": "giro, giro cheque",
  "n13382975": "paycheck, payroll check",
  "n13383090": "certified check, certified cheque",
  "n13383289": "personal check, personal cheque",
  "n13383439": "cashier's check, treasurer's check, cashier's cheque, treasurer's cheque",
  "n13383696": "blank check, blank cheque",
  "n13383855": "disability check, disability payment",
  "n13384018": "medicare check, medicare payment",
  "n13384164": "pension",
  "n13384341": "old-age pension, retirement pension, retirement check, retirement benefit, retirement fund, superannuation",
  "n13384557": "money",
  "n13384877": "money",
  "n13385080": "sterling",
  "n13385216": "boodle, bread, cabbage, clams, dinero, dough, gelt, kale, lettuce, lolly, lucre, loot, moolah, pelf, scratch, shekels, simoleons, sugar, wampum",
  "n13385466": "shinplaster",
  "n13385583": "subsidization, subsidisation",
  "n13385778": "token money",
  "n13385913": "currency",
  "n13386136": "Eurocurrency",
  "n13386334": "fractional currency",
  "n13386465": "cash, immediate payment",
  "n13386614": "cash, hard cash, hard currency",
  "n13386882": "hard currency",
  "n13387209": "paper money, folding money, paper currency",
  "n13387479": "change",
  "n13387689": "change",
  "n13387877": "coinage, mintage, specie, metal money",
  "n13388000": "small change, chickenfeed, chump change",
  "n13388111": "change",
  "n13388245": "coin",
  "n13389105": "bawbee",
  "n13389194": "bezant, bezzant, byzant, solidus",
  "n13389359": "denier",
  "n13389475": "ducat",
  "n13389577": "real",
  "n13389660": "piece of eight",
  "n13389763": "shilling",
  "n13389864": "crown",
  "n13389949": "half crown",
  "n13390040": "dime",
  "n13390139": "nickel",
  "n13390244": "quarter",
  "n13390405": "half dollar, fifty-cent piece",
  "n13390525": "halfpenny, ha'penny",
  "n13390626": "penny, cent, centime",
  "n13390788": "slug",
  "n13390857": "tenpence",
  "n13390963": "twopence, tuppence",
  "n13391118": "threepence",
  "n13391262": "fourpence, groat",
  "n13391373": "fivepence",
  "n13391452": "sixpence, tanner",
  "n13391610": "eightpence",
  "n13391693": "ninepence",
  "n13391774": "copper",
  "n13391841": "new penny",
  "n13391967": "dollar",
  "n13392156": "Susan B Anthony dollar",
  "n13392262": "silver dollar, cartwheel",
  "n13392357": "double eagle",
  "n13392472": "eagle",
  "n13392580": "half eagle",
  "n13392688": "guinea",
  "n13392786": "farthing",
  "n13392896": "doubloon",
  "n13392977": "louis d'or",
  "n13393059": "medallion",
  "n13393155": "stater",
  "n13393265": "sou",
  "n13393427": "Maundy money",
  "n13393599": "fiat money",
  "n13393762": "bill, note, government note, bank bill, banker's bill, bank note, banknote, Federal Reserve note, greenback",
  "n13394179": "silver certificate",
  "n13394326": "Treasury, Treasury obligations",
  "n13394592": "Treasury bill, T-bill",
  "n13394792": "Treasury bond",
  "n13394907": "Treasury note",
  "n13395074": "hundred dollar bill, c-note",
  "n13395187": "fifty dollar bill, fifty",
  "n13395296": "twenty dollar bill, twenty",
  "n13395407": "tenner, ten dollar bill",
  "n13395515": "fiver, five-spot, five dollar bill",
  "n13395652": "nickel, nickel note",
  "n13395799": "two dollar bill",
  "n13395897": "dollar, dollar bill, one dollar bill, buck, clam",
  "n13396054": "liabilities",
  "n13396276": "deficit",
  "n13396485": "budget deficit",
  "n13396603": "federal deficit",
  "n13396768": "trade deficit",
  "n13396861": "due",
  "n13397018": "limited liability",
  "n13397174": "debt",
  "n13397443": "arrears",
  "n13397519": "national debt",
  "n13397705": "public debt",
  "n13397932": "debt limit, debt ceiling",
  "n13398074": "national debt ceiling",
  "n13398241": "debt instrument, obligation, certificate of indebtedness",
  "n13398469": "note, promissory note, note of hand",
  "n13398768": "bad debt",
  "n13398859": "installment debt",
  "n13398953": "loan",
  "n13399275": "call loan, demand loan",
  "n13399379": "direct loan",
  "n13399570": "participation loan, loan participation, participation financing",
  "n13399782": "personal loan, consumer loan",
  "n13400024": "automobile loan, auto loan, car loan",
  "n13400150": "point",
  "n13400334": "real estate loan, mortgage loan",
  "n13400472": "time loan",
  "n13400580": "demand note",
  "n13400662": "principal",
  "n13400798": "charge",
  "n13401013": "lien",
  "n13401261": "artisan's lien",
  "n13401412": "federal tax lien",
  "n13401610": "general lien",
  "n13401746": "judgment lien",
  "n13401970": "landlord's lien",
  "n13402209": "mechanic's lien",
  "n13402389": "garageman's lien",
  "n13402529": "state tax lien",
  "n13402690": "tax lien",
  "n13402879": "warehouseman's lien",
  "n13403025": "encumbrance, incumbrance",
  "n13403146": "assessment",
  "n13403331": "document",
  "n13403520": "quittance",
  "n13403643": "record",
  "n13403964": "balance sheet",
  "n13404156": "expense record",
  "n13404248": "ledger, leger, account book, book of account, book",
  "n13404534": "cost ledger",
  "n13404655": "general ledger",
  "n13404879": "subsidiary ledger",
  "n13405015": "control account",
  "n13405166": "daybook, journal",
  "n13405296": "entry, accounting entry, ledger entry",
  "n13405480": "adjusting entry",
  "n13405646": "credit, credit entry",
  "n13405807": "debit, debit entry",
  "n13405962": "accounting, accounting system, method of accounting",
  "n13406374": "credit side",
  "n13406509": "debit side",
  "n13406638": "accrual basis",
  "n13406852": "cash basis",
  "n13407012": "pooling of interest",
  "n13407208": "accounts receivable",
  "n13407389": "note receivable",
  "n13407561": "accounts payable",
  "n13407718": "note payable",
  "n13407844": "profit and loss, profit and loss account",
  "n13408023": "dividend",
  "n13408256": "stock dividend",
  "n13408361": "extra dividend",
  "n13408473": "equalizing dividend",
  "n13408641": "divvy",
  "n13408776": "suspense account",
  "n13408980": "balance",
  "n13409160": "balance",
  "n13409363": "balance of trade, trade balance, visible balance, trade gap",
  "n13409647": "carry-over, carry-forward",
  "n13409850": "compensating balance, offsetting balance",
  "n13410269": "invisible balance",
  "n13410458": "balance of payments, balance of international payments",
  "n13410803": "current account",
  "n13411005": "trial balance",
  "n13411157": "audited account, audit",
  "n13411385": "limited audit",
  "n13411533": "review, limited review",
  "n13411785": "analytical review",
  "n13411943": "expense account, travel and entertainment account",
  "n13412131": "payslip",
  "n13412321": "register",
  "n13412533": "inventory",
  "n13412721": "payroll, paysheet",
  "n13412877": "payroll, paysheet",
  "n13413016": "peanuts",
  "n13413159": "purse",
  "n13413294": "purse",
  "n13413493": "value, economic value",
  "n13413840": "mess of pottage",
  "n13413986": "premium",
  "n13414159": "bankbook, passbook",
  "n13414310": "checkbook, chequebook",
  "n13414443": "pawn ticket",
  "n13414554": "escrow",
  "n13414770": "escrow funds",
  "n13414849": "commercial paper",
  "n13415043": "municipal note",
  "n13415168": "IOU",
  "n13415271": "time note",
  "n13415381": "floater",
  "n13415547": "hotel plan, meal plan",
  "n13415744": "American plan",
  "n13415866": "modified American plan",
  "n13415998": "Bermuda plan",
  "n13416106": "European plan, continental plan",
  "n13416241": "devise",
  "n13416345": "security, certificate",
  "n13416897": "scrip",
  "n13417071": "stock certificate, stock",
  "n13417289": "tax-exempt security, tax-exempt",
  "n13417410": "bond, bond certificate",
  "n13418047": "Premium Bond",
  "n13418219": "warrant, stock warrant, stock-purchase warrant",
  "n13418616": "perpetual warrant",
  "n13418713": "subscription warrant",
  "n13418823": "zero-coupon security, zero coupon security",
  "n13419032": "partnership certificate",
  "n13419178": "proprietorship certificate",
  "n13419325": "convertible, convertible security",
  "n13419535": "letter security",
  "n13419755": "investment letter",
  "n13419951": "treasury stock, treasury shares, reacquired stock",
  "n13420221": "voting stock",
  "n13420359": "watered stock",
  "n13420544": "letter stock",
  "n13420722": "letter bond",
  "n13420900": "listed security",
  "n13421095": "unlisted security, over the counter security, OTC security",
  "n13421286": "over the counter stock, OTC stock, unlisted stock",
  "n13421462": "budget",
  "n13421702": "balanced budget",
  "n13421832": "budget",
  "n13422061": "Civil List",
  "n13422234": "operating budget",
  "n13422411": "petty cash",
  "n13422545": "pocket money, pin money, spending money",
  "n13422684": "ready cash, cold cash, ready money",
  "n13422905": "sight draft, sight bill",
  "n13423007": "time draft, time bill",
  "n13423118": "matching funds",
  "n13423267": "bottom line",
  "n13423405": "ablactation",
  "n13423489": "ablation",
  "n13423615": "abrasion, attrition, corrasion, detrition",
  "n13423779": "abscission",
  "n13423922": "absorption, soaking up",
  "n13424183": "absorption",
  "n13424477": "accession",
  "n13424643": "acclimatization, acclimatisation, acclimation",
  "n13424865": "accretion, accumulation",
  "n13425067": "accretion",
  "n13425245": "accretion",
  "n13425425": "accretion",
  "n13425637": "acetylation",
  "n13425839": "Acheson process",
  "n13425980": "acid-base equilibrium, acid-base balance",
  "n13426238": "acidification",
  "n13426376": "activation",
  "n13426591": "active birth",
  "n13426726": "active transport",
  "n13426948": "acylation",
  "n13427078": "adaptation, adaption, adjustment",
  "n13427356": "addition reaction",
  "n13427481": "adiabatic process",
  "n13427633": "administrative data processing, business data processing",
  "n13427789": "adsorption, surface assimilation",
  "n13427989": "advection",
  "n13428159": "aeration",
  "n13428308": "agenesis, agenesia",
  "n13428421": "agglutination",
  "n13428608": "agglutination",
  "n13428804": "agglutination, agglutinating activity",
  "n13429006": "aging, ageing, senescence",
  "n13429194": "aldol reaction",
  "n13429300": "alluvion",
  "n13429432": "alpha decay",
  "n13429585": "alternative birth, alternative birthing",
  "n13429780": "amelogenesis",
  "n13429888": "Americanization, Americanisation",
  "n13430074": "amitosis",
  "n13430262": "ammonification",
  "n13430391": "amylolysis",
  "n13430495": "anabolism, constructive metabolism",
  "n13430786": "anaglyphy",
  "n13431043": "anamorphism",
  "n13431221": "anamorphosis, anamorphism",
  "n13431380": "anaphase",
  "n13431569": "anastalsis",
  "n13431722": "androgenesis, androgeny",
  "n13431992": "angiogenesis",
  "n13432085": "Anglicization, Anglicisation",
  "n13432249": "anisogamy",
  "n13432443": "anovulation",
  "n13432647": "anthropogenesis, anthropogeny",
  "n13432802": "antiredeposition",
  "n13432903": "antisepsis, asepsis",
  "n13433061": "aphaeresis, apheresis",
  "n13433283": "aphesis",
  "n13433462": "apogamy",
  "n13433727": "apomixis",
  "n13433948": "apposition",
  "n13434120": "asexual reproduction, agamogenesis",
  "n13434406": "assibilation",
  "n13434537": "assimilation",
  "n13434688": "assimilation, absorption",
  "n13434878": "assimilation, absorption",
  "n13435152": "association",
  "n13435377": "asynchronous operation",
  "n13435548": "attack",
  "n13435770": "autocatalysis",
  "n13435918": "autolysis, self-digestion",
  "n13436063": "automatic data processing, ADP",
  "n13436206": "autoradiography",
  "n13436361": "autotype, autotypy",
  "n13436503": "autoregulation",
  "n13436682": "auxesis",
  "n13436809": "auxiliary operation, off-line operation",
  "n13436992": "background processing, backgrounding",
  "n13437181": "backup",
  "n13437379": "bacteriolysis",
  "n13437495": "bacteriostasis",
  "n13437610": "basal metabolic rate, BMR",
  "n13437753": "basal metabolism",
  "n13437902": "batch processing",
  "n13438006": "beach erosion",
  "n13438088": "bed-wetting",
  "n13438223": "Bessemer process",
  "n13438521": "beta decay",
  "n13438671": "biochemical mechanism",
  "n13438817": "biosynthesis, biogenesis",
  "n13438979": "blastogenesis",
  "n13439088": "blaze, blazing",
  "n13439230": "blood coagulation, blood clotting",
  "n13439390": "blooming, bloom",
  "n13439570": "blossoming, flowering, florescence, inflorescence, anthesis, efflorescence",
  "n13439807": "blowing",
  "n13439935": "bluing, blueing",
  "n13440063": "bodily process, body process, bodily function, activity",
  "n13440779": "boiling",
  "n13440935": "boolean operation, binary operation, binary arithmetic operation",
  "n13441154": "bottom fermentation",
  "n13441387": "bowel movement, movement, bm",
  "n13441536": "Bradley method of childbirth, Bradley method",
  "n13441812": "brooding, incubation",
  "n13441959": "budding",
  "n13442168": "buildup",
  "n13442312": "calcification",
  "n13442476": "calcination",
  "n13442639": "calving",
  "n13442715": "capture",
  "n13442867": "capture",
  "n13442998": "carbonation",
  "n13443120": "carbon cycle",
  "n13443228": "carbon cycle",
  "n13443370": "carbonization, carbonisation",
  "n13443537": "carriage return",
  "n13443709": "catabiosis",
  "n13443787": "catabolism, katabolism, dissimilation, destructive metabolism",
  "n13444131": "catalysis, contact action",
  "n13444513": "cavity, caries, dental caries, tooth decay",
  "n13444703": "cell division, cellular division",
  "n13444940": "cenogenesis, kenogenesis, caenogenesis, cainogenesis, kainogenesis",
  "n13445296": "centrifugation",
  "n13445479": "chain reaction",
  "n13445624": "chain reaction",
  "n13445831": "chelation",
  "n13445972": "chelation",
  "n13446197": "chemical equilibrium, equilibrium",
  "n13446390": "chemical process, chemical change, chemical action",
  "n13447361": "chemical reaction, reaction",
  "n13447923": "chemisorption, chemosorption",
  "n13448179": "chemosynthesis",
  "n13448334": "childbirth, childbearing, accouchement, vaginal birth",
  "n13448622": "chlorination",
  "n13448778": "chromatography",
  "n13448970": "civilization, civilisation",
  "n13449156": "cleavage, segmentation",
  "n13449319": "cleavage",
  "n13449450": "climate change, global climate change",
  "n13449566": "clouding, clouding up",
  "n13449714": "cohesion",
  "n13449892": "cold fusion",
  "n13450070": "column chromatography",
  "n13450206": "combustion, burning",
  "n13450417": "deflagration",
  "n13450636": "compensation",
  "n13450862": "computer operation, machine operation",
  "n13451073": "concretion",
  "n13451204": "concurrent operation",
  "n13451348": "condensation",
  "n13451508": "congealment, congelation",
  "n13451665": "conspicuous consumption",
  "n13451804": "consumption, economic consumption, usance, use, use of goods and services",
  "n13452117": "control operation, control function",
  "n13452347": "convalescence, recuperation, recovery",
  "n13452614": "convection",
  "n13452750": "convection",
  "n13452947": "conversion",
  "n13453160": "cooling, chilling, temperature reduction",
  "n13453428": "corrosion, corroding, erosion",
  "n13453640": "corruption",
  "n13453737": "cost-pull inflation",
  "n13453861": "cracking",
  "n13454130": "crossing over, crossover",
  "n13454318": "cultivation",
  "n13454479": "curdling, clotting, coagulation",
  "n13454789": "cyanide process",
  "n13454950": "cytogenesis, cytogeny",
  "n13455086": "cytolysis",
  "n13455234": "data mining",
  "n13455487": "data processing",
  "n13455906": "dealignment",
  "n13456071": "deamination, deaminization",
  "n13456252": "decalcification",
  "n13456367": "decarboxylation",
  "n13456567": "decay, decline",
  "n13456715": "decay",
  "n13456899": "decay, radioactive decay, disintegration",
  "n13457181": "decentralization, decentalisation",
  "n13457378": "decline, diminution",
  "n13457665": "decoction",
  "n13457831": "decoction process, decoction mashing",
  "n13458019": "decomposition, rot, rotting, putrefaction",
  "n13458268": "decomposition, decomposition reaction, chemical decomposition reaction",
  "n13458571": "decrease, decrement",
  "n13458840": "dedifferentiation",
  "n13458968": "deepening",
  "n13459088": "defecation, laxation, shitting",
  "n13459322": "defense mechanism, defense reaction, defence mechanism, defence reaction, defense, defence",
  "n13459821": "deflation",
  "n13460034": "defoliation",
  "n13460129": "degaussing",
  "n13460299": "degeneration, devolution",
  "n13460568": "dehydration, desiccation, drying up, evaporation",
  "n13460863": "de-iodination",
  "n13460991": "demagnetization, demagnetisation",
  "n13461162": "demand",
  "n13461390": "demand-pull inflation",
  "n13461525": "demineralization, demineralisation",
  "n13461722": "denazification, de-Nazification",
  "n13461951": "denial",
  "n13462084": "deossification",
  "n13462191": "deposition, deposit",
  "n13462387": "derivation",
  "n13462680": "eponymy",
  "n13462795": "desalination, desalinization, desalinisation",
  "n13462989": "desertification",
  "n13463255": "desensitization, desensitisation",
  "n13463490": "desorption",
  "n13463656": "destalinization, de-Stalinization, destalinisation, de-Stalinisation",
  "n13464031": "destructive distillation",
  "n13464204": "deterioration, decline in quality, declension, worsening",
  "n13464440": "detumescence",
  "n13464557": "development, developing",
  "n13464820": "development, evolution",
  "n13465264": "diakinesis",
  "n13465382": "diastrophism",
  "n13465530": "diffusion",
  "n13465809": "digestion",
  "n13465998": "digestion",
  "n13466170": "digital photography",
  "n13466312": "dilapidation, ruin",
  "n13466449": "diplotene",
  "n13466586": "discharge, emission, expelling",
  "n13466849": "disinflation",
  "n13467009": "displacement",
  "n13467224": "displacement, displacement reaction",
  "n13467443": "dissimilation",
  "n13467700": "dissociation",
  "n13467916": "dissolution, disintegration",
  "n13468094": "dissolving, dissolution",
  "n13468306": "distillation, distillment",
  "n13468542": "distributed data processing, remote-access data processing, teleprocessing",
  "n13468786": "dithering",
  "n13468954": "domestication",
  "n13469066": "double decomposition, double decomposition reaction, metathesis",
  "n13469317": "double replacement reaction",
  "n13469526": "doubling",
  "n13469674": "drift",
  "n13469893": "drift",
  "n13470015": "dry plate, dry plate process",
  "n13470193": "melioration",
  "n13470392": "dyadic operation",
  "n13470491": "ebb, ebbing, wane",
  "n13470687": "eburnation",
  "n13470868": "ecchymosis",
  "n13471052": "economic growth",
  "n13471206": "economic process",
  "n13471517": "effacement",
  "n13471681": "effervescence",
  "n13471815": "ejaculation",
  "n13471958": "electrodeposition",
  "n13472125": "electrolysis",
  "n13472341": "electronic data processing, EDP",
  "n13472518": "electrophoresis, cataphoresis, dielectrolysis, ionophoresis",
  "n13472862": "electrostatic precipitation",
  "n13473097": "elimination, evacuation, excretion, excreting, voiding",
  "n13473392": "elimination reaction",
  "n13473536": "elision",
  "n13473716": "ellipsis, eclipsis",
  "n13473836": "elution",
  "n13474130": "emergent evolution",
  "n13474290": "emission",
  "n13474495": "encapsulation",
  "n13474615": "endoergic reaction",
  "n13474734": "endothermic reaction",
  "n13474858": "enuresis, urinary incontinence",
  "n13475072": "epigenesis",
  "n13475205": "epilation",
  "n13475320": "epitaxy",
  "n13475538": "erosion, eroding, eating away, wearing, wearing away",
  "n13475944": "erosion",
  "n13476111": "erythropoiesis",
  "n13476267": "establishment, ecesis",
  "n13476440": "Europeanization, Europeanisation",
  "n13476590": "eutrophication",
  "n13477023": "evolution, organic evolution, phylogeny, phylogenesis",
  "n13477462": "execution, instruction execution",
  "n13477691": "exoergic reaction",
  "n13477812": "exothermic reaction",
  "n13477934": "expectoration",
  "n13478055": "exponential decay, exponential return",
  "n13478205": "expression",
  "n13478342": "extinction",
  "n13478525": "extraction",
  "n13478813": "extravasation",
  "n13479034": "farrow, farrowing",
  "n13479169": "fat metabolism",
  "n13479380": "feedback",
  "n13479605": "feminization, feminisation",
  "n13479889": "festering, suppuration, maturation",
  "n13480176": "fibrinolysis",
  "n13480394": "field emission",
  "n13480541": "filling",
  "n13480667": "filtration",
  "n13480848": "fire, flame, flaming",
  "n13481224": "fission, nuclear fission",
  "n13481408": "fission",
  "n13481580": "fissiparity",
  "n13481727": "fixed-cycle operation",
  "n13481883": "flare",
  "n13481994": "floating-point operation, flop",
  "n13482187": "flocculation",
  "n13482330": "flow",
  "n13482580": "flowage",
  "n13482781": "focalization, focalisation",
  "n13482940": "fold, folding",
  "n13483061": "foliation",
  "n13483190": "foliation, leafing",
  "n13483331": "foreground processing, foregrounding",
  "n13483488": "formation",
  "n13483726": "fossilization, fossilisation",
  "n13483971": "fractional distillation",
  "n13484082": "fractionation, fractional process",
  "n13484303": "fragmentation",
  "n13484644": "freeze, freezing",
  "n13484937": "freeze-drying, lyophilization, lyophilisation",
  "n13485270": "frost, icing",
  "n13485408": "fructification",
  "n13485525": "fusion, nuclear fusion, nuclear fusion reaction",
  "n13485767": "fusion",
  "n13485890": "galactosis",
  "n13485968": "galvanization, galvanisation",
  "n13486115": "gametogenesis",
  "n13486270": "gasification",
  "n13486431": "gassing",
  "n13486520": "gastric digestion",
  "n13486671": "gastrulation",
  "n13486838": "geological process, geologic process",
  "n13487207": "germination, sprouting",
  "n13487409": "glaciation",
  "n13487638": "Riss glaciation",
  "n13487791": "Saale glaciation",
  "n13487952": "Wolstonian glaciation",
  "n13488110": "globalization, globalisation",
  "n13488310": "global warming",
  "n13488498": "glycogenesis",
  "n13488606": "glycolysis",
  "n13488844": "growing",
  "n13489037": "growth, growing, maturation, development, ontogeny, ontogenesis",
  "n13490343": "growth",
  "n13490487": "gynogenesis",
  "n13490683": "Haber process, Haber-Bosch process",
  "n13490909": "habit",
  "n13491060": "hardening, solidifying, solidification, set, curing",
  "n13491464": "hatch, hatching",
  "n13491616": "healing",
  "n13491794": "heat dissipation",
  "n13491876": "heating, warming",
  "n13492136": "hemagglutination, haemagglutination",
  "n13492287": "hematochezia, haematochezia",
  "n13492453": "hematopoiesis, haematopoiesis, hemopoiesis, haemopoiesis, hemogenesis, haemogenesis, hematogenesis, haematogenesis, sanguification",
  "n13492789": "hemimetamorphosis, hemimetabolism, hemimetaboly",
  "n13493003": "heterometabolism, heterometaboly",
  "n13493213": "hemolysis, haemolysis, hematolysis, haematolysis",
  "n13493399": "heredity",
  "n13493566": "heterospory",
  "n13493692": "holometabolism, holometaboly",
  "n13493874": "homospory",
  "n13493998": "human process",
  "n13494192": "humification",
  "n13494325": "hydration",
  "n13494472": "hydrocracking",
  "n13494707": "hydrogenation",
  "n13494919": "hydrolysis",
  "n13495209": "hyperhidrosis, hyperidrosis, polyhidrosis",
  "n13495333": "hypersecretion",
  "n13495413": "hypostasis",
  "n13495507": "hypostasis, epistasis",
  "n13495636": "idealization, idealisation",
  "n13495873": "ignition",
  "n13496017": "imbibition",
  "n13496167": "immunoelectrophoresis",
  "n13496286": "implantation, nidation",
  "n13496517": "impregnation, saturation",
  "n13496771": "inactivation",
  "n13496972": "incontinence, incontinency",
  "n13497135": "increase, increment, growth",
  "n13497650": "incrustation, encrustation",
  "n13497784": "induction heating",
  "n13497928": "industrial process",
  "n13498213": "indweller",
  "n13498404": "infection",
  "n13498666": "infection",
  "n13498828": "inflation, rising prices",
  "n13499165": "deflation",
  "n13499393": "inflationary spiral",
  "n13499590": "deflationary spiral",
  "n13499782": "inflow, influx",
  "n13499923": "infructescence",
  "n13500023": "infusion",
  "n13500214": "inhibition",
  "n13500437": "inpouring, inpour, inrush",
  "n13500557": "inspissation",
  "n13500674": "insufflation",
  "n13500860": "integrated data processing, IDP",
  "n13501059": "intellectualization, intellectualisation",
  "n13501261": "internal combustion",
  "n13501405": "intrusion",
  "n13501548": "intumescence, intumescency, swelling",
  "n13501738": "intussusception",
  "n13501941": "invagination, introversion, intussusception, infolding",
  "n13502205": "inversion",
  "n13502397": "involution",
  "n13502556": "iodination",
  "n13502714": "ion exchange",
  "n13502909": "ionization, ionisation",
  "n13503226": "irreversible process",
  "n13503345": "isoagglutination",
  "n13503513": "isogamy",
  "n13503673": "isolation",
  "n13503908": "iteration, looping",
  "n13504173": "iteration, loop",
  "n13504403": "juvenescence",
  "n13504497": "cytokinesis",
  "n13504739": "karyokinesis",
  "n13504947": "karyolysis",
  "n13505069": "katamorphism",
  "n13505249": "keratinization, keratinisation",
  "n13505467": "Krebs cycle, Krebs citric acid cycle, citric acid cycle, tricarboxylic acid cycle",
  "n13505843": "lacrimation, lachrymation, tearing, watering",
  "n13505987": "lactation",
  "n13506119": "Lamaze method of childbirth, Lamaze method",
  "n13506477": "laying, egg laying",
  "n13506587": "leach, leaching",
  "n13506727": "leak, wetting, making water, passing water",
  "n13506906": "Leboyer method of childbirth, Leboyer method",
  "n13507210": "leeway",
  "n13507336": "leptotene",
  "n13507472": "lexicalization, lexicalisation",
  "n13507617": "libration",
  "n13507827": "life cycle",
  "n13508005": "light reaction",
  "n13508183": "line feed",
  "n13508333": "linguistic process",
  "n13508651": "liquefaction",
  "n13508830": "list processing",
  "n13508950": "lithuresis",
  "n13509042": "logic operation, logical operation",
  "n13509196": "loss",
  "n13509383": "lymphopoiesis",
  "n13509528": "lysis",
  "n13509843": "lysis",
  "n13509967": "lysogenization, lysogenisation",
  "n13510152": "maceration",
  "n13510263": "macroevolution",
  "n13510433": "magnetization, magnetisation, magnetic induction",
  "n13510638": "majority operation",
  "n13510829": "malabsorption",
  "n13510946": "marginalization, marginalisation",
  "n13511243": "market forces",
  "n13511368": "Markov chain, Markoff chain",
  "n13511507": "Markov process, Markoff process",
  "n13511755": "masculinization, masculinisation, virilization, virilisation",
  "n13512036": "materialization, materialisation",
  "n13512238": "maturation, ripening, maturement",
  "n13512506": "mechanism, chemical mechanism",
  "n13512725": "meiosis, miosis, reduction division",
  "n13513222": "mellowing",
  "n13513362": "meltdown, nuclear meltdown",
  "n13513540": "menorrhagia, hypermenorrhea",
  "n13513747": "menstruation, menses, menstruum, catamenia, period, flow",
  "n13514314": "metabolism, metabolic process",
  "n13514648": "metamorphism",
  "n13514880": "metamorphosis, metabolism",
  "n13515149": "metaphase",
  "n13515251": "metaphase",
  "n13515353": "metastasis",
  "n13515520": "metathesis",
  "n13515678": "microevolution",
  "n13515828": "microphoning",
  "n13515958": "micturition, urination",
  "n13516176": "mildew, mold, mould",
  "n13516312": "mineral extraction, mineral processing, mineral dressing, ore processing, ore dressing, beneficiation",
  "n13516597": "mitosis",
  "n13516842": "molt, molting, moult, moulting, ecdysis",
  "n13517080": "monadic operation, unary operation",
  "n13517199": "monogenesis, sporulation",
  "n13517385": "morphallaxis",
  "n13517553": "morphogenesis",
  "n13517717": "multiplex operation",
  "n13517843": "multiplication",
  "n13518140": "multiprocessing, parallel processing",
  "n13518279": "multiprogramming, concurrent execution",
  "n13518432": "myelinization, myelinisation",
  "n13518563": "narrowing",
  "n13518656": "natural childbirth",
  "n13518963": "natural process, natural action, action, activity",
  "n13520426": "Nazification",
  "n13520629": "necrolysis",
  "n13520731": "negative feedback",
  "n13520849": "neoplasia",
  "n13520981": "neurogenesis",
  "n13521072": "neutralization, neutralisation, neutralization reaction, neutralisation reaction",
  "n13521444": "new line",
  "n13521616": "nitrification",
  "n13521873": "nitrification",
  "n13522084": "nitrogen cycle",
  "n13522305": "nitrogen fixation",
  "n13522485": "nocturia, nycturia",
  "n13522611": "nocturnal emission",
  "n13522744": "nondevelopment",
  "n13522879": "nondisjunction",
  "n13523088": "nosedive",
  "n13523208": "nuclear reaction",
  "n13523503": "nucleosynthesis",
  "n13523661": "nutrition",
  "n13523983": "obsolescence",
  "n13524191": "oligomenorrhea",
  "n13524295": "oliguria",
  "n13524399": "omission, deletion",
  "n13524612": "oogenesis",
  "n13524686": "opacification",
  "n13524804": "open-hearth process",
  "n13524925": "operation",
  "n13525549": "operation, functioning, performance",
  "n13525912": "opsonization, opsonisation",
  "n13526110": "organic process, biological process",
  "n13527726": "organification",
  "n13527817": "orogeny",
  "n13527965": "oscillation",
  "n13528100": "osmosis",
  "n13528441": "reverse osmosis",
  "n13528750": "ossification",
  "n13528852": "ossification",
  "n13529015": "ossification",
  "n13529167": "osteolysis",
  "n13529295": "outflow, effluence, efflux",
  "n13529450": "overactivity",
  "n13529616": "overcompensation",
  "n13529831": "overflow incontinence",
  "n13530108": "overheating",
  "n13530237": "ovulation",
  "n13530408": "oxidation, oxidization, oxidisation",
  "n13530799": "oxidation-reduction, oxidoreduction, redox",
  "n13530989": "oxidative phosphorylation",
  "n13531149": "oxygenation",
  "n13531318": "pachytene",
  "n13531435": "pair production, pair creation, pair formation",
  "n13531652": "palingenesis, recapitulation",
  "n13531927": "paper chromatography",
  "n13532059": "paper electrophoresis, carrier electrophoresis",
  "n13532196": "parallel operation, simultaneous operation",
  "n13532356": "parthenocarpy",
  "n13532501": "parthenogenesis, parthenogeny",
  "n13532747": "parthenogenesis, parthenogeny, virgin birth",
  "n13532886": "parturition, birth, giving birth, birthing",
  "n13533186": "passive transport",
  "n13533349": "pathogenesis",
  "n13533470": "pathologic process, pathological process",
  "n13533704": "peace process",
  "n13533886": "peeing, pee, pissing, piss",
  "n13534098": "peptization, peptisation",
  "n13534274": "percolation, infiltration",
  "n13534608": "perennation",
  "n13534773": "peristalsis, vermiculation",
  "n13534954": "permeation, pervasion, suffusion",
  "n13535261": "perspiration, sweating, diaphoresis, sudation, hidrosis",
  "n13535583": "petrifaction, petrification",
  "n13535837": "phagocytosis",
  "n13536016": "phase change, phase transition, state change, physical change",
  "n13536299": "phase of cell division",
  "n13536523": "photochemical reaction",
  "n13536661": "photoelectric emission",
  "n13536794": "photography",
  "n13537176": "photomechanics, photoplate making",
  "n13537429": "photosynthesis",
  "n13537622": "pigmentation",
  "n13537761": "pinocytosis",
  "n13537894": "pitting, roughness, indentation",
  "n13538080": "placentation",
  "n13538182": "planation",
  "n13538314": "plastination",
  "n13538757": "polymerization, polymerisation",
  "n13538996": "population growth",
  "n13539143": "irruption",
  "n13539262": "positive feedback, regeneration",
  "n13539422": "potentiation",
  "n13539585": "powder photography, powder method, powder technique",
  "n13539894": "precession of the equinoxes",
  "n13540091": "prechlorination",
  "n13540199": "precipitation",
  "n13540322": "precocious dentition",
  "n13540429": "premature ejaculation",
  "n13540610": "preservation",
  "n13540791": "printing operation",
  "n13540975": "priority processing",
  "n13541167": "processing",
  "n13541491": "professionalization, professionalisation",
  "n13541798": "projection",
  "n13541975": "proliferation",
  "n13542114": "proliferation",
  "n13542374": "prophase",
  "n13542474": "prophase",
  "n13542688": "proteolysis",
  "n13542855": "psilosis",
  "n13542947": "psychoanalytic process",
  "n13543093": "psychogenesis",
  "n13543231": "psychogenesis",
  "n13543418": "psychomotor development",
  "n13543564": "psychosexual development",
  "n13543871": "ptyalism",
  "n13543968": "pullulation",
  "n13544073": "pullulation, gemmation",
  "n13544279": "pyrochemical process, pyrochemistry",
  "n13544433": "quadrupling",
  "n13544537": "quellung, quellung reaction",
  "n13544783": "quickening",
  "n13544930": "quintupling",
  "n13545036": "radiant heating",
  "n13545184": "radiation, radioactivity",
  "n13545382": "radiography, skiagraphy",
  "n13545627": "radiolysis",
  "n13545733": "rain-wash",
  "n13545845": "rally",
  "n13545973": "random walk",
  "n13546169": "rationalization, rationalisation",
  "n13546416": "reaction formation",
  "n13546768": "Read method of childbirth, Read method",
  "n13547048": "real-time processing, real-time operation",
  "n13547199": "rectification",
  "n13547380": "redeposition",
  "n13547513": "reducing",
  "n13547677": "reduction, reducing",
  "n13547925": "refilling, replenishment, replacement, renewal",
  "n13548105": "refining, refinement, purification",
  "n13548350": "reflation",
  "n13548531": "refrigeration, infrigidation",
  "n13548734": "regeneration",
  "n13548931": "regression",
  "n13549105": "regulation",
  "n13549311": "relaxation, relaxation behavior",
  "n13549488": "relaxation",
  "n13549672": "release",
  "n13549916": "replication",
  "n13550089": "repression",
  "n13550318": "reproduction",
  "n13550504": "resorption, reabsorption",
  "n13550755": "reticulation",
  "n13550940": "retrieval",
  "n13551117": "stovepiping",
  "n13551396": "reuptake, re-uptake",
  "n13551617": "reversible process",
  "n13551823": "rigor mortis",
  "n13551965": "ripening, aging, ageing",
  "n13552124": "rooting",
  "n13552270": "rust, rusting",
  "n13552528": "salivation",
  "n13552644": "saltation",
  "n13552858": "saponification",
  "n13553083": "scanning",
  "n13553251": "scattering",
  "n13553418": "schizogony",
  "n13553560": "search, lookup",
  "n13553758": "secondary emission",
  "n13553916": "secretion, secernment",
  "n13554121": "segregation",
  "n13554343": "sensitization, sensitisation",
  "n13554586": "sequestration",
  "n13554800": "serial operation, sequential operation, consecutive operation",
  "n13554984": "serial processing",
  "n13555101": "sericulture",
  "n13555240": "sexual reproduction, amphimixis",
  "n13555446": "shaping, defining",
  "n13555599": "shedding, sloughing",
  "n13555775": "shit, dump",
  "n13555915": "sink",
  "n13556151": "sinking spell",
  "n13556249": "slippage",
  "n13556377": "slippage",
  "n13556509": "slump, slack, drop-off, falloff, falling off",
  "n13556893": "smoke, smoking",
  "n13557158": "soak, soakage, soaking",
  "n13557451": "social process",
  "n13557766": "softening",
  "n13558003": "soil erosion",
  "n13558125": "solvation",
  "n13558325": "Solvay process",
  "n13558490": "sorption",
  "n13558696": "sort, sorting",
  "n13558953": "source",
  "n13559177": "origin",
  "n13559409": "souring",
  "n13559509": "spallation",
  "n13559782": "specialization, specialisation, differentiation",
  "n13560079": "speciation",
  "n13560191": "spiral",
  "n13560329": "spermatogenesis",
  "n13560417": "spoilage, spoiling",
  "n13560586": "spontaneous combustion",
  "n13560738": "stagflation",
  "n13560911": "Stalinization, Stalinisation",
  "n13561198": "stationary stochastic process",
  "n13561396": "steel production",
  "n13561521": "stiffening, rigidifying, rigidification",
  "n13561719": "stimulation",
  "n13561896": "stochastic process",
  "n13562133": "storage",
  "n13562328": "stratification",
  "n13562474": "stress incontinence",
  "n13562680": "subduction",
  "n13562862": "succession, ecological succession",
  "n13563139": "summation",
  "n13563382": "superposition",
  "n13563522": "supply",
  "n13563647": "suppression",
  "n13563746": "survival, survival of the fittest, natural selection, selection",
  "n13563948": "symphysis",
  "n13564045": "synapsis",
  "n13564215": "syncretism",
  "n13564501": "synchronous operation",
  "n13564639": "syneresis, synaeresis",
  "n13564795": "syneresis, synaeresis",
  "n13564910": "synergy, synergism",
  "n13565201": "synizesis, synezesis",
  "n13565379": "synthesis",
  "n13565622": "tanning",
  "n13565781": "teething, dentition, odontiasis",
  "n13565940": "telophase",
  "n13566041": "telophase",
  "n13566212": "temperature change",
  "n13566436": "teratogenesis",
  "n13566535": "thaw, melt, thawing, melting",
  "n13566928": "thermionic emission, thermal emission",
  "n13567063": "thermocoagulation",
  "n13567177": "thermonuclear reaction",
  "n13567344": "threshold operation",
  "n13567515": "thrombolysis",
  "n13567627": "top fermentation",
  "n13567960": "transamination",
  "n13568109": "transamination",
  "n13568238": "transcription",
  "n13568524": "transduction",
  "n13568796": "transduction",
  "n13568983": "translation",
  "n13569226": "protein folding, folding",
  "n13569464": "translocation",
  "n13569664": "translocation",
  "n13569774": "transpiration",
  "n13569905": "transpiration",
  "n13570072": "transpiration",
  "n13570236": "transport",
  "n13570472": "tripling",
  "n13570574": "tumefaction",
  "n13570783": "ulceration",
  "n13570923": "ultracentrifugation",
  "n13571041": "underdevelopment",
  "n13571217": "unfolding, flowering",
  "n13571365": "union, conglutination",
  "n13571580": "uptake",
  "n13571763": "urbanization, urbanisation",
  "n13571943": "urge incontinence",
  "n13572226": "urochesia, urochezia",
  "n13572324": "variation",
  "n13572436": "vaporization, vaporisation, vapor, vapour, evaporation",
  "n13572860": "vascularization, vascularisation",
  "n13573057": "vegetation",
  "n13573181": "vesiculation, vesication, blistering",
  "n13573368": "vicious circle, vicious cycle",
  "n13573499": "video digitizing",
  "n13573666": "vinification",
  "n13573805": "vitrification",
  "n13573915": "vulcanization, vulcanisation",
  "n13574193": "washout, wash",
  "n13574452": "wastage",
  "n13574582": "Westernization, Westernisation",
  "n13574804": "widening, broadening",
  "n13574928": "word processing",
  "n13575109": "zygotene",
  "n13575226": "zymosis",
  "n13575433": "zymosis, zymolysis, fermentation, fermenting, ferment",
  "n13575869": "fundamental quantity, fundamental measure",
  "n13576101": "definite quantity",
  "n13576355": "indefinite quantity",
  "n13576982": "relative quantity",
  "n13577171": "system of measurement, metric",
  "n13577544": "system of weights and measures",
  "n13577731": "British Imperial System, English system, British system",
  "n13577934": "metric system",
  "n13578140": "cgs, cgs system",
  "n13578267": "Systeme International d'Unites, Systeme International, SI system, SI, SI unit, International System of Units, International System",
  "n13578830": "United States Customary System",
  "n13579166": "point system",
  "n13579287": "information measure",
  "n13579510": "bandwidth",
  "n13579679": "baud, baud rate",
  "n13579829": "cordage",
  "n13579933": "octane number, octane rating",
  "n13580058": "utility",
  "n13580228": "marginal utility",
  "n13580415": "enough, sufficiency",
  "n13580723": "fill",
  "n13580909": "normality, N",
  "n13581067": "majority, absolute majority",
  "n13581218": "plurality, relative majority",
  "n13581465": "absolute value, numerical value",
  "n13581598": "acid value",
  "n13581871": "chlorinity",
  "n13582013": "number",
  "n13583030": "quire",
  "n13583137": "ream",
  "n13583272": "solubility",
  "n13583478": "toxicity",
  "n13583611": "cytotoxicity",
  "n13583724": "unit of measurement, unit",
  "n13584646": "measuring unit, measuring block",
  "n13584746": "denier",
  "n13584918": "diopter, dioptre",
  "n13585127": "karat, carat, kt",
  "n13585309": "decimal",
  "n13585429": "constant",
  "n13585819": "Avogadro's number, Avogadro number",
  "n13585997": "Boltzmann's constant",
  "n13586122": "coefficient",
  "n13586455": "absorption coefficient, coefficient of absorption, absorptance",
  "n13586831": "drag coefficient, coefficient of drag",
  "n13587030": "coefficient of friction",
  "n13587236": "coefficient of mutual induction, mutual inductance",
  "n13587525": "coefficient of self induction, self-inductance",
  "n13587763": "modulus",
  "n13587963": "coefficient of elasticity, modulus of elasticity, elastic modulus",
  "n13588233": "bulk modulus",
  "n13588381": "modulus of rigidity",
  "n13588497": "Young's modulus",
  "n13588625": "coefficient of expansion, expansivity",
  "n13588819": "coefficient of reflection, reflection factor, reflectance, reflectivity",
  "n13589003": "transmittance, transmission",
  "n13589140": "coefficient of viscosity, absolute viscosity, dynamic viscosity",
  "n13589321": "weight, weighting",
  "n13589546": "cosmological constant",
  "n13589681": "equilibrium constant",
  "n13589957": "dissociation constant",
  "n13590079": "gas constant, universal gas constant, R",
  "n13590327": "gravitational constant, universal gravitational constant, constant of gravitation, G",
  "n13590598": "Hubble's constant, Hubble constant, Hubble's parameter, Hubble parameter",
  "n13590974": "ionic charge",
  "n13591123": "Planck's constant, h",
  "n13591314": "oxidation number, oxidation state",
  "n13591586": "cardinality",
  "n13591761": "count",
  "n13591998": "complement",
  "n13592105": "blood count",
  "n13592219": "body count",
  "n13592384": "circulation",
  "n13592598": "circulation",
  "n13592764": "head count, headcount",
  "n13592871": "pollen count",
  "n13593065": "sperm count",
  "n13593219": "factor",
  "n13593438": "conversion factor",
  "n13593634": "factor of proportionality, constant of proportionality",
  "n13593908": "Fibonacci number",
  "n13594005": "prime, prime quantity",
  "n13594136": "prime factor",
  "n13594302": "prime number",
  "n13594417": "composite number",
  "n13594585": "score",
  "n13595094": "stroke",
  "n13595276": "birdie",
  "n13595414": "bogey",
  "n13595550": "deficit",
  "n13595699": "double-bogey",
  "n13595844": "duck, duck's egg",
  "n13595968": "eagle",
  "n13596106": "double eagle",
  "n13596235": "game",
  "n13596422": "lead",
  "n13596569": "love",
  "n13596673": "match",
  "n13596756": "par",
  "n13596986": "record",
  "n13597147": "compound number",
  "n13597280": "ordinal number, ordinal, no.",
  "n13597444": "first, number one, number 1",
  "n13597585": "cardinal number, cardinal",
  "n13597794": "base, radix",
  "n13598013": "floating-point number",
  "n13598128": "fixed-point number",
  "n13598237": "frequency, absolute frequency",
  "n13598408": "googol",
  "n13598556": "googolplex",
  "n13598715": "atomic number",
  "n13598960": "magic number",
  "n13599114": "baryon number",
  "n13599348": "quota",
  "n13599469": "long measure",
  "n13599547": "magnetization, magnetisation",
  "n13599730": "magnetic flux",
  "n13599871": "absorption unit",
  "n13599982": "acceleration unit",
  "n13600097": "angular unit",
  "n13600404": "area unit, square measure",
  "n13600822": "volume unit, capacity unit, capacity measure, cubage unit, cubic measure, cubic content unit, displacement unit, cubature unit",
  "n13601370": "cubic inch, cu in",
  "n13601483": "cubic foot, cu ft",
  "n13601596": "computer memory unit",
  "n13602401": "cord",
  "n13602526": "electromagnetic unit, emu",
  "n13602922": "explosive unit",
  "n13603065": "force unit",
  "n13603305": "linear unit, linear measure",
  "n13604275": "metric unit, metric",
  "n13604572": "miles per gallon",
  "n13604718": "monetary unit",
  "n13607187": "megaflop, MFLOP, million floating point operations per second",
  "n13607405": "teraflop, trillion floating point operations per second",
  "n13607616": "MIPS, million instructions per second",
  "n13607873": "pain unit",
  "n13607985": "pressure unit",
  "n13608207": "printing unit",
  "n13608319": "sound unit",
  "n13608480": "telephone unit",
  "n13608598": "temperature unit",
  "n13608788": "weight unit, weight",
  "n13609214": "mass unit",
  "n13609390": "unit of viscosity",
  "n13609507": "work unit, heat unit, energy unit",
  "n13609871": "langley",
  "n13609948": "Brinell number",
  "n13610046": "Brix scale",
  "n13610162": "point",
  "n13610426": "advantage",
  "n13610540": "set point",
  "n13610669": "match point",
  "n13610815": "sabin",
  "n13610987": "circular measure",
  "n13611082": "mil",
  "n13611207": "degree, arcdegree",
  "n13611395": "second, arcsecond",
  "n13611567": "minute, arcminute, minute of arc",
  "n13611740": "microradian",
  "n13611884": "milliradian",
  "n13612042": "radian, rad",
  "n13612319": "grad, grade",
  "n13612427": "oxtant",
  "n13612550": "sextant",
  "n13612688": "straight angle",
  "n13612791": "steradian, sr",
  "n13612943": "square inch, sq in",
  "n13613060": "square foot, sq ft",
  "n13613177": "square yard, sq yd",
  "n13613294": "square meter, square metre, centare",
  "n13613408": "square mile",
  "n13613504": "section",
  "n13613592": "quarter section",
  "n13613742": "acre",
  "n13613862": "are, ar",
  "n13613985": "hectare",
  "n13614143": "arpent",
  "n13614256": "barn, b",
  "n13614467": "dessiatine",
  "n13614572": "morgen",
  "n13614679": "perch, rod, pole",
  "n13614764": "liquid unit, liquid measure",
  "n13615036": "dry unit, dry measure",
  "n13615235": "United States liquid unit",
  "n13615557": "British capacity unit, Imperial capacity unit",
  "n13616054": "metric capacity unit",
  "n13616340": "ardeb",
  "n13616427": "arroba",
  "n13616560": "bath",
  "n13616688": "cran",
  "n13616787": "ephah, epha",
  "n13616926": "field capacity",
  "n13617046": "homer, kor",
  "n13617207": "hin",
  "n13617308": "fathom, fthm",
  "n13617468": "acre-foot",
  "n13617630": "acre inch",
  "n13617713": "board measure",
  "n13617835": "board foot",
  "n13617952": "standard",
  "n13618076": "cubic yard, yard",
  "n13618180": "last",
  "n13618298": "mutchkin",
  "n13618418": "oka",
  "n13618508": "minim",
  "n13618629": "fluidram, fluid dram, fluid drachm, drachm",
  "n13618849": "fluidounce, fluid ounce",
  "n13619028": "gill",
  "n13619168": "cup",
  "n13619307": "pint",
  "n13619475": "fifth",
  "n13619592": "quart",
  "n13619764": "gallon, gal",
  "n13619920": "barrel, bbl",
  "n13620154": "United States dry unit",
  "n13620404": "pint, dry pint",
  "n13620549": "quart, dry quart",
  "n13620713": "peck",
  "n13620871": "bushel",
  "n13621011": "minim",
  "n13621190": "fluidram, fluid dram, fluid drachm, drachm",
  "n13621418": "fluidounce, fluid ounce",
  "n13621660": "gill",
  "n13621850": "pint",
  "n13622035": "quart",
  "n13622209": "gallon, Imperial gallon, congius",
  "n13622451": "peck",
  "n13622591": "bushel",
  "n13622769": "firkin",
  "n13622914": "kilderkin",
  "n13623054": "quarter",
  "n13623205": "hogshead",
  "n13623329": "chaldron",
  "n13623455": "cubic millimeter, cubic millimetre",
  "n13623636": "milliliter, millilitre, mil, ml, cubic centimeter, cubic centimetre, cc",
  "n13623856": "centiliter, centilitre, cl",
  "n13624026": "deciliter, decilitre, dl",
  "n13624190": "liter, litre, l, cubic decimeter, cubic decimetre",
  "n13624509": "dekaliter, dekalitre, decaliter, decalitre, dal, dkl",
  "n13624705": "hectoliter, hectolitre, hl",
  "n13624873": "kiloliter, kilolitre, cubic meter, cubic metre",
  "n13625063": "cubic kilometer, cubic kilometre",
  "n13625237": "bit",
  "n13625482": "parity bit, parity, check bit",
  "n13625884": "nybble, nibble",
  "n13626013": "byte",
  "n13626240": "sector",
  "n13626464": "block",
  "n13626789": "bad block",
  "n13626942": "allocation unit",
  "n13627114": "partition",
  "n13627327": "word",
  "n13627516": "kilobyte, kibibyte, K, KB, kB, KiB",
  "n13627681": "kilobyte, K, KB, kB",
  "n13627810": "kilobit, kbit, kb",
  "n13627935": "kibibit, kibit",
  "n13628056": "megabyte, mebibyte, M, MB, MiB",
  "n13628246": "megabyte, M, MB",
  "n13628419": "megabit, Mbit, Mb",
  "n13628592": "mebibit, Mibit",
  "n13628761": "gigabyte, gibibyte, G, GB, GiB",
  "n13628955": "gigabyte, G, GB",
  "n13629132": "gigabit, Gbit, Gb",
  "n13629309": "gibibit, Gibit",
  "n13629482": "terabyte, tebibyte, TB, TiB",
  "n13629676": "terabyte, TB",
  "n13629854": "terabit, Tbit, Tb",
  "n13630036": "tebibit, Tibit",
  "n13630213": "petabyte, pebibyte, PB, PiB",
  "n13630387": "petabyte, PB",
  "n13630545": "petabit, Pbit, Pb",
  "n13630707": "pebibit, Pibit",
  "n13630864": "exabyte, exbibyte, EB, EiB",
  "n13631037": "exabyte, EB",
  "n13631194": "exabit, Ebit, Eb",
  "n13631355": "exbibit, Eibit",
  "n13631512": "zettabyte, zebibyte, ZB, ZiB",
  "n13631687": "zettabyte, ZB",
  "n13631845": "zettabit, Zbit, Zb",
  "n13632007": "zebibit, Zibit",
  "n13632164": "yottabyte, yobibyte, YB, YiB",
  "n13632320": "yottabyte, YB",
  "n13632461": "yottabit, Ybit, Yb",
  "n13632606": "yobibit, Yibit",
  "n13632744": "capacitance unit",
  "n13632961": "charge unit, quantity unit",
  "n13633229": "conductance unit",
  "n13633375": "current unit",
  "n13633596": "elastance unit",
  "n13633704": "field strength unit",
  "n13633851": "flux density unit",
  "n13634033": "flux unit, magnetic flux unit",
  "n13634205": "inductance unit",
  "n13634418": "light unit",
  "n13634615": "magnetomotive force unit",
  "n13634784": "potential unit",
  "n13635108": "power unit",
  "n13635336": "radioactivity unit",
  "n13635553": "resistance unit",
  "n13635698": "electrostatic unit",
  "n13635854": "picofarad",
  "n13635985": "microfarad",
  "n13636135": "millifarad",
  "n13636286": "farad, F",
  "n13636529": "abfarad",
  "n13636648": "coulomb, C, ampere-second",
  "n13636866": "abcoulomb",
  "n13636989": "ampere-minute",
  "n13637124": "ampere-hour",
  "n13637240": "mho, siemens, reciprocal ohm, S",
  "n13637376": "ampere, amp, A",
  "n13637613": "milliampere, mA",
  "n13637722": "abampere, abamp",
  "n13637841": "ampere, international ampere",
  "n13637988": "daraf",
  "n13638094": "gamma",
  "n13638244": "oersted",
  "n13638375": "maxwell, Mx",
  "n13638577": "weber, Wb",
  "n13638706": "microgauss",
  "n13638847": "gauss",
  "n13639009": "tesla",
  "n13639147": "abhenry",
  "n13639274": "millihenry",
  "n13639405": "henry, H",
  "n13639647": "illumination unit",
  "n13639808": "luminance unit",
  "n13639927": "luminous flux unit",
  "n13640050": "luminous intensity unit, candlepower unit",
  "n13640206": "exposure",
  "n13640371": "footcandle",
  "n13640520": "lambert, L",
  "n13640716": "lux, lx",
  "n13640866": "phot",
  "n13641023": "nit",
  "n13641175": "foot-lambert, ft-L",
  "n13641299": "lumen, lm",
  "n13641534": "candle, candela, cd, standard candle",
  "n13641855": "international candle",
  "n13642000": "gilbert, Gb, Gi",
  "n13642122": "ampere-turn",
  "n13642337": "magneton",
  "n13642464": "abvolt",
  "n13642596": "millivolt, mV",
  "n13642729": "microvolt",
  "n13642856": "nanovolt",
  "n13642982": "picovolt",
  "n13643109": "femtovolt",
  "n13643276": "volt, V",
  "n13643770": "kilovolt, kV",
  "n13643894": "rydberg, rydberg constant, rydberg unit",
  "n13644047": "wave number",
  "n13644165": "abwatt",
  "n13644379": "milliwatt",
  "n13644522": "watt, W",
  "n13644761": "kilowatt, kW",
  "n13644894": "megawatt",
  "n13645010": "horsepower, HP, H.P.",
  "n13645132": "volt-ampere, var",
  "n13645331": "kilovolt-ampere",
  "n13645465": "millicurie",
  "n13645599": "curie, Ci",
  "n13645812": "gray, Gy",
  "n13646031": "roentgen, R",
  "n13646223": "rutherford",
  "n13646367": "REM",
  "n13646557": "rad",
  "n13646694": "abohm",
  "n13646816": "ohm",
  "n13647097": "megohm",
  "n13647215": "kiloton",
  "n13647353": "megaton",
  "n13647498": "dyne",
  "n13647667": "newton, N",
  "n13647888": "sthene",
  "n13647997": "poundal, pdl",
  "n13648184": "pound, lbf.",
  "n13648378": "pounder",
  "n13648596": "g, gee, g-force",
  "n13648793": "gal",
  "n13648947": "Beaufort scale",
  "n13649054": "astronomy unit",
  "n13649268": "metric linear unit",
  "n13649626": "nautical linear unit",
  "n13649791": "inch, in",
  "n13650045": "foot, ft",
  "n13650225": "footer",
  "n13650447": "yard, pace",
  "n13650735": "yarder",
  "n13650921": "perch, rod, pole",
  "n13651072": "furlong",
  "n13651218": "mile, statute mile, stat mi, land mile, international mile, mi",
  "n13651520": "miler",
  "n13651804": "half mile, 880 yards",
  "n13651931": "quarter mile, 440 yards",
  "n13652066": "league",
  "n13652203": "ligne",
  "n13652335": "nail",
  "n13652443": "archine",
  "n13652529": "kos, coss",
  "n13652661": "vara",
  "n13652799": "verst",
  "n13652886": "cable, cable length, cable's length",
  "n13652994": "chain",
  "n13653154": "Gunter's chain",
  "n13653261": "engineer's chain",
  "n13653349": "cubit",
  "n13653461": "finger, fingerbreadth, finger's breadth, digit",
  "n13653615": "fistmele",
  "n13653821": "body length",
  "n13653902": "extremum, peak",
  "n13654093": "hand",
  "n13654232": "handbreadth, handsbreadth",
  "n13654380": "head",
  "n13654571": "lea",
  "n13654674": "li",
  "n13654777": "link",
  "n13654889": "mesh",
  "n13655089": "mil",
  "n13655262": "mile, mil, Swedish mile",
  "n13655414": "mile, Roman mile",
  "n13655570": "Roman pace",
  "n13655790": "geometric pace",
  "n13655908": "military pace",
  "n13656075": "palm",
  "n13656204": "span",
  "n13656345": "survey mile",
  "n13656520": "light year, light-year",
  "n13656749": "light hour",
  "n13656913": "light minute",
  "n13657080": "light second",
  "n13657244": "Astronomical Unit, AU",
  "n13657489": "parsec, secpar",
  "n13657691": "femtometer, femtometre, fermi",
  "n13657849": "picometer, picometre, micromicron",
  "n13658027": "angstrom, angstrom unit, A",
  "n13658278": "nanometer, nanometre, nm, millimicron, micromillimeter, micromillimetre",
  "n13658496": "micron, micrometer",
  "n13658657": "millimeter, millimetre, mm",
  "n13658828": "centimeter, centimetre, cm",
  "n13658998": "decimeter, decimetre, dm",
  "n13659162": "meter, metre, m",
  "n13659419": "decameter, dekameter, decametre, dekametre, dam, dkm",
  "n13659604": "hectometer, hectometre, hm",
  "n13659760": "kilometer, kilometre, km, klick",
  "n13659943": "myriameter, myriametre, mym",
  "n13660084": "nautical chain",
  "n13660178": "fathom, fthm",
  "n13660337": "nautical mile, mile, mi, naut mi, knot, international nautical mile, air mile",
  "n13660619": "nautical mile, naut mi, mile, mi, geographical mile, Admiralty mile",
  "n13660868": "sea mile, mile",
  "n13661045": "halfpennyworth, ha'p'orth",
  "n13661163": "pennyworth, penn'orth",
  "n13661273": "dollar",
  "n13661820": "euro",
  "n13662190": "franc",
  "n13662703": "fractional monetary unit, subunit",
  "n13663077": "Afghan monetary unit",
  "n13663229": "afghani",
  "n13663340": "pul",
  "n13663448": "Argentine monetary unit",
  "n13663562": "austral",
  "n13663694": "Thai monetary unit",
  "n13663820": "baht, tical",
  "n13663933": "satang",
  "n13664041": "Panamanian monetary unit",
  "n13664153": "balboa",
  "n13664283": "Ethiopian monetary unit",
  "n13664396": "birr",
  "n13664521": "cent",
  "n13664808": "centesimo",
  "n13665027": "centimo",
  "n13665256": "centavo",
  "n13665965": "centime",
  "n13666675": "Venezuelan monetary unit",
  "n13666790": "bolivar",
  "n13666922": "Ghanian monetary unit",
  "n13667048": "cedi",
  "n13667150": "pesewa",
  "n13667255": "Costa Rican monetary unit",
  "n13667372": "colon, Costa Rican colon",
  "n13667523": "El Salvadoran monetary unit",
  "n13667643": "colon, El Salvadoran colon",
  "n13667797": "Brazilian monetary unit",
  "n13667908": "real",
  "n13668034": "Gambian monetary unit",
  "n13668161": "dalasi",
  "n13668266": "butut, butat",
  "n13668380": "Algerian monetary unit",
  "n13668491": "Algerian dinar, dinar",
  "n13668632": "Algerian centime",
  "n13668751": "Bahrainian monetary unit",
  "n13668864": "Bahrain dinar, dinar",
  "n13669006": "fils",
  "n13669237": "Iraqi monetary unit",
  "n13669342": "Iraqi dinar, dinar",
  "n13669479": "Jordanian monetary unit",
  "n13669590": "Jordanian dinar, dinar",
  "n13669733": "Kuwaiti monetary unit",
  "n13669860": "Kuwaiti dinar, dinar",
  "n13669998": "Kuwaiti dirham, dirham",
  "n13670156": "Libyan monetary unit",
  "n13670281": "Libyan dinar, dinar",
  "n13670399": "Libyan dirham, dirham",
  "n13670521": "Tunisian monetary unit",
  "n13670668": "Tunisian dinar, dinar",
  "n13670790": "Tunisian dirham, dirham",
  "n13670935": "millime",
  "n13671047": "Yugoslavian monetary unit",
  "n13671182": "Yugoslavian dinar, dinar",
  "n13671310": "para",
  "n13671416": "Moroccan monetary unit",
  "n13671527": "Moroccan dirham, dirham",
  "n13671674": "United Arab Emirate monetary unit",
  "n13671813": "United Arab Emirate dirham, dirham",
  "n13671967": "Australian dollar",
  "n13672077": "Bahamian dollar",
  "n13672177": "Barbados dollar",
  "n13672274": "Belize dollar",
  "n13672367": "Bermuda dollar",
  "n13672462": "Brunei dollar",
  "n13672555": "sen",
  "n13672794": "Canadian dollar, loonie",
  "n13672967": "Cayman Islands dollar",
  "n13673080": "Dominican dollar",
  "n13673178": "Fiji dollar",
  "n13673267": "Grenada dollar",
  "n13673362": "Guyana dollar",
  "n13673455": "Hong Kong dollar",
  "n13673554": "Jamaican dollar",
  "n13673650": "Kiribati dollar",
  "n13673747": "Liberian dollar",
  "n13673843": "New Zealand dollar",
  "n13673946": "Singapore dollar",
  "n13674045": "Taiwan dollar",
  "n13674138": "Trinidad and Tobago dollar",
  "n13674257": "Tuvalu dollar",
  "n13674350": "United States dollar",
  "n13674479": "Eurodollar",
  "n13674660": "Zimbabwean dollar",
  "n13674759": "Vietnamese monetary unit",
  "n13674890": "dong",
  "n13674994": "hao",
  "n13675093": "Greek monetary unit",
  "n13675218": "drachma, Greek drachma",
  "n13675349": "lepton",
  "n13675456": "Sao Thome e Principe monetary unit",
  "n13675591": "dobra",
  "n13675708": "Cape Verde monetary unit",
  "n13675824": "Cape Verde escudo, escudo",
  "n13675976": "Portuguese monetary unit",
  "n13676108": "Portuguese escudo, escudo",
  "n13676285": "conto",
  "n13676396": "Hungarian monetary unit",
  "n13676544": "forint",
  "n13676650": "filler",
  "n13676758": "pengo",
  "n13676897": "Belgian franc",
  "n13677019": "Benin franc",
  "n13677109": "Burundi franc",
  "n13677222": "Cameroon franc",
  "n13677337": "Central African Republic franc",
  "n13677469": "Chadian franc",
  "n13677579": "Congo franc",
  "n13677692": "Djibouti franc",
  "n13677788": "French franc",
  "n13677889": "Gabon franc",
  "n13677998": "Ivory Coast franc, Cote d'Ivoire franc",
  "n13678145": "Luxembourg franc",
  "n13678273": "Madagascar franc",
  "n13678373": "Mali franc",
  "n13678480": "Niger franc",
  "n13678589": "Rwanda franc",
  "n13678700": "Senegalese franc",
  "n13678816": "Swiss franc",
  "n13678931": "Togo franc",
  "n13679038": "Burkina Faso franc",
  "n13679161": "Haitian monetary unit",
  "n13679273": "gourde",
  "n13679377": "Haitian centime",
  "n13679494": "Paraguayan monetary unit",
  "n13679608": "guarani",
  "n13679739": "Dutch monetary unit",
  "n13679855": "guilder, gulden, florin, Dutch florin",
  "n13680032": "Surinamese monetary unit",
  "n13680146": "guilder, gulden, florin",
  "n13680273": "Peruvian monetary unit",
  "n13680381": "inti",
  "n13680463": "Papuan monetary unit",
  "n13680599": "kina",
  "n13680712": "toea",
  "n13680823": "Laotian monetary unit",
  "n13680948": "kip",
  "n13681048": "at",
  "n13681142": "Czech monetary unit",
  "n13681275": "koruna",
  "n13681407": "haler, heller",
  "n13681530": "Slovakian monetary unit",
  "n13681661": "koruna",
  "n13681749": "haler, heller",
  "n13681863": "Icelandic monetary unit",
  "n13681993": "Icelandic krona, krona",
  "n13682116": "eyrir",
  "n13682221": "Swedish monetary unit",
  "n13682330": "Swedish krona, krona",
  "n13682450": "ore",
  "n13682631": "Danish monetary unit",
  "n13682740": "Danish krone, krone",
  "n13682860": "Norwegian monetary unit",
  "n13682971": "Norwegian krone, krone",
  "n13683093": "Malawian monetary unit",
  "n13683221": "Malawi kwacha, kwacha",
  "n13683342": "tambala",
  "n13683451": "Zambian monetary unit",
  "n13683578": "Zambian kwacha, kwacha",
  "n13683700": "ngwee",
  "n13683805": "Angolan monetary unit",
  "n13683932": "kwanza",
  "n13684037": "lwei",
  "n13684140": "Myanmar monetary unit",
  "n13684298": "kyat",
  "n13684402": "pya",
  "n13684503": "Albanian monetary unit",
  "n13684632": "lek",
  "n13684735": "qindarka, qintar",
  "n13684853": "Honduran monetary unit",
  "n13684965": "lempira",
  "n13685096": "Sierra Leone monetary unit",
  "n13685216": "leone",
  "n13685346": "Romanian monetary unit",
  "n13685475": "leu",
  "n13685578": "ban",
  "n13685678": "Bulgarian monetary unit",
  "n13685809": "lev",
  "n13685913": "stotinka",
  "n13686023": "Swaziland monetary unit",
  "n13686137": "lilangeni",
  "n13686268": "Italian monetary unit",
  "n13686376": "lira, Italian lira",
  "n13686526": "British monetary unit",
  "n13686660": "British pound, pound, British pound sterling, pound sterling, quid",
  "n13686877": "British shilling, shilling, bob",
  "n13687015": "Turkish monetary unit",
  "n13687160": "lira, Turkish lira",
  "n13687278": "kurus, piaster, piastre",
  "n13687420": "asper",
  "n13687524": "Lesotho monetary unit",
  "n13687652": "loti",
  "n13687756": "sente",
  "n13687906": "German monetary unit",
  "n13688033": "mark, German mark, Deutsche Mark, Deutschmark",
  "n13688190": "pfennig",
  "n13688319": "Finnish monetary unit",
  "n13688447": "markka, Finnish mark",
  "n13688577": "penni",
  "n13688695": "Mozambique monetary unit",
  "n13688811": "metical",
  "n13688944": "Nigerian monetary unit",
  "n13689073": "naira",
  "n13689178": "kobo",
  "n13689282": "Bhutanese monetary unit",
  "n13689411": "ngultrum",
  "n13689518": "chetrum",
  "n13689630": "Mauritanian monetary unit",
  "n13689765": "ouguiya",
  "n13689875": "khoum",
  "n13689984": "Tongan monetary unit",
  "n13690109": "pa'anga",
  "n13690214": "seniti",
  "n13690321": "Macao monetary unit",
  "n13690445": "pataca",
  "n13690549": "avo",
  "n13690650": "Spanish monetary unit",
  "n13690758": "peseta, Spanish peseta",
  "n13690911": "Bolivian monetary unit",
  "n13691022": "boliviano",
  "n13691154": "Nicaraguan monetary unit",
  "n13691269": "cordoba",
  "n13691401": "Chilean monetary unit",
  "n13691509": "Chilean peso, peso",
  "n13691651": "Colombian monetary unit",
  "n13691764": "Colombian peso, peso",
  "n13691909": "Cuban monetary unit",
  "n13692014": "Cuban peso, peso",
  "n13692151": "Dominican monetary unit",
  "n13692278": "Dominican peso, peso",
  "n13692437": "Guinea-Bissau monetary unit",
  "n13692559": "Guinea-Bissau peso, peso",
  "n13692713": "Mexican monetary unit",
  "n13692822": "Mexican peso, peso",
  "n13692963": "Philippine monetary unit",
  "n13693084": "Philippine peso, peso",
  "n13693237": "Uruguayan monetary unit",
  "n13693349": "Uruguayan peso, peso",
  "n13693514": "Cypriot monetary unit",
  "n13693641": "Cypriot pound, pound",
  "n13693781": "mil",
  "n13693908": "Egyptian monetary unit",
  "n13694017": "Egyptian pound, pound",
  "n13694160": "piaster, piastre",
  "n13694367": "penny",
  "n13694552": "Irish monetary unit",
  "n13694657": "Irish pound, Irish punt, punt, pound",
  "n13694825": "Lebanese monetary unit",
  "n13694936": "Lebanese pound, pound",
  "n13695081": "Maltese monetary unit",
  "n13695189": "lira, Maltese lira",
  "n13695307": "Sudanese monetary unit",
  "n13695420": "Sudanese pound, pound",
  "n13695567": "Syrian monetary unit",
  "n13695674": "Syrian pound, pound",
  "n13695815": "Botswana monetary unit",
  "n13695945": "pula",
  "n13696050": "thebe",
  "n13696155": "Guatemalan monetary unit",
  "n13696270": "quetzal",
  "n13696402": "South African monetary unit",
  "n13696523": "rand",
  "n13696652": "Iranian monetary unit",
  "n13696777": "Iranian rial, rial",
  "n13696893": "Iranian dinar, dinar",
  "n13697011": "Omani monetary unit",
  "n13697134": "riyal-omani, Omani rial, rial",
  "n13697262": "baiza, baisa",
  "n13697380": "Yemeni monetary unit",
  "n13697505": "Yemeni rial, rial",
  "n13697621": "Yemeni fils, fils",
  "n13697727": "Cambodian monetary unit",
  "n13697840": "riel",
  "n13697963": "Malaysian monetary unit",
  "n13698076": "ringgit",
  "n13698202": "Qatari monetary unit",
  "n13698327": "Qatari riyal, riyal",
  "n13698445": "Qatari dirham, dirham",
  "n13698567": "Saudi Arabian monetary unit",
  "n13698706": "Saudi Arabian riyal, riyal",
  "n13698838": "qurush",
  "n13698949": "Russian monetary unit",
  "n13699076": "ruble, rouble",
  "n13699189": "kopek, kopeck, copeck",
  "n13699313": "Armenian monetary unit",
  "n13699442": "dram",
  "n13699527": "lumma",
  "n13699612": "Azerbaijani monetary unit",
  "n13699747": "manat",
  "n13699836": "qepiq",
  "n13699925": "Belarusian monetary unit",
  "n13700056": "rubel",
  "n13700142": "kapeika",
  "n13700233": "Estonian monetary unit",
  "n13700362": "kroon",
  "n13700448": "sent",
  "n13700533": "Georgian monetary unit",
  "n13700662": "lari",
  "n13700766": "tetri",
  "n13700870": "Kazakhstani monetary unit",
  "n13701005": "tenge",
  "n13701094": "tiyin",
  "n13701183": "Latvian monetary unit",
  "n13701310": "lats",
  "n13701394": "santims",
  "n13701482": "Lithuanian monetary unit",
  "n13701615": "litas",
  "n13701703": "centas",
  "n13701793": "Kyrgyzstani monetary unit",
  "n13701928": "som",
  "n13702015": "tyiyn",
  "n13702102": "Moldovan monetary unit",
  "n13702231": "leu",
  "n13702315": "ban",
  "n13702396": "Tajikistani monetary unit",
  "n13702531": "ruble",
  "n13702620": "tanga",
  "n13702707": "Turkmen monetary unit",
  "n13702840": "manat",
  "n13702931": "tenge",
  "n13703022": "Ukranian monetary unit",
  "n13703151": "hryvnia",
  "n13703258": "kopiyka",
  "n13703370": "Uzbekistani monetary unit",
  "n13703505": "som",
  "n13703592": "tiyin",
  "n13703679": "Indian monetary unit",
  "n13703804": "Indian rupee, rupee",
  "n13703942": "paisa",
  "n13704143": "Pakistani monetary unit",
  "n13704274": "Pakistani rupee, rupee",
  "n13704418": "anna",
  "n13704529": "Mauritian monetary unit",
  "n13704643": "Mauritian rupee, rupee",
  "n13704788": "Nepalese monetary unit",
  "n13704897": "Nepalese rupee, rupee",
  "n13705037": "Seychelles monetary unit",
  "n13705153": "Seychelles rupee, rupee",
  "n13705300": "Sri Lankan monetary unit",
  "n13705415": "Sri Lanka rupee, rupee",
  "n13705559": "Indonesian monetary unit",
  "n13705674": "rupiah",
  "n13705800": "Austrian monetary unit",
  "n13705929": "schilling, Austrian schilling",
  "n13706068": "groschen",
  "n13706183": "Israeli monetary unit",
  "n13706310": "shekel",
  "n13706415": "agora",
  "n13706521": "Kenyan monetary unit",
  "n13706628": "Kenyan shilling, shilling",
  "n13706753": "Somalian monetary unit",
  "n13706864": "Somalian shilling, shilling",
  "n13706993": "Tanzanian monetary unit",
  "n13707106": "Tanzanian shilling, shilling",
  "n13707237": "Ugandan monetary unit",
  "n13707346": "Ugandan shilling, shilling",
  "n13707473": "Ecuadoran monetary unit",
  "n13707585": "sucre",
  "n13707713": "Guinean monetary unit",
  "n13707804": "Guinean franc",
  "n13707897": "Bangladeshi monetary unit",
  "n13708014": "taka",
  "n13708141": "Western Samoan monetary unit",
  "n13708282": "tala",
  "n13708392": "sene",
  "n13708500": "Mongolian monetary unit",
  "n13708631": "tugrik, tughrik",
  "n13708748": "mongo",
  "n13708855": "North Korean monetary unit",
  "n13708992": "North Korean won, won",
  "n13709118": "chon",
  "n13709223": "South Korean monetary unit",
  "n13709360": "South Korean won, won",
  "n13709486": "chon",
  "n13709591": "Japanese monetary unit",
  "n13709700": "yen",
  "n13709819": "Chinese monetary unit",
  "n13709992": "yuan, kwai",
  "n13710101": "jiao",
  "n13710219": "fen",
  "n13710317": "Zairese monetary unit",
  "n13710443": "zaire",
  "n13710546": "likuta",
  "n13710651": "Polish monetary unit",
  "n13710777": "zloty",
  "n13710881": "grosz",
  "n13710986": "dol",
  "n13711060": "standard atmosphere, atmosphere, atm, standard pressure",
  "n13711303": "pascal, Pa",
  "n13711416": "torr, millimeter of mercury, mm Hg",
  "n13711570": "pounds per square inch, psi",
  "n13711663": "millibar",
  "n13711855": "bar",
  "n13712120": "barye, bar absolute, microbar",
  "n13712286": "point",
  "n13712428": "em, pica em, pica",
  "n13712592": "en, nut",
  "n13712689": "em, em quad, mutton quad",
  "n13712890": "cicero",
  "n13713003": "agate line, line",
  "n13713159": "milline",
  "n13713300": "column inch, inch",
  "n13713408": "linage, lineage",
  "n13713521": "Bel, B",
  "n13713633": "decibel, dB",
  "n13713817": "sone",
  "n13713983": "phon",
  "n13714082": "Erlang",
  "n13714184": "degree",
  "n13714376": "millidegree",
  "n13714491": "degree centigrade, degree Celsius, C",
  "n13714641": "degree Fahrenheit, F",
  "n13714756": "kelvin, K",
  "n13714905": "Rankine",
  "n13715001": "degree day",
  "n13715124": "standard temperature",
  "n13715240": "poise",
  "n13715495": "atomic mass unit",
  "n13715614": "mass number, nucleon number",
  "n13715755": "system of weights, weight",
  "n13715921": "avoirdupois, avoirdupois weight",
  "n13716084": "avoirdupois unit",
  "n13716490": "troy, troy weight",
  "n13716686": "troy unit",
  "n13716878": "apothecaries' unit, apothecaries' weight",
  "n13717155": "metric weight unit, weight unit",
  "n13717619": "arroba",
  "n13717728": "catty, cattie",
  "n13717914": "crith",
  "n13718046": "frail",
  "n13718178": "last",
  "n13718267": "maund",
  "n13718451": "obolus",
  "n13718574": "oka",
  "n13718675": "picul",
  "n13718838": "pood",
  "n13718946": "rotl",
  "n13719102": "slug",
  "n13719291": "tael",
  "n13719410": "tod",
  "n13719529": "welterweight",
  "n13719683": "grain",
  "n13719808": "dram",
  "n13719922": "ounce, oz.",
  "n13720096": "pound, lb",
  "n13720302": "pound",
  "n13720405": "half pound",
  "n13720501": "quarter pound",
  "n13720600": "stone",
  "n13720852": "quarter",
  "n13721003": "hundredweight, cwt, long hundredweight",
  "n13721177": "hundredweight, cwt, short hundredweight, centner, cental, quintal",
  "n13721387": "long ton, ton, gross ton",
  "n13721529": "short ton, ton, net ton",
  "n13721695": "kiloton",
  "n13721804": "megaton",
  "n13721893": "grain",
  "n13722060": "scruple",
  "n13722198": "pennyweight",
  "n13722340": "dram, drachm, drachma",
  "n13722522": "ounce, troy ounce, apothecaries' ounce",
  "n13722757": "troy pound, apothecaries' pound",
  "n13722929": "microgram, mcg",
  "n13723061": "milligram, mg",
  "n13723189": "nanogram, ng",
  "n13723304": "grain, metric grain",
  "n13723470": "decigram, dg",
  "n13723577": "carat",
  "n13723712": "gram, gramme, gm, g",
  "n13723899": "gram atom, gram-atomic weight",
  "n13724081": "gram molecule, mole, mol",
  "n13724350": "dekagram, decagram, dkg, dag",
  "n13724474": "hectogram, hg",
  "n13724582": "kilogram, kg, kilo",
  "n13724838": "key",
  "n13724977": "myriagram, myg",
  "n13725108": "centner",
  "n13725271": "hundredweight, metric hundredweight, doppelzentner, centner",
  "n13725457": "quintal",
  "n13725588": "metric ton, MT, tonne, t",
  "n13725726": "erg",
  "n13725902": "electron volt, eV",
  "n13726074": "joule, J, watt second",
  "n13726296": "calorie, gram calorie, small calorie",
  "n13726562": "Calorie, kilogram calorie, kilocalorie, large calorie, nutritionist's calorie",
  "n13726947": "British thermal unit, BTU, B.Th.U.",
  "n13727209": "therm",
  "n13727333": "watt-hour",
  "n13727478": "kilowatt hour, kW-hr, Board of Trade unit, B.T.U.",
  "n13727683": "foot-pound",
  "n13727841": "foot-ton",
  "n13727931": "foot-poundal",
  "n13728074": "horsepower-hour",
  "n13728203": "kilogram-meter",
  "n13728367": "natural number",
  "n13728499": "integer, whole number",
  "n13728818": "addend",
  "n13728924": "augend",
  "n13729031": "minuend",
  "n13729135": "subtrahend",
  "n13729236": "remainder, difference",
  "n13729428": "complex number, complex quantity, imaginary number, imaginary",
  "n13729732": "complex conjugate",
  "n13729902": "real number, real",
  "n13730054": "pure imaginary number",
  "n13730189": "imaginary part, imaginary part of a complex number",
  "n13730377": "modulus",
  "n13730469": "rational number, rational",
  "n13730584": "irrational number, irrational",
  "n13730756": "transcendental number",
  "n13730902": "algebraic number",
  "n13731021": "square, second power",
  "n13731241": "cube, third power",
  "n13731356": "biquadrate, biquadratic, quartic, fourth power",
  "n13731530": "radical",
  "n13731669": "root",
  "n13731840": "square root",
  "n13731959": "cube root",
  "n13732078": "fraction",
  "n13732295": "common fraction, simple fraction",
  "n13732871": "numerator",
  "n13732953": "dividend",
  "n13733066": "denominator",
  "n13733167": "divisor",
  "n13733316": "quotient",
  "n13733402": "divisor, factor",
  "n13733663": "remainder",
  "n13733818": "multiplier, multiplier factor",
  "n13733979": "multiplicand",
  "n13734085": "scale factor",
  "n13734202": "time-scale factor",
  "n13734349": "equivalent-binary-digit factor",
  "n13734629": "aliquot, aliquot part",
  "n13734816": "aliquant, aliquant part",
  "n13734992": "common divisor, common factor, common measure",
  "n13735163": "greatest common divisor, greatest common factor, highest common factor",
  "n13735355": "common multiple",
  "n13735476": "common denominator",
  "n13735622": "modulus",
  "n13735800": "improper fraction",
  "n13735921": "proper fraction",
  "n13736057": "complex fraction, compound fraction",
  "n13736197": "decimal fraction, decimal",
  "n13736376": "circulating decimal, recurring decimal, repeating decimal",
  "n13736550": "continued fraction",
  "n13736799": "one-half, half",
  "n13736997": "fifty percent",
  "n13737089": "moiety, mediety",
  "n13737190": "one-third, third, tierce",
  "n13737378": "two-thirds",
  "n13737480": "one-fourth, fourth, one-quarter, quarter, fourth part, twenty-five percent, quartern",
  "n13737700": "three-fourths, three-quarters",
  "n13737830": "one-fifth, fifth, fifth part, twenty percent",
  "n13737952": "one-sixth, sixth",
  "n13738043": "one-seventh, seventh",
  "n13738140": "one-eighth, eighth",
  "n13738235": "one-ninth, ninth",
  "n13738327": "one-tenth, tenth, tenth part, ten percent",
  "n13738459": "one-twelfth, twelfth, twelfth part, duodecimal",
  "n13738585": "one-sixteenth, sixteenth, sixteenth part",
  "n13738705": "one-thirty-second, thirty-second, thirty-second part",
  "n13738840": "one-sixtieth, sixtieth",
  "n13738939": "one-sixty-fourth, sixty-fourth",
  "n13739051": "one-hundredth, hundredth, one percent",
  "n13739170": "one-thousandth, thousandth",
  "n13739278": "one-ten-thousandth, ten-thousandth",
  "n13739396": "one-hundred-thousandth",
  "n13739507": "one-millionth, millionth",
  "n13739612": "one-hundred-millionth",
  "n13739721": "one-billionth, billionth",
  "n13739826": "one-trillionth, trillionth",
  "n13739934": "one-quadrillionth, quadrillionth",
  "n13740051": "one-quintillionth, quintillionth",
  "n13740168": "nothing, nil, nix, nada, null, aught, cipher, cypher, goose egg, naught, zero, zilch, zip, zippo",
  "n13740591": "nihil",
  "n13740765": "bugger all, fuck all, Fanny Adams, sweet Fanny Adams",
  "n13741022": "digit, figure",
  "n13741512": "binary digit",
  "n13741603": "octal digit",
  "n13741698": "decimal digit",
  "n13741797": "duodecimal digit",
  "n13741903": "hexadecimal digit",
  "n13742011": "significant digit, significant figure",
  "n13742358": "zero, 0, nought, cipher, cypher",
  "n13742573": "one, 1, I, ace, single, unity",
  "n13742840": "monad, monas",
  "n13742980": "singleton",
  "n13743100": "mate, fellow",
  "n13743269": "two, 2, II, deuce",
  "n13743460": "craps, snake eyes",
  "n13743605": "couple, pair, twosome, twain, brace, span, yoke, couplet, distich, duo, duet, dyad, duad",
  "n13743869": "doubleton",
  "n13744044": "three, 3, III, trio, threesome, tierce, leash, troika, triad, trine, trinity, ternary, ternion, triplet, tercet, terzetto, trey, deuce-ace",
  "n13744304": "four, 4, IV, tetrad, quatern, quaternion, quaternary, quaternity, quartet, quadruplet, foursome, Little Joe",
  "n13744521": "five, 5, V, cinque, quint, quintet, fivesome, quintuplet, pentad, fin, Phoebe, Little Phoebe",
  "n13744722": "six, 6, VI, sixer, sise, Captain Hicks, half a dozen, sextet, sestet, sextuplet, hexad",
  "n13744916": "seven, 7, VII, sevener, heptad, septet, septenary",
  "n13745086": "eight, 8, VIII, eighter, eighter from Decatur, octad, ogdoad, octonary, octet",
  "n13745270": "nine, 9, IX, niner, Nina from Carolina, ennead",
  "n13745420": "large integer",
  "n13746419": "double digit",
  "n13746512": "ten, 10, X, tenner, decade",
  "n13746672": "eleven, 11, XI",
  "n13746785": "twelve, 12, XII, dozen",
  "n13746946": "boxcars",
  "n13747114": "teens",
  "n13747199": "thirteen, 13, XIII, baker's dozen, long dozen",
  "n13747348": "fourteen, 14, XIV",
  "n13747469": "fifteen, 15, XV",
  "n13747606": "sixteen, 16, XVI",
  "n13747725": "seventeen, 17, XVII",
  "n13747865": "eighteen, 18, XVIII",
  "n13747989": "nineteen, 19, XIX",
  "n13748128": "twenty, 20, XX",
  "n13748246": "twenty-one, 21, XXI",
  "n13748367": "twenty-two, 22, XXII",
  "n13748493": "twenty-three, 23, XXIII",
  "n13748622": "twenty-four, 24, XXIV, two dozen",
  "n13748763": "twenty-five, 25, XXV",
  "n13748890": "twenty-six, 26, XXVI",
  "n13749017": "twenty-seven, 27, XXVII",
  "n13749146": "twenty-eight, 28, XXVIII",
  "n13749278": "twenty-nine, 29, XXIX",
  "n13749407": "thirty, 30, XXX",
  "n13749527": "forty, 40, XL",
  "n13749644": "fifty, 50, L",
  "n13749778": "sixty, 60, LX",
  "n13749894": "seventy, 70, LXX",
  "n13750033": "seventy-eight, 78, LXXVIII",
  "n13750164": "eighty, 80, LXXX, fourscore",
  "n13750297": "ninety, 90, XC",
  "n13750415": "hundred, 100, C, century, one C",
  "n13750504": "gross, 144",
  "n13750574": "long hundred, great hundred, 120",
  "n13750712": "five hundred, 500, D",
  "n13750844": "thousand, one thousand, 1000, M, K, chiliad, G, grand, thou, yard",
  "n13751036": "millenary",
  "n13751158": "great gross, 1728",
  "n13751265": "ten thousand, 10000, myriad",
  "n13751404": "hundred thousand, 100000, lakh",
  "n13751533": "million, 1000000, one thousand thousand, meg",
  "n13751686": "crore",
  "n13751829": "billion, one thousand million, 1000000000",
  "n13752033": "milliard",
  "n13752172": "billion, one million million, 1000000000000",
  "n13752443": "trillion, one million million, 1000000000000",
  "n13752679": "trillion, one million million million",
  "n13752911": "quadrillion",
  "n13753067": "quadrillion",
  "n13753274": "quintillion",
  "n13753430": "sextillion",
  "n13753585": "septillion",
  "n13753740": "octillion",
  "n13753894": "aleph-null, aleph-nought, aleph-zero",
  "n13754008": "pi",
  "n13754165": "e",
  "n13754293": "addition, increase, gain",
  "n13754565": "accretion",
  "n13754778": "bag",
  "n13754948": "breakage",
  "n13755053": "capacity",
  "n13755342": "formatted capacity",
  "n13755627": "unformatted capacity",
  "n13755822": "catch, haul",
  "n13755957": "correction, fudge factor",
  "n13756125": "containerful",
  "n13757249": "footstep, pace, step, stride",
  "n13757582": "headspace",
  "n13757724": "large indefinite quantity, large indefinite amount",
  "n13758197": "chunk",
  "n13758296": "limit, limitation",
  "n13758556": "limit, limit point, point of accumulation",
  "n13758745": "output, yield, production",
  "n13759014": "cutoff",
  "n13759146": "region, neighborhood",
  "n13759428": "outage",
  "n13759558": "picking, pick",
  "n13759773": "reserve",
  "n13759941": "pulmonary reserve",
  "n13760129": "run",
  "n13760316": "small indefinite quantity, small indefinite amount",
  "n13760980": "crumb",
  "n13761171": "dab, splash, splatter",
  "n13761407": "spot, bit",
  "n13761603": "hair's-breadth, hairsbreadth, hair, whisker",
  "n13761801": "modicum",
  "n13761966": "scattering, sprinkling",
  "n13762178": "shoestring, shoe string",
  "n13762305": "spray, spraying",
  "n13762458": "nose",
  "n13762579": "step, stone's throw",
  "n13762717": "little",
  "n13762836": "shtik, shtick, schtik, schtick",
  "n13763058": "shtikl, shtickl, schtikl, schtickl",
  "n13763185": "tad, shade",
  "n13763384": "minimum, lower limit",
  "n13763626": "skeleton",
  "n13763806": "spillage",
  "n13763888": "spoilage",
  "n13763970": "tankage",
  "n13764086": "ullage",
  "n13764213": "top-up",
  "n13764342": "worth",
  "n13764540": "armful",
  "n13764639": "bag, bagful",
  "n13764764": "barrel, barrelful",
  "n13764897": "barrow, barrowful",
  "n13764998": "barnful",
  "n13765086": "basin, basinful",
  "n13765207": "basket, basketful",
  "n13765306": "bin, binful",
  "n13765396": "bottle, bottleful",
  "n13765531": "bowl, bowlful",
  "n13765624": "box, boxful",
  "n13765749": "bucket, bucketful",
  "n13765866": "busload",
  "n13765990": "can, canful",
  "n13766080": "capful",
  "n13766166": "carful",
  "n13766252": "cartload",
  "n13766337": "carton, cartonful",
  "n13766436": "case, caseful",
  "n13766547": "cask, caskful",
  "n13766637": "crate, crateful",
  "n13766733": "cup, cupful",
  "n13766896": "dish, dishful",
  "n13767042": "dustpan, dustpanful",
  "n13767146": "flask, flaskful",
  "n13767239": "glass, glassful",
  "n13767350": "handful, fistful",
  "n13767455": "hatful",
  "n13767545": "headful",
  "n13767691": "houseful",
  "n13767822": "jar, jarful",
  "n13767956": "jug, jugful",
  "n13768064": "keg, kegful",
  "n13768154": "kettle, kettleful",
  "n13768250": "lapful",
  "n13768343": "mouthful",
  "n13768440": "mug, mugful",
  "n13768537": "pail, pailful",
  "n13768630": "pipeful",
  "n13768748": "pitcher, pitcherful",
  "n13768850": "plate, plateful",
  "n13768946": "pocketful",
  "n13769033": "pot, potful",
  "n13769123": "roomful",
  "n13769206": "sack, sackful",
  "n13769317": "scoop, scoopful",
  "n13769428": "shelfful",
  "n13769549": "shoeful",
  "n13769672": "skinful",
  "n13769870": "shovel, shovelful, spadeful",
  "n13769994": "skepful",
  "n13770076": "split",
  "n13770169": "spoon, spoonful",
  "n13770310": "tablespoon, tablespoonful",
  "n13770416": "dessertspoon, dessertspoonful",
  "n13770529": "tank, tankful",
  "n13770635": "teacup, teacupful",
  "n13770729": "teaspoon, teaspoonful",
  "n13770829": "thimble, thimbleful",
  "n13770926": "tub, tubful",
  "n13771034": "morsel",
  "n13771154": "handful, smattering",
  "n13771290": "couple",
  "n13771404": "drop, drib, driblet",
  "n13771828": "droplet",
  "n13771929": "eyedrop, eye-drop",
  "n13772020": "dollop",
  "n13772106": "dose, dosage",
  "n13772313": "load",
  "n13772468": "load, loading",
  "n13772653": "precipitation",
  "n13772876": "trainload",
  "n13772971": "dreg",
  "n13773047": "jack, doodly-squat, diddly-squat, diddlysquat, diddly-shit, diddlyshit, diddly, diddley, squat, shit",
  "n13773250": "nip, shot",
  "n13773361": "trace, hint, suggestion",
  "n13773539": "spark",
  "n13773725": "shred, scintilla, whit, iota, tittle, smidgen, smidgeon, smidgin, smidge",
  "n13773906": "tot",
  "n13774010": "snuff",
  "n13774115": "touch, hint, tinge, mite, pinch, jot, speck, soupcon",
  "n13774311": "barrels",
  "n13774404": "batch, deal, flock, good deal, great deal, hatful, heap, lot, mass, mess, mickle, mint, mountain, muckle, passel, peck, pile, plenty, pot, quite a little, raft, sight, slew, spate, stack, tidy sum, wad",
  "n13775093": "battalion, large number, multitude, plurality, pack",
  "n13775319": "billyo, billyoh, billy-ho, all get out",
  "n13775523": "boatload, shipload, carload",
  "n13775706": "flood, inundation, deluge, torrent",
  "n13775939": "haymow",
  "n13776060": "infinitude",
  "n13776137": "maximum, upper limit",
  "n13776342": "mile",
  "n13776432": "million, billion, trillion, zillion, jillion, gazillion",
  "n13776621": "much",
  "n13776726": "myriad",
  "n13776854": "reservoir",
  "n13776971": "ocean, sea",
  "n13777098": "ream",
  "n13777211": "small fortune",
  "n13777344": "supply",
  "n13777509": "tons, dozens, heaps, lots, piles, scores, stacks, loads, rafts, slews, wads, oodles, gobs, scads, lashings",
  "n13777764": "room, way, elbow room",
  "n13778089": "breathing room, breathing space",
  "n13778261": "headroom, headway, clearance",
  "n13778400": "houseroom",
  "n13778530": "living space, lebensraum",
  "n13778671": "parking",
  "n13778827": "sea room",
  "n13778907": "swath",
  "n13779032": "volume",
  "n13779244": "volume",
  "n13779374": "capacity, content",
  "n13779570": "vital capacity",
  "n13779804": "population",
  "n13780180": "proof",
  "n13780339": "STP, s.t.p.",
  "n13780449": "relations, dealings",
  "n13780606": "causality",
  "n13780719": "relationship, human relationship",
  "n13781164": "function",
  "n13781348": "partnership",
  "n13781670": "personal relation, personal relationship",
  "n13781820": "bonding",
  "n13782033": "obligation, indebtedness",
  "n13782208": "female bonding",
  "n13782329": "male bonding",
  "n13782533": "maternal-infant bonding",
  "n13782769": "association",
  "n13783038": "logical relation",
  "n13783259": "contradictory",
  "n13783421": "contrary",
  "n13783581": "mathematical relation",
  "n13783816": "function, mathematical function, single-valued function, map, mapping",
  "n13784366": "expansion",
  "n13784537": "inverse function",
  "n13784763": "Kronecker delta",
  "n13784906": "metric function, metric",
  "n13785136": "transformation",
  "n13785404": "reflection",
  "n13785557": "rotation",
  "n13785736": "translation",
  "n13785965": "affine transformation",
  "n13786187": "isometry",
  "n13786413": "operator",
  "n13786595": "linear operator",
  "n13786748": "identity, identity element, identity operator",
  "n13786960": "trigonometric function, circular function",
  "n13787379": "sine, sin",
  "n13787551": "arc sine, arcsine, arcsin, inverse sine",
  "n13787723": "cosine, cos",
  "n13787853": "arc cosine, arccosine, arccos, inverse cosine",
  "n13788035": "tangent, tan",
  "n13788182": "arc tangent, arctangent, arctan, inverse tangent",
  "n13788369": "cotangent, cotan",
  "n13788502": "arc cotangent, arccotangent, inverse cotangent",
  "n13788690": "secant, sec",
  "n13788820": "arc secant, arcsecant, arcsec, inverse secant",
  "n13789002": "cosecant, cosec",
  "n13789136": "arc cosecant, arccosecant, inverse cosecant",
  "n13789281": "threshold function",
  "n13789462": "exponential, exponential function",
  "n13789627": "exponential equation",
  "n13789751": "exponential curve",
  "n13789849": "exponential expression",
  "n13790004": "exponential series",
  "n13790133": "parity",
  "n13790492": "evenness",
  "n13790592": "oddness",
  "n13790712": "foundation",
  "n13790912": "footing, basis, ground",
  "n13791122": "common ground",
  "n13791250": "grass roots",
  "n13791389": "connection, connexion, connectedness",
  "n13791910": "series",
  "n13792183": "alliance, bond",
  "n13792468": "silver cord",
  "n13792579": "linkage",
  "n13792692": "link, nexus",
  "n13792842": "communication",
  "n13793127": "concatenation",
  "n13793330": "bridge",
  "n13793504": "involvement",
  "n13793776": "implication",
  "n13794034": "inclusion, comprehension",
  "n13794226": "unconnectedness",
  "n13794417": "relevance, relevancy",
  "n13794658": "materiality",
  "n13794793": "cogency",
  "n13794885": "point",
  "n13795042": "germaneness",
  "n13795180": "applicability, pertinence, pertinency",
  "n13795489": "relatedness",
  "n13795695": "bearing",
  "n13795834": "irrelevance, irrelevancy",
  "n13796053": "inapplicability",
  "n13796216": "immateriality",
  "n13796403": "unrelatedness",
  "n13796585": "extraneousness",
  "n13796779": "grammatical relation",
  "n13797142": "linguistic relation",
  "n13797313": "agreement, concord",
  "n13797561": "number agreement",
  "n13797747": "person agreement",
  "n13797856": "case agreement",
  "n13797985": "gender agreement",
  "n13798118": "transitivity, transitiveness",
  "n13798301": "intransitivity, intransitiveness",
  "n13798491": "transitivity",
  "n13798814": "reflexivity, reflexiveness",
  "n13799063": "coreference",
  "n13799245": "reflexivity, reflexiveness",
  "n13799392": "conjunction",
  "n13799655": "coordinating conjunction",
  "n13799796": "subordinating conjunction",
  "n13799967": "copulative conjunction",
  "n13800115": "disjunctive conjunction",
  "n13800267": "adversative conjunction",
  "n13800418": "complementation",
  "n13800539": "coordination",
  "n13800675": "subordination",
  "n13800801": "modification, qualifying, limiting",
  "n13801040": "restrictiveness",
  "n13801217": "apposition",
  "n13801424": "mood, mode, modality",
  "n13801700": "indicative mood, indicative, declarative mood, declarative, common mood, fact mood",
  "n13801937": "subjunctive mood, subjunctive",
  "n13802098": "optative mood, optative",
  "n13802306": "imperative mood, imperative, jussive mood, imperative form",
  "n13802485": "interrogative mood, interrogative",
  "n13802634": "modality, mode",
  "n13802805": "anaphoric relation",
  "n13802920": "voice",
  "n13803158": "active voice, active",
  "n13803431": "passive voice, passive",
  "n13803782": "inflection, inflexion",
  "n13804085": "conjugation",
  "n13804203": "declension",
  "n13804375": "paradigm",
  "n13804507": "pluralization, pluralisation",
  "n13804669": "aspect",
  "n13804889": "perfective, perfective aspect",
  "n13805020": "imperfective, imperfective aspect",
  "n13805179": "durative, durative aspect",
  "n13805336": "progressive aspect",
  "n13805456": "inchoative, inchoative aspect",
  "n13805594": "iterative, iterative aspect",
  "n13805734": "progressive, progressive tense, imperfect, imperfect tense, continuous tense",
  "n13805974": "present progressive, present progressive tense",
  "n13806140": "perfective, perfective tense, perfect, perfect tense",
  "n13806404": "present perfect, present perfect tense",
  "n13806614": "preterit, preterite",
  "n13806735": "past perfect, past perfect tense, pluperfect, pluperfect tense",
  "n13806964": "past progressive, past progressive tense",
  "n13807178": "future perfect, future perfect tense",
  "n13807403": "future progressive, future progressive tense",
  "n13807636": "semantic relation",
  "n13807853": "hyponymy, subordination",
  "n13808004": "hypernymy, superordination",
  "n13808161": "synonymy, synonymity, synonymousness",
  "n13808405": "antonymy",
  "n13808566": "holonymy, whole to part relation",
  "n13808708": "meronymy, part to whole relation",
  "n13808868": "troponymy",
  "n13808981": "homonymy",
  "n13809207": "part, portion, component part, component, constituent",
  "n13809769": "basis, base",
  "n13809920": "detail, particular, item",
  "n13810141": "highlight, high spot",
  "n13810323": "unit",
  "n13810615": "member",
  "n13810818": "remainder, balance, residual, residue, residuum, rest",
  "n13811184": "leftover, remnant",
  "n13811340": "subpart",
  "n13811410": "affinity, kinship",
  "n13811740": "rapport, resonance",
  "n13811900": "sympathy",
  "n13812173": "mutual understanding, mutual affection",
  "n13812296": "affinity, phylogenetic relation",
  "n13812607": "kinship, family relationship, relationship",
  "n13813042": "descent, line of descent, lineage, filiation",
  "n13813283": "affinity",
  "n13813483": "steprelationship",
  "n13813591": "consanguinity, blood kinship, cognation",
  "n13813765": "parentage, birth",
  "n13813898": "fatherhood, paternity",
  "n13814041": "motherhood, maternity",
  "n13814184": "sisterhood, sistership",
  "n13814336": "brotherhood",
  "n13814456": "bilateral descent",
  "n13814601": "unilateral descent",
  "n13814755": "matrilineage, enation, cognation",
  "n13814898": "patrilineage, agnation",
  "n13815030": "marital relationship, marital bed",
  "n13815152": "magnitude relation, quantitative relation",
  "n13815342": "scale",
  "n13815449": "proportion, ratio",
  "n13815742": "proportion",
  "n13816071": "case-fatality proportion",
  "n13816313": "case-to-infection proportion, case-to-infection ratio",
  "n13816521": "content",
  "n13816649": "rate",
  "n13816862": "scale",
  "n13817066": "golden section, golden mean",
  "n13817279": "commensurateness, correspondence, proportionateness",
  "n13817526": "percentage, percent, per centum, pct",
  "n13817761": "absentee rate",
  "n13817872": "batting average, hitting average",
  "n13818143": "batting average",
  "n13818354": "fielding average",
  "n13818551": "occupancy rate",
  "n13818736": "hospital occupancy",
  "n13818829": "hotel occupancy",
  "n13818916": "vacancy rate",
  "n13819074": "unemployment rate",
  "n13819207": "ratio",
  "n13820000": "abundance",
  "n13820239": "abundance",
  "n13820432": "albedo, reflective power",
  "n13820544": "aspect ratio",
  "n13820655": "average",
  "n13820826": "cephalic index, breadth index, cranial index",
  "n13820993": "efficiency",
  "n13821118": "figure of merit",
  "n13821272": "facial index",
  "n13821408": "focal ratio, f number, stop number, speed",
  "n13821570": "frequency, relative frequency",
  "n13821766": "hematocrit, haematocrit, packed cell volume",
  "n13821977": "intelligence quotient, IQ, I.Q.",
  "n13822249": "adult intelligence",
  "n13822364": "borderline intelligence",
  "n13822569": "load factor",
  "n13822735": "loss ratio",
  "n13822876": "Mach number",
  "n13822995": "magnification",
  "n13823150": "mechanical advantage",
  "n13823287": "mileage, fuel consumption rate, gasoline mileage, gas mileage",
  "n13823483": "odds, betting odds",
  "n13823651": "order of magnitude, magnitude",
  "n13823968": "output-to-input ratio",
  "n13824114": "prevalence",
  "n13824340": "price-to-earnings ratio, P/E ratio",
  "n13824500": "productivity",
  "n13824675": "proportionality",
  "n13824815": "quotient",
  "n13824929": "refractive index, index of refraction",
  "n13825080": "relative humidity",
  "n13825292": "respiratory quotient",
  "n13825490": "safety factor, factor of safety",
  "n13825664": "signal-to-noise ratio, signal-to-noise, signal/noise ratio, signal/noise, S/N",
  "n13825840": "stoichiometry",
  "n13826084": "time constant",
  "n13826221": "employee turnover, turnover rate, turnover",
  "n13826426": "loading",
  "n13826614": "power loading",
  "n13826732": "span loading",
  "n13826845": "wing loading",
  "n13826959": "incidence, relative incidence",
  "n13827103": "morbidity",
  "n13827205": "control",
  "n13827426": "direction",
  "n13827943": "frontage",
  "n13828075": "orientation",
  "n13828343": "attitude",
  "n13828520": "trim",
  "n13828681": "horizontal",
  "n13828795": "vertical",
  "n13828905": "opposition",
  "n13829047": "orthogonality, perpendicularity, orthogonal opposition",
  "n13829243": "antipodal, antipodal opposition, diametrical opposition",
  "n13829408": "enantiomorphism, mirror-image relation",
  "n13829586": "windward",
  "n13829720": "to windward, windward side, weatherboard, weather side",
  "n13829848": "leeward",
  "n13829980": "to leeward, leeward side",
  "n13830084": "seaward",
  "n13830166": "quarter",
  "n13830305": "compass point, point",
  "n13831000": "cardinal compass point",
  "n13831176": "north, due north, northward, N",
  "n13831310": "north by east, NbE",
  "n13831441": "north, magnetic north, compass north",
  "n13831572": "north northeast, nor'-nor'-east, NNE",
  "n13831717": "northeast by north, NEbN",
  "n13831843": "northeast, nor'-east, northeastward, NE",
  "n13831994": "northeast by east, NEbE",
  "n13832118": "east northeast, ENE",
  "n13832236": "east by north, EbN",
  "n13832355": "east, due east, eastward, E",
  "n13832480": "east by south, EbS",
  "n13832599": "east southeast, ESE",
  "n13832717": "southeast by east, SEbE",
  "n13832841": "southeast, sou'-east, southeastward, SE",
  "n13832993": "southeast by south, SEbS",
  "n13833119": "south southeast, sou'-sou'-east, SSE",
  "n13833256": "south by east, SbE",
  "n13833375": "south, due south, southward, S",
  "n13833504": "south by west, SbW",
  "n13833623": "south southwest, sou'-sou'-west, SSW",
  "n13833760": "southwest by south, SWbS",
  "n13833886": "southwest, sou'-west, southwestward, SW",
  "n13834038": "southwest by west, SWbW",
  "n13834162": "west southwest, WSW",
  "n13834280": "west by south, WbS",
  "n13834399": "west, due west, westward, W",
  "n13834524": "west by north, WbN",
  "n13834643": "west northwest, WNW",
  "n13834761": "northwest by west, NWbW",
  "n13834885": "northwest, nor'-west, northwestward, NW",
  "n13835037": "northwest by north, NWbN",
  "n13835163": "north northwest, nor'-nor'-west, NNW",
  "n13835308": "north by west, NbW",
  "n13835427": "north",
  "n13835546": "northeast",
  "n13835664": "east",
  "n13835781": "southeast",
  "n13835899": "south",
  "n13836018": "southwest",
  "n13836136": "west",
  "n13836253": "northwest",
  "n13836371": "angular position",
  "n13836550": "elevation, EL, altitude, ALT",
  "n13836715": "depression",
  "n13836841": "business relation",
  "n13837009": "competition",
  "n13837200": "price war, price competition",
  "n13837351": "clientage",
  "n13837439": "professional relation",
  "n13837667": "medical relation",
  "n13837840": "doctor-patient relation",
  "n13838025": "nurse-patient relation",
  "n13838205": "legal relation",
  "n13838386": "fiduciary relation",
  "n13838766": "bank-depositor relation",
  "n13838930": "confidential adviser-advisee relation",
  "n13839120": "conservator-ward relation",
  "n13839287": "director-stockholder relation",
  "n13839468": "executor-heir relation",
  "n13839662": "lawyer-client relation, attorney-client relation",
  "n13839888": "partner relation",
  "n13840039": "receiver-creditor relation",
  "n13840231": "trustee-beneficiary relation",
  "n13840404": "academic relation",
  "n13840553": "teacher-student relation",
  "n13840719": "politics, political relation",
  "n13840958": "chemistry, interpersonal chemistry, alchemy",
  "n13841213": "reciprocality, reciprocity",
  "n13841467": "complementarity",
  "n13841651": "correlation, correlativity",
  "n13841863": "mutuality, interdependence, interdependency",
  "n13842241": "commensalism",
  "n13842415": "parasitism",
  "n13842622": "symbiosis, mutualism",
  "n13842841": "trophobiosis",
  "n13843001": "additive inverse",
  "n13843173": "multiplicative inverse, reciprocal",
  "n13843401": "mutuality, mutualness",
  "n13843601": "reciprocal",
  "n13843786": "sharing",
  "n13843920": "sharing",
  "n13844057": "time sharing",
  "n13844212": "interrelation, interrelationship, interrelatedness",
  "n13844518": "psychodynamics",
  "n13844690": "temporal relation",
  "n13844923": "antecedent, forerunner",
  "n13845114": "chronology",
  "n13845239": "synchronism, synchrony, synchronicity, synchroneity, synchronization, synchronisation, synchronizing",
  "n13845838": "asynchronism, asynchrony, desynchronization, desynchronisation, desynchronizing",
  "n13846199": "first, number one",
  "n13846356": "former",
  "n13846546": "second",
  "n13846708": "latter",
  "n13846900": "third",
  "n13847124": "fourth",
  "n13847240": "fifth",
  "n13847402": "sixth",
  "n13847498": "seventh",
  "n13847616": "eighth",
  "n13847715": "ninth",
  "n13847858": "tenth",
  "n13847954": "eleventh",
  "n13848052": "twelfth",
  "n13848149": "thirteenth",
  "n13848249": "fourteenth",
  "n13848349": "fifteenth",
  "n13848466": "sixteenth",
  "n13848565": "seventeenth",
  "n13848684": "eighteenth",
  "n13848784": "nineteenth",
  "n13848884": "twentieth",
  "n13848983": "thirtieth",
  "n13849082": "fortieth",
  "n13849180": "fiftieth",
  "n13849296": "sixtieth",
  "n13849394": "seventieth",
  "n13849512": "eightieth",
  "n13849611": "ninetieth",
  "n13849710": "hundredth",
  "n13849810": "thousandth",
  "n13849913": "millionth",
  "n13850019": "billionth",
  "n13850148": "last",
  "n13850304": "scale, scale of measurement, graduated table, ordered series",
  "n13850674": "Beaufort scale, wind scale",
  "n13851067": "index",
  "n13851245": "logarithmic scale",
  "n13851426": "Mercalli scale",
  "n13851677": "Mohs scale",
  "n13851865": "Richter scale",
  "n13852083": "moment magnitude scale",
  "n13852395": "temperature scale",
  "n13852600": "Celsius scale, international scale, centigrade scale",
  "n13852820": "Fahrenheit scale",
  "n13853002": "Kelvin scale, absolute scale",
  "n13853198": "Rankine scale",
  "n13853407": "Reaumur scale",
  "n13853546": "wage scale, wage schedule",
  "n13853680": "sliding scale",
  "n13853808": "comparison",
  "n13853947": "imaginative comparison",
  "n13854101": "gauge, standard of measurement",
  "n13854318": "baseline",
  "n13854488": "norm",
  "n13854649": "opposition, oppositeness",
  "n13855100": "antipode",
  "n13855230": "antithesis",
  "n13855377": "conflict",
  "n13855627": "contrast, direct contrast",
  "n13855828": "flip side",
  "n13856110": "mutual opposition, polarity",
  "n13856320": "gradable opposition",
  "n13856449": "polar opposition",
  "n13856574": "polarity, sign",
  "n13856857": "positivity, positiveness",
  "n13857008": "negativity, negativeness",
  "n13857159": "ungradable opposition",
  "n13857314": "complementarity",
  "n13857486": "contradictoriness",
  "n13857676": "contradiction",
  "n13857804": "dialectic",
  "n13858045": "incompatibility, mutual exclusiveness, inconsistency, repugnance",
  "n13858270": "contrary",
  "n13858392": "contrariety",
  "n13858481": "tertium quid",
  "n13858604": "reverse, contrary, opposite",
  "n13858833": "inverse, opposite",
  "n13859043": "change",
  "n13859307": "difference",
  "n13859512": "gradient",
  "n13859692": "concentration gradient",
  "n13859925": "gravity gradient",
  "n13860145": "temperature gradient",
  "n13860281": "implication, logical implication, conditional relation",
  "n13860548": "antagonism",
  "n13860793": "solid",
  "n13861050": "plane, sheet",
  "n13861449": "Cartesian plane",
  "n13861580": "facet plane",
  "n13861686": "midplane, midline",
  "n13861806": "orbital plane",
  "n13861934": "picture plane",
  "n13862144": "tangent plane",
  "n13862282": "natural shape",
  "n13862407": "leaf shape, leaf form",
  "n13862552": "equilateral",
  "n13862644": "flare, flair",
  "n13862780": "figure",
  "n13863020": "pencil",
  "n13863186": "plane figure, two-dimensional figure",
  "n13863473": "solid figure, three-dimensional figure",
  "n13863602": "subfigure",
  "n13863771": "line",
  "n13864035": "bulb",
  "n13864153": "convex shape, convexity",
  "n13864423": "camber",
  "n13864542": "entasis",
  "n13864763": "angular shape, angularity",
  "n13864965": "concave shape, concavity, incurvation, incurvature",
  "n13865298": "cylinder",
  "n13865483": "round shape",
  "n13865786": "conglomeration, conglobation",
  "n13865904": "heart",
  "n13866144": "polygon, polygonal shape",
  "n13866515": "isogon",
  "n13866626": "convex polygon",
  "n13866827": "concave polygon",
  "n13867005": "reentrant polygon, reentering polygon",
  "n13867133": "regular polygon",
  "n13867276": "distorted shape, distortion",
  "n13867492": "amorphous shape",
  "n13867641": "curve, curved shape",
  "n13868248": "closed curve",
  "n13868371": "simple closed curve, Jordan curve",
  "n13868515": "S-shape",
  "n13868607": "catenary",
  "n13868813": "Cupid's bow",
  "n13868944": "wave, undulation",
  "n13869045": "extrados",
  "n13869129": "gooseneck",
  "n13869243": "intrados",
  "n13869327": "bend, crook, twist, turn",
  "n13869547": "hook, crotchet",
  "n13869676": "uncus",
  "n13869788": "envelope",
  "n13869896": "bight",
  "n13869991": "straight line",
  "n13870414": "geodesic, geodesic line",
  "n13870680": "perpendicular",
  "n13870805": "connection, connexion, link",
  "n13870935": "asymptote",
  "n13871134": "tangent",
  "n13871338": "secant",
  "n13871452": "perimeter",
  "n13871538": "radius",
  "n13871717": "diameter",
  "n13871967": "centerline, center line",
  "n13872072": "dome",
  "n13872211": "pit, fossa",
  "n13872421": "recess, recession, niche, corner",
  "n13872592": "cone, conoid, cone shape",
  "n13872822": "funnel, funnel shape",
  "n13872975": "conic section, conic",
  "n13873213": "intersection",
  "n13873361": "oblong",
  "n13873502": "circle",
  "n13873849": "circlet",
  "n13873917": "circle",
  "n13874073": "equator",
  "n13874251": "semicircle, hemicycle",
  "n13874384": "arc",
  "n13874558": "scallop, crenation, crenature, crenel, crenelle",
  "n13874927": "chord",
  "n13875027": "sector",
  "n13875185": "disk, disc, saucer",
  "n13875392": "ring, halo, annulus, doughnut, anchor ring",
  "n13875571": "loop",
  "n13875884": "bight",
  "n13875970": "coil, whorl, roll, curl, curlicue, ringlet, gyre, scroll",
  "n13876371": "spiral",
  "n13876561": "helix, spiral",
  "n13876753": "double helix",
  "n13876921": "perversion",
  "n13877129": "eccentricity",
  "n13877408": "element",
  "n13877547": "element of a cone",
  "n13877667": "element of a cylinder",
  "n13877785": "helix angle",
  "n13877918": "kink, twist, twirl",
  "n13878112": "whirl, swirl, vortex, convolution",
  "n13878306": "ellipse, oval",
  "n13878634": "square, foursquare",
  "n13878951": "square",
  "n13879049": "quadrate",
  "n13879126": "quadrilateral, quadrangle, tetragon",
  "n13879320": "triangle, trigon, trilateral",
  "n13879634": "triangle",
  "n13879816": "acute triangle, acute-angled triangle",
  "n13879947": "equilateral triangle, equiangular triangle",
  "n13880102": "delta",
  "n13880199": "isosceles triangle",
  "n13880295": "oblique triangle",
  "n13880415": "obtuse triangle, obtuse-angled triangle",
  "n13880551": "right triangle, right-angled triangle",
  "n13880704": "scalene triangle",
  "n13880811": "hexagram",
  "n13880994": "parallel",
  "n13881175": "parallelogram",
  "n13881381": "trapezium",
  "n13881512": "trapezoid",
  "n13881644": "star",
  "n13881810": "asterism",
  "n13882008": "pentacle, pentagram, pentangle",
  "n13882201": "pentagon",
  "n13882276": "hexagon",
  "n13882385": "regular hexagon",
  "n13882487": "heptagon",
  "n13882563": "octagon",
  "n13882639": "nonagon",
  "n13882713": "decagon",
  "n13882804": "undecagon",
  "n13882883": "dodecagon",
  "n13882961": "rhombus, rhomb, diamond",
  "n13883133": "rhomboid",
  "n13883346": "rectangle",
  "n13883494": "box",
  "n13883603": "spherical polygon",
  "n13883763": "spherical triangle",
  "n13883885": "polyhedron",
  "n13884261": "convex polyhedron",
  "n13884384": "concave polyhedron",
  "n13884511": "prism",
  "n13884740": "parallelepiped, parallelopiped, parallelepipedon, parallelopipedon",
  "n13884930": "cuboid",
  "n13885011": "quadrangular prism",
  "n13885111": "triangular prism",
  "n13885207": "sinuosity, sinuousness",
  "n13885370": "tortuosity, tortuousness, torsion, contortion, crookedness",
  "n13885700": "warp, buckle",
  "n13885836": "knot, gnarl",
  "n13886133": "arch",
  "n13886260": "bell, bell shape, campana",
  "n13886371": "parabola",
  "n13886575": "hyperbola",
  "n13886724": "furcation, forking",
  "n13886910": "bifurcation",
  "n13887056": "bifurcation",
  "n13887211": "jog",
  "n13887319": "zigzag, zig, zag",
  "n13887509": "angle",
  "n13888387": "complementary angles",
  "n13888491": "angular distance",
  "n13888783": "hour angle, HA",
  "n13889066": "true anomaly",
  "n13889216": "plane angle",
  "n13889331": "spherical angle",
  "n13889461": "solid angle",
  "n13889602": "inclination, angle of inclination",
  "n13889843": "inclination, inclination of an orbit",
  "n13890064": "reentrant angle, reentering angle",
  "n13890226": "salient angle",
  "n13890391": "interior angle, internal angle",
  "n13890535": "exterior angle, external angle",
  "n13890679": "hip",
  "n13890869": "angle of incidence, incidence angle",
  "n13891082": "angle of attack",
  "n13891242": "critical angle",
  "n13891371": "angle of reflection",
  "n13891547": "angle of refraction",
  "n13891735": "angle of extinction, extinction angle",
  "n13891937": "acute angle",
  "n13892048": "obtuse angle",
  "n13892142": "dogleg",
  "n13892237": "right angle",
  "n13892383": "oblique angle",
  "n13892562": "reflex angle",
  "n13892674": "perigon, round angle",
  "n13892765": "cutting angle",
  "n13892897": "dip, angle of dip, magnetic dip, magnetic inclination, inclination",
  "n13893129": "lead",
  "n13893310": "magnetic declination, magnetic variation, variation",
  "n13893502": "azimuth, AZ",
  "n13893694": "bowl, trough",
  "n13893786": "groove, channel",
  "n13894154": "rut",
  "n13894306": "scoop, pocket",
  "n13894434": "bulge, bump, hump, swelling, gibbosity, gibbousness, jut, prominence, protuberance, protrusion, extrusion, excrescence",
  "n13895262": "belly",
  "n13895362": "caput",
  "n13895549": "mogul",
  "n13895622": "nub, nubble",
  "n13895745": "snag",
  "n13895852": "wart",
  "n13895986": "node, knob, thickening",
  "n13896100": "bow, arc",
  "n13896217": "crescent",
  "n13896369": "depression, impression, imprint",
  "n13896695": "dimple",
  "n13896854": "hyperboloid",
  "n13897002": "paraboloid",
  "n13897198": "ellipsoid",
  "n13897377": "flank",
  "n13897528": "hypotenuse",
  "n13897657": "altitude",
  "n13897837": "base",
  "n13897996": "balance, equilibrium, equipoise, counterbalance",
  "n13898207": "conformation",
  "n13898315": "symmetry, proportion",
  "n13898509": "disproportion",
  "n13898645": "spheroid, ellipsoid of revolution",
  "n13898899": "sphere",
  "n13899109": "hemisphere",
  "n13899200": "sphere",
  "n13899404": "ball, globe, orb",
  "n13899735": "spherule",
  "n13899804": "cylinder",
  "n13900088": "torus, toroid",
  "n13900287": "toroid",
  "n13900422": "column, tower, pillar",
  "n13900760": "plume",
  "n13900914": "columella",
  "n13901055": "hoodoo",
  "n13901211": "barrel, drum",
  "n13901321": "pipe, tube",
  "n13901423": "pellet",
  "n13901490": "bolus",
  "n13901585": "drop, bead, pearl",
  "n13901858": "dewdrop",
  "n13901925": "teardrop",
  "n13902048": "ridge",
  "n13902229": "corrugation",
  "n13902336": "rim",
  "n13902482": "point, tip, peak",
  "n13902793": "taper",
  "n13902905": "quadric, quadric surface",
  "n13903079": "boundary, edge, bound",
  "n13903387": "margin, border, perimeter",
  "n13903576": "periphery, fringe, outer boundary",
  "n13903738": "brink, threshold, verge",
  "n13903855": "upper bound",
  "n13904011": "lower bound",
  "n13904164": "diagonal, bias",
  "n13904325": "diagonal",
  "n13904506": "dip",
  "n13904665": "cup",
  "n13904843": "incision, scratch, prick, slit, dent",
  "n13905121": "incisure, incisura",
  "n13905275": "notch",
  "n13905405": "score, scotch",
  "n13905572": "sag, droop",
  "n13905792": "wrinkle, furrow, crease, crinkle, seam, line",
  "n13906345": "crow's foot, crow's feet, laugh line",
  "n13906484": "dermatoglyphic",
  "n13906669": "frown line",
  "n13906767": "line of life, life line, lifeline",
  "n13906936": "line of heart, heart line, love line, mensal line",
  "n13907104": "line of fate, line of destiny, line of Saturn",
  "n13907272": "crevice, cranny, crack, fissure, chap",
  "n13907415": "fold, crease, plication, flexure, crimp, bend",
  "n13907847": "pucker, ruck",
  "n13908021": "indentation, indenture",
  "n13908201": "cleft",
  "n13908348": "stria, striation",
  "n13908580": "roulette, line roulette",
  "n13908777": "cycloid",
  "n13908954": "curate cycloid",
  "n13909071": "prolate cycloid",
  "n13909190": "sine curve, sinusoid",
  "n13909296": "epicycle",
  "n13909488": "epicycloid",
  "n13909632": "cardioid",
  "n13909752": "hypocycloid",
  "n13909889": "shapelessness",
  "n13910019": "blob",
  "n13910116": "void, vacancy, emptiness, vacuum",
  "n13910384": "space",
  "n13910895": "hollow",
  "n13911045": "node",
  "n13911151": "articulation, join, joint, juncture, junction",
  "n13911429": "hole",
  "n13911517": "cavity, enclosed space",
  "n13911660": "pocket",
  "n13911872": "point, dot",
  "n13912115": "pore",
  "n13912260": "tree, tree diagram",
  "n13912424": "cladogram",
  "n13912540": "stemma",
  "n13912686": "thalweg",
  "n13912839": "spur, spine, acantha",
  "n13912992": "constriction, bottleneck, chokepoint",
  "n13913164": "facet",
  "n13913263": "vector",
  "n13913427": "ray",
  "n13913566": "cast, mold, mould, stamp",
  "n13913849": "branch, leg, ramification",
  "n13914141": "brachium",
  "n13914265": "fork, crotch",
  "n13914473": "pouch, sac, sack, pocket",
  "n13914608": "block, cube",
  "n13914837": "pyramid",
  "n13915023": "ovoid",
  "n13915113": "tetrahedron",
  "n13915209": "pentahedron",
  "n13915305": "hexahedron",
  "n13915417": "rhombohedron",
  "n13915594": "octahedron",
  "n13915690": "decahedron",
  "n13915784": "dodecahedron",
  "n13915883": "icosahedron",
  "n13915999": "regular polyhedron, regular convex solid, regular convex polyhedron, Platonic body, Platonic solid, ideal solid",
  "n13916363": "polyhedral angle",
  "n13916495": "face angle",
  "n13916603": "regular tetrahedron",
  "n13916721": "cube, regular hexahedron",
  "n13916927": "tesseract",
  "n13917022": "quadrate",
  "n13917094": "regular dodecahedron",
  "n13917214": "regular octahedron",
  "n13917334": "regular icosahedron",
  "n13917457": "frustum",
  "n13917690": "truncated pyramid",
  "n13917785": "truncated cone",
  "n13917874": "prismatoid",
  "n13918095": "prismoid",
  "n13918274": "tail, tail end",
  "n13918387": "tongue, knife",
  "n13918584": "tilt angle",
  "n13918717": "trapezohedron",
  "n13918816": "vertical angle",
  "n13918965": "verticil",
  "n13919059": "view angle, angle of view",
  "n13919173": "washout",
  "n13919394": "wave angle",
  "n13919547": "wedge, wedge shape, cuneus",
  "n13919685": "projection",
  "n13919919": "keel",
  "n13920012": "cleavage",
  "n13920169": "medium",
  "n13920322": "ornamentation",
  "n13920429": "condition",
  "n13920835": "condition, status",
  "n13923030": "conditions",
  "n13923219": "conditions",
  "n13923440": "condition",
  "n13923582": "anchorage",
  "n13923779": "health",
  "n13923929": "mode",
  "n13924069": "conditionality",
  "n13924196": "ground state",
  "n13924336": "niche, ecological niche",
  "n13924536": "noise conditions",
  "n13924659": "participation, involvement",
  "n13924863": "prepossession",
  "n13925015": "regularization, regularisation",
  "n13925188": "saturation",
  "n13925340": "saturation point",
  "n13925550": "silence",
  "n13925752": "situation, position",
  "n13926068": "ski conditions",
  "n13926168": "niche",
  "n13926329": "election",
  "n13926451": "nationhood",
  "n13926535": "nomination",
  "n13926786": "place, shoes",
  "n13926932": "poverty trap",
  "n13927112": "soup",
  "n13927231": "stymie, stymy",
  "n13927383": "situation, state of affairs",
  "n13928191": "absurd, the absurd",
  "n13928388": "relationship",
  "n13928668": "relationship",
  "n13928943": "tribalism",
  "n13929037": "account, business relationship",
  "n13929336": "short account",
  "n13929477": "blood brotherhood",
  "n13929588": "company, companionship, fellowship, society",
  "n13929852": "confidence, trust",
  "n13930073": "freemasonry",
  "n13930245": "acquaintance, acquaintanceship",
  "n13930385": "affiliation, association, tie, tie-up",
  "n13930725": "anaclisis",
  "n13930928": "assimilation",
  "n13931145": "friendship, friendly relationship",
  "n13931356": "intrigue",
  "n13931436": "love affair, romance",
  "n13931627": "membership",
  "n13931765": "sexual relationship",
  "n13931889": "affair, affaire, intimacy, liaison, involvement, amour",
  "n13932045": "utopia",
  "n13932213": "dystopia",
  "n13932421": "acceptance",
  "n13932739": "ballgame, new ballgame",
  "n13932948": "challenge",
  "n13933103": "childlessness",
  "n13933221": "complication",
  "n13933391": "conflict of interest",
  "n13933560": "crisis",
  "n13933841": "crowding",
  "n13934070": "crunch",
  "n13934274": "disequilibrium",
  "n13934465": "element",
  "n13934596": "environment",
  "n13934900": "equilibrium",
  "n13935227": "exclusion",
  "n13935400": "goldfish bowl, fish bowl, fishbowl",
  "n13935572": "hornet's nest, hornets' nest",
  "n13935753": "hotbed",
  "n13935910": "hot potato",
  "n13936030": "how-do-you-do, how-d'ye-do",
  "n13936153": "imbroglio, embroilment",
  "n13936304": "inclusion",
  "n13936441": "intestacy",
  "n13936557": "Mexican standoff",
  "n13936676": "nightmare, incubus",
  "n13936783": "no-win situation",
  "n13936939": "pass, strait, straits",
  "n13937075": "picture, scene",
  "n13937284": "prison, prison house",
  "n13937406": "purgatory",
  "n13937554": "rejection",
  "n13937727": "size, size of it",
  "n13937918": "square one",
  "n13938210": "status quo",
  "n13938296": "swamp",
  "n13938440": "standardization, standardisation",
  "n13938663": "stigmatism",
  "n13938899": "astigmatism, astigmia",
  "n13939158": "stratification, social stratification",
  "n13939353": "wild, natural state, state of nature",
  "n13939604": "way",
  "n13939734": "isomerism",
  "n13939892": "degree, level, stage, point",
  "n13940311": "ladder",
  "n13940456": "acme, height, elevation, peak, pinnacle, summit, superlative, meridian, tiptop, top",
  "n13941125": "extent",
  "n13941337": "resultant, end point",
  "n13941469": "standard of living, standard of life",
  "n13941806": "plane",
  "n13941924": "state of the art",
  "n13942104": "ultimacy, ultimateness",
  "n13942405": "extremity",
  "n13942554": "profoundness",
  "n13942743": "ordinary",
  "n13942875": "circumstance",
  "n13943053": "homelessness",
  "n13943236": "vagrancy",
  "n13943400": "event, case",
  "n13943598": "hinge",
  "n13943740": "playing field",
  "n13943968": "thing",
  "n13944096": "time bomb, ticking bomb",
  "n13944337": "tinderbox",
  "n13944516": "urgency",
  "n13944747": "congestion, over-crowding",
  "n13944914": "reinstatement",
  "n13945102": "office, power",
  "n13945495": "executive clemency",
  "n13945682": "war power",
  "n13945919": "status, position",
  "n13946760": "equality, equivalence, equation, par",
  "n13947010": "egality, egalite",
  "n13947174": "tie",
  "n13947272": "deuce",
  "n13947415": "social station, social status, social rank, rank",
  "n13947645": "place, station",
  "n13947867": "place",
  "n13948026": "quality",
  "n13948136": "standing",
  "n13948441": "high status",
  "n13948654": "center stage, centre stage",
  "n13948766": "stardom",
  "n13948912": "championship, title",
  "n13949080": "triple crown",
  "n13949275": "triple crown",
  "n13949463": "high ground",
  "n13949576": "seniority, senior status, higher status, higher rank",
  "n13949802": "precedence, precedency, priority",
  "n13950143": "back burner",
  "n13950282": "front burner",
  "n13950440": "transcendence, transcendency, superiority",
  "n13950703": "high profile",
  "n13950812": "Holy Order, Order",
  "n13951215": "low status, lowness, lowliness",
  "n13951444": "inferiority, lower status, lower rank",
  "n13951629": "backseat",
  "n13951794": "shade",
  "n13951984": "subordinateness, subsidiarity",
  "n13952171": "handmaid, handmaiden, servant",
  "n13952386": "junior status",
  "n13952466": "subservience, subservientness",
  "n13952601": "legal status",
  "n13952792": "civil death",
  "n13953057": "villeinage, villainage",
  "n13953215": "bastardy, illegitimacy, bar sinister",
  "n13953356": "left-handedness",
  "n13953467": "citizenship",
  "n13953608": "command",
  "n13953787": "nationality",
  "n13953936": "footing, terms",
  "n13954118": "retirement",
  "n13954253": "being, beingness, existence",
  "n13954818": "actuality",
  "n13955152": "entelechy",
  "n13955341": "genuineness",
  "n13955461": "reality, realness, realism",
  "n13955874": "fact",
  "n13956097": "reality",
  "n13956362": "historicalness",
  "n13956488": "truth, the true, verity, trueness",
  "n13956905": "eternity, timelessness, timeless existence",
  "n13957117": "preexistence",
  "n13957268": "coexistence",
  "n13957390": "eternal life, life eternal",
  "n13957493": "subsistence",
  "n13957601": "presence",
  "n13957974": "immanence, immanency",
  "n13958138": "inherence, inherency",
  "n13958353": "ubiety",
  "n13958456": "ubiquity, ubiquitousness, omnipresence",
  "n13958672": "hereness",
  "n13958800": "thereness",
  "n13958921": "thereness",
  "n13959069": "occurrence",
  "n13959289": "allopatry",
  "n13959463": "sympatry",
  "n13959642": "shadow",
  "n13959818": "nonbeing",
  "n13959931": "nonexistence, nonentity",
  "n13960117": "unreality, irreality",
  "n13960357": "cloud",
  "n13960464": "falsity, falseness",
  "n13960833": "spuriousness",
  "n13960974": "absence",
  "n13961179": "nonoccurrence",
  "n13961273": "awayness",
  "n13961399": "life",
  "n13961642": "animation, life, living, aliveness",
  "n13962048": "skin",
  "n13962166": "survival, endurance",
  "n13962360": "subsistence",
  "n13962498": "death",
  "n13962765": "rest, eternal rest, sleep, eternal sleep, quietus",
  "n13963032": "extinction, defunctness",
  "n13963192": "life",
  "n13963342": "ghetto",
  "n13963565": "transcendence, transcendency",
  "n13963757": "marital status",
  "n13963970": "marriage, matrimony, union, spousal relationship, wedlock",
  "n13964466": "bigamy",
  "n13964591": "civil union",
  "n13964879": "common-law marriage",
  "n13965049": "endogamy, intermarriage, inmarriage",
  "n13965274": "exogamy, intermarriage",
  "n13965515": "marriage of convenience",
  "n13965627": "misalliance",
  "n13965780": "mesalliance",
  "n13965888": "monandry",
  "n13966007": "monogamy, monogamousness",
  "n13966201": "monogyny",
  "n13966340": "serial monogamy",
  "n13966496": "open marriage",
  "n13966683": "cuckoldom",
  "n13966795": "polyandry",
  "n13966925": "polygamy",
  "n13967089": "polygyny",
  "n13967215": "sigeh",
  "n13967507": "celibacy",
  "n13967581": "virginity",
  "n13967715": "bachelorhood",
  "n13967827": "spinsterhood",
  "n13967970": "widowhood",
  "n13968092": "employment, employ",
  "n13968308": "unemployment",
  "n13968547": "order",
  "n13968859": "civil order, polity",
  "n13968973": "rule of law",
  "n13969101": "tranquillity, tranquility, quiet",
  "n13969243": "harmony, concord, concordance",
  "n13969700": "peace",
  "n13969854": "comity",
  "n13969988": "comity of nations",
  "n13970124": "stability",
  "n13970236": "peace",
  "n13970460": "amity",
  "n13970584": "peaceableness, peacefulness",
  "n13970764": "mollification",
  "n13970912": "armistice, cease-fire, truce",
  "n13971065": "agreement, accord",
  "n13971382": "community, community of interests",
  "n13971561": "conciliation",
  "n13971802": "concurrence, meeting of minds",
  "n13971901": "consensus",
  "n13972180": "sense of the meeting",
  "n13972387": "unanimity",
  "n13972505": "unison",
  "n13972601": "social contract",
  "n13972797": "disorder",
  "n13973059": "anarchy, lawlessness",
  "n13973320": "nihilism",
  "n13973490": "cytopenia",
  "n13973632": "hematocytopenia, haematocytopenia",
  "n13973769": "pancytopenia",
  "n13973990": "immunological disorder",
  "n13974144": "immunocompetence",
  "n13974317": "immunodeficiency",
  "n13974620": "immunosuppression",
  "n13974937": "bloodiness",
  "n13975037": "incompatibility",
  "n13975292": "histoincompatibility",
  "n13975444": "Rh incompatibility",
  "n13975659": "instability",
  "n13975752": "confusion",
  "n13975988": "demoralization, demoralisation",
  "n13976182": "bluster",
  "n13976322": "chaos, pandemonium, bedlam, topsy-turvydom, topsy-turvyness",
  "n13976527": "balagan",
  "n13976731": "hugger-mugger",
  "n13976907": "schemozzle, shemozzle",
  "n13977043": "rioting, riot",
  "n13977184": "rowdiness, rowdyism, roughness, disorderliness",
  "n13977366": "disturbance, disruption, commotion, flutter, hurly burly, to-do, hoo-ha, hoo-hah, kerfuffle",
  "n13977732": "convulsion, turmoil, upheaval",
  "n13977870": "earthquake",
  "n13978033": "incident",
  "n13978166": "stir, splash",
  "n13978344": "storm, tempest",
  "n13978601": "storm center, storm centre",
  "n13978709": "tumult, tumultuousness, uproar, garboil",
  "n13978914": "combustion",
  "n13979064": "discord, strife",
  "n13979173": "turbulence, upheaval, Sturm und Drang",
  "n13979503": "agitation, ferment, fermentation, tempestuousness, unrest",
  "n13979786": "roller coaster",
  "n13979977": "violence",
  "n13980128": "rage",
  "n13980288": "hostility, enmity, antagonism",
  "n13980596": "latent hostility, tension",
  "n13980845": "conflict",
  "n13981137": "clash, friction",
  "n13981271": "clash",
  "n13981403": "war, state of war",
  "n13981763": "proxy war",
  "n13981885": "hot war",
  "n13982000": "cold war",
  "n13982156": "Cold War",
  "n13982357": "disagreement, dissension, dissonance",
  "n13982588": "disunity",
  "n13982692": "divide",
  "n13982839": "suspicion",
  "n13982999": "cloud",
  "n13983147": "illumination",
  "n13983304": "light, lighting",
  "n13983515": "dark, darkness",
  "n13983717": "night",
  "n13983807": "total darkness, lightlessness, blackness, pitch blackness, black",
  "n13984082": "blackout, brownout, dimout",
  "n13984285": "semidarkness",
  "n13984468": "cloudiness, overcast",
  "n13984613": "shade, shadiness, shadowiness",
  "n13984944": "shadow",
  "n13985098": "umbra",
  "n13985219": "penumbra",
  "n13985323": "dimness, duskiness",
  "n13985462": "gloom, somberness, sombreness",
  "n13985671": "obscurity, obscureness",
  "n13985818": "emotional state, spirit",
  "n13986189": "embarrassment",
  "n13986372": "ecstasy, rapture, transport, exaltation, raptus",
  "n13986679": "gratification, satisfaction",
  "n13986943": "quality of life",
  "n13987219": "comfort",
  "n13987423": "happiness, felicity",
  "n13987719": "blessedness, beatitude, beatification",
  "n13987905": "bliss, blissfulness, cloud nine, seventh heaven, walking on air",
  "n13988101": "ecstasy, rapture",
  "n13988224": "nirvana, enlightenment",
  "n13988498": "state",
  "n13988663": "unhappiness",
  "n13988871": "embitterment",
  "n13989051": "sadness, sorrow, sorrowfulness",
  "n13989280": "mourning, bereavement",
  "n13989443": "poignance, poignancy",
  "n13989627": "innocence",
  "n13989863": "blamelessness, inculpability, inculpableness, guiltlessness",
  "n13990064": "purity, pureness, sinlessness, innocence, whiteness",
  "n13990389": "cleanness",
  "n13990502": "clear",
  "n13990675": "guilt, guiltiness",
  "n13990960": "blameworthiness, culpability, culpableness",
  "n13991121": "bloodguilt",
  "n13991245": "complicity",
  "n13991346": "criminalism, criminality, criminalness",
  "n13991516": "guilt by association",
  "n13991688": "impeachability, indictability",
  "n13991823": "freedom",
  "n13992194": "academic freedom",
  "n13992387": "enfranchisement",
  "n13992514": "autonomy, liberty",
  "n13992738": "self-government, self-determination, self-rule",
  "n13992935": "sovereignty",
  "n13993030": "local option",
  "n13993210": "home rule",
  "n13993356": "autarky, autarchy",
  "n13993517": "fragmentation",
  "n13993685": "free hand, blank check",
  "n13993842": "free rein, play",
  "n13994014": "freedom of the seas",
  "n13994148": "independence, independency",
  "n13994456": "liberty",
  "n13994806": "license, licence",
  "n13995016": "poetic license",
  "n13995148": "latitude",
  "n13995378": "license, licence",
  "n13995662": "civil liberty, political liberty",
  "n13995824": "discretion",
  "n13995935": "run",
  "n13996061": "liberty",
  "n13996211": "svoboda",
  "n13996300": "subjugation, subjection",
  "n13996571": "repression",
  "n13996719": "oppression",
  "n13996909": "yoke",
  "n13997050": "enslavement, captivity",
  "n13997253": "bondage, slavery, thrall, thralldom, thraldom",
  "n13997529": "bondage",
  "n13997778": "bonded labor",
  "n13998014": "servitude",
  "n13998186": "peonage",
  "n13998263": "serfdom, serfhood, vassalage",
  "n13998395": "encapsulation",
  "n13998576": "confinement",
  "n13998781": "constraint, restraint",
  "n13998966": "cage",
  "n13999114": "iron cage",
  "n13999206": "captivity, imprisonment, incarceration, immurement",
  "n13999663": "detention, detainment, hold, custody",
  "n13999941": "solitary confinement, solitary",
  "n14000105": "durance",
  "n14000200": "life imprisonment",
  "n14000302": "internment",
  "n14000403": "representation, delegacy, agency",
  "n14000642": "free agency",
  "n14000820": "legal representation",
  "n14001031": "autonomy, self-direction, self-reliance, self-sufficiency",
  "n14001213": "separateness",
  "n14001348": "dependence, dependance, dependency",
  "n14001629": "helplessness",
  "n14001728": "reliance",
  "n14001852": "subordination",
  "n14001973": "contingency",
  "n14002109": "polarization, polarisation",
  "n14002279": "balance",
  "n14002481": "tension",
  "n14002813": "balance of power",
  "n14002915": "dynamic balance",
  "n14003167": "homeostasis",
  "n14003432": "isostasy",
  "n14003599": "Nash equilibrium",
  "n14003880": "poise",
  "n14004017": "thermal equilibrium",
  "n14004149": "imbalance, instability, unbalance",
  "n14004317": "motion",
  "n14004572": "shaking, shakiness, trembling, quiver, quivering, vibration, palpitation",
  "n14004958": "tremolo",
  "n14005137": "tremor",
  "n14005302": "essential tremor",
  "n14005549": "perpetual motion",
  "n14005728": "precession",
  "n14005892": "stream, flow",
  "n14006179": "motionlessness, stillness, lifelessness",
  "n14006490": "stationariness, immobility, fixedness",
  "n14006684": "rootage",
  "n14006821": "dead letter, non-issue",
  "n14006945": "action, activity, activeness",
  "n14007546": "agency",
  "n14007750": "Frankenstein",
  "n14007864": "virus",
  "n14008050": "busyness, hum",
  "n14008342": "behavior, behaviour",
  "n14008567": "eruption, eructation, extravasation",
  "n14008806": "operation",
  "n14009020": "overdrive",
  "n14009274": "commission",
  "n14009481": "running",
  "n14009618": "idle",
  "n14009763": "play",
  "n14009946": "swing",
  "n14010148": "inaction, inactivity, inactiveness",
  "n14010636": "abeyance, suspension",
  "n14010845": "anergy",
  "n14010927": "arrest, check, halt, hitch, stay, stop, stoppage",
  "n14011408": "calcification",
  "n14011557": "deep freeze",
  "n14011724": "desuetude",
  "n14011811": "dormancy, quiescence, quiescency",
  "n14012173": "extinction",
  "n14012311": "holding pattern",
  "n14012501": "rest",
  "n14012667": "stagnation, stagnancy, doldrums",
  "n14013005": "stagnation, stagnancy",
  "n14013249": "stasis",
  "n14013368": "recession",
  "n14013549": "cold storage",
  "n14013646": "deferral, recess",
  "n14013751": "moratorium",
  "n14013841": "standdown, stand-down",
  "n14013981": "hibernation",
  "n14014162": "estivation, aestivation",
  "n14014425": "acathexia",
  "n14014521": "angiotelectasia",
  "n14014621": "torpor, torpidity",
  "n14014849": "hibernation",
  "n14014990": "lethargy, lassitude, sluggishness",
  "n14015148": "slumber",
  "n14015266": "countercheck",
  "n14015361": "deadlock, dead end, impasse, stalemate, standstill",
  "n14015596": "logjam",
  "n14015731": "temporary state",
  "n14015996": "case",
  "n14016114": "state of mind, frame of mind",
  "n14016275": "thinking cap",
  "n14016361": "fatigue, weariness, tiredness",
  "n14016863": "eyestrain, asthenopia",
  "n14017033": "jet lag",
  "n14017206": "exhaustion",
  "n14017332": "depletion",
  "n14017451": "salt depletion",
  "n14017665": "electrolyte balance",
  "n14017871": "nitrogen balance",
  "n14018055": "frazzle",
  "n14018203": "mental exhaustion, brain-fag",
  "n14018318": "grogginess",
  "n14018432": "loginess, logginess",
  "n14018567": "drunkenness, inebriation, inebriety, intoxication, tipsiness, insobriety",
  "n14018918": "grogginess",
  "n14019039": "sottishness",
  "n14019138": "soberness, sobriety",
  "n14019296": "acardia",
  "n14019441": "acephalia, acephaly, acephalism",
  "n14019600": "acidosis",
  "n14019840": "ketoacidosis, diabetic acidosis",
  "n14020023": "metabolic acidosis",
  "n14020271": "respiratory acidosis, carbon dioxide acidosis",
  "n14020558": "starvation acidosis",
  "n14020753": "acidemia",
  "n14020936": "alkalemia",
  "n14021118": "alkalinuria, alkaluria",
  "n14021266": "alkalosis",
  "n14021471": "metabolic alkalosis",
  "n14021622": "respiratory alkalosis",
  "n14021853": "acorea",
  "n14021936": "acromicria, acromikria",
  "n14022068": "acromphalus",
  "n14022199": "agalactia, agalactosis",
  "n14022369": "amastia",
  "n14022520": "ankylosis, anchylosis",
  "n14022661": "aneuploidy",
  "n14022849": "anorchism, anorchidism, anorchia",
  "n14022959": "wakefulness, sleeplessness",
  "n14023236": "hypersomnia",
  "n14023374": "insomnia",
  "n14023491": "anesthesia, anaesthesia",
  "n14023804": "anhidrosis, anhydrosis",
  "n14023901": "aplasia",
  "n14023997": "arousal",
  "n14024185": "arteriectasis, arteriectasia",
  "n14024296": "arthropathy",
  "n14024391": "asynergy, asynergia",
  "n14024581": "asystole, cardiac arrest, cardiopulmonary arrest",
  "n14024882": "sleep, slumber",
  "n14025258": "sleep terror disorder, pavor nocturnus",
  "n14025478": "orthodox sleep, nonrapid eye movement sleep, NREM sleep, nonrapid eye movement, NREM",
  "n14025755": "paradoxical sleep, rapid eye movement sleep, REM sleep, rapid eye movement, REM",
  "n14025993": "sleep, sopor",
  "n14026089": "shuteye",
  "n14026166": "abulia, aboulia",
  "n14026285": "anhedonia",
  "n14026376": "depersonalization, depersonalisation",
  "n14026592": "hypnosis",
  "n14026781": "self-hypnosis",
  "n14026869": "cryoanesthesia, cryoanaesthesia",
  "n14026981": "general anesthesia, general anaesthesia",
  "n14027204": "local anesthesia, local anaesthesia",
  "n14027396": "conduction anesthesia, conduction anaesthesia, nerve block anesthesia, nerve block anaesthesia, block anesthesia, block anaesthesia",
  "n14027674": "regional anesthesia, regional anaesthesia",
  "n14028087": "topical anesthesia, topical anaesthesia",
  "n14028291": "acroanesthesia, acroanaesthesia",
  "n14028406": "caudal anesthesia, caudal anaesthesia, caudal block",
  "n14028652": "epidural anesthesia, epidural anaesthesia, epidural",
  "n14028967": "paracervical block",
  "n14029163": "pudendal block",
  "n14029405": "spinal anesthesia, spinal anaesthesia, spinal",
  "n14029686": "saddle block anesthesia, saddle block anaesthesia",
  "n14029943": "inhalation anesthesia",
  "n14030084": "twilight sleep",
  "n14030291": "fugue",
  "n14030435": "sleepiness, drowsiness, somnolence",
  "n14030671": "oscitancy, oscitance",
  "n14030820": "imminence, imminency, imminentness, impendence, impendency, forthcomingness",
  "n14031108": "readiness, preparedness, preparation",
  "n14031523": "ready",
  "n14031660": "alert, qui vive",
  "n14031922": "air alert",
  "n14032100": "red alert",
  "n14032293": "strip alert",
  "n14032480": "diverticulosis",
  "n14032599": "emergency",
  "n14032737": "clutch",
  "n14032868": "Dunkirk",
  "n14033054": "exigency",
  "n14033185": "juncture, critical point, crossroads",
  "n14033481": "desperate straits, dire straits",
  "n14033587": "criticality",
  "n14033802": "flash point, flashpoint",
  "n14033917": "flux, state of flux",
  "n14034177": "physical condition, physiological state, physiological condition",
  "n14035298": "drive",
  "n14035502": "elastosis",
  "n14035695": "flatulence, flatulency, gas",
  "n14035909": "flexure, flection, flexion",
  "n14036043": "alertness, alerting",
  "n14036203": "emotional arousal",
  "n14036356": "excitation, innervation, irritation",
  "n14036539": "anger, angriness",
  "n14036735": "rage",
  "n14036892": "fever pitch",
  "n14037011": "excitement, excitation, inflammation, fervor, fervour",
  "n14037480": "sensation",
  "n14037619": "sexual arousal",
  "n14037816": "cybersex",
  "n14037925": "eroticism, erotism",
  "n14038027": "horniness, hotness, hot pants",
  "n14038185": "erection, hard-on",
  "n14038264": "estrus, oestrus, heat, rut",
  "n14038482": "anestrus, anestrum, anoestrus, anoestrum",
  "n14038743": "diestrus, diestrum",
  "n14038993": "desire",
  "n14039179": "rage, passion",
  "n14039310": "materialism, philistinism",
  "n14039534": "hunger, hungriness",
  "n14039828": "bulimia",
  "n14039974": "emptiness",
  "n14040071": "edacity, esurience, ravenousness, voracity, voraciousness",
  "n14040310": "starvation, famishment",
  "n14040490": "undernourishment, malnourishment",
  "n14040660": "thirst, thirstiness",
  "n14040846": "dehydration",
  "n14040966": "polydipsia",
  "n14041106": "sex drive",
  "n14041256": "hypoxia",
  "n14041508": "anemic hypoxia",
  "n14041631": "hypoxic hypoxia",
  "n14041763": "ischemic hypoxia, stagnant hypoxia",
  "n14041940": "hypercapnia, hypercarbia",
  "n14042165": "hypocapnia, acapnia",
  "n14042423": "asphyxia",
  "n14042735": "oxygen debt",
  "n14042914": "altitude sickness",
  "n14043092": "mountain sickness",
  "n14043243": "anoxia",
  "n14043458": "anemic anoxia",
  "n14043579": "anoxic anoxia",
  "n14043708": "ischemic anoxia, stagnant anoxia",
  "n14043882": "suffocation, asphyxiation",
  "n14044174": "hyperthermia, hyperthermy",
  "n14044372": "normothermia",
  "n14044454": "hypothermia",
  "n14044592": "flux",
  "n14044719": "muscularity",
  "n14044852": "myasthenia",
  "n14044930": "impotence, impotency",
  "n14045141": "erectile dysfunction, male erecticle dysfunction, ED",
  "n14045347": "barrenness",
  "n14045507": "sterility, infertility",
  "n14045835": "cacogenesis",
  "n14045954": "dysgenesis",
  "n14046038": "false pregnancy, pseudocyesis",
  "n14046202": "pregnancy, gestation, maternity",
  "n14046751": "trouble",
  "n14046861": "gravidity, gravidness, gravidation",
  "n14047009": "gravida",
  "n14047171": "parity, para",
  "n14047440": "abdominal pregnancy",
  "n14047547": "ovarian pregnancy",
  "n14047641": "tubal pregnancy",
  "n14047740": "ectopic pregnancy, extrauterine pregnancy, ectopic gestation, extrauterine gestation, eccyesis, metacyesis",
  "n14048015": "entopic pregnancy",
  "n14048134": "quickening",
  "n14048309": "premature labor, premature labour",
  "n14048441": "parturiency, labor, labour, confinement, lying-in, travail, childbed",
  "n14048847": "placenta previa",
  "n14049098": "asynclitism, obliquity",
  "n14049262": "atresia",
  "n14049420": "rigor mortis",
  "n14049552": "vitalization, vitalisation",
  "n14049711": "good health, healthiness",
  "n14050011": "wholeness, haleness",
  "n14050143": "energy, vim, vitality",
  "n14050434": "juice",
  "n14050559": "qi, chi, ch'i, ki",
  "n14050871": "bloom, blush, flush, rosiness",
  "n14051056": "freshness, glow",
  "n14051201": "radiance",
  "n14051381": "sturdiness",
  "n14051494": "fertility, fecundity",
  "n14051728": "potency, potence",
  "n14051917": "pathological state",
  "n14052046": "ill health, unhealthiness, health problem",
  "n14052403": "disorder, upset",
  "n14053384": "functional disorder",
  "n14053558": "organic disorder",
  "n14053717": "dyscrasia",
  "n14053850": "blood dyscrasia",
  "n14053965": "abocclusion",
  "n14054102": "abruptio placentae",
  "n14054262": "achlorhydria",
  "n14054465": "acholia, cholestasis",
  "n14054642": "achylia, achylia gastrica",
  "n14054762": "acute brain disorder, acute organic brain syndrome",
  "n14055052": "adult respiratory distress syndrome, ARDS, wet lung, white lung",
  "n14055408": "ailment, complaint, ill",
  "n14055623": "eating disorder",
  "n14055796": "anorexia",
  "n14055977": "pica",
  "n14056143": "astereognosis, tactile agnosia",
  "n14056280": "attention deficit disorder, ADD, attention deficit hyperactivity disorder, ADHD, hyperkinetic syndrome, minimal brain dysfunction, minimal brain damage, MBD",
  "n14056588": "anorgasmia",
  "n14056685": "bulimarexia, binge-purge syndrome, binge-vomit syndrome, bulima nervosa",
  "n14057001": "bulimia, binge-eating syndrome",
  "n14057221": "bladder disorder",
  "n14057371": "cardiovascular disease",
  "n14057569": "carpal tunnel syndrome",
  "n14057828": "celiac disease",
  "n14058066": "cheilosis, perleche",
  "n14058252": "choking",
  "n14058456": "colpoxerosis",
  "n14058563": "degenerative disorder",
  "n14058770": "demyelination",
  "n14058934": "dysaphia",
  "n14059021": "dysosmia, parosamia, olfactory impairment",
  "n14059143": "dysphagia",
  "n14059252": "dysuria",
  "n14059336": "dystrophy",
  "n14059481": "osteodystrophy",
  "n14059663": "failure",
  "n14059825": "fantods",
  "n14059928": "glandular disease, gland disease, glandular disorder, adenosis",
  "n14060256": "hyperactivity",
  "n14060417": "impaction, impacted tooth",
  "n14060575": "impaction",
  "n14060688": "learning disorder, learning disability",
  "n14060929": "malocclusion",
  "n14061097": "overbite",
  "n14061263": "anorexia nervosa",
  "n14061462": "cellularity",
  "n14061599": "hypercellularity",
  "n14061703": "hypocellularity",
  "n14061805": "illness, unwellness, malady, sickness",
  "n14062332": "invagination, introversion",
  "n14062490": "invalidism",
  "n14062565": "biliousness",
  "n14062725": "addiction, dependence, dependance, dependency, habituation",
  "n14063089": "suspended animation",
  "n14063290": "anabiosis",
  "n14063475": "cryptobiosis",
  "n14063633": "dilatation, distension, distention",
  "n14063877": "tympanites",
  "n14064044": "ectasia, ectasis",
  "n14064186": "lymphangiectasia, lymphangiectasis",
  "n14064296": "alveolar ectasia",
  "n14064408": "drug addiction, white plague",
  "n14064644": "alcoholism, alcohol addiction, inebriation, drunkenness",
  "n14064984": "cocaine addiction",
  "n14065071": "heroin addiction",
  "n14065156": "caffein addiction",
  "n14065243": "nicotine addiction",
  "n14065332": "ague",
  "n14065412": "roots",
  "n14065693": "amyloidosis",
  "n14065903": "anuresis, anuria",
  "n14066005": "catastrophic illness",
  "n14066203": "collapse, prostration",
  "n14066492": "breakdown, crack-up",
  "n14066661": "nervous breakdown",
  "n14066806": "nervous exhaustion, nervous prostration",
  "n14066959": "neurasthenia",
  "n14067076": "shock",
  "n14067481": "cardiogenic shock",
  "n14067575": "hypovolemic shock",
  "n14067681": "obstructive shock",
  "n14067786": "distributive shock",
  "n14067902": "anaphylactic shock",
  "n14068177": "insulin shock, insulin reaction",
  "n14068344": "decompression sickness, aeroembolism, air embolism, gas embolism, caisson disease, bends",
  "n14068528": "fluorosis",
  "n14068685": "food poisoning, gastrointestinal disorder",
  "n14068894": "botulism",
  "n14069057": "mushroom poisoning",
  "n14069212": "gammopathy",
  "n14069383": "glossolalia",
  "n14069541": "ptomaine, ptomaine poisoning",
  "n14069747": "salmonellosis",
  "n14069895": "lead poisoning, plumbism, saturnism",
  "n14070055": "lead colic, painter's colic",
  "n14070206": "catalepsy",
  "n14070360": "disease",
  "n14071235": "disease of the neuromuscular junction",
  "n14071419": "angiopathy",
  "n14071559": "aspergillosis",
  "n14071758": "acanthocytosis",
  "n14071896": "agranulocytosis, agranulosis, granulocytopenia",
  "n14072126": "analbuminemia",
  "n14072239": "Banti's disease, Banti's syndrome",
  "n14072423": "anthrax",
  "n14072625": "cutaneous anthrax, malignant pustule",
  "n14072934": "pulmonary anthrax, inhalation anthrax, anthrax pneumonia, ragpicker's disease, ragsorter's disease, woolsorter's pneumonia, woolsorter's disease",
  "n14073331": "blackwater",
  "n14073505": "Argentine hemorrhagic fever",
  "n14073647": "blackwater fever",
  "n14073795": "jungle fever",
  "n14073906": "cat scratch disease",
  "n14074041": "complication",
  "n14074267": "crud",
  "n14074394": "endemic, endemic disease",
  "n14074606": "enteropathy",
  "n14074715": "idiopathic disease, idiopathic disorder, idiopathy",
  "n14074877": "monogenic disorder, monogenic disease",
  "n14075199": "polygenic disorder, polygenic disease",
  "n14075358": "hypogonadism",
  "n14075660": "male hypogonadism, eunuchoidism",
  "n14075838": "Kallman's syndrome",
  "n14076011": "valvular incompetence",
  "n14076126": "incompetence",
  "n14076270": "Kawasaki disease, mucocutaneous lymph node syndrome",
  "n14076479": "plague, pestilence, pest",
  "n14076667": "pycnosis, pyknosis",
  "n14076809": "hyalinization, hyalinisation",
  "n14076992": "hyperparathyroidism",
  "n14077269": "hypoparathyroidism",
  "n14077454": "hyperpituitarism",
  "n14077693": "vacuolization, vacuolisation, vacuolation",
  "n14077830": "malaria",
  "n14078120": "Marseilles fever, Kenya fever, Indian tick fever, boutonneuse fever",
  "n14078421": "Meniere's disease",
  "n14078613": "milk sickness",
  "n14078733": "mimesis",
  "n14078871": "myasthenia gravis, myasthenia",
  "n14079156": "Lambert-Eaton syndrome, Eaton-Lambert syndrome, myasthenic syndrome, carcinomatous myopathy",
  "n14079493": "occupational disease, industrial disease",
  "n14079740": "onycholysis",
  "n14079861": "onychosis",
  "n14079971": "Paget's disease, osteitis deformans",
  "n14080210": "rheumatism",
  "n14080352": "periarteritis nodosa, polyarteritis nodosa",
  "n14080622": "periodontal disease, periodontitis",
  "n14080836": "pyorrhea, pyorrhoea, pyorrhea alveolaris, Riggs' disease",
  "n14081025": "pericementoclasia",
  "n14081124": "alveolar resorption",
  "n14081216": "gingivitis",
  "n14081297": "ulatrophia",
  "n14081375": "attack",
  "n14081665": "anxiety attack",
  "n14081789": "flare",
  "n14081941": "seizure, ictus, raptus",
  "n14082186": "touch, spot",
  "n14082303": "stroke, apoplexy, cerebrovascular accident, CVA",
  "n14082595": "convulsion",
  "n14082788": "paroxysm, fit, convulsion",
  "n14082984": "hysterics",
  "n14083061": "clonus",
  "n14083200": "epileptic seizure",
  "n14083368": "grand mal, generalized seizure, epilepsia major",
  "n14083559": "petit mal, epilepsia minor",
  "n14083790": "mental disorder, mental disturbance, disturbance, psychological disorder, folie",
  "n14084291": "Asperger's syndrome",
  "n14084502": "metabolic disorder",
  "n14084709": "alkaptonuria, alcaptonuria",
  "n14084880": "nervous disorder, neurological disorder, neurological disease",
  "n14085220": "brain damage",
  "n14085474": "akinesis, akinesia",
  "n14085592": "alalia",
  "n14085708": "brain disorder, encephalopathy, brain disease",
  "n14086036": "cystoplegia, cystoparalysis",
  "n14086143": "epilepsy",
  "n14086528": "akinetic epilepsy",
  "n14086626": "cortical epilepsy, focal epilepsy",
  "n14086924": "focal seizure",
  "n14087096": "raptus hemorrhagicus",
  "n14087208": "diplegia",
  "n14087321": "protuberance",
  "n14087513": "grand mal epilepsy, grand mal, generalized epilepsy, epilepsia major",
  "n14087758": "Jacksonian epilepsy",
  "n14087938": "myoclonus epilepsy, Lafora's disease",
  "n14088127": "petit mal epilepsy, petit mal, epilepsia minor",
  "n14088412": "absence, absence seizure",
  "n14088758": "complex absence",
  "n14088918": "pure absence, simple absence",
  "n14089080": "subclinical absence",
  "n14089240": "musicogenic epilepsy",
  "n14089339": "photogenic epilepsy",
  "n14089450": "posttraumatic epilepsy, traumatic epilepsy",
  "n14089592": "procursive epilepsy",
  "n14089719": "progressive vaccinia, vaccinia gangrenosa",
  "n14089974": "psychomotor epilepsy, temporal lobe epilepsy",
  "n14090245": "reflex epilepsy",
  "n14090416": "sensory epilepsy",
  "n14090530": "status epilepticus",
  "n14090727": "tonic epilepsy",
  "n14090842": "Erb's palsy, Erb-Duchenne paralysis",
  "n14091018": "nympholepsy",
  "n14091126": "apraxia",
  "n14091254": "ataxia, ataxy, dyssynergia, motor ataxia",
  "n14091525": "Friedreich's ataxia, herediatry spinal ataxia",
  "n14091757": "hereditary cerebellar ataxia",
  "n14091965": "atopognosia, atopognosis",
  "n14092114": "brachydactyly, brachydactylia",
  "n14092247": "cryptorchidy, cryptorchidism, cryptorchism",
  "n14092456": "monorchism, monorchidism",
  "n14092577": "dyskinesia",
  "n14092704": "tardive dyskinesia",
  "n14092925": "deviated septum",
  "n14093096": "deviated nasal septum",
  "n14093325": "dextrocardia",
  "n14093464": "ectrodactyly",
  "n14093591": "enteroptosis",
  "n14093723": "erethism",
  "n14093874": "fetal distress, foetal distress",
  "n14094068": "multiple sclerosis, MS, disseminated sclerosis, disseminated multiple sclerosis",
  "n14094350": "paralysis agitans, Parkinsonism, Parkinson's disease, Parkinson's syndrome, Parkinson's, shaking palsy",
  "n14094653": "cerebral palsy, spastic paralysis",
  "n14094881": "chorea",
  "n14095128": "choriomeningitis",
  "n14095273": "flaccid paralysis",
  "n14095439": "orthochorea",
  "n14095568": "Sydenham's chorea, Saint Vitus dance, St. Vitus dance",
  "n14095742": "tarantism",
  "n14095948": "agraphia, anorthography, logagraphia",
  "n14096123": "acataphasia",
  "n14096328": "aphagia",
  "n14096412": "amaurosis",
  "n14096598": "amblyopia",
  "n14096724": "ametropia",
  "n14096957": "emmetropia",
  "n14097162": "aniseikonia",
  "n14097321": "anorthopia",
  "n14097432": "aphakia",
  "n14097574": "aphasia",
  "n14097893": "auditory aphasia, acoustic aphasia, word deafness",
  "n14098078": "conduction aphasia, associative aphasia",
  "n14098347": "global aphasia, total aphasia",
  "n14098458": "motor aphasia, Broca's aphasia, ataxic aphasia, expressive aphasia, nonfluent aphasia",
  "n14098663": "nominal aphasia, anomic aphasia, anomia, amnesic aphasia, amnestic aphasia",
  "n14098864": "transcortical aphasia",
  "n14099050": "visual aphasia, alexia, word blindness",
  "n14099172": "Wernicke's aphasia, fluent aphasia, receptive aphasia, sensory aphasia, impressive aphasia",
  "n14099439": "dyscalculia",
  "n14099552": "dysgraphia",
  "n14099643": "dyslexia",
  "n14099785": "dysphasia",
  "n14099933": "agnosia",
  "n14100092": "anarthria",
  "n14100243": "auditory agnosia",
  "n14100370": "visual agnosia",
  "n14100494": "Creutzfeldt-Jakob disease, CJD, Jakob-Creutzfeldt disease",
  "n14100769": "occlusion",
  "n14100957": "laryngospasm",
  "n14101083": "embolism",
  "n14101317": "air embolism, aeroembolism, gas embolism",
  "n14101568": "fat embolism",
  "n14101801": "pulmonary embolism",
  "n14101935": "thromboembolism",
  "n14102075": "thrombosis",
  "n14102347": "cerebral thrombosis",
  "n14102454": "coronary occlusion",
  "n14102631": "coronary heart disease",
  "n14102831": "coronary thrombosis, coronary",
  "n14103018": "milk leg, white leg, phlegmasia alba dolens",
  "n14103180": "hepatomegaly, megalohepatia",
  "n14103288": "heart disease, cardiopathy",
  "n14103510": "high blood pressure, hypertension",
  "n14103785": "inversion",
  "n14103998": "transposition, heterotaxy",
  "n14104118": "keratectasia",
  "n14104218": "keratoconus",
  "n14104367": "orthostatic hypotension, postural hypotension",
  "n14104522": "hypotension",
  "n14104645": "essential hypertension, hyperpiesia, hyperpiesis",
  "n14104830": "portal hypertension",
  "n14105091": "malignant hypertension",
  "n14105386": "secondary hypertension",
  "n14105504": "white-coat hypertension",
  "n14105629": "amyotrophia, amyotrophy",
  "n14105737": "amyotrophic lateral sclerosis, ALS, Lou Gehrig's disease",
  "n14106025": "aneurysm, aneurism",
  "n14106352": "aortic aneurysm",
  "n14106456": "abdominal aortic aneurysm, AAA",
  "n14106609": "aortic stenosis",
  "n14106727": "enterostenosis",
  "n14106823": "laryngostenosis",
  "n14106917": "pulmonary stenosis",
  "n14107085": "pyloric stenosis",
  "n14107249": "rhinostenosis",
  "n14107374": "stenosis, stricture",
  "n14107653": "cerebral aneurysm",
  "n14107750": "intracranial aneurysm",
  "n14107849": "ventricular aneurysm",
  "n14108039": "angina pectoris, angina",
  "n14108232": "arteriolosclerosis",
  "n14108324": "arteriosclerosis, arterial sclerosis, hardening of the arteries, induration of the arteries, coronary-artery disease",
  "n14108576": "atherogenesis",
  "n14108713": "atherosclerosis, coronary artery disease",
  "n14108938": "athetosis",
  "n14109101": "kuru",
  "n14109485": "nerve compression",
  "n14109691": "nerve entrapment",
  "n14109889": "arteriosclerosis obliterans",
  "n14110025": "ascites",
  "n14110146": "azymia",
  "n14110219": "bacteremia, bacteriemia, bacteriaemia",
  "n14110411": "sclerosis, induration",
  "n14110674": "cardiac arrhythmia, arrhythmia",
  "n14110966": "cardiomyopathy, myocardiopathy",
  "n14111133": "hypertrophic cardiomyopathy",
  "n14111355": "flutter",
  "n14111540": "gallop rhythm, cantering rhythm",
  "n14111715": "mitral valve prolapse",
  "n14111915": "mitral stenosis, mitral valve stenosis",
  "n14112101": "circulatory failure",
  "n14112255": "heart failure, coronary failure",
  "n14112466": "valvular heart disease",
  "n14112719": "congestive heart failure",
  "n14112855": "heart attack",
  "n14113021": "myocardial infarction, myocardial infarct, MI",
  "n14113228": "kidney disease, renal disorder, nephropathy, nephrosis",
  "n14113471": "insufficiency",
  "n14113636": "coronary insufficiency, cardiac insufficiency",
  "n14113798": "nephritis, Bright's disease",
  "n14113938": "nephrosclerosis, nephroangiosclerosis",
  "n14114186": "polycystic kidney disease, PKD",
  "n14114365": "polyuria",
  "n14114555": "renal failure, kidney failure",
  "n14114773": "renal insufficiency",
  "n14114886": "acute renal failure, acute kidney failure",
  "n14115086": "chronic renal failure, chronic kidney failure",
  "n14115245": "cholelithiasis",
  "n14115351": "enterolithiasis",
  "n14115454": "nephrocalcinosis",
  "n14115648": "nephrolithiasis, renal lithiasis",
  "n14115799": "lipomatosis",
  "n14115914": "lithiasis",
  "n14116078": "glomerulonephritis",
  "n14116321": "liver disease",
  "n14116482": "cirrhosis, cirrhosis of the liver",
  "n14116672": "fatty liver",
  "n14116908": "Addison's disease, Addison's syndrome, hypoadrenalism, hypoadrenocorticism",
  "n14117173": "adenopathy",
  "n14117317": "aldosteronism, hyperaldosteronism",
  "n14117449": "Cushing's disease, hyperadrenalism",
  "n14117668": "Cushing's syndrome, hyperadrenocorticism",
  "n14117805": "diabetes",
  "n14118138": "diabetes mellitus, DM",
  "n14118423": "type I diabetes, insulin-dependent diabetes mellitus, IDDM, juvenile-onset diabetes, juvenile diabetes, growth-onset diabetes, ketosis-prone diabetes, ketoacidosis-prone diabetes, autoimmune diabetes",
  "n14118936": "type II diabetes, non-insulin-dependent diabetes mellitus, NIDDM, non-insulin-dependent diabetes, ketosis-resistant diabetes mellitus, ketosis-resistant diabetes, ketoacidosis-resistant diabetes mellitus, ketoacidosis-resistant diabetes, adult-onset diabetes mellitus, adult-onset diabetes, maturity-onset diabetes mellitus, maturity-onset diabetes, mature-onset diabetes",
  "n14119598": "nephrogenic diabetes insipidus",
  "n14119770": "diabetes insipidus",
  "n14120096": "latent diabetes, chemical diabetes",
  "n14120310": "angioedema, atrophedema, giant hives, periodic edema, Quincke's edema",
  "n14120614": "lymphedema",
  "n14120767": "hyperthyroidism, thyrotoxicosis",
  "n14121058": "Graves' disease, exophthalmic goiter",
  "n14121276": "hypothyroidism",
  "n14121500": "myxedema, myxoedema",
  "n14121667": "cretinism",
  "n14121804": "achondroplasia, achondroplasty, osteosclerosis congenita, chondrodystrophy",
  "n14122053": "communicable disease",
  "n14122235": "contagious disease, contagion",
  "n14122497": "influenza, flu, grippe",
  "n14122670": "Asian influenza, Asiatic flu",
  "n14122813": "swine influenza, swine flu",
  "n14123044": "measles, rubeola, morbilli",
  "n14123259": "German measles, rubella, three-day measles, epidemic roseola",
  "n14123510": "diphtheria",
  "n14123759": "exanthema subitum, roseola infantum, roseola infantilis, pseudorubella",
  "n14124070": "scarlet fever, scarlatina",
  "n14124232": "pox",
  "n14124423": "smallpox, variola, variola major",
  "n14124688": "alastrim, variola minor, pseudosmallpox, pseudovariola, milk pox, white pox, West Indian smallpox, Cuban itch, Kaffir pox",
  "n14124931": "Vincent's angina, Vincent's infection, trench mouth",
  "n14125159": "blastomycosis",
  "n14125333": "chromoblastomycosis",
  "n14125466": "tinea, ringworm, roundworm",
  "n14125774": "dhobi itch",
  "n14125881": "kerion",
  "n14126078": "tinea pedis, athlete's foot",
  "n14126181": "tinea barbae, barber's itch",
  "n14126293": "tinea capitis",
  "n14126412": "tinea corporis",
  "n14126519": "tinea cruris, jock itch, eczema marginatum",
  "n14126660": "blindness, sightlessness, cecity",
  "n14126908": "legal blindness",
  "n14127100": "tinea unguium",
  "n14127211": "infectious disease",
  "n14127782": "AIDS, acquired immune deficiency syndrome",
  "n14128029": "brucellosis, undulant fever, Malta fever, Gibraltar fever, Rock fever, Mediterranean fever",
  "n14128331": "agammaglobulinemia",
  "n14128536": "anergy",
  "n14128650": "hypogammaglobulinemia",
  "n14128812": "severe combined immunodeficiency, severe combined immunodeficiency disease, SCID",
  "n14129223": "ADA-SCID",
  "n14129351": "X-linked SCID, X-SCID",
  "n14129579": "cholera, Asiatic cholera, Indian cholera, epidemic cholera",
  "n14129784": "dengue, dengue fever, dandy fever, breakbone fever",
  "n14129999": "dysentery",
  "n14130166": "epidemic disease",
  "n14130354": "hepatitis",
  "n14130518": "viral hepatitis",
  "n14130661": "hepatitis A, infectious hepatitis",
  "n14130942": "hepatitis B, serum hepatitis",
  "n14131247": "hepatitis delta, delta hepatitis",
  "n14131353": "hepatitis C",
  "n14131651": "liver cancer, cancer of the liver",
  "n14131950": "herpes",
  "n14132102": "herpes simplex",
  "n14132375": "oral herpes, herpes labialis, cold sore, fever blister",
  "n14132524": "genital herpes, herpes genitalis",
  "n14132829": "herpes zoster, zoster, shingles",
  "n14132975": "chickenpox, varicella",
  "n14133159": "venereal disease, VD, venereal infection, social disease, Cupid's itch, Cupid's disease, Venus's curse, dose, sexually transmitted disease, STD",
  "n14133543": "gonorrhea, gonorrhoea, clap",
  "n14133750": "granuloma inguinale, granuloma venereum",
  "n14133985": "syphilis, syph, pox, lues venerea, lues",
  "n14134351": "primary syphilis",
  "n14134482": "secondary syphilis",
  "n14134623": "tertiary syphilis",
  "n14134819": "tabes dorsalis, locomotor ataxia",
  "n14135065": "neurosyphilis",
  "n14135181": "tabes",
  "n14135277": "infectious mononucleosis, mononucleosis, mono, glandular fever, kissing disease",
  "n14135623": "Ebola hemorrhagic fever, Ebola fever, Ebola",
  "n14135956": "Lassa fever",
  "n14136187": "leprosy, Hansen's disease",
  "n14136544": "tuberculoid leprosy",
  "n14136674": "lepromatous leprosy",
  "n14136868": "necrotizing enterocolitis, NEC",
  "n14137066": "listeriosis, listeria meningitis",
  "n14137408": "lymphocytic choriomeningitis",
  "n14137561": "lymphogranuloma venereum, LGV, lymphopathia venereum",
  "n14137829": "meningitis",
  "n14138178": "mumps, epidemic parotitis",
  "n14138364": "cerebrospinal meningitis, epidemic meningitis, brain fever, cerebrospinal fever",
  "n14138538": "paratyphoid, paratyphoid fever",
  "n14138691": "plague, pestilence, pest, pestis",
  "n14139015": "bubonic plague, pestis bubonica, glandular plague",
  "n14139333": "ambulant plague, ambulatory plague, pestis ambulans",
  "n14139462": "Black Death, Black Plague",
  "n14139661": "pneumonic plague, pulmonic plague, plague pneumonia",
  "n14139959": "septicemic plague",
  "n14140176": "poliomyelitis, polio, infantile paralysis, acute anterior poliomyelitis",
  "n14140391": "Pott's disease",
  "n14140533": "ratbite fever",
  "n14140781": "rickettsial disease, rickettsiosis",
  "n14141062": "typhus, typhus fever",
  "n14141238": "murine typhus, rat typhus, urban typhus, endemic typhus",
  "n14141490": "spotted fever",
  "n14141656": "Rocky Mountain spotted fever, mountain fever, tick fever",
  "n14141821": "Q fever",
  "n14141912": "rickettsialpox",
  "n14142207": "trench fever",
  "n14142326": "tsutsugamushi disease, scrub typhus",
  "n14142459": "relapsing fever, recurrent fever",
  "n14142768": "rheumatic fever",
  "n14142983": "rheumatic heart disease",
  "n14143239": "sweating sickness, miliary fever",
  "n14143415": "tuberculosis, TB, T.B.",
  "n14143818": "miliary tuberculosis",
  "n14144064": "pulmonary tuberculosis, consumption, phthisis, wasting disease, white plague",
  "n14144247": "scrofula, struma, king's evil",
  "n14144416": "typhoid, typhoid fever, enteric fever",
  "n14144626": "whooping cough, pertussis",
  "n14144761": "yaws, frambesia, framboesia",
  "n14144960": "yellow jack, yellow fever, black vomit",
  "n14145095": "respiratory disease, respiratory illness, respiratory disorder",
  "n14145501": "cold, common cold",
  "n14145759": "head cold",
  "n14145911": "asthma, asthma attack, bronchial asthma",
  "n14146128": "status asthmaticus",
  "n14146273": "bronchitis",
  "n14146459": "bronchiolitis",
  "n14146571": "chronic bronchitis",
  "n14146774": "chronic obstructive pulmonary disease",
  "n14147014": "coccidioidomycosis, coccidiomycosis, valley fever, desert rheumatism",
  "n14147212": "cryptococcosis",
  "n14147380": "emphysema, pulmonary emphysema",
  "n14147627": "pneumonia",
  "n14147964": "atypical pneumonia, primary atypical pneumonia, mycoplasmal pneumonia",
  "n14148231": "bronchopneumonia, bronchial pneumonia",
  "n14148411": "double pneumonia",
  "n14148510": "interstitial pneumonia",
  "n14148646": "lobar pneumonia",
  "n14148834": "Legionnaires' disease",
  "n14149167": "pneumococcal pneumonia",
  "n14149290": "pneumocytosis, pneumocystis pneumonia, pneumocystis carinii pneumonia, interstitial plasma cell pneumonia",
  "n14149536": "pneumothorax",
  "n14149773": "psittacosis, parrot fever, ornithosis",
  "n14149963": "pneumoconiosis, pneumonoconiosis",
  "n14150210": "anthracosis, black lung, black lung disease, coal miner's lung",
  "n14150363": "asbestosis",
  "n14150470": "siderosis",
  "n14150612": "silicosis",
  "n14150740": "respiratory distress syndrome, respiratory distress syndrome of the newborn, hyaline membrane disease",
  "n14151139": "genetic disease, genetic disorder, genetic abnormality, genetic defect, congenital disease, inherited disease, inherited disorder, hereditary disease, hereditary condition",
  "n14151884": "abetalipoproteinemia",
  "n14152169": "ablepharia",
  "n14152279": "albinism",
  "n14152492": "macrencephaly",
  "n14152617": "anencephaly, anencephalia",
  "n14152803": "adactylia, adactyly, adactylism",
  "n14152924": "ametria",
  "n14153010": "color blindness, colour blindness, color vision deficiency, colour vision deficiency",
  "n14153285": "diplopia, double vision",
  "n14153468": "epispadias",
  "n14153616": "dichromacy, dichromatism, dichromatopsia, dichromia, dichromasy",
  "n14153982": "red-green dichromacy, red-green color blindness, red-green colour blindness",
  "n14154168": "deuteranopia, Daltonism, green-blindness",
  "n14154421": "protanopia, red-blindness",
  "n14154669": "yellow-blue dichromacy, yellow-blue color blindness",
  "n14154832": "tetartanopia, yellow-blindness",
  "n14155031": "tritanopia, blue-blindness",
  "n14155274": "monochromacy, monochromatism, monochromatic vision, monochromia, monochromasy",
  "n14155506": "cystic fibrosis, CF, fibrocystic disease of the pancreas, pancreatic fibrosis, mucoviscidosis",
  "n14155834": "inborn error of metabolism",
  "n14156134": "galactosemia",
  "n14156345": "Gaucher's disease",
  "n14156488": "Hirschsprung's disease, congenital megacolon",
  "n14156740": "Horner's syndrome",
  "n14156976": "Huntington's chorea, Huntington's disease",
  "n14157163": "Hurler's syndrome, Hurler's disease, gargoylism, dysostosis multiplex, lipochondrodystrophy",
  "n14157527": "mucopolysaccharidosis",
  "n14157782": "malignant hyperthermia",
  "n14157967": "Marfan's syndrome",
  "n14158179": "neurofibromatosis, von Recklinghausen's disease",
  "n14158424": "osteogenesis imperfecta",
  "n14158594": "hyperbetalipoproteinemia",
  "n14158812": "hypobetalipoproteinemia",
  "n14158997": "ichthyosis",
  "n14159153": "clinocephaly, clinocephalism",
  "n14159318": "clinodactyly",
  "n14159459": "macroglossia",
  "n14159623": "mongolism, mongolianism, Down's syndrome, Down syndrome, trisomy 21",
  "n14159887": "maple syrup urine disease, branched chain ketoaciduria",
  "n14160179": "McArdle's disease",
  "n14160365": "muscular dystrophy, dystrophy",
  "n14160665": "oligodactyly",
  "n14160786": "oligodontia",
  "n14160903": "otosclerosis",
  "n14161075": "Becker muscular dystrophy",
  "n14161515": "distal muscular dystrophy",
  "n14161795": "Duchenne's muscular dystrophy, pseudohypertrophic dystrophy",
  "n14162025": "autosomal dominant disease, autosomal dominant disorder",
  "n14162275": "autosomal recessive disease, autosomal recessive defect",
  "n14162563": "limb-girdle muscular dystrophy",
  "n14162943": "lysinemia",
  "n14163182": "myotonic muscular dystrophy, myotonic dystrophy, myotonia atrophica, Steinert's disease",
  "n14163676": "oculopharyngeal muscular dystrophy",
  "n14164007": "Niemann-Pick disease",
  "n14164190": "oxycephaly, acrocephaly",
  "n14164341": "aplastic anemia, aplastic anaemia",
  "n14164548": "erythroblastosis fetalis",
  "n14164866": "Fanconi's anemia, Fanconi's anaemia, congenital pancytopenia",
  "n14165081": "favism",
  "n14165240": "hemolytic anemia, haemolytic anaemia",
  "n14165373": "hyperchromic anemia, hyperchromic anaemia",
  "n14165544": "hypochromic anemia, hypochromic anaemia",
  "n14165730": "hypoplastic anemia, hypoplastic anaemia",
  "n14165909": "iron deficiency anemia, iron deficiency anaemia",
  "n14166118": "ischemia, ischaemia",
  "n14166358": "ischemic stroke, ischaemic stroke",
  "n14166589": "transient ischemic attack, TIA",
  "n14166775": "chlorosis, greensickness",
  "n14166968": "macrocytic anemia, macrocytic anaemia",
  "n14167123": "microcytic anemia, microcytic anaemia",
  "n14167298": "parasitemia, parasitaemia",
  "n14167426": "pernicious anemia, pernicious anaemia, malignant anemia, malignant anaemia",
  "n14167773": "megaloblastic anemia, megaloblastic anaemia",
  "n14168010": "metaplastic anemia, metaplastic anaemia",
  "n14168176": "refractory anemia, refractory anaemia",
  "n14168447": "sideroblastic anemia, sideroblastic anaemia, siderochrestic anemia, siderochrestic anaemia",
  "n14168653": "sideropenia",
  "n14168792": "sickle-cell anemia, sickle-cell anaemia, sickle-cell disease, crescent-cell anemia, crescent-cell anaemia, drepanocytic anemia, drepanocytic anaemia",
  "n14169128": "Spielmeyer-Vogt disease, juvenile amaurotic idiocy",
  "n14169364": "Tay-Sachs disease, Tay-Sachs, Sachs disease, infantile amaurotic idiocy",
  "n14169717": "thrombasthenia",
  "n14169897": "tyrosinemia",
  "n14170070": "Werdnig-Hoffman disease",
  "n14170337": "hemophilia, haemophilia, bleeder's disease",
  "n14170623": "afibrinogenemia",
  "n14170772": "hemophilia A, haemophilia A, classical hemophilia, classical haemophilia",
  "n14170987": "hemophilia B, haemophilia B, Christmas disease",
  "n14171176": "von Willebrand's disease, angiohemophilia, vascular hemophilia",
  "n14171492": "congenital afibrinogenemia",
  "n14171682": "inflammatory disease",
  "n14172005": "gastroenteritis, stomach flu, intestinal flu",
  "n14172217": "cholera infantum",
  "n14172383": "cholera morbus, collywobbles",
  "n14172558": "pelvic inflammatory disease, PID",
  "n14172873": "empyema",
  "n14173013": "pleurisy",
  "n14173175": "purulent pleurisy",
  "n14173295": "pleuropneumonia",
  "n14173397": "pyelitis",
  "n14173484": "sore throat, pharyngitis, raw throat",
  "n14173625": "angina",
  "n14173823": "quinsy, peritonsillar abscess",
  "n14174011": "croup, spasmodic laryngitis",
  "n14174208": "glossoptosis",
  "n14174316": "hypermotility",
  "n14174455": "indisposition",
  "n14174549": "infection",
  "n14175165": "amebiasis, amoebiasis, amebiosis, amoebiosis",
  "n14175313": "amebic dysentery, amoebic dysentery",
  "n14175579": "chlamydia",
  "n14175727": "fascioliasis, fasciolosis",
  "n14175903": "fasciolopsiasis",
  "n14176051": "Guinea worm disease, Guinea worm, dracunculiasis",
  "n14176372": "enterobiasis",
  "n14176570": "felon, whitlow",
  "n14176715": "focal infection",
  "n14176895": "fungal infection, mycosis",
  "n14177210": "giardiasis",
  "n14177423": "hemorrhagic fever, haemorrhagic fever, viral hemorrhagic fever, viral haemorrhagic fever, VHF",
  "n14177848": "herpangia",
  "n14178077": "leishmaniasis, leishmaniosis, kala azar",
  "n14178326": "nonsocial infection, cross infection",
  "n14178482": "opportunistic infection",
  "n14178794": "paronychia",
  "n14178913": "protozoal infection",
  "n14179104": "respiratory tract infection, respiratory infection",
  "n14179276": "lower respiratory infection",
  "n14179390": "Crimea-Congo hemorrhagic fever",
  "n14179569": "Rift Valley fever",
  "n14179744": "HIV",
  "n14179839": "viral pneumonia",
  "n14179946": "severe acute respiratory syndrome, SARS",
  "n14180213": "upper respiratory infection",
  "n14180327": "scabies, itch",
  "n14180565": "schistosomiasis, bilharzia, bilharziasis",
  "n14180848": "sepsis",
  "n14181049": "visceral leishmaniasis, kala-azar, Assam fever, dumdum fever",
  "n14181187": "cutaneous leishmaniasis, Old World leishmaniasis, oriental sore, tropical sore, Delhi boil, Aleppo boil",
  "n14181409": "mucocutaneous leishmaniasis, New World leishmaniasis, American leishmaniasis, leishmaniasis americana, nasopharyngeal leishmaniasis",
  "n14181713": "candidiasis, moniliasis, monilia disease",
  "n14181948": "dermatomycosis, dermatophytosis",
  "n14182103": "favus",
  "n14182242": "keratomycosis",
  "n14182332": "phycomycosis",
  "n14182453": "sporotrichosis",
  "n14182568": "thrush",
  "n14182697": "focus, focal point, nidus",
  "n14182871": "aspergillosis",
  "n14183065": "sore",
  "n14183210": "boil, furuncle",
  "n14183337": "gumboil",
  "n14183420": "blain",
  "n14183522": "chilblain, chilblains, pernio",
  "n14183692": "kibe",
  "n14183774": "carbuncle",
  "n14183926": "cartilaginification",
  "n14184067": "chancre",
  "n14184254": "fester, suppurating sore",
  "n14184390": "gall",
  "n14184506": "saddle sore",
  "n14184608": "shigellosis, bacillary dysentery",
  "n14184801": "staphylococcal infection",
  "n14184986": "streptococcal sore throat, strep throat, streptococcus tonsilitis, septic sore throat, throat infection",
  "n14185219": "sty, stye, hordeolum, eye infection",
  "n14185353": "superinfection",
  "n14185503": "suprainfection",
  "n14185620": "tapeworm infection",
  "n14185803": "tetanus, lockjaw",
  "n14186046": "toxoplasmosis",
  "n14186256": "trichomoniasis",
  "n14186340": "viral infection, virus infection",
  "n14186541": "arthritis",
  "n14186738": "rheumatoid arthritis, atrophic arthritis, rheumatism",
  "n14187229": "rheumatoid factor",
  "n14187378": "autoimmune disease, autoimmune disorder",
  "n14187869": "psoriatic arthritis",
  "n14188030": "Still's disease, juvenile rheumatoid arthritis",
  "n14188238": "osteoarthritis, degenerative arthritis, degenerative joint disease",
  "n14188467": "osteosclerosis",
  "n14188568": "housemaid's knee",
  "n14188702": "cystitis",
  "n14188804": "gout, gouty arthritis, urarthritis",
  "n14189065": "spondylarthritis",
  "n14189204": "blood disease, blood disorder",
  "n14189837": "blood poisoning, septicemia, septicaemia",
  "n14190132": "sapremia, sapraemia",
  "n14190284": "ozone sickness",
  "n14190493": "puerperal fever, childbed fever",
  "n14190736": "pyemia, pyaemia",
  "n14190907": "toxemia, toxaemia",
  "n14191037": "toxemia of pregnancy, toxaemia of pregnancy, toxemia, toxaemia",
  "n14191284": "eclampsia",
  "n14191442": "preeclampsia, pre-eclampsia",
  "n14191646": "eosinopenia",
  "n14191756": "erythroblastosis",
  "n14191903": "hemoglobinemia, haemoglobinemia",
  "n14192034": "hemoglobinopathy, haemoglobinopathy",
  "n14192199": "hemoptysis, haemoptysis",
  "n14192376": "Hand-Schuller-Christian disease, Schuller-Christian disease",
  "n14192672": "Haverhill fever",
  "n14192790": "histiocytosis",
  "n14192942": "hydatid mole, hydatidiform mole, molar pregnancy",
  "n14193202": "hydramnios",
  "n14193325": "water on the knee",
  "n14193421": "hydremia",
  "n14193571": "hydrocele",
  "n14193711": "lipidosis",
  "n14193925": "lipemia, lipaemia, lipidemia, lipidaemia, lipoidemia, lipoidaemia, hyperlipemia, hyperlipaemia, hyperlipidemia, hyperlipidaemia, hyperlipoidemia, hyperlipoidaemia",
  "n14194183": "lysine intolerance",
  "n14194345": "lysogeny, lysogenicity",
  "n14194690": "hypothrombinemia",
  "n14194942": "hypervolemia, hypervolaemia",
  "n14195112": "hypovolemia, hypovolaemia",
  "n14195315": "anemia, anaemia",
  "n14195715": "thalassemia, thalassaemia, Mediterranean anemia, Mediterranean anaemia",
  "n14195939": "Cooley's anemia, Cooley's anaemia, thalassemia major, thalassaemia major",
  "n14196221": "leukocytosis, leucocytosis",
  "n14196405": "leukopenia, leucopenia",
  "n14196543": "neutropenia",
  "n14196722": "cyclic neutropenia",
  "n14196823": "lymphocytopenia, lymphopenia",
  "n14196965": "lymphocytosis",
  "n14197099": "microcytosis",
  "n14197315": "polycythemia",
  "n14197468": "purpura, peliosis",
  "n14197628": "nonthrombocytopenic purpura",
  "n14197780": "essential thrombocytopenia",
  "n14198019": "thrombocytopenia, thrombopenia",
  "n14198200": "deficiency disease",
  "n14198380": "fibrocystic breast disease, fibrocystic disease of the breast, cystic breast disease, cystic mastitis",
  "n14198576": "avitaminosis, hypovitaminosis",
  "n14198829": "hypervitaminosis",
  "n14198997": "hypospadias",
  "n14199146": "lagophthalmos",
  "n14199264": "beriberi",
  "n14199389": "kakke disease",
  "n14199477": "goiter, goitre, struma, thyromegaly",
  "n14199700": "malnutrition",
  "n14199934": "kwashiorkor",
  "n14200098": "marasmus",
  "n14200301": "mental abnormality",
  "n14200418": "nanophthalmos",
  "n14200548": "organic brain syndrome",
  "n14200704": "zinc deficiency",
  "n14200873": "pellagra, Alpine scurvy, mal de la rosa, mal rosso, maidism, mayidism, Saint Ignatius' itch",
  "n14201311": "rickets, rachitis",
  "n14201520": "scurvy, scorbutus",
  "n14201682": "dermoid cyst",
  "n14201845": "galactocele",
  "n14201989": "hemorrhagic cyst, blood cyst, hematocyst",
  "n14202119": "hydatid",
  "n14202272": "nabothian cyst, nabothian follicle",
  "n14202417": "ovarian cyst",
  "n14202520": "chalazion, Meibomian cyst",
  "n14202673": "ranula",
  "n14202763": "sebaceous cyst, pilar cyst, wen, steatocystoma",
  "n14202996": "cyst",
  "n14203269": "pip",
  "n14203346": "motion sickness, kinetosis",
  "n14203605": "airsickness, air sickness",
  "n14203813": "car sickness",
  "n14203942": "seasickness, mal de mer, naupathia",
  "n14204095": "heatstroke, heat hyperpyrexia",
  "n14204253": "heat exhaustion, heat prostration",
  "n14204441": "algidity",
  "n14204586": "sunstroke, insolation, thermic fever, siriasis",
  "n14204763": "endometriosis, adenomyosis",
  "n14204950": "pathology",
  "n14206224": "adhesion",
  "n14206375": "symphysis",
  "n14206477": "synechia",
  "n14206722": "anterior synechia",
  "n14206826": "posterior synechia",
  "n14206929": "hemochromatosis, iron-storage disease, iron overload, bronzed diabetes",
  "n14207256": "classic hemochromatosis, idiopathic hemochromatosis",
  "n14207388": "acquired hemochromatosis",
  "n14207561": "infarct, infarction",
  "n14207711": "macrocytosis",
  "n14207809": "fibrosis",
  "n14207978": "myelofibrosis",
  "n14208065": "malacia",
  "n14208176": "osteomalacia",
  "n14208342": "mastopathy, mazopathy",
  "n14208438": "neuropathy",
  "n14208587": "Charcot-Marie-Tooth disease, hereditary motor and sensory neuropathy",
  "n14208988": "mononeuropathy",
  "n14209087": "multiple mononeuropathy",
  "n14209201": "myopathy",
  "n14209348": "dermatomyositis",
  "n14209635": "polymyositis",
  "n14209836": "inclusion body myositis",
  "n14210119": "osteopetrosis, Albers-Schonberg disease, marble bones disease",
  "n14210354": "osteoporosis",
  "n14210564": "priapism",
  "n14210716": "demineralization, demineralisation",
  "n14210888": "pyorrhea, pyorrhoea",
  "n14210971": "uremia, uraemia, azotemia, azotaemia",
  "n14211212": "azoturia",
  "n14211294": "lesion",
  "n14211440": "tubercle",
  "n14211609": "ulcer, ulceration",
  "n14211985": "aphthous ulcer",
  "n14212126": "bedsore, pressure sore, decubitus ulcer",
  "n14212303": "chancroid",
  "n14212402": "peptic ulcer, peptic ulceration",
  "n14212579": "duodenal ulcer",
  "n14212670": "gastric ulcer",
  "n14212759": "canker, canker sore",
  "n14212941": "noli-me-tangere",
  "n14213044": "noma",
  "n14213199": "affliction",
  "n14213328": "curvature",
  "n14213512": "deformity, malformation, misshapenness",
  "n14213867": "Arnold-Chiari deformity",
  "n14214001": "clawfoot, pes cavus",
  "n14214209": "cleft foot",
  "n14214355": "cleft lip, harelip, cheiloschisis",
  "n14214486": "cleft palate",
  "n14214584": "clubfoot, talipes",
  "n14214819": "talipes valgus",
  "n14214938": "talipes equinus",
  "n14215046": "talipes calcaneus",
  "n14215199": "pigeon breast, chicken breast",
  "n14215331": "blight",
  "n14215908": "alder blight",
  "n14216054": "apple blight, apple canker",
  "n14216152": "beet blight",
  "n14216234": "blister blight",
  "n14216318": "blister blight",
  "n14216404": "cane blight",
  "n14216546": "celery blight",
  "n14216653": "chestnut blight, chestnut canker, chestnut-bark disease",
  "n14216793": "coffee blight",
  "n14216888": "collar blight",
  "n14217002": "fire blight, pear blight",
  "n14217147": "blight canker",
  "n14217253": "halo blight, halo spot, bean blight",
  "n14217360": "halo blight",
  "n14217473": "head blight",
  "n14217581": "wheat scab",
  "n14217676": "late blight",
  "n14217897": "leaf blight",
  "n14218057": "leaf disease",
  "n14218158": "needle blight, needle cast, leaf cast",
  "n14218293": "oak blight",
  "n14218401": "peach blight",
  "n14218493": "potato blight, potato mold, potato disease, potato mildew, potato murrain",
  "n14218637": "rim blight",
  "n14218717": "spinach blight",
  "n14218805": "spur blight",
  "n14218887": "stem blight",
  "n14218991": "stripe blight",
  "n14219068": "thread blight",
  "n14219187": "tomato blight, tomato yellows",
  "n14219290": "twig blight",
  "n14219394": "walnut blight",
  "n14219487": "sandfly fever, pappataci fever, phlebotomus",
  "n14219661": "skin disease, disease of the skin, skin disorder",
  "n14220308": "lupus vulgaris",
  "n14220474": "ankylosing spondylitis, Marie-Strumpell disease, rheumatoid spondylitis",
  "n14220735": "discoid lupus erythematosus, DLE",
  "n14220998": "Hashimoto's disease",
  "n14221138": "lupus erythematosus, LE",
  "n14221311": "systemic lupus erythematosus, SLE, disseminated lupus erythematosus",
  "n14221601": "acantholysis",
  "n14221741": "acanthosis",
  "n14221924": "acanthosis nigricans, keratosis nigricans",
  "n14222112": "acne",
  "n14222352": "acne rosacea, rosacea",
  "n14222541": "acne vulgaris",
  "n14222685": "actinic dermatitis",
  "n14222788": "atopic dermatitis, atopic eczema",
  "n14222918": "bubble gum dermatitis",
  "n14223074": "contact dermatitis",
  "n14223296": "Rhus dermatitis",
  "n14223492": "poison ivy",
  "n14223638": "poison oak",
  "n14223752": "poison sumac",
  "n14223870": "cradle cap",
  "n14223978": "diaper rash, diaper dermatitis",
  "n14224170": "hypericism",
  "n14224341": "neurodermatitis",
  "n14224547": "schistosome dermatitis, swimmer's itch",
  "n14224757": "dermatitis",
  "n14225045": "dermatosis",
  "n14225215": "baker's eczema",
  "n14225346": "allergic eczema",
  "n14225463": "eczema herpeticum",
  "n14225632": "eczema vaccinatum, Kaposi's varicelliform eruption",
  "n14225877": "lichtenoid eczema, chronic eczema, eczema hypertrophicum",
  "n14226056": "eczema",
  "n14226303": "erythema",
  "n14226567": "erythema multiforme",
  "n14226709": "erythema nodosum",
  "n14226862": "hickey, love bite",
  "n14227015": "erythema nodosum leprosum, ENL",
  "n14227218": "erythroderma",
  "n14227357": "flare",
  "n14227488": "furunculosis",
  "n14227613": "impetigo",
  "n14227806": "jungle rot",
  "n14227906": "keratoderma, keratodermia",
  "n14228038": "keratonosis",
  "n14228148": "keratosis",
  "n14228376": "actinic keratosis",
  "n14228512": "keratosis blennorrhagica, keratoderma blennorrhagica",
  "n14228712": "keratosis follicularis, Darier's disease",
  "n14228885": "keratosis pilaris",
  "n14229067": "seborrheic keratosis",
  "n14229248": "leukoderma",
  "n14229403": "lichen",
  "n14229604": "lichen planus, lichen ruber planus",
  "n14229788": "livedo",
  "n14229912": "lupus",
  "n14230083": "melanosis, melanism",
  "n14230232": "molluscum",
  "n14230358": "molluscum contagiosum",
  "n14230588": "necrobiosis lipoidica, necrobiosis lipoidica diabeticorum",
  "n14230800": "pemphigus",
  "n14231022": "pityriasis",
  "n14231207": "dandruff",
  "n14231330": "pityriasis alba",
  "n14231504": "pityriasis rosea",
  "n14231640": "prurigo",
  "n14231794": "psoriasis",
  "n14232034": "rhagades",
  "n14232157": "Saint Anthony's fire",
  "n14232299": "erysipelas",
  "n14232460": "scleredema",
  "n14232596": "seborrhea",
  "n14232740": "seborrheic dermatitis, seborrheic eczema",
  "n14232945": "vitiligo",
  "n14233155": "xanthelasma",
  "n14233267": "xanthoma",
  "n14233545": "xanthoma disseminatum",
  "n14233717": "xanthomatosis, xanthoma multiplex, cholesterosis cutis, lipid granulomatosis, lipoid granulomatosis",
  "n14233974": "xanthosis",
  "n14234074": "growth",
  "n14234317": "exostosis",
  "n14234436": "polyp, polypus",
  "n14234611": "adenomatous polyp",
  "n14234867": "sessile polyp",
  "n14234950": "pedunculated polyp",
  "n14235066": "peduncle",
  "n14235200": "tumor, tumour, neoplasm",
  "n14235667": "acanthoma, skin tumor",
  "n14235793": "adenoma",
  "n14235928": "angioma",
  "n14236130": "chondroma",
  "n14236226": "benign tumor, benign tumour, nonmalignant tumor, nonmalignant tumour, nonmalignant neoplasm",
  "n14236595": "blastoma, blastocytoma, embryonal carcinosarcoma",
  "n14236743": "brain tumor, brain tumour",
  "n14236872": "glioblastoma, spongioblastoma",
  "n14237032": "glioma",
  "n14237148": "carcinoid",
  "n14237351": "carcinosarcoma",
  "n14237489": "celioma",
  "n14237561": "malignancy, malignance",
  "n14237818": "granulation, granulation tissue",
  "n14238041": "enchondroma",
  "n14238211": "fibroadenoma",
  "n14238393": "fibroid tumor, fibroid",
  "n14238528": "fibroma",
  "n14238639": "granuloma",
  "n14238820": "gumma",
  "n14238953": "hamartoma",
  "n14239097": "keratoacanthoma",
  "n14239322": "lipoma, adipose tumor",
  "n14239425": "malignant tumor, malignant neoplasm, metastatic tumor",
  "n14239743": "meningioma",
  "n14239918": "cancer, malignant neoplastic disease",
  "n14240262": "angiosarcoma",
  "n14240487": "chondrosarcoma",
  "n14240587": "Ewing's sarcoma, Ewing's tumor, Ewing's tumour, endothelial myeloma",
  "n14240775": "Kaposi's sarcoma",
  "n14241088": "leiomyosarcoma",
  "n14241241": "liposarcoma",
  "n14241319": "myosarcoma",
  "n14241400": "neurosarcoma, malignant neuroma",
  "n14241556": "osteosarcoma, osteogenic sarcoma",
  "n14241733": "lymphadenoma",
  "n14241825": "lymphadenopathy",
  "n14241968": "lymphoma",
  "n14242132": "Hodgkin's disease",
  "n14242337": "carcinoma",
  "n14242788": "cancroid, squamous cell carcinoma",
  "n14242922": "leukemia, leukaemia, leucaemia, cancer of the blood",
  "n14243268": "acute leukemia",
  "n14243393": "acute lymphocytic leukemia, acute lymphoblastic leukemia",
  "n14243649": "acute myelocytic leukemia, acute myeloid leukemia",
  "n14243877": "chronic leukemia",
  "n14244003": "chronic lymphocytic leukemia",
  "n14244160": "chronic myelocytic leukemia, myeloid leukemia",
  "n14244353": "lymphocytic leukemia",
  "n14244542": "lymphoblastic leukemia",
  "n14244726": "monocytic leukemia, monocytic leukaemia, monoblastic leukemia, monoblastic leukaemia, histiocytic leukemia, histiocytic leukaemia",
  "n14244991": "myeloblastic leukemia",
  "n14245163": "myelocytic leukemia, granulocytic leukemia",
  "n14245405": "rhabdomyosarcoma, rhabdosarcoma",
  "n14245594": "embryonal rhabdomyosarcoma, embryonal rhabdosarcoma",
  "n14245758": "alveolar rhabdomyosarcoma, alveolar rhabdosarcoma",
  "n14245928": "pleomorphic rhabdomyosarcoma, pleomorphic rhabdosarcoma",
  "n14246097": "Wilms' tumor, Wilms tumour, adenomyosarcoma, nephroblastoma, embryoma of the kidney",
  "n14246359": "sarcoma",
  "n14246710": "adenocarcinoma, glandular cancer, glandular carcinoma",
  "n14246899": "breast cancer",
  "n14247035": "carcinoma in situ, preinvasive cancer",
  "n14247239": "colon cancer",
  "n14247360": "embryonal carcinoma",
  "n14247458": "endometrial carcinoma, endometrial cancer",
  "n14247575": "hemangioma, haemangioma",
  "n14247743": "strawberry hemangioma, strawberry haemangioma",
  "n14247954": "lymphangioma",
  "n14248069": "spider angioma, spider nevus, vascular spider",
  "n14248256": "myeloma",
  "n14248422": "multiple myeloma",
  "n14248541": "myoma",
  "n14248669": "myxoma",
  "n14248787": "neurinoma",
  "n14248899": "neuroblastoma",
  "n14249037": "neuroepithelioma",
  "n14249138": "neurofibroma, neurilemoma",
  "n14249262": "neuroma",
  "n14249384": "leiomyoma",
  "n14249512": "rhabdomyoma",
  "n14249601": "osteoblastoma",
  "n14249785": "osteochondroma",
  "n14249937": "osteoma",
  "n14250081": "papilloma, villoma, papillary tumor, papillary tumour",
  "n14250232": "pheochromocytoma, phaeochromocytoma",
  "n14250433": "pinealoma",
  "n14250514": "plasmacytoma",
  "n14250622": "psammoma, sand tumor",
  "n14250850": "retinoblastoma",
  "n14251045": "teratoma",
  "n14251174": "hepatoma, malignant hepatoma, hepatocarcinoma, hepatocellular carcinoma",
  "n14251317": "lung cancer",
  "n14251435": "mesothelioma",
  "n14251616": "oat cell carcinoma, small cell carcinoma",
  "n14251845": "oral cancer",
  "n14251981": "pancreatic cancer",
  "n14252067": "prostate cancer, prostatic adenocarcinoma",
  "n14252184": "seminoma, testicular cancer",
  "n14252320": "skin cancer",
  "n14252464": "epithelioma",
  "n14252564": "melanoma, malignant melanoma",
  "n14252722": "trophoblastic cancer",
  "n14252864": "eye disease",
  "n14253124": "animal disease",
  "n14254102": "actinomycosis",
  "n14254397": "cervicofacial actinomycosis, lumpy jaw",
  "n14254570": "cataract",
  "n14254768": "macular edema",
  "n14254926": "cystoid macular edema",
  "n14255064": "drusen",
  "n14255234": "glaucoma",
  "n14255536": "acute glaucoma, closed-angle glaucoma, angle-closure glaucoma",
  "n14255848": "normal tension glaucoma",
  "n14256014": "cortical cataract",
  "n14256117": "nuclear cataract",
  "n14256227": "posterior subcapsular cataract",
  "n14256346": "chronic glaucoma, open-angle glaucoma",
  "n14256585": "keratonosus",
  "n14256668": "macular degeneration",
  "n14256889": "age-related macular degeneration, AMD",
  "n14257014": "retinopathy",
  "n14257147": "diabetic retinopathy",
  "n14257377": "trachoma",
  "n14257563": "leukoma, leucoma",
  "n14257687": "adenitis",
  "n14257779": "alveolitis",
  "n14257993": "alveolitis, dry socket",
  "n14258176": "angiitis",
  "n14258293": "aortitis",
  "n14258391": "rheumatic aortitis",
  "n14258512": "appendicitis",
  "n14258609": "arteritis",
  "n14258762": "periarteritis",
  "n14258865": "polyarteritis",
  "n14258957": "Takayasu's arteritis, pulseless disease",
  "n14259133": "temporal arteritis",
  "n14259320": "ophthalmia, ophthalmitis",
  "n14259431": "ophthalmia neonatorum",
  "n14259616": "thoracic actinomycosis",
  "n14259739": "abdominal actinomycosis",
  "n14259865": "farmer's lung, thresher's lung",
  "n14260044": "anaplasmosis",
  "n14260182": "anthrax, splenic fever",
  "n14260349": "aspergillosis, brooder pneumonia",
  "n14260550": "aspiration pneumonia",
  "n14260774": "bagassosis, bagascosis",
  "n14260898": "balanitis",
  "n14260991": "balanoposthitis",
  "n14261112": "bighead",
  "n14261244": "blepharitis",
  "n14261384": "bursitis",
  "n14261508": "brucellosis, contagious abortion, Bang's disease",
  "n14261717": "bluetongue",
  "n14261846": "bovine spongiform encephalitis, BSE, mad cow disease",
  "n14262045": "bull nose",
  "n14262153": "camelpox",
  "n14262336": "canine chorea, chorea",
  "n14262419": "catarrhal fever",
  "n14262585": "chronic wasting disease",
  "n14262740": "costiasis",
  "n14262882": "cowpox, vaccinia",
  "n14263089": "creeps",
  "n14263280": "hemorrhagic septicemia, pasteurellosis",
  "n14263440": "fistulous withers, fistula",
  "n14263562": "fowl cholera",
  "n14263753": "fowl pest",
  "n14263944": "hog cholera",
  "n14264042": "distemper",
  "n14264203": "canine distemper",
  "n14264352": "equine distemper, strangles",
  "n14264517": "enterotoxemia",
  "n14264664": "foot-and-mouth disease, hoof-and-mouth disease",
  "n14264858": "foot rot",
  "n14265006": "black disease, sheep rot, liver rot, distomatosis",
  "n14265205": "glanders",
  "n14265350": "heaves, broken wind",
  "n14265508": "Lyme disease, Lyme arthritis",
  "n14265722": "Marburg disease, Marburg hemorrhagic fever, green monkey disease",
  "n14265958": "albuminuria, proteinuria",
  "n14266168": "aminoaciduria",
  "n14266313": "ammoniuria",
  "n14266400": "hematocyturia, haematocyturia",
  "n14266521": "Jacquemier's sign",
  "n14266677": "Kayser-Fleischer ring",
  "n14266835": "keratomalacia",
  "n14267017": "Kernig's sign",
  "n14267205": "ketonemia, ketosis, acetonemia",
  "n14267358": "Koplik's spots",
  "n14267588": "fructosuria",
  "n14267722": "glucosuria",
  "n14267841": "glycosuria",
  "n14267994": "lymphuria",
  "n14268084": "monocytosis",
  "n14268221": "thrombocytosis",
  "n14268439": "ochronosis",
  "n14268623": "hypercalcemia, hypercalcaemia",
  "n14268871": "hypocalcemia, hypocalcaemia",
  "n14269090": "hypercalciuria, hypercalcinuria",
  "n14269319": "hypercholesterolemia, hypercholesteremia",
  "n14269556": "hyperkalemia",
  "n14269772": "hypokalemia",
  "n14270016": "kalemia",
  "n14270127": "kaliuresis, kaluresis",
  "n14270241": "natriuresis",
  "n14270380": "hyperlipoproteinemia",
  "n14270614": "hypolipoproteinemia",
  "n14270846": "hypoproteinemia",
  "n14271015": "hypernatremia",
  "n14271173": "hyponatremia",
  "n14271322": "hypersplenism",
  "n14271480": "ketonuria, ketoaciduria, acetonuria",
  "n14271667": "rabies, hydrophobia, lyssa, madness",
  "n14271920": "red water",
  "n14272023": "rhinotracheitis",
  "n14272141": "rinderpest, cattle plague",
  "n14272347": "scours",
  "n14272421": "scrapie",
  "n14272620": "shipping fever, shipping pneumonia",
  "n14272915": "spavin",
  "n14273084": "blood spavin",
  "n14273183": "bog spavin",
  "n14273277": "bone spavin",
  "n14273365": "swamp fever, leptospirosis",
  "n14273601": "canicola fever",
  "n14273753": "Weil's disease",
  "n14273860": "loco disease, locoism",
  "n14274068": "looping ill",
  "n14274217": "mange",
  "n14274451": "moon blindness, mooneye",
  "n14274592": "murrain",
  "n14274701": "myxomatosis",
  "n14274801": "Newcastle disease",
  "n14274905": "pip",
  "n14274975": "psittacosis, parrot disease",
  "n14275077": "pullorum disease, bacillary white diarrhea, bacillary white diarrhoea",
  "n14275240": "saddle sore, gall",
  "n14275405": "sand crack",
  "n14275560": "toe crack",
  "n14275661": "quarter crack",
  "n14275764": "staggers, blind staggers",
  "n14275978": "sweating sickness",
  "n14276081": "Texas fever",
  "n14276201": "trembles, milk sickness",
  "n14276360": "tularemia, tularaemia, rabbit fever, deer fly fever, yatobyo",
  "n14276649": "zoonosis, zoonotic disease",
  "n14276936": "plant disease",
  "n14277655": "rust",
  "n14277833": "blister rust, white-pine rust, white pine blister rust",
  "n14278102": "blackheart",
  "n14278219": "black knot",
  "n14278360": "black rot",
  "n14278500": "black spot",
  "n14278642": "bottom rot",
  "n14278773": "brown rot",
  "n14278953": "brown rot gummosis, gummosis",
  "n14279099": "gummosis",
  "n14279235": "ring rot, ring disease, tobacco wilt",
  "n14279417": "canker",
  "n14279543": "cotton ball",
  "n14279632": "crown gall",
  "n14279845": "hairy root",
  "n14280011": "crown wart",
  "n14280151": "damping off",
  "n14280298": "dieback",
  "n14280504": "dry rot",
  "n14280639": "heartrot",
  "n14280775": "mosaic",
  "n14280986": "potato mosaic",
  "n14281086": "tobacco mosaic",
  "n14281216": "tomato streak",
  "n14281397": "rhizoctinia disease",
  "n14281549": "little potato, rosette, russet scab, stem canker",
  "n14281678": "pink disease",
  "n14281851": "potato wart",
  "n14281992": "root rot",
  "n14282107": "scorch",
  "n14282258": "leaf scorch",
  "n14282384": "sweet-potato ring rot",
  "n14282524": "sclerotium disease, sclerotium rot",
  "n14282698": "Dutch elm disease",
  "n14282796": "ergot",
  "n14282890": "foot rot",
  "n14283002": "granville wilt",
  "n14283097": "pinkroot",
  "n14283178": "wilt, wilt disease",
  "n14283423": "fusarium wilt",
  "n14283525": "verticilliosis",
  "n14283632": "smut",
  "n14283909": "loose smut",
  "n14284026": "bunt, stinking smut",
  "n14284193": "flag smut",
  "n14284309": "green smut, false smut",
  "n14284450": "soft rot",
  "n14284581": "leak",
  "n14284688": "yellow dwarf",
  "n14284876": "yellow dwarf of potato, potato yellow dwarf",
  "n14285008": "onion yellow dwarf",
  "n14285113": "yellow spot",
  "n14285276": "trauma, psychic trauma",
  "n14285461": "birth trauma",
  "n14285662": "injury, hurt, harm, trauma",
  "n14286339": "raw wound",
  "n14286435": "stigmata",
  "n14286549": "abrasion, scratch, scrape, excoriation",
  "n14286885": "graze",
  "n14286995": "rope burn",
  "n14287113": "cut, gash, slash, slice",
  "n14287314": "laceration",
  "n14287408": "bite",
  "n14287567": "dog bite",
  "n14287647": "snakebite",
  "n14287741": "bee sting",
  "n14287823": "flea bite",
  "n14287904": "mosquito bite",
  "n14287995": "birth trauma",
  "n14288107": "blast trauma",
  "n14288235": "bleeding, hemorrhage, haemorrhage",
  "n14288561": "hemorrhagic stroke, haemorrhagic stroke",
  "n14288725": "blunt trauma",
  "n14288871": "bruise, contusion",
  "n14289079": "ecchymosis",
  "n14289193": "petechia",
  "n14289387": "shiner, black eye, mouse",
  "n14289504": "bump",
  "n14289590": "burn",
  "n14289829": "electric burn",
  "n14289942": "scorch, singe",
  "n14290107": "scald",
  "n14290212": "sedation",
  "n14290365": "sunburn, erythema solare",
  "n14290534": "tan, suntan, sunburn, burn",
  "n14290751": "windburn",
  "n14290881": "hyperpigmentation",
  "n14291010": "hypopigmentation",
  "n14291173": "first-degree burn",
  "n14291295": "second-degree burn",
  "n14291435": "third-degree burn",
  "n14291561": "dislocation",
  "n14291823": "electric shock",
  "n14292090": "fracture, break",
  "n14292571": "comminuted fracture",
  "n14292688": "complete fracture",
  "n14292796": "compound fracture, open fracture",
  "n14292944": "compression fracture",
  "n14293093": "depressed fracture",
  "n14293207": "displaced fracture",
  "n14293352": "fatigue fracture, stress fracture",
  "n14293505": "hairline fracture, capillary fracture",
  "n14293678": "greenstick fracture",
  "n14293840": "incomplete fracture",
  "n14293985": "impacted fracture",
  "n14294117": "simple fracture, closed fracture",
  "n14294271": "abarticulation",
  "n14294354": "diastasis",
  "n14294517": "spondylolisthesis",
  "n14294678": "frostbite, cryopathy",
  "n14294861": "intravasation",
  "n14294964": "penetrating trauma, penetrating injury",
  "n14295140": "pinch",
  "n14295248": "rupture",
  "n14295389": "hernia, herniation",
  "n14295691": "colpocele, vaginocele",
  "n14295829": "diverticulum",
  "n14295986": "Meckel's diverticulum",
  "n14296142": "eventration",
  "n14296254": "exomphalos",
  "n14296399": "hiatus hernia, hiatal hernia, diaphragmatic hernia",
  "n14296579": "herniated disc, ruptured intervertebral disc, slipped disc",
  "n14296802": "inguinal hernia",
  "n14296966": "cystocele, colpocystocele",
  "n14297150": "rectocele, proctocele",
  "n14297329": "keratocele",
  "n14297406": "laparocele",
  "n14297489": "umbilical hernia, omphalocele",
  "n14297696": "sleep disorder",
  "n14297870": "sting, bite, insect bite",
  "n14298102": "strain",
  "n14298247": "strangulation",
  "n14298420": "whiplash, whiplash injury",
  "n14298620": "wale, welt, weal, wheal",
  "n14298815": "wound, lesion",
  "n14299070": "wrench, twist, pull",
  "n14299336": "sprain",
  "n14299480": "trench foot, immersion foot",
  "n14299637": "symptom",
  "n14301785": "sign",
  "n14302005": "vital sign",
  "n14302261": "amenorrhea, amenorrhoea, amenia",
  "n14302460": "aura",
  "n14302652": "chloasma, melasma, mask of pregnancy",
  "n14302847": "clubbing",
  "n14303009": "cyanosis",
  "n14303249": "diuresis",
  "n14303448": "acrocyanosis, Raynaud's sign",
  "n14303647": "primary amenorrhea",
  "n14303743": "secondary amenorrhea",
  "n14303877": "prodrome, prodroma",
  "n14304060": "syndrome",
  "n14304722": "cervical disc syndrome, cervical root syndrome",
  "n14304964": "Chinese restaurant syndrome",
  "n14305149": "Conn's syndrome",
  "n14305458": "fetal alcohol syndrome, FAS",
  "n14305699": "Gulf War syndrome, Persian Gulf illness",
  "n14305990": "regional enteritis, regional ileitis, Crohn's disease",
  "n14306245": "irritable bowel syndrome, spastic colon, mucous colitis",
  "n14306475": "Klinefelter's syndrome, Klinefelter syndrome, XXY-syndrome",
  "n14306802": "ulcerative colitis",
  "n14307035": "malabsorption syndrome",
  "n14307275": "Munchausen's syndrome, Munchausen syndrome",
  "n14307468": "narcolepsy",
  "n14307724": "nephrotic syndrome, nephrosis",
  "n14308012": "Noonan's syndrome",
  "n14308168": "phantom limb syndrome",
  "n14308311": "premenstrual syndrome, PMS",
  "n14308472": "radiation sickness, radiation syndrome, radiation",
  "n14308945": "Ramsay Hunt syndrome",
  "n14309183": "Reiter's syndrome, Reiter's disease",
  "n14309394": "restless legs syndrome, restless legs, Ekbom syndrome",
  "n14309653": "Reye's syndrome",
  "n14309938": "scalenus syndrome",
  "n14310172": "neonatal death",
  "n14310292": "sudden infant death syndrome, SIDS, infant death, crib death, cot death",
  "n14310504": "tetany, tetanilla, intermittent tetanus, intermittent cramp, apyretic tetanus",
  "n14310830": "thoracic outlet syndrome",
  "n14310987": "Tietze's syndrome",
  "n14311117": "Tourette's syndrome, Gilles de la Tourette syndrome",
  "n14311348": "effect",
  "n14311574": "aftereffect",
  "n14311709": "bummer",
  "n14311801": "hairy tongue, furry tongue, black tongue",
  "n14311977": "side effect",
  "n14312172": "abscess",
  "n14312363": "abscessed tooth",
  "n14312481": "head",
  "n14312600": "purulence, purulency",
  "n14312776": "water blister",
  "n14312889": "blood blister",
  "n14313017": "exophthalmos",
  "n14313154": "festination",
  "n14313342": "furring",
  "n14313440": "gangrene, sphacelus, slough",
  "n14313661": "dry gangrene, cold gangrene, mumification necrosis, mummification",
  "n14313943": "gas gangrene, clostridial myonecrosis, emphysematous gangrene, emphysematous phlegmon, gangrenous emphysema, gas phlegmon, progressive emphysematous necrosis",
  "n14314320": "hematuria, haematuria",
  "n14314483": "hemoglobinuria, haemoglobinuria",
  "n14314597": "hemosiderosis, haemosiderosis",
  "n14314756": "nebula",
  "n14314850": "sneeze, sneezing, sternutation",
  "n14315071": "enlargement",
  "n14315192": "swelling, puffiness, lump",
  "n14315549": "bloat",
  "n14315722": "bubo",
  "n14315897": "anasarca",
  "n14316048": "chemosis",
  "n14316164": "papilledema",
  "n14316352": "bunion",
  "n14316470": "palsy",
  "n14316583": "pyuria",
  "n14316714": "edema, oedema, hydrops, dropsy",
  "n14317064": "cerebral edema, brain edema",
  "n14317221": "hematocele, haematocele, hematocoele, haematocoele",
  "n14317433": "hematocolpometra, haematocolpometra",
  "n14317562": "hematocolpos, haematocolpos",
  "n14317720": "intumescence, intumescency",
  "n14317943": "iridoncus",
  "n14318030": "lymphogranuloma",
  "n14318116": "oscheocele, oscheocoele",
  "n14318210": "tumescence",
  "n14318360": "tumidity, tumidness",
  "n14318552": "cephalhematoma, cephalohematoma",
  "n14318714": "hematoma, haematoma",
  "n14318819": "proud flesh",
  "n14318927": "hyperbilirubinemia",
  "n14319093": "hyperbilirubinemia of the newborn, neonatal hyperbilirubinemia",
  "n14319299": "hyperglycemia, hyperglycaemia",
  "n14319454": "hypoglycemia, hypoglycaemia",
  "n14319684": "jaundice, icterus",
  "n14320016": "jaundice of the newborn, physiological jaundice of the newborn, icterus neonatorum",
  "n14320218": "kernicterus",
  "n14320394": "congestion",
  "n14320602": "hydrothorax",
  "n14320813": "hemothorax, haemothorax",
  "n14320984": "hyperemia, hyperaemia",
  "n14321135": "engorgement",
  "n14321243": "pulmonary congestion",
  "n14321333": "stuffiness",
  "n14321469": "eruption",
  "n14321653": "enanthem, enanthema",
  "n14321814": "exanthem, exanthema, skin eruption",
  "n14321953": "rash, roseola, efflorescence, skin rash",
  "n14322106": "prickly heat, heat rash, miliaria",
  "n14322248": "urtication, urticaria, hives, nettle rash",
  "n14322528": "numbness",
  "n14322699": "pain, hurting",
  "n14323683": "ache, aching",
  "n14323974": "toothache, odontalgia",
  "n14324099": "aerodontalgia",
  "n14324274": "agony, suffering, excruciation",
  "n14324572": "arthralgia",
  "n14324672": "throe",
  "n14324795": "paresthesia, paraesthesia",
  "n14325006": "formication",
  "n14325220": "Passion, Passion of Christ",
  "n14325335": "backache",
  "n14325437": "burn, burning",
  "n14325575": "causalgia",
  "n14325732": "colic, intestinal colic, gripes, griping",
  "n14325902": "chest pain",
  "n14325976": "chiralgia",
  "n14326072": "distress",
  "n14326190": "dysmenorrhea",
  "n14326305": "primary dysmenorrhea",
  "n14326458": "secondary dysmenorrhea",
  "n14326607": "headache, head ache, cephalalgia",
  "n14326880": "glossalgia, glossodynia",
  "n14326969": "growing pains",
  "n14327125": "hemorrhoid, haemorrhoid, piles",
  "n14327266": "stomachache, stomach ache, bellyache, gastralgia",
  "n14327435": "earache, otalgia",
  "n14327543": "histamine headache, cluster headache",
  "n14327707": "migraine, megrim, sick headache, hemicrania",
  "n14327878": "sick headache",
  "n14327970": "sinus headache",
  "n14328105": "tension headache",
  "n14328290": "lumbago, lumbar pain",
  "n14328513": "keratalgia",
  "n14328588": "labor pain",
  "n14328720": "mastalgia",
  "n14328794": "melagra",
  "n14328894": "meralgia",
  "n14328966": "metralgia",
  "n14329040": "myalgia, myodynia",
  "n14329159": "nephralgia",
  "n14329262": "neuralgia, neuralgy",
  "n14329460": "odynophagia",
  "n14329578": "orchidalgia",
  "n14329654": "pang",
  "n14329762": "pang, sting",
  "n14329888": "photalgia, photophobia",
  "n14330046": "pleurodynia, pleuralgia, costalgia",
  "n14330201": "podalgia",
  "n14330265": "proctalgia",
  "n14330340": "epidemic pleurodynia, epidemic myalgia, myosis, diaphragmatic pleurisy, Bornholm disease",
  "n14330588": "trigeminal neuralgia, tic douloureux",
  "n14330727": "sciatica",
  "n14330833": "birth pangs, labor pains, labour pains",
  "n14330992": "afterpains",
  "n14331137": "palilalia",
  "n14331271": "palmature",
  "n14331373": "referred pain",
  "n14331690": "renal colic",
  "n14331873": "smart, smarting, smartness",
  "n14332085": "sting, stinging",
  "n14332339": "stitch",
  "n14332448": "rebound tenderness",
  "n14332617": "tenderness, soreness, rawness",
  "n14333058": "thermalgesia",
  "n14333136": "chafe",
  "n14333334": "throb",
  "n14333433": "torture, torment",
  "n14333575": "ulalgia",
  "n14333645": "urodynia",
  "n14333721": "postnasal drip",
  "n14333863": "papule",
  "n14334011": "papulovesicle, vesicopapule",
  "n14334122": "pustule",
  "n14334306": "pimple, hickey, zit",
  "n14334511": "pock",
  "n14334631": "cardiomegaly, megalocardia, megacardia, enlarged heart",
  "n14334814": "heart murmur, cardiac murmur, murmur",
  "n14335006": "systolic murmur",
  "n14335097": "palpitation",
  "n14335223": "heartburn, pyrosis",
  "n14335485": "gastroesophageal reflux, esophageal reflux, oesophageal reflux",
  "n14335641": "hepatojugular reflux",
  "n14335761": "ureterorenal reflux",
  "n14335884": "vesicoureteral reflux",
  "n14336004": "reflux",
  "n14336169": "hot flash, flush",
  "n14336317": "indigestion, dyspepsia, stomach upset, upset stomach",
  "n14336539": "inflammation, redness, rubor",
  "n14338848": "amyxia",
  "n14338942": "carditis",
  "n14339094": "endocarditis",
  "n14339219": "subacute bacterial endocarditis",
  "n14339387": "myocardial inflammation, myocarditis",
  "n14339536": "pancarditis",
  "n14339682": "pericarditis",
  "n14339772": "catarrh",
  "n14339914": "cellulitis",
  "n14340091": "cervicitis",
  "n14340182": "cheilitis",
  "n14340287": "chill, shivering",
  "n14340462": "ague, chills and fever",
  "n14340635": "quartan",
  "n14340734": "cholangitis",
  "n14340822": "cholecystitis",
  "n14340914": "chorditis",
  "n14341001": "chorditis",
  "n14341091": "colitis, inflammatory bowel disease",
  "n14341253": "colpitis",
  "n14341334": "colpocystitis",
  "n14341432": "conjunctivitis, pinkeye",
  "n14341563": "corditis",
  "n14341652": "costochondritis",
  "n14341769": "dacryocystitis",
  "n14341923": "diverticulitis",
  "n14342132": "encephalitis, cephalitis, phrenitis",
  "n14342533": "encephalomyelitis",
  "n14342638": "endarteritis",
  "n14342742": "acute hemorrhagic encephalitis",
  "n14342882": "equine encephalitis, equine encephalomyelitis",
  "n14343062": "herpes simplex encephalitis, herpes encephalitis, acute inclusion body encephalitis",
  "n14343302": "leukoencephalitis",
  "n14343411": "meningoencephalitis, cerebromeningitis, encephalomeningitis",
  "n14343597": "panencephalitis",
  "n14343735": "sleeping sickness, sleepy sickness, epidemic encephalitis, lethargic encephalitis, encephalitis lethargica",
  "n14344033": "West Nile encephalitis",
  "n14344189": "subacute sclerosing panencephalitis, SSPE, inclusion body encephalitis, subacute inclusion body encephalitis, sclerosing leukoencephalitis, subacute sclerosing leukoencephalitis, Bosin's disease, Dawson's encephalitis, Van Bogaert encephalitis",
  "n14344720": "rubella panencephalitis",
  "n14344881": "endocervicitis",
  "n14344997": "enteritis",
  "n14345168": "necrotizing enteritis",
  "n14345304": "epicondylitis",
  "n14345455": "epididymitis",
  "n14345552": "epiglottitis",
  "n14345719": "episcleritis",
  "n14345815": "esophagitis, oesophagitis",
  "n14345958": "fibrositis",
  "n14346080": "fibromyositis",
  "n14346190": "folliculitis",
  "n14346280": "funiculitis",
  "n14346416": "gastritis",
  "n14346606": "acute gastritis",
  "n14346731": "chronic gastritis",
  "n14346909": "glossitis",
  "n14347045": "acute glossitis",
  "n14347190": "chronic glossitis",
  "n14347334": "Moeller's glossitis, glossodynia exfoliativa",
  "n14347540": "hydrarthrosis",
  "n14347695": "ileitis",
  "n14347774": "iridocyclitis",
  "n14347886": "iridokeratitis",
  "n14347993": "iritis",
  "n14348070": "jejunitis",
  "n14348176": "jejunoileitis",
  "n14348300": "keratitis",
  "n14348429": "keratoconjunctivitis",
  "n14348538": "keratoiritis",
  "n14348647": "keratoscleritis",
  "n14348757": "labyrinthitis, otitis interna",
  "n14348895": "laminitis, founder",
  "n14349043": "laryngitis",
  "n14349208": "laryngopharyngitis",
  "n14349311": "laryngotracheobronchitis",
  "n14349446": "leptomeningitis",
  "n14349541": "lymphadenitis",
  "n14349628": "lymphangitis",
  "n14349717": "mastitis",
  "n14349807": "mastoiditis",
  "n14349892": "metritis, endometritis",
  "n14350023": "monoplegia",
  "n14350106": "myelatelia",
  "n14350206": "myelitis",
  "n14350292": "myositis",
  "n14350448": "myometritis",
  "n14350536": "trichinosis, trichiniasis, myositis trichinosa",
  "n14350837": "neuritis",
  "n14351002": "oophoritis",
  "n14351094": "orchitis",
  "n14351219": "ophthalmoplegia",
  "n14351321": "osteitis",
  "n14351490": "osteomyelitis",
  "n14351629": "otitis",
  "n14351759": "otitis externa",
  "n14351886": "otitis media",
  "n14351995": "ovaritis",
  "n14352077": "otorrhea",
  "n14352163": "ozena, ozaena",
  "n14352334": "pancreatitis",
  "n14352455": "parametritis",
  "n14352570": "parotitis",
  "n14352687": "peritonitis, peritoneal inflammation",
  "n14352801": "phalangitis",
  "n14352890": "phlebitis",
  "n14353008": "phlebothrombosis, venous thrombosis",
  "n14353314": "polyneuritis, multiple neuritis",
  "n14353479": "retrobulbar neuritis",
  "n14353622": "Guillain-Barre syndrome, infectious polyneuritis, Landry's paralysis",
  "n14353862": "thrombophlebitis",
  "n14353995": "pneumonitis",
  "n14354121": "posthitis",
  "n14354257": "proctitis",
  "n14354462": "prostatitis",
  "n14354642": "rachitis",
  "n14354751": "radiculitis",
  "n14354847": "chorioretinitis",
  "n14354960": "retinitis",
  "n14355060": "rhinitis, coryza",
  "n14355238": "sinusitis",
  "n14355356": "pansinusitis",
  "n14355459": "salpingitis",
  "n14355643": "scleritis",
  "n14355725": "sialadenitis",
  "n14355819": "splenitis",
  "n14355901": "spondylitis",
  "n14356044": "stomatitis",
  "n14356167": "vesicular stomatitis",
  "n14356328": "synovitis",
  "n14356497": "tarsitis",
  "n14356578": "tendinitis, tendonitis, tenonitis",
  "n14356720": "tennis elbow, lateral epicondylitis, lateral humeral epicondylitis",
  "n14356993": "tenosynovitis, tendosynovitis, tendonous synovitis",
  "n14357160": "thyroiditis",
  "n14357251": "tonsillitis",
  "n14357370": "tracheitis",
  "n14357454": "tracheobronchitis",
  "n14357607": "tympanitis",
  "n14357693": "ulitis",
  "n14357770": "ureteritis",
  "n14357853": "uveitis",
  "n14357942": "uvulitis",
  "n14358022": "vaccinia, vaccina, variola vaccine, variola vaccinia, variola vaccina",
  "n14358335": "vaginitis",
  "n14358474": "valvulitis",
  "n14358622": "vasculitis",
  "n14358709": "vasovesiculitis",
  "n14358860": "vesiculitis",
  "n14358993": "vulvitis",
  "n14359073": "vulvovaginitis",
  "n14359174": "cough, coughing",
  "n14359459": "hiccup, hiccough, singultus",
  "n14359816": "meningism",
  "n14359952": "nausea, sickness",
  "n14360156": "morning sickness",
  "n14360320": "queasiness, squeamishness, qualm",
  "n14360459": "spasm, cramp, muscle spasm",
  "n14360742": "charley horse, charley-horse",
  "n14360915": "writer's cramp, graphospasm",
  "n14361064": "blepharospasm",
  "n14361182": "crick, kink, rick, wrick",
  "n14361415": "myoclonus",
  "n14361513": "opisthotonos",
  "n14361664": "twitch, twitching, vellication",
  "n14361972": "tic",
  "n14362075": "blepharism",
  "n14362179": "fibrillation",
  "n14362373": "atrial fibrillation",
  "n14362510": "bradycardia",
  "n14362593": "heart block, Adams-Stokes syndrome, Stokes-Adams syndrome, atrioventricular block",
  "n14362841": "premature ventricular contraction, PVC",
  "n14363027": "tachycardia",
  "n14363139": "ventricular fibrillation",
  "n14363367": "fasciculation",
  "n14363483": "scar, cicatrix, cicatrice",
  "n14363785": "callus",
  "n14363913": "keloid, cheloid",
  "n14364065": "pockmark",
  "n14364217": "sword-cut",
  "n14364306": "vaccination",
  "n14364432": "hardening",
  "n14364566": "callosity, callus",
  "n14364802": "corn, clavus",
  "n14364980": "calcification",
  "n14365119": "musca volitans, muscae volitantes, floater, spots",
  "n14365356": "fever, febrility, febricity, pyrexia, feverishness",
  "n14365619": "hyperpyrexia",
  "n14365741": "atrophy, wasting, wasting away",
  "n14365950": "dysplasia",
  "n14366225": "fibrous dysplasia of bone",
  "n14366486": "Albright's disease, polyostotic fibrous dysplasia",
  "n14366633": "monostotic fibrous dysplasia",
  "n14366759": "hypertrophy",
  "n14367005": "adenomegaly",
  "n14367080": "cor pulmonale",
  "n14367249": "dactylomegaly",
  "n14367341": "elephantiasis",
  "n14367556": "elephantiasis neuromatosa",
  "n14367649": "elephantiasis scroti, chyloderma",
  "n14367797": "nevoid elephantiasis, pachyderma",
  "n14368032": "filariasis",
  "n14368192": "splenomegaly",
  "n14368288": "giantism, gigantism, overgrowth",
  "n14368483": "acromegaly, acromegalia",
  "n14368805": "hyperplasia",
  "n14368917": "benign prostatic hyperplasia, BPH",
  "n14369064": "hypoplasia",
  "n14369194": "anaplasia",
  "n14369408": "apnea",
  "n14369530": "periodic apnea of the newborn",
  "n14369744": "dyspnea, dyspnoea",
  "n14369913": "orthopnea",
  "n14370122": "shortness of breath, SOB, breathlessness",
  "n14370267": "sleep apnea",
  "n14370391": "cerebral hemorrhage",
  "n14370526": "blood extravasation",
  "n14370726": "hyphema",
  "n14370825": "metrorrhagia",
  "n14370992": "nosebleed, epistaxis",
  "n14371082": "ulemorrhagia",
  "n14371161": "constipation, irregularity",
  "n14371449": "dyschezia",
  "n14371620": "fecal impaction",
  "n14371770": "obstipation",
  "n14371913": "diarrhea, diarrhoea, looseness of the bowels, looseness",
  "n14372286": "the shits, the trots",
  "n14372399": "Montezuma's revenge",
  "n14372513": "dizziness, giddiness, lightheadedness, vertigo",
  "n14372738": "anemia, anaemia",
  "n14372855": "wheeziness",
  "n14373057": "withdrawal symptom",
  "n14373246": "thrombus",
  "n14373407": "embolus",
  "n14373582": "psychological state, psychological condition, mental state, mental condition",
  "n14374273": "morale",
  "n14374432": "anxiety, anxiousness",
  "n14374764": "castration anxiety",
  "n14375005": "hypochondria, hypochondriasis",
  "n14375166": "overanxiety",
  "n14375241": "hallucinosis",
  "n14375363": "identity crisis",
  "n14375576": "nervousness, nerves",
  "n14375761": "jitters, heebie-jeebies, screaming meemies",
  "n14375890": "strain, mental strain, nervous strain",
  "n14376188": "tension, tenseness, stress",
  "n14376497": "yips",
  "n14376695": "breaking point",
  "n14376855": "delusion, psychotic belief",
  "n14377177": "delusions of grandeur",
  "n14377375": "delusions of persecution",
  "n14377617": "hallucination",
  "n14377830": "auditory hallucination, acousma",
  "n14377965": "chromatism",
  "n14378064": "pink elephants",
  "n14378177": "pseudohallucination",
  "n14378311": "trip",
  "n14378439": "visual hallucination",
  "n14378586": "zoopsia",
  "n14378712": "nihilistic delusion, nihilism",
  "n14378897": "somatic delusion",
  "n14379017": "zoanthropy",
  "n14379130": "mental health",
  "n14379360": "mental soundness, mental balance",
  "n14379501": "sanity, saneness",
  "n14379703": "lucidity",
  "n14379829": "rationality, reason, reasonableness",
  "n14380140": "mental illness, mental disease, psychopathy",
  "n14380473": "anxiety disorder",
  "n14380717": "generalized anxiety disorder, GAD, anxiety reaction",
  "n14381017": "obsessive-compulsive disorder",
  "n14381249": "panic disorder",
  "n14381416": "phobia, phobic disorder, phobic neurosis",
  "n14381732": "acarophobia",
  "n14381840": "agoraphobia",
  "n14381997": "androphobia",
  "n14382075": "arachnophobia",
  "n14382159": "gynophobia",
  "n14382238": "simple phobia",
  "n14382766": "acrophobia",
  "n14382871": "algophobia",
  "n14382967": "aquaphobia",
  "n14383067": "astraphobia",
  "n14383163": "automysophobia",
  "n14383252": "claustrophobia",
  "n14383362": "cryophobia",
  "n14383444": "cyberphobia",
  "n14383544": "hydrophobia",
  "n14383624": "hydrophobia",
  "n14383759": "hypnophobia",
  "n14383848": "mysophobia",
  "n14383961": "neophobia",
  "n14384041": "nyctophobia",
  "n14384133": "phobophobia",
  "n14384227": "phonophobia, acousticophobia",
  "n14384351": "photophobia",
  "n14384431": "pyrophobia",
  "n14384509": "taphephobia",
  "n14384602": "thanatophobia",
  "n14384684": "triskaidekaphobia",
  "n14384796": "zoophobia",
  "n14385002": "ailurophobia",
  "n14385082": "cynophobia",
  "n14385160": "entomophobia",
  "n14385243": "lepidophobia",
  "n14385325": "musophobia",
  "n14385403": "social phobia",
  "n14385734": "satanophobia",
  "n14385818": "school phobia",
  "n14385919": "traumatophobia",
  "n14386022": "xenophobia",
  "n14386130": "posttraumatic stress disorder, PTSD",
  "n14386475": "psychosomatic disorder",
  "n14386590": "aberration",
  "n14386697": "conversion disorder, conversion reaction, conversion hysteria",
  "n14386968": "glove anesthesia",
  "n14387202": "delirium",
  "n14387349": "delusional disorder",
  "n14387477": "encopresis",
  "n14387604": "folie a deux",
  "n14387807": "personality disorder",
  "n14387975": "maladjustment",
  "n14388139": "antisocial personality disorder, sociopathic personality, psychopathic personality",
  "n14388596": "battle fatigue, combat fatigue, combat neurosis, shell shock",
  "n14388758": "schizotypal personality, schizoid",
  "n14388910": "affective disorder, major affective disorder, emotional disorder, emotional disturbance",
  "n14389240": "depressive disorder, clinical depression, depression",
  "n14389584": "agitated depression",
  "n14389739": "anaclitic depression",
  "n14389909": "dysthymia, dysthymic depression",
  "n14390109": "endogenous depression",
  "n14390249": "exogenous depression, reactive depression",
  "n14390466": "major depressive episode",
  "n14390839": "involutional depression",
  "n14390967": "neurotic depression",
  "n14391112": "psychotic depression",
  "n14391319": "retarded depression",
  "n14391480": "unipolar depression",
  "n14391660": "mania, manic disorder",
  "n14391876": "craze, delirium, frenzy, fury, hysteria",
  "n14392143": "mass hysteria, epidemic hysertia",
  "n14392318": "megalomania",
  "n14392491": "melancholia",
  "n14392639": "bipolar disorder, manic depression, manic depressive illness, manic-depressive psychosis",
  "n14392862": "cyclothymia, cyclothymic disorder, cyclic disorder",
  "n14393033": "schizothymia",
  "n14393161": "neurosis, neuroticism, psychoneurosis",
  "n14393438": "hysteria, hysterical neurosis",
  "n14393716": "anxiety hysteria",
  "n14393862": "hysterocatalepsy",
  "n14393958": "anxiety neurosis",
  "n14394094": "depersonalization, depersonalisation, depersonalization disorder, depersonalisation disorder, depersonalization neurosis, depersonalisation neurosis",
  "n14394479": "fugue, psychogenic fugue",
  "n14394783": "split personality, multiple personality",
  "n14395018": "insanity",
  "n14395240": "lunacy, madness, insaneness",
  "n14395403": "dementia, dementedness",
  "n14395597": "alcoholic dementia, alcohol amnestic disorder, Korsakoff's psychosis, Korsakoff's syndrome, Korsakov's psychosis, Korsakov's syndrome, polyneuritic psychosis",
  "n14395955": "presenile dementia",
  "n14396096": "Alzheimer's disease, Alzheimer's, Alzheimers",
  "n14396431": "Pick's disease",
  "n14396707": "senile dementia, senile psychosis",
  "n14396890": "rhinopathy",
  "n14397040": "rhinophyma, hypertrophic rosacea, toper's nose, brandy nose, rum nose, rum-blossom, potato nose, hammer nose, copper nose",
  "n14397372": "Wernicke's encephalopathy",
  "n14397563": "irrationality, unreason",
  "n14397714": "derangement, mental unsoundness, unbalance",
  "n14397889": "craziness, daftness, flakiness",
  "n14398067": "psychosis",
  "n14398279": "delirium tremens, DTs",
  "n14398390": "paranoia",
  "n14398523": "schizophrenia, schizophrenic disorder, schizophrenic psychosis, dementia praecox",
  "n14398929": "borderline schizophrenia, latent schizophrenia",
  "n14399116": "catatonic schizophrenia, catatonic type schizophrenia, catatonia",
  "n14399438": "hebephrenia, hebephrenic schizophrenia, disorganized schizophrenia, disorganized type schizophrenia",
  "n14399852": "paranoid schizophrenia, paranoic type schizophrenia, paraphrenic schizophrenia, paraphrenia",
  "n14400325": "acute schizophrenic episode, reactive schizophrenia",
  "n14400507": "aphonia, voicelessness",
  "n14400677": "speech disorder, speech defect, defect of speech",
  "n14400979": "sprue, tropical sprue, psilosis",
  "n14401240": "flaccid bladder",
  "n14401512": "neurogenic bladder",
  "n14401644": "spastic bladder",
  "n14401878": "cataphasia",
  "n14402015": "dysarthria",
  "n14402184": "dyslogia",
  "n14402377": "dysphonia",
  "n14402488": "lallation",
  "n14402625": "lambdacism",
  "n14402763": "lisp",
  "n14402922": "stammer, stutter",
  "n14403107": "agitation",
  "n14403282": "disturbance, perturbation, upset",
  "n14403560": "fret, stew, sweat, lather, swither",
  "n14403772": "dither, pother, fuss, tizzy, flap",
  "n14404042": "tailspin",
  "n14404160": "depression",
  "n14404460": "blues, blue devils, megrims, vapors, vapours",
  "n14404610": "funk, blue funk",
  "n14404722": "melancholy",
  "n14404831": "slough of despond",
  "n14404941": "low spirits",
  "n14405061": "dumps, mopes",
  "n14405225": "elation",
  "n14405452": "high",
  "n14405621": "high",
  "n14405774": "cold sweat",
  "n14405931": "panic, scare",
  "n14406209": "red scare",
  "n14406303": "fit, tantrum, scene, conniption",
  "n14406465": "areflexia",
  "n14406573": "irritation, annoyance, vexation, botheration",
  "n14406900": "bummer",
  "n14407070": "huff, miff, seeing red",
  "n14407211": "pinprick",
  "n14407283": "restlessness, impatience",
  "n14407435": "snit",
  "n14407536": "enchantment, spell, trance",
  "n14407795": "possession",
  "n14407899": "fascination, captivation",
  "n14408086": "difficulty",
  "n14408519": "bitch",
  "n14408646": "predicament, quandary, plight",
  "n14408951": "corner, box",
  "n14409137": "hot water",
  "n14409285": "rattrap",
  "n14409371": "pinch",
  "n14409489": "fix, hole, jam, mess, muddle, pickle, kettle of fish",
  "n14409718": "dog's breakfast, dog's dinner",
  "n14409880": "hard time, rough sledding",
  "n14410089": "stress, strain",
  "n14410365": "mire",
  "n14410605": "problem, job",
  "n14410918": "race problem",
  "n14411079": "balance-of-payments problem",
  "n14411243": "situation",
  "n14411772": "quicksand",
  "n14411884": "vogue",
  "n14411981": "recognition, acknowledgment, acknowledgement",
  "n14412374": "approval, favorable reception, favourable reception",
  "n14412564": "appro",
  "n14412725": "acceptation",
  "n14412882": "contentedness, content",
  "n14413162": "acquiescence",
  "n14413265": "welcome",
  "n14413411": "apostasy, renunciation, defection",
  "n14413644": "disfavor, disfavour",
  "n14413831": "wilderness",
  "n14413993": "excommunication, exclusion, censure",
  "n14414146": "reprobation",
  "n14414294": "separation",
  "n14414503": "discreteness, distinctness, separateness, severalty",
  "n14414715": "isolation",
  "n14414980": "solitude",
  "n14415072": "solitude, purdah",
  "n14415163": "loneliness, solitariness",
  "n14415335": "quarantine",
  "n14415518": "insulation, insularity, insularism, detachment",
  "n14415773": "alienation, estrangement",
  "n14415898": "anomie, anomy",
  "n14416089": "privacy, privateness, secrecy, concealment",
  "n14416349": "hiddenness, covertness",
  "n14416473": "bosom",
  "n14416668": "confidentiality",
  "n14416845": "dissociation, disassociation",
  "n14417146": "compartmentalization, compartmentalisation",
  "n14417300": "dissociative disorder",
  "n14417551": "discontinuity",
  "n14417697": "disjunction, disjuncture, disconnection, disconnectedness",
  "n14417993": "separability",
  "n14418103": "incoherence, incoherency",
  "n14418290": "disjointedness",
  "n14418395": "union, unification",
  "n14418662": "coalition, fusion",
  "n14418822": "alliance, confederation",
  "n14418970": "federalization, federalisation",
  "n14419164": "connection, link, connectedness",
  "n14419510": "contact",
  "n14419737": "concatenation",
  "n14419889": "osculation",
  "n14420074": "tangency",
  "n14420240": "interconnection, interconnectedness",
  "n14420464": "coherence, coherency, cohesion, cohesiveness",
  "n14420780": "consistency",
  "n14420954": "junction, conjunction, conjugation, colligation",
  "n14421139": "association",
  "n14421373": "disassociation",
  "n14421585": "marriage",
  "n14421724": "syncretism",
  "n14422035": "continuity",
  "n14422179": "improvement, melioration",
  "n14422488": "decline, declination",
  "n14422751": "betterment",
  "n14422871": "development",
  "n14423268": "underdevelopment",
  "n14423428": "neglect, disuse",
  "n14423614": "omission",
  "n14423740": "twilight",
  "n14423870": "wreck",
  "n14424087": "reformation",
  "n14424384": "counterreformation",
  "n14424517": "renovation, restoration, refurbishment",
  "n14424780": "maturity, matureness",
  "n14425103": "adulthood",
  "n14425319": "manhood",
  "n14425414": "parenthood, parentage",
  "n14425601": "ripeness",
  "n14425715": "womanhood, muliebrity",
  "n14425853": "youth",
  "n14425974": "immaturity, immatureness",
  "n14426325": "post-maturity, post-menopause",
  "n14426449": "greenness",
  "n14426568": "callowness, jejuneness, juvenility",
  "n14426736": "prematureness, prematurity",
  "n14426910": "adolescence",
  "n14427065": "childhood, puerility",
  "n14427239": "infancy, babyhood",
  "n14427408": "embrace",
  "n14427537": "encompassment",
  "n14427633": "banishment, ostracism, Coventry",
  "n14427991": "debarment",
  "n14428160": "grade, level, tier",
  "n14428404": "biosafety level",
  "n14428655": "biosafety level 1",
  "n14428796": "biosafety level 2",
  "n14428970": "biosafety level 3",
  "n14429150": "biosafety level 4",
  "n14429382": "rating",
  "n14429484": "ranking",
  "n14429608": "gradation, step",
  "n14429885": "cut",
  "n14429985": "rank",
  "n14430916": "second class",
  "n14431015": "A level",
  "n14431169": "General Certificate of Secondary Education, GCSE, O level",
  "n14431338": "college level",
  "n14431471": "military rank, military rating, paygrade, rating",
  "n14431637": "flag rank",
  "n14431738": "caste",
  "n14431902": "dignity",
  "n14432032": "nobility, noblesse",
  "n14432167": "ordination",
  "n14432271": "purple",
  "n14432389": "pedestal",
  "n14432527": "archidiaconate",
  "n14432623": "baronetcy, barony",
  "n14432744": "dukedom",
  "n14432875": "earldom",
  "n14433001": "kingship",
  "n14433115": "princedom",
  "n14433232": "viscountcy, viscounty",
  "n14433361": "leadership",
  "n14433505": "ennoblement",
  "n14433587": "prominence",
  "n14433769": "limelight, spotlight, glare, public eye",
  "n14434022": "salience, saliency, strikingness",
  "n14434219": "conspicuousness",
  "n14434329": "visibility, profile",
  "n14434522": "low profile",
  "n14434681": "importance, grandness",
  "n14434866": "emphasis, accent",
  "n14435187": "stress, focus",
  "n14435353": "primacy",
  "n14435445": "eminence, distinction, preeminence, note",
  "n14435670": "king",
  "n14435809": "prestige, prestigiousness",
  "n14436029": "obscurity",
  "n14436285": "anonymity, namelessness",
  "n14436438": "humbleness, unimportance, obscureness, lowliness",
  "n14436667": "nowhere",
  "n14436769": "oblivion, limbo",
  "n14436875": "honor, honour, laurels",
  "n14437134": "glory, glorification",
  "n14437386": "fame, celebrity, renown",
  "n14437552": "esteem, regard, respect",
  "n14437845": "disesteem",
  "n14437976": "stature",
  "n14438125": "repute, reputation",
  "n14438419": "black eye",
  "n14438541": "stock",
  "n14438693": "character",
  "n14438788": "name",
  "n14438898": "fame",
  "n14439012": "infamy",
  "n14439149": "notoriety, ill fame",
  "n14439294": "reputation",
  "n14439447": "dishonor, dishonour",
  "n14439745": "disrepute, discredit",
  "n14440035": "corruptness",
  "n14440137": "shame, disgrace, ignominy",
  "n14440488": "humiliation",
  "n14440623": "abasement, degradation, abjection",
  "n14440875": "degeneracy, degeneration, decadence, decadency",
  "n14441083": "depth",
  "n14441230": "infamy, opprobrium",
  "n14441498": "obloquy, opprobrium",
  "n14441610": "odium",
  "n14441714": "reproach",
  "n14441825": "dominance, ascendance, ascendence, ascendancy, ascendency, control",
  "n14442361": "ascendant, ascendent",
  "n14442530": "domination, mastery, supremacy",
  "n14442749": "predominance, predomination, prepotency",
  "n14442933": "dominion, rule",
  "n14443228": "paramountcy",
  "n14443345": "raj",
  "n14443434": "regulation",
  "n14443532": "reign, sovereignty",
  "n14443676": "scepter, sceptre",
  "n14443786": "suzerainty",
  "n14443912": "absolutism, tyranny, despotism",
  "n14444114": "monopoly",
  "n14444326": "monopoly",
  "n14444644": "monopsony",
  "n14444825": "oligopoly",
  "n14445072": "corner",
  "n14445226": "bane, curse, scourge, nemesis",
  "n14445379": "comfort, comfortableness",
  "n14445749": "relief, ease",
  "n14446034": "reprieve, respite",
  "n14446161": "solace, solacement",
  "n14446298": "coziness, cosiness, snugness",
  "n14446493": "convenience",
  "n14446652": "discomfort, uncomfortableness",
  "n14446878": "inconvenience, incommodiousness",
  "n14447019": "malaise, unease, uneasiness",
  "n14447165": "hangover, katzenjammer",
  "n14447303": "wretchedness",
  "n14447525": "wellbeing, well-being, welfare, upbeat, eudaemonia, eudaimonia",
  "n14447816": "fool's paradise",
  "n14447908": "health, wellness",
  "n14448200": "ill-being",
  "n14448333": "misery, wretchedness, miserableness",
  "n14448692": "concentration camp",
  "n14448825": "living death",
  "n14448910": "suffering, woe",
  "n14449022": "anguish",
  "n14449126": "need, demand",
  "n14449405": "lack, deficiency, want",
  "n14449865": "dearth, famine, shortage",
  "n14449960": "deficit",
  "n14450172": "mineral deficiency",
  "n14450339": "shortness",
  "n14450540": "stringency, tightness",
  "n14450691": "necessity",
  "n14450904": "requisiteness",
  "n14451020": "urgency",
  "n14451189": "hurry, haste",
  "n14451349": "imperativeness, insistence, insistency, press, pressure",
  "n14451672": "criticality, criticalness, cruciality",
  "n14451911": "fullness",
  "n14452151": "repletion, satiety, satiation",
  "n14452294": "surfeit, excess, overabundance",
  "n14452442": "solidity",
  "n14452616": "infestation",
  "n14453066": "acariasis, acariosis, acaridiasis",
  "n14453175": "ascariasis",
  "n14453290": "coccidiosis",
  "n14453414": "echinococcosis, hydatid disease, hydatidosis",
  "n14453554": "helminthiasis",
  "n14453676": "hookworm, hookworm disease",
  "n14453840": "myiasis",
  "n14454030": "onchocerciasis, river blindness",
  "n14454311": "opisthorchiasis",
  "n14454450": "pediculosis, lousiness",
  "n14454661": "pediculosis capitis, head lice",
  "n14454773": "pediculosis corporis",
  "n14454874": "pediculosis pubis, crabs",
  "n14454988": "trombiculiasis",
  "n14455074": "trichuriasis",
  "n14455206": "emptiness",
  "n14455419": "blankness",
  "n14455552": "hollowness",
  "n14455700": "nothingness, void, nullity, nihility",
  "n14455847": "thin air",
  "n14455966": "vacancy",
  "n14456054": "vacuum, vacuity",
  "n14456138": "nakedness, nudity, nudeness",
  "n14456435": "nude",
  "n14456586": "raw, altogether, birthday suit",
  "n14456752": "undress",
  "n14456893": "bareness",
  "n14457041": "baldness, phalacrosis",
  "n14457218": "hairlessness, depilation",
  "n14457361": "alopecia",
  "n14457630": "alopecia areata",
  "n14457707": "male-patterned baldness, male pattern baldness",
  "n14457838": "dishabille, deshabille",
  "n14457976": "shirtsleeves",
  "n14458181": "grace, saving grace, state of grace",
  "n14458593": "damnation, eternal damnation",
  "n14458763": "fire and brimstone",
  "n14458943": "omniscience",
  "n14459093": "God's Wisdom",
  "n14459185": "omnipotence",
  "n14459332": "God's Will",
  "n14459422": "perfection, flawlessness, ne plus ultra",
  "n14459715": "dream",
  "n14459824": "polish, refinement, culture, cultivation, finish",
  "n14460257": "fare-thee-well",
  "n14460407": "intactness",
  "n14460565": "integrity, unity, wholeness",
  "n14460974": "completeness",
  "n14461231": "entirety, entireness, integrality, totality",
  "n14461519": "comprehensiveness, fullness",
  "n14461679": "whole shebang, whole kit and caboodle, kit and caboodle, whole kit and boodle, kit and boodle, whole kit, whole caboodle, whole works, works, full treatment",
  "n14462064": "partialness",
  "n14462193": "incompleteness, rawness",
  "n14462565": "sketchiness",
  "n14462666": "imperfection, imperfectness",
  "n14462946": "failing, weakness",
  "n14463170": "flaw",
  "n14463331": "tragic flaw, hamartia",
  "n14463471": "insufficiency, inadequacy",
  "n14463676": "fatigue",
  "n14463826": "flaw",
  "n14464005": "defect",
  "n14464203": "defect, fault, flaw",
  "n14464512": "blister",
  "n14464675": "bug, glitch",
  "n14464793": "hole",
  "n14464883": "wart",
  "n14465048": "birth defect, congenital anomaly, congenital defect, congenital disorder, congenital abnormality",
  "n14465768": "hydrocephalus, hydrocephaly",
  "n14466122": "hydronephrosis",
  "n14466258": "abrachia",
  "n14466344": "amelia",
  "n14466432": "meromelia",
  "n14466567": "phocomelia, seal limbs",
  "n14466846": "encephalocele",
  "n14466974": "familial hypercholesterolemia",
  "n14467172": "meningocele",
  "n14467395": "myelomeningocele",
  "n14467685": "plagiocephaly",
  "n14467846": "polysomy",
  "n14467975": "hermaphroditism, hermaphrodism",
  "n14468167": "pseudohermaphroditism",
  "n14468348": "progeria",
  "n14468508": "scaphocephaly",
  "n14468693": "valgus",
  "n14468870": "varus",
  "n14469014": "congenital heart defect",
  "n14469154": "septal defect",
  "n14469334": "atrial septal defect",
  "n14469466": "ventricular septal defect",
  "n14469766": "tetralogy of Fallot, Fallot's tetralogy, Fallot's syndrome",
  "n14470144": "toxic shock, toxic shock syndrome, TSS",
  "n14470496": "Waterhouse-Friderichsen syndrome",
  "n14470754": "Williams syndrome",
  "n14471054": "Zollinger-Ellison syndrome",
  "n14471224": "spina bifida, rachischisis, schistorrhachis",
  "n14471507": "spinocerebellar disorder",
  "n14471724": "polydactyly, hyperdactyly",
  "n14471926": "syndactyly, syndactylism",
  "n14472111": "tongue tie, ankyloglossia",
  "n14472299": "defectiveness, faultiness",
  "n14472526": "bugginess",
  "n14472624": "crudeness, crudity, primitiveness, primitivism, rudeness",
  "n14472900": "lameness",
  "n14473054": "sickness",
  "n14473222": "fortune, destiny, fate, luck, lot, circumstances, portion",
  "n14473655": "good fortune, luckiness, good luck",
  "n14473917": "providence",
  "n14474052": "prosperity, successfulness",
  "n14474264": "blessing, boon",
  "n14474435": "mercy",
  "n14474556": "strength",
  "n14474718": "weakness",
  "n14474894": "success",
  "n14475154": "big time",
  "n14475273": "pay dirt",
  "n14475405": "misfortune, bad luck, tough luck, ill luck",
  "n14475661": "adversity, hardship, hard knocks",
  "n14475992": "gutter, sewer, toilet",
  "n14476205": "hard cheese",
  "n14476290": "catastrophe, disaster",
  "n14476521": "extremity",
  "n14476697": "bitter end",
  "n14476852": "distress",
  "n14477076": "pressure",
  "n14477207": "throe",
  "n14477342": "affliction",
  "n14477509": "victimization",
  "n14477667": "cross, crown of thorns",
  "n14477877": "failure",
  "n14478143": "solvency",
  "n14478291": "insolvency",
  "n14478433": "bankruptcy, failure",
  "n14478684": "bankruptcy",
  "n14478862": "bank failure",
  "n14478975": "crop failure",
  "n14479086": "dead duck",
  "n14479320": "receivership",
  "n14479488": "ownership",
  "n14479615": "state of matter, state",
  "n14480065": "phase, form",
  "n14480420": "liquid, liquidness, liquidity, liquid state",
  "n14480772": "solid, solidness, solid state",
  "n14481080": "gas, gaseous state",
  "n14481511": "plasma",
  "n14481929": "possibility, possibleness",
  "n14482299": "conceivableness, conceivability",
  "n14482444": "achievability, attainability, attainableness",
  "n14482620": "potential, potentiality, potency",
  "n14482852": "latency",
  "n14482968": "prospect, chance",
  "n14483126": "impossibility, impossibleness",
  "n14483348": "inconceivability, inconceivableness",
  "n14483508": "unattainableness",
  "n14483620": "hopefulness",
  "n14483744": "confidence",
  "n14483917": "opportunity, chance",
  "n14484378": "brass ring",
  "n14484516": "day",
  "n14484646": "fresh start, clean slate, tabula rasa",
  "n14484778": "hearing, audience",
  "n14484969": "hunting ground",
  "n14485064": "occasion",
  "n14485249": "opening",
  "n14485436": "room",
  "n14485526": "say",
  "n14485673": "shot, crack",
  "n14485811": "street",
  "n14485990": "throw",
  "n14486122": "anticipation, expectation",
  "n14486274": "despair, desperation",
  "n14486533": "dejection",
  "n14486639": "nadir, low-water mark",
  "n14486767": "purity, pureness",
  "n14487026": "plainness",
  "n14487184": "impurity, impureness",
  "n14487443": "adulteration, debasement",
  "n14487598": "admixture, alloy",
  "n14487731": "contamination, taint",
  "n14487902": "dust contamination",
  "n14488004": "dirtiness",
  "n14488118": "feculence",
  "n14488200": "putridity",
  "n14488317": "financial condition",
  "n14488594": "economic condition",
  "n14488813": "boom",
  "n14488912": "credit crunch, liquidity crisis, squeeze",
  "n14489113": "depression, slump, economic crisis",
  "n14489361": "Great Depression",
  "n14489523": "full employment",
  "n14489699": "prosperity",
  "n14489859": "softness",
  "n14490110": "obligation",
  "n14490319": "indebtedness, liability, financial obligation",
  "n14490564": "debt",
  "n14490685": "arrears",
  "n14490801": "account payable, payable",
  "n14491034": "score",
  "n14491171": "scot and lot",
  "n14491271": "wealth, wealthiness",
  "n14491625": "affluence, richness",
  "n14491889": "ease, comfort",
  "n14492116": "lap of luxury",
  "n14492249": "inherited wealth",
  "n14492373": "luxury, luxuriousness, opulence, sumptuousness",
  "n14492634": "mammon",
  "n14492723": "silver spoon, old money",
  "n14492953": "sufficiency",
  "n14493145": "poverty, poorness, impoverishment",
  "n14493426": "privation, want, deprivation, neediness",
  "n14493613": "destitution",
  "n14493716": "indigence, need, penury, pauperism, pauperization",
  "n14494032": "beggary, mendicancy, mendicity",
  "n14494186": "impecuniousness, pennilessness, penuriousness",
  "n14494358": "shakeout",
  "n14494594": "wage setter",
  "n14494716": "sanitary condition",
  "n14494893": "sanitariness",
  "n14495080": "hygiene",
  "n14495230": "asepsis, antisepsis, sterility, sterileness",
  "n14495466": "sanitation",
  "n14495589": "unsanitariness",
  "n14495761": "filth, filthiness, foulness, nastiness",
  "n14495980": "dunghill",
  "n14496063": "tilth",
  "n14496193": "cleanness",
  "n14496451": "cleanliness",
  "n14496563": "spotlessness, immaculateness",
  "n14496710": "orderliness, order",
  "n14496977": "spit and polish",
  "n14497120": "kilter, kelter",
  "n14497233": "tidiness",
  "n14497365": "neatness, spruceness",
  "n14497511": "trim, trimness",
  "n14497679": "shambles",
  "n14497763": "dirtiness, uncleanness",
  "n14498096": "dirt, filth, grime, soil, stain, grease, grunge",
  "n14498404": "befoulment, defilement, pollution",
  "n14498567": "dinginess, dinge",
  "n14498733": "dustiness",
  "n14498843": "griminess, grubbiness",
  "n14498972": "smuttiness, sootiness",
  "n14499111": "sordidness, squalor, squalidness",
  "n14499262": "disorderliness, disorder",
  "n14499594": "untidiness",
  "n14499734": "sloppiness, slovenliness, unkemptness",
  "n14499953": "shagginess",
  "n14500047": "mess, messiness, muss, mussiness",
  "n14500341": "disorganization, disorganisation, disarrangement",
  "n14500567": "clutter, jumble, muddle, fuddle, mare's nest, welter, smother",
  "n14500819": "rummage",
  "n14500908": "normality, normalcy",
  "n14501172": "averageness",
  "n14501375": "commonness, expectedness",
  "n14501545": "typicality",
  "n14501726": "abnormality, abnormalcy",
  "n14502896": "atelectasis",
  "n14503060": "atypicality, untypicality",
  "n14503234": "anoxemia",
  "n14503354": "arrested development, fixation, infantile fixation, regression",
  "n14503528": "coprolalia",
  "n14503665": "aberrance, aberrancy, aberration, deviance",
  "n14503990": "cyclopia",
  "n14504103": "chromosomal aberration, chromosomal anomaly, chrosomal abnormality, chromosonal disorder",
  "n14504405": "monosomy",
  "n14504558": "trisomy",
  "n14504726": "deflection, warp",
  "n14504889": "spinal curvature",
  "n14505069": "kyphosis, humpback, hunchback",
  "n14505215": "lordosis, hollow-back",
  "n14505364": "scoliosis",
  "n14505469": "dowager's hump",
  "n14505667": "subnormality",
  "n14505821": "anomaly, anomalousness",
  "n14506020": "gynecomastia",
  "n14506233": "cross-eye, crossed eye, convergent strabismus, esotropia",
  "n14506403": "dwarfism, nanism",
  "n14506533": "pycnodysostosis",
  "n14506656": "lactose intolerance, lactase deficiency, milk intolerance",
  "n14506920": "lactosuria",
  "n14507050": "myoglobinuria",
  "n14507148": "oliguria",
  "n14507375": "phenylketonuria, PKU",
  "n14507651": "porphyria",
  "n14507787": "infantilism",
  "n14507951": "obstruction, blockage",
  "n14508149": "intestinal obstruction, ileus",
  "n14508368": "tamponade, tamponage",
  "n14508578": "cardiac tamponade",
  "n14508799": "ateleiosis, ateliosis",
  "n14508974": "macrocephaly, megacephaly, megalocephaly",
  "n14509219": "microbrachia",
  "n14509299": "microcephaly, microcephalus, nanocephaly",
  "n14509491": "pachycheilia",
  "n14509583": "phimosis",
  "n14509712": "poisoning, toxic condition, intoxication",
  "n14510104": "alkali poisoning",
  "n14510247": "caffeinism, caffeine intoxication",
  "n14510401": "carbon monoxide poisoning",
  "n14510648": "cyanide poisoning",
  "n14510814": "malathion poisoning",
  "n14510955": "ergotism",
  "n14511234": "mercury poisoning",
  "n14511491": "Minamata disease",
  "n14511730": "naphthalene poisoning",
  "n14511862": "nicotine poisoning",
  "n14512009": "ophidism",
  "n14512088": "paraquat poisoning",
  "n14512270": "parathion poisoning",
  "n14512508": "pesticide poisoning",
  "n14512674": "salicylate poisoning",
  "n14512817": "context, circumstance, setting",
  "n14513062": "ecology",
  "n14513259": "setting, background, scope",
  "n14513489": "canvas, canvass",
  "n14513694": "home",
  "n14513944": "milieu, surroundings",
  "n14514039": "sphere, domain, area, orbit, field, arena",
  "n14514392": "distaff",
  "n14514491": "front",
  "n14514685": "lotusland, lotus land",
  "n14514805": "kingdom, land, realm",
  "n14515041": "lap",
  "n14515177": "lap of the gods",
  "n14515344": "political arena, political sphere",
  "n14515463": "preserve",
  "n14515633": "province, responsibility",
  "n14515816": "ecclesiastical province",
  "n14516082": "showcase, show window",
  "n14516256": "street",
  "n14516501": "environmental condition",
  "n14516743": "pollution",
  "n14517067": "biodegradable pollution",
  "n14517243": "nonbiodegradable pollution",
  "n14517412": "air pollution",
  "n14517629": "acid rain, acid precipitation",
  "n14517852": "industrial air pollution",
  "n14518010": "miasma, miasm",
  "n14518219": "small-particle pollution",
  "n14518377": "smog, smogginess",
  "n14518539": "noise pollution, sound pollution",
  "n14518671": "thermal pollution",
  "n14518820": "water pollution",
  "n14518924": "erosion",
  "n14519097": "deforestation",
  "n14519208": "depopulation",
  "n14519366": "climate, clime",
  "n14519673": "glaciation",
  "n14519942": "inhospitableness",
  "n14520123": "meteorological conditions",
  "n14520278": "atmosphere, atmospheric state",
  "n14520518": "air mass",
  "n14520670": "high",
  "n14520829": "low, depression",
  "n14521021": "anticyclone",
  "n14521302": "cyclone",
  "n14521648": "fog, fogginess, murk, murkiness",
  "n14521954": "fug",
  "n14522113": "good weather",
  "n14522265": "calmness",
  "n14522408": "mildness, clemency",
  "n14522570": "balminess, softness",
  "n14522809": "stillness, windlessness",
  "n14522956": "lull, quiet",
  "n14523090": "bad weather, inclemency, inclementness",
  "n14523344": "raw weather",
  "n14523436": "storminess",
  "n14523669": "boisterousness",
  "n14523787": "breeziness, windiness",
  "n14523923": "tempestuousness",
  "n14524029": "choppiness, roughness, rough water",
  "n14524198": "cloudiness, cloud cover, overcast",
  "n14524383": "turbulence",
  "n14524487": "clear-air turbulence",
  "n14524661": "climate, mood",
  "n14524849": "atmosphere, ambiance, ambience",
  "n14525108": "cloud",
  "n14525274": "genius loci",
  "n14525365": "gloom, gloominess, glumness",
  "n14525548": "bleakness, desolation, bareness, nakedness",
  "n14525777": "Hollywood",
  "n14526032": "miasma, miasm",
  "n14526182": "spirit, tone, feel, feeling, flavor, flavour, look, smell",
  "n14526635": "Zeitgeist",
  "n14526764": "unsusceptibility, immunity",
  "n14526954": "resistance",
  "n14527171": "immunity, resistance",
  "n14527430": "immunogenicity",
  "n14527553": "active immunity",
  "n14527721": "passive immunity",
  "n14527977": "autoimmunity",
  "n14528193": "acquired immunity",
  "n14528571": "natural immunity, innate immunity",
  "n14528752": "racial immunity",
  "n14528873": "exemption, freedom",
  "n14529082": "amnesty",
  "n14529212": "diplomatic immunity",
  "n14529408": "indemnity",
  "n14529524": "impunity",
  "n14529612": "grandfather clause",
  "n14529835": "subservience",
  "n14530061": "susceptibility, susceptibleness",
  "n14530403": "liability",
  "n14530563": "taxability",
  "n14530659": "ratability, rateability",
  "n14530836": "capability, capacity",
  "n14531056": "habitus",
  "n14531203": "activity",
  "n14531392": "irritation",
  "n14531598": "retroversion, retroflection, retroflexion",
  "n14531772": "sensitivity, predisposition",
  "n14531983": "sensitization, sensitisation",
  "n14532142": "food allergy",
  "n14532250": "immediate allergy, atopy, atopic allergy, type I allergic reaction",
  "n14532458": "serum sickness, serum disease",
  "n14532659": "delayed allergy, type IV allergic reaction",
  "n14532816": "allergy, allergic reaction",
  "n14533106": "cryesthesia, cryaesthesia",
  "n14533203": "hypersensitivity reaction",
  "n14533547": "anaphylaxis",
  "n14533796": "hypersensitivity",
  "n14533937": "allergic rhinitis",
  "n14534076": "eosinophilia",
  "n14534199": "hay fever, pollinosis",
  "n14534333": "diathesis",
  "n14534457": "reactivity",
  "n14534571": "suggestibility",
  "n14534696": "wetness",
  "n14535056": "wateriness, muddiness, sloppiness",
  "n14535431": "moisture, wet",
  "n14535643": "humidity, humidness",
  "n14535807": "mugginess",
  "n14535905": "damp, dampness, moistness",
  "n14536085": "dankness, clamminess",
  "n14536207": "rawness",
  "n14536331": "sogginess",
  "n14536438": "dryness, waterlessness, xerotes",
  "n14536831": "dehydration, desiccation",
  "n14537054": "drought, drouth",
  "n14537250": "aridity, aridness, thirstiness",
  "n14537472": "sereness",
  "n14537563": "xeroderma, xerodermia",
  "n14537735": "xeroderma pigmentosum",
  "n14538113": "xerophthalmia, xerophthalmus, xeroma, conjunctivitis arida",
  "n14538329": "xerostomia, dry mouth",
  "n14538472": "safety",
  "n14538811": "biosafety",
  "n14538908": "risklessness",
  "n14539029": "invulnerability, impregnability",
  "n14539165": "salvation",
  "n14539268": "security",
  "n14539524": "peace, public security",
  "n14539693": "secureness",
  "n14539826": "indemnity, insurance",
  "n14539960": "protection, shelter",
  "n14540220": "collective security",
  "n14540318": "Pax Romana",
  "n14540455": "radioprotection",
  "n14540564": "cause of death, killer",
  "n14540765": "danger",
  "n14541044": "danger",
  "n14541247": "clear and present danger",
  "n14541538": "hazardousness, perilousness",
  "n14541677": "insecurity",
  "n14541852": "hazard, jeopardy, peril, risk, endangerment",
  "n14542320": "health hazard",
  "n14542441": "biohazard",
  "n14542579": "moral hazard",
  "n14542858": "occupational hazard",
  "n14542983": "sword of Damocles",
  "n14543145": "powder keg",
  "n14543231": "menace, threat",
  "n14543411": "yellow peril",
  "n14543552": "riskiness, peril",
  "n14543756": "speculativeness",
  "n14543931": "vulnerability, exposure",
  "n14544159": "insecureness",
  "n14544335": "tension, tensity, tenseness, tautness",
  "n14544672": "tonicity, tonus, tone",
  "n14545045": "catatonia",
  "n14545215": "muscular tonus, muscle tone",
  "n14545353": "myotonia",
  "n14545548": "acromyotonia",
  "n14545685": "myotonia congenita, Thomsen's disease",
  "n14545845": "atonicity, atony, atonia, amyotonia",
  "n14546042": "laxness, laxity",
  "n14546227": "condition, shape",
  "n14546432": "fitness, physical fitness",
  "n14546596": "fettle",
  "n14546701": "repair",
  "n14546844": "soundness",
  "n14547036": "seaworthiness, fitness",
  "n14547188": "airworthiness",
  "n14547369": "unfitness, softness",
  "n14547643": "infirmity, frailty, debility, feebleness, frailness, valetudinarianism",
  "n14547976": "asthenia, astheny",
  "n14548105": "cachexia, cachexy, wasting",
  "n14548343": "disability, disablement, handicap, impairment",
  "n14548913": "disability of walking",
  "n14549070": "abasia",
  "n14549284": "abasia trepidans",
  "n14549382": "atactic abasia, ataxic abasia",
  "n14549491": "choreic abasia",
  "n14549600": "paralytic abasia",
  "n14549709": "paroxysmal trepidant abasia",
  "n14549823": "spastic abasia",
  "n14549937": "lameness, limping, gimp, gimpiness, gameness, claudication",
  "n14550195": "intermittent claudication",
  "n14550366": "astasia",
  "n14550469": "amputation",
  "n14550614": "sequela",
  "n14550797": "hearing impairment, hearing disorder",
  "n14550987": "deafness, hearing loss",
  "n14551146": "conductive hearing loss, conduction deafness, middle-ear deafness",
  "n14551321": "hyperacusis, hyperacusia, auditory hyperesthesia",
  "n14551565": "sensorineural hearing loss, nerve deafness",
  "n14551704": "tone deafness, tin ear",
  "n14551822": "deaf-mutism, deaf-muteness",
  "n14551986": "mutism, muteness",
  "n14552205": "analgesia",
  "n14552355": "dysomia",
  "n14552477": "anosmia",
  "n14552718": "hyposmia",
  "n14552802": "visual impairment, visual defect, vision defect, visual disorder",
  "n14553290": "myopia, nearsightedness, shortsightedness",
  "n14553590": "astigmatism, astigmia",
  "n14553873": "anopia",
  "n14554011": "hyperopia, hypermetropia, hypermetropy, farsightedness, longsightedness",
  "n14554345": "hemeralopia, day blindness",
  "n14554459": "hemianopia, hemianopsia",
  "n14554591": "quadrantanopia",
  "n14554695": "metamorphopsia",
  "n14554853": "nyctalopia, night blindness, moon blindness",
  "n14555059": "photoretinitis",
  "n14555214": "presbyopia, farsightedness",
  "n14555414": "eye condition",
  "n14555560": "anisometropia",
  "n14555688": "isometropia",
  "n14555790": "snowblindness, snow-blindness",
  "n14555962": "retinal detachment, detachment of the retina, detached retina",
  "n14556203": "scotoma",
  "n14556426": "annular scotoma",
  "n14556552": "central scotoma",
  "n14556656": "hemianopic scotoma",
  "n14556765": "paracentral scotoma",
  "n14556879": "scintillating scotoma, flittering scotoma",
  "n14557091": "tunnel vision",
  "n14557206": "eyelessness",
  "n14557315": "figural blindness",
  "n14557415": "strabismus, squint",
  "n14557573": "walleye, divergent strabismus, exotropia",
  "n14557719": "torticollis, wryneck",
  "n14557898": "dysfunction, disfunction",
  "n14558226": "paralysis, palsy",
  "n14558599": "paresis",
  "n14558700": "paraparesis",
  "n14558801": "metroptosis, descensus uteri",
  "n14558899": "nephroptosis, nephroptosia",
  "n14558995": "ptosis",
  "n14559134": "brow ptosis",
  "n14559208": "prolapse, prolapsus, descensus",
  "n14559495": "paraplegia",
  "n14559644": "hemiplegia, unilateral paralysis",
  "n14559757": "quadriplegia",
  "n14559852": "hypoesthesia, hypesthesia",
  "n14559983": "knock-knee, genu valgum, tibia valga",
  "n14560115": "pigeon toes",
  "n14560253": "bow leg, bow legs, bandy legs",
  "n14560360": "unsoundness",
  "n14560538": "disrepair",
  "n14560612": "decay",
  "n14560926": "putrefaction, rot",
  "n14561102": "putrescence, putridness, rottenness, corruption",
  "n14561327": "decomposition, disintegration",
  "n14561461": "disintegration",
  "n14561618": "deterioration, impairment",
  "n14561839": "rancidity",
  "n14561995": "corrosion",
  "n14562142": "devastation, desolation",
  "n14562324": "ruin, ruination",
  "n14562541": "decrepitude, dilapidation",
  "n14562683": "wear",
  "n14562854": "blight",
  "n14562960": "end, destruction, death",
  "n14563222": "foulness",
  "n14563475": "impropriety",
  "n14563564": "iniquity, wickedness, darkness, dark",
  "n14563784": "light, illumination",
  "n14563928": "malady",
  "n14564056": "revocation, annulment",
  "n14564165": "merchantability",
  "n14564306": "sale",
  "n14564519": "urinary hesitancy",
  "n14564779": "wall",
  "n14564934": "sarcoidosis",
  "n14565121": "morphea",
  "n14565196": "scleroderma, dermatosclerosis",
  "n14565417": "thrombocytopenic purpura, idiopathic thrombocytopenic purpura, purpura hemorrhagica, Werlhof's disease",
  "n14565696": "sex-linked disorder",
  "n14565927": "Turner's syndrome",
  "n14566129": "urinary tract infection",
  "n14566308": "pyelonephritis",
  "n14566476": "acute pyelonephritis",
  "n14566606": "carotenemia, xanthemia",
  "n14566766": "chronic pyelonephritis",
  "n14566996": "nongonococcal urethritis, NGU",
  "n14567134": "rhinorrhea",
  "n14567281": "rhinosporidiosis",
  "n14567421": "urethritis",
  "n14567571": "nonspecific urethritis, NSU",
  "n14567690": "sodoku, spirillum fever",
  "n14567812": "stasis",
  "n14567956": "steatorrhea",
  "n14568194": "stridor",
  "n14568359": "tinnitus",
  "n14568516": "climax",
  "n14568621": "serration",
  "n14568762": "turgor",
  "n14568986": "shin splints",
  "n14569121": "hepatolenticular degeneration, Wilson's disease",
  "n14569337": "homozygosity",
  "n14569508": "heterozygosity",
  "n14569683": "neotony",
  "n14569844": "plurality",
  "n14570091": "polyvalence, polyvalency",
  "n14570330": "polyvalence, polyvalency, multivalence, multivalency",
  "n14570581": "amphidiploidy",
  "n14570676": "diploidy",
  "n14570761": "haploidy",
  "n14570846": "heteroploidy",
  "n14570939": "polyploidy",
  "n14571028": "mosaicism",
  "n14571181": "orphanage, orphanhood",
  "n14571377": "kraurosis",
  "n14571504": "kraurosis vulvae",
  "n14571618": "oligospermia",
  "n14571714": "tenesmus",
  "n14571930": "stigmatism",
  "n14572074": "transsexualism",
  "n14572265": "trismus",
  "n14572353": "uratemia",
  "n14572503": "uraturia",
  "n14572637": "ureterocele",
  "n14572770": "ureterostenosis",
  "n14572854": "urethrocele",
  "n14572951": "uricaciduria",
  "n14573099": "urocele",
  "n14573196": "uropathy",
  "n14573303": "varicocele",
  "n14573432": "varicosis",
  "n14573553": "varicosity",
  "n14573713": "varix",
  "n14573846": "viremia, viraemia",
  "n14574014": "volvulus",
  "n14574192": "xanthopsia",
  "n14574349": "absolution",
  "n14574504": "automation",
  "n14574675": "brutalization, brutalisation",
  "n14574846": "condemnation",
  "n14575063": "deification",
  "n14575180": "diversification",
  "n14575399": "exoneration",
  "n14575531": "facilitation",
  "n14575700": "frizz",
  "n14575849": "fruition",
  "n14575952": "hiding",
  "n14576053": "hospitalization",
  "n14576242": "hypertonia, hypertonus, hypertonicity",
  "n14576468": "hypotonia, hypotonus, hypotonicity",
  "n14576690": "hypertonicity",
  "n14576860": "hypotonicity",
  "n14577046": "identification",
  "n14577331": "impaction",
  "n14577469": "ionization, ionisation",
  "n14577753": "irradiation",
  "n14577872": "leakiness",
  "n14578104": "lubrication",
  "n14578302": "mechanization, mechanisation",
  "n14578471": "motivation",
  "n14578634": "mummification",
  "n14578792": "paternity",
  "n14578940": "preservation",
  "n14579063": "prognathism",
  "n14579215": "rustication",
  "n14579365": "rustiness",
  "n14579489": "scandalization, scandalisation",
  "n14579623": "slot",
  "n14579827": "toehold",
  "n14580090": "submission",
  "n14580334": "urbanization, urbanisation",
  "n14580476": "utilization",
  "n14580597": "substance",
  "n14580752": "chemistry",
  "n14580897": "material, stuff",
  "n14582025": "ylem",
  "n14582220": "dark matter",
  "n14582535": "antimatter",
  "n14582716": "micronutrient",
  "n14582870": "philosopher's stone, philosophers' stone, elixir",
  "n14583066": "precursor",
  "n14583228": "phlogiston",
  "n14583400": "chyme",
  "n14583573": "glop",
  "n14583670": "impurity, dross",
  "n14583843": "acceptor",
  "n14584016": "adduct",
  "n14584110": "actinoid, actinide, actinon",
  "n14584260": "allergen",
  "n14584390": "assay",
  "n14584502": "pyrogen, pyrectic",
  "n14584658": "pyrogen",
  "n14584765": "aldehyde",
  "n14585048": "alkapton, alcapton, homogentisic acid",
  "n14585223": "antiknock",
  "n14585392": "ragweed pollen",
  "n14585519": "atom, molecule, particle, corpuscle, mote, speck",
  "n14585808": "ammunition",
  "n14585960": "floccule, floc",
  "n14586119": "HAZMAT",
  "n14586258": "mixture",
  "n14586769": "alloy, metal",
  "n14587688": "botulinum toxin",
  "n14587886": "botulinum toxin A, Botox",
  "n14588219": "colloid",
  "n14588492": "composition",
  "n14588646": "dispersed phase, dispersed particles",
  "n14588797": "dispersing phase, dispersion medium, dispersing medium",
  "n14588986": "mechanical mixture",
  "n14589114": "eutectic",
  "n14589223": "solution",
  "n14589912": "aqueous solution",
  "n14589994": "Ringer's solution, Ringer solution",
  "n14590293": "saline solution, saline",
  "n14590441": "polyelectrolyte",
  "n14590542": "gel, colloidal gel",
  "n14590685": "hydrogel",
  "n14590795": "sol, colloidal solution, colloidal suspension",
  "n14590975": "hydrocolloid",
  "n14591091": "suspension",
  "n14591333": "slurry",
  "n14591481": "resuspension",
  "n14591635": "precipitate",
  "n14591804": "sludge",
  "n14591901": "domoic acid",
  "n14592028": "acrylonitrile-butadiene-styrene, ABS",
  "n14592217": "Mylar",
  "n14592309": "pina cloth",
  "n14592405": "Plasticine",
  "n14592610": "plastic",
  "n14593162": "plasticizer, plasticiser",
  "n14593344": "thermoplastic, thermoplastic resin",
  "n14593545": "saran",
  "n14593671": "acrylic, acrylic resin, acrylate resin",
  "n14593874": "polymethyl methacrylate",
  "n14594032": "Lucite, Perspex",
  "n14594174": "Plexiglas, plexiglass",
  "n14594313": "polyethylene, polythene",
  "n14594456": "acetyl, acetyl group, acetyl radical, ethanoyl group, ethanoyl radical",
  "n14594708": "acyl, acyl group",
  "n14594925": "aggregate",
  "n14595076": "alcohol group, alcohol radical",
  "n14595176": "aldehyde group, aldehyde radical",
  "n14595279": "polyvinyl acetate, PVA",
  "n14595403": "polyvinyl chloride, PVC",
  "n14595543": "styrene, cinnamene, phenylethylene, vinylbenzene",
  "n14595711": "polystyrene",
  "n14595933": "Styrofoam",
  "n14596063": "thermosetting compositions, thermosetting resin",
  "n14596216": "Bakelite",
  "n14596398": "Teflon, polytetrafluoroethylene",
  "n14596599": "Vinylite",
  "n14596700": "raw material, staple",
  "n14596839": "feedstock",
  "n14596956": "sorbate",
  "n14597158": "sorbent, sorbent material",
  "n14597413": "absorbent material, absorbent",
  "n14597628": "absorbate",
  "n14597758": "sponge",
  "n14598079": "absorber",
  "n14598251": "absorbent cotton",
  "n14598383": "absorption indicator",
  "n14598525": "adsorbent, adsorbent material",
  "n14598704": "adsorbate",
  "n14598834": "acaricide, acaracide",
  "n14598937": "acaroid resin, accaroid resin, accroides, accroides resin, accroides gum, gum accroides",
  "n14599168": "acetic acid, ethanoic acid",
  "n14599494": "acetin",
  "n14599641": "acetum",
  "n14599806": "acetate, ethanoate",
  "n14599938": "dichlorodiphenyltrichloroethane, DDT",
  "n14600135": "larvacide",
  "n14600238": "lead arsenate",
  "n14600357": "tetraethyl lead, lead tetraethyl",
  "n14600504": "acetone, propanone, dimethyl ketone",
  "n14600742": "acetylene, ethyne, alkyne",
  "n14600927": "adobe",
  "n14601039": "Agent Orange",
  "n14601178": "alicyclic compound",
  "n14601294": "aliphatic compound",
  "n14601505": "alkylbenzene",
  "n14601646": "alkyl halide, haloalkane",
  "n14601829": "amino acid, aminoalkanoic acid",
  "n14602490": "alanine",
  "n14602597": "argil",
  "n14602721": "arsenical",
  "n14602851": "asparagine",
  "n14602987": "aspartic acid",
  "n14603143": "canavanine",
  "n14603236": "chlorobenzene",
  "n14603497": "chlorofluorocarbon, CFC",
  "n14603798": "chlorobenzylidenemalononitrile, CS gas",
  "n14604038": "chloroacetophenone, CN gas",
  "n14604184": "citrate",
  "n14604286": "citrulline",
  "n14604454": "cysteine",
  "n14604616": "cystine",
  "n14604763": "diamagnet",
  "n14604857": "diamine",
  "n14604959": "dopa, dihydroxyphenylalanine",
  "n14605132": "L-dopa, levodopa, Bendopa, Brocadopa, Larodopa",
  "n14605415": "endonuclease",
  "n14605590": "enol",
  "n14605787": "essential amino acid",
  "n14606137": "exonuclease",
  "n14606290": "gamma aminobutyric acid, GABA",
  "n14606482": "glutamic acid, glutaminic acid",
  "n14606704": "glutamine",
  "n14606839": "glutathione peroxidase",
  "n14606993": "glycine",
  "n14607131": "hydroxyproline",
  "n14607250": "iodoamino acid",
  "n14607378": "ornithine",
  "n14607521": "acid",
  "n14609198": "acid-base indicator",
  "n14609443": "alpha-linolenic acid",
  "n14609689": "alpha-naphthol",
  "n14609784": "alpha-naphthol test, Molisch's test, Molisch test, Molisch reaction",
  "n14610088": "aromatic compound",
  "n14610261": "arsenate",
  "n14610347": "arsenic acid",
  "n14610443": "arsenide",
  "n14610548": "cerotic acid, hexacosanoic acid",
  "n14610703": "chlorate",
  "n14610782": "chloric acid",
  "n14610914": "chlorous acid",
  "n14611031": "monobasic acid",
  "n14611158": "dibasic acid",
  "n14611279": "dibasic salt",
  "n14611397": "tribasic acid",
  "n14611521": "tritium",
  "n14611679": "tetrabasic acid",
  "n14611804": "fulminic acid",
  "n14611980": "gamma acid",
  "n14612077": "heavy metal",
  "n14612317": "hexanedioic acid, adipic acid",
  "n14612444": "HMG-CoA reductase, 5-hydroxy-3-methylglutaryl-coenzyme A reductase",
  "n14612618": "horseradish peroxidase",
  "n14612764": "hydrazoic acid, azoimide, hydrogen azide, HN",
  "n14612972": "hydriodic acid",
  "n14613130": "hydrochlorofluorocarbon, HCFC",
  "n14613368": "hydrogen cyanide",
  "n14613643": "hydrocyanic acid, prussic acid",
  "n14613841": "hydrolysate",
  "n14613922": "hydroxy acid",
  "n14614139": "hydroxybenzoic acid",
  "n14614245": "hypochlorite",
  "n14614378": "hyponitrous acid",
  "n14614490": "hypophosphoric acid",
  "n14614594": "hypophosphorous acid, phosphorous acid, orthophosphorous acid",
  "n14614744": "juniperic acid",
  "n14614864": "lysergic acid",
  "n14615006": "manganic acid",
  "n14615135": "metaphosphoric acid",
  "n14615264": "polyphosphoric acid",
  "n14615400": "pyrogallol, pyrogallic acid",
  "n14615602": "pyrophosphoric acid",
  "n14615724": "pyrophosphate",
  "n14615822": "methacrylic acid, 2-methylpropenoic acid",
  "n14615971": "mucic acid",
  "n14616073": "selenic acid",
  "n14616181": "suberic acid, octanedioic acid",
  "n14616292": "succinic acid",
  "n14616410": "sulfonate",
  "n14616508": "sulfonic acid, sulphonic acid",
  "n14616620": "titanic acid",
  "n14616740": "titanium dioxide, titanium oxide, titanic oxide, titania",
  "n14616939": "adulterant, adulterator",
  "n14617189": "alkyl, alkyl group, alkyl radical",
  "n14617427": "allyl, allyl group, allyl radical",
  "n14617597": "amino, amino group",
  "n14617715": "aminomethane",
  "n14617834": "amyl",
  "n14617944": "anionic compound",
  "n14618050": "anionic detergent, anionic, non-ionic detergent",
  "n14618253": "base, alkali",
  "n14618674": "base metal",
  "n14618834": "binary compound",
  "n14619033": "chelate, chelate compound",
  "n14619225": "atom",
  "n14619658": "isotope",
  "n14619857": "radioisotope",
  "n14620066": "label",
  "n14620257": "halon",
  "n14620549": "bromoform, tribromomethane",
  "n14620654": "fluoroform, trifluoromethane",
  "n14620781": "iodoform, tri-iodomethane",
  "n14620895": "haloform",
  "n14621080": "monad",
  "n14621194": "azido group, azido radical",
  "n14621319": "azo group, azo radical",
  "n14621446": "group, radical, chemical group",
  "n14622141": "fullerene",
  "n14622350": "buckminsterfullerene, buckyball",
  "n14622491": "carbon nanotube, nanotube",
  "n14622623": "benzyl, benzyl group, benzyl radical",
  "n14622768": "benzoyl group, benzoyl radical",
  "n14622893": "chemical element, element",
  "n14624025": "allotrope",
  "n14624191": "transuranic element",
  "n14624369": "noble gas, inert gas, argonon",
  "n14624659": "helium group",
  "n14624743": "rare earth, rare-earth element, lanthanoid, lanthanide, lanthanon",
  "n14624944": "terbium metal",
  "n14625110": "actinide series",
  "n14625297": "lanthanide series",
  "n14625458": "metallic element, metal",
  "n14627081": "noble metal",
  "n14627246": "nonmetal",
  "n14627373": "transactinide",
  "n14627529": "metallized dye",
  "n14627655": "actinium, Ac, atomic number 89",
  "n14627820": "aluminum, aluminium, Al, atomic number 13",
  "n14628119": "alum, potassium alum, potash alum",
  "n14628307": "alum, ammonia alum, ammonium alum",
  "n14628494": "americium, Am, atomic number 95",
  "n14628668": "antimony, Sb, atomic number 51",
  "n14628920": "argon, Ar, atomic number 18",
  "n14629149": "arsenic, As, atomic number 33",
  "n14629561": "astatine, At, atomic number 85",
  "n14629772": "atrazine",
  "n14629998": "barium, Ba, atomic number 56",
  "n14630204": "baryta",
  "n14630363": "barium hydroxide",
  "n14630492": "barium monoxide, barium oxide, barium protoxide",
  "n14630630": "barium dioxide, barium peroxide",
  "n14630769": "base pair",
  "n14631134": "berkelium, Bk, atomic number 97",
  "n14631295": "beryllium, Be, glucinium, atomic number 4",
  "n14631502": "bismuth, Bi, atomic number 83",
  "n14631757": "bohrium, Bh, element 107, atomic number 107",
  "n14631871": "boron, B, atomic number 5",
  "n14632129": "bromine, Br, atomic number 35",
  "n14632444": "cadmium, Cd, atomic number 48",
  "n14632648": "calcium, Ca, atomic number 20",
  "n14633037": "californium, Cf, atomic number 98",
  "n14633206": "carbon, C, atomic number 6",
  "n14633781": "carbon atom",
  "n14633856": "radiocarbon, carbon 14",
  "n14633957": "cerium, Ce, atomic number 58",
  "n14634232": "cesium, caesium, Cs, atomic number 55",
  "n14634477": "cesium 137",
  "n14634591": "chlorine, Cl, atomic number 17",
  "n14634999": "radiochlorine",
  "n14635092": "chromium, Cr, atomic number 24",
  "n14635290": "cobalt, Co, atomic number 27",
  "n14635542": "cobalt 60",
  "n14635722": "copper, Cu, atomic number 29",
  "n14636220": "curium, Cm, atomic number 96",
  "n14636392": "darmstadtium, Ds, element 110, atomic number 110",
  "n14636523": "dubnium, Db, hahnium, element 105, atomic number 105",
  "n14636647": "dysprosium, Dy, atomic number 66",
  "n14636822": "einsteinium, Es, E, atomic number 99",
  "n14636988": "erbium, Er, atomic number 68",
  "n14637176": "europium, Eu, atomic number 63",
  "n14637339": "fermium, Fm, atomic number 100",
  "n14637507": "fluorine, F, atomic number 9",
  "n14637864": "francium, Fr, atomic number 87",
  "n14638041": "gadolinium, Gd, atomic number 64",
  "n14638256": "gallium, Ga, atomic number 31",
  "n14638517": "germanium, Ge, atomic number 32",
  "n14638799": "gold, Au, atomic number 79",
  "n14639269": "18-karat gold",
  "n14639368": "22-karat gold",
  "n14639467": "24-karat gold, pure gold",
  "n14639556": "hafnium, Hf, atomic number 72",
  "n14639795": "hassium, Hs, element 108, atomic number 108",
  "n14639921": "helium, He, atomic number 2",
  "n14640222": "holmium, Ho, atomic number 67",
  "n14640434": "hydrogen, H, atomic number 1",
  "n14640756": "hydrogen atom",
  "n14640890": "acidic hydrogen, acid hydrogen",
  "n14641046": "deuterium, heavy hydrogen",
  "n14641223": "indium, In, atomic number 49",
  "n14641397": "iodine, iodin, I, atomic number 53",
  "n14641797": "iodine-131",
  "n14642005": "iodine-125",
  "n14642219": "iridium, Ir, atomic number 77",
  "n14642417": "iron, Fe, atomic number 26",
  "n14642916": "krypton, Kr, atomic number 36",
  "n14643118": "lanthanum, La, atomic number 57",
  "n14643323": "lawrencium, Lr, atomic number 103",
  "n14643467": "lead, Pb, atomic number 82",
  "n14643793": "lithium, Li, atomic number 3",
  "n14644063": "lutetium, lutecium, Lu, atomic number 71",
  "n14644249": "magnesium, Mg, atomic number 12",
  "n14644654": "manganese, Mn, atomic number 25",
  "n14644963": "meitnerium, Mt, element 109, atomic number 109",
  "n14645092": "mendelevium, Md, Mv, atomic number 101",
  "n14645346": "mercury, quicksilver, hydrargyrum, Hg, atomic number 80",
  "n14645661": "molybdenum, Mo, atomic number 42",
  "n14645882": "neodymium, Nd, atomic number 60",
  "n14646152": "neon, Ne, atomic number 10",
  "n14646403": "neptunium, Np, atomic number 93",
  "n14646610": "nickel, Ni, atomic number 28",
  "n14646942": "niobium, Nb, atomic number 41",
  "n14647154": "columbium",
  "n14647235": "nitrogen, N, atomic number 7",
  "n14647623": "azote",
  "n14647722": "nobelium, No, atomic number 102",
  "n14647907": "osmium, Os, atomic number 76",
  "n14648100": "oxygen, O, atomic number 8",
  "n14648526": "liquid oxygen, LOX",
  "n14648754": "palladium, Pd, atomic number 46",
  "n14649036": "phosphor",
  "n14649197": "phosphorus, P, atomic number 15",
  "n14649543": "platinum, Pt, atomic number 78",
  "n14649775": "plutonium, Pu, atomic number 94",
  "n14650137": "plutonium 239",
  "n14650366": "weapons plutonium, weapon-grade plutonium",
  "n14650556": "polonium, Po, atomic number 84",
  "n14650807": "potassium, K, atomic number 19",
  "n14651212": "praseodymium, Pr, atomic number 59",
  "n14651479": "promethium, Pm, atomic number 61",
  "n14651708": "protactinium, protoactinium, Pa, atomic number 91",
  "n14651921": "radium, Ra, atomic number 88",
  "n14652104": "radon, Rn, atomic number 86",
  "n14652390": "rhenium, Re, atomic number 75",
  "n14652623": "rhodium, Rh, atomic number 45",
  "n14652824": "roentgenium, Rg, element 111, atomic number 111",
  "n14652954": "rubidium, Rb, atomic number 37",
  "n14653242": "ruthenium, Ru, atomic number 44",
  "n14653416": "rutherfordium, Rf, unnilquadium, Unq, element 104, atomic number 104",
  "n14653596": "samarium, Sm, atomic number 62",
  "n14653831": "scandium, Sc, atomic number 21",
  "n14654058": "seaborgium, Sg, element 106, atomic number 106",
  "n14654175": "selenium, Se, atomic number 34",
  "n14654541": "silicon, Si, atomic number 14",
  "n14654954": "silver, Ag, atomic number 47",
  "n14655371": "sodium, Na, atomic number 11",
  "n14655731": "strontium, Sr, atomic number 38",
  "n14656001": "strontium 90",
  "n14656219": "sulfur, S, sulphur, atomic number 16",
  "n14656666": "tantalum, Ta, atomic number 73",
  "n14656926": "taurine",
  "n14657047": "technetium, Tc, atomic number 43",
  "n14657228": "tellurium, Te, atomic number 52",
  "n14657566": "terbium, Tb, atomic number 65",
  "n14657818": "thallium, Tl, atomic number 81",
  "n14658109": "thorium, Th, atomic number 90",
  "n14658422": "thorium-228, radiothorium",
  "n14658546": "thulium, Tm, atomic number 69",
  "n14658855": "tin, Sn, atomic number 50",
  "n14659211": "titanium, Ti, atomic number 22",
  "n14659512": "tungsten, wolfram, W, atomic number 74",
  "n14659794": "ununbium, Uub, element 112, atomic number 112",
  "n14659922": "ununhexium, Uuh, element 116, atomic number 116",
  "n14660052": "ununpentium, Uup, element 115, atomic number 115",
  "n14660183": "ununquadium, Uuq, element 114, atomic number 114",
  "n14660314": "ununtrium, Uut, element 113, atomic number 113",
  "n14660443": "uranium, U, atomic number 92",
  "n14660700": "uranium 235",
  "n14660835": "uranium 238",
  "n14661020": "vanadium, V, atomic number 23",
  "n14661274": "xenon, Xe, atomic number 54",
  "n14661482": "ytterbium, Yb, atomic number 70",
  "n14661740": "yttrium, Y, atomic number 39",
  "n14661977": "zinc, Zn, atomic number 30",
  "n14662281": "zirconium, Zr, atomic number 40",
  "n14662574": "mineral",
  "n14665102": "ader wax, earth wax, mineral wax, ozokerite, ozocerite",
  "n14665351": "alabaster",
  "n14665469": "alabaster, oriental alabaster, onyx marble, Mexican onyx",
  "n14665605": "amblygonite",
  "n14665767": "amphibole",
  "n14666012": "amphibole group",
  "n14666369": "amphibolite",
  "n14666510": "apatite",
  "n14666758": "aragonite",
  "n14666885": "argentite",
  "n14667017": "argillite",
  "n14667181": "argyrodite",
  "n14667328": "arsenopyrite, mispickel",
  "n14667455": "asphalt, mineral pitch",
  "n14667668": "augite",
  "n14667855": "azurite",
  "n14667949": "baddeleyite",
  "n14668065": "bastnasite, bastnaesite",
  "n14668277": "bauxite",
  "n14668539": "beryl",
  "n14668743": "biotite",
  "n14668895": "bone black, bone char, animal black, animal charcoal",
  "n14669084": "borax",
  "n14669242": "bornite, peacock ore",
  "n14669413": "carnallite",
  "n14669677": "carnotite",
  "n14669823": "caspase",
  "n14669927": "cassiterite",
  "n14670061": "celestite",
  "n14670178": "cerussite, white lead ore",
  "n14670344": "chalcocite, copper glance",
  "n14670481": "chalcopyrite, copper pyrites",
  "n14670639": "china clay, china stone, kaolin, kaoline, porcelain clay, terra alba",
  "n14670954": "chlorite",
  "n14671125": "chromite",
  "n14671253": "chromogen",
  "n14671372": "chrysoberyl",
  "n14671587": "cinnabar",
  "n14671744": "cobalt blue, cobalt ultramarine",
  "n14671895": "cobaltite",
  "n14672023": "sodium chloride, common salt",
  "n14672224": "halite, rock salt",
  "n14672373": "columbite-tantalite, coltan",
  "n14672544": "cordierite",
  "n14672717": "corundom, corundum",
  "n14672893": "cristobalite",
  "n14673032": "crocolite",
  "n14673150": "cryolite, Greenland spar",
  "n14673325": "cuprite",
  "n14673462": "cutin",
  "n14673669": "damourite",
  "n14673747": "dolomite, bitter spar",
  "n14673978": "earth color",
  "n14674143": "emery",
  "n14674408": "emulsifier",
  "n14674584": "emulsion",
  "n14674776": "erythrite, cobalt bloom",
  "n14675012": "fergusonite",
  "n14675216": "fluorapatite",
  "n14675356": "fluorite, fluorspar, fluor",
  "n14675569": "gadolinite, ytterbite",
  "n14675910": "galena",
  "n14676042": "garnet",
  "n14676326": "garnierite",
  "n14676486": "almandite, almandine",
  "n14676608": "germanite",
  "n14676756": "gesso",
  "n14676943": "gibbsite",
  "n14677144": "glauconite",
  "n14677314": "goethite, gothite",
  "n14677485": "greaves, crackling",
  "n14677610": "greenockite, cadmium sulphide",
  "n14677778": "gypsum",
  "n14678068": "hausmannite",
  "n14678230": "heavy spar, barite, barytes, barium sulphate",
  "n14678406": "hemimorphite, calamine",
  "n14678551": "ilmenite",
  "n14678762": "jadeite",
  "n14678952": "kainite",
  "n14679147": "kaolinite",
  "n14679267": "kernite",
  "n14679447": "kieserite",
  "n14679584": "kyanite, cyanite",
  "n14679780": "lactate",
  "n14679882": "langbeinite",
  "n14680040": "lecithin",
  "n14680261": "lepidolite",
  "n14680415": "lepidomelane",
  "n14680514": "magnesite",
  "n14680661": "malachite",
  "n14680836": "maltha, mineral tar",
  "n14680963": "manganese tetroxide",
  "n14681100": "manganite",
  "n14681249": "marl",
  "n14681445": "meerschaum, sepiolite",
  "n14681555": "mica, isinglass",
  "n14681987": "millerite",
  "n14682133": "molecule",
  "n14682469": "molybdenite",
  "n14682642": "monazite",
  "n14682963": "monomer",
  "n14683085": "muscovite",
  "n14683204": "nepheline, nephelite",
  "n14683447": "nephelinite",
  "n14683595": "nephrite",
  "n14683859": "niobite, columbite",
  "n14684017": "nitrocalcite",
  "n14684111": "olivine",
  "n14684272": "olivenite",
  "n14684427": "ozonide",
  "n14684607": "perchlorate",
  "n14684690": "perchloric acid",
  "n14684798": "proline",
  "n14684918": "biomass",
  "n14685017": "butane",
  "n14685172": "char",
  "n14685296": "charcoal, wood coal",
  "n14685475": "coal gas",
  "n14685641": "town gas",
  "n14685768": "coke",
  "n14685881": "diesel oil, diesel fuel",
  "n14686020": "derv",
  "n14686186": "fire",
  "n14686352": "fossil fuel",
  "n14686585": "fuel oil, heating oil",
  "n14686723": "gasohol",
  "n14686913": "gasoline, gasolene, gas, petrol",
  "n14687261": "leaded gasoline, leaded petrol",
  "n14687513": "illuminant",
  "n14687633": "kerosene, kerosine, lamp oil, coal oil",
  "n14687818": "methanol, methyl alcohol, wood alcohol, wood spirit",
  "n14688089": "nuclear fuel",
  "n14688234": "opal",
  "n14688500": "ore",
  "n14688705": "oroide, oreide",
  "n14688831": "orpiment",
  "n14688978": "osmiridium, iridosmine",
  "n14689226": "paragonite",
  "n14689325": "paraldehyde, ethanal trimer",
  "n14689493": "peat",
  "n14689672": "pentlandite",
  "n14689817": "triose",
  "n14689940": "tetrose",
  "n14690063": "pentose",
  "n14690186": "hexose",
  "n14690335": "pentoxide",
  "n14690447": "peptone",
  "n14690607": "periclase, magnesia, magnesium oxide",
  "n14690790": "phlogopite",
  "n14690938": "pinite",
  "n14691085": "pollucite",
  "n14691231": "quaternary ammonium compound",
  "n14691445": "proenzyme, zymogen",
  "n14691686": "propane",
  "n14691822": "propellant, propellent",
  "n14692026": "rocket fuel, rocket propellant, rocket propellent",
  "n14692202": "propylthiouracil",
  "n14692342": "psilomelane",
  "n14692510": "pyridine",
  "n14692682": "pyrite, iron pyrite, fool's gold",
  "n14692844": "pyrites",
  "n14692973": "pyrolusite",
  "n14693124": "pyromorphite, green lead ore",
  "n14693275": "pyrophyllite",
  "n14693396": "pyroxene",
  "n14693575": "pyrrhotite, pyrrhotine, magnetic pyrites",
  "n14693733": "quartz",
  "n14694161": "quartzite",
  "n14694293": "rock crystal, transparent quartz",
  "n14694453": "rhinestone",
  "n14694571": "cairngorm, smoky quartz",
  "n14694672": "rag paper",
  "n14694765": "reactant",
  "n14694910": "realgar",
  "n14695069": "red clay",
  "n14695184": "red fire",
  "n14695351": "regosol",
  "n14695497": "regur, regur soil",
  "n14695588": "residual soil, residual clay",
  "n14695737": "topsoil, surface soil",
  "n14695838": "subsoil, undersoil",
  "n14695953": "resinoid",
  "n14696035": "rhodochrosite",
  "n14696180": "rhodonite",
  "n14696331": "ribose",
  "n14696463": "ricin, ricin toxin",
  "n14696690": "road metal",
  "n14696793": "rock, stone",
  "n14697485": "arenaceous rock",
  "n14697600": "argillaceous rock",
  "n14697708": "rudaceous rock",
  "n14697839": "breccia",
  "n14698000": "sedimentary rock",
  "n14698233": "sial",
  "n14698405": "sima",
  "n14698568": "marlite, marlstone",
  "n14698698": "metamorphic rock",
  "n14698884": "gravel, crushed rock",
  "n14699068": "hornfels, hornstone",
  "n14699209": "ballast",
  "n14699321": "bank gravel, pit-run gravel, pit run",
  "n14699441": "caliche",
  "n14699574": "shingle",
  "n14699752": "gem, gemstone, stone",
  "n14700046": "cabochon",
  "n14700162": "slate",
  "n14700280": "shingling",
  "n14700438": "pumice, pumice stone",
  "n14700594": "grit, gritrock, gritstone",
  "n14700745": "animal product",
  "n14700985": "ambergris",
  "n14701143": "lac",
  "n14701412": "garnet lac",
  "n14701521": "gum-lac",
  "n14701628": "shellac",
  "n14701826": "mosaic gold, stannic sulfide",
  "n14701949": "stick lac",
  "n14702117": "seed lac",
  "n14702256": "Sonora lac",
  "n14702416": "adhesive material, adhesive agent, adhesive",
  "n14702703": "birdlime, lime",
  "n14702875": "glue, gum, mucilage",
  "n14703190": "animal glue",
  "n14703345": "casein glue",
  "n14703458": "Crazy Glue",
  "n14703566": "fish glue",
  "n14703709": "marine glue",
  "n14703797": "putty",
  "n14703999": "iron putty",
  "n14704152": "red-lead putty",
  "n14704328": "spirit gum",
  "n14704465": "binder",
  "n14704640": "cement",
  "n14704851": "mastic",
  "n14704966": "paste, library paste",
  "n14705183": "wafer",
  "n14705287": "paste",
  "n14705386": "rubber cement",
  "n14705533": "sealing material",
  "n14705718": "sealant, sealer",
  "n14706026": "filler",
  "n14706214": "lute, luting",
  "n14706372": "size, sizing",
  "n14706574": "purine",
  "n14706749": "purine",
  "n14706889": "adenine, A",
  "n14707101": "adenosine",
  "n14707361": "adenosine deaminase, ADA",
  "n14707710": "adenosine monophosphate, AMP, adenylic acid",
  "n14707903": "adenosine diphosphate, ADP",
  "n14708042": "adenosine triphosphate, ATP",
  "n14708232": "agate",
  "n14708413": "moss agate",
  "n14708531": "Alar, daminozide",
  "n14708720": "alcohol",
  "n14709102": "ether",
  "n14709265": "ethyl alcohol, ethanol, fermentation alcohol, grain alcohol",
  "n14709706": "spirits of wine",
  "n14709791": "absolute alcohol",
  "n14709907": "isopropyl alcohol, isopropanol",
  "n14710024": "denaturant",
  "n14710164": "denatured alcohol",
  "n14710325": "ethyl, ethyl group, ethyl radical",
  "n14710501": "aldohexose",
  "n14710662": "aldose",
  "n14710805": "acetal",
  "n14710935": "acetaldol",
  "n14711008": "acetaldehyde, ethanal",
  "n14711197": "acetamide, ethanamide",
  "n14711364": "acrylamide",
  "n14711647": "agglomerate",
  "n14711799": "aldol, aldehyde-alcohol",
  "n14712036": "alkali",
  "n14712224": "alkali metal, alkaline metal",
  "n14712489": "alkaline earth, alkaline-earth metal",
  "n14712692": "alkaloid",
  "n14713120": "alkene, olefine, olefin",
  "n14713247": "alkylbenzenesulfonate",
  "n14713341": "cinchonine",
  "n14713487": "ephedrine",
  "n14713748": "ergonovine, Ergotrate Maleate",
  "n14714028": "ergotamine",
  "n14714213": "pseudoephedrine",
  "n14714353": "epsilon toxin, Clostridium perfringens epsilon toxin",
  "n14714645": "aflatoxin",
  "n14714817": "nicotine",
  "n14715008": "strychnine",
  "n14715189": "brucine",
  "n14715356": "shag",
  "n14715453": "Turkish tobacco",
  "n14715601": "latakia",
  "n14715679": "alexandrite",
  "n14715786": "alloy iron, alloy cast iron",
  "n14716042": "alloy steel",
  "n14716358": "Alnico",
  "n14716550": "amalgam, dental amalgam",
  "n14716845": "fusible metal",
  "n14716997": "brass",
  "n14717275": "bronze",
  "n14717634": "gunmetal",
  "n14717785": "phosphor bronze",
  "n14717925": "cupronickel",
  "n14718017": "electrum",
  "n14718099": "pewter",
  "n14718233": "pinchbeck",
  "n14718362": "pot metal",
  "n14718483": "hard solder",
  "n14718626": "silver solder",
  "n14718715": "soft solder",
  "n14718822": "solder",
  "n14719025": "gold dust",
  "n14719169": "white gold",
  "n14719291": "Monel metal, Monell metal",
  "n14719458": "type metal",
  "n14719597": "white metal, bearing metal",
  "n14719725": "alluvial soil",
  "n14719893": "allyl alcohol, propenyl alcohol",
  "n14720101": "allyl resin",
  "n14720238": "alpha-beta brass, Muntz metal, yellow metal",
  "n14720474": "alpha brass",
  "n14720588": "alpha bronze",
  "n14720692": "alpha iron",
  "n14720833": "alpha-tocopheral",
  "n14720962": "carotenoid",
  "n14721236": "carotene",
  "n14721342": "lycopene",
  "n14721470": "beta-carotene",
  "n14721635": "xanthophyll, xanthophyl, lutein",
  "n14721781": "zeaxanthin",
  "n14721951": "betaine",
  "n14722056": "beta iron",
  "n14722240": "gamma iron",
  "n14722426": "delta iron",
  "n14722589": "amethyst",
  "n14722738": "amygdaloid",
  "n14722912": "aluminum bronze, aluminium bronze",
  "n14723079": "activator",
  "n14723425": "activating agent",
  "n14723628": "catalyst, accelerator",
  "n14723909": "biocatalyst",
  "n14724025": "sensitizer, sensitiser",
  "n14724264": "amide",
  "n14724436": "inhibitor",
  "n14724645": "antioxidant",
  "n14724916": "rust inhibitor",
  "n14725024": "anticatalyst",
  "n14725217": "actinolite",
  "n14725364": "andesite",
  "n14725444": "anthophyllite",
  "n14725591": "asbestos",
  "n14725809": "chrysotile",
  "n14725941": "tremolite",
  "n14726124": "hornblende",
  "n14726315": "aphanite",
  "n14726484": "aplite",
  "n14726642": "afterdamp",
  "n14726823": "dacite",
  "n14726998": "firedamp",
  "n14727153": "carrier",
  "n14727355": "moderator",
  "n14727508": "heavy water, deuterium oxide",
  "n14727670": "organic compound",
  "n14728724": "protein",
  "n14729633": "recombinant protein",
  "n14729737": "actomyosin",
  "n14729953": "aleurone",
  "n14730105": "amyloid",
  "n14730301": "apoenzyme",
  "n14730421": "beta-naphthol",
  "n14730553": "gelatin, gelatine",
  "n14730802": "chondrin",
  "n14730955": "mucin",
  "n14731135": "conjugated protein, compound protein",
  "n14731334": "actin",
  "n14731509": "albumin, albumen",
  "n14731716": "lactalbumin",
  "n14731799": "serum albumin",
  "n14731964": "alpha globulin",
  "n14732116": "serum globulin",
  "n14732299": "C-reactive protein, CRP",
  "n14732472": "keratin, ceratin",
  "n14732722": "chitin",
  "n14732946": "enzyme",
  "n14733941": "fibrin",
  "n14734164": "filaggrin",
  "n14734348": "growth factor",
  "n14734487": "nerve growth factor, NGF",
  "n14734625": "haptoglobin",
  "n14734761": "iodoprotein, iodinated protein",
  "n14734905": "nucleoprotein",
  "n14735077": "opsin",
  "n14735225": "phosphoprotein",
  "n14735347": "casein",
  "n14735457": "amylase",
  "n14735642": "angiotensin converting enzyme, angiotensin-converting enzyme, ACE",
  "n14735822": "cholinesterase",
  "n14735953": "coagulase",
  "n14736079": "collagenase",
  "n14736201": "complement",
  "n14736359": "plasma protein",
  "n14736510": "prostate specific antigen, PSA",
  "n14736854": "proteome",
  "n14736972": "simple protein",
  "n14737249": "thrombin",
  "n14737365": "tumor necrosis factor, tumour necrosis factor, TNF",
  "n14737639": "catalase",
  "n14737847": "cyclooxygenase, Cox",
  "n14738052": "cyclooxygenase-1, Cox-1",
  "n14738295": "cyclooxygenase-2, Cox-2",
  "n14738521": "ptyalin",
  "n14738623": "rennet",
  "n14738752": "ferment",
  "n14738892": "substrate",
  "n14739004": "amine, aminoalkane",
  "n14739271": "azadirachtin",
  "n14739360": "carboxylic acid",
  "n14739734": "saccharic acid",
  "n14739861": "sebacic acid, decanedioic acid",
  "n14739978": "sorbic acid",
  "n14740094": "valeric acid, pentanoic acid",
  "n14740227": "fatty acid",
  "n14740587": "saturated fatty acid",
  "n14740915": "unsaturated fatty acid",
  "n14741124": "trans fatty acid",
  "n14741474": "monounsaturated fatty acid",
  "n14741730": "polyunsaturated fatty acid",
  "n14742028": "omega-3 fatty acid, omega-3",
  "n14742271": "omega-6 fatty acid, omega-6",
  "n14742476": "margaric acid, heptadecanoic acid",
  "n14742601": "ricinoleic acid",
  "n14742737": "fibrinopeptide",
  "n14742917": "polypeptide",
  "n14743046": "peptide",
  "n14743289": "aminobenzoic acid",
  "n14743381": "amino plastic, aminoplast, amino resin",
  "n14743582": "ammonia",
  "n14743727": "ammine",
  "n14743840": "ammonia water, ammonia, ammonium hydroxide",
  "n14743976": "ammoniac, gum ammoniac",
  "n14744120": "ammonium, ammonium ion",
  "n14744291": "ammonium carbamate",
  "n14744417": "ammonium carbonate",
  "n14744589": "ammonium chloride, sal ammoniac",
  "n14744698": "amyl alcohol",
  "n14744841": "phytohormone, plant hormone, growth regulator",
  "n14745057": "auxin",
  "n14745222": "gibberellin",
  "n14745368": "gibberellic acid",
  "n14745477": "kinin, cytokinin",
  "n14745635": "steroid hormone, steroid, sex hormone",
  "n14745891": "corticosterone",
  "n14746048": "progesterone, Lipo-Lutin",
  "n14746270": "megestrol, megestrol acetate",
  "n14746417": "norethindrone, norethindrone acetate, norethandrolone, Norlutin",
  "n14746793": "norethynodrel",
  "n14747007": "norgestrel",
  "n14747168": "medroxyprogesterone, Provera",
  "n14747338": "progestin, progestogen",
  "n14747587": "androgen, androgenic hormone",
  "n14747838": "adrenosterone",
  "n14747981": "androsterone",
  "n14748117": "methyltestosterone",
  "n14748335": "nandrolone, Durabolin, Kabolin",
  "n14748577": "testosterone",
  "n14748765": "follicle-stimulating hormone, FSH",
  "n14749030": "human chorionic gonadotropin, human chorionic gonadotrophin, HCG",
  "n14749272": "luteinizing hormone, LH, interstitial cell-stimulating hormone, ICSH",
  "n14749543": "prolactin, lactogenic hormone, luteotropin",
  "n14749794": "estrogen, oestrogen",
  "n14750122": "diethylstilbestrol, diethylstilboestrol, stilbestrol, stilboestrol, DES",
  "n14750316": "estradiol, oestradiol",
  "n14750622": "estriol, oestriol",
  "n14750782": "estrone, oestrone, theelin, Estronol",
  "n14751055": "hexestrol",
  "n14751216": "mestranol",
  "n14751417": "corticosteroid, corticoid, adrenal cortical steroid",
  "n14751863": "mineralocorticoid",
  "n14752057": "glucocorticoid",
  "n14752323": "glucosamine",
  "n14752511": "aldosterone",
  "n14752702": "hydrocortisone, cortisol, Hydrocortone, Cortef",
  "n14752952": "cortisone, Cortone Acetate",
  "n14753188": "prednisolone, Pediapred, Prelone",
  "n14753414": "prednisone, Orasone, Deltasone, Liquid Pred, Meticorten",
  "n14753808": "dexamethasone, Decadron, Dexamethasone Intensol, Dexone, Hexadrol, Oradexon",
  "n14754192": "spironolactone, Aldactone",
  "n14754378": "acid dye",
  "n14754505": "aniline, aniline oil, aminobenzine, phenylamine",
  "n14754705": "alizarin yellow",
  "n14754860": "anil, indigo, indigotin",
  "n14754985": "aniline dye",
  "n14755077": "animal oil",
  "n14755641": "drying oil",
  "n14755804": "animal material",
  "n14756039": "animal pigment",
  "n14756200": "arsine",
  "n14756369": "bilirubin, hematoidin, haematoidin",
  "n14756606": "urobilin",
  "n14756754": "urobilinogen, stercobilinogen",
  "n14757011": "luciferin",
  "n14757172": "melanin",
  "n14757382": "dentine, dentin",
  "n14757547": "ivory, tusk",
  "n14757754": "fluff",
  "n14757848": "bone, osseous tissue",
  "n14758027": "horn",
  "n14758252": "whalebone, baleen",
  "n14758420": "tortoiseshell",
  "n14758536": "shell",
  "n14758706": "mother-of-pearl, nacre",
  "n14758842": "animal skin",
  "n14759003": "parchment, sheepskin, lambskin",
  "n14759147": "vellum",
  "n14759275": "hide, fell",
  "n14759444": "cowhide",
  "n14759515": "goatskin",
  "n14759588": "rawhide",
  "n14759722": "leather",
  "n14760339": "grain",
  "n14760447": "alligator",
  "n14760555": "buckskin",
  "n14760696": "buff",
  "n14760813": "ooze leather",
  "n14760965": "Russia leather",
  "n14761122": "caffeine, caffein",
  "n14761334": "calf, calfskin",
  "n14761450": "white leather, whit leather",
  "n14761578": "cassava, cassava starch, manioc, manioca",
  "n14761806": "chamois, chamois leather, chammy, chammy leather, shammy, shammy leather",
  "n14762038": "wash leather",
  "n14762145": "cordovan",
  "n14762248": "cowhide, cowskin",
  "n14762366": "crushed leather, crush",
  "n14762486": "deerskin",
  "n14762572": "doeskin",
  "n14762664": "glove leather",
  "n14762758": "horsehide",
  "n14762846": "kid, kidskin",
  "n14762969": "mocha",
  "n14763059": "morocco",
  "n14763218": "Levant, Levant morocco",
  "n14763329": "patent leather",
  "n14763424": "pigskin",
  "n14763508": "sheepskin, fleece",
  "n14763674": "Golden Fleece",
  "n14763879": "shoe leather",
  "n14763964": "suede, suede leather",
  "n14764061": "fur, pelt",
  "n14764440": "astrakhan",
  "n14764518": "bearskin",
  "n14764617": "beaver, beaver fur",
  "n14764715": "chinchilla",
  "n14764820": "ermine",
  "n14764910": "fox",
  "n14765016": "lambskin",
  "n14765148": "broadtail",
  "n14765240": "Persian lamb",
  "n14765324": "lapin, rabbit",
  "n14765422": "leopard",
  "n14765497": "mink",
  "n14765575": "muskrat, muskrat fur",
  "n14765687": "hudson seal",
  "n14765785": "otter",
  "n14765856": "raccoon",
  "n14765946": "sable",
  "n14766040": "seal, sealskin",
  "n14766189": "squirrel",
  "n14766265": "anime, gum anime",
  "n14766364": "antifreeze",
  "n14766532": "nitric acid, aqua fortis",
  "n14766725": "nitrous acid",
  "n14766890": "nitrogen oxide",
  "n14767087": "nitrogen dioxide",
  "n14767202": "nitric oxide",
  "n14767291": "anhydride",
  "n14767477": "aqua regia, nitrohydrochloric acid",
  "n14767663": "aquamarine",
  "n14767796": "arginine",
  "n14767996": "aromatic hydrocarbon",
  "n14768201": "arsenic, arsenic trioxide, arsenous anhydride, arsenous oxide, white arsenic, ratsbane",
  "n14768480": "artificial blood",
  "n14768640": "acetic anhydride",
  "n14768765": "phthalic anhydride",
  "n14768854": "art paper",
  "n14768989": "asafetida, asafoetida",
  "n14769160": "ash",
  "n14769331": "fly ash",
  "n14769465": "asphalt",
  "n14769644": "mineral wool, rock wool",
  "n14769760": "austenite",
  "n14769962": "austenitic steel",
  "n14770146": "axle grease",
  "n14770248": "azide",
  "n14770400": "hydrazoite",
  "n14770500": "azo dye",
  "n14770631": "congo red",
  "n14770838": "gentian violet, crystal violet",
  "n14771088": "thiazine",
  "n14771270": "methylene blue, methylthionine chloride",
  "n14771482": "methyl orange",
  "n14771643": "phenothiazine, thiodiphenylamine",
  "n14771831": "diazonium",
  "n14772007": "Babbitt metal, babbitt",
  "n14772185": "bactericide, bacteriacide",
  "n14772317": "bagasse",
  "n14772461": "baking powder",
  "n14772594": "banana oil",
  "n14772716": "barbituric acid, malonylurea",
  "n14772877": "barium sulphate, barium sulfate, blanc fixe",
  "n14773022": "basalt",
  "n14773275": "basic dye, basic color, basic colour",
  "n14773447": "basic iron",
  "n14773618": "basic slag",
  "n14773787": "bath water",
  "n14773865": "battery acid, electrolyte acid",
  "n14773989": "bearing brass",
  "n14774090": "beebread, ambrosia",
  "n14774228": "royal jelly",
  "n14774397": "beef tallow",
  "n14774491": "beet sugar",
  "n14774575": "bell metal",
  "n14774699": "benzene, benzine, benzol",
  "n14774894": "benzene formula, benzene ring, benzene nucleus, Kekule formula",
  "n14775067": "benzoate",
  "n14775173": "benzoate of soda, sodium benzoate",
  "n14775321": "benzoic acid",
  "n14775430": "benzoyl peroxide",
  "n14775626": "beryllium bronze",
  "n14775729": "bicarbonate, hydrogen carbonate",
  "n14775995": "bicarbonate of soda, sodium hydrogen carbonate, sodium bicarbonate, baking soda, saleratus",
  "n14776237": "bimetal",
  "n14776391": "binder's board, binder board",
  "n14776523": "bitter principle",
  "n14776686": "black opal",
  "n14776804": "active agent, active",
  "n14776924": "Alka-seltzer",
  "n14777104": "Brioschi",
  "n14777188": "Bromo-seltzer",
  "n14777277": "lansoprazole, Prevacid",
  "n14777441": "Maalox",
  "n14777523": "Mylanta",
  "n14777606": "omeprazole, Prilosec",
  "n14777768": "Pepto-bismal",
  "n14777856": "Rolaids",
  "n14777939": "Tums",
  "n14778019": "antacid, gastric antacid, alkalizer, alkaliser, antiacid",
  "n14778436": "agent",
  "n14778982": "reagent",
  "n14779082": "bacteriostat",
  "n14779205": "bleaching agent, bleach, blanching agent, whitener",
  "n14779550": "chemical agent",
  "n14779796": "desiccant, drying agent, drier, siccative",
  "n14780040": "oxidant, oxidizer, oxidiser, oxidizing agent",
  "n14780267": "reducing agent, reducer, reductant",
  "n14780605": "bleaching clay, bleaching earth",
  "n14780737": "mud pie",
  "n14780850": "bleaching powder, chlorinated lime, chloride of lime",
  "n14781062": "bleach liquor",
  "n14781225": "hydrogen peroxide, peroxide",
  "n14781513": "blister copper",
  "n14781631": "bloodstone, heliotrope",
  "n14781752": "blotting paper, blotter",
  "n14781872": "blow gas, blowing gas",
  "n14781989": "blubber",
  "n14782150": "blueprint paper",
  "n14782252": "blue vitriol, blue copperas, blue stone, chalcanthite",
  "n14782403": "bog soil",
  "n14782529": "bond, bond paper",
  "n14782689": "bone ash",
  "n14782833": "bonemeal",
  "n14782919": "neem cake",
  "n14783038": "bone fat",
  "n14783251": "bone oil, Dippel's oil",
  "n14783436": "bone oil",
  "n14783588": "borate",
  "n14783670": "boric acid, boracic acid",
  "n14783808": "boric acid, orthoboric acid",
  "n14784111": "boron trifluoride",
  "n14784198": "borosilicate",
  "n14784290": "bouncing putty",
  "n14784506": "bowstring hemp",
  "n14784601": "box calf",
  "n14784705": "brewer's yeast",
  "n14784831": "bottom fermenting yeast",
  "n14784952": "top fermenting yeast",
  "n14785065": "bricks and mortar",
  "n14785197": "brushwood",
  "n14785325": "brimstone, native sulfur, native sulphur",
  "n14785436": "Britannia metal",
  "n14785524": "bromic acid",
  "n14785625": "bromide",
  "n14785840": "brownstone",
  "n14785941": "buffer",
  "n14786110": "buffer solution",
  "n14786257": "starting buffer",
  "n14786361": "phosphate buffer solution, PBS",
  "n14786479": "building material",
  "n14786943": "lagging",
  "n14787093": "butylene, butene",
  "n14787270": "isobutylene",
  "n14787386": "polybutylene, polybutene",
  "n14787520": "animal fat",
  "n14787707": "butterfat",
  "n14787816": "cabinet wood",
  "n14787954": "butanone, methyl ethyl ketone",
  "n14788200": "butyl alcohol, butanol",
  "n14788332": "butyric acid, butanoic acid",
  "n14788485": "butyrin",
  "n14788596": "tributyrin",
  "n14788714": "cacodyl, cacodyl group, cacodyl radical, arsenic group",
  "n14788875": "cacodyl, tetramethyldiarsine",
  "n14789087": "calcium carbide",
  "n14789220": "calcium-cyanamide, cyanamide",
  "n14789365": "calcium hypochlorite",
  "n14789504": "calcium lactate",
  "n14789724": "calcium nitrate",
  "n14789885": "calcium oxide, quicklime, lime, calx, calcined lime, fluxing lime, unslaked lime, burnt lime",
  "n14790137": "calcium phosphate",
  "n14790259": "calcium stearate, calcium octadecanoate",
  "n14790526": "carbonyl",
  "n14790693": "carbonyl group",
  "n14790796": "carboxyl, carboxyl group",
  "n14790979": "camphor",
  "n14791202": "camphor oil",
  "n14791292": "candelilla wax",
  "n14791453": "cane sugar",
  "n14791560": "cannabin, cannabis resin",
  "n14791722": "cannel coal",
  "n14791830": "capric acid, decanoic acid",
  "n14791988": "caproic acid, hexanoic acid",
  "n14792146": "caprylic acid",
  "n14792281": "carbamate",
  "n14792389": "carbamic acid",
  "n14792550": "carbide",
  "n14792703": "carbohydrate, saccharide, sugar",
  "n14793223": "Carboloy",
  "n14793393": "carbonado, black diamond",
  "n14793533": "carbon black, lampblack, soot, smut, crock",
  "n14793812": "carcinogen",
  "n14793921": "cellulose",
  "n14794203": "carboxymethyl cellulose",
  "n14794304": "diethylaminoethyl cellulose, DEAE cellulose",
  "n14794418": "pulp",
  "n14794534": "cartridge brass",
  "n14794693": "case-hardened steel",
  "n14794823": "cellulose ester",
  "n14794993": "cellulose nitrate, nitrocellulose, guncotton, nitrocotton",
  "n14795263": "collodion",
  "n14795432": "pyrocellulose",
  "n14795598": "pyroxylin, pyroxyline",
  "n14795749": "glycogen, animal starch",
  "n14795959": "inulin",
  "n14796073": "carbolic acid, phenol, hydroxybenzene, oxybenzene, phenylic acid",
  "n14796359": "activated carbon, activated charcoal",
  "n14796575": "graphite, black lead, plumbago",
  "n14796748": "pencil",
  "n14796969": "carbon dioxide, CO2, carbonic acid gas",
  "n14797264": "chokedamp, blackdamp",
  "n14797436": "carbon disulfide",
  "n14797641": "carbon monoxide, carbon monoxide gas, CO",
  "n14797813": "carbon paper, carbon",
  "n14798039": "carbon tetrachloride, carbon tet, tetrachloromethane, perchloromethane",
  "n14798331": "carbon tetrahalide",
  "n14798450": "carbonate",
  "n14798709": "fulminate",
  "n14798815": "mercury fulminate, fulminate of mercury, fulminating mercury",
  "n14799084": "carbonic acid",
  "n14799244": "abrasive, abradant, abrasive material",
  "n14799476": "carborundum",
  "n14799601": "cardboard, composition board",
  "n14799833": "cartridge paper",
  "n14799927": "cartridge paper",
  "n14800034": "card",
  "n14800138": "cocarboxylase, thiamine pyrophosphate",
  "n14800277": "coenzyme",
  "n14800539": "coenzyme A",
  "n14800706": "cofactor",
  "n14800842": "congener",
  "n14800979": "corrugated board, corrugated cardboard",
  "n14801146": "paperboard, poster board, posterboard",
  "n14801271": "pasteboard",
  "n14801402": "millboard",
  "n14801493": "strawboard",
  "n14801594": "carnelian, cornelian",
  "n14801711": "carrageenin, carrageenan",
  "n14801845": "ingot iron",
  "n14801921": "cast iron",
  "n14802167": "pot metal",
  "n14802262": "wrought iron",
  "n14802450": "steel",
  "n14802921": "stainless steel, stainless, chromium steel",
  "n14803074": "carbon steel",
  "n14803283": "crucible steel",
  "n14803438": "Damascus steel, Damask steel",
  "n14803578": "steel wool, wire wool",
  "n14803695": "cationic detergent, invert soap",
  "n14803886": "cat's eye",
  "n14804039": "cellulosic",
  "n14804175": "cement",
  "n14804487": "cement",
  "n14804641": "cement",
  "n14804797": "reinforced concrete, ferroconcrete",
  "n14804958": "hydraulic cement, Portland cement",
  "n14805145": "cementite, iron carbide",
  "n14805322": "ceresin",
  "n14805412": "cerulean blue",
  "n14805550": "cetrimide",
  "n14805676": "chad",
  "n14805899": "chaff, husk, shuck, stalk, straw, stubble",
  "n14806176": "bran",
  "n14806333": "chalcedony, calcedony",
  "n14806598": "chalk",
  "n14806728": "chamosite",
  "n14806838": "chemical, chemical substance",
  "n14807410": "neurochemical",
  "n14807558": "neurotransmitter",
  "n14807737": "monoamine neurotransmitter",
  "n14807929": "catecholamine",
  "n14808147": "chromophore",
  "n14808273": "serotonin, 5-hydroxytryptamine",
  "n14808418": "acetylcholine",
  "n14808617": "acyl anhydrides, acid anhydrides",
  "n14808752": "acyl halide, acid halide",
  "n14808911": "acetyl chloride, ethanoyl chloride",
  "n14809057": "endorphin",
  "n14809247": "beta endorphin",
  "n14809373": "enkephalin",
  "n14809521": "cheoplastic metal",
  "n14809666": "chernozemic soil",
  "n14809756": "chisel steel",
  "n14809843": "chlorine dioxide",
  "n14810032": "chlorine water",
  "n14810168": "chloropicrin, nitrochloroform",
  "n14810356": "nitrochloromethane",
  "n14810466": "chlorpyrifos",
  "n14810561": "choline",
  "n14810704": "chrome",
  "n14810854": "chrome-nickel steel, Elinvar",
  "n14811083": "chrome-tungsten steel",
  "n14811196": "chrome green",
  "n14811347": "Windsor green",
  "n14811435": "Hooker's green",
  "n14811556": "chrome yellow",
  "n14811706": "chromic acid",
  "n14811826": "chromate",
  "n14811932": "lead chromate",
  "n14812047": "chrysolite",
  "n14812233": "chrysoprase",
  "n14812359": "chylomicron",
  "n14812566": "cigarette paper, rolling paper",
  "n14812763": "cinder pig",
  "n14812872": "citric acid",
  "n14813067": "citrine",
  "n14813182": "clay",
  "n14813560": "pipeclay, terra alba",
  "n14813709": "bentonite",
  "n14813843": "fireclay",
  "n14813957": "Kitty Litter",
  "n14814125": "potter's clay, potter's earth",
  "n14814295": "slip",
  "n14814419": "claystone",
  "n14814531": "clunch",
  "n14814616": "coal",
  "n14814921": "anthracite, anthracite coal, hard coal",
  "n14815086": "bituminous coal, soft coal",
  "n14815263": "lignite, brown coal, wood coal",
  "n14815423": "sea coal",
  "n14815530": "steam coal",
  "n14815628": "Clorox",
  "n14815728": "coagulant, coagulator",
  "n14815867": "cod-liver oil, cod liver oil",
  "n14816064": "cod oil",
  "n14816181": "lanolin, wool fat, wool grease",
  "n14816401": "codon",
  "n14816613": "coin silver",
  "n14816745": "combustible, combustible material",
  "n14816899": "complementary DNA, cDNA",
  "n14817141": "provirus",
  "n14817346": "dscDNA",
  "n14817419": "episome",
  "n14817592": "complex, coordination compound",
  "n14817783": "composite material",
  "n14817964": "hydrochloride",
  "n14818101": "compost",
  "n14818238": "compound, chemical compound",
  "n14819856": "allomorph",
  "n14820052": "computer paper",
  "n14820180": "concrete",
  "n14820425": "conjugate solution, conjugate",
  "n14820641": "conima",
  "n14820747": "constantan, Eureka",
  "n14820933": "construction paper",
  "n14821043": "conductor",
  "n14821248": "semiconductor, semiconducting material",
  "n14821590": "insulator, dielectric, nonconductor",
  "n14821852": "glass wool",
  "n14821984": "contaminant, contamination",
  "n14822141": "coolant",
  "n14822563": "copper-base alloy",
  "n14822762": "copper oxide",
  "n14822839": "copper sulfate, copper sulphate, cupric sulfate, cupric sulphate",
  "n14823036": "coral, red coral, precious coral",
  "n14823227": "cork",
  "n14823376": "corkboard",
  "n14823547": "phellem, cork",
  "n14823690": "corn sugar",
  "n14823788": "corrosive",
  "n14823944": "alumina, aluminum oxide, aluminium oxide",
  "n14824140": "aluminate",
  "n14824238": "aluminum hydroxide, aluminium hydroxide, hydrated aluminum oxide, hydrated aluminium oxide",
  "n14824469": "alundum",
  "n14824556": "cottonseed cake, cotton cake",
  "n14824702": "coumarone-indene resin, coumarone resin",
  "n14824906": "indene",
  "n14825062": "covering material",
  "n14825243": "creatine, creatin",
  "n14825487": "creosol",
  "n14825631": "creosote, coal-tar creosote",
  "n14825812": "creosote",
  "n14825982": "cresol, methyl phenol",
  "n14826173": "crepe, crepe paper",
  "n14826328": "cryogen",
  "n14826482": "cyanamide, cyanamid",
  "n14826613": "cyanic acid",
  "n14826767": "cyanide",
  "n14826904": "sodium cyanide",
  "n14827017": "cyanogen",
  "n14827191": "cyano group, cyano radical, cyanide group, cyanide radical",
  "n14827346": "nitrile, nitril, cyanide",
  "n14827505": "cyanohydrin",
  "n14827680": "cyanuric acid",
  "n14827763": "cyclohexanol",
  "n14827871": "cyclohexanol phthalate",
  "n14827979": "cymene",
  "n14828193": "cytokine",
  "n14828396": "cytolysin",
  "n14828511": "cytosine, C",
  "n14828683": "daub",
  "n14828815": "decarboxylase",
  "n14828927": "defoliant",
  "n14829073": "de-iodinase",
  "n14829172": "demantoid",
  "n14829264": "andradite",
  "n14829470": "demerara",
  "n14829565": "deoxyadenosine monophosphate, A",
  "n14829765": "deoxycytidine monophosphate, C",
  "n14829964": "deoxyguanosine monophosphate, G",
  "n14830164": "deoxythymidine monophosphate, T",
  "n14830364": "deoxyribonucleic acid, desoxyribonucleic acid, DNA",
  "n14830992": "exon, coding DNA",
  "n14831178": "intron, noncoding DNA",
  "n14831338": "junk DNA",
  "n14831479": "recombinant deoxyribonucleic acid, recombinant DNA",
  "n14831665": "sticky end",
  "n14831812": "transposon, jumping gene",
  "n14832046": "ribonuclease, ribonucleinase, RNase",
  "n14832193": "ribonucleic acid, RNA",
  "n14832770": "messenger RNA, mRNA, template RNA, informational RNA",
  "n14833027": "nuclear RNA, nRNA",
  "n14833143": "transfer RNA, tRNA, acceptor RNA, soluble RNA",
  "n14833514": "deoxyribose",
  "n14833618": "dental gold",
  "n14833729": "depilatory",
  "n14833885": "derivative",
  "n14834013": "desert soil, desertic soil",
  "n14834132": "dew",
  "n14834325": "dextrin",
  "n14834563": "diamond, adamant",
  "n14834714": "digestive",
  "n14834807": "mono-iodotyrosine",
  "n14834906": "di-iodotyrosine",
  "n14835004": "iodotyrosine",
  "n14835125": "iodothyronine",
  "n14835230": "tri-iodothyronine",
  "n14835333": "dilutant, diluent, thinner",
  "n14835478": "dilution",
  "n14835569": "dimer",
  "n14835686": "dimethylglyoxime",
  "n14835817": "dimpled chad, pregnant chad, dimple",
  "n14835980": "diol, glycol, dihydric alcohol",
  "n14836127": "dioxide",
  "n14836308": "dioxin",
  "n14836468": "disaccharidase",
  "n14836642": "disaccharide",
  "n14836851": "dishwater",
  "n14836960": "distillate, distillation",
  "n14837258": "distilled water",
  "n14837364": "exhaust, exhaust fumes, fumes",
  "n14837506": "dressed ore, concentrate",
  "n14837678": "driftwood",
  "n14837786": "drill steel, drill rod",
  "n14837900": "docosahexaenoic acid",
  "n14838055": "dolomite",
  "n14838217": "dopamine, Dopastat, Intropin",
  "n14838539": "dottle",
  "n14838677": "dragon's blood",
  "n14838821": "drawing paper",
  "n14838951": "drilling mud, drilling fluid",
  "n14839206": "dubbin",
  "n14839322": "Duralumin",
  "n14839439": "particulate, particulate matter",
  "n14839698": "chalk dust",
  "n14839846": "dust",
  "n14840092": "dust",
  "n14840342": "elaidic acid",
  "n14840583": "elastomer",
  "n14840755": "element",
  "n14841056": "elixir",
  "n14841267": "air",
  "n14841770": "breath",
  "n14841923": "hot air",
  "n14842019": "halitosis",
  "n14842091": "halitus, exhalation",
  "n14842226": "compressed gas",
  "n14842378": "compressed air",
  "n14842560": "air cushion",
  "n14842703": "air",
  "n14842847": "fire",
  "n14842992": "earth, ground",
  "n14843295": "diatomaceous earth, diatomite, kieselguhr",
  "n14843493": "sienna",
  "n14843640": "bister, bistre",
  "n14843767": "burnt sienna",
  "n14843877": "raw sienna",
  "n14843986": "ocher, ochre",
  "n14844169": "sinopis, sinopia, sinoper",
  "n14844281": "yellow ocher, yellow ochre",
  "n14844414": "earth",
  "n14844560": "saprolite",
  "n14844693": "soil, dirt",
  "n14845383": "soil conditioner",
  "n14845578": "caliche, hardpan",
  "n14845743": "water, H2O",
  "n14846517": "holy water",
  "n14846646": "musk",
  "n14846779": "nectar",
  "n14846890": "pheromone",
  "n14847103": "quintessence, ether",
  "n14847357": "water",
  "n14847503": "ground water, spring water, well water",
  "n14847654": "eicosapentaenoic acid",
  "n14847810": "eleostearic acid, elaeostearic acid",
  "n14847929": "electrolyte",
  "n14848149": "eluate",
  "n14848245": "Fehling's solution",
  "n14848479": "formalin, formol",
  "n14848642": "gargle, mouthwash",
  "n14848785": "infusion, extract",
  "n14848995": "pancreatin",
  "n14849215": "injection, injectant",
  "n14849367": "isotonic solution, isosmotic solution",
  "n14849540": "elastase",
  "n14849655": "emerald",
  "n14849789": "emery cloth",
  "n14849880": "emery paper, sandpaper",
  "n14850051": "emery stone, emery rock",
  "n14850190": "enterokinase",
  "n14850334": "erythropoietin",
  "n14850483": "ester",
  "n14850826": "ethane, C2H6",
  "n14850948": "ethyl acetate",
  "n14851157": "ethylene, ethene",
  "n14851414": "trichloroethylene, trichloroethane, TCE",
  "n14851668": "ethylenediaminetetraacetic acid, EDTA",
  "n14851850": "ethylene glycol, glycol, ethanediol",
  "n14852003": "propylene glycol, propanediol",
  "n14852312": "euphorbium, gum eurphorbium",
  "n14852450": "discharge, emission",
  "n14852646": "emmenagogue",
  "n14852747": "eutectoid steel",
  "n14852913": "exudate, exudation",
  "n14853112": "transudate, transudation",
  "n14853210": "high explosive",
  "n14853392": "low explosive",
  "n14853530": "effluvium",
  "n14853647": "rheum",
  "n14853796": "vaginal discharge",
  "n14853947": "body waste, excretion, excreta, excrement, excretory product",
  "n14854262": "fecal matter, faecal matter, feces, faeces, BM, stool, ordure, dejection",
  "n14854581": "crap, dirt, shit, shite, poop, turd",
  "n14854764": "pigeon droppings",
  "n14854847": "droppings, dung, muck",
  "n14855066": "cow pie, cowpie",
  "n14855150": "meconium",
  "n14855280": "melena, melaena",
  "n14855428": "fecula",
  "n14855512": "wormcast",
  "n14855635": "human waste",
  "n14855724": "urine, piss, pee, piddle, weewee, water",
  "n14855992": "vomit, vomitus, puke, barf",
  "n14856134": "detritus",
  "n14856263": "waste, waste material, waste matter, waste product",
  "n14856752": "filth, crud, skank",
  "n14856893": "sewage, sewerage",
  "n14857021": "effluent, wastewater, sewer water",
  "n14857151": "garbage, refuse, food waste, scraps",
  "n14857278": "pollutant",
  "n14857497": "rubbish, trash, scrap",
  "n14857737": "scrap metal",
  "n14857897": "debris, dust, junk, rubble, detritus",
  "n14858099": "slack",
  "n14858292": "litter",
  "n14858451": "slop",
  "n14858637": "toxic waste, toxic industrial waste",
  "n14858794": "extravasation",
  "n14858950": "fallout, radioactive dust",
  "n14859100": "fencing material, fencing",
  "n14859201": "ferrite",
  "n14859344": "fertilizer, fertiliser, plant food",
  "n14859622": "gallamine, Flaxedil",
  "n14859838": "organic, organic fertilizer, organic fertiliser",
  "n14860102": "flux",
  "n14860280": "soldering flux",
  "n14860455": "foryml",
  "n14860603": "sodium nitrate, soda niter",
  "n14860749": "pearl ash",
  "n14860842": "potassium bicarbonate, potassium acid carbonate, potassium hydrogen carbonate",
  "n14861042": "potassium chloride, potassium muriate, potash muriate, K-Dur 20, Kaochlor, K-lor, Klorvess, K-lyte",
  "n14861355": "potassium nitrate, saltpeter, saltpetre, niter, nitre",
  "n14861566": "potassium bromide",
  "n14861716": "potassium carbonate",
  "n14861887": "potassium chlorate",
  "n14862064": "potassium cyanide",
  "n14862192": "potassium dichromate",
  "n14862316": "potassium iodide",
  "n14862481": "producer gas, air gas",
  "n14862640": "proprionamide, propanamide",
  "n14862753": "propionic acid, propanoic acid",
  "n14863031": "pudding stone, conglomerate",
  "n14863177": "putrescine",
  "n14863326": "pyroligneous acid, wood vinegar",
  "n14863521": "manure",
  "n14863788": "chicken manure",
  "n14863883": "cow manure",
  "n14863970": "green manure",
  "n14864079": "horse manure",
  "n14864170": "night soil",
  "n14864259": "facial tissue",
  "n14864360": "fat",
  "n14864731": "cocoa butter",
  "n14864825": "leaf fat, leaf lard",
  "n14864961": "feldspar, felspar",
  "n14865221": "orthoclase",
  "n14865316": "plagioclase, oligoclase",
  "n14865533": "albite, white feldspar",
  "n14865667": "anorthite",
  "n14865800": "ferric oxide",
  "n14865934": "ferricyanic acid",
  "n14866043": "ferricyanide",
  "n14866166": "ferritin",
  "n14866369": "ferrocerium",
  "n14866490": "ferrocyanic acid",
  "n14866605": "ferrocyanide",
  "n14866769": "fiberglass, fibreglass",
  "n14866889": "fiber, fibre",
  "n14867365": "string",
  "n14867545": "fish meal",
  "n14867690": "buntal",
  "n14867858": "fibril, filament, strand",
  "n14868116": "fieldstone",
  "n14868243": "filling, fill",
  "n14868464": "filter paper",
  "n14868564": "filtrate",
  "n14868771": "firelighter",
  "n14868916": "fire opal, girasol",
  "n14869035": "fish-liver oil, fish oil",
  "n14869177": "fixative",
  "n14869327": "fixing agent, fixer",
  "n14869513": "flavone",
  "n14869658": "flavonoid",
  "n14869829": "flax",
  "n14869975": "flyspeck",
  "n14870078": "cotton, cotton fiber, cotton wool",
  "n14870325": "long-staple cotton",
  "n14870424": "short-staple cotton",
  "n14870525": "chert",
  "n14870665": "taconite",
  "n14870821": "firestone",
  "n14870924": "flavin",
  "n14871078": "flint",
  "n14871268": "flintstone",
  "n14871370": "flooring",
  "n14871464": "floor wax",
  "n14871601": "fluoride",
  "n14871791": "fluoroboric acid",
  "n14871883": "fluoroboride",
  "n14871968": "fluorocarbon",
  "n14872226": "fluorocarbon plastic",
  "n14872325": "fluosilicate",
  "n14872408": "fluosilicic acid, hydrofluosilicic acid",
  "n14872572": "flypaper",
  "n14872697": "foam",
  "n14872875": "foam rubber",
  "n14873056": "fomentation",
  "n14873196": "formaldehyde, methanal",
  "n14873344": "formic acid",
  "n14873641": "formulation, preparation",
  "n14873951": "frankincense, olibanum, gum olibanum, thus",
  "n14874196": "free radical, radical",
  "n14874564": "freezing mixture",
  "n14874731": "Freon",
  "n14874932": "fructose, fruit sugar, levulose, laevulose",
  "n14875077": "fuel",
  "n14875707": "fuller's earth",
  "n14875878": "fulvic acid",
  "n14876066": "fumaric acid",
  "n14876206": "fumigant",
  "n14876318": "furan, furane, furfuran",
  "n14876473": "furfural, furfuraldehyde",
  "n14876679": "galactagogue",
  "n14876771": "galactose, brain sugar",
  "n14876872": "galbanum, gum albanum",
  "n14876993": "gallic acid",
  "n14877100": "galvanized iron",
  "n14877234": "greenhouse gas, greenhouse emission",
  "n14877486": "carbuncle",
  "n14877585": "gas",
  "n14878483": "liquefied petroleum gas, bottled gas",
  "n14878650": "water gas",
  "n14878844": "ghatti, ghatti gum",
  "n14878980": "kraft, kraft paper",
  "n14879115": "butcher paper",
  "n14879247": "gift wrap",
  "n14879356": "gilding metal",
  "n14879492": "gilgai soil",
  "n14879605": "natural glass",
  "n14879750": "quartz glass, quartz, vitreous silica, lechatelierite, crystal",
  "n14879996": "opal glass, milk glass",
  "n14880107": "optical glass",
  "n14880273": "optical crown, crown glass, optical crown glass",
  "n14880425": "optical flint, flint glass",
  "n14880557": "crown glass",
  "n14880690": "tektite",
  "n14880777": "volcanic glass",
  "n14880960": "obsidian",
  "n14881143": "pitchstone",
  "n14881224": "tachylite",
  "n14881303": "glass",
  "n14881759": "soft glass",
  "n14881861": "ground glass",
  "n14882000": "ground glass",
  "n14882117": "lead glass",
  "n14882248": "paste",
  "n14882370": "safety glass, laminated glass, shatterproof glass",
  "n14882578": "soluble glass, water glass, sodium silicate",
  "n14882793": "stained glass",
  "n14882935": "Tiffany glass",
  "n14883098": "wire glass",
  "n14883206": "crystal",
  "n14883517": "twins",
  "n14883661": "enamine",
  "n14883766": "enantiomorph, enantiomer",
  "n14883954": "exotherm",
  "n14884120": "glucose",
  "n14884336": "dextrose, dextroglucose, grape sugar",
  "n14884481": "blood sugar, blood glucose",
  "n14884581": "glutamate",
  "n14884669": "glutamic oxalacetic transaminase, glutamic oxaloacetic transaminase",
  "n14884820": "glyceraldehyde, glyceric aldehyde",
  "n14884963": "glyceric acid",
  "n14885088": "glyceride, acylglycerol",
  "n14885369": "triglyceride",
  "n14885684": "glycerol, glycerin, glycerine",
  "n14885857": "glycerinated gelatin",
  "n14886037": "glycerin jelly",
  "n14886180": "glycerite, glycerole",
  "n14886297": "glycerogelatin, glycerogel",
  "n14886443": "glyceryl",
  "n14886579": "nitroglycerin, nitroglycerine, trinitroglycerin, glyceryl trinitrate, Nitrospan, Nitrostat",
  "n14887026": "glyceryl ester",
  "n14887305": "glycoside",
  "n14887464": "amygdalin",
  "n14887633": "laetrile",
  "n14887801": "glucoside",
  "n14887925": "saponin",
  "n14888124": "glycolic acid, glycollic acid, hydroxyacetic acid",
  "n14888310": "glycoprotein",
  "n14888529": "cluster of differentiation 4, CD4",
  "n14888720": "cluster of differentiation 8, CD8",
  "n14888884": "hemoprotein, haemoprotein",
  "n14889074": "lectin",
  "n14889271": "gneiss",
  "n14889371": "schist",
  "n14889479": "rust",
  "n14889728": "goitrogen",
  "n14889857": "goldstone",
  "n14889973": "gondang wax, fig wax",
  "n14890099": "goose grease",
  "n14890183": "graph paper",
  "n14890286": "granite",
  "n14890485": "granular pearlite, globular pearlite",
  "n14890659": "lubricant, lubricator, lubricating substance, lube",
  "n14890945": "grease, lubricating oil",
  "n14891132": "greaseproof paper",
  "n14891255": "Greek fire",
  "n14891425": "green gold",
  "n14891581": "greisen",
  "n14891678": "groundmass",
  "n14891868": "grid metal",
  "n14891988": "grout",
  "n14892138": "guanine, G",
  "n14892289": "guano",
  "n14892405": "guinea gold",
  "n14892510": "gunite",
  "n14892655": "essential oil, volatile oil",
  "n14892989": "attar, atar, athar, ottar",
  "n14893128": "attar of roses, rose oil",
  "n14893271": "clove oil, oil of cloves",
  "n14893406": "costus oil",
  "n14893537": "eucalyptus oil",
  "n14893652": "turpentine, oil of turpentine, spirit of turpentine, turps",
  "n14893881": "wormwood oil, absinthe oil",
  "n14894029": "linalool",
  "n14894140": "resin, rosin",
  "n14894481": "natural resin",
  "n14894880": "amber",
  "n14895003": "urea-formaldehyde resin",
  "n14895189": "copal",
  "n14895373": "copalite, copaline, fossil copal",
  "n14895498": "congo copal, congo gum",
  "n14895599": "kauri, kauri copal, kauri resin, kauri gum",
  "n14895807": "dammar, gum dammar, damar, dammar resin",
  "n14896018": "Zanzibar copal, anime",
  "n14896128": "colophony",
  "n14896321": "mastic",
  "n14896441": "oleoresin",
  "n14896714": "balsam",
  "n14896923": "balm",
  "n14897083": "balm of Gilead",
  "n14897164": "Canada balsam",
  "n14897369": "turpentine, gum terpentine",
  "n14897521": "Chian turpentine",
  "n14897620": "copaiba, copaiba balsam, balsam capivi",
  "n14897751": "gum resin",
  "n14897941": "benzoin, gum benzoin, benjamin, gum benjamin, asa dulcis",
  "n14898101": "benzofuran, coumarone, cumarone",
  "n14898273": "bdellium",
  "n14898364": "gamboge",
  "n14898470": "gum",
  "n14899152": "medium",
  "n14899328": "culture medium, medium",
  "n14899530": "medium",
  "n14899687": "contrast medium, contrast material",
  "n14899888": "medium",
  "n14900008": "agar, agar-agar",
  "n14900184": "agar, nutrient agar",
  "n14900342": "blood agar",
  "n14900454": "algin, alginic acid",
  "n14900571": "cherry-tree gum",
  "n14900695": "chicle, chicle gum",
  "n14900817": "guar gum",
  "n14900963": "gum arabic, gum acacia",
  "n14901158": "Senegal gum",
  "n14901265": "gum butea, butea gum, butea kino, Bengal kino",
  "n14901411": "kino, gum kino, kino gum",
  "n14901565": "mesquite gum",
  "n14901679": "mucilage",
  "n14901793": "sterculia gum, karaya gum",
  "n14901959": "synthetic, synthetic substance",
  "n14902141": "synthetic resin",
  "n14902500": "alkyd, alkyd resin",
  "n14902627": "phenolic resin, phenolic, phenoplast",
  "n14902733": "epoxy, epoxy resin, epoxy glue",
  "n14902949": "copolymer",
  "n14903095": "polyurethane, polyurethan",
  "n14903339": "polyfoam, polyurethane foam",
  "n14903466": "cinnamon stone, essonite, hessonite",
  "n14903596": "gumbo, gumbo soil",
  "n14903763": "gutta-percha",
  "n14903942": "terra alba",
  "n14904052": "gummite",
  "n14904226": "halibut-liver oil",
  "n14904359": "halide",
  "n14904528": "halocarbon",
  "n14904661": "halogen",
  "n14904956": "hanging chad",
  "n14905076": "hard lead",
  "n14905197": "hard lead, antimonial lead",
  "n14905333": "hard steel",
  "n14905422": "hard water",
  "n14905607": "harlequin opal",
  "n14905720": "hematite, haematite",
  "n14905911": "hemiacetal",
  "n14906085": "hemlock",
  "n14906261": "hemolysin, haemolysin, erythrolysin, erythrocytolysin",
  "n14906500": "hemp",
  "n14906656": "heptane",
  "n14906850": "herbicide, weedkiller, weed killer",
  "n14907122": "hexane",
  "n14907273": "high brass",
  "n14907349": "high-density lipoprotein, HDL, alpha-lipoprotein",
  "n14907703": "high-level radioactive waste",
  "n14907867": "high-speed steel, hot-work steel",
  "n14908027": "hip tile, hipped tile",
  "n14908146": "histidine",
  "n14908297": "histaminase",
  "n14908422": "homogenate",
  "n14908584": "horsehair",
  "n14908683": "humectant",
  "n14908822": "humus",
  "n14908977": "humate",
  "n14909120": "humic acid",
  "n14909319": "humic substance",
  "n14909480": "humin",
  "n14909584": "hyacinth, jacinth",
  "n14909703": "hyaline, hyalin",
  "n14909895": "hyaluronic acid",
  "n14910165": "hyaluronidase, spreading factor, Hyazyme",
  "n14910440": "hydrate",
  "n14910581": "hydrazine",
  "n14910748": "hydride",
  "n14910926": "hydrobromic acid",
  "n14911057": "hydrocarbon",
  "n14911530": "bitumen",
  "n14911704": "pitch, tar",
  "n14911899": "coal tar",
  "n14912092": "butadiene",
  "n14912245": "chloroprene",
  "n14912387": "hydrochloric acid, chlorohydric acid",
  "n14912556": "hydrofluorocarbon, HFC",
  "n14912710": "hydrogen bromide",
  "n14912840": "hydrogen chloride",
  "n14912935": "hydrogen fluoride",
  "n14913136": "hydrofluoric acid",
  "n14913278": "hydrogen iodide",
  "n14913403": "hydroiodic acid",
  "n14913518": "hydrogen sulfide",
  "n14913633": "hyper-eutectoid steel",
  "n14913744": "hypnagogue",
  "n14913860": "hypo, sodium thiosulphate, sodium thiosulfate",
  "n14914014": "hypochlorous acid",
  "n14914191": "hypo-eutectoid steel",
  "n14914301": "hypoglycemic agent, hypoglycaemic agent",
  "n14914544": "hydrazo group, hydrazo radical",
  "n14914670": "hydroxide",
  "n14914775": "hydroxyl, hydroxyl group, hydroxyl radical",
  "n14914945": "hydroxide ion, hydroxyl ion",
  "n14915072": "hydroxymethyl",
  "n14915184": "ice, water ice",
  "n14915462": "black ice",
  "n14915622": "frost, hoar, hoarfrost, rime",
  "n14915804": "hailstone",
  "n14915928": "icicle",
  "n14916053": "Iceland spar",
  "n14916185": "identification particle",
  "n14916369": "Inconel",
  "n14916512": "ideal gas, perfect gas",
  "n14916670": "imidazole, iminazole, glyoxaline",
  "n14916796": "impregnation",
  "n14916962": "indelible ink",
  "n14917080": "India ink, drawing ink",
  "n14917208": "indicator",
  "n14917561": "indurated clay",
  "n14917635": "ink",
  "n14917859": "magnetic ink",
  "n14918023": "printer's ink, printing ink",
  "n14918162": "writing ink",
  "n14918264": "Indian red",
  "n14918380": "Indian red",
  "n14918529": "indoleacetic acid, IAA",
  "n14918654": "indolebutyric acid",
  "n14918784": "inducer",
  "n14918883": "ivory black",
  "n14918994": "incense",
  "n14919156": "inhalant",
  "n14919272": "inoculant, inoculum",
  "n14919511": "inorganic compound",
  "n14919635": "inosine",
  "n14919819": "inositol",
  "n14919948": "insecticide, insect powder",
  "n14920252": "insectifuge, insect repellent, insect repellant",
  "n14920388": "repellent, repellant",
  "n14920586": "repellent, repellant",
  "n14920728": "instillation",
  "n14920844": "insulating material, insulant, insulation",
  "n14921090": "interleukin",
  "n14921272": "intermediate",
  "n14921414": "Invar",
  "n14921622": "invertase, saccharase, sucrase",
  "n14921777": "invert sugar",
  "n14921974": "Javelle water, Javel water, eau de Javelle",
  "n14922107": "fraction",
  "n14922252": "iodic acid",
  "n14922371": "iodide",
  "n14922529": "iodocompound",
  "n14922637": "thyroprotein",
  "n14922779": "thyroglobulin",
  "n14922960": "iron blue, Prussian blue",
  "n14923060": "iron blue, steel grey, steel gray, Davy's grey, Davy's gray",
  "n14923207": "Payne's grey, Payne's gray",
  "n14923337": "iron disulfide",
  "n14923458": "iron ore",
  "n14923733": "iron perchloride",
  "n14923872": "isocyanate",
  "n14923962": "isocyanic acid",
  "n14924067": "isoleucine",
  "n14924188": "isomer",
  "n14924359": "isomerase",
  "n14924473": "itaconic acid",
  "n14924602": "jade, jadestone",
  "n14924786": "Japan wax, Japan tallow",
  "n14924915": "jargoon, jargon",
  "n14925033": "jasper",
  "n14925198": "jelly",
  "n14925378": "jet",
  "n14925526": "joss stick",
  "n14925645": "jute",
  "n14925776": "kapok, silk cotton, vegetable silk",
  "n14925945": "red silk cotton",
  "n14926086": "paraffin, paraffin oil",
  "n14926180": "keratohyalin",
  "n14926294": "ketone",
  "n14926595": "ketone body, acetone body",
  "n14926937": "ketone group",
  "n14927051": "acetoacetic acid",
  "n14927250": "beta-hydroxybutyric acid",
  "n14927451": "hydroxybutyric acid, oxybutyric acid",
  "n14927587": "ketohexose",
  "n14927739": "ketose",
  "n14927881": "kinase",
  "n14928008": "Kleenex",
  "n14928180": "kunzite",
  "n14928332": "Kwell",
  "n14928518": "labdanum, gum labdanum",
  "n14928729": "lacquer",
  "n14928885": "lactase, Lactaid",
  "n14929062": "lactic acid",
  "n14929209": "lactifuge",
  "n14929350": "lactogen",
  "n14929462": "lactose, milk sugar",
  "n14929662": "lamellar mixture",
  "n14929797": "lapis lazuli, lazuli",
  "n14929897": "lard oil",
  "n14930073": "larvicide",
  "n14930165": "laterite",
  "n14930312": "lath and plaster",
  "n14930476": "lauric acid, dodecanoic acid",
  "n14930670": "lauryl alcohol, 1-dodecanol",
  "n14930824": "latten",
  "n14930989": "lava",
  "n14931212": "tuff, tufa",
  "n14931323": "tufa, calc-tufa",
  "n14931472": "aa",
  "n14931559": "pahoehoe",
  "n14931633": "pillow lava",
  "n14931795": "magma",
  "n14931879": "igneous rock",
  "n14932222": "adesite",
  "n14932303": "batholith, batholite, pluton, plutonic rock",
  "n14932554": "diorite",
  "n14932645": "gabbro",
  "n14932741": "pegmatite",
  "n14932943": "peridotite",
  "n14933086": "kimberlite",
  "n14933236": "rhyolite",
  "n14933314": "volcanic rock",
  "n14933569": "leaded bronze",
  "n14933663": "lead ore",
  "n14933755": "massicot, massicotite",
  "n14933911": "leaf mold, leaf mould, leaf soil",
  "n14934031": "leaven, leavening",
  "n14934229": "ledger paper",
  "n14934353": "lepidocrocite",
  "n14934527": "laid paper",
  "n14934655": "wove paper",
  "n14934758": "letter paper",
  "n14934864": "lindane",
  "n14934998": "linen, linen paper",
  "n14935128": "leucine",
  "n14935319": "lignin",
  "n14935555": "hydroxide, hydrated oxide",
  "n14935751": "calcite",
  "n14936010": "calcium hydroxide, lime, slaked lime, hydrated lime, calcium hydrate, caustic lime, lime hydrate",
  "n14936226": "limestone",
  "n14936474": "rottenstone, tripoli",
  "n14936630": "dripstone",
  "n14936790": "calcium bicarbonate",
  "n14936905": "calcium carbonate",
  "n14937093": "limewater",
  "n14937225": "calcium chloride",
  "n14937366": "calcium hydride, hydrolith",
  "n14937521": "calcium sulphate, calcium sulfate",
  "n14937658": "limonene",
  "n14937801": "limonite",
  "n14937943": "linoleic acid, linolic acid",
  "n14938199": "linolenic acid",
  "n14938389": "linoleum, lino",
  "n14938467": "lint",
  "n14938578": "linuron",
  "n14938687": "lipase",
  "n14938907": "lipid, lipide, lipoid",
  "n14939230": "lipoprotein",
  "n14939445": "fluid",
  "n14939663": "grume",
  "n14939755": "ichor",
  "n14939900": "fluid",
  "n14940100": "liquid",
  "n14940386": "liquid",
  "n14940750": "liquid nitrogen",
  "n14940856": "liquid air",
  "n14940952": "liquid bleach",
  "n14941087": "liquid crystal",
  "n14941230": "liquor",
  "n14941407": "litmus, litmus test",
  "n14941642": "litmus paper",
  "n14941787": "lithia water",
  "n14941884": "lithium carbonate, Lithane, Lithonate, Eskalith",
  "n14942223": "loam",
  "n14942411": "lodestone, loadstone",
  "n14942631": "loess",
  "n14942762": "log",
  "n14942924": "low brass",
  "n14943009": "low-density lipoprotein, LDL, beta-lipoprotein",
  "n14943359": "low-level radioactive waste",
  "n14943580": "lumber, timber",
  "n14943814": "lye",
  "n14943950": "lymphokine",
  "n14944182": "lysine",
  "n14944317": "Lysol",
  "n14944455": "lysozyme, muramidase",
  "n14944617": "Mace, Chemical Mace",
  "n14944888": "macromolecule, supermolecule",
  "n14945137": "magnesium bicarbonate",
  "n14945254": "magnesium carbonate",
  "n14945399": "magnesium hydroxide",
  "n14945517": "magnesium nitride",
  "n14945624": "magnesium sulfate",
  "n14945725": "Epsom salts, bitter salts",
  "n14945914": "magnetite, magnetic iron-ore",
  "n14946082": "Malathion",
  "n14946251": "maleate",
  "n14946424": "maleic acid",
  "n14946618": "maltose, malt sugar",
  "n14946750": "manifold paper, manifold",
  "n14946939": "manila, manila paper, manilla, manilla paper",
  "n14947125": "manganese bronze, high-strength brass",
  "n14947255": "manganese steel, austenitic manganese steel",
  "n14947445": "manganate",
  "n14947558": "Manila hemp, Manilla hemp, abaca",
  "n14947702": "maple sugar",
  "n14947807": "marble",
  "n14948055": "verd antique, verde antique",
  "n14948156": "marking ink",
  "n14948265": "marsh gas",
  "n14948377": "martensite",
  "n14948645": "mash",
  "n14948812": "sour mash",
  "n14948983": "matchwood",
  "n14949075": "matchwood, splinters",
  "n14949227": "matrix",
  "n14949384": "matte",
  "n14949522": "medium steel",
  "n14949608": "megilp, magilp",
  "n14949746": "melamine, cyanuramide",
  "n14949886": "melamine resin",
  "n14950055": "meltwater",
  "n14950129": "menhaden oil",
  "n14950300": "menstruum",
  "n14950394": "menthol",
  "n14950694": "mercuric chloride, mercury chloride, bichloride of mercury, corrosive sublimate",
  "n14950937": "calomel, mercurous chloride",
  "n14951110": "message pad, writing pad",
  "n14951229": "methane",
  "n14951377": "methane series, alkane series, alkane, paraffin series, paraffin",
  "n14951682": "methyl bromide",
  "n14951814": "methylated spirit",
  "n14951983": "methylene group, methylene radical, methylene",
  "n14952122": "methyl, methyl group, methyl radical",
  "n14952290": "methionine",
  "n14952441": "methyl salicylate, birch oil, sweet-birch oil",
  "n14952654": "Microtaggant",
  "n14952878": "mild steel, low-carbon steel, soft-cast steel",
  "n14953005": "mine pig",
  "n14953091": "mineral oil",
  "n14953240": "misch metal",
  "n14953357": "mitogen",
  "n14953441": "motor oil",
  "n14953564": "mold, mould",
  "n14953674": "molybdenum steel",
  "n14953813": "monoamine",
  "n14953968": "monoamine oxidase, MAO",
  "n14954152": "monohydrate",
  "n14954284": "monosaccharide, monosaccharose, simple sugar",
  "n14954601": "monoxide",
  "n14954733": "montan wax",
  "n14954822": "moonstone",
  "n14955030": "mordant",
  "n14955246": "chrome alum",
  "n14955391": "tartar emetic, antimony potassium tartrate",
  "n14955559": "tartrate",
  "n14955682": "bitartrate",
  "n14955776": "morganite",
  "n14955889": "mortar",
  "n14956027": "mucoid",
  "n14956143": "mucopolysaccharide",
  "n14956325": "mud, clay",
  "n14956523": "slop, mire",
  "n14956661": "sludge, slime, goo, goop, gook, guck, gunk, muck, ooze",
  "n14956921": "sapropel",
  "n14957057": "muriatic acid",
  "n14957152": "music paper, score paper",
  "n14957270": "mustard gas, mustard agent, blistering agent, dichloroethyl sulfide, sulfur mustard",
  "n14957546": "muton",
  "n14957649": "nitrogen mustard",
  "n14957795": "mutton tallow",
  "n14957893": "myelin, myeline, medulla",
  "n14958102": "myristic acid, tetradecanoic acid",
  "n14958254": "napalm",
  "n14958405": "naphtha",
  "n14958564": "naphthalene",
  "n14958800": "naphthol",
  "n14958937": "pyrene",
  "n14959058": "man-made fiber, synthetic fiber",
  "n14959234": "natural fiber, natural fibre",
  "n14959472": "animal fiber, animal fibre",
  "n14959644": "plant fiber, plant fibre",
  "n14959939": "straw",
  "n14960090": "natural gas, gas",
  "n14960261": "naval brass, Admiralty brass, Admiralty Metal, Tobin bronze",
  "n14960457": "neat's-foot oil",
  "n14960606": "nebula",
  "n14960721": "nerve gas, nerve agent",
  "n14960963": "VX gas",
  "n14961375": "sarin, GB",
  "n14961512": "neuromuscular blocking agent",
  "n14961722": "newspaper, newsprint",
  "n14961896": "Nichrome",
  "n14962117": "nickel-base alloy, nickel alloy",
  "n14962291": "nickel bronze",
  "n14962387": "nickel silver, German silver",
  "n14962521": "nickel steel",
  "n14962612": "nicotinamide adenine dinucleotide, NAD",
  "n14962837": "nicotinamide adenine dinucleotide phosphate, NADP",
  "n14963051": "Ni-hard, Ni-hard iron",
  "n14963187": "Ni-resist, Ni-resist iron",
  "n14963317": "nitride",
  "n14963486": "nitrobenzene",
  "n14963634": "nitrofuran",
  "n14963762": "nitrogenase",
  "n14963916": "nuclease",
  "n14964129": "nucleic acid",
  "n14964367": "nucleoside",
  "n14964590": "nucleotide, base",
  "n14964925": "nurse log",
  "n14965135": "cellulose acetate",
  "n14965321": "cellulose triacetate, triacetate",
  "n14965501": "celluloid",
  "n14965780": "cellulose xanthate, viscose",
  "n14965958": "acrylic fiber, acrylic",
  "n14966058": "polyamide, polymeric amide",
  "n14966192": "nylon",
  "n14966334": "oakum",
  "n14966530": "octane",
  "n14966667": "oil",
  "n14967478": "fixed oil, fatty oil",
  "n14967577": "fusel oil",
  "n14967730": "gas oil",
  "n14967893": "stand oil",
  "n14968038": "neroli oil",
  "n14968185": "tall oil",
  "n14968308": "oil-hardened steel",
  "n14968402": "oilpaper",
  "n14968526": "oleic acid",
  "n14968795": "oleo oil",
  "n14968926": "oleoresin capiscum",
  "n14969044": "oligosaccharide",
  "n14969254": "onionskin, flimsy",
  "n14969421": "india paper",
  "n14969523": "onyx",
  "n14969666": "opaque gem",
  "n14969892": "opopanax",
  "n14969994": "organophosphate",
  "n14970136": "organophosphate nerve agent",
  "n14970538": "ormolu",
  "n14970645": "oxalacetate, oxaloacetate",
  "n14970752": "oxalacetic acid, oxaloacetic acid",
  "n14970920": "oxalate",
  "n14971022": "oxalic acid, ethanedioic acid",
  "n14971234": "oxidase",
  "n14971359": "oxidation-reduction indicator",
  "n14971519": "oxide",
  "n14971969": "oxidoreductase",
  "n14972092": "oxime",
  "n14972184": "oxyacetylene",
  "n14972359": "oxyacid, oxygen acid",
  "n14972582": "periodic acid",
  "n14972681": "oxygenase",
  "n14972807": "ozone",
  "n14973133": "pad, pad of paper, tablet",
  "n14973303": "palmitic acid, hexadecanoic acid",
  "n14973492": "palmitin",
  "n14973585": "pantothenic acid, pantothen",
  "n14973833": "papain",
  "n14973965": "para aminobenzoic acid, PABA",
  "n14974153": "paraquat",
  "n14974264": "paper",
  "n14975266": "paper tape",
  "n14975351": "paper toweling",
  "n14975444": "papier-mache, paper-mache",
  "n14975598": "papyrus",
  "n14975779": "parchment",
  "n14975872": "rice paper",
  "n14976001": "roofing paper, tar paper",
  "n14976170": "ola, olla",
  "n14976323": "ticker tape",
  "n14976448": "packing material, packing, wadding",
  "n14976634": "excelsior, wood shavings",
  "n14976759": "pantile",
  "n14976871": "blacktop, blacktopping",
  "n14977075": "macadam",
  "n14977188": "tarmacadam, tarmac",
  "n14977358": "parget, pargeting, pargetting",
  "n14977504": "paving, pavement, paving material",
  "n14977741": "pay dirt",
  "n14977845": "pearlite",
  "n14978114": "pectic acid",
  "n14978232": "pectin",
  "n14978453": "pediculicide",
  "n14978544": "penicillinase, beta-lactamase",
  "n14978729": "pepsin",
  "n14978887": "pepsinogen",
  "n14979052": "perboric acid",
  "n14979132": "perfluorocarbon, PFC",
  "n14979285": "Permalloy",
  "n14979410": "permanganate",
  "n14979588": "permanganic acid",
  "n14979730": "peroxidase",
  "n14979961": "peridot",
  "n14980087": "peroxide",
  "n14980215": "pesticide",
  "n14980468": "petrochemical",
  "n14980579": "petroleum, crude oil, crude, rock oil, fossil oil, oil",
  "n14980784": "residual oil, resid",
  "n14980910": "petrolatum, petroleum jelly, mineral jelly",
  "n14981183": "system",
  "n14981461": "phenolic plastic, phenolic urea",
  "n14981579": "phenylalanine",
  "n14981847": "phosgene",
  "n14981983": "phosphatase",
  "n14982129": "phosphine",
  "n14982265": "phosphate, orthophosphate, inorganic phosphate",
  "n14982421": "phosphocreatine, creatine phosphate, creatine phosphoric acid",
  "n14982681": "phospholipid",
  "n14982907": "phosphoric acid, orthophosphoric acid",
  "n14983036": "phthalic acid",
  "n14983143": "phytochemical",
  "n14983326": "picric acid",
  "n14983491": "pig iron",
  "n14983694": "pig lead",
  "n14983774": "plasmin, fibrinolysin",
  "n14983911": "plasminogen",
  "n14984066": "plasminogen activator, urokinase",
  "n14984244": "platinum black",
  "n14984378": "polymerase",
  "n14984584": "DNA polymerase",
  "n14984687": "transcriptase, RNA polymerase",
  "n14984799": "reverse transcriptase",
  "n14984973": "coloring material, colouring material, color, colour",
  "n14985383": "dye, dyestuff",
  "n14986004": "tincture",
  "n14986144": "argent",
  "n14986277": "alizarin, alizarine",
  "n14986438": "alizarin carmine, alizarin crimson, alizarin red",
  "n14986593": "bluing, blueing, blue",
  "n14986719": "bromophenol blue, bromphenol blue, tetrabromo-phenolsulfonephthalein",
  "n14986890": "bromothymol blue, bromthymol blue",
  "n14987025": "cochineal",
  "n14987150": "cyanine dye",
  "n14987341": "direct dye, substantive dye",
  "n14987484": "eosin, bromeosin",
  "n14987695": "fluorescein, fluoresceine, fluorescent dye, resorcinolphthalein",
  "n14988150": "fluorescein isothiocyanate, fluorescein isocyanate",
  "n14988337": "fluorochrome",
  "n14988502": "hair dye, hair coloring",
  "n14988635": "hematochrome",
  "n14988741": "henna",
  "n14988854": "rinse",
  "n14988961": "Kendal green, Kendal",
  "n14989106": "lac dye",
  "n14989243": "lead acetate, sugar of lead",
  "n14989430": "orchil, archil, cudbear",
  "n14989545": "phenol",
  "n14989820": "pigment",
  "n14990723": "pigment",
  "n14990864": "bole",
  "n14991004": "lake",
  "n14991106": "lake",
  "n14991210": "orange",
  "n14991319": "watercolor, water-color, watercolour, water-colour",
  "n14991476": "pine tar",
  "n14991611": "pisang wax",
  "n14991712": "plant material, plant substance",
  "n14991927": "plant product",
  "n14992165": "plasma",
  "n14992287": "plaster",
  "n14992613": "plaster of Paris, plaster",
  "n14992945": "puddle",
  "n14993137": "podzol, podzol soil, podsol, podsol soil, podsolic soil",
  "n14993378": "poison gas",
  "n14993559": "polyester",
  "n14993713": "polyester",
  "n14993868": "polyester fiber",
  "n14994004": "polysaccharide, polyose",
  "n14994328": "polymer",
  "n14994726": "polyphosphate",
  "n14994824": "polyunsaturated fat",
  "n14995061": "potassium ferrocyanide, yellow prussiate of potash",
  "n14995278": "potassium permanganate, permanganate of potash",
  "n14995541": "sandstone",
  "n14995812": "bluestone",
  "n14995918": "greensand",
  "n14996020": "polish",
  "n14996158": "polypropylene, polypropene",
  "n14996295": "polyvinyl-formaldehyde",
  "n14996395": "porphyry, porphyritic rock",
  "n14996579": "porpoise oil",
  "n14996709": "dolphin oil",
  "n14996843": "potash, caustic potash, potassium hydroxide",
  "n14997012": "powder, pulverization, pulverisation",
  "n14997274": "prairie soil",
  "n14997393": "precipitant",
  "n14997529": "preservative",
  "n14997699": "product",
  "n14997888": "percolate",
  "n14998042": "propanal, propionaldehyde",
  "n14998142": "propanol, propyl alcohol",
  "n14998290": "propenal, acrolein",
  "n14998421": "propenoate, acrylate",
  "n14998522": "propenoic acid, acrylic acid",
  "n14998677": "propenonitrile, acrylonitrile, vinyl cyanide",
  "n14998826": "propylene, propene",
  "n14998965": "propyl, propyl group, propyl radical",
  "n14999106": "protease, peptidase, proteinase, proteolytic enzyme",
  "n14999411": "ptomaine, ptomain",
  "n14999614": "Pyrex",
  "n14999801": "pyrimidine",
  "n14999913": "pyrimidine",
  "n15000060": "pyrope",
  "n15000149": "pyrophoric alloy",
  "n15000334": "pyruvic acid",
  "n15000475": "quassia",
  "n15000689": "quenched steel",
  "n15000858": "quercitron",
  "n15000973": "quinone, benzoquinone",
  "n15001212": "radiopaque dye",
  "n15001394": "rhodolite",
  "n15001500": "safranine, safranin, saffranine",
  "n15001640": "pheno-safranine",
  "n15001753": "Tyrian purple",
  "n15001887": "vat dye, vat color",
  "n15002138": "woad",
  "n15002233": "radioactive material",
  "n15002346": "radioactive waste",
  "n15002546": "raffia, raphia",
  "n15002679": "raffinose",
  "n15002814": "rauwolfia",
  "n15002959": "raveling, ravelling",
  "n15003139": "red brass, guinea gold",
  "n15003329": "red lead, minium",
  "n15003496": "red tide",
  "n15003645": "reductase",
  "n15003797": "refrigerant",
  "n15003969": "remover",
  "n15004149": "renin",
  "n15004317": "rennin, chymosin",
  "n15004501": "residue",
  "n15004715": "resorcinol",
  "n15004874": "restrainer",
  "n15005071": "restriction endonuclease, restriction nuclease, restriction enzyme",
  "n15005386": "retinene, retinal",
  "n15005577": "ridge tile",
  "n15005716": "roofing material",
  "n15005895": "rose quartz",
  "n15006012": "roughcast",
  "n15006118": "latex",
  "n15006258": "rubber, natural rubber, India rubber, gum elastic, caoutchouc",
  "n15006682": "crepe rubber",
  "n15006789": "rubber, synthetic rubber",
  "n15007016": "silicone rubber",
  "n15007190": "cold rubber",
  "n15007377": "neoprene",
  "n15007534": "hard rubber, vulcanite, ebonite",
  "n15007677": "Para rubber",
  "n15007803": "buna, buna rubber",
  "n15007917": "butyl rubber",
  "n15008025": "butyl",
  "n15008142": "ruby",
  "n15008287": "ruddle, reddle, raddle",
  "n15008399": "rutile",
  "n15008607": "rain, rainwater",
  "n15008765": "condensate",
  "n15008847": "seawater, saltwater, brine",
  "n15009192": "evaporite",
  "n15009326": "fresh water, freshwater",
  "n15009474": "Rochelle salt, Rochelle salts, potassium sodium tartrate",
  "n15009637": "Seidlitz powder, Seidlitz powders, Rochelle powder",
  "n15009843": "salicylate",
  "n15010038": "salicylic acid, 2-hydroxybenzoic acid",
  "n15010277": "salmon oil",
  "n15010430": "salol, phenyl salicylate",
  "n15010703": "salt",
  "n15011987": "double salt",
  "n15012214": "parathion",
  "n15012327": "Paris green",
  "n15012478": "rotenone",
  "n15012668": "samarskite",
  "n15012810": "sapphirine",
  "n15012999": "bile salt",
  "n15013139": "Glauber's salt, Glauber's salts",
  "n15013269": "cream of tartar, tartar, potassium bitartrate, potassium hydrogen tartrate",
  "n15013450": "sodium chlorate",
  "n15013577": "dichromic acid",
  "n15013764": "bichromate, dichromate",
  "n15013875": "sodium dichromate, sodium bichromate",
  "n15014012": "ammonium nitrate",
  "n15014170": "silver nitrate",
  "n15014406": "lunar caustic",
  "n15014529": "caustic",
  "n15014696": "caulk, caulking",
  "n15014845": "roan",
  "n15014990": "sodium hydroxide, caustic soda",
  "n15015192": "silver bromide",
  "n15015355": "silver iodide",
  "n15015501": "nitrate",
  "n15015872": "nitro group",
  "n15015963": "nitrite",
  "n15016123": "sodium nitrite",
  "n15016314": "gunpowder, powder",
  "n15016503": "smokeless powder, Ballistite",
  "n15016726": "microcosmic salt",
  "n15016852": "chloride",
  "n15016984": "trichloride",
  "n15017121": "nitrogen trichloride, Agene",
  "n15017343": "dichloride, bichloride",
  "n15017466": "perchloride",
  "n15017604": "chloride",
  "n15017878": "aluminum chloride, aluminium chloride",
  "n15018013": "methylene chloride, dichloromethane",
  "n15018169": "obidoxime chloride",
  "n15018318": "silver chloride",
  "n15018448": "stannic chloride",
  "n15018572": "stannous fluoride",
  "n15018688": "staple, staple fiber, staple fibre",
  "n15018889": "starch",
  "n15019030": "sand",
  "n15019267": "sangapenum, gum sangapenum",
  "n15019357": "water sapphire",
  "n15019483": "sapphire",
  "n15019645": "sarcosine",
  "n15019734": "sard, sardine, sardius",
  "n15019844": "sardine oil",
  "n15019972": "sardonyx",
  "n15020107": "sawdust",
  "n15020203": "saw log",
  "n15020297": "saxitoxin",
  "n15020528": "scale wax, paraffin scale",
  "n15020628": "scavenger",
  "n15020794": "scheelite",
  "n15020905": "schorl",
  "n15020974": "scrap iron",
  "n15021085": "notepad",
  "n15021189": "scratch pad, scratch paper, scribbling block",
  "n15021381": "seal oil",
  "n15021560": "secretase",
  "n15021824": "sedimentary clay",
  "n15021927": "sepia",
  "n15022035": "serine",
  "n15022171": "globulin",
  "n15022389": "gamma globulin, human gamma globulin",
  "n15022617": "myosin",
  "n15022776": "coagulation factor, clotting factor",
  "n15023156": "fibrinogen, factor I",
  "n15023294": "releasing factor, releasing hormone, RF",
  "n15023537": "growth hormone-releasing factor, GHRF",
  "n15023720": "intrinsic factor",
  "n15023959": "platelet, blood platelet, thrombocyte",
  "n15024124": "porphyrin",
  "n15024240": "hemoglobin, haemoglobin, Hb",
  "n15024606": "myoglobin",
  "n15024773": "oxyhemoglobin, oxyhaemoglobin",
  "n15024997": "heme, haem, hematin, haemitin, protoheme",
  "n15025228": "hemin, protohemin",
  "n15025397": "heterocyclic compound, heterocyclic, heterocycle",
  "n15025571": "cytochrome",
  "n15025815": "cytochrome c",
  "n15025942": "globin, hematohiston, haematohiston",
  "n15026155": "glutelin",
  "n15026256": "histone",
  "n15026420": "prolamine",
  "n15026508": "protamine",
  "n15026716": "scleroprotein, albuminoid",
  "n15026963": "hemosiderin, haemosiderin",
  "n15027189": "antibody",
  "n15027618": "autoantibody",
  "n15027762": "precipitin",
  "n15027888": "ABO antibodies",
  "n15027970": "Rh antibody",
  "n15028050": "antitoxin",
  "n15028208": "antivenin, antivenene",
  "n15028378": "tetanus antitoxin",
  "n15028555": "toxin antitoxin",
  "n15028684": "agglutinin",
  "n15028818": "isoagglutinin",
  "n15029002": "agglutinogen",
  "n15029147": "isoagglutinogen",
  "n15029292": "heterophil antibody, heterophile antibody, Forssman antibody",
  "n15029482": "isoantibody, alloantibody",
  "n15029646": "lysin",
  "n15029781": "monoclonal antibody, monoclonal",
  "n15030022": "infliximab, Remicade",
  "n15030264": "opsonin",
  "n15030481": "immunoglobulin, Ig, immune serum globulin, immune gamma globulin, immune globulin",
  "n15030853": "immunoglobulin A, IgA",
  "n15031073": "immunoglobulin D, IgD",
  "n15031231": "immunoglobulin E, IgE",
  "n15031418": "reagin",
  "n15031705": "immunoglobulin G, IgG",
  "n15031866": "immunoglobulin M, IgM",
  "n15032071": "tetanus immunoglobulin, tetanus immune globulin",
  "n15032376": "poison, toxicant, poisonous substance",
  "n15032661": "chemical irritant",
  "n15032829": "capsaicin",
  "n15033063": "gingerol",
  "n15033189": "piperin, piperine",
  "n15033367": "isothiocyanate",
  "n15033662": "fetoprotein, foetoprotein",
  "n15033846": "alpha fetoprotein, alpha foetoprotein, AFP",
  "n15034074": "toxin",
  "n15034454": "anatoxin, toxoid",
  "n15034741": "animal toxin, zootoxin",
  "n15034939": "bacterial toxin",
  "n15035123": "botulin, botulinus toxin, botulismotoxin",
  "n15035367": "cytotoxin",
  "n15035505": "endotoxin",
  "n15035697": "enterotoxin",
  "n15035832": "exotoxin",
  "n15035975": "mephitis",
  "n15036103": "hepatotoxin",
  "n15036211": "nephrotoxin",
  "n15036321": "neurotoxin, neurolysin",
  "n15036536": "mycotoxin",
  "n15036638": "plant toxin, phytotoxin",
  "n15036916": "venom",
  "n15037127": "kokoi venom",
  "n15037249": "snake venom",
  "n15037339": "antigen",
  "n15037664": "antigenic determinant, determinant, epitope",
  "n15037877": "rhesus factor, Rh factor, Rh",
  "n15038127": "sap",
  "n15038316": "scabicide",
  "n15038426": "sewer gas",
  "n15038520": "shale",
  "n15038680": "humic shale",
  "n15038789": "oil shale",
  "n15038892": "shale oil",
  "n15038987": "shark oil, shark-liver oil",
  "n15039179": "sheep dip",
  "n15039335": "Shetland wool",
  "n15039445": "shingle, shake",
  "n15039568": "shoe polish, blacking",
  "n15039703": "shot metal",
  "n15039827": "siderite, chalybite",
  "n15039935": "silicic acid",
  "n15040033": "silicate",
  "n15040129": "silicide",
  "n15040266": "silicon carbide",
  "n15040493": "silicone, silicone polymer",
  "n15040809": "silicone resin",
  "n15040899": "siloxane",
  "n15041050": "siding",
  "n15041190": "silex",
  "n15041277": "silica, silicon oxide, silicon dioxide",
  "n15041688": "silica gel",
  "n15041793": "silicon bronze",
  "n15041911": "silk",
  "n15042052": "silt",
  "n15042194": "siltstone",
  "n15042295": "silvex",
  "n15042429": "simazine",
  "n15042542": "Simoniz",
  "n15042654": "sisal, sisal hemp",
  "n15042772": "ski wax",
  "n15042856": "slag, scoria, dross",
  "n15043002": "slate, slating",
  "n15043118": "smaltite",
  "n15043308": "slush",
  "n15043399": "smelling salts",
  "n15043579": "snake oil",
  "n15043763": "snow",
  "n15044016": "snuff",
  "n15044122": "corn snow",
  "n15044232": "crud",
  "n15044327": "soapstone, soaprock, soap-rock, steatite",
  "n15044523": "soda lime",
  "n15044648": "sodalite",
  "n15044844": "sodium carbonate, washing soda, sal soda, soda ash, soda",
  "n15045030": "sodium carboxymethyl cellulose",
  "n15045216": "sodium fluoride",
  "n15045379": "sodium hydride",
  "n15045490": "sodium hypochlorite",
  "n15045677": "sodium iodide",
  "n15045782": "sodium lauryl sulphate, sodium lauryl sulfate, SLS",
  "n15046077": "sodium pyrophosphate, tetrasodium pyrophosphate",
  "n15046250": "sodium sulphate, sodium sulfate",
  "n15046421": "sodium tripolyphosphate",
  "n15046568": "sodium phosphate, sodium orthophosphate",
  "n15046716": "soft water",
  "n15046900": "solid",
  "n15047167": "dry ice",
  "n15047313": "solvent, dissolvent, dissolver, dissolving agent, resolvent",
  "n15047849": "solute",
  "n15047988": "solvate",
  "n15048184": "solvating agent",
  "n15048293": "viricide, virucide",
  "n15048463": "alkahest, alcahest, universal solvent",
  "n15048623": "soup",
  "n15048748": "sourdough",
  "n15048888": "spackle, spackling compound",
  "n15049096": "spar",
  "n15049309": "sparkle metal",
  "n15049401": "spiegeleisen, spiegel, spiegel iron",
  "n15049594": "spill",
  "n15049745": "spelter",
  "n15049902": "sperm oil",
  "n15050011": "spice",
  "n15050161": "stacte",
  "n15050320": "staff",
  "n15050516": "staphylococcal enterotoxin",
  "n15050695": "staphylococcal enterotoxin B, SEB",
  "n15050898": "spinel",
  "n15051129": "spinel ruby, ruby spinel",
  "n15051301": "almandine",
  "n15051392": "balas, balas ruby",
  "n15051503": "Ceylonite, pleonaste",
  "n15051608": "rubicelle",
  "n15051705": "solid solution, primary solid solution",
  "n15052063": "spirits of ammonia, sal volatile",
  "n15052223": "spodumene",
  "n15052413": "hiddenite",
  "n15052555": "spray",
  "n15052667": "stabilizer",
  "n15052852": "stachyose",
  "n15052970": "stain",
  "n15053212": "counterstain",
  "n15053373": "Gram's solution",
  "n15053568": "stannite, tin pyrites",
  "n15053703": "star sapphire",
  "n15053867": "starch, amylum",
  "n15054347": "arrowroot",
  "n15054467": "cornstarch, cornflour",
  "n15054607": "sago",
  "n15054770": "pearl sago",
  "n15054862": "amyloid",
  "n15055046": "Otaheite arrowroot, Otaheite arrowroot starch",
  "n15055181": "steam",
  "n15055342": "live steam",
  "n15055442": "water vapor, water vapour",
  "n15055633": "vapor, vapour",
  "n15055936": "softener",
  "n15056112": "water softener",
  "n15056372": "soman, GD",
  "n15056541": "spray",
  "n15056749": "sea spray",
  "n15056827": "spindrift, spoondrift",
  "n15056938": "stearic acid, octadecanoic acid",
  "n15057103": "stearin",
  "n15057212": "Stellite",
  "n15057433": "sterling silver",
  "n15057558": "sternutator, sternutatory",
  "n15057744": "steroid",
  "n15058023": "nonsteroid, nonsteroidal",
  "n15058163": "ketosteroid",
  "n15058310": "sterol, steroid alcohol",
  "n15058544": "cholesterol, cholesterin",
  "n15058755": "HDL cholesterol",
  "n15058987": "LDL cholesterol",
  "n15059237": "oxidized LDL cholesterol",
  "n15059404": "ergosterol",
  "n15059552": "bile acid",
  "n15059694": "cholic acid",
  "n15059797": "bilge, bilge water",
  "n15059939": "cardiac glycoside, cardiac glucoside",
  "n15060131": "digitalis, digitalis glycoside, digitalin",
  "n15060326": "render",
  "n15060449": "stibnite",
  "n15060569": "sticks and stone",
  "n15060688": "wattle and daub",
  "n15060825": "stiffener",
  "n15060937": "streptodornase",
  "n15061171": "streptokinase",
  "n15061428": "streptolysin",
  "n15061550": "stripper",
  "n15061674": "strophanthin",
  "n15061936": "strontianite",
  "n15062057": "stucco",
  "n15062284": "sublimate",
  "n15062468": "tallow",
  "n15062641": "vegetable tallow",
  "n15062778": "sucrose, saccharose",
  "n15062955": "jaggery, jagghery, jaggary",
  "n15063108": "structural iron",
  "n15063224": "structural steel",
  "n15063358": "sulfanilic acid, sulphanilic acid",
  "n15063493": "sulfate, sulphate",
  "n15063699": "sulfide, sulphide",
  "n15063946": "sulfur oxide, sulphur oxide",
  "n15064053": "sulfur dioxide, sulphur dioxide",
  "n15064327": "sulfur hexafluoride, sulphur hexafluoride",
  "n15064560": "sunstone, aventurine",
  "n15064729": "superoxide, superoxide anion",
  "n15064917": "superoxide",
  "n15065025": "superoxide dismutase, SOD",
  "n15065280": "surgical spirit",
  "n15065483": "Swedish iron",
  "n15065584": "swinging chad",
  "n15065713": "sylvanite, graphic tellurium",
  "n15065928": "sylvite, sylvine",
  "n15066125": "tabun, GA",
  "n15066367": "talc, talcum",
  "n15066666": "French chalk",
  "n15066796": "rensselaerite",
  "n15066906": "tallow oil",
  "n15067025": "tannin, tannic acid",
  "n15067237": "catechin",
  "n15067375": "tantalite",
  "n15067576": "tartaric acid",
  "n15067739": "racemic acid",
  "n15067877": "tear gas, teargas, lacrimator, lachrymator",
  "n15068134": "telluride",
  "n15068282": "telomerase",
  "n15068436": "tenderizer, tenderiser",
  "n15068635": "terpene",
  "n15068754": "tetrachloride",
  "n15068891": "tetrafluoroethylene",
  "n15069046": "tetrahalide",
  "n15069161": "tetrasaccharide",
  "n15069338": "tetrodotoxin",
  "n15069472": "tetroxide",
  "n15069584": "tetryl, nitramine",
  "n15069712": "thatch",
  "n15069820": "thickening, thickener",
  "n15069998": "thiouracil",
  "n15070098": "thiocyanate",
  "n15070234": "thiocyanic acid",
  "n15070366": "thorite",
  "n15070550": "thortveitite",
  "n15070699": "threonine",
  "n15070897": "prothrombin, factor II",
  "n15071035": "thromboplastin, thrombokinase, factor III",
  "n15071229": "calcium ion, factor IV",
  "n15071366": "proaccelerin, prothrombin accelerator, accelerator factor, factor V",
  "n15071503": "proconvertin, cothromboplastin, stable factor, factor VII",
  "n15071684": "antihemophilic factor, antihaemophilic factor, antihemophilic globulin, antihaemophilic globulin, factor VIII, Hemofil",
  "n15071960": "Christmas factor, factor IX",
  "n15072099": "prothrombinase, factor X",
  "n15072331": "plasma thromboplastin antecedent, factor XI",
  "n15072491": "Hageman factor, factor XII",
  "n15072657": "fibrinase, factor XIII",
  "n15072857": "thymine, T",
  "n15073018": "deoxyadenosine",
  "n15073147": "deoxycytidine, cytidine",
  "n15073285": "deoxyguanosine, guanosine",
  "n15073424": "deoxythymidine, thymidine",
  "n15073563": "thymol, thyme camphor, thymic acid",
  "n15073784": "melanocyte-stimulating hormone, MSH",
  "n15073973": "thyrotropin, thyrotropic hormone, thyrotrophin, thyrotrophic hormone, thyroid-stimulating hormone, TSH",
  "n15074203": "thyrotropin-releasing hormone, TRH, thyrotropin-releasing factor, TRF, protirelin",
  "n15074456": "thyronine",
  "n15074568": "tile, roofing tile",
  "n15074772": "till, boulder clay",
  "n15074962": "tissue, tissue paper",
  "n15075141": "toilet tissue, toilet paper, bathroom tissue",
  "n15075298": "toilet roll",
  "n15075378": "toluene, methylbenzene",
  "n15075596": "toluic acid",
  "n15075691": "tombac, tombak, tambac",
  "n15075867": "toner",
  "n15076011": "toner",
  "n15076180": "tool steel",
  "n15076349": "topaz",
  "n15076523": "topaz, false topaz, common topaz",
  "n15076619": "tourmaline",
  "n15076931": "trace element",
  "n15077109": "tracing paper",
  "n15077226": "tragacanth",
  "n15077338": "transaminase, aminotransferase, aminopherase",
  "n15077571": "transferase",
  "n15077752": "transfer paper",
  "n15077917": "transferrin, beta globulin, siderophilin",
  "n15078050": "transparent gem",
  "n15078550": "transparent substance, translucent substance",
  "n15078768": "triamcinolone, Aristocort, Aristopak, Kenalog",
  "n15079030": "triazine",
  "n15079184": "tri-chad",
  "n15079310": "trichloroacetic acid, trichloracetic acid",
  "n15079445": "margarin, glycerol trimargarate",
  "n15079557": "tridymite",
  "n15079656": "triolein, olein",
  "n15079795": "trimer",
  "n15079925": "trioxide",
  "n15080055": "tripalmitin, glycerol tripalmitate",
  "n15080168": "triphosphopyridine",
  "n15080277": "triphosphopyridine nucleotide",
  "n15080382": "triphosphoric acid",
  "n15080571": "trisaccharide",
  "n15080747": "trisodium phosphate, trisodium orthophosphate, tribasic sodium phosphate",
  "n15080942": "tristearin, glycerol tristearate",
  "n15081052": "trypsin",
  "n15081221": "trypsinogen",
  "n15081417": "tryptophan, tryptophane",
  "n15081595": "tuna oil",
  "n15081675": "tundra soil",
  "n15081828": "tungsten steel, wolfram steel",
  "n15081957": "tungstate",
  "n15082036": "tungstic acid",
  "n15082198": "turquoise",
  "n15082382": "typewriter paper, typing paper",
  "n15082524": "tyramine",
  "n15082766": "tyrosine",
  "n15082890": "ubiquinone, coenzyme Q",
  "n15083146": "ultramarine, ultramarine blue",
  "n15083283": "French blue, French ultramarine, French ultramarine blue",
  "n15083429": "umber",
  "n15083533": "raw umber",
  "n15083611": "burnt umber",
  "n15083713": "undecylenic acid",
  "n15083819": "unleaded gasoline, unleaded petrol",
  "n15083974": "undercut",
  "n15084110": "urease",
  "n15084277": "urethane",
  "n15084357": "uracil, U",
  "n15084545": "uraninite, pitchblende",
  "n15084824": "uranium ore",
  "n15084999": "uranyl, uranyl group, uranyl radical",
  "n15085175": "uranyl nitrate",
  "n15085327": "uranyl oxalate",
  "n15085472": "urea, carbamide",
  "n15085682": "uric acid",
  "n15085836": "urate",
  "n15085925": "tophus, chalkstone",
  "n15086088": "valine",
  "n15086247": "linseed, flaxseed",
  "n15086352": "linseed oil, flaxseed oil",
  "n15086545": "tung oil, Chinese wood oil",
  "n15086672": "chaulmoogra oil",
  "n15086811": "vanadate",
  "n15086939": "vanadinite",
  "n15087090": "vanadium pentoxide, vanadic acid",
  "n15087244": "vanadium steel",
  "n15087384": "vellum",
  "n15087486": "vermiculite",
  "n15087632": "very low density lipoprotein, VLDL",
  "n15087904": "vesuvianite, vesuvian, idocrase",
  "n15088111": "vinegar",
  "n15088183": "vinyl",
  "n15088308": "vinyl, vinyl group, vinyl radical",
  "n15088440": "vinyl polymer, vinyl resin, polyvinyl resin",
  "n15088669": "iodopsin",
  "n15088869": "visual purple, rhodopsin, retinal purple",
  "n15089104": "photopigment",
  "n15089258": "vitamin",
  "n15089472": "fat-soluble vitamin",
  "n15089645": "water-soluble vitamin",
  "n15089803": "vitamin A, antiophthalmic factor, axerophthol, A",
  "n15090065": "vitamin A1, retinol",
  "n15090238": "vitamin A2, dehydroretinol",
  "n15090379": "provitamin",
  "n15090535": "provitamin A, carotene, carotin",
  "n15090742": "B-complex vitamin, B complex, vitamin B complex, vitamin B, B vitamin, B",
  "n15091129": "vitamin B1, thiamine, thiamin, aneurin, antiberiberi factor",
  "n15091304": "vitamin B12, cobalamin, cyanocobalamin, antipernicious anemia factor",
  "n15091473": "vitamin B2, vitamin G, riboflavin, lactoflavin, ovoflavin, hepatoflavin",
  "n15091669": "vitamin B6, pyridoxine, pyridoxal, pyridoxamine, adermin",
  "n15091846": "vitamin Bc, vitamin M, folate, folic acid, folacin, pteroylglutamic acid, pteroylmonoglutamic acid",
  "n15092059": "niacin, nicotinic acid",
  "n15092227": "vitamin D, calciferol, viosterol, ergocalciferol, cholecalciferol, D",
  "n15092409": "vitamin E, tocopherol, E",
  "n15092650": "biotin, vitamin H",
  "n15092751": "vitamin K, naphthoquinone, antihemorrhagic factor",
  "n15092942": "vitamin K1, phylloquinone, phytonadione",
  "n15093049": "vitamin K3, menadione",
  "n15093137": "vitamin P, bioflavinoid, citrin",
  "n15093298": "vitamin C, C, ascorbic acid",
  "n15093482": "vitriol, oil of vitriol, sulfuric acid, sulphuric acid",
  "n15093735": "volatile",
  "n15093938": "wallpaper",
  "n15094053": "waste paper",
  "n15094136": "water of crystallization, water of crystallisation, water of hydration",
  "n15094294": "wax",
  "n15094824": "beeswax",
  "n15095014": "Ghedda wax",
  "n15095103": "cerumen, earwax",
  "n15095237": "paraffin, paraffin wax",
  "n15095389": "spermaceti",
  "n15095496": "vegetable wax",
  "n15095639": "shellac wax, lac wax",
  "n15095771": "cadaverine",
  "n15095926": "cadmium sulfide",
  "n15096048": "cadmium yellow",
  "n15096233": "cadmium yellow pale",
  "n15096331": "cadmium orange",
  "n15096426": "zinc cadmium sulfide",
  "n15096524": "verdigris, cupric acetate",
  "n15096661": "wax paper",
  "n15096783": "wetting agent, wetter, surfactant, surface-active agent",
  "n15097017": "detergent",
  "n15097209": "builder, detergent builder",
  "n15097430": "whale oil, train oil",
  "n15097578": "whey, milk whey",
  "n15097722": "white lead, ceruse, lead carbonate",
  "n15097849": "wicker",
  "n15097994": "wiesenboden",
  "n15098161": "wood",
  "n15100023": "dyewood",
  "n15100112": "hardwood",
  "n15100257": "softwood, deal",
  "n15100397": "pulpwood",
  "n15100479": "raw wood",
  "n15100570": "hardtack",
  "n15100644": "firewood",
  "n15100834": "cordwood",
  "n15100945": "backlog",
  "n15101059": "Yule log",
  "n15101157": "brand, firebrand",
  "n15101271": "pine knot",
  "n15101361": "igniter, ignitor, lighter",
  "n15101586": "kindling, tinder, touchwood, spunk, punk",
  "n15101723": "punk",
  "n15101854": "board, plank",
  "n15102167": "lemongrass, lemon grass, lemongrass oil",
  "n15102359": "planking",
  "n15102455": "chipboard, hardboard",
  "n15102622": "deal",
  "n15102712": "knot",
  "n15102894": "knothole",
  "n15103007": "clapboard, weatherboard, weatherboarding",
  "n15103226": "wolframite, iron manganese tungsten",
  "n15103485": "wollastonite",
  "n15103634": "wood pulp",
  "n15103780": "wood sugar, xylose",
  "n15103911": "Wood's metal, Wood's alloy",
  "n15104088": "wood tar",
  "n15104217": "wool",
  "n15104488": "raw wool",
  "n15104598": "alpaca",
  "n15104669": "cashmere",
  "n15104752": "fleece",
  "n15104860": "shoddy",
  "n15104933": "vicuna",
  "n15105008": "virgin wool",
  "n15105122": "wrapping paper",
  "n15105268": "writing paper",
  "n15105638": "wool oil",
  "n15105733": "wulfenite",
  "n15105849": "wurtzite",
  "n15105955": "xenotime",
  "n15106143": "xylene, xylol",
  "n15106271": "yeast, barm",
  "n15106529": "yellowcake, U308",
  "n15106674": "mother",
  "n15106867": "zeolite",
  "n15107196": "chabazite, chabasite",
  "n15107362": "heulandite",
  "n15107524": "natrolite",
  "n15107677": "phillipsite",
  "n15107876": "zinc blende, blende, sphalerite",
  "n15108087": "zinc oxide, flowers of zinc, philosopher's wool, philosophers' wool",
  "n15108324": "zinc sulfate, zinc sulphate, white vitriol, zinc vitriol",
  "n15108538": "zinc sulfide, zinc sulphide",
  "n15108745": "zinc white, Chinese white",
  "n15108897": "zinkenite",
  "n15109016": "zinnwaldite",
  "n15109127": "zircon, zirconium silicate",
  "n15109391": "zirconium oxide, zirconia, zirconium dioxide",
  "n15109586": "zymase",
  "n15109745": "emanation",
  "n15109931": "ectoplasm",
  "n15110096": "essence",
  "n15110307": "ligand",
  "n15110454": "enamel",
  "n15110666": "imide",
  "n15110796": "metabolite",
  "n15110956": "vegetable matter",
  "n15111116": "anabolic steroid",
  "n15111437": "pregnanediol",
  "n15111609": "tubocurarine, curare",
  "n15111903": "tuberculin",
  "n15112086": "vehicle",
  "n15112239": "vesicant, vesicatory",
  "n15112425": "vernix, vernix caseosa",
  "n15112565": "vitrification",
  "n15112702": "wad",
  "n15112828": "xanthate",
  "n15112932": "xanthic acid",
  "n15113050": "xanthine",
  "n15113229": "time period, period of time, period",
  "n15115926": "trial period, test period",
  "n15116095": "time frame",
  "n15116283": "geological time, geologic time",
  "n15116532": "biological time",
  "n15116724": "cosmic time",
  "n15116910": "civil time, standard time, local time",
  "n15117246": "daylight-saving time, daylight-savings time, daylight saving, daylight savings",
  "n15117516": "hours",
  "n15117621": "downtime",
  "n15117809": "uptime",
  "n15118011": "24/7",
  "n15118100": "hours",
  "n15118228": "work time",
  "n15118453": "time off",
  "n15118724": "face time",
  "n15118935": "compensatory time",
  "n15119072": "bout",
  "n15119185": "hospitalization",
  "n15119374": "travel time",
  "n15119536": "present, nowadays",
  "n15119919": "now",
  "n15120050": "here and now, present moment, moment",
  "n15120223": "date",
  "n15120346": "times",
  "n15120528": "modern times, present times, modern world, contemporary world",
  "n15120715": "Roman times",
  "n15120823": "past, past times, yesteryear",
  "n15121088": "yore",
  "n15121153": "bygone, water under the bridge",
  "n15121302": "old",
  "n15121406": "history",
  "n15121625": "future, hereafter, futurity, time to come",
  "n15121880": "kingdom come",
  "n15122011": "musical time",
  "n15122231": "time",
  "n15122648": "Elizabethan age",
  "n15122853": "Victorian age",
  "n15123115": "day",
  "n15123362": "dead",
  "n15123512": "hard times",
  "n15123589": "incarnation",
  "n15123754": "continuum",
  "n15123996": "history",
  "n15124183": "Phanerozoic, Phanerozoic eon, Phanerozoic aeon",
  "n15124361": "Cenozoic, Cenozoic era, Age of Mammals",
  "n15124545": "Quaternary, Quaternary period, Age of Man",
  "n15124713": "Holocene, Holocene epoch, Recent, Recent epoch",
  "n15124864": "Pleistocene, Pleistocene epoch, Glacial epoch",
  "n15125097": "Tertiary, Tertiary period",
  "n15125323": "Pliocene, Pliocene epoch",
  "n15125519": "Miocene, Miocene epoch",
  "n15125679": "Oligocene, Oligocene epoch",
  "n15125845": "Eocene, Eocene epoch",
  "n15126000": "Paleocene, Paleocene epoch",
  "n15126175": "Mesozoic, Mesozoic era, Age of Reptiles",
  "n15126361": "Cretaceous, Cretaceous period",
  "n15126595": "Jurassic, Jurassic period",
  "n15126750": "Triassic, Triassic period",
  "n15126931": "Paleozoic, Paleozoic era",
  "n15127165": "Permian, Permian period",
  "n15127307": "Carboniferous, Carboniferous period",
  "n15127507": "Pennsylvanian, Pennsylvanian period, Upper Carboniferous, Upper Carboniferous period",
  "n15127729": "Mississippian, Missippian period, Lower Carboniferous, Lower Carboniferous period",
  "n15127982": "Devonian, Devonian period, Age of Fishes",
  "n15128200": "Silurian, Silurian period",
  "n15128363": "Ordovician, Ordovician period",
  "n15128549": "Cambrian, Cambrian period",
  "n15128711": "Precambrian, Precambrian eon, Precambrian aeon, Precambrian period",
  "n15128997": "Proterozoic, Proterozoic eon, Proterozoic aeon",
  "n15129220": "Archean, Archean eon, Archean aeon, Archeozoic, Archaeozoic, Archeozoic eon, Archaeozoic aeon",
  "n15129572": "Hadean, Hadean time, Hadean eon, Hadean aeon, Priscoan, Priscoan eon, Priscoan aeon",
  "n15129927": "clock time, time",
  "n15130205": "Greenwich Mean Time, Greenwich Time, GMT, universal time, UT, UT1",
  "n15130434": "coordinated universal time, UTC",
  "n15130612": "Earth-received time, ERT",
  "n15130749": "one-way light time, OWLT",
  "n15130926": "round-trip light time, RTLT",
  "n15131123": "elapsed time",
  "n15131269": "transmission time, TRM",
  "n15131448": "spacecraft event time, SCET",
  "n15131598": "spacecraft clock time, SCLK",
  "n15131734": "Atlantic Time, Atlantic Standard Time",
  "n15131994": "Eastern Time, Eastern Standard Time, EST",
  "n15132201": "Central Time, Central Standard Time, CST",
  "n15132408": "Mountain Time, Mountain Standard Time, MST",
  "n15132638": "Pacific Time, Pacific Standard Time, PST",
  "n15132865": "Alaska Standard Time, Yukon Time",
  "n15133069": "Hawaii Time, Hawaii Standard Time",
  "n15133289": "Bering Time, Bering Standard Time",
  "n15133488": "duration, continuance",
  "n15133621": "duration, continuance",
  "n15133903": "clocking",
  "n15134054": "longueur",
  "n15134233": "residence time",
  "n15134340": "span",
  "n15134490": "stretch, stint",
  "n15134691": "time scale",
  "n15134913": "value, time value, note value",
  "n15135057": "extended time scale, slow time scale",
  "n15135258": "fast time scale",
  "n15135434": "time being, nonce",
  "n15135536": "biological clock",
  "n15135700": "circadian rhythm",
  "n15135822": "fourth dimension, time",
  "n15135996": "workweek, week",
  "n15136147": "week, calendar week",
  "n15136342": "midweek",
  "n15136453": "day",
  "n15136723": "workday, working day",
  "n15136923": "workday, working day, work day",
  "n15137047": "rest day, day of rest",
  "n15137194": "overtime",
  "n15137295": "turnaround, turnaround time",
  "n15137425": "spare time, free time",
  "n15137556": "day off",
  "n15137676": "leisure, leisure time",
  "n15137890": "vacation, holiday",
  "n15138241": "half-term",
  "n15138401": "vac",
  "n15138496": "half-holiday",
  "n15138600": "playtime, playday",
  "n15138691": "field day, outing, picnic",
  "n15138809": "field day",
  "n15138903": "honeymoon",
  "n15139018": "paid vacation",
  "n15139130": "leave, leave of absence",
  "n15139432": "furlough",
  "n15139552": "pass",
  "n15139691": "compassionate leave",
  "n15139849": "sabbatical, sabbatical leave",
  "n15139983": "sabbatical year",
  "n15140080": "shore leave, liberty",
  "n15140190": "sick leave",
  "n15140294": "terminal leave",
  "n15140405": "life, lifetime, life-time, lifespan",
  "n15140744": "life",
  "n15140892": "life",
  "n15141059": "days, years",
  "n15141213": "millennium, millenary",
  "n15141375": "bimillennium, bimillenary",
  "n15141486": "occupation",
  "n15141693": "past",
  "n15141894": "shelf life",
  "n15142025": "life expectancy",
  "n15142167": "birth",
  "n15142452": "cradle",
  "n15142568": "puerperium",
  "n15142836": "lactation",
  "n15143012": "incipiency, incipience",
  "n15143276": "death, last",
  "n15143477": "death, dying, demise",
  "n15143726": "grave",
  "n15143864": "afterlife, hereafter",
  "n15143984": "kingdom come",
  "n15144094": "immortality",
  "n15144178": "period",
  "n15144371": "time of life",
  "n15145012": "summer",
  "n15145171": "age, eld",
  "n15145499": "neonatal period",
  "n15145586": "infancy, babyhood, early childhood",
  "n15145782": "anal stage, anal phase",
  "n15146004": "genital stage, genital phase",
  "n15146260": "latency stage, latency phase, latency period",
  "n15146545": "oral stage, oral phase",
  "n15146828": "phallic stage, phallic phase",
  "n15147097": "childhood",
  "n15147330": "girlhood, maidenhood, maidhood",
  "n15147504": "boyhood",
  "n15147598": "schooldays, schooltime",
  "n15147713": "youth",
  "n15147850": "adolescence",
  "n15148081": "prepuberty",
  "n15148295": "puberty, pubescence",
  "n15148467": "teens",
  "n15148584": "twenties, mid-twenties",
  "n15148707": "1900s",
  "n15148787": "1530s",
  "n15148867": "twenties, 1920s",
  "n15148958": "1820s",
  "n15149038": "thirties, mid-thirties, thirty-something",
  "n15149180": "thirties, 1930s",
  "n15149271": "1830s",
  "n15149351": "forties, mid-forties",
  "n15149472": "forties, 1940s",
  "n15149562": "1840s",
  "n15149642": "fifties, mid-fifties",
  "n15149763": "fifties, 1950s",
  "n15149853": "1850s",
  "n15149933": "1750s",
  "n15150013": "sixties, mid-sixties",
  "n15150134": "sixties, 1960s",
  "n15150224": "1860s",
  "n15150304": "1760s",
  "n15150384": "golden years",
  "n15150493": "seventies, mid-seventies",
  "n15150618": "seventies, 1970s",
  "n15150710": "1870s",
  "n15150790": "1770s",
  "n15150870": "eighties, mid-eighties",
  "n15150993": "eighties, 1980s",
  "n15151084": "eighties, 1880s",
  "n15151175": "1780s",
  "n15151255": "nineties, mid-nineties",
  "n15151379": "nineties, 1990s",
  "n15151470": "nineties, 1890s",
  "n15151561": "1790s",
  "n15151641": "bloom, bloom of youth, salad days",
  "n15151764": "age of consent",
  "n15152062": "majority, legal age",
  "n15152261": "minority, nonage",
  "n15152409": "prime, prime of life",
  "n15152531": "drinking age",
  "n15152674": "voting age",
  "n15152817": "adulthood, maturity",
  "n15153225": "maturity, maturity date, due date",
  "n15153353": "bachelorhood",
  "n15153472": "middle age",
  "n15153667": "widowhood",
  "n15153787": "old age, years, age, eld, geezerhood",
  "n15154190": "dotage, second childhood, senility",
  "n15154376": "deathbed",
  "n15154462": "menopause, climacteric, change of life",
  "n15154646": "climacteric",
  "n15154774": "time unit, unit of time",
  "n15155220": "day, twenty-four hours, twenty-four hour period, 24-hour interval, solar day, mean solar day",
  "n15155747": "night",
  "n15155891": "tomorrow",
  "n15156001": "today",
  "n15156187": "yesterday",
  "n15156311": "morrow",
  "n15156424": "eve",
  "n15156537": "mean time, mean solar time",
  "n15156746": "terrestrial time, TT, terrestrial dynamical time, TDT, ephemeris time",
  "n15157041": "calendar day, civil day",
  "n15157225": "day",
  "n15158189": "Admission Day",
  "n15158359": "Arbor Day",
  "n15158450": "Cinco de Mayo",
  "n15158691": "commencement day, degree day",
  "n15158816": "November 5",
  "n15158997": "Guy Fawkes Day",
  "n15159139": "Bonfire Night, Guy Fawkes Night",
  "n15159265": "Inauguration Day, January 20",
  "n15159426": "leap day, bissextile day, February 29",
  "n15159583": "date, day of the month",
  "n15159819": "date",
  "n15160076": "future date",
  "n15160216": "rain date",
  "n15160418": "sell-by date",
  "n15160579": "date",
  "n15160866": "quarter day",
  "n15161165": "fast day",
  "n15161284": "major fast day",
  "n15161430": "minor fast day",
  "n15161631": "feast day, fete day",
  "n15161872": "Succoth, Sukkoth, Succos, Feast of Booths, Feast of Tabernacles, Tabernacles",
  "n15162210": "religious festival, church festival",
  "n15162388": "festival",
  "n15162523": "D-day, 6 June 1944",
  "n15162640": "V-day, Victory Day",
  "n15162764": "V-E Day, 8 May 1945",
  "n15162882": "V-J Day, 15 August 1945",
  "n15163005": "day of the week",
  "n15163157": "weekday",
  "n15163408": "feria",
  "n15163797": "Sunday, Lord's Day, Dominicus, Sun",
  "n15163979": "Monday, Mon",
  "n15164105": "Tuesday, Tues",
  "n15164233": "Wednesday, Midweek, Wed",
  "n15164354": "Thursday, Th",
  "n15164463": "Friday, Fri",
  "n15164570": "Saturday, Sabbatum, Sat",
  "n15164750": "Sabbath",
  "n15164957": "day, daytime, daylight",
  "n15165289": "morning, morn, morning time, forenoon",
  "n15165490": "noon, twelve noon, high noon, midday, noonday, noontide",
  "n15165637": "mealtime",
  "n15165805": "breakfast time",
  "n15165917": "lunchtime, lunch period",
  "n15166070": "dinnertime, suppertime",
  "n15166191": "afternoon",
  "n15166352": "midafternoon",
  "n15166462": "evening, eve, even, eventide",
  "n15166742": "guest night",
  "n15166897": "prime time",
  "n15167027": "night, nighttime, dark",
  "n15167349": "night",
  "n15167474": "night",
  "n15167577": "night",
  "n15167675": "weeknight",
  "n15167778": "eve",
  "n15167906": "evening",
  "n15168080": "late-night hour",
  "n15168185": "midnight",
  "n15168369": "small hours",
  "n15168475": "bedtime",
  "n15168569": "lights-out",
  "n15168665": "closing time",
  "n15168790": "dawn, dawning, morning, aurora, first light, daybreak, break of day, break of the day, dayspring, sunrise, sunup, cockcrow",
  "n15169136": "early-morning hour",
  "n15169248": "sunset, sundown",
  "n15169421": "twilight, dusk, gloaming, gloam, nightfall, evenfall, fall, crepuscule, crepuscle",
  "n15169759": "night",
  "n15169873": "week, hebdomad",
  "n15170178": "week from Monday",
  "n15170331": "fortnight, two weeks",
  "n15170504": "weekend",
  "n15170786": "rag, rag week",
  "n15171008": "rag day",
  "n15171145": "red-letter day",
  "n15171307": "Judgment Day, Judgement Day, Day of Judgment, Day of Judgement, Doomsday, Last Judgment, Last Judgement, Last Day, eschaton, day of reckoning, doomsday, crack of doom, end of the world",
  "n15171738": "off-day",
  "n15171857": "access time",
  "n15172212": "distance, space",
  "n15172423": "distance",
  "n15172664": "embolism, intercalation",
  "n15172782": "payday",
  "n15172881": "polling day, election day",
  "n15173064": "church year, Christian year",
  "n15173259": "field day",
  "n15173353": "field day",
  "n15173479": "calendar",
  "n15173712": "timekeeping",
  "n15173830": "Roman calendar",
  "n15173974": "ides",
  "n15174122": "market day",
  "n15174218": "Gregorian calendar, New Style calendar",
  "n15174885": "Julian calendar, Old Style calendar",
  "n15175202": "Revolutionary calendar",
  "n15175640": "Revolutionary calendar month",
  "n15175987": "Vendemiaire",
  "n15176161": "Brumaire",
  "n15176319": "Frimaire",
  "n15176476": "Nivose",
  "n15176630": "Pluviose",
  "n15176785": "Ventose",
  "n15176937": "Germinal",
  "n15177091": "Floreal",
  "n15177244": "Prairial",
  "n15177396": "Messidor",
  "n15177549": "Thermidor",
  "n15177705": "Fructidor",
  "n15177866": "Jewish calendar, Hebrew calendar",
  "n15178417": "lunar calendar",
  "n15178546": "lunisolar calendar",
  "n15178694": "solar calendar",
  "n15178841": "Islamic calendar, Muhammadan calendar, Mohammedan calendar, Moslem calendar, Muslim calendar",
  "n15179415": "Hindu calendar",
  "n15179888": "date, particular date",
  "n15180082": "deadline",
  "n15180209": "curfew",
  "n15180304": "anachronism, mistiming, misdating",
  "n15180528": "point, point in time",
  "n15180934": "arrival time, time of arrival",
  "n15181094": "departure time, time of departure",
  "n15181282": "checkout, checkout time",
  "n15181444": "Holy Week, Passion Week",
  "n15181556": "Holy Year",
  "n15181718": "church calendar, ecclesiastical calendar",
  "n15181977": "Walpurgis Night",
  "n15182053": "New Year's Eve, December 31",
  "n15182189": "New Year's Day, New Year's, January 1",
  "n15182319": "New Year",
  "n15182402": "Martin Luther King Jr's Birthday, Martin Luther King Day",
  "n15182569": "Robert E Lee's Birthday, Robert E Lee Day, Lee's Birthday, January 19",
  "n15182724": "Hogmanay",
  "n15182805": "Rosh Hashanah, Rosh Hashana, Rosh Hashonah, Rosh Hashona, Jewish New Year",
  "n15183085": "Rosh Hodesh, Rosh Chodesh",
  "n15183266": "Tet",
  "n15183428": "holiday",
  "n15183802": "religious holiday, holy day",
  "n15184008": "High Holy Day, High Holiday",
  "n15184170": "Christian holy day",
  "n15184755": "Jewish holy day",
  "n15185007": "holy day of obligation",
  "n15185290": "movable feast, moveable feast",
  "n15185471": "Yom Kippur, Day of Atonement",
  "n15185721": "Saint Agnes's Eve, January 20",
  "n15185837": "Martinmas, St Martin's Day, 11 November",
  "n15185996": "Indian summer, Saint Martin's summer",
  "n15186147": "Annunciation, Lady Day, Annunciation Day, March 25",
  "n15186412": "Michaelmas, Michaelmas Day, September 29",
  "n15186596": "Michaelmastide",
  "n15186681": "Candlemas, Candlemas Day, Feb 2",
  "n15186871": "Groundhog Day, February 2",
  "n15187077": "Lincoln's Birthday, February 12",
  "n15187250": "Valentine Day, Valentine's Day, Saint Valentine's Day, St Valentine's Day, February 14",
  "n15187451": "Washington's Birthday, February 22",
  "n15187619": "Presidents' Day",
  "n15187800": "Texas Independence Day, March 2",
  "n15187988": "St Patrick's Day, Saint Patrick's Day, March 17",
  "n15188154": "Easter",
  "n15188359": "Easter Sunday, Easter Day",
  "n15188505": "April Fools', April Fools' day, All Fools' day",
  "n15188688": "Pan American Day, April 14",
  "n15188852": "Patriot's Day",
  "n15189033": "May Day, First of May, May 1",
  "n15189249": "Mother's Day",
  "n15189347": "Armed Forces Day",
  "n15189452": "Memorial Day, Decoration Day",
  "n15189684": "Jefferson Davis' Birthday, Davis' Birthday, June 3",
  "n15189838": "Flag Day, June 14",
  "n15189982": "Father's Day",
  "n15190084": "Independence Day, Fourth of July, July 4",
  "n15190228": "Lammas, Lammas Day, August 1",
  "n15190424": "Lammastide",
  "n15190520": "Labor Day",
  "n15190652": "Citizenship Day, September 17",
  "n15190779": "American Indian Day",
  "n15190895": "Columbus Day, Discovery Day, October 12",
  "n15191080": "United Nations Day, October 24",
  "n15191233": "Halloween, Hallowe'en, Allhallows Eve",
  "n15191401": "Pasch, Pascha",
  "n15191494": "Pasch, Pascha",
  "n15191587": "Eastertide",
  "n15191661": "Palm Sunday",
  "n15191739": "Passion Sunday",
  "n15191827": "Good Friday",
  "n15191924": "Low Sunday",
  "n15192008": "Holy Saturday",
  "n15192116": "Holy Innocents' Day, Innocents' Day",
  "n15192272": "Septuagesima, Septuagesima Sunday",
  "n15192406": "Quinquagesima, Quinquagesima Sunday",
  "n15192544": "Quadragesima, Quadrigesima Sunday",
  "n15192649": "Trinity Sunday",
  "n15192736": "Rogation Day",
  "n15192890": "Solemnity of Mary, January 1",
  "n15193052": "Ascension, Ascension Day, Ascension of the Lord",
  "n15193271": "Circumcision, Feast of the Circumcision, January 1",
  "n15193526": "Maundy Thursday, Holy Thursday",
  "n15193660": "Corpus Christi",
  "n15193776": "Saints Peter and Paul, June 29",
  "n15193908": "Assumption, Assumption of Mary, August 15",
  "n15194194": "Dormition, Feast of Dormition",
  "n15194506": "Epiphany, Epiphany of Our Lord, Twelfth day, Three Kings' Day, January 6",
  "n15194739": "Saint Joseph, St Joseph, March 19",
  "n15194860": "Twelfthtide",
  "n15194940": "Twelfth night",
  "n15195059": "All Saints' Day, Allhallows, November 1, Hallowmas, Hallowmass",
  "n15195259": "Immaculate Conception, December 8",
  "n15195388": "Allhallowtide",
  "n15195477": "All Souls' Day, November 2",
  "n15195622": "Ash Wednesday",
  "n15195834": "Ember Day",
  "n15195928": "Passover, Pesach, Pesah, Feast of the Unleavened Bread",
  "n15196186": "Christmas, Christmas Day, Xmas, Dec 25",
  "n15196444": "Christmas Eve, Dec 24",
  "n15196537": "Christmas, Christmastide, Christmastime, Yule, Yuletide, Noel",
  "n15196746": "Boxing Day",
  "n15196870": "Purim",
  "n15197042": "Shavous, Shabuoth, Shavuoth, Shavuot, Pentecost, Feast of Weeks",
  "n15197302": "Shimchath Torah, Simchat Torah, Simhath Torah, Simhat Torah, Simchas Torah, Rejoicing over the Law, Rejoicing of the Law, Rejoicing in the Law",
  "n15197658": "Tishah b'Av, Tishah b'Ab, Tisha b'Av, Tisha b'Ab, Ninth of Av, Ninth of Ab, Fast of Av, Fast of Ab",
  "n15197954": "Fast of Gedaliah",
  "n15198136": "Fast of Tevet",
  "n15198401": "Fast of Esther",
  "n15198662": "Fast of the Firstborn",
  "n15198872": "Fast of Tammuz",
  "n15199033": "Hanukkah, Hanukah, Hannukah, Chanukah, Chanukkah, Channukah, Channukkah, Festival of Lights, Feast of Lights, Feast of Dedication, Feast of the Dedication",
  "n15199406": "Lag b'Omer",
  "n15199592": "legal holiday, national holiday, public holiday",
  "n15200032": "bank holiday",
  "n15200164": "Commonwealth Day, Empire day, May 24",
  "n15200314": "Dominion Day, July 1",
  "n15200493": "Bastille Day, 14 July",
  "n15200661": "Remembrance Day, Remembrance Sunday, Poppy Day",
  "n15200896": "Veterans Day, Veterans' Day, Armistice Day, November 11",
  "n15201116": "Thanksgiving, Thanksgiving Day",
  "n15201384": "Victoria Day",
  "n15201505": "year",
  "n15201822": "anomalistic year",
  "n15201994": "year-end",
  "n15202131": "common year, 365 days",
  "n15202230": "leap year, intercalary year, 366 days, bissextile year",
  "n15202424": "off year",
  "n15202532": "off year",
  "n15202634": "calendar year, civil year",
  "n15202806": "solar year, tropical year, astronomical year, equinoctial year",
  "n15203017": "lunar year",
  "n15203120": "fiscal year, financial year",
  "n15203229": "school, schooltime, school day",
  "n15203565": "school year, academic year",
  "n15203791": "year, twelvemonth, yr",
  "n15204201": "annum",
  "n15204297": "year",
  "n15204485": "semester",
  "n15204609": "bimester",
  "n15204720": "Olympiad",
  "n15204907": "lustrum",
  "n15204983": "decade, decennary, decennium",
  "n15205532": "century",
  "n15205719": "quadrennium",
  "n15205799": "quinquennium",
  "n15205880": "quattrocento",
  "n15206004": "twentieth century",
  "n15206097": "half-century",
  "n15206195": "quarter-century",
  "n15206296": "month",
  "n15206590": "quarter",
  "n15206744": "phase of the moon",
  "n15206943": "new moon, new phase of the moon",
  "n15207103": "half-moon",
  "n15207302": "first quarter",
  "n15207430": "last quarter",
  "n15207556": "full moon, full-of-the-moon, full phase of the moon, full",
  "n15207770": "harvest moon",
  "n15207872": "lunar month, moon, lunation, synodic month",
  "n15208036": "anomalistic month",
  "n15208162": "sidereal time",
  "n15208333": "sidereal day, day",
  "n15208540": "day",
  "n15208740": "lunar day",
  "n15208898": "sidereal year",
  "n15209074": "sidereal hour",
  "n15209174": "sidereal month",
  "n15209316": "solar month",
  "n15209413": "calendar month, month",
  "n15209706": "Gregorian calendar month",
  "n15210045": "January, Jan",
  "n15210383": "mid-January",
  "n15210486": "February, Feb",
  "n15210765": "mid-February",
  "n15210870": "March, Mar",
  "n15211090": "mid-March",
  "n15211189": "April, Apr",
  "n15211385": "mid-April",
  "n15211484": "May",
  "n15211711": "mid-May",
  "n15211806": "June",
  "n15212070": "mid-June",
  "n15212167": "July",
  "n15212358": "mid-July",
  "n15212455": "August, Aug",
  "n15212638": "mid-August",
  "n15212739": "September, Sep, Sept",
  "n15213008": "mid-September",
  "n15213115": "October, Oct",
  "n15213303": "mid-October",
  "n15213406": "November, Nov",
  "n15213669": "mid-November",
  "n15213774": "December, Dec",
  "n15213963": "mid-December",
  "n15214068": "Jewish calendar month",
  "n15214419": "Tishri",
  "n15214639": "Heshvan",
  "n15214840": "Kislev, Chislev",
  "n15215068": "Tebet, Tevet",
  "n15215270": "Shebat, Shevat",
  "n15215480": "Adar",
  "n15215693": "Veadar, Adar Sheni",
  "n15215816": "Nisan, Nissan",
  "n15215994": "Iyar, Iyyar",
  "n15216189": "Sivan, Siwan",
  "n15216386": "Tammuz, Thammuz",
  "n15216563": "Ab, Av",
  "n15216760": "Elul, Ellul",
  "n15216966": "Islamic calendar month",
  "n15217308": "Muharram, Moharram, Muharrum",
  "n15217443": "Safar, Saphar",
  "n15217563": "Rabi I",
  "n15217674": "Rabi II",
  "n15217787": "Jumada I, Jomada I",
  "n15217911": "Jumada II, Jomada II",
  "n15218037": "Rajab",
  "n15218149": "Sha'ban, Shaaban",
  "n15218272": "Ramadan",
  "n15218448": "Id al-Fitr",
  "n15218551": "Shawwal",
  "n15218663": "Dhu'l-Qa'dah, Dhu al-Qadah",
  "n15218798": "Dhu'l-Hijja, Dhu'l-Hijjah, Dhu al-Hijja, Dhu al-Hijjah",
  "n15219022": "Id al-Adha, Feast of Sacrifice",
  "n15219351": "Hindu calendar month",
  "n15219694": "Chait, Caitra",
  "n15219874": "Ramanavami",
  "n15220027": "Baisakh, Vaisakha",
  "n15220149": "Jeth, Jyaistha",
  "n15220267": "Asarh, Asadha",
  "n15220385": "Sawan, Sravana",
  "n15220503": "Bhadon, Bhadrapada",
  "n15220625": "Asin, Asvina",
  "n15220743": "Kartik, Karttika",
  "n15220864": "Aghan, Margasivsa",
  "n15220985": "Pus, Pansa",
  "n15221099": "Magh, Magha",
  "n15221286": "Mesasamkranti",
  "n15221474": "Phagun, Phalguna",
  "n15221596": "saint's day",
  "n15221718": "name day",
  "n15221818": "solstice",
  "n15222012": "summer solstice, June 21, midsummer",
  "n15222202": "Midsummer Day, Midsummer's Day, St John's Day, June 24",
  "n15222369": "Midsummer Eve, Midsummer Night, St John's Eve, St John's Night, June 23",
  "n15222540": "school day",
  "n15222686": "speech day",
  "n15222840": "washday, washing day",
  "n15222951": "wedding day",
  "n15223048": "wedding night",
  "n15223190": "winter solstice",
  "n15223343": "equinox",
  "n15223574": "vernal equinox, March equinox, spring equinox",
  "n15223750": "autumnal equinox, September equinox, fall equinox",
  "n15223916": "Noruz, Nowruz, Nowrooz",
  "n15224156": "time limit",
  "n15224293": "limitation",
  "n15224486": "term",
  "n15224692": "prison term, sentence, time",
  "n15224978": "hard time",
  "n15225076": "life sentence, life",
  "n15225249": "school term, academic term, academic session, session",
  "n15225526": "summer school",
  "n15225668": "midterm",
  "n15225797": "semester",
  "n15225929": "trimester",
  "n15226046": "quarter",
  "n15226214": "gestation, gestation period",
  "n15226451": "term, full term",
  "n15226625": "midterm",
  "n15226732": "trimester",
  "n15226972": "first trimester",
  "n15227133": "second trimester",
  "n15227261": "third trimester",
  "n15227391": "refractory period",
  "n15227593": "bell, ship's bell",
  "n15227846": "hour, hr, 60 minutes",
  "n15228162": "half-hour, 30 minutes",
  "n15228267": "quarter-hour, 15 minutes",
  "n15228378": "hour, time of day",
  "n15228787": "none",
  "n15228910": "hour",
  "n15229019": "happy hour",
  "n15229144": "rush hour",
  "n15229300": "zero hour",
  "n15229408": "canonical hour",
  "n15229677": "matins, morning prayer",
  "n15229784": "prime",
  "n15229875": "terce, tierce",
  "n15229974": "sext",
  "n15230076": "nones",
  "n15230180": "vespers, evensong",
  "n15230363": "compline, complin",
  "n15230482": "man hour, person hour",
  "n15230617": "silly season",
  "n15230790": "Golden Age",
  "n15231031": "silver age",
  "n15231263": "bronze age",
  "n15231415": "Bronze Age",
  "n15231634": "iron age",
  "n15231765": "Iron Age",
  "n15231964": "Stone Age",
  "n15232236": "Eolithic Age, Eolithic",
  "n15232406": "Paleolithic Age, Paleolithic, Palaeolithic",
  "n15232712": "Lower Paleolithic",
  "n15232899": "Middle Paleolithic",
  "n15233047": "Upper Paleolithic",
  "n15233239": "Mesolithic Age, Mesolithic, Epipaleolithic",
  "n15233411": "Neolithic Age, Neolithic, New Stone Age",
  "n15233614": "great year, Platonic year",
  "n15233778": "regulation time",
  "n15233989": "overtime, extra time",
  "n15234212": "extra innings",
  "n15234348": "overtime period",
  "n15234469": "tiebreaker",
  "n15234587": "sudden death",
  "n15234764": "minute, min",
  "n15234942": "quarter",
  "n15235126": "second, sec, s",
  "n15235334": "leap second",
  "n15235540": "attosecond",
  "n15235687": "femtosecond",
  "n15235853": "picosecond",
  "n15236015": "nanosecond",
  "n15236176": "microsecond",
  "n15236338": "millisecond, msec",
  "n15236475": "season, time of year",
  "n15236859": "fall, autumn",
  "n15237044": "spring, springtime",
  "n15237250": "summer, summertime",
  "n15237567": "dog days, canicule, canicular days",
  "n15237782": "winter, wintertime",
  "n15238074": "midwinter",
  "n15238169": "growing season",
  "n15238271": "seedtime",
  "n15238371": "sheepshearing",
  "n15238472": "holiday season",
  "n15238570": "high season, peak season",
  "n15238761": "off-season",
  "n15238895": "rainy season",
  "n15239033": "monsoon",
  "n15239174": "dry season",
  "n15239292": "season",
  "n15239579": "season",
  "n15240119": "preseason",
  "n15240291": "spring training",
  "n15240389": "baseball season",
  "n15240503": "triple-crown season",
  "n15240633": "basketball season",
  "n15240733": "exhibition season",
  "n15240888": "fishing season",
  "n15240998": "football season",
  "n15241094": "hockey season",
  "n15241186": "hunting season",
  "n15241311": "social season",
  "n15241405": "theatrical season",
  "n15241507": "Advent",
  "n15241655": "Advent Sunday",
  "n15241777": "Shrovetide",
  "n15241898": "Mardi Gras, Shrove Tuesday, pancake day",
  "n15242029": "Lent, Lententide",
  "n15242209": "Pentecost, Whitsunday",
  "n15242432": "Whitmonday, Whitsun Monday",
  "n15242599": "Whit-Tuesday, Whitsun Tuesday",
  "n15242719": "Whitsun, Whitsuntide, Whitweek",
  "n15242955": "long time, age, years",
  "n15243202": "month of Sundays",
  "n15243351": "long run, long haul",
  "n15243590": "eon, aeon",
  "n15243730": "eon, aeon",
  "n15243976": "eternity, infinity",
  "n15244094": "alpha and omega",
  "n15244200": "blue moon",
  "n15244351": "year dot",
  "n15244505": "drought, drouth",
  "n15244650": "moment, minute, second, instant",
  "n15244942": "eleventh hour, last minute",
  "n15245144": "moment of truth",
  "n15245244": "moment of truth",
  "n15245382": "pinpoint",
  "n15245515": "time",
  "n15245711": "high time",
  "n15245829": "occasion",
  "n15245990": "meal",
  "n15246135": "psychological moment",
  "n15246258": "wee",
  "n15246353": "while, piece, spell, patch",
  "n15246683": "cold spell, cold snap",
  "n15246775": "hot spell",
  "n15246853": "moment, mo, minute, second, bit",
  "n15247110": "blink of an eye, flash, heartbeat, instant, jiffy, split second, trice, twinkling, wink, New York minute",
  "n15247410": "ephemera",
  "n15247518": "period, geological period",
  "n15248020": "era, geological era",
  "n15248269": "epoch",
  "n15248564": "era, epoch",
  "n15248812": "Caliphate",
  "n15249096": "Christian era, Common era",
  "n15249236": "day",
  "n15249458": "year of grace",
  "n15249547": "Y2K",
  "n15249636": "generation",
  "n15249799": "anniversary, day of remembrance",
  "n15250178": "birthday",
  "n15250312": "jubilee",
  "n15250468": "diamond jubilee",
  "n15250580": "silver jubilee",
  "n15250691": "wedding anniversary",
  "n15250890": "silver wedding anniversary",
  "n15250991": "golden wedding anniversary",
  "n15251092": "diamond wedding anniversary, diamond wedding",
  "n15251212": "semicentennial, semicentenary",
  "n15251336": "centennial, centenary",
  "n15251489": "sesquicentennial",
  "n15251600": "bicentennial, bicentenary",
  "n15251757": "tercentennial, tercentenary, triennial",
  "n15251892": "quatercentennial, quatercentenary",
  "n15252021": "quincentennial, quincentenary",
  "n15252146": "millennium, millenary",
  "n15252300": "bimillennium, bimillenary",
  "n15252422": "birthday, natal day",
  "n15252524": "time immemorial, time out of mind",
  "n15252635": "auld langsyne, langsyne, old times, good old days",
  "n15252770": "by-and-by",
  "n15252907": "chapter",
  "n15253139": "antiquity",
  "n15253895": "golden age",
  "n15254028": "historic period, age",
  "n15254550": "prehistory, prehistoric culture",
  "n15254831": "modern era",
  "n15254933": "information age",
  "n15255195": "ice age, glacial period, glacial epoch",
  "n15255439": "Jazz Age",
  "n15255641": "chukker, chukka",
  "n15255804": "inning, frame",
  "n15256022": "top, top of the inning",
  "n15256245": "bottom, bottom of the inning",
  "n15256417": "set",
  "n15256567": "game",
  "n15256714": "turn, bout, round",
  "n15256915": "playing period, period of play, play",
  "n15257416": "first period",
  "n15257553": "second period",
  "n15257692": "final period",
  "n15257829": "half",
  "n15258091": "first half",
  "n15258179": "second half, last half",
  "n15258281": "period",
  "n15258450": "quarter",
  "n15258694": "over",
  "n15258945": "maiden over, maiden",
  "n15259076": "Baroque, Baroque era, Baroque period",
  "n15259284": "Middle Ages, Dark Ages",
  "n15259812": "Renaissance, Renascence",
  "n15260130": "Italian Renaissance",
  "n15260277": "Industrial Revolution, technological revolution",
  "n15260436": "Reign of Terror",
  "n15260651": "reign of terror",
  "n15260814": "reign",
  "n15260964": "reign",
  "n15261169": "turn of the century",
  "n15261306": "Harlem Renaissance",
  "n15261471": "New Deal",
  "n15261658": "Reconstruction, Reconstruction Period",
  "n15261868": "Restoration",
  "n15261972": "print run, press run",
  "n15262120": "run",
  "n15262360": "run-time",
  "n15262510": "run-time",
  "n15262624": "split run",
  "n15262794": "space age",
  "n15262921": "today",
  "n15263045": "tonight",
  "n15263138": "yesterday",
  "n15263283": "millennium",
  "n15263675": "offing",
  "n15263795": "tomorrow",
  "n15263925": "manana",
  "n15264010": "common time, four-four time, quadruple time, common measure",
  "n15264168": "duple time",
  "n15264264": "triple time",
  "n15264363": "tempo, pacing",
  "n15264607": "in time",
  "n15264726": "accelerando",
  "n15264891": "allegretto",
  "n15265021": "allegro",
  "n15265135": "allegro con spirito",
  "n15265219": "andante",
  "n15265331": "meno mosso",
  "n15265423": "rubato",
  "n15265518": "beginning, commencement, first, outset, get-go, start, kickoff, starting time, showtime, offset",
  "n15266034": "youth, early days",
  "n15266164": "terminus a quo, starting point",
  "n15266265": "presidency, presidential term, administration",
  "n15266568": "vice-presidency, vice-presidential term",
  "n15266685": "middle",
  "n15266911": "end, ending",
  "n15267373": "deep",
  "n15267536": "stopping point, finale, finis, finish, last, conclusion, close",
  "n15267945": "dawn",
  "n15268094": "evening",
  "n15268239": "cease",
  "n15268367": "fag end, tail, tail end",
  "n15268547": "last gasp",
  "n15268682": "termination, expiration, expiry",
  "n15268857": "terminus ad quem, terminal point, limit",
  "n15268993": "threshold",
  "n15269128": "seek time",
  "n15269331": "track-to-track seek time",
  "n15269513": "time interval, interval",
  "n15269996": "time constant",
  "n15270245": "time slot, slot",
  "n15270431": "time",
  "n15270697": "lunitidal interval",
  "n15270862": "absence",
  "n15271008": "pause, intermission, break, interruption, suspension",
  "n15271417": "lapse",
  "n15271619": "blackout",
  "n15271732": "caesura",
  "n15271901": "dead air",
  "n15272029": "delay, hold, time lag, postponement, wait",
  "n15272382": "extension",
  "n15272571": "halftime",
  "n15272685": "interlude",
  "n15272791": "entr'acte",
  "n15272887": "interim, meantime, meanwhile, lag",
  "n15273101": "latent period",
  "n15273241": "reaction time, response time, latency, latent period",
  "n15273406": "eternity",
  "n15273522": "interregnum",
  "n15273626": "sleep, nap",
  "n15273875": "beauty sleep",
  "n15273955": "kip",
  "n15274074": "respite, rest, relief, rest period",
  "n15274305": "time-out",
  "n15274441": "letup, lull",
  "n15274695": "breath, breather, breathing place, breathing space, breathing spell, breathing time",
  "n15274863": "lease, term of a contract",
  "n15275094": "half life, half-life",
  "n15275315": "relaxation time",
  "n15275466": "moratorium",
  "n15275598": "retardation",
  "n15275727": "tide, lunar time period",
  "n15275851": "acceleration",
  "n15276023": "centripetal acceleration",
  "n15276171": "deceleration",
  "n15276307": "attrition rate, rate of attrition",
  "n15276427": "birthrate, birth rate, fertility, fertility rate, natality",
  "n15276642": "bits per second, bps",
  "n15276801": "crime rate",
  "n15276959": "data rate",
  "n15277118": "deathrate, death rate, mortality, mortality rate, fatality rate",
  "n15277358": "dose rate",
  "n15277462": "erythrocyte sedimentation rate, ESR, sedimentation rate, sed rate",
  "n15277730": "flow, flow rate, rate of flow",
  "n15277926": "cardiac output",
  "n15278132": "flux",
  "n15278281": "frequency, frequence, oftenness",
  "n15278691": "gigahertz, GHz, gigacycle per second, gigacycle, Gc",
  "n15278825": "growth rate, rate of growth",
  "n15278960": "isometry",
  "n15279104": "hertz, Hz, cycle per second, cycles/second, cps, cycle",
  "n15279299": "inflation rate, rate of inflation",
  "n15279480": "jerk",
  "n15279596": "kilohertz, kHz, kilocycle per second, kilocycle, kc",
  "n15279767": "kilometers per hour, kilometres per hour, kph, km/h",
  "n15279957": "megahertz, MHz, megacycle per second, megacycle, Mc",
  "n15280108": "terahertz, THz",
  "n15280201": "metabolic rate",
  "n15280346": "miles per hour, mph",
  "n15280497": "pace, gait",
  "n15280695": "pulse, pulse rate, heart rate",
  "n15280964": "femoral pulse",
  "n15281071": "radial pulse",
  "n15281176": "rate of return",
  "n15281329": "return on invested capital, return on investment, ROI",
  "n15281653": "respiratory rate, rate of respiration",
  "n15281870": "revolutions per minute, rpm, rev",
  "n15282032": "sampling rate",
  "n15282188": "Nyquist rate",
  "n15282378": "solar constant",
  "n15282534": "spacing",
  "n15282696": "speed, velocity",
  "n15283097": "tempo, pace",
  "n15283224": "quick time",
  "n15283327": "double time",
  "n15283433": "airspeed",
  "n15283554": "escape velocity",
  "n15283675": "groundspeed",
  "n15283780": "hypervelocity",
  "n15283920": "muzzle velocity",
  "n15284043": "peculiar velocity",
  "n15284158": "radial velocity",
  "n15284285": "speed of light, light speed, c",
  "n15284553": "steerageway",
  "n15284704": "terminal velocity",
  "n15284878": "miles per hour, mph",
  "n15284999": "attendance",
  "n15285180": "count per minute, counts/minute",
  "n15285279": "sampling frequency",
  "n15285456": "Nyquist frequency",
  "n15285622": "infant deathrate, infant mortality, infant mortality rate",
  "n15285772": "neonatal mortality, neonatal mortality rate",
  "n15285910": "words per minute, wpm",
  "n15286042": "beats per minute, bpm, metronome marking, M.M.",
  "n15286249": "rate",
  "n15287073": "channel capacity",
  "n15287199": "neutron flux",
  "n15287351": "radiant flux",
  "n15287486": "luminous flux",
  "n15287578": "incubation",
  "n15287830": "cycle, rhythm, round",
  "n15288111": "menstrual cycle",
  "n15288489": "fertile period, fertile phase",
  "n15288707": "menstrual phase",
  "n15288943": "musth",
  "n15289208": "secretory phase, luteal phase",
  "n15289524": "lead time",
  "n15289779": "period",
  "n15289944": "orbit period",
  "n15290132": "phase, phase angle",
  "n15290337": "phase, stage",
  "n15290930": "generation",
  "n15291110": "multistage",
  "n15291199": "apogee, culmination",
  "n15291416": "seedtime",
  "n15291498": "tenure, term of office, incumbency",
  "n15291714": "episcopate",
  "n15291801": "shift, work shift, duty period",
  "n15292069": "go, spell, tour, turn",
  "n15292260": "trick",
  "n15292336": "watch",
  "n15292502": "watch",
  "n15292617": "dogwatch",
  "n15292722": "day shift",
  "n15292829": "evening shift, swing shift",
  "n15292960": "night shift, graveyard shift",
  "n15293091": "split shift",
  "n15293227": "peacetime",
  "n15293328": "wartime",
  "n15293435": "graveyard watch, middle watch, midwatch, night watch",
  "n15293590": "enlistment, hitch, term of enlistment, tour of duty, duty tour, tour",
  "n15293788": "honeymoon",
  "n15293931": "indiction",
  "n15294085": "float",
  "n15294211": "Depression, Great Depression",
  "n15294382": "prohibition, prohibition era",
  "n15294607": "incubation period",
  "n15294745": "rainy day",
  "n15294884": "novitiate, noviciate",
  "n15295045": "flower, prime, peak, heyday, bloom, blossom, efflorescence, flush",
  "n15295267": "golden age",
  "n15295416": "rule",
  "n15295603": "Regency",
  "n15295778": "running time",
  "n15295884": "show time",
  "n15296039": "safe period",
  "n15296258": "octave",
  "n15296354": "then",
  "n15296489": "shiva, shivah, shibah",
  "n15296687": "epoch, date of reference",
  "n15296920": "clotting time",
  "n15297069": "rotational latency, latency",
  "n15297303": "probation",
  "n15297472": "probation",
  "n15297672": "processing time",
  "n15297859": "air alert",
  "n15298011": "command processing overhead time, command processing overhead, command overhead, overhead",
  "n15298283": "Great Schism",
  "n15298507": "question time",
  "n15298695": "real time",
  "n15298852": "real time",
  "n15298995": "regency",
  "n15299097": "snap",
  "n15299225": "study hall",
  "n15299367": "Transfiguration, Transfiguration Day, August 6",
  "n15299585": "usance",
  "n15299783": "window",
  "n15300051": "9/11, 9-11, September 11, Sept. 11, Sep 11",
}